/*
* Copyright c                  Realtek Semiconductor Corporation, 2005  
* All rights reserved.
* 
* Program : Benchmarking time functions
* Abstract : 
* Author : Edward Jin-Ru Chen (jzchen@realtek.com.tw)               
* $Id: benchmarkTime.c,v 1.1.1.1 2006/07/26 05:38:21 blue Exp $
*/

#include "../rtl865x/asicRegs.h"
#include "../rtl865x/rtl_glue.h"
#include "../rtl865x/rtl8651_tblAsicDrv.h"
#include "benchmarkTime.h"

/* Local variables */
volatile uint64 bcp3C0, bcp3C1, bcp3C2, bcp3C3;
volatile uint32 cp3MonCtrl;
uint32 prevBenchTC1CNT, currBenchTC1CNT;

static uint32 backupWatchDogReg;

void benchCp3CounterGet( void )
{
	__asm__ \
	("	;\
		la		$8, bcp3C0;\
		mfc3	$9, $9			;\
		sw		$9, 0($8)		;\
		mfc3	$9, $8			;\
		sw		$9, 4($8)		;\
		la		$8, bcp3C1;\
		mfc3	$9, $11			;\
		sw		$9, 0($8)		;\
		mfc3	$9, $10			;\
		sw		$9, 4($8)		;\
		la		$8, bcp3C2;\
		mfc3	$9, $13			;\
		sw		$9, 0($8)		;\
		mfc3	$9, $12			;\
		sw		$9, 4($8)		;\
		la		$8, bcp3C3;\
		mfc3	$9, $15			;\
		sw		$9, 0($8)		;\
		mfc3	$9, $14			;\
		sw		$9, 4($8)		;\
	");
}

void benchCp3CounterResetAndStart(uint32 counterDefine) {

	cp3MonCtrl = counterDefine;

	__asm__ __volatile__\
	("  ;\
		mfc0	$8, $12			;\
		nop						;\
		la		$9, 0x80000000	;\
		or		$8, $9			;\
		mtc0	$8, $12			;\
		nop						;\
		nop						;\
		mtc3	$0, $8			;\
		mtc3	$0, $9			;\
		mtc3	$0, $10			;\
		mtc3	$0, $11			;\
		mtc3	$0, $12			;\
		mtc3	$0, $13			;\
		mtc3	$0, $14			;\
		mtc3	$0, $15			;\
		la		$8, cp3MonCtrl		;\
		lw		$9, 0($8)			;\
		ctc3 	$9, $0				;\
		nop						;\
	");
}

void benchCp3CounterStop(void) {
	__asm__ __volatile__\
	("	;\
		ctc3 	$0, $0			;\
		nop					;\
	");
}

void benchCp3CounterStopAndGet(uint64 * pcp3Cntr) {
	if(pcp3Cntr) {
		currBenchTC1CNT = READ_MEM32(TC1CNT);
		benchCp3CounterStop();
		benchCp3CounterGet();
		pcp3Cntr[0] = bcp3C0;
		pcp3Cntr[1] = bcp3C1;
		pcp3Cntr[2] = bcp3C2;
		pcp3Cntr[3] = bcp3C3;
	}
}

/* Input is the register value of Mon*ctr */
void benchmarkTime_start(uint32 counterDefine) {

	backupWatchDogReg = READ_MEM32( WDTCNR );
	WRITE_MEM32( WDTCNR,  WDSTOP_PATTERN<<WDTE_OFFSET );
	rtlglue_drvMutexLock();
	benchCp3CounterResetAndStart(counterDefine);
	prevBenchTC1CNT = READ_MEM32(TC1CNT);
}

/* Return number of 10ms */
unsigned int benchmarkTime_end(uint64 * pcp3Cntr) {
	uint32 jifPassed = 0;
	uint32 tuRemind = 0;

	if(pcp3Cntr) {
		currBenchTC1CNT = READ_MEM32(TC1CNT);
		benchCp3CounterStop();
		benchCp3CounterGet();
		pcp3Cntr[0] = bcp3C0;
		pcp3Cntr[1] = bcp3C1;
		pcp3Cntr[2] = bcp3C2;
		pcp3Cntr[3] = bcp3C3;
	}

	rtlglue_drvMutexUnlock();
	WRITE_MEM32( WDTCNR,  backupWatchDogReg );

	if ( prevBenchTC1CNT >= currBenchTC1CNT )
		tuRemind += (prevBenchTC1CNT-currBenchTC1CNT)>>TCD_OFFSET; /* how many units are passed since last check? */
	else
		tuRemind += (prevBenchTC1CNT+(0xffffff00-currBenchTC1CNT)+(0x1<<TCD_OFFSET))>>TCD_OFFSET; /* how many units are passed since last check? */
 
	while( tuRemind >= (READ_MEM32(TC0DATA)>>TCD_OFFSET) )
	{
		tuRemind -= (READ_MEM32(TC0DATA)>>TCD_OFFSET);
		jifPassed++;
	}
	return jifPassed;
}

void benchmarkTime_display(uint32 counterDefine, uint64 * pcp3Cntr) { 
	uint32 i;

	for(i=0; i<4; i++) {
		switch((counterDefine>>(i<<3))&0x1F) {
			case 0x10:
				printk("Counter %d %lld cycles\n", i, pcp3Cntr[i]);
				break;
			case 0x11:
				printk("Counter %d %lld new instructions\n", i, pcp3Cntr[i]);
				break;
			case 0x12:
				printk("Counter %d %lld instruction fetch cache misses\n", i, pcp3Cntr[i]);
				break;
			case 0x13:
				printk("Counter %d %lld inst fetch miss busy cycle\n", i, pcp3Cntr[i]);
				break;
			case 0x14:
				printk("Counter %d %lld data store instructions\n", i, pcp3Cntr[i]);
				break;
			case 0x15:
				printk("Counter %d %lld data load instructions\n", i, pcp3Cntr[i]);
				break;
			case 0x16:
				printk("Counter %d %lld data load or store instructions\n", i, pcp3Cntr[i]);
				break;
			case 0x17:
				printk("Counter %d %lld exact retired instructions\n", i, pcp3Cntr[i]);
				break;
			case 0x18:
				printk("Counter %d %lld pipe A retired instructions\n", i, pcp3Cntr[i]);
				break;
			case 0x19:
				printk("Counter %d %lld pipe B retired instructions\n", i, pcp3Cntr[i]);
				break;
			case 0x1A:
				printk("Counter %d %lld load or store cache miss\n", i, pcp3Cntr[i]);
				break;
			case 0x1B:
				printk("Counter %d %lld load or store miss busy cycles\n", i, pcp3Cntr[i]);
				break;
			default:
				printk("Unknown counter %d counter id %x\n", i, (counterDefine>>(i<<3))&0x1F);
				break;
		}
	}
}


