/*
* Copyright c                  Realtek Semiconductor Corporation, 2005  
* All rights reserved.
* 
* Program : Bogo MIPS
* Abstract : Porting from linux initialized bogomips
* Author : Edward Jin-Ru Chen (jzchen@realtek.com.tw)               
* $Id: bogomips.c,v 1.1.1.1 2006/07/26 05:38:21 blue Exp $
*/
#include "benchmarkTime.h"
#include <asm/param.h>
#include <linux/sched.h>
#include <linux/delay.h>

/* this should be approx 2 Bo*oMips to start (note initial shift), and will
   still work even if initially too large, it will just take slightly longer */
/* unsigned long loops_per_jiffy = (1<<12); */
extern unsigned long loops_per_jiffy;

/* This is the number of bits of precision for the loops_per_jiffy.  Each
   bit takes on average 1.5/HZ seconds.  This (like the original) is a little
   better than 1% */
#define LPS_PREC 8

void bogoMips(void)
{
	unsigned long ticks, loopbit;
	int lps_precision = LPS_PREC;
	unsigned int testRound;
	unsigned int cp3CtrlReg[4] = {0x1a121110, 0x1b131710, 0x19181710, 0x16151410};
	uint64 cp3Cntr[16];
	
	loops_per_jiffy = (1<<12);

	printk("Calibrating delay loop... ");
	while (loops_per_jiffy <<= 1) {
		/* wait for "start of" clock tick */
		ticks = jiffies;
		while (ticks == jiffies)
			/* nothing */;
		/* Go .. */
		ticks = jiffies;
		__delay(loops_per_jiffy);
		ticks = jiffies - ticks;
		if (ticks)
			break;
	}

/* Do a binary approximation to get loops_per_jiffy set to equal one clock
   (up to lps_precision bits) */
	loops_per_jiffy >>= 1;
	loopbit = loops_per_jiffy;
	while ( lps_precision-- && (loopbit >>= 1) ) {
		loops_per_jiffy |= loopbit;
		ticks = jiffies;
		while (ticks == jiffies);
		ticks = jiffies;
		__delay(loops_per_jiffy);
		if (jiffies != ticks)	/* longer than 1 tick */
			loops_per_jiffy &= ~loopbit;
	}

	for(testRound=0; testRound<4; testRound++) {
		benchmarkTime_start(cp3CtrlReg[testRound]);
		__delay(loops_per_jiffy);
		benchmarkTime_end(&cp3Cntr[testRound<<2]); 
	}
	for(testRound = 0; testRound<4; testRound++)
		benchmarkTime_display(cp3CtrlReg[testRound], &cp3Cntr[testRound<<2]);
	

/* Round the value and print it */	
	printk("%lu.%02lu BogoMIPS\n",
		loops_per_jiffy/(500000/HZ),
		(loops_per_jiffy/(5000/HZ)) % 100);
}

