/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : Model code for pattern match
* Abstract : 
* Author : Louis Yung-Chieh Lo (yjlou@realtek.com.tw)               
* $Id: gdmaTest.c,v 1.1.1.1 2006/07/26 05:38:22 blue Exp $
*/

#include "rtl_types.h"
#include "asicRegs.h"
#ifdef CONFIG_RTL865XC
#include "rtl865xC_tblAsicDrv.h"
#else
#include "rtl8651_tblAsicDrv.h"
#endif
#include "icExport.h"
#include "gdma.h"
#include "kmp.h"
#include "gdmaModel.h"
#include "virtualMac.h"
#include "rtl_glue.h"
#include "rtl_utils.h"
#include "icTest.h"
#include "drvTest.h"
#ifdef RTL865X_MODEL_USER
#include "errno.h"
#include <time.h>
#endif


#if 0
	#define EXPECT_ACCEPT( s ) do {} while (0)
	#define EXPECT_NOT_ACCEPT( s ) do {} while (0)
#else
#define EXPECT_ACCEPT( state ) \
	do { \
		if ( ( state&STATE_ESCAPE ) != STATE_ESCAPE ) \
		{ \
			rtlglue_printf( "[%s():%d] Expect ACCEPT, but not accept (0x%06x)\n", __FUNCTION__, __LINE__, state ); \
			return FAILED; \
		} \
	} while (0)

#define EXPECT_NOT_ACCEPT( state ) \
	do { \
		if ( ( state&STATE_ESCAPE ) == STATE_ESCAPE ) \
		{ \
			rtlglue_printf( "[%s():%d] Expect NOT ACCEPT, but accepted (0x%06x)\n", __FUNCTION__, __LINE__, state ); \
			return FAILED; \
		} \
	} while (0)
#endif



int32 displayStateMachine( pmVsm_t* pVsm )
{
	char* pDot;
	int32 retval = SUCCESS;

	retval = vsmGenDOT( pVsm, &pDot );
	ASSERT( retval==SUCCESS );
	ASSERT( pDot != NULL );

#ifdef RTL865X_MODEL_USER
	if ( getenv( "DISPLAY" ) )
	{
		/* DISPLAY is set, and we asuume that user have X window to show picture. */
		FILE *fp;
		char* cmd = "/usr/bin/dot -Tgif /dev/stdin -o /dev/stdout | /usr/X11R6/bin/xview /dev/stdin";
		
		fp = popen( cmd, "w" );
		if ( fp == NULL )
		{
			rtlglue_printf( "%s():%d cmd[%s] error:%s\n", __FUNCTION__, __LINE__, cmd, strerror(errno) );
			ASSERT ( fp!=NULL );
		}

		fprintf( fp, "%s", pDot );
		fclose( fp );
	}
	else
	{
		rtlglue_printf( "To show graphic of state machine, setenv DISPLAY in your environment variable.\n" );
		rtlglue_printf( "Generated DOT:\n%s\n", pDot );
	}
#else
	rtlglue_printf( "Generated DOT:\n%*s\n", strlen(pDot), pDot );
#endif
	if ( pDot )
		rtlglue_free( pDot );

	return retval;
}


/******************************************************************************
 ******************************************************************************
 **                                                                          **
 **  test bench                                                              **
 **                                                                          **
 ******************************************************************************
 ******************************************************************************/




/******************************************************************************
 **  case 1                                                                  **
 ******************************************************************************/
/* KMP: nano */
subStateMachine_t subSM1 =
{
	rules:
	{
		{ 
			type:0,
			not: 0,
			cpu: 0,
			jmpState: 2,
			matchChar: 'n'
		},
		{ 
			type:0,
			not: 1,
			cpu: 0,
			jmpState: 0,
			matchChar: 'n'
		},
		{ 
			type:0,
			not: 0,
			cpu: 0,
			jmpState: 5,
			matchChar: 'a'
		},
		{
			type:0,
			not: 0,
			cpu: 0,
			jmpState: 0,
			matchChar: 'n'
		},
		{ 
			type:0,
			not: 1,
			cpu: 0,
			jmpState: 0,
			matchChar: 'a'
		},
		{ 
			type:0,
			not: 0,
			cpu: 0,
			jmpState: 7,
			matchChar: 'n'
		},
		{ 
			type:0,
			not: 1,
			cpu: 0,
			jmpState: 0,
			matchChar: 'n'
		},
		{ 
			type:0,
			not: 0,
			cpu: 1,
			jmpState: STATE_ACCEPT,
			matchChar: 'o'
		},
		{ 
			type:0,
			not: 0,
			cpu: 0,
			jmpState: 5,
			matchChar: 'a'
		},
		{ 
			type:0,
			not: 1,
			cpu: 0,
			jmpState: 2,
			matchChar: 'o'
		},
	},
};
stateMachine_t SM1 =
{
	subsm:
	{
		&subSM1,
		NULL
	}
};
uint8 testString1[] = "nanano";
descriptor_t descTestString1[] = 
{
	{
		pData: (void*)testString1,
		ldb: 1,
		length: sizeof(testString1)/sizeof(testString1[0])-1
	},
};

int32 testGdmaPatternMatch1(uint32 caseNo)
{
	int i;
	int32 retval;

	rtlglue_printf( "TEST 1 (KMP:nano,Expect:ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */
		
		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, ((uint32*)descTestString1)[0] );
		WRITE_MEM32( GDMASBL0, (descTestString1[0].ldb<<31)|(descTestString1[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)&SM1 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN );
		while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */
	
		EXPECT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), 5, __FUNCTION__, __LINE__ );
	}

	return SUCCESS;
}


/******************************************************************************
 **  case 2                                                                  **
 ******************************************************************************/
/* RE: {ab|bc} */
subStateMachine_t subSM2 =
{
	rules:
	{
		{/*0:A*/
			not: 0,
			cpu: 0,
			matchChar: 'a',
			jmpState: 3,
		},
		{/*1*/
			not: 0,
			cpu: 0,
			matchChar: 'b',
			jmpState: 6,
		},
		{/*2*/
			not: 1,
			cpu: 0,
			matchChar: 'b',
			jmpState: 0,
		},
		{/*3:B*/
			not: 0,
			cpu: 0,
			matchChar: 'a',
			jmpState: 3,
		},
		{/*4*/
			not: 0,
			cpu: 1,
			matchChar: 'b',
			jmpState: STATE_ACCEPT,
		},
		{/*5*/
			not: 1,
			cpu: 0,
			matchChar: 'b',
			jmpState: 0,
		},
		{/*6:C*/
			not: 0,
			cpu: 0,
			matchChar: 'a',
			jmpState: 3,
		},
		{/*7*/
			not: 0,
			cpu: 0,
			matchChar: 'b',
			jmpState: 6,
		},
		{/*8*/
			not: 0,
			cpu: 1,
			matchChar: 'c',
			jmpState: STATE_ACCEPT,
		},
		{/*9*/
			not: 1,
			cpu: 0,
			matchChar: 'c',
			jmpState: 0,
		},
	},
};
stateMachine_t SM2 =
{
	subsm:
	{
		&subSM2,
		NULL
	},
};
uint8 testString2[] = "aacacccb";
descriptor_t descTestString2[] = 
{
	{
		pData: (void*)testString2,
		ldb: 1,
		length: sizeof(testString2)/sizeof(testString2[0])-1
	},
};
uint8 testString21[] = "aacabbccc";
descriptor_t descTestString21[] = 
{
	{
		pData: (void*)testString21,
		ldb: 1,
		length: sizeof(testString21)/sizeof(testString21[0])-1
	},
};
uint8 testString22[] = "aacbcaacc";
descriptor_t descTestString22[] = 
{
	{
		pData: (void*)testString22,
		ldb: 1,
		length: sizeof(testString22)/sizeof(testString22[0])-1
	},
};
uint8 testString23[] = "acbaaccaccbaaccaxyyyyyyyyyy";
descriptor_t descTestString23[] = 
{
	{
		pData: (void*)testString23,
		ldb: 1,
		length: sizeof(testString23)/sizeof(testString23[0])-1
	},
};
uint8 testString241[] = "aaaccc";
uint8 testString242[] = "acb";
uint8 testString243[] = "ccc";
descriptor_t descTestString24[] = 
{
	{
		pData: (void*)testString241,
		ldb: 0,
		length: sizeof(testString241)/sizeof(testString241[0])-1
	},
	{
		pData: (void*)testString242,
		ldb: 0,
		length: sizeof(testString242)/sizeof(testString242[0])-1
	},
	{
		pData: (void*)testString243,
		ldb: 1,
		length: sizeof(testString243)/sizeof(testString243[0])-1
	},
};

int32 testGdmaPatternMatch2(uint32 caseNo)
{
	int i;
	int32 retval;
	
	rtlglue_printf( "TEST 2 (RE:{ab|bc},Expect:NOT ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */
		
		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString2[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString2[0].ldb<<31)|(descTestString2[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)&SM2 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_NOT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), descTestString2[0].length, __FUNCTION__, __LINE__ );
	}

	rtlglue_printf( "TEST 2-1 (RE:{ab|bc},Expect:ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */
		
		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString21[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString21[0].ldb<<31)|(descTestString21[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)&SM2 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), 4, __FUNCTION__, __LINE__ );
	}

	rtlglue_printf( "TEST 2-2 (RE:{ab|bc},Expect:ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString22[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString22[0].ldb<<31)|(descTestString22[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)&SM2 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), 4, __FUNCTION__, __LINE__ );
	}

	rtlglue_printf( "TEST 2-3 (RE:{ab|bc},Expect:NOT ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString23[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString23[0].ldb<<31)|(descTestString23[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)&SM2 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_NOT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), descTestString23[0].length, __FUNCTION__, __LINE__ );
	}
	
	rtlglue_printf( "TEST 2-4 (RE:{ab|bc},Expect:ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */
		
		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString24[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString24[0].ldb<<31)|(descTestString24[0].length) );
		WRITE_MEM32( GDMASBP1, (uint32)(descTestString24[1].pData) );
		WRITE_MEM32( GDMASBL1, (descTestString24[1].ldb<<31)|(descTestString24[1].length) );
		WRITE_MEM32( GDMASBP2, (uint32)(descTestString24[2].pData) );
		WRITE_MEM32( GDMASBL2, (descTestString24[2].ldb<<31)|(descTestString24[2].length) );
		WRITE_MEM32( GDMADBP0, (uint32)&SM2 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), 9, __FUNCTION__, __LINE__ );
	}
	
	return SUCCESS;
}

	
/******************************************************************************
 **  case 3                                                                  **
 ******************************************************************************/
/* RE: ^abcde */
subStateMachine_t subSM3 =
{
	rules:
	{
		{/*0:A*/
			not: 0,
			cpu: 0,
			matchChar: 'a',
			jmpState: 2,
		},
		{/*1:!A*/
			not: 1,
			cpu: 1,
			matchChar: 'a',
			jmpState: STATE_DEAD,
		},
		{/*2:B*/
			not: 0,
			cpu: 0,
			matchChar: 'b',
			jmpState: 4,
		},
		{/*3:!B*/
			not: 1,
			cpu: 1,
			matchChar: 'b',
			jmpState: STATE_DEAD,
		},
		{/*4:C*/
			not: 0,
			cpu: 0,
			matchChar: 'c',
			jmpState: 6,
		},
		{/*5:!C*/
			not: 1,
			cpu: 1,
			matchChar: 'c',
			jmpState: STATE_DEAD,
		},
		{/*6:D*/
			not: 0,
			cpu: 0,
			matchChar: 'd',
			jmpState: 8,
		},
		{/*7:!D*/
			not: 1,
			cpu: 1,
			matchChar: 'd',
			jmpState: STATE_DEAD,
		},
		{/*8:E*/
			not: 0,
			cpu: 1,
			matchChar: 'e',
			jmpState: STATE_ACCEPT,
		},
		{/*9:!E*/
			not: 1,
			cpu: 1,
			matchChar: 'e',
			jmpState: STATE_DEAD,
		},
	},
};
stateMachine_t SM3 =
{
	subsm:
	{
		&subSM3,
		NULL
	}
};
uint8 testString31[] = "ab";
descriptor_t descTestString31[] = 
{
	{
		pData: (void*)testString31,
		ldb: 1,
		length: sizeof(testString31)/sizeof(testString31[0])-1
	},
};
uint8 testString32[] = "c";
descriptor_t descTestString32[] = 
{
	{
		pData: (void*)testString32,
		ldb: 1,
		length: sizeof(testString32)/sizeof(testString32[0])-1
	},
};
uint8 testString33[] = "de";
descriptor_t descTestString33[] = 
{
	{
		pData: (void*)testString33,
		ldb: 1,
		length: sizeof(testString33)/sizeof(testString33[0])-1
	},
};

int32 testGdmaPatternMatch3(uint32 caseNo)
{
	int i;
	int32 retval;
	
	rtlglue_printf( "TEST 3 (RE:^abcde,fragment packet,Expect 1st:NOT ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString31[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString31[0].ldb<<31)|(descTestString31[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)&SM3 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_NOT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), descTestString31[0].length, __FUNCTION__, __LINE__ );
	}
	
	rtlglue_printf( "                                  Expect 2nd:NOT ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */
		
		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, READ_MEM32( GDMAICVL ) );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString32[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString32[0].ldb<<31)|(descTestString32[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)&SM3 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_NOT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), descTestString32[0].length, __FUNCTION__, __LINE__ );
	}
	
	rtlglue_printf( "                                  Expect 3rd:ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, READ_MEM32( GDMAICVL ) );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString33[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString33[0].ldb<<31)|(descTestString33[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)&SM3 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), 1, __FUNCTION__, __LINE__ );
	}
	
	return SUCCESS;
}
	

/******************************************************************************
 **  case 4                                                                  **
 ******************************************************************************/
/* RE: ^ab */
subStateMachine_t subSM40 =
{
	rules:
	{
		{/*0x0000:A*/
			not: 0,
			cpu: 0,
			matchChar: 'a',
			jmpState: 0x0100,
		},
		{/*0x0001:!A*/
			not: 1,
			cpu: 1,
			matchChar: 'a',
			jmpState: STATE_DEAD,
		},
	},
};
subStateMachine_t subSM41 =
{
	rules:
	{
		{/*0x0100:B*/
			not: 0,
			cpu: 1,
			matchChar: 'b',
			jmpState: STATE_ACCEPT,
		},
		{/*0x0101:!B*/
			not: 1,
			cpu: 1,
			matchChar: 'b',
			jmpState: STATE_DEAD,
		},
	},
};
stateMachine_t SM4 =
{
	subsm:
	{
		&subSM40,
		&subSM41,
		NULL
	},
};
uint8 testString4[] = "ab";
descriptor_t descTestString4[] = 
{
	{
		pData: (void*)testString4,
		ldb: 1,
		length: sizeof(testString4)/sizeof(testString4[0])-1
	},
};

int32 testGdmaPatternMatch4(uint32 caseNo)
{
	int i;
	int32 retval;

	rtlglue_printf( "TEST 4 (RE:^ab,Inter-sub-SM,Expect:ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString4[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString4[0].ldb<<31)|(descTestString4[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)&SM4 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), 1, __FUNCTION__, __LINE__ );
	}

	return SUCCESS;
}


/******************************************************************************
 **  case 5                                                                  **
 ******************************************************************************/
/* KMP: nanano */
uint8 testString51[] = "aaaoonoananananoaa";
descriptor_t descTestString51[] = 
{
	{
		pData: (void*)testString51,
		ldb: 1,
		length: sizeof(testString51)/sizeof(testString51[0])-1
	},
};
uint8 testString52[] = "nanananananaonananaoonanaao";
descriptor_t descTestString52[] = 
{
	{
		pData: (void*)testString52,
		ldb: 1,
		length: sizeof(testString52)/sizeof(testString52[0])-1
	},
};

int32 testGdmaPatternMatch5(uint32 caseNo)
{
	int i;
	int32 retval;
	pmVsm_t vsm;

	vsmInit( &vsm );
	vsmGenFromKMP( &vsm, "nanano", 6 );

	rtlglue_printf( "TEST 5-1 (KMP:nanano,Expect:ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString51[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString51[0].ldb<<31)|(descTestString51[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)vsm.prefix[0].pStateMachine );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), 15, __FUNCTION__, __LINE__ );
	}
	
	rtlglue_printf( "TEST 5-2 (KMP:nanano,Expect:NOT ACCEPT):" );
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		WRITE_MEM32( GDMACNR, 0 );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)(descTestString52[0].pData) );
		WRITE_MEM32( GDMASBL0, (descTestString52[0].ldb<<31)|(descTestString52[0].length) );
		WRITE_MEM32( GDMADBP0, (uint32)vsm.prefix[0].pStateMachine );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
		EXPECT_NOT_ACCEPT( READ_MEM32( GDMAICVL ) );
		IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), descTestString52[0].length, __FUNCTION__, __LINE__ );
	}
	
	vsmFree( &vsm );
	return SUCCESS;
}
	

/******************************************************************************
 **  case 6                                                                  **
 ******************************************************************************/
/* KMP: porn-porn-.... */
uint8 testString6[1500];
descriptor_t descTestString6[] = 
{
	{
		pData: (void*)testString6,
		ldb: 1,
		length: sizeof(testString6)/sizeof(testString6[0])-1
	},
};

int32 testGdmaPatternMatch6(uint32 caseNo)
{
	int i, j, k;
	int accept, notaccept;
	int32 retval;
	pmVsm_t vsm;
	char *pat = "nanonino"; /*"porn-porn-porn-porn-porn-porn-porn-porn-porn-porn-porn-porn.info/free-hardcore-interracial-porn-gallery-hardcore-interracial-porn/free-hardcore-interracial-porn-gallery.php";*/
	uint32 t; /* time_t */
	uint8 *found;

	rtlglue_printf( "TEST 6 (KMP:pron-pron-...,Expect:as same as strstr()):\n" );

	rtlglue_srandom( rtlglue_time(&t) );
	vsmInit( &vsm );
	retval = vsmGenFromKMP( &vsm, pat, strlen(pat) );
	ASSERT( retval==SUCCESS );

	displayStateMachine( &vsm );
	
	accept = notaccept = 0;
	for( i = 0; i < 10000; i++ )
	{
		for( j = 0; j < sizeof(testString6)-1; j++ )
			testString6[j] = rtlglue_random()%26 + 'a';
		testString6[j] = '\0';
		if ( rtlglue_random()%2 )
		{
			k = rtlglue_random()%(sizeof(testString6)-1-strlen(pat));
			memcpy( &testString6[k], pat, strlen(pat) );
		}
		
		for( k = IC_TYPE_MIN; k < IC_TYPE_MAX; k++ )
		{
			retval = model_setTestTarget( k );
			if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

			WRITE_MEM32( GDMACNR, 0 );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE );
			WRITE_MEM32( GDMAICVL, 0 );
			WRITE_MEM32( GDMAICVR, 0 );
			WRITE_MEM32( GDMASBP0, (uint32)(descTestString6[0].pData) );
			WRITE_MEM32( GDMASBL0, (descTestString6[0].ldb<<31)|(descTestString6[0].length) );
			WRITE_MEM32( GDMADBP0, (uint32)vsm.prefix[0].pStateMachine );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_PATTERN);
	
			if ( ( found = strstr( testString6, pat ) ) )
			{
				accept++;
				EXPECT_ACCEPT( READ_MEM32( GDMAICVL ) );
				IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), (found-testString6)+strlen(pat)-1, __FUNCTION__, __LINE__ );
			}
			else
			{
				notaccept++;
				EXPECT_NOT_ACCEPT( READ_MEM32( GDMAICVL ) );
				IS_EQUAL_INT("Processed is not matched", READ_MEM32(GDMAICVR), descTestString6[0].length, __FUNCTION__, __LINE__ );
			}
		}
	}
	/*modelGDMA_dump();*/
	vsmFree( &vsm );

	rtlglue_printf("Accepted:%d  Not Accepted:%d\n", accept, notaccept );

	return SUCCESS;
}


#if 0
/******************************************************************************
 **  read Clam AV database file 'virus.txt'                                  **
 ******************************************************************************/
int32 testGdmaPatternMatch7(uint32 caseNo)
{
	static uint8 buf[4096];
	static pmVsm_t vsm[30000]; 
	uint32 cntVsm = 0;
	char* filename = "virus.txt";
	FILE* fp;
	int32 i;
	int32 maxPatternLen, maxCntInst;
	double maxRatio;

	maxPatternLen = maxCntInst = maxRatio = 0;
	
	/* generate state machine from pattern */
	fp = fopen( filename, "r" );
	if ( fp==NULL )
	{
		char buf[256];
		snprintf( buf, sizeof(buf)-1, "open '%s' file error.", filename );
		perror( buf );
		return FAILED;
	}
	else
	{
		while( fgets( buf, sizeof(buf)-1, fp ) )
		{
			uint8 str[512];
			uint32 cntInst;
			int32 lenPattern;
			
			vsmInit( &vsm[cntVsm] );
			for( i = 0, lenPattern = 0; buf[i]&&buf[i+1]; i+=2, lenPattern++ )
			{
				char ch[3];

				ch[0] = buf[i];
				ch[1] = buf[i+1];
				ch[2] = 0;

				str[lenPattern] = strtol( ch, NULL, 16 );
			}
			if ( vsmGenFromKMP( &vsm[cntVsm], str, lenPattern )==FAILED )
			{
				rtlglue_printf( "vsmGenFromKMP(%d) error.\n", cntVsm );
				return FAILED;
			}

			cntInst = vsm[cntVsm].prefix[0].pStateMachine->cntTotalRules;
			rtlglue_printf("[%6d] patternLen=%d cntInst=%d  ratio=%.2f\n", cntVsm, lenPattern, cntInst, (cntInst*1.0)/lenPattern );

			if ( maxPatternLen < lenPattern ) maxPatternLen = lenPattern;
			if ( maxCntInst < cntInst ) maxCntInst = cntInst;
			if ( maxRatio < (cntInst*1.0)/lenPattern ) maxRatio = (cntInst*1.0)/lenPattern;
			
			cntVsm++;
		}
		fclose( fp );
	}

	rtlglue_printf("MaxPatternLen=%d MaxCntInst=%d MaxRatio=%.2f\n", maxPatternLen, maxCntInst, maxRatio );
	
#if 0
	{
		static uint8 packet[1500] = "123456";
		descriptor_t descPacket[] = 
		{
			{
				pData: (void*)packet,
				end: 1,
				length: sizeof(packet)/sizeof(packet[0])
			},
		};
		/* parse packet .... */
		for( i = 0; i < cntSubSM; i++ )
		{
			WRITE_MEM32( GDMACR, 1<<28 );
			WRITE_MEM32( GDMASDR, (uint32)descPacket );
			WRITE_MEM32( GDMADDR, (uint32)vsm.prefix[0].pStateMachine );
			if ( modelGDMA() != SUCCESS ) rtlglue_printf( "\n*** State Machine ERROR ***\n" );
		}
#endif

	/* free VSM */
	{
		int i;
		for( i = 0; i < cntVsm; i++ )
		{
			vsmFree( &vsm[i] );
		}
	}

	return SUCCESS;
}


/******************************************************************************
 **  read URL black list with 'url.txt '                                    **
 ******************************************************************************/
int32 testGdmaPatternMatch8(uint32 caseNo)
{
	static uint8 buf[4096];
	static pmVsm_t vsm[30000];
	char *filename = "urls.txt";
	FILE* fp;
	int32 maxPatternLen, maxCntInst;
	double maxRatio;
	uint32 cntVsm = 0;

	maxPatternLen = maxCntInst = maxRatio = 0;
	
	/* generate state machine from pattern */
	fp = fopen( filename, "r" );
	if ( fp==NULL )
	{
		char buf[256];
		snprintf( buf, sizeof(buf)-1, "open '%s' file error.", filename );
		perror( buf );
		return FAILED;
	}
	else
	{
		while( fgets( buf, sizeof(buf)-1, fp ) && cntVsm<sizeof(vsm)/sizeof(vsm[0]) )
		{
			uint32 cntInst;
			int32 lenPattern;

			vsmInit( &vsm[cntVsm] );
			lenPattern = strlen( buf );

			if ( vsmGenFromKMP( &vsm[cntVsm], buf, lenPattern )==FAILED )
			{
				rtlglue_printf( "vsmGenFromKMP(%d) error.\n", cntVsm );
				return FAILED;
			}

			cntInst = vsm[cntVsm].prefix[0].pStateMachine->cntTotalRules;
			rtlglue_printf("[%6d] patternLen=%d cntInst=%d  ratio=%.2f\n", cntVsm, lenPattern, cntInst, (cntInst*1.0)/lenPattern );

			if ( maxPatternLen < lenPattern ) maxPatternLen = lenPattern;
			if ( maxCntInst < cntInst ) maxCntInst = cntInst;
			if ( maxRatio < (cntInst*1.0)/lenPattern ) maxRatio = (cntInst*1.0)/lenPattern;
			
			cntVsm++;
		}
		fclose( fp );
	}

	rtlglue_printf("MaxPatternLen=%d MaxCntInst=%d MaxRatio=%.2f\n", maxPatternLen, maxCntInst, maxRatio );

#if 0
	{
		static uint8 packet[1500] = "123456";
		descriptor_t descPacket[] = 
		{
			{
				pData: (void*)packet,
				end: 1,
				length: sizeof(packet)/sizeof(packet[0])
			},
		};
		int32 i;
		/* parse packet .... */
		for( i = 0; i < cntSubSM; i++ )
		{
			WRITE_MEM32( GDMACR, 1<<28 );
			WRITE_MEM32( GDMASDR, (uint32)descPacket );
			WRITE_MEM32( GDMADDR, (uint32)vsm.prefix[0].pStateMachine );
			if ( modelGDMA() != SUCCESS ) rtlglue_printf( "\n*** State Machine ERROR ***\n" );
		}
	}
#endif

	/* free VSM */
	{
		int i;
		for( i = 0; i < cntVsm; i++ )
		{
			vsmFree( &vsm[i] );
		}
	}

	return SUCCESS;
}


/******************************************************************************
 **  Test case to evaluate the performance of KMP algorithm                  **
 **  The test case is diverted from 'drv romeperf bmp'                       **
 ******************************************************************************/
uint8 testString9[1501];
descriptor_t descTestString9[] = 
{
	{
		pData: (void*)testString9,
		end: 1,
		length: sizeof(testString9)/sizeof(testString9[0])-1
	},
};

int32 testGdmaPatternMatch9(uint32 caseNo)
{
	int i, j;
	int32 retval;
	pmVsm_t vsm;
	char pattern[] = "free-quit-kit.how-quit-smoking.com/picture-of-someone-smoking-a-cigarette.how-quit-smoking.com";
	int patLen;
	int loop = 1000000;
	uint32 seed = 0x5566;
	char randChar[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz:/?&-";

	rtlglue_printf( "TEST 9 (To evaluate the performance of KMP algorithm):\n" );
	rtlglue_printf( "  Test Loop=%d\n", loop );
	rtlglue_printf( "  Random Seed=%d\n", seed );
	rtlglue_printf( "  pattern='%s'\n", pattern );

	
	for( patLen = 1; patLen < strlen(pattern); patLen+=8 )
	{
		vsmInit( &vsm );
		retval = vsmGenFromKMP( &vsm, pattern, patLen );
		ASSERT( retval==SUCCESS );
		displayStateMachine( &vsm );
		
		for( loop = 1000000; loop > 0; loop-- )
		{
			char subChar[sizeof(pattern)];

			sched_yield();
			
			for( j = 0; j < sizeof(testString9)-1; j++ )
			{
				testString9[j] = randChar[seed%(sizeof(randChar)-1)];
				seed = seed * 0x13579bdf + 0xfdb97531;
			}
			testString9[j] = '\0';
			//rtlglue_printf( "  packet='%s'\n", testString9 );
			
			WRITE_MEM32( GDMACR, 1<<28 );
			WRITE_MEM32( GDMAICVL, 0 );
			WRITE_MEM32( GDMAICVR, 0 );
			WRITE_MEM32( GDMASDR, (uint32)descTestString9 );
			WRITE_MEM32( GDMADDR, (uint32)vsm.prefix[0].pStateMachine );
			if ( modelGDMA() != SUCCESS ) rtlglue_printf( "\n*** State Machine ERROR ***\n" );

			/* sub-string of pattern */
			strncpy( subChar, pattern, patLen );
			subChar[patLen] = '\0';
			if ( strstr( testString9, subChar ) )
				EXPECT_ACCEPT( READ_MEM32( GDMAICVL ) );
			else
				EXPECT_NOT_ACCEPT( READ_MEM32( GDMAICVL ) );
		}

		rtlglue_printf( "patLen=%d\n", patLen );
		modelGDMA_dump();
		vsmFree( &vsm );
	}

	return SUCCESS;
}

#endif


#define HEAPSIZE (8*4*1024)
static uint8 sb1[HEAPSIZE]; /* source heap (expected) */
static uint8 db1[HEAPSIZE]; /* destination heap (expected) */
static uint8 sb2[HEAPSIZE]; /* source heap (to be compared) */
static uint8 db2[HEAPSIZE]; /* destination heap (to be compared) */


/*
 *
 */
int32 compare_data_block( void )
{
	if ( memcmp( sb1, sb2, HEAPSIZE ) != 0 )
	{
		rtlglue_printf( "sb1(%p) sb2(%p) is different.\n", sb1, sb2 );
		return FAILED;
	}
	if ( memcmp( db1, db2, HEAPSIZE ) != 0 )
	{
		rtlglue_printf( "db1(%p) db2(%p) is different.\n", db1, db2 );
		return FAILED;
	}
	return SUCCESS;
}


/*
 *	This function will enumulate block alignment from continued data range.
 *	Results are stored in either source block pointer/length or destination block pointer/length.
 *	(4BA: 4 bytes aligned)
 *		pointer: 4BA, 		length: 4BA
 *		pointer: 4BA+1,		length: 4BA
 *		pointer: 4BA+2,		length: 4BA
 *		pointer: 4BA+3,		length: 4BA
 *	#block = { 1, 2, 3, 4, 5, 6, 7, 8 };
 *	pointer = { 4BA+0, 4BA+1, 4BA+2, 4BA+3, 16BA+3, FIFOBoundary, Bank+1, ChipBoundary };
 *	length = { 4BA+0, 4BA+1, 4BA+2, 4BA+3, FIFOSize, BankSize,  };
 */
int32 enumBlockAlignment( int32 en, uint8 *data, int32 len, int32 bSrc )
{
	return SUCCESS;
}

/******************************************************************************
 ******************************************************************************
 **                                                                          **
 **  Memory Functions                                                        **
 **                                                                          **
 ******************************************************************************
 ******************************************************************************/
/*
 *  +----------------+
 *  | AAAAAAAAAAAAAAA|  Block A/B/C/D is 64-bytes long and located at offset 1.
 *  |AAAAAAAAAAAAAAAA|  We will test 4 memory functions together, and alignment issue.
 *  |AAAAAAAAAAAAAAAA|  
 *  |AAAAAAAAAAAAAAAA|  0. init A with meaningful string
 *  |ABBBBBBBBBBBBBBB|  1. memcpy from A to B
 *  |BBBBBBBBBBBBBBBB|  2. memset C and D with 0xa5
 *  |BBBBBBBBBBBBBBBB|  3. memXor (A,B) and (C,D)
 *  |BBBBBBBBBBBBBBBB|  4. memcmp with C and D, and expected identical
 *  |BCCCCCCCCCCCCCCC|
 *  |CCCCCCCCCCCCCCCC|
 *  |CCCCCCCCCCCCCCCC|
 *  |CCCCCCCCCCCCCCCC|
 *  |CDDDDDDDDDDDDDDD|
 *  |DDDDDDDDDDDDDDDD|
 *  |DDDDDDDDDDDDDDDD|
 *  |DDDDDDDDDDDDDDDD|
 *  |D               |
 *  +----------------+
 */
int32 testGdmaMemory1(uint32 caseNo)
{
	int i, j;
	int32 retval;
	uint8 str[64] = "Hi, everybody. I am Louis. I come from Taiwan. Glad to meet you.";
	
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		/* clean memory */
		memset( sb1, 0, HEAPSIZE );
		memset( sb2, 0, HEAPSIZE );
		memset( db1, 0, HEAPSIZE );
		memset( db2, 0, HEAPSIZE );

		/* 0. init A with meaningful string */
		memcpy( sb2+1, str, sizeof(str) );

		/* 1. memcpy from A to B */
		WRITE_MEM32( GDMACNR, 0 ); /* Reset GDMA */
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_MEMCPY );
		WRITE_MEM32( GDMAIMR, 0 );
		WRITE_MEM32( GDMAISR, 0xffffffff );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)sb2+1+sizeof(str)*0 ); /* A */
		WRITE_MEM32( GDMASBL0, GDMA_LDB|sizeof(str) );
		WRITE_MEM32( GDMADBP0, (uint32)sb2+1+sizeof(str)*1 ); /* B */
		WRITE_MEM32( GDMADBL0, GDMA_LDB|sizeof(str) );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_MEMCPY );

		/* Expected data */
		memcpy( sb1+1+sizeof(str)*0, str, sizeof(str) );
		memcpy( sb1+1+sizeof(str)*1, str, sizeof(str) );
		while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

		IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
		IS_EQUAL_INT("Source block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, 0, __FUNCTION__, __LINE__ );
		ASSERT( compare_data_block() == SUCCESS );
		
		/* 2. memset C and D with 0xa5 */
		WRITE_MEM32( GDMACNR, 0 ); /* Reset GDMA */
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_MEMSET );
		WRITE_MEM32( GDMAIMR, 0 );
		WRITE_MEM32( GDMAISR, 0xffffffff );
		WRITE_MEM32( GDMAICVL, 0xa5 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP1, 0 );
		WRITE_MEM32( GDMASBL1, GDMA_LDB );
		WRITE_MEM32( GDMADBP0, (uint32)sb2+1+sizeof(str)*2 ); /* C */
		WRITE_MEM32( GDMADBL0, sizeof(str) );
		WRITE_MEM32( GDMADBP1, (uint32)sb2+1+sizeof(str)*3 ); /* D */
		WRITE_MEM32( GDMADBL1, GDMA_LDB|sizeof(str) );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_MEMSET );

		/* Expected data */
		memset( sb1+1+sizeof(str)*2, 0xa5, sizeof(str) );
		memset( sb1+1+sizeof(str)*3, 0xa5, sizeof(str) );
		while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

		IS_EQUAL_INT("Dest block should under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, GDMA_DBUDRIP, __FUNCTION__, __LINE__ );
		IS_EQUAL_INT("Source block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, 0, __FUNCTION__, __LINE__ );
		ASSERT( compare_data_block() == SUCCESS );
		
		/* 3. memXor (A,B) and (C,D) */
		WRITE_MEM32( GDMACNR, 0 ); /* Reset GDMA */
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_MEMXOR );
		WRITE_MEM32( GDMAIMR, 0 );
		WRITE_MEM32( GDMAISR, 0xffffffff );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)sb2+1+sizeof(str)*0 ); /* A,B */
		WRITE_MEM32( GDMASBL0, GDMA_LDB|(sizeof(str)*2) );
		WRITE_MEM32( GDMADBP0, (uint32)sb2+1+sizeof(str)*2 ); /* C,D */
		WRITE_MEM32( GDMADBL0, GDMA_LDB|(sizeof(str)*2) );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_MEMXOR );

		/* Expected data */
		for( j = 0; j < sizeof(str)*2; j++ )
		{
			sb1[1+sizeof(str)*2+j] ^= sb1[1+sizeof(str)*0+j];
		}
		while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

		IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
		IS_EQUAL_INT("Source block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, 0, __FUNCTION__, __LINE__ );
		ASSERT( compare_data_block() == SUCCESS );
		
		/* 4. memcmp with C and D */
		WRITE_MEM32( GDMACNR, 0 ); /* Reset GDMA */
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_MEMCMP );
		WRITE_MEM32( GDMAIMR, 0 );
		WRITE_MEM32( GDMAISR, 0xffffffff );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)sb2+1+sizeof(str)*2 ); /* C */
		WRITE_MEM32( GDMASBL0, GDMA_LDB|(sizeof(str)*1) );
		WRITE_MEM32( GDMADBP0, (uint32)sb2+1+sizeof(str)*3 ); /* D */
		WRITE_MEM32( GDMADBL0, GDMA_LDB|(sizeof(str)*1) );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_MEMCMP );
		
		IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
		IS_EQUAL_INT("Source block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, 0, __FUNCTION__, __LINE__ );
		IS_EQUAL_INT("C and D should be identical", READ_MEM32(GDMAICVL), 0, __FUNCTION__, __LINE__ );
		ASSERT( compare_data_block() == SUCCESS );
		
	}

	return SUCCESS;
}


/******************************************************************************
 ******************************************************************************
 **                                                                          **
 **  Checksum Offload                                                        **
 **                                                                          **
 ******************************************************************************
 ******************************************************************************/
/*
 * Test fixed pattern
 */
int32 testGdmaChecksum1(uint32 caseNo)
{
	struct TEST_VECTOR
	{
		uint32 icv;
		char *source;
		int len;
		uint32 sum;
	}
	static testvec[] =
	{
		{
			icv: 0xdeadc0de, /* test initial value */
			source: "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0",
			len: 20,
			sum: 0xdeadc0de,
		},
		{
			icv: 0x0011+0x004c, /* 0 + protocol + length */
			source: "\
\x45\x00\x00\x60\x27\x98\x00\x00\x7f\x11\x0b\x68\xc0\xa8\x9a\xcf\
\xac\x0a\x01\x0b\x01\xff\x00\x89\x00\x4c\x93\xdf\x82\x76\x40\x00\
\x00\x01\x00\x00\x00\x00\x00\x01\x20\x41\x42\x41\x43\x46\x50\x46\
\x50\x45\x4e\x46\x44\x45\x43\x46\x43\x45\x50\x46\x48\x46\x44\x45\
\x46\x46\x50\x46\x50\x41\x43\x41\x42\x00\x00\x20\x00\x01\xc0\x0c\
\x00\x20\x00\x01\x00\x04\x93\xe0\x00\x06\xe0\x00\xc0\xa8\x00\xb9\
",
			len: 96,
			sum: 0x000afff5,
		},
		{
			icv: 0x0006+0x0086, /* 0 + protocol + length */
			source: "\
\x45\x00\x00\x9a\xe4\x9e\x40\x00\x7d\x06\x0f\xd8\xac\x15\x01\x9c\
\xc0\xa8\x9a\x8d\x00\x50\x04\x90\x4b\x16\x78\xeb\x8e\xa6\x25\x8f\
\x80\x18\xfe\x61\xcb\x10\x00\x00\x01\x01\x08\x0a\x03\x9c\x48\x43\
\x00\x70\x2b\xbb\x48\x54\x54\x50\x2f\x31\x2e\x31\x20\x33\x30\x34\
\x20\x4f\x4b\x0d\x0a\x53\x65\x72\x76\x65\x72\x3a\x20\x4d\x69\x63\
\x72\x6f\x73\x6f\x66\x74\x2d\x49\x49\x53\x2f\x35\x2e\x30\x0d\x0a\
\x44\x61\x74\x65\x3a\x20\x54\x68\x75\x2c\x20\x30\x34\x20\x41\x75\
\x67\x20\x32\x30\x30\x35\x20\x31\x34\x3a\x33\x33\x3a\x32\x33\x20\
\x47\x4d\x54\x0d\x0a\x43\x6f\x6e\x6e\x65\x63\x74\x69\x6f\x6e\x3a\
\x20\x63\x6c\x6f\x73\x65\x0d\x0a\x0d\x0a",
			len: 154,
			sum: 0x0013ffec,
		},
	};
	int i, j;
	int32 retval;
	
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		for( j = 0; j < sizeof(testvec)/sizeof(testvec[0]); j++ )
		{
			int _offset[] = { 0, 2, -1 };
			int *offset;
			
			for( offset = _offset; *offset>=0; offset++ )
			{
				/* clean memory */
				memset( sb1, 0, HEAPSIZE );
				memset( sb2, 0, HEAPSIZE );
				memset( db1, 0, HEAPSIZE );
				memset( db2, 0, HEAPSIZE );

				/* Reset GDMA */
				WRITE_MEM32( GDMACNR, 0 );
				WRITE_MEM32( GDMACNR, GDMA_ENABLE );
				
				memcpy( sb2+*offset, testvec[j].source, testvec[j].len );
				WRITE_MEM32( GDMAIMR, 0 );
				WRITE_MEM32( GDMAISR, 0xffffffff );
				WRITE_MEM32( GDMAICVL, testvec[j].icv );
				WRITE_MEM32( GDMAICVR, 0 );
				WRITE_MEM32( GDMASBP0, (uint32)sb2+0x0c+*offset ); /* SIP + DIP */
				if ( (testvec[j].len-20) > 0 )
				{
					WRITE_MEM32( GDMASBL0, 8 );
					WRITE_MEM32( GDMASBP1, (uint32)sb2+0x14+*offset ); /* UDP + payload */
					WRITE_MEM32( GDMASBL1, GDMA_LDB|(testvec[j].len-20) );
				}
				else
				{
					WRITE_MEM32( GDMASBL0, GDMA_LDB|8 );
				}
				WRITE_MEM32( GDMADBP0, 0 );
				WRITE_MEM32( GDMADBL0, 0 );
				WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_CHKOFF );

				/* Expected data */
				memcpy( sb1+*offset, testvec[j].source, testvec[j].len );
				while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

				IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
				IS_EQUAL_INT("Source block should under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, GDMA_SBUDRIP, __FUNCTION__, __LINE__ );
				IS_EQUAL_INT("Checksum not equal!", READ_MEM32(GDMAICVL), testvec[j].sum, __FUNCTION__, __LINE__ );
				ASSERT( compare_data_block() == SUCCESS );
			}
		}
	}

	return SUCCESS;
}


/******************************************************************************
 ******************************************************************************
 **                                                                          **
 **  Checksum Offload                                                        **
 **                                                                          **
 ******************************************************************************
 ******************************************************************************/
/*
 * Test fixed pattern
 */
int32 testGdmaSequentialTCAM(uint32 caseNo)
{
	enum IC_TYPE i;
	int j;
	int32 retval;
	struct TEST_VECTOR
	{
		uint32 entSize; /* unit: byte */
		uint8 dst[32];
		uint8 src[64*8];
		int count;
		int32 match; /* -1 means no match */
	}
	static testvec[] =
	{
		{
			entSize: 4,
			dst: { 192,168,154,254 },
			src: { 192,168,154,241,255,255,255,255,
			       192,168,154,242,255,255,255,255,
			       192,168,251,  0,255,255,255,  0,
			       192,168,154,  0,255,255,255,  0,
			         0,  0,  0,  0,  0,  0,  0,  0,
			     },
			count: 5,
			match: 3,
		},
		{
			entSize: 8,
			dst: { 192,168,154,141,                192,168, 87,  1 },
			src: { 192,168,154,241,255,255,255,255,192,168,154,242,255,255,255,255,
			       192,168,154,254,255,255,255,254,192,168, 87,111,255,255,255,254,
			       192,168,251,  0,255,255,255,  0,192,168, 87,  0,255,255,255,  0,
			       192,168,154,  0,255,255,255,  0,192,168,251,  0,255,255,255,  0,
			       192,168,154,254,255,255,255,  0,192,168, 87,111,255,255,255,  0,
			       192,168,111,111,255,255,  0,  0,192,168,111,111,255,255,  0,  0,
			     },
			count: 6,
			match: 4,
		},
		{
			entSize: 16,
			dst: { 0x20,0x02,0x8c,0x71,                     0x1b,0x32,0x00,0x00,                     0x00,0x00,0x00,0x00,                     0x8c,0x71,0x1b,0x32 },
			src: { 0xfe,0x80,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*Link-Local*/
			       0xfe,0xC0,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*Site-Local*/
			       0x20,0x01,0x00,0x00,0xff,0xff,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*Commercial*/
			       0x3F,0xFE,0x00,0x00,0xff,0xff,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*For Experiment*/
			       0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x01,0xff,0xff,0xff,0xff, /*Localhost*/
			       0x20,0x02,0x8c,0x71,0xff,0xff,0xff,0xff, 0x1b,0x32,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*6-to-4 Auto Tunnel*/
			       0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*Any Address*/
			     },
			count: 7,
			match: 5,
		},
		{
			entSize: 32,
			dst: { 0x20,0x02,0x8c,0x71,                     0x1b,0x32,0x00,0x00,                     0x00,0x00,0x00,0x00,                     0x8c,0x71,0x1b,0x32,                     0x20,0x02,0x8c,0x71,                     0x1b,0x32,0x00,0x00,                     0x00,0x00,0x00,0x00,                     0x55,0xaa,0x77,0x88 },
			src: { 0xfe,0x80,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0xfe,0x80,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*Link-Local*/
			       0xfe,0xC0,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0xfe,0xC0,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*Site-Local*/
			       0x20,0x01,0x00,0x00,0xff,0xff,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x20,0x01,0x00,0x00,0xff,0xff,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*Commercial*/
			       0x3F,0xFE,0x00,0x00,0xff,0xff,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x3F,0xFE,0x00,0x00,0xff,0xff,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*For Experiment*/
			       0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x01,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff, 0x00,0x00,0x00,0x01,0xff,0xff,0xff,0xff, /*Localhost*/
			       0x20,0x02,0x8c,0x71,0xff,0xff,0xff,0xff, 0x1b,0x32,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x20,0x02,0x8c,0x71,0xff,0xff,0xff,0xff, 0x1b,0x32,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x55,0xaa,0x77,0x88,0xff,0xff,0xff,0xff, /*6-to-4 Auto Tunnel*/
			       0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, /*Any Address*/
			     },
			count: 7,
			match: 5,
		},
	};

	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		for( j = 0; j < sizeof(testvec)/sizeof(testvec[0]); j++ )
		{
			/* clean memory */
			memset( sb1, 0, HEAPSIZE );
			memset( sb2, 0, HEAPSIZE );
			memset( db1, 0, HEAPSIZE );
			memset( db2, 0, HEAPSIZE );

			/* Reset GDMA */
			WRITE_MEM32( GDMACNR, 0 );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE );
			
			memcpy( sb2, testvec[j].src, (testvec[j].count*testvec[j].entSize*2) );
			WRITE_MEM32( GDMAIMR, 0 );
			WRITE_MEM32( GDMAISR, 0xffffffff );
			WRITE_MEM32( GDMAICVL, 0 );
			WRITE_MEM32( GDMAICVR, 0 );
			WRITE_MEM32( GDMASBP0, (uint32)sb2 );
			WRITE_MEM32( GDMASBL0, GDMA_LDB|(testvec[j].count*testvec[j].entSize*2) );
			WRITE_MEM32( GDMADBP0, big_endian(((uint32*)testvec[j].dst)[0]) );
			WRITE_MEM32( GDMADBL0, 0 );
			WRITE_MEM32( GDMADBP1, big_endian(((uint32*)testvec[j].dst)[1]) );
			WRITE_MEM32( GDMADBL1, 0 );
			WRITE_MEM32( GDMADBP2, big_endian(((uint32*)testvec[j].dst)[2]) );
			WRITE_MEM32( GDMADBL2, 0 );
			WRITE_MEM32( GDMADBP3, big_endian(((uint32*)testvec[j].dst)[3]) );
			WRITE_MEM32( GDMADBL3, 0 );
			WRITE_MEM32( GDMADBP4, big_endian(((uint32*)testvec[j].dst)[4]) );
			WRITE_MEM32( GDMADBL4, 0 );
			WRITE_MEM32( GDMADBP5, big_endian(((uint32*)testvec[j].dst)[5]) );
			WRITE_MEM32( GDMADBL5, 0 );
			WRITE_MEM32( GDMADBP6, big_endian(((uint32*)testvec[j].dst)[6]) );
			WRITE_MEM32( GDMADBL6, 0 );
			WRITE_MEM32( GDMADBP7, big_endian(((uint32*)testvec[j].dst)[7]) );
			WRITE_MEM32( GDMADBL7, 0 );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_STCAM| (testvec[j].entSize==32?GDMA_ENTSIZ256:
			                                                         testvec[j].entSize==16?GDMA_ENTSIZ128:
			                                                         testvec[j].entSize== 8?GDMA_ENTSIZ64:GDMA_ENTSIZ32) );

			/* Expected data */
			memcpy( sb1, testvec[j].src, (testvec[j].count*testvec[j].entSize*2) );
			while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

			IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
			if ( testvec[j].match >= 0 )
			{
				/* should matched !!! */
				IS_EQUAL_INT("Source block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, 0, __FUNCTION__, __LINE__ );
				IS_EQUAL_INT("Match Index is not right!", READ_MEM32(GDMAICVL), testvec[j].match, __FUNCTION__, __LINE__ );
			}
			else
			{
				/* should NOT matched */
				IS_EQUAL_INT("Source block should under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, GDMA_SBUDRIP, __FUNCTION__, __LINE__ );
			}
			ASSERT( compare_data_block() == SUCCESS );
		}
	}

	return SUCCESS;
}


/******************************************************************************
 ******************************************************************************
 **                                                                          **
 **  Wireless MIC                                                            **
 **                                                                          **
 ******************************************************************************
 ******************************************************************************/
/*
 * Test fixed pattern
 */
int32 testGdmaWirelessMIC(uint32 caseNo)
{
	struct TEST_VECTOR
	{
		uint32 icvl;
		uint32 icvr;
		char *source;
		uint32 len;
		uint32 outl;
		uint32 outr;
	}
	static testvec[] =
	{
		{
			icvl: 0x000000000,
			icvr: 0x000000000,
			source: "\x5a\x00\x00\x00\x00\x00\x00\x00",
			len: 8,
			outl: 0x1c5c9282,
			outr: 0xb830d1a1,
		},
		{
			icvl: 0x1c5c9282,
			icvr: 0xb830d1a1,
			source: "M\x5a\x00\x00\x00\x00\x00\x00",
			len: 8,
			outl: 0xca214743,
			outr: 0x3f9b6340,
		},
		{
			icvl: 0xca214743,
			icvr: 0x3f9b6340,
			source: "Mi\x5a\x00\x00\x00\x00\x00",
			len: 8,
			outl: 0xcabef9e8,
			outr: 0x295d7ee9,
		},
		{
			icvl: 0xcabef9e8,
			icvr: 0x295d7ee9,
			source: "Mic\x5a\x00\x00\x00\x00",
			len: 8,
			outl: 0xc68f0390,
			outr: 0xdbc113cf,
		},
		{
			icvl: 0xc68f0390,
			icvr: 0xdbc113cf,
			source: "Mich\x5a\x00\x00\x00\x00\x00\x00\x00",
			len: 12,
			outl: 0x05105eD5,
			outr: 0x86891210,
		},
		{
			icvl: 0x05105eD5,
			icvr: 0x86891210,
			source: "Michael\x5a\x00\x00\x00\x00",
			len: 12,
			outl: 0x122b940a,
			outr: 0x46a5ca4e,
		},
	};
	int i, j;
	int32 retval;
	
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		for( j = 0; j < sizeof(testvec)/sizeof(testvec[0]); j++ )
		{
			int _offset[] = { 0, -1 };
			int *offset;
			
			for( offset = _offset; *offset>=0; offset++ )
			{
				/* clean memory */
				memset( sb1, 0, HEAPSIZE );
				memset( sb2, 0, HEAPSIZE );
				memset( db1, 0, HEAPSIZE );
				memset( db2, 0, HEAPSIZE );

				/* Reset GDMA */
				WRITE_MEM32( GDMACNR, 0 );
				WRITE_MEM32( GDMACNR, GDMA_ENABLE );
				
				memcpy( sb2+*offset, testvec[j].source, testvec[j].len );
				WRITE_MEM32( GDMAIMR, 0 );
				WRITE_MEM32( GDMAISR, 0xffffffff );
				WRITE_MEM32( GDMAICVL, testvec[j].icvl );
				WRITE_MEM32( GDMAICVR, testvec[j].icvr );
				WRITE_MEM32( GDMASBP0, (uint32)sb2+*offset ); /* SIP + DIP */
				WRITE_MEM32( GDMASBL0, GDMA_LDB|testvec[j].len );
				WRITE_MEM32( GDMADBP0, 0 );
				WRITE_MEM32( GDMADBL0, 0 );
				WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_MIC );

				/* Expected data */
				memcpy( sb1+*offset, testvec[j].source, testvec[j].len );
				while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */
				IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
				IS_EQUAL_INT("Source block should under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, GDMA_SBUDRIP, __FUNCTION__, __LINE__ );
				IS_EQUAL_INT("Checksum not equal!", READ_MEM32(GDMAICVL), testvec[j].outl, __FUNCTION__, __LINE__ );
				IS_EQUAL_INT("Checksum not equal!", READ_MEM32(GDMAICVR), testvec[j].outr, __FUNCTION__, __LINE__ );
				ASSERT( compare_data_block() == SUCCESS );
			}
		}
	}

	return SUCCESS;
}


/******************************************************************************
 ******************************************************************************
 **                                                                          **
 **  Base 64                                                                 **
 **                                                                          **
 ******************************************************************************
 ******************************************************************************/
/*
 * Test fixed pattern
 */
int32 testGdmaBase64Encode1(uint32 caseNo)
{
	struct TEST_VECTOR
	{
		char *decoded;
		char *encoded;
	}
	static testvec[] =
	{
		{
			decoded: "\x01\x02\x04",
			encoded: "\x41\x51\x49\x45",
		},
		{
			decoded: "\n",
			encoded: "Cg==",
		},
		{
			decoded: "01",
			encoded: "MDE=",
		},
		{
			decoded: "\xFF\xFF\xFF\xFF",
			encoded: "/////w==",
		},
		{
			decoded: "This is a test for base64\n",
			encoded: "VGhpcyBpcyBhIHRlc3QgZm9yIGJhc2U2NAo=",
		},
		{
			decoded: "\
#include<stdio.h>\n\
#include<math.h>\n\
\n\
int main( int argc, char** argv )\n\
{\n\
	double dR;\n\
	double total = 0;\n\
	double i;\n\
\n\
	dR = atof( argv[1] );\n\
	printf(\"R=%lf\\n\", dR );\n\
\n\
	for( i = 1; i <= 50; i++ )\n\
	{\n\
		total += 6600 / ( pow( 1 + (dR/10000)/12.0, i ) );\n\
	}\n\
\n\
	printf(\"total=%lf\\n\", total );\n\
\n\
	return 0;\n\
}\n",
			encoded: "\
I2luY2x1ZGU8c3RkaW8uaD4KI2luY2x1ZGU8bWF0aC5oPgoKaW50IG1haW4oIGludCBhcmdjLCBj\r\n\
aGFyKiogYXJndiApCnsKCWRvdWJsZSBkUjsKCWRvdWJsZSB0b3RhbCA9IDA7Cglkb3VibGUgaTsK\r\n\
CglkUiA9IGF0b2YoIGFyZ3ZbMV0gKTsKCXByaW50ZigiUj0lbGZcbiIsIGRSICk7CgoJZm9yKCBp\r\n\
ID0gMTsgaSA8PSA1MDsgaSsrICkKCXsKCQl0b3RhbCArPSA2NjAwIC8gKCBwb3coIDEgKyAoZFIv\r\n\
MTAwMDApLzEyLjAsIGkgKSApOwoJfQoKCXByaW50ZigidG90YWw9JWxmXG4iLCB0b3RhbCApOwoK\r\n\
CXJldHVybiAwOwp9Cg==",
		},
	};
	int i, j;
	int32 retval;
	int maxLen = 76;
	
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		for( j = 0; j < sizeof(testvec)/sizeof(testvec[0]); j++ )
		{
			/* clean memory */
			memset( sb1, 0, HEAPSIZE );
			memset( sb2, 0, HEAPSIZE );
			memset( db1, 0, HEAPSIZE );
			memset( db2, 0, HEAPSIZE );

			/* Reset GDMA */
			WRITE_MEM32( GDMACNR, 0 );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE );
				
			memcpy( sb2, testvec[j].decoded, strlen(testvec[j].decoded) );
			WRITE_MEM32( GDMAIMR, 0 );
			WRITE_MEM32( GDMAISR, 0xffffffff );
			WRITE_MEM32( GDMAICVL, maxLen/4-1 ); /* Line Length */
			WRITE_MEM32( GDMAICVR, 0 );
			WRITE_MEM32( GDMASBP0, (uint32)sb2 );
			WRITE_MEM32( GDMASBL0, GDMA_LDB|strlen(testvec[j].decoded) );
			WRITE_MEM32( GDMADBP0, (uint32)db2 );
#if 1
			WRITE_MEM32( GDMADBL0, GDMA_LDB|(strlen(testvec[j].encoded)) );
#else
			/* old fashion to compute output length */
			WRITE_MEM32( GDMADBL0, GDMA_LDB|(strlen(testvec[j].decoded)*4/3+(strlen(testvec[j].decoded)*4%3?3:0)+strlen(testvec[j].decoded)*4/3/maxLen*2) );
#endif
			WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_B64ENC );

			/* Expected data */
			memcpy( sb1, testvec[j].decoded, strlen(testvec[j].decoded) );
			memcpy( db1, testvec[j].encoded, strlen(testvec[j].encoded) );

			while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

			IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Source block should under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, GDMA_SBUDRIP, __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Processed length not match!", READ_MEM32(GDMAICVL)>>16, strlen(testvec[j].decoded), __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Output length not match!", READ_MEM32(GDMAICVL)&0xffff, strlen(testvec[j].encoded), __FUNCTION__, __LINE__ );
			ASSERT( compare_data_block() == SUCCESS );
		}
	}

	return SUCCESS;
}


/*
 * Test fixed pattern
 */
int32 testGdmaBase64Decode1(uint32 caseNo)
{
	struct TEST_VECTOR
	{
		char *decoded;
		char *encoded;
	}
	static testvec[] =
	{
		{
			decoded: "\x01\x02\x04",
			encoded: "\x41\x51\x49\x45",
		},
		{
			decoded: "\n",
			encoded: "Cg=",
		},
		{
			decoded: "01",
			encoded: "MDE=",
		},
		{
			decoded: "\x01\x02\x04",
			encoded: "=\x41===\x51-----\x49*****\x45",
		},
		{
			decoded: "\x01\x02\x04",
			encoded: "-----\x41()\x51\\\x49\x45*****\r\n",
		},
		{
			decoded: "\n",
			encoded: "=Cg=",
		},
		{
			decoded: "\n",
			encoded: "=C=g=",
		},
		{
			decoded: "\n",
			encoded: "=#C---=g=",
		},
		{
			decoded: "01",
			encoded: "MDE=",
		},
		{
			decoded: "01",
			encoded: "M=---D---E=",
		},
		{
			decoded: "\xFF\xFF\xFF\xFF",
			encoded: "/////w=",
		},
		{
			decoded: "\xFF\xFF\xFF\xFF",
			encoded: "\\/\\/\\/\\/\\/w=",
		},
		{
			decoded: "\x04\x10\x41\x08\x20\x82",
			encoded: "BBBB=CCCC",
		},
		{
			decoded: "\x04\x10\x41\x08\x20\x82",
			encoded: "BBBB==CCCC",
		},
		{
			decoded: "\x04\x10\x41\x08\x20\x82",
			encoded: "BBBBC=CCC",
		},
		{
			decoded: "\x04\x10\x41\x08\x20\x82",
			encoded: "BBBBC==CCC",
		},
		{
			decoded: "\x04\x10\x41\x08\x20\x82",
			encoded: "BBBB=====CCCC",
		},
		{
			decoded: "This is a test for base64\n",
			encoded: "VGhpcyBpcyBhIHRlc3QgZm9yIGJhc2U2NAo=",
		},
		{
			decoded: "\
#include<stdio.h>\n\
#include<math.h>\n\
\n\
int main( int argc, char** argv )\n\
{\n\
	double dR;\n\
	double total = 0;\n\
	double i;\n\
\n\
	dR = atof( argv[1] );\n\
	printf(\"R=%lf\\n\", dR );\n\
\n\
	for( i = 1; i <= 50; i++ )\n\
	{\n\
		total += 6600 / ( pow( 1 + (dR/10000)/12.0, i ) );\n\
	}\n\
\n\
	printf(\"total=%lf\\n\", total );\n\
\n\
	return 0;\n\
}\n",
			encoded: "\
I2luY2x1ZGU8c3RkaW8uaD4KI2luY2x1ZGU8bWF0aC5oPgoKaW50IG1haW4oIGludCBhcmdjLCBj\r\n\
aGFyKiogYXJndiApCnsKCWRvdWJsZSBkUjsKCWRvdWJsZSB0b3RhbCA9IDA7Cglkb3VibGUgaTsK\r\n\
CglkUiA9IGF0b2YoIGFyZ3ZbMV0gKTsKCXByaW50ZigiUj0lbGZcbiIsIGRSICk7CgoJZm9yKCBp\r\n\
ID0gMTsgaSA8PSA1MDsgaSsrICkKCXsKCQl0b3RhbCArPSA2NjAwIC8gKCBwb3coIDEgKyAoZFIv\r\n\
MTAwMDApLzEyLjAsIGkgKSApOwoJfQoKCXByaW50ZigidG90YWw9JWxmXG4iLCB0b3RhbCApOwoK\r\n\
CXJldHVybiAwOwp9Cg=",
		},
	};
	int i, j;
	int32 retval;
	int maxLen = 76;
	
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		for( j = 0; j < sizeof(testvec)/sizeof(testvec[0]); j++ )
		{
			/* clean memory */
			memset( sb1, 0, HEAPSIZE );
			memset( sb2, 0, HEAPSIZE );
			memset( db1, 0, HEAPSIZE );
			memset( db2, 0, HEAPSIZE );
			
			/* Reset GDMA */
			WRITE_MEM32( GDMACNR, 0 );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE );
				
			memcpy( sb2, testvec[j].encoded, strlen(testvec[j].encoded) );
			WRITE_MEM32( GDMAIMR, 0 );
			WRITE_MEM32( GDMAISR, 0xffffffff );
			WRITE_MEM32( GDMAICVL, maxLen/4-1 ); /* Line Length */
			WRITE_MEM32( GDMAICVR, 0 );
			WRITE_MEM32( GDMASBP0, (uint32)sb2 );
			WRITE_MEM32( GDMASBL0, GDMA_LDB|strlen(testvec[j].encoded) );
			WRITE_MEM32( GDMADBP0, (uint32)db2 );
			WRITE_MEM32( GDMADBL0, GDMA_LDB|strlen(testvec[j].decoded) );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_B64DEC );

			/* Expected data */
			memcpy( sb1, testvec[j].encoded, strlen(testvec[j].encoded) );
			memcpy( db1, testvec[j].decoded, strlen(testvec[j].decoded) );

			while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

			/*memDump( sb2, strlen(testvec[j].encoded), "sd2" );*/
			/*memDump( db2, strlen(testvec[j].decoded), "dd2" );*/
			if ( testvec[j].encoded[strlen(testvec[j].encoded)-1] == '=' )
				IS_EQUAL_INT("Source block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, 0, __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Processed length not match!", READ_MEM32(GDMAICVL)>>16, strlen(testvec[j].encoded), __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Output length not match!", READ_MEM32(GDMAICVL)&0xffff, strlen(testvec[j].decoded), __FUNCTION__, __LINE__ );
			ASSERT( compare_data_block() == SUCCESS );
		}
	}

	return SUCCESS;
}


/*
 * Test fixed pattern
 */
int32 testGdmaQuotedPrintableEncode1(uint32 caseNo)
{
	struct TEST_VECTOR
	{
		char *decoded;
		char *encoded;
	}
	static testvec[] =
	{
		{
			decoded: "=",
			encoded: "=3D",
		},
		{
			decoded: "0123456789012345678901234567890123456789012345678901234567890123456789012\xAA",
			encoded: "0123456789012345678901234567890123456789012345678901234567890123456789012=\r\n=AA",
		},
		{
			decoded: "01234567890123456789012345678901234567890123456789012345678901234567890123\xAA",
			encoded: "01234567890123456789012345678901234567890123456789012345678901234567890123=\r\n=AA",
		},
		{
			decoded: "012345678901234567890123456789012345678901234567890123456789012345678901234\xAA",
			encoded: "012345678901234567890123456789012345678901234567890123456789012345678901234=\r\n=AA",
		},
		{
			decoded: "\n0123456789012345678901234567890123456789012345678901234567890123456789012345\r0123456789012345678901234567890123456789012345678901234567890123456789012345",
			encoded: "\n012345678901234567890123456789012345678901234567890123456789012345678901234=\r\n5\r012345678901234567890123456789012345678901234567890123456789012345678901234=\r\n5",
		},
		{
			decoded: "Ttq",
			encoded: "=A4T=B0=EA=BAt=B8q",
		},
		{
			decoded: "\x0a\x0d\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f""0123456789\x3a\x3b\x3c\x3e\x3f\x40""ABCDEFGHIJKLMNOPQRSTUVWXYZ\x5b\x5c\x5d\x5e\x5f\x60""abcdefghijklmnopqrstuvwxyz\x7b\x7c\x7d\x7e",
			encoded: "\x0a\x0d\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f""0123456789\x3a\x3b\x3c\x3e\x3f\x40""ABCDEFGHIJKLMNOPQRSTUVWXYZ\x5b\x5c\x5d\x5e\x5f\x60""abcdefghijkl=\r\nmnopqrstuvwxyz\x7b\x7c\x7d\x7e",
		},
		{
			decoded: "Fw:  [K] x178y!?",
			encoded: "Fw:=20=20[=B1K=A7=DE]=20=A7=EB=C4x=BE=F71=A4=C0=C4=C178=C1=FB=B2y!?",
		},
		{
			decoded: "\
rFoɤjs@nĸix̡A}lΥkƨgⴡxءI@}ljaQLo|~F@ADLQF¡HuOıoL欰ITC\r\n\
oɧڭťxƨgoXixġu??B??B??vA٬ݨWyܡA10B20B25B34o˺ƨgɡIڭ̤~DLbFCطPıuöέpqݮ]Ū԰OƭȧֳtɡC\r\n\
rFkⴡ֤FAⴡ֤FSkAڭ̦b~YgA٤ѴLGuѤU5FI5B4B3B2.............vjaƨ2ɡArFMjs@nG u@~~~~~~~~~~I(GhF~~~~~~~~~~I)v\r\n\
Mɶ@knbxظ̡CrFoɳݩIIݤjaGuˡIF`aIv\r\n",
			encoded: "\
=ACr=BDF=B3o=AE=C9=A4j=A5s=A4@=C1n=BD=C4=B8=F5=B6i=A7=EB=C4x=BE=F7=B8=CC=A1=\r\n\
A=B6}=A9l=A5=CE=A5k=A4=E2=BA=C6=A8g=A9=E2=B4=A1=C4x=AE=D8=A1I=A4@=B6}=A9l=\r\n\
=A4j=AEa=B3Q=A5L=B3o=AD=D3=C1|=B0=CA=C0~=A4F=A4@=B8=F5=A1A=A4=A3=AA=BE=B9D=\r\n\
=A5L=B7Q=B7F=B3=C2=A1H=A5u=ACO=C4=B1=B1o=A5L=AA=BA=A6=E6=AC=B0=A6=B3=C2I=B5=\r\n\
T=C1=B6=A1C\r\n\
=B3o=AE=C9=A7=DA=AD=CC=C5=A5=A8=EC=A7=EB=C4x=BE=F7=BA=C6=A8g=B5o=A5X=B6i=C4=\r\n\
x=AA=BA=AD=B5=AE=C4=A1u?\?=A1B?\?=A1B?\?=A1v=A1A=C1=D9=AC=DD=A8=EC=A4W=AD=B1=\r\n\
=AA=BA=B2y=BC=C6=C5=E3=A5=DC=A1A10=C1=FB=A1B20=C1=FB=A1B25=C1=FB=A1B34=C1=\r\n\
=FB=B3o=BC=CB=BA=C6=A8g=B4=A3=A4=C9=A1I=A7=DA=AD=CC=A4~=AA=BE=B9D=A5L=A6b=\r\n\
=B7F=B9=C0=A1C=A8=BA=BA=D8=B7P=C4=B1=AFu=A6=B3=B0=F7=B9=B3=B0=F2=AF=C3=B6=\r\n\
=A4=AA=F8=A5=CE=ADp=B6q=BE=B9=AC=DD=AE]=AE=A9=AA=C5=AA=BA=BE=D4=B0=AB=A4O=\r\n\
=BC=C6=AD=C8=A7=D6=B3t=B4=A3=A4=C9=A1C\r\n\
=ACr=BDF=A5k=A4=E2=B4=A1=B2=D6=A4F=B4=AB=A5=AA=A4=E2=A1A=A5=AA=A4=E2=B4=A1=\r\n\
=B2=D6=A4F=A4S=B4=AB=A5k=A4=E2=A1A=A7=DA=AD=CC=A6b=A5~=C0Y=A4=A3=B0=B1=A8g=\r\n\
=AF=BA=A1A=C1=D9=A4=A3=A7=D1=B4=A3=BF=F4=A5L=A1G=A1u=B3=D1=A4U5=AC=ED=A4F=\r\n\
=A1I5=A1B4=A1B3=A1B2.............=A1v=A5=BF=B7=ED=A4j=AEa=BC=C6=A8=EC2=AE=\r\n\
=C9=A1A=ACr=BDF=AC=F0=B5M=A4j=A5s=A4@=C1n=A1G=20=A1u=A4@=AD=FA=B3=E9~~~~~~~=\r\n\
~~~=A1I(=A4=E9=A4=E5=A1G=A5h=A4F~~~~~~~~~~=A1I)=A1v\r\n\
=B5M=AB=E1=AE=C9=B6=A1=A4@=A8=EC=A5k=A4=E2=B3n=A6b=C4x=AE=D8=B8=CC=A1C=ACr=\r\n\
=BDF=B3o=AE=C9=B3=DD=A9I=A9I=AA=BA=B0=DD=A4j=AEa=A1G=A1u=AB=E7=BB=F2=BC=CB=\r\n\
=A1I=BCF=AE`=A7a=A1I=A1v\r\n\
",
		},
		{	/* 75 spaces + 76 spaces + 77 spaces */
			decoded: "\
                                                                           \r\n\
                                                                            \r\n\
                                                                             \r\n\
",
			encoded: "\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20\r\n\
",
		},
	};
	int i, j;
	int32 retval;
	int maxLen = 76;
	
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		for( j = 0; j < sizeof(testvec)/sizeof(testvec[0]); j++ )
		{
			/* clean memory */
			memset( sb1, 0, HEAPSIZE );
			memset( sb2, 0, HEAPSIZE );
			memset( db1, 0, HEAPSIZE );
			memset( db2, 0, HEAPSIZE );
			
			/* Reset GDMA */
			WRITE_MEM32( GDMACNR, 0 );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE );
				
			memcpy( sb2, testvec[j].decoded, strlen(testvec[j].decoded) );
			WRITE_MEM32( GDMAIMR, 0 );
			WRITE_MEM32( GDMAISR, 0xffffffff );
			WRITE_MEM32( GDMAICVL, maxLen/4-1 ); /* Line Length */
			WRITE_MEM32( GDMAICVR, 0 );
			WRITE_MEM32( GDMASBP0, (uint32)sb2 );
			WRITE_MEM32( GDMASBL0, GDMA_LDB|strlen(testvec[j].decoded) );
			WRITE_MEM32( GDMADBP0, (uint32)db2 );
			WRITE_MEM32( GDMADBL0, GDMA_LDB|(strlen(testvec[j].encoded)) );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_QPENC );

			/* Expected data */
			memcpy( sb1, testvec[j].decoded, strlen(testvec[j].decoded) );
			memcpy( db1, testvec[j].encoded, strlen(testvec[j].encoded) );

			while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

			/*memDump( sb1, strlen(testvec[j].decoded), "sd1" );*/
			/*memDump( sb2, strlen(testvec[j].decoded), "sd2" );*/
			/*memDump( db1, strlen(testvec[j].encoded), "dd1" );*/
			/*memDump( db2, strlen(testvec[j].encoded), "dd2" );*/
			IS_EQUAL_INT("Source block should under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, GDMA_SBUDRIP, __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Processed length not match!", READ_MEM32(GDMAICVL)>>16, strlen(testvec[j].decoded), __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Output length not match!", READ_MEM32(GDMAICVL)&0xffff, strlen(testvec[j].encoded), __FUNCTION__, __LINE__ );
			ASSERT( compare_data_block() == SUCCESS );
		}
	}

	return SUCCESS;
}

/*
 * Test fixed pattern
 */
int32 testGdmaQuotedPrintableDecode1(uint32 caseNo)
{
	struct TEST_VECTOR
	{
		char *decoded;
		char *encoded;
	}
	static testvec[] =
	{
		{
			decoded: "=",
			encoded: "=3D",
		},
		{
			decoded: "0\r\n12",
			encoded: "0\r\n12",
		},
		{
			decoded: "0123456789ABCDEF",
			encoded: "0=\r\n12=\r\n345=\r\n6789=\r\nABCDEF",
		},
		{
			decoded: "AB",
			encoded: "A=\r\nB",
		},
		{
			decoded: "Ttq",
			encoded: "=A4T=B0=EA=BAt=B8q",
		},
		{
			decoded: "\x0a\x0d\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f""0123456789\x3a\x3b\x3c\x3e\x3f\x40""ABCDEFGHIJKLMNOPQRSTUVWXYZ\x5b\x5c\x5d\x5e\x5f\x60""abcdefghijklmnopqrstuvwxyz\x7b\x7c\x7d\x7e",
			encoded: "\x0a\x0d\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f""0123456789\x3a\x3b\x3c\x3e\x3f\x40""ABCDEFGHIJKLMNOPQRSTUVWXYZ\x5b\x5c\x5d\x5e\x5f\x60""abcdefghijkl=\r\nmnopqrstuvwxyz\x7b\x7c\x7d\x7e",
		},
		{
			decoded: "Fw:  [K] x178y!?",
			encoded: "Fw:=20=20[=B1K=A7=DE]=20=A7=EB=C4x=BE=F71=A4=C0=C4=C178=C1=FB=B2y!?",
		},
		{
			decoded: "\
rFoɤjs@nĸix̡A}lΥkƨgⴡxءI@}ljaQLo|~F@ADLQF¡HuOıoL欰ITC\r\n\
oɧڭťxƨgoXixġu??B??B??vA٬ݨWyܡA10B20B25B34o˺ƨgɡIڭ̤~DLbFCطPıuöέpqݮ]Ū԰OƭȧֳtɡC\r\n\
rFkⴡ֤FAⴡ֤FSkAڭ̦b~YgA٤ѴLGuѤU5FI5B4B3B2.............vjaƨ2ɡArFMjs@nG u@~~~~~~~~~~I(GhF~~~~~~~~~~I)v\r\n\
Mɶ@knbxظ̡CrFoɳݩIIݤjaGuˡIF`aIv\r\n",
			encoded: "\
=ACr=BDF=B3o=AE=C9=A4j=A5s=A4@=C1n=BD=C4=B8=F5=B6i=A7=EB=C4x=BE=F7=B8=CC=A1=\r\n\
A=B6}=A9l=A5=CE=A5k=A4=E2=BA=C6=A8g=A9=E2=B4=A1=C4x=AE=D8=A1I=A4@=B6}=A9l=\r\n\
=A4j=AEa=B3Q=A5L=B3o=AD=D3=C1|=B0=CA=C0~=A4F=A4@=B8=F5=A1A=A4=A3=AA=BE=B9D=\r\n\
=A5L=B7Q=B7F=B3=C2=A1H=A5u=ACO=C4=B1=B1o=A5L=AA=BA=A6=E6=AC=B0=A6=B3=C2I=B5=\r\n\
T=C1=B6=A1C\r\n\
=B3o=AE=C9=A7=DA=AD=CC=C5=A5=A8=EC=A7=EB=C4x=BE=F7=BA=C6=A8g=B5o=A5X=B6i=C4=\r\n\
x=AA=BA=AD=B5=AE=C4=A1u?\?=A1B?\?=A1B?\?=A1v=A1A=C1=D9=AC=DD=A8=EC=A4W=AD=B1=\r\n\
=AA=BA=B2y=BC=C6=C5=E3=A5=DC=A1A10=C1=FB=A1B20=C1=FB=A1B25=C1=FB=A1B34=C1=\r\n\
=FB=B3o=BC=CB=BA=C6=A8g=B4=A3=A4=C9=A1I=A7=DA=AD=CC=A4~=AA=BE=B9D=A5L=A6b=\r\n\
=B7F=B9=C0=A1C=A8=BA=BA=D8=B7P=C4=B1=AFu=A6=B3=B0=F7=B9=B3=B0=F2=AF=C3=B6=\r\n\
=A4=AA=F8=A5=CE=ADp=B6q=BE=B9=AC=DD=AE]=AE=A9=AA=C5=AA=BA=BE=D4=B0=AB=A4O=\r\n\
=BC=C6=AD=C8=A7=D6=B3t=B4=A3=A4=C9=A1C\r\n\
=ACr=BDF=A5k=A4=E2=B4=A1=B2=D6=A4F=B4=AB=A5=AA=A4=E2=A1A=A5=AA=A4=E2=B4=A1=\r\n\
=B2=D6=A4F=A4S=B4=AB=A5k=A4=E2=A1A=A7=DA=AD=CC=A6b=A5~=C0Y=A4=A3=B0=B1=A8g=\r\n\
=AF=BA=A1A=C1=D9=A4=A3=A7=D1=B4=A3=BF=F4=A5L=A1G=A1u=B3=D1=A4U5=AC=ED=A4F=\r\n\
=A1I5=A1B4=A1B3=A1B2.............=A1v=A5=BF=B7=ED=A4j=AEa=BC=C6=A8=EC2=AE=\r\n\
=C9=A1A=ACr=BDF=AC=F0=B5M=A4j=A5s=A4@=C1n=A1G=20=A1u=A4@=AD=FA=B3=E9~~~~~~~=\r\n\
~~~=A1I(=A4=E9=A4=E5=A1G=A5h=A4F~~~~~~~~~~=A1I)=A1v\r\n\
=B5M=AB=E1=AE=C9=B6=A1=A4@=A8=EC=A5k=A4=E2=B3n=A6b=C4x=AE=D8=B8=CC=A1C=ACr=\r\n\
=BDF=B3o=AE=C9=B3=DD=A9I=A9I=AA=BA=B0=DD=A4j=AEa=A1G=A1u=AB=E7=BB=F2=BC=CB=\r\n\
=A1I=BCF=AE`=A7a=A1I=A1v\r\n\
",
		},
		{	/* 75 spaces + 76 spaces + 77 spaces */
			decoded: "\
                                                                           \r\n\
                                                                            \r\n\
                                                                             \r\n\
",
			encoded: "\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=20=\r\n\
=20=20\r\n\
",
		},
		{	/* Outlook: 75 spaces + 76 spaces + 77 spaces */
			decoded: "\
                                                                           \r\n\
                                                                            \r\n\
                                                                             \r\n\
",
			encoded: "\
                                                                         =\r\n\
 =20\r\n\
                                                                         =\r\n\
  =20\r\n\
                                                                         =\r\n\
   =20\r\n\
",
		},
		{	/* IRA News */
			decoded: "\
IRAŧG^Z˧ܪ ګww HB @ ͵CL    \r\n\
                          \r\n\
                  issD j s@i\r\n\
                                  \r\n\
                             \r\n\
                             \r\n\
                        \
HɤOqn_RΤ@uR@MxvIRAAx_ɶQѲ`]ŧGANM^FTQ~Z˧ܪAHFvPDͧP覡Bz_RDF^ꭺۥܺ(sDB)gIRAoMwu㦳ūevNqvAڶ]`wALֽפ]jաAIRAu@PvA~ҩL̽T꦳DͧP۷N]fݳɡ^\r\n\
",
			encoded: "\
IRA=AB=C5=A7G=B5=B2=A7=F4=B9=EF=AD^=AA=BA=AAZ=B8=CB=A7=DC=AA=A7 =\r\n\
=B0=EA=BB=DA=ABw=AA=ED=C5w=AA=EF =B1H=B5=B9=AAB=A4=CD =A1@ =\r\n\
=A4=CD=B5=BD=A6C=A6L   =20\r\n\
                         =20\r\n\
                  =A1i=A4=A4=BCs=B7s=BBD=BA=F4 =A1j =BCs=A1@=A7i\r\n\
                                 =20\r\n\
                            =20\r\n\
                            =20\r\n\
                        =\r\n\
=AA=F8=B4=C1=A5H=BC=C9=A4O=A4=E2=ACq=AA=A7=A8=FA=ABn=A5_=B7R=BA=B8=C4=F5=B2=\r\n\
=CE=A4@=AA=BA=A1u=B7R=BA=B8=C4=F5=A6@=A9M=ADx=A1vIRA=A1A=A5x=A5_=AE=C9=B6=\r\n\
=A1=ACQ=A4=D1=B2`=A9]=A5=BF=A6=A1=AB=C5=A7G=A1A=B1N=B5=B2=A7=F4=A9M=AD^=B0=\r\n\
=EA=AA=F8=B9F=A4T=A4Q=A6~=AA=BA=AAZ=B8=CB=A7=DC=AA=A7=A1A=A7=EF=A5H=ACF=AA=\r\n\
v=BBP=A5=C1=A5D=BD=CD=A7P=A4=E8=A6=A1=B3B=B2z=A5_=B7R=BA=B8=C4=F5=B0=DD=C3=\r\n\
D=A1F=AD^=B0=EA=AD=BA=AC=DB=A5=AC=B5=DC=BA=B8(=B7s=BBD=A1B=BA=F4=AF=B8)=C6=\r\n\
g=B4=ADIRA=AA=BA=B3o=B6=B5=A8M=A9w=A1u=A8=E3=A6=B3=AA=C5=ABe=AA=BA=BE=FA=A5=\r\n\
v=B7N=B8q=A1v=A1A=B0=EA=BB=DA=B6=A1=A4]=B2`=AA=ED=C5w=AA=EF=A1A=A4=A3=B9L=\r\n\
=B0=EA=BB=DA=C1=D6=BD=D7=A4]=B1j=BD=D5=A1AIRA=A5=B2=B6=B7=A1u=A8=A5=A6=E6=\r\n\
=A4@=ADP=A1v=A1A=A4~=AF=E0=C3=D2=A9=FA=A5L=AD=CC=BDT=B9=EA=A6=B3=A5=C1=A5=\r\n\
D=BD=CD=A7P=AA=BA=B8=DB=B7N=A1]=B8=AD=ACf=BC=DD=B3=F8=BE=C9=A1^\r\n\
",
		},
	};
	int i, j;
	int32 retval;
	int maxLen = 76;
	
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		for( j = 0; j < sizeof(testvec)/sizeof(testvec[0]); j++ )
		{
			/* clean memory */
			memset( sb1, 0, HEAPSIZE );
			memset( sb2, 0, HEAPSIZE );
			memset( db1, 0, HEAPSIZE );
			memset( db2, 0, HEAPSIZE );
			
			/* Reset GDMA */
			WRITE_MEM32( GDMACNR, 0 );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE );
				
			memcpy( sb2, testvec[j].encoded, strlen(testvec[j].encoded) );
			WRITE_MEM32( GDMAIMR, 0 );
			WRITE_MEM32( GDMAISR, 0xffffffff );
			WRITE_MEM32( GDMAICVL, maxLen/4-1 ); /* Line Length */
			WRITE_MEM32( GDMAICVR, 0 );
			WRITE_MEM32( GDMASBP0, (uint32)sb2 );
			WRITE_MEM32( GDMASBL0, GDMA_LDB|strlen(testvec[j].encoded) );
			WRITE_MEM32( GDMADBP0, (uint32)db2 );
			WRITE_MEM32( GDMADBL0, GDMA_LDB|(strlen(testvec[j].decoded)) );
			WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_QPDEC );

			/* Expected data */
			memcpy( sb1, testvec[j].encoded, strlen(testvec[j].encoded) );
			memcpy( db1, testvec[j].decoded, strlen(testvec[j].decoded) );

			while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

			/*memDump( sb1, strlen(testvec[j].encoded), "sd1" );*/
			/*memDump( sb2, strlen(testvec[j].encoded), "sd2" );*/
			/*memDump( db1, strlen(testvec[j].decoded), "dd1" );*/
			/*memDump( db2, strlen(testvec[j].decoded), "dd2" );*/
			IS_EQUAL_INT("Source block should under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, GDMA_SBUDRIP, __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Processed length not match!", READ_MEM32(GDMAICVL)>>16, strlen(testvec[j].encoded), __FUNCTION__, __LINE__ );
			IS_EQUAL_INT("Output length not match!", READ_MEM32(GDMAICVL)&0xffff, strlen(testvec[j].decoded), __FUNCTION__, __LINE__ );
			ASSERT( compare_data_block() == SUCCESS );
		}
	}

	return SUCCESS;
}


/******************************************************************************
 ******************************************************************************
 **                                                                          **
 **  Byte Swap                                                               **
 **                                                                          **
 ******************************************************************************
 ******************************************************************************/
int32 testGdmaByteSwap1(uint32 caseNo)
{
	int i, j;
	int32 retval;
	uint8 str[64] = "Hi, everybody. I am Louis. I come from Taiwan. Glad to meet you.";
	
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		/* clean memory */
		memset( sb1, 0, HEAPSIZE );
		memset( sb2, 0, HEAPSIZE );
		memset( db1, 0, HEAPSIZE );
		memset( db2, 0, HEAPSIZE );

		/* 0. init A with meaningful string */
		memcpy( sb2, str, sizeof(str) );

		/* Byte Swap */
		WRITE_MEM32( GDMACNR, 0 ); /* Reset GDMA */
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_MEMCPY );
		WRITE_MEM32( GDMAIMR, 0 );
		WRITE_MEM32( GDMAISR, 0xffffffff );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)sb2 );
		WRITE_MEM32( GDMASBL0, GDMA_LDB|sizeof(str) );
		WRITE_MEM32( GDMADBP0, (uint32)db2 );
		WRITE_MEM32( GDMADBL0, GDMA_LDB|sizeof(str) );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_BYTESWAP|GDMA_SWAPTYPE0);

		/* Expected data */
		memcpy( sb1, str, sizeof(str) );
		memcpy( db1, str, sizeof(str) );
		for( j = 0; j < sizeof(str); j+=2 )
		{
			uint8 tmp;
			tmp = db1[j];
			db1[j] = db1[j+1];
			db1[j+1] = tmp;
		}
		while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

		IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
		IS_EQUAL_INT("Source block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, 0, __FUNCTION__, __LINE__ );
		ASSERT( compare_data_block() == SUCCESS );
	}

	return SUCCESS;
}


int32 testGdmaByteSwap2(uint32 caseNo)
{
	int i, j;
	int32 retval;
	uint8 str[64] = "Hi, everybody. I am Louis. I come from Taiwan. Glad to meet you.";
	
	for( i = IC_TYPE_MIN; i < IC_TYPE_MAX; i++ )
	{
		retval = model_setTestTarget( i );
		if ( retval == VMAC_ERROR_NON_AVAILABLE ) continue; /* Not support, we don't compare. */

		/* clean memory */
		memset( sb1, 0, HEAPSIZE );
		memset( sb2, 0, HEAPSIZE );
		memset( db1, 0, HEAPSIZE );
		memset( db2, 0, HEAPSIZE );

		/* 0. init A with meaningful string */
		memcpy( sb2, str, sizeof(str) );

		/* Byte Swap */
		WRITE_MEM32( GDMACNR, 0 ); /* Reset GDMA */
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_MEMCPY );
		WRITE_MEM32( GDMAIMR, 0 );
		WRITE_MEM32( GDMAISR, 0xffffffff );
		WRITE_MEM32( GDMAICVL, 0 );
		WRITE_MEM32( GDMAICVR, 0 );
		WRITE_MEM32( GDMASBP0, (uint32)sb2 );
		WRITE_MEM32( GDMASBL0, GDMA_LDB|sizeof(str) );
		WRITE_MEM32( GDMADBP0, (uint32)db2 );
		WRITE_MEM32( GDMADBL0, GDMA_LDB|sizeof(str) );
		WRITE_MEM32( GDMACNR, GDMA_ENABLE|GDMA_POLL|GDMA_BYTESWAP|GDMA_SWAPTYPE1);

		/* Expected data */
		memcpy( sb1, str, sizeof(str) );
		memcpy( db1, str, sizeof(str) );
		for( j = 0; j < sizeof(str); j+=4 )
		{
			uint8 tmp;
			tmp = db1[j];
			db1[j] = db1[j+3];
			db1[j+3] = tmp;
			tmp = db1[j+1];
			db1[j+1] = db1[j+2];
			db1[j+2] = tmp;
		}
		while( ( READ_MEM32( GDMAISR ) & GDMA_COMPIP ) == 0 ); /* wait for complete. */

		IS_EQUAL_INT("Dest block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_DBUDRIP, 0, __FUNCTION__, __LINE__ );
		IS_EQUAL_INT("Source block should NOT under-run", READ_MEM32(GDMAISR)&GDMA_SBUDRIP, 0, __FUNCTION__, __LINE__ );
		ASSERT( compare_data_block() == SUCCESS );
	}

	return SUCCESS;
}





/*
 * The last data block has NO LDB bit.
 */

