/*
* ----------------------------------------------------------------
* Copyright c              Realtek Semiconductor Corporation, 2002
* All rights reserved.
* 
* Abstract: header stamp module header file.
*
* $Author: blue $
* $id$
*/

#ifndef _HS_MODEL_H_
#define _HS_MODEL_H_

#include "rtl_types.h"

/* To directly access ASIC-mapped header stamp, we must set pack pragma. */
#pragma pack(push,1)

/* Header Stamp Before */
struct hsb_s
{
	uint32 spa:3;		/* [2:0] */
	uint32 res:2;		/* [1:0] */
	uint32 len:15;		/* [15:0] */
	uint32 vid:12;		/* [11:0] */
	uint32 tagif:1;
	uint32 pppoeif:1;
	uint32 sip:32;		/* [31:0] */
	uint32 sprt:16;		/* [15:0] */
	uint32 dip:32;		/* [31:0] */
	uint32 dprt:16;		/* [15:0] */
	uint32 ipptl:8;		/* [7:0] */
	uint32 ipfg:3;		/* [2:0]=[IP_NONFG,IP_DF, IP_MF] */
	uint32 iptos:8;		/* [7:0] */
	uint32 tcpfg:8;		/* [7:0]=TCP flag or ICMP code */
	uint32 type:3;		/* [2:0] */
	uint32 patmatch:1;
	uint32 ethtype:16;	/* [15:0] */
	uint32 da0:8;		/* [47:0] */
	uint32 da1:8;
	uint32 da2:8;
	uint32 da3:8;
	uint32 da4:8;
	uint32 da5:8;
	uint32 sa0:8;		/* [47:0] */
	uint32 sa1:8;
	uint32 sa2:8;
	uint32 sa3:8;
	uint32 sa4:8;
	uint32 sa5:8;
	uint32 hiprior:3;	/* [2:0] */
	uint32 snap:1;
	uint32 udpnocs:1;
	uint32 ttlst:2;		/* [1:0] */
	uint32 dirtx:1;
	uint32 l3csok:1;
	uint32 l4csok:1;
	uint32 ipfo0_n:1;
	uint32 llcothr:1;
	uint32 urlmch:1;
	uint32 extspa:2;	/* [1:0] */
	uint32 extl2:1;
};
typedef struct hsb_s hsb_t;


/* Header Stamp After */
typedef struct 
{
	uint32 nhmac0:8;	/* [47:0] */
	uint32 nhmac1:8;
	uint32 nhmac2:8;
	uint32 nhmac3:8;
	uint32 nhmac4:8;
	uint32 nhmac5:8;
	uint32 trip:32;		/* [31:0] */
	uint32 port:32;		/* [15:0] */
	uint32 l3csdt:16;	/* [15:0] */
	uint32 l4csdt:16;	/* [15:0] */
	uint32 egif:1;
	uint32 l2tr:1;
	uint32 l34tr:1;
	uint32 dirtxo:1;
	uint32 typeo:3;		/* [2:0] */
	uint32 llc:1;
	uint32 vlanproc:2;	/* [1:0] */
	uint32 dvidx:12;	/* [11:0] */
	uint32 pppoeifo:2;	/* [1:0] */
	uint32 pppidx:3;	/* [2:0] */
	uint32 leno:15;		/* [14:0] */
	uint32 l3csoko:1;
	uint32 l4csoko:1;
	uint32 frag:1;
	uint32 lastfrag:1;
	uint32 ipmcastr:1;
	uint32 svidx:12;	/* [11:0] */
	uint32 fragpkt:1;
	uint32 ttl_1if:6;	/* [5:0] */
	uint32 dpc:3;		/* [2:0] */
	uint32 spao:3;		/* [2:0] */
	uint32 reso:2;		/* [1:0] */
	uint32 ttl_1if2:3;	/* [8:6] 3 extPorts */
	uint32 dpext:4;		/* [3:0] */
	uint32 spaext:2;	/* [1:0] 0~3 */
	uint32 why2cpu:16;	/* [15:0] */
	uint32 dp:7; /* destinatio port mask*/
} hsa_t;

#pragma pack(pop)
/* Recover pragma setting */


int32 modelGetHsb( hsb_t* hsb );
int32 modelSetHsb( hsb_t* hsb );
int32 modelGetHsa( hsa_t* hsa );
int32 modelSetHsa( hsa_t* hsa );
#define HSB_TYPE_TCP  5
#define HSB_TYPE_UDP 6

#endif /* _HS_H_ */
