/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : Source File for IC to Export Registers
* Abstract : 
* Author : Louis Yung-Chieh Lo (yjlou@realtek.com.tw)               
* $Id: icExport.c,v 1.1.1.1 2006/07/26 05:38:22 blue Exp $
*/

#include "rtl_types.h"
#include "asicRegs.h"
#include "assert.h"
#include "rtl_glue.h"
#include "icTest.h"
#include "icExport.h"
#include "gdma.h"
#include "virtualMac.h"
#include <cle_struct.h>
#ifdef RTL865X_MODEL_USER
#include <sys/types.h> /* for file access */
#include <sys/stat.h> /* for file access */
#include <fcntl.h> /* for file access */
#include <unistd.h> /* for file access */
#include <errno.h> /* strerr */
#else
#ifdef CONFIG_RTL865XB_EXP_CRYPTOENGINE
#include "rtl865xc_cryptoEngine.h"
#endif
#endif


uint32 expGroup = 0;

/* Output File */
char expFileName[256];
static int fdFile = -1; /* File Descriptor of Output File */

enum EXPORT_REGISTER_FORMAT expFormat = 0; /* default for none */
enum IC_TYPE expIcType;


#define MAX_VIRTUAL_MAP 16
typedef struct VIRTUAL_MAP_S
{
	uint32 virtualStart, virtualEnd;  /* if virtualStart==virtualEnd, mean invalid */
	uint32 physicalStart;
} VIRTUAL_MAP_T;
VIRTUAL_MAP_T virtualMap[MAX_VIRTUAL_MAP];

/* for runModelExport() to check */
uint32 defaultModelToPhysicalAdded;

/*===========================================================================*
 *   VIRTUAL ADDRESS TO PHYSICAL ADDRESS                                     *
 *===========================================================================*/
/*
 * map back to physical address, for example:
 *   [pVirtualSWReg,pVirtualSWReg+64K} --> REAL_SWCORE_BASE
 */
uint32 queryModelToPhysical( uint32 virtual )
{
	int i;
	for( i = 0; i < MAX_VIRTUAL_MAP; i++ )
	{
		if ( virtual >= virtualMap[i].virtualStart && virtual < virtualMap[i].virtualEnd )
		{
			return virtual - virtualMap[i].virtualStart + virtualMap[i].physicalStart;
		}
	}

	return virtual;
}

/*
 * add a pair of [virtual,physical].
 */
int32 addModelToPhysical( uint32 virtualStart, uint32 length, uint32 physicalStart )
{
	int i;
	for( i = MAX_VIRTUAL_MAP-1; i >= 0 ; i-- )
	{
		if ( virtualMap[i].virtualStart == virtualMap[i].virtualEnd )
		{
			virtualMap[i].virtualStart = virtualStart;
			virtualMap[i].virtualEnd = virtualStart+length;
			virtualMap[i].physicalStart = physicalStart;
			return SUCCESS;
		}
	}
	return FAILED;
}

/*
 * delete a pair of [virtual,physical].
 */
int32 delModelToPhysical( uint32 virtualStart )
{
	int i;
	for( i = 0; i < MAX_VIRTUAL_MAP; i++ )
	{
		if ( virtualMap[i].virtualStart == virtualStart )
		{
			virtualMap[i].virtualStart = virtualMap[i].virtualEnd;
			virtualMap[i].physicalStart = 0xDeadC0de;
			return SUCCESS;
		}
	}
	return FAILED;
}


/*===========================================================================*
 *   EXPORT-RELATED FUNCTIONS                                                *
 *===========================================================================*/
/*
 *	This function is used to dump 'addr' and 'value'.
 *	'addr' will be converted if it is a model-space variable, for example, 0x08091CA4 --> 0x00F91CA4.
 *	'value' will be raw-output.
 */
char* modelExportValue( uint32 addr, uint32 value )
{
	static char buf[256];
	enum IC_TYPE icType;

	/* IC type is not matched. */
	model_getTestTarget( &icType );
	if ( icType != expIcType ) return NULL;

	switch( expFormat )
	{
		case EXPORT_RTK_SIM:
			sprintf( buf, "memw %08X %08X;", queryModelToPhysical(addr), value );
			break;
			
		case EXPORT_RTK_EXPECT:
			sprintf( buf, "memr %08X %08X;", queryModelToPhysical(addr), value );
			break;
			
		case EXPORT_CLESHELL:
			sprintf( buf, "memory write %08x %08x;", queryModelToPhysical(addr), value );
			break;
	}

	if ( fdFile != -1 )
	{
#ifdef RTL865X_MODEL_USER
		write( fdFile, buf, strlen(buf) );
		write( fdFile, "\n", 1 );
#else
		rtlglue_printf( "%s\n", buf );
#endif
	}

	return buf;
}


/*
 *	This function is used if the 'addr' is a register that pointed to SDRAM.
 *	The value of this register will be converted to physical address of SDRAM.
 */
char* modelExportPointerRegister( uint32 addr )
{
	char *buf;
	uint32 value;

	value = READ_VIR32( addr );
	buf = modelExportValue( addr, queryModelToPhysical(value) );

	return buf;
}


/*
 *	This function is used for 'addr' is a normal register.
 */
char* modelExportRegister( uint32 addr )
{
	char *buf;
	uint32 value;

	value = READ_VIR32( addr );
	buf = modelExportValue( addr, value );

	return buf;
}


/*
 *	This function is for burst memory dump.
 *	The 'addr' will be converted to physical address, either SDRAM or ASIC register.
 *	The unit of len is bytes.
 */
char* modelExportMemory( uint32 addr, uint32 len )
{
	static char buf[1024*16];
	static char tmp[32];
	uint32 value;
	enum IC_TYPE icType;
	int i, j;

	/* IC type is not matched. */
	model_getTestTarget( &icType );
	if ( icType != expIcType ) return NULL;

	if ( len & 3 ) 
	{
		rtlglue_printf( "Currently, we do not support len(%d) that is not times of 4. ", len );
		len = (len+3)&~3;
		rtlglue_printf( "Round up to %d.\n", len );
	}

	buf[0] = '\0';
	if ( len == 0 ) return buf;
	switch( expFormat )
	{
		case EXPORT_RTK_SIM:
			for( i = 0; i < len; i+=j )
			{
				buf[0] = '\0';
				sprintf( tmp, "memw %08X ", queryModelToPhysical(addr+i) );
				strcat( buf, tmp );
				for( j = 0; (j<31*4)&&(i+j+4)<len; j+=4 )
				{
					value = READ_VIR32( addr+i+j );
					sprintf( tmp, "%08X,\n              ", value );
					strcat( buf, tmp );					
				}
				value = READ_VIR32( addr+i+j );
				sprintf( tmp, "%08X;\n", value );
				strcat( buf, tmp );
				j+=4;

				if ( fdFile != -1 )
				{
#ifdef RTL865X_MODEL_USER
					write( fdFile, buf, strlen(buf) );
#else
					rtlglue_printf( "%s", buf );
#endif
				}

			}
			break;
			
		case EXPORT_RTK_EXPECT:
			for( i = 0; i < len; i+=j )
			{
				buf[0] = '\0';
				
				sprintf( tmp, "memr %08X ", queryModelToPhysical(addr+i) );
				strcat( buf, tmp );
				for( j = 0; (j<31*4)&&(i+j+4)<len; j+=4 )
				{
					value = READ_VIR32( addr+i+j );
					sprintf( tmp, "%08X,\n              ", value );
					strcat( buf, tmp );					
				}
				value = READ_VIR32( addr+i+j );
				sprintf( tmp, "%08X;\n", value );
				strcat( buf, tmp );
				j+=4;

				if ( fdFile != -1 )
				{
#ifdef RTL865X_MODEL_USER
					write( fdFile, buf, strlen(buf) );
#else
					rtlglue_printf( "%s", buf );
#endif
				}
			}
			break;
			
		case EXPORT_CLESHELL:
			for( i = 0; i < len; i+=4 )
			{
				value = READ_VIR32( addr + i );
				sprintf( buf, "memory write %08x %08x;\n", addr, value );
			}
			break;
	}

	buf[0] = '\0';
	return buf;
}


/*
 *	This function is provided to model code.
 *	It will output human-readable comment according to 'format' field.
 *	Leaving pszComment with NULL pointer will output a empty new line.
 */
char* modelExportComment( char* pszComment )
{
	static char buf[256];
	enum IC_TYPE icType;

	/* IC type is not matched. */
	model_getTestTarget( &icType );
	if ( icType != expIcType ) return NULL;

	switch( expFormat )
	{
		case EXPORT_RTK_SIM:
			if ( pszComment )
				sprintf( buf, "// %s", pszComment );
			else
				buf[0]= '\0';
			break;
			
		case EXPORT_RTK_EXPECT:
			if ( pszComment )
				sprintf( buf, "// %s", pszComment );
			else
				buf[0]= '\0';
			break;
			
		case EXPORT_CLESHELL:
			if ( pszComment )
				sprintf( buf, "// %s", pszComment );
			else
				buf[0]= '\0';
			break;
	}

	if ( fdFile != -1 )
	{
#ifdef RTL865X_MODEL_USER
		write( fdFile, buf, strlen(buf) );
		write( fdFile, "\n", 1 );
#else
		rtlglue_printf( "%s\n", buf );
#endif
	}

	return buf;
}
/*===========================================================================*
 *   END OF EXPORT FUNCTIONS                                                 *
 *===========================================================================*/



/*===========================================================================*
 *   IC-SPECIFIC EXPORT FUNCTION                                             *
 *===========================================================================*/
/*
 *	This function only dumps GDMA related registers and memory content.
 */
#define BURST_ACCESS
int32 modelExportGdmaRegisters( void )
{
	{
		uint32 i, len;
#ifdef BURST_ACCESS /* use burst access for designer */
#else
		uint32 j;
#endif
		
		/* export source data block content */
		for( i = 0; i < 8; i++ )
		{
			uint32 start, end;

			start = READ_VIR32(GDMASBP0+i*8);
			len = READ_VIR32( GDMASBL0+i*8 ) & GDMA_BLKLENMASK;
			end = start + len;

			/* align to 4-bytes */
			start &= ~3;
			end = ( end + 3 ) & ~3;
			
#ifdef BURST_ACCESS /* use burst access for designer */
			if ( end > start ) modelExportMemory( start, end-start );
#else
			for( j = start; j < end; j += 4 )
			{
				modelExportRegister( j );
			}
#endif

			if ( READ_VIR32( GDMASBL0+i*8 ) & GDMA_LDB ) break;
		}

		/* If this is pattern match, we need special process for destination data block. */
		if ( ( READ_VIR32(GDMACNR) & GDMA_FUNCMASK ) == GDMA_PATTERN )
		{
			uint32 **pVsm;
			uint32 *pSub;
			uint32 inst;
			uint32 i, j;
			
			pVsm = (uint32 **)READ_VIR32( GDMADBP0 );

			for( i = 0; i < SUB_STATE_MACHINE_NUMBER; i++ )
			{
				if (( pSub = (uint32*)big_endian(READ_VIR32((uint32)(pVsm+i))) ))
				{
					modelExportValue( (uint32)(pVsm+i), queryModelToPhysical((uint32)pSub) );
					for( j = 0; j < INTERNAL_RULE_NUMBER; j++ )
					{
						if (( inst = big_endian(READ_VIR32((uint32)(pSub+j))) ))
						{
							modelExportValue( (uint32)(pSub+j), inst );
						}
					}
				}
				else
				{
					break;
				}
			}
		}
		else
		{
			/* export destination data block content */
			for( i = 0; i < 8; i++ )
			{
				uint32 start, end;

				start = READ_VIR32(GDMADBP0+i*8);
				len = READ_VIR32( GDMADBL0+i*8 ) & GDMA_BLKLENMASK;
				if ( len > 0 )
				{
					end = start + len;

					/* align to 4-bytes */
					start &= ~3;
					end = ( end + 3 ) & ~3;
					
#ifdef BURST_ACCESS /* use burst access for designer */
					if ( end > start ) modelExportMemory( start, end-start );
#else
					for( j = start; j < end; j += 4 )
					{
						modelExportRegister( j );
					}
#endif
				}

				if ( READ_VIR32( GDMADBL0+i*8 ) & GDMA_LDB ) break;
			}
		}

		/* Export register */
		modelExportRegister( GDMAIMR );
		modelExportRegister( GDMAISR );
		modelExportRegister( GDMAICVL );
		modelExportRegister( GDMAICVR );

		/* export source data block register */
		for( i = 0; i < 8; i++ )
		{
			modelExportPointerRegister( GDMASBP0+i*8 );
			modelExportRegister( GDMASBL0+i*8 );
			if ( READ_VIR32( GDMASBL0+i*8 ) & GDMA_LDB ) break;
		}

		/* export destination data block register */
		for( i = 0; i < 8; i++ )
		{
			modelExportPointerRegister( GDMADBP0+i*8 );
			modelExportRegister( GDMADBL0+i*8 );
			if ( READ_VIR32( GDMADBL0+i*8 ) & GDMA_LDB ) break;
		}

		modelExportRegister( GDMACNR );
	}

	return SUCCESS;
}


#ifdef CONFIG_RTL865XB_EXP_CRYPTOENGINE
/*
 *	This function only dumps Crypto related registers and memory content.
 */
int32 modelExportCryptoRegisters( void )
{
	{
		rtl865xc_crypto_source_t *ipssdar;
		rtl865xc_crypto_dest_t *ipsddar;

		/* export source and destination descriptor */
		ipssdar = (rtl865xc_crypto_source_t *)KSEG1_ADDRESS(READ_VIR32( IPSSDAR ));
		do
		{
			modelExportMemory( (uint32)ipssdar, sizeof(*ipssdar) );
			if ( ipssdar->sbl )
			{
				uint32 start, end;

				start = KSEG1_ADDRESS(ipssdar->sdbp);
				end = start + ipssdar->sbl;
				/* align to 4-bytes */
				start &= ~3;
				end = ( end + 3 ) & ~3;
				modelExportMemory( start, end-start );
			}
		} while ( (ipssdar++)->eor == 0 );

		ipsddar = (rtl865xc_crypto_dest_t *)KSEG1_ADDRESS(READ_VIR32( IPSDDAR ));
		do
		{
			modelExportMemory( (uint32)ipsddar, sizeof(*ipsddar) );
			if ( ipsddar->dbl )
			{
				uint32 start, end;

				start = KSEG1_ADDRESS(ipsddar->ddbp);
				end = start + ipsddar->dbl;
				/* align to 4-bytes */
				start &= ~3;
				end = ( end + 3 ) & ~3;
				modelExportMemory( start, end-start );
			}
		} while ( (ipsddar++)->eor == 0 );

		/* Dump registers */
		modelExportRegister( IPSCTR );
		modelExportPointerRegister( IPSSDAR );
		modelExportPointerRegister( IPSDDAR );
		modelExportRegister( IPSCSR );
	}
	return SUCCESS;
}
#endif


/*
 *	This function is called by program.
 *	It will output all IC registers.
 *	So, if your module needs to support export function, please add your code here.
 */
int32 modelExportAllRegisters( void )
{
	modelExportGdmaRegisters();
	return SUCCESS;
}


int32 modelExportSetOutputForm( enum EXPORT_REGISTER_FORMAT newFormat )
{
	expFormat = newFormat;
	return SUCCESS;
}


int32 modelExportGetOutputForm( enum EXPORT_REGISTER_FORMAT *oldFormat )
{
	if ( oldFormat )
		*oldFormat = expFormat;
	return SUCCESS;
}


int32 modelExportSetIcType( enum IC_TYPE newType )
{
	expIcType = newType;
	return SUCCESS;
}


int32 modelExportGetIcType( enum IC_TYPE *oldType )
{
	if ( oldType )
		*oldType = expIcType;
	return SUCCESS;
}


int32 modelExportOpenFile( char* pszFilename )
{
#ifdef RTL865X_MODEL_USER
	if ( fdFile != -1 )
		modelExportCloseFile();

	if ( !strcmp( pszFilename, "stdout" ) )
		pszFilename = "/dev/stdout";

	/* open log file */
	fdFile = open( pszFilename, O_CREAT | O_RDWR | O_APPEND, 0666 );
	if ( fdFile == -1 )
	{
		rtlglue_printf( "Open export logfile '%s' error: %s\n", pszFilename, strerror(errno) );
		return FAILED;
	}

	rtlglue_printf( "Export logfile '%s' is opened.\n", pszFilename );
#else
	fdFile = 0; /* set 0 to print out */
#endif
	strcpy( expFileName, pszFilename );
	
	return SUCCESS;
}


int32 modelExportCloseFile( void )
{
#ifdef RTL865X_MODEL_USER
	if ( fdFile != -1 )
	{
		close( fdFile );
		fdFile = -1;
	}
#else
	fdFile = -1; /* set -1 not to print out */
#endif

	rtlglue_printf( "Export log file '%s' is closed.\n", expFileName );
	return SUCCESS;
}


int32 runModelExport(uint32 userId,  int32 argc,int8 **saved)
{
	int8 *nextToken;
	int32 size;

#ifdef RTL865X_MODEL_USER
	expIcType = IC_TYPE_MODEL;
#else /*RTL865X_MODEL_KERNEL*/
	expIcType = IC_TYPE_REAL;
#endif

	if ( defaultModelToPhysicalAdded == FALSE )
	{
		addModelToPhysical( (uint32)REAL_SWCORE_BASE, VIRTUAL_SWCORE_REG_SIZE, PHYSICAL_ADDRESS(REAL_SWCORE_BASE) );
		addModelToPhysical( (uint32)REAL_SYSTEM_BASE, VIRTUAL_SYSTEM_REG_SIZE, PHYSICAL_ADDRESS(REAL_SYSTEM_BASE) );
#ifdef RTL865X_MODEL_USER
		addModelToPhysical( 0x08000000, 0x02000000, 0x00000000 );
#else /*RTL865X_MODEL_KERNEL*/
		addModelToPhysical( 0x80000000, 0x02000000, 0x00000000 ); /* KSEG0 */
		addModelToPhysical( 0xA0000000, 0x02000000, 0x00000000 ); /* KSEG1 */
#endif
		addModelToPhysical( (uint32)pVirtualSWReg, VIRTUAL_SWCORE_REG_SIZE, PHYSICAL_ADDRESS(REAL_SWCORE_BASE) );
		addModelToPhysical( (uint32)pVirtualSysReg, VIRTUAL_SYSTEM_REG_SIZE, PHYSICAL_ADDRESS(REAL_SYSTEM_BASE) );

		defaultModelToPhysicalAdded = TRUE;
	}
	
	cle_getNextCmdToken(&nextToken,&size,saved);
	if ( !strcmp( nextToken, "snapshot" ) )
	{
		/* to be implemented */
	}
	else if ( !strcmp( nextToken, "open" ) )
	{
		/* get filename */
		cle_getNextCmdToken(&nextToken,&size,saved);
		return modelExportOpenFile( nextToken );
	}
	else if ( !strcmp( nextToken, "close" ) )
	{
		return modelExportCloseFile();
	}
	else if ( !strcmp( nextToken, "map" ) )
	{
		int i;

		rtlglue_printf("Model Address Space | Physical\n" );
		rtlglue_printf("--------------------+-------------\n" );
		for( i = 0; i < MAX_VIRTUAL_MAP; i++ )
		{
			if ( virtualMap[i].virtualStart != virtualMap[i].virtualEnd )
				rtlglue_printf("[%08x-%08x] | %08x\n", virtualMap[i].virtualStart, virtualMap[i].virtualEnd, virtualMap[i].physicalStart );
		}
	}
	else
		return FAILED;

	return SUCCESS;
}


