/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : Source File for Layer2 Model Code
* Abstract : 
* Author : Louis Yung-Chieh Lo (yjlou@realtek.com.tw)               
* $Id: l2Model.c,v 1.1.1.1 2006/07/26 05:38:22 blue Exp $
*/

#include <rtl_glue.h>
#include "assert.h"
#include "hsModel.h"
#include "asicRegs.h"
#include "icModel.h"
#include "modelTrace.h"
#ifdef CONFIG_RTL865XC
#include "rtl865xC_tblAsicDrv.h"
#else
#include "rtl8651_tblAsicDrv.h"
#endif
#include "l2Model.h"
static int32 _cmpTableMac(ether_addr_t * srcP, ether_addr_t * dstP);
static uint32 l2SwitchTable_getHashIndex(uint8 *,uint8);
static uint16 internalVlanID,DMACisGateway;
static uint8 MultiLayerMode;
rtl865x_tblAsicDrv_vlanParam_t  GlobalNetif;
/* 
@func uint16 | modelMatchGatewayMac	| This function will return the match or no
@parm null
@rvalue enum MODEL_BOOLEAN_VALUE|
@comm 
*/
enum MODEL_BOOLEAN_VALUE modelMatchGatewayMAC()
{
	if (DMACisGateway) return MMATCH;
	else return MNOTMATCH;
}
int32 modelGetVlanID()
{
	return internalVlanID;
}

int32 modelGetNetIf(void* pointer)
{
	*(int*)(pointer)=(int)&GlobalNetif;
	return SUCCESS;
}
/*
@func void | 	| This function will set cpu port
@parm hsb_t* | hsb | Header Stamp of Before (for input)
@parm hsa_t* | hsa | Header Stamp of After (for output)
@comm 
*/
void modelCPUport(hsb_t* hsb,hsa_t* hsa)
{
	hsa->dp=MODEL_CPU_PORT_MASK;
}
/*
@func void | 	| This function will add cpu port
@parm hsb_t* | hsb | Header Stamp of Before (for input)
@parm hsa_t* | hsa | Header Stamp of After (for output)
@comm 
*/
void modelAddCPUport(hsb_t* hsb,hsa_t* hsa)
{
	hsa->dp|=MODEL_CPU_PORT_MASK;
}


/*
func uint16 | modelFindVlanId	| This function will return the vlan ID which this packet belongs.
parm hsb_t* | hsb | Header Stamp of Before (for input)
parm hsa_t* | hsa | Header Stamp of After (for output)
rvalue uint16 | VLAN ID (0~4095)
comm 
According the following procedure to decide VLAN ID:
 1. 802.1q VLAN tag existed
 2. protocol-based VLAN
 3. Port-based VLAN (PVID)
*/

static int32 modelFindVlanId( hsb_t* hsb, hsa_t* hsa )
{
	uint8 i,valid = FALSE;
	uint32 vlanId;
	rtl865x_tblAsicDrv_vlanParam_t  vlan;
	memset((void*)&vlan,0,sizeof(vlan));
	if (( hsb->tagif)  && ((READ_MEM32(VCR) &EN_1QUNWARE)!=EN_1QUNWARE))
	{
		/* 802.1q VLAN ID present */
		vlanId = hsb->vid;
		goto out;
	}
	else
	{
		/* traverse protocol-based vlan */
		uint8 ptypep, ptype5, ptype6;
		uint16 pvalue5, pvalue6;
	
		switch ( hsb->ethtype )
		{
			case 0x8137: /* IPX */
				rtl8651_getProtocolBasedVLAN( 1, hsb->spa, &valid, &vlanId );
				break;
			case 0x8040: /* NetBios */
				rtl8651_getProtocolBasedVLAN( 2, hsb->spa, &valid, &vlanId );
				break;
			case 0x8863: /* PPPoE Control */
				rtl8651_getProtocolBasedVLAN( 3, hsb->spa, &valid, &vlanId );
				break;
			case 0x8864: /* PPPoE Data */
				rtl8651_getProtocolBasedVLAN( 4, hsb->spa, &valid, &vlanId );
				break;
			default:

				rtl8651_queryProtocolBasedVLAN( 5, &ptype5, &pvalue5 );
				rtl8651_queryProtocolBasedVLAN( 6, &ptype6, &pvalue6 );

				if ( hsb->llcothr )
					ptypep = 0x2; /* LLC-other, see PBVCR and PBVCR2 */
				else
					ptypep = 0x0; /* Ethernet II, see PBVCR and PBVCR2 */
				
				if ( ptype5==ptypep && hsb->ethtype==pvalue5 )
					rtl8651_getProtocolBasedVLAN( 5, hsb->spa, &valid, &vlanId);
				else if ( ptype6==ptypep && hsb->ethtype==pvalue6 )
					rtl8651_getProtocolBasedVLAN( 6, hsb->spa, &valid, &vlanId);

		}
		
		if ( valid )
		{
			goto out;
		}

		/* Use PVID */
		rtl8651_getAsicPvid(hsb->spa,&vlanId);

	}

out: /* assert vlanIdx is set. */
	
		DMACisGateway=0;
		MultiLayerMode =  (READ_MEM32(SWTCR0)>>16)&0x3;
		switch(MultiLayerMode)
		{
			case VLAN_BASE_MODE:
				for (i=0;i<RTL865XC_NETINTERFACE_NUMBER;i++)
				{
					rtl865xc_getAsicNetInterface(i,&vlan);
					if (vlan.valid && vlan.vid==vlanId)
						if ( (vlan.macAddr.octet[5]==hsb->da5) && (vlan.macAddr.octet[4]==hsb->da4) &&
					  	(vlan.macAddr.octet[3]==hsb->da3) &&(vlan.macAddr.octet[2]==hsb->da2) &&
						(vlan.macAddr.octet[1]==hsb->da1) &&(vlan.macAddr.octet[0]==hsb->da0) 
						)		
						{
							DMACisGateway=1;
							break;
						}

				}
				break;
			case PORT_BASE_MODE:	
				rtl865xc_getAsicNetInterface((READ_MEM32(PLITIMR) >> ((hsb->spa-1)*3))&0x3,&vlan);
				if (vlan.valid && vlan.vid==vlanId)
				if ( (vlan.macAddr.octet[5]==hsb->da5) && (vlan.macAddr.octet[4]==hsb->da4) &&
				  	(vlan.macAddr.octet[3]==hsb->da3) &&(vlan.macAddr.octet[2]==hsb->da2) &&
					(vlan.macAddr.octet[1]==hsb->da1) &&(vlan.macAddr.octet[0]==hsb->da0) 
					)		
					{
						DMACisGateway=1;
						break;
					}
				break;
			case MAC_BASE_MODE:
				for (i=0;i<RTL865XC_NETINTERFACE_NUMBER;i++)
				{
					rtl865xc_getAsicNetInterface(i,&vlan);
					if (vlan.valid && vlan.vid)
						if ( (vlan.macAddr.octet[5]==hsb->da5) && (vlan.macAddr.octet[4]==hsb->da4) &&
					  	(vlan.macAddr.octet[3]==hsb->da3) &&(vlan.macAddr.octet[2]==hsb->da2) &&
						(vlan.macAddr.octet[1]==hsb->da1) &&(vlan.macAddr.octet[0]==hsb->da0) 
						)		
						{
							DMACisGateway=1;
							break;
						}
				}
				break;
			default:
				return MRET_TOCPU;
				
		}

	memcpy((void*)&GlobalNetif,(void*)&vlan,sizeof(vlan));
	internalVlanID=(vlanId);
	return (vlanId);

}


/*
func enum MODEL_RETURN_VALUE | model802_1dAddress	| model code to trap 802.1d reserved address
parm hsb_t* | hsb | Header Stamp of Before (for input)
parm hsa_t* | hsa | Header Stamp of After (for output)
rvalue enum MODEL_RETURN_VALUE |
comm 
*/
enum MODEL_RETURN_VALUE model802_1dAddress( hsb_t* hsb, hsa_t* hsa )
{
	if ( hsb->da0==0x01 && hsb->da1==0x80 && hsb->da2==0xC2 &&
	     hsb->da3==0x00 && hsb->da4==0x00 && (hsb->da5&0xf0)==0x00 )
	{
/*		hsa->dpext = PN_PORT_CPU-PN_PORT_EXT0; to CPU */
		modelCPUport(hsb,hsa);
		MT_TOCPU( "" );
		return MRET_TOCPU;
	}
	return MRET_OK;
}


/*
func enum MODEL_RETURN_VALUE | modelIngress802_1X	| model code to trap 802.1X Control
parm hsb_t* | hsb | Header Stamp of Before (for input)
parm hsa_t* | hsa | Header Stamp of After (for output)
rvalue enum MODEL_RETURN_VALUE |
comm 
*/
enum MODEL_RETURN_VALUE modelIngress802_1X( hsb_t* hsb, hsa_t* hsa )
{
	uint32 status;
	status= READ_MEM32(DOT1XPORTCR) ;
	status = (status>>(hsb->spa*3)) &03;
	if ((status& EN_8021X_PORT_CONTROL)==0) 
		return MRET_OK;
	if (status &EN_8021X_PORT_AUTH)
		return MRET_OK;
	modelCPUport(hsb,hsa);
	MT_TOCPU( "Port Not Unauthized" );
	return MRET_TOCPU;

	return MRET_OK;
}
/*
func enum MODEL_RETURN_VALUE | modelEgress802_1X	| model code to trap 802.1X Control
parm hsb_t* | hsb | Header Stamp of Before (for input)
parm hsa_t* | hsa | Header Stamp of After (for output)
rvalue enum MODEL_RETURN_VALUE |
comm 
*/
enum MODEL_RETURN_VALUE modelEgress802_1X( hsb_t* hsb, hsa_t* hsa )
{
	uint32 status;
	status= READ_MEM32(DOT1XPORTCR) ;
	status = (status>>(hsb->spa*3)) &03;
	if ((status& EN_8021X_PORT_CONTROL)==0) 
		return MRET_OK;
	if (status &EN_8021X_PORT_AUTH)
		return MRET_OK;
	if(status & EN_8021X_PORT_DIR_IN)
		return MRET_OK;
	modelCPUport(hsb,hsa);
	MT_TOCPU( "Egress Drop Port Not Unauthized" );
	return MRET_TOCPU;
}


/*
func enum MODEL_RETURN_VALUE | modelIngressCheck	| model code for ingress check
parm hsb_t* | hsb | Header Stamp of Before (for input)
parm hsa_t* | hsa | Header Stamp of After (for output)
rvalue enum MODEL_RETURN_VALUE |
comm 
*/
enum MODEL_RETURN_VALUE modelIngressCheck( hsb_t* hsb, hsa_t* hsa )
{
	rtl865xc_tblAsic_vlanTable_t vlan;
	uint32 vid;
	char msg[128];					
	uint32 spa=hsb->spa;	
	vid = ( modelFindVlanId( hsb, hsa ) );
	hsa->dvidx= vid;
	/* Get the VLAN index which packet belongs. */
	_rtl8651_readAsicEntry(TYPE_VLAN_TABLE, vid, &vlan);
	if ( (vlan.memberPort==0) &&(vlan.extMemberPort==0) &&( READ_MEM32(SWTCR0)&EN_UKNOWNVID_TO_CPU))
	{

			modelCPUport(hsb,hsa);
			sprintf( msg, "VLAN(%d)  No member port.", vid );
			MT_TOCPU( msg );
			return MRET_TOCPU;

	}
	/* We check VLAN ingress filtering only when enabled. */
	if ( READ_MEM32(SWTMCR) & EN_VLAN_INGRESS_FILTER )
	{

		/* Verify the member port of VLAN. Drop this packet if the packet is coming from the non-existed member port. */
		if ( (( (1<<spa) & vlan.memberPort ) ==0) &&(( (1<<spa) & vlan.extMemberPort) == 0) )		
		{
			hsa->dp = 0;
			hsa->dpext=0;
			sprintf( msg, "source port %d is not in VLAN(%d)'s member port.", spa, vid );
			MT_DROP( msg );
			return MRET_DROP;
		}
	}
	return MRET_OK;
}
/*
@func enum MODEL_RETURN_VALUE | modelEgressCheck	| model code for ingress check
@parm hsb_t* | hsb | Header Stamp of Before (for input)
@parm hsa_t* | hsa | Header Stamp of After (for output)
@rvalue enum MODEL_RETURN_VALUE |
@comm 
*/
enum MODEL_RETURN_VALUE modelEgressCheck( hsb_t* hsb, hsa_t* hsa )
{
	rtl865xc_tblAsic_vlanTable_t vlan;
	uint32 vid;
	uint32 dp=hsa->dp;
	uint32 extdp=hsa->dpext;
	vid=hsa->dvidx;
	/* We check VLAN ingress filtering only when enabled. */
	/* Get the VLAN index which packet belongs. */
	_rtl8651_readAsicEntry(TYPE_VLAN_TABLE, vid, &vlan);
	/* Verify the member port of VLAN. Drop this packet if the packet is coming from the non-existed member port. */
	if ( ((dp & vlan.memberPort ) ==0) &&((extdp & vlan.extMemberPort) == 0) )		
	{
		char msg[256];					
		sprintf( msg, "destination  port list  %x is not in VLAN(%d)'s member port.", dp, vid );
		rtlglue_printf("destination  port list  %x is not in VLAN(%d)'s member port %x.\n", dp, vid,vlan.memberPort);
		MT_DROP( msg );
		return MRET_DROP;

	}
	return MRET_OK;
}

static int32 _cmpTableMac(ether_addr_t * srcP, ether_addr_t * dstP)
{
    return (     (srcP->octet[5] == dstP->octet[5]) &&
	   		(srcP->octet[4] == dstP->octet[4]) &&
	   		(srcP->octet[3] == dstP->octet[3]) &&
	   		(srcP->octet[2] == dstP->octet[2]) &&	   		
	   		(srcP->octet[1] == dstP->octet[1]) &&
	   		(srcP->octet[0] == dstP->octet[0]));
}
static uint32 l2SwitchTable_getHashIndex(uint8 *mac,uint8 fid)
{
    return  (mac[0] ^mac[1] ^mac[2]^mac[3]^mac[4]^mac[5] ^ fid)<<2;
                   
}
/*
func enum MODEL_RETURN_VALUE | modelLayer2Switching	| model code for L2 switching
parm hsb_t* | hsb | Header Stamp of Before (for input)
parm hsa_t* | hsa | Header Stamp of After (for output)
rvalue enum MODEL_RETURN_VALUE |
comm 
*/
enum MODEL_RETURN_VALUE modelLayer2Switching( hsb_t* hsb, hsa_t* hsa )
{
	uint32  eidx;
	uint32  i;
	ether_addr_t mac,entryMac;
	uint8 fid;
	rtl865xc_tblAsic_l2Table_t   entryContent;
	   rtl865xc_tblAsic_vlanTable_t vlanTable;   	
	_rtl8651_readAsicEntry(TYPE_VLAN_TABLE, internalVlanID,&vlanTable);
	fid=vlanTable.fid;
	/* Destitntoin Mac Searching*/
	mac.octet[0]=hsb->da0;	mac.octet[1]=hsb->da1;
	mac.octet[2]=hsb->da2;	mac.octet[3]=hsb->da3;
	mac.octet[4]=hsb->da4;	mac.octet[5]=hsb->da5;
	eidx = l2SwitchTable_getHashIndex((uint8*)&mac,fid);
	hsa->dp = vlanTable.memberPort &(~ (1<<hsb->spa ));
	hsa->dpext=vlanTable.extMemberPort &(~(1<<hsb->extspa));
	/* Walk through 4 ways to find out */

	for (i=0; i<4; i++)
	{
	   	_rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, eidx + i, &entryContent);
		entryMac.octet[0] = entryContent.mac47_40;
		entryMac.octet[1] = entryContent.mac39_24>>8;
		entryMac.octet[2] = entryContent.mac39_24&0xff;
		entryMac.octet[3] = entryContent.mac23_8>>8;
		entryMac.octet[4] = entryContent.mac23_8&0xff;	   
		entryMac.octet[5]=	 entryMac.octet[0]^entryMac.octet[1]^entryMac.octet[2]^entryMac.octet[3]^entryMac.octet[4]^((uint8)(entryContent.fid&0x3))^(eidx>>2);		
	       if ( _cmpTableMac(&entryMac, &mac)  &&( entryContent.fid == (fid&0x3)) )
	        	{

	    			hsa->dp = entryContent.memberPort &(~ (1<<hsb->spa ));;
	    			hsa->dpext= entryContent.extMemberPort & (~(1<<hsb->extspa));;
	 		   	entryContent.agingTime= 0x3;
				break;
	        	}
	}	
	/* Unknow unicast control*/
	if ((i==4) && ( (hsb->da0&0x1)!=0x1) && (  (READ_MEM32(FFCR)&EN_UNUNICAST_TOCPU)==EN_UNUNICAST_TOCPU))
	{
		modelAddCPUport(hsb,hsa);
	}
	/* Unknow muliticast control*/
	if ((i==4) && ( (hsb->da0&0x1)==0x1) && (  (READ_MEM32(FFCR)&EN_UNMCAST_TOCPU)==EN_UNMCAST_TOCPU))
	{
		modelAddCPUport(hsb,hsa);
	}

	
	/* Source MAC Learning*/
	
	if (hsb->sa0&0x1)     return MRET_OK;

	mac.octet[0]=hsb->sa0;    mac.octet[1]=hsb->sa1;
	mac.octet[2]=hsb->sa2;    mac.octet[3]=hsb->sa3;
	mac.octet[4]=hsb->sa4;    mac.octet[5]=hsb->sa5;
	eidx = l2SwitchTable_getHashIndex((uint8*)&mac,fid);
	 /* Walk through 4 ways to find available slot */
	for (i=0; i<4; i++)
	{
	    _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, eidx + i, &entryContent);
		entryMac.octet[0] = entryContent.mac47_40;
		entryMac.octet[1] = entryContent.mac39_24>>8;
		entryMac.octet[2] = entryContent.mac39_24&0xff;
		entryMac.octet[3] = entryContent.mac23_8>>8;
		entryMac.octet[4] = entryContent.mac23_8&0xff;
		entryMac.octet[5]=	entryMac.octet[0]^entryMac.octet[1]^entryMac.octet[2]^entryMac.octet[3]^entryMac.octet[4]^(eidx>>2);
	        if ( _cmpTableMac(&entryMac, &mac) && (fid==entryContent.fid))
	        	{

				if (hsb->spa>RTL8651_PHY_NUMBER)
				{
					entryContent.extMemberPort=1<< hsb->extspa;
					entryContent.memberPort=0;
				}
				else
				{
					entryContent.memberPort=1<< hsb->spa;				
					entryContent.extMemberPort=0;					
				}					
				entryContent.agingTime= 0x3;
		       return MRET_OK;
	        	}

	}

	for (i=0; i<4; i++)
	{
	    _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, eidx + i, &entryContent);
	    if ( entryContent.isStatic )
		continue;
	    if ( entryContent.agingTime)	
		continue;
	    break;
	}    
	if ( ( (READ_MEM32(TEACR)&EN_L2LRUHASH) != EN_L2LRUHASH) && (i==4))
	{

			 return MRET_OK;		
	}
	if ( i == 4 )
	{
		/* All of 4 ways are occupied  */
		   // find non static Entry
	    for (i=0; i<4; i++)
	    {
			_rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, eidx + i, &entryContent);
			if ( entryContent.isStatic )
				continue;
		break;
	     }

	 }	
	entryContent.mac47_40= (mac.octet[0]);
	entryContent.mac39_24= ((mac.octet[1]) << 8) | mac.octet[2];
	entryContent.mac23_8 = (mac.octet[3]<<8)|mac.octet[4];
	if (hsb->spa>RTL8651_PHY_NUMBER)
	{
			entryContent.extMemberPort=1<<hsb->extspa;
			entryContent.memberPort=0;
	}
	else
	{
				entryContent.memberPort=1<<hsb->spa;
				entryContent.extMemberPort=0;
	}

	entryContent.isStatic=0;
	entryContent.agingTime= 0x3;
	entryContent.fid=vlanTable.fid;
	_rtl8651_addAsicEntry(TYPE_L2_SWITCH_TABLE, eidx + i, &entryContent);
	if ((hsa->dp==0 )&& (hsa->dpext==0))
	{
		 return MRET_DROP;
	}
	 return MRET_OK;
}

