/*
* Copyright c                  Realtek Semiconductor Corporation, 2005
* All rights reserved.
* 
* Program : Model code for packet parser
* Abstract : 
* Author : Tony Tzong-yn Su (tysu@realtek.com.tw)               
* $Id: parserTest.c,v 1.1.1.1 2006/07/26 05:38:22 blue Exp $
*/

#include "rtl_types.h"
#include "asicRegs.h"
#ifdef CONFIG_RTL865XC
#include "rtl865xC_tblAsicDrv.h"
#else
#include "rtl8651_tblAsicDrv.h"
#endif
#include "gdma.h"
#include "kmp.h"
#include "gdmaModel.h"
#include "virtualMac.h"
#include "rtl_glue.h"
#include "rtl_utils.h"
#include "drvTest.h"
#include "icTest.h"


#include "swNic2.h"

struct rtl_mBuf * _genPktHdr(void *buffer, uint32 id, uint32 bufsize,uint32 datalen, uint16 align){

	struct rtl_mBuf *m=NULL;
	struct rtl_pktHdr *ph=NULL;	

	m=(struct rtl_mBuf *)rtlglue_malloc(sizeof(struct rtl_mBuf));
	ph =(struct rtl_pktHdr *)rtlglue_malloc(sizeof(struct rtl_pktHdr));
	memset(m,0,sizeof(struct rtl_mBuf));
	memset(ph,0,sizeof(struct rtl_pktHdr));

	ph->ph_mbuf = m;
	ph->ph_len = datalen;
	ph->ph_flags= PKTHDR_USED;
	*((uint16 *)((uint8 *)ph)+8)=0;

	ph->ph_iphdrOffset=0;
	ph->ph_pppeTagged = 0;
	ph->ph_LLCTagged = 0;
	ph->ph_vlanTagged = 0;

	ph->ph_rxdesc=32767;
	m->m_next = NULL;
	m->m_pkthdr = ph;
	m->m_len=datalen;
	m->m_flags=MBUF_USED|MBUF_EXT|MBUF_PKTHDR;
	m->m_data = ((uint8 *)buffer)+align;
	m->m_extbuf=buffer;
	m->m_extClusterId = id;
	m->m_extsize= bufsize;

	return m;
 }

enum  flagType
{
	L2_ETH=0,
	L2_PPPOE,
	L2_SNAP,
	L2_SNAP_PPPOE,
	L2_VLAN,
	L2_VLAN_PPPOE		
};
int32 _createSourcePacket(char *data,int flags)
{
	char l2data_eth[] = {
	0x00, 0x00, 0xe8, 0x66, 0x46, 0x85,   //DMAC
	0x00, 0x0c, 0x6e, 0x3a, 0x5d, 0x66,   //SMAC
	0x08, 0x00  //Ether Type
	};

	char l2data_pppoe[] = {
	0x00, 0x00, 0xe8, 0x66, 0x46, 0x85,   //DMAC
	0x00, 0x0c, 0x6e, 0x3a, 0x5d, 0x66,   //SMAC
	0x88, 0x63,  // PPPoE Type
	0,0,0,0,0,0,  // PPPoE Header
	0,0x21 // ppp (IP Type)
	};		


	char l2data_snap[] = {
	0x00, 0x00, 0xe8, 0x66, 0x46, 0x85,   //DMAC
	0x00, 0x0c, 0x6e, 0x3a, 0x5d, 0x66,   //SMAC
	0x05,0xff,0xaa,0xaa,0x03,0,0,0, // for SNAP	
	0x08, 0x00  //Ether Type
	};	


	char l2data_snap_pppoe[] = {
	0x00, 0x00, 0xe8, 0x66, 0x46, 0x85,   //DMAC
	0x00, 0x0c, 0x6e, 0x3a, 0x5d, 0x66,   //SMAC
	0x05,0xff,0xaa,0xaa,0x03,0,0,0, // for SNAP	
	0x88, 0x63,  // PPPoE Type
	0,0,0,0,0,0,  // PPPoE Header
	0,0x21 // ppp (IP Type)
	};		


	char l2data_vlan[] = {
	0x00, 0x00, 0xe8, 0x66, 0x46, 0x85,   //DMAC
	0x00, 0x0c, 0x6e, 0x3a, 0x5d, 0x66,   //SMAC
	0x81,0x00, // VLAN Tag
	0,0xf0, // Priority, CFI, VID
	0x08, 0x00  //Ether Type
	};	

	char l2data_vlan_pppoe[] = {
	0x00, 0x00, 0xe8, 0x66, 0x46, 0x85,   //DMAC
	0x00, 0x0c, 0x6e, 0x3a, 0x5d, 0x66,   //SMAC
	0x81,0x00, // VLAN Tag
	0,0xf0, // Priority, CFI, VID
	0x88, 0x63,  // PPPoE Type
	0,0,0,0,0,0,  // PPPoE Header
	0,0x21 // ppp (IP Type)
	};		

	char l34data[] = {

	/* The following packet is a HTTP Request;
	===================================================================================
	GET / HTTP/1.1
	Accept: * / *
	Accept-Language: zh-tw
	Accept-Encoding: gzip, deflate
	User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; .NET CLR 1.1.4322)
	Host: www.google.com.tw
	Connection: Keep-Alive
	Cookie: PREF=ID=0c6c50c80799b799:TB=2:LD=zh-TW:TM=1116899852:LM=1119873714:GM=1:S=lPKPB0qpAAJvXssT
	===================================================================================*/
	
	//IP header
	0x45, 0x00, 0x01, 0x67, 0xa4, 0xd2, 0x40, 0x00, 
	0x80, 0x06, 0xaf, 0xde, 0xc0, 0xa8, 0x9a, 0x6e, 
	0x42, 0x66, 0x07, 0x63, 

	// TCP header
	0x10, 0x46, 0x00, 0x50, 0x92, 0x11, 
	0xa2, 0xdf, 0x4c, 0xa5, 0x5c, 0xdd, 0x50, 0x18,
	0x44, 0x70, 0xbc, 0x2c, 0x00, 0x00,

	// Payload
	0x47, 0x45, 0x54, 0x20, 0x2f, 0x20, 0x48, 0x54, 	0x54, 0x50, 0x2f, 0x31, 0x2e, 0x31, 0x0d, 0x0a, 	0x41, 0x63, 0x63, 0x65, 0x70, 0x74, 0x3a, 0x20, 	
	0x2a, 0x2f, 0x2a, 0x0d, 0x0a, 0x41, 0x63, 0x63, 	0x65, 0x70, 0x74, 0x2d, 0x4c, 0x61, 0x6e, 0x67, 	0x75, 0x61, 0x67, 0x65, 0x3a, 0x20, 0x7a, 0x68, 	
	0x2d, 0x74, 0x77, 0x0d, 0x0a, 0x41, 0x63, 0x63, 	0x65, 0x70, 0x74, 0x2d, 0x45, 0x6e, 0x63, 0x6f, 	0x64, 0x69, 0x6e, 0x67, 0x3a, 0x20, 0x67, 0x7a, 	
	0x69, 0x70, 0x2c, 0x20, 0x64, 0x65, 0x66, 0x6c, 	0x61, 0x74, 0x65, 0x0d, 0x0a, 0x55, 0x73, 0x65, 	0x72, 0x2d, 0x41, 0x67, 0x65, 0x6e, 0x74, 0x3a, 	
	0x20, 0x4d, 0x6f, 0x7a, 0x69, 0x6c, 0x6c, 0x61, 	0x2f, 0x34, 0x2e, 0x30, 0x20, 0x28, 0x63, 0x6f, 	0x6d, 0x70, 0x61, 0x74, 0x69, 0x62, 0x6c, 0x65, 	
	0x3b, 0x20, 0x4d, 0x53, 0x49, 0x45, 0x20, 0x36, 	0x2e, 0x30, 0x3b, 0x20, 0x57, 0x69, 0x6e, 0x64, 	0x6f, 0x77, 0x73, 0x20, 0x4e, 0x54, 0x20, 0x35, 	
	0x2e, 0x30, 0x3b, 0x20, 0x2e, 0x4e, 0x45, 0x54, 	0x20, 0x43, 0x4c, 0x52, 0x20, 0x31, 0x2e, 0x31, 	0x2e, 0x34, 0x33, 0x32, 0x32, 0x29, 0x0d, 0x0a, 	
	0x48, 0x6f, 0x73, 0x74, 0x3a, 0x20, 0x77, 0x77, 	0x77, 0x2e, 0x67, 0x6f, 0x6f, 0x67, 0x6c, 0x65, 	0x2e, 0x63, 0x6f, 0x6d, 0x2e, 0x74, 0x77, 0x0d, 	
	0x0a, 0x43, 0x6f, 0x6e, 0x6e, 0x65, 0x63, 0x74, 	0x69, 0x6f, 0x6e, 0x3a, 0x20, 0x4b, 0x65, 0x65, 	0x70, 0x2d, 0x41, 0x6c, 0x69, 0x76, 0x65, 0x0d, 	
	0x0a, 0x43, 0x6f, 0x6f, 0x6b, 0x69, 0x65, 0x3a, 	0x20, 0x50, 0x52, 0x45, 0x46, 0x3d, 0x49, 0x44, 	0x3d, 0x30, 0x63, 0x36, 0x63, 0x35, 0x30, 0x63, 	
	0x38, 0x30, 0x37, 0x39, 0x39, 0x62, 0x37, 0x39, 	0x39, 0x3a, 0x54, 0x42, 0x3d, 0x32, 0x3a, 0x4c, 	0x44, 0x3d, 0x7a, 0x68, 0x2d, 0x54, 0x57, 0x3a, 	
	0x54, 0x4d, 0x3d, 0x31, 0x31, 0x31, 0x36, 0x38, 	0x39, 0x39, 0x38, 0x35, 0x32, 0x3a, 0x4c, 0x4d, 	0x3d, 0x31, 0x31, 0x31, 0x39, 0x38, 0x37, 0x33, 	
	0x37, 0x31, 0x34, 0x3a, 0x47, 0x4d, 0x3d, 0x31, 	0x3a, 0x53, 0x3d, 0x6c, 0x50, 0x4b, 0x50, 0x42, 	0x30, 0x71, 0x70, 0x41, 0x41, 0x4a, 0x76, 0x58, 	
	0x73, 0x73, 0x54, 0x0d, 0x0a, 0x0d, 0x0a 
	};

	switch(flags)
	{
		case L2_ETH:
			memcpy(data,l2data_eth,sizeof(l2data_eth));
			memcpy(data+sizeof(l2data_eth),l34data,sizeof(l34data));	
			return sizeof(l2data_eth)+sizeof(l34data);
			break;
			
		case L2_PPPOE:
			memcpy(data,l2data_pppoe,sizeof(l2data_pppoe));
			memcpy(data+sizeof(l2data_pppoe),l34data,sizeof(l34data));	
			return sizeof(l2data_pppoe)+sizeof(l34data);			
			break;
			
		case L2_SNAP:
			memcpy(data,l2data_snap,sizeof(l2data_snap));
			memcpy(data+sizeof(l2data_snap),l34data,sizeof(l34data));	
			return sizeof(l2data_snap)+sizeof(l34data);			
			break;
			
		case L2_SNAP_PPPOE:
			memcpy(data,l2data_snap_pppoe,sizeof(l2data_snap_pppoe));
			memcpy(data+sizeof(l2data_snap_pppoe),l34data,sizeof(l34data));	
			return sizeof(l2data_snap_pppoe)+sizeof(l34data);			
			break;
			
		case L2_VLAN:
			memcpy(data,l2data_vlan,sizeof(l2data_vlan));
			memcpy(data+sizeof(l2data_vlan),l34data,sizeof(l34data));	
			return sizeof(l2data_vlan)+sizeof(l34data);			
			break;
			
		case L2_VLAN_PPPOE:
			memcpy(data,l2data_vlan_pppoe,sizeof(l2data_vlan_pppoe));
			memcpy(data+sizeof(l2data_vlan_pppoe),l34data,sizeof(l34data));	
			return sizeof(l2data_vlan_pppoe)+sizeof(l34data);				
			break;
	}


	return 0;
}


void _initHsb(hsb_t *hsb)
{
	hsb->spa=0x7;
	hsb->vid=0x0;
	hsb->tagif=0x0;
	hsb->pppoeif=0x0;
	hsb->sip=0xc0a89a6e;
	hsb->sprt=0xafde;
	hsb->dprt=0xbc2c;
	hsb->dip=0x42660763;
	hsb->ipptl=0x6;
	hsb->ipfg=0x0;
	hsb->iptos=0x0;
	hsb->tcpfg=0x18;
	hsb->type=0x5;
	hsb->patmatch=0x0;
	hsb->ethtype=0x0;
	hsb->da0=0x0; hsb->da1=0x0; hsb->da2=0xe8; hsb->da3=0x66; hsb->da4=0x46; hsb->da5=0x85;
	hsb->sa0=0x0; hsb->sa1=0xc; hsb->sa2=0x6e; hsb->sa3=0x3a; hsb->sa4=0x5d; hsb->sa5=0x66;
	hsb->hiprior=0x0;
	hsb->snap=0x0;
	hsb->llcothr=0x0;
	hsb->udpnocs=0x0;
	hsb->ttlst=0x2;
	hsb->dirtx=0x1;
	hsb->l3csok=0x0;
	hsb->l4csok=0x0;
	hsb->ipfg=0x0;
	hsb->ipfo0_n=0x0;
	hsb->urlmch=0x0;
	hsb->extspa=0x0;
	hsb->extl2=0x0;


	


}

int32 parserTagForDirectTX(uint32 caseNo)
{

	struct rtl_pktHdr	*pktHdr;
	struct rtl_mBuf	*Mblk;


	hsb_t hsb,rhsb;	
	char data[2048]={0};

//	int i,j;

	int retval=0;

	// Direct TX Correct Result (Ether type)

	_initHsb(&hsb);

	// Ether Type Parser 
	hsb.len=_createSourcePacket(data,L2_ETH);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);

	if(retval==4)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;

	// PPPoE Type Parser 
	hsb.pppoeif=0;
	hsb.len=_createSourcePacket(data,L2_PPPOE);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);
	if(retval==4)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;

	// PPPoE Type DirectTX
	hsb.len=_createSourcePacket(data,L2_PPPOE);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	pktHdr->ph_pppeTagged=1;
	pktHdr->ph_pppoeIdx=7;

	hsb.ethtype=(pktHdr->ph_portlist<<5)|(pktHdr->ph_vlanTagged<<4)|(pktHdr->ph_pppeTagged<<3)|(pktHdr->ph_pppoeIdx); // for Direct TX, this field is { portmask(7), DoVLANTAG, DoPPPoETAG, PPPoEID }
	hsb.pppoeif=1;
	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);
	if(retval==4)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;

	// SNAP Parser 
	hsb.ethtype=0;
	hsb.pppoeif=0;
	hsb.snap=0;	

	hsb.len=_createSourcePacket(data,L2_SNAP);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	
	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);

	if(retval==4)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;

	// SNAP DirectTX
	hsb.len=_createSourcePacket(data,L2_SNAP);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	pktHdr->ph_LLCTagged=1;
	hsb.snap=1;
	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);
	if(retval==4)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;

	// SNAP & PPPoE  Parser 
	hsb.pppoeif=0;
	hsb.snap=0;	
	hsb.len=_createSourcePacket(data,L2_SNAP_PPPOE);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);
	if(retval==4)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;


	//SNAP & PPPoE DirectTX
	hsb.len=_createSourcePacket(data,L2_SNAP_PPPOE);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	pktHdr->ph_LLCTagged=1;
	pktHdr->ph_pppeTagged=1;
	pktHdr->ph_pppoeIdx=7;
	hsb.snap=1;
	hsb.ethtype=(pktHdr->ph_portlist<<5)|(pktHdr->ph_vlanTagged<<4)|(pktHdr->ph_pppeTagged<<3)|(pktHdr->ph_pppoeIdx); // for Direct TX, this field is { portmask(7), DoVLANTAG, DoPPPoETAG, PPPoEID }
	hsb.pppoeif=1;
	
	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);
	if(retval==4)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;

	hsb.pppoeif=0;
	hsb.ethtype=0;
	hsb.snap=0;

	// VLAN parser
	hsb.len=_createSourcePacket(data,L2_VLAN);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);
	if(retval==4)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;

	// VLAN DirectTX
	hsb.len=_createSourcePacket(data,L2_VLAN);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	

	pktHdr=Mblk->m_pkthdr;
	pktHdr->ph_vlanTagged=1;
	pktHdr->ph_vlanID=10; 
	hsb.ethtype=(pktHdr->ph_portlist<<5)|(pktHdr->ph_vlanTagged<<4)|(pktHdr->ph_pppeTagged<<3)|(pktHdr->ph_pppoeIdx); // for Direct TX, this field is { portmask(7), DoVLANTAG, DoPPPoETAG, PPPoEID }
	hsb.vid=pktHdr->ph_vlanID;	

	
	hsb.snap=0;
	hsb.tagif=1;
	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);
	if(retval==4)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;

	//VLAN PPPOE parser
	hsb.tagif=0;
	hsb.pppoeif=0;

	pktHdr->ph_pppeTagged=0;	
	pktHdr->ph_vlanTagged=0;
	pktHdr->ph_vlanID=0;	
	

	hsb.len=_createSourcePacket(data,L2_VLAN_PPPOE);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	


	pktHdr=Mblk->m_pkthdr;
	hsb.vid=pktHdr->ph_vlanID;
	hsb.ethtype=(pktHdr->ph_portlist<<5)|(pktHdr->ph_vlanTagged<<4)|(pktHdr->ph_pppeTagged<<3)|(pktHdr->ph_pppoeIdx); // for Direct TX, this field is { portmask(7), DoVLANTAG, DoPPPoETAG, PPPoEID }	
	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);
	if(retval==4)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;	

	// VLAN DirectTX
	hsb.tagif=1;
	hsb.pppoeif=1;
	hsb.snap=0;	
	hsb.len=_createSourcePacket(data,L2_VLAN_PPPOE);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	
	pktHdr=Mblk->m_pkthdr;
	pktHdr->ph_pppeTagged=1;	
	pktHdr->ph_pppoeIdx=7;
	pktHdr->ph_vlanTagged=1;
	pktHdr->ph_vlanID=10; 
	pktHdr->ph_portlist=0x1f;

	hsb.vid=pktHdr->ph_vlanID; 
	hsb.ethtype=(pktHdr->ph_portlist<<5)|(pktHdr->ph_vlanTagged<<4)|(pktHdr->ph_pppeTagged<<3)|(pktHdr->ph_pppoeIdx); // for Direct TX, this field is { portmask(7), DoVLANTAG, DoPPPoETAG, PPPoEID }

	retval=modelPktParser(data,7,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);
	if(retval==4) //Trap to CPU
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;	
	

	return SUCCESS;

}

int32 parserPatternMatch(uint32 caseNo)
{

	struct rtl_pktHdr	*pktHdr;
	struct rtl_mBuf	*Mblk;
	hsb_t hsb,rhsb;	
	char data[2048]={0};	
	int i;
	int retval=0;
	
	//set Pattern Match
	{
		union {
			char pat[4];
			uint32 pattern;
		}u;

		u.pat[0]='T';
		u.pat[1]='T';
		u.pat[2]='P';
		u.pat[3]='/';

		for(i=0; i<RTL8651_PORT_NUMBER; i++) {
			int32 retval;
			retval = rtl8651_setAsicPortPatternMatch(i,u.pattern, 0xffffffff, 0x2);
			assert(retval == SUCCESS);
		}
	}

	// Direct TX Correct Result (Ether type)

	_initHsb(&hsb);
	hsb.spa=0x1;
	hsb.sprt=4166;
	hsb.dprt=80;
	hsb.patmatch=0x1;	
	hsb.ethtype=0x0800;
	hsb.dirtx=0x0;	
	hsb.l3csok=0x1;
	hsb.l4csok=0x1;

	// Ether Type Parser 
	hsb.len=_createSourcePacket(data,L2_ETH);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	retval=modelPktParser(data,1,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);

	if(retval==2)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;

	return SUCCESS;
}

int32 parserChecksumCheck(uint32 caseNo)
{
	struct rtl_pktHdr	*pktHdr;
	struct rtl_mBuf	*Mblk;
	hsb_t hsb,rhsb;	
	char data[2048]={0};	
	int retval=0;
	
	// Direct TX Correct Result (Ether type)

	_initHsb(&hsb);
	hsb.spa=0x1;
	hsb.sprt=4166;
	hsb.dprt=80;
	hsb.patmatch=0x1;	
	hsb.ethtype=0x0800;
	hsb.dirtx=0x0;	
	hsb.l3csok=0x1;
	hsb.l4csok=0x1;

	// Ether Type Parser (L34 OK)
	hsb.len=_createSourcePacket(data,L2_ETH);
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	retval=modelPktParser(data,1,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);

	if(retval==2)
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;


	// Ether Type Parser (L3 OK,L4 Checksum Error)
	hsb.l4csok=0x0;
	hsb.len=_createSourcePacket(data,L2_ETH);
	data[hsb.len-1]=0x0b; //modify layer 4 content
	
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	retval=modelPktParser(data,1,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);

	if(retval==0) //Checksum Error , Drop the Packet
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;


	// Ether Type Parser (L3 L4 Checksum Error)
	hsb.l3csok=0x0;
	hsb.l4csok=0x0;	
	hsb.len=_createSourcePacket(data,L2_ETH);
	data[26]=1; //modify the IP address of IP packet
	hsb.sip=0x01a89a6e;
	
	Mblk=_genPktHdr((void*)data,(uint32)0,2048, hsb.len,0);	
	pktHdr=Mblk->m_pkthdr;
	retval=modelPktParser(data,1,0,pktHdr->ph_len,pktHdr);
	rtlglue_free(Mblk);
	rtlglue_free(pktHdr);

	if(retval==0) //Checksum Error , Drop the Packet
	{
		modelGetHsb(&rhsb);
		retval=compHsb(&rhsb,&hsb);	
		if(retval!=SUCCESS) return retval;
	}	
	else return FAILED;

	return SUCCESS;	
}

int32 parserUrlTrap(uint32 caseNo)
{

	/* not be implemented */
	
	return SUCCESS;
}

