/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : Header File for Virtual MAC
* Abstract : 
* Author : Louis Yung-Chieh Lo (yjlou@realtek.com.tw)               
* $Id: virtualMac.c,v 1.1.1.1 2006/07/26 05:38:22 blue Exp $
*/

#include <rtl_types.h>
#include <rtl_glue.h>
#ifdef CONFIG_RTL865XC
#include "rtl865xC_tblAsicDrv.h"
#else
#include "rtl8651_tblAsicDrv.h"
#endif
#include "virtualMac.h"
#include "icModel.h"
#include "hsModel.h"


/* The current emulated IC type.
 *	IC_TYPE_MODEL:
 *		MacInput(), MacOutput(), HSA, HSB, READ_MEM32(), and WRITE_MEM32() functions are mapped to 
 *		model code, not the phisical IC/address.
 *
 *	IC_TYPE_REAL:
 *		MacInput(), MacOutput(), HSA, HSB, READ_MEM32(), and WRITE_MEM32() functions are mapped to 
 *		real IC/address.
 */
enum IC_TYPE curIcType;


#if defined(RTL865X_MODEL_KERNEL)
static int32 cpu_status_is_set;
static uint32 cpu_status;
#endif

int32 model_setTestTarget( enum IC_TYPE ic_type )
{
	switch ( ic_type )
	{
		case IC_TYPE_MODEL:

#if defined(RTL865X_MODEL_KERNEL)
			/* Disable All interrupt to avoid race condition. */
			cpu_status = read_c0_status();
			cpu_status_is_set = TRUE;
			/*rtlglue_printf( "Status saved: 0x%08x\n", cpu_status );*/
#endif

			_rtl8651_mapToVirtualRegSpace();
			curIcType = ic_type;
			break;
			
#if defined(RTL865X_MODEL_USER)
		case IC_TYPE_REAL:
			return VMAC_ERROR_NON_AVAILABLE;
#elif defined(RTL865X_MODEL_KERNEL)
		case IC_TYPE_REAL:
			_rtl8651_mapToRealRegSpace();
			curIcType = ic_type;

			/* Enable All interrupt to avoid race condition. */
			if ( cpu_status_is_set )
			{
				/*rtlglue_printf( "Status recovered: 0x%08x\n", cpu_status );*/
				cpu_status_is_set = FALSE;
			}
			
			return SUCCESS;
			break;
#endif

		default:
			return VMAC_ERROR_NON_AVAILABLE;
	}
	
	return SUCCESS;
}


int32 model_getTestTarget( enum IC_TYPE *pic_type )
{
	if ( pic_type ) *pic_type = curIcType;
	return SUCCESS;
}


/*
* The Implementation of Virtual Mac Interface 
* 
* This interface provides uniform interface to input/output packets.
* The user of this interface is test bench and 865xB emulation board.
* And this interface will abstract the difference between IC model code, real IC (865xC), and FPGA.
* 
*/
int32 virtualMacInput( enum PORT_NUM fromPort, uint8* packet, int32 len )
{
	switch ( curIcType )
	{
		case IC_TYPE_MODEL:

			{
				int srcPortNum,srcExtPortNum;

				if(fromPort<=PN_PORT5)
				{
					srcPortNum=fromPort;
					srcExtPortNum=0;
					modelPktParser(packet,srcPortNum,srcExtPortNum,len,NULL);
				}

				else if((fromPort>=PN_PORT_EXT0)&&(fromPort<=PN_PORT_EXT2))
				{
					// from ExtPort
					srcPortNum=7;				
					srcExtPortNum=fromPort - PN_PORT_EXT0 + 1;					
				}

				else if(fromPort==PN_PORT_CPU)
				{

					// Direct TX
					srcPortNum=7;				
					srcExtPortNum=0;					
				}
			}

			/* HSB is updated. */
			modelStart();
			
			break;
#ifdef RTL865X_MODEL_USER
		case IC_TYPE_REAL:
			return VMAC_ERROR_NON_AVAILABLE;
#elif defined(RTL865X_MODEL_KERNEL)
		case IC_TYPE_REAL:
			break;
#endif
		default:
			return FAILED;
	}

	return SUCCESS;
}


int32 virtualMacOutput( enum PORT_MASK *toPort, uint8* packet, int32 *len )
{
	switch ( curIcType )
	{
		case IC_TYPE_MODEL:
			return modelPktTranslator( toPort, packet, len );
#ifdef RTL865X_MODEL_USER
		case IC_TYPE_REAL:
			return VMAC_ERROR_NON_AVAILABLE;
#elif defined(RTL865X_MODEL_KERNEL)
		case IC_TYPE_REAL:
			return SUCCESS;
#endif
		default:
			return FAILED;
	}
}


int32 virtualMacGetHsb( hsb_t* hsb )
{
	switch ( curIcType )
	{
		case IC_TYPE_MODEL:
			return modelGetHsb( hsb );
#ifdef RTL865X_MODEL_USER
		case IC_TYPE_REAL:
			return VMAC_ERROR_NON_AVAILABLE;
#elif defined(RTL865X_MODEL_KERNEL)
		case IC_TYPE_REAL:
			return SUCCESS;
#endif
		default:
			return FAILED;
	}
}


int32 virtualMacSetHsb( hsb_t* hsb )
{
	switch ( curIcType )
	{
		case IC_TYPE_MODEL:
			return modelSetHsb( hsb );
#ifdef RTL865X_MODEL_USER
		case IC_TYPE_REAL:
			return VMAC_ERROR_NON_AVAILABLE;
#elif defined(RTL865X_MODEL_KERNEL)
		case IC_TYPE_REAL:
			return SUCCESS;
#endif
		default:
			return FAILED;
	}
}


int32 virtualMacGetHsa( hsa_t* hsa )
{
	switch ( curIcType )
	{
		case IC_TYPE_MODEL:
			return modelGetHsa( hsa );
#ifdef RTL865X_MODEL_USER
		case IC_TYPE_REAL:
			return VMAC_ERROR_NON_AVAILABLE;
#elif defined(RTL865X_MODEL_KERNEL)
		case IC_TYPE_REAL:
			return SUCCESS;
#endif
		default:
			return FAILED;
	}
}


int32 virtualMacSetHsa( hsa_t* hsa )
{
	switch ( curIcType )
	{
		case IC_TYPE_MODEL:
			return modelSetHsa( hsa );
#ifdef RTL865X_MODEL_USER
		case IC_TYPE_REAL:
			return VMAC_ERROR_NON_AVAILABLE;
#elif defined(RTL865X_MODEL_KERNEL)
		case IC_TYPE_REAL:
			return SUCCESS;
#endif
		default:
			return FAILED;
	}
}



