/*
* Copyright c                  Realtek Semiconductor Corporation, 2004  
* All rights reserved.
* 
* Program : Control MII & MDC/MDIO connected 8305SB/8305SC commands
* Abstract : 
* Author : Edward Jin-Ru Chen (jzchen@realtek.com.tw)               
* $Id: rtl8305s_cmd.c,v 1.1.1.1 2006/07/26 05:38:23 blue Exp $
*/
#include "rtl_types.h"
#include "types.h"
#include "rtl_cle.h" 
#include "rtl8305s.h"
#ifdef CONFIG_RTL8305SB
#include "rtl8305sb.h"
#endif
#ifdef CONFIG_RTL8305SC
#include "rtl8305sc.h"
#endif

static int32 _rtl8305s_readCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken;
	uint8 phyAd, regAd;
	uint16 regData, expData;
	int32 size, repeatTimes, i;
	
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(size) {
		phyAd = U8_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		regAd = U8_value(nextToken);
		rtl8305s_smiRead(phyAd, regAd, &regData);
		printk("Read 8305S PHY address %02X Register address %02X data %04X\n", phyAd, regAd, regData);
	}
	else if (strcmp(nextToken, "repeat") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		repeatTimes = U32_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		phyAd = U8_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		regAd = U8_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		expData = U16_value(nextToken);
		for(i=0; i<repeatTimes; i++) {
			rtl8305s_smiRead(phyAd, regAd, &regData);
			printk("\rRead %04dth times ", i);
			if(regData != expData)
				printk(" Read %04X but expect %04X\n", regData, expData);
		}
		printk(" read finished!\n");
	} 
	return SUCCESS;
}

static int32 _rtl8305s_writeCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken;
	uint8 phyAd, regAd;
	uint16 regData;
	int32 size;
	
	rtl8305s_init();
	cle_getNextCmdToken(&nextToken,&size,saved); 
	phyAd = U8_value(nextToken);
	cle_getNextCmdToken(&nextToken,&size,saved); 
	regAd = U8_value(nextToken);
	cle_getNextCmdToken(&nextToken,&size,saved); 
	regData = U16_value(nextToken);
	printk("Write 8305S PHY address %02X Register address %02X data %04X\n", phyAd, regAd, regData);
	rtl8305s_smiWrite(phyAd, regAd, regData);
	rtl8305s_smiRead(phyAd, regAd, &regData);
	printk("Read back 8305S PHY address %02X Register address %02X data %04X\n", phyAd, regAd, regData);
	return SUCCESS;
}

static int32 _rtl8305s_phyCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken, autoNegotiation, fullDuplex, status;
	uint16 port;
	int32 size;
	uint32 advCapability, speed; 
	
	cle_getNextCmdToken(&nextToken,&size,saved); 
	port = U16_value(nextToken);
	printk("Operation over port %d\n", port);
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if (strcmp(nextToken, "info") == 0) {
		if(rtl8305s_getAsicEthernetPHY(port, &autoNegotiation, &advCapability, &speed, &fullDuplex) == FAILED)
			return FAILED;
		rtl8305s_getAsicPHYLinkStatus(port, &status);
		printk("Link %s\n", status == TRUE? "Up":"Down");
		if(autoNegotiation == TRUE) {
			printk("Auto-negotiation enabled and advertised capability is ");
			switch(advCapability) {
				case RTL8305S_ETHER_AUTO_100FULL:
					printk("100Mbps Full-duplex\n");
				break;
				case RTL8305S_ETHER_AUTO_100HALF:
					printk("100Mbps Half-duplex\n");
				break;
				case RTL8305S_ETHER_AUTO_10FULL	:
					printk("10Mbps Full-duplex\n");
				break;
				case RTL8305S_ETHER_AUTO_10HALF:
					printk("10Mbps Half-duplex\n");
				break;
				default:
				printk("Unknown\n");
				break;
			}
			if(status == TRUE) {
				printk("Link current speed %dMbps %s-duplex\n", speed, fullDuplex==TRUE?"Full":"Half");
			}
			rtl8305s_getAsicPHYAutoNegotiationDone(port, &status);
			printk("Auto-negotiation is %s\n", status==TRUE?"Completed":"not complteted");
		}
		else
			printk("Force mode at %dMbps %s-duplex", speed, fullDuplex==TRUE?"Full":"Half");
		rtl8305s_getAsicPHYLoopback(port, &status);
		printk("Loopback %s\n", status==TRUE?"Enabled":"Disabled");
	} 
	else if (strcmp(nextToken, "loopback") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "on") == 0)
			rtl8305s_setAsicPHYLoopback(port, TRUE);
		else if (strcmp(nextToken, "off") == 0) 
			rtl8305s_setAsicPHYLoopback(port, FALSE);
	} 
	else if (strcmp(nextToken,"auto-negotiated") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken,"100Full") == 0)
			rtl8305s_setAsicEthernetPHY(port, TRUE, RTL8305S_ETHER_AUTO_100FULL, 100, TRUE);
		else if (strcmp(nextToken,"100Half") == 0)
			rtl8305s_setAsicEthernetPHY(port, TRUE, RTL8305S_ETHER_AUTO_100HALF, 100, TRUE);
		else if (strcmp(nextToken,"10Full") == 0)
			rtl8305s_setAsicEthernetPHY(port, TRUE, RTL8305S_ETHER_AUTO_10FULL, 100, TRUE);
		else if (strcmp(nextToken,"10Half") == 0)
			rtl8305s_setAsicEthernetPHY(port, TRUE, RTL8305S_ETHER_AUTO_10HALF, 100, TRUE);
		else
			rtl8305s_setAsicEthernetPHY(port, TRUE, RTL8305S_ETHER_AUTO_100FULL, 100, TRUE);
	}
	else if (strcmp(nextToken,"force") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken,"100Full") == 0)
			rtl8305s_setAsicEthernetPHY(port, FALSE, RTL8305S_ETHER_AUTO_100FULL, 100, TRUE);
		else if (strcmp(nextToken,"100Half") == 0)
			rtl8305s_setAsicEthernetPHY(port, FALSE, RTL8305S_ETHER_AUTO_100FULL, 100, FALSE);
		else if (strcmp(nextToken,"10Full") == 0)
			rtl8305s_setAsicEthernetPHY(port, FALSE, RTL8305S_ETHER_AUTO_100FULL, 10, TRUE);
		else if (strcmp(nextToken,"10Half") == 0)
			rtl8305s_setAsicEthernetPHY(port, FALSE, RTL8305S_ETHER_AUTO_100FULL, 10, FALSE);
		else
			rtl8305s_setAsicEthernetPHY(port, FALSE, RTL8305S_ETHER_AUTO_100FULL, 100, TRUE);
	}
	return SUCCESS;
}

#ifdef CONFIG_RTL8305SC

static int32 _rtl8305sc_vlanCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken, enabled;
	uint16 vid, memberPortMask, port, value, index;
	int32 size;
	
	rtl8305s_init();
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(size) {
		vid = U16_value(nextToken);
		printk("Operation over VLAN ID %d\n", vid);
	}
	else if (strcmp(nextToken, "info") == 0) {
		rtl8305sc_getAsicVlanEnable(&enabled);
		printk("VLAN is %s\n", enabled==TRUE? "Enabled":"Disabled");
		rtl8305sc_getAsicVlanTagAware(&enabled);
		printk("VLAN is Tag %s\n", enabled==TRUE? "awared":"not awared");
		rtl8305sc_getAsicVlanIngressFilter(&enabled);
		printk("VLAN ingress filter is %s\n", enabled==TRUE?"Enabled":"Disabled");
		rtl8305sc_getAsicVlanTaggedOnly(&enabled);
		printk("VLAN support %s\n", enabled==TRUE?"Tagged only":"Both tagged and non-tagged");
		rtl8305sc_getAsicVlanPriorityThreshold(&value);
		printk("Packet 802.1p priority threshold %d\n", value);
		for(index= 0; index < 16; index++) {
			rtl8305sc_getAsicVlan(index, &vid, &memberPortMask);
			printk("[%2d] VLAN ID %3d Member port ", index, vid);
			for(port = 0; port < RTL8305S_PORT_NUMBER; port ++) {
				if(memberPortMask & (1<<port))
					printk(" %d", port);
			}
			printk("\n");
		}
		for(port = 0; port < RTL8305S_PORT_NUMBER; port++) {
			rtl8305sc_getAsicPortVlanIndex(port, &index);
			rtl8305sc_getAsicVlan(index, &vid, &memberPortMask);
			printk("Port %d VLAN index %2d VID %3d\n", port, index, vid);
		}
	}
	else if (strcmp(nextToken, "threshold") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		value = U16_value(nextToken);
		rtl8305sc_setAsicVlanPriorityThreshold(value);
	}
	else if (strcmp(nextToken, "set") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "port") == 0) {
			cle_getNextCmdToken(&nextToken,&size,saved); 
			port = U16_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			index = U16_value(nextToken);
			rtl8305sc_setAsicPortVlanIndex(port, index);
		}
		else {
			cle_getNextCmdToken(&nextToken,&size,saved); 
			index = U16_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			vid = U16_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			memberPortMask= U16_value(nextToken);
			rtl8305sc_setAsicVlan(index, vid, memberPortMask);
		}
		rtl8305sc_setAsicVlanPriorityThreshold(value);
	}
	return SUCCESS;
}

static int32 _rtl8305sc_macAddrTblCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken, macAddr[6], isStatic, isValid;
	uint16 port, age, index;
	int32 size;
	
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(size) {
		index = U16_value(nextToken);
		if(rtl8305sc_getAsicMacAddressTableEntry(index, macAddr, &age, &isStatic, &port, &isValid) == FAILED)
			return FAILED;
		if(isValid == TRUE) {
			printk("MAC Addr: %02X:%02X:%02X:%02X:%02X:%02X Age: %03d sec %s Port: %d\n",
				macAddr[0]&0xFF, macAddr[1]&0xFF, macAddr[2]&0xFF, macAddr[3]&0xFF, macAddr[4]&0xFF, macAddr[5]&0xFF, 
				age, isStatic==TRUE?"STATIC":"", port);
		}
		else
			printk("Index %d does not have valid entry\n", index);
		
	}
	else if (strcmp(nextToken, "all") == 0) {
		int32 i;
		for(index = 0; index<1024; index++) {
			for(i=0; i<1000; i++);
			if(rtl8305sc_getAsicMacAddressTableEntry(index, macAddr, &age, &isStatic, &port, &isValid) == FAILED)
				return FAILED;
			if(isValid == TRUE) {
				printk("[%04d] MAC Addr: %02X:%02X:%02X:%02X:%02X:%02X Age: %03d sec %s Port: %d\n",
					index, macAddr[0]&0xFF, macAddr[1]&0xFF, macAddr[2]&0xFF, macAddr[3]&0xFF, macAddr[4]&0xFF, macAddr[5]&0xFF, 
					age, isStatic==TRUE?"STATIC":"", port);
			}
		}
	}
	else if (strcmp(nextToken, "dynamic") == 0)  {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		memcpy(macAddr, nextToken, 6);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		port = U16_value(nextToken);
		printk("Add Dynamic MAC Addr: %02X:%02X:%02X:%02X:%02X:%02X Port: %d\n", 
			macAddr[0]&0xFF, macAddr[1]&0xFF, macAddr[2]&0xFF, macAddr[3]&0xFF, macAddr[4]&0xFF, macAddr[5]&0xFF, 
			port);
		rtl8305sc_setAsicMacAddressTableEntry(macAddr, 300, FALSE, port);
	}
	else if (strcmp(nextToken, "static") == 0)  {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		memcpy(macAddr, nextToken, 6);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		port = U16_value(nextToken);
		printk("Add Static MAC Addr: %02X:%02X:%02X:%02X:%02X:%02X Port: %d\n", 
			macAddr[0]&0xFF, macAddr[1]&0xFF, macAddr[2]&0xFF, macAddr[3]&0xFF, macAddr[4]&0xFF, macAddr[5]&0xFF, 
			port);
		rtl8305sc_setAsicMacAddressTableEntry(macAddr, 300, TRUE, port);
	}
	else if (strcmp(nextToken, "delete") == 0)  {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		memcpy(macAddr, nextToken, 6);
		printk("Delete MAC Addr: %02X:%02X:%02X:%02X:%02X:%02X \n", 
			macAddr[0]&0xFF, macAddr[1]&0xFF, macAddr[2]&0xFF, macAddr[3]&0xFF, macAddr[4]&0xFF, macAddr[5]&0xFF);
			rtl8305sc_setAsicMacAddressTableEntry(macAddr, 0, FALSE, 0);
	}
	return SUCCESS;
}

static int32 _rtl8305sc_rtctCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken, isTxDir;
	uint16 port;
	int32 size;
	uint32 cableStatus, length;
	
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(size) {
		port = U16_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "tx") == 0)
			isTxDir = TRUE;
		else
			isTxDir = FALSE;
		if(rtl8305sc_asicRealtekCableTester(port, isTxDir, &cableStatus, &length) == FAILED) 
			printk("Unable to get port %d %s Realtek cable testing result\n", port, isTxDir==TRUE?"Tx":"Rx");
		else
			printk("Port %x %s status: %s Length %d m %d cm\n", port, isTxDir==TRUE?"Tx":"Rx", 
				cableStatus==RTL8305SC_RTCT_OPEN ? "Open":cableStatus==RTL8305SC_RTCT_SHORT? "Short":"Normal", 
				length/2, (length%2)*50);
	}
	else if (strcmp(nextToken, "all") == 0) {
		for(port=0; port<RTL8305S_PORT_NUMBER; port++)
			for(isTxDir=0; isTxDir<2; isTxDir++) {
				if(rtl8305sc_asicRealtekCableTester(port, isTxDir, &cableStatus, &length) == FAILED) 
					printk("Unable to get port %d %s Realtek cable testing result\n", port, isTxDir==TRUE?"Tx":"Rx");
				else
					printk("Port %x %s status: %s Length %d m %d cm\n", port, isTxDir==TRUE?"Tx":"Rx", 
						cableStatus==RTL8305SC_RTCT_OPEN ? "Open":cableStatus==RTL8305SC_RTCT_SHORT? "Short":"Normal", 
						length/2, (length%2)*50);
			}
	}
	return SUCCESS;
}

static int32 _rtl8305sc_miscCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken, enabled;
	int32 size;
	
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if (strcmp(nextToken, "info") == 0) {
		rtl8305sc_getAsicLoopDetection(&enabled);
		printk("Loopback detection is %s\n", enabled==TRUE?"Enabled":"Disabled");
		rtl8305sc_getAsicAgressiveBackoff(&enabled);
		printk("Aggressive back-off is %s\n", enabled==TRUE?"Enabled":"Disabled"); 
		rtl8305sc_getAsicLongReceivedPacketLength(&enabled);
		printk("Received packet length is %d\n", enabled==TRUE?1552:1536);
		rtl8305sc_getAsicBroadcastInputDrop(&enabled);
		printk("Drop broadcast packet at %s\n", enabled==TRUE?"Input":"Output");
		rtl8305sc_getAsicForward1dReservedAddress(&enabled);
		printk("%s 802.1d reserved MAC addresses\n", enabled==TRUE?"Forward":"Filter");
		rtl8305sc_getAsicLeakyVlan(&enabled);
		printk("Leaky VLAN is %s\n", enabled==TRUE?"Enabled":"Disabled");
		rtl8305sc_getAsicArpVlan(&enabled);
		printk("ARP VLAN is %s\n", enabled==TRUE?"Enabled":"Disabled");
		rtl8305sc_getAsicBroadcastStormControl(&enabled);
		printk("Broadcast storm control is %s\n", enabled==TRUE?"Enabled":"Disabled");
	}
	else if (strcmp(nextToken, "loopbackDetection") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "enable") == 0)
			rtl8305sc_setAsicLoopDetection(TRUE);
		else
			rtl8305sc_setAsicLoopDetection(FALSE);
	}
	else if (strcmp(nextToken, "aggressiveBackoff") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "enable") == 0)
			rtl8305sc_setAsicAgressiveBackoff(TRUE);
		else
			rtl8305sc_setAsicAgressiveBackoff(FALSE);
	}
	else if (strcmp(nextToken, "longReceive") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "enable") == 0)
			rtl8305sc_setAsicLongReceivedPacketLength(TRUE);
		else
			rtl8305sc_setAsicLongReceivedPacketLength(FALSE);
	}
	else if (strcmp(nextToken, "broadcastInputDrop") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "enable") == 0)
			rtl8305sc_setAsicBroadcastInputDrop(TRUE);
		else
			rtl8305sc_setAsicBroadcastInputDrop(FALSE);
	}
	else if (strcmp(nextToken, "forward1d") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "enable") == 0)
			rtl8305sc_setAsicForward1dReservedAddress(TRUE);
		else
			rtl8305sc_setAsicForward1dReservedAddress(FALSE);
	}
	else if (strcmp(nextToken, "leakyVlan") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "enable") == 0)
			rtl8305sc_setAsicLeakyVlan(TRUE);
		else
			rtl8305sc_setAsicLeakyVlan(FALSE);
	}
	else if (strcmp(nextToken, "arpVlan") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "enable") == 0)
			rtl8305sc_setAsicArpVlan(TRUE);
		else
			rtl8305sc_setAsicArpVlan(FALSE);
	}
	else if (strcmp(nextToken, "broadcastStormControl") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "enable") == 0)
			rtl8305sc_setAsicBroadcastStormControl(TRUE);
		else
			rtl8305sc_setAsicBroadcastStormControl(FALSE);
	}
	else if (strcmp(nextToken, "softReset") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "enable") == 0)
			rtl8305sc_asicSoftReset();
	}
	return SUCCESS;
}

#endif

cle_exec_t rtl8305s_cmds[] = {
	{
		"read",
		"Read register",
		"{ %d'PHY address' %d'Reg address' | repeat %d'Repeat reat times' %d'PHY address' %d'Reg address' %d'Expected read value' } ",
		_rtl8305s_readCmd,
		CLE_USECISCOCMDPARSER,	
		0,
		NULL
	},
	{
		"write",
		"Configure register",
		"%d'PHY address' %d'Reg address' %d'16-bit data to configure' ",
		_rtl8305s_writeCmd,
		CLE_USECISCOCMDPARSER,	
		0,
		NULL
	},
	{
		"phy",
		"Configure and get PHY setting.",
		"%d'Port number(s)' { info' Port information' | loopback'Configure port into loopback mode' { on | off } | auto-negotiated { [ 100Full | 100Half | 10Full | 10Half ] } | force { 10Half | 10Full | 100Half | 100Full } }",
		_rtl8305s_phyCmd,
		CLE_USECISCOCMDPARSER,	
		0,
		NULL
	},
#ifdef CONFIG_RTL8305SB
#endif
#ifdef CONFIG_RTL8305SC
	{
		"vlan",
		"Configure VLAN.",
		" {  %d'VLAN ID' | info | threshold %d'The priority threshold' | set { port %d %d'VLAN index' | vindex %d'VLAN index' %d'VID' %d'Member port mask' } } ",
		_rtl8305sc_vlanCmd,
		CLE_USECISCOCMDPARSER,	
		0,
		NULL
	},
	{
		"macAddressTable",
		"Dump MAC address table",
		" {  %d'Entry index' | all | static %mac'MAC to configure' %d'Port of this host address' | dynamic %mac'MAC to configure' %d'Port of this host address' | delete %mac'MAC address to delete' } ",
		_rtl8305sc_macAddrTblCmd,
		CLE_USECISCOCMDPARSER,	
		0,
		NULL
	},
	{
		"rtct",
		"Realtek Cable Tester",
		" { %d'Port to execute cable tester' { tx | rx } | all } ",
		_rtl8305sc_rtctCmd,
		CLE_USECISCOCMDPARSER,	
		0,
		NULL
	},
	{
		"misc",
		"Other controls",
		" { info | { loopbackDetection | aggressiveBackoff | longReceive | broadcastInputDrop | forward1d | leakyVlan | arpVlan | broadcastStormControl | softReset } { enable | disable } } ",
		_rtl8305sc_miscCmd,
		CLE_USECISCOCMDPARSER,	
		0,
		NULL
	},
#endif
};


