/*
* Copyright c                  Realtek Semiconductor Corporation, 2004  
* All rights reserved.
* 
* Program : Control MII & MDC/MDIO connected 8305SC
* Abstract : 
* Author : Edward Jin-Ru Chen (jzchen@realtek.com.tw)               
* $Id: rtl8305sc.c,v 1.1.1.1 2006/07/26 05:38:23 blue Exp $
*/

/*
	Implementation note:
	1. Using GPIO Port A Pin 6 to emulate MDIO and GPIO Port A Pin 7 to emulate MDC
	2. 8305SC Port 4 connect to 8650B Port 5 (MII)
*/

/*	@doc RTL8305SC_API

	@module rtl8305sc.c - RTL8305SC 5-port switch controller driver API documentation	|
	This document explains the API interface of the driver module. 
	API included in this document is special APIs for RTL8305SC excluding common APIs for 
	both RTL8305SB and RTL8305SC.

	@normal Edward Jin-Ru Chen (jzchen@realtek.com.tw) <date>

	Copyright <cp>2004 Realtek<tm> Semiconductor Cooperation, All Rights Reserved.

 	@head3 List of Symbols |
 	Here is a list of all functions and variables in this module.

 	@index | RTL8305SC_API
*/

#include "rtl8305s.h"
#include "rtl8305sc.h"
#include "assert.h"

#define RTL8305SC_ASIC_DRIVER

/*
@func int32 | rtl8305sc_setAsicVlanEnable | Configure switch to be VLAN switch.
@parm int8 | enabled | Configure 8305SC to be VLAN switch or not.
@comm
*/
int32 rtl8305sc_setAsicVlanEnable(int8 enabled) {
	rtl8305s_smiWriteBit(0, 18, 8, enabled==FALSE?1:0);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicVlanEnable | Get whether switch is VLAN switch.
@parm int8 * | enabled | 8305SC is VLAN switch or not.
@comm
*/
int32 rtl8305sc_getAsicVlanEnable(int8 *enabled) {
	uint8 bitValue;
	
	rtl8305s_smiReadBit(0, 18, 8, &bitValue);
	*enabled = bitValue == 0? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicVlanTagAware | Configure switch to be VLAN tag awared.
@parm int8 | enabled | Configure 8305SC VLAN tag awared
@comm
*/
int32 rtl8305sc_setAsicVlanTagAware(int8 enabled) {
	rtl8305s_smiWriteBit(0, 16, 10, enabled==FALSE?1:0);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicVlanTagAware | Get switch VLAN tag awared setting.
@parm int8 * | enabled | 8305SC is VLAN tag awared or not.
@comm
*/
int32 rtl8305sc_getAsicVlanTagAware(int8 *enabled) {
	uint8 bitValue;
	
	rtl8305s_smiReadBit(0, 16, 10, &bitValue);
	*enabled = bitValue == 0? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicVlanIngressFilter | Configure switch to be VLAN ingress filtered or not.
@parm int8 | enabled | Configure 8305SC VLAN ingress filter
@comm
*/
int32 rtl8305sc_setAsicVlanIngressFilter(int8 enabled) {
	rtl8305s_smiWriteBit(0, 16, 9, enabled==FALSE?1:0);
	return SUCCESS;
}
/*
@func int32 | rtl8305sc_getAsicVlanIngressFilter | Get switch VLAN ingress filter setting.
@parm int8 * | enabled | 8305SC is VLAN ingress filter or not.
@comm
*/
int32 rtl8305sc_getAsicVlanIngressFilter(int8 *enabled) {
	uint8 bitValue;
	
	rtl8305s_smiReadBit(0, 16, 9, &bitValue);
	*enabled = bitValue == 0? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicVlanTaggedOnly | Configure switch to be VLAN tagged only or not.
@parm int8 | enabled | Configure 8305SC receive VLAN tagged only
@comm
*/
int32 rtl8305sc_setAsicVlanTaggedOnly(int8 enabled) {
	rtl8305s_smiWriteBit(0, 16, 8, enabled==FALSE?1:0);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicVlanTaggedOnly | Get switch VLAN tagged only setting.
@parm int8 * | enabled | 8305SC is VLAN tagged only or not.
@comm
*/
int32 rtl8305sc_getAsicVlanTaggedOnly(int8 *enabled) {
	uint8 bitValue;
	
	rtl8305s_smiReadBit(0, 16, 8, &bitValue);
	*enabled = bitValue == 0? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicVlanPriorityThreshold | Set the VLAN priority threshold.
@parm uint16 | highPriorityThreshold | The priroty threshold value
@comm
This function is used to configure the threshold value used to compare with VLAN tag priority.
If priority is larger or equal to the specified value, packet is classified to be high priority.
*/
int32 rtl8305sc_setAsicVlanPriorityThreshold(uint16 highPriorityThreshold) {
	uint16 value;

	rtl8305s_smiRead(4, 21, &value);
	value = (value & 0x1FFF) | ((highPriorityThreshold<<13)&0xE000);
	rtl8305s_smiWrite(4, 21, value);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicVlanPriorityThreshold | Get the VLAN priority threshold.
@parm uint16 * | highPriorityThreshold | The priroty threshold value
@comm
This function is used to get the threshold value used to compare with VLAN tag priority.
If priority is larger or equal to this value, packet is classified to be high priority.
*/
int32 rtl8305sc_getAsicVlanPriorityThreshold(uint16 * highPriorityThreshold) {
	
	rtl8305s_smiRead(4, 21, highPriorityThreshold);
	*highPriorityThreshold = (*highPriorityThreshold >> 13)&0x7;

	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicVlan | Configure switch VLAN ID and corresponding member ports.
@parm uint16 | vlanIndex | Specify the position to configure VLAN
@parm uint16 | vid | Specify the VLAN ID
@parm uint16 | memberPortMask | Specify the member port list of corresponding VLAN ID
@comm
	Total have 16 entries to specify the VID and corresponding member port. This API will not
check whether configured VID is duplicated or not.
*/
int32 rtl8305sc_setAsicVlan(uint16 vlanIndex, uint16 vid, uint16 memberPortMask) {
	uint16 regValue;

	switch(vlanIndex) {
		case 0:/* VLAN[A] */
		rtl8305s_smiRead(0, 25, &regValue);
		regValue = (regValue & 0xF000) | (vid & 0xFFF);
		rtl8305s_smiWrite(0, 25, regValue);
		rtl8305s_smiRead(0, 24, &regValue);
		regValue = (regValue & 0xFFE0) | (memberPortMask & 0x1F);
		rtl8305s_smiWrite(0, 24, regValue);
		break;
		case 1:/* VLAN[B] */
		rtl8305s_smiRead(1, 25, &regValue);
		regValue = (regValue & 0xF000) | (vid & 0xFFF);
		rtl8305s_smiWrite(1, 25, regValue);
		rtl8305s_smiRead(1, 24, &regValue);
		regValue = (regValue & 0xFFE0) | (memberPortMask & 0x1F);
		rtl8305s_smiWrite(1, 24, regValue);
		break;
		case 2:/* VLAN[C] */
		rtl8305s_smiRead(2, 25, &regValue);
		regValue = (regValue & 0xF000) | (vid & 0xFFF);
		rtl8305s_smiWrite(2, 25, regValue);
		rtl8305s_smiRead(2, 24, &regValue);
		regValue = (regValue & 0xFFE0) | (memberPortMask & 0x1F);
		rtl8305s_smiWrite(2, 24, regValue);
		break;
		case 3:/* VLAN[D] */
		rtl8305s_smiRead(3, 25, &regValue);
		regValue = (regValue & 0xF000) | (vid & 0xFFF);
		rtl8305s_smiWrite(3, 25, regValue);
		rtl8305s_smiRead(3, 24, &regValue);
		regValue = (regValue & 0xFFE0) | (memberPortMask & 0x1F);
		rtl8305s_smiWrite(3, 24, regValue);
		break;
		case 4:/* VLAN[E] */
		rtl8305s_smiRead(4, 25, &regValue);
		regValue = (regValue & 0xF000) | (vid & 0xFFF);
		rtl8305s_smiWrite(4, 25, regValue);
		rtl8305s_smiRead(4, 24, &regValue);
		regValue = (regValue & 0xFFE0) | (memberPortMask & 0x1F);
		rtl8305s_smiWrite(4, 24, regValue);
		break;
		case 5:/* VLAN[F] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(0, 27, vid);
		rtl8305s_smiWrite(0, 26, memberPortMask);
		break;
		case 6:/* VLAN[G] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(1, 27, vid);
		rtl8305s_smiWrite(1, 26, memberPortMask);
		break;
		case 7:/* VLAN[H] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(2, 27, vid);
		rtl8305s_smiWrite(2, 26, memberPortMask);
		break;
		case 8:/* VLAN[I] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(3, 27, vid);
		rtl8305s_smiWrite(3, 26, memberPortMask);
		break;
		case 9:/* VLAN[J] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(4, 27, vid);
		rtl8305s_smiWrite(4, 26, memberPortMask);
		break;
		case 10:/* VLAN[K] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(0, 29, vid);
		rtl8305s_smiWrite(0, 28, memberPortMask);
		break;
		case 11:/* VLAN[L] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(1, 29, vid);
		rtl8305s_smiWrite(1, 28, memberPortMask);
		break;
		case 12:/* VLAN[M] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(2, 29, vid);
		rtl8305s_smiWrite(2, 28, memberPortMask);
		break;
		case 13:/* VLAN[N] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(3, 29, vid);
		rtl8305s_smiWrite(3, 28, memberPortMask);
		break;
		case 14:/* VLAN[O] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(4, 29, vid);
		rtl8305s_smiWrite(4, 28, memberPortMask);
		break;
		case 15:/* VLAN[P] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiWrite(0, 31, vid);
		rtl8305s_smiWrite(0, 30, memberPortMask);
		break;
		default:
			return FAILED;
	}
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicVlan | Configure switch VLAN ID and corresponding member ports.
@parm uint16 | vlanIndex | Specify the position to configure VLAN
@parm uint16 * | vid | Specify the VLAN ID
@parm uint16 * | memberPortMask | Specify the member port list of corresponding VLAN ID
@comm
	Total have 16 entries to store the VID and corresponding member port. This API will not
check whether entry already configured or not.
*/
int32 rtl8305sc_getAsicVlan(uint16 vlanIndex, uint16 *vid, uint16 *memberPortMask) {

	if(vid == NULL || memberPortMask == NULL)
		return FAILED;
	switch(vlanIndex) {
		case 0:/* VLAN[A] */
		rtl8305s_smiRead(0, 25, vid);
		rtl8305s_smiRead(0, 24, memberPortMask);
		break;
		case 1:/* VLAN[B] */
		rtl8305s_smiRead(1, 25, vid);
		rtl8305s_smiRead(1, 24, memberPortMask);
		break;
		case 2:/* VLAN[C] */
		rtl8305s_smiRead(2, 25, vid);
		rtl8305s_smiRead(2, 24, memberPortMask);
		break;
		case 3:/* VLAN[D] */
		rtl8305s_smiRead(3, 25, vid);
		rtl8305s_smiRead(3, 24, memberPortMask);
		break;
		case 4:/* VLAN[E] */
		rtl8305s_smiRead(4, 25, vid);
		rtl8305s_smiRead(4, 24, memberPortMask);
		break;
		case 5:/* VLAN[F] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(0, 27, vid);
		rtl8305s_smiRead(0, 26, memberPortMask);
		break;
		case 6:/* VLAN[G] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(1, 27, vid);
		rtl8305s_smiRead(1, 26, memberPortMask);
		break;
		case 7:/* VLAN[H] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(2, 27, vid);
		rtl8305s_smiRead(2, 26, memberPortMask);
		break;
		case 8:/* VLAN[I] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(3, 27, vid);
		rtl8305s_smiRead(3, 26, memberPortMask);
		break;
		case 9:/* VLAN[J] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(4, 27, vid);
		rtl8305s_smiRead(4, 26, memberPortMask);
		break;
		case 10:/* VLAN[K] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(0, 29, vid);
		rtl8305s_smiRead(0, 28, memberPortMask);
		break;
		case 11:/* VLAN[L] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(1, 29, vid);
		rtl8305s_smiRead(1, 28, memberPortMask);
		break;
		case 12:/* VLAN[M] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(2, 29, vid);
		rtl8305s_smiRead(2, 28, memberPortMask);
		break;
		case 13:/* VLAN[N] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(3, 29, vid);
		rtl8305s_smiRead(3, 28, memberPortMask);
		break;
		case 14:/* VLAN[O] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(4, 29, vid);
		rtl8305s_smiRead(4, 28, memberPortMask);
		break;
		case 15:/* VLAN[P] */
		rtl8305s_smiWriteBit(0, 16, 15, 1); /* Change to page 1 */
		rtl8305s_smiRead(0, 31, vid);
		rtl8305s_smiRead(0, 30, memberPortMask);
		break;
		default:
			return FAILED;
	}
	*vid = *vid & 0xFFF;
	*memberPortMask = *memberPortMask & 0x1F;
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicPortVlanIndex | Configure switch port un-tagged packet belonged VLAN index
@parm uint16 | port | Specify the port to configure
@parm uint16 | vlanIndex | Specify the VLAN index
@comm
There are 16 valid index. This function does not check whether VLAN index do configured before port belong to it.
*/
int32 rtl8305sc_setAsicPortVlanIndex(uint16 port, uint16 vlanIndex) {
	uint16 regValue;

	if(port>=RTL8305S_PORT_NUMBER)
		return FAILED;
	rtl8305s_smiRead(port, 24, &regValue);
	regValue = (regValue & 0xFFF) | (vlanIndex<<12);
	rtl8305s_smiWrite(port, 24, regValue);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicPortVlanIndex | Get configured switch port un-tagged packet belonged VLAN index
@parm uint16 | port | Specify the portto configure VLAN index
@parm uint16 * | vlanIndex | Port belonged VLAN index
@comm
This function only get the configured register value and do not check whether VLAN do exist
*/
int32 rtl8305sc_getAsicPortVlanIndex(uint16 port, uint16 *vlanIndex) {
	if(port>=RTL8305S_PORT_NUMBER || vlanIndex == NULL)
		return FAILED;
	rtl8305s_smiRead(port, 24, vlanIndex);
	*vlanIndex = (*vlanIndex>>12)&0xF;
	return SUCCESS;
}


/*
@func int32 | rtl8305sc_getAsicMacAddressTableEntry | Get switch MAC address table entry.
@parm uint16 | index | Position of address table to get MAC address information
@parm int8 * | macAddress | Point to the 6-byte space to store MAC address.
@parm uint16 * | age | Space to store aging time, translated to seconds
@parm int8 * | isStatic | Specify whether corresponding address is static configured or not.
@parm uint16 * | port | number of corresponding address.
@parm int8 * | isValid | Specify whether retrieved entry information is valid or not. In 8305SC, the entry valid is decided by aging non-zero or static
@comm
	Using position to get the MAC address table entry.
*/
int32 rtl8305sc_getAsicMacAddressTableEntry(uint16 index, int8 * macAddress, uint16 *age, int8 * isStatic, uint16 * port, int8 * isValid) {
	uint16 value;
	uint32 idleCounter;

	if(macAddress == NULL || age == NULL || isStatic == NULL || port == NULL || isValid == NULL)
		return FAILED;
	rtl8305s_smiWriteBit(0, 16, 13, 1);/* Enable lookup table accessable */
	rtl8305s_smiWrite(4, 18, (index&0xFF)<<8 | ((index>>8)&0x3)); /* Set the index in [0:9]*/
	rtl8305s_smiWrite(4, 16, 0x3); /* Read operation and start to operate */
	idleCounter = 0;
	do {
		rtl8305s_smiRead(4, 16, &value);
		idleCounter++;
	}while((value &0x2) && idleCounter < RTL8305S_IDLE_TIMEOUT);
	if(idleCounter == RTL8305S_IDLE_TIMEOUT)
		return FAILED;
	rtl8305s_smiRead(4, 17, &value);
	*port = value & 0xF;
	switch((value>>4)&0x3) {
		case 0: *age = 0; break;
		case 1: *age = 100; break;
		case 2: *age = 300; break;
		case 3: *age = 200; break;
	}
	*isStatic = value&0x40?TRUE:FALSE;
	*isValid =value&0x70?TRUE:FALSE;
	rtl8305s_smiRead(4, 18, &value);
	macAddress[4] = (int8)value&0xFF;
	macAddress[5] = (value>>8)&0xFF;
	rtl8305s_smiRead(4, 19, &value);
	macAddress[2] = (int8)value&0xFF;
	macAddress[3] = (value>>8)&0xFF;
	rtl8305s_smiRead(4, 20, &value);
	macAddress[0] = (int8)value&0xFF;
	macAddress[1] = (value>>8)&0xFF;
	
	rtl8305s_smiWriteBit(0, 16, 13, 0);/* Disable lookup table accessable */
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicMacAddressTableEntry | Set switch MAC address table entry.
@parm int8 * | macAddress | Point to the 6-byte space to store MAC address.
@parm uint16 | age | Space to store aging time, translated to seconds
@parm int8 | isStatic | Specify whether corresponding address is static configured or not.
@parm uint16 * | port | number of corresponding address.
@comm
	Use this API to configure the MAC address table.
*/
int32 rtl8305sc_setAsicMacAddressTableEntry(int8 * macAddress, uint16 age, int8 isStatic, uint16 port) {
	uint16 value;
	uint32 idleCounter;

	if(macAddress == NULL)
		return FAILED;
	rtl8305s_smiWriteBit(0, 16, 13, 1);/* Enable lookup table accessable */
	value = port & 0xF;
	if(age>200)
		value |= 0x20;
	else if(age>100)
		value |= 0x30;
	else if(age>0)
		value |= 0x10;
	if(isStatic == TRUE) 
		value |= 0x40;
	rtl8305s_smiWrite(4, 17, value);
	rtl8305s_smiWrite(4, 18, macAddress[5]<<8 | macAddress[4]);
	rtl8305s_smiWrite(4, 19, macAddress[3]<<8 | macAddress[2]);
	rtl8305s_smiWrite(4, 20, macAddress[1]<<8 | macAddress[0]);
	rtl8305s_smiWrite(4, 16, 0x2); /* Write operation and start to operate */
	idleCounter=0;
	do {
		rtl8305s_smiRead(4, 16, &value);
		idleCounter++;
	}while((value &0x2)&& idleCounter<RTL8305S_IDLE_TIMEOUT);
	if(idleCounter == RTL8305S_IDLE_TIMEOUT)
		return FAILED;
	
	rtl8305s_smiWriteBit(0, 16, 13, 0);/* Disable lookup table accessable */
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicLoopDetection | Configure switch loop detection mechanism.
@parm int8 | enabled | Configure 8305SC loop detection mechanism
@comm
*/
int32 rtl8305sc_setAsicLoopDetection(int8 enabled) {
	rtl8305s_smiWriteBit(0, 16, 2, enabled==TRUE?1:0);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicLoopDetection | Get switch loop detection mechanism setting.
@parm int8 * | enabled | 8305SC loop detection mechanism is enabled only or not.
@comm
*/
int32 rtl8305sc_getAsicLoopDetection(int8 *enabled) {
	uint8 bitValue;

	if(enabled == NULL)
		return FAILED;
	rtl8305s_smiReadBit(0, 16, 2, &bitValue);
	*enabled = bitValue == 1? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_asicSoftReset | Trigger 8305SC soft reset.
@comm
	When call this API, RTL8305SC will reset all internal registers except PHY registers, and will
not configurations from EEPROM or strapping pings. Software reset is designed to provide a 
convenient way for users to change the configuration via SMI.
*/
int32 rtl8305sc_asicSoftReset(void) {
	rtl8305s_smiWriteBit(0, 16, 12, 1);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicAgressiveBackoff | Configure switch aggressive back-off mechanism.
@parm int8 | enabled | Configure 8305SC aggressive back-off mechanism
@comm
*/
int32 rtl8305sc_setAsicAgressiveBackoff(int8 enabled) {
	rtl8305s_smiWriteBit(0, 18, 15, enabled==TRUE?1:0);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicAgressiveBackoff | Get switch aggressive back-off mechanism setting.
@parm int8 * | enabled | 8305SC aggressive back-off mechanism is enabled only or not.
@comm
*/
int32 rtl8305sc_getAsicAgressiveBackoff(int8 *enabled) {
	uint8 bitValue;

	if(enabled==NULL)
		return FAILED;
	rtl8305s_smiReadBit(0, 18, 15, &bitValue);
	*enabled = bitValue == 1? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicLongReceivedPacketLength | Configure switch maximum received packet length.
@parm int8 | enabled1552 | Configure 8305SC to extend the received packet length from 1536 to 1552.
@comm
	If input is TRUE, receiving packet length is 1552 bytes. If input is FALSE, receiving
packet length is 1536.
*/
int32 rtl8305sc_setAsicLongReceivedPacketLength(int8 enabled1552) {

	rtl8305s_smiWriteBit(0, 18, 14, enabled1552==TRUE?0:1);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicLongReceivedPacketLength | Get switch maximum received packet length.
@parm int8 * | enabled1552 | 8305SC can received packet length up to 1552 or only 1536.
@comm
	If output is TRUE, receiving packet length is 1552 bytes. If output is FALSE, receiving
packet length is 1536.
*/
int32 rtl8305sc_getAsicLongReceivedPacketLength(int8 *enabled1552) {
	uint8 bitValue;

	if(enabled1552==NULL)
		return FAILED;
	rtl8305s_smiReadBit(0, 18, 14, &bitValue);
	*enabled1552 = bitValue == 0? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicBroadcastInputDrop | Configure switch maximum received packet length.
@parm int8 | enabled | Configure 8305SC broadcast input drop..
@comm
*/
int32 rtl8305sc_setAsicBroadcastInputDrop(int8 enabled) {

	rtl8305s_smiWriteBit(0, 18, 13, enabled==TRUE?1:0);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicBroadcastInputDrop | Get switch broadcast input drop.
@parm int8 * | enabled | 8305SC broacast input drop.
@comm
*/
int32 rtl8305sc_getAsicBroadcastInputDrop(int8 *enabled) {
	uint8 bitValue;

	if(enabled==NULL)
		return FAILED;
	rtl8305s_smiReadBit(0, 18, 13, &bitValue);
	*enabled = bitValue == 1? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicForward1dReservedAddress | Configure switch to forward 802.1d reserved MAC addresses.
@parm int8 | enabled | Configure 8305SC to forward 802.1d reserved MAC addresses
@comm
The reserved address is ranged from 01-80-c2-00-00-04 to 01-80-c2-00-00-0f.
*/
int32 rtl8305sc_setAsicForward1dReservedAddress(int8 enabled) {

	rtl8305s_smiWriteBit(0, 18, 12, enabled==TRUE?1:0);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicForward1dReservedAddress | Get switch whether forward 802.1d reserved addresses or not..
@parm int8 * | enabled | Whether RTL8305SC forwards 802.1d reserved addresses or not..
@comm
The reserved address is ranged from 01-80-c2-00-00-04 to 01-80-c2-00-00-0f.
*/
int32 rtl8305sc_getAsicForward1dReservedAddress(int8 *enabled) {
	uint8 bitValue;

	if(enabled==NULL)
		return FAILED;
	rtl8305s_smiReadBit(0, 18, 12, &bitValue);
	*enabled = bitValue == 1? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicLeakyVlan | Configure switch to forward unicast frames to other VLANs.
@parm int8 | enabled | Configure 8305SC to forward unicast frames to other VLANs.
@comm
*/
int32 rtl8305sc_setAsicLeakyVlan(int8 enabled) {

	rtl8305s_smiWriteBit(0, 18, 11, enabled==FALSE?1:0);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicLeakyVlan | Get switch whether forward unicast frames to other VLANs.
@parm int8 * | enabled | Whether RTL8305SC forwards unicast frames to other VLANs.
@comm
*/
int32 rtl8305sc_getAsicLeakyVlan(int8 *enabled) {
	uint8 bitValue;

	if(enabled==NULL)
		return FAILED;
	rtl8305s_smiReadBit(0, 18, 11, &bitValue);
	*enabled = bitValue == 0? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicLeakyVlan | Configure switch to enable broadcast storm control.
@parm int8 | enabled | Configure 8305SC to enable broadcast storm control.
@comm
*/
int32 rtl8305sc_setAsicBroadcastStormControl(int8 enabled) {

	rtl8305s_smiWriteBit(0, 18, 2, enabled==FALSE?1:0);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicLeakyVlan | Get switch whether enable broadcast storm control.
@parm int8 * | enabled | Whether RTL8305SC enable broadcast storm control.
@comm
*/
int32 rtl8305sc_getAsicBroadcastStormControl(int8 *enabled) {
	uint8 bitValue;

	if(enabled==NULL)
		return FAILED;
	rtl8305s_smiReadBit(0, 18, 2, &bitValue);
	*enabled = bitValue == 0? TRUE: FALSE;

	return SUCCESS;
}

/*
@func int32 | rtl8305sc_setAsicLeakyVlan | Configure switch to forward arp frames to other VLANs.
@parm int8 | enabled | Configure 8305SC to forward arp frames to other VLANs.
@comm
*/
int32 rtl8305sc_setAsicArpVlan(int8 enabled) {

	rtl8305s_smiWriteBit(0, 18, 10, enabled==FALSE?1:0);
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_getAsicLeakyVlan | Get switch whether forward arp frames to other VLANs.
@parm int8 * | enabled | Whether RTL8305SC forwards arp frames to other VLANs.
@comm
*/
int32 rtl8305sc_getAsicArpVlan(int8 *enabled) {
	uint8 bitValue;

	if(enabled==NULL)
		return FAILED;
	rtl8305s_smiReadBit(0, 18, 10, &bitValue);
	*enabled = bitValue == 0? TRUE: FALSE;
		
	return SUCCESS;
}

/*
@func int32 | rtl8305sc_asicRealtekCableTester | Testing port by Realtek Cable Tester
@parm uint16 | port | Specify the port to test
@parm int8 | isTxDir | Specify the port testing direction either Tx or Rx
@parm uint32 * | cableStatus | Store the cable status of testing result
@parm uint32 * | length | Store the detected length of testing result, in unit of 0.5m
@comm
This function will enable the Realtek Cable Tester feature and test specified port.
The port status and measured length information will be returned, and the length unit is 0.5 m.
*/
int32 rtl8305sc_asicRealtekCableTester(uint16 port, int8 isTxDir, uint32 * cableStatus, uint32 * length) {
	uint32 idleCounter;
	uint16 value;
	
	if(port >= RTL8305S_PORT_NUMBER || cableStatus == NULL || length == NULL)
		return FAILED;

	rtl8305s_smiWrite(3, 31, 0x0000);
	rtl8305s_smiWrite(3, 30, 0x07FA);
	rtl8305s_smiWrite(3, 29, 0xFC29);
	rtl8305s_smiWrite(3, 28, 0x7AB5);
	rtl8305s_smiWrite(3, 27, 0xA458);
	
	rtl8305s_smiWriteBit(0, 16, 11, 1);
	rtl8305s_smiWrite(0, 28, 0x300 | (port & 0x7));
	rtl8305s_smiWrite(4, 27, 0);
	rtl8305s_smiWrite(4, 27, isTxDir==TRUE?0x1000: 0x9000);
	idleCounter = 0;
	do {
		rtl8305s_smiRead(4, 27, &value);
		idleCounter++;
	}while(!(value&0x800) && idleCounter<RTL8305S_IDLE_TIMEOUT);
	if(idleCounter == RTL8305S_IDLE_TIMEOUT)
		return FAILED;
	if(value & 0x400)
		*cableStatus = RTL8305SC_RTCT_SHORT;
	else if (value & 0x200)
		*cableStatus = RTL8305SC_RTCT_OPEN;
	else
		*cableStatus = RTL8305SC_RTCT_NORMAL;
	*length = value & 0x1FF;
	
	rtl8305s_smiWriteBit(0, 16, 11, 1);
	return SUCCESS;
}

#define RTL8305SC_TABLE_DRIVER

