/*
* Copyright c                  Realtek Semiconductor Corporation, 2003  
* All rights reserved.
* 
* Program :  Multiple Session L2TP ALG (for L2TP Protocol ver 2)
* Abstract : 
* Creator : 
* Author :
* $Id: rtl8651_alg_l2tp.c,v 1.1.1.1 2006/07/26 05:38:23 blue Exp $
* $Log: rtl8651_alg_l2tp.c,v $
* Revision 1.1.1.1  2006/07/26 05:38:23  blue
*
*
* Revision 1.1.1.1  2006/02/14 03:09:05  jason
* backup the source code from v1036
*
* Revision 1.1.1.1  2006/01/16 02:39:43  jason
* backup source,because now we directly use romedriver source code, i open a new directory;
*
* Revision 1.1.1.1  2005/11/21 05:20:03  chenyl
* +: WPNT834
* 	- System provider	: Sercomm
* 	- Brand Name		: Netgear
* 	- 3rt Party		: Airgo Gen 3
* 	- SDK			: 066-3
* 	- ROMEDRV		: 3.6-3
*
*
* Revision 1.8  2004/09/07 14:52:13  chenyl
* *: bug fix: napt-redirect fragment packet checksum-recalculate
* *: bug fix: conflict between protocol stack flow cache and protocol stack action
* *: bug fix: protocol stack action mis-process to UDP-zero-checksum packet
* *: separate the header file:
*         - internal : rtl8651_tblDrvFwdLocal.h
*         - external : rtl8651_tblDrvFwd.h
*
* Revision 1.7  2004/07/23 13:19:57  tony
* *: remove all warning messages.
*
* Revision 1.6  2004/04/20 03:44:02  tony
* if disable define "RTL865X_OVER_KERNEL" and "RTL865X_OVER_LINUX", __KERNEL__ and __linux__ will be undefined.
*
* Revision 1.5  2004/03/19 12:17:21  cfliu
* Rename all osk_malloc, printk, printf, printfByPolling as rtlglue_printf
*
* Revision 1.4  2004/03/19 08:12:53  tony
* make IPSec, PPTP, L2TP are able to set session numbers dynamically by user.
*
* Revision 1.3  2004/03/16 07:53:06  tony
* fix l2tp alg to support module test(endian free).
*
* Revision 1.2  2004/03/15 11:33:19  tony
* fix l2tp alg bug
*
* Revision 1.1  2004/02/25 14:26:33  chhuang
* *** empty log message ***
*
* Revision 1.2  2004/02/25 14:24:52  chhuang
* *** empty log message ***
*
* Revision 1.7  2004/01/29 01:39:47  orlando
* checkin acl's dsid support, not the final version
*
* Revision 1.6  2004/01/13 06:27:49  tony
* make l2tp is able to support multiple dial session.
*
* Revision 1.5  2004/01/12 11:59:34  tony
* make ALG is able to enable/disable for multiple dial session.
*
* Revision 1.4  2004/01/09 08:03:21  tony
* make the code architecture of ALG is able to support multiple dial session.
*
* Revision 1.3  2003/12/31 04:59:26  tony
* fix bug: turn L2TP ALG on/off dynamically.
*
* Revision 1.2  2003/12/22 11:59:14  tony
* fix bug: unalignment issue.
*
* Revision 1.1  2003/12/08 03:37:26  tony
* add new ALG: L2TP v2 multiple-session
*
*
*
*/

#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_alg_init.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_l2tp.h"

static uint32 last_ip=0;
static uint32 session_cnt=0;
static struct l2tp_session *session=NULL;


int32 rtl8651_l4AliasL2tpInit(void *arg)
{
	if(session) rtlglue_free(session);
	session = (struct l2tp_session *)rtlglue_malloc(sizeof(struct l2tp_session)*algMaxSession.l2tpMaxSession);
	memset((void*)session,0,sizeof(struct l2tp_session)*algMaxSession.l2tpMaxSession);	
	last_ip=0;
	session_cnt=0;
	return SUCCESS;
}


int32 rtl8651_l4AliasHandleL2tpClientOutbound(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tb)
{
	struct udphdr *udp;
	int32 tlen;
	struct l2tp_header *l2tp;
	
	
	//printk("alg_support & RTL_L2TP= %x\n",alg_supports&RTL8651_ALG_L2TP);

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_L2TP)==0) return FWDENG_USER_DROP;

	//printk("L2TP_OUT\n");	

	udp = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));

	if(udp->uh_sport!=udp->uh_dport) return SUCCESS;	

	tlen = ntohs(pip->ip_len); //total_length

	if(tlen-36<0) return SUCCESS; //20(ip header)+8(udp header)+8(l2tp header)

	//control message
	l2tp = (struct l2tp_header *)(udp+1);
	
	if(last_ip==0) last_ip=PKTGET_UINT32_UNALIGNED(&pip->ip_src.s_addr);

	#ifdef _RTL_UPNP
	/* add PNP port map */
	if(tb->insideLocalPort==1701)
	{
		rtl8651_addUpnpMap(UPNP_ONESHOT, tb->dstIpAddr, 1701, 
			tb->insideGlobalIpAddr , 1701,
			tb->insideLocalIpAddr , 1701);
			tb->insideGlobalPort=1701;
	}
	#endif /* _RTL_UPNP */	

	if((ntohs(l2tp->flag)&0xf4ff)==0xc002) 
	{
		if((l2tp->session_id==0)&&(l2tp->tunnel_id==0))
		{
			last_ip=PKTGET_UINT32_UNALIGNED(&pip->ip_src.s_addr);
			//printk("save last ip=%x\n",last_ip);
		}
	}
	
	return SUCCESS;
}


int32 rtl8651_l4AliasHandleL2tpClientInbound(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tb)
{

	struct udphdr *udp;
	int32 i;
	int32 tlen;
	struct l2tp_header *l2tp;

	//printk("alg_support & RTL_L2TP= %x\n",alg_supports&RTL8651_ALG_L2TP);	
	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_L2TP)==0) return FWDENG_USER_DROP;

	//printk("L2TP_IN\n");
	if(last_ip==0)  return SUCCESS;

	udp = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));

	if(udp->uh_sport!=udp->uh_dport) return SUCCESS;

	tlen = ntohs(pip->ip_len); //total_length
	if(tlen-36<0)  return SUCCESS;  //20(ip header)+8(udp header)+8(l2tp header)
	l2tp = (struct l2tp_header *)(udp+1);
	
	//already have session Id
	for(i=0;i<algMaxSession.l2tpMaxSession;i++)
	{
		if((session[i].tunnel_id==l2tp->tunnel_id)&&(l2tp->tunnel_id!=0))
		{
			tb->insideLocalIpAddr=session[i].redirect_ip; //change local ip address by session ID
			//printk("session=%d, redirect to session ip=%x\n",i,session[i].redirect_ip);
			return SUCCESS;
		}
	}

	//do not have session id, save it.
	if((ntohs(l2tp->flag)&0xf4ff)==0xc002) //control message 
	{
		for(i=0;i<algMaxSession.l2tpMaxSession;i++) // old session re-login
		{	
			if(session[i].redirect_ip==last_ip)
			{
				session[i].tunnel_id=l2tp->tunnel_id;
				//printk("save tunnel id=%d fwd to last_ip=%x ,have old record\n",l2tp->tunnel_id,last_ip);
				break;
			}
		}

		if(i==algMaxSession.l2tpMaxSession) //new session login
		{
			session[session_cnt].tunnel_id=l2tp->tunnel_id;
			session[session_cnt].redirect_ip=last_ip;								
			//printk("save tunnel id=%d fwd to last_ip=%x\n",l2tp->tunnel_id,last_ip);
			++session_cnt;
			session_cnt%=algMaxSession.l2tpMaxSession;
		}
	}

	tb->insideLocalIpAddr=last_ip;
	return SUCCESS;
}



