/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : 
* Abstract : 
* Creator : 
* Author :  
* $Id: rtl8651_alg_pptp.c,v 1.1.1.1 2006/07/26 05:38:23 blue Exp $
* $Log: rtl8651_alg_pptp.c,v $
* Revision 1.1.1.1  2006/07/26 05:38:23  blue
*
*
* Revision 1.1.1.1  2006/02/14 03:09:05  jason
* backup the source code from v1036
*
* Revision 1.1.1.1  2006/01/16 02:39:43  jason
* backup source,because now we directly use romedriver source code, i open a new directory;
*
* Revision 1.2  2005/11/22 15:50:54  chenyl
* *: Add mutex lock/unlock in log r/w
* *: new idle-timeout mechanism for VLAN / PPPOE / PPTP / L2TP wan types
* *: bug fix: PPTP passthrough
* *: bug fix: sip / ymsg typo error
* *: for ARP-reply module in RomeDrv, use newly allocated mbuf/pkt instead of original one ( To prevent from treated as packet forwarding instead of reply).
* *: minor coding style, sync with main trunk
*
* +: ioctl case for Mutex Lock/Unlock called by user level
* +: add 2 APIs to enable / disable
* 	1. Error messages from protocol stack to WAN : TCP RST, ICMP ICMP_UNREACH/ICMP_SOURCEQUENCH/ICMP_TIMXCEED/ICMP_PARAMPROB
* 	2. All packet from protocol stack to WAN
*    filtering to disable / enable idle-timer update triggered by these kinds of packets.
* +: TFTP ALG
* +: rtl8651_debug.h for debugging usage
* +: merge code about rtl_glue.c / re_core.c with Sercomm's version ( Airgo modification ?)
* +:
*
* Revision 1.32  2005/10/20 14:10:15  jzchen
* +: Debug message can be selective enable/disable through API.
* +: Session termination through timeout
*
* Revision 1.31  2005/10/18 15:55:09  jzchen
* *: Server outbound only have software instead of expected hardware forwarding
* +: Entry does not deleted when disconnect information appear. Configurable timeout after these information appear is applied.
*
* Revision 1.30  2005/10/17 16:32:02  chenyl
* *: bug fix: fix typo error: It would cause system crash when PPTP-passthrough via HW.
*
* Revision 1.29  2005/10/17 14:51:58  jzchen
* *: Change initial search ID to be more random
* *: Free link list chaining bug is fixed
* *: Consider about Stop control connection message
*
* Revision 1.28  2005/10/14 14:28:27  jzchen
* *: Bug fix: Return to 1.25
*
* Revision 1.27  2005/10/14 06:41:15  jzchen
* *: Bug fix: tracking inbound list but typeo to track outbound list
*
* Revision 1.26  2005/10/13 14:55:49  jzchen
* *: Server outbound traffic when Call ID mismatch, clean previous configured one
*
* Revision 1.25  2005/10/03 15:44:30  jzchen
* *: Bug fix, when duplicated creation appear under WAN side PPTP client connect to LAN side PPTP server, link list hooking does not consider the situation that duplicated message packets may enter and cause link list looping.
*
* Revision 1.24  2005/09/29 06:05:49  tony
* *: bug fixed: when PPTP do not receive a stop-connection-request, the entry is keep, so if next
* connection dial up, if the call id is the same previous call id. then the GRE can't send correctly.
*
* Revision 1.23  2005/09/02 02:56:17  tony
* *: bug fixed: pptp passthrough in multiple pppoe.
*
* Revision 1.22  2005/08/12 13:30:07  tony
* *: bug fixed: for pptp housekeeping when reinit.
*
* Revision 1.21  2005/06/01 13:56:18  tony
* *: support alg reinit.
*
* Revision 1.20  2005/03/01 08:20:34  tony
* *: support PPTP/IPSec ALG config dividedly.
*
* Revision 1.19  2004/10/14 10:43:18  tony
* *: fix PPTP can't modify Call ID and Peer's ID in PPTP_WanErrorNotifya, PPTP_SetLinkInfo, PPTP_CallClearRequest, PPTP_CallDiscNotify.
*
* Revision 1.18  2004/09/22 13:50:31  tony
* +: close then data channel(GRE) when disable PPTP pass-throught ALG,
*    this function is supported for HW/SW and server in and client out.
*
* Revision 1.17  2004/09/16 11:50:20  tony
* +: add PPTP pass-throught software (not change call ID) case.
*
* Revision 1.16  2004/09/07 14:52:13  chenyl
* *: bug fix: napt-redirect fragment packet checksum-recalculate
* *: bug fix: conflict between protocol stack flow cache and protocol stack action
* *: bug fix: protocol stack action mis-process to UDP-zero-checksum packet
* *: separate the header file:
*         - internal : rtl8651_tblDrvFwdLocal.h
*         - external : rtl8651_tblDrvFwd.h
*
* Revision 1.15  2004/07/27 07:31:42  chenyl
* *: pptp/ipSec pass-through, GRE/ESP fragment packets process.
* *: if DoS Ignore Type is matched, we check whether system can sent fragment packet
*    regardless of the TearDrop Setting.
*
* Revision 1.14  2004/07/26 08:49:49  tony
* *: remove debug message
*
* Revision 1.13  2004/07/23 13:19:57  tony
* *: remove all warning messages.
*
* Revision 1.12  2004/07/21 12:45:51  tony
* +: add new features: PPTP Pass-through support server in.
*
* Revision 1.11  2004/07/08 05:48:57  tony
* *: if disable pptp ALG , drop all the port 1723 packets.
*
* Revision 1.10  2004/06/29 14:47:55  jzchen
* +: Add PPTP passthrough modify ID display
*
* Revision 1.9  2004/06/24 03:23:42  jzchen
* +: Change hash algorithm to meet 865xB PPTP pass-through algorithm
* +: Configure software entry to ASIC when ASIC entry is empty
* +: Remove ASIC entry and replaced with candidate entry when software entry is removed
* +: Aging software entry, which does not configured to ASIC
* +: Heavy load swapping. When software entry load is too heavy, swap the first one meet the condition to ASIC
*
* Revision 1.8  2004/05/14 02:29:26  tony
* let HW calculate checksum and set ip_sum to zero.
*
* Revision 1.7  2004/05/13 14:35:00  tony
* make pptp passthru is able to run under rtl865XB. (fix checksum error, when ttl-
* 1)
*
* Revision 1.6  2004/05/12 09:56:26  tony
* fix bug: pptp will crash when sessions are overflow.
*
* Revision 1.5  2004/04/20 03:44:03  tony
* if disable define "RTL865X_OVER_KERNEL" and "RTL865X_OVER_LINUX", __KERNEL__ and __linux__ will be undefined.
*
* Revision 1.4  2004/03/19 12:19:26  cfliu
* Rename all osk_malloc, printk, printf, printfByPolling as rtlglue_printf
*
* Revision 1.3  2004/03/19 08:12:53  tony
* make IPSec, PPTP, L2TP are able to set session numbers dynamically by user.
*
* Revision 1.2  2004/03/18 02:00:41  tony
* make pptp alg support endian free.
*
* Revision 1.1  2004/02/25 14:26:33  chhuang
*
*/

/*	@doc RTL8651_PPTP_PASS_THROUGH_API

	@module rtl8651_alg_pptp.c - RTL8651 Home gateway controller PPTP pass-through documentation	|
	This document explains the internal and external API interface of the PPTP pass-through module. Functions with _rtl8651 prefix
	are internal functions and functions with rtl8651_ prefix are external functions.
	@normal Jin-Ru Chen (jzchen@realtek.com.tw) <date>

	Copyright <cp>2004 Realtek<tm> Semiconductor Cooperation, All Rights Reserved.

 	@head3 List of Symbols |
 	Here is a list of all functions and variables in this module.

 	@index | RTL8651_TABLEDRIVER_API
*/


#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_pptp.h"
#include "rtl8651_tblAsicDrv.h"

//#undef RTL8650B_PPTP_HASH
#define RTL8650B_PPTP_HASH

/* PPTP definitions */

struct grehdr						   /* Enhanced GRE header. */
{
	uint16    gh_flags;				   /* Flags. */
	uint16    gh_protocol;			   /* Protocol type. */
	uint16    gh_length;			   /* Payload length. */
	uint16    gh_call_id;			   /* Call ID. */
	uint32    gh_seq_no;			   /* Sequence number (optional). */
	uint32    gh_ack_no;			   /* Acknowledgment number (optional). */
};
typedef struct grehdr GreHdr;

/* The PPTP protocol ID used in the GRE 'proto' field. */
#define PPTP_GRE_PROTO          0x880b

/* Bits that must be set a certain way in all PPTP/GRE packets. */
#define PPTP_INIT_VALUE		((0x2001 << 16) | PPTP_GRE_PROTO)
#define PPTP_INIT_MASK		0xef7fffff

#define PPTP_MAGIC		0x1a2b3c4d
#define PPTP_CTRL_MSG_TYPE	1

enum
{
	PPTP_StartCtrlConnRequest = 1,
	PPTP_StartCtrlConnReply = 2,
	PPTP_StopCtrlConnRequest = 3,
	PPTP_StopCtrlConnReply = 4,
	PPTP_EchoRequest = 5,
	PPTP_EchoReply = 6,
	PPTP_OutCallRequest = 7,
	PPTP_OutCallReply = 8,
	PPTP_InCallRequest = 9,
	PPTP_InCallReply = 10,
	PPTP_InCallConn = 11,
	PPTP_CallClearRequest = 12,
	PPTP_CallDiscNotify = 13,
	PPTP_WanErrorNotify = 14,
	PPTP_SetLinkInfo = 15
};

int reInit=0;

  /* Message structures  */
struct pptpMsgHead
{
	uint16    length;				   /* total length */
	uint16    msgType;				   /* PPTP message type */
	uint32    magic;				   /* magic cookie */
	uint16    type;					   /* control message type */
	uint16    resv0;				   /* reserved */
};
typedef struct pptpMsgHead *PptpMsgHead;

struct pptpCodes
{
	uint8     resCode;				   /* Result Code */
	uint8     errCode;				   /* Error Code */
};
typedef struct pptpCodes *PptpCode;

struct pptpCallIds
{
	uint16    cid1;					   /* Call ID field #1 */
	uint16    cid2;					   /* Call ID field #2 */
};
typedef struct pptpCallIds *PptpCallId;

typedef struct _rtl8651_pptp_passthrough_para_s {
	uint16 idelTimeout;//Compared with per-entry idleTimer
	uint16 heavyLoadSwappingThreshold;//Compared with per-entry pktCnt;
} _rtl8651_pptp_passthrough_para_t;

_rtl8651_pptp_passthrough_para_t pptpPassthroughSysPara;

typedef struct _rtl8651_pptpGreEntry_s {
	uint32 outsideGlobalIpAddr;
	uint32 insideGlobalIpAddr;
	uint32 insideLocalIpAddr;

	uint16 outboundGlobalId;
	uint16 outboundLocalId;
	uint16 inboundGlobalId;
	uint16 inboundLocalId;

	uint16 inboundIdleTimer;	//Used to evaluate how long packet does not use this inbound flow
	uint16 inboundPktCnt;	//Used to evaluate the number of software forwaded packet number
	uint16 outboundIdleTimer;	//Used to evaluate how long packet does not use this outbound flow
	uint16 outboundPktCnt;	//Used to evaluate the number of software forwaded packet number

	uint32 dsid;
	
	uint16 inboundIdInitiated:1;//When outgoing-call-request passed, the inboundId information is created
	uint16 inboundIdClosed:1;//When call-clear-request passed, the inboundId closed state is set
	uint16 inboundToAsic:1;//Whether inbound PPTP translation configured to ASIC
	uint16 outboundIdInitiated:1;//When outgoing-call-reply passed, based on IP address and Peer's ID, the created session can be found and can calculate the used outbound ID
	uint16 outboundIdClosed:1;//When call-disconnect-notify passed, the outboundId cloased state is set
	uint16 outboundToAsic:1;//Whether outbound PPTP translation configured to ASIC
	uint16 controlIdleTimer;
	struct _rtl8651_pptpGreEntry_s * inboundNext, * outboundNext;
}_rtl8651_pptpGreEntry_t;

_rtl8651_pptpGreEntry_t * pptpGreInboundHeadPtr[RTL8651_ICMPTBL_SIZE], * pptpGreOutboundHeadPtr[RTL8651_ICMPTBL_SIZE], 
	* freePptpGreEntryHeadPtr=NULL ,*softwarePptpGreEntryHeadPtr=NULL;
uint16 pptpPassthroughControlShortTimeout = 10, pptpPassthroughControlMiddleTimeout = 60;
uint8 pptpPassthroughClientInboundMsg = 0, pptpPassthroughClientOutboundMsg = 0,
	 pptpPassthroughServerInboundMsg = 0, pptpPassthroughServerOutboundMsg = 0,
	 pptpPassthroughGreInboundMsg = 0, pptpPassthroughGreOutboundMsg = 0;

/*
@func int32	| _rtl8651_addAsicPptpPassthroughInboundEntry	|  Add PPTP passthrough inbound entry to switch core
@parm _rtl8651_pptpGreEntry_t *	| entryPtr		| PPTP passthrough entry
@parm struct ip *			| pip		| IP header of the arrived packet
@comm 
1. If entry already in ICMP table, definite success
2. Fond both inbound and outbound list, if there do exist some entry in ASIC, stop writing to ASIC
3. Configure to ASIC

*/
int32 _rtl8651_addAsicPptpPassthroughInboundEntry(_rtl8651_pptpGreEntry_t * entryPtr) {
	_rtl8651_pptpGreEntry_t *trackPtr;
	rtl865x_tblAsicDrv_naptIcmpParam_t pptpAsicEntry;
	uint32 tblIdx = entryPtr->inboundGlobalId&((1<<RTL8651_ICMPTBL_BITS)-1);
	
	if(entryPtr->inboundToAsic)
		return SUCCESS;

	trackPtr = pptpGreInboundHeadPtr[tblIdx];
	while(trackPtr) {
		if(trackPtr->inboundIdInitiated == 1 && trackPtr->inboundToAsic==1)
			break;
		trackPtr = trackPtr->inboundNext;
	}
	if(trackPtr == NULL) {
		trackPtr = pptpGreOutboundHeadPtr[tblIdx];
		while(trackPtr) {
			if(trackPtr->outboundIdInitiated == 1 && trackPtr->outboundToAsic==1)
				break;
			trackPtr = trackPtr->outboundNext;
		}
	}
	if(trackPtr==NULL) {//Neigher inbound nor outbound configured to ASIC
		memset(&pptpAsicEntry, 0, sizeof(rtl865x_tblAsicDrv_naptIcmpParam_t));
		pptpAsicEntry.insideLocalId = entryPtr->inboundLocalId;
		pptpAsicEntry.insideLocalIpAddr = entryPtr->insideLocalIpAddr;
		pptpAsicEntry.offset = (entryPtr->inboundGlobalId>>10) & 0x3F;
		pptpAsicEntry.isCollision = 1;
		pptpAsicEntry.isStatic = 1;
		pptpAsicEntry.isValid = 1;
		pptpAsicEntry.isPptp = 1;
		pptpAsicEntry.isSpi = 0;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx , &pptpAsicEntry);
		entryPtr->inboundToAsic = 1;
		return SUCCESS;
	}
	return FAILED;
}

/*
@func int32	| _rtl8651_delAsicPptpPassthroughInboundEntry	|  Delete PPTP passthrough inbound entry to switch core
@parm _rtl8651_pptpGreEntry_t *	| entryPtr		| PPTP passthrough entry
@parm struct ip *			| pip		| IP header of the arrived packet
@comm 
1. If entry already not in ICMP table, definite success
2. Remove the entry from ASIC. ASIC corresonding entry kept collision but not valid
3. First search for inbound pure software flow and then outbound pure software flow

*/
int32 _rtl8651_delAsicPptpPassthroughInboundEntry(_rtl8651_pptpGreEntry_t * entryPtr) {
	_rtl8651_pptpGreEntry_t *trackPtr;
	rtl865x_tblAsicDrv_naptIcmpParam_t pptpAsicEntry;
	uint32 tblIdx = entryPtr->inboundGlobalId&((1<<RTL8651_ICMPTBL_BITS)-1);
	memset(&pptpAsicEntry, 0, sizeof(rtl865x_tblAsicDrv_naptIcmpParam_t));

	if(entryPtr->inboundToAsic == 0)
		return SUCCESS;

	entryPtr->inboundToAsic = 0;
	trackPtr = pptpGreInboundHeadPtr[tblIdx];
	while(trackPtr) {
		if(trackPtr->inboundIdInitiated == 1 && trackPtr->inboundToAsic==0) {
			pptpAsicEntry.insideLocalId = trackPtr->inboundLocalId;
			pptpAsicEntry.offset = (trackPtr->inboundGlobalId>>10) & 0x3F;
			entryPtr->inboundToAsic = 1;
			break;
		}
		trackPtr = trackPtr->inboundNext;
	}
	if(trackPtr == NULL) {
		trackPtr = pptpGreOutboundHeadPtr[tblIdx];
		while(trackPtr) {
			if(trackPtr->outboundIdInitiated == 1 && trackPtr->outboundToAsic==0) {
				pptpAsicEntry.insideLocalId = trackPtr->outboundLocalId;
				pptpAsicEntry.offset = (trackPtr->outboundGlobalId>>10) & 0x3F;
				entryPtr->outboundToAsic = 1;
				break;
			}
			trackPtr = trackPtr->outboundNext;
		}
	}
	if(trackPtr) {//Candidate entry found
		pptpAsicEntry.insideLocalIpAddr = trackPtr->insideLocalIpAddr;
		pptpAsicEntry.isCollision = 1;
		pptpAsicEntry.isStatic = 1;
		pptpAsicEntry.isValid = 1;
		pptpAsicEntry.isPptp = 1;
		pptpAsicEntry.isSpi = 0;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx , &pptpAsicEntry);
	} else {
		memset(&pptpAsicEntry, 0, sizeof(rtl865x_tblAsicDrv_naptIcmpParam_t));
		pptpAsicEntry.isCollision = 1;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx, &pptpAsicEntry);
	}
	return SUCCESS;
}

int32 _rtl8651_addAsicPptpPassthroughOutboundEntry(_rtl8651_pptpGreEntry_t * entryPtr) {
	_rtl8651_pptpGreEntry_t *trackPtr;
	rtl865x_tblAsicDrv_naptIcmpParam_t pptpAsicEntry;
	uint32 tblIdx = entryPtr->outboundGlobalId&((1<<RTL8651_ICMPTBL_BITS)-1);

	if(entryPtr->outboundToAsic)
		return SUCCESS;
	
	trackPtr = pptpGreInboundHeadPtr[tblIdx];
	while(trackPtr) {
		if(trackPtr->inboundIdInitiated == 1 && trackPtr->inboundToAsic==1)
			break;
		trackPtr = trackPtr->inboundNext;
	}
	if(trackPtr == NULL) {
		trackPtr = pptpGreOutboundHeadPtr[tblIdx];
		while(trackPtr) {
			if(trackPtr->outboundIdInitiated == 1 && trackPtr->outboundToAsic==1)
				break;
			trackPtr = trackPtr->outboundNext;
		}
	}
	if(trackPtr==NULL) {//Neigher inbound nor outbound configured to ASIC
		memset(&pptpAsicEntry, 0, sizeof(rtl865x_tblAsicDrv_naptIcmpParam_t));
		pptpAsicEntry.insideLocalId = entryPtr->outboundLocalId;
		pptpAsicEntry.insideLocalIpAddr = entryPtr->insideLocalIpAddr;
		pptpAsicEntry.offset = (entryPtr->outboundGlobalId>>10) & 0x3F;
		pptpAsicEntry.isCollision = 1;
		pptpAsicEntry.isStatic = 1;
		pptpAsicEntry.isValid = 1;
		pptpAsicEntry.isPptp = 1;
		pptpAsicEntry.isSpi = 0;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx, &pptpAsicEntry);
		entryPtr->outboundToAsic = 1;
		//printk("add asic ok\n");
		return SUCCESS;
	}
	//printk("add asic failed\n");	
	return FAILED;
}


int32 _rtl8651_delAsicPptpPassthroughOutboundEntry(_rtl8651_pptpGreEntry_t * entryPtr) {
	_rtl8651_pptpGreEntry_t *trackPtr;
	rtl865x_tblAsicDrv_naptIcmpParam_t pptpAsicEntry;
	uint32 tblIdx = entryPtr->outboundGlobalId&((1<<RTL8651_ICMPTBL_BITS)-1);

	if(entryPtr->outboundToAsic == 0)
		return SUCCESS;

	entryPtr->outboundToAsic = 0;
	trackPtr = pptpGreOutboundHeadPtr[tblIdx];
	while(trackPtr) {
		if(trackPtr->outboundIdInitiated == 1 && trackPtr->outboundToAsic==0) {
			pptpAsicEntry.insideLocalId = trackPtr->outboundLocalId;
			pptpAsicEntry.offset = (trackPtr->outboundGlobalId>>10) & 0x3F;
			trackPtr->outboundToAsic = 1;
			break;
		}
		trackPtr = trackPtr->outboundNext;
	}
	if(trackPtr == NULL) {
		trackPtr = pptpGreInboundHeadPtr[tblIdx];
		while(trackPtr) {
			if(trackPtr->inboundIdInitiated == 1 && trackPtr->inboundToAsic==0) {
				pptpAsicEntry.insideLocalId = trackPtr->inboundLocalId;
				pptpAsicEntry.offset = (trackPtr->inboundGlobalId>>10) & 0x3F;
				trackPtr->inboundToAsic = 1;
				break;
			}
			trackPtr = trackPtr->inboundNext;
		}
	}
	if(trackPtr) {//Candidate entry found
		pptpAsicEntry.insideLocalIpAddr = trackPtr->insideLocalIpAddr;
		pptpAsicEntry.isCollision = 1;
		pptpAsicEntry.isStatic = 1;
		pptpAsicEntry.isValid = 1;
		pptpAsicEntry.isPptp = 1;
		pptpAsicEntry.isSpi = 0;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx , &pptpAsicEntry);
	} else {
		memset(&pptpAsicEntry, 0, sizeof(rtl865x_tblAsicDrv_naptIcmpParam_t));
		pptpAsicEntry.isCollision = 1;
		rtl8651_setAsicNaptIcmpTable(TRUE, tblIdx, &pptpAsicEntry);
	}
	return SUCCESS;
}

/*
@func int32 | _rtl8651_l4GreNaptAliasIn	|  Create PPTP GRE inbound entry
@parm struct ip *	| pip		| IP header pointer
@parm uint16 *	| ptype		| Variable to store Control Message Type
@comm 
	When received outgoing-call-request, create inbound session.
*/
static int32 _rtl8651_l4GreNaptAliasIn(struct rtl_pktHdr *phdr,struct ip * pip) {
	_rtl8651_pptpGreEntry_t * tempPtr, *checkPtr, * asicPtr;
	GreHdr   *gr;
	uint32 hdr_bit;
	uint16 gh_call_id;
	uint32 srcIp, dstIp;
	uint8 software=0;
	int i;

	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;


	
	gr = (GreHdr *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
	hdr_bit = PKTGET_UINT32_UNALIGNED(gr);
	gh_call_id = ntohs(gr->gh_call_id);
	srcIp = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	dstIp = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

	if(pptpPassthroughGreInboundMsg)
		printk("PPTP-PTH: Data GRE In ID=%d DIP %d.%d.%d.%d SIP %d.%d.%d.%d\n",
			gh_call_id, (srcIp>>24)&0xff, (srcIp>>16)&0xff, (srcIp>>8)&0xff, srcIp&0xff,
			(dstIp>>24)&0xff, (dstIp>>16)&0xff, (dstIp>>8)&0xff, dstIp&0xff);	

	/* Check GRE header bits. */
	if ((hdr_bit & PPTP_INIT_MASK) != PPTP_INIT_VALUE)
		return FWDENG_GRE_IGNORE;


	if(software)
	{

		tempPtr=softwarePptpGreEntryHeadPtr;
		for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
		{
			if(tempPtr->inboundIdInitiated &&  tempPtr->inboundGlobalId == gh_call_id && tempPtr->insideGlobalIpAddr == dstIp && tempPtr->outsideGlobalIpAddr == srcIp)
				break;
		}
		if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;

		if(tempPtr)
		{
			uint8 *dip;
			tempPtr->inboundIdleTimer = 0;
			/* Restore original IP address. */					 
			dip=(uint8 *)&pip->ip_dst;
			dip[0]=(tempPtr->insideLocalIpAddr>>24)&0xff;
			dip[1]=(tempPtr->insideLocalIpAddr>>16)&0xff;
			dip[2]=(tempPtr->insideLocalIpAddr>>8)&0xff;
			dip[3]=tempPtr->insideLocalIpAddr&0xff;
			return SUCCESS;
		}
		
	}

	else //hardware
	{
		tempPtr = pptpGreInboundHeadPtr[gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1)];
/*		printk("PPTP-PTH: GRE IN ID Idx = %d\n",gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1)); */

		while(tempPtr) {
/*
			printk("PPTP-PTH: GRE Search Entry GID:%d Inside GIP %d.%d.%d.%d Outside GIP %d.%d.%d.%d\n",
					tempPtr->inboundGlobalId, 
					(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff,
					(tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,
					(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff,
					(tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff); */
			if(tempPtr->inboundIdInitiated &&  tempPtr->inboundGlobalId == gh_call_id && tempPtr->insideGlobalIpAddr == dstIp && tempPtr->outsideGlobalIpAddr == srcIp)
				break;
			tempPtr = tempPtr->inboundNext;
		}
		if (tempPtr)	{
			uint8 *dip;
			/* De-alias the Peer's Call Id. */
			gr->gh_call_id = htons(tempPtr->inboundLocalId);

			/* Restore original IP address. */					 
			dip=(uint8 *)&pip->ip_dst;
			dip[0]=(tempPtr->insideLocalIpAddr>>24)&0xff;
			dip[1]=(tempPtr->insideLocalIpAddr>>16)&0xff;
			dip[2]=(tempPtr->insideLocalIpAddr>>8)&0xff;
			dip[3]=tempPtr->insideLocalIpAddr&0xff;
			phdr->ph_dip = tempPtr->insideLocalIpAddr;
			//printk("gre in original ip= %x %x\n",dip,tempPtr->insideLocalIpAddr);

			tempPtr->inboundIdleTimer = 0;
			tempPtr->inboundPktCnt++;
			if(tempPtr->inboundPktCnt == pptpPassthroughSysPara.heavyLoadSwappingThreshold) {
				asicPtr = NULL;
				checkPtr = pptpGreInboundHeadPtr[gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1)];
				while(checkPtr) {
					if(checkPtr->inboundToAsic == 1) {
						assert(asicPtr == NULL);
						asicPtr = checkPtr;
					}
					if(checkPtr != tempPtr && 
						checkPtr->inboundPktCnt >= pptpPassthroughSysPara.heavyLoadSwappingThreshold)
						break;
					checkPtr = checkPtr->inboundNext;
				}
				if(checkPtr == NULL) {
					checkPtr = pptpGreOutboundHeadPtr[gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1)];
					while(checkPtr) {
						if(checkPtr->outboundToAsic == 1) {
							assert(asicPtr == NULL);
							asicPtr = checkPtr;
						}
						if(checkPtr->outboundPktCnt >= pptpPassthroughSysPara.heavyLoadSwappingThreshold)
							break;
						checkPtr = checkPtr->outboundNext;
					}
				}
				if(checkPtr == NULL) {//There is no previous swapped entry
					assert(asicPtr);
					//printk("PPTP-PTH: Swap inbound [%x]\n", gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1));
					if(asicPtr->inboundToAsic==1 && 
						(asicPtr->inboundGlobalId & ((1<<RTL8651_ICMPTBL_BITS)-1)) == (gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1))) {
						asicPtr->inboundToAsic=0;
						//printk("PPTP-PTH: Org inbound Src %d.%d.%d.%d Dst %d.%d.%d.%d GID:%x LID:%x Age:%u Pkt:%u\n",
						//	(asicPtr->outsideGlobalIpAddr>>24)&0xff, (asicPtr->outsideGlobalIpAddr>>16)&0xff, (asicPtr->outsideGlobalIpAddr>>8)&0xff, asicPtr->outsideGlobalIpAddr&0xff,
						//	(asicPtr->insideGlobalIpAddr>>24)&0xff, (asicPtr->insideGlobalIpAddr>>16)&0xff, (asicPtr->insideGlobalIpAddr>>8)&0xff, asicPtr->insideGlobalIpAddr&0xff,
						//	asicPtr->inboundGlobalId, asicPtr->inboundLocalId, asicPtr->inboundIdleTimer, asicPtr->inboundPktCnt);
					}
					else {
						assert(asicPtr->outboundToAsic==1);
						assert((asicPtr->outboundGlobalId& ((1<<RTL8651_ICMPTBL_BITS)-1)) == (gh_call_id&((1<<RTL8651_ICMPTBL_BITS)-1)));
						asicPtr->outboundToAsic=0;
						//printk("PPTP-PTH: Org outbound Src %d.%d.%d.%d Dst %d.%d.%d.%d GID:%x LID:%x Age:%u Pkt:%u\n",
						//	(asicPtr->outsideGlobalIpAddr>>24)&0xff, (asicPtr->outsideGlobalIpAddr>>16)&0xff, (asicPtr->outsideGlobalIpAddr>>8)&0xff, asicPtr->outsideGlobalIpAddr&0xff,
						//	(asicPtr->insideGlobalIpAddr>>24)&0xff, (asicPtr->insideGlobalIpAddr>>16)&0xff, (asicPtr->insideGlobalIpAddr>>8)&0xff, asicPtr->insideGlobalIpAddr&0xff,
						//	asicPtr->outboundGlobalId, asicPtr->outboundLocalId, asicPtr->inboundIdleTimer, asicPtr->outboundPktCnt);
					}
					//printk("PPTP-PTH: New Src %d.%d.%d.%d Dst %d.%d.%d.%d GID:%x LID:%x Age:%u Pkt:%u\n",
					//	(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
					//	(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,
					//	tempPtr->inboundGlobalId, tempPtr->inboundLocalId, tempPtr->inboundIdleTimer, tempPtr->inboundPktCnt);
					_rtl8651_addAsicPptpPassthroughInboundEntry(tempPtr);
				}
			}
			/* Compute IP checksum for revised packet */
			//pip->ip_sum = 0;				
			//pip->ip_sum = _rtl8651_ipChecksum(pip);								
			//printk("in checksum=%x\n",pip->ip_sum);
			//phdr->ph_flags|=CSUM_IP;
			//phdr->ph_flags&=~CSUM_L4;
			//phdr->ph_proto=PKTHDR_IP;		
			return SUCCESS;
		}
		else {
			if(pptpPassthroughGreInboundMsg) printk("PPTP-PTH: GRE IN no existed entry \n");
		}
	}
	return FWDENG_GRE_IGNORE;
}


/*
@func _rtl8651_pptpGreEntry_t *	| _rtl8651_createPptpGreInboundEntry	|  Create PPTP GRE inbound entry
@parm struct ip *	| pip		| IP header pointer
@parm uint16 *	| ptype		| Variable to store Control Message Type
@comm 
	When received outgoing-call-request, create inbound session.
*/
static int32 _rtl8651_l4GreNaptAliasOut(struct rtl_pktHdr *phdr,struct ip *pip)
{
	_rtl8651_pptpGreEntry_t * tempPtr, *checkPtr, * asicPtr;
	GreHdr   *gr;
	uint32 hdr_bit, tblIdx;
	uint16 gh_call_id;
	uint32 srcIp, dstIp;
	uint8 software=0;
	int i;

	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;
	gr = (GreHdr *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
	hdr_bit = PKTGET_UINT32_UNALIGNED(gr);
	gh_call_id =  ntohs(gr->gh_call_id) ;
	srcIp = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	dstIp = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

	if(pptpPassthroughGreOutboundMsg)
		printk("PPTP-PTH: Data GRE Out ID=%d DIP %d.%d.%d.%d SIP %d.%d.%d.%d\n",
			gh_call_id, (dstIp>>24)&0xff, (dstIp>>16)&0xff, (dstIp>>8)&0xff, dstIp&0xff,
			(srcIp>>24)&0xff, (srcIp>>16)&0xff, (srcIp>>8)&0xff, srcIp&0xff );	

	/* Check GRE header bits.  */	
	if ((hdr_bit & PPTP_INIT_MASK) != PPTP_INIT_VALUE)
		return FWDENG_GRE_IGNORE;

	if(software)
	{
		tempPtr=softwarePptpGreEntryHeadPtr;
		for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
		{
			if(tempPtr->outboundIdInitiated && tempPtr->insideLocalIpAddr == srcIp && tempPtr->outboundLocalId == gh_call_id && tempPtr->outsideGlobalIpAddr == dstIp)
				break;
		}
		if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;

		if(tempPtr)
		{
			uint8 *sip;
			tempPtr->outboundIdleTimer = 0;
			sip=(uint8 *)&pip->ip_src;		
			sip[0]=(tempPtr->insideGlobalIpAddr>>24)&0xff;
			sip[1]=(tempPtr->insideGlobalIpAddr>>16)&0xff;
			sip[2]=(tempPtr->insideGlobalIpAddr>>8)&0xff;
			sip[3]=tempPtr->insideGlobalIpAddr&0xff;
			
			return SUCCESS;
		}
		
	}
	else
	{

		rtl8651_naptIcmpTableIndex(srcIp, gh_call_id, dstIp, &tblIdx);	
		//printk("gre out tblIdx=%d srcip=%x gh_callid=%x dstIP=%x\n",tblIdx,srcIp,gh_call_id,dstIp);
		tempPtr = pptpGreOutboundHeadPtr[tblIdx];


		while(tempPtr) {
			//printk("gre out tempPtr->insideLocalIpAddr(%x) == srcIp(%x) && tempPtr->outboundLocalId(%x) == gh_call_id(%x) && tempPtr->outsideGlobalIpAddr(%x) == dstIp(%x)\n",
			//	tempPtr->insideLocalIpAddr,srcIp,tempPtr->outboundLocalId,gh_call_id,tempPtr->outsideGlobalIpAddr,dstIp);
			
			if(tempPtr->outboundIdInitiated && tempPtr->insideLocalIpAddr == srcIp && tempPtr->outboundLocalId == gh_call_id && tempPtr->outsideGlobalIpAddr == dstIp)
				break;
			tempPtr = tempPtr->outboundNext;
		}
		if (tempPtr) {
			uint8 *sip;

			gr->gh_call_id = htons(tempPtr->outboundGlobalId);

			sip=(uint8 *)&pip->ip_src;		
			sip[0]=(tempPtr->insideGlobalIpAddr>>24)&0xff;
			sip[1]=(tempPtr->insideGlobalIpAddr>>16)&0xff;
			sip[2]=(tempPtr->insideGlobalIpAddr>>8)&0xff;
			sip[3]=tempPtr->insideGlobalIpAddr&0xff;
			phdr->ph_sip = tempPtr->insideGlobalIpAddr;
			//printk("new gre out sip=%x\n",sip);

			tempPtr->outboundIdleTimer = 0;
			tempPtr->outboundPktCnt++;
			if(tempPtr->outboundPktCnt == pptpPassthroughSysPara.heavyLoadSwappingThreshold) {			
				asicPtr = NULL;
				checkPtr = pptpGreOutboundHeadPtr[tblIdx];
				while(checkPtr) {
					if(checkPtr->outboundToAsic == 1) {
						assert(asicPtr == NULL);
						asicPtr = checkPtr;
					}
					if(checkPtr != tempPtr && 
						checkPtr->outboundPktCnt >= pptpPassthroughSysPara.heavyLoadSwappingThreshold)
						break;
					checkPtr = checkPtr->outboundNext;
				}
				if(checkPtr == NULL) {
					checkPtr = pptpGreInboundHeadPtr[tblIdx];
					while(checkPtr) {
						if(checkPtr->inboundToAsic == 1) {
							assert(asicPtr == NULL);
							asicPtr = checkPtr;
						}
						if(checkPtr->inboundPktCnt >= pptpPassthroughSysPara.heavyLoadSwappingThreshold)
							break;
						checkPtr = checkPtr->inboundNext;
					}
				}
				if(checkPtr == NULL) {//There is no previous swapped entry
					assert(asicPtr);
					//printk("PPTP-PTH: Swap outbound [%x]\n", tblIdx);
					if(asicPtr->inboundToAsic==1 && (asicPtr->inboundGlobalId & ((1<<RTL8651_ICMPTBL_BITS)-1)) == tblIdx) {
						asicPtr->inboundToAsic=0;
						//printk("PPTP-PTH: Org inbound Src %d.%d.%d.%d Dst %d.%d.%d.%d GID:%x LID:%x Age:%u Pkt:%u\n",
						//	(asicPtr->outsideGlobalIpAddr>>24)&0xff, (asicPtr->outsideGlobalIpAddr>>16)&0xff, (asicPtr->outsideGlobalIpAddr>>8)&0xff, asicPtr->outsideGlobalIpAddr&0xff,
						//	(asicPtr->insideGlobalIpAddr>>24)&0xff, (asicPtr->insideGlobalIpAddr>>16)&0xff, (asicPtr->insideGlobalIpAddr>>8)&0xff, asicPtr->insideGlobalIpAddr&0xff,
						//	asicPtr->inboundGlobalId, asicPtr->inboundLocalId, asicPtr->inboundIdleTimer, asicPtr->inboundPktCnt);
					}
					else {
						assert(asicPtr->outboundToAsic==1);
						assert((asicPtr->outboundGlobalId & ((1<<RTL8651_ICMPTBL_BITS)-1)) == tblIdx);
						asicPtr->outboundToAsic=0;
						//printk("PPTP-PTH: Org outbound Src %d.%d.%d.%d Dst %d.%d.%d.%d GID:%x LID:%x Age:%u Pkt:%u\n",
						//	(asicPtr->outsideGlobalIpAddr>>24)&0xff, (asicPtr->outsideGlobalIpAddr>>16)&0xff, (asicPtr->outsideGlobalIpAddr>>8)&0xff, asicPtr->outsideGlobalIpAddr&0xff,
						//	(asicPtr->insideGlobalIpAddr>>24)&0xff, (asicPtr->insideGlobalIpAddr>>16)&0xff, (asicPtr->insideGlobalIpAddr>>8)&0xff, asicPtr->insideGlobalIpAddr&0xff,
						//	asicPtr->outboundGlobalId, asicPtr->outboundLocalId, asicPtr->inboundIdleTimer, asicPtr->outboundPktCnt);
					}
					//printk("PPTP-PTH: New Src %d.%d.%d.%d Dst %d.%d.%d.%d GID:%x LID:%x Age:%u Pkt:%u\n",
					//	(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
					//	(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,
					//	tempPtr->inboundGlobalId, tempPtr->inboundLocalId, tempPtr->inboundIdleTimer, tempPtr->inboundPktCnt);
					_rtl8651_addAsicPptpPassthroughOutboundEntry(tempPtr);
				}
			}
			/* Compute IP checksum for revised packet */
			//pip->ip_sum = 0;				
			//pip->ip_sum = _rtl8651_ipChecksum(pip);				
			//phdr->ph_flags|=CSUM_IP;
			//phdr->ph_flags&=~CSUM_L4;
			//phdr->ph_proto=PKTHDR_IP;
			
			return SUCCESS;	
		}
	}

	return FWDENG_GRE_IGNORE;
}



/*
@func PptpCallId	| _rtl8651_aliasVerifyPptp	|  Analyze PPTP control packet (TCP) and return ID when corresponding fields do exist.
@parm struct ip *	| pip		| IP header pointer
@parm uint16 *	| ptype		| Variable to store Control Message Type
@comm 
	Using length information to decide whether required message do exist. If exist, return the Call ID and Peer's Call ID pointer.
*/
static PptpCallId _rtl8651_aliasVerifyPptp(struct ip *pip, uint16 * ptype)	/* IP packet to examine/patch */
{
	int32     hlen, tlen, dlen;
	PptpMsgHead hptr;
	struct tcphdr *tc;

	uint32 magic;

	/* Calculate some lengths */
	tc =  (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));//Uising IP header address plus header length to get TCP header 
	hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >>4)&0xf)) << 2;//IP + TCP header length
	tlen = ntohs(pip->ip_len);//Total IP packet length include IP header length, this is got from IP header
	dlen = tlen - hlen;//TCP data part length

	/* Verify data length, if the length is unable to analyze outgoing-call-request or outgoing-call-reply, this should be invalid PPTP control packet  */
	if (dlen < (sizeof(struct pptpMsgHead) + sizeof(struct pptpCallIds)))//pptpMsgHead and pptpCallIds are just enough to get session ID and Peer's call ID
		return (NULL);

	/* Move up to PPTP message header */
	hptr = (PptpMsgHead) (((int8 *) pip) + hlen);

	/* Return the control message type */
	*ptype = ntohs(hptr->type);
	
	/* Verify PPTP Control Message  */

	magic = PKTGET_UINT32_UNALIGNED(&hptr->magic);
	
	if ((ntohs(hptr->msgType) != PPTP_CTRL_MSG_TYPE) ||
		(magic != PPTP_MAGIC))
		return (NULL);

	/* Verify data length. */
	if ((*ptype == PPTP_OutCallReply || *ptype == PPTP_InCallReply) &&
		(dlen < sizeof(struct pptpMsgHead) + sizeof(struct pptpCallIds) +
		 sizeof(struct pptpCodes)))//When packet is reply type, must have result code and error code to decide whether ID field is valid
		return (NULL);
	else
		return (PptpCallId) (hptr + 1);

}



/*
@func int32	| rtl8651_l4AliasHandlePptpClientOutbound	|  Analysis outbound traffic
@parm struct rtl_pktHdr *	| pkt		| Packet header of the arrived packet
@parm struct ip *			| pip		| IP header of the arrived packet
@parm struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *	| link		| Session to store the ALG control channel entry
@comm 
1. Create PPTP GRE entry when Outgoing-call-request is received

*/
int32 rtl8651_l4AliasHandlePptpClientOutbound(
		struct rtl_pktHdr *pkt,
		struct ip *pip,
		struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
	_rtl8651_pptpGreEntry_t * tempPtr=NULL;
	PptpCallId cptr;
	uint16    ctl_type;				   /* control message type */
	uint32 i, entryIdx, tblIdx;
	uint8 software=0;
	static uint32 searchIdx = 0;
	
	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;
	if((alg_supports[pkt->ph_dsid]&RTL8651_ALG_PPTP)==0) return FWDENG_USER_DROP;

	/* Verify valid PPTP control message */
	if ((cptr = _rtl8651_aliasVerifyPptp(pip, &ctl_type)) == NULL)
		return SUCCESS;


	/* Modify certain PPTP messages */
	switch (ctl_type)
	{
		case PPTP_OutCallRequest:
			if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH PKT: Client Outbound OutCallRequest\n");
			if(!software)
			{
				for(i=0, tempPtr = NULL; tempPtr==NULL && i<RTL8651_ICMPTBL_SIZE; i++) {
					tempPtr = pptpGreInboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideLocalIpAddr == link->insideLocalIpAddr &&
							tempPtr->inboundLocalId == ntohs(cptr->cid1) &&	tempPtr->inboundIdInitiated == 1)
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
			}				
			else //software
			{

				tempPtr=softwarePptpGreEntryHeadPtr;

				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{	
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideLocalIpAddr == link->insideLocalIpAddr &&
					   tempPtr->inboundLocalId == ntohs(cptr->cid1) &&	tempPtr->inboundIdInitiated == 1)
							break;
				}
				
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				else
				{
//					printk("kill previous entry...\n");
					memset(tempPtr,0,sizeof(tempPtr));
					tempPtr=NULL;
				}
				
			
			}
	
			
			if(tempPtr == NULL) {	/* Establish PPTP link for address and Call ID found in control message.  */
				if(!software)
				{
					if(freePptpGreEntryHeadPtr == NULL)
						return FWDENG_USER_DROP; //entry full				
					tempPtr = freePptpGreEntryHeadPtr;
					freePptpGreEntryHeadPtr = freePptpGreEntryHeadPtr->inboundNext;
				}
				else //find a free entry to put
				{
					_rtl8651_pptpGreEntry_t empty;
					memset(&empty,0,sizeof(empty));
	
					tempPtr=softwarePptpGreEntryHeadPtr;
					for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
					{
						if(memcmp(tempPtr,&empty,sizeof(empty))==0)
								break;
					}
					if(i==algMaxSession.pptpMaxSession) return SUCCESS; //entry full
					
				}
				
				tempPtr->outsideGlobalIpAddr = link->dstIpAddr;
				tempPtr->insideGlobalIpAddr = link->insideGlobalIpAddr;
				tempPtr->insideLocalIpAddr = link->insideLocalIpAddr;
				tempPtr->inboundLocalId = ntohs(cptr->cid1);
				tempPtr->inboundIdInitiated = 1;
				tempPtr->outboundIdInitiated = 0;
				tempPtr->inboundToAsic = 0;
				tempPtr->outboundToAsic = 0;
				tempPtr->inboundIdClosed = 0;
				tempPtr->outboundIdClosed = 0;
				tempPtr->controlIdleTimer = 0;
				tempPtr->inboundIdleTimer = 0;
				tempPtr->inboundPktCnt = 0;
				tempPtr->outboundIdleTimer = 0;
				tempPtr->outboundPktCnt = 0;
				tempPtr->dsid = pkt->ph_dsid;

				if(software==1)
				{
					tempPtr->inboundGlobalId = tempPtr->inboundLocalId;
				}
				else
				{//Calculate the inbound Global Id from local IP and local ID
					_rtl8651_pptpGreEntry_t * trackPtr;

					tempPtr->inboundGlobalId = rtl8651_naptIcmpTableIndex(tempPtr->insideLocalIpAddr, tempPtr->inboundLocalId, tempPtr->outsideGlobalIpAddr, &tblIdx);
					if(pptpPassthroughClientOutboundMsg)
						printk("PPTP-PH: Client from LAN OutCallRequest SIP: %d.%d.%d.%d DIP:%d.%d.%d.%d ID %d\n",
							(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff,
							(tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff,
							(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff,
							(tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
							tempPtr->inboundGlobalId);
					if(pptpGreInboundHeadPtr[tblIdx]) {//Must find the free one only when list not empty 
						uint32 offsetMask[2], i;
					
						trackPtr = pptpGreInboundHeadPtr[tblIdx];
						offsetMask[0] = offsetMask[1] = 0;
						while(trackPtr) {
							assert(trackPtr->inboundIdInitiated);
							offsetMask[(trackPtr->inboundGlobalId>>10)&0x1] |= (1<<((trackPtr->inboundGlobalId>>11)&0x1F));
							trackPtr = trackPtr->inboundNext;
						}
						for(i=searchIdx; i!=((searchIdx+63)&0x3F); i=(i+1)&0x3F) {
							if((offsetMask[i&0x1] & (1<<((i>>1)&0x1F))) == 0)
								break;
						}
						searchIdx = (i+1)&0x3F;
						assert(i<64);
						tempPtr->inboundGlobalId |= (i<<10);
					}
					if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH: Add inbound GID:%d LID:%d\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					_rtl8651_addAsicPptpPassthroughInboundEntry(tempPtr);
					tempPtr->inboundNext = pptpGreInboundHeadPtr[tempPtr->inboundGlobalId &((1<<RTL8651_ICMPTBL_BITS)-1)];
					pptpGreInboundHeadPtr[tempPtr->inboundGlobalId &((1<<RTL8651_ICMPTBL_BITS)-1)] = tempPtr;

				}
				
			}
			else {
				if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH: Established inbound GID:%d LID:%d\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
				tempPtr->inboundIdClosed = 0;
				tempPtr->controlIdleTimer = 0;
			}
			cptr->cid1 = htons(tempPtr->inboundGlobalId);
			
			//printk("PPTP-PTH: Modify inbound from LID:%x to GID:%x\n", tempPtr->inboundLocalId, tempPtr->inboundGlobalId);
			break;
		case PPTP_CallClearRequest:
			if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH PKT: Client Outbound CallClearRequest\n");
			if(!software)
			{
			//Get corresponding inbound session
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreInboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->inboundLocalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
				if(tempPtr) {
#if 0					
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreInboundHeadPtr[entryIdx] == tempPtr) {
						pptpGreInboundHeadPtr[entryIdx] = tempPtr->inboundNext;
					} else {
						trackPtr = pptpGreInboundHeadPtr[entryIdx];
						while(trackPtr && trackPtr->inboundNext != tempPtr)
							trackPtr = trackPtr->inboundNext;
						assert(trackPtr);//trackPtr->inboundNext must point to the
						trackPtr->inboundNext = tempPtr->inboundNext;
					}
					tempPtr->inboundIdInitiated = 0;
					//printk("PPTP-PTH: Del inbound GIP:%d.%d.%d.%d LIP:%d. %d.%d.%d GID:%x LID:%x\n", 
					//	(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
					//	(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
					//	tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
					if(tempPtr->outboundIdInitiated)//Not empty
						tempPtr->inboundNext = NULL;
					else {
						tempPtr->inboundNext = freePptpGreEntryHeadPtr;
						freePptpGreEntryHeadPtr = tempPtr;
					}
#endif
					if(tempPtr->inboundIdClosed == 1 && pptpPassthroughClientOutboundMsg)
						printk("PPTP-PTH:(Call Clear Request) Inbound Closed ID GID:%d LID: %d\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					else if (pptpPassthroughClientOutboundMsg)
						printk("PPTP-PTH:(Call Clear Request) Inbound ID GID: %d LID: %d Closed\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					tempPtr->inboundIdClosed = 1;
					tempPtr->controlIdleTimer = 0;
					cptr->cid1 = htons(tempPtr->inboundGlobalId);
				}
				
			}
			else
			{

				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->inboundLocalId == ntohs(cptr->cid1))
					break;
				}
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				
				if(tempPtr)
				{
					tempPtr->inboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
					tempPtr->outboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
				}
			}
			break;
		case PPTP_WanErrorNotify:// Send by PAC to PNS to indicate WAN error condition(??)
		case PPTP_SetLinkInfo:// Send by PNS to PAC to set PPP-negotiated options.		
			if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH PKT: Client Outbound SetLinkInfo\n");
			if(!software)
			{
				//Get corresponding inbound session
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->outboundLocalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}

				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				if(tempPtr)
				{
					cptr->cid1=htons(tempPtr->outboundGlobalId);
					tempPtr->controlIdleTimer = 0;
				}					
			}
			break;
		case PPTP_StopCtrlConnRequest:
		case PPTP_StopCtrlConnReply:
			if(pptpPassthroughClientOutboundMsg) printk("PPTP-PTH PKT: Client Outbound StopCtrlConn\n");
			if(!software)	{
				do {
					for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
						entryIdx = i;
						tempPtr = pptpGreInboundHeadPtr[i];
						while(tempPtr) {
							if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && tempPtr->inboundIdClosed == 0)
								break;
							tempPtr = tempPtr->inboundNext;
						}
					}
					if(tempPtr) {
						if (pptpPassthroughClientOutboundMsg)
							printk("PPTP-PTH:(Stop Control Connection) Inbound ID GID: %d LID: %d Closed\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
						tempPtr->inboundIdClosed = 1;
						tempPtr->controlIdleTimer = 0;
						cptr->cid1 = htons(tempPtr->inboundGlobalId);
					}
				}while(tempPtr);
			}				
			break;
		case PPTP_InCallRequest:
		case PPTP_InCallReply:
		case PPTP_InCallConn: //Send by PAC to PNS in response to a received Incoming-Call-Reply			
			//printk("PPTP-PTH: Current not supported control message type: %x\n", ctl_type);
		default:
			return SUCCESS;
	}

	return SUCCESS;
}

/* Server inbound is equivalent to client outbound setting */
int32 rtl8651_l4AliasHandlePptpServerInbound(
	struct rtl_pktHdr *pkt, 
	struct ip *pip,
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s * link)	
{

	_rtl8651_pptpGreEntry_t * tempPtr=NULL;
	PptpCallId cptr;
	uint16    ctl_type;				   /* control message type */
	uint32 i, entryIdx, tblIdx;
	uint8 software=0;
	static uint32 searchIdx = 0, lastId=0;
	
	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;

	if((alg_supports[pkt->ph_dsid]&RTL8651_ALG_PPTP_SERVER)==0) return FWDENG_USER_DROP;

	/* Verify valid PPTP control message */
	if ((cptr = _rtl8651_aliasVerifyPptp(pip, &ctl_type)) == NULL)
		return SUCCESS;

	/* Modify certain PPTP messages */
	switch (ctl_type)
	{
		case PPTP_OutCallRequest:
			if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH PKT: Server Inbound OutCallRequest\n");
			if(!software)
			{
				for(i=0, tempPtr = NULL; tempPtr==NULL && i<RTL8651_ICMPTBL_SIZE; i++) {
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideLocalIpAddr == link->insideLocalIpAddr &&
							tempPtr->outboundGlobalId == ntohs(cptr->cid1) &&	tempPtr->outboundIdInitiated == 1)
							break;
						tempPtr = tempPtr->outboundNext;
					}
				}
			}
			else
			{				
				tempPtr=softwarePptpGreEntryHeadPtr;

				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{	
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideLocalIpAddr == link->insideLocalIpAddr &&
						tempPtr->outboundGlobalId == ntohs(cptr->cid1) &&	tempPtr->outboundIdInitiated == 1)
						break;
				}
				
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				
			}
			
			if(tempPtr == NULL) {	/* Establish PPTP link for address and Call ID found in control message.  */
				
				//find free entry
				if(!software)
				{						
					if(freePptpGreEntryHeadPtr == NULL)
						return FWDENG_USER_DROP;
					tempPtr = freePptpGreEntryHeadPtr;
					freePptpGreEntryHeadPtr = freePptpGreEntryHeadPtr->inboundNext;
				}
				else
				{
					_rtl8651_pptpGreEntry_t empty;
					memset(&empty,0,sizeof(empty));
	
					tempPtr=softwarePptpGreEntryHeadPtr;
					for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
					{
						if(memcmp(tempPtr,&empty,sizeof(empty))==0)
							break;
					}
					if(i==algMaxSession.pptpMaxSession) return SUCCESS; //entry full
				}
				
				tempPtr->outsideGlobalIpAddr = link->dstIpAddr;
				tempPtr->insideGlobalIpAddr = link->insideGlobalIpAddr;
				tempPtr->insideLocalIpAddr = link->insideLocalIpAddr;
				tempPtr->outboundGlobalId = ntohs(cptr->cid1);
				tempPtr->outboundIdInitiated = 1;
				
				tempPtr->inboundIdInitiated = 0;
				tempPtr->inboundToAsic = 0;
				tempPtr->outboundToAsic = 0;
				tempPtr->inboundIdClosed = 0;
				tempPtr->outboundIdClosed = 0;
				tempPtr->inboundIdleTimer = 0;
				tempPtr->inboundPktCnt = 0;
				tempPtr->outboundIdleTimer = 0;
				tempPtr->outboundPktCnt = 0;
				tempPtr->controlIdleTimer = 0; /* Reset control Idle timer*/
				
				tempPtr->dsid = pkt->ph_dsid;

				if(software)
				{
					tempPtr->outboundLocalId = tempPtr->outboundGlobalId;
					
				}
				else
				{
					{//Calculate the inbound Global Id from local IP and local ID
						_rtl8651_pptpGreEntry_t * trackPtr;
						uint16 localid;
						uint16 targetHashIdx=tempPtr->outboundGlobalId & 0x3ff;
						lastId = ((lastId +1)*(tempPtr->outboundGlobalId+1))&0xFFFF;
						if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH: Server Inbound try ID start at %d\n", lastId);
						for(localid = lastId;localid != ((lastId+65535)&0xFFFF);localid = (localid +1)&0xFFFF) 	{
							uint16 hashIdx;
							//tempPtr->outboundGlobalId = 
							hashIdx=rtl8651_naptIcmpTableIndex(tempPtr->insideLocalIpAddr, localid, tempPtr->outsideGlobalIpAddr, &tblIdx);
							if(hashIdx==targetHashIdx)
								{
									tempPtr->outboundLocalId=localid;
									break;
								}
						}
						lastId = localid;
						if(pptpGreOutboundHeadPtr[tblIdx]) {//Must find the free one only when list not empty 
							uint32 offsetMask[2], i;
							
							trackPtr = pptpGreOutboundHeadPtr[tblIdx];
							offsetMask[0] = offsetMask[1] = 0;
							while(trackPtr) {
								assert(trackPtr->outboundIdInitiated);
								offsetMask[(trackPtr->outboundLocalId>>10)&0x1] |= (1<<((trackPtr->outboundLocalId>>11)&0x1F));
								trackPtr = trackPtr->outboundNext;
							}
							if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH: Server Inbound try ID offset start at %d\n", searchIdx);
							for(i=searchIdx; i!=((searchIdx+63)&0x3F); i=(i+1)&0x3f) {
								if((offsetMask[i&0x1] & (1<<((i>>1)&0x1F))) == 0)
									break;
							}
							searchIdx = (i+1)&0x3F;
							assert(i<64);
							tempPtr->outboundLocalId |= (i<<10);
							lastId = tempPtr->outboundLocalId;
						}
						
						if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH: Add outbound GID:%d LID:%d to table index %d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId, tblIdx);
						_rtl8651_addAsicPptpPassthroughOutboundEntry(tempPtr);
					}
					//tempPtr->outboundNext = pptpGreOutboundHeadPtr[tempPtr->outboundGlobalId &((1<<RTL8651_ICMPTBL_BITS)-1)];
					//pptpGreOutboundHeadPtr[tempPtr->outboundGlobalId &((1<<RTL8651_ICMPTBL_BITS)-1)] = tempPtr;
					
					rtl8651_naptIcmpTableIndex(tempPtr->insideLocalIpAddr, tempPtr->outboundLocalId, tempPtr->outsideGlobalIpAddr, &tblIdx);
					tempPtr->outboundNext = pptpGreOutboundHeadPtr[tblIdx];
					pptpGreOutboundHeadPtr[tblIdx] = tempPtr;
				}
			}
			else {
				if(tempPtr->outboundIdClosed == 1 && pptpPassthroughServerInboundMsg)
					printk("PPTP-PTH: Closed Outbound ID GID:%d LID:%d refreshed to normal state\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
				tempPtr->outboundIdClosed = 0;
				tempPtr->controlIdleTimer = 0; /* Reset control Idle timer*/
			}
			cptr->cid1 = htons(tempPtr->outboundLocalId);
			if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH: Modify Server Inbound  Outcall Request from GID:%d to LID:%d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
			break;
		case PPTP_CallClearRequest:
			if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH PKT: Server Inbound CallClearRequest\n");
			if(!software)
			{
				//Get corresponding inbound session
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->outboundGlobalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->outboundNext;
					}
				}
				if(tempPtr) {
#if 0					
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreOutboundHeadPtr[entryIdx] == tempPtr) {
						pptpGreOutboundHeadPtr[entryIdx] = tempPtr->outboundNext;
					} else {
						trackPtr = pptpGreOutboundHeadPtr[entryIdx];
						while(trackPtr && trackPtr->outboundNext != tempPtr)
							trackPtr = trackPtr->outboundNext;
						assert(trackPtr);//trackPtr->inboundNext must point to the
						trackPtr->outboundNext = tempPtr->outboundNext;
					}
					tempPtr->outboundIdInitiated = 0;
					printk("PPTP-PTH: Del outbound GIP:%d.%d.%d.%d LIP:%d. %d.%d.%d GID:%d LID:%d\n", 
						(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
						(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
						tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
					_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
					if(tempPtr->inboundIdInitiated)//Not empty
						tempPtr->outboundNext = NULL;
					else {
						tempPtr->inboundNext = freePptpGreEntryHeadPtr;
						freePptpGreEntryHeadPtr = tempPtr;
					}
#endif
					if(tempPtr->outboundIdClosed == 1 && pptpPassthroughServerInboundMsg)
						printk("PPTP-PTH:(Call Clear Request) Outbound Closed ID GID:%d LID: %d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
					else if (pptpPassthroughServerInboundMsg)
						printk("PPTP-PTH:(Call Clear Request) Outbound ID GID: %d LID: %d Closed\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
					tempPtr->outboundIdClosed = 1;
					tempPtr->controlIdleTimer = 0; /* Reset control Idle timer*/
					cptr->cid1 = htons(tempPtr->outboundLocalId);
				}
			}			
			else
			{

				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->outboundGlobalId == ntohs(cptr->cid1))
					break;
				}
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
			
				if(tempPtr)
				{
					tempPtr->inboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
					tempPtr->outboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
				}
			}			
			break;
		case PPTP_StopCtrlConnRequest:
		case PPTP_StopCtrlConnReply:
			if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH PKT: Server Inbound StopCtrlConnRequest and StopCtrlConnReply\n");
			if(!software)
			{
				do {
					//Get corresponding inbound session
					for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
						entryIdx = i;
						tempPtr = pptpGreOutboundHeadPtr[i];
						while(tempPtr) {
							if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && tempPtr->outboundIdClosed == 0)
								break;
							tempPtr = tempPtr->outboundNext;
						}
					}
					if(tempPtr) {
#if 0
						_rtl8651_pptpGreEntry_t * trackPtr;
						if(pptpGreOutboundHeadPtr[entryIdx] == tempPtr) {
							pptpGreOutboundHeadPtr[entryIdx] = tempPtr->outboundNext;
						} else {
							trackPtr = pptpGreOutboundHeadPtr[entryIdx];
							while(trackPtr && trackPtr->outboundNext != tempPtr)
								trackPtr = trackPtr->outboundNext;
							assert(trackPtr);//trackPtr->inboundNext must point to the
							trackPtr->outboundNext = tempPtr->outboundNext;
						}
						tempPtr->outboundIdInitiated = 0;
						printk("PPTP-PTH: Del outbound GIP:%d.%d.%d.%d LIP:%d. %d.%d.%d GID:%d LID:%d\n", 
							(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
							(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
							tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
						_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
						if(tempPtr->inboundIdInitiated)//Not empty
							tempPtr->outboundNext = NULL;
						else {
							tempPtr->inboundNext = freePptpGreEntryHeadPtr;
							freePptpGreEntryHeadPtr = tempPtr;
						}
#endif
						if(pptpPassthroughServerInboundMsg) printk("PPTP-PTH:(Stop Control Connection) Outbound ID GID: %d LID: %d Closed\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
						tempPtr->outboundIdClosed = 1;
						tempPtr->controlIdleTimer = 0; /* Reset control Idle timer*/
					}
				} while(tempPtr);
			}			
			break;
		case PPTP_WanErrorNotify:// Send by PAC to PNS to indicate WAN error condition(??)
		case PPTP_SetLinkInfo:// Send by PNS to PAC to set PPP-negotiated options.		

			if(!software)
			{
				//Get corresponding inbound session
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {

						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->inboundGlobalId == ntohs(cptr->cid1))						
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
				
				if(tempPtr)
				{
					if(tempPtr->outboundIdClosed == 1 && pptpPassthroughServerInboundMsg)
						printk("PPTP-PTH: Closed Inbound ID GID:%d LID:%d packet keep send Call Clear Request\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					cptr->cid1=htons(tempPtr->inboundLocalId);
					tempPtr->controlIdleTimer = 0; /* Reset control Idle timer*/
				}					
			}


			break;
		case PPTP_InCallRequest:
		case PPTP_InCallReply:
		case PPTP_InCallConn: //Send by PAC to PNS in response to a received Incoming-Call-Reply						
			//printk("PPTP-PTH: Current not supported control message type: %x\n", ctl_type);
		default:
			return SUCCESS;
	}

	return SUCCESS;
}

/*
@func int32	| rtl8651_l4AliasHandlePptpClientInbound	|  Analysis inbound traffic to get insideGlobalId and update created PPTP GRE entry
@parm struct rtl_pktHdr *	| pkt		| Packet header of the arrived packet
@parm struct ip *			| pip		| IP header of the arrived packet
@parm struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *	| link		| Session to store the ALG control channel entry
@comm 
	
*/
int32 rtl8651_l4AliasHandlePptpClientInbound(struct rtl_pktHdr *pkt, struct ip *pip, 
					   struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s * link)	
{
	_rtl8651_pptpGreEntry_t * tempPtr=NULL, *trackPtr;
	PptpCallId cptr;
	uint16    ctl_type;	/* control message type */
	uint32 i, entryIdx;
	PptpCode  codes;
	uint8 software=0;
	static uint32 lastId=0;
	
	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;
	if((alg_supports[pkt->ph_dsid]&RTL8651_ALG_PPTP)==0) return FWDENG_USER_DROP;

	/* Verify valid PPTP control message  */
	if ((cptr = _rtl8651_aliasVerifyPptp(pip, &ctl_type)) == NULL)
		return SUCCESS;
	
	switch (ctl_type)
	{
		case PPTP_OutCallReply:
			if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH PKT: Client Inbound OutCallReply\n");
			//Get corresponding inbound session
			if(!software)
			{
				tempPtr = pptpGreInboundHeadPtr[ntohs(cptr->cid2) &((1<<RTL8651_ICMPTBL_BITS)-1)];
				while(tempPtr) {
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->inboundIdInitiated == 1 && tempPtr->inboundGlobalId == ntohs(cptr->cid2))
						break;
					tempPtr = tempPtr->inboundNext;
				}
			}
			else
			{
				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->inboundIdInitiated == 1 && tempPtr->inboundGlobalId == ntohs(cptr->cid2))
						break;
				}
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
			}

			
			if(tempPtr) {
				uint32 tblIdx=0;

				codes = (PptpCode) (cptr + 1);
				if(codes->resCode == 1) {
				
					cptr->cid2 = htons(tempPtr->inboundLocalId);
					if(tempPtr->outboundIdInitiated == 0) 
					{					
						if(pptpPassthroughClientInboundMsg)
							printk("PPTP-PH: Server from WAN OutCallReply SIP: %d.%d.%d.%d DIP: %d.%d.%d.%d ID %d\n",
								(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff,
								(tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
								(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff,
								(tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,
								tempPtr->inboundGlobalId);
						tempPtr->outboundGlobalId = ntohs(cptr->cid1);
						lastId = ((lastId +1)*(tempPtr->outboundGlobalId+1))&0xFFFF;
						if(!software)
						{
							uint32 localId, targetHashIdx, hashIdx, tempTblIdx;
							tblIdx = tempPtr->outboundGlobalId & ((1<<RTL8651_ICMPTBL_BITS)-1);
							targetHashIdx = tempPtr->outboundGlobalId & 0x3FF;
							for(localId = lastId; localId!=((lastId+65535)&0xFFFF); localId=(localId+1)&0xFFFF) {
								hashIdx = rtl8651_naptIcmpTableIndex(tempPtr->insideLocalIpAddr, localId, tempPtr->outsideGlobalIpAddr, &tempTblIdx);
								if(hashIdx == targetHashIdx && tempTblIdx == tblIdx) {
									trackPtr =  pptpGreOutboundHeadPtr[tblIdx];
									while(trackPtr) {
										if(trackPtr->outboundIdInitiated == 1 && trackPtr->outboundLocalId == localId)
											break;
										trackPtr = trackPtr->outboundNext;
									}
									if(trackPtr == NULL)//There does not exist same local ID
										break;
								}
							}
							lastId = localId;
							tempPtr->outboundLocalId =  localId;
							cptr->cid1 = htons(tempPtr->outboundLocalId);
							tempPtr->outboundIdInitiated = 1;
							tempPtr->inboundIdClosed = 0;
							tempPtr->controlIdleTimer = 0;
							if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH: Add outbound from GID:%d to LID:%d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
							_rtl8651_addAsicPptpPassthroughOutboundEntry(tempPtr);
						}
						else //software
						{

							tempPtr->outboundLocalId =  tempPtr->outboundGlobalId;							
							tempPtr->outboundIdInitiated = 1;
						}

//						printk("PPTP-PTH: Add outbound GID:%x LID:%x\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
//						rtl8651_l4AliasPptpDump();

						if(!software)
						{
							tempPtr->outboundNext = pptpGreOutboundHeadPtr[tblIdx];
							pptpGreOutboundHeadPtr[tblIdx] = tempPtr;
						}
					}
					else {
						if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH: Established outbound from GID:%d to LID:%d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
						cptr->cid1 = htons(tempPtr->outboundLocalId);
						tempPtr->inboundIdClosed = 0;
						tempPtr->controlIdleTimer = 0;
					}
				}
				else 
				{//Session not created, drop the previous created entry

					if(software)
					{
						memset(tempPtr,0,sizeof(_rtl8651_pptpGreEntry_t));						
					}
					else //hardware
					{
						//Get corresponding inbound session
						for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
							entryIdx = i;
							tempPtr = pptpGreInboundHeadPtr[i];
							while(tempPtr) {
								if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
									tempPtr->inboundGlobalId == ntohs(cptr->cid2))
									break;
								tempPtr = tempPtr->inboundNext;
							}
						}
						if(tempPtr) 
						{
#if 0						
							_rtl8651_pptpGreEntry_t * trackPtr;
							if(pptpGreInboundHeadPtr[entryIdx] == tempPtr) {
								pptpGreInboundHeadPtr[entryIdx] = tempPtr->inboundNext;
							} else {
								trackPtr = pptpGreInboundHeadPtr[entryIdx];
								while(trackPtr && trackPtr->inboundNext != tempPtr)
									trackPtr = trackPtr->inboundNext;
								assert(trackPtr);//trackPtr->inboundNext must point to the
								trackPtr->inboundNext = tempPtr->inboundNext;
							}
							tempPtr->inboundIdInitiated = 0;
							//printk("PPTP-PTH: Not create successful Del inbound GID:%x LID:%x\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
							_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
							if(tempPtr->outboundIdInitiated)//Not empty
								tempPtr->inboundNext = NULL;
							else {
								tempPtr->inboundNext = freePptpGreEntryHeadPtr;
								freePptpGreEntryHeadPtr = tempPtr;
							}
#endif
							tempPtr->inboundIdClosed = 1;
							tempPtr->controlIdleTimer = 0;
						}
					}
				}
			}
			break;
		case PPTP_CallDiscNotify:	   //Send by PAC to PNS when call is disconnected
			if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH PKT: Client Inbound Call Disconnect Notify\n");
			if(!software)
			{
				//Get corresponding inbound session
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->outboundGlobalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
				if(tempPtr) {
#if 0
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreOutboundHeadPtr[entryIdx] == tempPtr) {
						pptpGreOutboundHeadPtr[entryIdx] = tempPtr->outboundNext;
					} else {
						trackPtr = pptpGreOutboundHeadPtr[entryIdx];
						while(trackPtr && trackPtr->outboundNext != tempPtr)
							trackPtr = trackPtr->outboundNext;
						assert(trackPtr);//trackPtr->inboundNext must point to the
						trackPtr->outboundNext = tempPtr->outboundNext;
					}
					tempPtr->outboundIdInitiated = 0;
					//printk("PPTP-PTH: Del outbound GIP:%d.%d.%d.%d LIP:%d. %d.%d.%d GID:%x LID:%x\n", 
					//	(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
					//	(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
					//	tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
					_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
					if(tempPtr->inboundIdInitiated)//Not empty
						tempPtr->outboundNext = NULL;
					else {
						tempPtr->inboundNext = freePptpGreEntryHeadPtr;
						freePptpGreEntryHeadPtr = tempPtr;
					}
#endif
					tempPtr->outboundIdClosed = 1;
					tempPtr->controlIdleTimer = 0;
					cptr->cid1 = htons(tempPtr->outboundLocalId);
				}
			}
			else
			{

			printk("PPTP_CallDiscNotify.....client in\n");
				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->outboundLocalId == ntohs(cptr->cid1))
						break;
				}

				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				if(tempPtr)
				{
				printk("kill this entry..........link->dstIpAddr=%x link->insideGlobalIpAddr=%x localid=%x\n", link->dstIpAddr, link->insideGlobalIpAddr ,ntohs(cptr->cid1));
					tempPtr->inboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
					tempPtr->outboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
				}				
			}
			break;
		case PPTP_StopCtrlConnRequest:
		case PPTP_StopCtrlConnReply:
			if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH PKT: Client Inbound Stop Control Connection\n");
			if(!software)
			{
				do {
					for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
						entryIdx = i;
						tempPtr = pptpGreOutboundHeadPtr[i];
						while(tempPtr) {
							if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
								tempPtr->outboundIdClosed == 0)
								break;
							tempPtr = tempPtr->inboundNext;
						}
					}
					if(tempPtr) {
						tempPtr->outboundIdClosed = 1;
						tempPtr->controlIdleTimer = 0;
						cptr->cid1 = htons(tempPtr->outboundLocalId);
					}
				} while (tempPtr);
			}
			break;
		case PPTP_WanErrorNotify:// Send by PAC to PNS to indicate WAN error condition(??)
		case PPTP_SetLinkInfo:// Send by PNS to PAC to set PPP-negotiated options.		
			if(pptpPassthroughClientInboundMsg) printk("PPTP-PTH PKT: Client Inbound SetLinkInfo\n");
			//modify Global Peer's Call ID to Local Peer's Call ID
			if(!software)
			{
				//Get corresponding inbound session
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->inboundGlobalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
				tempPtr->controlIdleTimer = 0;
				if(tempPtr) cptr->cid1=htons(tempPtr->inboundLocalId);
			}
			break;			

		case PPTP_InCallRequest:
		case PPTP_InCallReply:
		case PPTP_InCallConn: //Send by PAC to PNS in response to a received Incoming-Call-Reply
			//printk("PPTP-PTH: Current not supported control message type: %x\n", ctl_type);
		default:
			return SUCCESS;
	}

	return SUCCESS;
}



int32 rtl8651_l4AliasHandlePptpServerOutbound(
		struct rtl_pktHdr *pkt,
		struct ip *pip,
		struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link){
		
	_rtl8651_pptpGreEntry_t * tempPtr=NULL, *trackPtr;
	PptpCallId cptr;
	uint16    ctl_type;	/* control message type */
	uint32 i, entryIdx , tblIdx;
	PptpCode  codes;
	uint8 software=0;
	static uint32 searchIdx = 0;

	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;

	if((alg_supports[pkt->ph_dsid]&RTL8651_ALG_PPTP_SERVER)==0) return FWDENG_USER_DROP;
		

	/* Verify valid PPTP control message  */
	if ((cptr = _rtl8651_aliasVerifyPptp(pip, &ctl_type)) == NULL)
		return SUCCESS;
	
	switch (ctl_type)
	{
		case PPTP_OutCallReply:
			if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH PKT: Server Outbound OutCallReply\n");
			if(!software)
			{
				//Get corresponding inbound session
				//tempPtr = pptpGreOutboundHeadPtr[ntohs(cptr->cid2) &((1<<RTL8651_ICMPTBL_BITS)-1)];
				rtl8651_naptIcmpTableIndex(link->insideLocalIpAddr, ntohs(cptr->cid2), link->dstIpAddr, &tblIdx);
				tempPtr = pptpGreOutboundHeadPtr[tblIdx];			
						
				
/*				printk("pptp server out idx=%d cid2=%x\n",tblIdx,ntohs(cptr->cid2)); */
				while(tempPtr) {
/*					printk("pptp server out tempPtr->outsideGlobalIpAddr(%x) == link->dstIpAddr(%x) && tempPtr->insideGlobalIpAddr(%x) == link->insideGlobalIpAddr(%x) && tempPtr->outboundLocalId(%x) == ntohs(cptr->cid2)(%x))\n",	tempPtr->outsideGlobalIpAddr,link->dstIpAddr, tempPtr->insideGlobalIpAddr,link->insideGlobalIpAddr,tempPtr->outboundLocalId,ntohs(cptr->cid2));					 */
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->outboundIdInitiated == 1 && tempPtr->outboundLocalId == ntohs(cptr->cid2))
						break;
					tempPtr = tempPtr->outboundNext;
				}
			}
			else
			{
				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->outboundIdInitiated == 1 && tempPtr->outboundLocalId == ntohs(cptr->cid2))
						break;
				}
				if(i==algMaxSession.pptpMaxSession) tempPtr=NULL;
				
			}
			
			if(tempPtr) {
				uint32 tblIdx;

				if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH: Exist outbound GID:%d LID:%d inboundInitiated %d\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId, tempPtr->inboundIdInitiated);
				codes = (PptpCode) (cptr + 1);
				if(codes->resCode == 1) {
			
					cptr->cid2 = htons(tempPtr->outboundGlobalId);


					if(tempPtr->inboundIdInitiated == 0) 
					{
						tempPtr->inboundLocalId = ntohs(cptr->cid1);
						if(pptpPassthroughServerOutboundMsg)
							printk("PPTP-PH: Server from LAN reply DIP:%d.%d.%d.%d SIP:%d.%d.%d.%d ID:%d\n",
								(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff,
								(tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
								(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff,
								(tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff,
								tempPtr->inboundLocalId);
						if(!software)
						{
//							uint32 globalId, targetHashIdx, hashIdx, tempTblIdx;							

							tempPtr->inboundGlobalId = rtl8651_naptIcmpTableIndex(tempPtr->insideLocalIpAddr, tempPtr->inboundLocalId, tempPtr->outsideGlobalIpAddr, &tblIdx);
							
							if(pptpGreInboundHeadPtr[tblIdx]) {//Must find the free one only when list not empty 
								uint32 offsetMask[2], i;
								
								trackPtr = pptpGreInboundHeadPtr[tblIdx];
								offsetMask[0] = offsetMask[1] = 0;
								while(trackPtr) {
									assert(trackPtr->inboundIdInitiated);
									offsetMask[(trackPtr->inboundGlobalId>>10)&0x1] |= (1<<((trackPtr->inboundGlobalId>>11)&0x1F));
									trackPtr = trackPtr->inboundNext;
								}
								if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH: Server Inbound try ID offset start at %d\n", searchIdx);
								for(i=searchIdx; i!=((searchIdx+63)&0x3F); i=(i+1)&0x3f) {
									if((offsetMask[i&0x1] & (1<<((i>>1)&0x1F))) == 0)
										break;
								}
								searchIdx = (i+1)&0x3F;
								assert(i<64);
								tempPtr->inboundGlobalId |= (i<<10);
							}

							cptr->cid1 = htons(tempPtr->inboundGlobalId);
							if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH: Modify inbound from LID:%d to GID:%d inseart to table idx: %d\n", tempPtr->inboundLocalId, tempPtr->inboundGlobalId, tblIdx); 
							_rtl8651_addAsicPptpPassthroughInboundEntry(tempPtr);
							tempPtr->inboundNext = pptpGreInboundHeadPtr[tblIdx];
							pptpGreInboundHeadPtr[tblIdx] = tempPtr;
							tempPtr->inboundIdInitiated = 1;
							tempPtr->inboundIdClosed = 0;
							tempPtr->controlIdleTimer = 0;
						}
						else {
							tempPtr->inboundGlobalId =  tempPtr->inboundLocalId;
							tempPtr->inboundIdInitiated = 1;
						}
					}
					else {/* inbound Id Initialized and modify ID and corresponding flag*/
						cptr->cid1 = htons(tempPtr->inboundGlobalId);
						tempPtr->inboundIdClosed = 0;
						tempPtr->controlIdleTimer = 0;
					}
				}
				else {//Session not created, drop the previous created entry

					if(software)
					{
						memset(tempPtr,0,sizeof(_rtl8651_pptpGreEntry_t));						
					}
					else //hardware
					{
						//Get corresponding inbound session
						for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
							entryIdx = i;
							tempPtr = pptpGreOutboundHeadPtr[i];
							while(tempPtr) {
								if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
									tempPtr->outboundGlobalId == ntohs(cptr->cid2))
									break;
								tempPtr = tempPtr->outboundNext;
							}
						}
					
					
						if(tempPtr) {
#if 0
							_rtl8651_pptpGreEntry_t * trackPtr;
							if(pptpGreOutboundHeadPtr[entryIdx] == tempPtr) {
								pptpGreOutboundHeadPtr[entryIdx] = tempPtr->outboundNext;
							} else {
								trackPtr = pptpGreOutboundHeadPtr[entryIdx];
								while(trackPtr && trackPtr->outboundNext != tempPtr)
									trackPtr = trackPtr->outboundNext;
								assert(trackPtr);//trackPtr->inboundNext must point to the
								trackPtr->outboundNext = tempPtr->outboundNext;
							}
							tempPtr->outboundIdInitiated = 0;
/*							printk("PPTP-PTH: Not create successful Del outbound GID:%x LID:%x\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId); */
							_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
							if(tempPtr->inboundIdInitiated)//Not empty
								tempPtr->outboundNext = NULL;
							else {
								tempPtr->inboundNext = freePptpGreEntryHeadPtr;
								freePptpGreEntryHeadPtr = tempPtr;
							}
#endif							
							tempPtr->outboundIdClosed = 1;
							tempPtr->controlIdleTimer = 0;
						}
					}
				}
			}
			break;
		case PPTP_CallDiscNotify:	   //Send by PAC to PNS when call is disconnected
			if(pptpPassthroughServerOutboundMsg)  printk("PPTP-PTH PKT: Server Outbound CallDiscNotify\n");
			if(!software)
			{
				//Get corresponding inbound session
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreInboundHeadPtr[i];
					while(tempPtr) {
						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->inboundLocalId == ntohs(cptr->cid1))
							break;
						tempPtr = tempPtr->outboundNext;
					}
				}
				if(tempPtr) {
#if 0					
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreInboundHeadPtr[entryIdx] == tempPtr) {
						pptpGreInboundHeadPtr[entryIdx] = tempPtr->inboundNext;
					} else {
						trackPtr = pptpGreInboundHeadPtr[entryIdx];
						while(trackPtr && trackPtr->inboundNext != tempPtr)
							trackPtr = trackPtr->inboundNext;
						assert(trackPtr);//trackPtr->inboundNext must point to the
						trackPtr->inboundNext = tempPtr->inboundNext;
					}
					tempPtr->inboundIdInitiated = 0;
					printk("PPTP-PTH: Del inbound GIP:%d.%d.%d.%d LIP:%d.%d.%d.%d GID:%d LID:%d\n", 
						(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
						(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
						tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
					if(tempPtr->outboundIdInitiated)//Not empty
						tempPtr->inboundNext = NULL;
					else {
						tempPtr->inboundNext = freePptpGreEntryHeadPtr;
						freePptpGreEntryHeadPtr = tempPtr;
					}
#endif
					if(tempPtr->inboundIdClosed == 1 && pptpPassthroughServerOutboundMsg)
						printk("PPTP-PTH:(Call Disconnect Notify) Inbound closed session GID: %d LID: %d\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					else if (pptpPassthroughServerOutboundMsg)
						printk("PPTP-PTH:(Call Disconnect Notify) Inbound session GID: %d LID: %d closed\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					tempPtr->inboundIdClosed = 1;
					tempPtr->controlIdleTimer = 0;
					
					cptr->cid1 = htons(tempPtr->inboundGlobalId);
				}
			}
			else
			{
				tempPtr=softwarePptpGreEntryHeadPtr;
				for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
				{
					if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
						tempPtr->inboundLocalId == ntohs(cptr->cid1))
					break;
				}
				
				if(tempPtr)
				{
					tempPtr->inboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
					tempPtr->outboundIdleTimer= pptpPassthroughSysPara.idelTimeout-3;
				}
			}
			break;
		case PPTP_StopCtrlConnRequest:
		case PPTP_StopCtrlConnReply:
			if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH PKT: Server Outbound StopCtrlConnRequest and StopCtrlConnReply\n");
			if(!software)
			{
				do {
					for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
						entryIdx = i;
						tempPtr = pptpGreInboundHeadPtr[i];
						while(tempPtr) {
							if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && tempPtr->inboundIdClosed == 0)
								break;
							tempPtr = tempPtr->outboundNext;
						}
					}
					if(tempPtr) {
#if 0						
						_rtl8651_pptpGreEntry_t * trackPtr;
						if(pptpGreInboundHeadPtr[entryIdx] == tempPtr) {
							pptpGreInboundHeadPtr[entryIdx] = tempPtr->inboundNext;
						} else {
							trackPtr = pptpGreInboundHeadPtr[entryIdx];
							while(trackPtr && trackPtr->inboundNext != tempPtr)
								trackPtr = trackPtr->inboundNext;
							assert(trackPtr);//trackPtr->inboundNext must point to the
							trackPtr->inboundNext = tempPtr->inboundNext;
						}
						tempPtr->inboundIdInitiated = 0;
						printk("PPTP-PTH: Del inbound GIP:%d.%d.%d.%d LIP:%d.%d.%d.%d GID:%d LID:%d\n", 
							(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
							(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
							tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
						_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
						if(tempPtr->outboundIdInitiated)//Not empty
							tempPtr->inboundNext = NULL;
						else {
							tempPtr->inboundNext = freePptpGreEntryHeadPtr;
							freePptpGreEntryHeadPtr = tempPtr;
						}
#endif
						if(pptpPassthroughServerOutboundMsg) printk("PPTP-PTH:(Stop Control Connection) Inbound session GID: %d LID: %d closed\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
						tempPtr->inboundIdClosed = 1;
						tempPtr->controlIdleTimer = 0;
					}
				} while(tempPtr);
			}
			break;
		case PPTP_WanErrorNotify:// Send by PAC to PNS to indicate WAN error condition(??)
		case PPTP_SetLinkInfo:// Send by PNS to PAC to set PPP-negotiated options.			
			if(!software)
			{
				//Get corresponding inbound session
				for(i=0, tempPtr = NULL; tempPtr == NULL &&  i < RTL8651_ICMPTBL_SIZE; i++) {
					entryIdx = i;
					tempPtr = pptpGreOutboundHeadPtr[i];
					while(tempPtr) {

						if(tempPtr->outsideGlobalIpAddr == link->dstIpAddr && tempPtr->insideGlobalIpAddr == link->insideGlobalIpAddr && 
							tempPtr->outboundLocalId == ntohs(cptr->cid1))						
							break;
						tempPtr = tempPtr->inboundNext;
					}
				}
				
				if(tempPtr)
				{
					cptr->cid1=htons(tempPtr->outboundGlobalId);
				}					
			}
			break;
			
		case PPTP_InCallRequest:
		case PPTP_InCallReply:
		case PPTP_InCallConn: //Send by PAC to PNS in response to a received Incoming-Call-Reply			
			//printk("PPTP-PTH: Current not supported control message type: %x\n", ctl_type);
		default:
			return SUCCESS;
	}

	return SUCCESS;
}

void rtl8651_l4AliasPptpReInit(void)
{
	reInit=1;
	pptpPassthroughSysPara.idelTimeout=0;
	rtl8651_l4AliasPptpHouseKeeping();
	pptpPassthroughSysPara.idelTimeout = 1800; // 30-min
}


/*
@func int32	| rtl8651_l4AliasPptpInit	|  Initiate PPTP pass-through.
@parm void *	| arg		| Meaningless
@comm 
	The memory allocated for PPTP pass-through is defined in global variable algMaxSession member pptpMaxSession.
*/

static int algPptpInited=0;
int32 rtl8651_l4AliasPptpInit(void *arg){
	_rtl8651_pptpGreEntry_t * pptpPassThroughEntryPtr;
	uint32 i;

	if(algPptpInited) return SUCCESS;

	algPptpInited=1;
	for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
		pptpGreInboundHeadPtr[i] = NULL;
		pptpGreOutboundHeadPtr[i] = NULL;
	}

	freePptpGreEntryHeadPtr = NULL;

	for(i=0; i<algMaxSession.pptpMaxSession; i++) {
		pptpPassThroughEntryPtr = (_rtl8651_pptpGreEntry_t *) rtlglue_malloc(sizeof(_rtl8651_pptpGreEntry_t));
		
		if(pptpPassThroughEntryPtr) {
			pptpPassThroughEntryPtr->inboundNext = freePptpGreEntryHeadPtr;
			freePptpGreEntryHeadPtr = pptpPassThroughEntryPtr;
		}
	}

	if(softwarePptpGreEntryHeadPtr!=NULL) rtlglue_free(softwarePptpGreEntryHeadPtr);
	softwarePptpGreEntryHeadPtr = (_rtl8651_pptpGreEntry_t *) rtlglue_malloc(sizeof(_rtl8651_pptpGreEntry_t)*algMaxSession.pptpMaxSession);
	memset(softwarePptpGreEntryHeadPtr,0,sizeof(_rtl8651_pptpGreEntry_t)*algMaxSession.pptpMaxSession);
	
	pptpPassthroughSysPara.heavyLoadSwappingThreshold = 100;
	pptpPassthroughSysPara.idelTimeout = 1800; // 30-min
	rtl8651_l4GreNaptAliasIn = _rtl8651_l4GreNaptAliasIn;
	rtl8651_l4GreNaptAliasOut = _rtl8651_l4GreNaptAliasOut;
	return SUCCESS;
}


/*********************************************************************************
* Routine Name :  rtl8651_l4AliasPptpHouseKeeping                                   
* Description :                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
void  rtl8651_l4AliasPptpHouseKeeping(){ 
	_rtl8651_pptpGreEntry_t * trackEntryPtr, * tempPtr, *nextPtr;
	uint32 i,j;
	uint8 software=0;

	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;

	//software house keeping
	//if(tempPtr->outboundIdleTimer >= pptpPassthroughSysPara.idelTimeout) 
	if(software)
	{
		tempPtr=softwarePptpGreEntryHeadPtr;
		for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
		{
			if((tempPtr->inboundIdInitiated==1)&&(tempPtr->outboundIdInitiated==1))
			{				
				tempPtr->inboundIdleTimer++;
				tempPtr->outboundIdleTimer++;
				if((tempPtr->inboundIdleTimer >= pptpPassthroughSysPara.idelTimeout) &&
					(tempPtr->outboundIdleTimer >= pptpPassthroughSysPara.idelTimeout))
				{
					memset(tempPtr,0,sizeof(_rtl8651_pptpGreEntry_t));					
				}

				for(j=0;j<RTL8651_MAX_DIALSESSION;j++)
				{
					if((alg_supports[j]&RTL8651_ALG_PPTP)==0)
					{
						if(tempPtr->dsid==j)
						{
							memset(tempPtr,0,sizeof(_rtl8651_pptpGreEntry_t));							
						}
					}
				}
				
			}
		}
	}
	else
	{
		//hardware house keeping 
		for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
			/* Handle GRE session timeout*/
			tempPtr = pptpGreInboundHeadPtr[i];
			while(tempPtr) {

				
				uint8 disabled=FALSE;
				assert(tempPtr->inboundIdInitiated);
				tempPtr->inboundPktCnt = 0;

				for(j=0;j<RTL8651_MAX_DIALSESSION;j++)
				{
					if((alg_supports[j]&RTL8651_ALG_PPTP)==0)
					{
						if(tempPtr->dsid==j)
						{
							disabled=TRUE;
							break;
						}
					}
				}

				if((tempPtr->inboundToAsic==1)&&(!disabled)&&(reInit==0))
                                       tempPtr = tempPtr->inboundNext;
                               else				
				{
					reInit=0;
					if(tempPtr->inboundToAsic==1)
					{
						//kill hardware entry
						tempPtr->inboundIdInitiated = 0;
						_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
					}
					
					tempPtr->inboundIdleTimer++;	
					if((tempPtr->inboundIdleTimer >= pptpPassthroughSysPara.idelTimeout)||(disabled)) {
						//printk("PPTP-PTH: Timeout remove inbound session GID:%x LID:%x\n", tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
						nextPtr = tempPtr->inboundNext;
						if(pptpGreInboundHeadPtr[i] == tempPtr) {
							pptpGreInboundHeadPtr[i] = tempPtr->inboundNext;
						} else {
							trackEntryPtr = pptpGreInboundHeadPtr[i];
							while(trackEntryPtr) {
								if(trackEntryPtr->inboundNext == tempPtr) {
									trackEntryPtr->inboundNext = tempPtr->inboundNext;
									break;
								}
								trackEntryPtr = trackEntryPtr->inboundNext;
							}
						}
						tempPtr->inboundIdInitiated = 0;
						if(tempPtr->outboundIdInitiated == 0) {//Not linked to any list, move it to free list
							tempPtr->inboundNext = freePptpGreEntryHeadPtr;
							freePptpGreEntryHeadPtr = tempPtr;
						}
						tempPtr = nextPtr;
					} 
					else 
						tempPtr = tempPtr->inboundNext;
				}
			}
			/* Handle outbound GRE session timeout */
			tempPtr = pptpGreOutboundHeadPtr[i];
			while(tempPtr) {
				uint8 disabled=FALSE;
				assert(tempPtr->outboundIdInitiated);
				tempPtr->outboundPktCnt = 0;

				for(j=0;j<RTL8651_MAX_DIALSESSION;j++)
				{
					if((alg_supports[j]&RTL8651_ALG_PPTP)==0)
					{
						if(tempPtr->dsid==j)
						{
							disabled=TRUE;
							break;
						}
					}
				}				
				
			      if((tempPtr->outboundToAsic == 1) &&(!disabled)&&(reInit==0))
                                     tempPtr = tempPtr->outboundNext;
                              else 
				{
					reInit=0;
					if(tempPtr->outboundToAsic == 1)
					{
						//kill hardware entry
						tempPtr->outboundIdInitiated = 0;						
						_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
					}
					tempPtr->outboundIdleTimer++;					
					
					if((tempPtr->outboundIdleTimer >= pptpPassthroughSysPara.idelTimeout)||(disabled)) {
						//printk("PPTP-PTH: Timeout remove outbound session GID:%x LID:%x\n", tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
						nextPtr = tempPtr->outboundNext;
						if(pptpGreOutboundHeadPtr[i] == tempPtr) {
							pptpGreOutboundHeadPtr[i] = tempPtr->outboundNext;
						} else {
							trackEntryPtr = pptpGreOutboundHeadPtr[i];
							while(trackEntryPtr) {
								if(trackEntryPtr->outboundNext == tempPtr) {
									trackEntryPtr->outboundNext = tempPtr->outboundNext;
									break;
								}
								trackEntryPtr = trackEntryPtr->outboundNext;
							}
						}
						tempPtr->outboundIdInitiated = 0;
						if(tempPtr->inboundIdInitiated == 0) {//Not linked to any list, move it to free list
												
							tempPtr->inboundNext = freePptpGreEntryHeadPtr;
							freePptpGreEntryHeadPtr = tempPtr;
						}
						tempPtr = nextPtr;
					} 
					else 
						tempPtr = tempPtr->outboundNext;
				}
			}
			/* Handle inbound control session timeout */
			tempPtr = pptpGreInboundHeadPtr[i];
			while(tempPtr) {
				uint8 inboundNotOpen, outboundNotOpen;
				inboundNotOpen = (tempPtr->inboundIdClosed || !tempPtr->inboundIdInitiated )? 1: 0;
				outboundNotOpen = (tempPtr->outboundIdClosed  || !tempPtr->outboundIdInitiated )? 1: 0;
				if( inboundNotOpen || outboundNotOpen)
					tempPtr->controlIdleTimer++;
				if(	((inboundNotOpen && outboundNotOpen) && tempPtr->controlIdleTimer >= pptpPassthroughControlShortTimeout) ||
					((inboundNotOpen || outboundNotOpen) &&tempPtr->controlIdleTimer >= pptpPassthroughControlShortTimeout)) {
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreInboundHeadPtr[i] == tempPtr) {
						pptpGreInboundHeadPtr[i] = tempPtr->inboundNext;
					} else {
						trackPtr = pptpGreInboundHeadPtr[i];
						while(trackPtr && trackPtr->inboundNext != tempPtr)
							trackPtr = trackPtr->inboundNext;
						assert(trackPtr);//trackPtr->inboundNext must point to the
						trackPtr->inboundNext = tempPtr->inboundNext;
					}
					tempPtr->inboundIdInitiated = 0;
					if(pptpPassthroughClientInboundMsg ||  pptpPassthroughServerOutboundMsg)
						printk("PPTP-PTH: Del inbound GIP:%d.%d.%d.%d LIP:%d.%d.%d.%d GID:%d LID:%d\n", 
							(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
							(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
							tempPtr->inboundGlobalId, tempPtr->inboundLocalId);
					_rtl8651_delAsicPptpPassthroughInboundEntry(tempPtr);
					if(tempPtr->outboundIdInitiated)//Not empty
						tempPtr->inboundNext = NULL;
					else {
						tempPtr->inboundNext = freePptpGreEntryHeadPtr;
						freePptpGreEntryHeadPtr = tempPtr;
					}
					tempPtr = pptpGreInboundHeadPtr[i];/* Tracking from start of link list */
				}
				else 
					tempPtr = tempPtr->inboundNext;
			}
			/* Since add function will check whether any entry every configured to ASIC, just force write to avoid entry empty*/
			if(pptpGreInboundHeadPtr[i])
				_rtl8651_addAsicPptpPassthroughInboundEntry(pptpGreInboundHeadPtr[i]);
			/* Handle inbound control session timeout */
			tempPtr = pptpGreOutboundHeadPtr[i];
			while(tempPtr) {
				uint8 inboundNotOpen, outboundNotOpen;
				inboundNotOpen = (tempPtr->inboundIdClosed == 1 || tempPtr->inboundIdInitiated == 0)? 1: 0;
				outboundNotOpen = (tempPtr->outboundIdClosed == 1 || tempPtr->outboundIdInitiated == 0)? 1: 0;
				if( !tempPtr->inboundIdInitiated)
					tempPtr->controlIdleTimer++;
				/* Since aging operation is done in inbound list tracking, needless to check it*/
				if(	((inboundNotOpen && outboundNotOpen) && tempPtr->controlIdleTimer >= pptpPassthroughControlShortTimeout) ||
					((inboundNotOpen || outboundNotOpen) &&tempPtr->controlIdleTimer >= pptpPassthroughControlShortTimeout)) {
					_rtl8651_pptpGreEntry_t * trackPtr;
					if(pptpGreOutboundHeadPtr[i] == tempPtr) {
						pptpGreOutboundHeadPtr[i] = tempPtr->outboundNext;
					} else {
						trackPtr = pptpGreOutboundHeadPtr[i];
						while(trackPtr && trackPtr->outboundNext != tempPtr)
							trackPtr = trackPtr->outboundNext;
						assert(trackPtr);//trackPtr->inboundNext must point to the
						trackPtr->outboundNext = tempPtr->outboundNext;
					}
					tempPtr->outboundIdInitiated = 0;
					if(pptpPassthroughClientOutboundMsg || pptpPassthroughServerInboundMsg)
						printk("PPTP-PTH: Del outbound GIP:%d.%d.%d.%d LIP:%d. %d.%d.%d GID:%d LID:%d\n", 
							(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff, 
							(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff, 
							tempPtr->outboundGlobalId, tempPtr->outboundLocalId);
					_rtl8651_delAsicPptpPassthroughOutboundEntry(tempPtr);
					assert(tempPtr->inboundIdInitiated==0);/* Must be cleared at previous stage*/
					tempPtr->inboundNext = freePptpGreEntryHeadPtr;
					freePptpGreEntryHeadPtr = tempPtr;
					tempPtr = pptpGreOutboundHeadPtr[i];
				}
				else
					tempPtr = tempPtr->outboundNext;
			}
			/* Since add function will check whether any entry every configured to ASIC, just force write to avoid entry empty*/
			if(pptpGreOutboundHeadPtr[i])
				_rtl8651_addAsicPptpPassthroughInboundEntry(pptpGreOutboundHeadPtr[i]);

		}
	}

}

/* 3 value , -1 means not configure, 0 means disable, >1 means enable */
void rtl8651_l4AliasPptpMsg( int8 clientInbound, int8 clientOutbound, int8 serverInbound, int8 serverOutbound, int8 greInbound, int8 greOutbound) {
	if(clientInbound>=0)
		pptpPassthroughClientInboundMsg = clientInbound==0?0:1;
	if(clientOutbound>=0)
		pptpPassthroughClientOutboundMsg = clientOutbound==0?0:1;
	if(serverInbound>=0)
		pptpPassthroughServerInboundMsg = serverInbound==0?0:1;
	if(serverOutbound>=0)
		pptpPassthroughServerOutboundMsg = serverOutbound==0?0:1;
	if(greInbound>=0)
		pptpPassthroughGreInboundMsg = greInbound==0?0:1;
	if(greOutbound>=0)
		pptpPassthroughGreOutboundMsg = greOutbound==0?0:1;	
}

void rtl8651_l4AliasPptpDump(void) {
	_rtl8651_pptpGreEntry_t * tempPtr;
	uint32 i;
	uint8 software=0;

	if((alg_supports[0]&RTL8651_ALG_PPTP_SOFTWARE)!=0) software=1;

	if(software) {
		if(softwarePptpGreEntryHeadPtr!=NULL)
		{
			printk("PPTP pure software current session information\n");

			tempPtr=softwarePptpGreEntryHeadPtr;
			for(i=0;i<algMaxSession.pptpMaxSession;i++,tempPtr++)
			{
				if(tempPtr->outboundIdInitiated&&tempPtr->inboundIdInitiated)
				{
					printk("outsideGlobal:%d.%d.%d.%d:%d insideGlobal:%d.%d.%d.%d:%d  outsideLocal:%d insideLocal %d.%d.%d.%d:%d Age:%u inPkt:%u outPkt:%u\n",
						(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,tempPtr->outboundGlobalId
						,(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,tempPtr->inboundGlobalId,tempPtr->outboundLocalId
						,(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff,tempPtr->inboundLocalId
						, tempPtr->inboundIdleTimer, tempPtr->inboundPktCnt, tempPtr->outboundPktCnt);
				}
			}	
		}
	}
	else {
		printk("PPTP Passthrough Software session information\n");
		printk("Inbound session\n");
		for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
			tempPtr = pptpGreInboundHeadPtr[i];
			if(tempPtr)
				printk("[%02d]:\n", i);
			while(tempPtr) {
				if(!tempPtr->inboundToAsic)
					printk("Src %d.%d.%d.%d Dst %d.%d.%d.%d GID:%d LID:%d %s\n\tAge:%u Pkt:%u Inbound:%c%c Outbound:%c%c CAge: %u\n",
						(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
						(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,
						tempPtr->inboundGlobalId, tempPtr->inboundLocalId, 
						tempPtr->inboundToAsic?"ASIC":"", tempPtr->inboundIdleTimer, tempPtr->inboundPktCnt,
						tempPtr->inboundIdInitiated?'I':' ', tempPtr->inboundIdClosed?'C':' ', 
						tempPtr->outboundIdInitiated?'I':' ', tempPtr->outboundIdClosed?'C':' ', 
						tempPtr->controlIdleTimer);
				tempPtr = tempPtr->inboundNext;
			}
		}
		printk("Outbound session\n");
		for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
			tempPtr = pptpGreOutboundHeadPtr[i];
			if(tempPtr)
				printk("[%02d]:\n", i);
			while(tempPtr) {
				if(!tempPtr->inboundToAsic)
					printk("Src %d.%d.%d.%d Dst %d.%d.%d.%d LID %d GID %d %s\n\tAge:%u Pkt:%u Inbound:%c%c Outbound:%c%cCAge: %u\n",
						(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff,
						(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
						tempPtr->outboundLocalId, tempPtr->outboundGlobalId, 
						tempPtr->outboundToAsic?"ASIC":"", tempPtr->outboundIdleTimer, tempPtr->outboundPktCnt,
						tempPtr->inboundIdInitiated?'I':' ', tempPtr->inboundIdClosed?'C':' ', 
						tempPtr->outboundIdInitiated?'I':' ', tempPtr->outboundIdClosed?'C':' ', 
						tempPtr->controlIdleTimer);
				tempPtr = tempPtr->outboundNext;
			}
		}
		printk("PPTP Pass-through current session information\n");
		printk("Inbound session\n");
		for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
			tempPtr = pptpGreInboundHeadPtr[i];
			if(tempPtr)
				printk("[%02d]:\n", i);
			while(tempPtr) {
				if(tempPtr->inboundToAsic)
					printk("Src %d.%d.%d.%d Dst %d.%d.%d.%d GID:%d LID:%d %s\n\tAge:%u Pkt:%u Inbound:%c%c Outbound:%c%c CAge: %u\n",
						(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
						(tempPtr->insideGlobalIpAddr>>24)&0xff, (tempPtr->insideGlobalIpAddr>>16)&0xff, (tempPtr->insideGlobalIpAddr>>8)&0xff, tempPtr->insideGlobalIpAddr&0xff,
						tempPtr->inboundGlobalId, tempPtr->inboundLocalId, 
						tempPtr->inboundToAsic?"ASIC":"", tempPtr->inboundIdleTimer, tempPtr->inboundPktCnt,
						tempPtr->inboundIdInitiated?'I':' ', tempPtr->inboundIdClosed?'C':' ', 
						tempPtr->outboundIdInitiated?'I':' ', tempPtr->outboundIdClosed?'C':' ', 
						tempPtr->controlIdleTimer);

				tempPtr = tempPtr->inboundNext;
			}
		}
		printk("Outbound session\n");
		for(i=0; i<RTL8651_ICMPTBL_SIZE; i++) {
			tempPtr = pptpGreOutboundHeadPtr[i];
			if(tempPtr)
				printk("[%02d]:\n", i);
			while(tempPtr) {
				if(tempPtr->inboundToAsic)
					printk("Src %d.%d.%d.%d Dst %d.%d.%d.%d LID %d GID %d %s\n\tAge:%u Pkt:%u Inbound:%c%c Outbound:%c%cCAge: %u\n",
						(tempPtr->insideLocalIpAddr>>24)&0xff, (tempPtr->insideLocalIpAddr>>16)&0xff, (tempPtr->insideLocalIpAddr>>8)&0xff, tempPtr->insideLocalIpAddr&0xff,
						(tempPtr->outsideGlobalIpAddr>>24)&0xff, (tempPtr->outsideGlobalIpAddr>>16)&0xff, (tempPtr->outsideGlobalIpAddr>>8)&0xff, tempPtr->outsideGlobalIpAddr&0xff,
						tempPtr->outboundLocalId, tempPtr->outboundGlobalId, 
						tempPtr->outboundToAsic?"ASIC":"", tempPtr->outboundIdleTimer, tempPtr->outboundPktCnt,
						tempPtr->inboundIdInitiated?'I':' ', tempPtr->inboundIdClosed?'C':' ', 
						tempPtr->outboundIdInitiated?'I':' ', tempPtr->outboundIdClosed?'C':' ', 
						tempPtr->controlIdleTimer);
				tempPtr = tempPtr->outboundNext;
			}
		}
		tempPtr = freePptpGreEntryHeadPtr;
		i = 0;
		while(tempPtr) {
			i++;
			tempPtr = tempPtr->inboundNext;
		}
		printk("Total %u free entry in the free list\n\n", i);
	}
}


