

/*
* Copyright c                  Realtek Semiconductor Corporation, 2004
* All rights reserved.
* 
* Program : RealAudio UDP data channel
* Abstract : 
* Creator : 
* Author : $
* $Id: rtl8651_alg_rtsp.c,v 1.1.1.1 2006/07/26 05:38:23 blue Exp $
* $Log: rtl8651_alg_rtsp.c,v $
* Revision 1.1.1.1  2006/07/26 05:38:23  blue
*
*
* Revision 1.1.1.1  2006/02/14 03:09:05  jason
* backup the source code from v1036
*
* Revision 1.1.1.1  2006/01/16 02:39:43  jason
* backup source,because now we directly use romedriver source code, i open a new directory;
*
* Revision 1.2  2005/11/22 04:20:46  shliu
* *: specify case-sensitive when calling mBuf_getBMjmpTable() and
*    mBuf_getBMjmpTable().
*
* Revision 1.1.1.1  2005/11/21 05:20:03  chenyl
* +: WPNT834
* 	- System provider	: Sercomm
* 	- Brand Name		: Netgear
* 	- 3rt Party		: Airgo Gen 3
* 	- SDK			: 066-3
* 	- ROMEDRV		: 3.6-3
*
*
* Revision 1.3  2005/08/25 04:01:35  chenyl
* *: rearrange code, add MUTEX lock/unlock for external functions, add `inline' property to some critical functions
*
* Revision 1.2  2004/10/04 09:09:48  chhuang
* *: bug fixed for rtsp
*
* Revision 1.1  2004/09/30 14:13:27  chhuang
* +:add rtsp
*
* Revision 1.1  2004/09/14 13:24:29  chhuang
* +: RTSP alg support
*
*
*/

#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwd.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_rtsp.h"

#define RTSP_NUM_SESSION		8
#define RTSP_IDLE_TIME			60

typedef struct rtsp_Entry_s
{
	uint32	sessionID;
	uint16	CstartPort;
	uint16	CendPort;
	uint16	SstartPort;
	uint16	SendPort;
	uint32 	idle_time;
} rtsp_Entry_t;

static uint16 patternBuf[256];
static rtsp_Entry_t	tblRtsp_[RTSP_NUM_SESSION];
static int8 pattern[] = "client_port=";
static int8 pattern1[] = "server_port=";
static int8 pattern2[] = "destination=";
static int8 sessionPattern[] = "Session: ";

extern rtl8651_tblDrv_naptTcpUdpFlowEntry_t *  
_rtl8651_findNaptConnectionByExtIpPort( int8 isTcp, ipaddr_t ExtIp, uint16 ExtPort,
                                        ipaddr_t remoteIp, uint16 remotePort ) ;


void _rtl8651_getRTSPPortRange(uint8 *cp, uint16 *startPort, uint16 *endPort, uint32 *SSID)
{
	int32 flag=0;

	if (SSID)
		*SSID = 0;
	else {	*startPort = *endPort = 0;	}
	while(*cp != '\r' && *cp != ';') {
		if (*cp == '-') {
			flag = 1;
			cp++;
			continue;
		}
		if (SSID) 
			*SSID = *SSID * 10 + (*cp-'0');
		else {
			if (flag == 0) 
				*startPort = *startPort * 10 + (*cp - '0');
			else *endPort = *endPort * 10 + (*cp - '0');
		}
		cp ++;
	}
}



int32 rtl8651_l4AliasHandleRTSPClientOutbound(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tb)
{
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t *napt_t;
	uint16 startPort, endPort;
	uint8 *rtsp_hdr, buf[128], *cp;
	struct tcphdr *tcph;
	uint32 sessionId;
	int32 pos, i, diff;
	

	rtsp_hdr = (uint8 *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2) + 20);
	if (rtsp_hdr[0]=='S' && rtsp_hdr[1]=='E' && rtsp_hdr[2]=='T' && rtsp_hdr[3]=='U' && rtsp_hdr[4]=='P')
		goto SETUP_;
	if (rtsp_hdr[0]=='T'&&rtsp_hdr[1]=='E'&&rtsp_hdr[2]=='A'&&rtsp_hdr[3]=='R'&&rtsp_hdr[4]=='D'&&rtsp_hdr[5]=='O'&&
	    rtsp_hdr[6]=='W'&&rtsp_hdr[7]=='N')
	    goto TEARDOWN_;
	return FAILED;

SETUP_:

	mBuf_getBMjmpTable(pattern, patternBuf, 12, TRUE);
	pos = mBuf_BMpatternMatch(pkthdrPtr->ph_mbuf, pkthdrPtr->ph_len, (uint8*)pattern, 12, patternBuf, TRUE);
	if (pos < 1)
		return pos;
	_rtl8651_getRTSPPortRange(&pkthdrPtr->ph_mbuf->m_data[pos+12], &startPort, &endPort, NULL);
	rtl8651_addUpnpMap(UPNP_PERSIST, 0, 0, tb->insideGlobalIpAddr, startPort, tb->insideLocalIpAddr, startPort);
	rtl8651_addUpnpMap(UPNP_PERSIST, 0, 0, tb->insideGlobalIpAddr, endPort, tb->insideLocalIpAddr, endPort);
	for(i=0; i<RTSP_NUM_SESSION; i++) {
		if (tblRtsp_[i].CstartPort == 0) {
			tblRtsp_[i].CstartPort 	= startPort;
			tblRtsp_[i].CendPort	= endPort;
			tblRtsp_[i].idle_time	= 0;
			break;
		}
	}

	/* Check if present 'Destination' parameter */
	mBuf_getBMjmpTable(pattern2, patternBuf, 12, TRUE);
	pos = mBuf_BMpatternMatch(pkthdrPtr->ph_mbuf, pkthdrPtr->ph_len, (uint8*)pattern2, 12, patternBuf, TRUE);
	if (pos < 1)
		return SUCCESS;
	cp = &pkthdrPtr->ph_mbuf->m_data[pos+12];
	/* Present destination parameter */
	sprintf(buf,"%u.%u.%u.%u;client_port=%u-%u", tb->insideGlobalIpAddr>>24, (tb->insideGlobalIpAddr&0x00ff0000)>>16,
			(tb->insideGlobalIpAddr&0x0000ff00)>>8, tb->insideGlobalIpAddr&0xff, startPort, endPort);
	diff = 0;
	while(buf[diff] != ';')
		diff++;
	while(*cp!=';' && *cp!='\r') {
		cp++; diff--;
	}
	rtl8651_memcpy(&pkthdrPtr->ph_mbuf->m_data[pos+12], buf, strlen(buf));
	cp = &pkthdrPtr->ph_mbuf->m_data[pos+12] + strlen(buf);
	*cp='\r'; cp++, *cp='\n'; cp++; *cp='\r'; cp++; *cp='\n';

	/* Modify packet length in mbuf */
	pkthdrPtr->ph_len += diff;
	pkthdrPtr->ph_mbuf->m_len += diff;

	tcph = (struct tcphdr *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
	/* handle seq modification */
	{
		int32 delta;
		tb->tcpAckModified = 1;
		delta = _rtl8651_l4GetDeltaSeqOut(pip, tcph, tb);
		_rtl8651_l4AddSeq(pip, tb, delta + diff);
	}

	pip->ip_len	=htons(ntohs(pip->ip_len)+diff);	
	/* handle checksum */
	tcph->th_sum 	= 0;
	pip->ip_sum 	= 0;				
	tcph->th_sum 	= _rtl8651_tcpChecksum(pip); 
	pip->ip_sum 	= _rtl8651_ipChecksum(pip);
	return SUCCESS;


TEARDOWN_:

	/* Get Session ID */
	mBuf_getBMjmpTable(sessionPattern, patternBuf, 9, TRUE);
	pos = mBuf_BMpatternMatch(pkthdrPtr->ph_mbuf, pkthdrPtr->ph_len, (uint8*)sessionPattern, 9, patternBuf, TRUE);
	if (pos < 1)
		return pos;
	_rtl8651_getRTSPPortRange(&pkthdrPtr->ph_mbuf->m_data[pos+9], NULL, NULL, &sessionId);

	for(i=0; i<RTSP_NUM_SESSION; i++) {
		if (tblRtsp_[i].sessionID == sessionId)
			break;
	}
	if (i == RTSP_NUM_SESSION)
		return FAILED; /* Entry not found */
	
	napt_t = _rtl8651_findNaptConnectionByExtIpPort(FALSE, tb->insideGlobalIpAddr, tblRtsp_[i].CstartPort, tb->dstIpAddr, tblRtsp_[i].SstartPort);
	if (napt_t)	_rtl8651_delNaptConnection(napt_t);
	napt_t = _rtl8651_findNaptConnectionByExtIpPort(FALSE, tb->insideGlobalIpAddr, tblRtsp_[i].CendPort, tb->dstIpAddr, tblRtsp_[i].SendPort);
	if (napt_t) _rtl8651_delNaptConnection(napt_t);

	startPort 	= tblRtsp_[i].CstartPort;
	endPort		= tblRtsp_[i].CendPort;
	bzero(&tblRtsp_[i], sizeof(struct rtsp_Entry_s));
	/* Decide if to remove uPnP Mapping */	
	for(i=0; i<RTSP_NUM_SESSION; i++) {
		if (tblRtsp_[i].CstartPort==startPort && tblRtsp_[i].CendPort==endPort)
			break;
	}
	if (i == RTSP_NUM_SESSION) {
		rtl8651_delUpnpMap(FALSE, 0, 0, tb->insideGlobalIpAddr, startPort, tb->insideLocalIpAddr, startPort);
		rtl8651_delUpnpMap(FALSE, 0, 0, tb->insideGlobalIpAddr, endPort, tb->insideLocalIpAddr, endPort);
	}
	return SUCCESS;
}



int32 rtl8651_l4AliasHandleRTSPClientInbound(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tb)
{
	uint16 startPort, endPort;
	uint32 sessionId;
	uint8 *http_hdr;
	int32 pos, i;

	http_hdr = (uint8 *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2)+20);
	if (http_hdr[0]!='R'&&http_hdr[1]!='T'&&http_hdr[2]!='S'&&http_hdr[3]!='P'&&http_hdr[4]!='/'&&
		http_hdr[9]!='2'&&http_hdr[10]!='0'&&http_hdr[11]!='0')
		return FAILED;

	/* Get Client Port range */
	mBuf_getBMjmpTable(pattern, patternBuf, 12, TRUE);
	pos = mBuf_BMpatternMatch(pkthdrPtr->ph_mbuf, pkthdrPtr->ph_len, (uint8*)pattern, 12, patternBuf, TRUE);
	if (pos < 1)
		return FAILED;
	_rtl8651_getRTSPPortRange(&pkthdrPtr->ph_mbuf->m_data[pos+12], &startPort, &endPort, NULL);

	/* Find entry created by SETUP command */
	for(i=0; i<RTSP_NUM_SESSION; i++) {
		if (tblRtsp_[i].CstartPort==startPort && tblRtsp_[i].CendPort==endPort && tblRtsp_[i].sessionID==0)
			break;
	}
	if (i == RTSP_NUM_SESSION)
		return FAILED; /* Out of entry */

	/* Get Server Port range */
	mBuf_getBMjmpTable(pattern1, patternBuf, 12, TRUE);
	pos = mBuf_BMpatternMatch(pkthdrPtr->ph_mbuf, pkthdrPtr->ph_len, (uint8*)pattern1, 12, patternBuf, TRUE);
	if (pos < 1)
		return pos;
	_rtl8651_getRTSPPortRange(&pkthdrPtr->ph_mbuf->m_data[pos+12], &startPort, &endPort, NULL);

	/* Get Session ID */
	mBuf_getBMjmpTable(sessionPattern, patternBuf, 9, TRUE);
	pos = mBuf_BMpatternMatch(pkthdrPtr->ph_mbuf, pkthdrPtr->ph_len, (uint8*)sessionPattern, 9, patternBuf, TRUE);
	if (pos < 1)
		return pos;
	_rtl8651_getRTSPPortRange(&pkthdrPtr->ph_mbuf->m_data[pos+9], NULL, NULL, &sessionId);

	/* Complete entry */
	tblRtsp_[i].SendPort	= endPort;
	tblRtsp_[i].SstartPort	= startPort;
	tblRtsp_[i].sessionID	= sessionId;
	tblRtsp_[i].idle_time	= 0;
	return SUCCESS;
}


void rtl8651_l4AliasRtspHouseKeeping(void)
{
	int32 entry;

	for(entry=0; entry<RTSP_NUM_SESSION; entry++) {
		if (tblRtsp_[entry].CstartPort == 0 || tblRtsp_[entry].sessionID) 
			continue;
		tblRtsp_[entry].idle_time ++;
		if (tblRtsp_[entry].sessionID == 0 && tblRtsp_[entry].idle_time>RTSP_IDLE_TIME) {
			bzero(&tblRtsp_[entry], sizeof(struct rtsp_Entry_s));
		}
	}

}



int32 rtl8651_l4AliasRTSPInit(void *arg){
	bzero(tblRtsp_, sizeof(struct rtsp_Entry_s)*RTSP_NUM_SESSION);
	return SUCCESS;
}


