/*
* Copyright c                  Realtek Semiconductor Corporation, 2003  
* All rights reserved.
* 
* Program : SIP ALG (for single session)
* Abstract : 
* Creator : 
* Author : $
* $Id: rtl8651_alg_sip.c,v 1.1.1.1 2006/07/26 05:38:23 blue Exp $
* $Log: rtl8651_alg_sip.c,v $
* Revision 1.1.1.1  2006/07/26 05:38:23  blue
*
*
* Revision 1.1.1.1  2006/02/14 03:09:05  jason
* backup the source code from v1036
*
* Revision 1.1.1.1  2006/01/16 02:39:43  jason
* backup source,because now we directly use romedriver source code, i open a new directory;
*
* Revision 1.2  2005/11/22 15:50:54  chenyl
* *: Add mutex lock/unlock in log r/w
* *: new idle-timeout mechanism for VLAN / PPPOE / PPTP / L2TP wan types
* *: bug fix: PPTP passthrough
* *: bug fix: sip / ymsg typo error
* *: for ARP-reply module in RomeDrv, use newly allocated mbuf/pkt instead of original one ( To prevent from treated as packet forwarding instead of reply).
* *: minor coding style, sync with main trunk
*
* +: ioctl case for Mutex Lock/Unlock called by user level
* +: add 2 APIs to enable / disable
* 	1. Error messages from protocol stack to WAN : TCP RST, ICMP ICMP_UNREACH/ICMP_SOURCEQUENCH/ICMP_TIMXCEED/ICMP_PARAMPROB
* 	2. All packet from protocol stack to WAN
*    filtering to disable / enable idle-timer update triggered by these kinds of packets.
* +: TFTP ALG
* +: rtl8651_debug.h for debugging usage
* +: merge code about rtl_glue.c / re_core.c with Sercomm's version ( Airgo modification ?)
* +:
*
* Revision 1.1.1.1  2005/11/21 05:20:03  chenyl
* +: WPNT834
* 	- System provider	: Sercomm
* 	- Brand Name		: Netgear
* 	- 3rt Party		: Airgo Gen 3
* 	- SDK			: 066-3
* 	- ROMEDRV		: 3.6-3
*
*
* Revision 1.11  2005/08/25 04:01:35  chenyl
* *: rearrange code, add MUTEX lock/unlock for external functions, add `inline' property to some critical functions
*
* Revision 1.10  2005/06/01 13:56:18  tony
* *: support alg reinit.
*
* Revision 1.9  2004/12/22 05:19:15  tony
* +: support new features: make ALG(FTP,SIP,NetMeeting) support MNQueue.
*
* Revision 1.8  2004/09/07 14:52:13  chenyl
* *: bug fix: napt-redirect fragment packet checksum-recalculate
* *: bug fix: conflict between protocol stack flow cache and protocol stack action
* *: bug fix: protocol stack action mis-process to UDP-zero-checksum packet
* *: separate the header file:
*         - internal : rtl8651_tblDrvFwdLocal.h
*         - external : rtl8651_tblDrvFwd.h
*
* Revision 1.7  2004/07/23 13:19:57  tony
* *: remove all warning messages.
*
* Revision 1.6  2004/07/20 06:25:47  tony
* *: remove debug message
*
* Revision 1.5  2004/07/20 06:19:53  chhuang
* *: fix sip bug
*
* Revision 1.4  2004/04/20 03:44:03  tony
* if disable define "RTL865X_OVER_KERNEL" and "RTL865X_OVER_LINUX", __KERNEL__ and __linux__ will be undefined.
*
* Revision 1.3  2004/03/19 12:17:21  cfliu
* Rename all osk_malloc, printk, printf, printfByPolling as rtlglue_printf
*
* Revision 1.2  2004/03/15 09:08:54  tony
* support for test module(endian free), fix sip bug, support for src port not 5060
*
* Revision 1.1  2004/02/25 14:26:33  chhuang
* *** empty log message ***
*
* Revision 1.2  2004/02/25 14:24:52  chhuang
* *** empty log message ***
*
* Revision 1.8  2004/01/29 01:39:47  orlando
* checkin acl's dsid support, not the final version
*
* Revision 1.7  2004/01/27 08:34:55  tony
* make SIP ALG support incoming call.
*
* Revision 1.6  2004/01/12 11:59:34  tony
* make ALG is able to enable/disable for multiple dial session.
*
* Revision 1.5  2003/12/22 11:59:14  tony
* fix bug: unalignment issue.
*
* Revision 1.4  2003/12/19 11:40:01  tony
* change memcpy to rtl8651_memcpy
*
* Revision 1.3  2003/12/17 10:30:36  hiwu
* add checksum function
*
* Revision 1.2  2003/12/12 11:33:24  tony
* modify SIP ALG, make it support CISCO IP Phone 7960 dial from LAN.
*
* Revision 1.1  2003/12/08 13:56:52  tony
* add new ALG: support SIP protocol 
*
*
*/


#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"

#include "rtl8651_alg_sip.h"

#define _RTL8651_L4_MAX_SIP_ALIAS_BUF_SIZE 1520

static uint8 *_rtl8651_l4_sip_alias_buf[1];

static uint16 o_port=0,i_port=0;

int32 rtl8651_l4AliasHandleSipClientOutbound(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tb)
{
	struct ip *spip;
	struct udphdr *udp,*sudp;
	int32 tlen, dlen, off_set, i ,j, k;
	int32 size_delta=0;
	int8 *sptr, *newpacket,*data;

	char localip[16];
	int8	localip_len;
	char globalip[16];
	int8	globalip_len;

	struct rtl_mBuf *m = pkthdrPtr->ph_mbuf;

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_SIP)==0) return SUCCESS;

	//printk("into sip client out\n");

	// private ip do nothing, at the same submask
	//if((ntohl(pip->ip_dst.s_addr)&0x0a000000)==0x0a000000) return SUCCESS; //10.x.x.x
	//if((ntohl(pip->ip_dst.s_addr)&0xc0a80000)==0xc0a80000) return SUCCESS; //192.168.x.x
	//if((ntohl(pip->ip_dst.s_addr)&0xac13c000)==0xac130000) return SUCCESS; //172.19.16.x~172.19.32.x

		
	udp = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));	
	data = (uint8 *)(udp+1);
	
	tlen = ntohs(pip->ip_len);
	dlen = tlen-sizeof(struct udphdr)-((pip->ip_vhl &0xf) << 2);

			
	off_set = ((uint8 *)pip) - m->m_data;
	if(off_set < 0 || off_set > 40 || tb == NULL) return FAILED;


	// assigned new packet buffer ; use the latter part of _rtl8651_l4_sip_alias_buf 
	newpacket =  (int8 *)_rtl8651_l4_sip_alias_buf[0];
	
	// copy out the mbuf data 
	mBuf_copyToUserBuffer(m, off_set, m->m_pkthdr->ph_len - off_set, newpacket);

	// Place string pointer and beginning of data 
	spip = (struct ip*)newpacket;
	sudp = (struct udphdr *) ((int8 *) spip + ((spip->ip_vhl &0xf) << 2));
	sptr = (int8 *)(sudp+1);
	
	sprintf(localip,"%u.%u.%u.%u",(tb->insideLocalIpAddr>>24)&0xff,(tb->insideLocalIpAddr>>16)&0xff,(tb->insideLocalIpAddr>>8)&0xff,(tb->insideLocalIpAddr)&0xff);
	localip_len=strlen(localip);

	sprintf(globalip,"%u.%u.%u.%u",(tb->insideGlobalIpAddr>>24)&0xff,(tb->insideGlobalIpAddr>>16)&0xff,(tb->insideGlobalIpAddr>>8)&0xff,(tb->insideGlobalIpAddr)&0xff);
	globalip_len=strlen(globalip);	
			
	for(i=0,j=0;i<dlen;i++,j++)
	{
		if(data[i]<0x20&&data[i]!=0x0d&&data[i]!=0x0a)  return SUCCESS; //not sip packet
	
		if(i+4<=dlen)
		{
			if(memcmp(&data[i],"\r\nm=",4)==0)
			{
				int flag=0;
				i_port=0;
				for(k=i+4;k<dlen;k++)
				{
					if(data[k]==' ')
					{
						if(flag==0) flag=k;
						else
						{
							//printk("out- in_port=[%d] out_port=%d\n",i_port,o_port);
							_rtl8651_addAlgQosUpnpMap(UPNP_PERSIST|UPNP_ALG_QOS, 0, 0, 
							tb->insideGlobalIpAddr, i_port,
							tb->insideLocalIpAddr, i_port,RTL8651_ALG_SIP_IDX);
							//o_port=0;
							break;
						}
					}
					else if(flag!=0) i_port=(data[k]-'0')+i_port*10;
				}
			}
		}
		
		if((i+localip_len<=dlen)&&(memcmp(&data[i],localip,localip_len)==0))
		{
			//printk("out-match ip address = %s-->%s\n",localip,globalip);
			rtl8651_memcpy(&sptr[j],globalip,globalip_len);
			i+=(localip_len-1);
			j+=(globalip_len-1);
			size_delta+=(globalip_len-localip_len);				
		}
		else
		{
			sptr[j]=data[i];
		}

	}

	//printk("old len=%d, delta=%d\n",spip->ip_len,size_delta);
	spip->ip_len = htons(ntohs(spip->ip_len)+size_delta);	
	sudp->uh_ulen = htons(ntohs(sudp->uh_ulen)+size_delta);	


	//stcp->th_sum = 0;
	//printk("len=%d offset=%d\n",ntohs(spip->ip_len),off_set);
	
	/* Compute TCP/IP checksum for revised packet */
	sudp->uh_sum = 0;
	spip->ip_sum = 0;				
	sudp->uh_sum = _rtl8651_tcpChecksum(spip); /* tcp/udp checksum function is the same */
	spip->ip_sum = _rtl8651_ipChecksum(spip);	
		
	mBuf_copyToMbuf(m, off_set, ntohs(spip->ip_len), newpacket);
	
	assert(m->m_len == m->m_pkthdr->ph_len);

#ifdef _RTL_UPNP
	if(tb->insideLocalPort==5060)
	{
		/* add PNP port map */
		rtl8651_addUpnpMap(UPNP_PERSIST, 0,5060, 
			tb->insideGlobalIpAddr, 5060,
			tb->insideLocalIpAddr, 5060);
		tb->insideGlobalPort=5060;	
	}
#endif /* _RTL_UPNP */	
	

	return SUCCESS;
}


int32 rtl8651_l4AliasHandleSipClientInbound(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tb)
{
	struct ip *spip;
	struct udphdr *udp,*sudp;
	int32 tlen, dlen, off_set, i ,j, k;
	int32 size_delta=0;
	int8 *sptr, *newpacket,*data;

	char localip[16];
	int8	localip_len;
	char globalip[16];
	int8	globalip_len;
	struct rtl_mBuf *m = pkthdrPtr->ph_mbuf;

	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_SIP)==0) return SUCCESS;

	//printk("into sip client in\n");
	
	// private ip do nothing
	//if((ntohl(pip->ip_src.s_addr)&0x0a000000)==0x0a000000) return SUCCESS; //10.x.x.x
	//if((ntohl(pip->ip_src.s_addr)&0xc0a80000)==0xc0a80000) return SUCCESS; //192.168.x.x
	//if((ntohl(pip->ip_src.s_addr)&0xac13c000)==0xac130000) return SUCCESS; //172.19.16.x~172.19.32.x

	udp = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));	
	data = (uint8 *)(udp+1);
	
	tlen = ntohs(pip->ip_len);
	dlen = tlen-sizeof(struct udphdr)-((pip->ip_vhl &0xf) << 2);

			
	off_set = ((uint8 *)pip) - m->m_data;
	if(off_set < 0 || off_set > 40 || tb == NULL)  return FAILED;


	// assigned new packet buffer ; use the latter part of _rtl8651_l4_sip_alias_buf 
	newpacket =  (int8 *)_rtl8651_l4_sip_alias_buf[0];
	
	// copy out the mbuf data 
	mBuf_copyToUserBuffer(m, off_set, m->m_pkthdr->ph_len - off_set, newpacket);

	// Place string pointer and beginning of data 
	spip = (struct ip*)newpacket;
	sudp = (struct udphdr *) ((int8 *) spip + ((spip->ip_vhl &0xf) << 2));
	sptr = (int8 *)(sudp+1);

	sprintf(localip,"%u.%u.%u.%u",(tb->insideLocalIpAddr>>24)&0xff,(tb->insideLocalIpAddr>>16)&0xff,(tb->insideLocalIpAddr>>8)&0xff,(tb->insideLocalIpAddr)&0xff);
	localip_len=strlen(localip);

	sprintf(globalip,"%u.%u.%u.%u",(tb->insideGlobalIpAddr>>24)&0xff,(tb->insideGlobalIpAddr>>16)&0xff,(tb->insideGlobalIpAddr>>8)&0xff,(tb->insideGlobalIpAddr)&0xff);
	globalip_len=strlen(globalip);
		
	for(i=0,j=0;i<dlen;i++,j++)
	{
		if(data[i]<0x20&&data[i]!=0x0d&&data[i]!=0x0a)  return SUCCESS; //not sip packet
		
	
		if(i+4<=dlen)
		{
			if(memcmp(&data[i],"\r\nm=",4)==0)
			{
				int flag=0;
				o_port=0;
				for(k=i+4;k<dlen;k++)
				{
					if(data[k]==' ')
					{
						if(flag==0) flag=k;
						else
						{							
							#ifdef _RTL_UPNP
							//printk("in- out_port=[%d]\n",i_port);
							/* add PNP port map */
							/*
							rtl8651_addUpnpMap(UPNP_PERSIST, 0, o_port, 
							tb->insideGlobalIpAddr, i_port,
							tb->insideLocalIpAddr, i_port);
							*/
							#endif /* _RTL_UPNP */	
							
							break;
						}
					}
					else if(flag!=0) o_port=(data[k]-'0')+o_port*10;
				}
			}
		}

		if(((i+globalip_len<=dlen))&&(memcmp(&data[i],globalip,globalip_len)==0))
		{
			//printk("match ip in address = %s-->%s\n",globalip,localip);
			rtl8651_memcpy(&sptr[j],localip,localip_len);
			i+=(globalip_len-1);
			j+=(localip_len-1);
			size_delta+=(localip_len-globalip_len);	
				
		}
		else
		{
			sptr[j]=data[i];
		}
	}
	
	spip->ip_len = htons(ntohs(spip->ip_len)+size_delta);	
	sudp->uh_ulen = htons(ntohs(sudp->uh_ulen)+size_delta);

	/* Compute TCP/IP checksum for revised packet */
	sudp->uh_sum = 0;
	spip->ip_sum = 0;				
	sudp->uh_sum = _rtl8651_tcpChecksum(spip); /* tcp/udp checksum function is the same */
	spip->ip_sum = _rtl8651_ipChecksum(spip);	

	mBuf_copyToMbuf(m, off_set, ntohs(spip->ip_len), newpacket);
	
	assert(m->m_len == m->m_pkthdr->ph_len);


	return SUCCESS;	
}

static int32 init_sip = 0;

void rtl8651_l4AliasSipReInit(void)
{
	o_port=0;
	i_port=0;
}

int32 rtl8651_l4AliasSipInit(void *arg){
	
	/* malloc global variable */
	if(init_sip)
		return SUCCESS;

	_rtl8651_l4_sip_alias_buf[0] = (uint8 *)rtlglue_malloc(_RTL8651_L4_MAX_SIP_ALIAS_BUF_SIZE);
	assert(_rtl8651_l4_sip_alias_buf[0]);

	init_sip = 1;
	return SUCCESS;
}


