/*
* Copyright c                  Realtek Semiconductor Corporation, 2005
* All rights reserved.
* 
* Program : TFTP ALG (for single session)
* Abstract : 
* Creator : Yi-Lun Chen (chenyl@realtek.com.tw)
* Author : Yi-Lun Chen (chenyl@realtek.com.tw)
*/


#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"

#include "rtl8651_alg_tftp.h"

/*
	To handle LAN to WAN TFTP client flow:

	With destination port being 69.
	- We need to check if this flow is TFTP or not. If it is, we would allow the following flow chart:

	LAN                                              GW                                              WAN

	--(Source Port X to Dport 69)-->
							             --(Source Port X' to Dport 69)-->

						                    <--(Source Port Y to Dport X')---
       <--(Source Port Y to Dport X)---
*/
int32 rtl8651_l4AliasHandleTftpClientOutbound (
	struct rtl_pktHdr *pkthdrPtr,
	struct ip *pip,
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tb)
{
	struct udphdr *ud = NULL;
	struct tftpRequest *tftp = NULL;

	if ((alg_supports[pkthdrPtr->ph_dsid] & RTL8651_ALG_TFTP) == 0)
	{
		return SUCCESS;
	}

	if (	( pkthdrPtr == NULL ) ||
		( pip == NULL ) ||
		( tb == NULL ))
	{
		goto out;
	}

	/* <---------- Check if this packet is TFTP RRQ/WRQ or not ----------> */
	if (	( pip->ip_p != IPPROTO_UDP ) ||
		( IS_FRAGMENT_REMAIN(pip->ip_off) ) )
	{
		/*
			TFTP ALG: Only for UDP, non-fragmented packet
		*/
		goto out;
	}

	ud = (struct udphdr*)L4HDR_ADDR(pip);

	if ( ud->uh_dport != TFTP_UDP_DPORT )
	{
		goto out;
	}

	tftp = (struct tftpRequest *)((uint32)ud + sizeof(struct udphdr));

	switch ( ntohs(tftp->opcode) )
	{
		case TFTP_RRQ:
		case TFTP_WRQ:
			/* We only use OPCODE to decide this packet is TFTP or NOT */
			break;
		default:
			goto out;
	}

	/* <---------- For TFTP RRQ/WRQ : get its sport to further packet input ----------> */
	rtl8651_addUpnpMapLeaseTime(
		UPNP_ONESHOT,				/* Only for 1 flow from TFTP server to LAN */
		tb->dstIpAddr,				/* Only allow respose from TFTP server */
		0,							/* We don't know the source port of TFTP server response */
		tb->insideGlobalIpAddr,		/* TFTP server would response to GW's external IP */
		tb->insideGlobalPort,			/* TFTP server would response to GW's external port */
		tb->insideLocalIpAddr,			/* We need to transfer TFTP packet to TFTP client's IP */
		tb->insideLocalPort,			/* We need to transfer TFTP packet to TFTP client's port */
		TFTP_FLOW_CACHE_TIMEOUT	/* Flow cache age-timeout */
		);
out:
	return SUCCESS;
}

