/*
* Copyright c                  Realtek Semiconductor Corporation, 2003  
* All rights reserved.
* 
* Program : 
* Abstract : 
* Creator : 
* Author :
* $Id: rtl8651_alg_ymsg.c,v 1.1.1.1 2006/07/26 05:38:23 blue Exp $
* $Log: rtl8651_alg_ymsg.c,v $
* Revision 1.1.1.1  2006/07/26 05:38:23  blue
*
*
* Revision 1.1.1.1  2006/02/14 03:09:06  jason
* backup the source code from v1036
*
* Revision 1.1.1.1  2006/01/16 02:39:43  jason
* backup source,because now we directly use romedriver source code, i open a new directory;
*
* Revision 1.2  2005/11/22 15:50:54  chenyl
* *: Add mutex lock/unlock in log r/w
* *: new idle-timeout mechanism for VLAN / PPPOE / PPTP / L2TP wan types
* *: bug fix: PPTP passthrough
* *: bug fix: sip / ymsg typo error
* *: for ARP-reply module in RomeDrv, use newly allocated mbuf/pkt instead of original one ( To prevent from treated as packet forwarding instead of reply).
* *: minor coding style, sync with main trunk
*
* +: ioctl case for Mutex Lock/Unlock called by user level
* +: add 2 APIs to enable / disable
* 	1. Error messages from protocol stack to WAN : TCP RST, ICMP ICMP_UNREACH/ICMP_SOURCEQUENCH/ICMP_TIMXCEED/ICMP_PARAMPROB
* 	2. All packet from protocol stack to WAN
*    filtering to disable / enable idle-timer update triggered by these kinds of packets.
* +: TFTP ALG
* +: rtl8651_debug.h for debugging usage
* +: merge code about rtl_glue.c / re_core.c with Sercomm's version ( Airgo modification ?)
* +:
*
* Revision 1.1.1.1  2005/11/21 05:20:03  chenyl
* +: WPNT834
* 	- System provider	: Sercomm
* 	- Brand Name		: Netgear
* 	- 3rt Party		: Airgo Gen 3
* 	- SDK			: 066-3
* 	- ROMEDRV		: 3.6-3
*
*
* Revision 1.9  2005/08/25 04:01:35  chenyl
* *: rearrange code, add MUTEX lock/unlock for external functions, add `inline' property to some critical functions
*
* Revision 1.8  2005/08/15 10:53:28  chenyl
* +: new TRIGGER PORT mechanism
* 	- sip check when activate Trigger Port.
* 	- age timer for Triggered Port Mapping.
*
* Revision 1.7  2004/09/07 14:52:13  chenyl
* *: bug fix: napt-redirect fragment packet checksum-recalculate
* *: bug fix: conflict between protocol stack flow cache and protocol stack action
* *: bug fix: protocol stack action mis-process to UDP-zero-checksum packet
* *: separate the header file:
*         - internal : rtl8651_tblDrvFwdLocal.h
*         - external : rtl8651_tblDrvFwd.h
*
* Revision 1.6  2004/07/23 13:19:57  tony
* *: remove all warning messages.
*
* Revision 1.5  2004/04/20 03:44:03  tony
* if disable define "RTL865X_OVER_KERNEL" and "RTL865X_OVER_LINUX", __KERNEL__ and __linux__ will be undefined.
*
* Revision 1.4  2004/03/19 12:17:21  cfliu
* Rename all osk_malloc, printk, printf, printfByPolling as rtlglue_printf
*
* Revision 1.3  2004/03/18 09:22:26  tony
* add alg reinit function, it will be called when fwdEng ReInit.
*
* Revision 1.2  2004/03/03 09:15:44  tony
* update to endian free version for test module
*
* Revision 1.1  2004/02/25 14:26:33  chhuang
* *** empty log message ***
*
* Revision 1.2  2004/02/25 14:24:52  chhuang
* *** empty log message ***
*
* Revision 1.13  2004/02/06 12:33:37  tony
* add rtl8651a_addTriggerPortRule and rtl8651a_flushTriggerPortRules API
*
* Revision 1.12  2004/01/30 12:03:05  tony
* make Special Application is able to support multiple session.
*
* Revision 1.11  2004/01/29 01:39:47  orlando
* checkin acl's dsid support, not the final version
*
* Revision 1.10  2004/01/12 11:59:34  tony
* make ALG is able to enable/disable for multiple dial session.
*
* Revision 1.9  2003/12/22 07:22:12  tony
* fix bug: unalignment issue appear in YMSG ALG.
*
* Revision 1.8  2003/12/19 11:40:01  tony
* change memcpy to rtl8651_memcpy
*
* Revision 1.7  2003/12/17 10:27:07  hiwu
* add checksum function
*
* Revision 1.6  2003/12/12 01:36:22  tony
* make YMSG ALG support protocol v10 and v11 at the same time.
*
* Revision 1.5  2003/11/13 12:29:39  tony
* fix trigger port 5050 to 5055
*
* Revision 1.4  2003/11/07 05:34:36  tony
* modify YMSG ALG, it'll not have to cooperate with trigger port.
*
* Revision 1.3  2003/11/04 03:01:44  tony
* small change
*
* Revision 1.2  2003/11/03 12:53:40  tony
* modify alg:yahoo messenger bug
*
* Revision 1.1  2003/11/03 08:30:30  tony
* add ALG:Yahoo Messenger v5.6 (Peer-to-Peer File transfer)
*

*
*/

#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_ymsg.h"

#define _RTL8651_L4_MAX_YMSG_ALIAS_BUF_SIZE 576

static uint8 *_rtl8651_l4_ymsg_alias_buf[1]; 
static ipaddr_t triggerIP=0;

int32 rtl8651_l4AliasHandleYmsgClientOutbound(struct rtl_pktHdr *pkthdrPtr, struct ip *pip,
			struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tb)
{
	struct ip *spip;
	struct tcphdr *tcp,*stcp;
	struct ymsg_command *ymsg;
	int32 hlen, tlen, dlen, off_set, i ,size_delta;
	int8 *sptr, *newpacket;
	int8 sflag=0, http[8]="http://",*ymsg_content;	
	int8 route_ip[22]={0};
	ipaddr_t src,dest;	
	struct rtl_mBuf *m = pkthdrPtr->ph_mbuf;


	if((alg_supports[pkthdrPtr->ph_dsid]&RTL8651_ALG_YAHOOMSG)==0) return SUCCESS;
	
	src=ntohl(PKTGET_UINT32_UNALIGNED(&pip->ip_src.s_addr));
	dest=ntohl(PKTGET_UINT32_UNALIGNED(&pip->ip_dst.s_addr));


#ifdef _RTL_TRIGGER_PORT
	{
		triggerIP = src;		
		rtl8651a_addTriggerPortRule(pkthdrPtr->ph_dsid,2,5000,5001,0,0,0,htonl(triggerIP));
		rtl8651a_addTriggerPortRule(pkthdrPtr->ph_dsid,2,5055,5055,0,0,0,htonl(triggerIP));
	}
#endif	

	// disabled 2003/3/3
	// private ip do nothing
	//if((dest&0x0a000000)==0x0a000000) return SUCCESS; //10.x.x.x
	//if((dest&0xc0a80000)==0xc0a80000) return SUCCESS; //192.168.x.x
	//if((dest&0xac13c000)==0xac130000) return SUCCESS; //172.19.16.x~172.19.32.x
		
	tcp = (struct tcphdr *) ((int8 *) pip + ((pip->ip_vhl &0xf) << 2));	
	ymsg = (struct ymsg_command *)((uint8 *)tcp + ((tcp->th_off_x & 0xf0) >> 2));	
	ymsg_content = (void *)ymsg + sizeof(struct ymsg_command);
	
	off_set = ((uint8 *)pip) - m->m_data;
	
	if(off_set < 0 || off_set > 40 || tb == NULL) return FAILED;

	if(memcmp(ymsg->ymsg,"YMSG",4)!=0)  return SUCCESS;	
	if(ntohs(ymsg->service)!=YAHOO_SERVICE_P2PFILEXFER) return SUCCESS;
	
	//if((ymsg->version<0x000b0000)||(ymsg->version>0x00ff0000)) return SUCCESS; //only support YMSG protocol >=11
	//printk("modify ymsg buffer\n");
	
	/* test the packet length */
	if (m->m_pkthdr->ph_len > _RTL8651_L4_MAX_YMSG_ALIAS_BUF_SIZE) return FAILED;
	
	/* assigned new packet buffer ; use the latter part of _rtl8651_l4_ftp_alias_buf */
	newpacket =  (int8 *)_rtl8651_l4_ymsg_alias_buf[0];
	
	/* copy out the mbuf data */
	mBuf_copyToUserBuffer(m, off_set, m->m_pkthdr->ph_len - off_set, newpacket);
	
	tlen = ntohs(pip->ip_len);
	hlen = ((pip->ip_vhl & 0xf) + (( tcp->th_off_x >>4)&0xf)) << 2;//Edward: Fix for alignment issue
	dlen = tlen - hlen;

	/* Place string pointer and beginning of data */
	spip = (struct ip*)newpacket;
	stcp = (struct tcphdr *) ((int8 *) spip + ((spip->ip_vhl &0xf) << 2));	
	sptr = (int8 *) newpacket+hlen+sizeof(struct ymsg_command); // pointer to ysmg content
			
	//printk("YMSG version=%x length=%d dlen=%d service=%x status=%x\n",ymsg->version,ymsg->length,dlen,ymsg->service,ymsg->status);
		
	for(i=0;i<ntohs(ymsg->length)-10;i++)
	{
		if((sflag!=0)&&(ymsg_content[i]=='/')) break;		
		if((sflag==0)&&(memcmp((void *)(ymsg_content+i),http,7)==0)) {  i+=7; sflag=i; }
	}

	if(sflag==0) return SUCCESS;	//support for old version YMSG protocol 

	sprintf(route_ip,"%u.%u.%u.%u:5002",(tb->insideGlobalIpAddr>>24)&0xff,
		(tb->insideGlobalIpAddr>>16)&0xff,
		(tb->insideGlobalIpAddr>>8)&0xff,
		(tb->insideGlobalIpAddr)&0xff);

	size_delta=strlen(route_ip)-(i-sflag);
	//printk("rip:%s sd:%d\n",route_ip,size_delta);	

	sptr+=sflag;
	rtl8651_memcpy(sptr,route_ip,strlen(route_ip));
	rtl8651_memcpy(sptr+strlen(route_ip),ymsg_content+i,ntohs(ymsg->length)-i+1);
			
	/* Save information regarding modified seq and ack numbers */
	{
		int32     delta;
		tb->tcpAckModified = 1;
		delta = _rtl8651_l4GetDeltaSeqOut(spip, stcp, tb);
		_rtl8651_l4AddSeq(spip, tb, delta + size_delta);
	}

	spip->ip_len = htons(ntohs(spip->ip_len)+size_delta);
	
	/* Compute TCP/IP checksum for revised packet */
	stcp->th_sum = 0;
	spip->ip_sum = 0;				
	stcp->th_sum = _rtl8651_tcpChecksum(spip);
	spip->ip_sum = _rtl8651_ipChecksum(spip);	

	mBuf_copyToMbuf(m, off_set, ntohs(spip->ip_len), newpacket);
	
	assert(m->m_len == m->m_pkthdr->ph_len);

#ifdef _RTL_UPNP
		/* add PNP port map */
		rtl8651_addUpnpMap(UPNP_TCP|UPNP_ONESHOT, 0, 0, 
				tb->insideGlobalIpAddr, 5002,
				tb->insideLocalIpAddr, 80);	
#endif /* _RTL_UPNP */

	return SUCCESS;
}

static int32 init_ymsg = 0;
int32 rtl8651_l4AliasYmsgInit(void *arg){
	
	/* malloc global variable */
	if(init_ymsg)
		return SUCCESS;

	_rtl8651_l4_ymsg_alias_buf[0] = (uint8 *)rtlglue_malloc(_RTL8651_L4_MAX_YMSG_ALIAS_BUF_SIZE+64);
	assert(_rtl8651_l4_ymsg_alias_buf[0]);

	init_ymsg = 1;	
	return SUCCESS;
}

void rtl8651_l4AliasYmsgReInit()
{
	triggerIP=0;
}


