/*
* Copyright c                  Realtek Semiconductor Corporation, 2005
* All rights reserved.
* 
* Program :	rtl8651_dns.c
* Abstract :	DNS related process in ROMEDRV
* Creator :	Yi-Lun Chen (chenyl@realtek.com.tw)
* Author :	Yi-Lun Chen (chenyl@realtek.com.tw)
*
*/


#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_errno.h"
#include "rtl_glue.h"
#include "mbuf.h"

#include "mbuf.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_dns_local.h"

/* =============================================================
	Internal Function declarations
    ============================================================= */
/*
  *		Utilties of DNS process system
  *
  */

/* <============ Global Control Variables ============> */
static	_rtl8651_dnsDB_ctrl_t _rtl8651_dns_ctrlBlock;
uint32	_rtl8651_dns_enableTTLSuppression			= FALSE;	/* enable / disable RR's TTL field suppression */
uint32	_rtl8651_dns_suppressedTTL					= 0;		/* suppress DNS RR's TTL field to this value */

static int __inet_aton(const char *cp, ipaddr_t *addrptr);
static ipaddr_t __inet_addr(const char *cp);

static _rtl8651_dnsDB_t *__rtl8651_dns_getDnsDB(uint8 process, char *domain, _rtl8651_dnsDB_t *start, uint8 isExactlyMatch);
static _rtl8651_dnsDB_t *__rtl8651_dns_addAndLinkDnsDB(uint8 process);
static int32 __rtl8651_dns_UnlinkAndDelDnsDB(_rtl8651_dnsDB_t *dnsDb_p);
static int32 __rtl8651_dns_destroyDnsDB(uint8 process, _rtl8651_dnsDB_t *dnsDb_p);
static int32 __rtl8651_dns_flushDataBase(void);
static int32 _rtl8651_dns_dataBase_init(rtl8651_fwdEngineInitPara_t *para);
static int32 _rtl8651_dns_dataBase_reinit(void);
static int32 _rtl8651_dns_funcCallBack(char *mappedDomain, ipaddr_t mappedIp);


/*
  *		DNS packet process:
  *
  *			Parse DNS message and get DomainName <-->IP mapping to central control sub-system.
  *
  */
static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameMap_p = NULL;
static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameLru_h = NULL;
static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameLru_t = NULL;
static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameFree_p = NULL;
static uint32 _rtl8651_dns_cnameTimeout = 0;

static int32 _rtl8651_dns_cnameMapping_init(rtl8651_fwdEngineInitPara_t *para);
static void _rtl8651_dns_cnameMapping_reinit(void);
static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameMapping_find(char *name);
static inline _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameMapping_getEntry(void);
static void _rtl8651_dns_cnameMapping_refresh(_rtl8651_dns_cnameMapping_t *entry);
static int32 _rtl8651_dns_cnameMapping_add(char *name1, char *name2);
static void _rtl8651_dns_cnameMapping_del(_rtl8651_dns_cnameMapping_t *delEntry);
static void _rtl8651_dns_cnameMapping_TimeUpdate(uint32 secPassed);
static int32 _rtl8651_dns_procDnsCallBack(ipaddr_t mapped_ip, char *mapped_domain, int32 (*dns_proc)(char *, ipaddr_t));

static int32 _RTL8651_DNS_GET_NAMELEN(uint8 *name, uint32 maxLen, uint32 *len_p);
static int32 _RTL8651_DNS_GET_INFO(_rtl8651_dns_dnsDomain_t domainType, uint8 *outputPtr, uint32 maxOutputLen, uint8 *inputPtr, uint8 *dnsData_p, uint32 dnsDataLen);
static int32 _rtl8651_dns_parseDnsPkt(uint8 *dns_p, uint32 dnsLen, uint8 *pktModified, int32 (*dns_proc)(char *, ipaddr_t));
static int32 _rtl8651_dns_parseDnsRR(uint8 *dns_p, uint32 totalLen, uint8 *rr_p, uint32 maxLen, uint32 *rrLen_p, uint8 *pktModified, int32 (*dns_proc)(char *, ipaddr_t));

/* =============================================================
	Funtion Implementations
    ============================================================= */

/* =========================================================================
  *
  *
  *					Internal functions.
  *
  *
  * ========================================================================= */
#define RTL8651_DNS_DATABASE

#define __rtl_isupper(c)	(((c) >= 'A') && ((c) <= 'Z'))

/*
	copy string from 'src' to 'dst' and set all alphabets to lower-case in 'dst'.
*/
static inline void __strlowerncpy(char *dst, const char *src, int32 len)
{
	int32 cpLen = 0;
	char p;

	while ((cpLen < len) && (src[cpLen] != 0))
	{
		p = src[cpLen];
 		dst[cpLen] = __rtl_isupper(p)?(p-'A'+'a'):p;
		cpLen ++;
	}
}

static inline int __inet_aton(const char *cp, ipaddr_t *addrptr)
{
	ipaddr_t addr;
	int value;
	int part;

	addr = 0;
	for (part = 1; part <= 4; part++)
	{
		if (!isdigit(*cp))
			return FAILED;

		value = 0;
		while (isdigit(*cp))
		{
			value *= 10;
			value += *cp++ - '0';
			if (value > 255)
				return FAILED;
		}

		if (part < 4)
		{
			if (*cp++ != '.')
				return FAILED;
		} else
		{
			char c = *cp++;
			if (c != '\0' && !isspace(c))
			return FAILED;
		}

		addr <<= 8;
		addr |= value;
	}

	/*  W. Richard Stevens in his book UNIX Network Programming,
	 *  Volume 1, second edition, on page 71 says:
	 *
	 *  An undocumented feature of inet_aton is that if addrptr is
	 *  a null pointer, the function still performs it validation
	 *  of the input string, but does not store the result.
	 */
	if (addrptr)
	{
		PKTSET_UINT32_UNALIGNED(addr, addrptr);
	}

	return SUCCESS;
}

static ipaddr_t __inet_addr(const char *cp)
{
	ipaddr_t a;

	if (__inet_aton(cp, &a) != SUCCESS)
	{
		return 0;
	}else
	{
		return a;
	}
}

/*
  *	Get DNS DB entry using 'domain' and 'isExactlyMatch' for process 'process'.
  *	If the pointer 'start' is given, it means caller want to find DNS entry started from "NEXT entry of ptr", or
  *	the search would start from the first entry of store in database for process 'process'.
  */
static _rtl8651_dnsDB_t *__rtl8651_dns_getDnsDB(
	uint8 process,
	char *domain,
	_rtl8651_dnsDB_t *start,
	uint8 isExactlyMatch)
{
	_rtl8651_dnsDB_t *ptr = NULL;
	uint32 domainLen = 0;

	if (process >= _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		RTL_DNS_ERR("Invalid process %d", process);
		goto out;
	}

	domainLen = strlen(domain);

	if ((domainLen >= RTL8651_DNS_DOMAIN_MAXLENGTH) || (domainLen == 0))
	{
		goto out;
	}

	ptr = (start)?(start->next):_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process];

	while (ptr)
	{
		void *matchPos;

		/* For Exactly Match: these two string length MUST be the same. or they would not be the same. */
		if (	(isExactlyMatch == TRUE) && (strlen(ptr->domain) != strlen(domain)))
		{
			goto next;
		}

		if ((matchPos = (void*)strstr(domain, ptr->domain)) != NULL)
		{
			/*	For Exactly Match: if the string is matched but it just a substring of ptr->domain,
				this is not exactly match.												*/
			if (	(isExactlyMatch == TRUE) && (matchPos != domain))
			{
				goto next;
			}
			/* String match successfully: return this entry */
			goto out;
		}
	next:
		ptr = ptr->next;
	}

out:
	return ptr;
}

/*
	Allocate new DNS DB entry and link it to inused linked list.
*/
static _rtl8651_dnsDB_t *__rtl8651_dns_addAndLinkDnsDB(uint8 process)
{
	_rtl8651_dnsDB_t *ptr = NULL;

	if (process >= _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		RTL_DNS_ERR("Invalid process %d", process);
		goto out;
	}

	if (_rtl8651_dns_ctrlBlock.freeDnsDB)
	{
		/* allocate */
		ptr = _rtl8651_dns_ctrlBlock.freeDnsDB;
		_rtl8651_dns_ctrlBlock.freeDnsDB = (_rtl8651_dns_ctrlBlock.freeDnsDB)->next;

		/* link to corresponding process */
		ptr->hdr = &(_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process]);
		ptr->prev = NULL;
		ptr->next = _rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process];

		if (_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process])
		{
			(_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process])->prev = ptr;
		}

		_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[process] = ptr;

		/* init other data */
		memset(ptr->domain, 0, RTL8651_DNS_DOMAIN_MAXLENGTH);
		ptr->procData = NULL;

	}
out:
	return ptr;
}

/*
	Release DNS DB entry back to free list.

	Note that: caller need to clear 'procData' manually and clear this pointer,
	or error message would occur.
*/
static int32 __rtl8651_dns_UnlinkAndDelDnsDB(_rtl8651_dnsDB_t *dnsDb_p)
{
	_rtl8651_dnsDB_t *hdr = NULL;

	if (	(dnsDb_p == NULL			/* dnsDB pointer is not provided */) ||
		(dnsDb_p->procData != NULL	/* procData is not be cleaned by user */))
	{
		RTL_DNS_ERR("Invalid input:");
		RTL_DNS_ERR("\tDnsDB: %p", dnsDb_p);
		if (dnsDb_p)
		{
			RTL_DNS_ERR("\tDnsDB procData: %p (must be cleaned before delete this DB entry)", dnsDb_p->procData);
		}
		return FAILED;
	}

	/* unlink */
	if (dnsDb_p->next)
	{
		dnsDb_p->next->prev = dnsDb_p->prev;
	}
	if (dnsDb_p->prev)
	{
		dnsDb_p->prev->next = dnsDb_p->next;
	}

	hdr = *(_rtl8651_dnsDB_t**)(dnsDb_p->hdr);
	if (hdr == dnsDb_p)
	{
		*(dnsDb_p->hdr) = dnsDb_p->next;
	}

	/* clear this dns DB and give it back to free list */
	memset(dnsDb_p, 0, sizeof(_rtl8651_dnsDB_t));
	dnsDb_p->next = _rtl8651_dns_ctrlBlock.freeDnsDB;
	_rtl8651_dns_ctrlBlock.freeDnsDB = dnsDb_p;

	return SUCCESS;

}

/*
	call the DESTRUCTOR of process to free process's data tied with DNS DB entry and free that DNS DB entry.
*/
static int32 __rtl8651_dns_destroyDnsDB(uint8 process, _rtl8651_dnsDB_t *dnsDb_p)
{
	/* 1. delete user-specific procData */
	if (dnsDb_p->procData)
	{
		/* run destructor */
		if (_rtl8651_dns_ctrlBlock.callBack[process].destructor)
		{
			RTL_DNS_INFO("[START] destructor(%p) for process (%d), procData (%p)",
							_rtl8651_dns_ctrlBlock.callBack[process].destructor,
							process,
							dnsDb_p->procData);

			_rtl8651_dns_ctrlBlock.callBack[process].destructor(dnsDb_p->procData);

			RTL_DNS_INFO("[END] destructor OK");
		} else
		{
			RTL_DNS_WARN("No destructor for process (%d), procData (%p)",
							process,
							dnsDb_p->procData);
		}

		/* clear procData */
		dnsDb_p->procData = NULL;
	}

	return __rtl8651_dns_UnlinkAndDelDnsDB(dnsDb_p);
}

/*
	flush DNS DB.

	NOTE: all related processes' destructors would be called to free entries.
*/
static int32 __rtl8651_dns_flushDataBase(void)
{
	int32 idx, retval = FAILED;

	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		while (_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx])
		{
			retval = _rtl8651_dns_delProcessEntry(	idx,
													TRUE,	/* delete by pointer */
													_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx],
													NULL,
													NULL);
			if (retval != SUCCESS)
			{
				RTL_DNS_ERR("FATAL ERROR, clear DNS structure FAILED");
				goto out;
			}
		}
	}

	/* all deletion success */
	retval = SUCCESS;
out:
	return retval;
}

/*
	Init function of DNS DataBase
*/
static int32 _rtl8651_dns_dataBase_init(rtl8651_fwdEngineInitPara_t *para)
{
	/* init control block */
	memset(&_rtl8651_dns_ctrlBlock, 0 , sizeof(_rtl8651_dns_ctrlBlock));

	/* check parameters */
	if (	(para == NULL) ||
		(para->dns_dnsDB_tblSize == 0))
	{
		if (para)
		{
			RTL_DNS_WARN("Variable err: dns table size [%d]", para->dns_dnsDB_tblSize);
		} else
		{
			RTL_DNS_WARN("Variable err: para [%p]", para);
		}
		goto err_out;
	}

	_rtl8651_dns_ctrlBlock.dnsDB_unusedProcCnt = para->dns_dnsDB_processCnt;

	/* allocate DNS DB entries according to information in 'para' */
	MEM_ALLOC(_rtl8651_dns_ctrlBlock.freeDnsDB, sizeof(_rtl8651_dnsDB_t) * para->dns_dnsDB_tblSize, FAILED);
	DNS_LIST_INIT(_rtl8651_dns_ctrlBlock.freeDnsDB, _rtl8651_dnsDB_t, para->dns_dnsDB_tblSize, next);

	MEM_ALLOC(_rtl8651_dns_ctrlBlock.rtl8651_inusedDB, sizeof(_rtl8651_dnsDB_t*) * para->dns_dnsDB_processCnt, FAILED);
	MEM_ALLOC(_rtl8651_dns_ctrlBlock.callBack, sizeof(_rtl8651_dnsFuncCtrl_t*) * para->dns_dnsDB_processCnt, FAILED);

	/* backup callBack initial parameters */
	memcpy(&(_rtl8651_dns_ctrlBlock.initPara), para, sizeof(rtl8651_fwdEngineInitPara_t));

	return SUCCESS;

err_out:
	return FAILED;
}

/*
  *	Re-Init function of DNS DataBase
  *
  */
static int32 _rtl8651_dns_dataBase_reinit(void)
{
	return __rtl8651_dns_flushDataBase();
}

static int32 _rtl8651_dns_funcCallBack(char *mappedDomain, ipaddr_t mappedIp)
{
	int32 idx;
	_rtl8651_dnsDB_t *db_p;

	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.callBack[idx].proc)
		{
			db_p = NULL;
			while ((db_p = __rtl8651_dns_getDnsDB(
							idx,
							mappedDomain,
							db_p,
							(_rtl8651_dns_ctrlBlock.callBack[idx].property&RTL8651_DNSDB_PROPERTY_EXACTLYMATCH)?
							TRUE:FALSE))
					!= NULL)
			{
				_rtl8651_dns_ctrlBlock.callBack[idx].proc(db_p->domain, mappedDomain, mappedIp, db_p->procData);
			}
		}
	}
	return SUCCESS;
}

/*
@func void			| rtl8651_dns_dataBase_dump | Dump DNS database's information
@comm
	Dump information in DNS database module.
*/
void rtl8651_dns_dataBase_dump(void)
{
	int32 idx;
	_rtl8651_dnsDB_t *db_p;

	uint32 inusedCnt = 0;

	rtlglue_drvMutexLock();

	printk("\nDNS database table:\n\n");

	for (idx = 0 ; idx < (_rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt - _rtl8651_dns_ctrlBlock.dnsDB_unusedProcCnt); idx ++)
	{
		printk("Process [%d]\n", idx);

		printk("\tProperty:\t\t\t\t0x%x\n", _rtl8651_dns_ctrlBlock.callBack[idx].property);
		printk("\tREINIT call-back function:\t\t%p\n", _rtl8651_dns_ctrlBlock.callBack[idx].reinit);
		printk("\tPROC call-back function:\t\t%p\n", _rtl8651_dns_ctrlBlock.callBack[idx].proc);
		printk("\tTimeUpdate call-back function:\t\t%p\n", _rtl8651_dns_ctrlBlock.callBack[idx].timeUpdate);
		printk("\tDELETE call-back function:\t\t%p\n", _rtl8651_dns_ctrlBlock.callBack[idx].destructor);

		printk("\n");

		db_p = _rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx];
		if (db_p)
		{
			printk("\t[DATA PTR]\t[DOMAIN]\n");
			printk("\t===================================================\n");
			while (db_p)
			{
				printk("\t%p\t%s\n", db_p->procData, db_p->domain);
				inusedCnt ++;
				db_p = db_p->next;
			}
			printk("\n");
		}
	}
	printk("\n\tInused:\t%d\n", inusedCnt);
	printk("\tFree:\t%d\n", _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_tblSize - inusedCnt);

	rtlglue_drvMutexUnlock();
}

/* =========================================================================
  *
  *
  *					External functions.
  *
  *
  * ========================================================================= */
#define RTL8651_DNS_EXTERNAL

/*
@func int32	| rtl8651_dns_setDnsTtlSuppression	| Enable/Disable DNS TTL Suppression in DNS system
@parm uint8	| enable							| TRUE: Enable, FALSE: Disable
@parm uint32	| suppressedTtl					| TTL upper-bound for DNS A/PTR/CNAME record.
@rvalue TRUE									| TTL suppression is enabled.
@rvalue FALSE								| TTL suppression is disabled.
@comm
	Enable / Disable TTL suppression in DNS system:
	Due to many domain-based functions need to user issue DNS query in short period,
	DNS system would suppress DNS A/PTR/CNAME record's TTL to a short number.
	User can call this function to enable/disable this function.
*/
int32 rtl8651_dns_setDnsTtlSuppression(uint8 enable, uint32 suppressedTtl)
{
	rtlglue_drvMutexLock();
	_rtl8651_dns_enableTTLSuppression = enable?TRUE:FALSE;
	_rtl8651_dns_suppressedTTL = suppressedTtl;
	rtlglue_drvMutexUnlock();

	return enable;
}

/*
@func int32				| rtl8651_dns_registerProcess				| MOUDLES register related information and callBack funtions to DNS system
@parm uint8*														| process			| process number given by DNS system when registration success.
@parm int32(*)(void *para)											| init			| init function of process.
@parm void (*)(void)												| reinit			| re-init function of process
@parm int32 (*)(char *orig_domain, char *domain, ipaddr_t ip, void *data)	| proc			| domain-ip mapping proc function of process
@parm void (*)(uint32 secpassed)									| timeUpdate		| time-update function of process
@parm int32 (*)(void *data)											| destructor		| destructor of process's data
@rvalue SUCCESS													| registration success.
@rvalue FAILED													| registration failed.
@comm
	DNS system allow other MODULEs to register callback functions and DNS system would help them
	to parse DNS packet and provide (domain name <-> ip) mapping to them.

	If registration success, the sequence number of registrated process would be set in <p process> by DNS system.
*/
int32 rtl8651_dns_registerProcess(
	uint8 *process,
	uint32 property,
	int32 (*init)(void *para),
	void (*reinit)(void),
	int32 (*proc)(char *orig_domain, char *domain, ipaddr_t ip, void *data),
	void (*timeUpdate)(uint32 secpassed),
	int32 (*destructor)(void *data)
	)
{
	int32 retval;

	rtlglue_drvMutexLock();
	retval = _rtl8651_dns_registerProcess(	process,
											property,
											init,
											reinit,
											proc,
											timeUpdate,
											destructor);
	rtlglue_drvMutexUnlock();
	return retval;
}

/*
  *
  *	Let each functions register its callBack functions to process DNS.
  *
  *		System would run 'init' function when one function register its callback function.
  *
  */
int32 _rtl8651_dns_registerProcess(
	uint8 *process,
	uint32 property,
	int32 (*init)(void *para),
	void (*reinit)(void),
	int32 (*proc)(char *orig_domain, char *domain, ipaddr_t ip, void *data),
	void (*timeUpdate)(uint32 secpassed),
	int32 (*destructor)(void *data)
	)
{
	if (_rtl8651_dns_ctrlBlock.dnsDB_unusedProcCnt == 0)
	{
		RTL_DNS_ERR("NO Free entry for new DNS process! Please increase \"dns_dnsDB_processCnt\" of forwarding engine parameter.");
		return FAILED;
	}

	assert(process);

	*process = (_rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt - _rtl8651_dns_ctrlBlock.dnsDB_unusedProcCnt);

	RTL_DNS_INFO("Register function into DNS module.");
	RTL_DNS_INFO("================================");
	RTL_DNS_INFO("property:\t\t%d", property);
	RTL_DNS_INFO("init callBack:\t%p", init);
	RTL_DNS_INFO("reinit callBack:\t%p", reinit);
	RTL_DNS_INFO("proc callBack:\t%p", proc);
	RTL_DNS_INFO("timeUpdate:\t%p", timeUpdate);
	RTL_DNS_INFO("desructor:\t%p", destructor);
	RTL_DNS_INFO("================================");

	_rtl8651_dns_ctrlBlock.callBack[*process].property = property;
	_rtl8651_dns_ctrlBlock.callBack[*process].reinit = reinit;
	_rtl8651_dns_ctrlBlock.callBack[*process].proc = proc;
	_rtl8651_dns_ctrlBlock.callBack[*process].timeUpdate = timeUpdate;
	_rtl8651_dns_ctrlBlock.callBack[*process].destructor = destructor;

	if (init)
	{
		DNS_UNLOCK_ERROR_CHECK(init(&(_rtl8651_dns_ctrlBlock.initPara)), SUCCESS, TRUE, FAILED);	
	}

	_rtl8651_dns_ctrlBlock.dnsDB_unusedProcCnt --;

	RTL_DNS_INFO("--------> Registration OK.");

	return SUCCESS;
}

/*
@func int32	| rtl8651_dns_addProcessEntry		| Add entry for specific process
@parm uint8	| process							| process sequence to add entry
@parm char*	| domain							| domain of added entry
@parm void*	| procData						| process-owned data of added entry
@parm void**	| entryPtr						| ptr to allow process to drop( or get information from) this added entry by it.
@rvalue SUCCESS								| Entry addition success.
@rvalue FAILED								| Entry addition failed.
@comm
	Add entry for <p domain> / <p process> into DNS system.
	If addition success, related pointer would be assigned in <p destroyPtr> and returned to caller.
*/
int32 rtl8651_dns_addProcessEntry(
	uint8 process,
	char *domain,
	void *procData,
	void **entryPtr)
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_dns_addProcessEntry(	process,
											domain,
											procData,
											entryPtr);
	rtlglue_drvMutexUnlock();
	return retval;
}


int32 _rtl8651_dns_addProcessEntry(
	uint8 process,
	char *domain,
	void *procData,
	void **entryPtr)
{
	_rtl8651_dnsDB_t * dnsDb_p = NULL;
	int32 idx;
	int32 origTrapStatus = FAILED;
	uint32 domainLen = 0;

	if (process >= _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		RTL_DNS_ERR("Invalid process %d", process);
		goto err_out;
	}

	domainLen = strlen(domain);

	if ((domainLen >= RTL8651_DNS_DOMAIN_MAXLENGTH) || (domainLen == 0))
	{
		RTL_DNS_ERR("Domain name (%s) length is invalid: %d", domain, domainLen);
		goto err_out;
	}

	/*
		check redundancy:

			NOTE:
			We treat the "redundancy" as "entry has same `domain name' and `procData' to new entry".
	*/
	dnsDb_p = NULL;
	while ((dnsDb_p = __rtl8651_dns_getDnsDB(	process,
												domain,
												dnsDb_p,
												TRUE)) != NULL)
	{
		if (dnsDb_p->procData == procData)
		{
			RTL_DNS_WARN("Duplicate Entry: process: %d domain: %s", process, domain);
			goto ok_out;
		}
	}

	/* Before new entry addition, we check if there is any entry in DNS system */
	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx])
		{
			origTrapStatus = TRUE;
			break;
		}
	}

	/* allocate new entry */
	if ((dnsDb_p = __rtl8651_dns_addAndLinkDnsDB(process)) != NULL)
	{
		/* set process specific information: we ALWAYS set alphabets to lower-case */
		__strlowerncpy(dnsDb_p->domain, domain, RTL8651_DNS_DOMAIN_MAXLENGTH);
		dnsDb_p->procData = procData;
	}

	/*
		If original trap status == FAILED (no any entry in DNS system)
		and we have add ONE entry in DB now, it means we need to trap
		DNS packet to protocol stack.
	*/
	_rtl8651_arrangeAllACL();

ok_out:

	if (entryPtr)
	{
		*entryPtr = (void*)dnsDb_p;
	}

	return SUCCESS;

err_out:
	return FAILED;
}

/*
@func int32	| rtl8651_dns_delProcessEntry		| Delete entry from specific process
@parm uint8	| process							| process sequence to delete entry
@parm uint8	| byDbPtr						| indicate the entry deletion using <p pointer type>(TRUE) or <p information type>(FALSE)
@parm void*	| db_p							| checked when <p byDbPtr> is TRUE. remove DNS entry by this pointer.
@parm char*	| domain							| checked when <p byDbPtr> is FALSE. domain of entry to delete.
@parm void*	| procData						| checked when <p byDbPtr> is FALSE. user's data pointer of entry to delete.
@rvalue SUCCESS								| Entry deletion success.
@rvalue FAILED								| Entry deletion failed.
@comm
	Delete entry for <p domain> / <p procData> or <p db_p> in the <p process> sequence from DNS system.
	If <p byDbPtr> is TRUE, system only delete entry pointer by <p db_p>. This pointer can be gotten in <p rtl8651_dns_addProcessEntry()>'s
	parameter <p entryPtr>.
	If <p byDbPtr> is FALSE, system will delete ALL entris which domain is <p domain> and user-defined pointer is <p procData>.
	Note that the <p destructor> of the process would be called to release <p procData>.
*/
int32 rtl8651_dns_delProcessEntry(
	uint8 process,
	uint8 byDbPtr,						/* delete process by DNS-DB's pointer (gotten from _rtl8651_dns_addProcessEntry) */
	void *db_p,
	char *domain,
	void *procData)
{
	int32 retval;

	rtlglue_drvMutexLock();
	retval = _rtl8651_dns_delProcessEntry(	process,
											byDbPtr,
											db_p,
											domain,
											procData);
	rtlglue_drvMutexUnlock();
	return retval;
}


int32 _rtl8651_dns_delProcessEntry(
	uint8 process,
	uint8 byDbPtr,						/* delete process by DNS-DB's pointer (gotten from _rtl8651_dns_addProcessEntry) */
	void *db_p,
	char *domain,
	void *procData)
{
	_rtl8651_dnsDB_t * dnsDb_p = NULL;
	int32 idx;
	uint8 domainCmp[RTL8651_DNS_DOMAIN_MAXLENGTH];

	if (process >= _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		RTL_DNS_ERR("Invalid process %d", process);
		goto err_out;
	}

	if (byDbPtr == TRUE)
	{
		if (db_p == NULL)
		{
			RTL_DNS_ERR("delete null entry.");
			goto err_out;
		}
		/* DELETE entry */
		if (__rtl8651_dns_destroyDnsDB(process, (_rtl8651_dnsDB_t *)db_p) != SUCCESS)
		{
			goto err_out;
		}
	} else
	{
		uint32 domainLen = 0;

		domainLen = strlen(domain);

		if ((domainLen >= RTL8651_DNS_DOMAIN_MAXLENGTH) || (domainLen == 0))
		{
			RTL_DNS_ERR("Domain name (%s) length is invalid: %d", domain, domainLen);
			goto err_out;
		}

		/* we would translate upper-case-alphabets to lower-case */
		memset(domainCmp, 0, sizeof(domainCmp));
		__strlowerncpy(domainCmp, domain, strlen(domain));

		/* DELETE entry */
		dnsDb_p = NULL;
		while ((dnsDb_p = __rtl8651_dns_getDnsDB(	process,
													domain,
													dnsDb_p,
													TRUE)) != NULL)
		{
			if (procData && (dnsDb_p->procData != procData))
				continue;

			if (__rtl8651_dns_destroyDnsDB(process, dnsDb_p) != SUCCESS)
			{
				goto err_out;
			}
		}
	}

	/*
		After entry deletion, we check if there is any entry in DNS system, if there is no any
		entry in DNS system, we don't trap DNS packet anymore.
	*/
	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx])	break;
	}
	if (idx == _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		_rtl8651_arrangeAllACL();
	}

	return SUCCESS;

err_out:
	return FAILED;
}

/*
@func int32	| rtl8651_dns_findProcessEntry		| Find entry in DNS system
@parm uint8	| process							| process sequence to find entry
@parm void*	| procData						| user's data pointer of entry to find (if NULL, means don't care).
@parm char*	| domain							| domain of entry to find.
@parm char**	| returnedDomain					| when entry is found, point to it's really domain.
@parm void**	| returnedProcData				| when entry is found, point to it's really process-owned data entry.
@parm void**	| start							| starting entry to find (if NULL, search starts from first entry).
@rvalue SUCCESS								| Entry found.
@rvalue FAILED								| Entry not found.
@comm
	Find entry using <p domain> ( and <p procData> ) in the <p process> sequence from DNS system.
	DNS system would use the match type (KEYWORD or EXACTLY match) specified when process registration to find entry
	which fit condition. If <p procData> is NULL, system would only use <p domain> to find entry.

	When entry is found in DNS system, the really domain and procData would saved
	in <p returnedDomain> and <p returnedProcData> for further comparison.

	The <p start> is used for continueous search.
	For example, if user want to find all entris fit a condition(ex. domain is 'abc.com'). He can use the following code to do this:

		======================================================
		void *ptr = NULL;

		while (rtl8651_dns_findProcessEntry(	process,
										"abc.com",
										NULL,
										&origDomain,
										&origData,
										&ptr) == TRUE)
		{
			...... (anything).....
		}
		======================================================

*/
int32 rtl8651_dns_findProcessEntry(
	uint8 process,
	char *domain,
	void *procData,
	char **returnedDomain,
	void **returnedProcData,
	void **start)
{
	int32 retval;

	rtlglue_drvMutexLock();
	retval = _rtl8651_dns_findProcessEntry(	process,
											domain,
											procData,
											returnedDomain,
											returnedProcData,
											start);
	rtlglue_drvMutexUnlock();
	return retval;
}

int32 _rtl8651_dns_findProcessEntry(
	uint8 process,
	char *domain,
	void *procData,
	char **returnedDomain,
	void **returnedProcData,
	void **start)
{
	_rtl8651_dnsDB_t * dnsDb_p = NULL;

	if (process >= _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt)
	{
		RTL_DNS_ERR("Invalid process %d", process);
		return FALSE;
	}

	dnsDb_p = start?(*start):NULL;

	while ((dnsDb_p = __rtl8651_dns_getDnsDB(	process,
												domain,
												dnsDb_p,
												TRUE)) != NULL)
	{	
		if (procData && (procData != dnsDb_p->procData))
			continue;

		break;	/* found !*/
	}

	if (start)
	{
		*start = dnsDb_p;
	}
	if (returnedProcData)
	{
		*returnedProcData = dnsDb_p?dnsDb_p->procData:NULL;
	}
	if (returnedDomain)
	{
		*returnedDomain = dnsDb_p?dnsDb_p->domain:NULL;
	}
	return (dnsDb_p)?TRUE:FALSE;
}

/*
@func char*	| rtl8651_dns_getProcessEntryDomainInfo	| Return the Domain of given DNS Database entry pointer
@parm void*	| db_p								| DNS database entry's pointer
@rvalue (NOT NULL)								| pointer of domain string of the entry.
@rvalue NULL										| pointer is INVALID (NULL) or domain doest not exist.
@comm
	Return the domain string of given DNS system entry.
	This pointer can be gotten in <p rtl8651_dns_addProcessEntry()>'s parameter <p entryPtr>.
*/

char *rtl8651_dns_getProcessEntryDomainInfo(void *db_p)
{
	char *ret;

	rtlglue_drvMutexLock();
	ret = _rtl8651_dns_getProcessEntryDomainInfo(db_p);
	rtlglue_drvMutexUnlock();
	return ret;
}

char *_rtl8651_dns_getProcessEntryDomainInfo(void *db_p)
{
	if (db_p)
	{
		return ((_rtl8651_dnsDB_t*)db_p)->domain;
	}

	return NULL;
}

/* ==========================================================================
  *
  *				DNS packet processor
  *
  *					1. Need to handle all DNS related proces
  *						* Trap DNS packet to CPU
  *						* parse DNS packet to get informations
  *
  *					2. Provide ALL (hostname) <--> (ip) mappings from DNS packets.
  *						* Need to handle CNAME
  *
  * ========================================================================== */
#define RTL8651_DNS_PKTPROC

/*
@func void			| rtl8651_dns_dnsProc_dump | Dump DNS packet processing system's information
@comm
	Dump information in DNS packet processing system.
	Both packet process information and CNAME mapping informaton are included.
*/
void rtl8651_dns_dnsProc_dump(void)
{
	_rtl8651_dns_cnameMapping_t *ptr, *hdr;
	uint32 inusedCnt = 0, freeCnt = 0;

	rtlglue_drvMutexLock();

	printk("\nDNS Process module:\n\n");

	printk("\tTTL Suppression:\t%s\n", _rtl8651_dns_enableTTLSuppression==TRUE?"TRUE":"FALSE");
	if ( _rtl8651_dns_enableTTLSuppression==TRUE )
	{
		printk("\tTTL upper bound:\t%d\n", _rtl8651_dns_suppressedTTL);
	}
	printk("\n");


	ptr = _rtl8651_dns_cnameLru_h;
	while ( ptr )
	{
		inusedCnt ++;
		ptr = ptr->lru_next;
	}
	ptr = _rtl8651_dns_cnameFree_p;
	while ( ptr )
	{
		freeCnt ++;
		ptr = ptr->next;
	}

	printk("\nDNS CNAME mapping table:\n\n");
	printk("\tInused:\t\t%d\n", inusedCnt);
	printk("\tFree:\t\t%d\n", freeCnt);
	printk("\tTIMEOUT:\t%d\n", _rtl8651_dns_cnameTimeout);
	if (inusedCnt > 0)
	{
		int32 idx = 1;
		printk("\n\t[NO]\t[Timeout]\t[DOMAIN NAME]\n");
		printk("\t============================================================\n");
		hdr = _rtl8651_dns_cnameMap_p;
		while ( hdr )
		{
			printk("\t( %d )\n", idx ++);
			ptr = hdr;
			while ( ptr )
			{
				printk("\t\t%d\t\t%s\n", 	(ptr->updateTime + _rtl8651_dns_cnameTimeout) - _sysUpSeconds,
										ptr->addr);
				ptr = ptr->next;
			}
			hdr = hdr->hdr_next;
		}
	}
	printk("\t============================================================\n");

	rtlglue_drvMutexUnlock();
}


static int32 _rtl8651_dns_cnameMapping_init(rtl8651_fwdEngineInitPara_t *para)
{

	assert(para);

	_rtl8651_dns_cnameMap_p = NULL;
	_rtl8651_dns_cnameLru_h = NULL;
	_rtl8651_dns_cnameLru_t = NULL;
	_rtl8651_dns_cnameFree_p = NULL;
	_rtl8651_dns_cnameTimeout = 0;

	if (para->dns_dnsDB_cnameMapTblSize == 0)
	{
		RTL_DNS_WARN(	"** Disable CNAME mapping: para->dns_dnsDB_tblSize = %d",
						para->dns_dnsDB_tblSize);
		return SUCCESS;
	}

	MEM_ALLOC(_rtl8651_dns_cnameFree_p, para->dns_dnsDB_cnameMapTblSize * sizeof(_rtl8651_dns_cnameMapping_t), FAILED);
	DNS_LIST_INIT(_rtl8651_dns_cnameFree_p, _rtl8651_dns_cnameMapping_t, para->dns_dnsDB_cnameMapTblSize, next);
 
	/* set timeout */
	_rtl8651_dns_cnameTimeout = para->dns_dnsDB_cnameMapTimeout;

	return SUCCESS;
}

static void _rtl8651_dns_cnameMapping_reinit(void)
{
	while (_rtl8651_dns_cnameMap_p)
	{
		_rtl8651_dns_cnameMapping_del(_rtl8651_dns_cnameMap_p);
	}
}

static _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameMapping_find(char *name)
{
	_rtl8651_dns_cnameMapping_t *ptr, *hdr;
	uint32 nameLen = strlen(name);

	hdr = _rtl8651_dns_cnameMap_p;

	while (hdr)
	{
		ptr = hdr;

		while (ptr)
		{
			if (strlen(ptr->addr) != nameLen)
			{
				goto next;
			}

			if (strncmp(ptr->addr, name, nameLen) == 0)
			{
				return ptr;
			}

		next:
			ptr = ptr->next;
		}

		hdr = hdr->hdr_next;
	}

	return NULL;
}

static inline _rtl8651_dns_cnameMapping_t *_rtl8651_dns_cnameMapping_getEntry(void)
{
	_rtl8651_dns_cnameMapping_t *entry = NULL;

	if (_rtl8651_dns_cnameFree_p)
	{	/* get from free list */
		entry = _rtl8651_dns_cnameFree_p;
		_rtl8651_dns_cnameFree_p = _rtl8651_dns_cnameFree_p->next;
	} else
	{	/* get from LRU list */
		if (_rtl8651_dns_cnameLru_h)
		{
			_rtl8651_dns_cnameMapping_del(_rtl8651_dns_cnameLru_h);
			assert(entry);
		}
	}

	/* init entry before return it */
	if (entry)
	{
		memset(entry, 0, sizeof(_rtl8651_dns_cnameMapping_t));
	}

	return entry;
}

static void _rtl8651_dns_cnameMapping_refresh(_rtl8651_dns_cnameMapping_t *entry)
{
	entry->updateTime = _sysUpSeconds;

	if (_rtl8651_dns_cnameLru_t != entry)
	{
		/* remove from LRU list */
		if (entry->lru_next)
		{
			entry->lru_next->lru_prev = entry->lru_prev;
		}
		if (entry->lru_prev)
		{
			entry->lru_prev->lru_next = entry->lru_next;
		}
		if (_rtl8651_dns_cnameLru_h == entry)
		{
			_rtl8651_dns_cnameLru_h = entry->lru_next;
		}

		entry->lru_next = NULL;
		entry->lru_prev = NULL;

		/* move back to the tail of LRU list */
		if (_rtl8651_dns_cnameLru_t)
		{
			_rtl8651_dns_cnameLru_t->lru_next = entry;
			entry->lru_prev = _rtl8651_dns_cnameLru_t;
		} else
		{
			assert(_rtl8651_dns_cnameLru_h == NULL);
			_rtl8651_dns_cnameLru_h = entry;
		}

		_rtl8651_dns_cnameLru_t = entry;
	}
}

/*
	Add CNAME mapping entry(s) for 'name1' and 'name2' .

	NOTE: if entry for 'name1' or(and) 'name2' exist(s), it's timer would be refreshed.
*/
static int32 _rtl8651_dns_cnameMapping_add(char *name1, char *name2)
{
	_rtl8651_dns_cnameMapping_t *name1_p, *name2_p, *exist_p, *new_p;

	name1_p = _rtl8651_dns_cnameMapping_find(name1);
	name2_p = _rtl8651_dns_cnameMapping_find(name2);

	if ((name1_p) && (name2_p))
	{
		/* check if name1 and name2 are in the same group */
		_rtl8651_dns_cnameMapping_t *first, *ptr;
		uint8 name1_ok, name2_ok;

		/* find from the first entry of this group */
		first = name1_p;
		while (first->prev)
		{
			first = first->prev;
		}

		ptr = first;
		name1_ok = FALSE;
		name2_ok = FALSE;
		while (ptr)
		{
			if (ptr == name1_p)
			{
				name1_ok = TRUE;
			}
			if (ptr == name2_p)
			{
				name2_ok = TRUE;
			}
			ptr = ptr->next;
		}

		if ((name1_ok != TRUE) || (name2_ok != TRUE))
		{	/* name 1 and name 2 are NOT in the same group ! remove both of them and add an new group for them */
			_rtl8651_dns_cnameMapping_del(name1_p);
			_rtl8651_dns_cnameMapping_del(name2_p);
			goto new_group;
		}

		/* name 1 and name 2 are in the same group, we simply update their time record and refresh LRU position */
		_rtl8651_dns_cnameMapping_refresh(name1_p);
		_rtl8651_dns_cnameMapping_refresh(name2_p);

	} else if (name1_p || name2_p)
	{
		if ((new_p = _rtl8651_dns_cnameMapping_getEntry()) == NULL)
		{
			goto out;
		}
		if (name1_p)
		{
			exist_p = name1_p;
			strncpy(new_p->addr, name2, RTL8651_DNS_DOMAIN_MAXLENGTH);
		} else
		{
			exist_p = name2_p;
			strncpy(new_p->addr, name1, RTL8651_DNS_DOMAIN_MAXLENGTH);
		}

		goto link_to_exist;

	} else
	{	/* create a new relationship for name1 and name2 */
new_group:
		/* allocate and fill entry for NAME 1 and NAME 2 , let NAME 1's entry as exist one */
		if ((new_p = _rtl8651_dns_cnameMapping_getEntry()) == NULL)
		{
			goto out;
		}
		if ((exist_p = _rtl8651_dns_cnameMapping_getEntry()) == NULL)
		{
			_rtl8651_dns_cnameMapping_del(new_p);
			goto out;
		}

		strncpy(exist_p->addr, name1, RTL8651_DNS_DOMAIN_MAXLENGTH);

		strncpy(new_p->addr, name2, RTL8651_DNS_DOMAIN_MAXLENGTH);

		/* put 'exist_p' in CNAME-MAPPING table as a new group */
		if (_rtl8651_dns_cnameMap_p)
		{
			_rtl8651_dns_cnameMap_p->hdr_prev = exist_p;
			exist_p->hdr_next = _rtl8651_dns_cnameMap_p;
		}
		_rtl8651_dns_cnameMap_p = exist_p;

link_to_exist:
		/* put the "new_p" into the group list of "exist_p" and refresh both of them */
		if (exist_p->next)
		{
			exist_p->next->prev =new_p;
		}
		new_p->next = exist_p->next;
		new_p->prev = exist_p;
		exist_p->next = new_p;

		_rtl8651_dns_cnameMapping_refresh(new_p);
		_rtl8651_dns_cnameMapping_refresh(exist_p);
	}

out:
	return SUCCESS;
}

/*
	Delete the CNAME mapping entry.
*/
static void _rtl8651_dns_cnameMapping_del(_rtl8651_dns_cnameMapping_t *delEntry)
{
	_rtl8651_dns_cnameMapping_t *first = NULL, *entry = delEntry;

proc_del:

	/* before delete this entry, we get its first entry of the group after delete "entry" */
	first = entry;
	while (first->prev)
	{
		first = first->prev;
	}
	if (first == entry)
	{
		first = entry->next;
	}

	/* released from entry CNAME mapping list */
	if (entry->prev)
	{
		entry->prev->next = entry->next;
	}
	if (entry->next)
	{
		entry->next->prev = entry->prev;

		if (entry->hdr_next)
		{
			entry->next->hdr_next = entry->hdr_next;
			entry->hdr_next->hdr_prev = entry->next;
		}
		if (entry->hdr_prev)
		{
			entry->next->hdr_prev = entry->hdr_prev;
			entry->hdr_prev->hdr_next = entry->next;
		}
		if (_rtl8651_dns_cnameMap_p == entry)
		{
			_rtl8651_dns_cnameMap_p = entry->next;
		}
	} else
	{	/* no next entry */
		if (entry->hdr_next)
		{
			entry->hdr_next->hdr_prev = entry->hdr_prev;
		}
		if (entry->hdr_prev)
		{
			entry->hdr_prev->hdr_next = entry->hdr_next;
		}
		if (_rtl8651_dns_cnameMap_p == entry)
		{
			_rtl8651_dns_cnameMap_p = entry->hdr_next;
		}
	}

	/* released from entry LRU list */
	if (entry->lru_prev)
	{
		entry->lru_prev->lru_next = entry->lru_next;
	}
	if (entry->lru_next)
	{
		entry->lru_next->lru_prev = entry->lru_prev;
	}
	if (_rtl8651_dns_cnameLru_h == entry)
	{
		_rtl8651_dns_cnameLru_h = entry->lru_next;
	}
	if (_rtl8651_dns_cnameLru_t == entry)
	{
		_rtl8651_dns_cnameLru_t = entry->lru_prev;
	}

	memset(entry, 0, sizeof(_rtl8651_dns_cnameMapping_t));

	/* add to free list */
	entry->next = _rtl8651_dns_cnameFree_p;
	_rtl8651_dns_cnameFree_p = entry;

	/* If there is only ONE entry in a group, it has no any meaning to keep this group */
	if (first)
	{
		if (first->next == NULL)	/* no any entry following first entry in the group */
		{
			entry = first;
			goto proc_del;		/* kill this 'first' entry, too */
		}
	}
}

/*
	Remove the timed out CNAME mapping entry from database.
*/
static void _rtl8651_dns_cnameMapping_TimeUpdate(uint32 secPassed)
{
	while (	(_rtl8651_dns_cnameLru_h) &&
			((_rtl8651_dns_cnameLru_h->updateTime + _rtl8651_dns_cnameTimeout) < _sysUpSeconds))
	{
		_rtl8651_dns_cnameMapping_del(_rtl8651_dns_cnameLru_h);
	}
}
 /*
	Parse NAME field in DNS message and get the field length from it.

	NOTE: this function don't trace POINTER, it just simply add 2 octect and return when the pointer-type is recognized.
*/
static int32 _RTL8651_DNS_GET_NAMELEN(uint8 *name, uint32 maxLen, uint32 *len_p)
{
	uint8 *ptr = name;
	uint32 len = 0, nameLenCheck = 0;

	while (1)
	{
		if ((*((uint8*)ptr) & 0xc0) == 0xc0)
		{	/* pointer */

			/**********************************************************************

				RFC 1035 ( Chap 4.1.4 , p.30 ) :

				The pointer takes the form of a two octet sequence:

				    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				    | 1    1  |                                 OFFSET                                   |
				    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

			  **********************************************************************/

			len += 2;			/* size of POINTER */
			break;	/* ok */
		} else
		{	/* sequence of labels */
			uint32 added;

			/**********************************************************************

				RFC 1035 ( Chap 4.1.2 , p.28 ) :

				Sequence of labels, where each label consists of a length octet followed by that
				number of octets. It terminates with the zero length octet for the null label of
				the root. Note that this field may be an odd number of octets; no padding is used.

				    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				    |        Length of octets N           |                  Octet[1]               |
				    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				    |                  Octet[2]               |                  Octet[3]               |
				    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
		   	           /                                                                                            /
		   	           /                                                                                            /
		   	           /                  Octet[N]               |
				    +--+--+--+--+--+--+--+--+


				RFC 1035 ( Chap 4.1.4 , p.31 ) :

					domain name : F.ISI.ARPA

				        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				    20 |                     1                      |                     F                     |
				        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				    22 |                     3                      |                     I                      |
				        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				    24 |                     S                      |                     I                     |
				        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				    26 |                     4                      |                     A                     |
				        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				    28 |                     R                     |                      P                     |
				        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				    30 |                     A                     |                      0                     |
				        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+


			  **********************************************************************/

			added = (1 + (uint32)(*(uint8*)ptr));	/* size of LENGTH + LABELS */

			len += added;

			if (added == 1)	/* it means length of this sequence == 0, and this is the "NULL ROOT LABEL" (the last sequence of NAME) */
			{
				break;	/* ok */
			}

			/* check phase */
			{	/* We treat the added '1' (size of LENGTH) as the ASCII '.' (separator of labels) */
				nameLenCheck += added;

				/****************************************************************
					RFC 2181 "Clarifications to the DNS Specification" 

					A full domain
						name is limited to 255 octets (including the separators).  The zero
						length full name is defined as representing the root of the DNS tree,
						and is typically written and displayed as ".".

						Note that the last '.' (DNS root) is included in domain length calculation.
				  ****************************************************************/

				if (nameLenCheck > RTL8651_DNS_DOMAIN_MAXLENGTH)
				{
					RTL_DNS_ERR("Decoded Domain Length (%d) is larger than Domain name max length (%d), ignore.",
								nameLenCheck,
								RTL8651_DNS_DOMAIN_MAXLENGTH);
					return FAILED;
				}
			}

			if (len > maxLen)
			{
				RTL_DNS_WARN("Domain Name parse failed.");
				return FAILED;
			}

			ptr = (uint8*)(added + (uint32)ptr);	/* find next segment */
		}

	}

	if (len_p)
	{
		*len_p = len;
	}

	return SUCCESS;
}

/*
	Parse NAME field in DNS message and return the label-sequence from it.
*/
static int32 _RTL8651_DNS_GET_INFO(
	_rtl8651_dns_dnsDomain_t domainType,
	uint8 *outputPtr,
	uint32 maxOutputLen,
	uint8 *inputPtr,
	uint8 *dnsData_p,
	uint32 dnsDataLen)
{
	uint8 outTmp[RTL8651_DNS_DOMAIN_MAXLENGTH];
	uint8 *inPtr = inputPtr;
	uint8 *outPtr = outputPtr;
	uint32 outLen = 0;
	uint8 addDot = FALSE;

	/*
		1.
			If the domainType is IPV4, it means this domain is as the type '1.1.113.140.in-addr.arpa.'.
			In this case, we need to store the string '140.113.1.1'  into outputPtr and we store the original
			string into temporary storage first.

		2.
			If the domainType is normal domain name, we need to translate alphabets to lower-case.
			Therefore, we would store the string gotten from DNS message in temporary storage and
			translate alphabets when we copy string from temporary storage to actual output.
	*/
	outPtr = outTmp;
	addDot = FALSE;	/* for the first label sequence, we don't add DOT before it */

	while (1)
	{
		if ((*((uint8*)inPtr) & 0xc0) == 0xc0)
		{	/* pointer */
			uint16 offset;

			/* byte by byte copy, for alignment issue */
			offset = (*(uint8*)inPtr)<<8 | (*(uint8*)((uint32)inPtr + 1));
			offset &= 0x3fff;	/* get the low-14 bits */
			offset = ntohs(offset) + 0;

			if (offset > dnsDataLen)
			{
				RTL_DNS_WARN("offset : %d, dnsDataLen : %d\n", offset, dnsDataLen);
				goto err_out;
			}

			inPtr = (uint8*)((uint32)dnsData_p + offset);

		} else
		{	/* labels */
			uint8 *origInPtr, *origOutPtr;
			uint8 labelLen;

			labelLen = *((uint8*)inPtr);

			PTR_CHK_SHIFT_EXIT(	(dnsDataLen + (uint32)dnsData_p - (uint32)inPtr),
									inPtr,
									sizeof(uint8),
									err_out);	/* move to the start address of LABEL sequence */

			/* The added LABEL is 'zero root label', finish */
			if (labelLen == 0)
			{
				break;
			}

			if (addDot == TRUE)
			{
				/* we need to add '.' to output stream before copy current label to it */
				origOutPtr = outPtr;
				PTR_CHK_SHIFT_EXIT(	(maxOutputLen - outLen),
										outPtr,
										sizeof(uint8),
										err_out);

				/* update outLen after moving pointer */
				outLen += sizeof(uint8);
				/* Everything is OK, we start to copy label to output stream */
				*(char*)origOutPtr = '.';
			}

			/* we move pointer and check remaining length budget first, if anything is ok, we start to copy */
			origInPtr = inPtr;
			origOutPtr = outPtr;
			
			PTR_CHK_SHIFT_EXIT(	(dnsDataLen + (uint32)dnsData_p - (uint32)inPtr),
									inPtr,
									labelLen,
									err_out);
			PTR_CHK_SHIFT_EXIT(	(maxOutputLen - outLen),
									outPtr,
									labelLen,
									err_out);

			/* update ouput length after moving pointer */
			outLen += labelLen;

			/* Everything is OK, we start to copy label to output stream */
			memcpy(origOutPtr, origInPtr, labelLen);

			addDot = TRUE;	/* We can add '.' before next label sequence */

		}
	}

	/* set last octect of output as '\0' */
	{
		uint8 *origOutPtr = outPtr;

		PTR_CHK_SHIFT_EXIT(	(maxOutputLen - outLen),
								outPtr,
								sizeof(uint8),
								err_out);

		/* update outLen after moving pointer */
		outLen += sizeof(uint8);
		*(uint8*)origOutPtr = 0;
	}

	/* <-------------- parse finish  --------------> */
	if (domainType == _RTL8651_DNS_GET_TYPE_INET_IPV4)
	{
		/*
			We need to reverse the the output string,
			the following comment use the string '1.1.113.140.in-addr.arpa.' as example to explain it.
		*/
		uint32 separator[16];
		uint8 digitBeforeSeparator[16];
		uint32 separatorCnt, sepIdx;
		uint32 digitLen;
		uint8 *ptr;

		uint8 *actOutputPtr = outputPtr;
		uint32 remOutLen = maxOutputLen;
		uint8 *origOutPtr;

		/* 1. We need to separate the numeral part "1.1.113.140." from original string
			  and store the position of the separator '.' between each numeral.
		*/
		ptr = outTmp;
		digitLen = 0;
		memset(separator, 0, sizeof(separator));
		memset(digitBeforeSeparator, 0 ,sizeof(digitBeforeSeparator));
		separatorCnt = 0;

		while (1)
		{
			if (	(*(uint8*)ptr == 0) ||
				(digitLen > outLen))
			{
				RTL_DNS_WARN("ptr : %p, digitLen %d, outLen %d", ptr, digitLen, outLen);
				goto err_out;
			}

			if (	(!isdigit(*(char*)ptr)) &&
				(*(char*)ptr != '.'))
			{
				if (	(digitLen == 0)||
					(*(char*)(ptr-1) != '.'))
				{
					/* error situation: "1.1.113.140abc.-blah-blah" */
					RTL_DNS_WARN("digitLen : %d, char (ptr - 1) : %c -- Error string.", digitLen, *(char*)(ptr-1));
					goto err_out;
				}
				assert(*(char*)ptr == 'i');	/* "i"n-addr */
				break;
			}

			if (*(char*)ptr == '.')
			{
				if (separatorCnt >= (sizeof(separator) / sizeof(separator[0])))
				{
					RTL_DNS_WARN("separatorCount: %d , size of separator: %d , sizeof separator[0]: %d",
									separatorCnt,
									sizeof(separator),
									sizeof(separator[0]));
					goto err_out;
				}

				separator[separatorCnt] = digitLen;

				if (separatorCnt == 0)
				{
					digitBeforeSeparator[separatorCnt] = digitLen;
				} else
				{
					digitBeforeSeparator[separatorCnt] = digitLen - separator[separatorCnt - 1] - 1;
				}

				separatorCnt ++;
			}

			ptr ++;
			digitLen ++;
		}

		/*	Now, ptr points to the char 'i' in the target string,
			digitLen is the string length of "1.1.113.140." ,
			and separator[] stores all positions of '.' in "1.1.113.140.".
			==> NOTE: Last "." (after "140") is stored in separator[], too.
		*/

		/* 2. We need to get the string "140", "113", "1", "1" sequentially and store it to outputPtr */

		sepIdx = separatorCnt;
		
		while (sepIdx > 0)
		{
			uint8 *numPtr = &(outTmp[(separator[sepIdx - 1] + 1)]);

			if (numPtr == ptr)
			{	/* last '.' */
				goto nextlabel;
			}

			if (	(sepIdx >= (sizeof(digitBeforeSeparator) - sizeof(digitBeforeSeparator[0]))) ||
				(digitBeforeSeparator[sepIdx] == 0))
			{	/* no digit to copy */
				goto nextlabel;
			}

			/* copy current digit part to real output */
			origOutPtr = actOutputPtr;

			PTR_SHIFT_EXIT(	remOutLen,
								actOutputPtr,
								digitBeforeSeparator[sepIdx],
								err_out);

			memcpy(origOutPtr, numPtr, digitBeforeSeparator[sepIdx]);

			/* add "." to real output */
			origOutPtr = actOutputPtr;

			PTR_SHIFT_EXIT(	remOutLen,
								actOutputPtr,
								1,
								err_out);

			*(char*)origOutPtr = '.';

		nextlabel:
			sepIdx --;
		}

		/* Add the digit "digitBeforeSeparator[0]" to real output */
		if (digitBeforeSeparator[0] > 0)
		{
			origOutPtr = actOutputPtr;

			PTR_SHIFT_EXIT(	remOutLen,
								actOutputPtr,
								digitBeforeSeparator[0],
								err_out);

			memcpy(origOutPtr, outTmp, digitBeforeSeparator[0]);
		}

		/* set last octect of output as '\0' */
		{
			origOutPtr = actOutputPtr;

			PTR_CHK_SHIFT_EXIT(	remOutLen,
									actOutputPtr,
									sizeof(uint8),
									err_out);

			/* update outLen after moving pointer */
			outLen += sizeof(uint8);
			*(uint8*)origOutPtr = 0;
		}

	}else	/* Normal domain name */
	{	/* we copy string from temporary storage to actual output */
		__strlowerncpy(outputPtr, outTmp, outLen);
	}

	RTL_DNS_INFO("Get string [%s]", outputPtr);

	return SUCCESS;

err_out:

	RTL_DNS_WARN("process failed");
	
	return FAILED;
}


/* =====================
		DNS packet parser
    ===================== */

/*
	Parse DNS message
*/
static int32 _rtl8651_dns_parseDnsPkt(
	uint8 *dns_p,
	uint32 dnsLen,
	uint8 *pktModified,
	int32 (*dns_proc)(char *, ipaddr_t))
{
	uint8 *ptr = NULL;
	uint32 idx = 0, remLen = 0;
	_rtl8651_dns_dnsHeader_t *dnsHdr_p = NULL;
	uint16 qdcount = 0, ancount = 0, nscount = 0, arcount = 0;

	RTL_DNS_INFO("Parse DNS message (size %d)\n", dnsLen);
	RTL_DNS_INFO("*=======================================*");

	if (	(dns_p == NULL) ||
		(dns_proc == NULL) ||
		(dnsLen < sizeof(_rtl8651_dns_dnsHeader_t)))
	{
		goto err_out;
	}

	ptr = dns_p;
	remLen = dnsLen;

	/*************************************************************************************

		RFC 1035 ( Chap 4.1 , p.25 ) :
		All communications inside of domain protocol are carried in a single format called a MESSAGE,
		The top level format of MESSAGE is divided into 5 sections shown below:

						    +---------------------+
						    |             Header              |
						    +---------------------+
						    |            Question             | the question for the name server
						    +---------------------+
						    |             Answer              | RRs answering the question
						    +---------------------+
						    |           Authority              | RRs pointing toward an authority
						    +---------------------+
						    |           Additional             | RRs holding additional information
						    +---------------------+

		We would parse these sections step by step.

	   *************************************************************************************/


	/*************************************************************************************
		[HEADER]

		RFC 1035 ( 4.1.1. , p.26-p.27 ) :

		The header contains the following fields:

		                                                                1    1    1    1    1    1
		       0    1    2    3    4    5    6    7    8    9    0    1    2    3    4    5
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
		    |                                            ID                                           |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
		    |QR|      Opcode     |AA|TC|RD|RA|       Z      |      RCODE     |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
		    |                                       QDCOUNT                                     |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
		    |                                       ANCOUNT                                      |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
		    |                                       NSCOUNT                                      |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
		    |                                       ARCOUNT                                      |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

		where:

		ID			Identifier
		QR			query (0), response (1).
		OPCODE		Define the kind of this message.
		AA			Indicate if this message is sent by the AUTHORITY server for first replied query or not.
		TC			Indicate if this packet is truncated or not.
		RD			Request DNS server to pursue this query recursively
		RA			Indicate if the DNS server support recursive process or not.
		Z			reserved, must be 0.
		RCODE		return value: 0 means NO ERROR, others mean something wrong.
		QDCOUNT	# of entries in the question section.
		ANCOUNT	# of resource records in the answer section.
		NSCOUNT	# of name server resource records in the authority records section.
		ARCOUNT	# of resource records in the additional records section.

	*************************************************************************************/
	dnsHdr_p = (_rtl8651_dns_dnsHeader_t*)ptr;

	RTL_DNS_INFO("");
	RTL_DNS_INFO("Header:");
	RTL_DNS_INFO("");
	RTL_DNS_INFO("\tID:\t\t%d", ntohs(dnsHdr_p->id));
	RTL_DNS_INFO("\tQR:\t\t%s", (dnsHdr_p->qr == 0)?"QUERY":"RESPONSE");
	RTL_DNS_INFO("\tOPCODE:\t\t%d", dnsHdr_p->opcode);
	RTL_DNS_INFO("\tAA:\t\t%d", dnsHdr_p->aa);
	RTL_DNS_INFO("\tTC:\t\t%d", dnsHdr_p->tc);
	RTL_DNS_INFO("\tRD:\t\t%d", dnsHdr_p->rd);
	RTL_DNS_INFO("\tRA:\t\t%d", dnsHdr_p->ra);
	RTL_DNS_INFO("\tRCODE:\t\t%d", dnsHdr_p->rcode);
	RTL_DNS_INFO("\tQDCOUNT:\t%d", ntohs(dnsHdr_p->qdcount));
	RTL_DNS_INFO("\tANCOUNT:\t%d", ntohs(dnsHdr_p->ancount));
	RTL_DNS_INFO("\tQNSOUNT:\t%d", ntohs(dnsHdr_p->nscount));
	RTL_DNS_INFO("\tQAROUNT:\t%d", ntohs(dnsHdr_p->arcount));

	/* check fields */

	/* --> We only process response due to query don't bring any information about (hostname<-->IP) mapping */
	if (dnsHdr_p->qr != _RTL8651_DNSHDR_QR_RESPONSE)
	{
		goto ok_out;
	}

	/* --> Generally, opcode == _RTL8651_DNSHDR_OPCODE_QUERY, Need further survey to study IQUERY and STATUS, we don't filter packets using OPCODE there. */
	if (dnsHdr_p->opcode != _RTL8651_DNSHDR_OPCODE_QUERY)
	{
		RTL_DNS_INFO("opcode = %d", dnsHdr_p->opcode);
	}

	/* --> If this DNS packet is truncated, we ignore it now. */
	if (dnsHdr_p->tc == 1)
	{
		RTL_DNS_WARN("TC == 1, this DNS message is truncated");
		goto ok_out;
	}

	/* --> we only parse response which indicates NO error in query */
	if (dnsHdr_p->opcode == _RTL8651_DNSHDR_RCODE_OK)
	{
		/* The query corresponding to this response is OK, so we continuously process packet */
	} else
	{
		if ((dnsHdr_p->ancount + dnsHdr_p->nscount + dnsHdr_p->arcount) > 0)
		{	/* there are some AN/NS/AR sections for ERROR query? */
			RTL_DNS_WARN("Response with opcode (%d): ignore information -- ANCNT(%d) NSCNT(%d) ARCNT(%d)",
								dnsHdr_p->opcode,
								dnsHdr_p->ancount,
								dnsHdr_p->nscount,
								dnsHdr_p->arcount);
		}
		goto ok_out;
	}

	/* --> get QD/AN/NS/AR count */
	qdcount = dnsHdr_p->qdcount;
	ancount = dnsHdr_p->ancount;
	nscount = dnsHdr_p->nscount;
	arcount = dnsHdr_p->arcount;

	PTR_SHIFT_EXIT(remLen, ptr, sizeof(_rtl8651_dns_dnsHeader_t),  err_out);

	/*************************************************************************************
		[Question]

		RFC 1035 ( 4.1.2. , p.28 ) :

		The question section is used to carry the "QUESTION" in most queries, i.e.,
		the parameters that define what is being asked. The section contains QDCOUNT (usually 1)
		entries, each of the following format:

		                                                                1    1    1    1    1    1
		       0    1    2    3    4    5    6    7    8    9    0    1    2    3    4    5
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
		    |                                                                                           |
		    /                                        QNAME                                         /
		    /                                                                                           /
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
		    |                                        QTYPE                                          |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
		    |                                        QCLASS                                       |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

		where:

		QNAME		domain Name represented as a sequence of labels (for PTR query, this seems just like '1.1.113.140.in-addr.arpa' )
		QTYPE		indicate the type of the query
		QCLASS		indicate the class of the query.

	*************************************************************************************/

	/* we always ignore QUERY field due to the needed mappings are in RR record instead of QUERY field */
	for ( idx = 0 ; idx < qdcount ; idx ++ )
	{
		uint32 nameLen;

		/* QNAME */
		if (_RTL8651_DNS_GET_NAMELEN(ptr, remLen, &nameLen) != SUCCESS)
		{
			goto err_out;
		}

#ifdef RTL_DNS_DEBUG
		RTL_DNS_INFO("");
		RTL_DNS_INFO("QUERY:");
		RTL_DNS_INFO("");
		{
			uint16 qtype, qclass;

			qtype = PKTGET_UINT16_UNALIGNED((uint16*)((uint32)ptr+nameLen));
			qclass = PKTGET_UINT16_UNALIGNED((uint16*)((uint32)ptr+nameLen+sizeof(uint16)));

			RTL_DNS_INFO("\tQNAME Length:\t%d", nameLen);
			RTL_DNS_INFO("\tQTYPE:\t\t%d", qtype);
			RTL_DNS_INFO("\tQCLASS:\t\t%d", qclass);
		}
#endif
		PTR_SHIFT_EXIT(	remLen,
							ptr,
							(nameLen /* QNAME */ + sizeof(uint16) /* QTYPE */ + sizeof(uint16) /* QCLASS */),
							err_out);
	}

	/*************************************************************************************
		[Answer]
		[Authority]
		[Additional]

		RFC 1035 ( 4.1.3. , p.29 ) :

		The answer, authority, and additional sections all share the same format:
		a variable number of resource records (RR), where the number of records is
		specified in the corresponding count field in the header.
		Each resource record has the following format:

	*************************************************************************************/
	for ( idx = 0 ; idx < (ancount + nscount + arcount) ; idx ++ )
	{
		uint32 rrLen = 0;

		if (_rtl8651_dns_parseDnsRR(dns_p, dnsLen, ptr, remLen, &rrLen, pktModified, dns_proc) != SUCCESS)
		{
			goto err_out;
		}

		PTR_SHIFT_EXIT(remLen, ptr, rrLen,  err_out);
	}

ok_out:

	RTL_DNS_INFO("*=======================================*");
	return SUCCESS;
err_out:

	RTL_DNS_INFO("*=======================================*");
	return FAILED;
}

/*
	Parse DNS's RR record.

		3 kinds of RR are recognized:

			- A record.
			- PTR record.
			- CNAME record.

*/
static int32 _rtl8651_dns_parseDnsRR(
	uint8 *dns_p, uint32 totalLen,	/* information about DNS message: for compression mode NAME search */
	uint8 *rr_p,					/* pointer for RR message */
	uint32 maxLen,				/* max remain data length about RR parsing */
	uint32 *rrLen_p,
	uint8 *pktModified,
	int32 (*dns_proc)(char *, ipaddr_t)
	)
{
	/*************************************************************************************

		RFC 1035 ( 4.1.3. , p.29 ) :

		                                                                1    1    1    1    1    1
		       0    1    2    3    4    5    6    7    8    9    0    1    2    3    4    5
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                  |                                                                                           |
                  /                                                                                           /
                  /                                          NAME                                         /
                  |                                                                                           |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                  |                                          TYPE                                          |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                  |                                         CLASS                                        |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                  |                                           TTL                                           |
                  |                                                                                           |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                  |                                      RDLENGTH                                     |
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                  /                                         RDATA                                        /
                  /                                                                                           /
		    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

		where:

		NAME		domain Name represented as a sequence of labels (for PTR query, this seems just like '1.1.113.140.in-addr.arpa' )
		TYPE		indicate the RR type codes. This field specifies the meaning of the data in the RDATA field.
		CLASS		indicate the class of the data in the RDATA field.
		TTL			specifies the time interval (in seconds) that the RR may be cached before it should be discarded.
		RDLENGTH	specifies the length in octets of the RDATA field.
		RDATA		A variable length string of octets that describes the resource. The format of this information varies according to the TYPE and CLASS of the resource record.

	*************************************************************************************/
	_rtl8651_dns_dnsRR_fixedField_t fixFieldBody, *rrFixField_p = &fixFieldBody;
	uint32 nameLen = 0, rrLen = 0;
	uint8 *namePtr = NULL;

	char mapped_domain[RTL8651_DNS_DOMAIN_MAXLENGTH];
	char mapped_cname[RTL8651_DNS_DOMAIN_MAXLENGTH];		/* for CNAME mapping */
	ipaddr_t mapped_ip = 0;

	/* we get the NAME field length first. */
	DNS_UNLOCK_ERROR_CHECK_JMP(	_RTL8651_DNS_GET_NAMELEN(rr_p, maxLen, &nameLen),
									SUCCESS,
									err_out); 
	namePtr = rr_p;

	/* check remaining field */

	DNS_UNLOCK_COND_JMP(maxLen < (nameLen + 10), err_out);

	rtl8651_memcpy(rrFixField_p, (void*)((uint32)rr_p + nameLen), 10);

	/* check total length */
	rrLen =	nameLen +					/* name length */
			10 +							/* fixed field entry length */
			ntohs(rrFixField_p->rdlength);	/* RDATA length */

	DNS_UNLOCK_COND_JMP((rrLen > maxLen), err_out);

	RTL_DNS_INFO("");
	RTL_DNS_INFO("RR field:");
	RTL_DNS_INFO("");
	RTL_DNS_INFO("\tNAME Length:\t%d", nameLen);
	RTL_DNS_INFO("\tTYPE:\t\t%d", ntohs(rrFixField_p->type));
	RTL_DNS_INFO("\tCLASS:\t\t%d", ntohs(rrFixField_p->class));
	RTL_DNS_INFO("\tTTL:\t\t%d", ntohl(rrFixField_p->ttl));
	RTL_DNS_INFO("\tRDLEN:\t\t%d", ntohs(rrFixField_p->rdlength));

	/* type */
	switch (rrFixField_p->type)
	{
		case _RTL8651_DNSTYPE_A:
			
		case _RTL8651_DNSTYPE_CNAME:
		case _RTL8651_DNSTYPE_PTR:
			break;
		default:							/* we don't get information of other TYPE */
			goto ok_out;
	}

	/* class */
	switch (rrFixField_p->class)
	{
		case _RTL8651_DNSCLASS_IN:
			break;
		default:						/* we don't get information of other CLASS */
			goto ok_out;
	}

	/* <================ Start to parse RR and get information ================> */

	/* clear variables */
	memset(mapped_domain, 0, RTL8651_DNS_DOMAIN_MAXLENGTH);
	memset(mapped_cname, 0, RTL8651_DNS_DOMAIN_MAXLENGTH);
	mapped_ip = 0;

	switch (rrFixField_p->type)
	{
		case _RTL8651_DNSTYPE_A:
			/*
				RFC 1035 (3.4.1. A RDATA format)

				+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				|                                   ADDRESS                                          |
				+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

				where:

				ADDRESS	A 32 bit Internet address.
			*/
			rtl8651_memcpy(&mapped_ip, (void*)((uint32)namePtr + sizeof(_rtl8651_dns_dnsRR_fixedField_t)), sizeof(ipaddr_t));
			if (_RTL8651_DNS_GET_INFO(	_RTL8651_DNS_GET_TYPE_DOMAIN,
										mapped_domain,
										RTL8651_DNS_DOMAIN_MAXLENGTH,
										namePtr,
										dns_p, totalLen) != SUCCESS)	/* get Domain from NAME field */
			{
				goto err_out;
			}

			if (_rtl8651_dns_procDnsCallBack(mapped_ip, mapped_domain, dns_proc) != SUCCESS)
			{
				goto err_out;
			}

			break;
		case _RTL8651_DNSTYPE_PTR:
			/*
				RFC 1035 (3.3.12. PTR RDATA format)

				+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				/                                     PTRDNAME                                      /
				+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

				where:

				PTRDNAME	A <domain-name> which points to some location in the
							domain name space.

			*/
			if (	(_RTL8651_DNS_GET_INFO(	_RTL8651_DNS_GET_TYPE_INET_IPV4,
											mapped_cname,
											RTL8651_DNS_DOMAIN_MAXLENGTH,
											namePtr,
											dns_p, totalLen) != SUCCESS) ||/* get Text based IP address using 'mapped_cname' as temp */
				(_RTL8651_DNS_GET_INFO(	_RTL8651_DNS_GET_TYPE_DOMAIN,
											mapped_domain,
											RTL8651_DNS_DOMAIN_MAXLENGTH,
											(void*)((uint32)namePtr + sizeof(_rtl8651_dns_dnsRR_fixedField_t)),
											dns_p, totalLen) != SUCCESS)	/* get Domain from RDATA field */
				)
			{
				goto err_out;
			}

			/* get IP address and clean tmp memory */
			mapped_ip = __inet_addr(mapped_cname);
			if (_rtl8651_dns_procDnsCallBack(mapped_ip, mapped_domain, dns_proc) != SUCCESS)
			{
				goto err_out;
			}

			break;
		case _RTL8651_DNSTYPE_CNAME:
			/*
				RFC 1035 (3.3.1. CNAME RDATA format)

				+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
				/                                         CNAME                                        /
				/                                                                                           /
				+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

				where:

				CNAME	A <domain-name> which specifies the canonical or primary
						name for the owner.  The owner name is an alias.
			*/
			if (	(_RTL8651_DNS_GET_INFO(	_RTL8651_DNS_GET_TYPE_DOMAIN,
											mapped_domain,
											RTL8651_DNS_DOMAIN_MAXLENGTH,
											namePtr,
											dns_p, totalLen) != SUCCESS) ||
				(_RTL8651_DNS_GET_INFO(	_RTL8651_DNS_GET_TYPE_DOMAIN,
											mapped_cname,
											RTL8651_DNS_DOMAIN_MAXLENGTH,
											(void*)((uint32)namePtr + sizeof(_rtl8651_dns_dnsRR_fixedField_t)),
											dns_p, totalLen) != SUCCESS)	)
			{
				goto err_out;
			}

			if (_rtl8651_dns_cnameMapping_add(mapped_domain, mapped_cname) != SUCCESS)
			{
				assert(0);
				goto err_out;
			}

			break;

		default:
			goto ok_out;
	}

	/* try to suppress TTL for A/PTR/CNAME records */
	if (	(_rtl8651_dns_enableTTLSuppression == TRUE) &&
		(rrFixField_p->ttl > _rtl8651_dns_suppressedTTL)	)
	{
		_rtl8651_dns_dnsRR_fixedField_t *tmp;
		tmp = (_rtl8651_dns_dnsRR_fixedField_t *)((uint32)rr_p + nameLen);
		PKTSET_UINT32_UNALIGNED(_rtl8651_dns_suppressedTTL, &(tmp->ttl));

		assert(pktModified);
		*pktModified = TRUE;		/* indicate this packet is modified */
	}
ok_out:

	/* return the RR's length back to caller */
	assert(rrLen);

	if (rrLen_p)
	{
		*rrLen_p = rrLen;
	}

	return SUCCESS;
err_out:
	return FAILED;
}

/*
	For all (domain name) <--> (ip) mapping, this function would call the 'dns_proc' callback function to handle this mapping.
	Note that is function can handle mappings using CNAME information.
*/
static int32 _rtl8651_dns_procDnsCallBack(
	ipaddr_t mapped_ip,
	char *mapped_domain,
	int32 (*dns_proc)(char *, ipaddr_t)
	)
{
	int32 retval;
	_rtl8651_dns_cnameMapping_t *ptr;

	/* check mapped IP/domain */
	if (mapped_ip == 0)		/* maybe mapped_ip gotten from DNS message is canonical */
	{
		RTL_DNS_INFO("mapping ignored: [%s]:(%u.%u.%u.%u).", mapped_domain, NIPQUAD(mapped_ip));
		return SUCCESS;
	}

	/* call this callback function for current mapping */
	if ((retval = dns_proc(mapped_domain, mapped_ip)) != SUCCESS)
	{
		return retval;
	}
	/* trace CNAME mapping table to find out all of matched entries */
	ptr = _rtl8651_dns_cnameMapping_find(mapped_domain);

	if (ptr)
	{
		_rtl8651_dns_cnameMapping_t *trace = ptr;

		/* get the header of this cname mapping entry */
		while (trace->prev)
		{
			trace = trace->prev;
		}

		while (trace)
		{
			if (trace != ptr)	/* other "synonymous" DOMAIN name is found ! */
			{
				if ((retval = dns_proc(trace->addr, mapped_ip)) != SUCCESS)
				{
					return retval;
				}
			}
			trace = trace->next;
		}
	}

	return SUCCESS;
}

/* =========================================================================
  *
  *
  *					Glue functions with TBLDRV
  *
  *
  * ========================================================================= */

#define RTL8651_DNS_TBLDRV_GLUE

#define SET_DNSACL_MATCHLAYER(rule) \
	rule.pktOpApp = RTL8651_ACLTBL_L3_AND_L4;

/*
	If DNS MODULE is enabled, it will add ACL rule to trap all packets with SPORT == 53 (dns response) to CPU
	and process it via DNS packet parsing module.
*/
uint32 _rtl8651_tryToAddAclRuleForDns(uint32 aclStart, uint32 aclEnd, int8 *enoughFlag)
{
	int32 idx, maxIdx;
	_rtl8651_tblDrvAclRule_t rule;
	uint32 addCount = 0, aclIdx = aclStart;
	uint32 trapDns = FALSE;

	/* if there is any entry is waiting for DNS message, we trap all DNS packets */
	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.rtl8651_inusedDB[idx])
		{
			trapDns = TRUE;
			break;
		}
	}

	/* We don't check DNS packet */
	if (trapDns == FALSE)
	{
		goto out;
	}

#ifdef ENABLE_TCP_DNS
	maxIdx = 2;
#else
	maxIdx = 1;
#endif

	for (idx = 0 ; idx < maxIdx ; idx ++)
	{
		if (aclIdx > aclEnd) {
			*enoughFlag = 0;	/* Not enough space for adding IPSec ACL rules. */
			goto out;
		}

		bzero(&rule, sizeof(_rtl8651_tblDrvAclRule_t));

		rule.actionType_				= RTL8651_ACL_CPU;

		if ( idx == 0 )
		{
			/* we trap packet with "SPORT==53" (response from DNS server) */
			rule.ruleType_				= RTL8651_ACL_UDP;
			rule.udpSrcPortUB_		= 53;
			rule.udpSrcPortLB_			= 53;
			rule.udpDstPortUB_		= 0xffff;
			rule.udpDstPortLB_			= 0;
		} else
		{
#ifdef ENABLE_TCP_DNS
			/* we trap packet with "SPORT==53" (response from DNS server) */
			rule.ruleType_				= RTL8651_ACL_TCP;
			rule.tcpSrcPortUB_			= 53;
			rule.tcpSrcPortLB_			= 53;
			rule.tcpDstPortUB_			= 0xffff;
			rule.tcpDstPortLB_			= 0;
#endif
		}

		SET_DNSACL_MATCHLAYER(rule);
		rtl8651_setAsicAclRule(aclIdx, &rule);
		_rtl8651_registerACLDB( aclIdx, _rtl8651_dns_procDnsPkt, NULL );

		addCount ++;
		aclIdx ++;
	}

out:
	return addCount;
}

/*
  *	Init function of DNS system
  *
  */
int32 _rtl8651_dns_init(rtl8651_fwdEngineInitPara_t *para)
{
	RTL_DNS_INFO("Init DNS analysis module of ROMEDRV.");

	DNS_UNLOCK_ERROR_CHECK(_rtl8651_dns_dataBase_init(para), SUCCESS, TRUE, FAILED);
	DNS_UNLOCK_ERROR_CHECK(_rtl8651_dns_cnameMapping_init(para), SUCCESS, TRUE, FAILED);

	return SUCCESS;
}

/*
  *	Re-Init function of DNS system
  *
  */
void _rtl8651_dns_reinit(void)
{
	int32 idx;

	RTL_DNS_INFO("Re-init DNS analysis module of ROMEDRV.");

	/* before flush databases, system call reinit function for each processes */
	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.callBack[idx].reinit)
		{
			_rtl8651_dns_ctrlBlock.callBack[idx].reinit();
		}
	}

	_rtl8651_dns_dataBase_reinit();
	_rtl8651_dns_cnameMapping_reinit();
}

void _rtl8651_dns_timeUpdate(uint32 secPassed)
{
	int32 idx;

	/* run time update for each function */
	for ( idx = 0 ; idx < _rtl8651_dns_ctrlBlock.initPara.dns_dnsDB_processCnt ; idx ++ )
	{
		if (_rtl8651_dns_ctrlBlock.callBack[idx].timeUpdate)
		{
			_rtl8651_dns_ctrlBlock.callBack[idx].timeUpdate(secPassed);
		}
	}

	_rtl8651_dns_cnameMapping_TimeUpdate(secPassed);
}

/* =========================================================================
  *
  *
  *					Glue functions with FWDENG
  *
  *
  * ========================================================================= */

#define RTL8651_DNS_FWDENG_GLUE

/*
	Callback function to FWDENG.
	
	If packet with SPORT == 53 is trapped, forwarding engine will call this function to process DNS packets
*/
int32 _rtl8651_dns_procDnsPkt(int32 ruleNo, struct rtl_pktHdr *pktHdr, struct ip *pip, void *usrDefined)
{
	uint8 *ptr;
	uint32 dataLen;
	uint8 pktModified = FALSE;
	struct tcphdr *tc = NULL;
	struct udphdr *ud = NULL;

	RTL_DNS_INFO("Proc DNS packet trapped by ACL rule (%d)", ruleNo);

	if (IS_FRAGMENT_REMAIN(pip->ip_off))
	{
		goto out;
	}

	switch (pip->ip_p)
	{
		case IPPROTO_TCP:
		{
			tc = (struct tcphdr *)L4HDR_ADDR(pip);
			ptr = (uint8*)TCPDATA_ADDR(tc);			
			break;
		}
		case IPPROTO_UDP:
		{
			ud = (struct udphdr *)L4HDR_ADDR(pip);
			ptr = (uint8*)((uint32)ud + sizeof(struct udphdr));
			break;
		}
		default:
			goto out;
	}

	/* calculate the L4 payload length */
	dataLen = ntohs(pip->ip_len) - ((uint32)ptr - (uint32)pip);

	if (_rtl8651_dns_parseDnsPkt(	ptr,
									dataLen,
									&pktModified,
									_rtl8651_dns_funcCallBack) != SUCCESS)
	{
		RTL_DNS_WARN("_rtl8651_dns_parseDnsPkt return FAILED.");
	}

	if (pktModified == TRUE)
	{
		switch(pip->ip_p)
		{
			case IPPROTO_TCP:
				tc->th_sum = 0;
				tc->th_sum = _rtl8651_tcpChecksum(pip);
				break;
			case IPPROTO_UDP:
				if (ud->uh_sum)
				{
					ud->uh_sum = 0;
					ud->uh_sum = _rtl8651_tcpChecksum(pip);
				}
				break;
		}
	}

out:
	return FAILED;	/* indicate this packet need be continueously processed */
}

