#include "types.h"
#include "rtl_glue.h"
#include "mbuf.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_timer.h"
#include "rtl8651_dos.h"
#include "rtl8651_dos_local.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_ipQueue.h"
#include "rtl8651_ipQueue_local.h"

// hash library
#include "rtl8651_jhash.h"

/* IP flags. */
#ifdef RTL865X_DEBUG
static int IpReasmReqds;
static int IpReasmFails;
static int IpReasmTimeout;
static int IpReasmOKs;
#define IP_INC_STATS_BH(x) x++
#define DLOG(fmt,args...) rtlglue_printf(fmt,#args)
#else
#define DLOG(fmt,args...) 
#define IP_INC_STATS_BH(x)
#endif

/* Hash table. */
#define IPQ_HASHSZ	64
/* Per-bucket lock is easy to add now. */
static struct ipq *ipq_hash[IPQ_HASHSZ];
static uint32 ipfrag_hash_rnd=31;

/* negative ipq list */
static struct ipq *negative_ipq[IPQ_HASHSZ];

/* positive ipq list */
static struct ipq *positive_ipq[IPQ_HASHSZ];

/*
	Because timer will starved and expiration will not occur if system is under heavy load,
	so we must maintain a lru (least recently used) list of ipq, too.

	When Queue system is full, we must trace lru list and drop packets!
*/
static struct ipq *lru_list = NULL;			/* for incomplete ipq list */
static struct ipq *lru_negative_list = NULL;	/* for negative ipq list */
static struct ipq *lru_positive_list = NULL;	/* for positive ipq list */

//static struct rtl8651_timer_list ipfrag_secret_timer;
// rupert int sysctl_ipfrag_secret_interval = 10 * 60 * HZ;

// Variables
static uint32 sysctl_ipfrag_time = RTL8651_DefaultDrvMaxFragTimeOut;				/* fragment timeout */
static uint32 negative_list_time = RTL8651_DefaultDrvMaxNegativeListTimeOut;			/* negative entry existing time */
static uint32 rtl8651_drvMaxNegativeListCnt = RTL8651_DefaultDrvMaxNegativeListCnt;	/* Max negative entry count in queue system */
static uint32 rtl8651_drvMaxPositiveListCnt = RTL8651_DefaultDrvMaxPositiveListCnt;	/* Max positive entry count in queue system */
static uint32 rtl8651_drvMaxFragSubPktCnt = RTL8651_DefaultDrvMaxFragSubPktCnt;	/* Max frame count of each fragment packet */
static uint32 rtl8651_drvMaxFragPoolCnt = RTL8651_DefaultDrvMaxFragPoolCnt;		/* Max fragment Pool count */
static uint32 rtl8651_drvMaxFragPktCnt = RTL8651_DefaultDrvMaxFragPktCnt;			/* Max packet count */
static uint32 negative_list_cnt = 0;												/* negative entry count */
static uint32 positive_list_cnt = 0;												/* positive entry count */
static uint32 frag_flow_cnt = 0;													/* fragment flow count in queue system */
static uint32 frag_pkt_cnt = 0;													/* total packet queued in fragment packet */
static uint32 ip_frag_mem = 0;													/* Memory used for fragments */

static uint8 system_enable = TRUE;
/*
	For static IP_Queue System, we maintain static array for ipq allocation
*/
struct ipq *ipq_pool;
static struct ipq* freelist;

// variables defined in rtl8651_tblDrvfwd.c for UDP Blocking and Queue system interaction with fwdEngine
extern uint16				*_rtl8651_udpSizeThreshValue;
extern uint8				_rtl8651_fragQueueProcess;
extern uint8				_rtl8651_fragQueueDropChain;
extern int8				rtl8651_drvIcmpRoutingMsgEnable;
#ifdef _RTL_NAPT_REDIRECT
extern ipaddr_t			_rtl8651_fragQueueNaptRedirect_MatchedIP;
#endif /* _RTL_NAPT_REDIRECT */

#define FRAG_CB(pktHdrPtr)	((uint16)(pktHdrPtr->ph_fragOffset))


/* Describe an entry in the "incomplete packets" queue. */
struct ipq {
	struct ipq		*next;				/* linked list pointers */
	struct ipq		**pprev;			/* linked list pointers */
	uint32			saddr;
	uint32			daddr;
	uint16			id;
	uint8			protocol;
	uint8			last_in;
	#define LAST_IN			0x01		/* Last packet has arrived */
	#define FIRST_IN			0x02		/* First packet has arrived */
	#define NEGATIVE_LIST	0x04		/* This entry in negative list */
	#define POSITIVE_LIST	0x08		/* This entry in positive list */
	struct rtl_pktHdr 	*fragments;			/* linked list of received fragments */
	int				subPktCnt;			/* sub-packet count of this fragment packet */
	int				len;					/* total length of original datagram */
	int				meat;
#if 0
	int				refcnt;
#endif
	struct rtl8651_timer_list	timer;				/* when will this queue expire? */

	unsigned int		hash_idx;			/* index in hash table */

	/* for lru linked list */
	struct ipq *prev_lru;					/* drop iqp if queue-system is full (because timer starvation will occur when system is under high load) */
	struct ipq *next_lru;
//	int				iif;
};



/******************************************************
	declaration of functions
******************************************************/
#if 0
static void dump_ipqhash(void);
#endif
static __inline__ unsigned int ipqhashfn(uint16, uint32, uint32, uint8);
static __inline__ struct ipq *ipq_alloc(void);
static __inline__ void ipq_link(struct ipq *, unsigned int);
static __inline__ void ipq_unlink(struct ipq *);
static __inline__ void ipq_free(struct ipq *);
static __inline__ void frag_freepktMbuf(struct rtl_pktHdr *);
static __inline__  void frag_freeMbufChain(struct ipq *);
static __inline__ void negative_unlink(struct ipq *);
static __inline__ void negative_free(struct ipq *);
static void frag_delNegative(struct ipq *);
static void frag_ipq2Negative(struct ipq *);
static __inline__ void positive_unlink(struct ipq *);
static __inline__ void positive_free(struct ipq *);
static void frag_delPositive(struct ipq *);
static void frag_ipq2Positive(struct ipq *);
static void frag_destroy(struct ipq *);
static void frag_tryDelNegative(void);
static int32 frag_tryExpire(void);
static void frag_expire(unsigned long);
static struct ipq *ip_frag_create(unsigned int, struct ip *);
static struct ipq *ip_frag_find(struct ip *);
int32 ip_frag_queue(struct ipq *, struct rtl_pktHdr *,struct ip *);
static int32 _rtl8651_flushFragQueue(struct rtl_pktHdr *, struct rtl_pktHdr *);
static __inline__ void _rtl8651_IpFragQueueSetInitPara(rtl8651_fwdEngineInitPara_t *para);
static __inline__ void ipqPool_init(void);
static __inline__ void ipqPool_reinit(void);
static void * ipqPool_malloc(uint32);
static void ipqPool_free(void*);
/******************************************************
				Queue System management
******************************************************/
/*
@func int32		| rtl8651_IpFragQueueSetTimeout	| Set time related parameters of IP Queue System
@parm 		_rtl8651_IpFragQueueTimeout_t * | para | Parameters to set
@comm 
Set timeout parameters in IP-Queue System, including fragment reassembling timeout/negative list entry timeout
 */
void rtl8651_IpFragQueueSetTimeout(_rtl8651_IpFragQueueTimeout_t *para)
{
	if (!para)
		return;

	rtlglue_drvMutexLock();//Lock resource
	if (para->MaxFragTimeout)
	{
		sysctl_ipfrag_time = para->MaxFragTimeout;
	}
	if (para->MaxNegativeListTimeOut)
	{
		negative_list_time = para->MaxNegativeListTimeOut;
	}
	rtlglue_drvMutexUnlock();//UnLock resource
}
void _rtl8651_IpFragQueueSetInitPara(rtl8651_fwdEngineInitPara_t *para)
{
	if (!para)
		return;

	rtl8651_drvMaxFragPktCnt = para->MaxFragPktCnt;
	rtl8651_drvMaxFragSubPktCnt = para->MaxFragSubPktCnt;
	rtl8651_drvMaxNegativeListCnt = para->MaxNegativeListEntryCnt;
	rtl8651_drvMaxPositiveListCnt = para->MaxPositiveListEntryCnt;
	rtl8651_drvMaxFragPoolCnt = para->MaxFragPoolCnt;

	if (para->MaxFragTimeOut)
	{
		sysctl_ipfrag_time = para->MaxFragTimeOut;
	}
	if (para->MaxNegativeListTimeOut)
	{
		negative_list_time = para->MaxNegativeListTimeOut;
	}

	if ((!rtl8651_drvMaxFragPktCnt) ||
		((!rtl8651_drvMaxFragSubPktCnt) && (!rtl8651_drvMaxFragPoolCnt)))
	{
		system_enable = FALSE;
	}
}
/*
@func int32		| rtl8651_IpFragQueueReinit	| Re-initialize IP Queue System
@comm
Always return SUCCESS
Re-initialize IP-Queue System, release all queued mbuf, reset all entries
 */
int32 rtl8651_IpFragQueueReinit(void)
{
	int32 index;

	if (system_enable == FALSE)
	{
		return SUCCESS;
	}

	rtlglue_drvMutexLock();//Lock resource

	for (index = 0 ; index < IPQ_HASHSZ ; index ++)
	{
		// ipq_hash
		while (ipq_hash[index])
		{
			frag_destroy(ipq_hash[index]);	// destroy the first ipq in current index
		}
		//negtive ipq
		while(negative_ipq[index])
		{
			frag_delNegative(negative_ipq[index]);
		}
		// positive ipq
		while(positive_ipq[index])
		{
			frag_delPositive(positive_ipq[index]);
		}
	}

	ipqPool_reinit();

	assert(lru_list == NULL);
	assert(lru_negative_list == NULL);
	assert(lru_positive_list == NULL);
	assert(negative_list_cnt == 0);
	assert(positive_list_cnt == 0);
	assert(frag_flow_cnt == 0);
	assert(frag_pkt_cnt == 0);
	assert(ip_frag_mem == 0);

	rtlglue_drvMutexUnlock();//UnLock resource

	return SUCCESS;
}
/*
@func int32		| rtl8651_IpFragQueueInit	| Initialize IP Queue System
@parm 		rtl8651_fwdEngineInitPara_t * | para | Parameters to set
@comm 
Always return SUCCESS
Initialize IP-Queue System for fragment packet process, release all queued mbuf, reset all entries and modify parameters
 */
int32 _rtl8651_IpFragQueueInit(rtl8651_fwdEngineInitPara_t *para)
{
	if (para)
	{
		_rtl8651_IpFragQueueSetInitPara(para);
	}
	if (system_enable == FALSE)
	{
		return SUCCESS;
	}
	ipqPool_init();
	rtl8651_IpFragQueueReinit();
	return SUCCESS;
}
/******************************************************
				Hash Algorithm
*******************************************************/
#if 0
void dump_ipqhash(void)
{
	struct rtl_pktHdr *pktHdrPtr;
	struct ipq *next;
	int pktnum,i;
	DLOG(" ipqhash:\n");
	for (i=0;i<IPQ_HASHSZ;i++)
	{
		
		 next = ipq_hash[i];
		 while (next)
		{
			
			pktHdrPtr= next->fragments ;
			pktnum = 0;
			while (pktHdrPtr)
			{
				pktnum++;
				pktHdrPtr=(struct rtl_pktHdr *)pktHdrPtr->ph_nextHdr;
			}
			DLOG(" %d: %x queue pkts %d \n",i,(uint32)next,pktnum);
			next=next->next;
		}
	}
}
#endif

static __inline__ unsigned int ipqhashfn(uint16 id, uint32 saddr, uint32 daddr, uint8 prot)
{
	return jhash_3words((uint32)id << 16 | prot, saddr, daddr,
			    ipfrag_hash_rnd) & (IPQ_HASHSZ - 1);
}
/******************************************************
			ip-queue pool management
*******************************************************/
void ipqPool_init(void)
{
	rtlglue_drvMutexLock();//Lock resource

	ipq_pool = (struct ipq *)rtlglue_malloc(sizeof(struct ipq) * (rtl8651_drvMaxFragPktCnt + rtl8651_drvMaxNegativeListCnt + rtl8651_drvMaxPositiveListCnt));
	freelist = ipq_pool;

	rtlglue_drvMutexUnlock();//UnLock resource
}
void ipqPool_reinit(void)
{
	uint32 idx;
	struct ipq *ptr;

	if ((rtl8651_drvMaxFragPktCnt + rtl8651_drvMaxNegativeListCnt + rtl8651_drvMaxPositiveListCnt) == 0)
	{
		freelist = NULL;
		return;
	}

	// clear ipq pool
	memset(ipq_pool, 0, (sizeof(struct ipq) * (rtl8651_drvMaxFragPktCnt + rtl8651_drvMaxNegativeListCnt + rtl8651_drvMaxPositiveListCnt)));

	// link all ipq into free list
	freelist = ipq_pool;

	ptr = ipq_pool;

	for (idx = 0 ; idx < (rtl8651_drvMaxFragPktCnt + rtl8651_drvMaxNegativeListCnt + rtl8651_drvMaxPositiveListCnt) ; idx ++, ptr ++)
	{
		/*
			Only maintain one-directional linked list
		*/
		if (idx == ((rtl8651_drvMaxFragPktCnt + rtl8651_drvMaxNegativeListCnt + rtl8651_drvMaxPositiveListCnt) - 1))	// last one
		{
			ptr->next = NULL;
		}else
		{
			ptr->next = (ptr+1);
		}
	}
	return;
}
void * ipqPool_malloc(uint32 ipq_sz)
{
	struct ipq *ret;

	assert(ipq_sz == sizeof(struct ipq));

	if (!freelist)
	{
		return NULL;
	}

	ret = freelist;

	freelist = ret->next;

	return (void *)ret;
}
void ipqPool_free(void* ipq)
{
	memset(ipq, 0, sizeof(struct ipq));
	((struct ipq *)ipq)->next = freelist;
	freelist = ipq;
}
/******************************************************
			IP queue management
*******************************************************/
/*
	Allocate an ip-queue structure for a new fragment packet
*/
static __inline__ struct ipq *ipq_alloc(void)
{
	struct ipq *qp;

	qp = (struct ipq*)ipqPool_malloc(sizeof(struct ipq));

	if(!qp)
	{
		goto err;;
	}

	/* initial pointer related entry */
	qp->next = NULL;
	qp->pprev = NULL;

	/* update counter */
	ip_frag_mem += sizeof(struct ipq);
	frag_flow_cnt ++;

	return qp;
err:
	return NULL;
}
/*
	Add ipq to ipq list and initial timer
*/
static __inline__ void ipq_link(struct ipq *qp, unsigned int hash)
{
	struct ipq *orig;

	/* Set and Initialize the timer for this ipq. */
	_rtl8651_init_timer(&qp->timer);
	qp->timer.data = (uint32) qp;				/* pointer to queue */
	qp->timer.function =(void*)frag_expire;		/* expire function	*/
	_rtl8651_mod_timer(&qp->timer, _sysUpSeconds + sysctl_ipfrag_time);

	/* Put this new qp in the front of hash list */
	orig = ipq_hash[hash];				// original first ipq entry
	qp->next = orig;
	qp->pprev = &ipq_hash[hash];			// for first entry, it's pprev is hash table entry
	if (orig)
	{
		orig->pprev = &(qp->next);		// qp->next == qp (because "next" is the first parameter of qp)
	}
	ipq_hash[hash] = qp;

}
/*
	Remove ipq from ipq list and remove timer
*/
static __inline__ void ipq_unlink(struct ipq *qp)
{
	/* remove timer of this ipq */
	_rtl8651_del_timer(&qp->timer);

	/* remove ipq from hash list */
	if(qp->next)
	{
		qp->next->pprev = qp->pprev;
	}
	if (qp->pprev)
	{
		*qp->pprev = qp->next;
	}

	qp->next = NULL;
	qp->pprev = NULL;

	/* remove ipq from lru list */
	if (qp->next_lru)
	{
		qp->next_lru->prev_lru = qp->prev_lru;
	}
	if (qp->prev_lru)
	{
		qp->prev_lru->next_lru = qp->next_lru;
	}
	if (lru_list == qp)
	{
		lru_list = qp->next_lru;
	}
	qp->next_lru = NULL;
	qp->prev_lru = NULL;

}
/*
	Release qp => will update frag_flow_cnt here: don't use this function for Negative List management
*/
static __inline__ void ipq_free(struct ipq *qp)
{
	/* free ipq */
	ipqPool_free(qp);

	/* update counter */
	ip_frag_mem -= sizeof(struct ipq);
	frag_flow_cnt --;
}
/*
	Release pkthdr and mbuf of pkt
*/
static __inline__ void frag_freepktMbuf(struct rtl_pktHdr *pkt)
{
	frag_pkt_cnt --;
	ip_frag_mem -= pkt->ph_mbuf->m_extsize;
	pkt->ph_nextHdr = NULL;
	mBuf_freeMbufChain(pkt->ph_mbuf);
}
/*
	Release an Ip-queue, decrese ip_frag_mem and free all pkt
*/
static __inline__  void frag_freeMbufChain(struct ipq *qp)
{
	struct rtl_pktHdr *head, *fp;

	if ((head = qp->fragments)!=NULL)
	{
		for ( ; head ; )
		{
			fp = head->ph_nextHdr;
			frag_freepktMbuf(head);
			head = fp;
		}
	}
	return;
}
/*
	Release ALL pkts of ip queue and put ipq to Negative List
*/
void frag_destroy(struct ipq *qp)
{
	/* unlink it */
	ipq_unlink(qp);

	/* free related mbuf chain */
	frag_freeMbufChain(qp);

	/* Put ipq to negative list */
	frag_ipq2Negative(qp);

	/* update counter for new fragment pkts */
	frag_flow_cnt --;
}

/*
	Try to delete ipq with Event Triggering
*/
int32 frag_tryExpire(void)
{
	uint8 expireOK = FALSE;

	while (lru_list)
	{
		if (_sysUpSeconds >= lru_list->timer.expires)
		{	// expire it !
			frag_expire((unsigned long)lru_list);
			expireOK = TRUE;
		}else
		{
			break;
		}
	}

	if (expireOK)
	{
		return SUCCESS;
	}

//no_expire:
	return FAILED;
}
/*
	Oops, a fragment queue timed out.  Kill it and send an ICMP reply.
*/
void frag_expire(unsigned long arg)
{
	struct ipq *qp = (struct ipq *) arg;

 	// statistics
	IP_INC_STATS_BH(IpReasmTimeout);
	IP_INC_STATS_BH(IpReasmFails);

	// generate icmp error message
	if (rtl8651_drvIcmpRoutingMsgEnable)
	{
		// error code 1 : Fragment reassembly timeout
		_rtl8651_IcmpPktField_t icmp_info;
		struct rtl_pktHdr *pkthdr = qp->fragments;
		struct ip *pip = (struct ip *)((uint8 *)(pkthdr->ph_mbuf->m_data) + (pkthdr->ph_iphdrOffset));
		rtl8651_tblDrv_vlanTable_t *vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
		uint16 vid = vlanp->vid;

		icmp_info.ic_type = ICMP_TIMXCEED;	// Time exceed
		icmp_info.ic_code = 1;	// Fragment reassembly timeout
		icmp_info.ic_mtu = 0;

		// use the first fragment packet to generate icmp error message
		qp->fragments = pkthdr->ph_nextHdr;
		pkthdr->ph_nextHdr = NULL;
		ip_frag_mem -= pkthdr->ph_mbuf->m_extsize;
		frag_pkt_cnt --;

		if (_rtl8651_drvIcmpErrorGeneration(pkthdr, pip, &icmp_info, vid) != SUCCESS)
		{
			mBuf_freeMbufChain(pkthdr->ph_mbuf);
		}
	}

	// destroy queue
	frag_destroy(qp);
}

/*
	Add an entry to the 'ipq' queue for a newly received IP packet.
*/
struct ipq *ip_frag_create(unsigned int hash, struct ip *pip)
{
	struct ipq *qp;
	struct ipq *lru;
	struct ipq *ipq_recycle;


	// Can we add more new queue for this new fragment packet?
	if ((frag_flow_cnt >= rtl8651_drvMaxFragPktCnt) &&
		(frag_tryExpire() != SUCCESS))
	{
		goto out_err;
	}

	// Allocate space for new queue struct
	if ((qp = ipq_alloc()) == NULL)
	{
		goto out_nomem;
	}

	// Initial packet information
	qp->protocol		= pip->ip_p;
	qp->id			= ntohs(pip->ip_id);
	qp->saddr		= PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	qp->daddr		= PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
	qp->last_in		= 0;
	qp->len			= 0;
	qp->meat		= 0;
	qp->fragments	= NULL;
	qp->subPktCnt	= 0;
	qp->hash_idx		= hash;
	qp->prev_lru		= NULL;
	qp->next_lru		= NULL;

	// link this ipq into list and initial it's timer
	ipq_link(qp, hash);

	// link this ipq into the tail of lru_list
	if (!lru_list)
	{
		lru_list = qp;
	}else
	{
		lru = lru_list;
		while(lru->next_lru)
		{
			lru = lru->next_lru;
		}
		lru->next_lru = qp;
		qp->prev_lru = lru;
	}
	return qp;

out_err:
	// no. of fragment packets is too much, try to add negative list
	if ((ipq_recycle = ipqPool_malloc(sizeof(struct ipq))) != NULL)
	{
		ip_frag_mem += sizeof(struct ipq);
		goto add_negative;
	}
	
out_nomem:
	// add this packet to negative list, due to ipq allocate fail, so we try to drop first negative list entry and add this new one
	if (lru_negative_list)
	{
		ipq_recycle = lru_negative_list;
		// unlink the first negtive entry
		negative_unlink(ipq_recycle);
		negative_list_cnt --;
	}else
	{	// no mem for negative entry, goto out
		goto out;
	}

add_negative:
		// change content of this ipq structure as the new pkt
		ipq_recycle->protocol		= pip->ip_p;
		ipq_recycle->id			= ntohs(pip->ip_id);
		ipq_recycle->saddr		= PKTGET_UINT32_UNALIGNED(&pip->ip_src);
		ipq_recycle->daddr		= PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
		ipq_recycle->last_in		= 0;
		ipq_recycle->len			= 0;
		ipq_recycle->meat			= 0;
		ipq_recycle->fragments		= NULL;
		ipq_recycle->subPktCnt		= 0;
		ipq_recycle->hash_idx		= hash;
		ipq_recycle->prev_lru		= NULL;
		ipq_recycle->next_lru		= NULL;

		// add the new negative entry into negative list (negative_list_cnt will increase here)
		frag_ipq2Negative(ipq_recycle);

out:
	return NULL;
}

/******************************************************
			Positive List management
*******************************************************/
/*
	unlink positive ipq
*/
static __inline__ void positive_unlink(struct ipq *qp)
{
	/* remove ipq from positive list */
	if(qp->next)
	{
		qp->next->pprev = qp->pprev;
	}
	if (qp->pprev)
	{
		*qp->pprev = qp->next;
	}

	qp->next = NULL;
	qp->pprev = NULL;

	/* remove ipq from lru list */
	if (qp->next_lru)
	{
		qp->next_lru->prev_lru = qp->prev_lru;
	}
	if (qp->prev_lru)
	{
		qp->prev_lru->next_lru = qp->next_lru;
	}
	if (lru_positive_list == qp)
	{
		lru_positive_list = qp->next_lru;
	}
	qp->next_lru = NULL;
	qp->prev_lru = NULL;
}
/*
	free positive ipq
*/
static __inline__ void positive_free(struct ipq *qp)
{
	/* free ipq */
	ipqPool_free(qp);

	/* update counter */
	ip_frag_mem -= sizeof(struct ipq);
}
/*
	unlink and free Positive ipq
*/
void frag_delPositive(struct ipq *qp)
{
	/* unlink qp */
	positive_unlink(qp);
	/* free qp */
	positive_free(qp);
	/* update counter */
	positive_list_cnt --;
}
/*
	put ipq to positive list to allow remaining packets
*/
void frag_ipq2Positive(struct ipq *qp)
{
	unsigned int hash;
	struct ipq *orig;
	struct ipq *lru_positive;

	/* If negative list is full, release first one to contain this new one */
	if (positive_list_cnt >= rtl8651_drvMaxPositiveListCnt)
	{
		if (lru_positive_list)
		{	// release first entry (pointed by "lru_positive_list")
			frag_delPositive(lru_positive_list);
		} else
		{
			// we can NOT remove old pq, so we move it to negative entry
			frag_ipq2Negative(qp);
			goto out;
		}
	}

	qp->last_in = POSITIVE_LIST;	// this entry is in positive list;

	/* Put this new positive qp in the front of hash list */
	hash = qp->hash_idx;
	orig = positive_ipq[hash];					// original first entry
	qp->next = orig;
	qp->pprev = &positive_ipq[hash];			// for first entry, it's pprev is positive table entry
	if (orig)
	{
		orig->pprev = &(qp->next);		// qp->next == qp (because "next" is the first parameter of qp)
	}
	positive_ipq[hash] = qp;

	/* Put this new positive qp in the tail of lru_negative_list */
	qp->next_lru = NULL;
	qp->prev_lru = NULL;
	
	if (!lru_positive_list)
	{
		lru_positive_list = qp;
	}else
	{
		lru_positive = lru_positive_list;
		while(lru_positive->next_lru)
		{
			lru_positive = lru_positive->next_lru;
		}
		lru_positive->next_lru = qp;
		qp->prev_lru = lru_positive;
	}
	positive_list_cnt ++;

out:
	return;
}

/******************************************************
			Negative List management
*******************************************************/
/*
	unlink negative ipq
*/
static __inline__ void negative_unlink(struct ipq *qp)
{
	_rtl8651_del_timer(&qp->timer);

	/* remove ipq from negative list */
	if(qp->next)
	{
		qp->next->pprev = qp->pprev;
	}
	if (qp->pprev)
	{
		*qp->pprev = qp->next;
	}

	qp->next = NULL;
	qp->pprev = NULL;

	/* remove ipq from lru list */
	if (qp->next_lru)
	{
		qp->next_lru->prev_lru = qp->prev_lru;
	}
	if (qp->prev_lru)
	{
		qp->prev_lru->next_lru = qp->next_lru;
	}
	if (lru_negative_list == qp)
	{
		lru_negative_list = qp->next_lru;
	}
	qp->next_lru = NULL;
	qp->prev_lru = NULL;
}
/*
	free negative ipq
*/
static __inline__ void negative_free(struct ipq *qp)
{
	/* free ipq */
	ipqPool_free(qp);

	/* update counter */
	ip_frag_mem -= sizeof(struct ipq);
}
/*
	unlink and free Negative ipq
*/
void frag_delNegative(struct ipq *qp)
{
	/* unlink qp */
	negative_unlink(qp);
	/* free qp */
	negative_free(qp);
	/* update counter */
	negative_list_cnt --;
}
/*
	put ipq to negative list to prevent remaining illegal packets
*/
void frag_ipq2Negative(struct ipq *qp)
{
	unsigned int hash;
	struct ipq *orig;
	struct ipq *lru_negative;

	/* If negative list is full, release first one to contain this new one */
	if (negative_list_cnt >= rtl8651_drvMaxNegativeListCnt)
	{
		if (lru_negative_list)
		{	// release first entry (pointed by "lru_negative_list")
			frag_delNegative(lru_negative_list);
		} else
		{
			// we can NOT remove old pq, so we drop this newly-incoming negative entry
			negative_free(qp);
			goto out;
		}
	}

	qp->last_in = NEGATIVE_LIST;	// this entry is in negative list;

	/* Set and Initialize the timer for this ipq. */
	_rtl8651_init_timer(&qp->timer);
	qp->timer.data = (uint32) qp;						/* pointer to queue */
	qp->timer.function =(void*)frag_delNegative;		/* ipq free function	*/
	_rtl8651_mod_timer(&qp->timer, _sysUpSeconds + negative_list_time);

	/* Put this new negative qp in the front of hash list */
	hash = qp->hash_idx;
	orig = negative_ipq[hash];					// original first entry
	qp->next = orig;
	qp->pprev = &negative_ipq[hash];			// for first entry, it's pprev is negative table entry
	if (orig)
	{
		orig->pprev = &(qp->next);		// qp->next == qp (because "next" is the first parameter of qp)
	}
	negative_ipq[hash] = qp;

	/* Put this new negative qp in the tail of lru_negative_list */
	qp->next_lru = NULL;
	qp->prev_lru = NULL;
	
	if (!lru_negative_list)
	{
		lru_negative_list = qp;
	}else
	{
		lru_negative = lru_negative_list;
		while(lru_negative->next_lru)
		{
			lru_negative = lru_negative->next_lru;
		}
		lru_negative->next_lru = qp;
		qp->prev_lru = lru_negative;
	}

	negative_list_cnt ++;

out:
	return;
}

/*
	Try to delete Negative list with Event Triggering
*/
void frag_tryDelNegative(void)
{
	while (lru_negative_list)
	{
		if (_sysUpSeconds >= lru_negative_list->timer.expires)
		{	// drop!
			frag_delNegative(lru_negative_list);
		}else
		{
			goto out;
		}
	}
out:
	return;
}

/* Find the correct entry in the "incomplete datagrams" queue for
 * this IP datagram, and create new one, if nothing is found.
 */
struct ipq *ip_frag_find(struct ip *pip)
{

	uint16 id = ntohs(pip->ip_id);
	uint32 saddr = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	uint32 daddr = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
	uint8 protocol = pip->ip_p;
	unsigned int hash = ipqhashfn(id, saddr, daddr, protocol);
	struct ipq *qp;
	// try to expire negative ipq list
	frag_tryDelNegative();

	/* check for positive entry */
	if (positive_ipq[hash] == NULL)
	{
		goto check_negative;
	}

	for (qp = positive_ipq[hash]; qp; qp = qp->next)
	{
		if(qp->id == id		&&
		   qp->saddr == saddr	&&
		   qp->daddr == daddr	&&
		   qp->protocol == protocol)
		{
			return qp;	/* Positive List match: forward it! */
		}
	}

check_negative:
	/* check for negative entry */
	if (negative_ipq[hash] == NULL)
	{
		goto check_incomplete;
	}

	for (qp = negative_ipq[hash]; qp; qp = qp->next)
	{
		if(qp->id == id		&&
		   qp->saddr == saddr	&&
		   qp->daddr == daddr	&&
		   qp->protocol == protocol)
		{
 			return NULL;	/* Neagtive List matched : return NULL to drop it !!! */
		}
	}

check_incomplete:

	/* check for positive entry */
	if (ipq_hash[hash] == NULL)
	{
		goto new_one;
	}
	// find existing entry
	for (qp = ipq_hash[hash]; qp; qp = qp->next)
	{
		if(qp->id == id		&&
		   qp->saddr == saddr	&&
		   qp->daddr == daddr	&&
		   qp->protocol == protocol)
		{
 			return qp;
		}
	}

// not found, create a new entry
new_one:
	return ip_frag_create(hash, pip);
}

/*
	Add new segment to existing queue

	Return	FWDENG_QUEUED			:	packet is queued, pktHdrPtr is useless
			FWDENG_QUEUED_TO_GW	:	pktHdrPtr brings packets which must be trapped to protocol stack
			SUCCESS				:	packet is processed(dropped/send), pktHdrPtr is useless
*/
int32 ip_frag_queue(struct ipq *qp, struct rtl_pktHdr *pktHdrPtr,struct ip * ip)
{
	struct rtl_pktHdr  *prev, *next;
	int flags, offset;
	int ihl, end;
	int32 retval;
	rtl8651_tblDrv_vlanTable_t *vlanp = NULL;
	uint8 direction = 0;
	uint32 dosIgnore = FALSE;
	uint32 ipQueuePassThrough = TRUE;

	// Can we add more new queue for this new fragment packet?
	if ((rtl8651_drvMaxFragSubPktCnt > 0) &&
		(qp->subPktCnt >= rtl8651_drvMaxFragSubPktCnt))
	{
		goto err_destroy_pkt_qp;
	}
	if ((rtl8651_drvMaxFragPoolCnt > 0) &&
		(frag_pkt_cnt >= rtl8651_drvMaxFragPoolCnt))
	{
		goto err_destroy_pkt_qp;
	}

	/*************************************************
			dos check first
	*************************************************/
	if (rtl8651a_dosFragmentCheck(pktHdrPtr->ph_dsid, ip, pktHdrPtr) == SUCCESS)
	{	// dos Attack!
		goto err_dos_detected;
	}
	/*************************************************
			get packet information
	*************************************************/
 	offset = ntohs(ip->ip_off);
	flags = offset & ~IP_OFFMASK;							/* ip flags field */
	offset &= IP_OFFMASK;									/* ip offset field */
	offset <<= 3;											/* offset is in 8-byte chunks */
 	ihl = ((ip->ip_vhl & 0x0f)<< 2);							/* ip header length */
	pktHdrPtr->ph_dataLen = ntohs(ip->ip_len) - ihl;			/* ip payload length */

 	end = offset + pktHdrPtr->ph_dataLen;				/* determine the start/end of this fragment packet */

	// Is this the final fragment?
	if ((flags & IP_MF) == 0) {
		if ((end < qp->len) ||
			((qp->last_in & LAST_IN) && (end != qp->len)))
		{	// some bits beyond end, or have different end : Tear Drop attack
			goto err_TearDrop;
		}
		// set "LAST_IN" bit and length
		qp->last_in |= LAST_IN;
		qp->len = end;
	} else {
		if (end & 0x07) {
			end &= ~0x07;
		}
		if (end > qp->len) {
			if (qp->last_in & LAST_IN)
			{	// Some bits beyond end : Tear Drop attack
				goto err_TearDrop;
			}
			qp->len = end;
		}
	}

	/*********************************************************************
		Find out which fragments are in front and at the back of us
		in the chain of fragments so far.
					prev			next			enqueue position
					===================================
					  NULL		  NULL		First packet
					  NULL		!NULL		Head of queue
					!NULL		  NULL		Tail of queue
					!NULL		!NULL		between prev and next
	 *********************************************************************/
	prev = NULL;
	if (qp->fragments == NULL)
	{
		next = NULL;
	}else
	{
		for(next = qp->fragments ; next ; next = next->ph_nextHdr)
		{
			if (FRAG_CB(next) >= offset)
			{
				break;	/* bingo! */
			}
			prev = next;
		}
	}

	/*************************************************
		We found where to put this one.  Check for overlap with
		preceding fragment, and, if needed, align things so that
		any overlaps are eliminated.
	*************************************************/
	if (prev) {
		int i = (FRAG_CB(prev) + prev->ph_dataLen) - offset;

		if (i > 0) {
 			goto err_TearDrop;
 		}
	}

	if (next && (FRAG_CB(next) < end))
	{
		int i = end - FRAG_CB(next);

		if (i < next->ph_dataLen) {
			 goto err_TearDrop;
 		}
 	}

	pktHdrPtr->ph_fragOffset=offset;

	/*************************************************
		Insert this fragment in the chain of fragments
	*************************************************/
	pktHdrPtr->ph_nextHdr = next;
	if (prev)
	{
		prev->ph_nextHdr = pktHdrPtr;
	}else
	{
		qp->fragments = pktHdrPtr;
	}

	qp->meat += pktHdrPtr->ph_dataLen;
	qp->subPktCnt ++;
	ip_frag_mem += pktHdrPtr->ph_mbuf->m_extsize;
	frag_pkt_cnt ++;

	if (offset == 0)
	{
		qp->last_in |= FIRST_IN;
	}

	/*************************************************
		Check if this chain can be sent or not
	**************************************************/
	/* get direction */
	vlanp = &DrvTbl.vlan[pktHdrPtr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	direction = (vlanp->internal?0:1);

	dosIgnore= (_rtl8651_dos_enable & _rtl8651_dosIgnoreTypeCheck(direction));

	/* default : pass through packet */
	ipQueuePassThrough = TRUE;
	if (dosIgnore == TRUE)
	{
		/* if DOS System is ignored, system only queue packet when UDP Blocking is working */
		if ((qp->protocol == IPPROTO_UDP) && (_rtl8651_udpSizeThreshValue[qp->fragments->ph_dsid] != 0xffff))
			ipQueuePassThrough = FALSE;
	}else
	{
		/*
			if DOS System is enabled, system queue packet when ....
				1. Teardrop is enabled.
				2. UDP Blocking is working.
				3. UDP Bomb is enabled.
		*/
	
		if (_rtl8651_dos_sw[qp->fragments->ph_dsid] & RTL8651_DOS_TEARDROP)
			ipQueuePassThrough = FALSE;
		else if (qp->protocol == IPPROTO_UDP)
		{
			if ((_rtl8651_udpSizeThreshValue[qp->fragments->ph_dsid] != 0xffff) ||
					(_rtl8651_dos_sw[qp->fragments->ph_dsid] & RTL8651_DOS_UDPBOMB))
				ipQueuePassThrough = FALSE;
		}
	}

	/* if dos is ignored and udp blocking is not enabled, we can dequeue packets if first packet is arrived */
	if (ipQueuePassThrough == TRUE)
	{
		if (qp->last_in & FIRST_IN)
		{
			uint32 pktComplete;

			IP_INC_STATS_BH(IpReasmOKs);
			ipq_unlink(qp);	// unlink this ipq

			if ((qp->last_in == (FIRST_IN|LAST_IN)) &&
				(qp->meat == qp->len))
			{
				pktComplete = TRUE;
			}else
			{
				pktComplete = FALSE;
			}

			frag_pkt_cnt -= qp->subPktCnt;
			ip_frag_mem -= ((qp->fragments->ph_mbuf->m_extsize) * (qp->subPktCnt));

			retval = _rtl8651_flushFragQueue(qp->fragments, pktHdrPtr);

			if (retval == FAILED)
			{
				retval = SUCCESS;
				if (pktComplete == TRUE)
				{
					ipq_free(qp);
				}else
				{
					frag_ipq2Negative(qp);	// set this packet into negative list
					frag_flow_cnt --;
				}
			}else
			{
				if (pktComplete == TRUE)
				{
					ipq_free(qp);
				}else
				{
					frag_ipq2Positive(qp);		// set this packet into positive list
					frag_flow_cnt --;
				}
			}
			goto out;
		}
	} else
	{	// dos enable or udp blocking enable: we send packet only when all packets have arrived
		if ((qp->last_in == (FIRST_IN|LAST_IN)) &&
			(qp->meat == qp->len))
		{
			/******************************
				Fragment Packet dequeue
			******************************/
			IP_INC_STATS_BH(IpReasmOKs);
			
			ipq_unlink(qp);	// unlink this ipq to prevent timer-expire

			/*
				Error Checking
			*/
			if (qp->protocol == IPPROTO_UDP)
			{
				/* Check for udp blocking */
				if ((_rtl8651_udpSizeThreshValue[qp->fragments->ph_dsid] != 0xffff) &&	/* udp blocking feature is enabled */
					(qp->len > _rtl8651_udpSizeThreshValue[qp->fragments->ph_dsid]))	/* flow's size is larger than it */
				{	// we block this packet
					frag_freeMbufChain(qp);
					retval = SUCCESS;
					goto flow_done;
				}

				/* Check for udp bomb */
				{
					struct rtl_pktHdr *first = qp->fragments;
					struct ip *p;
					struct udphdr *u;

					p = (struct ip *)((uint32)(first->ph_mbuf->m_data) + (uint32)(first->ph_iphdrOffset));
					u = (struct udphdr *)((uint32)p + (uint32)((p->ip_vhl & 0x0f) << 2));

					if (rtl8651a_dosFragUdpBombCheck(first->ph_dsid, vlanp->vid, p, u, direction, qp->len, first) == SUCCESS)
					{	// attack detected
						frag_freeMbufChain(qp);
						retval = SUCCESS;
						goto flow_done;
					}
				}
			}
			{	// we try to send this packet
				frag_pkt_cnt -= qp->subPktCnt;
				ip_frag_mem -= ((qp->fragments->ph_mbuf->m_extsize) * (qp->subPktCnt));
				retval = _rtl8651_flushFragQueue(qp->fragments, pktHdrPtr);
				if (retval == FAILED)
					retval = SUCCESS;
			}
flow_done:
			ipq_free(qp);	// release this ipq
			goto out;
		}
	}

	{
		/****************************************
			Fragment Packet collection is incomplete
		****************************************/
		if ((rtl8651_drvMaxFragSubPktCnt > 0) && 
			(qp->subPktCnt >= rtl8651_drvMaxFragSubPktCnt))
		{	// PktCnt >= MaxFragSubPktCnt, and we can't send it, we drop this chain now
			goto err_destroy_qp;
		}else
		{
			// The newly incoming packet must be queued, so we allocate a new packet for queue and free the original one
			// for RX Ring
			struct rtl_pktHdr *newHdr;
			struct rtl_mBuf  *newmb=NULL, *oldMbuf;
			uint8 *newCluster;
			uint32 newClusterId;
			int16 rxdesc;

			if (pktHdrPtr->ph_rxdesc >= 0)	// only exchange pkthdr when this packet is from RX Ring
			{
				oldMbuf = pktHdrPtr->ph_mbuf;

				//get a new pkt hdr and mbuf.
				if((newmb = mBuf_get(MBUF_DONTWAIT,MBUFTYPE_DATA, 1)) == NULL)
				{
					goto err_destroy_qp;
				}
				if((mBuf_getPkthdr(newmb, MBUF_DONTWAIT) == (struct rtl_mBuf *) NULL)||
				(mBuf_reserve(newmb,rtl8651_fwdEnginePara.mbufHeadroom)))
				{
					//we can't get a pkthdr or can't reserve enough headroom
					mBuf_freeMbufChain(newmb);//free new mbuf we just get.
					goto err_destroy_qp;
				}
	
				assert(newmb);
				// store new cluster/mbuf info. for exchange
				newHdr = newmb->m_pkthdr;
				newCluster = newmb->m_extbuf;
				newClusterId = newmb->m_extClusterId;
				assert(newHdr);
				*newmb = *oldMbuf;
				newmb->m_next = NULL;
				newmb->m_pkthdr = newHdr;
		
				rxdesc = newHdr->ph_rxdesc;
				*newHdr = *pktHdrPtr;
				newHdr->ph_mbuf = newmb;
				newHdr->ph_rxdesc = rxdesc;
				oldMbuf->m_extClusterId = newClusterId;
				oldMbuf->m_data = newCluster;
				oldMbuf->m_extbuf = newCluster;
				oldMbuf->m_len = 0;
				pktHdrPtr->ph_nextHdr = 0;
				mBuf_freeMbufChain(pktHdrPtr->ph_mbuf);		/* give this mbuf/pkthdr back to RX Ring */
		
				/* Insert it into queue chain */
				newHdr->ph_nextHdr = next;
				if (prev)
					{
					prev->ph_nextHdr = newHdr;
				}else
				{
					qp->fragments = newHdr;
				}
			}
			retval = FWDENG_QUEUED;
		}
	}

out:
 	return retval;

err_TearDrop:
	// log tearDrop event here
	{
		struct rtl_pktHdr *pkthdr;
		struct ip *pip;
		ipaddr_t dip;

		if (offset)
		{	// this packet is NOT the first one of fragment packets
			pkthdr = qp->fragments;	// get the first packet in ip-queue for logging
			pip = (struct ip *)((uint8 *)(pkthdr->ph_mbuf->m_data) + (pkthdr->ph_iphdrOffset));
		}else
		{	// this packet is the first one of fragment packet
			pkthdr = pktHdrPtr;
			pip = ip;
		}

		dip = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

		if (vlanp == NULL)
		{	/* direction/vlan is not obtained yet */
			vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
			direction = (vlanp->internal?0:1);	/* make sure packet direction  0:LAN_to_WAN 1:WAN_to_LAN */
		}
		rtl8651a_dosTearDropLogging(pkthdr->ph_dsid, vlanp->vid, pip, direction, pkthdr);
	}
err_dos_detected:
err_destroy_pkt_qp:
	mBuf_freeMbufChain(pktHdrPtr->ph_mbuf);		// drop current packet
err_destroy_qp:
	IP_INC_STATS_BH(IpReasmFails);
	// must drop all queued packets here!
	frag_destroy(qp);
	return SUCCESS;
}

/*************************************************************
	For fast path before fragment system processing:
	
		The flow of the packet has some other pieces in
		fragment system or in negative list,
		this packet can NOT be forwarded.
**************************************************************/
inline int32 _rtl8651_frag_NeedFragmentProcess(uint16 id, ipaddr_t sip, ipaddr_t dip, uint8 l4proto)
{
	unsigned int hash;
	struct ipq *qp;

	/* in dequeue state, we always return FALSE */
	if (_rtl8651_fragQueueProcess)
		return FALSE;

	hash = ipqhashfn(id, sip, dip, l4proto);

	/* check entry negative entry */
	if (negative_ipq[hash] == NULL)
	{
		goto check_incomplete;
	}

	for (qp = negative_ipq[hash]; qp; qp = qp->next)
	{
		if(qp->id == id		&&
		   qp->saddr == sip	&&
		   qp->daddr == dip	&&
		   qp->protocol == l4proto)
		{
 			return TRUE;	/* Neagtive List matched !!! */
		}
	}

check_incomplete:

	/* check incomplete entry */
	if (ipq_hash[hash] == NULL)
		return FALSE;

	// find existing entry
	for (qp = ipq_hash[hash]; qp; qp = qp->next)
	{
		if(qp->id == id		&&
		   qp->saddr == sip	&&
		   qp->daddr == dip	&&
		   qp->protocol == l4proto)
		{
 			return TRUE;
		}
	}
	return FALSE;
}


/*
	Process an incoming IP datagram fragment.

	If queue collection is not complete, we exchange the pkthdr

	retval:
			FWDENG_IP_CANTFRAG		:	system is disabled
			SUCCESS				:	process successful
			FWDENG_QUEUED			:	packet is queued
			FWDENG_QUEUED_TO_GW	:	packet is to protocol stack
			FAILED					:	packet in postive list: continue process it
*/
int32 _rtl8651_ip_defrag(struct rtl_pktHdr * pkthdrPtr, struct ip *pip)
{
	struct ipq *qp;

	if (system_enable == FALSE)
	{
		return FWDENG_IP_CANTFRAG;
	}

	// Increase counter
	IP_INC_STATS_BH(IpReasmReqds);

 	/* Lookup (or create) queue header */
	if ((qp = ip_frag_find(pip)) != NULL) {

		if (qp->last_in & POSITIVE_LIST)
		{	// forward it
			return FAILED;
		}

		// put this packet into queue and process queue
		return ip_frag_queue(qp, pkthdrPtr, pip);
	}else
	{
		// we cannot find corresponding fragment entry, drop this packet
		mBuf_freeMbufChain(pkthdrPtr->ph_mbuf);
	}
//err:
	IP_INC_STATS_BH(IpReasmFails);
	return SUCCESS;
}

static int32 _rtl8651_flushFragQueue(struct rtl_pktHdr *head, struct rtl_pktHdr *currentPkt)
{
	struct rtl_pktHdr  *hp, *fp;
	int val;

	if (!head)
	{
		return FAILED;
	}

	_rtl8651_fragQueueDropChain = FALSE;
	_rtl8651_fragQueueProcess = TRUE;
#ifdef _RTL_NAPT_REDIRECT
	_rtl8651_fragQueueNaptRedirect_MatchedIP = 0;
#endif /* _RTL_NAPT_REDIRECT */

	hp = head;
	while(hp)
	{
		fp = hp->ph_nextHdr;
		hp->ph_nextHdr = NULL;
		val = rtl8651_fwdEngineInput(hp);

		if (_rtl8651_fragQueueDropChain == TRUE)
		{	// one packet is dropped, flush all queued packets directly
			hp = fp;
			goto errout;
		}

		if (val != SUCCESS)
		{
			if (hp == head)
			{	// first packet must be trapped to protocol stack, we trap whole queue to protocol stack
				hp->ph_nextHdr = fp;
				goto concatenate;
			}else
			{	// some packet is sent, but one of remaining packets is failed, drop all of then
				hp->ph_nextHdr = fp;
				goto errout;
			}
		}
		hp = fp;
	}
//out:
	_rtl8651_fragQueueProcess = FALSE;
	return SUCCESS;

errout:
	/* must set hp before entering this block */
	while(hp)
	{
		fp = hp->ph_nextHdr;
		hp->ph_nextHdr = NULL;
		mBuf_freeMbufChain(hp->ph_mbuf);
		hp = fp;
	}
	_rtl8651_fragQueueProcess = FALSE;
	return FAILED;

concatenate:
	// concate all packets in ip-queue and let currentPkt points it's head
	{

		if (head == currentPkt)
		{
			goto concatenateOut;
		}

		// when send packet queue to protocol stack, we must make sure that packet is in-order
		// and currentPkt points to first fragment packet
		hp = head;
		while (hp)
		{
			fp = hp->ph_nextHdr;
	
			if (fp == currentPkt)
			{
				if (hp == head)
				{	// head->nextHdr == currentPkt
					head->ph_nextHdr = currentPkt->ph_nextHdr;
					currentPkt->ph_nextHdr = head;
				}else
				{
					struct rtl_pktHdr *tmpHdr;

					hp->ph_nextHdr = head;
					// exchange next ptr of head/currentPkt
					tmpHdr = head->ph_nextHdr;
					head->ph_nextHdr = currentPkt->ph_nextHdr;
					currentPkt->ph_nextHdr = tmpHdr;
				}
				goto exchange_pkt;
			}
			hp = fp;
		}

		// error ! drop!
		hp = head;
		goto errout;

exchange_pkt:
		// exchange "cluster" of <head> and <currentPkt>
		{
			struct rtl_mBuf *mbuf_h;
			struct rtl_mBuf *mbuf_c;
			void *tmp_s;
			uint32 tmp_u32;
			uint16 tmp_u16;
			int8 buff[sizeof(struct rtl_pktHdr)];
			int16 rxdesc_h, rxdesc_c;
			struct rtl_pktHdr *nextHdr_h, *nextHdr_c;

			// exchange cluster informations
			mbuf_h = head->ph_mbuf;
			mbuf_c = currentPkt->ph_mbuf;

			/* pkthdr */
			rxdesc_h = head->ph_rxdesc;
			rxdesc_c = currentPkt->ph_rxdesc;
			nextHdr_h = head->ph_nextHdr;
			nextHdr_c = currentPkt->ph_nextHdr;

			bzero((void *)buff,sizeof(buff));
			rtl8651_memcpy((void *)buff, (void *)head, sizeof(struct rtl_pktHdr));
			rtl8651_memcpy((void *)head, (void *)currentPkt, sizeof(struct rtl_pktHdr));
			rtl8651_memcpy((void *)currentPkt, buff, sizeof(struct rtl_pktHdr));

			head->ph_mbuf = mbuf_h;
			currentPkt->ph_mbuf = mbuf_c;
			head->ph_rxdesc = rxdesc_h;
			currentPkt->ph_rxdesc = rxdesc_c;
			head->ph_nextHdr = nextHdr_h;
			currentPkt->ph_nextHdr = nextHdr_c;

			/* mbuf */
			// m_data
			tmp_s = mbuf_h->m_data;
			mbuf_h->m_data = mbuf_c->m_data;
			mbuf_c->m_data = tmp_s;
			// m_extbuf
			tmp_s = mbuf_h->m_extbuf;
			mbuf_h->m_extbuf = mbuf_c->m_extbuf;
			mbuf_c->m_extbuf = tmp_s;
			// m_extClusterId
			tmp_u32 = mbuf_h->m_extClusterId;
			mbuf_h->m_extClusterId = mbuf_c->m_extClusterId;
			mbuf_c->m_extClusterId = tmp_u32;
			// m_extsize
			tmp_u16 = mbuf_h->m_extsize;
			mbuf_h->m_extsize = mbuf_c->m_extsize;
			mbuf_c->m_extsize = tmp_u16;
			// m_len
			tmp_u16 = mbuf_h->m_len;
			mbuf_h->m_len = mbuf_c->m_len;
			mbuf_c->m_len = tmp_u16;
		}
	}

concatenateOut:	
	/*
		before we trap these packets, we must check if we need to modify them
			1. for napt redirect
	*/
#ifdef _RTL_NAPT_REDIRECT
	/* it means the first packet match napt-redirect table and the matched redirect IP is set */
	if ( _rtl8651_fragQueueNaptRedirect_MatchedIP )
	{	/* for napt redirect */
		struct rtl_pktHdr *m_pkt;
		struct ip *m_ip;
		ipaddr_t dip;

		m_pkt = currentPkt;
		dip = 0;
		while (m_pkt)
		{
			m_ip = (struct ip *)(((uint32)(m_pkt->ph_mbuf->m_data) + (m_pkt->ph_iphdrOffset)));

			/* we only modify remaining packets because the first packet is modified by forwarding engine */
			if (IS_FRAGMENT_REMAIN(m_ip->ip_off))
			{	/* all remaining packets only need to modify L3 information */
				if (!dip)
					dip = PKTGET_UINT32_UNALIGNED(&m_ip->ip_dst);

				PKTSET_UINT32_UNALIGNED(_rtl8651_fragQueueNaptRedirect_MatchedIP, &m_ip->ip_dst);
				_RTL8651_L3_ADJUST_CHECKSUM(_rtl8651_fragQueueNaptRedirect_MatchedIP, dip, m_ip->ip_sum);
			}
			/* next packet */
			m_pkt = m_pkt->ph_nextHdr;
		}
	}
#endif /* _RTL_NAPT_REDIRECT */

	_rtl8651_fragQueueProcess = FALSE;
	return FWDENG_QUEUED_TO_GW;
}




