/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : IPSec-related function of ROME Driver
* Abstract : 
* Author : Louis Yung-Chieh Lo (yjlou@realtek.com.tw)               
* $Id: rtl8651_ipsec.c,v 1.1.1.1 2006/07/26 05:38:23 blue Exp $
*/


#include "rtl_types.h"
#include "rtl8651_ipsec.h"
#include "rtl8651_ipsec_local.h"
#include "rtl_errno.h"
#include "assert.h"
#include "rtl8651b_authEngine.h"
#include "rtl8651b_cryptoEngine.h"
#include "aes.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_aclLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_multicast.h"
#ifdef CONFIG_RTL865X_ROMEPERF
#include "../romeperf.h"
#endif

#ifdef RTL865X_TEST
#ifdef CYGWIN
typedef int size_t;
#endif
#endif


/*
 * 0 --- no debug information
 * 1 --- print only when API is called
 * 2 --- print debug info only when error
 * 3 --- less info
 * 4 --- more details
 * 5 --- debug all
 */
#define _IPSEC_DEBUG_ 0

#undef _LOUIS_DEBUG_

/*  _DISMISS_TO_FWDENG_
 *
 *  If defined, the invalid IPSEC packets, e.g. IPSEC passthru, will be returned back to 
 *    forwarding engine for further process.
 *  If not defined, the invalid packets will be silently dropped.
 */
#define _DISMISS_TO_FWDENG_


int32 rtl865x_ipsecMtu = 65536; /* default: no limit */
int32 rtl865x_ipsecMaxConcurrentSpi; /* Maximum concurrent SPI number */
int32 rtl865x_ipsecMaxConcurrentSpiGrp; /* Maximum concurrent SPI Group number */
int32 rtl865x_ipsecMaxConcurrentERoute; /* Maximum concurrent ERoute number */

/*
 * _SW_AES_USE_DRAM_
 * If defined, software AES will copy packet to DRAM, and encrypt. It cost 1876 bytes of DRAM.
 * If not defined, software AES will encrypt in uncached area.
 */
#undef _SW_AES_USE_DRAM_

#ifdef _SW_AES_USE_DRAM_
__DRAM_CRYPTO uint8 rtl8651_AES_buf[1600];
__DRAM_CRYPTO uint8 rtl8651_AES_iv[32];
__DRAM_CRYPTO static AES_KEY rtl8651_AES_key;
#else
static AES_KEY rtl8651_AES_key;
#endif

static int32 rtl8651_cryptoKeyLen[] =
{
	0,
	8,	/*IPSEC_CBC_DES */
	24,	/*IPSEC_CBC_3DES*/
	8,	/*IPSEC_ECB_DES */
	24,	/*IPSEC_ECB_3DES*/
	16,	/*IPSEC_CBC_AES128*/
	24,	/*IPSEC_CBC_AES192*/
	32,	/*IPSEC_CBC_AES256*/
};

static int32 rtl8651_cryptoIVLen[] =
{
	0,
	sizeof(cbcHdr64_t),		/*IPSEC_CBC_DES */
	sizeof(cbcHdr64_t),		/*IPSEC_CBC_3DES*/
	sizeof(cbcHdr64_t),		/*IPSEC_ECB_DES */
	sizeof(cbcHdr64_t),		/*IPSEC_ECB_3DES*/
	sizeof(cbcHdr128_t),	/*IPSEC_CBC_AES128*/
	sizeof(cbcHdr128_t),	/*IPSEC_CBC_AES192*/
	sizeof(cbcHdr128_t),	/*IPSEC_CBC_AES256*/
};

static int32 rtl8651_cryptoMaxPadLen[] =
{
	0,
	8,	/*IPSEC_CBC_DES */
	8,	/*IPSEC_CBC_3DES*/
	8,	/*IPSEC_ECB_DES */
	8,	/*IPSEC_ECB_3DES*/
	16,	/*IPSEC_CBC_AES128*/
	16,	/*IPSEC_CBC_AES192*/
	16,	/*IPSEC_CBC_AES256*/
};

static int32 rtl8651_authKeyLen[] =
{
	0,
	16,	/*IPSEC_HASH_MD5 */
	20,	/*IPSEC_HASH_SHA1*/
	16,	/*IPSEC_HMAC_MD5 */
	20,	/*IPSEC_HMAC_SHA1*/
};


static int32 rtl8651_authDataLen[] =
{
	0,
	12,	/*IPSEC_HASH_MD5 */
	12,	/*IPSEC_HASH_SHA1*/
	12,	/*IPSEC_HMAC_MD5 */
	12,	/*IPSEC_HMAC_SHA1*/
};

static char* strIpsecProto[] =
{
	"unknwon",
	"ESP",
	"IPIP",
};

static char* strIpsecCryptoType[] =
{
	"unknown",
	"CBC_DES",
	"CBC_3DES",
	"ECB_DES",
	"ECB_3DES",
	"CBC_AES128",
	"CBC_AES192",
	"CBC_AES256",
};

static char* strIpsecAuthType[] =
{
	"unknown",
	"HASH_MD5",
	"HASH_SHA1",
	"HMAC_MD5",
	"HMAC_SHA1",
};


#define _IPIP_OVERHEAD_LEN_(cryptoType) (sizeof(struct ip))
#define _ESP_OVERHEAD_LEN_(cryptoType) (sizeof(espHdr_t)+rtl8651_cryptoIVLen[cryptoType])
#define _TUNNEL_OVERHEAD_LEN_(cryptoType) (_IPIP_OVERHEAD_LEN_(cryptoType)+_ESP_OVERHEAD_LEN_(cryptoType))
#define _MAX_POSSIBLE_PADLEN_(cryptoType) rtl8651_cryptoMaxPadLen[cryptoType] /* MAX possible padding */
#ifdef RTL865X_TEST
#define IPSEC_UNCACHED_ADDR(addr) (addr)
#else
#define IPSEC_UNCACHED_ADDR(addr) ((void*)((uint32)(addr)|0xa0000000))
#endif


/* global variables */
ipsecEntryList_t IpsecEntry;



#ifdef _LOUIS_DEBUG_
typedef uint32 in_addr_t;
static int inet_aton(const char *cp, struct in_addr *addrptr)
{
	in_addr_t addr;
	int value;
	int part;

	addr = 0;
	for (part = 1; part <= 4; part++) {

		if (!isdigit(*cp))
			return 0;

		value = 0;
		while (isdigit(*cp)) {
			value *= 10;
			value += *cp++ - '0';
			if (value > 255)
				return 0;
		}

		if (part < 4) {
			if (*cp++ != '.')
				return 0;
		} else {
			char c = *cp++;
			if (c != '\0' && !isspace(c))
			return 0;
		}

		addr <<= 8;
		addr |= value;
	}

	/*  W. Richard Stevens in his book UNIX Network Programming,
	 *  Volume 1, second edition, on page 71 says:
	 *
	 *  An undocumented feature of inet_aton is that if addrptr is
	 *  a null pointer, the function still performs it validation
	 *  of the input string, but does not store the result.
	 */
	if (addrptr) {
	    addrptr->s_addr = htonl(addr);
	}

	return 1;
}

static unsigned long inet_addr(const char *cp)
{
	struct in_addr a;

	if (!inet_aton(cp, &a))
		return -1;
	else
		return a.s_addr;
}
#endif/*_LOUIS_DEBUG_*/



int32 _rtl8651_initIpsec( void )
{
#ifdef _LOUIS_DEBUG_
	int32 retval;
	uint32 encKey[] = {0x40ee9860,0xd890027c,0xa8b32072,0x3cbb0855,0x8ee22f98,0x11adca08};
	uint32 authKey[] = {0xac56a5c2,0xbe6188cd,0x078c290b,0x89b3498c};
	ipaddr_t extIp = ntohl(inet_addr("192.168.154.143"));
	ipaddr_t remIp = ntohl(inet_addr("192.168.154.241"));
#endif

#if _IPSEC_DEBUG_>=1
	rtlglue_printf( "%s():%d\n", __FUNCTION__, __LINE__ );
#endif

#ifdef RTL865X_TEST
#else
	rtl8651b_authEngine_init( 32, TRUE );
	rtl8651b_cryptoEngine_init( 32, TRUE );
#endif
	_rtl8651_flushIpsec();


#ifdef _LOUIS_DEBUG_
	/************************** Outbound *************************/
	/* ADD EROUTE */
	retval = rtl8651_addIpsecERoute( ntohl(inet_addr("192.168.143.0")), ntohl(inet_addr("255.255.255.0")),
	                                 ntohl(inet_addr("192.168.241.0")), ntohl(inet_addr("255.255.255.0")),
	                                 remIp, 0x1005, IPSEC_IPIP, 
	                                 0 );
	assert( retval == SUCCESS );
	
	/* ADD IPIP */
	retval = rtl8651_addIpsecSpi( remIp, 0x1005, IPSEC_IPIP, 
	                              extIp, 0, NULL, 0, NULL,
	                              extIp, 15 , 0, 0 );
	assert( retval == SUCCESS );
	/* ADD ESP */
	retval = rtl8651_addIpsecSpi( remIp, 0x1005, IPSEC_ESP, 
	                              extIp, IPSEC_CBC_3DES, encKey, IPSEC_HMAC_MD5, authKey,
	                              ntohl(inet_addr("0.0.0.0")), 15 , 0, IPSEC_AUTH_DATA_PRESENT );
	assert( retval == SUCCESS );

	/* ADD SPI GRP */
	retval = rtl8651_addIpsecSpiGrp( remIp, 0x1005, IPSEC_IPIP, 
								     remIp, 0x1005, IPSEC_ESP, 
								     0, 0, 0, 
								     0, 0, 0, 
	                                 0 );
	assert( retval == SUCCESS );


	/************************** Inbound *************************/
	/* ADD EROUTE */
	retval = rtl8651_addIpsecERoute( ntohl(inet_addr("192.168.241.0")), ntohl(inet_addr("255.255.255.0")),
	                                 ntohl(inet_addr("192.168.143.0")), ntohl(inet_addr("255.255.255.0")),
	                                 extIp, 0x1005, IPSEC_IPIP, 
	                                 0 );
	assert( retval == SUCCESS );

	/* ADD ESP */
	retval = rtl8651_addIpsecSpi( extIp, 0x1005, IPSEC_ESP, 
	                              remIp, IPSEC_CBC_3DES, encKey, IPSEC_HMAC_MD5, authKey,
	                              ntohl(inet_addr("0.0.0.0")), 15 , 0, IPSEC_AUTH_DATA_PRESENT );
	assert( retval == SUCCESS );
	/* ADD IPIP */
	retval = rtl8651_addIpsecSpi( extIp, 0x1005, IPSEC_IPIP, 
	                              remIp, 0, NULL, 0, NULL,
	                              extIp, 15 , 0, 0 );
	assert( retval == SUCCESS );

	/* ADD SPI GRP */
	retval = rtl8651_addIpsecSpiGrp( extIp, 0x1005, IPSEC_IPIP, 
								     extIp, 0x1005, IPSEC_ESP, 
								     0, 0, 0, 
								     0, 0, 0, 
	                                 0 );
	assert( retval == SUCCESS );
#endif

	return SUCCESS;
}


int32 _rtl8651_allocIpsec( int32 spiNum, int32 spiGrpNum, int32 eRouteNum )
{
	int32 i;
	
#if _IPSEC_DEBUG_>=1
	rtlglue_printf( "%s(#spi=%d,#spiGrp=%d,#eRoute=%d):%d\n", __FUNCTION__, spiNum, spiGrpNum, eRouteNum, __LINE__ );
#endif

	/*==========================================================
	  * Allocate memory for IPSEC SPI entry 
	  *==========================================================*/
	{
		ipsecSpi_t *pool;
		
		CTAILQ_INIT( &IpsecEntry.freeList.spi );
		CTAILQ_INIT( &IpsecEntry.inuseList.spi );
		TBL_MEM_ALLOC( pool, ipsecSpi_t, spiNum );
		memset( pool, 0, sizeof(ipsecSpi_t)*spiNum );		
		// chain all entries are in free list.
		for( i=0; i<spiNum; i++ ) 
			CTAILQ_INSERT_TAIL( &IpsecEntry.freeList.spi, &pool[i], next );
	}
	
	/*==========================================================
	  * Allocate memory for IPSEC SPIGRP entry 
	  *==========================================================*/
	{
		ipsecSpiGrp_t *pool;
		
		CTAILQ_INIT( &IpsecEntry.freeList.spigrp );
		CTAILQ_INIT( &IpsecEntry.inuseList.spigrp );
		TBL_MEM_ALLOC( pool, ipsecSpiGrp_t, spiGrpNum );
		memset( pool, 0, sizeof(ipsecSpiGrp_t)*spiGrpNum );		
		// chain all entries are in free list.
		for( i=0; i<spiGrpNum; i++ ) 
			CTAILQ_INSERT_TAIL( &IpsecEntry.freeList.spigrp, &pool[i], next );
	}
	
	/*==========================================================
	  * Allocate memory for IPSEC EROUTE entry 
	  *==========================================================*/
	{
		ipsecERoute_t *pool;
		
		CTAILQ_INIT( &IpsecEntry.freeList.eroute );
		CTAILQ_INIT( &IpsecEntry.inuseList.eroute );
		TBL_MEM_ALLOC( pool, ipsecERoute_t, eRouteNum );
		memset( pool, 0, sizeof(ipsecERoute_t)*eRouteNum );		
		// chain all entries are in free list.
		for( i=0; i<eRouteNum; i++ ) 
			CTAILQ_INSERT_TAIL( &IpsecEntry.freeList.eroute, &pool[i], next );
	}
	
	return SUCCESS;
}


/****************************************************************
 *******   FOR CLE
 ****************************************************************/
int32 _rtl8651_dumpIpsec()
{
	ipsecSpi_t *spi;
	ipsecSpiGrp_t *spigrp;
	ipsecERoute_t *eroute;
	int32 i;
	
	rtlglue_printf(">>> IPSec\n");
	
	rtlglue_printf("\n==== IPSec SPI entry =============================================\n");
	CTAILQ_FOREACH( spi, &IpsecEntry.inuseList.spi, next )
	{
		assert( spi->valid );

		rtlglue_printf( "0x%08x@"NIPQUAD2STR" %-4s ",
			spi->spi, NIPQUAD2((spi->edstIp)), strIpsecProto[spi->proto] );
		rtlglue_printf( "%s ", spi->spiGrp?"[linked]":"[      ]" );
		rtlglue_printf( "%s AD:%d antiReplay:%d\n", 
			spi->outbound?"outbound":" inbound", 
			spi->authDataPresent, spi->IHaveAntiReplay );
		rtlglue_printf( "  srcIp="NIPQUAD2STR"  seq=%d  winSiz=%d  life=%d\n", 
		                NIPQUAD2((spi->srcIp)), spi->seq , spi->replayWindow, spi->lifetime );
		if ( spi->proto==IPSEC_ESP )
		{
			rtlglue_printf( "  %s:%s(", spi->outbound?"encrypt":"decrypt", strIpsecCryptoType[spi->cryptoType] );
			for( i=0; i<rtl8651_cryptoKeyLen[spi->cryptoType]; i++ )
			{
				rtlglue_printf( "%02x", spi->cryptoKey[i] );
				if ( (i&3)==3 ) rtlglue_printf( "_" );
			}
			rtlglue_printf( ")\n" );
			
			rtlglue_printf( "     auth:%s(", strIpsecAuthType[spi->authType] );
			for( i=0; i<rtl8651_authKeyLen[spi->authType]; i++ )
			{
				rtlglue_printf( "%02x", spi->authKey[i] );
				if ( (i&3)==3 ) rtlglue_printf( "_" );
			}
			rtlglue_printf( ")\n" );
		}
		else if ( spi->proto==IPSEC_IPIP )
		{
			rtlglue_printf( "  dstIp="NIPQUAD2STR"\n", NIPQUAD2((spi->dstIp)) );
		}
		rtlglue_printf( "------------------------------------------------------------\n" );
	}

	rtlglue_printf("\n==== IPSec SPIGRP entry =============================================\n");
	CTAILQ_FOREACH( spigrp, &IpsecEntry.inuseList.spigrp, next )
	{
		assert( spigrp->valid );

		for( i = 0; i < sizeof(spigrp->entry)/sizeof(spigrp->entry[0]); i++ )
		{
			if ( spigrp->entry[i].valid )
			{
				rtlglue_printf( "[%d]%s0x%08x@"NIPQUAD2STR"[%s] ", i,
				    strIpsecProto[spigrp->entry[i].proto],
					spigrp->entry[i].spi, NIPQUAD2((spigrp->entry[i].ip)),
					spigrp->entry[i].linkedSpi?"*":" " );
			}
		}
		rtlglue_printf( "\n" );
	}

	rtlglue_printf("\n==== IPSec EROUTE entry =============================================\n");
	CTAILQ_FOREACH( eroute, &IpsecEntry.inuseList.eroute, next )
	{
		assert( eroute->valid );

		rtlglue_printf( "0x%08x@"NIPQUAD2STR" %s %s acl:%d\n",
			eroute->spi, NIPQUAD2((eroute->edstIp)), strIpsecProto[eroute->proto],
			eroute->linkedSpi?"[linked]":"[      ]",
			(eroute->linkedSpi)?eroute->linkedSpi->aclRule:-1);
		rtlglue_printf(	"  srcNet="NIPQUAD2STR"/"NIPQUAD2STR" --> dstNet="NIPQUAD2STR"/"NIPQUAD2STR"\n",
			NIPQUAD2((eroute->srcIp)), NIPQUAD2((eroute->srcIpMask)),
			NIPQUAD2((eroute->dstIp)), NIPQUAD2((eroute->dstIpMask)) );
		rtlglue_printf(	"  Success:%d  InvERoute:%d  RouteFail:%d  ErrorSeq:%d  NotIPIP:%d\n",
			eroute->counter.cntSuccess, eroute->counter.cntInvalidEroute, eroute->counter.cntRouteFailed, eroute->counter.cntErrorSeq, eroute->counter.cntNotAnESPPacket );
		rtlglue_printf(	"  ErrorAuth:%d  CrypEngErr:%d  AuthEngErr:%d  SApolicy:%d\n",
			eroute->counter.cntErrorAuth, eroute->counter.cntCryptoEngineError, eroute->counter.cntAuthEngineError, eroute->counter.cntSourceAddressNotMatched );
		rtlglue_printf(	"  FragError:%d\n",
			eroute->counter.cntFragError );
	}

	/* EROUTE */
	return SUCCESS;
}


/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * -=  SPI Function
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*
@func int32		| _rtl8651_findIpsecSpi	| find an IPSEC SPI 
@parm ipaddr_t	| edstIp		| Effective destination IP Address
@parm uint32	| spi			| Outbound Security Parameter Index
@parm uint32	| proto			| proto (ESP or IPIP)
@parm ipsecSpi_t**	| pRet		| Found spi entry
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_findIpsecSpi( ipaddr_t edstIp, uint32 spi, uint32 proto,
                             ipsecSpi_t **pRet )
{
	int32 retval = SUCCESS;
	ipsecSpi_t *pSpi;
	
	CTAILQ_FOREACH( pSpi, &IpsecEntry.inuseList.spi, next )
	{
		assert( pSpi->valid );

		if ( pSpi->edstIp == edstIp &&
		     pSpi->spi == spi &&
		     pSpi->proto == proto )
		{
			/* found */
			if ( pRet ) *pRet = pSpi;
			goto out;
		}
	}

	/* not found */
	if ( pRet ) *pRet = NULL;
	retval = TBLDRV_EIPSEC_SPI_NOT_FOUNT;
out:
	return retval;
}


/*
@func int32		| _rtl8651_addIpsecSpi	| add IPSEC SPI 
@parm ipaddr_t	| edstIp		| Effective destination IP Address
@parm uint32	| spi			| Outbound Security Parameter Index
@parm uint32	| proto			| proto (ESP or IPIP)
@parm ipaddr_t	| srcIp			| Source IP Address
@parm int32		| cryptoType	| IPSEC_CBC_DES, IPSEC_CBC_3DES, IPSEC_ECB_DES, IPSEC_ECB_3DES
@parm void*		| encKey		| pointer to encrypt key
@parm int32		| authType		| IPSEC_HASH_MD5, IPSEC_HASH_SHA1, IPSEC_HMAC_MD5, IPSEC_HMAC_SHA1
@parm void*		| authKey		| pointer to auth key
@parm ipaddr_t	| dstIp			| Destination IP Address
@parm int32		| lifetime		| Lifetime
@parm int32		| replayWindow	| Allowed Replay Window Size.
@parm uint32	| flags			| some flags for different configuration
@parm ipsecSpi_t**	| pRet		| Newly added spi entry
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_addIpsecSpi( ipaddr_t edstIp, uint32 spi, uint32 proto,
                            ipaddr_t srcIp, uint32 cryptoType, void* cryptoKey, uint32 authType, void* authKey,
                            ipaddr_t dstIp, int32 lifetime, int32 replayWindow, uint32 flags,
                            ipsecSpi_t **pRet )
{
	int32 retval;
	ipsecSpi_t *pSpi;

#if _IPSEC_DEBUG_>=1
	rtlglue_printf( "%s(0x%08x@"NIPQUAD2STR" %-4s,\n", __FUNCTION__, spi, NIPQUAD2(edstIp), strIpsecProto[proto] );
	rtlglue_printf( "    src="NIPQUAD2STR" %-4s%-4s,\n", NIPQUAD2(srcIp), strIpsecCryptoType[cryptoType], strIpsecAuthType[authType] );
	rtlglue_printf( "    dst="NIPQUAD2STR" life=%d replay=%d flags=0x%08x)\n", NIPQUAD2(dstIp), lifetime, replayWindow, flags );
#endif

	if ( replayWindow < 32 )
	{
		/* Invalid window size (See RFC2046 3.4.3) */
		retval = TBLDRV_EIPSEC_INVALID_WINSIZE;
		goto out;
	}
	if ( replayWindow > MAX_WINDOW_SIZE )
	{
		retval = TBLDRV_EIPSEC_INVALID_WINSIZE;
		goto out;
	}

	if ( _rtl8651_findIpsecSpi( edstIp, spi, proto, &pSpi ) == SUCCESS )
	{
		/* found, clear pointer to force link. */
		pSpi->spiGrp = NULL;
	}
	else
	{
		/* not found, alloc one. */
		if ( ( pSpi = CTAILQ_FIRST( &IpsecEntry.freeList.spi ) ) )
		{
			/* chain to inuse list */
			CTAILQ_REMOVE( &IpsecEntry.freeList.spi, pSpi, next );
			CTAILQ_INSERT_TAIL(&IpsecEntry.inuseList.spi, pSpi, next );
		}
		else
		{
			retval = TBLDRV_EIPSEC_NO_MORE_SPI;
			goto out;
		}
	}

	pSpi->valid = TRUE;
	pSpi->outbound = _rtl8651_extIpAddr( edstIp, NULL )?FALSE:TRUE;
	pSpi->authDataPresent = (flags&IPSEC_AUTH_DATA_PRESENT)?TRUE:FALSE;
	pSpi->IHaveAntiReplay = (flags&IPSEC_I_HAVE_ANTI_REPLAY)?TRUE:FALSE;
	pSpi->edstIp = edstIp;
	pSpi->spi = spi;
	pSpi->proto = proto;
	pSpi->srcIp = srcIp;
	if ( proto==IPSEC_ESP )
	{
		pSpi->cryptoType = cryptoType;
		if ( cryptoKey )
		{
			memcpy( pSpi->cryptoKey, cryptoKey, rtl8651_cryptoKeyLen[cryptoType] );
		}
		pSpi->authType = authType;
		if ( authType )
		{
			int i;
			
			memcpy( pSpi->authKey, authKey, rtl8651_authKeyLen[authType] );

			/* pre-compute ipad and opad */
			memset( pSpi->ipad, 0x36, RTL8651B_MAX_MD_CBLOCK );
			memset( pSpi->opad, 0x5c, RTL8651B_MAX_MD_CBLOCK );
			for( i = 0; i < rtl8651_authKeyLen[authType]; i++ ) 
			{
				pSpi->ipad[i] ^= ((uint8*)authKey)[i];
				pSpi->opad[i] ^= ((uint8*)authKey)[i];
			}
		}
	}
	else if ( proto==IPSEC_IPIP )
	{
		pSpi->dstIp = dstIp;
	}
	pSpi->lifetime = lifetime;
	pSpi->replayWindow = replayWindow;
	pSpi->seq = 1;

	/* re-binding */
	_rtl8651_ipsecAutoLink();
	_rtl8651_arrangeAllACL();
	
	if ( CTAILQ_TOTAL( &IpsecEntry.inuseList.spi ) > rtl865x_ipsecMaxConcurrentSpi )
	{
		rtl865x_ipsecMaxConcurrentSpi = CTAILQ_TOTAL( &IpsecEntry.inuseList.spi );
	}
	
	retval = SUCCESS;
out:
	return retval;
}


/*
@func int32		| _rtl8651_delIpsecSpi	| delete IPSEC SPI 
@parm ipaddr_t	| edstIp		| Effective destination IP Address
@parm uint32	| spi			| Outbound Security Parameter Index
@parm uint32	| proto			| proto (ESP or IPIP)
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_delIpsecSpi( ipaddr_t edstIp, uint32 spi, uint32 proto )
{
	int32 retval;
	ipsecSpi_t *pSpi;

#if _IPSEC_DEBUG_>=1
	rtlglue_printf( "%s(0x%08x@"NIPQUAD2STR" %-4s)\n", __FUNCTION__, spi, NIPQUAD2(edstIp), strIpsecProto[proto] );
#endif

	if ( _rtl8651_findIpsecSpi( edstIp, spi, proto, &pSpi ) != SUCCESS )
	{
		/* not found, return. */
		retval = TBLDRV_EIPSEC_SPI_NOT_FOUNT;
		goto out;
	}

	memset( pSpi, 0, offsetof( ipsecSpi_t, next ) );
	CTAILQ_REMOVE( &IpsecEntry.inuseList.spi, pSpi, next );
	CTAILQ_INSERT_TAIL(&IpsecEntry.freeList.spi, pSpi, next );

	/* re-binding */
	_rtl8651_ipsecAutoLink();
	_rtl8651_arrangeAllACL();
	
	retval = SUCCESS;
out:
	return retval;
}


/*
@func int32		| rtl8651_addIpsecSpi	| add IPSEC SPI 
@parm ipaddr_t	| edstIp		| Effective destination IP Address
@parm uint32	| spi			| Outbound Security Parameter Index
@parm uint32	| proto			| proto (ESP or IPIP)
@parm ipaddr_t	| srcIp			| Source IP Address
@parm int32		| cryptoType	| IPSEC_CBC_DES, IPSEC_CBC_3DES, IPSEC_ECB_DES, IPSEC_ECB_3DES
@parm void*		| encKey		| pointer to encrypt key
@parm int32		| authType		| IPSEC_HASH_MD5, IPSEC_HASH_SHA1, IPSEC_HMAC_MD5, IPSEC_HMAC_SHA1
@parm void*		| authKey		| pointer to auth key
@parm ipaddr_t	| dstIp			| Destination IP Address
@parm int32		| lifetime		| Lifetime
@parm int32		| replayWindow	| Allowed Replay Window Size.
@parm uint32	| flags			| some flags for different configuration
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 rtl8651_addIpsecSpi( ipaddr_t edstIp, uint32 spi, uint32 proto,
                           ipaddr_t srcIp, uint32 cryptoType, void* cryptoKey, uint32 authType, void* authKey,
                           ipaddr_t dstIp, int32 lifetime, int32 replayWindow, uint32 flags )
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_addIpsecSpi( edstIp, spi, proto,
                                   srcIp, cryptoType, cryptoKey, authType, authKey,
                                   dstIp, lifetime, replayWindow, flags, NULL );
	rtlglue_drvMutexUnlock();
	return retval;
}


/*
@func int32		| rtl8651_updateIpsecSpi	| Update an IPSEC SPI 
@parm ipaddr_t	| edstIp		| Effective destination IP Address
@parm uint32	| spi			| Outbound Security Parameter Index
@parm uint32	| proto			| proto (ESP or IPIP)
@parm ipaddr_t	| srcIp			| Source IP Address
@parm int32		| cryptoType	| IPSEC_CBC_DES, IPSEC_CBC_3DES, IPSEC_ECB_DES, IPSEC_ECB_3DES
@parm void*		| encKey		| pointer to encrypt key
@parm int32		| authType		| IPSEC_HASH_MD5, IPSEC_HASH_SHA1, IPSEC_HMAC_MD5, IPSEC_HMAC_SHA1
@parm void*		| authKey		| pointer to auth key
@parm ipaddr_t	| dstIp			| Destination IP Address
@parm int32		| lifetime		| Lifetime
@parm int32		| replayWindow	| Allowed Replay Window Size.
@parm uint32	| flags			| some flags for different configuration
@rvalue SUCCESS	| success
@comm
Alias to _rtl8651_addIpsecSpi().
@devnote
 */
int32 rtl8651_updateIpsecSpi( ipaddr_t edstIp, uint32 spi, uint32 proto,
                              ipaddr_t srcIp, uint32 cryptoType, void* cryptoKey, uint32 authType, void* authKey,
                              ipaddr_t dstIp, int32 lifetime, int32 replayWindow, uint32 flags )
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_addIpsecSpi( edstIp, spi, proto,
                                   srcIp, cryptoType, cryptoKey, authType, authKey,
                                   dstIp, lifetime, replayWindow, flags, NULL );
	rtlglue_drvMutexUnlock();
	return retval;
}


/*
@func int32		| rtl8651_delIpsecSpi	| delete IPSEC SPI 
@parm ipaddr_t	| edstIp		| Effective destination IP Address
@parm uint32	| spi			| Outbound Security Parameter Index
@parm uint32	| proto			| proto (ESP or IPIP)
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 rtl8651_delIpsecSpi( ipaddr_t edstIp, uint32 spi, uint32 proto )
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_delIpsecSpi( edstIp, spi, proto );
	rtlglue_drvMutexUnlock();
	return retval;
}


/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * -=  SPIGRP Function
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*
@func int32		| _rtl8651_findIpsecSpiGrp	| add IPSEC SPI 
@parm ipaddr_t	| edstIp		| Effective destination IP Address
@parm uint32	| spi			| Outbound Security Parameter Index
@parm uint32	| proto			| proto (ESP or IPIP)
@parm ipsecSpiGrp_t**	| pRet		| Found spigrp entry
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_findIpsecSpiGrp( ipaddr_t edstIp, uint32 spi, uint32 proto,
                                ipsecSpiGrp_t **pRet )
{
	int32 retval = SUCCESS;
	ipsecSpiGrp_t *pSpiGrp;
	int32 i;
	
	CTAILQ_FOREACH( pSpiGrp, &IpsecEntry.inuseList.spigrp, next )
	{
		assert( pSpiGrp->valid );

		for( i = 0; i < sizeof(pSpiGrp->entry)/sizeof(pSpiGrp->entry[0]); i++ )
		{
			if ( pSpiGrp->entry[i].valid &&
			     pSpiGrp->entry[i].ip == edstIp &&
			     pSpiGrp->entry[i].spi == spi &&
			     pSpiGrp->entry[i].proto == proto )
			{
				/* found */
				if ( pRet ) *pRet = pSpiGrp;
				goto out;
			}
		}
	}

	/* not found */
	if ( pRet ) *pRet = NULL;
	retval = TBLDRV_EIPSEC_SPIGRP_NOT_FOUNT;
out:
	return retval;
}


/*
@func int32		| _rtl8651_addIpsecSpiGrp	| add IPSEC SPI GRP
@parm ipaddr_t	| edstIp1		| Effective destination IP Address 1
@parm uint32	| spi1			| Outbound Security Parameter Index 1
@parm uint32	| proto1		| proto 1 (ESP or IPIP)
@parm ipaddr_t	| edstIp2		| Effective destination IP Address 2
@parm uint32	| spi2			| Outbound Security Parameter Index 2
@parm uint32	| proto2		| proto 2 (ESP or IPIP)
@parm ipaddr_t	| edstIp3		| Effective destination IP Address 3
@parm uint32	| spi3			| Outbound Security Parameter Index 3
@parm uint32	| proto3		| proto 3 (ESP or IPIP)
@parm ipaddr_t	| edstIp4		| Effective destination IP Address 4
@parm uint32	| spi4			| Outbound Security Parameter Index 4
@parm uint32	| proto4		| proto 4 (ESP or IPIP)
@parm uint32	| flags			| some flags for different configuration
@parm ipsecSpiGrp_t**	| pRet	| Newly added spigrp entry
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_addIpsecSpiGrp( ipaddr_t edstIp1, uint32 spi1, uint32 proto1,
                               ipaddr_t edstIp2, uint32 spi2, uint32 proto2,
                               ipaddr_t edstIp3, uint32 spi3, uint32 proto3,
                               ipaddr_t edstIp4, uint32 spi4, uint32 proto4,
                               uint32 flags,
                               ipsecSpiGrp_t **pRet )
{
	int32 retval;
	ipsecSpiGrp_t *pSpiGrp;

#if _IPSEC_DEBUG_>=1
	rtlglue_printf( "%s(0x%08x@"NIPQUAD2STR" %-4s,\n", __FUNCTION__, spi1, NIPQUAD2(edstIp1), strIpsecProto[proto1] );
	rtlglue_printf( "    0x%08x@"NIPQUAD2STR" %-4s,\n", spi2, NIPQUAD2(edstIp2), strIpsecProto[proto2] );
	rtlglue_printf( "    0x%08x@"NIPQUAD2STR" %-4s,\n", spi3, NIPQUAD2(edstIp3), strIpsecProto[proto3] );
	rtlglue_printf( "    0x%08x@"NIPQUAD2STR" %-4s,\n", spi4, NIPQUAD2(edstIp4), strIpsecProto[proto4] );
	rtlglue_printf( "    flags=0x%08x)\n", flags );
#endif

	if ( _rtl8651_findIpsecSpiGrp( edstIp1, spi1, proto1, &pSpiGrp ) == SUCCESS )
	{
		/* found, clear pointer to force link. */
		pSpiGrp->entry[0].linkedSpi = NULL;
		pSpiGrp->entry[1].linkedSpi = NULL;
		pSpiGrp->entry[2].linkedSpi = NULL;
		pSpiGrp->entry[3].linkedSpi = NULL;
	}
	else
	{
		/* not found, alloc one. */
		if ( ( pSpiGrp = CTAILQ_FIRST( &IpsecEntry.freeList.spigrp ) ) )
		{
			/* chain to inuse list */
			CTAILQ_REMOVE( &IpsecEntry.freeList.spigrp, pSpiGrp, next );
			CTAILQ_INSERT_TAIL(&IpsecEntry.inuseList.spigrp, pSpiGrp, next );
		}
		else
		{
			retval = TBLDRV_EIPSEC_NO_MORE_SPIGRP;
			goto out;
		}
	}

	pSpiGrp->valid = TRUE;
	pSpiGrp->entry[0].valid = (edstIp1)?TRUE:FALSE;
	pSpiGrp->entry[0].ip = edstIp1;
	pSpiGrp->entry[0].spi = spi1;
	pSpiGrp->entry[0].proto = proto1;
	pSpiGrp->entry[1].valid = (edstIp2)?TRUE:FALSE;
	pSpiGrp->entry[1].ip = edstIp2;
	pSpiGrp->entry[1].spi = spi2;
	pSpiGrp->entry[1].proto = proto2;
	pSpiGrp->entry[2].valid = (edstIp3)?TRUE:FALSE;
	pSpiGrp->entry[2].ip = edstIp3;
	pSpiGrp->entry[2].spi = spi3;
	pSpiGrp->entry[2].proto = proto3;
	pSpiGrp->entry[3].valid = (edstIp4)?TRUE:FALSE;
	pSpiGrp->entry[3].ip = edstIp4;
	pSpiGrp->entry[3].spi = spi4;
	pSpiGrp->entry[3].proto = proto4;

	/* re-binding */
	_rtl8651_ipsecAutoLink();
	_rtl8651_arrangeAllACL();

	if ( CTAILQ_TOTAL( &IpsecEntry.inuseList.spigrp ) > rtl865x_ipsecMaxConcurrentSpiGrp )
	{
		rtl865x_ipsecMaxConcurrentSpiGrp = CTAILQ_TOTAL( &IpsecEntry.inuseList.spigrp );
	}

	retval = SUCCESS;
out:
	return retval;
}


/*
@func int32		| _rtl8651_delIpsecSpiGrp	| delete IPSEC SPI GRP
@parm ipaddr_t	| edstIp1		| Effective destination IP Address 1
@parm uint32	| spi1			| Outbound Security Parameter Index 1
@parm uint32	| proto1		| proto 1 (ESP or IPIP)
@rvalue SUCCESS	| success
@comm 
Given the first SA, this function will un-group the existing SPI group.
@devnote
 */
int32 _rtl8651_delIpsecSpiGrp( ipaddr_t edstIp1, uint32 spi1, uint32 proto1 )
{
	int32 retval;
	ipsecSpiGrp_t *pSpiGrp;

#if _IPSEC_DEBUG_>=1
	rtlglue_printf( "%s(0x%08x@"NIPQUAD2STR" %-4s)\n", __FUNCTION__, spi1, NIPQUAD2(edstIp1), strIpsecProto[proto1] );
#endif

	if ( _rtl8651_findIpsecSpiGrp( edstIp1, spi1, proto1, &pSpiGrp ) != SUCCESS )
	{
		/* not found, return. */
		retval = TBLDRV_EIPSEC_SPIGRP_NOT_FOUNT;
		goto out;
	}

	memset( pSpiGrp, 0, offsetof( ipsecSpiGrp_t, next ) );
	CTAILQ_REMOVE( &IpsecEntry.inuseList.spigrp, pSpiGrp, next );
	CTAILQ_INSERT_TAIL(&IpsecEntry.freeList.spigrp, pSpiGrp, next );

	/* re-binding */
	_rtl8651_ipsecAutoLink();
	_rtl8651_arrangeAllACL();
	
	retval = SUCCESS;
out:
	return retval;
}


/*
@func int32		| rtl8651_addIpsecSpiGrp	| add IPSEC SPI GRP
@parm ipaddr_t	| edstIp1		| Effective destination IP Address 1
@parm uint32	| spi1			| Outbound Security Parameter Index 1
@parm uint32	| proto1		| proto 1 (ESP or IPIP)
@parm ipaddr_t	| edstIp2		| Effective destination IP Address 2
@parm uint32	| spi2			| Outbound Security Parameter Index 2
@parm uint32	| proto2		| proto 2 (ESP or IPIP)
@parm ipaddr_t	| edstIp3		| Effective destination IP Address 3
@parm uint32	| spi3			| Outbound Security Parameter Index 3
@parm uint32	| proto3		| proto 3 (ESP or IPIP)
@parm ipaddr_t	| edstIp4		| Effective destination IP Address 4
@parm uint32	| spi4			| Outbound Security Parameter Index 4
@parm uint32	| proto4		| proto 4 (ESP or IPIP)
@parm uint32	| flags			| some flags for different configuration
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 rtl8651_addIpsecSpiGrp( ipaddr_t edstIp1, uint32 spi1, uint32 proto1,
                              ipaddr_t edstIp2, uint32 spi2, uint32 proto2,
                              ipaddr_t edstIp3, uint32 spi3, uint32 proto3,
                              ipaddr_t edstIp4, uint32 spi4, uint32 proto4,
                              uint32 flags )
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_addIpsecSpiGrp( edstIp1, spi1, proto1,
                                      edstIp2, spi2, proto2,
                                      edstIp3, spi3, proto3,
                                      edstIp4, spi4, proto4,
                                      flags, NULL );
	rtlglue_drvMutexUnlock();
	return retval;
}


/*
@func int32		| rtl8651_delIpsecSpiGrp	| delete IPSEC SPI GRP
@parm ipaddr_t	| edstIp1		| Effective destination IP Address 1
@parm uint32	| spi1			| Outbound Security Parameter Index 1
@parm uint32	| proto1		| proto 1 (ESP or IPIP)
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 rtl8651_delIpsecSpiGrp( ipaddr_t edstIp1, uint32 spi1, uint32 proto1 )
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_delIpsecSpiGrp( edstIp1, spi1, proto1 );
	rtlglue_drvMutexUnlock();
	return retval;
}


/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * -=  EROUTE Function
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*
@func int32		| _rtl8651_findIpsecERoute	| find IPSEC ERoute 
@parm ipaddr_t	| srcIp		| source IP address
@parm ipaddr_t	| srcIpMask	| source IP Mask
@parm ipaddr_t	| dstIp		| destination IP address
@parm ipaddr_t	| dstIpMask	| destination IP Mask
@parm ipsecERoute_t**	| pRet		| Found eroute entry
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_findIpsecERoute( ipaddr_t srcIp, ipaddr_t srcIpMask,
                                ipaddr_t dstIp, ipaddr_t dstIpMask,
                                ipsecERoute_t **pRet )
{
	int32 retval = SUCCESS;
	ipsecERoute_t *pERoute;
	
	CTAILQ_FOREACH( pERoute, &IpsecEntry.inuseList.eroute, next )
	{
		assert( pERoute->valid );

		if ( pERoute->srcIp == srcIp &&
		     pERoute->srcIpMask == srcIpMask &&
		     pERoute->dstIp == dstIp &&
		     pERoute->dstIpMask == dstIpMask )
		{
			/* found */
			if ( pRet ) *pRet = pERoute;
			goto out;
		}
	}

	/* not found */
	if ( pRet ) *pRet = NULL;
	retval = TBLDRV_EIPSEC_EROUTE_NOT_FOUNT;
out:
	return retval;
}


/*
@func int32		| _rtl8651_findIpsecERouteBySrcDstIp	| find IPSEC ERoute 
@parm ipaddr_t	| srcIp		| source IP address
@parm ipaddr_t	| dstIp		| destination IP address
@parm ipsecERoute_t**	| pRet		| Found eroute entry
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_findIpsecERouteBySrcDstIp( ipaddr_t srcIp, ipaddr_t dstIp, ipsecERoute_t **pRet )
{
	int32 retval = SUCCESS;
	ipsecERoute_t *pERoute;
	
	CTAILQ_FOREACH( pERoute, &IpsecEntry.inuseList.eroute, next )
	{
		assert( pERoute->valid );

		/*rtlglue_printf("srcIp="NIPQUAD2STR"("NIPQUAD2STR")  dstIp="NIPQUAD2STR"("NIPQUAD2STR")\n", 
		               NIPQUAD2(srcIp), NIPQUAD2(pERoute->srcIp), NIPQUAD2(dstIp), NIPQUAD2(pERoute->dstIp) );*/
		if ( ( srcIp & pERoute->srcIpMask ) == ( pERoute->srcIp & pERoute->srcIpMask ) &&
		     ( dstIp & pERoute->dstIpMask ) == ( pERoute->dstIp & pERoute->dstIpMask ) )
		{
			/* found */
			if ( pRet ) *pRet = pERoute;
			goto out;
		}
	}

	/* not found */
	if ( pRet ) *pRet = NULL;
	retval = TBLDRV_EIPSEC_EROUTE_NOT_FOUNT;
out:
	return retval;
}


/*
@func int32		| _rtl8651_findIpsecERouteBySrcDstIp	| find IPSEC ERoute 
@parm ipaddr_t	| srcIp		| source IP address (remote gateway)
@parm ipaddr_t	| dstIp		| Effective destination IP address (our gateway)
@parm ipsecERoute_t**	| pRet		| Found eroute entry
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_findIpsecERouteBySrcEdstIp( ipaddr_t srcIp, ipaddr_t edstIp, ipsecERoute_t **pRet )
{
	int32 retval = SUCCESS;
	ipsecERoute_t *pERoute;
	
	CTAILQ_FOREACH( pERoute, &IpsecEntry.inuseList.eroute, next )
	{
		assert( pERoute->valid );

		/*rtlglue_printf("srcIp="NIPQUAD2STR"("NIPQUAD2STR")  dstIp="NIPQUAD2STR"("NIPQUAD2STR")\n", 
		               NIPQUAD2(srcIp), NIPQUAD2(pERoute->srcIp), NIPQUAD2(dstIp), NIPQUAD2(pERoute->dstIp) );*/
		if ( ( pERoute->linkedSpi && srcIp == pERoute->linkedSpi->srcIp ) &&
		     ( edstIp == pERoute->edstIp  ) )
		{
			/* found */
			if ( pRet ) *pRet = pERoute;
			goto out;
		}
	}

	/* not found */
	if ( pRet ) *pRet = NULL;
	retval = TBLDRV_EIPSEC_EROUTE_NOT_FOUNT;
out:
	return retval;
}


/*
@func int32		| _rtl8651_addIpsecSpiGrp	| add IPSEC EROUTE
@parm ipaddr_t	| srcIp			| source IP address
@parm ipaddr_t	| srcIpMask		| source IP Mask
@parm ipaddr_t	| dstIp			| destination IP address
@parm ipaddr_t	| dstIpMask		| destination IP Mask
@parm ipaddr_t	| edstIp1		| Effective destination IP Address 1
@parm uint32	| spi1			| Outbound Security Parameter Index 1
@parm uint32	| proto1		| proto 1 (ESP or IPIP)
@parm uint32	| flags			| some flags for different configuration
@parm ipsecERoute_t**	| pRet	| Newly added spigrp entry
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_addIpsecERoute( ipaddr_t srcIp, ipaddr_t srcIpMask,
                               ipaddr_t dstIp, ipaddr_t dstIpMask,
                               ipaddr_t edstIp1, uint32 spi1, uint32 proto1,
                               uint32 flags,
                               ipsecERoute_t **pRet )
{
	int32 retval;
	ipsecERoute_t *pERoute;

#if _IPSEC_DEBUG_>=1
	rtlglue_printf( "%s(src="NIPQUAD2STR"/"NIPQUAD2STR", dst="NIPQUAD2STR"/"NIPQUAD2STR",\n", __FUNCTION__, NIPQUAD2(srcIp), NIPQUAD2(srcIpMask), NIPQUAD2(dstIp), NIPQUAD2(dstIpMask) );
	rtlglue_printf( "    %08x@"NIPQUAD2STR" %-4s,flags=0x%08x)\n", spi1, NIPQUAD2(edstIp1), strIpsecProto[proto1], flags );
#endif

	if ( _rtl8651_findIpsecERoute( srcIp, srcIpMask, dstIp, dstIpMask, &pERoute ) == SUCCESS )
	{
		/* found, clear pointer to force link. */
		pERoute->linkedSpi = NULL;
	}
	else
	{
		/* not found, alloc one. */
		if ( ( pERoute = CTAILQ_FIRST( &IpsecEntry.freeList.eroute ) ) )
		{
			/* chain to inuse list */
			CTAILQ_REMOVE( &IpsecEntry.freeList.eroute, pERoute, next );
			/* always chain to insert head, therefore the newer ERoute will be arranged to earlier ACL rule. */
			CTAILQ_INSERT_HEAD(&IpsecEntry.inuseList.eroute, pERoute, next );
		}
		else
		{
			retval = TBLDRV_EIPSEC_NO_MORE_EROUTE;
			goto out;
		}
	}

	pERoute->valid = TRUE;
#ifdef _LOUIS_DEBUG_
	pERoute->outbound = edstIp1==inet_addr("192.168.154.143")?FALSE:TRUE;
#else
	pERoute->outbound = _rtl8651_extIpAddr( edstIp1, NULL )?FALSE:TRUE;;
#endif
	pERoute->srcIp = srcIp;
	pERoute->srcIpMask = srcIpMask;
	pERoute->dstIp = dstIp;
	pERoute->dstIpMask = dstIpMask;
	pERoute->edstIp = edstIp1;
	pERoute->spi = spi1;
	pERoute->proto = proto1;

	/* re-binding */
	_rtl8651_ipsecAutoLink();
	_rtl8651_arrangeAllACL();

	if ( CTAILQ_TOTAL( &IpsecEntry.inuseList.eroute ) > rtl865x_ipsecMaxConcurrentERoute )
	{
		rtl865x_ipsecMaxConcurrentERoute = CTAILQ_TOTAL( &IpsecEntry.inuseList.eroute );
	}

	retval = SUCCESS;
out:
	return retval;
}


/*
@func int32		| _rtl8651_delIpsecERoute	| delete IPSEC EROUTE
@parm ipaddr_t	| srcIp			| source IP address
@parm ipaddr_t	| srcIpMask		| source IP Mask
@parm ipaddr_t	| dstIp			| destination IP address
@parm ipaddr_t	| dstIpMask		| destination IP Mask
@rvalue SUCCESS	| success
@comm 
Given the first SA, this function will un-group the existing SPI group.
@devnote
 */
int32 _rtl8651_delIpsecERoute( ipaddr_t srcIp, ipaddr_t srcIpMask,
                               ipaddr_t dstIp, ipaddr_t dstIpMask )
{
	int32 retval;
	ipsecERoute_t *pERoute;

#if _IPSEC_DEBUG_>=1
	rtlglue_printf( "%s(src="NIPQUAD2STR"/"NIPQUAD2STR", dst="NIPQUAD2STR"/"NIPQUAD2STR")\n", __FUNCTION__, NIPQUAD2(srcIp), NIPQUAD2(srcIpMask), NIPQUAD2(dstIp), NIPQUAD2(dstIpMask) );
#endif

	if ( _rtl8651_findIpsecERoute( srcIp, srcIpMask, dstIp, dstIpMask, &pERoute ) != SUCCESS )
	{
		/* not found, return. */
		retval = TBLDRV_EIPSEC_EROUTE_NOT_FOUNT;
		goto out;
	}

	memset( pERoute, 0, offsetof( ipsecERoute_t, next ) );
	CTAILQ_REMOVE( &IpsecEntry.inuseList.eroute, pERoute, next );
	CTAILQ_INSERT_TAIL(&IpsecEntry.freeList.eroute, pERoute, next );

	/* re-binding */
	_rtl8651_ipsecAutoLink();
	_rtl8651_arrangeAllACL();
	
	retval = SUCCESS;
out:
	return retval;
}


/*
@func int32		| rtl8651_addIpsecERoute	| add IPSEC EROUTE
@parm ipaddr_t	| srcIp			| source IP address
@parm ipaddr_t	| srcIpMask		| source IP Mask
@parm ipaddr_t	| dstIp			| destination IP address
@parm ipaddr_t	| dstIpMask		| destination IP Mask
@parm ipaddr_t	| edstIp1		| Effective destination IP Address 1
@parm uint32	| spi1			| Outbound Security Parameter Index 1
@parm uint32	| proto1		| proto 1 (ESP or IPIP)
@parm uint32	| flags			| some flags for different configuration
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 rtl8651_addIpsecERoute( ipaddr_t srcIp, ipaddr_t srcIpMask,
                              ipaddr_t dstIp, ipaddr_t dstIpMask,
                              ipaddr_t edstIp1, uint32 spi1, uint32 proto1,
                              uint32 flags )
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_addIpsecERoute( srcIp, srcIpMask,
	                                  dstIp, dstIpMask,
	                                  edstIp1, spi1, proto1,
	                                  flags, NULL );
	rtlglue_drvMutexUnlock();
	return retval;
}


/*
@func int32		| rtl8651_updateIpsecERoute	| update IPSEC EROUTE
@parm ipaddr_t	| srcIp			| source IP address
@parm ipaddr_t	| srcIpMask		| source IP Mask
@parm ipaddr_t	| dstIp			| destination IP address
@parm ipaddr_t	| dstIpMask		| destination IP Mask
@parm ipaddr_t	| edstIp1		| Effective destination IP Address 1
@parm uint32	| spi1			| Outbound Security Parameter Index 1
@parm uint32	| proto1		| proto 1 (ESP or IPIP)
@parm uint32	| flags			| some flags for different configuration
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 rtl8651_updateIpsecERoute( ipaddr_t srcIp, ipaddr_t srcIpMask,
                                 ipaddr_t dstIp, ipaddr_t dstIpMask,
                                 ipaddr_t edstIp1, uint32 spi1, uint32 proto1,
                                 uint32 flags )
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_addIpsecERoute( srcIp, srcIpMask,
	                                  dstIp, dstIpMask,
	                                  edstIp1, spi1, proto1,
	                                  flags, NULL );
	rtlglue_drvMutexUnlock();
	return retval;
}


/*
@func int32		| rtl8651_delIpsecSpiGrp	| add IPSEC SPI GRP
@parm ipaddr_t	| srcIp			| source IP address
@parm ipaddr_t	| srcIpMask		| source IP Mask
@parm ipaddr_t	| dstIp			| destination IP address
@parm ipaddr_t	| dstIpMask		| destination IP Mask
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 rtl8651_delIpsecERoute( ipaddr_t srcIp, ipaddr_t srcIpMask,
                              ipaddr_t dstIp, ipaddr_t dstIpMask )
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_delIpsecERoute( srcIp, srcIpMask, dstIp, dstIpMask );
	rtlglue_drvMutexUnlock();
	return retval;
}


/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * -=  Integraion Function
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*
@func int32		| _rtl8651_flushIpsec	| Flush all tables
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_flushIpsec( void )
{
	ipsecSpi_t *spi;
	ipsecSpiGrp_t *spigrp;
	ipsecERoute_t *eroute;

#if _IPSEC_DEBUG_>=1
	rtlglue_printf( "%s():%d\n", __FUNCTION__, __LINE__ );
#endif

	for( spi = CTAILQ_FIRST( &IpsecEntry.inuseList.spi );
	     spi;
	     spi = CTAILQ_FIRST( &IpsecEntry.inuseList.spi ) )
	{
		CTAILQ_REMOVE( &IpsecEntry.inuseList.spi, spi, next );
		CTAILQ_INSERT_TAIL( &IpsecEntry.freeList.spi, spi, next );
		memset( spi, 0, offsetof( ipsecSpi_t, next ) );
	}
	rtl865x_ipsecMaxConcurrentSpi = 0;

	for( spigrp = CTAILQ_FIRST( &IpsecEntry.inuseList.spigrp );
	     spigrp;
	     spigrp = CTAILQ_FIRST( &IpsecEntry.inuseList.spigrp ) )
	{
		CTAILQ_REMOVE( &IpsecEntry.inuseList.spigrp, spigrp, next );
		CTAILQ_INSERT_TAIL( &IpsecEntry.freeList.spigrp, spigrp, next );
		memset( spigrp, 0, offsetof( ipsecSpiGrp_t, next ) );
	}
	rtl865x_ipsecMaxConcurrentSpiGrp = 0;

	for( eroute = CTAILQ_FIRST( &IpsecEntry.inuseList.eroute );
	     eroute;
	     eroute = CTAILQ_FIRST( &IpsecEntry.inuseList.eroute ) )
	{
		CTAILQ_REMOVE( &IpsecEntry.inuseList.eroute, eroute, next );
		CTAILQ_INSERT_TAIL( &IpsecEntry.freeList.eroute, eroute, next );
		memset( eroute, 0, offsetof( ipsecERoute_t, next ) );
	}
	rtl865x_ipsecMaxConcurrentERoute = 0;

	return SUCCESS;
}


/*
@func int32		| rtl8651_flushIpsec	| Flush all tables
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 rtl8651_flushIpsec( void )
{
	int32 retval;
	rtlglue_drvMutexLock();
	retval = _rtl8651_flushIpsec();
	rtlglue_drvMutexUnlock();
	return retval;
}


/*
@func int32		| _rtl8651_ipsecAutoLink	| Automatically link and unlink SPI/SPIGRP/EROUTE tables.
@rvalue SUCCESS	| success
@comm 
@devnote
 */
int32 _rtl8651_ipsecAutoLink( void )
{
	ipsecSpi_t *spi;
	ipsecSpiGrp_t *spigrp;
	ipsecERoute_t *eroute;
	int32 i;

	for( spi = CTAILQ_FIRST( &IpsecEntry.inuseList.spi );
	     spi;
	     spi = CTAILQ_NEXT( spi, next ) )
	{
		if ( spi->valid==FALSE )
		{
			/* This entry is just invalidated. Delete it.
			 * Including:
			 *  "spi->valid==FALSE && spi->spiGrp==NULL" and
			 *  "spi->valid==FALSE && spi->spiGrp!=NULL"
			 */
			_rtl8651_delIpsecSpi( spi->edstIp, spi->spi, spi->proto );
		}
		else if ( spi->valid==TRUE && spi->spiGrp==NULL )
		{
			/* Try to link SPI Group */
			_rtl8651_findIpsecSpiGrp( spi->edstIp, spi->spi, spi->proto, &spi->spiGrp );
			/* printf(" spi->edstIp=%x, spi->spi=%x, spi->proto=%d, %p\n", spi->edstIp, spi->spi, spi->proto, spi->spiGrp ); */
		}
		else /*if ( spi->valid==TRUE && spi->spiGrp!=NULL )*/
		{
			if ( spi->spiGrp->valid==FALSE )
			{
				/* SPI Group has been invalid, unlink this pointer. */
				spi->spiGrp = NULL;
				/* printf(" spi->edstIp=%x, spi->spi=%x, spi->proto=%d, UNLINK\n", spi->edstIp, spi->spi, spi->proto ); */
			}
		}
	}

	for( spigrp = CTAILQ_FIRST( &IpsecEntry.inuseList.spigrp );
	     spigrp;
	     spigrp = CTAILQ_NEXT( spigrp, next ) )
	{
		if ( spigrp->valid==FALSE )
		{
			/* This entry is just invalidated. Delete it. */
			_rtl8651_delIpsecSpiGrp( spi->edstIp, spi->spi, spi->proto );
		}
		else
		{	/* spigrp->valid==TRUE */
			for( i = 0; i < sizeof(spigrp->entry)/sizeof(spigrp->entry[0]); i++ )
			{
				if ( spigrp->entry[i].valid )
				{
					if ( spigrp->entry[i].linkedSpi==NULL )
					{
						/* Pointer is null, try to link SPI Group */
						_rtl8651_findIpsecSpi( spigrp->entry[i].ip, spigrp->entry[i].spi,
						                       spigrp->entry[i].proto, &spigrp->entry[i].linkedSpi );
					}
					else
					{
						/* Pointer is valid, verify it. */
						if ( spigrp->entry[i].linkedSpi->valid==FALSE )
						{
							/* SPI Group has been invalid, unlink this pointer. */
							spigrp->entry[i].linkedSpi = NULL;
						}
						else
						{
							/* it is still valid, keep the link. */
						}
					}
				}
				else
				{
					/* entry is invalid, skip it. */
				}
			}
		}
	}

	for( eroute = CTAILQ_FIRST( &IpsecEntry.inuseList.eroute );
	     eroute;
	     eroute = CTAILQ_NEXT( eroute, next ) )
	{
		if ( eroute->valid==FALSE )
		{
			/* This entry is just invalidated. Delete it.
			 * Including:
			 *  "spi->valid==FALSE && spi->spiGrp==NULL" and
			 *  "spi->valid==FALSE && spi->spiGrp!=NULL"
			 */
			_rtl8651_delIpsecERoute( eroute->srcIp, eroute->srcIpMask, eroute->dstIp, eroute->dstIpMask );
		}
		else if ( eroute->valid==TRUE && eroute->linkedSpi==NULL )
		{
			/* Try to link SPI Group */
			_rtl8651_findIpsecSpi( eroute->edstIp, eroute->spi, eroute->proto, &eroute->linkedSpi );
		}
		else /*if ( eroute->valid==TRUE && eroute->linkedSpi!=NULL )*/
		{
			if ( eroute->linkedSpi->valid==FALSE )
			{
				/* SPI Group has been invalid, unlink this pointer. */
				eroute->linkedSpi = NULL;
			}
		}
	}

	return SUCCESS;
}





/****************************************************************
 *******   FOR ACL
 ****************************************************************/
/*
@func int32		| _rtl8651_tryToAddAclRuleForIpsecOutbound	| This function is called by _rtl8651_arrangeAllACL() 
@parm uint16	| vid		| vlan id
@parm uint32	| aclStart	| Starting ACL number
@parm uint32	| aclEnd	| Ending ACL number
@rvalue 0		| no more space to add ACL rule
@rvalue other	| total number of added ACL rule 
@comm 
@devnote
 */
uint32 _rtl8651_tryToAddAclRuleForIpsecOutbound( uint16 vid, uint32 aclStart, uint32 aclEnd, int8 *enoughFlag)
{
	_rtl8651_tblDrvAclRule_t rule;
	ipsecERoute_t *eroute;
	int32 addCount = 0;
	int32 retval;

	CTAILQ_FOREACH( eroute, &IpsecEntry.inuseList.eroute, next )
	{
		assert( eroute->valid );

		if ( eroute->outbound &&
		     eroute->linkedSpi )
		{
			
			/* internal LAN --> opponent LAN */
			if (aclStart > aclEnd) {
				*enoughFlag = 0;	// Not enough space for adding IPSec ACL rules.
				goto out;
			}
			bzero(&rule, sizeof(_rtl8651_tblDrvAclRule_t));
			rule.ruleType_				= RTL8651_ACL_IP;
			rule.srcIpAddr_				= eroute->srcIp;
			rule.srcIpAddrMask_			= eroute->srcIpMask;
			rule.dstIpAddr_				= eroute->dstIp;
			rule.dstIpAddrMask_			= eroute->dstIpMask;
			rule.tos_					= 0;
			rule.tosMask_				= 0;
			rule.ipProto_				= 0;
			rule.ipProtoMask_			= 0;
			rule.ipFlag_				= 0;
			rule.ipFlagMask_			= 0;
			rule.ipFOP_					= 0;
			rule.ipFOM_					= 0;
			rule.ipHttpFilter_			= 0;
			rule.ipHttpFilterM_			= 0;
			rule.ipIdentSrcDstIp_		= 0;
			rule.ipIdentSrcDstIpM_		= 0;
			rule.ipDF_					= 0;
			rule.ipMF_					= 0;
			rule.actionType_			= RTL8651_ACL_CPU;
			retval = rtl8651_setAsicAclRule(aclStart, &rule);
			assert( retval == SUCCESS );
			
			eroute->linkedSpi->aclRule = aclStart;
			_rtl8651_registerACLDB( aclStart, _rtl8651_encapIpsec, eroute );
			addCount++;
			aclStart++;
		}
	}
out:
	return addCount;
}


/*
@func int32		| _rtl8651_tryToAddAclRuleForIpsecInbound	| This function is called by _rtl8651_arrangeAllACL() 
@parm uint16	| vid		| vlan id
@parm uint32	| aclStart	| Starting ACL number
@parm uint32	| aclEnd	| Ending ACL number
@rvalue 0		| no more space to add ACL rule
@rvalue other	| total number of added ACL rule 
@comm 
@devnote
 */
uint32 _rtl8651_tryToAddAclRuleForIpsecInbound( uint16 vid, uint32 aclStart, uint32 aclEnd, int8 *enoughFlag )
{
	_rtl8651_tblDrvAclRule_t rule;
	ipsecERoute_t *eroute;
	int32 addCount = 0;
	int32 retval;

	CTAILQ_FOREACH( eroute, &IpsecEntry.inuseList.eroute, next )
	{
		assert( eroute->valid );

		if ( eroute->outbound==FALSE &&
		     eroute->linkedSpi )
		{
			/* opponent LAN --> internal LAN */
			if (aclStart > aclEnd) {
				*enoughFlag = 0;	// Not enough space for adding IPSec ACL rules.
				goto out;
			}
			bzero(&rule, sizeof(_rtl8651_tblDrvAclRule_t));
			rule.ruleType_				= RTL8651_ACL_IP;
			rule.srcIpAddr_				= eroute->linkedSpi->srcIp;
			rule.srcIpAddrMask_			= 0xffffffff;
			rule.dstIpAddr_				= eroute->edstIp;
			rule.dstIpAddrMask_			= 0xffffffff;
			rule.tos_					= 0;
			rule.tosMask_				= 0;
			rule.ipProto_				= IPPROTO_ESP;
			rule.ipProtoMask_			= 0xff;
			rule.ipFlag_				= 0;
			rule.ipFlagMask_			= 0;
			rule.ipFOP_					= 0;
			rule.ipFOM_					= 0;
			rule.ipHttpFilter_			= 0;
			rule.ipHttpFilterM_			= 0;
			rule.ipIdentSrcDstIp_		= 0;
			rule.ipIdentSrcDstIpM_		= 0;
			rule.ipDF_					= 0;
			rule.ipMF_					= 0;
			rule.actionType_			= RTL8651_ACL_CPU;
			retval = rtl8651_setAsicAclRule(aclStart, &rule);
			assert( retval == SUCCESS );
			
			eroute->linkedSpi->aclRule = aclStart;
			_rtl8651_registerACLDB( aclStart, _rtl8651_decapIpsec, eroute );
			addCount++;
			aclStart++;
		}
	}
out:
	return addCount;
}


/****************************************************************
 *******   FOR FWD_ENGINE
 ****************************************************************/
int32 _rtl8651_mtuFragment2( struct rtl_pktHdr *pkthdrPtr, uint32 mtu, 
                             struct rtl_pktHdr** retPktHdr, int sizeOfRet )
{
	int32 count = 0;
	struct ip *ip = (struct ip *)(pkthdrPtr->ph_mbuf->m_data + pkthdrPtr->ph_iphdrOffset);
	uint32 local_ph_reason;

#if _IPSEC_DEBUG_>=4
	rtlglue_printf("[%s():%d] need fragment(mtu=%d)\n", __FUNCTION__, __LINE__, mtu );
#endif

	local_ph_reason = 0x0064;

	switch (local_ph_reason & 0x007f) {
		struct rtl_mBuf *new_m;
		struct ip *new_ip;
		int32 offset;
		int32 ip_hlen;
		uint8 *cp1;
		uint32 payloadLen, hdrLen;
		uint32 segLen;
		uint32 seg_seq;
	case 0x0064: /* unicast over MTU */		
		/* get ip header length */
		ip_hlen = (ip->ip_vhl &0xf) << 2;

		/* re-generate ip/tcp/udp/icmp checksum */
		if (IS_FRAGMENT(ip->ip_off))
			pkthdrPtr->ph_proto = PKTHDR_IP;

		switch(pkthdrPtr->ph_proto){
			struct tcphdr *tcp_p;
			struct udphdr*udp_p;
			struct icmp *icmp_p;
			int32 totalLen ;
		case PKTHDR_TCP:
			totalLen = ntohs(ip->ip_len);
			tcp_p = (struct tcphdr *) ((int8 *) ip + ip_hlen);
			tcp_p->th_sum = 0;
			ip->ip_sum = 0;
			tcp_p->th_sum = htons(_rtl8651_tcpChecksum(ip));
			ip->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)ip));
			break;
		case PKTHDR_UDP:
			udp_p = (struct udphdr *) ((int8 *) ip + ip_hlen);
			totalLen = ntohs(ip->ip_len);
			ip->ip_sum = 0;
			if (udp_p->uh_sum)
			{	/* don't re-calculate checksum if udp checksum is zero */
				udp_p->uh_sum = 0;
				udp_p->uh_sum = htons(_rtl8651_tcpChecksum(ip));
			}
			ip->ip_sum = htons(_rtl8651_ipChecksum(ip));
			break;
		case PKTHDR_ICMP:	
			totalLen = ntohs(ip->ip_len);
			icmp_p = (struct icmp *) ((int8 *) ip + ip_hlen);			
			icmp_p->icmp_cksum = 0;
			ip->ip_sum = 0;		
			icmp_p->icmp_cksum = htons(_rtl8651_icmpChecksum(ip));
			ip->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)ip));
			break;
#if 0	/* if ip header length is not 20, we can say that the ip checksum is done by previous process */
		case PKTHDR_IP:
			totalLen = ntohs(ip->ip_len);
			if (!IPHDRLEN_IS_TWENTY(ip))
			{
				ip->ip_sum = 0;
				ip->ip_sum = htons(_rtl8651_ipChecksum(ip));
			}
			break;
#endif
		default:
			break;
		}


#if 0
		/* Original Packet for callback */
		if ((rtl8651_drvIcmpRoutingMsgEnable == TRUE) &&
			(rtl8651_drvIcmpAlwaysReplyErrForFrag == TRUE || DONT_FRAG(ip->ip_off))) {
			struct ip *iphdr;
			rtl8651_tblDrv_vlanTable_t *vlan_t;
			rtl8651_tblDrv_arpEntry_t *arp_t;
//			ether_addr_t src_mac;
			uint32 ip_hlen, totalLen, ipAddr;
//			uint8 *temp_buf;

			if ((new_m=mBuf_dupPacket(pkthdrPtr->ph_mbuf, MBUF_DONTWAIT)) == NULL) 
				return FAILED;
			orgPkthdr = new_m->m_pkthdr;

			/* flush iphdr setting:
					FIXME:	we assume the sender of this fragmented packet is via VLAN-type interface,
							so ICMP error message will NEVER add pppoe header now. */
			orgPkthdr->ph_flags &= ~(PKTHDR_PPPOE_AUTOADD|PKTHDR_VLAN_AUTOADD|CSUM_IP|CSUM_L4);

			iphdr = (struct ip *)(new_m->m_data + orgPkthdr->ph_iphdrOffset);
			iphdr->ip_ttl ++;
			if (isNapt == TRUE) {
				PKTSET_UINT32_UNALIGNED(orgPkthdr->ph_sip, &iphdr->ip_src);
			}
			else {
				rtl8651_memcpy(&ipAddr, &iphdr->ip_src, 4);		
				rtl8651_memcpy(&iphdr->ip_src, &iphdr->ip_dst, 4);	
				rtl8651_memcpy(&iphdr->ip_dst, &ipAddr, 4);			
			}
			ip_hlen = (iphdr->ip_vhl &0xf) << 2;
			totalLen = ntohs(iphdr->ip_len);
			
			/* original ether header */
			if ((arp_t=_rtl8651_getArpEntry(NULL, orgPkthdr->ph_sip, FALSE))) {
				vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(orgPkthdr->ph_vlanIdx)];
				rtl8651_memcpy((new_m->m_data+6), &arp_t->macInfo->macAddr, 6);
				rtl8651_memcpy(new_m->m_data, &vlan_t->macAddr, 6);
			}
			
			switch(orgPkthdr->ph_proto){
				struct tcphdr *tcp_p;
				struct udphdr*udp_p;
				struct icmp *icmp_p;
			case PKTHDR_TCP:
				tcp_p = (struct tcphdr *) ((int8 *) iphdr + ip_hlen);
				tcp_p->th_sport = htons(orgPkthdr->ph_sport);
				tcp_p->th_sum = 0;
				iphdr->ip_sum = 0;
				tcp_p->th_sum = htons(_rtl8651_tcpChecksum((struct ip*)iphdr));
				iphdr->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)iphdr));
				break;
				
			case PKTHDR_UDP:
				udp_p = (struct udphdr *) ((int8 *) iphdr + ip_hlen);
				udp_p->uh_sport = htons(orgPkthdr->ph_sport);
				iphdr->ip_sum = 0;
				if (udp_p->uh_sum)
				{
					udp_p->uh_sum = 0;
					udp_p->uh_sum= htons(_rtl8651_tcpChecksum((struct ip*)iphdr));
				}
				iphdr->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)iphdr));			
				break;
				
			case PKTHDR_ICMP:
				icmp_p = (struct icmp *) ((int8 *) iphdr + ip_hlen);
				icmp_p->icmp_id = htons(orgPkthdr->ph_sid);
				icmp_p->icmp_cksum = 0;
				iphdr->ip_sum = 0;		
				icmp_p->icmp_cksum = htons(_rtl8651_icmpChecksum((struct ip*)iphdr));
				iphdr->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)iphdr));
				break;

			case PKTHDR_IP:
				if (!IPHDRLEN_IS_TWENTY(iphdr))
				{
					iphdr->ip_sum = 0;
					iphdr->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)iphdr));
				}
			default: break;

			}
			/* run callback function */
			if (p_mtuHandler)
			{
				p_mtuHandler(orgPkthdr);
			}else
			{
				_rtl8651_IcmpPktField_t icmpInfo;
				icmpInfo.ic_type	= ICMP_UNREACH;
				icmpInfo.ic_code	= ICMP_UNREACH_NEEDFRAG;
				icmpInfo.ic_mtu	= mtu;
				vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(orgPkthdr->ph_vlanIdx)];
				orgPkthdr->ph_portlist = srcPortList;
				if (_rtl8651_drvIcmpErrorGeneration(orgPkthdr, iphdr, &icmpInfo, vlan_t->vid) != SUCCESS)
				{
					mBuf_freeMbufChain(orgPkthdr->ph_mbuf);
				}
			}
		}
#endif
		
		if (DONT_FRAG(ip->ip_off))
		{
#if _IPSEC_DEBUG_>=2
			rtlglue_printf("[%s():%d] Don't Frag bit is set.\n", __FUNCTION__, __LINE__ );
#endif
			retPktHdr[count++] = pkthdrPtr;
			retPktHdr[count] = NULL;
			return FAILED;
		}
		
		/* get ip pointer offset */
		offset = (uint8 *)ip - (uint8 *)pkthdrPtr->ph_mbuf->m_data;
		if(offset <= 0 ){
#if 0
			if (orgPkthdr)
				mBuf_freeMbufChain(orgPkthdr->ph_mbuf);
#endif
#if _IPSEC_DEBUG_>=2
			rtlglue_printf("[%s():%d] IP offset is negtive.\n", __FUNCTION__, __LINE__ );
#endif
			retPktHdr[count++] = pkthdrPtr;
			retPktHdr[count] = NULL;
			return FAILED;
		}

		hdrLen = offset + ip_hlen;			/* ether header + ip header length */
		payloadLen = ntohs(ip->ip_len) - ip_hlen;		/* payload length */
		segLen = ((mtu - ip_hlen)>>3)<<3; 	/* MTU not include ether header: 8-byte alignment */
		
		for(seg_seq=0 ; payloadLen > segLen; payloadLen-=segLen, seg_seq++) {

			if ( count >= sizeOfRet-1-1 ) /* we need to reserve the last one and terminator pointer(NULL). */
			{
#if _IPSEC_DEBUG_>=2
				rtlglue_printf("[%s():%d] retPktHdr[%d] is not big enough.\n", __FUNCTION__, __LINE__, sizeOfRet );
#endif
				retPktHdr[count++] = pkthdrPtr;
				retPktHdr[count] = NULL;
				return FAILED;
			}
		
			if ((new_m=mBuf_dupPacket(pkthdrPtr->ph_mbuf, MBUF_DONTWAIT)) == NULL) {
				//rtlglue_printf("enter %s %d\n",__FUNCTION__,__LINE__);
#if 0
				if (orgPkthdr)
					mBuf_freeMbufChain(orgPkthdr->ph_mbuf);
#endif
#if _IPSEC_DEBUG_>=2
				rtlglue_printf("[%s():%d] mBuf_dupPacket() failed.\n", __FUNCTION__, __LINE__ );
#endif
				retPktHdr[count++] = pkthdrPtr;
				retPktHdr[count] = NULL;
				return FAILED;
			}
			{
				/* Louis: chain it ... */
#if 1
				retPktHdr[count] = new_m->m_pkthdr;
				retPktHdr[count+1] = NULL;
#else
				struct rtl_pktHdr *prev = pkthdrPtr->ph_nextHdr;
				pkthdrPtr->ph_nextHdr = new_m->m_pkthdr;
				new_m->m_pkthdr->ph_nextHdr = prev;
#endif
				count++;
			}

#if 0
			new_m->m_pkthdr->ph_flags &= ~(CSUM_L4);
			if (ip_hlen == 20)
				new_m->m_pkthdr->ph_flags |= (CSUM_IP);
			else
				new_m->m_pkthdr->ph_flags &= ~(CSUM_IP);
#endif
			if (seg_seq > 0) {
				cp1 = (uint8 *)new_m->m_data + (seg_seq*segLen);
				rtl8651_memcpy(cp1, (uint8 *)new_m->m_data, hdrLen);
				mBuf_adjHead(new_m,  segLen*seg_seq);
			}
			mBuf_adjTail(new_m,  payloadLen-segLen);
			/* get new ip header address */
			new_ip = (struct ip *)(new_m->m_data + offset);
			{
				uint16 ip_len = ntohs(new_ip->ip_len);
				uint16 ip_off =ntohs(ip->ip_off);
				/* modify iplen */
				ip_len = ip_hlen + segLen;
				new_ip->ip_len = htons(ip_len);
				/* modify offset */
				ip_off = seg_seq*(segLen>>3);
				new_ip->ip_off += htons(ip_off) ;
				new_ip->ip_off |= htons(IP_MF);
			}	
			assert(new_m->m_pkthdr);
			new_ip->ip_sum = 0;
			new_ip->ip_sum = ntohs(_rtl8651_ipChecksum( new_ip ));
			//new_m->m_pkthdr->ph_rxdesc = -3;
#if 0
			new_m->m_pkthdr->ph_proto = PKTHDR_IP;
			if ((retval=_rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)(new_m->m_pkthdr),dvid,-1)) != SUCCESS) {
				mBuf_freeMbufChain(new_m);
#if 0				
				if (orgPkthdr)
					mBuf_freeMbufChain(orgPkthdr->ph_mbuf);
#endif
				return retval;
			}
#endif
		}


		cp1 = (uint8 *)pkthdrPtr->ph_mbuf->m_data + (seg_seq*segLen);
		rtl8651_memcpy(cp1, (uint8 *)pkthdrPtr->ph_mbuf->m_data, hdrLen);
		mBuf_adjHead(pkthdrPtr->ph_mbuf,  segLen*seg_seq);
		/* get new ip header address */
		ip = (struct ip *)(pkthdrPtr->ph_mbuf->m_data + offset);
		{
			uint16 ip_len = ntohs(ip->ip_len);
			uint16 ip_off =ntohs(ip->ip_off);
			/* modify iplen */
			ip_len = ip_hlen + payloadLen;
			ip->ip_len = htons(ip_len);
			/* modify offset */
			ip_off = seg_seq*(segLen>>3);
			ip->ip_off += htons(ip_off);
		}
		ip->ip_sum = 0;
		ip->ip_sum = ntohs(_rtl8651_ipChecksum( ip ));

#if 0
		pkthdrPtr->ph_flags &= ~(CSUM_L4);
		if (ip_hlen == 20)
			pkthdrPtr->ph_flags |= (CSUM_IP);
		else
			pkthdrPtr->ph_flags &= ~(CSUM_IP);
#endif
		retPktHdr[count] = pkthdrPtr;
		count++;
#if 0
		if (p_mtuHandler) {
			if ((retval = _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdrPtr,dvid,-1)) != SUCCESS) {
				mBuf_freeMbufChain(orgPkthdr->ph_mbuf);
				return retval;
			}
			p_mtuHandler(orgPkthdr);
			return retval;
		}
		pkthdrPtr->ph_proto = PKTHDR_IP;
#endif
#if 0
		return _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdrPtr,dvid,-1);
#endif
		break;
		
	case 0x0066: /* multicast over MTU */	 
#if _IPSEC_DEBUG_>=2
		rtlglue_printf("[%s():%d] multicast over MTU.\n", __FUNCTION__, __LINE__ );
#endif
		retPktHdr[count++] = pkthdrPtr;
		retPktHdr[count] = NULL;
		return FAILED;
	}

#if _IPSEC_DEBUG_>=4
	rtlglue_printf("[%s():%d] count=%d\n", __FUNCTION__, __LINE__, count );
#endif

#ifdef RTL865X_TEST
	{
		/* Dump fragmented packet */
		struct rtl_pktHdr *pHdr;
		int i = 0;

		while( retPktHdr[i] )
		{
#if _IPSEC_DEBUG_>=4
			rtlglue_printf("[%s():%d] packet%d: ph_len=%d\n", __FUNCTION__, __LINE__, i, retPktHdr[i]->ph_len );
#endif
#if _IPSEC_DEBUG_>=5
			dumpMbufPkt( retPktHdr[i]->ph_mbuf );
#endif
			i++;
		}
	}
#endif

	retPktHdr[count] = NULL; /* terminated */

	return SUCCESS;
}


/*
@func int32	| _rtl8651_encapIpsec	| This function is called by FWDEng to encap packet for IPSec tunnel. 
@rvalue SUCCESS	| This packet has been sent.
@rvalue FAILED	| This packet is not sent.
@comm 
@devnote
 */
int32 _rtl8651_encapIpsec( int32 ruleNo, struct rtl_pktHdr *pkthdr, struct ip *pip, void *userDefined )
{
	ipsecERoute_t *eroute = (ipsecERoute_t *)userDefined;
	ipsecSpi_t *linkedspi;
	ipsecSpiGrp_t *spigrp;
	ipsecSpi_t *spi;
	struct ether_header copyEthHdr;
	struct ip *ip1, *ip2;
	/*int32 authPadLen;*/
	int32 cryptoPadLen;
	uint8 *pAuthData;
	espHdr_t *esphdr;
	cbcHdr128_t *cbchdr; /* Use maximum IV structure */
	uint8 newDigest[RTL8651_MAX_IPSEC_KEY_LENGTH];
	uint16 ip2Len;
	int32 i;
	struct rtl_pktHdr *fragedPackets[16];
	int32 fragSeq = 0;
	int32 retval;
	rtl8651_tblDrv_pppTable_t * pppoeEntry;
	int16 mtu;
	int16 smallerMtu;
	

	if ( pip==NULL )
	{
		rtlglue_printf("[%s():%d] exception, pip is NULL!\n", __FUNCTION__, __LINE__ );
		retval = FAILED;
		goto out;
	}

	/*if( (((uint32)pip)&0xf0000000)!=0xa0000000 ) printk("pip=%p\n",pip);*/
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_BEFORE_CRYPTO_ENCAP);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_ENCAP);
#endif

#if _IPSEC_DEBUG_>=3
	rtlglue_printf("[%s():%d] ruleNo=%d, pip->ip_src=%d.%d.%d.%d, pip->ip_dst=%d.%d.%d.%d, pTunnel->remIp=%d.%d.%d.%d, Len=%d\n", __FUNCTION__, __LINE__, 
	               ruleNo, NIPQUAD2(ntohl(pip->ip_src.s_addr)), NIPQUAD2(ntohl(pip->ip_dst.s_addr)),
	               NIPQUAD2(eroute->edstIp), ntohs(pip->ip_len) );
#endif
#if _IPSEC_DEBUG_>=5
	memDump( (uint8*)IPSEC_UNCACHED_ADDR(pkthdr->ph_mbuf->m_data), pkthdr->ph_mbuf->m_len, ">>>>> IPSec (Origin)");
#endif

	assert( eroute );

	if ( ( eroute->valid==TRUE ) &&
	     ( linkedspi = eroute->linkedSpi ) &&
	     ( spigrp = linkedspi->spiGrp ) &&
 	     ( spi = spigrp->entry[1].linkedSpi ) )
	{
		/* all needed pointer is ready and valid. */
	}
	else
	{
#if _IPSEC_DEBUG_>=2
		rtlglue_printf("[%s():%d] eroute is not valid(valid=%d,linkedSpi=%p,spiGrp=%p,spi=%p)\n", __FUNCTION__, __LINE__, 
		                eroute->valid,
		                eroute->linkedSpi,
		                eroute->linkedSpi?eroute->linkedSpi->spiGrp:NULL,
		                eroute->linkedSpi&&eroute->linkedSpi->spiGrp?eroute->linkedSpi->spiGrp->entry[1].linkedSpi:NULL );
#endif
		if ( eroute->valid )
		{
			eroute->counter.cntInvalidEroute++;
		}
		retval = FAILED;
		goto out;
	}

	/* Get WAN's MTU */
	pppoeEntry = _rtl8651_getPppoeSessionByIpAddr(spi->srcIp);
	if (pppoeEntry)
		mtu = pppoe_mss[pppoeEntry->dsid] + sizeof(struct ip) + sizeof(struct tcphdr);
	else
		mtu = rtl8651_mss + sizeof(struct ip) + sizeof(struct tcphdr);

	if ( rtl865x_ipsecMtu < mtu )
	{
		smallerMtu = mtu - rtl865x_ipsecMtu;
		mtu = rtl865x_ipsecMtu;
	}
	else
	{
		smallerMtu = 0; /* mtu is equal to WAN's mtu. */
	}

	/* Modify MSS */
	if ( pip->ip_p == IPPROTO_TCP && IS_FRAGMENT_REMAIN(pip->ip_off)==0 )
	{
		struct tcphdr *tc;
		
		tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
		if (tc->th_flags & TH_SYN)
		{
			_rtl8651_l4TcpAdjustMss( pip, tc, eroute->edstIp, 
			                         -( _TUNNEL_OVERHEAD_LEN_(spi->cryptoType) + _MAX_POSSIBLE_PADLEN_(spi->cryptoType) + 
			                            rtl8651_authDataLen[spi->authType] + smallerMtu ) );
		}
	}

	/* check if the packet is needed to be fragment. */
	{
#if _IPSEC_DEBUG_>=4
		rtlglue_printf("[%s():%d] pktLen=%d mtu=%d(maxPayload=%d).\n", __FUNCTION__, __LINE__, 
		               ntohs(pip->ip_len), mtu,
		               mtu-(_TUNNEL_OVERHEAD_LEN_(spi->cryptoType)+_MAX_POSSIBLE_PADLEN_(spi->cryptoType)+rtl8651_authDataLen[spi->authType]) );
#endif
		if ( ntohs(pip->ip_len) > mtu-(_TUNNEL_OVERHEAD_LEN_(spi->cryptoType)+_MAX_POSSIBLE_PADLEN_(spi->cryptoType)+
				                       rtl8651_authDataLen[spi->authType]) )
		{
#if _IPSEC_DEBUG_>=4
			rtlglue_printf("[%s():%d] packet is too big, we need fragment.\n", __FUNCTION__, __LINE__ );
#endif
			/* packet is too big, we need fragment. */
			retval = _rtl8651_mtuFragment2( pkthdr, 
			                                mtu-(_TUNNEL_OVERHEAD_LEN_(spi->cryptoType)+_MAX_POSSIBLE_PADLEN_(spi->cryptoType)+rtl8651_authDataLen[spi->authType]), 
			                                fragedPackets, sizeof(fragedPackets)/sizeof(fragedPackets[0]) );
			if ( retval != SUCCESS )
			{
#if _IPSEC_DEBUG_>=2
				rtlglue_printf("[%s():%d] _rtl8651_mtuFragment2() FAILED, free all packets.\n", __FUNCTION__, __LINE__ );
#endif
				while( ( pkthdr = fragedPackets[fragSeq++] ) )
				{
					mBuf_freeMbufChain( pkthdr->ph_mbuf );		
				}

				eroute->counter.cntFragError++;
				retval = SUCCESS;
				goto out;
			}
		}
		else
		{
			/* packet is small, thus we uniform pkthdr into fragedPackets[]. */
			fragedPackets[0] = pkthdr;
			fragedPackets[1] = NULL;
		}
	}

nextPkt:
	pkthdr = fragedPackets[fragSeq];
	if ( pkthdr == NULL )
	{
		retval = SUCCESS;
		goto out;
	}
	/* update pip */
	pip = (struct ip *) &(((char*)pkthdr->ph_mbuf->m_data)[sizeof(struct ether_header)]);

	/* assign space */
	ip2 = IPSEC_UNCACHED_ADDR(pip);
	ip1 = (struct ip*)((char*)ip2-_TUNNEL_OVERHEAD_LEN_(spi->cryptoType));
	esphdr = (espHdr_t *)(((uint8*)ip1)+sizeof(struct ip));
	cbchdr = (cbcHdr128_t *)(((uint8*)esphdr)+sizeof(espHdr_t));
	
	/*
	 *  OriginIP | padding | padLen | NextHeader|
	 *           |<------- cryptoPadLen ------->|
	 *               include tailing 2 bytes
	 */
	ip2Len = ntohs(ip2->ip_len); /* save length before encryption */
	/*authPadLen = ((ip2Len+8+63)&~63)-(ip2Len);*/
	cryptoPadLen = ((ip2Len+2+rtl8651_cryptoMaxPadLen[spi->cryptoType]-1)&~(rtl8651_cryptoMaxPadLen[spi->cryptoType]-1))-(ip2Len);
	assert( cryptoPadLen>=2 && (cryptoPadLen<=rtl8651_cryptoMaxPadLen[spi->cryptoType]+1) );
	pAuthData = ((uint8*)ip2) + ip2Len + cryptoPadLen;

#if _IPSEC_DEBUG_>=4
	rtlglue_printf("ip2->ip_len=%d, cryptoPadLen=%d\n",
	               ip2Len, cryptoPadLen );
#endif

	/* Encap the packet */
	copyEthHdr = *(struct ether_header*)pkthdr->ph_mbuf->m_data;
	pkthdr->ph_mbuf->m_data -= _TUNNEL_OVERHEAD_LEN_(spi->cryptoType); /* move m_data */
	pkthdr->ph_mbuf->m_len += _TUNNEL_OVERHEAD_LEN_(spi->cryptoType) + cryptoPadLen + rtl8651_authDataLen[spi->authType];
	pkthdr->ph_len += _TUNNEL_OVERHEAD_LEN_(spi->cryptoType) + cryptoPadLen + rtl8651_authDataLen[spi->authType];
	*(struct ether_header*)pkthdr->ph_mbuf->m_data = copyEthHdr;

	/* build new ip header */
	//*ip1 = *ip2;
	ip1->ip_vhl = 0x45;
	ip1->ip_tos = 0x00;
	ip1->ip_len = ntohs( ip2Len + _TUNNEL_OVERHEAD_LEN_(spi->cryptoType) + cryptoPadLen + rtl8651_authDataLen[spi->authType] );
	ip1->ip_id = swaps16(ntohs(fragSeq)) ^ ip2->ip_id;
	ip1->ip_off = 0;//ip2->ip_off;
	ip1->ip_ttl = ip2->ip_ttl;
	ip1->ip_p = IPPROTO_ESP;
	//ip1->ip_sum = 
	pkthdr->ph_flags |= CSUM_IP;
	pkthdr->ph_flags &= ~CSUM_L4;
	pkthdr->ph_proto = PKTHDR_IP;
	ip1->ip_src.s_addr = ntohl( spi->srcIp );
	ip1->ip_dst.s_addr = ntohl( spi->edstIp );

	/***********************************************************************
	 * Generate ESP and IV
	 ***********************************************************************/
	esphdr->spi = ntohl( spi->spi );
	esphdr->seq = ntohl( spi->seq );
	cbchdr->iv[0] = ntohl( (ip2->ip_id<<16) | (~ip2->ip_id) );
	cbchdr->iv[1] = ~cbchdr->iv[0];
	switch ( spi->cryptoType )
	{
		case IPSEC_CBC_AES256:
		case IPSEC_CBC_AES192:
		case IPSEC_CBC_AES128:
			cbchdr->iv[2] = ntohl( (ip2->ip_id<<16) | (~ip2->ip_id) ) ^ esphdr->spi;
			cbchdr->iv[3] = cbchdr->iv[0] ^ esphdr->seq;
			break;
	}
	
	/***********************************************************************
	 * Encryption
	 ***********************************************************************/
	/*if ( cryptoPadLen >= 2 )*/
	{
		for( i = 0; i < cryptoPadLen-2; i++ )
		{
			((char*)ip2)[ip2Len+i] = i+1; /* padding */
		}
		((char*)ip2)[ip2Len+(i++)] = cryptoPadLen-2; /* padding length */
		((char*)ip2)[ip2Len+(i++)] = IPPROTO_IPIP; /* next_header */
	}
	
#if _IPSEC_DEBUG_>=5
	memDump( (uint8*)IPSEC_UNCACHED_ADDR(pkthdr->ph_mbuf->m_data), pkthdr->ph_mbuf->m_len, ">>>>> IPSec (Before Cipher)");
#endif
	
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_ENCAP_CRYPTO_ENGINE);
#endif

#ifdef RTL865X_TEST
	{	/* XOR the packet for debug ... */
		int i;
		char* pAddr = (char*)ip2;
		int len = ip2Len+cryptoPadLen;
		
		for( i = 0; i < len; i++ )
		{
			pAddr[i] = pAddr[i] ^ 0xff;
		}
		retval = SUCCESS;
	}
#else
	switch ( spi->cryptoType )
	{
		case IPSEC_CBC_DES:
		case IPSEC_CBC_3DES:
		case IPSEC_ECB_DES:
		case IPSEC_ECB_3DES:
			retval = rtl8651b_cryptoEngine_des( (spi->cryptoType-1)|4/*encrypto*/,
			                                    (char*)ip2,
			                                    ip2Len+cryptoPadLen,
			                                    spi->cryptoKey,
			                                    (int8*)cbchdr->iv );
			break;
			
		case IPSEC_CBC_AES128:
		case IPSEC_CBC_AES192:
		case IPSEC_CBC_AES256:
			{
				cbcHdr128_t cbchdrBackup; /* since AES_cbc_encrypt() will modify ivec, we must backup it. */
				
				cbchdrBackup = *cbchdr;
				retval = AES_set_encrypt_key( spi->cryptoKey, rtl8651_cryptoKeyLen[spi->cryptoType]*8, &rtl8651_AES_key );
				if ( retval != 0 ) { rtlglue_printf("[%s:%d] AES_set_encrypt_key()=%d\n", __FUNCTION__, __LINE__, retval ); }

#ifdef _SW_AES_USE_DRAM_
				memcpy( rtl8651_AES_buf, ip2, ip2Len+cryptoPadLen );
				memcpy( rtl8651_AES_iv, cbchdr->iv, sizeof(cbchdr->iv) );
				AES_cbc_encrypt( rtl8651_AES_buf, rtl8651_AES_buf, ip2Len+cryptoPadLen, &rtl8651_AES_key, rtl8651_AES_iv, AES_ENCRYPT );
				memcpy( ip2, rtl8651_AES_buf, ip2Len+cryptoPadLen );
#else
				AES_cbc_encrypt( (char*)ip2, (char*)ip2, ip2Len+cryptoPadLen, &rtl8651_AES_key, (uint8*)cbchdr->iv, AES_ENCRYPT );
#endif

				*cbchdr = cbchdrBackup;
				retval = SUCCESS;
			}
			break;
			
		default:
			retval = FAILED;
			assert( 0 );
	}
#endif

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_ENCAP_CRYPTO_ENGINE);
#endif

	if ( retval != SUCCESS )
	{
		mBuf_freeMbufChain( pkthdr->ph_mbuf );
		eroute->counter.cntCryptoEngineError++;
		retval = SUCCESS;
		goto PktFailed;
	}

#if _IPSEC_DEBUG_>=5
	memDump( (uint8*)IPSEC_UNCACHED_ADDR(pkthdr->ph_mbuf->m_data), pkthdr->ph_mbuf->m_len, ">>>>> IPSec (After Cipher, but no authdata)");
#endif
	
	/***********************************************************************
	 * Generate Authentication Data (ICV)
	 ***********************************************************************/
	if ( spi->authDataPresent )
	{
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_ENCAP_AUTH_ENGINE);
#endif

#ifdef RTL865X_TEST
	{	/* authorized by fixed pattern */
		memset( newDigest, 0x55, rtl8651_authDataLen[spi->authType] );
		retval = SUCCESS;
	}
#else
		retval = rtl8651b_authEngine( (spi->authType-1) | RTL8651_AUTH_IOPAD_READY, 
		                              (char*)esphdr,
		                              ip2Len+sizeof(espHdr_t)+rtl8651_cryptoIVLen[spi->cryptoType]+cryptoPadLen,
		                              (uint8*)IPSEC_UNCACHED_ADDR(spi->ipad), 0/*rtl8651_authKeyLen[spi->authType]*/, 
		                              (uint8*)IPSEC_UNCACHED_ADDR(newDigest) );
#endif

#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfExitPoint(ROMEPERF_INDEX_ENCAP_AUTH_ENGINE);
#endif

		if ( retval != SUCCESS )
		{
			mBuf_freeMbufChain( pkthdr->ph_mbuf );
			eroute->counter.cntAuthEngineError++;
			retval = SUCCESS;
			goto PktFailed;
		}

		/* append digest to end of packet */
		memcpy( pAuthData, (uint8*)IPSEC_UNCACHED_ADDR(newDigest), rtl8651_authDataLen[spi->authType] );

#if _IPSEC_DEBUG_>=4
		memDump( (uint8*)IPSEC_UNCACHED_ADDR(pkthdr->ph_mbuf->m_data), pkthdr->ph_mbuf->m_len, ">>>>> IPSec (After Cipher && authdata)");
#endif
	}

	pkthdr->ph_routeIdx = 0; /* invalid routing cache */
	rtl8651_drvPacketFastSend = TRUE;	/* Enter fast mode */
	retval = _rtl8651_l3FastActRouting(FWDENG_FROMDRV, pkthdr, ntohl(ip1->ip_dst.s_addr), ip1, NULL );
	rtl8651_drvPacketFastSend = FALSE;	/* Exit fast mode */
	if ( retval == SUCCESS )
	{
		/* packet routed, */
#if _IPSEC_DEBUG_>=4
		rtlglue_printf("packet routed\n");
#endif
		spi->seq++;
		eroute->counter.cntSuccess++;
		retval = SUCCESS;
		goto PktFailed;
	}
	else
	{
		/* routing failed .... drop it */
#if _IPSEC_DEBUG_>=2
		rtlglue_printf("routing failed\n");
#endif
		mBuf_freeMbufChain( pkthdr->ph_mbuf );
		eroute->counter.cntRouteFailed++;
		retval = SUCCESS;
		goto PktFailed;
	}

PktFailed:
	fragSeq++;
	goto nextPkt;
	
out:

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_ENCAP);
#endif

	return retval;
}


/*
@func int32	| _rtl8651_decapIpsec	| This function is called by FWDEng to encap packet for IPSec tunnel. 
@rvalue SUCCESS	| This packet has been sent.
@rvalue FAILED	| This packet is not sent.
@comm 
@devnote
 */
int32 _rtl8651_decapIpsec( int32 ruleNo, struct rtl_pktHdr *pkthdr, struct ip *pip, void *userDefined )
{
	ipsecERoute_t *eroute = (ipsecERoute_t *)userDefined;
	ipsecSpi_t *linkedspi;
	ipsecSpiGrp_t *spigrp;
	ipsecSpi_t *spi;
	espHdr_t *esphdr;
	uint32 newSeq;
	uint32 shiftOffsetOfNewPacket;
	uint8 backupDigest[RTL8651_MAX_IPSEC_KEY_LENGTH];
	uint8 verifyDigest[RTL8651_MAX_IPSEC_KEY_LENGTH];
	struct ether_header backupEthHdr;
	int32 retval;

	if ( pip==NULL )
	{
		rtlglue_printf("[%s():%d] exception, pip is NULL!\n", __FUNCTION__, __LINE__ );
		retval = FAILED;
		goto out;
	}

	/*if( (((uint32)pip)&0xf0000000)!=0xa0000000 ) printk("pip=%p\n",pip);*/
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_BEFORE_CRYPTO_DECAP);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_DECAP);
#endif
	
#if _IPSEC_DEBUG_>=3
	rtlglue_printf("[%s():%d] ruleNo=%d, pip->ip_src=%d.%d.%d.%d, pip->ip_dst=%d.%d.%d.%d, pTunnel->remIp=%d.%d.%d.%d, Len=%d\n", __FUNCTION__, __LINE__, 
	               ruleNo, NIPQUAD2(ntohl(pip->ip_src.s_addr)), NIPQUAD2(ntohl(pip->ip_dst.s_addr)), 
	               NIPQUAD2((eroute->edstIp)), ntohs(pip->ip_len) );
#endif

	if ( ( eroute->valid==TRUE ) &&
	     ( linkedspi = eroute->linkedSpi ) &&
	     ( spigrp = linkedspi->spiGrp ) &&
 	     ( spi = spigrp->entry[1].linkedSpi ) )
	{
		/* all needed pointer is ready and valid. */
	}
	else
	{
#if _IPSEC_DEBUG_>=2
		rtlglue_printf("[%s():%d] eroute is not valid(valid=%d,linkedSpi=%p,spiGrp=%p)\n", __FUNCTION__, __LINE__, 
		                eroute->valid,
		                eroute->linkedSpi,
		                eroute->linkedSpi?eroute->linkedSpi->spiGrp:NULL);
#endif
		retval = FAILED;
		goto out;
	}

#if _IPSEC_DEBUG_>=4
	rtlglue_printf("[%s():%d] ip_len=%d  crypto_len=%d  auth_len=%d\n", __FUNCTION__, __LINE__, 
					ntohs(pip->ip_len), 
					ntohs(pip->ip_len)-(sizeof(struct ip)+sizeof(espHdr_t)+rtl8651_cryptoIVLen[spi->cryptoType]+rtl8651_authDataLen[spi->authType]), 
					ntohs(pip->ip_len)-(sizeof(struct ip)+rtl8651_authDataLen[spi->authType]) );
#endif

	pip = (struct ip *)IPSEC_UNCACHED_ADDR(pip);
	esphdr = (espHdr_t*)((char*)pip+sizeof(struct ip));
#if _IPSEC_DEBUG_>=5
	memDump( pip, ntohs(pip->ip_len), "<<<<< IPSec (Original)");
#endif

	/***********************************************************************
	 * Check SPI 
	 ***********************************************************************/
	if ( esphdr->spi != spi->spi )
	{
		ipsecSpi_t *newSpi;
#if _IPSEC_DEBUG_>=3
		rtlglue_printf("[%s():%d] SPI not match (pkt=0x%08x, pTunnel=0x%08x), maybe peer uses new SPI.\n", __FUNCTION__, __LINE__, esphdr->spi, spi->spi );
#endif
		if ( _rtl8651_findIpsecSpi( ntohl(pip->ip_dst.s_addr), esphdr->spi, IPSEC_ESP, &newSpi )==SUCCESS )
		{
#if _IPSEC_DEBUG_>=3
			rtlglue_printf("[%s():%d] spi is changed from 0x%08x to 0x%08x.\n", __FUNCTION__, __LINE__,
			               spi->spi, newSpi->spi );
#endif
			spi = newSpi;
		}
		else
		{
#if _IPSEC_DEBUG_>=2
			rtlglue_printf("[%s():%d] SPI not match (pkt=0x%08x, pTunnel=0x%08x), go on for PS.\n", __FUNCTION__, __LINE__, esphdr->spi, spi->spi );
#endif
			retval = FAILED;
			goto out;
		}
	}
	
	/***********************************************************************
	 * Anti-replay 
	 *  RFC2406 - 3.4.3 Sequence Number Verification
	 *  To enable anti-replay check, authentication must be enabled (for integrity).
	 *
	 ***********************************************************************/
	newSeq = ntohl( esphdr->seq );
	shiftOffsetOfNewPacket = newSeq - spi->seq;
	if ( spi->authDataPresent && spi->IHaveAntiReplay )
	{
		/* check if it is in the left side. */
		if ( newSeq <= ( spi->seq - spi->replayWindow ) )
		{
#if _IPSEC_DEBUG_>=2
			rtlglue_printf("[%s():%d] Sequence number is in the left side of window (pkt=%d, newest=%d, winSize=%d), drop it.\n", __FUNCTION__, __LINE__,
			               newSeq, spi->seq, spi->replayWindow );
#endif

#ifdef _DISMISS_TO_FWDENG_/* Orlando and chenyl ask Louis: never free packet even if it is wrong. Because the packet may be IPSec passthru. */
			eroute->counter.cntErrorSeq++;
			retval = FAILED;
#else
			mBuf_freeMbufChain( pkthdr->ph_mbuf );
			eroute->counter.cntErrorSeq++;
			retval = SUCCESS;
#endif
			goto out;
		}
		else if ( newSeq > spi->seq )
		{
			/* new seq, do authenticated later. */
		}
		else 
		{
			/* the seq is in the window, check if it is duplicated. */
			shiftOffsetOfNewPacket = spi->seq - newSeq;
			if ( spi->recvWindow[shiftOffsetOfNewPacket>>5] & (1<<(shiftOffsetOfNewPacket&0x1f )) )
			{
				
#if _IPSEC_DEBUG_>=2
				rtlglue_printf("[%s():%d] Sequence number is duplicated (pkt=%d, newest=%d), drop it.\n", __FUNCTION__, __LINE__,
				               newSeq, spi->seq );
#endif

#ifdef _DISMISS_TO_FWDENG_ /* Orlando and chenyl ask Louis: never free packet even if it is wrong. Because the packet may be IPSec passthru. */
				eroute->counter.cntDupSeq++;
				retval = FAILED;
#else
				mBuf_freeMbufChain( pkthdr->ph_mbuf );
				eroute->counter.cntDupSeq++;
				retval = SUCCESS;
#endif
				goto out;
			}
		}
	}
	 
	/***********************************************************************
	 * If authentication data is present in ESP packet, check it. 
	 ***********************************************************************/
	if ( spi->authDataPresent )
	{
		/*
		 * Since authEngine will modify the original packet,
		 *   we must back the authentication data.
		 */
		memcpy( backupDigest, &(((char*)pip)[ntohs(pip->ip_len)-rtl8651_authDataLen[spi->authType]]), rtl8651_authDataLen[spi->authType] );

#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_DECAP_AUTH_ENGINE);
#endif

#ifdef RTL865X_TEST
		{	/* authorized by fixed pattern */
			memset( verifyDigest, 0x55, rtl8651_authDataLen[spi->authType] );
			retval = SUCCESS;
		}
#else
		retval = rtl8651b_authEngine( (spi->authType-1) | RTL8651_AUTH_IOPAD_READY, 
		                              ((char*)pip)+sizeof(struct ip),
		                              ntohs(pip->ip_len)-(sizeof(struct ip)+rtl8651_authDataLen[spi->authType]),
		                              (uint8*)IPSEC_UNCACHED_ADDR(spi->ipad), 0/*rtl8651_authKeyLen[spi->authType]*/, 
		                              (uint8*)IPSEC_UNCACHED_ADDR(verifyDigest) );
		if ( retval != SUCCESS )
		{
#ifdef _DISMISS_TO_FWDENG_ /* Orlando and chenyl ask Louis: never free packet even if it is wrong. Because the packet may be IPSec passthru. */
			eroute->counter.cntAuthEngineError++;
			retval = FAILED;
#else
			mBuf_freeMbufChain( pkthdr->ph_mbuf );
			eroute->counter.cntAuthEngineError++;
			retval = SUCCESS;
#endif
			goto out;
		}
#endif

#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfExitPoint(ROMEPERF_INDEX_DECAP_AUTH_ENGINE);
#endif

#if _IPSEC_DEBUG_>=5
		memDump( backupDigest, rtl8651_authDataLen[spi->authType], "backup Digest" );
		memDump( IPSEC_UNCACHED_ADDR(verifyDigest), rtl8651_authDataLen[spi->authType], "verify Digest" );
#endif

		if ( memcmp( backupDigest, IPSEC_UNCACHED_ADDR(verifyDigest), rtl8651_authDataLen[spi->authType] ) == 0 )
		{
			/* the same, authenticated. */
#if _IPSEC_DEBUG_>=4
			rtlglue_printf("Good! Authenticated!\n" );
#endif
		}
		else
		{
			/* NOT the same, failed to authenticate. */
#if _IPSEC_DEBUG_>=2
			rtlglue_printf("BAD BAD BAD! Authentication data is not identical!\n" );
#endif

#ifdef _DISMISS_TO_FWDENG_ /* Orlando and chenyl ask Louis: never free packet even if it is wrong. Because the packet may be IPSec passthru. */
			eroute->counter.cntErrorAuth++;
			retval = FAILED;
#else
			mBuf_freeMbufChain( pkthdr->ph_mbuf );
			eroute->counter.cntErrorAuth++;
			retval = SUCCESS;
#endif
			goto out;
		}
	}

	/* packet authenticated, update recvSeq */
	/* _rtl8651_ipsecShiftWindow( spi->recvWindow, shiftOffsetOfNewPacket ); */
	{
		int32 shiftBytes, shiftBits;
		int32 i;
		
		/* mark received */
		if ( newSeq > spi->seq )
		{
			/* new window range */
			shiftBits = shiftOffsetOfNewPacket & 0x1f;
			shiftBytes = shiftOffsetOfNewPacket >> 5;
			
			for( i = min( sizeof(spi->recvWindow)/sizeof(spi->recvWindow[0]), 
			              (spi->replayWindow+31)/(sizeof(spi->recvWindow[0])*8) ) - 1;
			     i >= shiftBytes;
			     i-- )
			{
				spi->recvWindow[i] = spi->recvWindow[i-shiftBytes]<<shiftBits;
				if ( ( ( i-shiftBytes ) != 0 ) && ( shiftBits > 0 ) )
					spi->recvWindow[i] |= spi->recvWindow[i-shiftBytes-1]>>(32-shiftBits);
			}
			/* clear new-spanned recv window */
			for( ;
			     i >= 0;
			     i-- )
			{
				spi->recvWindow[i] = 0;
			}

			/* mark the newest packet */
			spi->recvWindow[0] |= 0x00000001; 

			/* unmask the bits ouitside the replay window. */
			if ( spi->replayWindow < MAX_WINDOW_SIZE )
				spi->recvWindow[spi->replayWindow>>5] &= ~(0xffffffff<<(spi->replayWindow&0x1f));

			/* forward the right-side boundary. */ 
			spi->seq = newSeq;
		}
		else
		{
			/* in-order seq */
			shiftOffsetOfNewPacket = spi->seq - newSeq;
			spi->recvWindow[shiftOffsetOfNewPacket>>5] |= (1<<(shiftOffsetOfNewPacket&0x1f ));
		}
	}

	/***********************************************************************
	 * Ensure it is IPIP packet.
	 ***********************************************************************/
	if ( pip->ip_p != IPPROTO_ESP )
	{
#if _IPSEC_DEBUG_>=2
		rtlglue_printf("This is not an IPIP packet! ip_p=0x%02x\n", pip->ip_p );
#endif

#ifdef _DISMISS_TO_FWDENG_ /* Orlando and chenyl ask Louis: never free packet even if it is wrong. Because the packet may be IPSec passthru. */
		eroute->counter.cntNotAnESPPacket++;
		retval = FAILED;
#else
		mBuf_freeMbufChain( pkthdr->ph_mbuf );
		eroute->counter.cntNotAnESPPacket++;
		retval = SUCCESS;
#endif
		goto out;
	}
	
	/***********************************************************************
	 * Descryption
	 ***********************************************************************/
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_DECAP_CRYPTO_ENGINE);
#endif

#ifdef RTL865X_TEST
	{	/* XOR the packet for debug ... */
		int i;
		char* pAddr = ((char*)pip)+sizeof(struct ip)+sizeof(espHdr_t)+rtl8651_cryptoIVLen[spi->cryptoType];
		int len = ntohs(pip->ip_len)-(sizeof(struct ip)+sizeof(espHdr_t)+rtl8651_cryptoIVLen[spi->cryptoType]+rtl8651_authDataLen[spi->authType]);
		
		for( i = 0; i < len; i++ )
		{
			pAddr[i] = pAddr[i] ^ 0xff;
		}
		retval = SUCCESS;
	}
#else
	switch ( spi->cryptoType )
	{
		case IPSEC_CBC_DES:
		case IPSEC_CBC_3DES:
		case IPSEC_ECB_DES:
		case IPSEC_ECB_3DES:
			retval = rtl8651b_cryptoEngine_des( (spi->cryptoType-1)|0,
			                                    ((char*)pip)+sizeof(struct ip)+sizeof(espHdr_t)+rtl8651_cryptoIVLen[spi->cryptoType],
			                                    ntohs(pip->ip_len)-(sizeof(struct ip)+sizeof(espHdr_t)+rtl8651_cryptoIVLen[spi->cryptoType]+rtl8651_authDataLen[spi->authType]),
			                                    spi->cryptoKey,
			                                    ((char*)pip)+sizeof(struct ip)+sizeof(espHdr_t) );
			break;
			
		case IPSEC_CBC_AES128:
		case IPSEC_CBC_AES192:
		case IPSEC_CBC_AES256:
			{
				char* ip2 = ((char*)pip)+sizeof(struct ip)+sizeof(espHdr_t)+rtl8651_cryptoIVLen[spi->cryptoType];
				int32 cryptoLen = ntohs(pip->ip_len)-(sizeof(struct ip)+sizeof(espHdr_t)+rtl8651_cryptoIVLen[spi->cryptoType]+rtl8651_authDataLen[spi->authType]);
				char* iv = ((char*)pip)+sizeof(struct ip)+sizeof(espHdr_t);
				
				retval = AES_set_decrypt_key( spi->cryptoKey, rtl8651_cryptoKeyLen[spi->cryptoType]*8, &rtl8651_AES_key );
				if ( retval != 0 ) { rtlglue_printf("[%s:%d] AES_set_decrypt_key()=%d\n", __FUNCTION__, __LINE__, retval ); }

#ifdef _SW_AES_USE_DRAM_
				memcpy( rtl8651_AES_buf, ip2, cryptoLen );
				memcpy( rtl8651_AES_iv, iv, sizeof(iv) );
				AES_cbc_encrypt( rtl8651_AES_buf, rtl8651_AES_buf, cryptoLen, &rtl8651_AES_key, iv, AES_DECRYPT );
				memcpy( ip2, rtl8651_AES_buf, cryptoLen );
#else
				AES_cbc_encrypt( (char*)ip2, (char*)ip2, cryptoLen, &rtl8651_AES_key, iv, AES_DECRYPT );
#endif
				
				retval = SUCCESS;
			}
			break;
			
		default:
			retval = FAILED;
			assert( 0 );
	}
	
	if ( retval != SUCCESS )
	{
#ifdef _DISMISS_TO_FWDENG_ /* Orlando and chenyl ask Louis: never free packet even if it is wrong. Because the packet may be IPSec passthru. */
		eroute->counter.cntCryptoEngineError++;
		retval = FAILED;
#else
		mBuf_freeMbufChain( pkthdr->ph_mbuf );
		eroute->counter.cntCryptoEngineError++;
		retval = SUCCESS;
#endif
		goto out;
	}
#endif

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_DECAP_CRYPTO_ENGINE);
#endif
	
#if _IPSEC_DEBUG_>=5
	memDump( pip, ntohs(pip->ip_len), ">>>>> IPSec (After Decipher)");
#endif

	/***********************************************************************
	 * We had decipher the packet, now, forward it.
	 ***********************************************************************/
	
	/* Decap the packet */
	backupEthHdr = *(struct ether_header*)pkthdr->ph_mbuf->m_data;
	pkthdr->ph_mbuf->m_data += _TUNNEL_OVERHEAD_LEN_(spi->cryptoType); /* move m_data */
	pkthdr->ph_mbuf->m_len -= _TUNNEL_OVERHEAD_LEN_(spi->cryptoType);
	pkthdr->ph_len -= _TUNNEL_OVERHEAD_LEN_(spi->cryptoType);
	*(struct ether_header*)pkthdr->ph_mbuf->m_data = backupEthHdr;

	/*
	 * We only route IP packet. (Louis: It must be IP, because it is IP-in-IP.
	 */
	if ( 1 )
	{
		rtl8651_tblDrv_pppTable_t * pppoeEntry;
		int16 mtu;
		int16 smallerMtu;
	
		/* pip is updated to inner IP */
		pip = (struct ip*)((char*)pip+_TUNNEL_OVERHEAD_LEN_(spi->cryptoType));
		
		if ( ( ntohl(pip->ip_src.s_addr) & eroute->srcIpMask ) != ( eroute->srcIp & eroute->srcIpMask ) )
		{
			/* 
			 * Source IP address is not matched with eroute, it is fine for multiple IPSEC tunnel.
			 * Because the source and destination IP addresses in ACL rules are identical.
			 * So, we should use the whole tunnel pair to query database again.
			 */
			ipsecERoute_t *eroute2;

			if ( _rtl8651_findIpsecERouteBySrcDstIp( ntohl(pip->ip_src.s_addr),
			                                         ntohl(pip->ip_dst.s_addr),
			                                         &eroute2 ) == SUCCESS )
			{
#if _IPSEC_DEBUG_>=3
				rtlglue_printf("Packet source does not match EROUTE SA (src:"NIPQUAD2STR"-->"NIPQUAD2STR")! But it is fine that we find it out in database.\n", NIPQUAD2(pip->ip_src.s_addr), NIPQUAD2(pip->ip_dst.s_addr) );
#endif
				assert( eroute2 );
				if ( ( ntohl(pip->ip_src.s_addr) & eroute2->srcIpMask ) == ( eroute2->srcIp & eroute2->srcIpMask ) )
				{
					/* matched, nothing happen */
					eroute = eroute2;
					goto sa_ok;
				}
			}

#if _IPSEC_DEBUG_>=2
			rtlglue_printf("Packet source does not match SA! pkt:"NIPQUAD2STR" srcIp:"NIPQUAD2STR"/"NIPQUAD2STR"\n",
			               NIPQUAD2(ntohl(pip->ip_src.s_addr)), NIPQUAD2(eroute->srcIp), NIPQUAD2(eroute->srcIpMask));
#endif

#ifdef _DISMISS_TO_FWDENG_ /* Orlando and chenyl ask Louis: never free packet even if it is wrong. Because the packet may be IPSec passthru. */
			eroute->counter.cntSourceAddressNotMatched++;
			retval = FAILED;
#else
			mBuf_freeMbufChain( pkthdr->ph_mbuf );
			eroute->counter.cntSourceAddressNotMatched++;
			retval = SUCCESS;
#endif
			goto out;
		}
sa_ok:

#if 0
		pkthdr->ph_mbuf->m_len = pkthdr->ph_len = pip->ip_len+sizeof(struct ether_header); /* re-gen length, due to encryption padding */
#endif

		/* Get WAN's MTU */
		pppoeEntry = _rtl8651_getPppoeSessionByIpAddr(spi->srcIp);
		if (pppoeEntry)
			mtu = pppoe_mss[pppoeEntry->dsid] + sizeof(struct ip) + sizeof(struct tcphdr);
		else
			mtu = rtl8651_mss + sizeof(struct ip) + sizeof(struct tcphdr);

		if ( rtl865x_ipsecMtu < mtu )
		{
			smallerMtu = mtu - rtl865x_ipsecMtu;
			mtu = rtl865x_ipsecMtu;
		}
		else
		{
			smallerMtu = 0; /* mtu is equal to WAN's mtu. */
		}
		
		/* Modify MSS */
		if ( pip->ip_p == IPPROTO_TCP && IS_FRAGMENT_REMAIN(pip->ip_off)==0 )
		{
			struct tcphdr *tc;
			
			tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
			if (tc->th_flags & TH_SYN)
			{
				_rtl8651_l4TcpAdjustMss( pip, tc, eroute->edstIp, 
				                         -( _TUNNEL_OVERHEAD_LEN_(spi->cryptoType) + _MAX_POSSIBLE_PADLEN_(spi->cryptoType) + 
				                            rtl8651_authDataLen[spi->authType] + smallerMtu ) );
			}
		}

		/* The inner IP address might be multicast address, therefore,
		 *   we must update DMAC also.
		 */
		{	
			/* chenyl: patch for multicast forward */
			ipaddr_t dest_ip = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
			if (IN_MULTICAST(dest_ip))
			{
				MCAST_IP2MAC(dest_ip, pkthdr->ph_mbuf->m_data);
			}
		}

		pkthdr->ph_routeIdx = 0; /* invalid routing cache */
		rtl8651_drvPacketFastSend = TRUE;	/* Enter fast mode */
		retval = _rtl8651_l3FastActRouting(FWDENG_FROMDRV, pkthdr, ntohl(pip->ip_dst.s_addr), pip, NULL );
		rtl8651_drvPacketFastSend = FALSE;	/* Exit fast mode */
		if ( retval == SUCCESS )
		{
			/* packet routed, */
#if _IPSEC_DEBUG_>=4
			rtlglue_printf("packet routed\n" );
#endif
			eroute->counter.cntSuccess++;
			retval = SUCCESS;
			goto out;
		}
		else
		{
			/* routing failed .... drop it */
#if _IPSEC_DEBUG_>=2
			rtlglue_printf("routing failed\n");
#endif

#ifdef _DISMISS_TO_FWDENG_ /* Orlando and chenyl ask Louis: never free packet even if it is wrong. Because the packet may be IPSec passthru. */
			eroute->counter.cntRouteFailed++;
			retval = FAILED;
#else
			mBuf_freeMbufChain( pkthdr->ph_mbuf );
			eroute->counter.cntRouteFailed++;
			retval = SUCCESS;
#endif
			goto out;
		}
	}

	assert( 0 ); /* Should never come here. */
	retval = FAILED;
out:

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_DECAP);
#endif

	return retval;
}


