/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : rtl8651_layer2.c
* Abstract : RTL8651 Home gateway controller Layer2  driver
* Author : Chun-Feng Liu(cfliu@realtek.com.tw)
* $Id: rtl8651_layer2.c,v 1.1.1.1 2006/07/26 05:38:23 blue Exp $
*/
#include "rtl_types.h"
#include "assert.h"
#include "rtl_errno.h"
#include "types.h"
#include "rtl8651_layer2.h"
#include "rtl8651_tblAsicDrv.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "asicRegs.h"
#include "asicTabs.h"
#include "rtl_queue.h"
#include "rtl_utils.h"
#ifdef RTL865X_TEST
#include <string.h>
#endif
#ifdef RTL865X_MODEL_USER
#include "rtl_glue.h"
#include "rtl8651_tblDrvPatch.h"
#include <stdio.h>
#include <string.h>
#endif

uint32 _sysUpSeconds; /* System uptime in seconds */
uint32 rtl8651_tblDrvCpuMirrorPortMask;
uint32 phyPortLinkStatus;

/* SPT control */
uint32 rtl8651_tblDrvSpanningTreeEnable;
uint32 rtl8651_tblDrvSpanningTreePortFwdStatus;

uint32 arpAgingTime; /* default 5 mins */
rtl8651_tblDrvInitPara_t rtl8651_tblDrvPara;
__DRAM_L2_FWD struct rtl8651_driverTables DrvTbl;
int8 pureSoftFwd;	/* TRUE: pure software forwarding. When pureSoftFwd is TRUE, naptAutoLearn is implicitly disabled even though naptAutoLearn=TRUE
				 			 * FALSE: software and ASIC cooperation*/

struct AsicTableUsage_S AsicTableUsage;

uint32 rtl8651_l2protoPassThrough = 0;	//bitmask of Layer2 protocol passthrough. Would affect data path throughput.
int8 rtl8651_drvDropUnknownPppoePADT		= FALSE;		// TRUE if not forward unknown PPPoE Discovery from WAN side
struct ether_header rtl8651_knownPppoePassthruEthHdr;		// Save ether header of unknown PPPoE passthru session


//should be static
void (*rtl8651_linkStatusNotifier)(uint32 port, int8 linkUp)=NULL;
void (*rtl8651_L2statusNotifier)(void *l2entry_t, void *arg)=NULL;
int32 _rtl8651_delVlanRefFilterDatabaseEntry(uint16 l2Type, uint16 fid, ether_addr_t * macAddr);

//was rtl8651_installEtherrnetLinkStatusNotifier
/*
@func int32 | rtl8651_installPortStatusChangeNotifier | Set a call back function to notify Layer 3 driver link status change events.
@parm void (*)(uint32, int8) | notifier | A call back function which type is "void (*)(uint32, int8)".
@rvalue SUCCESS | A call back function was successfully registered.
@comm
If L3/4 layer driver/applications need to be notified when link changes, an application call back function should be registered to Table Driver.
*/
int32 rtl8651_installPortStatusChangeNotifier(void (*notifier)(uint32 port, int8 linkUp)){
	rtl8651_linkStatusNotifier = notifier;
	return SUCCESS;
}

/*
@func int32 | rtl8651_installFDBEntryChangeNotifier | Set a call back function to notify Layer 3 driver software L2 table entry modified/changed events.
@parm void (*)(uint32, int8) | notifier | A call back function which type is "void (*)(uint32, int8)".
@rvalue SUCCESS | A call back function was successfully registered.
@comm
If L3/4 layer driver/applications need to be notified when link changes, an application call back function should be registered to Table Driver.
*/

int32 rtl8651_installFDBEntryChangeNotifier( void (*notifier)(void *l2entry_t, void *arg)){
	rtl8651_L2statusNotifier=notifier;
	return SUCCESS;
}

/*
@func void | rtl8651_setEthernetPortLinkStatus | Set ethernet port link status (extension ports included).
@parm uint32 | port | port number.
@parm int8 | linkUp | link up (TRUE) or link down (FALSE).
@comm
This API should be used to extension ports. In normal case, don't apply this API to physical port.
*/
 void rtl8651_setEthernetPortLinkStatus(uint32 port, int8 linkUp) {
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	rtl8651_tblDrv_ethernet_t *phyPortPtr;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	int8 ethernetNotify;
	uint32 i;
#ifdef RTL865XB_DCUT_SWVLAN
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;
#endif

	if(port >= RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum)
		return;
	rtlglue_drvMutexLock();//Lock resource
	phyPortPtr = &DrvTbl.phyPort[port];
	if(rtl8651_linkStatusNotifier && phyPortPtr->linkUp != (linkUp==TRUE? 1: 0))
		ethernetNotify = TRUE;
	else
		ethernetNotify = FALSE;
	//change port status
	if (linkUp == TRUE) {		
		phyPortPtr->linkUp = 1;
		phyPortLinkStatus |= (1 << port);
	} else {
		phyPortPtr->linkUp = 0;
		phyPortLinkStatus &= ~(1 << port);
	}

	//Update aggregator status
	for(i=0,aggrePtr=&DrvTbl.aggregator[0]; i < RTL8651_PORT_NUMBER + rtl8651_totalExtPortNum ; i++,aggrePtr++)//Check aggregator
		if(aggrePtr->ethernetPortMask & (1<<port)) {
			if(linkUp==TRUE)
				aggrePtr->ethernetPortUpStatus |= (1<<port);
			else
				aggrePtr->ethernetPortUpStatus &= ~(1<<port);
		}
	//update VLAN member port status
	for(i=0,vlanPtr=&DrvTbl.vlan[0]; i<RTL8651_VLAN_NUMBER; i++,vlanPtr++) {
		if (vlanPtr->valid == 0)
			continue;
		if(vlanPtr->memberPortMask & (1<<port)) {
			if(linkUp==TRUE)
				vlanPtr->memberPortUpStatus |= (1<<port);
			else
				vlanPtr->memberPortUpStatus &= ~(1<<port);
			RTL8651_SETASICVLAN(vlanPtr);
		}
	}

#ifdef RTL865XB_DCUT_SWVLAN
	// update software VLAN member port status
	for (i = 0, swVlanPtr = &DrvTbl.swVlan[0]; i < RTL8651_SWVLAN_NUMBER; i++, swVlanPtr++) {
		if (swVlanPtr->valid == 0)
			continue;
		if (swVlanPtr->memberPortMask & (1 << port)) {
			if (linkUp == TRUE)
				swVlanPtr->memberPortUpStatus |= (1 << port);
			else
				swVlanPtr->memberPortUpStatus &= ~(1 << port);
		}
	}
#endif
	
	memset(&cachedDA,0,sizeof(ether_addr_t)); //clear port cache in rtl8651_asicL2DAlookup

#if 1 && defined(CONFIG_RTL865X_MULTILAYER_BSP) //CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
	if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
	{
		// update Protocol-based VLAN of PPPoE Passthru 

		/*  For CDRouter.
		 *  Add PPPoE Ctrl packet to Protocol-based VLAN
		 *   when we disabled "drop unknown PPPoE packet" function. */
		if ( 0 && rtl8651_drvDropUnknownPppoePADT )
			rtl8651_setProtocolBasedVLAN( RTL8651_PBV_RULE_PPPOE_CONTROL, port, FALSE, rtl8651_vlanTableIndex(rtl8651_gPPPoEPassthruVlanID) );
		else
			rtl8651_setProtocolBasedVLAN( RTL8651_PBV_RULE_PPPOE_CONTROL, port, linkUp, rtl8651_vlanTableIndex(rtl8651_gPPPoEPassthruVlanID) );

		/* PPPoE Data packet is always added to Protocol-based VLAN. */
		rtl8651_setProtocolBasedVLAN( RTL8651_PBV_RULE_PPPOE_SESSION, port, linkUp, rtl8651_vlanTableIndex(rtl8651_gPPPoEPassthruVlanID) );
	}
#else //CONFIG_RTL865XB
#endif//CONFIG_RTL865XB

#if 1 && defined(CONFIG_RTL865X_MULTILAYER_BSP) //CONFIG_RTL865XB_EXP_IPV6_PASSTHRU
	if ( rtl8651_l2protoPassThrough & PASSTHRU_IPV6 )
	{
		// update Protocol-based VLAN of IPv6 Passthru 
		if ( rtl8651_pbv_ipv6passthru != RTL8651_IPV6PASSTHRU_PBVID_IS_NOT_SET )
			rtl8651_setProtocolBasedVLAN( rtl8651_pbv_ipv6passthru, port, linkUp, rtl8651_vlanTableIndex(rtl8651_gIPv6PassthruVlanID) );
	}
#else //CONFIG_RTL865XB
#endif//CONFIG_RTL865XB

#if 1 && defined(CONFIG_RTL865X_MULTILAYER_BSP) //CONFIG_RTL865XB_EXP_IPX_PASSTHRU
	if ( rtl8651_l2protoPassThrough & PASSTHRU_IPX )
	{
		// update Protocol-based VLAN of IPX Passthru 
		if ( rtl8651_pbv_ipxpassthru != RTL8651_IPXPASSTHRU_PBVID_IS_NOT_SET )
			rtl8651_setProtocolBasedVLAN( rtl8651_pbv_ipxpassthru, port, linkUp, rtl8651_vlanTableIndex(rtl8651_gIPXPassthruVlanID) );
	}
#else //CONFIG_RTL865XB
#endif//CONFIG_RTL865XB

#if 1 && defined(CONFIG_RTL865X_MULTILAYER_BSP) //CONFIG_RTL865XB_EXP_NETBIOS_PASSTHRU
	if ( rtl8651_l2protoPassThrough & PASSTHRU_NETBIOS )
	{
		// update Protocol-based VLAN of Netbios Passthru 
		if ( rtl8651_pbv_netbiospassthru != RTL8651_NETBIOSPASSTHRU_PBVID_IS_NOT_SET )
			rtl8651_setProtocolBasedVLAN( rtl8651_pbv_netbiospassthru, port, linkUp, rtl8651_vlanTableIndex(rtl8651_gNetbiosPassthruVlanID) );
	}
#else //CONFIG_RTL865XB
#endif//CONFIG_RTL865XB

	/* For RTL865xB cut-d & after that */
	if (RtkHomeGatewayChipRevisionID >= 0x02)
	{
		int8 autoNego, fullDuplex;
		uint32 speed, advCap;
		rtl8651_getAsicEthernetPHY(port, &autoNego, &advCap, &speed, &fullDuplex);
		phyPortPtr->speed = (speed==1)? 1: 0;
		phyPortPtr->duplex = (fullDuplex==TRUE)? 1: 0;
		phyPortPtr->autoNegotiation = (autoNego==1)? 1: 0;
		phyPortPtr->autoAdvCapability = advCap;
		
		//printk("%s(%d): speed: %s, %s, %s\n",  __FUNCTION__, __LINE__, (phyPortPtr->speed==1? "100Ml": "10M"), (phyPortPtr->duplex==1? "full": "half"), (phyPortPtr->autoNegotiation==1? "auto": "non-auto"));
	}
	/* For RTL865xB cut-b & before that */
	else
	{
		if (phyPortPtr->autoAdvCapability < 0x4) { /* Only for N-Way */
			int8 autoNego, fullDuplex;
			uint32 speed, advCap;
			rtl8651_getAsicEthernetPHY(port, &autoNego, &advCap, &speed, &fullDuplex);
			phyPortPtr->speed = (speed==1)? 1: 0;
			phyPortPtr->duplex = (fullDuplex==TRUE)? 1: 0;
			phyPortPtr->autoNegotiation = (autoNego==1)? 1: 0;
			//phyPortPtr->autoAdvCapability = advCap;
		}
	}

#ifdef CONFIG_RTL865X_MULTILAYER_BSP
	/* modify multicast table */
	if (linkUp == FALSE) {
		_rtl8651_delMulticastByPort(port);
	}
#endif
	rtlglue_drvMutexUnlock();//Unlock resource
	//notify upper layer for status change.	
	if(ethernetNotify == TRUE)
		rtl8651_linkStatusNotifier(port, linkUp);

}

static void _rtl8651_updateAsicPHY(uint32 port) {
	rtl8651_setAsicEthernetPHY(port, DrvTbl.phyPort[port].autoNegotiation==1? TRUE: FALSE, DrvTbl.phyPort[port].autoAdvCapability, DrvTbl.phyPort[port].speed, DrvTbl.phyPort[port].duplex==1? TRUE: FALSE);
}


/*
@func int32 | rtl8651_getEthernetPortLinkStatus | Get ethernet port status.
@parm uint32 | port | physical port number.
@parm int8 * | linkUp | After calling the API, <p linkUp> will be set to TRUE for link up or FALSE for link down. 
@parm uint16 * | speed | The <p speed> records link data rate, 10Mbps or 100Mbps.
@parm int8 * | fullduplex | TRUE for duplex mode and FALSE for half mode.
@parm int8 * | autoNeg | Wheteher the port has auto-negotiage capability.
@rvalue SUCCESS | The API call was success.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@comm
Specify a physical port number, the port current status will be returned.
*/
int32 rtl8651_getEthernetPortLinkStatus(uint32 port, int8 * linkUp, uint16 *speed, int8 *fullduplex, int8 *autoNeg)
{
	rtl8651_tblDrv_ethernet_t *phyPortPtr;

	UNLOCK_ERROR_CHECK((port>=RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
	rtlglue_drvMutexLock();//Lock resource
	phyPortPtr = &DrvTbl.phyPort[port];

	if(linkUp)
		*linkUp = phyPortPtr->linkUp==1? TRUE: FALSE;
	if(speed){
		int32 _speed[]={ 10, 100, 1000 };
		*speed = _speed[phyPortPtr->speed];
	}
	if(fullduplex)
		*fullduplex = phyPortPtr->duplex==1? TRUE: FALSE;

	/* For RTL865xB cut-d & after that. */
	if (RtkHomeGatewayChipRevisionID >= 0x02)
	{
		if(autoNeg)
			*autoNeg = phyPortPtr->autoNegotiation==1? TRUE: FALSE;
	}
	/* For RTL865xB cut-b & before that. */
	else	
	{
		if (autoNeg) {
			if (phyPortPtr->fake_forceMode == 1)		// If fake_forceMode bit is turned on, it means that user turns off N-way.
				*autoNeg = FALSE;
			/*  ====================================================
				User's setting is either "turning on N-way" or "turning off N-way".
				Don't report phyPortPtr->autoNegotiation because it is sync. to PHY reg config.
			      ==================================================== */
			else	
				*autoNeg = TRUE;
		}
	}

	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

int32 _rtl8651_setEthernetPortAutoNegotiation(uint32 port, int8 autoNegotiation, uint32 advCapability) {
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	rtl8651_tblDrv_ethernet_t *phyPortPtr;
	int32 i;

	/* For RTL865xB cut-d & after that. */
	if (RtkHomeGatewayChipRevisionID >= 0x02)
	{
		/* If the port is a member of an aggregator, forbid setting autoNegotation */
		for(i=0,aggrePtr=&DrvTbl.aggregator[0]; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++, aggrePtr++) {
			LOCK_ERROR_CHECK(
				( (aggrePtr->aggregated==1)&&(aggrePtr->ethernetPortMask&(1<<port)) ),
				TBLDRV_EPORTAUTONEGOCANNOTAGGREGATE
			);		
		}
		phyPortPtr = &DrvTbl.phyPort[port];
		phyPortPtr->autoNegotiation = autoNegotiation==TRUE? 1: 0;
		phyPortPtr->autoAdvCapability = advCapability;
		_rtl8651_updateAsicPHY(port);
	}
	/* For RTL865xB cut-b & before that. */
	else
	{
		/* If the port is a member of an aggregator, forbid setting autoNegotation */
		for(i=0,aggrePtr=&DrvTbl.aggregator[0]; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++, aggrePtr++) {
			LOCK_ERROR_CHECK(
				( (aggrePtr->aggregated==1)&&(aggrePtr->ethernetPortMask&(1<<port)) ),
				TBLDRV_EPORTAUTONEGOCANNOTAGGREGATE
			);		
		}	

		phyPortPtr = &DrvTbl.phyPort[port];
		/* ===============================================================
		    In order to patch an IC bug (switch core crash due to turning on in_queue flow control of a port), 
	           we cannot let user turn off autoNegotiation. 
	           
	           Therefore, if user wants to turn off autoNegotiation, this API will set phyPortPtr->fake_forceMode 1.
	           If user uses rtl8651_getEthernetPortLinkStatus(), ROMEDRV will tell a lie to the user that PHY is 
	           already configured in force mode.
		    Actually, PHY is always configured in N-way mode.
	           =============================================================== */
		if (autoNegotiation == FALSE)
			phyPortPtr->fake_forceMode = 1;	
		else
			phyPortPtr->fake_forceMode = 0;

		phyPortPtr->autoNegotiation = 1;			// No matter what user did, we always turn on autoNegotiation.
		phyPortPtr->autoAdvCapability = advCapability;
		if (advCapability == RTL8651_ETHER_FORCE_100FULL) {
			phyPortPtr->speed = 1;
			phyPortPtr->duplex = 1;
		}
		else if (advCapability == RTL8651_ETHER_FORCE_100HALF) {
			phyPortPtr->speed = 1;
			phyPortPtr->duplex = 0;
		}
		else if (advCapability == RTL8651_ETHER_FORCE_10FULL) {
			phyPortPtr->speed = 0;
			phyPortPtr->duplex = 1;
		}
		else if (advCapability == RTL8651_ETHER_FORCE_10HALF) {
			phyPortPtr->speed = 0;
			phyPortPtr->duplex = 0;
		}
		_rtl8651_updateAsicPHY(port);
	}

	return SUCCESS;
}


/*
@func int32 | rtl8651_setEthernetPortDuplexMode | Set a physical port to full/half duplex.
@parm uint32 | port | Physical port number.
@parm int8 | fullDuplex | TRUE for full duplex. FALSE for half duplex.
@rvalue SUCCESS | Port operation mode was set successfully.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@rvalue TBLDRV_EPORTOPMODECANNOTAGGREGATE | Can not set a port's operation mode if the port is aggregated.
@comm
Set a port to be full/half duplex. If specify an aggregated port, the API will return fail.
*/
int32 rtl8651_setEthernetPortDuplexMode(uint32 port, int8 fullDuplex) {
	/* For RTL865xB cut-d & after that. */
	if (RtkHomeGatewayChipRevisionID >= 0x02)
	{
		rtl8651_tblDrv_linkAggregation_t *aggrePtr;
		uint32 i;

		UNLOCK_ERROR_CHECK((port>=RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
		rtlglue_drvMutexLock();//Lock resource
		/*cfliu: If this port has been aggregated, forbid change to half duplex mode*/
		if(!fullDuplex){
			//Check whether this port is a member of aggregator 
			for(i=0,aggrePtr=&DrvTbl.aggregator[0]; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++,aggrePtr++) {
				LOCK_ERROR_CHECK(
					( (aggrePtr->aggregated == 1)&&(aggrePtr->ethernetPortMask&(1<<port)) ),
					TBLDRV_EPORTOPMODECANNOTAGGREGATE
				);
			}
		}
		DrvTbl.phyPort[port].duplex = fullDuplex==TRUE? 1: 0;
		_rtl8651_updateAsicPHY(port);
		rtlglue_drvMutexUnlock();//Unlock resource
		return SUCCESS;
	}
	/* For RTL865xB cut-b & before that. */
	else	
	{
		int32 retval;
		uint32 advCapability;

		UNLOCK_ERROR_CHECK((port >= RTL8651_PORT_NUMBER + rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);

		rtlglue_drvMutexLock();		// Lock resource


		
		if (DrvTbl.phyPort[port].speed == 1 && fullDuplex)
			advCapability = RTL8651_ETHER_FORCE_100FULL;
		else if (DrvTbl.phyPort[port].speed == 1 && !fullDuplex)
			advCapability = RTL8651_ETHER_FORCE_100HALF;
		else if (DrvTbl.phyPort[port].speed == 0 && fullDuplex)
			advCapability = RTL8651_ETHER_FORCE_10FULL;
		else
			advCapability = RTL8651_ETHER_FORCE_10HALF;

		retval = _rtl8651_setEthernetPortAutoNegotiation(port, FALSE, advCapability);

		rtlglue_drvMutexUnlock();	// Unlock resource
		return retval;
	}
}


/*
@func int32 | rtl8651_setEthernetPortSpeed | Set a physical port's speed.
@parm uint32 | port | Physical port number.
@parm uint32 | speed | Port speed, 10: 10Mbps, 100: 100Mbps.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@rvalue TBLDRV_EUNKNOWPORTSPEED | Unknow port speed. The port speed should be either 10 or 100.
@rvalue TBLDRV_EPORTSPEEDCANNOTAGGREGATE | The specified port's speed can not be changed if the port is aggreaged.
@comm
Set a physical port's speed to be 10Mbps or 100Mbps. If the port belongs to a trunking port, the API will return fail.
*/
int32 rtl8651_setEthernetPortSpeed(uint32 port, uint32 speed) {
	/* For RTL865xB cut-d & after that. */
	if (RtkHomeGatewayChipRevisionID >= 0x02)
	{
		rtl8651_tblDrv_linkAggregation_t *aggrePtr;
		rtl8651_tblDrv_ethernet_t *phyPortPtr;
		uint32 newSpeed=0, i;

		UNLOCK_ERROR_CHECK((port>=RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
		UNLOCK_ERROR_CHECK(((speed!=10)&&(speed!=100)), TBLDRV_EUNKNOWPORTSPEED);
		rtlglue_drvMutexLock();//Lock resource
		switch(speed) {
			case 10:	newSpeed = 0; break;
			case 100:	newSpeed = 1; break;
			default:	assert(0);//Speed mismatch should return failed
		}
		phyPortPtr = &DrvTbl.phyPort[port];
		/*cfliu: if this port is a member of an aggregator, forbid link speed to be changed*/
		if(phyPortPtr->speed != newSpeed) {
			for(i=0,aggrePtr=&DrvTbl.aggregator[0]; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++,aggrePtr++) {
				LOCK_ERROR_CHECK(
					( (aggrePtr->aggregated == 1)&&(aggrePtr->ethernetPortMask&(1<<port)) ),
					TBLDRV_EPORTSPEEDCANNOTAGGREGATE;
				);
			}
		}
		phyPortPtr->speed = newSpeed;
		_rtl8651_updateAsicPHY(port);
		rtlglue_drvMutexUnlock();//Unlock resource
		return SUCCESS;
	}
	/* For RTL865xB cut-b & before that. */
	else	
	{
		int32 retval;
		uint32 advCapability;

		UNLOCK_ERROR_CHECK((port >= RTL8651_PORT_NUMBER + rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
		UNLOCK_ERROR_CHECK(((speed != 10) && (speed!=100)), TBLDRV_EUNKNOWPORTSPEED);
		rtlglue_drvMutexLock();		//Lock resource

		if (speed == 100 && DrvTbl.phyPort[port].duplex)
			advCapability = RTL8651_ETHER_FORCE_100FULL;
		else if (speed == 100 && !DrvTbl.phyPort[port].duplex)
			advCapability = RTL8651_ETHER_FORCE_100HALF;
		else if (speed == 10 && DrvTbl.phyPort[port].duplex)
			advCapability = RTL8651_ETHER_FORCE_10FULL;
		else
			advCapability = RTL8651_ETHER_FORCE_10HALF;

		retval = _rtl8651_setEthernetPortAutoNegotiation(port, FALSE, advCapability);
		rtlglue_drvMutexUnlock();	//Unlock resource
		return retval;
	}
}

/*
@func int32 | rtl8651_setEthernetPortAutoNegotiation | Configure physical port's auto-negotiation and advertise capability.
@parm uint32 | port | Physical port number.
@parm int8 | autoNegotiation | TRUE for auto-negotiation and FALSE for non auto-negotiation.
@parm uint32 | advCapability | TRUE for advertising and FALSE for non advertising.
@rvalue SUCCESS | API success.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@rvalue TBLDRV_EPORTAUTONEGOCANNOTAGGREGATE | Can not change a port's auto-negotiation capability if the port is aggregated.
@comm
Configure a port's auto-negotiation and advertising capability. If the port belongs to a trunking port, the API will return fail.
*/
int32 rtl8651_setEthernetPortAutoNegotiation(uint32 port, int8 autoNegotiation, uint32 advCapability) {
	int32 retval;

	/* For RTL865xB cut-d & after that. */
	if (RtkHomeGatewayChipRevisionID >= 0x02)
	{
		UNLOCK_ERROR_CHECK((port>=RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
	}
	/* For RTL865xB cut-d & before that. */
	else	
	{
	    	/* For RTL8650B bug */
		UNLOCK_ERROR_CHECK((port>=RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
		if ((autoNegotiation == TRUE && advCapability > 0x03) || (autoNegotiation == FALSE && advCapability < 0x04))
			return TBLDRV_EINVALIDINPUT;
	}
	rtlglue_drvMutexLock();//Lock resource

	retval = _rtl8651_setEthernetPortAutoNegotiation(port, autoNegotiation, advCapability);

	rtlglue_drvMutexUnlock();//Unlock resource
	return retval;
}

/*
@func int32 | rtl8651_setEthernetPortBandwidthControl | Limit a physical port's bandwidth.
@parm uint32 | port | Physical port number.
@parm int8 | input | TRUE for limit receving rate and FALSE for sending reate.
@parm uint32 | rate | Limit rate. RTL8651_BC_FULL: Full rate, RTL8651_BC_128K: 128K, RTL8651_BC_256K:256K, RTL8651_BC_512K: 512K, RTL8651_BC_1M: 1MB, RTL8651_BC_2M: 2MB, RTL8651_BC_4M: 4MB, RTL8651_BC_8M: 8MB.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@rvalue TBLDRV_EINVALIDDATARATE | Invalid data rate.
@comm
	Configure the limit a port's receiving rate or sending rate. 
Process sequence:
1. Check the input port number. The port number should be physical port.
2. Check the condifured rate and new configured rate. The rate setting must not have different X4 configuration.
3. Update driver configuration
4. Configure all port bandwidth control setting and rate value X4 setting
*/

int32 rtl8651_setEthernetPortBandwidthControl(uint32 port, int8 input, uint32 rate) {
	rtl8651_tblDrv_ethernet_t *phyPortPtr;
	uint32 rangeIdx, i, j, cfgRate;
	uint32 asicRate=0;	

	UNLOCK_ERROR_CHECK((port >= RTL8651_PORT_NUMBER), TBLDRV_EINVALIDPORT);

	/* 
		Before RTL865xB c-cut, "128K, 256K" are provided by x1 scale.
						       "512K, 1M, 2M, 4M, 8M, 16M, 32M" are provided by x4 scale.
		After RTL865xB c-cut, "128K, 256K, 512K" are provided by x1 scale.
						     "1M, 2M, 4M, 8M, 16M, 32M, 64M" are provided by x8 scale.

		The following code seems "fat" for distinguishing the chip version.
		However, most part needs to modify a bit.
		In other words, we need to add additional chip version judgement code in many portion of this API 
		to seperate the code of different chip version.
		
		Fix?
	*/	
	if (RtkHomeGatewayChipRevisionID >= 0x02) {
		UNLOCK_ERROR_CHECK((rate > RTL8651_BC_64M), TBLDRV_EINVALIDDATARATE);
		
		rtlglue_drvMutexLock();
		{/* Analysis current configured port rate setting and make sure the X4 range will not conflict with current status*/
			rangeIdx = 0;
			for(i=0; i<RTL8651_PORT_NUMBER; i++) {
				phyPortPtr = &DrvTbl.phyPort[i];
				if(i != port) {
					if(phyPortPtr->inputBandwidthControl == RTL8651_BC_128K || phyPortPtr->inputBandwidthControl == RTL8651_BC_256K ||
						phyPortPtr->inputBandwidthControl == RTL8651_BC_512K || phyPortPtr->outputBandwidthControl == RTL8651_BC_512K ||
						phyPortPtr->outputBandwidthControl == RTL8651_BC_128K || phyPortPtr->outputBandwidthControl == RTL8651_BC_256K) {
						assert(rangeIdx==0 || rangeIdx == 1);
						rangeIdx = 1;
					} else if (phyPortPtr->inputBandwidthControl == RTL8651_BC_16M || phyPortPtr->inputBandwidthControl == RTL8651_BC_32M ||
						phyPortPtr->inputBandwidthControl == RTL8651_BC_64M || phyPortPtr->outputBandwidthControl == RTL8651_BC_64M ||
						phyPortPtr->outputBandwidthControl == RTL8651_BC_16M || phyPortPtr->outputBandwidthControl == RTL8651_BC_32M ) {
						assert(rangeIdx==0 || rangeIdx == 2);
						rangeIdx = 2;
					}
				} else if (input == TRUE) {
					if(phyPortPtr->outputBandwidthControl == RTL8651_BC_128K || phyPortPtr->outputBandwidthControl == RTL8651_BC_256K ||
						phyPortPtr->outputBandwidthControl == RTL8651_BC_512K) {
						assert(rangeIdx==0 || rangeIdx == 1);
						rangeIdx = 1;
					} else if (phyPortPtr->outputBandwidthControl == RTL8651_BC_16M || phyPortPtr->outputBandwidthControl == RTL8651_BC_32M ||
						phyPortPtr->outputBandwidthControl == RTL8651_BC_64M) {
						assert(rangeIdx==0 || rangeIdx == 2);
						rangeIdx = 2;
					}
				
				} else {
					if(phyPortPtr->inputBandwidthControl == RTL8651_BC_128K || phyPortPtr->inputBandwidthControl == RTL8651_BC_256K ||
						phyPortPtr->inputBandwidthControl == RTL8651_BC_512K) {
						assert(rangeIdx==0 || rangeIdx == 1);
						rangeIdx = 1;
					} else if (phyPortPtr->inputBandwidthControl == RTL8651_BC_16M || phyPortPtr->inputBandwidthControl == RTL8651_BC_32M ||
						phyPortPtr->inputBandwidthControl == RTL8651_BC_64M) {
						assert(rangeIdx==0 || rangeIdx == 2);
						rangeIdx = 2;
					}
				}
			}
			if (((rate == RTL8651_BC_128K || rate == RTL8651_BC_256K || rate == RTL8651_BC_512K) && rangeIdx == 2) ||
				((rate == RTL8651_BC_16M || rate == RTL8651_BC_32M || rate == RTL8651_BC_64M) && rangeIdx == 1)) {
				rtlglue_drvMutexUnlock();
				return(TBLDRV_EINVALIDDATARATE);
			}
		}
		phyPortPtr = &DrvTbl.phyPort[port];
		if(input == TRUE)
			phyPortPtr->inputBandwidthControl = rate;
		else phyPortPtr->outputBandwidthControl = rate;
		
		if(rangeIdx == 2 || rate == RTL8651_BC_16M || rate == RTL8651_BC_32M || rate == RTL8651_BC_64M) {
			rtl8651_setAsicEthernetBandwidthControlX8(TRUE);
			for(i=0; i<RTL8651_PORT_NUMBER; i++) 
				for(j=0; j<2; j++) {
					phyPortPtr = &DrvTbl.phyPort[i];
					if(j)	cfgRate = phyPortPtr->inputBandwidthControl;
					else	cfgRate = phyPortPtr->outputBandwidthControl;
					switch(cfgRate){
						case RTL8651_BC_FULL:	asicRate = BW_FULL_RATE;
							break;
						case RTL8651_BC_1M:		asicRate = BW_128K;
							break;
						case RTL8651_BC_2M:		asicRate = BW_256K;
							break;
						case RTL8651_BC_4M:		asicRate = BW_512K;
							break;
						case RTL8651_BC_8M:		asicRate = BW_1M;
							break;
						case RTL8651_BC_16M:	asicRate = BW_2M;
							break;
						case RTL8651_BC_32M:	asicRate = BW_4M;
							break;
						case RTL8651_BC_64M:	asicRate = BW_8M;
							break;
						case RTL8651_BC_128K:
						case RTL8651_BC_256K:
						case RTL8651_BC_512K:
						default:
							assert(0);
					}
					rtl8651_setAsicEthernetBandwidthControl(i, j, asicRate);
				}
		}
		else {
			rtl8651_setAsicEthernetBandwidthControlX8(FALSE);
			for(i=0; i<RTL8651_PORT_NUMBER; i++) 
				for(j=0; j<2; j++) {
					phyPortPtr = &DrvTbl.phyPort[i];
					if(j)	cfgRate = phyPortPtr->inputBandwidthControl;
					else	cfgRate = phyPortPtr->outputBandwidthControl;
					switch(cfgRate){
						case RTL8651_BC_FULL:	asicRate = BW_FULL_RATE;
							break;
						case RTL8651_BC_128K:	asicRate = BW_128K;
							break;
						case RTL8651_BC_256K:	asicRate = BW_256K;
							break;
						case RTL8651_BC_512K:	asicRate = BW_512K;
							break;
						case RTL8651_BC_1M:		asicRate = BW_1M;
							break;
						case RTL8651_BC_2M:		asicRate = BW_2M;
							break;
						case RTL8651_BC_4M:		asicRate = BW_4M;
							break;
						case RTL8651_BC_8M:		asicRate = BW_8M;
							break;
						case RTL8651_BC_16M:
						case RTL8651_BC_32M:
						case RTL8651_BC_64M:
						default:
							assert(0);
					}
					rtl8651_setAsicEthernetBandwidthControl(i, j, asicRate);
				}
		}	
	}
	/* For RTL865xB b-cut & before chip version */
	else {
		UNLOCK_ERROR_CHECK((rate > RTL8651_BC_32M), TBLDRV_EINVALIDDATARATE);
		rtlglue_drvMutexLock();
		{/* Analysis current configured port rate setting and make sure the X4 range will not conflict with current status*/
			rangeIdx = 0;
			for(i=0; i<RTL8651_PORT_NUMBER; i++) {
				phyPortPtr = &DrvTbl.phyPort[i];
				if(i != port) {
					if(phyPortPtr->inputBandwidthControl == RTL8651_BC_128K || phyPortPtr->inputBandwidthControl == RTL8651_BC_256K ||
						phyPortPtr->outputBandwidthControl == RTL8651_BC_128K || phyPortPtr->outputBandwidthControl == RTL8651_BC_256K) {
						assert(rangeIdx==0 || rangeIdx == 1);
						rangeIdx = 1;
					} else if (phyPortPtr->inputBandwidthControl == RTL8651_BC_16M || phyPortPtr->inputBandwidthControl == RTL8651_BC_32M ||
						phyPortPtr->outputBandwidthControl == RTL8651_BC_16M || phyPortPtr->outputBandwidthControl == RTL8651_BC_32M ) {
						assert(rangeIdx==0 || rangeIdx == 2);
						rangeIdx = 2;
					}
				} else if (input == TRUE) {
					if(phyPortPtr->outputBandwidthControl == RTL8651_BC_128K || phyPortPtr->outputBandwidthControl == RTL8651_BC_256K) {
						assert(rangeIdx==0 || rangeIdx == 1);
						rangeIdx = 1;
					} else if (phyPortPtr->outputBandwidthControl == RTL8651_BC_16M || phyPortPtr->outputBandwidthControl == RTL8651_BC_32M ) {
						assert(rangeIdx==0 || rangeIdx == 2);
						rangeIdx = 2;
					}
				
				} else {
					if(phyPortPtr->inputBandwidthControl == RTL8651_BC_128K || phyPortPtr->inputBandwidthControl == RTL8651_BC_256K) {
						assert(rangeIdx==0 || rangeIdx == 1);
						rangeIdx = 1;
					} else if (phyPortPtr->inputBandwidthControl == RTL8651_BC_16M || phyPortPtr->inputBandwidthControl == RTL8651_BC_32M) {
						assert(rangeIdx==0 || rangeIdx == 2);
						rangeIdx = 2;
					}
				}
			}
			if (((rate == RTL8651_BC_128K || rate == RTL8651_BC_256K) && rangeIdx == 2) ||
				((rate == RTL8651_BC_16M || rate == RTL8651_BC_32M) && rangeIdx == 1)) {
				rtlglue_drvMutexUnlock();
				return(TBLDRV_EINVALIDDATARATE);
			}
		}
		phyPortPtr = &DrvTbl.phyPort[port];
		if(input == TRUE)
			phyPortPtr->inputBandwidthControl = rate;
		else phyPortPtr->outputBandwidthControl = rate;
		
		if(rangeIdx == 2 || rate == RTL8651_BC_16M || rate == RTL8651_BC_32M) {
			rtl8651_setAsicEthernetBandwidthControlX4(TRUE);
			for(i=0; i<RTL8651_PORT_NUMBER; i++) 
				for(j=0; j<2; j++) {
					phyPortPtr = &DrvTbl.phyPort[i];
					if(j)	cfgRate = phyPortPtr->inputBandwidthControl;
					else	cfgRate = phyPortPtr->outputBandwidthControl;
					switch(cfgRate){
						case RTL8651_BC_FULL:	asicRate = BW_FULL_RATE;
							break;
						case RTL8651_BC_512K:	asicRate = BW_128K;
							break;
						case RTL8651_BC_1M:		asicRate = BW_256K;
							break;
						case RTL8651_BC_2M:		asicRate = BW_512K;
							break;
						case RTL8651_BC_4M:		asicRate = BW_1M;
							break;
						case RTL8651_BC_8M:		asicRate = BW_2M;
							break;
						case RTL8651_BC_16M:	asicRate = BW_4M;
							break;
						case RTL8651_BC_32M:	asicRate = BW_8M;
							break;
						case RTL8651_BC_128K:
						case RTL8651_BC_256K:
						default:
							assert(0);
					}
					rtl8651_setAsicEthernetBandwidthControl(i, j, asicRate);
				}
		}
		else {
			rtl8651_setAsicEthernetBandwidthControlX4(FALSE);
			for(i=0; i<RTL8651_PORT_NUMBER; i++) 
				for(j=0; j<2; j++) {
					phyPortPtr = &DrvTbl.phyPort[i];
					if(j)	cfgRate = phyPortPtr->inputBandwidthControl;
					else	cfgRate = phyPortPtr->outputBandwidthControl;
					switch(cfgRate){
						case RTL8651_BC_FULL:	asicRate = BW_FULL_RATE;
							break;
						case RTL8651_BC_128K:	asicRate = BW_128K;
							break;
						case RTL8651_BC_256K:	asicRate = BW_256K;
							break;
						case RTL8651_BC_512K:	asicRate = BW_512K;
							break;
						case RTL8651_BC_1M:		asicRate = BW_1M;
							break;
						case RTL8651_BC_2M:		asicRate = BW_2M;
							break;
						case RTL8651_BC_4M:		asicRate = BW_4M;
							break;
						case RTL8651_BC_8M:		asicRate = BW_8M;
							break;
						case RTL8651_BC_16M:	
						case RTL8651_BC_32M:	
						default:
							assert(0);
					}
					rtl8651_setAsicEthernetBandwidthControl(i, j, asicRate);
				}
		}	
	}
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

static uint32 mPortMask = 0;
static uint32 mRxMirrorMask = 0;
static uint32 mTxMirrorMask = 0;

static int32 _rtl8651_delMirrorPort(uint32 port) {
	rtl8651_tblDrv_linkAggregation_t *aggregate_t;
	rtl8651_tblDrv_vlanTable_t *vlan_t;
	uint32 portMask, i;

	portMask = (1<<port);
	UNLOCK_ERROR_CHECK(!(mPortMask&portMask), TBLDRV_EMIRRORPORTNOTFOUND);
	aggregate_t = &DrvTbl.aggregator[port];
	assert(aggregate_t->aggregated == 0);
	assert(aggregate_t->ethernetPortMask > 0);
	aggregate_t->pvid = 1; /* default PVID */
	/* Find default VLAN (vlan1), it should exist */
	for(i=0,vlan_t=&DrvTbl.vlan[0]; i<RTL8651_VLAN_NUMBER; i++,vlan_t++) 
		if (vlan_t->valid == 1 && vlan_t->vid == 1)
			break;
	assert(i < RTL8651_VLAN_NUMBER);
	vlan_t->memberPortMask |= portMask;
	vlan_t->memberPortUpStatus |= portMask;
	vlan_t->untagPortMask |= portMask;
	RTL8651_SETASICVLAN(vlan_t);
	mPortMask &= ~portMask;
	rtl8651_setAsicPortMirror(mTxMirrorMask, mRxMirrorMask, mPortMask);
	return SUCCESS;
}

/*
@func int32 | rtl8651_addRxTxMirror | Set a port to be Rx/Tx mirrored.
@parm int8 | isRxMirror | TRUE for Rx Mirror, FALSE for Tx Mirror.
@parm uint32 | port | Specified port number.
@rvalue SUCCESS | The specified port was successfully add to Rx/Tx mirror.
@rvalue TBLDRV_EPORTISAMIRRORPORT | Fail due to the specified port is a mirror port.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@comm
If a port is set to Rx/Tx mirrored, all traffic received/transmitted from the port will be copied to mirror port. Hence a 
mirror port can see all traffic recieved/transmitted from a Rx/Tx mirrored port.
*/
int32 rtl8651_addRxTxMirror(int8 isRxMirror, uint32 port) {
	uint32 portMask;

	UNLOCK_ERROR_CHECK(port>=RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum, TBLDRV_EINVALIDPORT);
	rtlglue_drvMutexLock();
	portMask = (1<<port);
	/* If the specified port belongs to a mirror port, forbid! */
	LOCK_ERROR_CHECK((mPortMask&portMask), TBLDRV_EPORTISAMIRRORPORT);
	if (isRxMirror == TRUE)
		mRxMirrorMask |= portMask;
	else mTxMirrorMask = portMask; /* Only one Tx Mirror is allowed */
	if (mPortMask != 0)
		rtl8651_setAsicPortMirror(mTxMirrorMask, mRxMirrorMask, mPortMask);
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32 | rtl8651_delRxTxMirror | Remove a port from Rx/Tx mirrored port.
@parm int8 | isRxMirror | TRUE for mirroring received packet. FALSE for mirroring transmitted packet.
@parm uint32 | port | Physical port number.
@rvalue SUCCESS | API success.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@rvalue TBLDRV_EPORTISAMIRRORPORT | The specified port is a mirror port. Hence it can not set to Rx/Tx mirrored port.
@comm
If a port is destined for a Rx/Tx mirrored port, the port can not be a mirror port. If a port is a Rx mirrored port, all traffic received from this port will mirror to another port, the mirror port.
Similarly, if a port is a Tx mirrored port, all traffic transmitted from this port will be mirrored to mirror port.
*/
int32 rtl8651_delRxTxMirror(int8 isRxMirror, uint32 port) {
	uint32 portMask;

	UNLOCK_ERROR_CHECK(port>=RTL8651_PORT_NUMBER, TBLDRV_EINVALIDPORT);
	rtlglue_drvMutexLock();
	portMask = (1<<port);
	/* If the specified port belongs to a mirror port, forbid! */
	LOCK_ERROR_CHECK((mPortMask&portMask), TBLDRV_EPORTISAMIRRORPORT);
	if (isRxMirror == TRUE)
		mRxMirrorMask &= ~portMask;
	else mTxMirrorMask = 0;
	if (mPortMask != 0)
		rtl8651_setAsicPortMirror(mTxMirrorMask, mRxMirrorMask, mPortMask);
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32 | rtl8651_addMirrorPort | Set a port to be a mirror port.
@parm uint32 | port | Physical port number.
@rvalue SUCCESS | API success.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@rvalue TBLDRV_EPORTISARXTXMIRRORED | The specified port is a Rx/Tx mirrored port.
@rvalue TBLDRV_EPORTCANTBEAGGREGATED | Can not set an aggregated port to a mirror port.
@comm
A mirror port will see all traffic in Rx/Tx mirrored port. Note: If a port is set to a mirror port, the port will not belong to any VLAN anymore 
and if a mirror port is removed, the port will add to default VLAN, here VLAN 1.
*/
int32 rtl8651_addMirrorPort(uint32 port) {
	rtl8651_tblDrv_linkAggregation_t *aggregate_t;
	rtl8651_tblDrv_vlanTable_t *vlan_t;
	uint32 portMask, i;

	/* NOTE:
	 * If a port is a Mirror Port, the PVID of the Mirror port is 
	 * MEANINGLESS.
	 */
	UNLOCK_ERROR_CHECK(port>=RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum, TBLDRV_EINVALIDPORT);
	rtlglue_drvMutexLock();
	portMask = (1<<port);
	/* If the specified port belongs to Mirror Rx/Tx, forbid! */
	LOCK_ERROR_CHECK(((mRxMirrorMask&portMask)||(mTxMirrorMask&portMask)), TBLDRV_EPORTISARXTXMIRRORED);
	LOCK_ERROR_CHECK((mPortMask&portMask), SUCCESS);
	/* If the specified port is an aggregated port, borbid! */
	aggregate_t = &DrvTbl.aggregator[0];
	for(i=0; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++, aggregate_t++) {
		if (aggregate_t->aggregated == 0)
			continue;
		LOCK_ERROR_CHECK((aggregate_t->ethernetPortMask&portMask), TBLDRV_EPORTCANTBEAGGREGATED);
	}
	/* If a port is a Mirror Port, the port cannot be any VLAN's member port */
	for(i=0,vlan_t=&DrvTbl.vlan[0]; i<RTL8651_VLAN_NUMBER; i++,vlan_t++) {
		if (vlan_t->valid == 0)
			continue;
		vlan_t->memberPortMask &= ~portMask;
		vlan_t->untagPortMask &= ~portMask;
		vlan_t->memberPortUpStatus &= ~portMask;
		RTL8651_SETASICVLAN(vlan_t);
	}
	mPortMask |= portMask;
	rtl8651_setAsicPortMirror(mTxMirrorMask, mRxMirrorMask, mPortMask);
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32 | rtl8651_delMirrorPort | Remove a mirror port.
@parm uint32 | port | Physical port number.
@rvalue  TBLDRV_EINVALIDPORT | Invalid port number.
@comm
Remove a mirror port from mirror port list. Note: if a mirror port is removed, the port will belong to default port, here VLAN 1.
*/
int32 rtl8651_delMirrorPort(uint32 port) {
	int32 retval;
	UNLOCK_ERROR_CHECK(port>=RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum, TBLDRV_EINVALIDPORT);
	rtlglue_drvMutexLock();
	retval = _rtl8651_delMirrorPort(port);
	rtlglue_drvMutexUnlock();
	return retval;
}

/*
@func int32 | rtl8651_resetRxTxMirror | Remove all mirror port, Rx/Tx mirrored port.
@rvalue SUCCESS | API success.
@comm
This API will remove all mirror ports from mirror port list and all Rx/Tx mirrored ports. All mirror function will be canceled.
*/
int32 rtl8651_resetRxTxMirror(void) {
	uint32 port;
	mRxMirrorMask = mTxMirrorMask = 0;
	rtl8651_setAsicPortMirror(0, 0, 0);
	/* If mirror port exists, add to default vlan */
	for(port=0; port<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; port++)
		if (mPortMask&(1<<port))
			_rtl8651_delMirrorPort(port);
	return SUCCESS;
}

/*
@func int32 | rtl8651_setVlanFwdTxMirror | Set Fworwarding Tx Mirror.
@parm uint16 | vid | VLAN ID.
@parm int8 | isFwdTxMirror | TRUE to enable Tx Mirror. FALSE to disable Tx Mirror.
@rvalue SUCCESS | The specified VLAN was successfully set as Fwd Tx Mirror.
@rvalue TBLDRV_EINVALIDVLANID | Invalid VLAN ID.
@rvalue TBLDRV_EVLANISREFERENCEDBYNETIF | VLAN is referecned by network interface. Forbid to set Tx Mirror.
@rvalue TBLDRV_EVLANISREFERENCEDBYPPPOE | VLAN is referenced by PPPoE. Forbid to set Tx Mirror.
@rvalue TBLDRV_EVLANPROMISSET | VLAN promiscuous mode is enabled. Forbid to set Tx Mirror.
@comm
The <p rtl8651_setVlanFwdTxMirror()> set the specified VLAN as Fwd Tx Mirror. It causes that all traffic transmitted from CPU will be copied to
the Fwd Tx Mirror VLAN.
*/
int32 rtl8651_setVlanFwdTxMirror(uint16 vid, int8 isFwdTxMirror) {
	rtl8651_tblDrv_networkIntfTable_t *netIfPtr;
	rtl8651_tblDrv_pppTable_t *pppoePtr;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	uint32 i;

	UNLOCK_ERROR_CHECK((vid == 0 || vid > 4094), TBLDRV_EINVALIDVLANID);
	rtlglue_drvMutexLock();//Lock resource
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	LOCK_ERROR_CHECK((vlanPtr->valid == 0 || vlanPtr->vid != vid), TBLDRV_EINVALIDVLANID);
	LOCK_ERROR_CHECK((vlanPtr->fwdTxMirror == (isFwdTxMirror==TRUE? 1: 0)), SUCCESS);
	if(isFwdTxMirror == TRUE) {//Checking network interface and pppoe
		if(rtl8651_getAsicOperationLayer()>2){
			for(i=0, netIfPtr=&DrvTbl.netIface[0]; i< rtl8651_tblDrvPara.networkIntfTableSize; i++,netIfPtr++)
				if(netIfPtr->valid == 1 && netIfPtr->vid == vid) {
					rtlglue_drvMutexUnlock();//Unlock resource
					return(TBLDRV_EVLANISREFERENCEDBYNETIF);
				}
			for(i=0, pppoePtr=&DrvTbl.pppoe[0]; i<RTL8651_PPPOE_NUMBER; i++, pppoePtr++)
				if(pppoePtr->valid == 1 && pppoePtr->vid == vid) {
					rtlglue_drvMutexUnlock();//Unlock resource
					return(TBLDRV_EVLANISREFERENCEDBYPPPOE);
				}
		}
		if(vlanPtr->promiscuous == 1)
			return (TBLDRV_EVLANPROMISSET);
	}
	vlanPtr->fwdTxMirror= isFwdTxMirror==TRUE? 1: 0;
	if(isFwdTxMirror == TRUE)
		rtl8651_tblDrvCpuMirrorPortMask |= vlanPtr->memberPortMask;
	else
		rtl8651_tblDrvCpuMirrorPortMask &= ~vlanPtr->memberPortMask;
	RTL8651_SETASICVLAN(vlanPtr);
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}




/*=======================================================================================
 * RTL8651 Spanning Tree APIs:
 *	1. External APIs
 *	  	. rtl8651_addSpanningTreeInstance()
 *	   	. rtl8651_setSpanningTreeInstanceProtocolWorking()
 *	   	. rtl8651_setSpanningTreeInstancePortState()
 *	   	. rtl8651_delSpanningTreeInstance()
 *=======================================================================================*/
#undef STP_DEBUG
#define __STPSTATE(s)	(((s)==0)?"DISABLE":((s)==1)?"BLK":((s)==2)?"LTN":((s)==3)?"LRN":((s)==4)?"FWD":"UNKNOW")

/*
@func int32 | rtl8651_addSpanningTreeInstance | Add a spanning tree.
@parm uint16 | sid | Spanning Tree ID.
@rvalue TBLDRV_EINVALIDSID | Invalid Spanning Tree ID. Only 0 is allowed currently.
@rvalue tBLDRV_ESIDALREADYEXIST | Spanning Tree ID already exists.
@comm
Currently only one Spanning Tree is allowed. If a Spanning Tree is added, the Default VLAN (VLAN 1) and a 
Filter Database 0 must exists.
*/
int32 rtl8651_addSpanningTreeInstance(uint16 sid) {
	rtl8651_tblDrv_spanningTreeTable_t *spanTreePtr, *stPtr;
	uint32 i;
	
	/*
	 * Driver Assumption:
	 * The default VLAN (VLAN 1), Filter database 0 and STI 0 (Spanning Tree Instance) 
	 * must exist after table driver initialization.
	 */

	 
	UNLOCK_ERROR_CHECK((sid >= RTL8651_FDB_NUMBER), TBLDRV_EINVALIDSID);
	rtlglue_drvMutexLock();//Lock resource
	spanTreePtr = &DrvTbl.spanningTree[sid];
	LOCK_ERROR_CHECK((spanTreePtr->valid == 1), TBLDRV_ESIDALREADYEXIST);
	for(i = 0, stPtr=&DrvTbl.spanningTree[0]; i<RTL8651_FDB_NUMBER; i++, stPtr++) 
		if(stPtr->valid == 1)
			break;
	if(i==RTL8651_FDB_NUMBER)//First installed spanning tree instance
	{
		rtl8651_setAsicSpanningEnable(FALSE);
		/* not enable spanning tree working yet, so the spanning tree procces is still disabled */
		rtl8651_tblDrvSpanningTreeEnable = FALSE;
	}
	spanTreePtr->valid = 1;
	spanTreePtr->protocolWorking = 0;//Protocol does not working
	for(i=0;i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++) {
		spanTreePtr->portState[i] = RTL8651_PORTSTA_FORWARDING;//in forwarding state at initial time
		rtl8651_setAsicMulticastSpanningTreePortState(i, RTL8651_PORTSTA_FORWARDING);// sync to multicast spanning tree status
		rtl8651_tblDrvSpanningTreePortFwdStatus |= (1<<i);	// sync to SW fwdStatus
	}
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

/*
@func int32 | rtl8651_setSpanningTreeInstanceProtocolWorking | Enable/Disable Spanning Tree Protocol.
@parm uint16 | sid | Spanning Tree ID.
@parm int8 | working | TRUE: Enable, FALSE: Disable.
@rvalue TBLDRV_EINVALIDSID | Invalid Spanning Tree ID.
@comm
The API is used to Enable or Disable a Spanning Tree Protocol.
*/
int32 rtl8651_setSpanningTreeInstanceProtocolWorking(uint16 sid, int8 working) {
	rtl8651_tblDrv_spanningTreeTable_t *spanTreePtr, *tmpPtr;
	uint32 i, workingCount;
	int8 previousWorking;
	
	UNLOCK_ERROR_CHECK((sid >= RTL8651_STI_NUMBER), TBLDRV_EINVALIDSID);
	rtlglue_drvMutexLock();//Lock resource
	spanTreePtr = &DrvTbl.spanningTree[sid];
	LOCK_ERROR_CHECK((spanTreePtr->valid == 0), TBLDRV_EINVALIDSID);
	previousWorking = spanTreePtr->protocolWorking == 1? TRUE: FALSE;
	spanTreePtr->protocolWorking = working == TRUE? 1: 0;
	workingCount = 0;
	for(i=0,tmpPtr=&DrvTbl.spanningTree[0]; i< RTL8651_STI_NUMBER; i++, tmpPtr++)
		if(tmpPtr->valid == 1 && tmpPtr->protocolWorking == 1)
			workingCount++;
	if(previousWorking == TRUE && working == FALSE && workingCount == 0) {
		rtl8651_setAsicSpanningEnable(FALSE);
		/* disable corresponding spanning tree process */
		rtl8651_tblDrvSpanningTreeEnable = FALSE;
	}
	if(previousWorking == FALSE && working == TRUE && workingCount == 1) {
		rtl8651_setAsicSpanningEnable(TRUE);
		/* enable and initial spanning tree process */
		rtl8651_tblDrvSpanningTreeEnable = TRUE;
	}
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

//Configure VLAN specific port (link aggregator) port state. 1. Use vid to find fid. 2. Use fid to find MSID. 3. Configure MSID port state

/*
@func int32 | rtl8651_setSpanningTreeInstancePortState | Configure Spanning Tree port state.
@parm uint16 | stid | Spanning Tree ID.
@parm uint16 | port | Physical port number.
@parm uint8 | portState | <p portState> could be RTL8651_PORTSTA_DISABLED, RTL8651_PORTSTA_BLOCKING, RTL8651_PORTSTA_LISTENING, RTL8651_PORTSTA_LEARNING, RTL8651_PORTSTA_FORWARDING.
@rvalue SUCCESS | API Success.
@rvalue TBLDRV_EINVALIDPORT | Invalid Port number.
@rvalue TBLDRV_EINVALIDSID | Invalid Spanning Tree ID.
@rvalue TBLDRV_EINVALIDPORTSTATE | Invalid Port State.
@comm
For each port, it has five states: Disabled, Blocking, Listening, Learning and Forwarding. This API will set each port's state. Note that currently ONLY SID=0 is allowed. 
*/
int32 rtl8651_setSpanningTreeInstancePortState(uint16 stid, uint16 port, uint8 portState) {
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	rtl8651_tblDrv_spanningTreeTable_t *spanTreePtr;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	uint16 i, vid;

	#ifdef STP_DEBUG
	uint32 oldState = rtl8651_tblDrvSpanningTreePortFwdStatus;
	printk("[%s]\n\t=> stid %d port %d portState %d\n", __FUNCTION__, stid, port, portState);
	#endif
	
	UNLOCK_ERROR_CHECK((port >= RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
	UNLOCK_ERROR_CHECK((stid >= RTL8651_STI_NUMBER), TBLDRV_EINVALIDSID);
	UNLOCK_ERROR_CHECK((portState > 4), TBLDRV_EINVALIDPORTSTATE);
	rtlglue_drvMutexLock();//Lock resource
	aggrePtr = &DrvTbl.aggregator[port];
	spanTreePtr = &DrvTbl.spanningTree[stid];
	LOCK_ERROR_CHECK((aggrePtr->ethernetPortMask == 0), TBLDRV_EINVALIDPORT);
	LOCK_ERROR_CHECK((spanTreePtr->valid == 0), TBLDRV_EINVALIDSID);
	//Configure all link aggraged port with the same port
	for(i = 0; i< RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++)
		if(aggrePtr->ethernetPortMask & (1<<i)) {

			#ifdef STP_DEBUG
			uint32 tempOld = rtl8651_tblDrvSpanningTreePortFwdStatus;
			printk("[%s]\n\t=> Update port[%d] State: [%s] -> [%s]\n",	__FUNCTION__,
															i,
															__STPSTATE(spanTreePtr->portState[i]),
															__STPSTATE(portState));
			#endif

			/* update date structure */
			spanTreePtr->portState[i] = portState; //Set member of link aggregator port state
			rtl8651_setAsicMulticastSpanningTreePortState(i, portState);	// sync to multicast spanning tree port state
			// sync to forwarding engine's forwarding datebase
			rtl8651_tblDrvSpanningTreePortFwdStatus =
				(portState==RTL8651_PORTSTA_FORWARDING)?
					(rtl8651_tblDrvSpanningTreePortFwdStatus|(1<<i)):
					(rtl8651_tblDrvSpanningTreePortFwdStatus&~(1<<i));


			#ifdef STP_DEBUG
			if (rtl8651_tblDrvSpanningTreePortFwdStatus != tempOld) {
				printk("\tUpdate Port [%d]: %s\n",
					i,
					(portState==RTL8651_PORTSTA_FORWARDING)?
					"BLK -> FWD":"FWD -> BLK");
			}
			#endif
		}

	#ifdef STP_DEBUG
	if (oldState != rtl8651_tblDrvSpanningTreePortFwdStatus) {
		printk("[%s]\n\t=>Forwarding state update: (%x) --> (%x)\n",
					__FUNCTION__,
					oldState,
					rtl8651_tblDrvSpanningTreePortFwdStatus);
	} else
		printk("[%s]\n\t=>Forwarding state not change (%x)\n", __FUNCTION__, oldState);
	#endif

	//Configure port state into actual VLAN table (8651 model)
	for(vid = 0, vlanPtr=&DrvTbl.vlan[0]; vid<RTL8651_VLAN_NUMBER; vid++, vlanPtr++)
		if(vlanPtr->valid == 1) {
			assert(vlanPtr->fid<RTL8651_FDB_NUMBER);
			assert(DrvTbl.filterDB[vlanPtr->fid].valid == 1);
			assert(DrvTbl.filterDB[vlanPtr->fid].sid<RTL8651_STI_NUMBER);
			assert(DrvTbl.spanningTree[DrvTbl.filterDB[vlanPtr->fid].sid].valid == 1);
			if(DrvTbl.filterDB[vlanPtr->fid].sid == stid)
				RTL8651_SETASICVLAN(vlanPtr);
		}

	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

/*
@func int32 | rtl8651_delSpanningTreeInstance | Remove a Spanning Tree.
@parm uint16 | sid | Spanning Tree ID.
@rvalue SUCCESS | API Success.
@rvalue TBLDRV_EINVALIDSID | Invalid Spanning Tree ID.
@rvalue TBLDRV_ESIDISREFERENCEEDBYFID | Spanning Tree is being referenced by Filter Database.
@comm
Currently only Spanning Tree ID = 0 is allowed.
*/
int32 rtl8651_delSpanningTreeInstance(uint16 sid) {
	rtl8651_tblDrv_spanningTreeTable_t *spanTreePtr;
	rtl8651_tblDrv_filterDbTable_t *fdbPtr;
	uint16 fid;

	/* Invalid SID: either out of SID or SID doesn't exist */
	UNLOCK_ERROR_CHECK((sid >= RTL8651_FDB_NUMBER), TBLDRV_EINVALIDSID);
	rtlglue_drvMutexLock();//Lock resource
	spanTreePtr = &DrvTbl.spanningTree[sid];
	LOCK_ERROR_CHECK((spanTreePtr->valid == 0), TBLDRV_EINVALIDSID);
	for(fid=0, fdbPtr=&DrvTbl.filterDB[0]; fid<RTL8651_FDB_NUMBER; fid++, fdbPtr++) {
		/* Spanning tree instance is referenced by some filter database */
		LOCK_ERROR_CHECK((fdbPtr->valid==1&&fdbPtr->sid==sid), TBLDRV_ESIDISREFERENCEEDBYFID);
		}
	spanTreePtr->valid = 0;//Invalidate corresponding spanning tree instance
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}


#define RTL8651_FILTER_DATABASE_APIs
/*=======================================================================================
 * RTL8651 Filter DataBase APIs:
 *	1. External APIs
 *		. rtl8651_addFilterDatabase()
 *		. rtl8651_specifyFilterDatabaseSpanningTreeInstance()
 *		. rtl8651_delFilterDatabase()
 *=======================================================================================*/
/*
@func int32 | rtl8651_addFilterDatabase | Add a filter database.
@parm uint16 | fid | Filter Database ID.
@rvalue TBLDRV_EINVALIDFID | Invalid Filter Database ID.
@rvalue TBLDRV_EFIDALREADYEXISTS | The specified Filter Database ID already exists.
@comm
Currently only one Filter Database(FID=0)is allowed.
*/
int32 rtl8651_addFilterDatabase(uint16 fid) {
	rtl8651_tblDrv_filterDbTable_t *fdbPtr;
	ether_addr_t mac = { {0xff, 0xff, 0xff, 0xff, 0xff, 0xff} };
	int32 retval, i;
	uint32 portmask=0xffffffff;

	/* Filter database id out of range */
	UNLOCK_ERROR_CHECK((fid >= RTL8651_FDB_NUMBER), TBLDRV_EINVALIDFID);
	rtlglue_drvMutexLock();//Lock resource	
	fdbPtr = &DrvTbl.filterDB[fid];
	LOCK_ERROR_CHECK((fdbPtr->valid == 1), TBLDRV_EFIDALREADYEXISTS);
	fdbPtr->valid = 1;	//Enable filter database
	assert(DrvTbl.spanningTree[0].valid);//Spanning tree instance 0 should exist
	fdbPtr->sid = 0;	//Using first spanning tree database
	for(i=0; i<RTL8651_L2TBL_ROW; i++)
		SLIST_INIT(&(fdbPtr->database[i]));
	/* Once a filter database was added, add a filter database entry automatically */
	retval = _rtl8651_addVlanRefFilterDatabaseEntry(RTL8651_L2_TYPEII, fid, 0, &mac, RTL8651_FORWARD_MAC, portmask);

	assert(retval == SUCCESS);
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

/*
@func int32 | rtl8651_specifyFilterDatabaseSpanningTreeInstance | Specify which Spanning Tree is used by a Filter database.
@parm uint16 | fid | Filter Database ID.
@parm uint16 | sid | Spanning Tree ID.
@rvalue TBLDRV_EINVALIDFID | Invalid Filter Database ID.
@rvalue TBLDRV_EINVALIDSID | Invalid Spanning Tree ID.
*/
int32 rtl8651_specifyFilterDatabaseSpanningTreeInstance(uint16 fid, uint16 sid) {
	rtl8651_tblDrv_filterDbTable_t *fdbPtr;
	int32 retval = SUCCESS;
	
	UNLOCK_ERROR_CHECK((fid >= RTL8651_FDB_NUMBER), TBLDRV_EINVALIDFID);
	UNLOCK_ERROR_CHECK((sid >= RTL8651_FDB_NUMBER), TBLDRV_EINVALIDSID);
	rtlglue_drvMutexLock();//Lock resource
	fdbPtr = &DrvTbl.filterDB[fid];
	if(fdbPtr->valid == 0){
		retval = TBLDRV_EINVALIDFID;
	}else if(DrvTbl.spanningTree[sid].valid == 0)
		retval = TBLDRV_EINVALIDSID;
	else fdbPtr->sid = sid;	//Using specified spanning tree database
	rtlglue_drvMutexUnlock();//Unlock resource
	return retval;
}

/*
@func int32 | rtl8651_delFilterDatabase | Remove a Filter Database.
@parm uint16 | fid | Filter database ID.
@rvalue TBLDRV_EINVALIDFID | Invlaid Filter database ID.
@rvalue TBLDRV_EFIDISREFERENCEDBYVLAN | Filter Database is being referenced by VLAN.
@rvalue TBLDRV_EFIDISNOTEMPTY | Filter Database is not empty.
*/
int32 rtl8651_delFilterDatabase(uint16 fid) {
	ether_addr_t mac = { {0xff, 0xff, 0xff, 0xff, 0xff, 0xff} };
	rtl8651_tblDrv_filterDbTable_t *fdbPtr;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	int32 retval, i;

	/* Filter database out of range */
	UNLOCK_ERROR_CHECK((fid >= RTL8651_FDB_NUMBER), TBLDRV_EINVALIDFID);
	rtlglue_drvMutexLock();//Lock resource
	fdbPtr = &DrvTbl.filterDB[fid];
	LOCK_ERROR_CHECK((fdbPtr->valid == 0), TBLDRV_EINVALIDFID);
	//Forbid filter database delete when VLAN reference
	for(i=0, vlanPtr=&DrvTbl.vlan[0]; i<RTL8651_VLAN_NUMBER; i++, vlanPtr++)
		LOCK_ERROR_CHECK((vlanPtr->valid == 1 && vlanPtr->fid == fid), TBLDRV_EFIDISREFERENCEDBYVLAN);
	for(i=0; i<RTL8651_L2TBL_ROW; i++) //Forbid database deletion when database entry exist
		LOCK_ERROR_CHECK((SLIST_FIRST(&(fdbPtr->database[i]))), TBLDRV_EFIDISNOTEMPTY);
	/* Once a filter database was removed, remove a default L2 entry automatically */
	retval = _rtl8651_delVlanRefFilterDatabaseEntry(RTL8651_L2_TYPEII, fid, &mac);
	if (retval == SUCCESS)
		fdbPtr->valid = 0;	//Disable filter database
	rtlglue_drvMutexUnlock();//Unlock resource
	return retval;
}


static void _rtl8651_arrangeAsicL2Entry(uint16 fid, uint32 l2Index) {
	rtl8651_tblDrv_filterDbTableEntry_t *l2_old[RTL8651_L2TBL_COLUMN];
	rtl8651_tblDrv_filterDbTableEntry_t *l2_new[RTL8651_L2TBL_COLUMN];
	rtl8651_tblDrv_filterDbTableEntry_t *l2Entry;
	rtl8651_tblDrv_filterDbTable_t *fdbPtr= &DrvTbl.filterDB[fid];
	int8 cpu , srcBlk , occupied[RTL8651_L2TBL_COLUMN] = { FALSE, FALSE, FALSE, FALSE };
	uint32 colIdx, i, j;
	int32 useNxtHopFlag=0;
	for(colIdx=0; colIdx<RTL8651_L2TBL_COLUMN; colIdx++) 
		l2_old[colIdx] = l2_new[colIdx] = NULL;
	SLIST_FOREACH(l2Entry, &(fdbPtr->database[l2Index]), nextFDB) 
		if (l2Entry->configToAsic == 1)
			l2_old[l2Entry->asicPos] = l2Entry;

	colIdx = 0;
	/* TYPE II has high priority */
	SLIST_FOREACH(l2Entry, &(fdbPtr->database[l2Index]), nextFDB) {
		if (colIdx == RTL8651_L2TBL_COLUMN)
			goto __pass;
		if (l2Entry->l2type & RTL8651_L2_TYPEII) 
			l2_new[colIdx++] = l2Entry; 
	}
	/* Then TYPE III */
	SLIST_FOREACH(l2Entry, &(fdbPtr->database[l2Index]), nextFDB)  {
		if (colIdx == RTL8651_L2TBL_COLUMN)
			goto __pass;
		if (l2Entry->l2type & RTL8651_L2_TYPEII) 
			continue; /* This entry has been taken into consideration */
		if (l2Entry->l2type & RTL8651_L2_TYPEIII)
			l2_new[colIdx++] = l2Entry;
	}
	/* Last TYPE I */
	SLIST_FOREACH(l2Entry, &(fdbPtr->database[l2Index]), nextFDB) {
		if (colIdx == RTL8651_L2TBL_COLUMN)
			goto __pass;
		if (l2Entry->l2type & RTL8651_L2_TYPEII || l2Entry->l2type & RTL8651_L2_TYPEIII)
			continue; /* This entry has been taken into consideration */
		if (l2Entry->l2type & RTL8651_L2_TYPEI)
			l2_new[colIdx++] = l2Entry;
	}

__pass:

	for(i=0; i<RTL8651_L2TBL_COLUMN; i++)
		for(j=0; l2_new[i]&&j<RTL8651_L2TBL_COLUMN; j++)
			if (l2_new[i] == l2_old[j]) {
				occupied[l2_old[j]->asicPos] = TRUE;
				l2_new[i] = l2_old[j] = NULL;
				break;
		}

	colIdx = 0;
	for(i=0; i<RTL8651_L2TBL_COLUMN; i++) {
		if ((l2Entry=l2_new[i]) == NULL)
			continue;
		while(colIdx < RTL8651_L2TBL_COLUMN && occupied[colIdx]==TRUE)
			colIdx++;
		assert(colIdx < RTL8651_L2TBL_COLUMN);
		l2Entry->configToAsic = 1;
		l2Entry->asicPos = colIdx;
		occupied[colIdx] = TRUE;
		switch(l2Entry->process) 
		{
		default:
		case RTL8651_FORWARD_MAC:
		case RTL8651_DSTBLOCK_MAC:  
			cpu = srcBlk = FALSE;
			break;
		case RTL8651_SRCBLOCK_MAC:  
			cpu = FALSE; srcBlk = TRUE;
			break;
		case RTL8651_TOCPU_MAC: 	
			cpu = TRUE; srcBlk = FALSE;			
		}
//#ifdef CONFIG_RTL865X_WLAN_ACCELERATION					
		 //if chip is 8650B, we don't care about port member since it would be auto learnt.
		if(peripheralExtPortMask)
			useNxtHopFlag=1;  /* set to 1 for using nextHostFlag orlando 2003/11/20 */
//#endif
		rtl8651_setAsicL2Table_Patch(
					l2Index, 
					l2Entry->asicPos, 
					&l2Entry->macAddr, 
					cpu, 
					srcBlk, 
					l2Entry->memberPortMask, 
					arpAgingTime, 
					TRUE, 
					useNxtHopFlag
					);
		if(rtl8651_L2statusNotifier)
			rtl8651_L2statusNotifier(l2Entry,NULL);
	}
	for(i=0; i<RTL8651_L2TBL_COLUMN; i++)
		if (l2_old[i]) {
			l2_old[i]->configToAsic = 0;
			if(rtl8651_L2statusNotifier)
				rtl8651_L2statusNotifier(l2_old[i],NULL);
		}
	for(i=0; i<RTL8651_L2TBL_COLUMN; i++)
		if (occupied[i] == FALSE)
			rtl8651_delAsicL2Table(l2Index, i);
}

	
int32 _rtl8651_addVlanRefFilterDatabaseEntry(uint16 l2Type, uint16 fid, uint16 vid, ether_addr_t * macAddr, uint32 type, uint32 portMask) {
	rtl8651_tblDrv_filterDbTable_t *fdb_t = &DrvTbl.filterDB[fid];
	rtl8651_tblDrv_filterDbTableEntry_t * l2entry_t, *tmpL2;
	uint32 rowIdx = rtl8651_filterDbIndex(macAddr), typeII = 0;

#ifdef RTL865XB_DCUT_SWVLAN
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr = NULL;
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];

	if (vlanPtr->valid == 0 || vlanPtr->vid != vid) {
		swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
		if (swVlanPtr->valid == 0)
			swVlanPtr = NULL;
	}
#endif

	#if 0
	rtlglue_printf("(1) macAddr: %x.%x.%x.%x.%x.%x\n", 
		macAddr->octet[0], macAddr->octet[1], macAddr->octet[2], 
		macAddr->octet[3], macAddr->octet[4], macAddr->octet[5]);
	#endif

	SLIST_FOREACH(l2entry_t, &(fdb_t->database[rowIdx]), nextFDB) {
		/* Check how many TypeII entries have been added */
		if (l2entry_t->l2type & RTL8651_L2_TYPEII)
			typeII++;
	}
	/* TypeII entry can not exceed RTL8651_L2TBL_COLUMN */
	if (typeII == RTL8651_L2TBL_COLUMN && l2Type == RTL8651_L2_TYPEII) 
		return TBLDRV_ENOFREEBUFFER;
	
	
	SLIST_FOREACH(l2entry_t, &(fdb_t->database[rowIdx]), nextFDB) 
		if (!memcmp(&l2entry_t->macAddr, macAddr, sizeof(ether_addr_t))) 
			break; /* Duplicate entry found */
	if (l2entry_t != (rtl8651_tblDrv_filterDbTableEntry_t *)NULL) {
		switch(l2Type)
		{
		case RTL8651_L2_TYPEI:	/* Referenced by ARP/PPPoE. 2nd priority */
			if (l2entry_t->process != 0)	
				return FAILED;
			if (l2entry_t->memberPortMask != portMask) {
				/* always trust new information */
				l2entry_t->memberPortMask = portMask;
#if defined(RTL865X_MODEL_USER) || defined(RTL865X_MODEL_KERNEL)
				/* In model code, we don't want to compile tblDrv.
				   Thus, do not call _rtl8651_removeArpByL2Entry(). */
#else
	#ifdef CONFIG_RTL865X_MULTILAYER_BSP
				/* Only in the multiple layer, we need to call this function. */
				_rtl8651_removeArpByL2Entry(l2entry_t);
	#endif
#endif
				rtl8651_setAsicL2Table_Patch(
					rowIdx, 
					l2entry_t->asicPos, 
					&l2entry_t->macAddr, 
					FALSE, 
					FALSE, 
					l2entry_t->memberPortMask, 
					arpAgingTime, 
					TRUE, 
					FALSE
					);
				return SUCCESS;
			}
#if 0		
			 if (l2entry_t->process != 0 || l2entry_t->memberPortMask != portMask) {
				rtlglue_printf("[new]: %x:%x:%x:%x:%x:%x portMask: %x, vid:%d\n", macAddr->octet[0], macAddr->octet[1], macAddr->octet[2], macAddr->octet[3], macAddr->octet[4], macAddr->octet[5], portMask, vid);
				rtlglue_printf("[old]: %x:%x:%x:%x:%x:%x portMask: %x, refCount: %d, vid: %d\n\ttype: %x, pro: %x\n",
						l2entry_t->macAddr.octet[0], l2entry_t->macAddr.octet[1], l2entry_t->macAddr.octet[2], 
						l2entry_t->macAddr.octet[3], l2entry_t->macAddr.octet[4], l2entry_t->macAddr.octet[5],
						l2entry_t->memberPortMask, l2entry_t->refCount, l2entry_t->vid, l2entry_t->l2type,
						l2entry_t->process);
				return FAILED;
			 }
#endif

#ifdef RTL865XB_DCUT_SWVLAN
			if (swVlanPtr != NULL)
				rtl8651_setAsicL2Table_Patch(
					rowIdx, 
					l2entry_t->asicPos, 
					&l2entry_t->macAddr, 
					FALSE, 
					FALSE, 
					l2entry_t->memberPortMask, 
					arpAgingTime,
					TRUE,
					FALSE
					);
#endif

			 l2entry_t->l2type |= RTL8651_L2_TYPEI;
			 l2entry_t->refCount++;
			 return SUCCESS;
			 
		case RTL8651_L2_TYPEII:	/* Referenced by Protocol. Has highest priority to write to ASIC*/
			 if (l2entry_t->l2type & RTL8651_L2_TYPEII || l2entry_t->memberPortMask != portMask)
			 	return TBLDRV_EL2ENTRYEXISTS;
			 l2entry_t->l2type |= RTL8651_L2_TYPEII;
			 break;
			 
		case RTL8651_L2_TYPEIII:	/* Referenced by PCI/Extension Port. Lowest priority */
			 if (l2entry_t->l2type & RTL8651_L2_TYPEIII || l2entry_t->memberPortMask != portMask)
			 	return TBLDRV_EL2ENTRYEXISTS;
			 l2entry_t->l2type |= RTL8651_L2_TYPEIII;
		}	
		_rtl8651_arrangeAsicL2Entry(fid, rowIdx);
		return SUCCESS;
	}

	/* Entry doesn't exist, allocate one */
	if(SLIST_FIRST(&DrvTbl.freeList.filterDBentry) == NULL) 
		return(TBLDRV_ENOFREEBUFFER);	
	
	/* Allocate free entry */
	l2entry_t = SLIST_FIRST(&DrvTbl.freeList.filterDBentry);
	SLIST_REMOVE_HEAD(&DrvTbl.freeList.filterDBentry, nextFDB);
	/* Initial entry content*/
	//l2entry_t->refCount = (RTL8651_L2_TYPEI == TRUE)? 1: 0;
	l2entry_t->refCount = ((l2Type == RTL8651_L2_TYPEI)? 1: 0);
	l2entry_t->macAddr = *macAddr;
	l2entry_t->vid = vid;
	l2entry_t->configToAsic = 0;
	l2entry_t->l2type = l2Type;
	l2entry_t->linkId=0; //default on link #0
	
	switch(type) {	
	case RTL8651_FORWARD_MAC:
	 	 l2entry_t->process = RTL8651_FORWARD_MAC;
		 l2entry_t->memberPortMask = portMask;
		 break;
	case RTL8651_DSTBLOCK_MAC:
		 l2entry_t->process = RTL8651_DSTBLOCK_MAC;
		 l2entry_t->memberPortMask = 0;
		 break;
	case RTL8651_SRCBLOCK_MAC:
		 l2entry_t->process = RTL8651_SRCBLOCK_MAC;
		 l2entry_t->memberPortMask = 0;
		 break;
	case RTL8651_TOCPU_MAC:
		 l2entry_t->process = RTL8651_TOCPU_MAC;
		 l2entry_t->memberPortMask = 0;
	 	 break;
	default: assert(0);
	}
	if (SLIST_FIRST(&(fdb_t->database[rowIdx]))) {
		SLIST_FOREACH(tmpL2, &(fdb_t->database[rowIdx]), nextFDB) 
			if (tmpL2->nextFDB.sle_next == NULL) {
				tmpL2->nextFDB.sle_next = l2entry_t;
				l2entry_t->nextFDB.sle_next = NULL;
			break;
		}
	} else 
		SLIST_INSERT_HEAD(&(fdb_t->database[rowIdx]), l2entry_t, nextFDB);

//#ifdef CONFIG_RTL865X_WLAN_ACCELERATION
	//In 8650B, Type III entries are NOT used.
	if(l2Type&RTL8651_L2_TYPEIII &&rtl8651_L2statusNotifier){
		//new station found. If chip is 8650/51, add a patch rule to ACL.
		rtl8651_L2statusNotifier(NULL,NULL);
	}
//#endif

	_rtl8651_arrangeAsicL2Entry(fid, rowIdx);
	return SUCCESS;
}

 int32 _rtl8651_modifyVlanRefFilterDatabaseEntry(uint32 fid, ether_addr_t *mac, uint32 *memberPortMask, uint16 *process, uint32 *linkId, uint32 attr){
	rtl8651_tblDrv_filterDbTableEntry_t * thisL2entry;
	rtl8651_tblDrv_filterDbTable_t *fdbPtr;
	uint32 rowIdx = rtl8651_filterDbIndex(mac);
	assert(fid < RTL8651_FDB_NUMBER);
	assert(mac);
	fdbPtr = &DrvTbl.filterDB[fid];
	assert(fdbPtr->valid == 1);

	//Check whether entry exist
	SLIST_FOREACH(thisL2entry, &(fdbPtr->database[rowIdx]), nextFDB){ 
		if(memcmp(&thisL2entry->macAddr, mac, sizeof(ether_addr_t)) == 0)
			break;
	}
	if(!thisL2entry)
		return FAILED;
	if(memberPortMask){
		//user wants to change member portmask.
		thisL2entry->memberPortMask=*memberPortMask;
	}
	if(process)
		thisL2entry->process=*process;
	if(linkId)
		thisL2entry->linkId=*linkId;
	if(thisL2entry->configToAsic){
		rtl865x_tblAsicDrv_l2Param_t asic_l2;
		bzero(&asic_l2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		rtl8651_getAsicL2Table_Patch(rowIdx, thisL2entry->asicPos, &asic_l2);
//#ifdef CONFIG_RTL865X_WLAN_ACCELERATION	
		if(peripheralExtPortMask)
			asic_l2.nhFlag=1;
//#endif
		if (attr == L2_FORCE_STATIC) {
			asic_l2.isStatic = TRUE;
			asic_l2.nhFlag = 0;
		}
		else if (attr == L2_FORCE_DYNAMIC) {
			asic_l2.isStatic = FALSE;
			asic_l2.nhFlag = 0;
		}
		asic_l2.macAddr = thisL2entry->macAddr;
		asic_l2.memberPortMask = thisL2entry->memberPortMask&(RTL8651_PHYSICALPORTMASK|rtl8651_allExtPortMask);
		rtl8651_setAsicL2Table(rowIdx, thisL2entry->asicPos, &asic_l2);
	}
//#ifdef CONFIG_RTL865X_WLAN_ACCELERATION
	if(rtl8651_L2statusNotifier)
		rtl8651_L2statusNotifier(NULL,NULL);
//#endif	
return SUCCESS;
}


int32 _rtl8651_delVlanRefFilterDatabaseEntry(uint16 l2Type, uint16 fid, ether_addr_t * macAddr) {
	rtl8651_tblDrv_filterDbTable_t *fdb_t = &DrvTbl.filterDB[fid];
	rtl8651_tblDrv_filterDbTableEntry_t * l2entry_t;
	uint32 rowIdx = rtl8651_filterDbIndex(macAddr);

	SLIST_FOREACH(l2entry_t, &(fdb_t->database[rowIdx]), nextFDB)
		if (!memcmp(&l2entry_t->macAddr, macAddr, sizeof(ether_addr_t)))
			break;
	if (l2entry_t == (rtl8651_tblDrv_filterDbTableEntry_t *)NULL)
		return TBLDRV_EL2ENTRYNOTFOUND;
		
	switch(l2Type) 
	{
	case RTL8651_L2_TYPEI:
		 if (!(l2entry_t->l2type & RTL8651_L2_TYPEI))
			 return TBLDRV_EL2ENTRYNOTFOUND;
		 assert(l2entry_t->refCount > 0);
		 if (--l2entry_t->refCount == 0)
		 	l2entry_t->l2type &= ~RTL8651_L2_TYPEI;
		 break;
		 
	case RTL8651_L2_TYPEII:
		 if (!(l2entry_t->l2type & RTL8651_L2_TYPEII))
			 return TBLDRV_EL2ENTRYNOTFOUND;
		 l2entry_t->l2type &= ~RTL8651_L2_TYPEII;
		 break;
	
	case RTL8651_L2_TYPEIII:
		 if (!(l2entry_t->l2type & RTL8651_L2_TYPEIII))
			 return TBLDRV_EL2ENTRYNOTFOUND;
		 l2entry_t->l2type &= ~RTL8651_L2_TYPEIII;
		 break;
	}

	if (l2entry_t->l2type == 0x00) {
		SLIST_REMOVE(
			&(fdb_t->database[rowIdx]), 
			l2entry_t, 
			 rtl8651_tblDrv_filterDbTableEntry_s,
			 nextFDB
		);
		SLIST_INSERT_HEAD(&DrvTbl.freeList.filterDBentry, l2entry_t, nextFDB);
	}
	_rtl8651_arrangeAsicL2Entry(fid, rowIdx);
	return SUCCESS;
}
				
 rtl8651_tblDrv_filterDbTableEntry_t * 
_rtl8651_getVlanFilterDatabaseEntry(uint16 vid, ether_addr_t * macAddr) {
	rtl8651_tblDrv_filterDbTableEntry_t * tempFilterDbEntry;
	rtl8651_tblDrv_filterDbTable_t *fdbPtr;

	fdbPtr = &DrvTbl.filterDB[0];
	
	//assert(vid&&vid<4095);
	//vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	//fdbPtr = &DrvTbl.filterDB[vlanPtr->fid];
	//If no such VLAN exist, 
	//if(vlanPtr->valid == 0 || vlanPtr->vid != vid)
	//	return (rtl8651_tblDrv_filterDbTableEntry_t *)NULL;
	//assert(vlanPtr->fid < RTL8651_FDB_NUMBER);
	//assert(fdbPtr->valid == 1);
	//Check whether entry exist
	#if 0
	printk("_rtl8651_getVlanFilterDatabaseEntry mac: %x.%x.%x.%x.%x.%x\n", 
		macAddr->octet[0], macAddr->octet[1], macAddr->octet[2],
		macAddr->octet[3], macAddr->octet[4], macAddr->octet[5]);
	#endif
	SLIST_FOREACH(tempFilterDbEntry, &(fdbPtr->database[rtl8651_filterDbIndex(macAddr)]), nextFDB) 
		if(memcmp(&tempFilterDbEntry->macAddr, macAddr, sizeof(ether_addr_t)) == 0)
			return tempFilterDbEntry;
	return (rtl8651_tblDrv_filterDbTableEntry_t *)NULL;
}

/*
@func int32 | rtl8651_addFilterDatabaseEntry | Add a Filter database entry.
@parm uint16 | fid | Filter database ID.
@parm uint16 | vid | VLAN ID.
@parm ether_addr_t * | macAddr | Pointer to a MAC Address.
@parm uint32 | type | Action type, could be RTL8651_FORWARD_MAC, RTL8651_DSTBLOCK_MAC, RTL8651_SRCBLOCK_MAC, RTL8651_TOCPU_MAC.
@parm uint32 | portMask | Port mask.
@rvalue SUCCESS | API Success.
@rvalue TBLDRV_EINVALIDACTIONTYPE | Invlaid port action type.
@rvalue TBLDRV_EINVALIDFID | Invalid Filter database ID.
@rvalue TBLDRV_ENULLMACADDR | The specified MAC address is NULL.
@rvalue TBLDRV_ENOFREEBUFFER | Out of free buffer.
@rvalue TBLDRV_EL2ENTRYEXISTS | The filter database entry already exists.
*/
int32 rtl8651_addFilterDatabaseEntry(uint16 fid, ether_addr_t * macAddr, uint32 type, uint32 portMask) {
	int32 retval;

	UNLOCK_ERROR_CHECK(
		( (type != RTL8651_FORWARD_MAC)&&(type != RTL8651_DSTBLOCK_MAC)&&
		  (type != RTL8651_SRCBLOCK_MAC)&&(type != RTL8651_TOCPU_MAC)
		), TBLDRV_EINVALIDACTIONTYPE
	);
	UNLOCK_ERROR_CHECK((fid >= RTL8651_FDB_NUMBER), TBLDRV_EINVALIDFID);
	UNLOCK_ERROR_CHECK((macAddr == (ether_addr_t *)NULL), TBLDRV_ENULLMACADDR);	
	rtlglue_drvMutexLock();//Lock resource
	LOCK_ERROR_CHECK((DrvTbl.filterDB[fid].valid == 0), TBLDRV_EINVALIDFID);
	retval = _rtl8651_addVlanRefFilterDatabaseEntry(RTL8651_L2_TYPEII, fid, 0, macAddr, type, portMask);
	rtlglue_drvMutexUnlock();//Unlock resource
	return retval;
}

/*
@func int32 | rtl8651_delFilterDatabaseEntry | Remove a filter database entry.
@parm uint16 | fid | Filter database entry.
@parm ether_addr_t * | macAddr | Pointer to a MAC Address.
@rvalue TBLDRV_EINVALIDFID | Filter database ID.
@rvalue TBLDRV_ENULLMACADDR | The specified MAC address is NULL.
@rvalue TBLDRV_EL2ENTRYNOTFOUND | The specified filter database entry was not found.
@rvalue TBLDRV_EINVALIDINPUT | Invalid input parameter.
*/
int32 rtl8651_delFilterDatabaseEntry(uint16 fid, ether_addr_t * macAddr) {
	int32 retval;
	
	UNLOCK_ERROR_CHECK((fid >=RTL8651_FDB_NUMBER), TBLDRV_EINVALIDFID);
	UNLOCK_ERROR_CHECK((macAddr == NULL), TBLDRV_EINVALIDINPUT);
	rtlglue_drvMutexLock();//Lock resource
	LOCK_ERROR_CHECK((DrvTbl.filterDB[fid].valid == 0), TBLDRV_EINVALIDFID);
	retval = _rtl8651_delVlanRefFilterDatabaseEntry(RTL8651_L2_TYPEII, fid, macAddr);
	rtlglue_drvMutexUnlock();//Unlock resource
	return retval;
}



	
#define RTL8651_PORT_AGGREAGATOR_APIs
/*=======================================================================================
 * RTL8651 Port Aggregator APIs:
 *	1. Static Functions
 *		. _rtl8651_returnPortMaskPortNumber()
 *	2. External APIs
 *		. rtl8651_getAggregatorActiveStatus()
 *		. rtl8651_setAggregatorIndividual()
 *		. rtl8651_setPortAggregator()
 *		. rtl8651_setPvid()
 *=======================================================================================*/
static uint32 _rtl8651_returnPortMaskPortNumber(uint32 portMask) {
	uint32 i, portNumber;

	for(i=0, portNumber = 0; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++)
		if(portMask & (1<<i))
			portNumber++;
	return portNumber;
}

/*
@func int32 | rtl8651_getAggregatorActiveStatus | Get aggregator status.
@parm uint16 | aggregator | Aggregator ID.
@parm int8 * | isActive | TRUE for aggregator active, FALSE for aggregator inactive.
@rvalue SUCCESS | API Success.
@rvalue TBLDRV_EINVALIDAGGREGATORID | Invalid Aggregator ID.
@rvalue TBLDRV_EINVALIDINPUT | <p isActive> can not be NULL.
@comm
If ports belong to an aggregator are inactive, the aggregator is inactive. Otherwise the aggregator is active.
*/
int32 rtl8651_getAggregatorActiveStatus(uint16 aggregator, int8 * isActive) {

	UNLOCK_ERROR_CHECK((aggregator >= RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDAGGREGATORID);
	UNLOCK_ERROR_CHECK((isActive == NULL), TBLDRV_EINVALIDINPUT);
	rtlglue_drvMutexLock();//Lock resource
	*isActive = DrvTbl.aggregator[aggregator].ethernetPortUpStatus != 0? TRUE: FALSE;
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

/*
@func int32 | rtl8651_setAggregatorIndividual | Enable/Disable trunking.
@parm uint16 | aggregator | Aggregator ID.
@parm int8 | individual | TRUE: disable trunking capability. FALSE: enable trunking capability.
@rvalue SUCCESS | API Success.
@rvalue TBLDRV_EINVALIDAGGREGATORID | Invalid aggregator ID.
@rvalue TBLDRV_EALREADYTRUNKING | Aggregator already aggregates more than one port.
@rvalue TBLDRV_EAGGREGATORHASNOPORT | Not allow aggregator without port member.
@comm
If an aggregator is set to individual, then the aggregator cannot trunk more than one port together.
*/
int32 rtl8651_setAggregatorIndividual(uint16 aggregator, int8 individual) {
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	
	UNLOCK_ERROR_CHECK((aggregator >=RTL8651_PORT_NUMBER+ rtl8651_totalExtPortNum), TBLDRV_EINVALIDAGGREGATORID);
	rtlglue_drvMutexLock();//Lock resource
	aggrePtr = &DrvTbl.aggregator[aggregator];
	/* If original state and new configured value are equal, no need to modify */
	LOCK_ERROR_CHECK(
		( (aggrePtr->individual == 1 && individual == TRUE) ||
		  (aggrePtr->individual == 0 && individual == FALSE)
		),	SUCCESS
	);
	/* If condition not allow individual state, forbid this configuration */
	LOCK_ERROR_CHECK((aggrePtr->aggregated == 1&&individual == TRUE), TBLDRV_EALREADYTRUNKING);
	/* Not allow aggregator without port member to be configured with properties */
	LOCK_ERROR_CHECK((aggrePtr->ethernetPortMask == 0), TBLDRV_EAGGREGATORHASNOPORT);
	aggrePtr->individual = (individual == TRUE) ? 1: 0;
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

/*
@func int32 | rtl8651_setPortAggregator | Aggregate more than one port to an aggregating port(trunking port or logical port).
@parm uint32 | port | Physical port number.
@parm uint16 | aggregator | Aggregator ID.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@rvalue TBLDRV_EINVALIDAGGREGATORID | Invalid aggregator ID.
@rvalue TBLDRV_EAGGREGATORIDGTPORTNO | Aggreagator ID > Port Number. The aggregator ID > port number is disallowed.
@rvalue TBLDRV_EAGGREGATORSETINDIVIDUAL | Aggregator was set to individual, hence the aggregator can not aggregate ports.
@rvalue TBLDRV_EAGGREGATOREXCEED | The number of aggregator excees the system capability.
@rvalue TBLDRV_EDIFFBROADCASTDOMAIN | Aggregator and specified port are not in the same broadcast domain.
@comm
<p rtl8651_setPortAggregator()> aggregates more than one port to form an aggregating port(trunking port or logical port). Currently, only one aggregating port 
is allowed and aggregator ID can not be greater than specified port number. Note that if no ports are aggregated, the physical port and aggregating port are 
one-to-one mapping.
*/
int32 rtl8651_setPortAggregator(uint32 port, uint16 aggregator) {
	uint32 i, j, orgAggregator, fid, aggregaedNumber;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	rtl8651_tblDrv_filterDbTable_t *fdbPtr;
	rtl8651_tblDrv_spanningTreeTable_t *spanTreePtr;
	rtl8651_tblDrv_linkAggregation_t *orgAggrePtr, *aggrePtr, *tmpAggrePtr;
	int8 portLinkUp;
	
	UNLOCK_ERROR_CHECK((port >= RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
	UNLOCK_ERROR_CHECK((aggregator >= RTL8651_AGGREGATOR_NUMBER), TBLDRV_EINVALIDAGGREGATORID);
	/* Not allow aggregator larger than port...Port should select 
	 * aggregator less or equal to itself
	 */
	UNLOCK_ERROR_CHECK((aggregator > port), TBLDRV_EAGGREGATORIDGTPORTNO);
	rtlglue_drvMutexLock();//Lock resource
	aggrePtr = &DrvTbl.aggregator[aggregator];
	orgAggrePtr = &DrvTbl.aggregator[0];
	for(orgAggregator=0; orgAggregator<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; orgAggregator++, orgAggrePtr++) {
		if(orgAggrePtr->ethernetPortMask & (1<<port))
			break;
	}
	assert(orgAggregator < RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum);
	/* No need to change aggregator */
	LOCK_ERROR_CHECK((orgAggregator == aggregator), SUCCESS);
	/* Original aggregator not allow to aggregate with other
	 * Target aggregator not allow to be aggregated
	 */
	LOCK_ERROR_CHECK(
		(orgAggrePtr->individual == 1 ||aggrePtr->individual == 1), 
		TBLDRV_EAGGREGATORSETINDIVIDUAL
	);
	//Check whether the number of aggregator exceeds system capacity
	tmpAggrePtr = &DrvTbl.aggregator[0];
	for(i=0, aggregaedNumber = 0; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++, tmpAggrePtr++) {
		assert(tmpAggrePtr->aggregated == 1 ? _rtl8651_returnPortMaskPortNumber(tmpAggrePtr->ethernetPortMask)>1 : _rtl8651_returnPortMaskPortNumber(tmpAggrePtr->ethernetPortMask)<= 1);
		if(tmpAggrePtr->aggregated == 1) 
			aggregaedNumber++;
	}
	LOCK_ERROR_CHECK(
		( (aggregaedNumber>=RTL8651_MAX_AGGREGATION_NUM)&& //Unable to add new aggregated aggregator
		  (aggrePtr->aggregated==0)&&//Target aggregator is not already aggregated, unable to operate without increasing aggreaged aggreagtor
		  (orgAggrePtr->aggregated == 0)//Original link is not aggreaged
		), TBLDRV_EAGGREGATOREXCEED
	);
	/* Orignal link is aggreaged but not change to not aggregated after port removal */
	LOCK_ERROR_CHECK(
		(_rtl8651_returnPortMaskPortNumber(orgAggrePtr->ethernetPortMask)>2),
		TBLDRV_EAGGREGATOREXCEED
	);
	for(i=0,vlanPtr=&DrvTbl.vlan[0]; i<RTL8651_VLAN_NUMBER; i++, vlanPtr++) {
		LOCK_ERROR_CHECK(
			( (vlanPtr->valid==1) &&
			  ((vlanPtr->memberPortMask&(1<<port))==0)!=((vlanPtr->memberPortMask&aggrePtr->ethernetPortMask)==0)
			), TBLDRV_EDIFFBROADCASTDOMAIN
		);
		}
	//Start to modify related tables
	orgAggrePtr->ethernetPortMask &= (~(1<<port));
	if(orgAggrePtr->ethernetPortUpStatus & (1<<port))
		portLinkUp = TRUE;
	else
		portLinkUp = FALSE;
	orgAggrePtr->ethernetPortUpStatus &= ~(1<<port);
	if(_rtl8651_returnPortMaskPortNumber(orgAggrePtr->ethernetPortMask)<= 1 && orgAggrePtr->aggregated == 1) {
		orgAggrePtr->aggregated = 0;
		rtl8651_setAsicLinkAggregator(0);//Clear ASIC trunk assignment
	}
	aggrePtr->ethernetPortMask |= (1<<port);
	if(portLinkUp==TRUE)
		aggrePtr->ethernetPortUpStatus |= (1<<port);
	if(_rtl8651_returnPortMaskPortNumber(aggrePtr->ethernetPortMask) > 1) {
		aggrePtr->aggregated = 1;
		rtl8651_setAsicLinkAggregator(aggrePtr->ethernetPortMask); //Assign ASIC trunk assignment
	}
	//Processes when port removal causes aggregator to become invalid. 
	// Change VLAN, filter database . The information about original aggreator is hooked to target aggregator
	if((orgAggrePtr->ethernetPortMask & (~(1<<port))) == 0) {
		assert(orgAggrePtr->aggregated == 0);
		for(i=0,spanTreePtr=&DrvTbl.spanningTree[0]; i<RTL8651_STI_NUMBER; i++,spanTreePtr++) //Store aggregator member port state equal to aggregator state
			if(spanTreePtr->valid == 1)
				for(j=0; j< RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; j++)
					if((aggrePtr->ethernetPortMask & (1<<j) ) && j != aggregator)
						spanTreePtr->portState[j] = spanTreePtr->portState[aggregator];
		//Traverse all filter database to move orgAggregator to new aggregator
		for(fid=0,fdbPtr=&DrvTbl.filterDB[0]; fid<RTL8651_FDB_NUMBER; fid++, fdbPtr++)
			if(fdbPtr->valid == 1) {
				rtl8651_tblDrv_filterDbTableEntry_t * tempFilterDbEntry;
				for(i=0; i<RTL8651_L2TBL_ROW; i++) {
					int32 useNxtHopFlag=0;
					//tempFilterDbEntry = fdbPtr->database[i];
					SLIST_FOREACH(tempFilterDbEntry, &(fdbPtr->database[i]), nextFDB) {
						if(tempFilterDbEntry->memberPortMask & (1<<port)) {
							tempFilterDbEntry->memberPortMask &= (~(1<<port));
							tempFilterDbEntry->memberPortMask |= (1<<aggregator);
							//Write modified entry into ASIC
							// if turn on nextHostFlag, use next line orlando 2003/11/20
//#ifdef CONFIG_RTL865X_WLAN_ACCELERATION								
useNxtHopFlag=1;
//#endif
							rtl8651_setAsicL2Table_Patch(rtl8651_filterDbIndex(&tempFilterDbEntry->macAddr), tempFilterDbEntry->asicPos, &tempFilterDbEntry->macAddr, tempFilterDbEntry->process == RTL8651_SRCBLOCK_MAC?TRUE : FALSE, tempFilterDbEntry->process == RTL8651_TOCPU_MAC ? TRUE: FALSE, tempFilterDbEntry->memberPortMask, arpAgingTime, TRUE, 	useNxtHopFlag /* set to 1 for using nextHostFlag orlando 2003/11/20 */ );
						}
					}
				}
			}
	}
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

uint16  rtl8651_getPvid(uint32 port){
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	aggrePtr = &DrvTbl.aggregator[port];
	return aggrePtr->pvid;
}

int32 _rtl8651_setPvid(uint32 port, uint16 vid) {
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	int32 i;

	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	aggrePtr = &DrvTbl.aggregator[port];
	if (vlanPtr->valid == 0 || vlanPtr->vid != vid)
		return TBLDRV_EINVALIDVLANID;
	//Check whether aggregator valid. The way to check the validation of aggregator is whether at least one port belong to aggregator
	if (aggrePtr->ethernetPortMask == 0)
		return TBLDRV_EAGGREGATORHASNOPORT;
	aggrePtr->pvid = vid;	
	for(i = 0; i< RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++)
		if(aggrePtr->ethernetPortMask & (1<<i))
			rtl8651_setAsicPvid(i, rtl8651_vlanTableIndex(vid));
	return SUCCESS;
}


/*
@func int32 | rtl8651_setPvid | Configure link aggregator VID(logical port's PVID).
@parm uint32 | port | Logical port number(Aggregator ID).
@parm uint16 | vid | VLAN ID.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@rvalue TBLDRV_EINVALIDVLANID | Invlaid VLAN ID.
@rvalue TBLDRV_EAGGREGATORHASNOPORT | Each aggregator at least should have one member port.
@comm
<p rtl8651_setPvid()> configures each Logical port's(aggregator) PVID. Note that if no ports are aggregated, the logical port and physical port
are one-to-one mapping.
*/

int32 rtl8651_setPvid(uint32 port, uint16 vid) 
{
	int32 retval;	
	
	if  (port >= RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum)
		return TBLDRV_EINVALIDPORT;
	if (vid == 0 || vid > 4094)
		return TBLDRV_EINVALIDVLANID;

	rtlglue_drvMutexLock();//Lock resource
	retval = _rtl8651_setPvid(port, vid);
	rtlglue_drvMutexUnlock();//Unlock resource
	return retval;
}




#define RTL8651_VLAN_TABLE_APIs
/*=======================================================================================
 * RTL8651 VLAN APIs:
 *	1. Static Functions
 *		. _rtl8651_updateAsicAllVlan()
 *	2. External APIs
 *		. rtl8651_addVlan()
 *		. rtl8651_delVlan()
 *		. rtl8651_specifyVlanFilterDatabase()
 *		. rtl8651_setVlanPromiscuous()
 *		. rtl8651_setVlanFwdTxMirror()
 *		. rtl8651_getVlanActiveStatus()
 *		. rtl8651_addVlanPortMember()
 *		. rtl8651_setVlanPortUntag()
 *		. rtl8651_assignVlanMacAddress()
 *		. rtl8651_allocateVlanMacAddress()
 *		. rtl8651_getVlanMacAddress()
 *		. rtl8651_delVlanPortMember()
 *=======================================================================================*/

#if 0
#define RTL8651_SETASICVLAN(_vlan_t)\
{\
	rtl8651_setAsicVlan(\
				(_vlan_t)->vid,\
				&((_vlan_t)->macAddr),\
				\
				/*																		\
				 * NOTE:																\
				 *	If a port of a vlan port member is down, we should remove			\
				 *	the port from the vlan's member port list. The reason is:			\
				 *	=> eg:																\
				 *	       1. vlan1 has three port: port0, port1, port2					\
				 *		   2. suppose port0 is down										\
  				 *		   3. an outging packet from the vlan will send to port0,		\
  				 *			  port1 and port2											\
  				 *		   4. because port0 is down, it will use 10Mbps to transmit		\
  				 *			  packet=> Tx complete = outgoing packet using 10Mbps		\
  				 *		   5. port1 and port2											\
  				 *			  Tx complete = outgoing packet using 100Mbps				\
  				 *		   6. finally, wait Tx complete = ouging packet using 10Mbps	\
  				 *		   7. To resolve this problem, we remove down link port to		\
  				 *			  avoid waiting long Tx complete time.						\
				 */																		\
				(_vlan_t)->memberPortMask & phyPortLinkStatus,\
				\
				(_vlan_t)->inAclStart,\
				(_vlan_t)->inAclEnd,\
				(_vlan_t)->outAclStart,\
				(_vlan_t)->outAclEnd,\
				(_vlan_t)->internal == 1 ? TRUE: FALSE,\
				pureSoftFwd == TRUE? FALSE: ((_vlan_t)->ipAttached==1? TRUE: FALSE),\
				DrvTbl.spanningTree[DrvTbl.filterDB[(_vlan_t)->fid].sid].portState,\
				TRUE,\
				(_vlan_t)->promiscuous==1? TRUE: FALSE,\
				(_vlan_t)->untagPortMask,\
				(_vlan_t)->macAddrNumber,\
				1500\
	);\
}
#endif



static void _rtl8651_updateAsicAllVlan(void) {
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	uint32 i;

	for(i=0, vlanPtr=&DrvTbl.vlan[0]; i<RTL8651_VLAN_NUMBER; i++, vlanPtr++) {
		if(vlanPtr->valid == 1) {
			if (pureSoftFwd)
				vlanPtr->softRoute = 1;
			else vlanPtr->softRoute = 0;
			RTL8651_SETASICVLAN(vlanPtr);
		}
	}
}

/*
@func int32	| rtl8651_pureSoftwareFwd	|  Disable/Enable ASIC acceleration. Let software forward all L3/L4 traffic.
@parm int8	| isSoftFwd		| TRUE to turn off ASIC accelerating for L3/4 traffic, FALSE to keep software & ASIC work together for best system performance.
@comm 
If <p isSoftFwd> is TRUE, routing on all created VLANs are disabled. Therefore, all Layer3/Layer4 packets goes to software and ASIC is a pure layer2 switch.
If <p isSoftFwd> is FALSE, enable ASIC to help offload routing/NAT/NAPT tasks from software.

Note: Using pure software forwarding automatically overrides smart ASIC NAPT allocation & arrangement feature. See <p rtl8651_enableSmartNaptCreate()>.

Warning: Using pure software forwarding would slowdown your system performance. 
*/
int32 rtl8651_pureSoftwareFwd(int8 isSoftFwd) {
	rtlglue_drvMutexLock();//Lock resource
	pureSoftFwd = isSoftFwd;
	_rtl8651_updateAsicAllVlan();
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}


int32 _rtl8651_addVlan(uint16 vid)
{
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	uint32 vlanCount, vidx;

	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	UNLOCK_ERROR_CHECK((vlanPtr->valid == 1), TBLDRV_EVLANALREADYEXISTS);
	vlanPtr->valid = 1; //Occupy entry
	vlanPtr->broadcastToCpu = 1; // default is broadcast to CPU
	vlanPtr->vid = vid; // Configure vid
	vlanPtr->netif_t = NULL;
	assert(DrvTbl.filterDB[0].valid == 1); //Filter database 0 should exist
	vlanPtr->fid = 0; //Using default filtering database
	vlanPtr->memberPortMask = 0; //No member inside
	vlanPtr->untagPortMask = 0; //No port exist
	vlanPtr->memberPortUpStatus = 0; //No member port cause not member port up
	vlanPtr->macAddrNumber = 0;//No mac address arranged
	vlanPtr->internal = 1;//Default set network to be internal
	vlanPtr->ipAttached = 0;//Default no IP interface attached over it.
	vlanPtr->promiscuous = 0;//Default not promiscuous. When enabled, no IP, PPPoE or other can attached over it.
	vlanPtr->fwdTxMirror = 0;//Default not mirror Forwarding engine tx packets
	vlanPtr->softRoute = 0; // default enable ASIC accelerate
	vlanPtr->fvlan_t = NULL;
	vlanPtr->mtu = 1500;
	assert(vlanPtr->fid < RTL8651_FDB_NUMBER);// Driver should store correct fid in vlan table
	assert(DrvTbl.filterDB[vlanPtr->fid].valid == 1); // Driver should store valid fid in vlan table
	assert(DrvTbl.spanningTree[DrvTbl.filterDB[vlanPtr->fid].sid].valid == 1); // Driver should store valid sid in filtering database

	/* When VLAN is created, set its default ACL rule to PERMIT */
 	vlanPtr->inAclStart = vlanPtr->inAclEnd = RTL8651_ACLTBL_PERMIT_ALL;
	vlanPtr->outAclStart = vlanPtr->outAclEnd = RTL8651_ACLTBL_PERMIT_ALL;	
	RTL8651_SETASICVLAN(vlanPtr);
	vlanCount = 0;
	for(vidx = 0, vlanPtr=&DrvTbl.vlan[0]; vidx<RTL8651_VLAN_NUMBER; vidx++, vlanPtr++)
		if(vlanPtr->valid == 1)
			vlanCount++;
	if(vlanCount == 1)//Add and success and final count == 1. Change from no VLAN to VLAN exist
		rtl8651_setAsicOperationLayer(2);
	return SUCCESS;
}

#ifdef RTL865XB_DCUT_SWVLAN
int32 _rtl8651_addSwVlan(uint16 vid)
{
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;

	swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
	UNLOCK_ERROR_CHECK((swVlanPtr->valid == 1), TBLDRV_EVLANALREADYEXISTS);
	swVlanPtr->valid = 1;
	swVlanPtr->vid = vid;
	assert(DrvTbl.filterDB[0].valid == 1);
	swVlanPtr->fid = 0;
	swVlanPtr->memberPortMask = 0;
	swVlanPtr->memberPortUpStatus = 0;
	swVlanPtr->untagPortMask = 0;
	assert(swVlanPtr->fid < RTL8651_FDB_NUMBER);
	assert(DrvTbl.filterDB[swVlanPtr->fid].valid == 1);
	assert(DrvTbl.spanningTree[DrvTbl.filterDB[swVlanPtr->fid].sid].valid == 1);

	return SUCCESS;	
}

/* 
    Add software VLAN. This kind of VLAN is used to perform layer-2 forwarding only, 
    which is different from multi-layer VLAN in ASIC.  Therefore, this VLAN inforamtion 
    will not be written into ASIC VLAN table. If one VLAN has been added into ASIC 
    VLAN table for multi-layer forwarding, this VLAN should not be added in software. 
*/
int32 rtl8651_addSwVlan(uint16 vid)
{
	int32 retval;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	
	/* vid should be legal vlan ID */
	UNLOCK_ERROR_CHECK((vid < 1 || vid > 4094), TBLDRV_EINVALIDVLANID);
	rtlglue_drvMutexLock();//Lock resource
	/* Check if this vid has been added into multi-layer VLAN table. */
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	LOCK_ERROR_CHECK((vlanPtr->valid == 1 && vlanPtr->vid == vid), TBLDRV_EVLANALREADYEXISTS);
	/* Start to add this vid into software (layer-2 forwarding) VLAN table. */
	retval = _rtl8651_addSwVlan(vid);
	rtlglue_drvMutexUnlock();//Unlock resource
	return retval;
}

int32 _rtl8651_delSwVlan(uint16 vid) 
{
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;

	swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
	UNLOCK_ERROR_CHECK((swVlanPtr->valid == 0), TBLDRV_EINVALIDVLANID);
	swVlanPtr->valid = 0;

	return SUCCESS;
}

int32 rtl8651_delSwVlan(uint16 vid) 
{
	int32 retval;
	
	UNLOCK_ERROR_CHECK((vid < 1 || vid > 4094), TBLDRV_EINVALIDVLANID);
	rtlglue_drvMutexLock();
	retval = _rtl8651_delSwVlan( vid );
	rtlglue_drvMutexUnlock();
	return retval;
}

#if 0	// Redundant
int32 rtl8651_specifySwVlanFilterDatabase(uint16 vid, uint16 fid) {
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;

	UNLOCK_ERROR_CHECK((vid == 0 || vid > 4094), TBLDRV_EINVALIDVLANID);
	UNLOCK_ERROR_CHECK((fid >= RTL8651_FDB_NUMBER), TBLDRV_EINVALIDFID);
	rtlglue_drvMutexLock();
	swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
	LOCK_ERROR_CHECK((swVlanPtr->valid == 0), TBLDRV_EINVALIDVLANID);
	LOCK_ERROR_CHECK((swVlanPtr->memberPortMask), TBLDRV_EVLANHASHPORTMEMBER);
	LOCK_ERROR_CHECK((DrvTbl.filterDB[fid].valid == 0), TBLDRV_EINVALIDFID);
	swVlanPtr->fid = fid;
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}


int32 rtl8651_addSwVlanPortMember(uint16 vid, uint32 port) {
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	
	UNLOCK_ERROR_CHECK((vid == 0 || vid > 4094), TBLDRV_EINVALIDVLANID);
	UNLOCK_ERROR_CHECK((port >= RTL8651_PORT_NUMBER + rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
	rtlglue_drvMutexLock();
	swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
	aggrePtr = &DrvTbl.aggregator[port];
	LOCK_ERROR_CHECK((swVlanPtr->valid == 0), TBLDRV_EINVALIDVLANID);
	LOCK_ERROR_CHECK((DrvTbl.aggregator[port].ethernetPortMask == 0), TBLDRV_EAGGREGATORHASNOPORT);	
	LOCK_ERROR_CHECK((swVlanPtr->memberPortMask & aggrePtr->ethernetPortMask), SUCCESS);
	swVlanPtr->memberPortMask |= aggrePtr->ethernetPortMask;
	swVlanPtr->untagPortMask |= aggrePtr->ehternetPortMask;	// By default, we add the port to untag list
	swVlanPtr->memberPortUpStatus |= aggrePtr->ethernetPortUpStatus;
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

int32 rtl8651_setSwVlanPortUntag(uint16 vid, uint32 port, int8 untag) {
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;

	//Impossible VID or port number checking
	UNLOCK_ERROR_CHECK((vid == 0 || vid > 4094), TBLDRV_EINVALIDVLANID);
	UNLOCK_ERROR_CHECK((port >= RTL8651_PORT_NUMBER + rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
	rtlglue_drvMutexLock();//Lock resource
	swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
	aggrePtr = &DrvTbl.aggregator[port];
	LOCK_ERROR_CHECK(((swVlanPtr->memberPortMask & aggrePtr->ethernetPortMask) == 0), 
		TBLDRV_EINVALIDPORT
	);
	if(untag == TRUE)
		swVlanPtr->untagPortMask |= aggrePtr->ethernetPortMask;
	else swVlanPtr->untagPortMask &= ~aggrePtr->ethernetPortMask;
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}
#endif	// End of "#if 0"
#endif

int32 _rtl8651_delVlan(uint16 vid) 
{
	rtl8651_tblDrv_networkIntfTable_t *netIfPtr;
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	rtl8651_tblDrv_pppTable_t *pppoePtr;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	uint32 i, vlanCount;

	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	UNLOCK_ERROR_CHECK((vlanPtr->valid == 0 || vlanPtr->vid != vid), TBLDRV_EINVALIDVLANID);
	if(rtl8651_getAsicOperationLayer()>2){
		/* When IP interface using this VLAN, forbid to delete this VLAN */
		netIfPtr = &DrvTbl.netIface[0];
		for(i=0; i<rtl8651_tblDrvPara.networkIntfTableSize; i++, netIfPtr++) {
			UNLOCK_ERROR_CHECK(
				( (netIfPtr->valid == 1) && (netIfPtr->linkLayerType != 0) && 
				  (netIfPtr->vid == vid) 
				), TBLDRV_EVLANISREFERENCEDBYNETIF
			);
		}
		/* PPPoE session references this vlan, forbid */
		for(i=0, pppoePtr=&DrvTbl.pppoe[0]; i<RTL8651_PPPOE_NUMBER; i++,pppoePtr++) {
			UNLOCK_ERROR_CHECK((pppoePtr->valid == 1&& pppoePtr->vid == vid), TBLDRV_EVLANISREFERENCEDBYPPPOE);
		}
	}
	/* Removing vlan should not allow mac address attached */
	UNLOCK_ERROR_CHECK((vlanPtr->macAddrNumber), TBLDRV_EVLANHASMACALLOCATED);
	/* Since vlan 1 always exist and ASIC must have one valid VLAN as PVID */
	for(i=0,aggrePtr=&DrvTbl.aggregator[0]; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++, aggrePtr++)
		if(aggrePtr->ethernetPortMask && aggrePtr->pvid == vid)
			rtl8651_setPvid(i, 1);
	vlanPtr->valid = 0;
	rtl8651_delAsicVlan(vid);//Write to ASIC table
	//Update operation layer
	vlanCount = 0;
	for(i=0, vlanPtr=&DrvTbl.vlan[0]; i<RTL8651_VLAN_NUMBER; i++, vlanPtr++)
		if(vlanPtr->valid == 1)
			vlanCount++;
	if(vlanCount == 0)//Delete and success and final count == 0. Change from VLAN exist to no VLAN exist
		rtl8651_setAsicOperationLayer(1);

	return SUCCESS;
}



/*
@func int32 | rtl8651_addVlan | Add a VLAN.
@parm uint16 | vid | VLAN ID.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDVLANID | Invalid VLAN ID.
@rvalue TBLDRV_EVLANALREADYEXISTS | Vlan already exists.
@comm
If a vlan is successfully added, the switch will change to layer-2 or high layer switch.
*/
int32 rtl8651_addVlan(uint16 vid) {
	int32 retval;
	/* vid should be legal vlan ID */
	UNLOCK_ERROR_CHECK((vid < 1 || vid > 4094), TBLDRV_EINVALIDVLANID);
	rtlglue_drvMutexLock();//Lock resource
#ifdef RTL865XB_DCUT_SWVLAN
	/* Check if this VLAN has already been added into software VLAN table. */
	{
		rtl8651_tblDrv_swVlanTable_t *swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
		LOCK_ERROR_CHECK((swVlanPtr->valid == 1), TBLDRV_EVLANALREADYEXISTS);
	}
	
#endif
	retval = _rtl8651_addVlan(vid);
	rtlglue_drvMutexUnlock();//Unlock resource
	return retval;
}

/*
@func int32 | rtl8651_delVlan | Remove a VLAN.
@parm uint16 | vid | VLAN ID.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDVLANID | Invlaid VLAN ID.
@rvalue TBLDRV_ECANNOTREMOVEDEFVLAN | Default VLAN cannot be removed.
@rvalue TBLDRV_EVLANISREFERENCEDBYNETIF | Vlan is referenced by network interface.
@rvalue TBLDRV_EVLANISREFERENCEDBYPPPOE | Vlan is referenced by PPPoE.
@rvalue TBLDRV_EVLANHASMACALLOCATED | Vlan still has MAC address allocated.
@comm
<p rtl8651_delVlan()> removes a VLAN. A removed VLAN can not be referenced and can not has MAC address allocated.
*/
int32 rtl8651_delVlan(uint16 vid) 
{
	rtl8651_tblDrv_vlanTable_t *vlan_t;
	int32 retval;
	
	UNLOCK_ERROR_CHECK((vid < 1 || vid > 4094), TBLDRV_EINVALIDVLANID);
	/* System assumption, not permit vlan1 to be removed */
	UNLOCK_ERROR_CHECK((vid == 1), TBLDRV_ECANNOTREMOVEDEFVLAN);
	rtlglue_drvMutexLock();
	vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	if (vlan_t->fvlan_t != NULL) {
		retval = TBLDRV_EINVALIDVLANID;
		goto out;
	}
	retval = _rtl8651_delVlan( vid );
out:
	rtlglue_drvMutexUnlock();
	return retval;
}


#ifdef RTL865X_PSEUDO_VLAN

/* For SerCmom Spec. */
int32 rtl8651_addSubVlan(uint16 fvid, uint16 vid, uint32 portMask)
{
	rtl8651_tblDrv_vlanTable_t *vlan_t, *fvlan_t;
	int32 retval, port, vidx;
	
	if (vid<1 || vid>4094)
		return TBLDRV_EINVALIDVLANID;
	rtlglue_drvMutexLock();
	/* Forbid wan port to be set */
	for(vlan_t=DrvTbl.vlan, vidx=0; vidx<RTL8651_VLAN_NUMBER; vidx++, vlan_t++) {
		if (vlan_t->valid != 1 || vlan_t->internal == 1)
			continue;
		/* External Vlan */
		if (vlan_t->memberPortMask & portMask) {
			retval = TBLDRV_EINVALIDVLANID;
			goto out;
		}
	}
	
	fvlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(fvid)];
	/* The specified fvid should exist */
	if (fvlan_t->valid == 0) {
		retval = TBLDRV_EENTRYNOTFOUND;
		goto out;
	}
	/* The specified fvid should be configured to layer3 or above */
	if (fvlan_t->netif_t == NULL) {
		retval = TBLDRV_EINVALIDVLANID;
		goto out;
	}
	if ((retval=_rtl8651_addVlan(vid)) == SUCCESS) {
		vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
		vlan_t->netif_t 		= fvlan_t->netif_t;
		vlan_t->fvlan_t			= fvlan_t;
		vlan_t->ipAttached		= 1;
		vlan_t->inAclStart		= fvlan_t->inAclStart;
		vlan_t->inAclEnd		= fvlan_t->inAclEnd;
		vlan_t->outAclStart		= fvlan_t->outAclStart;
		vlan_t->outAclEnd		= fvlan_t->outAclEnd;
		vlan_t->macAddr			= fvlan_t->macAddr;
		vlan_t->macAddrNumber	= fvlan_t->macAddrNumber;
		for(port=0; port<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; port++) {
			if (portMask & (1<<port)) {
				vlan_t->memberPortMask |= (1<<port);
				_rtl8651_setPvid(port, vid);
			}
		}
		vlan_t->untagPortMask	= vlan_t->memberPortMask;
		RTL8651_SETASICVLAN(vlan_t);
	}

out:
	rtlglue_drvMutexUnlock();
	return retval;
}

int32 rtl8651_delSubVlan(uint16 vid)
{
	rtl8651_tblDrv_vlanTable_t *vlan_t;
	int32 retval, port;
	
	if (vid<1 || vid>4094)
		return TBLDRV_EINVALIDVLANID;
	rtlglue_drvMutexLock();
	vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	if (vlan_t->valid == 0 || vlan_t->fvlan_t == NULL) {
		retval = TBLDRV_EENTRYNOTFOUND;
		goto out;
	}
	vlan_t->macAddrNumber = 0;
	if ((retval=_rtl8651_delVlan(vid)) == SUCCESS) {
		for(port=0; port<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; port++)
			if (vlan_t->memberPortMask & (1<<port))
				_rtl8651_setPvid(port, vlan_t->fvlan_t->vid);
	} else vlan_t->macAddrNumber = 1;
	
out:
	rtlglue_drvMutexUnlock();
	return retval;
}

#endif

/*   Parameter internal: 
  *         (1) TURE, if vlan table has at least one internal vlan, this API return one internal vlan index.
  *                        if vlan table does not have any internal vlan, this API return one external vlan index.
  *                        if vlan table is empty, this API return vlan index 0.
  *         (2) FALSE, if vlan table has at least one external vlan, this API return one external vlan index.
  *                         if vlan table does not have any external vlan, this API return one internal vlan index.
  *                         if vlan table is empty, this API return index 0.
  */
uint16 _rtl8651_getOneVidx(int8 internal)
{
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	uint16 bakup_vidx = 0;
	uint32 i;

	for(i = 0, vlanPtr = &DrvTbl.vlan[0];  i < RTL8651_VLAN_NUMBER; i++, vlanPtr++) {
		if (vlanPtr->valid == 1) {
			if (vlanPtr->internal == internal)
				return rtl8651_vlanTableIndex(vlanPtr->vid);
			else
				bakup_vidx = rtl8651_vlanTableIndex(vlanPtr->vid);
		}
	}
	return bakup_vidx;
}

int8 _rtl8651_isInternal(uint16 vid) {
	uint16 vidx = rtl8651_vlanTableIndex(vid);

	if (DrvTbl.vlan[vidx].internal== 1)
		return TRUE;
	return FALSE;
}


/*
@func int32 | rtl8651_specifyVlanFilterDatabase | Specify a Filter database the VLAN uses.
@parm uint16 | vid | VLAN ID.
@parm uint16 | fid | Filter Database ID.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDVLANID | Invlaid VLAN ID.
@rvalue TBLDRV_EINVALIDFID | Invalid Filter database ID.
@rvalue TBLDRV_VLANHASHPORTMEMBER | Vlan can not has member port while specifying its Filter database.
@comm
<p rtl8651_specifyVlanFilterDatabase()> specifies which Filter Database the VLAN should use. Note that 
if the specified VLAN has member port, this API will retuan fail.
*/
int32 rtl8651_specifyVlanFilterDatabase(uint16 vid, uint16 fid) {
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
#ifdef RTL865XB_DCUT_SWVLAN
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;
#endif
	
	UNLOCK_ERROR_CHECK((vid == 0 || vid > 4094), TBLDRV_EINVALIDVLANID);
	UNLOCK_ERROR_CHECK((fid >= RTL8651_FDB_NUMBER), TBLDRV_EINVALIDFID);
	rtlglue_drvMutexLock();//Lock resource
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];

#ifdef RTL865XB_DCUT_SWVLAN
	swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
	LOCK_ERROR_CHECK((DrvTbl.filterDB[fid].valid == 0), TBLDRV_EINVALIDFID);

	if (vlanPtr->valid == 1 && vlanPtr->vid == vid) {
		LOCK_ERROR_CHECK((vlanPtr->memberPortMask), TBLDRV_EVLANHASHPORTMEMBER);
		vlanPtr->fid = fid;
	}
	else if (swVlanPtr->valid == 1) {
		LOCK_ERROR_CHECK((swVlanPtr->memberPortMask), TBLDRV_EVLANHASHPORTMEMBER);
		swVlanPtr->fid = fid;
	}
	else {
		rtlglue_drvMutexUnlock();
		return TBLDRV_EINVALIDVLANID;
	}
#else
	//Check whether vlan do exist
	LOCK_ERROR_CHECK((vlanPtr->valid == 0 || vlanPtr->vid != vid), TBLDRV_EINVALIDVLANID);
	LOCK_ERROR_CHECK((vlanPtr->memberPortMask), TBLDRV_EVLANHASHPORTMEMBER);
	LOCK_ERROR_CHECK((DrvTbl.filterDB[fid].valid == 0), TBLDRV_EINVALIDFID);
	vlanPtr->fid = fid;
#endif
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

/*
@func int32 | rtl8651_setVlanPromiscuous | Set the specified VLAN to operate in promiscuous mode.
@parm uint16 | vid | VLAN ID.
@parm int8 | isPromiscuous | TRUE to enable promiscuous mode, FALSE to disable promiscuous mode.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDVLANID | Invalid VLAN ID.
@rvalue TBLDRV_VLANTXMIRRORSET | Vlan has Tx mirror set hence can not set to promiscuous mode.
@rvalue TBLDRV_EVLANISREFERENCEDBYPPPOE | See comment.
@rvalue TBLDRV_EVLANISREFERENCEDBYNETIF | See comment.
@rvalue TBLDRV_EFWDTXMIRRORSET | VLAN Forwarding Tx Mirror was set. Forbid to enable promiscuous mode.
@comm
<p rtl8651_setVlanPromiscuous()> sets VLAN to operate in promiscuous mode. If the VLAN has Tx mirror set, the VLAN is disallowed to operate in promiscuous mode.
Note that Currently if the specified VLAN has network interface attached or PPPoE referenced, the VLAN can not change to promiscuous mode.
*/
int32 rtl8651_setVlanPromiscuous(uint16 vid, int8 isPromiscuous) {
	rtl8651_tblDrv_networkIntfTable_t *netIfPtr;
	rtl8651_tblDrv_pppTable_t *pppoePtr;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	uint32 i;

	UNLOCK_ERROR_CHECK((vid == 0 || vid > 4094), TBLDRV_EINVALIDVLANID);
	rtlglue_drvMutexLock();//Lock resource
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	LOCK_ERROR_CHECK((vlanPtr->valid == 0 || vlanPtr->vid != vid), TBLDRV_EINVALIDVLANID);
	LOCK_ERROR_CHECK((vlanPtr->promiscuous == (isPromiscuous==TRUE? 1: 0)), SUCCESS);
	if(isPromiscuous == TRUE) {//Checking network interface and pppoe
		if(rtl8651_getAsicOperationLayer()>2){
			netIfPtr = &DrvTbl.netIface[0];
			for(i=0; i< rtl8651_tblDrvPara.networkIntfTableSize; i++,netIfPtr++) {
				LOCK_ERROR_CHECK(
					( (netIfPtr->valid == 1) && (netIfPtr->vid == vid) && 
					  (netIfPtr->linkLayerType == RTL8651_LL_VLAN)
					), TBLDRV_EVLANISREFERENCEDBYNETIF
				);
				}
			for(i=0, pppoePtr=&DrvTbl.pppoe[0]; i<RTL8651_PPPOE_NUMBER; i++, pppoePtr++) {
				LOCK_ERROR_CHECK((pppoePtr->valid == 1 && pppoePtr->vid == vid), 
					TBLDRV_EVLANISREFERENCEDBYPPPOE
				);
				}
		}
		LOCK_ERROR_CHECK((vlanPtr->fwdTxMirror == 1), TBLDRV_EFWDTXMIRRORSET);
	}
	vlanPtr->promiscuous = isPromiscuous==TRUE? 1: 0;
	RTL8651_SETASICVLAN(vlanPtr);
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

/*
@func int32 | rtl8651_getVlanActiveStatus | Check if VLAN is active or inactive.
@parm uint16 | vid | VLAN ID.
@parm int8 * | isActive | TRUE for active and FALSE for inactive.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDVLANID | Invlaid VLAN ID.
@rvalue TBLDRV_EINVALIDINPUT | Invalid Input paramater.
@comm
If one of all specified VLAN member port is active, <p rtl8651_getVlanActiveStatus()> API regards the VLAN as ACTIVE status.
Contrarily, if non of all specified VLAN member port is active, the VLAN is treated as INACTIVE.
*/
int32 rtl8651_getVlanActiveStatus(uint16 vid, int8 * isActive) {
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
#ifdef RTL865XB_DCUT_SWVLAN
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;
#endif
	
	UNLOCK_ERROR_CHECK((vid == 0 || vid > 4094) , TBLDRV_EINVALIDVLANID);
	UNLOCK_ERROR_CHECK((isActive==NULL), TBLDRV_EINVALIDINPUT);
	rtlglue_drvMutexLock();//Lock resource
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
#ifdef RTL865XB_DCUT_SWVLAN
	swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
	if (vlanPtr->valid == 1 && vlanPtr->vid == vid)
		*isActive = vlanPtr->memberPortUpStatus != 0 ? TRUE: FALSE;
	else if (swVlanPtr->valid == 1)
		*isActive = swVlanPtr->memberPortUpStatus != 0 ? TRUE: FALSE;
	else {
		rtlglue_drvMutexUnlock();
		return TBLDRV_EINVALIDVLANID;
	}
#else
	LOCK_ERROR_CHECK((vlanPtr->valid == 0 || vlanPtr->vid != vid), TBLDRV_EINVALIDVLANID);
	*isActive = vlanPtr->memberPortUpStatus!= 0? TRUE: FALSE;
#endif
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

/*
@func int32 | rtl8651_addVlanPortMember | Add Vlan Port member.
@parm uint16 | vid | VLAN ID.
@parm uint32 | port | Logical port number(Trunking port or aggregating port).
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDVLANID | Invalid VLAN ID.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@rvalue TBLDRV_EAGGREGATORHASNOPORT | The logical port doesn't map to a physical port.
*/
int32 rtl8651_addVlanPortMember(uint16 vid, uint32 port) {
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
#ifdef RTL865XB_DCUT_SWVLAN
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;
#endif

	//Impossible VID or port number checking
	UNLOCK_ERROR_CHECK((vid == 0 || vid > 4094), TBLDRV_EINVALIDVLANID);
	UNLOCK_ERROR_CHECK((port >= RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
	rtlglue_drvMutexLock();//Lock resource
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	aggrePtr = &DrvTbl.aggregator[port];

#ifdef RTL865XB_DCUT_SWVLAN
	swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
	LOCK_ERROR_CHECK((DrvTbl.aggregator[port].ethernetPortMask == 0), TBLDRV_EAGGREGATORHASNOPORT);
	
	if (vlanPtr->valid == 1 && vlanPtr->vid == vid) {
		/* Aggregator has alread belong to VLAN */
		LOCK_ERROR_CHECK((vlanPtr->memberPortMask & aggrePtr->ethernetPortMask), SUCCESS);
		//Put port into corresponding VLAN port member set
		vlanPtr->memberPortMask |= aggrePtr->ethernetPortMask;
		vlanPtr->untagPortMask |= aggrePtr->ethernetPortMask;//Default add  port to untag list
		vlanPtr->memberPortUpStatus |= aggrePtr->ethernetPortUpStatus;
		if(vlanPtr->fwdTxMirror == 1)
			rtl8651_tblDrvCpuMirrorPortMask |= aggrePtr->ethernetPortMask;
		//Release all allocated 
		RTL8651_SETASICVLAN(vlanPtr);
	}
	else if (swVlanPtr->valid == 1) {
		LOCK_ERROR_CHECK((swVlanPtr->memberPortMask & aggrePtr->ethernetPortMask), SUCCESS);
		swVlanPtr->memberPortMask |= aggrePtr->ethernetPortMask;
		swVlanPtr->untagPortMask |= aggrePtr->ethernetPortMask;	// By default, we add the port to untag list
		swVlanPtr->memberPortUpStatus |= aggrePtr->ethernetPortUpStatus;
	}
	else {
		rtlglue_drvMutexUnlock();
		return TBLDRV_EINVALIDVLANID;
	}
#else
	LOCK_ERROR_CHECK((vlanPtr->valid == 0 || vlanPtr->vid != vid), TBLDRV_EINVALIDVLANID);
	//Check whether link aggregator valid
	LOCK_ERROR_CHECK((DrvTbl.aggregator[port].ethernetPortMask == 0), TBLDRV_EAGGREGATORHASNOPORT);
	/* Aggregator has alread belong to VLAN */
	LOCK_ERROR_CHECK((vlanPtr->memberPortMask & aggrePtr->ethernetPortMask), SUCCESS);
	//Put port into corresponding VLAN port member set
	vlanPtr->memberPortMask |= aggrePtr->ethernetPortMask;
	vlanPtr->untagPortMask |= aggrePtr->ethernetPortMask;//Default add  port to untag list
	vlanPtr->memberPortUpStatus |= aggrePtr->ethernetPortUpStatus;
	if(vlanPtr->fwdTxMirror == 1)
		rtl8651_tblDrvCpuMirrorPortMask |= aggrePtr->ethernetPortMask;
	//Release all allocated 
	RTL8651_SETASICVLAN(vlanPtr);
#endif

#ifdef CONFIG_RTL865X_MULTILAYER_BSP
	/* rearrange multicast port status */
	_rtl8651_arrangeMulticastPortStatus();
#endif	
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

/*
@func int32 | rtl8651_setVlanPortUntag | Set VLAN member ports to tag or untag.
@parm uint16 | vid | VLAN ID.
@parm uint32 | port | Logical port number. If physical port and logical port are one-to-one mapping, then physical port is the logical port.
@parm int8 | untag | TRUE to set member port untagging and FALSE to set tagging.
@rvalue TBLDRV_EINVALIDVLANID | Invalid VLAN ID.
@rvalue TBLDRV_EINVALIDPORT | Invalie port number.
@rvalue TBLDRV_EINVALIDPORT | The specified port is not a member port of specified VLAN.
@comm
If a port is set to a member port of a VLAN, the port is set to untag by default. The <p rtl8651_setVlanPortUntag()> can be used to set
a VLAN's member port to tag.
*/
int32 rtl8651_setVlanPortUntag(uint16 vid, uint32 port, int8 untag) {
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
#ifdef RTL865XB_DCUT_SWVLAN
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;
#endif

	//Impossible VID or port number checking
	UNLOCK_ERROR_CHECK((vid == 0 || vid > 4094), TBLDRV_EINVALIDVLANID);
	UNLOCK_ERROR_CHECK((port >= RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
	rtlglue_drvMutexLock();//Lock resource
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	aggrePtr = &DrvTbl.aggregator[port];
	
#ifdef RTL865XB_DCUT_SWVLAN
	swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];

	if (vlanPtr->valid == 1 && vlanPtr->vid == vid) {
		LOCK_ERROR_CHECK(((vlanPtr->memberPortMask & aggrePtr->ethernetPortMask) == 0), 
			TBLDRV_EINVALIDPORT
		);
		if(untag == TRUE)
			vlanPtr->untagPortMask |= aggrePtr->ethernetPortMask;
		else vlanPtr->untagPortMask &= ~aggrePtr->ethernetPortMask;
		RTL8651_SETASICVLAN(vlanPtr);
	}
	else if (swVlanPtr->valid == 1) {
		LOCK_ERROR_CHECK(((swVlanPtr->memberPortMask & aggrePtr->ethernetPortMask) == 0), 
			TBLDRV_EINVALIDPORT
		);		
		if(untag == TRUE)
			swVlanPtr->untagPortMask |= aggrePtr->ethernetPortMask;
		else swVlanPtr->untagPortMask &= ~aggrePtr->ethernetPortMask;
	}
	else {
		rtlglue_drvMutexUnlock();
		return TBLDRV_EINVALIDVLANID;
	}
#else
	LOCK_ERROR_CHECK(((vlanPtr->memberPortMask & aggrePtr->ethernetPortMask) == 0), 
		TBLDRV_EINVALIDPORT
	);
	if(untag == TRUE)
		vlanPtr->untagPortMask |= aggrePtr->ethernetPortMask;
	else vlanPtr->untagPortMask &= ~aggrePtr->ethernetPortMask;
	RTL8651_SETASICVLAN(vlanPtr);
#endif

	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}



int32 _rtl8651_assignVlanMacAddress(uint16 vid, ether_addr_t * macAddress, uint16 macAddrNumber) 
{
	rtl8651_tblDrv_vlanTable_t *vlan_t;
	int32 retval = SUCCESS;

	if (vid < 1 || vid > 4094)
		return TBLDRV_EINVALIDVLANID;
	if (macAddrNumber > 1 && (macAddress->octet[5]&(macAddrNumber-1)))
		return TBLDRV_EINVALIDINPUT;

	vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	if (vlan_t->valid == 0 || vlan_t->vid != vid) {
		retval = TBLDRV_EINVALIDVLANID;
		goto out;
	}
	//Set new VLAN MAC address
	vlan_t->macAddr 			= *macAddress;
	vlan_t->macAddrNumber 	= macAddrNumber;
	vlan_t->manualAssign 		= 1;
	RTL8651_SETASICVLAN(vlan_t);
out:
	return retval;
}

/*
@func int32 | rtl8651_assignVlanMacAddress | Assign VLAN MAC address instead of automatic allocation.
@parm uint16 | vid | VLAN ID.
@parm ether_addr_t * | macAddress | Pointer to MAC address.
@parm uint16 | macAddrNumber | The number of MAC addresses assigned to a VLAN.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDVLANID | Invalid VLAN ID.
@rvalue TBLDRV_EINVALIDINPUT | The logical AND of <p macAddrNumber> and the last type of MAC address should be zero.
@rvalue TBLDRV_EVLANMACREFERENCEDBYPPPOE | VLAN MAC is referenced by PPPoE.
@rvalue TBLDRV_EVLANMACREFERENCEDBYNETIF | VLAN MAC is referenced by network interface.
@rvalue TBLDRV_EINVVLANMAC | Specified VLAN MAC number should be (0, 1, 2, 4, 8).
*/
int32 rtl8651_assignVlanMacAddress(uint16 vid, ether_addr_t * macAddress, uint16 macAddrNumber) {
#if 1

	int32 retval = SUCCESS;

	if (vid < 1 || vid > 4094)
		return TBLDRV_EINVALIDVLANID;
	if (macAddrNumber > 1 && (macAddress->octet[5]&(macAddrNumber-1)))
		return TBLDRV_EINVALIDINPUT;

	rtlglue_drvMutexLock();
	retval = _rtl8651_assignVlanMacAddress( vid, macAddress, macAddrNumber );
	rtlglue_drvMutexUnlock();
	return retval;
	
#else
	rtl8651_tblDrv_networkIntfTable_t *netIfPtr;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	rtl8651_tblDrv_pppoeTable_t *pppoePtr;
	uint32 i;
	
	UNLOCK_ERROR_CHECK((vid < 1 || vid > 4094), TBLDRV_EINVALIDVLANID);
	if (macAddrNumber>1) /* add by rupert*/
	UNLOCK_ERROR_CHECK((macAddress->octet[5] & (macAddrNumber-1)), TBLDRV_EINVALIDINPUT);
	rtlglue_drvMutexLock();//Lock resource
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	LOCK_ERROR_CHECK((vlanPtr->valid == 0 || vlanPtr->vid != vid), TBLDRV_EINVALIDVLANID);
	if(rtl8651_getAsicOperationLayer()>2){
		for(i=0, pppoePtr=&DrvTbl.pppoe[0]; i<RTL8651_PPPOE_NUMBER; i++, pppoePtr++) {
			LOCK_ERROR_CHECK(
				(pppoePtr->valid == 1 && pppoePtr->vid == vid), 
				TBLDRV_EVLANMACREFERENCEDBYPPPOE
			);
		}
		netIfPtr = &DrvTbl.netIface[0];
		for(i=0;i< rtl8651_tblDrvPara.networkIntfTableSize; i++,netIfPtr++) {
			LOCK_ERROR_CHECK(
				( (netIfPtr->valid == 1) && (netIfPtr->vid == vid) &&
				  (netIfPtr->linkLayerType == RTL8651_LL_VLAN)
				), TBLDRV_EVLANMACREFERENCEDBYNETIF
			);
		}
	}
	//Allocate new mac address
	vlanPtr->macAddr = *macAddress;
	vlanPtr->macAddrNumber = macAddrNumber;
	vlanPtr->manualAssign = 1;
	RTL8651_SETASICVLAN(vlanPtr);
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
#endif
}

		
/*
@func int32 | rtl8651_getVlanMacAddress | Get VLAN MAC address.
@parm uint16 | vid | VLAN ID.
@parm ether_addr_t * | macAddr | VLAN MAC address.
@parm uint16 | macAddrNumber | Specify the number of MACs in the specified VLAN.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDVLANID | Invalid VLAN ID.
@rvalue TBLDRV_EVLANNOMAC | Vlan has no MAC Address.
*/
int32 rtl8651_getVlanMacAddress(uint16 vid, ether_addr_t * macAddr, uint16 * macAddrNumber) {
	rtl8651_tblDrv_vlanTable_t *vlanPtr; 
	
	UNLOCK_ERROR_CHECK((vid < 1 || vid > 4094), TBLDRV_EINVALIDVLANID);
	rtlglue_drvMutexLock();//Lock resource
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	LOCK_ERROR_CHECK((vlanPtr->valid == 0 || vlanPtr->vid != vid ), TBLDRV_EINVALIDVLANID);
	//Release previous allocated mac address
	if(vlanPtr->macAddrNumber > 0) {
		*macAddr = vlanPtr->macAddr;
		if(macAddrNumber)
			*macAddrNumber = vlanPtr->macAddrNumber;
		rtlglue_drvMutexUnlock();//Unlock resource
		return SUCCESS;
	}
	rtlglue_drvMutexUnlock();//Unlock resource
	return(TBLDRV_EVLANNOMAC);
}

/*
@func int32 | rtl8651_delVlanPortMember | Remove a member port from a VLAN.
@parm uint16 | vid | VLAN ID.
@parm uint32 | port | Logical port.
@rvalue SUCCESS | A member port was removed successfully.
@rvalue TBLDRV_EINVALIDVLANID | Invalid VLAN ID.
@rvalue TBLDRV_EINVALIDPORT | Invalid port number.
@rvalue TBLDRV_EAGGREGATORHASNOPORT | Not allow aggregator without port member .
@rvalue TBLDRV_ENOTVLANPORTMEMBER | Specified port is not a member port of specified vlan.
*/
int32 rtl8651_delVlanPortMember(uint16 vid, uint32 port) {
	rtl8651_tblDrv_linkAggregation_t *aggrePtr;
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
#ifdef RTL865XB_DCUT_SWVLAN
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr;
#endif

	//Impossible VID or port number checking
	UNLOCK_ERROR_CHECK((vid == 0 || vid > 4094), TBLDRV_EINVALIDVLANID);
	UNLOCK_ERROR_CHECK((port >= RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum), TBLDRV_EINVALIDPORT);
	rtlglue_drvMutexLock();//Lock resource
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
	aggrePtr = &DrvTbl.aggregator[port];

#ifdef RTL865XB_DCUT_SWVLAN
	swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
	//Check whether link aggregator valid
	LOCK_ERROR_CHECK((aggrePtr->ethernetPortMask == 0), TBLDRV_EAGGREGATORHASNOPORT);

	if (vlanPtr->valid == 1 && vlanPtr->vid == vid) {
		//Check whether port in corresponding VLAN port member set
		LOCK_ERROR_CHECK(
			((vlanPtr->memberPortMask & aggrePtr->ethernetPortMask) == 0), 
			TBLDRV_ENOTVLANPORTMEMBER
		);
		//Remove port from member set
		vlanPtr->memberPortMask &=  ~aggrePtr->ethernetPortMask;
		vlanPtr->untagPortMask &=  ~aggrePtr->ethernetPortMask;
		vlanPtr->memberPortUpStatus &= ~aggrePtr->ethernetPortUpStatus;
		if(vlanPtr->fwdTxMirror == 1)
			rtl8651_tblDrvCpuMirrorPortMask &= ~aggrePtr->ethernetPortMask;
		RTL8651_SETASICVLAN(vlanPtr);
	}
	else if (swVlanPtr->valid == 1) {
		//Check whether port in corresponding VLAN port member set
		LOCK_ERROR_CHECK(
			((swVlanPtr->memberPortMask & aggrePtr->ethernetPortMask) == 0), 
			TBLDRV_ENOTVLANPORTMEMBER
		);
		//Remove port from member set
		swVlanPtr->memberPortMask &=  ~aggrePtr->ethernetPortMask;
		swVlanPtr->untagPortMask &=  ~aggrePtr->ethernetPortMask;
		swVlanPtr->memberPortUpStatus &= ~aggrePtr->ethernetPortUpStatus;
	}
	else {
		rtlglue_drvMutexUnlock();
		return TBLDRV_EINVALIDVLANID;
	}
	
#else
	LOCK_ERROR_CHECK((vlanPtr->valid == 0 || vlanPtr->vid != vid), TBLDRV_EINVALIDVLANID);
	//Check whether link aggregator valid
	LOCK_ERROR_CHECK((aggrePtr->ethernetPortMask == 0), TBLDRV_EAGGREGATORHASNOPORT);
	//Check whether port in corresponding VLAN port member set
	LOCK_ERROR_CHECK(
		((vlanPtr->memberPortMask & aggrePtr->ethernetPortMask) == 0), 
		TBLDRV_ENOTVLANPORTMEMBER
	);
	//Remove port from member set
	vlanPtr->memberPortMask &=  ~aggrePtr->ethernetPortMask;
	vlanPtr->untagPortMask &=  ~aggrePtr->ethernetPortMask;
	vlanPtr->memberPortUpStatus &= ~aggrePtr->ethernetPortUpStatus;
	if(vlanPtr->fwdTxMirror == 1)
		rtl8651_tblDrvCpuMirrorPortMask &= ~aggrePtr->ethernetPortMask;
	RTL8651_SETASICVLAN(vlanPtr);
#endif

#ifdef CONFIG_RTL865X_MULTILAYER_BSP
	/* rearrange multicast port status */
	_rtl8651_arrangeMulticastPortStatus();
#endif	
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}



#define RTL8651_MISC_APIs
//peripheralExtPortMask is the portmask of loopback(in 8651) or ext port(in 8650B)
//which already has a USB/PCI device registered.
uint32 peripheralExtPortMask=0;
int32 rtl8651_addPeripheralExtPort(uint32 port){
	if(port>=RTL8651_AGGREGATOR_NUMBER){
		rtlglue_printf("Reserve port %d too large (Max:%d)for peripheral device use failed.\n",port, RTL8651_AGGREGATOR_NUMBER);
		return FAILED;
	}
	peripheralExtPortMask|=1<<port;
	rtlglue_printf("Reserve port %d for peripheral device use. (0x%x)\n",port, peripheralExtPortMask);

	return SUCCESS;
}

int32 rtl8651_delPeripheralExtPort(uint32 port){
	if(port>=RTL8651_AGGREGATOR_NUMBER)
		return FAILED;
	peripheralExtPortMask&=~(1<<port);
	rtlglue_printf("Delete port %d from peripheral port set. (0x%x)\n",port, peripheralExtPortMask);
	return SUCCESS;
}


/*=======================================================================================
 * RTL8651 MISC. APIs:
 *	1. Static Functions
 *
 *	2. Internal APIs
 *
 *	3. External APIs
 *
 *=======================================================================================*/

int32 rtl8651_setLoopbackPortPHY(uint32 phy, int32 flag){
	//phy number can't exceed RTL8651's 6 ports (from 0~5) and can't change extension port phy
	//when current setting phy is still active
	if(phy>RTL8651_PHY_NUMBER)
		return FAILED;
	if(phy==RTL8651_MII_PORTNUMBER){
		//enable MII port
		if(rtl8651_setAsicEthernetMII(8, P5_LINK_MII_PHY,flag))
			return FAILED;
	}else{
		//use ports other than MII port as our loopback port
		if(rtl8651_setAsicEthernetPHYLoopback(phy, flag))
			return FAILED;
	}
	//port state of loopback port is always set by software.
	rtl8651_setEthernetPortLinkStatus(phy, flag);
	return SUCCESS;
}

#define RTL8651_PROTOCOL_TRAP_APIs
/*=======================================================================================
 * RTL8651 Protocol Trap APIs:
 *	1. Internal APIs
 *		. _rtl8651_queryProtoTrap()
 *	2. External APIs
 *		. rtl8651_addProtoTrap()
 *		. rtl8651_delProtoTrap()
 *=======================================================================================*/
#define ILLEGAL_PROTOTRAP_TYPE(type)\
	((type)!=RTL8651_PROTOTRAP_ETHER && (type)!=RTL8651_PROTOTRAP_IP &&\
	 (type)!=RTL8651_PROTOTRAP_TCPPORT && (type)!=RTL8651_PROTOTRAP_UDPPORT)


/*
@func int32  | _rtl8651_queryProtoTrap | Internal use only. Query Protocl-Trap Table.
@parm uint8 | protoType | Protocol Type incluses: RTL8651_PROTOTRAP_ETHER, RTL8651_PROTOTRAP_IP, RTL8651_PROTOTRAP_TCPPORT, RTL8651_PROTOTRAP_UDPPORT.
@parm uint16 | protoContent | The protoContent is the comparison content and is defined by protoType. RTL8651_PROTOTRAP_ETHER: protoContent means EtherType. RTL8651_PROTOTRAP_IP: protoContent means IP protocol. RTL8651_PROTOTRAP_TCPPORT: protoContent means TCP Port number. RTL8651_PROTOTRAP_UDPPORT: protoContent means UDP Port number.
@rvalue Negative value | Protocol trap entry not found
@rvalue Non-negative value | Protocol trap entry number found
@comm
Given <p protoType> and <p protoContent>, the internal API will decide if the trap exists(and also enabled) in the ASIC or not.
*/
int32 _rtl8651_queryProtoTrap(uint8 protoType, uint16 protoContent) {
#if 0
	rtl8651_tblDrvprotoTrapAsicArrangementEnty_t *ptPtr;
	int8 isEnable;
	int32 index,asicPosition ;
	
	if (ILLEGAL_PROTOTRAP_TYPE(protoType))
		return FAILED;
	asicPosition=rtl8651_getHardwiredProtoTrap(protoType, protoContent, &isEnable);
	if ( asicPosition>=0 && isEnable==TRUE)
		return asicPosition;
	for(index=0,ptPtr=&AsicTableUsage.protoTrap[0]; index<RTL8651_PROTOCOLTRAP_SIZE; index++,ptPtr++)
		if ((ptPtr->valid==1)&&(protoType==(uint8)ptPtr->trapType)&&(protoContent==ptPtr->trapContent))
			return RTL8651_HARDWIRED_PROTOTRAP_SIZE+index;
#endif
	return FAILED;	
}

/*
@func int32 | rtl8651_addProtoTrap | Add a trap to trap user-interest packets to CPU.
@parm uint8 | protoType | Protocol Type incluses: RTL8651_PROTOTRAP_ETHER, RTL8651_PROTOTRAP_IP, RTL8651_PROTOTRAP_TCPPORT, RTL8651_PROTOTRAP_UDPPORT.
@parm uint16 | protoContent | The protoContent is the comparison content and is defined by protoType. RTL8651_PROTOTRAP_ETHER: protoContent means EtherType. RTL8651_PROTOTRAP_IP: protoContent means IP protocol. RTL8651_PROTOTRAP_TCPPORT: protoContent means TCP Port number. RTL8651_PROTOTRAP_UDPPORT: protoContent means UDP Port number.
@rvalue SUCCESS | The user specified trap was successfully added into the ASIC.
@rvalue TBLDRV_EUNKNOWPROTOTYPE | Unknown Protocol type.
@rvalue TBLDRV_EENTRYALREADYEXIST | The trap user specified is already in the ASIC.
@rvalue TBLDRV_ENOFREEBUFFER | Out of memory buffer.
@comm
In normal case, ASIC always processes each packet as possible as it can. For such packets, they won't be awared by CPU/Protocol Stack. 
Hence for Protocl-Stack-interest packets, the protocol stack can add a trap to trap packets to Protocol Stack. 
*/
int32 rtl8651_addProtoTrap(uint8 protoType, uint16 protoContent) {
#if 0
	rtl8651_tblDrvprotoTrapAsicArrangementEnty_t *ptPtr, *addPtr;
	uint32 protoIndex = RTL8651_PROTOCOLTRAP_SIZE, i = 0;

	UNLOCK_ERROR_CHECK(ILLEGAL_PROTOTRAP_TYPE(protoType), TBLDRV_EUNKNOWPROTOTYPE);
#ifdef CONFIG_RTL865X_WLAN_ACCELERATION	
	return FAILED;
#endif
	rtlglue_drvMutexLock();//Lock resource
	if (rtl8651_turnOnHardwiredProtoTrap(protoType, protoContent) == SUCCESS) {
		rtlglue_drvMutexUnlock();
		return SUCCESS;
	}	
	for(ptPtr=&AsicTableUsage.protoTrap[0]; i<RTL8651_PROTOCOLTRAP_SIZE; i++, ptPtr++) {
		/* Duplicate checking */
		LOCK_ERROR_CHECK(
			( (ptPtr->valid==1) && (protoType==(uint8)ptPtr->trapType) && 
			  (protoContent==ptPtr->trapContent)
			), TBLDRV_EENTRYALREADYEXIST
		);
		if (ptPtr->valid == 0 && protoIndex == RTL8651_PROTOCOLTRAP_SIZE) {
			protoIndex = i;
			addPtr = ptPtr;
		}
	}
	LOCK_ERROR_CHECK((protoIndex == RTL8651_PROTOCOLTRAP_SIZE), TBLDRV_ENOFREEBUFFER);
	addPtr->valid = 1;
	addPtr->trapType = protoType;
	addPtr->trapContent = protoContent;
	convert_setAsicProtoTrap(protoIndex, protoType, protoContent);
	rtlglue_drvMutexUnlock();//Unlock resource
#endif
	rtlglue_printf("%s is legacy now and deprecated\n", __FUNCTION__);
	return SUCCESS;
}


/*
@func int32 | rtl8651_delProtoTrap | Remove a trap from ASIC.
@parm uint8 | protoType | Protocol Type incluses: RTL8651_PROTOTRAP_ETHER, RTL8651_PROTOTRAP_IP, RTL8651_PROTOTRAP_TCPPORT, RTL8651_PROTOTRAP_UDPPORT.
@parm uint16 | protoContent | The protoContent is the comparison content and is defined by protoType. RTL8651_PROTOTRAP_ETHER: protoContent means EtherType. RTL8651_PROTOTRAP_IP: protoContent means IP protocol. RTL8651_PROTOTRAP_TCPPORT: protoContent means TCP Port number. RTL8651_PROTOTRAP_UDPPORT: protoContent means UDP Port number.
@rvalue SUCCESS | The user specified trap was successfully removed from ASIC.
@rvalue TBLDRV_EUNKNOWPROTOTYPE | Unknown Protocol type.
@rvalue TBLDRV_EENTRYNOTFOUND | User-specified trap is not found.
@comm
The API can remove a trap added by Protocol-Stack from ASIC.
*/
int32 rtl8651_delProtoTrap(uint8 protoType, uint16 protoContent) {
#if 0
	rtl8651_tblDrvprotoTrapAsicArrangementEnty_t *ptPtr;
	uint32 protoIndex;
	
 	UNLOCK_ERROR_CHECK((ILLEGAL_PROTOTRAP_TYPE(protoType)), TBLDRV_EUNKNOWPROTOTYPE);
	rtlglue_drvMutexLock();//Lock resource	
	LOCK_ERROR_CHECK((rtl8651_turnOffHardwiredProtoTrap(protoType, protoContent)==SUCCESS), SUCCESS);
 	ptPtr=&AsicTableUsage.protoTrap[0];
	for(protoIndex=0 ;protoIndex<RTL8651_PROTOCOLTRAP_SIZE; protoIndex++, ptPtr++) 
		if ((ptPtr->valid==1)&&(protoType==(uint8)ptPtr->trapType)&&(protoContent==ptPtr->trapContent))
			break;
	LOCK_ERROR_CHECK((protoIndex == RTL8651_PROTOCOLTRAP_SIZE), TBLDRV_EENTRYNOTFOUND);
	ptPtr->valid = 0;
	rtl8651_delAsicProtoTrap(protoIndex);
	rtlglue_drvMutexUnlock();//Unlock resource
#endif
	rtlglue_printf("%s is legacy now and deprecated\n", __FUNCTION__);
	return SUCCESS;
}



#define RTL8651_BROADCAST_STORM_API
/*
@func int32 | rtl8651_enableBroadCastStormPrevention | Enable/Disable Broadcast storm prevention.
@parm int8 | enable | TRUE to enable Broadcast storm prevention and FALSE to disable it.
@rvalue SUCCESS | Success.
@rvalue TBLDRV_EINVALIDINPUT | Invalid input parameter.
*/
int32 rtl8651_enableBroadCastStormPrevention(int8 enable) {
	UNLOCK_ERROR_CHECK((enable!=TRUE && enable!=FALSE), TBLDRV_EINVALIDINPUT);
	rtlglue_drvMutexLock();
	rtl8651_setBroadCastStormReg(enable);
	rtlglue_drvMutexUnlock();//Unlock resource
	return SUCCESS;
}

int32 rtl8651_layer2_alloc(void){
	int32 index;
	memset(&AsicTableUsage, 0, sizeof(struct AsicTableUsage_S)) ;
	//Allocate arrays	
	TBL_MEM_ALLOC(AsicTableUsage.mac, rtl8651_tblDrv_macAddressEntry_t, rtl8651_tblDrvPara.macAddressDbSize);

	memset( &DrvTbl, 0, sizeof( DrvTbl ) );
	TBL_MEM_ALLOC(DrvTbl.phyPort, rtl8651_tblDrv_ethernet_t, RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum);
	TBL_MEM_ALLOC(DrvTbl.aggregator, rtl8651_tblDrv_linkAggregation_t, RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum);
	TBL_MEM_ALLOC(DrvTbl.vlan, rtl8651_tblDrv_vlanTable_t, RTL8651_VLAN_NUMBER);
#ifdef RTL865XB_DCUT_SWVLAN
	TBL_MEM_ALLOC(DrvTbl.swVlan, rtl8651_tblDrv_swVlanTable_t, RTL8651_SWVLAN_NUMBER);
#endif
	TBL_MEM_ALLOC(DrvTbl.filterDB, rtl8651_tblDrv_filterDbTable_t, RTL8651_FDB_NUMBER);
  	TBL_MEM_ALLOC(DrvTbl.spanningTree, rtl8651_tblDrv_spanningTreeTable_t, RTL8651_FDB_NUMBER);

	//Allocate lists and put all entries to free list.

	
	{//Initial free filter database entry
		rtl8651_tblDrv_filterDbTableEntry_t * tempFilterDb;
		TBL_MEM_ALLOC(tempFilterDb, rtl8651_tblDrv_filterDbTableEntry_t, rtl8651_tblDrvPara.filterDbSize);
		SLIST_INIT(&DrvTbl.freeList.filterDBentry);
		for(index=0; index<rtl8651_tblDrvPara.filterDbSize; index++)
			SLIST_INSERT_HEAD(&DrvTbl.freeList.filterDBentry, &tempFilterDb[index], nextFDB);
	}

	return SUCCESS;
}

//Initialization all Layer2 tables and linked lists. For software reset purpose.
int32 rtl8651_layer2_init(void){
	int32 index, index1,i;

	/////////////Initialize L2 static variables.//////////////////

	memset(&rtl8651AsicEthernetTable[0], 0, (RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum)*sizeof(rtl8651_tblAsic_ethernet_t));
		
	//Enable link up/down interrupt
	linkChangePendingCount = 0;
	rtl8651_l2protoPassThrough=0;
	miiPhyAddress=-1;//set to -1 means not usable.
	rtl8651_tblDrvCpuMirrorPortMask=0;
	arpAgingTime = 30;//300;	/* default 5 mins. Init at L2 becoz it's also L2 entry's default timeout
	/* Physical port link status */
#ifndef RTL865X_TEST	
	phyPortLinkStatus = 0;
#else
	phyPortLinkStatus = 0xffffffff; //test code relies on this...
#endif

	/* Initially disable spanning tree control and all ports are assume as in forwarding state */
	rtl8651_tblDrvSpanningTreeEnable = 0;
	rtl8651_tblDrvSpanningTreePortFwdStatus = 0xffffffff;

	pureSoftFwd = TRUE; //by default, disable ASIC L3/4 acceleration at L2. 
	//If L3/4 is initializaed later, this variable would be changed to TRUE
	rtl8651_L2statusNotifier= NULL;
	rtl8651_linkStatusNotifier = NULL;	 //Notify L3 driver
#ifdef CONFIG_RTL865X_MULTILAYER_BSP
	rtl8651_externalMulticastPortMask = 0;	// reset all multicast port back to internal Mode, ASIC will be set at below code
#endif
	/////////////Initialize L2 allocated arrays.//////////////////

	 //Initial VLAN Table
	memset((int8 *)DrvTbl.vlan, 0, sizeof(rtl8651_tblDrv_vlanTable_t)*RTL8651_VLAN_NUMBER);

#ifdef RTL865XB_DCUT_SWVLAN
	// Init software vlan table
	memset((int8 *)DrvTbl.swVlan, 0, sizeof(rtl8651_tblDrv_swVlanTable_t) * RTL8651_SWVLAN_NUMBER);
#endif

	memset(&AsicTableUsage.mac[0],0,rtl8651_tblDrvPara.macAddressDbSize*sizeof(rtl8651_tblDrv_macAddressEntry_t));
	{ ////Get hardware stored (flash or ROM) MAC address
		rtl8651_tblDrv_macAddressEntry_t *asicMac;
		ether_addr_t macAddress;
		uint32 number;
		rtlglue_getMacAddress(&macAddress, &number);
		//assert((macAddress.octet[5] & 0x7) == 0);
		for(i=0,asicMac=&AsicTableUsage.mac[0]; i< number && i < rtl8651_tblDrvPara.macAddressDbSize; i++,asicMac++) {
			asicMac->mac = macAddress;
			asicMac->mac.octet[4] += (uint8) (i>>8);
			asicMac->mac.octet[5] += (uint8) (i&0xff);
			asicMac->allocated = 0;
			asicMac->valid = 1;
		}
	}



	{ //Invalidate all spanning tree table
		rtl8651_tblDrv_spanningTreeTable_t *spt_t=&DrvTbl.spanningTree[0];
		for(index=0; index<RTL8651_FDB_NUMBER; index++, spt_t++)
			spt_t->valid = 0;
	}	

	{ //Initial Phy Port and Aggregator
		rtl8651_tblDrv_ethernet_t *phyPtr=&DrvTbl.phyPort[0];
		rtl8651_tblDrv_linkAggregation_t *aggrePtr=&DrvTbl.aggregator[0];
		
		for(i=0,phyPtr=&DrvTbl.phyPort[0],aggrePtr=&DrvTbl.aggregator[0]; i<(RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum); i++,phyPtr++,aggrePtr++) {//Configure Ethernet port default values
			//Initial Ethernet ports
			phyPtr->speed = 1;	//100Mbps default
			phyPtr->duplex = 1; //full-duplex mode default
			phyPtr->autoNegotiation = 1; //Default auto-negotiation enabled
			phyPtr->fake_forceMode = 0; //Default fake_forceMode disabled
			phyPtr->autoAdvCapability = 0; //Default RTL8651_ETHER_AUTO_100FULL
			phyPtr->aggregatorIndex = i;
			phyPtr->linkUp = 0;			
			phyPtr->inputBandwidthControl = RTL8651_BC_FULL;
			phyPtr->outputBandwidthControl = RTL8651_BC_FULL;
			//Initial Link Aggregator
			aggrePtr->pvid = 1; //Default VLAN 1 IEEE 802.1q p.69
			aggrePtr->individual = 0; //Initial to be not-individual, but the total number of disabling individual property is limited by RTL8651_MAX_AGGREGATOR_NUM
			aggrePtr->aggregated = 0; //Initial not aggregated
			aggrePtr->ethernetPortMask = 1<<i;//Initially, the port and link aggragator are 1-to-1 mapping
			aggrePtr->ethernetPortUpStatus = 0;//Initially, the port state of the link aggregator is down
		}
	}
	
	{ //Initial Filter database
		rtl8651_tblDrv_filterDbTable_t *fdb_t=&DrvTbl.filterDB[0];
		for(index=0; index<RTL8651_FDB_NUMBER; index++) {
			for(index1=0; index1<RTL8651_L2TBL_ROW; index1++)
				SLIST_INIT(&(fdb_t->database[index1]));
			fdb_t->valid = 0;
		}
	}

	//ASIC Initialization
	for(i=0; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++) {
		rtl8651_setAsicMulticastPortInternal(i, TRUE);
		rtl8651_setAsicMulticastSpanningTreePortState(i, RTL8651_PORTSTA_FORWARDING);
		rtl8651_setAsicEthernetBandwidthControl(i, TRUE, RTL8651_BC_FULL);
		rtl8651_setAsicEthernetBandwidthControl(i, FALSE, RTL8651_BC_FULL);
	}

	/* chhuang: for phy, temporarily solution */
	{
                int8 autoNego, fullDuplex;
                uint32 speed, advCap, port;
		 rtl8651_tblDrv_ethernet_t *phyPortPtr;
	
		for(port=0; port<RTL8651_PORT_NUMBER; port++) {
		 	phyPortPtr = &DrvTbl.phyPort[port];
                	rtl8651_getAsicEthernetPHY(port, &autoNego, &advCap, &speed, &fullDuplex);
                	phyPortPtr->speed = (speed==1)? 1: 0;
                	phyPortPtr->duplex = (fullDuplex==TRUE)? 1: 0;
                	phyPortPtr->autoNegotiation = (autoNego==1)? 1: 0;
                	phyPortPtr->autoAdvCapability = advCap;
		}

        }

	/* These two global variable is used to patch IC bug in RTL865xB cut-b & before: switch core crash 
	    due to turning off in_queue flow control. */
	memset(fake_DisableFlowControl, 0, sizeof(uint32) * RTL8651_PHY_NUMBER);
	memset(prePortState, 0, sizeof(uint32) * RTL8651_PHY_NUMBER);

	return SUCCESS;	
}

int32 rtl8651_layer2_collect(void) {
	uint32 entry, index, index1;

	{ //Collect Filter Database Entries
		rtl8651_tblDrv_filterDbTableEntry_t * tempFilterDbPtr;
		rtl8651_tblDrv_filterDbTable_t *filterDbPtr;
		
		for(index=0,filterDbPtr=&DrvTbl.filterDB[0]; index<RTL8651_FDB_NUMBER; index++,filterDbPtr++) {
			for(index1=0; index1<RTL8651_L2TBL_ROW; index1++)
				while(SLIST_FIRST(&(filterDbPtr->database[index1]))) {
					tempFilterDbPtr = SLIST_FIRST(&(filterDbPtr->database[index1]));
					SLIST_REMOVE_HEAD(&(filterDbPtr->database[index1]), nextFDB);
					SLIST_INSERT_HEAD(&DrvTbl.freeList.filterDBentry, tempFilterDbPtr, nextFDB);
				}
		}
		entry = 0;
		SLIST_FOREACH(tempFilterDbPtr, &DrvTbl.freeList.filterDBentry, nextFDB) 
			entry++;
		assert(entry == rtl8651_tblDrvPara.filterDbSize);
	}

	
	return SUCCESS;
}


/*
@func int32 | rtl8651_tblDrvInit | Initiate Table Driver.
@parm rtl8651_tblDrvInitPara_t * | configPara | Initial parameters setting.
@rvalue SUCCESS | Table Driver Initialization was success.
@rvalue FAILED | Table DRiver Initialization failed.
@comm
This API should be called to initiate Table Driver when in the RTL8651 initialization state.
*/
int32 rtl8651_initDrvParam(rtl8651_tblDrvInitPara_t * configPara) {
	if(configPara == NULL) {
		rtl8651_tblDrvPara.macAddressDbSize = 8;
		rtl8651_tblDrvPara.filterDbSize = 1024;
		rtl8651_tblDrvPara.networkIntfTableSize = 8;
		rtl8651_tblDrvPara.ipIntfEntryNumber = 16;
		rtl8651_tblDrvPara.ipIntfIpAddrNumber = 24;
		rtl8651_tblDrvPara.arpTableSize = 512;
		rtl8651_tblDrvPara.routingTableSize = 32;
		rtl8651_tblDrvPara.natSize = 8;
		rtl8651_tblDrvPara.naptServerPortEntryNumber = 64;
		rtl8651_tblDrvPara.naptTcpUdpFlowSize = 4096;
		
		rtl8651_tblDrvPara.naptIcmpFlowSize = 128;
		rtl8651_tblDrvPara.aclSize = 256;
		rtl8651_tblDrvPara.algSize = 128;
		rtl8651_tblDrvPara.urlfilterDBSize=16; //max number of filter string
		rtl8651_tblDrvPara.urlfilterStringLen=32; //max length of each filter string
		rtl8651_tblDrvPara.pathfilterStringLen=32; //max length of each path string of url
		rtl8651_tblDrvPara.urlPktfilterDBSize=64; // max number of further packet filter in url filter function
		rtl8651_tblDrvPara.dialSessionNumber=4; // supported dsid: 0,1
		rtl8651_tblDrvPara.dmzHostsNum=2; // supported dsid: 0,1
		rtl8651_tblDrvPara.policyRules = 8;
		rtl8651_tblDrvPara.dynamicPolicy = 32;
		rtl8651_tblDrvPara.pptpSize = 2;
		rtl8651_tblDrvPara.l2tpSize = 2;
		rtl8651_tblDrvPara.pppoePassthruSize = 8;
		rtl8651_tblDrvPara.protocolBasedNatSize = 8;
		rtl8651_tblDrvPara.mCastTblSize = 64;
		rtl8651_tblDrvPara.macAccessControllSize = 8;
		rtl8651_tblDrvPara.ipUnnumberSize = 1;
	}
	else {
		rtl8651_tblDrvPara = *configPara;
		if(rtl8651_tblDrvPara.macAddressDbSize == 0)
			rtl8651_tblDrvPara.macAddressDbSize = 8;
		if(rtl8651_tblDrvPara.filterDbSize == 0)
			rtl8651_tblDrvPara.filterDbSize = 1024;
		if(rtl8651_tblDrvPara.networkIntfTableSize == 0)
			rtl8651_tblDrvPara.networkIntfTableSize = 8;
		if(rtl8651_tblDrvPara.ipIntfEntryNumber == 0)
			rtl8651_tblDrvPara.ipIntfEntryNumber = 16;
		if(rtl8651_tblDrvPara.ipIntfIpAddrNumber == 0)
			rtl8651_tblDrvPara.ipIntfIpAddrNumber = 24;
		if(rtl8651_tblDrvPara.arpTableSize == 0)
			rtl8651_tblDrvPara.arpTableSize = 512;
		if(rtl8651_tblDrvPara.routingTableSize == 0)
			rtl8651_tblDrvPara.routingTableSize = 64;
		if(rtl8651_tblDrvPara.natSize == 0)
			rtl8651_tblDrvPara.natSize = 8;
		if(rtl8651_tblDrvPara.naptServerPortEntryNumber == 0)
			rtl8651_tblDrvPara.naptServerPortEntryNumber = 8;
		if(rtl8651_tblDrvPara.naptTcpUdpFlowSize == 0)
			rtl8651_tblDrvPara.naptTcpUdpFlowSize = 4096;
		if(rtl8651_tblDrvPara.naptIcmpFlowSize == 0)
			rtl8651_tblDrvPara.naptIcmpFlowSize = 128;
		if(rtl8651_tblDrvPara.aclSize == 0)
			rtl8651_tblDrvPara.aclSize = 256;
		if(rtl8651_tblDrvPara.algSize == 0)
			rtl8651_tblDrvPara.algSize = 128;
		if(rtl8651_tblDrvPara.dialSessionNumber == 0)
			rtl8651_tblDrvPara.dialSessionNumber = 4;
		if(rtl8651_tblDrvPara.dmzHostsNum== 0)
			rtl8651_tblDrvPara.dmzHostsNum= 2;
		if (rtl8651_tblDrvPara.urlfilterDBSize == 0)
			rtl8651_tblDrvPara.urlfilterDBSize = 16;
		if (rtl8651_tblDrvPara.urlfilterStringLen == 0)
			rtl8651_tblDrvPara.urlfilterStringLen = 32;
		if (rtl8651_tblDrvPara.pathfilterStringLen == 0)
			rtl8651_tblDrvPara.pathfilterStringLen = 32;
		if (rtl8651_tblDrvPara.urlPktfilterDBSize == 0)
			rtl8651_tblDrvPara.urlPktfilterDBSize = 64;
		if (rtl8651_tblDrvPara.policyRules == 0)
			rtl8651_tblDrvPara.policyRules = 8;
		if (rtl8651_tblDrvPara.dynamicPolicy == 0)
			rtl8651_tblDrvPara.dynamicPolicy = 32;
		if (rtl8651_tblDrvPara.pptpSize == 0)
			rtl8651_tblDrvPara.pptpSize = 2;
		if (rtl8651_tblDrvPara.l2tpSize == 0)
			rtl8651_tblDrvPara.l2tpSize = 2;
		if (rtl8651_tblDrvPara.pppoePassthruSize == 0)
			rtl8651_tblDrvPara.pppoePassthruSize = 8;
		if (rtl8651_tblDrvPara.protocolBasedNatSize == 0)
			rtl8651_tblDrvPara.protocolBasedNatSize = 8;
		if (rtl8651_tblDrvPara.mCastTblSize == 0)
			rtl8651_tblDrvPara.mCastTblSize = 64;
		if (rtl8651_tblDrvPara.macAccessControllSize == 0)
			rtl8651_tblDrvPara.macAccessControllSize = 8;
		if (rtl8651_tblDrvPara.ipUnnumberSize == 0)
			rtl8651_tblDrvPara.ipUnnumberSize = 1;
		
	}
	//rtl8651_tblDrvPara.routingTableSize = RTL8651_ROUTINGTBL_SIZE;	
	//Unacceptable parameter setting checking
	if(rtl8651_tblDrvPara.macAddressDbSize==0)
		return FAILED;
	if(rtl8651_tblDrvPara.filterDbSize < RTL8651_L2TBL_ROW * RTL8651_L2TBL_COLUMN)
		return FAILED;
/*	chenyl: Now we allow TcpUdpFlow Size < ASIC Table size
	if(rtl8651_tblDrvPara.naptTcpUdpFlowSize < RTL8651_TCPUDPTBL_SIZE)
		return FAILED;
*/
	if(rtl8651_tblDrvPara.naptIcmpFlowSize < RTL8651_ICMPTBL_SIZE)
		return FAILED;
	assert(RTL8651_FDB_NUMBER >= 1);	//Must have at least one filter database
	_sysUpSeconds	=0;  //initialize software timer value.
	return SUCCESS;
}



int32 _rtl8651_portStatusPatch(uint32 port) 
{
	/* For RTL865xB cut-c & before that. */
	if (RtkHomeGatewayChipRevisionID <= 0x02)
	{
		rtl8651_tblDrv_vlanTable_t *vlan_t;
		uint32 vidx;

		for(vidx=0,vlan_t=DrvTbl.vlan; vidx<RTL8651_VLAN_NUMBER; vidx++, vlan_t++) {
			if (vlan_t->valid == 0)
				continue;
			if (vlan_t->memberPortMask & (1 << port)) {	/* this port is one member of a vlan */
				if ((vlan_t->memberPortUpStatus & (1 << port)) == 0)	/* driver records this port is link down */
					return RTL8651_LINKPATCH_DRIVERDOWN;
				else												/* driver records this port is link up */
					return RTL8651_LINKPATCH_DRIVERUP;
			}
		}
		return RTL8651_LINKPATCH_NOTINVLAN;
	}
	/* For RTL865xB cut-d & after that. Actually, this function should not be used in RTL865xB cut-d. 
	     Because it is designed to patch for RTL865xB cut-b IC bug (link change interrupt disappear) */
	else
		return RTL8651_LINKPATCH_NOTBCUT;
}

