/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : rtl8651_layer2fwd.c
* Abstract : RTL8651 Home gateway controller Layer2  Fast Forward Engine
* Author : Chun-Feng Liu(cfliu@realtek.com.tw)
* $Id: rtl8651_layer2fwd.c,v 1.1.1.1 2006/07/26 05:38:23 blue Exp $
* $Log: rtl8651_layer2fwd.c,v $
* Revision 1.1.1.1  2006/07/26 05:38:23  blue
*
*
* Revision 1.1.1.1  2006/02/14 03:09:06  jason
* backup the source code from v1036
*
* Revision 1.1.1.1  2006/01/16 02:39:44  jason
* backup source,because now we directly use romedriver source code, i open a new directory;
*
* Revision 1.2  2005/11/22 15:50:54  chenyl
* *: Add mutex lock/unlock in log r/w
* *: new idle-timeout mechanism for VLAN / PPPOE / PPTP / L2TP wan types
* *: bug fix: PPTP passthrough
* *: bug fix: sip / ymsg typo error
* *: for ARP-reply module in RomeDrv, use newly allocated mbuf/pkt instead of original one ( To prevent from treated as packet forwarding instead of reply).
* *: minor coding style, sync with main trunk
*
* +: ioctl case for Mutex Lock/Unlock called by user level
* +: add 2 APIs to enable / disable
* 	1. Error messages from protocol stack to WAN : TCP RST, ICMP ICMP_UNREACH/ICMP_SOURCEQUENCH/ICMP_TIMXCEED/ICMP_PARAMPROB
* 	2. All packet from protocol stack to WAN
*    filtering to disable / enable idle-timer update triggered by these kinds of packets.
* +: TFTP ALG
* +: rtl8651_debug.h for debugging usage
* +: merge code about rtl_glue.c / re_core.c with Sercomm's version ( Airgo modification ?)
* +:
*
* Revision 1.1.1.1  2005/11/21 05:20:03  chenyl
* +: WPNT834
* 	- System provider	: Sercomm
* 	- Brand Name		: Netgear
* 	- 3rt Party		: Airgo Gen 3
* 	- SDK			: 066-3
* 	- ROMEDRV		: 3.6-3
*
*
* Revision 1.157  2005/09/27 05:59:40  chenyl
* *: modify IRAM / DRAM layout:
* 	IRAM/DRAM-FWD		: external used, swNic ... blahblah
* 	IRAM/DRAM-L2-FWD	: fwdengine internal used, L2 and below (ex. preprocess, postprocess)
* 	IRAM/DRAM-L34-FWD	: fwdengine internal used, L3/L4 process only (ex. Routing, decision table...etc).
*
* 	=> If L34 is used, we strongly suggest L2 must be used, too.
*
* Revision 1.156  2005/09/09 06:01:22  shliu
* *: remove "extern struct _RTL8651_PS_ACTION_ENTRY *rtl8651_ProtoStackActions;"
*
* Revision 1.155  2005/08/24 02:22:31  chenyl
* *: L2Input() would be forward function instead of EXTDRV function
*
* Revision 1.154  2005/08/23 14:38:26  chenyl
* +: apply prioirty IRAM/DRAM usage
*
* Revision 1.153  2005/08/19 08:40:35  shliu
* *: modify SW-vlan's vid parameter passing to _rtl8651_fwdEngineSend_ExtDevSend
*    in _rtl8651_fwdEngineSend.
*
* Revision 1.152  2005/08/18 09:56:27  chenyl
* *: declare variables at the top of block (prevent from compile error).
*
* Revision 1.151  2005/08/10 09:07:02  shliu
* *: change _rtl8651_rxPktPreprocess_stripVlan function to non-static.
*
* Revision 1.150  2005/07/27 10:34:02  shliu
* *: add RTL865xB d-cut software vlan support
*
* Revision 1.149  2005/07/13 14:06:08  chenyl
* *: bug fix: In order to fix ASIC's bug (UDP with zero checksum from extension port would be dropped by ASIC),
* 	    We always parse packet from Extension port and calculate UDP checksum (if it's zero) before sending to ASIC.
*
* Revision 1.148  2005/05/31 15:59:42  chenyl
* *: bug fix: we can not limit the extension port==6
*
* Revision 1.147  2005/05/31 15:37:51  chenyl
* *: bug fix: ph_srcExtPortNum is Port Number and we don't need to translate it with "extPortMaskToPortNum" anymore.
*
* Revision 1.146  2005/04/30 04:19:49  chenyl
* +: support (L2-Packet-filling + L3-Forwarding) offload Engine
*
* Revision 1.145  2005/04/18 03:21:45  chenyl
* *: unify the extended device's LinkID type to uint32
* 	- register/unregister
* 	- packet from extended device to ROMEDRV
*
* 	- Note that: the packet from ROMEDRV to extended device is via
* 		     rtlglue_ interface, and for the control purpose,
* 		     'int32' type of LinkID is used there.
*
* Revision 1.144  2005/04/06 07:29:52  cfliu
* -: Remove CONFIG_RTL865X_CACHED_NETWORK_IO code
*
* Revision 1.143  2005/03/22 11:06:50  tony
* *: support Software LSO and Protocol-Stack fast path(tunning for USB->SAMBA performance)
*
* Revision 1.142  2005/03/16 06:23:51  chenyl
* *: fix pptp idle-timeout problem
*
* Revision 1.141  2005/03/08 01:52:23  chenyl
* *: change the check mechanism in _rtl8651_fwdEngineSend()
* 	- If Packet is from Protocol Stack
* 		Org: check its rxdesc (-1)
* 		New: check the parameter 'fromDrv'
*
* Revision 1.140  2005/03/03 12:30:02  chenyl
* *: reset l2tp buffer state if new l2tp control packet is sniffed.
*
* Revision 1.139  2005/03/03 03:00:49  chenyl
* *: Only trap L3/4 processed IP packet to CPU for IP-Range ACL rules
*
* Revision 1.138  2005/02/25 07:49:06  chenyl
* *: spanning tree function patch up
* 	- update multicast spanning tree port state with unicast port state simultaneously.
* 	- trap 80211D packet instead of broadcast it.
* 	- set FWD state to non-bridge port
* 	- Don't support Extension port spanning tree management, so we set FWD state to Extension ports.
* 	- enable spanning tree process in SW ROMEDRV packet forwarding ( filter all NON-FWD-State port for
* 	  ROMEDRV packet forwarding packets and protocol stack sending packet).
*
* Revision 1.137  2005/02/02 01:37:57  chenyl
* *: fix L2TP ICCN/ZLB erroneous judgement problem
* +: forge L2TP header if ICCN has sent by protocol stack and sensed by ROMEDRV
*
* Revision 1.136  2005/01/28 12:09:23  yjlou
* *: abstract _rtl8651_fwdEngineSend() into :
*    _rtl8651_fwdEngineSend_PptpL2tp()
*    _rtl8651_fwdEngineSend_FromPS()
*    _rtl8651_fwdEngineSend_ExtDevSend()
*
* Revision 1.135  2005/01/26 08:27:57  chenyl
* *: fix PPTP/L2TP WAN type bugs:
*
* 	1. After decapsulating, and the DIP is broadcast IP (0xffffffff), we must let DMAC been ff-ff-ff-ff-ff-ff
* 	2. System MUST forbid L2 broadcasting if packet is from PPTP/L2TP loopback Port.
* 	3. Before trapping packet to protocol stack, system MUST modify packet's source VLAN if this packet is from
* 	   PPTP/L2TP loopback Port.
*
* Revision 1.134  2005/01/17 08:32:02  chenyl
* *: kick Watch-Dog when PCI-receiving
*
* Revision 1.133  2005/01/16 10:44:08  cfliu
* *** empty log message ***
*
* Revision 1.132  2005/01/12 08:18:12  chenyl
* *: re-add the code to strip VLAN-Tag / Priority-Tag
* 	because:
* 		1. forwarding engine don't fully support it yet (support Tagged-packet receive, instead of
* 			Tagged-packet forward)
* 		2. Linux kernel don't support it yet.
* *: modify ArpAgent process code
* 	get ARP information and reply-arp request according to
* 	etherType field's position
*
* Revision 1.131  2005/01/11 16:18:22  cfliu
* *** empty log message ***
*
* Revision 1.130  2005/01/11 08:51:18  chenyl
* *: check forwarding engine's multicast switch before L2-broadcasting to Extension port
*
* Revision 1.129  2005/01/11 07:49:14  chenyl
* *: bug fix : PPTP/L2TP header encap/decap's bug
* 	- packet from loopbackVlan --> outbound
* 	- packet from wanVlan --> inbound
* 	- modify m_data when decapsulating PPTP/L2TP header
*
* Revision 1.128  2005/01/10 08:16:59  chenyl
* *: fix testing module's error
*
* Revision 1.127  2005/01/10 07:45:15  yjlou
* *: merge rtl8651_drvPppoePassthruEnable, rtl8651_drvIpv6PassthruEnable, rtl8651_drvIpxPassthruEnable, rtl8651_drvNetbiosPassthruEnable into rtl8651_l2protoPassThrough.
*
* Revision 1.126  2005/01/10 03:58:40  cfliu
* +:Split rtl8651_fwdEngineExtPortRecv to fast and slow path.
* +:Modify rtl8651_fwdEngineExtPortRecv's fast path to chk ext port number and extportlist
*   Slow path named as _rtl8651_extPortRecvL2Lookup(). Only bcast or non-IP pkt will do L2 lookup.
* +:Add new functions: _rtl8651_layer2Passthrough, _rtl8651_pptpL2tpInput, _rtl8651_l2switch
* +:Put _rtl8651_fwdEnginePciSend to IRAM
*
* Revision 1.125  2005/01/09 08:46:04  cfliu
* +:Always put to I-RAM: rtl8651_rxPktPreprocess, rtl8651_txPktPostProcessing
* +:Add _rtl8651_layer2fwd_debug(),_rtl8651_rxPktPreprocess_stripVlan() to save fast path code size.
* +:Add rtl8651_fwdEngineAddWlanSTA, rtl8651_fwdEngineDelWlanSTA: New APIs for adding/deleting WLAN STAs
* +:Add rtl8651_fwdEngineExtPortUcastFastRecv(): extreme fast path for WLAN->LAN unicast pkt already in mbuf form.
*
* Revision 1.124  2005/01/07 11:07:58  yjlou
* *: rewrite for un-mark -DRTL865X_DEBUG.
*    Therefore, we can reduce file size.
*
* Revision 1.123  2005/01/06 09:27:22  yjlou
* *: change the index name of ROMEPERF
*
* Revision 1.122  2005/01/03 08:33:07  danwu
* * declare before operation.
*
* Revision 1.121  2004/12/27 01:48:15  yjlou
* +: add ROMEPERF code to measure performance.
*
* Revision 1.120  2004/12/20 12:57:21  yjlou
* +: add romeperf point to measure performance.
*
* Revision 1.119  2004/12/02 04:58:04  yjlou
* *: fix the IPX and Netbios ether type
*
* Revision 1.118  2004/12/02 03:08:37  yjlou
* +: support IPX/Netbios Passthru function
*
* Revision 1.117  2004/11/30 04:50:16  cfliu
* *:bugfix for multicast pkt bcast to ext port which may cause a storm
*
* Revision 1.116  2004/11/22 04:36:09  chenyl
* *: move TX-alignment code from layer2fwd to swNic2
*
* Revision 1.115  2004/11/22 03:38:02  chenyl
* *: fix patch-bug
*
* Revision 1.114  2004/11/22 02:57:00  chenyl
* *: ph_len is uint16 insteads of uint32, so 0xFFFC is better than 0xFFFFFFFC
*
* Revision 1.113  2004/11/22 02:49:21  jzchen
* +: Add NIC tx may over-run issue software patch
*
* Revision 1.112  2004/11/18 14:33:12  chhuang
* *: add a parameter to _rtl8651_modifyVlanRefFilterDatabaseEntry() to specify static/dynamic l2 entry
*
* Revision 1.111  2004/11/18 08:59:54  chenyl
* *: fix some unalignment code
* *: bug fix: enable forwarding engine's icmp-reply function for PPTP/L2TP/PPPoE Wan type
* *: re-write some comments and code
*
* Revision 1.110  2004/11/18 01:54:44  chenyl
* -: remove some comments
* *: return SUCCESS and release mbuf when some data-path process fail
* *: Fix some unalignment codes
*
* Revision 1.109  2004/11/16 02:30:30  shliu
* *** empty log message ***
*
* Revision 1.108  2004/11/15 06:09:45  chenyl
* *: bug fix : multicast flow-cache update bugs
* 		- flow-cache might cause age-counting error in some cases
* 		- must filter source vlan/port when add multicast member to flow-cache
* 		- don't remove node as multicast upstream when that node send LEAVE
*
* Revision 1.107  2004/11/10 08:14:32  cfliu
* *:fix STA move between different linkID problem
*
* Revision 1.106  2004/11/10 05:25:18  cfliu
* -: Move symbol export statements to rtl_glue.c
*
* Revision 1.105  2004/11/01 06:34:25  tony
* *: support rtl8651_addProtoStackServerPortRange
*
* Revision 1.104  2004/10/29 02:23:12  yjlou
* *: cfliu says Passthru VLAN index modification should be implemented in re_core.c, instead of layer2fwd.c.
* +: add _modifyPassthruVid()
* -: remove _modifyPPPoEPassthruVid()
*
* Revision 1.103  2004/10/22 11:46:44  cfliu
* *: Bugfix: (cfliu): We should check port link status by VLAN since port link status differs from VLAN if SPT is enabled but use this global variable for now until we turn on STP.
*
* Revision 1.102  2004/10/21 06:12:59  chenyl
* *: bug fix: free L2-broadcasted Mcast packet when we send it fail.
*
* Revision 1.101  2004/10/21 03:21:25  yjlou
* *: consider peripheral ExtPort in PPPoE/IPv6 Passthru forwarding.
*
* Revision 1.100  2004/10/21 03:06:48  yjlou
* +: add PPPOE_PASSTHRU_NEW_MODEL: use simple mechanism to forward PPPoE and IPv6 Passthru packets.
*
* Revision 1.99  2004/10/14 02:41:01  chhuang
* *: support routing entry > 8 in nexthop function
*
* Revision 1.98  2004/10/12 07:14:20  yjlou
* *: change PPPoE Passthru VLAN ID of gateway's packet in layer2. For IP multicast.
*
* Revision 1.97  2004/10/08 10:58:11  yjlou
* *: more proper portlist in PPPoE Passthru.
*
* Revision 1.96  2004/10/08 03:14:56  yjlou
* *: fixed the bug of PPPoE Passthru that sends packets to Mii Port.
*
* Revision 1.95  2004/10/05 09:40:56  cfliu
* add idle timeout patch to avoid device->LAN pkt updates idle timer
*
* Revision 1.94  2004/10/05 03:30:28  chenyl
* *: bug fix: layer2 forward bouncing
*
* Revision 1.93  2004/10/04 02:16:23  chenyl
* +: pptp/l2tp header refill function
*
* Revision 1.92  2004/09/21 07:08:40  chenyl
* *: modify layer2fwd.c forbid error-pptp/l2tp encapsulation
*
* Revision 1.91  2004/09/20 15:41:56  chenyl
* *: fix pptp/l2tp multicast port bouncing problem
*
* Revision 1.90  2004/09/16 05:43:32  chenyl
* *: multicast: remove/modify temporary APIs
* 		- rtl8651_multicastSetUpStream: remove 'port'
* 		- remove temp APIs add/del-ExternalMulticastPort();
* +: multicast: dynmaic upstream port selection
* 		- if upstream vlan (set by rtl8651_multicastSetUpStream) has multiple WAN port,
* 		  system will dynamically select one active port of it to forward.
*
* Revision 1.89  2004/09/15 16:14:28  chenyl
* +: enable multicast forward cache
*
* Revision 1.88  2004/09/15 14:38:18  yjlou
* *: commit for CDRouter case 228: PPPoE Passthru Unknown MAC Address
*   +: add rtl8651_enableDropUnknownPppoePADT()
*   +: In this mode, 0x8863 packets must be trapped to CPU, and unknown dmac will be dropped.
*
* Revision 1.87  2004/09/07 14:52:13  chenyl
* *: bug fix: napt-redirect fragment packet checksum-recalculate
* *: bug fix: conflict between protocol stack flow cache and protocol stack action
* *: bug fix: protocol stack action mis-process to UDP-zero-checksum packet
* *: separate the header file:
*         - internal : rtl8651_tblDrvFwdLocal.h
*         - external : rtl8651_tblDrvFwd.h
*
* Revision 1.86  2004/08/31 03:31:02  chhuang
* +: new feature: nexthop
*
* Revision 1.85  2004/08/30 05:19:47  tony
* *: fix bug - the path of PPTP server to protocol stack is failed.
*
* Revision 1.84  2004/08/30 02:54:20  tony
* *: fix bug - pptp can't handle LCP packet when un-compress ppp mode.
*
* Revision 1.83  2004/08/24 05:08:35  tony
* *: fix rtl8651_pptp_ppp_compress's default value to FALSE.
*
* Revision 1.82  2004/08/24 03:49:41  tony
* *: add function: rtl8651_pptpPppCompress
*
* Revision 1.81  2004/08/24 03:04:53  tony
* +: support pptp ppp un-compress mode.
*
* Revision 1.80  2004/08/12 01:52:13  cfliu
* *: modify debug message...
* *: bug fix for PPTP+WLAN mix use problem
*
* Revision 1.79  2004/08/10 10:04:08  chenyl
* *: pptp/l2tp checksum calculation
*
* Revision 1.78  2004/08/09 02:15:54  chhuang
* *: code re-architecture--phase 1
*
* Revision 1.77  2004/08/05 06:39:11  chenyl
* +: fragment shortest path forwarding
* *: UDPBomb don't aware ignore Type Checking result
*
* Revision 1.76  2004/07/28 12:34:24  cfliu
* *:[bugfix] don't send to extension port for multicast pkt come from protocol stack.
*
* Revision 1.75  2004/07/27 10:43:47  cfliu
* -: Remove all code which supports RTL865x WLAN using external loopback
*
* Revision 1.74  2004/07/23 13:19:57  tony
* *: remove all warning messages.
*
* Revision 1.73  2004/07/23 13:17:15  cfliu
* -: Remove legacy workaround code for 8650Bv1
*
* Revision 1.72  2004/07/23 09:22:03  cfliu
* *: solve WLAN internal loop of IP multicast pkts  by fix ph_srcExtPortNum field.
*
* Revision 1.71  2004/07/18 12:13:05  chhuang
* *: Disable debug message
*
* Revision 1.70  2004/07/16 03:05:27  tony
* *: fix rxdesc check bug.
*
* Revision 1.69  2004/07/15 11:16:02  tony
* +: new features: make remote management support new action: PS_DONT_TIMEOUT
*
* Revision 1.68  2004/07/15 07:46:09  tony
* *: fix bug: PPP LCP packets make gateway can't silent timeout.
*
* Revision 1.67  2004/07/13 13:52:06  yjlou
* +: ROME driver supports IPv6 passthru forwarding even when protocol-based VLAN is created.
*
* Revision 1.66  2004/07/13 02:53:16  tony
* *: fix protostack packets be drop in pptp/l2tp mode.
*
* Revision 1.65  2004/07/12 04:25:36  chenyl
* *: extend existing port scan mechanism
*
* Revision 1.64  2004/07/12 03:29:38  tony
* *: make remote management is able to run in pppoe mode.
*
* Revision 1.63  2004/07/08 09:24:33  cfliu
* *: Fix src port number error
*
* Revision 1.62  2004/07/08 04:11:05  yjlou
* +: ROME driver supports PPPoE passthru forwarding even when protocol-based VLAN is created.
* -: PPPoE passthru no more monitor the control messages.
*
* Revision 1.61  2004/07/06 14:19:44  cfliu
* +: remove all __IRAM
*
* Revision 1.60  2004/07/05 06:44:32  cfliu
* *: add vlan tag check code for fre 8650Bv2 chips
*
* Revision 1.59  2004/07/04 14:59:30  cfliu
* +: move protocol stack action code to L3/4 fwd engine.
*
*/

#include "rtl_types.h"
#include "assert.h"
#include "rtl8651_tblAsicDrv.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvPatch.h"
#include "rtl8651_tblDrvStatistic.h"
#include "asicRegs.h"
#include "asicTabs.h"
#include "rtl_glue.h"
#include "rtl_queue.h"
#include "rtl_utils.h"
#ifdef RTL865X_TEST
#include <string.h>
#endif
#ifdef CONFIG_RTL865X_ROMEPERF
#include "../romeperf.h"
#endif

//8650Bv1 has CRC error problem on Rx ==> Add VLAN tag
//has extra VLAN tag problem on Tx ==> Remove VLAN tag.
#ifdef CONFIG_RTL865XB
#define RTL865XB_EXTPORT_VLAN_PATCH
#endif
#define CHIP_BUG_PATCH

//extern uint16 rtl8651_ProtoStackServerUsedTcpPorts[RTL8651_MAX_PROTO_STACK_SERVER_USED_TCP_PORTS];

static int32  _rtl8651_fwdEnginePciSend(void * data, int16 dvid, int32 iphdrOffset);
static int32  (*_arpAgent)(struct rtl_pktHdr *pkthdrPtr,struct rtl_mBuf *mbufPtr,uint16 vid)=NULL;
static void _rtl8651_ExtPortChecksumProcess(struct rtl_mBuf *mbuf);

void rtl8651_installArpAgent(int32  (*arpAgent)(struct rtl_pktHdr *pkthdrPtr,struct rtl_mBuf *mbufPtr,uint16 vid)){
	_arpAgent=arpAgent;
}
#if 0
void config_dram(unsigned int base, unsigned int size)
{
    __asm__ __volatile__(
        ".set\tnoreorder\n\t"
        "mfc0\t$8,$12\n\t"
        "nop\n\t"
        "nop\n\t"
        "or\t$8,0x80000000\n\t"
        "mtc0\t$8,$12\n\t"
        "nop\n\t"
        "nop\n\t"
        "mtc0\t$0,$20\n\t"
        "nop\n\t"
        "nop\n\t"
        "li\t$8,0x3\n\t"
        "mtc0\t$8,$20\n\t"
        "nop\n\t"
        "nop\n\t"
        "move\t$8,$4\n\t"
        "la\t$9,0x0ffffc00\n\t"
        "and\t$8,$8,$9\n\t'"
        "mtc3\t$8,$4\n\t"
        "nop\n\t"
        "nop\n\t"
        "mtc3\t$8,$4\n\t" //writing again to make it work, do not know why???
        "nop\n\t"
        "nop\n\t"
        "add\t$8,$8,$5\n\t"
        "addiu\t$8,$8,-1\n\t"
        "mtc3\t$8,$5\n\t"
        "nop\n\t"
        "nop\n\t"
        "mtc0\t$0,$20\n\t"
        "nop\n\t"
        "nop\n\t"
        ".set\treorder\n\t");
}
#endif

#define _RTL865XB_EXTPORTMASKS   7
static uint8 extPortMaskToPortNum[_RTL865XB_EXTPORTMASKS+1]={
	0, 1, 2, 0, 3, 0, 0, 0
};


#ifdef RTL865X_DEBUG
static void	_rtl8651_layer2fwd_debug(int idx, struct rtl_pktHdr * pPkthdr){
	struct rtl_mBuf *pMbuf=pPkthdr->ph_mbuf;
	uint8 *data = pMbuf->m_data;
	if(idx==1){
		//begin of rtl8651_rxPktPreprocess
		printk("(%d)Rcvd===============================len:%d\n",pPkthdr->ph_rxdesc,pPkthdr->ph_len);
		printk("  vidx: %d srcExt:0x%x, dstExt:0x%x, portlist: 0x%x vtag=%d\n", 
			pPkthdr->ph_vlanIdx, pPkthdr->ph_srcExtPortNum,
			pPkthdr->ph_extPortList, pPkthdr->ph_portlist, pPkthdr->ph_vlanTagged);
		memDump(data, 64, "NIC Rx");
		if(wlan_acc_debug>4)
			memDump(pPkthdr, 32, "pkthdr");
	}else if (idx==2){
		//end of rtl8651_rxPktPreprocess
		rtlglue_printf("  ===Recv len:%d\n",pPkthdr->ph_len);
		rtlglue_printf("  vidx: %d srcExt:0x%x, dstExt:0x%x, portlist: 0x%x vtag=%d\n", 
			pPkthdr->ph_vlanIdx, pPkthdr->ph_srcExtPortNum,
			pPkthdr->ph_extPortList, pPkthdr->ph_portlist, pPkthdr->ph_vlanTagged);
		if(wlan_acc_debug>3 &&pPkthdr->ph_portlist==6){
			memDump(data, 64, "NIC Rx");
		}
	}else if (idx==3){
		//end of tx post processing
		uint8 macStr[18];	
		rtlglue_printf("\nNIC Tx ");
		if(ether_ntoa_r((ether_addr_t *)&data[6], macStr)==0)
			rtlglue_printf(" 0x%x %s->", pPkthdr->ph_portlist, macStr);
		if(ether_ntoa_r((ether_addr_t *)&data[0], macStr)==0)
			rtlglue_printf("%s  %02x%02x %dB",macStr,data[12],data[13], pPkthdr->ph_len);
		rtlglue_printf("(v%d s%d d%d p%d)...\n",pPkthdr->ph_vlanIdx,pPkthdr->ph_srcExtPortNum,pPkthdr->ph_extPortList,pPkthdr->ph_portlist);
		if(wlan_acc_debug>3){
			memDump(&data[12],64,"NIC Tx content");
		}
	}else if(idx==4){
		uint8 macStr[18];
		ether_ntoa_r((ether_addr_t *)&data[6], macStr);
		rtlglue_printf("\nNIC Rx:(%d)%s->",pPkthdr->ph_rxdesc,macStr);
		ether_ntoa_r((ether_addr_t *)&data[0], macStr);
		rtlglue_printf("%s s%d d%d p%d %dB\n",macStr, pPkthdr->ph_srcExtPortNum,pPkthdr->ph_extPortList,pPkthdr->ph_portlist,pPkthdr->ph_len);
	}else if(idx==5){
		//start of rtl8651_fwdEngineExtPortRecv
		uint8 macStr[18];
		ether_ntoa_r((ether_addr_t *)&data[6], macStr);
		rtlglue_printf("PCI Rx/NIC Tx,%s->", macStr);
		ether_ntoa_r((ether_addr_t *)&data[0], macStr);
		rtlglue_printf("%s  %x %dB\n",macStr,*((uint16 *)(data+12)), pPkthdr->ph_len);
		memDump(&data[12],40,"");
	}
}
#endif

void _rtl8651_rxPktPreprocess_stripVlan(struct rtl_pktHdr * pPkthdr){
		struct rtl_mBuf *thisMbuf=pPkthdr->ph_mbuf;
		uint8 *data=thisMbuf->m_data;
		uint8 dasa[12];
		//printk("Extracting vlan tag...\n");
		memcpy(dasa,data,12);
		memcpy(data+4,dasa,12);
		pPkthdr->ph_len-=4;
		thisMbuf->m_len-=4;
		thisMbuf->m_data+=4;
}

__IRAM_L2_FWD int32 rtl8651_rxPktPreprocess(struct rtl_pktHdr * pPkthdr)
{
	uint32 destExtPortList , srcPortNum;
	
#ifdef CONFIG_RTL865XB
	destExtPortList = pPkthdr->ph_extPortList;
	srcPortNum=pPkthdr->ph_portlist;
	pPkthdr->ph_extPortList&=0xf; //ignore TTL-1 bits.
	if(srcPortNum>=7) {
		//src extension port field is meaningful only when portlist is 7.
		//assert(srcPortNum==7);//if not 7, it's an ASIC bug!
		//assert(pPkthdr->ph_srcExtPortNum>=1); //if not > 1, it's an ASIC bug!
		pPkthdr->ph_extPortList &= ~(1<<(pPkthdr->ph_srcExtPortNum-1)); //avoid looping back to src ext port again.
		if(pPkthdr->ph_extPortList==0){
			//No need to pass the pkt to upper layer. It's just a loopback pkt.
			//free the packet immediately.
			return FAILED;
		}else if (pPkthdr->ph_extPortList<8){
			//if dest Ext port 0x1 => to dst ext port 1 => from src port 1+5=6
			//if dest Ext port 0x2 => to dst ext port 2 => from src port 2+5=7
			//if dest Ext port 0x4 => to dst ext port 3 => from src port 3+5=8
			srcPortNum=extPortMaskToPortNum[pPkthdr->ph_extPortList]+RTL8651_PORT_NUMBER-1;
			pPkthdr->ph_portlist=srcPortNum;

#if 1
			/*
				chenyl (2005/10/14)

				The problem for 865xBv1: Packet which dest to Ext port would carry the vlanIdx being
				"source vlan Index" instead of "destination vlan Index". (Design has changed after C-Cut.)

				=>	Driver MUST modify the vlan Index to destination vlan index if we've detected this packet
					is sent to extension port.

				=> (2005/10/14)
					Originally, we check if we need to change vlanIdx according to the "TTL-1" bit.
					It is because that if the "TTL-1 bit set" means this packet has been routed/NAPTed by HW,
					so the vlanIdx modification is needed. (L2 bridge done by HW, TTL-1 is not set, but, vlanIdx is NOT needed, too.)

					However, for some productions, TTL-1 is turned OFF for L3/4 process.
					In this case, the patch would be skip and error occurs.
					So we always patch vlanIdx issue for all packets to extsion port.
			*/
			if (RtkHomeGatewayChipRevisionID < 0x02)
#else
			if((destExtPortList&0x78)!=0x8)
				//TTL-1 done. Pkt has been routed/NAPT. Must change vlan idx due to 865xBv1 bug
#endif
			{
				pPkthdr->ph_vlanIdx=DrvTbl.aggregator[srcPortNum].pvid&(RTL8651_VLAN_NUMBER-1);
			}
		}else{
			//>8: has CPU bit, pkt is original pkt from port 6~8
			pPkthdr->ph_portlist=pPkthdr->ph_srcExtPortNum+RTL8651_PORT_NUMBER-1;
		}
		//keep ph_srcExtPortNum info to prune incoming port when sending bcast pkt.
	}else{
		//otherwise, pkt is rcvd from PHY
		pPkthdr->ph_srcExtPortNum=0;
		if((pPkthdr->ph_extPortList&0xf)<8){ //No CPU bit, only dest ext mbr port...
			//if dest Ext port 0x1 => to dst ext port 1 => from src port 1+5=6
			//if dest Ext port 0x2 => to dst ext port 2 => from src port 2+5=7
			//if dest Ext port 0x4 => to dst ext port 3 => from src port 3+5=8
			if(pPkthdr->ph_extPortList){
				//redefine src port number
				srcPortNum=extPortMaskToPortNum[pPkthdr->ph_extPortList]+RTL8651_PORT_NUMBER-1;
				pPkthdr->ph_portlist=srcPortNum;
			}
#if 1
			/*
				=> chenyl (2005/10/14)
					Originally, we check if we need to change vlanIdx according to the "TTL-1" bit.
					It is because that if the "TTL-1 bit set" means this packet has been routed/NAPTed by HW,
					so the vlanIdx modification is needed. (L2 bridge done by HW, TTL-1 is not set, but, vlanIdx is NOT needed, too.)

					However, for some productions, TTL-1 is turned OFF for L3/4 process.
					In this case, the patch would be skip and error occurs.
					So we always patch vlanIdx issue for all packets to extsion port.
			*/
			if (RtkHomeGatewayChipRevisionID < 0x02)
#else
			if(destExtPortList&0x70)
				//TTL-1 done. Pkt has been routed/NAPT. Must change vlan idx due to 865xBv1 bug
#endif
			{
				pPkthdr->ph_vlanIdx=DrvTbl.aggregator[srcPortNum].pvid&(RTL8651_VLAN_NUMBER-1);
			}
		}
	}		


	if(pPkthdr->ph_vlanTagged) {
		/* WMM (Wi-Fi Multimedia) needs the priority information of vlan-tag. */
		// Magic number 14 means 3rd byte of vlan-tag.
		pPkthdr->ph_priority = (ntohs(*((uint16 *)(pPkthdr->ph_mbuf->m_data + 14))) & 0xe000) >> 13;

		/* 
			For RTL865xB b-cut & before, IC has a DMA bug when handling undersized packet that destined to extension port. 
			This is one step of that patch.
			Because RTL865xB d-cut has fixed this IC bug, and in order to support unknown vid in RTL865xB d-cut, 
			we don't strip vlan-tag.
		*/
		if (RtkHomeGatewayChipRevisionID < 0x02) {
			//For 865xBv1 bug
			_rtl8651_rxPktPreprocess_stripVlan(pPkthdr);
			pPkthdr->ph_vlanTagged=0;
		}
	}
	else
		pPkthdr->ph_priority = 0;

	//assert(pPkthdr->ph_portlist<=RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum-1);
	//If pkt is original pkt, pPkthdr->ph_extPortList(extension port list) has 0x8 bit set

#endif /*CONFIG_RTL865XB*/
	
	return SUCCESS;
}


__IRAM_L2_FWD int32 rtl8651_txPktPostProcessing(struct rtl_pktHdr * pPkt)
{
	uint32 destPortMask=pPkt->ph_portlist;

	destPortMask&= ~0x40; //clear CPU port.

	if(destPortMask){
		//if portmask contains MII port. MII port must be configured before we can send pkt to it.
		if(destPortMask & RTL8651_MII_PORTMASK && miiPhyAddress<0)
			destPortMask&=~RTL8651_MII_PORTMASK;

		pPkt->ph_srcExtPortNum=pPkt->ph_extPortList=0;
		pPkt->ph_flags &= ~(PKTHDR_BRIDGING|PKTHDR_HWLOOKUP);
		destPortMask&=~rtl8651_allExtPortMask;
	}
#ifdef CONFIG_RTL865XB
	else if(pPkt->ph_srcExtPortNum){
		pPkt->ph_flags |=PKTHDR_HWLOOKUP; //turn on hardware L3/4 offloading for pkts from extension port.
		destPortMask=7; //send from extension port. destination portmask unknown

	}else
#endif /*CONFIG_RTL865XB*/
		pPkt->ph_flags &= ~(PKTHDR_BRIDGING|PKTHDR_HWLOOKUP); //always clear these bits in 8650/51
	pPkt->ph_portlist=destPortMask;

	return SUCCESS;

}	

	

int32 rtl8651_bridgeEnable;
int32 rtl8651_fwdEngineEnableWDS(int32 flag){
	rtl8651_bridgeEnable=flag;
	return SUCCESS;
}

//Remove all IP based hosts from this external link ID
//our current driver limitation forbids removing L2 entries directly so hosts
//running protocols such as IPX, Appletalk won't be removed after 'linkID' removed.
int32 rtl8651_fwdEngineRemoveHostsOnExtLinkID(uint32 linkID){
	int32 retval=FAILED;
	if(linkID==0)
		return retval;
	rtlglue_drvMutexLock();
	_rtl8651_removeArpAndNaptFlowWhenLinkDown(-1,&linkID);
	rtlglue_drvMutexUnlock();
	return retval;
}

 int32 rtl8651_fwdEngineAddWlanSTA(uint8 *smac,uint16 myvid, uint32 myportmask, uint32 linkId){
	rtl8651_tblDrv_vlanTable_t *thisVlan;
	rtl8651_tblDrv_filterDbTableEntry_t * thisFDBentry;
	int32 retval=FAILED;

	if (linkId==0)
		return retval;
	
	rtlglue_drvMutexLock();

	//L2 lookup
	thisVlan= &DrvTbl.vlan[ myvid&(RTL8651_VLAN_NUMBER-1)];
	assert(thisVlan);
	 thisFDBentry=_rtl8651_getVlanFilterDatabaseEntry(myvid, (ether_addr_t*)smac);
	
	//First, check source MAC address. If not yet in L2 table, it's a new station on PCI/WLAN. Add to L2 table.
	//use TYPEIII, so add an patch ACL is added later.
	if(!thisFDBentry){
		int32 retval;
		int32 l2type;
		//SA learning, but don't learn gateway MAC.
		if(thisVlan->macAddrNumber){
			//Edward says support only one MAC is ok, although 8650 supports more than 1 mac on each vlan..04/02/19
			if(memcmp(smac, &thisVlan->macAddr,6)==0) 
				goto out;//strange case. return FAILED so NIC would drop it in case looping happened on PCI interface.
		}
		if(smac[0]&1){ //Don't learn broadcast SA pkts.
			goto out;
		}
		//Not gateway MAC, so we can add it.

//#ifdef CONFIG_RTL865XB_EXP_WLAN
		//TypeIII is used in 8651 which needs ACL patch, in 8650B, we just use TypeII.
		assert(myportmask>RTL8651_PHYSICALPORTMASK);
		l2type=RTL8651_L2_TYPEII;
		retval=_rtl8651_addVlanRefFilterDatabaseEntry(l2type, 0, myvid, (ether_addr_t *)smac, RTL8651_FORWARD_MAC,myportmask);
//#endif
		if(!retval)//success, now specify link ID.
			_rtl8651_modifyVlanRefFilterDatabaseEntry(0,  (ether_addr_t *)smac,NULL,NULL, &linkId, L2_AUTO);

		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug)
		{
			uint8 macStr[18];
			ether_ntoa_r((ether_addr_t *)smac, macStr);
			rtlglue_printf("=====adding %s on PCI to L2 table , linkId=%d, errno=%d.\n ", macStr,linkId,retval);
		}
		#endif
		retval=SUCCESS;
	}	else if(thisFDBentry->memberPortMask!=myportmask ||(thisFDBentry->linkId!=linkId)){
		#ifdef RTL865X_DEBUG
		//a station moved from LAN to PCI/WLAN port
		uint32 oldmask=thisFDBentry->memberPortMask;
		uint32 oldid=thisFDBentry->linkId;
		//detected station port changed.

		if(wlan_acc_debug)
		{
			uint8 macStr[18];
			ether_ntoa_r((ether_addr_t *)smac, macStr);
			rtlglue_printf("=====SA %s moved from portmask 0x%x(%d) to 0x%x(%d)....", macStr, oldmask,oldid, myportmask, linkId);
		}
		#endif
		_rtl8651_modifyVlanRefFilterDatabaseEntry(0, (ether_addr_t *)smac,&myportmask, NULL,&linkId, L2_AUTO);
		retval=SUCCESS;
	}
out:
	rtlglue_drvMutexUnlock();
	return retval;
 }

	
 int32 rtl8651_fwdEngineDelWlanSTA(uint8 *mac,uint16 myvid){
 	rtl8651_tblDrv_vlanTable_t *thisVlan;
	rtl8651_tblDrv_filterDbTableEntry_t * thisFDBentry;
	int32 retval=FAILED;	
	rtlglue_drvMutexLock();
	//L2 lookup
	thisVlan= &DrvTbl.vlan[ myvid&(RTL8651_VLAN_NUMBER-1)];
	assert(thisVlan);
	 thisFDBentry=_rtl8651_getVlanFilterDatabaseEntry(myvid, (ether_addr_t*)mac);
	if(thisFDBentry){
		retval=_rtl8651_delVlanRefFilterDatabaseEntry(RTL8651_L2_TYPEII,0,(ether_addr_t *)mac);
	}
	rtlglue_drvMutexUnlock();	
	return retval;
 }


__IRAM_EXTDEV int32 rtl8651_fwdEngineExtPortUcastFastRecv(struct rtl_pktHdr *pkt,uint16 myvid, uint32 myportmask)
{
	rtlglue_drvMutexLock();
	pkt->ph_flags &= ~(CSUM_IP| CSUM_L4|PKTHDR_BRIDGING);
	pkt->ph_portlist = myportmask;
	pkt->ph_vlanIdx= myvid & (RTL8651_VLAN_NUMBER-1);
	pkt->ph_srcExtPortNum= extPortMaskToPortNum[ (myportmask>>RTL8651_PORT_NUMBER)&_RTL865XB_EXTPORTMASKS];
	pkt->ph_portlist=0; 
	pkt->ph_portlist |= rtl8651_tblDrvCpuMirrorPortMask;

	/* check some checksum related issue before sending to ASIC */
	_rtl8651_ExtPortChecksumProcess(pkt->ph_mbuf);

	//send pkt here
	rtlglue_drvSend((void*)pkt);
	rtlglue_drvMutexUnlock();
	return SUCCESS;
 }

static  void _rtl8651_extPortRecvL2Lookup(uint8 *data, uint16 myvid, uint32 myportmask, uint32 linkID){
	uint8 *smac=data+6; //sixth byte should be SMAC.
	rtl8651_tblDrv_filterDbTableEntry_t * thisFDBentry;
	rtl8651_tblDrv_vlanTable_t *thisVlan;
	//L2 lookup
	thisVlan= &DrvTbl.vlan[ myvid&(RTL8651_VLAN_NUMBER-1)];
	assert(thisVlan);
	 thisFDBentry=_rtl8651_getVlanFilterDatabaseEntry(myvid, (ether_addr_t*)smac);
	
	//First, check source MAC address. If not yet in L2 table, it's a new station on PCI/WLAN. Add to L2 table.
	//use TYPEIII, so add an patch ACL is added later.
	if(!thisFDBentry){
		int32 retval;
		int32 l2type;
		//SA learning, but don't learn gateway MAC.
		if(thisVlan->macAddrNumber){
			//Edward says support only one MAC is ok, although 8650 supports more than 1 mac on each vlan..04/02/19
			if(memcmp(smac, &thisVlan->macAddr,6)==0) 
				return;//strange case. return FAILED so NIC would drop it in case looping happened on PCI interface.
		}
		if(smac[0]&1) //Don't learn broadcast SA pkts.
			return;
		//Not gateway MAC, so we can add it.

//#ifdef CONFIG_RTL865XB_EXP_WLAN
		//TypeIII is used in 8651 which needs ACL patch, in 8650B, we just use TypeII.
		if(myportmask>RTL8651_PHYSICALPORTMASK)
			l2type=RTL8651_L2_TYPEII;
		else
			l2type=RTL8651_L2_TYPEIII;
		retval=_rtl8651_addVlanRefFilterDatabaseEntry(l2type, 0, myvid, (ether_addr_t *)smac, RTL8651_FORWARD_MAC,myportmask);
//#endif
		if(!retval)//success, now specify link ID.
			_rtl8651_modifyVlanRefFilterDatabaseEntry(0,  (ether_addr_t *)smac,NULL,NULL, &linkID, L2_AUTO);

		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug)
		{
			uint8 macStr[18];
			ether_ntoa_r((ether_addr_t *)smac, macStr);
			rtlglue_printf("=====adding %s on PCI to L2 table , linkId=%d, errno=%d.\n ", macStr,linkID,retval);
		}
		#endif
		
	} else if(thisFDBentry->memberPortMask!=myportmask ||(thisFDBentry->linkId!=linkID)){
		#ifdef RTL865X_DEBUG
		//a station moved from LAN to PCI/WLAN port
		uint32 oldmask=thisFDBentry->memberPortMask;
		uint32 oldid=thisFDBentry->linkId;
		//detected station port changed.

		if(wlan_acc_debug)
		{
			uint8 macStr[18];
			ether_ntoa_r((ether_addr_t *)smac, macStr);
			rtlglue_printf("=====SA %s moved from portmask 0x%x(%d) to 0x%x(%d)....", macStr, oldmask,oldid, myportmask, linkID);
		}
		#endif

		_rtl8651_modifyVlanRefFilterDatabaseEntry(0, (ether_addr_t *)smac,&myportmask, NULL,&linkID, L2_AUTO);
	}
}

__IRAM_EXTDEV static void _rtl8651_ExtPortChecksumProcess(struct rtl_mBuf *mbuf)
{
	/* handler for the case: (UDP checksum == 0) */
	uint8 *pkt = mbuf->m_data;
	uint16 byte2;
	uint32 etherPos = 12;
	struct ip *pip = NULL;

nextRound:
	byte2 = *(uint16*)(&pkt[etherPos]);
	switch (byte2)
	{
		case 0x0800:
			pip = (struct ip *)(&pkt[(etherPos+2)]);
			break;
		case 0x8864:
			byte2 = ntohs(*((uint16 *)(&pkt[etherPos + 8])));
			if (byte2 == 0x2100)
			{
				pip = (struct ip *)(&pkt[(etherPos + 10)]);
				break;
			} else
			{
				goto check_done;
			}
		case 0x8100:	/* Tagged */
			if ( pkt[etherPos + 2] & 0x10 )
			{	/* CFI bit is set, ignore */
				goto check_done;
			}
			etherPos += 4;
			goto nextRound;
		default:	/* other cases, including LLC, other Ethertypes .... we don't process them or don't support them (ex.LLC) here */
			goto check_done;
	}

	assert(pip);

	if ((pip->ip_vhl & 0xf0) != 0x40)
	{
		goto check_done;
	}

	if (	(!IS_FRAGMENT(pip->ip_off)) &&
		(pip->ip_p == IPPROTO_UDP))
	{
		struct udphdr *ud = (struct udphdr*)L4HDR_ADDR(pip);

		if (ud->uh_sum == 0)
		{
			ud->uh_sum = 0;
			pip->ip_sum = 0;
			ud->uh_sum = _rtl8651_tcpChecksum(pip);
			pip->ip_sum = _rtl8651_ipChecksum(pip);
		}
	}

check_done:
	return;
}

int32 _WdsLinkFwd(uint8 *data,struct rtl_pktHdr* pktHdr, uint32 linkID, uint16 myvid, uint8 *vlanMac ){
	if(data[0]&1){
		struct rtl_mBuf *newMbuf= mBuf_dupPacket(pktHdr->ph_mbuf, MBUF_DONTWAIT);
		if(newMbuf){
			#ifdef RTL865X_DEBUG
			if(wlan_acc_debug>1){
				rtlglue_printf("wds bcast fwd %d -> 0",linkID);
			}
			#endif
			rtlglue_extDeviceSend(myvid, -((int32)(linkID)),newMbuf->m_pkthdr);
		}
		#ifdef RTL865X_DEBUG
		else if(wlan_acc_debug>1){
			rtlglue_printf("wds bcast from %d failed. No mbuf\n",linkID);
		}
		#endif
	}else if(memcmp(data, vlanMac,6)!=0){
		rtl8651_tblDrv_filterDbTableEntry_t *dal2Entry;

		//unicast, DA lookup
		dal2Entry=_rtl8651_getVlanFilterDatabaseEntry(myvid, (ether_addr_t *)data); //DA lookup.
		if(dal2Entry&&dal2Entry->linkId>0){
			int32 UnicastToWdsLinkID= dal2Entry->linkId; //linkId>0 means DA is another WLAN station. 
#ifdef RTL865X_DEBUG
			if(wlan_acc_debug>1)
				rtlglue_printf("wds unicast fwd %d ->%d",linkID, UnicastToWdsLinkID );
#endif
			rtlglue_extDeviceSend(myvid, UnicastToWdsLinkID, pktHdr);//would be freed in pci driver
			return 1;/*pkt sent. */
		}
	}
	return 0; /*pkt still needs to be sent to switch port*/
}


//From extension port (wireless/PCI) to swcore
__IRAM_EXTDEV int32 rtl8651_fwdEngineExtPortRecv(void *id, uint8 *data,  uint32 len, uint16 myvid, uint32 myportmask, uint32 linkID)
{
#define UNCACHE_MASK		0xa0000000
#define UNCACHE(x)		((uint32)x|UNCACHE_MASK)
	struct rtl_pktHdr 	*pktHdr=NULL;	
	struct rtl_mBuf	*m=NULL;
	int32 retval=FAILED;
	rtl8651_tblDrv_vlanTable_t *thisVlan;
	thisVlan= &DrvTbl.vlan[ myvid&(RTL8651_VLAN_NUMBER-1)];

#ifdef CONFIG_RTL865XB	
	/*reset watchdog timer.*/
	REG32(WDTCNR)|=WDTCLR;
#endif

	if (linkID==0) {
		rtlglue_printf("packet from illegal-device-LinkID (%d)\n", linkID);
		goto out;
	}

	assert(data);

	rtlglue_drvMutexLock();

	if(data[0]&1 || *((uint16 *)(data+12))!=0x0800){
		//only check L2 table when broadcast pkt or non-IP pkt is received.
		_rtl8651_extPortRecvL2Lookup(data, myvid, myportmask, linkID);
	}

	//get a new mbuf/pkthdr pair to bind with given packet
	m=mBuf_attachHeader((void*)UNCACHE(data),(uint32)(id),m_clusterSize, len,0);
	if(!m){
		#ifdef RTL865X_DEBUG
		rtlglue_printf("X");
		#endif/* RTL865X_DEBUG*/
		goto out;
	}
	SETBITS(m->m_flags, MBUF_EXT);
	pktHdr=m->m_pkthdr;
	pktHdr->ph_portlist = myportmask;

	//Don't touch higher layer cksum fields.
	pktHdr->ph_flags &= ~(CSUM_IP| CSUM_L4|PKTHDR_BRIDGING);

	//When WDS is enabled,is unicast pkt,is not GW mac, check if it is to another WLAN station.
	if(rtl8651_bridgeEnable){
		if(_WdsLinkFwd(data,pktHdr,linkID,myvid,(uint8 *)&thisVlan->macAddr)){
			retval=SUCCESS;
			goto out;
		}
	}

	/* check some checksum related issue before sending to ASIC */
	_rtl8651_ExtPortChecksumProcess(m);

	//Finally, send to swcore directly
	_rtl8651_fwdEnginePciSend((void*)pktHdr, myvid,-1);
//#if defined(RTL865X_DEBUG)&& !defined(RTL865X_TEST)
//	if(wlan_acc_debug>1){
//		rtlglue_printf("\tRes:0x%x\n", hsb_dport(myportmask>RTL8651_PHYSICALPORTMASK));
//	}
//#endif
	retval = SUCCESS;
out:
	rtlglue_drvMutexUnlock();
	return retval;//pkt consumed.
}

int rtl8651_pptp_ppp_compress=FALSE;
/*
@func int32		| _rtl8651_pptpPppCompress	| PPTP PPP Compress mode toggle.
@parm int | enable | TURE: enable compress, FALSE: disable compress
@comm 
This flag will make PPTP wan interface have different PPP content.
when this flag enable, PPP's content is 0x21.
else PPP's content is 0xff,0x03,0x00,0x21.
 */
void rtl8651_pptpPppCompress(int enable)
{
	rtl8651_pptp_ppp_compress=enable;
}

//#endif /*WLAN ACCELERATION*/

#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)

#define RTL8651_MAX_PPTPL2TP_BUFFERSIZE	60

static int32 _rtl8651_pptpL2tpInput(int32 inbound, struct rtl_pktHdr *pkthdrPtr,rtl8651_tblDrv_vlanTable_t *local_vlanp, uint8 *m_data, uint16 ethtype, rtl8651_tblDrv_pppTable_t **ppp);

#define RTL8651_NEW_L2TPBUFF_FILL_CODE	/* re-implement by chenyl (20050201) */

#ifdef RTL8651_NEW_L2TPBUFF_FILL_CODE
static inline void _rtl8651_forgeL2tpBuffer(uint8 **l2tpBuff, uint8 *pL2tpLength, uint8 *pL2tpLengthBit, ipaddr_t gatewayIp, ipaddr_t serverIp, uint16 tunnelID, uint16 sessionID)
{
	void *buff;
	struct ip *pip;
	struct udphdr *udp;
	struct wan_l2tp_header *l2tp;
	uint8 *ppp;

	/* get buffer and init offset */
	if (*l2tpBuff == NULL)
	{
		*l2tpBuff = rtlglue_malloc(RTL8651_MAX_PPTPL2TP_BUFFERSIZE);

		assert(*l2tpBuff);
	}
	buff = *l2tpBuff;

	*pL2tpLength = 0;

	/* forge IP header */
	pip = (struct ip *)buff;

	pip->ip_vhl = 0x45;
	pip->ip_tos = 0;
	pip->ip_len = 0;			/* this field will be adjusted with different packets */
	pip->ip_id = 1;			/* gotten from the air */
	pip->ip_off = 0;			/* we don't set DF bit here */
	pip->ip_ttl = 0xff;
	pip->ip_p = IPPROTO_UDP;	/* l2tp is based on UDP packet */
	pip->ip_sum = 0;			/* this field will be adjusted with different packets */
	PKTSET_UINT32_UNALIGNED(gatewayIp, &pip->ip_src);
	PKTSET_UINT32_UNALIGNED(serverIp, &pip->ip_dst);

	*pL2tpLength += sizeof(struct ip);

	/* forge UDP header */
	udp = (struct udphdr *)((uint32)buff + sizeof(struct ip));

	udp->uh_sport = htons(1701);	/* L2TP source port */
	udp->uh_dport = htons(1701);	/* L2TP destination port */
	udp->uh_ulen = 0;				/* this field will be adjusted with different packets */
	udp->uh_sum = 0;				/* this field will be adjusted with different packets */

	*pL2tpLength += sizeof(struct udphdr);

	/* forge L2TP header */
	l2tp = (struct wan_l2tp_header*)((uint32)udp + sizeof(struct udphdr));

	l2tp->flags_type = 0;		/* data type */
	l2tp->flags_length = 1;		/* length field is present */
	l2tp->flags_rsv1 = 0;		/* reserve bits must be 0 */
	l2tp->flags_sequence = 0;	/* generally we don't provide Ns/Nr */
	l2tp->flags_rsv2 = 0;		/* reserve bits must be 0 */
	l2tp->flags_offset = 0;		/* generally we don't provide offset */
	l2tp->flags_priority = 0;	/* generally we don't set priority bit */
	l2tp->flags_rsv3 = 0;		/* reserve bits must be 0 */
	l2tp->flags_version = 2;	/* RFC2661(3.1): Ver MUST be 2, indicating the version of the L2TP data message header described in this document. */

	l2tp->length = 0;						/* this field will be adjusted with different packets */
	l2tp->tunnel_id = ntohs(tunnelID);		/* tunnel ID is given from caller function */
	l2tp->session_id = ntohs(sessionID);	/* session ID is given from caller function */

	*pL2tpLengthBit = TRUE;	/* L2TP length field is present */
	*pL2tpLength += sizeof(struct wan_l2tp_header);

	/* forge PPP header */
	ppp = (uint8 *)((uint32)l2tp + sizeof(struct wan_l2tp_header));

	ppp[0] = 0xff;
	ppp[1] = 0x03;
	ppp[2] = 0x00;
	ppp[3] = 0x21;

	*pL2tpLength += 4;

}
#endif /* RTL8651_NEW_L2TPBUFF_FILL_CODE */

int32 _rtl8651_fwdEngineSend_PptpL2tp( struct rtl_pktHdr * pkthdrPtr,
                                       rtl8651_tblDrv_vlanTable_t *local_vlanp,
                                       struct rtl_mBuf *mbuf,
                                       struct ip *ip,
                                       rtl8651_tblDrv_routeTable_t *rt_t,
                                       uint32* pDestPortMask )
{
	rtl8651_tblDrv_pppTable_t *pppEntry_t;
	uint32 destPortMask = *pDestPortMask;
	int32 retval;

#if 1
	pppEntry_t=rt_t->rt_nhSession_t;
#else
	pppEntry_t = (rtl8651_tblDrv_pppTable_t *)rt_t->rt_nhstart_t->nexthop_t;
#endif
	assert( pppEntry_t );

	//	printk("send local_vlanp->vid=%d pppEntry_t->initiated=%d pkthdrPtr->ph_portlist=%x\n",local_vlanp->vid,pppEntry_t->initiated,pkthdrPtr->ph_portlist);	
	if((pppEntry_t->initiated)&&(local_vlanp->vid==tunnel.loopBackVid))
	{	
		//SMAC=PPTP_Server_MAC
		if(memcmp(mbuf->m_data+6,pppEntry_t->macInfo->macAddr.octet,6)==0)
		{
#ifdef CONFIG_RTL865XB_PPTPL2TP					
			if(tunnel.loopbackPort>=RTL8651_PORT_NUMBER)
			{
				pkthdrPtr->ph_vlanIdx= tunnel.loopBackVid & (RTL8651_VLAN_NUMBER-1);
				pkthdrPtr->ph_srcExtPortNum=extPortMaskToPortNum[CONFIG_RTL865XB_PPTPL2TP_PORT-(RTL8651_PORT_NUMBER-1)];
				pkthdrPtr->ph_portlist= pkthdrPtr->ph_extPortList=0;					

				/* check some checksum related issue before sending to ASIC */
				_rtl8651_ExtPortChecksumProcess(mbuf);
			}
#endif /* CONFIG_RTL865XB_PPTPL2TP	*/	
			destPortMask = pkthdrPtr->ph_portlist;

			retval = FWDENG_SEND_GOTO_SEND;
			goto out;
		}
#ifdef CONFIG_RTL865XB_PPTPL2TP			
		//else if(pkthdrPtr->ph_portlist>RTL8651_PHYSICALPORTMASK)
		else
		{// LAN->loopback: If destination portlist > physical port, this implies we are using
		 //8650B's extension port mechanism.
			//loopback to L2 fwd engine since 8650B's extension port can not be loopbacked as 8651's MII loopback solution.
			{
				ip->ip_sum=0;
				ip->ip_sum=_rtl8651_ipChecksum(ip);
				if (!IS_FRAGMENT(ip->ip_off))
				{
					switch(ip->ip_p)
					{
						struct tcphdr *tc;
						struct udphdr *ud;
						struct icmp *ic;
						struct igmp *ig;
						case IPPROTO_TCP:
							tc = (struct tcphdr *)L4HDR_ADDR(ip);
							tc->th_sum = 0;
							tc->th_sum = _rtl8651_tcpChecksum(ip);
							break;
						case IPPROTO_UDP:
							ud = (struct udphdr *)L4HDR_ADDR(ip);
							if (ud->uh_sum)
							{
								ud->uh_sum = 0;
								ud->uh_sum = _rtl8651_tcpChecksum(ip);
							}
							break;
						case IPPROTO_ICMP:
							ic = (struct icmp *)L4HDR_ADDR(ip);
							ic->icmp_cksum = 0;
							ic->icmp_cksum = _rtl8651_icmpChecksum(ip);
							break;
						case IPPROTO_IGMP:
							ig = (struct igmp *)L4HDR_ADDR(ip);
							ig->igmp_cksum = 0;
							ig->igmp_cksum = _rtl8651_icmpChecksum(ip);
							break;
					}
				}
		 	}
			pkthdrPtr->ph_vlanIdx= tunnel.loopBackVid & (RTL8651_VLAN_NUMBER-1);
			pkthdrPtr->ph_srcExtPortNum=pkthdrPtr->ph_extPortList=0;
			pkthdrPtr->ph_portlist= CONFIG_RTL865XB_PPTPL2TP_PORT;
			if (_rtl8651_pptpL2tpInput(FALSE/* outbound*/, pkthdrPtr, &DrvTbl.vlan[pkthdrPtr->ph_vlanIdx], pkthdrPtr->ph_mbuf->m_data, 0x0800, &pppEntry_t) != SUCCESS)
//					if ( rtl8651_fwdEngineL2Input(pkthdrPtr, &DrvTbl.vlan[pkthdrPtr->ph_vlanIdx], pkthdrPtr->ph_mbuf->m_data, 0x0800) != SUCCESS)
			{
				mBuf_freeMbufChain(pkthdrPtr->ph_mbuf);
			}
			retval = SUCCESS;
			goto out;
		}
#endif	/* CONFIG_RTL865XB_PPTPL2TP	*/		
	}
	else if(local_vlanp->vid==tunnel.wanVid)
	{
		uint16 iplen=ntohs(ip->ip_len);
		//memDump(ip, 64,"ip");
		//printk("vid==wanvid iplen=%d pppEntry_t->initiated=%d\n",iplen,pppEntry_t->initiated);
		if (rt_t->netif_t->linkLayerType == RTL8651_LL_PPTP)
		{
			if (memcmp(local_vlanp->macAddr.octet, mbuf->m_data + 6 , 6))
			{
				/* FIXME: we don't support forward to local subnet under PPTP/L2TP configuration now, so we drop it here */
				pkthdrPtr->ph_portlist = 0;
				pkthdrPtr->ph_srcExtPortNum = 0;	/* this implies it need to be dropped */
			}else
			{	/* source mac == gateway mac, encapsulate and send it */
				if(pppEntry_t->greBuf1_==NULL)
				{	
					#if 1
					pppEntry_t->greBuf1_=(uint8*)rtlglue_malloc(40);
					assert(pppEntry_t->greBuf1_);
					#else
					pppEntry_t->greBuf1_=(uint8*)rtlglue_malloc(1061);
					pppEntry_t->greBuf1_=(uint32)pppEntry_t->greBuf1_&0xfffffc00;
					pppEntry_t->greBuf1_+=1024;
					config_dram(pppEntry_t->greBuf1_, 37);
					#endif				
				}
				// Encapsulating GRE header				
				//if((pppEntry_t->initiated)&&(pkthdrPtr->ph_portlist!=(1<<(tunnel.wanPort)))&& // not from protcol stack
				if((pppEntry_t->initiated)&&(pkthdrPtr->ph_rxdesc>=0)&& // not from protcol stack
					(iplen>=20)&&(iplen<=local_vlanp->mtu))
				{
					struct ip* iphdr;
					uint32 newSeq;
					static int ip_id = 1;

					int offset=40;

					if(rtl8651_pptp_ppp_compress==TRUE)
					{
						offset-=3;
					}
					//printk("encapsulating\n");

					pppEntry_t->delta_++;
					//printk("delta=%d\n",pppEntry_t->delta_);
					newSeq=pppEntry_t->lastSeq_+pppEntry_t->delta_;
					pppEntry_t->greBuf1_[21]|=0x80; // enable ack bit

					if(rtl8651_pptp_ppp_compress==TRUE)
					{
						pppEntry_t->greBuf1_[24]=((iplen+1)>>8)&0xff;
						pppEntry_t->greBuf1_[25]=(iplen+1)&0xff;						
					}
					else						
					{
						pppEntry_t->greBuf1_[24]=((iplen+4)>>8)&0xff;
						pppEntry_t->greBuf1_[25]=(iplen+4)&0xff;
					}
					
					pppEntry_t->greBuf1_[28]=((newSeq)>>24)&0xff;
					pppEntry_t->greBuf1_[29]=((newSeq)>>16)&0xff;
					pppEntry_t->greBuf1_[30]=((newSeq)>>8)&0xff;
					pppEntry_t->greBuf1_[31]=(newSeq)&0xff;						
					pppEntry_t->greBuf1_[32]=(pppEntry_t->lastReceiveSeq_>>24)&0xff;
					pppEntry_t->greBuf1_[33]=(pppEntry_t->lastReceiveSeq_>>16)&0xff;
					pppEntry_t->greBuf1_[34]=(pppEntry_t->lastReceiveSeq_>>8)&0xff;
					pppEntry_t->greBuf1_[35]=pppEntry_t->lastReceiveSeq_&0xff;

					if(rtl8651_pptp_ppp_compress==TRUE)
					{
						pppEntry_t->greBuf1_[36]=0x21; // PPP TYPE:IP
					}
					else
					{
						pppEntry_t->greBuf1_[36]=0xff;
						pppEntry_t->greBuf1_[37]=0x03;
						pppEntry_t->greBuf1_[38]=0x00; // PPP TYPE:IP
						pppEntry_t->greBuf1_[39]=0x21; // PPP TYPE:IP
					}

					iphdr=(struct ip *)pppEntry_t->greBuf1_;

					pkthdrPtr->ph_proto = PKTHDR_IP; //IP
					pkthdrPtr->ph_flags |= CSUM_IP;	// enable hardware L3 checksum calculating
					pkthdrPtr->ph_flags &= ~CSUM_L4; //disable hardware L4 checksum calculating

					iphdr->ip_id = htons(ip_id ++);
					iphdr->ip_len = htons(iplen+offset);
					//printk("encapsulate\n");
					rtl8651_memcpy(mbuf->m_data-offset,mbuf->m_data,14);
					rtl8651_memcpy(mbuf->m_data-offset+14,pppEntry_t->greBuf1_,offset);
					if(offset<=128) mbuf->m_data-=offset;
					
					/************************* Patch for multicast *************************/
					if (mbuf->m_data[0]==0x01 && mbuf->m_data[1]==0x00 && mbuf->m_data[2]==0x5e)
					{
						rtl8651_memcpy(mbuf->m_data, &(pppEntry_t->macInfo->macAddr), 6);
					}
					/*******************************************************************/

					pkthdrPtr->ph_len += offset;
					mbuf->m_len += offset;
									
					pkthdrPtr->ph_portlist=1<<(tunnel.wanPort);
					destPortMask = pkthdrPtr->ph_portlist;
					//printk("ext->WAN(PPTP)");

					retval = FWDENG_SEND_GOTO_SEND;
					goto out;			
				}
				
				//GET PS PPTP INFO
				//else if((ip->ip_p==IPPROTO_GRE)&&(pkthdrPtr->ph_portlist!=(1<<(tunnel.wanPort))) //from PS
				else if((ip->ip_p==IPPROTO_GRE)&&(pkthdrPtr->ph_rxdesc<0) //from PS				
					&&(iplen>=28))
				{
					struct wan_gre_header *gre;
					uint16 iphlen = IPHDR_LEN(ip);
					gre = (struct wan_gre_header *) ((int8 *) ip + iphlen);

					if(gre->flags&16) // Sequence
					{
						if(gre->ver&128) //Acknowledge
						{
							pppEntry_t->lastSeq_=PKTGET_UINT32_UNALIGNED(&gre->seq);
							pppEntry_t->lastAck_=PKTGET_UINT32_UNALIGNED(&gre->ack);
							//printk("GRE PS seq=%d ack=%d delta=%d\n",pppEntry_t->lastSeq_,pppEntry_t->lastAck_,pppEntry_t->delta_);
					 		PKTSET_UINT32_UNALIGNED(pppEntry_t->lastSeq_+pppEntry_t->delta_,&gre->seq);
						 	PKTSET_UINT32_UNALIGNED(pppEntry_t->lastAck_+pppEntry_t->delta_,&gre->ack);
						}
						else
						{
							pppEntry_t->lastSeq_=PKTGET_UINT32_UNALIGNED(&gre->seq);
							//printk("GRE PS seq=%d delta=%d\n",pppEntry_t->lastSeq_,pppEntry_t->delta_);
							PKTSET_UINT32_UNALIGNED(pppEntry_t->lastSeq_+pppEntry_t->delta_,&gre->seq);
						}
						
						if((pppEntry_t->greBuf1_[26]!=((ntohs(gre->call_id)>>8)&0xff))|| //if call id change, copy gre header
							(pppEntry_t->greBuf1_[27]!=((ntohs(gre->call_id)&0xff)))||(pppEntry_t->peerCallId_==0))
						{
							pppEntry_t->peerCallId_=ntohs(gre->call_id);								
							rtl8651_memcpy(pppEntry_t->greBuf1_,ip,28);						
							pppEntry_t->greBuf1_[22]=0x88; //ppp protocol
							pppEntry_t->greBuf1_[23]=0x0b; //ppp protocol
							pppEntry_t->greBuf1_[26]=(pppEntry_t->peerCallId_>>8)&0xff;
							pppEntry_t->greBuf1_[27]=pppEntry_t->peerCallId_&0xff;
							//printk("callId=%d\n",pppEntry_t->peerCallId_);
						}						
				
					}
					else if(gre->ver&128) // Acknowledge
					{
						pppEntry_t->lastAck_=PKTGET_UINT32_UNALIGNED(&gre->seq);
					 	PKTSET_UINT32_UNALIGNED(pppEntry_t->lastAck_+pppEntry_t->delta_,&gre->seq);
					}
					
					retval = FWDENG_SEND_GOTO_SEND;
					goto out;
				}	//end ps			
			}// smac == gateway mac
		}//end PPTP path
		else if (rt_t->netif_t->linkLayerType == RTL8651_LL_L2TP)
		{
			if (memcmp(local_vlanp->macAddr.octet, mbuf->m_data + 6 , 6))
			{
				/* FIXME: we don't support forward to local subnet under PPTP/L2TP configuration now, so we drop it here */
				pkthdrPtr->ph_portlist = 0;
				pkthdrPtr->ph_srcExtPortNum = 0;	/* this implies it need to be dropped */
			}else
			{	/* source mac == gateway mac, encapsulate and send it */
				if(pppEntry_t->l2tpBuf1_==NULL)
				{	
					pppEntry_t->l2tpBuf1_=(uint8*)rtlglue_malloc(RTL8651_MAX_PPTPL2TP_BUFFERSIZE);
					assert(pppEntry_t->l2tpBuf1_);						
				}

				// Encapsulating L2TP header
				//if((pppEntry_t->initiated)&&(pkthdrPtr->ph_portlist!=(1<<tunnel.wanPort))&& // not from protcol stack
				if((pppEntry_t->initiated)&&(pkthdrPtr->ph_rxdesc>=0)&& // not from protcol stack
					(iplen>=20)&&(iplen<=1452))
				{				

					struct ip *iphdr=(struct ip *)pppEntry_t->l2tpBuf1_;
					struct udphdr *udph = (struct udphdr *) ((int8 *) iphdr + ((iphdr->ip_vhl &0xf) << 2));
					static uint16 ip_id = 1;
													
					
					pkthdrPtr->ph_proto=6; //UDP
					pkthdrPtr->ph_flags|=CSUM_IP;
					pkthdrPtr->ph_flags|=CSUM_L4;
					iphdr->ip_len=htons(iplen+pppEntry_t->l2tpOffset_);
					iphdr->ip_id = htons(ip_id ++);
					udph->uh_ulen=htons(iplen+18); // udp(8)+l2tp(6)+ppp(4)	
						
					if(pppEntry_t->l2tpLengthBit_==1)
					{
						udph->uh_ulen=htons(ntohs(udph->uh_ulen)+2);
						if((pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-2]==0)&&(pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-3]==0x03))
						{
							pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-10]=((iplen+12)>>8)&0xff; // 12:l2tp(8)+ppp(4)
							pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-9]=(iplen+12)&0xff;
						}
						else
						{
							pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-7]=((iplen+9)>>8)&0xff; // 12:l2tp(8)+ppp(1)
							pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-6]=(iplen+9)&0xff;						
						}
					}

					rtl8651_memcpy(mbuf->m_data-pppEntry_t->l2tpOffset_,mbuf->m_data,14);
					rtl8651_memcpy(mbuf->m_data-pppEntry_t->l2tpOffset_+14,pppEntry_t->l2tpBuf1_,pppEntry_t->l2tpOffset_);
					if(pppEntry_t->l2tpOffset_<=128) mbuf->m_data-=pppEntry_t->l2tpOffset_;
				
					/************************* Patch for multicast *************************/
					if (mbuf->m_data[0]==0x01 && mbuf->m_data[1]==0x00 && mbuf->m_data[2]==0x5e)
					{
						rtl8651_memcpy(mbuf->m_data, &(pppEntry_t->macInfo->macAddr), 6);
					}
					/*******************************************************************/

					pkthdrPtr->ph_len += pppEntry_t->l2tpOffset_;
					mbuf->m_len += pppEntry_t->l2tpOffset_;	
					pkthdrPtr->ph_portlist=1<<(tunnel.wanPort);	
					destPortMask = pkthdrPtr->ph_portlist;

					retval = FWDENG_SEND_GOTO_SEND;
					goto out;
				}

				//34:ip(20)+udp(8)+l2tp(6) , from protocol stack, capture first data packet
				else if((ip->ip_p==IPPROTO_UDP)&&(iplen>=34)) 
				{			
					struct udphdr *udp;
					udp = (struct udphdr *)L4HDR_ADDR(ip);
					if(ntohs(udp->uh_dport)==1701)
					{				
						struct wan_l2tp_header *l2tp=(struct wan_l2tp_header*)(udp+1);	
						//printk("from ps flagtype=%d priority=%d l2tpOffset=%d iplen=%d l2tp->length=%d\n",l2tp->flags_type,l2tp->flags_priority,pppEntry_t->l2tpOffset_,iplen,l2tp->length);
						if((ntohs(l2tp->tunnel_id)==0)&&(ntohs(l2tp->session_id)==0))
						{
							pppEntry_t->l2tpBuf1State_=0;
							pppEntry_t->l2tpOffset_=0;
							pppEntry_t->tunnelId_=0;
							pppEntry_t->l2tpSessionId_=0;
							pppEntry_t->l2tpServerIp_=0;
							pppEntry_t->l2tpLengthBit_=0;
						}

#ifdef RTL8651_NEW_L2TPBUFF_FILL_CODE
						/*
							data message, we get the first packet from protocol stack (offset==sequence==0)
							as the template of L2TP header
						*/
						if (	(pppEntry_t->l2tpBuf1State_ == 0) /* real l2tpBuffer1 is not be recorded (might be forged before)*/ &&
						       (l2tp->flags_type == 0) &&
							(l2tp->flags_offset == 0) &&
							(l2tp->flags_sequence == 0) )
						{
							if (pppEntry_t->l2tpBuf1_ == NULL)
							{
								pppEntry_t->l2tpBuf1_ = rtlglue_malloc(RTL8651_MAX_PPTPL2TP_BUFFERSIZE);
								assert(pppEntry_t->l2tpBuf1_);
							}

							if (iplen < RTL8651_MAX_PPTPL2TP_BUFFERSIZE)
							{
								rtl8651_memcpy(pppEntry_t->l2tpBuf1_, ip, iplen);
								pppEntry_t->l2tpOffset_ = (uint32)l2tp - (uint32)ip + 10/* simplest l2tp(6)+ppp(4) */;

								if(l2tp->flags_length)
								{
									pppEntry_t->l2tpOffset_ += 2;
									pppEntry_t->l2tpLengthBit_= 1;
								}else
								{
									pppEntry_t->l2tpLengthBit_ = 0;
								}
								pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_ - 4] = 0xff;
								pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_ - 3] = 0x03;
								pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_ - 2] = 0x00;
								pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_ - 1] = 0x21;

								/* indicate that the real l2tpBuffer1 is recorded */
								pppEntry_t->l2tpBuf1State_ = 1;
							}
						} else if (l2tp->flags_type == 1)	/* control message */
						{
							int32 pureL2tpHdrLen = 8;	/* Mandatory fields : Flags, Ver, Length, TunnelID, SessionID */
						
							assert(l2tp->flags_length == 1);	/* RFC 2661 3.1: Length(L) bit MUST be set to 1 for control messages */

							if (l2tp->flags_sequence)	/* Ns and Nr are present */
							{
								pureL2tpHdrLen += 4;
							}

							if (ntohs(l2tp->length) <= pureL2tpHdrLen)
							{	/* no AVPs are present: might be ZLB-ACK */
							}else
							{	/* AVPs (Attribute, Value, Pairs) are present, parse them */
								int32 totalAVPLengthBudget;
								uint8 *AVP;

								AVP = (uint8 *)((uint32)l2tp + pureL2tpHdrLen);
								/* total length that AVP entries presented */
								totalAVPLengthBudget = ntohs(l2tp->length) - pureL2tpHdrLen;
								assert(totalAVPLengthBudget > 0);

								/* loop to parse AVP */
								while (totalAVPLengthBudget > 0)
								{
									int32 currentAVPLength;
									uint16 attributeType;

									currentAVPLength = ntohs(*(uint16 *)AVP) & 0x3ff;	/* Length: 10 bits */
									assert(currentAVPLength >= 6);	/* M-bit, H-bit, resv, Length, VenderID, Attribute Type */

									attributeType = ntohs(*(uint16 *)((uint32)AVP + 4));/* Attribute Type */

									/* Dispatch different AVPs */
									switch (attributeType)
									{
										case L2TP_AVP_ATTRIBUTE_CTRLMSG:	/* control message */
										{
											/*	RFC.2661 (4.4.1)
											  *	The Attribute Value field for this AVP has the following format:
											  *     0                                       1
											  *     0  1   2  3  4  5   6  7  8  9   0  1   2  3  4  5
											  *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      											  *   |                  Message Type                        |
											  *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
											  */
											if (currentAVPLength == 8)
											{	/* RFC.2661 (4.4.1) The Length of this AVP is 8 */
												uint16 msgType;

												msgType = ntohs(*(uint16 *)((uint32)AVP + 6));/* Attribute Value */

												switch (msgType)
												{
													case L2TP_CTRLMSG_ICCN:
													{	/*	Incoming-Call-Connected:
															System can update tunnelID and sessionID according to the l2tp header.

															Note: copy from [packet content] to [packet buffer], no endian change is needed.
														*/
														if ((pppEntry_t->l2tpBuf1_) && (pppEntry_t->l2tpOffset_ >= 8))
														{
															*(uint16*)&(pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_ - 8]) = l2tp->tunnel_id;
															*(uint16*)&(pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_ - 6]) = l2tp->session_id;
														}else
														{
															uint16 tunnelID = ntohs(l2tp->tunnel_id);
															uint16 sessionID = ntohs(l2tp->session_id);
															ipaddr_t gatewayIp = PKTGET_UINT32_UNALIGNED(&ip->ip_src);
															ipaddr_t serverIp = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);

															_rtl8651_forgeL2tpBuffer(&pppEntry_t->l2tpBuf1_,
																					 &pppEntry_t->l2tpOffset_,
																					 &pppEntry_t->l2tpLengthBit_,
																					 gatewayIp,
																					 serverIp,
																					 tunnelID,
																					 sessionID);
														}
													}
													break;
												}
											}
											
										}
										break;
									}
									/* check next AVP */
									AVP = (uint8 *)((uint32)AVP + currentAVPLength);

									assert(totalAVPLengthBudget >= currentAVPLength);
									totalAVPLengthBudget -= currentAVPLength;
								}
							}
						}
#else /* RTL8651_NEW_L2TPBUFF_FILL_CODE */
						if((pppEntry_t->l2tpOffset_==0)&&(l2tp->flags_type==0)&&(l2tp->flags_offset==0)&&(l2tp->flags_sequence==0)) //copy first data pkt
						{
							//printk("first l2tp packet\n");
							rtl8651_memcpy(pppEntry_t->l2tpBuf1_,ip,iplen);
							pppEntry_t->l2tpOffset_=(uint32)l2tp-(uint32)ip+10; //10: l2tp(6)+ppp(4)
							if(l2tp->flags_length) 
							{
								pppEntry_t->l2tpOffset_+=2;
								pppEntry_t->l2tpLengthBit_=1;
							}
							pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-4]=0xff;
							pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-3]=0x03;
							pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-2]=0x00;
							pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-1]=0x21;						
						}
						
						else if((l2tp->flags_type==1)&&(pppEntry_t->l2tpOffset_)&&(iplen>=28+l2tp->length)) // ip(20)+udp(8)+l2tp_hdr(x)
						{
							uint8 *avp=(uint8 *)(l2tp)+8;
							if(l2tp->flags_sequence) avp+=4;
							//printk("ctrl pkt tunnelid=%d sessionid=%d type=%02x %02x value=%02x %02x\n",l2tp->tunnel_id,l2tp->session_id,avp[4],avp[5],avp[6],avp[7]);							
							if(avp[4]==0&&avp[5]==0&&avp[6]==0) // [4][5]:AVP type=0 
							{
								if(avp[7]==12) //Incoming-Call-Connected
								{
									pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-8]=htons((ntohs(l2tp->tunnel_id)>>8)&0xff);
									pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-7]=htons(ntohs(l2tp->tunnel_id)&0xff);
									pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-6]=htons((ntohs(l2tp->session_id)>>8)&0xff);
									pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-5]=htons(ntohs(l2tp->session_id)&0xff);
									//printk("set new tunnel id\n");
								}
							}
						}
#endif /* RTL8651_NEW_L2TPBUFF_FILL_CODE */
						retval = FWDENG_SEND_GOTO_SEND;
						goto out;
					}
				}
			}// source mac == gateway mac
		} //end l2tp
	}		
	//destPortMask = pkthdrPtr->ph_portlist;
	//goto send;

	retval = FWDENG_SEND_PASSED;
out:
	*pDestPortMask = destPortMask;
	return retval;
}
#endif // endif CONFIG_RTL865X(B)_PPTPL2TP


/*
 * This function is abstracted from _rtl8651_fwdEngineSend().
 * Processing packets that are sent by protocol stack.
 */
int32 _rtl8651_fwdEngineSend_FromPS( int16 dvid, 
                                     struct rtl_pktHdr * pkthdrPtr,
                                     rtl8651_tblDrv_vlanTable_t *local_vlanp,
                                     struct rtl_mBuf *mbuf,
                                     uint32* pDestPortMask )
{
	rtl8651_tblDrv_filterDbTableEntry_t *dal2Entry;
	int32 retval;
	uint32 destPortMask = *pDestPortMask;

	/*pkt is generated by localhost's protocol stack. 
	must lookup local L2 table if unicast pkt
	All headers should already been modified/added, cksum already been recalculated.*/
	struct rtl_mBuf *newMbuf;

	/*when use 8651 loopback port, pkt must be sent to PHY directly if unicast pkt
	when use 8650B ext port, pkt must not send to port 6-8
	*/

	//		#ifdef RTL865X_DEBUG
	//		if(wlan_acc_debug>1) {
	//			uint8 macStr[18];
	//			ether_ntoa_r((ether_addr_t *)&mbuf->m_data[6], macStr);
	//			rtlglue_printf("\nPS Tx:(%d)%s->",pkthdrPtr->ph_rxdesc,macStr);
	//			ether_ntoa_r((ether_addr_t *)&mbuf->m_data[0], macStr);
	//			rtlglue_printf("%s 0x%04x Len%d..extDev=%x\n",macStr,*((uint16 *)(data+12)),pkthdrPtr->ph_len,peripheralExtPortMask );
	//		}
	//		#endif

	if((mbuf->m_data[0]&1)==0){
		//unicast pkt
		dal2Entry=_rtl8651_getVlanFilterDatabaseEntry(dvid, (ether_addr_t *)&mbuf->m_data[0]);
		if(dal2Entry && (dal2Entry->memberPortMask&local_vlanp->memberPortMask)){
			//DA found and is in the egress vlan.

			//shall we send this unicast pkt to peripheral devices?
			if(dal2Entry->memberPortMask&peripheralExtPortMask)
			{
				#ifdef RTL865X_DEBUG
				if(wlan_acc_debug>1) 
					rtlglue_printf("\tPS unicast to link ID %d...\n", dal2Entry->linkId);
				#endif				

				rtlglue_extDeviceSend(local_vlanp->vid, (int32)(dal2Entry->linkId), pkthdrPtr);//would be freed in pci driver
				retval = SUCCESS;
				goto out;
			}

			#ifdef RTL865X_DEBUG
			if(wlan_acc_debug>1) 
				rtlglue_printf("\tPS unicast to port 0x%x  (peripheralExtPortMask=%x)...\n", dal2Entry->memberPortMask,peripheralExtPortMask);
			#endif

			destPortMask =dal2Entry->memberPortMask;
			//unicast to swcore
			retval = FWDENG_SEND_GOTO_TOSWCORE;
			goto out;

		}

		#ifdef RTL865X_DEBUG			
		if(wlan_acc_debug>1)
			rtlglue_printf("\tunknown unicast "); 
		#endif
		//for unknown unicast, send to both sides, fall through
	}

	//broadcast or unknown unicast, 
	//if this VLAN has peripheral device ports, send to both sides...
	if(destPortMask & local_vlanp->memberPortMask & peripheralExtPortMask)
	{
		newMbuf= mBuf_dupPacket(mbuf, MBUF_DONTWAIT);
		if(newMbuf){	
			#ifdef RTL865X_DEBUG
			if(wlan_acc_debug>1) 
				rtlglue_printf("\tPS bcast to extPorts %x->%x",local_vlanp->memberPortMask, peripheralExtPortMask);
			#endif
			
			rtlglue_extDeviceSend(local_vlanp->vid, 0, newMbuf->m_pkthdr);//would be freed in pci driver
		}
	}
	//filter extension ports. 
	destPortMask&=local_vlanp->memberPortMask&RTL8651_PHYSICALPORTMASK;
	//if chip is 8650B, all ports can be bcasted.
	destPortMask &= ~peripheralExtPortMask; //send to all member ports except loopback/ext port.
	
	#ifdef RTL865X_DEBUG
	if(wlan_acc_debug>1) 
		rtlglue_printf("\tPS bcast to %s port 0x%x...\n", local_vlanp->internal?"LAN":"WAN", destPortMask);
	#endif
	pkthdrPtr->ph_srcExtPortNum=0;//Don't send to extension port again.
	retval = FWDENG_SEND_GOTO_SEND;
	/* fall thru */

out:
	*pDestPortMask = destPortMask;
	return retval;
}


/*
 * This function is abstracted from _rtl8651_fwdEngineSend().
 * Processing packets that are sent to extension port.
 */
int32 _rtl8651_fwdEngineSend_ExtDevSend( int16 dvid, 
                                         struct rtl_pktHdr * pkthdrPtr,
                                         rtl8651_tblDrv_vlanTable_t *local_vlanp,
                                         struct rtl_mBuf *mbuf,
                                         struct ip *ip,
                                         uint32* pDestPortMask )
{
	rtl8651_tblDrv_filterDbTableEntry_t *dal2Entry;
	int32 retval;
	uint32 destPortMask = *pDestPortMask;
	uint32 vlanMbrPortMask;
	struct rtl_mBuf *newMbuf;
	uint32 linkId=0;
	//Get link ID

#ifdef RTL865XB_DCUT_SWVLAN
	if (local_vlanp == NULL) {
		rtl8651_tblDrv_swVlanTable_t *swVlanPtr;
		
		// Software vlan table should have this vid if configured.
		swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(dvid)];
		assert(swVlanPtr->valid == 1);
		vlanMbrPortMask = swVlanPtr->memberPortMask;
	}
	else
		vlanMbrPortMask = local_vlanp->memberPortMask;
#else
	vlanMbrPortMask = local_vlanp->memberPortMask;
#endif
	
	dal2Entry=_rtl8651_getVlanFilterDatabaseEntry(dvid, (ether_addr_t *)&mbuf->m_data[0]);
//	if(dal2Entry && (dal2Entry->memberPortMask&local_vlanp->memberPortMask))		// shliu: variable replacement
	if(dal2Entry && (dal2Entry->memberPortMask & vlanMbrPortMask))
		linkId=dal2Entry->linkId;
	//must calculate L3/4 cksum on our own since 8650B xtension port can't do loopback
	if(ip){
		ip->ip_sum=0;
		ip->ip_sum=_rtl8651_ipChecksum(ip);
		if (!IS_FRAGMENT(ip->ip_off))
		{
			switch(ip->ip_p)
			{
				struct tcphdr *tc;
				struct udphdr *ud;
				struct icmp *ic;
				struct igmp *ig;
				case IPPROTO_TCP:
					tc = (struct tcphdr *)L4HDR_ADDR(ip);
					tc->th_sum = 0;
					tc->th_sum = _rtl8651_tcpChecksum(ip);
					break;
				case IPPROTO_UDP:
					ud = (struct udphdr *)L4HDR_ADDR(ip);
					if (ud->uh_sum)
					{
						ud->uh_sum = 0;
						ud->uh_sum = _rtl8651_tcpChecksum(ip);
					}
					break;
				case IPPROTO_ICMP:
					ic = (struct icmp *)L4HDR_ADDR(ip);
					ic->icmp_cksum = 0;
					ic->icmp_cksum = _rtl8651_icmpChecksum(ip);
					break;
				case IPPROTO_IGMP:
					ig = (struct igmp *)L4HDR_ADDR(ip);
					ig->igmp_cksum = 0;
					ig->igmp_cksum = _rtl8651_icmpChecksum(ip);
					break;
			}
		}
	}
	if((destPortMask&RTL8651_PHYSICALPORTMASK)==0){
		//loopback/extension port is the only destination, so no need to copy
		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug>1) 
			rtlglue_printf("\tunicast ");
		#endif
#ifdef RTL865XB_DCUT_SWVLAN
		if (local_vlanp == NULL)
			rtlglue_extDeviceSend(dvid, (int32)linkId, pkthdrPtr);//would be freed in pci driver
		else
			rtlglue_extDeviceSend(local_vlanp->vid, (int32)linkId, pkthdrPtr);//would be freed in pci driver
#else
		rtlglue_extDeviceSend(local_vlanp->vid, (int32)linkId, pkthdrPtr);//would be freed in pci driver
#endif
		retval = SUCCESS;
		goto out;
	}
	//there are more ports so we must copy
	newMbuf= mBuf_dupPacket(mbuf, MBUF_DONTWAIT);
	if(newMbuf){
		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug>1) 
			rtlglue_printf("\tbcast ");
		#endif

#ifdef RTL865XB_DCUT_SWVLAN
		if (local_vlanp == NULL)
			rtlglue_extDeviceSend(dvid, (int32)linkId, newMbuf->m_pkthdr);//would be freed in pci driver
		else
			rtlglue_extDeviceSend(local_vlanp->vid, (int32)linkId, newMbuf->m_pkthdr);//would be freed in pci driver
#else
		rtlglue_extDeviceSend(local_vlanp->vid, (int32)linkId, newMbuf->m_pkthdr);//would be freed in pci driver
#endif		
		//fall through, there are other receivers on switch ports.
	}
	
	//clear extension ports that are destined to peripheral devices (already sent)
	//destPortMask &= ~(destPortMask&peripheralExtPortMask&local_vlanp->memberPortMask);	// shliu: variable replacement
	destPortMask &= ~(destPortMask & peripheralExtPortMask & vlanMbrPortMask);
	destPortMask &= ~(1<<(pkthdrPtr->ph_srcExtPortNum-1+RTL8651_MAC_NUMBER));

	#ifdef RTL865X_DEBUG
	if(wlan_acc_debug>2) 
		rtlglue_printf("\tsend to 0x%04x\n", destPortMask);
	#endif

	retval = FWDENG_SEND_PASSED;
out:
	*pDestPortMask = destPortMask;
	return retval;
}


/*
@func int32		| _rtl8651_fwdEngineSend	| Layer 2  fwdEngine send function
@parm uint32 | property | The packet's property : FWDENG_FROMDRV: from ROMEDRV, FWDENG_FROMPS: Protocol Stack. FWDENG_OFFLOAD: Offload traffic
@parm void * | pkthdr | Pointer to sending packet's pkthdr.
@parm int16 | dvid | packet's destination VLAN id.
@parm int32 | iphdrOffset | IP header oddset. If this value is negative, egress ACL filter would bypass IP ACL rules.
@rvalue FWDENG_ACL_OUT_DROP	| 	Packet is dropped due to egress ACL.
@rvalue SUCCESS	| 	Packet passed egress ACL and sent.
@comm 
Common L2 send function for any packet sending from gateway to network.
If <p iphdrOffset> value is negative, egress IP ACL ruls would be bypassed, otherwise,
egress ACL would check the packet and determine if it could be sent.
 */
__IRAM_L2_FWD int32 _rtl8651_fwdEngineSend(uint32 property, void * data, int16 dvid, int32 iphdrOffset)
{
	struct rtl_pktHdr * pkthdrPtr=(struct rtl_pktHdr *)data;
	struct ip *ip=NULL;
	uint32 destPortMask=pkthdrPtr->ph_portlist;
	uint32 vlanMbrPortMask;
	struct rtl_mBuf *mbuf;
	rtl8651_tblDrv_vlanTable_t *local_vlanp;	
	int32 retval=SUCCESS;
#ifdef RTL865XB_DCUT_SWVLAN
	rtl8651_tblDrv_vlanTable_t *vlanPtr;
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr = NULL;
#endif

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_FWDENG_SEND);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_EGRESS_ACL);
#endif
	

	assert(dvid);	
	assert((destPortMask & rtl8651_tblDrvCpuMirrorPortMask) == 0);
	assert(pkthdrPtr->ph_mbuf);
	mbuf=pkthdrPtr->ph_mbuf;

#ifdef RTL865XB_DCUT_SWVLAN
	vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(dvid)];
	if (vlanPtr->valid == 0 || vlanPtr->vid != dvid) {
		pkthdrPtr->ph_vlanIdx = 0;		// don't care
		local_vlanp = NULL;				
		// Software vlan table should have this vid if configured.
		swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(dvid)];
		assert(swVlanPtr->valid == 1);
		vlanMbrPortMask = swVlanPtr->memberPortMask;
		
		goto to_swcore;
	}
	else {
		pkthdrPtr->ph_vlanIdx= dvid & (RTL8651_VLAN_NUMBER - 1);
		local_vlanp = &DrvTbl.vlan[pkthdrPtr->ph_vlanIdx];
		vlanMbrPortMask = local_vlanp->memberPortMask;
	}
#else
	pkthdrPtr->ph_vlanIdx= dvid & (RTL8651_VLAN_NUMBER-1);
	local_vlanp = &DrvTbl.vlan[pkthdrPtr->ph_vlanIdx];
	vlanMbrPortMask = local_vlanp->memberPortMask;
#endif

//	printk("fwdenginesend porlist=%x pkthdrPtr->ph_rxdesc=%d local_vlanp->vid=%d\n",pkthdrPtr->ph_portlist,pkthdrPtr->ph_rxdesc,local_vlanp->vid);

	if(pkthdrPtr->ph_iphdrOffset)
		ip=(struct ip *)&mbuf->m_data[pkthdrPtr->ph_iphdrOffset];
	else if(iphdrOffset>=0){
		ip=(struct ip *)&mbuf->m_data[iphdrOffset];
	}

	/* Apply egress filter */
	if(rtl8651_drvPacketFastSend == FALSE) {
		retval=_rtl8651_egressAcl(pkthdrPtr->ph_dsid,local_vlanp,dvid,(uint8 *)mbuf->m_data,(uint8 *)ip);
		if(retval&FWDENG_DROP){
//			printk("Egress filter drop. Reason=%d\n",retval );
			return retval;
		}
	}
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_EGRESS_ACL);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_PPTPL2TP_ENCAP);
#endif
	{
		rtl8651_tblDrv_routeTable_t *rt_t;

		rt_t = DrvTbl.route+(rtl8651_tblDrvPara.routingTableSize-1);
		if (rt_t->valid != 0 && rt_t->ifrt_flag==0)
		{
			/*if default route exists and it's PPTP or L2TP,get default session ptr*/
			if (((rt_t->netif_t->linkLayerType == RTL8651_LL_PPTP) || (rt_t->netif_t->linkLayerType == RTL8651_LL_L2TP)) &&
			    (ip) ) 
			{
				retval = _rtl8651_fwdEngineSend_PptpL2tp( pkthdrPtr, local_vlanp, mbuf, ip, rt_t, &destPortMask );
				switch ( retval )
				{
					case FWDENG_SEND_GOTO_SEND:
						goto send;
					case FWDENG_SEND_PASSED:
						break;
					case SUCCESS:
						return SUCCESS;
					default:
						assert( 0 );
						return retval;
				}
			}
		}
	}

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_PPTPL2TP_ENCAP);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_FROM_PS);
#endif

	/* TX from protocol stack */
	if(property == FWDENG_FROMPS)
	{ 
		retval = _rtl8651_fwdEngineSend_FromPS( dvid, pkthdrPtr, local_vlanp, mbuf, &destPortMask );
		switch ( retval )
		{
			case FWDENG_SEND_GOTO_SEND:
				goto send;
			case FWDENG_SEND_GOTO_TOSWCORE:
				goto to_swcore;
			case SUCCESS:
				return SUCCESS;
			default:
				assert( 0 );
				return retval;
		}
	}
	
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_FROM_PS);
#endif

to_swcore:
	
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_EXTDEV_SEND);
#endif
	//destPortMask &= local_vlanp->memberPortMask;  //vlan egress port filter		// shliu: variable replacement
	destPortMask &= vlanMbrPortMask;
	//should this pkt be send to peripheral devices?
//	if(destPortMask&peripheralExtPortMask&local_vlanp->memberPortMask){	// shliu: variable replacement
	if(destPortMask & peripheralExtPortMask & vlanMbrPortMask){
		retval = _rtl8651_fwdEngineSend_ExtDevSend( dvid, pkthdrPtr, local_vlanp, mbuf, ip, &destPortMask );
		switch ( retval )
		{
			case FWDENG_SEND_PASSED:
				break;
			case SUCCESS:
				return SUCCESS;
			default:
				assert( 0 );
				return retval;
		}
	}
	pkthdrPtr->ph_srcExtPortNum=0;
send:
		
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_EXTDEV_SEND);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_FRAG_2ND_HALF);
#endif

	assert((destPortMask&peripheralExtPortMask)==0);
	destPortMask |= rtl8651_tblDrvCpuMirrorPortMask;	
	
	//cfliu: We should check port link status by VLAN since port link status differs from VLAN if SPT is enabled.
	//but use this global variable for now until we turn on STP.
	destPortMask &= phyPortLinkStatus;

	/*
		Spanning tree process need to block all non-BPDU packets
		if the destination port is not in the forwarding state
	*/
	if (	rtl8651_tblDrvSpanningTreeEnable &&
		/* 1. (packet forwarded by ROMEDRV) 2. (normal packet from protocol stack) need spanning tree filtering */
		(property & (FWDENG_FROMDRV|FWDENG_OFFLOAD) || ((pkthdrPtr->ph_pkt_property & PH_PKT_PROPER_STPCTL) == 0)))
	{
		destPortMask &= rtl8651_tblDrvSpanningTreePortFwdStatus;
	}

	/* drop packet if its destination port mask is zero */
	if(destPortMask==0&&pkthdrPtr->ph_srcExtPortNum==0){
		//printk("srcExt:%d, ph_flags:%x, dport=0 dropped!!\n", pkthdrPtr->ph_srcExtPortNum, pkthdrPtr->ph_flags);
		mBuf_freeMbufChain(mbuf);
		return SUCCESS;
	}

#ifdef RTL865XB_DCUT_SWVLAN
	if (local_vlanp == NULL)
	{
		/* handle untag set in software */
		pkthdrPtr->ph_flags &= ~(PKTHDR_VLAN_AUTOADD|CSUM_IP|CSUM_L4);
		pkthdrPtr->ph_vlanTagged = 0;
	} else
	{
		/*update Idle timeout by session*/
		assert(pkthdrPtr->ph_dsid<8);

		{
			uint32 swNicRxIndex;

#ifdef RTL865X_TEST
			swNicRxIndex = 0xffff;	/* we make sure all packet's rxdesc < 0xffff */
#else

#ifdef CONFIG_RTL865XC
			rtlglue_getRingSize(&swNicRxIndex, NULL, 0); /* must modify */
#else
			rtlglue_getRingSize(&swNicRxIndex, NULL);
#endif

#endif

			if(	(pkthdrPtr->ph_rxdesc >= 0) &&
				(pkthdrPtr->ph_rxdesc < swNicRxIndex) &&
				(local_vlanp->internal == FALSE) &&
				(pkthdrPtr->ph_srcExtPortNum == 0))
		{
				_rtl8651_recordTraffic(pkthdrPtr->ph_dsid, TRUE /* to WAN */);
			}
		}
	}
#else
	/*update Idle timeout by session*/
	assert(pkthdrPtr->ph_dsid<8);
	{
		uint32 swNicRxIndex;

#ifdef RTL865X_TEST
			swNicRxIndex = 0xffff;	/* we make sure all packet's rxdesc < 0xffff */
#else

#ifdef CONFIG_RTL865XC
		rtlglue_getRingSize(&swNicRxIndex, NULL, 0); /* must modify */
#else
		rtlglue_getRingSize(&swNicRxIndex, NULL);
#endif

#endif
		if(	(pkthdrPtr->ph_rxdesc >= 0) &&
			(pkthdrPtr->ph_rxdesc < swNicRxIndex) &&
			(local_vlanp->internal == FALSE) &&
			(pkthdrPtr->ph_srcExtPortNum == 0))
		{
			_rtl8651_recordTraffic(pkthdrPtr->ph_dsid, TRUE /* to WAN */);
		}
	}
#endif

	//send pkt here
	#ifdef RTL865X_DEBUG
	if(wlan_acc_debug>2) 
		rtlglue_printf("send to swcore 0x%04x\n", destPortMask);
	#endif

	pkthdrPtr->ph_portlist=destPortMask;

#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
	if ((pkthdrPtr->ph_rxdesc >= 0) && ip)
		_rtl8651_fragmentShortestPathCacheSet_secondHalf(pkthdrPtr, pkthdrPtr->ph_mbuf->m_data, ip);
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_FRAG_2ND_HALF);
#endif
	rtlglue_drvSend((void*)data);

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_FWDENG_SEND);
#endif
	
	return SUCCESS;
}


static int32  _rtl8651_layer2Passthrough(struct rtl_pktHdr *pkthdrPtr,rtl8651_tblDrv_vlanTable_t *local_vlanp, uint8 *m_data, uint16 ethtype){
	/* CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU */
	/* If this is a PPPoE control packet and the dst MAC address is not this gateway,
	 *   then it is a PPPoE passthru packet.
	 */
	if ( ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE && ( ethtype == 0x8863 || ethtype == 0x8864 ) ) ||
	     ( rtl8651_l2protoPassThrough & PASSTHRU_IPV6 && ethtype == 0x86DD ) ||
	     ( rtl8651_l2protoPassThrough & PASSTHRU_IPX && ethtype == 0xE0E0 ) ||
	     ( rtl8651_l2protoPassThrough & PASSTHRU_NETBIOS && ethtype == 0xF0F0 ) )
	{
		ether_addr_t vlanGMAC;
		
		//memDump(pkthdrPtr->ph_mbuf->m_data, 0x40 , "PPPoE 0x8863" );

		// get VLAN Gateway MAC address
		rtl8651_getVlanMacAddress( local_vlanp->vid, &vlanGMAC, NULL );

		// the dst MAC address is NOT GMAC address ==> it is PPPoE passthru packet.
		if ( memcmp( pkthdrPtr->ph_mbuf->m_data, &vlanGMAC, sizeof(ether_addr_t) ) != 0 )
		{
			int32 retval;
			rtl8651_tblDrv_filterDbTableEntry_t *macinfo;
			struct ether_header *ethHdr;
			uint16 srcport;
#if PPPOE_PASSTHRU_NEW_MODEL==1
#else/*PPPOE_PASSTHRU_NEW_MODEL*/
			uint32 i, j;
			uint16 dvid[RTL8651_VLANTBL_SIZE]; /* maybe multiple-value in WAN vlan */
			int32 dvid_count;
			uint16 dportmask;
			
			dvid_count = 0;
#endif/*PPPOE_PASSTHRU_NEW_MODEL*/

			srcport = pkthdrPtr->ph_portlist;
			ethHdr = (struct ether_header*)m_data;
			
			/*
			 * For CDRouter, if this packet is PPPoE Discovery nad the destination MAC is unknown, drop it.
			 */
			if ( rtl8651_drvDropUnknownPppoePADT &&
			     ethtype == 0x8863 )
			{
				pppoeHdr_t *pppoeHdr;
				
				pppoeHdr = (pppoeHdr_t *)(((char*)ethHdr)+sizeof(struct ether_header));

				/*rtlglue_printf("PPPoE Packet: internal=%d code=%02x\n", local_vlanp->internal, pppoeHdr->code );*/
				/* If the packet is from WAN, and is PPPoE Discovery, then verify destination MAC. */
				if ( local_vlanp->internal==FALSE &&
				     pppoeHdr->code==0xa7/*PADT*/ )
				{
					/*rtlglue_printf("PPPoE PADT !!! dMAC %02x:%02x:%02x:%02x:%02x:%02x !!\n",
						ethHdr->ether_dhost[0],
						ethHdr->ether_dhost[1],
						ethHdr->ether_dhost[2],
						ethHdr->ether_dhost[3],
						ethHdr->ether_dhost[4],
						ethHdr->ether_dhost[5]);*/

					/* If the dMAC is not the inside Host, drop it! */
					if ( memcmp( ethHdr->ether_dhost, rtl8651_knownPppoePassthruEthHdr.ether_shost, ETHER_ADDR_LEN ) )
					{
						/*rtlglue_printf("\n**** Bingo! unknown PPPoE PADT !!!\n\n");*/
						return FWDENG_PPPOE8863;
					}
					else
					{
						/*rtlglue_printf("Fine! It is a known PADT, forward it!\n");*/
					}
				}

				/* If the packet is from LAN, remember the source/dest MAC. */
				if ( local_vlanp->internal==TRUE )
				{
					rtl8651_knownPppoePassthruEthHdr = *ethHdr;
					/*rtlglue_printf("learn SMAC %02x:%02x:%02x:%02x:%02x:%02x !!\n",
						rtl8651_knownPppoePassthruEthHdr.ether_shost[0],
						rtl8651_knownPppoePassthruEthHdr.ether_shost[1],
						rtl8651_knownPppoePassthruEthHdr.ether_shost[2],
						rtl8651_knownPppoePassthruEthHdr.ether_shost[3],
						rtl8651_knownPppoePassthruEthHdr.ether_shost[4],
						rtl8651_knownPppoePassthruEthHdr.ether_shost[5] );*/
				}
				
			}

#ifdef RTL865X_TEST
			/* simulate ASIC's behavior.
			 * In test code, we force vid equal to PPPoE Passthru VLAN index.
			 */
			uint16 backupVid = local_vlanp->vid;
			local_vlanp->vid = rtl8651_gPPPoEPassthruVlanID;
#endif/*RTL865X_TEST*/

#if PPPOE_PASSTHRU_NEW_MODEL==1

#if 0/* Louis note: source vlan is modified before fwdInput() */
			/* we assert the source vlan is passthru. */
			assert( local_vlanp->vid == rtl8651_gPPPoEPassthruVlanID ||
			        local_vlanp->vid == rtl8651_gIPv6PassthruVlanID );
			if ( local_vlanp->vid != rtl8651_gPPPoEPassthruVlanID &&
			     local_vlanp->vid != rtl8651_gIPv6PassthruVlanID )
				rtlglue_printf("[%s:%d] local_vlanp->vid=%d, rtl8651_gPPPoEPassthruVlanID=%d, rtl8651_gIPv6PassthruVlanID=%d\n",
				                __FILE__, __LINE__,
				                local_vlanp->vid, rtl8651_gPPPoEPassthruVlanID, rtl8651_gIPv6PassthruVlanID );
#endif/* 0 */

			/* Check if we know where is DMAC? */
			macinfo = _rtl8651_getVlanFilterDatabaseEntry( 0/*vid,no use*/, (struct ether_addr_s*) ethHdr->ether_dhost );
			if ( macinfo == NULL )
			{
				/* We don't know .... broadcast it ! */
				switch ( ethtype )
				{
					case 0x86DD:
						/* IPv6 */
						pkthdrPtr->ph_portlist = ( DrvTbl.vlan[rtl8651_vlanTableIndex(rtl8651_gIPv6PassthruVlanID)].memberPortMask | peripheralExtPortMask ) & (~(1<<srcport));
						break;
						
					case 0xE0E0:
						/* IPX */
						pkthdrPtr->ph_portlist = ( DrvTbl.vlan[rtl8651_vlanTableIndex(rtl8651_gIPXPassthruVlanID)].memberPortMask | peripheralExtPortMask ) & (~(1<<srcport));
						break;
						
					case 0xF0F0:
						/* Netbios */
						pkthdrPtr->ph_portlist = ( DrvTbl.vlan[rtl8651_vlanTableIndex(rtl8651_gNetbiosPassthruVlanID)].memberPortMask | peripheralExtPortMask ) & (~(1<<srcport));
						break;
						
					default:
						/* PPPoE */
						pkthdrPtr->ph_portlist = ( DrvTbl.vlan[rtl8651_vlanTableIndex(rtl8651_gPPPoEPassthruVlanID)].memberPortMask | peripheralExtPortMask ) & (~(1<<srcport));
						break;
				}
			}
			else
			{
				/* rtlglue_printf("macinfo != NULL   macinfo->memberPortMask=%x\n",macinfo->memberPortMask); */
				/* dmac is in L2 table, forward it. */
				pkthdrPtr->ph_portlist = macinfo->memberPortMask & (~(1<<srcport));
			}

			if ( ethtype == 0x8864 )
			{
				pkthdrPtr->ph_pppeTagged = 0; // tell ASIC do not decap PPPoE tag
				pkthdrPtr->ph_flags &= ~(PKTHDR_PPPOE_AUTOADD|CSUM_IP|CSUM_L4);
			}
			
			/* rtlglue_printf( "dvid=%d  pkthdrPtr->ph_portlist=%x\n", 
			                ethtype==0x86DD?rtl8651_gIPv6PassthruVlanID:rtl8651_gPPPoEPassthruVlanID,
			                pkthdrPtr->ph_portlist ); */
			
			retval = _rtl8651_fwdEngineSend( FWDENG_FROMDRV,
							     (void*)pkthdrPtr, 
			                                 ethtype==0x86DD?rtl8651_gIPv6PassthruVlanID:
			                                 ethtype==0xE0E0?rtl8651_gIPXPassthruVlanID:
			                                 ethtype==0xF0F0?rtl8651_gNetbiosPassthruVlanID:
			                                 rtl8651_gPPPoEPassthruVlanID,
			                                 -1 );

			if ( retval != SUCCESS )
			{
				mBuf_freeMbufChain( pkthdrPtr->ph_mbuf );
			}

#else/*PPPOE_PASSTHRU_NEW_MODEL*/

			// find destination Port List
			macinfo = _rtl8651_getVlanFilterDatabaseEntry( 0/*vid,no use*/, (struct ether_addr_s*) ethHdr->ether_dhost );
			if ( macinfo == NULL )
			{
				/* Not in L2 table, broadcast this packet. */
				pkthdrPtr->ph_portlist = 0;
				
				// traverse all VLAN to find out reasonable member port
				for( i = 0;
				     i < RTL8651_VLANTBL_SIZE;
				     i++ )
				{
					if ( DrvTbl.vlan[i].valid )
					{
						pkthdrPtr->ph_portlist |= DrvTbl.vlan[i].memberPortMask;
						/* printf("vid=%d, port:%x, mp=%x\n",DrvTbl.vlan[i].vid, pkthdrPtr->ph_portlist, DrvTbl.vlan[i].memberPortMask); */
					}
				}

				/* except the source port */
				pkthdrPtr->ph_portlist &= ~(1<<srcport);
				/*rtlglue_printf("broadcast: pkthdrPtr->ph_portlist=0x%x\n",pkthdrPtr->ph_portlist);*/
			}
			else
			{
				/* dmac is in L2 table, forward it. */
				pkthdrPtr->ph_portlist = macinfo->memberPortMask;
				/*rtlglue_printf("get L2: pkthdrPtr->ph_portlist=0x%x\n",pkthdrPtr->ph_portlist);*/
			}

			/* find out all possible dvid */
			dportmask = pkthdrPtr->ph_portlist;
			for( i = 0;
			     i < RTL8651_VLANTBL_SIZE;
			     i++ )
			{
				if ( dportmask == 0x0000 ) break;
				if ( DrvTbl.vlan[i].valid &&
				     DrvTbl.vlan[i].vid != local_vlanp->vid )
				{
					if ( dportmask & DrvTbl.vlan[i].memberPortMask )
					{
						dvid[dvid_count++] = DrvTbl.vlan[i].vid;
						dportmask &= ~DrvTbl.vlan[i].memberPortMask;
					}
				}
			}
			if ( dvid_count == 0 )
			{
				rtlglue_printf( "PPPoEPassthru: dvid is not set.\n" );
				
				mBuf_freeMbufChain( pkthdrPtr->ph_mbuf );
				goto no_forward_pppoe_passthru;
#if 0
				dvid[dvid_count++] = 1; /* insert DEFAULT VLAN */
#endif/*0*/
			}

			/* forward packet to the destination vid */
			for( i = 0;
			     i < RTL8651_VLANTBL_SIZE;
			     i++ )
			{
				if ( DrvTbl.vlan[i].valid &&
				     DrvTbl.vlan[i].vid != local_vlanp->vid  )
				{
					/* rtlglue_printf("DrvTbl.vlan[i].vid=%d  pkthdrPtr->ph_portlist=%x\n",DrvTbl.vlan[i].vid,pkthdrPtr->ph_portlist); */
					for( j = 0; j < dvid_count; j++ )
					{
						if ( dvid[j] == DrvTbl.vlan[i].vid )
						{
							struct rtl_mBuf *dup_mBuf = NULL;
							
							/* final dvid is no need to duplicate */
							if ( j != dvid_count-1 )
								dup_mBuf = mBuf_dupPacket( pkthdrPtr->ph_mbuf, MBUF_DONTWAIT );

							/* forward it */
							pkthdrPtr->ph_portlist = DrvTbl.vlan[i].memberPortMask & ~(1<<srcport);
							if ( ethtype == 0x8864 )
							{
								pkthdrPtr->ph_pppeTagged = 0; // tell ASIC do not decap PPPoE tag
								pkthdrPtr->ph_flags &= ~(PKTHDR_PPPOE_AUTOADD|CSUM_IP|CSUM_L4);
							}
							/*rtlglue_printf( "Forward PPPoE! vid=%d, dvid=%d, portlist=%x\n", local_vlanp->vid, dvid[j], pkthdrPtr->ph_portlist );*/
							retval = _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdrPtr, dvid[j], -1 );

							if ( retval == SUCCESS )
							{
								/* DO NOT forward the packet again if has been send to those ports */
								if ( dup_mBuf )
									dup_mBuf->m_pkthdr->ph_portlist &= ~DrvTbl.vlan[i].memberPortMask;
							}
							else
							{
								/* Forwarding Engine send error, free pkthdrPtr */
								mBuf_freeMbufChain( pkthdrPtr->ph_mbuf );
							}

							/* header has been freed (no matter success or failed),
							   recover with duplicated packet. */
							if ( dup_mBuf ) 
								pkthdrPtr = dup_mBuf->m_pkthdr;
							else 
								pkthdrPtr = NULL;
						}
					}
				}
			}

			/* In the final, packet shoule already be freed. */
			assert( pkthdrPtr == NULL );
no_forward_pppoe_passthru:
#endif/*PPPOE_PASSTHRU_NEW_MODEL*/

#ifdef RTL865X_TEST
			/* Recover local_vlanp->vid */
			local_vlanp->vid = backupVid;
#endif/*RTL865X_TEST*/

			return SUCCESS;
		}
		else
		{
			/* rtlglue_printf("PPPoE/IPv6 Packet for gateway...\n"); */
			/* _modifyPPPoEPassthruVid( pkthdrPtr, ethtype, local_vlanp ); */
		}
	}
	return FAILED;
}

#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
static int32 _rtl8651_pptpL2tpInput(int32 inbound, struct rtl_pktHdr *pkthdrPtr,rtl8651_tblDrv_vlanTable_t *local_vlanp, uint8 *m_data, uint16 ethtype, rtl8651_tblDrv_pppTable_t **ppp){
	rtl8651_tblDrv_routeTable_t *rt_t=NULL;
	rtl8651_tblDrv_pppTable_t *pppEntry_t=NULL;	

	rt_t = DrvTbl.route+(rtl8651_tblDrvPara.routingTableSize-1);
	if (rt_t->valid != 0 && rt_t->ifrt_flag==0)
	{
		if ((rt_t->netif_t->linkLayerType == RTL8651_LL_PPTP) || (rt_t->netif_t->linkLayerType == RTL8651_LL_L2TP))
		{
#if 1
			pppEntry_t=rt_t->rt_nhSession_t;
#else
			pppEntry_t = (rtl8651_tblDrv_pppTable_t *)rt_t->rt_nhstart_t->nexthop_t;
#endif
		}
	}
	
	*ppp=pppEntry_t;

	//printk("input local_vlanp->vid=%d \n",local_vlanp->vid );
	// decapsulating PPTP or L2TP	
	if(inbound==TRUE)//if(local_vlanp->vid==tunnel.wanVid)
	{
		struct rtl_mBuf *mbufPtr=pkthdrPtr->ph_mbuf;

		if((pppEntry_t)&&(ethtype==0x0800))
		{
			struct ip *ip=NULL,*ip2=NULL;
			
			if(pkthdrPtr->ph_iphdrOffset)
				ip=(struct ip *)&m_data[pkthdrPtr->ph_iphdrOffset];
			
			if(ip)
			{
				if(rt_t->netif_t->linkLayerType == RTL8651_LL_PPTP)
				{
					//printk("pptp decap iplen=%d ip->ip_p=%d\n",ntohs(ip->ip_len),ip->ip_p);
					if((ip->ip_p==IPPROTO_GRE)&&(ntohs(ip->ip_len)>=28)) 
					{

						struct wan_gre_header *gre;
//						uint32 temp,i;
						uint16 iphlen = IPHDR_LEN(ip);
						uint16 offset=iphlen+8,ip2len=0; // gre header:(8)
						uint8 *ptr;
						uint32 gre_seq, gre_ack;
						uint8 gre_type;
				
						gre = (struct wan_gre_header *) ((int8 *) ip + iphlen);
						gre_seq = 0;
						gre_ack = 0;
						gre_type = 0;

						if(gre->flags&16) //GRE SEQ
						{
							offset+=4;
							pppEntry_t->lastReceiveSeq_= PKTGET_UINT32_UNALIGNED(&gre->seq);
							gre_seq = pppEntry_t->lastReceiveSeq_;
							gre_type |= RTL8651_PPTPHDR_SEQ_SET;
							//printk("receive GRE seq=%d",pppEntry_t->lastReceiveSeq_);						

							if(gre->ver&128) // GRE ACK
							{
								offset+=4;
								gre_ack = PKTGET_UINT32_UNALIGNED(&gre->ack);
								gre_type |= RTL8651_PPTPHDR_ACK_SET;
								//printk(", ack=%d",temp);						
								PKTSET_UINT32_UNALIGNED(gre_ack - pppEntry_t->delta_, &gre->ack);
							}
							//printk("\n");
						}
						else if(gre->ver&128) // GRE ACK(at seq address)
						{
							offset+=4;
							gre_ack = PKTGET_UINT32_UNALIGNED(&gre->seq);
							gre_type |= RTL8651_PPTPHDR_ACK_SET;
							PKTSET_UINT32_UNALIGNED(gre_ack - pppEntry_t->delta_,&gre->seq);						
							//printk("receive GRE ack=%d",temp);
						}
					
						ptr=(uint8 *)ip;	
						
						if((pppEntry_t->initiated)&&(gre->payload_len))
						{
							/*
								chenyl (2005/10/27):

								Based on RFC 1662,chap 3.1, the PPP frame format is:

									+----------+----------+----------+
									|      Flag       |   Address   |     Control   |
									|   01111110  |  11111111 |   00000011  |
									+----------+----------+----------+
									+----------+-------------+---------+
									|    Protocol   |   Information    |  Padding  |
									|   8/16 bits  |           *           |       *       |
									+----------+-------------+---------+
									+----------+----------+-----------------
									|       FCS      |       Flag     | Inter-frame Fill
									|   16/32 bits |   01111110 | or next Address
									+----------+----------+-----------------

								Because the Flag field is used between two frames, so it would not
								exist in PPTP case (pleae refer RFC2637,chap4, and find "No HDLC flags").
								Therefore, the format would become:

								+----------+----------+----------+-------------
								|   Address   |     Control   |    Protocol   |   Information
								|  11111111 |   00000011  |   8/16 bits  |           *
								+----------+----------+----------+-------------

								Besides, based on RFC 1662,chap 3.2, we can do the
								"Address-and-Control-Field-Compression" in PPTP case.
								In this case, the Address and Control fields can be simply omitted,
								and the PPP header would become:

								+----------+-------------
								|    Protocol   |   Information
								|   8/16 bits  |           *
								+----------+-------------


								==> We need to check 4 types of PPP encapsulation:

									0xff 0x03 0x00 0x21
									0xff 0x03 0x21
									0x00 0x21
									0x21

							*/
							if ((ptr[offset] == 0xff) && (ptr[offset + 1] == 0x03))
							{	/* RFC 1662 3.1, Address & Control Field */
								offset += 2;
							}

							if (	((ptr[offset] == 0x00) && (ptr[offset + 1] == 0x21) /* 16 bits case */) ||
								(ptr[offset] == 0x21 /* 8 bits case */))
							{
								/* decapsulation */
								if(ptr[offset]==0x21)
								{	/* 8 bits case */
									offset+=1;
								} else
								{	/* 16 bits case */
									offset += 2;
								}

								ip2=(struct ip *)(ptr+offset);
								ip2len = ntohs(ip2->ip_len);

								if(ip2len<=local_vlanp->mtu)
								{	/* before PPTP decapsulation, we record its header, sequence and ack */
									{
										/* save sequence and ack */
										uint16 id = ntohs(ip2->ip_id);
										uint16 offset = ntohs(ip2->ip_off);
										_rtl8651_pptpHdrCache_Add(id, offset, gre_seq, gre_ack, gre_type);

										/* save header */
										if (pppEntry_t->greOffset2_ == 0)
										{
											/* allocate buffer if it does not exist */
											if (pppEntry_t->greBuf2_ == NULL)
											{
												assert(((uint32)ip2 - (uint32)ip) <= RTL8651_MAX_PPTPL2TP_BUFFERSIZE);
												pppEntry_t->greBuf2_ = rtlglue_malloc(RTL8651_MAX_PPTPL2TP_BUFFERSIZE);	/* all pptp header length is less than RTL8651_MAX_PPTPL2TP_BUFFERSIZE */
											}
											/* copy the header */
											if (pppEntry_t->greBuf2_)
											{
												pppEntry_t->greOffset2_ = ((uint32)ip2 - (uint32)ip);
												rtl8651_memcpy(pppEntry_t->greBuf2_, (void *)ip, pppEntry_t->greOffset2_);
											}
										}else if (pppEntry_t->greBuf2_)
										{	/* we always copy the call ID */
											struct ip *ipBuf;
											struct wan_gre_header *greBuf;
											ipBuf = (struct ip *)(pppEntry_t->greBuf2_);
											greBuf = (struct wan_gre_header *)L4HDR_ADDR(ipBuf);
											greBuf->call_id = gre->call_id;
										}
									}
									//printk("decapsulating\n");
									rtl8651_memcpy((uint8 *)(ip2)-14,m_data,14);
									m_data=(uint8 *)(ip2)-14;

									{	/* chenyl: patch for multicast forward */
										ipaddr_t dest_ip = PKTGET_UINT32_UNALIGNED(&ip2->ip_dst);
										if (IN_MULTICAST(dest_ip))
										{
											MCAST_IP2MAC(dest_ip, m_data);
										}
										/* chenyl: patch for broadcast forward */
										if (dest_ip == 0xffffffff)
										{
											BCAST_MAC(m_data);
										}
									}

									pkthdrPtr->ph_flags&=~CSUM_L4;
									pkthdrPtr->ph_flags&=~CSUM_IP;

									pkthdrPtr->ph_len -= offset;
									mbufPtr->m_len -= offset;
									mbufPtr->m_data = m_data;

									if (_rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdrPtr,tunnel.loopBackVid,-1) != SUCCESS)
									{
										mBuf_freeMbufChain(pkthdrPtr->ph_mbuf);
									}
									return SUCCESS;
								}
							}
						}
					}
				}
				else if(rt_t->netif_t->linkLayerType == RTL8651_LL_L2TP)
				{
					rtl8651_tblDrv_vlanTable_t *extVlan = &DrvTbl.vlan[(tunnel.loopBackVid & (RTL8651_VLAN_NUMBER-1))];
					assert(extVlan->valid);

					if((ip->ip_p==IPPROTO_UDP)&&(ntohs(ip->ip_len)>=35)&&(ntohs(ip->ip_len)<=extVlan->mtu)) //35:ip(20)+udp(8)+l2tp(6)+ppp(1)
					{
						struct udphdr *udp;
						udp = (struct udphdr *)L4HDR_ADDR(ip);
						if(ntohs(udp->uh_dport)==1701)
						{
							struct wan_l2tp_header *l2tp=(struct wan_l2tp_header*)(udp+1);

							if (l2tp->flags_type==0) /* DATA channel */
							{
								uint8 *u8p_l2tp = (uint8*)l2tp;
								struct ip *ip2=NULL;
								uint32 l2tpPppChkLen = 6 /* L2TP's least header length is 6 */; /* trace of the length of L2TP+PPP header */

								if (l2tp->flags_length)
								{
									l2tpPppChkLen += 2;
								}
								if (l2tp->flags_offset)
								{
									l2tpPppChkLen += 2;
								}

								/* for PPP header format, please refer the code of PPTP shown above */
								if ((u8p_l2tp[l2tpPppChkLen] == 0xff) && (u8p_l2tp[l2tpPppChkLen + 1] == 0x03))
								{	/* RFC 1662 3.1, Address & Control Field */
									l2tpPppChkLen += 2;
								}
								if (u8p_l2tp[l2tpPppChkLen] == 0x21)
								{	/* 8 bits case */
									ip2 = (struct ip *)((uint32)u8p_l2tp + l2tpPppChkLen + 1);
								} else if ((u8p_l2tp[l2tpPppChkLen] == 0x00) && (u8p_l2tp[l2tpPppChkLen + 1] == 0x21))
								{	/* 16 bits case */
									ip2 = (struct ip *)((uint32)u8p_l2tp + l2tpPppChkLen + 2);
								}
								if(ip2)
								{
									uint16 ip2len=ntohs(ip2->ip_len);
								
									if(ip2len<=extVlan->mtu)
									{
										/* before L2TP decapsulation, we record its header */
										if (pppEntry_t->l2tpOffset2_ == 0)
										{
											/* allocate buffer if it does not exist */
											if (pppEntry_t->l2tpBuf2_ == NULL)
											{
												assert(((uint32)ip2 - (uint32)ip) <= RTL8651_MAX_PPTPL2TP_BUFFERSIZE);
												pppEntry_t->l2tpBuf2_ = rtlglue_malloc(RTL8651_MAX_PPTPL2TP_BUFFERSIZE);	/* all l2tp header length is less than RTL8651_MAX_PPTPL2TP_BUFFERSIZE */
											}
											/* copy the header */
											if (pppEntry_t->l2tpBuf2_)
											{
												pppEntry_t->l2tpOffset2_ = ((uint32)ip2 - (uint32)ip);
												rtl8651_memcpy(pppEntry_t->l2tpBuf2_, (void *)ip, pppEntry_t->l2tpOffset2_);
											}
										}

										switch(ip2->ip_p)
										{
											case IPPROTO_ICMP:
												pkthdrPtr->ph_proto = PKTHDR_ICMP;												
												break;
											case IPPROTO_IGMP:
												pkthdrPtr->ph_proto = PKTHDR_IGMP;
												break;
											case IPPROTO_TCP:									
												pkthdrPtr->ph_proto = PKTHDR_TCP;
												break;
											case IPPROTO_UDP:
												pkthdrPtr->ph_proto = PKTHDR_UDP;
												break;										
										}

										pkthdrPtr->ph_flags&=~CSUM_L4;
										pkthdrPtr->ph_flags&=~CSUM_IP;

										//decapsulate
										rtl8651_memcpy((uint8 *)(ip2)-14,m_data,14);
										m_data=(uint8 *)(ip2)-14;

										{	/* chenyl: patch for multicast forward */
											ipaddr_t dest_ip = PKTGET_UINT32_UNALIGNED(&ip2->ip_dst);
											if (IN_MULTICAST(dest_ip))
											{
												MCAST_IP2MAC(dest_ip, m_data);
											}
											/* chenyl: patch for broadcast forward */
											if (dest_ip == 0xffffffff)
											{
												BCAST_MAC(m_data);
											}
										}

										pkthdrPtr->ph_len -= ((uint32)ip2-(uint32)ip);
										mbufPtr->m_len -= ((uint32)ip2-(uint32)ip);
										mbufPtr->m_data = m_data;

										if (_rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdrPtr,tunnel.loopBackVid,-1) != SUCCESS)
										{
											mBuf_freeMbufChain(pkthdrPtr->ph_mbuf);
										}
										return SUCCESS;
									}
								} //ip2->ip_len boundary check
							}// flags_type==0
/*							
							else //Ctrl Channel
							{
								if((pppEntry_t->initiated)&&(l2tp->flags_length)&&(ntohs(ip->ip_len)>=28+ntohs(l2tp->length)))  // ip(20)+udp(8)+l2tp_hdr(x)
								{
									uint8 *avp=(uint8 *)(l2tp)+8;
									if(l2tp->flags_sequence) avp+=4;
									 printk("ctrl pkt tunnelid=%d sessionid=%d type=%02x %02x value=%02x %02x\n",l2tp->tunnel_id,l2tp->session_id,avp[4],avp[5],avp[6],avp[7]);
									if(avp[4]==0&&avp[5]==0&&avp[6]==0) // [4][5]:AVP type=0 
									{
										if(avp[7]==2) //SCCRP(Start-Control-Connection-Reply)
										{
											pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-8]=htons((ntohs(l2tp->tunnel_id)>>8)&0xff);
											pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-7]=htons(ntohs(l2tp->tunnel_id)&0xff);
											//printk("set new tunnel id\n");
										}
										if(avp[7]==11) //ICRP(Incoming-Call-Reply)
										{
											pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-6]=htons((ntohs(l2tp->session_id)>>8)&0xff);
											pppEntry_t->l2tpBuf1_[pppEntry_t->l2tpOffset_-5]=htons(ntohs(l2tp->session_id)&0xff);
											//printk("set new session id\n");
										}								
									}								
								}
							}//end ctrl channel
*/							
						}//end udpport=1701
					}
				}
			}
		}
	}
	else if((pppEntry_t) && (pppEntry_t->initiated))
	{
		//DMAC = PPTP server , LAN -> WAN
		//printk("match 2050 vlan. send to fwdeng\n");

		/* FIXME: for Multiple WAN configuration(ex. multiple PPTP...etc), we check source mac when destination mac is mutlicast mac will cause wrong output! */
		if((memcmp(pppEntry_t->macInfo->macAddr.octet,m_data,6)==0) ||
			(m_data[0]==0x01 && m_data[1]==0x00 && m_data[2]==0x5e && memcmp(pppEntry_t->macInfo->macAddr.octet, &m_data[6], 6)!=0)){
			//printk("LAN->ext\n");
			
			return _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdrPtr,tunnel.wanVid,-1); //LAN->ext WAN
		}else{
			//printk("WAN->ext\n");
			//WAN->ext WAN, DA is always WAN's GMAC
		}
	}
	return FAILED;
}
#endif //CONFIG_RTL865X(B)_PPTPL2TP

static int32 _rtl8651_l2switch(struct rtl_pktHdr *pkthdrPtr,rtl8651_tblDrv_vlanTable_t *local_vlanp, uint8 *m_data, uint16 ethtype);

/* From swcore to protocol stack or extension port(PCI/WLAN) */
__IRAM_L2_FWD int32 rtl8651_fwdEngineL2Input(struct rtl_pktHdr *pkthdrPtr,rtl8651_tblDrv_vlanTable_t *local_vlanp, uint8 *m_data, uint16 ethtype)
{
	int32 compare;

#ifdef RTL865XB_DCUT_SWVLAN
	if (local_vlanp == NULL) {
		#if 0
		rtl8651_tblDrv_swVlanTable_t *swVlanPtr;
		// Because "local_vlanp" is NULL, we are sure that this packet is vlan-tagged (with unknown vid to ASIC).
		uint16 vid = ntohs(*((uint16 *)(m_data + 14))) & 0x0fff;
		
		// Software vlan table should have this vid if configured.
		swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
		if (swVlanPtr->valid == 0)
			return FWDENG_VLANINVALID;
		#endif
		return _rtl8651_l2switch(pkthdrPtr, local_vlanp, m_data, ethtype);
	}
#endif

	if (rtl8651_l2protoPassThrough)
	{
		int32 ret=_rtl8651_layer2Passthrough(pkthdrPtr,local_vlanp, m_data, ethtype);

		if(ret!=FAILED)
		{
			return ret;
	}
	}

#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
	{
		rtl8651_tblDrv_pppTable_t *pppEntry_t=NULL;	
		if (local_vlanp->vid==tunnel.loopBackVid)
		{
			int32 ret=_rtl8651_pptpL2tpInput(FALSE/*outbound*/, pkthdrPtr,local_vlanp, m_data, ethtype, &pppEntry_t);

			if(ret!=FAILED)
			{
				return ret;
			}
		}else if (local_vlanp->vid==tunnel.wanVid)
		{
			int32 ret=_rtl8651_pptpL2tpInput(TRUE/*inbound*/, pkthdrPtr,local_vlanp, m_data, ethtype, &pppEntry_t);

			if(ret!=FAILED)
			{
				return ret;
		}
	}
	}
#endif
	
	#ifdef RTL865X_DEBUG
	if (wlan_acc_debug>1)
	{
		rtlglue_printf("\n(%d)", pkthdrPtr->ph_rxdesc);
		if(wlan_acc_debug>2)
		{
			_rtl8651_layer2fwd_debug(4, pkthdrPtr);
	}
	}
	#endif

	/*
		If this packet is processed by ASIC and ASIC has decided this packet would be sent to a specific Extension port,
		we forward it directly to the extension device which registered to ROMEDRV.
	*/
	/* ps.Edward says support only one MAC is ok, although 8650 supports more than 1 mac on each vlan..04/02/19 */
	if (	((1<<pkthdrPtr->ph_portlist) & peripheralExtPortMask)	/* To Peripheral Extension Port */ &&
		(pkthdrPtr->ph_extPortList < 8)							/* It's not to CPU port (b'1000), it means ASIC has done the FWD process */ )
	{
		/* SA=Gateway MAC???  */

			/* At least one external peripheral device exists on loopback/ext port */
			uint32 linkId=0;	/* 0: default broadcast. */

			rtl8651_tblDrv_filterDbTableEntry_t * dal2Entry;
			/*
			  * pkt already translated by switch. Send to PCI directly.
			  * lookup L2 table to see which WDS link this MAC belongs to.
			  * if DA not found, keep linkId=0. This would broadcast pkt to all WDS links.
			  * dal2Entry=_rtl8651_getVlanFilterDatabaseEntry(vid, (ether_addr_t *)&m_data[0]);
			  */
			SLIST_FOREACH(dal2Entry, &(DrvTbl.filterDB[0].database[rtl8651_filterDbIndex((ether_addr_t *)m_data)]), nextFDB)
			{
				if (memcmp(&dal2Entry->macAddr, m_data, sizeof(ether_addr_t)) == 0)
				{
					break;
				}
			}
			if (dal2Entry)
			{
				linkId=dal2Entry->linkId;
			}
			rtlglue_extDeviceSend(local_vlanp->vid, (int32)linkId, pkthdrPtr);	/* pkt would be freed in pci driver */
 
			return SUCCESS;
		}

	/* DA=Gateway MAC??? */
	compare = memcmp(&m_data[0], &local_vlanp->macAddr, sizeof(ether_addr_t));	/* check DMAC with base MAC */
	if (compare == 0)
	{	/* DMAC is gateway MAC. This is normal case if VLAN has only one MAC */
		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug>1)
		{
			rtlglue_printf("to myself\n");
		}
		#endif

		return FAILED;	/* to myself....return to fwd engine. */
	} else
	{
		/* L2 bridging or L2 broadcast pkt process. slow path */
		return _rtl8651_l2switch(pkthdrPtr,local_vlanp,m_data, ethtype);
	}
}

static int32 _rtl8651_l2PhysicalPortRelay(struct rtl_mBuf *inMbufPtr)
{
	struct rtl_pktHdr *phdr;

	phdr = inMbufPtr->m_pkthdr;

#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
	if (DrvTbl.vlan[phdr->ph_vlanIdx].vid == tunnel.loopBackVid)
	{
		/*	chenyl:
			    If multicast packet is from PPTP/L2TP loopBack VLAN, It means there is no
			    other member port in this VLAN and we CAN NOT call fwdEngineSend to do
			    L2 forwarding														*/
	} else
#endif
	{

		struct rtl_mBuf *mbufPtr;
	
		mbufPtr = mBuf_dupPacket(inMbufPtr, MBUF_DONTWAIT);

		if(mbufPtr == NULL)
		{
			#ifdef RTL865X_DEBUG
			rtlglue_printf("No more mbufs for relaying brdcast pkt\n");
			#endif
			return FAILED;
		}

		/* disable L3/4 checksum */
		mbufPtr->m_pkthdr->ph_flags &= (~(CSUM_IP | CSUM_L4));
		/* relay to all member ports except the incoming port */
		mbufPtr->m_pkthdr->ph_portlist = DrvTbl.vlan[phdr->ph_vlanIdx].memberPortUpStatus & ~(1<<phdr->ph_portlist);
		mbufPtr->m_pkthdr->ph_portlist &=~rtl8651_allExtPortMask; /* filter Extension devices */
		if (_rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)mbufPtr->m_pkthdr, DrvTbl.vlan[phdr->ph_vlanIdx].vid, -1) != SUCCESS)
		{
			mBuf_driverFreeMbufChain(mbufPtr);
		}
	}

	return SUCCESS;
}


 int32 _rtl8651_l2switch(struct rtl_pktHdr *pkthdrPtr,rtl8651_tblDrv_vlanTable_t *local_vlanp, uint8 *m_data, uint16 ethtype){
	uint16 vid;
//	uint32 portmask;
	uint32 vlanMbrPortMask;
	struct rtl_mBuf *mbufPtr;	
	rtl8651_tblDrv_filterDbTableEntry_t *dal2Entry=NULL;		
	struct rtl_mBuf *newMbuf;
#ifdef RTL865XB_DCUT_SWVLAN
	rtl8651_tblDrv_swVlanTable_t *swVlanPtr = NULL;
	struct rtl_mBuf *newMbuf2;
#endif

	//...incoming port maybe any switch port, including the loopback port

#ifdef RTL865XB_DCUT_SWVLAN
	if (local_vlanp == NULL) {		
		// Because "local_vlanp" is NULL, we are sure that this packet is vlan-tagged (with unknown vid to ASIC).
		vid = ntohs(*((uint16 *)(m_data + 14))) & 0x0fff;
		// rtlglue_printf("In %s, vid: %d\n", __FUNCTION__, vid);

		// Software vlan table should have this vid if configured.
		swVlanPtr = &DrvTbl.swVlan[rtl8651_swVlanTableIndex(vid)];
		if (swVlanPtr->valid == 0)
			return FWDENG_VLANINVALID;
		vlanMbrPortMask = swVlanPtr->memberPortMask;
		//rtlglue_printf("vlanMbrPortMask: %d\n", vlanMbrPortMask);
	}
	else {
		vid = local_vlanp->vid;
		vlanMbrPortMask = local_vlanp->memberPortMask;
	}
#else
	vid = local_vlanp->vid;
	vlanMbrPortMask = local_vlanp->memberPortMask;
#endif

//	portmask = 1<<pkthdrPtr->ph_portlist;
	mbufPtr=pkthdrPtr->ph_mbuf;	

	//DMAC is not gateway MAC. so this is a pure L2 pkt.
	//L2 unicast or bcast pkt from
	// a)switch port to other ports(not to GW). src port is port 0~4, but not loopback port
	// b) PCI to other switch port, but trapped to CPU due to URL filter. Just do L2 forward.
	//  1. L2 unicast to other switch ports in same VLAN.
	//  2. L2 unicast to PCI/WLAN port if in same VLAN.
	//  3. 1st unknown unicast pkt: send to stack and PCI/WLAN if on same vlan
	//  4. bcast-addr pkt: send to stack and PCI/WLAN if on same vlan
	//Pure L2 pkt. Don't touch higher layer cksum field.
	if( (m_data[0]&1)==0){
		//unicast pkt. DA lookup
		#if 0
		rtlglue_printf("dmac lookup: %x.%x.%x.%x.%x.%x\n", 
			((ether_addr_t *)&m_data[0])->octet[0], ((ether_addr_t *)&m_data[0])->octet[1],
			((ether_addr_t *)&m_data[0])->octet[2], ((ether_addr_t *)&m_data[0])->octet[3],
			((ether_addr_t *)&m_data[0])->octet[4], ((ether_addr_t *)&m_data[0])->octet[5]);
		#endif
		dal2Entry=_rtl8651_getVlanFilterDatabaseEntry(vid, (ether_addr_t *)&m_data[0]);		

		if(dal2Entry){
			//rtlglue_printf("dal2Entry->memberPortMask: %d\n", dal2Entry->memberPortMask);
			if(dal2Entry->memberPortMask & vlanMbrPortMask) {
				
				if((dal2Entry->memberPortMask&peripheralExtPortMask)==0)
				{
					//Case 1: DA found on switch's non-loopback/non-extension port. 
					#ifdef RTL865X_DEBUG
					if(wlan_acc_debug>1)
						#ifdef RTL865XB_DCUT_SWVLAN
							{
								if (local_vlanp == NULL)
									rtlglue_printf("L2 unicast to vlan %d\n", vid);
								else
									rtlglue_printf("L2 unicast to %s\n", local_vlanp->internal?"LAN":"WAN");
							}
						#else
							rtlglue_printf("L2 unicast to %s\n", local_vlanp->internal?"LAN":"WAN");
						#endif
					#endif

#ifdef RTL865XB_DCUT_SWVLAN
					if (local_vlanp == NULL) {
						/* ========= smac learning =========== */
						_rtl8651_addVlanRefFilterDatabaseEntry(RTL8651_L2_TYPEI, swVlanPtr->fid, 
							vid, (ether_addr_t *)(m_data + 6), RTL8651_FORWARD_MAC, 1 << pkthdrPtr->ph_portlist);
					}
#endif					
					mbufPtr->m_pkthdr->ph_portlist = dal2Entry->memberPortMask & ~(1<<mbufPtr->m_pkthdr->ph_portlist);
					pkthdrPtr->ph_iphdrOffset=0;
					//LAN/WAN L2 switching, null ip header to stop egress ACL.
					//Pure L2 pkt. Don't touch higher layer cksum field.
					pkthdrPtr->ph_flags &= ~(CSUM_IP| CSUM_L4);

#ifdef RTL865XB_DCUT_SWVLAN
					if (local_vlanp == NULL) {
						/* ========= new mbuf for tag-set port =========*/
						newMbuf = mBuf_dupPacket(mbufPtr, MBUF_DONTWAIT);
						if (newMbuf) {
							newMbuf->m_pkthdr->ph_portlist &= ~(swVlanPtr->untagPortMask);
							_rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void *)newMbuf->m_pkthdr, vid, -1);
						}

						/* ========= mbuf for untag-set port =========*/
						mbufPtr->m_pkthdr->ph_portlist &= swVlanPtr->untagPortMask;
						_rtl8651_rxPktPreprocess_stripVlan(mbufPtr->m_pkthdr);
						return _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void *)pkthdrPtr, vid, -1);
					}
					else
						return _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void *)pkthdrPtr, DrvTbl.vlan[pkthdrPtr->ph_vlanIdx].vid, -1);
#else
					return _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdrPtr,DrvTbl.vlan[pkthdrPtr->ph_vlanIdx].vid,-1 );
#endif
				}
				if(peripheralExtPortMask){
					//case 2: DA on PCI/WAN
					if((1<<pkthdrPtr->ph_portlist)&peripheralExtPortMask)//Loopback, drop it.
						goto drop;

					#ifdef RTL865X_DEBUG
					rtlglue_printf("L2 unicast to PCI,");
					#endif /* RTL865X_DEBUG */

#ifdef RTL865XB_DCUT_SWVLAN
					if (local_vlanp == NULL) {
						/* ========= new mbuf for tag-set port =========*/
						newMbuf = mBuf_dupPacket(mbufPtr, MBUF_DONTWAIT);
						if (newMbuf) {
							newMbuf->m_pkthdr->ph_portlist &= ~(swVlanPtr->untagPortMask);
							rtlglue_extDeviceSend(vid, (int32)(dal2Entry->linkId), newMbuf->m_pkthdr);//would be freed in pci driver
						}
						/* ========= mbuf for untag-set port =========*/
						mbufPtr->m_pkthdr->ph_portlist &= swVlanPtr->untagPortMask;
						_rtl8651_rxPktPreprocess_stripVlan(mbufPtr->m_pkthdr);
						rtlglue_extDeviceSend(vid, (int32)(dal2Entry->linkId), pkthdrPtr);//would be freed in pci driver
					}
					else
						rtlglue_extDeviceSend(vid, (int32)(dal2Entry->linkId), pkthdrPtr);//would be freed in pci driver
#else
					rtlglue_extDeviceSend(vid, (int32)(dal2Entry->linkId), pkthdrPtr);//would be freed in pci driver
#endif
					return SUCCESS;
				}
				assert(0); //should not come here
			}
		}
	}
	else if (m_data[0] & 1)
	{	/* some special broadcast mac need further process */
#ifdef RTL865XB_DCUT_SWVLAN
		if (local_vlanp == NULL) 
			goto bcast;
#endif

#ifdef CONFIG_RTL865X_MULTICAST
		/*
			Mulicast:

			If forwarding engine's multicast process is enabled, we must return multicast packet here
			to prevent from extension port L2 broadcast.
		*/
		if (	rtl8651_drvMulticastProcessEnable &&
			(m_data[0] == 0x01 && m_data[1] == 0x00 && m_data[2] == 0x5e))
		{
			return FWDENG_IP_MULTICAST;
		}
#endif /* CONFIG_RTL865X_MULTICAST */

		/*
			IEEE 802.1D/Q reserved MAC:

			2 dmac ranges need to reserve:

				01:80:c2:00:00:00 ~ 01:80:c2:00:00:10
				01:80:c2:00:00:20 ~ 01:80:c2:00:00:2f

			If forwarding engine's spanning tree process is enabled, we must return control packet here
			to prevent from extension port L2 broadcast.
		*/
		if (	rtl8651_tblDrvSpanningTreeEnable &&
			m_data[0] == 0x01 &&
			m_data[1] == 0x80 &&
			m_data[2] == 0xc2 &&
			m_data[3] == 0x00 &&
			m_data[4] == 0x00 &&
			((m_data[5] <= 0x10) || ((m_data[5] & 0xf0) == 0x20)) )
		{
			return FWDENG_L2_CTRLMSG;
		}
	}
#ifdef RTL865XB_DCUT_SWVLAN
bcast:

#endif
	/*
		All other packets will arrive here:
			1. bcast packet after filtering
			2. Unknown unicast packet

			Process : They might need to L2 broadcast.
	*/

	/* broadcast packet to extension devices */
	//rtlglue_printf("peripheralExtPortMask: %d\n", peripheralExtPortMask);
	if (vlanMbrPortMask & peripheralExtPortMask)
	{
		/* filter source port of packet */
		if(pkthdrPtr->ph_srcExtPortNum)
		{	//filter loopback packet. for 8650B only
			//uint32 filtered=local_vlanp->memberPortMask & ~(1<<(pkthdrPtr->ph_srcExtPortNum-1+RTL8651_MAC_NUMBER));	// shliu: variable replacement
			uint32 filtered = vlanMbrPortMask & ~(1 << (pkthdrPtr->ph_srcExtPortNum - 1 + RTL8651_MAC_NUMBER));
			if((filtered&rtl8651_allExtPortMask)==0)
			{	/* It's from extension port and no other extension port need to receive this packet, so we skip extension device send */
				goto physical_port_proc;
			}
		}

		/* loopback/extension port in pkt's vlan. clone it and send to PCI. */
		newMbuf= mBuf_dupPacket(mbufPtr, MBUF_DONTWAIT);
		if(newMbuf)
		{
			#ifdef RTL865X_DEBUG
			if(wlan_acc_debug>1){
				#ifdef RTL865XB_DCUT_SWVLAN
					if((m_data[0]&1)==0) {
						if (local_vlanp == NULL)
							rtlglue_printf("L2 unknown ucast from vlan %d to PCI,", vid);
						else
							rtlglue_printf("L2 unknown ucast from %s to PCI,", local_vlanp->internal?"LAN":"WAN");
					}
					else {
						if (local_vlanp == NULL)
							rtlglue_printf("L2 bcast from vlan %d to PCI,", vid);
						else
							rtlglue_printf("L2 bcast from %s to PCI,", local_vlanp->internal?"LAN":"WAN");
					}
				#else
					if((m_data[0]&1)==0)
						rtlglue_printf("L2 unknown ucast from %s to PCI,", local_vlanp->internal?"LAN":"WAN");
					else
						rtlglue_printf("L2 bcast from %s to PCI,", local_vlanp->internal?"LAN":"WAN");
				#endif
			}
			#endif
			
#ifdef RTL865XB_DCUT_SWVLAN
			if (local_vlanp == NULL) {
				/* ========= new mbuf for tag-set port =========*/
				newMbuf2 = mBuf_dupPacket(newMbuf, MBUF_DONTWAIT);
				if (newMbuf2) {
					newMbuf2->m_pkthdr->ph_portlist &= ~(swVlanPtr->untagPortMask);
					rtlglue_extDeviceSend(vid, 0, newMbuf2->m_pkthdr);//would be freed in pci driver
				}
				/* ========= mbuf for untag-set port =========*/
				newMbuf->m_pkthdr->ph_portlist &= swVlanPtr->untagPortMask;
				_rtl8651_rxPktPreprocess_stripVlan(newMbuf->m_pkthdr);
				rtlglue_extDeviceSend(vid, 0, newMbuf->m_pkthdr);//would be freed in pci driver
			}
			else
				rtlglue_extDeviceSend(vid, 0, newMbuf->m_pkthdr);//would be freed in pci driver
#else
			rtlglue_extDeviceSend(vid, 0, newMbuf->m_pkthdr);//would be freed in pci driver
#endif
		}

		/*
			pass original packet down:
				for physical port broadcast, to protocol stack(bcast) or drop(unknown ucast).
		*/
	}

physical_port_proc:	

	if( m_data[0]&1 )
	{
#ifdef RTL865XB_DCUT_SWVLAN
		if (local_vlanp == NULL)
			goto unknown_ucast;
#endif
		/* broadcast mac: try to process 1. multicast process 2. L2 broadcast if we need it */
		
		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug>1)
			rtlglue_printf("vid %d to protostack\n", vid);
		#endif

		/* distinguish between multicast and broadcast packet */
		if ( m_data[0] == 0x01 && m_data[1] == 0x00 && m_data[2] == 0x5e )
		{
			/*
				Multicast process
			*/
			_rtl8651_l2PhysicalPortRelay(mbufPtr);
 			return FWDENG_IP_MULTICAST;
		}

		/*
			Normal L2 broadcast process
		*/
		/* 1. check if we need to do L2 broadcast */
		if ( rtl8651_drvSoftwareBroadcastEnable[vid & (RTL8651_VLAN_NUMBER - 1)]==TRUE )
		{
			_rtl8651_l2PhysicalPortRelay(mbufPtr);
		}

		/* 2. process ro specific ether type */
		switch (ethtype)
		{

			case 0x8863:		/* broadcast PPPoE PADI */
			{
				if (m_data[15] == 0x7)
				{	/* detect PPPoE offer, we add server's mac to L2 table automatically */

					#ifdef RTL865X_DEBUG
					uint8 svrMac[24];
					ether_aton_r(m_data+6,svrMac);
					rtlglue_printf("detected PPPoE svr %s on port %d.\n",svrMac, pkthdrPtr->ph_portlist);
					#endif

					_rtl8651_addVlanRefFilterDatabaseEntry(RTL8651_L2_TYPEI, local_vlanp->fid, 
						local_vlanp->vid, (ether_addr_t *)m_data+6, RTL8651_FORWARD_MAC, 1<<pkthdrPtr->ph_portlist);
				}
				return FWDENG_PPPOE8863;
			}
				break;
			case 0x0806:
			{
				if (_arpAgent)
				{
					return _arpAgent (pkthdrPtr, mbufPtr, vid);
				}
			}
				break;
		}

		/* all broadcast packet trap to protocol stack */
		return FWDENG_BROADCAST;

	}

	/* unknow unicast process */
#ifdef RTL865XB_DCUT_SWVLAN
unknown_ucast:
	if (local_vlanp == NULL) {
		#if 0
		rtlglue_printf("smac learning mac: %x.%x.%x.%x.%x.%x portmask: %d\n", 
			((ether_addr_t *)(m_data + 6))->octet[0], ((ether_addr_t *)(m_data + 6))->octet[1],
			((ether_addr_t *)(m_data + 6))->octet[2], ((ether_addr_t *)(m_data + 6))->octet[3],
			((ether_addr_t *)(m_data + 6))->octet[4], ((ether_addr_t *)(m_data + 6))->octet[5],
			pkthdrPtr->ph_portlist);
		#endif
		/* (1) smac learning */
		_rtl8651_addVlanRefFilterDatabaseEntry(RTL8651_L2_TYPEI, swVlanPtr->fid, 
			vid, (ether_addr_t *)(m_data + 6), RTL8651_FORWARD_MAC, 1 << pkthdrPtr->ph_portlist);
		
		/* (2) L2 broadcast */
		pkthdrPtr->ph_portlist = swVlanPtr->memberPortUpStatus & ~(1 << pkthdrPtr->ph_portlist);
		pkthdrPtr->ph_portlist &= ~rtl8651_allExtPortMask; /* filter Extension devices */

		/* ========= new mbuf for tag-set port =========*/
		newMbuf = mBuf_dupPacket(mbufPtr, MBUF_DONTWAIT);
		if (newMbuf) {
			newMbuf->m_pkthdr->ph_portlist &= ~(swVlanPtr->untagPortMask);			
			_rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void *)newMbuf->m_pkthdr, vid, -1);
		}
		/* ========= mbuf for untag-set port =========*/		
		pkthdrPtr->ph_portlist &= swVlanPtr->untagPortMask;
		_rtl8651_rxPktPreprocess_stripVlan(pkthdrPtr);
		_rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void *)pkthdrPtr, vid, -1);
	}
#endif

	#ifdef RTL865X_DEBUG
	if(wlan_acc_debug>1)
		rtlglue_printf("unknown unicast. Dropped!\n");
	#endif
drop:
	mBuf_driverFreeMbufChain(pkthdrPtr->ph_mbuf);
	
	return SUCCESS;
}


__IRAM_EXTDEV static int32  _rtl8651_fwdEnginePciSend(void * data, int16 dvid, int32 iphdrOffset)
{
	struct rtl_pktHdr * pkthdrPtr=(struct rtl_pktHdr *)data;

	struct rtl_mBuf *mbuf;
	rtl8651_tblDrv_vlanTable_t *local_vlanp;	

	mbuf=pkthdrPtr->ph_mbuf;
	pkthdrPtr->ph_vlanIdx= dvid & (RTL8651_VLAN_NUMBER-1);
	local_vlanp = &DrvTbl.vlan[pkthdrPtr->ph_vlanIdx ];

	if(peripheralExtPortMask>RTL8651_PHYSICALPORTMASK){
		//specify a NULL portlist and a src ext port number to let NIC driver know that this pkt is to be 
		//sent via ext port.
		pkthdrPtr->ph_srcExtPortNum=
			extPortMaskToPortNum[ (pkthdrPtr->ph_portlist>>RTL8651_PORT_NUMBER)&_RTL865XB_EXTPORTMASKS];
	#ifdef RTL865X_DEBUG
		if(wlan_acc_debug>1)
			rtlglue_printf("\tpci to swcore from ext %d ", pkthdrPtr->ph_srcExtPortNum);
	#endif
		pkthdrPtr->ph_portlist=0; 
	}else
		pkthdrPtr->ph_srcExtPortNum=0;
	pkthdrPtr->ph_portlist |= rtl8651_tblDrvCpuMirrorPortMask;
	//send pkt here
	rtlglue_drvSend((void*)data);
		
	return SUCCESS;
}

#if 0
static int32 _rtl8651_l2Relay(struct rtl_mBuf *inMbufPtr) {
	struct rtl_mBuf *mbufPtr;
	struct rtl_pktHdr *phdr=inMbufPtr->m_pkthdr;

	// hiwu :disable checksum regen
	phdr->ph_flags &= (~(CSUM_IP_OK | CSUM_L4_OK));
	
	mbufPtr = mBuf_dupPacket(inMbufPtr, MBUF_DONTWAIT);

	if(mbufPtr == NULL){
		rtlglue_printf("No more mbufs for relaying brdcast pkt\n");
		return FAILED;
	}

	//relay to all member ports except the incoming port	
	mbufPtr->m_pkthdr->ph_portlist = DrvTbl.vlan[phdr->ph_vlanIdx].memberPortUpStatus & ~(1<<phdr->ph_portlist);
	return _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)mbufPtr->m_pkthdr,DrvTbl.vlan[phdr->ph_vlanIdx].vid,-1 );
}
#endif

