/*
* Copyright c                  Realtek Semiconductor Corporation, 2003
* All rights reserved.
* 
* Program : Switch core table driver rtl8651_multicast.c
* Abstract : 
* Author : Yi-Lun Chen (chenyl@realtek.com.tw)
*
*/

#include "rtl_types.h"
#include "rtl8651_multicast_local.h"
#include "mbuf.h"

/******************************
Utility
*******************************/
#define MAX(a,b) ((a>b)?a:b)
#define MIN(a,b) ((a>b)?b:a)

/******************************
for IGMP message generate
*******************************/
static uint16					system_id							= 1;		// id field in ip-hdr of new generated packet
#ifdef RTL865X_TEST
#define						SYS_ID								system_id
#else
#define						SYS_ID								(system_id ++)
#endif /* RTL865X_TEST */

/******************************
Global Variable Declaration
*******************************/
static rtl8651_tblDrvMCastGlobal_t rtl8651_GlobalVariable;
static rtl8651_tblDrvMCast_t *	rtl8651_MCastTable					= NULL;									/* ptr to linked list of multicast table */
static uint32					rtl8651_drvMaxMCastCnt				= RTL8651_DefaultDrvMaxMCastCnt;			/* Max mcast group cnt */
static uint32					rtl8651_drvMaxMCastMemberCnt		= RTL8651_DefaultDrvMaxMCastMemberCnt;	/* Max per-group mcast subscriber cnt */
static uint32					rtl8651_drvMaxTotalMemberCnt		= RTL8651_DefaultDrvMaxTotalMemberCnt;	/* Max total subscriber count */
static uint32					rtl8651_drvMaxTotalSourceCnt		= RTL8651_DefaultDrvMaxTotalSourceCnt;		/* Max total source count */
//static uint32					rtl8651_drvMCastSystemRsvSourceCnt= 0;										/* Systerm Reserved Source list count */
static uint32					rtl8651_drvMCastMember_ExpireTime	= RTL8651_DefaultDrvMcastMemberTimeout;	/* timeout of member */
static uint32					rtl8651_drvMCastUpstream_Timeout	= RTL8651_DefaultDrvMCastUpstreamTimeout;	/* timeout of multicast entry without upstream */
static uint8					rtl8651_drvMCastVersion				= RTL8651_MCAST_IGMP_MAX_VERSION;

static uint32					rtl8651_igmp_qi						= RTL8651_IGMP_QI;
static uint32					rtl8651_igmp_qri					= RTL8651_IGMP_QRI;
static uint32					rtl8651_igmp_oqpi					= RTL8651_IGMP_OQPI;

//	=> if default upstream is set, it will not broadcast to default Upstream
static uint32					rtl8651_igmp_sqc					= RTL8651_IGMP_SQC;
static uint32					rtl8651_igmp_sqi					= RTL8651_IGMP_SQI;

static uint32					rtl8651_igmp_lmqc					= RTL8651_IGMP_LMQC;
static uint32					rtl8651_igmp_lmqi					= RTL8651_IGMP_LMQI;

static uint32					rtl8651_igmp_gmi					= RTL8651_IGMP_GMI;
static uint32					rtl8651_igmp_lmqt					= (RTL8651_IGMP_LMQC * RTL8651_IGMP_LMQI);

#define						rtl8651_drvMCastMember_DestroyTime	(((rtl8651_igmp_lmqc - 1) * rtl8651_igmp_lmqi) + rtl8651_igmp_qri + rtl8651_igmp_response_tolerance_delay)

static uint32					rtl8651_igmp_general_query			= RTL8651_IGMP_GENERAL_QUERY;
static uint32					rtl8651_igmp_group_query_interval	= RTL8651_IGMP_GROUP_QUERY_INTERVAL;
static uint32					rtl8651_igmp_response_tolerance_delay	= RTL8651_IGMP_RESPONSE_TOLERANCE_DELAY;
// counter
static uint32					rtl8651_MCastEntryCnt				= 0;				/* Counter of total mulicast entry */
static uint32					rtl8651_MemberCnt					= 0;				/* Counter of total member */
static uint32					rtl8651_SourceListCnt				= 0;				/* Counter of source list */
static uint8					system_enable						= TRUE;

/*
	For member entry, we must consider (RTL8651_DefaultDrvMaxTotalMemberCnt) and (RTL8651_DefaultDrvMaxMCastCnt*RTL8651_DefaultDrvMaxMCastMemberCnt)
	For source entry, we must consider (RTL8651_DefaultDrvMaxTotalSourceCnt) and (RTL8651_DefaultDrvMaxMCastMemberCnt*RTL8651_DefaultDrvMaxMCastSourceCnt)

	to determine the actual count of these two entries
*/
#define						rtl8651_addedMCastSourceCnt		(4 * rtl8651_drvMaxTotalSourceCnt)
static uint32					rtl8651_SystemMCastMemberCnt		= MIN(RTL8651_DefaultDrvMaxTotalMemberCnt, (RTL8651_DefaultDrvMaxMCastCnt*RTL8651_DefaultDrvMaxMCastMemberCnt));
static uint32					rtl8651_SystemMCastSourceCnt		= (5 * RTL8651_DefaultDrvMaxTotalSourceCnt);


/*
	Multicast process option, masked by definitions in rtl8651_multicast.h
*/
static uint32					rtl8651_multicastProcessOptions		= RTL8651_MCAST_PROCOPT_DEFAULT;

/******************************
IGMP Proxy Data Structure
*******************************/
#define	MCast_DATA_TYPE_MCastEntry	0xff
#define	MCast_DATA_TYPE_MCastMember	0xfe
#define	MCast_DATA_TYPE_MCastSource	0xfd

// pools
static rtl8651_tblDrvMCast_t			*MCastEntryPool			= NULL;
static rtl8651_tblDrvMCastMember_t	*MCastMemberPool		= NULL;
static rtl8651_tblDrvMCastSource_t	*MCastSourcePool		= NULL;
// ptrs
static rtl8651_tblDrvMCast_t			*MCastEntryPoolPtr		= NULL;
static rtl8651_tblDrvMCastMember_t	*MCastMemberPoolPtr	= NULL;
static rtl8651_tblDrvMCastSource_t	*MCastSourcePoolPtr		= NULL;


/******************************
Multicast Group Filter Data Structure
*******************************/
static rtl8651_tblDrvMCastFilter_t	*MCastFilter[RTL8651_MCastFilterHashSize]	= {NULL};
static rtl8651_tblDrvMCastFilter_t *MCastFilter_FreeHdr						= NULL;

/*********************************************
			Internal function Declaration
  *********************************************/
/**************************	
Initiate
**************************/
static __inline__ int32 _rtl8651_multicastSetInitPara(rtl8651_fwdEngineInitPara_t *);
/**************************
	Resource Managment
**************************/
static __inline__ int32 MCastPool_init(void);
static __inline__ void MCastPool_reinit(void);
static void *MCast_malloc(uint32);
static void MCast_free(void *);
/************************
	Utilization
************************/
static uint16 MCast_igmpChecksum(struct ip *);
static void MCast_BroadCastQuery(void);
static int32 MCast_setPkthdrProperty(struct rtl_pktHdr *, int32);
/************************
	Structure Maintenance
************************/
// Multicast Entry
static rtl8651_tblDrvMCast_t *MCast_FindMCastEntry(ipaddr_t, uint8, uint8);
static rtl8651_tblDrvMCast_t *MCast_CreateMCastEntry(ipaddr_t);
static __inline__ uint8 MCast_MCastEntryMinVersion(rtl8651_tblDrvMCast_t *);
static __inline__ void MCast_SetMCastEntryUpStream(rtl8651_tblDrvMCast_t *,rtl8651_tblDrv_vlanTable_t *, uint16, uint8);
static void MCast_ClrMCastEntryUpStream(uint32);
static void MCast_RemoveMCastEntry(rtl8651_tblDrvMCast_t *);
static __inline__ void MCast_Entrylink(rtl8651_tblDrvMCast_t *);
static __inline__ void MCast_EntryUnlink(rtl8651_tblDrvMCast_t *);
// Member
static rtl8651_tblDrvMCastMember_t *MCast_FindMember(rtl8651_tblDrvMCast_t *, rtl8651_tblDrv_vlanTable_t *, uint16, uint8, uint8, uint8);
static rtl8651_tblDrvMCastMember_t *MCast_CreateMember(rtl8651_tblDrvMCast_t *, rtl8651_tblDrv_vlanTable_t *, uint16, uint8);
static void MCast_RemoveMember(rtl8651_tblDrvMCast_t *, rtl8651_tblDrvMCastMember_t *);
static void MCast_DestroyMember(rtl8651_tblDrvMCastMember_t *);
static __inline__ void MCast_MemberLink(rtl8651_tblDrvMCast_t *, rtl8651_tblDrvMCastMember_t *);
static __inline__ int32 MCast_MemberUnLink(rtl8651_tblDrvMCastMember_t *);
static void MCast_MemberExcludeExpire(uint32);
static void MCast_MemberExpire(uint32);
static void MCast_MemberSendQuery(rtl8651_tblDrvMCastMember_t *);
// Source List
static __inline__ rtl8651_tblDrvMCastSource_t *MCast_SourceList_Create(void);
static __inline__ void MCast_SourceList_Link(rtl8651_tblDrvMCast_t *, rtl8651_tblDrvMCastMember_t *, rtl8651_tblDrvMCastSource_t *);
static __inline__ void MCast_SourceList_UnLink(rtl8651_tblDrvMCastSource_t *);
static rtl8651_tblDrvMCastSource_t* MCast_SourceList_inList(ipaddr_t , rtl8651_tblDrvMCastSource_t *);
static __inline__ uint32 MCast_SourceList_EntryCnt(rtl8651_tblDrvMCastSource_t *);
static rtl8651_tblDrvMCastSource_t *MCast_SourceList_freeCurrent(rtl8651_tblDrvMCastSource_t *);
static void MCast_SourceList_freeAll(rtl8651_tblDrvMCastSource_t *);
static uint32 MCast_SourceList_removal(rtl8651_tblDrvMCastSource_t *,uint8 ,uint8 ,rtl8651_tblDrvMCastSource_t *,uint8 ,uint8 ,rtl8651_tblDrvMCastSource_t **);
static void MCast_SourceList_addList(rtl8651_tblDrvMCastMember_t *, rtl8651_tblDrvMCastSource_t *, uint8, uint8);
static int32 MCast_SourceList_dup(rtl8651_tblDrvMCastMember_t *, rtl8651_tblDrvMCastSource_t *, rtl8651_tblDrvMCastSource_t **);
static void MCast_SourceList_Expire(rtl8651_tblDrvMCastSource_t *);
static uint32 MCast_SourceList_getCache(rtl8651_tblDrvMCast_t *, rtl8651_tblDrvMCastSource_t **);
/******************
	Pkt Process
******************/
static __inline__ ipaddr_t MCast_GetIpByVlan(rtl8651_tblDrv_vlanTable_t *, uint32);
static __inline__ int32 MCast_SetRouteInfoAndGetIntfIp(ipaddr_t *, rtl8651_tblDrvMCastRouteInfo_t *);
static int32 MCast_query(struct rtl_pktHdr *, struct ip *, struct igmp *);
static int32 MCast_report(struct rtl_pktHdr *, struct ip *, struct igmp *);
static int32 MCast_leave(struct rtl_pktHdr *, struct ip *, struct igmp *);

static int32 MCast_QueryPack(rtl8651_tblDrvMCastMember_t *,rtl8651_tblDrvMCastRouteInfo_t *, uint8, rtl8651_tblDrvMCastSource_t *, uint16, struct rtl_pktHdr *);
static int32 MCast_ReportPack(rtl8651_tblDrvMCast_t *, uint8, uint8, rtl8651_tblDrvMCastSource_t *, uint8, rtl8651_tblDrvMCastSource_t *, struct rtl_pktHdr *);

static int32 MCast_LeavePack(rtl8651_tblDrvMCast_t *, struct rtl_pktHdr *);
static int32 MCast_l3ActRouting(struct rtl_pktHdr *, rtl8651_tblDrvMCastRouteInfo_t *);

static __inline__ void MCast_Proc_ISIN(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);
static __inline__ void MCast_Proc_ISEX(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);
static __inline__ void MCast_Proc_TOIN(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);
static __inline__ void MCast_Proc_TOEX(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);
static __inline__ void MCast_Proc_ALLOW(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);
static __inline__ void MCast_Proc_BLOCK(rtl8651_tblDrvMCastMember_t*, rtl8651_tblDrvMCastSource_t *, uint16 ,uint8);

/************************
	Multicast group filter
************************/
static __inline__ uint32 _rtl8651_multicast_Filter_Hash(ipaddr_t);
rtl8651_tblDrvMCastFilter_t *_rtl8651_multicast_FindFilter(ipaddr_t);
/*********************************************
			Implementation
  *********************************************/
/**************************
	Initialize
**************************/
static __inline__ int32 MCastPool_init(void)
{
	int32 retval = SUCCESS;	/* default : SUCCESS */

	rtlglue_drvMutexLock();	/* Lock resource */

	if (	(rtl8651_drvMaxMCastCnt == 0) ||
		(rtl8651_SystemMCastMemberCnt == 0) ||
		(rtl8651_SystemMCastSourceCnt == 0)	)
	{
		goto out;	/* return SUCCESS */
	}

	/* Allocate memory */
	MCastEntryPool = (rtl8651_tblDrvMCast_t *)rtlglue_malloc(sizeof(rtl8651_tblDrvMCast_t) * rtl8651_drvMaxMCastCnt);
	MCastMemberPool =
		(rtl8651_tblDrvMCastMember_t *)rtlglue_malloc(sizeof(rtl8651_tblDrvMCastMember_t) * rtl8651_SystemMCastMemberCnt);
	MCastSourcePool = 
		(rtl8651_tblDrvMCastSource_t *)rtlglue_malloc(sizeof(rtl8651_tblDrvMCastSource_t) * rtl8651_SystemMCastSourceCnt);

	if (	(MCastEntryPool == NULL) ||
		(MCastMemberPool == NULL) ||
		(MCastSourcePool == NULL)	)
	{
		/* return FAILED */
		retval = FAILED;
		goto out;
	}

	MCastPool_reinit();

	/* return SUCCESS */

out:
	rtlglue_drvMutexUnlock();	/* UnLock resource */

	return retval;
}
void MCastPool_reinit(void)
{
	uint32 idx;
	rtl8651_tblDrvMCast_t *mCastPtr;
	rtl8651_tblDrvMCastMember_t *mCastMemberPtr;
	rtl8651_tblDrvMCastSource_t *mCastSourcePtr;

	rtlglue_drvMutexLock();//Lock resource

	/* Multicast Entry */
	bzero(MCastEntryPool, (rtl8651_drvMaxMCastCnt * sizeof(rtl8651_tblDrvMCast_t)));
	MCastEntryPoolPtr = MCastEntryPool;
	mCastPtr = MCastEntryPool;
	for (idx = 0 ; idx < rtl8651_drvMaxMCastCnt ; idx ++, mCastPtr ++)
	{
		if (idx == (rtl8651_drvMaxMCastCnt - 1))
			mCastPtr->next = NULL;
		else
			mCastPtr->next = mCastPtr + 1;
	}
	/* Multicast Member */
	bzero(MCastMemberPool, (rtl8651_SystemMCastMemberCnt * sizeof(rtl8651_tblDrvMCastMember_t)));
	MCastMemberPoolPtr = MCastMemberPool;
	mCastMemberPtr = MCastMemberPool;
	for (idx = 0 ; idx < rtl8651_SystemMCastMemberCnt ; idx ++, mCastMemberPtr ++)
	{
		if (idx == (rtl8651_SystemMCastMemberCnt - 1))
			mCastMemberPtr->next = NULL;
		else
			mCastMemberPtr->next = mCastMemberPtr + 1;
	}
	/* Source List */
	bzero(MCastSourcePool, (rtl8651_SystemMCastSourceCnt * sizeof(rtl8651_tblDrvMCastSource_t)));
	MCastSourcePoolPtr = MCastSourcePool;
	mCastSourcePtr = MCastSourcePool;
	for (idx = 0 ; idx < rtl8651_SystemMCastSourceCnt ; idx ++, mCastSourcePtr ++)
	{
		if (idx == (rtl8651_SystemMCastSourceCnt - 1))
			mCastSourcePtr->next = NULL;
		else
			mCastSourcePtr->next = mCastSourcePtr + 1;
	}

	rtlglue_drvMutexUnlock();//UnLock resource
}

#define RTL8651_MULTICAST_EXTERNAL_APIS
/*
@func int32		| rtl8651_multicastSetPara	| Set Parameters of Multicast System
@parm 		rtl8651_tblDrvMCastPara_t * | para | Timer Parameters to set
@rvalue SUCCESS	| 	parameter setting successfully.
@rvalue FAILED	| 	setting failed.
@comm 
Dynamically set timing parameters of multicast system.
*/
int32 rtl8651_multicastSetPara(rtl8651_tblDrvMCastPara_t *para)
{
	if (!para)
		return FAILED;

	rtlglue_drvMutexLock();//Lock resource

	if (para->MCastMember_ExpireTime)
		rtl8651_drvMCastMember_ExpireTime = para->MCastMember_ExpireTime;
	if (para->MCastUpstream_Timeout)
		rtl8651_drvMCastUpstream_Timeout = para->MCastUpstream_Timeout;
	if (para->igmp_qi)
		rtl8651_igmp_qi = para->igmp_qi;
	if (para->igmp_qri)
		rtl8651_igmp_qri = para->igmp_qri;
	if (para->igmp_gmi)
		rtl8651_igmp_gmi = para->igmp_gmi;
	if (para->igmp_group_query_interval)
		rtl8651_igmp_group_query_interval = para->igmp_group_query_interval;

	rtl8651_igmp_oqpi = para->igmp_oqpi;
	rtl8651_igmp_sqc = para->igmp_sqc;
	rtl8651_igmp_sqi = para->igmp_sqi;
	rtl8651_igmp_lmqc = para->igmp_lmqc;
	rtl8651_igmp_lmqi = para->igmp_lmqi;
	rtl8651_igmp_lmqt = (rtl8651_igmp_lmqc * rtl8651_igmp_lmqi);

	rtl8651_igmp_response_tolerance_delay = para->igmp_response_tolerance_delay;

	if (rtl8651_GlobalVariable.remainingQuery)	// system in startup-phase
		rtl8651_GlobalVariable.remainingQuery = rtl8651_igmp_sqc;

	rtlglue_drvMutexUnlock();//UnLock resource	
	return SUCCESS;
}

#ifdef RTL865X_TEST
void _rtl8651_setMcastProcessOptions(uint32 option)
{
	rtl8651_multicastProcessOptions = option;
}
#endif

// set initial parameters
int32 _rtl8651_multicastSetInitPara(rtl8651_fwdEngineInitPara_t *para)
{
	uint32 c1, c2;
	// error check
	if (!para)
		return FAILED;

	rtl8651_multicastProcessOptions		= para->MCastProcessOption;
	rtl8651_drvMaxMCastCnt				= para->MaxMCastCnt; 
	rtl8651_drvMaxMCastMemberCnt		= para->MaxMCastMemberCnt;
	rtl8651_drvMaxTotalMemberCnt		= para->MaxMCastTotalMemberCnt;
	rtl8651_drvMaxTotalSourceCnt		= para->MaxMCastTotalSourceCnt;
	rtl8651_SystemMCastSourceCnt		= rtl8651_drvMaxTotalSourceCnt + rtl8651_addedMCastSourceCnt;

	if ((para->MCastDefaultVersion > RTL8651_MCAST_IGMP_MAX_VERSION) ||
			(para->MCastDefaultVersion == 0))
	{
		rtl8651_drvMCastVersion = RTL8651_MCAST_IGMP_MAX_VERSION;
	}else
	{
		rtl8651_drvMCastVersion = para->MCastDefaultVersion;
	}
		

	if (para->MCastMember_ExpireTime)
		rtl8651_drvMCastMember_ExpireTime	= para->MCastMember_ExpireTime;
	if (para->MCastUpstream_Timeout)
		rtl8651_drvMCastUpstream_Timeout	= para->MCastUpstream_Timeout;
	if (para->igmp_qi)
		rtl8651_igmp_qi						= para->igmp_qi;
	if (para->igmp_qri)
		rtl8651_igmp_qri					= para->igmp_qri;
	if (para->igmp_gmi)
		rtl8651_igmp_gmi					= para->igmp_gmi;
	if (para->igmp_group_query_interval)
		rtl8651_igmp_group_query_interval	= para->igmp_group_query_interval;

	rtl8651_igmp_oqpi						= para->igmp_oqpi;
	rtl8651_igmp_sqc						= para->igmp_sqc;
	rtl8651_igmp_sqi						= para->igmp_sqi;
	rtl8651_igmp_lmqc						= para->igmp_lmqc;
	rtl8651_igmp_lmqi						= para->igmp_lmqi;
	rtl8651_igmp_lmqt = (rtl8651_igmp_lmqc * rtl8651_igmp_lmqi);

	rtl8651_igmp_response_tolerance_delay	= para->igmp_response_tolerance_delay;

	{
		// set total member count
		c1 = rtl8651_drvMaxTotalMemberCnt;
		c2 = (rtl8651_drvMaxMCastCnt * rtl8651_drvMaxMCastMemberCnt);

		if (c1 == 0)
			rtl8651_SystemMCastMemberCnt = c2;
		else if (c2 == 0)
			rtl8651_SystemMCastMemberCnt = c1;
		else
			rtl8651_SystemMCastMemberCnt = MIN(c1, c2);
	}

	return SUCCESS;
}

/*
@func void		| rtl8651_multicastReinit	| Re-initialize Multicast System
@comm
Re-initialize Multicast System, release all multicast entries and re-broadcast General Query to all ports
 */
void rtl8651_multicastReinit(void)
{
	if (system_enable == FALSE)
		return;

	rtlglue_drvMutexLock();//Lock resource
	// clear all multicast entry
	while (rtl8651_MCastTable)
		MCast_RemoveMCastEntry(rtl8651_MCastTable);

	// remove all-other mcast entries
	_rtl8651_delMulticast(0, 0, 0);

	assert(rtl8651_MCastEntryCnt == 0);
	assert(rtl8651_MemberCnt == 0);

	// reset system iphdr id-field
	system_id = 1;
	// Reset multicast entry pool
	MCastPool_reinit();

	// Reset process options
	rtl8651_multicastProcessOptions = RTL8651_MCAST_PROCOPT_DEFAULT;

	// re-set Startup/periodic query
	rtl8651_GlobalVariable.remainingQuery = rtl8651_igmp_sqc;	// initial
	// set timer
	if (rtl8651_igmp_sqc)
	{
		rtl8651_GlobalVariable.query_timer.data	= (uint32)0;
		rtl8651_GlobalVariable.query_timer.function	= (void*)MCast_BroadCastQuery;
		_rtl8651_mod_timer(&(rtl8651_GlobalVariable.query_timer), _sysUpSeconds + RTL8651_IGMP_STARTUP_QUERY_DELAY);
	}

	rtlglue_drvMutexUnlock();//UnLock resource
}
/*
@func int32		| rtl8651_multicastInit	| Initialize Multicast System
@parm 		rtl8651_fwdEngineInitPara_t * | para | parameter to initiate multicast system
@rvalue SUCCESS	| 	parameter setting successfully.
@rvalue FAILED	| 	setting failed.
@comm 
Initialize Multicast System, clear all multicast entries and broadcast General Query to all ports. Besides
Set Initial Parameters of Multicast System, including all timers and Maximum amount of multicast entry/member, (NULL)
when apply default values
*/
int32 _rtl8651_multicastInit(rtl8651_fwdEngineInitPara_t * para)
{
	int32 retval = SUCCESS;

	system_enable = FALSE;

	/* Initial global variables */
	rtl8651_MCastTable = NULL;
	bzero(&rtl8651_GlobalVariable, sizeof(rtl8651_tblDrvMCastGlobal_t));
	rtl8651_GlobalVariable.MCastUpStream.isSet	= FALSE;
	_rtl8651_init_timer(&(rtl8651_GlobalVariable.query_timer));

	/* initiate system variables */
	if (para)
		retval = _rtl8651_multicastSetInitPara(para);

	/* check database status */
	if (
		(!rtl8651_drvMaxMCastCnt) ||
		(!rtl8651_SystemMCastMemberCnt) ||
		(!rtl8651_SystemMCastSourceCnt)
	)
		goto out;

	/* initiate memory pool */
	if ((retval = MCastPool_init()) != SUCCESS)
	{
		goto out;
	}

	/* start system */
	system_enable = TRUE;

	/* initiate others */
	rtl8651_multicastReinit();
out:
	return retval;
}
/*
@func int32		| rtl8651_multicastSetUpStream	| Set Default UpStream of multicast System
@rvalue SUCCESS	| upstream setting successfully.
@rvalue FAILED	| setting failed.
@comm
Set Default UpStream with indicate Network Interface Name and sessionid(for PPPOE/PPTP/L2TP) or ip address(for VLAN)
 */
int32 rtl8651_multicastSetUpStream(int8* netifName, uint32 para)
{
	rtl8651_tblDrvMCastUpStreamInfo_t fake_UpStream;
	rtl8651_tblDrv_networkIntfTable_t *netif;

	rtlglue_printf("\nSet IGMP Default Upstream interface ");

	rtlglue_drvMutexLock();//Lock resource

	memset(&fake_UpStream, 0, sizeof(rtl8651_tblDrvMCastUpStreamInfo_t));
	if ((!netifName) ||
		((netif = GET_NETWORK_INTERFACE(netifName, NULL)) == NULL))
	{
		goto err;
	}
	rtlglue_printf("(%s) ... ", netifName);

	fake_UpStream.netif = netif;
	fake_UpStream.vlan = &DrvTbl.vlan[netif->vid & (RTL8651_VLAN_NUMBER - 1)];
	fake_UpStream.portMask = fake_UpStream.vlan->memberPortMask;
	
	switch (netif->linkLayerType)
	{
		case RTL8651_LL_VLAN:
			{
				rtl8651_tblDrv_ipIntfEntry_t * ipIntfPtr;
				rtl8651_tblDrv_ipIntfIpAddrEntry_t * ipAddrPtr;

				if (para)
				{	// check for correctness
					SLIST_FOREACH(ipIntfPtr, &(netif->IpHead), nextIp)
					{
						if (!ipIntfPtr)
						{
							goto err;
						}
						SLIST_FOREACH(ipAddrPtr, &ipIntfPtr->ipAddr, nextIpAddr)
						{
							if (!ipAddrPtr)
							{
								goto err;
							}
							if (ipAddrPtr->ipAddr == para)
							{
								fake_UpStream.M_intfIp= para;
								goto set_para;
							}
						}
					}
					goto err;
				}else
				{
#if 0
					ipIntfPtr = SLIST_FIRST(&netif->IpHead);
					if (!ipIntfPtr)
					{
						fake_UpStream.M_intfIp = 0;
						goto set_para;
					}
					ipAddrPtr = SLIST_FIRST(&ipIntfPtr->ipAddr);
					if (!ipAddrPtr)
					{
						fake_UpStream.M_intfIp = 0;
						goto set_para;
					}
					assert(ipAddrPtr->ipAddr);
					fake_UpStream.M_intfIp = ipAddrPtr->ipAddr;
#else
					fake_UpStream.M_intfIp = 0;	// don't set it
#endif
				}
			}
			break;
		case RTL8651_LL_PPPOE:
			{
				rtl8651_tblDrvSession_t	*sessionPtr;
			
				if (para)
				{	// tarce for correctness or para
					CTAILQ_FOREACH(sessionPtr, &(netif->sessionHead), nextSession)
					{
						assert(sessionPtr->session_t);
						if (sessionPtr->session_t->sessionId == para)
						{
							fake_UpStream.M_pppoeSessionPtr = sessionPtr;
							goto set_para;
						}
					}
					goto err;
				}else
				{
					fake_UpStream.M_pppoeSessionPtr = CTAILQ_FIRST(&netif->sessionHead);
				}
			}
			break;
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		case RTL8651_LL_PPTP:
		case RTL8651_LL_L2TP:
			{
				rtl8651_tblDrvSession_t	*sessionPtr;
				if (para)
				{
					CTAILQ_FOREACH(sessionPtr, &(netif->sessionHead), nextSession)
					{
						assert(sessionPtr->session_t);
						if (sessionPtr->session_t->sessionId == para)
						{
							fake_UpStream.M_sessionPtr = sessionPtr;
							goto set_para;
						}
					}
					goto err;
				}else
				{
					fake_UpStream.M_sessionPtr = CTAILQ_FIRST(&netif->sessionHead);
				}
			}
			break;
		#endif /* CONFIG_RTL865X_PPTPL2TP */
		default:
			goto err;
	}
set_para:
	rtl8651_GlobalVariable.MCastUpStream		= fake_UpStream;
	rtl8651_GlobalVariable.MCastUpStream.isSet	= TRUE;

	rtlglue_drvMutexUnlock();//UnLock resource

	rtlglue_printf("SUCCESS!!\n");
	return SUCCESS;
err:
	rtlglue_drvMutexUnlock();//UnLock resource
	rtlglue_printf("FAILED!!\n");
	return FAILED;
}
/*
@func int32		| rtl8651_multicastPeriodicQuery	| Turn ON/OFF periodic-General-Query broadcast
@parm 		int8 | enable | enable(1) or disable(0)
@rvalue SUCCESS	| Status switching successful
@comm 
Turn on/off periodic-general-query broadcast, if ON, gateway will broadcast IGMPv2 general query to each interfaces
(excepting default Upstream if it is set)
*/
int32 rtl8651_multicastPeriodicQuery(int8 enable)
{
	if (enable)
	{
		/*
			modify timer only when original switch is turned off and system is not in startup state
		*/
		if (rtl8651_igmp_general_query == FALSE)
		{
			if (rtl8651_GlobalVariable.remainingQuery == 0)
			{
				_rtl8651_mod_timer(&(rtl8651_GlobalVariable.query_timer), _sysUpSeconds + rtl8651_igmp_qi);
			}
			rtl8651_igmp_general_query = TRUE;
			/*
				--> Send Query immediately
			*/
			MCast_BroadCastQuery();
		}
	}else
	{
		rtl8651_igmp_general_query = FALSE;
		if (rtl8651_GlobalVariable.remainingQuery == 0)
			_rtl8651_del_timer(&(rtl8651_GlobalVariable.query_timer));
	}
	return SUCCESS;
}
/**************************
	Resource Managment
**************************/
void *MCast_malloc(uint32 size)
{
	void *ret = NULL;

	switch (size)
	{
		case (sizeof (rtl8651_tblDrvMCast_t)):
			if (MCastEntryPoolPtr)
			{
				ret = MCastEntryPoolPtr;
				MCastEntryPoolPtr = MCastEntryPoolPtr->next;
				bzero(ret, sizeof(rtl8651_tblDrvMCast_t));
				((rtl8651_tblDrvMCast_t *)ret)->_type = MCast_DATA_TYPE_MCastEntry;
			}
			break;
		case (sizeof (rtl8651_tblDrvMCastMember_t)):
			if (MCastMemberPoolPtr)
			{
				ret = MCastMemberPoolPtr;
				MCastMemberPoolPtr = MCastMemberPoolPtr->next;
				bzero(ret, sizeof(rtl8651_tblDrvMCastMember_t));
				((rtl8651_tblDrvMCastMember_t *)ret)->_type = MCast_DATA_TYPE_MCastMember;
			}
			break;
		case (sizeof (rtl8651_tblDrvMCastSource_t)):
			if (MCastSourcePoolPtr)
			{
				ret = MCastSourcePoolPtr;
				MCastSourcePoolPtr = MCastSourcePoolPtr->next;
				bzero(ret, sizeof(rtl8651_tblDrvMCastSource_t));
				((rtl8651_tblDrvMCastSource_t *)ret)->_type = MCast_DATA_TYPE_MCastSource;
			}
			break;
		default:
			assert(0);
	}

	return ret;
}
void MCast_free(void *ptr)
{
	if (!ptr)
		return;

	switch (((rtl8651_tblDrvMCast_t *)ptr)->_type)
	{
		case MCast_DATA_TYPE_MCastEntry:
			((rtl8651_tblDrvMCast_t *)ptr)->next = MCastEntryPoolPtr;
			MCastEntryPoolPtr = (rtl8651_tblDrvMCast_t *)ptr;
			break;
		case MCast_DATA_TYPE_MCastMember:
			((rtl8651_tblDrvMCastMember_t *)ptr)->next = MCastMemberPoolPtr;
			MCastMemberPoolPtr = (rtl8651_tblDrvMCastMember_t *)ptr;
			break;
		case MCast_DATA_TYPE_MCastSource:
			((rtl8651_tblDrvMCastSource_t *)ptr)->next = MCastSourcePoolPtr;
			MCastSourcePoolPtr = (rtl8651_tblDrvMCastSource_t *)ptr;
			break;
		default:
			assert(0);
	}
}
/**************************
	Utility
**************************/
/* shoud continue memory */
uint16 MCast_igmpChecksum(struct ip * pip)
{
	uint16   *ptr;
	int32  nhdr, nbytes;
	int32  sum, i;


	nhdr =  ((pip->ip_vhl &0xf) << 2);
	nbytes = ntohs(pip->ip_len) - nhdr;
	ptr = (uint16 *) ((char *) pip + nhdr);
	sum = 0;

	/* Note: we always padding here!! */
	*(((uint8 *)pip) + nhdr + nbytes) = (uint8)0;
	for (i=0; i<nbytes ; i=i+2){
		sum +=  (unsigned long) *ptr++;
	}

	/* Roll over carry bits */
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);

	/* Return checksum */
	return ((uint16) ~ sum);
}
/*
	Broadcast IGMPv2 General Query to All Vlan/Port
	to all internal vlan
	excepting default UpStream Vlan/Port

*/
void MCast_BroadCastQuery(void)
{

	if ((system_enable == FALSE) || (rtl8651_drvMulticastProcessEnable == FALSE))
		return;
	
	if ((rtl8651_igmp_general_query == TRUE) ||
		(rtl8651_GlobalVariable.remainingQuery > 0))
	{
		// send General Query to each port
		rtl8651_tblDrvMCastRouteInfo_t	info;
		uint8						idx;
		rtl8651_tblDrv_vlanTable_t *		vlanp;// = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];

		#ifdef RTL8651_MCAST_DEBUG
		if (rtl8651_GlobalVariable.remainingQuery > 0)
			rtlglue_printf("[MCAST] Start-up Broadcast General Query: remaining[%d]\n", rtl8651_GlobalVariable.remainingQuery);
		else
			rtlglue_printf("[MCAST] Periodically broadcast General Query .....\n");
		#endif

		info.m_addr	= IGMP_ALL_HOSTS;

		vlanp = DrvTbl.vlan;
		for (idx = 0 ; idx < RTL8651_VLAN_NUMBER ; idx ++)
		{	// only broadcast to internal vlan
			if ((vlanp->valid) && (vlanp->internal))
			{
				info.netif		= NULL;
				info.ifInfo		= 0;
				info.vlan		= vlanp;
				info.port		= vlanp->memberPortMask;
				info.isPortMask= TRUE;
				if ((rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE) &&
					(rtl8651_GlobalVariable.MCastUpStream.vlan == vlanp))
				{
					info.port	&= ~(rtl8651_GlobalVariable.MCastUpStream.portMask);
					if (info.port == 0)
						goto next_vlan;
				}
				MCast_QueryPack(NULL, &info, rtl8651_drvMCastVersion, NULL, 0, NULL);
			}
		// trace to next vlan
		next_vlan:
			vlanp ++;
		}

		// set timer for next Query
		if (rtl8651_GlobalVariable.remainingQuery != 0)
		{	// start-up broadcast
			_rtl8651_mod_timer(&(rtl8651_GlobalVariable.query_timer), _sysUpSeconds + rtl8651_igmp_sqi);
			// decrease counter
			rtl8651_GlobalVariable.remainingQuery --;
		}
	}

	if (rtl8651_GlobalVariable.remainingQuery == 0)
	{
		if (rtl8651_igmp_general_query == TRUE)
			_rtl8651_mod_timer(&(rtl8651_GlobalVariable.query_timer), _sysUpSeconds + rtl8651_igmp_qi);
		else
			_rtl8651_del_timer(&(rtl8651_GlobalVariable.query_timer));
	}
	return;
}

/************************
	Structure Maintenance
************************/
// dump Multicast Entry Structure
void MCast_DumpMCast(void)
{
	rtl8651_tblDrvMCast_t *m_entry;
	rtl8651_tblDrvMCastMember_t *member;
	rtl8651_tblDrvMCastSource_t *source;

	rtlglue_printf("\nIGMP System report:\n");
	// Global Variable
	rtlglue_printf("System support latest IGMP version:\t\t[ IGMPv%d ]\n", RTL8651_MCAST_IGMP_MAX_VERSION);
	rtlglue_printf("System default IGMP version:\t\t\t[ IGMPv%d ]\n", rtl8651_drvMCastVersion);

	if (system_enable == FALSE)
	{
		rtlglue_printf("\nIP-multicast system Enable:\t\t[ NO ]\n\n");
		return;
	}

	// Timer
	rtlglue_printf("Timer:\n");
	rtlglue_printf("\tSystem specific:\n");
	if (rtl8651_drvMCastUpstream_Timeout == 0xffffffff)
		rtlglue_printf("\t\tUp-stream No-response timeout:\t\t[INFINITE]\n");
	else
		rtlglue_printf("\t\tUp-stream No-response timeout:\t\t%d sec.\n", rtl8651_drvMCastUpstream_Timeout);

	rtlglue_printf("\t\tDown-stream No-report timeout:\t\t%d sec.\n", rtl8651_drvMCastMember_ExpireTime);

	rtlglue_printf("\tIGMP msg related:\n");
	rtlglue_printf("\t\tQuery Interval:\t\t\t\t%d sec.\n", rtl8651_igmp_group_query_interval);
	rtlglue_printf("\t\tQuery Response Interval:\t\t%d sec.\n", rtl8651_igmp_qri);
	rtlglue_printf("\t\tOther Querier Present Interval:\t\t%d sec.\n", rtl8651_igmp_oqpi);
	rtlglue_printf("\t\tIGMP msg tolerance delay:\t\t%d sec\n", rtl8651_igmp_response_tolerance_delay);
	rtlglue_printf("\t\tSystem-up Query Count:\t\t\t%d\n", rtl8651_igmp_sqc);
	rtlglue_printf("\t\tSystem-up Query Interval:\t\t%d sec.\n", rtl8651_igmp_sqi);
	rtlglue_printf("\t\tLast Member Query Count:\t\t%d\n", rtl8651_igmp_lmqc);
	rtlglue_printf("\t\tLast Member Query Interval:\t\t%d sec.\n", rtl8651_igmp_lmqi);

	rtlglue_printf("\n\t\tPeriodic General Query:\t");
	if (rtl8651_igmp_general_query == TRUE)
		rtlglue_printf("[ on ]\n\t\t\tQuery Interval:\t\t\t%d sec.\n\n", rtl8651_igmp_qi);
	else
		rtlglue_printf("[ off ]\n\n");

	/* system information */
	rtlglue_printf("System status:\n");
	rtlglue_printf("\t\tRemaining Start-up Query Count:\t\t%d\n", rtl8651_GlobalVariable.remainingQuery);
	if (rtl8651_igmp_general_query == TRUE)
		rtlglue_printf("\t\tNext general query remaining time:\t%d sec.\n", (rtl8651_GlobalVariable.query_timer.expires - _sysUpSeconds));
	rtlglue_printf("\n");

	/* Counter */
	rtlglue_printf("Resource Utilization:\n");
	rtlglue_printf("\t\t\t\tTotal\tUsage\tFree\n");
	rtlglue_printf("\tMulticast Entry\t\t%d\t%d\t%d\n", rtl8651_drvMaxMCastCnt, rtl8651_MCastEntryCnt, (rtl8651_drvMaxMCastCnt-rtl8651_MCastEntryCnt));
	rtlglue_printf("\tMulticast Member\t%d\t%d\t%d\n", rtl8651_SystemMCastMemberCnt, rtl8651_MemberCnt, (rtl8651_SystemMCastMemberCnt-rtl8651_MemberCnt));
	rtlglue_printf("\tSource List\t\t%d\t%d\t%d\n", rtl8651_SystemMCastSourceCnt, rtl8651_SourceListCnt, (rtl8651_SystemMCastSourceCnt-rtl8651_SourceListCnt));

	rtlglue_printf("Multicast Entries:\n");
	rtlglue_printf("===============================================================\n");

	rtlglue_drvMutexLock();//Lock resource

	m_entry = rtl8651_MCastTable;
	while (m_entry)
	{	// dump m_entry info
		rtlglue_printf("(%3u.%3u.%3u.%3u) [IGMPv%d] <Up ", NIPQUAD(m_entry->mcast_addr), m_entry->MulticastRouterVersion);
		if (m_entry->up_vlan)	// upstream exists
			rtlglue_printf("Vid[0x%02X] pMsk[0x%02X] ", m_entry->up_vlan->vid, m_entry->up_portMask);
		else
			rtlglue_printf("NOT Set ");

		if ((rtl8651_drvMCastUpstream_Timeout == 0xffffffff) || !(m_entry->up_vlan))
			rtlglue_printf("(NO Exp) >");
		else
			rtlglue_printf("(Exp: %d) >", (m_entry->timer_upstream.expires - _sysUpSeconds));

		if (rtl8651_drvMaxMCastMemberCnt)
			rtlglue_printf(" MEMBER[%d:%d]\n", m_entry->memberCnt, rtl8651_drvMaxMCastMemberCnt);
		else
			rtlglue_printf(" MEMBER[%d]\n", m_entry->memberCnt);

		// member
		member = m_entry->member;
		while (member)
		{	// dump member info
			if (member->isForwarder == TRUE)
				rtlglue_printf("\t=> [O] ");
			else
				rtlglue_printf("\t=> [X] ");
			rtlglue_printf("VID[0x%02X] MBR[0x%02X] [v%d] %s(%d) ",
				member->vlan->vid, member->portMask, member->MulticastRouterVersion,
				((member->FilterMode == IGMP_FILTER_INCLUDE)?"Inc":"Exc"),
				((member->FilterMode == IGMP_FILTER_INCLUDE)?0:(member->timer_filterMode.expires - _sysUpSeconds)));

			if (member->state == IGMP_MEMBER_STATE_NORMAL)
			{
				rtlglue_printf("(NORMAL) NextQ(%d) ", (member->timer_query.expires - _sysUpSeconds));
				if (member->timer_query_expire.list.next)
					rtlglue_printf("Exp:(%d)\n", (member->timer_query_expire.expires - _sysUpSeconds));
				else
					rtlglue_printf("Exp:(%d)\n", (member->timer_state.expires - _sysUpSeconds));
			}else
				rtlglue_printf("(EXPIRE) Rem Q(%d) Exp(%d)\n", member->remainingQuery, (member->timer_state.expires - _sysUpSeconds));
			if (member->FilterMode == IGMP_FILTER_INCLUDE)
			{
				source = member->slist;
				while (source)
				{
					rtlglue_printf("\t\t+ %u.%u.%u.%u (Exp %d)\n", NIPQUAD(source->source), (source->timer.expires - _sysUpSeconds));
					source = source->next;
				}
			}else if (member->FilterMode == IGMP_FILTER_EXCLUDE)
			{
				source = member->slist;
				while (source)
				{
					if (source->timer.expires <= _sysUpSeconds)
						rtlglue_printf("\t\t- %u.%u.%u.%u\n", NIPQUAD(source->source));
					source = source->next;
				}
				source = member->slist;
				while (source)
				{
					if (source->timer.expires > _sysUpSeconds)
						rtlglue_printf("\t\t+ %u.%u.%u.%u (Exp %d)\n", NIPQUAD(source->source), (source->timer.expires - _sysUpSeconds));
					source = source->next;
				}
			}

			// next member
			member = member->next;
		}

		// next m_entry
		m_entry = m_entry->next;
	}

	rtlglue_drvMutexUnlock();//UnLock resource

	rtlglue_printf("===============================================================\n\n");
	return;
}
// find Multicast Entry
rtl8651_tblDrvMCast_t *MCast_FindMCastEntry(ipaddr_t mcast_addr, uint8 create, uint8 refresh)
{
	rtl8651_tblDrvMCast_t *m_entry = rtl8651_MCastTable;

	if (!IN_MULTICAST(mcast_addr))
		return NULL;

	while (m_entry)
	{
		if (m_entry->mcast_addr == mcast_addr)
		{	// found
			if (refresh == TRUE)
			{	// reset timer
				if (rtl8651_drvMCastUpstream_Timeout == 0xffffffff)
					_rtl8651_del_timer(&m_entry->timer_upstream);
				else if (m_entry->up_vlan)
					_rtl8651_mod_timer(&m_entry->timer_upstream, _sysUpSeconds + rtl8651_drvMCastUpstream_Timeout);
			}
			return m_entry;
		}
		m_entry = m_entry->next;
	}

	if (create == TRUE)
		return MCast_CreateMCastEntry(mcast_addr);	// create new entry when we cannot find old entry
	else
		return NULL;
}
// create new MCast Entry
rtl8651_tblDrvMCast_t *MCast_CreateMCastEntry(ipaddr_t mcast_addr)
{
	rtl8651_tblDrvMCast_t *new_entry;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Create MCast Entry (%u.%u.%u.%u) System Entry Count: [%d]\n", NIPQUAD(mcast_addr), rtl8651_MCastEntryCnt);
	#endif

	rtlglue_drvMutexLock();//Lock resource

	// can we allocate new multicast entry for this new one?
	if (rtl8651_MCastEntryCnt >= rtl8651_drvMaxMCastCnt)
		goto errout;

	// allocate new entry
	new_entry = (rtl8651_tblDrvMCast_t *)MCast_malloc(sizeof(rtl8651_tblDrvMCast_t));
	if (!new_entry)
		goto errout;

	// set values
	new_entry->mcast_addr				= mcast_addr;
	new_entry->MulticastRouterVersion		= rtl8651_drvMCastVersion;	/* default version */
	new_entry->memberCnt				= 0;
	new_entry->member					= NULL;
	new_entry->up_vlan					= NULL;
	new_entry->up_portMask				= 0;

	/* Initially, cached filter mode is NONE */
	new_entry->FilterMode					= IGMP_FILTER_NONE;
	new_entry->slist						= NULL;

	new_entry->next						= NULL;
	new_entry->prev						= NULL;

	// set timer
	_rtl8651_init_timer(&new_entry->timer_upstream);
	new_entry->timer_upstream.data = (uint32)new_entry;
	new_entry->timer_upstream.function =(void*)MCast_ClrMCastEntryUpStream;

	// link this new entry
	MCast_Entrylink(new_entry);

	// add counter
	rtl8651_MCastEntryCnt ++;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Create MCast Entry OK!! System Entry Count: [%d]\n", rtl8651_MCastEntryCnt);
	#endif

	rtlglue_drvMutexUnlock();//UnLock resource
	return new_entry;
errout:
	rtlglue_drvMutexUnlock();//UnLock resource
	return NULL;
}
// return the min version of m_entry
uint8 MCast_MCastEntryMinVersion(rtl8651_tblDrvMCast_t *m_entry)
{
	rtl8651_tblDrvMCastMember_t *mptr;
	uint8 v = RTL8651_MCAST_IGMP_MAX_VERSION;

	if (!m_entry)
		goto out;
	mptr = m_entry->member;
	while (mptr)
	{
		if (mptr->MulticastRouterVersion < v)
			v = mptr->MulticastRouterVersion;
		mptr = mptr->next;
	}
out:
	return v;
}
// set upstream information of multicast entry
void MCast_SetMCastEntryUpStream(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrv_vlanTable_t *vlan, uint16 up_port, uint8 RouterVersion)
{
	m_entry->up_vlan = vlan;
	m_entry->up_portMask = (1<<up_port);
	m_entry->MulticastRouterVersion = RouterVersion;
	if (rtl8651_drvMCastUpstream_Timeout != 0xffffffff)
		_rtl8651_mod_timer(&m_entry->timer_upstream, _sysUpSeconds + rtl8651_drvMCastUpstream_Timeout);
	m_entry->timer_upstream.data = (uint32)m_entry;
	m_entry->timer_upstream.function = (void*)MCast_ClrMCastEntryUpStream;
}
// clear upstream information
void MCast_ClrMCastEntryUpStream(uint32 expire)
{
	rtl8651_tblDrvMCast_t *m_entry = (rtl8651_tblDrvMCast_t *)expire;
	m_entry->up_vlan = NULL;
	m_entry->up_portMask = 0;
	m_entry->MulticastRouterVersion = rtl8651_drvMCastVersion;
}
// remove Multicast entry
void MCast_RemoveMCastEntry(rtl8651_tblDrvMCast_t *m_entry)
{
	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Remove MCast Entry (%u.%u.%u.%u) System Entry Count: [%d]\n", NIPQUAD(m_entry->mcast_addr), rtl8651_MCastEntryCnt);
	#endif

	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("[MCAST] *FLOW* Remove {%u.%u.%u.%u}\n",
		NIPQUAD(m_entry->mcast_addr));
	#endif
	rtlglue_drvMutexLock();//Lock resource

	_rtl8651_delMulticast(m_entry->mcast_addr, 0, 0);
	// send leave message to upstream
	if (m_entry->MulticastRouterVersion == IGMPV2)
		MCast_LeavePack(m_entry, NULL);
	if (m_entry->MulticastRouterVersion == IGMPV3)
		MCast_ReportPack(m_entry, m_entry->MulticastRouterVersion, IGMP_FILTER_EXCLUDE, NULL, IGMP_FILTER_INCLUDE, NULL, NULL);
	// clear member
	while (m_entry->member)
		MCast_RemoveMember(m_entry, m_entry->member);
	// remove forwarding cache
	MCast_SourceList_freeAll(m_entry->slist);
	// delete timer
	_rtl8651_del_timer(&m_entry->timer_upstream);
	MCast_EntryUnlink(m_entry);
	MCast_free(m_entry);
	// decrease counter
	rtl8651_MCastEntryCnt --;
	rtlglue_drvMutexUnlock();//UnLock resource

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Remove MCast Entry OK!! System Entry Count: [%d]\n", rtl8651_MCastEntryCnt);
	#endif
}

// link MCast Entry in the front of Multicast table
void MCast_Entrylink(rtl8651_tblDrvMCast_t *entry)
{
	rtlglue_drvMutexLock();//Lock resource
	entry->next = rtl8651_MCastTable;
	if (rtl8651_MCastTable)
		rtl8651_MCastTable->prev = entry;
	rtl8651_MCastTable = entry;
	rtlglue_drvMutexUnlock();//UnLock resource
}
// unlink MCast Entry from the Multicast table
void MCast_EntryUnlink(rtl8651_tblDrvMCast_t *entry)
{
	rtlglue_drvMutexLock();//Lock resource
	if (rtl8651_MCastTable == entry)
		rtl8651_MCastTable = entry->next;
	if (entry->next)
		entry->next->prev = entry->prev;
	if (entry->prev)
		entry->prev->next = entry->next;
	entry->prev = NULL;
	entry->next = NULL;
	rtlglue_drvMutexUnlock();//UnLock resource
}
// find member entry of a specific multicast entry
// "version" is for "create" and "refresh", we need to determine new timeout to refresh/create members
rtl8651_tblDrvMCastMember_t *MCast_FindMember(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrv_vlanTable_t *vlan, uint16 port,  uint8 create, uint8 refresh, uint8 ModifyDataFlow)
{
	rtl8651_tblDrvMCastMember_t *member = m_entry->member;
	uint16 portMask = (1<<port);

	while (member)
	{
		if ((member->vlan == vlan) &&
			(member->portMask == portMask))
		{
			if (refresh == TRUE)
			{
				uint8 oldState = member->state;
				member->state = IGMP_MEMBER_STATE_NORMAL;

				/* set expire time */
				_rtl8651_mod_timer(&(member->timer_state), _sysUpSeconds + rtl8651_drvMCastMember_ExpireTime);
				member->timer_state.data = (uint32)member;
				member->timer_state.function =(void*)MCast_MemberExpire;

				/* set query interval */
				_rtl8651_mod_timer(&(member->timer_query), _sysUpSeconds + rtl8651_igmp_group_query_interval);
				member->timer_query.data = (uint32)member;
				member->timer_query.function = (void*)MCast_MemberSendQuery;

				/* remove query expiration timer */
				_rtl8651_del_timer(&(member->timer_query_expire));

				/* modify active multicast table entry for this refreshed member */
				if ((oldState != IGMP_MEMBER_STATE_NORMAL) && (rtl8651_multicastProcessOptions & MCAST_PROC_EXPIRE_STATE_DONT_FWD))
				{
					#ifdef RTL8651_MCAST_DATA_DEBUG
						rtlglue_printf("[MCAST] *FLOW* Add Member(0x%X)[0x%X] to {%u.%u.%u.%u}\n",
						member->vlan->vid, member->portMask, NIPQUAD(m_entry->mcast_addr));
					#endif
					_rtl8651_addMulticastMember(m_entry->mcast_addr, (1<<((member->vlan->vid)&(RTL8651_VLAN_NUMBER-1))), member->portMask, FALSE);
				}

			}
			return member;
		}
		member = member->next;
	}
	if (create == TRUE)
		// create new entry when system cannot find new entry
		return MCast_CreateMember(m_entry, vlan, portMask, ModifyDataFlow);
	else
		return NULL;
}
// create new member entry for a specific multicast entry
rtl8651_tblDrvMCastMember_t *MCast_CreateMember(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrv_vlanTable_t *vlan, uint16 portMask, uint8 ModifyDataFlow)
{
	rtl8651_tblDrvMCastMember_t *member;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Create Member (%u.%u.%u.%u) Member Count: [%d]\n", NIPQUAD(m_entry->mcast_addr), m_entry->memberCnt);
	#endif

	rtlglue_drvMutexLock();//Lock resource

	if ((!m_entry) ||
		((rtl8651_drvMaxMCastMemberCnt)&&(m_entry->memberCnt >= rtl8651_drvMaxMCastMemberCnt)) ||
		((rtl8651_drvMaxTotalMemberCnt)&&(rtl8651_MemberCnt >= rtl8651_drvMaxTotalMemberCnt)))
		goto errout;

	if ((member = (rtl8651_tblDrvMCastMember_t *)MCast_malloc(sizeof(rtl8651_tblDrvMCastMember_t))) == NULL)
		goto errout;

	// set default parameter
	member->MulticastRouterVersion	= rtl8651_drvMCastVersion;		/* default version */
	member->vlan				= vlan;							/* set vlan */
	member->state				= IGMP_MEMBER_STATE_NORMAL;	/* default state */
	member->isForwarder			= TRUE;							/* default, is forwarder */

	member->next				= NULL;
	member->prev				= NULL;
	
	member->portMask			= portMask;						/* set port mask of this member */

	member->FilterMode			= IGMP_FILTER_NONE;
	member->slist				= NULL;

	// link and  for this member
	MCast_MemberLink(m_entry, member);

	// set timer
	_rtl8651_init_timer(&member->timer_state);
	member->timer_state.data = (uint32) member;						/* pointer to member */
	member->timer_state.function = (void*)MCast_MemberExpire;		/* expire function */
	_rtl8651_mod_timer(&member->timer_state, _sysUpSeconds + rtl8651_drvMCastMember_ExpireTime);

	_rtl8651_init_timer(&member->timer_query);
	member->timer_query.data = (uint32) member;
	member->timer_query.function = (void*)MCast_MemberSendQuery;	/* periodic Query sending function */
	_rtl8651_mod_timer(&member->timer_query, _sysUpSeconds + rtl8651_igmp_group_query_interval);

	_rtl8651_init_timer(&member->timer_filterMode);
	member->timer_filterMode.data = (uint32)member;
	member->timer_filterMode.function = (void*)MCast_MemberExcludeExpire;	/* expire function of EXCLUDE mode */
	_rtl8651_mod_timer(&member->timer_filterMode, _sysUpSeconds + rtl8651_igmp_gmi);

	// for query expire function, we only activate it when we sending query
	_rtl8651_init_timer(&(member->timer_query_expire));
	member->timer_query_expire.data = (uint32) member;
	member->timer_query_expire.function = (void*)MCast_MemberExpire;		/* expire function */


	// add counter
	m_entry->memberCnt ++;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Create Member OK !! Member Count: [%d]\n", m_entry->memberCnt);
	#endif

	// add member count
	rtl8651_MemberCnt ++;

	if (ModifyDataFlow == TRUE)
	{
	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("[MCAST] *FLOW* Add Member(0x%X)[0x%X] to {%u.%u.%u.%u}\n",
		member->vlan->vid, member->portMask, NIPQUAD(m_entry->mcast_addr));
	#endif
	_rtl8651_addMulticastMember(m_entry->mcast_addr, (1<<((member->vlan->vid)&(RTL8651_VLAN_NUMBER-1))), member->portMask, FALSE);
	}

	rtlglue_drvMutexUnlock();//UnLock resource
	return member;
errout:
	rtlglue_drvMutexUnlock();//UnLock resource
	return NULL;
}
// remove [member] from [m_entry]
void MCast_RemoveMember(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrvMCastMember_t *member)
{
	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Remove Member (%u.%u.%u.%u) Member Count: [%d]\n", NIPQUAD(m_entry->mcast_addr), m_entry->memberCnt);
	#endif

	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("[MCAST] *FLOW* Remove Member(0x%X)[0x%X] from {%u.%u.%u.%u}\n",
		member->vlan->vid, member->portMask, NIPQUAD(m_entry->mcast_addr));
	#endif

	rtlglue_drvMutexLock();//Lock resource

	_rtl8651_delMulticast(m_entry->mcast_addr, (1<<((member->vlan->vid)&(RTL8651_VLAN_NUMBER-1))), member->portMask);

	// del timer
	_rtl8651_del_timer(&member->timer_filterMode);
	_rtl8651_del_timer(&member->timer_state);
	_rtl8651_del_timer(&member->timer_query);
	_rtl8651_del_timer(&member->timer_query_expire);

	MCast_MemberUnLink(member);

	// release all slist
	MCast_SourceList_freeAll(member->slist);

	MCast_free(member);

	// decrease counter
	m_entry->memberCnt --;
	rtl8651_MemberCnt --;

	rtlglue_drvMutexUnlock();//UnLock resource

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Remove Member OK !! Member Count: [%d]\n", m_entry->memberCnt);
	#endif
}
// destory [member] and try to destroy m_entry of this member
void MCast_DestroyMember(rtl8651_tblDrvMCastMember_t *member)
{
	rtl8651_tblDrvMCast_t *m_entry;

	rtlglue_drvMutexLock();//Lock resource

	m_entry = member->m_entry;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Destroy Member! (%u.%u.%u.%u) Member Count: [%d]\n", NIPQUAD(m_entry->mcast_addr), m_entry->memberCnt);
	#endif

	MCast_RemoveMember(m_entry, member);
	if (m_entry->memberCnt == 0)	// destory m_entry
		MCast_RemoveMCastEntry(m_entry);

	rtlglue_drvMutexUnlock();//UnLock resource
}
// link member into m_entry's member list
void MCast_MemberLink(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrvMCastMember_t *member)
{
	rtlglue_drvMutexLock();//Lock resource
	member->m_entry = m_entry;
	if (m_entry->member)
	{
		member->next = m_entry->member;
		m_entry->member->prev = member;
	}
	m_entry->member = member;
	rtlglue_drvMutexUnlock();//Lock resource
}
// unlinke member from m_entry's member list
int32 MCast_MemberUnLink(rtl8651_tblDrvMCastMember_t *member)
{
	rtl8651_tblDrvMCast_t *m_entry;

	rtlglue_drvMutexLock();//Lock resource
	m_entry = member->m_entry;
	if (!m_entry)
		goto errout;

	if (m_entry->member == member)
		m_entry->member = member->next;
	if (member->next)
		member->next->prev = member->prev;
	if (member->prev)
		member->prev->next = member->next;

	member->m_entry = NULL;
	member->prev = NULL;
	member->next = NULL;

	rtlglue_drvMutexUnlock();//UnLock resource
	return SUCCESS;
errout:
	rtlglue_drvMutexUnlock();//UnLock resource
	return FAILED;
}
// for IGMP v3 , filter mode change
/*
	For V3
		1. for Exclude mode:
			1) change to Include mode
			2) remove the source list which were expired
*/
void MCast_MemberExcludeExpire(uint32 expire)
{
	rtl8651_tblDrvMCastMember_t *member = (rtl8651_tblDrvMCastMember_t *)expire;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Member FilterMode Change! (%u.%u.%u.%u) time [%d] \n", NIPQUAD(member->m_entry->mcast_addr), _sysUpSeconds);
	#endif

	if (member->FilterMode == IGMP_FILTER_EXCLUDE)
	{
		rtl8651_tblDrvMCastSource_t *ptr;
		member->FilterMode = IGMP_FILTER_INCLUDE;
		ptr = member->slist;
		while (ptr)
		{
			if (ptr->timer.expires <= _sysUpSeconds)
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
				ptr = ptr->next;
		}
		if (member->MulticastRouterVersion == IGMPV3)
		{
			_rtl8651_delMulticast(member->m_entry->mcast_addr, 0, 0);
			if (!member->slist)
				MCast_MemberExpire((uint32)member);	// INCLUDE {} -> expire member
		}
	}

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Member FilterMode Change OK! time [%d] \n", _sysUpSeconds);
	#endif

}
// expire function of member
/*
	1. member enter IGMP_MEMBER_STATE_WILLTIMEOUT
	2. initial and send first Group-Specific/General Query
	3. initial timer
*/
void MCast_MemberExpire(uint32 expire)
{
	rtl8651_tblDrvMCastMember_t *member = (rtl8651_tblDrvMCastMember_t *)expire;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Member EXPIRE! (%u.%u.%u.%u) time [%d] --------> migth be Destroyed at [%d]\n", NIPQUAD(member->m_entry->mcast_addr), _sysUpSeconds, _sysUpSeconds + rtl8651_drvMCastMember_DestroyTime);
	#endif

	member->state			= IGMP_MEMBER_STATE_WILLTIMEOUT;
	member->remainingQuery	= rtl8651_igmp_lmqc;

	/* remove query expiration timer */
	_rtl8651_del_timer(&(member->timer_query_expire));

	if (!member->remainingQuery)
	{
		MCast_DestroyMember(member);
		return;
	}

	/* if this member is in expire state, we don't forward to it if this option is turned-on */
	if (rtl8651_multicastProcessOptions & MCAST_PROC_EXPIRE_STATE_DONT_FWD)
	{
		#ifdef RTL8651_MCAST_DATA_DEBUG
		rtlglue_printf("[MCAST] *FLOW* Remove Member(0x%X)[0x%X] from {%u.%u.%u.%u}\n",
		member->vlan->vid, member->portMask, NIPQUAD(member->m_entry->mcast_addr));
		#endif
		_rtl8651_delMulticast(member->m_entry->mcast_addr, (1<<((member->vlan->vid)&(RTL8651_VLAN_NUMBER-1))), member->portMask);
	}



	member->timer_state.data = (uint32) member;						/* pointer to member */
	member->timer_state.function = (void*)MCast_DestroyMember;		/* destory member and MCast Entry */
	_rtl8651_mod_timer(&member->timer_state, _sysUpSeconds + rtl8651_drvMCastMember_DestroyTime);

	if (member->isForwarder == TRUE)
		MCast_MemberSendQuery(member);
}
// periodic send query
void MCast_MemberSendQuery(rtl8651_tblDrvMCastMember_t *member)
{
	uint32 expire_time;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Send Query to (%u.%u.%u.%u) time [%d]\n", NIPQUAD(member->m_entry->mcast_addr), _sysUpSeconds);
	#endif

	// reset isForwarder
	member->isForwarder = TRUE;

	if (member->state == IGMP_MEMBER_STATE_NORMAL)
	{	// in normal state, send query periodically
		expire_time = rtl8651_igmp_group_query_interval;
		// for member at state "IGMP_MEMBER_STATE_NORMAL", we set expire timer "timer_query_expire" if no response in "rtl8651_igmp_qri" interval
		_rtl8651_mod_timer(&(member->timer_query_expire), _sysUpSeconds + rtl8651_igmp_qri + rtl8651_igmp_response_tolerance_delay);
		member->timer_query_expire.data = (uint32)member;
		member->timer_query_expire.function = (void*)MCast_MemberExpire;		/* expire function */
	}else
	{	// member->state == IGMP_MEMBER_STATE_WILLTIMEOUT
		if (!member->remainingQuery)
			return;
		expire_time = rtl8651_igmp_lmqi;
		member->remainingQuery --;
	}

	// re-initial timer
	member->timer_query.data = (uint32)member;
	member->timer_query.function = (void*)MCast_MemberSendQuery;
	_rtl8651_mod_timer(&member->timer_query, _sysUpSeconds + expire_time);

	// check for state-transition timeout, system cannot change state before query expiration
	if ((_sysUpSeconds + rtl8651_igmp_qri + rtl8651_igmp_response_tolerance_delay) > (member->timer_state.expires))
		_rtl8651_mod_timer(&member->timer_state, _sysUpSeconds + rtl8651_igmp_qri + rtl8651_igmp_response_tolerance_delay);

	// send Query
	if (member->MulticastRouterVersion == IGMPV1)
	{	// send general query
		rtl8651_tblDrvMCastRouteInfo_t info;

		info.netif		= NULL;
		info.ifInfo		= 0;
		info.m_addr	= IGMP_ALL_HOSTS;
		info.netif		= NULL;
		info.ifInfo		= 0;
		info.vlan		= member->vlan;
		info.port		= member->portMask;
		info.isPortMask= TRUE;

		MCast_QueryPack(NULL, &info, IGMPV1, NULL, 0, NULL);
	}else
		// IGMPV2/3, send Group specific query
		MCast_QueryPack(member, NULL, member->MulticastRouterVersion, NULL, 0, NULL);
}

/******************
	source-list
*******************/
// create
rtl8651_tblDrvMCastSource_t *MCast_SourceList_Create()
{
	rtl8651_tblDrvMCastSource_t *ret = NULL;
	if ((ret = MCast_malloc(sizeof(rtl8651_tblDrvMCastSource_t))) == NULL)
		return NULL;
	_rtl8651_init_timer(&ret->timer);
	ret->timer.data = (uint32)ret;
	ret->timer.function = (void*)MCast_SourceList_Expire;
	rtl8651_SourceListCnt ++;

	return ret;
}
// link
void MCast_SourceList_Link(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrvMCastMember_t *member, rtl8651_tblDrvMCastSource_t *src)
{
	rtl8651_tblDrvMCastSource_t **listpp;
	if ((!src) || (m_entry && member) || (!m_entry && !member))
		return;

	if (m_entry)
		listpp = &m_entry->slist;
	else
	{
		listpp = &member->slist;
		src->mbr = member;
	}

	src->prev = NULL;
	src->next = (*listpp);
	if (*listpp)
		(*listpp)->prev = src;
	*listpp = src;
}

// unlink
void MCast_SourceList_UnLink(rtl8651_tblDrvMCastSource_t *src)
{
	if (src->next)
		src->next->prev = src->prev;
	if (src->prev)
		src->prev->next = src->next;
	if (src->mbr && (src->mbr->slist == src))
		src->mbr->slist = src->next;

	src->prev = NULL;
	src->next = NULL;
}
// return if "addr" in "slist" or not
rtl8651_tblDrvMCastSource_t* MCast_SourceList_inList(ipaddr_t addr, rtl8651_tblDrvMCastSource_t *slist)
{
	rtl8651_tblDrvMCastSource_t *ptr = slist;

	if (!ptr)
		return ptr;

	while (ptr)
	{
		if (ptr->source == addr)
			return ptr;
		ptr = ptr->next;
	}

	return NULL;
}

// return entry count of one source List
uint32 MCast_SourceList_EntryCnt(rtl8651_tblDrvMCastSource_t *slist)
{
	rtl8651_tblDrvMCastSource_t *ptr = slist;
	uint32 cnt = 0;

	while (ptr)
	{
		cnt ++;
		ptr = ptr->next;
	}

	return cnt;
}

// free current entry and return next entry
// might cause slist become empty: it is deired because NULL source list is OK
rtl8651_tblDrvMCastSource_t *MCast_SourceList_freeCurrent(rtl8651_tblDrvMCastSource_t *s_entry)
{
	rtl8651_tblDrvMCastSource_t *ret;
	if (!s_entry)
		return NULL;

	// unlink it
	ret = s_entry->next;

	MCast_SourceList_UnLink(s_entry);
	_rtl8651_del_timer(&s_entry->timer);
	MCast_free(s_entry);
	rtl8651_SourceListCnt --;

	return ret;
}

// free all entries in source list "slist"
void MCast_SourceList_freeAll(rtl8651_tblDrvMCastSource_t *slist)
{
	rtl8651_tblDrvMCastSource_t *ptr = slist;
	if (!ptr)
		return;

	while(ptr->prev)		// trace to the first entry
		ptr = ptr->prev;

	if (ptr->mbr)
	{
		assert(ptr->mbr->slist == ptr);
		ptr->mbr->slist = NULL;
	}

	while (ptr)
		ptr = MCast_SourceList_freeCurrent(ptr);

	return;
}

// fill *removalResult with {S1-S2} and return {list count of removalResult}
uint32 MCast_SourceList_removal(
	rtl8651_tblDrvMCastSource_t *s1,
	uint8 s1constraint,
	uint8 s1action,
	rtl8651_tblDrvMCastSource_t *s2,
	uint8 s2constraint,
	uint8 newSrcAction,
	rtl8651_tblDrvMCastSource_t **removalResult)
{
	rtl8651_tblDrvMCastSource_t *tmp;
	rtl8651_tblDrvMCastSource_t *ptr = s1;
	rtl8651_tblDrvMCastSource_t *ret = NULL;
	rtl8651_tblDrvMCastSource_t* newp = NULL;
	uint32 count = 0;

	if (!ptr)
		goto out;

	while(ptr)
	{
		if (
			((s1constraint == IGMP_SRCLIST_CONSTRAINT_NONEXPIRE) && (ptr->timer.expires <= _sysUpSeconds)) ||
			((s1constraint == IGMP_SRCLIST_CONSTRAINT_EXPIRE) && (ptr->timer.expires > _sysUpSeconds))
		)
			goto next_s1;
		if ((tmp = MCast_SourceList_inList(ptr->source, s2)) == NULL)	// source is not in list
		{
match:
			if ((newp = MCast_SourceList_Create()) == NULL)
				goto out;	// cannot allocate new entry: return exist source list

			newp->next = ret;
			if (ret)
				ret->prev = newp;
			ret = newp;

			newp->source = ptr->source;
			newp->mbr = NULL;
			_rtl8651_init_timer(&newp->timer);

			// action
			if (s1action == IGMP_SRCLIST_ACTION_SETLMQT)
				_rtl8651_mod_timer(&ptr->timer, _sysUpSeconds + rtl8651_igmp_lmqt);

			if (newSrcAction== IGMP_SRCLIST_ACTION_SETLMQT)
				_rtl8651_mod_timer(&newp->timer, _sysUpSeconds + rtl8651_igmp_lmqt);

			count ++;
		}else
		{
			if (
				((s2constraint == IGMP_SRCLIST_CONSTRAINT_NONEXPIRE) && (tmp->timer.expires <= _sysUpSeconds)) ||
				((s2constraint == IGMP_SRCLIST_CONSTRAINT_EXPIRE) && (tmp->timer.expires > _sysUpSeconds))
				)
				goto match;
		}
next_s1:
		ptr = ptr->next;
	}
out:
	*removalResult = ret;
	return count;
}

// add one source list to member
// refresh timer to GMI if "refresh" is set
// the "add" source list will be destroyed after calling this function
void MCast_SourceList_addList(rtl8651_tblDrvMCastMember_t *member, rtl8651_tblDrvMCastSource_t *add, uint8 refreshAdded, uint8 refreshExist)
{
	rtl8651_tblDrvMCastSource_t *tmp, *ptr = add;

	while(ptr)
	{
		if ((tmp = MCast_SourceList_inList(ptr->source, member->slist)) != NULL)
		{	// remove added entry
			ptr = MCast_SourceList_freeCurrent(ptr);
			if (refreshExist == TRUE)
				_rtl8651_mod_timer(&tmp->timer, (_sysUpSeconds + rtl8651_igmp_gmi));
		}else
		{	// move this entry from old linked list to new linked list
			// unlink old linked list
			tmp = ptr;
			ptr = ptr->next;
			MCast_SourceList_UnLink(tmp);
			MCast_SourceList_Link(NULL, member, tmp);
			if (refreshAdded == TRUE)
				_rtl8651_mod_timer(&tmp->timer, (_sysUpSeconds + rtl8651_igmp_gmi));
		}
	}
}

// duplicate org to dup,
// if member != NULL, link dup to member and reset their Timeout!
// if member == NULL, return with dup
int32 MCast_SourceList_dup(rtl8651_tblDrvMCastMember_t *member, rtl8651_tblDrvMCastSource_t *org, rtl8651_tblDrvMCastSource_t **dup)
{
	rtl8651_tblDrvMCastSource_t *ptr, *newp;

	if ((!org) || (!dup && !member) || (dup && member))
		return FAILED;

	if (dup)
		*dup = NULL;

	ptr = org;

	while (ptr)
	{
		newp = MCast_SourceList_Create();
		if (!newp)
			return FAILED;
		newp->source = ptr->source;
		if (member)
			MCast_SourceList_addList(member, newp, TRUE, TRUE);
		else
		{
			newp->next = *dup;
			if (*dup)
				(*dup)->prev = newp;
			*dup = newp;
		}
		ptr = ptr->next;
	}
	return SUCCESS;
}

/*
	function to process expired s_entry
*/
void MCast_SourceList_Expire(rtl8651_tblDrvMCastSource_t *s_entry)
{
	rtl8651_tblDrvMCastMember_t *mbr;
	if ((!s_entry) || (!s_entry->mbr))
		return;

	mbr = s_entry->mbr;

	if (mbr->FilterMode == IGMP_FILTER_EXCLUDE)
	{
		if (mbr->MulticastRouterVersion == IGMPV3)
			_rtl8651_delMulticast(mbr->m_entry->mcast_addr, 0, 0);
	}else if (mbr->FilterMode == IGMP_FILTER_INCLUDE)
	{
		// Expired and its corresponding member entry's filter mode is INCLUDE
		MCast_SourceList_freeCurrent(s_entry);
		if (mbr->MulticastRouterVersion == IGMPV3)
		{
			// flush mulitcast cache for source list modification
			_rtl8651_delMulticast(mbr->m_entry->mcast_addr, 0, 0);

			if (!mbr->slist)
				MCast_MemberExpire((uint32)mbr);
		}
	}else
	{
		assert(0);
	}
}

/*
	MCast_SourceList_getCache

	calculate multicast's source list cache
	and return current Filter mode
*/
uint32 MCast_SourceList_getCache(rtl8651_tblDrvMCast_t *m_entry, rtl8651_tblDrvMCastSource_t **newCache)
{
	rtl8651_tblDrvMCastMember_t *mptr, *firstmptr, *firstExcludeMember;
	uint8 mFilterMode = IGMP_FILTER_NONE;
	uint8 allMemberNone;
	rtl8651_tblDrvMCastSource_t  *ret = NULL, *tmp, *tmp_2, *newp;

	if (!m_entry)
		goto list_add_done;

	assert(m_entry->memberCnt > 0);
	firstmptr = m_entry->member;
	// trace member for new FilterMode
	firstExcludeMember = NULL;				// if we find EXCLUDE, we must record it for further source list construction
	allMemberNone = TRUE;	// if all members have "NONE" mode, then the cache of this mulitcast entry is "NONE"
	mptr = firstmptr;
	while (mptr)
	{	// if any one member of m_entry has the "EXCLUDE" FilterMode, this m_entry's FilterMode will be "EXCLUDE"

		// refresh filter mode (member exist with INCLUDE{NULL}: it must be IGMPv1/v2 previously)
		if ((mptr->FilterMode == IGMP_FILTER_INCLUDE) && (mptr->slist == NULL))
			mptr->FilterMode = IGMP_FILTER_EXCLUDE;

		if (mptr->FilterMode == IGMP_FILTER_EXCLUDE)
		{
			mFilterMode = IGMP_FILTER_EXCLUDE;
			if (!firstExcludeMember)
				firstExcludeMember = mptr;
		}
		if (mptr->FilterMode != IGMP_FILTER_NONE)
			allMemberNone = FALSE;
		mptr = mptr->next;
	}
	if (allMemberNone == TRUE)
	{
		mFilterMode = IGMP_FILTER_NONE;
		ret = NULL;
		goto list_add_done; 
	}
	
	if (mFilterMode == IGMP_FILTER_NONE)	// no any member has filtermode "EXCLUDE", so the cached filtermode is "INCLUDE"
		mFilterMode = IGMP_FILTER_INCLUDE;

	// we get the new filter mode, now we can determine new source list of m_entry
	ret = NULL;	// init source list
	if (mFilterMode == IGMP_FILTER_INCLUDE)
	{	//	RFC 3376: $3.2
		//	new mFilterMode is "INCLUDE", it means that all of members have "INCLUDE" filter mode.
		//	So the new Source List of m_entry is the union all source lists of all members
		mptr = firstmptr;
		while (mptr)
		{
			assert(mptr->MulticastRouterVersion >= IGMPV3);
			tmp = mptr->slist;
			while (tmp)
			{
				if (!MCast_SourceList_inList(tmp->source, ret))
				{
					if ((newp = MCast_SourceList_Create()) == NULL)
						goto list_add_done;
					// copy source and link to "ret"
					newp->source = tmp->source;
					newp->next = ret;
					if (ret)
						ret->prev = newp;
					ret = newp;
				}
				tmp = tmp->next;
			}
			mptr = mptr->next;
		}
	}else
	{	//	RFC 3376: $3.2
		//	new mFilterMode is "EXCLUDE", it means that some of members have "EXCLUDE" filter mode.
		//	So we must set the default source list to one of the EXCLUDE source list,
		//	and trace all member's source list:
		//		INCLUDE{A}:
		//			remove {A} from current list
		//		EXCLUDE{A}:
		//			{current list} = {current list} & {A}

		// set the initial source list: get source list from "firstExcludeMember" (recorded from previous member tracing)
		assert(firstExcludeMember);
		tmp = firstExcludeMember->slist;
		while (tmp)
		{
			if ((newp = MCast_SourceList_Create()) == NULL)
				goto list_init_done;

			newp->source = tmp->source;
			newp->next = ret;
			if (ret)
				ret->prev = newp;
			ret = newp;

			tmp = tmp->next;			
		}
	list_init_done:
		// start trace all members
		mptr = m_entry->member;
		while (mptr)
		{
			if (mptr->MulticastRouterVersion < IGMPV3)	// ignore it
				goto ex_next_member;
			if (!ret)
				goto list_add_done;
			if (mptr != firstExcludeMember)
			{
				if (mptr->FilterMode == IGMP_FILTER_INCLUDE)
				{
					tmp = mptr->slist;
					while (tmp)
					{
						if ((tmp_2 = MCast_SourceList_inList(tmp->source, ret)) != NULL)
						{
							if (tmp_2 == ret)
								ret = MCast_SourceList_freeCurrent(tmp_2);	// change ret to next entry
							else
								MCast_SourceList_freeCurrent(tmp_2);		// don't change ret
						}
						if (!ret)
							goto list_add_done;

						tmp = tmp->next;
					}
				}else
				{	// (mptr->FilterMode == IGMP_FILTER_EXCLUDE)
					tmp = ret;
					while (tmp)
					{
						if (!MCast_SourceList_inList(tmp->source, mptr->slist))
						{	// remove {new source list} - {this member's source list}
							if (tmp == ret)
							{
								ret = MCast_SourceList_freeCurrent(tmp);	// change ret to next entry
								tmp = ret;
							} else
								tmp = MCast_SourceList_freeCurrent(tmp);	// don't change ret
						} else
							tmp = tmp->next;
						if (!ret)
							goto list_add_done;
					}
				}
			}
	ex_next_member:
			mptr = mptr->next;
		}
	}
list_add_done:
	if (newCache)
		*newCache = ret;
	return mFilterMode;
}

/******************
	Pkt Process
******************/
// Get interface IP by vlan
ipaddr_t MCast_GetIpByVlan(rtl8651_tblDrv_vlanTable_t *vlanp, uint32 vid)
{
	rtl8651_tblDrv_networkIntfTable_t	*netif;
	rtl8651_tblDrv_ipIntfEntry_t			*ipIf;
	rtl8651_tblDrv_ipIntfIpAddrEntry_t	*ipEntry;

	if (!vlanp && !vid)
	{
		return 0;
	}

	if (vid)
		netif = _rtl8651_getNetworkIntfEntryByVid(vid);
	else
		netif = _rtl8651_getNetworkIntfEntryByVid(vlanp->vid);

	if (netif)
	{
		SLIST_FOREACH(ipIf, &netif->IpHead, nextIp)
		{
			SLIST_FOREACH(ipEntry, &ipIf->ipAddr, nextIpAddr)
			{
				if (ipEntry->ipAddr)
					return (ipEntry->ipAddr);
			}
		}
	}
	return 0;
}
// get truth vlan/port to set ASIC
void inline MCast_asicSetActualVlanPort(uint16 *pPortMask, uint32 *pVlanMask, rtl8651_tblDrv_networkIntfTable_t *netif, rtl8651_tblDrv_vlanTable_t* vlan, uint16 orgPortMask, uint32 intfInfo)
{
	assert(pPortMask && pVlanMask);
	assert(netif && vlan);

	switch (netif->linkLayerType) {
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		case RTL8651_LL_PPTP:
		case RTL8651_LL_L2TP:
		{
			rtl8651_tblDrv_pppTable_t *ptb;
			assert(intfInfo);
			
			ptb = ((rtl8651_tblDrvSession_t *)intfInfo)->session_t;

			*pVlanMask = (ptb->initiated)?(1<<(tunnel.loopBackVid & (RTL8651_VLAN_NUMBER - 1))):0;
			*pPortMask = (ptb->initiated)?(1<<CONFIG_RTL865XB_PPTPL2TP_PORT):0;
		}
			break;
#endif /* CONFIG_RTL865X_PPTPL2TP , CONFIG_RTL865XB_PPTPL2TP */
		default:
			*pPortMask = orgPortMask;
			*pVlanMask = 1<<(vlan->vid & (RTL8651_VLAN_NUMBER - 1));
	}
}

// Set routing Information about (netif/ifInfo) and get SIP for an input routing information
int32 MCast_SetRouteInfoAndGetIntfIp(ipaddr_t *gip, rtl8651_tblDrvMCastRouteInfo_t *info)
{	
	ipaddr_t ip;

	assert(info->vlan);	// (vlan) of info must be set

	if ((!(info->netif)) && 
		((info->netif = _rtl8651_getNetworkIntfEntryByVid(info->vlan->vid)) == NULL))
	{
		return FAILED;
	}
	switch (info->netif->linkLayerType)
	{
		case RTL8651_LL_VLAN:
				if (info->ifInfo)
				{	// Gateway's Interface ip is set
					ip = info->ifInfo;
				}else
				{
					if (SLIST_FIRST(&(info->netif->IpHead)) == NULL)
						return FAILED;
					if (SLIST_FIRST(&(SLIST_FIRST(&(info->netif->IpHead)))->ipAddr) == NULL)
						return FAILED;
					if ((SLIST_FIRST(&(SLIST_FIRST(&(info->netif->IpHead)))->ipAddr))->ipAddr == 0)
						return FAILED;
					ip = info->ifInfo = SLIST_FIRST(&(SLIST_FIRST(&(info->netif->IpHead))->ipAddr))->ipAddr;
				}
			break;
		case RTL8651_LL_PPPOE:
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		case RTL8651_LL_PPTP:
		case RTL8651_LL_L2TP:
#endif /* CONFIG_RTL865X_PPTPL2TP , CONFIG_RTL865XB_PPTPL2TP */
				if (!info->ifInfo)
				{
					if ((info->ifInfo = (uint32)(CTAILQ_FIRST(&(info->netif->sessionHead)))) == 0)
					{
						return FAILED;
					}
				}
				if ((!((rtl8651_tblDrvSession_t*)(info->ifInfo))->ip_t) ||
					(!((rtl8651_tblDrvSession_t*)(info->ifInfo))->ip_t->ipAddr))
				{
					return FAILED;
				}
				ip = ((rtl8651_tblDrvSession_t*)(info->ifInfo))->ip_t->ipAddr;
			break;
		default:
			return FAILED;
	}

	if (gip)
		*gip = ip;
	return SUCCESS;
}
/*
	Process each RecordType and modify member entry's source list
*/
// ISIN
void MCast_Proc_ISIN(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	// INCLUDE(A)	IS_IN(B)		-> (B) = GMI
	// EXCLUDE(X, Y)	IS_IN(A)		-> (A) = GMI
	MCast_SourceList_addList(member, slist, TRUE, TRUE);
}
// ISEX
static inline void MCast_Proc_ISEX(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	rtl8651_tblDrvMCastSource_t *ptr, *tmp;
	// INCLUDE(A)	IS_EX(B)		-> (B-A) = 0, Delete (A-B), Group Timer = GMI
	// EXCLUDE(X,Y)	IS_EX(A)		-> (A-X-Y) = GMI, Delete (X-A), Delete (Y-A), Group Timer = GMI
	if (member->FilterMode == IGMP_FILTER_INCLUDE)
	{
		// delete (A-B)
		ptr = member->slist;
		while(ptr)
		{
			if (!MCast_SourceList_inList(ptr->source, slist))
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
				ptr = ptr->next;
		}
		// (B-A) = 0
		ptr = slist;
		while(ptr)
		{
			if (!MCast_SourceList_inList(ptr->source, member->slist))
			{	// add this entry to member's source list and set it timer to 0
				tmp = ptr->next;
				MCast_SourceList_UnLink(ptr);
				MCast_SourceList_Link(NULL, member, ptr);
				_rtl8651_del_timer(&ptr->timer);	// set the timer to 0
				ptr->timer.expires	= _sysUpSeconds;
				ptr = tmp;
			}else
			{	// (B*A), remove ptr
				ptr = MCast_SourceList_freeCurrent(ptr);
			}
		}
	}else		// (member->FilterMode == IGMP_FILTER_EXCLUDE)
	{
		// delete (X-A) and (Y-A)
		ptr = member->slist;
		while(ptr)
		{
			if (!MCast_SourceList_inList(ptr->source, slist))
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
				ptr = ptr->next;
		}
		// (A-X-Y) = GMI :add A to member's source list, but only set the newly-added source list timeout to GMI
		MCast_SourceList_addList(member, slist, TRUE, FALSE);
	}
	member->FilterMode = IGMP_FILTER_EXCLUDE;										// modify filter mode
	_rtl8651_mod_timer(&member->timer_filterMode, (_sysUpSeconds + rtl8651_igmp_gmi));	// reset member's state timer
}
// TOIN
static inline void MCast_Proc_TOIN(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	rtl8651_tblDrvMCastSource_t *sendQsrc = NULL;
	uint32 sendQsrcCnt = 0;

	// INCLUDE(A)	TOIN(B)		-> (B) = GMI, Send Q(G, A-B)
	// EXCLUDE(X, Y)	TOIN(A)		-> (A) = GMI, Send Q(G, X-A), Send Q(G)
	// generate Query List
	if (member->FilterMode == IGMP_FILTER_INCLUDE)	// we must set new timeout to LMQT for entries in member
		sendQsrcCnt = MCast_SourceList_removal(member->slist, 0, IGMP_SRCLIST_ACTION_SETLMQT, slist, 0, 0, &sendQsrc);
	else	// (member->FilterMode == IGMP_FILTER_EXCLUDE)
		sendQsrcCnt = MCast_SourceList_removal(member->slist, IGMP_SRCLIST_CONSTRAINT_NONEXPIRE, IGMP_SRCLIST_ACTION_SETLMQT, slist, 0, 0, &sendQsrc);
	// update list
	MCast_SourceList_addList(member, slist, TRUE, TRUE);
	if (sendQ == TRUE)
	{
		if (member->FilterMode == IGMP_FILTER_INCLUDE)
		{
			if (sendQsrcCnt)	// RFC 3376 $6.4.2: for Source-and-Group Specific query, if source list is zero, we don't send it!
				MCast_QueryPack(member, NULL, member->MulticastRouterVersion, sendQsrc, sendQsrcCnt, NULL);
		}else
		{	// (member->FilterMode == IGMP_FILTER_EXCLUDE)
			if (sendQsrcCnt > 0)	// RFC 3376 $6.4.2: for Source-and-Group Specific query, if source list is zero, we don't send it!
				MCast_QueryPack(member, NULL, member->MulticastRouterVersion, sendQsrc, sendQsrcCnt, NULL);
//			MCast_QueryPack(member, NULL, member->MulticastRouterVersion, NULL, 0, NULL);  // don't send Query here, we enter Expire state directly
			MCast_MemberExpire((uint32)member);
		}
	}
	// free temp src list
	MCast_SourceList_freeAll(sendQsrc);
}
// TOEX
static inline void MCast_Proc_TOEX(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	rtl8651_tblDrvMCastSource_t *ptr;
	rtl8651_tblDrvMCastSource_t *sendQsrc = NULL;
	uint32 sendQsrcCnt = 0;

	// INCLUDE(A)	TOEX(B)		-> (B-A) = 0, Delete (A-B), Send Q(G, A*B), Group Timer=GMI
	// EXCLUDE(X, Y)	TOEX(A)		-> (A-X-Y) = Group Timer, Delete (X-A), Delete (Y-A), SendQ(G,A-Y), Group timer=GMI
	if (member->FilterMode == IGMP_FILTER_INCLUDE)
	{
		// delete (A-B)
		ptr = member->slist;
		while(ptr)
		{
			if (!MCast_SourceList_inList(ptr->source, slist))
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
				ptr = ptr->next;
		}

		sendQsrcCnt = 0;
		ptr = member->slist;
		while (ptr)
		{	// modify timer and add count
			_rtl8651_mod_timer(&ptr->timer, (_sysUpSeconds + rtl8651_igmp_lmqt));
			sendQsrcCnt ++;
			ptr = ptr->next;
		}
		if ((sendQ == TRUE) && (sendQsrcCnt > 0))	// source list counter count and timer modify complete: send Query
			MCast_QueryPack(member, NULL, member->MulticastRouterVersion, member->slist, sendQsrcCnt, NULL);
		// add (B-A) to member's group (All of timers in slist are "zero"(expired))
		MCast_SourceList_addList(member, slist, FALSE, FALSE);
		// set member's filter mode and reset timer
		member->FilterMode = IGMP_FILTER_EXCLUDE;
		_rtl8651_mod_timer(&member->timer_filterMode, (_sysUpSeconds + rtl8651_igmp_gmi));
	}else
	{	// (member->FilterMode == IGMP_FILTER_EXCLUDE)
		// delete (X-A)/(Y-A)
		ptr = member->slist;
		while(ptr)
		{
			if (!MCast_SourceList_inList(ptr->source, slist))
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
				ptr = ptr->next;
		}
		// send Query
		// sendQ (G, A-Y)
		sendQsrcCnt = MCast_SourceList_removal(slist, IGMP_SRCLIST_CONSTRAINT_NONE, IGMP_SRCLIST_ACTION_NONE, member->slist, IGMP_SRCLIST_CONSTRAINT_EXPIRE, IGMP_SRCLIST_ACTION_NONE, &sendQsrc);
		if ((sendQ == TRUE) && (sendQsrcCnt > 0))
		{
			MCast_QueryPack(member, NULL, member->MulticastRouterVersion, sendQsrc, sendQsrcCnt, NULL);
			MCast_SourceList_freeAll(sendQsrc);
		}
		// modify timer (X*A) to LQMT
		ptr = member->slist;
		while (ptr)
		{
			if (ptr->timer.expires <= _sysUpSeconds)
				goto next_ptr;
			if (MCast_SourceList_inList(ptr->source, slist))
				_rtl8651_mod_timer(&ptr->timer, (_sysUpSeconds * rtl8651_igmp_lmqt));
			next_ptr:
				ptr = ptr->next;
		}
		// add (A-X-Y) to member source group and set (A-X-Y) timeout = Group timer
		ptr = slist;
		while(ptr)
		{
			if (MCast_SourceList_inList(ptr->source, member->slist))
			{	// exist, remove ptr
				ptr = MCast_SourceList_freeCurrent(ptr);
			}else
			{	// move from slist to member's slist and set its timeout
				rtl8651_tblDrvMCastSource_t *tmp;
				tmp = ptr->next;
				MCast_SourceList_UnLink(ptr);
				MCast_SourceList_Link(NULL, member, ptr);
				if (member->timer_filterMode.expires > _sysUpSeconds)
					_rtl8651_mod_timer(&ptr->timer, (member->timer_filterMode.expires));
				else
					_rtl8651_mod_timer(&ptr->timer, (_sysUpSeconds + 1));	//expired in next second
				ptr = tmp;
			}
		}
		// Group Timer = GMI
		_rtl8651_mod_timer(&member->timer_filterMode, (_sysUpSeconds + rtl8651_igmp_gmi));
	}
}
// ALLOW
static inline void MCast_Proc_ALLOW(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	// INCLUDE(A)	ALLOW(B)	-> (B) = GMI
	// EXCLUDE(X, Y)	ALLOW(A)	-> (A) = GMI
	MCast_SourceList_addList(member, slist, TRUE, TRUE);
}
// BLOCK
static inline void MCast_Proc_BLOCK(
	rtl8651_tblDrvMCastMember_t	*member,
	rtl8651_tblDrvMCastSource_t	*slist,
	uint16 srcCnt,
	uint8 sendQ
)
{
	rtl8651_tblDrvMCastSource_t *slistP = slist;
	rtl8651_tblDrvMCastSource_t *ptr, *tmp;
	rtl8651_tblDrvMCastSource_t *sendQsrc = NULL;
	uint32 sendQsrcCnt = 0;
	
	// INCLUDE(A)	BLOCK(B)	-> Send Q(G, A*B)
	// EXCLUDE(X, Y)	BLOCK(A)	-> (A-X-Y) = Group Timer, Send Q(G, A-Y)
	if (member->FilterMode == IGMP_FILTER_INCLUDE)
	{
		ptr = member->slist;
		sendQsrcCnt = 0;
		sendQsrc = NULL;
		while (ptr)
		{
			if ((tmp = MCast_SourceList_inList(ptr->source, slistP)) != NULL)
			{	// move this entry from "slist" to "sendQsrc", and set new timer (LMQT) to ptr
				if (slistP == tmp)
					slistP = tmp->next;
				MCast_SourceList_UnLink(tmp);
				tmp->next = sendQsrc;
				if (sendQsrc)
					sendQsrc->prev = tmp;
				sendQsrc = tmp;
				sendQsrcCnt ++;

				_rtl8651_mod_timer(&ptr->timer, (_sysUpSeconds + rtl8651_igmp_lmqt));
			}
			ptr = ptr->next;
		}
		if ((sendQ == TRUE) && (sendQsrcCnt > 0))		// pack Query and send it
			MCast_QueryPack(member, NULL, member->MulticastRouterVersion, sendQsrc, sendQsrcCnt, NULL);
		// remove remaining slist
		MCast_SourceList_freeAll(slistP);
		if (sendQsrcCnt > 0)
			MCast_SourceList_freeAll(sendQsrc);
	}else
	{	// (member->FilterMode == IGMP_FILTER_EXCLUDE)

		// send Q(G, A-Y)
		sendQsrcCnt = MCast_SourceList_removal(slist, IGMP_SRCLIST_CONSTRAINT_NONE, IGMP_SRCLIST_ACTION_NONE, member->slist, IGMP_SRCLIST_CONSTRAINT_EXPIRE, IGMP_SRCLIST_ACTION_NONE, &sendQsrc);
		if ((sendQ == TRUE) && (sendQsrcCnt>0))	// pack Query and send it
			MCast_QueryPack(member, NULL, member->MulticastRouterVersion, sendQsrc, sendQsrcCnt, NULL);

		if (sendQsrcCnt > 0)
		MCast_SourceList_freeAll(sendQsrc);

		ptr = slist;
		while (ptr)
		{	// (A-X-Y) = Group timer
			if (MCast_SourceList_inList(ptr->source, member->slist))
				ptr = MCast_SourceList_freeCurrent(ptr);
			else
			{	// move to member's source list and set it timer to group timer
				tmp = ptr->next;
				MCast_SourceList_UnLink(ptr);
				MCast_SourceList_Link(NULL, member, ptr);
				if (member->timer_filterMode.expires > _sysUpSeconds)
					_rtl8651_mod_timer(&ptr->timer, (member->timer_filterMode.expires));
				else
					_rtl8651_mod_timer(&ptr->timer, (_sysUpSeconds + 1));	//expired in next second
				ptr = tmp;
			}
		}
	}
}

/*
	Process Newly incoming Group Record and update multicast entry's cache for this change

	* if (sendQ) == FALSE, we suppress the send of IGMPv3 Query
	* return the new filter mode of this multicast entry and linke the new source list to *newCache
*/
uint32 MCast_ProcessGroupRecord(
	rtl8651_tblDrvMCast_t *m_entry,
	rtl8651_tblDrvMCastMember_t *member,
	uint8 RecordType,
	rtl8651_tblDrvMCastSource_t *slist,
	uint16 srcCnt,
	uint8 sendQ,
	rtl8651_tblDrvMCastSource_t **newCache)
{
	uint8 mFilterMode = IGMP_FILTER_EXCLUDE;

	if (newCache)
		*newCache = NULL;

	if ((!m_entry) || (!member))
		goto out;

	if (member->FilterMode == IGMP_FILTER_NONE)
	{
		MCast_SourceList_freeAll(member->slist);
		MCast_SourceList_dup(member, slist, NULL);
		switch (RecordType)
		{
			case IGMP_MODE_ISIN:
			case IGMP_MODE_TOIN:
			case IGMP_MODE_ALLOW:
				member->FilterMode = IGMP_FILTER_INCLUDE;
				break;
			case IGMP_MODE_ISEX:
			case IGMP_MODE_TOEX:
			case IGMP_MODE_BLOCK:
				member->FilterMode = IGMP_FILTER_EXCLUDE;
				break;
		}
	}
	// update member's source list, send corresponding query, and remove "slist" entries
	switch (RecordType)
	{
		case IGMP_MODE_ISIN:
			MCast_Proc_ISIN(member, slist, srcCnt, sendQ);
			break;
		case IGMP_MODE_ISEX:
			MCast_Proc_ISEX(member, slist, srcCnt, sendQ);
			break;
		case IGMP_MODE_TOIN:
			MCast_Proc_TOIN(member, slist, srcCnt, sendQ);
			break;
		case IGMP_MODE_TOEX:
			MCast_Proc_TOEX(member, slist, srcCnt, sendQ);
			break;
		case IGMP_MODE_ALLOW:
			MCast_Proc_ALLOW(member, slist, srcCnt, sendQ);
			break;
		case IGMP_MODE_BLOCK:
			MCast_Proc_BLOCK(member, slist, srcCnt, sendQ);
			break;
		default:
			goto out;
	}

	// re-calculate multicast entry m_entry's source list cache
	mFilterMode = MCast_SourceList_getCache(m_entry, newCache);
out:
	return mFilterMode;
}

// Process Query
int32 MCast_query(struct rtl_pktHdr *pkthdr, struct ip *ip, struct igmp *ig)
{
	uint16						igmp_length = 0;
	uint8						version = 0;
	ipaddr_t						dip = 0;
	ipaddr_t						mcast_addr = 0;
	rtl8651_tblDrvMCast_t *			m_entry;
	rtl8651_tblDrvMCastMember_t *	member;
	rtl8651_tblDrv_vlanTable_t *		vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	uint8						keep_pkt = FALSE;	// keep packet for protocol stack

	/*
		Query version Distinctions (RFC 3376):
			- IGMPv1 Query: length = 8 octets AND Max Resp Code field is zero
			- IGMPv2 Query: length = 8 octets AND Max Resp Code field is non-zero
			- IGMPv3 Query: length >= 12 octets
	*/

	igmp_length = ntohs(ip->ip_len) - ((ip->ip_vhl&0xf)<<2);

	if (igmp_length == 8)
	{
		if (ig->igmp_code)	// Max Resp Code
			version = IGMPV2;
		else
			version = IGMPV1;
	}else if (igmp_length >= 12)
		version = IGMPV3;
	else
		goto errout;	// unknown query version: send to protocol stack

	// get mcast addr
	mcast_addr = PKTGET_UINT32_UNALIGNED(&ig->igmp_group);

	/* check filter database first */
	{
		rtl8651_tblDrvMCastFilter_t *filter;
		filter = _rtl8651_multicast_FindFilter(mcast_addr);
		if (filter)
		{
			switch (filter->action)
			{
				case (MCAST_FILTER_ACT_MIRROR|MCAST_FILTER_ACT_NOPROC):
					// trap this packet to protocol stack without process it
					goto errout;
					break;
				case (MCAST_FILTER_ACT_MIRROR):
					// duplicate this packet, process it and trap this packet to protocol stack
					keep_pkt = TRUE;
					break;
				case (MCAST_FILTER_ACT_NOPROC):
					// ignore this packet silently
					goto freeout;
					break;
			}
		}
	}

	if (!IN_MULTICAST(mcast_addr))
	{
		#ifdef RTL8651_MCAST_DEBUG
		rtlglue_printf("##  [MCAST] Receive [General Query] <%d>\n", _sysUpSeconds);
		#endif
		dip = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);
		if (dip == IGMP_ALL_HOSTS)
		{	// general query
			#ifdef RTL8651_MCAST_DEBUG
			rtlglue_printf("##  [MCAST] Receive [General Query] <%d>\n", _sysUpSeconds);
			#endif
			goto GeneralQuery;
		}else
		{	// unknown query: send to protocol stack
			goto errout;
		}
	}

	if (IN_MULTICAST_RESV(mcast_addr))
		goto errout;		// query for Reserve Mutlicast Group: send to protocol stack

/***
	Group-Specific Query
***/

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("##  [MCAST] Receive [Group-Specific Query] for (%u.%u.%u.%u) <%d>\n", NIPQUAD(mcast_addr), _sysUpSeconds);
	#endif

	if ((m_entry = MCast_FindMCastEntry(mcast_addr, FALSE, TRUE)) == NULL)	// system has no entry for this group, ignore it
		goto freeout;

	if ((m_entry->up_vlan) &&
		((m_entry->up_vlan != vlanp) || (m_entry->up_portMask != (1<<pkthdr->ph_portlist))))
	{	// upstream has been set, check if we has member in this port
		member = MCast_FindMember(m_entry, vlanp, pkthdr->ph_portlist, FALSE, FALSE, FALSE);
		if (member)
		{	// Querier Election
			ipaddr_t sip;

			sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);

			if (sip && (MCast_GetIpByVlan(member->vlan, 0) > sip) && rtl8651_igmp_oqpi)
			{
				member->isForwarder = FALSE;
				// re-initial timer
				member->timer_query.data = (uint32)member;
				member->timer_query.function = (void*)MCast_MemberSendQuery;
				_rtl8651_mod_timer(&member->timer_query, _sysUpSeconds + rtl8651_igmp_oqpi);
			}
			// update current version of this member
			member->MulticastRouterVersion = version;
		}
		goto freeout;
	}else
	{	// upstream is not set yet or this message is from upstream
		uint8 ver;
	
		member = MCast_FindMember(m_entry, vlanp, pkthdr->ph_portlist, FALSE, FALSE, FALSE);
		if (member)
		{	// remove member
			MCast_RemoveMember(m_entry, member);
			if (m_entry->memberCnt == 0)	/* entry have no member */
			{	// remove multicast entry and return
				MCast_RemoveMCastEntry(m_entry);
				goto freeout;
			}
		}
		// reset upstream
		MCast_SetMCastEntryUpStream(m_entry, vlanp, pkthdr->ph_portlist, version);
		if (version >= IGMPV3)
		{	// for igmpv3 query, we must update cache before send report
			MCast_SourceList_freeAll(m_entry->slist);
			m_entry->FilterMode = MCast_SourceList_getCache(m_entry, &(m_entry->slist));
		}
		// generate and send report
		ver = MCast_MCastEntryMinVersion(m_entry);
		if (ver < IGMPV3)
			MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, IGMP_FILTER_EXCLUDE, NULL, ((keep_pkt == TRUE)?NULL:pkthdr));
		else
			MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, m_entry->FilterMode, m_entry->slist, ((keep_pkt == TRUE)?NULL:pkthdr));
	}
	goto out;

/***
	General Query
***/
GeneralQuery:
	{
		struct rtl_pktHdr *pkt = pkthdr;
		uint8 proc_entry = TRUE;
		uint16 portlist = pkthdr->ph_portlist;
		uint8 memberVersion = rtl8651_drvMCastVersion;

		/*
			If upstream is set, we only allow General Query from Upstream
		*/
		if ((rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE) &&
			((rtl8651_GlobalVariable.MCastUpStream.vlan != vlanp) ||
			(!(rtl8651_GlobalVariable.MCastUpStream.portMask & (1<<pkt->ph_portlist)))))
			goto freeout;

		m_entry = rtl8651_MCastTable;
		while (m_entry)
		{	//report all entries
			//	Two different kinds of multicast entry need to report:
			//			1. multicast entry without upstream							- report, and set upstream to this interface
			//			2. multicast entry those upstream is where this packet come from	- report
			proc_entry = TRUE;
			if ((!m_entry->up_vlan) ||
				((m_entry->up_vlan) && (m_entry->up_vlan == vlanp) && (m_entry->up_portMask == (1<<portlist))))
			{	// set new upstream for all m_entry without upstream / refresh m_entry with upstream is Default upstream
				if (!m_entry->up_vlan)
				{	// set upstream
					member = MCast_FindMember(m_entry, vlanp, portlist, FALSE, FALSE, FALSE);
					if (member)
					{	// remove member
						MCast_RemoveMember(m_entry, member);
						if (m_entry->memberCnt == 0)	/* entry have no member */
						{	// remove multicast entry and return
							rtl8651_tblDrvMCast_t *next = m_entry->next;
							MCast_RemoveMCastEntry(m_entry);
							m_entry = next;
							proc_entry = FALSE;
						}
					}
				}
				if (proc_entry == TRUE)
				{
				#ifdef RTL8651_MCAST_DEBUG
				rtlglue_printf("##  [MCAST] Process Multicast entry (%u.%u.%u.%u) for [General Query] <%d>\n", NIPQUAD(m_entry->mcast_addr), _sysUpSeconds);
				#endif
					// reset upstream version
					MCast_SetMCastEntryUpStream(m_entry, vlanp, portlist, version);
					// get min member's version
					memberVersion = MCast_MCastEntryMinVersion(m_entry);
					if ((memberVersion >= IGMPV3) && (version >= IGMPV3))
					{	// for igmpv3 query and if all members are IGMPV3, we must update cache before send report
						MCast_SourceList_freeAll(m_entry->slist);
						m_entry->FilterMode = MCast_SourceList_getCache(m_entry, &(m_entry->slist));
					}
					// send report
					if (pkt && (keep_pkt == FALSE))
					{
						if (memberVersion < IGMPV3)
							MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, IGMP_FILTER_EXCLUDE, NULL, pkt);
						else
							MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, m_entry->FilterMode, m_entry->slist, pkt);
						pkt = NULL;
					}else
					{
						if (memberVersion < IGMPV3)
							MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, IGMP_FILTER_EXCLUDE, NULL, NULL);
						else
							MCast_ReportPack(m_entry, version, IGMP_FILTER_NONE, NULL, m_entry->FilterMode, m_entry->slist, NULL);
					}
				}			
			}
			if (proc_entry == TRUE)
			{
				// reset upstream timeout
				if ((rtl8651_drvMCastUpstream_Timeout != 0xffffffff) && (m_entry->up_vlan))
					_rtl8651_mod_timer(&m_entry->timer_upstream, _sysUpSeconds + rtl8651_drvMCastUpstream_Timeout);
				// to next (if proc_query == FALSE, next m_entry has been set)
				m_entry = m_entry->next;
			}
		}
		// goto out: free pkt
		if (pkt && (keep_pkt == FALSE))
			mBuf_freeMbufChain(pkt->ph_mbuf);
	}	// general query
out:
	if (keep_pkt == TRUE)
		return FWDENG_IGMPPKT;
	return SUCCESS;
freeout:
	if (keep_pkt == TRUE)
		return FWDENG_IGMPPKT;

	mBuf_freeMbufChain(pkthdr->ph_mbuf);
	return SUCCESS;
errout:
//	mBuf_freeMbufChain(pkthdr->ph_mbuf);
//	return FAILED;
	return FWDENG_IGMPPKT;
}

// process report
/*
	SUCCESS	: successfully update database
	FAILED		: database update failed
*/
int32 MCast_report(struct rtl_pktHdr *pkthdr, struct ip *ip, struct igmp *ig)
{
	ipaddr_t						mcast_addr;
	uint8						version;
	rtl8651_tblDrvMCast_t *			m_entry;
	rtl8651_tblDrvMCastMember_t *	member;
	rtl8651_tblDrv_vlanTable_t *		vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	uint16						portlist = pkthdr->ph_portlist;
	uint8						keep_pkt = FALSE;
	// l4 length count to prevent IGMPV3 buffer overflow attack
	uint32						l4Len = 0;
	uint32						l4ClaimLen = ntohs(ip->ip_len) - ((*(uint8*)ip&0xf) << 2); //claimed l4 length of this packet
	struct igmp_gr *				gr = NULL;
	int32						LoopCnt = 0;
	uint8						mirrorThis = FALSE;
	// source list
	rtl8651_tblDrvMCastSource_t *	slist;
	uint16						srcCnt;
	uint8						RecordType;
	uint8						newFilterMode = IGMP_FILTER_EXCLUDE;
	rtl8651_tblDrvMCastSource_t *	newSlist = NULL;
	

	// chenyl : 04/04/22: to prevent member from wan
	if (!vlanp->internal)	// ignore IGMP report from WAN
		goto freeout;

	switch (ig->igmp_type)
	{
		case IGMP_V1_REPORT:
			version = IGMPV1;
			break;
		case IGMP_V2_REPORT:
			version = IGMPV2;
			break;
		case IGMP_V3_REPORT:
			version = IGMPV3;
			break;
		default:
			// undefined report: send to protocol stack
			goto errout;
	}

	/*
	 *		Parse packet using while loop,
	 *		before actuall run the while loop to parse packet, we set the initial value of this function
	*/
	if (version >= IGMPV3)
	{	// initial setting, get group record position and loop count is set to number of group record
		LoopCnt	= ntohs(ig->igmp_rnofg);
		gr		= &(ig->igmp_grlist);
	}else
		LoopCnt	= 1;	// for IGMPv1/v2, only loop 1 time

	l4Len	= 0;
	keep_pkt	= FALSE;

	while (LoopCnt > 0)
	{
		if (version < IGMPV3)
			mcast_addr = PKTGET_UINT32_UNALIGNED(&ig->igmp_group);		// V1/V2
		else
			mcast_addr = PKTGET_UINT32_UNALIGNED(&gr->igmp_gr_group);	// V3

		/* check filter database first */
		mirrorThis = FALSE;
		{
			rtl8651_tblDrvMCastFilter_t *filter;
			filter = _rtl8651_multicast_FindFilter(mcast_addr);
			if (filter)
			{
				switch (filter->action)
				{
					case (MCAST_FILTER_ACT_MIRROR|MCAST_FILTER_ACT_NOPROC):
						// system must trap this packet to protocol stack and must ignore this group record
						keep_pkt = TRUE;
						mirrorThis = TRUE;
						goto next_loop;
						break;
					case (MCAST_FILTER_ACT_MIRROR):
						// system must duplicate this packet, process it and trap this packet to protocol stack
						keep_pkt = TRUE;
						mirrorThis = TRUE;
						break;
					case (MCAST_FILTER_ACT_NOPROC):
						// ignore this group record silently
						goto next_loop;
						break;
				}
			}
		}

		// request for reserved multicast group
		if ((!IN_MULTICAST(mcast_addr)) || IN_MULTICAST_RESV(mcast_addr))
		{	// send this packet to protocol stack and process next
			keep_pkt = TRUE;
			goto next_loop;
		}
		// can not create new multicast entry for this report, ignore it
		if ((m_entry = MCast_FindMCastEntry(mcast_addr, TRUE, FALSE)) == NULL)
			goto next_loop;
		// report from upStream of this multicast group, ignore it
		if ((m_entry->up_vlan == vlanp) &&
			(m_entry->up_portMask == (1<<portlist)))
			goto next_loop;
		// can not create new multicast member for this report, ignore it
		if ((member = MCast_FindMember(m_entry, vlanp, portlist, TRUE, TRUE, TRUE)) == NULL)
			goto next_loop;
		// set toCpu bit
		if (mirrorThis == TRUE)
			_rtl8651_setMGroupAttribute(mcast_addr, TRUE);

		// If any new version arrived, we change our version to fit that
		member->MulticastRouterVersion = version;

		// Update Member's source list and multicast entry's source list cache
		if (version < IGMPV3)
		{	// V1/V2 report:	IS_EX{}	[RFC3376 $7.3.2]
			slist			= NULL;
			srcCnt		= 0;
			RecordType	= IGMP_MODE_ISEX;
		}else
		{	// V3
			rtl8651_tblDrvMCastSource_t *s;
			ipaddr_t*	addrPtr = &(gr->src_list);
			uint16 i = 0;

			RecordType	= gr->igmp_gr_rt;
			srcCnt		= ntohs(gr->igmp_gr_nofs);
			slist			= NULL;
			// parse V3 Group record to get source list
			for (i = 0 ; i < srcCnt ; i ++)
			{
				// we have no space to store source anymore: ignore other source
				if ((s = MCast_SourceList_Create()) == NULL)
					break;
				// set the content of this source entry
				s->source	= PKTGET_UINT32_UNALIGNED(addrPtr);
				s->next		= slist;
				if (slist)
					slist->prev = s;
				slist			= s;
				// point to next address
				addrPtr ++;
			}
		}
		// before processing Group Record, we need to update current m_entry cache
		MCast_SourceList_freeAll(m_entry->slist);
		m_entry->FilterMode = MCast_SourceList_getCache(m_entry, &m_entry->slist);
		// process Group Record and update multicast entry cache
		newFilterMode = MCast_ProcessGroupRecord(m_entry, member, RecordType, slist, srcCnt, ((version == IGMPV3)?TRUE:FALSE), &newSlist);

		// send solicited Leave and Report to upstream if upstream of this group is not set
		if ((!m_entry->up_vlan) &&										// m_entry has no entry yet
			(rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE) &&			// default upstream exists
			((rtl8651_GlobalVariable.MCastUpStream.vlan != vlanp) ||((rtl8651_GlobalVariable.MCastUpStream.portMask& ~(1<<portlist)) != 0)))	// default upstream and report are not from the same interface
			// send report to default upstream
		{
			// check all members in m_entry
			uint8 ver = MCast_MCastEntryMinVersion(m_entry);

			if (ver < IGMPV3)
				MCast_ReportPack(m_entry, m_entry->MulticastRouterVersion, m_entry->FilterMode, m_entry->slist, IGMP_FILTER_EXCLUDE, NULL, NULL);
			else
				MCast_ReportPack(m_entry, m_entry->MulticastRouterVersion, m_entry->FilterMode, m_entry->slist, newFilterMode, newSlist, NULL);
		}

		// update m_entry cache
		MCast_SourceList_freeAll(m_entry->slist);
		m_entry->slist = newSlist;
		m_entry->FilterMode = newFilterMode;

		// If this member is IGMPV3, it means there might be some member changes for some sip in mutlicast cache
		// so we  flush mulitcast cache for source list modification
		if (version >= IGMPV3)
			_rtl8651_delMulticast(mcast_addr, 0, 0);

next_loop:
		// end of one loop, system must check some variables and set some variable for next loop
		if (version < IGMPV3)
		{	// V1/V2
			l4Len += 8;	// count l4Length
		}else
		{	// V3
			l4Len += (8 + (uint32)(gr->igmp_gr_auxlen) + (ntohs(gr->igmp_gr_nofs) * sizeof(ipaddr_t)));	// count l4Length
			// point to next group record
			if (LoopCnt > 1)	// not the last loop
				gr = (struct igmp_gr*)(8 /* hdr len */ + (uint32)(gr->igmp_gr_auxlen) + (uint32)(ntohs(gr->igmp_gr_nofs) * sizeof(ipaddr_t))/* source list size */ + (uint32)gr);
		}
		// check for buffer overflow
		if (l4Len > l4ClaimLen)
		{
			assert(0);
			keep_pkt = FALSE;	// free this packet
			goto freeout;
		}
		// decrease Loop Count
		LoopCnt --;
	}

// the original packet arrives here and drop if we don't want to trap it: goto freeout
freeout:
	if (keep_pkt == TRUE)
		return FWDENG_IGMPPKT;

	mBuf_freeMbufChain(pkthdr->ph_mbuf);
	return SUCCESS;
errout:
//	mBuf_freeMbufChain(pkthdr->ph_mbuf);
//	return FAILED;
	return FWDENG_IGMPPKT;
}

// process leave
int32 MCast_leave(struct rtl_pktHdr *pkthdr, struct ip *ip, struct igmp *ig)
{
	rtl8651_tblDrvMCast_t *			m_entry;
	rtl8651_tblDrvMCastMember_t *	member;
	rtl8651_tblDrv_vlanTable_t *		vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	uint16						portlist = pkthdr->ph_portlist;
	ipaddr_t mcast_addr;
	uint8 keep_pkt = FALSE;

	// chenyl : 04/04/22: to prevent member from wan
	if (!vlanp->internal)
		goto freeout;

	mcast_addr = PKTGET_UINT32_UNALIGNED(&(ig->igmp_group));

	/* check filter database first */
	{
		rtl8651_tblDrvMCastFilter_t *filter;
		filter = _rtl8651_multicast_FindFilter(mcast_addr);
		if (filter)
		{
			switch (filter->action)
			{
				case (MCAST_FILTER_ACT_MIRROR|MCAST_FILTER_ACT_NOPROC):
					// trap this packet to protocol stack without process it
					goto errout;
					break;
				case (MCAST_FILTER_ACT_MIRROR):
					// duplicate this packet, process it and trap this packet to protocol stack
					keep_pkt = TRUE;
					break;
				case (MCAST_FILTER_ACT_NOPROC):
					// ignore this packet silently
					goto freeout;
					break;
			}
		}
	}

	if ((!IN_MULTICAST(mcast_addr)) || IN_MULTICAST_RESV(mcast_addr))	// client want to leave group which is reserved : send to protocol stack
		goto errout;

	if ((m_entry = MCast_FindMCastEntry(mcast_addr, TRUE, FALSE)) == NULL)	// can not create new multicast entry for this leave pkt: ignore
		goto freeout;
	else if ((m_entry->up_vlan == vlanp) &&
			(m_entry->up_portMask== (1<<portlist)))		// leave from upstream, ignore
		goto freeout;

	if ((member = MCast_FindMember(m_entry, vlanp, portlist, TRUE, FALSE, (rtl8651_multicastProcessOptions & MCAST_PROC_EXPIRE_STATE_DONT_FWD)?FALSE:TRUE)) == NULL)
		// can not create new multicast member entry for this leave pkt: ignore
		goto freeout;
	// set toCPU bit
	if (keep_pkt == TRUE)
		_rtl8651_setMGroupAttribute(mcast_addr, TRUE);

	member->MulticastRouterVersion = IGMPV2;

	// process source list
	{
		uint8 newFilterMode;
		rtl8651_tblDrvMCastSource_t *newSlist;
		newFilterMode = MCast_ProcessGroupRecord(m_entry, member, IGMP_MODE_TOIN, NULL, 0, FALSE, &newSlist);
	}

	// Set State to Expire state : enter IGMP_MEMBER_STATE_WILLTIMEOUT state
	MCast_MemberExpire((uint32) member);
#if 0
	// Remove Related Flow with sip is source of this packet
	{
		ipaddr_t sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);

		#ifdef RTL8651_MCAST_DATA_DEBUG
		rtlglue_printf("[MCAST] *FLOW* Try to remove Source(%u.%u.%u.%u)(0x%X)[0x%X] from {%u.%u.%u.%u}\n",
			NIPQUAD(sip), (1<<((vlanp->vid)&(RTL8651_VLAN_NUMBER-1))), (1<<portlist), NIPQUAD(mcast_addr));
		#endif
		_rtl8651_delMulticastUpStream(mcast_addr, sip, (1<<((vlanp->vid)&(RTL8651_VLAN_NUMBER-1))), (1<<portlist));
	}
#endif
	// goto freeout;
freeout:
	if (keep_pkt == TRUE)
		return FWDENG_IGMPPKT;

	mBuf_freeMbufChain(pkthdr->ph_mbuf);
	return SUCCESS;
errout:
//	mBuf_freeMbufChain(pkthdr->ph_mbuf);
//	return FAILED;
	return FWDENG_IGMPPKT;
}

/*
	Packet Packets

	Query
		- IGMPv1/2/3
		- parameters:
			[member]	: member to send Query, if [member] == NULL, it's general Query, we use routing information of [info]
			[info]		: routing information for this Query packet, ignored if [member] exists
			[version]		: version of Query to be sent
			[sourceList]	: for IGMPv3, if set, it generate the IGMPv3 Group-Source Specific Query.
			[pkthdr]		: existing pkt to send, if it is NULL, then we initial a new packet
		
	Report
		- IGMPv1/2/3
		- parameters:
			[mEntry]		: for IGMPv1/2, specify a specific multicast entry to report, for IGMPv3, it's ignored
			[version]		: version of Report to be sent
			
	Leave
		- for V2 only
		- parameters:
			[mEntry]		: specify a specific multicast entry to leave
*/
// allocate New Packet or clear old packet
static inline struct rtl_pktHdr *MCast_procPkt(struct rtl_pktHdr *pkt)
{
	if (pkt)
	{
		// clear pkt
		mBuf_adjTail(pkt->ph_mbuf, mBuf_getPktlen(pkt->ph_mbuf));
		return pkt;
	}else
	{
		struct rtl_mBuf *mbuf = NULL;
		if((mbuf = mBuf_get(MBUF_DONTWAIT, MBUFTYPE_DATA, 1)) == NULL)
			return NULL;
		if((mBuf_getPkthdr(mbuf, MBUF_DONTWAIT) == (struct rtl_mBuf *) NULL)||
		(mBuf_reserve(mbuf,rtl8651_fwdEnginePara.mbufHeadroom)))
		{
			mBuf_freeMbufChain(mbuf);
			return NULL;
		}
		return mbuf->m_pkthdr;
	}
}

// pack Query packet
/*
	[member]	: member to send Query, if [member] == NULL, it's general Query, we use routing information of [info]
	[info]		: routing information for this Query packet, ignored if [member] exists
	[version]		: version of Query to be sent
	[sourceList]	: for IGMPv3, if set, it generate the IGMPv3 Group-Source Specific Query.
	[NumOfSrc]	: source list number
	[pkthdr]		: existing pkt to send, if it is NULL, then we initial a new packet
*/
int32 MCast_QueryPack(
	rtl8651_tblDrvMCastMember_t *member,
	rtl8651_tblDrvMCastRouteInfo_t *info,
	uint8 version,
	rtl8651_tblDrvMCastSource_t *sourceList,
	uint16 NumOfSrc,
	struct rtl_pktHdr *pkthdr)
{
	rtl8651_tblDrvMCastRouteInfo_t McastRouteInfo, *route_info = &McastRouteInfo;
	struct rtl_pktHdr	*pkt = pkthdr;
	struct rtl_mBuf	*mbuf;
	ipaddr_t			sip;
	struct ip			*pip;
	struct igmp		*ig;
	uint8			*trailer;
	uint8			isGeneral = FALSE;
	uint32			pktLen = 0;
	rtl8651_tblDrvMCastSource_t *sptr = sourceList;
	uint32			remNumOfSrc = NumOfSrc;

	// Error Checking
	if (
		(!member && !info) ||
		((version == IGMPV3) && sourceList && !member) ||
		((version == IGMPV3) && NumOfSrc && !member))
		goto errout;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Generate IGMPv%d [Query] <%d>\n", version, _sysUpSeconds);
	#endif

	// get sip/maddr
	if (member)
	{	// Group-specific Query
		isGeneral = FALSE;	

		route_info->netif		= NULL;
		route_info->ifInfo		= 0;
		route_info->m_addr	= member->m_entry->mcast_addr;
		route_info->vlan		= member->vlan;
		route_info->port		= member->portMask;
		route_info->isPortMask	= TRUE;
	}else
	{	// General Query
		isGeneral = TRUE;

		*route_info = *info;
		route_info->m_addr = IGMP_ALL_HOSTS;
	}
	// get interface IP
	if (MCast_SetRouteInfoAndGetIntfIp(&sip, route_info) != SUCCESS)
		goto errout;

sendNext:

	// Initiate empty packet
	pkt = MCast_procPkt(pkt);
	if (pkt)
		mbuf = pkt->ph_mbuf;
	else
		goto errout;

	/*
		Construct Packet : traditional vlan packet

		total length :
		V1:
			14 (ethernet) + 20 (ip) + 8 (igmp query) + 18 (trailer) = 60
		V2:
			14 (ethernet) + 24 (ip + Router Alert Option) + 8 (igmp query) + 14 (trailer) = 60
		V3:
			14 (ethernet) + 24 (ip + Router Alter Option) + 12 (igmp query) + (4 * source list count), min = 60
	*/
	pktLen = 0;

	trailer = (uint8 *)(mbuf->m_data + 46);

	// L2
	// smac/dmac is set by MCast_l3ActRouting()
	// Ether type
	*(uint16 *)(mbuf->m_data + 12) = htons(0x0800);
	pktLen += 14;

	// L3
	pip = (struct ip *)(mbuf->m_data + pktLen);
	if (version == IGMPV1)
		pip->ip_vhl	= 0x45;
	else
		pip->ip_vhl	= 0x46;
	pip->ip_tos	= 0;
	pip->ip_id	= htons(SYS_ID);
	pip->ip_off	= htons(0);
	pip->ip_ttl	= 1;
	pip->ip_p	= IPPROTO_IGMP;
	pip->ip_sum	= 0;

	PKTSET_UINT32_UNALIGNED(sip, &pip->ip_src);
	if (isGeneral == TRUE)
		PKTSET_UINT32_UNALIGNED(IGMP_ALL_HOSTS, &pip->ip_dst);
	else
		PKTSET_UINT32_UNALIGNED(route_info->m_addr, &pip->ip_dst);

	if (version == IGMPV1)
		pktLen += 20;		// normal IP Hdr
	else
	{	// V2/3 Router Alert option
		*(uint16*)((uint8*)pip+20) = htons(0x9404);
		*(uint16*)((uint8*)pip+22) = htons(0x0000);
		pktLen += 24;		// IP Hdr + Option
	}

	// L4
	ig = (struct igmp *)(mbuf->m_data + pktLen);
	ig->igmp_type = IGMP_QUERY;
	ig->igmp_cksum = 0;
	if (isGeneral)
		PKTSET_UINT32_UNALIGNED(0, &ig->igmp_group);
	else
		PKTSET_UINT32_UNALIGNED(route_info->m_addr, &ig->igmp_group);

	switch (version)
	{
		case IGMPV1:
			ig->igmp_code = 0;
			pktLen += 8;
			break;
		case IGMPV2:
			{
				if (rtl8651_igmp_qri)
					ig->igmp_code = ((rtl8651_igmp_qri * 10) & 0xff);		// It units of 0.1 second
				else
					ig->igmp_code = 10;	// default 1 second
				pktLen += 8;
			}
			break;
		case IGMPV3:
			{
				uint32 NumOfSrcThisPkt;
				ipaddr_t *pkt_srcPtr;
				
				if (rtl8651_igmp_qri)
					ig->igmp_code = ((rtl8651_igmp_qri * 10) & 0xff);		// It units of 0.1 second
				else
					ig->igmp_code = 10;	// default 1 second

				ig->igmp_qrsq	= 0x01;									// only set Robustness variable to 1
				if (member)
					ig->igmp_qqic	= (uint8)(IGMPV3_QQIC(rtl8651_igmp_group_query_interval));	// group query interval
				else
					ig->igmp_qqic	= (uint8)(IGMPV3_QQIC(rtl8651_igmp_qi));	// general query interval

				// can we put all source list into this packet ?
				if (route_info->netif->mtu < (pktLen + 12 + (remNumOfSrc * sizeof(ipaddr_t)) - 14))
				{
					NumOfSrcThisPkt = ((route_info->netif->mtu - (pktLen + 12)) % sizeof(ipaddr_t));
					remNumOfSrc -= NumOfSrcThisPkt;
				}else
				{
					NumOfSrcThisPkt = remNumOfSrc;
					remNumOfSrc = 0;
				}

				ig->igmp_qnofs	= htons(NumOfSrcThisPkt); 				// source List Number
				pktLen += (12 + (NumOfSrcThisPkt * sizeof(ipaddr_t)));		// IGMP Query + (source list)

				// construct source list
				pkt_srcPtr = (ipaddr_t *)(&(ig->igmp_qsrclist));
				while (NumOfSrcThisPkt)
				{
					assert(sptr);
					if (!sptr)
						goto errout;

					PKTSET_UINT32_UNALIGNED(sptr->source, pkt_srcPtr);
					sptr = sptr->next;
					pkt_srcPtr ++;
					NumOfSrcThisPkt --;
				}
			}
			break;
		default:
			goto errout;
	}

	// set L3 ip_len here
	pip->ip_len = htons((pktLen - 14));	// ip_len field is (total length - ether hdr length)

	// trailer
	if (pktLen < 60)
	{
		trailer = (uint8*)(mbuf->m_data + pktLen);
		memset(trailer, 0, (60 - pktLen));
		pktLen = 60;
	}

	// calculate L3/L4 checksum
	ig->igmp_cksum	= MCast_igmpChecksum(pip);
	pip->ip_sum		= _rtl8651_ipChecksum(pip);
	pkt->ph_flags		&= ~(CSUM_IP|CSUM_L4);

	// pkthdr
	pkt->ph_len = pktLen;
	pkt->ph_mbuf->m_len = pktLen;
	pkt->ph_portlist = 0;
	pkt->ph_pppeTagged = 0;
	pkt->ph_iphdrOffset = 14;
	pkt->ph_proto = PKTHDR_IGMP;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("\t=> Send IGMPv%d [Query] to VLAN[0x%X] Mask[0x%X] <%d>\n", version, route_info->vlan->vid, route_info->port, _sysUpSeconds);
	#endif

	if (MCast_l3ActRouting(pkt, route_info) != SUCCESS)
		goto errout;

	if ((version == IGMPV3) && (remNumOfSrc > 0))
	{	// there are some source list not be sent
		pkt = NULL;	// force system to allocate new packet
		goto sendNext;
	}
	return SUCCESS;
errout:
	if (pkt)
		mBuf_freeMbufChain(pkt->ph_mbuf);
	return FAILED;
}

// pack Report packet
/*
 *	IGMPv3 report			oldFilterMode		oldSourceList		newFilterMode		newSourceList
 *	-------------------------------------------------------------------------------------
 *	current-state			NONE		NULL				Current Mode		CurrentList
*/
int32 MCast_ReportPack(
	rtl8651_tblDrvMCast_t *m_entry,
	uint8 version,
	uint8 oldFilterMode,
	rtl8651_tblDrvMCastSource_t *oldSourceList,
	uint8 newFilterMode,
	rtl8651_tblDrvMCastSource_t *newSourceList,
	struct rtl_pktHdr *pkthdr)
{
	rtl8651_tblDrvMCastRouteInfo_t McastRouteInfo, *route_info = &McastRouteInfo;
	struct rtl_pktHdr	*pkt = pkthdr;
	struct rtl_mBuf	*mbuf;
	ipaddr_t			sip = 0;
	struct ip			*pip;
	struct igmp		*ig;
	uint8			*trailer;
	uint32			pktLen;

	rtl8651_tblDrvMCastSource_t*	tempList[2] = {NULL, NULL};

	rtl8651_tblDrvMCastSource_t*	SourceList[3] = {NULL, NULL, NULL};
	uint8						RecordType[3] = {IGMP_MODE_NONE, IGMP_MODE_NONE, IGMP_MODE_NONE};
	uint32						SourceListCnt[3] = {0, 0, 0};

	rtl8651_tblDrvMCastSource_t	**CurrentSourceListp = NULL;
	uint32						*CurrentSourceListCntp = NULL;
	uint8						*CurrentRecordTypep = NULL;

	if 	((!m_entry) ||
		(
			(version == IGMPV3) &&
			(newFilterMode != IGMP_FILTER_EXCLUDE) &&
			(newFilterMode != IGMP_FILTER_INCLUDE))
		)
		goto release_out;

	/****************************************************************************
		For IGMPv3, we decide what the packet need to be generated

		Old State			New State			State-Change Record Sent
		-----------------------------------------------------------
	(1)	(NULL)			INCLUDE(A)			IS_IN(A)
	(2)	(NULL)			EXCLUDE(A)			IS_EX(A)
	(3)	INCLUDE(A)		INCLUDE(B)			ALLOW(B-A), BLOCK(A-B)
	(4)	EXCLUDE(A)		EXCLUDE(B)			ALLOW(A-B), BLOCK(B-A)
	(5)	INCLUDE(A)		EXCLUDE(B)			TO_EX(B)
	(6)	EXCLUDE(A)		INCLUDE(B)			TO_IN(B)
	****************************************************************************/
	if (version == IGMPV3)
	{
		if (oldFilterMode == IGMP_FILTER_NONE)
		{	// case (1), (2)
			RecordType[0]		= ((newFilterMode == IGMP_FILTER_EXCLUDE)?IGMP_MODE_ISEX:IGMP_MODE_ISIN);
			SourceList[0]			= newSourceList;
			SourceListCnt[0]		= MCast_SourceList_EntryCnt(newSourceList);
			RecordType[1]		= IGMP_MODE_NONE;
			SourceList[1]			= NULL;
			SourceListCnt[1]		= 0;
		}else if (oldFilterMode != newFilterMode)
		{	// case (5), (6)
			RecordType[0]		= ((newFilterMode == IGMP_FILTER_EXCLUDE)?IGMP_MODE_TOEX:IGMP_MODE_TOIN);
			SourceList[0]			= newSourceList;
			SourceListCnt[0]		= MCast_SourceList_EntryCnt(newSourceList);
			RecordType[1]		= IGMP_MODE_NONE;
			SourceList[1]			= NULL;
			SourceListCnt[1]		= 0;
		}else
		{	// oldFilterMode == newFilterMode, case (3), (4)
			RecordType[0]	= IGMP_MODE_ALLOW;
			RecordType[1]	= IGMP_MODE_BLOCK;
			if (oldFilterMode == IGMP_FILTER_INCLUDE)
			{
				SourceListCnt[0] = MCast_SourceList_removal(newSourceList, 0, 0, oldSourceList, 0, 0, &tempList[0]);
				SourceListCnt[1] = MCast_SourceList_removal(oldSourceList, 0, 0, newSourceList, 0, 0, &tempList[1]);
			}else
			{
				SourceListCnt[0] = MCast_SourceList_removal(oldSourceList, 0, 0, newSourceList, 0, 0, &tempList[0]);
				SourceListCnt[1] = MCast_SourceList_removal(newSourceList, 0, 0, oldSourceList, 0, 0, &tempList[1]);
			}
			SourceList[0] = tempList[0];
			SourceList[1] = tempList[1];
		}

		/* set initial value */
		CurrentSourceListp		= &SourceList[0];
		CurrentSourceListCntp	= &SourceListCnt[0];
		CurrentRecordTypep	= &RecordType[0];
	}

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Generate IGMPv%d [Report] <%d>\n", version, _sysUpSeconds);
	#endif

	if (version < IGMPV3)
		route_info->m_addr	= m_entry->mcast_addr;
	else
		route_info->m_addr	= IGMP_V3_REPORT_ADDR;

	// generate route info
	if (m_entry->up_vlan)
	{
		route_info->netif		= NULL;
		route_info->ifInfo		= 0;
		route_info->vlan		= m_entry->up_vlan;
		route_info->port		= m_entry->up_portMask;
		route_info->isPortMask	= TRUE;
	}else if (rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE)	// send to default
	{
		rtl8651_tblDrvMCastMember_t *member;

		route_info->netif	= rtl8651_GlobalVariable.MCastUpStream.netif;
		route_info->vlan	= rtl8651_GlobalVariable.MCastUpStream.vlan;
		route_info->port	= (rtl8651_GlobalVariable.MCastUpStream.portMask & phyPortLinkStatus);
		if (!route_info->port)
			goto release_out;
		{
			uint16 tmp = ~(route_info->port);
			route_info->port = route_info->port & (tmp + 1);	/* get the lowest bit */
		}
		route_info->ifInfo	= (uint32)(rtl8651_GlobalVariable.MCastUpStream.M_intfIp);	// IP / session /pppoe session

		// trace m_entry's member to get member vlan/port, we cannot send report to vlan/port to where we have member
		member = m_entry->member;
		while (member)
		{
			if (member->vlan == route_info->vlan)
			{
				route_info->port &= ~(member->portMask);
				if (!(route_info->port))
					goto release_out;
			}
			member = member->next;
		}
		route_info->isPortMask	= TRUE;
	}else
		goto release_out;

	if (MCast_SetRouteInfoAndGetIntfIp(&sip, route_info) != SUCCESS)
		goto errout;

sendNext:
	/*
		Initiate empty packet
	*/
	pkt = MCast_procPkt(pkt);
	if (pkt)
		mbuf = pkt->ph_mbuf;
	else
		goto errout;

	/*
		Construct Packet : traditional vlan packet

		total length :
		V1:
			14 (ethernet) + 20 (ip) + 8 (igmp report) + 18 (trailer) = 60
		V2:
			14 (ethernet) + 24 (ip + Router Alert Option) + 8 (igmp report) + 14 (trailer) = 60
		V3:
			14 (ethernet) + 24 (ip + Router Alter Option) + 8 (igmp report) + (total group record length), min = 60
	*/
	pktLen = 0;
	// L2
	// smac/dmac is set by MCast_l3ActRouting()
	// Ether type
	*(uint16 *)(mbuf->m_data + 12) = htons(0x0800);
	pktLen += 14;
	pip = (struct ip *)(mbuf->m_data + pktLen);

	// L3
	if (version == IGMPV1)
		pip->ip_vhl	= 0x45;
	else
		pip->ip_vhl	= 0x46;
	pip->ip_tos	= 0;
	pip->ip_id	= htons(SYS_ID);
	pip->ip_off	= htons(0);
	pip->ip_ttl	= 1;
	pip->ip_p	= IPPROTO_IGMP;
	pip->ip_sum	= 0;

	PKTSET_UINT32_UNALIGNED(sip, &pip->ip_src);
	PKTSET_UINT32_UNALIGNED(route_info->m_addr, &pip->ip_dst);

	if (version == IGMPV1)
	{
		pktLen += 20;		// normal IP hdr length
	}else
	{
		// option
		*(uint16*)((uint8*)pip+20) = htons(0x9404);
		*(uint16*)((uint8*)pip+22) = htons(0x0000);
		pktLen += 24;		// IP hdr length + Router Alert Option
	}

	ig = (struct igmp *)(mbuf->m_data + pktLen);
	// L4
	ig->igmp_cksum = 0;
	ig->igmp_code = 0;

	switch (version)
	{
		case IGMPV1:
			PKTSET_UINT32_UNALIGNED(route_info->m_addr, &ig->igmp_group);
			ig->igmp_type = IGMP_V1_REPORT;
			pktLen += 8;
			break;
		case IGMPV2:
			PKTSET_UINT32_UNALIGNED(route_info->m_addr, &ig->igmp_group);
			ig->igmp_type = IGMP_V2_REPORT;
			pktLen += 8;
			break;
		case IGMPV3:
			{
				struct igmp_gr *	currGr, *gr;
				ipaddr_t *		addrPtr;
				uint32			mtu;
				uint32			nofg;
				uint32			nofs;
				uint32			genNextGroup;	// indicate system must generate next Group Record or not
				
				ig->igmp_type	= IGMP_V3_REPORT;
				ig->igmp_rresv	= 0;
				pktLen			+= 8;		// igmp report header, include nofg

				mtu			= route_info->netif->mtu;
				currGr		= &(ig->igmp_grlist);			// initial position of group record
				nofg			= 0;
				genNextGroup	= TRUE;						// initially, we continuously generate group record

				while ((*CurrentRecordTypep != IGMP_MODE_NONE) && (genNextGroup == TRUE))
				{
					if (((*CurrentRecordTypep == IGMP_MODE_ALLOW) || (*CurrentRecordTypep == IGMP_MODE_BLOCK)) &&
						(*CurrentSourceListCntp == 0))
					{	// ALLOW/BLOCK with Null source list : ignore it (RFC3376 $5.1)
						// move to next Group Record
						CurrentSourceListp ++;
						CurrentSourceListCntp ++;
						CurrentRecordTypep ++;
						continue;
					}
						
					gr		= currGr;
					addrPtr	= &(gr->src_list);
					nofs		= 0;
					nofg		++;
					// fill Group Record header and Source Address
					gr->igmp_gr_rt = *CurrentRecordTypep;
					// Set Group Record
					gr->igmp_gr_auxlen = 0;
					PKTSET_UINT32_UNALIGNED(m_entry->mcast_addr, &(gr->igmp_gr_group));
					pktLen += 8;	// group record header: including nofs field
					while(*CurrentSourceListCntp > 0)
					{
						assert(*CurrentSourceListp);
						if (((pktLen + sizeof(ipaddr_t)-14) > mtu) || (nofs == 0xff))
						{	// packet larger than mtu or nofs is arrive MAX -> sent this packet (remaining list will be sent in next packet)
							genNextGroup		= FALSE;		// in current packet, we don't generate next group record anymore
							goto SlistAdd_End;
						}
						PKTSET_UINT32_UNALIGNED((*CurrentSourceListp)->source, addrPtr);	// set source list
						addrPtr ++;
						nofs ++;
						(*CurrentSourceListCntp) --;
						(*CurrentSourceListp) = (*CurrentSourceListp)->next;
					}
					// move to next Group Record
					CurrentSourceListp ++;
					CurrentSourceListCntp ++;
					CurrentRecordTypep ++;
					currGr = (struct igmp_gr *)addrPtr;	// point to next Group Record
	SlistAdd_End:
					gr->igmp_gr_nofs = htons(nofs);
					pktLen += (nofs * sizeof(ipaddr_t));
				}
				ig->igmp_rnofg = htons(nofg);	// finally, we determine the number of group record
				if (nofg == 0)
					goto release_out;
			}
			break;
		default:
			assert(0);
			goto errout;
	}

	pip->ip_len	= htons(pktLen - 14);	// iplen = {total packet length} - {ether hdr length}

	// trailer
	if (pktLen < 60)
	{
		trailer = (uint8*)(mbuf->m_data + pktLen);
		memset(trailer, 0, (60 - pktLen));
		pktLen = 60;
	}

	// calculate L3/L4 checksum
	ig->igmp_cksum	= MCast_igmpChecksum(pip);
	pip->ip_sum		= _rtl8651_ipChecksum(pip);
	pkt->ph_flags		&= ~(CSUM_IP|CSUM_L4);

	// pkthdr
	pkt->ph_len = pktLen;
	pkt->ph_mbuf->m_len = pktLen;
	pkt->ph_portlist = 0;
	pkt->ph_pppeTagged = 0;
	pkt->ph_iphdrOffset = 14;
	pkt->ph_proto = PKTHDR_IGMP;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("\t=> Send IGMPv%d [Report] to VLAN[0x%X] Mask[0x%X] <%d>\n", version, route_info->vlan->vid, route_info->port, _sysUpSeconds);
	#endif

	if (MCast_l3ActRouting(pkt, route_info) != SUCCESS)
		goto errout;

	if (
		(version == IGMPV3) &&
		(*CurrentSourceListCntp > 0) &&
		(*CurrentRecordTypep != IGMP_MODE_ISEX) &&	// RFC 3376 $4.2.16 If type is ISEX/TOEX, a single Group Record is sent, containing as many source address as can fit
		(*CurrentRecordTypep != IGMP_MODE_TOEX))		// and the remaining source addresses are not reported
	{	// some packets need to be sent in next packet
		pkt = NULL;	// force system to allocate new packet
		goto sendNext;
	}
	// Report pack and send OK, exit
	goto out;

release_out:
	if (pkt)
		mBuf_freeMbufChain(pkt->ph_mbuf);
out:
	if (tempList[0])
			MCast_SourceList_freeAll(tempList[0]);
	if (tempList[1])
			MCast_SourceList_freeAll(tempList[1]);
	return SUCCESS;
errout:
	if (pkt)
		mBuf_freeMbufChain(pkt->ph_mbuf);
	if (tempList[0])
			MCast_SourceList_freeAll(tempList[0]);
	if (tempList[1])
			MCast_SourceList_freeAll(tempList[1]);
	return FAILED;
}

// Pack V2 Leave pkt
int32 MCast_LeavePack(rtl8651_tblDrvMCast_t *m_entry, struct rtl_pktHdr *pkthdr)
{
	struct rtl_pktHdr	*pkt = pkthdr;
	struct rtl_mBuf	*mbuf;
	ipaddr_t			sip;
	struct ip			*pip;
	struct igmp		*ig;
	uint8			*trailer;
	rtl8651_tblDrvMCastRouteInfo_t McastRouteInfo, *route_info = &McastRouteInfo;

	if (!m_entry)
		goto errout;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("[MCAST] Generate IGMP [Leave] <%d>\n", _sysUpSeconds);
	#endif

	// generate route info
	route_info->m_addr	= m_entry->mcast_addr;
	if (m_entry->up_vlan)
	{
		route_info->netif		= NULL;
		route_info->ifInfo		= 0;
		route_info->vlan		= m_entry->up_vlan;
		route_info->port		= m_entry->up_portMask;
		route_info->isPortMask = TRUE;
	}else
	{	// default
		if (rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE)
		{
			rtl8651_tblDrvMCastMember_t *member;

			route_info->netif		= rtl8651_GlobalVariable.MCastUpStream.netif;
			route_info->vlan		= rtl8651_GlobalVariable.MCastUpStream.vlan;
			route_info->port		= rtl8651_GlobalVariable.MCastUpStream.portMask;
			if (!route_info->port)
				goto release_out;
			{
				uint16 tmp = ~(route_info->port);
				route_info->port = route_info->port & (tmp + 1);	/* get the lowest bit */
			}
			route_info->ifInfo	= (uint32)(rtl8651_GlobalVariable.MCastUpStream.M_intfIp);	// IP / session /pppoe session

			// trace m_entry's member to get member vlan/port, we cannot send report to vlan/port to where we have member
			member = m_entry->member;
			while (member)
			{
				if (member->vlan == route_info->vlan)
				{
					route_info->port = route_info->port & ~(member->portMask);
					if (!(route_info->port))
					{
						goto release_out;
					}
				}
				member = member->next;
			}
			route_info->isPortMask	= TRUE;
		}else
		{
			goto errout;
		}
	}

	if (MCast_SetRouteInfoAndGetIntfIp(&sip, route_info) != SUCCESS)
	{
		goto errout;
	}

	/*
		Initiate empty packet
	*/
	pkt = MCast_procPkt(pkthdr);
	if (pkt)
		mbuf = pkt->ph_mbuf;
	else
		goto errout;

	/*
		Construct Packet : traditional vlan packet

		total length : 14 (ethernet) + 24 (ip + Router Alert Option) + 8 (igmp query) + 14 (trailer) = 60
	*/
	pkt->ph_len = 60;
	mbuf->m_len = 60;
	pip = (struct ip *)(mbuf->m_data + 14);
	ig = (struct igmp *)(mbuf->m_data + 38);
	trailer = (uint8 *)(mbuf->m_data + 46);

	// L2
	// smac/dmac is set by MCast_l3ActRouting()
	// Ether type
	*(uint16 *)(mbuf->m_data + 12) = htons(0x0800);

	// L3
	pip->ip_vhl	= 0x46;
	pip->ip_tos	= 0;
	pip->ip_len	= htons((24+8));	// ip_len filed only include (iphdr len + igmp hdr len)
	pip->ip_id	= htons(SYS_ID);
	pip->ip_off	= htons(0);
	pip->ip_ttl	= 1;
	pip->ip_p	= IPPROTO_IGMP;
	pip->ip_sum	= 0;

	PKTSET_UINT32_UNALIGNED(sip, &pip->ip_src);
	PKTSET_UINT32_UNALIGNED(IGMP_ALL_ROUTER, &pip->ip_dst);
	// option
	*(uint16*)((uint8*)pip+20) = htons(0x9404);
	*(uint16*)((uint8*)pip+22) = htons(0x0000);

	// L4
	ig->igmp_type = IGMP_V2_LEAVE;
	ig->igmp_code = 0;
	ig->igmp_cksum = 0;
	PKTSET_UINT32_UNALIGNED(route_info->m_addr, &ig->igmp_group);


	// trailer
	memset(trailer, 0, 14);

	// calculate L3/L4 checksum
	ig->igmp_cksum	= MCast_igmpChecksum(pip);
	pip->ip_sum		= _rtl8651_ipChecksum(pip);
	pkt->ph_flags		&= ~(CSUM_IP|CSUM_L4);

	// pkthdr
	pkt->ph_portlist = 0;
	pkt->ph_pppeTagged = 0;
	pkt->ph_iphdrOffset = 14;
	pkt->ph_proto = PKTHDR_IGMP;

	#ifdef RTL8651_MCAST_DEBUG
	rtlglue_printf("\t=> Send IGMPv%d [Leave] to VLAN[0x%X] Mask[0x%X] <%d>\n", IGMPV2, route_info->vlan->vid, route_info->port, _sysUpSeconds);
	#endif

	if (MCast_l3ActRouting(pkt, route_info) != SUCCESS)
	{
		goto errout;
	}

	return SUCCESS;
release_out:
	if (pkt)
		mBuf_freeMbufChain(pkt->ph_mbuf);
	return SUCCESS;
errout:
	if (pkt)
		mBuf_freeMbufChain(pkt->ph_mbuf);
	return FAILED;
}

int32 MCast_l3ActRouting(struct rtl_pktHdr* pkthdr, rtl8651_tblDrvMCastRouteInfo_t *info)
{
	int8 *ptr;
	rtl8651_tblDrv_networkIntfTable_t *netif;
	uint16 srcPortList;

	if ((!(info->vlan)) ||
		(!(info->port)) ||
		(!(info->m_addr)))
	{
		return FAILED;
	}

	// get interface type
	if (info->netif)
	{
		netif = info->netif;
	}else if ((netif = _rtl8651_getNetworkIntfEntryByVid(info->vlan->vid)) == NULL)
	{
		return FAILED;
	}

	if(pkthdr->ph_pppeTagged)
	{
		uint16 etherType = htons(0x0800);
		// PPPoE to Ethernet
		mBuf_adjHead(pkthdr->ph_mbuf, 8);
		// set ptr
		ptr = (int8*)(pkthdr->ph_mbuf->m_data);
		// change ethertype
		rtl8651_memcpy(ptr+12, &etherType, 2);
		// remove pppoe Tagged on packet header
		pkthdr->ph_pppeTagged = 0;
		pkthdr->ph_iphdrOffset = 14;
	}
	assert(pkthdr->ph_iphdrOffset == 14);

	// set ptr
	ptr = (int8*)(pkthdr->ph_mbuf->m_data);

	switch (netif->linkLayerType)
	{
		case RTL8651_LL_VLAN:
			// dmac
			{
				ether_addr_t	ether;

				ether.octet[0] = 0x01;
				ether.octet[1] = 0x00;
				ether.octet[2] = 0x5e;
				ether.octet[3] = (uint8)((info->m_addr >> 16) & 0xff);
				ether.octet[4] = (uint8)((info->m_addr >> 8) & 0xff);
				ether.octet[5] = (uint8)(info->m_addr & 0xff);
				ether.octet[3] &= 0x7f;
				rtl8651_memcpy(ptr, &ether, 6);
			}

			break;
		case RTL8651_LL_PPPOE:
			// get pppoe Idx
			{
				rtl8651_tblDrv_pppTable_t *poe_tb;

				if (info->ifInfo)
				{
					poe_tb = ((rtl8651_tblDrvSession_t*)(info->ifInfo))->session_t;
				}else
				{
					poe_tb = CTAILQ_FIRST(&netif->sessionHead)->session_t;
				}

				// set pppoe Index
				pkthdr->ph_pppoeIdx = PPPOE_TABLE_INDEX(poe_tb); //poe_tb->asicPppoeTblIdx;
				pkthdr->ph_dsid=poe_tb->dsid;
				// pppoe autoadd
				pkthdr->ph_flags |= (PKTHDR_PPPOE_AUTOADD);
				// set dmac
				rtl8651_memcpy(ptr, &(poe_tb->macInfo->macAddr), 6);
			}
			break;
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		case RTL8651_LL_PPTP:
		case RTL8651_LL_L2TP:
			{
				rtl8651_tblDrv_pppTable_t *ptb;
				if (info->ifInfo)
				{
					ptb = ((rtl8651_tblDrvSession_t *)(info->ifInfo))->session_t;
				}else
				{
					ptb = ((rtl8651_tblDrvSession_t *)CTAILQ_FIRST(&(netif->sessionHead)))->session_t;
				}
				// set dmac
				rtl8651_memcpy(ptr, &(ptb->macInfo->macAddr), 6);

				// modify destination port/vlan
				info->vlan = &DrvTbl.vlan[netif->vid & (RTL8651_VLAN_NUMBER - 1)];
				info->port = info->vlan->memberPortMask;
				info->isPortMask = TRUE;
			}
			break;
		#endif /* CONFIG_RTL865X_PPTPL2TP, CONFIG_RTL865XB_PPTPL2TP */
		default:
			return FAILED;
	}
	// set smac
	if (info->vlan->macAddrNumber == 0)
	{
		return FAILED;
	}
	rtl8651_memcpy(ptr+6, &(info->vlan->macAddr), 6);

	// set pkthdr
	srcPortList = pkthdr->ph_portlist;
	if (info->isPortMask == TRUE)
	{
		pkthdr->ph_portlist = (info->port);
	}else
	{
		pkthdr->ph_portlist = (1<<(info->port));
	}
	if (pkthdr->ph_portlist == 0)
	{
		return FAILED;
	}
	pkthdr->ph_flags &= ~(CSUM_IP|CSUM_L4);
	MCast_setPkthdrProperty(pkthdr, PKTHDR_ASLOOPBACK);	// set packet as loopback

	{
		uint32 mtu, tol_len;
		mtu = netif->mtu;
		tol_len = mBuf_getPktlen(pkthdr->ph_mbuf);	
		if (tol_len > (mtu+14))
		{
			int32 retval;
			int8 old_IcmpRoutingMsgEnable;

			/* strore call back mode */
			old_IcmpRoutingMsgEnable = rtl8651_drvIcmpRoutingMsgEnable;
			/* always turn-off call back mode for multicast packet */
			rtl8651_drvIcmpRoutingMsgEnable = FALSE;
			/* run fragment */
			retval = _rtl8651_mtuFragment(FWDENG_FROMDRV, pkthdr, mtu, (info->vlan->vid), FALSE, srcPortList);
			/* restore call back mode */
			rtl8651_drvIcmpRoutingMsgEnable = old_IcmpRoutingMsgEnable;
			return retval;
		}
	}
	return _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdr, (info->vlan->vid),-1);
}

/*Set pkt's property*/
//user want change property of packet, this function modify packet's parameter to fit user's requirement
int32 MCast_setPkthdrProperty(struct rtl_pktHdr *ph, int32 property)
{
	uint32 swNicRxIndex = 0;

	if (!ph)
		return FAILED;

#ifdef CONFIG_RTL865XC
	rtlglue_getRingSize(&swNicRxIndex, NULL,0); //must modify
#else
	rtlglue_getRingSize(&swNicRxIndex, NULL);
#endif

	if (!swNicRxIndex)
		return FAILED;

	switch (property)
	{
		case PKTHDR_ASLOOPBACK:	/* let packet seen as loopback from physical port */
			if ((ph->ph_rxdesc>=0)&&(ph->ph_rxdesc<swNicRxIndex))
				return FAILED;
			ph->ph_rxdesc = swNicRxIndex;
			break;
		default:
			return FAILED;
	}
	return SUCCESS;
}

/*********************************************
				External Function
  *********************************************/
int32 _rtl8651_igmp_process(struct rtl_pktHdr *pkthdr, struct ip *ip, struct igmp *ig)
{
	if ((system_enable == FALSE) || (rtl8651_drvMulticastProcessEnable == FALSE))
	{
		return FWDENG_IGMPPKT;
	}

	switch (ig->igmp_type)
	{
		case IGMP_V1_REPORT:
		case IGMP_V2_REPORT:
		case IGMP_V3_REPORT:
			#ifdef RTL8651_MCAST_DEBUG
			rtlglue_printf("##  [MCAST] Receive IGMP [Report] Port [0x%02X] <%d>\n", (1<<pkthdr->ph_portlist), _sysUpSeconds);
			#endif
			return MCast_report(pkthdr, ip, ig);
		case IGMP_V2_LEAVE:
			#ifdef RTL8651_MCAST_DEBUG
			rtlglue_printf("##  [MCAST] Receive IGMP [Leave] Port [0x%02X] <%d>\n", (1<<pkthdr->ph_portlist), _sysUpSeconds);
			#endif
			return MCast_leave(pkthdr, ip, ig);
		case IGMP_QUERY:
			#ifdef RTL8651_MCAST_DEBUG
			rtlglue_printf("##  [MCAST] Receive IGMP [Query] Port [0x%02X] <%d>\n", (1<<pkthdr->ph_portlist), _sysUpSeconds);
			#endif
			return MCast_query(pkthdr, ip, ig);
	}
	return FWDENG_IGMPPKT;
}

#define RTL8651_MULTICAST_FWD_FUNCTIONS
int32 _rtl8651_multicast_process(struct rtl_pktHdr *pkthdr, struct ip *ip)
{
	ipaddr_t						sip_org, sip;
	uint16						portSrc = pkthdr->ph_portlist;
	uint16						portMask = (1<<pkthdr->ph_portlist);
	uint16						MemberPortMask;
	ipaddr_t						mcast_addr;
	ipaddr_t						extIp;
	rtl8651_tblDrvMCast_t *			m_entry;
	rtl8651_tblDrvMCastMember_t *	member;
	rtl8651_tblDrv_vlanTable_t *		vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	rtl8651_tblDrvMCastRouteInfo_t	info;
	rtl8651_tblDrv_vlanTable_t *		FwdVlan[RTL8651_VLAN_NUMBER];		// vlan to fwd
	uint16						FwdPortMask[RTL8651_VLAN_NUMBER];	// portmask to fwd
	uint32						CacheVlanMask = 0;	// vlan mask for forwarding-cache
	uint32						CachePortMask = 0;	// port mask for forwarding-cache
	uint8						idx;
	uint8						fwdPktCnt;	// total packet count to forward
#ifndef RTL8651_MCAST_ALWAYS2UPSTREAM
	uint8						totalUpCnt = 0;
#endif /* RTL8651_MCAST_ALWAYS2UPSTREAM */
	rtl8651_tblDrvMCastUpStreamInfo_t upstreamInfo, *up = &upstreamInfo;
	struct rtl_pktHdr				*fwdHdr = pkthdr;
	struct rtl_mBuf				*fwdMbuf = pkthdr->ph_mbuf;
	struct ip						*fwdIp = ip;
	uint8						l4proto;
	/* for multicast ip filter */
	uint8						keep_pkt = FALSE;
	/* for upstream cache */
	uint8						ignoreCacheUpStream;
	uint16						upPortMask = 0;

	if ((system_enable == FALSE) || (rtl8651_drvMulticastProcessEnable == FALSE))
		return FWDENG_IP_MULTICAST;

	/* init variables for forwarding cache */
	CacheVlanMask= 0;
	CachePortMask = 0;
	extIp = 0;
	ignoreCacheUpStream = FALSE;
	/* init variable for upload packet */
	up->isSet = FALSE;
	/* init l4proto */
	l4proto = ip->ip_p;

	/* add counter */
	switch(l4proto)
	{
		case IPPROTO_TCP:
			rtl8651_fwdEngineCounter.tcpReceive ++;
			break;
		case IPPROTO_UDP:
			rtl8651_fwdEngineCounter.udpReceive ++;
			break;
	}

	mcast_addr = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);
	sip_org = PKTGET_UINT32_UNALIGNED(&ip->ip_src);
	assert(IN_MULTICAST(mcast_addr));

	/****************************************************************************************

		1. We check filter data base first

	****************************************************************************************/
	{
		rtl8651_tblDrvMCastFilter_t *filter;
		filter = _rtl8651_multicast_FindFilter(mcast_addr);
		if (filter)
		{
			switch (filter->action)
			{
				case (MCAST_FILTER_ACT_MIRROR|MCAST_FILTER_ACT_NOPROC):
					// trap this packet to protocol stack without process it: return FAILED here
					return FWDENG_IP_MULTICAST;
				case (MCAST_FILTER_ACT_MIRROR):
					// duplicate this packet, process it and trap this packet to protocol stack
					{
						struct rtl_mBuf *nbuf = NULL;
						nbuf = mBuf_dupPacket(pkthdr->ph_mbuf, MBUF_DONTWAIT);
						if (nbuf)
						{
							fwdMbuf = nbuf;
							fwdHdr= nbuf->m_pkthdr;
							fwdIp = (struct ip *)((uint32)nbuf->m_data + ((uint32)ip - (uint32)(pkthdr->ph_mbuf->m_data)));
							keep_pkt = TRUE;
						}
					}
					break;
				case (MCAST_FILTER_ACT_NOPROC):
					// ignore this packet silently
					goto release_out;
			}
		}
	}

	/*
		destination ip is in reserved space: trap to protocol stack
	*/
	if (IN_MULTICAST_RESV(mcast_addr))
		goto release_out;

	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("[MCAST] *FLOW* Slow path: Flow{%u.%u.%u.%u} from (%u.%u.%u.%u)vlan[0x%X]/port[%X]\n",
		NIPQUAD(mcast_addr), NIPQUAD(sip_org), vlanp->vid, fwdHdr->ph_portlist);
	#endif

	/****************************************************************************************

		2. We check if we need to send this packet to other member

	****************************************************************************************/
	if ((m_entry = MCast_FindMCastEntry(mcast_addr, FALSE, TRUE)) == NULL)
	{	/* check if we can forward to default upstream */
		if ((rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE) && ((rtl8651_GlobalVariable.MCastUpStream.vlan != vlanp) || !(rtl8651_GlobalVariable.MCastUpStream.portMask & portMask)))
		{
				up = &(rtl8651_GlobalVariable.MCastUpStream);
		}
	}else
	{	/* related group record is found */

		/* A. we decide if we need to send to upstrem or not */
		/***************************************************************************************/
		if ((m_entry->up_vlan) && ((m_entry->up_vlan != vlanp) || (m_entry->up_portMask != portMask)))
		{	/* we will forward this packet to m_entry's upstream */
			up->isSet = TRUE;
			up->vlan = m_entry->up_vlan;
			up->portMask = m_entry->up_portMask;
			up->netif = _rtl8651_getNetworkIntfEntryByVid(m_entry->up_vlan->vid);
			up->M_intfIp = 0;
		}else if ((rtl8651_GlobalVariable.MCastUpStream.isSet == TRUE) && ((rtl8651_GlobalVariable.MCastUpStream.vlan != vlanp) || !(rtl8651_GlobalVariable.MCastUpStream.portMask & portMask)))
		{
			up = &(rtl8651_GlobalVariable.MCastUpStream);
		}

		/* B. forward this packet to all members */
		/***************************************************************************************/
	member = m_entry->member;
	memset(FwdPortMask, 0, (sizeof(uint16)*RTL8651_VLAN_NUMBER));
	memset(FwdVlan, 0, (sizeof(uint32)*RTL8651_VLAN_NUMBER));
	fwdPktCnt = 0;
	while (member)
	{
			/* check if we are the forwarder of this member, if not, we ignore it */
		if (member->isForwarder == FALSE)
			goto next_member;

			/* if this member is in expire state, user can choose if they want to forward it to this member or not */
			if ((rtl8651_multicastProcessOptions & MCAST_PROC_EXPIRE_STATE_DONT_FWD) &&
				(member->state != IGMP_MEMBER_STATE_NORMAL))
				goto next_member;

			/* for igmp v3 we must check filter mode and source list */
		if (member->MulticastRouterVersion == IGMPV3)
		{
			rtl8651_tblDrvMCastSource_t *tmp;
			if ((member->FilterMode == IGMP_FILTER_INCLUDE) && (!MCast_SourceList_inList(sip_org, member->slist)))
				goto next_member;
			if (member->FilterMode == IGMP_FILTER_EXCLUDE)
			{
				tmp = MCast_SourceList_inList(sip_org, member->slist);
				if (tmp && (tmp->timer.expires <= _sysUpSeconds))
					goto next_member;
			}
		}

			/* decide the port and vlan of this member to add, if conflict with incoming packet, ignore it */
		MemberPortMask = member->portMask;
		if (member->vlan == vlanp)
			MemberPortMask &= ~(portMask);
		if (MemberPortMask == 0)
			goto next_member;

			/* After check: add member list to send here */
		if (FwdVlan[((member->vlan->vid) & (RTL8651_VLAN_NUMBER - 1))] == NULL)
		{
				fwdPktCnt ++; /* because packet forward count is based on each vlan, so only add fwdPktCnt if a new vlan is added */
			FwdVlan[((member->vlan->vid) & (RTL8651_VLAN_NUMBER - 1))] = member->vlan;
		}
			FwdPortMask[((member->vlan->vid) & (RTL8651_VLAN_NUMBER - 1))] |= MemberPortMask;

			#ifdef RTL8651_MCAST_DATA_DEBUG
			rtlglue_printf("[MCAST] *FLOW* Slow path: Flow{%u.%u.%u.%u} can to: Vlan[%d]portMask{0x%x}\n",
				NIPQUAD(mcast_addr), member->vlan->vid, MemberPortMask);
			#endif
		
	next_member:
		member = member->next;
	}

#ifndef RTL8651_MCAST_ALWAYS2UPSTREAM
		totalUpCnt = fwdPktCnt;
#endif /* RTL8651_MCAST_ALWAYS2UPSTREAM */

		/* after checking all members, we start to forward packet */
	if (fwdPktCnt == 0)
		goto fwd_to_upstream;

		/* init variables */
	idx = 0;
	info.m_addr		= mcast_addr;
	info.isPortMask	= TRUE;
	info.netif			= NULL;
	info.ifInfo			= 0;

		/* Set pkthdr data */
	fwdHdr->ph_flags &= ~(CSUM_IP|CSUM_L4);

		/* start to forward packet */
	while (fwdPktCnt > 0)
	{
			uint8 needDup = FALSE;
			uint8 needDecrTTL = FALSE;
			uint8 needIgnoreThisFwd = FALSE;
		
		while (FwdVlan[idx] == NULL)
		{
			idx ++;
			assert((idx < RTL8651_VLAN_NUMBER));
		}

			/*
				we save this vlan/mask to forwarding cache.

				chenyl: We must write "ALL" vlans and "ALL" ports information into ASIC and can NOT consider
						TTL issue!
						Because TTL issue is independent from Where we must forward the multicast packet.
						And sender can change TTL dynamically.
			*/
			CacheVlanMask |= (1<<idx);
			CachePortMask|= FwdPortMask[idx];

			/* set vlan/port info */
		info.vlan	= FwdVlan[idx];
		info.port	= FwdPortMask[idx];
			assert(info.vlan);
			assert(info.port);	/* we checked the forward port mask in above code */


			/* if there are other packet need to send (to other vlan/to upstream), we must duplicate packet here */
			if ((fwdPktCnt > 1)
#ifdef RTL8651_MCAST_ALWAYS2UPSTREAM
				|| (up->isSet == TRUE)
#endif /* RTL8651_MCAST_ALWAYS2UPSTREAM */
			)
		{
				needDup = TRUE;
		}


			#if 0	/* don't support LAN to WAN forwarding here beause we assume down-stream always from internal vlan */
		if ((!info.vlan->internal) && (vlanp->internal))	// LAN to WAN
		{
				ipaddr_t sip;
				if ((sip = MCast_GetIpByVlan(info.vlan, 0)) == 0)
				{	// failed
					goto fail_loop;
				}
				// set sip
				PKTSET_UINT32_UNALIGNED(sip, &fwdIp->ip_src);
		}
		#endif

			/************************************************************************
				Here we must consider about TTL issue, there are 2 difference situations:

				1. All destination ports don't need (ttl-1) because (destination vlan == source vlan)
				2. All destination ports need (ttl-1) because (destination vlan != source vlan)
			************************************************************************/
			if (FwdVlan[idx] != vlanp)
			{	/* need (ttl-1) */
				if (fwdIp->ip_ttl > 1)
				{
					needDecrTTL = TRUE;
				}else
				{
					#ifdef RTL8651_MCAST_DATA_DEBUG
					rtlglue_printf("[MCAST] \t\tTTL(%d) too low ==> Ignore it!!!\n", fwdIp->ip_ttl);
					#endif
					needIgnoreThisFwd = TRUE;
				}
			}

			#ifdef RTL8651_MCAST_DATA_DEBUG
			rtlglue_printf("[MCAST] *FLOW* Send {%u.%u.%u.%u} to: Vlan[%d]/portMask{0x%x} -- ttl[%d]\n",
				NIPQUAD(mcast_addr), info.vlan->vid, info.port, fwdIp->ip_ttl);
			#endif

			if (needIgnoreThisFwd == TRUE)
			{	/* we must ignore this packet, if we must free packet after this forwarding process, we free it */
				if (needDup == FALSE)
				{
					assert(fwdPktCnt == 1);
					mBuf_freeMbufChain(fwdMbuf);
				}
			} else
			{	/* we MUST send packet */
				if (needDup == TRUE)
		{
			struct rtl_mBuf *newMbuf;

					if ((newMbuf = mBuf_dupPacket(fwdMbuf, MBUF_DONTWAIT)) != NULL)
					{
						/*************************** TTL issue *****************************/
						if (needDecrTTL == TRUE)
						{
							struct ip *newIp = (struct ip*)(((uint32)fwdIp - (uint32)(fwdMbuf->m_data)) + (uint32)(newMbuf->m_data)); 

							_RTL8651_TTL_DEC_1(newIp->ip_ttl, newIp->ip_sum);

							#ifdef RTL8651_MCAST_DATA_DEBUG
							rtlglue_printf("[MCAST] \t\t==> decrease TTL to %d ...\n", newIp->ip_ttl);
							#endif
						}
						#ifdef RTL8651_MCAST_DATA_DEBUG
						else
							rtlglue_printf("[MCAST] \t\t==> We Keep TTL ...\n");
						#endif
						/******************************************************************/				
			if (MCast_l3ActRouting(newMbuf->m_pkthdr, &info) != SUCCESS)
			{
				mBuf_freeMbufChain(newMbuf);
							#ifdef RTL8651_MCAST_DATA_DEBUG
							rtlglue_printf("[MCAST] \t\t==> FAILED!!!\n");
							#endif
						}
						#ifdef RTL8651_MCAST_DATA_DEBUG
						else
							rtlglue_printf("[MCAST] \t\t==> SUCCESS\n");
						#endif
			}
		}else
		{
					/*************************** TTL issue *****************************/
					if (needDecrTTL == TRUE)
					{
						_RTL8651_TTL_DEC_1(fwdIp->ip_ttl, fwdIp->ip_sum);
						#ifdef RTL8651_MCAST_DATA_DEBUG
						rtlglue_printf("[MCAST] \t\t==> decrease TTL to %d ...\n", fwdIp->ip_ttl);
						#endif
					}
					#ifdef RTL8651_MCAST_DATA_DEBUG
					else
						rtlglue_printf("[MCAST] \t\t==> We Keep TTL ...\n");
					#endif
					/******************************************************************/

					if (MCast_l3ActRouting(fwdHdr, &info) != SUCCESS)
					{	/* this is the last one and we don't need to forward this packet to upstream, free it */
						mBuf_freeMbufChain(fwdMbuf);
						#ifdef RTL8651_MCAST_DATA_DEBUG
						rtlglue_printf("[MCAST] \t\t==> FAILED!!!\n");
						#endif
					}
					#ifdef RTL8651_MCAST_DATA_DEBUG
					else
						rtlglue_printf("[MCAST] \t\t==> SUCCESS\n");
					#endif
				}
			}
//		next_send:
			fwdPktCnt --;
			idx ++;
		}
	}

	/****************************************************************************************

		3. We try to upload packet to upstream

	****************************************************************************************/
fwd_to_upstream:
#ifndef RTL8651_MCAST_ALWAYS2UPSTREAM
	if (totalUpCnt == 0)	/* only send to upstream if totalUpCnt == 0 */
	{
#endif	/* RTL8651_MCAST_ALWAYS2UPSTREAM */
		if (up->isSet == TRUE)
		{
			#ifdef RTL8651_MCAST_DATA_DEBUG
			rtlglue_printf("[MCAST] *FLOW* Slow path: Flow{%u.%u.%u.%u} from{%u.%u.%u.%u},vlan[%d]/port[%d] to upstream...\n",
				NIPQUAD(mcast_addr), NIPQUAD(sip_org), vlanp->vid, fwdHdr->ph_portlist);
			rtlglue_printf("\t\t==> Upstream :\n\t\t\tVid [%d]\n\t\t\tPort [0x%x]\n", up->vlan->vid, up->portMask);
			#endif

			/* FIXME: if we introduce "session", then we can NOT exclude port directly, we can only exclude source port only the source/upstream "session" are equal */
			info.port		= ((up->portMask & ~(portMask)) & phyPortLinkStatus);
			if (info.port == 0)	/* no tx port, ignore */
			{
				#ifdef RTL8651_MCAST_DATA_DEBUG
				rtlglue_printf("[MCAST] *FLOW* Source port[0x%x] is same as destination port[0x%x], ignore.\n", portMask, up->portMask);
				#endif
				ignoreCacheUpStream = TRUE;
				goto upstream_fwd_fail;
			} else
			{
				uint16 tmp = ~(info.port);
				info.port = info.port & (tmp + 1);	/* get the lowest bit */
				upPortMask = info.port;
			}
			info.m_addr		= mcast_addr;
			info.vlan			= up->vlan;
			info.netif			= up->netif;
			info.ifInfo 		= (uint32)(up->M_intfIp);		/* ip/session/pppoe session */
			info.isPortMask	= TRUE;

			/* for default upstream or normal upstream, we must check the integrity of route information */
			if (MCast_SetRouteInfoAndGetIntfIp(&sip, &info) != SUCCESS)
			{
				ignoreCacheUpStream = TRUE;
				goto upstream_fwd_fail;
			}

			if ((!(portMask&rtl8651_externalMulticastPortMask)) &&
					((info.port)&rtl8651_externalMulticastPortMask))
			{	/* internal port to external port, modify SIP */

				/* set source ip */
				extIp = sip;

				/* set sip */
				PKTSET_UINT32_UNALIGNED(sip, &fwdIp->ip_src);
				/* adjust L3 checksum */
				_RTL8651_L3_ADJUST_CHECKSUM(sip, sip_org, fwdIp->ip_sum);
				if (!(ntohs(fwdIp->ip_off) & IP_OFFMASK))
				{
					uint8 *l4hdr = (uint8 *) ((int8 *) fwdIp + ((*(uint8*)fwdIp&0xf) << 2));
					uint16 *cksum = NULL;

					switch(fwdHdr->ph_proto)
					{
						case PKTHDR_TCP:
	 						cksum = &(((struct tcphdr*)l4hdr)->th_sum);
							break;
						case PKTHDR_UDP:
							cksum = &(((struct udphdr*)l4hdr)->uh_sum);
							break;
						default:
							goto skip_l4_checksum;
					}
					/* adjust L4 checksum */
					_RTL8651_L3_ADJUST_CHECKSUM(sip, sip_org, *cksum);
				skip_l4_checksum:
					; /* remove warning: deprecated use of label at end of compound statement */
				}
			}

			/* ttl issues */
			if (up->vlan != vlanp)
			{	/* port property: source != destination, need to decrease TTL */
				if (fwdIp->ip_ttl > 1)
				{
					_RTL8651_TTL_DEC_1(fwdIp->ip_ttl, fwdIp->ip_sum);
					#ifdef RTL8651_MCAST_DATA_DEBUG
					rtlglue_printf("[MCAST] \t\t==> decrease TTL to %d ...\n", fwdIp->ip_ttl);
					#endif
				}else
				{	/* ignore, but still cache it */
					#ifdef RTL8651_MCAST_DATA_DEBUG
					rtlglue_printf("[MCAST] \t\tTTL(%d) too low ==> Ignore it!!!\n", fwdIp->ip_ttl);
					#endif
					goto upstream_fwd_fail;
				}
			}

			if (MCast_l3ActRouting(fwdHdr, &info) != SUCCESS)
			{
	upstream_fwd_fail:
				mBuf_freeMbufChain(fwdMbuf);
			}
		}else
		{
			goto release_out;
		}
	}
	/* (NOTE) fwdMbuf is free or received now */

	/****************************************************************************************

		4. We record this flow to multicast cache

	****************************************************************************************/
//recordCache:
#ifdef RTL8651_MCAST_ALWAYS2UPSTREAM
	if (((up->isSet == TRUE) && (ignoreCacheUpStream == FALSE)) || (CacheVlanMask && CachePortMask))
	{
		uint16 recUpPortMask;
		uint32 recUpVlanMask;
		if ((up->isSet == TRUE) && (ignoreCacheUpStream == FALSE))
		{	/* packet need to be forwarded to upstream */
			MCast_asicSetActualVlanPort(&recUpPortMask, &recUpVlanMask, up->netif, up->vlan, upPortMask, (uint32)(up->M_intfIp));

			if ((recUpPortMask = (recUpPortMask & ~(portMask))) != 0)
			{	/* we only simply add the upstream's vlan/port into flow cache */
				CacheVlanMask |= recUpVlanMask;
				CachePortMask |= recUpPortMask;
			}				
		}
		}
#else	/* RTL8651_MCAST_ALWAYS2UPSTREAM */
	/* for upstream cache */
	if ((totalUpCnt == 0) && (up->isSet == TRUE) && (ignoreCacheUpStream == FALSE))
	{
		uint16 recUpPortMask;
		uint32 recUpVlanMask;

		assert(!CacheVlanMask && !CachePortMask);

		MCast_asicSetActualVlanPort(&recUpPortMask, &recUpVlanMask, up->netif, up->vlan, upPortMask, (uint32)(up->M_intfIp));

		if ((recUpPortMask = (recUpPortMask & ~(portMask))) != 0)
		{
			CacheVlanMask |= recUpVlanMask;
			CachePortMask |= recUpPortMask;
		}
	}
#endif	/* RTL8651_MCAST_ALWAYS2UPSTREAM */
	if (CacheVlanMask && CachePortMask)
	{	/* Start to cache this multicast forwarding entry */
		#ifdef RTL8651_MCAST_DATA_DEBUG
		{
			int32 retval;

			rtlglue_printf("[MCAST] *FLOW* Add Flow{%u.%u.%u.%u} (%u.%u.%u.%u)(0x%X)[0x%X]=>(0x%X)[0x%X]\n",
				NIPQUAD(mcast_addr), NIPQUAD(sip_org), vlanp->vid, portSrc, CacheVlanMask, CachePortMask);
		#ifdef RTL8651_MCAST_ALWAYS2UPSTREAM
				retval = _rtl8651_addMulticast(mcast_addr, sip_org, vlanp->vid, portSrc, CacheVlanMask, CachePortMask, extIp, ((keep_pkt == TRUE)?TRUE:FALSE), 0);
		#else
				retval = _rtl8651_addMulticast(mcast_addr, sip_org, vlanp->vid, portSrc, CacheVlanMask, CachePortMask, extIp, ((keep_pkt == TRUE)?TRUE:FALSE), (totalUpCnt == 0)?RTL865X_MULTICAST_UPLOADONLY:0);
		#endif /* RTL8651_MCAST_ALWAYS2UPSTREAM */
			if (retval == SUCCESS)
				rtlglue_printf("\t=>SUCCESS!\n");
			else
				rtlglue_printf("\t=>FAILED!\n");
		}
	#else
		#ifdef RTL8651_MCAST_ALWAYS2UPSTREAM
		_rtl8651_addMulticast(mcast_addr, sip_org, vlanp->vid, portSrc, CacheVlanMask, CachePortMask, extIp, ((keep_pkt == TRUE)?TRUE:FALSE), 0);
		#else
		_rtl8651_addMulticast(mcast_addr, sip_org, vlanp->vid, portSrc, CacheVlanMask, CachePortMask, extIp, ((keep_pkt == TRUE)?TRUE:FALSE),  (totalUpCnt == 0)?RTL865X_MULTICAST_UPLOADONLY:0);
		#endif /* RTL8651_MCAST_ALWAYS2UPSTREAM */
	#endif
	}
ret:
	/* Add SUCCESS Counter */
	{
		switch(l4proto)
		{
			case IPPROTO_TCP:
				rtl8651_fwdEngineCounter.tcpSend ++;
				break;
			case IPPROTO_UDP:
				rtl8651_fwdEngineCounter.udpSend ++;
				break;
		}
	}
	if (keep_pkt == TRUE)
		return FWDENG_IP_MULTICAST;
	else
		return SUCCESS;

release_out:
	mBuf_freeMbufChain(fwdMbuf);
	goto ret;
}


int32 _rtl8651_multicast_Fastfwd(struct rtl_pktHdr *pkthdr, struct ip *ip, rtl8651_tblDrv_vlanTable_t *vlanp)
{
	rtl8651_tblDrvMCastRouteInfo_t info;
	rtl8651_tblDrv_mCast_t *m_data;
	ipaddr_t mcast_addr;
	ipaddr_t sip;
	uint32 vlanMask;
	uint16 idx;
	uint8 l4proto;
	uint8 keep_pkt = FALSE;

	if ((system_enable == FALSE) || (rtl8651_drvMulticastProcessEnable == FALSE))
		return FAILED;

	if (IN_MULTICAST_RESV(pkthdr->ph_dip))	// error dip (reserved): return FAILED
		return FAILED;

	mcast_addr = pkthdr->ph_dip;
	sip = pkthdr->ph_sip;
	l4proto = ip->ip_p;

	/* check filter database first */
	{
		rtl8651_tblDrvMCastFilter_t *filter;
		filter = _rtl8651_multicast_FindFilter(mcast_addr);
		if (filter)
		{
			switch (filter->action)
			{
				case (MCAST_FILTER_ACT_MIRROR|MCAST_FILTER_ACT_NOPROC):
					// trap this packet to protocol stack without process it: return FAILED here
					return FWDENG_IP_MULTICAST;
				case (MCAST_FILTER_ACT_MIRROR):
					// duplicate this packet, process it and trap this packet to protocol stack
					keep_pkt = TRUE;
					break;
				case (MCAST_FILTER_ACT_NOPROC):
					// ignore this packet silently
					goto errout;
			}
		}
	}
	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("[MCAST] *FLOW* Find Flow{%u.%u.%u.%u} from (%u.%u.%u.%u)(0x%X)[0x%X]\n",
		NIPQUAD(mcast_addr), NIPQUAD(sip), vlanp->vid, pkthdr->ph_portlist);
	#endif

	m_data = (rtl8651_tblDrv_mCast_t *)_rtl8651_findMCast(mcast_addr, sip, vlanp->vid, pkthdr->ph_portlist);

	if(m_data == NULL )
	{
		#ifdef RTL8651_MCAST_DATA_DEBUG
		rtlglue_printf("\t=>NOT Found......... To Slow Path\n");
		#endif
		return FAILED;
	}
	#ifdef RTL8651_MCAST_DATA_DEBUG
	rtlglue_printf("\t=>Entry FOUND: (0x%X) [0x%X]\n", m_data->vlanMask, m_data->mbr);
	#endif

	/* if we must modify source ip of this packet, we process it in slow path */
	if (m_data->flag & RTL865X_MULTICAST_EXTIP_SET)
	{
		#ifdef RTL8651_MCAST_DATA_DEBUG
		rtlglue_printf("\t=>Skip due to source IP modification...... To Slow Path\n");
		#endif
		return FAILED;
	}
	/*
		if the properties of each ports are not the same, it means some of port need to be forwarded with (ttl-1),
		and some don't. So we skip it and process this packet in slow path
	*/
	/* check if some VLANs are different form the source vlan of this packet and some VLANs are same as the it? */
	if (((pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)) & m_data->vlanMask)
		&& (m_data->vlanMask & ~(pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1))))
	{
		#ifdef RTL8651_MCAST_DATA_DEBUG
		rtlglue_printf("\t=>Skip due to complex TTL handling...... To Slow Path\n");
		#endif
		return FAILED;
	}else
	{	/* all vlans are not same as soure vlan, need ttl-1 */
		if (m_data->vlanMask & ~(pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)))
		{	/* ttl must - 1 */
			if (ip->ip_ttl > 1)
			{
				_RTL8651_TTL_DEC_1(ip->ip_ttl, ip->ip_sum);					
			}else
			{	/* ttl expire: free packet */
				goto freeout;
			}
		}
	}
	
	// set route info and forward
	info.m_addr		= mcast_addr;
	info.port			= m_data->mbr;
	info.isPortMask	= TRUE;
	info.netif			= NULL;
	info.ifInfo			= 0;

	vlanMask = m_data->vlanMask;
	idx = 0;

	if ((!vlanMask) || (!(m_data->mbr)))
		return FAILED;

	// Set pkthdr data
	pkthdr->ph_flags &= ~(CSUM_IP|CSUM_L4);

	while (vlanMask)
	{
		if (vlanMask & 1)	// last bit is '1'
		{
			info.vlan = &DrvTbl.vlan[idx];

			// forward packet
			if ((keep_pkt == FALSE) && (vlanMask == 1))
			{	// last packet to forward
				if (MCast_l3ActRouting(pkthdr, &info) != SUCCESS)
					goto errout;
			}else
			{	// there are other vlans which need to forward
				struct rtl_mBuf *newMbuf;
				newMbuf = mBuf_dupPacket(pkthdr->ph_mbuf, MBUF_DONTWAIT);

				if (MCast_l3ActRouting(newMbuf->m_pkthdr, &info) != SUCCESS)
					mBuf_freeMbufChain(newMbuf);
					// send to next vlan
			}
		}
		// next vlan
		vlanMask = vlanMask >> 1;
		idx ++;
	}

	/* Add SUCCESS Counter */
	{
		switch(l4proto)
		{
			case IPPROTO_TCP:
				rtl8651_fwdEngineCounter.tcpSend ++;
				break;
			case IPPROTO_UDP:
				rtl8651_fwdEngineCounter.udpSend ++;
				break;
		}
	}

	if (keep_pkt == TRUE)
		return FWDENG_IP_MULTICAST;
	else
		return SUCCESS;
errout:
//FIXME: Add FAILED Counter
freeout:
	mBuf_freeMbufChain(pkthdr->ph_mbuf);
	return SUCCESS;
}

#define RTL8651_MULTICAST_FILTER_FUNCTIONS

/**************************************************************************************

	Multicast filter:

		For some group ip (ex. 239.255.255.250), protocl stack(ex. uPNP sddp server) want to mirror
		packets realted to it, or to ignore it silently, so we must provide APIs to do this.

**************************************************************************************/
int32 _rtl8651_multicast_FilterInit(rtl8651_fwdEngineInitPara_t *para)
{
	uint32 idx;
	rtl8651_tblDrvMCastFilter_t *ptr;

	rtlglue_drvMutexLock();//Lock resource

	// flush Hash Table
	for (idx = 0 ; idx < RTL8651_MCastFilterHashSize ; idx ++)
		MCastFilter[idx] = NULL;

	// init Hdr
	MCastFilter_FreeHdr = NULL;

	if (!para->MCastFilterEntryCnt)
		goto out;

	// allocate Hdr
	MCastFilter_FreeHdr = (rtl8651_tblDrvMCastFilter_t *)rtlglue_malloc(para->MCastFilterEntryCnt * sizeof(rtl8651_tblDrvMCastFilter_t));

	if (!MCastFilter_FreeHdr)
		goto out;

	// init-free filter entry list
	bzero(MCastFilter_FreeHdr, (para->MCastFilterEntryCnt * sizeof(rtl8651_tblDrvMCastFilter_t)));
	
	ptr = MCastFilter_FreeHdr;
	for (idx = 0 ; idx < para->MCastFilterEntryCnt ; idx ++, ptr++)
	{
		if (idx < (para->MCastFilterEntryCnt - 1))
			ptr->next = ptr + 1;
		else
			ptr->next = NULL;
	}

out:
	rtlglue_drvMutexUnlock();//Release resource
	return SUCCESS;
}
void rtl8651_multicast_FilterReinit(void)
{
	uint32 idx;
	
	for (idx = 0 ; idx < RTL8651_MCastFilterHashSize ; idx ++)
	{
		while (MCastFilter[idx])	// remove this entry
			rtl8651_multicast_RemoveFilter(MCastFilter[idx]->gip, MCAST_FILTER_ACT_ALL);
		assert(MCastFilter[idx] == NULL);
	}
}

uint32 _rtl8651_multicast_Filter_Hash(ipaddr_t gip)
{
	return (gip % RTL8651_MCastFilterHashSize);
}

rtl8651_tblDrvMCastFilter_t *_rtl8651_multicast_FindFilter(ipaddr_t gip)
{
	rtl8651_tblDrvMCastFilter_t *ptr;

	ptr = MCastFilter[_rtl8651_multicast_Filter_Hash(gip)];

	while(ptr && (ptr->gip != gip))
		ptr = ptr->next;

	return ptr;
}

int32 rtl8651_multicast_AddFilter(ipaddr_t gip, uint32 action)
{
	rtl8651_tblDrvMCastFilter_t *ptr;
	uint32 idx;
	uint32 orgAction = 0;	// no any action

	/* check action to filter some mutual exclusive settings */
	MCASTFILTER_UNLOCK_ACTIONCHECK(action, (MCAST_FILTER_ACT_NOPROC|MCAST_FILTER_ACT_L2FWD), FAILED);

	rtlglue_drvMutexLock();//Lock resource

	ptr = _rtl8651_multicast_FindFilter(gip);
	if (ptr)
	{	// entry exist
		orgAction = ptr->action;
		ptr->action |= action;
	}else
	{	// entry not exist, allocate new entry for this new filter
		if (!action)	// no action: we don't need to add entry
			goto out;
		if (!MCastFilter_FreeHdr)	// no new entry: we CANNOT add entry
			goto errout;

		// allocate entry
		ptr = MCastFilter_FreeHdr; 
		MCastFilter_FreeHdr = MCastFilter_FreeHdr->next;

		// assign parameter
		ptr->gip		= gip;
		ptr->action	= action;

		// attach this entry to hash table
		idx = _rtl8651_multicast_Filter_Hash(gip);
		ptr->next = MCastFilter[idx];
		MCastFilter[idx] = ptr;
	}

	// Modify tblDrv multicast entry according to new action
	_rtl8651_setMGroupAttribute(gip, ((ptr->action & MCAST_FILTER_ACT_MIRROR)?TRUE:FALSE));
	if (ptr->action & (MCAST_FILTER_ACT_NOPROC|MCAST_FILTER_ACT_L2FWD))
	{
		rtl8651_tblDrvMCast_t *m_entry = MCast_FindMCastEntry(gip, FALSE, FALSE);
		if (m_entry)
		{
			MCast_RemoveMCastEntry(m_entry);
			_rtl8651_delMulticast(gip, 0, 0);
		}
	}

out:
	rtlglue_drvMutexUnlock();//Release resource
	return SUCCESS;
errout:
	rtlglue_drvMutexUnlock();//Release resource
	return FAILED;
}
int32 rtl8651_multicast_RemoveFilter(ipaddr_t gip, uint32 action)
{
	rtl8651_tblDrvMCastFilter_t *ptr, *prev;
	uint32	idx;
	uint8	TurnMirrorOff = FALSE;

	rtlglue_drvMutexLock();//Lock resource

	/* find entry */
	idx = _rtl8651_multicast_Filter_Hash(gip);
	prev = NULL;
	ptr = MCastFilter[idx];
	while (ptr && (ptr->gip != gip))
	{
		ptr	= ptr->next;
		prev	= ptr;
	}
	// bingo(process this entry) or entry not found(exit)

	/* bingo case, entry found */
	if (ptr)
	{
		uint32 orgAction = ptr->action;
	
		ptr->action &= ~action;

		if ((orgAction & MCAST_FILTER_ACT_MIRROR) && (!(ptr->action & MCAST_FILTER_ACT_MIRROR)))
			TurnMirrorOff = TRUE;
		
		if (ptr->action)
			goto out;

		// No more action needed: remove this entry
		// update linked list
		if (prev)
			prev->next = ptr->next;

		if (MCastFilter[idx] == ptr)
			MCastFilter[idx] = ptr->next;

		// put this entry back to free list
		ptr->next = MCastFilter_FreeHdr;
		MCastFilter_FreeHdr = ptr;
	}

out:

	// Modify tblDrv multicast entry according to new action
	if (TurnMirrorOff == TRUE)
		_rtl8651_setMGroupAttribute(gip, FALSE);

	rtlglue_drvMutexUnlock();//Release resource
	return SUCCESS;
}


