#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "mbuf.h"
#include "PktProc/pktDump.h"
#include "mbufGen.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvFwd.h"

//#define _MSG

int buffcmp(const void *b1, const void *b2, size_t length)
{
	const char *p1 = b1, *p2 = b2;
	size_t len = length;

	if (len == 0)
		return(0);
        do {
		if (*p1++ != *p2++)
			break;
	} while (--len);
	return(len);
}


/*
	copy mbuf data to buffer
*/
int8 *_mBuf2pkt(struct rtl_mBuf *mbuf, uint32* pktLen){
	void *buf;
	uint32 totalLen;

	if (!mbuf)
	{
		printf("\t* ERROR!! NULL mbuf!\n");
		return NULL;
	}

	totalLen = mBuf_getPktlen(mbuf); // get pkt length

	// when *pktLen is not assigned or is illegal, set to the default packet length
	if ((*pktLen == 0) || (*pktLen > totalLen))
		*pktLen = totalLen;

	buf = malloc(*pktLen);		// allocate buffer for mbuf data

	mBuf_copyToUserBuffer(mbuf, 0, *pktLen, buf);	// cp mbuf to buffer

	return (int8 *)buf;
}
/*
	get mbuf and copy data from buffer to mbuf
*/
struct rtl_mBuf *_pkt2mBuf(int8 *buf, uint32 pktLen){
	struct rtl_mBuf *mbuf, *thisMbuf;
	struct rtl_pktHdr *pktHdr;
	void *ptr;
	uint32 remLen;

	if (!buf)
	{
		printf("\t* ERROR!! NULL packet buffer!\n");
		return NULL;
	}

	#if 0
	mbuf = mBuf_getm(0, pktLen, MBUF_ALLOCPKTHDR, 0);	// allocate mbuf and pkthdr
	if (mbuf == 0){
		printf("\t* buf2ptk: ERROR! Allocate mbuf failed!\n");
		return NULL;
	}
	#endif

	if((mbuf = mBuf_get(MBUF_DONTWAIT,MBUFTYPE_DATA, 1)) == NULL){
		printf("\t* ERROR!! mbuf allocate error!\n");
		return NULL;
	}
	if(mBuf_getPkthdr(mbuf, MBUF_DONTWAIT) == (struct rtl_mBuf *) NULL) {
		printf("\t* ERROR!! pkthdr allocate error!\n");
		freeMbuf(mbuf);
		return NULL;
	}
	if(mbuf->m_extsize < pktLen)
	{
		printf("\t* ERROR!! pkt is larger than cluster size! Cannot create this packet\n");
		freeMbuf(mbuf);
		return NULL;
	}

	pktHdr = mbuf->m_pkthdr;

	thisMbuf = mbuf;
	ptr = buf;
	remLen = pktLen;

	while (thisMbuf){
		if (thisMbuf->m_extsize < remLen){
			thisMbuf->m_data = thisMbuf->m_extbuf + 128;
			memcpy(thisMbuf->m_data, ptr, thisMbuf->m_extsize);

			thisMbuf->m_len = thisMbuf->m_extsize;
			remLen -= thisMbuf->m_extsize;
			ptr += thisMbuf->m_extsize;
		}else{
			thisMbuf->m_data = thisMbuf->m_extbuf+128;
			memcpy(thisMbuf->m_data, ptr, remLen);

			thisMbuf->m_len = remLen;
			break;
		}
		thisMbuf = thisMbuf->m_next;
	}
	pktHdr = mbuf->m_pkthdr;
	pktHdr->ph_len = pktLen;
	return mbuf;
}
/*
	set ptkHdr information from config structure
*/
int32 _setPktHdr(struct rtl_mBuf* mbuf, rtl8651_PktConf_t* conf){
	struct rtl_pktHdr *pktHdr;

	if (!mbuf || !mbuf->m_pkthdr){
		printf("\t* FAILED!! no mbuf or pkthdr, exit...\n");
		return FAILED;
	}

	pktHdr = mbuf->m_pkthdr;

	if (conf->l2Flag & L2_VLAN){
		pktHdr->ph_vlanTagged = 1;
	}else{
		pktHdr->ph_vlanTagged = 0;
	}
	if (conf->l2Flag & L2_SNAP){
		pktHdr->ph_LLCTagged = 1;
	}else{
		pktHdr->ph_LLCTagged = 0;
	}
	if (conf->l2Flag & L2_PPPoE){
		pktHdr->ph_pppeTagged = 1;
	}
	pktHdr->ph_pppoeIdx = conf->conf_pppoeIdx; 
	/*
		NOTE:	ETHER/IPX/ARP/PPTP/IGMP don't set L3/L4 CSUM bits
		chenyl : set by conf_Flags now
	*/
	#if 0
	switch (conf->pktType){
		case _PKT_TYPE_TCP:
		case _PKT_TYPE_UDP:
		case _PKT_TYPE_ICMP:
			if (conf->ErrFlag & L4_CKSUM_ERR){	// tcp/udp/icmp checksum
				pktHdr->ph_flags &=~CSUM_L4_OK;
			}else{
				pktHdr->ph_flags |= CSUM_L4_OK;
			}
		case _PKT_TYPE_IP:
		case _PKT_TYPE_IGMP:
			if (conf->ErrFlag & L3_CKSUM_ERR){	// L3 ip checksum
				pktHdr->ph_flags &= ~CSUM_IP_OK;
			}else{
				pktHdr->ph_flags |= CSUM_IP_OK;
			}
	}
	#endif
	if (conf->relayType & L2_BCAST){
		pktHdr->ph_flags |= PKT_BCAST;
	}else{
		pktHdr->ph_flags &= ~PKT_BCAST;
	}
	if (conf->relayType & L2_MCAST){
		pktHdr->ph_flags |= PKT_MCAST;
	}else{
		pktHdr->ph_flags &= ~PKT_MCAST;
	}

	/* always clear the PKTHDR_HPRIORITY bit */
#ifdef CONFIG_RTL865XC 
	pktHdr->ph_flags &= ~PKTHDR_QUEUE_ID;
#else
	pktHdr->ph_flags &= ~PKTHDR_HPRIORITY;
#endif

	pktHdr->ph_flags |= conf->conf_Flags;	// other flags

	switch (conf->pktType){
        case _PKT_TYPE_ETHER: 
        case _PKT_TYPE_IPX:
        case _PKT_TYPE_ARP:
		pktHdr->ph_proto = PKTHDR_ETHERNET;
		break;
        case _PKT_TYPE_IP:
		pktHdr->ph_proto = PKTHDR_IP;
		break;
        case _PKT_TYPE_PPTP:
		pktHdr->ph_proto = PKTHDR_PPTP;
		break;
        case _PKT_TYPE_ICMP:
		pktHdr->ph_proto = PKTHDR_ICMP;
		break;
        case _PKT_TYPE_IGMP:
		pktHdr->ph_proto = PKTHDR_IGMP;
		break;
        case _PKT_TYPE_TCP:
		pktHdr->ph_proto = PKTHDR_TCP;
		break;
        case _PKT_TYPE_UDP:
		pktHdr->ph_proto = PKTHDR_UDP;
		break;
	case _PKT_TYPE_UDP_OPTION:
		pktHdr->ph_proto = PKTHDR_UDP;
		break;
	default:
		printf("\t* FAILED!! unknown type, exit...\n");
		return FAILED;
	}
#ifdef _MSG
	if (!conf->conf_vlanIdx)		printf("\t* vlanIdx == 0\n");
#endif
#ifdef CONFIG_RTL865XC 
#else
	pktHdr->ph_vlanIdx 		= conf->conf_vlanIdx;
#endif
	#ifdef RTL8651B
#ifdef _MSG
	if (!conf->conf_extPortList)	printf("\t* extPortList == 0\n");
#endif
	pktHdr->ph_extPortList		= conf->conf_extPortList;
#ifdef _MSG
	if (!conf->conf_srcExtPortNum)	printf("\t* srcExtPortNum == 0\n");
#endif
	pktHdr->ph_srcExtPortNum	= conf->conf_srcExtPortNum;
	#endif
#ifdef _MSG
	if (!conf->conf_portlist)		printf("\t* portlist == 0\n");
#endif
	pktHdr->ph_portlist 		= conf->conf_portlist;
#ifdef _MSG
	if (!conf->conf_reason)		printf("\t* reason == 0\n");
#endif
	pktHdr->ph_reason 		= conf->conf_reason;

	pktHdr->ph_category		= conf->conf_category;
	
#ifdef _MSG
	printf("\t* SUCCESS!!\n");
#endif
	return SUCCESS;
}
/*
	Dump pktHdr information
*/
int32 _dumpPktHdr(struct rtl_mBuf* mbuf){
	struct rtl_pktHdr *pktHdr;

	printf("\t* dump pkt info...\n");
	if (!mbuf || !mbuf->m_pkthdr){
		printf("\t* FAILED!! no mbuf or pkthdr, exit...\n");
		return FAILED;
	}
	pktHdr = mbuf->m_pkthdr;
	// pkt length
	if (pktHdr->ph_len != mbuf->m_len)
		printf("\t** ERROR: length check un-match! hdr [%d] != mbuf [%d]\n", pktHdr->ph_len, mbuf->m_len);
	printf("\t* Length\t\t[%d]\n", pktHdr->ph_len);
	// flags
	printf("\t* FLAG\n");
	if (pktHdr->ph_flags & PKTHDR_FREE)		{printf("\t\tPKTHDR_FREE\n");}
	if (pktHdr->ph_flags & PKTHDR_USED)		{printf("\t\tPKTHDR_USED\n");}
	if (pktHdr->ph_flags & PKTHDR_DRIVERHOLD)	{printf("\t\tPKTHDR_DRIVERHOLD\n");}
	if (pktHdr->ph_flags & PKTHDR_CPU_OWNED)	{printf("\t\tPKTHDR_CPU_OWNED\n");}
	//if (pktHdr->ph_flags & PKT_INCOMING)		{printf("\t\tPKT_INCOMING\n");}
	//if (pktHdr->ph_flags & PKT_OUTGOING)		{printf("\t\tPKT_OUTGOING\n");}
	if (pktHdr->ph_flags & PKT_BCAST)		{printf("\t\tPKT_BCAST\n");}
	if (pktHdr->ph_flags & PKT_MCAST)		{printf("\t\tPKT_MCAST\n");}
	#ifdef RTL8651B
	if (pktHdr->ph_flags & PKTHDR_BRIDGING)		{printf("\t\tPKTHDR_BRIDGING\n");}
	if (pktHdr->ph_flags & PKTHDR_HWLOOKUP)		{printf("\t\tPKTHDR_HWLOOKUP\n");}
	#endif
	if (pktHdr->ph_flags & PKTHDR_PPPOE_AUTOADD)	{printf("\t\tPKTHDR_PPPOE_AUTOADD\n");}
	if (pktHdr->ph_flags & PKTHDR_VLAN_AUTOADD)	{printf("\t\tPKTHDR_VLAN_AUTOADD\n");}
	if (pktHdr->ph_flags & CSUM_IP_OK)		{printf("\t\tCSUM_IP_OK\n");}
	if (pktHdr->ph_flags & CSUM_L4_OK)		{printf("\t\tCSUM_L4_OK\n");}
	if (pktHdr->ph_flags & CSUM_IP)			{printf("\t\tCSUM_IP\n");}
	if (pktHdr->ph_flags & CSUM_L4)			{printf("\t\tCSUM_L4\n");}

	// proto
	printf("\t* Proto\t\t");
	switch (pktHdr->ph_proto){
		case PKTHDR_ETHERNET:
			printf("PKTHDR_ETHERNET\n");
			break;
		case PKTHDR_PPTP:
			printf("PKTHDR_PPTP\n");
			break;
		case PKTHDR_IP:
			printf("PKTHDR_IP\n");
			break;
		case PKTHDR_ICMP:
			printf("PKTHDR_ICMP\n");
			break;
		case PKTHDR_IGMP:  
			printf("PKTHDR_IGMP\n");
			break;
		case PKTHDR_TCP:
			printf("PKTHDR_TCP\n");
			break;
		case PKTHDR_UDP:
			printf("PKTHDR_UDP\n");
			break;
		default:
			printf("\n\t* FAILED!! unknown type, exit...n");
			return FAILED;
	}
	// tags
	printf("\t* TAG\n");
	if (pktHdr->ph_vlanTagged)	printf("\t\tVLAN\n");
	if (pktHdr->ph_LLCTagged)	printf("\t\tLLC\n");
	if (pktHdr->ph_pppeTagged)	printf("\t\tpppoe\n");
	printf("\t* pppoe index\t\t[%d]\n", pktHdr->ph_pppoeIdx);

#ifdef CONFIG_RTL865XC 
#else
	printf("\t* vlan index\t\t[%d]\n", pktHdr->ph_vlanIdx);
#endif
	// reason field
	printf("\t* reason bits\t\t%x\n", pktHdr->ph_reason);
	#ifdef RTL8651B
	printf("\t* extPortList\t\t%x\n", pktHdr->ph_extPortList);
	printf("\t* srcExtPortNum\t%x\n", pktHdr->ph_srcExtPortNum);
	#endif
	printf("\t* portlist\t\t%x\n", pktHdr->ph_portlist);

	return SUCCESS;
}


/*
	Free mbuf
*/
int32 freeMbuf(struct rtl_mBuf* mbuf){
	return mBuf_freeMbufChain(mbuf);
}


/*
	Generate mbuf and packet header

	Input : rtl8651_PktConf_t*
	Output: rtl_mBuf*
*/
struct rtl_mBuf* genMbufPkt(rtl8651_PktConf_t* conf){
	int8 buff[4096];
	uint32 pktLen;
	struct rtl_mBuf *mbuf = NULL;

	pktLen = pktGen(conf, buff);				/* generate pkt in buff */
	if (pktLen == 0){
		printf("\t* Null packet...exit!!\n");
		return NULL;
	}
	mbuf = _pkt2mBuf(buff, pktLen);				/* convert buff into mbuf */
	if (mbuf == NULL){
		printf("\t* Get mBuf error...exit!!\n");
		return NULL;
	}
	if(_setPktHdr(mbuf,conf) != 0){/* success*/
		printf("\t* Set pktHdr error...exit!!\n");
		freeMbuf(mbuf);
		return NULL;
	}
	return mbuf;
}
/*
	Generate ICMP error message : mbuf and packet header

	Input : rtl8651_PktConf_t* err_pkt
		    rtl8651_PktConf_t* icmp_pkt

	Output: rtl_mBuf*
*/
struct rtl_mBuf* genIcmpErrMbufPkt(rtl8651_PktConf_t* err_pkt, rtl8651_PktConf_t* icmp_pkt){
	int8 buff[4096], buff_err[4096];
	uint32 pktLen, errLen, errIphLen;
	struct rtl_mBuf *mbuf = NULL;
	uint32 ipoffset;
	struct ip* ip;
	struct tcphdr* tc;
	struct udphdr* ud;
	struct icmp* ic;
	rtl8651_PktConf_t fake_icmp_pkt;

	// check icmp_pkt parameters
	if (icmp_pkt->pktType != _PKT_TYPE_ICMP)
	{
		printf("\t* ERROR!! gen ICMP packet for non-icmp packet configuration!\n");
		return NULL;
	}
	switch (icmp_pkt->icmp.type)
	{
		case ICMP_UNREACH:
		case ICMP_TIMXCEED:
		case ICMP_PARAMPROB:
		case ICMP_SOURCEQUENCH:
		case ICMP_REDIRECT:
			break;
		default:
			printf("\t* WARNING!! gen ICMP error message for non-icmp error message configuration\n");
			return genMbufPkt(icmp_pkt);
	}
	if ((icmp_pkt->payload.content != NULL)&&(icmp_pkt->payload.length != 0))
	{
		printf("\t* WARNING!! icmp error message with content! Ignore it.\n");
	}

	// gen error packet
	errLen = pktGen(err_pkt, buff_err);
	if (errLen == 0)
	{
		printf("\t* Null error message...exit!!\n");
		return NULL;
	}

	// get ip offset / error packet content
	ipoffset = 14;	// ether header
	if (err_pkt->l2Flag & L2_VLAN)		ipoffset += 4;	// VlanTagged
	if (err_pkt->l2Flag & L2_SNAP)		ipoffset += 8;	// LLCTagged
	if (err_pkt->l2Flag & L2_PPPoE)		// PppoeTagged
	{
		if (((*(int8*)(buff_err+ipoffset)) != 0x45)&&((*(int8*)(buff_err+ipoffset)) != 0x46))
		{	// chenyl: roughly check ip version/hdr len
			ipoffset += 8;
		}
	}
	ip = (struct ip*)(buff_err+ipoffset);
	errIphLen=  ((ip->ip_vhl & 0x0f) << 2);

	// ip checksum of error message is useless, clear it
	ip->ip_sum= 0;

	// l4 checksum of error message is useless, clear it
	switch (ip->ip_p){
		case IPPROTO_TCP:
			tc = (struct tcphdr *) ((int8 *) ip + ((*(int8*)ip&0xf) << 2));
			tc->th_sum = 0;
			break;
		case IPPROTO_UDP:
			ud = (struct udphdr *) ((int8 *)ip + ((*(int8*)ip&0xf)<<2));
			ud->uh_sum = 0;	
			break;
		case IPPROTO_ICMP:
			ic = (struct icmp*)((int8*)ip+((*(int8*)ip&0xf)<<2));
			ic->icmp_cksum = 0;
			break;
	}

	// calculate offset
	if ((((uint32)buff_err + errLen) - ((uint32)ip + errIphLen)) >= 8)
	{
		errLen = (uint32)(errIphLen + 8);
	}else
	{
		errLen = (uint32)(errLen - ((uint32)ip - (uint32)buff_err));
	}

	// gen fake icmp configuration
	memcpy(&fake_icmp_pkt, icmp_pkt, sizeof(rtl8651_PktConf_t));
	fake_icmp_pkt.payload.content = (int8*)ip;
	fake_icmp_pkt.payload.length = errLen;

	// gen icmp error message
	pktLen = pktGen(&fake_icmp_pkt, buff);			/* generate pkt in buff */
	if (pktLen == 0)
	{
		printf("\t* Null packet...exit!!\n");
		return NULL;
	}
	mbuf = _pkt2mBuf(buff, pktLen);				/* convert buff into mbuf */
	if (mbuf == NULL)
	{
		printf("\t* Get mBuf error...exit!!\n");
		return NULL;
	}
	if(_setPktHdr(mbuf, &fake_icmp_pkt) != 0)
	{
		printf("\t* Set pktHdr error...exit!!\n");
		freeMbuf(mbuf);
		return NULL;
	}
	return mbuf;
}

/*
	Dump mbuf informations (pktHdr and packet content)
*/
int32 dumpMbufPkt(struct rtl_mBuf* mbuf){
	struct rtl_pktHdr *pktHdr;
	int8 * buff;
	uint32 pktLen;

	// dump pktHdr
	if (_dumpPktHdr(mbuf) != 0){	// failed
		return FAILED;
	}

	pktHdr = mbuf->m_pkthdr;
	buff = _mBuf2pkt(mbuf, &pktLen);

	// dump packet content
	dumpInfo(buff, pktLen);
	// free buff
	free(buff);
	
	return SUCCESS;
}
/*
	Dump mbuf system status
*/
int32 dump_mBuf(){
	struct rtl_mBufStatus mbs;
 	if (mBuf_getBufStat(&mbs) != 0){
		printf(">>mbuf status get error !!\n");
		return FAILED;
	}
	printf(">>Mbuf Status:\n");
	printf("\t\t\tSizeof\t\tTotal\t\tFree\n");
	printf("\tCluster\t\t%d\t\t%d\t\t%d\n", mbs.m_mclbytes, mbs.m_totalclusters, mbs.m_freeclusters);
	printf("\tMbuf\t\t%d\t\t%d\t\t%d\n", mbs.m_msize, mbs.m_totalmbufs, mbs.m_freembufs);
	printf("\tpktHdr\t\t%d\t\t%d\t\t%d\n", mbs.m_pkthdrsize, mbs.m_totalpkthdrs, mbs.m_freepkthdrs);
	printf("\n");
	printf("\t* Waited times for space:\t%d\n", mbs.m_wait);	  /* times waited for space, includes mbuf, pkthdr and cluster */
	return SUCCESS;
}
/*
	get_mbuf_IpId
	For icmp packets : get sip/dip/id info form mbuf
*/
int32 get_mbuf_IpId(struct rtl_mBuf* mbuf, ipaddr_t *sip, ipaddr_t *dip, uint16 *id){
	struct rtl_pktHdr *hdr = mbuf->m_pkthdr;
	int8* buff;
	uint32 len;
	int ipoffset;
	struct ip* ip;
	struct icmp* ic;

	buff = _mBuf2pkt(mbuf, &len);

	// get ip offset
	ipoffset = 14;	// ether header

	if (hdr->ph_vlanTagged)	ipoffset += 4;
	if (hdr->ph_LLCTagged)	ipoffset += 8;
	if (hdr->ph_pppeTagged){
		if (((*(int8*)(buff+ipoffset)) != 0x45)&&((*(int8*)(buff+ipoffset)) != 0x46))
		{	// chenyl: roughly check ip version/hdr len
			ipoffset += 8;
		}
	}

	ip = (struct ip*)(buff+ipoffset);

	// get ip info
	if (sip)	*sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);
	if (dip)	*dip = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);

	// get port info
	switch (ip->ip_p){
		case IPPROTO_ICMP:
			ic = (struct icmp *) ((int8*)ip + ((*(int8*)ip&0xf)<<2));
			if (id)	*id = PKTGET_UINT16_UNALIGNED(&ic->icmp_id);
			break;
		default:
			free(buff);
			return FAILED;
	}
	free(buff);
	return SUCCESS;
}

/*
	get_mbuf_IpPort
	For tcp/udp packets : get sip/sport/dip/dport info form mbuf
*/
int32 get_mbuf_IpPort(struct rtl_mBuf* mbuf, ipaddr_t *sip, uint16 *sport, ipaddr_t *dip, uint16 *dport){
	struct rtl_pktHdr *hdr = mbuf->m_pkthdr;
	int8* buff;
	uint32 len = 0;
	int ipoffset;
	struct ip* ip;
	struct tcphdr* tc;
	struct udphdr* ud;


	buff = _mBuf2pkt(mbuf, &len);

	// get ip offset
	ipoffset = 14;	// ether header

	if (hdr->ph_vlanTagged)	ipoffset += 4;
	if (hdr->ph_LLCTagged)	ipoffset += 8;
	if (hdr->ph_pppeTagged){
		if (((*(int8*)(buff+ipoffset)) != 0x45)&&((*(int8*)(buff+ipoffset)) != 0x46))
		{	// chenyl: roughly check ip version/hdr len
			ipoffset += 8;
		}
	}

	ip = (struct ip*)(buff+ipoffset);

	// get ip info
	if (sip)	*sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);
	if (dip)	*dip = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);

	// get port info
	switch (ip->ip_p){
		case IPPROTO_TCP:
			tc = (struct tcphdr *) ((int8*)ip + ((*(int8*)ip&0xf)<<2));
			if (sport)	*sport = PKTGET_UINT16_UNALIGNED(&tc->th_sport);
			if (dport)	*dport = PKTGET_UINT16_UNALIGNED(&tc->th_dport);
			break;
		case IPPROTO_UDP:
			ud = (struct udphdr *) ((int8*)ip + ((*(int8*)ip&0xf)<<2));
			if (sport)	*sport = PKTGET_UINT16_UNALIGNED(&ud->uh_sport);
			if (dport)	*dport = PKTGET_UINT16_UNALIGNED(&ud->uh_dport);
			break;
		default:
			free(buff);
			return FAILED;
	}
	free(buff);
	return SUCCESS;
}

/*
	compare mbuf1 and mbuf2
	if mbuf1 == mbuf2, return SUCCESS
	if mbuf1 != mbuf2, return FAILED
*/
int32 mbuf_cmp(struct rtl_mBuf* m1, struct rtl_mBuf* m2, int8* _fun_, uint32 _line_){
	struct rtl_pktHdr *h1, *h2;
	int8 *b1, *b2;
	int32 diffAddr;
	uint32 l1 = 0, l2 = 0;
	// compare pktHdr
	assert(m1 && m2);
	h1 = m1->m_pkthdr;
	h2 = m2->m_pkthdr;

	// modify pkt length : for ALG
	{
		int nCount;
		int32 len = 100;
		int8** buf_arr[2]             = {&b1, &b2};
		struct rtl_pktHdr** ph_arr[2] = {&h1, &h2};

		b1 = _mBuf2pkt(m1, &len);	// get first 100 bytes of pkt
		b2 = _mBuf2pkt(m2, &len);	// get first 100 bytes of pkt

		for (nCount = 0 ; nCount < 2 ; nCount ++)
		{
			struct ip* ip;
/* Unused variables.
			struct tcphdr* tc;
			struct udphdr* ud;
			struct icmp* ic;
*/
			int ipoffset = 14;
			int8 *buf = *(int8**)(buf_arr[nCount]);
			struct rtl_pktHdr *pkthdr = *(struct rtl_pktHdr **)(ph_arr[nCount]);

			if (pkthdr->ph_vlanTagged) ipoffset += 4;
			if (pkthdr->ph_LLCTagged)  ipoffset += 6;
			if (pkthdr->ph_pppeTagged){
				if (((*(int8*)(buf+ipoffset)) != 0x45) && ((*(int8*)(buf+ipoffset)) != 0x46)){	// chenyl: roughly check ip version/hdr len
					ipoffset += 8;
				}
			}
			if (((*(int8*)(buf+ipoffset)) & 0xf0) != 0x40){
				break;
			}
			ip = ((struct ip *)(buf+ipoffset));
			if ((ipoffset+ntohs(ip->ip_len)) != pkthdr->ph_len)
			{
				#if 0
				// packet has some padding data: notify it
				printf("\t[Note] Mbuf length is modified!!\n");
				printf("\told length: %d\tnew length: %d\n", pkthdr->ph_len, (ipoffset+ntohs(ip->ip_len)));
				#endif
				if ((ipoffset+ntohs(ip->ip_len)) < pkthdr->ph_len)
					mBuf_adjTail(pkthdr->ph_mbuf, ((pkthdr->ph_len) - (ipoffset+ntohs(ip->ip_len))));
				else
					pkthdr->ph_len = (ipoffset+ntohs(ip->ip_len));
			}
		}
		free(b1);
		free(b2);
	}

	IS_EQUAL_INT("ph_len", h1->ph_len, h2->ph_len, _fun_, _line_);
	IS_EQUAL_INT("ph_flags", (h1->ph_flags&(0x001f)), (h2->ph_flags&(0x001f)),_fun_, _line_);	// PKTHDR_HPRIORITY, pppoe_autoadd, vlan_autoadd, csum_ip_ok, csum_l4_ok, csum_ip, csum_l4
	IS_EQUAL_INT("ph_proto", h1->ph_proto, h2->ph_proto, _fun_, _line_);
	IS_EQUAL_BOOL("ph_vlanTagged", h1->ph_vlanTagged, h2->ph_vlanTagged, _fun_, _line_);
	IS_EQUAL_BOOL("ph_LLCTagged", h1->ph_LLCTagged, h2->ph_LLCTagged, _fun_, _line_);
	IS_EQUAL_BOOL("ph_pppeTagged", h1->ph_pppeTagged, h2->ph_pppeTagged, _fun_, _line_);
	IS_EQUAL_INT("ph_pppoeIdx", h1->ph_pppoeIdx, h2->ph_pppoeIdx, _fun_, _line_);
#ifdef CONFIG_RTL865XC 

#else
	IS_EQUAL_INT("ph_vlanIdx", h1->ph_vlanIdx, h2->ph_vlanIdx, _fun_, _line_);
#endif
	#ifdef RTL8651B
	IS_EQUAL_INT("ph_extPortList", h1->ph_extPortList, h2->ph_extPortList, _fun_, _line_);
	IS_EQUAL_INT("ph_srcExtPortNum", h1->ph_srcExtPortNum, h2->ph_srcExtPortNum, _fun_, _line_);
	#endif
	IS_EQUAL_INT("ph_portlist", h1->ph_portlist, h2->ph_portlist, _fun_, _line_);

	IS_EQUAL_INT("ph_category", h1->ph_category, h2->ph_category, _fun_, _line_);

	// compare mbuf
	b1 = _mBuf2pkt(m1, &l1);
	b2 = _mBuf2pkt(m2, &l2);

	IS_EQUAL_INT("mbuf length", l1, l2, _fun_, _line_);
	// clear L3/L4 checksum: assume it's VLAN type
	{
		int ipoffset1 = 14;
		int etherPos1 = 12;
		int ipoffset2 = 14;
		int etherPos2 = 12;
		struct ip* ip;
		struct tcphdr* tc;
		struct udphdr* ud;
		struct icmp* ic;
		/*
			b1
		*/
		if (h1->ph_vlanTagged) ipoffset1 += 4;
		if (h1->ph_vlanTagged) etherPos1 += 4;
		if (h1->ph_LLCTagged) ipoffset1 += 6;
		if (h1->ph_LLCTagged) etherPos1 += 8;
		if (h1->ph_pppeTagged){
			if (((*(int8*)(b1+ipoffset1)) != 0x45)&&((*(int8*)(b1+ipoffset1)) != 0x46))
			{	// chenyl: roughly check ip version/hdr len
				ipoffset1 += 8;
			}
		}

		switch (ntohs(*((uint16*)(b1+etherPos1)))){
			case 0x0806:
				// erase Trailer
				{
					uint32 tailer;
					/* HW type:2, Proto type:2, HW size:1, proto size:1, opcode:2,
					   smac:6 sip:4 dmac:6 dmac:4 */
					tailer = (uint32)b1 + ipoffset1 + 28;
					memset((void*)tailer, 0, ((uint32)b1 + l1 - tailer));
				}
				goto b2;
			default:
				break;
		}

		ip = (struct ip*)(b1+ipoffset1);
		ip->ip_sum = 0;
		switch (ip->ip_p){
			case IPPROTO_TCP:
				tc = (struct tcphdr *) ((int8 *) ip + ((*(int8*)ip&0xf) << 2));
				tc->th_sum = 0;
				break;
			case IPPROTO_UDP:
				ud = (struct udphdr *) ((int8 *)ip + ((*(int8*)ip&0xf)<<2));
				ud->uh_sum = 0;
				break;
			case IPPROTO_ICMP:
				ic = (struct icmp*)((int8*)ip+((*(int8*)ip&0xf)<<2));
				ic->icmp_cksum = 0;
				switch (ic->icmp_type)
				{
					case ICMP_PARAMPROB:
					case ICMP_TIMXCEED:
					case ICMP_UNREACH:
					case ICMP_SOURCEQUENCH:
						goto b1IcmpErr;
					default:
						break;
				}
				break;
			default:
				break;
		}
		goto b2;
	b1IcmpErr:
		{
			// erase icmp error packet payload checksum
			struct ip *ip_err = &(ic->icmp_ip);
			struct tcphdr* tc_err;
			struct udphdr* ud_err;
			struct icmp* ic_err;

			ip_err->ip_sum = 0;	// L3 cksum

			switch (ip_err->ip_p)
			{
				case IPPROTO_TCP:
					tc_err = (struct tcphdr *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					tc_err->th_sum = 0;
					break;
				case IPPROTO_UDP:
					ud_err = (struct udphdr *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					ud_err->uh_sum = 0;
					break;
				case IPPROTO_ICMP:
					ic_err = (struct icmp *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					ic_err->icmp_cksum = 0;
					break;
				default:
					break;
			}
		} // b1IcmpErr
	b2:
		/*
			b2
		*/
		if (h2->ph_vlanTagged) ipoffset2 += 4;
		if (h2->ph_vlanTagged) etherPos2 += 4;
		if (h2->ph_LLCTagged) ipoffset2 += 6;
		if (h2->ph_LLCTagged) etherPos2 += 8;
		if (h2->ph_pppeTagged){
			if (((*(int8*)(b2+ipoffset2)) != 0x45)&&((*(int8*)(b2+ipoffset2)) != 0x46))
			{	// chenyl: roughly check ip version/hdr len
				ipoffset2 += 8;
			}
		}

		switch (ntohs(*((uint16*)(b2+etherPos2)))){
			case 0x0806:
				// erase Trailer
				{
					uint32 tailer;
					/* HW type:2, Proto type:2, HW size:1, proto size:1, opcode:2,
					   smac:6 sip:4 dmac:6 dmac:4 */
					tailer = (uint32)b2 + ipoffset2 + 28;
					memset((void *)tailer, 0, ((uint32)b2 + l2 - tailer));
				}
				goto cmp;
			default:
				break;
		}

		ip = (struct ip*)(b2+ipoffset2);
		ip->ip_sum = 0;
		switch (ip->ip_p){
			case IPPROTO_TCP:
				tc = (struct tcphdr *) ((int8 *) ip + ((*(int8*)ip&0xf) << 2));
				tc->th_sum = 0;
				break;
			case IPPROTO_UDP:
				ud = (struct udphdr *) ((int8 *)ip + ((*(int8*)ip&0xf)<<2));
				ud->uh_sum = 0;
				break;
			case IPPROTO_ICMP:
				ic = (struct icmp*)((int8*)ip+((*(int8*)ip&0xf)<<2));
				ic->icmp_cksum = 0;
				switch (ic->icmp_type)
				{
					case ICMP_PARAMPROB:
					case ICMP_TIMXCEED:
					case ICMP_UNREACH:
					case ICMP_SOURCEQUENCH:
						goto b2IcmpErr;
					default:
						break;
				}
				break;
			default:
				break;
		}
		goto cmp;

	b2IcmpErr:
		{
			// erase icmp error packet payload checksum
			struct ip *ip_err = &(ic->icmp_ip);
			struct tcphdr* tc_err;
			struct udphdr* ud_err;
			struct icmp* ic_err;

			ip_err->ip_sum = 0;	// L3 cksum

			switch (ip_err->ip_p)
			{
				case IPPROTO_TCP:
					tc_err = (struct tcphdr *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					tc_err->th_sum = 0;
					break;
				case IPPROTO_UDP:
					ud_err = (struct udphdr *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					ud_err->uh_sum = 0;
					break;
				case IPPROTO_ICMP:
					ic_err = (struct icmp *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					ic_err->icmp_cksum = 0;
					break;
				default:
					break;
			}
		}// b2IcmpErr
	}

	cmp:

	if ((diffAddr = buffcmp(b1, b2, l1)) != 0){
		dumpInfo(b1, l1);
		dumpInfo(b2, l2);
		printf("\t%s(%u): %s: mbuf is different: [%p],[%p]\n",
				_fun_, _line_, "mbuf data", (b1+ABSOLUTE(l1 - diffAddr)), (b2+ABSOLUTE(l1 - diffAddr)));
		free(b1);
		free(b2);
		return FAILED;
	}
	free(b1);
	free(b2);
	return SUCCESS;
}

/*
	compare ICMP Error Message mbuf1 and mbuf2
	if mbuf1 == mbuf2, return SUCCESS
	if mbuf1 != mbuf2, return FAILED
*/
int32 icmpErrPkt_cmp(struct rtl_mBuf* m1, struct rtl_mBuf* m2, int8* _fun_, uint32 _line_){

	// compre icmp packet and error packet (by mbuf compare)
	return mbuf_cmp(m1, m2, _fun_, _line_);
}
/*
	compare checksum
*/
int32 checksum_cmp(struct rtl_mBuf* m1, struct rtl_mBuf* m2, uint32 flag, int8* _fun_, uint32 _line_)
{
	struct rtl_pktHdr *h1, *h2;
	uint8 *d1, *d2;
	uint8 l41, l42;
	uint16 l3cs1, l3cs2, l4cs1, l4cs2, icmperr1, icmperr2;
	int32 retval;
	l4cs1 = l4cs2 = icmperr1 = icmperr2 = 0; /* just for initialize */

	if (!m1)
	{
		printf("%s(%d) first mbuf is NULL.\n", _fun_, _line_);
		goto errout;
	}
	
	if (!m2)
	{
		printf("%s(%d) second mbuf is NULL.\n", _fun_, _line_);
		goto errout;
	}

	h1 = m1->m_pkthdr;
	h2 = m2->m_pkthdr;
	d1 = m1->m_data;
	d2 = m2->m_data;

	{
		int ipoffset1 = 14;
		int etherPos1 = 12;
		int ipoffset2 = 14;
		int etherPos2 = 12;
		struct ip* ip;
		struct tcphdr* tc;
		struct udphdr* ud;
		struct icmp* ic;
		/*
			b1
		*/
		if (h1->ph_vlanTagged) ipoffset1 += 4;
		if (h1->ph_vlanTagged) etherPos1 += 4;
		if (h1->ph_LLCTagged) ipoffset1 += 6;
		if (h1->ph_LLCTagged) etherPos1 += 8;
		if (h1->ph_pppeTagged){
			if (((*(int8*)(d1+ipoffset1)) != 0x45)&&((*(int8*)(d1+ipoffset1)) != 0x46))
			{	// chenyl: roughly check ip version/hdr len
				ipoffset1 += 8;
			}
		}

		ip = (struct ip*)(d1+ipoffset1);
		l3cs1 = ntohs(ip->ip_sum);
		l41 = ip->ip_p;
		switch (ip->ip_p){
			case IPPROTO_TCP:
				tc = (struct tcphdr *) ((int8 *) ip + ((*(int8*)ip&0xf) << 2));
				l4cs1 = ntohs(tc->th_sum);
				break;
			case IPPROTO_UDP:
				ud = (struct udphdr *) ((int8 *)ip + ((*(int8*)ip&0xf)<<2));
				l4cs1 = ntohs(ud->uh_sum);
				break;
			case IPPROTO_ICMP:
				ic = (struct icmp*)((int8*)ip+((*(int8*)ip&0xf)<<2));
				l4cs1 = ntohs(ic->icmp_cksum);
				switch (ic->icmp_type)
				{
					case ICMP_PARAMPROB:
					case ICMP_TIMXCEED:
					case ICMP_UNREACH:
					case ICMP_SOURCEQUENCH:
						goto b1IcmpErr;
					default:
						break;
				}
				break;
			default:
				break;
		}
		goto b2;
	b1IcmpErr:
		{
			// erase icmp error packet payload checksum
			struct ip *ip_err = &(ic->icmp_ip);
/* Unused variables.
			struct tcphdr* tc_err;
			struct udphdr* ud_err;
			struct icmp* ic_err;
*/
			icmperr1 = ntohs(ip_err->ip_sum);	// L3 cksum

			/* we ingore L4 checksum in error payload */
			#if 0
			switch (ip_err->ip_p)
			{
				case IPPROTO_TCP:
					tc_err = (struct tcphdr *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					tc_err->th_sum = 0;
					break;
				case IPPROTO_UDP:
					ud_err = (struct udphdr *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					ud_err->uh_sum = 0;
					break;
				case IPPROTO_ICMP:
					ic_err = (struct icmp *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					ic_err->icmp_cksum = 0;
					break;
				default:
					break;
			}
			#endif
		} // b1IcmpErr
	b2:
		/*
			b2
		*/
		if (h2->ph_vlanTagged) ipoffset2 += 4;
		if (h2->ph_vlanTagged) etherPos2 += 4;
		if (h2->ph_LLCTagged) ipoffset2 += 6;
		if (h2->ph_LLCTagged) etherPos2 += 8;
		if (h2->ph_pppeTagged){
			if (((*(int8*)(d2+ipoffset2)) != 0x45)&&((*(int8*)(d2+ipoffset2)) != 0x46))
			{	// chenyl: roughly check ip version/hdr len
				ipoffset2 += 8;
			}
		}

		ip = (struct ip*)(d2+ipoffset2);
		l3cs2 = ntohs(ip->ip_sum);
		l42 = ip->ip_p;
		switch (ip->ip_p){
			case IPPROTO_TCP:
				tc = (struct tcphdr *) ((int8 *) ip + ((*(int8*)ip&0xf) << 2));
				l4cs2 = ntohs(tc->th_sum);
				break;
			case IPPROTO_UDP:
				ud = (struct udphdr *) ((int8 *)ip + ((*(int8*)ip&0xf)<<2));
				l4cs2 = ntohs(ud->uh_sum);
				break;
			case IPPROTO_ICMP:
				ic = (struct icmp*)((int8*)ip+((*(int8*)ip&0xf)<<2));
				l4cs2 = ntohs(ic->icmp_cksum);
				switch (ic->icmp_type)
				{
					case ICMP_PARAMPROB:
					case ICMP_TIMXCEED:
					case ICMP_UNREACH:
					case ICMP_SOURCEQUENCH:
						goto b2IcmpErr;
					default:
						break;
				}
				break;
			default:
				break;
		}
		goto cmp;

	b2IcmpErr:
		{
			// erase icmp error packet payload checksum
			struct ip *ip_err = &(ic->icmp_ip);
/* Unused variables.
			struct tcphdr* tc_err;
			struct udphdr* ud_err;
			struct icmp* ic_err;
*/
			icmperr2 = ntohs(ip_err->ip_sum);	// L3 cksum

			/* we ingore L4 checksum in error payload */
			#if 0
			switch (ip_err->ip_p)
			{
				case IPPROTO_TCP:
					tc_err = (struct tcphdr *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					tc_err->th_sum = 0;
					break;
				case IPPROTO_UDP:
					ud_err = (struct udphdr *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					ud_err->uh_sum = 0;
					break;
				case IPPROTO_ICMP:
					ic_err = (struct icmp *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					ic_err->icmp_cksum = 0;
					break;
				default:
					break;
			}
			#endif
		}// b2IcmpErr
	}
cmp:

	retval = SUCCESS;
	
	if (!flag)
	{
		printf("%s(%d) Checksum compare flags is %d: No checksum is compared.\n", _fun_, _line_, flag);
	}
	if ((flag&TEST_CSUM_IP) && (l3cs1 != l3cs2))
	{
		if ((flag & TEST_CSUM_1SCMODE) &&
			((l3cs1 == 0xffff && l3cs2 == 0)||(l3cs1 == 0 && l3cs2 == 0xffff)))
		{
			printf("[checksum_cmp] in 1's complement, [0x%X] == [0x%X]\n", ntohs(l3cs1), ntohs(l3cs2));
		}else
		{
			retval = FAILED;
		}
	}
	if (flag & TEST_CSUM_IGNOREUDP0)
	{
		if (flag & TEST_CSUM_L4)
		{
			if ((l41 == l42) && (l41 == IPPROTO_UDP) && ((l4cs1 == 0)||(l4cs2 == 0)))
			{
				printf("[checksum_cmp] least one of UDP checksum is zero, ignore it: [0x%X] [0x%X]\n", ntohs(l4cs1), ntohs(l4cs2));
			}else if (l4cs1 != l4cs2)
			{
				if ((flag & TEST_CSUM_1SCMODE) &&
					((l4cs1 == 0xffff && l4cs2 == 0)||(l4cs1 == 0 && l4cs2 == 0xffff)))
				{
					printf("[checksum_cmp] in 1's complement, [0x%X] == [0x%X]\n", ntohs(l4cs1), ntohs(l4cs2));
				}else
				{
					retval = FAILED;
				}
			}
		}
	}else if ((flag&TEST_CSUM_L4) && (l4cs1 != l4cs2))
	{
		if ((flag & TEST_CSUM_1SCMODE) &&
			((l4cs1 == 0xffff && l4cs2 == 0)||(l4cs1 == 0 && l4cs2 == 0xffff)))
		{
			printf("[checksum_cmp] in 1's complement, [0x%X] == [0x%X]\n", ntohs(l4cs1), ntohs(l4cs2));
		}else
		{
			retval = FAILED;
		}
	}
	if ((flag&TEST_CSUM_ICMPERR) && (icmperr1!= icmperr2))
	{
		if ((flag & TEST_CSUM_1SCMODE) &&
			((icmperr1 == 0xffff && icmperr2 == 0)||(icmperr1 == 0 && icmperr2== 0xffff)))
		{
			printf("[checksum_cmp] in 1's complement, [0x%X] == [0x%X]\n", ntohs(icmperr1), ntohs(icmperr2));
		}else
		{
			retval = FAILED;
		}
	}

	if (retval == FAILED)
	{
		dumpMbufPkt(m1);
		dumpMbufPkt(m2);
	}

	if ((flag&TEST_CSUM_IP) && (l3cs1 != l3cs2))
	{
		if (!((flag & TEST_CSUM_1SCMODE) &&
				((l3cs1 == 0xffff && l3cs2 == 0)||(l3cs1 == 0 && l3cs2 == 0xffff))))
			printf("%s(%d) Layer3 checksums are not equal: 0x%x 0x%x\n", _fun_, _line_, l3cs1, l3cs2);
	}
	if ((flag&TEST_CSUM_L4) && (l4cs1 != l4cs2))
	{

		if (!((flag & TEST_CSUM_1SCMODE) &&
			((l4cs1 == 0xffff && l4cs2 == 0)||(l4cs1 == 0 && l4cs2 == 0xffff))))
		{
			if ((!(flag & TEST_CSUM_IGNOREUDP0)) ||
					(!((l41 == l42) && (l41 == IPPROTO_UDP) && ((l4cs1 == 0)||(l4cs2 == 0)))))
				printf("%s(%d) Layer4 checksums are not equal: 0x%x 0x%x\n", _fun_, _line_, l4cs1, l4cs2);
		}
	}
	if ((flag&TEST_CSUM_ICMPERR) && (icmperr1!= icmperr2))
	{
		if ((flag & TEST_CSUM_1SCMODE) &&
				((icmperr1 == 0xffff && icmperr2 == 0)||(icmperr1 == 0 && icmperr2== 0xffff)))
			printf("%s(%d) Icmp Error message payload checksums are not equal: 0x%x 0x%x\n", _fun_, _line_, icmperr1, icmperr2);
	}

	if (retval != SUCCESS)
		goto errout;

/* Unused label.
out:
*/
	return SUCCESS;
errout:
	return FAILED;
}
/*
	chear checksum
*/
int32 clr_checksum(struct rtl_mBuf* m, uint32 flag)
{
	uint8 *d;
	int ipoffset = 14;
	int etherPos = 12;
	struct rtl_pktHdr *h;
	struct ip* ip;
	struct tcphdr* tc;
	struct udphdr* ud;
	struct icmp* ic;

	if (!m)
		goto errout;

	h = m->m_pkthdr;
	d = m->m_data;

	if (h->ph_vlanTagged) ipoffset+= 4;
	if (h->ph_vlanTagged) etherPos += 4;
	if (h->ph_LLCTagged) ipoffset += 6;
	if (h->ph_LLCTagged) etherPos += 8;
	if (h->ph_pppeTagged)
	{
		if (((*(int8*)(d+ipoffset)) != 0x45)&&((*(int8*)(d+ipoffset)) != 0x46))
		{	// chenyl: roughly check ip version/hdr len
			ipoffset += 8;
		}
	}

	ip = (struct ip*)(d+ipoffset);
	if (flag&TEST_CSUM_IP)
		ip->ip_sum = 0;

		switch (ip->ip_p)
		{
			case IPPROTO_TCP:
				tc = (struct tcphdr *) ((int8 *) ip + ((*(int8*)ip&0xf) << 2));
				if (flag&TEST_CSUM_L4)
					tc->th_sum = 0;
				break;
			case IPPROTO_UDP:
				ud = (struct udphdr *) ((int8 *)ip + ((*(int8*)ip&0xf)<<2));
				if (flag&TEST_CSUM_L4)
					ud->uh_sum = 0;
				break;
			case IPPROTO_ICMP:
				ic = (struct icmp*)((int8*)ip+((*(int8*)ip&0xf)<<2));
				if (flag&TEST_CSUM_L4)
					ic->icmp_cksum = 0;
				if (flag&TEST_CSUM_ICMPERR)
				{
					switch (ic->icmp_type)
					{
						case ICMP_PARAMPROB:
						case ICMP_TIMXCEED:
						case ICMP_UNREACH:
						case ICMP_SOURCEQUENCH:
							goto IcmpErr;
						default:
							break;
					}
				}
				break;
			default:
				break;
		}
		goto out;
	IcmpErr:
		{
			// erase icmp error packet payload checksum
			struct ip *ip_err = &(ic->icmp_ip);
			struct tcphdr* tc_err;
			struct udphdr* ud_err;
			struct icmp* ic_err;

			ip_err->ip_sum = 0;	// L3 cksum

			switch (ip_err->ip_p)
			{
				case IPPROTO_TCP:
					tc_err = (struct tcphdr *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					tc_err->th_sum = 0;
					break;
				case IPPROTO_UDP:
					ud_err = (struct udphdr *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					ud_err->uh_sum = 0;
					break;
				case IPPROTO_ICMP:
					ic_err = (struct icmp *) ((int8 *) ip_err + ((*(int8*)ip_err&0xf) << 2));
					ic_err->icmp_cksum = 0;
					break;
				default:
					break;
			}
		} // IcmpErr
out:
	return SUCCESS;
errout:
	return FAILED;
}

/*
	mbuf list
		stored by rtlglue_drvSend() and analysed by test functions
*/
// Initial mbufList
void mbufList_init(void){
	memset(mbufList, (int)0, sizeof(mbufList));
	mbufListCur = mbufList;
	nmbufList = 0;
}
void mbufList_Reinit(void){

	// free stored mbuf
	if (nmbufList != 0)
	{
		while (mbufListCur < &mbufList[nmbufList])
		{
			freeMbuf(mbufList_dequeue());
		}
	}
	memset(mbufList, (int)0, sizeof(mbufList));
	mbufListCur = mbufList;
	nmbufList = 0;
}
int32 mbufList_enqueue(struct rtl_mBuf* mbuf){
	if (mbufListCur != mbufList){
		printf("You don't Re-init mbufList, system re-init it...\n");
		mbufList_Reinit();
	}
	if (nmbufList < total_nmbufList){
		mbufList[nmbufList] = mbuf;
		nmbufList ++;
	}else{
		printf("Error!! MbufList Full !!\n\n");
		return FAILED;
	}
	return SUCCESS;
}
struct rtl_mBuf* mbufList_dequeue(void){
	struct rtl_mBuf* ret_mbuf = NULL;
	if (mbufListCur < &mbufList[nmbufList]){
		ret_mbuf = *mbufListCur;
		*mbufListCur = NULL;
		mbufListCur++;
	}

	/* init some information in the ret_mbuf */
	if (ret_mbuf && ret_mbuf->m_pkthdr)
	{
		ret_mbuf->m_pkthdr->ph_category = RTL8651_DEFAULT_TRAPPING_DISPATCH_CATEGORY;	/* packet TXed, we don't care its categorey */
	}

	
	return ret_mbuf;
}

int32 mbufList_nmbuf(void){
	return nmbufList;
}

