#include <linux/config.h>
#include "../rtl_types.h"
#include "../asicRegs.h"
#include "../hsRegs.h"
#include "../rtl8651_tblAsicDrv.h"
#include "../mbuf.h"
#include "veri.h"
#include "pktgen/pktgen.h"
#include "pktgen/pg_utility.h"
#include "pktgen/pg_dumpPkt.h"
#include "pktgen/pg_protoForm.h"


/* Local Definition
*/
#define VERI_IFACE  0 //write both to dut and model, read from dut
#define MODEL_IFACE 1 //write to model, read from dut


/* Static Variable 
*/
static int32 glueIface = VERI_IFACE;


/* Extern Subroutine 
*/
extern void writeMemory(uint32* addr, int32 size, uint32 value);
extern void readMemory(uint32* addr, uint32* buf, int32 size);
extern void writeTable(uint32* addr, uint32* buf);
extern void clearTable(uint32* addr, uint32 entryNum);
extern void mii_writeToMii(unsigned int port, unsigned char* buf, unsigned long len);
extern void mii_readFrMii(unsigned int port, unsigned char* buf, unsigned int *len);
extern void nicSendPacket(unsigned char* pkthdrBuf, unsigned char* buf, unsigned long len);

#ifndef __linux__
extern void procEngine_writeTable(uint32 tableType, uint32 eidx, void *entryContent_P);
extern void procEngine_readTable(uint32 tableType, uint32 eidx, void *entryContent_P);
extern void procEngine_clearTable(uint32 tableType, uint32 startEidx, uint32 endEidx);
extern void procEngine_write_mem32(uint32 reg, uint32 value);
extern uint32 procEngine_read_mem32(uint32 addr);
extern void procEngine_sendPacket(uint32 portNum, uint8 *pktBuf, uint32 pktLen);
extern void procEngine_receivePacket(uint32 portNum, uint8 *pktBuf, uint32 *pktLen);
extern void procEngine_readHsb(hsb_param_watch_t *hsbp);
extern void procEngine_readHsa(hsa_param_watch_t *hsap);
#else
#define procEngine_sendPacket(a,b,c)
#define procEngine_receivePacket(a,b,c)
#define procEngine_readHsb(a)
#define procEngine_readHsa(a)
#endif //__linux__


void glue_setModelIface(void)
{
    glueIface = MODEL_IFACE;
}

void glue_setVeriIface(void)
{
    glueIface = VERI_IFACE;
}

#ifndef __linux__
int32 _rtl8651_addAsicEntry(uint32 tableType, uint32 eidx, void *entryContent_P) {
    if (glueIface != MODEL_IFACE)
        writeTable((uint32*)((uint32) rtl8651_asicTableAccessAddrBase(tableType) + eidx * RTL8651_ASICTABLE_ENTRY_LENGTH), 
            (uint32*) entryContent_P);
    procEngine_writeTable(tableType, eidx, entryContent_P);
    return SUCCESS;
}

int32 _rtl8651_forceAddAsicEntry(uint32 tableType, uint32 eidx, void *entryContent_P) {
    if (glueIface != MODEL_IFACE)
        writeTable((uint32*)((uint32) rtl8651_asicTableAccessAddrBase(tableType) + eidx * RTL8651_ASICTABLE_ENTRY_LENGTH), 
            (uint32*) entryContent_P);
    procEngine_writeTable(tableType, eidx, entryContent_P);
    return SUCCESS;
}

int32 _rtl8651_readAsicEntry(uint32 tableType, uint32 eidx, void *entryContent_P) {
    if (glueIface == VERI_IFACE)
        readMemory((uint32*)((uint32) rtl8651_asicTableAccessAddrBase(tableType) + eidx * RTL8651_ASICTABLE_ENTRY_LENGTH), 
            (uint32*) entryContent_P, RTL8651_ASICTABLE_ENTRY_LENGTH/4);
    else if (glueIface == MODEL_IFACE)
        procEngine_readTable(tableType, eidx, entryContent_P);
    return SUCCESS;
}

int32 _rtl8651_delAsicEntry(uint32 tableType, uint32 startEidx, uint32 endEidx) {
    if (glueIface != MODEL_IFACE)
        clearTable((uint32*)((uint32) rtl8651_asicTableAccessAddrBase(tableType) + startEidx * RTL8651_ASICTABLE_ENTRY_LENGTH), 
            endEidx - startEidx + 1);
    procEngine_clearTable(tableType, startEidx, endEidx);
    return SUCCESS;
}

void WRITE_MEM32(uint32 reg, uint32 value)
{
    /* workaround: disabling and enabling aging multiple times makes auto-learn 
        of tcp/udp table fail */
    if (reg == TEATCR)
        return;
        
    if (glueIface != MODEL_IFACE)
        writeMemory((uint32*) reg, 1, value);
    procEngine_write_mem32(reg, value);
    return;
}

uint32 READ_MEM32(uint32 addr)
{
    uint32 value;
    if (glueIface == VERI_IFACE)
        readMemory((uint32*) addr,&value,1);
    else if (glueIface == MODEL_IFACE)
        value = procEngine_read_mem32(addr);
    return value;
}
#endif //__linux__

void miiLikeTestStart()
{
    if (glueIface == VERI_IFACE)
//        WRITE_MEM32(TMCR, 0xffff0000);
{
uint32 i;
for(i=0;i<5;i++)
{
    WRITE_MEM32(TCR0, READ_MEM32(PORT0_PHY_CONTROL+(i*0x20)) | (1<<14));
    WRITE_MEM32(SWTAA, PORT0_PHY_CONTROL+(i*0x20));
    WRITE_MEM32(SWTACR, ACTION_START|CMD_FORCE);
    while((READ_MEM32(SWTACR)&ACTION_MASK)!=ACTION_DONE);
}
}
    return;
}

void miiLikeTestEnd()
{
    if (glueIface == VERI_IFACE)
        WRITE_MEM32(TMCR, 0);
    return;
}

void miiLikeTestSend(uint32 portNum, uint8 *pktBuf, uint32 pktLen)
{
    if (glueIface == VERI_IFACE)
//        mii_writeToMii(portNum, pktBuf, pktLen);
{
struct rtl_pktHdr ph;
struct rtl_mBuf mb;
bzero((void*)&ph,sizeof(ph));
ph.ph_mbuf=&mb;
ph.ph_flags=PKTHDR_USED|PKT_OUTGOING;
bzero((void*)&mb,sizeof(mb));
mb.m_pkthdr=&ph;
mb.m_flags=MBUF_USED|MBUF_PKTHDR|MBUF_EOR;
ph.ph_portlist=1<<portNum;
ph.ph_len=mb.m_len=pktLen;
mb.m_data=pktBuf;
swNic_write((void*)&ph);
}
    else if (glueIface == MODEL_IFACE)
        procEngine_sendPacket(portNum, pktBuf, pktLen);
    return;
}

static struct rtl_pktHdr *rph[MAX_PORT_NUMBER];
static int8 rphInited=0;
void glue_rx(struct rtl_pktHdr *ph)
{
    if (!rphInited)
    {
        bzero((void*)rph, sizeof(struct rtl_pktHdr*) * MAX_PORT_NUMBER);
        rphInited = 1;
    }
    if (rph[ph->ph_portlist] != NULL)
        mBuf_freeMbufChain(rph[ph->ph_portlist]->ph_mbuf);
    rph[ph->ph_portlist] = ph;
}

void miiLikeTestReceive(uint32 portNum, uint8 *pktBuf, uint32 *pktLen)
{
    if (glueIface == VERI_IFACE)
//        mii_readFrMii(portNum, pktBuf, pktLen);
{
if (rph[portNum] == NULL)
    *pktLen = 0;
else
{
    *pktLen = rph[portNum]->ph_len;
    memcpy((void*) pktBuf, rph[portNum]->ph_mbuf->m_data, *pktLen);
}
}
    else if (glueIface == MODEL_IFACE)
        procEngine_receivePacket(portNum, pktBuf, pktLen);
    return;
}

#if 0 //To be removed
void nicSend(nicPktHdrFormat_t *pktHdrt, uint8 *pktBuf)
{
    struct rtl_pktHdr pktHdr;
    uint8 *pkthdr = (uint8*) &pktHdr + 4;
    pktHdr.ph_flags = PKTHDR_USED | PKT_OUTGOING | 
                        (pktHdrt->csumIp==1? CSUM_IP: 0) | 
                        (pktHdrt->csumL4==1? CSUM_L4: 0) | 
                        (pktHdrt->ph_vlanTagged==1? PKTHDR_VLAN_AUTOADD: 0) | 
                        (pktHdrt->ph_pppeTagged==1? PKTHDR_PPPOE_AUTOADD: 0) | 
                        (pktHdrt->hwLookUp==1? PKTHDR_HWLOOKUP: 0) | 
                        (pktHdrt->bridgingOnly==1? PKTHDR_BRIDGING: 0);
    pktHdr.ph_len = pktHdrt->ph_len;
    pktHdr.ph_proto = pktHdrt->ph_proto;
    pktHdr.ph_vlanTagged = pktHdr.ph_LLCTagged = pktHdr.ph_pppeTagged = 0;
    pktHdr.ph_pppoeIdx = pktHdrt->ph_pppoeIdx;
    pktHdr.ph_vlanIdx = pktHdrt->ph_vlanIdx;
    pktHdr.ph_extPortList = pktHdrt->ph_extPortList;
    pktHdr.ph_srcExtPortNum = pktHdrt->ph_srcExtPortNum;
    pktHdr.ph_portlist = pktHdrt->ph_portlist;
    nicSendPacket(pkthdr, pktBuf, pktHdrt->ph_len);
    return;
}

void nicReceive(nicPktHdrFormat_t *pktHdrt, uint8 *pktBuf, uint32 *pktLen)
{
    struct rtl_pktHdr pktHdr;
    uint8 *pkthdr = (uint8*) &pktHdr + 4;
    nicReceivePacket(pkthdr, pktBuf, pktLen);
    if (*pktLen == 0)
        return;
    bzero(pktHdrt,sizeof(pktHdrt));
    pktHdrt->csumIp = pktHdr.ph_flags&CSUM_IP? 1: 0;
    pktHdrt->csumL4 = pktHdr.ph_flags&CSUM_L4? 1: 0;
    pktHdrt->hwLookUp = pktHdr.ph_flags&PKTHDR_HWLOOKUP? 1: 0;
    pktHdrt->bridgingOnly = pktHdr.ph_flags&PKTHDR_BRIDGING? 1: 0;
    pktHdrt->ph_len = pktHdr.ph_len;
    pktHdrt->ph_proto = pktHdr.ph_proto;
    pktHdrt->ph_vlanTagged = pktHdr.ph_vlanTagged;
    pktHdrt->ph_LLCTagged = pktHdr.ph_LLCTagged;
    pktHdrt->ph_pppeTagged = pktHdr.ph_pppeTagged;
    pktHdrt->ph_pppoeIdx = pktHdr.ph_pppoeIdx;
    pktHdrt->ph_vlanIdx = pktHdr.ph_vlanIdx;
    pktHdrt->ph_reason = pktHdr.ph_reason;
    pktHdrt->ph_extPortList = pktHdr.ph_extPortList;
    pktHdrt->ph_srcExtPortNum = pktHdr.ph_srcExtPortNum;
    pktHdrt->ph_portlist = pktHdr.ph_portlist;
    return;
}
#endif

void readHsb(hsb_param_watch_t *hsbp)
{
    if (glueIface == VERI_IFACE)
    {
#define NUM_OF_WORDS (sizeof(hsb_param_t)/4+1)
        uint32 hsb[NUM_OF_WORDS];
        uint32 i;

        for (i=0;i<NUM_OF_WORDS;i++)
            hsb[i] = READ_MEM32((uint32)HSB_BASE+4*i);
        hs_readHsb(hsbp,(hsb_param_t*)hsb);
    }
    else if (glueIface == MODEL_IFACE)
        procEngine_readHsb(hsbp);
    return;
}

void readHsa(hsa_param_watch_t *hsap)
{
    if (glueIface == VERI_IFACE)
    {
#define NUM_OF_WORDS (sizeof(hsa_param_t)/4+1)
        uint32 hsa[NUM_OF_WORDS];
        uint32 i;

        for (i=0;i<NUM_OF_WORDS;i++)
            hsa[i] = READ_MEM32((uint32)HSA_BASE+4*i);
        hs_readHsa(hsap,(hsa_param_t*)hsa);
    }
    else if (glueIface == MODEL_IFACE)
        procEngine_readHsa(hsap);
    return;
}

int32 cmpMac(ether_addr_t *mac, uint8 *str)
{
    ether_addr_t tmp;
    strtomac(&tmp, str);
    return memcmp(mac, &tmp, 6);
}

int32 cmpIp(uint8 *ip, uint8 *str)
{
    ipaddr_t tmp;
    strtoip(&tmp, str);
    return ((ip[0]<<24|ip[1]<<16|ip[2]<<8|ip[3]) != tmp);
}

int32 cmpIp2(ipaddr_t ip, uint8 *str)
{
    ipaddr_t tmp;
    strtoip(&tmp, str);
    return (ip != tmp);
}

uint32 pktGenArp(uint8 *pktBuf, pktGenArpFormat_t *arpt)
{
    uint32 pktLen;
    uint32 l2Flag;

    int8	zeroMac[6];

    memset(zeroMac, 0x00, 6);
    
    l2Flag = 0;
    if (arpt->vlanTag)
        l2Flag |= L2_VLAN;

    		if (arpt->op == ARPOP_REQUEST) {
    			protoForm_vs_arp(pktBuf, l2Flag, arpt->dmac.octet, arpt->smac.octet, arpt->vid, 0, 0, arpt->smac.octet,
    				arpt->sip, zeroMac, arpt->dip, arpt->op, &pktLen);
    			} else if (arpt->op == ARPOP_REPLY) {
    			protoForm_vs_arp(pktBuf, l2Flag, arpt->dmac.octet, arpt->smac.octet, arpt->vid, 0, 0, arpt->smac.octet,
    				arpt->sip, arpt->dmac.octet, arpt->dip, arpt->op, &pktLen);
    				}
    			
		
		return pktLen;
}

uint32 pktParseArp(uint8 *pktBuf, pktGenArpFormat_t *arpt, uint32 pktLen)
{
    pktParse_t parset;
    
    pktParse((int8*)pktBuf, &parset, pktLen);
    
    if (parset.arp == NULL)
        return FAILED;
    bzero(arpt, sizeof(pktGenArpFormat_t));
    memcpy(arpt->dmac.octet, parset.ether2->da, 6);
    memcpy(arpt->smac.octet, parset.ether2->sa, 6);
    if (parset.vlan)
    {
        arpt->vlanTag = 1;
        arpt->vid = ntohs(parset.vlan->tag) & 0xfff;
    }
    arpt->sip = ntohl(parset.ip->ip_src);
    arpt->dip = ntohl(parset.ip->ip_dst);
    arpt->op = ntohs(parset.arp->ar_op);
		
		return SUCCESS;
}

uint32 pktGenIp(uint8 *pktBuf, pktGenIpFormat_t *ipt)
{
    uint32 pktLen;
    uint32 l2Flag;
    uint32 i;
    static uint8 content[2048];
    
    l2Flag = 0;
    if (ipt->vlanTag)
        l2Flag |= L2_VLAN;
    if (ipt->pppoeTag)
        l2Flag |= L2_PPPoE;
    
    for(i=0; i<ipt->padLen;i++)
        //content[i] = i; Daniel 031202: chksum error bug
        content[i] = 0;
    pktLen = ipt->padLen;
    
		protoForm_vsp_ip(pktBuf, l2Flag, ipt->dmac.octet, ipt->smac.octet, ipt->vid, 0, 0, 0x8864, ipt->sid, 4, 20, 0, 0, 0, 0, 0, 16,
		    ipt->sip, ipt->dip, ipt->proto, content, 0, &pktLen);
		
		return pktLen;
}

uint32 pktParseIp(uint8 *pktBuf, pktGenIpFormat_t *ipt, uint32 pktLen)
{
    pktParse_t parset;
    
    pktParse((int8*)pktBuf, &parset, pktLen);
    
    if (parset.ip == NULL)
        return FAILED;
    bzero(ipt, sizeof(pktGenIpFormat_t));
    memcpy(ipt->dmac.octet, parset.ether2->da, 6);
    memcpy(ipt->smac.octet, parset.ether2->sa, 6);
    if (parset.vlan)
    {
        ipt->vlanTag = 1;
        ipt->vid = ntohs(parset.vlan->tag) & 0xfff;
    }
    if (parset.pppoe)
    {
        ipt->pppoeTag = 1;
        ipt->sid = ntohs(parset.pppoe->sessionId);
    }
    ipt->sip = ntohl(parset.ip->ip_src);
    ipt->dip = ntohl(parset.ip->ip_dst);
    ipt->proto = parset.ip->ip_p;
		
		return SUCCESS;
}

uint32 pktGenIcmp(uint8 *pktBuf, pktGenIcmpFormat_t *icmpt)
{
    uint32 pktLen;
    uint32 l2Flag;

    static uint8 content[2048];
    
    l2Flag = 0;
    if (icmpt->vlanTag)
        l2Flag |= L2_VLAN;
    if (icmpt->pppoeTag)
        l2Flag |= L2_PPPoE;

    
		protoForm_vsp_icmp(pktBuf, l2Flag, icmpt->dmac.octet, icmpt->smac.octet,  icmpt->vid, 0, 0, 0x8864, icmpt->sid, 4, 20, 0, 0, 0, 0, 0, 16,
		    icmpt->sip, icmpt->dip, icmpt->type, icmpt->code, icmpt->id, icmpt->seq, content, 0, &pktLen);
		/* remove crc */
		pktLen -= 4;
		
		return pktLen;
}

uint32 pktParseIcmp(uint8 *pktBuf, pktGenIcmpFormat_t *icmpt, uint32 pktLen)
{
    pktParse_t parset;
    
    pktParse((int8*)pktBuf, &parset, pktLen);
    
    if (parset.icmp == NULL)
        return FAILED;
    bzero(icmpt, sizeof(pktGenIcmpFormat_t));
    memcpy(icmpt->dmac.octet, parset.ether2->da, 6);
    memcpy(icmpt->smac.octet, parset.ether2->sa, 6);
    if (parset.vlan)
    {
        icmpt->vlanTag = 1;
        icmpt->vid = ntohs(parset.vlan->tag) & 0xfff;
    }
    if (parset.pppoe)
    {
        icmpt->pppoeTag = 1;
        icmpt->sid = ntohs(parset.pppoe->sessionId);
    }
    icmpt->sip = ntohl(parset.ip->ip_src);
    icmpt->dip = ntohl(parset.ip->ip_dst);
    icmpt->type = parset.icmp->icmp_type;
    icmpt->code = parset.icmp->icmp_code;
    icmpt->id = htonl(parset.icmp->content.echo_s.icd_id);
    icmpt->seq = htonl(parset.icmp->content.echo_s.icd_seq);
		
		return SUCCESS;
}

uint32 pktGenIgmp(uint8 *pktBuf, pktGenIgmpFormat_t *igmpt)
{
    uint32 pktLen;
    uint32 l2Flag;
    
    l2Flag = 0;
    if (igmpt->vlanTag)
        l2Flag |= L2_VLAN;
    if (igmpt->pppoeTag)
        l2Flag |= L2_PPPoE;

    
		protoForm_vsp_igmp(pktBuf, l2Flag, igmpt->dmac.octet, igmpt->smac.octet, igmpt->vid, 0, 1, 0x8864, igmpt->sid, 4, 20, 0, 0, 0, 0, 0, 16,
		    igmpt->sip, igmpt->dip, igmpt->type, igmpt->respTime, igmpt->gaddr, 0, &pktLen);
		/* remove crc */
		pktLen -= 4;
		
		return pktLen;
}

uint32 pktParseIgmp(uint8 *pktBuf, pktGenIgmpFormat_t *igmpt, uint32 pktLen)
{
    pktParse_t parset;
    
    pktParse((int8*)pktBuf, &parset, pktLen);
    
    if (parset.igmp == NULL)
        return FAILED;
    bzero(igmpt, sizeof(pktGenIgmpFormat_t));
    memcpy(igmpt->dmac.octet, parset.ether2->da, 6);
    memcpy(igmpt->smac.octet, parset.ether2->sa, 6);
    if (parset.vlan)
    {
        igmpt->vlanTag = 1;
        igmpt->vid = ntohs(parset.vlan->tag) & 0xfff;
    }
    if (parset.pppoe)
    {
        igmpt->pppoeTag = 1;
        igmpt->sid = ntohs(parset.pppoe->sessionId);
    }
    igmpt->sip = ntohl(parset.ip->ip_src);
    igmpt->dip = ntohl(parset.ip->ip_dst);
    igmpt->type = parset.igmp->igmp_type;
    igmpt->respTime = parset.igmp->igmp_respTime;
    igmpt->gaddr = htonl(parset.igmp->igmp_group);
		
		return SUCCESS;
}

uint32 pktGenTcp(uint8 *pktBuf, pktGenTcpFormat_t *tcpt)
{
    uint32 pktLen;
    uint32 l2Flag;
    uint32 i;
    static uint8 content[2048];
    
    l2Flag = 0;
    if (tcpt->vlanTag)
        l2Flag |= L2_VLAN;
    if (tcpt->pppoeTag)
        l2Flag |= L2_PPPoE;
    
    for(i=0; i<tcpt->padLen;i++)
        //content[i] = i; Daniel 031202: chksum error bug
        content[i] = 0;
    pktLen = tcpt->padLen;
    
		protoForm_vsp_tcp(pktBuf, l2Flag, tcpt->dmac.octet, tcpt->smac.octet, tcpt->vid, 0, 0, 0x8864, tcpt->sid, 4, 20, 0, 0, 0, 0, 0, 16,
		    tcpt->sip, tcpt->dip, tcpt->sport, tcpt->dport, 0x55aa, 1, tcpt->flag, 4096, 
		    content, 0, &pktLen);
		/* remove crc */
		pktLen -= 4;
		
		return pktLen;
}

uint32 pktParseTcp(uint8 *pktBuf, pktGenTcpFormat_t *tcpt, uint32 pktLen)
{
    pktParse_t parset;
    
    pktParse((int8*)pktBuf, &parset, pktLen);
    
    if (parset.tcp == NULL)
        return FAILED;
    bzero(tcpt, sizeof(pktGenTcpFormat_t));
    memcpy(tcpt->dmac.octet, parset.ether2->da, 6);
    memcpy(tcpt->smac.octet, parset.ether2->sa, 6);
    if (parset.vlan)
    {
        tcpt->vlanTag = 1;
        tcpt->vid = ntohs(parset.vlan->tag) & 0xfff;
    }
    if (parset.pppoe)
    {
        tcpt->pppoeTag = 1;
        tcpt->sid = ntohs(parset.pppoe->sessionId);
    }
    tcpt->sip = ntohl(parset.ip->ip_src);
    tcpt->dip = ntohl(parset.ip->ip_dst);
    tcpt->sport = ntohs(parset.tcp->th_sport);
    tcpt->dport = ntohs(parset.tcp->th_dport);
    tcpt->flag = parset.tcp->th_flags;
		
		return SUCCESS;
}

uint32 pktGenUdp(uint8 *pktBuf, pktGenUdpFormat_t *udpt)
{
    uint32 pktLen;
    uint32 l2Flag;
    uint32 i;
    int8 content[1500];
    
    l2Flag = 0;
    if (udpt->vlanTag)
        l2Flag |= L2_VLAN;
    if (udpt->pppoeTag)
        l2Flag |= L2_PPPoE;
    
    for(i=0; i<udpt->padLen;i++)
        //content[i] = i; Daniel 031202: chksum error bug
        content[i] = 0;
    pktLen = udpt->padLen;
    
		protoForm_vsp_udp(pktBuf, l2Flag, udpt->dmac.octet, udpt->smac.octet, udpt->vid, 0, 0, 0x8864, udpt->sid, 4, 20, 0, 0, 0, 0, 0, 16,
		    udpt->sip, udpt->dip, udpt->sport, udpt->dport, 
		    content, 0, &pktLen);
		/* remove crc */
		pktLen -= 4;
		
		return pktLen;
}

uint32 pktParseUdp(uint8 *pktBuf, pktGenUdpFormat_t *udpt, uint32 pktLen)
{
    pktParse_t parset;
    
    pktParse((int8*)pktBuf, &parset, pktLen);
    
    if (parset.udp == NULL)
        return FAILED;
    bzero(udpt, sizeof(pktGenUdpFormat_t));
    memcpy(udpt->dmac.octet, parset.ether2->da, 6);
    memcpy(udpt->smac.octet, parset.ether2->sa, 6);
    if (parset.vlan)
    {
        udpt->vlanTag = 1;
        udpt->vid = ntohs(parset.vlan->tag) & 0xfff;
    }
    if (parset.pppoe)
    {
        udpt->pppoeTag = 1;
        udpt->sid = ntohs(parset.pppoe->sessionId);
    }
    udpt->sip = ntohl(parset.ip->ip_src);
    udpt->dip = ntohl(parset.ip->ip_dst);
    udpt->sport = ntohs(parset.udp->uh_sport);
    udpt->dport = ntohs(parset.udp->uh_dport);
		
		return SUCCESS;
}

uint32 pktGenSetVlanCfi(uint8 *pktBuf, uint32 pktLen)
{
    pktParse_t parset;
    
    /* make sure this packet is vlan tagged */
    pktParse((int8*)pktBuf, &parset, pktLen);
    if (parset.vlan == NULL)
        return FAILED;
    parset.vlan->tag |= ntohs(1<<12);
    return SUCCESS;
}

uint32 pktGenSetIpVerHdrLen(uint8 *pktBuf, uint16 ipVerHdrLen, uint32 pktLen)
{
    pktParse_t parset;
    
    /* make sure this packet has an ip header */
    pktParse((int8*)pktBuf, &parset, pktLen);
    if (parset.ip == NULL)
        return FAILED;
    parset.ip->ip_v = ipVerHdrLen>>4;
    parset.ip->ip_hl = ipVerHdrLen&0xf;
    return SUCCESS;
}

void strtomac(ether_addr_t *mac, uint8 *str)
{
    int32 t1, t2, t3, t4, t5, t6;
    sscanf(str, "%x-%x-%x-%x-%x-%x", &t1, &t2, &t3, &t4, &t5, &t6);
    mac->octet[0]=t1; mac->octet[1]=t2; mac->octet[2]=t3; mac->octet[3]=t4; mac->octet[4]=t5; mac->octet[5]=t6;
}

void strtoip(ipaddr_t *ip, uint8 *str)
{
    int32 t1, t2, t3, t4;
    sscanf(str, "%d.%d.%d.%d", &t1, &t2, &t3, &t4);
    *ip = (t1<<24)|(t2<<16)|(t3<<8)|t4;
}

uint32 getL2TableRowFromMacStr(uint8 *str)
{
    int32 t1, t2, t3, t4, t5, t6;
    ether_addr_t mac;
    sscanf(str, "%x-%x-%x-%x-%x-%x", &t1, &t2, &t3, &t4, &t5, &t6);
    mac.octet[0]=t1; mac.octet[1]=t2; mac.octet[2]=t3; mac.octet[3]=t4; mac.octet[4]=t5; mac.octet[5]=t6;
    return rtl8651_filterDbIndex(&mac);
}

uint32 pktGenGeneral(uint8 *pktBuf, pktGenGeneralPktFormat_t *pktt)
{

	uint32	pktLen;
	pktLen = pktt->len;
	
	switch(pktt->protocol)
		{
			case	U_PROTO_ETHER:
				protoForm_vsp_ether(pktBuf, pktt->l2ether.l2Flag,
					pktt->l2ether.dmac.octet, pktt->l2ether.smac.octet,
					pktt->l2vlan.vid, pktt->l2vlan.vlanCfi, pktt->l2vlan.priority,
					pktt->l2llc.dsap, pktt->l2llc.ssap,
					pktt->l2pppoe.pppoeType,
					pktt->l2pppoe.pppoeSession,
					pktt->l2ether.etherType, 
					pktt->content,
					&pktLen);
			break;

			case	U_PROTO_IPX:
				protoForm_vsp_ipx(pktBuf, pktt->l2ether.l2Flag,
					pktt->l2ether.dmac.octet, pktt->l2ether.smac.octet,
					pktt->l2vlan.vid, pktt->l2vlan.vlanCfi, pktt->l2vlan.priority,
					pktt->l2pppoe.pppoeType,
					pktt->l2pppoe.pppoeSession,
					pktt->l3ipx.ipxtype,
					pktt->l3ipx.dnet, pktt->l3ipx.dnode, pktt->l3ipx.dsock,
					pktt->l3ipx.snet, pktt->l3ipx.snode, pktt->l3ipx.ssock,
					pktt->content, &pktLen);
			break;

			case	U_PROTO_ARP:
				protoForm_vs_arp(pktBuf, pktt->l2ether.l2Flag, 
					pktt->l2ether.dmac.octet, pktt->l2ether.smac.octet,
					pktt->l2vlan.vid, pktt->l2vlan.vlanCfi, pktt->l2vlan.priority, 
					pktt->l3arp.sha.octet, pktt->l3arp.spa,
					pktt->l3arp.tha.octet, pktt->l3arp.tpa,
					pktt->l3arp.op, &pktLen);
			break;

			case	U_PROTO_IP:
				protoForm_vsp_ip(pktBuf, pktt->l2ether.l2Flag,
					pktt->l2ether.dmac.octet, pktt->l2ether.smac.octet,
					pktt->l2vlan.vid, pktt->l2vlan.vlanCfi, pktt->l2vlan.priority,
					pktt->l2pppoe.pppoeType,
					pktt->l2pppoe.pppoeSession,
					pktt->l3ip.ipv,
					pktt->l3ip.iphl,
					pktt->l3ip.tos, pktt->l3ip.ipid,
					pktt->l3ip.df, pktt->l3ip.mf,
					pktt->l3ip.ipoff,
					pktt->l3ip.ttl,
					pktt->l3ip.sip, pktt->l3ip.dip,
					pktt->l3ip.proto, pktt->content,
					pktt->l3ip.errorFlag, &pktLen);
			break;

			case	U_PROTO_PPTP:
				protoForm_vsp_pptp(pktBuf, pktt->l2ether.l2Flag,
					pktt->l2ether.dmac.octet, pktt->l2ether.smac.octet,
					pktt->l2vlan.vid, pktt->l2vlan.vlanCfi, pktt->l2vlan.priority,
					pktt->l2pppoe.pppoeType,
					pktt->l2pppoe.pppoeSession,
					pktt->l3ip.ipv,
					pktt->l3ip.iphl,
					pktt->l3ip.tos, pktt->l3ip.ipid,
					pktt->l3ip.df, pktt->l3ip.mf,
					pktt->l3ip.ipoff,
					pktt->l3ip.ttl,
					pktt->l3ip.sip, pktt->l3ip.dip,
					pktt->l4pptp.callid, pktt->l4pptp.seqno, pktt->l4pptp.ackno,
					pktt->content,
					pktt->l3ip.errorFlag, &pktLen);
			break;

			case	U_PROTO_ICMP:
				protoForm_vsp_icmp(pktBuf, pktt->l2ether.l2Flag,
					pktt->l2ether.dmac.octet, pktt->l2ether.smac.octet,
					pktt->l2vlan.vid, pktt->l2vlan.vlanCfi, pktt->l2vlan.priority,
					pktt->l2pppoe.pppoeType,
					pktt->l2pppoe.pppoeSession,
					pktt->l3ip.ipv,
					pktt->l3ip.iphl,
					pktt->l3ip.tos, pktt->l3ip.ipid,
					pktt->l3ip.df, pktt->l3ip.mf,
					pktt->l3ip.ipoff,
					pktt->l3ip.ttl,
					pktt->l3ip.sip, pktt->l3ip.dip,
					pktt->l4icmp.type, pktt->l4icmp.code,
					pktt->l4icmp.id, pktt->l4icmp.seq,
					pktt->content,
					pktt->l3ip.errorFlag, &pktLen);
			break;

			case	U_PROTO_IGMP:
				protoForm_vsp_igmp(pktBuf, pktt->l2ether.l2Flag,
					pktt->l2ether.dmac.octet, pktt->l2ether.smac.octet,
					pktt->l2vlan.vid, pktt->l2vlan.vlanCfi, pktt->l2vlan.priority,
					pktt->l2pppoe.pppoeType,
					pktt->l2pppoe.pppoeSession,
					pktt->l3ip.ipv,
					pktt->l3ip.iphl,
					pktt->l3ip.tos, pktt->l3ip.ipid,
					pktt->l3ip.df, pktt->l3ip.mf,
					pktt->l3ip.ipoff,
					pktt->l3ip.ttl,
					pktt->l3ip.sip, pktt->l3ip.dip,
					pktt->l4igmp.type, pktt->l4igmp.respTime,
					pktt->l4igmp.gaddr, 
					pktt->l3ip.errorFlag, &pktLen);
			break;

			case	U_PROTO_TCP:
				protoForm_vsp_tcp(pktBuf, pktt->l2ether.l2Flag,
					pktt->l2ether.dmac.octet, pktt->l2ether.smac.octet,
					pktt->l2vlan.vid, pktt->l2vlan.vlanCfi, pktt->l2vlan.priority,
					pktt->l2pppoe.pppoeType,
					pktt->l2pppoe.pppoeSession,
					pktt->l3ip.ipv,
					pktt->l3ip.iphl,
					pktt->l3ip.tos, pktt->l3ip.ipid,
					pktt->l3ip.df, pktt->l3ip.mf,
					pktt->l3ip.ipoff,
					pktt->l3ip.ttl,
					pktt->l3ip.sip, pktt->l3ip.dip,
					pktt->l4tcpudp.sport, pktt->l4tcpudp.dport,
					pktt->l4tcpudp.tcp_seq, pktt->l4tcpudp.tcp_ack,
					pktt->l4tcpudp.tcpFlag,
					pktt->l4tcpudp.tcp_win, 
					pktt->content, 
					pktt->l3ip.errorFlag, &pktLen);
			break;

			case	U_PROTO_UDP:
				protoForm_vsp_udp(pktBuf, pktt->l2ether.l2Flag,
					pktt->l2ether.dmac.octet, pktt->l2ether.smac.octet,
					pktt->l2vlan.vid, pktt->l2vlan.vlanCfi, pktt->l2vlan.priority,
					pktt->l2pppoe.pppoeType,
					pktt->l2pppoe.pppoeSession,
					pktt->l3ip.ipv,
					pktt->l3ip.iphl,
					pktt->l3ip.tos, pktt->l3ip.ipid,
					pktt->l3ip.df, pktt->l3ip.mf,
					pktt->l3ip.ipoff,
					pktt->l3ip.ttl,
					pktt->l3ip.sip, pktt->l3ip.dip,
					pktt->l4tcpudp.sport, pktt->l4tcpudp.dport,
					pktt->content,
					pktt->l3ip.errorFlag, &pktLen);
			break;

			default:
				return (-1);
			break;
	}

	pktLen -= 4;
	
	return	pktLen;
	
}

uint32	pktGenGeneralDefaultPacket(pktGenGeneralPktFormat_t *pktt)
{
	uint8	sa[6],da[6];
	int32	len;

	len = 60;
	
	// protocol
	pktt->protocol = U_PROTO_UDP ;
	// Layer2
	// Layer2 l2ether
	pktt->l2ether.l2Flag = 0;
	da[0] = 0x00; da[1] = 0x01; da[2] = 0x01;
	da[3] = 0x01; da[4] = 0x01; da[5] = 0x01;
	memcpy(pktt->l2ether.dmac.octet,da,sizeof(da));
	sa[0] = 0x02; sa[1] = 0x02; sa[2] = 0x02;
	sa[3] = 0x02; sa[4] = 0x02; sa[5] = 0x02;
	memcpy(pktt->l2ether.smac.octet,sa,sizeof(sa));
	pktt->l2ether.etherType = 0x0800 ;
	// Layer2 LLC Other
	pktt->l2llc.dsap = 0xaa;
	pktt->l2llc.ssap = 0xaa;
	// Layer2 vlan
	pktt->l2vlan.vid = 24 ;
	pktt->l2vlan.vlanCfi = 0 ;
	pktt->l2vlan.priority = 0 ;
	// Layer2 PPPoE
	pktt->l2pppoe.pppoeType = 0x8864;
	pktt->l2pppoe.pppoeSession = 5 ;
	// Layer3
	// Layer3 IPX
	pktt->l3ipx.ipxtype = 0x00;
	pktt->l3ipx.dnet = 0x1;
	memset(pktt->l3ipx.dnode,255,6);
	pktt->l3ipx.dsock = 0x0452;
	pktt->l3ipx.snet = 0x1;
	memcpy(pktt->l3ipx.snode, sa,sizeof(sa));
	pktt->l3ipx.ssock = 0x0452;
	// Layer3 IP
	pktt->l3ip.sip = ntohl(inet_addr("192.168.0.1")) ;
	pktt->l3ip.dip = ntohl(inet_addr("192.168.0.100")) ;
	pktt->l3ip.ipv = 4;
	pktt->l3ip.iphl = 20 ;
	pktt->l3ip.tos = 0 ;
	pktt->l3ip.ipid = 0 ;
	pktt->l3ip.df = 0 ;
	pktt->l3ip.mf = 0 ;
	pktt->l3ip.ipoff = 0;
	pktt->l3ip.ttl = 16 ;
	pktt->l3ip.proto = 89 ;
	pktt->l3ip.errorFlag = 0 ;
	// Layer3 Arp
	pktt->l3arp.op = 1 ;
	memset(pktt->l3arp.sha.octet,0,6);
	pktt->l3arp.spa = ntohl(inet_addr("192.168.0.1")) ;
	memset(pktt->l3arp.tha.octet,0,6);
	pktt->l3arp.tpa = ntohl(inet_addr("192.168.0.100")) ;
	// Layer4
	// PPTP GRE
	pktt->l4pptp.callid = 1;
	pktt->l4pptp.seqno = 1;
	pktt->l4pptp.ackno = 1;
	// Layer4 ICMP
	pktt->l4icmp.type = ICMP_ECHO;
	pktt->l4icmp.code = 0 ;
	pktt->l4icmp.id = 12 ;
	pktt->l4icmp.seq = 1 ;
	// Layer4 IGMP
	pktt->l4igmp.type = IGMP_MEMBERSHIP_QUERY ;
	pktt->l4igmp.respTime = 0 ;
	pktt->l4igmp.gaddr = ntohl(inet_addr("224.1.1.1"));
	// Layer4 TCP UDP
	pktt->l4tcpudp.sport = 1 ;
	pktt->l4tcpudp.dport = 2 ;
	pktt->l4tcpudp.tcp_seq = 2 ;
	pktt->l4tcpudp.tcp_ack = 1 ;
	pktt->l4tcpudp.tcpFlag = (pktt->l4tcpudp.tcp_ack)?0x10:0 ;
	pktt->l4tcpudp.tcp_win = 4096 ;
	// Payload
	pktt->len = len ;
	memset(pktt->content,0x00,2048) ;

	return 0;
}

uint32 pktParseGeneral(uint8 *pktBuf, pktGenGeneralPktFormat_t *pktt, uint32 pktLen)
{
    pktParse_t parset;
    
    pktParse((int8*)pktBuf, &parset, pktLen);

    /* MUST be an Ethernet Frame */
    if(parset.ether2 == NULL)
    	return FAILED;

    bzero(pktt, sizeof(pktGenGeneralPktFormat_t));

    /* L2 Information */
    memcpy(pktt->l2ether.dmac.octet,parset.ether2->da,6);
    memcpy(pktt->l2ether.smac.octet,parset.ether2->sa,6);
    pktt->l2ether.etherType = parset.ether2->etherType;
    pktt->protocol = U_PROTO_ETHER;

    // VLAN Header
    if(parset.vlan != NULL) {
    	pktt->l2ether.l2Flag |= L2_VLAN;
	pktt->l2vlan.vid = ntohs(parset.vlan->tag)&0xfff;
	pktt->l2vlan.vlanCfi = (ntohs(parset.vlan->tag)>>12)&0x1;
	pktt->l2vlan.priority = (ntohs(parset.vlan->tag)>>13)&0x7;
    	}

    // LLC/SNAP Header
    if(parset.llc != NULL) {
    	pktt->l2ether.l2Flag |= L2_SNAP;
    	pktt->l2llc.dsap = parset.llc->llc_dsap;
    	pktt->l2llc.ssap = parset.llc->llc_ssap;    	
    	}

    // PPPoE Header
    if(parset.pppoe != NULL) {
    	pktt->l2ether.l2Flag |= L2_PPPoE;
    	pktt->l2pppoe.pppoeSession= ntohs(parset.pppoe->sessionId);
    	}

    // ARP
    if(parset.arp != NULL) {
    	pktt->l3arp.op = ntohs(parset.arp->ar_op);
    	memcpy(pktt->l3arp.sha.octet,parset.arp->ar_sha,6);
    	pktt->l3arp.spa = ntohl((ipaddr_t)parset.arp->ar_spa);
    	memcpy(pktt->l3arp.tha.octet,parset.arp->ar_tha,6);
    	pktt->l3arp.tpa = ntohl((ipaddr_t)parset.arp->ar_tpa);
    	pktt->protocol = U_PROTO_ARP;
    	return SUCCESS;
    	}

    // IPX
    if(parset.ipx != NULL) {
    	pktt->l3ipx.ipxtype = parset.ipx->ipx_type;
    	pktt->l3ipx.dnet = ntohl(parset.ipx->ipx_dnet);
    	memcpy(pktt->l3ipx.dnode,parset.ipx->ipx_dnode,6);
    	pktt->l3ipx.dsock = ntohs(parset.ipx->ipx_dsock);
    	pktt->l3ipx.snet = ntohl(parset.ipx->ipx_snet);
    	memcpy(pktt->l3ipx.snode,parset.ipx->ipx_snode,6);
    	pktt->l3ipx.ssock = ntohs(parset.ipx->ipx_ssock);
    	pktt->protocol = U_PROTO_IPX;
    	return SUCCESS;
    	}
    
    // IP
    if(parset.ip != NULL) {
    	pktt->l3ip.sip = ntohl(parset.ip->ip_src);
    	pktt->l3ip.dip = ntohl(parset.ip->ip_dst);
    	pktt->l3ip.ipv = parset.ip->ip_v;
    	pktt->l3ip.iphl = parset.ip->ip_hl;
    	pktt->l3ip.tos = parset.ip->ip_tos;
    	pktt->l3ip.ipid = ntohs(parset.ip->ip_id);
    	pktt->l3ip.df = (ntohs(parset.ip->ip_off)&0x4000) >> 14;
    	pktt->l3ip.mf = (ntohs(parset.ip->ip_off)&0x2000) >> 13;
    	pktt->l3ip.ipoff = (ntohs(parset.ip->ip_off)&0x1fff)*8;
    	pktt->l3ip.ttl = parset.ip->ip_ttl;
    	pktt->l3ip.proto = parset.ip->ip_p;
    	if(parset.l3Flag == PKT_PARSE_L3_CKSUM_ERR)
    		pktt->l3ip.errorFlag |= L3_CKSUM_ERR;
    	pktt->len = parset.dataLen;
    	pktt->ipDataLen = parset.ipDataLen;
    	pktt->protocol = U_PROTO_IP;
    	}

    // PPTP
    if(parset.pptp != NULL) {
    	pktt->l4pptp.callid = ntohs(parset.pptp->gre_callid);
    	pktt->l4pptp.seqno = ntohl(parset.pptp->gre_seqno);
    	pktt->l4pptp.ackno = ntohl(parset.pptp->gre_ackno);
    	pktt->len = parset.dataLen;
    	pktt->protocol = U_PROTO_PPTP;
    	}

    // ICMP
    if(parset.icmp != NULL) {
    	pktt->l4icmp.type = parset.icmp->icmp_type;
    	pktt->l4icmp.code = parset.icmp->icmp_code;
    	pktt->l4icmp.id = ntohs(parset.icmp->content.echo_s.icd_id);
    	pktt->l4icmp.seq = ntohs(parset.icmp->content.echo_s.icd_seq);
    	if(parset.l4Flag == PKT_PARSE_L4_CKSUM_ERR)
    		pktt->l3ip.errorFlag |= L4_CKSUM_ERR;
    	pktt->len = parset.dataLen;
    	pktt->protocol = U_PROTO_ICMP;
    	}

    // IGMP
    if(parset.igmp != NULL) {
    	pktt->l4igmp.type = parset.igmp->igmp_type;
    	pktt->l4igmp.respTime = parset.igmp->igmp_respTime;
    	pktt->l4igmp.gaddr = ntohl(parset.igmp->igmp_group);
    	if(parset.l4Flag == PKT_PARSE_L4_CKSUM_ERR)
    		pktt->l3ip.errorFlag |= L4_CKSUM_ERR;
    	pktt->len = parset.dataLen;
    	pktt->protocol = U_PROTO_IGMP;
    	}

    // TCP
    if(parset.tcp != NULL) {
    	pktt->l4tcpudp.sport = ntohs(parset.tcp->th_sport);
    	pktt->l4tcpudp.dport = ntohs(parset.tcp->th_dport);
    	pktt->l4tcpudp.tcp_seq = ntohl(parset.tcp->th_seq);
    	pktt->l4tcpudp.tcp_ack = ntohl(parset.tcp->th_ack);
    	pktt->l4tcpudp.tcpFlag = parset.tcp->th_flags;
    	pktt->l4tcpudp.tcp_win = ntohs(parset.tcp->th_win);
    	if(parset.l4Flag == PKT_PARSE_L4_CKSUM_ERR)
    		pktt->l3ip.errorFlag |= L4_CKSUM_ERR;
    	pktt->len = parset.dataLen;
    	pktt->protocol = U_PROTO_TCP;
    	}

    // UDP
    if(parset.udp != NULL) {
    	pktt->l4tcpudp.sport = ntohs(parset.udp->uh_sport);
    	pktt->l4tcpudp.dport = ntohs(parset.udp->uh_dport);
    	if(parset.l4Flag == PKT_PARSE_L4_CKSUM_ERR)
    		pktt->l3ip.errorFlag |= L4_CKSUM_ERR;
    	pktt->len = parset.dataLen;
    	pktt->protocol = U_PROTO_UDP;
    	}

	memset(pktt->content,0,2048);
    	memcpy(pktt->content,parset.payload,parset.dataLen);
    	
    	return SUCCESS;

}
