/*
* ----------------------------------------------------------------
* Copyright c              Realtek Semiconductor Corporation, 2002
* All rights reserved.
* 
* $Header: /home/cvsroot/SBU_III_2/wpnt834/ntg/src_from_v1036/uClinux-dist/linux-2.4.x/drivers/net/re865x/rtl865x/veri/pktgen/pg_dumpPkt.c,v 1.1.1.1 2006/07/26 05:38:30 blue Exp $
*
* Abstract: packet generation module. various protocol
*           analyze and dump routine.
*
* $Author: blue $
*
* $Log: pg_dumpPkt.c,v $
* Revision 1.1.1.1  2006/07/26 05:38:30  blue
*
*
* Revision 1.1.1.1  2006/02/14 03:09:26  jason
* backup the source code from v1036
*
* Revision 1.1.1.1  2006/01/16 02:40:11  jason
* backup source,because now we directly use romedriver source code, i open a new directory;
*
* Revision 1.1.1.1  2005/11/21 05:20:04  chenyl
* +: WPNT834
* 	- System provider	: Sercomm
* 	- Brand Name		: Netgear
* 	- 3rt Party		: Airgo Gen 3
* 	- SDK			: 066-3
* 	- ROMEDRV		: 3.6-3
*
*
* Revision 1.1  2004/03/26 06:31:34  danwu
* init
*
* Revision 1.10  2004/02/02 07:03:35  danwu
* no message
*
* Revision 1.9  2004/01/08 14:54:56  waynelee
* add "dump IPX SAP"
*
* Revision 1.8  2003/12/30 10:34:43  waynelee
* add pptp, ipx support and provide more flexible packet format
*
* Revision 1.7  2003/12/19 02:24:17  danwu
* no message
*
* Revision 1.6  2003/09/29 14:14:58  waynelee
* dumpUdp bug fix (dport and sport)
*
* Revision 1.5  2003/05/02 09:10:41  cfliu
* no message
*
* Revision 1.4  2002/08/29 03:50:00  orlando
* Finetune pg_pktDump().
*
* Revision 1.3  2002/08/28 14:07:41  henryCho
* Correct printf() argument of pg_pktDump().
*
* Revision 1.2  2002/08/28 08:43:30  orlando
* Add pg_pktDump().
*
* Revision 1.1  2002/08/28 03:43:28  orlando
* packet generation module initial check-in.
*
* ---------------------------------------------------------------
*/

#include "../../types.h"
#include "pktgen.h" // exported prototype declaration

void dumpEther(etherHdr_t * ether) {

	if(ether == NULL)
		return;

	printf("DMAC: ");
	printf("%02x-%02x-%02x-%02x-%02x-%02x ",ether->da[0],ether->da[1],ether->da[2],ether->da[3],ether->da[4],ether->da[5]);
	printf("SMAC: ");
	printf("%02x-%02x-%02x-%02x-%02x-%02x ",ether->sa[0],ether->sa[1],ether->sa[2],ether->sa[3],ether->sa[4],ether->sa[5]);
	if(ntohs(ether->etherType)<0x0600)
		printf("Length: %04x (%u bytes)\n", ntohs(ether->etherType), ntohs(ether->etherType));
	else {
		printf("EtherType: %04x ", ntohs(ether->etherType));
		switch(ntohs(ether->etherType)) {
		case 0x8137:
			printf("(IPX)\n");
			break;
		case 0x0800:
			printf("(IP)\n");
			break;
		case 0x0806:
			printf("(ARP)\n");
			break;
		case 0x8863:
			printf("(PPPoE Discovery)\n");
			break;
		case 0x8864:
			printf("(PPPoE Session)\n");
			break;
		case 0x8035:
			printf("(RARP)\n");
			break;
		case 0x8100:
			printf("(VLAN)\n");
			break;
		default:
			printf("\n");
			break;
		}
	}
}

void dumpVlan(vlanHdr_t * vlan) {
	if(vlan==NULL)
		return;
	printf("Vlan priority %d CFI:%d VID:%d\n",
		(ntohs(vlan->tag)>>13)&0x7, (ntohs(vlan->tag)>>12)&0x1,
		ntohs(vlan->tag)&0xfff);
	if(ntohs(vlan->etherType)<0x0600)
		printf("Length: %04x (%u bytes)\n", ntohs(vlan->etherType), ntohs(vlan->etherType));
	else {
		printf("EtherType: %04x ", ntohs(vlan->etherType));
		switch(ntohs(vlan->etherType)) {
		case 0x8137:
			printf("(IPX)\n");
			break;
		case 0x0800:
			printf("(IP)\n");
			break;
		case 0x0806:
			printf("(ARP)\n");
			break;
		case 0x8863:
			printf("(PPPoE Discovery)\n");
			break;
		case 0x8864:
			printf("(PPPoE Session)\n");
			break;
		case 0x8035:
			printf("(RARP)\n");
			break;
		case 0x8100:
			printf("(VLAN)\n");
			break;
		default:
			printf("\n");
			break;
		}
	}
}

void dumpLlc(llcHdr_t * llc) {
	uint32 i;

	if(llc == NULL)
		return;
	printf("DSAP:%02x SSAP:%02x CTL:%x\n",llc->llc_dsap, llc->llc_ssap, llc->ctrl);
	if(llc->llc_dsap == 0xe0 && llc->llc_ssap == 0xe0) {
		printf("IPX SAP\n");
		return;
	}

	printf("Org Code: ");
	for(i=0;i<3; i++)
		printf("%02x:",llc->org_code[i]);
	printf(" EtherType: %04x ", ntohs(llc->etherType));
	switch(ntohs(llc->etherType)) {
	case 0x8137:
		printf("(IPX)\n");
		break;
	case 0x0800:
		printf("(IP)\n");
		break;
	case 0x0806:
		printf("(ARP)\n");
		break;
	case 0x8863:
		printf("(PPPoE Discovery)\n");
		break;
	case 0x8864:
		printf("(PPPoE Session)\n");
		break;
	case 0x8035:
		printf("(RARP)\n");
		break;
	case 0x8100:
		printf("(VLAN)\n");
		break;
	default:
		printf("\n");
		break;
	}
}

void dumpPppoe(pppoeHdr_t * pppoe) {
	if(pppoe == NULL)
		return;

	printf("Ver: %x Type: %x Code:%x\n",pppoe->ver, pppoe->type, pppoe->code);
	printf("Session:%x Length:%x PPP Protocol:%04x ",
			ntohs(pppoe->sessionId), ntohs(pppoe->length), ntohs(pppoe->proto));
	switch(ntohs(pppoe->proto)) {
	case 0x8137:
		printf("(IPX)\n");
		break;
	case 0x0021:
		printf("(IP)\n");
		break;
	default:
		printf("\n");
		break;
	}
}

void dumpIpx(ipxHdr_t * ipx) {

	if(ipx == NULL)
		return;
	printf("IPX Checksum:0x%04x, Length:%d, TCtrl:0x%02x, Type:0x%02x\n",
		ntohs(ipx->ipx_sum), ntohs(ipx->ipx_len), ipx->ipx_tctrl, ipx->ipx_type);

	printf("dest net:%d, node:%02x-%02x-%02x-%02x-%02x-%02x, sock: 0x%04x\n",
		ntohl(ipx->ipx_dnet),ipx->ipx_dnode[0],ipx->ipx_dnode[1],ipx->ipx_dnode[2],ipx->ipx_dnode[3],ipx->ipx_dnode[4],ipx->ipx_dnode[5],
		ntohs(ipx->ipx_dsock));

	printf("sour net:%d, node:%02x-%02x-%02x-%02x-%02x-%02x, sock: 0x%04x\n",
		ntohl(ipx->ipx_snet),ipx->ipx_snode[0],ipx->ipx_snode[1],ipx->ipx_snode[2],ipx->ipx_snode[3],ipx->ipx_snode[4],ipx->ipx_snode[5],
		ntohs(ipx->ipx_ssock));

	printf("\n");
}

void dumpArp(arpHdr_t * arp) {
	if(arp == NULL)
		return;
	printf("Hardware:%x Protocol:%x HardwareLen:%x ProtocolLen:%x ",
		ntohs(arp->ar_hrd), ntohs(arp->ar_pro), arp->ar_hln, arp->ar_pln);
	switch(ntohs(arp->ar_op)) {
	case ARPOP_REQUEST:
		printf("ARP request\n");
		break;
	case ARPOP_REPLY:
		printf("ARP reply\n");
		break;
	case ARPOP_REVREQUEST:
		printf("RARP Request\n");
		break;
	case ARPOP_REVREPLY:
		printf("RARP Reply\n");
		break;
	}
	printf("Sender MAC: ");
	printf("%02x-%02x-%02x-%02x-%02x-%02x ",arp->ar_sha[0],arp->ar_sha[1],arp->ar_sha[2],arp->ar_sha[3],arp->ar_sha[4],arp->ar_sha[5]);
	printf("Sender IP: ");
	printf("%d.%d.%d.%d",arp->ar_spa[0],arp->ar_spa[1],arp->ar_spa[2],arp->ar_spa[3]);
	printf("\n");

	printf("Target MAC: ");
	printf("%02x-%02x-%02x-%02x-%02x-%02x ",arp->ar_tha[0],arp->ar_tha[1],arp->ar_tha[2],arp->ar_tha[3],arp->ar_tha[4],arp->ar_tha[5]);
	printf("Target IP: ");
	printf("%d.%d.%d.%d",arp->ar_tpa[0],arp->ar_tpa[1],arp->ar_tpa[2],arp->ar_tpa[3]);
	printf("\n");
}

void dumpIp(ipHdr_t * ip, uint32 l3Flag, uint16 l3Cksum) {
	struct in_addr ina;

	if(ip == NULL)
		return;
	printf("Ver:%d Header Len:%d TOS:%02x Total Length:%d\n",
		ip->ip_v, ip->ip_hl*4, ip->ip_tos, ntohs(ip->ip_len));
	printf("ID:0x%x flag: ", ntohs(ip->ip_id));
	if(ntohs(ip->ip_off) & IP_DF)
		printf("Don't fragment ");
	if(ntohs(ip->ip_off) & IP_MF)
		printf("More fragment ");
	printf(" Offset: %d bytes\n",(uint16)(ntohs(ip->ip_off)&0x1fff)*8);
	printf("TTL:%d Layer4 Protocol:",ip->ip_ttl);
	switch(ip->ip_p) {
	case IP_PROTO_PPTP:
		printf("PPTP ");
		break;
	case IP_PROTO_ICMP:
		printf("ICMP ");
		break;
	case IP_PROTO_IGMP:
		printf("IGMP ");
		break;
	case IP_PROTO_TCP:
		printf("TCP ");
		break;
	case IP_PROTO_UDP:
		printf("UDP ");
		break;
	default:
		printf("%d",ip->ip_p);
		break;
	}
	if(l3Flag & PKT_PARSE_L3_CKSUM_ERR)
		printf("Checksum:%04x Error! (%04x)\n", ntohs(ip->ip_sum), l3Cksum);
	else if(l3Flag & PKT_PARSE_L3_CKSUM_OK)
		printf("Checksum:%04x Correct!\n", ntohs(ip->ip_sum));
	else
		printf("Cecksum:%04x Unchecked!\n", ntohs(ip->ip_sum));
	ina.s_addr = ip->ip_src;
	printf("Source IP: %s ",inet_ntoa(ina));
	ina.s_addr = ip->ip_dst;
	printf("Destination IP: %s\n",inet_ntoa(ina));
}

void dumpPptp(greHdr_t * pptp) {
	if(pptp == NULL)
		return;

	printf("PPTP GRE ");
	printf("Key(HW)Payload Length: %d, Key(LW)Call ID: %d\n",ntohs(pptp->gre_length),ntohs(pptp->gre_callid));
	printf("seq. no: %u, ack no: %u\n",ntohl(pptp->gre_seqno),ntohl(pptp->gre_ackno));
}



void dumpIcmp(icmpHdr_t * icmp, uint32 l4Flag, uint16 l4Cksum) {
	if(icmp == NULL)
		return;
	switch(icmp->icmp_type) {
	case ICMP_ECHOREPLY:
		printf("Echo reply ");
		if(icmp->icmp_code != 0)
			printf("Error code value! (%d) ", icmp->icmp_code);
		if(l4Flag & PKT_PARSE_L4_CKSUM_ERR)
			printf("Checksum %04x Error! (%04x)", ntohs(icmp->icmp_cksum), l4Cksum);
		else if(l4Flag & PKT_PARSE_L4_CKSUM_OK)
			printf("Checksum %04x Correct!", ntohs(icmp->icmp_cksum));
		else
			printf("Checksum %04x Unchecked!", ntohs(icmp->icmp_cksum));
		printf("ID:%d Seq:%d\n", ntohs(icmp->content.echo_s.icd_id),
			ntohs(icmp->content.echo_s.icd_seq));
		break;
	case ICMP_ECHO:
		printf("Echo request ");
		if(icmp->icmp_code != 0)
			printf("Error code value! (%d) ", icmp->icmp_code);
		if(l4Flag & PKT_PARSE_L4_CKSUM_ERR)
			printf("Checksum %04x Error! (%04x)", ntohs(icmp->icmp_cksum), l4Cksum);
		else if(l4Flag & PKT_PARSE_L4_CKSUM_OK)
			printf("Checksum %04x Correct!", ntohs(icmp->icmp_cksum));
		else
			printf("Checksum %04x Unchecked!", ntohs(icmp->icmp_cksum));
		printf("ID:%d Seq:%d\n", ntohs(icmp->content.echo_s.icd_id),
			ntohs(icmp->content.echo_s.icd_seq));
		break;
	default:
		if(l4Flag & PKT_PARSE_L4_CKSUM_ERR)
			printf("Type:%d code:%d Checksum:%04x Error! (%04x)\n",
				icmp->icmp_type, icmp->icmp_code, icmp->icmp_cksum, l4Cksum);
		else if(l4Flag & PKT_PARSE_L4_CKSUM_OK)
			printf("Type:%d code:%d Checksum:%04x Correct!\n",
				icmp->icmp_type, icmp->icmp_code, icmp->icmp_cksum);
		else
			printf("Type:%d code:%d Checksum:%04x Unchecked!\n",
				icmp->icmp_type, icmp->icmp_code, icmp->icmp_cksum);
		break;
	}
}

void dumpIgmp(igmpHdr_t * igmp, uint32 l4Flag, uint16 l4Cksum) {
	struct in_addr ina;
	
	if(igmp == NULL)
		return;

	ina.s_addr = igmp->igmp_group ;

	printf("IGMP Group Address: %s\n",inet_ntoa(ina));
	switch(igmp->igmp_type) {
	case IGMP_MEMBERSHIP_QUERY:
		printf("Member query Max response time %d ",igmp->igmp_respTime);
		break;
	case IGMP_V1_MEMBERSHIP_REPORT:
		printf("Version1 Member report ");
		if(igmp->igmp_respTime != 0)
			printf("Error Max Response Time value (%d) ", igmp->igmp_respTime);
		break;
	case IGMP_V2_MEMBERSHIP_REPORT:
		printf("Version2 Member report ");
		if(igmp->igmp_respTime != 0)
			printf("Error Max Response Time value (%d) ", igmp->igmp_respTime);
		break;
	case IGMP_V2_LEAVE_GROUP:
		printf("Leave report ");
		if(igmp->igmp_respTime != 0)
			printf("Error Max Response Time value (%d) ", igmp->igmp_respTime);
		break;
	case IGMP_DVMRP:
		printf("DVMRP ");
		if(igmp->igmp_respTime != 0)
			printf("Error Max Response Time value (%d) ", igmp->igmp_respTime);
		break;
	case IGMP_PIM:
		printf("PIM ");
		if(igmp->igmp_respTime != 0)
			printf("Error Max Response Time value (%d) ", igmp->igmp_respTime);
		break;
	}
	if(l4Flag & PKT_PARSE_L4_CKSUM_ERR)
		printf("Checksum %04x Error! (%04x)\n", ntohs(igmp->igmp_cksum), l4Cksum);
	else if(l4Flag & PKT_PARSE_L4_CKSUM_OK)
		printf("Checksum %04x Correct! \n", ntohs(igmp->igmp_cksum));
	else
		printf("Checksum %04x Unchecked! \n", ntohs(igmp->igmp_cksum));
}


void dumpTcp(tcpHdr_t * tcp, uint32 l4Flag, uint16 l4Cksum) {
	if(tcp == NULL)
		return;
	printf("Source port:%u Destination port:%u Header Len:%u\n", ntohs(tcp->th_sport), ntohs(tcp->th_dport), tcp->th_off*4);
	printf("Sequence number:%u Acknowledge number:%u\n",ntohl(tcp->th_seq), ntohl(tcp->th_ack));
	printf("Flag:");
	if(tcp->th_flags & TH_URG)
		printf("Urgent ");
	if(tcp->th_flags & TH_ACK)
		printf("Ack ");
	if(tcp->th_flags & TH_PUSH)
		printf("Push ");
	if(tcp->th_flags & TH_RST)
		printf("Reset ");
	if(tcp->th_flags & TH_SYN)
		printf("Sync ");
	if(tcp->th_flags & TH_FIN)
		printf("Fin ");
	printf("Window size %u\n", ntohs(tcp->th_win));
	if(l4Flag & PKT_PARSE_L4_CKSUM_ERR)
		printf("Checksum:%04x Error! (%04x) Urgent pointer:%u\n",ntohs(tcp->th_sum), l4Cksum, ntohs(tcp->th_urp));
	else if(l4Flag & PKT_PARSE_L4_CKSUM_OK)
		printf("Checksum:%04x Correct! Urgent pointer:%u\n",ntohs(tcp->th_sum), ntohs(tcp->th_urp));
	else
		printf("Checksum:%04x Unchecked! Urgent pointer:%u\n",ntohs(tcp->th_sum), ntohs(tcp->th_urp));
}

void dumpUdp(udpHdr_t * udp, uint32 l4Flag, uint16 l4Cksum) {
	if(udp == NULL)
		return;
	printf("Source port:%d Destination port:%d\n",
		ntohs(udp->uh_sport), ntohs(udp->uh_dport));

	if(l4Flag & PKT_PARSE_UDP_NO_CKSUM)
		printf("Payload len:%d Checksum:%04x (No Cksum)\n", ntohs(udp->uh_ulen), ntohs(udp->uh_sum));
	else if (l4Flag & PKT_PARSE_L4_CKSUM_ERR)
		printf("Payload len:%d Checksum:%04x Error! (%04x)\n", ntohs(udp->uh_ulen), ntohs(udp->uh_sum), l4Cksum);
	else if(l4Flag & PKT_PARSE_L4_CKSUM_OK)
		printf("Payload len:%d Checksum:%04x Correct! \n", ntohs(udp->uh_ulen), ntohs(udp->uh_sum));
	else
		printf("Payload len:%d Checksum:%04x Unchecked! \n", ntohs(udp->uh_ulen), ntohs(udp->uh_sum));
}

/*
 * dump raw packet contents
 */
void pg_pktDump(int8 * pktBuf, uint32 pktLen)
{
	uint32 i;
	for (i=0; i<pktLen; i++)
	{
		printf("%02x",(uint8)pktBuf[i]);
		// print a new line for every 16 bytes
		if ((i & 0x0000000f) == 0xf)
			printf("\n");
		else
			printf("-");
	}
		
}

/* end of pg_dumpPkt.c */
