/*
* ----------------------------------------------------------------
* Copyright c              Realtek Semiconductor Corporation, 2002
* All rights reserved.
* 
* $Header: /home/cvsroot/SBU_III_2/wpnt834/ntg/src_from_v1036/uClinux-dist/linux-2.4.x/drivers/net/re865x/rtl865x/veri/pktgen/pg_pktForm.c,v 1.1.1.1 2006/07/26 05:38:30 blue Exp $
*
* Abstract: packet generator module. packet form.
*
* $Author: blue $
*
* $Log: pg_pktForm.c,v $
* Revision 1.1.1.1  2006/07/26 05:38:30  blue
*
*
* Revision 1.1.1.1  2006/02/14 03:09:26  jason
* backup the source code from v1036
*
* Revision 1.1.1.1  2006/01/16 02:40:11  jason
* backup source,because now we directly use romedriver source code, i open a new directory;
*
* Revision 1.1.1.1  2005/11/21 05:20:04  chenyl
* +: WPNT834
* 	- System provider	: Sercomm
* 	- Brand Name		: Netgear
* 	- 3rt Party		: Airgo Gen 3
* 	- SDK			: 066-3
* 	- ROMEDRV		: 3.6-3
*
*
* Revision 1.2  2004/04/21 07:18:28  danwu
* *** empty log message ***
*
* Revision 1.1  2004/03/26 06:31:34  danwu
* init
*
* Revision 1.5  2004/01/08 14:55:59  waynelee
* correct L3_IPX and L3_ARP l3Offset caculation
*
* Revision 1.4  2003/12/30 10:34:42  waynelee
* add pptp, ipx support and provide more flexible packet format
*
* Revision 1.3  2003/05/02 09:57:40  cfliu
* remove DIV, MOD, MUL
*
* Revision 1.2  2003/04/30 15:46:09  cfliu
* remove macro.h, use macros defined in types.h instead
*
* Revision 1.1  2002/08/28 03:43:54  orlando
* packet generation module initial check-in.
*
* ---------------------------------------------------------------
*/

#include "../../types.h"
#include "pktgen.h" // exported prototype declaration
#include "pg_l2pkt.h"
#include "pg_l3pkt.h"
#include "pg_l4pkt.h"
#include "pg_pktForm.h"
#include "pg_utility.h"
#include "pg_crc32.h"

void l2FormEtherAddress(int8 * pktBuf, int8 * da, int8 * sa) {
	etherHdr_t * etherPtr = (etherHdr_t *)pktBuf;
	memcpy(etherPtr->da, da, 6);
	memcpy(etherPtr->sa, sa, 6);
}

void l2FormCrc(uint32 flag, int8 * pktBuf, uint32 payloadSize)
{
	uint32 crc, i;
	uint32 crcVerify;

	crc32(pktBuf,payloadSize,&crc);

	if (flag & L2_CRC_ERR)
		crc = ~crc;
	
	for(i=0;i<4;i++)
	{
		#if 0
		// for little endian machine
		pktBuf[payloadSize+i] = (crc>>((3-i)*8)) & 0xff;
		#else
		// for big endian machine
		pktBuf[payloadSize+i] = (crc>>(i*8)) & 0xff;
		#endif
	}

} /* end l2FormCrc */

void l2FormVlanTag(uint32 flag, int8 * pktBuf, uint16 vlanTag, uint8 vlanCfi, uint16 vlanPriority) {
	vlanHdr_t * vlanPtr;
	etherHdr_t * etherPtr;
	uint16 tag;

	if((flag & L2_VLAN) == NULL)
		return;
	
	etherPtr = (etherHdr_t *)pktBuf;
	etherPtr->etherType = htons(0x8100);

	vlanPtr = (vlanHdr_t *)(pktBuf + sizeof(etherHdr_t));

	if(vlanCfi !=0)
		vlanCfi =1;

	tag = (vlanTag & 0x0fff) | (vlanPriority & 0x7)<<13 | (vlanCfi <<12) ;
	vlanPtr->tag = htons(tag);
}

void l2FormLlc(uint32 flag, uint8 dsap, uint8 ssap, int8 * pktBuf) {
	int8 * tmpPtr;
	llcHdr_t * llcPtr;
	
	if((flag & L2_SNAP) == NULL)
		return;

	tmpPtr = pktBuf + sizeof(etherHdr_t);//Ethernet is the MUST
	if((flag & L2_VLAN) == L2_VLAN)
		tmpPtr += sizeof(vlanHdr_t);
	
	llcPtr = (llcHdr_t *)tmpPtr;
	llcPtr->llc_dsap = dsap;
	llcPtr->llc_ssap = ssap;
	llcPtr->ctrl = 0x03;
	memset(llcPtr->org_code, 0, 3);
}

void l2FormPppoeSession(uint32 flag, int8 * pktBuf, uint16 etherType, uint16 sessionId) {
	int8 * tmpPtr = pktBuf + sizeof(etherHdr_t);//Ethernet is the MUST
	pppoeHdr_t * pppoePtr;
	llcHdr_t * llcPtr;
	vlanHdr_t * vlanPtr;
	etherHdr_t * etherPtr;

	if((flag & L2_PPPoE) == NULL)
		return;

	if((flag & L2_VLAN) == NULL && (flag & L2_SNAP) == NULL) {
		etherPtr = (etherHdr_t *)pktBuf;
		etherPtr->etherType = htons(etherType);
	}
	if((flag & L2_VLAN) == L2_VLAN) {
		vlanPtr = (vlanHdr_t *)tmpPtr;
		tmpPtr += sizeof(vlanHdr_t);
		if((flag & L2_SNAP) == NULL)
			vlanPtr->etherType = htons(etherType);
	}
	if((flag & L2_SNAP) == L2_SNAP) {
		llcPtr = (llcHdr_t *)tmpPtr;
		tmpPtr += sizeof(llcHdr_t);
		llcPtr->etherType = htons(etherType);
	}

	pppoePtr = (pppoeHdr_t*)tmpPtr;
	pppoePtr->ver = 0x1;
	pppoePtr->type = 0x1;
	pppoePtr->code = 0x0;
	pppoePtr->sessionId = htons(sessionId);
}

//
void l2FormLayer2Len(uint32 flag, int8 * pktBuf, uint16 payloadLen) {
	int8 * tmpPtr = pktBuf + sizeof(etherHdr_t);//Ethernet is the MUST
	uint16 llcLen;
	pppoeHdr_t * pppoePtr;
	llcHdr_t * llcPtr;
	vlanHdr_t * vlanPtr;
	etherHdr_t * etherPtr;

	if((flag & L2_SNAP) == L2_SNAP) {
		if((flag & L2_PPPoE) == L2_PPPoE)
			llcLen = payloadLen + sizeof(llcHdr_t) + sizeof(pppoeHdr_t);
		else
			llcLen = payloadLen + sizeof(llcHdr_t);

		if ((flag & L2_VLAN) == L2_VLAN) {
			vlanPtr = (vlanHdr_t*)tmpPtr;
			vlanPtr->etherType = htons(llcLen);
		}
		else {
			etherPtr = (etherHdr_t*)pktBuf;
			etherPtr->etherType = htons(llcLen);
		}
	}
	if((flag & L2_PPPoE) == L2_PPPoE) {
		if ((flag & L2_VLAN) == L2_VLAN)
			tmpPtr += sizeof(vlanHdr_t);
		if((flag & L2_SNAP) == L2_SNAP)
			tmpPtr += sizeof(llcHdr_t);
		pppoePtr = (pppoeHdr_t *)tmpPtr;
		pppoePtr->length = htons(payloadLen + 2);//2 is the PPP header
	}
}

void l2FormLayer3Type(uint32 flag, int8 * pktBuf, uint16 type) {
	int8 * tmpPtr = pktBuf + sizeof(etherHdr_t);//Ethernet is the MUST
	pppoeHdr_t * pppoePtr;
	llcHdr_t * llcPtr;
	vlanHdr_t * vlanPtr;
	etherHdr_t * etherPtr;

	if((flag&(L2_VLAN | L2_SNAP | L2_PPPoE)) == 0) {//Ethernet only
		etherPtr = (etherHdr_t*)pktBuf;
		etherPtr->etherType = htons(type);
		return;
	}
	if ((flag & L2_VLAN) == L2_VLAN) {
		vlanPtr = (vlanHdr_t *)tmpPtr;
		tmpPtr += sizeof(vlanHdr_t);
		if((flag & L2_SNAP) == NULL && (flag & L2_PPPoE) == NULL) {
			vlanPtr->etherType = htons(type);
			return;
		}
	}
	if((flag & L2_SNAP) == L2_SNAP) {
		llcPtr = (llcHdr_t*)tmpPtr;
		tmpPtr += sizeof(llcHdr_t);
		if((flag & L2_PPPoE) == NULL) {
			llcPtr->etherType = htons(type);
			return;
		}
		else {
			pppoePtr = (pppoeHdr_t*)tmpPtr;
			pppoePtr->proto = htons(type);
			return;
		}
	}

	if((flag & L2_PPPoE) == L2_PPPoE) {
		pppoePtr = (pppoeHdr_t*)tmpPtr;
		pppoePtr->proto = htons(type);
		return;
	}
}

static int32 l3Offset(uint32 flag) {
	int32 offset;
	
	offset = sizeof(etherHdr_t);

	if ((flag & L2_VLAN) == L2_VLAN) 
		offset += sizeof(vlanHdr_t);

	if((flag & L2_SNAP) == L2_SNAP) {
		if((flag & L3_IPX) == L3_IPX)
			offset += sizeof(llcHdr_t)-5; // IPX Specific
		else
		offset += sizeof(llcHdr_t);
		}

	if((flag & L2_PPPoE) == L2_PPPoE)
		offset += sizeof(pppoeHdr_t);

	return offset;
}

static int32 bitCount(uint32 flags) {
	uint32 i, cntr = 0;

	for(i=0;i<32;i++)
		if(flags & (0x1<<i))
			cntr++;
	return cntr;
}

// user defined etherType
void l2FormOtherEtherType(uint32 flag, int8 * pktBuf, uint16 etherType, int8 * content, uint32 payloadLen) {
	
	int8 * tmpPtr = pktBuf + l3Offset(flag);

	memcpy(tmpPtr,content,payloadLen);
	l2FormLayer2Len(flag, pktBuf, payloadLen);
	l2FormLayer3Type(flag, pktBuf, etherType);

}

void l3FormIpxBasic(uint32 flag, uint8 type, int8 * pktBuf) {
	int8 * tmpPtr = pktBuf + l3Offset(flag);
	ipxHdr_t * ipxPtr;

	ipxPtr = (ipxHdr_t*)tmpPtr;

	// memset(tmpPtr, 1, sizeof(ipxHdr_t));
	ipxPtr->ipx_sum = htons(65535);
	ipxPtr->ipx_tctrl = 0; /* used by NetWare Router */
	ipxPtr->ipx_type = type;

	l2FormLayer3Type(flag,pktBuf,0x8137);
}

void l3FormIpxDnet(uint32 flag, uint32 dnet, uint8 * dnode, uint16 dsock, int8 * pktBuf) {
	int8 * tmpPtr = pktBuf + l3Offset(flag);
	ipxHdr_t * ipxPtr;

	ipxPtr = (ipxHdr_t*)tmpPtr;
	ipxPtr->ipx_dnet = htonl(dnet);
	memcpy(ipxPtr->ipx_dnode,dnode,6);
	ipxPtr->ipx_dsock = htons(dsock);
}

void l3FormIpxSnet(uint32 flag, uint32 snet, uint8 * snode, uint16 ssock, int8 * pktBuf) {
	int8 * tmpPtr = pktBuf + l3Offset(flag);
	ipxHdr_t * ipxPtr;

	ipxPtr = (ipxHdr_t*)tmpPtr;
	ipxPtr->ipx_snet = htonl(snet);
	memcpy(ipxPtr->ipx_snode,snode,6);
	ipxPtr->ipx_ssock = htons(ssock);
}

void l3FormIpxContent(uint32 flag, int8 * pktBuf, int8 * content, uint32 size) {
	int8 * tmpPtr = pktBuf + l3Offset(flag) + sizeof(ipxHdr_t);
	ipxHdr_t * ipxPtr = (ipxHdr_t*)(pktBuf + l3Offset(flag));

	memcpy(tmpPtr, content, size);
	ipxPtr->ipx_len = htons(46);

	l2FormLayer2Len(flag,pktBuf,sizeof(ipxHdr_t)+size);
}

static int32 l3FlagValidArp(uint32 flag) {
	if(bitCount(flag & L3_MASK) > 1 || (flag & L3_ARP) != L3_ARP)
		return 0;
	return 1;
}

void l3FormArpBasic(uint32 flag, uint16 op, int8 * pktBuf) {
	int8 * tmpPtr = pktBuf + l3Offset(flag);
	arpHdr_t * arpPtr;

	if(!l3FlagValidArp(flag) || (flag & L2_PPPoE) == L2_PPPoE) // PPPoE do not have Ethernet ARP
		return;

	memset(tmpPtr, 0, sizeof(arpHdr_t));
	arpPtr = (arpHdr_t*)tmpPtr;
	arpPtr->ar_hrd = htons(ARPHRD_ETHER);
	arpPtr->ar_pro = htons(0x0800);
	arpPtr->ar_hln = 0x6;
	arpPtr->ar_pln = 0x4;
//	arpPtr->ar_hln = 0x6;
	// 0x0806 is ARP ; 8035 is RARP
	if(op == ARPOP_REQUEST || op == ARPOP_REPLY || op == ARPOP_INARPREQUEST || op == ARPOP_INARPREPLY)
		l2FormLayer3Type(flag, pktBuf, 0x0806);
	else
		l2FormLayer3Type(flag, pktBuf, 0x8035);

	l2FormLayer2Len(flag, pktBuf, sizeof(arpHdr_t));
}


void l3FormArpSelfInfo(uint32 flag, int8 * pktBuf, uint32 ipAddr, int8 * macAddr) {
	int8 * tmpPtr = pktBuf + l3Offset(flag);
	uint32 i;
	arpHdr_t * arpPtr;

	if(!l3FlagValidArp(flag))
		return;

	arpPtr = (arpHdr_t*)tmpPtr;
	memcpy(arpPtr->ar_sha, macAddr, 6);
	for(i=0;i<4;i++)
		arpPtr->ar_spa[i] = (ipAddr >> (8*(3-i)))&0xff;
}

void l3FormArpPeerInfo(uint32 flag, int8 * pktBuf, uint16 op, uint32 ipAddr, int8 * macAddr) {
	int8 * tmpPtr = pktBuf + l3Offset(flag);
	uint32 i;
	arpHdr_t * arpPtr;

	if(!l3FlagValidArp(flag))
		return;

	arpPtr = (arpHdr_t*)tmpPtr;
	arpPtr->ar_op = htons(op);
	if(op == ARPOP_REPLY)
		memcpy(arpPtr->ar_tha, macAddr, 6);
	else if(op == ARPOP_REQUEST)
		memset(arpPtr->ar_tha, 0, 6);
	else
		memcpy(arpPtr->ar_tha, macAddr, 6);
	for(i=0;i<4;i++)
		arpPtr->ar_tpa[i] = (ipAddr >> (8*(3-i)))&0xff;
}

static int32 l3FlagValidIp(uint32 flag) {
	if(bitCount(flag & L3_MASK) > 1 || (flag & L3_IP) != L3_IP)
		return 0;
	return 1;
}

void l3FormIpBasic(uint32 flag, int8 * pktBuf, uint8 ver, uint8 len) {
	int8 * tmpPtr = pktBuf + l3Offset(flag);
	ipHdr_t * ipPtr;
	uint32 ipHeaderLen;

	if(!l3FlagValidIp(flag))
		return;

	if(len%4)
	//if (MOD(len,4))
		ipHeaderLen = len/4 + 1;
	else
		ipHeaderLen = len/4;

	memset(tmpPtr, 0, ipHeaderLen*4);
	ipPtr = (ipHdr_t*) tmpPtr;
	ipPtr->ip_hl = ipHeaderLen;
	ipPtr->ip_v = ver;

	if((flag & L2_PPPoE) == L2_PPPoE)
			l2FormLayer3Type(flag, pktBuf, 0x0021);
	else
			l2FormLayer3Type(flag, pktBuf, 0x0800);
	l2FormLayer2Len(flag, pktBuf, ipHeaderLen*4);
	l3FormIpChecksum(flag, pktBuf);
}

void l3FormIpFlow(uint32 flag, int8 *pktBuf, uint8 tos, uint8 ttl) {
	ipHdr_t * ipPtr = (ipHdr_t *)(pktBuf + l3Offset(flag));

	if(!l3FlagValidIp(flag))
		return;

	ipPtr->ip_tos = tos;
	ipPtr->ip_ttl = ttl;
	l3FormIpChecksum(flag, pktBuf);
}

void l3FormIpTotalLen(uint32 flag, int8 * pktBuf, uint16 len) {
	ipHdr_t * ipPtr = (ipHdr_t *)(pktBuf + l3Offset(flag));

	if(!l3FlagValidIp(flag))
		return;

	ipPtr->ip_len = htons(len + ipPtr->ip_hl*4);
	l3FormIpChecksum(flag, pktBuf);
	l2FormLayer2Len(flag, pktBuf, len + ipPtr->ip_hl*4);
}

void l3FormIpIdFrag(uint32 flag, int8 * pktBuf, uint16 id, uint8 df, uint8 mf, uint16 offset) {
	uint16	ip_off;
	ipHdr_t * ipPtr = (ipHdr_t *)(pktBuf + l3Offset(flag));

	if(!l3FlagValidIp(flag))
		return;

	ip_off = 0x0;
	if(df)
		ip_off |= IP_DF;
	else
		ip_off &= ~IP_DF;

	if(mf)
		ip_off |= IP_MF;
	else
		ip_off &= ~IP_MF;

	ip_off |= offset & IP_OFFMASK;

	ipPtr->ip_id = htons(id);
	ipPtr->ip_off = htons(ip_off);
	l3FormIpChecksum(flag, pktBuf);
}


void l3FormIpProtocol(uint32 flag, int8 * pktBuf, int8 protocol){
	int8 * tmpPtr = pktBuf + l3Offset(flag);
	ipHdr_t * ipPtr;

	if(!l3FlagValidIp(flag))
		return;

	ipPtr = (ipHdr_t*) tmpPtr;
	ipPtr->ip_p = protocol;
	l3FormIpChecksum(flag, pktBuf);
}

void l3FormIpAddr(uint32 flag, int8 * pktBuf, uint32 sip, uint32 dip){
	ipHdr_t * ipPtr = (ipHdr_t*)(pktBuf + l3Offset(flag));

	if(!l3FlagValidIp(flag))
		return;

	ipPtr->ip_src = htonl(sip);
	ipPtr->ip_dst = htonl(dip);
	l3FormIpChecksum(flag, pktBuf);
}

void l3FormIpContent(uint32 flag, int8 * pktBuf, int8 * content, uint32 size) {
	ipHdr_t * ipPtr = (ipHdr_t*)(pktBuf + l3Offset(flag));
	int8 * tmpPtr = pktBuf + l3Offset(flag) + ipPtr->ip_hl * 4;
	
	memcpy(tmpPtr, content, size);
	l3FormIpTotalLen(flag, pktBuf, size);
	l3FormIpChecksum(flag, pktBuf);
}

void l3FormIpChecksum(uint32 flag, int8 *pktBuf) {
	int8 * tmpPtr = pktBuf + l3Offset(flag);
	ipHdr_t * ipPtr;
	uint16 cksum;

	if(!l3FlagValidIp(flag))
		return;

	ipPtr = (ipHdr_t*) tmpPtr;
	ipPtr->ip_sum = 0x0;
	cksum = ipcsum((uint16*)ipPtr, ipPtr->ip_hl * 4, 0);

	if(flag & L3_CKSUM_ERR)
		ipPtr->ip_sum = htons(cksum);
	else
		ipPtr->ip_sum = htons(~cksum);
}

static int32 l4Offset(uint32 flag, int8 * pktBuf) {
	int8 * tmpPtr = pktBuf + l3Offset(flag);
	ipHdr_t * ipPtr = (ipHdr_t *)tmpPtr;
	
	if((flag & L3_MASK) == 0)
		return l3Offset(flag);
	else if (flag & L3_ARP)
		return l3Offset(flag) + sizeof(arpHdr_t);

	return l3Offset(flag) + ipPtr->ip_hl * 4;
}

static uint16 l4PayloadLen(uint32 flag, int8 * pktBuf) {
	ipHdr_t * ipPtr = (ipHdr_t *)(pktBuf + l3Offset(flag));

	if((flag & L4_MASK) == 0)
		return 0;

	return ntohs(ipPtr->ip_len) - ipPtr->ip_hl * 4;
}

static int32 l4FlagValidPptp(uint32 flag) {
	if(!l3FlagValidIp(flag) ||
		bitCount(flag & L4_MASK) > 1 || (flag & L4_PPTP) != L4_PPTP)
		return 0;
	return 1;
}

void l4FormPptpBasic(uint32 flag, int8 * pktBuf, 
					 uint8 Cbit, uint8 Rbit, uint8 Kbit, uint8 Sbit, uint8 sbit,
					 uint8 recur, uint8	Abit, uint8	otherflags, uint8 ver) {
	uint8	pptpflagsx1,pptpflagsx2;
	int8 * tmpPtr = (int8 *)(pktBuf + l4Offset(flag, pktBuf));
	greHdr_t * pptpPtr = (greHdr_t *)tmpPtr;

	if(!l4FlagValidPptp(flag))
		return;

	memset(tmpPtr, 0, sizeof(greHdr_t));

	pptpflagsx1 = 0x00;

	if(Cbit)
		pptpflagsx1 |= GRE_C_BIT;
	if(Rbit)
		pptpflagsx1 |= GRE_R_BIT;
	if(Kbit)
		pptpflagsx1 |= GRE_K_BIT;
	if(Sbit)
		pptpflagsx1 |= GRE_S_BIT;
	if(sbit)
		pptpflagsx1 |= GRE_s_BIT;
	
	pptpflagsx1 |= (recur & ~GRE_FLAGS_x1);

	pptpPtr->gre_flags_x1 = pptpflagsx1;

	pptpflagsx2 = 0x00;

	if(Abit)
		pptpflagsx2 |= 0x80;

	pptpflagsx2 |= ((otherflags & 0x0f)<<3);
	pptpflagsx2 |= (ver & 0x07);
	pptpPtr->gre_flags_x2 = pptpflagsx2;
	pptpPtr->gre_protocol = htons(0x880b);

}

void l4FormPptpKey (uint32 flag, int8 * pktBuf, 
					uint16 callid, uint32 seqno, uint32 ackno) {
	int8 * tmpPtr = (int8 *)(pktBuf + l4Offset(flag, pktBuf));
	greHdr_t * pptpPtr = (greHdr_t *)tmpPtr;

	if(!l4FlagValidPptp(flag))
		return;

	pptpPtr->gre_callid = htons(callid);
	pptpPtr->gre_seqno = htonl(seqno);
	pptpPtr->gre_ackno = htonl(ackno);
}

void l4FormPptpOther (uint32 flag, int8 * pktBuf, int8 * content, uint32 size) {
	int8 * cntPtr = (int8 *)(pktBuf + l4Offset(flag, pktBuf) + sizeof(greHdr_t));
	int8 * tmpPtr = (int8 *)(pktBuf + l4Offset(flag, pktBuf));
	greHdr_t * pptpPtr = (greHdr_t *)tmpPtr;

	if(!l4FlagValidPptp(flag))
		return;

	pptpPtr->gre_length = htons(size);

	memcpy(cntPtr, content, size);
	l3FormIpProtocol(flag, pktBuf, IP_PROTO_PPTP);
	l3FormIpTotalLen(flag, pktBuf, size + sizeof(greHdr_t));
}


static int32 l4FlagValidIcmp(uint32 flag) {
	if(!l3FlagValidIp(flag) || 
		bitCount(flag & L4_MASK) > 1 || (flag & L4_ICMP) != L4_ICMP)
		return 0;
	return 1;
}

void l4FormIcmpBasic(uint32 flag, int8 * pktBuf, uint8 type, uint8 code) {
	int8 * tmpPtr = pktBuf + l4Offset(flag, pktBuf);
	icmpHdr_t * icmpPtr = (icmpHdr_t *)tmpPtr;

	if(!l4FlagValidIcmp(flag))
		return;

	memset(tmpPtr, 0, sizeof(icmpHdr_t));
	icmpPtr->icmp_type = type;
	icmpPtr->icmp_code = code;
	l3FormIpProtocol(flag, pktBuf, IP_PROTO_ICMP);
	l3FormIpTotalLen(flag, pktBuf, sizeof(icmpHdr_t));
	l4FormIcmpChecksum(flag, pktBuf);
}

void l4FormIcmpEcho(uint32 flag, int8 * pktBuf, uint16 id, uint16 seq) {
	int8 * tmpPtr = pktBuf + l4Offset(flag, pktBuf);
	icmpHdr_t * icmpPtr = (icmpHdr_t *)tmpPtr;

	if(!l4FlagValidIcmp(flag))
		return;

	icmpPtr->content.echo_s.icd_id = htons(id);
	icmpPtr->content.echo_s.icd_seq = htons(seq);
	l3FormIpTotalLen(flag, pktBuf, sizeof(icmpHdr_t));
	l4FormIcmpChecksum(flag, pktBuf);
}

void l4FormIcmpData(uint32 flag, int8 * pktBuf, int8 * data, uint32 size) {
	int8 * tmpPtr = (int8 *)(pktBuf + l4Offset(flag, pktBuf) + sizeof(icmpHdr_t));
//	int8 * tmpPtr = (int8 *)(pktBuf + l4Offset(flag, pktBuf));
	icmpHdr_t * icmpPtr = (icmpHdr_t *)(pktBuf + l4Offset(flag, pktBuf));

	if(!l4FlagValidIcmp(flag))
		return;

	memcpy(tmpPtr, data, size);

	l3FormIpTotalLen(flag, pktBuf, size+sizeof(icmpHdr_t));
	l4FormIcmpChecksum(flag, pktBuf);
}

void l4FormIcmpChecksum(uint32 flag, int8 * pktBuf) {
	ipHdr_t * ipPtr = (ipHdr_t *)(pktBuf + l3Offset(flag));
	icmpHdr_t * icmpPtr = (icmpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));
	uint16 cksum;

	if(!l4FlagValidIcmp(flag))
		return;

	icmpPtr->icmp_cksum = 0;
	cksum = ipcsum((uint16*)icmpPtr,  ntohs(ipPtr->ip_len) - ipPtr->ip_hl*4, 0);
	
	if(flag & L4_CKSUM_ERR)
		icmpPtr->icmp_cksum = htons(cksum);
	else
		icmpPtr->icmp_cksum = htons(~cksum);
}

static int32 l4FlagValidIgmp(uint32 flag) {
	if(!l3FlagValidIp(flag) || 
		bitCount(flag & L4_MASK) > 1 || (flag & L4_IGMP) != L4_IGMP)
		return 0;
	return 1;
}

void l4FormIgmpBasic(uint32 flag, int8 * pktBuf, uint8 type, uint8 respTime, uint32 addr) {
	ipHdr_t * ipPtr = (ipHdr_t *)(pktBuf + l3Offset(flag));
	igmpHdr_t * igmpPtr = (igmpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));

	if(!l4FlagValidIgmp(flag))
		return;

	l3FormIpProtocol(flag, pktBuf, IP_PROTO_IGMP);
	igmpPtr->igmp_type = type;
	igmpPtr->igmp_respTime = respTime;
	igmpPtr->igmp_cksum = 0;
	igmpPtr->igmp_group = htonl(addr);
	l3FormIpTotalLen(flag, pktBuf, sizeof(igmpHdr_t));
	l4FormIgmpChecksum(flag, pktBuf);
}

void l4FormIgmpContent(uint32 flag, int8 * pktBuf, int8 * content, uint32 size) {
	int8 * cntPtr = (int8 *) (pktBuf + l3Offset(flag) + sizeof(igmpHdr_t));
	
	if(!l4FlagValidIgmp(flag))
		return;

	memcpy(cntPtr, content, size);
	l3FormIpTotalLen(flag, pktBuf, size + sizeof(igmpHdr_t));
	l4FormIgmpChecksum(flag, pktBuf);
}

void l4FormIgmpChecksum(uint32 flag, int8 * pktBuf) {
	ipHdr_t * ipPtr = (ipHdr_t *)(pktBuf + l3Offset(flag));
	igmpHdr_t * igmpPtr = (igmpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));
	uint16 cksum;

	if(!l4FlagValidIgmp(flag))
		return;

	igmpPtr->igmp_cksum = 0;

	cksum = ipcsum((uint16*)igmpPtr, ntohs(ipPtr->ip_len) - ipPtr->ip_hl*4, 0);

	if(flag & L4_CKSUM_ERR)
		igmpPtr->igmp_cksum = htons(cksum);
	else
		igmpPtr->igmp_cksum = htons(~cksum);
}

static int32 l4FlagValidUdp(uint32 flag) {
	if(!l3FlagValidIp(flag) || 
		bitCount(flag & L4_MASK) > 1 || (flag & L4_UDP) != L4_UDP)
		return 0;
	return 1;
}

void l4FormUdpBasic(uint32 flag, int8 * pktBuf, uint16 sPort, uint16 dPort) {
	udpHdr_t * udpPtr = (udpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));

	if(!l4FlagValidUdp(flag))
		return;
	
	udpPtr->uh_sport = htons(sPort);
	udpPtr->uh_dport = htons(dPort);
	l3FormIpProtocol(flag, pktBuf, IP_PROTO_UDP);
	l3FormIpTotalLen(flag, pktBuf, sizeof(udpHdr_t));
	l4FormUdpChecksum(flag, pktBuf);
}

void l4FormUdpContent(uint32 flag, int8 * pktBuf, int8 * content, uint32 size) {
	int8 * cntPtr = pktBuf + l4Offset(flag, pktBuf) + sizeof(udpHdr_t);
	udpHdr_t * udpPtr = (udpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));

	if(!l4FlagValidUdp(flag))
		return;
	
	memcpy(cntPtr, content, size);
	udpPtr->uh_ulen = htons(size+sizeof(udpHdr_t));
	l3FormIpTotalLen(flag, pktBuf, size+sizeof(udpHdr_t));
	l4FormUdpChecksum(flag, pktBuf);
}

void l4FormUdpChecksum(uint32 flag, int8 * pktBuf) {
	ipHdr_t * ipPtr = (ipHdr_t *)(pktBuf + l3Offset(flag));
	udpHdr_t * udpPtr = (udpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));
	uint16 tmp, csum;

	if(!l4FlagValidUdp(flag))
		return;

	csum = ipcsum((uint16 *)&ipPtr->ip_src, 8, 0);
	tmp = htons(IP_PROTO_UDP);
	csum = ipcsum(&tmp, 2, csum);
	csum = ipcsum((uint16 *)&udpPtr->uh_ulen, 2, csum);
	udpPtr->uh_sum = 0;
	csum = ipcsum((uint16 *)udpPtr, ntohs(udpPtr->uh_ulen), csum);
	if(flag & L4_NO_CKSUM)
		udpPtr->uh_sum = 0;
	else if(flag & L4_CKSUM_ERR)
		udpPtr->uh_sum = htons(csum);
	else
		udpPtr->uh_sum = htons(~csum);
}

static int32 l4FlagValidTcp(uint32 flag) {
	if(!l3FlagValidIp(flag) || 
		bitCount(flag & L4_MASK) > 1 || (flag & L4_TCP) != L4_TCP)
		return 0;
	return 1;
}

void l4FormTcpBasic(uint32 flag, int8 * pktBuf, uint16 sPort, uint16 dPort, uint8 hlen) {
	tcpHdr_t * tcpPtr = (tcpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));
	uint32 tcpHlen;

	if(!l4FlagValidTcp(flag))
		return;
	
	tcpPtr->th_sport = htons(sPort);
	tcpPtr->th_dport = htons(dPort);


	if(hlen%4)
	//if(MOD(hlen,4))
		tcpHlen = hlen/4 + 1;
	else
		tcpHlen = hlen/4;

	//memset(tcpPtr,tcpHlen*4, 0);
	memset(tcpPtr,tcpHlen<<2, 0);
	tcpPtr->th_off = tcpHlen;

	l3FormIpProtocol(flag, pktBuf, IP_PROTO_TCP);
	l3FormIpTotalLen(flag, pktBuf, sizeof(tcpHdr_t));
	l4FormTcpChecksum(flag, pktBuf);
}
void l4FormTcpFlow(uint32 flag, int8 * pktBuf, uint32 seq, uint32 ack, uint16 winSize, uint8 tcpFlag, uint16 urgPtr) {
	tcpHdr_t * tcpPtr = (tcpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));

	if(!l4FlagValidTcp(flag))
		return;

	tcpPtr->th_seq = htonl(seq);
	tcpPtr->th_ack = htonl(ack);
	tcpPtr->th_win = htons(winSize);
	tcpPtr->th_flags = tcpFlag;
	tcpPtr->th_urp = htons(urgPtr);
	l3FormIpTotalLen(flag, pktBuf, sizeof(tcpHdr_t));
	l4FormTcpChecksum(flag, pktBuf);
}

void l4FormTcpOption(uint32 flag, int8 * pktBuf, int8 * option, uint8 size) {
	tcpHdr_t * tcpPtr = (tcpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));
	int8 * optPtr = pktBuf + l4Offset(flag, pktBuf) + sizeof(tcpHdr_t);

	if(!l4FlagValidTcp(flag))
		return;
	
	if(size + sizeof(tcpHdr_t) > tcpPtr->th_off*4)
		return;
	
	memcpy(optPtr, option, size);
	l3FormIpTotalLen(flag, pktBuf, sizeof(tcpHdr_t));
	l4FormTcpChecksum(flag, pktBuf);
}

void l4FormTcpContent(uint32 flag, int8 * pktBuf, int8 * content, uint32 size) {
	tcpHdr_t * tcpPtr = (tcpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));
	int8 * cntPtr = pktBuf + l4Offset(flag, pktBuf) + tcpPtr->th_off*4;

	if(!l4FlagValidTcp(flag))
		return;

	memcpy(cntPtr, content, size);
	l3FormIpTotalLen(flag, pktBuf, size+tcpPtr->th_off*4);
	l4FormTcpChecksum(flag, pktBuf);
}

void l4FormTcpChecksum(uint32 flag, int8 * pktBuf) {
	ipHdr_t * ipPtr = (ipHdr_t *)(pktBuf + l3Offset(flag));
	tcpHdr_t * tcpPtr = (tcpHdr_t*)(pktBuf + l4Offset(flag, pktBuf));
	uint16 tmp, csum;

	if(!l4FlagValidTcp(flag))
		return;

	csum = ipcsum((uint16 *)&ipPtr->ip_src, 8, 0);
	tmp = htons(IP_PROTO_TCP);
	csum = ipcsum(&tmp, 2, csum);
	tmp = htons(l4PayloadLen(flag, pktBuf));
	csum = ipcsum(&tmp, 2, csum);
	tcpPtr->th_sum = 0;
	csum = ipcsum((uint16 *)tcpPtr, l4PayloadLen(flag, pktBuf), csum);
	if(flag & L4_CKSUM_ERR) 
		tcpPtr->th_sum = htons(csum);
	else
		tcpPtr->th_sum = htons(~csum);
}

uint32 formedPktLen(uint32 flag, int8 * pktBuf) {
	if(flag & L4_MASK)
		return l4PayloadLen(flag, pktBuf) + l4Offset(flag, pktBuf);

	return l4Offset(flag, pktBuf);
}

uint32 formedFrameLen(uint32 flag, int8 * pktBuf, uint32 payloadLen) {

	if((flag & L3_IPX) == L3_IPX)
		return l3Offset(flag) + sizeof(ipxHdr_t) + payloadLen;

	if((flag & L3_ARP) == L3_ARP)
		return l3Offset(flag) + sizeof(arpHdr_t) + payloadLen;

	return l3Offset(flag) + payloadLen;

}

/* end of pg_pktForm.c */
