/*
* ----------------------------------------------------------------
* Copyright c              Realtek Semiconductor Corporation, 2002
* All rights reserved.
* 
* $Header: /home/cvsroot/SBU_III_2/wpnt834/ntg/src_from_v1036/uClinux-dist/linux-2.4.x/drivers/net/re865x/rtl865x/veri/pktgen/pg_pktParse.c,v 1.1.1.1 2006/07/26 05:38:30 blue Exp $
*
* Abstract: packet generation module. packet parse.
*
* $Author: blue $
*
* $Log: pg_pktParse.c,v $
* Revision 1.1.1.1  2006/07/26 05:38:30  blue
*
*
* Revision 1.1.1.1  2006/02/14 03:09:26  jason
* backup the source code from v1036
*
* Revision 1.1.1.1  2006/01/16 02:40:11  jason
* backup source,because now we directly use romedriver source code, i open a new directory;
*
* Revision 1.1.1.1  2005/11/21 05:20:04  chenyl
* +: WPNT834
* 	- System provider	: Sercomm
* 	- Brand Name		: Netgear
* 	- 3rt Party		: Airgo Gen 3
* 	- SDK			: 066-3
* 	- ROMEDRV		: 3.6-3
*
*
* Revision 1.1  2004/03/26 06:31:34  danwu
* init
*
* Revision 1.6  2004/01/08 14:56:33  waynelee
* debug: parsing IPX SAP
*
* Revision 1.5  2003/12/30 10:34:42  waynelee
* add pptp, ipx support and provide more flexible packet format
*
* Revision 1.4  2003/12/19 02:23:49  danwu
* make pktParse routine public
*
* Revision 1.3  2003/04/30 15:46:08  cfliu
* remove macro.h, use macros defined in types.h instead
*
* Revision 1.2  2002/08/28 08:43:38  orlando
* Add pg_pktDump().
*
* Revision 1.1  2002/08/28 03:44:07  orlando
* packet generation module initial check-in.
*
* ---------------------------------------------------------------
*/

#include "../../types.h"
#include "pktgen.h" // exported prototype declaration

// prototype declaration
//static void pktParse(int8 * pkt, pktParse_t * result);

void pg_pktParse(int8 * pktBuf, uint32 pktLen)
{
	pktParse_t result;

	//printf("Packet Length %d :\n", pktLen);
	pktParse(pktBuf, &result, pktLen);
	printf("\nBegin parse-----------------\n");
	dumpEther(result.ether2);
	dumpVlan(result.vlan);
	dumpLlc(result.llc);
	dumpPppoe(result.pppoe);
	dumpIpx(result.ipx);
	dumpArp(result.arp);
	dumpIp(result.ip, result.l3Flag, result.l3Cksum);
	dumpPptp(result.pptp);
	dumpIcmp(result.icmp, result.l4Flag, result.l4Cksum);
	dumpIgmp(result.igmp, result.l4Flag, result.l4Cksum);
	dumpTcp(result.tcp, result.l4Flag, result.l4Cksum);
	dumpUdp(result.udp, result.l4Flag, result.l4Cksum);
	printf("\nEnd parse-------------------\n");	
	
} /* end pktParseMain */


/*
 * local routine
 */
/*static*/ void pktParse(int8 * pkt, pktParse_t * result, uint32 totalLen) {
	uint32	type, len;
	uint16 tmp, csum;
	int8 * nextPtr;

	memset(result, 0, sizeof(pktParse_t));

	// total length
	result->totalLen = totalLen;

	//Parse layer2 headers
	result->ether2 = (etherHdr_t *)pkt;
	nextPtr = pkt + sizeof(etherHdr_t);
	type = ntohs(result->ether2->etherType);

	if(ntohs(result->ether2->etherType) == 0x8100) {
		result->vlan = (vlanHdr_t *)nextPtr;
		result->pktFlag |= L2_VLAN;
		nextPtr += sizeof(vlanHdr_t);
		type = ntohs(result->vlan->etherType);
	}

	if(type < 0x0600) {//802.3-2000 p.79
		result->llc = (llcHdr_t *)nextPtr;
		result->pktFlag |= L2_SNAP;
		if((result->llc->llc_dsap == 0xe0) && (result->llc->llc_ssap == 0xe0)) {
			nextPtr += (sizeof(llcHdr_t)-5);
			type = 0x8137;
		} else {
			nextPtr += sizeof(llcHdr_t);
			type = ntohs(result->llc->etherType);
		}
	}

	if(type == 0x8864 || type == 0x8863 ) {
		result->pppoe = (pppoeHdr_t *)nextPtr;
		result->pktFlag |= L2_PPPoE;
		nextPtr += sizeof(pppoeHdr_t);
		type = ntohs(result->pppoe->proto);
	}

	//Parse layer3 headers, IP, ARP
	if( ((result->pktFlag & L2_PPPoE) == L2_PPPoE && type == 0x0021) ||
		((result->pktFlag & L2_PPPoE) == NULL && 
		  (((result->pktFlag & L2_SNAP) == L2_SNAP && result->llc->llc_dsap == 0xAA &&	result->llc->llc_ssap == 0xAA && result->llc->ctrl == 0x03) || 
		   (result->pktFlag & L2_SNAP) ==NULL) &&
			type == 0x0800) ) {
		result->ip = (ipHdr_t *)nextPtr;
		result->pktFlag = 0;
		result->pktFlag |= L3_IP;
		nextPtr += result->ip->ip_hl*4;
		type = result->ip->ip_p;
		len = ntohs(result->ip->ip_len) - result->ip->ip_hl*4;
		// store ip data length to dataLen and ipDataLen
		result->ipDataLen = len;
		result->dataLen = len;
		// ip payload
		result->ipPayload = (uint8 *)nextPtr;
		result->payload = (uint8 *)nextPtr;
		if(ipcsum((uint16 *)result->ip, result->ip->ip_hl*4, 0) == 0)
			result->l3Flag |= PKT_PARSE_L3_CKSUM_OK;
		else {
			uint16 orgCksum = result->ip->ip_sum;
			
			// printf("Find l3 header error\n");
			result->l3Flag |= PKT_PARSE_L3_CKSUM_ERR;
			result->ip->ip_sum = 0;
			result->l3Cksum = ~ipcsum((uint16 *)result->ip, result->ip->ip_hl*4, 0);
			result->ip->ip_sum = orgCksum;
		}
	}
	else if  ((result->pktFlag & L2_PPPoE) == NULL && (type == 0x0806 || type == 0x8035)) {
		result->arp = (arpHdr_t *)nextPtr;
		result->pktFlag |= L3_ARP;
		nextPtr = NULL;//Arp do not parse futher
		return;
	}
	else if  (type == 0x8137) {
		result->ipx = (ipxHdr_t *)nextPtr;
		nextPtr = NULL;
		return;
	}
	else {
		result->payload = (uint8 *)nextPtr;
		// caculate the data len
		result->dataLen = totalLen - sizeof(etherHdr_t);
		if((result->pktFlag & L2_VLAN) == L2_VLAN)
			result->dataLen -= sizeof(vlanHdr_t);
		if((result->pktFlag & L2_SNAP) == L2_SNAP)
			result->dataLen -= sizeof(llcHdr_t);
		if((result->pktFlag & L2_PPPoE) == L2_PPPoE)
			result->dataLen -= sizeof(pppoeHdr_t);
		nextPtr = NULL;
		return;
	}
	//Parse layer4 headers
	if(((result->pktFlag & L3_IP) == L3_IP) && ((uint16)(ntohs(result->ip->ip_off)&0x1fff) ==0)) {
		uint16 orgCksum, phCksum;

		result->l4Flag = 0;
		switch(type) {
		case IP_PROTO_PPTP:
			result->pptp = (greHdr_t *) nextPtr;
			nextPtr += sizeof(greHdr_t);
			result->payload = (uint8 *)nextPtr;
			result->dataLen -= sizeof(greHdr_t);
			result->pktFlag |= L4_PPTP;
			break;
		case IP_PROTO_ICMP:
			result->icmp = (icmpHdr_t *) nextPtr;
			if(ipcsum((uint16 *)result->icmp, len, 0)==0)
				result->l4Flag |= PKT_PARSE_L4_CKSUM_OK;
			else {
				orgCksum = result->icmp->icmp_cksum;

				result->l4Flag |= PKT_PARSE_L4_CKSUM_ERR;
				result->icmp->icmp_cksum = 0;
				result->l4Cksum = ~ipcsum((uint16 *)result->icmp, len, 0);
				result->icmp->icmp_cksum = orgCksum;
			}
			nextPtr += sizeof(icmpHdr_t);
			result->payload = (uint8 *)nextPtr;
			result->dataLen -= sizeof(icmpHdr_t);
			result->pktFlag |= L4_ICMP;
			break;
		case IP_PROTO_IGMP:
			result->igmp = (igmpHdr_t *) nextPtr;
			if(ipcsum((uint16 *)result->igmp, len, 0)==0)
				result->l4Flag |= PKT_PARSE_L4_CKSUM_OK;
			else {
				orgCksum = result->igmp->igmp_cksum;
				result->l4Flag |= PKT_PARSE_L4_CKSUM_ERR;
				result->igmp->igmp_cksum = 0;
				result->l4Cksum = ~ipcsum((uint16 *)result->igmp, len, 0);
				result->igmp->igmp_cksum = orgCksum;
			}
			nextPtr += sizeof(igmpHdr_t);
//			result->payload = (uint8 *)nextPtr;
//			result->dataLen -= sizeof(igmpHdr_t);
			result->pktFlag |= L4_IGMP;
			break;
		case IP_PROTO_TCP:
			result->tcp = (tcpHdr_t *) nextPtr;
			phCksum = ipcsum((uint16 *)&result->ip->ip_src, 8, 0);
			tmp = htons(IP_PROTO_TCP);
			phCksum = ipcsum(&tmp, 2, phCksum);
			tmp = len;
			tmp = htons(tmp);
			phCksum = ipcsum(&tmp, 2, phCksum);
			csum = ipcsum((uint16 *)result->tcp, len, phCksum);
			if(csum == 0)
				result->l4Flag |= PKT_PARSE_L4_CKSUM_OK;
			else {
				orgCksum = result->tcp->th_sum;
				result->l4Flag |= PKT_PARSE_L4_CKSUM_ERR;
				result->tcp->th_sum = 0;
				result->l4Cksum = ~ipcsum((uint16 *)result->tcp, len, phCksum);
				result->tcp->th_sum = orgCksum;
			}
			// nextPtr += (sizeof(tcpHdr_t) + result->tcp->th_off*4);
			nextPtr += result->tcp->th_off*4;
			result->pktFlag |= L4_TCP;
			result->dataLen -= result->tcp->th_off*4;
			result->payload = (uint8 *)nextPtr;
			break;
		case IP_PROTO_UDP:
			result->udp = (udpHdr_t *) nextPtr;
			if((uint16)len != ntohs(result->udp->uh_ulen))
				result->l4Flag |= PKT_PARSE_UDP_LEN_ERR;
			phCksum = ipcsum((uint16 *)&result->ip->ip_src, 8, 0);
			tmp = htons(IP_PROTO_UDP);
			phCksum = ipcsum(&tmp, 2, phCksum);
			phCksum = ipcsum((uint16 *)&result->udp->uh_ulen, 2, phCksum);
			csum = ipcsum((uint16 *)result->udp, len, phCksum);
			if(result->udp->uh_sum == 0)
				result->l4Flag |= PKT_PARSE_UDP_NO_CKSUM;
			else if(csum == 0)
				result->l4Flag |= PKT_PARSE_L4_CKSUM_OK;
			else {
				orgCksum = result->udp->uh_sum;
				result->l4Flag |= PKT_PARSE_L4_CKSUM_ERR;
				result->udp->uh_sum = 0;
				result->l4Cksum = ~ipcsum((uint16 *)result->udp, len, phCksum);
				result->udp->uh_sum = orgCksum;
			}
			nextPtr += sizeof(udpHdr_t);
			result->pktFlag |= L4_UDP;
			result->dataLen -= sizeof(udpHdr_t);
			result->payload = (uint8 *)nextPtr;
			break;
		}
	}
}


/* end of pg_pktParse.c */
