#include <linux/config.h>
#include "../../rtl_types.h"
#include "../../asicRegs.h"
#include "../../hsRegs.h"
#include "../../rtl8651_tblAsicDrv.h"
#include "../veri.h"


/* LOCAL DEFINITIONS 
*/
/* gateway */
#define WAN0_VID 8
#define GW_WAN0_MAC "00-00-00-00-88-88"
#define WAN1_VID 9
#define GW_WAN1_MAC "00-00-00-00-88-89"
#define GW_WAN1_IP "192.168.3.1"
#define GW_WAN1_SUBNET_MASK "255.255.255.248"
#define LAN0_VID 10
#define GW_LAN0_MAC "00-00-00-00-88-8a"
#define GW_LAN0_IP "192.168.2.1"
#define GW_LAN0_SUBNET_MASK "255.255.255.248"
#define LAN1_VID 11
#define GW_LAN1_MAC "00-00-00-00-88-8b"
#define GW_LAN1_IP "192.168.1.1"
#define GW_LAN1_SUBNET_MASK "255.255.255.248"
/* isp 0 */
#define WAN0_ISP_MAC "00-00-00-00-99-9c"
#define WAN0_ISP_PPPOE_SID 0x55
#define WAN0_ISP_PUBLIC_IP "192.168.4.1"
/* isp 1 */
#define WAN1_ISP_PUBLIC_IP "192.168.3.1"
/* wan 0 hosts */
#define HOST5_MAC "00-00-00-00-00-10"
#define HOST5_IP "192.168.4.5"
/* wan 1 hosts */
#define HOST4_MAC "00-00-00-00-00-20"
#define HOST4_IP "192.168.3.2"
/* lan 0 hosts */
#define HOST3_MAC "00-00-00-00-00-30"
#define HOST3_IP "192.168.2.2"
/* lan 1 hosts */
#define HOST1_MAC "00-00-00-00-00-40"
#define HOST1_IP "192.168.1.2"
#define HOST2_MAC "00-00-00-00-00-50"
#define HOST2_IP "192.168.1.3"
#define HOST6_MAC "00-00-00-00-00-60"
#define HOST6_IP "192.168.1.4"


/* LOCAL SUBROUTINES 
*/
static uint8 hp[MAX_PORT_NUMBER + 1];


void l34_tplg(uint8 *hostPortNumArray)
{
/* Locally declaring parameter structures overflows the stack */
extern rtl865x_tblAsicDrv_vlanParam_t vlant;
extern _rtl8651_tblDrvAclRule_t aclt;
extern rtl865x_tblAsicDrv_l2Param_t l2t;
extern rtl865x_tblAsicDrv_pppoeParam_t pppoet;
extern rtl865x_tblAsicDrv_arpParam_t arpt;
extern rtl865x_tblAsicDrv_routingParam_t routet;
extern rtl865x_tblAsicDrv_extIntIpParam_t ipt;
extern rtl865x_tblAsicDrv_serverPortParam_t srvt;
extern rtl865x_tblAsicDrv_algParam_t algt;
    ipaddr_t ip32;
    uint32 i;
    uint32 pvcr;
    
    memcpy(&hp[1], hostPortNumArray, MAX_PORT_NUMBER);
    VERI_MSG(("Topology: H1 H2 H3 H4 H5 H6\n         "));
    for (i=1; i<=MAX_PORT_NUMBER; i++)
         VERI_MSG(("  %d", hp[i]));
    VERI_MSG(("\n"));
    
    /* switch core misc control reg: enable L4 hash, multicast MTU = 0x6ff */
    WRITE_MEM32(MISCCR, 0x16ff);
    /* module switch control reg: enable L2/L3/L4/STP, trap 802.1D/GARP to CPU */
    rtl8651_setAsicOperationLayer(4);
    /* TTL control reg: enable TTL-1 operation */
    WRITE_MEM32(TTLCR, EN_TTL1);
    /* VLAN tagging control reg: accept untagged */
    WRITE_MEM32(VLANTCR, 0);
    /* l4 table aging control reg */
    WRITE_MEM32(TEATCR, 0xffffffff);
    /* l4 table offset control reg: start = 20, end = 20 */
    WRITE_MEM32(OCR, (20<<OCR_START_OFFSET)|(20<<OCR_END_OFFSET));
    /* switch table misc control reg: enable napt auto learn/delete, enable VLAN ingress filtering, 
      trap non-TCP/UDP/ICMP NAPT packets to CPU, enable multicast table, accept broadcast packets, 
      enable PPPoE auto-encapsulation and auto-decapsulation */
    WRITE_MEM32(SWTMCR, 0x7600001e);
    /* gidx mapping control */
    WRITE_MEM32(GIDXMCR, 0);
    /* disable 50b mode */
    WRITE_MEM32(SWTECR, 0);
    
    /* port based vlan config reg */
    pvcr = 0 << (hp[5] * 3); //host 5 pvid 0
    pvcr |= 1 << (hp[4] * 3); //host 4 pvid 1
    pvcr |= 2 << (hp[3] * 3); //host 3 pvid 2
    pvcr |= 3 << (hp[2] * 3); //host 2 pvid 3
    pvcr |= 3 << (hp[1] * 3); //host 1 pvid 3
    pvcr |= 3 << (hp[6] * 3); //host 6 pvid 3
    WRITE_MEM32(PVCR, pvcr);

    rtl8651_clearAsicAllTable();
    
    /* config vlan table */
    bzero((void*) &vlant, sizeof(vlant));
    bzero((void*) &aclt, sizeof(aclt));
    /* wan 0 */
    strtomac(&vlant.macAddr, GW_WAN0_MAC);
    vlant.memberPortMask = vlant.untagPortMask = (1<<hp[5]);
    vlant.macAddrNumber = 1;
    vlant.vid = WAN0_VID;
    for(i=0;i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum;i++)
        vlant.portState[i] = RTL8651_PORTSTA_FORWARDING;
    vlant.mtu = 1522;
    vlant.inAclStart = 0;
    vlant.inAclEnd = 2;
    vlant.outAclStart = vlant.outAclEnd = 3;
    vlant.internal = 0;
    vlant.enableRoute = 1;
    vlant.valid = 1;
    rtl8651_setAsicVlan(&vlant);
    /* ingress filter gateway mac */
    aclt.ruleType_ = RTL8651_ACL_MAC;
    memcpy(aclt.srcMac_.octet, vlant.macAddr.octet, 6);
    memset(aclt.srcMacMask_.octet, 0xff, 6);
    aclt.actionType_ = RTL8651_ACL_CPU;
    rtl8651_setAsicAclRule(0, &aclt);
    /* wan 1 */
    strtomac(&vlant.macAddr, GW_WAN1_MAC);
    vlant.memberPortMask = vlant.untagPortMask = (1<<hp[4]);
    vlant.vid = WAN1_VID;
    vlant.inAclStart = 4;
    vlant.inAclEnd = 6;
    vlant.outAclStart = vlant.outAclEnd = 7;
    vlant.internal = 0;
    rtl8651_setAsicVlan(&vlant);
    /* ingress filter gateway mac */
    aclt.ruleType_ = RTL8651_ACL_MAC;
    memcpy(aclt.srcMac_.octet, vlant.macAddr.octet, 6);
    aclt.actionType_ = RTL8651_ACL_CPU;
    rtl8651_setAsicAclRule(4, &aclt);
    /* lan 0 */
    strtomac(&vlant.macAddr, GW_LAN0_MAC);
    vlant.memberPortMask = (1<<hp[3]);
    vlant.untagPortMask = 0; //tagged
    vlant.vid = LAN0_VID;
    vlant.inAclStart = 8;
    vlant.inAclEnd = 10;
    vlant.outAclStart = vlant.outAclEnd = 11;
    vlant.internal = 1;
    rtl8651_setAsicVlan(&vlant);
    /* ingress filter gateway mac */
    aclt.ruleType_ = RTL8651_ACL_MAC;
    memcpy(aclt.srcMac_.octet, vlant.macAddr.octet, 6);
    aclt.actionType_ = RTL8651_ACL_CPU;
    rtl8651_setAsicAclRule(8, &aclt);
    /* lan 1 */
    strtomac(&vlant.macAddr, GW_LAN1_MAC);
    vlant.memberPortMask = vlant.untagPortMask = (1<<hp[2])|(1<<hp[1])|(1<<hp[6]);
    vlant.vid = LAN1_VID;
    vlant.inAclStart = 12;
    vlant.inAclEnd = 14;
    vlant.outAclStart = vlant.outAclEnd = 15;
    vlant.internal = 1;
    rtl8651_setAsicVlan(&vlant);
    /* ingress filter gateway mac */
    aclt.ruleType_ = RTL8651_ACL_MAC;
    memcpy(aclt.srcMac_.octet, vlant.macAddr.octet, 6);
    aclt.actionType_ = RTL8651_ACL_CPU;
    rtl8651_setAsicAclRule(12, &aclt);
    
    /* config l2 table */
    bzero((void*) &l2t, sizeof(l2t));
    /* wan 0 isp modem */
    strtomac(&l2t.macAddr, WAN0_ISP_MAC);
    l2t.memberPortMask = (1<<hp[5]);
    l2t.isStatic = 1;
    rtl8651_setAsicL2Table(getL2TableRowFromMacStr(WAN0_ISP_MAC), 0, &l2t);
    /* wan 1 host */
    strtomac(&l2t.macAddr, HOST4_MAC);
    l2t.memberPortMask = (1<<hp[4]);
    rtl8651_setAsicL2Table(getL2TableRowFromMacStr(HOST4_MAC), 0, &l2t);
    /* lan 0 host */
    strtomac(&l2t.macAddr, HOST3_MAC);
    l2t.memberPortMask = (1<<hp[3]);
    rtl8651_setAsicL2Table(getL2TableRowFromMacStr(HOST3_MAC), 0, &l2t);
    /* lan 1 host 1 */
    strtomac(&l2t.macAddr, HOST2_MAC);
    l2t.memberPortMask = (1<<hp[2]);
    rtl8651_setAsicL2Table(getL2TableRowFromMacStr(HOST2_MAC), 0, &l2t);
    /* lan 1 host 0 */
    strtomac(&l2t.macAddr, HOST1_MAC);
    l2t.memberPortMask = (1<<hp[1]);
    rtl8651_setAsicL2Table(getL2TableRowFromMacStr(HOST1_MAC), 0, &l2t);
    /* lan 1 host 5 */
    strtomac(&l2t.macAddr, HOST6_MAC);
    l2t.memberPortMask = (1<<hp[6]);
    rtl8651_setAsicL2Table(getL2TableRowFromMacStr(HOST6_MAC), 0, &l2t);
    
    /* config pppoe table */
    bzero((void*) &pppoet, sizeof(pppoet));
    pppoet.sessionId = WAN0_ISP_PPPOE_SID;
    rtl8651_setAsicPppoe(0, &pppoet);
    
    /* config arp table */
    bzero((void*) &arpt, sizeof(arpt));
    /* wan 1 host */
    arpt.nextHopRow = getL2TableRowFromMacStr(HOST4_MAC);
    strtoip(&ip32, HOST4_IP);
    rtl8651_setAsicArp(ip32 & 0x7, &arpt);
    /* lan 0 host */
    arpt.nextHopRow = getL2TableRowFromMacStr(HOST3_MAC);
    strtoip(&ip32, HOST3_IP);
    rtl8651_setAsicArp(8 + (ip32 & 0x7), &arpt);
    /* lan 1 host 1 */
    arpt.nextHopRow = getL2TableRowFromMacStr(HOST2_MAC);
    strtoip(&ip32, HOST2_IP);
    rtl8651_setAsicArp(16 + (ip32 & 0xff), &arpt);
    rtl8651_setAsicArp(ip32 & 0xff, &arpt);
    /* lan 1 host 0 */
    arpt.nextHopRow = getL2TableRowFromMacStr(HOST1_MAC);
    strtoip(&ip32, HOST1_IP);
    rtl8651_setAsicArp(16 + (ip32 & 0xff), &arpt);
    /* lan 1 host 5 */
    arpt.nextHopRow = getL2TableRowFromMacStr(HOST6_MAC);
    strtoip(&ip32, HOST6_IP);
    rtl8651_setAsicArp(16 + (ip32 & 0xff), &arpt);
    
    /* config routing table */
    bzero((void*) &routet, sizeof(routet));
    /* route to wan 1: use arp table [0:7] */
    strtoip(&routet.ipAddr, GW_WAN1_IP);
    strtoip(&routet.ipMask, GW_WAN1_SUBNET_MASK);
    routet.process = PROCESS_INDIRECT;
    routet.vidx = rtl8651_vlanTableIndex(WAN1_VID);
    routet.arpStart = 0;
    routet.arpEnd = 0;
    rtl8651_setAsicRouting(0, &routet);
    /* route to lan 0: use arp table [8:15] */
    strtoip(&routet.ipAddr, GW_LAN0_IP);
    strtoip(&routet.ipMask, GW_LAN0_SUBNET_MASK);
    routet.process = PROCESS_INDIRECT;
    routet.vidx = rtl8651_vlanTableIndex(LAN0_VID);
    routet.arpStart = 1;
    routet.arpEnd = 1;
    rtl8651_setAsicRouting(1, &routet);
    /* route to lan 1: use arp table [16:23] */
    strtoip(&routet.ipAddr, GW_LAN1_IP);
    strtoip(&routet.ipMask, GW_LAN1_SUBNET_MASK);
    routet.process = PROCESS_INDIRECT;
    routet.vidx = rtl8651_vlanTableIndex(LAN1_VID);
    routet.arpStart = 2;
    routet.arpEnd = 2;
    rtl8651_setAsicRouting(2, &routet);
    /* default route to wan 0: direct to modem */
    bzero((void*) &routet, sizeof(routet));
    strtoip(&routet.ipAddr, "0.0.0.0");
    strtoip(&routet.ipMask, "0.0.0.0");
    routet.process = PROCESS_PPPOE;
    routet.vidx = rtl8651_vlanTableIndex(WAN0_VID);
    routet.nextHopRow = getL2TableRowFromMacStr(WAN0_ISP_MAC);
    routet.pppoeIdx = 0;
    rtl8651_setAsicRouting(7, &routet);
    
    /* config ip table */
    bzero((void*) &ipt, sizeof(ipt));
    /* wan 0 napt */
    strtoip(&ipt.intIpAddr, "0.0.0.0");
    strtoip(&ipt.extIpAddr, WAN0_ISP_PUBLIC_IP);
    rtl8651_setAsicExtIntIpTable(0, &ipt);
    /* wan 1 nat */
    strtoip(&ipt.intIpAddr, HOST1_IP);
    strtoip(&ipt.extIpAddr, WAN1_ISP_PUBLIC_IP);
    ipt.nat = 1;
    rtl8651_setAsicExtIntIpTable(1, &ipt);
    
    /* config server port table */
    bzero((void*) &srvt, sizeof(srvt));
    /* server port 1020 translated to wan 0 isp port 20 */
    strtoip(&srvt.intIpAddr, HOST2_IP);
    strtoip(&srvt.extIpAddr, WAN0_ISP_PUBLIC_IP);
    srvt.extPort = 20;
    srvt.intPort = 1020;
    srvt.valid = 1;
    rtl8651_setAsicServerPortTable(0, &srvt);
    
    /* config alg table */
    bzero((void*) &algt, sizeof(algt));
    algt.port = 80; //trap http
    rtl8651_setAsicAlg(0, &algt);

    return;
}

uint32 l34_testNatThruEth()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    
    VERI_MSG(("NAT through Ethernet ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_LAN1_MAC);
    strtomac(&tcpt.smac, HOST1_MAC);
    strtoip(&tcpt.dip, HOST4_IP);
    strtoip(&tcpt.sip, HOST1_IP);
    tcpt.dport = 200;
    tcpt.sport = 100;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[1], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[4], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST4_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_WAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST4_IP, errCount);
        CHK_PKT_IP(tcpt.sip, WAN1_ISP_PUBLIC_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 200, errCount);
        CHK_PKT_DEC(tcpt.sport, 100, errCount);
    }

    VERI_MSG(("Inbound NAT through Ethernet ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_WAN1_MAC);
    strtomac(&tcpt.smac, HOST4_MAC);
    strtoip(&tcpt.dip, WAN1_ISP_PUBLIC_IP);
    strtoip(&tcpt.sip, HOST4_IP);
    tcpt.dport = 100;
    tcpt.sport = 200;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[4], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[1], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST1_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_LAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST1_IP, errCount);
        CHK_PKT_IP(tcpt.sip, HOST4_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 100, errCount);
        CHK_PKT_DEC(tcpt.sport, 200, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testNatThruPppoe()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    
    VERI_MSG(("NAT through PPPoE ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_LAN1_MAC);
    strtomac(&tcpt.smac, HOST1_MAC);
    strtoip(&tcpt.dip, HOST5_IP);
    strtoip(&tcpt.sip, HOST1_IP);
    tcpt.dport = 200;
    tcpt.sport = 100;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[1], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[5], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;
        }
        CHK_PKT_MAC(tcpt.dmac, WAN0_ISP_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_WAN0_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 1, errCount);
        CHK_PKT_HEX(tcpt.sid, WAN0_ISP_PPPOE_SID, errCount);
        CHK_PKT_IP(tcpt.dip, HOST5_IP, errCount);
        CHK_PKT_IP(tcpt.sip, WAN1_ISP_PUBLIC_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 200, errCount);
        CHK_PKT_DEC(tcpt.sport, 100, errCount);
    }

    VERI_MSG(("Inbound NAT through PPPoE ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_WAN0_MAC);
    strtomac(&tcpt.smac, HOST5_MAC);
    tcpt.pppoeTag = 1;
    tcpt.sid = WAN0_ISP_PPPOE_SID;
    strtoip(&tcpt.dip, WAN1_ISP_PUBLIC_IP);
    strtoip(&tcpt.sip, HOST5_IP);
    tcpt.dport = 100;
    tcpt.sport = 200;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[5], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[1], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST1_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_LAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST1_IP, errCount);
        CHK_PKT_IP(tcpt.sip, HOST5_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 100, errCount);
        CHK_PKT_DEC(tcpt.sport, 200, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testLanRoute()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    
    VERI_MSG(("Inter-LAN routing ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_LAN1_MAC);
    strtomac(&tcpt.smac, HOST2_MAC);
    strtoip(&tcpt.dip, HOST3_IP);
    strtoip(&tcpt.sip, HOST2_IP);
    tcpt.dport = 200;
    tcpt.sport = 100;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[2], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[3], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST3_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_LAN0_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 1, errCount);
        CHK_PKT_DEC(tcpt.vid, LAN0_VID, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST3_IP, errCount);
        CHK_PKT_IP(tcpt.sip, HOST2_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 200, errCount);
        CHK_PKT_DEC(tcpt.sport, 100, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testNaptThruEth()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    uint16 extPort;
    
    VERI_MSG(("NAPT through Ethernet ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_LAN1_MAC);
    strtomac(&tcpt.smac, HOST2_MAC);
    strtoip(&tcpt.dip, HOST4_IP);
    strtoip(&tcpt.sip, HOST2_IP);
    tcpt.dport = 200;
    tcpt.sport = 100;
    tcpt.flag = TCPFLAG_SYN;
    tcpt.padLen = 100;
    extPort = (20<<10) | rtl8651_naptTcpUdpTableIndex(1, tcpt.sip, tcpt.sport, tcpt.dip, tcpt.dport);
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[2], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[4], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST4_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_WAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST4_IP, errCount);
        CHK_PKT_IP(tcpt.sip, WAN0_ISP_PUBLIC_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 200, errCount);
        CHK_PKT_DEC(tcpt.sport, extPort, errCount);
    }

    VERI_MSG(("Inbound NAPT through Ethernet ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_WAN1_MAC);
    strtomac(&tcpt.smac, HOST4_MAC);
    strtoip(&tcpt.dip, WAN0_ISP_PUBLIC_IP);
    strtoip(&tcpt.sip, HOST4_IP);
    tcpt.dport = extPort;
    tcpt.sport = 200;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[4], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[2], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST2_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_LAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST2_IP, errCount);
        CHK_PKT_IP(tcpt.sip, HOST4_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 100, errCount);
        CHK_PKT_DEC(tcpt.sport, 200, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testNaptThruPppoe()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    uint16 extPort;
    
    VERI_MSG(("NAPT through PPPoE ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_LAN1_MAC);
    strtomac(&tcpt.smac, HOST2_MAC);
    strtoip(&tcpt.dip, HOST5_IP);
    strtoip(&tcpt.sip, HOST2_IP);
    tcpt.dport = 200;
    tcpt.sport = 100;
    tcpt.flag = TCPFLAG_SYN;
    tcpt.padLen = 100;
    extPort = (20<<10) | rtl8651_naptTcpUdpTableIndex(1, tcpt.sip, tcpt.sport, tcpt.dip, tcpt.dport);
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[2], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[5], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;
        }
        CHK_PKT_MAC(tcpt.dmac, WAN0_ISP_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_WAN0_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 1, errCount);
        CHK_PKT_HEX(tcpt.sid, WAN0_ISP_PPPOE_SID, errCount);
        CHK_PKT_IP(tcpt.dip, HOST5_IP, errCount);
        CHK_PKT_IP(tcpt.sip, WAN0_ISP_PUBLIC_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 200, errCount);
        CHK_PKT_DEC(tcpt.sport, extPort, errCount);
    }

    VERI_MSG(("Inbound NAPT through PPPoE ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_WAN0_MAC);
    strtomac(&tcpt.smac, HOST5_MAC);
    tcpt.pppoeTag = 1;
    tcpt.sid = WAN0_ISP_PPPOE_SID;
    strtoip(&tcpt.dip, WAN0_ISP_PUBLIC_IP);
    strtoip(&tcpt.sip, HOST5_IP);
    tcpt.dport = extPort;
    tcpt.sport = 200;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[5], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[2], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST2_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_LAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST2_IP, errCount);
        CHK_PKT_IP(tcpt.sip, HOST5_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 100, errCount);
        CHK_PKT_DEC(tcpt.sport, 200, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testLanRouteStripVlan()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    
    VERI_MSG(("Inter-LAN routing, striping VLAN tag ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_LAN0_MAC);
    strtomac(&tcpt.smac, HOST3_MAC);
    tcpt.vlanTag = 1;
    tcpt.vid = LAN0_VID;
    strtoip(&tcpt.dip, HOST1_IP);
    strtoip(&tcpt.sip, HOST3_IP);
    tcpt.dport = 200;
    tcpt.sport = 100;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[3], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[1], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST1_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_LAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST1_IP, errCount);
        CHK_PKT_IP(tcpt.sip, HOST3_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 200, errCount);
        CHK_PKT_DEC(tcpt.sport, 100, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testNaptThruEthStripVlan()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    uint16 extPort;
    
    VERI_MSG(("NAPT through Ethernet, striping VLAN tag ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_LAN0_MAC);
    strtomac(&tcpt.smac, HOST3_MAC);
    tcpt.vlanTag = 1;
    tcpt.vid = LAN0_VID;
    strtoip(&tcpt.dip, HOST4_IP);
    strtoip(&tcpt.sip, HOST3_IP);
    tcpt.dport = 200;
    tcpt.sport = 100;
    tcpt.flag = TCPFLAG_SYN;
    tcpt.padLen = 100;
    extPort = (20<<10) | rtl8651_naptTcpUdpTableIndex(1, tcpt.sip, tcpt.sport, tcpt.dip, tcpt.dport);
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[3], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[4], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST4_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_WAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST4_IP, errCount);
        CHK_PKT_IP(tcpt.sip, WAN0_ISP_PUBLIC_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 200, errCount);
        CHK_PKT_DEC(tcpt.sport, extPort, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testNaptThruPppoeStripVlan()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    uint16 extPort;
    
    VERI_MSG(("NAPT through PPPoE, striping VLAN tag ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_LAN0_MAC);
    strtomac(&tcpt.smac, HOST3_MAC);
    tcpt.vlanTag = 1;
    tcpt.vid = LAN0_VID;
    strtoip(&tcpt.dip, HOST5_IP);
    strtoip(&tcpt.sip, HOST3_IP);
    tcpt.dport = 200;
    tcpt.sport = 100;
    tcpt.flag = TCPFLAG_SYN;
    tcpt.padLen = 100;
    extPort = (20<<10) | rtl8651_naptTcpUdpTableIndex(1, tcpt.sip, tcpt.sport, tcpt.dip, tcpt.dport);
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[3], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[5], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;
        }
        CHK_PKT_MAC(tcpt.dmac, WAN0_ISP_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_WAN0_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 1, errCount);
        CHK_PKT_HEX(tcpt.sid, WAN0_ISP_PPPOE_SID, errCount);
        CHK_PKT_IP(tcpt.dip, HOST5_IP, errCount);
        CHK_PKT_IP(tcpt.sip, WAN0_ISP_PUBLIC_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 200, errCount);
        CHK_PKT_DEC(tcpt.sport, extPort, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testSrvPrtThruEth()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    
    VERI_MSG(("Inbound server port through Ethernet ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_WAN1_MAC);
    strtomac(&tcpt.smac, HOST4_MAC);
    strtoip(&tcpt.dip, WAN0_ISP_PUBLIC_IP);
    strtoip(&tcpt.sip, HOST4_IP);
    tcpt.dport = 20;
    tcpt.sport = 200;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[4], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[2], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST2_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_LAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST2_IP, errCount);
        CHK_PKT_IP(tcpt.sip, HOST4_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 1020, errCount);
        CHK_PKT_DEC(tcpt.sport, 200, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testSrvPrtThruPppoe()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    
    VERI_MSG(("Inbound server port through PPPoE ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_WAN0_MAC);
    strtomac(&tcpt.smac, HOST5_MAC);
    tcpt.pppoeTag = 1;
    tcpt.sid = WAN0_ISP_PPPOE_SID;
    strtoip(&tcpt.dip, WAN0_ISP_PUBLIC_IP);
    strtoip(&tcpt.sip, HOST5_IP);
    tcpt.dport = 20;
    tcpt.sport = 200;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[5], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[2], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST2_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_LAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST2_IP, errCount);
        CHK_PKT_IP(tcpt.sip, HOST5_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 1020, errCount);
        CHK_PKT_DEC(tcpt.sport, 200, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testWanRoute()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    
    VERI_MSG(("Inter-WAN routing, adding PPPoE tag ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_WAN1_MAC);
    strtomac(&tcpt.smac, HOST4_MAC);
    strtoip(&tcpt.dip, HOST5_IP);
    strtoip(&tcpt.sip, HOST4_IP);
    tcpt.dport = 100;
    tcpt.sport = 200;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[4], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[5], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;;
        }
        CHK_PKT_MAC(tcpt.dmac, WAN0_ISP_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_WAN0_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 1, errCount);
        CHK_PKT_HEX(tcpt.sid, WAN0_ISP_PPPOE_SID, errCount);
        CHK_PKT_IP(tcpt.dip, HOST5_IP, errCount);
        CHK_PKT_IP(tcpt.sip, HOST4_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 100, errCount);
        CHK_PKT_DEC(tcpt.sport, 200, errCount);
    }
    
    VERI_MSG(("Inter-WAN routing, striping PPPoE tag ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_WAN0_MAC);
    strtomac(&tcpt.smac, HOST5_MAC);
    tcpt.pppoeTag = 1;
    tcpt.sid = WAN0_ISP_PPPOE_SID;
    strtoip(&tcpt.dip, HOST4_IP);
    strtoip(&tcpt.sip, HOST5_IP);
    tcpt.dport = 100;
    tcpt.sport = 200;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[5], pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(hp[4], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++errCount;
        goto _EXIT;;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++errCount;
            goto _EXIT;;
        }
        CHK_PKT_MAC(tcpt.dmac, HOST4_MAC, errCount);
        CHK_PKT_MAC(tcpt.smac, GW_WAN1_MAC, errCount);
        CHK_PKT_DEC(tcpt.vlanTag, 0, errCount);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, errCount);
        CHK_PKT_IP(tcpt.dip, HOST4_IP, errCount);
        CHK_PKT_IP(tcpt.sip, HOST5_IP, errCount);
        CHK_PKT_DEC(tcpt.dport, 100, errCount);
        CHK_PKT_DEC(tcpt.sport, 200, errCount);
    }
    
_EXIT:
    return errCount;
}

uint32 l34_testAlgTrap()
{
    uint32 errCount = 0;
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    
    VERI_MSG(("ALG trapping ...\n"));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_LAN0_MAC);
    strtomac(&tcpt.smac, HOST3_MAC);
    tcpt.vlanTag = 1;
    tcpt.vid = LAN0_VID;
    strtoip(&tcpt.dip, HOST5_IP);
    strtoip(&tcpt.sip, HOST3_IP);
    tcpt.dport = 80;
    tcpt.sport = 100;
    tcpt.flag = 0;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(hp[3], pktBuf, pktLen);
    /* check trapped packet */
    pktLen=2048;
    miiLikeTestReceive(hp[3], pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet trapped\n"));
        ++errCount;
        goto _EXIT;
    }
    
_EXIT:
    return errCount;
}
