#include <linux/config.h>
#include "../../rtl_types.h"
#include "../../asicRegs.h"
#include "../../hsRegs.h"
#include "../../rtl8651_tblAsicDrv.h"
#include "../veri.h"


/* LOCAL DEFINITIONS 
*/
/* gateway */
#define WAN_VID 8
#define GW_WAN_MAC "00-00-00-00-88-88"
#define LAN_VID 9
#define GW_LAN_MAC "00-00-00-00-88-8b"
#define GW_LAN_IP "192.168.1.254"
#define GW_LAN_SUBNET_MASK "255.255.255.0"
/* isp */
#define WAN_ISP_MAC "00-00-00-00-99-9c"
#define WAN_ISP_PPPOE_SID 0x55
#define WAN_ISP_NPE_IP "203.49.1.11"
/* lan hosts */
#define LAN_NPI_HOST1_MAC "00-00-00-00-00-40"
#define LAN_NPI_HOST1_IP "192.168.1.161"
#define LAN_NPI_HOST2_MAC "00-00-00-00-00-50"
#define LAN_NPI_HOST2_IP "192.168.1.214"
/* rp */
#define RP_IP "211.72.254.7"
/* sessions 4-way-hashed into the same slot */
#define LAN_NPI_HOST1_PORT1 10
#define RP_PORT1 20
#define LAN_NPI_HOST1_PORT2 778
#define RP_PORT2 20
#define LAN_NPI_HOST1_PORT3 779
#define RP_PORT3 21
#define LAN_NPI_HOST1_PORT4 776
#define RP_PORT4 22
#define LAN_NPI_HOST1_PORT5 522
#define RP_PORT5 20
#define LAN_NPI_HOST1_PORT6 523
#define RP_PORT6 21
#define LAN_NPI_HOST1_PORT7 266
#define RP_PORT7 20
/* sessions 4-way-hashed into previous slot */
#define LAN_NPI_HOST1_PORT_PRV 11
#define RP_PORT_PRV 20
/* sessions 4-way-hashed into next slot */
#define LAN_NPI_HOST1_PORT_NXT 773
#define RP_PORT_NXT 20


/* LOCAL SUBROUTINES 
*/
static int32 sessionTest(uint32 sessionId, uint16 rpPort, uint16 intPort, uint32 dp, uint32 slotNum, uint32 *errCountp);


void napt4WayHash_tplg()
{
extern rtl865x_tblAsicDrv_vlanParam_t vlant;
extern _rtl8651_tblDrvAclRule_t aclt;
extern rtl865x_tblAsicDrv_l2Param_t l2t;
extern rtl865x_tblAsicDrv_pppoeParam_t pppoet;
extern rtl865x_tblAsicDrv_arpParam_t arpt;
extern rtl865x_tblAsicDrv_routingParam_t routet;
extern rtl865x_tblAsicDrv_extIntIpParam_t ipt;
    ipaddr_t ip32;
    uint32 i;
    
    /* switch core misc control reg: enable L4 hash, multicast MTU = 0x6ff */
    WRITE_MEM32(MISCCR, 0x16ff);
    /* module switch control reg: enable L2/L3/L4/STP, trap 802.1D/GARP to CPU */
    rtl8651_setAsicOperationLayer(4);
    /* TTL control reg: enable TTL-1 operation */
    WRITE_MEM32(TTLCR, EN_TTL1);
    /* VLAN tagging control reg: accept untagged */
    WRITE_MEM32(VLANTCR, 0);
    /* l4 table aging control reg */
    WRITE_MEM32(TEATCR, 0xffffffff);
    /* l4 table offset control reg: start = 20, end = 20 */
    WRITE_MEM32(OCR, (20<<OCR_START_OFFSET)|(20<<OCR_END_OFFSET));
    /* switch table misc control reg: enable napt auto learn/delete, enable VLAN ingress filtering, 
      trap non-TCP/UDP/ICMP NAPT packets to CPU, enable multicast table, accept broadcast packets, 
      enable PPPoE auto-encapsulation and auto-decapsulation */
    WRITE_MEM32(SWTMCR, 0x7600001e);
    /* gidx mapping control */
    WRITE_MEM32(GIDXMCR, 0);
    
    /* enable L4 TCP/UDP table 4-way hash */
    WRITE_MEM32(SWTECR, EN_TCPUDP_4WAY_HASH);
    
    /* port based vlan config reg */
    WRITE_MEM32(PVCR, (1<<3) | (0<<0));

    rtl8651_clearAsicAllTable();
    
    /* config vlan table */
    /* wan */
    bzero((void*) &vlant, sizeof(vlant));
    strtomac(&vlant.macAddr, GW_WAN_MAC);
    vlant.memberPortMask = 0x01;
    vlant.untagPortMask = 0x3f;
    vlant.macAddrNumber = 1;
    vlant.vid = WAN_VID;
    for(i=0;i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum;i++)
        vlant.portState[i] = RTL8651_PORTSTA_FORWARDING;
    vlant.mtu = 1522;
    vlant.inAclStart = 0;
    vlant.inAclEnd = 2;
    vlant.outAclStart = vlant.outAclEnd = 3;
    vlant.internal = 0;
    vlant.enableRoute = 1;
    vlant.valid = 1;
    rtl8651_setAsicVlan(&vlant);
    /* ingress filter gateway mac */
    aclt.ruleType_ = RTL8651_ACL_MAC;
    memcpy(aclt.srcMac_.octet, vlant.macAddr.octet, 6);
    memset(aclt.srcMacMask_.octet, 0xff, 6);
    aclt.actionType_ = RTL8651_ACL_CPU;
    rtl8651_setAsicAclRule(0, &aclt);
    /* lan */
    strtomac(&vlant.macAddr, GW_LAN_MAC);
    vlant.memberPortMask = 0x3e;
    vlant.vid = LAN_VID;
    vlant.inAclStart = 4;
    vlant.inAclEnd = 6;
    vlant.outAclStart = vlant.outAclEnd = 7;
    vlant.internal = 1;
    rtl8651_setAsicVlan(&vlant);
    /* ingress filter gateway mac */
    aclt.ruleType_ = RTL8651_ACL_MAC;
    memcpy(aclt.srcMac_.octet, vlant.macAddr.octet, 6);
    aclt.actionType_ = RTL8651_ACL_CPU;
    rtl8651_setAsicAclRule(4, &aclt);
    
    /* config l2 table */
    /* wan isp modem */
    bzero((void*) &l2t, sizeof(l2t));
    strtomac(&l2t.macAddr, WAN_ISP_MAC);
    l2t.memberPortMask = 0x01;
    l2t.isStatic = 1;
    rtl8651_setAsicL2Table(getL2TableRowFromMacStr(WAN_ISP_MAC), 0, &l2t);
    /* lan npi host 1 */
    strtomac(&l2t.macAddr, LAN_NPI_HOST1_MAC);
    l2t.memberPortMask = 0x02;
    rtl8651_setAsicL2Table(getL2TableRowFromMacStr(LAN_NPI_HOST1_MAC), 0, &l2t);
    /* lan npi host 2 */
    strtomac(&l2t.macAddr, LAN_NPI_HOST2_MAC);
    l2t.memberPortMask = 0x02;
    rtl8651_setAsicL2Table(getL2TableRowFromMacStr(LAN_NPI_HOST2_MAC), 0, &l2t);
    
    /* config pppoe table */
    bzero((void*) &pppoet, sizeof(pppoet));
    pppoet.sessionId = WAN_ISP_PPPOE_SID;
    rtl8651_setAsicPppoe(0, &pppoet);
    
    /* config arp table */
    /* lan npi host 1 */
    bzero((void*) &arpt, sizeof(arpt));
    arpt.nextHopRow = getL2TableRowFromMacStr(LAN_NPI_HOST1_MAC);
    strtoip(&ip32, LAN_NPI_HOST1_IP);
    rtl8651_setAsicArp(ip32 & 0xff, &arpt);
    /* lan npi host 2 */
    bzero((void*) &arpt, sizeof(arpt));
    arpt.nextHopRow = getL2TableRowFromMacStr(LAN_NPI_HOST2_MAC);
    strtoip(&ip32, LAN_NPI_HOST2_IP);
    rtl8651_setAsicArp(ip32 & 0xff, &arpt);
    
    /* config routing table */
    /* route to lan: use arp table [0:255] */
    bzero((void*) &routet, sizeof(routet));
    strtoip(&routet.ipAddr, GW_LAN_IP);
    strtoip(&routet.ipMask, GW_LAN_SUBNET_MASK);
    routet.process = PROCESS_INDIRECT;
    routet.vidx = rtl8651_vlanTableIndex(LAN_VID);
    routet.arpStart = 0;
    routet.arpEnd = 31;
    rtl8651_setAsicRouting(0, &routet);
    /* default route to wan: direct to modem */
    bzero((void*) &routet, sizeof(routet));
    strtoip(&routet.ipAddr, "0.0.0.0");
    strtoip(&routet.ipMask, "0.0.0.0");
    routet.process = PROCESS_PPPOE;
    routet.vidx = rtl8651_vlanTableIndex(WAN_VID);
    routet.nextHopRow = getL2TableRowFromMacStr(WAN_ISP_MAC);
    routet.pppoeIdx = 0;
    rtl8651_setAsicRouting(7, &routet);
    
    /* config ip table */
    /* napt */
    bzero((void*) &ipt, sizeof(ipt));
    strtoip(&ipt.extIpAddr, WAN_ISP_NPE_IP);
    rtl8651_setAsicExtIntIpTable(0, &ipt);
    
    VERI_MSG(("*****\n* NAPT 4-way hash test topology ready!\n*****\n"));
    return;
}

/* LAN to WAN by NAPT */
#define NAPT4WAYHASH_TEST1_TARGET1 \
"Verify that 4 sessions that are 4-way-hashed into the same slot can all be learnt even some of them are 1-way-hashed into the same entry. Verify that NAPT/R traffic works, indicating that hash result being masked out 2 lsb's is done correctly\n"
#define NAPT4WAYHASH_TEST1_TARGET2 \
"Verify that 5th session that is 4-way-hashed into the same slot results in a collision and is trapped to CPU. Read the table to make sure that the collision bit is correctly set on the 1-way-hashed entry.\n"
#define NAPT4WAYHASH_TEST1_TARGET3 \
"Verify that 6th session that H4(6th session)=H4(5th session) should be trapped to CPU even though there is already an empty entry.\n"
#define NAPT4WAYHASH_TEST1_TARGET4 \
"Verify that 7th session that is 4-way-hashed into the same slot can be learnt.\n"
#define NAPT4WAYHASH_TEST1_TARGET5 \
"Write an SPI entry with all same parameters as that of 7th session. Verify that 7th session does not pass by it.\n"
#define NAPT4WAYHASH_TEST1_TARGET6 \
"Boundry Test. Verify that 2 sessions that are individually 1-way-hash to previous and next of the four entries are autolearnt to the last entry of their 4-way-hash slot.\n"
uint32 napt4WayHash_test1()
{
    uint32 errCount = 0;
    rtl865x_tblAsicDrv_naptTcpUdpParam_t tcptbl;
    
    /* clear napt tcp/udp table */
    rtl8651_delAsicNaptTcpUdpTable(0, RTL8651_TCPUDPTBL_SIZE - 1);
    
    VERI_MSG(("*****\n"));
    VERI_MSG((NAPT4WAYHASH_TEST1_TARGET1));
    
    /* H1(1st session)=479, H4(1st session)=476, actually to 479 */
    if ((sessionTest(1, //sessionId
                RP_PORT1, //rpPort
                LAN_NPI_HOST1_PORT1, //intPort
                1, //dp
                479, //slotNum
                &errCount) != SUCCESS) || (errCount != 0))
        goto _EXIT;

    /* H1(2nd session)=476, H4(2nd session)=476, actually to 478 */
    if ((sessionTest(2, //sessionId
                RP_PORT2, //rpPort
                LAN_NPI_HOST1_PORT2, //intPort
                1, //dp
                478, //slotNum
                &errCount) != SUCCESS) || (errCount != 0))
        goto _EXIT;
        
    /* H1(3rd session)=476, H4(3rd session)=476, actually to 477 */
    if ((sessionTest(3, //sessionId
                RP_PORT3, //rpPort
                LAN_NPI_HOST1_PORT3, //intPort
                1, //dp
                477, //slotNum
                &errCount) != SUCCESS) || (errCount != 0))
        goto _EXIT;
        
    /* H1(4th session)=476, H4(4th session)=476, actually to 476 */
    if ((sessionTest(4, //sessionId
                RP_PORT4, //rpPort
                LAN_NPI_HOST1_PORT4, //intPort
                1, //dp
                476, //slotNum
                &errCount) != SUCCESS) || (errCount != 0))
        goto _EXIT;
        
    VERI_MSG(("Test Done\n"));
    VERI_MSG(("*****\n"));
    
    VERI_MSG(("*****\n"));
    VERI_MSG((NAPT4WAYHASH_TEST1_TARGET2));
    
    /* H1(5th session)=477, H4(5th session)=476, actually collision */
    if ((sessionTest(5, //sessionId
                RP_PORT5, //rpPort
                LAN_NPI_HOST1_PORT5, //intPort
                0x40, //dp
                0, //slotNum
                &errCount) != SUCCESS) || (errCount != 0))
        goto _EXIT;
    /* check table */
    rtl8651_getAsicNaptTcpUdpTable(477, &tcptbl);
    if (!tcptbl.isValid || !tcptbl.isCollision)
    {
        VERI_MSG(("Error! table content incorrect\n"));
        ++errCount;
        goto _EXIT;
    }
        
    VERI_MSG(("Test Done\n"));
    VERI_MSG(("*****\n"));
    
    VERI_MSG(("*****\n"));
    VERI_MSG((NAPT4WAYHASH_TEST1_TARGET3));
    
    /* remove an non-collision entry 476 */
    rtl8651_delAsicNaptTcpUdpTable(476, 476);
    /* H1(6th session)=477, H4(6th session)=476, actually still collision */
    if ((sessionTest(6, //sessionId
                RP_PORT6, //rpPort
                LAN_NPI_HOST1_PORT6, //intPort
                0x40, //dp
                0, //slotNum
                &errCount) != SUCCESS) || (errCount != 0))
        goto _EXIT;
        
    VERI_MSG(("Test Done\n"));
    VERI_MSG(("*****\n"));
    
    VERI_MSG(("*****\n"));
    VERI_MSG((NAPT4WAYHASH_TEST1_TARGET4));
    
    /* H1(7th session)=478, H4(7th session)=476, actually to 476 */
    if ((sessionTest(7, //sessionId
                RP_PORT7, //rpPort
                LAN_NPI_HOST1_PORT7, //intPort
                1, //dp
                476, //slotNum
                &errCount) != SUCCESS) || (errCount != 0))
        goto _EXIT;
        
    VERI_MSG(("Test Done\n"));
    VERI_MSG(("*****\n"));
    
    VERI_MSG(("*****\n"));
    VERI_MSG((NAPT4WAYHASH_TEST1_TARGET5));
    
    /* change entry 476 into spi */
    rtl8651_getAsicNaptTcpUdpTable(476, &tcptbl);
    tcptbl.isDedicated = 1;
    rtl8651_setAsicNaptTcpUdpTable(1, 476, &tcptbl);
    /* H1(7th session)=478, H4(7th session)=476, actually collision */
    if ((sessionTest(7, //sessionId
                RP_PORT7, //rpPort
                LAN_NPI_HOST1_PORT7, //intPort
                0x40, //dp
                0, //slotNum
                &errCount) != SUCCESS) || (errCount != 0))
        goto _EXIT;
    /* check table */
    rtl8651_getAsicNaptTcpUdpTable(476, &tcptbl);
    if (!tcptbl.isValid || !tcptbl.isDedicated)
    {
        VERI_MSG(("Error! table content incorrect\n"));
        ++errCount;
        goto _EXIT;
    }
    rtl8651_getAsicNaptTcpUdpTable(478, &tcptbl);
    if (!tcptbl.isValid || !tcptbl.isCollision)
    {
        VERI_MSG(("Error! table content incorrect\n"));
        ++errCount;
        goto _EXIT;
    }
        
    VERI_MSG(("Test Done\n"));
    VERI_MSG(("*****\n"));
    
    VERI_MSG(("*****\n"));
    VERI_MSG((NAPT4WAYHASH_TEST1_TARGET6));
    
    /* H1(8st session)=475, H4(8st session)=472, actually to 475 */
    if ((sessionTest(8, //sessionId
                RP_PORT_PRV, //rpPort
                LAN_NPI_HOST1_PORT_PRV, //intPort
                1, //dp
                475, //slotNum
                &errCount) != SUCCESS) || (errCount != 0))
        goto _EXIT;
    /* H1(9st session)=480, H4(9st session)=480, actually to 483 */
    if ((sessionTest(9, //sessionId
                RP_PORT_NXT, //rpPort
                LAN_NPI_HOST1_PORT_NXT, //intPort
                1, //dp
                483, //slotNum
                &errCount) != SUCCESS) || (errCount != 0))
        goto _EXIT;
        
    VERI_MSG(("Test Done\n"));
    VERI_MSG(("*****\n"));
    
_EXIT:
    /* clear napt tcp/udp table */
    rtl8651_delAsicNaptTcpUdpTable(0, RTL8651_TCPUDPTBL_SIZE - 1);
    
    if(errCount) VERI_MSG(("! %d errors\n", errCount));
    
    return errCount;
}

static int32 sessionTest(uint32 sessionId, uint16 rpPort, uint16 intPort, uint32 dp, uint32 slotNum, uint32 *errCountp)
{
    uint8 pktBuf[2048];
    uint32 pktLen;
    pktGenTcpFormat_t tcpt;
    
    /* outbound */
    VERI_MSG(("Outbound packet of session #%d ...\n", sessionId));
    bzero((void*) &tcpt, sizeof(tcpt));
    strtomac(&tcpt.dmac, GW_LAN_MAC);
    strtomac(&tcpt.smac, LAN_NPI_HOST1_MAC);
    strtoip(&tcpt.dip, RP_IP);
    strtoip(&tcpt.sip, LAN_NPI_HOST1_IP);
    tcpt.dport = rpPort;
    tcpt.sport = intPort;
    tcpt.flag = TCPFLAG_SYN;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(1, pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    if (dp != 0x40)
        miiLikeTestReceive(0, pktBuf, &pktLen);
    else
        miiLikeTestReceive(1, pktBuf, &pktLen); //directly trapped
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++*errCountp;
        return FAILED;
    }
    else
    {
        if (dp == 0x40)
            return SUCCESS;
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++*errCountp;
            return FAILED;
        }
        CHK_PKT_MAC(tcpt.dmac, WAN_ISP_MAC, *errCountp);
        CHK_PKT_MAC(tcpt.smac, GW_WAN_MAC, *errCountp);
        CHK_PKT_DEC(tcpt.vlanTag, 0, *errCountp);
        CHK_PKT_DEC(tcpt.pppoeTag, 1, *errCountp);
        CHK_PKT_HEX(tcpt.sid, WAN_ISP_PPPOE_SID, *errCountp);
        CHK_PKT_IP(tcpt.dip, RP_IP, *errCountp);
        CHK_PKT_IP(tcpt.sip, WAN_ISP_NPE_IP, *errCountp);
        CHK_PKT_DEC(tcpt.dport, rpPort, *errCountp);
        CHK_PKT_DEC(tcpt.sport, (20<<10)|slotNum, *errCountp);
    }
    if (*errCountp > 0)
        return FAILED;
    /* inbound */
    VERI_MSG(("Inbound packet of session #%d ...\n", sessionId));
    bzero((void*) &tcpt, sizeof(tcpt));
    tcpt.pppoeTag = 1;
    tcpt.sid = WAN_ISP_PPPOE_SID;
    strtomac(&tcpt.dmac, GW_WAN_MAC);
    strtomac(&tcpt.smac, WAN_ISP_MAC);
    strtoip(&tcpt.dip, WAN_ISP_NPE_IP);
    strtoip(&tcpt.sip, RP_IP);
    tcpt.dport = (20<<10)|slotNum;
    tcpt.sport = rpPort;
    tcpt.flag = TCPFLAG_ACK;
    tcpt.padLen = 100;
    pktLen = pktGenTcp(pktBuf, &tcpt);
    miiLikeTestSend(0, pktBuf, pktLen);
    /* check received packet */
    pktLen=2048;
    miiLikeTestReceive(1, pktBuf, &pktLen);
    if (pktLen == 0)
    {
        VERI_MSG(("Error! no packet received\n"));
        ++*errCountp;
        return FAILED;
    }
    else
    {
        if (pktParseTcp(pktBuf, &tcpt, pktLen) != SUCCESS)
        {
            VERI_MSG(("Error! received packet is not TCP\n"));
            ++*errCountp;
            return FAILED;
        }
        CHK_PKT_MAC(tcpt.dmac, LAN_NPI_HOST1_MAC, *errCountp);
        CHK_PKT_MAC(tcpt.smac, GW_LAN_MAC, *errCountp);
        CHK_PKT_DEC(tcpt.vlanTag, 0, *errCountp);
        CHK_PKT_DEC(tcpt.pppoeTag, 0, *errCountp);
        CHK_PKT_IP(tcpt.dip, LAN_NPI_HOST1_IP, *errCountp);
        CHK_PKT_IP(tcpt.sip, RP_IP, *errCountp);
        CHK_PKT_DEC(tcpt.dport, intPort, *errCountp);
        CHK_PKT_DEC(tcpt.sport, rpPort, *errCountp);
    }
    return SUCCESS;
}
