#include "../types.h"
#include "../asicRegs.h"
#include "../asicTabs.h"
#include "../rtl8651_tblAsicDrv.h"
#include "../hsRegs.h"
#include "../rtl_utils.h"
#include "veri.h"

static uint32 stoi(uint8 *s)
{
    uint32 i = 0;
    while ((*s >= '0') && (*s <= '9'))
        i = i*10 + *(s++) - '0';
    return i;
}

static int32 _parseVlan(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_vlanTable_t    entry;
    ether_addr_t           mac;
    
    if ( idx >= RTL8651_VLANTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_VLAN_TABLE, idx, &entry);
    
    if (!force && !entry.valid)
        return FAILED;
    
    printf("[%d] ", idx);
    printf("mac=%02x:%02x:%02x:%02x:%02x:%02x",
                entry.mac47_32>>8, entry.mac47_32&0xff, entry.mac31_16>>8, 
                entry.mac31_16&0xff, entry.mac15_0>>8, entry.mac15_0&0xff);
    printf(",valid=%d", entry.valid);
    printf(",mbr=0x%x", entry.memberPort);
    printf(",vhid=%d(vid=%d)", entry.vhid, (entry.vhid<<3) | idx);
    printf(",acl=[%d:%d,%d:%d]", entry.inACLStart, entry.inACLEnd, entry.outACLStart, entry.outACLEnd);
    printf(",internal=%d", entry.isInternal);
    printf(",enroute=%d", entry.enHWRoute);
    printf(",spt=%d%d%d%d%d%d", (entry.STPStatus>>10)&3, (entry.STPStatus>>8)&3, (entry.STPStatus>>6)&3
                        , (entry.STPStatus>>4)&3, (entry.STPStatus>>2)&3, entry.STPStatus&3);
    printf(",bc2cpu=%d", entry.bcastToCPU);
    printf(",promiscuous=%d", entry.promiscuous);
    printf(",untag=0x%x", entry.egressUntag);
    printf(",macmask=%d", entry.macMask);
    printf(",mtu=%d", (entry.mtuH << 8) | (entry.mtuL & 0xff));
#ifdef CONFIG_RTL865XB
    printf(",extmbr=%d", entry.extMemberPort);
    printf(",extuntag=%d", entry.extEgressUntag);
    printf(",extstp=%d%d%d", (entry.extSTPStatus>>4)&3, (entry.extSTPStatus>>2)&3, entry.extSTPStatus&3);
    printf(",dmz=%d", entry.isDMZ);
    switch ((entry.isInternal<<1) | entry.isDMZ)
    {
        case 0: printf("(wan)"); break;
        case 1: printf("(dmz)"); break;
        case 2: printf("(lan)"); break;
        case 3: printf("(rlan)"); break;
    }
    printf(",macnon=%d", entry.macNotExist);
#endif /*CONFIG_RTL865XB*/

    printf("\n");
    return SUCCESS;
}

static int32 _parseProtoTrap(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_protoTrapTable_t   entry;
    
    if ( idx >= RTL8651_PROTOTRAPTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_PROTOCOL_TRAP_TABLE, idx, &entry);
    
    if (!force && !entry.valid)
        return FAILED;
    
    printf("[%d] ", idx);
    printf("valid=%d", entry.valid);
    printf(",proto=%d", entry.trapProtocol);
    switch (entry.trapProtocol)
    {
        case TYPE_TRAP_ETHERNET:
            printf("(eth)");
            break;
        case TYPE_TRAP_IP:
            printf("(ip)");
            break;
        case TYPE_TRAP_TCP:
            printf("(tcp");
            break;
        case TYPE_TRAP_UDP:
            printf("(udp)");
            break;
    }
    printf(",content=0x%x", entry.trapContent);

    printf("\n");
    return SUCCESS;
}

static int32 _parseServerPort(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_srvPortTable_t entry;
    
    if ( idx >= RTL8651_SERVERPORTTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_SERVER_PORT_TABLE, idx, &entry);
    
    if (!force && !entry.valid)
        return FAILED;
    
    printf("[%d] ", idx);
    printf("valid=%d", entry.valid);
    printf(",intip=%d.%d.%d.%d", (entry.internalIP>>24)&0xff, (entry.internalIP>>16)&0xff
                        , (entry.internalIP>>8)&0xff, entry.internalIP&0xff);
    printf(",extip=%d.%d.%d.%d", (entry.externalIP>>24)&0xff, (entry.externalIP>>16)&0xff
                        , (entry.externalIP>>8)&0xff, entry.externalIP&0xff);
    printf(",intprt=0x%x", entry.internalPort);
    printf(",extprt=0x%x", entry.externalPort);
    
#ifdef CONFIG_RTL865XB
    printf(",nxthop=%d", entry.nextHop);
    {
        rtl8651_tblAsic_nextHopTable_t nxthop;
        rtl8651_tblAsic_pppoeTable_t pppoe;
        
        _rtl8651_readAsicEntry(TYPE_NEXT_HOP_TABLE, entry.nextHop, &nxthop);
        printf("(dvidx=%d", nxthop.dstVid);
        printf(",pppidx=%d", nxthop.PPPoEIndex);
        _rtl8651_readAsicEntry(TYPE_PPPOE_TABLE, nxthop.PPPoEIndex, &pppoe);
        printf("->%d", pppoe.sessionID);
        printf(",l2=%d", nxthop.nextHop);
        {
            rtl8651_tblAsic_l2Table_t l2;
            ether_addr_t mac;
        
            _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, nxthop.nextHop, &l2);
        		mac.octet[0] = l2.mac47_40;
        		mac.octet[1] = l2.mac39_24 >> 8;
        		mac.octet[2] = l2.mac39_24 & 0xff;
        		mac.octet[3] = l2.mac23_8 >> 8;
        		mac.octet[4] = l2.mac23_8 & 0xff;
        		mac.octet[5] = (nxthop.nextHop >> 2) ^ mac.octet[0] ^ mac.octet[1] ^ mac.octet[2] ^ mac.octet[3] ^ mac.octet[4];
            printf("->mac=%02x:%02x:%02x:%02x:%02x:%02x)",
                    mac.octet[0], mac.octet[1], mac.octet[2], 
                    mac.octet[3], mac.octet[4], mac.octet[5]);
        }
    }
    printf(",isrange=%d", entry.isPortRange);
#endif /*CONFIG_RTL865XB*/

    printf("\n");
    return SUCCESS;
}

static int32 _parseExtIntIp(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_extIpTable_t     entry;
    
    if ( idx >= RTL8651_IPTABLE_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_EXT_INT_IP_TABLE, idx, &entry);
    
    if (!force && !entry.valid)
        return FAILED;
    
    printf("[%d] ", idx);
    printf("valid=%d", entry.valid);
    printf(",intip=%d.%d.%d.%d", (entry.internalIP>>24)&0xff, (entry.internalIP>>16)&0xff
                        , (entry.internalIP>>8)&0xff, entry.internalIP&0xff);
    printf(",extip=%d.%d.%d.%d", (entry.externalIP>>24)&0xff, (entry.externalIP>>16)&0xff
                        , (entry.externalIP>>8)&0xff, entry.externalIP&0xff);
    printf(",type=%d", (entry.isLocalPublic<<1) | entry.isOne2One);
    if (entry.isLocalPublic) printf("(local public)");
    if (entry.isOne2One) printf("(NAT)");
    
#ifdef CONFIG_RTL865XB
    printf(",nxthop=%d", entry.nextHop);
    {
        rtl8651_tblAsic_nextHopTable_t nxthop;
        rtl8651_tblAsic_pppoeTable_t pppoe;
        
        _rtl8651_readAsicEntry(TYPE_NEXT_HOP_TABLE, entry.nextHop, &nxthop);
        printf("(dvidx=%d", nxthop.dstVid);
        printf(",pppidx=%d", nxthop.PPPoEIndex);
        _rtl8651_readAsicEntry(TYPE_PPPOE_TABLE, nxthop.PPPoEIndex, &pppoe);
        printf("->%d", pppoe.sessionID);
        printf(",l2=%d", nxthop.nextHop);
        {
            rtl8651_tblAsic_l2Table_t l2;
            ether_addr_t mac;
        
            _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, nxthop.nextHop, &l2);
        		mac.octet[0] = l2.mac47_40;
        		mac.octet[1] = l2.mac39_24 >> 8;
        		mac.octet[2] = l2.mac39_24 & 0xff;
        		mac.octet[3] = l2.mac23_8 >> 8;
        		mac.octet[4] = l2.mac23_8 & 0xff;
        		mac.octet[5] = (nxthop.nextHop >> 2) ^ mac.octet[0] ^ mac.octet[1] ^ mac.octet[2] ^ mac.octet[3] ^ mac.octet[4];
            printf("->mac=%02x:%02x:%02x:%02x:%02x:%02x)",
                    mac.octet[0], mac.octet[1], mac.octet[2], 
                    mac.octet[3], mac.octet[4], mac.octet[5]);
        }
    }
#endif /*CONFIG_RTL865XB*/

    printf("\n");
    return SUCCESS;
}

static int32 _parseAlg(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_algTable_t entry;
    
    if ( idx >= RTL8651_ALGTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_ALG_TABLE, idx, &entry);
    
    if (!force && !entry.valid)
        return FAILED;
    
    printf("[%d] ", idx);
    printf("valid=%d", entry.valid);
    printf(",prt=0x%x", entry.L4Port);

    printf("\n");
    return SUCCESS;
}

static int32 _parseL3(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_l3RouteTable_t  entry;
    
    if ( idx >= RTL8651_ROUTINGTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_L3_ROUTING_TABLE, idx, &entry);
    
    if (!force && !entry.linkTo.ARPEntry.valid)
        return FAILED;
    
    printf("[%d] ", idx);
    printf("valid=%d", entry.linkTo.ARPEntry.valid);
    printf(",ip=%d.%d.%d.%d", (entry.IPAddr>>24)&0xff, (entry.IPAddr>>16)&0xff
                        , (entry.IPAddr>>8)&0xff, entry.IPAddr&0xff);
    printf(",mask=%d", entry.linkTo.ARPEntry.IPMask);
    printf(",hp=%d", entry.linkTo.ARPEntry.hPriority);
    printf(",process=%d", entry.linkTo.ARPEntry.process);
    switch (entry.linkTo.ARPEntry.process)
    {
        case PROCESS_DIRECT:
            printf("(dir)");
            printf(",nxthop=%d", entry.linkTo.L2Entry.nextHop);
            {
                rtl8651_tblAsic_l2Table_t l2;
                ether_addr_t mac;
            
                _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, entry.linkTo.L2Entry.nextHop, &l2);
            		mac.octet[0] = l2.mac47_40;
            		mac.octet[1] = l2.mac39_24 >> 8;
            		mac.octet[2] = l2.mac39_24 & 0xff;
            		mac.octet[3] = l2.mac23_8 >> 8;
            		mac.octet[4] = l2.mac23_8 & 0xff;
            		mac.octet[5] = (entry.linkTo.L2Entry.nextHop >> 2) ^ mac.octet[0] ^ mac.octet[1] ^ mac.octet[2] ^ mac.octet[3] ^ mac.octet[4];
                printf("(mac=%02x:%02x:%02x:%02x:%02x:%02x)",
                        mac.octet[0], mac.octet[1], mac.octet[2], 
                        mac.octet[3], mac.octet[4], mac.octet[5]);
            }
            printf(",dvidx=%d", entry.linkTo.ARPEntry.vid);
            break;
        case PROCESS_INDIRECT:
            printf("(arp)");
            printf(",arpsta=%d", entry.linkTo.ARPEntry.ARPStart);
            printf(",arpend=%d", entry.linkTo.ARPEntry.ARPEnd);
            printf("(%d~%d)", entry.linkTo.ARPEntry.ARPStart<<3, (entry.linkTo.ARPEntry.ARPEnd<<3) + 7);
            printf(",dvidx=%d", entry.linkTo.ARPEntry.vid);
            break;
        case PROCESS_PPPOE:
            printf("(pppoe)");
            printf(",nxthop=%d", entry.linkTo.PPPoEEntry.nextHop);
            {
                rtl8651_tblAsic_l2Table_t l2;
                ether_addr_t mac;
            
                _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, entry.linkTo.PPPoEEntry.nextHop, &l2);
            		mac.octet[0] = l2.mac47_40;
            		mac.octet[1] = l2.mac39_24 >> 8;
            		mac.octet[2] = l2.mac39_24 & 0xff;
            		mac.octet[3] = l2.mac23_8 >> 8;
            		mac.octet[4] = l2.mac23_8 & 0xff;
            		mac.octet[5] = (entry.linkTo.PPPoEEntry.nextHop >> 2) ^ mac.octet[0] ^ mac.octet[1] ^ mac.octet[2] ^ mac.octet[3] ^ mac.octet[4];
                printf("(mac=%02x:%02x:%02x:%02x:%02x:%02x)",
                        mac.octet[0], mac.octet[1], mac.octet[2], 
                        mac.octet[3], mac.octet[4], mac.octet[5]);
            }
            printf(",pppidx=%d", entry.linkTo.PPPoEEntry.PPPoEIndex);
            {
                rtl8651_tblAsic_pppoeTable_t pppoe;
                
                _rtl8651_readAsicEntry(TYPE_PPPOE_TABLE, entry.linkTo.PPPoEEntry.PPPoEIndex, &pppoe);
                printf("(session=%d)", pppoe.sessionID);
            }
            printf(",dvidx=%d", entry.linkTo.ARPEntry.vid);
            break;
        case PROCESS_S_DROP:
            printf("(sdrop)");
            break;
        case PROCESS_S_CPU:
            printf("(scpu)");
            break;
#ifdef CONFIG_RTL865XB
        case PROCESS_NXT_HOP:
            printf("(nxthop)");
            printf(",nhsta=%d", entry.linkTo.NxtHopEntry.nhStart);
            printf(",nhnum=%d(%d entries)", entry.linkTo.NxtHopEntry.nhNum, 2^entry.linkTo.NxtHopEntry.nhNum);
            printf(",nhnxt=%d", entry.linkTo.NxtHopEntry.nhNxt);
            printf(",nhalgo=%d", entry.linkTo.NxtHopEntry.nhAlgo);
            switch (entry.linkTo.NxtHopEntry.nhAlgo)
            {
                case 0: printf("(per-pkt)"); break;
                case 1: printf("(per-session)"); break;
                case 2: printf("(per-sip)"); break;
            }
            printf(",ipdomain=%d", entry.linkTo.NxtHopEntry.IPDomain);
            switch (entry.linkTo.NxtHopEntry.IPDomain)
            {
                case 0: printf("(entries 0~3)"); break;
                case 1: printf("(entries 4~7)"); break;
                case 2: printf("(entries 8~11)"); break;
                case 3: printf("(entries 12~15)"); break;
                case 4: printf("(entries 0~7)"); break;
                case 5: printf("(entries 8~15)"); break;
            }
            break;
#else
        case PROCESS_N_DROP:
            printf("(ndrop)");
            break;
        case PROCESS_N_CPU:
            printf("(ncpu)");
            break;
#endif /*CONFIG_RTL865XB*/
    }

    printf("\n");
    return SUCCESS;
}

static int32 _parseArp(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_arpTable_t entry;
    
    if ( idx >= RTL8651_ARPTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_ARP_TABLE, idx, &entry);
    
    if (!force && !entry.valid)
        return FAILED;
    
    printf("[%d] ", idx);
    printf("valid=%d", entry.valid);
    printf(",nxthop=%d", entry.nextHop);
    {
        rtl8651_tblAsic_l2Table_t l2;
        ether_addr_t mac;
    
        _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, entry.nextHop, &l2);
    		mac.octet[0] = l2.mac47_40;
    		mac.octet[1] = l2.mac39_24 >> 8;
    		mac.octet[2] = l2.mac39_24 & 0xff;
    		mac.octet[3] = l2.mac23_8 >> 8;
    		mac.octet[4] = l2.mac23_8 & 0xff;
    		mac.octet[5] = (entry.nextHop >> 2) ^ mac.octet[0] ^ mac.octet[1] ^ mac.octet[2] ^ mac.octet[3] ^ mac.octet[4];
        printf("(mac=%02x:%02x:%02x:%02x:%02x:%02x)",
                mac.octet[0], mac.octet[1], mac.octet[2], 
                mac.octet[3], mac.octet[4], mac.octet[5]);
    }

    printf("\n");
    return SUCCESS;
}

static int32 _parsePPPoE(uint32 idx)
{
    rtl8651_tblAsic_pppoeTable_t entry;
    
    if ( idx >= RTL8651_PPPOETBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_PPPOE_TABLE, idx, &entry);
    
    printf("[%d] ", idx);
    printf("session=%d", entry.sessionID);
    
#ifdef CONFIG_RTL865XB
    printf(",age=%d", entry.ageTime);
#endif /*CONFIG_RTL865XB*/

    printf("\n");
    return SUCCESS;
}

static int32 _parseL2(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_l2Table_t   entry;
    ether_addr_t           mac;
    
    if ( idx >= RTL8651_MACTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, idx, &entry);
    
    if (!force && !entry.isStatic && !entry.agingTime)
        return FAILED;
    
    printf("[%d] ", idx);
		mac.octet[0] = entry.mac47_40;
		mac.octet[1] = entry.mac39_24 >> 8;
		mac.octet[2] = entry.mac39_24 & 0xff;
		mac.octet[3] = entry.mac23_8 >> 8;
		mac.octet[4] = entry.mac23_8 & 0xff;
		mac.octet[5] = (idx >> 2) ^ mac.octet[0] ^ mac.octet[1] ^ mac.octet[2] ^ mac.octet[3] ^ mac.octet[4];
    printf("mac=%02x:%02x:%02x:%02x:%02x:%02x",
                mac.octet[0], mac.octet[1], mac.octet[2], 
                mac.octet[3], mac.octet[4], mac.octet[5]);
    printf(",mbr=0x%x", entry.memberPort);
    printf(",hp=%d", entry.hPriority);
    printf(",cpu=%d", entry.toCPU);
    printf(",static=%d", entry.isStatic);
    printf(",age=%d", entry.agingTime);
    printf(",srcblk=%d", entry.srcBlock);
    printf(",nxthop=%d", entry.nxtHostFlag);
#ifdef CONFIG_RTL865XB
    printf(",extmbr=%d", entry.extMemberPort);
#endif

    printf("\n");
    return SUCCESS;
}

static int32 _parseMcast(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_ipMulticastTable_t entry;
    
    if ( idx >= RTL8651_MULTICASTTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_MULTICAST_TABLE, idx, &entry);
    
    if (!force && !entry.valid)
        return FAILED;
    
    printf("[%d] ", idx);
    printf("valid=%d", entry.valid);
    printf(",intip=%d.%d.%d.%d", (entry.srcIPAddr>>24)&0xff, (entry.srcIPAddr>>16)&0xff
                        , (entry.srcIPAddr>>8)&0xff, entry.srcIPAddr&0xff);
    printf(",intip=224.%d.%d.%d", (entry.destIPAddrLsbs>>16)&0xff
                        , (entry.destIPAddrLsbs>>8)&0xff, entry.destIPAddrLsbs&0xff);
    printf(",svid=%d", entry.srcVid);
    printf(",spa=%d", (entry.srcPortL & 1) | (entry.srcPortH << 1));
    printf(",mbr=0x%x", entry.portList);
    printf(",extip=%d", entry.extIPIndex);
    printf(",cpu=%d", entry.toCPU);
    
#ifdef CONFIG_RTL865XB
    printf(",extspa=%d", entry.srcPortExt);
    printf(",extmbr=%d", entry.extPortList);
    printf(",age=%d", entry.ageTime);
    printf(",extiph=%d", entry.extIPIndexH);
    {
        rtl8651_tblAsic_extIpTable_t extIp;
        
        _rtl8651_readAsicEntry(TYPE_EXT_INT_IP_TABLE, entry.extIPIndex|(entry.extIPIndexH<<3), &extIp);
        printf("(extip[%d]=%d.%d.%d.%d)", entry.extIPIndex|(entry.extIPIndexH<<3), 
                   	   	   	(extIp.externalIP>>24)&0xff, (extIp.externalIP>>16)&0xff, 
                            (extIp.externalIP>>8)&0xff, extIp.externalIP&0xff);
    }
#endif /*CONFIG_RTL865XB*/

    printf("\n");
    return SUCCESS;
}

int32 _parseTcpUdp(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_naptTcpUdpTable_t entry;
    
    if ( idx >= RTL8651_TCPUDPTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_L4_TCP_UDP_TABLE, idx, &entry);
    
#ifndef CONFIG_RTL865XB
    if (!force && !entry.valid)
#else
    if (!force && !entry.valid && !entry.dedicate)
#endif /*CONFIG_RTL865XB*/
        return FAILED;
    
    printf("[%d] ", idx);
    printf("valid=%d", entry.valid);
    printf(",intip=%d.%d.%d.%d", (entry.intIPAddr>>24)&0xff, (entry.intIPAddr>>16)&0xff
                        , (entry.intIPAddr>>8)&0xff, entry.intIPAddr&0xff);
    printf(",intprt=0x%x", entry.intPort);
    printf(",col=%d", entry.collision);
    printf(",age=0x%x", entry.agingTime);
    printf(",offset=%d", entry.offset);
    printf(",static=%d", entry.isStatic);
    printf(",tcpflag=%d", entry.TCPFlag);
#ifdef CONFIG_RTL865XB
    if (entry.valid && entry.dedicate)
    {
        if (entry.TCPFlag & 4)
            printf("(quiet");
        else
            printf("(trap-syn/fin/rst");
        if (entry.TCPFlag & 2)
            if (entry.TCPFlag & 1)
                printf(",outbound)");
            else
                printf(",inbound)");
        else
            printf(",bidirection)");
    }
#endif /*CONFIG_RTL865XB*/
    printf(",istcp=%d", entry.isTCP);
#ifdef CONFIG_RTL865XB
    printf(",dedicated=%d", entry.dedicate);
    if ( (entry.valid == 0) && (entry.dedicate == 1) )
        printf("(spi)");
    else if ( (entry.valid == 1) && (entry.dedicate == 0) )
        printf("(hash1)");
    else if ( (entry.valid == 1) && (entry.dedicate == 1) )
        printf("(hash2)");
    printf(",col2=%d", entry.collision2);
    printf(",selipidx=%d", entry.selIPIdx);
    if (entry.valid && entry.dedicate)
    {
        rtl8651_tblAsic_extIpTable_t extIp;
        
        _rtl8651_readAsicEntry(TYPE_EXT_INT_IP_TABLE, entry.selIPIdx, &extIp);
        printf("(extip=%d.%d.%d.%d)", (extIp.externalIP>>24)&0xff, (extIp.externalIP>>16)&0xff
                            , (extIp.externalIP>>8)&0xff, extIp.externalIP&0xff);
    }
    printf(",seleidx=%d", entry.selEIdx);
    if (entry.valid && entry.dedicate && (entry.TCPFlag & 3 == 3))
    {
        printf("(inbound-index)");
    }
    else if ( entry.valid && !entry.dedicate && entry.isStatic && (entry.selEIdx & 1) )
    {
        printf("(nxthop:");
        {
            rtl8651_tblAsic_nextHopTable_t nxthop;
            rtl8651_tblAsic_pppoeTable_t pppoe;
            
            _rtl8651_readAsicEntry(TYPE_NEXT_HOP_TABLE, (entry.selEIdx >> 1) & 0x1f, &nxthop);
            printf("extip=%d", nxthop.IPIndex);
            {
                rtl8651_tblAsic_extIpTable_t extIp;
                
                _rtl8651_readAsicEntry(TYPE_EXT_INT_IP_TABLE, nxthop.IPIndex, &extIp);
                printf("->%d.%d.%d.%d", (extIp.externalIP>>24)&0xff, (extIp.externalIP>>16)&0xff
                                    , (extIp.externalIP>>8)&0xff, extIp.externalIP&0xff);
            }
            printf(",dvidx=%d", nxthop.dstVid);
            printf(",pppidx=%d", nxthop.PPPoEIndex);
            _rtl8651_readAsicEntry(TYPE_PPPOE_TABLE, nxthop.PPPoEIndex, &pppoe);
            printf("->%d", pppoe.sessionID);
            printf(",l2=%d", nxthop.nextHop);
            {
                rtl8651_tblAsic_l2Table_t l2;
                ether_addr_t mac;
            
                _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, nxthop.nextHop, &l2);
            		mac.octet[0] = l2.mac47_40;
            		mac.octet[1] = l2.mac39_24 >> 8;
            		mac.octet[2] = l2.mac39_24 & 0xff;
            		mac.octet[3] = l2.mac23_8 >> 8;
            		mac.octet[4] = l2.mac23_8 & 0xff;
            		mac.octet[5] = (nxthop.nextHop >> 2) ^ mac.octet[0] ^ mac.octet[1] ^ mac.octet[2] ^ mac.octet[3] ^ mac.octet[4];
                printf("->mac=%02x:%02x:%02x:%02x:%02x:%02x)",
                        mac.octet[0], mac.octet[1], mac.octet[2], 
                        mac.octet[3], mac.octet[4], mac.octet[5]);
            }
        }
    }
#endif /*CONFIG_RTL865XB*/

    printf("\n");
    return SUCCESS;
}

static int32 _parseIcmp(uint32 idx, uint32 force)
{
    rtl8651_tblAsic_naptIcmpTable_t entry;
    
    if ( idx >= RTL8651_ICMPTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_L4_ICMP_TABLE, idx, &entry);
    
    if (!force && !entry.valid)
        return FAILED;
    
    printf("[%d] ", idx);
    printf("valid=%d", entry.valid);
    printf(",intip=%d.%d.%d.%d", (entry.intIPAddr>>24)&0xff, (entry.intIPAddr>>16)&0xff
                        , (entry.intIPAddr>>8)&0xff, entry.intIPAddr&0xff);
    printf(",intprt=0x%x", (entry.ICMPIDH << 15) | (entry.ICMPIDL & 0x7fff));
    printf(",col=%d", entry.collision);
    printf(",age=0x%x", entry.agingTime);
    printf(",offset=%d", entry.offset);
    printf(",static=%d", entry.isStatic);
    
#ifdef CONFIG_RTL865XB
    printf(",type=%d", entry.type);
    if (entry.type==1)
        printf("(spi)");
    else if (entry.type>1)
        printf("(pptp)");
#else
    printf(",count=%d", entry.count);
#endif /*CONFIG_RTL865XB*/

    printf("\n");
    return SUCCESS;
}

static int32 _parseAcl(uint32 idx)
{
    rtl8651_tblAsic_aclTable_t entry;
    
    if ( idx >= RTL8651_ACLTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_ACL_RULE_TABLE, idx, &entry);
    
    printf("[%d] ", idx);
    printf("ruletype=%d", entry.ruleType);
    switch (entry.ruleType)
    {
        case RULE_ETHERNET:
            printf("(eth)");
            printf(",dmacp=%02x:%02x:%02x:%02x:%02x:%02x",
                        entry.is.ETHERNET.dMacP47_32>>8, entry.is.ETHERNET.dMacP47_32&0xff, entry.is.ETHERNET.dMacP31_16>>8, 
                        entry.is.ETHERNET.dMacP31_16&0xff, entry.is.ETHERNET.dMacP15_0>>8, entry.is.ETHERNET.dMacP15_0&0xff);
            printf(",dmacm=%02x:%02x:%02x:%02x:%02x:%02x",
                        entry.is.ETHERNET.dMacM47_32>>8, entry.is.ETHERNET.dMacM47_32&0xff, entry.is.ETHERNET.dMacM31_16>>8, 
                        entry.is.ETHERNET.dMacM31_16&0xff, entry.is.ETHERNET.dMacM15_0>>8, entry.is.ETHERNET.dMacM15_0&0xff);
            printf(",smacp=%02x:%02x:%02x:%02x:%02x:%02x",
                        entry.is.ETHERNET.sMacP47_32>>8, entry.is.ETHERNET.sMacP47_32&0xff, entry.is.ETHERNET.sMacP31_16>>8, 
                        entry.is.ETHERNET.sMacP31_16&0xff, entry.is.ETHERNET.sMacP15_0>>8, entry.is.ETHERNET.sMacP15_0&0xff);
            printf(",smacm=%02x:%02x:%02x:%02x:%02x:%02x",
                        entry.is.ETHERNET.sMacM47_32>>8, entry.is.ETHERNET.sMacM47_32&0xff, entry.is.ETHERNET.sMacM31_16>>8, 
                        entry.is.ETHERNET.sMacM31_16&0xff, entry.is.ETHERNET.sMacM15_0>>8, entry.is.ETHERNET.sMacM15_0&0xff);
            break;
        case RULE_IFSEL:
            printf("(ifsel)");
            printf(",gidxsel=%d", entry.is.IFSEL.gidxSel);
            break;
        case RULE_IP:
            printf("(ip)");
            printf(",iptosp=0x%x", entry.is.L3L4.is.IP.IPTOSP);
            printf(",iptosm=0x%x", entry.is.L3L4.is.IP.IPTOSM);
            printf(",ipprotop=0x%x", entry.is.L3L4.is.IP.IPProtoP);
            printf(",ipprotom=0x%x", entry.is.L3L4.is.IP.IPProtoM);
            printf(",ipflagp=0x%x", entry.is.L3L4.is.IP.IPFlagP);
            printf(",ipflagm=0x%x", entry.is.L3L4.is.IP.IPFlagM);
#ifdef CONFIG_RTL865XB
            printf(",FOP=%d", entry.is.L3L4.is.IP.FOP);
            printf(",FOM=%d", entry.is.L3L4.is.IP.FOM);
            printf(",HTTPP=%d", entry.is.L3L4.is.IP.HTTPP);
            printf(",HTTPP=%d", entry.is.L3L4.is.IP.HTTPP);
            printf(",idsipdipp=%d", entry.is.L3L4.is.IP.identSDIPP);
            printf(",idsipdipm=%d", entry.is.L3L4.is.IP.identSDIPM);
#endif /*CONFIG_RTL865XB*/
            goto l3l4_shared;
        case RULE_ICMP:
            printf("(icmp)");
            printf(",iptosp=0x%x", entry.is.L3L4.is.ICMP.IPTOSP);
            printf(",iptosm=0x%x", entry.is.L3L4.is.ICMP.IPTOSM);
            printf(",icmptypep=0x%x", entry.is.L3L4.is.ICMP.ICMPTypeP);
            printf(",icmptypem=0x%x", entry.is.L3L4.is.ICMP.ICMPTypeM);
            printf(",icmpcodep=0x%x", entry.is.L3L4.is.ICMP.ICMPCodeP);
            printf(",icmpcodem=0x%x", entry.is.L3L4.is.ICMP.ICMPCodeM);
            goto l3l4_shared;
        case RULE_IGMP:
            printf("(igmp)");
            printf(",iptosp=0x%x", entry.is.L3L4.is.IGMP.IPTOSP);
            printf(",iptosm=0x%x", entry.is.L3L4.is.IGMP.IPTOSM);
            printf(",igmptypep=0x%x", entry.is.L3L4.is.IGMP.IGMPTypeP);
            printf(",igmptypem=0x%x", entry.is.L3L4.is.IGMP.IGMPTypeM);
            goto l3l4_shared;
        case RULE_TCP:
            printf("(tcp)");
            printf(",iptosp=0x%x", entry.is.L3L4.is.TCP.IPTOSP);
            printf(",iptosm=0x%x", entry.is.L3L4.is.TCP.IPTOSM);
            printf(",tcpflagp=0x%x", entry.is.L3L4.is.TCP.TCPFlagP);
            printf(",tcpflagm=0x%x", entry.is.L3L4.is.TCP.TCPFlagM);
            printf(",tcpspub=%d", entry.is.L3L4.is.TCP.TCPSPUB);
            printf(",tcpsplb=%d", entry.is.L3L4.is.TCP.TCPSPLB);
            printf(",tcpdpub=%d", entry.is.L3L4.is.TCP.TCPDPUB);
            printf(",tcpdplb=%d", entry.is.L3L4.is.TCP.TCPDPLB);
            goto l3l4_shared;
        case RULE_UDP:
            printf("(udp)");
            printf(",iptosp=0x%x", entry.is.L3L4.is.UDP.IPTOSP);
            printf(",iptosm=0x%x", entry.is.L3L4.is.UDP.IPTOSM);
            printf(",udpspub=%d", entry.is.L3L4.is.UDP.UDPSPUB);
            printf(",udpsplb=%d", entry.is.L3L4.is.UDP.UDPSPLB);
            printf(",udpdpub=%d", entry.is.L3L4.is.UDP.UDPDPUB);
            printf(",udpdplb=%d", entry.is.L3L4.is.UDP.UDPDPLB);
            goto l3l4_shared;
            
#ifdef CONFIG_RTL865XB
        case RULE_SRC_FILTER:
            printf("(src-filter)");
            printf(",smacp=%02x:%02x:%02x:%02x:%02x:%02x",
                        entry.is.SRC_FILTER.sMacP47_32>>8, entry.is.SRC_FILTER.sMacP47_32&0xff, entry.is.SRC_FILTER.sMacP31_16>>8, 
                        entry.is.SRC_FILTER.sMacP31_16&0xff, entry.is.SRC_FILTER.sMacP15_0>>8, entry.is.SRC_FILTER.sMacP15_0&0xff);
            printf(",smacm=%02x:%02x:%02x:%02x:%02x:%02x",
                        entry.is.SRC_FILTER.sMacM47_32>>8, entry.is.SRC_FILTER.sMacM47_32&0xff, entry.is.SRC_FILTER.sMacM31_16>>8, 
                        entry.is.SRC_FILTER.sMacM31_16&0xff, entry.is.SRC_FILTER.sMacM15_0>>8, entry.is.SRC_FILTER.sMacM15_0&0xff);
            printf(",spap=0x%x", entry.is.SRC_FILTER.spaP);
            printf(",spam=0x%x", entry.is.SRC_FILTER.spaM);
            printf(",svidxp=%d", entry.is.SRC_FILTER.sVidxP);
            printf(",svidxm=%d", entry.is.SRC_FILTER.sVidxM);
            printf(",prototype=%d", entry.is.SRC_FILTER.protoType);
            printf(",sipp=%d.%d.%d.%d", (entry.is.SRC_FILTER.sIPP>>24)&0xff, (entry.is.SRC_FILTER.sIPP>>16)&0xff
                                , (entry.is.SRC_FILTER.sIPP>>8)&0xff, entry.is.SRC_FILTER.sIPP&0xff);
            printf(",sipm=%d.%d.%d.%d", (entry.is.SRC_FILTER.sIPM>>24)&0xff, (entry.is.SRC_FILTER.sIPM>>16)&0xff
                                , (entry.is.SRC_FILTER.sIPM>>8)&0xff, entry.is.SRC_FILTER.sIPM&0xff);
            printf(",sportub=%d", entry.is.SRC_FILTER.SPORTUB);
            printf(",sportlb=%d", entry.is.SRC_FILTER.SPORTLB);
            break;
            
        case RULE_DST_FILTER:
            printf("(dst-filter)");
            printf(",dmacp=%02x:%02x:%02x:%02x:%02x:%02x",
                        entry.is.DST_FILTER.dMacP47_32>>8, entry.is.DST_FILTER.dMacP47_32&0xff, entry.is.DST_FILTER.dMacP31_16>>8, 
                        entry.is.DST_FILTER.dMacP31_16&0xff, entry.is.DST_FILTER.dMacP15_0>>8, entry.is.DST_FILTER.dMacP15_0&0xff);
            printf(",dmacm=%02x:%02x:%02x:%02x:%02x:%02x",
                        entry.is.DST_FILTER.dMacM47_32>>8, entry.is.DST_FILTER.dMacM47_32&0xff, entry.is.DST_FILTER.dMacM31_16>>8, 
                        entry.is.DST_FILTER.dMacM31_16&0xff, entry.is.DST_FILTER.dMacM15_0>>8, entry.is.DST_FILTER.dMacM15_0&0xff);
            printf(",vidxp=%d", entry.is.DST_FILTER.vidxP);
            printf(",vidxm=%d", entry.is.DST_FILTER.vidxM);
            printf(",prototype=%d", entry.is.DST_FILTER.protoType);
            printf(",dipp=%d.%d.%d.%d", (entry.is.DST_FILTER.dIPP>>24)&0xff, (entry.is.DST_FILTER.dIPP>>16)&0xff
                                , (entry.is.DST_FILTER.dIPP>>8)&0xff, entry.is.DST_FILTER.dIPP&0xff);
            printf(",dipm=%d.%d.%d.%d", (entry.is.DST_FILTER.dIPM>>24)&0xff, (entry.is.DST_FILTER.dIPM>>16)&0xff
                                , (entry.is.DST_FILTER.dIPM>>8)&0xff, entry.is.DST_FILTER.dIPM&0xff);
            printf(",dportub=%d", entry.is.DST_FILTER.DPORTUB);
            printf(",dportlb=%d", entry.is.DST_FILTER.DPORTLB);
            break;
#endif /*CONFIG_RTL865XB*/
            
l3l4_shared:
            printf(",sipp=%d.%d.%d.%d", (entry.is.L3L4.sIPP>>24)&0xff, (entry.is.L3L4.sIPP>>16)&0xff
                                , (entry.is.L3L4.sIPP>>8)&0xff, entry.is.L3L4.sIPP&0xff);
            printf(",sipm=%d.%d.%d.%d", (entry.is.L3L4.sIPM>>24)&0xff, (entry.is.L3L4.sIPM>>16)&0xff
                                , (entry.is.L3L4.sIPM>>8)&0xff, entry.is.L3L4.sIPM&0xff);
            printf(",dipp=%d.%d.%d.%d", (entry.is.L3L4.dIPP>>24)&0xff, (entry.is.L3L4.dIPP>>16)&0xff
                                , (entry.is.L3L4.dIPP>>8)&0xff, entry.is.L3L4.dIPP&0xff);
            printf(",dipm=%d.%d.%d.%d", (entry.is.L3L4.dIPM>>24)&0xff, (entry.is.L3L4.dIPM>>16)&0xff
                                , (entry.is.L3L4.dIPM>>8)&0xff, entry.is.L3L4.dIPM&0xff);
            break;
    }
    
    printf(",actiontype=%d", entry.actionType);
    switch (entry.actionType)
    {
        case ACTION_PERMIT:
            printf("(permit)");
            goto common_act;
        case ACTION_REDIRECT:
            printf("(redirect)");
            goto common_act;
        case ACTION_MIRROR:
            printf("(mirror)");
            goto common_act;
        case ACTION_PPPOE_REDIRECT:
            printf("(pppoe-redirect)");
            goto common_act;
        case ACTION_S_DROP:
            printf("(sdrop)");
            goto common_act;
        case ACTION_S_CPU:
            printf("(scpu)");
            goto common_act;
        case ACTION_N_DROP:
            printf("(ndrop)");
            goto common_act;
        case ACTION_N_CPU:
#ifndef CONFIG_RTL865XB
            printf("(ncpu)");
#else
            printf("(drop-to-cpu-for-log)");
#endif
            goto common_act;
#ifdef CONFIG_RTL865XB
        case ACTION_DEFAULT_REDIRECT:
            printf("(default-redirect)");
            printf(",nxthop=%d", entry.nextHop & 0x1f);
            printf("\n  nxthop:");
            _parseNextHop(entry.nextHop & 0x1f);
            break;
        case ACTION_DEFAULT_MIRROR:
            printf("(default-mirror)");
            goto common_act;
        case ACTION_DROP_RATE_PPS:
            printf("(drop-if-rate-exceed-pps)");
            goto rate_act;
        case ACTION_LOG2CPU_RATE_PPS:
            printf("(log-to-cpu-if-rate-exceed-pps)");
            goto rate_act;
        case ACTION_DROP_RATE_BPS:
            printf("(drop-if-rate-exceed-bps)");
            goto rate_act;
        case ACTION_LOG2CPU_RATE_BPS:
            printf("(log-to-cpu-if-rate-exceed-bps)");
            goto rate_act;
#endif /*CONFIG_RTL865XB*/

common_act:
        printf(",hp=%d", entry.hPriority);
        printf(",dvidx=%d", entry.vid);
        printf(",nxthop=%d", entry.nextHop);
        if ((entry.actionType == ACTION_REDIRECT) || (entry.actionType == ACTION_PPPOE_REDIRECT))
        {
            rtl8651_tblAsic_l2Table_t l2;
            ether_addr_t mac;
        
            _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, entry.nextHop, &l2);
        		mac.octet[0] = l2.mac47_40;
        		mac.octet[1] = l2.mac39_24 >> 8;
        		mac.octet[2] = l2.mac39_24 & 0xff;
        		mac.octet[3] = l2.mac23_8 >> 8;
        		mac.octet[4] = l2.mac23_8 & 0xff;
        		mac.octet[5] = (entry.nextHop >> 2) ^ mac.octet[0] ^ mac.octet[1] ^ mac.octet[2] ^ mac.octet[3] ^ mac.octet[4];
            printf("(mac=%02x:%02x:%02x:%02x:%02x:%02x)",
                    mac.octet[0], mac.octet[1], mac.octet[2], 
                    mac.octet[3], mac.octet[4], mac.octet[5]);
        }
        printf(",pppidx=%d", entry.PPPoEIndex);
        if (entry.actionType == ACTION_PPPOE_REDIRECT)
        {
            rtl8651_tblAsic_pppoeTable_t pppoe;
            
            _rtl8651_readAsicEntry(TYPE_PPPOE_TABLE, entry.PPPoEIndex, &pppoe);
            printf("(session=%d)", pppoe.sessionID);
        }
        break;
        
#ifdef CONFIG_RTL865XB
rate_act:
        printf(",ratelimit=%d", entry.nextHop & 0x1f);
        printf("\n  ratelimit:");
        _parseRateLimit(entry.nextHop & 0x1f);
        break;
#endif /*CONFIG_RTL865XB*/
}

#ifdef CONFIG_RTL865XB
    printf(",pktopapp=%d", entry.pktOpApp);
#endif /*CONFIG_RTL865XB*/
    

    printf("\n");
    return SUCCESS;
}

#ifdef CONFIG_RTL865XB
static int32 _parseNextHop(uint32 idx)
{
    rtl8651_tblAsic_nextHopTable_t entry;
    
    if ( idx >= RTL8651_NXTHOPTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_NEXT_HOP_TABLE, idx, &entry);
    
    printf("[%d] ", idx);
    printf("type=%d", entry.type);
    if (entry.type == 0)
        printf("(eth)");
    else if (entry.type == 1)
        printf("(pppoe)");
    printf(",ipidx=%d", entry.IPIndex);
    {
        rtl8651_tblAsic_extIpTable_t extIp;
        
        _rtl8651_readAsicEntry(TYPE_EXT_INT_IP_TABLE, entry.IPIndex, &extIp);
        printf("(extip=%d.%d.%d.%d)", (extIp.externalIP>>24)&0xff, (extIp.externalIP>>16)&0xff
                            , (extIp.externalIP>>8)&0xff, extIp.externalIP&0xff);
    }
    printf(",dvidx=%d", entry.dstVid);
    printf(",pppidx=%d", entry.PPPoEIndex);
    {
        rtl8651_tblAsic_pppoeTable_t pppoe;
        
        _rtl8651_readAsicEntry(TYPE_PPPOE_TABLE, entry.PPPoEIndex, &pppoe);
        printf("(session=%d)", pppoe.sessionID);
    }
    printf(",nxthop=%d", entry.nextHop);
    {
        rtl8651_tblAsic_l2Table_t l2;
        ether_addr_t mac;
    
        _rtl8651_readAsicEntry(TYPE_L2_SWITCH_TABLE, entry.nextHop, &l2);
    		mac.octet[0] = l2.mac47_40;
    		mac.octet[1] = l2.mac39_24 >> 8;
    		mac.octet[2] = l2.mac39_24 & 0xff;
    		mac.octet[3] = l2.mac23_8 >> 8;
    		mac.octet[4] = l2.mac23_8 & 0xff;
    		mac.octet[5] = (entry.nextHop >> 2) ^ mac.octet[0] ^ mac.octet[1] ^ mac.octet[2] ^ mac.octet[3] ^ mac.octet[4];
        printf("(mac=%02x:%02x:%02x:%02x:%02x:%02x)",
                mac.octet[0], mac.octet[1], mac.octet[2], 
                mac.octet[3], mac.octet[4], mac.octet[5]);
    }

    printf("\n");
    return SUCCESS;
}

static int32 _parseRateLimit(uint32 idx)
{
    rtl8651_tblAsic_rateLimitTable_t entry;
    
    if ( idx >= RTL8651_RATELIMITTBL_SIZE )
    {
        printf("Invalid index\n");
        return FAILED;
    }
    
    _rtl8651_readAsicEntry(TYPE_RATE_LIMIT_TABLE, idx, &entry);
    
    printf("[%d] ", idx);
    printf("token=0x%x", entry.token);
    printf(",maxtoken=0x%x", entry.maxToken);
    printf(",refill=0x%x", entry.refill);
    printf(",refilltime=%d", entry.refillTime);
    printf(",timeremain=%d", entry.refillRemainTime);

    printf("\n");
    return SUCCESS;
}
#endif /*CONFIG_RTL865XB*/

void tabparse_main(void* arg)
{
    struct {int argc; char **argv;} *p = arg;
    int argc = p->argc;
    char **argv = p->argv;
    uint32 start=0, end=0, i;
    uint32 nxttoken;
    int8 all=0, force=0;
    
    if (argc < 2)
    {
        printf("Usage: tabparse [-af] TYPE [INDEX [INDEX]]\n");
        printf("-f     Force. Parse even invalid entries.\n");
        printf("-a     All. Display all entries.\n");
        printf("TYPE   Table type. Options are <l2|arp|l3|mc|pt|vlan|ip|alg|srv|tcp|icmp|ppp|acl|nxthop|rate>.\n");
        printf("INDEX  Table index. Give the second index to specify a range. If -a is given, the index is neglected\n");
        return;
    }

    nxttoken = 1;
    if (argv[1][0] == '-')
    {
        for (i=1;i<strlen(argv[1]);i++)
            if (argv[1][i] == 'a')
                all = 1;
            else if (argv[1][i] == 'f')
                force = 1;
            else
            {
                printf("Unknown option.\n");
                return;
            }
        nxttoken = 2;
    }

    if (argc == (nxttoken+1))
        start = end = 0;
    else if (argc == (nxttoken+2))
        start = end = stoi(argv[nxttoken+1]);
    else
    {
        start = stoi(argv[nxttoken+1]);
        end = stoi(argv[nxttoken+2]);
        if (end < start)
        {
            printf("Invalid range.\n");
            return;
        }
    }

    if (all)
        start = 0;
    if (strncmp(argv[nxttoken], "l2", sizeof("l2")) == 0)
    {
        if (all)
            end = RTL8651_MACTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseL2(i, force);
    }
    else if (strncmp(argv[nxttoken], "arp", sizeof("arp")) == 0)
    {
        if (all)
            end = RTL8651_ARPTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseArp(i, force);
    }
    else if (strncmp(argv[nxttoken], "l3", sizeof("l3")) == 0)
    {
        if (all)
            end = RTL8651_ROUTINGTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseL3(i, force);
    }
    else if (strncmp(argv[nxttoken], "mc", sizeof("mc")) == 0)
    {
        if (all)
            end = RTL8651_MULTICASTTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseMcast(i, force);
    }
    else if (strncmp(argv[nxttoken], "pt", sizeof("pt")) == 0)
    {
        if (all)
            end = RTL8651_PROTOTRAPTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseProtoTrap(i, force);
    }
    else if (strncmp(argv[nxttoken], "vlan", sizeof("vlan")) == 0)
    {
        if (all)
            end = RTL8651_VLANTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseVlan(i, force);
    }
    else if (strncmp(argv[nxttoken], "ip", sizeof("ip")) == 0)
    {
        if (all)
            end = RTL8651_IPTABLE_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseExtIntIp(i, force);
    }
    else if (strncmp(argv[nxttoken], "alg", sizeof("alg")) == 0)
    {
        if (all)
            end = RTL8651_ALGTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseAlg(i, force);
    }
    else if (strncmp(argv[nxttoken], "srv", sizeof("srv")) == 0)
    {
        if (all)
            end = RTL8651_SERVERPORTTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseServerPort(i, force);
    }
    else if (strncmp(argv[nxttoken], "tcp", sizeof("tcp")) == 0)
    {
        if (all)
            end = RTL8651_TCPUDPTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseTcpUdp(i, force);
    }
    else if (strncmp(argv[nxttoken], "icmp", sizeof("icmp")) == 0)
    {
        if (all)
            end = RTL8651_ICMPTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseIcmp(i, force);
    }
    else if (strncmp(argv[nxttoken], "ppp", sizeof("ppp")) == 0)
    {
        if (all)
            end = RTL8651_PPPOETBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parsePPPoE(i);
    }
    else if (strncmp(argv[nxttoken], "acl", sizeof("acl")) == 0)
    {
        if (all)
            end = RTL8651_ACLTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseAcl(i);
    }
#ifdef CONFIG_RTL865XB
    else if (strncmp(argv[nxttoken], "nxthop", sizeof("nxthop")) == 0)
    {
        if (all)
            end = RTL8651_NXTHOPTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseNextHop(i);
    }
    else if (strncmp(argv[nxttoken], "rate", sizeof("rate")) == 0)
    {
        if (all)
            end = RTL8651_RATELIMITTBL_SIZE - 1;
        for (i=start; i<=end; i++)
            _parseRateLimit(i);
    }
#endif //CONFIG_RTL865XB
    else
        printf("Unknown table type.\n");
    
    return;
}
