#include "../types.h"
#include "../asicRegs.h"
#include "../hsRegs.h"
#include "../rtl8651_tblAsicDrv.h"
#include "veri.h"

/* Global Variables 
*/
rtl865x_tblAsicDrv_vlanParam_t vlant;
_rtl8651_tblDrvAclRule_t aclt;
rtl865x_tblAsicDrv_l2Param_t l2t;
rtl865x_tblAsicDrv_pppoeParam_t pppoet;
rtl865x_tblAsicDrv_arpParam_t arpt;
rtl865x_tblAsicDrv_routingParam_t routet;
rtl865x_tblAsicDrv_extIntIpParam_t ipt;
rtl865x_tblAsicDrv_serverPortParam_t srvt;
rtl865x_tblAsicDrv_algParam_t algt;
rtl865x_tblAsicDrv_nextHopParam_t nxthpt;

/* Static Variables
*/
static int quiet;

/* Static Routines 
*/
static void setQuiet(int q)
{
    quiet = q;
}

int veri_quiet(void)
{
    return quiet;
}

void veri_main(void* arg)
{
    struct {int argc; char **argv;} *p = arg;
    int argc = p->argc;
    char **argv = p->argv;
    uint32 token;
    static int8 asic_inited=0;
    uint32 errCount=0;
    uint32 i;
    struct case_desc_s {char *caseString; char *descString;} cases[] = {
        {"start", "Enter test environment"},
        {"end", "Leave test environment"},
        {"l34", "8650-compatible L3/4 test"},
        {"napt4WayHash", "NAPT TCP/UDP table 4-way hash test"},
        {"routCentric", "Route-centric architecture test, including next hop table"},
        {"hs", "Dump header stamps for debugging"}};
#define CASE_NUM (sizeof(cases)/sizeof(struct case_desc_s))
    
    if (argc < 2)
    {
        printf("Usage: veri -q CASE\n");
        printf("CASE:\n");
        for (i=0; i<CASE_NUM; i++)
            printf("  %s: %s\n", cases[i].caseString, cases[i].descString);
        return;
    }
    
    if (!strncmp(argv[1],"-q",2))
    {
        setQuiet(1);
        token = 2;
    }
    else
    {
        setQuiet(0);
        token = 1;
    }

    if (!strncmp(argv[token],cases[0].caseString,sizeof(cases[0].caseString))) //start
    {
        if (!asic_inited)
        {
            rtl8651_initAsic();
            rtl8651_clearAsicAllTable();
            swNicPoll_init(128,128,16,2048);
            asic_inited = 1;
        }
        miiLikeTestStart();
    }
    else if (!strncmp(argv[token],cases[1].caseString,sizeof(cases[1].caseString))) //end
        miiLikeTestEnd();
    else if (!strncmp(argv[token],cases[2].caseString,sizeof(cases[2].caseString))) //l34
    {
#define NUM_OF_TOPOLOGIES 3
        uint8 port_number_of_host[NUM_OF_TOPOLOGIES][MAX_PORT_NUMBER] = {{0,1,2,3,4,5},{4,3,2,1,0,5},{1,0,3,2,4,5}};
        uint32 tplg = 0;
        while (tplg < NUM_OF_TOPOLOGIES )
        {
            l34_tplg(port_number_of_host[tplg++]);
            errCount += l34_testNatThruEth();
            errCount += l34_testNatThruPppoe();
            errCount += l34_testLanRoute();
            errCount += l34_testNaptThruEth();
            errCount += l34_testNaptThruPppoe();
            errCount += l34_testLanRouteStripVlan();
            errCount += l34_testNaptThruEthStripVlan();
            errCount += l34_testNaptThruPppoeStripVlan();
            errCount += l34_testSrvPrtThruEth();
            errCount += l34_testSrvPrtThruPppoe();
            errCount += l34_testWanRoute();
            errCount += l34_testAlgTrap();
        }
        if (errCount == 0)
            printf("PASS!\n");
        else
            printf("FAIL! Total Error %d\n", errCount);
    }
    else if (!strncmp(argv[token],cases[3].caseString,sizeof(cases[3].caseString))) //napt4WayHash
    {
        napt4WayHash_tplg();
        errCount += napt4WayHash_test1();
        if (errCount == 0)
            printf("PASS!\n");
        else
            printf("FAIL! Total Error %d\n", errCount);
    }
    else if (!strncmp(argv[token],cases[4].caseString,sizeof(cases[4].caseString))) //routCentric
    {
        routCentric_tplg();
        errCount += routCentric_testNatNextHop();
        errCount += routCentric_testSrvPrtNextHop();
        errCount += routCentric_testNaptLoadBalance();
        errCount += routCentric_testSrvPrtRange();
        errCount += routCentric_testNaptStaticHash1();
        errCount += routCentric_testNaptHash2();
        errCount += routCentric_testNaptHash1Vs2();
        errCount += routCentric_testDefaultRedirectNextHop();
        if (errCount == 0)
            printf("PASS!\n");
        else
            printf("FAIL! Total Error %d\n", errCount);
    }
    else if (!strncmp(argv[token],cases[CASE_NUM-1].caseString,sizeof(cases[CASE_NUM-1].caseString))) //hs
    {
        hs_dumpHsb(1);
        hs_dumpHsa(1);
    }
    else
        printf("Unknown case!\n");
    
    return;
}
