#include "pktgen/pg_flagDef.h"
#include "pktgen/pg_pktForm.h"

#ifdef __linux__
#define printf printk
#endif

#define CHK_HDR_STAMP_DEC(a,b,errCount) \
if((a)!=(b)) { \
    printf("Header Stamp Error! %s=%d (expected %d)\n", #a, a, b); \
    ++(errCount); \
}
#define CHK_HDR_STAMP_HEX(a,b,errCount) \
if((a)!=(b)) { \
    printf("Header Stamp Error! %s=0x%x (expected 0x%x)\n", #a, a, b); \
    ++(errCount); \
}
#define CHK_HDR_STAMP_MAC(a,b,errCount) \
if(cmpMac((ether_addr_t*)&a, b)) { \
    printf("Header Stamp Error! %s=%02x-%02x-%02x-%02x-%02x-%02x (expected %s)\n",#a,(a)[0],(a)[1],(a)[2],(a)[3],(a)[4],(a)[5],b); \
    ++(errCount); \
}
#define CHK_HDR_STAMP_IP(a,b,errCount) \
if(cmpIp(a, b)) { \
    printf("Header Stamp Error! %s=%d.%d.%d.%d (expected %s)\n",#a,(a)[0],(a)[1],(a)[2],(a)[3],b); \
    ++(errCount); \
}
#define CHK_PKT_DEC(a,b,errCount) \
if((a)!=(b)) { \
    printf("Packet Error! %s=%d (expected %d)\n", #a, a, b); \
    ++(errCount); \
}
#define CHK_PKT_HEX(a,b,errCount) \
if((a)!=(b)) { \
    printf("Packet Error! %s=0x%x (expected 0x%x)\n", #a, a, b); \
    ++(errCount); \
}
#define CHK_PKT_MAC(a,b,errCount) \
if(cmpMac((ether_addr_t*)(a).octet, b)) { \
    printf("Packet Error! %s=%02x-%02x-%02x-%02x-%02x-%02x (expected %s)\n",#a,(a).octet[0],(a).octet[1],(a).octet[2],(a).octet[3],(a).octet[4],(a).octet[5],b); \
    ++(errCount); \
}
#define CHK_PKT_IP(a,b,errCount) \
if(cmpIp2(a, b)) { \
    printf("Packet Error! %s=%d.%d.%d.%d (expected %s)\n",#a,((a)>>24)&0xff,((a)>>16)&0xff,((a)>>8)&0xff,(a)&0xff,b); \
    ++(errCount); \
}
#define VERI_MSG(format) \
{ \
if (!veri_quiet()) \
    printf format; \
}

//uint32 READ_MEM32(uint32 addr);
//void WRITE_MEM32(uint32 addr, uint32 value);
void miiLikeTestStart();
void miiLikeTestEnd();
/* pktLen does not include CRC */
void miiLikeTestSend(uint32 portNum, uint8 *pktBuf, uint32 pktLen);
/* pktLen: input - size of pktBuf, output - packet length (excluding CRC) */
void miiLikeTestReceive(uint32 portNum, uint8 *pktBuf, uint32 *pktLen);
typedef struct nicPktHdrFormat_s {
    uint16 ph_len; //packet length, excluding CRC
    uint16 ph_reason;
    uint32 csumIp:1, //rx: ip chksum ok, tx: offload ip chksum calculation
           csumL4:1, //rx: icmp/tcp/udp chksum ok, tx: offload icmp/tcp/udp chksum calculation
           ph_vlanTagged:1, //rx: tag attached, tx: offload attachment
           ph_LLCTagged:1, //rx: tag attached
           ph_pppeTagged:1, //rx: header attached, tx: offload attachment
           hwLookUp:1,
           bridgingOnly:1,
           resrv:25;
    uint16 ph_vlanIdx;
    uint16 ph_pppoeIdx;
    uint16 ph_proto;
#define PKTHDR_ETHERNET      0
#define PKTHDR_PPTP          1
#define PKTHDR_IP            2
#define PKTHDR_ICMP          3
#define PKTHDR_IGMP          4
#define PKTHDR_TCP           5
#define PKTHDR_UDP           6
    uint16 ph_extPortList; //dest extension port list
    uint16 ph_srcExtPortNum; //source extension port number
    uint16 ph_portlist; //rx: source port number, tx: destination portmask
} nicPktHdrFormat_t;
void nicSend(nicPktHdrFormat_t *pktHdrt, uint8 *pktBuf);
/* pktLen: input - size of pktBuf, output - packet length (excluding CRC) */
void nicReceive(nicPktHdrFormat_t *pktHdrt, uint8 *pktBuf, uint32 *pktLen);
void readHsb(hsb_param_watch_t *hsbp);
void readHsa(hsa_param_watch_t *hsap);
int32 cmpMac(ether_addr_t *mac, uint8 *str);
int32 cmpIp(uint8 *ip, uint8 *str);
int32 cmpIp2(ipaddr_t ip, uint8 *str);
typedef struct pktGenArpFormat_s {
    ether_addr_t dmac;
    ether_addr_t smac;
    uint32 vlanTag:1,
           resrv:31;
    uint32 vid; //vlan id
    ipaddr_t dip;
    ipaddr_t sip;
    uint32 op; //opcode
#define	ARPOP_REQUEST	1
#define	ARPOP_REPLY	2
#define	ARPOP_REVREQUEST 3
#define	ARPOP_REVREPLY 4
} pktGenArpFormat_t;
uint32 pktGenArp(uint8 *pktBuf, pktGenArpFormat_t *arpt);
uint32 pktParseArp(uint8 *pktBuf, pktGenArpFormat_t *arpt, uint32 pktLen);
typedef struct pktGenIpFormat_s {
    ether_addr_t dmac;
    ether_addr_t smac;
    uint32 vlanTag:1,
           pppoeTag:1,
           resrv:30;
    uint32 vid; //vlan id
    uint32 sid; //pppoe session id
    ipaddr_t dip;
    ipaddr_t sip;
    uint32 proto; //protocol
    uint32 padLen; //padding length
} pktGenIpFormat_t;
uint32 pktGenIp(uint8 *pktBuf, pktGenIpFormat_t *ipt);
uint32 pktParseIp(uint8 *pktBuf, pktGenIpFormat_t *ipt, uint32 pktLen);
typedef struct pktGenIcmpFormat_s {
    ether_addr_t dmac;
    ether_addr_t smac;
    uint32 vlanTag:1,
           pppoeTag:1,
           resrv:30;
    uint32 vid; //vlan id
    uint32 sid; //pppoe session id
    ipaddr_t dip;
    ipaddr_t sip;
    uint32 type; //type
#define ICMPTYPE_ECHO_REQUEST   8
#define ICMPTYPE_ECHO_REPLY     0
    uint32 code; //code
    uint32 id; //id
    uint32 seq; //seq
} pktGenIcmpFormat_t;
uint32 pktGenIcmp(uint8 *pktBuf, pktGenIcmpFormat_t *icmpt);
uint32 pktParseIcmp(uint8 *pktBuf, pktGenIcmpFormat_t *icmpt, uint32 pktLen);

typedef struct pktGenIgmpFormat_s {
    ether_addr_t dmac;
    ether_addr_t smac;
    uint32 vlanTag:1,
           pppoeTag:1,
           resrv:30;
    uint32 vid; //vlan id
    uint32 sid; //pppoe session id
    ipaddr_t dip;
    ipaddr_t sip;
    uint32 dport; //L4 destination port
    uint32 sport; //L4 source port
    uint32 type; //type
    uint32 respTime; //respTime
    uint32 gaddr; //gaddr
} pktGenIgmpFormat_t;
uint32 pktGenIgmp(uint8 *pktBuf, pktGenIgmpFormat_t *igmpt);
uint32 pktParseIgmp(uint8 *pktBuf, pktGenIgmpFormat_t *igmpt, uint32 pktLen);
typedef struct pktGenTcpFormat_s {
    ether_addr_t dmac;
    ether_addr_t smac;
    uint32 vlanTag:1,
           pppoeTag:1,
           resrv:30;
    uint32 vid; //vlan id
    uint32 sid; //pppoe session id
    ipaddr_t dip;
    ipaddr_t sip;
    uint32 dport; //L4 destination port
    uint32 sport; //L4 source port
    uint32 flag; //TCP flag
#define	TCPFLAG_FIN	0x01
#define	TCPFLAG_SYN	0x02
#define	TCPFLAG_RST	0x04
#define	TCPFLAG_PUSH 0x08
#define	TCPFLAG_ACK	0x10
#define	TCPFLAG_URG	0x20
    uint32 padLen; //padding length
} pktGenTcpFormat_t;
uint32 pktGenTcp(uint8 *pktBuf, pktGenTcpFormat_t *tcpt);
uint32 pktParseTcp(uint8 *pktBuf, pktGenTcpFormat_t *tcpt, uint32 pktLen);
typedef struct pktGenUdpFormat_s {
    ether_addr_t dmac;
    ether_addr_t smac;
    uint32 vlanTag:1,
           pppoeTag:1,
           resrv:30;
    uint32 vid; //vlan id
    uint32 sid; //pppoe session id
    ipaddr_t dip;
    ipaddr_t sip;
    uint32 dport; //L4 destination port
    uint32 sport; //L4 source port
    uint32 padLen; //padding length
} pktGenUdpFormat_t;
uint32 pktGenUdp(uint8 *pktBuf, pktGenUdpFormat_t *udpt);
uint32 pktParseUdp(uint8 *pktBuf, pktGenUdpFormat_t *udpt, uint32 pktLen);
uint32 pktGenSetVlanCfi(uint8 *pktBuf, uint32 pktLen); //set vlan tag cfi = 1
uint32 pktGenSetIpVerHdrLen(uint8 *pktBuf, uint16 ipVerHdrLen, uint32 pktLen); //ip version together with header length
void strtomac(ether_addr_t *mac, uint8 *str);
void strtoip(ipaddr_t *ip, uint8 *str);
uint32 getL2TableRowFromMacStr(uint8 *str);

// General Packet Format

// Layer 2

typedef struct pktGenGeneralPktFormatL2ether_s
{
	uint32	l2Flag;
	ether_addr_t	dmac;
	ether_addr_t	smac;
	uint16	etherType;
} pktGenGeneralPktFormatL2ether_t;

typedef struct pktGenGeneralPktFormatL2llc_s
{
	uint8	dsap;
	uint8	ssap;
} pktGenGeneralPktFormatL2llc_t;

typedef struct pktGenGeneralPktFormatL2vlan_s
{
	uint16	vid;
	uint8	vlanCfi;
	uint16	priority;
} pktGenGeneralPktFormatL2vlan_t;

typedef struct pktGenGeneralPktFormatL2pppoe_s
{
	uint16	pppoeType;
	uint16	pppoeSession;
} pktGenGeneralPktFormatL2pppoe_t;

// Layer 3

typedef struct pktGenGeneralPktFormatL3ipx_s
{
	uint8	ipxtype;
	uint32	dnet;
	uint8	dnode[6];
	uint16	dsock;
	uint32	snet;
	uint8	snode[6];
	uint16	ssock;
} pktGenGeneralPktFormatL3ipx_t;


typedef struct pktGenGeneralPktFormatL3ip_s
{
	ipaddr_t	sip;
	ipaddr_t	dip;
	uint8	ipv;
	uint8	iphl;
	uint8	tos;
	uint16	ipid;
	uint8	df;
	uint8	mf;
	uint32	ipoff;
	uint8	ttl;
	int8	proto;
	uint32	errorFlag;
} pktGenGeneralPktFormatL3ip_t;

typedef struct pktGenGeneralPktFormatL3arp_s
{
	uint16	op;
	ether_addr_t	sha;
	ipaddr_t		spa;
	ether_addr_t	tha;
	ipaddr_t		tpa;
} pktGenGeneralPktFormatL3arp_t;

// Layer 4
typedef struct pktGenGeneralPktFormatL4pptp_s
{
	uint16	callid;
	uint32	seqno;
	uint32	ackno;
} pktGenGeneralPktFormatL4pptp_t;

typedef struct pktGenGeneralPktFormatL4icmp_s
{
	uint8	type;
	uint8	code;
	uint16	id;
	uint16	seq;
} pktGenGeneralPktFormatL4icmp_t;

typedef struct pktGenGeneralPktFormatL4igmp_s
{
	uint8	type;
	uint8	respTime;
	uint32	gaddr;
} pktGenGeneralPktFormatL4igmp_t;

typedef struct pktGenGeneralPktFormatL4TcpUdp_s
{
	uint16	sport;
	uint16	dport;
	uint32	tcp_seq;
	uint32	tcp_ack;
	uint8	tcpFlag;
	uint16	tcp_win;
} pktGenGeneralPktFormatL4TcpUdp_t;

	#define		U_PROTO_ETHER		0
	#define		U_PROTO_IPX		1
	#define		U_PROTO_ARP		2
	#define		U_PROTO_IP			3
	#define		U_PROTO_ICMP		4
	#define		U_PROTO_IGMP		5
	#define		U_PROTO_TCP		6
	#define		U_PROTO_UDP		7
	#define		U_PROTO_PPTP		8
	
typedef struct pktGenGeneralPktFormat_s {
	int					protocol;
// Layer2
	pktGenGeneralPktFormatL2ether_t	l2ether;
	pktGenGeneralPktFormatL2llc_t		l2llc;
	pktGenGeneralPktFormatL2vlan_t		l2vlan;
	pktGenGeneralPktFormatL2pppoe_t	l2pppoe;
	// Layer 3
	pktGenGeneralPktFormatL3ipx_t		l3ipx;
	pktGenGeneralPktFormatL3ip_t		l3ip;
	pktGenGeneralPktFormatL3arp_t		l3arp;
	// Layer 4
	pktGenGeneralPktFormatL4pptp_t		l4pptp;
	pktGenGeneralPktFormatL4icmp_t		l4icmp;
	pktGenGeneralPktFormatL4igmp_t	l4igmp;
	pktGenGeneralPktFormatL4TcpUdp_t	l4tcpudp;
	// payload
	uint8 				content[2048];
	uint32 				len;
	uint32				ipDataLen; // Only for parsing packet
} pktGenGeneralPktFormat_t;

uint32	pktGenGeneral(uint8 *pktBuf, pktGenGeneralPktFormat_t *pktt);
uint32	pktGenGeneralDefaultPacket(pktGenGeneralPktFormat_t *pktt);
uint32 pktParseGeneral(uint8 *pktBuf, pktGenGeneralPktFormat_t *pktt, uint32 pktLen);

