/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : VoIP Supporting in ROME Driver
* Abstract : 
* $Id: voip_support.c,v 1.1.1.1 2006/07/26 05:38:26 blue Exp $
*
*/

#include "rtl_types.h"
#include "rtl_queue.h"
#include "rtl8651_tblAsicDrv.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvFwd.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "voip_support.h"
#include "asicRegs.h"
#ifdef CONFIG_RTL865X_PCM
#include "../voip/pcm2.h"
#endif


#undef _VOIP_DEBUG_


typedef SLIST_HEAD( , voip_callback_s) ListCallback;
ListCallback inuseCallback;
ListCallback freeCallback;
typedef SLIST_HEAD( , voip_rtp_port_s) ListRtp;
ListRtp inuseRtp;
ListRtp freeRtp;


/*
@func int32 | voip_init | Initialize
@parm int32 | numChannel | Number of channel allocated at init.
@parm int32 | numRtp | Number of RTP information allocated at init.
@rvalue SUCCESS	| Initialization success
@rvalue FAILED | Initialization failure
@comm 
@devnote
*/
int32 voip_init( int32 numChannel, int32 numRtp )
{
	voip_callback_t *spoolCallback;
	voip_rtp_port_t *spoolRtp;
	int i;
	
	SLIST_INIT( &inuseCallback );
	SLIST_INIT( &freeCallback );
	SLIST_INIT( &inuseRtp );
	SLIST_INIT( &freeRtp );
	
	TBL_MEM_ALLOC( spoolCallback, voip_callback_t, numChannel );
	for( i = 0; i<numChannel; i++ )
		SLIST_INSERT_HEAD( &freeCallback, &spoolCallback[i], nextCallback );

	TBL_MEM_ALLOC( spoolRtp, voip_rtp_port_t, numRtp );
	for( i = 0; i<numRtp; i++ )
		SLIST_INSERT_HEAD( &freeRtp, &spoolRtp[i], nextRtp );
	
	return SUCCESS;
}


/*
@func int32 | voip_reinit | Collect entries
@rvalue SUCCESS	| Initialization success
@rvalue FAILED | Initialization failure
@comm 
@devnote
*/
int32 voip_collect( void )
{
	voip_callback_t *one;
	voip_rtp_port_t *two;
	
	while ( ( one = SLIST_FIRST( &inuseCallback ) ) )
 	{
		SLIST_REMOVE( &inuseCallback, one, voip_callback_s, nextCallback );
		SLIST_INSERT_HEAD( &freeCallback, one, nextCallback );
	}

	while ( ( two = SLIST_FIRST( &inuseRtp ) ) )
 	{
		SLIST_REMOVE( &inuseRtp, two, voip_rtp_port_s, nextRtp );
		SLIST_INSERT_HEAD( &freeRtp, two, nextRtp );
	}

	SLIST_INIT( &inuseCallback );
	SLIST_INIT( &inuseRtp );
	
	return SUCCESS;
}


/*
@func int32 | voip_register_callback | Register callback functions.
@parm uint32 | chid | Channel ID
@parm voip_callback_t | *pReg | Functions and parameters to register.
@rvalue SUCCESS	| Register success
@rvalue FAILED | Register failure
@comm 
@devnote
*/
int32 voip_register_callback( uint32 chid, voip_callback_t *pReg )
{
	voip_callback_t *one;
	
	for( one = SLIST_FIRST( &inuseCallback );
	     one;
 	     one = SLIST_NEXT( one, nextCallback ) )
 	{
 		if ( one->chid == chid )
 		goto found;
 	}

 	/* not found the channel is not exists, allocate one */
 	one = SLIST_FIRST( &freeCallback );
 	if ( one == NULL ) return FAILED;
	SLIST_REMOVE( &freeCallback, one, voip_callback_s, nextCallback );
	SLIST_INSERT_HEAD( &inuseCallback, one, nextCallback );

found:

	if ( pReg ) /* for "add" mode */
	{
		one->chid = chid;
		one->fpPcmInt = pReg->fpPcmInt;
		one->fpPktRx = pReg->fpPktRx;
		one->fpTick = pReg->fpTick;
		one->msTick = pReg->msTick;
		one->msRemain = pReg->msTick;
	}
	else        /* for "del" mode */
	{
		SLIST_REMOVE( &inuseCallback, one, voip_callback_s, nextCallback );
		SLIST_INSERT_HEAD( &freeCallback, one, nextCallback );
	}
	
	return SUCCESS;
}


/*
@func int32 | voip_dump_callback | Dump for debug
*/
void voip_dump_callback( void )
{
	voip_callback_t *one;
	
	rtlglue_printf( "chid  tick(ms)   pcmInt   pktRx\n" );
	rtlglue_printf( "--------------------------------------------------\n" );
	for( one = SLIST_FIRST( &inuseCallback );
	     one;
 	     one = SLIST_NEXT( one, nextCallback ) )
 	{
		rtlglue_printf( "  %1d  %p(%d) %p %p\n", 
		                one->chid, one->fpTick, one->msTick,
		                one->fpPcmInt, one->fpPktRx );
 	}
}


/*
@func int32 | voip_ms_routine | check if any msRoutine is arrived.
@parm int32 | msPassed | micro second passed
@rvalue SUCCESS	| Success
@rvalue FAILED | Failure
@comm 
@devnote
*/
int32 voip_ms_routine( int32 msPassed )
{
	voip_callback_t *one;
	
	for( one = SLIST_FIRST( &inuseCallback );
	     one;
 	     one = SLIST_NEXT( one, nextCallback ) )
 	{
 		one->msRemain -= msPassed;
 		if ( one->msRemain <= 0 )
 		{
 			if ( one->fpTick )
	 			one->fpTick( one->chid, msPassed );
 			one->msRemain += one->msTick;
 		}
 	}
 	
	return SUCCESS;
}


/*
@func int32 | voip_pollPcmInt | Poll all PCM interrupt and callback corresponding handler.
@rvalue SUCCESS	| Success
@rvalue FAILED | Failure
@comm 
@devnote
*/
int32 voip_pollPcmInt( void )
{
	voip_callback_t *one;
	uint32 pisr;

	pisr = READ_MEM32( PCM_ISR );
	WRITE_MEM32( PCM_ISR, pisr ); /* clear interrupt pending bits to avoid infinite loop */
	for( one = SLIST_FIRST( &inuseCallback );
	     one;
 	     one = SLIST_NEXT( one, nextCallback ) )
 	{
 		/* If any bit of this channel is set, call the call back function 
 		 * It is bad to hardcode const in the code, but it is OK if we need performance.
 		 */
 		if ( pisr & (0xff000000>>(one->chid*8)) )
 		{
#ifdef CONFIG_RTL865X_PCM
			voip_cntPollPcmInt++;
#endif
 			if ( one->fpPcmInt )
 			{
	 			one->fpPcmInt( pisr );/*voip_session_pcmInt, rx_to_voip*/
 			}
 		}
 	}
 	
	return SUCCESS;
}


/*
@func int32 | voip_query_RTPport | Query RTP information with IP address and transport port.
@parm uint32 | *chid | (output) Address to store channel ID
@parm uint32 | *mid | (output) Address to store media ID
@parm voip_rtp_port_t | *pRtp | (input) IP and transport port.
@rvalue SUCCESS	| RTP found.
@rvalue FAILED | RTP not found.
@comm 
@devnote
*/
int32 voip_query_RTPport( uint32 *chid, uint32 *mid, voip_rtp_port_t *pRtp )
{
	voip_rtp_port_t *one;
	
#ifdef _VOIP_DEBUG_
	rtlglue_printf( "%s( isTcp=%d, remIp=0x%08x, rempPort=%d, extIp=0x%08x, extPort=%d ).\n\n", 
	                __FUNCTION__, pRtp->isTcp, pRtp->remIp, pRtp->remPort, pRtp->extIp, pRtp->extPort );
#endif
	
	for( one = SLIST_FIRST( &inuseRtp );
	     one;
 	     one = SLIST_NEXT( one, nextRtp ) )
 	{
 		if ( one->chid == pRtp->chid &&
 		     one->mid == pRtp->mid )
 		goto found;
 	}

 	/* not found */
	return FAILED;

found:
	if ( chid ) *chid = one->chid;
	if ( mid ) *mid = one->mid;
	
	return SUCCESS;
}


/*
@func int32 | voip_register_RTPport | Register RTP information with IP address and transport port.
@parm uint32 | chid | (input) Address to store channel ID
@parm uint32 | mid | (input) Address to store media ID
@parm voip_rtp_port_t | *pRtp | (input) IP and transport port.
@rvalue SUCCESS	| RTP information added.
@rvalue FAILED | RTP information not added.
@comm 
@devnote
In current, we do not support TCP.
*/
int32 voip_register_RTPport( uint32 chid, uint32 mid, voip_rtp_port_t* pRtp )
{
	voip_rtp_port_t *one;
	int32 ret;

	if ( pRtp && pRtp->isTcp )
	{
		/* In current, we don't support TCP yet. */
		ret = FAILED;
		goto out;
	}
	
#ifdef _VOIP_DEBUG_
	rtlglue_printf( "%s( isTcp=%d, remIp=0x%08x, rempPort=%d, extIp=0x%08x, extPort=%d ).\n\n", 
	                __FUNCTION__, pRtp->isTcp, pRtp->remIp, pRtp->remPort, pRtp->extIp, pRtp->extPort );
#endif
	
	for( one = SLIST_FIRST( &inuseRtp );
	     one;
 	     one = SLIST_NEXT( one, nextRtp ) )
 	{
 		if ( one->chid == chid &&
 		     one->mid == mid )
 		goto found;
 	}

 	/* not found, allocate one */
 	one = SLIST_FIRST( &freeRtp );
 	if ( one == NULL )
 	{
		printk("SLIST_FIRST failed\n");
 		ret = FAILED;
 		goto out;
 	}
	SLIST_REMOVE( &freeRtp, one, voip_rtp_port_s, nextRtp );
	SLIST_INSERT_HEAD( &inuseRtp, one, nextRtp );

found:
	if ( pRtp )
	{
		rtl8651_tblDrvAclRule_t acl;
		rtl8651_tblDrv_networkIntfTable_t *intf;
		
		one->chid = chid;
		one->mid = mid;
		one->isTcp = pRtp->isTcp;
		one->remIp = pRtp->remIp;
		one->remPort = pRtp->remPort;
		one->extIp = pRtp->extIp;
		one->extPort = pRtp->extPort;

		/* register to trap RTP packet. */
		ret = rtl8651_registerTrappingDispatchEntry( one->isTcp?IPPROTO_TCP:IPPROTO_UDP,
		                                             one->remIp, one->extIp,
		                                             one->remPort, one->extPort,
		                                             VOIP_CATEGORY( one->chid, one->mid ) );
		if ( ret != SUCCESS )
		{
			printk("rtl8651_registerTrappingDispatchEntry failed\n");
			goto out;
		}

		/* query interface name */
		intf = _rtl8651_getIpInfo( one->extIp, NULL, NULL, NULL );
		if ( intf==NULL )
		{
			printk("_rtl8651_getIpInfo failed\n");
			ret = FAILED;
			goto out;
		}
		
		/* register ACL */
		memset((void*)&acl,0,sizeof(acl));
		acl.ruleType_=RTL8651_ACL_UDP;
		acl.srcIpAddr_		= one->remIp;
		acl.srcIpAddrMask_	= 0xffffffff;
		acl.dstIpAddr_		= one->extIp;
		acl.dstIpAddrMask_	= 0xffffffff;
		acl.udpSrcPortLB_=one->remPort;
		acl.udpSrcPortUB_=one->remPort;
		acl.udpDstPortLB_=one->extPort;
		acl.udpDstPortUB_=one->extPort;
		acl.actionType_ = RTL8651_ACL_CPU;
		ret = rtl8651a_addAclRule( 0, intf->ifName, TRUE,  &acl );
		if ( ret != SUCCESS )
		{
			printk("Add acl rule fail\n");
			goto out;
		}
	}
	else
	{
		/* unregister to trap RTP packet. */
		ret = rtl8651_unregisterTrappingDispatchEntry( one->isTcp?IPPROTO_TCP:IPPROTO_UDP,
		                                               one->remIp, one->extIp,
		                                               one->remPort, one->extPort );

		SLIST_REMOVE( &inuseRtp, one, voip_rtp_port_s, nextRtp );
		SLIST_INSERT_HEAD( &freeRtp, one, nextRtp );
	}

	ret = SUCCESS;
out:
	return ret;
}


/*
@func int32 | voip_register_RTPport | Register RTP information with IP address and transport port.
@parm uint32 | chid | (input) Address to store channel ID
@parm uint32 | mid | (input) Address to store media ID
@parm voip_rtp_port_t | *pRtp | (input) IP and transport port.
@rvalue SUCCESS	| RTP information added.
@rvalue FAILED | RTP information not added.
@comm 
@devnote
In current, we do not support TCP.
*/
int32 voip_pktRx( struct rtl_pktHdr *pPkt )
{
	voip_callback_t *one;
	uint32 chid, mid;

	VOIP_CATEGORY_MAP( pPkt->ph_category, &chid, &mid );
	
#ifdef _VOIP_DEBUG_
	rtlglue_printf( "%s( pPkt->ph_category=%08x ), chid=%d, mid=%d.\n", 
	                __FUNCTION__, pPkt->ph_category, chid, mid );
#endif

	for( one = SLIST_FIRST( &inuseCallback );
	     one;
 	     one = SLIST_NEXT( one, nextCallback ) )
 	{
#ifdef _VOIP_DEBUG_
		rtlglue_printf( "%s():%d one->chid=%d.\n", 
		                __FUNCTION__, __LINE__, one->chid );
#endif
 		if ( one->chid == chid )
 		{
 			/* found */
 			struct ip *ipHdr;
			struct udphdr *udp;

 			ipHdr = (struct ip *)(pPkt->ph_mbuf->m_data + pPkt->ph_iphdrOffset);
			udp = (struct udphdr *)L4HDR_ADDR(ipHdr);
 			
			if ( one->fpPktRx )
			{
#ifdef _VOIP_DEBUG_
				rtlglue_printf( "%s():%d udp payload length=%d.\n", 
				                __FUNCTION__, __LINE__, ntohs(udp->uh_ulen)-sizeof(*udp) );				
#endif
				one->fpPktRx( chid, mid, ((uint8*)udp)+sizeof(*udp), ntohs(udp->uh_ulen)-sizeof(*udp), TYPE_NONE );
			}
 			
 			break;
 		}
 	}

	mBuf_freeMbufChain( pPkt->ph_mbuf );
	return SUCCESS;
}


/*
@func int32 | voip_dump_RTPport | Dump for debug
*/
void voip_dump_RTPport( void )
{
	voip_rtp_port_t *one;
	
	rtlglue_printf( "chid mid Pro ExternalIp:port  RemoteIp:port\n" );
	rtlglue_printf( "--------------------------------------------------\n" );
	for( one = SLIST_FIRST( &inuseRtp );
	     one;
 	     one = SLIST_NEXT( one, nextRtp ) )
 	{
		rtlglue_printf( "  %1d   %1d  %3s "NIPQUAD2STR2":%5d "NIPQUAD2STR2":%5d\n", 
		                one->chid, one->mid, one->isTcp?"TCP":"UDP",
		                NIPQUAD2(one->extIp), one->extPort, NIPQUAD2(one->remIp), one->remPort );
 	}
}


/*
@func int32 | voip_TxRtpPkt | Send a packet
@parm uint32 | chid | (input) Address to store channel ID
@parm uint32 | mid | (input) Address to store media ID
@parm void | *packet | (input) IP and transport port.
@parm uint32 | pktLen | (input) Packet Length
@parm uint32 | flags | (input) Flags
@rvalue SUCCESS	| RTP information added.
@rvalue FAILED | RTP information not added.
@comm 
@devnote
In current, we do not support TCP.
*/
int32 voip_TxRtpPkt( uint32 chid, uint32 mid, void* packet, uint32 pktLen, enum VOIP_PKT_TYPE flags )
{
	voip_rtp_port_t *one;
	int32 ret;
	static uint16 ipIdCount = 0x5566; /* any number is OK. */
	
#ifdef _VOIP_DEBUG_
	rtlglue_printf( "%s(chid=%d, mid=%d, packet=%p, pktLen=%d, flags=0x%08x).\n\n", 
	                __FUNCTION__, chid, mid, packet, pktLen, flags );
#endif

	for( one = SLIST_FIRST( &inuseRtp );
	     one;
 	     one = SLIST_NEXT( one, nextRtp ) )
 	{
 		if ( one->chid == chid &&
 		     one->mid == mid )
 		break;
 	}

	if ( one )
	{
		struct ip *pip;
		struct udphdr *udp;
		uint8 *rtp;
		struct rtl_pktHdr *pkt;
		struct rtl_mBuf  *mbuf;

		/* get a new pkt hdr and mbuf. */
		if( (mbuf = mBuf_get(MBUF_DONTWAIT,MBUFTYPE_DATA, 1)) == NULL )
		{
			ret = FAILED;
#ifdef _VOIP_DEBUG_
			rtlglue_printf( "%s():%d return %d.\n\n", 
			                __FUNCTION__, __LINE__, ret );
#endif
			goto out;
		}
		if( (mBuf_getPkthdr(mbuf, MBUF_DONTWAIT) == (struct rtl_mBuf *) NULL)||
			(mBuf_reserve(mbuf,128/*headroom*/+sizeof(struct ether_header)+2/*align*/)) )
		{
			mBuf_freeMbufChain(mbuf);/* free new mbuf we just get. */
			ret = FAILED;
#ifdef _VOIP_DEBUG_
			rtlglue_printf( "%s():%d return %d.\n\n", 
			                __FUNCTION__, __LINE__, ret );
#endif
			goto out;
		}

		pkt = mbuf->m_pkthdr;
		
		pip = (struct ip *)((uint8*)mbuf->m_data);
		udp = (struct udphdr *)(((uint8*)pip)+sizeof(struct ip));
		rtp = (uint8*)((uint8*)udp+sizeof(struct udphdr));

		/* build new ip header */
		pip->ip_vhl = 0x45;
		pip->ip_tos = 0x00;
		pip->ip_len = ntohs( sizeof(struct ip) + sizeof(struct udphdr) + pktLen );
#ifdef RTL865X_TEST
		pip->ip_id = ntohs(0x7788); /* for test code only */
#else
		pip->ip_id = ntohs(((uint8*)packet)[0]<<8|((uint8*)packet)[1])^((uint32)packet)^((uint32)packet>>16)^(ipIdCount++); /* as random as possible */
#endif
		pip->ip_off = 0;
		pip->ip_ttl = 147; /* just for identification. */
		pip->ip_p = IPPROTO_UDP;
		//pip->ip_sum = 
		pip->ip_src.s_addr = ntohl( one->extIp );
		pip->ip_dst.s_addr = ntohl( one->remIp );

		/* build new UDP header */
		udp->uh_ulen = ntohs( pktLen + sizeof(struct udphdr) );
		udp->uh_sport = ntohs( one->extPort );
		udp->uh_dport = ntohs( one->remPort );
		udp->uh_sum = 0x5566; /* must not be 0 to force re-compute */

		/* copy RTP header and payload */
		memcpy( rtp, packet, pktLen );

		ret = rtl8651_offloadEngineL3Send( pkt, pktLen+sizeof(struct ip)+sizeof(struct udphdr), 
		                                   OFFLOAD_IP_CKSUM | OFFLOAD_L4_CKSUM, 0 );
		if ( ret != SUCCESS )
		{
#ifdef _VOIP_DEBUG_
			rtlglue_printf( "%s():%d return %d.\n\n", 
			                __FUNCTION__, __LINE__, ret );
#endif
			goto out;
		}                                     
	}
	else
	{
		/* not found */
		ret = FAILED;
		goto out;
	}

	ret = SUCCESS;
out:
	return ret;
}


