

#
# cp3.S - Utilities for Coprocessor 3
#


#include "mips.h"

	.data
cp3_ctrl:
	.word	0x0

	.text

 ######################################################################
 # cache flush for Lexra LX4180
 #

	.globl	flush_cache_r1
	.ent	flush_cache_r1
flush_cache_r1:

	.set	reorder

	#--- transition to kseg1 from undetermined kernel segment
	la	t1,1f
	or	t1,0xa0000000
	jr	t1
	nop
1:

	/*
	** This section handles invalidating Lexra's lx-4080 core
	** The code is modeled after the code found in the LX-4080
        ** Evaluation System Users Guide.
	*/
	li		t0, 2		/* Ccntl bit for Icache invalidate */
	beqz	a0, 1f
	li		t0, 1		/* Ccntl bit for Dcache invalidate */

	.set	noreorder

1:	mtc0	zero, $20	/* CCTL */
	nop
	nop
	mtc0	t0, $20
	nop
	nop
	mtc0	zero, $20	/* CCTL */
	nop
	nop

	j		ra
	nop

	.set	reorder
	
	.end	flush_cache_r1

	.globl	cp3_init
	.ent	cp3_init


#
# void cp3_init(void)
#
cp3_init:
	# set cp3 usuable bit in status register
	#
	mfc0	t1, status
	nop
	la		t2, 0x80000000
	or		t1, t2
	mtc0	t1, status
	nop

	# stop counters
	#
	ctc3	zero, $0
	
	# clear counters
	#
	mtc3	zero, $8
	mtc3	zero, $9
	mtc3	zero, $10
	mtc3	zero, $11
	mtc3	zero, $12
	mtc3	zero, $13
	mtc3	zero, $14
	mtc3	zero, $15

	# clear cp3 conrol word
	#
	la		t2, cp3_ctrl
	sw		zero, 0(t2)
		
	j		ra
	
	.end	cp3_init
	
	
	.globl	cp3_stop
	.ent	cp3_stop


#
# void cpe_stop(void)
#	
cp3_stop:
	# stop counters
	#
	ctc3	zero, $0
	
	j		ra
	
	.end	cp3_stop
	
	
	.globl	cp3_start
	.ent	cp3_start


#
# void cp3_start(void)
#	
cp3_start:
	# load the counter countrol word to cp3
	#
	la		t4, cp3_ctrl
	lw		t5, 0(t4)
	nop
	ctc3	t5, $0
	nop
	
	j		ra
	
	.end	cp3_start
	

#
# void cp3_set_counter(int counter, unsigned int control_bits)
#
	.globl	cp3_set_counter
	.ent	cp3_set_counter

cp3_set_counter:
	# load control bits
	#
	la		t1, cp3_ctrl
	lw		t2, 0(t1)
	
	# compute mask bits
	#
	sll		a0, 3	
	li		t0, 0x1f
	sllv	t0, t0, a0
	not		t0

	# clear original bits
	#
	and		t2, t0
	
	# set new bits
	#
	sllv	t0, a1, a0
	or		t2, t0
	
	# save control bits
	#
	sw		t2, 0(t1)
	
	j		ra
	
	.end	cp3_set_counter
	
	
#
# unsigned int cp3_get_counter_high(int counter)
#
	.globl	cp3_get_counter_high
	.ent	cp3_get_counter_high
	
cp3_get_counter_high:
	bnez	a0, 1f
	mfc3	v0, $9
	j		2f
1:
	bne		a0, 1, 1f
	mfc3	v0, $11
	j		2f
	
1:
	bne		a0, 2, 1f
	mfc3	v0, $13
	j		2f
	
1:
	mfc3	v0, $15

2:
	nop
	andi	v0, 0x0000ffff
	j		ra
	
	.end	cp3_get_counter_high


#
# unsigned int cp3_get_counter_low(int counter)
#
	.globl	cp3_get_counter_low
	.ent	cp3_get_counter_low
	
cp3_get_counter_low:
	bnez	a0, 1f
	mfc3	v0, $8
	j		2f
1:
	bne		a0, 1, 1f
	mfc3	v0, $10
	j		2f
	
1:
	bne		a0, 2, 1f
	mfc3	v0, $12
	j		2f
	
1:
	mfc3	v0, $14

2:
	j		ra
	
	.end	cp3_get_counter_low


#
# void cp3_set_iram_addr(void* base_addr, void* top_addr)
#
	.globl	cp3_set_iram_addr
	.ent	cp3_set_iram_addr
	
cp3_set_iram_addr:
	addi	sp, sp, -20
	sw		ra, 16(sp)
	sw		a0, 12(sp)
#	move	a3, a0		# czc : store a0
	
	# flush cache first
	#
	#define ICACHE	0
 	li		a0, ICACHE
	jal		flush_cache

	# init cp3 i-ram base and top registers
	#
	.set	noreorder
#	move	a0, a3		# czc : restore a0
	lw		a0, 12(sp)
    nop
	and		t0, a0, 0x0fffffff			# translate to physical address
	and		t1, a1, 0x0fffffff			# czc	
 	mtc3	t0, $0						# $0: i-ram base
 	mtc3	t1, $1						# $1: i-ram top		czc
	nop
	nop
	nop
	nop
	.set	reorder
	
	lw		ra, 16(sp)
	addi	sp, sp, 20
	
	j		ra
	
	.end	cp3_set_iram_addr


#
# void cp3_set_dram_addr(void* base_addr, void* top_addr)
#
	.globl	cp3_set_dram_addr_8601
	.ent	cp3_set_dram_addr_8601
	
cp3_set_dram_addr_8601:

	addi	sp, sp, -20
	sw		ra, 16(sp)
	sw		a0, 12(sp)
#	move	a3, a0		# czc	
	
	# flush cache first
	#
	#define DCACHE	1
 	li		a0, DCACHE
	jal		flush_cache

	# init cp3 d-ram base and top registers
	#
	.set	noreorder
#	move	a0, a3		# czc
	lw		a0, 12(sp)
	nop
	and		t0, a0, 0x0fffffff			# translate to physical address
	and		t1, a1, 0x0fffffff			# czc
 	mtc3	t0, $4						# $4: d-ram base
 	mtc3	t1, $5						# $5: d-ram top
	nop
	nop
	nop
	nop
	.set	reorder
	
	lw		ra, 16(sp)
	addi	sp, sp, 20
	
	j		ra
	
	.end	cp3_set_dram_addr_8601

#
# void cp3_set_dram_addr_8181(void* base_addr, void* top_addr)
#
	.global	cp3_set_dram_addr_8181
	.ent	cp3_set_dram_addr_8181
	
cp3_set_dram_addr_8181:

	#
	# Invalidate dcache first
	#
	mfc0 	t0, $20 # read CCTL
	nop
	nop
	ori 	t0, t0, 0x1 # bit 0: DInval
	mtc0 	t0, $20
	nop
	nop
		
	#
	# Make sure Coprocessor 3 is usable
	#
	mfc0 	t0, $12 # read STATUS
	nop
	nop
	or 		t0, t0, 0x80000000 # enable CP3
	mtc0 	t0, $12
	nop
	nop
	
	#
	# Set D-RAM base and top to enable D-RAM
	#	
	and 	t1, a0, 0x1fffffff # translate to physical address	
	move	t2, a1
	//sub 	t2, t2, 1
	and 	t2, t2, 0x1fffffff # translate to physical address
	mtc3 	t1, $4 # base, must be 1KB alignment
	mtc3 	t2, $5 # top
	nop
	nop
	
	jr		ra
	nop
	
	.end	cp3_set_dram_addr_8181
