/* tools for cp3 and debug used */
#include <cyg/infra/diag.h>
#include "cp3.h"
#include "typedef.h"

int strlen(const char* str)
{
	int len = 0;

	while (*str)
	{
		len++;
		str++;
	}

	return len;
}

char *strcpy(char * dst, const char * src)
{
	int i, j;
	
	i = strlen(src) + 1;
	for (j = 0; j < i; j++)
		*(dst + j) = *(src + j);

	return dst;
}

void *memcpy(void *out, const void *in, unsigned int n)
{
	char *dst = (char *) out;
  	char *src = (char *) in;
  	void *tmp;
  	
  	tmp = out;
  	
  	while (n--)
    {
      *dst++ = *src++;
    }  	
	
	return tmp;
}

void *memset(void *out, const int in, unsigned int n)
{
	char *dst = (char *) out;
	unsigned int i;
	void *tmp = out;
	
	for (i=0; i<n; i++)
		*dst++ = in;
	
	return tmp;
}

void delay_s(int n)
{
	int i,j;
	for(i=0; i<n; i++)
		for (j=0; j<1000; j++);
}

void MyWrite(char* pStr)
{
	/*
	unsigned int* pR1Reg = (unsigned int*) 0xbd010204;	
	CDspcodecResponse* pSendBuffer = (CDspcodecResponse*)RECV_ADDR;
		
	while(pSendBuffer->nOwner == 1)
		delay_s(300);		
	strcpy(pSendBuffer->pPrintBuff, pStr);
	pSendBuffer->uAck = DSPCODEC_ACK_PRINTMSG;
	pSendBuffer->nOwner = 1;
	*pR1Reg = 0x00000002;
	*/
	diag_printf(pStr);
}

char* itoa(unsigned int nValue, unsigned int nRadix, char* pStr, int nBufSize)
{
	int i, nLen = 0;
	char c, pBuf[32];

	if (nValue == 0)
	{
		nLen = 1;
		pBuf[0] = '0';
		pBuf[1] = '\0';
	}
	else
	{	
		while (1)
		{
			if (nValue == 0)
				break;
			
			c = nValue % nRadix;
			if (c < 10)
				pBuf[nLen] = '0' + c;
			else
				pBuf[nLen] = 'a' + c - 10;
			
			nValue /= nRadix;
			nLen++;
		}
	}

	if (nLen >= nBufSize - 1)
		//return NULL;
		return 0;
		
	for (i = 0; i < nLen; i++)
		pStr[i] = pBuf[nLen - i - 1];
		
	pStr[nLen] = '\0';
	return pStr;
}



void WriteNumber(unsigned int nNumber)
{
	char buf[32], *p;
	
	p = itoa(nNumber, 10, buf, 32);
	if (p)
		MyWrite(p);
}

void WriteNumberHex(unsigned int nNumber, int bPrefix)
{
	char buf[32], *p;
		
	p = itoa(nNumber, 16, buf, 32);
	if (p)
	{
		int i, nLen;
		char buf2[32];
		
		nLen = strlen(buf);
		for (i = 0; i < 8 - nLen; i++)
			buf2[i] = '0';
		for (i = 0; i <= nLen; i++)
			buf2[i + 8 - nLen] = buf[i];
			
		if (bPrefix)	
			MyWrite("0x");

		MyWrite(buf2);
	}
}

void WriteNumberLongLongHex(unsigned long long nNumber, int bPrefix)
{
	unsigned int n;
	
	n = nNumber >> 32;
	WriteNumberHex(n, bPrefix);
	
	n = nNumber & 0xffffffff;
	WriteNumberHex(n, 0);
}

/****************************************************/

///////////////////////////////////////////////////////////////////////
// operations for long long (64-bit) type
//
#define BIT63		(1LL << 63)

// multiply of two long long values
//
unsigned long long ll_mul(unsigned long long x, unsigned long long y)
{
	unsigned long long m = 0;
	int i;
	
	for (i = 0; i < 64; i++)
	{
		m <<= 1;

		if (y & BIT63)
			m += x;

		y <<= 1;	
	}
	
	return m;
}

// division of two long long values
//
unsigned long long ll_div(unsigned long long x, unsigned long long y)
{
	unsigned long long d, xx, yy;
	int i, msby, msbx;
	
	// find the msb of y
	//
	yy = y;
	msby = -1;
	for (i = 0; i < 64; i++)
	{
		if (yy & BIT63)
		{
			msby = 63 - i;
			break;
		}
		
		yy <<= 1;
	}
	
	if (msby == -1)
	{
		// y = 0, divided by zero
		// generate an exception
		//
		i = msby / 0;
	}

	// find the msb of x
	//
	xx = x;
	msbx = -1;
	for (i = 0; i < 64; i++)
	{
		if (xx & (1LL << 63))
		{
			msbx = 63 - i;
			break;
		}
		
		xx <<= 1;
	}
	
	if (msbx == 0)
	{
		// x = 0
		//
		return 0LL;
	}
	
	d = 0;
	yy = y << (msbx - msby);
	
	for (i = 0; i <= msbx - msby; i ++)
	{
		d <<= 1;
		
		if (x > yy)
		{
			x -= yy;
			d |= 1;
		}
		
		yy >>= 1;
	}
	
	return d;
}

// module of two long long values
//
unsigned long long ll_mod(unsigned long long x, unsigned long long y)
{
	return x - ll_mul(y, ll_div(x, y));
}

///////////////////////////////////////////////////////////////////////
// utilities for cp3
//

static unsigned long long cp3_counter[4];
static unsigned int cp3_usage[4];

void cp3_set_counter_usage(int counter, unsigned int usage)
{
	if (counter < 0 || counter > 3)
		return;
		
	cp3_usage[counter] = usage;
	
	cp3_set_counter(counter, usage);
}

void cp3_get_counters(void)
{
	int i;
	unsigned long long value;
	
	for (i = 0; i < 4; i++)
	{
		value = ((unsigned long long)cp3_get_counter_high(i)) << 32;
		cp3_counter[i] = value + cp3_get_counter_low(i);
	}
}

static char counter_titles[][64] =
{
	"stopped",
	"clock cycles",
	"new instruction fetches",
	"instruction fetch cache misses",
	"instruction fetch miss busy cycles",
	"data store instructions",
	"data load instructions",
	"load or store instructions",
	"load or store cache misses",
	"load or store miss busy cycles",
	"unknown"
};

const char* cp3_get_usage_title(int counter)
{
	int title;
	
	if (counter < 0 || counter > 3)
		return 0;

	switch (cp3_usage[counter])
	{
	case CP3_COUNT_STOP:
		title = 0;
		break;
		
	case CP3_COUNT_CYCLE:
		title = 1;
		break;
		
	case CP3_COUNT_NEW_INST_FETCH:
		title = 2;
		break;
		
	case CP3_COUNT_INST_CACHE_MISS:
		title = 3;
		break;
		
	case CP3_COUNT_INST_CACHE_MISS_BUSY:
		title = 4;
		break;
		
	case CP3_COUNT_DATA_STORE:
		title = 5;
		break;
		
	case CP3_COUNT_DATA_LOAD:
		title = 6;
		break;
		
	case CP3_COUNT_LOAD_STORE:
		title = 7;
		break;
		
	case CP3_COUNT_LOAD_STORE_MISS:
		title = 8;
		break;
		
	case CP3_COUNT_LOAD_STORE_MISS_BUSY:
		title = 9;
		break;
		
	default:
		title = 10;
		break;
	}

	return counter_titles[title];	
}

// show a counter value, as well as ratio to another counter indexed by peer_counter.
void cp3_show_counter(int counter, int peer_counter)
{
	if (counter < 0 || counter > 3)
		return;

	MyWrite("\nCounter ");	
	WriteNumber(counter);
	MyWrite(": ");
	
	MyWrite((char *)cp3_get_usage_title(counter));
	MyWrite("\n  ");
	WriteNumberLongLongHex(cp3_counter[counter], 1);
	
	if (peer_counter >=0 && peer_counter <= 3 && peer_counter != counter)
	{
		unsigned long long ratio, integer, decimal;
		
		MyWrite("\n  Ratio: ");
		ratio = ll_div(ll_mul(cp3_counter[counter], 10000), cp3_counter[peer_counter]);		
		integer = ll_div(ratio, 100);
		decimal = ratio - ll_mul(integer, 100);
		
		WriteNumber((unsigned int)integer);
		MyWrite(".");
		if (decimal < 10) {
			MyWrite("0");
		}
		WriteNumber((unsigned int)decimal);
		
		MyWrite("% (to ");
		MyWrite((char *)cp3_get_usage_title(peer_counter));
		MyWrite(")");		
	}

	MyWrite("\n");
}


/////////////////////////////////////////////////
