/*
**
** File:        "cst_lbc.h"
**
** Description:  This file contains global definition of the SG15
**    LBC Coder for 6.3/5.3 kbps.
**
*/


/*
    ITU-T G.723 Speech Coder   ANSI-C Source Code     Version 5.00
    copyright (c) 1995, AudioCodes, DSP Group, France Telecom,
    Universite de Sherbrooke.  All rights reserved.
*/
#ifndef __CST_LBC_H__
#define __CST_LBC_H__

#define  False 0
#define  True  1

/* Definition of the working mode */
typedef enum {Both, Cod, Dec} Wmode;

/* Coder rate */
typedef enum {Rate63, Rate53} Crate;

/* Coder global constants */
#define  Frame       240
#define  LpcFrame    180
#define  SubFrames   4
#define  SubFrLen    (Frame/SubFrames)

/* LPC constants */
#define  LpcOrder          10
#define  RidgeFact         10
#define  CosineTableSize   512
#define  PreCoef           (Word16) 0xc000            /* -0.25*2 */

#define  LspPrd0           12288
#define  LspPrd1           23552

#define  LspQntBands       3
#define  LspCbSize         256
#define  LspCbBits         8

/* LTP constants */
#define  PitchMin          18
#define  PitchMax          (PitchMin+127)
#define  PwConst           (Word16) 0x2800
#define  PwRange           3
#define  ClPitchOrd        5
#define  Pstep             1
#define NbFilt085          85
#define NbFilt170          170

/* MP-MLQ constants */
#define  Sgrid             2
#define  MaxPulseNum       6
#define  MlqSteps          2

/* acelp constants */
#define SubFrLen2          (SubFrLen +4)
#define DIM_RR             416
#define NB_POS             8
#define STEP               8
#define MSIZE              64
#define threshold          16384  /* 0.5 = 16384 in Q15 */
#define max_time           120

/* Gain constant */
#define  NumOfGainLev      24

/* FER constant */
#define  ErrMaxNum         3

/* CNG constants  */
#define NbAvAcf            3  /* Nb of frames for Acf average               */
#define NbAvGain           3  /* Nb of frames for gain average              */
#define ThreshGain         3  /* Theshold for quantized gains               */
#define FracThresh         7000   /* Itakura dist threshold: frac. part     */
#define NbPulsBlk          11 /* Nb of pulses in 2-subframes blocks         */

#define InvNbPulsBlk       2979 /* 32768/NbPulsBlk                          */
#define NbFilt             50 /* number of filters for CNG exc generation   */
#define LpcOrderP1         (LpcOrder+1)
#define SizAcf             ((NbAvAcf+1)*LpcOrderP1) /* size of array Acf (44)   */	
#define SubFrLenD          (2*SubFrLen)
#define Gexc_Max           5000  /* Maximum gain for fixed CNG excitation   */

/* Taming constants */
#define NbFilt085_min      51
#define NbFilt170_min      93
#define SizErr             5
#define Err0               (Word32)4  /* scaling factor */
#define ThreshErr          0x40000000L
#define DEC                (30 - 7)

/*
   Used structures
*/
typedef  struct   {
   /* High pass variables */
   Word16   HpfZdl   ;
   Word16	blank0;
   Word32   HpfPdl   ;							// +4
   Word16	blank[3];			// not used, just for data alignment

   /* Sine wave detector */
   Word16   SinDet   ;							// +14   

   /* Lsp previous vector */
   Word16   PrevLsp[LpcOrder] ;					// +16
   Word16	blank2[2];

   /* All pitch operation buffers */
   Word16   PrevWgt[PitchMax] ;					// +40
   Word16	blank3[3];
   Word16   PrevErr[PitchMax] ;					// +336
   Word16	blank4[3];
   Word16   PrevExc[PitchMax] ;					// +632
   Word16	blank5[3];

   /* Required memory for the delay */
   Word16   PrevDat[LpcFrame-SubFrLen] ;		// +928

   /* Used delay lines */
   Word16   WghtFirDl[LpcOrder] ;				// +1168
   Word16	blank6[2];
   Word16   WghtIirDl[LpcOrder] ;				// +1192
   Word16	blank7[2];
   Word16   RingFirDl[LpcOrder] ;				// +1216
   Word16	blank8[2];
   Word16   RingIirDl[LpcOrder] ;				// +1240
   Word16	blank9[2];

   /* Taming procedure errors */
   Word32 Err[SizErr];							// +1264

   } CODSTATDEF  ;

typedef  struct   {
   Word16   Ecount ;
   Word16   InterGain ;
   Word16   InterIndx ;
   Word16   Rseed ;
   Word16   Park  ;
   Word16   Gain  ;
   /* Lsp previous vector */
   Word16   PrevLsp[LpcOrder] ;				// +12

   /* Used delay lines */
   Word16   SyntIirDl[LpcOrder] ;			// +32
   Word16   PostFirDl[LpcOrder] ;			// +52
   Word16   PostIirDl[LpcOrder] ;			// +72
   Word32	blank0;
   /* All pitch operation buffers */
   Word16   PrevExc[PitchMax] ;				// +96

   } DECSTATDEF  ;

   /* subframe coded parameters */
typedef  struct   {
   Word16   AcLg  ;
   Word16   AcGn  ;
   Word16   Mamp  ;
   Word16   Grid  ;
   Word16   Tran  ;
   Word16   Pamp  ;
   Word32   Ppos  ;
   } SFSDEF ;

   /* frame coded parameters */
typedef  struct   {
   Word16   Crc   ;
   Word16	blank ;
   Word32   LspId ;
   Word16   Olp[SubFrames/2] ;	//+8
   SFSDEF   Sfs[SubFrames] ;	//+12
   } LINEDEF ;

   /* harmonic noise shaping filter parameters */
typedef  struct   {
   Word16   Indx  ;
   Word16   Gain  ;
   } PWDEF  ;

    /* pitch postfilter parameters */
typedef  struct   {
   Word16   Indx  ;
   Word16   Gain  ;
   Word16   ScGn  ;
   } PFDEF  ;

    /* best excitation vector parameters for the high rate */
typedef  struct {
   Word32   MaxErr   ;
   Word16   GridId   ;
   Word16   MampId   ;
   Word16   UseTrn   ;
   Word16	blank	 ;	// for data alignment
   Word16   Ploc[MaxPulseNum] ;//+12
   Word16   Pamp[MaxPulseNum] ;//+24
   } BESTDEF ;

    /* VAD static variables */
typedef struct {
    Word16  Hcnt ;
    Word16  Vcnt ;		//+2
    Word32  Penr ;		//+4
    Word32  Nlev ;		//+8
    Word16  Aen ;		//+12
	Word16	blank;				//for data alignment
    Word16  Polp[4] ;	//+16
    Word16  NLpc[LpcOrder] ;	//+24
} VADSTATDEF ;


/* CNG features */

/* Coder part */
typedef struct {
    Word16 CurGain;
    Word16 PastFtyp;
    Word16 Acf[SizAcf];			//+4
	Word32 blank1;	
    Word16 ShAcf[NbAvAcf+1];	//+96
    Word16 LspSid[LpcOrder] ;	//+104
	Word32 blank2;
    Word16 SidLpc[LpcOrder] ;	//+128
	Word32 blank3;				
    Word16 RC[LpcOrderP1];		//+152
    Word16 ShRC;				//+174
    Word16 Ener[NbAvGain];		//+176
	Word16 blank4;
    Word16 NbEner;				//+184
    Word16 IRef;				//+186
    Word16 SidGain;				//+188
    Word16 RandSeed;			//+190
} CODCNGDEF;

/* Decoder part */
typedef struct {
    Word16 CurGain;
    Word16 PastFtyp;
    Word16 LspSid[LpcOrder] ;	//+4
    Word16 SidGain;				//+24
    Word16 RandSeed;			//+26
} DECCNGDEF;

/* all Coder used memory */
typedef struct {	
	Crate	WrkRate;					//work rate (Rate63/Rate53)
	Word32	blank0;
	Word16	DataBuff[PitchMax+Frame+1];	//PCM in buf		//align 8
	Word32	blank1;
	CODSTATDEF CodStat;					//Coder state		//align 8 +784
	Word32	blank2;
	CODCNGDEF  CodCng;					//Coder CNG state	//align 8 +2072
	VADSTATDEF VadStat ;				//vad state			//align 8 +2264
} CoderMem;

/* all Decoder used memory */
typedef struct {
	DECSTATDEF DecStat;			// align 8
	Word32 blank0;
	DECCNGDEF DecCng;			// align 8 +392
} DecodMem;

#endif

