/*
**
** File:        "coder.c"
**
** Description:     Top-level source code for G.723 dual-rate coder
**
** Functions:       Init_Coder()
**                  Coder()
**
**
*/
/*
    ITU-T G.723 Speech Coder   ANSI-C Source Code     Version 5.00
    copyright (c) 1995, AudioCodes, DSP Group, France Telecom,
    Universite de Sherbrooke.  All rights reserved.
*/


//#include <stdlib.h>
//#include <stdio.h>
//#include <string.h>

#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "cst_lbc.h"
#include "tab_lbc.h"
#include "coder.h"
#include "lpc.h"
#include "lsp.h"
#include "exc_lbc.h"
#include "util_lbc.h"
#include "vad.h"
#include "cod_cng.h"
#include "tame.h"
#include "../basic_op/cp_tools.h"// 2005/04/20 BrianBin
/*
   This file includes the coder main functions
*/

//CODSTATDEF  CodStat  ;
//CODSTATDEF	*pCodStat = &CodStat;
/*
**
** Function:        Init_Coder()
**
** Description:     Initializes non-zero state variables
**          for the coder.
**
** Links to text:   Section 2.21
** 
** Arguments:       None
**
** Outputs:     None
** 
** Return value:    None
**
*/
void  Init_Coder( void)
{
    int   i ;

    /* Initialize encoder data structure with zeros */
    memset(&pCoderMem->CodStat, 0, sizeof(CODSTATDEF));

    /* Initialize the previously decoded LSP vector to the DC vector */
    for ( i = 0 ; i < LpcOrder ; i ++ )
        pCoderMem->CodStat.PrevLsp[i] = LspDcTable[i] ;

    /* Initialize the taming procedure */
    for(i=0; i<SizErr; i++) 
		pCoderMem->CodStat.Err[i] = Err0;

    return;
}

/*
**
** Function:        Coder()
**
** Description:     Implements G.723 dual-rate coder for    a frame
**          of speech
**
** Links to text:   Section 2
**
** Arguments:
**
**  Word16 DataBuff[]   frame (480 bytes, 240 samples)
**

** Outputs:
**
**  Word16 Vout[]       Encoded frame (20/24 bytes)
**
** Return value:
**
**  Flag            Always True
**
*/
int  Coder( Word16 *DataBuff, char *Vout, Word32 FrCnt )
{
    int     i,j   ;
	int		frame_len=0;

    /*
      Local variables
    */
    Word16   UnqLpc[SubFrames*LpcOrder] ;
    //Word16   QntLpc[SubFrames*LpcOrder] ;
    extern   Word16 QntLpc[];
    Word16   PerLpc[2*SubFrames*LpcOrder] ;

    Word16   LspVect[LpcOrder] ;
    LINEDEF  Line  ;
    PWDEF    Pw[SubFrames]  ;

#ifdef VC6
    Word16   ImpResp[SubFrLen] ;
#else
    extern Word16 ImpResp[];
#endif

	extern 	Word16 Temp[], Temp1[];
    Word16   *Dpnt  ;
	Word16	 *pExc, *pErr;

    Word16   Ftyp = 1 ;
	
    /*
      Coder Start
    */
    Line.Crc = (Word16) 0 ;

    Rem_Dc( DataBuff ) ;

    /* Compute the Unquantized Lpc set for whole frame */
#ifdef VC6
    Comp_Lpc( UnqLpc, pCoderMem->CodStat.PrevDat, DataBuff ) ;
#else
	Comp_Lpcs( UnqLpc, pCoderMem->CodStat.PrevDat, DataBuff ) ;
#endif

    /* Convert to Lsp */
#ifdef VC6
    AtoLsp( LspVect, &UnqLpc[LpcOrder*(SubFrames-1)], pCoderMem->CodStat.PrevLsp ) ;
#else
	AtoLsps( LspVect, &UnqLpc[LpcOrder*(SubFrames-1)], pCoderMem->CodStat.PrevLsp ) ;
#endif

    /* Compute the Vad */
#ifdef VC6
    Ftyp = (Word16) Comp_Vad( DataBuff ) ;
#else
	Ftyp = (Word16) Comp_Vads( DataBuff ) ;
#endif

    /* VQ Lsp vector */
    Line.LspId = Lsp_Qnt( LspVect, pCoderMem->CodStat.PrevLsp ) ;

    Mem_Shift( pCoderMem->CodStat.PrevDat, DataBuff ) ;	// PrevDat -> Dpnt[240:360]
												// DataBuff -> Dpnt[60:300]

    /* Compute Perceptual filter Lpc coefficients */
#ifdef VC6
    Wght_Lpc( PerLpc, UnqLpc ) ;
#else
	Wght_Lpcs( PerLpc, UnqLpc ) ;
#endif

    /* Apply the perceptual weighting filter */
#ifdef VC6
    Error_Wght( DataBuff, PerLpc ) ;			// DataBuff'
#else
	Error_Wghts( DataBuff, PerLpc ) ;			// DataBuff'
#endif
    /*
    // Compute Open loop pitch estimates
    */
    //Dpnt = (Word16 *) malloc( sizeof(Word16)*(PitchMax+Frame) ) ;
	Dpnt = Temp;

    /* Construct the buffer */
    for ( i = 0 ; i < PitchMax ; i ++ )
        Dpnt[i] = pCoderMem->CodStat.PrevWgt[i] ;
    for ( i = 0 ; i < Frame ; i ++ )
        Dpnt[PitchMax+i] = DataBuff[i] ;

#ifdef VC6    
	Vec_Norm( Dpnt, Temp1, (Word16) (PitchMax+Frame) ) ;
#else 
	Vec_Norms( Dpnt, Temp1, (Word16) (PitchMax+Frame) ) ;
#endif

    j = PitchMax ;
    for ( i = 0 ; i < SubFrames/2 ; i ++ ) {
#ifdef VC6        
		Line.Olp[i] = Estim_Pitch( Temp1, (Word16) j ) ;		
#else
		Line.Olp[i] = Estim_Pitchs( Temp1, (Word16) j ) ;
#endif
        pCoderMem->VadStat.Polp[i+2] = Line.Olp[i] ;
        j += 2*SubFrLen ;
    }
        
    if(Ftyp != 1) {

        /*
        // Case of inactive signal
        */
        //free ( (char *) Dpnt ) ;

        /* Save PrevWgt */
        for ( i = 0 ; i < PitchMax ; i ++ )
            pCoderMem->CodStat.PrevWgt[i] = DataBuff[i+Frame-PitchMax] ;

        /* CodCng => Ftyp = 0 (untransmitted) or 2 (SID) */
        Cod_Cng(DataBuff, &Ftyp, &Line, QntLpc);

        /* Update the ringing delays */
        Dpnt = DataBuff;
        for( i = 0 ; i < SubFrames; i++ ) {

            /* Update exc_err */
            Update_Err(Line.Olp[i>>1], Line.Sfs[i].AcLg, Line.Sfs[i].AcGn);

            Upd_Ring( Dpnt, &QntLpc[i*LpcOrder], &PerLpc[i*2*LpcOrder],
                                                        pCoderMem->CodStat.PrevErr ) ;
            Dpnt += SubFrLen;
        }
    }

    else {

        /*
        // Case of Active signal  (Ftyp=1)
        */
		
        /* Compute the Hmw */
        j = PitchMax ;
        for ( i = 0 ; i < SubFrames ; i ++ ) {
#ifdef VC6			
			Comp_Pw( Temp1, (Word16) j, Line.Olp[i>>1], &Pw[i] ) ;
#else
			Comp_Pws( Temp1, (Word16) j, Line.Olp[i>>1], &Pw[i] ) ;
#endif
            j += SubFrLen ;
        }


        /* Reload the buffer */
        /*
		for ( i = 0 ; i < PitchMax ; i ++ )
            Dpnt[i] = pCoderMem->CodStat.PrevWgt[i] ;
        for ( i = 0 ; i < Frame ; i ++ )
            Dpnt[PitchMax+i] = DataBuff[i] ;
		*/

        /* Save PrevWgt */
        for ( i = 0 ; i < PitchMax ; i ++ )
            pCoderMem->CodStat.PrevWgt[i] = Dpnt[Frame+i] ;

        /* Apply the Harmonic filter */
        j = 0 ;
        for ( i = 0 ; i < SubFrames ; i ++ ) {
            Filt_Pw( DataBuff, Dpnt, (Word16) j , Pw[i] ) ;
            j += SubFrLen ;
        }
        //free ( (char *) Dpnt ) ;

        /* Inverse quantization of the LSP */
        Lsp_Inq( LspVect, pCoderMem->CodStat.PrevLsp, Line.LspId, Line.Crc ) ;

        /* Interpolate the Lsp vectors */
        Lsp_Int( QntLpc, LspVect, pCoderMem->CodStat.PrevLsp ) ;

        /* Copy the LSP vector for the next frame */
        for ( i = 0 ; i < LpcOrder ; i ++ )
            pCoderMem->CodStat.PrevLsp[i] = LspVect[i] ;		

        /*
        // Start the sub frame processing loop
        */
        Dpnt = DataBuff ;

		for (i=0; i<PitchMax; i++)
		{
			Temp[i] = pCoderMem->CodStat.PrevExc[i];
			Temp1[i] = pCoderMem->CodStat.PrevErr[i];
		}
		pExc = &Temp[0];
		pErr = &Temp1[0];

        for ( i = 0 ; i < SubFrames ; i ++ ) {

            /* Compute full impulse response */
#ifdef VC6
            Comp_Ir( ImpResp, &QntLpc[i*LpcOrder], &PerLpc[i*2*LpcOrder], &Pw[i] ) ;
#else
			Comp_Irs( ImpResp, &QntLpc[i*LpcOrder], &PerLpc[i*2*LpcOrder], &Pw[i] ) ;
#endif

            /* Subtract the ringing of previous sub-frame */
#ifdef VC6
            Sub_Ring( Dpnt, &QntLpc[i*LpcOrder], &PerLpc[i*2*LpcOrder],
                                                   pErr/*pCodStat->PrevErr*/, &Pw[i] ) ;
#else 
			Sub_Rings( Dpnt, &QntLpc[i*LpcOrder], &PerLpc[i*2*LpcOrder],
                                                  pErr/*pCodStat->PrevErr*/, &Pw[i] ) ;
#endif

            /* Compute adaptive code book contribution */
#ifdef VC6
            Find_Acbk( Dpnt, ImpResp, pExc/*pCodStat->PrevExc*/, &Line, (Word16) i ) ;
#else
			Find_Acbks( Dpnt, ImpResp, pExc/*pCodStat->PrevExc*/, &Line, (Word16) i ) ;
#endif

            /* Compute fixed code book contribution */
            Find_Fcbk( Dpnt, ImpResp, &Line, (Word16) i ) ;

            /* Reconstruct the excitation */
#ifdef VC6            
            Decod_Acbk( ImpResp, pExc/*pCodStat->PrevExc*/, Line.Olp[i>>1],
                                    Line.Sfs[i].AcLg, Line.Sfs[i].AcGn ) ;
#else
            Decod_Acbks( ImpResp, pExc/*pCodStat->PrevExc*/, Line.Olp[i>>1],
                                    Line.Sfs[i].AcLg, Line.Sfs[i].AcGn ) ;
#endif                                   

            //for ( j = SubFrLen ; j < PitchMax ; j ++ )
            //    pCoderMem->CodStat.PrevExc[j-SubFrLen] = pCoderMem->CodStat.PrevExc[j] ;

            for ( j = 0 ; j < SubFrLen ; j ++ ) {
//                Dpnt[j] = shl( Dpnt[j], (Word16) 1 ) ;
				Dpnt[j] = add(Dpnt[j], Dpnt[j]);
                Dpnt[j] = add( Dpnt[j], ImpResp[j] ) ;
                //pCodStat->PrevExc[PitchMax-SubFrLen+j] = Dpnt[j] ;
				pExc[PitchMax+j] = Dpnt[j];
            }

            /* Update exc_err */
            Update_Err(Line.Olp[i>>1], Line.Sfs[i].AcLg, Line.Sfs[i].AcGn);

            /* Update the ringing delays */
#ifdef VC6
            Upd_Ring( Dpnt, &QntLpc[i*LpcOrder], &PerLpc[i*2*LpcOrder],
                                                       pErr/*pCodStat->PrevErr*/ ) ;			
#else 
           Upd_Rings( Dpnt, &QntLpc[i*LpcOrder], &PerLpc[i*2*LpcOrder],
                                                       pErr/*pCodStat->PrevErr*/ ) ;			
#endif

            Dpnt += SubFrLen ;

			pExc += SubFrLen;
			pErr += SubFrLen;
        }  /* end of subframes loop */

		for (i=0; i<PitchMax; i++)
		{
			pCoderMem->CodStat.PrevExc[i] = Temp[Frame+i];
			pCoderMem->CodStat.PrevErr[i] = Temp1[Frame+i];
		}

        /*
        // Save Vad information and reset CNG random generator
        */
        pCoderMem->CodCng.PastFtyp = 1;
        pCoderMem->CodCng.RandSeed = 12345;

    } /* End of active frame case */

    /* Pack the Line structure */
    frame_len = Line_Pack( &Line, Vout, Ftyp ) ;

    return frame_len ;
}

void g7231_setCoderRate53(void){
	pCoderMem->WrkRate = Rate53;
}

void g7231_setCoderRate63(void){
	pCoderMem->WrkRate = Rate63;
}

Word16 *g7231_getCoderBuff(void){
	return pCoderMem->DataBuff;
}
