/*
**
** File:        "decod.c"
**
** Description:     Top-level source code for G.723 dual-rate decoder
**
** Functions:       Init_Decod()
**                  Decod()
**
**
*/

/*
    ITU-T G.723 Speech Coder   ANSI-C Source Code     Version 5.00
    copyright (c) 1995, AudioCodes, DSP Group, France Telecom,
    Universite de Sherbrooke.  All rights reserved.
*/


//#include <stdlib.h>
//#include <stdio.h>
//#include <string.h>
#include <linux/kernel.h>// 2005/04/20 BrianBin
#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "cst_lbc.h"
#include "tab_lbc.h"
#include "lbccodec.h"
#include "decod.h"
#include "util_lbc.h"
#include "lpc.h"
#include "lsp.h"
#include "exc_lbc.h"
#include "dec_cng.h"
#include "../basic_op/cp_tools.h"// 2005/04/20 BrianBin
#include "../basic_op/cp3.h"

/*
   The following structure contains all the static decoder
      variables.
*/

/*#ifdef GHS_TOOLS
#pragma	ghs section bss=".DecStat"
#endif*/

//DECSTATDEF  DecStat  ;
//DECSTATDEF	*pDecStat = &DecStat;

/*#ifdef GHS_TOOLS
#pragma ghs section bss=default
#endif*/

/*
**
** Function:        Init_Decod()
** 
** Description:     Initializes non-zero state variables
**          for the decoder.
**
** Links to text:   Section 3.11
** 
** Arguments:       None
** 
** Outputs:     None
** 
** Return value:    None
**
*/
void  Init_Decod( )
{
    int   i  ;

    /* Initialize encoder data structure with zeros */
    memset(&pDecodMem->DecStat, 0, sizeof(DECSTATDEF));

    /* Initialize the previously decoded LSP vector to the DC vector */
    for ( i = 0 ; i < LpcOrder ; i ++ )
        pDecodMem->DecStat.PrevLsp[i] = LspDcTable[i] ;

    /* Initialize the gain scaling unit memory to a constant */
    pDecodMem->DecStat.Gain = (Word16) 0x1000 ;

    return;
}


/*
**
** Function:        Decod()
**
** Description:     Implements G.723 dual-rate decoder for  a frame
**          of speech
**
** Links to text:   Section 3
**
** Arguments:
**
**  Word16 *DataBuff    Empty buffer
**  Word16 Vinp[]       Encoded frame (22/26 bytes)
**

** Outputs:
**
**  Word16 DataBuff[]   Decoded frame (480 bytes)
**
** Return value:
**
**  Flag            Always True
**
*/

/*#ifdef GHS_TOOLS
 #pragma ghs section bss=".QntLpc"
#endif*/
#ifdef VC6
Word16 QntLpc[SubFrames*LpcOrder] ;
#else
extern Word16 QntLpc[];
#endif
/*#ifdef GHS_TOOLS
 #pragma ghs section bss=".Temp"
#endif*/
#ifdef VC6
Word16 Temp[PitchMax+Frame+1];
#else
extern Word16 Temp[];
#endif
/*#ifdef GHS_TOOLS
 #pragma ghs section bss=".Temp1"
#endif*/
#ifdef VC6
Word16 Temp1[PitchMax+Frame+1];
#else
extern Word16 Temp1[];
#endif

#ifdef VC6
Word16 Temp2[PitchMax+Frame+1];
#else
extern Word16 Temp2[];
#endif

#ifdef VC6
Word16 Temp3[PitchMax+Frame+1];
#else
extern Word16 Temp3[];
#endif

/*#ifdef GHS_TOOLS
#pragma ghs section bss=default
#endif*/
Flag Decod(Word16 *DataBuff, char *Vinp, Word16 Crc)
{
    int   i, j;
    Word32   Senr ;
//    Word16   QntLpc[SubFrames*LpcOrder] ;
    Word16   AcbkCont[SubFrLen] ;
    Word16   LspVect[LpcOrder] ;
//    Word16   Temp[PitchMax+Frame+1]={0} ;	// +1 for data alignment
//	Word16	 Temp1[PitchMax+Frame+1]={0};	//920529
    Word16  *Dpnt ;

    LINEDEF  Line ;
    PFDEF    Pf[SubFrames] ;

    Word16   Ftyp;

    /*
    * Decode the packed bitstream for the frame.  (Text: Section 4;
    * pars of sections 2.17, 2.18)
    */
    Line = Line_Unpk( Vinp, &Ftyp, Crc ) ;

    /*
    * Update the frame erasure count (Text: Section 3.10)
    */
    if ( Line.Crc != (Word16) 0 ) {
        if(pDecodMem->DecCng.PastFtyp == 1) Ftyp = 1;  /* active */
        else Ftyp = 0;  /* untransmitted */
    }

    if(Ftyp != 1) {

        /* Silence frame : do noise generation */
        Dec_Cng(Ftyp, &Line, DataBuff, QntLpc);
    }

    else {

        /*
        * Update the frame erasure count (Text: Section 3.10)
        */
        if ( Line.Crc != (Word16) 0 )
            pDecodMem->DecStat.Ecount = add( pDecodMem->DecStat.Ecount, (Word16) 1 ) ;
        else
            pDecodMem->DecStat.Ecount = (Word16) 0 ;

        if ( pDecodMem->DecStat.Ecount > (Word16) ErrMaxNum )
            pDecodMem->DecStat.Ecount = (Word16) ErrMaxNum ;
            
        /*
        * Decode the LSP vector for subframe 3.  (Text: Section 3.2)
        */
        Lsp_Inq( LspVect, pDecodMem->DecStat.PrevLsp, Line.LspId, Line.Crc ) ;
        
        /*
        * Interpolate the LSP vectors for subframes 0--2.  Convert the
        * LSP vectors to LPC coefficients.  (Text: Section 3.3)
        */
        Lsp_Int( QntLpc, LspVect, pDecodMem->DecStat.PrevLsp ) ;
        
        /* Copy the LSP vector for the next frame */
        for ( i = 0 ; i < LpcOrder ; i ++ )
            pDecodMem->DecStat.PrevLsp[i] = LspVect[i] ;

        /*
        * In case of no erasure, update the interpolation gain memory.
        * Otherwise compute the interpolation gain (Text: Section 3.10)
        */
        if ( pDecodMem->DecStat.Ecount == (Word16) 0 ) {
            pDecodMem->DecStat.InterGain = add( Line.Sfs[SubFrames-2].Mamp,
                                            Line.Sfs[SubFrames-1].Mamp ) ;
            pDecodMem->DecStat.InterGain = shr( pDecodMem->DecStat.InterGain, (Word16) 1 ) ;
            pDecodMem->DecStat.InterGain = FcbkGainTable[pDecodMem->DecStat.InterGain] ;
        }
        else
            pDecodMem->DecStat.InterGain = mult_r( pDecodMem->DecStat.InterGain, (Word16) 0x6000 ) ;


        /*
        * Generate the excitation for the frame
        */
        for ( i = 0 ; i < PitchMax ; i ++ )
            Temp[i] = pDecodMem->DecStat.PrevExc[i] ;

        Dpnt = &Temp[PitchMax] ;

        if ( pDecodMem->DecStat.Ecount == (Word16) 0 ) {
// jason++ 20050316				
//            for ( i = 0 ; i < SubFrames - 1 ; i ++ ) {
            for ( i = 0 ; i < SubFrames ; i ++ ) {

                /* Generate the fixed codebook excitation for a
                   subframe. (Text: Section 3.5) */
                Fcbk_Unpk( Dpnt, Line.Sfs[i], Line.Olp[i>>1], (Word16) i ) ;

                /* Generate the adaptive codebook excitation for a
                   subframe. (Text: Section 3.4) */
#ifdef VC6
                Decod_Acbk( AcbkCont, &Temp[SubFrLen*i], Line.Olp[i>>1],
                                    Line.Sfs[i].AcLg, Line.Sfs[i].AcGn ) ;
#else
                Decod_Acbks( AcbkCont, &Temp[SubFrLen*i], Line.Olp[i>>1],
                                    Line.Sfs[i].AcLg, Line.Sfs[i].AcGn ) ;
#endif

                /* Add the adaptive and fixed codebook contributions to
                   generate the total excitation. */
                for ( j = 0 ; j < SubFrLen ; j ++ ) {
//                    Dpnt[j] = shl( Dpnt[j], (Word16) 1 ) ;
					Dpnt[j] = add(Dpnt[j], Dpnt[j]);
                    Dpnt[j] = add( Dpnt[j], AcbkCont[j] ) ;
					DataBuff[i*SubFrLen+j] = Dpnt[j];		// 920529:merge
                }

                Dpnt += SubFrLen ;
            }
		//printk("DataBuff: 0x%04x, 0x%04x\n", (unsigned short) DataBuff[0], (unsigned short) DataBuff[1]);

            /* Save the excitation */
            /*for ( j = 0 ; j < Frame ; j ++ )		// 920529:merge to upper loop
                DataBuff[j] = Temp[PitchMax+j] ;*/

            /* Compute interpolation index. (Text: Section 3.10) */
            /* Use pDecCng->SidGain and pDecCng->CurGain to store    */
            /* excitation energy estimation                      */
#ifdef VC6
            pDecodMem->DecStat.InterIndx = Comp_Info( Temp, Temp1, Line.Olp[SubFrames/2-1],
                                       &pDecodMem->DecCng.SidGain, &pDecodMem->DecCng.CurGain ) ;
#else
			pDecodMem->DecStat.InterIndx = Comp_Infos( Temp, Temp1, Line.Olp[SubFrames/2-1],
                                     &pDecodMem->DecCng.SidGain, &pDecodMem->DecCng.CurGain ) ;
#endif

            /* Compute pitch post filter coefficients.  (Text: Section 3.6) */            
            if ( UsePf )
                for ( i = 0 ; i < SubFrames ; i ++ )
#ifdef VC6
					Comp_Lpf( Temp1, Line.Olp[i>>1], (Word16) i, &Pf[i]);
#else	
					Comp_Lpfs( Temp1, Line.Olp[i>>1], (Word16) i, &Pf[i]);
#endif

            /* Reload the original excitation */		// cz 920529
            /*for ( j = 0 ; j < PitchMax ; j ++ )
                Temp[j] = DecStat.PrevExc[j] ;
            for ( j = 0 ; j < Frame ; j ++ )
                Temp[PitchMax+j] = DataBuff[j] ;*/

            /* Perform pitch post filtering for the frame.  (Text: Section
               3.6) */
            if ( UsePf )
                for ( i = 0 ; i < SubFrames ; i ++ )
#ifdef VC6                
                    Filt_Lpf( DataBuff, Temp, &Pf[i], (Word16) i ) ;
#else
                    Filt_Lpfs( DataBuff, Temp, &Pf[i], (Word16) i ) ;
#endif           

            /* Save Lsps --> LspSid */
            for(i=0; i< LpcOrder; i++)
                pDecodMem->DecCng.LspSid[i] = pDecodMem->DecStat.PrevLsp[i];
        }

        else {

            /* If a frame erasure has occurred, regenerate the
               signal for the frame. (Text: Section 3.10) */
            Regen( DataBuff, Temp, pDecodMem->DecStat.InterIndx, pDecodMem->DecStat.InterGain,
                                        pDecodMem->DecStat.Ecount, &(pDecodMem->DecStat.Rseed) ) ;
        }

        /* Update the previous excitation for the next frame */
        for ( j = 0 ; j < PitchMax ; j ++ )
            pDecodMem->DecStat.PrevExc[j] = Temp[Frame+j] ;

        /* Resets random generator for CNG */
       pDecodMem->DecCng.RandSeed = 12345;
    }

    /* Save Ftyp information for next frame */
    pDecodMem->DecCng.PastFtyp = Ftyp;

	/*
    * Synthesize the speech for the frame
    */
//	memDump(QntLpc, SubFrames*LpcOrder, "QntLpc");
	Dpnt = DataBuff ;
	for ( i = 0 ; i < SubFrames ; i ++ ) 
	{
		/* Compute the synthesized speech signal for a subframe.
         * (Text: Section 3.7)
         */
#ifdef VC6
		Synt( Dpnt, &QntLpc[i*LpcOrder] ) ;
#else
		Synts2( Dpnt, &QntLpc[i*LpcOrder]);
#endif
    	printk("SubFrame=%d [79]=0x%04x [89]=0x%04x [155]=0x%04x [234]=0x%04x\n", i, DataBuff[79], DataBuff[89], DataBuff[155], DataBuff[234]);
 
		if ( UsePf ) 
		{

            /* Do the formant post filter. (Text: Section 3.8) */
#ifdef VC6
			Senr = Spf( Dpnt, &QntLpc[i*LpcOrder] ) ;
#else
			Senr = Spfs( Dpnt, &QntLpc[i*LpcOrder] ) ;	// cause a little diff in 111638 char
#endif

          /* Do the gain scaling unit.  (Text: Section 3.9) */
#ifdef VC6
			Scale( Dpnt, Senr ) ;
#else
			Scales( Dpnt, Senr ) ;
#endif
        }

		Dpnt += SubFrLen ;
    }
	return (Flag) True ;
}

