/*
**
** File:        "lbccodec.c"
**
** Description:     Top-level source code for G.723 dual-rate codec
**
** Functions:       main
**                  Process_files()
**
**
*/


/*
    ITU-T G.723 Speech Coder   ANSI-C Source Code     Version 5.00
    copyright (c) 1995, AudioCodes, DSP Group, France Telecom,
    Universite de Sherbrooke.  All rights reserved.
*/

/*
	Profiling: 10 sec test sequence, 333 frames
	(1) coder: cost 3.56s to processed @ 100 MHz
		(3.56/10)*100MHz = 35.6 MIPS
	(2) decoder: cost 0.48s to processed @ 100MHz
		(0.48/10)*100Mhz = 4.8 MIPS	
*/

//#include <cyg/infra/diag.h>	// diag_printf()
//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>
//#include <time.h>

#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "cst_lbc.h"
#include "tab_lbc.h"
#include "lbccodec.h"
#include "coder.h"
#include "decod.h"
#include "exc_lbc.h"
#include "util_lbc.h"
#include "cod_cng.h"
#include "dec_cng.h"
#include "vad.h"

#include "../basic_op/cp3.h"
#include "../basic_op/cp_tools.h"



#ifndef VC6
void set_mmd(Word16);
#endif

void Tests(void);
void Firs(Word32 *Acc0, Word32 *Acc1, Word16 *ptr1, Word16 *ptr2);

/* Global variables */
#ifdef GNU
__attribute__ ((aligned (8)))	// for GNU
#endif
static CoderMem Coder0;    //channel 0 coder memory
CoderMem *pCoderMem = &Coder0;
#ifdef GNU
__attribute__ ((aligned (8)))	// for GNU
#endif
static DecodMem Decoder0;  //channel 0 decoder memory
DecodMem *pDecodMem = &Decoder0;


Wmode	WrkMode = Both;
//enum  Crate    WrkRate = Rate63 ;

int   PackedFrameSize[2] = {
   24 ,
   20
   } ;

Flag    UseHp = True ;
Flag    UsePf = True ;
Flag    UseVx = False ;
Flag    UsePr = True ;

/*#ifdef GHS_TOOLS
#pragma ghs section bss=".DataBuff"*/
//Word16  DataBuff[PitchMax+Frame+1] ;	// +1 for data alignment
/*#ifdef VC6
Word16  DataBuff[PitchMax+Frame+1] ;	// +1 for data alignment
#else
extern Word16 DataBuff[];
#endif
*/
BESTDEF Best;
LINEDEF Line1;

/*#pragma ghs section bss=default

#else
Word16  DataBuff[Frame] ;

#endif*/


#if DANWU
/*
   This function processes the argument parameters. The function
      opens the IO files, and sets the global arguments accordingly
      to the command line parameters.
*/
long  Process_Files( FILE **Ifp, FILE **Ofp, FILE **Fep, FILE **Ratp,
                                        int Argc, char *Argv[] )
{
    int     i ;
    long    Flen ;
    char    *FerFileName = NULL ;
    char    *RateFileName = NULL ;

    /*
      Process the argument list, if any
    */
    if (Argc < 3 ) {
        diag_printf("Usage: %s [options] inputfile outputfile \n", Argv[0]);
        assert(0);
    }

    for ( i = 1 ; i < Argc-2 ; i ++ ) {

        /* Check the coder rate */
        if ( ! strncmp( "-r", Argv[i], 2) ) {
            if ( ! strcmp("63", Argv[i]+2) ) {
                pCoderMem->WrkRate = Rate63 ;
                continue ;
            }

            else if ( ! strcmp("53", Argv[i]+2) ) {
                pCoderMem->WrkRate = Rate53 ;
                continue ;
            }

            else {
                RateFileName = &Argv[i][2] ;
                continue ;
            }
        }

        /* Check Working mode */
        if ( ! strcmp("-b", Argv[i]) ) {
            WrkMode = Both ;
            continue ;
        }

        if ( ! strcmp("-c", Argv[i]) ) {
            WrkMode = Cod ;
            continue ;
        }

        if ( ! strcmp("-d", Argv[i]) ) {
            WrkMode = Dec ;
            continue ;
        }

        if ( ! strcmp("-v", Argv[i]) ) {
            UseVx = True ;
            continue ;
        }

        if ( ! strcmp("-Noh", Argv[i]) ) {
            UseHp = False;
            continue ;
        }

        if ( ! strcmp("-Nop", Argv[i]) ) {
            UsePf = False;
            continue ;
        }

        if ( ! strncmp( "-f", Argv[i], 2) ) {
            FerFileName = &Argv[i][2] ;
            continue ;
        }

        if ( ! strcmp("-n", Argv[i]) ) {
            UsePr = False;
            continue ;
        }

        fprintf(stderr, "Illegal argument, %s\n", Argv[i] ) ;
        assert(0) ;
    }

    *Ifp = fopen( Argv[Argc-2], "rb") ;
    if ( *Ifp == NULL ) {
        fprintf(stderr, "Invalid input file name: %s\n", Argv[Argc-2] ) ;
        assert(0) ;
    }

    if ( UsePr )
        diag_printf("Input  file:     %s\n", Argv[Argc-2] ) ;

    *Ofp = fopen( Argv[Argc-1], "wb") ;
    if ( *Ofp == NULL ) {
        fprintf(stderr, "Can't open output file: %s\n", Argv[Argc-1] ) ;
        assert(0) ;
    }
    if ( UsePr )
        diag_printf("Output file:     %s\n", Argv[Argc-1] ) ;

    /* Open Fer file if required */
    if ( WrkMode == Dec ) {

        if ( FerFileName != NULL ) {
            *Fep = fopen( FerFileName, "rb" ) ;
            if ( *Fep == NULL ) {
                fprintf(stderr, "Can't open FER file: %s\n", FerFileName ) ;
                assert(0) ;
            }
            if ( UsePr )
                diag_printf("FER    file:     %s\n", FerFileName ) ;
        }
    }
    else {
        if ( RateFileName != NULL ) {
            *Ratp = fopen( RateFileName, "rb" ) ;
            if ( *Ratp == NULL ) {
                fprintf(stderr, "Can't open Rate file: %s\n", RateFileName ) ;
                assert(0) ;
            }
            if ( UsePr )
                diag_printf("Rate   file:     %s\n", RateFileName ) ;
        }
    }

    /* Options report */
    if ( UsePr ) {

        diag_printf("Options:\n");
        if (WrkMode == Both )
            diag_printf("Encoder/Decoder\n");
        else if (WrkMode == Cod )
            diag_printf("Encoder\n");
        else
            diag_printf("Decoder\n");


        if( WrkMode != Cod ) {
            if (UsePf == 0 )
                diag_printf("Postfilter disabled\n");
            else
                diag_printf("Postfilter enabled\n");
        }

        if( WrkMode <= Cod ) {
            if(*Ratp == NULL) {
                if (pCoderMem->WrkRate == Rate63 )
                    diag_printf("Rate 6.3 kb/s\n");
                else
                    diag_printf("Rate 5.3 kb/s\n");
            }
            if (UseHp == 0 )
                diag_printf("Highpassfilter disabled\n");
            else
                diag_printf("Highpassfilter enabled\n");
            if (UseVx == 0 )
                diag_printf("VAD/CNG disabled\n");
            else
                diag_printf("VAD/CNG enabled\n");
        }
    }

    /*
      Compute the file length
    */
    fseek( *Ifp, 0L, SEEK_END ) ;
    Flen = ftell( *Ifp ) ;
    rewind( *Ifp ) ;
    if ( WrkMode == Dec )
        Flen = 0x7fffffffL ;
    else
        Flen /= sizeof(Word16)*Frame ;

    return Flen ;
}



unsigned int readreg(unsigned int addr, short bytes)
{
	char *paddr;
	unsigned int data;
	
	paddr = (char *)(addr);
	data = 0;
	while(bytes-- > 0)
	{
		data = (data << 8) | (*paddr++ & 0x000000ff);
	}
	return data;
}

void writereg(unsigned int addr, unsigned int data, short bytes)
{
	char *paddr, *pdata;
	
	paddr = (char*)(addr);
	pdata = (char *)(&data);
	while(bytes-- > 0)
	{
		*paddr++ = *pdata++;
	}
	return;
}


char  SignOn[] = "ACL/USH/FT/DSPG ANSI C CODEC ITU LBC Ver 5.00\n" ;
#ifdef VC6
int testmain( int argc, char *argv[] )
#else
int testmain (void)
#endif
{
    FILE    *Ifp=NULL, *Ofp=NULL ;            /* I/O File pointers */
    FILE    *Fep  = NULL;           /* Frame erasures file pointer */
    FILE    *Ratp = NULL;           /* Rate file pointer */

    long    FrCnt = 0 ;
    long    FlLen ;
    Flag	reset;
    
    Word16  Crc, i, j;
    char 	Rate_Rd;
	char    Line[24] ;
	int		indx=0;
	extern 	char r63vbits[];
	int		good=0, bad=0, gbflag=0;
	
	    
#ifndef VC6
	set_mmd(0x2);		// set to integer mode and 32 bit saturation

	// check endian system
  	{
  		Word16	pattern = 0x1234;
  		char	*ptr = (char *)&pattern;
  	
  		if (*ptr == 0x12)
  			diag_printf("Big-endian target ...\n");
  		else if (*ptr == 0x34)
  			diag_printf("Little-endian target ...\n");
  		else {
  			diag_printf("Unknow system ... exit()\n");
  			assert(0);
  		}
  	}
  	
  	// get system status and turn off some interrupts
	{		
		unsigned int gimr,gpio;
		
		gimr = readreg(0xbd010000, sizeof(int));
		diag_printf("GIMR=0x%x\n", gimr);		
		//gimr = gimr & 0xfffffff6;				//off UART and timer intr(often-)
		//gimr = gimr & 0xfffffff7;				//off UART intr (often)
		//gimr = gimr & 0xfffffffe;				//off timer intr (often)		
		gimr = gimr & 0xffffff00;				//off timer nic uart (never)
		//gimr = gimr & 0xffffff08;				//off timer nic (never)
		writereg(0xbd010000, gimr, sizeof(int));
		gimr = readreg(0xbd010000, sizeof(int));
		diag_printf("GIMR=0x%x\n", gimr);		
		gpio = readreg(0xbd010040, sizeof(int));
		diag_printf("PABDIR=0x%x\n", gpio);		
		gpio = readreg(0xbd010044, sizeof(int));
		diag_printf("PABDATA=0x%x\n", gpio);	
		
	}
#endif

	/* cp3 setting */
	cp3_init();	
	cp3_set_counter_usage(0, CP3_COUNT_CYCLE);
	cp3_set_counter_usage(1, CP3_COUNT_NEW_INST_FETCH);
	cp3_set_counter_usage(2, CP3_COUNT_INST_CACHE_MISS);
	cp3_set_counter_usage(3, CP3_COUNT_LOAD_STORE_MISS);		
	//cp3_set_iram_addr((void*)0x00601000, (void*)(0x00601000 + 0x1ff0));		// 8k IRAM
	//cp3_set_dram_addr_8181((void*)0x800d4000, (void*)(0x800d4000 + 0x0ff0));		// 4k DRAM
	cp3_set_dram_addr_8181((void*)0x800fc000, (void*)(0x800fc000 + 0x0ff0));		// 4k DRAM			
	cp3_start();

            
    /* in case of decoder, women63v.rco */
 while(1)
 {
 	
 	// clear variable
 	/*{
 		int	len=(0x80101000-0x800fc000)/4;
 		int	i=0;
 		unsigned int *ptr;
 		
 		ptr = (unsigned int *)0x800fc000;
 		 		
 		for(i=0;i<len; i++)
 			*ptr++ = 0;	
 		
 		len = (0x80132d10-0x80132598)/4;
 		ptr = (unsigned int *)0x80132598;
 		for(i=0;i<len;i++)
 			*ptr++ = 0;
 		
 	}*/
 	
    /*
      Init coder and the decoder
    */
    Init_Coder( ) ;
    Init_Decod( ) ;

    /* Init Comfort Noise Functions */
    if( UseVx ) {
        Init_Vad();
        Init_Cod_Cng( );
    }
    Init_Dec_Cng( );
    
    // set codec parameters
    FlLen = 333;
    pCoderMem->WrkRate = Rate53;
    WrkMode = Cod;
    FrCnt = 0;
    reset = 1;
    

    /* Process all the input file */
    do {

        switch ( WrkMode ) {

            case Both:
                if(Ratp != NULL) {
                    fread((char *)&Rate_Rd, sizeof(char), 1, Ratp);
                    pCoderMem->WrkRate = (Crate)Rate_Rd;
                }
                if ( pCoderMem->WrkRate == Rate53) reset_max_time();
                Read_lbc( pCoderMem->DataBuff, Frame,reset) ;
                i=Coder( pCoderMem->DataBuff, Line, FrCnt ) ;
                Decod( Line, pCoderMem->DataBuff, (Word16) 0 ) ;
                Write_lbc( pCoderMem->DataBuff, Frame, Ofp ) ;
            break ;

            case Cod :
            	/*
                if(Ratp != NULL) {
                    fread((char *)&Rate_Rd, sizeof(char), 1, Ratp);
                    WrkRate = (enum Crate)Rate_Rd;
                }*/
                if ( pCoderMem->WrkRate == Rate53) reset_max_time();
                Read_lbc( pCoderMem->DataBuff, Frame, reset) ;
                
                Coder( pCoderMem->DataBuff, Line, FrCnt ) ;

                if(reset == 1)
                	reset = 0;
                
                //Line_Wr( Line, Ofp ) ;
            break ;

            case Dec :
                /*if(Line_Rd( Line, Ifp ) == (-1)) {
                    FlLen = FrCnt;
                    break;
                }
                if ( Fep == NULL )
                    Crc = (Word16) 0 ;
                else
                    fread( (char *)&Crc, sizeof(Word16), 1, Fep ) ;
                */
                Line_Rd(Line, r63vbits);
                Crc = (Word16) 0;
                Decod( Line, pCoderMem->DataBuff, Crc ) ;
                
                if(reset ==1)
                	reset = 0;
                //Write_lbc( DataBuff, Frame, Ofp ) ;
            break ;
        }

        FrCnt ++ ;


	}   while ( FrCnt < FlLen ) ;
	
	// cp3 stop	
	cp3_stop();
	cp3_get_counters();
	MyWrite("\n\nCoprocessor 3 statitics:\n");
	cp3_show_counter(0, -1);
	cp3_show_counter(1, 0);
	cp3_show_counter(2, 1);
	cp3_show_counter(3, 1);
	
	
if(WrkMode == Dec)
{
	MyWrite("decoder segment:\n");
	for(i=0;i<Frame;i+=4)
		diag_printf("[%d]=0x%x [%d]=0x%x [%d]=0x%x [%d]=0x%x\n", i,pCoderMem->DataBuff[i], i+1,pCoderMem->DataBuff[i+1],
								i+2, pCoderMem->DataBuff[i+2], i+3, pCoderMem->DataBuff[i+3]); 
}	
else {
    MyWrite("coder segment:\n");
    if(pCoderMem->WrkRate == Rate63)
    	j=24;
    else
    	j=20;
    	
    for(i=0;i<j; i+=4)
    	diag_printf("[%d]=0x%x [%d]=0x%x [%d]=0x%x [%d]=0x%x\n", i,Line[i], i+1,Line[i+1],
									i+2, Line[i+2], i+3, Line[i+3]); 
	
	// compare enc output (6.3k)
	gbflag = 0;
	
    if(pCoderMem->WrkRate == Rate63)
    {
    	extern char r63vbits[];
    	static int	ind=0;
    	int	i,j;                	
    	
    	if (reset == 1)
    		ind = 0;
    	
    	for(i=0;i<24;i++)
    	{
    		if(Line[i] != r63vbits[ind++])
    		{
    			//diag_printf("FrCnt=%ld\n", FrCnt);
    			gbflag=1;
    			FrCnt = 333;
    			for(j=0; j<24; j++)
    				diag_printf("Linr[%d]=0x%x\n", j, Line[j]);
    			//assert(0);
    			break;
    		}
    	}		
    }
                
	// compare last frame enc output (5.3k)
	if (pCoderMem->WrkRate == Rate53)
	{
		char women5k3[]={0xbd, 0x6d, 0x9c, 0x8c, 0x10, 0x19, 0xac, 0x48, 0x7a, 0x98, 0x17, 0x5a, 
							0x20, 0x7d, 0x42, 0x69, 0xd7, 0x8d, 0x85, 0x5d};
		Word16 i;
		
		for(i=0; i<20; i++)
		{
			if(Line[i] != women5k3[i])
			{
				diag_printf("women 5.3k coder output error!!\n");
				gbflag = 1;
				break;
			}
		}
	}									
	
}							

	if(gbflag == 0)
		good++;
	else
		bad++;		
	
	diag_printf("\nGood: %d, Bad: %d\n", good, bad);
	
	//while(bad != 0) ;
		
	
 }

    if(Ifp) { (void)fclose(Ifp); }
    if(Ofp) { (void)fclose(Ofp); }
    if(Fep) { (void)fclose(Fep); }
    if(Ratp) { (void)fclose(Ratp); }
    return 0 ;
}
 #endif

void g7231_setHighPassFilter(Flag Hp)
{
	UseHp = Hp;
}


