/*
**

** File:        "vad.c"
**
** Description:     Voice Activity Detection
**
** Functions:       Init_Vad()
**                  Vad()
**
**
*/
/*
    ITU-T G.723 Speech Coder   ANSI-C Source Code     Version 5.00
    copyright (c) 1995, AudioCodes, DSP Group, France Telecom,
    Universite de Sherbrooke.  All rights reserved.
*/

//#include <stdio.h>

#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "cst_lbc.h"
#include "tab_lbc.h"
#include "lsp.h"
#include "vad.h"
#include "coder.h"
#include "lbccodec.h"

//VADSTATDEF  VadStat ;
//VADSTATDEF  *pVadStat = &VadStat ;

void    Init_Vad(void)
{
    int i ;
    pCoderMem->VadStat.Hcnt = 3 ;
    pCoderMem->VadStat.Vcnt = 0 ;
    pCoderMem->VadStat.Penr = 0x00000400L ;
    pCoderMem->VadStat.Nlev = 0x00000400L ;

    pCoderMem->VadStat.Aen = 0 ;

    pCoderMem->VadStat.Polp[0] = 1 ;
    pCoderMem->VadStat.Polp[1] = 1 ;
    pCoderMem->VadStat.Polp[2] = SubFrLen ;
    pCoderMem->VadStat.Polp[3] = SubFrLen ;

    for(i=0; i < LpcOrder; i++) pCoderMem->VadStat.NLpc[i] = 0;

}

Flag Comp_Vad( Word16 *Dpnt)
{
    int i,j ;

    Word32  Acc0,Acc1 ;
    Word16  Tm0, Tm1, Tm2 ;
    Word16  Minp ;
    Flag    VadState = 1 ;
	Word16	Tmp[LpcOrder+2];	//+2 for alignment

    static  Word16  ScfTab[11] = {
         9170 ,
         9170 ,
         9170 ,
         9170 ,
        10289 ,
        11544 ,
        12953 ,
        14533 ,
        16306 ,
        18296 ,
        20529 ,
    } ;

    if ( !UseVx )
        return VadState ;

    /* Find Minimum pitch period */
    Minp = PitchMax ;
    for ( i = 0 ; i < 4 ; i ++ ) {
        if ( Minp > pCoderMem->VadStat.Polp[i] )
            Minp = pCoderMem->VadStat.Polp[i] ;
    }

    /* Check that all are multiplies of the minimum */
    Tm2 = 0 ;
    for ( i = 0 ; i < 4 ; i ++ ) {
        Tm1 = Minp ;
        for ( j = 0 ; j < 8 ; j ++ ) {
            Tm0 = sub( Tm1, pCoderMem->VadStat.Polp[i] ) ;
            Tm0 = abs_s( Tm0 ) ;
            if ( Tm0 <= 3 )
                Tm2 ++ ;
            Tm1 = add( Tm1, Minp ) ;
        }
    }

    /* Update adaptation enable counter if not periodic and not sine */
    if ( (Tm2 == 4) || (pCoderMem->CodStat.SinDet < 0) )	//SinDet < 0 means a sine tone present
        pCoderMem->VadStat.Aen += 2 ;
    else
        pCoderMem->VadStat.Aen -- ;

    /* Clip it */
    if ( pCoderMem->VadStat.Aen > 6 )
        pCoderMem->VadStat.Aen = 6 ;
    if ( pCoderMem->VadStat.Aen < 0 )
        pCoderMem->VadStat.Aen = 0 ;

    /* Inverse filter the data */
	for(i=0; i<LpcOrder; i++)
		Tmp[LpcOrder-1-i] = pCoderMem->VadStat.NLpc[i];
	Tmp[LpcOrder]=Tmp[LpcOrder+1]=0;	//pad 2 zero

    Acc1 = 0L ;
    for ( i = SubFrLen ; i < Frame ; i ++ ) {
		Acc0 = (Word32) 0;
        //for ( j = 0 ; j < LpcOrder ; j ++ )
        //    Acc0 = L_msu( Acc0, Dpnt[i-j-1], Tmp[j] ) ;
		for (j=LpcOrder-1; j>=0; j--)
			Acc0 = L_msu( Acc0, Dpnt[i-j-1], Tmp[LpcOrder-1-j]);
		//Acc0 = L_mult( Dpnt[i], 0x2000 ) ;
		Acc0 = L_mac(Acc0, Dpnt[i], 0x2000);
        Tm0 = round ( Acc0 ) ;
        Acc1 = L_mac( Acc1, Tm0, Tm0 ) ;
    }

    /* Scale the rezidual energy */
    Acc1 = L_mls( Acc1, (Word16) 2913 ) ;

    /* Clip noise level in any case */
    if ( pCoderMem->VadStat.Nlev > pCoderMem->VadStat.Penr ) {
        Acc0 = L_sub( pCoderMem->VadStat.Penr, L_shr( pCoderMem->VadStat.Penr, 2 ) ) ;
        pCoderMem->VadStat.Nlev = L_add( Acc0, L_shr( pCoderMem->VadStat.Nlev, 2 ) ) ;
    }


    /* Update the noise level, if adaptation is enabled */
    if ( !pCoderMem->VadStat.Aen ) {
        pCoderMem->VadStat.Nlev = L_add( pCoderMem->VadStat.Nlev, L_shr( pCoderMem->VadStat.Nlev, 5 ) ) ;
    }
    /* Decay Nlev by small amount */
    else {
        pCoderMem->VadStat.Nlev = L_sub( pCoderMem->VadStat.Nlev, L_shr( pCoderMem->VadStat.Nlev,11 ) ) ;
    }

    /* Update previous energy */
    pCoderMem->VadStat.Penr = Acc1 ;

    /* CLip Noise Level */
    if ( pCoderMem->VadStat.Nlev < 0x00000080L )
        pCoderMem->VadStat.Nlev = 0x00000080L ;
    if ( pCoderMem->VadStat.Nlev > 0x0001ffffL )
        pCoderMem->VadStat.Nlev = 0x0001ffffL ;

    /* Compute the treshold */
    Acc0 = L_shl( pCoderMem->VadStat.Nlev, 13 ) ;	// Nlev > 0
    Tm0 = norm_l( Acc0 ) ;
    Acc0 = L_shl( Acc0, Tm0 ) ;
    Acc0 &= 0x3f000000L ;
    Acc0 <<= 1 ;
    Tm1 = extract_h( Acc0 ) ;
    Acc0 = L_deposit_h( ScfTab[Tm0] ) ;
    Acc0 = L_mac( Acc0, Tm1, ScfTab[Tm0-1] ) ;
    Acc0 = L_msu( Acc0, Tm1, ScfTab[Tm0] ) ;
    Tm1 = extract_h( Acc0 ) ;
    Tm0 = extract_l( L_shr( pCoderMem->VadStat.Nlev, 2 ) ) ;
    Acc0 = L_mult( Tm0, Tm1 ) ;
    Acc0 >>= 11 ;

    /* Compare with the treshold */
    if ( Acc0 > Acc1 )
        VadState = 0 ;

    /* Do the various counters */
    if ( VadState ) {
        pCoderMem->VadStat.Vcnt ++ ;
        pCoderMem->VadStat.Hcnt ++ ;
    }
    else {
        pCoderMem->VadStat.Vcnt -- ;
        if ( pCoderMem->VadStat.Vcnt < 0 )
            pCoderMem->VadStat.Vcnt = 0 ;
    }

    if ( pCoderMem->VadStat.Vcnt >= 2 ) {
        pCoderMem->VadStat.Hcnt = 6 ;
        if ( pCoderMem->VadStat.Vcnt >= 3 )
            pCoderMem->VadStat.Vcnt = 3 ;
    }

    if ( pCoderMem->VadStat.Hcnt ) {
        VadState = 1 ;
        if ( pCoderMem->VadStat.Vcnt == 0 )
            pCoderMem->VadStat.Hcnt -- ;
    }

    /* Update Periodicy detector */
    pCoderMem->VadStat.Polp[0] = pCoderMem->VadStat.Polp[2] ;
    pCoderMem->VadStat.Polp[1] = pCoderMem->VadStat.Polp[3] ;

    return VadState ;
}







