/*
   ITU-T G.729A Speech Coder with Annex B    ANSI-C Source Code
   Version 1.3    Last modified: August 1997

   Copyright (c) 1996,
   AT&T, France Telecom, NTT, Universite de Sherbrooke, Lucent Technologies,
   Rockwell International
   All rights reserved.
*/

/*-----------------------------------------------------------------*
 *   Functions Coder_ld8a and Init_Coder_ld8a                      *
 *             ~~~~~~~~~~     ~~~~~~~~~~~~~~~                      *
 *                                                                 *
 *  Init_Coder_ld8a(void);                                         *
 *                                                                 *
 *   ->Initialization of variables for the coder section.          *
 *                                                                 *
 *                                                                 *
 *  Coder_ld8a(Word16 ana[]);                                      *
 *                                                                 *
 *   ->Main coder function.                                        *
 *                                                                 *
 *                                                                 *
 *  Input:                                                         *
 *                                                                 *
 *    80 speech data should have been copy to vector new_speech[]. *
 *    This vector is global and is declared in this function.      *
 *                                                                 *
 *  Ouputs:                                                        *
 *                                                                 *
 *    ana[]      ->analysis parameters.                            *
 *                                                                 *
 *-----------------------------------------------------------------*/

#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "g729_ld8a.h"
#include "g729_vad.h"
#include "g729_dtx.h"
#include "g729_sid.h"
#include "mystruct.h"

/* temp used */
/* should align 8 */
/*
#ifdef GHS_TOOLS
#pragma ghs section data=".CodStatic"
#endif
*/

#ifdef GNU
	__attribute__ ((aligned (4096)))	/* for GNU */
	__attribute__ ((section(".bss.h1")))
#endif
Word16 h1[L_SUBFR]; /*Impulse response h1[40] L_SUBFR = 40*/

#ifdef GNU
__attribute__ ((aligned (8)))	/* for GNU */
__attribute__ ((section(".bss.xn")))
#endif
/* Target vector for pitch search : adaptive    */
Word16 xn[L_SUBFR+4]; 

#ifdef GNU
__attribute__ ((aligned (8)))	/* for GNU */
__attribute__ ((section(".bss.xn2")))
#endif
/* Target vector for codebook search : for ACELP  */  
Word16 xn2[L_SUBFR+4];

/*-----------------------------------------------------------*
 *    Coder constant parameters (defined in "ld8a.h")        *
 *-----------------------------------------------------------*
 *   L_WINDOW    : LPC analysis window size.                 *
 *   L_NEXT      : Samples of next frame needed for autocor. *
 *   L_FRAME     : Frame size.                               *
 *   L_SUBFR     : Sub-frame size.                           *
 *   M           : LPC order.                                *
 *   MP1         : LPC order+1                               *
 *   L_TOTAL     : Total size of speech buffer.              *
 *   PIT_MIN     : Minimum pitch lag.                        *
 *   PIT_MAX     : Maximum pitch lag.                        *
 *   L_INTERPOL  : Length of filter for interpolation        *
 *-----------------------------------------------------------*/

/*--------------------------------------------------------*
 *         Static memory allocation.                      *
 *--------------------------------------------------------*/

        /* Speech vector */

// static Word16 old_speech[L_TOTAL];
// static Word16 *speech, *p_window;
// Word16 *new_speech;                    /* Global variable */

        /* Weighted speech vector */

// static Word16 old_wsp[L_FRAME+PIT_MAX];
// static Word16 *wsp;

        /* Excitation vector */

// static Word16 old_exc[L_FRAME+PIT_MAX+L_INTERPOL];
// static Word16 *exc;

        /* Lsp (Line spectral pairs) */

// static Word16 lsp_old[M]={
//              30000, 26000, 21000, 15000, 8000, 0, -8000,-15000,-21000,-26000};
// static Word16 lsp_old_q[M];

        /* Filter's memory */

// static Word16  mem_w0[M], mem_w[M], mem_zero[M];
// static Word16  sharp;

        /* For G.729B */
        /* DTX variables */
// static Word16 pastVad;   
// static Word16 ppastVad;
// static Word16 seed;

// Word16 dram1[M];	// just for Lx5280 load twinword consideration,
					// not need to store these variables
/*
void StoreCodld8aStaticVar(Codld8aStaticVar *ptr)
{
	int i;
	
	for (i=0; i<L_TOTAL; i++)
		ptr->old_speech[i] = old_speech[i];

	ptr->speech = speech; ptr->p_window = p_window;

	for (i=0; i<L_FRAME+PIT_MAX; i++)
		ptr->old_wsp[i] = old_wsp[i];

	ptr->wsp = wsp;

	for (i=0; i<L_FRAME+PIT_MAX+L_INTERPOL; i++)
		ptr->old_exc[i] = old_exc[i];

	ptr->exc = exc;

	for (i=0; i<M; i++)
	{
		ptr->lsp_old[i] = lsp_old[i];
		ptr->lsp_old_q[i] = lsp_old_q[i];
		ptr->mem_w0[i] = mem_w0[i];
		ptr->mem_w[i] = mem_w[i];
//		ptr->mem_zero[i] = mem_zero[i];
	}

	ptr->sharp = sharp;

	ptr->pastVad = pastVad;
	ptr->ppastVad = ppastVad;
	ptr->seed = seed;
}

void LoadCodld8aStaticVar(Codld8aStaticVar *ptr)
{
	int i;
	
	for (i=0; i<L_TOTAL; i++)
		old_speech[i] = ptr->old_speech[i];

	speech = ptr->speech ; p_window = ptr->p_window ;

	for (i=0; i<L_FRAME+PIT_MAX; i++)
		old_wsp[i] = ptr->old_wsp[i];

	wsp = ptr->wsp;

	for (i=0; i<L_FRAME+PIT_MAX+L_INTERPOL; i++)
		old_exc[i] = ptr->old_exc[i];

	exc = ptr->exc;

	for (i=0; i<M; i++)
	{
		lsp_old[i] = ptr->lsp_old[i];
		lsp_old_q[i] = ptr->lsp_old_q[i];
		mem_w0[i] = ptr->mem_w0[i];
		mem_w[i] = ptr->mem_w[i];
//		mem_zero[i] = ptr->mem_zero[i];
	}

	sharp = ptr->sharp;

	pastVad = ptr->pastVad;
	ppastVad = ptr->ppastVad;
	seed = ptr->seed;
}*/

/*-----------------------------------------------------------------*
 *   Function  Init_Coder_ld8a                                     *
 *            ~~~~~~~~~~~~~~~                                      *
 *                                                                 *
 *  Init_Coder_ld8a(void);                                         *
 *                                                                 *
 *   ->Initialization of variables for the coder section.          *
 *       - initialize pointers to speech buffer                    *
 *       - initialize static  pointers                             *
 *       - set static vectors to zero                              *
 *                                                                 *
 *-----------------------------------------------------------------*/
void Init_Coder_ld8a(void)
{
  int i;
  p729Cod->CodStat.lsp_old[0] =  30000;
  p729Cod->CodStat.lsp_old[1] =  26000;
  p729Cod->CodStat.lsp_old[2] =  21000;
  p729Cod->CodStat.lsp_old[3] =  15000;
  p729Cod->CodStat.lsp_old[4] =   8000;
  p729Cod->CodStat.lsp_old[5] =      0;
  p729Cod->CodStat.lsp_old[6] =  -8000;
  p729Cod->CodStat.lsp_old[7] = -15000;
  p729Cod->CodStat.lsp_old[8] = -21000;
  p729Cod->CodStat.lsp_old[9] = -26000;

  p729Cod->LpcStat.old_A[0]   = 0x1000;/*Q.12 1.0*/
  for( i = 1 ; i < M+1 ; i++ )
    p729Cod->LpcStat.old_A[i] = 0;	
  p729Cod->LpcStat.old_rc[0] = 0;
  p729Cod->LpcStat.old_rc[1] = 0;
	
  for(i=0; i<4; i++)
    p729Cod->QGainStat.past_qua_en[i] = -14336;
	
  /*----------------------------------------------------------------------*
  *      Initialize pointers to speech vector.                            *
  *   |--------------------|-------------|-------------|------------|     *
  *     previous speech(120)      sf1(40)      sf2(40)      L_NEXT(40)    *
  *                                                                       *
  *   <----------------  Total speech vector(L_TOTAL 240)----------->     *
  *   <----------------  LPC analysis window (L_WINDOW)  ----------->     *
  *   |                   <-- present frame (L_FRAME) -->                 *
  * old_speech            |              <-- new speech (L_FRAME) -->     *
  * p_window              |              |                                *
  *                     speech           |                                *
  *                             new_speech                                *
  *-----------------------------------------------------------------------*/

  p729Cod->GloVar.new_speech = p729Cod->CodStat.old_speech + L_TOTAL - L_FRAME;         /* New speech     */
  p729Cod->CodStat.speech    = p729Cod->GloVar.new_speech  - L_NEXT;                    /* Present frame  */
  p729Cod->CodStat.p_window  = p729Cod->CodStat.old_speech + L_TOTAL - L_WINDOW;        /* For LPC window */

  /* Initialize static pointers */

  /*----------------------------------------------------------------------*
  *      Initialize pointers to wsp vector.                               *
  *   |-|---------------------------------------|--------------------|    *
  *   | |              (-143)                   |        (80)             *
  *  old_wsp                                    wsp                       *
  *     |                                                                 *
  *     wsp[-PIT_MAX]                                                     *
  *-----------------------------------------------------------------------*/
  p729Cod->CodStat.wsp    = p729Cod->CodStat.old_wsp + PIT_MAX + 1;	//+1 for alignment

  /*----------------------------------------------------------------------*
  *      Initialize pointers to exc vector.                               *
  *   |-------------------------------|------------|--------------------| *
  *   |         (PIT_MAX)             | (IINTERPOL)|     (L_FRAME=80)     *
  *   old_exc                                      exc                    *
  *-----------------------------------------------------------------------*/
  p729Cod->CodStat.exc    = p729Cod->CodStat.old_exc + PIT_MAX + L_INTERPOL;


  /* Static vectors to zero */
  Set_zero(p729Cod->CodStat.old_speech, L_TOTAL);
  Set_zero(p729Cod->CodStat.old_exc, PIT_MAX+L_INTERPOL);
  Set_zero(p729Cod->CodStat.old_wsp, PIT_MAX+1);
  Set_zero(p729Cod->CodStat.mem_w,   M);
  Set_zero(p729Cod->CodStat.mem_w0,  M);
  Set_zero(p729Cod->CodStat.mem_zero,M);
  p729Cod->CodStat.sharp = SHARPMIN;

  /* Initialize lsp_old_q[] */
  Copy(p729Cod->CodStat.lsp_old, p729Cod->CodStat.lsp_old_q, M);
  
  Lsp_encw_reset();
  Init_exc_err();

  /* For G.729B */
  /* Initialize VAD/DTX parameters */
  p729Cod->CodStat.pastVad = 1;
  p729Cod->CodStat.ppastVad = 1;
  p729Cod->CodStat.seed = INIT_SEED;/*11111*/
  vad_init();
  Init_lsfq_noise();
  return;
}

void g729_cod_ld8a_sub1s(Word16 *xn, Word16 *y1, Word16 gain_pit, Word16 *xn2);

void g729_cod_ld8a_sub1(
	Word16 *xn,		//align 8
	Word16 *y1,		//align 4
	Word16 gain_pit, 
	Word16 *xn2)	//align 8
{
	int i;
	Word32 L_temp;
    for (i = 0; i < L_SUBFR; i++)
    {
		L_temp = L_mult(y1[i], gain_pit);
		L_temp = L_shl(L_temp, 1);/* gain_pit in Q14 */
		xn2[i] = sub(xn[i], extract_h(L_temp));
    }
}

void g729_cod_ld8a_sub2s(Word16 *pexc, Word16 gain_pit, Word16 *code, Word16 gain_code);
void Copy_a4_154(Word16 x[], Word16 y[], Word16 L);

void g729_cod_ld8a_sub2(
	Word16 *pexc, 		//align 4
	Word16 gain_pit, 
	Word16 *code, 		//align 4
	Word16 gain_code)
{
	Word32 i, L_temp;
	
    for (i = 0; i < L_SUBFR;  i++)
    {
      /* exc[i] = gain_pit*exc[i] + gain_code*code[i]; */
      /* exc[i]  in Q0   gain_pit in Q14               */
      /* code[i] in Q13  gain_cod in Q1                */

      L_temp = L_mult(pexc[i], gain_pit);
      L_temp = L_mac(L_temp, code[i], gain_code);
      L_temp = L_shl(L_temp, 1);
      pexc[i] = round(L_temp);
    }
}
/*-----------------------------------------------------------------*
 *   Functions Coder_ld8a                                          *
 *            ~~~~~~~~~~                                           *
 *  Coder_ld8a(Word16 ana[]);                                      *
 *                                                                 *
 *   ->Main coder function.                                        *
 *                                                                 *
 *                                                                 *
 *  Input:                                                         *
 *                                                                 *
 *    80 speech data should have been copied to vector new_speech[].*
 *    This vector is global and is declared in this function.      *
 *                                                                 *
 *  Ouputs:                                                        *
 *                                                                 *
 *    ana[]      ->analysis parameters.                            *
 *                                                                 *
 *-----------------------------------------------------------------*/

/* DMEM */
/* To avoid stack overflow, we move the following cariables from 
    the local variables of Coder_ld8a() into global. */
	/* LPC analysis */
	Word16 Aq_t[(MP1)*2]; /* A(z)   quantized for the 2 subframes 11x2 */  
	Word16 Ap_t[(MP1)*2]; /* A(z/gamma)       for the 2 subframes 11x2 */  
	/* Other vectors */

	Word16 code[L_SUBFR];   /* Fixed codebook excitation          */
	Word16   y1[L_SUBFR];   /* Filtered adaptive excitation       */
	Word16   y2[L_SUBFR];   /* Filtered fixed codebook excitation */
	Word16     g_coeff[4];  /* Correlations between xn & y1       */
	Word16     g_coeff_cs[5];
	Word16 exp_g_coeff_cs[5];/* Correlations between xn, y1, & y2
                <y1,y1>, -2<xn,y1>, <y2,y2>, -2<xn,y2>, 2<y1,y2> */
	
		/* Temporary vectors */
		Word16 r_l[NP+1], r_h[NP+1];     /* Autocorrelations low and hi          */
		Word16 rc[M];                    /* Reflection coefficients.             */
		Word16 lsp_new[M], lsp_new_q[M]; /* LSPs at 2th subframe                 */

		/* For G.729B */
		Word16 rh_nbe[MP1];             
		Word16 lsf_new[M];
		Word16 lsfq_mem[MA_NP][M];
		Word16 exp_R0;
		Word16 Vad;

void Coder_ld8a(
     Word16 ana[],     /* output  : Analysis parameters[12] */
     Word16 frame,     /* input   : frame counter (VAD)     */
     Word16 vad_enable /* input   : VAD enable flag     */
)
{

	/* LPC analysis */

	Word16 *Aq, *Ap;      /* Pointer on Aq_t and Ap_t             */

	/* Scalars */

	Word16 i, j, k, i_subfr;
	Word16 T_op, T0, T0_min, T0_max, T0_frac;
	Word16 gain_pit, gain_code, index;
	Word16 temp, taming;

	/* pad zeros L_SUBFR:40*/
	 xn[L_SUBFR]= xn[L_SUBFR+1]= xn[L_SUBFR+2]= xn[L_SUBFR+3]=0;
	xn2[L_SUBFR]=xn2[L_SUBFR+1]=xn2[L_SUBFR+2]=xn2[L_SUBFR+3]=0;

	/*------------------------------------------------------------------------*
	 *  - Perform LPC analysis:                                               *
	 *       * autocorrelation + lag windowing                                *
	 *       * Levinson-durbin algorithm to find a[]                          *
	 *       * convert a[] to lsp[]                                           *
	 *       * quantize and code the LSPs                                     *
	 *       * find the interpolated LSPs and convert to a[] for the 2        *
	 *         subframes (both quantized and unquantized)                     *
	 *------------------------------------------------------------------------*/
	{

		/* LP analysis */
#ifdef VC6
		Autocorr(p729Cod->CodStat.p_window, /*NP,*/ r_h, r_l, &exp_R0);
#else
		Autocorrs(p729Cod->CodStat.p_window, /*NP,*/ r_h, r_l, &exp_R0);
#endif
		Copy(r_h, rh_nbe, MP1);		
		Lag_window(NP, r_h, r_l);            /* Lag windowing    */
#ifdef VC6
		temp = Levinson(r_h, r_l, Ap_t, rc); /* Levinson Durbin  */
#else
		temp = Levinsons(r_h, r_l, Ap_t, rc);/* Levinson Durbin  */
#endif
		/* From A(z) to lsp Section 3.2.3 */
		Az_lsp(Ap_t, lsp_new, p729Cod->CodStat.lsp_old);

#ifdef VC6
		/* For G.729B VAD function lsp = cos(lsf)*/
		Lsp_lsf(lsp_new, lsf_new, M);
#else
		Lsp_lsfs(lsp_new, lsf_new, M);
#endif
		/*-----------------------------------------------------------------*
		 * Functions vad                                                   *
		 *           ~~~                                                   *
		 * Input:                                                          *
		 *   rc            : reflection coefficient                        *
		 *   lsf[]         : unquantized lsf vector                        *
		 *   r_h[]         : upper 16-bits of the autocorrelation vector   *
		 *   r_l[]         : lower 16-bits of the autocorrelation vector   *
		 *   exp_R0        : exponent of the autocorrelation vector        *
		 *   sigpp[]       : preprocessed input signal                     *
		 *   frm_count     : frame counter                                 *
		 *   prev_marker   : VAD decision of the last frame                *
		 *   pprev_marker  : VAD decision of the frame before last frame   *
		 *-----------------------------------------------------------------*/
		vad( rc[1], lsf_new, r_h, r_l, exp_R0, p729Cod->CodStat.p_window, frame,
                                               p729Cod->CodStat.pastVad, 
                                               p729Cod->CodStat.ppastVad, &Vad);

		Update_cng(rh_nbe, exp_R0, Vad);
    
		/* ---------------------- */
		/* Case of Inactive frame */
		/* ---------------------- */
		if ((Vad == 0) && (vad_enable == 1))
		{
			Get_freq_prev(lsfq_mem);
			Cod_cng(p729Cod->CodStat.exc, 
				    p729Cod->CodStat.pastVad, 
				    p729Cod->CodStat.lsp_old_q, 
				    Aq_t, ana, lsfq_mem, &p729Cod->CodStat.seed);
			Update_freq_prev(lsfq_mem);
			p729Cod->CodStat.ppastVad = p729Cod->CodStat.pastVad;
			p729Cod->CodStat.pastVad = Vad;

			/* Update wsp, mem_w and mem_w0 */
			Aq = Aq_t;
			for( i_subfr = 0 ; i_subfr < L_FRAME ; i_subfr += L_SUBFR) 
			{
				/* Residual signal in xn */
#ifdef VC6
				Residu(Aq, &p729Cod->CodStat.speech[i_subfr], xn, L_SUBFR);
#else
				Residus(Aq, &p729Cod->CodStat.speech[i_subfr], xn, L_SUBFR);
#endif

#ifdef VC6
				Weight_Az(Aq, GAMMA1, M, Ap_t);
#else
				Weight_Azs(Aq, GAMMA1, M, Ap_t);
#endif
        
				/* Compute wsp and mem_w */
				/*Ap = Ap_t + MP1;*/
				Ap = Ap_t + M;
				/*Ap[0] = 4096;*/
				/*for(i=1; i<=M; i++)*/    /* Ap[i] = Ap_t[i] - 0.7 * Ap_t[i-1]; */
				/*Ap[i] = sub(Ap_t[i], mult(Ap_t[i-1], 22938));*/
				Ap[0] = sub(Ap_t[0], mult(0x1000, 22938));
				for(i=1; i<M; i++)
					Ap[i] = sub(Ap_t[i], mult(Ap_t[i-1], 22938));
				/*Do the synthesis filtering 1/A(z)*/
#ifdef VC6
				Syn_filt(Ap, xn, &p729Cod->CodStat.wsp[i_subfr], L_SUBFR, p729Cod->CodStat.mem_w, 1);
#else
				Syn_filts(Ap, xn, &p729Cod->CodStat.wsp[i_subfr], L_SUBFR, p729Cod->CodStat.mem_w, 1);
#endif

				/* Compute mem_w0 */
				for(i=0; i<L_SUBFR; i++) {
					xn[i] = sub(xn[i], p729Cod->CodStat.exc[i_subfr+i]);  /* residu[] - exc[] */
				}
#ifdef VC6
				Syn_filt(Ap_t, xn, xn, L_SUBFR, p729Cod->CodStat.mem_w0, 1);
#else
				Syn_filts(Ap_t, xn, xn, L_SUBFR, p729Cod->CodStat.mem_w0, 1);
#endif
				//Aq += MP1;
				Aq += M;
			}/*case of subframe END*/
			p729Cod->CodStat.sharp = SHARPMIN;
			/* Update memories for next frames */
			Copy_a8_m8(&p729Cod->CodStat.old_speech[L_FRAME], &p729Cod->CodStat.old_speech[0], L_TOTAL-L_FRAME);
			Copy(&p729Cod->CodStat.old_wsp[L_FRAME+1], &p729Cod->CodStat.old_wsp[1], PIT_MAX);
			Copy_a4_154(&p729Cod->CodStat.old_exc[L_FRAME], &p729Cod->CodStat.old_exc[0], PIT_MAX+L_INTERPOL);
			return;
		}  /* End of inactive frame case */

	    /* ----------------------------------------------------- */
	    /* Case of Active frame */
	    /* ----------------------------------------------------- */
		*ana++ = 1;/* *ana = 1 then ana++ */
		p729Cod->CodStat.seed = INIT_SEED;/*INIT_SEED=11111*/
		p729Cod->CodStat.ppastVad = p729Cod->CodStat.pastVad;
		p729Cod->CodStat.pastVad = Vad;

		/* -----------------------------------------------------
		 * LSP quantization Section 3.2.4
		 * output L0 L1 L2 L3 in ana`[0] ana`[1]
		 * -----------------------------------------------------*/
		Qua_lsp(lsp_new, lsp_new_q, ana);
		ana += 2;

		/*---------------------------------------------------------------*
	     * Find interpolated LPC parameters in all subframes             *
	     * The interpolated parameters are in array Aq_t[].              *
	     *---------------------------------------------------------------*/
#ifdef VC6
		Int_qlpc( p729Cod->CodStat.lsp_old_q, lsp_new_q, Aq_t);
#else
		Int_qlpcs(p729Cod->CodStat.lsp_old_q, lsp_new_q, Aq_t);
#endif

		/* Compute A(z/gamma)
         * the perceptual weighting filter is based on the quantized LP(Aq_t)
         */
#ifdef VC6
		Weight_Az(&Aq_t[0], GAMMA1, M, &Ap_t[0]);
		Weight_Az(&Aq_t[M], GAMMA1, M, &Ap_t[M]);
#else
		Weight_Azs(&Aq_t[0], GAMMA1, M, &Ap_t[0]);
		Weight_Azs(&Aq_t[M], GAMMA1, M, &Ap_t[M]);
#endif
		/* update the LSPs for the next frame */
		Copy(lsp_new,   p729Cod->CodStat.lsp_old,   M);
		Copy(lsp_new_q, p729Cod->CodStat.lsp_old_q, M);
	}
	
	/*----------------------------------------------------------------------*
	 * - Find the weighted input speech w_sp[] for the whole speech frame   *
	 * - Find the open-loop pitch delay                                     *
	 * - AnnexA is completely different than G.729 orginal                  *
	 *----------------------------------------------------------------------*/
	{
#ifdef VC6
	Residu(&Aq_t[0], &p729Cod->CodStat.speech[0],       &p729Cod->CodStat.exc[0],       L_SUBFR);
	Residu(&Aq_t[M], &p729Cod->CodStat.speech[L_SUBFR], &p729Cod->CodStat.exc[L_SUBFR], L_SUBFR);
#else
	Residus(&Aq_t[0], &p729Cod->CodStat.speech[0],       &p729Cod->CodStat.exc[0],       L_SUBFR);
	Residus(&Aq_t[M], &p729Cod->CodStat.speech[L_SUBFR], &p729Cod->CodStat.exc[L_SUBFR], L_SUBFR);
#endif

	Word16 Ap1[M]; /* Used for modified A(z) with low-pass */
	/*
	 *  Find weighted speech for subframe 1 ( a low-pass filter is embedded in it )
	 */
	Ap = Ap_t;
	Ap1[0] = sub(Ap[0], mult(0x1000, 22938));
	for(i=1; i<M; i++)
		Ap1[i] = sub(Ap[i], mult(Ap[i-1], 22938));
#ifdef VC6
	Syn_filt(Ap1, &p729Cod->CodStat.exc[0], &p729Cod->CodStat.wsp[0], L_SUBFR, p729Cod->CodStat.mem_w, 1);
#else
	Syn_filts(Ap1, &p729Cod->CodStat.exc[0], &p729Cod->CodStat.wsp[0], L_SUBFR, p729Cod->CodStat.mem_w, 1);
#endif
	/*
	 *  Find weighted speech for subframe 2
	 */
	Ap += M;
	Ap1[0] = sub(Ap[0], mult(0x1000, 22938));
	for(i=1; i<M; i++)
		Ap1[i] = sub(Ap[i], mult(Ap[i-1], 22938));
#ifdef VC6
	Syn_filt(Ap1, &p729Cod->CodStat.exc[L_SUBFR], &p729Cod->CodStat.wsp[L_SUBFR], L_SUBFR, p729Cod->CodStat.mem_w, 1);
#else
	Syn_filts(Ap1, &p729Cod->CodStat.exc[L_SUBFR], &p729Cod->CodStat.wsp[L_SUBFR], L_SUBFR, p729Cod->CodStat.mem_w, 1);
#endif
	}
	
    /* Find open loop pitch lag(delay) */
#ifdef VC6  
	T_op = Pitch_ol_fast( p729Cod->CodStat.wsp);
#else
	T_op = Pitch_ol_fasts(p729Cod->CodStat.wsp);
#endif

	/* Range for closed loop pitch search in 1st subframe */
	T0_min = sub(T_op, 3);
	if (sub(T0_min,PIT_MIN)<0) {
		T0_min = PIT_MIN;
	}
	T0_max = T0_min + 6;
	if (sub(T0_max ,PIT_MAX)>0)
	{
		T0_max = PIT_MAX;
		T0_min = sub(T0_max, 6);
	}
	 /*------------------------------------------------------------------------*
	  *          Loop for every subframe in the analysis frame                 *
	  *------------------------------------------------------------------------*
	  *  To find the pitch and innovation parameters. The subframe size is     *
	  *  L_SUBFR and the loop is repeated 2 times.                             *
	  *     - find the weighted LPC coefficients                               *
	  *     - find the LPC residual signal res[]                               *
	  *     - compute the target signal for pitch search                       *
	  *     - compute impulse response of weighted synthesis filter (h1[])     *
	  *     - find the closed-loop pitch parameters                            *
	  *     - encode the pitch delay                                           *
	  *     - find target vector for codebook search                           *
	  *     - codebook search                                                  *
	  *     - VQ of pitch and codebook gains                                   *
	  *     - update states of weighting filter                                *
	  *------------------------------------------------------------------------*/

	Aq = Aq_t;    /* pointer to interpolated quantized LPC parameters     */
	Ap = Ap_t;    /* pointer to weighted LPC coefficients                 */
	              /* in AnnexA Ap is not only weighted but also quantized */
	for (i_subfr = 0;  i_subfr < L_FRAME; i_subfr += L_SUBFR)
	{
	   /*---------------------------------------------------------------*
        * Compute impulse response, h1[], of weighted synthesis filter  *
        *---------------------------------------------------------------*/
		h1[0] = 4096;/*An impulse in Q12*/
		Set_zero(&h1[1], L_SUBFR-1);
#ifdef VC6
		Syn_filt( Ap, h1, h1, L_SUBFR, &h1[2]/*dram1*/, 0);	/*h1[2:12]=0*/
#else
		Syn_filts(Ap, h1, h1, L_SUBFR, &h1[2]/*dram1*/, 0);	/*h1[2:12]=0*/
#endif
	   /*----------------------------------------------------------------------*
	    *  Find the target vector for pitch search:                            *
	    *----------------------------------------------------------------------*/
#ifdef VC6
		Syn_filt( Ap, &p729Cod->CodStat.exc[i_subfr], xn, L_SUBFR, p729Cod->CodStat.mem_w0, 0);
#else
		Syn_filts(Ap, &p729Cod->CodStat.exc[i_subfr], xn, L_SUBFR, p729Cod->CodStat.mem_w0, 0);
#endif
	    /*---------------------------------------------------------------------*
	     *                 Closed-loop fractional pitch search                 *
	     *---------------------------------------------------------------------*/
		T0 = Pitch_fr3_fast(&p729Cod->CodStat.exc[i_subfr], xn, h1, L_SUBFR,
		   T0_min, T0_max, i_subfr, &T0_frac);
		index = Enc_lag3(T0, T0_frac, &T0_min, &T0_max,PIT_MIN,PIT_MAX,i_subfr);
#if 0 /* brian_bin add for value check */
		if(p729DebugInfo->SpeechFrameIndex==38){
			printk("Open loop Top %d \n",T_op);
			printk("Pitch delay %d \n",T0);
			printk("P1 index %d \n",index);
			while(1){};
		}
#endif		
		*ana++ = index;
		if (i_subfr == 0) {
			*ana++ = Parity_Pitch(index);
		}
	   /*-----------------------------------------------------------------*
	    *   - find filtered pitch exc                                     *
	    *   - compute pitch gain and limit between 0 and 1.2              *
	    *   - update target vector for codebook search                    *
	    *-----------------------------------------------------------------*/
#ifdef VC6
		Syn_filt(Ap, &p729Cod->CodStat.exc[i_subfr], y1, L_SUBFR, p729Cod->CodStat.mem_zero, 0);
#else
		Syn_filts(Ap, &p729Cod->CodStat.exc[i_subfr], y1, L_SUBFR, p729Cod->CodStat.mem_zero, 0);
#endif
#ifdef VC6
		gain_pit = G_pitch(xn, y1, g_coeff, L_SUBFR);
#else
		gain_pit = G_pitchs(xn, y1, g_coeff, L_SUBFR);
#endif

		/* clip pitch gain if taming is necessary */
		taming = test_err(T0, T0_frac);
		if( taming == 1){
			if (sub(gain_pit, GPCLIP) > 0) {
				gain_pit = GPCLIP;
			}
		}
		/* xn2[i]   = xn[i] - y1[i] * gain_pit  */
#ifdef VC6
		g729_cod_ld8a_sub1(xn, y1, gain_pit, xn2);
#else
		g729_cod_ld8a_sub1s(xn, y1, gain_pit, xn2);
#endif

	   /*-----------------------------------------------------*
	    * - Innovative codebook search.                       *
	    *-----------------------------------------------------*/
		index = ACELP_Code_A(xn2, h1, T0, p729Cod->CodStat.sharp, code, y2, &i);
		*ana++ = index;        /* Positions index */
		*ana++ = i;            /* Signs index     */
	   /*-----------------------------------------------------*
	    * - Quantization of gains.                            *
	    *-----------------------------------------------------*/
		g_coeff_cs[0]     = g_coeff[0];            /* <y1,y1> */
		exp_g_coeff_cs[0] = negate(g_coeff[1]);    /* Q-Format:XXX -> JPN */
		g_coeff_cs[1]     = negate(g_coeff[2]);    /* (xn,y1) -> -2<xn,y1> */
		exp_g_coeff_cs[1] = negate(add(g_coeff[3], 1)); /* Q-Format:XXX -> JPN */
#ifdef VC6
		Corr_xy2( xn, y1, y2, g_coeff_cs, exp_g_coeff_cs );  /* Q0 Q0 Q12 ^Qx ^Q0 */
        /* g_coeff_cs[3]:exp_g_coeff_cs[3] = <y2,y2>   */
        /* g_coeff_cs[4]:exp_g_coeff_cs[4] = -2<xn,y2> */
        /* g_coeff_cs[5]:exp_g_coeff_cs[5] = 2<y1,y2>  */
#else
		Corr_xy2s( xn, y1, y2, g_coeff_cs, exp_g_coeff_cs );  /* Q0 Q0 Q12 ^Qx ^Q0 */
#endif

#ifdef VC6
		*ana++ = Qua_gain(code, g_coeff_cs, exp_g_coeff_cs, &gain_pit, &gain_code, taming);
#else
		*ana++ = Qua_gains(code, g_coeff_cs, exp_g_coeff_cs, &gain_pit, &gain_code, taming);
#endif
		/*------------------------------------------------------------*
		 * - Update pitch sharpening "sharp" with quantized gain_pit  *
		 *------------------------------------------------------------*/
		p729Cod->CodStat.sharp = gain_pit;
		if (sub(p729Cod->CodStat.sharp,  SHARPMAX) > 0) 
			{   p729Cod->CodStat.sharp = SHARPMAX; }
		if (sub(p729Cod->CodStat.sharp,  SHARPMIN) < 0) 
			{   p729Cod->CodStat.sharp = SHARPMIN; }


		/*------------------------------------------------------*
		 * - Find the total excitation                          *
		 * - update filters memories for finding the target     *
		 *   vector in the next subframe                        *
		 *------------------------------------------------------*/
#ifdef VC6    
		g729_cod_ld8a_sub2(&p729Cod->CodStat.exc[i_subfr], gain_pit, code, gain_code);
#else
		g729_cod_ld8a_sub2s(&p729Cod->CodStat.exc[i_subfr], gain_pit, code, gain_code);
#endif

		update_exc_err(gain_pit, T0);

		for (i = L_SUBFR-M, j = 0; i < L_SUBFR; i++, j++)
		{
			temp       = extract_h(L_shl( L_mult(y1[i], gain_pit),  1) );
			k          = extract_h(L_shl( L_mult(y2[i], gain_code), 2) );
			p729Cod->CodStat.mem_w0[j]  = sub(xn[i], add(temp, k));
		}
 //		Aq += MP1;           /* interpolated LPC parameters for next subframe */
 //		Ap += MP1;
		Aq += M;
		Ap += M;
	}
	 /*--------------------------------------------------*
	  * Update signal for next frame.                    *
	  * -> shift to the left by L_FRAME:                 *
	  *     speech[], wsp[] and  exc[]                   *
	  *--------------------------------------------------*/
	Copy_a8_m8(&p729Cod->CodStat.old_speech[L_FRAME], &p729Cod->CodStat.old_speech[0], L_TOTAL-L_FRAME);  
	Copy(&p729Cod->CodStat.old_wsp[L_FRAME+1], &p729Cod->CodStat.old_wsp[1], PIT_MAX);
	Copy_a4_154(&p729Cod->CodStat.old_exc[L_FRAME], &p729Cod->CodStat.old_exc[0], PIT_MAX+L_INTERPOL);
	return;
}

