/*
   ITU-T G.729A Speech Coder with Annex B    ANSI-C Source Code
   Version 1.3    Last modified: August 1997

   Copyright (c) 1996,
   AT&T, France Telecom, NTT, Universite de Sherbrooke, Lucent Technologies,
   Rockwell International
   All rights reserved.
*/

/*-----------------------------------------------------------------*
 *   Functions Init_Decod_ld8a  and Decod_ld8a                     *
 *-----------------------------------------------------------------*/

#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "g729_ld8a.h"

#include "g729_dtx.h"
#include "g729_sid.h"
#include "mystruct.h"

/*---------------------------------------------------------------*
 *   Decoder constant parameters (defined in "g729_ld8a.h")           *
 *---------------------------------------------------------------*
 *   L_FRAME     : Frame size.                                   *
 *   L_SUBFR     : Sub-frame size.                               *
 *   M           : LPC order.                                    *
 *   MP1         : LPC order+1                                   *
 *   PIT_MIN     : Minimum pitch lag.                            *
 *   PIT_MAX     : Maximum pitch lag.                            *
 *   L_INTERPOL  : Length of filter for interpolation            *
 *   PRM_SIZE    : Size of vector containing analysis parameters *
 *---------------------------------------------------------------*/

/*--------------------------------------------------------*
 *         Static memory allocation.                      *
 *--------------------------------------------------------*/

        /* Excitation vector */
// static Word16 old_exc[L_FRAME+PIT_MAX+L_INTERPOL];
// static Word16 *exc;

        /* Lsp (Line spectral pairs) */
// static Word16 lsp_old[M]={
//             30000, 26000, 21000, 15000, 8000, 0, -8000,-15000,-21000,-26000};

        /* Filter's memory */
 //static Word16 mem_syn[M];
// static Word16 sharp;           /* pitch sharpening of previous frame */
 //static Word16 old_T0;          /* integer delay of previous frame    */
 //static Word16 gain_code;       /* Code gain                          */
// static Word16 gain_pitch;      /* Pitch gain                         */
 
/* for G.729B */
//static Word16 seed_fer;
/* CNG variables */
//static Word16 past_ftyp;
//static Word16 seed;
//static Word16 sid_sav, sh_sid_sav;
/*
void StoreDecld8aStaticVar(Decld8aStaticVar *ptr)
{
	int i;
	for (i=0; i<L_FRAME+PIT_MAX+L_INTERPOL; i++)
		ptr->old_exc[i] = old_exc[i];
	ptr->exc = exc;
	for (i=0; i<M; i++)
	{
		ptr->lsp_old[i] = lsp_old[i];
		ptr->mem_syn[i] = mem_syn[i];
	}
	ptr->sharp = sharp;
	ptr->old_T0 = old_T0;
	ptr->gain_code = gain_code;
	ptr->gain_pitch = gain_pitch;
	ptr->seed_fer = seed_fer;
	ptr->past_ftyp = past_ftyp;
	ptr->seed = seed;
	ptr->sid_sav = sid_sav;
	ptr->sh_sid_sav = sh_sid_sav;		
}

void LoadDecld8aStaticVar(Decld8aStaticVar *ptr)
{
	int i;
	for (i=0; i<L_FRAME+PIT_MAX+L_INTERPOL; i++)
		old_exc[i] = ptr->old_exc[i];
	exc = ptr->exc;
	for (i=0; i<M; i++)
	{
		lsp_old[i] = ptr->lsp_old[i];
		mem_syn[i] = ptr->mem_syn[i];
	}
	sharp = ptr->sharp;
	old_T0 = ptr->old_T0;
	gain_code = ptr->gain_code;
	gain_pitch = ptr->gain_pitch;
	seed_fer = ptr->seed_fer;
	past_ftyp = ptr->past_ftyp;
	seed = ptr->seed;
	sid_sav = ptr->sid_sav;
	sh_sid_sav = ptr->sh_sid_sav;		
}
*/

/*-----------------------------------------------------------------*
 *   Function Init_Decod_ld8a                                      *
 *            ~~~~~~~~~~~~~~~                                      *
 *                                                                 *
 *   ->Initialization of variables for the decoder section.        *
 *                                                                 *
 *-----------------------------------------------------------------*/

void Init_Decod_ld8a(void)
{
	int	i;

	p729Dec->DecStat.lsp_old[0] = 30000;
	p729Dec->DecStat.lsp_old[1] = 26000;
	p729Dec->DecStat.lsp_old[2] = 21000;
	p729Dec->DecStat.lsp_old[3] = 15000;
	p729Dec->DecStat.lsp_old[4] = 8000;
	p729Dec->DecStat.lsp_old[5] = 0;
	p729Dec->DecStat.lsp_old[6] = -8000;
	p729Dec->DecStat.lsp_old[7] = -15000;
	p729Dec->DecStat.lsp_old[8] = -21000;
	p729Dec->DecStat.lsp_old[9] = -26000;

	for(i=0;i<4;i++)
		p729Dec->DGnStat.past_qua_en[i] = -14336;

  /* Initialize static pointer */

  p729Dec->DecStat.exc = p729Dec->DecStat.old_exc + PIT_MAX + L_INTERPOL;

  /* Static vectors to zero */

  Set_zero(p729Dec->DecStat.old_exc, PIT_MAX+L_INTERPOL);
  Set_zero(p729Dec->DecStat.mem_syn, M);

  p729Dec->DecStat.sharp  = SHARPMIN;
  p729Dec->DecStat.old_T0 = 60;
  p729Dec->DecStat.gain_code = 0;
  p729Dec->DecStat.gain_pitch = 0;

  Lsp_decw_reset();

  /* for G.729B */
  p729Dec->DecStat.seed_fer = 21845;
  p729Dec->DecStat.past_ftyp = 1;
  p729Dec->DecStat.seed = INIT_SEED;
  p729Dec->DecStat.sid_sav = 0;
  p729Dec->DecStat.sh_sid_sav = 1;
  Init_lsfq_noise();

  return;
}

void Decod_ld8a_sub1(Word16 code[], Word16 i_subfr)
{
	Word16 i;
	Word32 L_temp;
		for (i = 0; i < L_SUBFR;  i++)
          {
            /* exc[i] = gain_pitch*exc[i] + gain_code*code[i]; */
            /* exc[i]  in Q0   gain_pitch in Q14               */
            /* code[i] in Q13  gain_codeode in Q1              */
            
            L_temp = L_mult(p729Dec->DecStat.exc[i+i_subfr], p729Dec->DecStat.gain_pitch);           
            L_temp = L_mac(L_temp, code[i], p729Dec->DecStat.gain_code);           
			L_temp = L_add(L_temp, L_temp);
            p729Dec->DecStat.exc[i+i_subfr] = round(L_temp);
          }
	return;
}

void Decod_ld8a_sub2(void)
{
	Word32 L_temp;
	Word16 i;
    L_temp = 0L;
    for(i=0; i<L_FRAME; i++) {
      L_temp = L_mac(L_temp, p729Dec->DecStat.exc[i], p729Dec->DecStat.exc[i]);
    } /* may overflow => last level of SID quantizer */
    p729Dec->DecStat.sh_sid_sav = norm_l(L_temp);
    p729Dec->DecStat.sid_sav = round(L_shl(L_temp, p729Dec->DecStat.sh_sid_sav));
    //p729Dec->DecStat.sh_sid_sav = sub(16, p729Dec->DecStat.sh_sid_sav);
	p729Dec->DecStat.sh_sid_sav = 16 - p729Dec->DecStat.sh_sid_sav;
	return;
}

/*-----------------------------------------------------------------*
 *   Function Decod_ld8a                                           *
 *           ~~~~~~~~~~                                            *
 *   ->Main decoder routine. Input is a frame                      *
 *                                                                 *
 *-----------------------------------------------------------------*/

void Decod_ld8a(
  Word16  parm[],  /* (i): vector of synthesis parameters
                           parm[0] = bad frame indicator (bfi)  
                           parm[1] = ftyp indicator             */
  Word16  synth[], /* (o): synthesis speech                     */
  Word16  A_t[],   /* (o): decoded LP filter in 2 subframes     */
  Word16  *T2,     /* (o): decoded pitch lag in 2 subframes     */
  Word16  *Vad     /* (o): frame type                           */
)
{
	Word16  *Az;                  /* Pointer on A_t   */
	static Word16  lsp_new[M];           /* LSPs             */
	static Word16  code[L_SUBFR];        /* ACELP codevector (40)*/
	/* Scalars */
	Word16  i;
	Word16  j;
	Word16  i_subfr;
	Word16  T0;
	Word16  T0_frac;
	Word16  index;
	Word16  bfi;                  /* bad frame indicator */
	Word16  bad_pitch;            /* bad pitch indicator */
	/* for G.729B */
	Word16  ftyp;				  /* frame type */
	static Word16  lsfq_mem[MA_NP][M];

	/* Test bad frame indicator (bfi) */
	/* bfi      = parm[0]
	   parm[0] <= parm[1] 
	   ftpe     = parm[0] = parm[1](original);
	 */
	bfi  = *parm++;
	/* for G.729B */
	ftyp = *parm;

	/* if this is a bad frame make speech according to prev_frame 
	   this might be caused by network quality so bfi flag is 1
	 */
	if(bfi == 1) {
		if(p729Dec->DecStat.past_ftyp == 1) ftyp = 1;
		else ftyp = 0;
		*parm = ftyp;  /* modification introduced in version V1.3 */
	}

	*Vad = ftyp;

	/* Processing non active frames ftype != 1 (SID frame) */
	if(ftyp != 1) {
		Get_decfreq_prev(lsfq_mem);/* Get p729Dec->DLspStat.freq_prev[][] */
		Dec_cng(p729Dec->DecStat.past_ftyp, 
			    p729Dec->DecStat.sid_sav, 
			    p729Dec->DecStat.sh_sid_sav, parm, 
			    p729Dec->DecStat.exc, 
			    p729Dec->DecStat.lsp_old, A_t, 
			   &p729Dec->DecStat.seed, lsfq_mem);
		Update_decfreq_prev(lsfq_mem);
		Az = A_t;
		for (i_subfr = 0; i_subfr < L_FRAME; i_subfr += L_SUBFR) {
			Overflow = 0;
#ifdef VC6
			Syn_filt(Az, &p729Dec->DecStat.exc[i_subfr], &synth[i_subfr], L_SUBFR, p729Dec->DecStat.mem_syn, 0);
#else
			Syn_filts(Az, &p729Dec->DecStat.exc[i_subfr], &synth[i_subfr], L_SUBFR, p729Dec->DecStat.mem_syn, 0);
#endif
	  
			if(Overflow != 0) {
				// In case of overflow in the synthesis          
				// -> Scale down vector exc[] and redo synthesis 
				for(i=0; i<PIT_MAX+L_INTERPOL+L_FRAME; i++)
				p729Dec->DecStat.old_exc[i] = shr(p729Dec->DecStat.old_exc[i], 2);
#ifdef VC6        
				Syn_filt(Az, &p729Dec->DecStat.exc[i_subfr], &synth[i_subfr], L_SUBFR, p729Dec->DecStat.mem_syn, 1);
#else
				Syn_filts(Az, &p729Dec->DecStat.exc[i_subfr], &synth[i_subfr], L_SUBFR, p729Dec->DecStat.mem_syn, 1);
#endif
			}
			else {
				Copy(&synth[i_subfr+L_SUBFR-M], p729Dec->DecStat.mem_syn, M);
			}
      		//Az += MP1;
			Az += M;
			*T2++ = p729Dec->DecStat.old_T0;
		}
		p729Dec->DecStat.sharp = SHARPMIN;
	}
	else { 
		/* Processing active frame */
		p729Dec->DecStat.seed = INIT_SEED;//11111
		parm++;
		/* Decode the LSPs */
		D_lsp(parm, lsp_new, add(bfi, p729Dec->GloVar.bad_lsf));
    	parm += 2;
		/* -----------------------------------------------------
		 * Note: "bad_lsf" is introduced in case the standard is used with
		 * channel protection.
		 * ----------------------------------------------------- */
		/* Interpolation of LPC for the 2 subframes */
#ifdef VC6    
		Int_qlpc(p729Dec->DecStat.lsp_old, lsp_new, A_t);
#else
		Int_qlpcs(p729Dec->DecStat.lsp_old, lsp_new, A_t);
#endif    
		/* update the LSFs for the next frame */
		Copy(lsp_new, p729Dec->DecStat.lsp_old, M);
    
 /*------------------------------------------------------------------------*
  *          Loop for every subframe in the analysis frame                 *
  *------------------------------------------------------------------------*
  * The subframe size is L_SUBFR and the loop is repeated L_FRAME/L_SUBFR  *
  *  times                                                                 *
  *     - decode the pitch delay                                           *
  *     - decode algebraic code                                            *
  *     - decode pitch and codebook gains                                  *
  *     - find the excitation and compute synthesis speech                 *
  *------------------------------------------------------------------------*/
    
		Az = A_t;            /* pointer to interpolated LPC parameters */
    
		for (i_subfr = 0; i_subfr < L_FRAME; i_subfr += L_SUBFR)
		{
			index = *parm++;        /* pitch index */
			if(i_subfr == 0)
			{
				i = *parm++;        /* get parity check result */
				//bad_pitch = add(bfi, i);
				bad_pitch = bfi + i;
				if( bad_pitch == 0)/*if no bad pitch*/ {
                	Dec_lag3(index, PIT_MIN, PIT_MAX, i_subfr, &T0, &T0_frac);
                	p729Dec->DecStat.old_T0 = T0;
              	}
            	else {     /* Bad frame, or parity error */
                	T0  =  p729Dec->DecStat.old_T0;
               		T0_frac = 0;                	//p729Dec->DecStat.old_T0 = add( p729Dec->DecStat.old_T0, 1);
					p729Dec->DecStat.old_T0 += 1;           	//if( sub(p729Dec->DecStat.old_T0, PIT_MAX) > 0) {
					if(p729Dec->DecStat.old_T0 > PIT_MAX) {
                  		p729Dec->DecStat.old_T0 = PIT_MAX;
                	}
              	}
          	}
        	else {    /* second subframe */
            	if( bfi == 0)/*if no bad pitch*/
              	{
                	Dec_lag3(index, PIT_MIN, PIT_MAX, i_subfr, &T0, &T0_frac);
                	p729Dec->DecStat.old_T0 = T0;
              	}
            	else {
                	T0  =  p729Dec->DecStat.old_T0;
                	T0_frac = 0;
                	//p729Dec->DecStat.old_T0 = add( p729Dec->DecStat.old_T0, 1);
					p729Dec->DecStat.old_T0 += 1;
                	//if( sub(p729Dec->DecStat.old_T0, PIT_MAX) > 0) {
					if(p729Dec->DecStat.old_T0 > PIT_MAX) {
                  		p729Dec->DecStat.old_T0 = PIT_MAX;
                	}
              	}
          	}
        	*T2++ = T0;
        /*-------------------------------------------------*
         * - Find the adaptive codebook vector.            *
         *-------------------------------------------------*/
#ifdef VC6
        	Pred_lt_3(&p729Dec->DecStat.exc[i_subfr], T0, T0_frac, L_SUBFR);
#else
			Pred_lt_3s(&p729Dec->DecStat.exc[i_subfr], T0, T0_frac, L_SUBFR);
#endif       

        /*-------------------------------------------------------*
         * - Decode innovative codebook.                         *
         * - Add the fixed-gain pitch contribution to code[].    *
         *-------------------------------------------------------*/

        	if(bfi != 0)            /* Bad frame */
          	{
#ifdef VC6
            	parm[0] = Random(&p729Dec->DecStat.seed_fer) & (Word16)0x1fff; /* 13 bits random */
            	parm[1] = Random(&p729Dec->DecStat.seed_fer) & (Word16)0x000f; /*  4 bits random */
#else
				parm[0] = Randoms(&p729Dec->DecStat.seed_fer) & (Word16)0x1fff; /* 13 bits random */
            	parm[1] = Randoms(&p729Dec->DecStat.seed_fer) & (Word16)0x000f; /*  4 bits random */
#endif
          	}

#ifdef VC6
        	Decod_ACELP(parm[1]/*sign*/, parm[0]/*index*/, code/*out*/);
#else
			Decod_ACELPs(parm[1], parm[0], code);
#endif
        	parm +=2;

//        	j = shl(sharp, 1);      /* From Q14 to Q15 */
			j = (p729Dec->DecStat.sharp<<1);
        
			if(T0 < L_SUBFR) {
          		for (i = T0; i < L_SUBFR; i++) {
            		code[i] = add(code[i], mult(code[i-T0], j));
          		}
        	}

        /*-------------------------------------------------*
         * - Decode pitch and codebook gains.              *
         *-------------------------------------------------*/
        	index = *parm++;        /* index of energy VQ */
        	Dec_gain(index, code, L_SUBFR, bfi, &p729Dec->DecStat.gain_pitch, &p729Dec->DecStat.gain_code);

        /*-------------------------------------------------------------*
         * - Update pitch sharpening "sharp" with quantized gain_pitch *
         *-------------------------------------------------------------*/
        	p729Dec->DecStat.sharp = p729Dec->DecStat.gain_pitch;
        	if (p729Dec->DecStat.sharp > SHARPMAX) { p729Dec->DecStat.sharp = SHARPMAX;  }
        	if (p729Dec->DecStat.sharp < SHARPMIN) { p729Dec->DecStat.sharp = SHARPMIN;  }

        /*-------------------------------------------------------*
         * - Find the total excitation.                          *
         * - Find synthesis speech corresponding to exc[].       *
         *-------------------------------------------------------*/
#ifdef VC6
			Decod_ld8a_sub1(code, i_subfr);
#else
			Decod_ld8a_sub1s(code, i_subfr);
#endif
#if 0
        	for (i = 0; i < L_SUBFR;  i++)
          	{
            	/* exc[i] = gain_pitch*exc[i] + gain_code*code[i]; */
            	/* exc[i]  in Q0   gain_pitch in Q14               */
            	/* code[i] in Q13  gain_codeode in Q1              */
            
            	L_temp = L_mult(p729Dec->DecStat.exc[i+i_subfr], p729Dec->DecStat.gain_pitch);
            	L_temp = L_mac(L_temp, code[i], p729Dec->DecStat.gain_code);
            	//L_temp = L_shl(L_temp, 1);
				L_temp = L_add(L_temp, L_temp);
            	p729Dec->DecStat.exc[i+i_subfr] = round(L_temp);
          	}
#endif        
        	Overflow = 0;
#ifdef VC6
        	Syn_filt(Az, &p729Dec->DecStat.exc[i_subfr], &synth[i_subfr], L_SUBFR, p729Dec->DecStat.mem_syn, 0);
#else
			Syn_filts(Az, &p729Dec->DecStat.exc[i_subfr], &synth[i_subfr], L_SUBFR, p729Dec->DecStat.mem_syn, 0);
#endif
        	if(Overflow != 0)
          	{
            	// In case of overflow in the synthesis          
            	// -> Scale down vector exc[] and redo synthesis 

            	for(i=0; i<PIT_MAX+L_INTERPOL+L_FRAME; i++)
              		//p729Dec->DecStat.old_exc[i] = shr(p729Dec->DecStat.old_exc[i], 2);
			  		p729Dec->DecStat.old_exc[i] >>= 2;
#ifdef VC6
            	Syn_filt(Az, &p729Dec->DecStat.exc[i_subfr], &synth[i_subfr], L_SUBFR, p729Dec->DecStat.mem_syn, 1);
#else
				Syn_filts(Az, &p729Dec->DecStat.exc[i_subfr], &synth[i_subfr], L_SUBFR, p729Dec->DecStat.mem_syn, 1);
#endif
          	}
        	else {
          		Copy(&synth[i_subfr+L_SUBFR-M], p729Dec->DecStat.mem_syn, M);
			}

			//Az += MP1;              /* interpolated LPC parameters for next subframe */
			Az += M;
		}
	}
  
  /*------------*
   *  For G729b
   *-----------*/
	if(bfi == 0) {
#ifdef VC6
		Decod_ld8a_sub2();
#else
		Decod_ld8a_sub2s();
#endif
	}

    /*--------------------------------------------------*
     * Update signal for next frame.                    *
     * -> shift to the left by L_FRAME  exc[]           *
     *--------------------------------------------------*/
	//Copy(&p729Dec->DecStat.old_exc[L_FRAME], &p729Dec->DecStat.old_exc[0], PIT_MAX+L_INTERPOL);
	Copy_a8_154(&p729Dec->DecStat.old_exc[L_FRAME], &p729Dec->DecStat.old_exc[0], PIT_MAX+L_INTERPOL);
	/* for G729b */
	p729Dec->DecStat.past_ftyp = ftyp;
	return;
}






