/*
   ITU-T G.729A Annex B     ANSI-C Source Code
   Version 1.3    Last modified: August 1997
   Copyright (c) 1996, France Telecom, Rockwell International,
                       Universite de Sherbrooke.
   All rights reserved.
*/

/*
**
** File:            "g729_dec_sid.c"
**
** Description:     Comfort noise generation
**                  performed at the decoder part
**
*/
/**** Fixed point version ***/

//#include <stdio.h>
//#include <stdlib.h>
#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "g729_ld8a.h"
#include "g729_tab_ld8a.h"
#include "g729_vad.h"
#include "g729_dtx.h"
#include "g729_sid.h"
#include "g729_tab_dtx.h"
#include "mystruct.h"

//static Word16 cur_gain;
//static Word16 lspSid[M];
//static Word16 sid_gain;
/*
void StoreDecSidStaticVar(DecSidStaticVar *ptr)
{
	int i;

	ptr->cur_gain = cur_gain;
	for (i=0; i<M; i++)
		ptr->lspSid[i] = lspSid[i];
	ptr->sid_gain = sid_gain;
}

void LoadDecSidStaticVar(DecSidStaticVar *ptr)
{
	int i;

	cur_gain = ptr->cur_gain;
	for (i=0; i<M; i++)
		lspSid[i] = ptr->lspSid[i];
	sid_gain = ptr->sid_gain;
}
*/
/*
**
** Function:        Init_Dec_cng()
**
** Description:     Initialize dec_cng static variables
**
**
*/
void Init_Dec_cng(void)
{
  
  Copy( (Word16 *)freq_prev_reset, p729Cod->DSidStat.lspSid, M);
  p729Cod->DSidStat.sid_gain = tab_Sidgain[0];

  return;
}

/*-----------------------------------------------------------*
 * procedure Dec_cng:                                        *
 *           ~~~~~~~~                                        *
 *     Receives frame type                   *
 *     0  :  for untransmitted frames        *
 *     2  :  for SID frames                  *
 *     Decodes SID frames                    *
 *     Computes current frame excitation     *
 *     Computes current frame LSPs
 *-----------------------------------------------------------*/
void Dec_cng(
  Word16 past_ftyp,     /* (i)   : past frame type                      */
  Word16 sid_sav,       /* (i)   : energy to recover SID gain           */
  Word16 sh_sid_sav,    /* (i)   : corresponding scaling factor         */
  Word16 *parm,         /* (i)   : coded SID parameters                 */
  Word16 *exc,          /* (i/o) : excitation array                     */
  Word16 *lsp_old,      /* (i/o) : previous lsp                         */
  Word16 *A_t,          /* (o)   : set of interpolated LPC coefficients */
  Word16 *seed,         /* (i/o) : random generator seed                */
  Word16 freq_prev[MA_NP][M]
                        /* (i/o) : previous LPS for quantization        */
)
{
  Word16 temp;
  Word16 ind;
  Word16 dif;

  dif = sub(past_ftyp, 1);
  
  /* SID Frame */
  if(parm[0] != 0) {
    p729Cod->DSidStat.sid_gain = tab_Sidgain[(int)parm[4]];
    /* Inverse quantization of the LSP 
       parm[2]: quantized index
       parm[3]: quantized index
       p729Cod->DSidStat.lspSid: output quantized lsp vector
       freq_prev: memory of predictor
	 */
    sid_lsfq_decode(&parm[1], p729Cod->DSidStat.lspSid, freq_prev);
  }

  /* non SID Frame */
  else {
    
    /* Case of 1st SID frame erased : quantize-decode   */
    /* energy estimate stored in sid_gain         */
    if(dif == 0) {
      Qua_Sidgain(&sid_sav, &sh_sid_sav, 0, &temp, &ind);
      p729Cod->DSidStat.sid_gain = tab_Sidgain[(int)ind];
    }
    
  }
  
  if(dif == 0) {
    p729Cod->DSidStat.cur_gain = p729Cod->DSidStat.sid_gain;
  }
  else {
    p729Cod->DSidStat.cur_gain = mult_r(p729Cod->DSidStat.cur_gain, A_GAIN0);
    p729Cod->DSidStat.cur_gain = add(p729Cod->DSidStat.cur_gain, mult_r(p729Cod->DSidStat.sid_gain, A_GAIN1));
  }
//#ifdef VC6 
  Calc_exc_rand(p729Cod->DSidStat.cur_gain, exc, seed, FLAG_DEC);
//#else
//  Calc_exc_rands(p729Cod->DSidStat.cur_gain, exc, seed, FLAG_DEC); /* bug in ftyp=0*/
//#endif

  /* Interpolate the Lsp vectors */
#ifdef VC6  
  Int_qlpc(lsp_old, p729Cod->DSidStat.lspSid, A_t);
#else
  Int_qlpcs(lsp_old, p729Cod->DSidStat.lspSid, A_t);
#endif  
  Copy(p729Cod->DSidStat.lspSid, lsp_old, M);
  
  return;
}



/*---------------------------------------------------------------------------*
 * Function  Init_lsfq_noise                                                 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~                                                 *
 *                                                                           *
 * -> Initialization of variables for the lsf quantization in the SID        *
 *                                                                           *
 *---------------------------------------------------------------------------*/
void Init_lsfq_noise(void)			// ref by decoder
{
  Word16 i, j;
  Word32 acc0;

  /* initialize the noise_fg of encoder and decoder */
  for (i=0; i<4; i++)
  {
    Copy((Word16 *)fg[0][i], p729Dec->GloVar.noise_fg[0][i], M);
	Copy((Word16 *)fg[0][i], p729Cod->GloVar.noise_fg[0][i], M);
  }
  
  for (i=0; i<4; i++)
    for (j=0; j<M; j++){
      acc0 = L_mult(fg[0][i][j], 19660);
      acc0 = L_mac(acc0, fg[1][i][j], 13107);
      p729Dec->GloVar.noise_fg[1][i][j] = extract_h(acc0);
	  p729Cod->GloVar.noise_fg[1][i][j] = extract_h(acc0);
    }
}


void sid_lsfq_decode(
	Word16 *index,             /* (i) : quantized indices    */
    Word16 *lspq,              /* (o) : quantized lsp vector */
    Word16 freq_prev[MA_NP][M] /* (i) : memory of predictor  */
)
{
  Word32 acc0;
  Word16 i, j, k, lsfq[M], tmpbuf[M];

  /* get the lsf error vector in Q13 to tmpbuf*/
  Copy((Word16 *)lspcb1[PtrTab_1[index[1]]], tmpbuf, M);
  for (i=0; i<M/2; i++)
    tmpbuf[i] = add(tmpbuf[i], lspcb2[ PtrTab_2[0][index[2]] ][i]);
  for (i=M/2; i<M; i++)
    tmpbuf[i] = add(tmpbuf[i], lspcb2[ PtrTab_2[1][index[2]] ][i]);

  /* guarantee minimum distance of 0.0012 (~10 in Q13) between tmpbuf[j] 
     and tmpbuf[j-1] */
  for ( j = 1 ; j < M ; j++ ){
    acc0 = L_mult(tmpbuf[j-1],      16384);
    acc0 = L_mac (acc0, tmpbuf[j], -16384);
    acc0 = L_mac (acc0, 10,         16384);
    k    = extract_h(acc0);

    if (k > 0){
      tmpbuf[j-1] = sub(tmpbuf[j-1], k);
      tmpbuf[j]   = add(tmpbuf[j],   k);
    }
  }
  
  /* compute the quantized lsf vector */
#ifdef VC6
  Lsp_prev_compose(tmpbuf, lsfq, p729Dec->GloVar.noise_fg[index[0]], freq_prev, 
                   noise_fg_sum[index[0]]);
#else
  Lsp_prev_composes(tmpbuf, lsfq, p729Dec->GloVar.noise_fg[index[0]], freq_prev, 
                   noise_fg_sum[index[0]]);
#endif
  
  /* update the prediction memory */
  Lsp_prev_update(tmpbuf, freq_prev);
  
  /* lsf stability check */
  Lsp_stability(lsfq);

  /* convert lsf to lsp */
  Lsf_lsp2(lsfq, lspq, M);

}





