/*
   ITU-T G.729A Speech Coder with Annex B    ANSI-C Source Code
   Version 1.3    Last modified: March 1997

   Copyright (c) 1996,
   AT&T, France Telecom, NTT, Universite de Sherbrooke, Lucent Technologies,
   Rockwell International
   All rights reserved.
*/

/*-----------------------------------------------------------------*
 * Main program of the G.729A 8.0 kbit/s decoder.                  *
 *                                                                 *
 *     G729_decoder.c
 *                                                                 *
 *-----------------------------------------------------------------*/


#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "g729_ld8a.h"
#include "g729_dtx.h"
#include "g729_octet.h"
#include "mystruct.h"

//Word16 bad_lsf;        /* bad LSF indicator   */
/*
   This variable should be always set to zero unless transmission errors
   in LSP indices are detected.
   This variable is useful if the channel coding designer decides to
   perform error checking on these important parameters. If an error is
   detected on the  LSP indices, the corresponding flag is
   set to 1 signalling to the decoder to perform parameter substitution.
   (The flags should be set back to 0 for correct transmission).
*/
//Word16 synth_buf[L_FRAME+M], *synth; /* Synthesis                   */
//extern Word16 noise_fg[MODE][MA_NP][M];


void bit2parm(char *DecodIn, Word16 *parm, Flag bfi);

/*-----------------------------------------------------------------*
 *           Initialization of decoder                             *
 *-----------------------------------------------------------------*/

void Init_G729ab_Decoder(void)
{
	Word16 i;
	for (i=0; i<M; i++) 
		p729Dec->GloVar.synth_buf[i] = 0;
	p729Dec->GloVar.synth = p729Dec->GloVar.synth_buf + M;
	p729Dec->GloVar.bad_lsf = 0;          /* Initialize bad LSF indicator */
	Init_Decod_ld8a();
	Init_Post_Filter();
	Init_Post_Process();
	/* for G.729b */
	Init_Dec_cng();

	/*for (ch=0; ch<4; ch++)
	{
		StoreDecoderVar(ch);
	}*/
}


#ifdef VC6
void bit2parm(char *DecodIn,/*(i)decoder input bitstream*/
			Word16 *parm,   /*(o)*/
			Flag BadFrameIndicator
	)
{
	Word16 i,j = 0;	
	Word16 BitLength=0;
	Word16 serial[SERIAL_SIZE];/* SERIAL_SIZE = 80+2 */
	
	if (BadFrameIndicator == 1)
	{
		parm[0] = 1;/*Its a bad frame*/
		return ;
	}
	parm[0] = 0;/*assume no bad frame*/
	switch (DecodIn[0])
	{
	case 0x00: 	/* not TX (the frame is not transmitted)*/
		serial[1] = 0;
		BitLength = 0;
		break;
	case 0x01:	/* active a typical voice frame */
		serial[1] = 80;
		BitLength = 80;
		break;
	case 0x02:	/* SID silence insertion descriptor */
		serial[1] = 16;
		BitLength = 16;
		break;	
	default:	/* bad frame indicator as default*/
		serial[1] = 0;
		parm[0]   = 1;
	}
	j = 2;
	/********************************************************
	        0 1 2 3 4 5 6 7 8 9 10
	DecodIn G p p p p p p p p p p
	serial  ? L P P P P		.......................81
	
	        0 1 2 3 4 5 6 7 8 9 10 11 12
	parm    G T 
	********************************************************/
	/* Get 1 Word16 to represent 1 bit */
  	for ( i = 0 ; i < BitLength ; i ++ ) {
		serial[i+j] =  ( DecodIn[1 + (i>>3)] >> (7 - (i & 0x0007)) ) & 1;
	}
	/*
	 i  i+2  [1+(i>>3)]  (7-(i & 0x0007))
	 0    2		1			7
	 1    3		1			6
	 2    4		1			5
	 3    5		1			4
	 4    6		1			3
	 5    7		1			2
	 6    8		1			1
	 7    9		1			0
	 8   10		2			7
	 */
	// We are synchrounous with ITU-T codes right now
  	bits2prm_ld8k(&serial[1]/*in*/, parm/*out*/);

	/* parm[0] stores good frame or bad frame
	   parm[1] stores frame type 80 16 or 0
	 */

    /* check parity and put 1 in parm[5] if parity error
	   NOTE !! parm[5] be modified !! 
       parm[1]==1, active frame
       parm[1]==2, SID    frame
     */
  	if(parm[1] == 1) {
		/* checking the symbol P0 */
		parm[5] = Check_Parity_Pitch(parm[4], parm[5]);
  	}
}
#endif

void G729ab_Decoder(char *DecodIn, Word16 *decOut, Word16 bfi, int iframe)
{  
	static Word16  parm[PRM_SIZE+2]; /* Synthesis parameters 13     */
	static Word16  Az_dec[MP1*2];    /* Decoded Az for post-filter  */
	Word16  T2[2];            /* Pitch lag for 2 subframes   */
	Word16  Vad;
 
#ifndef OCTET_TX_MODE
	printk("OCTET TRANSMISSION MODE is disabled\n");
#endif
             /*11*/ /*13*/
	bit2parm(DecodIn, parm, bfi);		// parsing bitstream and reconstruct parameters	
    Decod_ld8a(parm, p729Dec->GloVar.synth, Az_dec, T2, &Vad);
#ifdef VC6
    Post_Filter(p729Dec->GloVar.synth, Az_dec, T2, Vad);        /* Post-filter */
#else
	Post_Filters(p729Dec->GloVar.synth, Az_dec, T2, Vad);        /* Post-filter */
#endif

#ifdef VC6
    Post_Process(p729Dec->GloVar.synth, L_FRAME);
#else
	Post_Processs(p729Dec->GloVar.synth, L_FRAME);
#endif
	/* Copy synthesis speech from synth to decout */
	Copy_a8_m8(p729Dec->GloVar.synth, decOut, L_FRAME);
	return;
}

/*
static GlobalVarD chdg[4];
static DecGainStaticVar chds1[4];
static Decld8aStaticVar chds2[4];
static LspDecStaticVar chds3[4];
static PostProStaticVar chds4[4];
static DecSidStaticVar chds5[4];
static TamingStaticVar chds6[4];
static PostfiltStaticVar chds7[4];
void StoreGlobalVarD(GlobalVarD *ptr)
{
	int i,j,k;
	ptr->bad_lsf = bad_lsf;
	for (i=0; i<L_FRAME+M; i++)
		ptr->synth_buf[i] = synth_buf[i];
	ptr->synth = synth;
	for (i=0; i<MODE; i++)
		for (j=0; j<MA_NP; j++)
			for (k=0; k<M; k++)
				ptr->noise_fg[i][j][k] = noise_fg[i][j][k];
}

void LoadGlobalVarD(GlobalVarD *ptr)
{
	int i,j,k;
	bad_lsf = ptr->bad_lsf;
	for (i=0; i<L_FRAME+M; i++)
		synth_buf[i] = ptr->synth_buf[i];
	synth = ptr->synth;
	for (i=0; i<MODE; i++)
		for (j=0; j<MA_NP; j++)
			for (k=0; k<M; k++)
				noise_fg[i][j][k] = ptr->noise_fg[i][j][k];
}
*/



