/*
   ITU-T G.729A Speech Coder    ANSI-C Source Code
   Version 1.1    Last modified: September 1996
   Copyright (c) 1996,
   AT&T, France Telecom, NTT, Universite de Sherbrooke, Lucent Technologies
   All rights reserved.
*/

#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "g729_ld8a.h"


/*-------------------------------------------------------------------*
 * Function  Convolve:                                               *
 *           ~~~~~~~~~                                               *
 *-------------------------------------------------------------------*
 * Perform the convolution between two vectors x[] and h[] and       *
 * write the result in the vector y[].                               *
 * All vectors are of length N.                                      *
 *-------------------------------------------------------------------*/

#if 0

void Convolve(
  Word16 x[],      /* (i)     : input vector                           */
  Word16 h[],      /* (i) Q12 : impulse response                       */
  Word16 y[],      /* (o)     : output vector                          */
  Word16 L         /* (i)     : vector size                            */
)
{
   Word16 i, n;
   Word32 s;

   for (n = 0; n < L; n++)
   {
     s = 0;
     for (i = 0; i <= n; i++)
       s = L_mac(s, x[i], h[n-i]);

     s    = L_shl(s, 3);                   /* h is in Q12 and saturation */
     y[n] = extract_h(s);
   }

   return;
}

#endif

/*
       void Iir(Word32 *Acc0, Word32 *Acc2, Word16 *pCoe, Word16 *pDat)
*/

static void Iir(
         Word32 *Acc0,	/* (i/o) first sample accumulator */
		 Word32 *Acc2,	/* (i/o) 2nd sample accumulator   */
		 Word16 *pCoe,	/* (i) coeff                      */
		 Word16 *pDat)	/* (i) data                       */
						/* processing 2 samples each time */
{
	Word32	Acc1, Acc3;
	
	*Acc0 = L_msu(*Acc0, pCoe[0], pDat[0]);	/* Acc0 = Acc0 - pCoe * pDat */
	*Acc0 = L_msu(*Acc0, pCoe[1], pDat[1]);
	*Acc0 = L_msu(*Acc0, pCoe[2], pDat[2]);	
	*Acc0 = L_msu(*Acc0, pCoe[3], pDat[3]);
	*Acc0 = L_msu(*Acc0, pCoe[4], pDat[4]);	
	*Acc0 = L_msu(*Acc0, pCoe[5], pDat[5]);
	*Acc0 = L_msu(*Acc0, pCoe[6], pDat[6]);	
	*Acc0 = L_msu(*Acc0, pCoe[7], pDat[7]);
	*Acc0 = L_msu(*Acc0, pCoe[8], pDat[8]);	
	*Acc0 = L_msu(*Acc0, pCoe[9], pDat[9]);        

	*Acc2 = L_msu(*Acc2, pCoe[1], pDat[0]);
	*Acc2 = L_msu(*Acc2, pCoe[2], pDat[1]);	
	*Acc2 = L_msu(*Acc2, pCoe[3], pDat[2]);
	*Acc2 = L_msu(*Acc2, pCoe[4], pDat[3]);	
	*Acc2 = L_msu(*Acc2, pCoe[5], pDat[4]);
	*Acc2 = L_msu(*Acc2, pCoe[6], pDat[5]);	
	*Acc2 = L_msu(*Acc2, pCoe[7], pDat[6]);
	*Acc2 = L_msu(*Acc2, pCoe[8], pDat[7]);	
	*Acc2 = L_msu(*Acc2, pCoe[9], pDat[8]);
	Acc1  = L_shl( *Acc0, (Word16) 3 ) ;
	pDat[-1] = round( Acc1 );
	*Acc2 = L_msu(*Acc2, pCoe[0], pDat[-1]);
	Acc3  = L_shl( *Acc2, (Word16) 3);
	pDat[-2] = round( Acc3 );

}

/*-----------------------------------------------------*
 * procedure Syn_filt:                                 *
 *           ~~~~~~~~                                  *
 * Do the synthesis filtering 1/A(z).                  *
 *-----------------------------------------------------*/
void Syn_filt(
  /*Word16 a[],*//*(i) Q12 : a[m+1] prediction coefficients   (m=10)  */
  Word16 a[],   /* (i) Q12 : a[m] prediction coefficients   (m=10)    */
  Word16 x[],   /* (i)     : input signal                             */
  Word16 y[],   /* (o)     : output signal                            */
  Word16 lg,    /* (i)     : size of filtering(40 or 22)              */
  Word16 mem[], /* (i/o)   : memory associated with this filtering.   */
  Word16 update /* (i)     : 0=no update, 1=update of memory.         */
)
{
	Word16 i;
	Word32 s, s2;
	Word16 tmp[L_SUBFR+M];/* This is usually done by memory allocation (lg+M) */
	Word16 *yy;

	/* Copy mem[] to yy[] */
	yy = &tmp[L_SUBFR];
	for( i = 0 ; i < M ; i++ )
	{ 
		yy[i] = mem[M-1-i];	/*9,8,7,....1,0*/
	}

	/* Do the filtering. */
	for( i = 0 ; i < lg ; i+=2 )
	{
		s      = L_mult(x[i],   /*a[0]*/ 0x1000 );
		s2     = L_mult(x[i+1], /*a[0]*/ 0x1000 );
		Iir(&s, &s2, &a[0]/*pCoe*/, yy/*pDat*/);
		y[i]   = yy[-1];
		y[i+1] = yy[-2];
		yy    -= 2;
	}
	/* Update of memory if update==1 */
	if(update != 0)
		for (i = 0; i < M; i++)
		{
			mem[i] = y[lg-M+i];       
		}

	return;
}

/*-----------------------------------------------------------------------*
 * procedure Residu:                                                     *
 *           ~~~~~~                                                      *
 * Compute the LPC residual  by filtering the input speech through A(z)  *
 *-----------------------------------------------------------------------*/

void Residu(
  Word16 a[],/* (i) Q12 : prediction coefficients                     */
  Word16 x[],/* (i)     : speech (values x[-m..-1] are needed         */
  Word16 y[],/* (o)     : residual signal                             */
  Word16 lg  /* (i)     : size of filtering : L_SUBFR                 */
)
{
	Word16 i,j;
	Word32 s, s2, s3, s4;
	Word16 coef[M+1+1], invx[L_SUBFR+M];

	/* copy a[] to coef[] , inverse sequence */
	for( i = 0 ; i < M ; i++ )
		coef[i] = a[M-1-i];
	coef[M]   = 0x1000;
 	coef[M+1] = 0x0000;
	/* copy x mem */
	for( i = 0 ; i < L_SUBFR + M /*=50*/ ; i++ )
		invx[i] = x[i-M];        /*-10,-9,....38,39*/
	/* processing */
	for( i = 0 ; i < L_SUBFR ; i+=4 )
	{
		s=0; s2=0; s3=0; s4=0;
		for( j = 0 ; j < M+2 ; j++ )
		{
			s  = L_mac(s,  coef[j], invx[i+j]  );
			s2 = L_mac(s2, coef[j], invx[i+1+j]);
			s3 = L_mac(s3, coef[j], invx[i+2+j]);
			s4 = L_mac(s4, coef[j], invx[i+3+j]);
		}
		s  = L_shl(s, 3);
		s2 = L_shl(s2,3);
		s3 = L_shl(s3,3);
		s4 = L_shl(s4,3);
		y[i]   = round(s );
		y[i+1] = round(s2);
		y[i+2] = round(s3);
		y[i+3] = round(s4);
	}
	return;
}




