/*
   ITU-T G.729A Speech Coder    ANSI-C Source Code
   Version 1.1    Last modified: September 1996

   Copyright (c) 1996,
   AT&T, France Telecom, NTT, Universite de Sherbrooke, Lucent Technologies
   All rights reserved.
*/

/*-------------------------------------------------------------------*
 * Function  Pred_lt_3()                                             *
 *           ~~~~~~~~~~~                                             *
 *-------------------------------------------------------------------*
 * Compute the result of long term prediction with fractional        *
 * interpolation of resolution 1/3.                                  *
 *                                                                   *
 * On return exc[0..L_subfr-1] contains the interpolated signal      *
 *   (adaptive codebook excitation)                                  *
 *-------------------------------------------------------------------*/
#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "g729_ld8a.h"
#include "g729_tab_ld8a.h"


void Pred_lt_3(
  Word16   exc[],       /* in/out: excitation buffer */
  Word16   T0,          /* input : integer pitch lag */
  Word16   frac,        /* input : fraction of lag all it can be -1 , 0, 1  */
  Word16   L_subfr      /* input : subframe size     */
)
{
  Word16   i, j, k;
  Word16   *x0, *px1, *px2, *c1, *c2;
  Word32   s, s1, s2, s3;
  Word16   fac1[L_INTER10+2], fac2[L_INTER10+2];	
  Word16   xl[L_SUBFR+L_INTER10+PIT_MAX+M], xr[L_SUBFR+L_INTER10+PIT_MAX+M];  
  Word16   dis;	//addr distance bw exc and x0

  x0 = &exc[-T0];

  frac = negate(frac);
  if (frac < 0)
  {
    frac = add(frac, UP_SAMP/*3*/);
    x0--;
  }

  dis = (Word16)(exc - x0);

  c1 = (Word16 *)&inter_3l[frac];
  c2 = (Word16 *)&inter_3l[sub(UP_SAMP,frac)];
    
  for( i=0, k=0 ; i < L_INTER10 ; i++, k += UP_SAMP )	/* L_INTER10=10 */
  {
	fac1[L_INTER10-1-i] = c1[k];	//inverse oder of c1[]
	fac2[i]             = c2[k];	
  }
  fac1[10]=fac1[11]=0;	//pad 2 zero
  fac2[10]=fac2[11]=0;	//pad 2 zero

  /* copy data L_SUBFR + L_INTER10 = 50 Copy50(x,y,lg) */
  Copy50(&x0[-9], xl, L_SUBFR + L_INTER10);
  Copy50( &x0[1], xr,  L_SUBFR + L_INTER10);
  px1 =xl;
  px2 =xr;

  for(j=0; j<L_SUBFR; j+=4)
  {
	  s = 0;	  
	  s1=0;
	  s2=0;
	  s3=0;
	  for(i=0; i<L_INTER10; i++)
	  {
		  s  = L_mac(s,  px1[i],   fac1[i]);		  
		  s1 = L_mac(s1, px1[i+1], fac1[i]);		  
		  s2 = L_mac(s2, px1[i+2], fac1[i]);		  
		  s3 = L_mac(s3, px1[i+3], fac1[i]);		  
	  }
	  for(i=0; i<L_INTER10; i++)
	  {
		  s = L_mac(s, px2[i], fac2[i]);		  
		  s1 = L_mac(s1, px2[i+1], fac2[i]);		  
		  s2 = L_mac(s2, px2[i+2], fac2[i]);		  
		  s3 = L_mac(s3, px2[i+3], fac2[i]);
	  }

	  exc[j] = round(s);	  
	  exc[j+1] =round(s1);
	  exc[j+2] = round(s2);
	  exc[j+3] = round(s3);
	  //if((dis+9+j) < 50)
		  xl[dis+9+j] = exc[j];
		  xl[dis+9+j+1] = exc[j+1];
		  xl[dis+9+j+2] = exc[j+2];
		  xl[dis+9+j+3] = exc[j+3];

	  //if((dis-1+j) < 50)
		  xr[dis-1+j] = exc[j];
		  xr[dis-1+j+1] = exc[j+1];
		  xr[dis-1+j+2] = exc[j+2];
		  xr[dis-1+j+3] = exc[j+3];
	  px1 += 4;
	  px2 += 4;
  }

  return;
}


