/*
   ITU-T G.729A Annex B     ANSI-C Source Code
   Version 1.3    Last modified: August 1997
   Copyright (c) 1996, France Telecom, Rockwell International,
                       Universite de Sherbrooke.
   All rights reserved.
*/

//#include <stdio.h>
#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "g729_ld8a.h"
#include "g729_oper_32b.h"
#include "g729_tab_ld8a.h"
#include "g729_vad.h"
#include "g729_dtx.h"
#include "g729_tab_dtx.h"
#include "mystruct.h"

/* local function */
static Word16 MakeDec(
               Word16 dSLE,    /* (i)  : differential low band energy */
               Word16 dSE,     /* (i)  : differential full band energy */
               Word16 SD,      /* (i)  : differential spectral distortion */
               Word16 dSZC     /* (i)  : differential zero crossing rate */
);

/* static variables */
//static Word16 MeanLSF[M];
//static Word16 Min_buffer[16];
//static Word16 Prev_Min, Next_Min, Min;
//static Word16 MeanE, MeanSE, MeanSLE, MeanSZC;
//static Word16 prev_energy;
//static Word16 count_sil, count_update, count_ext;
//static Word16 flag, v_flag, less_count;
/*
void StoreVadStaticVar(VadStaticVar *ptr)
{
	int i;
	for (i=0; i<M; i++)
		ptr->MeanLSF[i] = MeanLSF[i];
	for (i=0; i<16; i++)
		ptr->Min_buffer[i] = Min_buffer[i];
	ptr->Prev_Min = Prev_Min;
	ptr->Next_Min = Next_Min;	
	ptr->Min = Min;
	ptr->MeanE = MeanE;
	ptr->MeanSE = MeanSE;
	ptr->MeanSLE = MeanSLE;
	ptr->MeanSZC = MeanSZC;
	ptr->prev_energy = prev_energy;
	ptr->count_sil = count_sil;
	ptr->count_update = count_update;
	ptr->count_ext = count_ext;
	ptr->flag = flag;
	ptr->v_flag = v_flag;
	ptr->less_count = less_count;
}

void LoadVadStaticVar(VadStaticVar *ptr)
{
	int i;
	for (i=0; i<M; i++)
		MeanLSF[i] = ptr->MeanLSF[i];
	for (i=0; i<16; i++)
		Min_buffer[i] = ptr->Min_buffer[i];
	Prev_Min = ptr->Prev_Min;
	Next_Min = ptr->Next_Min;
	Min = ptr->Min;
	MeanE = ptr->MeanE;
	MeanSE = ptr->MeanSE;
	MeanSLE = ptr->MeanSLE;
	MeanSZC = ptr->MeanSZC;
	prev_energy = ptr->prev_energy;
	count_sil = ptr->count_sil;
	count_update = ptr->count_update;
	count_ext = ptr->count_ext;
	flag = ptr->flag;
	v_flag = ptr->v_flag;
	less_count = ptr->less_count;
}
*/
/*---------------------------------------------------------------------------*
 * Function  vad_init                                                        *
 * ~~~~~~~~~~~~~~~~~~                                                        *
 *                                                                           *
 * -> Initialization of variables for voice activity detection               *
 *                                                                           *
 *---------------------------------------------------------------------------*/
void vad_init(void)
{
  /* Static vectors to zero */
  Set_zero(p729Cod->VadStat.MeanLSF, M);

  /* Initialize VAD parameters */
  p729Cod->VadStat.MeanSE = 0;
  p729Cod->VadStat.MeanSLE = 0;
  p729Cod->VadStat.MeanE = 0;
  p729Cod->VadStat.MeanSZC = 0;
  p729Cod->VadStat.count_sil = 0;
  p729Cod->VadStat.count_update = 0;
  p729Cod->VadStat.count_ext = 0;
  p729Cod->VadStat.less_count = 0;
  p729Cod->VadStat.flag = 1;
  p729Cod->VadStat.Min = MAX_16;
}


/*-----------------------------------------------------------------*
 * Functions vad                                                   *
 *           ~~~                                                   *
 * Input:                                                          *
 *   rc            : reflection coefficient                        *
 *   lsf[]         : unquantized lsf vector                        *
 *   r_h[]         : upper 16-bits of the autocorrelation vector   *
 *   r_l[]         : lower 16-bits of the autocorrelation vector   *
 *   exp_R0        : exponent of the autocorrelation vector        *
 *   sigpp[]       : preprocessed input signal                     *
 *   frm_count     : frame counter                                 *
 *   prev_marker   : VAD decision of the last frame                *
 *   pprev_marker  : VAD decision of the frame before last frame   *
 *                                                                 *
 * Output:                                                         *
 *                                                                 *
 *   marker        : VAD decision of the current frame             * 
 *                                                                 *
 *-----------------------------------------------------------------*/
/*
vad( rc[1], lsf_new, r_h, r_l, exp_R0, p729Cod->CodStat.p_window, frame,
                                       p729Cod->CodStat.pastVad, 
                                       p729Cod->CodStat.ppastVad, &Vad);
*/
void vad(
         Word16 rc,
         Word16 *lsf, 
         Word16 *r_h,
         Word16 *r_l, 
         Word16 exp_R0,
         Word16 *sigpp,/* signal preprocessed */
         Word16 frm_count,
         Word16 prev_marker,
         Word16 pprev_marker,
         Word16 *marker)
{
  /* scalar */
  Word32 acc0;
  Word16 exp, frac;
  Word16 i, j;
  Word16 ENERGY, ENERGY_low, SD, ZC, dSE, dSLE, dSZC;
  Word16 COEF, C_COEF, COEFZC, C_COEFZC, COEFSD, C_COEFSD;

  /* compute the frame energy */
  acc0 = L_Comp(r_h[0], r_l[0]);
  Log2(acc0, &exp, &frac);
  acc0 = Mpy_32_16(exp, frac, 9864);/*if in Q15: 0.301025...*/
  i = sub(exp_R0, 1);  
  i = sub(i, 1);
  acc0 = L_mac(acc0, 9864, i);
  acc0 = L_shl(acc0, 11);
  ENERGY = extract_h(acc0);
  ENERGY = sub(ENERGY, 4875);       /*if in Q15: 0.148...*/

  /* compute the low band energy */
  acc0 = 0;
  for (i=1; i<=NP; i++)
    acc0 = L_mac(acc0, r_h[i], lbf_corr[i]);
  acc0 = L_shl(acc0, 1);
  acc0 = L_mac(acc0, r_h[0], lbf_corr[0]);
  Log2(acc0, &exp, &frac);
  acc0 = Mpy_32_16(exp, frac, 9864);
  i = sub(exp_R0, 1);  
  i = sub(i, 1);
  acc0 = L_mac(acc0, 9864, i);
  acc0 = L_shl(acc0, 11);
  ENERGY_low = extract_h(acc0);
  ENERGY_low = sub(ENERGY_low, 4875);
  
  /* compute SD */
  acc0 = 0;
  for (i=0; i<M; i++){
    j = sub(lsf[i], p729Cod->VadStat.MeanLSF[i]);
    acc0 = L_mac(acc0, j, j);
  }
  SD = extract_h(acc0);      /* Q15 */
  
  /* compute # zero crossing */
  ZC = 0;
  for (i=ZC_START+1; i<=ZC_END; i++)
    if (mult(sigpp[i-1], sigpp[i]) < 0)
      ZC = add(ZC, 410);     /* Q15 */

  /* Initialize and update Mins */
  if(sub(frm_count, 129) < 0){
    if (sub(ENERGY, p729Cod->VadStat.Min) < 0){
      p729Cod->VadStat.Min = ENERGY;
      p729Cod->VadStat.Prev_Min = ENERGY;
    }
    
    if((frm_count & 0x0007) == 0){
      i = sub(shr(frm_count,3),1);
      p729Cod->VadStat.Min_buffer[i] = p729Cod->VadStat.Min;
      p729Cod->VadStat.Min = MAX_16;
    }
  }

  if((frm_count & 0x0007) == 0){
    p729Cod->VadStat.Prev_Min = p729Cod->VadStat.Min_buffer[0];
    for (i=1; i<16; i++){
      if (sub(p729Cod->VadStat.Min_buffer[i], p729Cod->VadStat.Prev_Min) < 0)
        p729Cod->VadStat.Prev_Min = p729Cod->VadStat.Min_buffer[i];
    }
  }
  
  if(sub(frm_count, 129) >= 0){
    if(((frm_count & 0x0007) ^ (0x0001)) == 0){
      p729Cod->VadStat.Min = p729Cod->VadStat.Prev_Min;
      p729Cod->VadStat.Next_Min = MAX_16;
    }
    if (sub(ENERGY, p729Cod->VadStat.Min) < 0)
      p729Cod->VadStat.Min = ENERGY;
    if (sub(ENERGY, p729Cod->VadStat.Next_Min) < 0)
      p729Cod->VadStat.Next_Min = ENERGY;
    
    if((frm_count & 0x0007) == 0){
      for (i=0; i<15; i++)
        p729Cod->VadStat.Min_buffer[i] = p729Cod->VadStat.Min_buffer[i+1]; 
      p729Cod->VadStat.Min_buffer[15] = p729Cod->VadStat.Next_Min; 
      p729Cod->VadStat.Prev_Min = p729Cod->VadStat.Min_buffer[0];
      for (i=1; i<16; i++) 
        if (sub(p729Cod->VadStat.Min_buffer[i], p729Cod->VadStat.Prev_Min) < 0)
          p729Cod->VadStat.Prev_Min = p729Cod->VadStat.Min_buffer[i];
    }
    
  }

  if (sub(frm_count, INIT_FRAME/*32*/) <= 0)
  {
    if(sub(ENERGY, 3072) < 0){
      *marker = NOISE;
      p729Cod->VadStat.less_count++;
    }
    else{
      *marker = VOICE;
      acc0 = L_deposit_h(p729Cod->VadStat.MeanE);
      acc0 = L_mac(acc0, ENERGY, 1024);
      p729Cod->VadStat.MeanE = extract_h(acc0);
      acc0 = L_deposit_h(p729Cod->VadStat.MeanSZC);
      acc0 = L_mac(acc0, ZC, 1024);
      p729Cod->VadStat.MeanSZC = extract_h(acc0);
      for (i=0; i<M; i++){
        acc0 = L_deposit_h(p729Cod->VadStat.MeanLSF[i]);
        acc0 = L_mac(acc0, lsf[i], 1024);
        p729Cod->VadStat.MeanLSF[i] = extract_h(acc0);
      }
    }
  }
  
  if (sub(frm_count, INIT_FRAME) >= 0){
    if (sub(frm_count, INIT_FRAME) == 0){
      acc0 = L_mult(p729Cod->VadStat.MeanE, factor_fx[p729Cod->VadStat.less_count]);
      acc0 = L_shl(acc0, shift_fx[p729Cod->VadStat.less_count]);
      p729Cod->VadStat.MeanE = extract_h(acc0);

      acc0 = L_mult(p729Cod->VadStat.MeanSZC, factor_fx[p729Cod->VadStat.less_count]);
      acc0 = L_shl(acc0, shift_fx[p729Cod->VadStat.less_count]);
      p729Cod->VadStat.MeanSZC = extract_h(acc0);

      for (i=0; i<M; i++){
        acc0 = L_mult(p729Cod->VadStat.MeanLSF[i], factor_fx[p729Cod->VadStat.less_count]);
        acc0 = L_shl(acc0, shift_fx[p729Cod->VadStat.less_count]);
        p729Cod->VadStat.MeanLSF[i] = extract_h(acc0);
      }

      p729Cod->VadStat.MeanSE = sub(p729Cod->VadStat.MeanE, 2048);   /* Q11 */
      p729Cod->VadStat.MeanSLE = sub(p729Cod->VadStat.MeanE, 2458);  /* Q11 */
    }

    dSE = sub(p729Cod->VadStat.MeanSE, ENERGY);
    dSLE = sub(p729Cod->VadStat.MeanSLE, ENERGY_low);
    dSZC = sub(p729Cod->VadStat.MeanSZC, ZC);

    if(sub(ENERGY, 3072) < 0)
      *marker = NOISE;
    else 
      *marker = MakeDec(dSLE, dSE, SD, dSZC);

    p729Cod->VadStat.v_flag = 0;
    if((prev_marker==VOICE) && (*marker==NOISE) && (add(dSE,410) < 0) 
       && (sub(ENERGY, 3072)>0)){
      *marker = VOICE;
      p729Cod->VadStat.v_flag = 1;
    }

    if(p729Cod->VadStat.flag == 1){
      if((pprev_marker == VOICE) && 
         (prev_marker == VOICE) && 
         (*marker == NOISE) && 
         (sub(abs_s(sub(p729Cod->VadStat.prev_energy,ENERGY)), 614) <= 0)){
        p729Cod->VadStat.count_ext++;
        *marker = VOICE;
        p729Cod->VadStat.v_flag = 1;
        if(sub(p729Cod->VadStat.count_ext, 4) <= 0)
          p729Cod->VadStat.flag=1;
        else{
          p729Cod->VadStat.count_ext=0;
          p729Cod->VadStat.flag=0;
        }
      }
    }
    else
      p729Cod->VadStat.flag=1;
    
    if(*marker == NOISE)
      p729Cod->VadStat.count_sil++;

    if((*marker == VOICE) && (sub(p729Cod->VadStat.count_sil, 10) > 0) && 
       (sub(sub(ENERGY,p729Cod->VadStat.prev_energy), 614) <= 0)){
      *marker = NOISE;
      p729Cod->VadStat.count_sil=0;
    }

    if(*marker == VOICE)
      p729Cod->VadStat.count_sil=0;

    if ((sub(sub(ENERGY, 614), p729Cod->VadStat.MeanSE) < 0) && (sub(frm_count, 128) > 0)
        && (!p729Cod->VadStat.v_flag) && (sub(rc, 19661) < 0))
      *marker = NOISE;

    if ((sub(sub(ENERGY,614),p729Cod->VadStat.MeanSE) < 0) && (sub(rc, 24576) < 0)
        && (sub(SD, 83) < 0)){ 
      p729Cod->VadStat.count_update++;
      if (sub(p729Cod->VadStat.count_update, INIT_COUNT) < 0){
        COEF = 24576;
        C_COEF = 8192;
        COEFZC = 26214;
        C_COEFZC = 6554;
        COEFSD = 19661;
        C_COEFSD = 13017;
      } 
      else
        if (sub(p729Cod->VadStat.count_update, INIT_COUNT+10) < 0){
          COEF = 31130;
          C_COEF = 1638;
          COEFZC = 30147;
          C_COEFZC = 2621;
          COEFSD = 21299;
          C_COEFSD = 11469;
        }
        else
          if (sub(p729Cod->VadStat.count_update, INIT_COUNT+20) < 0){
            COEF = 31785;
            C_COEF = 983;
            COEFZC = 30802;
            C_COEFZC = 1966;
            COEFSD = 22938;
            C_COEFSD = 9830;
          }
          else
            if (sub(p729Cod->VadStat.count_update, INIT_COUNT+30) < 0){
              COEF = 32440;
              C_COEF = 328;
              COEFZC = 31457;
              C_COEFZC = 1311;
              COEFSD = 24576;
              C_COEFSD = 8192;
            }
            else
              if (sub(p729Cod->VadStat.count_update, INIT_COUNT+40) < 0){
                COEF = 32604;
                C_COEF = 164;
                COEFZC = 32440;
                C_COEFZC = 328;
                COEFSD = 24576;
                C_COEFSD = 8192;
              }
              else{
                COEF = 32604;
                C_COEF = 164;
                COEFZC = 32702;
                C_COEFZC = 66;
                COEFSD = 24576;
                C_COEFSD = 8192;
              }
      

      /* compute MeanSE */
      acc0 = L_mult(COEF, p729Cod->VadStat.MeanSE);
      acc0 = L_mac(acc0, C_COEF, ENERGY);
      p729Cod->VadStat.MeanSE = extract_h(acc0);

      /* compute MeanSLE */
      acc0 = L_mult(COEF, p729Cod->VadStat.MeanSLE);
      acc0 = L_mac(acc0, C_COEF, ENERGY_low);
      p729Cod->VadStat.MeanSLE = extract_h(acc0);

      /* compute MeanSZC */
      acc0 = L_mult(COEFZC, p729Cod->VadStat.MeanSZC);
      acc0 = L_mac(acc0, C_COEFZC, ZC);
      p729Cod->VadStat.MeanSZC = extract_h(acc0);
      
      /* compute MeanLSF */
      for (i=0; i<M; i++){
        acc0 = L_mult(COEFSD, p729Cod->VadStat.MeanLSF[i]);
        acc0 = L_mac(acc0, C_COEFSD, lsf[i]);
        p729Cod->VadStat.MeanLSF[i] = extract_h(acc0);
      }
    }

    if((sub(frm_count, 128) > 0) && (((sub(p729Cod->VadStat.MeanSE,p729Cod->VadStat.Min) < 0) &&
                        (sub(SD, 83) < 0)) || (sub(p729Cod->VadStat.MeanSE,p729Cod->VadStat.Min) > 2048))){
      p729Cod->VadStat.MeanSE = p729Cod->VadStat.Min;
      p729Cod->VadStat.count_update = 0;
    }
  }

  p729Cod->VadStat.prev_energy = ENERGY;

}

/* local function */  
static Word16 MakeDec(
               Word16 dSLE,    /* (i)  : differential low band energy */
               Word16 dSE,     /* (i)  : differential full band energy */
               Word16 SD,      /* (i)  : differential spectral distortion */
               Word16 dSZC     /* (i)  : differential zero crossing rate */
               )
{
  Word32 acc0;
  
  /* SD vs dSZC */
  acc0 = L_mult(dSZC, -14680);          /* Q15*Q23*2 = Q39 */  
  acc0 = L_mac(acc0, 8192, -28521);     /* Q15*Q23*2 = Q39 */
  acc0 = L_shr(acc0, 8);                /* Q39 -> Q31 */
  acc0 = L_add(acc0, L_deposit_h(SD));
  if (acc0 > 0) return(VOICE);

  acc0 = L_mult(dSZC, 19065);           /* Q15*Q22*2 = Q38 */
  acc0 = L_mac(acc0, 8192, -19446);     /* Q15*Q22*2 = Q38 */
  acc0 = L_shr(acc0, 7);                /* Q38 -> Q31 */
  acc0 = L_add(acc0, L_deposit_h(SD));
  if (acc0 > 0) return(VOICE);

  /* dSE vs dSZC */
  acc0 = L_mult(dSZC, 20480);           /* Q15*Q13*2 = Q29 */
  acc0 = L_mac(acc0, 8192, 16384);      /* Q13*Q15*2 = Q29 */
  acc0 = L_shr(acc0, 2);                /* Q29 -> Q27 */
  acc0 = L_add(acc0, L_deposit_h(dSE));
  if (acc0 < 0) return(VOICE);

  acc0 = L_mult(dSZC, -16384);          /* Q15*Q13*2 = Q29 */
  acc0 = L_mac(acc0, 8192, 19660);      /* Q13*Q15*2 = Q29 */
  acc0 = L_shr(acc0, 2);                /* Q29 -> Q27 */
  acc0 = L_add(acc0, L_deposit_h(dSE));
  if (acc0 < 0) return(VOICE);
 
  acc0 = L_mult(dSE, 32767);            /* Q11*Q15*2 = Q27 */
  acc0 = L_mac(acc0, 1024, 30802);      /* Q10*Q16*2 = Q27 */
  if (acc0 < 0) return(VOICE);
  
  /* dSE vs SD */
  acc0 = L_mult(SD, -28160);            /* Q15*Q5*2 = Q22 */
  acc0 = L_mac(acc0, 64, 19988);        /* Q6*Q14*2 = Q22 */
  acc0 = L_mac(acc0, dSE, 512);         /* Q11*Q9*2 = Q22 */
  if (acc0 < 0) return(VOICE);

  acc0 = L_mult(SD, 32767);             /* Q15*Q15*2 = Q31 */
  acc0 = L_mac(acc0, 32, -30199);       /* Q5*Q25*2 = Q31 */
  if (acc0 > 0) return(VOICE);

  /* dSLE vs dSZC */
  acc0 = L_mult(dSZC, -20480);          /* Q15*Q13*2 = Q29 */
  acc0 = L_mac(acc0, 8192, 22938);      /* Q13*Q15*2 = Q29 */
  acc0 = L_shr(acc0, 2);                /* Q29 -> Q27 */
  acc0 = L_add(acc0, L_deposit_h(dSE));
  if (acc0 < 0) return(VOICE);

  acc0 = L_mult(dSZC, 23831);           /* Q15*Q13*2 = Q29 */
  acc0 = L_mac(acc0, 4096, 31576);      /* Q12*Q16*2 = Q29 */
  acc0 = L_shr(acc0, 2);                /* Q29 -> Q27 */
  acc0 = L_add(acc0, L_deposit_h(dSE));
  if (acc0 < 0) return(VOICE);
  
  acc0 = L_mult(dSE, 32767);            /* Q11*Q15*2 = Q27 */
  acc0 = L_mac(acc0, 2048, 17367);      /* Q11*Q15*2 = Q27 */
  if (acc0 < 0) return(VOICE);
  
  /* dSLE vs SD */
  acc0 = L_mult(SD, -22400);            /* Q15*Q4*2 = Q20 */
  acc0 = L_mac(acc0, 32, 25395);        /* Q5*Q14*2 = Q20 */
  acc0 = L_mac(acc0, dSLE, 256);        /* Q11*Q8*2 = Q20 */
  if (acc0 < 0) return(VOICE);
  
  /* dSLE vs dSE */
  acc0 = L_mult(dSE, -30427);           /* Q11*Q15*2 = Q27 */
  acc0 = L_mac(acc0, 256, -29959);      /* Q8*Q18*2 = Q27 */
  acc0 = L_add(acc0, L_deposit_h(dSLE));
  if (acc0 > 0) return(VOICE);

  acc0 = L_mult(dSE, -23406);           /* Q11*Q15*2 = Q27 */
  acc0 = L_mac(acc0, 512, 28087);       /* Q19*Q17*2 = Q27 */
  acc0 = L_add(acc0, L_deposit_h(dSLE));
  if (acc0 < 0) return(VOICE);

  acc0 = L_mult(dSE, 24576);            /* Q11*Q14*2 = Q26 */
  acc0 = L_mac(acc0, 1024, 29491);      /* Q10*Q15*2 = Q26 */
  acc0 = L_mac(acc0, dSLE, 16384);      /* Q11*Q14*2 = Q26 */
  if (acc0 < 0) return(VOICE);

  return (NOISE);
}




