#include <linux/autoconf.h>
#include <asicRegs.h>
#include <rtl8651_tblAsicDrv.h>
#include "../rtl865x/rtl_glue.h"
#include "../rtl865x/rtl_types.h"
#include "../rtl865x/rtl_errno.h"
#include "../rtl865x/rtl_utils.h"
#include "pcm2.h"
#include "gpio.h"
#include "../rtl865x/voip_support.h"
#include "spi/slic.h"
#include "g729ab/mystruct.h"
#include "g729ab/g729.h"


/*
 *	Flag to switch optimized code.
 *	If defined, conservative code will be employed.
 *	If not defined, optimized code will be employed.
 */
#define _CONSERV_


/******  CONSTANT DEFINITION ******/
#define CHANNEL_NO 4
/******************************
Utility
*******************************/
#define MAX(a,b) ((a>b)?a:b)
#define MIN(a,b) ((a>b)?b:a)

static PCM_CHANNEL_t pcmChannel[CHANNEL_NO];
PCM_CHANNEL_t* pcmChan = pcmChannel;
uint32 voip_cntPktRx = 0;
uint32 voip_cntPcmInt = 0;
uint32 voip_cntPollPcmInt = 0;
uint32 voip_cntTxOk0 = 0;
uint32 voip_cntTxOk1 = 0;
uint32 voip_cntTxOk2 = 0;
uint32 voip_cntRxOk0 = 0;
uint32 voip_cntRxOk1 = 0;
uint32 voip_cntRxOk2 = 0;
uint32 voip_cntRxToVoip = 0;
uint32 voip_cntRxFrPage = 0;
uint32 voip_cntRxQEmpty = 0;
uint32 voip_cntTx = 0;
uint32 voip_testVal = 0;

int32 GPIO_PCM_enabled = 0;

/*
@ Enable 865x ASIC GPIO port B as pcm interface
*/
int32 pcm2_enableGCR_initRegs(void)
{
	int32 retval;
	int32 pin;
	/* GPIO: */
	if( GPIO_PCM_enabled == 1 )
		return SUCCESS;

	for( pin = 0; pin < 8; pin++ )
	{
		retval = _rtl865x_initGpioPin( 
			GPIO_ID(GPIO_PORT_B, pin), 
			GPIO_PERI_TYPE0, 
			GPIO_DIR_IN,
			GPIO_INT_DISABLE 
			);

		if ( retval != SUCCESS )
		{
			rtlglue_printf("_rtl865x_initGpioPin in int32 pcm2_enableGCR_initRegs(void) failed!\n");
			return FAILED;
		}
	}
	WRITE_MEM32(PCM_GCR,0x1800);/* 0->1 enable PCM							*/
	WRITE_MEM32(PCM_GCR,0x0000);/* 1->0  reset all logic to initial state	*/
	WRITE_MEM32(PCM_GCR,0x1800);/* 0->1 enable PCM							*/
	GPIO_PCM_enabled = 1;
	return SUCCESS;
}

/*
@ Enable the flag to be enable again 865x ASIC GPIO port B as pcm interface
*/
void pcm2_reset_gpio_enable_flag(void)
{
	GPIO_PCM_enabled = 0;
}

/*
	CMPE -> 0 : linear, 1 : alaw, 2 : ulaw
	channel -> 0, 1, 2, 3
	QueueSize -> Not specified
	PageSize -> Not specified
	timeSlot -> 0,1,....31
 */
int32 pcm2_enable_channel(
	uint32 channel,			/*(i)*/
	uint32 CMPE,			/*(i)*/
	uint32 RxQueueSize,		/*(i)*/
	uint32 TxQueueSize,		/*(i)*/
	uint32 *PageSize,		/*(i/o)*/
	uint32 timeSlot			/*(i)*/
)
{
	if( channel >= CHANNEL_NO ){rtlglue_printf("Channel input error\n");return FAILED;}
	if(timeSlot < 0) {timeSlot = 0 ;}
	if(timeSlot > 31){timeSlot = 31;}
	uint32 reg;
	uint32 nBsize;
	uint32 TruePage;
	uint32 tmp, tmp1;
	uint32 slot[CHANNEL_NO];

#ifdef _CONSERV_
	if ( PageSize == NULL ) {rtlglue_printf("PageSize input error\n");return FAILED;}
	/* We will strip the 2 LSBs in PageSize. */
	nBsize = *PageSize/4-1;
	TruePage = *PageSize & ~3;
	*PageSize = TruePage;
#else
	nBsize = ((*PageSize >> 1) - 2) >> 1;
	TruePage = ((nBsize << 1) + 2) << 1;
	*PageSize = TruePage;
#endif
	if(TruePage < 0)
		return FAILED;
	if( pcmChannel[channel].bEnable == TRUE )
	{
		rtlglue_printf("This channel was already setup!\n");
		return SUCCESS;
	}

	/* print and check PCM_GCR */
	rtlglue_printf("PCM GCR: %08X\n",READ_MEM32(PCM_GCR));

	/* IMR: disable pcm channnel interrupt======*/
	reg = READ_MEM32(PCM_ISR);/*Write '1' to clear*/
	WRITE_MEM32(PCM_ISR, reg | (0xFF000000 >> ( channel * 8 )));
	reg = READ_MEM32(PCM_IMR);
	WRITE_MEM32(PCM_IMR, reg & (0xFFFF0FFF >> ( channel * 4 )));/*<---Disable*/
	/* PCSCR: CMPE TE RE========================*/
	reg = READ_MEM32(PCM_SCR);
	switch(CMPE)
	{
	case PCM_ALAW:	
		tmp = 0x0F000000 >> channel * 8;
		break;
	case PCM_ULAW:	
		tmp = 0x0B000000 >> channel * 8;
		break;
	case PCM_LINEAR:	
		tmp = 0x03000000 >> channel * 8;
		break;
	default:
		tmp = 0; /* dummy */
		break;
	}
	reg = reg & ~(0xFF000000UL >> (channel * 8));
	reg = reg | tmp;
	WRITE_MEM32( PCM_SCR, reg);
	rtlglue_printf("PCM SCR: %08X\n",READ_MEM32(PCM_SCR));
	/* PTSAR:===================================*/
	reg = READ_MEM32(PCM_TSAR);

	/* read current slot assignment from PCM register. */
	for ( tmp = 0 ; tmp < sizeof(slot)/sizeof(slot[0]) ; tmp ++ )
	{
		tmp1 = reg & ( 0x1F000000UL >> (tmp * 8) );
		tmp1 = tmp1 >> (tmp * 8);
		slot[tmp] = tmp1;
	}

	/* Check if the other channel's TSA will conflict with timeSlot or not */
	for ( tmp = 0 ; tmp < sizeof(slot)/sizeof(slot[0]) ; tmp++ )
	{
		if ( slot[tmp] == timeSlot )/* Possible conflict */
		{
			if( channel == tmp )
				break; /* It is ok for self-use. */
			else 
			{
				rtlglue_printf("Should assign different time slot\n");
				return FAILED; /* It is duplicate. */
			}
		}
	}
	/* Assign channel's TSA = timeSlot */
	slot[channel] = timeSlot << (24 - channel * 8);
	reg = reg & ~(0xFF000000UL >> (channel * 8));
	reg = reg | slot[channel];
	WRITE_MEM32(PCM_TSAR, reg);
	rtlglue_printf("PCM TSA: %08X\n",READ_MEM32(PCM_TSAR));
	/* PBSIZE:==================================*/
	reg = READ_MEM32(PCM_BSIZE);
	reg = reg & ~(0xFF000000UL >> (channel * 8));
	nBsize = nBsize << (24 - channel * 8 );
	reg = reg | nBsize;
	WRITE_MEM32(PCM_BSIZE, reg);
	rtlglue_printf("PCM BSZ: %08X\n",READ_MEM32(PCM_BSIZE));
	/* CHxTXBSA==CHzRXBSA=======================*/
	/*RX Page allocation*/
	pcmChannel[channel].pRxPage0 = (uint8*) rtlglue_malloc(TruePage*2);
	rtlglue_printf("    Cached chid: %d Rx page0: %p\n",channel,pcmChannel[channel].pRxPage0);
	/*Cached to non-cached transformation*/
	pcmChannel[channel].pRxPage0 += 0x20000000;
	rtlglue_printf("Non-cached chid: %d Rx page0: %p\n",channel,pcmChannel[channel].pRxPage0);
	/*RX Page1 definition*/
	pcmChannel[channel].pRxPage1 = pcmChannel[channel].pRxPage0 + TruePage;
	pcmChannel[channel].dwRxPageSize = TruePage;
	if(pcmChannel[channel].pRxPage0==NULL)
	{
		rtlglue_printf("PCM RX page allocation failed!\n");
		return FAILED;
	}
	/*TX Page allocation*/
	pcmChannel[channel].pTxPage0 = (uint8*) rtlglue_malloc(TruePage*2);
	rtlglue_printf("    Cached chid: %d Tx page0: %p\n",channel,pcmChannel[channel].pTxPage0);
	/*Cached to non-cached transformation*/
	pcmChannel[channel].pTxPage0 += 0x20000000;
	rtlglue_printf("Non-cached chid: %d Tx page0: %p\n",channel,pcmChannel[channel].pTxPage0);
	/*TX Page1 definition*/
	pcmChannel[channel].pTxPage1 = pcmChannel[channel].pTxPage0 + TruePage;
	pcmChannel[channel].dwTxPageSize = TruePage;
	if(pcmChannel[channel].pTxPage0==NULL)
	{
		rtlglue_printf("PCM TX page allocation failed!\n");
		return FAILED;
	}
	rtlglue_printf("TruePageSize: %d bytes a page\n", TruePage);
	/*RX Page starting address writing*/
	WRITE_MEM32(PCM_CH0TXBSA + channel * 4, 
		(uint32) ((uint8*) ((uint32) pcmChannel[channel].pTxPage0| 0x20000000UL)) & 0x1fffffff);
	/*TX Page starting address writing*/
	WRITE_MEM32(PCM_CH0RXBSA + channel * 4, 
		(uint32) ((uint8*) ((uint32) pcmChannel[channel].pRxPage0| 0x20000000UL)) & 0x1fffffff);

	/* set owner PCM */
	reg = READ_MEM32(PCM_CH0TXBSA + channel * 4);
	WRITE_MEM32(PCM_CH0TXBSA + channel * 4, reg | 0x00000003);
	reg = READ_MEM32(PCM_CH0RXBSA + channel * 4);
	WRITE_MEM32(PCM_CH0RXBSA + channel * 4, reg | 0x00000003);
	rtlglue_printf("PCM RXBSA: %08X\n",READ_MEM32(PCM_CH0RXBSA + channel * 4));
	rtlglue_printf("PCM TXBSA: %08X\n",READ_MEM32(PCM_CH0TXBSA + channel * 4));
	/*Queue allocation*/
	pcmChannel[channel].dwRxQSize = RxQueueSize;
	pcmChannel[channel].dwTxQSize = TxQueueSize;
	pcmChannel[channel].pRxQStart = (uint8*) rtlglue_malloc( RxQueueSize );
	if(pcmChannel[channel].pRxQStart ==  NULL)
	{
		rtlglue_printf("PCM channel RX Queue not allocated\n");
		return FAILED;
	}
	pcmChannel[channel].pTxQStart = (uint8*) rtlglue_malloc( TxQueueSize );
	if(pcmChannel[channel].pTxQStart ==  NULL)
	{
		rtlglue_printf("PCM channel TX Queue not allocated\n");
		return FAILED;
	}
	pcmChannel[channel].dwRxQHead = pcmChannel[channel].dwRxQTail = 0;
	pcmChannel[channel].dwTxQHead = pcmChannel[channel].dwTxQTail = 0;

	pcmChannel[channel].bRxQEmpty  = TRUE;
	pcmChannel[channel].bTxQEmpty  = TRUE;
	pcmChannel[channel].bRxQFull   = FALSE;
	pcmChannel[channel].bTxQFull   = FALSE;

	/* IMR: enable pcm channnel interrupt=======*/
	reg = READ_MEM32(PCM_IMR);
	WRITE_MEM32(PCM_IMR, reg | (0xC000L >> (channel * 4)));
	rtlglue_printf("PCM IMR: %08X\n",READ_MEM32(PCM_IMR));
	/* finished */
	pcmChannel[channel].bEnable = TRUE;
	rtlglue_printf("PAGE_SIZE:%d BUFF_SIZE:%d\n",PCM_PAGE_SIZE, PCM_BUFF_SIZE);
	return SUCCESS;
}

int32 pcm2_disable_channel(uint32 channel)
{
	uint32 reg;
	if( channel >= CHANNEL_NO || channel < 0 )
		return FAILED;
	reg = READ_MEM32(PCM_IMR);
	/* disable int */
	WRITE_MEM32(PCM_IMR, reg & (0xFFFF0FFF >> ( channel * 4 )));
#ifdef _CONSERV_
	reg = READ_MEM32(PCM_ISR);
	/* clear int pending flag */
	WRITE_MEM32(PCM_IMR, reg | (0xFF000000 >> ( channel * 8 )));
	WRITE_MEM32(PCM_ISR, reg | (0xFF000000 >> ( channel * 8 )));
#else
	WRITE_MEM32(PCM_ISR, reg & (0xFF000000 >> ( channel * 8 )));
#endif
	reg = READ_MEM32(PCM_SCR);
	reg = reg & ~(0xFF000000UL >> channel * 8);
	WRITE_MEM32(PCM_SCR, reg);
	pcmChannel[channel].bRxQFull = FALSE;
	pcmChannel[channel].bTxQFull = FALSE;
	pcmChannel[channel].bRxQEmpty= FALSE;
	pcmChannel[channel].bTxQEmpty= FALSE;
	pcmChannel[channel].dwRxQHead = 0;
	pcmChannel[channel].dwRxQSize = 0;
	pcmChannel[channel].dwRxQTail = 0;
	pcmChannel[channel].dwTxQHead = 0;
	pcmChannel[channel].dwTxQSize = 0;
	pcmChannel[channel].dwTxQTail = 0;
	pcmChannel[channel].bEnable = FALSE;
	pcmChannel[channel].dwRxPageSize = 0;
	pcmChannel[channel].dwTxPageSize = 0;
	/*Noncached mem to cached mem transfer back */
	pcmChannel[channel].pRxPage0 -= 0x20000000;
	pcmChannel[channel].pTxPage0 -= 0x20000000;
	rtlglue_free(pcmChannel[channel].pRxQStart);
	rtlglue_free(pcmChannel[channel].pTxQStart);
	rtlglue_free(pcmChannel[channel].pRxPage0);
	rtlglue_free(pcmChannel[channel].pTxPage0);
	pcmChannel[channel].pRxQStart = NULL;
	pcmChannel[channel].pTxQStart = NULL;
	pcmChannel[channel].pRxPage0 = NULL;
	pcmChannel[channel].pTxPage0 = NULL;
	
	rtlglue_printf("PCM channel %d comapnder cleared\n",channel);
	rtlglue_printf("PCM channel %d page deleted\n",channel);
	rtlglue_printf("PCM channel %d queue deleted\n",channel);
	return SUCCESS;
}

/* you want to explode your queue? enlarge databuf
 G729 : 80 samples 160 bytes
 G7231: 240 samples 480 bytes
 G711 : 
 */
int32 pcm2_TxFrom_voip(uint32 channel/*(i)*/, 
 					   uint8* databuf/*(i)*/, 
 					   uint32 bufsize/*(i)*/)
{
	uint32 i;
	uint32 QSize  = pcmChannel[channel].dwTxQSize;
	uint32 QTail  = pcmChannel[channel].dwTxQTail;
	uint32 QHead  = pcmChannel[channel].dwTxQHead;
	uint8* QStart = pcmChannel[channel].pTxQStart;
	/*check validation*/
	if(channel<0 || channel>=4)		return FAILED;
	if(QStart ==NULL)				return TBLDRV_EPCM_QUEUE_UNAV;
	if(databuf==NULL)				return TBLDRV_EPCM_BUF_UNAV;

#if 1 /* Bin's thinking. */
	/*check queue status*/
	if( QTail > QHead )
	{
		if( ((QSize-QTail)+QHead) < bufsize )/*No need to -1*/
		{
			pcmChannel[channel].bTxQFull = TRUE;
			return TBLDRV_EPCM_QUEUE_FULL;
		}
	}
	else if( QTail == QHead )
	{
		if( pcmChannel[channel].bTxQEmpty == FALSE )
		{
			pcmChannel[channel].bTxQFull = TRUE;
			return TBLDRV_EPCM_QUEUE_FULL;
		}
	}
	else
	{
		if( (QHead - QTail) < bufsize )
		{
			pcmChannel[channel].bTxQFull = TRUE;
			return TBLDRV_EPCM_QUEUE_FULL;
		}
	}
#endif
#if 0 /* Lo's thinking. */
	if ( pcmChannel[channel].bTxQFull == TRUE )
		return TBLDRV_EPCM_QUEUE_FULL;
	if ( ( ( QTail + bufsize ) % QSize ) > QHead )
		return TBLDRV_EPCM_QUEUE_FULL;

	if ( ( ( QTail + bufsize ) % QSize ) == QHead ) 
		pcmChannel[channel].bTxQFull = TRUE;
	else
		pcmChannel[channel].bTxQFull = FALSE;
#endif

	/* Copying data to the tail of queue */
	for( i = 0 ; i < bufsize ; ++i )
		*(QStart + ( QTail + i ) % QSize) = *(databuf+i);

	/* revise queue status */
	pcmChannel[channel].bTxQEmpty = FALSE;
	pcmChannel[channel].dwTxQTail = ( QTail + bufsize ) % QSize;
	return SUCCESS;
}


/* Put data to pcm TX page */
int32 pcm2_TxTo_page(uint32  channel, /*(i)*/
					 uint32  page)
{
	uint32 i;
	uint32 QSize    = pcmChannel[channel].dwTxQSize;
	uint32 QTail    = pcmChannel[channel].dwTxQTail;
	uint32 QHead    = pcmChannel[channel].dwTxQHead;
	uint32 PageSize = pcmChannel[channel].dwTxPageSize;
	uint8* QStart   = pcmChannel[channel].pTxQStart;
	uint8* pPage    = pcmChannel[channel].pTxPage0;

	/* check validation */
	if(page!=0 && page!=1)			return FAILED;
	if(channel<0 || channel>=CHANNEL_NO)		return FAILED;
	if(QStart==NULL)		return TBLDRV_EPCM_QUEUE_UNAV;
	if(pPage ==NULL)		return TBLDRV_EPCM_PGAE_UNAV;
	if( page == 0 )
		pPage = pcmChannel[channel].pTxPage0;
	else
		pPage = pcmChannel[channel].pTxPage1;

#if 1 /* Bin's thinking. */
	/* check queue status */
	if( QTail > QHead )
	{
		if( (QTail-QHead) < PageSize )
		{
			for( i = 0 ; i < PageSize ; ++i )
				*( pPage + i ) = 0;
			return TBLDRV_EPCM_QUEUE_EMPTY;
		}
	}
	else if(QTail < QHead)
	{
		if( (QSize-1-QHead+QTail) < PageSize )
		{
			for( i = 0 ; i < PageSize ; ++i )
				*( pPage + i ) = 0;
			return TBLDRV_EPCM_QUEUE_EMPTY;
		}
	}
	else
	{
		if( pcmChannel[channel].bTxQEmpty == TRUE )
		{
			for( i = 0 ; i < PageSize ; ++i )
				*( pPage + i ) = 0;
			return TBLDRV_EPCM_QUEUE_EMPTY;
		}
	}
#endif
#if 0 /* Lo's thinking. */
	if ( pcmChannel[channel].bTxQEmpty == TRUE )
		return TBLDRV_EPCM_QUEUE_EMPTY;

	if ( ( ( QTail - QHead ) % QSize ) + QSize < PageSize )
		return TBLDRV_EPCM_QUEUE_EMPTY;
#endif

	/* copying data to TX page */
	for( i = 0 ; i < PageSize ; ++i )
	{
		*( pPage + i ) = *(QStart + ( QHead + i ) % QSize);
	}

	/* revise queue status */
	pcmChannel[channel].dwTxQHead = (QHead + PageSize ) % QSize;
	if( pcmChannel[channel].dwTxQHead == QTail )
	{
		pcmChannel[channel].bTxQEmpty = TRUE;
		pcmChannel[channel].bTxQFull  = FALSE;
	}
	else
	{
		pcmChannel[channel].bTxQEmpty = FALSE;
		pcmChannel[channel].bTxQFull  = FALSE;
	}
	return SUCCESS;
}

/* Put data in queue, and queue must pre-alloc memory in initial
   if page data is ready , I should take data out to queue
 */
int32 pcm2_RxFrom_page(uint32 channel/*(i)*/, uint32 page/*(i)*/)
{
	uint32 i;
	uint32 QSize    = pcmChannel[channel].dwRxQSize;
	uint32 QTail    = pcmChannel[channel].dwRxQTail;
	uint32 QHead    = pcmChannel[channel].dwRxQHead;
	uint32 PageSize = pcmChannel[channel].dwRxPageSize;
	uint8* QStart   = pcmChannel[channel].pRxQStart;
	uint8* pPage    = pcmChannel[channel].pRxPage0;

	/* check validation */
	if(page    != 0 && page    != 1)	return FAILED;
	if(channel <  0 || channel >= 4)	return FAILED;
	if(QStart==NULL)					return TBLDRV_EPCM_QUEUE_UNAV;
	if(pPage ==NULL)					return TBLDRV_EPCM_PGAE_UNAV;
	if( page == 0 )
		pPage = pcmChannel[channel].pRxPage0;
	else
		pPage = pcmChannel[channel].pRxPage1;

	/* check queue status */
	if( QTail > QHead )
	{
		if( ((QSize-1-QTail)+QHead) < PageSize )
		{
			pcmChannel[channel].bRxQFull = TRUE;
			return TBLDRV_EPCM_QUEUE_FULL;
		}
	}
	else if( QTail == QHead )
	{
		if( pcmChannel[channel].bRxQEmpty == FALSE )
		{
			pcmChannel[channel].bRxQFull = TRUE;
			return TBLDRV_EPCM_QUEUE_FULL;
		}
	}
	else
	{
		if( (QHead - QTail) < PageSize )
		{
			pcmChannel[channel].bRxQFull = TRUE;
			return TBLDRV_EPCM_QUEUE_FULL;
		}
	}

	/* copying data from page to queue */
	for( i = 0 ; i < PageSize ; ++i )
		*(QStart+(QTail+i)%QSize) = *( pPage + i );

	/* revise queue status */
	pcmChannel[channel].bRxQEmpty = FALSE;
	pcmChannel[channel].dwRxQTail = ( QTail + PageSize ) % QSize;
	return SUCCESS;
}

/* Take out data from queue, and caller must pre-alloc data memory in bufSize bytes */
int32 pcm2_RxTo_voip(uint32  channel, 
					 uint8*  databuf,/*(o)*/
					 uint32  bufsize /*(i)*/)
{
	uint32 i;
	uint32 QHead  = pcmChannel[channel].dwRxQHead;
	uint32 QTail  = pcmChannel[channel].dwRxQTail;
	uint32 QSize  = pcmChannel[channel].dwRxQSize;
	uint8* QStart = pcmChannel[channel].pRxQStart;
	if( databuf== NULL )				return FAILED;
	if( QStart == NULL )				return TBLDRV_EPCM_QUEUE_UNAV;
	if( channel < 0 || channel >= 4 )	return FAILED;

	/* check queue status */
	if( QTail > QHead )
	{
		if( (QTail-QHead) < bufsize )
		{
			for( i = 0 ; i < bufsize ; ++i )
				*( databuf + i ) = 0;
			return TBLDRV_EPCM_QUEUE_EMPTY;
		}
	}
	else if(QTail < QHead)
	{
		if( (QSize - 1 - QHead + QTail) < bufsize )
		{
			for( i = 0 ; i < bufsize ; ++i )
				*( databuf + i ) = 0;
			return TBLDRV_EPCM_QUEUE_EMPTY;
		}
	}
	else
	{
		if( pcmChannel[channel].bRxQEmpty == TRUE )
		{
			for( i = 0 ; i < bufsize ; ++i )
				*( databuf + i ) = 0;
			return TBLDRV_EPCM_QUEUE_EMPTY;
		}
	}

	/* copying data from queue to buffer */
	for( i = 0 ; i < bufsize ; i++ )
	{
		*(databuf+i) = *(QStart + (QHead + i) % QSize);
	}

	/* revise queue status */
	pcmChannel[channel].dwRxQHead = (QHead + bufsize ) % QSize;
	
	if( pcmChannel[channel].dwRxQHead == QTail )
	{
		pcmChannel[channel].bRxQEmpty = TRUE;
		pcmChannel[channel].bRxQFull  = FALSE;
	}
	else
	{
		pcmChannel[channel].bRxQEmpty = FALSE;
		pcmChannel[channel].bRxQFull  = FALSE;
	}
	return SUCCESS;
}

/*
@This function is to be registered in ROME driver
@it is called every 10 msec or pcm page interrupt
@especially used for slic channel 0
*/
int32 voip_session_pcmInt( uint32 pcm_isr )
{
#ifdef CONFIG_RTL865X_VOIP
#ifdef CONFIG_RTL865X_PCM
	int32 nRet = SUCCESS;
	int32 nRet_ToVoIP = SUCCESS;
	uint32 reg;

#ifdef __GNUC__
	__attribute__ ((aligned (8))) static uint8 pcmBuf[PCM_BUFF_SIZE];
	#ifdef CODEC729TEST
	#ifdef CONFIG_RTL865X_VOIP_G729
	__attribute__ ((aligned (8))) static unsigned char codedSpeech[11];
	#endif
	#endif
#endif
	static int ISRChannel = 0;
	static int ISRPage = 0;

	/*
	 PCM and SLIC channel must be matched 1-by-1
	 */
	for( ISRChannel = 0; ISRChannel < MIN(CHANNEL_NO,TOTAL_SLIC_CHANNEL) ; ISRChannel ++ )
	{
		if( slic_data[ISRChannel].bEnabled != TRUE || 
			slic_data[ISRChannel].bOffHook != TRUE )
			continue;
#ifdef STATE_ROLL
		if( slic_data[ISRChannel].nowState != TALKING )
			continue;
#endif
		for( ISRPage = 0 ; ISRPage < 2 ; ISRPage++ )
		{
			if( pcm_isr & ((0x20000000 >> ISRPage) >> (ISRChannel * 8)) ||
				(READ_MEM32(PCM_CH0RXBSA + ISRChannel * 4) & (1 << ISRPage))==0)
			{
				if(pcmChannel[ISRChannel].bEnable != TRUE)
					return TBLDRV_EPCM_CHANNEL_NULL;
				reg = READ_MEM32(PCM_CH0RXBSA + ISRChannel*4 );
				if((reg & (1<<ISRPage) )==0) /* CPU own */
				{
					nRet = pcm2_RxFrom_page(ISRChannel, ISRPage);
					SETBIT(reg, ISRPage);
					WRITE_MEM32(PCM_CH0RXBSA + ISRChannel*4, reg);/*set own by pcm*/
					if(nRet == SUCCESS || nRet == TBLDRV_EPCM_QUEUE_FULL)
					{
						do
						{
							nRet_ToVoIP = pcm2_RxTo_voip(ISRChannel, pcmBuf, 
								PCM_BUFF_SIZE);
							if(nRet_ToVoIP==SUCCESS)
							{
						#ifdef CODEC729TEST
							#ifdef CONFIG_RTL865X_VOIP_G729
							static unsigned int codLen = 0;
							G729ab_Encoder((Word16*)pcmBuf, codedSpeech, 1);
							switch(*(codedSpeech))
							{
							case 0x00:
								codLen = 1;
								break;
							case 0x01:
								codLen = 11;
								break;
							case 0x02:
								codLen = 3;
								break;
							}
							nRet = voip_TxRtpPkt(ISRChannel, 0, codedSpeech, 11,  TYPE_NONE);
							#endif
						#else /*Directly send raw data*/
							nRet = voip_TxRtpPkt(ISRChannel, 0, pcmBuf,
								PCM_BUFF_SIZE, TYPE_NONE);
						#endif
							}
						}while( nRet_ToVoIP == SUCCESS );
					}
				}
			}
	
			/* channel 0 TX page 0 */
			if( pcm_isr & ((0x80000000 >> ISRPage) >> (ISRChannel * 8)) || 
				(READ_MEM32(PCM_CH0TXBSA + ISRChannel * 4) & (1 << ISRPage))==0)
			{
				if(	pcmChannel[ISRChannel].bEnable != TRUE )
					return TBLDRV_EPCM_CHANNEL_NULL;
				reg = READ_MEM32(PCM_CH0TXBSA + ISRChannel*4 );
				if((reg & (1<<ISRPage) )==0)
				{
					nRet = pcm2_TxTo_page(ISRChannel, ISRPage);
					SETBIT(reg, ISRPage);
					WRITE_MEM32(PCM_CH0TXBSA + ISRChannel*4, reg);
				}
			}
		}/* End of each page */
	}/* End of each channel */
	return nRet;
#endif
#endif
}


/*
@ When recieved a packet, decode to queue and then to TX page
@ espcially used for slic channel 0
*/
int32 voip_session_pktRx
	(uint32 channel, uint32 mid, void* packet, uint32 pktLen, uint32 flags)
{
#ifdef CONFIG_RTL865X_VOIP
	extern PROSLIC_t* slic_data;
	int32 nRet = SUCCESS;
	static uint32 debugCnt = 0;
	uint32 reg;
	
#ifdef CODEC729TEST
	#ifdef CONFIG_RTL865X_VOIP_G729
	#ifdef __GNUC__
	__attribute__ ((aligned (8))) static int16 decSpeech[80];
	#else
	#error
	#endif
	#endif
#endif

	if( slic_data[channel].bEnabled != TRUE )
	{
		return SUCCESS;
	}
	
#ifdef STATE_ROLL
	static uint8 myResponse;
	if( pktLen == 1 && (*(uint8*)packet)!= I_AM_CNGCODE)
	{
		switch(*(uint8*)packet)
		{
		case I_SHOULD_RING:/*Requesting a call*/
			if( slic_data[channel].nowState == OFFHOOK )
			{
				myResponse = I_AM_BUSY;
				voip_TxRtpPkt( channel, mid, &myResponse, 1, TYPE_NONE);
			}
			if( slic_data[channel].nowState == ONHOOK ||
				slic_data[channel].nowState == DEMURE)
			{
				slic_ring_activate(channel);
				SLIC_ENTER_STATE(RING);
				myResponse = I_AM_RINGING;
				voip_TxRtpPkt( channel, mid, &myResponse, 1, TYPE_NONE);
			}
			break;
		case I_AM_RINGING:
			slic_tone_ringBack(channel);
			SLIC_ENTER_STATE(RINGTONE);
			break;
		case I_RECV_CALLING:
			slic_tone_diableOSCs(channel);
			SLIC_ENTER_STATE(TALKING);
			break;
		case I_AM_BUSY:
			slic_tone_busy(channel);
			SLIC_ENTER_STATE(BUSYTONE);
			break;
		case I_QUIT_TALK:
			slic_tone_busy(channel);
			SLIC_ENTER_STATE(BUSYTONE);
			break;
		case I_STOP_CALLING:
			rtlglue_printf("Stop Ring\n");
			slic_ring_stop(channel);
			SLIC_ENTER_STATE(ONHOOK);
			break;
		}
		return SUCCESS;
	}
#endif

	/* if no STATE_ROLL is defined, slic is directly talking */
	if( slic_data[channel].bOffHook != TRUE )/*Phone is hanging*/
	{
		return SUCCESS;
	}
	if(pcmChannel[channel].bEnable == TRUE)/*Slic is ok, Phone is taken up!*/
	{
		debugCnt ++;
#ifdef CODEC729TEST
		#ifdef CONFIG_RTL865X_VOIP_G729
		int iframe = 0;
		G729ab_Decoder(packet, decSpeech, 0, iframe);
		#endif
		nRet = pcm2_TxFrom_voip(channel, (uint8*)decSpeech, PCM_BUFF_SIZE);
#else
		nRet = pcm2_TxFrom_voip(channel, (uint8*)packet, pktLen);

#endif
		/* if the queue is full then you should have flush it
		   if it is not full, then give the TX to int
		*/
		if( nRet == TBLDRV_EPCM_QUEUE_FULL )
		{
			/* packet content delivering to CHANNEL 0 */
			reg = READ_MEM32(PCM_CH0TXBSA + channel * 4);
			if((reg & PAGE0_PCM_OWN)==0)/* page 0 now own by CPU */
			{
				if(pcm2_TxTo_page(channel, PCM_PAGE0)==SUCCESS)
					voip_cntTxOk2++;
				SETBIT(reg, 0);
				WRITE_MEM32(PCM_CH0TXBSA,reg);
			}
			reg = READ_MEM32(PCM_CH0TXBSA + channel * 4);
			if((reg & PAGE1_PCM_OWN)==0)/* page 0 now own by CPU */
			{
				if(pcm2_TxTo_page(channel, PCM_PAGE1)==SUCCESS)
					voip_cntTxOk2++;
				SETBIT(reg, 1);
				WRITE_MEM32(PCM_CH0TXBSA,reg);
			}
#ifdef CODEC729TEST
		#ifdef CONFIG_RTL865X_VOIP_G729
			nRet = pcm2_TxFrom_voip(channel, (uint8*)decSpeech, PCM_BUFF_SIZE);
		#endif
#else
			nRet = pcm2_TxFrom_voip(channel, (uint8*)packet, pktLen);
#endif
			if(nRet==SUCCESS)
				voip_cntTxOk1 ++;
		}
		else/* put to Queue */
		{
			voip_cntTxOk1 ++;
		}
	}
	else
	{
		rtlglue_printf("PCM not enabled when PktRcv!\n");
		nRet = FAILED;
	}
	return nRet;
#else 				/*ifndef CONFIG_RTL865X_VOIP*/
	rtlglue_printf("CONFIG_RTL865X_VOIP not define!\n");
#endif
}

int32 voip_session_pcmloopback(uint32 pcm_isr)
{
	uint32 reg;
	uint32 regRx;
	int32 ret = SUCCESS;
	int32 i;
	static uint8 page_buf[PCM_PAGE_SIZE];

	if(pcmChannel[PCM_CHANNEL0].bEnable != TRUE)
		return FAILED;
	if ( pcm_isr & PCM_CH0TXP0IP)
	{
		reg = READ_MEM32(PCM_CH0TXBSA);
		if((reg & PAGE0_PCM_OWN)==0) /* own by CPU */
		{
			regRx = READ_MEM32(PCM_CH0RXBSA);
			if((regRx & PAGE0_PCM_OWN)==0)
			{
				for( i = 0 ; i < PCM_PAGE_SIZE ; i++ )
					*(page_buf+i) = *(pcmChannel[0].pRxPage0+i);
				for( i = 0 ; i < PCM_PAGE_SIZE ; i++ )
					*(pcmChannel[0].pTxPage0+i) = *(page_buf+i);
				SETBIT(regRx,PCM_PAGE0);
				WRITE_MEM32(PCM_CH0RXBSA, regRx);
			}
			else if((regRx & PAGE1_PCM_OWN)==0)
			{
				for( i = 0 ; i < PCM_PAGE_SIZE ; i++ )
					*(page_buf+i) = *(pcmChannel[0].pRxPage1+i);
				for( i = 0 ; i < PCM_PAGE_SIZE ; i++ )
					*(pcmChannel[0].pTxPage0+i) = *(page_buf+i);
				SETBIT(regRx,PCM_PAGE1);
				WRITE_MEM32(PCM_CH0RXBSA, regRx);
			}
			SETBIT(reg,PCM_PAGE0);
			WRITE_MEM32(PCM_CH0TXBSA, reg);
		}
	}
	if ( pcm_isr & PCM_CH0TXP1IP)
	{
		reg = READ_MEM32(PCM_CH0TXBSA);
		if((reg & PAGE1_PCM_OWN)==0) /* own by CPU */
		{
			regRx = READ_MEM32(PCM_CH0RXBSA);
			if((regRx & PAGE0_PCM_OWN)==0)
			{
				for( i = 0 ; i < PCM_PAGE_SIZE ; i++ )
					*(page_buf+i) = *(pcmChannel[0].pRxPage0+i);
				for( i = 0 ; i < PCM_PAGE_SIZE ; i++ )
					*(pcmChannel[0].pTxPage1+i) = *(page_buf+i);
				SETBIT(regRx,PCM_PAGE0);
				WRITE_MEM32(PCM_CH0RXBSA, regRx);
			}
			else if((regRx & PAGE1_PCM_OWN)==0)
			{
				for( i = 0 ; i < PCM_PAGE_SIZE ; i++ )
					*(page_buf+i) = *(pcmChannel[0].pRxPage1+i);
				for( i = 0 ; i < PCM_PAGE_SIZE ; i++ )
					*(pcmChannel[0].pTxPage1+i) = *(page_buf+i);
				SETBIT(regRx,PCM_PAGE1);
				WRITE_MEM32(PCM_CH0RXBSA, regRx);
			}
			SETBIT(reg,PCM_PAGE1);
			WRITE_MEM32(PCM_CH0TXBSA, reg);
		}
	}
	return ret;
}

/* Configure GPIO_PortB as PCM port */
int32 pcm2_initGpioB(void)
{
	int32 nRet = SUCCESS;
	rtlglue_printf("Configure GPIO_PortB as PCM port.\n");
	_rtl865x_initGpioPin( GPIO_ID(GPIO_PORT_B, 0), GPIO_PERI_TYPE0, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* PCM_FSYNC */
	_rtl865x_initGpioPin( GPIO_ID(GPIO_PORT_B, 1), GPIO_PERI_TYPE0, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* PCM_CLK */
	_rtl865x_initGpioPin( GPIO_ID(GPIO_PORT_B, 2), GPIO_PERI_TYPE0, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* PCM_SDOUT */
	_rtl865x_initGpioPin( GPIO_ID(GPIO_PORT_B, 3), GPIO_PERI_TYPE0, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* PCM_SDIN */
	_rtl865x_initGpioPin( GPIO_ID(GPIO_PORT_B, 4), GPIO_PERI_TYPE0, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* PCM_DS0 */
	_rtl865x_initGpioPin( GPIO_ID(GPIO_PORT_B, 5), GPIO_PERI_TYPE0, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* PCM_DS1 */
	_rtl865x_initGpioPin( GPIO_ID(GPIO_PORT_B, 6), GPIO_PERI_TYPE0, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* PCM_DS2 */
	_rtl865x_initGpioPin( GPIO_ID(GPIO_PORT_B, 7), GPIO_PERI_TYPE0, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* PCM_DS3 */
	return nRet;
}

/* for testing purpose */
void pcm2_PlaySpeechData(short* data, uint32 dataLen, uint32 chid)
{
	uint32 i;
	uint32 RegValue;
	uint32 ProgressLen = 0;
		
	if(pcmChannel[chid].bEnable == TRUE)
	{
		RegValue = READ_MEM32(PCM_CH0TXBSA + chid*4);
		CLRBIT(RegValue, PCM_PAGE0);
		CLRBIT(RegValue, PCM_PAGE1);
		WRITE_MEM32(PCM_CH0TXBSA + chid*4, RegValue);
		do
		{
			RegValue = READ_MEM32(PCM_CH0TXBSA + chid*4);
			if((RegValue & PAGE0_PCM_OWN)==0)
			{
				/*Copy data*/
				for( i=0 ; i < pcmChannel[chid].dwTxPageSize ; i++ )
				{
					*(pcmChannel[chid].pTxPage0 + i) =
						*((uint8*)data + (ProgressLen * 2) + i);
				}
				ProgressLen += pcmChannel[chid].dwTxPageSize/2;
				SETBIT(RegValue, PCM_PAGE0);
				WRITE_MEM32(PCM_CH0TXBSA + chid*4, RegValue);
			}
			RegValue = READ_MEM32(PCM_CH0TXBSA + chid*4);
			if((RegValue & PAGE1_PCM_OWN)==0)
			{
				/*Copy data*/
				for( i=0 ; i < pcmChannel[chid].dwTxPageSize ; i++ )
				{
					*(pcmChannel[chid].pTxPage1 + i) =
						*((uint8*)data + (ProgressLen * 2) + i);
				}
				ProgressLen += pcmChannel[chid].dwTxPageSize/2;
				SETBIT(RegValue, PCM_PAGE1);
				WRITE_MEM32(PCM_CH0TXBSA + chid*4, RegValue);
			}

		}while(ProgressLen < dataLen);
		rtlglue_printf("PCM channel %d playing finished!\n",chid);
		rtlglue_printf("Playing %d:%d samples!\n",ProgressLen, dataLen);
	}
	else
	{
		rtlglue_printf("PCM channel %d is not enabled\n!",chid);
	}
}


