/* 
 * Program : SLIC Driver
 * Abstract : 
 * Author : Shao Huang Pin (brian_bin@realtek.com.tw)
 * $Id: slic.c,v 1.1.1.1 2006/07/26 05:38:36 blue Exp $
 */
#include <linux/autoconf.h>
#include <asicRegs.h>
#include <rtl8651_tblAsicDrv.h>
#include "../../rtl865x/voip_support.h"
#include "rtl_glue.h"
#include "rtl_types.h"
#include "rtl_errno.h"
#include "gpio.h"
#include "spi.h"
#include "slic.h"
#include "../pcm2.h"

static rtl8651_spi_dev_t dev[TOTAL_SLIC_CHANNEL];
rtl8651_spi_dev_t* spiDev = dev;
static PROSLIC_t chipData[TOTAL_SLIC_CHANNEL];
PROSLIC_t* slic_data = chipData;

/*
  The following Array contains: Reg# Label Initial Value
*/
indirectRegister  indirectRegisters[] =
{
	{	0,	"DTMF_ROW_0_PEAK",		0x55C2	},
	{	1,	"DTMF_ROW_1_PEAK",		0x51E6	},
	{	2,	"DTMF_ROW2_PEAK",		0x4B85	},
	{	3,	"DTMF_ROW3_PEAK",		0x4937	},
	{	4,	"DTMF_COL1_PEAK",		0x3333	},
	{	5,	"DTMF_FWD_TWIST",		0x0202	},
	{	6,	"DTMF_RVS_TWIST",		0x0202	},
	{	7,	"DTMF_ROW_RATIO",		0x0198	},
	{	8,	"DTMF_COL_RATIO",		0x0198	},
	{	9,	"DTMF_ROW_2ND_ARM",		0x0611	},
	{	10,	"DTMF_COL_2ND_ARM",		0x0202	},
	{	11,	"DTMF_PWR_MIN_",		0x00E5	},
	{	12,	"DTMF_OT_LIM_TRES",		0x0A1C	},
	{	13,	"OSC1_COEF",			0x7b30	},
	{	14,	"OSC1X",				0x0063	},
	{	15,	"OSC1Y",				0x0000	},
	{	16,	"OSC2_COEF",			0x7870	},
	{	17,	"OSC2X",				0x007d	},
	{	18,	"OSC2Y",				0x0000	},
	{	19,	"RING_V_OFF",			0x0000	},
	{	20,	"RING_OSC",				0x7EF0	},
	{	21,	"RING_X",				0x0160	},
	{	22,	"RING_Y",				0x0000	},
	{	23,	"PULSE_ENVEL",			0x2000	},
	{	24,	"PULSE_X",				0x2000	},
	{	25,	"PULSE_Y",				0x0000	},
	{	26,	"RECV_DIGITAL_GAIN",	0x4000	},
	{	27,	"XMIT_DIGITAL_GAIN",	0x4000	},
	{	28,	"LOOP_CLOSE_TRES",		0x1000	},
	{	29,	"RING_TRIP_TRES",		0x3600	},
	{	30,	"COMMON_MIN_TRES",		0x1000	},
	{	31,	"COMMON_MAX_TRES",		0x0080	},
	{	32,	"PWR_ALARM_Q1Q2",		0x07c0  },
	{	33,	"PWR_ALARM_Q3Q4",		0x376f	},
	{	34,	"PWR_ALARM_Q5Q6",		0x1B80	},
	{	35,	"LOOP_CLSRE_FlTER",		0x8000	},
	{	36,	"RING_TRIP_FILTER",		0x0320	},
	{	37,	"TERM_LP_POLE_Q1Q2",	0x008c  },
	{	38,	"TERM_LP_POLE_Q3Q4",	0x008c  },
	{	39,	"TERM_LP_POLE_Q5Q6",	0x0010	},
	{	40,	"CM_BIAS_RINGING",		0x0200	},
	{	41,	"DCDC_MIN_V",			0x0C00	},
	{	43,	"LOOP_CLOSE_TRES Low",	0x1000	},
	{	99,	"FSK 0 FREQ PARAM",		0x00DA	},
	{	100,"FSK 0 AMPL PARAM",		0x6B60	},
	{	101,"FSK 1 FREQ PARAM",		0x0074	},
	{	102,"FSK 1 AMPl PARAM",		0x79C0	},
	{	103,"FSK 0to1 SCALER",		0X1120	},
	{	104,"FSK 1to0 SCALER",		0x3BE0	},
	{	 97,"RCV_FLTR",				0},
	{     0,"",                     0},
};
/****************** COMMON TELEPHONEY TONES*********************************/
/* DialTone: when you take up the phone */
tone_struct DialTone = {  /* OSC1= 350 Hz OSC2= 440 Hz .0975 Volts -18 dBm */
	{0x7b30,0x0063,0,0,0,0,0},{0x7870,0x007d,0,0,0,0,0}
};
/* when caller does not dial properly */
tone_struct ReorderTone = {	/* OSC1= 480 Hz OSC2 = 620 .0975 Volts -18 dBm */
	{0x7700,0x0089,0,0x09,0x60,0x06,0x40},{0x7120,0x00b2,0,0x09,0x60,0x06,0x40}
};
/* when callee is busy */
tone_struct CongestionTone = { /* OSC1= 480 Hz OSC2 = 620 .0975 Volts -18 dBM */
	{0x7700,0x0089,0,0x06,0x40,0x09,0x60},{0x7120,0x00b2,0,0x06,0x40,0x09,0x60}
};
tone_struct RingbackPBX = {	/* OSC1 = 440 Hz OSC2= 480 .0975 Volts -18 dBM */
	{0x7870,0x007d,0,0x1f,0x40,0x5d,0xc0},{0x7700,0x0089,0,0x1f,0x40,0x5d,0xc0}
};
tone_struct RingbackNormal = { /* OSC1 = 440 Hz OSC2 = 480 .0975 Volts -18 dBm */
	{0x7870,0x007d,0,0x3e,0x80,0x7d,0x00},{0x7700,0x0089,0,0x3e,0x80,0x7d,0x00}
};
tone_struct BusySignal = { /* OSC1= 480  OSC2 = 620 .0975 Voltz -18 dBm 8*/
	{0x7700,0x0089,0,0x0f,0xa0,0x0f,0xa0},{0x7120,0x00b2,0,0x0f,0xa0,0x0f,0xa0}
};
tone_struct RingbackJapan = { /* OSC1 = 400 Hz OSC2 = 435 .0975 Volts -18 dBm */
	{0x79c0,0x00e9,0,0x1f,0x40,0x3e,0x80},{0x7940,0x00f2,0,0x1f,0x40,0x3e,0x80}
};
tone_struct BusyJapan = { /* OSC1 = 400 Hz OSC2 = 435 .0975 Volts -18 dBm */
	{0x79c0,0x00e9,0,0x0f,0xa0,0x0f,0xa0},{0,0,0,0,0,0,0}
};
tone_struct JapanDialTone = { /* OSC1 = 400 Hz OSC2 = 435 .0975 Volts -18 dBm */
	{0x79c0,0x00e9,0,0,0,0,0},{0,0,0,0,0,0,0}
};

int32 current_channel = 0;

/*
INFO: Freq = 400.000000    for n=1 or n=2
 OSCn = 0x79c0, OSCnX = 0x0071, OSCnY = 0x0000

C:\>tone 420 .0975

INFO: Freq = 420.000000    for n=1 or n=2
 OSCn = 0x7920, OSCnX = 0x0077, OSCnY = 0x0000

C:\>tone 415 .0975

INFO: Freq = 415.000000    for n=1 or n=2
 OSCn = 0x7940, OSCnX = 0x0076, OSCnY = 0x0000

C:\>tone 415 .2

INFO: Freq = 415.000000    for n=1 or n=2
 OSCn = 0x7940, OSCnX = 0x00f2, OSCnY = 0x0000

C:\>tone 400 .2

INFO: Freq = 400.000000    for n=1 or n=2
 OSCn = 0x79c0, OSCnX = 0x00e9, OSCnY = 0x0000
 */
/*
@ ZONE: register read/write, the functions in the following is
@ only used when slic generates kinds of tones
*/

uint8 _slic_read_DirectReg(uint32 reg, int channel)
{
	uint8 buf;
	buf = reg | 0x80;
	_rtl865x_spi_rawWrite( &dev[channel], &buf, 8);
	_rtl865x_spi_rawRead(  &dev[channel], &buf, 8);
	return buf;
}

void _slic_write_DirectReg(uint32 reg, uint8 data, int channel)
{
	uint8 buf;
	buf = reg;
	_rtl865x_spi_rawWrite( &dev[channel], &buf, 8 );
	buf = data;
	_rtl865x_spi_rawWrite( &dev[channel], &buf, 8 );
}

uint16 _slic_read_IndirectReg(uint32 reg, int channel)
{
	uint8 data8 = 0;
	uint16 data = 0;
	while(readDirectReg(31, channel));
	writeDirectReg(30, reg, channel);
	while(readDirectReg(31, channel));
	data8 = readDirectReg(28, channel);
	data  = data8;
	data8 = 0;
	data8 = readDirectReg(29, channel);
	data  = data | data8 << 8;
	return data;
}

void _slic_write_IndirectReg(uint32 reg, uint16 data, int channel)
{
	while(readDirectReg(31, channel));
	writeDirectReg(28,(uint8)(data & 0xFF), channel);
	writeDirectReg(29,(uint8)((data & 0xFF00)>>8), channel);
	writeDirectReg(30, reg, channel);
}


/*
@ ZONE: Initialization, the functions in the following is
@ only used in the start-up of slic communications
*/

/* INIT: FIRST, GPIO pins */
void slic_init_GPIO(uint32 channel)
{
	uint32 i;
	uint32 resetDelay = 50000000;

	/* SLIC 0 initialization */
	if(channel ==0)
	{
		/* Relay is used as when we use the phone and no electricity */
		rtlglue_printf("pin RELAY initialize ...");
		_rtl865x_initGpioPin(PIN_RELAY,	 GPIO_PERI_GPIO, GPIO_DIR_OUT, GPIO_INT_DISABLE); /* Relay */

		_rtl865x_initGpioPin(PIN_CS1,    GPIO_PERI_GPIO, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* CS_ */
		_rtl865x_initGpioPin(PIN_RESET1, GPIO_PERI_GPIO, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* RESET_ */

		for( i = resetDelay; i > 0; i-- );       /*50MHz ~= 250ms*/
		_rtl865x_setGpioDataBit( PIN_CS1,    0 );/* CS_ preset to high state*/
		_rtl865x_setGpioDataBit( PIN_RESET1, 1 );/* RESET_ */
		for( i = resetDelay; i > 0; i-- );
		_rtl865x_setGpioDataBit( PIN_RESET1, 0 );/* RESET_ Hold RESET_ low PCLK and FS is present and stable*/
		for( i = resetDelay; i > 0; i-- );
		_rtl865x_setGpioDataBit( PIN_RESET1, 1 );/* RESET_ */
		for( i = resetDelay; i > 0; i-- );/*Hold 2 ms or longer*/
		_rtl865x_spi_init( &dev[0],
			GPIO_ID(GPIO_PORT_E, 4),/*SLIC0: SPI CLK*/
			GPIO_ID(GPIO_PORT_E, 3),/*SLIC0: CS*/
			GPIO_ID(GPIO_PORT_E, 5),/*SLIC0: Data out from 865x*/
			GPIO_ID(GPIO_PORT_E, 6),/*SLIC0: Data int to   865x*/
			1000000 );
		
		_rtl865x_setGpioDataBit(PIN_CS1,   1);/* CS_ preset to high state(no selection)*/
		_rtl865x_setGpioDataBit(PIN_RELAY, 1);
		rtlglue_printf("PIN_RELAY(GPIO_E0) set to 1\n");
	}
	
	/* SLIC 1 initialization */
	if(channel==1)
	{
		_rtl865x_initGpioPin(PIN_RESET1, GPIO_PERI_GPIO, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* RESET_ */

		_rtl865x_initGpioPin(PIN_CS2,    GPIO_PERI_GPIO, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* CS_ */
		_rtl865x_initGpioPin(PIN_RESET2, GPIO_PERI_GPIO, GPIO_DIR_OUT, GPIO_INT_DISABLE );/* RESET_ */
		for( i = resetDelay; i > 0; i-- );       /* 50MHz ~= 250ms*/
		_rtl865x_setGpioDataBit( PIN_CS2,    0 );/* CS_ preset to high state*/
		_rtl865x_setGpioDataBit( PIN_RESET2, 1 );/* RESET_ */
		for( i = resetDelay; i > 0; i-- );
		_rtl865x_setGpioDataBit( PIN_RESET2, 0 );/* RESET_ Hold RESET_ low PCLK and FS is present and stable*/
		for( i = resetDelay; i > 0; i-- );
		_rtl865x_setGpioDataBit( PIN_RESET2, 1 );/* RESET_ */
		for( i = resetDelay; i > 0; i-- );
		_rtl865x_spi_init( &dev[1],
			GPIO_ID(GPIO_PORT_E, 4),/*SLIC1: SPI CLK*/
			GPIO_ID(GPIO_PORT_C, 0),/*SLIC1: CS*/
			GPIO_ID(GPIO_PORT_E, 5),/*SLIC1: Data out from 865x*/
			GPIO_ID(GPIO_PORT_E, 6),/*SLIC1: Data in  to   865x*/
			1000000 );
		_rtl865x_setGpioDataBit( PIN_CS2, 1 );/* CS_ preset to high state(no selection)*/
	}
}

int32 slic_init_Test(int channel)
{
	int32 ret = SUCCESS;
	if( readDirectReg(8,  channel) != 0x02 ||
		readDirectReg(11, channel) != 0x33 ||
		readDirectReg(64, channel) != 0x00 )
	{
		ret = FAILED;
	}
	return ret;
}

/*SECOND*/
int32 slic_init(uint32 channel)
{
	uint16 temp[5];
	int32 ret = SUCCESS;
	if(channel < 0 || channel > 1 )
	{
		rtlglue_printf("Given unavailable slic channel\n");
		return FAILED;
	}

	/*Test*/
	ret = slic_init_Test(channel);
	slic_read_PCMCLK(channel);
	if( ret == FAILED )
	{
		rtlglue_printf("SLIC not communicating\n");
		rtlglue_printf("Possible reason:\n");
		rtlglue_printf("	1	PCM:PCLK not providing clocks\n");
		rtlglue_printf("	2	Slic already init\n");
		return ret;
	}
	/* identify type & version */
	chipData[channel].version = 0xf & readDirectReg(0, channel);
	chipData[channel].type   = (0x30 & readDirectReg(0, channel)) >> 4;
	rtlglue_printf("SLIC %d, type : %d\t",channel,chipData[channel].type);
	rtlglue_printf("version : %c\n",'A'-1 + chipData[channel].version);

	/* initial indirect registers */
	slic_init_IndirectReg(channel);

	/* settings */
	writeDirectReg(8, 0, channel);/* audio path no loop back */
	writeDirectReg(108, 0xEB, channel);/* enhancement enable register */
	if (chipData[channel].type ==0) /* Si3210 only */
	{
		writeDirectReg(67,0x17, channel);/* Automatic control setting, why bit[3] is set to 0? which means no automatic switching to low battery in off-hook state */
		writeDirectReg(66,0x1, channel);/* VBAT tracks VRING */
	}
	if (chipData[channel].version <=2 )/*REVISION B and before*/
		writeDirectReg(73,4, channel);/* set common mode voltage to 6 volts */

	
	/*  Flush ProSLIC digital filters by setting Coefficients to clear */
	/*  Save OSC control, Atimer, Itimer of OSC1 OSC2 */
	temp[0] = readIndirectReg(35, channel);	writeIndirectReg(35, 0x8000, channel);
	temp[1] = readIndirectReg(36, channel);	writeIndirectReg(36, 0x8000, channel);
	temp[2] = readIndirectReg(37, channel);	writeIndirectReg(37, 0x8000, channel);
	temp[3] = readIndirectReg(38, channel);	writeIndirectReg(38, 0x8000, channel);
	temp[4] = readIndirectReg(39, channel);	writeIndirectReg(39, 0x8000, channel);

	/* Do Flush durring powerUp and calibrate */
	if (chipData[channel].type == 0 || chipData[channel].type==3) /*  Si3210 */
	{
		ret = slic_init_powerUp(channel);  /* Turn on the DC-DC converter and verify voltage. */
		if(ret!=SUCCESS)
		{
			rtlglue_printf("PowerUp Failed\t");
			return ret;
		}
		ret = slic_init_powerLeakTest(channel); /* Check for power leaks */
		if(ret!=SUCCESS)
		{
			rtlglue_printf("Power leak\t");
			return ret;
		}
		ret = slic_init_powerUp(channel); /* Turn on the DC-DC converter again */
		if(ret!=SUCCESS)
		{
			rtlglue_printf("PowerUp Failed\n");
			return ret;
		}
	}
	slic_init_DirectReg(channel);
	slic_init_calibrate(channel);
	writeIndirectReg(35, temp[0], channel);
	writeIndirectReg(36, temp[1], channel);
	writeIndirectReg(37, temp[2], channel);
	writeIndirectReg(38, temp[3], channel);
	writeIndirectReg(39, temp[4], channel);
	clearInterrupts(channel);/*maybe redundant ? */
	goActive(channel);
	slic_read_PCMCLK(channel);
	rtlglue_printf("ProSLIC device %d initialized\n", channel);
	return SUCCESS;
}

void clearInterrupts(int channel)
{
	writeDirectReg(	18	,	INIT_DR18	, channel);
	writeDirectReg(	19	,	INIT_DR19	, channel);
	writeDirectReg(	20	,	INIT_DR20	, channel);
}


/* indirect initials are defined according to silicon lab datasheet */
void slic_init_IndirectReg(int channel)
{
	unsigned char i=0;
	while (indirectRegisters[i].initial || 
		   indirectRegisters[i].address ) 
	{
		writeIndirectReg(
			indirectRegisters[i].address,
			indirectRegisters[i].initial,
			channel
			);
		i++;
	}
}

void slic_init_DirectReg(int channel)
{
	writeDirectReg(0,	INIT_DR0	, channel);/*0X00	Serial Interface */
	writeDirectReg(1,	INIT_DR1	, channel);/*0X28	PCM Mode */
	writeDirectReg(2,	INIT_DR2	, channel);/*0X00	PCM TX Clock Slot Low Byte (1 PCLK cycle/LSB) */
	writeDirectReg(3,	INIT_DR3	, channel);/*0x00	PCM TX Clock Slot High Byte */
	writeDirectReg(4,	INIT_DR4	, channel);/*0x00	PCM RX Clock Slot Low Byte (1 PCLK cycle/LSB) */
	writeDirectReg(5,	INIT_DR5	, channel);/*0x00	PCM RX Clock Slot High Byte */
	writeDirectReg(8,	INIT_DR8	, channel);/*0X00	Loopbacks (digital loopback default) */
	writeDirectReg(9,	INIT_DR9	, channel);/*0x00	Transmit and receive path gain and control */
	writeDirectReg(10,	INIT_DR10	, channel);/*0X28	Initialization Two-wire impedance (600  and enabled) */
	writeDirectReg(11,	INIT_DR11	, channel);/*0x33	Transhybrid Balance/Four-wire Return Loss */
	writeDirectReg(18,	INIT_DR18	, channel);/*0xff	Normal Oper. Interrupt Register 1 (clear with 0xFF) */
	writeDirectReg(19,	INIT_DR19	, channel);/*0xff	Normal Oper. Interrupt Register 2 (clear with 0xFF) */
	writeDirectReg(20,	INIT_DR20	, channel);/*0xff	Normal Oper. Interrupt Register 3 (clear with 0xFF) */
	writeDirectReg(21,	INIT_DR21	, channel);/*0xff	Interrupt Mask 1 */
	writeDirectReg(22,	INIT_DR22	, channel);/*0xff	Initialization Interrupt Mask 2 */
	writeDirectReg(23,	INIT_DR23	, channel);/*0xff	 Initialization Interrupt Mask 3 */
	writeDirectReg(32,	INIT_DR32	, channel);/*0x00	Oper. Oscillator 1 Controltone generation */
	writeDirectReg(33,	INIT_DR33	, channel);/*0x00	Oper. Oscillator 2 Controltone generation */
	writeDirectReg(34,	INIT_DR34	, channel);/*0X18	34 0x22 0x00 Initialization Ringing Oscillator Control */
	writeDirectReg(35,	INIT_DR35	, channel);/*0x00	Oper. Pulse Metering Oscillator Control */
	writeDirectReg(36,	INIT_DR36	, channel);/*0x00	36 0x24 0x00 Initialization OSC1 Active Low Byte (125 s/LSB) */
	writeDirectReg(37,	INIT_DR37	, channel);/*0x00	37 0x25 0x00 Initialization OSC1 Active High Byte (125 s/LSB) */
	writeDirectReg(38,	INIT_DR38	, channel);/*0x00	38 0x26 0x00 Initialization OSC1 Inactive Low Byte (125 s/LSB) */
	writeDirectReg(39,	INIT_DR39	, channel);/*0x00	39 0x27 0x00 Initialization OSC1 Inactive High Byte (125 s/LSB) */
	writeDirectReg(40,	INIT_DR40	, channel);/*0x00	40 0x28 0x00 Initialization OSC2 Active Low Byte (125 s/LSB) */
	writeDirectReg(41,	INIT_DR41	, channel);/*0x00	41 0x29 0x00 Initialization OSC2 Active High Byte (125 s/LSB) */
	writeDirectReg(42,	INIT_DR42	, channel);/*0x00	42 0x2A 0x00 Initialization OSC2 Inactive Low Byte (125 s/LSB) */
	writeDirectReg(43,	INIT_DR43	, channel);/*0x00	43 0x2B 0x00 Initialization OSC2 Inactive High Byte (125 s/LSB) */
	writeDirectReg(44,	INIT_DR44	, channel);/*0x00	44 0x2C 0x00 Initialization Pulse Metering Active Low Byte (125 s/LSB) */
	writeDirectReg(45,	INIT_DR45	, channel);/*0x00	45 0x2D 0x00 Initialization Pulse Metering Active High Byte (125 s/LSB) */
	writeDirectReg(46,	INIT_DR46	, channel);/*0x00	46 0x2E 0x00 Initialization Pulse Metering Inactive Low Byte (125 s/LSB) */
	writeDirectReg(47,	INIT_DR47	, channel);/*0x00	47 0x2F 0x00 Initialization Pulse Metering Inactive High Byte (125 s/LSB) */
	writeDirectReg(48,	INIT_DR48	, channel);/*0X80	48 0x30 0x00 0x80 Initialization Ringing Osc. Active Timer Low Byte (2 s,125 s/LSB) */
	writeDirectReg(49,	INIT_DR49	, channel);/*0X3E	49 0x31 0x00 0x3E Initialization Ringing Osc. Active Timer High Byte (2 s,125 s/LSB) */
	writeDirectReg(50,	INIT_DR50	, channel);/*0X00	50 0x32 0x00 0x00 Initialization Ringing Osc. Inactive Timer Low Byte (4 s, 125 s/LSB) */
	writeDirectReg(51,	INIT_DR51	, channel);/*0X7D	51 0x33 0x00 0x7D Initialization Ringing Osc. Inactive Timer High Byte (4 s, 125 s/LSB) */
	writeDirectReg(52,	INIT_DR52	, channel);/*0X00	52 0x34 0x00 Normal Oper. FSK Data Bit */
	writeDirectReg(63,	INIT_DR63	, channel);/*0X54	63 0x3F 0x54 Initialization Ringing Mode Loop Closure Debounce Interval */
	writeDirectReg(64,	INIT_DR64	, channel);/*0x00	64 0x40 0x00 Normal Oper. Mode Byteprimary control */
	writeDirectReg(65,	INIT_DR65	, channel);/*0X61	65 0x41 0x61 Initialization External Bipolar Transistor Settings */
	writeDirectReg(66,	INIT_DR66	, channel);/*0X03	66 0x42 0x03 Initialization Battery Control */
	writeDirectReg(67,	INIT_DR67	, channel);/*0X1F	67 0x43 0x1F Initialization Automatic/Manual Control */
	writeDirectReg(69,	INIT_DR69	, channel);/*0X0C	69 0x45 0x0A 0x0C Initialization Loop Closure Debounce Interval (1.25 ms/LSB) */
	writeDirectReg(70,	INIT_DR70	, channel);/*0X0A	70 0x46 0x0A Initialization Ring Trip Debounce Interval (1.25 ms/LSB) */
	writeDirectReg(71,	INIT_DR71	, channel);/*0X01	71 0x47 0x00 0x01 Initialization Off-Hook Loop Current Limit (20 mA + 3 mA/LSB) */
	writeDirectReg(72,	INIT_DR72	, channel);/*0X20	72 0x48 0x20 Initialization On-Hook Voltage (open circuit voltage) = 48 V(1.5 V/LSB) */
	writeDirectReg(73,	INIT_DR73	, channel);/*0X02	73 0x49 0x02 Initialization Common Mode VoltageVCM = 3 V(1.5 V/LSB) */
	writeDirectReg(74,	INIT_DR74	, channel);/*0X32	74 0x4A 0x32 Initialization VBATH (ringing) = 75 V (1.5 V/LSB) */
	writeDirectReg(75,	INIT_DR75	, channel);/*0X10	75 0x4B 0x10 Initialization VBATL (off-hook) = 24 V (TRACK = 0)(1.5 V/LSB) */
	if (chipData[current_channel].type != 3)
		writeDirectReg(92,INIT_DR92	, channel);/*0x7f	92 0x5C 0xFF 7F Initialization DCDC Converter PWM Period (61.035 ns/LSB) */
	else
		writeDirectReg(92,INIT_SI3210M_DR92,channel);/*0x60	92 0x5C 0xFF 7F Initialization DCDC Converter PWM Period (61.035 ns/LSB) */

	writeDirectReg(93,	INIT_DR93	, channel);/*0x14	93 0x5D 0x14 0x19 Initialization DCDC Converter Min. Off Time (61.035 ns/LSB) */
	writeDirectReg(96,	INIT_DR96	, channel);/*0x00	96 0x60 0x1F Initialization Calibration Control Register 1(written second and starts calibration) */
	writeDirectReg(97,	INIT_DR97	, channel);/*0X1F	97 0x61 0x1F Initialization Calibration Control Register 2(written before Register 96) */
	writeDirectReg(98,	INIT_DR98	, channel);/*0X10	98 0x62 0x10 Informative Calibration result (see data sheet) */
	writeDirectReg(99,	INIT_DR99	, channel);/*0X10	99 0x63 0x10 Informative Calibration result (see data sheet) */
	writeDirectReg(100,	INIT_DR100	, channel);/*0X11	100 0x64 0x11 Informative Calibration result (see data sheet) */
	writeDirectReg(101,	INIT_DR101	, channel);/*0X11	101 0x65 0x11 Informative Calibration result (see data sheet) */
	writeDirectReg(102,	INIT_DR102	, channel);/*0x08	102 0x66 0x08 Informative Calibration result (see data sheet) */
	writeDirectReg(103,	INIT_DR103	, channel);/*0x88	103 0x67 0x88 Informative Calibration result (see data sheet) */
	writeDirectReg(104,	INIT_DR104	, channel);/*0x00	104 0x68 0x00 Informative Calibration result (see data sheet) */
	writeDirectReg(105,	INIT_DR105	, channel);/*0x00	105 0x69 0x00 Informative Calibration result (see data sheet) */
	writeDirectReg(106,	INIT_DR106	, channel);/*0x20	106 0x6A 0x20 Informative Calibration result (see data sheet) */
	writeDirectReg(107,	INIT_DR107	, channel);/*0x08	107 0x6B 0x08 Informative Calibration result (see data sheet) */
	writeDirectReg(108,	INIT_DR108	, channel);/*0xEB	108 0x63 0x00 0xEB Initialization Feature enhancement register */
}

int32 slic_init_powerUp(int channel)
{
	int32 i;
	extern unsigned long jiffies;
	unsigned long initialTime, powerTime=0;
	if (chipData[channel].type == 3)  /* M version correction */
	{
		writeDirectReg(92,INIT_SI3210M_DR92, channel);/* M version */
		writeDirectReg(93,INIT_SI3210M_DR93, channel);/* M version */
	}
	else	
	{
		/* set the period of the DC-DC converter to 1/64 kHz  START OUT SLOW*/
		writeDirectReg(93, 0x12, channel);
		writeDirectReg(92, 0x7f, channel);
	}

	initialTime = jiffies;
	writeDirectReg(14, 0, channel); /* Engage the DC-DC converter */
	i = 0;
	while (readDirectReg(82, channel) < 0xC0)/* 0xC0 is 72 volt */
	{
		if( i++ > 2000 )
		{
			if( i == 2001 )
			{
				rtlglue_printf("TimeUp for powerup!\n");
			}
			if( (jiffies - initialTime) > 50 )/* 0.5 seconds */
			{
				rtlglue_printf("Exception: timeout for powerup! ::%02X\n", readDirectReg(82, channel));
				return FAILED;
			}
		}
	}
	powerTime = jiffies - initialTime;
	if ( powerTime > 50)/* 0.5 seconds */
	{
		rtlglue_printf("\nWarning Power Up took %lu milliseconds.\t",powerTime);
		rtlglue_printf("more than 0.5 seconds\n");
	}

	writeDirectReg(93, 0x80, channel);  /* DC-DC Calibration  */
	rtlglue_printf("Wait for DC-DC Calibration to complete\n");

	/* Wait for DC-DC Calibration to complete */
	while( 0x80 & readDirectReg(93, channel));
	
	rtlglue_printf("power up complete!\n");
	return SUCCESS;
}
 
int32 slic_init_powerLeakTest(int channel)
{
	unsigned long Leaktime;
	extern unsigned long volatile jiffies;
	unsigned char vBat;
	writeDirectReg(64, 0x00, channel);
	writeDirectReg(14, 0x10, channel);/*Force DC-DC circuitry off*/
	
	Leaktime = jiffies;
	while((jiffies - Leaktime)<100);/* 1 second */
	
	vBat = readDirectReg(82, channel);
	if ( vBat < 0x4)  /* 4x1.5 = 6 volts */
	{
		rtlglue_printf("Exception: power is leaking!\n");
		return FAILED;
	}
	rtlglue_printf("Power not leaking!\n");
	return SUCCESS;
}

void slic_init_calibrate(int channel)
{
	extern unsigned long volatile jiffies;
	unsigned long Caltime;
	unsigned char x,y,i=0,progress=0;
	/* Do Flush durring powerUp and calibrate */
	writeDirectReg(21,DISABLE_ALL_DR21, channel);/*  Disable all interupts in DR21 */
	writeDirectReg(22,DISABLE_ALL_DR22, channel);/*	Disable all interupts in DR21 */
	writeDirectReg(23,DISABLE_ALL_DR23, channel);/*	Disabel all interupts in DR21 */
	writeDirectReg(64,OPEN_DR64, channel);
	/*(0x18)Calibrations without the ADC and DAC offset and 
	  without common mode calibration.
	 */
	writeDirectReg(97,STANDARD_CAL_DR97, channel);
	
	/* Calibrate common mode and differential DAC mode DAC + ILIM */
	writeDirectReg(96,STANDARD_CAL_DR96, channel);
	while (readDirectReg(96, channel) != 0 );	
	/*
		Initialized DR 98 and 99 to get consistant results.
 		98 and 99 are the results registers and the search 
 		should have same intial conditions.
		The following is the manual gain mismatch calibration
		This is also available as a function
	 */
	Caltime= jiffies;
	while((jiffies-Caltime)<1);/* 0.01 seconds */
	writeIndirectReg(88,0, channel);
	writeIndirectReg(89,0, channel);
	writeIndirectReg(90,0, channel);
	writeIndirectReg(91,0, channel);
	writeIndirectReg(92,0, channel);
	writeIndirectReg(93,0, channel);
	/* This is necessary if the calibration occurs other than at reset time */
	writeDirectReg(98,0x10, channel);
	writeDirectReg(99,0x10, channel);
	for ( i=0x1f; i>0; i--)
	{
		writeDirectReg(98,i, channel);
		Caltime = jiffies;
		while((jiffies-Caltime)<4);/* 0.04 seconds */
		if((readDirectReg(88, channel)) == 0)
		{	
			progress|=1;
			x=i;
			break;
		}
	}
	for ( i=0x1f; i>0; i--)
	{
		writeDirectReg(99,i, channel);
		Caltime = jiffies;
		while((jiffies-Caltime)<4);
		if((readDirectReg(89, channel)) == 0)
		{
			progress|=2;
			y=i;
			break;
		}
	}

	/* The preceding is the manual gain mismatch calibration
  	   The following is the longitudinal Balance Cal
  	 */
	goActive(channel);
	if((readDirectReg(68, channel) & 0x3) & 4)
		return ;
	writeDirectReg(64,OPEN_DR64, channel);
	writeDirectReg(23,ENB2_DR23, channel);  /* enable interrupt for the balance Cal */
	writeDirectReg(97,BIT_CALCM_DR97, channel); /* this is a singular calibration bit for longitudinal calibration */
	writeDirectReg(96,0x40, channel);
	while(readDirectReg(96, channel) != 0 );

   	writeDirectReg(21,INIT_DR21, channel);
    writeDirectReg(22,INIT_DR22, channel);
    writeDirectReg(23,INIT_DR23, channel);
	/*The preceding is the longitudinal Balance Cal*/
}

/*THIRD*/

void slic_init_PCM(uint32 channel, int32 pcmFormat )
{
	uint8 modeReg = 0;
	writeDirectReg( 1, 0x00, channel);/*Disable slic pcm function */
	switch(pcmFormat)
	{
	case PCM_LINEAR:
		modeReg |= 0x18;
		break;
	case PCM_ULAW:
		modeReg |= 0x08;
		break;
	case PCM_ALAW:
		modeReg |= 0x00;
		break;
	default:/*Linear*/
		modeReg |= 0x18;
		break;
	}
	modeReg |= 0x01; /*Tri-state bit 0 on negative edge of PCLK*/
	modeReg |= 0x20; /*Enable pcm*/
	//modeReg |= 0x04; /* 16 bit tx mode */

	/* Set pcm time slot needs be careful */

	writeDirectReg( 2 , 1 + channel * 8, channel);/*'1' correspond time slot 0 in pcm */
	writeDirectReg( 3 , 0x00, channel);/*High byte TXS*/
	writeDirectReg( 4 , 1 + channel * 8, channel);
	writeDirectReg( 5 , 0x00, channel);/*High byte RXS*/
#ifdef STATE_ROLL
	chipData[channel].pcm_channel = channel;/* Slic channel must equal PCM channel */
	chipData[channel].mid = 0;
#endif
	writeDirectReg( 8 , 0x00, channel);/* disable loop back */
	writeDirectReg( 9 , 0x00, channel);/* TX RX gain = 0 db & digital mute */
	writeDirectReg( 10, 0x28, channel);
	writeDirectReg( 11, 0x00, channel);/*Hybrid adjustment 0dB*/
	writeDirectReg( 1 , modeReg, channel);/*pcm enabled*/
}

void slic_init_Value(uint32 channel, int32 pcm_mode)
{
	unsigned long timeEpoch;
	extern unsigned long volatile jiffies;
	int freq;
	char* freqs[ ] = 
		{"8192","4028","2048","1024","512","256","1536","768","32768"};

	slic_init_PCM(channel, pcm_mode);

	/* Set Ring cadence 1 sec ON , 2 sec OFF */
	slic_ring_standard(channel);
	/* Set loop current limit */
	writeDirectReg( 71 , 0x04, channel );

	/* Set external BJT and battery feed control and loop closure debounce */
	writeDirectReg( 65 , 0x11, channel );
	writeDirectReg( 66 , 0x01, channel );
	writeDirectReg( 69 , 0x0C, channel );
	/* print out the PCM clock frequecy and the revision */
	freq = readDirectReg(13, channel)>>4;  /* Read the frequency */
	rtlglue_printf("\nPCM clock = %s KHz Rev %c\n",
				freqs[freq], 'A'-1 + chipData[channel].version);

	/* Set Linefeed state to Forward on-hook transmission */
	writeDirectReg(64, 2, channel);

	if(slic_init_VerifyIReg(channel)!=SUCCESS)
	{
		rtlglue_printf("SLIC indirect register's initial values are lost\n");
	}
	//slic_ring_activate(channel);
	/* hold three seconds */
	timeEpoch = jiffies + 300;
	while(jiffies < timeEpoch);
	writeDirectReg(64, 0, channel);
	goActive(channel);
	
	if(readDirectReg(68, channel)&4)  /*ONHOOK: put down the phone*/
		chipData[channel].bOffHook = FALSE;
	else                     /*OFFHOOK: take off the phone*/
		chipData[channel].bOffHook = TRUE;
	
	chipData[channel].bEnabled = TRUE;
	
#ifdef STATE_ROLL
	chipData[channel].preState = DEMURE;
	if(chipData[channel].bOffHook == TRUE)
		chipData[channel].nowState = OFFHOOK;
	else
		chipData[channel].nowState = ONHOOK;
#endif
}

int32 slic_init_VerifyIReg(int channel)
{ 	
	int32 ret = SUCCESS;	
	uint16 i,j, initial;	
	for (i=0; i<42; i++) 	
	{		
		j = readIndirectReg((unsigned char) i, channel);
		initial= indirectRegisters[i].initial;		
		if ( j != initial )		
		{
			rtlglue_printf("The %d indirect register %04X(normal):%04X(lost)\n", i, initial,j);
			ret = FAILED;		
		}	
	}	
	return ret;
}

/*
@ ZONE: interrupt service, the functions in the following is
@ only used when kernel is polling at it. 
@ check state, take short response etc.
*/
int32 _slic_PollInt(void)
{
	uint8 isr_status;
	int32 ret = SUCCESS;
	int channel;
#ifdef STATE_ROLL
	static uint8 myPack;
#endif

	for( channel = 0 ; channel < TOTAL_SLIC_CHANNEL ; channel++ )
	{
		if( chipData[channel].bEnabled == FALSE )
		{
			continue;
		}
		/*------------------------------------------------
		 * ISR 20
		 *------------------------------------------------*/

		isr_status = readDirectReg(20, channel);
		if(isr_status & DTMFP)
		{
			rtlglue_printf("DTMF tone detected\n");
			isr_status |= DTMFP;
			writeDirectReg(20, isr_status, channel);
			slic_tone_diableOSCs(channel);
			dtmfAction(channel);
		}
		/*------------------------------------------------
		 * ISR 19
		 *------------------------------------------------*/
		/*
		   Loop closure detection, when you pick up or put down the phone
		   this interrupt is generated and inform you to act.
		 */
		isr_status = readDirectReg(19, channel);
		if(isr_status & LCIP)
		{
			isr_status |= LCIP;
			writeDirectReg(19, isr_status, channel);
	#ifdef STATE_ROLL
			if(readDirectReg(68, channel)&0x04)  /*ONHOOK: put down the phone*/
			{
				chipData[channel].bOffHook = FALSE;
				ENTER_STATE(ONHOOK);
				switch( chipData[channel].preState )
				{
				case BUSYTONE:
				case RINGTONE:
					slic_tone_diableOSCs(channel);
					myPack = I_STOP_CALLING;
					voip_TxRtpPkt(chipData[channel].pcm_channel, chipData[channel].mid, &myPack, 1, TYPE_NONE);
					rtlglue_printf("Stop Calling\n");
					ENTER_STATE(DEMURE);
					break;
				case TALKING:
					myPack = I_QUIT_TALK;
					voip_TxRtpPkt(chipData[channel].pcm_channel, chipData[channel].mid, &myPack, 1, TYPE_NONE);
					ENTER_STATE(DEMURE);
					break;
				default:
					slic_tone_diableOSCs(channel);
					ENTER_STATE(DEMURE);
					break;
				}
			}
			
			/* OFFHOOK: the phone is being taken up */
			
			else
			{
				chipData[channel].bOffHook = TRUE;
				ENTER_STATE(OFFHOOK);
				/* FIXME : ask for talk wait for status sure */
				switch( chipData[channel].preState )
				{
				case DEMURE:
				case ONHOOK:
					myPack = I_INIT_CALLING;
					voip_TxRtpPkt(chipData[channel].pcm_channel, chipData[channel].mid, &myPack, 1, TYPE_NONE);
					break;
				case RING:
					myPack = I_RECV_CALLING;
					voip_TxRtpPkt(chipData[channel].pcm_channel, chipData[channel].mid, &myPack, 1, TYPE_NONE);
					ENTER_STATE(TALKING);
					break;
				default:
					break;
				}
			}
	#endif
		}
		if(isr_status & RTIP)
		{
			rtlglue_printf("Phone picked up when RING~\n");
			isr_status |= RTIP;
			writeDirectReg(19, isr_status, channel);
	#ifdef STATE_ROLL
			if((readDirectReg(68, channel)&0x04)==0)  /*OFFHOOK*/
			{
				chipData[channel].bOffHook = TRUE;
				ENTER_STATE(OFFHOOK);
				/* FIXME : ask for talk wait for status sure */
				switch( chipData[channel].preState )
				{
				case DEMURE:
				case ONHOOK:
					break;
				case RING:
					myPack = I_RECV_CALLING;
					voip_TxRtpPkt(chipData[channel].pcm_channel, chipData[channel].mid, &myPack, 1, TYPE_NONE);
					ENTER_STATE(TALKING);
					break;
				default:
					break;
				}
			}
	#endif
		}
		
		/*------------------------------------------------
		 * ISR 18
		 *------------------------------------------------*/
		isr_status = readDirectReg(18, channel);
		if(isr_status & PMIP)
		{
			rtlglue_printf("P-Meter Inactive Time out\n");
			isr_status |= PMIP;
			writeDirectReg(18, isr_status, channel);
		}
		if(isr_status & PMAP)
		{
			rtlglue_printf("P-Meter Active Time out\n");
			isr_status |= PMAP;
			writeDirectReg(18, isr_status, channel);
		}
		if(isr_status & RGIP)
		{
			/*rtlglue_printf("Ring!!!\n");*/
			isr_status |= RGIP;
			writeDirectReg(18, isr_status, channel);
		}
		/* Ringing is now on!!and timer is count
		 * Next maybe pickup the phone
		 */
		if(isr_status & RGAP)
		{
			rtlglue_printf("Ring!!\n");
			isr_status |= RGAP;
			writeDirectReg(18, isr_status, channel);
		}
		if(isr_status & O2IP)
		{
			/*rtlglue_printf("Du-Du!!\n");*/
			isr_status |= O2IP;
			writeDirectReg(18, isr_status, channel);
		}
		if(isr_status & O2AP)
		{
			rtlglue_printf("Du-Du!!\n");
			isr_status |= O2AP;
			writeDirectReg(18, isr_status, channel);
		}
		if(isr_status & O1IP)
		{
			/*rtlglue_printf("Du-Du!\n");*/
			isr_status |= O1IP;
			writeDirectReg(18, isr_status, channel);
		}
		if(isr_status & O1AP)
		{
			rtlglue_printf("Du!\n");
			isr_status |= O1AP;
			writeDirectReg(18, isr_status, channel);
		}
	}
	return ret;
}

/*
@ ZONE: Ringing, the functions in the following is
@ only used when slic needs to ring the telephone
@ Wanna ring: _slic_ring_activate
@ Wanna stop: _slic_ring_stop
*/
void slic_ring_standard(int channel)
{ 	
 	/*
	 Enables ringing mode on ProSlic for standard North American ring
		RING_ON__LO	48
		RING_ON_HI	49
		RING_OFF_LO	50
		RING_OFF_HI	51
	 Active Timer
	*/
	writeDirectReg( RING_ON__LO, 0x80, channel);/*Active timer*/
	writeDirectReg( RING_ON_HI,  0x3E, channel);/*Active timer*/
	writeDirectReg( RING_OFF_LO, 0x00, channel);/*InActive timer*/
	writeDirectReg( RING_OFF_HI, 0x7D, channel);/*InActive timer*/
	writeDirectReg( 34, 0x18, channel);
}

void slic_ring_activate(int channel)
{
	writeDirectReg( 64, 0x04, channel);
}

void slic_ring_stop(int channel)
{
	if ((0xf & readDirectReg(0, channel))<=2 )  /* if REVISION B  */
	{
		/* Loop Debounce Register  = initial value */
		writeDirectReg(69,10, channel);
	}
	writeDirectReg(64,1, channel);
}

/*
@ ZONE: tone generation, the functions in the following is
@ only used when slic generates kinds of tones
*/
void slic_tone_diableOSCs(int channel)
{
#if 0
	uint8 regVal;
	regVal = readDirectReg(32);
	_slic_clear_DREGBIT(regVal, 2);/*disable OSC1*/
	writeDirectReg(32, regVal);
	regVal = readDirectReg(33);
	_slic_clear_DREGBIT(regVal, 2);/*disable OSC2*/
	writeDirectReg(33, regVal);
#endif
#if 1
	writeDirectReg(32, 0, channel);/*OSC1 control*/
	writeDirectReg(33, 0, channel);/*OSC2 control*/
	writeDirectReg(35, 0, channel);/*Pulse meter control*/
	
	writeDirectReg(36, 0, channel);/*OAT1 Low*/
	writeDirectReg(37, 0, channel);/*OAT1 High*/
	writeDirectReg(38, 0, channel);/*OIT1 Low*/
	writeDirectReg(39, 0, channel);/*OIT1 High*/
	
	writeDirectReg(40, 0, channel);/*OAT2 Low*/
	writeDirectReg(41, 0, channel);/*OAT2 High*/
	writeDirectReg(42, 0, channel);/*OIT2 Low*/
	writeDirectReg(43, 0, channel);/*OIT2 High*/
	
	writeDirectReg(44, 0, channel);/*PAT High*/
	writeDirectReg(45, 0, channel);/*PAt High*/
#endif
}


void slic_tone_generation(tone_struct tone, int channel)
{ 
	/* User function to extract values for oscillators 
	   from tone_struct
	   place them in the registers, and enable the oscillators.*/
	unsigned char osc1_ontimer_enable=0; 
	unsigned char osc1_offtimer_enable=0;
	unsigned char osc2_ontimer_enable=0;
	unsigned char osc2_offtimer_enable=0;
	int enable_osc2=0;
	/* Make sure the oscillators are not already on. */
	slic_tone_diableOSCs(channel);/* in DReg 32, 33*/

	if (tone.osc1.coeff == 0 || tone.osc1.x == 0) {
		rtlglue_printf("You passed me a invalid struct!\n");
		return;
	}
	/* Setup osc 1 */
	writeIndirectReg( OSC1_COEF, tone.osc1.coeff, channel);
	writeIndirectReg( OSC1X, tone.osc1.x, channel);
	writeIndirectReg( OSC1Y, tone.osc1.y, channel);
	/* Active Timer */
	if (tone.osc1.on_hi_byte != 0) {
		writeDirectReg( OSC1_ON__LO, tone.osc1.on_low_byte, channel);
		writeDirectReg( OSC1_ON_HI, tone.osc1.on_hi_byte, channel);
		osc1_ontimer_enable = 0x10;
	}
	/* Inactive Timer */
	if (tone.osc1.off_hi_byte != 0) {
		writeDirectReg( OSC1_OFF_LO, tone.osc1.off_low_byte, channel);
		writeDirectReg( OSC1_OFF_HI, tone.osc1.off_hi_byte, channel);
		osc1_offtimer_enable = 0x08;
	}
	
	if (tone.osc2.coeff != 0) {
		/* Setup OSC 2 */
		writeIndirectReg( OSC2_COEF, tone.osc2.coeff, channel);
		writeIndirectReg( OSC2X, tone.osc2.x, channel);
		writeIndirectReg( OSC2Y, tone.osc2.y, channel);
		
		/* Active Timer */
		if (tone.osc1.on_hi_byte != 0) {
			writeDirectReg( OSC2_ON__LO, tone.osc2.on_low_byte, channel);
			writeDirectReg( OSC2_ON_HI, tone.osc2.on_hi_byte, channel);
			osc2_ontimer_enable = 0x10;
		}
		/* Inactive Timer */
		if (tone.osc1.off_hi_byte != 0) {
			writeDirectReg( OSC2_OFF_LO, tone.osc2.off_low_byte, channel);
			writeDirectReg( OSC2_OFF_HI, tone.osc2.off_hi_byte, channel);
			osc2_offtimer_enable = 0x08;
		}
		enable_osc2 = 1;
	}
	writeDirectReg( OSC1, (unsigned char)(0x06 | osc1_ontimer_enable | osc1_offtimer_enable), channel);
	if (enable_osc2)
		writeDirectReg( OSC2, (unsigned char)(0x06 | osc2_ontimer_enable | osc2_offtimer_enable), channel);
	return;
}


void slic_tone_dial(int channel)
{
  writeIndirectReg(13,DIALTONE_IR13, channel);
  writeIndirectReg(14,DIALTONE_IR14, channel);
  writeIndirectReg(16,DIALTONE_IR16, channel);
  writeIndirectReg(17,DIALTONE_IR17, channel);
  writeDirectReg(36,  DIALTONE_DR36, channel);
  writeDirectReg(37,  DIALTONE_DR37, channel);
  writeDirectReg(38,  DIALTONE_DR38, channel);
  writeDirectReg(39,  DIALTONE_DR39, channel);
  writeDirectReg(40,  DIALTONE_DR40, channel);
  writeDirectReg(41,  DIALTONE_DR41, channel);
  writeDirectReg(42,  DIALTONE_DR42, channel);
  writeDirectReg(43,  DIALTONE_DR43, channel);

  writeDirectReg(32,  DIALTONE_DR32, channel);
  writeDirectReg(33,  DIALTONE_DR33, channel);
}

void slic_tone_busy(int channel)
{
 
  writeIndirectReg(13,BUSYTONE_IR13, channel);
  writeIndirectReg(14,BUSYTONE_IR14, channel);
  writeIndirectReg(16,BUSYTONE_IR16, channel);
  writeIndirectReg(17,BUSYTONE_IR17, channel);
  writeDirectReg(36,  BUSYTONE_DR36, channel);
  writeDirectReg(37,  BUSYTONE_DR37, channel);
  writeDirectReg(38,  BUSYTONE_DR38, channel);
  writeDirectReg(39,  BUSYTONE_DR39, channel);
  writeDirectReg(40,  BUSYTONE_DR40, channel);
  writeDirectReg(41,  BUSYTONE_DR41, channel);
  writeDirectReg(42,  BUSYTONE_DR42, channel);
  writeDirectReg(43,  BUSYTONE_DR43, channel);
  
  writeDirectReg(32,  BUSYTONE_DR32, channel);
  writeDirectReg(33,  BUSYTONE_DR33, channel);
}

void slic_tone_reorder(int channel)
{
  writeIndirectReg(13,REORDERTONE_IR13, channel);
  writeIndirectReg(14,REORDERTONE_IR14, channel);
  writeIndirectReg(16,REORDERTONE_IR16, channel);
  writeIndirectReg(17,REORDERTONE_IR17, channel);
  writeDirectReg(36,  REORDERTONE_DR36, channel);
  writeDirectReg(37,  REORDERTONE_DR37, channel);
  writeDirectReg(38,  REORDERTONE_DR38, channel);
  writeDirectReg(39,  REORDERTONE_DR39, channel);
  writeDirectReg(40,  REORDERTONE_DR40, channel);
  writeDirectReg(41,  REORDERTONE_DR41, channel);
  writeDirectReg(42,  REORDERTONE_DR42, channel);
  writeDirectReg(43,  REORDERTONE_DR43, channel);
  
  writeDirectReg(32,  REORDERTONE_DR32, channel);
  writeDirectReg(33,  REORDERTONE_DR33, channel);
 
}

void slic_tone_congestion(int channel)
{
  writeIndirectReg(13,CONGESTIONTONE_IR13, channel);
  writeIndirectReg(14,CONGESTIONTONE_IR14, channel);
  writeIndirectReg(16,CONGESTIONTONE_IR16, channel);
  writeIndirectReg(17,CONGESTIONTONE_IR17, channel);
  writeDirectReg(36,  CONGESTIONTONE_DR36, channel);
  writeDirectReg(37,  CONGESTIONTONE_DR37, channel);
  writeDirectReg(38,  CONGESTIONTONE_DR38, channel);
  writeDirectReg(39,  CONGESTIONTONE_DR39, channel);
  writeDirectReg(40,  CONGESTIONTONE_DR40, channel);
  writeDirectReg(41,  CONGESTIONTONE_DR41, channel);
  writeDirectReg(42,  CONGESTIONTONE_DR42, channel);
  writeDirectReg(43,  CONGESTIONTONE_DR43, channel);

  writeDirectReg(32,  CONGESTIONTONE_DR32, channel);
  writeDirectReg(33,  CONGESTIONTONE_DR33, channel);

}

void slic_tone_ringbackPbx(int channel)
{
  writeIndirectReg(13,RINGBACKPBXTONE_IR13, channel);
  writeIndirectReg(14,RINGBACKPBXTONE_IR14, channel);
  writeIndirectReg(16,RINGBACKPBXTONE_IR16, channel);
  writeIndirectReg(17,RINGBACKPBXTONE_IR17, channel);
  writeDirectReg(36,  RINGBACKPBXTONE_DR36, channel);
  writeDirectReg(37,  RINGBACKPBXTONE_DR37, channel);
  writeDirectReg(38,  RINGBACKPBXTONE_DR38, channel);
  writeDirectReg(39,  RINGBACKPBXTONE_DR39, channel);
  writeDirectReg(40,  RINGBACKPBXTONE_DR40, channel);
  writeDirectReg(41,  RINGBACKPBXTONE_DR41, channel);
  writeDirectReg(42,  RINGBACKPBXTONE_DR42, channel);
  writeDirectReg(43,  RINGBACKPBXTONE_DR43, channel);

  writeDirectReg(32,  RINGBACKPBXTONE_DR32, channel);
  writeDirectReg(33,  RINGBACKPBXTONE_DR33, channel);

}

void slic_tone_ringBack(int channel)
{
  writeIndirectReg(13,RINGBACKTONE_IR13, channel);
  writeIndirectReg(14,RINGBACKTONE_IR14, channel);
  writeIndirectReg(16,RINGBACKTONE_IR16, channel);
  writeIndirectReg(17,RINGBACKTONE_IR17, channel);
  writeDirectReg(36,  RINGBACKTONE_DR36, channel);
  writeDirectReg(37,  RINGBACKTONE_DR37, channel);
  writeDirectReg(38,  RINGBACKTONE_DR38, channel);
  writeDirectReg(39,  RINGBACKTONE_DR39, channel);
  writeDirectReg(40,  RINGBACKTONE_DR40, channel);
  writeDirectReg(41,  RINGBACKTONE_DR41, channel);
  writeDirectReg(42,  RINGBACKTONE_DR42, channel);
  writeDirectReg(43,  RINGBACKTONE_DR43, channel);
  
  writeDirectReg(32,  RINGBACKTONE_DR32, channel);
  writeDirectReg(33,  RINGBACKTONE_DR33, channel);
 
}

void slic_tone_ringBackJapan(int channel)
{
  writeIndirectReg(13,RINGBACKJAPANTONE_IR13, channel);
  writeIndirectReg(14,RINGBACKJAPANTONE_IR14, channel);
  writeIndirectReg(16,RINGBACKJAPANTONE_IR16, channel);
  writeIndirectReg(17,RINGBACKJAPANTONE_IR17, channel);
  writeDirectReg(36,  RINGBACKJAPANTONE_DR36, channel);
  writeDirectReg(37,  RINGBACKJAPANTONE_DR37, channel);
  writeDirectReg(38,  RINGBACKJAPANTONE_DR38, channel);
  writeDirectReg(39,  RINGBACKJAPANTONE_DR39, channel);
  writeDirectReg(40,  RINGBACKJAPANTONE_DR40, channel);
  writeDirectReg(41,  RINGBACKJAPANTONE_DR41, channel);
  writeDirectReg(42,  RINGBACKJAPANTONE_DR42, channel);
  writeDirectReg(43,  RINGBACKJAPANTONE_DR43, channel);
  writeDirectReg(32,  RINGBACKJAPANTONE_DR32, channel);
  writeDirectReg(33,  RINGBACKJAPANTONE_DR33, channel);
}

void slic_tone_busyJapan(int channel)
{
  writeIndirectReg(13,BUSYJAPANTONE_IR13, channel);
  writeIndirectReg(14,BUSYJAPANTONE_IR14, channel);
  writeIndirectReg(16,BUSYJAPANTONE_IR16, channel);
  writeIndirectReg(17,BUSYJAPANTONE_IR17, channel);
  writeDirectReg(36,  BUSYJAPANTONE_DR36, channel);
  writeDirectReg(37,  BUSYJAPANTONE_DR37, channel);
  writeDirectReg(38,  BUSYJAPANTONE_DR38, channel);
  writeDirectReg(39,  BUSYJAPANTONE_DR39, channel);
  writeDirectReg(40,  BUSYJAPANTONE_DR40, channel);
  writeDirectReg(41,  BUSYJAPANTONE_DR41, channel);
  writeDirectReg(42,  BUSYJAPANTONE_DR42, channel);
  writeDirectReg(43,  BUSYJAPANTONE_DR43, channel);
  writeDirectReg(32,  BUSYJAPANTONE_DR32, channel);
  writeDirectReg(33,  BUSYJAPANTONE_DR33, channel);
}

void slic_read_PCMCLK(int channel)
{
	char* freqs[ ] = 
	{
		"8192",	/*0*/
		"4028",	/*1*/
		"2048",	/*2*/
		"1024",	/*3*/
		"512",	/*4*/
		"256",	/*5*/
		"1536",	/*6*/
		"768",	/*7*/
		"32768"	/*8*/
	};
	rtlglue_printf("\nPCM clock = %s KHz \n",freqs[readDirectReg(13, channel)>>4]);
}

/*
@ ZONE: DTMF , the functions in the following is
@ only used when slic generates kinds of tones
*/

unsigned char digit(int channel)
{
	return readDirectReg(24, channel) & 0x0f;
}

void dtmfAction(int channel)
{  
	char rawDigit,asciiChar;
	rawDigit=digit(channel);
	/*asciiChar= '0' + digit;*/
	switch (rawDigit){
	case 0xA :
			asciiChar = '0';
			break;
	case 0xB:
			asciiChar = '*';
			break;
	case 0xC:
			asciiChar = '#';
			break;
	default:
		asciiChar= '0' + rawDigit;
		break;
	}
	
	if (chipData[channel].digit_count < 20)
	{
		chipData[channel].DTMF_digits[chipData[channel].digit_count] = asciiChar; 
		chipData[channel].digit_count++;
		chipData[channel].DTMF_digits[chipData[channel].digit_count]= 0;
		rtlglue_printf("Value= 0x%01x  String collected \"%s\" \n", 
			rawDigit, chipData[channel].DTMF_digits );
	}
}

/* Accessory functions */

void goActive(int channel)
{
	unsigned long timeEpoch;
	extern unsigned long volatile jiffies;
	writeDirectReg(64,1,channel);	
	/* LOOP STATE REGISTER SET TO ACTIVE */
	/* Active works for on-hook and off-hook see spec. */
	/* The phone hook-switch sets the off-hook and on-hook substate*/
	timeEpoch = jiffies + 10;
	while(jiffies < timeEpoch);
}

