#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include "rtl865x/rtl_types.h"
#include "../rtl8186/pcm.h"


//#define LOOPBACK
#define TEST_ALL_CHANNEL
#define TXBUF			0xa0600000
#define RXBUF			0xa0610000
#define TXBUF_PHYADDR  (TXBUF & 0x1fffffff)
#define RXBUF_PHYADDR  (RXBUF & 0x1fffffff)

#define RTL865X_REG_BASE (0xBD010000)

#define rtl_inb(offset) (*(volatile unsigned char *)(RTL865X_REG_BASE+offset))
#define rtl_inw(offset) (*(volatile unsigned short *)(RTL865X_REG_BASE+offset))
#define rtl_inl(offset) (*(volatile unsigned long *)(RTL865X_REG_BASE+offset))
#define rtl_outb(offset,val)	(*(volatile unsigned char *)(RTL865X_REG_BASE+offset) = val)
#define rtl_outw(offset,val)	(*(volatile unsigned short *)(RTL865X_REG_BASE+ offset) = val)
#define rtl_outl(offset,val)	(*(volatile unsigned long *)(RTL865X_REG_BASE+offset) = val)

#undef REG32
#define REG32(offset) 			(*((volatile uint32 *)(RTL865X_REG_BASE+offset)))

#define flush_dcache() \
	__asm__ __volatile__ ( \
		"mfc0 $9,$20\n\t" \
		"nop\n\t" \
		"la $10,0xfffffffd\n\t" \
		"and $9,$10\n\t" \
		"mtc0 $9,$12\n\t" \
		"nop\n\t" \
		"li $10,2\n\t" \
		"or $9,$10\n\t" \
		"mtc0 $9,$12\n\t" \
		"nop\n\t" \
		"nop\n\t" \
		: /* no output */ \
		: /* no input */ \
		)

#ifdef BIT
#error	"BIT define occurred earlier elsewhere!\n"
#endif

#define BIT(x)	( 1 << (x))
#define	PGCR			(0x7000+0x00)
#define PCSCR			(0x7000+0x04)
#define PTSAR			(0x7000+0x08)
#define PBSIZE			(0x7000+0x0c)
#define	CH0TXBSA		(0x7000+0x10)
#define	CH1TXBSA		(0x7000+0x14)
#define	CH2TXBSA		(0x7000+0x18)
#define	CH3TXBSA		(0x7000+0x1c)
#define	CH0RXBSA		(0x7000+0x20)
#define	CH1RXBSA		(0x7000+0x24)
#define	CH2RXBSA		(0x7000+0x28)
#define	CH3RXBSA		(0x7000+0x2c)
#define	PIMR			(0x7000+0x30)
#define	PISR			(0x7000+0x34)
//PGCR 
#define	PCMENABLE		BIT(12)
#define	PCMCLKDIR		BIT(11)
#define INTERNELPLL		BIT(11)		// name redefined by brian_bin
#define P0EXDSE			BIT(10)
#define	FSINV			BIT(9)
//PCSCR
#define	CH0ILBE			BIT(28)		//channel 0 loopback
#define	CH0CMPE			BIT(27)		//channel 0 compander enable
#define CH0Alaw			BIT(26)		//channel 0 u/A-law
#define CH0TE			BIT(25)		//channel 0 TX enable
#define	CH0RE			BIT(24)		//channel 0 RX enable

#define	CH1ILBE			BIT(20)		//channel 1 loopback
#define	CH1CMPE			BIT(19)		//channel 1 compander enable
#define CH1Alaw			BIT(18)		//channel 1 u/A-law
#define CH1TE			BIT(17)		//channel 1 TX enable
#define	CH1RE			BIT(16)		//channel 1 RX enable

#define	CH2ILBE			BIT(12)		//channel 0 loopback
#define	CH2CMPE			BIT(11)		//channel 0 compander enable
#define CH2Alaw			BIT(10)		//channel 0 u/A-law
#define CH2TE			BIT(9)		//channel 0 TX enable
#define	CH2RE			BIT(8)		//channel 0 RX enable

#define	CH3ILBE			BIT(4)		//channel 1 loopback
#define	CH3CMPE			BIT(3)		//channel 1 compander enable
#define CH3Alaw			BIT(2)		//channel 1 u/A-law
#define CH3TE			BIT(1)		//channel 1 TX enable
#define	CH3RE			BIT(0)		//channel 1 RX enable

#define	CH0_1			0x1
#define	CH0_2			0x2
#define	CH0_3			0x3	
#define	CH1_2			0x4
#define	CH1_3			0x5
#define	CH2_3			0x6	

//#define DEBUG

#define PAGE1_OWN	BIT(1)
#define PAGE0_OWN	BIT(0)

// if remark both, foreground polling mode, ok
//#define INT_MODE	1		// process in foreground, ok
#define PCM_PKT_SIZE 2048	// pre-allocated big-enoungh data buffer	
#define PAGESIZE 1024
#define CH_NUM 4
#define PKTSIZE 255	// real voice data size = 4 * (PKTSIZE+1), max.=1024 	
					// note: the larger data size, the larger delay
//extern union task_union sys_tasks[NUM_TASK];
void init_pcm(int channel);
void pcm_tx(int channel,int slot,int seed,int size);
void parse_pcm_rx(int channel,int size);
void prepare_pcm_data(unsigned char *buf, int size,int seed);
void polltx(int channel);

extern void i2c_task(void);

struct pcm_pkt {
	unsigned char 	tx_rx;
	unsigned char 	*payload;
};	
static struct pcm_pkt tx[CH_NUM];/*CH_NUM=4*/
static struct pcm_pkt rx[CH_NUM];/*CH_NUM=4*/
static int8 chanEnabled[CH_NUM];/*CH_NUM=4*/
static uint8 chanTxCurrPage[CH_NUM];/*CH_NUM=4*/
static uint8 chanRxCurrPage[CH_NUM];/*CH_NUM=4*/

static uint8 dingDong[] = {
0x00, 0x00, 0x01, 0x00, 0xff, 0xfe, 0x02, 0xff, 0xf6, 0xfe, 0x07, 0xfd, 0xfd, 0x04, 0xfc, 0xfc, 0x03, 0xfc, 0xff, 0x0b, 0x01, 0xf5, 0xf6, 0xfa, 0x03, 0x03, 0xff, 0x01, 0x00, 0xf6, 0xf7, 0x03, 
0x01, 0xf9, 0xff, 0x00, 0xfe, 0xfe, 0xfe, 0xfe, 0x01, 0x01, 0xfb, 0xfc, 0x01, 0xfd, 0xf8, 0x00, 0x06, 0xff, 0xfc, 0xfe, 0x00, 0x00, 0xfc, 0xfe, 0x03, 0x03, 0x01, 0xfc, 0xfc, 0x04, 0x05, 0x02, 
0x04, 0x00, 0xfb, 0x00, 0x02, 0x02, 0x02, 0xff, 0x02, 0x06, 0x02, 0x01, 0x02, 0x03, 0x04, 0x00, 0xfe, 0x02, 0x04, 0x03, 0x03, 0x04, 0x05, 0x02, 0x00, 0x04, 0x05, 0x03, 0x01, 0x02, 0x04, 0x04, 
0x00, 0x01, 0x04, 0x02, 0x03, 0x07, 0x02, 0xfe, 0x02, 0x01, 0xfe, 0x01, 0x06, 0x03, 0x01, 0x01, 0xfc, 0xff, 0x05, 0xfa, 0xf6, 0x18, 0x16, 0xf3, 0xfe, 0xf2, 0xe1, 0xf9, 0x00, 0xfe, 0x13, 0x1d, 
0x1b, 0x08, 0xe6, 0xca, 0xc9, 0xea, 0x11, 0x1d, 0x25, 0x3e, 0x1e, 0xe7, 0xc8, 0xc6, 0xd8, 0xfe, 0x1e, 0x27, 0x1a, 0x04, 0xe7, 0xd9, 0xda, 0xea, 0xff, 0x15, 0x2e, 0x21, 0xfb, 0xe0, 0xcd, 0xcb, 
0xe9, 0x10, 0x26, 0x39, 0x28, 0xff, 0xe1, 0xc1, 0xc5, 0xe5, 0x08, 0x32, 0x42, 0x19, 0xfd, 0xe8, 0xce, 0xc4, 0xd7, 0xfb, 0x31, 0x44, 0x22, 0x00, 0xe8, 0xda, 0xdc, 0xee, 0x07, 0x16, 0x21, 0x1a, 
0xfd, 0xe3, 0xd9, 0xdb, 0xf4, 0x18, 0x2d, 0x2e, 0x22, 0x00, 0xd6, 0xc9, 0xd9, 0xfa, 0x18, 0x2c, 0x33, 0x22, 0xfc, 0xd8, 0xc8, 0xd0, 0xf4, 0x25, 0x3e, 0x37, 0x22, 0x01, 0xd8, 0xbc, 0xc4, 0xec, 
0x1c, 0x46, 0x4d, 0x2f, 0x02, 0xd6, 0xbb, 0xc9, 0xf2, 0x1b, 0x33, 0x3a, 0x25, 0xfb, 0xd1, 0xcd, 0xe1, 0xff, 0x21, 0x32, 0x28, 0x0e, 0xe7, 0xcf, 0xd5, 0xe5, 0x04, 0x2d, 0x3d, 0x35, 0x13, 0xe0, 
0xc4, 0xcc, 0xe9, 0x0a, 0x26, 0x38, 0x34, 0x0e, 0xe6, 0xca, 0xc3, 0xde, 0x03, 0x23, 0x3e, 0x32, 0x0f, 0xf8, 0xd4, 0xbe, 0xda, 0xf9, 0x1b, 0x3c, 0x35, 0x13, 0xec, 0xc4, 0xc3, 0xdd, 0x09, 0x29, 
0x37, 0x34, 0x10, 0xdd, 0xbe, 0xc4, 0xe0, 0x04, 0x27, 0x39, 0x35, 0x17, 0xe3, 0xbb, 0xb6, 0xd9, 0x06, 0x26, 0x3d, 0x3e, 0x19, 0xe3, 0xbb, 0xb9, 0xdf, 0x0b, 0x27, 0x39, 0x35, 0x16, 0xe5, 0xbf, 
0xbb, 0xd9, 0x06, 0x2b, 0x3b, 0x2e, 0x04, 0xdc, 0xc3, 0xcd, 0xfb, 0x21, 0x2a, 0x2b, 0x14, 0xf0, 0xd9, 0xce, 0xce, 0xf3, 0x25, 0x39, 0x2f, 0x1f, 0xfc, 0xd9, 0xcb, 0xd6, 0xf5, 0x15, 0x22, 0x24, 
0x1f, 0x02, 0xe4, 0xd8, 0xde, 0xfb, 0x19, 0x1d, 0x1b, 0x14, 0xfa, 0xdf, 0xd9, 0xe8, 0x05, 0x1d, 0x25, 0x1e, 0x0b, 0xf7, 0xdf, 0xcc, 0xdd, 0xfc, 0x16, 0x26, 0x2c, 0x21, 0x03, 0xe1, 0xd0, 0xd6, 
0xf4, 0x11, 0x21, 0x2a, 0x1d, 0xf9, 0xdb, 0xd1, 0xe1, 0xf9, 0x0a, 0x1e, 0x31, 0x22, 0xf9, 0xdc, 0xcd, 0xdc, 0xff, 0x16, 0x21, 0x25, 0x1a, 0xff, 0xe1, 0xcd, 0xd5, 0xf2, 0x12, 0x30, 0x3a, 0x22, 
0xfb, 0xdc, 0xc8, 0xd4, 0xf4, 0x11, 0x2a, 0x36, 0x23, 0x04, 0xe0, 0xc9, 0xd5, 0xf4, 0x0e, 0x28, 0x34, 0x22, 0x00, 0xe0, 0xcd, 0xdc, 0x01, 0x13, 0x1b, 0x28, 0x1d, 0xf9, 0xdd, 0xd6, 0xe1, 0xfe, 
0x13, 0x18, 0x21, 0x1f, 0x01, 0xe2, 0xda, 0xe3, 0xf3, 0x09, 0x1c, 0x1e, 0x13, 0xfa, 0xe1, 0xde, 0xed, 0xfc, 0x06, 0x15, 0x23, 0x19, 0xf9, 0xe2, 0xdf, 0xe7, 0x01, 0x14, 0x15, 0x0f, 0x06, 0xf4, 
0xe8, 0xec, 0xf1, 0xf8, 0x0d, 0x1c, 0x1a, 0x11, 0xf8, 0xdc, 0xdc, 0xf0, 0x05, 0x12, 0x16, 0x14, 0x0d, 0xfa, 0xee, 0xea, 0xe7, 0xf8, 0x19, 0x26, 0x1b, 0x0b, 0xf6, 0xe8, 0xe7, 0xf0, 0xfd, 0x0c, 
0x18, 0x19, 0x0c, 0xfd, 0xec, 0xe3, 0xeb, 0xf8, 0x0b, 0x20, 0x21, 0x12, 0xfd, 0xe3, 0xdf, 0xfa, 0x07, 0x06, 0x10, 0x11, 0x04, 0xf8, 0xea, 0xe4, 0xf2, 0x04, 0x0d, 0x1d, 0x25, 0x0f, 0xf0, 0xdd, 
0xde, 0xf0, 0x02, 0x0c, 0x10, 0x12, 0x0b, 0xf9, 0xe3, 0xe1, 0xf3, 0x01, 0x12, 0x26, 0x20, 0xfd, 0xe6, 0xdf, 0xe8, 0xfc, 0x0a, 0x0d, 0x12, 0x18, 0x0a, 0xed, 0xdd, 0xe0, 0xf1, 0x09, 0x18, 0x20, 
0x1c, 0x05, 0xef, 0xda, 0xd4, 0xef, 0x14, 0x20, 0x22, 0x1d, 0x08, 0xf1, 0xdf, 0xd0, 0xd6, 0xf8, 0x1d, 0x33, 0x37, 0x1d, 0xe9, 0xcd, 0xd3, 0xea, 0x08, 0x1b, 0x22, 0x27, 0x11, 0xf1, 0xd9, 0xd5, 
0xe5, 0x04, 0x1c, 0x26, 0x27, 0x15, 0xf4, 0xd9, 0xd6, 0xec, 0x03, 0x1b, 0x2c, 0x24, 0x0d, 0xf0, 0xd6, 0xdb, 0xf1, 0x03, 0x1b, 0x31, 0x2f, 0x0f, 0xe4, 0xcd, 0xd4, 0xef, 0x08, 0x1a, 0x2b, 0x2a, 
0x0d, 0xe9, 0xd5, 0xd4, 0xe7, 0x09, 0x29, 0x30, 0x1d, 0xff, 0xe0, 0xcd, 0xd5, 0xed, 0x09, 0x24, 0x33, 0x26, 0x04, 0xdf, 0xc8, 0xc9, 0xe5, 0x09, 0x26, 0x32, 0x28, 0x06, 0xdd, 0xc9, 0xd3, 0xea, 
0x06, 0x20, 0x33, 0x29, 0x05, 0xde, 0xc6, 0xca, 0xea, 0x09, 0x22, 0x36, 0x2c, 0x07, 0xe2, 0xcf, 0xd0, 0xe9, 0x09, 0x1f, 0x2a, 0x24, 0x06, 0xe1, 0xd0, 0xd6, 0xed, 0x0f, 0x29, 0x33, 0x21, 0xfa, 
0xd6, 0xd0, 0xdd, 0xf4, 0x0f, 0x21, 0x28, 0x21, 0x00, 0xdd, 0xd2, 0xe0, 0xfc, 0x17, 0x27, 0x2c, 0x18, 0xf9, 0xde, 0xd4, 0xdf, 0xfd, 0x17, 0x27, 0x2a, 0x11, 0xed, 0xdd, 0xd9, 0xe4, 0x02, 0x1c, 
0x28, 0x2a, 0x14, 0xee, 0xd8, 0xd9, 0xeb, 0x07, 0x1a, 0x22, 0x24, 0x12, 0xf4, 0xde, 0xd6, 0xdd, 0xfd, 0x22, 0x2f, 0x28, 0x15, 0xf6, 0xda, 0xcf, 0xdf, 0xf8, 0x17, 0x2a, 0x29, 0x16, 0xf4, 0xd3, 
0xcf, 0xe0, 0xff, 0x1f, 0x2b, 0x22, 0x13, 0xf8, 0xdc, 0xd4, 0xe2, 0xfb, 0x16, 0x23, 0x1e, 0x0b, 0xf1, 0xdd, 0xdd, 0xef, 0x05, 0x19, 0x20, 0x1c, 0x10, 0xf4, 0xd7, 0xd3, 0xe1, 0xfe, 0x1a, 0x26, 
0x25, 0x15, 0xf6, 0xde, 0xdb, 0xe5, 0xf7, 0x10, 0x23, 0x24, 0x13, 0xfb, 0xe0, 0xd6, 0xe8, 0x02, 0x15, 0x20, 0x18, 0x08, 0xfa, 0xeb, 0xe0, 0xe3, 0xfa, 0x19, 0x2b, 0x28, 0x11, 0xef, 0xd6, 0xd6, 
0xe9, 0x01, 0x17, 0x25, 0x21, 0x10, 0xfb, 0xe5, 0xd9, 0xde, 0xfa, 0x1a, 0x2b, 0x26, 0x12, 0xf5, 0xdc, 0xd5, 0xe3, 0xf9, 0x10, 0x23, 0x28, 0x16, 0xf6, 0xdc, 0xd9, 0xe6, 0xfc, 0x15, 0x29, 0x29, 
0x15, 0xf6, 0xdc, 0xd1, 0xdb, 0xf5, 0x10, 0x25, 0x29, 0x17, 0xfe, 0xe2, 0xd1, 0xdc, 0xf8, 0x15, 0x29, 0x2b, 0x18, 0xf9, 0xdb, 0xce, 0xd8, 0xf8, 0x1a, 0x2b, 0x26, 0x11, 0xf7, 0xe1, 0xd7, 0xe0, 
0xf8, 0x13, 0x27, 0x2e, 0x17, 0xf2, 0xd8, 0xd6, 0xe8, 0x05, 0x17, 0x1e, 0x1c, 0x0c, 0xf5, 0xe3, 0xd8, 0xe3, 0x01, 0x20, 0x31, 0x2a, 0x0b, 0xe7, 0xcf, 0xd5, 0xed, 0x03, 0x13, 0x22, 0x26, 0x15, 
0xfa, 0xdb, 0xd0, 0xdc, 0xfb, 0x1d, 0x31, 0x2a, 0x0d, 0xec, 0xd9, 0xdb, 0xe8, 0xfc, 0x13, 0x25, 0x2b, 0x19, 0xf7, 0xd7, 0xcc, 0xdb, 0xfc, 0x1d, 0x35, 0x2f, 0x11, 0xed, 0xd8, 0xd5, 0xe2, 0xf9, 
0x11, 0x26, 0x2c, 0x18, 0xf6, 0xd9, 0xd1, 0xe1, 0x01, 0x1e, 0x2d, 0x28, 0x0f, 0xee, 0xd5, 0xd2, 0xe3, 0xfe, 0x17, 0x2a, 0x2a, 0x12, 0xf1, 0xd7, 0xd0, 0xe2, 0x01, 0x1e, 0x2d, 0x27, 0x0d, 0xee, 
0xda, 0xd7, 0xe1, 0xf8, 0x14, 0x26, 0x29, 0x17, 0xf6, 0xdd, 0xd4, 0xe1, 0x01, 0x1e, 0x27, 0x20, 0x07, 0xeb, 0xde, 0xe0, 0xea, 0xfd, 0x13, 0x21, 0x21, 0x0d, 0xf1, 0xdf, 0xdd, 0xe8, 0xff, 0x17, 
0x22, 0x1d, 0x0c, 0xf5, 0xe1, 0xda, 0xe4, 0xfb, 0x16, 0x25, 0x25, 0x0f, 0xf1, 0xdc, 0xd8, 0xe4, 0xff, 0x19, 0x27, 0x23, 0x0f, 0xf2, 0xdd, 0xda, 0xe8, 0xfb, 0x0f, 0x20, 0x24, 0x14, 0xf6, 0xdf, 
0xda, 0xe8, 0x03, 0x1a, 0x23, 0x1d, 0x09, 0xf2, 0xe4, 0xe2, 0xea, 0xfd, 0x12, 0x1b, 0x1b, 0x0f, 0xf7, 0xe4, 0xe2, 0xee, 0x03, 0x16, 0x1c, 0x18, 0x08, 0xf6, 0xec, 0xeb, 0xf0, 0xf9, 0x08, 0x1b, 
0x1d, 0x0e, 0xf8, 0xe6, 0xe2, 0xf0, 0x04, 0x12, 0x19, 0x16, 0x09, 0xf9, 0xeb, 0xe9, 0xed, 0xf9, 0x0e, 0x1c, 0x19, 0x09, 0xf8, 0xeb, 0xe8, 0xef, 0xfc, 0x08, 0x17, 0x1f, 0x15, 0x00, 0xe9, 0xdc, 
0xe4, 0xf7, 0x0b, 0x1c, 0x20, 0x12, 0xfe, 0xeb, 0xe3, 0xe9, 0xf7, 0x0c, 0x1d, 0x1b, 0x0a, 0xf8, 0xeb, 0xe4, 0xea, 0xfa, 0x0b, 0x1d, 0x23, 0x10, 0xf8, 0xe8, 0xe3, 0xeb, 0xfb, 0x0d, 0x1c, 0x1c, 
0x0b, 0xf4, 0xe7, 0xe3, 0xe8, 0xf8, 0x0c, 0x1f, 0x1f, 0x11, 0xfe, 0xe9, 0xdf, 0xe8, 0xfb, 0x0a, 0x15, 0x1a, 0x11, 0x00, 0xed, 0xe1, 0xe2, 0xf4, 0x0c, 0x1c, 0x19, 0x0c, 0xfa, 0xed, 0xe9, 0xf0, 
0xfc, 0x0a, 0x17, 0x16, 0x07, 0xf7, 0xea, 0xe5, 0xed, 0xfb, 0x0c, 0x1e, 0x20, 0x0c, 0xf6, 0xe9, 0xe5, 0xec, 0xf9, 0x0a, 0x18, 0x1b, 0x11, 0xfd, 0xea, 0xe1, 0xe6, 0xf5, 0x0b, 0x1b, 0x1c, 0x0f, 
0xfe, 0xef, 0xe9, 0xed, 0xfa, 0x09, 0x16, 0x17, 0x09, 0xf8, 0xed, 0xe8, 0xed, 0xfc, 0x0d, 0x19, 0x1a, 0x0f, 0xff, 0xef, 0xe4, 0xe7, 0xf5, 0x07, 0x17, 0x1c, 0x14, 0x03, 0xef, 0xe2, 0xe3, 0xf4, 
0x0a, 0x19, 0x1b, 0x12, 0x03, 0xef, 0xe3, 0xe9, 0xf7, 0x0a, 0x17, 0x17, 0x0d, 0xfe, 0xed, 0xe2, 0xe6, 0xf6, 0x0a, 0x1c, 0x20, 0x13, 0xfd, 0xeb, 0xe4, 0xe9, 0xf7, 0x0b, 0x18, 0x17, 0x08, 0xf9, 
0xed, 0xe4, 0xe8, 0xf9, 0x0f, 0x1f, 0x1c, 0x0b, 0xf7, 0xe8, 0xe0, 0xe7, 0xfa, 0x0f, 0x1d, 0x19, 0x09, 0xf9, 0xeb, 0xe2, 0xe9, 0xf9, 0x0d, 0x1a, 0x17, 0x09, 0xfb, 0xed, 0xe7, 0xed, 0xfe, 0x0b, 
0x11, 0x0f, 0x07, 0xfb, 0xf0, 0xe8, 0xeb, 0xfa, 0x0d, 0x18, 0x17, 0x0d, 0xfe, 0xf0, 0xe8, 0xec, 0xfa, 0x0a, 0x15, 0x13, 0x08, 0xfa, 0xef, 0xec, 0xf2, 0xfe, 0x0b, 0x12, 0x14, 0x0b, 0xfb, 0xef, 
0xec, 0xf1, 0xfc, 0x0a, 0x12, 0x0f, 0x06, 0xfb, 0xf5, 0xf2, 0xf1, 0xfa, 0x08, 0x13, 0x15, 0x0d, 0x00, 0xf5, 0xec, 0xee, 0xfb, 0x0a, 0x11, 0x10, 0x05, 0xf9, 0xf4, 0xf3, 0xf5, 0xff, 0x0d, 0x12, 
0x10, 0x0b, 0x01, 0xf3, 0xe9, 0xea, 0xf9, 0x0c, 0x14, 0x15, 0x0f, 0x00, 0xf1, 0xea, 0xeb, 0xf7, 0x08, 0x14, 0x19, 0x11, 0x00, 0xf1, 0xea, 0xec, 0xf7, 0x06, 0x10, 0x12, 0x11, 0x04, 0xf3, 0xe9, 
0xeb, 0xf6, 0x07, 0x14, 0x18, 0x12, 0x01, 0xef, 0xe7, 0xeb, 0xfa, 0x0b, 0x16, 0x16, 0x0b, 0xfb, 0xeb, 0xe3, 0xec, 0x01, 0x15, 0x1c, 0x15, 0x08, 0xf8, 0xea, 0xe6, 0xed, 0xfd, 0x10, 0x18, 0x12, 
0x04, 0xf8, 0xef, 0xeb, 0xf1, 0xff, 0x0c, 0x16, 0x15, 0x0b, 0xfb, 0xec, 0xe3, 0xeb, 0xfe, 0x0f, 0x13, 0x0d, 0x05, 0xfb, 0xf1, 0xed, 0xf3, 0x02, 0x0e, 0x13, 0x10, 0x04, 0xf3, 0xe9, 0xea, 0xf7, 
0x05, 0x0e, 0x10, 0x0f, 0x06, 0xf7, 0xec, 0xe9, 0xf3, 0x04, 0x11, 0x15, 0x12, 0x07, 0xf8, 0xed, 0xeb, 0xf1, 0xfe, 0x0b, 0x12, 0x10, 0x05, 0xf8, 0xf0, 0xef, 0xf6, 0x01, 0x0c, 0x11, 0x0c, 0x01, 
0xf4, 0xed, 0xf0, 0xfc, 0x07, 0x0e, 0x0f, 0x08, 0xfd, 0xf4, 0xef, 0xf0, 0xfc, 0x0b, 0x14, 0x11, 0x09, 0xfe, 0xf5, 0xee, 0xf0, 0xfb, 0x09, 0x10, 0x0e, 0x06, 0xfd, 0xf2, 0xeb, 0xf3, 0x04, 0x0f, 
0x13, 0x10, 0x07, 0xfc, 0xf1, 0xed, 0xf1, 0xfc, 0x0a, 0x10, 0x0c, 0x07, 0xfd, 0xf4, 0xf1, 0xf6, 0x01, 0x0a, 0x0c, 0x0b, 0x08, 0xff, 0xf4, 0xee, 0xef, 0xfb, 0x08, 0x0c, 0x0a, 0x04, 0xfa, 0xf3, 
0xf4, 0xfc, 0x05, 0x08, 0x08, 0x08, 0x03, 0xf9, 0xf1, 0xf0, 0xf8, 0x04, 0x0a, 0x0b, 0x07, 0xfc, 0xf2, 0xef, 0xf3, 0xfd, 0x07, 0x0f, 0x10, 0x08, 0xfc, 0xf3, 0xee, 0xee, 0xf6, 0x02, 0x0d, 0x11, 
0x08, 0xfc, 0xf3, 0xee, 0xf2, 0xfd, 0x08, 0x0f, 0x0d, 0x06, 0xff, 0xf7, 0xef, 0xef, 0xf9, 0x06, 0x0f, 0x0d, 0x04, 0xf9, 0xf2, 0xf1, 0xf5, 0xfc, 0x06, 0x0c, 0x0e, 0x0b, 0x03, 0xf6, 0xed, 0xf0, 
0xfc, 0x07, 0x0b, 0x0b, 0x06, 0xfd, 0xf6, 0xf3, 0xf9, 0x03, 0x0a, 0x0a, 0x06, 0x02, 0xfe, 0xf8, 0xf4, 0xf7, 0xff, 0x08, 0x0d, 0x0d, 0x04, 0xfa, 0xf1, 0xf0, 0xf6, 0x03, 0x0d, 0x11, 0x0f, 0x07, 
0xfa, 0xf1, 0xf1, 0xfa, 0x03, 0x08, 0x0b, 0x0a, 0x04, 0xf9, 0xf2, 0xf4, 0xfb, 0x06, 0x10, 0x15, 0x0f, 0x01, 0xf3, 0xed, 0xf0, 0xf9, 0x05, 0x0e, 0x11, 0x0b, 0xff, 0xf5, 0xf3, 0xf6, 0xfc, 0x05, 
0x0e, 0x12, 0x0e, 0x01, 0xf6, 0xef, 0xf0, 0xf9, 0x06, 0x0b, 0x0e, 0x09, 0x00, 0xf7, 0xf3, 0xf3, 0xfc, 0x08, 0x0f, 0x0e, 0x09, 0x00, 0xf7, 0xee, 0xef, 0xfa, 0x07, 0x0c, 0x0e, 0x08, 0xfd, 0xf5, 
0xf2, 0xf4, 0xfd, 0x09, 0x0d, 0x0c, 0x08, 0xfe, 0xf4, 0xef, 0xf3, 0xfe, 0x08, 0x0c, 0x0b, 0x03, 0xfa, 0xf6, 0xf3, 0xf6, 0xfe, 0x07, 0x0f, 0x11, 0x06, 0xf9, 0xf2, 0xf3, 0xf7, 0xff, 0x06, 0x0b, 
0x0a, 0x04, 0xf8, 0xef, 0xf2, 0xfb, 0x04, 0x0d, 0x10, 0x0d, 0x03, 0xf6, 0xed, 0xed, 0xf3, 0xff, 0x0b, 0x10, 0x0d, 0x05, 0xfa, 0xf2, 0xf1, 0xf8, 0x00, 0x08, 0x0a, 0x0a, 0x05, 0xfb, 0xf4, 0xf2, 
0xf6, 0xff, 0x06, 0x0a, 0x0b, 0x04, 0xf9, 0xf3, 0xf5, 0xf9, 0x00, 0x06, 0x0b, 0x0c, 0x06, 0xfc, 0xf4, 0xf0, 0xf6, 0x01, 0x08, 0x0b, 0x0a, 0x03, 0xf9, 0xf3, 0xf4, 0xfa, 0x03, 0x0a, 0x0d, 0x0b, 
0x02, 0xf8, 0xf1, 0xf0, 0xf9, 0x03, 0x09, 0x0c, 0x0a, 0x02, 0xfa, 0xf3, 0xf3, 0xf9, 0x03, 0x09, 0x0b, 0x0a, 0x03, 0xfa, 0xf5, 0xf5, 0xf8, 0x00, 0x07, 0x0b, 0x0a, 0x02, 0xf7, 0xf2, 0xf4, 0xfc, 
0x07, 0x0c, 0x0b, 0x08, 0x02, 0xf9, 0xf1, 0xf1, 0xf7, 0x02, 0x0b, 0x0f, 0x0c, 0x02, 0xf6, 0xef, 0xf3, 0xfc, 0x04, 0x0c, 0x0f, 0x0c, 0x02, 0xf7, 0xef, 0xf0, 0xf8, 0x04, 0x0b, 0x0d, 0x08, 0xfe, 
0xf4, 0xf3, 0xf7, 0xfd, 0x05, 0x0a, 0x0d, 0x0b, 0x02, 0xf6, 0xf0, 0xf0, 0xf8, 0x02, 0x09, 0x0b, 0x08, 0xff, 0xf9, 0xf5, 0xf5, 0xfc, 0x06, 0x0b, 0x0d, 0x09, 0x00, 0xf8, 0xf5, 0xf5, 0xf9, 0x01, 
0x09, 0x0b, 0x06, 0xff, 0xf8, 0xf6, 0xf9, 0xfe, 0x05, 0x08, 0x09, 0x05, 0xfe, 0xf6, 0xf2, 0xf6, 0x01, 0x09, 0x0c, 0x09, 0x01, 0xf9, 0xf5, 0xf5, 0xfb, 0x01, 0x05, 0x08, 0x09, 0x05, 0xfc, 0xf4, 
0xf1, 0xf8, 0x03, 0x08, 0x08, 0x08, 0x03, 0xfb, 0xf5, 0xf3, 0xf7, 0x00, 0x08, 0x0c, 0x0c, 0x07, 0xfd, 0xf3, 0xf1, 0xf8, 0xff, 0x06, 0x0a, 0x09, 0x04, 0xfe, 0xf8, 0xf6, 0xf9, 0xff, 0x07, 0x0a, 
0x08, 0x03, 0xfc, 0xf8, 0xf7, 0xf8, 0xfe, 0x03, 0x07, 0x08, 0x06, 0x00, 0xfb, 0xf7, 0xf9, 0xfe, 0x03, 0x05, 0x07, 0x06, 0x01, 0xfa, 0xf7, 0xf9, 0xff, 0x03, 0x04, 0x04, 0x02, 0xfe, 0xfc, 0xfd, 
0xfe, 0x01, 0x04, 0x05, 0x05, 0x03, 0xfe, 0xfa, 0xf9, 0xfa, 0xff, 0x03, 0x04, 0x05, 0x03, 0xff, 0xfb, 0xf9, 0xfc, 0x01, 0x04, 0x05, 0x07, 0x06, 0xff, 0xf9, 0xf8, 0xf9, 0xfd, 0x03, 0x07, 0x08, 
0x04, 0xff, 0xf9, 0xf7, 0xfc, 0x01, 0x04, 0x08, 0x08, 0x03, 0xfd, 0xf7, 0xf4, 0xf7, 0xff, 0x05, 0x07, 0x07, 0x06, 0x02, 0xfe, 0xfa, 0xf8, 0xfb, 0x01, 0x06, 0x08, 0x06, 0x02, 0xfc, 0xf7, 0xf7, 
0xfc, 0x01, 0x05, 0x05, 0x04, 0x02, 0xfe, 0xfa, 0xf9, 0xfc, 0x00, 0x04, 0x05, 0x03, 0xff, 0xfc, 0xfa, 0xfb, 0xff, 0x02, 0x02, 0x03, 0x01, 0xfc, 0xfa, 0xf8, 0xfb, 0x00, 0x05, 0x07, 0x05, 0x00, 
0xfa, 0xf5, 0xf4, 0xf8, 0x01, 0x06, 0x08, 0x06, 0x01, 0xfc, 0xf9, 0xf8, 0xf9, 0xff, 0x05, 0x07, 0x07, 0x04, 0xfe, 0xf8, 0xf6, 0xf9, 0xfc, 0x02, 0x07, 0x09, 0x07, 0x01, 0xfa, 0xf4, 0xf4, 0xf9, 
0x02, 0x08, 0x0a, 0x08, 0xff, 0xf7, 0xf2, 0xf3, 0xfb, 0x05, 0x0a, 0x0b, 0x08, 0x01, 0xf9, 0xf6, 0xf6, 0xfb, 0x03, 0x09, 0x09, 0x05, 0xff, 0xfa, 0xf6, 0xf6, 0xfb, 0x03, 0x08, 0x09, 0x06, 0xff, 
0xfc, 0xf9, 0xf7, 0xfb, 0x02, 0x07, 0x09, 0x07, 0x00, 0xf9, 0xf7, 0xf9, 0xfe, 0x04, 0x07, 0x06, 0x02, 0xfe, 0xfa, 0xf9, 0xfb, 0x00, 0x05, 0x07, 0x05, 0x01, 0xfb, 0xf7, 0xf7, 0xfb, 0x02, 0x08, 
0x0a, 0x07, 0x01, 0xfb, 0xf8, 0xf8, 0xfb, 0xff, 0x05, 0x09, 0x07, 0x04, 0xfd, 0xf7, 0xf6, 0xf9, 0xff, 0x07, 0x0a, 0x0a, 0x04, 0xfc, 0xf6, 0xf4, 0xf6, 0xfe, 0x07, 0x0c, 0x0b, 0x06, 0xfd, 0xf6, 
0xf1, 0xf6, 0xfe, 0x07, 0x0c, 0x0c, 0x07, 0xff, 0xf8, 0xf4, 0xf5, 0xfb, 0x03, 0x0a, 0x0b, 0x07, 0xff, 0xf7, 0xf3, 0xf7, 0xfd, 0x05, 0x09, 0x09, 0x03, 0xfd, 0xf7, 0xf4, 0xf7, 0xfe, 0x05, 0x08, 
0x09, 0x03, 0xf9, 0xf5, 0xf6, 0xfb, 0x02, 0x08, 0x09, 0x08, 0x02, 0xfa, 0xf6, 0xf5, 0xf9, 0xff, 0x06, 0x08, 0x05, 0xff, 0xfa, 0xf5, 0xf8, 0xfd, 0x03, 0x08, 0x07, 0x04, 0x00, 0xfb, 0xf7, 0xf6, 
0xfa, 0x02, 0x09, 0x09, 0x05, 0x00, 0xfa, 0xf5, 0xf5, 0xfa, 0x02, 0x08, 0x0a, 0x07, 0x03, 0xfd, 0xf7, 0xf6, 0xf9, 0xfe, 0x05, 0x09, 0x07, 0x02, 0xfc, 0xf8, 0xf8, 0xfb, 0x00, 0x05, 0x08, 0x07, 
0x03, 0xfe, 0xf9, 0xf8, 0xfb, 0x01, 0x07, 0x0a, 0x06, 0x00, 0xfb, 0xf7, 0xf7, 0xfc, 0x03, 0x08, 0x09, 0x07, 0x02, 0xfc, 0xf9, 0xf7, 0xfa, 0x02, 0x07, 0x09, 0x08, 0x03, 0xfc, 0xf6, 0xf6, 0xfb, 
0x01, 0x06, 0x07, 0x06, 0x03, 0xfd, 0xf8, 0xf6, 0xfc, 0x04, 0x07, 0x06, 0x04, 0x00, 0xfa, 0xf7, 0xf8, 0xfd, 0x04, 0x08, 0x07, 0x05, 0x00, 0xf8, 0xf4, 0xf5, 0xfe, 0x05, 0x08, 0x07, 0x05, 0xff, 
0xf9, 0xf5, 0xf7, 0xfe, 0x06, 0x09, 0x09, 0x05, 0xfe, 0xf8, 0xf4, 0xf6, 0xfd, 0x05, 0x08, 0x07, 0x04, 0xff, 0xf9, 0xf7, 0xf8, 0xfd, 0x03, 0x08, 0x08, 0x03, 0xfe, 0xfa, 0xf6, 0xf8, 0xfc, 0x01, 
0x07, 0x07, 0x03, 0xfe, 0xf9, 0xf7, 0xf9, 0xfe, 0x03, 0x07, 0x07, 0x06, 0x01, 0xf8, 0xf3, 0xf8, 0xff, 0x06, 0x09, 0x08, 0x03, 0xfd, 0xf7, 0xf5, 0xf8, 0xfe, 0x06, 0x0a, 0x08, 0x03, 0xfc, 0xf6, 
0xf5, 0xfa, 0x02, 0x08, 0x08, 0x05, 0x01, 0xfc, 0xf6, 0xf5, 0xf9, 0x00, 0x07, 0x0a, 0x08, 0x01, 0xfb, 0xf6, 0xf6, 0xfd, 0x05, 0x09, 0x09, 0x05, 0xfe, 0xf9, 0xf7, 0xf9, 0xfe, 0x03, 0x06, 0x09, 
0x07, 0x01, 0xfa, 0xf6, 0xf9, 0xff, 0x03, 0x06, 0x08, 0x04, 0xfe, 0xfa, 0xf8, 0xf9, 0xff, 0x04, 0x08, 0x09, 0x05, 0xff, 0xfa, 0xf8, 0xf9, 0xfd, 0x01, 0x05, 0x08, 0x07, 0x01, 0xfb, 0xf9, 0xf8, 
0xfc, 0x01, 0x04, 0x06, 0x05, 0x01, 0xfe, 0xfa, 0xf9, 0xfb, 0xff, 0x02, 0x03, 0x03, 0x00, 0xfd, 0xf9, 0xfa, 0xfe, 0x01, 0x04, 0x05, 0x03, 0xff, 0xfc, 0xf9, 0xf9, 0xfd, 0x01, 0x04, 0x05, 0x04, 
0x01, 0xfb, 0xf7, 0xf6, 0xfb, 0x02, 0x08, 0x09, 0x06, 0x00, 0xf9, 0xf6, 0xf9, 0xfe, 0x02, 0x05, 0x07, 0x04, 0x00, 0xfc, 0xf8, 0xf7, 0xfc, 0x03, 0x08, 0x08, 0x04, 0xff, 0xfc, 0xf9, 0xf8, 0xfc, 
0x00, 0x05, 0x08, 0x06, 0x01, 0xfc, 0xf9, 0xf8, 0xfc, 0x01, 0x07, 0x0a, 0x07, 0x01, 0xfd, 0xfa, 0xf8, 0xfb, 0xff, 0x04, 0x07, 0x06, 0x02, 0xfd, 0xf9, 0xfa, 0xfe, 0x01, 0x05, 0x08, 0x06, 0x03, 
0xfd, 0xf9, 0xf9, 0xfd, 0x01, 0x03, 0x05, 0x04, 0x01, 0xfc, 0xf7, 0xf8, 0xfd, 0x02, 0x07, 0x09, 0x05, 0xff, 0xfa, 0xf8, 0xf9, 0xfd, 0x02, 0x06, 0x08, 0x05, 0x00, 0xfa, 0xf7, 0xf8, 0xfe, 0x03, 
0x05, 0x05, 0x03, 0xff, 0xfb, 0xfa, 0xfc, 0xff, 0x02, 0x04, 0x03, 0x00, 0xfc, 0xf9, 0xf7, 0xfa, 0x00, 0x04, 0x05, 0x06, 0x03, 0xfd, 0xf9, 0xf7, 0xf9, 0xff, 0x04, 0x06, 0x06, 0x02, 0xfc, 0xf8, 
0xf7, 0xfa, 0xff, 0x03, 0x06, 0x05, 0x02, 0xfe, 0xfa, 0xf8, 0xfa, 0xff, 0x03, 0x05, 0x05, 0x02, 0xff, 0xfc, 0xfb, 0xfc, 0x00, 0x03, 0x03, 0x03, 0x01, 0xfd, 0xfb, 0xfd, 0xff, 0x01, 0x02, 0x04, 
0x04, 0x01, 0xfc, 0xf9, 0xfa, 0xff, 0x06, 0x08, 0x07, 0x03, 0xfe, 0xfb, 0xf9, 0xfb, 0xfe, 0x02, 0x05, 0x08, 0x06, 0x01, 0xfd, 0xfa, 0xf9, 0xfc, 0x02, 0x07, 0x07, 0x04, 0xff, 0xfb, 0xfb, 0xfc, 
0xfd, 0x01, 0x04, 0x06, 0x05, 0x00, 0xfb, 0xf7, 0xf8, 0xff, 0x06, 0x0a, 0x08, 0x04, 0x00, 0xfa, 0xf6, 0xf7, 0xfd, 0x03, 0x07, 0x06, 0x03, 0xfe, 0xf8, 0xf6, 0xf8, 0xfe, 0x05, 0x07, 0x05, 0x01, 
0xfd, 0xf9, 0xf8, 0xf9, 0xff, 0x03, 0x05, 0x04, 0x00, 0xfb, 0xf8, 0xf8, 0xfb, 0x01, 0x04, 0x07, 0x05, 0xff, 0xf9, 0xf7, 0xf8, 0xfc, 0x02, 0x07, 0x07, 0x04, 0x00, 0xfa, 0xf7, 0xf7, 0xfa, 0xff, 
0x05, 0x09, 0x08, 0x03, 0xfb, 0xf7, 0xf8, 0xfc, 0x00, 0x02, 0x05, 0x06, 0x02, 0xfd, 0xf9, 0xf8, 0xfa, 0x00, 0x06, 0x08, 0x06, 0x01, 0xfb, 0xf8, 0xfa, 0xfd, 0x00, 0x03, 0x05, 0x05, 0x02, 0xfc, 
0xf8, 0xf7, 0xfb, 0x02, 0x07, 0x08, 0x05, 0x01, 0xfc, 0xf7, 0xf6, 0xfa, 0x01, 0x05, 0x07, 0x06, 0x02, 0xfc, 0xf8, 0xf8, 0xfc, 0x01, 0x05, 0x07, 0x04, 0x00, 0xfc, 0xf8, 0xf9, 0xfe, 0x02, 0x06, 
0x08, 0x04, 0xfd, 0xf9, 0xf9, 0xfa, 0xfe, 0x03, 0x07, 0x08, 0x05, 0xff, 0xf8, 0xf8, 0xfb, 0xfe, 0x03, 0x06, 0x06, 0x03, 0xfe, 0xfb, 0xfa, 0xfc, 0xff, 0x04, 0x06, 0x05, 0x02, 0xfd, 0xfa, 0xfa, 
0xfb, 0xfe, 0x02, 0x05, 0x05, 0x03, 0xff, 0xfd, 0xfc, 0xfc, 0xfd, 0x02, 0x06, 0x07, 0x03, 0xff, 0xfd, 0xfc, 0xfc, 0xfe, 0x02, 0x05, 0x03, 0x00, 0xff, 0xfd, 0xfc, 0xfc, 0xfd, 0x01, 0x06, 0x06, 
0x04, 0x00, 0xfb, 0xfa, 0xfc, 0xfe, 0x01, 0x02, 0x03, 0x03, 0x00, 0xfc, 0xfc, 0xfd, 0xff, 0x03, 0x06, 0x05, 0x03, 0xfe, 0xf9, 0xfa, 0xfe, 0x01, 0x02, 0x04, 0x05, 0x02, 0xfe, 0xfa, 0xf9, 0xfc, 
0x00, 0x04, 0x08, 0x08, 0x04, 0xff, 0xfb, 0xfa, 0xfb, 0xfd, 0x01, 0x04, 0x05, 0x02, 0xff, 0xfd, 0xfc, 0xfc, 0xfe, 0x01, 0x03, 0x03, 0x01, 0xfe, 0xfd, 0xfe, 0x00, 0xff, 0x00, 0x00, 0xff, 0xff, 
0xff, 0xff, 0xfe, 0xfd, 0xfe, 0x00, 0x03, 0x03, 0x01, 0xfe, 0xfc, 0xfc, 0xfc, 0xfe, 0x00, 0x01, 0x01, 0x00, 0x00, 0xff, 0xfd, 0xfc, 0xfd, 0xff, 0x01, 0x02, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xfd, 
0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0xff, 0xfd, 0xfd, 0xff, 0x01, 0x00, 0xfe, 0xff, 0x01, 0x01, 0xfe, 0xfc, 0xfd, 0xff, 0x00, 0xff, 0xff, 0xff, 0xfe, 0xfd, 0xff, 0x01, 0x02, 0x01, 0xfd, 0xfd, 
0xfe, 0xfe, 0xfe, 0x01, 0x02, 0x01, 0x00, 0xfe, 0xfd, 0x02, 0x0b, 0x00, 0xf4, 0xfb, 0xfc, 0xf1, 0xf2, 0x02, 0x12, 0x16, 0x1a, 0x1f, 0x00, 0xd5, 0xd0, 0xd6, 0xee, 0x15, 0x37, 0x33, 0x23, 0x07, 
0xe4, 0xc9, 0xce, 0xd3, 0xe7, 0x17, 0x3f, 0x3e, 0x2f, 0x0f, 0xf3, 0xdc, 0xc7, 0xc1, 0xd3, 0xfe, 0x2d, 0x47, 0x47, 0x31, 0x08, 0xd7, 0xb9, 0xb7, 0xcf, 0xee, 0x16, 0x3c, 0x48, 0x39, 0x1f, 0xf3, 
0xcf, 0xb8, 0xbf, 0xe4, 0x0a, 0x25, 0x35, 0x3c, 0x31, 0x0c, 0xe3, 0xc4, 0xc5, 0xd5, 0xef, 0x08, 0x24, 0x38, 0x35, 0x1b, 0x04, 0xe5, 0xcc, 0xd0, 0xe3, 0xf1, 0x01, 0x22, 0x37, 0x2f, 0x19, 0xf9, 
0xdd, 0xcf, 0xcb, 0xd2, 0xf1, 0x1b, 0x3b, 0x39, 0x25, 0x13, 0xf4, 0xcd, 0xc1, 0xd3, 0xec, 0x0b, 0x24, 0x27, 0x21, 0x1e, 0x0e, 0xe7, 0xce, 0xd9, 0xe4, 0xf1, 0x08, 0x18, 0x20, 0x29, 0x25, 0x05, 
0xe6, 0xd9, 0xd5, 0xdc, 0xf2, 0x09, 0x1c, 0x2b, 0x2d, 0x17, 0xfb, 0xe9, 0xdc, 0xd2, 0xe4, 0x00, 0x0f, 0x1b, 0x2a, 0x26, 0x0d, 0xf3, 0xdb, 0xcd, 0xda, 0xee, 0x01, 0x1a, 0x30, 0x37, 0x21, 0xfe, 
0xde, 0xce, 0xd1, 0xe4, 0xfd, 0x18, 0x2f, 0x31, 0x1b, 0xfe, 0xe7, 0xd6, 0xd3, 0xe5, 0xf9, 0x0a, 0x1a, 0x2b, 0x2c, 0x12, 0xfe, 0xed, 0xd3, 0xcb, 0xe0, 0xfa, 0x15, 0x2c, 0x2c, 0x1d, 0x10, 0xf5, 
0xd1, 0xc6, 0xd8, 0xf6, 0x0f, 0x20, 0x29, 0x26, 0x1a, 0xfd, 0xde, 0xd1, 0xda, 0xf1, 0x07, 0x0d, 0x15, 0x1e, 0x1a, 0x06, 0xf2, 0xe1, 0xe7, 0xf8, 0xfe, 0x06, 0x0d, 0x08, 0x0c, 0x15, 0x0d, 0x02, 
0xf9, 0xea, 0xe7, 0xef, 0xf8, 0x05, 0x1a, 0x22, 0x16, 0x0a, 0xff, 0xe7, 0xdd, 0xe4, 0xf2, 0x06, 0x1b, 0x21, 0x1d, 0x15, 0x08, 0xf7, 0xe1, 0xd9, 0xe9, 0xf7, 0x05, 0x19, 0x24, 0x1f, 0x0f, 0xff, 
0xed, 0xdc, 0xdd, 0xea, 0xf6, 0x09, 0x1f, 0x2a, 0x25, 0x11, 0xf1, 0xdc, 0xda, 0xe1, 0xec, 0x01, 0x1f, 0x2c, 0x23, 0x0c, 0xf8, 0xe6, 0xdf, 0xe3, 0xf1, 0xf9, 0x0b, 0x1b, 0x21, 0x1a, 0x0c, 0xfb, 
0xe7, 0xde, 0xe3, 0xea, 0xfb, 0x12, 0x23, 0x22, 0x11, 0x05, 0xf0, 0xdf, 0xe4, 0xea, 0xef, 0x03, 0x15, 0x20, 0x26, 0x1a, 0xfe, 0xe4, 0xd7, 0xdf, 0xea, 0xef, 0x07, 0x1f, 0x28, 0x22, 0x0b, 0xf0, 
0xdf, 0xda, 0xe1, 0xf2, 0x02, 0x13, 0x1d, 0x23, 0x18, 0x05, 0xee, 0xdf, 0xe0, 0xe2, 0xe7, 0x00, 0x19, 0x2c, 0x2d, 0x18, 0xfa, 0xe5, 0xda, 0xd6, 0xdc, 0xf2, 0x0d, 0x27, 0x36, 0x28, 0x0c, 0xf5, 
0xe2, 0xd4, 0xd4, 0xe3, 0xfd, 0x19, 0x2f, 0x2c, 0x1c, 0x09, 0xeb, 0xd3, 0xd8, 0xe4, 0xf5, 0x0e, 0x1e, 0x27, 0x24, 0x11, 0xfa, 0xe7, 0xdc, 0xdc, 0xdf, 0xf2, 0x11, 0x26, 0x31, 0x28, 0x13, 0xf0, 
0xda, 0xd9, 0xde, 0xe3, 0xfa, 0x17, 0x30, 0x34, 0x1f, 0xfd, 0xe7, 0xdc, 0xd5, 0xd9, 0xf1, 0x11, 0x29, 0x2e, 0x24, 0x14, 0xfb, 0xde, 0xd5, 0xdc, 0xe7, 0xfe, 0x14, 0x20, 0x29, 0x21, 0x0c, 0xf6, 
0xe4, 0xd7, 0xd7, 0xe5, 0xfe, 0x1d, 0x2c, 0x26, 0x1e, 0x05, 0xea, 0xde, 0xdd, 0xdc, 0xec, 0x07, 0x21, 0x2b, 0x22, 0x0d, 0xfc, 0xe9, 0xe1, 0xe0, 0xe7, 0xfc, 0x10, 0x1d, 0x21, 0x20, 0x0f, 0xf4, 
0xe6, 0xe6, 0xe4, 0xe7, 0xf5, 0x0b, 0x1e, 0x22, 0x1a, 0x0b, 0xf7, 0xea, 0xe1, 0xdc, 0xec, 0x05, 0x17, 0x24, 0x27, 0x1a, 0x03, 0xef, 0xdf, 0xd4, 0xd9, 0xed, 0x09, 0x23, 0x29, 0x23, 0x18, 0x02, 
0xe4, 0xd1, 0xca, 0xd8, 0xfb, 0x20, 0x31, 0x30, 0x23, 0x09, 0xe8, 0xd1, 0xcb, 0xd1, 0xeb, 0x0d, 0x2a, 0x37, 0x35, 0x1d, 0xf7, 0xd5, 0xc6, 0xc5, 0xda, 0x00, 0x26, 0x3a, 0x37, 0x24, 0xff, 0xda, 
0xcf, 0xcf, 0xd1, 0xec, 0x14, 0x32, 0x33, 0x27, 0x11, 0xf5, 0xe2, 0xda, 0xd4, 0xdc, 0xfa, 0x18, 0x28, 0x29, 0x21, 0x0c, 0xf3, 0xe2, 0xd5, 0xcd, 0xe0, 0x06, 0x26, 0x2e, 0x2f, 0x25, 0x0b, 0xed, 
0xd5, 0xc6, 0xce, 0xed, 0x0f, 0x2f, 0x40, 0x37, 0x19, 0xf2, 0xd5, 0xc9, 0xc9, 0xe0, 0x06, 0x28, 0x3a, 0x36, 0x20, 0x00, 0xe2, 0xd3, 0xd1, 0xdc, 0xf8, 0x15, 0x2a, 0x30, 0x24, 0x09, 0xf3, 0xe7, 
0xde, 0xde, 0xe8, 0xfc, 0x13, 0x1f, 0x20, 0x1c, 0x0f, 0xfb, 0xec, 0xe0, 0xd9, 0xe2, 0xf7, 0x0f, 0x22, 0x28, 0x1f, 0x0f, 0xfb, 0xe7, 0xd8, 0xd4, 0xe2, 0xff, 0x1d, 0x2b, 0x31, 0x23, 0x01, 0xe7, 
0xd5, 0xc8, 0xd3, 0xf3, 0x14, 0x2e, 0x3b, 0x30, 0x0e, 0xf0, 0xd9, 0xca, 0xc8, 0xe3, 0x07, 0x26, 0x33, 0x30, 0x1e, 0x00, 0xe2, 0xd2, 0xcc, 0xd9, 0xf5, 0x13, 0x27, 0x2f, 0x24, 0x0d, 0xf7, 0xe4, 
0xd6, 0xd6, 0xe1, 0xf2, 0x0b, 0x25, 0x2e, 0x28, 0x14, 0xfc, 0xe3, 0xcd, 0xc6, 0xdb, 0xfa, 0x1d, 0x36, 0x3b, 0x27, 0x09, 0xe5, 0xcb, 0xc3, 0xd0, 0xee, 0x14, 0x2d, 0x3b, 0x32, 0x15, 0xf0, 0xd8, 
0xc9, 0xc8, 0xe2, 0x09, 0x25, 0x33, 0x2f, 0x1d, 0x02, 0xe6, 0xd6, 0xd1, 0xdd, 0xfa, 0x11, 0x1e, 0x24, 0x1f, 0x0e, 0xfd, 0xf0, 0xe2, 0xdb, 0xe5, 0xf5, 0x07, 0x19, 0x1e, 0x1b, 0x16, 0x0a, 0xf7, 
0xe7, 0xdc, 0xde, 0xed, 0x02, 0x16, 0x27, 0x2b, 0x20, 0x0a, 0xed, 0xd7, 0xd0, 0xd9, 0xf3, 0x12, 0x27, 0x2d, 0x25, 0x0e, 0xf5, 0xe0, 0xd2, 0xd8, 0xf0, 0x08, 0x1b, 0x24, 0x21, 0x12, 0xff, 0xee, 
0xe3, 0xe1, 0xea, 0xfd, 0x0d, 0x13, 0x17, 0x15, 0x0c, 0x02, 0xf6, 0xe7, 0xe6, 0xed, 0xf4, 0x01, 0x0e, 0x16, 0x18, 0x14, 0x0a, 0xfb, 0xeb, 0xe1, 0xdf, 0xee, 0x02, 0x14, 0x23, 0x26, 0x19, 0x02, 
0xe8, 0xd9, 0xd5, 0xde, 0xf6, 0x14, 0x28, 0x2f, 0x26, 0x0c, 0xf2, 0xda, 0xd0, 0xd9, 0xf2, 0x0d, 0x20, 0x28, 0x23, 0x11, 0xf9, 0xe4, 0xd8, 0xde, 0xf1, 0x05, 0x12, 0x18, 0x19, 0x15, 0x09, 0xf9, 
0xe8, 0xe2, 0xe7, 0xf4, 0xfe, 0x08, 0x11, 0x14, 0x13, 0x0c, 0xfc, 0xf0, 0xe9, 0xe1, 0xe8, 0xfc, 0x0b, 0x19, 0x22, 0x1b, 0x0e, 0xfb, 0xe3, 0xd3, 0xd6, 0xec, 0x06, 0x1a, 0x29, 0x2a, 0x18, 0xfe, 
0xe5, 0xd3, 0xcf, 0xe4, 0xff, 0x16, 0x23, 0x27, 0x1d, 0x05, 0xe9, 0xdd, 0xdd, 0xe7, 0xf8, 0x07, 0x11, 0x20, 0x23, 0x0f, 0xfd, 0xf1, 0xe6, 0xe7, 0xef, 0xf6, 0x04, 0x11, 0x16, 0x16, 0x13, 0x05, 
0xf4, 0xe9, 0xe3, 0xeb, 0xf9, 0x07, 0x15, 0x1e, 0x1e, 0x13, 0x01, 0xed, 0xdf, 0xde, 0xe9, 0xfd, 0x0f, 0x1f, 0x28, 0x23, 0x0d, 0xf3, 0xdd, 0xd1, 0xd8, 0xf0, 0x0c, 0x24, 0x32, 0x28, 0x0f, 0xf4, 
0xdf, 0xd4, 0xd8, 0xee, 0x07, 0x18, 0x28, 0x2c, 0x19, 0x00, 0xeb, 0xdf, 0xda, 0xe3, 0xf6, 0x0b, 0x18, 0x21, 0x1d, 0x11, 0x00, 0xef, 0xe0, 0xdd, 0xe6, 0xf5, 0x06, 0x1b, 0x28, 0x22, 0x13, 0xfd, 
0xe5, 0xd6, 0xd6, 0xe1, 0xfa, 0x16, 0x2c, 0x31, 0x24, 0x0c, 0xef, 0xd6, 0xc9, 0xd5, 0xec, 0x0a, 0x28, 0x3a, 0x33, 0x17, 0xf3, 0xd8, 0xcb, 0xcd, 0xdf, 0xfe, 0x1f, 0x33, 0x34, 0x25, 0x0c, 0xec, 
0xd2, 0xc9, 0xd8, 0xef, 0x0a, 0x1f, 0x2f, 0x2d, 0x1c, 0x00, 0xe6, 0xd6, 0xd3, 0xdf, 0xf4, 0x0b, 0x20, 0x2e, 0x2a, 0x19, 0xff, 0xe4, 0xd1, 0xd2, 0xdf, 0xf4, 0x0e, 0x2b, 0x36, 0x2b, 0x11, 0xf2, 
0xd9, 0xcb, 0xd1, 0xe6, 0x04, 0x20, 0x34, 0x34, 0x23, 0x05, 0xe4, 0xcb, 0xc7, 0xd9, 0xf6, 0x19, 0x33, 0x37, 0x25, 0x0b, 0xf0, 0xd7, 0xc7, 0xd3, 0xee, 0x09, 0x21, 0x2f, 0x2b, 0x1e, 0x06, 0xe7, 
0xd0, 0xd0, 0xe0, 0xf4, 0x0a, 0x23, 0x31, 0x27, 0x14, 0xfc, 0xe3, 0xd0, 0xd0, 0xdf, 0xf8, 0x13, 0x29, 0x31, 0x29, 0x14, 0xf4, 0xd9, 0xcb, 0xce, 0xe2, 0x01, 0x1f, 0x34, 0x34, 0x21, 0x03, 0xe5, 
0xcc, 0xc4, 0xd3, 0xf3, 0x13, 0x2b, 0x35, 0x2e, 0x16, 0xf4, 0xd7, 0xc7, 0xca, 0xe1, 0x02, 0x1c, 0x2f, 0x32, 0x24, 0x0b, 0xee, 0xd1, 0xc4, 0xd1, 0xed, 0x0a, 0x26, 0x37, 0x31, 0x1a, 0xfe, 0xe0, 
0xcb, 0xca, 0xd8, 0xf3, 0x13, 0x2d, 0x34, 0x29, 0x14, 0xf7, 0xdc, 0xcc, 0xd1, 0xe4, 0xfd, 0x17, 0x2e, 0x35, 0x2b, 0x0f, 0xee, 0xd3, 0xca, 0xd1, 0xe8, 0x06, 0x23, 0x37, 0x35, 0x21, 0x03, 0xe6, 
0xcc, 0xc7, 0xd4, 0xf1, 0x15, 0x31, 0x3b, 0x32, 0x16, 0xf2, 0xd3, 0xc3, 0xcc, 0xea, 0x0b, 0x26, 0x37, 0x32, 0x1e, 0x03, 0xe5, 0xce, 0xcc, 0xdd, 0xf8, 0x10, 0x26, 0x32, 0x2d, 0x18, 0xfb, 0xe0, 
0xcd, 0xd0, 0xdf, 0xfa, 0x18, 0x30, 0x36, 0x2d, 0x13, 0xf3, 0xd4, 0xc8, 0xcd, 0xe2, 0x00, 0x23, 0x3b, 0x3b, 0x27, 0x06, 0xe3, 0xc9, 0xc5, 0xd5, 0xf4, 0x16, 0x30, 0x39, 0x31, 0x17, 0xf4, 0xd6, 
0xc6, 0xcd, 0xe9, 0x07, 0x1e, 0x30, 0x35, 0x24, 0x07, 0xea, 0xd2, 0xcc, 0xd7, 0xef, 0x09, 0x20, 0x32, 0x31, 0x1c, 0xfe, 0xe1, 0xcc, 0xcc, 0xdf, 0xfc, 0x15, 0x2c, 0x37, 0x2a, 0x10, 0xf4, 0xd5, 
0xc6, 0xce, 0xe6, 0x04, 0x23, 0x35, 0x35, 0x24, 0x06, 0xe2, 0xc8, 0xc5, 0xd5, 0xf4, 0x13, 0x2c, 0x37, 0x2e, 0x16, 0xf9, 0xdb, 0xc9, 0xcb, 0xde, 0xfd, 0x1b, 0x30, 0x35, 0x29, 0x0c, 0xeb, 0xd1, 
0xc7, 0xd1, 0xeb, 0x09, 0x23, 0x33, 0x32, 0x1d, 0xfe, 0xe2, 0xd2, 0xcc, 0xd6, 0xf2, 0x0f, 0x27, 0x36, 0x30, 0x17, 0xf8, 0xd9, 0xc9, 0xcb, 0xdc, 0xf8, 0x16, 0x2f, 0x3a, 0x2e, 0x12, 0xf2, 0xd2, 
0xc0, 0xc4, 0xdf, 0x05, 0x28, 0x3f, 0x40, 0x26, 0x02, 0xdc, 0xc1, 0xbd, 0xd1, 0xf4, 0x19, 0x38, 0x45, 0x37, 0x16, 0xef, 0xce, 0xbe, 0xc5, 0xdf, 0x03, 0x25, 0x3c, 0x3c, 0x25, 0x04, 0xe5, 0xcf, 
0xc8, 0xd4, 0xf0, 0x0c, 0x24, 0x34, 0x33, 0x21, 0x07, 0xe7, 0xce, 0xc7, 0xd3, 0xee, 0x0d, 0x2d, 0x40, 0x3c, 0x20, 0xf9, 0xd2, 0xbd, 0xc1, 0xdc, 0x00, 0x24, 0x3d, 0x41, 0x2d, 0x0c, 0xe8, 0xc9, 
0xbc, 0xc8, 0xe8, 0x10, 0x34, 0x45, 0x3b, 0x1c, 0xf8, 0xd6, 0xc2, 0xc1, 0xd7, 0xfc, 0x1e, 0x36, 0x3c, 0x2b, 0x0c, 0xeb, 0xd1, 0xc5, 0xca, 0xe3, 0x03, 0x1f, 0x35, 0x3b, 0x26, 0x07, 0xe8, 0xcd, 
0xc1, 0xcd, 0xe9, 0x09, 0x27, 0x39, 0x37, 0x22, 0x03, 0xdf, 0xc5, 0xbf, 0xd0, 0xf0, 0x16, 0x38, 0x47, 0x38, 0x19, 0xf2, 0xcc, 0xb7, 0xbf, 0xde, 0x06, 0x2d, 0x47, 0x45, 0x2a, 0x04, 0xdb, 0xbd, 
0xb7, 0xcd, 0xf3, 0x1f, 0x40, 0x4b, 0x35, 0x10, 0xec, 0xcf, 0xc0, 0xc7, 0xe4, 0x0a, 0x29, 0x3a, 0x3a, 0x26, 0x05, 0xe4, 0xcf, 0xcb, 0xd6, 0xef, 0x0c, 0x28, 0x3a, 0x37, 0x1f, 0xfe, 0xde, 0xca, 
0xc7, 0xd7, 0xf5, 0x16, 0x31, 0x3e, 0x33, 0x19, 0xf5, 0xd4, 0xc2, 0xc6, 0xdd, 0x01, 0x27, 0x41, 0x41, 0x2a, 0x07, 0xe0, 0xc2, 0xba, 0xce, 0xf3, 0x19, 0x36, 0x43, 0x33, 0x12, 0xec, 0xcf, 0xc1, 
0xc9, 0xe4, 0x06, 0x24, 0x38, 0x38, 0x21, 0x02, 0xe5, 0xd0, 0xc8, 0xd4, 0xef, 0x0b, 0x24, 0x34, 0x30, 0x1a, 0xfd, 0xe1, 0xcf, 0xcc, 0xda, 0xf2, 0x0e, 0x28, 0x36, 0x2e, 0x18, 0xf9, 0xd9, 0xc7, 
0xcc, 0xe0, 0xfc, 0x1a, 0x34, 0x3b, 0x2a, 0x0d, 0xea, 0xcf, 0xc4, 0xcf, 0xea, 0x0d, 0x2b, 0x3c, 0x35, 0x1e, 0xfc, 0xd7, 0xc2, 0xc4, 0xdb, 0xfe, 0x20, 0x35, 0x39, 0x26, 0x08, 0xe9, 0xd2, 0xc8, 
0xd2, 0xeb, 0x0c, 0x27, 0x36, 0x30, 0x1c, 0xfe, 0xe3, 0xd3, 0xd2, 0xdf, 0xf3, 0x08, 0x1f, 0x30, 0x2d, 0x1b, 0x00, 0xe5, 0xd2, 0xcd, 0xd9, 0xf2, 0x13, 0x2f, 0x38, 0x2e, 0x17, 0xf3, 0xd2, 0xc4, 
0xcd, 0xe6, 0x07, 0x27, 0x3b, 0x37, 0x22, 0x01, 0xdf, 0xc9, 0xc7, 0xd8, 0xf9, 0x1b, 0x31, 0x34, 0x27, 0x10, 0xf2, 0xd7, 0xcb, 0xd2, 0xe8, 0x08, 0x23, 0x31, 0x2d, 0x1a, 0x00, 0xe9, 0xd9, 0xd5, 
0xde, 0xf3, 0x0b, 0x1f, 0x2b, 0x28, 0x18, 0xfd, 0xe4, 0xd5, 0xd5, 0xe1, 0xf6, 0x0e, 0x25, 0x2f, 0x27, 0x13, 0xf8, 0xe0, 0xd1, 0xd2, 0xe4, 0x01, 0x1b, 0x2c, 0x2f, 0x21, 0x08, 0xec, 0xd4, 0xcd, 
0xd9, 0xf2, 0x0e, 0x26, 0x2f, 0x28, 0x14, 0xf9, 0xe0, 0xcf, 0xd0, 0xe2, 0xff, 0x1a, 0x2d, 0x2e, 0x21, 0x08, 0xea, 0xd6, 0xd3, 0xdd, 0xf0, 0x08, 0x1f, 0x2a, 0x26, 0x15, 0xff, 0xea, 0xdc, 0xd8, 
0xe1, 0xf6, 0x0d, 0x21, 0x2b, 0x27, 0x16, 0xfc, 0xe3, 0xd6, 0xd6, 0xe3, 0xfa, 0x12, 0x26, 0x2d, 0x25, 0x0e, 0xf3, 0xdb, 0xcf, 0xd4, 0xe8, 0x05, 0x21, 0x30, 0x2e, 0x1d, 0x00, 0xe3, 0xd0, 0xcf, 
0xdf, 0xf9, 0x15, 0x2a, 0x2d, 0x22, 0x0d, 0xf2, 0xdc, 0xd5, 0xdc, 0xee, 0x05, 0x1b, 0x26, 0x23, 0x17, 0x03, 0xee, 0xe0, 0xdd, 0xe4, 0xf3, 0x07, 0x1a, 0x22, 0x1f, 0x15, 0x03, 0xef, 0xdf, 0xd8, 
0xe0, 0xf6, 0x0c, 0x20, 0x29, 0x25, 0x14, 0xfa, 0xe0, 0xd3, 0xd5, 0xe5, 0xfd, 0x17, 0x2a, 0x30, 0x24, 0x09, 0xea, 0xd5, 0xcf, 0xd9, 0xf0, 0x0c, 0x24, 0x2e, 0x29, 0x17, 0xfa, 0xdd, 0xd1, 0xd5, 
0xe7, 0x02, 0x1a, 0x25, 0x23, 0x1a, 0x07, 0xf1, 0xe0, 0xd8, 0xde, 0xf1, 0x07, 0x19, 0x20, 0x1f, 0x17, 0x05, 0xef, 0xe0, 0xdb, 0xe2, 0xf3, 0x07, 0x1a, 0x25, 0x23, 0x14, 0xff, 0xea, 0xdb, 0xd8, 
0xe3, 0xf9, 0x12, 0x24, 0x2a, 0x22, 0x0c, 0xf1, 0xde, 0xd5, 0xd9, 0xeb, 0x04, 0x1b, 0x28, 0x28, 0x1a, 0x02, 0xe8, 0xd7, 0xd4, 0xe1, 0xf8, 0x10, 0x21, 0x27, 0x23, 0x11, 0xf5, 0xe0, 0xda, 0xdf, 
0xed, 0x00, 0x13, 0x20, 0x23, 0x1c, 0x0a, 0xf5, 0xe3, 0xdb, 0xdf, 0xef, 0x05, 0x17, 0x21, 0x24, 0x18, 0x04, 0xf0, 0xe0, 0xd8, 0xdf, 0xf3, 0x0b, 0x1e, 0x29, 0x27, 0x14, 0xfb, 0xe3, 0xd5, 0xd5, 
0xe5, 0xfd, 0x15, 0x28, 0x2e, 0x23, 0x08, 0xeb, 0xd7, 0xd1, 0xdc, 0xf2, 0x0c, 0x23, 0x2e, 0x29, 0x15, 0xfa, 0xe1, 0xd5, 0xd7, 0xe6, 0xfe, 0x17, 0x27, 0x2b, 0x1f, 0x07, 0xec, 0xdc, 0xd8, 0xe0, 
0xf2, 0x0b, 0x1c, 0x23, 0x22, 0x15, 0x01, 0xed, 0xdf, 0xdb, 0xe6, 0xf8, 0x0c, 0x1b, 0x22, 0x1f, 0x11, 0xfc, 0xe9, 0xdc, 0xdb, 0xe9, 0x00, 0x14, 0x20, 0x24, 0x1c, 0x09, 0xf2, 0xe0, 0xd8, 0xdd, 
0xef, 0x06, 0x1a, 0x27, 0x26, 0x14, 0xfc, 0xe8, 0xdb, 0xd8, 0xe6, 0xfd, 0x14, 0x23, 0x26, 0x1c, 0x08, 0xf0, 0xdf, 0xdb, 0xe3, 0xf2, 0x07, 0x18, 0x21, 0x20, 0x13, 0x00, 0xec, 0xe0, 0xde, 0xe6, 
0xf8, 0x0c, 0x19, 0x1e, 0x1d, 0x10, 0xfc, 0xe9, 0xde, 0xdf, 0xe9, 0xfa, 0x0c, 0x1a, 0x23, 0x1e, 0x0c, 0xf7, 0xe5, 0xda, 0xdc, 0xec, 0x03, 0x17, 0x23, 0x25, 0x1b, 0x04, 0xec, 0xdc, 0xd7, 0xe1, 
0xf5, 0x0d, 0x20, 0x2b, 0x25, 0x11, 0xf7, 0xe2, 0xd8, 0xdb, 0xeb, 0x02, 0x17, 0x24, 0x26, 0x1a, 0x04, 0xed, 0xe0, 0xdc, 0xe4, 0xf5, 0x09, 0x18, 0x21, 0x20, 0x13, 0xff, 0xec, 0xdf, 0xdd, 0xe7, 
0xfb, 0x0d, 0x19, 0x20, 0x1d, 0x0f, 0xfb, 0xea, 0xdf, 0xde, 0xe7, 0xf9, 0x0f, 0x20, 0x27, 0x1f, 0x0d, 0xf5, 0xe0, 0xd5, 0xda, 0xee, 0x07, 0x1c, 0x2a, 0x29, 0x17, 0xfc, 0xe5, 0xd8, 0xd8, 0xe6, 
0xfb, 0x11, 0x21, 0x27, 0x1e, 0x0a, 0xf3, 0xe2, 0xda, 0xe0, 0xf0, 0x05, 0x16, 0x21, 0x21, 0x15, 0x01, 0xef, 0xe3, 0xe0, 0xe7, 0xf7, 0x09, 0x17, 0x1f, 0x1d, 0x10, 0xfd, 0xec, 0xe1, 0xdd, 0xe5, 
0xf8, 0x0c, 0x1b, 0x22, 0x1d, 0x0e, 0xfa, 0xe7, 0xdc, 0xdc, 0xe9, 0x00, 0x14, 0x23, 0x27, 0x1c, 0x06, 0xf0, 0xde, 0xd7, 0xe0, 0xf4, 0x0c, 0x1f, 0x27, 0x22, 0x11, 0xf8, 0xe4, 0xd9, 0xdb, 0xeb, 
0x02, 0x16, 0x23, 0x22, 0x16, 0x01, 0xee, 0xe3, 0xe1, 0xe8, 0xf6, 0x07, 0x13, 0x1a, 0x19, 0x10, 0x01, 0xf1, 0xe6, 0xe5, 0xea, 0xf7, 0x05, 0x12, 0x1b, 0x1a, 0x10, 0xff, 0xef, 0xe4, 0xe2, 0xe8, 
0xf8, 0x0a, 0x1b, 0x22, 0x1d, 0x0e, 0xfb, 0xe7, 0xdd, 0xdf, 0xed, 0x02, 0x15, 0x20, 0x21, 0x16, 0x01, 0xee, 0xe1, 0xe0, 0xe9, 0xfb, 0x0d, 0x1a, 0x1e, 0x18, 0x09, 0xf7, 0xe9, 0xe5, 0xe7, 0xf0, 
0x01, 0x0f, 0x17, 0x19, 0x12, 0x04, 0xf6, 0xeb, 0xe7, 0xeb, 0xf6, 0x01, 0x0e, 0x19, 0x19, 0x10, 0x01, 0xf3, 0xe8, 0xe5, 0xea, 0xf7, 0x06, 0x15, 0x1e, 0x1b, 0x0d, 0xfd, 0xed, 0xe2, 0xe1, 0xec, 
0xfd, 0x0f, 0x1d, 0x20, 0x17, 0x07, 0xf4, 0xe3, 0xdd, 0xe4, 0xf5, 0x0a, 0x1a, 0x22, 0x1d, 0x0d, 0xfa, 0xe9, 0xe0, 0xe2, 0xed, 0xfe, 0x0f, 0x1a, 0x1a, 0x12, 0x04, 0xf6, 0xea, 0xe5, 0xe7, 0xf2, 
0x02, 0x0f, 0x19, 0x1a, 0x12, 0x03, 0xf3, 0xe9, 0xe5, 0xe7, 0xf3, 0x04, 0x12, 0x1a, 0x1a, 0x10, 0x00, 0xef, 0xe4, 0xe1, 0xe9, 0xfa, 0x0b, 0x19, 0x1f, 0x19, 0x07, 0xf5, 0xe7, 0xe0, 0xe2, 0xf1, 
0x04, 0x16, 0x1d, 0x1b, 0x10, 0xff, 0xef, 0xe6, 0xe4, 0xeb, 0xfc, 0x0b, 0x15, 0x18, 0x13, 0x07, 0xf9, 0xef, 0xe9, 0xe8, 0xf0, 0xfe, 0x0b, 0x14, 0x17, 0x15, 0x0a, 0xfb, 0xef, 0xe8, 0xe9, 0xf2, 
0xfe, 0x0c, 0x19, 0x1e, 0x16, 0x06, 0xf6, 0xea, 0xe3, 0xe7, 0xf6, 0x08, 0x16, 0x1d, 0x1a, 0x10, 0x00, 0xf0, 0xe4, 0xe3, 0xed, 0xfd, 0x0d, 0x19, 0x1e, 0x18, 0x09, 0xf8, 0xe9, 0xe1, 0xe5, 0xf4, 
0x06, 0x14, 0x1b, 0x1a, 0x0f, 0x00, 0xf2, 0xe8, 0xe5, 0xec, 0xfa, 0x08, 0x14, 0x1a, 0x15, 0x0a, 0xfc, 0xf0, 0xe8, 0xe5, 0xed, 0xfc, 0x09, 0x15, 0x1b, 0x15, 0x09, 0xfb, 0xec, 0xe3, 0xe4, 0xee, 
0xfe, 0x11, 0x1e, 0x1f, 0x14, 0x04, 0xf2, 0xe4, 0xdf, 0xe7, 0xf7, 0x09, 0x18, 0x1e, 0x19, 0x0b, 0xfa, 0xeb, 0xe2, 0xe2, 0xed, 0xff, 0x10, 0x1c, 0x1d, 0x14, 0x05, 0xf5, 0xe8, 0xe2, 0xe7, 0xf5, 
0x06, 0x13, 0x1b, 0x19, 0x0e, 0xfe, 0xf0, 0xe7, 0xe6, 0xed, 0xfa, 0x0a, 0x17, 0x1c, 0x15, 0x0a, 0xfd, 0xef, 0xe5, 0xe6, 0xf0, 0xfe, 0x0d, 0x18, 0x1c, 0x15, 0x07, 0xf7, 0xe9, 0xe3, 0xe7, 0xf4, 
0x04, 0x13, 0x1d, 0x1b, 0x10, 0x00, 0xf0, 0xe4, 0xe1, 0xea, 0xfa, 0x0b, 0x18, 0x1d, 0x16, 0x09, 0xf9, 0xeb, 0xe2, 0xe5, 0xf1, 0x01, 0x11, 0x1a, 0x19, 0x10, 0x02, 0xf4, 0xe8, 0xe4, 0xeb, 0xf8, 
0x06, 0x13, 0x1a, 0x18, 0x0c, 0xfd, 0xef, 0xe7, 0xe6, 0xef, 0xfd, 0x0b, 0x16, 0x19, 0x12, 0x07, 0xf9, 0xec, 0xe5, 0xe8, 0xf3, 0x01, 0x0f, 0x19, 0x19, 0x10, 0x01, 0xf3, 0xe7, 0xe3, 0xea, 0xf7, 
0x08, 0x16, 0x1b, 0x15, 0x0a, 0xfb, 0xed, 0xe4, 0xe6, 0xef, 0xfe, 0x0c, 0x18, 0x1a, 0x12, 0x04, 0xf5, 0xea, 0xe7, 0xeb, 0xf4, 0x02, 0x10, 0x18, 0x17, 0x0e, 0x01, 0xf3, 0xe8, 0xe6, 0xeb, 0xf7, 
0x06, 0x13, 0x19, 0x17, 0x0d, 0xfe, 0xef, 0xe4, 0xe3, 0xee, 0xfe, 0x0d, 0x19, 0x1b, 0x15, 0x07, 0xf6, 0xe6, 0xe1, 0xe7, 0xf5, 0x05, 0x13, 0x1b, 0x19, 0x0e, 0xff, 0xf0, 0xe5, 0xe4, 0xed, 0xfb, 
0x0a, 0x17, 0x1b, 0x17, 0x0a, 0xfb, 0xed, 0xe6, 0xe8, 0xf1, 0xff, 0x0d, 0x18, 0x19, 0x12, 0x07, 0xf8, 0xeb, 0xe7, 0xea, 0xf5, 0x03, 0x10, 0x17, 0x19, 0x13, 0x05, 0xf4, 0xe8, 0xe6, 0xec, 0xf9, 
0x08, 0x14, 0x19, 0x16, 0x0c, 0xfe, 0xef, 0xe6, 0xe7, 0xf1, 0xff, 0x0e, 0x18, 0x1a, 0x15, 0x08, 0xf8, 0xeb, 0xe5, 0xe8, 0xf4, 0x05, 0x14, 0x1a, 0x18, 0x0f, 0x01, 0xf3, 0xe9, 0xe7, 0xef, 0xfc, 
0x09, 0x14, 0x17, 0x14, 0x0b, 0xfe, 0xf1, 0xea, 0xeb, 0xf2, 0xfe, 0x0c, 0x14, 0x15, 0x11, 0x08, 0xfa, 0xef, 0xea, 0xec, 0xf5, 0x02, 0x0d, 0x15, 0x16, 0x0f, 0x03, 0xf5, 0xeb, 0xe7, 0xeb, 0xf7, 
0x06, 0x12, 0x17, 0x15, 0x0b, 0xfc, 0xef, 0xe6, 0xe6, 0xf1, 0xff, 0x0d, 0x16, 0x18, 0x10, 0x03, 0xf5, 0xea, 0xe6, 0xeb, 0xf7, 0x05, 0x11, 0x16, 0x14, 0x0b, 0xfe, 0xf1, 0xe9, 0xe8, 0xef, 0xf9, 
0x05, 0x11, 0x16, 0x12, 0x08, 0xfc, 0xf0, 0xe8, 0xe8, 0xf0, 0xfd, 0x0c, 0x15, 0x17, 0x10, 0x04, 0xf7, 0xed, 0xe9, 0xec, 0xf4, 0x00, 0x0e, 0x16, 0x16, 0x0d, 0x01, 0xf2, 0xe8, 0xe6, 0xed, 0xfa, 
0x09, 0x14, 0x19, 0x16, 0x0a, 0xf9, 0xeb, 0xe4, 0xe7, 0xf3, 0x02, 0x10, 0x18, 0x18, 0x0f, 0x01, 0xf3, 0xe9, 0xe7, 0xee, 0xfb, 0x09, 0x12, 0x15, 0x12, 0x0a, 0xfe, 0xf2, 0xeb, 0xea, 0xf1, 0xfb, 
0x08, 0x11, 0x15, 0x12, 0x09, 0xfc, 0xf0, 0xe9, 0xe9, 0xf2, 0x00, 0x0e, 0x16, 0x19, 0x11, 0x04, 0xf4, 0xe9, 0xe6, 0xec, 0xf8, 0x08, 0x13, 0x18, 0x15, 0x0b, 0xfd, 0xef, 0xe8, 0xea, 0xf4, 0x01, 
0x0d, 0x13, 0x15, 0x10, 0x05, 0xf8, 0xef, 0xeb, 0xee, 0xf8, 0x05, 0x0e, 0x13, 0x13, 0x0d, 0x02, 0xf6, 0xee, 0xec, 0xf0, 0xfa, 0x06, 0x0e, 0x13, 0x12, 0x0a, 0x00, 0xf4, 0xec, 0xea, 0xf1, 0xfc, 
0x08, 0x11, 0x16, 0x13, 0x0a, 0xfc, 0xef, 0xe8, 0xea, 0xf3, 0x01, 0x0d, 0x16, 0x16, 0x0e, 0x01, 0xf5, 0xeb, 0xe7, 0xed, 0xfa, 0x07, 0x12, 0x16, 0x12, 0x0a, 0xfd, 0xf0, 0xe9, 0xea, 0xf3, 0xff, 
0x0b, 0x12, 0x13, 0x0e, 0x04, 0xf8, 0xee, 0xea, 0xee, 0xf7, 0x02, 0x0b, 0x11, 0x13, 0x0e, 0x03, 0xf5, 0xec, 0xeb, 0xef, 0xf8, 0x04, 0x10, 0x16, 0x14, 0x0a, 0xfe, 0xf1, 0xe8, 0xe7, 0xf0, 0xff, 
0x0c, 0x15, 0x16, 0x11, 0x07, 0xf8, 0xec, 0xe9, 0xed, 0xf7, 0x05, 0x10, 0x16, 0x13, 0x0a, 0xff, 0xf3, 0xea, 0xea, 0xf1, 0xfd, 0x09, 0x11, 0x14, 0x11, 0x08, 0xfc, 0xf1, 0xed, 0xee, 0xf4, 0xfe, 
0x0a, 0x11, 0x12, 0x0e, 0x06, 0xfb, 0xf0, 0xeb, 0xed, 0xf6, 0x02, 0x0d, 0x14, 0x15, 0x10, 0x04, 0xf5, 0xea, 0xe7, 0xec, 0xf9, 0x08, 0x14, 0x17, 0x14, 0x0b, 0xfd, 0xee, 0xe7, 0xe8, 0xf3, 0x01, 
0x0e, 0x16, 0x17, 0x11, 0x05, 0xf7, 0xec, 0xe8, 0xed, 0xf7, 0x05, 0x11, 0x17, 0x14, 0x0b, 0xff, 0xf2, 0xeb, 0xeb, 0xf3, 0xfe, 0x08, 0x10, 0x14, 0x11, 0x08, 0xfd, 0xf2, 0xeb, 0xeb, 0xf1, 0xfd, 
0x0a, 0x11, 0x14, 0x11, 0x08, 0xfa, 0xec, 0xe7, 0xeb, 0xf4, 0x01, 0x0e, 0x16, 0x17, 0x0f, 0x01, 0xf3, 0xe9, 0xe6, 0xed, 0xfb, 0x0a, 0x15, 0x18, 0x13, 0x08, 0xf8, 0xea, 0xe6, 0xe9, 0xf4, 0x03, 
0x10, 0x16, 0x16, 0x0d, 0xff, 0xf3, 0xeb, 0xe9, 0xee, 0xfb, 0x08, 0x12, 0x13, 0x10, 0x08, 0xfc, 0xf1, 0xec, 0xee, 0xf5, 0xfe, 0x07, 0x0e, 0x11, 0x0e, 0x06, 0xfc, 0xf3, 0xee, 0xed, 0xf4, 0xfe, 
0x07, 0x0f, 0x13, 0x10, 0x06, 0xf8, 0xee, 0xeb, 0xed, 0xf7, 0x03, 0x0f, 0x15, 0x14, 0x0a, 0xfe, 0xf2, 0xea, 0xe9, 0xf1, 0xff, 0x0a, 0x12, 0x14, 0x10, 0x03, 0xf5, 0xec, 0xea, 0xf0, 0xfb, 0x06, 
0x0e, 0x11, 0x0e, 0x05, 0xfc, 0xf5, 0xef, 0xef, 0xf5, 0xfe, 0x06, 0x0b, 0x0a, 0x04, 0xfc, 0xf5, 0xf1, 0xf1, 0xf7, 0xfe, 0x06, 0x0c, 0x0e, 0x0c, 0x05, 0xfc, 0xf4, 0xef, 0xf1, 0xf8, 0x00, 0x08, 
0x10, 0x13, 0x0e, 0x04, 0xf7, 0xee, 0xed, 0xf2, 0xfb, 0x08, 0x11, 0x14, 0x10, 0x08, 0xfb, 0xf1, 0xeb, 0xef, 0xf9, 0x05, 0x0d, 0x12, 0x12, 0x0b, 0x00, 0xf5, 0xef, 0xef, 0xf4, 0xfc, 0x07, 0x0e, 
0x10, 0x0c, 0x05, 0xfd, 0xf7, 0xf2, 0xf2, 0xf7, 0xfe, 0x05, 0x0b, 0x0f, 0x0d, 0x06, 0xfd, 0xf4, 0xef, 0xf0, 0xf7, 0x00, 0x09, 0x10, 0x12, 0x0d, 0x03, 0xf8, 0xee, 0xec, 0xf2, 0xfc, 0x05, 0x0f, 
0x13, 0x10, 0x07, 0xfc, 0xf1, 0xec, 0xed, 0xf5, 0x01, 0x0b, 0x11, 0x10, 0x0b, 0x01, 0xf5, 0xee, 0xee, 0xf4, 0xfd, 0x06, 0x0c, 0x0f, 0x0d, 0x05, 0xfc, 0xf5, 0xf2, 0xf3, 0xf7, 0xfe, 0x05, 0x0a, 
0x0c, 0x09, 0x04, 0xfd, 0xf7, 0xf5, 0xf5, 0xf8, 0xfd, 0x03, 0x0a, 0x0e, 0x0c, 0x05, 0xfd, 0xf5, 0xf0, 0xf0, 0xf7, 0x01, 0x0a, 0x11, 0x11, 0x0b, 0x01, 0xf5, 0xee, 0xed, 0xf4, 0xfe, 0x07, 0x0f, 
0x12, 0x0e, 0x06, 0xfa, 0xf1, 0xed, 0xf0, 0xf9, 0x04, 0x0b, 0x0e, 0x0d, 0x07, 0xff, 0xf6, 0xf0, 0xf1, 0xf7, 0xfd, 0x04, 0x0b, 0x0e, 0x0b, 0x04, 0xfd, 0xf7, 0xf2, 0xf3, 0xf7, 0xfe, 0x06, 0x0c, 
0x0e, 0x0c, 0x07, 0xfd, 0xf5, 0xf1, 0xf2, 0xf8, 0x02, 0x0b, 0x0f, 0x0f, 0x0a, 0x02, 0xf8, 0xf0, 0xee, 0xf2, 0xfc, 0x07, 0x0e, 0x11, 0x0f, 0x06, 0xf9, 0xf0, 0xed, 0xf1, 0xf8, 0x03, 0x0c, 0x10, 
0x0e, 0x07, 0xfd, 0xf5, 0xf0, 0xf0, 0xf6, 0x00, 0x07, 0x0a, 0x0c, 0x09, 0x02, 0xfa, 0xf6, 0xf4, 0xf6, 0xfa, 0x00, 0x04, 0x07, 0x09, 0x08, 0x04, 0xff, 0xf9, 0xf5, 0xf5, 0xf8, 0xfd, 0x02, 0x08, 
0x0b, 0x0a, 0x05, 0xfc, 0xf5, 0xf2, 0xf2, 0xf7, 0x00, 0x09, 0x0d, 0x0c, 0x08, 0xff, 0xf6, 0xf0, 0xf0, 0xf5, 0xfe, 0x06, 0x0d, 0x0e, 0x0a, 0x02, 0xf8, 0xf1, 0xf0, 0xf3, 0xfb, 0x04, 0x0a, 0x0d, 
0x0b, 0x05, 0xfd, 0xf6, 0xf3, 0xf5, 0xfa, 0x01, 0x05, 0x08, 0x0a, 0x08, 0x02, 0xfd, 0xf9, 0xf7, 0xf8, 0xfc, 0x00, 0x04, 0x08, 0x09, 0x08, 0x05, 0xff, 0xf9, 0xf6, 0xf6, 0xf9, 0xfe, 0x05, 0x0c, 
0x0d, 0x09, 0x02, 0xfb, 0xf5, 0xf2, 0xf4, 0xfb, 0x04, 0x0c, 0x0a, 0x05, 0xfc, 0xf5, 0xf3, 0xf5, 0xfb, 0x03, 0x0a, 0x0d, 0x0b, 0x06, 0xff, 0xf9, 0xf6, 0xf8, 0xfb, 0x00, 0x05, 0x09, 0x09, 0x07, 
0x04, 0xff, 0xfb, 0xfa, 0xfc, 0xfd, 0x00, 0x02, 0x05, 0x06, 0x07, 0x05, 0x01, 0xfe, 0xfa, 0xf8, 0xf9, 0xfd, 0x01, 0x07, 0x0a, 0x0a, 0x06, 0xff, 0xf8, 0xf4, 0xf4, 0xf8, 0xff, 0x07, 0x0b, 0x0b, 
0x06, 0xff, 0xf8, 0xf3, 0xf2, 0xf7, 0xfd, 0x05, 0x0a, 0x0b, 0x07, 0x01, 0xfa, 0xf5, 0xf5, 0xf8, 0xfd, 0x01, 0x05, 0x08, 0x07, 0x03, 0xff, 0xfb, 0xf8, 0xf8, 0xf9, 0xfc, 0xff, 0x02, 0x04, 0x06, 
0x05, 0x02, 0xfd, 0xf9, 0xf7, 0xf6, 0xf8, 0xfe, 0x04, 0x09, 0x0b, 0x07, 0x00, 0xf9, 0xf3, 0xf1, 0xf4, 0xfb, 0x02, 0x09, 0x0d, 0x0b, 0x03, 0xfa, 0xf4, 0xf1, 0xf2, 0xf9, 0x02, 0x0a, 0x0d, 0x0b, 
0x05, 0xfc, 0xf5, 0xf3, 0xf4, 0xf9, 0xff, 0x06, 0x09, 0x0a, 0x06, 0x01, 0xfb, 0xf8, 0xf8, 0xfb, 0xfe, 0x01, 0x04, 0x05, 0x06, 0x03, 0x00, 0xfe, 0xfb, 0xfa, 0xfb, 0xfc, 0xfe, 0x02, 0x06, 0x08, 
0x07, 0x04, 0xff, 0xfa, 0xf7, 0xf5, 0xf8, 0xff, 0x06, 0x0b, 0x0b, 0x08, 0x01, 0xf9, 0xf3, 0xf2, 0xf6, 0xfe, 0x06, 0x0c, 0x0d, 0x09, 0x03, 0xfb, 0xf5, 0xf3, 0xf7, 0xfd, 0x05, 0x0a, 0x0c, 0x0a, 
0x05, 0xfe, 0xf9, 0xf8, 0xf8, 0xfa, 0xff, 0x04, 0x08, 0x0a, 0x07, 0x03, 0xff, 0xfc, 0xfa, 0xf9, 0xfa, 0xfe, 0x03, 0x07, 0x09, 0x07, 0x04, 0x00, 0xfa, 0xf6, 0xf6, 0xf8, 0xfd, 0x02, 0x08, 0x0c, 
0x0a, 0x03, 0xfb, 0xf4, 0xf1, 0xf4, 0xfa, 0x02, 0x09, 0x0e, 0x0c, 0x06, 0xfd, 0xf6, 0xf2, 0xf3, 0xf8, 0x00, 0x07, 0x0c, 0x0c, 0x07, 0xff, 0xf9, 0xf4, 0xf3, 0xf7, 0xfd, 0x04, 0x08, 0x0a, 0x07, 
0x02, 0xfc, 0xf7, 0xf6, 0xf8, 0xfb, 0xff, 0x04, 0x06, 0x06, 0x04, 0x02, 0xfe, 0xfb, 0xf9, 0xf9, 0xfa, 0xfd, 0x00, 0x04, 0x07, 0x07, 0x04, 0xfe, 0xf9, 0xf5, 0xf4, 0xf7, 0xfd, 0x05, 0x0b, 0x0c, 
0x08, 0x00, 0xf8, 0xf2, 0xf0, 0xf5, 0xfd, 0x06, 0x0b, 0x0a, 0x03, 0xfa, 0xf4, 0xf2, 0xf6, 0xfc, 0x02, 0x09, 0x0c, 0x0a, 0x04, 0xfd, 0xf7, 0xf5, 0xf6, 0xfb, 0x00, 0x05, 0x08, 0x08, 0x06, 0x02, 
0xfd, 0xfa, 0xf9, 0xfa, 0xfb, 0xfe, 0x03, 0x07, 0x08, 0x06, 0x03, 0xfe, 0xf9, 0xf7, 0xf6, 0xf9, 0xff, 0x05, 0x0a, 0x0a, 0x08, 0x02, 0xfc, 0xf6, 0xf4, 0xf6, 0xfc, 0x05, 0x0c, 0x0f, 0x0c, 0x04, 
0xfc, 0xf6, 0xf4, 0xf6, 0xfc, 0x04, 0x0a, 0x0d, 0x0c, 0x05, 0xfd, 0xf8, 0xf5, 0xf7, 0xfc, 0x01, 0x06, 0x09, 0x09, 0x06, 0x01, 0xfc, 0xfa, 0xf9, 0xfb, 0xfe, 0x01, 0x04, 0x06, 0x06, 0x05, 0x03, 
0x00, 0xfc, 0xf8, 0xf7, 0xf8, 0xfd, 0x03, 0x08, 0x0a, 0x08, 0x03, 0xfc, 0xf6, 0xf3, 0xf5, 0xfc, 0x03, 0x0a, 0x0d, 0x0c, 0x06, 0xfe, 0xf7, 0xf2, 0xf3, 0xf8, 0x00, 0x09, 0x0e, 0x0d, 0x07, 0xff, 
0xf7, 0xf3, 0xf3, 0xf8, 0x00, 0x06, 0x09, 0x0a, 0x07, 0x01, 0xfc, 0xf8, 0xf6, 0xf7, 0xfb, 0xff, 0x03, 0x06, 0x07, 0x05, 0x01, 0xfd, 0xf9, 0xf6, 0xf7, 0xfa, 0xfe, 0x04, 0x08, 0x09, 0x08, 0x03, 
0xfc, 0xf6, 0xf3, 0xf4, 0xf9, 0x00, 0x07, 0x0c, 0x0c, 0x06, 0xfe, 0xf6, 0xf0, 0xf1, 0xf6, 0xff, 0x07, 0x0d, 0x0d, 0x08, 0x00, 0xf8, 0xf3, 0xf2, 0xf5, 0xfc, 0x04, 0x0a, 0x0c, 0x09, 0x03, 0xfd, 
0xf7, 0xf4, 0xf5, 0xf9, 0xff, 0x04, 0x08, 0x09, 0x07, 0x03, 0xfd, 0xf8, 0xf5, 0xf6, 0xf9, 0xfe, 0x03, 0x07, 0x09, 0x07, 0x03, 0xfd, 0xf8, 0xf5, 0xf5, 0xf9, 0x00, 0x07, 0x0b, 0x0b, 0x07, 0x01, 
0xfa, 0xf4, 0xf3, 0xf6, 0xfd, 0x05, 0x0b, 0x0a, 0x04, 0xfc, 0xf5, 0xf4, 0xf6, 0xfb, 0x03, 0x09, 0x0c, 0x0a, 0x05, 0xfe, 0xf8, 0xf5, 0xf6, 0xf9, 0xfe, 0x04, 0x08, 0x09, 0x07, 0x03, 0xfd, 0xf9, 
0xf7, 0xf7, 0xfa, 0xff, 0x04, 0x07, 0x09, 0x08, 0x03, 0xfe, 0xf9, 0xf6, 0xf5, 0xf9, 0x00, 0x06, 0x0b, 0x0c, 0x09, 0x02, 0xfa, 0xf5, 0xf3, 0xf5, 0xfc, 0x04, 0x0b, 0x0d, 0x0b, 0x05, 0xfd, 0xf7, 
0xf4, 0xf5, 0xf9, 0x00, 0x07, 0x0b, 0x0b, 0x08, 0x01, 0xfa, 0xf6, 0xf5, 0xf8, 0xfd, 0x03, 0x09, 0x0a, 0x08, 0x03, 0xfd, 0xf9, 0xf7, 0xf6, 0xf9, 0xff, 0x03, 0x06, 0x07, 0x07, 0x04, 0xfe, 0xf9, 
0xf6, 0xf6, 0xf9, 0xfe, 0x03, 0x07, 0x09, 0x08, 0x04, 0xfd, 0xf7, 0xf4, 0xf4, 0xf9, 0x00, 0x07, 0x0b, 0x0b, 0x07, 0xff, 0xf9, 0xf4, 0xf3, 0xf7, 0xfe, 0x05, 0x0a, 0x0b, 0x08, 0x02, 0xfc, 0xf6, 
0xf4, 0xf7, 0xfc, 0x02, 0x07, 0x09, 0x08, 0x04, 0xff, 0xfb, 0xf8, 0xf7, 0xf9, 0xfe, 0x03, 0x07, 0x08, 0x07, 0x04, 0x00, 0xfc, 0xf9, 0xf9, 0xfb, 0xff, 0x03, 0x07, 0x09, 0x09, 0x05, 0xff, 0xfb, 
0xf8, 0xf7, 0xf9, 0xff, 0x05, 0x09, 0x0b, 0x08, 0x02, 0xfc, 0xf7, 0xf5, 0xf6, 0xfd, 0x04, 0x09, 0x0b, 0x0a, 0x04, 0xfd, 0xf8, 0xf5, 0xf6, 0xfa, 0x01, 0x06, 0x08, 0x08, 0x05, 0x00, 0xfa, 0xf7, 
0xf6, 0xf9, 0xfe, 0x02, 0x06, 0x08, 0x06, 0x02, 0xff, 0xfb, 0xf8, 0xf8, 0xfa, 0xfe, 0x02, 0x04, 0x06, 0x06, 0x04, 0x00, 0xfb, 0xf7, 0xf7, 0xf9, 0xfd, 0x03, 0x08, 0x09, 0x07, 0x03, 0xfc, 0xf7, 
0xf5, 0xf5, 0xfa, 0x01, 0x07, 0x0a, 0x0a, 0x05, 0xfd, 0xf7, 0xf4, 0xf4, 0xf9, 0xff, 0x05, 0x08, 0x08, 0x05, 0x00, 0xfa, 0xf7, 0xf6, 0xf7, 0xfc, 0x01, 0x05, 0x07, 0x07, 0x03, 0xff, 0xfa, 0xf8, 
0xf7, 0xf9, 0xfe, 0x02, 0x05, 0x06, 0x05, 0x01, 0xfe, 0xfb, 0xf8, 0xf8, 0xfa, 0xfe, 0x01, 0x04, 0x06, 0x05, 0x02, 0xfd, 0xfa, 0xf8, 0xf7, 0xf8, 0xfd, 0x03, 0x08, 0x09, 0x07, 0x03, 0xfd, 0xf7, 
0xf6, 0xf7, 0xfb, 0x01, 0x06, 0x09, 0x08, 0x05, 0xff, 0xfa, 0xf8, 0xf7, 0xfa, 0xff, 0x03, 0x06, 0x07, 0x06, 0x02, 0xfe, 0xfa, 0xf9, 0xfa, 0xfc, 0x00, 0x04, 0x07, 0x08, 0x06, 0x02, 0xfe, 0xfc, 
0xfa, 0xfa, 0xfe, 0x01, 0x04, 0x06, 0x07, 0x05, 0x02, 0xfe, 0xfa, 0xf9, 0xfa, 0xfc, 0x01, 0x05, 0x08, 0x08, 0x05, 0x01, 0xfc, 0xf8, 0xf7, 0xfb, 0xff, 0x04, 0x07, 0x08, 0x06, 0x02, 0xfc, 0xf8, 
0xf7, 0xf9, 0xfd, 0x02, 0x06, 0x08, 0x07, 0x03, 0xff, 0xfc, 0xfa, 0xfa, 0xfc, 0x00, 0x03, 0x05, 0x05, 0x04, 0x01, 0xff, 0xfc, 0xfb, 0xfc, 0xfd, 0xfe, 0x01, 0x03, 0x05, 0x05, 0x02, 0xff, 0xfc, 
0xf9, 0xf9, 0xfc, 0xff, 0x02, 0x05, 0x06, 0x05, 0x01, 0xfd, 0xf9, 0xf7, 0xf9, 0xfc, 0x01, 0x05, 0x06, 0x05, 0x03, 0xfe, 0xf9, 0xf6, 0xf8, 0xfc, 0x00, 0x03, 0x05, 0x05, 0x02, 0xfe, 0xfb, 0xf9, 
0xfa, 0xfc, 0xff, 0x01, 0x04, 0x04, 0x03, 0x01, 0xff, 0xfd, 0xfc, 0xfc, 0xfd, 0xff, 0x01, 0x02, 0x03, 0x03, 0x01, 0x00, 0xfe, 0xfc, 0xfb, 0xfb, 0xfd, 0x00, 0x03, 0x04, 0x04, 0x02, 0xff, 0xfc, 
0xf9, 0xf9, 0xfb, 0xff, 0x02, 0x06, 0x07, 0x04, 0x00, 0xfc, 0xf9, 0xf9, 0xfb, 0xff, 0x03, 0x06, 0x06, 0x04, 0x01, 0xfd, 0xfa, 0xf9, 0xfc, 0xff, 0x02, 0x04, 0x04, 0x03, 0x02, 0xff, 0xfe, 0xfd, 
0xfd, 0xfe, 0x00, 0x01, 0x02, 0x02, 0x02, 0x02, 0x01, 0xff, 0xfd, 0xfd, 0xfc, 0xfe, 0x00, 0x02, 0x03, 0x04, 0x02, 0x00, 0xfe, 0xfb, 0xfa, 0xfa, 0xfd, 0x02, 0x05, 0x06, 0x05, 0x01, 0xfd, 0xf9, 
0xf9, 0xfb, 0xfe, 0x02, 0x05, 0x07, 0x06, 0x01, 0xfd, 0xfb, 0xf9, 0xfa, 0xfe, 0x03, 0x05, 0x05, 0x04, 0x02, 0xff, 0xfd, 0xfc, 0xfd, 0xff, 0x01, 0x02, 0x02, 0x03, 0x02, 0x00, 0xff, 0xff, 0xfe, 
0xfe, 0xfd, 0xfe, 0x00, 0x01, 0x03, 0x04, 0x03, 0x00, 0xfd, 0xfb, 0xfb, 0xfb, 0xfe, 0x01, 0x04, 0x05, 0x04, 0x02, 0xfe, 0xfa, 0xf9, 0xfa, 0xfc, 0x00, 0x04, 0x06, 0x05, 0x03, 0xff, 0xfb, 0xf9, 
0xfa, 0xfb, 0xff, 0x03, 0x05, 0x05, 0x02, 0xfe, 0xfc, 0xfb, 0xfb, 0xfd, 0xff, 0x01, 0x02, 0x02, 0x02, 0x00, 0xfe, 0xfd, 0xfc, 0xfc, 0xfd, 0xfe, 0xff, 0x01, 0x03, 0x02, 0x01, 0xff, 0xfe, 0xfc, 
0xfc, 0xfc, 0xfe, 0x00, 0x03, 0x04, 0x04, 0x02, 0xff, 0xfc, 0xfa, 0xfa, 0xfd, 0x00, 0x04, 0x06, 0x05, 0x03, 0x00, 0xfc, 0xf9, 0xf9, 0xfb, 0xfe, 0x02, 0x06, 0x06, 0x04, 0x01, 0xfd, 0xfb, 0xfa, 
0xfb, 0xfe, 0x01, 0x04, 0x05, 0x03, 0x02, 0x00, 0xfe, 0xfb, 0xfb, 0xfd, 0xfe, 0x00, 0x02, 0x03, 0x02, 0x01, 0x00, 0xfe, 0xfb, 0xfb, 0xfb, 0xfe, 0x00, 0x02, 0x04, 0x04, 0x03, 0xff, 0xfb, 0xfa, 
0xfa, 0xfc, 0xfe, 0x02, 0x06, 0x06, 0x05, 0x02, 0xfe, 0xfa, 0xf8, 0xfa, 0xfd, 0x02, 0x06, 0x07, 0x06, 0x03, 0xff, 0xfb, 0xf9, 0xfa, 0xfc, 0x00, 0x05, 0x06, 0x06, 0x04, 0x00, 0xfd, 0xfa, 0xfb, 
0xfd, 0x00, 0x03, 0x04, 0x04, 0x02, 0x00, 0xfe, 0xfd, 0xfc, 0xfd, 0xfe, 0x00, 0x01, 0x02, 0x02, 0x02, 0x01, 0x00, 0xff, 0xfd, 0xfc, 0xfc, 0xfd, 0x00, 0x03, 0x04, 0x04, 0x03, 0x00, 0xfd, 0xfa, 
0xf9, 0xfb, 0xff, 0x03, 0x06, 0x07, 0x05, 0x01, 0xfc, 0xf8, 0xf8, 0xfb, 0xfe, 0x03, 0x07, 0x07, 0x04, 0x01, 0xfc, 0xfa, 0xf9, 0xfb, 0xfd, 0x01, 0x05, 0x05, 0x04, 0x01, 0xff, 0xfc, 0xfb, 0xfb, 
0xfc, 0xff, 0x02, 0x03, 0x03, 0x02, 0x01, 0xfe, 0xfd, 0xfc, 0xfc, 0xfd, 0xff, 0x01, 0x03, 0x04, 0x05, 0x03, 0x00, 0xfd, 0xfc, 0xfc, 0xfd, 0x00, 0x03, 0x06, 0x07, 0x05, 0x02, 0xfe, 0xfa, 0xf9, 
0xfb, 0xff, 0x03, 0x06, 0x06, 0x05, 0x02, 0xfe, 0xfa, 0xf9, 0xfb, 0xfe, 0x01, 0x04, 0x06, 0x05, 0x01, 0xfe, 0xfb, 0xfa, 0xfb, 0xfd, 0x00, 0x03, 0x04, 0x02, 0x00, 0xff, 0xfd, 0xfb, 0xfb, 0xfd, 
0xfe, 0xff, 0x00, 0x02, 0x02, 0x01, 0x00, 0xfe, 0xfd, 0xfc, 0xfc, 0xfd, 0xff, 0x00, 0x02, 0x03, 0x02, 0x00, 0xfd, 0xfb, 0xfa, 0xfb, 0xfd, 0x00, 0x03, 0x04, 0x03, 0x00, 0xfd, 0xfb, 0xf9, 0xf9, 
0xfd, 0x00, 0x03, 0x04, 0x02, 0x01, 0xff, 0xfb, 0xf9, 0xfa, 0xfc, 0xff, 0x02, 0x03, 0x03, 0x00, 0xfd, 0xfc, 0xfb, 0xfb, 0xfd, 0xff, 0x01, 0x02, 0x01, 0x00, 0x00, 0xff, 0xfe, 0xfd, 0xfd, 0xfe, 
0xfe, 0xff, 0x00, 0x01, 0x02, 0x01, 0x00, 0xff, 0xfe, 0xfc, 0xfc, 0xfd, 0x00, 0x02, 0x04, 0x04, 0x03, 0x00, 0xfd, 0xfb, 0xfa, 0xfd, 0x01, 0x04, 0x05, 0x06, 0x03, 0x00, 0xfd, 0xfb, 0xfb, 0xfd, 
0x01, 0x04, 0x05, 0x04, 0x02, 0xff, 0xfe, 0xfd, 0xfe, 0xfe, 0x00, 0x02, 0x03, 0x03, 0x02, 0x00, 0xff, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0x01, 0x02, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 
0xff, 0xff, 0x01, 0x02, 0x02, 0x02, 0x02, 0x00, 0xfe, 0xfd, 0xfd, 0xff, 0x00, 0x01, 0x03, 0x03, 0x00, 0xfe, 0xfc, 0xfb, 0xfc, 0xfe, 0x01, 0x02, 0x02, 0x01, 0xff, 0xfe, 0xfd, 0xfd, 0xfd, 0xfe, 
0x00, 0x01, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xff, 0x00, 0x01, 0x03, 0x03, 0x01, 0xff, 0xfe, 0xfd, 0xfd, 
0xfe, 0x00, 0x02, 0x03, 0x03, 0x02, 0x00, 0xfe, 0xfc, 0xfc, 0xfd, 0x00, 0x02, 0x03, 0x04, 0x02, 0xff, 0xfc, 0xfb, 0xfc, 0xfe, 0x00, 0x02, 0x03, 0x02, 0x00, 0xfe, 0xfd, 0xfd, 0xfd, 0xff, 0x00, 
0x01, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xfe, 0xfd, 0xfd, 0xfd, 0xff, 0x00, 0x01, 0x01, 0x00, 0xff, 0xfd, 0xfb, 0xfc, 0xfd, 
0xff, 0x01, 0x02, 0x02, 0x00, 0xff, 0xfd, 0xfc, 0xfc, 0xfe, 0xff, 0x00, 0x01, 0x02, 0x00, 0xff, 0xfe, 0xfd, 0xfe, 0xfe, 0xff, 0x01, 0x01, 0x01, 0x01, 0xff, 0xff, 0x00, 0xff, 0xff, 0xff, 0xff, 
0x00, 0x00, 0x01, 0x02, 0x02, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x01, 0x03, 0x04, 0x03, 0x01, 0xfe, 0xfd, 0xfd, 0xfd, 0xff, 0x01, 0x03, 0x04, 0x03, 0x01, 0xfe, 0xfc, 0xfb, 0xfd, 0xff, 
0x02, 0x03, 0x04, 0x02, 0xff, 0xfd, 0xfc, 0xfc, 0xfe, 0x00, 0x02, 0x03, 0x03, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x02, 0x02, 0x01, 0x00, 0x00, 0xff, 0xfe, 0xff, 0x00, 0x00, 
0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0xfe, 0xfd, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x02, 0x00, 0xfe, 0xfe, 0xfc, 0xfc, 0xfd, 0xff, 0x01, 0x02, 0x02, 0x01, 0xff, 0xfd, 0xfc, 0xfc, 0xfd, 0xff, 
0x01, 0x02, 0x02, 0x00, 0xfe, 0xfd, 0xfd, 0xfd, 0xfe, 0xff, 0x00, 0x01, 0x00, 0x00, 0x00, 0xfe, 0xfe, 0xfe, 0xfd, 0xfe, 0xfe, 0xff, 0x01, 0x01, 0x01, 0x00, 0x00, 0xfe, 0xfd, 0xfd, 0xfd, 0xfe, 
0x00, 0x01, 0x01, 0x02, 0x01, 0xff, 0xfd, 0xfc, 0xfc, 0xfd, 0xff, 0x01, 0x02, 0x03, 0x02, 0x00, 0xfe, 0xfd, 0xfc, 0xfd, 0xff, 0x01, 0x02, 0x02, 0x02, 0x01, 0xff, 0xfe, 0xfd, 0xfd, 0xff, 0x01, 
0x02, 0x02, 0x02, 0x01, 0x00, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x01, 0x01, 0xff, 0xff, 0xfe, 0xfd, 0xfe, 0xff, 
0x00, 0x01, 0x02, 0x01, 0x01, 0xff, 0xfe, 0xfd, 0xfd, 0xfe, 0x00, 0x02, 0x02, 0x03, 0x01, 0xff, 0xff, 0xfe, 0xfe, 0x00, 0x01, 0x02, 0x02, 0x02, 0x01, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x01, 
0x02, 0x02, 0x01, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 
0x00, 0x01, 0x01, 0xff, 0xff, 0xfe, 0xfe, 0xfd, 0xff, 0x00, 0x01, 0x02, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x00, 0x00, 0xff, 0xfe, 0xfe, 0xfd, 0xfe, 0xff, 0x00, 0x00, 0x00, 
0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0xff, 0x00, 0x00, 0xff, 0xff, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0x01, 0x02, 
0x02, 0x01, 0x00, 0xfe, 0xfd, 0xfd, 0xfe, 0xff, 0x01, 0x01, 0x02, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0x01, 0x02, 0x02, 0x01, 0x00, 0xfe, 0xfd, 0xfe, 0xff, 0xff, 0x00, 0x01, 0x01, 0x00, 
0xff, 0xfe, 0xfd, 0xfe, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x02, 0x01, 0x00, 0xff, 0xfe, 0xfd, 0xfe, 0xff, 0x01, 0x03, 0x02, 
0x01, 0x00, 0xff, 0xfd, 0xfd, 0xfe, 0xff, 0x01, 0x02, 0x02, 0x01, 0x00, 0xfe, 0xfd, 0xfc, 0xfd, 0xfe, 0x00, 0x02, 0x02, 0x01, 0x00, 0xff, 0xfd, 0xfc, 0xfd, 0xff, 0x00, 0x01, 0x01, 0x01, 0x01, 
0xff, 0xfe, 0xfe, 0xfd, 0xfd, 0xfe, 0xff, 0x00, 0x01, 0x02, 0x01, 0x00, 0xfe, 0xfc, 0xfc, 0xfd, 0xfd, 0xff, 0x01, 0x02, 0x01, 0x01, 0x00, 0xfe, 0xfc, 0xfc, 0xfd, 0xff, 0x01, 0x01, 0x01, 0x01, 
0x00, 0xfe, 0xfd, 0xfc, 0xfd, 0xff, 0x01, 0x02, 0x02, 0x03, 0x02, 0xff, 0xfe, 0xfd, 0xfe, 0xff, 0xff, 0x00, 0x01, 0x01, 0x01, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x01, 0x01, 0x02, 0x01, 
0x01, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x01, 0x03, 0x03, 0x02, 0x01, 0x00, 0xfe, 0xfd, 0xfe, 0xff, 0x00, 0x02, 0x03, 0x03, 0x02, 0x00, 0xfe, 0xfd, 0xfd, 0xfe, 0x00, 0x01, 0x03, 0x03, 0x02, 
0x01, 0xff, 0xfe, 0xfe, 0xff, 0x01, 0x02, 0x03, 0x03, 0x01, 0x00, 0xff, 0xff, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0xff, 0xff, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x02, 0x01, 
0x01, 0x00, 0xfe, 0xfd, 0xfd, 0xfe, 0xfe, 0x00, 0x02, 0x02, 0x01, 0x00, 0xfd, 0xfc, 0xfc, 0xfc, 0xfe, 0x01, 0x03, 0x03, 0x01, 0xff, 0xfd, 0xfc, 0xfc, 0xfc, 0xfe, 0x00, 0x01, 0x01, 0x01, 0xff, 
0xfd, 0xfd, 0xfc, 0xfd, 0xfe, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xfe, 0xfe, 0xfd, 0xfd, 0xfe, 0xff, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfd, 0xfd, 0xfd, 0xfd, 0x00, 0x02, 0x02, 0x01, 
0x00, 0xff, 0xfc, 0xfb, 0xfb, 0xfd, 0x00, 0x03, 0x04, 0x03, 0x02, 0xff, 0xfd, 0xfc, 0xfc, 0xfd, 0x00, 0x03, 0x05, 0x04, 0x02, 0xff, 0xfd, 0xfd, 0xfe, 0xff, 0x00, 0x02, 0x03, 0x04, 0x03, 0x02, 
0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x01, 0x02, 0x03, 0x02, 0x02, 0x00, 0xff, 0xfe, 0xfe, 0xfd, 0xfe, 0x00, 0x03, 0x04, 0x04, 0x02, 0xff, 0xfd, 0xfc, 0xfc, 0xfd, 0x00, 0x01, 0x03, 0x04, 0x02, 
0x00, 0xfe, 0xfc, 0xfb, 0xfc, 0xff, 0x02, 0x03, 0x03, 0x03, 0x00, 0xff, 0xfd, 0xfc, 0xfd, 0xfe, 0x00, 0x02, 0x03, 0x03, 0x02, 0x00, 0xff, 0xfd, 0xfd, 0xfd, 0xfe, 0x00, 0x01, 0x02, 0x03, 0x03, 
0x02, 0x00, 0xfd, 0xfc, 0xfc, 0xfd, 0xff, 0x02, 0x03, 0x04, 0x03, 0x00, 0xfe, 0xfc, 0xfb, 0xfb, 0xfe, 0x01, 0x04, 0x04, 0x03, 0x01, 0xff, 0xfc, 0xfa, 0xfb, 0xfd, 0x00, 0x03, 0x04, 0x03, 0x02, 
0xff, 0xfd, 0xfb, 0xfb, 0xfc, 0xff, 0x01, 0x03, 0x03, 0x01, 0xff, 0xfe, 0xfc, 0xfb, 0xfb, 0xfc, 0xfe, 0xff, 0x02, 0x03, 0x02, 0x00, 0xfe, 0xfc, 0xfb, 0xfb, 0xfc, 0xfe, 0x02, 0x03, 0x03, 0x02, 
0xff, 0xfc, 0xfa, 0xf9, 0xfa, 0xfd, 0x00, 0x03, 0x04, 0x04, 0x01, 0xff, 0xfc, 0xfb, 0xfa, 0xfc, 0xff, 0x02, 0x04, 0x04, 0x02, 0x00, 0xfd, 0xfa, 0xfa, 0xfb, 0xfe, 0x01, 0x04, 0x05, 0x03, 0x02, 
0x00, 0xfd, 0xfb, 0xfc, 0xfd, 0xff, 0x02, 0x04, 0x04, 0x03, 0x02, 0xff, 0xfd, 0xfb, 0xfb, 0xfd, 0xff, 0x02, 0x04, 0x05, 0x05, 0x02, 0xff, 0xfc, 0xfa, 0xfa, 0xfc, 0x00, 0x03, 0x06, 0x06, 0x04, 
0x01, 0xfe, 0xfb, 0xfa, 0xfb, 0xfe, 0x03, 0x06, 0x07, 0x06, 0x03, 0x00, 0xfd, 0xfb, 0xfa, 0xfd, 0x01, 0x06, 0x08, 0x07, 0x05, 0x01, 0xfd, 0xfa, 0xfa, 0xfc, 0xff, 0x03, 0x06, 0x07, 0x07, 0x04, 
0x00, 0xfd, 0xfb, 0xfb, 0xfd, 0x00, 0x03, 0x06, 0x06, 0x05, 0x02, 0xff, 0xfb, 0xfa, 0xfb, 0xfe, 0x01, 0x04, 0x06, 0x06, 0x05, 0x01, 0xfd, 0xfb, 0xf9, 0xfa, 0xfd, 0x01, 0x05, 0x07, 0x06, 0x03, 
0xff, 0xfb, 0xf9, 0xf9, 0xfc, 0xff, 0x03, 0x06, 0x06, 0x04, 0x00, 0xfd, 0xf9, 0xf7, 0xf9, 0xfc, 0x00, 0x03, 0x04, 0x04, 0x01, 0xfe, 0xfa, 0xf9, 0xf8, 0xfa, 0xfd, 0x00, 0x03, 0x04, 0x03, 0x00, 
0xfd, 0xfa, 0xf7, 0xf7, 0xfa, 0xfd, 0x01, 0x04, 0x06, 0x04, 0x00, 0xfd, 0xfa, 0xf8, 0xf8, 0xfb, 0xff, 0x03, 0x06, 0x06, 0x04, 0xff, 0xfa, 0xf7, 0xf7, 0xfa, 0xfe, 0x03, 0x06, 0x07, 0x05, 0x02, 
0xfe, 0xfa, 0xf8, 0xf9, 0xfc, 0xff, 0x04, 0x06, 0x06, 0x04, 0x00, 0xfc, 0xf9, 0xf8, 0xfa, 0xfd, 0x01, 0x04, 0x05, 0x05, 0x02, 0xff, 0xfc, 0xf9, 0xf9, 0xfb, 0xfe, 0x02, 0x04, 0x06, 0x05, 0x03, 
0x00, 0xfc, 0xf9, 0xf9, 0xfb, 0xff, 0x03, 0x06, 0x07, 0x04, 0x01, 0xfd, 0xf9, 0xf8, 0xfa, 0xfd, 0x03, 0x06, 0x08, 0x07, 0x04, 0x00, 0xfb, 0xf8, 0xf9, 0xfc, 0x00, 0x04, 0x06, 0x06, 0x04, 0x02, 
0xfe, 0xfb, 0xf9, 0xfa, 0xfd, 0x00, 0x03, 0x05, 0x05, 0x04, 0x01, 0xfd, 0xfa, 0xf9, 0xfa, 0xfd, 0x00, 0x04, 0x06, 0x06, 0x04, 0x01, 0xfd, 0xf8, 0xf8, 0xf9, 0xfd, 0x02, 0x06, 0x07, 0x06, 0x03, 
0xfe, 0xf9, 0xf8, 0xf8, 0xfb, 0x00, 0x04, 0x07, 0x07, 0x04, 0x00, 0xfc, 0xf9, 0xf8, 0xf9, 0xfd, 0x01, 0x05, 0x06, 0x05, 0x02, 0xfe, 0xfa, 0xf8, 0xf8, 0xfa, 0xfe, 0x02, 0x06, 0x06, 0x05, 0x02, 
0xfd, 0xf9, 0xf8, 0xf9, 0xfc, 0x00, 0x04, 0x06, 0x07, 0x04, 0x00, 0xfd, 0xfa, 0xf9, 0xfb, 0xfe, 0x03, 0x07, 0x08, 0x07, 0x03, 0x00, 0xfc, 0xf9, 0xf9, 0xfd, 0x01, 0x05, 0x08, 0x08, 0x06, 0x01, 
0xfd, 0xfa, 0xf9, 0xfb, 0xff, 0x02, 0x06, 0x07, 0x06, 0x03, 0x00, 0xfc, 0xfa, 0xf9, 0xfc, 0xff, 0x03, 0x06, 0x06, 0x05, 0x02, 0xfe, 0xfb, 0xf9, 0xfa, 0xfc, 0xff, 0x03, 0x05, 0x06, 0x05, 0x02, 
0xfe, 0xfa, 0xf8, 0xf9, 0xfd, 0x01, 0x03, 0x06, 0x06, 0x04, 0x00, 0xfc, 0xf9, 0xf8, 0xfa, 0xfd, 0x02, 0x05, 0x06, 0x06, 0x03, 0xfe, 0xfb, 0xf8, 0xf7, 0xfa, 0xff, 0x04, 0x07, 0x07, 0x05, 0x01, 
0xfd, 0xf9, 0xf7, 0xf9, 0xfd, 0x02, 0x05, 0x06, 0x06, 0x02, 0xff, 0xfb, 0xf8, 0xf8, 0xfa, 0xfd, 0x02, 0x05, 0x07, 0x06, 0x02, 0xfe, 0xfa, 0xf8, 0xf8, 0xfb, 0xff, 0x04, 0x06, 0x06, 0x05, 0x01, 
0xfc, 0xf9, 0xf8, 0xf9, 0xfd, 0x01, 0x04, 0x07, 0x07, 0x04, 0xff, 0xfb, 0xf8, 0xf8, 0xfa, 0xff, 0x03, 0x06, 0x07, 0x06, 0x02, 0xfd, 0xf9, 0xf7, 0xf8, 0xfb, 0x00, 0x04, 0x06, 0x06, 0x04, 0x00, 
0xfc, 0xf9, 0xf8, 0xf9, 0xfd, 0x01, 0x05, 0x06, 0x06, 0x03, 0xff, 0xfb, 0xf9, 0xf9, 0xfb, 0xfe, 0x02, 0x06, 0x07, 0x05, 0x02, 0xfe, 0xfa, 0xf8, 0xf9, 0xfc, 0x00, 0x03, 0x05, 0x06, 0x05, 0x01, 
0xfd, 0xf9, 0xf8, 0xf9, 0xfd, 0x01, 0x04, 0x06, 0x06, 0x04, 0x00, 0xfb, 0xf8, 0xf8, 0xfa, 0xfe, 0x02, 0x06, 0x07, 0x06, 0x02, 0xfd, 0xf9, 0xf7, 0xf8, 0xfc, 0x00, 0x04, 0x07, 0x06, 0x04, 0x00, 
0xfc, 0xf9, 0xf8, 0xfa, 0xfe, 0x02, 0x05, 0x07, 0x05, 0x02, 0xfe, 0xfb, 0xf9, 0xf9, 0xfc, 0x00, 0x03, 0x06, 0x06, 0x05, 0x02, 0xfd, 0xfa, 0xf8, 0xf9, 0xfd, 0x00, 0x04, 0x06, 0x06, 0x04, 0x01, 
0xfc, 0xf9, 0xf9, 0xfa, 0xfd, 0x02, 0x05, 0x07, 0x06, 0x03, 0xff, 0xfb, 0xf8, 0xf8, 0xfb, 0xff, 0x03, 0x06, 0x06, 0x04, 0x01, 0xfc, 0xf9, 0xf7, 0xf9, 0xfd, 0x01, 0x05, 0x07, 0x06, 0x03, 0xff, 
0xfa, 0xf8, 0xf8, 0xfa, 0xfe, 0x02, 0x05, 0x06, 0x05, 0x02, 0xfe, 0xfa, 0xf8, 0xf8, 0xfb, 0xff, 0x03, 0x05, 0x06, 0x04, 0x00, 0xfc, 0xf9, 0xf7, 0xf9, 0xfc, 0x00, 0x04, 0x06, 0x06, 0x03, 0xff, 
0xfb, 0xf7, 0xf7, 0xf9, 0xfd, 0x02, 0x05, 0x07, 0x06, 0x03, 0xfd, 0xf9, 0xf7, 0xf9, 0xfc, 0x00, 0x05, 0x08, 0x08, 0x05, 0x01, 0xfd, 0xf9, 0xf8, 0xfb, 0xff, 0x03, 0x07, 0x08, 0x07, 0x04, 0x00, 
0xfc, 0xfa, 0xfa, 0xfc, 0x00, 0x04, 0x08, 0x08, 0x06, 0x03, 0xff, 0xfc, 0xfa, 0xfa, 0xfd, 0x02, 0x05, 0x08, 0x08, 0x06, 0x02, 0xfe, 0xfa, 0xf9, 0xfb, 0xff, 0x03, 0x07, 0x09, 0x08, 0x06, 0x01, 
0xfc, 0xf9, 0xf9, 0xfb, 0xff, 0x04, 0x07, 0x08, 0x07, 0x03, 0xfe, 0xfa, 0xf9, 0xfa, 0xfd, 0x01, 0x05, 0x08, 0x08, 0x05, 0x01, 0xfc, 0xf9, 0xf8, 0xfa, 0xfe, 0x02, 0x05, 0x07, 0x06, 0x03, 0xff, 
0xfb, 0xf9, 0xf9, 0xfa, 0xfe, 0x01, 0x05, 0x06, 0x05, 0x01, 0xfe, 0xfa, 0xf7, 0xf7, 0xfa, 0xfe, 0x02, 0x05, 0x06, 0x03, 0xff, 0xfa, 0xf7, 0xf6, 0xf8, 0xfb, 0x00, 0x04, 0x05, 0x04, 0x01, 0xfd, 
0xf9, 0xf6, 0xf7, 0xfa, 0xfe, 0x02, 0x04, 0x05, 0x03, 0x00, 0xfc, 0xf8, 0xf7, 0xf9, 0xfc, 0xff, 0x02, 0x05, 0x05, 0x03, 0x00, 0xfc, 0xf8, 0xf8, 0xfa, 0xfd, 0x01, 0x04, 0x06, 0x05, 0x02, 0xff, 
0xfb, 0xf9, 0xf8, 0xfb, 0xff, 0x03, 0x06, 0x07, 0x05, 0x02, 0xfe, 0xfb, 0xf9, 0xf9, 0xfc, 0x00, 0x05, 0x08, 0x07, 0x05, 0x01, 0xfd, 0xfa, 0xf9, 0xfc, 0x00, 0x04, 0x07, 0x08, 0x07, 0x04, 0x00, 
0xfc, 0xfa, 0xfb, 0xfd, 0x00, 0x03, 0x06, 0x07, 0x05, 0x02, 0x00, 0xfd, 0xfb, 0xfb, 0xfd, 0x00, 0x03, 0x06, 0x06, 0x05, 0x01, 0xfe, 0xfa, 0xf9, 0xfa, 0xfd, 0x01, 0x04, 0x06, 0x06, 0x03, 0x00, 
0xfc, 0xf9, 0xf9, 0xfb, 0xfe, 0x02, 0x05, 0x06, 0x05, 0x02, 0xfd, 0xfa, 0xf8, 0xf8, 0xfb, 0xff, 0x03, 0x05, 0x05, 0x03, 0x00, 0xfc, 0xf9, 0xf8, 0xfa, 0xfd, 0x01, 0x04, 0x05, 0x05, 0x02, 0xfe, 
0xfa, 0xf9, 0xf9, 0xfa, 0xfe, 0x02, 0x04, 0x05, 0x04, 0x01, 0xfd, 0xfa, 0xf8, 0xf9, 0xfb, 0xfe, 0x02, 0x04, 0x05, 0x02, 0xff, 0xfb, 0xf9, 0xf8, 0xfa, 0xfd, 0x01, 0x04, 0x05, 0x04, 0x01, 0xfe, 
0xfb, 0xf9, 0xfa, 0xfc, 0xff, 0x02, 0x04, 0x06, 0x04, 0x01, 0xfe, 0xfb, 0xf9, 0xfa, 0xfd, 0x00, 0x03, 0x05, 0x05, 0x03, 0x00, 0xfd, 0xfb, 0xfa, 0xfb, 0xfe, 0x01, 0x04, 0x05, 0x05, 0x03, 0xff, 
0xfc, 0xfb, 0xfb, 0xfc, 0xff, 0x03, 0x05, 0x07, 0x06, 0x03, 0xff, 0xfc, 0xfa, 0xfb, 0xfd, 0x01, 0x04, 0x06, 0x06, 0x04, 0x01, 0xfd, 0xfb, 0xfb, 0xfc, 0xff, 0x03, 0x05, 0x06, 0x06, 0x04, 0x00, 
0xfd, 0xfb, 0xfb, 0xfd, 0x00, 0x03, 0x06, 0x07, 0x05, 0x02, 0xff, 0xfb, 0xfa, 0xfa, 0xfd, 0x01, 0x03, 0x05, 0x05, 0x04, 0x00, 0xfd, 0xfa, 0xf9, 0xfb, 0xfd, 0x01, 0x04, 0x05, 0x04, 0x02, 0xfe, 
0xfb, 0xf9, 0xf9, 0xfb, 0xff, 0x03, 0x05, 0x05, 0x04, 0x01, 0xfd, 0xfa, 0xf9, 0xfa, 0xfc, 0x00, 0x03, 0x04, 0x03, 0x02, 0xff, 0xfd, 0xfa, 0xf9, 0xfb, 0xfe, 0x01, 0x04, 0x05, 0x04, 0x02, 0xfe, 
0xfc, 0xfa, 0xfa, 0xfb, 0xff, 0x03, 0x05, 0x05, 0x04, 0x01, 0xfe, 0xfb, 0xf9, 0xfa, 0xfc, 0x00, 0x03, 0x05, 0x05, 0x03, 0xff, 0xfc, 0xfa, 0xf9, 0xfb, 0xfe, 0x01, 0x04, 0x06, 0x05, 0x02, 0xff, 
0xfc, 0xfa, 0xfa, 0xfc, 0xff, 0x03, 0x05, 0x05, 0x04, 0x01, 0xfd, 0xfb, 0xfa, 0xfb, 0xfd, 0x00, 0x03, 0x05, 0x06, 0x03, 0x00, 0xfd, 0xfa, 0xf9, 0xfb, 0xfd, 0x00, 0x04, 0x06, 0x05, 0x02, 0xff, 
0xfb, 0xf9, 0xf9, 0xfb, 0xff, 0x03, 0x05, 0x06, 0x04, 0x01, 0xfe, 0xfb, 0xfa, 0xfb, 0xfd, 0x00, 0x04, 0x05, 0x05, 0x03, 0xff, 0xfd, 0xfa, 0xfa, 0xfc, 0xfe, 0x02, 0x04, 0x05, 0x04, 0x02, 0xff, 
0xfc, 0xfb, 0xfb, 0xfd, 0x00, 0x03, 0x05, 0x05, 0x03, 0x00, 0xfe, 0xfc, 0xfb, 0xfb, 0xfe, 0x01, 0x03, 0x05, 0x05, 0x03, 0x00, 0xfd, 0xfb, 0xfa, 0xfb, 0xfd, 0x01, 0x04, 0x05, 0x04, 0x01, 0xfe, 
0xfb, 0xfa, 0xfa, 0xfc, 0x00, 0x03, 0x04, 0x05, 0x03, 0x00, 0xfd, 0xfb, 0xfa, 0xfa, 0xfd, 0x00, 0x03, 0x04, 0x03, 0x01, 0xfe, 0xfb, 0xf9, 0xfa, 0xfc, 0xff, 0x02, 0x04, 0x04, 0x03, 0x01, 0xfe, 
0xfc, 0xfb, 0xfb, 0xfd, 0x00, 0x02, 0x04, 0x04, 0x02, 0x00, 0xfd, 0xfb, 0xfa, 0xfb, 0xfd, 0x00, 0x03, 0x05, 0x04, 0x02, 0xff, 0xfc, 0xfb, 0xfb, 0xfc, 0xfe, 0x01, 0x04, 0x05, 0x03, 0x01, 0xfd, 
0xfb, 0xfa, 0xfb, 0xfd, 0x00, 0x02, 0x04, 0x03, 0x02, 0xff, 0xfd, 0xfb, 0xfa, 0xfb, 0xfe, 0x01, 0x03, 0x03, 0x03, 0x01, 0xfe, 0xfc, 0xfb, 0xfc, 0xfd, 0x00, 0x02, 0x03, 0x03, 0x01, 0x00, 0xfe, 
0xfc, 0xfc, 0xfc, 0xfe, 0x00, 0x02, 0x03, 0x03, 0x02, 0x00, 0xfe, 0xfd, 0xfc, 0xfd, 0xff, 0x01, 0x03, 0x04, 0x03, 0x02, 0x00, 0xfe, 0xfd, 0xfc, 0xfe, 0x00, 0x02, 0x04, 0x04, 0x03, 0x02, 0xff, 
0xfd, 0xfd, 0xfd, 0xff, 0x02, 0x03, 0x04, 0x04, 0x02, 0x00, 0xfe, 0xfd, 0xfd, 0xfe, 0x00, 0x02, 0x04, 0x04, 0x03, 0x01, 0x00, 0xfe, 0xfd, 0xfe, 0xfe, 0x00, 0x02, 0x02, 0x03, 0x02, 0x00, 0xff, 
0xfe, 0xfd, 0xfd, 0xfe, 0x00, 0x01, 0x03, 0x03, 0x02, 0x00, 0xfe, 0xfd, 0xfd, 0xfd, 0xff, 0x01, 0x03, 0x03, 0x02, 0x00, 0xfe, 0xfd, 0xfc, 0xfc, 0xfe, 0x00, 0x02, 0x03, 0x03, 0x02, 0x00, 0xfe, 
0xfd, 0xfd, 0xfe, 0xff, 0x00, 0x02, 0x02, 0x01, 0x00, 0xff, 0xfe, 0xfd, 0xfd, 0xfe, 0x00, 0x01, 0x02, 0x02, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xfe, 0xff, 0x01, 0x02, 0x02, 0x01, 0x00, 0xfe, 
0xfd, 0xfd, 0xfd, 0xfe, 0x00, 0x01, 0x02, 0x02, 0x01, 0xff, 0xfe, 0xfc, 0xfc, 0xfe, 0xff, 0x01, 0x02, 0x02, 0x01, 0x00, 0xfe, 0xfd, 0xfd, 0xfd, 0xfe, 0x00, 0x01, 0x02, 0x01, 0x00, 0xff, 0xfd, 
0xfd, 0xfd, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfd, 0xfd, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x01, 0x00, 0x00, 0xfe, 0xfd, 0xfc, 0xfd, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x00, 0xfe, 
0xfd, 0xfc, 0xfc, 0xfd, 0xff, 0x01, 0x02, 0x02, 0x00, 0xff, 0xfd, 0xfc, 0xfc, 0xfd, 0xff, 0x00, 0x01, 0x01, 0x01, 0xff, 0xfe, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x00, 0x00, 0xff, 0xfe, 
0xfe, 0xfe, 0xff, 0xff, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x01, 0x02, 0x02, 0x01, 0x01, 0xff, 0xfe, 0xfe, 0xfe, 0x00, 0x01, 0x02, 0x03, 0x02, 0x01, 0x00, 
0xfe, 0xfe, 0xfe, 0x00, 0x01, 0x02, 0x02, 0x02, 0x01, 0x00, 0xff, 0xfe, 0xff, 0x00, 0x01, 0x02, 0x02, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 
0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x00, 0x01, 0x00, 0xff, 0xfd, 0xfd, 0xfd, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfd, 0xfd, 0xfd, 0xfe, 0x00, 0x01, 0x01, 0x00, 0xff, 0xfe, 
0xfe, 0xfd, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 
0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x01, 0x02, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x02, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x01, 0x02, 0x01, 0x00, 0x00, 
0xff, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0xff, 0xfe, 0xff, 0xff, 0xff, 0x01, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 
0xff, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0x01, 0x01, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 
0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x00, 
0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x01, 0x01, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 
0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x01, 0x00, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x00, 0x00, 
0x00, 0xff, 0xfe, 0xfe, 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 
0x00, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x00, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 
0x00, 0xff, 0xfe, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0xff, 0xff, 0xff, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x01, 
0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x00, 0x01, 0x01, 
0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 0x01, 0x00, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 
0x01, 0xff, 0xfe, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x01, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x01, 0x00, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x01, 0x01, 0x01, 
0x00, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x01, 0x01, 0x00, 0x00, 0xff, 0xfe, 0xfd, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 
0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x00, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x00, 0x01, 0x00, 
0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x01, 0x00, 0x00, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x01, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x01, 0x01, 0x01, 
0x01, 0x00, 0xff, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 0x01, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x01, 
0x00, 0xff, 0xff, 0xff, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0xff, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0x00, 0x00, 0x00, 0x00, 
0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 
0x00, 0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x01, 0x01, 0x00, 0x00, 0xff, 0xff, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 
0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x01, 0x01, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 0x01, 
0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 
0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0xff, 0xff, 0xfe, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 
0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xfe, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 
0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 
0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 
0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 
0x00, 0x00, 0x00, 0xff, 0xff, 0xfe, 0xff, 0xff, 0x00, 0x01, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 
0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 
0xff, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 
0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff, 0xff, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 
0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 
0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0xff, 0xff, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 
0x00, 0xff, 0xff, 0x00, 0xff, 0xff, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 
0xff, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff, 
0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 
0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0xff, 0x00, 0x00, 0xff, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 
0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 
};

/*********************************/
/*    VoIP Team G.711 source code*/
/*********************************/
#define G711_L_FRAME 160		// frame size
#define G711_PERIOD	  20		// 20 ms

#define	SIGN_BIT	(0x80)		/* Sign bit for a A-law byte. */
#define	QUANT_MASK	(0xf)		/* Quantization field mask. */
#define	NSEGS		(8)			/* Number of A-law segments. */
#define	SEG_SHIFT	(4)			/* Left shift for segment number. */
#define	SEG_MASK	(0x70)		/* Segment field mask. */

#define	PITCH_MIN	40		/* minimum allowed pitch, 200 Hz */
#define PITCH_MAX	120		/* maximum allowed pitch, 66 Hz */
#define	PITCHDIFF	(PITCH_MAX - PITCH_MIN)
#define	POVERLAPMAX	(PITCH_MAX >> 2)/* maximum pitch OLA window */
#define	HISTORYLEN	(PITCH_MAX * 3 + POVERLAPMAX) /* history buffer length*/
#define	NDEC		2		/* 2:1 decimation */
#define	CORRLEN		160		/* 20 ms correlation length */
#define	CORRBUFLEN	(CORRLEN + PITCH_MAX) /* correlation buffer length */
#define CORRMINPOWER	250	/* minimum power */
#define	EOVERLAPINCR	32	/* end OLA increment per frame, 4 ms */
#define	FRAMESZ		G711_L_FRAME /* 20 ms at 8 KHz */
#define ATTENFAC	6554	/* attenuation factor (0.2) per 10 ms frame */
#define	ATTENINCR	(ATTENFAC/FRAMESZ) /* attenuation per sample */
//#define BIAS        (0x84)      /* Bias for linear code. */
#define BIAS        (33)      /* Bias for linear code. */
//static  short seg_end[8] = {0xFF, 0x1FF, 0x3FF, 0x7FF, 0xFFF, 0x1FFF, 0x3FFF, 0x7FFF};
static  short useg_end[8] = {0x3F, 0x7F, 0xFF, 0x1FF, 0x3FF, 0x7FF, 0xFFF, 0x1FFF};
static  short aseg_end[8] = {0x1F, 0x3F, 0x7F, 0xFF, 0x1FF, 0x3FF, 0x7FF, 0xFFF};

static int search(int val, short *table, int size)
{
    int     i;

    for (i = 0; i < size; i++) {
        if (val <= *table++)
            return (i);
    }
    return (size);
}

/*
 * linear2alaw() - Convert a 16-bit linear PCM value to 8-bit A-law
 *
 * linear2alaw() accepts an 16-bit integer and encodes it as A-law data.
 *
 *      Linear Input Code   Compressed Code
 *  ------------------------    ---------------
 *  0000000wxyza            000wxyz
 *  0000001wxyza            001wxyz
 *  000001wxyzab            010wxyz
 *  00001wxyzabc            011wxyz
 *  0001wxyzabcd            100wxyz
 *  001wxyzabcde            101wxyz
 *  01wxyzabcdef            110wxyz
 *  1wxyzabcdefg            111wxyz
 */
unsigned char linear2alaw(int pcm_val)        /* 2's complement (16-bit range) */
{
    int     mask;
    int     seg;
    unsigned char   aval;

    if (pcm_val >= 0) {
//        mask = 0xD5;        /* sign (7th) bit = 1 */
        mask = 0x55;        /* sign (7th) bit = 0 */
    } else {
//        mask = 0x55;        /* sign bit = 0 */       
        mask = 0xD5;        /* sign bit = 1 */        
//        pcm_val = ~pcm_val;			
        pcm_val = ~pcm_val + 1;			
    }

    /* Convert the scaled magnitude to segment number. */
//    seg = search(pcm_val, seg_end, 8);
    seg = search(pcm_val, aseg_end, 8);

    /* Combine the sign, segment, and quantization bits. */

    if (seg >= 8)       /* out of range, return maximum value. */
        return (0x7F ^ mask);
    else {
        aval = seg << SEG_SHIFT;
        if (seg < 2)
//            aval |= (pcm_val >> 4) & QUANT_MASK;
            aval |= (pcm_val >> 1) & QUANT_MASK;
        else
//            aval |= (pcm_val >> (seg + 3)) & QUANT_MASK;
            aval |= (pcm_val >> seg) & QUANT_MASK;
        return (aval ^ mask);
    }
}

/*
 * linear2ulaw() - Convert a linear PCM value to u-law
 *
 * In order to simplify the encoding process, the original linear magnitude
 * is biased by adding 33 which shifts the encoding range from (0 - 8158) to
 * (33 - 8191). The result can be seen in the following encoding table:
 *
 *  Biased Linear Input Code    Compressed Code
 *  ------------------------    ---------------
 *  00000001wxyza           000wxyz
 *  0000001wxyzab           001wxyz
 *  000001wxyzabc           010wxyz
 *  00001wxyzabcd           011wxyz
 *  0001wxyzabcde           100wxyz
 *  001wxyzabcdef           101wxyz
 *  01wxyzabcdefg           110wxyz
 *  1wxyzabcdefgh           111wxyz
 *
 * Each biased linear code has a leading 1 which identifies the segment
 * number. The value of the segment number is equal to 7 minus the number
 * of leading 0's. The quantization interval is directly available as the
 * four bits wxyz.  * The trailing bits (a - h) are ignored.
 *
 * Ordinarily the complement of the resulting code word is used for
 * transmission, and so the code word is complemented before it is returned.
 *
 */
unsigned char linear2ulaw(int pcm_val)    /* 2's complement (16-bit range) */
{
    int     mask;
    int     seg;
    unsigned char   uval;

    /* Get the sign and the magnitude of the value. */
    if (pcm_val < 0) {
        pcm_val = BIAS - pcm_val;
        mask = 0x7F;
    } else {
        pcm_val += BIAS;
        mask = 0xFF;
    }

    /* Convert the scaled magnitude to segment number. */
//    seg = search(pcm_val, seg_end, 8);
    seg = search(pcm_val, useg_end, 8);

    /*
     * Combine the sign, segment, quantization bits;
     * and complement the code word.
     */
    if (seg >= 8)       /* out of range, return maximum value. */
        return (0x7F ^ mask);
    else {
//        uval = (seg << 4) | ((pcm_val >> (seg + 3)) & 0xF);
        uval = (seg << 4) | ((pcm_val >> (seg + 1)) & 0xF);
        return (uval ^ mask);
    }
}



#ifdef INT_MODE

static int pcm_int = 0;
static void pcm_interrupt(void)
{
	unsigned int status;
	unsigned int flags;

	status = rtl_inl(PISR);
	rtl_outl(PISR,status);	
	pcm_int = 1;
	return;
}
struct irqaction pcm_irq7 = { pcm_interrupt, NULL, 7, "pcm", NULL, NULL };

#endif

void pcm_task(uint32 mode, uint32 count)
{
	int		channel,slot;

#ifdef LOOPBACK
	static int	loop=0;
#else
	int 		i;
#endif

#ifdef INT_MODE	
	unsigned int 	flags;
#endif

	/* config GPIO as PCM */
	REG32(0x200c) |=  0x00ff0000UL;
	REG32(0x2054) &= ~0x00ff0000UL;
	/* disable watchdog */
	REG32(0x203c) = 0xa5000000;
	
	printk("This is pcm task init speaking\n");
	
	/*i2c_task();		// this is for configuring TI CODEC	*/
	channel = 0;
	slot = 0;
#ifdef LOOPBACK
	for(;;)
	{
		/*int32 retry = 500;*/
		/*reset each time before testing a channel*/
		init_pcm(channel); 
		/*All channel loop back mode */
		rtl_outl(PCSCR,rtl_inl(PCSCR)| CH0ILBE);
		printk("pcm loopback channel%d slot%d\n",channel,slot);
		pcm_tx(channel,slot,loop,PKTSIZE);
		for(;;) {
			/*detect if PCM controller finished his job*/
			if((rtl_inl(CH0RXBSA+channel*4) & 3) == 0)
				break;
		}
		parse_pcm_rx(channel,4*(PKTSIZE+1));/*compare data*/
	#ifdef TEST_ALL_CHANNEL
		channel ++;
		channel &= 0x3;
	#endif
		printk("Hello! This is %d-th pcm_task speaking\n",loop++);
	}
#else 	
	/* use channel 0 to do voice playback test */
	switch(mode){
		case 0:
			printk("linear mode\n");
			mode = 0;
			break;
		case 1:
			printk("A-law mode");
			mode =  CH0CMPE|CH0Alaw;
			break;
		case 2:
			printk("u-law mode");
			mode = CH0CMPE;
		default:
			mode = 0;
		}
	init_pcm(0);
	rtl_outl(PBSIZE,PKTSIZE<<24);
	rtl_outl(PCSCR,mode| CH0RE);/* enable channel 0 rx */
	for(i=0; i<count; i++) {
	#ifdef INT_MODE	/*interrupt mode*/
		 static int static_pcm_int ;
		 save_flags(flags);cli();
		 static_pcm_int = pcm_int;
		 restore_flags(flags);
		 if(static_pcm_int) {	
	#endif
			/* channel 0, page 0 is now full of data */
			for(;;) {
				if(!(rtl_inl(CH0RXBSA) & 1))
					break;
			}
			/* put RXBUF address in CH0TXBSA */
			rtl_outl(CH0TXBSA, RXBUF_PHYADDR | PAGE0_OWN);
			rtl_outl(PCSCR, mode| CH0TE|CH0RE);
			for(;;) {
				if(!(rtl_inl(CH0TXBSA) & 1))
					break;
			}	
			rtl_outl(CH0RXBSA,RXBUF_PHYADDR | PAGE0_OWN);
			// channel 0, page 1 has data inside
			for(;;) {
				if(!(rtl_inl(CH0RXBSA) & 2))
					break;
			}
			rtl_outl(CH0TXBSA,RXBUF_PHYADDR | PAGE1_OWN);
	                rtl_outl(PCSCR,mode| CH0TE|CH0RE);	
			for(;;) {
				if(!(rtl_inl(CH0TXBSA) & 2))
					break;
			}	
			rtl_outl(CH0RXBSA,RXBUF_PHYADDR | PAGE1_OWN);
	#ifdef INT_MODE
		   }	// static_pcm_int	
		   save_flags(flags);cli();	
		   if(pcm_int)
		     pcm_int = 0;
	 	   restore_flags(flags);
	#endif			
		}	// for(;;)
#endif
}

/*Added by brian_bin*/
void memcpy_bin(void* out, void *in, int nBytes)
{
  	int i;
  	for( i = 0 ; i < nBytes ; i++ )
  	{
  		if(*((short int*)in + i ) > 0x70){
			*((char*)out+i) = 0x70;
  		}
		else if(*((short int*)in + i ) < 0xFF) {
			*((char*)out+i) = 0xFF;
		}
		else{
	  		*((char*)out+i) = *((short int*)in + i );
		}
  	}
}
void memcpy_bin2(void* out, void *in, int nBytes)
{
  	int i;
  	for( i = 0 ; i < nBytes ; i++ )
		*((char*)out+i) = *((short int*)in + i) >> 8;
	if( nBytes < PAGESIZE ){
		for( ; i < PAGESIZE ; i++ )
			*((char*)out+i) = 0;
	}
}

    /* brian_bin 0511      */
void pcm_speakTest(short int* data, int samples, int channel)
{
	int samples_spoken = 0;
	uint8* txbuf;
	uint8* rxbuf;
	
	REG32(0x200c) |=  0x00ff0000UL;/* config GPIO as PCM          */
	REG32(0x2054) &= ~0x00ff0000UL;
	REG32(0x203c)  =  0xa5000000UL;/* disable watchdog            */
	/* PGCR */
	rtl_outl(PGCR,0x1800);	/* 0->1 enable PCM */
	rtl_outl(PGCR,0x0000);	/* 1->0  reset PCM */
	rtl_outl(PGCR,0x1800);	/* 0->1 enable PCM */

	/* PGCR */
	/* enable ch0 TX, and use compender */
	//rtl_outl(PCSCR, CH0CMPE|CH0Alaw|CH0TE);
	rtl_outl(PCSCR, CH0TE);

	/* PTSAR */
	rtl_outl(PTSAR, 0x00010203UL);

	/* PBSIZE */
	/* Set Channel 0 Page size 1024 bytes, PKTSIZE = FF */
	rtl_outl(PBSIZE, PKTSIZE << 24);

	/* allocate buffer */
	txbuf = (uint8*) rtlglue_malloc(4 * (PKTSIZE+1) * 2);
	rxbuf = (uint8*) rtlglue_malloc(4 * (PKTSIZE+1) * 2);
	tx[channel].payload = (uint8*) ((uint32) txbuf | 0x20000000UL);
	rx[channel].payload = (uint8*) ((uint32) rxbuf | 0x20000000UL);	

	/* allocate buffer address through SMMU */
	REG32(CH0TXBSA + channel * 4) = (uint32) tx[channel].payload & 0x1fffffff;
	REG32(CH0RXBSA + channel * 4) = (uint32) rx[channel].payload & 0x1fffffff;
	/* set pages owned by CPU       */
	rtl_outl(CH0TXBSA, rtl_inl(CH0TXBSA) & 0xFFFFFFFC);

	/* PIMR */
	/* enable all interrupt */
	rtl_outl(PIMR,0x0000ffff);

	/*PUT data to CH0 TX Page 0 & 1*/
	printk("Saying somthing...\n");	
	for(;;){
		/*if Ch0 page0 is own by CPU*/
		if((rtl_inl(CH0TXBSA) & 1)==0){
			if((samples - samples_spoken)>=PAGESIZE){
				memcpy_bin2( tx[0].payload , data+samples_spoken , PAGESIZE );
				samples_spoken += PAGESIZE;
			}
			else {
				memcpy_bin2( tx[0].payload , data+samples_spoken , (samples - samples_spoken) );
				samples_spoken += (samples - samples_spoken);
			}
			rtl_outl(CH0TXBSA,rtl_inl(CH0TXBSA) | PAGE0_OWN);
		}
		if(samples_spoken >= samples)
			break;
		
		/*if Ch0 page1 is own by CPU*/
		if((rtl_inl(CH0TXBSA) & 2)==0){
			if((samples - samples_spoken)>=PAGESIZE){
				memcpy_bin2( tx[0].payload + PAGESIZE , data+samples_spoken , PAGESIZE );
				samples_spoken += PAGESIZE;
			}
			else {
				memcpy_bin2( tx[0].payload + PAGESIZE , data+samples_spoken , (samples - samples_spoken) );
				samples_spoken += (samples - samples_spoken);
			}
			rtl_outl(CH0TXBSA,rtl_inl(CH0TXBSA) | PAGE1_OWN);
		}
		if(samples_spoken >= samples)
			break;
	}
	printk("Samples spoken: %d\n", samples_spoken);
}

void init_pcm(int channel)
{
	rtl_outl(PGCR,0x1800);	// 0->1 enable PCM
	rtl_outl(PGCR,0x0000);	// 1->0  reset PCM
	rtl_outl(PGCR,0x1800);	// 0->1 enable PCM
#if (INT_MODE)
	printk("request PCM irq\n");
	request_irq(7, (struct irqaction *)&pcm_irq7, NULL);
#endif
	tx[channel].payload = (unsigned char*) TXBUF;
	rx[channel].payload = (unsigned char*) RXBUF;	
	
	// allocate buffer address
	rtl_outl(CH0TXBSA+channel*4,(unsigned int)tx[channel].payload & 0xffffff);
	rtl_outl(CH0RXBSA+channel*4,(unsigned int)rx[channel].payload & 0xffffff);

#if (INT_MODE)	
	printk("enable PCM interrupt\n");
	rtl_outl(PIMR,0xffffffff);
#endif

	// set RX owned by PCM controller	
	rtl_outl(CH0RXBSA+channel*4, rtl_inl(CH0RXBSA+channel*4) | 0x3);
}

void pcm_tx(int channel,int slot,int seed,int size)
{
	unsigned char *buf;

	buf = tx[channel].payload;
	/* put un meaningful data on buffer */
	prepare_pcm_data(buf,4*(size+1),seed);

	/*write size register*/
	size = size << 24;
	rtl_outl(PBSIZE,size >> (8*channel));
	/*enable channel's TX RX Loopback*/
	polltx(channel);

}
void parse_pcm_rx(int channel,int size)
{
	//unsigned char ret;
	//int i;

	flush_dcache();
	if(memcmp(tx[channel].payload,rx[channel].payload,size)) {
		printk("compare tx %x rx %x\n",(unsigned int) tx[channel].payload,
					       (unsigned int) rx[channel].payload);	
		printk("channel%d data error\n",channel);
		while(1);
	}
	else
		printk("pcm tx = rx\n");
}
// generate sine wave based on freq and data size
void prepare_pcm_data(unsigned char *buf, int size, int seed)
{
	memset(buf,0,size);
	int i;
	// get PCM (8-bit) pattern 
	for(i=0;i<size;i++) { 
		*(buf+i) = i % 255 + seed;
	}

}

void polltx(int channel)
{
	printk("poll 1 channels\n");
	/* set owned by PCM controller */
	rtl_outl(CH0TXBSA+channel*4, rtl_inl(CH0TXBSA+channel*4)|3);
	switch(channel) {
	case 0:
		/* enable channel 0 TX/RX/LOOPBACK */
		rtl_outl(PCSCR,rtl_inl(PCSCR)| CH0RE|CH0TE);
		/*memDump(0xbd280000, 16, "pcm reg");*/
		break;	
	case 1: 	
		/* enable channel 1 TX/RX/LOOPBACK */
		rtl_outl(PCSCR,rtl_inl(PCSCR)| CH1RE|CH1TE);
		break;	
	case 2: 	
		/* enable channel 2 TX/RX/LOOPBACK */
		rtl_outl(PCSCR,rtl_inl(PCSCR)| CH2RE|CH2TE);
		break;	
	case 3: 	
		/* enable channel 3 TX/RX/LOOPBACK */
		rtl_outl(PCSCR,rtl_inl(PCSCR)| CH3RE|CH3TE);
		break;	
	}
}

int32 check_pcm_rx(int channel,int size)
{
	if(memcmp(tx[channel].payload,rx[channel].payload,size)) {
		printk("compare tx %x rx %x\n",(unsigned int) tx[channel].payload,
					       (unsigned int) rx[channel].payload);	
		printk("channel%d data error\n",channel);
		return FAILED;
	}
	else
		return SUCCESS;
}

void pcm_enableChan(uint32 chan)
{
	uint8 *txbuf;
	uint8 *rxbuf;

	if(chanEnabled[chan] == TRUE)
		return;

	/* allocate buffer */
	txbuf = (uint8*) rtlglue_malloc(4 * (PKTSIZE+1) * 2);
	rxbuf = (uint8*) rtlglue_malloc(4 * (PKTSIZE+1) * 2);
	tx[chan].payload = (uint8*) ((uint32) txbuf | 0x20000000UL);
	rx[chan].payload = (uint8*) ((uint32) rxbuf | 0x20000000UL);	

	/* setup buffer size */
	REG32(PBSIZE) &= (0xff << (chan * 8));
//	REG32(PBSIZE) |= (39 << (chan * 8));	/* (39+1)*4 = 160 bytes for g729 */
	REG32(PBSIZE) |= (119 << (chan * 8));	/* (119+1)*4 = 480 bytes for g723.1 */
	
	/* setup physical base address registers */
	REG32(CH0TXBSA + chan * 4) = (uint32) tx[chan].payload & 0x1fffffff;
	REG32(CH0RXBSA + chan * 4) = (uint32) rx[chan].payload & 0x1fffffff;

	/* set rx owned by asic */
	REG32(CH0RXBSA + chan * 4) |= (PAGE0_OWN | PAGE1_OWN);

	/* enable rx */
	REG32(PCSCR) |= ((CH0TE | CH0RE | CH0CMPE) >> (chan * 8));

	chanRxCurrPage[chan] = 0;
	chanEnabled[chan] = TRUE;
}

void pcm_disableChan(uint32 chan)
{
	if(chanEnabled[chan] == TRUE)
		return;

	/* free buffer */
	rtlglue_free((void*) tx[chan].payload);
	rtlglue_free((void*) rx[chan].payload);

	/* setup buffer size */
	REG32(PBSIZE) &= (0xff << (chan * 8));
	
	/* setup base address registers */
	REG32(CH0TXBSA + chan * 4) = 0;
	REG32(CH0RXBSA + chan * 4) = 0;

	/* disable tx,rx */
	REG32(PCSCR) &= ~((CH0TE | CH0RE) >> (chan * 8));

	chanEnabled[chan] = FALSE;
}

int32 pcm_write(uint32 chan, uint8 *buf)
{
	uint8 *txbuf;
	
	if(REG32(CH0TXBSA + chan * 4) & (PAGE0_OWN << chanTxCurrPage[chan]))
		return FAILED;

	txbuf = tx[chan].payload + (chanTxCurrPage[chan] * 4 * (PKTSIZE+1));
//	printk("write to channel %d tx page%d\n", chan, chanTxCurrPage[chan]);
	memcpy(txbuf, buf, 4 * (PKTSIZE+1));/*1024Bytes at a time*/
//	memDump(txbuf, 64, "");

	/* set own bit */
	REG32(CH0TXBSA + chan * 4) |= (PAGE0_OWN << chanTxCurrPage[chan]);

	chanTxCurrPage[chan] ^= 1;
	
	return SUCCESS;
}

/*
 *  RTL865X_REG_BASE (0xBD010000)
 *  REG32(offset) (*((volatile uint32 *)(RTL865X_REG_BASE+offset)))
 *  CH0RXBSA		 (0x7020)
 *
 */
int32 pcm_write2(uint32 chan, uint8 *buf)
{
	uint8 *txbuf;
	
	if(REG32(CH0TXBSA + chan * 4) & (PAGE0_OWN << chanTxCurrPage[chan]))
		return FAILED;

	txbuf = tx[chan].payload + (chanTxCurrPage[chan] * 4 * (39+1));	//g729
//	txbuf = tx[chan].payload + (chanTxCurrPage[chan] * 4 * (119+1));//g723.1
//	printk("write to channel %d tx page%d\n", chan, chanTxCurrPage[chan]);
	memcpy(txbuf, buf, 160);	//g729
//	memcpy(txbuf, buf, 480);	//g723.1
//	memDump(txbuf, 64, "");

	/* set own bit */
	REG32(CH0TXBSA + chan * 4) |= (PAGE0_OWN << chanTxCurrPage[chan]);

	chanTxCurrPage[chan] ^= 1;/*exclusive-or operation*/
	
	return SUCCESS;
}

int32 pcm_read(uint32 chan, uint8 *buf)
{
	uint8 *rxbuf;

	if(REG32(CH0RXBSA + chan * 4) & (PAGE0_OWN << chanRxCurrPage[chan]))
		return FAILED;
	rxbuf = rx[chan].payload + (chanRxCurrPage[chan] * 4 * (PKTSIZE+1));
	memcpy(buf, (void*) rxbuf, 4 * (PKTSIZE+1));
	/* reset own bit */
	REG32(CH0RXBSA + chan * 4) |= (PAGE0_OWN << chanRxCurrPage[chan]);
	chanRxCurrPage[chan] ^= 1;
	return SUCCESS;
}

int32 pcm_read2(uint32 chan, uint8 *buf)
{
	uint8 *rxbuf;
	
	if(REG32(CH0RXBSA + chan * 4) & (PAGE0_OWN << chanRxCurrPage[chan]))
		return FAILED;

	rxbuf = rx[chan].payload + (chanRxCurrPage[chan] * 4 * (39+1));	//g729
	rxbuf = rx[chan].payload + (chanRxCurrPage[chan] * 4 * (119+1));	//g723.1
//	printk("read from channel %d rx page%d\n", chan, chanRxCurrPage[chan]);
//	memDump((void*) rxbuf, 160, "");
//	memcpy(buf, (void*) rxbuf, 160);	//g729
	memcpy(buf, (void*) rxbuf, 480);	//g723.1

	/* reset own bit */
	REG32(CH0RXBSA + chan * 4) |= (PAGE0_OWN << chanRxCurrPage[chan]);

	chanRxCurrPage[chan] ^= 1;

	return SUCCESS;
}

int32 pcm_ioctl(uint32 chan, uint32 cmd, uint32* val)
{
	uint32 reg;
	if(chan >= CH_NUM)
		return FAILED;
	switch(cmd){
		case PCM_GET_MODE:
			reg = REG32(PCSCR);
			if(!(reg & (CH0CMPE >> (chan * 8))))
				*(uint32*)val =  PCM_MODE_UNIFORM;
			else if(reg & (CH0Alaw >> (chan * 8)))
				*(uint32*)val = PCM_MODE_ALAW;
			else
				*(uint32*)val = PCM_MODE_ULAW;
			return SUCCESS;
		case PCM_GET_LOOPBACK:
			reg = REG32(PCSCR);
			if((chan == 0) && (reg & CH0ILBE))
				*(uint32*)val = TRUE;
			else
				*(uint32*)val = FALSE;
			return SUCCESS;
		case PCM_GET_TIMESLOT:
			*(uint32*)val = REG32(PTSAR);
			return SUCCESS;
		case PCM_SET_MODE:
			switch(val){
				case PCM_MODE_UNIFORM:
					REG32(PCSCR) &= ~(CH0CMPE >> (chan * 8));
					break;
				case PCM_MODE_ULAW:
					reg = REG32(PCSCR);
					reg |= (CH0CMPE >> (chan * 8));
					reg &= ~(CH0Alaw >> (chan * 8));
					REG32(PCSCR) = reg;
					break;
				case PCM_MODE_ALAW:
					reg = REG32(PCSCR);
					reg |= (CH0CMPE >> (chan * 8));
					reg |= (CH0Alaw >> (chan * 8));
					REG32(PCSCR) = reg;
					break;
				default:
					return FAILED;
				}
			return SUCCESS;
		case PCM_SET_LOOPBACK:
			if(chan == 0){
				if(val == TRUE)
					REG32(PCSCR) |= CH0ILBE;
				else
					REG32(PCSCR) &= ~CH0ILBE;
				return SUCCESS;
				}
			else
				return FALSE;
		case PCM_SET_TIMESLOT:
			if(val > 0x1f)
				return FAILED;
			reg = REG32(PTSAR);
			reg &= ~(0xff << ((CH_NUM - 1 - chan) * 8));
			reg |= (val << ((CH_NUM - 1 - chan) * 8));
			REG32(PTSAR) = reg;
			return SUCCESS;
		default:
			return FAILED;
		}
	return FAILED;
}

int32 pcm_autoTest(uint32 chan, uint32 repeat, uint32 seed)
{
	uint32 i;
	static uint8 txbuf[4 * (PKTSIZE+1)];
	static uint8 rxbuf[4 * (PKTSIZE+1)];
	int32 retry = 200;
	printk("This re865x speaking\n");
	if(chan >= CH_NUM)
		return FAILED;
	
	for(i=0; i<repeat; i++)
	{
		prepare_pcm_data(txbuf, 4 * (PKTSIZE+1), seed);
		pcm_write(chan, txbuf);
		while(retry-- > 0)
			if(pcm_read(chan, rxbuf) == SUCCESS)
				break;
		/* compare data */
		if(memcmp(txbuf, rxbuf, 4 * (PKTSIZE+1))){
			printk("%d times tested\n", i);
			return FAILED;
			}
		printk("\b%c","-\\|/"[i%4]);
	}
	return SUCCESS;
}

int32 pcm_humanTest(uint32 chan, uint32 sec)
{
	uint32 osec;
	uint32 i = 0;
	uint8 buf[4 * (PKTSIZE+1)];
	
	if(chan >= CH_NUM)
		return FAILED;

	osec = getsectime();
	while(1)
	{
		while(1)
			if(pcm_read(chan, buf) == SUCCESS)
				break;
		pcm_write(chan, buf);
		printk("\b%c","-\\|/"[((i++)>>2)%4]);
		if(getsectime() - osec > sec)
			break;
	}
	return SUCCESS;
}
void pcm_run(uint32 mode)
{
	unsigned int i;
	int retry = 10000;
	/* config GPIO as PCM */
	REG32(0x200c) |= 0x00ff0000UL;//(*((volatile uint32 *)(RTL865X_REG_BASE+offset)))
	REG32(0x2054) &= ~0x00ff0000UL;
	/* disable watchdog */
	REG32(0x203c) = 0xa5000000;
	printk("Play Ding-Dong ");

	/* reset pcm */
	rtl_outl(PGCR,PCMENABLE|PCMCLKDIR);	// 0->1 enable PCM
	rtl_outl(PGCR,0x0000);	// 1->0 reset PCM
	rtl_outl(PGCR,PCMENABLE|PCMCLKDIR);	// 0->1 enable PCM
	tx[0].payload = (uint8*) dingDong;
	
	// allocate buffer address
	rtl_outl(CH0TXBSA,(unsigned int)tx[0].payload & 0xffffff);
	rtl_outl(PBSIZE,PKTSIZE*0x01010101UL);//4 chennels setting at long
	rtl_outl(CH0TXBSA,rtl_inl(CH0TXBSA) | PAGE0_OWN | PAGE1_OWN);

	printk(" now ...");
	rtl_outl(PCSCR,CH0TE|mode);
	
	for(i=0; i<5; i++){
		 while(rtl_inl(CH0TXBSA+4) & (PAGE0_OWN | PAGE1_OWN))
		 	if(--retry <= 0)
		 		return;
		rtl_outl(CH0TXBSA,(rtl_inl(CH0TXBSA) + 4*(PKTSIZE+1)*2) | PAGE0_OWN | PAGE1_OWN);
		}
	
	printk("Finished\n");
}

int32 pcm_randTest(uint32 seed, uint32 count)
{
#define MAX_BYTES_IN_PAGE	(8000 * 2 * 3/100)	//8kHz * 2bytes * 30ms
#define MAXERR 1024
#define ERRAVE 512
#define RETRY 10000
#define CHK_INTR(ip, chan, str) \
		retry = 0; \
		while(1){ \
			if((REG32(PISR) & (ip >> (chan * 8))) == (ip >> (chan * 8))) \
				break; \
			 if(++retry > RETRY){ \
			 	printk(str); \
			 	ret = FAILED; \
				goto _exit; \
			 	} \
			 for(j=0; j<1000; j++); \
			}
	int		chan, mode, frame;
	uint8 	*buf;
	uint8	*txbsa[CH_NUM];
	uint8	*rxbsa[CH_NUM];
	uint32	pgsize;
	int		retry;
	int 		i, j;
	int32	round;
	uint32	txshift, rxshift;
	int32	ret = SUCCESS;
	
	/* config GPIO as PCM */
	REG32(0x200c) |= 0x00ff0000UL;
	REG32(0x2054) &= ~0x00ff0000UL;

	/* stop watchdog */
	REG32(0x203c) = 0xa5000000;

	/* allocate buffer */
	buf = (uint8*) rtlglue_malloc(MAX_BYTES_IN_PAGE * 2 * CH_NUM * 2);
	for(i=0; i<CH_NUM; i++){
		buf += (MAX_BYTES_IN_PAGE * 2/*pages*/);
		txbsa[i] = (uint8*) ((uint32) buf | 0x20000000UL);
		REG32(CH0TXBSA + (i * 4)) = (uint32) buf & 0x1fffffffUL;
		}
	for(i=0; i<CH_NUM; i++){
		buf += (MAX_BYTES_IN_PAGE * 2/*pages*/);
		rxbsa[i] = (uint8*) ((uint32) buf | 0x20000000UL);
		REG32(CH0RXBSA + (i * 4)) = (uint32) buf & 0x1fffffffUL;
		}
		
	printk("PCM random loopback test start ...\n");
	printk("random seed = %d\n", seed);
	tauset(0, seed);
	round = 0;
	for(;;){
		/* reset pcm every time to get rid of in-FIFO rx data */
		REG32(PGCR) = PCMENABLE|PCMCLKDIR;	// 0->1 enable PCM
		REG32(PGCR) = 0;
		REG32(PGCR) = PCMENABLE|PCMCLKDIR;	// 0->1 enable PCM
		
		/* random variables */
		chan = taurand(CH_NUM);
		printk("Channel %d ", chan);
		mode = taurand(3);
		printk("%s ", (mode == PCM_MODE_UNIFORM)? "Linear mode": 
									(mode == PCM_MODE_ULAW)? "u-law": "A-law");
		frame = taurand(2);
		printk("%s ", (frame == 0)? "20ms": "30ms");

		/* calculate page size */
		if(!frame)
			pgsize = 8000 * 2/100;
		else
			pgsize = 8000 * 3/100;
		if(mode != PCM_MODE_UNIFORM)
			pgsize *= 2;
//		printk("pgsize=%d ", pgsize);

		/* random data */
		for(i=0; i<(pgsize * 2/*pages*/); i++)
			txbsa[chan][i] = taurand(256);

		/* configure registers */
		REG32(PBSIZE) = ((pgsize >> 2) -1) << ((CH_NUM - 1 - chan) * 8);
		REG32(CH0TXBSA + (chan * 4)) |= PAGE0_OWN | PAGE1_OWN;
		REG32(CH0RXBSA + (chan * 4)) |= PAGE0_OWN | PAGE1_OWN;
		REG32(PIMR) = 0xf000 >> (chan * 4);
		REG32(PCSCR) = CH0ILBE|
			((CH0TE|CH0RE|((mode == PCM_MODE_UNIFORM)? 0: (mode == PCM_MODE_ULAW)? (CH0CMPE): (CH0CMPE|CH0Alaw))) >> (chan * 8));
		/* wait for tx page 0 ok interrupt */
		CHK_INTR(0x80000000UL, chan, "ERROR: Tx page 0 OK intr miss\n");
		/* wait for rx page 0 ok interrupt */
		CHK_INTR(0x20000000UL, chan, "ERROR: Rx page 0 OK intr miss\n");
		/* wait for tx page 1 ok interrupt */
		CHK_INTR(0x40000000UL, chan, "ERROR: Tx page 1 OK intr miss\n");
		/* wait for rx page 1 ok interrupt */
		CHK_INTR(0x10000000UL, chan, "ERROR: Rx page 1 OK intr miss\n");
		/* wait for tx page 0 unavailable interrupt */
		CHK_INTR(0x08000000UL, chan, "ERROR: Rx page 1 unavailable intr miss\n");
		/* wait for rx page 0 unavailable interrupt */
		CHK_INTR(0x02000000UL, chan, "ERROR: Rx page 1 unavailable intr miss\n");
		/* check own bits */
		retry = 0;
		while(1){
			 if((REG32(CH0RXBSA+chan*4) & (PAGE0_OWN | PAGE1_OWN)) == 0)
				break;
			 if(++retry > RETRY){
			 	printk("ERROR: Rx not complete\n");
			 	ret = FAILED;
				goto _exit;
			 	}
			 for(j=0; j<1000; j++);
			}
		
		/* compare data */
//		memDump(txbsa[chan], 64, "tx page 0");
//		memDump(txbsa[chan] + pgsize, 64, "tx page 1");
//		memDump(rxbsa[chan], 64, "rx page 0");
//		memDump(rxbsa[chan] + pgsize, 64, "rx page 1");
		if(mode == PCM_MODE_UNIFORM){
			/* rx data might be one-byte ahead or behind of tx data */
			rxshift = txshift = 0;
			if(txbsa[chan][0] != rxbsa[chan][0]){
				if(txbsa[chan][0] == rxbsa[chan][1]){
					rxshift = 1;
					printk("(1st byte inserted:%d) ", rxbsa[chan][0]);
					}
				else if(txbsa[chan][1] == rxbsa[chan][0]){
					txshift = 1;
					printk("(1st byte lost:%d) ", txbsa[chan][0]);
					}
				}
			if(memcmp(txbsa[chan] + txshift, rxbsa[chan] + rxshift, (pgsize * 2) - (txshift + rxshift))){
				printk("ERROR: Data not match\n");
				for(j=0; j<(pgsize * 2) - (txshift + rxshift); j++)
					if(txbsa[chan][j + txshift] != rxbsa[chan][j + rxshift])
						printk("byte error! tx %02x(%x) != rx %02x(%x)\n", 
								txbsa[chan][j + txshift], (uint32) txbsa[chan] + j + txshift, rxbsa[chan][j + rxshift], (uint32) rxbsa[chan] + j + rxshift);
				ret = FAILED;
				goto _exit;
				}
			}
		else{
#define TWORDOF(chan, index) (*((int16*) txbsa[chan] + index))
#define RWORDOF(chan, index) (*((int16*) rxbsa[chan] + index))
#define ERR5WORDS(chan, tidx,ridx) \
	(abs(TWORDOF(chan,tidx)-RWORDOF(chan,ridx)) + \
	abs(TWORDOF(chan,tidx+1)-RWORDOF(chan,ridx+1)) + \
	abs(TWORDOF(chan,tidx+2)-RWORDOF(chan,ridx+2)) + \
	abs(TWORDOF(chan,tidx+3)-RWORDOF(chan,ridx+3)) + \
	abs(TWORDOF(chan,tidx+4)-RWORDOF(chan,ridx+4)) \
	)
			int16 tx, rx;
			uint32 error;
			uint32 maxErr = 0;
			uint32 errAve = 0;
			uint32 err5words;
			/* rx data might be one-word ahead or behind of tx data */
			rxshift = txshift = 0;
			err5words = ERR5WORDS(chan,0,0);
			if(ERR5WORDS(chan,0,1) < err5words){
				rxshift = 1;
				printk("(1st word inserted:%d)", RWORDOF(chan,0));
				}
			else if(ERR5WORDS(chan,1,0) < err5words){
				txshift = 1;
				printk("(1st word lost:%d)", TWORDOF(chan,0));
				}
			for(j=0; j<pgsize - (txshift + rxshift); j++){
				tx = TWORDOF(chan, j + txshift);
				rx = RWORDOF(chan, j + rxshift);
				error = abs(tx - rx);
				if(error > maxErr)
					maxErr = error;
				errAve += error;
				}
			errAve /= pgsize;
			printk("Max error=%d, average error=%d ", maxErr, errAve);
			if((maxErr > MAXERR) || (errAve > ERRAVE)){
				printk("ERROR: Data not match\n");
				for(j=0; j<pgsize - (txshift + rxshift); j++){
					tx = TWORDOF(chan, j + txshift);
					rx = RWORDOF(chan, j + rxshift);
					error = abs(tx - rx);
					if(error > MAXERR)
						printk("word error! tx %04x(%x) != rx %04x(%x)\n", 
								tx, (uint32) ((uint16*) txbsa[chan] + j + txshift), rx, (uint32) ((uint16*) rxbsa[chan] + j + rxshift));
					}
				ret = FAILED;
				goto _exit;
				}
			}
		
		REG32(CH0TXBSA + (chan * 4)) |= PAGE0_OWN;
		REG32(CH0RXBSA + (chan * 4)) |= PAGE0_OWN;
		REG32(PISR) = REG32(PISR);
		/* wait for tx page 0 ok interrupt */
		CHK_INTR(0x80000000UL, chan, "ERROR: Tx page 0 OK intr miss\n");
		/* wait for rx page 0 ok interrupt */
		CHK_INTR(0x20000000UL, chan, "ERROR: Rx page 0 OK intr miss\n");
		/* wait for tx page 1 unavailable interrupt */
		CHK_INTR(0x04000000UL, chan, "ERROR: Rx page 1 unavailable intr miss\n");
		/* wait for rx page 1 unavailable interrupt */
		CHK_INTR(0x01000000UL, chan, "ERROR: Rx page 1 unavailable intr miss\n");
		
		REG32(CH0TXBSA + (chan * 4)) |= PAGE1_OWN;
		REG32(CH0RXBSA + (chan * 4)) |= PAGE1_OWN;
		REG32(PISR) = REG32(PISR);
		/* wait for tx page 1 ok interrupt */
		CHK_INTR(0x40000000UL, chan, "ERROR: Tx page 0 OK intr miss\n");
		/* wait for rx page 1 ok interrupt */
		CHK_INTR(0x10000000UL, chan, "ERROR: Rx page 0 OK intr miss\n");
		/* wait for tx page 0 unavailable interrupt */
		CHK_INTR(0x08000000UL, chan, "ERROR: Rx page 1 unavailable intr miss\n");
		/* wait for rx page 0 unavailable interrupt */
		CHK_INTR(0x02000000UL, chan, "ERROR: Rx page 1 unavailable intr miss\n");
		
		REG32(PCSCR) = 0;
		printk("PASS\n");

		if(++round >= count)
			break;
	}

_exit:
	rtlglue_free(buf);
	printk("%d rounds\n", round);
	return ret;
}

/*mode: 0 for u-law, 1 for a-law, 2 for linear test*/
int32 pcm_compenderTest(int32 mode)
{
/*480 bytes in 240 samples*/
#define BYTES_IN_PAGE	(8000 * 2 * 3/100)	//8kHz * 2bytes * 30ms
#define RETRY 10000
	uint8 	*buf;
	int16	*txbsa;
	int8	*rxbsa;
	int		retry;
	int 		i, j;
	int32	ret = SUCCESS;
	
	/* config GPIO as PCM */
	REG32(0x200c) |= 0x00ff0000UL;
	REG32(0x2054) &= ~0x00ff0000UL;

	/* stop watchdog */
	REG32(0x203c) = 0xa5000000;

	/* allocate buffer */
	buf   = (uint8*) rtlglue_malloc(BYTES_IN_PAGE * 2 * 2);
	txbsa = (uint16*) ((uint32) buf | 0x20000000UL);
	REG32(CH0TXBSA) =  (uint32) buf & 0x1fffffffUL;
	
	buf  += (BYTES_IN_PAGE * 2/*pages*/);
	
	rxbsa =  (uint8*) ((uint32) buf | 0x20000000UL);
	REG32(CH1RXBSA) =  (uint32) buf & 0x1fffffffUL;
		
	printk("PCM encompender test start ...\n");
	
	/* reset pcm every time to get rid of in-FIFO rx data */
	REG32(PGCR) = PCMENABLE|PCMCLKDIR;	// 0->1 enable PCM
	REG32(PGCR) = 0;
	REG32(PGCR) = PCMENABLE|PCMCLKDIR;	// 0->1 enable PCM
	
	/* prepare data */
	for(i=0; i<BYTES_IN_PAGE/2; i++)
		txbsa[i] = i;
	txbsa[0] = -2460;
	txbsa[1] = -1505;
	txbsa[2] = -650;
	txbsa[3] = -338;
	txbsa[4] = -90;
	txbsa[5] = -1;
	if(mode == 0) {
		txbsa[6] = 102;
		txbsa[7] = 169;
		txbsa[8] = 420;
		txbsa[9] = 499;
		txbsa[10] = 980;
		txbsa[11] = 7000;
	}
	else if(mode == 1){
		txbsa[6] = 40;
		txbsa[7] = 102;
		txbsa[8] = 169;
		txbsa[9] = 420;
		txbsa[10] = 499;
		txbsa[11] = 980;
	}

	/* configure registers */
	REG32(PTSAR) = 0x00010203;
	REG32(PBSIZE) = (((BYTES_IN_PAGE >> 2) -1)  << 24) | ((BYTES_IN_PAGE >> 2) << 16);
	REG32(CH0TXBSA) |= PAGE0_OWN;
	REG32(CH0RXBSA + 4) |= PAGE0_OWN | PAGE1_OWN;
	switch(mode){
		case 2:
			REG32(PCSCR) = CH0TE|CH1RE;
			printk("linear mode\n");
			break;
		case 1: //a-law
			REG32(PCSCR) = CH0TE|CH1RE|CH0CMPE|CH0Alaw;
			printk("a-law mode\n");
			break;
		case 0: //u-law
		default:
			REG32(PCSCR) = CH0TE|CH1RE|CH0CMPE;
			printk("u-law mode\n");
			break;
		}
	/* end of configure registers */
	retry = 0;
	while(1){
		 if((REG32(CH1RXBSA) & (PAGE0_OWN | PAGE1_OWN)) == 0)
			break;
		 if(++retry > RETRY){
		 	printk("ERROR: Rx not complete\n");
		 	ret = FAILED;
			goto _exit;
		 	}
		 for(j=0; j<1000; j++);
	}
		
	/* compare data */
	memDump(txbsa, BYTES_IN_PAGE, "tx page 0");
	memDump(rxbsa, BYTES_IN_PAGE, "rx page 0");
//	memDump(rxbsa + BYTES_IN_PAGE, BYTES_IN_PAGE, "rx page 1");
	
	/* compute result */
	if(mode == 0){
		for(i=0; i<BYTES_IN_PAGE/2; i++)
			rxbsa[i] = linear2ulaw(txbsa[i]);
		memDump(rxbsa, BYTES_IN_PAGE, "software computation");
		}
	else if (mode == 1){
		for(i=0; i<BYTES_IN_PAGE/2; i++)
			rxbsa[i] = linear2alaw(txbsa[i]);
		memDump(rxbsa, BYTES_IN_PAGE, "software computation");
		}
	
 	printk("Fail! Check if partner setup correctly\n");
	ret = FAILED;

_exit:
	rtlglue_free(buf);
	return ret;
}

void pcm_compenderTestPartner(void)
{
	/* config GPIO as PCM */
	REG32(0x200c) |= 0x00ff0000UL;
	REG32(0x2054) &= ~0x00ff0000UL;
	
	/* reset pcm every time to get rid of in-FIFO rx data */
	REG32(PGCR) = PCMENABLE;	// 0->1 enable PCM
	REG32(PGCR) = 0;
	REG32(PGCR) = PCMENABLE|CH0_1;	// 0->1 enable PCM
	REG32(PCSCR) = CH0TE|CH0RE|CH1TE|CH1RE;
	REG32(PTSAR) = 0x00010203;
}

int pcm_probe(void)
{
	uint32 chan;

	/* config GPIO as PCM */
	REG32(0x200c) |= 0x00ff0000UL;
	REG32(0x2054) &= ~0x00ff0000UL;

	REG32(0x203c) = 0xa5000000;
	
	printk("Initialize PCM interface ...\n");
 	REG32(PGCR) = PCMENABLE | PCMCLKDIR;	// 0->1 enable PCM
	REG32(PGCR) = 0;	// 1->0 reset PCM
 	REG32(PGCR) = PCMENABLE | PCMCLKDIR;	// 0->1 enable PCM

	for (chan=0; chan<CH_NUM; chan++)
		pcm_enableChan(chan);

	return 0;
}

void pcm_exit(void)
{
	uint32 chan;

	printk("Exit PCM ...");
	for (chan=0; chan<CH_NUM; chan++)
		pcm_disableChan(chan);
	REG32(PGCR) = 0;	// 1->0 reset PCM
}


