/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : Switch core table driver fowrading engine rtl8651_tblDrvFwd.c
* Abstract : 
* Creator : Edward Jin-Ru Chen (jzchen@realtek.com.tw)
* Author :  
* $Id: rtl8651_tblDrvFwd.c,v 1.3 2006/09/21 13:00:02 blue Exp $
*
*/

/*	@doc RTL8651_FORWARDINGENGINE_API

	@module rtl8651_tblDrvFwd.c - RTL8651 Home gateway controller forwarding Engine API documentation	|
	This document explains the API interface of the forwarding engine module.
	@normal Chun-Feng Liu (cfliu@realtek.com.tw) <date>

	Copyright <cp>2003 Realtek<tm> Semiconductor Cooperation, All Rights Reserved.

 	@head3 List of Symbols |
 	Here is a list of all functions and variables in this module.

 	@index | RTL8651_FORWARDINGENGINE_API
*/

/********************************************************
	Compile flags for various features supported.
********************************************************/

#include "rtl_types.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_multicast_local.h"
#include "rtl8651_ipQueue_local.h"
#include "rtl8651_layer2local.h"
#include "rtl8651_dos_local.h"
#include "rtl8651_dos.h"
#include "asicRegs.h"
#include "rtl8651_tblAsicDrv.h"
#ifdef CONFIG_RTL865X_ROMEPERF
#include "../romeperf.h"
#endif
#ifdef CONFIG_RTL865X_IPSEC
#include "rtl8651_ipsec_local.h"
#endif
#include "rtl8651_dns_local.h"
#include "rtl8651_dns_domainBlock_local.h"

#include "rtl8651_debug.h"

/**********************************
  * Internal symbol definition
***********************************/
/* napt process */
#define	_RTL8651_L4_NAT_INPUT	1
#define	_RTL8651_L4_NAT_OUTPUT	2

#define _RTL8651_NAT_PASS		0 	/* change mbuf */
#define _RTL8651_NAT_IGNORE	1	/* not change mbuf */
#define _RTL8651_NAT_DROP		2   	/* drop packet */

#define FWDENG_INTERNAL_VARIABLES
/**********************************************************************************************************/
/********************            Internal function Prototype and data structure,variable definition           ********************/
/**********************************************************************************************************/
/* <------------------------- MISC -------------------------> */
int8 rtl8651_tblDrvFwd_Id[] = "$Id: rtl8651_tblDrvFwd.c,v 1.3 2006/09/21 13:00:02 blue Exp $";
rtl8651_tblDrvFwdEngineCounts_t rtl8651_fwdEngineCounter;
rtl8651_fwdEngineInitPara_t rtl8651_fwdEnginePara;
int8 rtl8651_drvPacketFastSend = FALSE;			/* indicate this packet must be sent fast */
static int8 _dropCases[FWDENG_TOTALERRS];			/* drop cases registration */
static int32 (*rtl8651_pppoeInput)(struct rtl_mBuf *);	/* pppoe client hooking function */
extern  int32 *rtl8651_wanStatus;					/* wan Status of each session */

/*******************************************************************************/
/* <---------------- Switches ----------------> */
static int8		rtl8651_drvIcmpReplyEnable = 0;
static int8		rtl8651_PPPoEEnable = 0;	// enable Realtek's PPP and PPPoE processing inside forwarding engine..
static int8		rtl8651_drvRoutingEnable = 1;
int8 			rtl8651_drvIcmpRoutingMsgEnable = 1;	// enable forwarding engine to reply routing-related ICMP message
static int8		rtl8651_drvIcmpAlwaysReplyErrForFrag = 0;	// always reply icmp host unreachable message regardless of Don't Fragment bit, it is only usable when rtl8651_drvIcmpRoutingMsgEnable = TRUE
static int8		rtl8651_DMZHostIcmpPassthroughEnable = 0;	// enable gateway forwards icmp message to DMZ host, if DMZ host is set
static int8		rtl8651_drvArpProcessEnable	= 0;	// enable forwarding engine reply the arp request to gateway
static int8		rtl8651_drvArpProxyProcessEnable = 0;	// enable forwarding engine's ARP Proxy function
static int8		rtl8651_drvProcessUpnpPortBouncing = FALSE;// enable forwarding engine apply upnp in Port Bouncing process
static int8		rtl8651_drvFwdOtherL4ToDMZEnable = FALSE;// enable non-TCP/UDP/ICMP packet forward to DMZ host
static int8		rtl8651_drvFwdUnicastIGMPEnable = FALSE;// enable forwarding engine forward (dip=gateway ip)'s multicast packet insteads of process it
int8			rtl8651_drvSoftwareBroadcastEnable[RTL8651_VLANTBL_SIZE] = {FALSE}; // enable software broadcast, default FALSE because its done by ASIC
static int8		rtl8651_drvEnPortBouncingPSaction = 1;	/* enable protocol stack action in Port Bouncing function */
static int8		rtl8651_drvPktDemandRouteTrigger = 1;	/* Enable to-protocol stack packet demand route triggering */
static int8		rtl8651_drvProtocolStackDontTimeout = FALSE;	/* don't refresh idle timer for ALL packets from protocol stack ? : Default : NO */
static int8		rtl8651_drvProtocolStackErrMsgDontTimeout = TRUE;	/* don't refresh idle timer for ERROR messages ? : Default : YES */
#if	_RTL_DROP_ERROR_MESSAGE
static int8		rtl8651_drvProtocolStackErrMsgDrop = TRUE;	/* don't refresh idle timer for ERROR messages ? : Default : YES */
#endif

#ifdef _RTL_NAPT_ADVANCED
//static int8	rtl8651_drvInexactTcpNaptEnable = 0;	// enable "Hash2 only" TCP connection: If it's 1, we establish NORMAL napt connection using Hash2 only
static int8		rtl8651_drvInexactUdpNaptEnable = 0;	// enable "Hash2 only" UDP connection: If it's 1, we establish NORMAL napt connection using Hash2 only
static int8		rtl8651_drvStaticUdpNaptPortTranslation = 0;// enable "Static Port Translation" for new UDP napt flow creations. Port will be kept if possible.
static int8		rtl8651_drvStaticTcpNaptPortTranslation = 0;// enable "Static Port Translation" for new TCP napt flow creations. Port will be kept if possible.
#endif /* _RTL_NAPT_ADVANCED */
#ifdef _RTL_WEAK_NAPT_CTRL
static int8		rtl8651_drvWeakTcpNaptEnable = 0; // enable forwarding engine's "weak" tcp napt mode
#endif	/* _RTL_WEAK_NAPT_CTRL */
/* Multicast */
#ifdef CONFIG_RTL865X_MULTICAST
int8			rtl8651_drvMulticastProcessEnable = 1;	// enable forwarding engint to allow IGMP proxy/IP-multicast
static uint8	IP_MULTICAST_SYSTEM_ENABLE = TRUE;	// ip-multicast system enable/disable
#endif /* CONFIG_RTL865X_MULTICAST */
/* IP fragment */
static int8		rtl8651_drvFragmentProcessEnable = 1;	// enable forwarding engine to allow IP-fragmented packet, if it's 0, fragment packet will be blocked
static uint8	IP_QUEUE_SYSTEM_ENABLE = TRUE;	// ip-fragment system enable/disable

/*******************************************************************************/
/* <------------------------- fragment process -------------------------> */
/* fragment cache */
#ifdef _RTL_NAPT_FRAG
#undef RTL8651_FRAGCACHE_DEBUG

static uint32 _rtl8651_max_fragCache_count = DEFAULT_MAX_FRAGPKT_NUM;
static uint32 _rtl8651_max_fragCache_age = DEFAULT_MAX_FRAGCACHE_AGE;
static struct _rtl8651_frag * _rtl8651_fragCache[RTL8651_FRAGCACHE_TYPE_COUNT];

static uint32 _rtl8651_fragCacheInit(void);
static void _rtl8651_fragCacheReinit(void);
static inline uint32 _rtl8651_getFragCache(int32 FlowType, ipaddr_t org_sip, ipaddr_t org_dip, uint16 ip_id, ipaddr_t *newIp, int8 *routeIdx, rtl8651_tblDrv_naptTcpUdpFlowEntry_t **naptEntryp, uint8* dsid);
static inline _rtl8651_frag_t *_rtl8651_getFragCacheEntry(int32 FlowType, ipaddr_t org_sip, ipaddr_t org_dip, uint16 ip_id);
static inline void _rtl8651_updateFragCache(int32 FlowType, ipaddr_t org_sip, ipaddr_t org_dip, uint16 ip_id, ipaddr_t ip_addr, int8 routeIdx, rtl8651_tblDrv_naptTcpUdpFlowEntry_t *naptEntry, uint8 dsid);
void _rtl8651_fragCache_timeUpdate(uint32 secPassed);
#endif /* _RTL_NAPT_FRAG */
/* fragment queueing system */
#ifdef _RTL_QUEUE_FRAG_PKT
/* for fragment packet cache */
uint8 _rtl8651_fragQueueProcess = FALSE;
uint8 _rtl8651_fragQueueDropChain = FALSE;			/* to indicate if remain packet must be dropped or not */
#ifdef _RTL_NAPT_REDIRECT
/* Record the matched redirectIp of napt-redirect table. Record by the first fragment packet and used when trap all
queued fragment packets to protocol stack with napt-redirect match */
ipaddr_t _rtl8651_fragQueueNaptRedirect_MatchedIP = 0;
#endif /* _RTL_NAPT_REDIRECT */
uint8 _rtl8651_fragQueueProcessOk = FALSE;			/* before processing fragment system, _rtl8651_fragQueueProcessOk == FALSE. Or _rtl8651_fragQueueProcessOk == TRUE */
uint16 *_rtl8651_udpSizeThreshValue;
uint8 _rtl8651_udpBlockingEnable = FALSE;
#endif /* _RTL_QUEUE_FRAG_PKT */

/*******************************************************************************/
/* <--------------------------- fragment shortest path ---------------------------> */
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
_rtl8651_FragmentShortestPathFwdCache_t fragmentShortestPathCache[RTL8651_FRAGMENTSHORTESTPATH_TBLSIZE];

static inline void _rtl8651_fragmentShortestPathCacheSet_firstHalf(struct rtl_pktHdr *pkthdr, uint8 *l4hdr, struct ip *pip, ipaddr_t sip, ipaddr_t dip, uint32 mode, ipaddr_t aliasIp);
static inline void _rtl8651_fragmentShortestPathCache_clear(struct rtl_pktHdr *pkthdr, struct ip*pip);
static void _rtl8651_fragmentShortestPathCache_timeUpdate(uint32 passed);
static inline int32 _rtl8651_fragmentShortestPathFwd(struct rtl_pktHdr*pkthdr, struct ip *pip);
#endif	/* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
/*******************************************************************************/
/* void _modifyPPPoEPassthruVid( struct rtl_pktHdr * pkthdrPtr,
                              uint16 ethType, rtl8651_tblDrv_vlanTable_t *local_vlanp ); */

/*******************************************************************************/
/* <--------------------------- TCP Connection Control ---------------------------> */

#ifdef DRAM_FWD
__DRAM_L34_FWD static uint32 *_ageTimer[TCP_CLOSED_FLOW+1];
static uint32 *__ageTimer[TCP_CLOSED_FLOW+1] ={ 
#else
static uint32 *_ageTimer[TCP_CLOSED_FLOW+1] ={
#endif
	NULL,	/* No such state */
	NULL, 	/*TCP_LISTEN_FLOW*/
	&rtl8651GlobalControl.tcpFastTimeout, /*TCP_OUTBOUND_SYN_RCVD_FLOW*/
	&rtl8651GlobalControl.tcpFastTimeout, /*TCP_INBOUND_SYN_RCVD_FLOW*/
	NULL, /* No such state */
	 &rtl8651GlobalControl.tcpLongTimeout, /*TCP_ESTABLISHED_FLOW*/
	 &rtl8651GlobalControl.tcpMediumTimeout, /*TCP_OUTBOUND_FIN_RCVD_FLOW*/
	 &rtl8651GlobalControl.tcpMediumTimeout, /*TCP_INBOUND_FIN_RCVD_FLOW*/
	 &rtl8651GlobalControl.tcpFinTimeout /*TCP_CLOSED_FLOW*/
};

#ifdef DRAM_FWD
__DRAM_L34_FWD static int8  _inNextState[2][8];
static int8  __inNextState[2][8]={
#else
static int8  _inNextState[2][8]={
#endif
	/*pktFlags={SYN,0} 		0:No action  -1:Invalid situation*/
	{ TCP_INBOUND_SYN_RCVD_FLOW, TCP_ESTABLISHED_FLOW, 0, -1, 0,0,0,TCP_INBOUND_SYN_RCVD_FLOW/*2msl reincarnation*/},
	/*pktFlags={SYN,FIN} or {0,FIN}, 	-1: Invalid situation*/
	{ 0, TCP_INBOUND_FIN_RCVD_FLOW,TCP_INBOUND_FIN_RCVD_FLOW,-1,TCP_INBOUND_FIN_RCVD_FLOW,TCP_CLOSED_FLOW,0,0}
};

#ifdef DRAM_FWD
__DRAM_L34_FWD static int8  _outNextState[2][8];
static int8  __outNextState[2][8]={
#else
static int8  _outNextState[2][8]={
#endif
	/*pktFlags={SYN,0} 		0:No action  -1:Invalid situation*/
	{ TCP_OUTBOUND_SYN_RCVD_FLOW, 0, TCP_ESTABLISHED_FLOW, -1, 0,0,0,TCP_OUTBOUND_SYN_RCVD_FLOW/*2msl reincarnation*/},
	/*pktFlags={SYN,FIN} or {0,FIN}, 	-1: Invalid situation*/
	{ TCP_OUTBOUND_FIN_RCVD_FLOW, TCP_OUTBOUND_FIN_RCVD_FLOW,TCP_OUTBOUND_FIN_RCVD_FLOW,-1,TCP_OUTBOUND_FIN_RCVD_FLOW,0, TCP_CLOSED_FLOW,0}
};

/*******************************************************************************/
/* <--------------------------- uPnP ---------------------------> */
#ifdef _RTL_UPNP
static struct _RTL8651_UPNP_ENTRY *upnp_tb;

static void _rtl8651_initUpnpDb(void);
static void _rtl8651_ReinitUpnpDb(void);
static inline int32 _rtl8651_findInboundUpnpMapping(	uint32 isTcp,
														ipaddr_t remote_ip, uint16 remote_port,
														ipaddr_t alias_ip, uint16 alias_port,
														ipaddr_t *local_ip, uint16 *local_port,
														uint32 *flags, uint32 *algIdx);
static inline int32 _rtl8651_findOutboundUpnpMapping(	uint32 isTcp,
														ipaddr_t IntIp, uint16 IntPort,
														ipaddr_t *ExtIp, uint16 *ExtPort,
														ipaddr_t RemoteIp, uint16 RemotePort,
														uint32 *flags, uint32 *algIdx);

static inline int32 _rtl8651_checkUsableUpnpEntry(	uint32 isTcp,
													ipaddr_t IntIp, uint32 IntPort,
													ipaddr_t ExtIp, uint32 ExtPort,
													ipaddr_t RemoteIp, uint32 RemotePort);


#endif /* _RTL_UPNP */

/*******************************************************************************/
/* <--------------------------- Trigger Port ---------------------------> */
#ifdef _RTL_TRIGGER_PORT

#undef TRIGGER_PORT_DEBUG

/* Trigger Port entry */
static uint32 triggerPort_entry_cnt = 0;
static _rtl8651_triggerPort_entry_t *triggerPort_entry_tbl = NULL;
static _rtl8651_triggerPort_entry_t *triggerPort_entry_freeList = NULL;
static _rtl8651_triggerPort_entry_t *triggerPort_entry_Hdr[RTL8651_MAX_DIALSESSION] = {NULL};
/* Trigger Port cache */
static uint32 triggerPort_cache_cnt = 0;
static _rtl8651_triggerPort_cache_t *triggerPort_cache_tbl = NULL;
static _rtl8651_triggerPort_cache_t *triggerPort_cache_freeList = NULL;
static _rtl8651_triggerPort_cache_t *triggerPort_cache_lruHdr = NULL;
static _rtl8651_triggerPort_cache_t *triggerPort_cache_lruTail = NULL;
#ifdef DRAM_FWD
__DRAM_L34_FWD static _rtl8651_triggerPort_cache_t *triggerPort_cache_hashTbl[RTL8651_TRIGGER_PORT_HASH_CACHE_TBLSIZE];
#else
static _rtl8651_triggerPort_cache_t *triggerPort_cache_hashTbl[RTL8651_TRIGGER_PORT_HASH_CACHE_TBLSIZE] = {NULL};
#endif
static _rtl8651_triggerPort_cache_t *triggerPort_cache_staticEntry = NULL;	/* header of static trigger port entry linked list */

static int32 _rtl8651_TriggerPortInit(rtl8651_fwdEngineInitPara_t *para);
static void _rtl8651_TriggerPortReinit(void);

static void _rtl8651_flushNetTriggerPortRules(uint32 netIntfId);
static int32 _rtl8651_addTriggerPortRule(uint32 netIntfId, uint8 inType, uint16 inPortStart, uint16 inPortEnd, _rtl8651_triggerPort_param_t *mappingInfo);

static inline void  _rtl8651_TriggerPort_FreeCache(uint8 isStatic, _rtl8651_triggerPort_cache_t *cache);
static inline void _rtl8651_TriggerPort_FreeEntry(_rtl8651_triggerPort_entry_t *entry);

static inline _rtl8651_triggerPort_cache_t * _rtl8651_TriggerPort_getCache(void);

static inline _rtl8651_triggerPort_cache_t *_rtl8651_getTriggerPortCache(uint8 refreshAge, uint8 isTcp, uint32 netIntfId, ipaddr_t extIp, uint16 extPort);
static inline void _rtl8651_addTriggerPortCache(_rtl8651_triggerPort_entry_t *entry, ipaddr_t extIp, ipaddr_t localIpAddr);

static inline int32 _rtl8651_tryActiveTriggerPort(uint8 isTcp, uint32 netIntfId, uint16 outPort, ipaddr_t extIp, ipaddr_t localIpAddr);
static inline int32 _rtl8651_getTriggeredIpAddr(uint8 isTcp, uint32 netIntfId, ipaddr_t extIp, uint16 inPort, ipaddr_t *localIpAddrp);

static inline int32 _rtl8651_checkUsableTriggeredPortEntry(uint32 isTcp, uint32 netIntfId, ipaddr_t intIp, uint32 extPort);

static void _rtl8651_TriggerPortTimeUpdate(uint32 secPassed);

#endif

/*******************************************************************************/
/* <--------------------------- Port Bouncing ---------------------------> */
#ifdef _RTL_PORT_BOUNCING
static struct _RTL8651_PORT_BOUNCING_ENTRY *portBouncingTable;
void _rtl8651_flushPortBouncingEntry(void);
static struct _RTL8651_PORT_BOUNCING_ENTRY * _rtl8651_addPortBouncingEntry(uint8 flowType,ipaddr_t cip,uint16 cport,ipaddr_t gip,uint16 gport,ipaddr_t sip,uint16 sport, uint16 id, uint32 tcpFlag, uint16 age);
void _rtl8651_portBouncingTimeUpdate(uint32 secPassed);
static int32 _rtl8651_l34PortBouncingIcmpErrMsg(struct rtl_pktHdr *, struct ip *, struct icmp *);
static inline int32 _rtl8651_l4NaptPortBouncing(struct rtl_pktHdr*, struct ip*);
static inline int32 _rtl8651_l4NatPortBouncing(struct rtl_pktHdr*, struct ip*);
static inline uint32 _rtl8651_portBouncingTcpMonitor(struct tcphdr *, uint32 *, int8);
#endif /* _RTL_PORT_BOUNCING */

/*******************************************************************************/
/* <--------------------------- MSS Modification ---------------------------> */
int32 rtl8651_mss=1460; //tony: default value for static & dhcpc connection mode
uint16 pppoe_mss[8]={1452,1452,1452,1452,1452,1452,1452,1452};

/*******************************************************************************/
/* <--------------------------- Protocol Stack Action ---------------------------> */
#ifdef _RTL_PS_ACTION
rtl8651_PS_Action_Entry_t *rtl8651_ProtoStackActions;

#if	_RTL_DROP_ERROR_MESSAGE	
#define	PS_SEND_ERROR_PKT		0x1000
#endif

static int32 _rtl8651_doProtoStackAction(struct ip *iph);
static int32 _rtl8651_aliasProtoStackSentPkt(uint32 outbound, uint32 dsid, struct ip *ip);
static int32 _rtl8651_aliasProtoStackRecvPkt(uint32 fromOutbound, struct ip *ip);
static inline rtl8651_PS_Action_Entry_t *_rtl8651_checkProtoStackRegistedPort(uint16 port, uint8 l4proto);
#endif /* _RTL_PS_ACTION */

/*******************************************************************************/
/* <--------------------------- Decision Table ---------------------------> */
#ifdef DRAM_FWD
__DRAM_L34_FWD static uint8 _RTL8651_PROC[_RTL8651_HOST_MAX][_RTL8651_HOST_MAX];
static uint8 __RTL8651_PROC[_RTL8651_HOST_MAX][_RTL8651_HOST_MAX] = {
#else
static uint8 _RTL8651_PROC[_RTL8651_HOST_MAX][_RTL8651_HOST_MAX] = {
#endif
/*						_RTL8651_HOST_NPI		_RTL8651_HOST_NI		_RTL8651_HOST_LP		_RTL8651_HOST_RP	_RTL8651_HOST_NPE		_RTL8651_HOST_NE		*/
/* _RTL8651_HOST_NPI */	{ _RTL8651_DEC_RT_NPNP,	_RTL8651_DEC_RT_NPN,	_RTL8651_DEC_RT_NPL,	_RTL8651_DEC_NPT,	_RTL8651_DEC_NPTBC_NP,	_RTL8651_DEC_NTBC_NP},
/* _RTL8651_HOST_NI  */	{ _RTL8651_DEC_RT_NNP,	_RTL8651_DEC_RT_NN,		_RTL8651_DEC_RT_NL,		_RTL8651_DEC_NT,	_RTL8651_DEC_NPTBC_N,	_RTL8651_DEC_NTBC_N},
/* _RTL8651_HOST_LP  */	{ _RTL8651_DEC_RT_LNP,	_RTL8651_DEC_RT_LN,		_RTL8651_DEC_RT_LL,		_RTL8651_DEC_RT_LR,	_RTL8651_DEC_NPTBC_L,	_RTL8651_DEC_NTBC_L},
/* _RTL8651_HOST_RP */	{_RTL8651_DEC_DP,		_RTL8651_DEC_DP,		_RTL8651_DEC_RT_RL,		_RTL8651_DEC_RT_RR,	_RTL8651_DEC_NPTR,		_RTL8651_DEC_NTR},
/* _RTL8651_HOST_NPE */	{_RTL8651_DEC_DP,		_RTL8651_DEC_DP,		_RTL8651_DEC_DP,		_RTL8651_DEC_DP,	_RTL8651_DEC_DP,		_RTL8651_DEC_DP},
/* _RTL8651_HOST_NE */	{_RTL8651_DEC_DP,		_RTL8651_DEC_DP,		_RTL8651_DEC_DP,		_RTL8651_DEC_DP,	_RTL8651_DEC_DP,		_RTL8651_DEC_DP}
};

/*******************************************************************************/
/* <--------------------------- L34 Additional execution ---------------------------> */

#ifdef DRAM_FWD
__DRAM_L34_FWD static uint8 _RTL8651_ADDITIONAL_PROC[_RTL8651_DEC_MAX];
static uint8 __RTL8651_ADDITIONAL_PROC[_RTL8651_DEC_MAX] = {0};
#else
static uint8 _RTL8651_ADDITIONAL_PROC[_RTL8651_DEC_MAX] = {0};
#endif

static void _rtl8651_l34AdditionalExec_init(void);
static inline int32 _rtl8651_l34AddtionalExec(int32 decision, ipaddr_t dip, void *pkt);

/*******************************************************************************/
/* <--------------------------- trapping packet dispatch system ---------------------------> */
#ifdef _RTL_TRAPPING_DISPATCH
_rtl8651_trapping_dispatch_entry_t *rtl8651_trappingDispatchSystem_inUsed;
_rtl8651_trapping_dispatch_entry_t *rtl8651_trappingDispatchSystem_freeHdr;

_rtl8651_trapping_dispatch_cache_t rtl8651_trappingDispatchSystem_cache;

static int32 _rtl8651_trappingDispatchInit(rtl8651_fwdEngineInitPara_t *para);
static void _rtl8651_trappingDispatchReinit(void);
static _rtl8651_trapping_dispatch_entry_t *_rtl8651_findTrappingDispatchEntry(uint8 l4Proto, ipaddr_t srcIp, ipaddr_t dstIp, int32 srcPort, int32 dstPort, uint8 isExactlyMatch);
static int32 _rtl8651_registerTrappingDispatchEntry(uint8 l4Proto, ipaddr_t srcIp, ipaddr_t dstIp, int32 srcPort, int32 dstPort, uint16 category);
static void _rtl8651_removeTrappingDispatchEntry(_rtl8651_trapping_dispatch_entry_t *delEntry);
static int32 _rtl8651_unregisterTrappingDispatchEntry(uint8 l4Proto, ipaddr_t srcIp, ipaddr_t dstIp, int32 srcPort, int32 dstPort);
static uint16 _rtl8651_getTrappingDispatchCategory(struct ip * pip);

#endif /* _RTL_TRAPPING_DISPATCH */

/*******************************************************************************/
/* <--------------------------- return value statistic ---------------------------> */
#ifdef _RTL8651_FWDENG_RETURN_STATISTIC
#define TOTAL_RETURNVALUE	(FWDENG_TOTALERRS + 3)
uint8 rtl8651_returnValue[TOTAL_RETURNVALUE];	/* FAILED, FWDENG_DROP, OTHER */
void _rtl8651_fwdEngineRetValueStatisticInit(void);
void _rtl8651_fwdEngineRetValueStatisticReset(void);
void _rtl8651_fwdEngineRetValueStatisticUpdate(int32);
int32 _rtl8651_fwdEngineRetValueStaticDump(void);
#endif /* _RTL8651_FWDENG_RETURN_STATISTIC */
/*******************************************************************************/
/* <--------------------------- Session Managment ---------------------------> */
static _rtl8651_pppoeActiveSession_entry_t * pppoeActiveSessionHead, * pppoeFreeActiveSession;
int8 rtl8651_drvPppoeNotExistedSessionTerminationEnable		= FALSE;		// the flag indicating whether pppoe not existed session termination is enabled.
void _rtl8651_freeAllPppoeActiveSession(void);
void _rtl8651_updatePppoeActiveSession(uint16 pppoeSessionId);
void _rtl8651_terminatePppoeActiveSession(uint16 pppoeSessionId);
int32 inline _rtl8651_pppoeSessionActive(uint16 pppoeSessionId);
uint16 * _rtl8651_getNextPppoeActiveSession(uint16 *curPppoeSessionId, uint16 * nextPppoeSessionId);

/*******************************************************************************/
/* <--------------------------- ALGs ---------------------------> */
static inline int32 _rtl8651_l4SeqDiff(uint32 x, uint32 y);
static int32 _rtl8651_l4GetDeltaAckIn(struct ip * pip, struct tcphdr *tc, rtl8651_tblDrv_naptTcpUdpFlowEntry_t* tb);

/*******************************************************************************/
/* <--------------------------- Internal packet process functions ---------------------------> */
static int32 _rtl8651_pktAccessGateway (ipaddr_t srcIpAddr, ipaddr_t dstIpAddr, uint16 vid);
static int32 _rtl8651_l2Relay(struct rtl_mBuf *arpMbufPtr);
static int32 _rtl8651_drvArpAgent(struct rtl_pktHdr *pkthdrPtr, struct rtl_mBuf *mbufPtr, uint16 vid);
static int32 _rtl8651_l3ActRouting(struct rtl_pktHdr * pkthdrPtr, ipaddr_t dst_ip, struct ip *ip);

static inline void _rtl8651_l34Decision(struct rtl_pktHdr *pkthdr, rtl8651_tblDrv_vlanTable_t *vlanp, uint32 *decision_p, uint32 *srcType_p, uint32 *dstType_p, uint32 *optLayer_p);
static inline int32 _rtl8651_l34PacketProcess(struct rtl_pktHdr *pkthdr, struct ip *pip, uint32 decision);
static inline int32 _rtl8651_l34ActExec(struct rtl_pktHdr *, struct ip* ,uint8, int32);
static inline int32 _rtl8651_l4napt_process(struct rtl_pktHdr *,struct ip  *ip, uint8, int32);
static inline int32 _rtl8651_l4nat_in_process(struct rtl_pktHdr *, struct ip*);
static inline int32 _rtl8651_l4nat_out_process(struct rtl_pktHdr *, struct ip*);

#ifdef _RTL_NAPT_SHORT_CUT			
static inline int32 _rtl8651_l4TcpFastAliasOut(struct rtl_pktHdr *, struct ip *pip);
static inline int32 _rtl8651_l4TcpFastAliasIn(struct rtl_pktHdr *, struct ip *pip);
static inline int32 _rtl8651_l4UdpFastAliasIn(struct rtl_pktHdr *, struct ip *pip);
static inline int32 _rtl8651_l4UdpFastAliasOut(struct rtl_pktHdr *, struct ip *pip);
#endif /* _RTL_NAPT_SHORT_CUT */

static int8 _rtl8651_l4CheckOccupiedConnection(uint32 isTcp, uint32 dsid, ipaddr_t IntIp, uint16 IntPort, ipaddr_t ExtIp, uint16 ExtPort, ipaddr_t RemoteIp, uint16 RemotePort);
static int32 _rtl8651_inboundL4DMZForward(struct rtl_pktHdr *phdr, struct ip *ip, ipaddr_t dip);
static int32 _rtl8651_outboundL4DMZForward(struct rtl_pktHdr * phdr, struct ip * ip, ipaddr_t sip);
static inline int32 _rtl8651_l4TcpMonitorIn(struct tcphdr *, rtl8651_tblDrv_naptTcpUdpFlowEntry_t *);
static inline int32 _rtl8651_l4TcpMonitorOut(struct tcphdr *, rtl8651_tblDrv_naptTcpUdpFlowEntry_t *);
static inline int32 _rtl8651_l4IcmpNaptAliasIn1(struct rtl_pktHdr *, struct ip *,struct icmp *ic);
static inline int32 _rtl8651_l4IcmpNaptAliasIn2(struct rtl_pktHdr *, struct ip *,struct icmp *ic);
static inline int32 _rtl8651_l4IcmpNaptAliasIn3(struct rtl_pktHdr *, struct ip *,struct icmp *ic);
static inline int32 _rtl8651_L4IcmpNaptAliasIn(struct rtl_pktHdr *, struct ip *);
static inline int32 _rtl8651_l4IcmpNaptAliasOut1(struct rtl_pktHdr *, struct ip *,struct icmp *);
static inline int32 _rtl8651_l4IcmpNaptAliasOut2(struct rtl_pktHdr *, struct ip *,struct icmp *);
static inline int32 _rtl8651_l4IcmpNaptAliasOut3(struct rtl_pktHdr *, struct ip *,struct icmp *);
static inline int32 _rtl8651_l4IcmpNaptAliasOut(struct rtl_pktHdr *, struct ip *);
static inline int32 _rtl8651_l4UdpNaptAliasOut(struct rtl_pktHdr *, struct ip *);
static inline int32 _rtl8651_l4UdpNaptAliasIn(struct rtl_pktHdr *, struct ip *);
static inline int32 _rtl8651_l4TcpNaptAliasOut(struct rtl_pktHdr *, struct ip *);
static inline int32 _rtl8651_l4TcpNaptAliasIn(struct rtl_pktHdr *, struct ip *);
static inline int32 _rtl8651_drvIcmpAgent(struct rtl_pktHdr *,struct ip*, uint16);
static inline int32 _rtl8651_drvIcmp(struct rtl_pktHdr *, struct ip*, struct icmp *, uint16);

static inline int32 _rtl8651_drvPktDemandRouteTrigger(struct ip *pip);

static inline int32 _rtl8651_l4UdpNatAlgAliasOut(struct rtl_pktHdr *, struct ip *, rtl8651_tblDrv_natEntry_t *);
static inline int32 _rtl8651_l4UdpNatAlgAliasIn(struct rtl_pktHdr *, struct ip *, rtl8651_tblDrv_natEntry_t *);
static inline int32 _rtl8651_l4TcpNatAlgAliasOut(struct rtl_pktHdr *, struct ip *, rtl8651_tblDrv_natEntry_t *);
static inline int32 _rtl8651_l4TcpNatAlgAliasIn(struct rtl_pktHdr *, struct ip *, rtl8651_tblDrv_natEntry_t *);
static int32 _rtl8651_uPnPOneSecondTimer(void);
static int32 _rtl8651_queuePacketSecondTimer(void);
extern int  _rtl8651_getPolicyRoutingDSID(struct rtl_pktHdr *pktHdr,uint32 ip_d,uint32 ip_s,uint8 proto,int ip_dp);

/* process packet from protocol stack */
static int32 _rtl8651_fwdEngineProtocolStackSend(void * pkthdr, int16 dvid, int32 iphdrOffset, int32 highPriority);
static void _rtl8651_fwdEnginePsFlowCacheCreate(struct ip *iphdr);

/*******************************************************************************/
/* <--------------------------- ARP Queue ---------------------------> */
#ifdef _RTL_QUEUE_ARP_PACKET
struct _RTL8651_QUEUE_PACKET_ENTRY *queue_packet_tb;

static int32 _rtl8651_protoStackActionInit(rtl8651_fwdEngineInitPara_t *para);
static int32 _rtl8651_queueInit(void);
static int32 _rtl8651_queueReinit(void);
static int32 _rtl8651_enQueuePacket(struct rtl_pktHdr * , struct ip*, ipaddr_t, uint8 );
static int32 _rtl8651_queuePacketSecondTimer(void);
static struct rtl_pktHdr *  _rtl8651_queryQuenePacket(struct ip*, uint8, uint32 *, struct ip** );
#endif /* _RTL_QUEUE_ARP_PACKET */

/*******************************************************************************/
/* <--------------------------- Packet Log ---------------------------> */
extern rtl8651_loggingFuncPtr_t _pUserLoggingFunc;
extern uint32 *_loggingModule;
#ifdef RTL8651_DSID_SUPPORT
extern uint8 *_rtl8651_acl_drop_nolog;
#endif

/*******************************************************************************/
/* <--------------------------- Passthrought ---------------------------> */
/* GRE NAPT plugin function */
int32 (*rtl8651_l4GreNaptAliasIn)(struct rtl_pktHdr *, struct ip *);
int32 (*rtl8651_l4GreNaptAliasOut)(struct rtl_pktHdr *, struct ip *);
/* ESP NAPT plugin function */
int32 (*rtl8651_l4EspNaptAliasIn)(struct rtl_pktHdr *, struct ip *);
int32 (*rtl8651_l4EspNaptAliasOut)(struct rtl_pktHdr *, struct ip *);
/* RSVP NAPT plugin function */
int32 (*rtl8651_l4RsvpNaptAliasIn)(struct rtl_pktHdr *, struct ip *);
int32 (*rtl8651_l4RsvpNaptAliasOut)(struct rtl_pktHdr *, struct ip *);

/*******************************************************************************/
/* <--------------------------- Napt Redirect ---------------------------> */
/************************************************
	Napt flow redirect function:
		Trap specific normal napt flow to protocol stack
************************************************/
#ifdef _RTL_NAPT_REDIRECT
/* napt redirect entry */
static _rtl8651_naptRedirectEntry_t *_rtl8651_naptRedirectEntry_inuse = NULL;
static _rtl8651_naptRedirectEntry_t *_rtl8651_naptRedirectEntry_freelist = NULL;
/* napt redirect flow */
static _rtl8651_naptRedirectFlow_t *_rtl8651_naptRedirectFlowCache_lru_head = NULL;
static _rtl8651_naptRedirectFlow_t *_rtl8651_naptRedirectFlowCache_lru_tail = NULL;
static _rtl8651_naptRedirectFlow_t *_rtl8651_naptRedirectFlowCache_freelist = NULL;
/* internal function declarations */
static int32 _rtl8651_naptRedirectInit(rtl8651_fwdEngineInitPara_t *para);
static void _rtl8651_naptRedirectReinit(void);
static inline int32 _rtl8651_findNaptRedirectEntry(uint8 isTcp, ipaddr_t sip, uint16 sport, ipaddr_t dip, uint16 dport, ipaddr_t *pNewDip, uint16 *pNewDport, uint8 createCache, uint16 vlanIdx);
static _rtl8651_naptRedirectEntry_t *_rtl8651_findExistNaptRedirectEntry(uint8 isExceptionList, uint8 isTcp, ipaddr_t sip, uint16 sport, ipaddr_t originalDip, uint16 originalDport);
static inline _rtl8651_naptRedirectFlow_t *_rtl8651_findNaptRedirectFlow(uint8 isTcp, ipaddr_t sip, uint16 sport, ipaddr_t originalDip, uint16 originalDport, ipaddr_t newDip, uint16 newDport, uint8 isExternalQuery, uint8 refresh);
static inline int32 _rtl8651_addNaptRedirectFlow(uint8 isTcp, ipaddr_t sip, uint16 sport, ipaddr_t originalDip, uint16 originalDport, ipaddr_t newDip, uint16 newDport);
static void _rtl8651_naptRedirectFlowTimeUpdate(uint32 secPassed);
static int32 _rtl8651_resolveProxyInboundFlow(rtl8651_tblDrv_vlanTable_t * dvlan, struct ip * pip);
static inline int32 _rtl8651_queryRedirectOutboundNaptFlow(uint8 isTcp, ipaddr_t sip, uint16 sport, ipaddr_t *pOriginalDip, uint16 *pOriginalDport, ipaddr_t newDip, uint16 newDport);

#endif /* _RTL_NAPT_REDIRECT */
/*******************************************************************************/
/* <--------------------------- PPTP/L2TP Header Refill Function ---------------------------> */
/**********************************************************************************
	pptp/l2tp header refill functions:
		refill pptp/l2tp header when packet is from pptp/l2tp interface and need to be trapped to
		protocol stack
**********************************************************************************/
#ifdef _RTL_PPTPL2TP_HDR_REFILL
_rtl8651_pptpHdrCacheEntry_t	*_rtl8651_pptpHdrCache_hashTbl[RTL8651_PPTPHDR_CACHE_HASH_TBLSIZE] = {NULL};
_rtl8651_pptpHdrCacheEntry_t	*_rtl8651_pptpHdrCache_lru_head = NULL;
_rtl8651_pptpHdrCacheEntry_t	*_rtl8651_pptpHdrCache_lru_tail = NULL;
_rtl8651_pptpHdrCacheEntry_t	*_rtl8651_pptpHdrCache_unused = NULL;
/* internal functions */
static int32 _rtl8651_pptpHdrCacheInit(rtl8651_fwdEngineInitPara_t *para);
static void _rtl8651_pptpHdrCacheReinit(void);
static void _rtl8651_pptpHdrCache_Del(_rtl8651_pptpHdrCacheEntry_t* ptr);
static inline int32 _rtl8651_pptpHdrCache_Find(uint16 ip_id, uint16 ip_off, uint32 *pSeq, uint32 *pAck, uint8 *pType);
static inline int32 _rtl8651_refillHeader(int32 result, struct rtl_pktHdr *pkthdr, struct ip*pip, rtl8651_tblDrv_vlanTable_t *svlan, uint16 srcPortList);
#endif /* _RTL_PPTPL2TP_HDR_REFILL */

#define FWDENG_EXTERNAL_FUNCTIONS
/**********************************************************************************************************/
/********************                           External  function implementation                                         ********************/
/**********************************************************************************************************/

static int32 _rtl8651_initFwdEngineParam(rtl8651_fwdEngineInitPara_t *configPara)
{

	memset(&rtl8651_fwdEnginePara, 0, sizeof(rtl8651_fwdEngineInitPara_t));

	{
		/* Fragment System */
		rtl8651_fwdEnginePara.MaxFragPktCnt =
			(configPara && configPara->MaxFragPktCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxFragPktCnt:RTL8651_DefaultDrvMaxFragPktCnt;
		rtl8651_fwdEnginePara.MaxFragSubPktCnt =
			(configPara && configPara->MaxFragSubPktCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxFragSubPktCnt:RTL8651_DefaultDrvMaxFragSubPktCnt;
		rtl8651_fwdEnginePara.MaxNegativeListEntryCnt =
			(configPara && configPara->MaxNegativeListEntryCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxNegativeListEntryCnt:RTL8651_DefaultDrvMaxNegativeListCnt;
		rtl8651_fwdEnginePara.MaxPositiveListEntryCnt =
			(configPara && configPara->MaxPositiveListEntryCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxPositiveListEntryCnt:RTL8651_DefaultDrvMaxPositiveListCnt;
		rtl8651_fwdEnginePara.MaxFragPoolCnt =
			(configPara && configPara->MaxFragPoolCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxFragPoolCnt:RTL8651_DefaultDrvMaxFragPoolCnt;
		rtl8651_fwdEnginePara.MaxFragTimeOut =
			(configPara && configPara->MaxFragTimeOut != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxFragTimeOut:RTL8651_DefaultDrvMaxFragTimeOut;
		rtl8651_fwdEnginePara.MaxNegativeListTimeOut =
			(configPara && configPara->MaxNegativeListTimeOut != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxNegativeListTimeOut:RTL8651_DefaultDrvMaxNegativeListTimeOut;

		/* Multicast System */
		rtl8651_fwdEnginePara.MCastDefaultVersion =
			(configPara && configPara->MCastDefaultVersion != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MCastDefaultVersion:/*2;//*/RTL8651_MCAST_IGMP_MAX_VERSION;
		rtl8651_fwdEnginePara.MCastProcessOption =
			(configPara && configPara->MCastProcessOption != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MCastProcessOption:RTL8651_MCAST_PROCOPT_DEFAULT;
		rtl8651_fwdEnginePara.MaxMCastCnt =
			(configPara && configPara->MaxMCastCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxMCastCnt:RTL8651_DefaultDrvMaxMCastCnt;
		rtl8651_fwdEnginePara.MaxMCastMemberCnt =
			(configPara && configPara->MaxMCastMemberCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxMCastMemberCnt:RTL8651_DefaultDrvMaxMCastMemberCnt;
		rtl8651_fwdEnginePara.MaxMCastTotalMemberCnt	 =
			(configPara && configPara->MaxMCastTotalMemberCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxMCastTotalMemberCnt:RTL8651_DefaultDrvMaxTotalMemberCnt;
		rtl8651_fwdEnginePara.MaxMCastTotalSourceCnt =
			(configPara && configPara->MaxMCastTotalSourceCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MaxMCastTotalSourceCnt:RTL8651_DefaultDrvMaxTotalSourceCnt;
		rtl8651_fwdEnginePara.MCastMember_ExpireTime =
			(configPara && configPara->MCastMember_ExpireTime != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MCastMember_ExpireTime:RTL8651_DefaultDrvMcastMemberTimeout;
		rtl8651_fwdEnginePara.MCastUpstream_Timeout =
			(configPara && configPara->MCastUpstream_Timeout != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MCastUpstream_Timeout:RTL8651_DefaultDrvMCastUpstreamTimeout;
		rtl8651_fwdEnginePara.MCastFilterEntryCnt =
			(configPara && configPara->MCastFilterEntryCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->MCastFilterEntryCnt:RTL8651_DefaultDrvMaxMCastFilterEntryCnt;
		rtl8651_fwdEnginePara.igmp_qi =
			(configPara && configPara->igmp_qi != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->igmp_qi:RTL8651_IGMP_QI;
		rtl8651_fwdEnginePara.igmp_qri =
			(configPara && configPara->igmp_qri != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->igmp_qri:/*1;//*/RTL8651_IGMP_QRI;
		rtl8651_fwdEnginePara.igmp_oqpi =
			(configPara && configPara->igmp_oqpi != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->igmp_oqpi:RTL8651_IGMP_OQPI;
		rtl8651_fwdEnginePara.igmp_sqc =
			(configPara && configPara->igmp_sqc != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->igmp_sqc:RTL8651_IGMP_SQC;
		rtl8651_fwdEnginePara.igmp_sqi =
			(configPara && configPara->igmp_sqi != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->igmp_sqi:RTL8651_IGMP_SQI;
		rtl8651_fwdEnginePara.igmp_lmqc =
			(configPara && configPara->igmp_lmqc != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->igmp_lmqc:RTL8651_IGMP_LMQC;
		rtl8651_fwdEnginePara.igmp_lmqi =
			(configPara && configPara->igmp_lmqi != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->igmp_lmqi:/*1;//*/RTL8651_IGMP_LMQI;
		rtl8651_fwdEnginePara.igmp_gmi =
			(configPara && configPara->igmp_gmi != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->igmp_gmi:RTL8651_IGMP_GMI;
		rtl8651_fwdEnginePara.igmp_group_query_interval =
			(configPara && configPara->igmp_group_query_interval != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->igmp_group_query_interval:RTL8651_IGMP_GROUP_QUERY_INTERVAL;
		rtl8651_fwdEnginePara.igmp_response_tolerance_delay =
			(configPara && configPara->igmp_response_tolerance_delay != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->igmp_response_tolerance_delay:/*3;//*/RTL8651_IGMP_RESPONSE_TOLERANCE_DELAY;

		/* DOS Detection System */
		rtl8651_fwdEnginePara.dos_SynFlood_trackCnt =
			(configPara && configPara->dos_SynFlood_trackCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->dos_SynFlood_trackCnt:RTL8651_DOS_SYN_TRACKCNT;
		rtl8651_fwdEnginePara.dos_StealthFin_trackCnt =
			(configPara && configPara->dos_StealthFin_trackCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->dos_StealthFin_trackCnt:RTL8651_DOS_FIN_TRACKCNT;
		rtl8651_fwdEnginePara.dos_UdpFlood_trackCnt =
			(configPara && configPara->dos_UdpFlood_trackCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->dos_UdpFlood_trackCnt:RTL8651_DOS_UDP_TRACKCNT;
		rtl8651_fwdEnginePara.dos_IcmpFlood_trackCnt =
			(configPara && configPara->dos_IcmpFlood_trackCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->dos_IcmpFlood_trackCnt:RTL8651_DOS_ICMP_TRACKCNT;
		rtl8651_fwdEnginePara.dos_total_trackCnt =
			(configPara && configPara->dos_total_trackCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->dos_total_trackCnt:RTL8651_DOS_TOTAL_TRACKCNT;
		rtl8651_fwdEnginePara.dos_total_portScan_trackCnt =
			(configPara && configPara->dos_total_portScan_trackCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->dos_total_portScan_trackCnt:RTL8651_DOS_SCAN_TOTAL_TRACKCNT;
		rtl8651_fwdEnginePara.dos_TcpConn_trackCnt =
			(configPara && configPara->dos_TcpConn_trackCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->dos_TcpConn_trackCnt:RTL8651_DOS_TCPCONN_TRACKCNT;
		rtl8651_fwdEnginePara.dos_UdpConn_trackCnt =
			(configPara && configPara->dos_UdpConn_trackCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->dos_UdpConn_trackCnt:RTL8651_DOS_UDPCONN_TRACKCNT;
		rtl8651_fwdEnginePara.dos_totalConn_trackCnt =
			(configPara && configPara->dos_totalConn_trackCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->dos_totalConn_trackCnt:RTL8651_DOS_TOTALCONN_TRACKCNT;


		/* DOS Process System */
		rtl8651_fwdEnginePara.dosProc_sipblock_count =
			(configPara && configPara->dosProc_sipblock_count != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->dosProc_sipblock_count:RTL8651_DOSPROC_SIPBLOCK_CNT;

		/* PPPoE Automatic Release System */
		rtl8651_fwdEnginePara.pppoeActiveSessionTrackingNumber =
			(configPara && configPara->pppoeActiveSessionTrackingNumber != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->pppoeActiveSessionTrackingNumber:4;	/* Track 4 Session : Get this number from the air. */

		/* mbuf head-reservation*/
		rtl8651_fwdEnginePara.mbufHeadroom =
			(configPara && configPara->mbufHeadroom != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->mbufHeadroom:FWDENG_MBUF_HEADROOM;

		/* Procotol Stack Action */
		rtl8651_fwdEnginePara.protoStackMaxActions =
			(configPara && configPara->protoStackMaxActions != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->protoStackMaxActions:RTL8651_MAX_PROTO_STACK_ACTIONS;

		/* NAPT-Redirect System */
#ifdef _RTL_NAPT_REDIRECT
		rtl8651_fwdEnginePara.redirectTblSize =
			(configPara && configPara->redirectTblSize != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->redirectTblSize:RTL8651_NAPTREDIRECT_DEFAULT_TABLESIZE;
		rtl8651_fwdEnginePara.redirectFlowCacheSize =
			(configPara && configPara->redirectFlowCacheSize != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
			configPara->redirectFlowCacheSize:RTL8651_NAPTREDIRECT_DEFAULT_FLOWCACHESIZE;
#endif /* _RTL_NAPT_REDIRECT */

		/* PPTP/L2TP Header-Refill process */
#ifdef _RTL_PPTPL2TP_HDR_REFILL
		rtl8651_fwdEnginePara.pptpHdrCacheSize =
		(configPara && configPara->pptpHdrCacheSize != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->pptpHdrCacheSize:RTL8651_PPTPHDR_CACHE_DEFAULT_ENTRYCNT;
#endif /* _RTL_PPTPL2TP_HDR_REFILL */

#ifdef _RTL_TRAPPING_DISPATCH
		rtl8651_fwdEnginePara.trappingPktDispatchTblSize =
		(configPara && configPara->trappingPktDispatchTblSize != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->trappingPktDispatchTblSize:RTL8651_TRAPPING_DISPATCH_DEFAULT_ENTRYCNT;
#endif /* _RTL_TRAPPING_DISPATCH */

		rtl8651_fwdEnginePara.dns_dnsDB_processCnt =
		(configPara && configPara->dns_dnsDB_processCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->dns_dnsDB_processCnt:RTL8651_DNSDB_DEFAULT_PROCESSCNT;
		rtl8651_fwdEnginePara.dns_dnsDB_tblSize =
		(configPara && configPara->dns_dnsDB_tblSize != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->dns_dnsDB_tblSize:RTL8651_DNSDB_DEFAULT_TABLESIZE;
		rtl8651_fwdEnginePara.dns_dnsDB_cnameMapTblSize =
		(configPara && configPara->dns_dnsDB_cnameMapTblSize != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->dns_dnsDB_cnameMapTblSize:RTL8651_DNSDB_DEFAULT_CNAMEMAPTABLESIZE;
		rtl8651_fwdEnginePara.dns_dnsDB_cnameMapTimeout =
		(configPara && configPara->dns_dnsDB_cnameMapTimeout != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->dns_dnsDB_cnameMapTimeout:RTL8651_DNSDB_DEFAULT_CNAMEMAPTIMEOUT;

		rtl8651_fwdEnginePara.domainBlock_tblSize =
		(configPara && configPara->domainBlock_tblSize != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->domainBlock_tblSize:RTL8651_DOMAINBLOCKING_DEFAULT_TABLESIZE;
		rtl8651_fwdEnginePara.domainBlock_maxBlockEntryCnt =
		(configPara && configPara->domainBlock_maxBlockEntryCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->domainBlock_maxBlockEntryCnt:RTL8651_DOMAINBLOCKING_DEFAULT_BLKENTRYCNT;
		rtl8651_fwdEnginePara.domainBlock_blockEntryTimeout =
		(configPara && configPara->domainBlock_blockEntryTimeout != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->domainBlock_blockEntryTimeout:RTL8651_DOMAINBLOCKING_DEFAULT_BLKENTRYTIMEOUT;

#ifdef _RTL_TRIGGER_PORT
		rtl8651_fwdEnginePara.triggerPortEntryCnt =
		(configPara && configPara->triggerPortEntryCnt != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->triggerPortEntryCnt:RTL8651_TRIGGER_PORT_DEFAULT_ENTRYCNT;
		rtl8651_fwdEnginePara.triggerPortCacheSize =
		(configPara && configPara->triggerPortCacheSize != RTL8651_FWDENGINE_PARA_UNDEFINED_VALUE)?
		configPara->triggerPortCacheSize:RTL8651_TRIGGER_PORT_DEFAULT_CACHESIZE;
#endif /* _RTL_TRIGGER_PORT */
	}

	/*
		Fragment system
	*/
	if ((!rtl8651_fwdEnginePara.MaxFragPktCnt) ||
		((!rtl8651_fwdEnginePara.MaxFragSubPktCnt) && (!rtl8651_fwdEnginePara.MaxFragPoolCnt)))
	{
		IP_QUEUE_SYSTEM_ENABLE = FALSE;
		rtl8651_drvFragmentProcessEnable = FALSE;

	}else
	{
		IP_QUEUE_SYSTEM_ENABLE = TRUE;
		rtl8651_drvFragmentProcessEnable = TRUE;
	}

#ifdef CONFIG_RTL865X_MULTICAST
	/*
		Multicast system
	*/
	if ((!rtl8651_fwdEnginePara.MaxMCastCnt) ||
		((!rtl8651_fwdEnginePara.MaxMCastMemberCnt) && (!rtl8651_fwdEnginePara.MaxMCastTotalMemberCnt)) ||
		(!rtl8651_fwdEnginePara.MaxMCastTotalSourceCnt))
	{
		IP_MULTICAST_SYSTEM_ENABLE = FALSE;
		rtl8651_drvMulticastProcessEnable = FALSE;
	}else
	{
		IP_MULTICAST_SYSTEM_ENABLE = TRUE;
		rtl8651_drvMulticastProcessEnable = FALSE;	// default disable
	}
#endif /* CONFIG_RTL865X_MULTICAST */
	return SUCCESS;
}

/* Reset the drop case setting */
static void _rtl8651_fwdEngineDropPktCaseReset(void)
{
	int32 trueFlag = TRUE;

	assert(sizeof(uint8)*256 >=FWDENG_TOTALERRS);

	/* By default, all error pkts would be sent to protocol stack. */
	memset(&_dropCases[0],0, sizeof(_dropCases));

	/* But These cases would be drop directly. */
	rtl8651_fwdEngineDropPktCase(FWDENG_VLANINVALID,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_VLANCFI,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_MACBLOCK,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_ACL_IN_DROP,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_ACL_OUT_DROP,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_IP_CANTQUEUEFRAGMENT,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_IP_BADCKSUM,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_ARP_OTHERS,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_NETIFACENOTFOUND,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_ICMP_BADCKSUM,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_BADCKSUM,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAPT_OUT_IGNORE,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAPT_IN_IGNORE,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAT_IN_ALGIGNORE,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAT_OUT_ALGIGNORE,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_BADCKSUM,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAPT_OUT_IGNORE,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAPT_IN_IGNORE,&trueFlag);	
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAT_OUT_ALGIGNORE,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAT_IN_ALGIGNORE,&trueFlag);
#ifndef RTL865X_TEST
	rtl8651_fwdEngineDropPktCase(FWDENG_ICMP_OUT_ADDFAIL, &trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_ICMP_IN_ADDFAIL, &trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAPT_OUT_ADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAPT_OUT_ADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAPT_IN_ADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAPT_IN_ADDFAIL,&trueFlag);

	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAPT_OUT_SVRALGADDFAIL, &trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAPT_OUT_CLIALGADDFAIL, &trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAPT_IN_ALGADDFAIL, &trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAT_OUT_SVRALGADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAT_OUT_CLIALGADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAT_IN_SVRALGADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_UDP_NAT_IN_CLIALGADDFAIL,&trueFlag);

	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAPT_OUT_SVRALGADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAPT_OUT_CLIALGADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAPT_IN_SVRALGADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAPT_IN_CLIALGADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAT_OUT_ADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_NAT_IN_ALGADDFAIL,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_DECISION_DROP,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_USER_DROP,&trueFlag);
	rtl8651_fwdEngineDropPktCase(FWDENG_TCP_URLFILTERED,&trueFlag);
#endif /* RTL865X_TEST */

}

#ifdef DRAM_FWD
static void _rtl8651_dramInit(void)
{
	memcpy(_ageTimer, __ageTimer, sizeof(_ageTimer));
	memcpy(_inNextState, __inNextState, sizeof(_inNextState));
	memcpy(_outNextState, __outNextState, sizeof(_outNextState));
	memset(triggerPort_cache_hashTbl, 0, sizeof(triggerPort_cache_hashTbl));
	memcpy(_RTL8651_PROC, __RTL8651_PROC, sizeof(_RTL8651_PROC));
	memcpy(_RTL8651_ADDITIONAL_PROC, __RTL8651_ADDITIONAL_PROC, sizeof(_RTL8651_ADDITIONAL_PROC));
}
#endif

/*
@func int32		| rtl8651_fwdEngineInit	| Initialize fowarding engine
@parm rtl8651_fwdEngineInitPara_t * | configPara | Initial parameters setting.
@rvalue SUCCESS	| 	module initialized successfully.
@rvalue FAILED	| 	initialization failed.
@comm 
Initialize fwd engine. Internal counters, DoS switches, default drop packet cases, ARP & PPPoE trap registered and many other variables are initialized here.
 */
int32 rtl8651_fwdEngineInit(rtl8651_fwdEngineInitPara_t * configPara)
{

	rtlglue_drvMutexLock();

	FWD_INIT_CHECK(_rtl8651_initFwdEngineParam(configPara));

	#ifdef DRAM_FWD
	_rtl8651_dramInit();
	#endif

	rtl8651_resetFwdEngineCounter();

	// NAT/NAPT library init
	/* Initialize packet aliasing software. */

	rtl8651_installArpAgent(_rtl8651_drvArpAgent);

#ifdef _RTL_NAPT_FRAG
	_rtl8651_fragCacheInit();
#endif /* _RTL_NAPT_FRAG */		

	/* asic relateive function */
	rtl8651_setNaptTcpUdpPortRange(32768,65535);
	_rtl8651_setl4NaptFindEmptyAsicRetry( 8 );
	rtl8651_enable4WayHash( TRUE );

#ifdef _RTL_UPNP	
	_rtl8651_initUpnpDb();
#endif /* RTL_UPNP */

#ifdef _RTL_TRIGGER_PORT
	FWD_INIT_CHECK(_rtl8651_TriggerPortInit(&rtl8651_fwdEnginePara));
#endif /* RTL_TRIGGER_PORT */


#ifdef _RTL_PORT_BOUNCING
	portBouncingTable = (struct _RTL8651_PORT_BOUNCING_ENTRY *)
				rtlglue_malloc(PORT_BOUNCING_ITEM  * sizeof (struct _RTL8651_PORT_BOUNCING_ENTRY));	
	_rtl8651_flushPortBouncingEntry();
#endif /* RTL_TRIGGER_PORT */

#ifdef _RTL_PPTPL2TP_HDR_REFILL
	_rtl8651_pptpHdrCacheInit(&rtl8651_fwdEnginePara);
#endif /* _RTL_PPTPL2TP_HDR_REFILL */

#ifdef _RTL_QUEUE_ARP_PACKET
	_rtl8651_queueInit();
#endif /* _RTL_QUEUE_ARP_PACKET */

#ifdef _RTL_PS_ACTION
	_rtl8651_protoStackActionInit(&rtl8651_fwdEnginePara);
#endif

#ifdef _RTL_DOS_PLUGIN
	_rtl8651_dos_init(&rtl8651_fwdEnginePara);
#endif /* _RTL_DOS_PLUGIN */

#ifdef _RTL_QUEUE_FRAG_PKT
	// Initiate Fragment system
	_rtl8651_IpFragQueueInit(&rtl8651_fwdEnginePara);
#endif /* _RTL_QUEUE_FRAG_PKT */

#ifdef CONFIG_RTL865X_MULTICAST
	// Initiate Multicast system
	_rtl8651_multicastInit(&rtl8651_fwdEnginePara);
	_rtl8651_multicast_FilterInit(&rtl8651_fwdEnginePara);
#endif /* CONFIG_RTL865X_MULTICAST */
	rtl8651_fwdEngineEnableWanRouting(TRUE);

#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
	memset(fragmentShortestPathCache, 0, (sizeof(_rtl8651_FragmentShortestPathFwdCache_t)*RTL8651_FRAGMENTSHORTESTPATH_TBLSIZE));
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */

#ifdef _RTL_NAPT_REDIRECT
	_rtl8651_naptRedirectInit(&rtl8651_fwdEnginePara);
#endif /* _RTL_NAPT_REDIRECT */

#ifdef _RTL_TRAPPING_DISPATCH
	_rtl8651_trappingDispatchInit(&rtl8651_fwdEnginePara);
#endif /* _RTL_TRAPPING_DISPATCH */

	_rtl8651_dns_init(&rtl8651_fwdEnginePara);
	_rtl8651_registerDomainBlock();					/* enable domain blocking */

	//WDS/bridging for WLAN
	rtl8651_bridgeEnable=0;

	// PPPoE Not existed session termination
	{
		_rtl8651_pppoeActiveSession_entry_t * tempPtr;
		uint32 cntr = rtl8651_fwdEnginePara.pppoeActiveSessionTrackingNumber;

		pppoeActiveSessionHead = pppoeFreeActiveSession = NULL;
		while(cntr--) {
			tempPtr = (_rtl8651_pppoeActiveSession_entry_t *) rtlglue_malloc(sizeof (_rtl8651_pppoeActiveSession_entry_t));
			tempPtr->next = pppoeFreeActiveSession;
			pppoeFreeActiveSession = tempPtr;
		}
		rtl8651_drvPppoeNotExistedSessionTerminationEnable = FALSE;
	}

#ifdef _RTL_WEAK_NAPT_CTRL
	// disable weak TCP
	rtl8651_drvWeakTcpNaptEnable = FALSE;
#endif

	/* udp blocking */
	_rtl8651_udpBlockingEnable = FALSE;

	/* turn-on always reply icmp error for fragment packet switch */
	rtl8651_drvIcmpAlwaysReplyErrForFrag = FALSE;

	/* Other L4 protocol DMZ forward */
	rtl8651_drvFwdOtherL4ToDMZEnable = FALSE;

	/* unicast IGMP forwardings */
	rtl8651_drvFwdUnicastIGMPEnable = FALSE;

	/* upnp por-tbouncing */
	rtl8651_drvProcessUpnpPortBouncing = FALSE;

	/* enable protocol stack action in Port Bouncing process */
	rtl8651_drvEnPortBouncingPSaction = TRUE;

	/* trapping packet's demand route triggering */
	rtl8651_drvPktDemandRouteTrigger = TRUE;
	/* don't refresh idle timer for ERROR messages */
	rtl8651_drvProtocolStackErrMsgDontTimeout = TRUE;
    
#if	_RTL_DROP_ERROR_MESSAGE
	/* drop pkt for ERROR messages */
	rtl8651_drvProtocolStackErrMsgDrop = TRUE;
#endif

	/* default disable sw broadcast */
	memset(&rtl8651_drvSoftwareBroadcastEnable, FALSE, (sizeof(int8) * RTL8651_VLAN_NUMBER));

#ifdef _RTL8651_FWDENG_RETURN_STATISTIC
	_rtl8651_fwdEngineRetValueStatisticInit();
#endif /* _RTL8651_FWDENG_RETURN_STATISTIC */

	/* Init drop case */
	_rtl8651_fwdEngineDropPktCaseReset();

	/* init database of additional permission checking for slow path */
	_rtl8651_l34AdditionalExec_init();

	rtlglue_drvMutexUnlock();

	return SUCCESS;
	
}

/*
@func int32		| rtl8651_fwdEngineReinit	| Re-initialize fowarding engine
@rvalue SUCCESS	| 	module re-initialized successfully.
@rvalue FAILED	| 	re-initialization failed.
@comm 
Reset fwd engine. Reset counters, DoS switches, default drop packet cases, ARP & PPPoE trap registered and many other variables are reset here.
 */
int32 rtl8651_fwdEngineReinit(void)
{

	rtlglue_drvMutexLock();

	#ifdef DRAM_FWD
	_rtl8651_dramInit();
	#endif

	rtl8651_resetFwdEngineCounter();

#ifdef _RTL_NAPT_FRAG
	_rtl8651_fragCacheReinit();
#endif /* _RTL_NAPT_FRAG */

	/* asic relateive function */
	rtl8651_setNaptTcpUdpPortRange(32768,65535);
	_rtl8651_setl4NaptFindEmptyAsicRetry( 8 );
	rtl8651_enable4WayHash( TRUE );

#ifdef _RTL_UPNP	
	_rtl8651_ReinitUpnpDb();
#endif /* RTL_UPNP */

#ifdef _RTL_TRIGGER_PORT
	_rtl8651_TriggerPortReinit();
#endif /* RTL_TRIGGER_PORT */


#ifdef _RTL_PORT_BOUNCING
	assert(portBouncingTable);
	_rtl8651_flushPortBouncingEntry();
#endif /* RTL_TRIGGER_PORT */

#ifdef _RTL_PPTPL2TP_HDR_REFILL
	_rtl8651_pptpHdrCacheReinit();
#endif /* _RTL_PPTPL2TP_HDR_REFILL */

#ifdef _RTL_QUEUE_ARP_PACKET
	_rtl8651_queueReinit();
#endif /* _RTL_QUEUE_ARP_PACKET */


#ifdef _RTL_DOS_PLUGIN
	_rtl8651_dos_Reinit();
#endif /* _RTL_DOS_PLUGIN */

#ifdef _RTL_QUEUE_FRAG_PKT
	rtl8651_IpFragQueueReinit();
#endif /* _RTL_QUEUE_FRAG_PKT */

#ifdef CONFIG_RTL865X_MULTICAST
	rtl8651_multicastReinit();
	rtl8651_multicast_FilterReinit();
#endif /* CONFIG_RTL865X_MULTICAST */

#ifdef _RTL_PS_ACTION
	rtl8651_flushProtoStackActions();
#endif /* _RTL_PS_ACTION */

#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
	memset(fragmentShortestPathCache, 0, (sizeof(_rtl8651_FragmentShortestPathFwdCache_t)*RTL8651_FRAGMENTSHORTESTPATH_TBLSIZE));
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */

#ifdef _RTL_NAPT_REDIRECT
	_rtl8651_naptRedirectReinit();
#endif /* _RTL_NAPT_REDIRECT */

#ifdef _RTL_WEAK_NAPT_CTRL
	// disable weak TCP
	rtl8651_drvWeakTcpNaptEnable = FALSE;
#endif

#ifdef _RTL_TRAPPING_DISPATCH
	_rtl8651_trappingDispatchReinit();
#endif /* _RTL_TRAPPING_DISPATCH */

	_rtl8651_dns_reinit();

	// Reinit alg-related space
	rtl8651_algReInit();

	// PPPoE Automatic release system
	{
		_rtl8651_freeAllPppoeActiveSession();		
		rtl8651_drvPppoeNotExistedSessionTerminationEnable = FALSE;
	}

	{
		uint32 index;
		rtl8651_mss=1460; 
		for(index=0; index<8; index++)
			pppoe_mss[index] = 1452;
	}

	/* turn-on always reply icmp error for fragment packet switch */
	rtl8651_drvIcmpAlwaysReplyErrForFrag = FALSE;
	
	/* udp blocking */
	_rtl8651_udpBlockingEnable = FALSE;

	/* Other L4 protocol DMZ forward */
	rtl8651_drvFwdOtherL4ToDMZEnable = FALSE;

	/* unicast IGMP forwardings */
	rtl8651_drvFwdUnicastIGMPEnable = FALSE;

	/* upnp por-tbouncing */
	rtl8651_drvProcessUpnpPortBouncing = FALSE;

	/* enable protocol stack action in Port Bouncing process */
	rtl8651_drvEnPortBouncingPSaction = TRUE;

	/* trapping packet's demand route triggering */
	rtl8651_drvPktDemandRouteTrigger = TRUE;
	/* don't refresh idle timer for ERROR messages */
	rtl8651_drvProtocolStackErrMsgDontTimeout = TRUE;

#if	_RTL_DROP_ERROR_MESSAGE
	/* drop pkt for ERROR messages */
	rtl8651_drvProtocolStackErrMsgDrop = FALSE;
#endif

	/* default disable sw broadcast */
	memset(&rtl8651_drvSoftwareBroadcastEnable, FALSE, (sizeof(int8) * RTL8651_VLAN_NUMBER));

#ifdef _RTL8651_FWDENG_RETURN_STATISTIC
	_rtl8651_fwdEngineRetValueStatisticReset();
#endif /* _RTL8651_FWDENG_RETURN_STATISTIC */

	/* Init drop case */
	_rtl8651_fwdEngineDropPktCaseReset();

	/* init database of additional permission checking for slow path */
	_rtl8651_l34AdditionalExec_init();

	rtlglue_drvMutexUnlock();

	return SUCCESS;	
}

/*
@func int32		| rtl8651_getFwdEngineCounter	| get counters in forwarding engine.
@parm rtl8651_tblDrvFwdEngineCounts_t* | counter | The variable to store counter
@rvalue SUCCESS	| 	<p getFwdEngineCounter> counter get succesffuly.
@rvalue FAILED	| 	<p counter> is point to NULL
@comm 
<p rtl8651_getFwdEngineCounter()> set the current value of forwarding engine counter to the input parameter <p counter>, return FAILED
if <p counter> is the NULL pointer
 */
int32 rtl8651_getFwdEngineCounter(rtl8651_tblDrvFwdEngineCounts_t * counter)
{

	rtlglue_drvMutexLock();

	if (counter == NULL)
	{
		rtlglue_drvMutexUnlock();
		return FAILED;
	}
	*counter = rtl8651_fwdEngineCounter;

	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_resetFwdEngineCounter	| reset counters in forwarding engine.
@rvalue SUCCESS	| 	<p resetFwdEngineCounter> always return SUCCESS
@comm 
<p rtl8651_resetFwdEngineCounter()> reset the counter in forwarding engine to zero
 */
int32 rtl8651_resetFwdEngineCounter(void)
{
	rtlglue_drvMutexLock();
	memset(&rtl8651_fwdEngineCounter, 0, sizeof(rtl8651_tblDrvFwdEngineCounts_t));
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineDropPktCase	| Drop/Trap packet by type.
@parm uint32 | fwdengDropCase | See return values of <p rtl8651_fwdEngineInput()>.
@parm int32 * | flag | NULL: Get status of <p fwdengDropCase> packet type. Non-NULL: Drop/Trap <p fwdengDropCase>.
@rvalue SUCCESS	| 	<p fwdengDropCase> status changed succesffuly.
@rvalue FAILED	| 	Failed to change status of <p fwdengDropCase>.
@rvalue 0	| 	<p fwdengDropCase> packet type would be trapped to protocol stack.
@rvalue 1	| 	<p fwdengDropCase> packet type would be dropped.
@comm 
<p rtl8651_fwdEngineInput()> by default returns packet to protocol stack if its return value is not SUCCESS.
User can use this API to drop exception packets without receiving them. For complete list of <p fwdengDropCase>, please see return values of <p rtl8651_fwdEngineInput()>.
@xref  <p rtl8651_fwdEngineInput()>
 */
int32 rtl8651_fwdEngineDropPktCase(uint32 fwdengDropCase, int32 *flag)
{

	rtlglue_drvMutexLock();

	if (fwdengDropCase>=FWDENG_TOTALERRS)
	{
		rtlglue_drvMutexUnlock();
		return FAILED;
	}
	if (!flag)
	{
		rtlglue_drvMutexUnlock();
		return _dropCases[fwdengDropCase]; //get status
	}

	/* set new status. */
	_dropCases[fwdengDropCase]= (*flag)? 1: 0;

	rtlglue_drvMutexUnlock();
	return SUCCESS;
}


/*
@func int32		| rtl8651_fwdEngineArp	| Enable/Disable forwarding engine to reply ARP requests to gateway.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwarding engine can reply ARP requests to local gateway if it is enabled here. However, even this feature is disabled, forwarding engine always auto learns sender's IP and MAC mapping in any ARP request received.
@xref  <p rtl8651_fwdEngineIcmp()>, <p rtl8651_fwdEngineProcessL34()>
*/
int32 rtl8651_fwdEngineArp(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvArpProcessEnable = enable;
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineIcmp	| Enable/Disable forwarding engine to reply ICMP echo request to gateway.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwarding engine can reply ICMP echo requests to local gateway if it is enabled here. If this feature is disabled, all ICMP requests to local gateway must be replied by upper layer protocol stack.
@xref  <p rtl8651_fwdEngineArp()>, <p rtl8651_fwdEngineIcmpRoutingMsg()>
*/
int32 rtl8651_fwdEngineIcmp(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvIcmpReplyEnable = enable;
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineIcmpRoutingMsg	| Enable/Disable forwarding engine to reply ICMP echo request to gateway.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwarding engine can reply ICMP echo requests to local gateway if it is enabled here. If this feature is disabled, all ICMP requests to local gateway must be replied by upper layer protocol stack.
@xref  <p rtl8651_fwdEngineIcmp()>, <p rtl8651_fwdEngineAlwaysReplyICMPErrForFragment()>
*/
int32 rtl8651_fwdEngineIcmpRoutingMsg(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvIcmpRoutingMsgEnable= enable;
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineAlwaysReplyICMPErrForFragment	| Enable/Disable forwarding engine to always reply ICMP host unreachable message for packet which need to be fragmented.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwarding engine will always reply ICMP host unreachable (Need Fragment) message for all packets which need to be fragmented if <p enable> is TRUE.
Or forwarding engine only reply ICMP message only when packet's "Don't Fragment" bit is set.
Note that this switch is usable only when rtl8651_fwdEngineIcmpRoutingMsg(TRUE) is called.
@xref  <p rtl8651_fwdEngineIcmpRoutingMsg()>, <p rtl8651_fwdEngineDMZHostIcmpPassThrough()>
*/
int32 rtl8651_fwdEngineAlwaysReplyICMPErrForFragment(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvIcmpAlwaysReplyErrForFrag = enable;
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineDMZHostIcmpPassThrough	| Enable/Disable forwarding engine to forward ICMP echo request to corresponding dmz host(if dmz host is set).
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
For gateway's external ip with its dmz host has been set. Forwarding engine will forward ICMP echo requests to DMZ host if this is enabled here. If this feature is disabled, ICMP requests to local gateway will always be replied by forwarding engine(if rtl8651_fwdEngineIcmp() is enabled) or gateway's protocol stack.
@xref  <p rtl8651_fwdEngineAlwaysReplyICMPErrForFragment()>, <p rtl8651_fwdEngineArpProxy()>
*/
int32 rtl8651_fwdEngineDMZHostIcmpPassThrough(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_DMZHostIcmpPassthroughEnable = enable;
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineArpProxy	| Enable/Disable forwarding engine's ARP proxy function.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
For arp requests from internal hosts, enable forwarding engine ARP Proxy for these requests
@xref  <p rtl8651_fwdEngineDMZHostIcmpPassThrough()>, <p rtl8651_fwdEngineProcessL34()>
*/
int32 rtl8651_fwdEngineArpProxy(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvArpProxyProcessEnable = enable;
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}
/*
@func int32		| rtl8651_fwdEngineProcessL34	| Process/Bypass layer3/4 packets in forwarding engine.
@parm int8 | routingEnable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
By default, forwarding engine forwards any packet received until exception occurs. User may bypass all
packets which needs to do L3~L7 operations to protocol stack if <p routingEnable>=FALSE. All layer 2 packets are still processed by fwd engine.
@xref  <p rtl8651_fwdEngineArpProxy()>, <p rtl8651_fwdEnginePPPoE()>
*/
int32 rtl8651_fwdEngineProcessL34(int8 routingEnable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvRoutingEnable = routingEnable;
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineProcessIPFragment	| Process/Bypass IP fragment packets in forwarding engine.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwading Engine will process all fragment packet process if <p enable>=TRUE.
If <p enable>=FALSE, forwarding engine will pass all fragment packet to protocol stack
@xref  <p rtl8651_fwdEngineDMZHostIcmpPassThrough()>, <p rtl8651_fwdEngineProcessIPMulticast()>
*/
int32 rtl8651_fwdEngineProcessIPFragment(int8 enable)
{
	rtlglue_drvMutexLock();
	if (IP_QUEUE_SYSTEM_ENABLE == FALSE)	// if multicast system is not initiated, always disable Multicast
	{
		rtl8651_drvFragmentProcessEnable = FALSE;
		rtlglue_drvMutexUnlock();
		return SUCCESS;
	}

	if (rtl8651_drvFragmentProcessEnable != enable)
	{
		rtl8651_IpFragQueueReinit();						// reset ipQueue system
		rtl8651_drvFragmentProcessEnable = enable;
	}

	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

#ifdef CONFIG_RTL865X_MULTICAST
/*
@func int32		| rtl8651_fwdEngineProcessIPMulticast	| Process/Bypass IP Multicast/IGMP packets in forwarding engine.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwading Engine will process all multicast packets and known IGMP packets if <p enable>=TRUE.
If <p enable>=FALSE, forwarding engine will L2-broadcast all multicast/IGMP packets and trap it to protocol stack
@xref  <p rtl8651_fwdEngineProcessIPFragment()>, <p rtl8651_fwdEngineSend()>
*/
int32 rtl8651_fwdEngineProcessIPMulticast(int8 enable)
{
	rtlglue_drvMutexLock();

	if (IP_MULTICAST_SYSTEM_ENABLE == FALSE)	// if multicast system is not initiated, always disable Multicast
	{
		rtl8651_drvMulticastProcessEnable = FALSE;

		rtlglue_drvMutexUnlock();

		return SUCCESS;
	}

	if (rtl8651_drvMulticastProcessEnable != enable)
	{
		/*
			Reset multicast system first, clear all entries
		*/
		rtl8651_multicastReinit();
		/*
			set the new state of multicast system
		*/
		rtl8651_drvMulticastProcessEnable = enable;
		/*
			Finally, set the multicast periodic Query state according to the new multicast state
		*/
		rtl8651_multicastPeriodicQuery(enable);
	}

	rtlglue_drvMutexUnlock();

	return SUCCESS;
}
#endif /* CONFIG_RTL865X_MULTICAST */

int32 rtl8651_fwdEnginePPPoE(int8 enable, int32 (*pppoeInput)(struct rtl_mBuf *)) {

	rtlglue_drvMutexLock();

	if (enable&&!pppoeInput)
	{
		rtlglue_drvMutexUnlock();
		return FAILED;
	}
	rtl8651_PPPoEEnable = enable;
	rtl8651_pppoeInput = pppoeInput;

	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEnginePppoeNotExistedSessionTermination	| Automatic terminat not existed PPPoE session enable/disable
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwading Engine will tracking active PPPoE session ID when <p rtl8651_drvPppoeNotExistedSessionTerminationEnable>= TRUE.
If <p enable>=FALSE, all previous active session ID will be released.
*/
int32 rtl8651_fwdEnginePppoeNotExistedSessionTermination(int8 enable)
{
	rtlglue_drvMutexLock();

	//If previous configuration is enabled and new setting is disabled, put all active session into free list, which means stop tracking.
	if (rtl8651_drvPppoeNotExistedSessionTerminationEnable == TRUE && enable== FALSE) {
		_rtl8651_freeAllPppoeActiveSession();		
		rtl8651_drvPppoeNotExistedSessionTerminationEnable = FALSE;
	} else 
		rtl8651_drvPppoeNotExistedSessionTerminationEnable = enable;

	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineEnableWanRouting	| Enable/Disable WAN routing. Default is on.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwading Engine will allow WAN side routing if <p enable>=TRUE.
If <p enable>=TRUE, forwarding engine will do WAN to WAN routing, otherwise, the packet is not processed by forwarding engine.
@xref  
*/
int32 rtl8651_fwdEngineEnableWanRouting(int8 enable)
{
	_rtl8651_l34AdditionalExec_t param;
	uint8 newProperty;

	rtlglue_drvMutexLock();
	_rtl8651_l34AdditionalExec_getParam(_RTL8651_HOST_RP, _RTL8651_HOST_RP, &param);
	if (enable)
	{
		newProperty = (param.property &~ _RTL8651_L34ADDTIONALEXEC_BLOCK);
	} else
	{
		newProperty = (param.property | _RTL8651_L34ADDTIONALEXEC_BLOCK);
	}

	if (newProperty != param.property)
	{
		param.property = newProperty;
		_rtl8651_l34AdditionalExec_setParam(_RTL8651_HOST_RP, _RTL8651_HOST_RP, &param);
	}
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}
/*
@func int32		| rtl8651_fwdEngineProcessPortBouncing | Enable/Disable PortBouncing. Default is on.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwarding Engine wll allow Port Bouncing packet if <p enable>=TRUE.
If <p enable>=TRUE, forwarding Engine will do port bouncing forwarding, otherwise, this packet will be trapped by forwarding engine.
@xref  
*/
int32 rtl8651_fwdEngineProcessPortBouncing(int8 enable)
{
	_rtl8651_l34AdditionalExec_t param;
	uint8 newProperty;
	int32 srcIdx, dstIdx;
	uint32 src[] = {_RTL8651_HOST_NPI, _RTL8651_HOST_NI, _RTL8651_HOST_LP};
	uint32 dst[] = {_RTL8651_HOST_NPE, _RTL8651_HOST_NE};

	rtlglue_drvMutexLock();

	if (enable == FALSE)
	{
		_rtl8651_flushPortBouncingEntry();
	}

	for (srcIdx = 0 ; srcIdx < (sizeof(src) / sizeof(uint32)) ; srcIdx ++)
	for (dstIdx = 0 ; dstIdx < (sizeof(dst) / sizeof(uint32)) ; dstIdx ++)
	{
		_rtl8651_l34AdditionalExec_getParam(src[srcIdx], dst[dstIdx], &param);
		if (enable)
		{
			newProperty = (param.property &~ _RTL8651_L34ADDTIONALEXEC_BLOCK);
		} else
		{
			newProperty = (param.property | _RTL8651_L34ADDTIONALEXEC_BLOCK);
		}

		if (newProperty != param.property)
		{
			param.property = newProperty;
			_rtl8651_l34AdditionalExec_setParam(src[srcIdx], dst[dstIdx], &param);
		}
	}

	rtlglue_drvMutexUnlock();

	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineProcessUpnpPortBouncing | Apply/Don't apply PortBouncing to upnp entry. Default is no.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwarding Engine wll check uPNP mapping when it do Port Bouncing packet if <p enable>=TRUE.
@xref  
*/
int32 rtl8651_fwdEngineProcessUpnpPortBouncing(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvProcessUpnpPortBouncing = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineForwardOtherL4ToDMZ | Enable/Disable Forward general L4 packet to DMZ. Default is off.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwarding Engine wll forward unknown non-TCP/UDP/ICMP/IGMP packets to DMZ if <p enable>=TRUE and DMZ host is set.
otherwise they will be trapped by forwarding engine. Note that forwarding engine only handle the L3 checksum mdification
for destination ip change.
@xref  
*/
int32 rtl8651_fwdEngineFwdGeneralL4ToDMZ(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvFwdOtherL4ToDMZEnable = enable;
	rtlglue_drvMutexUnlock();
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineFwdUnicastIGMPPkt | Choose if Forwarding engine need to forward Unicast IGMP packet or process it. Default is off.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwarding engine will forward the IGMP packet with UNICAST address if <p enable> = TRUE.
Note that if user enable this function, RFC 3376 (IGMPv3) 4.1.12 , 4.2.14 will be discarded,
and the multicast filter mechanism is ignored.
@xref  
*/
int32 rtl8651_fwdEngineFwdUnicastIGMPPkt(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvFwdUnicastIGMPEnable = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEnginePortBouncingDoProtocolStackAction | Choose if Forwarding engine apply protocol stack action module checking in Port Bouncing process path.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwarding engine will try to apply protocol stack action module in NAPT port bouncing process (Internal host --> Gateway NAPT IP) if <p enable> = TRUE.
@xref  
*/
int32 rtl8651_fwdEnginePortBouncingDoProtocolStackAction(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvEnPortBouncingPSaction = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineTrapPktDemandRouteTrigger | Choose if Forwarding engine want to trigger demand route when got packet which is sent to protocol stack.
@parm int8 | enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm 
Forwarding engine will try to trigger demand route when some special packet is trapped to protocol stack  if <p enable> = TRUE.
Now, this module would trigger:
	1. DNS packet ( dport 53, TCP/UDP )
@xref  
*/
int32 rtl8651_fwdEngineTrapPktDemandRouteTrigger(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvPktDemandRouteTrigger = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}

#ifdef _RTL_NAPT_ADVANCED
/*
@func int32		| rtl8651_fwdEngineInexactUdpFlow | Enable/Disable UDP Remote-Host-Regardless(Inexact) connection establish. Default is off.
@parm int8		| enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm
Forwarding Engine wll generate Normal UDP flow without consider remote host's IP/Port if <p enable>=TRUE.
For one NORMAL new UDP flow establishment, if <p enable>=TRUE, forwarding engine will establish this flow regardless of
it's remote host's IP/port. In other words, if this flow is established, system will allow other Remote IP/port to connect
to this entry via the external IP/port of this flow. Otherwise, forwarding engine will establish NORMAL new UDP flow with
the full information about internal/external/remote IP and port.
@xref  
*/
int32 rtl8651_fwdEngineInexactUdpFlow(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvInexactUdpNaptEnable = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}
#if 0
/*
@func int32		| rtl8651_fwdEngineInexactTcpFlow | Enable/Disable TCP Remote-Host-Regardless(Inexact) connection establish. Default is off.
@parm int8		| enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm
Forwarding Engine wll generate Normal TCP flow without consider remote host's IP/Port if <p enable>=TRUE.
For one NORMAL new TCP flow establishment, if <p enable>=TRUE, forwarding engine will establish this flow regardless of
it's remote host's IP/port. In other words, if this flow is established, system will allow other Remote IP/port to connect
to this entry via the external IP/port of this flow. Otherwise, forwarding engine will establish NORMAL new TCP flow with
the full information about internal/external/remote IP and port.
@xref  
*/
int32 rtl8651_fwdEngineInexactTcpFlow(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvInexactTcpNaptEnable = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}
#endif
/*
@func int32		| rtl8651_fwdEngineUdpStaticNaptPortTranslation | Enable/Disable UDP static NAPT port translation. Default is off.
@parm int8		| enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm
if <p enable>=TRUE, When processing normal UDP NAPT outbound flow, Forwarding Engine will NOT change
its source port number if this port is not be used now. But, if the source port is been used by other flow,
forwarding engine will assign a new one to this flow.
if <p enable>=FALSE, Forwarding engine will always assign external port for each normal UDP NAPT outbound flow.
@xref  
*/
int32 rtl8651_fwdEngineUdpStaticNaptPortTranslation(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvStaticUdpNaptPortTranslation = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}
/*
@func int32		| rtl8651_fwdEngineTcpStaticNaptPortTranslation | Enable/Disable TCP static NAPT port translation. Default is off.
@parm int8		| enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm
if <p enable>=TRUE, When processing normal TCP NAPT outbound flow, Forwarding Engine will NOT change
its source port number if this port is not be used now. But, if the source port is been used by other flow,
forwarding engine will assign a new one to this flow.
if <p enable>=FALSE, Forwarding engine will always assign external port for each normal TCP NAPT outbound flow.
@xref  
*/
int32 rtl8651_fwdEngineTcpStaticNaptPortTranslation(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvStaticTcpNaptPortTranslation = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}
#endif /* _RTL_NAPT_ADVANCED */

#ifdef _RTL_WEAK_NAPT_CTRL
/*
@func int32		| rtl8651_fwdEngineWeakTcpNaptProcess | Enable/Disable weak TCP NAPT flow creation. Default is off.
@parm int8		| enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm
if <p enable>=TRUE, When forwarding engine processes a outbound TCP non-SYN/no-entry packet,
it will create a new napt entry for this packet instead of it to protocol stack. For inbound TCP non-SYN/no-entry
packet, if dmz host/server port/upnp is assigned or trigger port is triggered, we will create a new napt entry
for this packet and forward it. if <p enable>=FALSE, Forwarding engine will always do SPI and trap illegal packets.
@xref  
*/
int32 rtl8651_fwdEngineWeakTcpNaptProcess(int8 enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvWeakTcpNaptEnable = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}
#endif	/* _RTL_WEAK_NAPT_CTRL */

/*
@func int32		| rtl8651_fwdEnginePsDontTimeout | Disable/Enable Idle-timer refresh for ALL Protocol Stack to-wan packets.
@parm int8		| enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm
if <p enable>=TRUE, all packets from protocol stack to WAN would NOT included in the idle-timer counting.
If <p enable>=FALSE. packets from protocol stack to wan would follow other decisions to decide if it can refresh idle-timer or not.
@xref  <p rtl8651_fwdEnginePsToWanErrMsgDontTimeout>
*/
int32 rtl8651_fwdEnginePsDontTimeout(int enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvProtocolStackDontTimeout = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEnginePsToWanErrMsgDontTimeout | Disable/Enable Idle-timer refresh for Protocol Stack to-wan-error message.
@parm int8		| enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm
if <p enable>=TRUE, the error messages from protocol stack to WAN would NOT included in the idle-timer counting.
If <p enable>=FALSE, the error messages would seem as normal packet and be included in idle-timer counting.
Note: Error message includes
	1. TCP RST packet
	2. ICMP error message
		1) ICMP_UNREACH
		2) ICMP_SOURCEQUENCH
		3) ICMP_TIMXCEED
		4) ICMP_PARAMPROB
@xref  <p rtl8651_fwdEnginePsDontTimeout>
*/
int32 rtl8651_fwdEnginePsToWanErrMsgDontTimeout(int enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvProtocolStackErrMsgDontTimeout = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}

#if	_RTL_DROP_ERROR_MESSAGE
/*
@func int32		| rtl8651_fwdEnginePsToWanErrMsgDrop | drop pkt for Protocol Stack to-wan-error message.
@parm int8		| enable | TRUE or FALSE. 
@rvalue SUCCESS	| 	Always returns SUCCESS.
@comm
if <p enable>=TRUE, the error messages from protocol stack to WAN would be silently droped.
If <p enable>=FALSE, the error messages would seem as normal packet and would NOT be silently droped.
Note: Error message includes
	1. TCP RST packet
	2. ICMP error message
		1) ICMP_UNREACH
		2) ICMP_SOURCEQUENCH
		3) ICMP_TIMXCEED
		4) ICMP_PARAMPROB
*/
int32 rtl8651_fwdEnginePsToWanErrMsgDrop(int enable)
{
	rtlglue_drvMutexLock();
	rtl8651_drvProtocolStackErrMsgDrop = enable;
	rtlglue_drvMutexUnlock();

	return SUCCESS;
}
#endif
/*
@func int32		| rtl8651_fwdEngineSend	| Send function for upper layer protocol stack.
@parm void * 		| pkthdr			| Pointer to sending packet's pkthdr.
@parm int16 		| dvid			| packet's destination VLAN id.
@parm int32 		| iphdrOffset		| IP header oddset. If this value is negative, egress ACL filter would bypass IP ACL rules.
@rvalue FWDENG_ACL_OUT_DROP	| 	Packet is dropped due to egress ACL.
@rvalue SUCCESS					| 	Packet passed egress ACL and sent.
@rvalue [non-SUCCESS]			|	Other failure.
@comm
All packets sent from protocol stack must call this function to send packet. Egress ACL would check
the packet and determine if it could be sent.
If <p iphdrOffset> value is negative, egress IP ACL ruls would be bypassed.
 */
/*
	Called by	:	protocol stack
	return	:	SUCCESS or drop reason
	note		:	pkt is always released in this function
 */
int32 rtl8651_fwdEngineSend(void * pkthdr, int16 dvid, int32 iphdrOffset)
{
	int32 retval = FAILED;

	rtlglue_drvMutexLock();		/* Lock resource */

	/* ROMEDRV actually send packet */
	retval = _rtl8651_fwdEngineProtocolStackSend(pkthdr,
												dvid,
												iphdrOffset,
												FALSE /* rtl8651_fwdEngineSend only use low priority TX queue */ );

	rtlglue_drvMutexUnlock();	/* Unlock resource */

	return retval;
}

/*
@func int32	| rtl8651_fwdEngineQosSend	| Send function for upper layer protocol stack with QoS capability.
@parm void *	| pkthdr					| Pointer to sending packet's pkthdr.
@parm int16	| dvid					| packet's destination VLAN id.
@parm int32	| iphdrOffset				| IP header oddset. If this value is negative, egress ACL filter would bypass IP ACL rules.
@parm int32	| qualityId				| Quality Identification, indicate the different quality types.
@rvalue FWDENG_ACL_OUT_DROP		| Packet is dropped due to egress ACL.
@rvalue SUCCESS						| Packet passed egress ACL and sent.
@rvalue [non-SUCCESS]				| Other failure.
@comm
This API do the same thing as <p rtl8651_fwdEngineSend> with QoS capability.
All packets sent from protocol stack must call this function to send packet. Egress ACL would check
the packet and determine if it could be sent.
If <p iphdrOffset> value is negative, egress IP ACL ruls would be bypassed.

QoS Issues:
		ROMEDRV support HW High/Low queue now, so:
			(qualityId > 0) means High queue is used.
			(qualityId == 0) means Low queue is used.		(same as <p rtl8651_fwdEngineSend>)
 */
/*
	Called by	:	protocol stack
	return	:	SUCCESS or drop reason
	note		:	pkt is always released in this function
 */
int32 rtl8651_fwdEngineQosSend(void *pkthdr, int16 dvid, int32 iphdrOffset, int32 qualityId)
{
	int32 retval = FAILED;

	rtlglue_drvMutexLock();		/* Lock resource */

	retval = _rtl8651_fwdEngineProtocolStackSend(pkthdr,
												dvid,
												iphdrOffset,
												qualityId?TRUE:FALSE);	/* ROMEDRV actually send packet */

	rtlglue_drvMutexUnlock();	/* Unlock resource */

	return retval;
}

void _calc_L34Checksum(struct rtl_pktHdr *pkthdrPtr, int32 IphdrOffset)
{
	struct rtl_mBuf *mbuf = pkthdrPtr->ph_mbuf;
	struct ip *iphdr = (struct ip *)&(mbuf->m_data[IphdrOffset]);
	uint16 org_l3cksum = 0, org_l4cksum = 0;

	if (pkthdrPtr->ph_pkt_property & PH_PKT_PROPER_IPIP_L3CSM) {
		org_l3cksum = iphdr->ip_sum;		// save original l3 checksum
		iphdr->ip_sum = 0;
		iphdr->ip_sum = _rtl8651_ipChecksum(iphdr);
	}
	if (pkthdrPtr->ph_pkt_property & PH_PKT_PROPER_IPIP_L4CSM) {
		uint16 ipcsm_tmp = iphdr->ip_sum;

		if (IS_FRAGMENT_REMAIN(iphdr->ip_off))
		{			
			goto l4cksum_done;
		}
		switch (iphdr->ip_p)
		{
			case IPPROTO_TCP:
				{
					struct tcphdr *tcp;
					tcp = (struct tcphdr *)L4HDR_ADDR(iphdr);

					org_l4cksum = tcp->th_sum;		// save original tcp ckecksum

					iphdr->ip_sum = 0;
					tcp->th_sum = 0;
					tcp->th_sum = _rtl8651_tcpChecksum(iphdr);
					iphdr->ip_sum = ipcsm_tmp;
				}
				break;
			case IPPROTO_UDP:
				{
					struct udphdr *udp;
					udp = (struct udphdr *)L4HDR_ADDR(iphdr);

					org_l4cksum = udp->uh_sum;		// save original udp checksum
					
					iphdr->ip_sum = 0;
					udp->uh_sum = 0;
					udp->uh_sum = _rtl8651_tcpChecksum(iphdr);
					iphdr->ip_sum = ipcsm_tmp;
				}
				break;
			case IPPROTO_ICMP:
				{
					struct icmp *ic;
					ic = (struct icmp *)L4HDR_ADDR(iphdr);

					org_l4cksum = ic->icmp_cksum;		// save original icmp ckecksum
					
					ic->icmp_cksum = 0;
					ic->icmp_cksum = _rtl8651_icmpChecksum(iphdr);
				}
				break;
			case IPPROTO_IGMP:
				{
					struct igmp *ig;
					ig = (struct igmp *)L4HDR_ADDR(iphdr);

					org_l4cksum = ig->igmp_cksum;		// save original igmp checksum
					
					ig->igmp_cksum = 0;
					ig->igmp_cksum = _rtl8651_icmpChecksum(iphdr);
				}
				break;
		}
	}
	
l4cksum_done:

	return;
}

/*
	Internal function:

		* Normal process for upper-layer procotol stack packet send.

		* It be called by external APIs
			(ex. rtl8651_fwdEngineSend for normal packet delivery and rtl8651_fwdEngineQosSend for
			 QoS-aware packet delivery).

		* Several things are done here:
			1. Rome driver private process to the sent packet.
			2. Rome driver Egress ACL check.
			3. Actual packet delivery.

		* Framework of packet delivery flow:

			Packet from UPPER LAYER--(don't need QoS)--> rtl8651_fwdEngineSend()------>------|
                                                           |---(need QoS)----> rtl8651_fwdEngineQosSend() --->------|
                                                                                                                                                         |
                                                                                                                                                         |
                                                     _rtl8651_fwdEngineProtocolStackSend() <----(call Internal API)----|
												|
												|
										(ROMEDRV process)
												|
												|
			Packet from ROMEDRV --------> _rtl8651_fwdEngineSend()

		* Parameters:
				pkthdr		- packet header structure of the packet to send
				dvid			- destination VLAN id
				iphdrOffset	- ip header offset of packet (for PPTP/L2TP packet, it need to set to outer IP header
				highPriority	- use HW high priority TX queue or not (decide packet header's "flags" set PKTHDR_HPRIORITY or not)

		* NOTE:
			No matter what return value this function return, packet is "processed" in this process
			(whether be delivered or be dropped).
*/
static int32 _rtl8651_fwdEngineProtocolStackSend(void * pkthdr, int16 dvid, int32 iphdrOffset, int32 highPriority)
{
	int32 retval=FAILED;
	rtl8651_tblDrv_vlanTable_t *local_vlanp;
	uint32 dsid=0, portmask=0,l2portmask=0; /*default dsid=0 if from protocol stack*/
#ifdef RTL8651_DSID_SUPPORT	
	uint32 sip=0;
	struct ip *iphdr=NULL;
#endif	
	struct rtl_mBuf *mbuf;
	struct rtl_pktHdr * pkthdrPtr=(struct rtl_pktHdr *)pkthdr;
	uint32 assignedPortMask = pkthdrPtr->ph_portlist;
	int32 realIphdrOffset;
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
	uint8 isToWanPptpL2tpOther = FALSE;	/* indicate this packet is PPTP/L2TP other packet (non IP data packet) within the its ip header and its sent to PPTP/L2TP interface.  */
#endif


	mbuf=pkthdrPtr->ph_mbuf;
	pkthdrPtr->ph_vlanIdx= dvid & (RTL8651_VLAN_NUMBER-1);
	local_vlanp = &DrvTbl.vlan[pkthdrPtr->ph_vlanIdx];
	pkthdrPtr->ph_rxdesc = -1;  /*pkt is from protocol stack.*/	
	realIphdrOffset = iphdrOffset;	/* initial value of real Iphdr Offset */

	/* <----- For PPTP/L2TP packet, we might need to check the inner-IP header for DATA Channel -----> */
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
	{
		rtl8651_tblDrv_routeTable_t *rt_t;

		/* get the default route and check its Layer2 wan type */
		rt_t = DrvTbl.route + (rtl8651_tblDrvPara.routingTableSize - 1);

		if (((rt_t->valid) && (rt_t->ifrt_flag == 0)) &&
			((rt_t->netif_t->linkLayerType == RTL8651_LL_PPTP) || (rt_t->netif_t->linkLayerType == RTL8651_LL_L2TP)))
		{	/* now we know the destination vlan of this packet uses PPTP/L2TP wan type */
			rtl8651_tblDrv_pppTable_t *pppEntry_t;
			pppEntry_t = rt_t->rt_nhSession_t;
			
			if (pppEntry_t && pppEntry_t->initiated && tunnel.valid && (local_vlanp->vid == tunnel.wanVid))
			{
				struct ip *outerIp;

				/* for DATA Channel, iphdrOffset is > 0 */
				if (realIphdrOffset <= 0)
				{
					goto TO_WAN_PPTPL2TP_OTHER;
				}

				outerIp = (struct ip *)((uint32)(mbuf->m_data) + realIphdrOffset);

				if ( IS_FRAGMENT_REMAIN(outerIp->ip_off) )
				{	/* outer IP payload has been fragmented */
					goto TO_WAN_PPTPL2TP_OTHER;
				}

				if (rt_t->netif_t->linkLayerType == RTL8651_LL_PPTP)
				{	/* PPTP */
					if (outerIp->ip_p == IPPROTO_GRE)
					{
						struct wan_gre_header *gre = (struct wan_gre_header*)L4HDR_ADDR(outerIp);
						uint32 l4Len = 8;
						uint8 *ptr = (uint8 *)gre;

						assert(ntohs(outerIp->ip_len) >= 28);

						if (gre->flags & 0x10)
							l4Len += 4;
						if (gre->ver & 0x80)
							l4Len += 4;

						if ((ptr[l4Len] == 0xff) && (ptr[l4Len + 1] == 0x03))
						{	/* RFC 1662 3.1, Address & Control Field */
							l4Len += 2;
						}

						if (	((ptr[l4Len] == 0x00) && (ptr[l4Len + 1] == 0x21) /* 16 bits case */) ||
							(ptr[l4Len] == 0x21 /* 8 bits case */))
						{
							struct ip *innerIp = NULL;

							if (ptr[l4Len] == 0x21)
							{	/* 8 bits case */
								l4Len += 1;
							} else
							{	/* 16 bits case */
								l4Len += 2;
							}

							innerIp = (struct ip*)((uint32)gre + l4Len);

							if ((uint32)ntohs(outerIp->ip_len) != (ntohs(innerIp->ip_len) + ((uint32)innerIp - (uint32)outerIp)))
							{
								goto TO_WAN_PPTPL2TP_OTHER;
							}

							/* <---------- Modify IP header offset here ----------> */
							realIphdrOffset = l4Len + ((uint32)gre - (uint32)(mbuf->m_data));

							/* Due to lack of checksum calculation in LSO, check ip-in-ip L3/L4 checksum here. */
							_calc_L34Checksum(pkthdrPtr, realIphdrOffset);
							goto IPHDROFFSET_MODIFY_DONE;
						}
					}
				} else if (rt_t->netif_t->linkLayerType == RTL8651_LL_L2TP)
				{	/* L2TP */
					if (outerIp->ip_p == IPPROTO_UDP)
					{
						struct udphdr *udp = (struct udphdr *)L4HDR_ADDR(outerIp);

						if (udp->uh_sport == htons(1701))
						{
							struct wan_l2tp_header *l2tp = (struct wan_l2tp_header*)(udp + 1);

							if (l2tp->flags_type == 0)	/* data packet, not control packet */
							{
								uint8 *u8p_l2tp = (uint8 *)l2tp;
								struct ip *innerIp = NULL;
								uint32 l2tpPppChkLen = 6 /* L2TP's least header length is 6 */; /* trace of the length of L2TP+PPP header */

								if (l2tp->flags_length)
								{
									l2tpPppChkLen += 2;
								}
								if (l2tp->flags_offset)
								{
									l2tpPppChkLen += 2;
								}

								if ((u8p_l2tp[l2tpPppChkLen] == 0xff) && (u8p_l2tp[l2tpPppChkLen + 1] == 0x03))
								{	/* RFC 1662 3.1, Address & Control Field */
									l2tpPppChkLen += 2;
								}

								if (u8p_l2tp[l2tpPppChkLen] == 0x21)
								{	/* 8 bits case */
									l2tpPppChkLen += 1;
								} else if ((u8p_l2tp[l2tpPppChkLen] == 0x00) && (u8p_l2tp[l2tpPppChkLen + 1] == 0x21))
								{	/* 16 bits case */
									l2tpPppChkLen += 2;
								}
								else
									goto TO_WAN_PPTPL2TP_OTHER;

								innerIp = (struct ip*)((uint32)u8p_l2tp + l2tpPppChkLen);

								if ((uint32)ntohs(outerIp->ip_len) != (ntohs(innerIp->ip_len) + ((uint32)innerIp-(uint32)outerIp)))
								{
									goto TO_WAN_PPTPL2TP_OTHER;
								}

								/* <---------- Modify IP header offset here ----------> */
								realIphdrOffset = ((uint32)innerIp - (uint32)mbuf->m_data);

								iphdr = (struct ip *)&(mbuf->m_data[realIphdrOffset]);

								/* ---------------------------------------------------------------------
									Due to lack of checksum calculation in LSO, check ip-in-ip L3/L4 checksum here. 
								    ---------------------------------------------------------------------- */
								if (pkthdrPtr->ph_pkt_property & PH_PKT_PROPER_IPIP_L3CSM) {
									uint16 org_l3cksum = iphdr->ip_sum;		/* save original l3 checksum */
									iphdr->ip_sum = 0;
									iphdr->ip_sum = _rtl8651_ipChecksum(iphdr);
									_RTL8651_DATA_ADJUST_CHECKSUM(iphdr->ip_sum, org_l3cksum, udp->uh_sum);
								}
								if (pkthdrPtr->ph_pkt_property & PH_PKT_PROPER_IPIP_L4CSM) {
									uint16 org_l4cksum;
									/* 
										Due to modification of ip checksum when calculating tcp/udp checksum, we have to save origial ip checksum first.
									    	After finishing calculation of l4 checksum, we have to restore original ip checksum.
									*/
									uint16 ipcsm_tmp = iphdr->ip_sum;

									if (IS_FRAGMENT_REMAIN(iphdr->ip_off))
										goto IPHDROFFSET_MODIFY_DONE;

									switch (iphdr->ip_p)
									{
										case IPPROTO_TCP:
											{
												struct tcphdr *inner_tcp;
												inner_tcp = (struct tcphdr *)L4HDR_ADDR(iphdr);
												org_l4cksum = inner_tcp->th_sum;		/* save original tcp ckecksum */
												iphdr->ip_sum = 0;
												inner_tcp->th_sum = 0;
												inner_tcp->th_sum = _rtl8651_tcpChecksum(iphdr);
												iphdr->ip_sum = ipcsm_tmp;
												/* adjust outer udp checksum */
												_RTL8651_DATA_ADJUST_CHECKSUM(inner_tcp->th_sum, org_l4cksum, udp->uh_sum);
											}
											break;
										case IPPROTO_UDP:
											{
												struct udphdr *inner_udp;
												inner_udp = (struct udphdr *)L4HDR_ADDR(iphdr);
												org_l4cksum = inner_udp->uh_sum;		/* save original udp checksum */
												iphdr->ip_sum = 0;
												inner_udp->uh_sum = 0;
												inner_udp->uh_sum = _rtl8651_tcpChecksum(iphdr);
												iphdr->ip_sum = ipcsm_tmp;
												/* adjust outer udp checksum */
												_RTL8651_DATA_ADJUST_CHECKSUM(inner_udp->uh_sum, org_l4cksum, udp->uh_sum);
											}
											break;
										case IPPROTO_ICMP:
											{
												struct icmp *inner_ic;
												inner_ic = (struct icmp *)L4HDR_ADDR(iphdr);
												org_l4cksum = inner_ic->icmp_cksum;		/* save original icmp ckecksum */
												inner_ic->icmp_cksum = 0;
												inner_ic->icmp_cksum = _rtl8651_icmpChecksum(iphdr);
												/* adjust outer udp checksum */
												_RTL8651_DATA_ADJUST_CHECKSUM(inner_ic->icmp_cksum, org_l4cksum, udp->uh_sum);
											}
											break;
										case IPPROTO_IGMP:
											{
												struct igmp *inner_ig;
												inner_ig = (struct igmp *)L4HDR_ADDR(iphdr);
												org_l4cksum = inner_ig->igmp_cksum;		/* save original igmp checksum */
												inner_ig->igmp_cksum = 0;
												inner_ig->igmp_cksum = _rtl8651_icmpChecksum(iphdr);
												/* adjust outer udp checksum */
												_RTL8651_DATA_ADJUST_CHECKSUM(inner_ig->igmp_cksum, org_l4cksum, udp->uh_sum);
											}
											break;
									}
								}
								
								goto IPHDROFFSET_MODIFY_DONE;
							}
						}
					}
				}
TO_WAN_PPTPL2TP_OTHER:
				/*
					If process arrives here, it imply 3 things:
						1. The WAN type of current configuration is activated PPTP or L2TP.
						2. This packet would be sent to PPTP/L2TP VLAN.
							- because (local_vlanp->vid == tunnel.wanVid)
						3. This packet is NOT PPTP/L2TP data packet.
							- because the IP data packets have been parsed and jumped to
							   IPHDROFFSET_MODIFY_DONE

					==> So we set the variable 'isPptpL2tpOther' to TRUE
				*/
				isToWanPptpL2tpOther = TRUE;
			}
		}
	}
IPHDROFFSET_MODIFY_DONE:
#endif
	/* <-------------------- End of realIphdrOffset modification --------------------> */

	/* Assign ph_pppoeIdx if PPPoE pkts. This ph_pppoeIdx is for ASIC to use when AutoAdd PPPoE header. */
	if (mbuf->m_data[12] == 0x88 && (mbuf->m_data[13] == 0x63 || mbuf->m_data[13] == 0x64)) {
		rtl8651_tblDrv_pppTable_t * pppoePtr;
		uint16 sessionId = (mbuf->m_data[16] << 8) + mbuf->m_data[17];

		pppoePtr = _rtl8651_getPppoeSessionByPppoeSessionId(sessionId);
		pkthdrPtr->ph_pppoeIdx = PPPOE_TABLE_INDEX(pppoePtr);
	}

	/* <--------- driver level packet checking ---------> */
	if(realIphdrOffset>=0){

		/* packet parsing and variables initiation*/
		iphdr=(struct ip *)&mbuf->m_data[realIphdrOffset];
#ifdef RTL8651_DSID_SUPPORT
		sip = PKTGET_UINT32_UNALIGNED(&iphdr->ip_src.s_addr);
		dsid= _rtl8651_getpppDSID(sip);
#endif

		/* Packet process for IP packets */

		/*
		  *	1. aliasProtoStackSentPkt
		  *		-- Process PPTP/L2TP/PPTP idle timeout
		  *		-- Process Port Modification for Remote Management
		  *
		  *	2. dos get Scan Complain
		  *		-- get the complain for DoS Scan attack
		  *
		  *	3. _rtl8651_resolveProxyInboundFlow
		  *		-- for NAPT redirect process, we recover the Source IP/Port of the packet sent back to NPI
		  *
		  *	4. protocol stack flow cache
		  *		-- cache the flow of ProtocolStack <--> Remote Host
		  *
		  *
		  *
		  *	===========================================================
		  */

#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		/*
		  *	Note:
		  *		@ For PPTP/L2TP non-IP packets to WAN, driver don't need to process the item 1/2/3
		  *
		  *			1. For ProtoStackSentPkt, Port modification is only for inner-IP packet, and only the
		  *			    Data packet (we only process IP Data packet yet) need to update idleTimer
		  *			2. for DoS Scan, we only process ICMP/TCP complain within PPTP/L2TP header if
		  *			    this packet sent to PPTP/L2TP interface. It implies that we only process inner-IP
		  *			    packets.
		  *			3. for ProxyInboundFlow, we only modify source ip/source port of inner-IP header
		  */
		if (isToWanPptpL2tpOther == FALSE)
#endif
		{

			/* ===============================================================
			  *	1. 	Remote control issues
			  *
			  *		pkt from protocol stack, see if we need to modify it
			  *
			  * ---------------------------------------------------------------------------------
			  *	This condition:
			  *		For PPPoE 0x8863 case: don't let it run ProtoStackSent		--> 0x8863's iphdrOffset must be -1 and would not arrive here natively
			  *		For PPPoE 0x8864 case: only allow packet with NCP==0x0021	--> We need to process this condition
			  *
			  *		==> So we only allow packet runing ProtoStackSentPkt when ...
			  *				1. {m_data[12], m_data[13]} != {0x8864} 	or
			  *				2. {m_data[12], m_data[13]} == {0x8864} && {m_data[20], m_data[21]} == {0x0021}
			  * ============================================================== */
#ifdef _RTL_PS_ACTION
			if((mbuf->m_data[12]!=0x88)||(mbuf->m_data[13]!=0x64)||((mbuf->m_data[20]==0x00) && (mbuf->m_data[21]==0x21)))
			{
				retval = _rtl8651_aliasProtoStackSentPkt( !local_vlanp->internal,dsid, iphdr );
			}
#endif
			/* ===============================================================
			  *	2. 	DoS complain collection
			  *
			  *		check the complain (ex. ICMP error message, TCP RST....) to find the SCAN attack out.
			  * ============================================================== */
#ifdef _RTL_DOS_PLUGIN
			/* check for TCP/UDP Scan */
			_rtl8651_dosScanCheckComplain(dsid, iphdr);	/* complain */
#endif

			/* ===============================================================
			  *	3. 	NAPT redirect recovery
			  *
			  *		recover source ip/port for redirected packet  from Gateway to NPI
			  * =============================================================== */
#ifdef _RTL_NAPT_REDIRECT
			/* for napt-redirect: recover source ip/port for redirected packet  from Gateway to NPI */
			_rtl8651_resolveProxyInboundFlow(local_vlanp, iphdr);
#endif
		}


			/* ===============================================================
			  *	4. 	Protocol Stack Flow Cache creation
			  *
			  *		Create the flow cache to notify our Driver that the flow is used by Protocol stack
			  *		and all packets belong to this flow need to be trapped to Protocol Stack.
			  * =============================================================== */

#ifdef _RTL_PS_FLOWCACHE
		if ((!local_vlanp->internal) && (!IS_FRAGMENT_REMAIN(iphdr->ip_off)))
		{
			_rtl8651_fwdEnginePsFlowCacheCreate(iphdr);
		}
#endif /* _RTL_PS_FLOWCACHE */
	}
	if (assignedPortMask==0)
	{
		l2portmask=rtl8651_asicL2DAlookup(mbuf->m_data);
		if (l2portmask == 0)	/* lookup miss: unknown dmac (might Mcast mac or unknown unicast mac), do broadcast */
		{
			l2portmask = local_vlanp->memberPortMask;
		}
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		else if (l2portmask == (1<<CONFIG_RTL865XB_PPTPL2TP_PORT))
		{
		    	l2portmask = (1<<tunnel.wanPort); //if this packet's mac in loopback vlan, force set it to wan.
		}
#endif
	} else
	{
		l2portmask = assignedPortMask;
	}
	portmask = local_vlanp->memberPortMask & l2portmask;

	if (!portmask)
	{
		retval = FWDENG_NOTXPORT;
		mBuf_driverFreeMbufChain(mbuf);
		goto out;
	}	
	
	pkthdrPtr->ph_portlist = portmask;
	pkthdrPtr->ph_dsid = dsid; 
	pkthdrPtr->ph_flags |= (0x8800 | ((highPriority == TRUE)?PKTHDR_HPRIORITY:0));	/* use high priority TX queue */

#if	_RTL_DROP_ERROR_MESSAGE
	if ( PS_SEND_ERROR_PKT==retval )
	{
		retval = SUCCESS;
		goto FreeAndSilentDropOut;
	}
#endif	

	/* in _rtl8651_fwdEngineSend, we MUST use the outer-IP-header for PPTP/L2TP process */
	retval= _rtl8651_fwdEngineSend(FWDENG_FROMPS, pkthdr, dvid, iphdrOffset);

	// free packets which need drop
	if (retval != SUCCESS)
	//if (retval & FWDENG_DROP)
	{		
#if	_RTL_DROP_ERROR_MESSAGE	
FreeAndSilentDropOut:
#endif		
		mBuf_driverFreeMbufChain(mbuf);
	}
out:
	return retval;
}


static void _rtl8651_fwdEnginePsFlowCacheCreate(struct ip *iphdr)
{
	/********************************************************************
		chenyl (04/09/01)
			
		For gateway's packets sent to External vlan, we must sniff them and
		record the cache. It's to prevent miscarriage with DMZ Host is assigned.

		We only cache TCP/UDP packets

			1. ICMP responding packets (ex. ICMP TIMESTAMP_REPLY, ECHO_REPLY, ICMP Error messages)
				to gateway will be trapped because of corresponding entry is not found.
				They will not be forwarded to DMZ Host.

			2.	Known Side-effect:
				Other L4 packets only be forwarded because of Protocol-based NAT.
				When it's set, gateway will not receive packets with that L4 protocol.
				We don't consider this case now.
	********************************************************************/
	uint8 protocol;

	protocol = iphdr->ip_p;

	if ((protocol == IPPROTO_TCP) || (protocol == IPPROTO_UDP))
	{
		struct tcphdr *tc;
		struct udphdr *ud;
		ipaddr_t sip;
		ipaddr_t dip;
		uint16 sport;
		uint16 dport;
		struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry;
		struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *tbl;

		tbl = &entry;

		sip = PKTGET_UINT32_UNALIGNED(&iphdr->ip_src);
		dip = PKTGET_UINT32_UNALIGNED(&iphdr->ip_dst);

		if (protocol == IPPROTO_TCP)
		{	/* tcp */
			tc = (struct tcphdr *)L4HDR_ADDR(iphdr);

			sport = ntohs(tc->th_sport);
			dport = ntohs(tc->th_dport);

			if (_rtl8651_getOutboundNaptFlow(TRUE, sip, sport, dip, dport, &tbl) != SUCCESS)
			{	/* entry not found, create a new one */
				int8 flowType;
				int32 retval;
				int16 retryCnt = 10;
				uint16 newSport = sport;

				if (tc->th_flags == TH_SYN)
				{
					flowType = TCP_OUTBOUND_SYN_RCVD_FLOW;
				}else
				{
					/*
						TH_SYN|TH_ACK:	This flow is initiated by corresponding host.
										It means flow can translate to established flow after sending this packet.
						other flags:		It can set to established flow, and modified in l4TcpMonitorOut()
					*/
					flowType = TCP_ESTABLISHED_FLOW;
				}

				tbl = NULL;

				/* check External IP/Port Reuse */
				while ( _rtl8651_getInboundNaptFlow(TRUE, sip, newSport, dip, dport, &tbl) == SUCCESS )
				{
					retryCnt --;
					newSport += 4096;	/* magic number : gotten from the AIR (multiple of 1024, index in napt table will not be changed) */
					tbl = NULL;
					if (retryCnt == 0) break;
				}

				/* add Napt Connection if we can */
				if ( (retryCnt > 0) &&
					((retval = _rtl8651_addNaptConnection(TRUE, (TRUE|PURE_SW|DONT_CHECK_INTIP|TO_PROTOCOL_STACK), flowType, sip, sport, &sip, &newSport, dip, dport)) == SUCCESS) )
				{
					tbl = &entry;
					if ((retval = _rtl8651_getInboundNaptFlow(TRUE, sip, newSport, dip, dport, &tbl)) != SUCCESS)
					{
#ifdef RTL865X_DEBUG
						rtlglue_printf("[FATAL] %s(%d) Add NAPT Success but get Inbound flow FAILED (ret:%d)!\n", __FUNCTION__, __LINE__, retval);
						rtlglue_printf("\tTCP %u.%u.%u.%u(%d) --- %u.%u.%u.%u(%d)\n", NIPQUAD(sip), newSport, NIPQUAD(dip), dport);
#endif	/* RTL865X_DEBUG */
						tbl = NULL;
					}
				}
				/* always send packet */
			}
			if (tbl)
			{
				if ( tbl->insideLocalPort != tbl->insideGlobalPort)
				{	/* modify source port if we need */
					assert(sport == tbl->insideLocalPort);
					tc->th_sport = htons(tbl->insideGlobalPort);
					_RTL8651_L4_ADJUST_CHECKSUM(0, 0, tbl->insideGlobalPort, sport, tc->th_sum);
					sport = tbl->insideGlobalPort;
				}
				assert(tbl->toProtocolStack == TRUE);
				_rtl8651_l4TcpMonitorOut(tc, tbl);
			}
		}else
		{	/* udp */
			ud = (struct udphdr *)L4HDR_ADDR(iphdr);

			sport = ntohs(ud->uh_sport);
			dport = ntohs(ud->uh_dport);

			if (_rtl8651_getOutboundNaptFlow(FALSE, sip, sport, dip, dport, &tbl) != SUCCESS)
			{	/* entry not found, create a new one */
				int32 retval;
				int16 retryCnt = 10;
				uint16 newSport = sport;

				tbl = NULL;

				/* check External IP/Port Reuse */
				while ( _rtl8651_getInboundNaptFlow(FALSE, sip, newSport, dip, dport, &tbl) == SUCCESS )
				{
					retryCnt --;
					newSport += 4096;	/* magic number : gotten from the AIR (multiple of 1024, index in napt table will not be changed) */
					tbl = NULL;
					if (retryCnt == 0) break;
				}

				/* add Napt Connection if we can */
				if ( (retryCnt > 0) && ((retval = _rtl8651_addNaptConnection(TRUE, (TRUE|PURE_SW|DONT_CHECK_INTIP|TO_PROTOCOL_STACK), UDP_FLOW,
					sip, sport, &sip, &newSport, dip, dport)) == SUCCESS) )
				{
					tbl = &entry;
					if ((retval = _rtl8651_getInboundNaptFlow(FALSE, sip, newSport, dip, dport, &tbl)) != SUCCESS)
					{
#ifdef RTL865X_DEBUG
						rtlglue_printf("[FATAL] %s(%d) Add NAPT Success but get Inbound flow FAILED (ret:%d)!\n", __FUNCTION__, __LINE__, retval);
						rtlglue_printf("\tUDP %u.%u.%u.%u(%d) --- %u.%u.%u.%u(%d)\n", NIPQUAD(sip), newSport, NIPQUAD(dip), dport);
#endif	/* RTL865X_DEBUG */
						tbl = NULL;
					}
				}

				/* always send packet */
			}
			if (tbl)
			{
				if ( tbl->insideLocalPort != tbl->insideGlobalPort)
				{	/* modify source port if we need */
					assert(sport == tbl->insideLocalPort);
					ud->uh_sport = htons(tbl->insideGlobalPort);
					if (ud->uh_sum);
						_RTL8651_L4_ADJUST_CHECKSUM(0, 0, tbl->insideGlobalPort, sport, ud->uh_sum);
					sport = tbl->insideGlobalPort;
				}
				assert(tbl->toProtocolStack == TRUE);
				_rtl8651_updateNaptConnection(tbl, UDP_FLOW, rtl8651GlobalControl.udpTimeout);
			}
		}
	}

	return;
}

int32 rtl8651_offloadEngineL3Send(void * pPkt, uint32 L3PktLength, uint32 l34CkumOffload, int32 highPriority)
{
	/*
		This function let ROME DRV offload the protocol stack traffic.

		Constraint:
			1. In packet content, the HEADROOM size (14+128=142) is needed.
			2. For fragmented packet, L4 checksum can NOT be offloaded, and this API will
			    NOT return fail when user use this wrong setting.
			3. Packet always be processed by offload engine (forward or drop),
			    but offload engine would return real fail result to user.
			4. UDP checksum field must not be 0 if you want to utilize L4_CSUM function of ASIC.

		Process:
			1. Fill the corresponding packet header information.
			2. Fill the packet as normal packet: DA - SA - ETHERTYPE - IPHDR.
			3. Forwarding engine would offload this packet as normal L3 forwarding process.
				- Decide its destination by routing table lookup.
				- Fill the l2 related header.
				- Add corresponding header (ex. PPPoE, VLAN, PPTP, L2TP).
				- Forward packet.
	*/
	struct rtl_pktHdr *pkthdr = (struct rtl_pktHdr*)pPkt;
	struct rtl_mBuf *mBuf = NULL;
	struct ip *iphdr = NULL;
	int32 retval = FAILED;

	/* ==============================================
		Check the correctness of packet and init variables
	     ============================================== */

	if (pkthdr==NULL||pkthdr->ph_mbuf==NULL)
	{
		return OFFENG_EINVALIDINPUT;
	}

	mBuf = pkthdr->ph_mbuf;
	#define L3OFFLOAD_MIN_HEADROOM	(14 + rtl8651_fwdEnginePara.mbufHeadroom)
	if (((uint32)(mBuf->m_data) - (uint32)(mBuf->m_extbuf)) < L3OFFLOAD_MIN_HEADROOM)
	{
		return OFFENG_EHEADROOM2SMALL;
	}

	iphdr = (struct ip*)(mBuf->m_data);

	/* ==============================================
		Forge the corresponding L2 information
	     ============================================== */
	mBuf->m_data -= 14;
	memset(mBuf->m_data, 0, 2 * sizeof(ether_addr_t));	/* init SA/DA, it would be filled in forwarding engine */
	*(uint16*)&(mBuf->m_data[12]) = htons(0x0800);		/* fill ethertype as IPv4 (0x0800) */

	/*
		if L3/4 checksum can NOT be offloaded by ASIC, we must calculate it here

		If User don't want to offload L3/L4 checksum,
		but we can process it in ASIC, then Offload Engine would still process L3/4 checksum offload
		because it has no any overhead.

	*/
	if (l34CkumOffload)
	{
		if (!IPHDRLEN_IS_TWENTY(iphdr))
		{	/* we must calculate L3 (L4) checksum here */
			if (l34CkumOffload & OFFLOAD_L4_CKSUM)
			{
				if (!IS_FRAGMENT(iphdr->ip_off))
				{
					switch (iphdr->ip_p)
					{
						case IPPROTO_TCP:
						{
							struct tcphdr *tc = (struct tcphdr*)L4HDR_ADDR(iphdr);
							tc->th_sum = 0;
							tc->th_sum = _rtl8651_tcpChecksum(iphdr);
						}
							break;
						case IPPROTO_UDP:
						{
							struct udphdr *ud = (struct udphdr*)L4HDR_ADDR(iphdr);
							ud->uh_sum = 0;
							ud->uh_sum = _rtl8651_tcpChecksum(iphdr);
						}
							break;
						case IPPROTO_ICMP:
						{
							struct icmp *ic = (struct icmp*)L4HDR_ADDR(iphdr);
							ic->icmp_cksum = 0;
							ic->icmp_cksum = _rtl8651_icmpChecksum(iphdr);
						}
							break;
						case IPPROTO_IGMP:
						{
							struct igmp *ig = (struct igmp*)L4HDR_ADDR(iphdr);
							ig->igmp_cksum = 0;
							ig->igmp_cksum = _rtl8651_icmpChecksum(iphdr);
						}
							break;
					}
				}
			}
			if (l34CkumOffload & OFFLOAD_IP_CKSUM)
			{
				iphdr->ip_sum = 0;
				iphdr->ip_sum = _rtl8651_ipChecksum(iphdr);
			}
		}
	}

	/* ==============================================
		Forge the corresponding packet header
	     ============================================== */
	pkthdr->ph_sip = PKTGET_UINT32_UNALIGNED(&iphdr->ip_src);
	pkthdr->ph_dip = PKTGET_UINT32_UNALIGNED(&iphdr->ip_dst);
	switch (iphdr->ip_p)
	{
		case IPPROTO_TCP:	pkthdr->ph_proto = PKTHDR_TCP;	break;
		case IPPROTO_UDP:	pkthdr->ph_proto = PKTHDR_UDP;	break;
		case IPPROTO_ICMP:	pkthdr->ph_proto = PKTHDR_ICMP;	break;
		case IPPROTO_IGMP:	pkthdr->ph_proto = PKTHDR_IGMP;	break;
		case IPPROTO_GRE:	pkthdr->ph_proto = PKTHDR_PPTP;	break;
		default:			pkthdr->ph_proto = PKTHDR_IP;	break;
	}

	pkthdr->ph_dataLen = 0;		/* useless */
	pkthdr->ph_fragOffset = 0;		/* useless */
	pkthdr->ph_reason = 2;		/* forge to : 'pre-ACL filtering trap' */
	pkthdr->ph_sport = 0;			/* useless */

	pkthdr->ph_flags &= ~(PKT_BCAST|PKT_MCAST|PKTHDR_BRIDGING|PKTHDR_HWLOOKUP|PKTHDR_HPRIORITY|PKTHDR_PPPOE_AUTOADD|PKTHDR_VLAN_AUTOADD|CSUM_IP|CSUM_L4);
	pkthdr->ph_flags |= ((highPriority == TRUE)?PKTHDR_HPRIORITY:0);	/* set high-priority bit */
	pkthdr->ph_portlist = RTL8651_PSOFFLOAD_RESV_PORT;
	pkthdr->ph_extPortList = 0;		/* forge as packet received from physical port, must be 0 */
	pkthdr->ph_srcExtPortNum = 0;	/* forge as packet received from physical port, must be 0 */
	pkthdr->ph_vlanIdx = 0;		/* useless */
	pkthdr->ph_pppoeIdx = 0;		/* useless */
	pkthdr->ph_unnumber = 0;		/* useless */
	pkthdr->ph_vlanTagged = 0;	/* forge as simple packet format: must be 0 */
	pkthdr->ph_LLCTagged = 0;		/* forge as simple packet format: must be 0 */

	pkthdr->ph_routeIdx = 0;
	pkthdr->ph_dsid = _rtl8651_getpppDSID(pkthdr->ph_sip);
	pkthdr->ph_iphdrOffset = 14;
	pkthdr->ph_len = mBuf->m_len = 14 + L3PktLength;

	/* ==============================================
		Try to match policy route
	     ============================================== */
	if (pkthdr->ph_dsid != _rtl8651_DefaultDialSessionId)
	{
		rtl8651_tblDrv_routeTable_t *rt_t = NULL;
		rtl8651_tblDrv_vlanTable_t *vlan_t;

		/*
			when we apply non-default session route,
			we MUST check if destination IP is Remote Public or not.

			NOTE:
					We DONT check if destination IP is gateway's IP here.
		*/
		rt_t = _rtl8651_getRoutingEntry(pkthdr->ph_dip);

		if (rt_t && rt_t->valid)
		{
			vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(rt_t->dvid)];
			if (vlan_t->internal == FALSE)
			{
				rtl8651_tblDrv_pppTable_t *ppp_tb = NULL;

				ppp_tb = _rtl8651_getPppoeSessionByDsid((uint32)(pkthdr->ph_dsid));
				if (ppp_tb)
				{
					uint32 rtIdx;

					rt_t = DrvTbl.route;
					for(rtIdx = 0; rtIdx < rtl8651_tblDrvPara.routingTableSize; rtIdx++,rt_t++)
					{
						if (	(rt_t->netif_t->linkLayerType == RTL8651_LL_PPPOE) &&
							(rt_t->rt_nhSession_t == ppp_tb))
						{
							/* check if the destination vlan of this routing entry is external */
							vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(rt_t->dvid)];
							if (vlan_t->internal == FALSE)
							{
								pkthdr->ph_routeIdx = (rt_t - &DrvTbl.route[0]) + 1;
								goto policyRoute_done;
							}
						}
					}
				}

			}
		}
	}
policyRoute_done:
	/* ==============================================
		Actually do L3 forwarding
	     ============================================== */

	rtlglue_drvMutexLock();
	retval = _rtl8651_l3FastActRouting(FWDENG_OFFLOAD, pkthdr, pkthdr->ph_dip, iphdr, NULL);
	rtlglue_drvMutexUnlock();

	if (retval != SUCCESS)
	{
		/* we always free mbuf in offload engine and report user with real reason */
		mBuf_freeMbufChain(mBuf);
	}

	return retval;

}

static int32 _rtl8651_fwdEnginePSOffload_L3ExtraProcess(struct rtl_pktHdr *pPkt, struct ip *iphdr, rtl8651_tblDrv_vlanTable_t *vlan_t)
{
	uint8 dsid = pPkt->ph_dsid;

#if	_RTL_DROP_ERROR_MESSAGE
	int32 retval = SUCCESS;
#endif	


	/* ===============================================================
	  *	1. 	Remote control issues
	  *
	  *		pkt from protocol stack, see if we need to modify it
	  *
	  * ---------------------------------------------------------------------------------
	  *	This condition:
	  *		For PPPoE 0x8863 case: don't let it run ProtoStackSent		--> 0x8863's iphdrOffset must be -1 and would not arrive here natively
	  *		For PPPoE 0x8864 case: only allow packet with NCP==0x0021	--> We need to process this condition
	  *
	  *		==> So we only allow packet runing ProtoStackSentPkt when ...
	  *				1. {m_data[12], m_data[13]} != {0x8864} 	or
	  *				2. {m_data[12], m_data[13]} == {0x8864} && {m_data[20], m_data[21]} == {0x0021}
	  *
	  *
	  *		==> NOTE: Here (_rtl8651_fwdEnginePSOffload_L3ExtraProcess) we can can sure L3 is IP
	  *
	  * ============================================================== */
#if	_RTL_DROP_ERROR_MESSAGE		  
	retval = _rtl8651_aliasProtoStackSentPkt(!vlan_t->internal, dsid, iphdr);
#else
	_rtl8651_aliasProtoStackSentPkt(!vlan_t->internal, dsid, iphdr);
#endif

	/* ===============================================================
	  *	2. 	DoS complain collection
	  *
	  *		check the complain (ex. ICMP error message, TCP RST....) to find the SCAN attack out.
	  * ============================================================== */
#ifdef _RTL_DOS_PLUGIN
	/* check for TCP/UDP Scan */
	_rtl8651_dosScanCheckComplain(dsid, iphdr);	/* complain */
#endif

	/* ===============================================================
	  *	3. 	NAPT redirect recovery
	  *
	  *		recover source ip/port for redirected packet  from Gateway to NPI
	  * =============================================================== */
#ifdef _RTL_NAPT_REDIRECT
	/* for napt-redirect: recover source ip/port for redirected packet  from Gateway to NPI */
	_rtl8651_resolveProxyInboundFlow(vlan_t, iphdr);
#endif

	/* ===============================================================
	  *	4. 	Protocol Stack Flow Cache creation
	  *
	  *		Create the flow cache to notify our Driver that the flow is used by Protocol stack
	  *		and all packets belong to this flow need to be trapped to Protocol Stack.
	  * =============================================================== */

#ifdef _RTL_PS_FLOWCACHE
	if ((!vlan_t->internal) && (!IS_FRAGMENT_REMAIN(iphdr->ip_off)))
	{
		_rtl8651_fwdEnginePsFlowCacheCreate(iphdr);
	}
#endif /* _RTL_PS_FLOWCACHE */

#if	_RTL_DROP_ERROR_MESSAGE
	return retval;
#else
	return SUCCESS;
#endif

}


static int32 _getEthTypeIpOffset(struct rtl_pktHdr *ph, uint8 *m_data, uint32 *ethtypeOffset, uint32 *iphdrOffset, uint32 *ncp){
	uint32 ip_offset = *iphdrOffset;
	uint32 ethTypePos= *ethtypeOffset;
	if( ph->ph_vlanTagged){		
		ip_offset+=4;
		ethTypePos+=4;
		/* check vlan CFI bit */
		if(m_data[ethTypePos-2] & 0x10 )
			return FWDENG_VLANCFI;
	}
	if(ph->ph_LLCTagged){
		ip_offset+=6;
		ethTypePos+=8;
	}
	if(ph->ph_pppeTagged){
		*ncp= ntohs(*((uint16 *)(m_data+ip_offset+6)));
		ip_offset+=8;
	}
	*ethtypeOffset=ethTypePos;
	*iphdrOffset=ip_offset;
	return SUCCESS;
}

/*
@func int32		| rtl8651_fwdEngineInput	| Forwarding engine input function called by NIC driver. 
@parm void * | pkthdr | Input packet's pkthdr pointer.
@rvalue FWDENG_NOTIPPACKET  |	Destined to gateway, but not IP packet.
@rvalue FWDENG_VLANINVALID	|	Invalid VLAN. Maybe not yet configured.
@rvalue FWDENG_VLANCFI		|	Vlan tagged pkt with CFI=1
@rvalue FWDENG_BROADCAST	|	L2 broadcast, send to protocol stack
@rvalue FWDENG_ACL_IN_DROP	|	Pkt dropped by ingress ACL
@rvalue FWDENG_ACL_OUT_DROP	|	Pkt dropped by egress ACL
@rvalue FWDENG_PPPOE8863	|	PPPoE discovery phase pkt
@rvalue FWDENG_PPPOE8864	|	PPPoE session phase pkt
@rvalue FWDENG_L34DISABLED	|	Pkt not processed due to fwd engine's layer3/4 is disabled. 
@rvalue FWDENG_L3UNKNOWNPROTO	|	unknown L3 protocol
@rvalue FWDENG_IP_CANTFRAG		|	Can't/don't process this IP fragment packet.
@rvalue FWDENG_IP_CANTQUEUEFRAGMENT	|	Out of fragment queue space
@rvalue FWDENG_IP_BADCKSUM		|	IP cksum error
@rvalue FWDENG_IP_TTLEXCEED		|	TTL exceed
@rvalue FWDENG_IP_INVALID		|	invalid  source IP or dest IP in IP header
@rvalue FWDENG_IP_MULTICAST		|	IP multicast pkt.(not IGMP)
@rvalue FWDENG_ARP_NOTREQ		|	Not a ARP request
@rvalue FWDENG_ARP_OTHERS		|	ARP request to other host
@rvalue FWDENG_ARP_REQUEST		|	ARP request to us, not yet replied.
@rvalue FWDENG_IGMPPKT			|	IGMP pkt not processed.
@rvalue FWDENG_NETIFACENOTFOUND	|	No layer3 network interface attached on receiving VLAN
@rvalue FWDENG_L4_UNKNOWNPROTO	|	unsupported L4 protocol
@rvalue FWDENG_L4_NAPT_UNSUPPORTPROTO	|	must do NAPT but packet is not TCP/UDP/ICMP
@rvalue FWDENG_L4_INVALIDPORT	|	Src or dst Port number is 0.
@rvalue FWDENG_L4_NATNOENTRY	|	NAT entry not found
@rvalue FWDENG_ICMP_OUT_ADDFAIL		|	out of ICMP entry space for this outbound ICMP echo request packet.
@rvalue FWDENG_ICMP_OUT_INVALIDIPID	|	out of ICMP ID when doing NAPT for this outbound ICMP echo request
@rvalue FWDENG_ICMP_OUT_ERRFLOWNOTFOUND	| error flow is not found for this outbound ICMP error mesg
@rvalue FWDENG_ICMP_IN_NOTFOUND  		|	icmp entry not found for this inbound ICMP packet.
@rvalue FWDENG_ICMP_IN_ERRFLOWNOTFOUND	|	icmp entry not found in inbound ICMP error mesg
@rvalue FWDENG_ICMP_IN_ADDFAIL	|	error occurred when system added entry for this inbound icmp packet
@rvalue FWDENG_ICMP_BADCKSUM	|	bad icmp checksum
@rvalue FWDENG_UDP_BADCKSUM	|	bad udp cksum
@rvalue FWDENG_UDP_ECHOCHARGENBOMB	|	suspicous UDP echo/chargen bomb pkt.
@rvalue FWDENG_UDP_NAPT_OUT_NOTFOUND	|	NAPT entry not found for this outbound UDP packet 
@rvalue FWDENG_UDP_NAPT_OUT_NOPORT	|	All usable port numbers exhausted for this outbound NAPT UDP packet.
@rvalue FWDENG_UDP_NAPT_OUT_ADDFAIL	|	Failed to add new UDP napt entry. Table full?
@rvalue FWDENG_UDP_NAPT_OUT_SVRALGADDFAIL	|	Failed to add new UDP server alg napt entry. Table full?
@rvalue FWDENG_UDP_NAPT_OUT_CLIALGADDFAIL	|	Failed to add new UDP client alg napt entry. Table full?
@rvalue FWDENG_UDP_NAPT_OUT_IGNORE			|	Ignored outbound UDP packet when doing NAPT.
@rvalue FWDENG_UDP_NAPT_IN_NOTFOUND		|	Inbound NAPT entry not found for this UDP packet.
@rvalue FWDENG_UDP_NAPT_IN_ALGADDFAIL		|	new ALG entry add fail for this inbound UDP packet.
@rvalue FWDENG_UDP_NAPT_IN_ADDFAIL			|	new NAPT entry add fail for this inbound UDP packet.
@rvalue FWDENG_UDP_NAPT_IN_SERVERPORTNOTFOUND |  server port entry not found for this UDP packet.
@rvalue FWDENG_UDP_NAPT_IN_IGNORE			|	Ignored inbound UDP packet when doing NAPT
@rvalue FWDENG_UDP_NAT_IN_CLIALGADDFAIL 	|	client ALG entry add failed during NAT for this inbound UDP packet.
@rvalue FWDENG_UDP_NAT_IN_SVRALGADDFAIL 	|	server ALG entry add failed during NAT for this inbound UDP packet
@rvalue FWDENG_UDP_NAT_IN_ALGNOTFOUND		|	ALG entry not found during NAT for this inbound UDP packet.
@rvalue FWDENG_UDP_NAT_IN_ALGIGNORE			|	Ignore inbound UDP ALG trapped packet when doing NAT 
@rvalue FWDENG_UDP_NAT_OUT_ALGIGNORE		|	Ignore outbound UDP ALG trapped packet when doing NAT 
@rvalue FWDENG_UDP_NAT_OUT_ALGNOTFOUND	|	ALG entry not found during NAT for this outbound UDP packet.
@rvalue FWDENG_UDP_NAT_OUT_CLIALGADDFAIL	|	client ALG entry add failed during NAT for this outbound UDP packet.
@rvalue FWDENG_UDP_NAT_OUT_SVRALGADDFAIL	|	server ALG entry add failed during NAT for this outbound UDP packet
@rvalue FWDENG_TCP_BADCKSUM	|	Bad TCP checksum
@rvalue FWDENG_TCP_URLFILTERED	|	URL filtered packet.
@rvalue FWDENG_TCP_NAPT_OUT_CLIALGADDFAIL 	|	adding a existed ALG control flow or ALG list full
@rvalue FWDENG_TCP_NAPT_OUT_SVRALGADDFAIL 	|	adding a existed ALG control flow or ALG list full
@rvalue FWDENG_TCP_NAPT_OUT_ALGNOTFOUND	|	ALG entry not found during NAPT for this outbound TCP packet.
@rvalue FWDENG_TCP_NAPT_OUT_NOIPPORT		|	No NAPT port number available for this outbound TCP packet
@rvalue FWDENG_TCP_NAPT_OUT_NOTFOUND		|	Outbound NAPT entry not found for this TCP packet.
@rvalue FWDENG_TCP_NAPT_OUT_ADDFAIL		|	Failed to add new NAPT entry for this outbound TCP packet. Table full?
@rvalue FWDENG_TCP_NAPT_OUT_IGNORE			|	Ignored outbound TCP packet when doing NAPT.
@rvalue FWDENG_TCP_NAPT_IN_CLIALGADDFAIL	|	client ALG entry add failed during NAT for this outbound TCP packet.
@rvalue FWDENG_TCP_NAPT_IN_IGNORE			|	inbound TCP packet ignored when doing NAPT.
@rvalue FWDENG_TCP_NAPT_IN_SVRALGADDFAIL	|	server ALG entry add failed during NAPT for this outbound TCP packet
@rvalue FWDENG_TCP_NAPT_IN_NOTFOUND		|	NAPT entry not found for this inbound TCP packet.
@rvalue FWDENG_TCP_NAPT_IN_ADDFAIL			|	new NAPT entry add fail for this inbound TCP packet.
@rvalue FWDENG_TCP_NAT_IN_ALGNOTFOUND		|	ALG entry not found during NAPT for this inbound TCP packet.
@rvalue FWDENG_TCP_NAT_IN_ALGADDFAIL		|	new ALG entry add fail when doing NAT for this inbound TCP packet.
@rvalue FWDENG_TCP_NAT_IN_ALGIGNORE			|	Ignore inbound TCP ALG trapped packet when doing NAT 
@rvalue FWDENG_TCP_NAT_OUT_ALGNOTFOUND	|	ALG entry not found during NAT for this outbound TCP packet.
@rvalue FWDENG_TCP_NAT_OUT_ALGIGNORE		|	Ignore inbound TCP ALG trapped packet when doing NAT 
@rvalue FWDENG_TCP_NAT_OUT_ADDFAIL			|	Failed to add new NAT entry for this outbound TCP packet. Table full?
@rvalue FWDENG_USER_IDSCHECKDROP			|	User defined DoS & IDS check
@rvalue FWDENG_TO_GW						|	Packet is indicated to trapped to CPU
@rvalue FWDENG_QUEUED_TO_GW				|	Queued Packets must be trapped to CPU
@rvalue FWDENG_L4_PORTBOUNCING				|	If Portbouncing process is disabled, forwarding engine trap all port bouncing packets with this return value
@rvalue SUCCESS	| 	packet processed by forwarding engine.
@rvalue FAILED	| 	Unknown error in forwarding engine.
@comm 
Forwarding engine input function called by NIC driver. 
 */
int32 rtl8651_fwdEngineInput(void * pkthdr)
{
	struct rtl_pktHdr * pkthdrPtr = (struct rtl_pktHdr *)pkthdr;
	struct rtl_mBuf *mbufPtr;
	ipaddr_t sip,dip;
	uint16 vid, ethType = 0;
	uint32 ethTypePos;
	uint8 *m_data;
	int32 local_val;
	uint8 ip_protocol;
	struct ip *ip=NULL;
	int32 ip_offset;
	uint32 ncp = 0;
	uint16 srcPortlist = 0;
	rtl8651_tblDrv_vlanTable_t *local_vlanp;
	int32 inAcl_done = FALSE;
	uint32 inAcl_index = 0;
	uint8 *l4hdr;
	uint8 direction; /* 0:LAN_to_WAN 1:WAN_to_LAN */
	struct tcphdr *tc;
	struct udphdr *ud;
	uint8 isMulticast;

	/* Decision table */
	uint32 srcType = _RTL8651_HOST_NONE;
	uint32 dstType = _RTL8651_HOST_NONE;
	uint32 decision = _RTL8651_DEC_NONE;
	uint32 optLayer = _RTL8651_OPERATION_OTHER;

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_RECVLOOP);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_BEFORE_CRYPTO_ENCAP);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_BEFORE_CRYPTO_DECAP);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_UNTIL_ACLDB);
#endif

	rtlglue_drvMutexLock();	/* Lock resource */

	local_vlanp = &DrvTbl.vlan[pkthdrPtr->ph_vlanIdx];	/* Get ASIC source vlan index */
	
	/* Initialize partial variables */
	mbufPtr = pkthdrPtr->ph_mbuf;
	m_data = mbufPtr->m_data;
	srcPortlist = pkthdrPtr->ph_portlist;

	assert((((uint32)m_data)&1)==0);	/* make sure pkt is 2 byte aligned. */
	ip_offset = 14;
	ethTypePos = 12;
	ncp=0;
	
	/*If pkt has PPPoE/VLAN/LLC*/
	if(( *((uint16 *)(((uint8 *)pkthdrPtr)+8))&0x1c00)==0)
	{
		pkthdrPtr->ph_iphdrOffset = 14;	/* most cases.... */
	} else
	{
		/* Yes...look closer... */
		local_val=_getEthTypeIpOffset(pkthdrPtr,m_data, &ethTypePos, &ip_offset, &ncp);
		if(local_val!=SUCCESS)
		{
			goto out;
		}
		pkthdrPtr->ph_iphdrOffset = ip_offset;
	}
	ethType = ntohs(*((uint16*)(m_data+ethTypePos)));
	ip = (struct ip *)(pkthdrPtr->ph_mbuf->m_data + ip_offset);	

	/*****************************************
				Reason Bit process
	******************************************/
	if ( ( pkthdrPtr->ph_reason & 0x1c ) == 0x04 )	/* ACL Match */
	{
		uint8 l4proto;
		l4proto = ip->ip_p;

		inAcl_done = FALSE;
		inAcl_index = ((pkthdrPtr->ph_reason >> 5) & 0x7f);

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_UNTIL_ACLDB);
#endif

#ifdef RTL865XB_ACLDB
		if ( _rtl8651_jumptoACLDB(inAcl_index, pkthdrPtr, ip) == SUCCESS )
		{
			switch (l4proto)
			{
				case IPPROTO_TCP:
					rtl8651_fwdEngineCounter.tcpReceive ++;
					rtl8651_fwdEngineCounter.tcpSend ++;
					break;
				case IPPROTO_UDP:
					rtl8651_fwdEngineCounter.udpReceive ++;
					rtl8651_fwdEngineCounter.udpSend ++;
					break;
				case IPPROTO_ICMP:
					rtl8651_fwdEngineCounter.icmpReceive ++;
					rtl8651_fwdEngineCounter.icmpSend ++;
					break;
			}
			local_val = SUCCESS;

			goto out;
		}
#endif /* RTL865XB_ACLDB */
	}else if ( pkthdrPtr->ph_reason & 0x18 )	/* ACL Passed */
	{
		rtl8651_fwdEngineCounter.inAclPermit ++;	/* update ingress ACL counter */
		inAcl_done = TRUE;
		inAcl_index = 0;
	} else /* pre-ACL trap */
	{
		inAcl_done = FALSE;
		inAcl_index = 0;

#ifdef RTL865XB_DCUT_SWVLAN
		/*
			(pre-ACL trap && vlan-tagged && vid not in VLAN TBL) == unknown vid pkt
			 => Try software vlan mechanism to handle this kind of packet.
		*/
		if (pkthdrPtr->ph_vlanTagged)
		{
			rtl8651_tblDrv_vlanTable_t *vlanPtr;
			uint16 vid = ntohs(*((uint16 *)(m_data + 14))) & 0x0fff;

			/* ASIC vlan table should not have this vid. */
			vlanPtr = &DrvTbl.vlan[rtl8651_vlanTableIndex(vid)];
			assert(vlanPtr->valid == 0 || vlanPtr->vid != vid);			
			/* Make this pointer NULL (it means that this packet is vlan-tagged with unknown vid to ASIC) */
			local_vlanp = NULL;
		}
#endif

	}

	/* 
		For RTL865xB b-cut, rtl8651_rxPktPreprocess will strip vlan tag if the pkt is vlan-tagged. 
		For RTL865xB c / d-cut, rtl8651_rxPktPreprocess will not strip vlan tag if the pkt is vlan-tagged.
		We strip vlan tag here for RTL865xB c / d-cut if the vid of the pkt is known to ASIC vlan.
	*/
	if (local_vlanp != NULL && pkthdrPtr->ph_vlanTagged)
	{
		_rtl8651_rxPktPreprocess_stripVlan(pkthdrPtr);
	}

	/* =========================================================================
	 *	Layer 2 processing.
	 *
	 *	Relay pure ethernet Broadcast/multicast to all other ports 
	 *	of this VLAN only if ACL passed.
	 *	  =FAILED: pkt is destined to gateway. Pass to L3/4 code
	 *	  =SUCCESS: pkt is L2 unicast and forwarded, or is a strange pkt and be dropped.
	 *	  All others: send to protocol stack
	    ========================================================================= */
	
	local_val=rtl8651_fwdEngineL2Input(pkthdrPtr,local_vlanp, m_data,ethType);

	if (local_val == SUCCESS)
	{
		goto out;
	}
	if (local_val != FAILED)
	{

#ifdef RTL865XB_DCUT_SWVLAN
		if (local_val == FWDENG_VLANINVALID)
			goto out;
#endif

		if (local_val==FWDENG_IP_MULTICAST)
		{	/* must set "isMulticast" for remaining process for multicast packet */
			isMulticast = TRUE;
		}else
		{
			/* L2 or IP broadcast packet, L2 control packet: trap to protocol stack */
			
			if(local_val==FWDENG_PPPOE8863)
				rtl8651_fwdEngineCounter.pppoeDiscoveryReceive++;
			goto out;
		}
	}

	vid = local_vlanp->vid;	
	pkthdrPtr->ph_routeIdx = 0;
#ifdef RTL865X_TEST
	pkthdrPtr->ph_sip=PKTGET_UINT32_UNALIGNED(&ip->ip_src);
	pkthdrPtr->ph_dip=PKTGET_UINT32_UNALIGNED(&ip->ip_dst);
#else
	/* unbelievable, using this way can save 200 CPU cycles. */
	pkthdrPtr->ph_sip=(ntohs(*((uint16*)&ip->ip_src))<<16)+ntohs(*(((uint16*)&ip->ip_src)+1));
	pkthdrPtr->ph_dip=(ntohs(*((uint16*)&ip->ip_dst))<<16)+ntohs(*(((uint16*)&ip->ip_dst)+1));
#endif

	/* system don't apply fragment system on the input packet yet */
	_rtl8651_fragQueueProcessOk = FALSE;

	/*==========================================
	  * TCP/UDP Fast Path is always the highest priority
	  *==========================================*/
	if ((rtl8651_drvRoutingEnable == TRUE) && (ip->ip_ttl > 1))
	{
		if (((ethType == 0x0800) || (ncp == 0x0021)))
		{
			uint8 l4proto;
			l4proto = ip->ip_p;
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
			if (l4proto == IPPROTO_UDP)
			{
				if (_rtl8651_fragmentShortestPathFwd(pkthdrPtr, ip) == SUCCESS)
				{
					rtl8651_fwdEngineCounter.udpReceive ++;
					rtl8651_fwdEngineCounter.udpSend ++;
					local_val = SUCCESS;
					goto out;
				}
			}else if (l4proto == IPPROTO_TCP)
			{
				if (_rtl8651_fragmentShortestPathFwd(pkthdrPtr, ip) == SUCCESS)
				{
					rtl8651_fwdEngineCounter.tcpReceive ++;
					rtl8651_fwdEngineCounter.tcpSend ++;
					local_val = SUCCESS;
					goto out;
				}
			}
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
			if (_rtl8651_dos_enable == TRUE)
			{
				/*
				 *		Source IP Blocking
				 *		Drop cases : must drop packet before do L2
				 *	cfliu: If we know what IPs to drop, why not use HW ACL but use software???
				 *    Ans:  HW ACL might be full, and the remaining blocked source are handled by software.
				 */
				//sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);
				if (_rtl8651_dosProc_blockSip_check(vid, pkthdrPtr->ph_sip) == SUCCESS)
				{
					local_val = FWDENG_DROP;
					goto out;
				}
			}
		}
		/* TCP/UDP Fast path */
#ifdef _RTL_NAPT_SHORT_CUT
		if ((_rtl8651_dos_enable == FALSE) && (_rtl8651_udpBlockingEnable == FALSE))
		{

#ifdef CONFIG_RTL865X_ROMEPERF
			rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_FASTPATH);
#endif

			rtl8651_drvPacketFastSend = TRUE;
			if (local_vlanp->internal) {
				if (ip->ip_p == IPPROTO_UDP)
				{
					local_val=_rtl8651_l4UdpFastAliasOut(pkthdrPtr, ip);
					assert(local_val != FWDENG_TO_GW);
					if (local_val == SUCCESS)
					{
						rtl8651_fwdEngineCounter.udpReceive ++;
						rtl8651_fwdEngineCounter.udpSend++;

#ifdef CONFIG_RTL865X_ROMEPERF
						rtl8651_romeperfExitPoint(ROMEPERF_INDEX_FASTPATH);
#endif
						goto out;
					}else if ((local_val > 0) && (local_val & FWDENG_DROP))
					{
						rtl8651_fwdEngineCounter.udpReceive ++;
						goto out;
					}
				} else if (ip->ip_p == IPPROTO_TCP)
				{
					local_val=_rtl8651_l4TcpFastAliasOut(pkthdrPtr,ip);
					assert(local_val != FWDENG_TO_GW);
					if (local_val == SUCCESS)
					{
						rtl8651_fwdEngineCounter.tcpReceive ++;
						rtl8651_fwdEngineCounter.tcpSend++;

#ifdef CONFIG_RTL865X_ROMEPERF
						rtl8651_romeperfExitPoint(ROMEPERF_INDEX_FASTPATH);
#endif
						goto out;
					}else if ((local_val > 0) && (local_val & FWDENG_DROP))
					{
						rtl8651_fwdEngineCounter.tcpReceive ++;
						goto out;
					}
				}
			}
			else	{
				if (ip->ip_p == IPPROTO_UDP)
				{
					local_val=_rtl8651_l4UdpFastAliasIn(pkthdrPtr, ip);
					assert(local_val != FWDENG_TO_GW);
					if (local_val == SUCCESS)
					{
						rtl8651_fwdEngineCounter.udpReceive ++;
						rtl8651_fwdEngineCounter.udpSend++;

#ifdef CONFIG_RTL865X_ROMEPERF
						rtl8651_romeperfExitPoint(ROMEPERF_INDEX_FASTPATH);
#endif
						goto out;
					}else if ((local_val > 0) && (local_val&FWDENG_DROP))
					{
						rtl8651_fwdEngineCounter.udpReceive ++;
						goto out;
					}
				} else if (ip->ip_p == IPPROTO_TCP)
				{
					local_val = _rtl8651_l4TcpFastAliasIn(pkthdrPtr,ip);
					assert(local_val != FWDENG_TO_GW);
					if (local_val == SUCCESS)
					{
						rtl8651_fwdEngineCounter.tcpReceive ++;
						rtl8651_fwdEngineCounter.tcpSend++;

#ifdef CONFIG_RTL865X_ROMEPERF
						rtl8651_romeperfExitPoint(ROMEPERF_INDEX_FASTPATH);
#endif
						goto out;
					}else if ((local_val > 0) && (local_val&FWDENG_DROP))
					{
						rtl8651_fwdEngineCounter.tcpReceive ++;
						goto out;
					}
				}
			}			
			rtl8651_drvPacketFastSend = FALSE;
		}
#endif /* _RTL_NAPT_SHORT_CUT */
	}

	/**************************************************/
	/*<--- Locate  IP header and ether type field position. --->*/
	/**************************************************/

	/* Local Variables Initialization */
	tc = NULL;
	ud = NULL;
	ip_protocol = 0;
	local_val = FAILED;
	l4hdr = NULL;
	direction = 0;
	isMulticast = FALSE;


/*
myself:
*/
	/* =========================================================================
	  * otherwise, pkt is for gateway itself to do L3/4 operation.. 
	  * both ethernet unicast(ACL not checked) and local copy of bcast pkts(ACL checked) fall through...
	  * The reason why I delayed ACL chk for unicast pkt is  I want to save per pkt ACL check time if pkt can go through
	  * fast path. Unicast pkt would not check ACL until it is to be L2 bridged or falls into slow path.
	    ========================================================================= */

	if(pkthdrPtr->ph_proto==PKTHDR_ETHERNET)
	{
		/* ingress ACL check */
		if(!inAcl_done) {
			local_val = _rtl8651_ingressAcl(pkthdrPtr->ph_dsid,inAcl_index,local_vlanp,vid,m_data,NULL, RTL8651_ACL_GENERIC, _RTL8651_OPERATION_OTHER);
			if(((local_val > 0) && (local_val&FWDENG_DROP))||(local_val==FWDENG_ACL_IN_USERTRAP))
			{
				goto out;
			}
			inAcl_done=TRUE;		
		}

		/* ingress ACL passed, start parse etherType */
		if(ethType == 0x0806) {
			/* unicast ARP packets: chenyl: It might be done in L2 Input if an function pointer registration process is done */
			local_val = _rtl8651_drvArpAgent(pkthdrPtr, mbufPtr, vid);
			goto out;
		}else if(ethType == 0x8863) {
			/* <----- PPPoE PADO/PADR/PADS -----> */
			if(rtl8651_drvPppoeNotExistedSessionTerminationEnable == TRUE) {
				pppoeHdr_t * pppoeHdrPtr;
				pppoeHdrPtr = (pppoeHdr_t *)&(mbufPtr->m_data[ethTypePos+2]);
				if(pppoeHdrPtr->code == 0x65) {/* PPPoE Active Discovery Session-confirmation (PADS), session ID become valid */
					rtlglue_printf("Active Session Detection receive PADS to update session ID %x\n", ntohs(pppoeHdrPtr->sessionId));
					_rtl8651_updatePppoeActiveSession(ntohs(pppoeHdrPtr->sessionId));
				} else if(pppoeHdrPtr->code == 0xa7) {/*PPPoE Active Discovery Terminate (PADT), session ID no longer valid */
					rtlglue_printf("Active Session Detection receive PADT to terminate session ID %x\n", ntohs(pppoeHdrPtr->sessionId));
					_rtl8651_terminatePppoeActiveSession(ntohs(pppoeHdrPtr->sessionId));
				}
			}
			rtl8651_fwdEngineCounter.pppoeDiscoveryReceive++;

			if (rtl8651_PPPoEEnable){
				/* use realtek's own pppoe/ppp implementation */
				assert( rtl8651_pppoeInput);
				if(rtl8651_pppoeInput(mbufPtr)) {
					local_val = FAILED;
				}else{
					rtl8651_fwdEngineCounter.pppoeDiscoveryProcessed++;
					local_val = SUCCESS;
				}
			}else
				local_val=FWDENG_PPPOE8863;

			/* ghhuang: diffserv remarking */

			
			goto out;
		}
	}

	if(pkthdrPtr->ph_pppeTagged){
		/* Input session checking */
		pppoeHdr_t * pppoeHdrPtr;
		rtl8651_tblDrv_pppTable_t * pppoePtr;

		pppoeHdrPtr = (pppoeHdr_t *)&(mbufPtr->m_data[ethTypePos+2]);		
		pppoePtr = _rtl8651_getPppoeSessionByPppoeSessionId(ntohs(pppoeHdrPtr->sessionId));
		if(pppoePtr == NULL) {
			if(rtl8651_drvPppoeNotExistedSessionTerminationEnable == TRUE && 
				_rtl8651_pppoeSessionActive(ntohs(pppoeHdrPtr->sessionId)) == FALSE) {//Input session packet without session previous exists. 
				uint16 i;
				uint8 charValue;

				/* PPPoE session not existe termination terminate corresponding session */
				rtlglue_printf("SID %x should be invalid send PADT\n", ntohs(pppoeHdrPtr->sessionId));
				
				for(i=0; i<6; i++) {/* Swap DMAC and SMAC */
					charValue = mbufPtr->m_data[i];
					mbufPtr->m_data[i] = mbufPtr->m_data[i+6];
					mbufPtr->m_data[i+6] = charValue;
				}
				mbufPtr->m_data[ethTypePos] = 0x88;
				mbufPtr->m_data[ethTypePos+1] = 0x63;
				pppoeHdrPtr->code = 0xa7; /* Change the code to PADT */
				pppoeHdrPtr->length = 0;
				pkthdrPtr->ph_flags &= ~(PKT_BCAST | PKT_MCAST | PKTHDR_BRIDGING | PKTHDR_HWLOOKUP	| PKTHDR_HPRIORITY |
										PKTHDR_PPPOE_AUTOADD | PKTHDR_VLAN_AUTOADD | CSUM_IP_OK | CSUM_L4_OK | CSUM_IP | CSUM_L4);
				pkthdrPtr->ph_pppeTagged = 0;
				pkthdrPtr->ph_proto = PKTHDR_ETHERNET;
				pkthdrPtr->ph_srcExtPortNum = 0;
				pkthdrPtr->ph_extPortList = 0;
				pkthdrPtr->ph_portlist = 1<<pkthdrPtr->ph_portlist;
				pkthdrPtr->ph_len = mbufPtr->m_len = 60;
				if(_rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdrPtr, vid,-1)!=SUCCESS) {
					mBuf_driverFreeMbufChain(pkthdrPtr->ph_mbuf);
				}
				local_val = SUCCESS;
				goto out;
			}
		}
		else {
			pppoePtr->rxBytes += pkthdrPtr->ph_len;
			pppoePtr->rxPackets ++;
		}
		assert(ethType==0x8864);
		/*** PPPoE session phase, all pkts are unicast ***/
			rtl8651_fwdEngineCounter.pppoeSessionReceive++;
		if(ncp!=0x0021){ /* not IPCP... */
			if (rtl8651_PPPoEEnable){
				/* use realtek's own pppoe/ppp implementation to handle cases such as LCP echo req/reply */
				assert( rtl8651_pppoeInput);
				if(rtl8651_pppoeInput(mbufPtr)){
					local_val=FAILED;
				}else{
					rtl8651_fwdEngineCounter.pppoeSessionProcessed++;
					local_val=SUCCESS;
				}
			}else
				local_val=FWDENG_PPPOE8864; /* Leave for upper layer stack */
			goto out;
		}
		else {/* PPPoE IP packet : normal case, continue processing */
		}
		rtl8651_fwdEngineCounter.pppoeSessionProcessed++;
	}else if(ethType!=0x0800){
		local_val=FWDENG_NOTIPPACKET;
		goto out;
	}

	/* ================================================
	  * IP protocol based packet processing
	  * ================================================ */

	if(rtl8651_drvRoutingEnable == FALSE){
		local_val=FWDENG_L34DISABLED;
		goto out;
	}
	/*
		Edward:
			Asking designer about IP checksum issue,
			jblin say the checksum calculation will be undefined
			when IP header length larger than 20-byte.
		chenyl:
			After test, the checksum OK bits (L3/L4) is set when IP header length larger than 20-byte.
	*/
	if(ISCLEARED(pkthdrPtr->ph_flags, CSUM_IP_OK)) {
		rtl8651_fwdEngineCounter.ipErrorReceive++;
		local_val = FWDENG_IP_BADCKSUM;
		goto out;	/* IP packet with layer3checksum error */
	}

	ip_protocol = ip->ip_p;
	l4hdr = (uint8 *) L4HDR_ADDR(ip);

	sip = pkthdrPtr->ph_sip;
	dip = pkthdrPtr->ph_dip;

	/* ASIC cannot verify the L4 protocol if ip header length is not 20, so we must check it by ourself */
	if (!IPHDRLEN_IS_TWENTY(ip))
	{
		switch(ip->ip_p)
		{
			case IPPROTO_TCP:
				pkthdrPtr->ph_proto = PKTHDR_TCP;
				break;
			case IPPROTO_UDP:
				pkthdrPtr->ph_proto = PKTHDR_UDP;
				break;
			case IPPROTO_ICMP:
				pkthdrPtr->ph_proto = PKTHDR_ICMP;
				break;
			case IPPROTO_IGMP:
				pkthdrPtr->ph_proto = PKTHDR_IGMP;
				break;
			/* default : don't modify original Pkt type */
		}
	}
	/*
		Filter out invalid IPs such as 0.0.0.0, 127.X.X.X, 255.255.255 as sip or dip
		also filter multicast IP as sip
		sip=0.0.0.0 is ok for DHCP and IGMP.
	*/
	if(!dip||((dip&0xff000000)==0x7f000000)){
		local_val=FWDENG_IP_INVALID;
		goto out;
	}
	if(sip&&(sip==0xffffffff|| (sip&0xff000000)==0x7f000000||(sip & 0xf0000000) == 0xe0000000)){
		local_val=FWDENG_IP_INVALID;
		goto out;
	}

	if (IN_MULTICAST(dip))
		isMulticast = TRUE;

	/* Process IGMP */
	if (pkthdrPtr->ph_proto == PKTHDR_IGMP)
	{
		struct igmp *ig = (struct igmp*)l4hdr;
		rtl8651_fwdEngineCounter.igmpReceive++;

		/* if forwarding engine's unicast IGMP forwarding is turned on : we ignore unicast IGMP process and continue its forwarding */
		if (!(rtl8651_drvFwdUnicastIGMPEnable && (IN_UNICAST(dip))))
		{
			local_val = FWDENG_IGMPPKT;
			#ifdef CONFIG_RTL865X_MULTICAST
			if (rtl8651_drvMulticastProcessEnable && (!IS_FRAGMENT(ip->ip_off)))
			{
				local_val = _rtl8651_igmp_process(pkthdr,  ip, ig);
			}
			#endif /* CONFIG_RTL865X_MULTICAST */
			goto out;
		}
	}

	/* IGMP's TTL always=1, multicast packet with TTL=1 must be forwarded through forwarding engine */
	if((ip->ip_ttl < 2) && (!isMulticast))
	{
		local_val=FWDENG_IP_TTLEXCEED;
		if (rtl8651_drvIcmpRoutingMsgEnable)
		{
			uint32 netIdx;
			if ((_rtl8651_extIpAddr(dip, &netIdx) == FALSE)
				&& (_rtl8651_intIpAddr(dip) == FALSE))
			{
				/* error code 0 : TTL equals 0 during transit */
				_rtl8651_IcmpPktField_t icmp_info;
				icmp_info.ic_type = ICMP_TIMXCEED;	/* Time exceed */
				icmp_info.ic_code = 0;	/* TTL equals 0 during transit */
				icmp_info.ic_mtu = 0;
				local_val = _rtl8651_drvIcmpErrorGeneration(pkthdrPtr, ip, &icmp_info, vid);
			}
		}
		goto out;
	}

	/* make sure packet direction  0:LAN_to_WAN 1:WAN_to_LAN */
	{
		if(!local_vlanp->internal)
			direction=1;
	}

#ifdef RTL8651_DSID_SUPPORT

	pkthdrPtr->ph_unnumber=0;
	if (!direction)
	{	/* try to match "LAN to WAN" */
		switch(pkthdrPtr->ph_proto)
		{
			case PKTHDR_TCP:
				tc=(struct tcphdr *)l4hdr;				
				pkthdrPtr->ph_dsid=_rtl8651_getPolicyRoutingDSID(pkthdrPtr,dip,sip,IPPROTO_TCP,ntohs(tc->th_dport));
				break;
			case PKTHDR_UDP:
				ud=(struct udphdr *)l4hdr;
				pkthdrPtr->ph_dsid=_rtl8651_getPolicyRoutingDSID(pkthdrPtr,dip,sip,IPPROTO_UDP,ntohs(ud->uh_dport));
				break;
			default:
				pkthdrPtr->ph_dsid=_rtl8651_getPolicyRoutingDSID(pkthdrPtr,dip,sip,pkthdrPtr->ph_proto,0);
				break;
		}
	} else
	{	/* from WAN */
		_rtl8651_getPolicyRoutingDSID(pkthdrPtr,0,dip,pkthdrPtr->ph_proto,0);
		pkthdrPtr->ph_dsid= _rtl8651_getpppDSID(dip);
	}

	pkthdrPtr->ph_unnumber|=direction<<1;
#endif

	/*
		process/reassemble fragment packet
	*/

#ifdef CONFIG_RTL865X_IPSEC
	/********* IPSEC *************************/
	{
		ipsecERoute_t *pIpsec;
		
		/* local to remote */
		if ( _rtl8651_findIpsecERouteBySrcDstIp( pkthdrPtr->ph_sip, pkthdrPtr->ph_dip, &pIpsec ) == SUCCESS )
		{
			local_val = _rtl8651_encapIpsec( -1, pkthdrPtr, ip, pIpsec );
			if(local_val == SUCCESS)
				goto out;
		}
		
		/* remote to local */
		if ( ip->ip_p==IPPROTO_ESP &&
		     _rtl8651_findIpsecERouteBySrcEdstIp( pkthdrPtr->ph_sip, pkthdrPtr->ph_dip, &pIpsec ) == SUCCESS )
		{
			local_val = _rtl8651_decapIpsec( -1, pkthdrPtr, ip, pIpsec );
			if(local_val == SUCCESS)
				goto out;
		}
	}
#endif

#ifdef _RTL_QUEUE_FRAG_PKT
	if (IS_FRAGMENT(ip->ip_off))
	{
		if (rtl8651_drvFragmentProcessEnable)
		{
			if (!_rtl8651_fragQueueProcess)	/* Dequeue from system Queue System */
			{
				pkthdrPtr->ph_dataLen = pkthdrPtr->ph_len - ip_offset;
				/* put this packet into queue system and determine if we need dequeue */
				local_val = _rtl8651_ip_defrag(pkthdrPtr, ip);
				if (local_val != FAILED)
				{
					/*
						if local_val == FAILED, it means this packet is the remaining packet of one positive
						fragment flow, forwarding engine must process it
					*/
					goto out;
				}
			}
		}else
		{	/* we don't process fragment packet */
			local_val = FWDENG_IP_CANTFRAG;
			goto out;
		}
		/* fragment system process is done */
		_rtl8651_fragQueueProcessOk = TRUE;

	}
#endif /* _RTL_QUEUE_FRAG_PKT */

	/* Multicast Fast Path */
	if (isMulticast == TRUE)
	{
#ifdef CONFIG_RTL865X_MULTICAST
		if (rtl8651_drvMulticastProcessEnable)
		{
			local_val = _rtl8651_multicast_Fastfwd(pkthdrPtr, ip, local_vlanp);
			if ((local_val == SUCCESS) || (local_val == FWDENG_IP_MULTICAST))
			{	/* process packet successful or mirror packet to protocol stack */
				goto out;
			}
		}else
#endif /* CONFIG_RTL865X_MULTICAST */
		{
			/*	If Multicast is disabled, Layer2 broadcast is
				done in Layer2 process, here forwarding engine only trap this packet to protocol	*/
			local_val = FWDENG_IP_MULTICAST;
			goto out;
		}
	}

	/* Forwarding engine process TCP first since most pkts are TCP. */
	if(pkthdrPtr->ph_proto == PKTHDR_TCP)
	{
		rtl8651_fwdEngineCounter.tcpReceive++;
		tc=(struct tcphdr *)l4hdr;
		pkthdrPtr->ph_sport = ntohs(tc->th_sport);

		/*	chenyl:
			After test, ASIC will set "CSUM_L4_OK" if this packet has no L4 header (fragment packet).
			So, we can check this flag no matter this packet is fragment or not.					*/
		if(ISCLEARED(pkthdrPtr->ph_flags, (CSUM_L4_OK)))
		{
			rtl8651_fwdEngineCounter.tcpErrorReceive++;
			local_val=FWDENG_TCP_BADCKSUM;
			goto out;	/* TCP packet with layer4 checksum error */
		}

		/* Layer 3 check */
		if(!sip || dip==0xffffffff)
		{
			local_val = FWDENG_IP_INVALID;
			goto out;
		}

		#ifdef _RTL_DOS_PLUGIN

		/* TCP DOS */
		if (rtl8651a_dosTcpCheck(pkthdrPtr->ph_dsid,vid,ip,&local_val,direction,
			rtl8651_wanStatus[pkthdrPtr->ph_dsid], pkthdrPtr, RTL8651_DOS_FAST_CHECK) == SUCCESS)
		{
			goto dos_out;
		}
		#endif /* _RTL_DOS_PLUGIN */

		#ifdef _RTL_NAPT_SHORT_CUT

		/* TCP Fast Path */
		rtl8651_drvPacketFastSend = TRUE;	/* Enter fast mode */
		{
			if(local_vlanp->internal)	/* check outbound packet. */
				local_val = _rtl8651_l4TcpFastAliasOut(pkthdrPtr,ip);
			else {		/* check inbound packet. */
				local_val = _rtl8651_l4TcpFastAliasIn(pkthdrPtr,ip);
			}
		}
		rtl8651_drvPacketFastSend = FALSE;	/* Exit fast mode */

		if (local_val == SUCCESS)
		{	/* forwarded in fast path. */
			rtl8651_fwdEngineCounter.tcpSend++;
			goto out;
		}

		if (local_val == FWDENG_TO_GW)
		{	/* with checking napt entry, system detects this packet must be trapped to protocol stack */
			goto out;
		}

		if (((local_val > 0) && (local_val & FWDENG_DROP)) || (local_val == FWDENG_TCP_URLFILTERED))
		{
			rtl8651_fwdEngineCounter.natdrop++;
			local_val |= FWDENG_DROP;
			goto out;
		}
		/* fast path FAILED: goto slow path */

		#endif 	/* _RTL_NAPT_SHORT_CUT */

		#ifdef _RTL_DOS_PLUGIN
		/* TCP DOS */
		if (rtl8651a_dosTcpCheck(pkthdrPtr->ph_dsid,vid,ip,&local_val,direction,
			rtl8651_wanStatus[pkthdrPtr->ph_dsid], pkthdrPtr, RTL8651_DOS_SLOW_CHECK) == SUCCESS)
		{
			goto dos_out;
		}
		/* check for TCP Scan */
		_rtl8651_dosScanTracking(pkthdrPtr, ip, IPPROTO_TCP, (void*)tc);	/* suspect source ? */
		#endif /* _RTL_DOS_PLUGIN */

		/* ========================== Make forwarding decision here ========================== */
		_rtl8651_l34Decision(	pkthdrPtr,
								local_vlanp,
								&decision,
								&srcType,
								&dstType,
								&optLayer);
		/* ======================================================================== */

		/* ingress ACL */
		if (inAcl_done == FALSE)
		{
			local_val = _rtl8651_ingressAcl(pkthdrPtr->ph_dsid,inAcl_index,local_vlanp,vid,m_data,(uint8 *)ip, RTL8651_ACL_GENERIC, optLayer);

			if (local_val & FWDENG_DROP)
			{
				goto out;
			}
			if (local_val == FWDENG_ACL_IN_USERTRAP)
			{

#ifdef _RTL_PS_ACTION
				
				/* Check if there's any rule for NAPT port bouncing pkt to protocol stack. */
				int32 val;

				val = _rtl8651_doProtoStackAction(ip);
				if (val == PS_ACTION_TO_PS)
					local_val = FWDENG_TO_GW;
				else if (val == PS_ACTION_DROP)
					local_val = FWDENG_USER_DROP;

#endif	/* _RTL_PS_ACTION */

				goto out;
			}
				
			inAcl_done = TRUE;
		}

#define DIFFSERV		1
#if DIFFSERV
		/* <ghhuang> */
		/* Check SYN, FIN, RST packets for matching ACL */
		{
			_rtl8651_tblDrvAclRule_t *rule_matched;

			rule_matched = (_rtl8651_tblDrvAclRule_t *) _rtl8651_matchDiffservACL(pkthdrPtr->ph_dsid, ip);
			if (rule_matched)
			{
				_rtl8651_markDiffservPkt(rule_matched, ip);
			}
		}
#endif
		/* goto IP normal process */
		goto normal_process;
	}else if (pkthdrPtr->ph_proto == PKTHDR_UDP)
	{
		rtl8651_fwdEngineCounter.udpReceive++;
		ud=(struct udphdr *)l4hdr;
		pkthdrPtr->ph_sport = ntohs(ud->uh_sport);

		/*
			ASIC will set "CSUM_L4_OK" if this packet has no L4 header (fragment packet).
			So, we can check this flag no matter this packet is fragment or not.
		*/
		if(ISCLEARED(pkthdrPtr->ph_flags, (CSUM_L4_OK)))
		{
			rtl8651_fwdEngineCounter.udpErrorReceive++;
			local_val = FWDENG_UDP_BADCKSUM;
			goto out;/* UDP packet with layer4 checksum error */
		}

		/* Layer 3 check */
		if (sip == 0)
		{
			/**************************************************************************
				If source IP is 0, we trap all packets to protocol stack except DHCP protocol packets.
				For fragment packet, we trap ALL packets with zero-source IP,
				but the side effect is, if the fragmented DHCP packet come, system will drop all
				remaining packets
			 **************************************************************************/
			if (IS_FRAGMENT_REMAIN(ip->ip_off) ||	/* no port can be checked*/
				((ntohs(ud->uh_sport) != 68)/*dhcp client to server*/ && (ntohs(ud->uh_sport) != 67/* dhcp server to client*/)))
			{
				local_val = FWDENG_IP_INVALID;
				goto out;
			}
		}

		#ifdef _RTL_DOS_PLUGIN
		/* UDP DOS */
		if (rtl8651a_dosUdpCheck(pkthdrPtr->ph_dsid,vid,ip,&local_val,direction,
			rtl8651_wanStatus[pkthdrPtr->ph_dsid], pkthdrPtr, RTL8651_DOS_FAST_CHECK) == SUCCESS)
		{
			goto dos_out;
		}
		#endif /* _RTL_DOS_PLUGIN */	

		#ifdef _RTL_NAPT_SHORT_CUT
		/* UDP fast path */
		rtl8651_drvPacketFastSend = TRUE;		/* Enter fast mode */
		{
			if (local_vlanp->internal)
				local_val = _rtl8651_l4UdpFastAliasOut(pkthdrPtr, ip);
			else
				local_val = _rtl8651_l4UdpFastAliasIn(pkthdrPtr, ip);
		}
		rtl8651_drvPacketFastSend = FALSE;		/* Exit fast mode */

		if (local_val == SUCCESS)
		{
			rtl8651_fwdEngineCounter.udpSend++;
			goto out;
		}

		if (local_val == FWDENG_TO_GW)
		{	/* with checking napt entry, system detects this packet must be trapped to protocol stack */
			goto out;
		}

		if ((local_val > 0) && (local_val & FWDENG_DROP))
		{
			rtl8651_fwdEngineCounter.natdrop++;
			goto out;
		}
		/* Fast path FAILED, try slow path */
		#endif /* _RTL_NAPT_SHORT_CUT */

		#ifdef _RTL_DOS_PLUGIN
		/* UDP DOS */
		if (rtl8651a_dosUdpCheck(pkthdrPtr->ph_dsid,vid,ip,&local_val,direction,
			rtl8651_wanStatus[pkthdrPtr->ph_dsid], pkthdrPtr, RTL8651_DOS_SLOW_CHECK) == SUCCESS)
		{
			goto dos_out;
		}
		/* check for UDP Scan */
		_rtl8651_dosScanTracking(pkthdrPtr, ip, IPPROTO_UDP, (void*)ud);	/* suspect source ? */
		#endif /* _RTL_DOS_PLUGIN */	

		/* ========================== Make forwarding decision here ========================== */
		_rtl8651_l34Decision(	pkthdrPtr,
								local_vlanp,
								&decision,
								&srcType,
								&dstType,
								&optLayer);
		/* ======================================================================== */

		/* ingress ACL */
		if (inAcl_done == FALSE)
		{
			local_val = _rtl8651_ingressAcl(pkthdrPtr->ph_dsid,inAcl_index,local_vlanp,vid,m_data,(uint8 *)ip, RTL8651_ACL_GENERIC, optLayer);
			if (	((local_val > 0) && (local_val&FWDENG_DROP)) ||
				(local_val == FWDENG_ACL_IN_USERTRAP)
			)
			{
				goto out;
			}
			inAcl_done=TRUE;
		}

		goto normal_process;
	}else if (pkthdrPtr->ph_proto==PKTHDR_ICMP)
	{
		struct icmp *ic;
		assert(ip);
		ic = (struct icmp *) l4hdr;
		pkthdrPtr->ph_sid = ntohs(ic->icmp_id);
		rtl8651_fwdEngineCounter.icmpReceive++;
		/*
			ASIC will set "CSUM_L4_OK" if this packet has no L4 header (fragment packet).
			So, we can check this flag no matter this packet is fragment or not.
		*/
		if(ISCLEARED(pkthdrPtr->ph_flags, (CSUM_L4_OK)))
		{
			rtl8651_fwdEngineCounter.icmpErrorReceive++;
			local_val=FWDENG_ICMP_BADCKSUM;
			goto out;/* ICMP packet with layer4 checksum error */
		}

		/* Layer 3 check */
		if (sip == 0)
		{
			local_val = FWDENG_IP_INVALID;
			goto out;
		}

		#ifdef _RTL_DOS_PLUGIN
		/* ICMP DOS */
		if (rtl8651a_dosIcmpCheck(pkthdrPtr->ph_dsid,vid,ip,&local_val,direction,
			rtl8651_wanStatus[pkthdrPtr->ph_dsid], pkthdrPtr) == SUCCESS)
		{
			goto dos_out;
		}
		#endif /* _RTL_DOS_PLUGIN */

		/* ========================== Make forwarding decision here ========================== */
		_rtl8651_l34Decision(	pkthdrPtr,
								local_vlanp,
								&decision,
								&srcType,
								&dstType,
								&optLayer);
		/* ======================================================================== */

		/* for ACL actionType is IP or ICMP , 2004/2/11 */
		if (inAcl_done == FALSE)
		{
			local_val=_rtl8651_ingressAcl(pkthdrPtr->ph_dsid,inAcl_index,local_vlanp,vid,m_data,(uint8 *)ip, RTL8651_ACL_GENERIC, optLayer);
			if (((local_val > 0) && (local_val&FWDENG_DROP))||(local_val==FWDENG_ACL_IN_USERTRAP))
			{
				goto out;
			}
			inAcl_done = TRUE;
		}

		if (_rtl8651_isLocalBcastAddr(vid, pkthdrPtr->ph_dip))
		{
			_rtl8651_l2Relay(mbufPtr);
		}
			
		goto normal_process;
	}else if (pkthdrPtr->ph_proto == PKTHDR_IP||pkthdrPtr->ph_proto == PKTHDR_PPTP)
	{
		rtl8651_fwdEngineCounter.ipReceive++;
#ifdef _RTL_DOS_PLUGIN
		/* IP DOS */
		/*
			chenyl: IP-DOS has been checked in DOS TCP/UDP/ICMP Check functions,
			so we do dosIpCheck only when ph_protol is IP/PPTP
		*/
		if (rtl8651a_dosIpCheck(pkthdrPtr->ph_dsid,vid,ip,&local_val,direction,rtl8651_wanStatus[pkthdrPtr->ph_dsid], pkthdrPtr)
			== SUCCESS)
		{
			goto dos_out;
		}
#endif /* _RTL_DOS_PLUGIN */	

		/* ========================== Make forwarding decision here ========================== */
		_rtl8651_l34Decision(	pkthdrPtr,
								local_vlanp,
								&decision,
								&srcType,
								&dstType,
								&optLayer);
		/* ======================================================================== */

		if (inAcl_done == FALSE)
		{
			local_val=_rtl8651_ingressAcl(pkthdrPtr->ph_dsid,inAcl_index,local_vlanp,vid,m_data,(uint8*)ip, RTL8651_ACL_GENERIC, optLayer);
			if (((local_val > 0) && (local_val&FWDENG_DROP))||(local_val==FWDENG_ACL_IN_USERTRAP))
			{
				goto out;
			}
			inAcl_done = TRUE;
		}

		/* if DIP is one of the IP broadcast addresses on this VLAN, broadcast */
		if (_rtl8651_isLocalBcastAddr(vid, pkthdrPtr->ph_dip))
		{
			_rtl8651_l2Relay(mbufPtr);
		}

		/*****************************************************************
				chenyl (2004/7/27):
					we assume that "Protocol Based NAT" process packet which
						L3 protocl "is" "IP"
						L4 protocol "is NOT" "TCP/UDP/ICMP"
					now, so we process protocol based NAT when ASIC tell me this
					packet's protocol is PKTHDR_IP or PKTHDR_PPTP
		******************************************************************/
		if (rtl8651_drvProtocolBasedNatEnable)
		{	/*	The ip protocol is not handled by our forwarding engine,
				maybe it is protocol-based NAT ? */
			if ( local_vlanp->internal )
			{
				/* from inernal */
				ipaddr_t newExtIp;
				if ( SUCCESS == _rtl8651_getOutboundProtocolBasedNAT(ip_protocol, &newExtIp, sip))
				{
					/* yes, this is protocol-based NAT. */
					PKTSET_UINT32_UNALIGNED(newExtIp, &ip->ip_src);
					if (!IPHDRLEN_IS_TWENTY(ip))
					{	/* adjust IP checksum */
						_RTL8651_L3_ADJUST_CHECKSUM(newExtIp, pkthdrPtr->ph_sip, ip->ip_sum);
					}
					pkthdrPtr->ph_proto = PKTHDR_IP;	/* disable L4 checksum offload */
					local_val = _rtl8651_l3ActRouting( pkthdr, pkthdrPtr->ph_dip, ip );
					goto out;
				}
			}else
			{
				/* from external */
				ipaddr_t newIntIp;
				if ( SUCCESS == _rtl8651_getInboundProtocolBasedNAT(ip_protocol, dip, &newIntIp))
				{
					/* yes, this is protocol-based NAT. */
					PKTSET_UINT32_UNALIGNED(newIntIp, &ip->ip_dst);
					if (!IPHDRLEN_IS_TWENTY(ip))
					{	/* adjust IP checksum */
						_RTL8651_L3_ADJUST_CHECKSUM(newIntIp, pkthdrPtr->ph_dip, ip->ip_sum);
					}
					pkthdrPtr->ph_proto = PKTHDR_IP;	/* disable L4 checksum offload */
					local_val = _rtl8651_l3ActRouting(pkthdr, newIntIp, ip );
					goto out;
				}
			}
		}

		goto normal_process;
	}else if (pkthdrPtr->ph_proto == PKTHDR_IGMP)
	{
		/* we only forward IGMP packet when rtl8651_drvFwdUnicastIGMPEnable  is enabled */
		assert(rtl8651_drvFwdUnicastIGMPEnable == TRUE);

		#ifdef _RTL_DOS_PLUGIN
		/* we only check IGMP with L3 dos */
		if (rtl8651a_dosIpCheck(pkthdrPtr->ph_dsid,vid,ip,&local_val,direction,rtl8651_wanStatus[pkthdrPtr->ph_dsid], pkthdrPtr)
			== SUCCESS)
		{
			goto dos_out;
		}
		#endif /* _RTL_DOS_PLUGIN */	

		/* ========================== Make forwarding decision here ========================== */
		_rtl8651_l34Decision(	pkthdrPtr,
								local_vlanp,
								&decision,
								&srcType,
								&dstType,
								&optLayer);
		/* ======================================================================== */

		if (inAcl_done == FALSE)
		{
			local_val=_rtl8651_ingressAcl(pkthdrPtr->ph_dsid,inAcl_index,local_vlanp,vid,m_data,(uint8*)ip, RTL8651_ACL_GENERIC, optLayer);
			if (((local_val > 0) && (local_val&FWDENG_DROP))||(local_val==FWDENG_ACL_IN_USERTRAP))
			{
				goto out;
			}
			inAcl_done = TRUE;
		}

		goto normal_process;
	}

	/* Unknown L3 protocol: Do ingress filter if not yet done. */
	if (inAcl_done == FALSE)
	{
		/* 2005.09.25 Because of unknown protocol, don't match ACL rules with packet operation layer defined. */
		local_val=_rtl8651_ingressAcl(pkthdrPtr->ph_dsid,inAcl_index,local_vlanp,vid,m_data,(uint8*)ip, RTL8651_ACL_GENERIC, _RTL8651_OPERATION_OTHER);
		if (((local_val > 0) && (local_val&FWDENG_DROP))||(local_val==FWDENG_ACL_IN_USERTRAP))
		{
			goto out;
		}
		inAcl_done = TRUE;
	}

	local_val = FWDENG_L3UNKNOWNPROTO;
	goto out;		

normal_process:

	assert(ip);

#ifdef CONFIG_RTL865X_MULTICAST
/* Multicast packet slowpath process */
	if ((isMulticast == TRUE) && 
		(rtl8651_drvMulticastProcessEnable))
	{
		local_val = _rtl8651_multicast_process(pkthdrPtr, ip);
		goto out;
	}
#endif /* CONFIG_RTL865X_MULTICAST */

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_FASTPATH);
#endif

	/***********************************************/
	/************** L3/4 Slow Path Process *************/
	/************************************************/

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_SLOWPATH);
#endif

	/********* Make forwarding decision here if it does not do it yet *********/
	if (decision == _RTL8651_DEC_NONE)
	{
		_rtl8651_l34Decision(	pkthdrPtr,
								local_vlanp,
								&decision,
								&srcType,
								&dstType,
								&optLayer);
	}


	/********* Advanced routing process *********/
	local_val = 0;
	local_val = _rtl8651_l34PacketProcess(pkthdrPtr, ip, decision);
	if(local_val != SUCCESS)
	{
		rtl8651_fwdEngineCounter.l34actionLookupFailed++;
		goto out;
	}

	/********* Slow Path packet dispatch *********/
	local_val = _rtl8651_l34ActExec(pkthdrPtr,ip,ip_protocol, decision);

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_SLOWPATH);
#endif

	/********* Other Processes *********/
	if(local_val != SUCCESS)
	{
		if ( local_val > 0 )
		{
			if ((local_val & ~FWDENG_DROP) == FWDENG_USER_IDSCHECKDROP)
			{
				goto dos_out;
			}
			if ( local_val & FWDENG_DROP )
			{
				rtl8651_fwdEngineCounter.l34actionFailed++;
				goto out;
			}
		}

		/********* Patch function for DNS packet damand route triggering *********/
		_rtl8651_drvPktDemandRouteTrigger(ip);

		/********* Forwarding engine icmp reply module *********/
		if (rtl8651_drvIcmpReplyEnable && ip_protocol == IPPROTO_ICMP)
		{
			if (_rtl8651_drvIcmpAgent(pkthdrPtr,ip,vid) == SUCCESS)
			{
				local_val = SUCCESS;
			}else
			{
				rtl8651_fwdEngineCounter.l34actionFailed++;
			}
			goto out;
		}

		/* vlanIdx of PPPoE/IPv6 passthru packet should be overwriten by WAN vid. */
		/* _modifyPPPoEPassthruVid( pkthdrPtr, ethType, local_vlanp ); */

		rtl8651_fwdEngineCounter.l34actionFailed++;
		goto out;
	}

	switch (ip_protocol)
	{
		case IPPROTO_TCP:
			rtl8651_fwdEngineCounter.tcpSend++;
			break;
		case IPPROTO_UDP:
			rtl8651_fwdEngineCounter.udpSend++;
			break;
		case IPPROTO_ICMP:			
			rtl8651_fwdEngineCounter.icmpSend++;
			break;
	}
	goto out;

dos_out:
	rtl8651_fwdEngineCounter.dosDrop++;

out:	
	/* always restore rtl8651_drvPacketFastSend when exiting forwarding engine */
	rtl8651_drvPacketFastSend = FALSE;

#ifdef _RTL8651_FWDENG_RETURN_STATISTIC
	if (local_val != SUCCESS)
	{
		_rtl8651_fwdEngineRetValueStatisticUpdate(local_val);
	}
#endif /* _RTL8651_FWDENG_RETURN_STATISTIC */

	/* local_val may be >0(FWDENG_XXX), ==0(SUCCESS) or <0(FAILED) */
	if(local_val > 0)
	{
		if (	((local_val & FWDENG_DROP) == 0) &&
			(_dropCases[local_val&(FWDENG_DROP-1)]))
		{
			local_val |= FWDENG_DROP;	/* user configured to drop this pkt. */
		}

		if (	((local_val & ~FWDENG_DROP) == FWDENG_ACL_DROP_NOTIFY) &&
			(pkthdrPtr->ph_proto!=PKTHDR_ETHERNET))
		{
			int32 DropBit = (local_val & FWDENG_DROP);
			_rtl8651_IcmpPktField_t icmpInfo;
			icmpInfo.ic_type	= ICMP_UNREACH;
			icmpInfo.ic_code	= ICMP_UNREACH_HOST;
			icmpInfo.ic_mtu		= 0;
			pkthdrPtr->ph_portlist	= srcPortlist;
			local_val = _rtl8651_drvIcmpErrorGeneration(pkthdrPtr, ip, &icmpInfo, local_vlanp->vid);
			if (local_val != SUCCESS)
			{
				local_val |= DropBit;
			}
		}
		
		if(local_val & FWDENG_DROP)
		{
			mBuf_driverFreeMbufChain(pkthdrPtr->ph_mbuf);
#ifdef _RTL_QUEUE_FRAG_PKT
			_rtl8651_fragQueueDropChain = TRUE;
#endif
			local_val = SUCCESS;
		}
		if(local_val == FWDENG_QUEUED)
		{
			local_val = SUCCESS;
		}

		/* vlanIdx of PPPoE/IPv6 passthru packet should be overwriten by WAN vid. */
		/* _modifyPPPoEPassthruVid( pkthdrPtr, ethType, local_vlanp ); */
	}

	if (local_val != SUCCESS)
	{
#ifdef _RTL_TRAPPING_DISPATCH
		{
			/* If there is no other categories matched, we use the default one */
			uint16 category = RTL8651_DEFAULT_TRAPPING_DISPATCH_CATEGORY;

			if ( (ethType == 0x0800 || ((ethType == 0x8864) && (ncp = 0x0021))) &&
				(pkthdrPtr->ph_iphdrOffset) &&
				(!_rtl8651_fragQueueProcess))
			{
				/*
					Note: we CAN NOT use the pointer 'ip' in this function call because that
					for fragment queue, it might not point to the first fragment packet in queue.
				*/
				category = _rtl8651_getTrappingDispatchCategory((struct ip *)(pkthdrPtr->ph_mbuf->m_data + pkthdrPtr->ph_iphdrOffset));
			}

			/* process fragment packet here */
			{
				struct rtl_pktHdr *pHdr = pkthdrPtr;

				while (pHdr)
				{
					pHdr->ph_category = category;	/* fill current packet */
					pHdr = pHdr->ph_nextHdr;		/* get next packet */
				}
			}
		}
#endif /* _RTL_TRAPPING_DISPATCH */

#ifdef _RTL_PPTPL2TP_HDR_REFILL
		if ( (ethType == 0x0800) &&
		      (pkthdrPtr->ph_iphdrOffset) &&
		      (!_rtl8651_fragQueueProcess))
		{
			/*
				In header refill system, fragment packets are processed.

				Note: we CAN NOT use the pointer 'ip' in this function call because that
				for fragment queue, it might not point to the first fragment packet in queue.
			*/
			_rtl8651_refillHeader( local_val,
								  pkthdrPtr,
								  (struct ip *)(pkthdrPtr->ph_mbuf->m_data + pkthdrPtr->ph_iphdrOffset),
								  local_vlanp,
								  srcPortlist);
		}
#endif /* _RTL_PPTPL2TP_HDR_REFILL */

#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		/*
			chenyl:
				restore VLAN index if packet is from PPTP/L2TP loopback vlan/port.
		*/
		{
			struct rtl_pktHdr *pHdr = pkthdrPtr;
			while (pHdr)
			{
				if (pHdr->ph_vlanIdx == rtl8651_vlanTableIndex(tunnel.loopBackVid))
				{
					pHdr->ph_vlanIdx = rtl8651_vlanTableIndex(tunnel.wanVid);
					pHdr->ph_portlist = tunnel.wanPort;
	}

				pHdr = pHdr->ph_nextHdr;	/* get next packet */
			}
		}
#endif	/* defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP) */

#ifdef RTL865X_PSEUDO_VLAN
		/*
			Protocol stack only know the main vlan,
			so if the packet is from one of the subVlan, we need to modify it.
		*/
		{
			struct rtl_pktHdr *pHdr = pkthdrPtr;
			rtl8651_tblDrv_vlanTable_t *srcVlan = NULL;

			while (pHdr)
			{
				srcVlan = &DrvTbl.vlan[pkthdrPtr->ph_vlanIdx];

				assert(srcVlan->valid);

				if (srcVlan->fvlan_t)	/* this packet is from pseudo-vlan and its soruce vlan index need to be mdofied */
				{
					pHdr->ph_vlanIdx = rtl8651_vlanTableIndex(srcVlan->fvlan_t->vid);
				}

				pHdr = pHdr->ph_nextHdr;	/* get next packet */
			}
		}
#endif	/* RTL865X_PSEUDO_VLAN */

	}

	rtlglue_drvMutexUnlock();/* Unlock resource */
	return local_val;	
}

/*
@func int32		| rtl8651_fwdEngineRegisterAlgModule	| Register ALG module to forwarding engine.
@parm uint16 | alg_port | ALG port number 
@parm uint16  | isTcp | TRUE for TCP, FALSE for UDP
@parm	alg_funcptr | alg_outboundClientcb | ALG hook function for client's outbound packet.
@parm	alg_funcptr | alg_inboundClientcb |	ALG hook function for client's inbound packet.
@parm	alg_funcptr | alg_outboundServercb |	ALG hook function for server's outbound packet.
@parm	alg_funcptr | alg_inboundServercb |	ALG hook function for server's inbound packet.
@comm 
Register ALG module by port number.
*/
 int32 rtl8651_fwdEngineRegisterAlgModule(uint16 alg_port, uint16 isTcp, 
	int32	(*alg_init)(void *),
	alg_funcptr alg_outboundClientcb,
	alg_funcptr alg_inboundClientcb,
	alg_funcptr alg_outboundServercb,
	alg_funcptr alg_inboundServercb) 	
{									
	
	rtl8651_tblDrv_algEntry_t *tb;
	int8 isServer, isClient;
	int32 retval = FAILED;

	rtlglue_drvMutexLock();

	tb = _rtl8651_getAlgEntry(isTcp, &isServer, &isClient, alg_port, alg_port);
	
	if (tb)
	{ /* link already exist */
		if ( tb->fwEngAdd)
		{
			retval = SUCCESS;
			goto out;
		}else
		{
			retval = FAILED;
			goto out;
		}
	}
	if (alg_init && (SUCCESS != alg_init(NULL)))
	{
		retval = FAILED;
		goto out;
	}

	isClient = TRUE;
	isServer = TRUE;
	retval = rtl8651_addALGRule(alg_port, alg_port, isTcp, isServer, isClient);

	if (retval != SUCCESS)
	{
		goto out;
	}

	tb = _rtl8651_getAlgEntry(isTcp, &isServer, &isClient, alg_port, alg_port);

	assert(tb);
	
	tb->fwEngAdd = 1; /* set add from forwarding engine */
	tb->alg_outboundClientcb = (alg_funcptr_t)alg_outboundClientcb; 
	tb->alg_inboundClientcb =  (alg_funcptr_t)alg_inboundClientcb;
	tb->alg_outboundServercb = (alg_funcptr_t)alg_outboundServercb; 
	tb->alg_inboundServercb =  (alg_funcptr_t)alg_inboundServercb;	

	retval = SUCCESS;

out:
	rtlglue_drvMutexUnlock();

	return retval;
	
}

/*
@func int32		| rtl8651_fwdEngineRemoveAlgModule	| Remove ALG module by port number.
@parm uint16 | alg_port | ALG port number 
@parm uint16  | isTcp | TRUE for TCP, FALSE for UDP
@comm 
Register ALG module by port number.
*/
int32 rtl8651_fwdEngineRemoveAlgModule(uint16 alg_port, uint16 isTcp)
{

	rtl8651_tblDrv_algEntry_t *tb;
	int8 isServer, isClient;
	int32 retval = FAILED;

	rtlglue_drvMutexLock();

	tb = _rtl8651_getAlgEntry(isTcp, &isServer, &isClient, alg_port, alg_port);
	
	if(tb)
	{ /* link already exist */
		if( tb->fwEngAdd == 0)
		{
			retval = FAILED;
			goto out;
		}
	} else
	{
		retval = FAILED;
		goto out;
	}

	retval = rtl8651_delALGRule(alg_port, alg_port, isTcp, isServer, isClient);

out:
	rtlglue_drvMutexUnlock();
	return retval;
}


void _rtl8651_fwdEngineTimeUpdate(uint32 secPassed)
{

#ifdef _RTL_QUEUE_ARP_PACKET
	_rtl8651_queuePacketSecondTimer();
#endif /* _RTL_QUEUE_ARP_PACKET */

#ifdef _RTL_UPNP
	_rtl8651_uPnPOneSecondTimer();
#endif /* _RTL_UPNP */

#ifdef _RTL_DOS_PLUGIN
	rtl8651_dosOneSecondTimer();
#endif /* _RTL_DOS_PLUGIN */

#ifdef _RTL_PORT_BOUNCING
	_rtl8651_portBouncingTimeUpdate(secPassed);
#endif

#ifndef RTL865X_TEST
	rtl8651_algOneSecondTimer();
#endif /* RTL865X_TEST */

#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
	_rtl8651_fragmentShortestPathCache_timeUpdate(secPassed);
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */

#ifdef _RTL_NAPT_REDIRECT
	_rtl8651_naptRedirectFlowTimeUpdate(secPassed);
#endif /* _RTL_NAPT_REDIRECT */ 

#ifdef _RTL_NAPT_FRAG
	_rtl8651_fragCache_timeUpdate(secPassed);
#endif /* _RTL_NAPT_FRAG */

#ifdef _RTL_TRIGGER_PORT
	_rtl8651_TriggerPortTimeUpdate(secPassed);
#endif /* _RTL_TRIGGER_PORT */

}



/*********************************************************************************
* Routine Name :    _rtl8651_l4SeqDiff                                 
* Description :                                   
* Input :                                                               
* Output :                                                             
* Return :                    
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4SeqDiff(uint32 x, uint32 y)
{

	/*
	    This function is encapsulated in case there are any unusual
	    arithmetic conditions that need to be considered.
	*/
	return ((y) - (x));
}

/*********************************************************************************
* Routine Name : _rtl8651_l4GetDeltaSeqOut
* Description :              
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
int32 _rtl8651_l4GetDeltaSeqOut(struct ip * pip, struct tcphdr *tc, rtl8651_tblDrv_naptTcpUdpFlowEntry_t  *tb)
{

	int32     i;
	int32     delta, seq_diff_min;
	uint32    seq;

	seq = PKTGET_UINT32_UNALIGNED(&tc->th_seq);

	delta = 0;
	seq_diff_min = -1;
	for (i = 0; i < L4_N_LINK_TCP_DATA; i++)
	{
		struct l4ack_data_record x;

		rtl8651_memcpy(&x,&tb->ack[i], sizeof(struct l4ack_data_record));
		
		if (x.active == 1)
		{
			int32     seq_diff;

			seq_diff = _rtl8651_l4SeqDiff(x.ack_old, seq);
			if (seq_diff >= 0)
			{
				if (seq_diff_min >= 0)
				{
					if (seq_diff < seq_diff_min)
					{
						delta = x.delta;
						seq_diff_min = seq_diff;
					}
				}
				else
				{
					delta = x.delta;
					seq_diff_min = seq_diff;
				}
			}
		}
	}

	return (delta);
}



#define FWDENG_INTERNAL_FUNCTIONS
////////////////////////////////////////////
//// Internal  function implementation
////////////////////////////////////////////

#define _rtl8651__INRANGE(c,l,h) (((uint32) ((c) - (l))) <= ((uint32) ((h)-(l))))

int32 _rtl8651_isdigit(int32 c)
{
	return  _rtl8651__INRANGE(c,'0','9');
}

int32 _rtl8651_isspace(int32 c )
{
	return (c==0x20) | _rtl8651__INRANGE(c,'\t','\r');
}

#ifdef _RTL_NAPT_FRAG
/*******************************************************************************/
/* <------------------------- fragment forwarding cache -------------------------> */

/* allocate memory space for fragment forwarding cache */
static uint32 _rtl8651_fragCacheInit(void)
{
	int i;

	/* init all fragment cache entry */
	for ( i = 0 ; i < RTL8651_FRAGCACHE_TYPE_COUNT ; i ++)
	{
		_rtl8651_fragCache[i] = NULL;
	}

	/* allocate fragment cache entry if we need */
	if (_rtl8651_max_fragCache_count)
	{
		for ( i = 0 ; i < RTL8651_FRAGCACHE_TYPE_COUNT ; i ++)
		{
			_rtl8651_fragCache[i] = (_rtl8651_frag_t *)rtlglue_malloc(sizeof(_rtl8651_frag_t) * (_rtl8651_max_fragCache_count));
			if (_rtl8651_fragCache[i] == NULL)
				return FAILED;
			/* clear entry */
			memset(_rtl8651_fragCache[i], 0, (sizeof(_rtl8651_frag_t) * (_rtl8651_max_fragCache_count)));
		}
	}

	return SUCCESS;
}

/* clean memory of fragment forwarding cache */
static void _rtl8651_fragCacheReinit(void)
{

	int i;

	for ( i = 0 ; i < RTL8651_FRAGCACHE_TYPE_COUNT ; i ++)
	{
		if (_rtl8651_fragCache[i])
		{
			memset(_rtl8651_fragCache[i], 0, (sizeof(_rtl8651_frag_t) * (_rtl8651_max_fragCache_count)));
		}
	}
}



/*
	get Exist fragment cache information

	Index :
			FlowType
			org_sip
			org_dip
			ip_id

	Gotten Info :
			newIp
			routeIdx
			naptEntry
			dsid

	return value:
			SUCCESS	: found
			FAILED		: not found
*/
static inline uint32 _rtl8651_getFragCache(int32 FlowType, ipaddr_t org_sip, ipaddr_t org_dip, uint16 ip_id, ipaddr_t *newIp, int8 *routeIdx, rtl8651_tblDrv_naptTcpUdpFlowEntry_t **naptEntryp, uint8 *dsid)
{
	int i;
	_rtl8651_frag_t *cachePtr;

	assert((FlowType >= 0) && (FlowType < RTL8651_FRAGCACHE_TYPE_COUNT));

#ifdef RTL8651_FRAGCACHE_DEBUG
	printk("[FRAG CACHE] -Get- Type[%d] sip[%u.%u.%u.%u] dip[%u.%u.%u.%u] id[%d]\n", FlowType, NIPQUAD(org_sip), NIPQUAD(org_dip), ip_id);
#endif /* RTL8651_FRAGCACHE_DEBUG */

	cachePtr = _rtl8651_fragCache[FlowType];

	for (i = 0 ; i < _rtl8651_max_fragCache_count ; i ++)
	{
#ifdef RTL8651_FRAGCACHE_DEBUG
	printk("[FRAG CACHE] -Entry Check- Entry [%d]\n\tage[%d]\n\torg sip[%u.%u.%u.%u] <-> [%u.%u.%u.%u]\n\torg dip[%u.%u.%u.%u] <-> [%u.%u.%u.%u]\n\tid[%d] <-> [%d]\n",
		i,
		cachePtr[i].age,
		NIPQUAD(cachePtr[i].org_sip), NIPQUAD(org_sip),
		NIPQUAD(cachePtr[i].org_dip), NIPQUAD(org_dip),
		cachePtr[i].ip_id, ip_id
		);
#endif /* RTL8651_FRAGCACHE_DEBUG */

		if (	(cachePtr[i].age > 0) &&
			(org_sip == cachePtr[i].org_sip) &&
			(org_dip == cachePtr[i].org_dip) &&
			(ip_id == cachePtr[i].ip_id)	)
		{
#ifdef RTL8651_FRAGCACHE_DEBUG
	printk("[FRAG CACHE] -Get OK- new Ip[%u.%u.%u.%u] rtIdx[%d] naptEntry[%p] dsid[%d]\n"
		, NIPQUAD(cachePtr[i].new_ip), cachePtr[i].routeIdx, cachePtr[i].napt_entry_t, cachePtr[i].dsid);
#endif /* RTL8651_FRAGCACHE_DEBUG */

			/* set information */
			if (newIp)
				*newIp = cachePtr[i].new_ip;
			if (routeIdx)
				*routeIdx = cachePtr[i].routeIdx;
			if (naptEntryp)
			{
				if (cachePtr[i].napt_entry_t && ((rtl8651_tblDrv_naptTcpUdpFlowEntry_t*)(cachePtr[i].napt_entry_t))->alive)
					*naptEntryp = cachePtr[i].napt_entry_t;
				else
					*naptEntryp = NULL;
			}
			if (dsid)
				*dsid = cachePtr[i].dsid;

			return SUCCESS;
		}
	}

#ifdef RTL8651_FRAGCACHE_DEBUG
	printk("[FRAG CACHE] -Get FAIL-\n");
#endif /* RTL8651_FRAGCACHE_DEBUG */

	return FAILED;
}

/*
	get Exist fragment cache entry

	Index :
			FlowType
			org_sip
			org_dip
			ip_id

	return value :
			pointer of found entry
*/
static inline _rtl8651_frag_t *_rtl8651_getFragCacheEntry(int32 FlowType, ipaddr_t org_sip, ipaddr_t org_dip, uint16 ip_id)
{
	int i;
	_rtl8651_frag_t *cachePtr;

	assert((FlowType >= 0) && (FlowType < RTL8651_FRAGCACHE_TYPE_COUNT));

	cachePtr = _rtl8651_fragCache[FlowType];

	for (i = 0 ; i < _rtl8651_max_fragCache_count ; i ++)
	{
		if (	(cachePtr[i].age > 0) &&
			(org_sip == cachePtr[i].org_sip) &&
			(org_dip == cachePtr[i].org_dip) &&
			(ip_id == cachePtr[i].ip_id)	)
		{
			return &cachePtr[i];
		}
	}
	return NULL;
}


/*
	update fragment cache information

	Index :
			FlowType

	Set Info :
			org_sip
			org_dip
			ip_id
			routeIdx
			naptEntry
*/
static inline void _rtl8651_updateFragCache(int32 FlowType, ipaddr_t org_sip, ipaddr_t org_dip, uint16 ip_id, ipaddr_t ip_addr, int8 routeIdx, rtl8651_tblDrv_naptTcpUdpFlowEntry_t *naptEntry, uint8 dsid)
{
	int i;
	_rtl8651_frag_t *cachePtr;
	_rtl8651_frag_t *entryToAdd;
	uint32 minAge = _rtl8651_max_fragCache_age + 1;

	assert((FlowType >= 0) && (FlowType < RTL8651_FRAGCACHE_TYPE_COUNT));
	assert(dsid < RTL8651_MAX_DIALSESSION);

#ifdef RTL8651_FRAGCACHE_DEBUG
	printk("[FRAG CACHE] -Update- Type[%d] sip[%u.%u.%u.%u] dip[%u.%u.%u.%u] id[%d]\n", FlowType, NIPQUAD(org_sip), NIPQUAD(org_dip), ip_id);
	printk("[FRAG CACHE] -data- new Ip[%u.%u.%u.%u] routeIdx[%d] naptEntry[%p] dsid[%d]\n", NIPQUAD(ip_addr), routeIdx, naptEntry, dsid);
#endif /* RTL8651_FRAGCACHE_DEBUG */

	/* try to get exist entry */
	cachePtr = _rtl8651_getFragCacheEntry(FlowType, org_sip, org_dip, ip_id);
	if (cachePtr)
	{	/* entry found */
#ifdef RTL8651_FRAGCACHE_DEBUG
	printk("[FRAG CACHE] -Exist- update IP[%s] -- RouteIdx[%s] -- NaptEntry[%s] -- dsid[%s]\n",
			(cachePtr->new_ip==ip_addr)?"No":"Yes",
			(cachePtr->routeIdx==routeIdx)?"No":"Yes",
			(cachePtr->napt_entry_t==naptEntry)?"No":"Yes",
			(cachePtr->dsid==dsid)?"No":"Yes"
			);
#endif /* RTL8651_FRAGCACHE_DEBUG */
		/* refill age */
		cachePtr->age = _rtl8651_max_fragCache_age;
		/* update information */
		cachePtr->new_ip = ip_addr;
		cachePtr->routeIdx = routeIdx;
		cachePtr->napt_entry_t = naptEntry;
		cachePtr->dsid = dsid;

		return;
	}

	/* entry is not found, we found an zero-age entry or min-age entry to fill new entry */
	cachePtr = _rtl8651_fragCache[FlowType];
	entryToAdd = NULL;

#ifdef RTL8651_FRAGCACHE_DEBUG
	int dbg_i = -1;
#endif /* RTL8651_FRAGCACHE_DEBUG */

	for ( i = 0 ; i < _rtl8651_max_fragCache_count ; i ++ )
	{
		if (cachePtr[i].age == 0)
		{
			entryToAdd = &cachePtr[i];

#ifdef RTL8651_FRAGCACHE_DEBUG
	printk("[FRAG CACHE] -Find Empty Entry to add- Entry index [%d]\n", i);
#endif /* RTL8651_FRAGCACHE_DEBUG */

			goto found;
		}
		if (cachePtr[i].age < minAge)
		{
			entryToAdd = &cachePtr[i];
			minAge = cachePtr[i].age;

#ifdef RTL8651_FRAGCACHE_DEBUG
			dbg_i = i;
#endif /* RTL8651_FRAGCACHE_DEBUG */

		}
	}
#ifdef RTL8651_FRAGCACHE_DEBUG
	printk("[FRAG CACHE] -Find Min age Entry to add- Entry index [%d]\n", dbg_i);
#endif /* RTL8651_FRAGCACHE_DEBUG */

found:
	assert(entryToAdd);
	/* refill age */
	entryToAdd->age = _rtl8651_max_fragCache_age;
	/* set index */
	entryToAdd->org_sip = org_sip;
	entryToAdd->org_dip = org_dip;
	entryToAdd->ip_id = ip_id;
	/* set information */
	entryToAdd->new_ip = ip_addr;
	entryToAdd->routeIdx = routeIdx;
	entryToAdd->napt_entry_t = naptEntry;
	entryToAdd->dsid = dsid;

	return;
}

/* update fragment timer age */
void _rtl8651_fragCache_timeUpdate(uint32 secPassed)
{
	int i;
	_rtl8651_frag_t *cachePtr;

	for ( i = 0 ; i < RTL8651_FRAGCACHE_TYPE_COUNT ; i ++)
	{
		cachePtr = _rtl8651_fragCache[i];
		if (cachePtr)
		{
			int j;
			for ( j = 0 ; j < _rtl8651_max_fragCache_count ; j ++ )
			{
				if (cachePtr[j].age > secPassed)
					cachePtr[j].age -= secPassed;
				else
				{
#ifdef RTL8651_FRAGCACHE_DEBUG
					if (cachePtr[j].age > 0)
						printk("[FRAG CACHE] -Expire - Type [%d] Entry index [%d] -- Time[%d]\n", i, j, _sysUpSeconds);
#endif /* RTL8651_FRAGCACHE_DEBUG */

					cachePtr[j].age = 0;
				}
			}	
		}
	}
}


#endif /* _RTL_NAPT_FRAG */

/****************************/

/*
	This function determines the packet is to gateway or not.

	Now, it's used for ICMP & ARP
*/
static int32 _rtl8651_pktAccessGateway (ipaddr_t srcIpAddr, ipaddr_t dstIpAddr, uint16 vid)
{
	uint32 natIdx;

	if (_rtl8651_extIpAddr(dstIpAddr, &natIdx) == TRUE)
	{
		goto true;
	}else if (_rtl8651_intIpAddr(dstIpAddr) == TRUE)
	{
		rtl8651_tblDrv_vlanTable_t *vlanp;

		vlanp = &DrvTbl.vlan[vid & (RTL8651_VLAN_NUMBER - 1)];

		if ((vlanp->internal) && (_rtl8651_intNetIpAddr(srcIpAddr) == TRUE))	/* only internal ip can send request to internal gip */
		{
			goto true;
 		}
	}

	return FALSE;
true:
	return TRUE;
}



/* <--- Layer2 forwarding ---> */
static int32 _rtl8651_l2Relay(struct rtl_mBuf *inMbufPtr)
{
	struct rtl_pktHdr *phdr;

	phdr = inMbufPtr->m_pkthdr;

#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
	if (DrvTbl.vlan[phdr->ph_vlanIdx].vid == tunnel.loopBackVid)
	{
		/*	chenyl:
			    If multicast packet is from PPTP/L2TP loopBack VLAN, It means there is no
			    other member port in this VLAN and we CAN NOT call fwdEngineSend to do
			    L2 forwarding														*/
	} else
	{
#endif
		struct rtl_mBuf *mbufPtr;
 		int32 retval;
	
		mbufPtr = mBuf_dupPacket(inMbufPtr, MBUF_DONTWAIT);

		if(mbufPtr == NULL)
		{
			#ifdef RTL865X_DEBUG
			rtlglue_printf("No more mbufs for relaying brdcast pkt\n");
			#endif
			return FAILED;
		}

		/* disable L3/4 checksum */
		mbufPtr->m_pkthdr->ph_flags &= (~(CSUM_IP | CSUM_L4));
		/* relay to all member ports except the incoming port */
		mbufPtr->m_pkthdr->ph_portlist = DrvTbl.vlan[phdr->ph_vlanIdx].memberPortUpStatus & ~(1<<phdr->ph_portlist);
		retval = _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)mbufPtr->m_pkthdr, DrvTbl.vlan[phdr->ph_vlanIdx].vid, -1);

		if (retval != SUCCESS)
		{
			mBuf_driverFreeMbufChain(mbufPtr);
		}

#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
	}
#endif

	return SUCCESS;
}


static int32 _rtl8651_drvArpGeneration(ipaddr_t ipAddr)
{
	uint32 natIdx;
	int32 retval;
	int8 *bufptr;
	ether_addr_t macAddr;
	rtl8651_tblDrv_networkIntfTable_t * netIntfPtr;
	rtl8651_tblDrv_ipIntfEntry_t * ipIntfPtr;
	struct rtl_mBuf *mbufPtr;

	if(_rtl8651_intNetIpAddr(ipAddr) == FALSE && _rtl8651_extNetIpAddr(ipAddr, &natIdx) == FALSE)
		return FAILED;
	if(_rtl8651_getArpEntry(NULL, ipAddr, TRUE) != NULL)	/* ARP entry exist */
		return FAILED;
	if((netIntfPtr =  _rtl8651_getNetworkIntfEntryByAddr(ipAddr)) == NULL)
		return FAILED;	/* No corresponding network interface exist */
	if((ipIntfPtr = _rtl8651_getIpIntfEntryByAddr(ipAddr)) == NULL)
		return FAILED;	/* No corresponding ip interface exist */
	if((mbufPtr = mBuf_get(MBUF_DONTWAIT, MBUFTYPE_DATA, 1)) == NULL)
		return FAILED;
	if(mBuf_getPkthdr(mbufPtr, MBUF_DONTWAIT) == (struct rtl_mBuf *) NULL)
	{
		mBuf_freeMbufChain(mbufPtr);
		return FAILED;
	}
	/* Casey, 2003/10/08 */
	mbufPtr->m_data = (void *)rtlglue_mbufClusterToData(mbufPtr->m_data);

	if (mBuf_reserve(mbufPtr,rtl8651_fwdEnginePara.mbufHeadroom))
	{
		mBuf_freeMbufChain(mbufPtr);
		return FAILED;
	}
	bufptr=(int8 *)mbufPtr->m_data;
	/* Construct destination MAC */
	memset(bufptr,0xff, 6);
	rtl8651_getVlanMacAddress(netIntfPtr->vid, &macAddr, NULL);
	/* Construct source MAC */
	rtl8651_memcpy(bufptr + 6,(int8 *)&macAddr.octet[0],6);
	/* construct Ethtype+ARP header */
	*(uint32 *)(bufptr + 12) = htonl(0x08060001);
	*(uint32 *)(bufptr + 16) = htonl(0x08000604);
	*(uint16 *)(bufptr + 20) = htons(0x0001);
	rtl8651_memcpy(bufptr + 22,(int8 *)&macAddr.octet[0],6);
	*(uint32 *)(bufptr + 28) = htonl(ipIntfPtr->ipAddr.slh_first->ipAddr);
	*(uint32 *)(bufptr + 32) = htonl(0x0);
	*(uint16 *)(bufptr + 36) = htonl(0x0);
	PKTSET_UINT32_UNALIGNED(ipAddr, bufptr+38);
	
	mbufPtr->m_len = 60;
	mbufPtr->m_pkthdr->ph_len = 60;
	mbufPtr->m_pkthdr->ph_portlist = (DrvTbl.vlan[netIntfPtr->vid&(RTL8651_VLAN_NUMBER-1)].memberPortMask);	
	retval = _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)mbufPtr->m_pkthdr,netIntfPtr->vid,-1 );
	if(retval == SUCCESS)
	{
		rtl8651_fwdEngineCounter.arpGenerate++;
	}else
	{
		mBuf_driverFreeMbufChain(mbufPtr);
	}
	return SUCCESS;	/* ARP generated */
}


static int32 _rtl8651_drvArpAgent(struct rtl_pktHdr *pkthdrPtr, struct rtl_mBuf *mbufPtr, uint16 vid)
{
	ipaddr_t srcIpAddr, dstIpAddr;
	int8 *bufptr;
	int32 retval;
	ether_addr_t macAddr;
	rtl8651_tblDrv_networkIntfTable_t * netIntfPtr;
	uint32 dummy = 14;
	uint32 etherTypePosition = 12;

	rtl8651_fwdEngineCounter.arpReceive++;	
	mbufPtr = pkthdrPtr->ph_mbuf;

	if ((netIntfPtr = _rtl8651_getNetworkIntfEntryByVid(vid)) == NULL)
	{
		return FWDENG_NETIFACENOTFOUND;	/* No layer3 network interface attached on receiving VLAN */
	}

	bufptr = (int8 *)mbufPtr->m_data;	

	if (_getEthTypeIpOffset(pkthdrPtr, bufptr, &etherTypePosition, &dummy, &dummy) != SUCCESS)
	{
		assert(0);
		return FWDENG_DROP;	/* never happen !*/
	}

	/* Get sender information to install ARP */
	memcpy(&macAddr, (bufptr+etherTypePosition+10),6);						/*	macAddr <- smac */
	srcIpAddr = PKTGET_UINT32_UNALIGNED(bufptr+(etherTypePosition+16));	/*	ipAddr	<- sip */

	if (srcIpAddr == 0)
	{
		return FWDENG_DROP;
	}

	if (_rtl8651_addLocalArp(TRUE, srcIpAddr, &macAddr, netIntfPtr, pkthdrPtr->ph_portlist) == SUCCESS)
	{
		rtl8651_fwdEngineCounter.arpAddSuccess++;

#ifdef _RTL_QUEUE_ARP_PACKET
		{
			struct rtl_pktHdr * q_pkt;
			struct ip *local_ip_p;
			uint32 local_dst_ip;
			int32 local_cnt = 0;
			struct ip fake_ip;
			fake_ip.ip_dst.s_addr = srcIpAddr;

			while (1)
			{
				local_cnt++;
				if (local_cnt > QUEUE_PACKET_ITEM)
				{
					_rtl8651_queuePacketSecondTimer();
					break;
				}
				q_pkt = _rtl8651_queryQuenePacket(&fake_ip, Q_ARP, &local_dst_ip, &local_ip_p);
				if (q_pkt == NULL)
				{
					break;
				}			
				
				if (_rtl8651_l3ActRouting(q_pkt, local_dst_ip, local_ip_p) != SUCCESS)
				{
					mBuf_freeMbufChain(q_pkt->ph_mbuf);
				}
			}
		}
#endif /* _RTL_QUEUE_ARP_PACKET */

	}else
	{
		rtl8651_fwdEngineCounter.arpAddFail ++;
	}
	
	if (*(uint16 *)(bufptr + (etherTypePosition+8)) == htons(0x1))
	{	/* ARP request */

		/* check if this packet is to Gateway or not */
		srcIpAddr = PKTGET_UINT32_UNALIGNED(bufptr + (etherTypePosition + 16));
		dstIpAddr = PKTGET_UINT32_UNALIGNED(bufptr + (etherTypePosition + 26));

		/*
			If this ARP request ...
				1. is to Gateway
				2. can access gateway
		*/
		if (_rtl8651_pktAccessGateway(srcIpAddr, dstIpAddr, vid) == FALSE)
		{
			/* This arp request does not query gateway */
			if (rtl8651_drvArpProxyProcessEnable)
			{	/* Arp Proxy */
				rtl8651_tblDrv_vlanTable_t *vlanp;
				vlanp = &DrvTbl.vlan[vid & (RTL8651_VLAN_NUMBER - 1)];
				if (vlanp->internal)	/* don't reply arp requests from external vlan */
				{
					rtl8651_tblDrv_routeTable_t *rt;
					rt =  _rtl8651_getRoutingEntry(dstIpAddr);
					if ((rt == NULL) || (rt->dvid != vid))	/* dip is in other vlan: we must reply gateway mac for this IP */
					{
						goto send_reply;
					}
				}
			}
			return FWDENG_ARP_OTHERS;	/* Not requesting myself */
		}
		/* This arp request asks for gateway mac */
		if (!rtl8651_drvArpProcessEnable)
		{	/* Check switch */
			return FWDENG_ARP_REQUEST;	
		}

send_reply:
		/* ==> ipAddr is the sender IP in arp reply */
		{
			struct rtl_mBuf *mbuf_arpReply = NULL;
			struct rtl_pktHdr *pkthdr_arpReply = NULL;
			uint8 *bufptr_reply;

			/*
				chenyl (2005 / 11 / 07):
					We always reply ARP request using newly allocated packet due to the idle-timer
					counting problem when we use the original packet to reply.
					=>	If we use the original packet to reply ARP request, its rxdesc must in the range of RX Ring.
						This would confuse the Tx module _rtl8651_fwdEngineSend() and treat this packet
						as a normal packet forwarding (then it might refresh idle-timer for this packet sending).
			*/
			if ((mbuf_arpReply = mBuf_dupPacket(pkthdrPtr->ph_mbuf, MBUF_DONTWAIT)) == NULL )
			{
				retval = FAILED;
				goto reply_out;
			}

			pkthdr_arpReply = mbuf_arpReply->m_pkthdr;
			bufptr_reply = mbuf_arpReply->m_data;

			rtl8651_memcpy((bufptr_reply + (etherTypePosition+20)), (bufptr_reply + (etherTypePosition+10)), 10);	/* Sender information change to receiver */
			*(uint16 *)(bufptr_reply + (etherTypePosition+8)) = htons(0x2);										/* Change OP to response */
			rtl8651_memcpy(bufptr_reply, (bufptr_reply + 6), 6);												/* Construct destination MAC: copy smac to dmac */
		rtl8651_getVlanMacAddress(vid, &macAddr, NULL);									/* Get source(gateway) MAC */
			rtl8651_memcpy((bufptr_reply+6), ((int8 *)&macAddr.octet[0]), 6);									/* Construct source MAC: use gateway mac of this vlan */
			rtl8651_memcpy((bufptr_reply+(etherTypePosition+10)), ((int8 *)&macAddr.octet[0]), 6);				/* Modify ARP sender field: gateway mac */
			PKTSET_UINT32_UNALIGNED(dstIpAddr, (bufptr_reply+(etherTypePosition+16)));						/* Modify ARP sender field: sip */

		/* Change port list definition */
			pkthdr_arpReply->ph_portlist = (1 << pkthdr_arpReply->ph_portlist);

			retval = _rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdr_arpReply, vid, -1);

reply_out:

		if(retval == SUCCESS)
		{
			rtl8651_fwdEngineCounter.arpReply++;
		}else
		{
				if ( mbuf_arpReply )
				{
					mBuf_driverFreeMbufChain( mbuf_arpReply );
		}
			}

			/* always free the original packet */
			mBuf_driverFreeMbufChain( pkthdrPtr->ph_mbuf );

			/* always return SUCCESS */
		return SUCCESS; /* Finish ARP process */
	}
	}
	return FWDENG_ARP_NOTREQ;
}

/* generate icmp error message */
int32 _rtl8651_drvIcmpErrorGeneration(struct rtl_pktHdr *pkthdrPtr, struct ip* ip, _rtl8651_IcmpPktField_t *icmp_info, uint16 vid) {
	rtl8651_tblDrv_networkIntfTable_t * netIntf;
	rtl8651_tblDrv_ipIntfEntry_t *ipIntf;
	int8 org_pkt[((ip->ip_vhl&0xf)<<2) + 8];
	uint16 org_pkt_len;
	struct icmp *ic;
	ipaddr_t sip;
	void *macHdr;
	ether_addr_t macAddr;
	int32 retval;
	uint16 trailer_len;		/* trailer length of icmp error message */
	uint16 icmp_ip_len;		/* ip length of icmp error message */
	uint16 icmp_total_len;	/* packet length of icmp error message (including ether-hdr, pppoe tag) */
	
	switch (icmp_info->ic_type)
	{
		case ICMP_TIMXCEED:
		case ICMP_UNREACH:
			break;
		default:
			return FAILED;	// Can't generate other icmp error message
	}


	// get network interface information
	netIntf = _rtl8651_getNetworkIntfEntryByVid(vid);
	if ((netIntf == NULL) ||(netIntf->valid == 0) || (SLIST_FIRST(&netIntf->IpHead) == NULL))
	{
		return FAILED;
	}
	ipIntf = SLIST_FIRST(&netIntf->IpHead);
	sip =ipIntf->ipAddr.slh_first->ipAddr;

	/*
		Copy the original packet content: IP header + 8 bytes payload
	*/
	{
		uint16 org_iphdr_len = (ip->ip_vhl&0xf)<<2;
		uint16 org_ip_len = ntohs(ip->ip_len);
		
		if ((org_ip_len - org_iphdr_len) > 8)
		{
			org_pkt_len = org_iphdr_len + 8;
			trailer_len = 0;
		}else
		{
			org_pkt_len = org_ip_len;
			trailer_len = 8 - (org_ip_len - org_iphdr_len);
		}
	}
	icmp_ip_len = 20 + 8 + org_pkt_len + trailer_len;		// icmp message's ip_len, ip hdr+icmp hdr+org pkt+trailer
	memset(org_pkt, 0, org_pkt_len);
	rtl8651_memcpy(org_pkt, ip, org_pkt_len);			// copy orginal packet to buffer

	// re-calculate pkt length
	icmp_total_len = pkthdrPtr->ph_iphdrOffset + icmp_ip_len;
	{
		uint16 org_total_len = pkthdrPtr->ph_len;
		if (org_total_len > icmp_total_len)
		{
			mBuf_adjTail(pkthdrPtr->ph_mbuf, (uint32)(org_total_len - icmp_total_len));
		}else
		{
			pkthdrPtr->ph_len = icmp_total_len;
			pkthdrPtr->ph_mbuf->m_len = icmp_total_len;
		}
	}

	/*
		set new icmp header
	*/
	ic = (struct icmp *)((uint32)ip + 20);	// new ip header length is 20 byte
	ic->icmp_type = icmp_info->ic_type;
	ic->icmp_code = icmp_info->ic_code;
	if (icmp_info->ic_mtu)
	{
		ic->icmp_pmvoid = 0;
		ic->icmp_nextmtu = htons(icmp_info->ic_mtu);
	}else
	{
		PKTSET_UINT32_UNALIGNED((uint32)0, &ic->icmp_void);
	}
	rtl8651_memcpy(&ic->icmp_ip, org_pkt, org_pkt_len);

	if (trailer_len)
	{
		memset((void*)((uint32)ip + (uint32)icmp_ip_len - (uint32)trailer_len), 0, trailer_len);
	}
	/*
		set new IP header
	*/
	ip->ip_vhl = 0x45;
	ip->ip_tos = 0;
	ip->ip_len = htons(icmp_ip_len);
	ip->ip_off = 0;
	ip->ip_ttl = 0x80;
	ip->ip_p = IPPROTO_ICMP;
	// exchange IP
	rtl8651_memcpy(&ip->ip_dst, &ip->ip_src, 4);			// set new dip
	PKTSET_UINT32_UNALIGNED(sip, &ip->ip_src);			// set new sip

	/*
		re-calculate L3/L4 checksum
	*/
	pkthdrPtr->ph_proto = PKTHDR_ICMP;					// let ASIC know how to do L4 checksum
	pkthdrPtr->ph_flags |= CSUM_L4;
	pkthdrPtr->ph_flags |= CSUM_IP;

	/*
		set new L2 header
	*/
	if( pkthdrPtr->ph_pppeTagged)
	{
		int8 tmpEther[14];
		/* backup ether header */
		rtl8651_memcpy(tmpEther, pkthdrPtr->ph_mbuf->m_data, 14);
		/* remove 8 bytes */
		mBuf_adjHead(pkthdrPtr->ph_mbuf, 8);
		/* copy backed-up ether header back */
		rtl8651_memcpy(pkthdrPtr->ph_mbuf->m_data, tmpEther, 14);
		/* change ethertype */
		PKTSET_UINT16_UNALIGNED(0x0800, (pkthdrPtr->ph_mbuf->m_data + 12));
		/* remove pppoe Tagged on packet header */
		pkthdrPtr->ph_pppeTagged = 0;
	}

	if (netIntf->linkLayerType == RTL8651_LL_PPPOE)		// packet from pppoe interface
	{
		/* pppoe header need to be added by ASIC */
		pkthdrPtr->ph_flags &= PKTHDR_PPPOE_AUTOADD;
	}

	// exchange mac
	macHdr = pkthdrPtr->ph_mbuf->m_data;
	rtl8651_memcpy(&macAddr, macHdr,6);			//Get source MAC
	rtl8651_memcpy(macHdr,macHdr+6,6);			//Construct destination MAC
	rtl8651_memcpy(macHdr+6,&macAddr,6);		//Construct source MAC

	pkthdrPtr->ph_portlist = (1<<pkthdrPtr->ph_portlist);
	retval=_rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdrPtr, vid,-1);

	if(retval==SUCCESS)
	{
		rtl8651_fwdEngineCounter.icmpReply++;
	}else
	{
		mBuf_driverFreeMbufChain(pkthdrPtr->ph_mbuf);
		rtl8651_fwdEngineCounter.icmpReplyFailed++;
	}
	return SUCCESS;
}

static inline int32 _rtl8651_drvIcmp(struct rtl_pktHdr *pkthdrPtr, struct ip* ip, struct icmp *icmp, uint16 vid) {
	rtl8651_tblDrv_networkIntfTable_t *netintf;
	ether_addr_t macAddr;
	ipaddr_t ipAddr;
	uint8 *cp = pkthdrPtr->ph_mbuf->m_data;
	int32 retval, tol_len;
	uint32 mtu;
	uint8 ttl;
	uint16 srcPortList;
	ipaddr_t sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);
	ipaddr_t dip = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);

	if (ntohs(ip->ip_off) & (IP_OFFMASK))
	{		
		if (_rtl8651_getFragCache(ICMP_REPLY_FRAG, sip, dip, ntohs(ip->ip_id), NULL, NULL, NULL, &(pkthdrPtr->ph_dsid)) == SUCCESS)
		{
			goto proc_icmp;
		}else
		{
			return FAILED;
		}
	}

	switch (icmp->icmp_type)
	{
		case ICMP_ECHO:		// driver level icmp ping responder
			break;
		default:
			return FAILED;	// don't support other icmp request now
	}

proc_icmp:

	// get network interface information
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
	if (tunnel.valid && vid == tunnel.loopBackVid && pkthdrPtr->ph_portlist == tunnel.loopbackPort) {
		netintf = _rtl8651_getNetworkIntfEntryByVid(tunnel.wanVid);
	}else
#endif
	{
		netintf = _rtl8651_getNetworkIntfEntryByVid(vid);
	}
	if ((netintf == NULL) || (netintf->valid == 0)/*|| (SLIST_FIRST(&netintf->IpHead) == NULL)*/)
	{
		return FAILED;
	}

	/*
		chenyl:
			We get mtu from network interface.
			For old TBLDRV model, mtu information store in netintf is the minimum mtu of all sessions, it's ok to work.
			For new TBLDRV model, mtu information in netintf is same as data in the corresponding sessions, it's ok.
		
	*/
	mtu = netintf->mtu;
	tol_len=mBuf_getPktlen(pkthdrPtr->ph_mbuf);

	// Exchange mac
	rtl8651_memcpy(&macAddr, cp,6);		//Get source MAC
	rtl8651_memcpy(cp,cp+6,6);			//Construct destination MAC
	rtl8651_memcpy(cp+6,&macAddr,6);	//Construct source MAC

	// Exchange IP
	rtl8651_memcpy(&ipAddr, &ip->ip_src, 4);		// get destination IP
	rtl8651_memcpy(&ip->ip_src, &ip->ip_dst, 4);	// construct new source IP
	rtl8651_memcpy(&ip->ip_dst, &ipAddr, 4);		// construct new destination IP

	ttl			= ip->ip_ttl;		/* store old ttl */
	ip->ip_ttl	= 0x80;			/* reset ttl */

	if (!(ntohs(ip->ip_off) & IP_OFFMASK))
	{
		icmp->icmp_type = ICMP_ECHOREPLY;	//*Change to ICMP Echo Reply */
	}

	pkthdrPtr->ph_flags |= (CSUM_IP|CSUM_L4);
	if ((ntohs(ip->ip_off) & (IP_MF|IP_OFFMASK)) || ((ip->ip_vhl & 0x0f) != 5))
	{
		if ((ip->ip_vhl & 0x0f) != 5)
		{	/* L3 checksum */
			pkthdrPtr->ph_flags &= ~CSUM_IP;
			_RTL8651_DATA_ADJUST_CHECKSUM((ip->ip_ttl << 8), (ttl << 8), ip->ip_sum);
		}
		if (!(ntohs(ip->ip_off) & IP_OFFMASK))
		{	/* L4 checksum if ip offset == 0*/
			_RTL8651_DATA_ADJUST_CHECKSUM((ICMP_ECHOREPLY<<8), (ICMP_ECHO<<8), icmp->icmp_cksum);
			if (ntohs(ip->ip_off) & IP_MF)
			{	/* first fragment packet: update fragment cache */
				_rtl8651_updateFragCache(ICMP_REPLY_FRAG, sip, dip, ntohs(ip->ip_id), (ipaddr_t)htonl(ipAddr), 0, NULL, pkthdrPtr->ph_dsid);
			}
		}
		pkthdrPtr->ph_flags &= ~CSUM_L4;
		pkthdrPtr->ph_proto = PKTHDR_IP;
	}

	pkthdrPtr->ph_flags |= CSUM_IP;

	srcPortList = pkthdrPtr->ph_portlist;
	pkthdrPtr->ph_portlist = (1<<pkthdrPtr->ph_portlist);

	if (tol_len > (mtu+14)) 
	{
		int8 old_IcmpRoutingMsgEnable = rtl8651_drvIcmpRoutingMsgEnable;
		rtl8651_drvIcmpRoutingMsgEnable = FALSE;
		retval = _rtl8651_mtuFragment(FWDENG_FROMDRV, pkthdrPtr, mtu, vid, FALSE, srcPortList);
		rtl8651_drvIcmpRoutingMsgEnable = old_IcmpRoutingMsgEnable;
	}else 
		retval=_rtl8651_fwdEngineSend(FWDENG_FROMDRV, (void*)pkthdrPtr, vid,-1);

	if(retval==SUCCESS)
	{
		rtl8651_fwdEngineCounter.icmpReply++;
		return SUCCESS;
	}
	else
	{
		rtl8651_fwdEngineCounter.icmpReplyFailed++;
		rtl8651_fwdEngineCounter.icmpDrop++;
		return FWDENG_DROP;
	}
}

int get_url_from_msg(char *msg,char *url)
{
	int i=0,len;
	char *cl,*c2;
	cl = &msg[12];
	c2 = url;
	if(cl==NULL)
	 return 0;
	len=(int)*cl;
	if(len == 0)
		return 0;
	while(len !=0)
	{
		cl++;
		for(i=0;i<len&&cl;i++,c2++,cl++)
			*c2=*cl;
		if(cl==NULL)
		  return 0;	
		len =(int) *cl;
		if(len !=0)
		{
				*c2='.';
		    c2++;
		}
		else
			break;		
	}
	*c2='\0';
	return 1;
}
/*
	Try to trigger demand route for some specific packets.
	=> Forwarding Engine would CONTINUE packet forwarding even the Demand Route is matched.
		=> Don't reply FWDENG_DEMANDRT_FOUND.
*/
static inline int32 _rtl8651_drvPktDemandRouteTrigger(struct ip *pip)
{
	if (rtl8651_drvPktDemandRouteTrigger != TRUE)
	{
		return SUCCESS;
	}

	{
		/*
			==> DNS request

				TCP / UDP , destination port == 53
		*/
		if (!IS_FRAGMENT_REMAIN(pip->ip_off))
		{
			switch (pip->ip_p)
			{
				case IPPROTO_TCP:
				{
					struct tcphdr *tc;
					ipaddr_t dip = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

					tc = (struct tcphdr *)L4HDR_ADDR(pip);

					if (ntohs(tc->th_dport) == 53)
					{
						_rtl8651_getDemandRoutingEntry(0, dip, IPPROTO_TCP, ntohs(tc->th_dport));
					}
				}
					break;
				case IPPROTO_UDP:
				{
					struct udphdr *ud;
					ipaddr_t dip = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

					ud = (struct udphdr *)L4HDR_ADDR(pip);

					if (ntohs(ud->uh_dport) == 53)
					{
						char *dnsmsg,tmpurl[128];
						dnsmsg=(char *)ud+sizeof(struct udphdr);
						if(dnsmsg!=NULL)
						  get_url_from_msg(dnsmsg,tmpurl);
						if(!(memcmp(tmpurl, "www.routerlogin.com", sizeof("www.routerlogin.com")) == 0
             ||memcmp(tmpurl, "www.routerlogin.net", sizeof("www.routerlogin.net")) == 0
             ||memcmp(tmpurl, "routerlogin.com", sizeof("routerlogin.com")) == 0
             ||memcmp(tmpurl, "routerlogin.net", sizeof("routerlogin.net")) == 0))
						_rtl8651_getDemandRoutingEntry(0, dip, IPPROTO_UDP, ntohs(ud->uh_dport));
					}
				}
			}
		}
	}

	return SUCCESS;
}

static int32 _rtl8651_drvIcmpAgent(struct rtl_pktHdr * pkthdrPtr, struct ip* ip, uint16 vid)
{

	if (_rtl8651_pktAccessGateway(pkthdrPtr->ph_sip, pkthdrPtr->ph_dip, vid) != TRUE)
	{
		return FAILED;
	}

	/*
		Try to reply ICMP back to sender.

		Note that fragment would be checked in _rtl8651_drvIcmp,
		so no any potential bug would be generated when we use
		L4HDR_ADDR (to get L4 header, might be wrong if this packet is not the first piece of fragment packet) before checking framgnetation issue */
	return _rtl8651_drvIcmp(pkthdrPtr, ip, (struct icmp *)L4HDR_ADDR(ip), vid);
}

int32 _rtl8651_ingressAcl(uint32 dsid, uint32 startIdx, rtl8651_tblDrv_vlanTable_t *pVlan, int32 vid, uint8 *m_data, uint8 *ip, uint32 matchType, uint32 optLayer){
	_rtl8651_tblDrvAclRule_t *acl_tb;
	rtl8651_tblDrv_networkIntfTable_t *netif_tb;
	int8 defPermit = TRUE;
	struct ip *pip=(struct ip*)ip;
	//rtlglue_printf("_rtl8651_ingressAcl(vid:%d)\n",vid);
	netif_tb = _rtl8651_getNetworkIntfEntryByVid(vid);
	if(netif_tb){
		acl_tb = _rtl8651_aclLookUp(dsid, netif_tb, TRUE, startIdx, (int8 *)m_data, (int8*)ip,&defPermit, matchType, optLayer);
		if(acl_tb){
			switch(acl_tb->actionType_){
				case RTL8651_ACL_PERMIT:
					rtl8651_fwdEngineCounter.inAclPermit++;
					return SUCCESS;
				case RTL8651_ACL_DROP:
					rtl8651_fwdEngineCounter.inAclDrop++;
					return FWDENG_ACL_IN_DROP|FWDENG_DROP;
				case RTL8651_ACL_DROP_LOG:
					rtl8651_fwdEngineCounter.inAclDrop++;
					if (  (_loggingModule[dsid]&RTL8651_LOGMODULE_ACL)
						&& _pUserLoggingFunc
						&& ip											/* we only log IP pkts */
#ifdef RTL8651_DSID_SUPPORT						
						&& (_rtl8651_acl_drop_nolog[dsid]==0)
#endif
					){
#ifdef _RTL_NEW_LOGGING_MODEL
						rtl8651_logInfo_t info;
						struct tcphdr *tc = NULL;
						struct udphdr *ud = NULL;
						struct icmp *ic = NULL;
						uint16 sport = 0xffff;
						uint16 dport = 0xffff;
						uint8 type = 0xff;
						uint16 id = 0xffff;

						if (!IS_FRAGMENT_REMAIN(pip->ip_off))
						{
							switch(pip->ip_p)
							{
								case IPPROTO_TCP:
									tc = (struct tcphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
									sport = ntohs(tc->th_sport);
									dport = ntohs(tc->th_dport);
									break;
								case IPPROTO_UDP:
									ud = (struct udphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
									sport = ntohs(ud->uh_sport);
									dport = ntohs(ud->uh_dport);
									break;
								case IPPROTO_ICMP:
									ic = (struct icmp *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
									type = ic->icmp_type;
									id = ntohs(ic->icmp_id);
									break;
							}
						}

						info.infoType		= RTL8651_LOG_INFO_PKT;
						info.pkt_dsid		= dsid;
						info.pkt_sip		= PKTGET_UINT32_UNALIGNED(&pip->ip_src);
						info.pkt_dip		= PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
						info.pkt_proto		= pip->ip_p;
						info.pkt_direction	= (pVlan->internal?RTL8651_LAN2WAN:RTL8651_WAN2LAN);
						info.pkt_sport		= sport;
						info.pkt_dport		= dport;
						info.pkt_icmpType	= type;
						info.pkt_icmpId	= id;
						info.action		= RTL8651_LOGACTION_DROP;
						(* _pUserLoggingFunc)(RTL8651_LOGMODULE_ACL, RTL8651_LOG_ACL_IngressAclDropLog, &info);
#else
						struct tcphdr * tc =NULL;
						//struct udphdr * ud=NULL;
						if (pip->ip_p==6||pip->ip_p==17) /* TCP/UDP */
							tc = (struct tcphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
						//else if (pip->ip_p==17) /* UDP */
						//	ud = (struct udphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
					  	(*_pUserLoggingFunc)(
					  		dsid,
						  	RTL8651_LOGMODULE_ACL,
						    (unsigned char)pip->ip_p, /* protocol */
						    (unsigned char)(pVlan->internal?RTL8651_LAN2WAN:RTL8651_WAN2LAN), /* direction */
							(unsigned long)PKTGET_UINT32_UNALIGNED(&pip->ip_src.s_addr),
							(unsigned long)PKTGET_UINT32_UNALIGNED(&pip->ip_dst.s_addr),
							(unsigned short)(tc?(ntohs(tc->th_sport)):0xffffffff),
							(unsigned short)(tc?(ntohs(tc->th_dport)):0xffffffff),
							0, /* type: only meaningful for ICMP */
							RTL8651_LOGACTION_DROP, /* action */
							"Ingress" /* msg */);
#endif
					}
					return (FWDENG_ACL_IN_DROP|FWDENG_DROP);
				case RTL8651_ACL_CPU:
					rtl8651_fwdEngineCounter.inAclCpu++;
					return FWDENG_ACL_IN_USERTRAP;
				case RTL8651_ACL_DROP_NOTIFY:
					rtl8651_fwdEngineCounter.inAclDrop++;
					return (FWDENG_ACL_DROP_NOTIFY|FWDENG_DROP);
				case RTL8651_ACL_REDIRECT:
				case RTL8651_ACL_REDIRECT_PPPOE:
					//cfliu: Must revise this. Should add code to handle redirect rule but trapped to CPU case.
					rtl8651_fwdEngineCounter.inAclDrop++;
					return FWDENG_ACL_IN_DROP|FWDENG_DROP;
				default:
					if(defPermit != TRUE){
						rtl8651_fwdEngineCounter.inAclDrop++;
						return FWDENG_ACL_IN_DROP|FWDENG_DROP;
					}else{
						rtl8651_fwdEngineCounter.inAclPermit++;
						return SUCCESS;
					}
			}	
		}
		#if 1
		else {
			/* no rule found, do default action */
			int8 defAction;
			int32 ret;
			ret=_rtl8651_getDefaultAcl(netif_tb->ifName,&defAction);
			assert(ret==SUCCESS);
			switch(defAction){
				case RTL8651_ACL_L34_DROP:
				case RTL8651_ACL_DROP:
					//rtlglue_printf("defAcl:drop\n");
					rtl8651_fwdEngineCounter.inAclDrop++;
					return FWDENG_ACL_IN_DROP|FWDENG_DROP;
				case RTL8651_ACL_DROP_LOG:
					//rtlglue_printf("defAcl:drop_log\n");
					rtl8651_fwdEngineCounter.inAclDrop++;
					if (
						(_loggingModule[dsid]&RTL8651_LOGMODULE_ACL)
						&&_pUserLoggingFunc
						&& ip											/* we only log IP pkts */
					)
					{
#ifdef _RTL_NEW_LOGGING_MODEL
						rtl8651_logInfo_t info;
						struct tcphdr *tc = NULL;
						struct udphdr *ud = NULL;
						struct icmp *ic = NULL;
						uint16 sport = 0xffff;
						uint16 dport = 0xffff;
						uint8 type = 0xff;
						uint16 id = 0xffff;

						if (!IS_FRAGMENT_REMAIN(pip->ip_off))
						{
							switch(pip->ip_p)
							{
								case IPPROTO_TCP:
									tc = (struct tcphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
									sport = ntohs(tc->th_sport);
									dport = ntohs(tc->th_dport);
									break;
								case IPPROTO_UDP:
									ud = (struct udphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
									sport = ntohs(ud->uh_sport);
									dport = ntohs(ud->uh_dport);
									break;
								case IPPROTO_ICMP:
									ic = (struct icmp *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
									type = ic->icmp_type;
									id = ntohs(ic->icmp_id);
									break;
							}
						}

						info.infoType		= RTL8651_LOG_INFO_PKT;
						info.pkt_dsid		= dsid;
						info.pkt_sip		= PKTGET_UINT32_UNALIGNED(&pip->ip_src);
						info.pkt_dip		= PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
						info.pkt_proto		= pip->ip_p;
						info.pkt_direction	= (pVlan->internal?RTL8651_LAN2WAN:RTL8651_WAN2LAN);
						info.pkt_sport		= sport;
						info.pkt_dport		= dport;
						info.pkt_icmpType	= type;
						info.pkt_icmpId	= id;
						info.action		= RTL8651_LOGACTION_DROP;
						(* _pUserLoggingFunc)(RTL8651_LOGMODULE_ACL, RTL8651_LOG_ACL_IngressAclDropLog, &info);
#else
						struct tcphdr * tc =NULL;
						//struct udphdr * ud=NULL;
						if (pip->ip_p==6||pip->ip_p==17) /* TCP/UDP */
							tc = (struct tcphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));						
						(*_pUserLoggingFunc)(
							dsid,
							RTL8651_LOGMODULE_ACL,
						    (unsigned char)pip->ip_p, /* protocol */
						    (unsigned char)(pVlan->internal?RTL8651_LAN2WAN:RTL8651_WAN2LAN), /* direction */
							(unsigned long)PKTGET_UINT32_UNALIGNED(&pip->ip_src.s_addr),
							(unsigned long)PKTGET_UINT32_UNALIGNED(&pip->ip_dst.s_addr),
							(unsigned short)(tc?(ntohs(tc->th_sport)):0xffffffff),
							(unsigned short)(tc?(ntohs(tc->th_dport)):0xffffffff),
							0, /* type: only meaningful for ICMP */
							RTL8651_LOGACTION_DROP, /* action */
							"Ingress" /* msg */);
#endif
					}					
					return FWDENG_ACL_IN_DROP|FWDENG_DROP;
				case RTL8651_ACL_PERMIT:
					//rtlglue_printf("defAcl:permit\n");
					rtl8651_fwdEngineCounter.inAclPermit++;
					return SUCCESS;					
				case RTL8651_ACL_CPU:
					//rtlglue_printf("defAcl***************:cpu\n");
					rtl8651_fwdEngineCounter.inAclPermit++;
					return SUCCESS;
				default:
					assert(0);
					break;
			}
		}
		#endif
	}
	return SUCCESS;
}

int32 _rtl8651_egressAcl(uint32 dsid,rtl8651_tblDrv_vlanTable_t *pVlan,int32 vid, uint8 *ethHdr, uint8 *ip){
	_rtl8651_tblDrvAclRule_t *acl_tb;
	rtl8651_tblDrv_networkIntfTable_t *netif_tb;
	int8 defPermit = TRUE;
	struct ip *pip=(struct ip*)ip;	
	netif_tb = _rtl8651_getNetworkIntfEntryByVid(vid);

	if(netif_tb){
		acl_tb = _rtl8651_aclLookUp(dsid,netif_tb,FALSE,0,(int8 *)ethHdr,(int8*)ip,&defPermit, 0, _RTL8651_OPERATION_OTHER);

		//if (ip != NULL && acl_tb) {
		//	rtlglue_printf("egress rule match.....\n");
		//}

		if(acl_tb) { /* replaced by next line 2004/1/19 orlando */
			switch(acl_tb->actionType_){
			case RTL8651_ACL_PERMIT:
				rtl8651_fwdEngineCounter.outAclPermit++;
				return SUCCESS;
			case RTL8651_ACL_DROP:
				rtl8651_fwdEngineCounter.outAclDrop++;
				return FWDENG_ACL_OUT_DROP|FWDENG_DROP;
			case RTL8651_ACL_DROP_LOG:
				rtl8651_fwdEngineCounter.outAclDrop++;
				//rtlglue_printf("_rtl8651_acl_drop_nolog[%d]:%d\n",dsid,_rtl8651_acl_drop_nolog[dsid]);
				if (  (_loggingModule[dsid]&RTL8651_LOGMODULE_ACL)
					&& _pUserLoggingFunc
					&& ip											/* we only log IP pkts */
#ifdef RTL8651_DSID_SUPPORT
					&& (_rtl8651_acl_drop_nolog[dsid]==0) 
#endif					
				){
#ifdef _RTL_NEW_LOGGING_MODEL
					rtl8651_logInfo_t info;
					struct tcphdr *tc = NULL;
					struct udphdr *ud = NULL;
					struct icmp *ic = NULL;
					uint16 sport = 0xffff;
					uint16 dport = 0xffff;
					uint8 type = 0xff;
					uint16 id = 0xffff;

					if (!IS_FRAGMENT_REMAIN(pip->ip_off))
					{
						switch(pip->ip_p)
						{
							case IPPROTO_TCP:
								tc = (struct tcphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
								sport = ntohs(tc->th_sport);
								dport = ntohs(tc->th_dport);
								break;
							case IPPROTO_UDP:
								ud = (struct udphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
								sport = ntohs(ud->uh_sport);
								dport = ntohs(ud->uh_dport);
								break;
							case IPPROTO_ICMP:
								ic = (struct icmp *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
								type = ic->icmp_type;
								id = ntohs(ic->icmp_id);
								break;
						}
					}

					info.infoType		= RTL8651_LOG_INFO_PKT;
					info.pkt_dsid		= dsid;
					info.pkt_sip		= PKTGET_UINT32_UNALIGNED(&pip->ip_src);
					info.pkt_dip		= PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
					info.pkt_proto		= pip->ip_p;
					info.pkt_direction	= (pVlan->internal?RTL8651_LAN2WAN:RTL8651_WAN2LAN);
					info.pkt_sport		= sport;
					info.pkt_dport		= dport;
					info.pkt_icmpType	= type;
					info.pkt_icmpId	= id;
					info.action		= RTL8651_LOGACTION_DROP;
					(* _pUserLoggingFunc)(RTL8651_LOGMODULE_ACL, RTL8651_LOG_ACL_EgressAclDropLog, &info);
#else
					struct tcphdr * tc =NULL;
					if (pip->ip_p==6||pip->ip_p==17) /* TCP/UDP */
						tc = (struct tcphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
					(*_pUserLoggingFunc)(
						dsid,
						RTL8651_LOGMODULE_ACL,
					    (unsigned char)pip->ip_p, /* protocol */
					    (unsigned char)(pVlan->internal?RTL8651_LAN2WAN:RTL8651_WAN2LAN), /* direction */
						(unsigned long)PKTGET_UINT32_UNALIGNED(&pip->ip_src.s_addr),
						(unsigned long)PKTGET_UINT32_UNALIGNED(&pip->ip_dst.s_addr),
						(unsigned short)(tc?(ntohs(tc->th_sport)):0xffffffff),
						(unsigned short)(tc?(ntohs(tc->th_dport)):0xffffffff),
						0, /* type: only meaningful for ICMP */
						RTL8651_LOGACTION_DROP, /* action */
						"Egress" /* msg */);
#endif
				}
				return (FWDENG_ACL_OUT_DROP|FWDENG_DROP);
			//case RTL8651_ACL_CPU:
			//	rtl8651_fwdEngineCounter.outAclCpu++;
			//	return FAILED;
			//case RTL8651_ACL_REDIRECT:
			//case RTL8651_ACL_REDIRECT_PPPOE:
			//	rtl8651_fwdEngineCounter.outAclDrop++;
			//	goto end_drop;			
			default:
				if(defPermit != TRUE){
					rtl8651_fwdEngineCounter.outAclDrop++;
					return FWDENG_ACL_OUT_DROP|FWDENG_DROP;
				}else{
					rtl8651_fwdEngineCounter.outAclPermit++;
					return SUCCESS;
				}
			}
		}
		else {
			/* no rule found, do default action */
			int8 defAction;
			int32 ret;
			ret=_rtl8651_getDefaultAcl(netif_tb->ifName,&defAction);
			assert(ret==SUCCESS);
			switch(defAction){
				case RTL8651_ACL_L34_DROP:
				case RTL8651_ACL_DROP:
					rtl8651_fwdEngineCounter.outAclDrop++;
					return FWDENG_ACL_OUT_DROP|FWDENG_DROP;
				case RTL8651_ACL_DROP_LOG:
					rtl8651_fwdEngineCounter.outAclDrop++;
					if (	(_loggingModule[dsid]&RTL8651_LOGMODULE_ACL)
						&& ip											/* we only log IP pkts */
						&&_pUserLoggingFunc
					)
					{
#ifdef _RTL_NEW_LOGGING_MODEL
						rtl8651_logInfo_t info;
						struct tcphdr *tc = NULL;
						struct udphdr *ud = NULL;
						struct icmp *ic = NULL;
						uint16 sport = 0xffff;
						uint16 dport = 0xffff;
						uint8 type = 0xff;
						uint16 id = 0xffff;

						if (!IS_FRAGMENT_REMAIN(pip->ip_off))
						{
							switch(pip->ip_p)
							{
								case IPPROTO_TCP:
									tc = (struct tcphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
									sport = ntohs(tc->th_sport);
									dport = ntohs(tc->th_dport);
									break;
								case IPPROTO_UDP:
									ud = (struct udphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
									sport = ntohs(ud->uh_sport);
									dport = ntohs(ud->uh_dport);
									break;
								case IPPROTO_ICMP:
									ic = (struct icmp *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
									type = ic->icmp_type;
									id = ntohs(ic->icmp_id);
									break;
							}
						}

						info.infoType		= RTL8651_LOG_INFO_PKT;
						info.pkt_dsid		= dsid;
						info.pkt_sip		= PKTGET_UINT32_UNALIGNED(&pip->ip_src);
						info.pkt_dip		= PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
						info.pkt_proto		= pip->ip_p;
						info.pkt_direction	= (pVlan->internal?RTL8651_LAN2WAN:RTL8651_WAN2LAN);
						info.pkt_sport		= sport;
						info.pkt_dport		= dport;
						info.pkt_icmpType	= type;
						info.pkt_icmpId	= id;
						info.action		= RTL8651_LOGACTION_DROP;
						(* _pUserLoggingFunc)(RTL8651_LOGMODULE_ACL, RTL8651_LOG_ACL_EgressAclDropLog, &info);
#else
						struct tcphdr * tc =NULL;
						//struct udphdr * ud=NULL;
						if (pip->ip_p==6||pip->ip_p==17) /* TCP/UDP */
							tc = (struct tcphdr *) ((int8 *) ip + ((*(uint8*)ip&0xf) << 2));
						(*_pUserLoggingFunc)(
							dsid,
							RTL8651_LOGMODULE_ACL,
						    (unsigned char)pip->ip_p, /* protocol */
						    (unsigned char)(pVlan->internal?RTL8651_LAN2WAN:RTL8651_WAN2LAN), /* direction */
							(unsigned long)PKTGET_UINT32_UNALIGNED(&pip->ip_src.s_addr),
							(unsigned long)PKTGET_UINT32_UNALIGNED(&pip->ip_dst.s_addr),
							(unsigned short)(tc?(ntohs(tc->th_sport)):0xffffffff),
							(unsigned short)(tc?(ntohs(tc->th_dport)):0xffffffff),
							0, /* type: only meaningful for ICMP */
							RTL8651_LOGACTION_DROP, /* action */
							"Egress" /* msg */);
#endif
					}					
					return FWDENG_ACL_OUT_DROP|FWDENG_DROP;
				case RTL8651_ACL_PERMIT:
				case RTL8651_ACL_CPU:
					rtl8651_fwdEngineCounter.outAclPermit++;
					return SUCCESS;
				default:
					assert(0);
					break;
			}
		}
	}
	return SUCCESS;
}

/*********************************************************************************
 * Routine Name :	_rtl8651_l34Decision
* Description :	decision table lookup
* Input :			pkthdr:		packet header data structure
				vlanp:		ingress vlan
				decision_p:	store the forward decision
				srcType_p:	store the soruce IP type
				dstType_p:	store the destination IP type
				optLayer_p:	store the operation layer of this packet
**********************************************************************************/
static inline void _rtl8651_l34Decision(	struct rtl_pktHdr *pkthdr,
										rtl8651_tblDrv_vlanTable_t *vlanp,
										uint32 *decision_p,
										uint32 *srcType_p,
										uint32 *dstType_p,
										uint32 *optLayer_p)
{
	ipaddr_t sip = pkthdr->ph_sip;
	ipaddr_t dip = pkthdr->ph_dip;

	assert(pkthdr && vlanp && decision_p && srcType_p && dstType_p && optLayer_p);

	/****************************** Determine Source Type ******************************/
	if (_rtl8651_localServerIpAddr(sip) == TRUE)
	{
		*srcType_p = _RTL8651_HOST_LP;
	} else
	{
	if (vlanp->internal)
		{	/* this packet is from Internal VLAN */
		if(_rtl8651_getNatInternalMappingEntry(sip, pkthdr))
		{
			*srcType_p = _RTL8651_HOST_NI;
		}else
		{
			*srcType_p = _RTL8651_HOST_NPI;
		}
	}else
		{	/* this packet is from External VLAN */
			*srcType_p = _RTL8651_HOST_RP;
		}
	}
	/****************************** Determine Destination Type ******************************/
	if (_rtl8651_getNatExternalMappingEntry(dip, pkthdr))
	{	/* nat external */
		*dstType_p = _RTL8651_HOST_NE;
	}else if (_rtl8651_localServerIpAddr(dip) == TRUE)
	{	/* local public */
		*dstType_p = _RTL8651_HOST_LP;
	}else if (_rtl8651_extIpAddr(dip, NULL) == TRUE)
	{	/* napt external */
		*dstType_p = _RTL8651_HOST_NPE;
	}else
	{	/* Now, the candidates of destination type are NI/NPI/RP */
		rtl8651_tblDrv_routeTable_t *rt = NULL;

		rt = _rtl8651_getRoutingEntry(dip);

		if (	(rt == NULL)
			|| (rt && (rt->ipAddr == 0))
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
			|| (rt->netif_t->linkLayerType == RTL8651_LL_PPTP)
			|| (rt->netif_t->linkLayerType == RTL8651_LL_L2TP) 
#endif
		)
		{
			/* Normal Routing Entry lookup miss : destination type must be Remote Public */
			*dstType_p = _RTL8651_HOST_RP;

			/*

				* <=========== DONT STORE ROUTING INFORMATION ==========> *

				In this case, advanced routing find in _rtl8651_l34PacketProcess() has higher prirority than normal routing.
				So we don't store the routing table lookup result here.
			*/

		} else
		{	/* determine destination type by matched routing entry */
			rtl8651_tblDrv_vlanTable_t *dst_vlanp;

			dst_vlanp = &DrvTbl.vlan[rt->dvid & (RTL8651_VLAN_NUMBER-1)];

			if (dst_vlanp->internal)
			{
				if (_rtl8651_getNatInternalMappingEntry(dip, pkthdr) != NULL)
				{	/* nat internal */
					*dstType_p = _RTL8651_HOST_NI;
				}else
				{	/* napt internal */
					*dstType_p = _RTL8651_HOST_NPI;
				}
			}else
			{	/* remote public */
				*dstType_p = _RTL8651_HOST_RP;
			}

			/* Normal routing match successful : store routing result first if we've got it */
			pkthdr->ph_routeIdx = (rt - &DrvTbl.route[0]) + 1;
		}
	}

	/****************************** Decision Table Lookup ******************************/
	assert( ((*dstType_p) < _RTL8651_HOST_MAX) && ((*srcType_p) < _RTL8651_HOST_MAX) );

	*decision_p = _RTL8651_PROC[*srcType_p][*dstType_p];
	*optLayer_p = _RTL8651_OPTLAYER(*decision_p);

	return;
}


/*********************************************************************************
 * Routine Name :	_rtl8651_l34PacketProcess
* Description :	Do policy routing / domain routing .... if its needed
* Input :			pkthdr:	packet header data structure
				l4hdr:	layer4 header
**********************************************************************************/
static inline int32 _rtl8651_l34PacketProcess(	struct rtl_pktHdr *pkthdr,
												struct ip *pip,
 												uint32 decision)
{
	ipaddr_t sip = pkthdr->ph_sip;
	ipaddr_t dip = pkthdr->ph_dip;

	if (pkthdr->ph_routeIdx == 0	/* The route entry is not be found now */ )
	{
		rtl8651_tblDrv_routeTable_t *policy_rt = NULL;
		rtl8651_tblDrv_routeTable_t *rt = NULL;

		/* for outbound NAPT, some additional policy might need to be applied */
		if (decision == _RTL8651_DEC_NPT)
		{
			if ( !IS_FRAGMENT_REMAIN(pip->ip_off) )
			{

				/* ================================================
									Policy route / demand route.
				      =============================================== */

				switch (pip->ip_p)
				{
					case IPPROTO_TCP:
					{
						struct tcphdr *tc;

						tc = (struct tcphdr *)L4HDR_ADDR(pip);

						if ((policy_rt = _rtl8651_getSoftRoutingEntry(sip, dip, IPPROTO_TCP, ntohs(tc->th_dport))) == NULL)
						{
							if (_rtl8651_getDemandRoutingEntry(sip, dip, IPPROTO_TCP, ntohs(tc->th_dport)))
							{
								return FAILED;
							}
						}
					}
						break;

					case IPPROTO_UDP:
					{
						struct udphdr *ud;

						ud = (struct udphdr *)L4HDR_ADDR(pip);

						if ((policy_rt = _rtl8651_getSoftRoutingEntry(sip, dip, IPPROTO_UDP, ntohs(ud->uh_dport))) == NULL)
						{
							if (_rtl8651_getDemandRoutingEntry(sip, dip, IPPROTO_UDP, ntohs(ud->uh_dport)))
							{
								return FAILED;
							}
						}
					}
						break;
					default:

						if ((policy_rt = _rtl8651_getSoftRoutingEntry(sip, dip, pip->ip_p, 0)) == NULL)
						{
							if (_rtl8651_getDemandRoutingEntry(sip, dip, pip->ip_p, 0))
							{
								return FAILED;
							}
						}
						break;
				}
			}
		} 

		/*
			If the policy route is not matched,
			and the forwarding process is not NAPT-R, NAT-R, NAPT-PortBouncing, NAT-PortBouncing, Drop:

			This means the packet is need to be processed, and its actual destination IP is known before doing
			L4 packet translation.

			==> We do routing table lookup here.
		*/

		if (	(policy_rt == NULL) &&
			(decision != _RTL8651_DEC_NTR) &&
			(decision != _RTL8651_DEC_NPTR) &&
			(decision != _RTL8651_DEC_NTBC_NP) &&
			(decision != _RTL8651_DEC_NTBC_N) &&
			(decision != _RTL8651_DEC_NTBC_L) &&
			(decision != _RTL8651_DEC_NPTBC_NP) &&
			(decision != _RTL8651_DEC_NPTBC_N) &&
			(decision != _RTL8651_DEC_NPTBC_L) &&
			(decision != _RTL8651_DEC_DP)	)
		{
			rt = _rtl8651_getRoutingEntry(dip);

			if (rt == NULL)
			{
				return FAILED;
			}
		}

		if ( policy_rt )
		{
			pkthdr->ph_routeIdx = (policy_rt - &DrvTbl.route[0]) + 1;
		}else if ( rt )
		{
			pkthdr->ph_routeIdx = (rt - &DrvTbl.route[0]) + 1;
		}

	}
 	return SUCCESS;
}

/* switch action */
static inline int32 _rtl8651_l34ActExec (	struct rtl_pktHdr * pkthdrPtr,  /* packet header */	
										struct ip* ip, /* pointer to ip header */
										uint8 proto,  /* ip protocol */
										int32 decision /*decision table result */	)
{
	int32 val;

	if ((val = _rtl8651_l34AddtionalExec(decision, pkthdrPtr->ph_dip, pkthdrPtr->ph_mbuf->m_data)) != SUCCESS)
	{
		return val;
	}

	switch ( decision )
	{
		/* =================================================================
											NAPT
		     ================================================================= */
		case _RTL8651_DEC_NPT:	/* L4 NAPT - outbound */
		{
			val = _rtl8651_l4napt_process(pkthdrPtr, ip, proto, _RTL8651_L4_NAT_OUTPUT);

			switch (val)
			{
				 case _RTL8651_NAT_PASS:
				 {
					if ((val = _rtl8651_l3ActRouting(pkthdrPtr,pkthdrPtr->ph_dip, ip)) == SUCCESS)
					{
						return SUCCESS;
					}
					/* val != SUCCESS : same as _RTL8651_NAT_DROP */
				 }
				case _RTL8651_NAT_DROP:
				{
					rtl8651_fwdEngineCounter.natdrop++;
					return FWDENG_DROP;
				}
				case _RTL8651_NAT_IGNORE:
				{
					return FAILED;
				}
			}

			/* other return values, it's for forwarding engine or protocol stack, return it */
			return val;
		}

		case _RTL8651_DEC_NPTR:	/* L4 NAPT - inbound */
		{
			val = _rtl8651_l4napt_process(pkthdrPtr, ip, proto, _RTL8651_L4_NAT_INPUT);

			switch (val)
			{
				 case _RTL8651_NAT_PASS:
				 {
				 	/* re-get dip and do routing again */
					ipaddr_t dip;

					dip = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);
					pkthdrPtr->ph_routeIdx = 0;					/* flush route result. must do it in "l3ActRouting". */

					if ((val = _rtl8651_l3ActRouting(pkthdrPtr, dip, ip)) == SUCCESS)
					{
						return SUCCESS;
					}
					/* val != SUCCESS : same as _RTL8651_NAT_DROP */
				 }
				case _RTL8651_NAT_DROP:
				{
					rtl8651_fwdEngineCounter.natdrop++;
					return FWDENG_DROP;
				}
				case _RTL8651_NAT_IGNORE:
				{
					return FAILED;
				}
			}

			/* other return values, it's for forwarding engine or protocol stack, return it */
			return val;
		}

		/* =================================================================
											NAT
		     ================================================================= */
		case _RTL8651_DEC_NT:	/* L4 NAT - outbound */
			return _rtl8651_l4nat_out_process(pkthdrPtr, ip);
		case _RTL8651_DEC_NTR:	/* L4 NAT - inbound */
			return _rtl8651_l4nat_in_process(pkthdrPtr, ip);

		/* =================================================================
											Routing
		     ================================================================= */
		case _RTL8651_DEC_RT_LR:	/* L3 Routing -- Local Public to Remote Host */
		case _RTL8651_DEC_RT_RL:	/* L3 Routing -- Remote host to Local Public */
		case _RTL8651_DEC_RT_LNP:	/* L3 Routing -- Local Public to NAPT Internal */
		case _RTL8651_DEC_RT_NPL:	/* L3 Routing -- NAPT Internal to Local Public */
		case _RTL8651_DEC_RT_LN:	/* L3 Routing -- Local Public to NAT Internal */
		case _RTL8651_DEC_RT_NL:	/* L3 Routing -- NAT Internal to Local Public */
		case _RTL8651_DEC_RT_LL:	/* L3 Routing -- Local Public to Local Public */
		case _RTL8651_DEC_RT_NN:	/* L3 Routing -- NAT Internal to NAT Internal */
		case _RTL8651_DEC_RT_NPNP:	/* L3 Routing -- NAPT Internal to NAPT Internal */
		case _RTL8651_DEC_RT_NPN:	/* L3 Routing -- NAPT Internal to NAT Internal */
		case _RTL8651_DEC_RT_NNP:	/* L3 Routing -- NAT Internal to NAPT Internal */
		case _RTL8651_DEC_RT_RR:	/* L3 Routing -- Remote Public to Remote Public (Wan Side routing) */

			return _rtl8651_l3ActRouting(pkthdrPtr, pkthdrPtr->ph_dip, ip);

		/* =================================================================
											Port bouncing
		     ================================================================= */

		case _RTL8651_DEC_NPTBC_NP:		/* port bouncing -- NAPT Internal to NPE */
		case _RTL8651_DEC_NPTBC_N:		/* port bouncing -- NAT Internal to NPE */
		case _RTL8651_DEC_NPTBC_L:		/* port bouncing -- Local Public to NPE */
			return _rtl8651_l4NaptPortBouncing(pkthdrPtr, ip);
		case _RTL8651_DEC_NTBC_NP:		/* port bouncing -- NAPT Internal to NE */
		case _RTL8651_DEC_NTBC_N:		/* port bouncing -- NAT Internal to NE */
		case _RTL8651_DEC_NTBC_L:		/* port bouncing -- Local Public to NE */
			return _rtl8651_l4NatPortBouncing(pkthdrPtr, ip);

		/* =================================================================
											Other
		     ================================================================= */

		case _RTL8651_DEC_DP:	/* Drop */
			return FWDENG_DECISION_DROP;

		default:
			assert(0);
			/* return FAILED */ 
	}

	return FAILED;  
}

int32 _rtl8651_mtuFragment(uint32 property, struct rtl_pktHdr *pkthdrPtr, uint32 mtu, uint16 dvid, int8 isNapt, uint16 srcPortList)
{
	struct ip *ip = (struct ip *)(pkthdrPtr->ph_mbuf->m_data + pkthdrPtr->ph_iphdrOffset);
	struct rtl_pktHdr *orgPkthdr=NULL;
	uint32 local_ph_reason;
	int32 retval;

	/* Assumption: packet has original source IP address and source Port/icmp ID */
#ifdef _RTL_IP_FRAG


	local_ph_reason = 0x0064;

	switch (local_ph_reason & 0x007f) {
		struct rtl_mBuf *new_m;
		struct ip *new_ip;
		int32 offset;
		int32 ip_hlen;
		uint8 *cp1;
		uint32 payloadLen, hdrLen;
		uint32 segLen;
		uint32 seg_seq;
	case 0x0064: /* unicast over MTU */		
		/* get ip header length */
		ip_hlen = (ip->ip_vhl &0xf) << 2;

		/* re-generate ip/tcp/udp/icmp checksum */
		if (IS_FRAGMENT(ip->ip_off))
			pkthdrPtr->ph_proto = PKTHDR_IP;

		switch(pkthdrPtr->ph_proto){
//			uint8 *temp_buf;
			struct tcphdr *tcp_p;
			struct udphdr*udp_p;
			struct icmp *icmp_p;
			int32 totalLen ;
		case PKTHDR_TCP:
			totalLen = ntohs(ip->ip_len);
			tcp_p = (struct tcphdr *) ((int8 *) ip + ip_hlen);
			tcp_p->th_sum = 0;
			ip->ip_sum = 0;
			tcp_p->th_sum = htons(_rtl8651_tcpChecksum(ip));
			ip->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)ip));
			break;
		case PKTHDR_UDP:
			udp_p = (struct udphdr *) ((int8 *) ip + ip_hlen);
			totalLen = ntohs(ip->ip_len);
			ip->ip_sum = 0;
			if (udp_p->uh_sum)
			{	/* don't re-calculate checksum if udp checksum is zero */
				udp_p->uh_sum = 0;
				udp_p->uh_sum = htons(_rtl8651_tcpChecksum(ip));
			}
			ip->ip_sum = htons(_rtl8651_ipChecksum(ip));
			break;
		case PKTHDR_ICMP:	
			totalLen = ntohs(ip->ip_len);
			icmp_p = (struct icmp *) ((int8 *) ip + ip_hlen);			
			icmp_p->icmp_cksum = 0;
			ip->ip_sum = 0;		
			icmp_p->icmp_cksum = htons(_rtl8651_icmpChecksum(ip));
			ip->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)ip));
			break;
#if 0	/* if ip header length is not 20, we can say that the ip checksum is done by previous process */
		case PKTHDR_IP:
			totalLen = ntohs(ip->ip_len);
			if (!IPHDRLEN_IS_TWENTY(ip))
			{
				ip->ip_sum = 0;
				ip->ip_sum = htons(_rtl8651_ipChecksum(ip));
			}
			break;
#endif
		default:
			break;
		}


		/* Original Packet for callback */
		if ((rtl8651_drvIcmpRoutingMsgEnable == TRUE) &&
			(property == FWDENG_FROMDRV) /* only generate icmp routing message when packet is processed by normal forwarding */ &&
			(rtl8651_drvIcmpAlwaysReplyErrForFrag == TRUE || DONT_FRAG(ip->ip_off))) {
			struct ip *iphdr;
			rtl8651_tblDrv_vlanTable_t *vlan_t;
			rtl8651_tblDrv_arpEntry_t *arp_t;
//			ether_addr_t src_mac;
			uint32 ip_hlen, totalLen, ipAddr;
//			uint8 *temp_buf;

			if ((new_m=mBuf_dupPacket(pkthdrPtr->ph_mbuf, MBUF_DONTWAIT)) == NULL) 
				return FAILED;
			orgPkthdr = new_m->m_pkthdr;

			/* flush iphdr setting:
					FIXME:	we assume the sender of this fragmented packet is via VLAN-type interface,
							so ICMP error message will NEVER add pppoe header now. */
			orgPkthdr->ph_flags &= ~(PKTHDR_PPPOE_AUTOADD|PKTHDR_VLAN_AUTOADD|CSUM_IP|CSUM_L4);

			iphdr = (struct ip *)(new_m->m_data + orgPkthdr->ph_iphdrOffset);
			iphdr->ip_ttl ++;
			if (isNapt == TRUE) {
				PKTSET_UINT32_UNALIGNED(orgPkthdr->ph_sip, &iphdr->ip_src);
			}
			else {
				rtl8651_memcpy(&ipAddr, &iphdr->ip_src, 4);		
				rtl8651_memcpy(&iphdr->ip_src, &iphdr->ip_dst, 4);	
				rtl8651_memcpy(&iphdr->ip_dst, &ipAddr, 4);			
			}
			ip_hlen = (iphdr->ip_vhl &0xf) << 2;
			totalLen = ntohs(iphdr->ip_len);
			
			/* original ether header */
			if ((arp_t=_rtl8651_getArpEntry(NULL, orgPkthdr->ph_sip, FALSE))) {
				vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(orgPkthdr->ph_vlanIdx)];
				rtl8651_memcpy((new_m->m_data+6), &arp_t->macInfo->macAddr, 6);
				rtl8651_memcpy(new_m->m_data, &vlan_t->macAddr, 6);
			}
			
			switch(orgPkthdr->ph_proto){
				struct tcphdr *tcp_p;
				struct udphdr*udp_p;
				struct icmp *icmp_p;
			case PKTHDR_TCP:
				tcp_p = (struct tcphdr *) ((int8 *) iphdr + ip_hlen);
				tcp_p->th_sport = htons(orgPkthdr->ph_sport);
				tcp_p->th_sum = 0;
				iphdr->ip_sum = 0;
				tcp_p->th_sum = htons(_rtl8651_tcpChecksum((struct ip*)iphdr));
				iphdr->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)iphdr));
				break;
				
			case PKTHDR_UDP:
				udp_p = (struct udphdr *) ((int8 *) iphdr + ip_hlen);
				udp_p->uh_sport = htons(orgPkthdr->ph_sport);
				iphdr->ip_sum = 0;
				if (udp_p->uh_sum)
				{
					udp_p->uh_sum = 0;
					udp_p->uh_sum= htons(_rtl8651_tcpChecksum((struct ip*)iphdr));
				}
				iphdr->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)iphdr));			
				break;
				
			case PKTHDR_ICMP:
				icmp_p = (struct icmp *) ((int8 *) iphdr + ip_hlen);
				icmp_p->icmp_id = htons(orgPkthdr->ph_sid);
				icmp_p->icmp_cksum = 0;
				iphdr->ip_sum = 0;		
				icmp_p->icmp_cksum = htons(_rtl8651_icmpChecksum((struct ip*)iphdr));
				iphdr->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)iphdr));
				break;

			case PKTHDR_IP:
				if (!IPHDRLEN_IS_TWENTY(iphdr))
				{
					iphdr->ip_sum = 0;
					iphdr->ip_sum = htons(_rtl8651_ipChecksum((struct ip*)iphdr));
				}
			default: break;

			}
			/* run callback function */
			if (p_mtuHandler)
			{
				p_mtuHandler(orgPkthdr);
			}else
			{
				_rtl8651_IcmpPktField_t icmpInfo;
				icmpInfo.ic_type	= ICMP_UNREACH;
				icmpInfo.ic_code	= ICMP_UNREACH_NEEDFRAG;
				icmpInfo.ic_mtu	= mtu;
				vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(orgPkthdr->ph_vlanIdx)];
				orgPkthdr->ph_portlist = srcPortList;
				if (_rtl8651_drvIcmpErrorGeneration(orgPkthdr, iphdr, &icmpInfo, vlan_t->vid) != SUCCESS)
				{
					mBuf_freeMbufChain(orgPkthdr->ph_mbuf);
				}
			}
		}
		
		if (DONT_FRAG(ip->ip_off))
			return FAILED;
		
		/* get ip pointer offset */
		offset = (uint8 *)ip - (uint8 *)pkthdrPtr->ph_mbuf->m_data;
		if(offset <= 0 ){
#if 0
			if (orgPkthdr)
				mBuf_freeMbufChain(orgPkthdr->ph_mbuf);
#endif
			return FAILED;
		}

		hdrLen = offset + ip_hlen;			/* ether header + ip header length */
		payloadLen = ntohs(ip->ip_len) - ip_hlen;		/* payload length */
		segLen = ((mtu - ip_hlen)>>3)<<3; 	/* MTU not include ether header: 8-byte alignment */
		
		for(seg_seq=0 ; payloadLen > segLen; payloadLen-=segLen, seg_seq++) {
			if ((new_m=mBuf_dupPacket(pkthdrPtr->ph_mbuf, MBUF_DONTWAIT)) == NULL) {
				//rtlglue_printf("enter %s %d\n",__FUNCTION__,__LINE__);
#if 0
				if (orgPkthdr)
					mBuf_freeMbufChain(orgPkthdr->ph_mbuf);
#endif
				return FAILED;
			}
			new_m->m_pkthdr->ph_flags &= ~(CSUM_L4);
			if (ip_hlen == 20)
				new_m->m_pkthdr->ph_flags |= (CSUM_IP);
			else
				new_m->m_pkthdr->ph_flags &= ~(CSUM_IP);
			if (seg_seq > 0) {
				cp1 = (uint8 *)new_m->m_data + (seg_seq*segLen);
				rtl8651_memcpy(cp1, (uint8 *)new_m->m_data, hdrLen);
				mBuf_adjHead(new_m,  segLen*seg_seq);
			}
			mBuf_adjTail(new_m,  payloadLen-segLen);
			/* get new ip header address */
			new_ip = (struct ip *)(new_m->m_data + offset);
			{
				uint16 ip_len = ntohs(new_ip->ip_len);
				uint16 ip_off =ntohs(ip->ip_off);
				/* modify iplen */
				ip_len = ip_hlen + segLen;
				new_ip->ip_len = htons(ip_len);
				/* modify offset */
				ip_off = seg_seq*(segLen>>3);
				new_ip->ip_off += htons(ip_off) ;
				new_ip->ip_off |= htons(IP_MF);
			}	
			assert(new_m->m_pkthdr);
			new_m->m_pkthdr->ph_proto = PKTHDR_IP;
			//new_m->m_pkthdr->ph_rxdesc = -3;
			if ((retval=_rtl8651_fwdEngineSend(property, (void*)(new_m->m_pkthdr),dvid,-1)) != SUCCESS) {
				mBuf_freeMbufChain(new_m);
#if 0				
				if (orgPkthdr)
					mBuf_freeMbufChain(orgPkthdr->ph_mbuf);
#endif
				return retval;
			}
		}


		cp1 = (uint8 *)pkthdrPtr->ph_mbuf->m_data + (seg_seq*segLen);
		rtl8651_memcpy(cp1, (uint8 *)pkthdrPtr->ph_mbuf->m_data, hdrLen);
		mBuf_adjHead(pkthdrPtr->ph_mbuf,  segLen*seg_seq);
		/* get new ip header address */
		ip = (struct ip *)(pkthdrPtr->ph_mbuf->m_data + offset);
		{
			uint16 ip_len = ntohs(ip->ip_len);
			uint16 ip_off =ntohs(ip->ip_off);
			/* modify iplen */
			ip_len = ip_hlen + payloadLen;
			ip->ip_len = htons(ip_len);
			/* modify offset */
			ip_off = seg_seq*(segLen>>3);
			ip->ip_off += htons(ip_off);
		}
		pkthdrPtr->ph_flags &= ~(CSUM_L4);
		if (ip_hlen == 20)
			pkthdrPtr->ph_flags |= (CSUM_IP);
		else
			pkthdrPtr->ph_flags &= ~(CSUM_IP);
#if 0
		if (p_mtuHandler) {
			if ((retval = _rtl8651_fwdEngineSend(property, (void*)pkthdrPtr,dvid,-1)) != SUCCESS) {
				mBuf_freeMbufChain(orgPkthdr->ph_mbuf);
				return retval;
			}
			p_mtuHandler(orgPkthdr);
			return retval;
		}
#endif
		pkthdrPtr->ph_proto = PKTHDR_IP;
		return _rtl8651_fwdEngineSend(property, (void*)pkthdrPtr,dvid,-1);
		
	case 0x0066: /* multicast over MTU */	 
		return FAILED;
	}
#endif /* _RTL_IP_FRAG */

	return FAILED;
}


/*
	Called by slow path packets, do some slow path check, then forwarding engine send it by _rtl8651_l3FastActRouting
*/
static int32 _rtl8651_l3ActRouting(struct rtl_pktHdr * pkthdrPtr, ipaddr_t dst_ip, struct ip *ip)
{
	{
		#ifdef _RTL_DOS_PLUGIN
		/* check for UDP/TCP Scan */
		_rtl8651_dosScanCheckComplain(pkthdrPtr->ph_dsid, ip);	/* complain ? */
		#endif
	}


	return _rtl8651_l3FastActRouting(FWDENG_FROMDRV, pkthdrPtr, dst_ip, ip, NULL);
}


__IRAM_L34_FWD int32 _rtl8651_l3FastActRouting(uint32 property, struct rtl_pktHdr * pkthdrPtr, ipaddr_t dst_ip, struct ip *ip, void  **arp_t)
{

	rtl8651_tblDrv_arpEntry_t *arp_tb;		
	rtl8651_tblDrv_vlanTable_t *vlan_t; 
	rtl8651_tblDrv_pppTable_t *nhSession_t;
	int8 *cp;
	ipaddr_t next_ip;
	rtl8651_tblDrv_routeTable_t *tb;
	int32 pktPPPoE = 0;
	uint32 mtu, tol_len;
	uint16 dvid;
	uint16 srcPortList;
#if	_RTL_DROP_ERROR_MESSAGE
	int32 retval = SUCCESS;
#endif	


#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_GET_MTU_AND_SOURCE_MAC);
#endif

	if (pkthdrPtr->ph_routeIdx >= 1)
	{
		tb = DrvTbl.route + (pkthdrPtr->ph_routeIdx - 1);
	}else
	{
		tb = _rtl8651_getRoutingEntry(dst_ip);			
		if(tb == NULL || ip == NULL)
		{
			return FAILED;
		}
	}

	/* set the initial value of MTU */
	mtu = tb->netif_t->mtu;

	/* Get source MAC Address */
	vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(tb->dvid)];
	if (vlan_t->macAddrNumber == 0)
		return FAILED;

	switch (tb->nextHopType)
	{
		case RTL8651_LL_NONE:
		case RTL8651_LL_VLAN:
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
		case RTL8651_LL_PPTP:
		case RTL8651_LL_L2TP:
#endif
			if( pkthdrPtr->ph_pppeTagged)
			{
				/* PPPoE to Ethernet */
				mBuf_adjHead( pkthdrPtr->ph_mbuf, 8);
				cp = (int8*)(pkthdrPtr->ph_mbuf->m_data);

				/* change ethertype */
				PKTSET_UINT16_UNALIGNED(0x0800, (cp+12));
				/* remove pppoe Tagged on packet header */
				pkthdrPtr->ph_pppeTagged = 0;
				pkthdrPtr->ph_iphdrOffset -= 8;
			} else	/* Ethernet to Ethernet: do nothing */
				cp = (int8*)(pkthdrPtr->ph_mbuf->m_data);

			nhSession_t = NULL; /* make compiler happy */
			break;
		case RTL8651_LL_PPPOE:
			pktPPPoE = 1;	
			cp = (int8*)(pkthdrPtr->ph_mbuf->m_data);
			if( pkthdrPtr->ph_pppeTagged == 0 )
			{	/* Ethernet to PPPoE */
				if( tb->rt_nhSession_t == NULL || tb->rt_nhSession_t->valid == 0) /* for old model */
#if 0
				nhSession_t = (rtl8651_tblDrv_pppTable_t *)tb->rt_nhstart_t->nexthop_t;
				if (nhSession_t == NULL || nhSession_t->valid == 0)
#endif
				{
					return FAILED;
				}
				#if 0
				/* set EtherType */
				*((uint16*)(cp+12)) = htons(0x8864);
				#endif
				/* 
				  * Currently, we use only one MAC address to dial up multiple PPPoE session if multiple
				  * session is configured. In the furture, if one-to-one mapping of MAC address and PPPoE
				  * session is needed, the following code should be changed.
				  */
				rtl8651_memcpy(cp+6, &(vlan_t->macAddr), sizeof(ether_addr_t));
			}else
			{
				/*
				  * PPPoE to PPPoE
				  * convert Session ID
				  * not support now
				  * do nothing ;
									*/
				 return FAILED;
			}
			break;
		default:
			return FAILED;
	}

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_GET_MTU_AND_SOURCE_MAC);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_PPTPL2TP_1);
#endif

	/* we save the source port of this packet */
	srcPortList = pkthdrPtr->ph_portlist;

#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)

	if(tb->nextHopType==RTL8651_LL_PPTP||tb->nextHopType==RTL8651_LL_L2TP)
	{
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD	/* we disable shortest path forwarding if the egress interface is PPTP/L2TP (packet content modification is needed) */
		_rtl8651_fragmentShortestPathCache_clear(pkthdrPtr, ip);
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
		if((tb-DrvTbl.route) != (rtl8651_tblDrvPara.routingTableSize-1))
		{
			 tb = &DrvTbl.route[rtl8651_tblDrvPara.routingTableSize-1];
			
			//goto pptpL2tpSend;
		}
#if 1
		if (tb->rt_nhSession_t->valid == 0 || tb->rt_nhSession_t->initiated == 0)
			return FAILED;
		rtl8651_memcpy(cp, &(tb->rt_nhSession_t->macInfo->macAddr), sizeof(ether_addr_t));
#else
		nhSession_t = (rtl8651_tblDrv_pppTable_t *)tb->rt_nhstart_t->nexthop_t;
		if (nhSession_t == NULL || nhSession_t->valid == 0 || nhSession_t->initiated == 0)
			return FAILED;
		rtl8651_memcpy(cp, &nhSession_t->macInfo->macAddr, 6);
#endif /* 1 */
		rtl8651_memcpy(cp+6, &(vlan_t->macAddr), sizeof(ether_addr_t));
		pkthdrPtr->ph_portlist = DrvTbl.vlan[tb->dvid & (RTL8651_VLAN_NUMBER - 1)].memberPortMask;		 
		pkthdrPtr->ph_dsid = 0;
		goto pptpL2tpSend;	
	}
#endif

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_PPTPL2TP_1);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_PPPOE_ARP_CACHE);
#endif

	if (pktPPPoE == 0)
	{
#if 1
		if (tb->ifrt_flag == 0)	/* Normal Route matched */
		{
			/*
				For Normal Route case:
					Both Next-Hop and Arp information can be gotten from ROUTING ENTRY.
			*/
			next_ip = tb->nextHop;
#if 1
			arp_tb = tb->rt_nharp_t; /* for old model */
#else
			arp_tb = (rtl8651_tblDrv_arpEntry_t *)tb->rt_nhstart_t->nexthop_t;
#endif /* 1 */
		} else				/* Interface Route matched */
		{
			next_ip = dst_ip;

			if (arp_t && *arp_t)
			{
				arp_tb = (rtl8651_tblDrv_arpEntry_t *)(*arp_t);
				if (	(arp_tb->age > 0) &&
					(arp_tb->vid ==tb->dvid) &&
					(arp_tb->ipAddr == next_ip)	)
				{
					/* <-- In this case, arp cache is valid and conformability check OK, we use it. --> */
				} else
				{	/* arp cache is out-of-date or is not conformable (might caused by Hash2 of NAPT), we re-lookup it */
					goto arp_lookup;
				}
			} else
			{
			arp_lookup:
				arp_tb = _rtl8651_getArpEntry(tb->netif_t, dst_ip, TRUE);
			}
		}

		/* update cache if caller want to get it */
		if (arp_t)
		{
			*arp_t = (void *)arp_tb;
		}

#else
		/*  For fast path, we use arp-cache mechanism  */
		if (arp_t == NULL || *arp_t == NULL)
		{
			if (tb->ifrt_flag == 0)
			{ /* Normal route */
#if 1
				arp_tb = tb->rt_nharp_t; /* for old model */
#else
				arp_tb = (rtl8651_tblDrv_arpEntry_t *)tb->rt_nhstart_t->nexthop_t;
#endif /* 1 */
				next_ip = tb->nextHop;
			}else
			{ /* Interface route */
				arp_tb = _rtl8651_getArpEntry(tb->netif_t, dst_ip, TRUE);
				next_ip = dst_ip;
			}
			if (arp_t)
				*arp_t = (void *)arp_tb;
		}else
		{
			arp_tb = (rtl8651_tblDrv_arpEntry_t *)(*arp_t);
			next_ip = arp_tb->ipAddr;
		}
#endif
		/* for large IP Class */
		if (tb->ipClass == 1 && arp_tb && tb->ipintf_t && !tb->ipintf_t->ipUnnumber)
			_rtl8651_ipClassMonitoring(next_ip, arp_tb);

		if (arp_tb == NULL || arp_tb->macInfo == NULL)
		{	/* destination is for gateway */
			rtl8651_tblDrv_ipIntfEntry_t *iftb;
			iftb = _rtl8651_getIpIntfEntryByAddr(next_ip);
			/* do not ask ourself arp  */
			if(iftb && iftb->ipAddr.slh_first )
			{
				if (iftb->ipAddr.slh_first->ipAddr == next_ip)
				{
					return FAILED;
				}
			}
			_rtl8651_drvArpGeneration(next_ip);
#ifdef _RTL_QUEUE_ARP_PACKET
			if(_rtl8651_enQueuePacket(pkthdrPtr,ip, next_ip,Q_ARP) == SUCCESS)
			{
				return SUCCESS;	
			}			
#endif /* _RTL_QUEUE_ARP_PACKET */ 

			return FAILED;
		}

		{
			uint32 mbrMask = arp_tb->macInfo->memberPortMask;
			rtl8651_memcpy(cp, &(arp_tb->macInfo->macAddr), sizeof(ether_addr_t));
			rtl8651_memcpy(cp+6, &vlan_t->macAddr, sizeof(ether_addr_t));
			if (mbrMask == 0xffffffff)
				pkthdrPtr->ph_portlist = DrvTbl.vlan[rtl8651_vlanTableIndex(tb->dvid)].memberPortMask;
			else
				pkthdrPtr->ph_portlist = mbrMask;
		}
	}else
	{	/* PPPoE Packet */
		rtl8651_tblDrv_pppTable_t *poe_tb;
		rtl8651_tblDrv_filterDbTableEntry_t *fb;		

#if 0
		if (pkthdrPtr->ph_unnumber & 0x1)
		{	
#endif
			poe_tb = (rtl8651_tblDrv_pppTable_t *)_rtl8651_getPppoeSessionByDsid(pkthdrPtr->ph_dsid);
			if(poe_tb == NULL)
			{
				return FAILED;
			}
#if 0
		} else /*poe_tb = nhSession_t;*/ poe_tb = tb->rt_nhSession_t;
#endif
		fb = poe_tb->macInfo;
		if(fb == NULL)
		{
			return FAILED;
		}
		poe_tb->txPackets++;
		poe_tb->txBytes += pkthdrPtr->ph_len;
		pkthdrPtr->ph_pppoeIdx=PPPOE_TABLE_INDEX(poe_tb); /* poe_tb->asicPppoeTblIdx; */
		pkthdrPtr->ph_portlist = fb->memberPortMask;
		rtl8651_memcpy(cp, &(fb->macAddr), sizeof(ether_addr_t));
		
		pkthdrPtr->ph_flags |= (PKTHDR_PPPOE_AUTOADD);
		pkthdrPtr->ph_dsid=poe_tb->dsid;

		mtu = (poe_tb->mss + 40);
	}

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_PPPOE_ARP_CACHE);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_PPTPL2TP_SEND);
#endif

#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
pptpL2tpSend:
#endif
	/* <-------------------------- ip ttl and checksum --------------------------> */
	if (property == FWDENG_FROMDRV)
	{
		ip->ip_ttl--;
	}

	/********************************* Checksum *********************************/
	if (IPHDRLEN_IS_TWENTY(ip))
	{
		/* l3 checksum */
		pkthdrPtr->ph_flags |= CSUM_IP;
		if ((pkthdrPtr->ph_proto == PKTHDR_IP) ||
			(property == FWDENG_FROMDRV && pkthdrPtr->ph_proto == PKTHDR_PPTP) ||
			(IS_FRAGMENT(ip->ip_off)))
		{
			/*	if already a fragmented packet, this means this packet
				doesn't have proper l4 information, in this case we
		  		should turn off h/w l4 csum recauculation 					*/
			pkthdrPtr->ph_flags &= (~CSUM_L4);
			pkthdrPtr->ph_proto = PKTHDR_IP;
		}else
		{
			pkthdrPtr->ph_flags |= CSUM_L4;
			if (property == FWDENG_FROMDRV && ip->ip_p == IPPROTO_UDP)
			{
				struct udphdr *ud;
				ud = (struct udphdr*)L4HDR_ADDR(ip);
				if (!ud->uh_sum)
					pkthdrPtr->ph_flags &= (~CSUM_L4);
			}
		}
	}else
	{	/* ASIC Cannot offload L3/L4 calculate, so we must handle it. In l3ActRouting,
		     we must modify ip cksum for ttl decrement */
		if (property == FWDENG_FROMDRV)
		{
			int32 accumulate = 0x0100;
			_RTL8651_ADJUST_CHECKSUM(accumulate, ip->ip_sum);
		}
		/*
			we disable L3/L4 ASIC offload here
		*/
		pkthdrPtr->ph_flags &= ~(CSUM_IP|CSUM_L4);
		pkthdrPtr->ph_proto = PKTHDR_IP;
	}
	/****************************************************************************/

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_PPTPL2TP_SEND);
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_FRAG);
#endif

	dvid = tb->dvid;

	/*
		Before processing packet L2 forwarding,
		we must do extra-process for Protocol Stack offload packet.
	*/
	if (property == FWDENG_OFFLOAD)
	{
#if	_RTL_DROP_ERROR_MESSAGE
		retval = _rtl8651_fwdEnginePSOffload_L3ExtraProcess(pkthdrPtr, ip, vlan_t);
#else
		_rtl8651_fwdEnginePSOffload_L3ExtraProcess(pkthdrPtr, ip, vlan_t);
#endif

	}

	/* If MTU < outgoing packet size, fragment it */
	tol_len = mBuf_getPktlen(pkthdrPtr->ph_mbuf);	
	if (tol_len > (mtu+14))
	{
		#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD	/* we disable shortest path forwarding if packet duplication is needed */
		_rtl8651_fragmentShortestPathCache_clear(pkthdrPtr, ip);
		#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
		return _rtl8651_mtuFragment(property, pkthdrPtr, mtu, tb->dvid, TRUE, srcPortList);
	}

#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_FRAG);
#endif

#if	_RTL_DROP_ERROR_MESSAGE
	if (retval == SUCCESS) {
#endif
	return _rtl8651_fwdEngineSend(property, (void*)pkthdrPtr,dvid,-1);
#if	_RTL_DROP_ERROR_MESSAGE	
	} else {
		mBuf_driverFreeMbufChain(pkthdrPtr->ph_mbuf);
		return SUCCESS;
	}
#endif

}


/*********************************************************************************
* Routine Name :  _rtl8651_l4TcpAdjustMss
* Description :   Adjust mss field of tcp packet
* Input : 
*         delta -- the delta value of base MSS.
*                  For example, IPSEC will use '-56' as a delta value to minus 56 from 1460 (ether type).
* Output :
* Return :
* Note :
**********************************************************************************/
inline void _rtl8651_l4TcpAdjustMss(struct ip *ip, struct tcphdr *tc, ipaddr_t intfIp, int32 delta)
{
	if (tc->th_flags & TH_SYN)
	{
		/*	Now, AutoADD/AutoDELETE is always turned off, and all SYN packet is trapped,	*/
		/*	so we don't need to check auto add is turned-ON or not.						*/
		uint8 *ptr;
		uint32 cnt;
		int32 optLen;
		int32 opt;

		ptr = (uint8 *)((uint8 *)tc + sizeof(struct tcphdr));			/* original: ptr point to the begin of tcp option */
		cnt = ((tc->th_off_x >> 2) & 0x3c) - sizeof(struct tcphdr);	/* original: cnt is the total length of tcp option */

		for (; cnt > 0 ; cnt -= optLen, ptr += optLen)
		{
			opt = (uint32)ptr[0];
			/* check return cases */
			if (opt == TCPOPT_EOL)
				return;

			/* if current option is MSS field, we modify it */
			if (opt == TCPOPT_MAXSEG)
			{
				uint16 mss, new_mss;
				rtl8651_tblDrv_pppTable_t * pppoeEntry;

				mss = ntohs(*(uint16*)(&ptr[2]));
				new_mss = mss;
				pppoeEntry=_rtl8651_getPppoeSessionByIpAddr(intfIp);
				if (pppoeEntry)
				{
					if (mss > pppoe_mss[pppoeEntry->dsid]+delta)
					{
						new_mss = pppoe_mss[pppoeEntry->dsid]+delta;
						*(uint16*)(&ptr[2]) = htons(new_mss);
					}
				}else if (mss > rtl8651_mss+delta)
				{
					new_mss = rtl8651_mss+delta;
					*(uint16*)(&ptr[2]) = htons(new_mss);
				}
				/*	always adjust tcp chskum, if mss is not been modified,					*/
				/*	system will not change checksum because (new_mss == mss) initially.		*/
				if ( 1 /*IS_FRAGMENT_FIRST(ip->ip_off)*/ )
				{
					_RTL8651_DATA_ADJUST_CHECKSUM(new_mss, mss, tc->th_sum);
				}
			}
			/* get next option */			
			if (opt == TCPOPT_NOP)
			{
				optLen = 1;
			}else
			{
				optLen= ptr[1];
				if (optLen <= 0)
					return;
			}
		}
	}
}


/*********************************************************************************
* Routine Name :  _rtl8651_l4TcpMonitorIn                                    
* Description :   Monitor inbound TCP flags   
* TCP Handling Routines
* This function is implemented according to "Draft Home Gateway ASIC Specification", pp65
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
* 	Reimplemented by cfliu 2003/7/8
**********************************************************************************/
static inline int32 _rtl8651_l4TcpMonitorIn(struct tcphdr *tc, rtl8651_tblDrv_naptTcpUdpFlowEntry_t* tb)
{
	uint32 pktFlags, currentState;
	int32 retval;
	/* software auto delete is alwyas on */

	pktFlags = tc->th_flags &  (TH_RST|TH_SYN|TH_FIN);
	currentState = tb->tcpFlag;
	//rtlglue_printf("inbound from %s:%d . current state %d, pktflags %d\n", ipstr, tb->dstPort,currentState, pktFlags);
	assert(currentState<=TCP_CLOSED_FLOW-1);
	if(pktFlags==0){ 
resetage:
		//No special flags in packet, just update age timer
		if (_ageTimer[currentState+1])
			retval=_rtl8651_updateNaptConnection( tb, 0, *_ageTimer[currentState+1]);
		else
			retval = SUCCESS;
		return retval;
	} 

	retval = FAILED;
	if (pktFlags>=TH_RST){
		retval=_rtl8651_updateNaptConnection( tb, TCP_CLOSED_FLOW, rtl8651GlobalControl.tcpFastTimeout);
	} else if(pktFlags==TH_SYN){
		//only SYN set
		int32 state=_inNextState[0][currentState];
		if(state>0){
			if (_ageTimer[state])
				retval=_rtl8651_updateNaptConnection( tb, state, *_ageTimer[state]);
			else
				retval = SUCCESS;
		}else if(state==0) //No action, just reset age timer
			goto resetage;
	}else{
		//At least FIN set
		int32 state=_inNextState[1][currentState];
		if(state>0){
			if (_ageTimer[state])
				retval=_rtl8651_updateNaptConnection( tb, state, *_ageTimer[state]);
			else
				retval = SUCCESS;
		}else if(state==0) //No action, just reset age timer
			goto resetage;
	}
	return retval;

}


/*********************************************************************************
* Routine Name :  _rtl8651_l4TcpMonitorOut                                    
* Description :   Monitor outbound TCP flags   
* TCP Handling Routines
* This function is implemented according to "Draft Home Gateway ASIC Specification", pp65
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
* 	Reimplemented by cfliu 2003/7/8
**********************************************************************************/
static inline int32 _rtl8651_l4TcpMonitorOut(struct tcphdr *tc, rtl8651_tblDrv_naptTcpUdpFlowEntry_t* tb)
{
	uint32 pktFlags, currentState;
	int32 retval;
	/* software auto delete is alwyas on */

	pktFlags = tc->th_flags &  (TH_RST|TH_SYN|TH_FIN);
	currentState = tb->tcpFlag;
	//rtlglue_printf("outbound to %s:%d . current state %d, pktflags %d\n", ipstr, tb->dstPort,currentState, pktFlags);
	assert(currentState<=TCP_CLOSED_FLOW-1);
	if(pktFlags==0){ 
resetage:
		//No SYN/FIN flags in packet, just update age timer
		if (_ageTimer[currentState+1])
			retval=_rtl8651_updateNaptConnection( tb, 0, *_ageTimer[currentState+1]);
		else
			retval = SUCCESS;
		return retval;
	}

	retval = FAILED;
	if (pktFlags >= TH_RST) {
		retval=_rtl8651_updateNaptConnection( tb, TCP_CLOSED_FLOW, rtl8651GlobalControl.tcpFastTimeout);
	}else if(pktFlags==TH_SYN){
		//only SYN set
		
		int32 state=_outNextState[0][currentState];
		if(state>0){
			if ( _ageTimer[state] )
				retval=_rtl8651_updateNaptConnection( tb, state, *_ageTimer[state]);
			else
				retval = SUCCESS;
		}else if(state==0) //No action, just reset age timer
			goto resetage;
	}else{
		//At least FIN set
		int32 state=_outNextState[1][currentState];
		if(state>0){
			if ( _ageTimer[state] )
				retval=_rtl8651_updateNaptConnection( tb, state, *_ageTimer[state]);
			else
				retval = SUCCESS;
		}
		else if(state==0) //No action, just reset age timer
			goto resetage;
	}
	return retval;
}

#ifdef _RTL_NAPT_ADVANCED
/*********************************************************************************
* Routine Name :	_rtl8651_l4CheckOccupiedConnection
* Description :	for TCP/UDP, check if the dsid/IP/Port is occpuied or not
* Input :
* Output :
			TRUE:	ip/port is inused
			FALSE:	ip/port is free
* Return :
* Note :
**********************************************************************************/
static int8 _rtl8651_l4CheckOccupiedConnection(uint32 isTcp, uint32 dsid, ipaddr_t IntIp, uint16 IntPort, ipaddr_t ExtIp, uint16 ExtPort, ipaddr_t RemoteIp, uint16 RemotePort)
{
	rtl8651_tblDrv_naptServerPortEntry_t *sp_t = NULL;

	/* (1) check NAPT connections */
	/*
		chenyl (2005/05/04):

			Because there are many different conditions to decide if this external Ip/Port can be used
			or not, so we just check this when we actually add this flow to NAPT system.

			==> If this function is used in other usage, we MUST add code to check NAPT system.
	*/

	/* (2) check external server port */
	if ((sp_t = _rtl8651_getOutsideNaptServerPortMapping(isTcp, ExtIp, ExtPort)) != NULL)
	{
		if (((IntPort + sp_t->endLocalPort - sp_t->localPort) != ExtPort) || (sp_t->localIp != IntIp))
		{
			return TRUE;
		}
	}

	/* (3) check protocol stack registered port */
	if (_rtl8651_checkProtoStackRegistedPort(ExtPort, (isTcp?IPPROTO_TCP:IPPROTO_UDP)))
	{
		return TRUE;
	}

	/* (4) check if upnp map is used by others or not */
	if (_rtl8651_checkUsableUpnpEntry(isTcp, IntIp, IntPort, ExtIp, ExtPort, RemoteIp, RemotePort) == FALSE)
	{
		return TRUE;
	}

	/* (5) check if trigger port is triggered to map others or not */
	if (_rtl8651_checkUsableTriggeredPortEntry(isTcp, dsid, IntIp, ExtPort) == FALSE)
	{
		return TRUE;
	}

	return FALSE;
}
#endif /* _RTL_NAPT_ADVANCED */

/*
	Check DMZ inbound forward for Non-TCP/UDP/ICMP packets
*/
static int32 _rtl8651_inboundL4DMZForward(struct rtl_pktHdr *phdr, struct ip *ip, ipaddr_t dip)
{
	ipaddr_t dmzIp = 0;

	if (!rtl8651_drvFwdOtherL4ToDMZEnable)
		goto errout;

	if (_rtl8651_getDmzHost(phdr->ph_dsid, dip, &dmzIp) != SUCCESS)
		goto errout;

	/* dmz host found, modify dip */
	PKTSET_UINT32_UNALIGNED(dmzIp, &ip->ip_dst);

	/* modify */
	if (!IPHDRLEN_IS_TWENTY(ip))
	{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
		/* for fragment packet, it don't need to calculate L3 checksum */
		_RTL8651_L3_ADJUST_CHECKSUM(dmzIp, dip, ip->ip_sum);	/* L3 Checksum */
		/* for IGMP packet, ip modification is unconcerned */
	}
	if (IS_FRAGMENT_FIRST(ip->ip_off))
	{	/* first packet of fragment flow */
		_rtl8651_updateFragCache(INBOUND_FRAG, phdr->ph_sip, phdr->ph_dip, ntohs(ip->ip_id), dmzIp, 0, NULL, phdr->ph_dsid);
	}

	return SUCCESS;

errout:
	return FAILED;
}

/*
	Check DMZ outbound forward for Non-TCP/UDP/ICMP packets
*/
static int32 _rtl8651_outboundL4DMZForward(struct rtl_pktHdr *phdr, struct ip *ip, ipaddr_t sip)
{
	ipaddr_t naptIp = 0;

	if (!rtl8651_drvFwdOtherL4ToDMZEnable)
		goto errout;

	if (_rtl8651_fromDmzHost(phdr->ph_dsid, sip, &naptIp) != SUCCESS)
		goto errout;

	/* dmz host found, modify dip */
	PKTSET_UINT32_UNALIGNED(naptIp, &ip->ip_src);

	/* modify */
	if (!IPHDRLEN_IS_TWENTY(ip))
	{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
		/* for fragment packet, it don't need to calculate L3 checksum */
		_RTL8651_L3_ADJUST_CHECKSUM(naptIp, sip, ip->ip_sum);	/* L3 Checksum */
		/* for IGMP packet, ip modification is unconcerned */
	}
	if (IS_FRAGMENT_FIRST(ip->ip_off))
	{	/* first packet of fragment flow */
		_rtl8651_updateFragCache(OUTBOUND_FRAG, phdr->ph_sip, phdr->ph_dip, ntohs(ip->ip_id), naptIp, 0, NULL, phdr->ph_dsid);
	}

	return SUCCESS;

errout:
	return FAILED;
}


/*********************************************************************************
* Routine Name : _rtl8651_l4IcmpNaptAliasIn1                                    
* Description :     
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4IcmpNaptAliasIn1(struct rtl_pktHdr *phdr, struct ip *pip,struct icmp *ic)
{

	rtl8651_tblDrv_naptIcmpFlowEntry_t * tb;
		 
	ipaddr_t  ip_org = 0;
	uint16 id, id_org = 0;
	
	id = ntohs(ic->icmp_id);

	tb = _rtl8651_getOutsideNaptIcmpEntry(phdr->ph_dip, id, phdr->ph_sip, TRUE, TRUE);

	if (tb != NULL)
	{
		id_org = tb->insideLocalId;
		ip_org= tb->insideLocalIpAddr;

		if (((pip->ip_vhl & 0x0f) != 5) || (ntohs(pip->ip_off) & IP_MF))
		{
			if ((pip->ip_vhl & 0x0f) != 5)
			{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
				/* for fragment packet, it don't need to calculate L3 checksum */
				_RTL8651_L3_ADJUST_CHECKSUM(ip_org, phdr->ph_sip, pip->ip_sum);	/* L3 Checksum */
			}
			/* for both fragment packet and non-ASIC offloaded packet, we must calculate L4 checksum by our self */
			_RTL8651_L4_ADJUST_CHECKSUM(0, 0, id_org, id, ic->icmp_cksum);	/* L4 Checksum: icmp don't include pseudo header */
		}
		#if 0
		/* l4 checksum */
		if (ntohs(pip->ip_off) & IP_MF)
		{
			_RTL8651_L4_ADJUST_CHECKSUM(0 , 0, id_org, id, ic->icmp_cksum);
		}
		#endif
		/* l3 Put original address back into IP header */
		PKTSET_UINT32_UNALIGNED(ip_org, &pip->ip_dst);


		/* l4 Put original sequence number back in */
		ic->icmp_id = htons(id_org);		
		return (SUCCESS);
	}
	
	return (FWDENG_ICMP_IN_NOTFOUND);
}


/*********************************************************************************
* Routine Name : _rtl8651_l4IcmpNaptAliasIn2                                    
* Description :         
*   Alias incoming ICMP error messages containing
*   IP header and first 64 bits of datagram.
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4IcmpNaptAliasIn2(struct rtl_pktHdr *phdr, struct ip *pip,struct icmp *ic)
{

/*
    Alias incoming ICMP error messages containing
    IP header and first 64 bits of datagram.
*/
	struct ip *ip;
	struct icmp *ic2=NULL;
	struct udphdr *ud=NULL;
	struct tcphdr *tc=NULL;
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry;

	rtl8651_tblDrv_naptIcmpFlowEntry_t * tb_icmp = NULL;
	rtl8651_tblDrv_naptTcpUdpFlowEntry_t * tb_tcpudp = &entry;

	ipaddr_t ip_s, ip_d;
	
	ip = &ic->icmp_ip;

	ip_s = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);
	ip_d =PKTGET_UINT32_UNALIGNED(&ip->ip_src);

	if (ip->ip_p == IPPROTO_UDP)
	{
		ud = (struct udphdr *) ((int8 *)ip + ((ip->ip_vhl & 0x0f) << 2));
		if(_rtl8651_getInboundNaptFlow(FALSE, ip_d, ntohs(ud->uh_sport),
					ip_s, ntohs(ud->uh_dport), &tb_tcpudp )!=SUCCESS)
		{
			tb_tcpudp = NULL;
		}
	} else if (ip->ip_p == IPPROTO_TCP){
		tc = (struct tcphdr *) ((int8 *) ip + ((ip->ip_vhl & 0x0f) << 2));
		if(_rtl8651_getInboundNaptFlow(TRUE, ip_d, ntohs(tc->th_sport),
					ip_s, ntohs(tc->th_dport), &tb_tcpudp )!=SUCCESS)
		{
			tb_tcpudp = NULL;
		}
	} else if (ip->ip_p == IPPROTO_ICMP)
	{
		ic2 = (struct icmp *) ((int8 *) ip + ( (ip->ip_vhl & 0x0f) << 2));
		tb_icmp = _rtl8651_getOutsideNaptIcmpEntry(ip_d, ntohs(ic2->icmp_id), ip_s, FALSE, FALSE);
		tb_tcpudp = NULL;
	} else
	{	/* other */
		#ifdef RTL865X_DEBUG
		rtlglue_printf("Icmp error message with payload L4 type is %d\n", (uint32)ip->ip_p);
		#endif
		tb_icmp = NULL;
		tb_tcpudp = NULL;
	}
	
	if (tb_tcpudp){
		ipaddr_t original_address;
		uint16    original_port;

		/**************************************************
			If this flow need to be trapped to protocol stack?
		**************************************************/
		if (tb_tcpudp->toProtocolStack == TRUE)
		{
			/*
				It means the original flow is need to be trapped
				to protocol stack, so the corresponding ICMP error message must be trapped, too.
			*/
			if (!IS_FRAGMENT_REMAIN(ip->ip_off))
			{
				uint16 orgSPort, orgDPort;
				uint16 *pnewSPort;
				uint16 *pnewDPort;
				if (ip->ip_p == IPPROTO_TCP)
				{
					orgSPort = ntohs(tc->th_sport);
					orgDPort = ntohs(tc->th_dport);
					pnewSPort = &(tc->th_sport);
					pnewDPort = &(tc->th_dport);

					if ( tb_tcpudp->insideLocalPort != orgSPort )
					{
						tc->th_sport = htons(tb_tcpudp->insideLocalPort);
						_RTL8651_L4_ADJUST_CHECKSUM(0, 0, tb_tcpudp->insideLocalPort, orgSPort, ic->icmp_cksum);
						orgSPort = tb_tcpudp->insideLocalPort;	
					}
				}else	/* udp */
				{
					orgSPort = ntohs(ud->uh_sport);
					orgDPort = ntohs(ud->uh_dport);
					pnewSPort = &(ud->uh_sport);
					pnewDPort = &(ud->uh_dport);

					if ( tb_tcpudp->insideLocalPort != orgSPort )
					{
						ud->uh_sport = htons(tb_tcpudp->insideLocalPort);
						_RTL8651_L4_ADJUST_CHECKSUM(0, 0, tb_tcpudp->insideLocalPort, orgSPort, ic->icmp_cksum);
						orgSPort = tb_tcpudp->insideLocalPort;	
					}
				}
				if (_rtl8651_aliasProtoStackRecvPkt(TRUE, ip) == PS_ACTION_TO_PS)
				{
					/*
						1. we only modify ICMP's checksum.
							- we don't modify L3 checksum because ip is not modified.
							- we don't modify L3 checksum in error payload because ip is not modified.
							- we don't modify L4 checksum in error payload because end-host cannot verify it.
						2. we don't care "doProtoStackAction" modify sport
							or dport, we modify all of them (if orgPort == newPort, checksum will not be modified) */
					_RTL8651_L4_ADJUST_CHECKSUM(0, 0, ntohs(*pnewSPort), orgSPort, ic->icmp_cksum);
					_RTL8651_L4_ADJUST_CHECKSUM(0, 0, ntohs(*pnewDPort), orgDPort, ic->icmp_cksum);
				}
			}
			return FWDENG_TO_GW;
		}

		original_address = tb_tcpudp->insideLocalIpAddr;
		original_port = tb_tcpudp->insideLocalPort;

		/******************************* Modify error content's checksum *******************************/
		 /* error content: only modify l3 checksum */
		 _RTL8651_L3_ADJUST_CHECKSUM(original_address, ip_d, ip->ip_sum);

		/********************************** Modify packet's checksum **********************************/
		if (((pip->ip_vhl & 0x0f) != 5) || (ntohs(pip->ip_off) & IP_MF))
		{
			if ((pip->ip_vhl & 0x0f) != 5)
			{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
				/* for fragment packet, it don't need to calculate L3 checksum */
				_RTL8651_L3_ADJUST_CHECKSUM(original_address, ip_d, pip->ip_sum);	/* L3 Checksum */
			}
			/* L4 Checksum: adjust for error content's modification */
			if (ip->ip_p == IPPROTO_TCP)
			{
				_RTL8651_L4_ADJUST_CHECKSUM(original_address, ip_d, original_port, ntohs(tc->th_sport), ic->icmp_cksum);
			}else	/* udp */
			{
				_RTL8651_L4_ADJUST_CHECKSUM(original_address, ip_d, original_port, ntohs(ud->uh_sport), ic->icmp_cksum);
			}
		}
		#if 0
		 /* l4 checksum */
		if (ntohs(pip->ip_off) & IP_MF)
		{
			if (ip->ip_p == IPPROTO_TCP)
			{
				_RTL8651_L4_ADJUST_CHECKSUM(original_address, ip_d, original_port, tc->th_sport, ic->icmp_cksum);
			}else
			{	// udp
				_RTL8651_L4_ADJUST_CHECKSUM(original_address, ip_d, original_port, ud->uh_sport, ic->icmp_cksum);
			}
		}
		#endif
		/*******************************************************************************************/
		/*l3 Un-alias address in IP header */
		PKTSET_UINT32_UNALIGNED(original_address, &pip->ip_dst);

		/*l4 Un-alias address and port number of original IP packet
		fragment contained in ICMP data section */
		PKTSET_UINT32_UNALIGNED(original_address, &ip->ip_src);
		if (ip->ip_p == IPPROTO_UDP)
		{
			ud->uh_sport = htons(original_port);
		}
		else if (ip->ip_p == IPPROTO_TCP)
		{
			tc->th_sport = htons(original_port);
		}
		return (SUCCESS);

	}else if (tb_icmp)
	{
		ipaddr_t original_address;
		uint16    original_id;

		original_address = tb_icmp->insideLocalIpAddr;
		original_id = tb_icmp->insideLocalId;

		/******************************* Modify error content's checksum *******************************/
		 /* error content: only modify l3 checksum */
		_RTL8651_L3_ADJUST_CHECKSUM(original_address, ip_d, ip->ip_sum);

		/********************************** Modify packet's checksum **********************************/
		if (((pip->ip_vhl & 0x0f) != 5) || (ntohs(pip->ip_off) & IP_MF))
		{
			if ((pip->ip_vhl & 0x0f) != 5)
			{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
				/* for fragment packet, it don't need to calculate L3 checksum */
				_RTL8651_L3_ADJUST_CHECKSUM(original_address, ip_d, pip->ip_sum);	/* L3 Checksum */
			}
			/* L4 Checksum: adjust for error content's modification */
			_RTL8651_L4_ADJUST_CHECKSUM(original_address, ip_d, original_id, ntohs(ic2->icmp_id), ic->icmp_cksum);
		}
		#if 0
		/* l4 checksum */
		if (ntohs(pip->ip_off) & IP_MF)
		{
			_RTL8651_L4_ADJUST_CHECKSUM(original_address, ip_d, original_id, ic2->icmp_id, ic->icmp_cksum);
		}
		#endif
		/*******************************************************************************************/

		/* l3 Un-alias address in IP header */
		PKTSET_UINT32_UNALIGNED(original_address, &pip->ip_dst);

		/* l4 Un-alias address of original IP packet and 
			sequence number of embedded ICMP datagram */
		PKTSET_UINT32_UNALIGNED(original_address, &ip->ip_src);
		ic2->icmp_id = htons(original_id);	
		return (SUCCESS);
	}else
	{
		return (FWDENG_ICMP_IN_ERRFLOWNOTFOUND);
	}
}

/*********************************************************************************
* Routine Name : _rtl8651_l4IcmpNaptAliasIn3                                 
* Description : Process Inbound ECHO/TSTAMP packet, forward to DMZ host and add icmp entry
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4IcmpNaptAliasIn3(struct rtl_pktHdr *phdr, struct ip *pip,struct icmp *ic)
{
	ipaddr_t ip_s, ip_d;
	ipaddr_t dmzhost;
	int r;
	
	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;

#ifdef _RTL_PS_ACTION
		r = _rtl8651_doProtoStackAction(pip);
		if(r==PS_ACTION_DROP) return FWDENG_USER_DROP;
#endif
	if (!rtl8651_DMZHostIcmpPassthroughEnable)
	{
		return FWDENG_ICMP_IN_NOTFOUND;
	}
	// forward this packet to dmzhost
	if (_rtl8651_getDmzHost(phdr->ph_dsid, ip_d, &dmzhost) == SUCCESS)
	{
		uint16 id;
		rtl8651_tblDrv_naptIcmpFlowEntry_t* icmp_tb;

		id = ntohs(ic->icmp_id);		// id of icmp packet
		icmp_tb = _rtl8651_getOutsideNaptIcmpEntry(ip_d, id, ip_s, TRUE, FALSE);

		if (!icmp_tb)
		{	
			{
				/* 
					Now, we are going to establish a new flow, 
					but we have to check the ingress ACL of "L4NEWFLOW" type first.
				*/
				int32 local_val;
				rtl8651_tblDrv_vlanTable_t *local_vlanp = &DrvTbl.vlan[phdr->ph_vlanIdx];
				local_val = _rtl8651_ingressAcl(phdr->ph_dsid, 0, local_vlanp, local_vlanp->vid, phdr->ph_mbuf->m_data, (uint8 *) pip, RTL8651_ACL_L4NEWFLOW, _RTL8651_OPERATION_OTHER);
				if (((local_val > 0) && (local_val & FWDENG_DROP)) ||(local_val == FWDENG_ACL_IN_USERTRAP))
					return local_val;
			}

			if( _rtl8651_addNaptIcmpNonCandidateFlow(TRUE, dmzhost, id, ip_d, id, ip_s)!= SUCCESS)
			{
				return FWDENG_ICMP_IN_ADDFAIL;
			}

			/* get flow structure: to store informations in the structure */
			icmp_tb = _rtl8651_getOutsideNaptIcmpEntry(ip_d, id, ip_s, TRUE, FALSE);
		}

		/* store caches into the flow structure */
		assert(icmp_tb);
		icmp_tb->dsid = phdr->ph_dsid;

		/* l4 checksum */
		/* for fragment packet, We don't change icmp ID and just adjust L3 IP, so icmp checksum is not changed */
		/* for packet whose ip header length != 20, we must recalculate it's L3 checksum */
		if ((pip->ip_vhl & 0x0f) != 5)
			_RTL8651_L3_ADJUST_CHECKSUM(dmzhost, ip_d, pip->ip_sum);

		/* Change destination address */		
		PKTSET_UINT32_UNALIGNED(dmzhost, &pip->ip_dst);
		return SUCCESS;
	} else
	{		
		return FWDENG_ICMP_IN_NOTFOUND;
	}
}


/*********************************************************************************
* Routine Name : _rtl8651_L4IcmpNaptAliasIn                                    
* Description : 
*   process icmp input packet
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_L4IcmpNaptAliasIn(struct rtl_pktHdr *phdr, struct ip *pip)
{
	int32     iresult;
	struct icmp *ic;	


	ic = (struct icmp *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));	
	switch (ic->icmp_type)
	{
		case ICMP_ECHOREPLY:
		case ICMP_TSTAMPREPLY:
			iresult = _rtl8651_l4IcmpNaptAliasIn1(phdr, pip, ic);
			break;
		case ICMP_UNREACH:
		case ICMP_SOURCEQUENCH:
		case ICMP_TIMXCEED:
		case ICMP_PARAMPROB:
			iresult = _rtl8651_l4IcmpNaptAliasIn2(phdr, pip, ic);		
			break;
		case ICMP_ECHO:
		case ICMP_TSTAMP:
			iresult = _rtl8651_l4IcmpNaptAliasIn3(phdr, pip, ic);
			break;
		default:
			return (FWDENG_ICMP_IN_NOTFOUND);
	}

	if (iresult == SUCCESS)
	{
#ifdef _RTL_NAPT_FRAG	
		if(ntohs(pip->ip_off) &  (IP_MF|IP_OFFMASK)){
			int32 ta;
			phdr->ph_proto = PKTHDR_IP;
			ta = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
			_rtl8651_updateFragCache(INBOUND_FRAG, phdr->ph_sip, phdr->ph_dip, ntohs(pip->ip_id), ta, 0, NULL, phdr->ph_dsid);
		}
#endif /* _RTL_NAPT_FRAG */	
	}
	return (iresult);
}


/*********************************************************************************
* Routine Name :  _rtl8651_l4IcmpNaptAliasOut1                                    
* Description :  
* Input :
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4IcmpNaptAliasOut1(struct rtl_pktHdr *phdr, struct ip *pip,struct icmp *ic)
{	    	
	rtl8651_tblDrv_naptIcmpFlowEntry_t * tb;

	ipaddr_t ip_s, ip_d, ip_alias = 0;
	uint16 id, id_alias = 0;

	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	id = ntohs(ic->icmp_id);
	
	tb = _rtl8651_getInsideNaptIcmpEntry(ip_s, id, ip_d, TRUE, FALSE);
	
	if(tb)
	{		
		ip_alias = tb->insideGlobalIpAddr;
		id_alias = tb->insideGlobalId;		
	}
	else
	{

		{
			/* 
				Now, we are going to establish a new flow, 
				but we have to check the ingress ACL of "L4NEWFLOW" type first.
			*/
			int32 local_val;
			rtl8651_tblDrv_vlanTable_t *local_vlanp = &DrvTbl.vlan[phdr->ph_vlanIdx];
			local_val = _rtl8651_ingressAcl(phdr->ph_dsid, 0, local_vlanp, local_vlanp->vid, phdr->ph_mbuf->m_data, (uint8 *) pip, RTL8651_ACL_L4NEWFLOW, _RTL8651_OPERATION_OTHER);
			if (((local_val > 0) && (local_val & FWDENG_DROP)) ||(local_val == FWDENG_ACL_IN_USERTRAP))
				return local_val;
		}

		if (_rtl8651_softIcmpPolicyRoute(ip_s, id, ip_d, &ip_alias, &id_alias) == SUCCESS)
		{
			int32 retval;
			while ((retval = _rtl8651_addNaptIcmpNonCandidateFlow(TRUE, ip_s, id, ip_alias, id_alias, ip_d))
					!= SUCCESS)
			{
				if (retval == TBLDRV_EEXTIDINUSE)
				{
					// assigned from a global valiable simply, so we can modify it as we wish
					id_alias ++;
				}
				else
				{
					return FWDENG_ICMP_OUT_ADDFAIL;
				}
			}
		}
		else
		{
			if (_rtl8651_getIcmpUsableExtIpAndID(ip_s, id, ip_d, 
				&ip_alias, &id_alias) != SUCCESS)
			{
				return FWDENG_ICMP_OUT_INVALIDIPID;				
			}
			if( _rtl8651_addNaptIcmpNonCandidateFlow(
				TRUE, ip_s, id, ip_alias, id_alias, ip_d)!= SUCCESS ){
				return FWDENG_ICMP_OUT_ADDFAIL;
			}
		}

		/* get flow structure: to store informations in the structure */
		tb = _rtl8651_getInsideNaptIcmpEntry(ip_s, id, ip_d, TRUE, FALSE);
	}	

	/* store caches into the flow structure */
	assert(tb);
	tb->dsid = phdr->ph_dsid;

	if (((pip->ip_vhl & 0x0f) != 5) || (ntohs(pip->ip_off) & IP_MF))
	{
		if ((pip->ip_vhl & 0x0f) != 5)
		{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
			/* for fragment packet, it don't need to calculate L3 checksum */
			_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, ip_s, pip->ip_sum);	/* L3 Checksum */
		}
		/* for both fragment packet and non-ASIC offloaded packet, we must calculate L4 checksum by our self */
		_RTL8651_L4_ADJUST_CHECKSUM(0, 0, id_alias, id, ic->icmp_cksum);	/* L4 Checksum: icmp don't include pseudo header */
	}
	#if 0
	/* l4 checksum */
	if (ntohs(pip->ip_off) & IP_MF)
	{	/* icmp don't include pseudo header */
		_RTL8651_L4_ADJUST_CHECKSUM(0, 0, id_alias, id, ic->icmp_cksum);
	}
	#endif
							
	/* Alias sequence number */
	ic->icmp_id = htons(id_alias);

	/* Change source address */		
	PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);	
	return (SUCCESS);
}

/*********************************************************************************
* Routine Name : _rtl8651_l4IcmpNaptAliasOut2
* Description :         
*   Alias outgoing ICMP error messages containing
*   IP header and first 64 bits of datagram.
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4IcmpNaptAliasOut2(struct rtl_pktHdr *phdr, struct ip *pip,struct icmp *ic)
{
/*
    Alias outgoing ICMP error messages containing
    IP header and first 64 bits of datagram.
*/
	struct ip *ip;
	struct icmp *ic2=NULL;
	struct udphdr *ud=NULL;
	struct tcphdr *tc=NULL;
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb_tcpudp = &entry;
	rtl8651_tblDrv_naptIcmpFlowEntry_t * tb_icmp = NULL;

	ipaddr_t ip_s, ip_d;

	ip = &ic->icmp_ip;

	ip_s = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);
	ip_d =PKTGET_UINT32_UNALIGNED(&ip->ip_src);

	if (ip->ip_p == IPPROTO_UDP)
	{
		ud = (struct udphdr *) ((int8 *) ip + ( (ip->ip_vhl & 0x0f) << 2));
		if(_rtl8651_getOutboundNaptFlow(FALSE,ip_s, ntohs(ud->uh_dport),
					ip_d,ntohs(ud->uh_sport),&tb_tcpudp )!=SUCCESS)
		{
			tb_tcpudp = NULL;
		}
	}else if (ip->ip_p == IPPROTO_TCP)
	{
		tc = (struct tcphdr *) ((int8 *) ip + ( (ip->ip_vhl & 0x0f) << 2));
		if(_rtl8651_getOutboundNaptFlow(TRUE,ip_s, ntohs(tc->th_dport),
					ip_d,ntohs(tc->th_sport),&tb_tcpudp )!=SUCCESS)
		{
			tb_tcpudp = NULL;
		}
	}else if (ip->ip_p == IPPROTO_ICMP)
	{
		ic2 = (struct icmp *) ((int8 *) ip + ( (ip->ip_vhl & 0x0f) << 2));
		tb_icmp = _rtl8651_getInsideNaptIcmpEntry(ip_s, ntohs(ic2->icmp_id), ip_d, FALSE, FALSE);
		tb_tcpudp = NULL;
	}else
	{	/* other */
		#ifdef RTL865X_DEBUG
		rtlglue_printf("Icmp error message with payload L4 type is %d\n", ip->ip_p);
		#endif
		tb_icmp = NULL;
		tb_tcpudp = NULL;
	}
	
	if (tb_tcpudp){
		ipaddr_t alias_address;
		uint16    alias_port;
//		int32 accumulate;

		alias_address = tb_tcpudp->insideGlobalIpAddr;
		alias_port = tb_tcpudp->insideGlobalPort;
		/* If the routing cache exists: should use cache route information */
		if (tb_tcpudp->out_rt_t)
		{
			phdr->ph_routeIdx = L3_TABLE_INDEX((rtl8651_tblDrv_routeTable_t *)tb_tcpudp->out_rt_t) + 1;
		}

		/******************************* Modify error content's checksum *******************************/
		 /* error content: only modify l3 checksum */
		 _RTL8651_L3_ADJUST_CHECKSUM(alias_address, ip_s, ip->ip_sum);

		/********************************** Modify packet's checksum **********************************/
		if (((pip->ip_vhl & 0x0f) != 5) || (ntohs(pip->ip_off) & IP_MF))
		{
			if ((pip->ip_vhl & 0x0f) != 5)
			{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
				/* for fragment packet, it don't need to calculate L3 checksum */
				_RTL8651_L3_ADJUST_CHECKSUM(alias_address, ip_s, pip->ip_sum);	/* L3 Checksum */
			}
			/* L4 Checksum: adjust for error content's modification */
			if (ip->ip_p == IPPROTO_TCP)
			{
				_RTL8651_L4_ADJUST_CHECKSUM(alias_address, ip_s, alias_port, ntohs(tc->th_dport), ic->icmp_cksum);
			}else	/* udp */
			{
				_RTL8651_L4_ADJUST_CHECKSUM(alias_address, ip_s, alias_port, ntohs(ud->uh_dport), ic->icmp_cksum);
			}
		}
		#if 0
		 /* l4 checksum */
		if (ntohs(pip->ip_off) && IP_MF)
		{
			if (ip->ip_p == IPPROTO_TCP)
			{
				_RTL8651_L4_ADJUST_CHECKSUM(alias_address, ip_s, alias_port, tc->th_dport, ic->icmp_cksum);
			}else
			{	// udp
				_RTL8651_L4_ADJUST_CHECKSUM(alias_address, ip_s, alias_port, ud->uh_dport, ic->icmp_cksum);
			}
		}
		#endif
		/*******************************************************************************************/
		/*l3 Un-alias address in IP header */
		PKTSET_UINT32_UNALIGNED(alias_address, &pip->ip_src);

		/*l4 Un-alias address and port number of original IP packet
		fragment contained in ICMP data section */
		PKTSET_UINT32_UNALIGNED(alias_address, &ip->ip_dst);
		if (ip->ip_p == IPPROTO_UDP)
		{
			ud->uh_dport = htons(alias_port);
		}
		else if (ip->ip_p == IPPROTO_TCP)
		{
			tc->th_dport = htons(alias_port);
		}

		/* Un-alias dsid */
		phdr->ph_dsid = tb_tcpudp->dsid;
		
		return (SUCCESS);

	}
	else if (tb_icmp){
		ipaddr_t alias_address;
		uint16    alias_id;
//		int32 accumulate;

		alias_address = tb_icmp->insideGlobalIpAddr;
		alias_id = tb_icmp->insideGlobalId;

		/******************************* Modify error content's checksum *******************************/
		 /* error content: only modify l3 checksum */
		 _RTL8651_L3_ADJUST_CHECKSUM(alias_address, ip_s, ip->ip_sum);

		/********************************** Modify packet's checksum **********************************/
		if (((pip->ip_vhl & 0x0f) != 5) || (ntohs(pip->ip_off) & IP_MF))
		{
			if ((pip->ip_vhl & 0x0f) != 5)
			{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
				/* for fragment packet, it don't need to calculate L3 checksum */
				_RTL8651_L3_ADJUST_CHECKSUM(alias_address, ip_s, pip->ip_sum);	/* L3 Checksum */
			}
			/* L4 Checksum: adjust for error content's modification */
			_RTL8651_L4_ADJUST_CHECKSUM(alias_address, ip_s, alias_id, ntohs(ic2->icmp_id), ic->icmp_cksum);
		}
		#if 0
		/* l4 checksum */
		if (ntohs(pip->ip_off) & IP_MF)
		{
			_RTL8651_L4_ADJUST_CHECKSUM(alias_address, ip_s, alias_id, ic2->icmp_id, ic->icmp_cksum);
		}
		#endif
		/*******************************************************************************************/
		/* l3 Un-alias address in IP header */
		PKTSET_UINT32_UNALIGNED(alias_address, &pip->ip_src);

		/* l4 Un-alias address of original IP packet and 
			sequence number of embedded ICMP datagram */
		PKTSET_UINT32_UNALIGNED(alias_address, &ip->ip_dst);
		ic2->icmp_id = htons(alias_id);

		/* Un-alias dsid */
		phdr->ph_dsid = tb_icmp->dsid;

		return (SUCCESS);
	}
	else{
		return (FWDENG_ICMP_OUT_ERRFLOWNOTFOUND);
	}
}

/*********************************************************************************
* Routine Name :  _rtl8651_l4IcmpNaptAliasOut3                                 
* Description :  
* Input :
* Output :
* Return :
* Note :
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4IcmpNaptAliasOut3(struct rtl_pktHdr *phdr, struct ip *pip,struct icmp *ic)
{	    	
	rtl8651_tblDrv_naptIcmpFlowEntry_t * icmp_tb;
	ipaddr_t ip_s, ip_d;
	uint16 id;
	
	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	id = ntohs(ic->icmp_id);

	icmp_tb = _rtl8651_getInsideNaptIcmpEntry(ip_s, id, ip_d, TRUE, TRUE);

	if(icmp_tb){
		ipaddr_t ip_alias;
		uint16 id_alias;

		ip_alias = icmp_tb->insideGlobalIpAddr;
		id_alias = icmp_tb->insideGlobalId;

		/* l4 checksum */
		if (((pip->ip_vhl & 0x0f) != 5) || (ntohs(pip->ip_off) & IP_MF))
		{
			if (((pip->ip_vhl) & 0x0f) != 5)
			{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
				/* for fragment packet, it don't need to calculate L3 checksum */
				_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, ip_s, pip->ip_sum);	/* L3 Checksum */
			}
			/* for both fragment packet and non-ASIC offloaded packet, we must calculate L4 checksum by our self */
			_RTL8651_L4_ADJUST_CHECKSUM(0, 0, id_alias, id, ic->icmp_cksum);	/* L4 Checksum: icmp don't include pseudo header */
		}
		#if 0
		if (ntohs(pip->ip_off) & IP_MF)
		{
			_RTL8651_L4_ADJUST_CHECKSUM( 0, 0, id_alias, id, ic->icmp_cksum);
		}
		#endif

		/* Alias sequence number */
		ic->icmp_id = htons(id_alias);

		/* Change source address */		
		PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);

		/* get dsid infomation */
		phdr->ph_dsid = icmp_tb->dsid;

		return (SUCCESS);
	}	
	return (FWDENG_ICMP_OUT_INVALIDIPID);
}


/*********************************************************************************
* Routine Name :  _rtl8651_l4IcmpNaptAliasOut                                   
* Description :                         
*    icmp output packet process
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4IcmpNaptAliasOut(struct rtl_pktHdr *phdr, struct ip *pip)
{
	int32     iresult;
	struct icmp *ic;
	
	ic = (struct icmp *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));

	switch (ic->icmp_type)
	{
		case ICMP_ECHO:
		case ICMP_TSTAMP:
			iresult = _rtl8651_l4IcmpNaptAliasOut1(phdr, pip, ic);
			break;
		case ICMP_UNREACH:
		case ICMP_SOURCEQUENCH:
		case ICMP_TIMXCEED:
		case ICMP_PARAMPROB:
			iresult = _rtl8651_l4IcmpNaptAliasOut2(phdr, pip, ic);
			break;
		case ICMP_ECHOREPLY:
		case ICMP_TSTAMPREPLY:
			iresult = _rtl8651_l4IcmpNaptAliasOut3(phdr, pip, ic);
			break;
		default:
			return (FWDENG_ICMP_OUT_INVALIDIPID);
	}

	if (iresult == SUCCESS)
	{
#ifdef _RTL_NAPT_FRAG
		if(IS_FRAGMENT(pip->ip_off))
		{
			int32 ta;
			phdr->ph_proto = PKTHDR_IP;
			ta = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
			_rtl8651_updateFragCache(OUTBOUND_FRAG, phdr->ph_sip, phdr->ph_dip, ntohs(pip->ip_id), ta, phdr->ph_routeIdx, NULL, phdr->ph_dsid);
		}	
#endif /* _RTL_NAPT_FRAG */	
	}
	return (iresult);
}


/*********************************************************************************
* Routine Name : _rtl8651_l4GetDeltaAckIn                                    
* Description :         
* Input :                                                               
* Output :                                                             
* Return :     
*    delta:number of delta in host order
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static int32 _rtl8651_l4GetDeltaAckIn(struct ip * pip, struct tcphdr *tc, rtl8651_tblDrv_naptTcpUdpFlowEntry_t* tb)
{

	int32     i;
	int32     delta, ack_diff_min;
	uint32    ack;

	ack = PKTGET_UINT32_UNALIGNED(&tc->th_ack);

	delta = 0;
	ack_diff_min = -1;
	for (i = 0; i < L4_N_LINK_TCP_DATA; i++)
	{
		struct l4ack_data_record x;

		rtl8651_memcpy(&x,&tb->ack[i],sizeof(struct l4ack_data_record));
		
		if (x.active == 1)
		{
			int32     ack_diff;

			ack_diff = _rtl8651_l4SeqDiff(x.ack_new, ack);
			if (ack_diff >= 0)
			{
				if (ack_diff_min >= 0)
				{
					if (ack_diff < ack_diff_min)
					{
						delta = x.delta;
						ack_diff_min = ack_diff;
					}
				}
				else
				{
					delta = x.delta;
					ack_diff_min = ack_diff;
				}
			}
		}
	}
	return (delta);
}


/*********************************************************************************
* Routine Name : _rtl8651_l4UdpNaptAliasIn                                     
* Description :  udp packet input process                                               
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4UdpNaptAliasIn(struct rtl_pktHdr *phdr, struct ip *pip)
{
	struct udphdr *ud;
	ipaddr_t ip_s, ip_d, dmzhost;
	ipaddr_t ip_org = 0;
	uint16 ip_sp, ip_dp, ip_orgp = 0;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb = &entry;
	rtl8651_tblDrv_algEntry_t *client_alg_link = NULL;
	rtl8651_tblDrv_algEntry_t *server_alg_link = NULL;
	rtl8651_tblDrv_naptServerPortEntry_t *tb2;
	uint32 actionFlag = 0,i,inTrigger=0;
	rtl8651_tblDrv_arpEntry_t *arp_t;
	int32 retval = FAILED;

	ud = (struct udphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));

	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(ud->uh_sport);
	ip_dp =  ntohs(ud->uh_dport);	

	//if tb == &entry, then this entry is a dynamic entry just learnt from ASIC
	//if tb != entry, then this is a already existed entry.
	
	if(_rtl8651_getInboundNaptFlow(FALSE,ip_d,ip_dp,ip_s,ip_sp,  &tb)!=SUCCESS){
 		tb=NULL;
 	}else
  	{	/* protocol stack cache checking */
		if (tb->toProtocolStack == TRUE)
		{	/* update timer and trap */
			if (ip_dp != tb->insideLocalPort)
			{
				ud->uh_dport = htons(tb->insideLocalPort);
				if (ud->uh_sum)
					_RTL8651_L4_ADJUST_CHECKSUM(0, 0, tb->insideLocalPort, ip_dp, ud->uh_sum);
			}
			_rtl8651_doProtoStackAction(pip);
			_rtl8651_updateNaptConnection(tb, UDP_FLOW, rtl8651GlobalControl.udpTimeout);
			return FWDENG_TO_GW;
		}
 	}

	tb2 = _rtl8651_getOutsideNaptServerPortMapping(FALSE,ip_d,ip_dp);
	// MAC Filtering
	if (tb2 && (arp_t=_rtl8651_getArpEntry(NULL, tb2->localIp, FALSE))) {
		if (_rtl8651_isMacAccessControlPermit(&arp_t->macInfo->macAddr) == FALSE)
			return FWDENG_MACBLOCK;
	}


	client_alg_link = _rtl8651_l4FindLinkAlg(ip_sp, FALSE);
	if(ip_sp!=ip_dp){
		server_alg_link = _rtl8651_l4FindLinkAlg(ip_dp, FALSE);
	}

	if(tb2 && server_alg_link && tb == NULL){
		tb=&entry;
		if(_rtl8651_getOutboundAlgCtrlFlow(FALSE,tb2->localIp,ip_dp-tb2->globalPort+tb2->localPort,ip_s,ip_sp, &tb)!=SUCCESS){
			tb=NULL;			
		}
	}


	if (tb == NULL) {
		/* 
			Now, we are going to establish a new flow, 
			but we have to check the ingress ACL of "L4NEWFLOW" type first.
		*/
		int32 local_val;
		rtl8651_tblDrv_vlanTable_t *local_vlanp = &DrvTbl.vlan[phdr->ph_vlanIdx];
		local_val = _rtl8651_ingressAcl(phdr->ph_dsid, 0, local_vlanp, local_vlanp->vid, phdr->ph_mbuf->m_data, (uint8 *) pip, RTL8651_ACL_L4NEWFLOW, _RTL8651_OPERATION_OTHER);
		if (((local_val > 0) && (local_val & FWDENG_DROP)) ||(local_val == FWDENG_ACL_IN_USERTRAP))
			return local_val;
	}

	
#define UDP_IN_UDP_ENTRY 0x01
#define UDP_IN_SERVER_PORT_ENTRY 0x02
#define UDP_IN_CLIENT_ALG 0x04
#define UDP_IN_SERVER_ALG	0x08

	if(tb)
		actionFlag |= UDP_IN_UDP_ENTRY;	
	
	if(tb2)
		actionFlag |= UDP_IN_SERVER_PORT_ENTRY;

#ifdef _RTL_TRIGGER_PORT
	if(actionFlag==0) //not UDP_IN_UDP_ENTRY and not UDP_IN_SERVER_PORT_ENTRY
	{
		if (_rtl8651_getTriggeredIpAddr(	FALSE,
										phdr->ph_dsid,
										0,			/* we don't include extIP as parameter in our trigger port check yet. */
										ip_dp,
										&ip_org) == SUCCESS)
		{
			if ((retval = _rtl8651_addNaptConnection(	TRUE,
													(TRUE|DOS_NAPT_ENCOUNT),
													UDP_FLOW,
													ip_org, ip_dp, & ip_d, &ip_dp, ip_s, ip_sp)) != SUCCESS)
			{
				return FWDENG_TCP_NAPT_IN_ADDFAIL;
			}
			tb = &entry;
			if (_rtl8651_getInboundNaptFlow(FALSE, ip_d, ip_dp, ip_s, ip_sp, &tb) != SUCCESS)
			{
				assert(0);
				return FWDENG_UDP_NAPT_IN_NOTFOUND;
			}
			inTrigger = 1;
		}
	}
#endif	

	if(inTrigger==0)
	{
		if(client_alg_link)
			actionFlag |= UDP_IN_CLIENT_ALG;
		
		if(server_alg_link)
			actionFlag |= UDP_IN_SERVER_ALG;
	}

#ifdef _TONY_DEBUG
{
	uint32 sip=GET_UINT32_UNALIGNED(&pip->ip_src);
	uint32 dip=GET_UINT32_UNALIGNED(&pip->ip_dst);	
	rtlglue_printf("l4UdpNaptAliasIn src:%u.%u.%u.%u:%d dest:%u.%u.%u.%u:%d actionFlag=%d\n",NIPQUAD(sip),ip_sp,NIPQUAD(dip),ip_dp,actionFlag);
}
#endif

	switch(actionFlag){
		case 0:
		case 4:
		case 8:
		case 12:
			if(inTrigger==1) break;
			/* client and server not found */
			/* but maybe UPNP incoming connection */
#ifdef _RTL_UPNP	
			{
			/* UPNP TCP port Mapping */		
			ipaddr_t local_ip = 0;
			uint16 local_port = 0;
			uint32 flags=0;
			uint32 algIdx=0;

			if (_rtl8651_findInboundUpnpMapping(FALSE, ip_s, ip_sp,ip_d, ip_dp, &local_ip, &local_port,&flags,&algIdx) != SUCCESS)
				goto newDmzHostConn;

			flags&=(UPNP_ALG_QOS);

			if(actionFlag==0)				
			{
				if((retval = _rtl8651_addAlgQosNaptConnection(TRUE,
						(TRUE|DOS_NAPT_ENCOUNT|flags),
						UDP_FLOW,local_ip, local_port,&ip_d, &ip_dp, ip_s, ip_sp,algIdx))
						!= SUCCESS)
				{
					return FWDENG_UDP_NAPT_IN_ADDFAIL;
				}

				tb=&entry;
				if(_rtl8651_getInboundNaptFlow(FALSE,ip_d,ip_dp,ip_s,ip_sp, &tb)!=SUCCESS)
				{
#ifdef _RTL_PS_ACTION			
					i=_rtl8651_doProtoStackAction(pip);
					if(i==PS_ACTION_DROP) return FWDENG_USER_DROP;		
#endif					
 					return FWDENG_UDP_NAPT_IN_NOTFOUND;
				}	
			}
 			else
 			{
				tb = _rtl8651_addAlgCtrlConnection(
						(TRUE|DOS_NAPT_ENCOUNT|flags),
						UDP_FLOW,local_ip,local_port,&ip_d, &ip_dp, ip_s, ip_sp);
				if(!tb)
				{
					return FWDENG_UDP_NAPT_IN_ALGADDFAIL;
				}
				if(tb)
				{
					tb->	tcpAckModified = 0;
					tb->	tcpFtpLastLineCRLFTermed = 0;
					tb->tcpStateINdex = 0;			
					memset(&tb->ack[0],0,sizeof(struct l4ack_data_record)*L4_N_LINK_TCP_DATA);				
				}
 			
				/* client alg process */
				if( (actionFlag & 4) && (client_alg_link->fwEngAdd) && (client_alg_link->alg_inboundClientcb)){
					client_alg_link->alg_inboundClientcb(phdr, pip, tb);
					tb->alg_client_in = client_alg_link->alg_inboundClientcb;
				}
				/* server alg process */
				if( (actionFlag & 8) && (server_alg_link->fwEngAdd) && (server_alg_link->alg_inboundServercb)){
					int32 r;					
					tb->alg_server_in = server_alg_link->alg_inboundServercb;
					r = server_alg_link->alg_inboundServercb(phdr, pip, tb);
					if(r==FWDENG_USER_DROP) {
						return FWDENG_USER_DROP;
					}
				}
 			}
			break; 				
 			
		}
#endif /* _RTL_UPNP */	
newDmzHostConn:


#ifdef _RTL_PS_ACTION			
			i=_rtl8651_doProtoStackAction(pip);
			if(i==PS_ACTION_DROP) return FWDENG_USER_DROP;
			if(i==PS_ACTION_TO_PS) return FWDENG_TCP_NAPT_IN_NOTFOUND;
#endif

		if (_rtl8651_getDmzHost(phdr->ph_dsid,ip_d, &dmzhost) == SUCCESS)
		{

			//if incoming packet not in entry, check the port whether the packet trap to protocol stack or not
#if 0			
			for(i=0;i<RTL8651_MAX_PROTO_STACK_USED_UDP_PORTS;i++)
			{				
				if(rtl8651_ProtoStackUsedUdpPorts[i]==0) 
				{
					break;
				}
				else if(rtl8651_ProtoStackUsedUdpPorts[i]==ntohs(ud->uh_sport)) 
				{
					//reply packet will be handled by protocol stack
					return FWDENG_UDP_NAPT_IN_NOTFOUND; // go to protocol stack
				}
			}
#endif

			
			if(_rtl8651_addNaptConnection(TRUE,
					(TRUE|DOS_NAPT_ENCOUNT), UDP_FLOW,
					dmzhost,ip_dp, &ip_d, &ip_dp, ip_s, ip_sp)
					!= SUCCESS)
			{
				return FWDENG_UDP_NAPT_IN_ADDFAIL;
			}
		
			tb=&entry;
			if(_rtl8651_getInboundNaptFlow(FALSE,ip_d,ip_dp,ip_s,ip_sp, &tb)!=SUCCESS) 
	 			return FWDENG_UDP_NAPT_IN_NOTFOUND;
			break;
		}else
			return FWDENG_UDP_NAPT_IN_NOTFOUND;
		
		case 3: //add on 2004/1/30
		case 1:
			/* entry already exist */
			if(tb->alg_client_in){
				int r;
				r=tb->alg_client_in(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}				
			}
			if(tb->alg_server_in){
				int r;
				r=tb->alg_server_in(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}
			break;
#if 0
		case 2:		
			/* server port without alg */			
			ud->th_dport = ip_dp-tb2->globalPort+tb2->localPort;
			PKTSET_UINT32_UNALIGNED(tb2->localIp, &pip->ip_dst);
			return (SUCCESS);	
#else			
//tony: new case 2 path, add a server port connection.
		case 2:

			if(_rtl8651_addNaptConnection(TRUE,
					(TRUE|DOS_NAPT_ENCOUNT),
					UDP_FLOW,tb2->localIp,ip_dp-tb2->globalPort+tb2->localPort, &ip_d, &ip_dp, ip_s, ip_sp)
					!= SUCCESS)
			{
				return FWDENG_UDP_NAPT_IN_ADDFAIL;
			}

			tb=&entry;
			
			if(_rtl8651_getInboundNaptFlow(FALSE,ip_d,ip_dp,ip_s,ip_sp, &tb)!=SUCCESS) 
	 			return FWDENG_UDP_NAPT_IN_NOTFOUND;
			break;
#endif	
		case 6:
		case 10:
			tb = _rtl8651_addAlgCtrlConnection((TRUE|DOS_NAPT_ENCOUNT),
				UDP_FLOW,tb2->localIp,ip_dp-tb2->globalPort+tb2->localPort,&ip_d, &ip_dp, ip_s, ip_sp);
			if(!tb)
			{
				return FWDENG_UDP_NAPT_IN_ALGADDFAIL;
			}
			if(tb)
			{
				tb->	tcpAckModified = 0;
				tb->	tcpFtpLastLineCRLFTermed = 0;
				tb->tcpStateINdex = 0;			
				memset(&tb->ack[0],0,sizeof(struct l4ack_data_record)*L4_N_LINK_TCP_DATA);				
			}
			break;
	
		case 7: //add on 2004/01/30
		case 5:		
			/* client alg process */
			if(client_alg_link->fwEngAdd && client_alg_link->alg_inboundClientcb){
				client_alg_link->alg_inboundClientcb(phdr, pip, tb);							
			}		
			break;
		case 11:
		case 9:			
			/* server alg process */
			 if(server_alg_link->fwEngAdd && server_alg_link->alg_inboundServercb){
				server_alg_link->alg_inboundServercb(phdr, pip, tb);							
		 	}			
			break;
		case 15: //add on 2004/01/30			
		case 13:
			if( (actionFlag & 4) && (client_alg_link->fwEngAdd) && (client_alg_link->alg_inboundClientcb)){
				client_alg_link->alg_inboundClientcb(phdr, pip, tb);					
			}
			/* server alg process */	
			if( (actionFlag & 8) && (server_alg_link->fwEngAdd) && (server_alg_link->alg_inboundServercb)){	
				server_alg_link->alg_inboundServercb(phdr, pip, tb);
			}
			break;

		default:			
			rtlglue_printf("%s: action=%d\n",__FUNCTION__,actionFlag );	
			return FWDENG_UDP_NAPT_IN_IGNORE;
		
	}
	assert(tb);

	/* update cache */
	tb->dsid = phdr->ph_dsid;

	// update timer
	_rtl8651_updateNaptConnection(tb, UDP_FLOW, rtl8651GlobalControl.udpTimeout);

	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(ud->uh_sport);
	ip_dp =  ntohs(ud->uh_dport);	
		
	ip_org = tb->insideLocalIpAddr;
	ip_orgp = tb->insideLocalPort;

	/********************************************************************************************/
	if ((!(actionFlag&UDP_IN_UDP_ENTRY)) && tb)
	{
		assert(tb->dosEnCount == TRUE && tb->dosFromInternal == FALSE);
	
		/*  chenyl: 2004/03/08 log for newly-added connection */
		if ((_loggingModule[phdr->ph_dsid]&RTL8651_LOGMODULE_NEWFLOW)&&_pUserLoggingFunc)
		{
#ifdef _RTL_NEW_LOGGING_MODEL
			rtl8651_logInfo_t info;
			info.infoType		= RTL8651_LOG_INFO_PKT;
			info.pkt_dsid		= phdr->ph_dsid;
			info.pkt_sip		= ip_s;
			info.pkt_dip		= ip_org;
			info.pkt_proto		= IPPROTO_UDP;
			info.pkt_direction	= 1;	/* direction: WAN to LAN */
			info.pkt_sport		= ip_sp;
			info.pkt_dport		= ip_orgp;
			info.action		= RTL8651_LOGACTION_FORWARD;
			(* _pUserLoggingFunc)(RTL8651_LOGMODULE_NEWFLOW, RTL8651_LOG_NEWFLOW_NewUdpNaptInbound, &info);
#else
			(*_pUserLoggingFunc)(
				phdr->ph_dsid,
				RTL8651_LOGMODULE_NEWFLOW,
				    (unsigned char)pip->ip_p, /* protocol */
				    (unsigned char)1, /* direction: WAN to LAN */
					(unsigned long)ip_s,
					(unsigned long)ip_org,
					(unsigned short)ip_sp,
					(unsigned short)ip_orgp,
					0, /* type: only meaningful for ICMP */
					RTL8651_LOGACTION_FORWARD, /* action */
					"New UDP Inbound Flow" /* msg */);
#endif
		}
	}
	/********************************************************************************************/

#ifdef _RTL_NAPT_FRAG
	if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK)) == IP_MF))
	{
		if ((pip->ip_vhl & 0x0f) != 5)
		{	/* ASIC cannot offload checksum if ip header length is not 20, so we must calculate L3/L4 checksum by ourself */
			_RTL8651_L3_ADJUST_CHECKSUM(ip_org, ip_d, pip->ip_sum);
		}
		/* for both first fragment packet and non-offloaded packet, we must calculate their L4 checksum here */
		if (ud->uh_sum)
		{
			_RTL8651_L4_ADJUST_CHECKSUM(ip_org, ip_d, ip_orgp, ip_dp, ud->uh_sum);
		}
		/* update fragment cache only if packet is the first fragment packet */
		if (IS_FRAGMENT(pip->ip_off))
		{
			phdr->ph_proto = PKTHDR_IP;
			_rtl8651_updateFragCache(INBOUND_FRAG, phdr->ph_sip, phdr->ph_dip, ntohs(pip->ip_id), ip_org, 0, NULL, phdr->ph_dsid);
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
			_rtl8651_fragmentShortestPathCacheSet_firstHalf(phdr, (uint8*)ud, pip, ip_s, ip_d, RTL8651_FRAGMENT_SHORTESTPATH_MODE_MODIFYDIP, ip_org);
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
		}
	}
#endif /* _RTL_NAPT_FRAG */	

	/* Put original port number back in */
	ud->uh_dport = htons(ip_orgp);
	/* Put original ip back in */		
	PKTSET_UINT32_UNALIGNED(ip_org, &pip->ip_dst);

	return (SUCCESS);
} 


/*********************************************************************************
* Routine Name : _rtl8651_l4UdpNatAlgAliasIn                                     
* Description :  udp packet input process                                               
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4UdpNatAlgAliasIn(struct rtl_pktHdr *phdr, struct ip *pip, rtl8651_tblDrv_natEntry_t *natEntry)
{
	struct udphdr *ud;
	ipaddr_t ip_s, ip_d, ip_org = 0;
	uint16 ip_sp, ip_dp, ip_orgp = 0;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb = &entry;
	rtl8651_tblDrv_algEntry_t *client_alg_link = NULL;
	rtl8651_tblDrv_algEntry_t *server_alg_link = NULL;
	uint32 actionFlag = 0;

	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;

	if (ip_d == natEntry->localIp)
	{	/* NI == NE, no ALG is needed */
		return FWDENG_UDP_NAT_IN_ALGNOTFOUND;
	}

	ud = (struct udphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));

	ip_sp = ntohs(ud->uh_sport);
	ip_dp =  ntohs(ud->uh_dport);	
	
	client_alg_link = _rtl8651_l4FindLinkAlg(ip_sp, FALSE);
	if(ip_sp != ip_dp){
		server_alg_link = _rtl8651_l4FindLinkAlg(ip_dp, FALSE);	
	}

	if(client_alg_link == NULL && server_alg_link == NULL){
		return FWDENG_UDP_NAT_IN_ALGNOTFOUND;
	}

	
	if(_rtl8651_getInboundNaptFlow(FALSE,ip_d,ip_dp,ip_s,ip_sp,  &tb)!=SUCCESS){
 		tb=NULL;
 	}

	
#define UDP_IN_UDP_ENTRY 0x01
#define UDP_IN_SERVER_PORT_ENTRY 0x02
#define UDP_IN_CLIENT_ALG 0x04
#define UDP_IN_SERVER_ALG	0x08

	if(tb)
		actionFlag |= UDP_IN_UDP_ENTRY;	
		 
	if(client_alg_link)
		actionFlag |= UDP_IN_CLIENT_ALG;
		
	if(server_alg_link)
		actionFlag |= UDP_IN_SERVER_ALG;

#ifdef _TONY_DEBUG
		rtlglue_printf("l4UdpNatAlgAliasIn src:%u.%u.%u.%u:%d dest:%u.%u.%u.%u:%d actionFlag=%d\n",NIPQUAD(ip_s),ip_sp,NIPQUAD(ip_d),ip_dp,actionFlag);
#endif

	switch(actionFlag){		
		case 1:
			/* entry already exist */
			if(tb->alg_client_in){
				int r;
				r=tb->alg_client_in(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}
			if(tb->alg_server_in){
				int r;
				r=tb->alg_server_in(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}			
			break;
		case 4:	
			/* first client alg process */
			tb = _rtl8651_addAlgCtrlConnection(
					(TRUE|DOS_NAPT_ENCOUNT),
					UDP_FLOW,natEntry->localIp, ip_dp, &ip_d, &ip_dp, ip_s, ip_sp);
			if(!tb)
				return FWDENG_UDP_NAT_IN_CLIALGADDFAIL;
			if(client_alg_link->fwEngAdd && client_alg_link->alg_inboundClientcb){
				tb->natlink = 1;
				client_alg_link->alg_inboundClientcb(phdr, pip, tb);							
			}
			break;			
		case 8:
			/* first ALG Server process */
			tb = _rtl8651_addAlgCtrlConnection(
					(TRUE|DOS_NAPT_ENCOUNT),
					UDP_FLOW,natEntry->localIp, ip_dp, &ip_d, &ip_dp, ip_s, ip_sp);
			if(!tb)
				return FWDENG_UDP_NAT_IN_SVRALGADDFAIL;
			 if( server_alg_link->fwEngAdd && server_alg_link->alg_outboundServercb){
			 	int r;
				tb->natlink = 1;
				r=server_alg_link->alg_outboundServercb(phdr, pip, tb);							
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			 }			
			break;
		case 5:		
			/* client alg process */
			if(client_alg_link->fwEngAdd && client_alg_link->alg_inboundClientcb){
				client_alg_link->alg_inboundClientcb(phdr, pip, tb);							
			}		
			break;
		case 9:		
			 /*ALG Server process*/
			 if(server_alg_link->fwEngAdd && server_alg_link->alg_inboundServercb){
				server_alg_link->alg_inboundServercb(phdr, pip, tb);							
			 }
			break;		
			
		default:			
			rtlglue_printf("%s: action=%d\n",__FUNCTION__,actionFlag);	
			return FWDENG_UDP_NAT_IN_ALGIGNORE;
	}

	assert(tb);

	// update timer
	_rtl8651_updateNaptConnection(tb, UDP_FLOW, rtl8651GlobalControl.udpTimeout);
	
	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(ud->uh_sport);
	ip_dp =  ntohs(ud->uh_dport);	
		
	ip_org = tb->insideLocalIpAddr;
	ip_orgp = tb->insideLocalPort;


#ifdef _RTL_NAPT_FRAG
	if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK)) == IP_MF))
	{
		if ((pip->ip_vhl & 0x0f) != 5)
		{
			_RTL8651_L3_ADJUST_CHECKSUM(ip_org, ip_d, pip->ip_sum);
		}
		if(ud->uh_sum)
		{
			_RTL8651_L4_ADJUST_CHECKSUM(ip_org, ip_d, ip_orgp, ip_dp, ud->uh_sum);
		}
	}
#endif /* _RTL_NAPT_FRAG */	

	/* Put original port number back in */
	ud->uh_dport = htons(ip_orgp);
	/* Put original ip back in */	
	PKTSET_UINT32_UNALIGNED(ip_org, &pip->ip_dst);
	
	return (SUCCESS);
} 


/*********************************************************************************
* Routine Name :  _rtl8651_l4UdpNaptAliasOut                                   
* Description :  
*    udp packet output process
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4UdpNaptAliasOut(struct rtl_pktHdr *phdr, struct ip *pip)
{
	struct udphdr *ud;
	ipaddr_t ip_s, ip_d, ip_alias = 0;
	uint16 ip_sp, ip_dp, ip_aliasp = 0;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb=&entry;
	rtl8651_tblDrv_algEntry_t *client_alg_link = NULL; 
	rtl8651_tblDrv_algEntry_t *server_alg_link = NULL; 
	//int32 accumulate;
	rtl8651_tblDrv_naptServerPortEntry_t *tb2;

	/*
		for outbound uPnP mapping:

			When outbound flow which match uPnP mapping,
			we will treat it same as Server Port process.
			And some notes are as below:

			1. Due to we treat this as a upnp mapping match process, therefore,
			    for One-Shot uPnP mapping, the uPnP entry will be removed after gotten its information.
			2. The check priority of NAPT outbound traffic:
				1) Server Port
				2) uPnP Mapping
				3) NAPT-redirect

																	chenyl 2005/05/06
	*/
	int32 isUpnpMapping = FALSE;
	ipaddr_t upnpExtIp;
	uint16 upnpExtPort;
	uint32 upnpFlags;
	uint32 upnpAlgIdx;

	uint32 actionFlag = 0;
	int inTrigger = 0;

	ud = (struct udphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));

	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(ud->uh_sport);
	ip_dp =  ntohs(ud->uh_dport);

	/*
		If this flow exist?
	*/
	if( _rtl8651_getOutboundNaptFlow(FALSE,ip_s,ip_sp,ip_d,ip_dp, &tb) != SUCCESS )
	{
		tb = &entry;
		if( _rtl8651_getOutboundAlgCtrlFlow(FALSE,ip_s,ip_sp,ip_d,ip_dp, &tb) != SUCCESS )
		{
			tb = NULL;
		}
	}

	/*
		If this flow is server port?
	*/
	tb2 = _rtl8651_getInsideNaptServerPortMapping(FALSE,ip_s,ip_sp);

	/*
		Check for client / server ALG
	*/
	client_alg_link = _rtl8651_l4FindLinkAlg(ip_dp, FALSE);

	if(ip_dp != ip_sp)
	{
		if( tb )
		{
			server_alg_link = _rtl8651_l4FindLinkAlg(tb->insideGlobalPort, FALSE);
		}else
		{
			server_alg_link = _rtl8651_l4FindLinkAlg(ip_sp, FALSE);
		}
	}

	/*
		If this packet:

			1. Have no entry for it.
			2. It's not in any server port mapping.

		We must check if it in upnp mapping or not, if this packet can fit any upnp mapping,
		we would use the corresponding external Ip/Port to modify Ip/Port.
	*/
	if ((tb == NULL) && (tb2 == NULL))
	{
		if (_rtl8651_findOutboundUpnpMapping(
			FALSE,
			ip_s,
			ip_sp,
			&upnpExtIp,
			&upnpExtPort,
			ip_d,
			ip_dp,
			&upnpFlags,
			&upnpAlgIdx) == SUCCESS)
		{
			isUpnpMapping = TRUE;
		}
	}

#ifdef _RTL_NAPT_REDIRECT
		/*
			For
				1. non-server port
				2. no-entry
				3.  non-upnp mapping

				packets, we must check napt redirect database and decide if we need to trap it or not
		*/
		if ((tb == NULL) && (tb2 == NULL) && (isUpnpMapping == FALSE))
		{
			ipaddr_t redirectIp;
			uint16 redirectPort;
			if (_rtl8651_findNaptRedirectEntry(FALSE, ip_s, ip_sp, ip_d, ip_dp, &redirectIp, &redirectPort, TRUE, phdr->ph_vlanIdx) == SUCCESS)
			{	/* this entry must be trapped to protocol stack for further process */
				/* modify dip/dport */
				PKTSET_UINT32_UNALIGNED(redirectIp, &pip->ip_dst);
				ud->uh_dport = htons(redirectPort);
				/* always modify L3/L4 checksum */
				if (ud->uh_sum)
					_RTL8651_L4_ADJUST_CHECKSUM(redirectIp, ip_d, redirectPort, ip_dp, ud->uh_sum);
				_RTL8651_L3_ADJUST_CHECKSUM(redirectIp, ip_d, pip->ip_sum);
				/* update fragment cache */
				if (IS_FRAGMENT_FIRST(pip->ip_off))
				{
					_rtl8651_updateFragCache(NAPT_REDIRECT_FRAG, ip_s, ip_d, ntohs(pip->ip_id), redirectIp, 0, NULL, phdr->ph_dsid);
				}
#ifdef _RTL_QUEUE_FRAG_PKT
				if (_rtl8651_fragQueueProcess == TRUE)
				{	/* if system is under fragment-dequeue-phase, we must store redirectIP for remaining trapped fragment packet */
					_rtl8651_fragQueueNaptRedirect_MatchedIP = redirectIp;
				}
#endif /* _RTL_QUEUE_FRAG_PKT */
				return FWDENG_TO_GW;
			}
		}
#endif /* _RTL_NAPT_REDIRECT */ 

	if (tb == NULL) {
		/* 
			Now, we are going to establish a new flow, 
			but we have to check the ingress ACL of "L4NEWFLOW" type first.
		*/
		int32 local_val;
		rtl8651_tblDrv_vlanTable_t *local_vlanp = &DrvTbl.vlan[phdr->ph_vlanIdx];
		local_val = _rtl8651_ingressAcl(phdr->ph_dsid, 0, local_vlanp, local_vlanp->vid, phdr->ph_mbuf->m_data, (uint8 *) pip, RTL8651_ACL_L4NEWFLOW, _RTL8651_OPERATION_OTHER);
		if (((local_val > 0) && (local_val & FWDENG_DROP)) ||(local_val == FWDENG_ACL_IN_USERTRAP))
			return local_val;
	}

#define UDP_OUT_UDP_ENTRY				0x01
#define UDP_OUT_SERVER_PORT_ENTRY	0x02
#define UDP_OUT_CLIENT_ALG				0x04
#define UDP_OUT_SERVER_ALG			0x08
#define UDP_OUT_UPNP_ENTRY			0x10

	if(tb)
	{
		actionFlag |= UDP_OUT_UDP_ENTRY;
	}
	
	if(tb2)
	{
		actionFlag |= UDP_OUT_SERVER_PORT_ENTRY;
	}

	if (isUpnpMapping == TRUE)
	{
		actionFlag |= UDP_OUT_UPNP_ENTRY;
	}


#ifdef _RTL_TRIGGER_PORT
	if ( actionFlag==0 ) /* not UDP_Entry and not Server_Port_Entry */
	{
		if (_rtl8651_tryActiveTriggerPort(	FALSE,
										phdr->ph_dsid,
										ip_dp,
										0,				/* we don't include extIP as parameter in our trigger port check yet. */
										ip_s) > 0)
		{
			inTrigger = 1;
		}
	}
#endif	

	/* tony: when Special AP Trigger Port is a ALG port, don't do ALG. */
	if(inTrigger==0)
	{
		if(client_alg_link)
		{
			actionFlag |= UDP_OUT_CLIENT_ALG;
		}
		
		if(server_alg_link)
		{
			actionFlag |= UDP_OUT_SERVER_ALG;
		}
	}

#ifdef _TONY_DEBUG
{
	uint32 sip=GET_UINT32_UNALIGNED(&pip->ip_src);
	uint32 dip=GET_UINT32_UNALIGNED(&pip->ip_dst);	
	rtlglue_printf("l4UdpNaptAliasOut src:%u.%u.%u.%u:%d dest:%u.%u.%u.%u:%d actionFlag=%d\n",NIPQUAD(sip),ip_sp,NIPQUAD(dip),ip_dp,actionFlag);
}
#endif

	/**********************************************************************

		Process Dispatcher
	
	 **********************************************************************/

	switch ( actionFlag )
	{

		case 0:

		/*********************************************************************************/
			_rtl8651_softNaptPolicyRoute(TRIGGER_DSTUDPPORT,ip_s,ip_d,ip_dp,&ip_alias,&ip_aliasp);
			if (ip_alias)
			{
#ifdef _RTL_NAPT_ADVANCED
				/*	if static ip translation is enabled,
					we ignore the policy route assigned external port if we can use the original port */
				if ((rtl8651_drvStaticUdpNaptPortTranslation == TRUE) &&
					(_rtl8651_l4CheckOccupiedConnection(FALSE, phdr->ph_dsid, ip_s, ip_sp, ip_alias, ip_sp, ip_d, ip_dp) == FALSE))
				{
					/* try to add napt entry with external port is original source port */
					if (_rtl8651_addNaptConnection(TRUE,
							(TRUE|PURE_SW|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL|DO_REDUNDANCY_CHECK),
							UDP_FLOW, ip_s, ip_sp, &ip_alias, &ip_sp,
							(rtl8651_drvInexactUdpNaptEnable?0:ip_d),(rtl8651_drvInexactUdpNaptEnable?0:ip_dp))
							== SUCCESS)
						ip_aliasp = ip_sp;	/* add successful */
					else
						goto UdpPolicyDynamicPortAssign; 	/* add failed, might the external ip/port is used by other flow? goto dynamic mode */
				}else
				{
UdpPolicyDynamicPortAssign:
					if(_rtl8651_addNaptConnection(TRUE, 
							(TRUE|PURE_SW|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
							UDP_FLOW,ip_s,ip_sp, &ip_alias,&ip_aliasp,
							(rtl8651_drvInexactUdpNaptEnable?0:ip_d),(rtl8651_drvInexactUdpNaptEnable?0:ip_dp))
							!= SUCCESS)  
						return FWDENG_UDP_NAPT_OUT_ADDFAIL;
				}
#else
				if(_rtl8651_addNaptConnection(TRUE,
						(TRUE|PURE_SW|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
						UDP_FLOW, ip_s, ip_sp, &ip_alias, &ip_aliasp, ip_d, ip_dp) != SUCCESS)
					return FWDENG_UDP_NAPT_OUT_ADDFAIL;
#endif /* _RTL_NAPT_ADVANCED */
			}else
			{
			/* create a udp entey */
#ifdef _RTL_NAPT_ADVANCED
				if (rtl8651_drvStaticUdpNaptPortTranslation == TRUE)
				{
					ip_alias = _rtl8651_getNaptInsideGlobalIpAddr(FALSE, ip_s, ip_sp, ip_d, ip_dp);

					if ((!ip_alias) || (_rtl8651_l4CheckOccupiedConnection(FALSE, phdr->ph_dsid, ip_s, ip_sp, ip_alias, ip_sp, ip_d, ip_dp) == TRUE))
						goto UdpNormalDynamicPortAssign;

					if (_rtl8651_addNaptConnection(TRUE,
							(TRUE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL|DO_REDUNDANCY_CHECK),
							UDP_FLOW, ip_s, ip_sp, &ip_alias, &ip_sp,
							(rtl8651_drvInexactUdpNaptEnable?0:ip_d), (rtl8651_drvInexactUdpNaptEnable?0:ip_dp))
							== SUCCESS)
						ip_aliasp = ip_sp;	/* add successful */
					else
						goto UdpNormalDynamicPortAssign;	/* add failed, might the external ip/port is used by other flow? goto dynamic mode */
				}else
				{
UdpNormalDynamicPortAssign:
					if(_rtl8651_addNaptConnection(TRUE,
							(FALSE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
							UDP_FLOW, ip_s, ip_sp, &ip_alias, &ip_aliasp,
							(rtl8651_drvInexactUdpNaptEnable?0:ip_d), (rtl8651_drvInexactUdpNaptEnable?0:ip_dp))
							!= SUCCESS )
						return FWDENG_UDP_NAPT_OUT_ADDFAIL;
				}
#else
				if(_rtl8651_addNaptConnection(TRUE,
						(FALSE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
						UDP_FLOW, ip_s, ip_sp,
						&ip_alias, &ip_aliasp, ip_d, ip_dp) != SUCCESS)
				{
					return FWDENG_UDP_NAPT_OUT_ADDFAIL;
				}
#endif /* _RTL_NAPT_ADVANCED */
			}
			tb=&entry;
			if(_rtl8651_getOutboundNaptFlow(FALSE,ip_s,ip_sp,ip_d,ip_dp, &tb)!=SUCCESS)
			{
 				return FWDENG_UDP_NAPT_OUT_NOTFOUND;
 			}
			break;
		/*********************************************************************************/

		case 1:
		case 3: /* for server port */
		case 17:	/* chenyl : add for upnp */

		/*********************************************************************************/
			/* entry have already create */
			/* entry already exist */
			if(tb->alg_client_out){
				int r;
				r=tb->alg_client_out(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}
			if(tb->alg_server_out){
				int r;
				r=tb->alg_server_out(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}					
			break;

		/*********************************************************************************/

		/* for server port */
		case 2:
		case 10:	 /* server alg + server port add on 2004/01/30(in this case, server port is miscarriage of justice) */

		/* for upnp */
		case 16:	/* chenyl : add for upnp */
		case 24:	/* chenyl : add for upnp */

		/*********************************************************************************/

			if (tb2)
			{	/* normal server port operation */
				assert(isUpnpMapping == FALSE);

				ip_alias = tb2->globalIp;
				ip_aliasp = ip_sp - tb2->localPort + tb2->globalPort;
			}else
			{	/* upnp process */
				assert(isUpnpMapping == TRUE);

				ip_alias = upnpExtIp;
				ip_aliasp = upnpExtPort;
			}

			if(_rtl8651_addNaptConnection(TRUE,
					(TRUE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
					UDP_FLOW,ip_s,ip_sp,&ip_alias,&ip_aliasp,ip_d,ip_dp)!=SUCCESS)
			{
				goto ServerPortOut;
			}

			tb=&entry;
			if(_rtl8651_getOutboundNaptFlow(FALSE,ip_s,ip_sp,ip_d,ip_dp, &tb)!=SUCCESS){
				goto ServerPortOut;
 			}

			if (actionFlag & UDP_OUT_SERVER_ALG)
				goto serverAlg_entry_run;
			break;

		ServerPortOut:
			/*
				chenyl:
					We can not add napt flow for this connection, but we still can
					translate source IP/Port of this packet because we have known them.
			*/
			ud->uh_sport = htons(ip_aliasp);
			PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);

			return SUCCESS;

		/*********************************************************************************/

		case 6: //add on 2004/01/30
		case 4:

		/*
			chenyl:
				Because the upnp process must synchronous with what server port done,
				when the case is (UPNP + Client ALG) == 20,
				we must process this packet just like case 6 (Server Port + Client ALG).
		*/
		case 20:	/* chenyl: add for upnp */

		/*********************************************************************************/

			_rtl8651_softNaptPolicyRoute(TRIGGER_DSTUDPPORT,ip_s,ip_d,ip_dp,&ip_alias,&ip_aliasp);
			if (ip_alias)
			{
				tb = _rtl8651_addAlgCtrlConnection(
						(TRUE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
						UDP_FLOW,ip_s,ip_sp,&ip_alias,&ip_aliasp,ip_d,ip_dp);
			}else
			{	/*ALG Client check*/
				tb = _rtl8651_addAlgCtrlConnection(
					(FALSE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
					UDP_FLOW,ip_s,ip_sp,&ip_alias,&ip_aliasp,ip_d,ip_dp);
			}
			
			if(!tb)
				return FWDENG_UDP_NAPT_OUT_CLIALGADDFAIL;			 
			if ( client_alg_link->fwEngAdd && client_alg_link->alg_outboundClientcb){
				int r;
				tb->alg_client_out=client_alg_link->alg_outboundClientcb;	
				r = client_alg_link->alg_outboundClientcb(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}
			break;

		/*********************************************************************************/
			
		case 7: /* add on 2004/01/30 */
		case 5:

		case 19:	/* chenyl: add for upnp, please reference the comment of case 20 */

		/*********************************************************************************/
			
			if ( client_alg_link->fwEngAdd && client_alg_link->alg_outboundClientcb){
				int r;
				tb->alg_client_out=client_alg_link->alg_outboundClientcb;
				r = client_alg_link->alg_outboundClientcb(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}				
			break;

		/*********************************************************************************/
	
		case 8: /* server alg */

		/*********************************************************************************/

			/* add server port alg link */
			tb = _rtl8651_addAlgCtrlConnection(
					(FALSE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
					UDP_FLOW,ip_s,ip_sp,&ip_alias,&ip_aliasp,ip_d,ip_dp);			
			if(!tb)
				return FWDENG_UDP_NAPT_OUT_SVRALGADDFAIL;			

			/* to be continue with case 9,11 */
		/*********************************************************************************/
			
		case 9: /* server alg, entry */
		case 11:	/* server alg, server port, entry, add on 2004/01/30 */

		/*********************************************************************************/

serverAlg_entry_run:
			 /*ALG Server process*/
			 if(server_alg_link->fwEngAdd && server_alg_link->alg_outboundServercb){
			 	int r;
			 	tb->alg_server_out=server_alg_link->alg_outboundServercb;	
				r=server_alg_link->alg_outboundServercb(phdr, pip, tb);							
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			 }
			break;

		/*********************************************************************************/
			
		case 12:/* server alg , client alg */
		case 14: /* server alg , client alg, server port, add on 2004/01/30(in this case, server port is miscarriage of justice) */

		case 28:	/* chenyl: add for upnp (server alg + client alg + upnp), please reference the comment of case 20 */

		/*********************************************************************************/
			tb = _rtl8651_addAlgCtrlConnection(
				(FALSE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
				UDP_FLOW,ip_s,ip_sp, &ip_alias,&ip_aliasp,ip_d,ip_dp);
			if(!tb)
				return FWDENG_UDP_NAPT_OUT_SVRALGADDFAIL;

		/* to be continue with case 13 */
		/*********************************************************************************/

		case 13:	 /* server alg , client alg, entry */
		case 15: /* server alg , client alg, entry , server port, add on 2004/01/30 */

		case 29:	/* chenyl: add for upnp (server alg + client alg + upnp + entry), please reference the comment of case 20 */

		/*********************************************************************************/
			if ( client_alg_link->fwEngAdd && client_alg_link->alg_outboundClientcb){
				int r;
				tb->alg_client_out=client_alg_link->alg_outboundClientcb;				
				r = client_alg_link->alg_outboundClientcb(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}
			if ( server_alg_link->fwEngAdd && server_alg_link->alg_outboundServercb){
				int r;
				tb->alg_server_out=server_alg_link->alg_outboundServercb;
				r = server_alg_link->alg_outboundServercb(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}
			break;

		/*********************************************************************************/

		default:			
			rtlglue_printf("%s: action=%d\n",__FUNCTION__,actionFlag );	
			return (FWDENG_UDP_NAPT_OUT_IGNORE);
	}

	assert(tb);

	/* update cache */	
	assert( phdr->ph_routeIdx);
	tb->out_rt_t = (void *)&DrvTbl.route[phdr->ph_routeIdx-1];
	tb->dsid = phdr->ph_dsid;

	// update timer
	_rtl8651_updateNaptConnection(tb, UDP_FLOW, rtl8651GlobalControl.udpTimeout);

	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(ud->uh_sport);
	ip_dp =  ntohs(ud->uh_dport);
	
	ip_alias = tb->insideGlobalIpAddr;
	ip_aliasp = tb->insideGlobalPort;

	/****************************************************************************************/
	if ((!(actionFlag&UDP_OUT_UDP_ENTRY)) && tb)
	{
		assert(tb->dosEnCount == TRUE && tb->dosFromInternal == TRUE);
	
		// chenyl: 2004/03/08 log for newly-added connection	
		if ((_loggingModule[phdr->ph_dsid]&RTL8651_LOGMODULE_NEWFLOW)&&_pUserLoggingFunc)
		{
#ifdef _RTL_NEW_LOGGING_MODEL
			rtl8651_logInfo_t info;

			info.infoType		= RTL8651_LOG_INFO_PKT;
			info.pkt_dsid		= phdr->ph_dsid;
			info.pkt_sip		= ip_s;
			info.pkt_dip		= ip_d;
			info.pkt_proto		= IPPROTO_UDP;
			info.pkt_direction	= 0;	/* direction: LAN to WAN */
			info.pkt_sport		= ip_sp;
			info.pkt_dport		= ip_dp;
			info.action		= RTL8651_LOGACTION_FORWARD;
			(* _pUserLoggingFunc)(RTL8651_LOGMODULE_NEWFLOW, RTL8651_LOG_NEWFLOW_NewUdpNaptOutbound, &info);
#else
			(*_pUserLoggingFunc)(
				phdr->ph_dsid,
				RTL8651_LOGMODULE_NEWFLOW,
				    (unsigned char)pip->ip_p, /* protocol */
				    (unsigned char)0, /* direction: LAN to WAN */
					(unsigned long)ip_s,
					(unsigned long)ip_d,
					(unsigned short)ip_sp,
					(unsigned short)ip_dp,
					0, /* type: only meaningful for ICMP */
					RTL8651_LOGACTION_FORWARD, /* action */
					"New UDP Outbound Flow" /* msg */);
#endif
		}
	}
	/****************************************************************************************/

#ifdef _RTL_NAPT_FRAG	
	if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK)) == IP_MF))
	{
		if ((pip->ip_vhl & 0x0f) != 5)
		{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
			/* for fragment packet, it don't need to calculate L3 checksum */
			_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, ip_s, pip->ip_sum);	/* L3 Checksum */
		}
		/* for both fragment packet and non-ASIC offloaded packet, we must calculate L4 checksum by our self */
		if( ud->uh_sum)
		{
			_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, ip_s, ip_aliasp, ip_sp, ud->uh_sum);	/* L4 Checksum */
		}

		if ((ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK)) == IP_MF)
		{	/* first packet of fragment flow */
			phdr->ph_proto = PKTHDR_IP;
			_rtl8651_updateFragCache(OUTBOUND_FRAG, ip_s, ip_d, ntohs(pip->ip_id), ip_alias, phdr->ph_routeIdx, NULL, phdr->ph_dsid);
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
			_rtl8651_fragmentShortestPathCacheSet_firstHalf(phdr, (uint8*)ud, pip, ip_s, ip_d, RTL8651_FRAGMENT_SHORTESTPATH_MODE_MODIFYSIP, ip_alias);
#endif /*RTL8651_FRAGMENT_SHORTESTPATH_FWD */
		}
	}

#endif /* _RTL_NAPT_FRAG */	

	/* Put alias port in UDP header */
	ud->uh_sport = htons(ip_aliasp);

	/* Change source address */
	PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);

	return (SUCCESS);
}



/*********************************************************************************
* Routine Name :  _rtl8651_l4UdpNatAlgAliasOut                                   
* Description :  
*    udp packet output process
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4UdpNatAlgAliasOut(struct rtl_pktHdr *phdr, struct ip *pip, rtl8651_tblDrv_natEntry_t *natEntry)
{
	struct udphdr *ud;
	ipaddr_t ip_s, ip_d, ip_alias = 0;
	uint16 ip_sp, ip_dp, ip_aliasp = 0;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb=&entry;
	rtl8651_tblDrv_algEntry_t *client_alg_link = NULL; 
	rtl8651_tblDrv_algEntry_t *server_alg_link = NULL; 
	//int32 accumulate;
	uint32 actionFlag = 0;

	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;

	if (ip_s == natEntry->globalIp)
	{	/* NI == NE, no ALG is needed */
		return FWDENG_UDP_NAT_OUT_ALGNOTFOUND;
	}

	ud = (struct udphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));

	ip_sp = ntohs(ud->uh_sport);
	ip_dp =  ntohs(ud->uh_dport);

	client_alg_link = _rtl8651_l4FindLinkAlg(ip_dp, FALSE);		
	if(ip_dp != ip_sp){
		server_alg_link = _rtl8651_l4FindLinkAlg(ip_sp, FALSE);
	}

	if(client_alg_link == NULL && server_alg_link == NULL){
		return FWDENG_UDP_NAT_OUT_ALGNOTFOUND;
	}
	
	if(_rtl8651_getOutboundAlgCtrlFlow(FALSE,ip_s,ip_sp,ip_d,ip_dp, &tb)!=SUCCESS){
			tb=NULL;
	}
	

#define UDP_OUT_UDP_ENTRY 0x01
#define UDP_OUT_SERVER_PORT_ENTRY 0x02
#define UDP_OUT_CLIENT_ALG 0x04
#define UDP_OUT_SERVER_ALG 0x08

	if(tb)
		actionFlag |= UDP_OUT_UDP_ENTRY;	
		 
	if(client_alg_link)
		actionFlag |= UDP_OUT_CLIENT_ALG;
		
	if(server_alg_link)
		actionFlag |= UDP_OUT_SERVER_ALG;

#ifdef _TONY_DEBUG
		rtlglue_printf("l4UdpNaptAlgAliasOut src:%u.%u.%u.%u:%d dest:%u.%u.%u.%u:%d actionFlag=%d\n",NIPQUAD(ip_s),ip_sp,NIPQUAD(ip_d),ip_dp,actionFlag);
#endif	

	switch(actionFlag){
		case 1:
			/* entry already exist */
			if(tb->alg_client_out){
				int r;
				r=tb->alg_client_out(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			}
			if(tb->alg_server_out){
				int r;
				r=tb->alg_server_out(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			}			
			break;		
		case 4:		
			 /* first ALG Client check*/
			 ip_alias = natEntry->globalIp;
			 ip_aliasp = ip_sp;
			tb = _rtl8651_addAlgCtrlConnection(
					(TRUE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
					UDP_FLOW,ip_s,ip_sp, &ip_alias,&ip_aliasp,ip_d,ip_dp);
			if(!tb)
				return FWDENG_UDP_NAT_OUT_CLIALGADDFAIL;
			if ( client_alg_link->fwEngAdd && client_alg_link->alg_outboundClientcb){
				int r;
				tb->natlink = 1;
				r=client_alg_link->alg_outboundClientcb(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			}		
			break;		
		case 5:		
			 /*ALG Client process*/
			if ( client_alg_link->fwEngAdd && client_alg_link->alg_outboundClientcb){
				int r;
				r=client_alg_link->alg_outboundClientcb(phdr, pip, tb);				
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			}				
			break;
		case 8:		
			 /* first ALG server check*/
			 ip_alias = natEntry->globalIp;
			 ip_aliasp = ip_sp;			 
			tb = _rtl8651_addAlgCtrlConnection(
					(TRUE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
					UDP_FLOW,ip_s,ip_sp, &ip_alias,&ip_aliasp,ip_d,ip_dp);
			if(!tb)
				return FWDENG_UDP_NAT_OUT_SVRALGADDFAIL;

			if ( server_alg_link->fwEngAdd && server_alg_link->alg_outboundServercb){							
				int r;
				tb->natlink = 1;
				r=server_alg_link->alg_outboundServercb(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			}		
			break;		
			
		case 9:		
			 /*ALG Server process*/
			 if(server_alg_link->fwEngAdd && server_alg_link->alg_outboundServercb){
			 	int r;
				r=server_alg_link->alg_outboundServercb(phdr, pip, tb);											
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			 }
			break;	

		default:		
			rtlglue_printf("%s: action=%d\n",__FUNCTION__,actionFlag );	
			return FWDENG_UDP_NAT_OUT_ALGIGNORE;
	}

	assert(tb);

	// update timer
	_rtl8651_updateNaptConnection(tb, UDP_FLOW, rtl8651GlobalControl.udpTimeout);
	
	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(ud->uh_sport);
	ip_dp =  ntohs(ud->uh_dport);
		
	ip_alias = tb->insideGlobalIpAddr;
	ip_aliasp = tb->insideGlobalPort;
						 
#ifdef _RTL_NAPT_FRAG
	if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF | IP_OFFMASK)) == IP_MF))
	{
		/* if ip header length != 20, we must calculate L3/4 checksum by ourself */
		if ((pip->ip_vhl & 0x0f) != 5)
		{
			_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, ip_s, pip->ip_sum);
		}
		/* L4 Checksum need to be calculated for first fragment packet and non-offloaded packet */
		if (ud->uh_sum)
		{
			_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, ip_s, ip_aliasp, ip_sp, ud->uh_sum);
		}
	}
#endif /* _RTL_NAPT_FRAG */	

	/* Put alias port in UDP header */
	ud->uh_sport = htons(ip_aliasp);

	/* Change source address */
	PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);

	return (SUCCESS);
}

/*********************************************************************************
* Routine Name : _rtl8651_l4TcpNaptAliasIn                                    
* Description :  TCP packet input process
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4TcpNaptAliasIn(struct rtl_pktHdr *phdr, struct ip *pip)
{
	struct tcphdr *tc;
	ipaddr_t ip_s, ip_d, ip_org = 0, dmzhost=0;
	uint16 ip_sp, ip_dp, ip_orgp = 0;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb = &entry;
	int32 accumulate = 0;
	rtl8651_tblDrv_algEntry_t *client_alg_link = NULL;
	rtl8651_tblDrv_algEntry_t *server_alg_link = NULL;
	rtl8651_tblDrv_naptServerPortEntry_t *tb2;
	uint32 actionFlag = 0,i,inTrigger=0;
	rtl8651_tblDrv_pppTable_t *thisPPPoE;
	rtl8651_tblDrv_arpEntry_t *arp_t;
	uint16 mtu;
	int8 initFlag = TCP_INBOUND_SYN_RCVD_FLOW;
	int32 retval = FAILED;


	tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
		
	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(tc->th_sport);
	ip_dp =  ntohs(tc->th_dport);

#define TCP_IN_TCP_SYN 0x01
#define TCP_IN_TCP_ENTRY 0x02
#define TCP_IN_SERVER_PORT_ENTRY 0x04
#define TCP_IN_CLIENT_ALG 0x08
#define TCP_IN_SERVER_ALG	0x10

	if(_rtl8651_getInboundNaptFlow(TRUE,ip_d,ip_dp,ip_s,ip_sp, &tb)!=SUCCESS){
 		tb=NULL;
 	}else
 	{	/* protocol stack cache checking */
		if (tb->toProtocolStack == TRUE)
		{	/* update tcp state and trap */
			if (ip_dp != tb->insideLocalPort)
			{
				tc->th_dport = htons(tb->insideLocalPort);
				_RTL8651_L4_ADJUST_CHECKSUM(0, 0, tb->insideLocalPort, ip_dp, tc->th_sum);
			}
			_rtl8651_doProtoStackAction(pip);
			_rtl8651_l4TcpMonitorIn(tc, tb);
			return FWDENG_TO_GW;
		}
 	}

	tb2 = _rtl8651_getOutsideNaptServerPortMapping(TRUE,ip_d,ip_dp);
	// MAC Filtering
	if (tb2 && (arp_t=_rtl8651_getArpEntry(NULL, tb2->localIp, FALSE))) {
		if (_rtl8651_isMacAccessControlPermit(&arp_t->macInfo->macAddr) == FALSE)
			return FWDENG_MACBLOCK;
	}
	
	client_alg_link = _rtl8651_l4FindLinkAlg(ip_sp, TRUE);	
	if(ip_sp != ip_dp){	
		server_alg_link = _rtl8651_l4FindLinkAlg(ip_dp, TRUE);
	}

#if 0 // disable by tony: this path do not need again. 2004/2/10
	if(tb2 && server_alg_link && tb == NULL){		
		tb=&entry;
		if(_rtl8651_getOutboundAlgCtrlFlow(TRUE,tb2->localIp,ip_dp-tb2->globalPort+tb2->localPort,ip_s,ip_sp, &tb)!=SUCCESS){
			tb=NULL;
		}
	}
#endif	

	if (tb == NULL) {
		/* 
			Now, we are going to establish a new flow, 
			but we have to check the ingress ACL of "L4NEWFLOW" type first.
		*/
		int32 local_val;
		rtl8651_tblDrv_vlanTable_t *local_vlanp = &DrvTbl.vlan[phdr->ph_vlanIdx];
		local_val = _rtl8651_ingressAcl(phdr->ph_dsid, 0, local_vlanp, local_vlanp->vid, phdr->ph_mbuf->m_data, (uint8 *) pip, RTL8651_ACL_L4NEWFLOW, _RTL8651_OPERATION_OTHER);
		if (((local_val > 0) && (local_val & FWDENG_DROP)) ||(local_val == FWDENG_ACL_IN_USERTRAP))
			return local_val;
	}




	if(tc->th_flags == TH_SYN)
		actionFlag |= TCP_IN_TCP_SYN;

	if(tb)
		actionFlag |= TCP_IN_TCP_ENTRY;	
	
	if(tb2)
		actionFlag |= TCP_IN_SERVER_PORT_ENTRY;	

#ifdef _RTL_WEAK_NAPT_CTRL
	if ((actionFlag == TCP_IN_TCP_SYN) || ((actionFlag == 0) && rtl8651_drvWeakTcpNaptEnable)) //not server port entry, not tcp entry, only have SYN or in weak NAPT mode
#else
	if (actionFlag == TCP_IN_TCP_SYN) //not server port entry, not tcp entry, only have SYN
#endif /* _RTL_WEAK_NAPT_CTRL */
	{

		// check is this port a trigger port in SYN packet.
#ifdef _RTL_TRIGGER_PORT

#ifdef _RTL_WEAK_NAPT_CTRL
		if (actionFlag == TCP_IN_TCP_SYN)
			initFlag = TCP_INBOUND_SYN_RCVD_FLOW;	// syn packet
		else
			initFlag = TCP_ESTABLISHED_FLOW;			// non syn packet
#endif /* _RTL_WEAK_NAPT_CTRL */

		if (_rtl8651_getTriggeredIpAddr(	TRUE,
										phdr->ph_dsid,
										0,			/* we don't include extIP as parameter in our trigger port check yet. */
										ip_dp,
										&ip_org) == SUCCESS)
		{
			if ((retval = _rtl8651_addNaptConnection(	TRUE,
													(TRUE|DOS_NAPT_ENCOUNT),
													initFlag,
													ip_org, ip_dp, &ip_d, &ip_dp, ip_s, ip_sp)) != SUCCESS)
			{
				return FWDENG_TCP_NAPT_IN_ADDFAIL;
			}
			tb = &entry;
			if (_rtl8651_getInboundNaptFlow(TRUE, ip_d, ip_dp, ip_s, ip_sp, &tb) != SUCCESS)
			{
				assert(0);
				return FWDENG_TCP_NAPT_IN_NOTFOUND;
			}
			inTrigger = 1;
		}
#endif			
	}

	//tony: when Special AP Incoming Port is a ALG port, don't do ALG.
	if(inTrigger==0)
	{
		if(client_alg_link)
			actionFlag |= TCP_IN_CLIENT_ALG;
		
		if(server_alg_link)
			actionFlag |= TCP_IN_SERVER_ALG;
	}

#ifdef _TONY_DEBUG
{
	uint32 sip=GET_UINT32_UNALIGNED(&pip->ip_src);
	uint32 dip=GET_UINT32_UNALIGNED(&pip->ip_dst);	
	rtlglue_printf("l4TcpNaptAliasIn src=%u.%u.%u.%u:%d dst=%u.%u.%u.%u:%d actionFlag=%d\n",NIPQUAD(sip),ip_sp,NIPQUAD(dip),ip_dp,actionFlag);
	if(tc->th_flags&TH_SYN) rtlglue_printf("SYN ");
	if(tc->th_flags&TH_ACK) rtlglue_printf("ACK ");
	if(tc->th_flags&TH_RST) rtlglue_printf("RST ");
	if(tc->th_flags&TH_FIN) rtlglue_printf("FIN ");
	if(tc->th_flags&TH_PUSH) rtlglue_printf("PUSH ");	
	rtlglue_printf("\n");
}
#endif	


// All of the following case is the same as case 0.
//	case 8: // Client ALG (unknow packet)
//	case 12: // Client ALG + Server Port (unknow packet)
//	case 16: // Server ALG (unknow packet)
//	case 20: // Server ALG + Server Port (unknow packet) 
//	case 24: // Server ALG + Client ALG  (unknow packet)
//	case 28: // Server ALG + Client ALG + Server Port (unknow packet)	
#ifdef _RTL_WEAK_NAPT_CTRL
	if((actionFlag&(TCP_IN_TCP_SYN|TCP_IN_TCP_ENTRY))==0)
	{
		if (actionFlag == TCP_IN_SERVER_PORT_ENTRY)
		{
			initFlag = TCP_ESTABLISHED_FLOW;
		}else if (rtl8651_drvWeakTcpNaptEnable)
		{
			/*
				When Weak Napt Control is turned on, for all non-SYN packets without entry we change the actionFlag
				for them to add new entries.Besides we modify initFlag to generate the ESTABLISH flow for these packets.
			*/
			initFlag = TCP_ESTABLISHED_FLOW;
			actionFlag |= TCP_IN_TCP_SYN;
		}else
		{
			actionFlag = 0;
		}
	}
#else /* _RTL_WEAK_NAPT_CTRL */
	if((actionFlag&(TCP_IN_TCP_SYN|TCP_IN_TCP_ENTRY)) == 0)
	{
		if (actionFlag == 4)
		{
			initFlag = TCP_ESTABLISHED_FLOW;
		}else
			actionFlag = 0;
	}
#endif /* _RTL_WEAK_NAPT_CTRL */

// All of the following case is the same as case 3.
//	case 7: // SYN + Entry + Server Port(software)
//	case 11:	// SYN + Entry + Client ALG
//	case 15: // SYN + Entry + Server Port + Client ALG 
//	case 19: // SYN + Entry + Server ALG
//	case 23: // SYN + Entry + Server Port + Server ALG 
//	case 27: // SYN + Entry +  Client ALG + Server ALG
//	case 31: // SYN + Entry + Server Port + Client ALG + Server ALG
if((actionFlag&(TCP_IN_TCP_SYN|TCP_IN_TCP_ENTRY))==(TCP_IN_TCP_SYN|TCP_IN_TCP_ENTRY))
{
	/* forbid the piggyback SYN packet */
	if ((tb->tcpFlag > (TCP_LISTEN_FLOW - 1)) &&
		(tb->tcpFlag < (TCP_CLOSED_FLOW - 1)) &&
		(tb->tcpFlag != (TCP_INBOUND_SYN_RCVD_FLOW - 1)))
	{
		return FWDENG_TCP_NAPT_IN_NOTFOUND;
	}
	actionFlag = 3;
}

	switch(actionFlag){
	case 0:// case 0,4,8,12,16,20,24,28 (unknow packet, fwd to PS)		
#ifdef _RTL_PS_ACTION
			i=_rtl8651_doProtoStackAction(pip);
			if(i==PS_ACTION_DROP) return FWDENG_USER_DROP;	
#endif	
		return FWDENG_TCP_NAPT_IN_NOTFOUND;
	case 1: // SYN
	case 9: // SYN + Client ALG(erroneous judgement: the src port is touch the ALG table) tony:2004/2/10
	case 17: // SYN + Server ALG(erroneous judgement: the dst port is touch the ALG table) tony:2004/2/10		
		if(inTrigger==1) {
			// is Trigger Port entry
			break;
		}
		
		// but maybe UPNP incoming connection
#ifdef _RTL_UPNP	
		{
			/* UPNP TCP port Mapping */		
			ipaddr_t local_ip = 0;
			uint16 local_port = 0;
			uint32 flags = 0, algIdx = 0;

			if (_rtl8651_findInboundUpnpMapping(TRUE, ip_s, ip_sp,ip_d, ip_dp, &local_ip, &local_port, &flags, &algIdx) != SUCCESS)
			{
				goto newDmzHostConn;
			}

			if(actionFlag==1)
			{
				struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s ttbe,*ttb;
				ttb=&ttbe;
				if(_rtl8651_getInboundNaptFlow(TRUE,ip_d,ip_dp,ip_s,20, &ttb)==SUCCESS) {
					if(ttb->tcpFlag==0)	_rtl8651_delNaptConnection(ttb);					
				}

				if(_rtl8651_addAlgQosNaptConnection(TRUE,
						(TRUE|DOS_NAPT_ENCOUNT|(flags & UPNP_ALG_QOS)),
						TCP_INBOUND_SYN_RCVD_FLOW,
						local_ip,local_port,
						&ip_d, &ip_dp,
						ip_s, ip_sp,
						algIdx) != SUCCESS)
					{
#if 0
						if((retval =
							_rtl8651_addNaptConnection(TRUE,
								(TRUE|DOS_NAPT_ENCOUNT), TCP_INBOUND_SYN_RCVD_FLOW,
								local_ip, local_port, &ip_d, &ip_dp, ip_s, ip_sp)) != SUCCESS)
						{
#endif
						return FWDENG_TCP_NAPT_IN_ADDFAIL;
#if 0
						}
#endif
					}						
				
				tb=&entry;
				if(_rtl8651_getInboundNaptFlow(TRUE,ip_d,ip_dp,ip_s,ip_sp, &tb) != SUCCESS)
				{
					rtlglue_printf("error: _rtl8651_getInboundNaptFlow connection in upnp \n");
		 			return FWDENG_TCP_NAPT_IN_NOTFOUND;
				}
				else
				{	
					//	tb->alg_server_out = server_alg_link->alg_outboundServercb;
					//tb->alg_server_out = server_alg_link->alg_outboundServercb;				
					break;
				}			
	 		}
			else
			{				
				tb = _rtl8651_addAlgCtrlConnection(
						(TRUE|DOS_NAPT_ENCOUNT),TCP_INBOUND_SYN_RCVD_FLOW,
						local_ip,local_port,&ip_d, &ip_dp, ip_s, ip_sp);
				if(!tb)
					return FWDENG_TCP_NAPT_IN_SVRALGADDFAIL;
				if(tb){	
					tb->	tcpAckModified = 0;
					tb->	tcpFtpLastLineCRLFTermed = 0;
					tb->tcpStateINdex = 0;			
					memset(&tb->ack[0],0,sizeof(struct l4ack_data_record)*L4_N_LINK_TCP_DATA);				
					}
				break;
			}
		}
#endif /* _RTL_UPNP */		
			
			
newDmzHostConn:
		/* client and server not found */

#ifdef _RTL_PS_ACTION
			i=_rtl8651_doProtoStackAction(pip);
			if(i==PS_ACTION_DROP) return FWDENG_USER_DROP;
			if(i==PS_ACTION_TO_PS) return FWDENG_TCP_NAPT_IN_NOTFOUND;
#endif
		if (_rtl8651_getDmzHost(phdr->ph_dsid,ip_d, &dmzhost) == SUCCESS)
		{
#if 0	
			for(i=0;i<RTL8651_MAX_PROTO_STACK_SERVER_USED_TCP_PORTS;i++)
			{				
				if(rtl8651_ProtoStackServerUsedTcpPorts[i]==0) 
				{
					break;
				}
				else if(rtl8651_ProtoStackServerUsedTcpPorts[i]==tc->th_dport) 
				{
					//reply packet will be handled by protocol stack
					return FWDENG_TCP_NAPT_IN_NOTFOUND; // go to protocol stack
				}
			}			
#endif			
			

			//if(_rtl8651_addNaptConnection(TRUE, TRUE, TCP_INBOUND_SYN_RCVD_FLOW,dmzhost,ip_dp,
			/* to make sure this napt flow in sw due to DoS checking needs */
			if (actionFlag!=1)
			{

				tb = _rtl8651_addAlgCtrlConnection(
					(TRUE|DOS_NAPT_ENCOUNT), TCP_INBOUND_SYN_RCVD_FLOW,
					dmzhost, ip_dp, &ip_d, &ip_dp, ip_s, ip_sp);
				if(!tb)
					return FWDENG_TCP_NAPT_IN_SVRALGADDFAIL;
				if(tb){	
					tb->	tcpAckModified = 0;
					tb->	tcpFtpLastLineCRLFTermed = 0;
					tb->tcpStateINdex = 0;			
					memset(&tb->ack[0],0,sizeof(struct l4ack_data_record)*L4_N_LINK_TCP_DATA);				
					}

			}
			else {
				if(_rtl8651_addNaptConnection(TRUE,
					(TRUE|DOS_NAPT_ENCOUNT), TCP_INBOUND_SYN_RCVD_FLOW,
					dmzhost,ip_dp, &ip_d, &ip_dp, ip_s, ip_sp) != SUCCESS)
				{
					printk("add connection failed\n");
					return FWDENG_TCP_NAPT_IN_ADDFAIL;
				}
				tb=&entry;
				if(_rtl8651_getInboundNaptFlow(TRUE,ip_d,ip_dp,ip_s,ip_sp, &tb)!=SUCCESS) {
					printk("getInboundflow failed\n");
		 			return FWDENG_TCP_NAPT_IN_NOTFOUND;
				}
			}
			
			break;
		}else
			return FWDENG_TCP_NAPT_IN_NOTFOUND;
	case 2: // Entry
	case 6: // Entry + Server Port

// tony: this path only for h245 special case, the port info is not save in ALG table. 
// but must handle the packet with ALG. h245 only use alg_client_in
			
		if(tb->alg_client_in){
			int r;
			r = tb->alg_client_in(phdr, pip, tb);
			if(r==FWDENG_USER_DROP) {
				return FWDENG_USER_DROP;
			}
		}
#if 0		
		if(tb->alg_server_in){
			int r;
			r = tb->alg_server_in(phdr, pip, tb);
			if(r==FWDENG_USER_DROP) {
				return FWDENG_USER_DROP;
			}
		}			
#endif			
		break;			
	case 3: // SYN + Entry + ... ( case 3,7,11,15,19,23,27,31 )
		// duplicate sync, by pass this packet
		break;				
	case 4: // Inbound Server Port, non-SYN
		if((retval =
			_rtl8651_addNaptConnection(TRUE,
				(TRUE|DOS_NAPT_ENCOUNT), TCP_ESTABLISHED_FLOW,
				tb2->localIp, ip_dp-tb2->globalPort+tb2->localPort, &ip_d, &ip_dp, ip_s, ip_sp)) != SUCCESS)
		{
			return FWDENG_TCP_NAPT_IN_ADDFAIL;
		}
		tb=&entry;
			
		if(_rtl8651_getInboundNaptFlow(TRUE,ip_d,ip_dp,ip_s,ip_sp, &tb)!=SUCCESS)
		{
			rtlglue_printf("error: _rtl8651_getInboundNaptFlow connection Failed\n");
 			return FWDENG_TCP_NAPT_IN_NOTFOUND;
		}
		break;
	case 5: // SYN + Server Port(software)

		if((retval = _rtl8651_addNaptConnection(TRUE,
				(TRUE|DOS_NAPT_ENCOUNT), TCP_INBOUND_SYN_RCVD_FLOW,
				tb2->localIp,ip_dp-tb2->globalPort+tb2->localPort,&ip_d, &ip_dp, ip_s, ip_sp)) != SUCCESS)
		{
			return FWDENG_TCP_NAPT_IN_ADDFAIL;
		}
		tb=&entry;
			
		if(_rtl8651_getInboundNaptFlow(TRUE,ip_d,ip_dp,ip_s,ip_sp, &tb)!=SUCCESS) {
			rtlglue_printf("error: _rtl8651_getInboundNaptFlow connection in upnp \n");
 			return FWDENG_TCP_NAPT_IN_NOTFOUND;
		}
		break;

	case 10: // Client ALG + Entry		 
	case 14: // Client ALG + Server Port + Entry	
		 if(client_alg_link->fwEngAdd && client_alg_link->alg_inboundClientcb){
			client_alg_link->alg_inboundClientcb(phdr, pip, tb);
		 }
		 break;
	
	case 13: // Client ALG + Server Port + SYN		
	case 21: // Server ALG + Server Port + SYN
	case 29: // Server ALG + Client ALG + Server Port + SYN
	
		tb=_rtl8651_addAlgCtrlConnection(
				(TRUE|DOS_NAPT_ENCOUNT),TCP_INBOUND_SYN_RCVD_FLOW,
				tb2->localIp,ip_dp+tb2->localPort-tb2->globalPort,&ip_d, &ip_dp, ip_s, ip_sp);
		if(!tb)
			return FWDENG_TCP_NAPT_IN_SVRALGADDFAIL;
		if(tb){	
			tb->	tcpAckModified = 0;
			tb->	tcpFtpLastLineCRLFTermed = 0;
			tb->tcpStateINdex = 0;			
			memset(&tb->ack[0],0,sizeof(struct l4ack_data_record)*L4_N_LINK_TCP_DATA);

#if 0 // nerver be touched here, diable by tony: 2004/2/10			
			 if(client_alg_link->fwEngAdd && client_alg_link->alg_inboundClientcb){
				client_alg_link->alg_inboundClientcb(phdr, pip, tb);							
			 }			
#endif			 
		}
		break;	


	case 18:	//Server ALG + Entry
	case 22: // Server ALG + Server Port + Entry
		 if(server_alg_link->fwEngAdd && server_alg_link->alg_inboundServercb){
			server_alg_link->alg_inboundServercb(phdr, pip, tb);							
		 }
		break;		
	default:	// unknow case
		rtlglue_printf("%s: action=%d, flag=%d\n",__FUNCTION__,actionFlag,tc->th_flags );	
		return FWDENG_TCP_NAPT_IN_IGNORE;
	}

	assert(tb);
	
#ifdef _RTL_WEAK_NAPT_CTRL
	/*
		execute Alg control block for weak napt flows
	*/
	if (rtl8651_drvWeakTcpNaptEnable && (initFlag == TCP_ESTABLISHED_FLOW) && (actionFlag != 4))
	{
		if (actionFlag & TCP_IN_SERVER_ALG)
		{
			 if(server_alg_link->fwEngAdd && server_alg_link->alg_inboundServercb)
			 {
				server_alg_link->alg_inboundServercb(phdr, pip, tb);							
			 }
		}
		if (actionFlag & TCP_IN_CLIENT_ALG)
		{
			 if(client_alg_link->fwEngAdd && client_alg_link->alg_inboundClientcb)
			 {
				client_alg_link->alg_inboundClientcb(phdr, pip, tb);
			 }
		}
	}
#endif /* _RTL_WEAK_NAPT_CTRL */

	/* tcp sync packet  adj mss value */		
	if(tc->th_flags & TH_SYN){
		int8 auto_add = FALSE,auto_delete;			
		rtl8651_getAsicNaptAutoAddDelete(&auto_add,&auto_delete);				
		if(auto_add == FALSE){							
			uint8 *cp;
			int32 opt, optlen,cnt;		
			uint16 mss;
		
			cp = (uint8 *)(tc+1);
			cnt = ((tc->th_off_x >>2)&0x3c)  -sizeof(struct tcphdr);//Edward: Fix for alignment issue
	
			for(; cnt > 0; cnt -= optlen, cp+= optlen){
				opt = cp[0];
				if(opt == TCPOPT_EOL)
					break;
				if(opt == TCPOPT_NOP)
					optlen = 1;
				else{
					optlen = cp[1];
					if(optlen <= 0)
						break;
				}

				switch (opt) {
				case TCPOPT_MAXSEG:
					mss = ntohs(*(uint16*)(&cp[2]));
//rtlglue_printf("call _rtl8651_getPppoeSessionByIpAddr()....2(tcpNaptAliasIn)\n");
					//thisPPPoE=_rtl8651_getPppoeSessionByIpAddr(PKTGET_UINT32_UNALIGNED(&ip_d));
					thisPPPoE=_rtl8651_getPppoeSessionByIpAddr(ip_d);
					if (thisPPPoE)
					{
						if (mss >pppoe_mss[thisPPPoE->dsid])
						{
							mtu=pppoe_mss[thisPPPoE->dsid];
							accumulate -=(mss-pppoe_mss[thisPPPoE->dsid]);
							*(uint16*)(&cp[2]) = htons(mtu);
						}
					}else if(mss > rtl8651_mss){
						mtu=rtl8651_mss;
						accumulate -=(mss-rtl8651_mss);
						*(uint16*)(&cp[2]) = htons(mtu);
					}					
					break;
				default:
					continue;
				}			
			}				
		}
	}

	/* update cache */
	tb->dsid = phdr->ph_dsid;

#if 0 //disable 2004/02/10
	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = tc->th_sport;
	ip_dp =  tc->th_dport;
#endif	
	ip_org = tb->insideLocalIpAddr;
	ip_orgp = tb->insideLocalPort;

	/* See if ACK number needs to be modified */
	if (tb->tcpAckModified == 1)
	{
		int32     delta;
		uint32    ack_32_x;
		
		delta = _rtl8651_l4GetDeltaAckIn(pip, tc, tb);
		accumulate += delta;
		if (delta != 0)
		{
			ack_32_x=PKTGET_UINT32_UNALIGNED(&tc->th_ack);
			ack_32_x = htonl(ntohl(ack_32_x) - delta);
			PKTSET_UINT32_UNALIGNED(ack_32_x, &tc->th_ack);
		}
	}

	/* Monitor TCP connection state */
	_rtl8651_l4TcpMonitorIn(tc, tb);

	/* chenyl : url filter checking is deferred to slow path */
#ifdef RTL865XB_URL_FILTER
	if((_urlFilterAclRuleNum>=0) &&
		(!tb->urlFiltered||((phdr->ph_reason&0xe)==4
		&& ((phdr->ph_reason&0xff0)>>5)==_urlFilterAclRuleNum)))
#else
	if(!tb->urlFiltered||phdr->ph_reason==65)
#endif
	{
		int32 position, len=64;
		int8 matched[64];
		int32 hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >>4)&0xf)) << 2;
		if ((ntohs(pip->ip_len))>hlen)
		{
			position=_rtl8651_urlFilter(phdr->ph_dsid,phdr,pip,RTL8651_WAN2LAN,matched, &len);
			if (position>0)
			{
				/*
					cfliu: If you prepare a string buffer for the filter function, 
					rtl8651_urlFilter() would return string matched, its length and its position found in mbuf chain
					you may insert a log function pointer here to keep this extra info
					or just return a error value if you don't care.
				*/
				#ifdef RTL865X_DEBUG
				rtlglue_printf("string '%s'(%d bytes) in URL at %dth byte. Filtered...\n",matched, len, position );
				#endif
				_rtl8651_updateNaptConnection(tb,TCP_CLOSED_FLOW, 3);	/* keep entry for 3 seconds. */
				return FWDENG_TCP_URLFILTERED;
			}
			tb->urlFiltered = TRUE; /* so we won't check this flow anymore. */
		}
	}


	/*******************************************************************************************/
	if ((!(actionFlag&TCP_IN_TCP_ENTRY)) && tb)
	{
		assert(tb->dosEnCount == TRUE && tb->dosFromInternal == FALSE);
	
		// chenyl: 2004/03/08 log for newly-added connection	
		if ((_loggingModule[phdr->ph_dsid]&RTL8651_LOGMODULE_NEWFLOW)&&_pUserLoggingFunc)
		{
#ifdef _RTL_NEW_LOGGING_MODEL
			rtl8651_logInfo_t info;
			info.infoType		= RTL8651_LOG_INFO_PKT;
			info.pkt_dsid		= phdr->ph_dsid;
			info.pkt_sip		= ip_s;
			info.pkt_dip		= ip_org;
			info.pkt_proto		= IPPROTO_TCP;
			info.pkt_direction	= 1;	/* direction: WAN to LAN */
			info.pkt_sport		= ip_sp;
			info.pkt_dport		= ip_orgp;
			info.action		= RTL8651_LOGACTION_FORWARD;
			(* _pUserLoggingFunc)(RTL8651_LOGMODULE_NEWFLOW, RTL8651_LOG_NEWFLOW_NewTcpNaptInbound, &info);
#else
			(*_pUserLoggingFunc)(
				phdr->ph_dsid,
				RTL8651_LOGMODULE_NEWFLOW,
				    (unsigned char)pip->ip_p, /* protocol */
				    (unsigned char)1, /* direction: WAN to LAN */
					(unsigned long)ip_s,
					(unsigned long)ip_org,
					(unsigned short)ip_sp,
					(unsigned short)ip_orgp,
					0, /* type: only meaningful for ICMP */
					RTL8651_LOGACTION_FORWARD, /* action */
					"New TCP Inbound Flow" /* msg */);
#endif
		}
	}
	/*******************************************************************************************/

#ifdef _RTL_NAPT_FRAG
	if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_OFFMASK|IP_MF)) == IP_MF))
	{
		if ((pip->ip_vhl & 0x0f) != 5)
		{	/* ASIC cannot offload checksum if ip header length is not 20, so we must calculate L3/L4 checksum by ourself */
			_RTL8651_L3_ADJUST_CHECKSUM(ip_org, ip_d, pip->ip_sum);
		}
		/* for both first fragment packet and non-offloaded packet, we must calculate their L4 checksum here */
		_RTL8651_L4_ADJUST_CHECKSUM(ip_org, ip_d, ip_orgp, ip_dp, tc->th_sum);
		_RTL8651_ADJUST_CHECKSUM(accumulate, tc->th_sum);	/* mss modification and delta modificaion */
		/* update fragment cache only if packet is the first fragment packet */
		if (IS_FRAGMENT(pip->ip_off))
		{
			_rtl8651_updateFragCache(INBOUND_FRAG, phdr->ph_sip, phdr->ph_dip, ntohs(pip->ip_id), ip_org, 0, NULL, phdr->ph_dsid);
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
			_rtl8651_fragmentShortestPathCacheSet_firstHalf(phdr, (uint8*)tc, pip, ip_s, ip_d, RTL8651_FRAGMENT_SHORTESTPATH_MODE_MODIFYDIP, ip_org);
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
		}
	}
#endif /* _RTL_NAPT_FRAG */	

	/* put otiginal port number back */
	tc->th_dport = htons(ip_orgp);
	/* put otiginal address back */
	PKTSET_UINT32_UNALIGNED(ip_org, &pip->ip_dst);

	return (SUCCESS);
}

/*********************************************************************************
* Routine Name : _rtl8651_l4TcpNatAlgAliasIn
* Description : 
*   TCP packet input process
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4TcpNatAlgAliasIn(struct rtl_pktHdr *phdr, struct ip *pip, rtl8651_tblDrv_natEntry_t *natEntry)
{

	struct tcphdr *tc;
	ipaddr_t ip_s, ip_d, ip_org = 0;
	uint16 ip_sp, ip_dp, ip_orgp = 0;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb = &entry;
	int32 accumulate = 0;
	rtl8651_tblDrv_algEntry_t *client_alg_link = NULL;
	rtl8651_tblDrv_algEntry_t *server_alg_link = NULL;
	uint32 actionFlag = 0;
	rtl8651_tblDrv_pppTable_t *thisPPPoE;
	uint16 mtu;

	tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
		
	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(tc->th_sport);
	ip_dp =  ntohs(tc->th_dport);

	if (ip_d == natEntry->localIp)
	{	/* NI == NE, no ALG is needed */
		tb = NULL;
		goto adjust_mss;
	}

#define TCP_IN_TCP_SYN 0x01
#define TCP_IN_TCP_ENTRY 0x02
#define TCP_IN_SERVER_PORT_ENTRY 0x04
#define TCP_IN_CLIENT_ALG 0x08
#define TCP_IN_SERVER_ALG	0x10

	client_alg_link = _rtl8651_l4FindLinkAlg(ip_sp, TRUE);		
	if(ip_sp != ip_dp){
		server_alg_link = _rtl8651_l4FindLinkAlg(ip_dp, TRUE);
	}
	//chenyl: mark for mss adjust
//	if(client_alg_link == NULL && server_alg_link == NULL){
//		return FWDENG_TCP_NAT_IN_ALGNOTFOUND;
//	}
		
	if(_rtl8651_getInboundNaptFlow(TRUE,ip_d,ip_dp,ip_s,ip_sp, &tb)!=SUCCESS){
 		tb=NULL;
 	}
	

	//if(tc->th_flags == TH_SYN)
	if(tc->th_flags & TH_SYN)
		actionFlag |= TCP_IN_TCP_SYN;
	
	if(tb)
		actionFlag |= TCP_IN_TCP_ENTRY;	
		 
	if(client_alg_link)
		actionFlag |= TCP_IN_CLIENT_ALG;
		
	if(server_alg_link)
		actionFlag |= TCP_IN_SERVER_ALG;

#ifdef _TONY_DEBUG
		rtlglue_printf("l4TcpNatAliasIn src:%u.%u.%u.%u:%d dest:%u.%u.%u.%u:%d actionFlag=%d\n",NIPQUAD(ip_s),ip_sp,NIPQUAD(ip_d),ip_dp,actionFlag);
#endif
	
	switch(actionFlag){
		case 2:
		/* cleint entry found */
			/* entry already exist */
			if(tb->alg_client_in){
				int r;
				r = tb->alg_client_in(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}
			if(tb->alg_server_in){
				int r;
				r = tb->alg_server_in(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}				
		case 10:
		case 11:
			 /*ALG Client process*/
			 if(client_alg_link->fwEngAdd && client_alg_link->alg_inboundClientcb){
				client_alg_link->alg_inboundClientcb(phdr, pip, tb);							
			 }
			 break;
		case 18:
		case 19:
			 /*ALG Server process*/
			 if(server_alg_link->fwEngAdd && server_alg_link->alg_inboundServercb){
				server_alg_link->alg_inboundServercb(phdr, pip, tb);							
			 }
			break;		
		case 9: /* client alg found ; sync found*/	
		case 17: /* server alg found ; sync found*/		
			/* add a new alg link entry */		
			tb = _rtl8651_addAlgCtrlConnection(
				(TRUE|DOS_NAPT_ENCOUNT), TCP_INBOUND_SYN_RCVD_FLOW,
				natEntry->localIp, ip_dp, &ip_d, &ip_dp, ip_s, ip_sp);										
			if(!tb)
				return FWDENG_TCP_NAT_IN_ALGADDFAIL;
			tb->natlink = 1;

			break;	
		case 1:
			goto adjust_mss;
		case 0:
			return FWDENG_TCP_NAT_IN_ALGNOTFOUND;
		default:	/* this should not happen */		
			rtlglue_printf("%s: action=%d, flag=%d\n",__FUNCTION__,actionFlag,tc->th_flags );
			return FWDENG_TCP_NAT_IN_ALGIGNORE;
	}

	assert(tb);
	/* Monitor TCP connection state */
	_rtl8651_l4TcpMonitorIn(tc, tb);

	/* tcp sync packet  adj mss value */		
adjust_mss:
	if(tc->th_flags & TH_SYN){
		int8 auto_add = FALSE,auto_delete;			
		rtl8651_getAsicNaptAutoAddDelete(&auto_add,&auto_delete);				
		if(auto_add == FALSE){							
			uint8 *cp;
			int32 opt, optlen,cnt;		
			uint16 mss;
		
			cp = (uint8 *)(tc+1);
			cnt = ((tc->th_off_x >>2)&0x3c)  -sizeof(struct tcphdr);//Edward: Fix for alignment issue
	
			for(; cnt > 0; cnt -= optlen, cp+= optlen){
				opt = cp[0];
				if(opt == TCPOPT_EOL)
					break;
				if(opt == TCPOPT_NOP)
					optlen = 1;
				else{
					optlen = cp[1];
					if(optlen <= 0)
						break;
				}

				switch (opt) {
				case TCPOPT_MAXSEG:
					mss = ntohs(*(uint16*)(&cp[2]));
//rtlglue_printf("call _rtl8651_getPppoeSessionByIpAddr()....3(l4TcpNatAliasIn)\n");
					//thisPPPoE=_rtl8651_getPppoeSessionByIpAddr(PKTGET_UINT32_UNALIGNED(&ip_d));
					thisPPPoE=_rtl8651_getPppoeSessionByIpAddr(ip_d);
					if (thisPPPoE)
					{
						if (mss >pppoe_mss[thisPPPoE->dsid])
						{ 		
							mtu=pppoe_mss[thisPPPoE->dsid];
							accumulate -=(mss-pppoe_mss[thisPPPoE->dsid]);
							*(uint16*)(&cp[2]) = htons(mtu);
						}
					}else if(mss > rtl8651_mss){ 						
						mtu=rtl8651_mss;
						accumulate -=(mss-rtl8651_mss);
						*(uint16*)(&cp[2]) = htons(mtu);
					}
					break;
				default:
					continue;
				}			
			}				
		}
	}


	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(tc->th_sport);
	ip_dp =  ntohs(tc->th_dport);
	
	
	if(tb){
		ip_org = tb->insideLocalIpAddr;
		ip_orgp = tb->insideLocalPort;

		/* put otiginal port number back */
		tc->th_dport = htons(ip_orgp);

		/* See if ACK number needs to be modified */
		if (tb->tcpAckModified == 1)
		{
			int32     delta;
			uint32    ack_32_x;
			
			delta = _rtl8651_l4GetDeltaAckIn(pip, tc, tb);
			accumulate += delta;
			if (delta != 0)
			{
				ack_32_x=PKTGET_UINT32_UNALIGNED(&tc->th_ack);
				ack_32_x = htonl(ntohl(ack_32_x) - delta);
				PKTSET_UINT32_UNALIGNED(ack_32_x, &tc->th_ack);
			}
		}
		PKTSET_UINT32_UNALIGNED(ip_org,&pip->ip_dst);
	}else{
		assert(natEntry);
		PKTSET_UINT32_UNALIGNED(natEntry->localIp,&pip->ip_dst);
	}

#ifdef _RTL_NAPT_FRAG	
	if(tb)
	{	/* nat alg */
		if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK)) == IP_MF))
		{
			if ((pip->ip_vhl & 0x0f) != 5)
			{
				_RTL8651_L3_ADJUST_CHECKSUM(ip_org, ip_d, pip->ip_sum);
			}
			_RTL8651_L4_ADJUST_CHECKSUM(ip_org, ip_d, ip_orgp, ip_dp, tc->th_sum);
			_RTL8651_ADJUST_CHECKSUM(accumulate, tc->th_sum);	/* mss modification and delta modificaion */
		}
	}
	else {
		/* non-ALG, syn pkt */
		ip_org = natEntry->localIp;

		if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK)) == IP_MF))
		{
			if ((pip->ip_vhl & 0x0f) != 5)
			{
				_RTL8651_L3_ADJUST_CHECKSUM(ip_org, ip_d, pip->ip_sum);
			}
			_RTL8651_L4_ADJUST_CHECKSUM(ip_org, ip_d, 0, 0, tc->th_sum);
			_RTL8651_ADJUST_CHECKSUM(accumulate, tc->th_sum);	/* mss modification and delta modificaion */
		}
	}
#endif /* _RTL_NAPT_FRAG */	

	return (SUCCESS);
}

/*********************************************************************************
* Routine Name :  _rtl8651_l4TcpNaptAliasOut                                   
* Description :  tcp packet output process
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4TcpNaptAliasOut(struct rtl_pktHdr *phdr, struct ip *pip)
{
	struct tcphdr *tc;
	ipaddr_t ip_s, ip_d, ip_alias = 0;
	uint16 ip_sp, ip_dp, ip_aliasp = 0;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb=&entry;
	rtl8651_tblDrv_algEntry_t *client_alg_link = NULL; 
	rtl8651_tblDrv_algEntry_t *server_alg_link = NULL; 
	int32 accumulate = 0;
	rtl8651_tblDrv_naptServerPortEntry_t *tb2;	
	uint32 actionFlag = 0;
	int inTrigger = 0;
	rtl8651_tblDrv_pppTable_t *thisPPPoE;
	uint16 mtu;
	int8 initFlag = TCP_OUTBOUND_SYN_RCVD_FLOW;
	int32 retval = FAILED;

	/*
		for outbound uPnP mapping:

			When outbound SYN packet which match uPnP mapping,
			we will treat it same as Server Port process.
			And some notes are as below:

			1. Due to we treat this as a upnp mapping match process, therefore,
			    for One-Shot uPnP mapping, the uPnP entry will be removed after gotten its information.
			2. The check priority of NAPT outbound traffic:
				1) Server Port
				2) uPnP Mapping
				3) NAPT-redirect

																	chenyl 2005/05/06
	*/
	int32 isUpnpMapping = FALSE;
	ipaddr_t upnpExtIp;
	uint16 upnpExtPort;
	uint32 upnpFlags;
	uint32 upnpAlgIdx;


#define TCP_OUT_TCP_SYN				0x01
#define TCP_OUT_TCP_ENTRY				0x02
#define TCP_OUT_SERVER_PORT_ENTRY	0x04
#define TCP_OUT_CLIENT_ALG				0x08
#define TCP_OUT_SERVER_ALG				0x10
#define TCP_OUT_UPNP_ENTRY			0x20

	tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));

	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(tc->th_sport);
	ip_dp =  ntohs(tc->th_dport);

	if(_rtl8651_getOutboundNaptFlow(TRUE,ip_s,ip_sp,ip_d,ip_dp, &tb)!=SUCCESS) {
		tb=&entry;
		if( _rtl8651_getOutboundAlgCtrlFlow(TRUE,ip_s,ip_sp,ip_d,ip_dp, &tb)!=SUCCESS) {
			tb=NULL;
		}
	} 

	tb2 = _rtl8651_getInsideNaptServerPortMapping(TRUE,ip_s,ip_sp);

	client_alg_link = _rtl8651_l4FindLinkAlg(ip_dp, TRUE);		

	if(ip_sp!= ip_dp)
	{
		/* tony: check Server ALG outbound packets by global port not source Port. 2004/2/18 */
		if(tb)
		{
			server_alg_link = _rtl8651_l4FindLinkAlg(tb->insideGlobalPort, TRUE);
		}else
		{
			server_alg_link = _rtl8651_l4FindLinkAlg(ip_sp, TRUE);
		}
	}

	/*
		If this packet:

			1. Have no entry for it.
			2. It's not in any server port mapping.

		We must check if it in upnp mapping or not, if this packet can fit any upnp mapping,
		we would use the corresponding external Ip/Port to modify Ip/Port.
	*/
	if ((tb == NULL) && (tb2 == NULL))
	{
		if (_rtl8651_findOutboundUpnpMapping(
			TRUE,
			ip_s,
			ip_sp,
			&upnpExtIp,
			&upnpExtPort,
			ip_d,
			ip_dp,
			&upnpFlags,
			&upnpAlgIdx) == SUCCESS)
		{
			isUpnpMapping = TRUE;
		}
	}

#ifdef _RTL_NAPT_REDIRECT
		/*
			For
				1. non-server port
				2. no-entry
				3.  non-upnp mapping

				packets, we must check napt redirect database and decide if we need to trap it or not
		*/
		if ((tb == NULL) && (tb2 == NULL) && (isUpnpMapping == FALSE))
		{
			ipaddr_t redirectIp;
			uint16 redirectPort;
			if (_rtl8651_findNaptRedirectEntry(TRUE, ip_s, ip_sp, ip_d, ip_dp, &redirectIp, &redirectPort, TRUE, phdr->ph_vlanIdx) == SUCCESS)
			{	/* this entry must be trapped to protocol stack for further process */
				/* modify dip/dport */
				PKTSET_UINT32_UNALIGNED(redirectIp, &pip->ip_dst);
				tc->th_dport = htons(redirectPort);
				/* always modify L3/L4 checksum */
				_RTL8651_L4_ADJUST_CHECKSUM(redirectIp, ip_d, redirectPort, ip_dp, tc->th_sum);
				_RTL8651_L3_ADJUST_CHECKSUM(redirectIp, ip_d, pip->ip_sum);
				/* update fragment cache */
				if (IS_FRAGMENT_FIRST(pip->ip_off))
				{
					_rtl8651_updateFragCache(NAPT_REDIRECT_FRAG, ip_s, ip_d, ntohs(pip->ip_id), redirectIp, 0, NULL, phdr->ph_dsid);
				}
#ifdef _RTL_QUEUE_FRAG_PKT
				if (_rtl8651_fragQueueProcess == TRUE)
				{	/* if system is under fragment-dequeue-phase, we must store redirectIP for remaining trapped fragment packet */
					_rtl8651_fragQueueNaptRedirect_MatchedIP = redirectIp;
				}
#endif /* _RTL_QUEUE_FRAG_PKT */
				return FWDENG_TO_GW;
			}
		}
#endif /* _RTL_NAPT_REDIRECT */ 


	if (tb == NULL) {
		/* 
			Now, we are going to establish a new flow, 
			but we have to check the ingress ACL of "L4NEWFLOW" type first.
		*/
		int32 local_val;
		rtl8651_tblDrv_vlanTable_t *local_vlanp = &DrvTbl.vlan[phdr->ph_vlanIdx];
		local_val = _rtl8651_ingressAcl(phdr->ph_dsid, 0, local_vlanp, local_vlanp->vid, phdr->ph_mbuf->m_data, (uint8 *) pip, RTL8651_ACL_L4NEWFLOW, _RTL8651_OPERATION_OTHER);
		if (((local_val > 0) && (local_val & FWDENG_DROP)) ||(local_val == FWDENG_ACL_IN_USERTRAP))
			return local_val;
	}

#ifdef _RTL_WEAK_NAPT_CTRL
	if (rtl8651_drvWeakTcpNaptEnable == FALSE)
#endif
	/*
		Server Port: Forbid outbound SYN+ACK packet if no connection exists
	*/
	if ((tc->th_flags & (TH_SYN|TH_ACK)) == (TH_SYN|TH_ACK) && tb == NULL)
	{
		return FWDENG_TCP_NAPT_OUT_NOTFOUND;
	}

	if(tc->th_flags == TH_SYN)
	{
		actionFlag |= TCP_OUT_TCP_SYN;
	}
	
	if(tb)
	{
		actionFlag |= TCP_OUT_TCP_ENTRY;
	}
	
	if(tb2)
	{
		actionFlag |= TCP_OUT_SERVER_PORT_ENTRY;
	}

	if (isUpnpMapping == TRUE)
	{
		actionFlag |= TCP_OUT_UPNP_ENTRY;
	}
	

#ifdef _RTL_TRIGGER_PORT
#ifdef _RTL_WEAK_NAPT_CTRL
		if ((actionFlag == TCP_OUT_TCP_SYN)||((actionFlag == 0) && rtl8651_drvWeakTcpNaptEnable)) //not TCP_ENTRY and not SERVER_PORT_ENTRY
#else
		if (actionFlag == TCP_OUT_TCP_SYN) /*not TCP_ENTRY and not SERVER_PORT_ENTRY */
#endif /* _RTL_WEAK_NAPT_CTRL */
		{
			if (_rtl8651_tryActiveTriggerPort(	TRUE,
											phdr->ph_dsid,
											ip_dp,
											0,
											ip_s) > 0)
			{
				inTrigger = 1;
			}
		}
#endif

	/* tony: when Special AP Trigger Port is a ALG port, don't do ALG. */
	if(inTrigger == 0)
	{
		if(client_alg_link)
			actionFlag |= TCP_OUT_CLIENT_ALG;
		
		if(server_alg_link)
			actionFlag |= TCP_OUT_SERVER_ALG;
	}

#ifdef _TONY_DEBUG
{
	uint32 sip=GET_UINT32_UNALIGNED(&pip->ip_src);
	uint32 dip=GET_UINT32_UNALIGNED(&pip->ip_dst);	
	rtlglue_printf("l4TcpNaptAliasOut src=%u.%u.%u.%u:%d dst=%u.%u.%u.%u:%d actionFlag=%d\n",NIPQUAD(sip),ip_sp,NIPQUAD(dip),ip_dp,actionFlag);
	if(tc->th_flags&TH_SYN) rtlglue_printf("SYN ");
	if(tc->th_flags&TH_ACK) rtlglue_printf("ACK ");
	if(tc->th_flags&TH_RST) rtlglue_printf("RST ");
	if(tc->th_flags&TH_FIN) rtlglue_printf("FIN ");
	if(tc->th_flags&TH_PUSH) rtlglue_printf("PUSH ");	
}
#endif

	/* set init TCP flow flag */
	initFlag = TCP_OUTBOUND_SYN_RCVD_FLOW;

/*
  *	All of the following case is the same as case 0.
  *	case 8: // Client ALG (unknow packet)
  *	case 12: // Client ALG + Server Port (unknow packet)
  *	case 16: // Server ALG (unknow packet)
  *	case 20: // Server ALG + Server Port (unknow packet)
  *	case 24: // Server ALG + Client ALG  (unknow packet)
  *	case 28: // Server ALG + Client ALG + Server Port (unknow packet)
  *
  *	case 40: // Client ALG + uPnP (unknow packet)
  *	case 48: // Server ALG + uPnP (unknow packet)
  *	case 56: // Server ALG + Client ALG + uPnP (unknow packet)

  */

#ifdef _RTL_WEAK_NAPT_CTRL
	if ((actionFlag & (TCP_OUT_TCP_SYN | TCP_OUT_TCP_ENTRY)) == 0)
	{
		if ((actionFlag == TCP_OUT_SERVER_PORT_ENTRY) || (actionFlag == TCP_OUT_UPNP_ENTRY))
		{
			initFlag = TCP_ESTABLISHED_FLOW;
		}else if (rtl8651_drvWeakTcpNaptEnable)
		{
			/*
				When Weak Napt Control is turned on, for all non-SYN packets without entry we change the actionFlag
				for them to add new entries.Besides we modify initFlag to generate the ESTABLISH flow for these packets.
			*/
			initFlag = TCP_ESTABLISHED_FLOW;
			actionFlag |= TCP_OUT_TCP_SYN;
		}else
		{
			actionFlag = 0;
		}
	}
#else
	if ((actionFlag&(TCP_OUT_TCP_SYN|TCP_OUT_TCP_ENTRY))==0)
	{
		if ((actionFlag == TCP_OUT_SERVER_PORT_ENTRY) || (actionFlag == TCP_OUT_UPNP_ENTRY))
		{
			initFlag = TCP_ESTABLISHED_FLOW;
		}else
		{
			actionFlag = 0;
		}
	}
#endif /* _RTL_WEAK_NAPT_CTRL */
/*
  * All of the following case is the same as case 3 (TCP_OUT_TCP_SYN|TCP_OUT_TCP_ENTRY).
  *
  *	case 7: // SYN + Entry + Server Port(software)
  *	case 11: // SYN + Entry + Client ALG
  *	case 15: // SYN + Entry + Server Port + Client ALG 
  *	case 19: // SYN + Entry + Server ALG
  *	case 23: // SYN + Entry + Server Port + Server ALG 
  *	case 27: // SYN + Entry +  Client ALG + Server ALG
  *	case 31: // SYN + Entry + Server Port + Client ALG + Server ALG
  *
  *	case 35: // SYN + Entry + uPnP
  *	case 43: // SYN + Entry + uPnP + Client ALG 
  *	case 51: // SYN + Entry + uPnP + Server ALG 
  *	case 59: // SYN + Entry + uPnP + Client ALG + Server ALG
  *
  */
	if((actionFlag&(TCP_OUT_TCP_SYN|TCP_OUT_TCP_ENTRY)) == (TCP_OUT_TCP_SYN|TCP_OUT_TCP_ENTRY))
	{
		actionFlag = (TCP_OUT_TCP_SYN|TCP_OUT_TCP_ENTRY);
	}

	/**********************************************************************

		Process Dispatcher
	
	 **********************************************************************/

	switch ( actionFlag )
	{
		case 0:

		/*********************************************************************************/

			_rtl8651_dosScanCheckTcpOutboundFailUnusual(phdr->ph_dsid, pip, tc);
			/* client no entry found */
			return FWDENG_TCP_NAPT_OUT_NOTFOUND;

		/*********************************************************************************/

		case 4:	/* Server Port, non-SYN */
		case 5:	/* SYN + Server Port */
		case 21:	/* SYN + Server Port + Server ALG(erroneous judgement: the src port is touch the ALG table and Server Port table) tony:2004/2/11 */

		case 32:	/* upnp + non-SYN, chenyl: add for upnp */
		case 33:	/* upnp + SYN, chenyl: add for upnp */
		case 49:	/* upnp + SYN + Server ALG, chenyl: add for upnp */

		/*********************************************************************************/
			if (tb2)
			{
				assert(isUpnpMapping == FALSE);
				ip_alias = tb2->globalIp;
				ip_aliasp = ip_sp - tb2->localPort + tb2->globalPort;
			} else
			{
				assert(isUpnpMapping == TRUE);
				ip_alias = upnpExtIp;
				ip_aliasp = upnpExtPort;
			}

			if((retval = _rtl8651_addNaptConnection(TRUE,
					(TRUE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
					initFlag, ip_s, ip_sp, &ip_alias, &ip_aliasp, ip_d, ip_dp)) != SUCCESS)
			{
				return FWDENG_TCP_NAPT_OUT_ADDFAIL;
			}
			tb=&entry;
			if(_rtl8651_getOutboundNaptFlow(TRUE, ip_s, ip_sp, ip_d, ip_dp, &tb)!=SUCCESS)
			{
				rtlglue_printf("error: _rtl8651_getOutboundNaptFlow connection Failed \n");
 				return FWDENG_TCP_NAPT_OUT_NOTFOUND;
			}
			tb->	tcpAckModified = 0;
			tb->	tcpFtpLastLineCRLFTermed = 0;
			tb->tcpStateINdex = 0;
			memset(&tb->ack[0],0,sizeof(struct l4ack_data_record)*L4_N_LINK_TCP_DATA);
			break;

		/*********************************************************************************/

		case 1:
		case 17: /* SYN + Server ALG(erroneous judgement: the src port is touch the ALG table) tony:2004/2/11 */

		/*********************************************************************************/

			_rtl8651_softNaptPolicyRoute(TRIGGER_DSTTCPPORT, ip_s, ip_d, ip_dp, &ip_alias, &ip_aliasp);
#ifdef _RTL_NAPT_ADVANCED
			if (actionFlag == 1)
			{	/* pure normal outbound napt flow, it might apply InexactTCP or Static port translation */
				if (ip_alias)
				{	/* policy route */
					if ((rtl8651_drvStaticTcpNaptPortTranslation == TRUE) &&
						(_rtl8651_l4CheckOccupiedConnection(TRUE, phdr->ph_dsid, ip_s, ip_sp, ip_alias, ip_sp, ip_d, ip_dp) == FALSE))
					{
						if (_rtl8651_addNaptConnection(TRUE,
								(TRUE|PURE_SW|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL|DO_REDUNDANCY_CHECK),
								initFlag, ip_s, ip_sp, &ip_alias, &ip_sp, ip_d, ip_dp/*(rtl8651_drvInexactTcpNaptEnable==TRUE?0:ip_d), (rtl8651_drvInexactTcpNaptEnable==TRUE?0:ip_dp)*/)
								== SUCCESS)
							ip_aliasp = ip_sp;
						else
							goto TcpPolicyDynamicPortAssign;
					}else
					{
TcpPolicyDynamicPortAssign:
						if( _rtl8651_addNaptConnection(TRUE,
								(TRUE|PURE_SW|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
								initFlag, ip_s, ip_sp, &ip_alias, &ip_aliasp,
								ip_d, ip_dp/*(rtl8651_drvInexactTcpNaptEnable==TRUE?0:ip_d), (rtl8651_drvInexactTcpNaptEnable==TRUE?0:ip_dp)*/)
								!= SUCCESS)
							return FWDENG_TCP_NAPT_OUT_ADDFAIL;
					}
				}else
				{	/* normal route */
					if (rtl8651_drvStaticTcpNaptPortTranslation == TRUE)
					{
						ip_alias = _rtl8651_getNaptInsideGlobalIpAddr(TRUE, ip_s, ip_sp, ip_d, ip_dp);

						if ((!ip_alias) ||(_rtl8651_l4CheckOccupiedConnection(TRUE, phdr->ph_dsid, ip_s, ip_sp, ip_alias, ip_sp, ip_d, ip_dp) == TRUE))
							goto TcpNormalDynamicPortAssign;

						if (_rtl8651_addNaptConnection(TRUE,
								(TRUE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL|DO_REDUNDANCY_CHECK),
								initFlag, ip_s, ip_sp, &ip_alias, &ip_sp, ip_d, ip_dp/*(rtl8651_drvInexactTcpNaptEnable==TRUE?0:ip_d), (rtl8651_drvInexactTcpNaptEnable==TRUE?0:ip_dp)*/)
								== SUCCESS)
							ip_aliasp = ip_sp;
						else
							goto TcpNormalDynamicPortAssign;
					}else
					{
TcpNormalDynamicPortAssign:
						if(_rtl8651_addNaptConnection(TRUE,
								(FALSE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
								initFlag, ip_s, ip_sp, &ip_alias, &ip_aliasp, ip_d, ip_dp/*(rtl8651_drvInexactTcpNaptEnable==TRUE?0:ip_d), (rtl8651_drvInexactTcpNaptEnable==TRUE?0:ip_dp)*/)
								!= SUCCESS)
							return FWDENG_TCP_NAPT_OUT_ADDFAIL;
					}
				}
			}else
			{	/* other cases, including ALG...etc */
				if (ip_alias)
				{
					if( _rtl8651_addNaptConnection(TRUE,
							(TRUE|PURE_SW|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
							initFlag, ip_s, ip_sp, &ip_alias, &ip_aliasp, ip_d, ip_dp) != SUCCESS)
						return FWDENG_TCP_NAPT_OUT_ADDFAIL;
				}
				else
				{
					/* client need add new entry */
					if( _rtl8651_addNaptConnection(TRUE,
							(FALSE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
							initFlag, ip_s, ip_sp, &ip_alias, &ip_aliasp, ip_d, ip_dp) != SUCCESS )
						return FWDENG_TCP_NAPT_OUT_ADDFAIL;
				}
			}
#else /* _RTL_NAPT_ADVANCED */
			if (ip_alias)
			{
				if( _rtl8651_addNaptConnection(TRUE,
							(TRUE|PURE_SW|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
							sinitFlag, ip_s, ip_sp, &ip_alias, &ip_aliasp, ip_d, ip_dp) != SUCCESS)
					return FWDENG_TCP_NAPT_OUT_ADDFAIL;
			}else
			{	/* client need add new entry */
				if( _rtl8651_addNaptConnection(TRUE,
						(FALSE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
						initFlag, ip_s, ip_sp, &ip_alias, &ip_aliasp, ip_d, ip_dp) != SUCCESS)
					return FWDENG_TCP_NAPT_OUT_ADDFAIL;
			} /* end _rtl8651_softNaptPolicyRoute*/
#endif /* _RTL_NAPT_ADVANCED */
			tb = &entry;
			if(_rtl8651_getOutboundNaptFlow(TRUE, ip_s, ip_sp, ip_d, ip_dp, &tb) == SUCCESS)
			{
				tb->	tcpAckModified = 0;
				tb->	tcpFtpLastLineCRLFTermed = 0;
				tb->tcpStateINdex = 0;
				memset(&tb->ack[0],0,sizeof(struct l4ack_data_record)*L4_N_LINK_TCP_DATA);
				break;
			}
			return FWDENG_TCP_NAPT_OUT_NOTFOUND;

		/*********************************************************************************/

		case 2:	/* Entry */
		case 6:	/* Entry + Server Port */

		case 34:	/* Entry + upnp */

		/*********************************************************************************/

/* tony: this path only for h245 special case, the port info is not save in ALG table. 
/ but must handle the packet with ALG. h245 only use alg_client_out, 			*/
#if 0
			if(tb->alg_server_out){
				int r;
				r = tb->alg_server_out(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}
#endif
			if(tb->alg_client_out)
			{
				int r;

				r = tb->alg_client_out(phdr, pip, tb);

				if(r == FWDENG_USER_DROP)
				{
					return FWDENG_USER_DROP;
				}
			}
			break;

		/*********************************************************************************/

		case 3:

		/*********************************************************************************/

			break;
#if 0 //tony disabled, case 5, 33 move to case 1			

		/*********************************************************************************/

		case 5:

		case 33:

		/*********************************************************************************/

			/* server port found; but no entry found*/
			/* direct send the packet */
			if (tb2)
			{
				tc->th_sport = ip_sp-tb2->localPort+tb2->globalPort;
				PKTSET_UINT32_UNALIGNED(tb2->globalIp, &pip->ip_src);
			}else
			{
				tc->th_sport = upnpExtPort;
				PKTSET_UINT32_UNALIGNED(upnpExtIp, &pip->ip_src);
			}

			return (SUCCESS);						

		/*********************************************************************************/

#endif

		case 9:	/* SYN + Client ALG */
		case 13:	/* SYN + Server Port + Client ALG */

		case 41:	/* SYN + upnp + client ALG */

		/*********************************************************************************/

			if (client_alg_link->fwEngAdd)
			{
				_rtl8651_softNaptPolicyRoute(TRIGGER_DSTTCPPORT,ip_s,ip_d,ip_dp,&ip_alias,&ip_aliasp);
				if (ip_alias)
				{
					tb = _rtl8651_addAlgCtrlConnection(
						(TRUE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
						initFlag, ip_s, ip_sp, &ip_alias, &ip_aliasp, ip_d, ip_dp);
				}else
				{
					tb = _rtl8651_addAlgCtrlConnection(
						(FALSE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
						initFlag, ip_s, ip_sp, &ip_alias, &ip_aliasp, ip_d, ip_dp);
				}
					if(!tb)
						return FWDENG_TCP_NAPT_OUT_CLIALGADDFAIL;
			}

			tb=&entry;
			if(_rtl8651_getOutboundAlgCtrlFlow(TRUE, ip_s, ip_sp, ip_d, ip_dp, &tb) == SUCCESS)
			{
				tb->	tcpAckModified = 0;
				tb->	tcpFtpLastLineCRLFTermed = 0;
				tb->tcpStateINdex = 0;
				memset(&tb->ack[0],0,sizeof(struct l4ack_data_record)*L4_N_LINK_TCP_DATA);
				break;
			}
			return FWDENG_TCP_NAPT_OUT_ALGNOTFOUND;								

		/*********************************************************************************/

		case 10:	/* Entry + Client ALG */
		case 14:	/* Entry + Client ALG + Server Port */

		case 42:	/* Entry + Client ALG + upnp */

		/*********************************************************************************/

			 if(client_alg_link->fwEngAdd && client_alg_link->alg_outboundClientcb)
			 {
			 	int r;
				r=client_alg_link->alg_outboundClientcb(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			 }
			 break;

		/*********************************************************************************/

#if 0 //tony: do not have this path, move to case 17			 
		case 21:

		case 49:

		/*********************************************************************************/

			/* add server port alg link */
			{
				uint16 globalPort;
				ipaddr_t globalIp;
				if (tb2)
				{
					assert(isUpnpMapping == FALSE);
					globalPort = tb2->globalPort;
					globalIp = tb2->globalIp;
				}else
				{
					assert(isUpnpMapping == TRUE);
					globalPort = upnpExtPort;
					globalIp = upnpExtIp;
				}
				tb=_rtl8651_addAlgCtrlConnection(
					(TRUE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),
					initFlag,ip_s,ip_sp,&globalIp,&globalPort,ip_d,ip_dp);
			}
			if(!tb)
			{
				return FWDENG_TCP_NAPT_OUT_SVRALGADDFAIL;
			}
			tb->	tcpAckModified = 0;
			tb->	tcpFtpLastLineCRLFTermed = 0;
			tb->tcpStateINdex = 0;			
			memset(&tb->ack[0],0,sizeof(struct l4ack_data_record)*L4_N_LINK_TCP_DATA);
			break;

		/*********************************************************************************/

#endif

		case 18: /* Entry + Server ALG */
		case 22: /* Entry + Server ALG + Server Port */

		case 50:	/* Entry + Server ALG + upnp */

		/*********************************************************************************/

			 if(server_alg_link->fwEngAdd && server_alg_link->alg_outboundServercb)
			 {
			 	int r;
				r = server_alg_link->alg_outboundServercb(phdr, pip, tb);
				if( r == FWDENG_USER_DROP )
				{
					return FWDENG_USER_DROP;
				}
			 }
			break;

		/*********************************************************************************/

		default:	/* this should not happen */
			rtlglue_printf("%s: action=%d, flag=%d\n",__FUNCTION__,actionFlag,tc->th_flags );		
			return FWDENG_TCP_NAPT_OUT_IGNORE;
	}
	assert(tb);	
	assert(phdr->ph_routeIdx);

#ifdef _RTL_WEAK_NAPT_CTRL
	/*
		execute Alg control block for weak napt flows
	*/
	if (	rtl8651_drvWeakTcpNaptEnable &&
		(initFlag == TCP_ESTABLISHED_FLOW) &&
		(actionFlag != TCP_OUT_SERVER_PORT_ENTRY) &&
		(actionFlag != TCP_OUT_UPNP_ENTRY)
	)
	{
		if (actionFlag & TCP_OUT_SERVER_ALG)
		{
			 if(server_alg_link->fwEngAdd && server_alg_link->alg_outboundServercb)
			 {
			 	int r;
				r=server_alg_link->alg_outboundServercb(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			 }
		}
		if (actionFlag & TCP_OUT_CLIENT_ALG)
		{
			 if(client_alg_link->fwEngAdd && client_alg_link->alg_outboundClientcb)
			 {
			 	int r;
				r=client_alg_link->alg_outboundClientcb(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			 }
		}
	}
#endif /* _RTL_WEAK_NAPT_CTRL */

	/* store cache */
	tb->out_rt_t = (void *)&DrvTbl.route[phdr->ph_routeIdx-1];	
	tb->dsid = phdr->ph_dsid;

	/*************************************************************************************/
	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;
	ip_sp = ntohs(tc->th_sport);
	ip_dp =  ntohs(tc->th_dport);
		
	ip_alias = tb->insideGlobalIpAddr;
	ip_aliasp = tb->insideGlobalPort;
	
	/* Monitor TCP connection state */
	_rtl8651_l4TcpMonitorOut(tc, tb);

	/* tcp sync packet  adj mss value */		
	if(tc->th_flags & TH_SYN){
		int8 auto_add = FALSE,auto_delete;			
		rtl8651_getAsicNaptAutoAddDelete(&auto_add,&auto_delete);			
		if(auto_add == FALSE){							
			uint8 *cp;
			int32 opt, optlen,cnt;		
			uint16 mss;
		
			cp = (uint8 *)(tc+1);
			cnt = ((tc->th_off_x >>2)&0x3c)  -sizeof(struct tcphdr);//Edward: Fix for alignment issue
	
			for(; cnt > 0; cnt -= optlen, cp+= optlen){
				opt = cp[0];
				if(opt == TCPOPT_EOL)
					break;
				if(opt == TCPOPT_NOP)
					optlen = 1;
				else{
					optlen = cp[1];
					if(optlen <= 0)
						break;
				}

				switch (opt) {
				case TCPOPT_MAXSEG:
					mss = ntohs(*(uint16*)(&cp[2]));
//rtlglue_printf("call _rtl8651_getPppoeSessionByIpAddr()....4(l4tcpNaptAliasOut)\n");
					//thisPPPoE=_rtl8651_getPppoeSessionByIpAddr(PKTGET_UINT32_UNALIGNED(&ip_alias));
					thisPPPoE=_rtl8651_getPppoeSessionByIpAddr(ip_alias);
					if (thisPPPoE)
					{
						if (mss >pppoe_mss[thisPPPoE->dsid]) 
						{ 
							mtu=pppoe_mss[thisPPPoE->dsid];
							accumulate -=(mss-pppoe_mss[thisPPPoE->dsid]);
							*(uint16*)(&cp[2]) = htons(mtu);
						}
					}else if(mss > rtl8651_mss){
						mtu=rtl8651_mss;
						accumulate -=(mss-rtl8651_mss);
						*(uint16*)(&cp[2]) = htons(mtu);
					}
					break;
				default:
					continue;
				}			
			}				
		}
	}

	/* Modify sequence number if necessary */
	if (tb->tcpAckModified == 1)
	{
		int32     delta;
		
		delta = _rtl8651_l4GetDeltaSeqOut(pip, tc, tb);
		accumulate += delta;
		if (delta != 0)
		{
			uint32 seq_32_x = PKTGET_UINT32_UNALIGNED(&tc->th_seq);
			seq_32_x  = htonl(ntohl(seq_32_x) + delta);
			PKTSET_UINT32_UNALIGNED( seq_32_x,&tc->th_seq);
		}
	}

	/* chenyl : url filter checking is deferred to slow path */
#ifdef RTL865XB_URL_FILTER
	if((_urlFilterAclRuleNum>=0) &&
		((phdr->ph_reason&0xe)==4
		&& ((phdr->ph_reason&0xff0)>>5)==_urlFilterAclRuleNum))
#else
	if(!tb->urlFiltered||phdr->ph_reason==65)
#endif
	{
		int32 position, len=64;
		int8 matched[64];
		int32 hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >>4)&0xf)) << 2;
		if ((ntohs(pip->ip_len))>hlen)
		{
			position=_rtl8651_urlFilter(phdr->ph_dsid,phdr,pip,RTL8651_LAN2WAN,matched, &len);
			if (position>0)
			{
				/*
					cfliu: If you prepare a string buffer for the filter function, 
					rtl8651_urlFilter() would return string matched, its length and its position found in mbuf chain
					you may insert a log function pointer here to keep this extra info
					or just return a error value if you don't care.
				*/
				#ifdef RTL865X_DEBUG
				rtlglue_printf("string '%s'(%d bytes) in URL at %dth byte. Filtered...\n",matched, len, position );
				#endif
				_rtl8651_updateNaptConnection(tb,TCP_CLOSED_FLOW, 3);	/* keep entry for 3 seconds. */
				return FWDENG_TCP_URLFILTERED;
			}
#ifndef RTL865XB_URL_FILTER
			tb->urlFiltered = TRUE;	/* so we won't check this flow anymore. */
#endif
		}
	}

	/*******************************************************************************************/
	if ((!(actionFlag&TCP_OUT_TCP_ENTRY)) && tb)
	{
		assert(tb->dosEnCount == TRUE && tb->dosFromInternal == TRUE);
	
		// chenyl: 2004/03/08 log for newly-added connection
		if ((_loggingModule[phdr->ph_dsid]&RTL8651_LOGMODULE_NEWFLOW)&&_pUserLoggingFunc)
		{
#ifdef _RTL_NEW_LOGGING_MODEL
			rtl8651_logInfo_t info;
			info.infoType		= RTL8651_LOG_INFO_PKT;
			info.pkt_dsid		= phdr->ph_dsid;
			info.pkt_sip		= ip_s;
			info.pkt_dip		= ip_d;
			info.pkt_proto		= IPPROTO_TCP;
			info.pkt_direction	= 0;	/* direction: LAN to WAN */
			info.pkt_sport		= ip_sp;
			info.pkt_dport		= ip_dp;
			info.action		= RTL8651_LOGACTION_FORWARD;
			(* _pUserLoggingFunc)(RTL8651_LOGMODULE_NEWFLOW, RTL8651_LOG_NEWFLOW_NewTcpNaptOutbound, &info);
#else
			(*_pUserLoggingFunc)(
				phdr->ph_dsid,
				RTL8651_LOGMODULE_NEWFLOW,
				    (unsigned char)pip->ip_p, /* protocol */
				    (unsigned char)0, /* direction: LAN to WAN */
					(unsigned long)ip_s,
					(unsigned long)ip_d,
					(unsigned short)ip_sp,
					(unsigned short)ip_dp,
					0, /* type: only meaningful for ICMP */
					RTL8651_LOGACTION_FORWARD, /* action */
					"New TCP Outbound Flow" /* msg */);
#endif
		}
	}
	/*******************************************************************************************/

#ifdef _RTL_NAPT_FRAG
	if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK)) == IP_MF))
	{
		if ((pip->ip_vhl & 0x0f) != 5)
		{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
			/* for fragment packet, it don't need to calculate L3 checksum */
			_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, ip_s, pip->ip_sum);	/* L3 Checksum */
		}
		/* for both fragment packet and non-ASIC offloaded packet, we must calculate L4 checksum by our self */
		_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, ip_s, ip_aliasp, ip_sp, tc->th_sum);	/* L4 Checksum */
		_RTL8651_ADJUST_CHECKSUM(accumulate, tc->th_sum);	/* mss modification and delta modification */

		if (IS_FRAGMENT(pip->ip_off))
		{	/* first packet of fragment flow */
			_rtl8651_updateFragCache(OUTBOUND_FRAG, phdr->ph_sip, phdr->ph_dip, ntohs(pip->ip_id), ip_alias, phdr->ph_routeIdx, NULL, phdr->ph_dsid);
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
			_rtl8651_fragmentShortestPathCacheSet_firstHalf(phdr, (uint8*)tc, pip, ip_s, ip_d, RTL8651_FRAGMENT_SHORTESTPATH_MODE_MODIFYSIP, ip_alias);
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
		}
	}
#endif /* _RTL_NAPT_FRAG */	

	/* source port is being altered */
	tc->th_sport = htons(ip_aliasp);
	/* source address is being altered */
	PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);

	return (SUCCESS);
}

/*********************************************************************************
* Routine Name :  _rtl8651_l4TcpNatAlgAliasOut                                   
* Description : 
*      tcp packet output process
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4TcpNatAlgAliasOut(struct rtl_pktHdr *phdr, struct ip *pip, rtl8651_tblDrv_natEntry_t *natEntry)
{
	struct tcphdr *tc;
	ipaddr_t ip_s, ip_d, ip_alias = 0;
	uint16 ip_sp, ip_dp, ip_aliasp = 0;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb=&entry;
	rtl8651_tblDrv_algEntry_t *client_alg_link = NULL; 
	rtl8651_tblDrv_algEntry_t *server_alg_link = NULL; 
	int32 accumulate = 0;
	uint32 actionFlag = 0;
	rtl8651_tblDrv_pppTable_t *thisPPPoE;
	uint16 mtu;

#define TCP_OUT_TCP_SYN 0x01
#define TCP_OUT_TCP_ENTRY 0x02
#define TCP_OUT_SERVER_PORT_ENTRY 0x04
#define TCP_OUT_CLIENT_ALG 0x08
#define TCP_OUT_SERVER_ALG	0x10

	tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));

	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;	
	ip_sp = ntohs(tc->th_sport);
	ip_dp =  ntohs(tc->th_dport);

	if (ip_s == natEntry->globalIp)
	{	/* NI == NE, so no ALG is needed */
		ip_alias = natEntry->globalIp;
		tb = NULL;
		goto adjust_mss;
	}

	client_alg_link = _rtl8651_l4FindLinkAlg(ip_dp, TRUE);		
	if(ip_sp != ip_dp)
	{
		server_alg_link = _rtl8651_l4FindLinkAlg(ip_sp, TRUE);	
	}

	//chenyl: mark for mss adjust
//	if(!client_alg_link && !server_alg_link){
//		return FWDENG_TCP_NAT_OUT_ALGNOTFOUND;
//	}
	
	if( _rtl8651_getOutboundAlgCtrlFlow(TRUE,ip_s,ip_sp,ip_d,ip_dp,  &tb)!=SUCCESS)
	{
		tb=NULL;
	}
	
	//if(tc->th_flags == TH_SYN)
	if(tc->th_flags & TH_SYN)
		actionFlag |= TCP_OUT_TCP_SYN;
	
	if(tb)
		actionFlag |= TCP_OUT_TCP_ENTRY;	
		 
	if(client_alg_link)
		actionFlag |= TCP_OUT_CLIENT_ALG;
		
	if(server_alg_link)
		actionFlag |= TCP_OUT_SERVER_ALG;

#ifdef _TONY_DEBUG
		rtlglue_printf("l4TcpNatAlgAliasOut src:%u.%u.%u.%u:%d dest:%u.%u.%u.%u:%d actionFlag=%d\n",NIPQUAD(ip_s),ip_sp,NIPQUAD(ip_d),ip_dp,actionFlag);
#endif		

	switch(actionFlag){

		case 2:
		/* cleint entry found */
			/* entry already exist */
			if(tb->alg_client_out){
				int r;
				r = tb->alg_client_out(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}
			if(tb->alg_server_out){
				int r;
				r = tb->alg_server_out(phdr, pip, tb);
				if(r==FWDENG_USER_DROP) {
					return FWDENG_USER_DROP;
				}
			}						
			break;
		case 9: /* client alg need add new entry */
		case 17 /* server alg need add new entry */:
			
			ip_alias = natEntry->globalIp;
			ip_aliasp = ip_sp;
			tb= _rtl8651_addAlgCtrlConnection(
				(TRUE|DOS_NAPT_ENCOUNT|DOS_NAPT_FROMINTERNAL),TCP_OUTBOUND_SYN_RCVD_FLOW,
				ip_s,ip_sp, &ip_alias,&ip_aliasp,ip_d,ip_dp);
			if(!tb)
				return FWDENG_TCP_NAT_OUT_ADDFAIL;

			tb->tcpAckModified = 0;
			tb->tcpFtpLastLineCRLFTermed = 0;
			tb->tcpStateINdex = 0;			
			memset(&tb->ack[0],0,sizeof(struct l4ack_data_record)*L4_N_LINK_TCP_DATA);
			tb->natlink = 1;
			break;

		case 10:
		case 11:
			 /*ALG Client process*/
			 if(client_alg_link->fwEngAdd && client_alg_link->alg_outboundClientcb){
			 	int r;
				r=client_alg_link->alg_outboundClientcb(phdr, pip, tb);	
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			 }
			 break;

		case 18:
		case 19:
			 /*ALG Server process*/
			 if(server_alg_link->fwEngAdd && server_alg_link->alg_outboundServercb){
			 	int r;
				r=server_alg_link->alg_outboundServercb(phdr, pip, tb);	
				if(r==FWDENG_USER_DROP) return FWDENG_USER_DROP;
			 }
			 break;
		case 1:
			ip_alias = natEntry->globalIp;
			goto adjust_mss;
		case 0:
			return FWDENG_TCP_NAT_OUT_ALGNOTFOUND;
		default:	/* this should not happen */
			rtlglue_printf("%s: action=%d, flag=%d\n",__FUNCTION__,actionFlag,tc->th_flags );		
			return FWDENG_TCP_NAT_OUT_ALGIGNORE;
	}
	
	assert(tb);
	ip_alias = tb->insideGlobalIpAddr;
	ip_aliasp = tb->insideGlobalPort;

	/* Monitor TCP connection state */
	_rtl8651_l4TcpMonitorOut(tc, tb);

	/* tcp sync packet  adj mss value */		
adjust_mss:
	if(tc->th_flags & TH_SYN)
	{
		int8 auto_add = FALSE,auto_delete;
		rtl8651_getAsicNaptAutoAddDelete(&auto_add,&auto_delete);			
		if(auto_add == FALSE){							
			uint8 *cp;
			int32 opt, optlen,cnt;		
			uint16 mss;
		
			cp = (uint8 *)(tc+1);
			cnt = ((tc->th_off_x >>2)&0x3c)  -sizeof(struct tcphdr);//Edward: Fix for alignment issue
	
			for(; cnt > 0; cnt -= optlen, cp+= optlen){
				opt = cp[0];
				if(opt == TCPOPT_EOL)
					break;
				if(opt == TCPOPT_NOP)
					optlen = 1;
				else{
					optlen = cp[1];
					if(optlen <= 0)
						break;
				}

				switch (opt) {
				case TCPOPT_MAXSEG:
					mss = ntohs(*(uint16*)(&cp[2]));
//rtlglue_printf("call _rtl8651_getPppoeSessionByIpAddr()....5(l4tcpNatAliasOut)\n");
					//thisPPPoE=_rtl8651_getPppoeSessionByIpAddr(PKTGET_UINT32_UNALIGNED(&ip_alias));
					thisPPPoE=_rtl8651_getPppoeSessionByIpAddr(ip_alias);
					// pppoe session might not be able to be decided due
					// to napt not turned on, has to filter this
					// condition further ...
					if (thisPPPoE)
					{	// pppoe wan config
						if (mss >pppoe_mss[thisPPPoE->dsid])
						{
							mtu=pppoe_mss[thisPPPoE->dsid];
							accumulate -=(mss-pppoe_mss[thisPPPoE->dsid]);
							*(uint16*)(&cp[2]) = htons(mtu);
						}						
					}else if(mss > rtl8651_mss){
						// static/dhcpc wan config						
						mtu=rtl8651_mss;
						accumulate -=(mss-rtl8651_mss);
						*(uint16*)(&cp[2]) = htons(mtu);
					}
					break;
				default:
					continue;
				}			
			}				
		}
	}

	/* and source address is being altered */
	if(tb){
		tc->th_sport = ip_aliasp;
		/* Modify sequence number if necessary */
		if (tb->tcpAckModified == 1)
		{
			int32     delta;
			
			delta = _rtl8651_l4GetDeltaSeqOut(pip, tc, tb);
			accumulate += delta;
			if (delta != 0){
				uint32 seq_32_x = PKTGET_UINT32_UNALIGNED(&tc->th_seq);
				seq_32_x  = htonl(ntohl(seq_32_x) + delta);
				PKTSET_UINT32_UNALIGNED( seq_32_x,&tc->th_seq);
			}
		}
		PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);
	}else{ //NAT, not ALG....
		assert(natEntry);
		PKTSET_UINT32_UNALIGNED(natEntry->globalIp, &pip->ip_src);
	}	
#ifdef _RTL_NAPT_FRAG			
	if(tb)
	{	/* ALG found */
		if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF | IP_OFFMASK)) == IP_MF))
		{
			/* if ip header length != 20, we must calculate L3/4 checksum by ourself */
			if ((pip->ip_vhl & 0x0f) != 5)
			{
				_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, ip_s, pip->ip_sum);
			}
			/* L4 Checksum need to be calculated for first fragment packet and non-offloaded packet */
			_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, ip_s, ip_aliasp, ip_sp, tc->th_sum);
			_RTL8651_ADJUST_CHECKSUM(accumulate, tc->th_sum);	/* mss modification and delta modification */
		}
	} else
	{	/* ALG Not found, but we come here because this is SYN packet */
		ip_alias = natEntry->globalIp;

		if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF | IP_OFFMASK)) == IP_MF))
		{
			/* if ip header length != 20, we must calculate L3/4 checksum by ourself */
			if ((pip->ip_vhl & 0x0f) != 5)
			{
				_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, ip_s, pip->ip_sum);
			}
			/* L4 Checksum need to be calculated for first fragment packet and non-offloaded packet */
			/* we only modify ip, so we don't adjust csum for port modify */
			_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, ip_s, 0, 0, tc->th_sum);
			_RTL8651_ADJUST_CHECKSUM(accumulate, tc->th_sum);	/* mss modification and delta modification */
		}
	}
#endif /* _RTL_NAPT_FRAG */	

	return (SUCCESS);
}

#ifdef _RTL_NAPT_SHORT_CUT			
/*********************************************************************************
* Routine Name : _rtl8651_l4TcpFastAliasIn                                    
* Description : 
*   TCP packet input process
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4TcpFastAliasIn(struct rtl_pktHdr *phdr, struct ip *pip)
{
	struct tcphdr *tc;
	ipaddr_t ip_org;
	ipaddr_t ip_s,  ip_d;
	uint16 ip_sp, ip_dp, ip_orgp;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb;
	
	ip_s=phdr->ph_sip;
	ip_d=phdr->ph_dip;
	
#ifdef _TONY_DEBUG
{
	uint32 sip=GET_UINT32_UNALIGNED(&pip->ip_src);
	uint32 dip=GET_UINT32_UNALIGNED(&pip->ip_dst);	
	rtlglue_printf("l4TcpFastAliasIn src:%u.%u.%u.%u:%d dest:%u.%u.%u.%u:%d seq=%u\n",NIPQUAD(sip),ip_sp,NIPQUAD(dip),ip_dp,GET_UINT32_UNALIGNED(&tc->th_seq));
}
#endif

#ifdef _RTL_NAPT_FRAG
		if (IS_FRAGMENT_REMAIN(pip->ip_off))
		{	/* deal with fragmented packet */
			uint32 org_ip;
			if (_rtl8651_getFragCache(INBOUND_FRAG, ip_s, ip_d, ntohs(pip->ip_id), &org_ip, &(phdr->ph_routeIdx), NULL, &(phdr->ph_dsid)) == SUCCESS)
			{
				PKTSET_UINT32_UNALIGNED(org_ip, &pip->ip_dst);
				if (!IPHDRLEN_IS_TWENTY(pip))
				{	/* Adjust L3 checksum by ourself */
					_RTL8651_L3_ADJUST_CHECKSUM(org_ip, ip_d, pip->ip_sum);
				}
				if (_rtl8651_l3FastActRouting(FWDENG_FROMDRV, phdr, org_ip, pip, NULL) == SUCCESS)
					return SUCCESS;
			}
			return FAILED;
		}
#endif /* _RTL_NAPT_FRAG */

	tb = &entry;
	tc = (struct tcphdr *)L4HDR_ADDR(pip);
	ip_sp = ntohs(tc->th_sport);
	ip_dp = ntohs(tc->th_dport);

	if (tc->th_flags & (TH_SYN|TH_FIN|TH_RST))
		return FAILED;

#ifdef _RTL_NAPT_FRAG
	if ((IS_FRAGMENT(pip->ip_off)) && (_rtl8651_fragQueueProcessOk == FALSE) && (_rtl8651_frag_NeedFragmentProcess(ntohs(pip->ip_id), ip_s, ip_d, IPPROTO_TCP) == TRUE))
		return FAILED;
#endif /* _RTL_NAPT_FRAG */
	
	if(_rtl8651_getInboundNaptFlow(TRUE,ip_d,ip_dp,ip_s,ip_sp,  &tb)!=SUCCESS || tb->inListType==NAPT_ALG_IN){
		return FAILED;
	}

	if (tb->toProtocolStack == TRUE)
	{	/* we can't process toProtocolStack here, for fragment queue process */
		return FAILED;
	}

	if(tb->alg_client_out || tb->alg_server_out || tb->alg_client_in || tb->alg_server_in ) 
		return FAILED; 

	if(tb->tcpFlag != (TCP_ESTABLISHED_FLOW-1))
	{
		return FAILED;
	}

#ifdef RTL865XB_URL_FILTER
	if ((_urlFilterAclRuleNum>=0) &&
		((phdr->ph_reason&0xe)==4
		&& ((phdr->ph_reason&0xff0)>>5)==_urlFilterAclRuleNum))
#else
	if (!tb->urlFiltered||phdr->ph_reason==65)
#endif
	{
		return FAILED;
	}

	ip_org = tb->insideLocalIpAddr;
	ip_orgp = tb->insideLocalPort;

	/* put otiginal port number back */
	tc->th_dport = htons(ip_orgp);

	/* put otiginal ip addr back */
	PKTSET_UINT32_UNALIGNED(ip_org, &pip->ip_dst);

	/* get dsid */
	phdr->ph_dsid = tb->dsid;

	if (tb->in_rt_t)
	{
		phdr->ph_routeIdx = L3_TABLE_INDEX((rtl8651_tblDrv_routeTable_t *)tb->in_rt_t) + 1;
	}else
	{
		rtl8651_tblDrv_routeTable_t *rt_entry;
		if (((rt_entry = _rtl8651_getSoftRoutingEntry(ip_s, ip_org, IPPROTO_TCP, ip_orgp)) == NULL) &&
			((rt_entry = _rtl8651_getRoutingEntry(ip_org)) == NULL))
			return FWDENG_DROP;
		tb->in_rt_t = rt_entry;
		phdr->ph_routeIdx = L3_TABLE_INDEX(rt_entry) + 1;
	}

	if(_rtl8651_l4TcpMonitorIn(tc, tb)!=SUCCESS)	 /* Monitor TCP connection state */
	{
		return FWDENG_DROP;
	}

	if ((IS_FRAGMENT(pip->ip_off)) || (!IPHDRLEN_IS_TWENTY(pip)))
	{
		if (!IPHDRLEN_IS_TWENTY(pip))
		{
			_RTL8651_L3_ADJUST_CHECKSUM(ip_org, ip_d, pip->ip_sum);
		}
		_RTL8651_L4_ADJUST_CHECKSUM(ip_org, ip_d, ip_orgp, ip_dp, tc->th_sum);
		if (IS_FRAGMENT(pip->ip_off))
		{
			_rtl8651_updateFragCache(INBOUND_FRAG, ip_s, ip_d, ntohs(pip->ip_id), ip_org, phdr->ph_routeIdx, NULL, phdr->ph_dsid);
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
			_rtl8651_fragmentShortestPathCacheSet_firstHalf(phdr, (uint8*)tc, pip, ip_s, ip_d, RTL8651_FRAGMENT_SHORTESTPATH_MODE_MODIFYDIP, ip_org);
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
		}
	}

	if( _rtl8651_l3FastActRouting(FWDENG_FROMDRV, phdr, ip_org, pip, &tb->in_macInfo_t) == SUCCESS)
	{
		return SUCCESS;
	}

	return FWDENG_DROP;
}



/*********************************************************************************
* Routine Name :  _rtl8651_l4TcpFastAliasOut                                   
* Description : 
*      tcp packet output process
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4TcpFastAliasOut(struct rtl_pktHdr *phdr, struct ip *pip)
{
	struct tcphdr *tc;
	ipaddr_t ip_alias;
	ipaddr_t ip_s, ip_d;
	uint16 ip_sp, ip_dp, ip_aliasp;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb;	

	ip_s=phdr->ph_sip;
	ip_d=phdr->ph_dip;

#ifdef _TONY_DEBUG
{
	uint32 sip=GET_UINT32_UNALIGNED(&pip->ip_src);
	uint32 dip=GET_UINT32_UNALIGNED(&pip->ip_dst);	
	rtlglue_printf("l4TcpFastAliasOut src:%u.%u.%u.%u:%d dest:%u.%u.%u.%u:%d seq=%u\n",NIPQUAD(sip),ip_sp,NIPQUAD(dip),ip_dp,GET_UINT32_UNALIGNED(&tc->th_seq));
	if(tc->th_flags&TH_SYN) rtlglue_printf("SYN ");
	if(tc->th_flags&TH_ACK) rtlglue_printf("ACK ");
	if(tc->th_flags&TH_RST) rtlglue_printf("RST ");
	if(tc->th_flags&TH_FIN) rtlglue_printf("FIN ");
	if(tc->th_flags&TH_PUSH) rtlglue_printf("PUSH ");	
}
#endif

#ifdef _RTL_NAPT_FRAG
	if (IS_FRAGMENT_REMAIN(pip->ip_off))
	{
		/* deal with fragmented packet */
		uint32 alias_ip;
		if (_rtl8651_getFragCache(OUTBOUND_FRAG, ip_s, ip_d, ntohs(pip->ip_id), &alias_ip, &(phdr->ph_routeIdx), NULL, &(phdr->ph_dsid)) == SUCCESS)
		{
			PKTSET_UINT32_UNALIGNED(alias_ip, &pip->ip_src);
			if (!IPHDRLEN_IS_TWENTY(pip))
			{	/* Adjust L3 checksum by ourself */
				_RTL8651_L3_ADJUST_CHECKSUM(alias_ip, ip_s, pip->ip_sum);
			}
			if (_rtl8651_l3FastActRouting(FWDENG_FROMDRV, phdr, ip_d, pip, NULL) == SUCCESS)
				return SUCCESS;
		}
		return FAILED;
	}
#endif /* _RTL_NAPT_FRAG */

	tb = &entry;
	tc = (struct tcphdr *)L4HDR_ADDR(pip);
	ip_sp = ntohs(tc->th_sport);
	ip_dp =  ntohs(tc->th_dport);

	if (tc->th_flags & (TH_SYN|TH_FIN|TH_RST)) 
		return FAILED;

#ifdef _RTL_NAPT_FRAG
	if ((IS_FRAGMENT(pip->ip_off)) && (_rtl8651_fragQueueProcessOk == FALSE) && (_rtl8651_frag_NeedFragmentProcess(ntohs(pip->ip_id), ip_s, ip_d, IPPROTO_TCP) == TRUE))
		return FAILED;
#endif /* _RTL_NAPT_FRAG */
	
	if(_rtl8651_getOutboundNaptFlow(TRUE,ip_s,ip_sp,ip_d,ip_dp,  &tb)!=SUCCESS||tb->outListType==NAPT_ALG_OUT)
		return FAILED;

	if (tb->toProtocolStack == TRUE)
	{
		assert(0);	/* packet from LAN and match "toProtocolStack" is not expected now */
		return FAILED;	/* we can't process to Protocol Stack in fast path, for fragment queue process */
	}

	if(tb->alg_client_out  || tb->alg_server_out)  
		return FAILED; 

	if(tb->tcpFlag != (TCP_ESTABLISHED_FLOW-1))
	{
		return FAILED;
	}

#ifdef RTL865XB_URL_FILTER
		if((_urlFilterAclRuleNum>=0) &&
			((phdr->ph_reason&0xe)==4
			&& ((phdr->ph_reason&0xff0)>>5)==_urlFilterAclRuleNum))
#else
	if(!tb->urlFiltered||phdr->ph_reason==65)
#endif
	{
		return FAILED;
	}

	ip_alias = tb->insideGlobalIpAddr;
	ip_aliasp = tb->insideGlobalPort;		

	/* get dsid */
	phdr->ph_dsid = tb->dsid;

	if(_rtl8651_l4TcpMonitorOut(tc, tb)!=SUCCESS)	 /* Monitor TCP connection state */	
	{
		return FWDENG_DROP;
	}

	/* and source port are being altered */		
	tc->th_sport = htons(ip_aliasp);
	PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);
	if (tb->out_rt_t)
	{
		phdr->ph_routeIdx = L3_TABLE_INDEX((rtl8651_tblDrv_routeTable_t *)tb->out_rt_t) + 1;
	}else
	{
		rtl8651_tblDrv_routeTable_t *rt_entry;
		if (((rt_entry = _rtl8651_getSoftRoutingEntry(ip_alias, ip_d, IPPROTO_TCP, ip_dp)) == NULL) &&
			((rt_entry = _rtl8651_getRoutingEntry(ip_d)) == NULL))
			return FWDENG_DROP;
		tb->out_rt_t = rt_entry;
		phdr->ph_routeIdx = L3_TABLE_INDEX(rt_entry) + 1;
	}

	if (IS_FRAGMENT_FIRST(pip->ip_off)||(!IPHDRLEN_IS_TWENTY(pip)))
	{
		if (!IPHDRLEN_IS_TWENTY(pip))
		{
			_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, ip_s, pip->ip_sum);
		}
		_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, ip_s, ip_aliasp, ip_sp, tc->th_sum);
		if (IS_FRAGMENT_FIRST(pip->ip_off))
		{
			_rtl8651_updateFragCache(OUTBOUND_FRAG, ip_s, ip_d, ntohs(pip->ip_id), ip_alias, phdr->ph_routeIdx, NULL, phdr->ph_dsid);
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
			_rtl8651_fragmentShortestPathCacheSet_firstHalf(phdr, (uint8*)tc, pip, ip_s, ip_d, RTL8651_FRAGMENT_SHORTESTPATH_MODE_MODIFYSIP, ip_alias);
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
		}
	}

	/* get cache route */
	if( _rtl8651_l3FastActRouting(FWDENG_FROMDRV, phdr, ip_d, pip, &tb->out_macInfo_t) == SUCCESS)
	{
		return SUCCESS;
	}
	return FWDENG_DROP;
}


/*********************************************************************************
* Routine Name : _rtl8651_l4UdpFastAliasIn                                     
* Description :  udp packet input process                                               
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4UdpFastAliasIn(struct rtl_pktHdr *phdr, struct ip *pip)
{
	struct udphdr *ud;
	ipaddr_t ip_org;
	uint16 ip_sp, ip_dp, ip_orgp;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb;
	ipaddr_t ip_s, ip_d;

	ip_s=phdr->ph_sip;
	ip_d=phdr->ph_dip;
	
#ifdef _TONY_DEBUG
{
	uint32 sip=GET_UINT32_UNALIGNED(&pip->ip_src);
	uint32 dip=GET_UINT32_UNALIGNED(&pip->ip_dst);	
	rtlglue_printf("l4UdpFastAliasIn src:%u.%u.%u.%u:%d dest:%u.%u.%u.%u:%d\n",NIPQUAD(sip),ip_sp,NIPQUAD(dip),ip_dp);
}
#endif

#ifdef _RTL_NAPT_FRAG
	if (IS_FRAGMENT_REMAIN(pip->ip_off))
	{
		ipaddr_t newIp;
		rtl8651_tblDrv_naptTcpUdpFlowEntry_t *napt_t = NULL;
		if (_rtl8651_getFragCache(INBOUND_FRAG, ip_s, ip_d, ntohs(pip->ip_id), &newIp, &(phdr->ph_routeIdx), &napt_t, &(phdr->ph_dsid)) == SUCCESS)
		{
			PKTSET_UINT32_UNALIGNED(newIp, &pip->ip_dst);
			if (!IPHDRLEN_IS_TWENTY(pip))
			{
				_RTL8651_L3_ADJUST_CHECKSUM(newIp, ip_d, pip->ip_sum);
			}
			if (_rtl8651_l3FastActRouting(FWDENG_FROMDRV, phdr, newIp, pip, napt_t?(&napt_t->in_macInfo_t):NULL) == SUCCESS)
				return SUCCESS;
			else
				return FWDENG_DROP;
		}
		return FAILED;
	}
#endif /* _RTL_NAPT_FRAG */

	tb = &entry;
	ud = (struct udphdr *)L4HDR_ADDR(pip);
	ip_sp = ntohs(ud->uh_sport);
	ip_dp = ntohs(ud->uh_dport);

#ifdef _RTL_NAPT_FRAG
	if ((IS_FRAGMENT(pip->ip_off)) && (_rtl8651_fragQueueProcessOk == FALSE) && (_rtl8651_frag_NeedFragmentProcess(ntohs(pip->ip_id), ip_s, ip_d, IPPROTO_UDP) == TRUE))
		return FAILED;
#endif /* _RTL_NAPT_FRAG */
	
	if(_rtl8651_getInboundNaptFlow(FALSE,ip_d,ip_dp,ip_s,ip_sp,  &tb)!=SUCCESS || tb->inListType==NAPT_ALG_IN){
		return FAILED;
	}

	if (tb->toProtocolStack == TRUE)
	{	/* we can't process toProtocolStack here, for fragment queue */
		return FAILED;
	}

	if (tb->alg_client_in != NULL || tb->alg_server_in)
		return FAILED;

	/* update timer */
	_rtl8651_updateNaptConnection(tb, UDP_FLOW, rtl8651GlobalControl.udpTimeout);

	ip_org = tb->insideLocalIpAddr;
	ip_orgp = tb->insideLocalPort;

	/* get dsid */
	phdr->ph_dsid = tb->dsid;

	/* Put original port number back in */
	ud->uh_dport = htons(ip_orgp);
	PKTSET_UINT32_UNALIGNED(ip_org, &pip->ip_dst);
	if (tb->in_rt_t)
	{
		phdr->ph_routeIdx = L3_TABLE_INDEX((rtl8651_tblDrv_routeTable_t *)tb->in_rt_t) + 1;
	}else
	{
		rtl8651_tblDrv_routeTable_t *rt_entry;
		if (((rt_entry = _rtl8651_getSoftRoutingEntry(ip_s, ip_org, IPPROTO_UDP, ip_orgp)) == NULL) &&
			((rt_entry = _rtl8651_getRoutingEntry(ip_org)) == NULL))
			return FWDENG_DROP;
		tb->in_rt_t = rt_entry;
		phdr->ph_routeIdx = L3_TABLE_INDEX(rt_entry) + 1;
	}

	if ((IS_FRAGMENT(pip->ip_off)) || (!IPHDRLEN_IS_TWENTY(pip)))
	{
		if (!IPHDRLEN_IS_TWENTY(pip))
		{
			_RTL8651_L3_ADJUST_CHECKSUM(ip_org, ip_d, pip->ip_sum);
		}
		if (ud->uh_sum)
		{
			_RTL8651_L4_ADJUST_CHECKSUM(ip_org, ip_d, ip_orgp, ip_dp, ud->uh_sum);
		}
		if (IS_FRAGMENT(pip->ip_off))
		{
			_rtl8651_updateFragCache(INBOUND_FRAG, ip_s, ip_d, ntohs(pip->ip_id), ip_org, phdr->ph_routeIdx, tb, phdr->ph_dsid);
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
			_rtl8651_fragmentShortestPathCacheSet_firstHalf(phdr, (uint8*)ud, pip, ip_s, ip_d, RTL8651_FRAGMENT_SHORTESTPATH_MODE_MODIFYDIP, ip_org);
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
		}
	}

	if( _rtl8651_l3FastActRouting(FWDENG_FROMDRV, phdr, ip_org, pip, &tb->in_macInfo_t) == SUCCESS)
		return SUCCESS;
	return FWDENG_DROP;	
 }

/*********************************************************************************
* Routine Name :  _rtl8651_l4UdpFastAliasOut                                   
* Description :  
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4UdpFastAliasOut(struct rtl_pktHdr *phdr, struct ip *pip)
{
	struct udphdr *ud;
	ipaddr_t ip_alias;
	uint16 ip_sp, ip_dp, ip_aliasp;	
	struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s entry, *tb;	
	ipaddr_t ip_s, ip_d;

	ip_s=phdr->ph_sip;
	ip_d=phdr->ph_dip;


#ifdef _TONY_DEBUG
{
	uint32 sip=GET_UINT32_UNALIGNED(&pip->ip_src);
	uint32 dip=GET_UINT32_UNALIGNED(&pip->ip_dst);	
	rtlglue_printf("l4UdpFastAliasOut src:%u.%u.%u.%u:%d dest:%u.%u.%u.%u:%d\n",NIPQUAD(sip),ip_sp,NIPQUAD(dip),ip_dp);
}
#endif

#ifdef _RTL_NAPT_FRAG
	if (IS_FRAGMENT_REMAIN(pip->ip_off))
	{
		ipaddr_t newIp;
		rtl8651_tblDrv_naptTcpUdpFlowEntry_t *napt_t = NULL;

		if (_rtl8651_getFragCache(OUTBOUND_FRAG, ip_s, ip_d, ntohs(pip->ip_id), &newIp, &(phdr->ph_routeIdx), &napt_t, &(phdr->ph_dsid)) == SUCCESS)
		{
			PKTSET_UINT32_UNALIGNED(newIp, &pip->ip_src);
			if (!IPHDRLEN_IS_TWENTY(pip))
			{
				_RTL8651_L3_ADJUST_CHECKSUM(newIp, ip_s, pip->ip_sum);
			}
			if (_rtl8651_l3FastActRouting(FWDENG_FROMDRV, phdr, ip_d, pip, napt_t?(&napt_t->out_macInfo_t):NULL) == SUCCESS)
				return SUCCESS;
			else
				return FWDENG_DROP;
		}
		return FAILED;
	}
#endif /* _RTL_NAPT_FRAG */
		
	tb = &entry;

	ud = (struct udphdr *)L4HDR_ADDR(pip);
	ip_sp = ntohs(ud->uh_sport);
	ip_dp =  ntohs(ud->uh_dport);

#ifdef _RTL_NAPT_FRAG
	if ((ntohs(pip->ip_off) & IP_MF) && (_rtl8651_fragQueueProcessOk == FALSE) && (_rtl8651_frag_NeedFragmentProcess(ntohs(pip->ip_id), ip_s, ip_d, IPPROTO_UDP) == TRUE))
		return FAILED;
#endif /* _RTL_NAPT_FRAG */

	if(_rtl8651_getOutboundNaptFlow(FALSE,ip_s,ip_sp,ip_d,ip_dp,  &tb)!=SUCCESS||tb->outListType==NAPT_ALG_OUT)
		return FAILED;

	if (tb->toProtocolStack == TRUE)
	{	/* we don't process toProtocolStack here, for fragment queue */
		assert(0);	/* packet from LAN and match "toProtocolStack" is not expected now */
		return FAILED;
	}

	if (tb->alg_client_out || tb->alg_server_out || tb->alg_client_in || tb->alg_server_in )
		return FAILED;

	/* update timer */
	_rtl8651_updateNaptConnection(tb, UDP_FLOW, rtl8651GlobalControl.udpTimeout);

	ip_alias = tb->insideGlobalIpAddr;
	ip_aliasp = tb->insideGlobalPort;

	/* Put alias port in UDP header */
	ud->uh_sport = htons(ip_aliasp);

	/* Change source address */
	PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);

	/* get dsid */
	phdr->ph_dsid = tb->dsid;

	/* Routing cache */
	if (tb->out_rt_t)
	{
		phdr->ph_routeIdx = L3_TABLE_INDEX((rtl8651_tblDrv_routeTable_t *)tb->out_rt_t) + 1;
	}else
	{
		rtl8651_tblDrv_routeTable_t *rt_entry;
		if (((rt_entry = _rtl8651_getSoftRoutingEntry(ip_alias, ip_d, IPPROTO_UDP, ip_dp)) == NULL) &&
			((rt_entry = _rtl8651_getRoutingEntry(ip_d)) == NULL))
			return FWDENG_DROP;
		tb->out_rt_t = rt_entry;
		phdr->ph_routeIdx = L3_TABLE_INDEX(rt_entry) + 1;
	}

	if ((IS_FRAGMENT(pip->ip_off)) || (!IPHDRLEN_IS_TWENTY(pip)))
	{
		if (!IPHDRLEN_IS_TWENTY(pip))
		{
			_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, ip_s, pip->ip_sum);
		}
		if (ud->uh_sum)
		{
			_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, ip_s, ip_aliasp, ip_sp, ud->uh_sum);
		}
		if (IS_FRAGMENT(pip->ip_off))
		{
			_rtl8651_updateFragCache(OUTBOUND_FRAG, ip_s, ip_d, ntohs(pip->ip_id), ip_alias, phdr->ph_routeIdx, tb, phdr->ph_dsid);
#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
			_rtl8651_fragmentShortestPathCacheSet_firstHalf(phdr, (uint8*)ud, pip, ip_s, ip_d, RTL8651_FRAGMENT_SHORTESTPATH_MODE_MODIFYSIP, ip_alias);
#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */
		}
	}

	if( _rtl8651_l3FastActRouting(FWDENG_FROMDRV, phdr, ip_d, pip, &tb->out_macInfo_t) == SUCCESS)
		return SUCCESS;

	return FWDENG_DROP;
}
#endif /* _RTL_NAPT_SHORT_CUT*/



/*********************************************************************************
* Routine Name : _rtl8651_l4AddSeq
* Description :        
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
void _rtl8651_l4AddSeq(struct ip *pip, rtl8651_tblDrv_naptTcpUdpFlowEntry_t *tb, int32 delta)
{

	struct tcphdr *tc;
	struct l4ack_data_record x;
	int32     hlen, tlen, dlen;
	int32     i;
	uint32  seq_32_x;

	if ((pip->ip_p != IPPROTO_TCP) || IS_FRAGMENT_REMAIN(pip->ip_off))
	{
		return;
	}

	tc = (struct tcphdr *)L4HDR_ADDR(pip);
	seq_32_x = PKTGET_UINT32_UNALIGNED(&tc->th_seq);
	hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >>4)&0xf)) << 2;//Edward: Fix for alignment issue	
	tlen = (pip->ip_len);
	dlen = tlen - hlen;

	x.ack_old = ((seq_32_x) + dlen);
	x.ack_new = ((seq_32_x) + dlen + delta);
	
	x.delta = delta;
	x.active = 1;

	i = tb->tcpStateINdex;
	i %= L4_N_LINK_TCP_DATA;
	rtl8651_memcpy(&tb->ack[i],&x,sizeof(struct l4ack_data_record));

	i++;
	if (i == L4_N_LINK_TCP_DATA)
		tb->tcpStateINdex = 0;
	else
		tb->tcpStateINdex = i;
}




/*********************************************************************************
* Routine Name : _rtl8651_l4FindLinkAlg                                     
* Description :                                             
* Input :                          
* Output :                                                             
* Return :
* Note :   
*        ThreadSafe: n
**********************************************************************************/
rtl8651_tblDrv_algEntry_t *_rtl8651_l4FindLinkAlg(uint16 alg_port, uint16 isTcp)								
{
	rtl8651_tblDrv_algEntry_t *tb;
	int8 isServer, isClient;

	tb = _rtl8651_getAlgEntry(isTcp, &isServer, &isClient, alg_port, alg_port);
	
	if(tb){ 
		if( tb->fwEngAdd)
			return tb;
		else
			return NULL;
	}

	return NULL;
}

/*********************************************************************************
* Routine Name :    _rtl8651_l4napt_process                                 
* Description :                                                 
* Input :         
* Output :                                                             
* Return :     
* Note : 
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4napt_process(struct rtl_pktHdr * pkthdrPtr , struct ip *ip,uint8 proto, int32 dir)
{
	int32     status;

	if (dir == _RTL8651_L4_NAT_OUTPUT){
		
#ifdef _RTL_NAPT_FRAG
		if(IS_FRAGMENT_REMAIN(ip->ip_off))
		{
			/* deal with fragmented packet */
			uint32 alias_ip;

			ipaddr_t org_sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);
			ipaddr_t org_dip = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);

			if (_rtl8651_getFragCache(OUTBOUND_FRAG, org_sip, org_dip, ntohs(ip->ip_id), &alias_ip, &(pkthdrPtr->ph_routeIdx), NULL, &(pkthdrPtr->ph_dsid)) == SUCCESS)
			{	/* cache found */
				PKTSET_UINT32_UNALIGNED(alias_ip, &ip->ip_src);
				if (!IPHDRLEN_IS_TWENTY(ip))
				{	/* Adjust L3 checksum by ourself */
					_RTL8651_L3_ADJUST_CHECKSUM(alias_ip, pkthdrPtr->ph_sip, ip->ip_sum);
				}
				pkthdrPtr->ph_proto = PKTHDR_IP;
				return _RTL8651_NAT_PASS;
			}else if (_rtl8651_getFragCache(NAPT_REDIRECT_FRAG, org_sip, org_dip, ntohs(ip->ip_id), &alias_ip, NULL, NULL, &(pkthdrPtr->ph_dsid)) == SUCCESS)
			{
				PKTSET_UINT32_UNALIGNED(alias_ip, &ip->ip_dst);
				_RTL8651_L3_ADJUST_CHECKSUM(alias_ip, pkthdrPtr->ph_dip, ip->ip_sum);
				return FWDENG_TO_GW;	/* trap to protocol stack */
			}
			return FWDENG_IP_CANTFRAG;
		}
#else		
		if( IS_FRAGMENT(ip->ip_off))
		{
			return FWDENG_IP_CANTFRAG;	
		}		
#endif /* _RTL_NAPT_FRAG */

		switch (proto){
			case IPPROTO_TCP:
				status = _rtl8651_l4TcpNaptAliasOut(pkthdrPtr, ip);
				break;
			case IPPROTO_UDP:
				status = _rtl8651_l4UdpNaptAliasOut(pkthdrPtr, ip);
				break;
			case IPPROTO_ICMP:
				status = _rtl8651_l4IcmpNaptAliasOut(pkthdrPtr, ip);
				break;
			case IPPROTO_GRE:
				if(rtl8651_l4GreNaptAliasOut)
				{
					ipaddr_t ip_s;	/* for fragment packets */
					ip_s = pkthdrPtr->ph_sip;
					/* process GRE Napt Alg */
					status = (*rtl8651_l4GreNaptAliasOut)(pkthdrPtr, ip);
					if (status == SUCCESS)
					{	/* Fragment and ip header process */
						if (!IPHDRLEN_IS_TWENTY(ip))
						{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
							/* for fragment packet, it don't need to calculate L3 checksum */
							_RTL8651_L3_ADJUST_CHECKSUM(pkthdrPtr->ph_sip, ip_s, ip->ip_sum);	/* L3 Checksum */
						}
						if (IS_FRAGMENT_FIRST(ip->ip_off))
						{	/* first packet of fragment flow */
							_rtl8651_updateFragCache(OUTBOUND_FRAG, ip_s, pkthdrPtr->ph_dip, ntohs(ip->ip_id), pkthdrPtr->ph_sip, pkthdrPtr->ph_routeIdx, NULL, pkthdrPtr->ph_dsid);
						}
						pkthdrPtr->ph_sip = ip_s;	/* we don't modify source ip's original information */
					}
				}else
				{
					status = FWDENG_L4_NAPT_UNSUPPORTPROTO;
				}
				if (status != SUCCESS)
				{
					if (_rtl8651_outboundL4DMZForward(pkthdrPtr, ip, pkthdrPtr->ph_sip) == SUCCESS)
						status = SUCCESS;
				}
				break;
			case IPPROTO_ESP:			
				if(rtl8651_l4EspNaptAliasOut)
				{
					ipaddr_t ip_s;	/* for fragment packets */
					ip_s = pkthdrPtr->ph_sip;
					/* process ESP Napt Alg */
					status = (*rtl8651_l4EspNaptAliasOut)(pkthdrPtr, ip);
					if (status == SUCCESS)
					{/* Fragment and ip header process */
						if (!IPHDRLEN_IS_TWENTY(ip))
						{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
							/* for fragment packet, it don't need to calculate L3 checksum */
							_RTL8651_L3_ADJUST_CHECKSUM(pkthdrPtr->ph_sip, ip_s, ip->ip_sum);	/* L3 Checksum */
						}
						if (IS_FRAGMENT_FIRST(ip->ip_off))
						{	/* first packet of fragment flow */
							_rtl8651_updateFragCache(OUTBOUND_FRAG, ip_s, pkthdrPtr->ph_dip, ntohs(ip->ip_id), pkthdrPtr->ph_sip, pkthdrPtr->ph_routeIdx, NULL, pkthdrPtr->ph_dsid);
						}
						pkthdrPtr->ph_sip = ip_s;	/* we don't modify source ip's original information */
					}
				}else
				{
					status = FWDENG_L4_NAPT_UNSUPPORTPROTO;
				}
				if (status != SUCCESS)
				{
					if (_rtl8651_outboundL4DMZForward(pkthdrPtr, ip, pkthdrPtr->ph_sip) == SUCCESS)
						status = SUCCESS;
				}
				break;
			case IPPROTO_RSVP:			
//				printk("rsvp...1\n");
				if(rtl8651_l4RsvpNaptAliasOut)
				{
					ipaddr_t ip_s;	/* for fragment packets */
					ip_s = pkthdrPtr->ph_sip;
					/* process ESP Napt Alg */
					status = (*rtl8651_l4RsvpNaptAliasOut)(pkthdrPtr, ip);
					if (status == SUCCESS)
					{/* Fragment and ip header process */
						if (!IPHDRLEN_IS_TWENTY(ip))
						{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
							/* for fragment packet, it don't need to calculate L3 checksum */
							_RTL8651_L3_ADJUST_CHECKSUM(pkthdrPtr->ph_sip, ip_s, ip->ip_sum);	/* L3 Checksum */
						}
						if (IS_FRAGMENT_FIRST(ip->ip_off))
						{	/* first packet of fragment flow */
							_rtl8651_updateFragCache(OUTBOUND_FRAG, ip_s, pkthdrPtr->ph_dip, ntohs(ip->ip_id), pkthdrPtr->ph_sip, pkthdrPtr->ph_routeIdx, NULL, pkthdrPtr->ph_dsid);
						}
						pkthdrPtr->ph_sip = ip_s;	/* we don't modify source ip's original information */
					}
				}else
				{
					status = FWDENG_L4_NAPT_UNSUPPORTPROTO;
				}
				if (status != SUCCESS)
				{
					if (_rtl8651_outboundL4DMZForward(pkthdrPtr, ip, pkthdrPtr->ph_sip) == SUCCESS)
						status = SUCCESS;
				}
				break;				
			default:
				if (_rtl8651_outboundL4DMZForward(pkthdrPtr, ip, pkthdrPtr->ph_sip) == SUCCESS)
					status = SUCCESS;
				else
				status = FWDENG_L4_UNKNOWNPROTO;
		}
		
	}else
	{
#ifdef _RTL_NAPT_FRAG		
		if( IS_FRAGMENT_REMAIN(ip->ip_off) )
		{
			/* deal with fragmented packet */
			uint32 org_ip;

			ipaddr_t org_sip = PKTGET_UINT32_UNALIGNED(&ip->ip_src);
			ipaddr_t org_dip = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);

			if (_rtl8651_getFragCache(INBOUND_FRAG, org_sip, org_dip, ntohs(ip->ip_id), &org_ip, &(pkthdrPtr->ph_routeIdx), NULL, &(pkthdrPtr->ph_dsid)) == SUCCESS)
			{
				PKTSET_UINT32_UNALIGNED(org_ip, &ip->ip_dst);
				if (!IPHDRLEN_IS_TWENTY(ip))
				{	/* Adjust L3 checksum by ourself */
					_RTL8651_L3_ADJUST_CHECKSUM(org_ip, pkthdrPtr->ph_dip, ip->ip_sum);
				}
				return _RTL8651_NAT_PASS;
			}
			return FWDENG_IP_CANTFRAG;
		}
#else
		if( IS_FRAGMENT(ip->ip_off))
		{
			return FWDENG_IP_CANTFRAG;	
		}				
#endif /* _RTL_NAPT_FRAG */		


		switch (proto){
		case IPPROTO_TCP:
			status = _rtl8651_l4TcpNaptAliasIn(pkthdrPtr, ip);
			break;
		case IPPROTO_UDP:
			status = _rtl8651_l4UdpNaptAliasIn(pkthdrPtr, ip);
			break;
		case IPPROTO_ICMP:				
			status = _rtl8651_L4IcmpNaptAliasIn(pkthdrPtr, ip);								
			break;
		case IPPROTO_GRE:	
			if(rtl8651_l4GreNaptAliasIn)
			{
				ipaddr_t ip_d;	/* for fragment packets */
				ip_d = pkthdrPtr->ph_dip;
				/* process GRE Napt Alg */
				status = (*rtl8651_l4GreNaptAliasIn)(pkthdrPtr, ip);
				if (status == SUCCESS)
				{
				/* Fragment and ip header process */
					if (!IPHDRLEN_IS_TWENTY(ip))
					{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
						/* for fragment packet, it don't need to calculate L3 checksum */
						_RTL8651_L3_ADJUST_CHECKSUM(pkthdrPtr->ph_dip, ip_d, ip->ip_sum);	/* L3 Checksum */
					}
					if (IS_FRAGMENT_FIRST(ip->ip_off))
					{	/* first packet of fragment flow */
						_rtl8651_updateFragCache(INBOUND_FRAG, pkthdrPtr->ph_sip, ip_d, ntohs(ip->ip_id), pkthdrPtr->ph_dip, 0, NULL, pkthdrPtr->ph_dsid);
					}
				}
			}else
			{
				status = FWDENG_L4_NAPT_UNSUPPORTPROTO;
			}
			if (status != SUCCESS)
			{
				if (_rtl8651_inboundL4DMZForward(pkthdrPtr, ip, pkthdrPtr->ph_dip) == SUCCESS)
					status = SUCCESS;
			}
			break;
		case IPPROTO_ESP:						
			if(rtl8651_l4EspNaptAliasIn)
			{
				ipaddr_t ip_d;	/* for fragment packets */
				ip_d = pkthdrPtr->ph_dip;
				/* process ESP Napt Alg */
				status = (*rtl8651_l4EspNaptAliasIn)(pkthdrPtr, ip);
				/* Fragment and ip header process */
				if (status == SUCCESS)
				{
					if (!IPHDRLEN_IS_TWENTY(ip))
					{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
						/* for fragment packet, it don't need to calculate L3 checksum */
						_RTL8651_L3_ADJUST_CHECKSUM(pkthdrPtr->ph_dip, ip_d, ip->ip_sum);	/* L3 Checksum */
					}
					if (IS_FRAGMENT_FIRST(ip->ip_off))
					{	/* first packet of fragment flow */
						_rtl8651_updateFragCache(INBOUND_FRAG, pkthdrPtr->ph_sip, ip_d, ntohs(ip->ip_id), pkthdrPtr->ph_dip, 0, NULL, pkthdrPtr->ph_dsid);
					}
				}
			}else
			{
				status = FWDENG_L4_NAPT_UNSUPPORTPROTO;
			}	
			if (status != SUCCESS)
			{
				if (_rtl8651_inboundL4DMZForward(pkthdrPtr, ip, pkthdrPtr->ph_dip) == SUCCESS)
					status = SUCCESS;
			}
			break;
		case IPPROTO_RSVP:
//			printk("RSVP.....\n");
			if(rtl8651_l4RsvpNaptAliasIn)
			{
				ipaddr_t ip_d;	/* for fragment packets */
				ip_d = pkthdrPtr->ph_dip;
				/* process ESP Napt Alg */
				status = (*rtl8651_l4RsvpNaptAliasIn)(pkthdrPtr, ip);
				/* Fragment and ip header process */
				if (status == SUCCESS)
				{
					if (!IPHDRLEN_IS_TWENTY(ip))
					{	/* ASIC cannot offload this packet, we must calucluate L3/L4 cksum by ourself */
						/* for fragment packet, it don't need to calculate L3 checksum */
						_RTL8651_L3_ADJUST_CHECKSUM(pkthdrPtr->ph_dip, ip_d, ip->ip_sum);	/* L3 Checksum */
					}
					if (IS_FRAGMENT_FIRST(ip->ip_off))
					{	/* first packet of fragment flow */
						_rtl8651_updateFragCache(INBOUND_FRAG, pkthdrPtr->ph_sip, ip_d, ntohs(ip->ip_id), pkthdrPtr->ph_dip, 0, NULL, pkthdrPtr->ph_dsid);
					}
				}
			}else
			{
				status = FWDENG_L4_NAPT_UNSUPPORTPROTO;
			}	
			if (status != SUCCESS)
			{
				if (_rtl8651_inboundL4DMZForward(pkthdrPtr, ip, pkthdrPtr->ph_dip) == SUCCESS)
					status = SUCCESS;
			}
			break;
		default:
			if (_rtl8651_inboundL4DMZForward(pkthdrPtr, ip, pkthdrPtr->ph_dip) == SUCCESS)
				status = SUCCESS;
			else
				status = FWDENG_L4_UNKNOWNPROTO;
		}
	
	}

	switch (status){	
	case SUCCESS:		
#ifdef _RTL_NAPT_FRAG
		if(ntohs(ip->ip_off) & (IP_MF| IP_OFFMASK))
		{
			pkthdrPtr->ph_proto = PKTHDR_IP;
		}
#endif
		return _RTL8651_NAT_PASS;
	case FWDENG_USER_DROP:
		//rtlglue_printf("FWDENG_USER_DROP should OK %s %d\n",__FILE__,__LINE__);
		return _RTL8651_NAT_DROP;
	default:
		return status;		
	}		
}

/*********************************************************************************
* Routine Name : _rtl8651_l4nat_in_process                                    
* Description : 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4nat_in_process(struct rtl_pktHdr *phdr, struct ip *pip)
{

	ipaddr_t ip_d, ip_org = 0;
	rtl8651_tblDrv_natEntry_t *tb =NULL;
	int32 iresult = SUCCESS;
//	int32 accumulate;
			
	ip_d = phdr->ph_dip;

	tb = _rtl8651_getNatExternalMappingEntry(ip_d,phdr);

	if(tb == NULL || tb->localIp == 0){
		return _RTL8651_NAT_IGNORE;
	}
			
	ip_org = tb->localIp;

	if ((pip->ip_off) & IP_OFFMASK)
		goto change_ip;

	switch(pip->ip_p){
		case IPPROTO_TCP:
#ifdef RTL865XB_URL_FILTER
			if((_urlFilterAclRuleNum>=0) && 
				((phdr->ph_reason&0xe)==4 && ((phdr->ph_reason&0xff0)>>5)==_urlFilterAclRuleNum))
#else
			if (phdr->ph_reason == 65)
#endif
			{
				struct tcphdr *tc=(struct tcphdr *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
				int32 hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >>4)&0xf)) << 2;

				if ((ntohs(pip->ip_len))>hlen)
				{
					int32 position, len=64;
					int8 matched[64];
					position=_rtl8651_urlFilter(phdr->ph_dsid,phdr,pip,RTL8651_WAN2LAN,matched,&len);
					if (position>0)
					{
						/*
							cfliu: If you prepare a string buffer for the filter function, 
							rtl8651_urlFilter() would return string matched, its length and its position found in mbuf chain
							you may insert a log function pointer here to keep this extra info
							or just return a error value if you don't care.
						*/
						#ifdef RTL865X_DEBUG
						rtlglue_printf("string '%s'(%d bytes) in URL at %dth byte. Filtered...\n",matched, len, position );
						#endif
						return FWDENG_TCP_URLFILTERED;
					}
				}
			}
			iresult = _rtl8651_l4TcpNatAlgAliasIn(phdr, pip, tb);
			break;
		case IPPROTO_UDP:
			iresult = _rtl8651_l4UdpNatAlgAliasIn(phdr, pip, tb);
			break;	
		case IPPROTO_ICMP:
			{	/* for icmp error message, we must change original packet content (port is not changed in ALG, we don't need to find the entry) */
				struct icmp *ic = (struct icmp *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
				struct ip *ip = &ic->icmp_ip;

				switch (ic->icmp_type)
				{
					case ICMP_UNREACH:
					case ICMP_SOURCEQUENCH:
					case ICMP_TIMXCEED:
					case ICMP_PARAMPROB:
						PKTSET_UINT32_UNALIGNED(ip_org, &ip->ip_src);
							_RTL8651_L3_ADJUST_CHECKSUM(ip_org, tb->globalIp, ip->ip_sum);
				}
			}
			/* no break, goto change_ip; */
		default:
			/* all other protocols just do NAT directly */
			goto change_ip;		
	}

	if(iresult ==FWDENG_TCP_NAT_IN_ALGNOTFOUND||
		iresult==FWDENG_UDP_NAT_IN_ALGNOTFOUND)
	{
change_ip:
		if (((pip->ip_vhl & 0x0f) != 5) || (((ntohs(pip->ip_off)) & (IP_MF|IP_OFFMASK))  == IP_MF))
		{
			struct tcphdr *tc;
			struct udphdr *ud;

			if ((pip->ip_vhl & 0x0f) != 5)
			{
				_RTL8651_L3_ADJUST_CHECKSUM(ip_org, phdr->ph_dip, pip->ip_sum);
			}

			if ((ntohs(pip->ip_off)&IP_OFFMASK) == 0)
			{
				switch(pip->ip_p)
				{
					case IPPROTO_TCP:
						tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
						_RTL8651_L4_ADJUST_CHECKSUM(ip_org, phdr->ph_dip, 0, 0, tc->th_sum);
						break;
					case IPPROTO_UDP:
						ud = (struct udphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
						if (ud->uh_sum)
						{
							_RTL8651_L4_ADJUST_CHECKSUM(ip_org, phdr->ph_dip, 0, 0, ud->uh_sum);
						}
						break;
				}
			}
		}
		PKTSET_UINT32_UNALIGNED(ip_org, &pip->ip_dst);
	} else if(iresult != SUCCESS)
	{
		rtlglue_printf("%s: errno=%d\n",__FUNCTION__, iresult);
		return iresult;
	}
#if 0
	if(iresult != SUCCESS && 
		((ntohs(pip->ip_off)) & (IP_MF|IP_OFFMASK))  == IP_MF){
		
		struct tcphdr *tc;
		struct udphdr *ud;

		/* l3 checksum */
		accumulate = (phdr->ph_dip & 0xffff);
		accumulate +=  ((phdr->ph_dip >> 16) & 0xffff);
		accumulate -= (ip_org & 0xffff);
		accumulate -=  ((ip_org >> 16) & 0xffff);

		PKTSET_UINT32_UNALIGNED(ip_org, &pip->ip_dst);
		
		switch(pip->ip_p){
			case IPPROTO_TCP:
				tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
				_RTL8651_ADJUST_CHECKSUM(accumulate, tc->th_sum);	
				break;
			case IPPROTO_UDP:
				ud = (struct udphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
				if (ud->uh_sum)
					_RTL8651_ADJUST_CHECKSUM(accumulate, ud->uh_sum);				
				break;		
		}		
	}
	else if(iresult ==FWDENG_TCP_NAT_IN_ALGNOTFOUND||
		iresult==FWDENG_UDP_NAT_IN_ALGNOTFOUND )
	{		
change_ip:

		/* l3 checksum */
		accumulate = (ip_d & 0xffff);
		accumulate +=  ((ip_d >> 16) & 0xffff);
		accumulate -= (ip_org & 0xffff);
		accumulate -=  ((ip_org >> 16) & 0xffff);
		
		PKTSET_UINT32_UNALIGNED(ip_org, &pip->ip_dst);
	}else if(iresult!=SUCCESS){
		rtlglue_printf("%s: errno=%d\n",__FUNCTION__, iresult);
	}
#endif

	if(ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK))
	{
		phdr->ph_proto = PKTHDR_IP;
	}

	phdr->ph_routeIdx = 0;//flush route result. must do it again.
	if(_rtl8651_l3ActRouting((struct rtl_pktHdr *)phdr,ip_org,pip) != SUCCESS){
		return _RTL8651_NAT_DROP;
	}
	return (_RTL8651_NAT_PASS);

}


/*********************************************************************************
* Routine Name :  _rtl8651_l4nat_out_process                                   
* Description : 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static inline int32 _rtl8651_l4nat_out_process(struct rtl_pktHdr *phdr, struct ip *pip)
{
	ipaddr_t ip_s, ip_d, ip_alias = 0;
	rtl8651_tblDrv_natEntry_t *tb;
	int32 iresult = SUCCESS;
//	int32 accumulate;
		
	ip_s = phdr->ph_sip;
	ip_d = phdr->ph_dip;


	tb = _rtl8651_getNatInternalMappingEntry(ip_s,phdr);
	/* patch for multiplePppoe/unnumber */
	#if 0
	if (tb==NULL && (phdr->ph_unnumber & 0x1)){
		tb=(rtl8651_tblDrv_natEntry_t*)rtlglue_malloc(sizeof(rtl8651_tblDrv_natEntry_t));
		assert(tb);
		memset(tb,0,sizeof(rtl8651_tblDrv_natEntry_t));
		//tb->nextNatEntry=NULL;
		if ((phdr->ph_unnumber & 0x2)==0){
			/* lan->wan*/			
			tb->localIp=ip_s;
			tb->globalIp=ip_s;
		}else{
			/* wan->lan*/
			tb->localIp=ip_d;
			tb->globalIp=ip_d;
		}
	}
	#endif
	if(tb == NULL || tb->localIp == 0){
		return FWDENG_L4_NATNOENTRY;
	}

	ip_alias = tb->globalIp;

	if (ntohs(pip->ip_off) & IP_OFFMASK)
		goto change_ip;

	switch (pip->ip_p)
	{
		case IPPROTO_TCP:
#ifdef RTL865XB_URL_FILTER
			if((_urlFilterAclRuleNum>=0) && 
				((phdr->ph_reason&0xe)==4 && ((phdr->ph_reason&0xff0)>>5)==_urlFilterAclRuleNum))
#else
			if (phdr->ph_reason == 65)
#endif				
			{
				struct tcphdr *tc=(struct tcphdr *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
				int32 hlen = ((pip->ip_vhl & 0xf) + (( tc->th_off_x >>4)&0xf)) << 2;

				if ((ntohs(pip->ip_len))>hlen)
				{
					int32 position, len=64;
					int8 matched[64];
					position=_rtl8651_urlFilter(phdr->ph_dsid,phdr,pip,RTL8651_LAN2WAN,matched, &len);
					if (position>0)
					{
						/*
							cfliu: If you prepare a string buffer for the filter function, 
							rtl8651_urlFilter() would return string matched, its length and its position found in mbuf chain
							you may insert a log function pointer here to keep this extra info
							or just return a error value if you don't care.
						*/
						#ifdef RTL865X_DEBUG
						rtlglue_printf("string '%s'(%d bytes) in URL at %dth byte. Filtered...\n",matched, len, position );
						#endif
						return FWDENG_TCP_URLFILTERED;
					}
				}
			}
			iresult = _rtl8651_l4TcpNatAlgAliasOut(phdr, pip, tb);
			break;
		case IPPROTO_UDP:
			iresult = _rtl8651_l4UdpNatAlgAliasOut(phdr, pip, tb);
			break;	
		case IPPROTO_ICMP:
			{	/* for icmp error message, we must change original packet content (port is not changed in ALG, we don't need to find the entry) */
				struct icmp *ic = (struct icmp *)((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
				struct ip *ip = &ic->icmp_ip;

				switch (ic->icmp_type)
				{
					case ICMP_UNREACH:
					case ICMP_SOURCEQUENCH:
					case ICMP_TIMXCEED:
					case ICMP_PARAMPROB:
						PKTSET_UINT32_UNALIGNED(ip_alias, &ip->ip_dst);
						_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, tb->localIp, ip->ip_sum);
				}
			}
			/* no break, goto change_ip; */
		default:
			/* all other protocols just do NAT directly */
			goto change_ip;		
	}

	if( (iresult == FWDENG_TCP_NAT_OUT_ALGNOTFOUND) ||
		(iresult == FWDENG_UDP_NAT_OUT_ALGNOTFOUND ))
	{
change_ip:
		if (((ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK))  == IP_MF) ||((pip->ip_vhl & 0x0f) != 5))
		{	/* need recalculate checksum */
			struct tcphdr *tc;
			struct udphdr *ud;
			/* L3 checksum */
			if ((pip->ip_vhl & 0x0f) != 5)
			{
				_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, phdr->ph_sip, pip->ip_sum);
			}
			/* L4 checksum */
			if ((ntohs(pip->ip_off) & IP_OFFMASK) == 0)
			{
				switch(pip->ip_p)
				{
					case IPPROTO_TCP:
						tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
						_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, phdr->ph_sip, 0, 0, tc->th_sum);
						break;
					case IPPROTO_UDP:
						ud = (struct udphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
						if (ud->uh_sum)
						{
							_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, phdr->ph_sip, 0, 0, ud->uh_sum);
						}
						break;
				}
			}
		}
		PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);
		//phdr->ph_sip = ip_alias;
	} else if (iresult != SUCCESS)
	{
		rtlglue_printf("%s: errno=%d\n",__FUNCTION__, iresult);
		return iresult;
	}
#if 0
	if((iresult != SUCCESS) &&
		(((ntohs(pip->ip_off) &  (IP_MF|IP_OFFMASK))  == IP_MF) ||
		((pip->ip_vhl & 0x0f) != 5)))
	{
		/* not ALG, and need to recalculate checksum */	
		struct tcphdr *tc;
		struct udphdr *ud;
		/* L3 checksum */
		if ((pip->ip_vhl & 0x0f) != 5)
			_RTL8651_L3_ADJUST_CHECKSUM(ip_alias, phdr->ph_sip, pip->ip_src);
		/* L4 checksum */
		switch(pip->ip_p)
		{
			case IPPROTO_TCP:
				tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
				_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, phdr->ph_sip, 0, 0, tc->th_sum);
				break;
			case IPPROTO_UDP:
				ud = (struct udphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
				_RTL8651_L4_ADJUST_CHECKSUM(ip_alias, phdr->ph_sip, 0, 0, ud->uh_sum);
				break;
		}

		/* change ip */
		PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);
		#if 0
		/* l3 checksum */
		accumulate = (phdr->ph_sip & 0xffff);
		accumulate +=  ((phdr->ph_sip >> 16) & 0xffff);	
		accumulate -= (ip_alias & 0xffff);
		accumulate -=  ((ip_alias >> 16) & 0xffff);
		PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);		
		switch(pip->ip_p){
		case IPPROTO_TCP:
			tc = (struct tcphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
			_RTL8651_ADJUST_CHECKSUM(accumulate, tc->th_sum);	
			break;
		case IPPROTO_UDP:
			ud = (struct udphdr *) ((int8 *) pip + ((*(uint8*)pip&0xf) << 2));
			_RTL8651_ADJUST_CHECKSUM(accumulate, ud->uh_sum);				
			break;		
		}
		#endif
	}
	else if(iresult ==FWDENG_TCP_NAT_OUT_ALGNOTFOUND||
		iresult==FWDENG_UDP_NAT_OUT_ALGNOTFOUND )
	{
change_ip:
		
		/* l3 checksum */
		accumulate = (phdr->ph_sip & 0xffff);
		accumulate +=  ((phdr->ph_sip >> 16) & 0xffff);	
		accumulate -= (ip_alias & 0xffff);
		accumulate -=  ((ip_alias >> 16) & 0xffff);

		PKTSET_UINT32_UNALIGNED(ip_alias, &pip->ip_src);
	}else if(iresult!=SUCCESS)
		rtlglue_printf("%s: errno=%d\n",__FUNCTION__, iresult);	

	if(ntohs(pip->ip_off) &  (IP_MF|IP_OFFMASK) ){
		phdr->ph_flags &= (~CSUM_L4);
		phdr->ph_proto = PKTHDR_IP;
	}
#endif

	if(ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK))
	{
		phdr->ph_flags &= (~CSUM_L4);
		phdr->ph_proto = PKTHDR_IP;
	}

	if(_rtl8651_l3ActRouting((struct rtl_pktHdr *)phdr,ip_d,pip) != SUCCESS)
	{
		return _RTL8651_NAT_DROP;
	}
	return (SUCCESS);
}

#define RTL8651_L34ADDITIONAL_EXECUTION_FUNCTIONS

/* set all initial values to additional execution function */
static void _rtl8651_l34AdditionalExec_init(void)
{
	memset(_RTL8651_ADDITIONAL_PROC, 0, sizeof(_RTL8651_ADDITIONAL_PROC));

	/* Initially we turn ON the SMAC Filtering for NAPT / NAT / Routing (LP - RP) -- for backward compatibility */
	_RTL8651_ADDITIONAL_PROC[_RTL8651_DEC_NPT] = _RTL8651_L34ADDTIONALEXEC_SMACFILTER;
	_RTL8651_ADDITIONAL_PROC[_RTL8651_DEC_NT] = _RTL8651_L34ADDTIONALEXEC_SMACFILTER;
	_RTL8651_ADDITIONAL_PROC[_RTL8651_DEC_RT_LR] = _RTL8651_L34ADDTIONALEXEC_SMACFILTER;

	/* Initially we don't block any decision */
}

/* run additional execution */
static inline int32 _rtl8651_l34AddtionalExec(int32 decision, ipaddr_t dip, void *pkt)
{
	int32 retval = SUCCESS;

	if (_RTL8651_DECISION_ISVALID(decision))
	{
		/* ========== check each property ========== */
		/* block this decision ? ==> return BLOCK */
		if (_RTL8651_ADDITIONAL_PROC[decision] & _RTL8651_L34ADDTIONALEXEC_BLOCK)
		{
			retval = FWDENG_DECISION_BLOCK;
			goto exception_check;
		}

		/* do smac filtering ? ==> check SMAC's permission */
		if (_RTL8651_ADDITIONAL_PROC[decision] & _RTL8651_L34ADDTIONALEXEC_SMACFILTER)
		{
			uint8 *ether = (uint8 *)pkt;
			assert( pkt );

			if (_rtl8651_isMacAccessControlPermit((ether_addr_t *)&ether[6]) == FALSE)
			{
				retval = FWDENG_MACBLOCK;
				goto exception_check;
			}

		}
	}else assert(0);

	/* continue processing */
	return SUCCESS;

exception_check:

	/* dst ip is gateway ip, we trap it to protocol stack instead of block it. */
	{
		int8 is_gwip;
		_rtl8651_getIpInfo(dip, NULL, NULL, &is_gwip);
		if (is_gwip == TRUE)
		{
			return FWDENG_TO_GW;
		}
	}

	return retval;
}

/* set parameters of L34 Additional Exec */
int32 _rtl8651_l34AdditionalExec_setParam(uint32 srcType, uint32 dstType, _rtl8651_l34AdditionalExec_t *para)
{
	int32 decision;

	if ((!_RTL8651_HOSTTYPE_ISVALID(srcType)) ||(!_RTL8651_HOSTTYPE_ISVALID(dstType)))
	{
		return FAILED;
	}

	decision = _RTL8651_PROC[srcType][dstType];

	assert( para && _RTL8651_DECISION_ISVALID(decision) );

	/* Because SMAC filtering always applied for NAT / NAPT / Routing (LP - RP), we check it here */
	if ((decision == _RTL8651_DEC_NPT) || (decision == _RTL8651_DEC_NT) || (decision == _RTL8651_DEC_RT_LR))
	{
		if ((para->property & _RTL8651_L34ADDTIONALEXEC_SMACFILTER) == 0)
		{
			#ifdef RTL865X_DEBUG
			rtlglue_printf("@ Error Configuration - User can not turn OFF SMAC filtering for outbound NAPT/NAT packet.\n");
			#endif
			return FAILED;
		}
	}

	/* set parameters into database */
	_RTL8651_ADDITIONAL_PROC[decision] = para->property;

	return SUCCESS;
}
/* get parameters of L34 Additional Exec */
int32 _rtl8651_l34AdditionalExec_getParam(uint32 srcType, uint32 dstType, _rtl8651_l34AdditionalExec_t *para)
{
	int32 decision;

	if ((!_RTL8651_HOSTTYPE_ISVALID(srcType)) ||(!_RTL8651_HOSTTYPE_ISVALID(dstType)))
	{
		return FAILED;
	}

	decision = _RTL8651_PROC[srcType][dstType];

	assert( para && _RTL8651_DECISION_ISVALID(decision) );

	/* record parameters from database */
	memset(para, 0, sizeof(_rtl8651_l34AdditionalExec_t));
	para->property = _RTL8651_ADDITIONAL_PROC[decision];

	return SUCCESS;
}


#define RTL8651_CHECKSUM_FUNCTIONS

/*********************************************************************************
* Routine Name :  _rtl8651_ipChecksum                                   
* Description : 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
uint16 
_rtl8651_ipChecksum(struct ip * pip)
{


	uint32     sum=0,oddbyte=0;
	uint16 *ptr = (uint16 *)pip;
	uint32 nbytes = ((pip->ip_vhl & 0xf) << 2);	

	while (nbytes > 1)
	{
		
		sum += (*ptr++);
		nbytes -= 2;
	}
	if (nbytes == 1)
	{
		//oddbyte = 0;
		//((uint8 *) & oddbyte)[0] = *(uint8 *) ptr;
		//((uint8 *) & oddbyte)[1] = 0;
		oddbyte = (*ptr & 0xff00);
		sum += oddbyte;
	}
	
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);
	
	return (~sum);


}


/*********************************************************************************
* Routine Name :  _rtl8651_tcpChecksum                                   
* Description : 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/

uint16 _rtl8651_tcpChecksum(struct ip *pip)
{
	int32 sum, nbytes, nhdr, i;
	uint16 *sip, *dip, *ptr;

	nhdr =  ((pip->ip_vhl & 0xf) << 2);
	nbytes = ntohs(pip->ip_len) - nhdr;
	ptr = (uint16 *) ((char *) pip + nhdr);
	sum = 0;

	/* Note: We always padding zero to the tail of the packet!! */
	*(((uint8 *)pip) + nhdr + nbytes) = (uint8)0;
	for (i=0; i<nbytes ;i=i+2){
		sum += (unsigned long)*ptr++;
	}

	/* "Pseudo-header" data */
	dip=(uint16 *)&pip->ip_dst;
	sum += *dip; dip++;
	sum += *dip;
	sip=(uint16 *)&pip->ip_src;
	sum += *sip; sip++;
	sum += *sip;	

	sum +=  nbytes;
	sum += ((uint16) pip->ip_p);

	/* Roll over carry bits */
	while (sum>>16)
		sum = (sum & 0xFFFF)+(sum >> 16);
	
	/* Take the one's complement of sum */
	sum = ~sum;
	return (uint16)sum;
	
}

#if 0
uint16 _rtl8651_tcpChecksum(struct ip *pip)
{
	uint32 l4_len = ntohs(pip->ip_len)-20;
	uint8 *buff = ((uint8 *)pip)+20;
	uint16 word16, *sip, *dip;
	uint32 sum=0, i;
	int8 padding=0;

	if (l4_len & 0x01) {
		padding = 1;
		buff[l4_len] = 0; //padding
	}
	for (i=0; i<l4_len+padding ;i=i+2){
		word16 =((buff[i]<<8)&0xFF00)+(buff[i+1]&0xFF);
		sum = sum + (unsigned long)word16;
	}

	/* "Pseudo-header" data */
	dip=(uint16 *)&pip->ip_dst;
	sum += *dip; dip++;
	sum += *dip;
	sip=(uint16 *)&pip->ip_src;
	sum += *sip; sip++;
	sum += *sip;	

	sum +=  l4_len;
	sum += ((uint16) pip->ip_p);

	/* Roll over carry bits */
	while (sum>>16)
		sum = (sum & 0xFFFF)+(sum >> 16);
	
	/* Take the one's complement of sum */
	sum = ~sum;
	return (uint16)sum;
}


uint16 
_rtl8651_tcpChecksum(struct ip * pip)
{
	uint16   *ptr;
	struct tcphdr *tc;
	int32     nhdr, ntcp, nbytes;
	int32     sum, oddbyte;

	nhdr =  ((pip->ip_vhl & 0xf) << 2);
	ntcp = ntohs(pip->ip_len) - nhdr;

	tc = (struct tcphdr *) ((char *) pip + nhdr);
	ptr = (uint16 *) tc;

/* Add up TCP header and data */
	nbytes = ntcp;
	sum = 0;
	while (nbytes > 1)
	{
		sum += *ptr++;
		nbytes -= 2;
	}
	if (nbytes == 1)
	{
		oddbyte = 0;
		((uint8 *) & oddbyte)[0] = *(uint8 *) ptr;
		((uint8 *) & oddbyte)[1] = 0;
		sum += oddbyte;
	}
	
/* "Pseudo-header" data */
	ptr = (uint16 *) & (pip->ip_dst);
	sum += *ptr++;
	sum += *ptr;
	ptr = (uint16 *) & (pip->ip_src);
	sum += *ptr++;
	sum += *ptr;

	sum += ((uint16) ntcp);
	sum += ((uint16) pip->ip_p);

/* Roll over carry bits */
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);

/* Return checksum */
	return ((uint16) ~ sum);
}
#endif


/*********************************************************************************
* Routine Name :  _rtl8651_icmpChecksum                                   
* Description : 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
/* shoud continue memory */
uint16 
_rtl8651_icmpChecksum(struct ip * pip)
{
	uint16   *ptr;
	int32  nhdr, nbytes;
	int32  sum, i;


	nhdr =  ((pip->ip_vhl &0xf) << 2);
	nbytes = ntohs(pip->ip_len) - nhdr;
	ptr = (uint16 *) ((char *) pip + nhdr);
	sum = 0;

	/* Note: we always padding here!! */
	*(((uint8 *)pip) + nhdr + nbytes) = (uint8)0;
	for (i=0; i<nbytes ; i=i+2){
		sum +=  (unsigned long) *ptr++;
	}

	/* Roll over carry bits */
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);

	/* Return checksum */
	return ((uint16) ~ sum);
}


#ifdef _RTL_UPNP
#define RTL8651_UPNP_FUNCTIONS

static void _rtl8651_initUpnpDb(void){
	upnp_tb = (struct _RTL8651_UPNP_ENTRY *)rtlglue_malloc(UPNP_ITEM * sizeof (struct _RTL8651_UPNP_ENTRY));
	memset(upnp_tb,0,UPNP_ITEM * sizeof (struct _RTL8651_UPNP_ENTRY));
}

static void _rtl8651_ReinitUpnpDb(void){
	memset(upnp_tb,0,UPNP_ITEM * sizeof (struct _RTL8651_UPNP_ENTRY));
}

/*
@func void		| rtl8651_addUpnpMap	| Add one record into UPNP Map Table.
@parm uint32 | inFalgs | UPNP_TCP: if the connection is UDP, don't set this flag, UPNP_ONESHOT: only for one shot connection,  UPNP_PERSIST: exclude with UPNP_ONESHOT, for persistent connection.
@parm ipaddr_t | remote_ip |The Remote IP address.
@parm uint16 | remote_port |The Remote IP port.
@parm ipaddr_t | alias_ip | The Gateway's WAN IP address.
@parm uint16 | alias_port | The Gateway's NAPT port of this connection.
@parm ipaddr_t | local_ip | The internal's local IP address.
@parm uint16 | local_port | The internal's local port.
@comm 
When add an Upnp Map record, someone from remote_ip:remote_port connect to the alias_ip:alias_port will redirect the packet to local_ip:local_port.
If the remote_ip or the remote_port are zero, it's mean the field will not to be checked.
@xref  <p rtl8651_delUpnpMap>
*/
void rtl8651_addUpnpMap(uint32 inFlags, 
	ipaddr_t remote_ip, uint16 remote_port, 
	ipaddr_t alias_ip, uint16 alias_port,
	ipaddr_t local_ip, uint16 local_port){
	int32 ii;
	struct _RTL8651_UPNP_ENTRY *tb;

	rtlglue_drvMutexLock();

	tb = upnp_tb;

	for (ii = 0 ; ii < UPNP_ITEM ; ii++, tb++){
		if( (tb->flags & UPNP_VALID) == 0 || ((tb->alias_port == alias_port)&&((tb->flags&UPNP_TCP) == (inFlags&UPNP_TCP)))){
			
			tb->remote_ip = remote_ip;
			tb->remote_port = remote_port; 
			tb->alias_ip = alias_ip;
			tb->alias_port = alias_port ;
			tb->local_ip = local_ip;
			tb->local_port = local_port;
			tb->algIdx = 0;
			
			tb->flags = UPNP_VALID;
			tb->flags |= inFlags;
			
			tb->age = UPNP_AGE;
			
			break;
		}
	}

	rtlglue_drvMutexUnlock();

}

void _rtl8651_addAlgQosUpnpMap(uint32 inFlags, 
	ipaddr_t remote_ip, uint16 remote_port, 
	ipaddr_t alias_ip, uint16 alias_port,
	ipaddr_t local_ip, uint16 local_port,uint32 algIdx){
	
	int32 ii;
	struct _RTL8651_UPNP_ENTRY *tb;

	rtlglue_drvMutexLock();

	tb = upnp_tb;

	for (ii = 0 ; ii < UPNP_ITEM ; ii++, tb++){
		if( (tb->flags & UPNP_VALID) == 0 || ((tb->alias_port == alias_port)&&((tb->flags&UPNP_TCP) == (inFlags&UPNP_TCP)))){
			
			tb->remote_ip = remote_ip;
			tb->remote_port = remote_port; 
			tb->alias_ip = alias_ip;
			tb->alias_port = alias_port ;
			tb->local_ip = local_ip;
			tb->local_port = local_port;
			tb->algIdx = algIdx;			
			tb->flags = UPNP_VALID;
			tb->flags |= inFlags;			
			tb->age = UPNP_AGE;
			
			break;
		}
	}

	rtlglue_drvMutexUnlock();
}

void rtl8651_addUpnpMapLeaseTime(uint32 inFlags, 
	ipaddr_t remote_ip, uint16 remote_port, 
	ipaddr_t alias_ip, uint16 alias_port,
	ipaddr_t local_ip, uint16 local_port,uint32 age){
	int32 ii;
	struct _RTL8651_UPNP_ENTRY *tb;

	rtlglue_drvMutexLock();

	tb = upnp_tb;

	for (ii = 0 ; ii < UPNP_ITEM ; ii++, tb++){
		if( (tb->flags & UPNP_VALID) == 0 || ((tb->alias_port == alias_port)&&((tb->flags&UPNP_TCP) == (inFlags&UPNP_TCP)))){
			
			tb->remote_ip = remote_ip;
			tb->remote_port = remote_port; 
			tb->alias_ip = alias_ip;
			tb->alias_port = alias_port ;
			tb->local_ip = local_ip;
			tb->local_port = local_port;
			tb->algIdx = 0;

			tb->flags = UPNP_VALID;
			tb->flags |= inFlags;

			if(age==0)
				tb->age = UPNP_AGE;
			else
				tb->age = age;
			
			break;
		}
	}

	rtlglue_drvMutexUnlock();

}

/*
@func void		| rtl8651_delUpnpMap	| Remove one record from UPNP Map Table.
@parm uint32 | uint32 | 1:TCP, 0:UDP
@parm ipaddr_t | remote_ip |The Remote IP address.
@parm uint16 | remote_port |The Remote IP port.
@parm ipaddr_t | alias_ip | The Gateway's WAN IP address.
@parm uint16 | alias_port | The Gateway's NAPT port of this connection.
@parm ipaddr_t | local_ip | The internal's local IP address.
@parm uint16 | local_port | The internal's local port.
@comm 
Remove one record from UPNP Map Table, please refer to rtl8651_addUpnpMap() to see more detail.
@xref  <p rtl8651_addUpnpMap>
*/

void rtl8651_delUpnpMap(uint32 isTcp, ipaddr_t remote_ip, uint16 remote_port, 
	ipaddr_t alias_ip, uint16 alias_port,
	ipaddr_t local_ip, uint16 local_port){
	int32 ii;
	struct _RTL8651_UPNP_ENTRY *tb;

	rtlglue_drvMutexLock();

	tb = upnp_tb;

	for (ii = 0 ; ii < UPNP_ITEM ; ii++, tb++){
		if( tb->flags & UPNP_VALID ){
			
			if(	(tb->remote_ip & remote_ip) == remote_ip && 
				(tb->remote_port &  remote_port)== remote_port &&
				tb->alias_ip == alias_ip &&
				tb->alias_port == alias_port &&
				((tb->local_ip == local_ip && tb->local_port == local_port) ||
				(tb->local_ip == local_ip && tb->local_port == 0) ||
				(tb->local_ip == 0 && tb->local_port == local_port) ||
				(tb->local_ip == 0 && tb->local_port == 0))
				){		
				
				if(isTcp == 0){
					if(tb->flags & UPNP_TCP) 
						continue;
				}
				
				tb->flags = 0;				
			}
		}
	}

	rtlglue_drvMutexUnlock();

}


static inline int32 _rtl8651_findInboundUpnpMapping (
	uint32 isTcp,
	ipaddr_t remote_ip, uint16 remote_port,
	ipaddr_t alias_ip, uint16 alias_port,
	ipaddr_t *local_ip, uint16 *local_port,
	uint32 *flags, uint32 *algIdx)
{
	int32 ii;
	struct _RTL8651_UPNP_ENTRY *tb = upnp_tb;	
	for (ii = 0 ; ii < UPNP_ITEM ; ii++, tb++){		
		if( tb->flags & UPNP_VALID ){
			if(	(tb->remote_ip & remote_ip) == tb->remote_ip && 
				(tb->remote_port &  remote_port)== tb->remote_port &&
				tb->alias_ip == alias_ip &&
				tb->alias_port == alias_port ){		

				if ((isTcp && (!(tb->flags&UPNP_TCP))) ||
						((!isTcp) && (tb->flags&UPNP_TCP)))
					continue;

				if(tb->flags & UPNP_ONESHOT)
				{
					tb->flags = 0;					/* Reset uPnP flag to invalidate this entry */
				}

				//tony: patch for CDRouter check problem.
				if(remote_port==20)
					return FAILED;

				assert(tb->local_ip);
				
				if (local_ip) *local_ip = tb->local_ip;
				if (local_port) *local_port = tb->local_port;
				if (flags) *flags = tb->flags;
				if (algIdx) *algIdx = tb->algIdx;
				
				return SUCCESS;
			}
		}
	}

	return FAILED;
}

/*
	Get External IP/Port for Upnp mapping

	return value:

		FAILED		- no uPnP entry found for this connection
		SUCCESS	- system found a uPnP entry for this connection
*/
static inline int32 _rtl8651_findOutboundUpnpMapping (
	uint32 isTcp,
	ipaddr_t IntIp, uint16 IntPort,
	ipaddr_t *ExtIp, uint16 *ExtPort,
	ipaddr_t RemoteIp, uint16 RemotePort,
	uint32 *flags, uint32 *algIdx)
{
	struct _RTL8651_UPNP_ENTRY *upnp_t = upnp_tb;
	int32 idx;

	/*
		trace uPnP table to find the entry which conforms to this flow.
	*/
	for (idx = 0 ; idx < UPNP_ITEM ; idx++, upnp_t++)
	{
		/* check for valid */
		if ((upnp_t->flags & UPNP_VALID) == FALSE)
			continue;

		/* check for TCP/UDP bit */
		if (	(isTcp && ((upnp_t->flags & UPNP_TCP) == FALSE)) ||
			((isTcp == FALSE) && ((upnp_t->flags & UPNP_TCP) == UPNP_TCP))
		   )
			continue;

		/* check for internal/remote IP/Port (we check port first because if 'port' has larger probability to find the difference out) */
		if (	(upnp_t->remote_port && (upnp_t->remote_port != RemotePort)) ||
			(upnp_t->remote_ip && (upnp_t->remote_ip != RemoteIp)) ||
			(upnp_t->local_port != IntPort) ||
			(upnp_t->local_ip != IntIp)
		   )
			continue;

		/* remove this entry if it's 'one-shot' entry */
		if (upnp_t->flags & UPNP_ONESHOT)
		{
			upnp_t->flags = 0;				/* Reset uPnP flag to invalidate this entry */
		}

		/*
			chenyl: copy this form _rtl8651_findInboundUpnpMapping()

				[ tony: patch for CDRouter check problem. ]
		*/
		if(RemotePort == 20)
		{
			return FAILED;
		}

		assert(upnp_t->local_ip);

		/*
			Entry exists: set it to parameters.
		*/
		if (ExtIp) 	*ExtIp = upnp_t->alias_ip;
		if (ExtPort)	*ExtPort = upnp_t->alias_port;
		if (flags)		*flags = upnp_t->flags;
		if (algIdx)	*algIdx = upnp_t->algIdx;

		return SUCCESS;		
	}

	return FAILED;
}

uint32 rtl8651_queryUpnpMapTimeAge(uint32 isTcp, ipaddr_t remote_ip, uint16 remote_port, 
	ipaddr_t alias_ip, uint16 alias_port){

	int32 ii;
	struct _RTL8651_UPNP_ENTRY *tb;	
	uint32 retval = 0;

	rtlglue_drvMutexLock();

	tb = upnp_tb;

	for (ii = 0 ; ii < UPNP_ITEM ; ii++, tb++){		
		if( tb->flags & UPNP_VALID ){
			if(	(tb->remote_ip & remote_ip) == tb->remote_ip && 
				(tb->remote_port &  remote_port)== tb->remote_port &&
				tb->alias_ip == alias_ip &&
				tb->alias_port == alias_port ){		

				if ((isTcp && (!(tb->flags&UPNP_TCP))) ||
						((!isTcp) && (tb->flags&UPNP_TCP)))
					continue;

				retval = tb->age;
				goto out;
			}
		}
	}

out:
	rtlglue_drvMutexUnlock();

	return retval;
}

/*
	Check if the specific isTcp/IntIp/IntPort/ExtIp/ExtPort/RemoteIp/RemotePort can be used in uPnP system or not
*/
static inline int32 _rtl8651_checkUsableUpnpEntry(uint32 isTcp, ipaddr_t IntIp, uint32 IntPort, ipaddr_t ExtIp, uint32 ExtPort, ipaddr_t RemoteIp, uint32 RemotePort)
{
	int32 i;
	struct _RTL8651_UPNP_ENTRY *tb = upnp_tb, *candidate = NULL;

	for (i = 0 ; i < UPNP_ITEM ; i ++ , tb ++)
	{
		if ((tb->flags&UPNP_VALID) &&												/* valid entry */
			(tb->alias_ip == ExtIp) &&												/* ip match */
			(tb->alias_port == ExtPort) &&											/* port match */
			(tb->remote_ip == 0 || tb->remote_ip == RemoteIp) &&					/* remote ip match */
			(tb->remote_port == 0 || tb->remote_port == RemotePort) &&				/* remote port match */
			((isTcp&&(tb->flags&UPNP_TCP)) ||((!isTcp)&&(!(tb->flags&UPNP_TCP))))	/* protocol match */
			)
		{
			if ((IntIp == tb->local_ip) && (IntPort == tb->local_port))
			{
				return TRUE;
			}
			/* the ExtIp/ExtPort is used by other internal Ip/Port */
			candidate = tb;
		}
	}

	/* <--- After trace upnp table, we don't find any entry use IntIp/IntPort ---> */

	/* the ExtIp/ExtPort is used by other IntIp/IntPort, so we can not use it */
	if (candidate)
		return FALSE;

	/* the ExtIp/ExtPort is not used by any entry */
	return TRUE;
}

static int32 _rtl8651_uPnPOneSecondTimer(){

	int32 ii;
	struct _RTL8651_UPNP_ENTRY *tb = upnp_tb;
	for (ii = 0 ; ii < UPNP_ITEM ; ii++, tb++){

		if( (tb->flags & (UPNP_PERSIST|UPNP_VALID)) ==  
			(UPNP_PERSIST|UPNP_VALID)){
			continue;
		}
		
		if( tb->flags & UPNP_VALID ){
			tb->age--;			
			if(tb->age == 0){
				tb->flags &= (~UPNP_VALID);
			}				
		}
	}

	return SUCCESS;
}
#endif /* _RTL_UPNP */

#define RTL8651_UDP_BLOCK_FUNCTIONS

/*
@func void		| rtl8651a_setUdpSizeThreasValue	| Set the UDP packet size thresh value
@parm uint32 | dsid | the dial session ID.
@parm uint32 | value | UDP Size
@comm 
Rome Driver will block udp packet when upd length is bigger than _rtl8651_udpSizeThreshValue
@xref  <p rtl8651a_setUdpSizeThreasValue>
*/

void rtl8651a_setUdpSizeThreshValue(uint32 SessionID, uint32 value)
{	
	uint32 dsid;
	uint32 idx;

	rtlglue_drvMutexLock();

	dsid=_rtl8651_SessionIDToDSID(SessionID);

	_rtl8651_udpSizeThreshValue[dsid]=(uint16)value;

	_rtl8651_udpBlockingEnable = FALSE;

	for (idx = 0 ; idx < rtl8651_tblDrvPara.dialSessionNumber ; idx ++)
	{
		if (_rtl8651_udpSizeThreshValue[idx] != 0xffff)
			_rtl8651_udpBlockingEnable = TRUE;
	}

	rtlglue_drvMutexUnlock();
}

#define RTL8651_WAN_STATUS_FUNCTIONS

/*
@func void		| rtl8651a_setWanStatus	| Set the connection status of WAN port.
@parm uint32 | Status | 0:Disconnect, 1:Connected
@parm uint32 | Status | 0:Disconnect, 1:Connected
@comm 
Rome Driver can't get the Connection Status from PPPoE or DHCPC daemon.
We must set the WAN Status when getting the WAN IP address, and connected.
@xref  <p rtl8651_setDosStatus>
*/
void rtl8651a_setWanStatus(uint32 SessionID, uint32 Status)
{	
	uint32 dsid;
	rtlglue_drvMutexLock();
	dsid=_rtl8651_SessionIDToDSID(SessionID);
	rtl8651_wanStatus[dsid]=Status;
	rtlglue_drvMutexUnlock();
}

/*
@func void		| rtl8651_setWanStatus	| Set the connection status of WAN port.
@parm uint32 | Status | 0:Disconnect, 1:Connected
@comm 
Rome Driver can't get the Connection Status from PPPoE or DHCPC daemon.
We must set the WAN Status when getting the WAN IP address, and connected.
@xref  <p rtl8651_setDosStatus>
*/
void rtl8651_setWanStatus(uint32 Status)
{	
	rtlglue_drvMutexLock();
	rtl8651_wanStatus[0]=Status;
	rtlglue_drvMutexUnlock();
}


#ifdef _RTL_PS_ACTION
#define RTL8651_PROTOCOL_STACK_ACTION_FUNCTIONS

static int32 _rtl8651_protoStackActionInit(rtl8651_fwdEngineInitPara_t *para){

	rtl8651_ProtoStackActions = (rtl8651_PS_Action_Entry_t *)rtlglue_malloc(para->protoStackMaxActions * sizeof (rtl8651_PS_Action_Entry_t));
	memset(rtl8651_ProtoStackActions,0,para->protoStackMaxActions * sizeof (rtl8651_PS_Action_Entry_t));	
	return SUCCESS;
}

/*
@func void		| rtl8651_flushProtoStackActions	| Flush Remote Management table
@comm 
This function will flush all the records in Remote Management table, before you add 
records, you must call this function to make remote management table free.
@xref  <p rtl8651_addProtoStackActions>
*/
void rtl8651_flushProtoStackActions(void)
{
	rtlglue_drvMutexLock();
	memset(rtl8651_ProtoStackActions,0,sizeof(rtl8651_PS_Action_Entry_t)*rtl8651_fwdEnginePara.protoStackMaxActions);
	rtlglue_drvMutexUnlock();
}

int32 _rtl8651_addProtocolStackActions(rtl8651_PS_Action_Entry_t *PSA)
{
	int i;
	rtl8651_PS_Action_Entry_t *psa_t = rtl8651_ProtoStackActions;

	/* dstIp must be external gateway ip. */
	if (PSA->dstIp != 0 && _rtl8651_extIpAddr(PSA->dstIp, NULL) == FALSE)
		return FWDENG_EPSA_DSTIPNOTEXTGWIP;;
	
	for(i = 0; i < rtl8651_fwdEnginePara.protoStackMaxActions; i++, psa_t++)
	{
		/* Check if there's any PS action table entry which is confilict with this "Protocol Stack Action" to be added. */
		if (psa_t->action != PS_ACTION_NONE) {
			if ((PSA->dstIp == 0 || psa_t->dstIp == 0 || PSA->dstIp == psa_t->dstIp) 
				&& PSA->role == psa_t->role && PSA->protocol == psa_t->protocol) {
				/*
					IP range overlapping, need to further check if port range is overlapping.
				*/
				if (!((PSA->upperIp < psa_t->lowerIp) || (PSA->lowerIp > psa_t->upperIp))) {
					/* 
						Port range overlapping, forbid to add this "Protocol Stack Action". 
						Because "port range" and "ip range" should not overlapping at the same time.
					*/
					if (!((PSA->orgPortEnd < psa_t->orgPort) ||(PSA->orgPort > psa_t->orgPortEnd)))
						return FWDENG_EPSA_IPPORTRANGE_CONFLICT;
					else {
						uint16 user_portRange = PSA->orgPortEnd - PSA->orgPort;
						uint16 tb_portRange = psa_t->orgPortEnd - psa_t->orgPort;

						/* Although port range is not overlapping, but fwd port range might be overlapping. check it! */
						if (!(((PSA->fwdPort + user_portRange) < psa_t->fwdPort) || (PSA->fwdPort > (psa_t->fwdPort + tb_portRange))))
							return FWDENG_EPSA_IPPORTRANGE_CONFLICT;
					}
				}
			}
			continue;
		}
		
		/* 2005.09.07  Currently, PS action table entries are continuous. */
		else /* No existing PS action table entries confict with this "Protocol Stack Action" to be added. */
		{
			psa_t->dstIp = PSA->dstIp;
			psa_t->action = PSA->action;
			psa_t->lowerIp = PSA->lowerIp;
			psa_t->upperIp = PSA->upperIp;
			psa_t->role = PSA->role;
			psa_t->protocol = PSA->protocol;
			psa_t->orgPort = PSA->orgPort;
			psa_t->orgPortEnd = PSA->orgPortEnd;
			psa_t->fwdPort = PSA->fwdPort;			
			return SUCCESS;
		}
	}
	return FWDENG_EPSA_TABLEFULL;
	
}

#define LEGALPORT_UB	65535
int32 _rtl8651_checkPSAction(rtl8651_PS_Action_Entry_t *PSAction)
{
	if (PSAction->lowerIp > PSAction->upperIp)
		return FWDENG_EPSA_IPRANGENOTCORRECT;
	if (PSAction->protocol == IPPROTO_ICMP && ((PSAction->role != 0) || (PSAction->orgPort != 0) || (PSAction->orgPortEnd != 0) || (PSAction->fwdPort != 0)))
		return FWDENG_EPSA_ICMPARGISNOTZERO;
	if (PSAction->orgPort > PSAction->orgPortEnd)
		return FWDENG_EPSA_PORTRANGENOTCORRECT;
	if ((PSAction->orgPortEnd - PSAction->orgPort) > (LEGALPORT_UB - PSAction->fwdPort))
		return FWDENG_EPSA_PORTRANGETOOLARGE;

	return SUCCESS;
}


/*
@func int32		| rtl8651_addGenericProtoStackActions	| Add Generic Protocol Stack Action table
@parm rtl8651_PS_Action_Entry_t * | PSAction | Passing the data structure of rtl8651_PS_Action_Entry_t to add an entry to PS action table
@rvalue SUCCESS | An PS action table entry was added successfully
@rvalue FWDENG_EPSA_IPPORTRANGE_CONFLICT | PS action table entry addition failed due to there's conflict with existing entries
@rvalue FWDENG_EPSA_TABLEFULL | When adding PSA, PS action table is full
@rvalue FWDENG_EPSA_IPRANGENOTCORRECT | When adding PSA, lower IP is larger than upper IP in IP range arguments
@rvalue FWDENG_EPSA_ICMPARGISNOTZERO | When adding PSA, ICMP protocol with incorrect arguments
@rvalue FWDENG_EPSA_PORTRANGENOTCORRECT | When adding PSA, lower port number is larger than upper port number
@rvalue FWDENG_EPSA_DSTIPNOTGWIP | When adding PSA, assigned dstIp is not gateway ip.
@rvalue FWDENG_EPSA_PORTRANGETOOLARGE	|Specified port range is larger than (legal port upper bound - fwdPort)
@comm 
The meaning of field in rtl8651_PS_Action_Entry is the same as parameter of rtl8651_addProtoStackActions,
except an additional field dstIp. This field represents the destination ip of packets destined to gateway.
The purpose of this additional field is to support multiple session remote management issue.
@xref  <p rtl8651_addProtoStackActions>
*/
int32 rtl8651_addGenericProtoStackActions(rtl8651_PS_Action_Entry_t *PSAction)
{
	int32 retval;

	/* input validation */
	if ((retval = _rtl8651_checkPSAction(PSAction)) != SUCCESS)
		return retval;
	
	rtlglue_drvMutexLock();
	retval = _rtl8651_addProtocolStackActions(PSAction);
	rtlglue_drvMutexUnlock();

	return retval;
}

/*
@func void		| rtl8651_addProtoStackActions	| Add Remote Management table
@parm ipaddr_t | lowerIp | Those two parameters will check the source IP of incoming packets, only matching rule will apply this rule, ip is 0 , mask is 0 will match all IP addresses.
@parm ipaddr_t | upperIp | please refer to paramter lowerIp.
@parm uint8 | action | Set PS_ACTION_TO_PS will trap to protocol stack, set PS_ACTION_DROP will drop match packets.
@parm uint8 | role | If your protocol stack is a server(ex: Web Server), please set PS_ACTION_SERVER, if your protocol stack is a client(ex: DNS relay), please set PS_ACTION_CLIENT. This parameters only valid in TCP/UDP protocol type.
@parm uint8 | protocol | protocol type, this type is the same as the protocol field of IP header. ICMP is 1, IGMP is 2, TCP is 6, UDP is 17...
@parm uint16 | orgPort | the original port number of incoming packet, if role is PS_ACTION_SERVER, this field is mean destination port. If role is PS_ACTION_CLIENT, this field is mean source port. This parameters only valid in TCP/UDP protocol type.
@parm uint16 | fwdPort | the original port will change to this port and forward this packets to protocol stack.
@comm 
This API allow users pass-through assigned packets to handle it by Proto Stack or  Drop.
If your rule not TCP/UDP protocol, please keep parameters role, orgPort, fwdPort to zero.
IF you don't want to forward port, please keep original port equal to forward port.
Example:
If you want to set port 8080 for your web server which is in gateway(port 80) for remote 
management,and only Some IP can access it(ex: 1.2.3.1~1.2.3.254). you can call APIs as following:
rtl8651_flushProtoStackActions();
rtl8651_addProtoStackActions(0x01020300,0x010203fe,PS_ACTION_TO_PS,PS_ACTION_SERVER,IPPROTO_TCP,8080,80);
You can also enable or disable ICMP echo reply by following API call:
to disable ICMP echo reply
rtl8651_addProtoStackActions(0,0xffffffff,PS_ACTION_DROP,0,IPPROTO_ICMP,0,0); 
to enable ICMP echo reply 
rtl8651_addProtoStackActions(0,0xffffffff,PS_ACTION_TO_PS,0,IPPROTO_ICMP,0,0);
@xref  <p rtl8651_flushProtoStackActions>
*/
int32 rtl8651_addProtoStackActions(ipaddr_t lowerIp, ipaddr_t upperIp, uint8 action, uint8 role, uint8 protocol, uint16 orgPort, uint16 fwdPort)
{
	int32 retval;
	rtl8651_PS_Action_Entry_t PSAction;

	memset(&PSAction, 0, sizeof(rtl8651_PS_Action_Entry_t));
	/* == data structure assignment == */
	PSAction.lowerIp = lowerIp;
	PSAction.upperIp = upperIp;
	PSAction.action = action;
	PSAction.role = role;
	PSAction.protocol = protocol;
	PSAction.orgPort = orgPort;
	PSAction.orgPortEnd = orgPort;
	PSAction.fwdPort = fwdPort;
	PSAction.dstIp = 0;	/* wild card */
	
	/* == input validation == */
	if ((retval = _rtl8651_checkPSAction(&PSAction)) != SUCCESS)
		return retval;

	rtlglue_drvMutexLock();
	retval = _rtl8651_addProtocolStackActions(&PSAction);
	rtlglue_drvMutexUnlock();

	return retval;
}

int32 rtl8651_addProtoStackServerPortRange(ipaddr_t lowerIp, ipaddr_t upperIp, uint8 protocol, uint16 PortStart, uint16 PortEnd, uint16 fwdPortStart)
{
	int32 retval;
	rtl8651_PS_Action_Entry_t PSAction;

	memset(&PSAction, 0, sizeof(rtl8651_PS_Action_Entry_t));
	/* == data structure assignment == */
	PSAction.lowerIp = lowerIp;
	PSAction.upperIp = upperIp;
	PSAction.action = PS_ACTION_TO_PS;
	PSAction.role = PS_ACTION_SERVER;
	PSAction.protocol = protocol;
	PSAction.orgPort = PortStart;
	PSAction.orgPortEnd = PortEnd;
	PSAction.fwdPort = fwdPortStart;
	PSAction.dstIp = 0;	/* wild card */

	/* == input validation == */
	if ((retval = _rtl8651_checkPSAction(&PSAction)) != SUCCESS)
		return retval;

	rtlglue_drvMutexLock();
	retval = _rtl8651_addProtocolStackActions(&PSAction);
	rtlglue_drvMutexUnlock();

	return retval;
}


static int32 _rtl8651_doProtoStackAction(struct ip *iph)
{
	int i;
	ipaddr_t sip = PKTGET_UINT32_UNALIGNED(&iph->ip_src);
	ipaddr_t dip = PKTGET_UINT32_UNALIGNED(&iph->ip_dst);
	uint16 *sport = NULL, *dport = NULL, *chksum = NULL;
	rtl8651_PS_Action_Entry_t *psa = rtl8651_ProtoStackActions;

	/* If no PS action table entry matched, bypass this packet */
	int32 retval = PS_ACTION_NONE;

#if 0
	/* redundant judgement because _rtl8651_doProtoStackAction will be executed in NAPT-R or Port Bouncing case. */
#endif
	/* if dip is not external gateway ip, forbid doing proto stack actions. */
	if (_rtl8651_extIpAddr(dip, NULL) == FALSE)
		return retval;

	/* special service for fragment packets */
	if (IS_FRAGMENT_REMAIN(iph->ip_off))
	{		
		ipaddr_t returnValue;

		if (_rtl8651_getFragCache(PROTOCOL_STACK_FRAG, sip, dip, ntohs(iph->ip_id), &returnValue, NULL, NULL, NULL) == SUCCESS)
			return (int32)returnValue;
		return retval;
	}

	/* retrieve L4 port number of this packet if can */
	switch (iph->ip_p)
	{
		case IPPROTO_TCP:
		{
			struct tcphdr *tcp;
			tcp = (struct tcphdr *) ((int8 *) iph + ((iph->ip_vhl &0xf) << 2));	
			sport = &tcp->th_sport;
			dport = &tcp->th_dport;
			chksum = &tcp->th_sum;
			break;
		}
		
		case IPPROTO_UDP:
		{
			struct udphdr *udp;
			udp = (struct udphdr *) ((int8 *) iph + ((iph->ip_vhl &0xf) << 2));	
			sport = &udp->uh_sport;
			dport = &udp->uh_dport;
			chksum = &udp->uh_sum;
			break;
		}
	}

	/* lookup the PS action table */
	for (i = 0; i < rtl8651_fwdEnginePara.protoStackMaxActions; i++, psa++)
	{
		if (psa->action == PS_ACTION_NONE)		/* invalid entry */
			break;
			
		if ((psa->dstIp == 0 || dip == psa->dstIp) && iph->ip_p == psa->protocol)	/* (1) dstIp & protocol match */
		{
			/* TCP/UDP: check dport if PS role is server; check sport if PS role is client */
			if (psa->protocol == IPPROTO_TCP || psa->protocol == IPPROTO_UDP)
			{
				uint16 *chkport = NULL;

				if (psa->role == PS_ACTION_SERVER)
					chkport = dport;
				else if (psa->role == PS_ACTION_CLIENT)
					chkport = sport;
				
				/* 
					2005.08.12
					
					Originally, protocol stack action entry is matched with "ip subnet". Now, we change it to "ip range".
				*/
				if ((psa->orgPort <= ntohs(*chkport)) && (psa->orgPortEnd >= ntohs(*chkport)))	/* (2) port range match */
				{
					if ((sip >= psa->lowerIp) && (sip <= psa->upperIp))	/* (3) ip range match => all key fields match now, must perform the action and stop  lookup */
					{
						if (psa->action == PS_ACTION_TO_PS)	/* PS rule */
						{
							if (psa->orgPort != psa->fwdPort)	/* Need port change */
							{
								uint16 newFwdPort = psa->fwdPort + (ntohs(*chkport) - psa->orgPort);
								
								if ((iph->ip_p != IPPROTO_UDP) || (*chksum != 0))	/* adjust l4 checksum */
								{
									_RTL8651_L4_ADJUST_CHECKSUM(0, 0, newFwdPort, ntohs(*chkport), *chksum);
								}
								*chkport = htons(newFwdPort);
							}
						}

						retval = psa->action;
						goto out;	/* must stop PS action table lookup due to rule match */
					}
					else		/* (3) ip range mis-match => set action to DROP, but still lookup the following rules */
						retval = PS_ACTION_DROP;
				}
			}
			else if(psa->protocol == IPPROTO_ICMP)
			{
				struct icmp *im;
				im = (struct icmp *) ((int8 *) iph + ((iph->ip_vhl &0xf) << 2));

				if (im->icmp_type == ICMP_ECHO)	/* Currently, we only handle ICMP ECHO REQUEST */
				{
					/* 
						2005.08.12
						
						Originally, protocol stack action entry is matched with "ip subnet". Now, we change it to "ip range".
					*/
					if ((sip >= psa->lowerIp) && (sip <= psa->upperIp))	/* (2) ip range match => all key fields match now, perform the corresponding action */
					{

						/*
						 * priority -
						 *			1. Drop
						 *			2. To DMZHost
						 *			3. To Protocol stack
						 */

						retval = psa->action;
						
						if ((psa->action != PS_ACTION_DROP) && rtl8651_DMZHostIcmpPassthroughEnable)
							retval = PS_ACTION_NONE;
						
						goto out;	/* must stop PS action table lookup due to rule match */
					}
					else		/* (2) ip range mis-match => set action to DROP, but still lookup the following rules */
						retval = PS_ACTION_DROP;
				}
				else		/* Not ICMP ECHO REQUEST message */
				{
					retval = PS_ACTION_NONE;
					goto out;
				}
			}
			else
			{
				/* 
					2005.08.12
					
					Originally, protocol stack action entry is matched with "ip subnet". Now, we change it to "ip range".
				*/				
				if ((sip >= psa->lowerIp) && (sip <= psa->upperIp))	/* (2) ip range match => all key fields match now, perform the corresponding action */
				{
					retval = psa->action;
					goto out;	/* must stop PS action table lookup due to rule match */
				}
				else		/* (2) ip range mis-match => set action to DROP, but still lookup the following rules */
					retval = PS_ACTION_DROP;
			}
		}
	}
out:
	if (IS_FRAGMENT_FIRST(iph->ip_off))
	{
		_rtl8651_updateFragCache(PROTOCOL_STACK_FRAG, sip, dip, ntohs(iph->ip_id), retval, 0, NULL, 0);
	}

	if (retval == PS_ACTION_DONT_TIMEOUT)
		retval = PS_ACTION_NONE;

	return retval;
}
/*
	Use port/l4proto as parameter, to find if there is any entry to fit this port/l4proto mapping or not.
	If matched, the first matched entry will be returned.
*/
static inline rtl8651_PS_Action_Entry_t *_rtl8651_checkProtoStackRegistedPort(uint16 port, uint8 l4proto)
{
	int i;
	rtl8651_PS_Action_Entry_t *psa;

	psa = rtl8651_ProtoStackActions;
	
	for ( i = 0 ; i < rtl8651_fwdEnginePara.protoStackMaxActions ; i++, psa++)
	{
		if (psa->action == PS_ACTION_NONE)
			continue;

		if (l4proto == psa->protocol)
		{
			if (((l4proto == IPPROTO_TCP) ||(l4proto == IPPROTO_UDP)) && ((port < psa->orgPort )||(port>psa->orgPortEnd)))
				continue;
			return psa;
		}
	}
	return NULL;
}

static int32 _rtl8651_aliasProtoStackSentPkt(uint32 outbound, uint32 dsid, struct ip *ip)
{
	if (rtl8651_getAsicOperationLayer() == 4) {
		if(outbound){
			/*send to WAN */
			int i=0,dont_timeout=FALSE;	
			uint16 *sport=NULL,*dport=NULL,*chkport=NULL,*chksum=NULL;
			rtl8651_PS_Action_Entry_t *psa = rtl8651_ProtoStackActions;
			struct ip *iph=ip;	
			ipaddr_t sip = PKTGET_UINT32_UNALIGNED(&iph->ip_src.s_addr);
			ipaddr_t dip = PKTGET_UINT32_UNALIGNED(&iph->ip_dst.s_addr);



			if (IS_FRAGMENT_REMAIN(iph->ip_off))
				return SUCCESS;

			switch(iph->ip_p)
			{
				case IPPROTO_TCP:
				{
					struct tcphdr *tcp;
					tcp = (struct tcphdr *) ((int8 *) iph + ((iph->ip_vhl &0xf) << 2));	
					sport=&tcp->th_sport;
					dport=&tcp->th_dport;
					chksum=&tcp->th_sum;

						if ( tcp->th_flags & TH_RST )
						{
#if	_RTL_DROP_ERROR_MESSAGE	
						    if ( rtl8651_drvProtocolStackErrMsgDrop == TRUE )
						    {
							return PS_SEND_ERROR_PKT;
						    }
#endif
						    if ( rtl8651_drvProtocolStackErrMsgDontTimeout == TRUE )
						    {
							/* Don't refresh idle-timer for RST packet */
							dont_timeout = TRUE;
						    }
					}

					break;
				}
				
				case IPPROTO_UDP:
				{
					struct udphdr *udp;
					udp = (struct udphdr *) ((int8 *) iph + ((iph->ip_vhl &0xf) << 2));	
					sport=&udp->uh_sport;
					dport=&udp->uh_dport;
					chksum=&udp->uh_sum;
					break;
				}
				case IPPROTO_ICMP:
				{
					/* If this packet is icmp error message, we need to perform port translation check. */
					struct icmp *ic = (struct icmp *)((int8 *)iph + ((*(uint8 *)iph & 0xf) << 2));

					if (ic->icmp_type == ICMP_UNREACH || ic->icmp_type == ICMP_SOURCEQUENCH || ic->icmp_type == ICMP_TIMXCEED || ic->icmp_type == ICMP_PARAMPROB) 
					{
#if	_RTL_DROP_ERROR_MESSAGE	
						if ( rtl8651_drvProtocolStackErrMsgDrop == TRUE )
						{
							return PS_SEND_ERROR_PKT;
						}
#endif

						if ( rtl8651_drvProtocolStackErrMsgDontTimeout == TRUE )
						{
							/* Don't refresh idle-timer for RST packet */
							dont_timeout = TRUE;
						}
					}
					break;
				}
			}		
			
			for(; i < rtl8651_fwdEnginePara.protoStackMaxActions; i++, psa++)
			{
				if(psa->action == PS_ACTION_NONE)
					break;
				
				if((psa->dstIp == 0 || sip == psa->dstIp) && iph->ip_p == psa->protocol)		/* (1) dstIp & protocol match */
				{			
					if(psa->protocol == IPPROTO_TCP || psa->protocol == IPPROTO_UDP)
					{
						if(psa->role == PS_ACTION_SERVER) 
							chkport = sport;
						else if(psa->role == PS_ACTION_CLIENT) 
							chkport = dport;

						if (ntohs(*chkport) >= psa->fwdPort && ntohs(*chkport) <= (psa->fwdPort + psa->orgPortEnd - psa->orgPort))	/* (2) port match */
						{
							if (dip >= psa->lowerIp && dip <= psa->upperIp)	/* (3) ip range match */
							{
								/* The condition that needs to change port */
								if((psa->action == PS_ACTION_TO_PS) && (psa->orgPort != psa->fwdPort))
								{
									uint16 newFwdPort = ntohs(*chkport) - psa->fwdPort + psa->orgPort;
									/* l4 checksum */
									if ((iph->ip_p != IPPROTO_UDP) || (*chksum != 0))
										_RTL8651_L4_ADJUST_CHECKSUM(0, 0, newFwdPort, ntohs(*chkport), *chksum);
									*chkport = htons(newFwdPort);
								}
																
								if(psa->action == PS_ACTION_DONT_TIMEOUT)
									dont_timeout = TRUE;
								
								goto _check_idleTimer;	/* this packet has met this PS action entry, no more PS action table lookup! */
							}
						}
					}			
				}			
			}
_check_idleTimer:
			if ((dont_timeout == FALSE) && (rtl8651_drvProtocolStackDontTimeout == FALSE))
			{
				_rtl8651_recordTraffic(dsid, TRUE /* to WAN */);
			}
		}
		else {	/* This code segment is only to handle the case that Protocol Stack sends remote management pkt to LAN. */
				/* Other packets should not be transparent to this code segment. */
			if (rtl8651_drvEnPortBouncingPSaction) {	/* make sure this is turned on */
				struct ip *iph = ip;
				int i = 0 ;
				uint16 *sport = NULL, *dport = NULL, *chkport = NULL, *chksum = NULL;
				ipaddr_t sip = PKTGET_UINT32_UNALIGNED(&iph->ip_src.s_addr);
				ipaddr_t dip = PKTGET_UINT32_UNALIGNED(&iph->ip_dst.s_addr);
				rtl8651_PS_Action_Entry_t *psa = rtl8651_ProtoStackActions;

				if (IS_FRAGMENT_REMAIN(iph->ip_off))
				return SUCCESS;

				switch(iph->ip_p) {
					case IPPROTO_TCP:
					{
						struct tcphdr *tcp;
						tcp = (struct tcphdr *) ((int8 *) iph + ((iph->ip_vhl &0xf) << 2));
						sport = &tcp->th_sport;
						dport = &tcp->th_dport;
						chksum = &tcp->th_sum;
						break;
					}

					case IPPROTO_UDP:
					{
						struct udphdr *udp;
						udp = (struct udphdr *) ((int8 *) iph + ((iph->ip_vhl &0xf) << 2));	
						sport=&udp->uh_sport;
						dport=&udp->uh_dport;
						chksum=&udp->uh_sum;
						break;
					}

					default:
					return SUCCESS;
				}

			for(; i < rtl8651_fwdEnginePara.protoStackMaxActions; i++, psa++) {
					if (psa->action == PS_ACTION_NONE)	/* Judge if this PSA entry is valid or not. */
						break;
#if 0
					if (psa->ip == 0 && psa->mask == 0)	// We want to match the PSA whose ip & mask field are defined specifically.
						continue;

				      printk("i: %d, psa->ip: %u.%u.%u.%u, psa->mask: %u.%u.%u.%u, srcip: %u.%u.%u.%u, psa->orgPort: %d, psa->fwdPort: %d\n",
					  	i, NIPQUAD(psa->ip), NIPQUAD(psa->mask), NIPQUAD(ntohl(iph->ip_src.s_addr)), psa->orgPort, psa->fwdPort);
				      printk("psa->protocol: %d, psa->action: %d, psa->role: %d, sport: %d\n", psa->protocol, psa->action, psa->role, *sport);
#endif

					if ((psa->dstIp == 0  || sip == psa->dstIp) && iph->ip_p == psa->protocol)	/* (1) dstIp & protocol match */
					{			
						if (psa->protocol == IPPROTO_TCP || psa->protocol == IPPROTO_UDP)
						{
							if (psa->role == PS_ACTION_SERVER)
								chkport = sport;
							else if (psa->role == PS_ACTION_CLIENT) 
								chkport = dport;

							if (ntohs(*chkport) >= psa->fwdPort && ntohs(*chkport) <= (psa->fwdPort + psa->orgPortEnd - psa->orgPort)) 	/* (2) port match */
							{
								/* (3) ip range match */
								if (dip >= psa->lowerIp && dip <= psa->upperIp)
								{
									/* The condition that needs to change port */
									if((psa->action == PS_ACTION_TO_PS) && (psa->orgPort != psa->fwdPort) && _rtl8651_extIpAddr(sip, NULL) == TRUE /* tricky part, because unnumbered gateway ip is external ip also internal ip. Use exteranl ip to make decision. */)
									{
										uint16 newFwdPort = ntohs(*chkport) - psa->fwdPort + psa->orgPort;
										/* l4 checksum */
										if ((iph->ip_p != IPPROTO_UDP) || (*chksum != 0))
											_RTL8651_L4_ADJUST_CHECKSUM(0, 0, newFwdPort, ntohs(*chkport), *chksum);
										*chkport = htons(newFwdPort);
									}
								return SUCCESS;	/* this packet has met this PS action entry, no more PS action table lookup */
								}
							}
						}
					}			
#if 0
					/* We want to find out the PSA which traps LAN pkt to web server on Protocol Stack. */
					// If the operation layer is 2 or 3, we don't have to change src port 80 to port 8080.
					if (_rtl8651_intIpAddr((PKTGET_UINT32_UNALIGNED(&iph->ip_src.s_addr))) == FALSE && (psa->orgPort != psa->fwdPort) &&
					      psa->protocol == IPPROTO_TCP && psa->action == PS_ACTION_TO_PS && psa->role == PS_ACTION_SERVER) {
							chkport = sport;	// Check the src port of this packet.
							if(psa->fwdPort == ntohs(*chkport)) {
								if ((iph->ip_p != IPPROTO_UDP) || (*chksum != 0))
									_RTL8651_L4_ADJUST_CHECKSUM(0, 0, psa->orgPort, ntohs(*chkport), *chksum);
								*chkport = htons(psa->orgPort);	// Change the src port of this packet.
							}
					}
#endif
				}
			}
		}
	}
    return SUCCESS;
}
/*
	chenyl (2004/12/6)
	When Forwarding Engine Receive the ICMP Error Message which is sent by protocol stack,
	the error packet content will be the original packet content sent by gateway.

	So, if this packet is modified by forwarding engine's protocol stack translation engine,
	we MUST translate it back.
*/
static int32 _rtl8651_aliasProtoStackRecvPkt(uint32 fromOutbound, struct ip *ip)
{
	uint16 *sportp = NULL, *dportp = NULL, *ckportp = NULL, *cksump = NULL;
	rtl8651_PS_Action_Entry_t * psa;
	ipaddr_t ip_s, ip_d;
	uint32 idx;
	int32 retval = PS_ACTION_NONE;
	ip_d = PKTGET_UINT32_UNALIGNED(&ip->ip_dst);
	ip_s = PKTGET_UINT32_UNALIGNED(&ip->ip_src);

	if ( !fromOutbound || (IS_FRAGMENT_REMAIN(ip->ip_off)) )
		goto out;

	switch (ip->ip_p)
	{
		case IPPROTO_TCP:
		{
			struct tcphdr * tc = (struct tcphdr *)L4HDR_ADDR(ip);
			sportp = &(tc->th_sport);
			dportp = &(tc->th_dport);
			cksump = &(tc->th_sum);

			break;
		}
		case IPPROTO_UDP:
		{
			struct udphdr * ud = (struct udphdr *)L4HDR_ADDR(ip);
			sportp = &(ud->uh_sport);
			dportp = &(ud->uh_dport);
			cksump = &(ud->uh_sum);

			break;
		}
		default:	/* for other L4 protocols, we don't care about them because we don't modify them in _rtl8651_aliasProtoStackSentPkt */
			goto out;
	}

	psa = rtl8651_ProtoStackActions;
	for ( idx = 0 ; idx < rtl8651_fwdEnginePara.protoStackMaxActions ; idx++, psa++ )
	{
		if (psa->action == PS_ACTION_NONE)
			continue;

		/* TCP or UDP here */
		if ((psa->dstIp == 0 || psa->dstIp == ip_s) && ip->ip_p == psa->protocol)	/* (1) dstIp & protocol match */
		{
			if(psa->role == PS_ACTION_SERVER)
				ckportp = sportp;
			else if (psa->role == PS_ACTION_CLIENT)
				ckportp = dportp;

			if (ntohs(*ckportp) >= psa->orgPort && ntohs(*ckportp) <= psa->orgPortEnd) 	/* (2) port match */
			{
				if (ip_d >= psa->lowerIp && ip_d <= psa->upperIp)		/* (3) ip range match */
				{
					/* the condition to perform port translation */
					if ((psa->action == PS_ACTION_TO_PS) && (psa->orgPort != psa->fwdPort))
					{	/* l4 checksum */
						uint16 newFwdPort = ntohs(*ckportp) - psa->orgPort + psa->fwdPort;						
						if ( (ip->ip_p != IPPROTO_UDP) || (*cksump != 0) )
							_RTL8651_L4_ADJUST_CHECKSUM( 0, 0, newFwdPort, ntohs(*ckportp), *cksump);
						*ckportp = htons(newFwdPort);
					}
					return psa->action;
				}
			}
		}
	}

out:
	return retval;
}
#endif /* _RTL_PS_ACTION */


#ifdef _RTL_PORT_BOUNCING
#define RTL8651_PORT_BOUNCING_FUNCTIONS

void _rtl8651_flushPortBouncingEntry(void)
{
	memset(portBouncingTable,0,PORT_BOUNCING_ITEM * sizeof (struct _RTL8651_PORT_BOUNCING_ENTRY));
}

/* the definitions is referenced by the priority to be allocated as new entry, and it can NOT been changed arbitrarily */
#define _RTL8651_PB_CANDIDATE_CLOSETCP	0
#define _RTL8651_PB_CANDIDATE_ICMP		1
#define _RTL8651_PB_CANDIDATE_UDP		2
#define _RTL8651_PB_CANDIDATE_OTHERTCP	3
#define _RTL8651_PB_CANDIDATE_CNT			4

static struct _RTL8651_PORT_BOUNCING_ENTRY * _rtl8651_addPortBouncingEntry(uint8 flowType,ipaddr_t cip,uint16 cport,ipaddr_t gip,uint16 gport,ipaddr_t sip,uint16 sport, uint16 id, uint32 tcpFlag, uint16 age)
{
	int i;
	struct _RTL8651_PORT_BOUNCING_ENTRY *ptr;
	struct _RTL8651_PORT_BOUNCING_ENTRY *candidate[_RTL8651_PB_CANDIDATE_CNT] = {NULL};

	ptr = portBouncingTable;

	for(i=0;i<PORT_BOUNCING_ITEM;i++, ptr++)
	{
		if(ptr->age == 0)
		{
setNewEntry:
			ptr->clientIpAddr		= cip;
			ptr->clientPort		= cport;
			ptr->serverIpAddr		= sip;
			ptr->serverPort		= sport;
			ptr->gatewayIpAddr	= gip;
			ptr->gatewayPort		= gport;
			ptr->id				= id;
			ptr->age				= age;
			ptr->flowType			= flowType;
			if (tcpFlag)
			{
				ptr->tcpFlag		= (tcpFlag - 1);
			}else
			{
				ptr->tcpFlag		= 0;
			}
			return ptr;
		}else
		{	/* for the TCP entry which doesn't be aged out yet and its state in CLOSE_STATE,
				if there is no any free entry for this new port bouncing flow, port bouncing system will
				destroy this TCP entry immediately and let it to store that new flow */
			if (ptr->flowType == PKTHDR_TCP)
			{
				if (ptr->tcpFlag == (TCP_CLOSED_FLOW - 1))
					candidate[_RTL8651_PB_CANDIDATE_CLOSETCP] = ptr;
				else if (	(candidate[_RTL8651_PB_CANDIDATE_OTHERTCP] == NULL) ||
						(candidate[_RTL8651_PB_CANDIDATE_OTHERTCP]->age > ptr->age)	)
					candidate[_RTL8651_PB_CANDIDATE_OTHERTCP] = ptr;
			} else if (ptr->flowType == PKTHDR_UDP)
			{
				if (		(candidate[_RTL8651_PB_CANDIDATE_UDP] == NULL) ||
						(candidate[_RTL8651_PB_CANDIDATE_UDP]->age > ptr->age)	)
					candidate[_RTL8651_PB_CANDIDATE_UDP] = ptr;
			} else if (ptr->flowType == PKTHDR_ICMP)
			{
				if (		(candidate[_RTL8651_PB_CANDIDATE_ICMP] == NULL) ||
						(candidate[_RTL8651_PB_CANDIDATE_ICMP]->age > ptr->age)	)
					candidate[_RTL8651_PB_CANDIDATE_ICMP] = ptr;
			}
		}
	}
	/*
		system didn't found aged-out entry, so they try reuse entry:

		Priority:
			1. fin-ed tcp
			2. icmp with minimum age
			3. udp with minimum age
			4. other tcp with minimum age
	*/
	for (i = 0 ; i < _RTL8651_PB_CANDIDATE_CNT ; i ++)
	{
		if (candidate[i])
		{
			ptr = candidate[i];
			goto setNewEntry;
		}
	}

	/* no any entry free, add failed */
	return NULL;
}

void _rtl8651_portBouncingTimeUpdate(uint32 secPassed)
{
	struct _RTL8651_PORT_BOUNCING_ENTRY *pb_t;
	int i;
	
	for(i=0, pb_t=portBouncingTable;i<PORT_BOUNCING_ITEM; i++, pb_t++)
	{
//		if (portBouncingTable[i].age < secPassed)
//			portBouncingTable[i].age = 0;
//		else
//			portBouncingTable[i].age -= secPassed;
		if (pb_t->age < secPassed)
		{
			pb_t->age = 0;
		} else pb_t->age -= secPassed;
	}
}

static int32 _rtl8651_l34PortBouncingIcmpErrMsg(struct rtl_pktHdr *pkthdr, struct ip *pip, struct icmp *ic)
{
	uint32 nCount;
	uint16 errFlowType;
	struct ip *eip;
//	ipaddr_t eip_addr;
	struct tcphdr *etc=NULL;
	struct udphdr *eud=NULL;
	uint16 esport=0, edport=0;
	struct icmp *eic;
	uint16 eid=0;

	// only icmp error message is accepted
	switch (ic->icmp_type)
	{
		case ICMP_TIMXCEED:
		case ICMP_UNREACH:
		case ICMP_SOURCEQUENCH:
		case ICMP_PARAMPROB:
			break;
		default:
			goto to_gw;				/* other icmp packet are rejected and trapped to protocol stack */
	}

	eip = &ic->icmp_ip;
	switch (eip->ip_p)
	{
		case IPPROTO_TCP:
			errFlowType = PKTHDR_TCP;
			etc = (struct tcphdr*) ((int8 *) eip + ((*(uint8*)eip&0xf) << 2));
			esport = ntohs(etc->th_sport);
			edport = ntohs(etc->th_dport);
			break;
		case IPPROTO_UDP:
			errFlowType = PKTHDR_UDP;
			eud = (struct udphdr*) ((int8 *) eip + ((*(uint8*)eip&0xf) << 2));
			esport = ntohs(eud->uh_sport);
			edport = ntohs(eud->uh_dport);
			break;
		case IPPROTO_ICMP:
			errFlowType = PKTHDR_ICMP;
			eic = (struct icmp*) ((int8 *) eip + ((*(uint8*)eip&0xf) << 2));
			eid = ntohs(eic->icmp_id);
			break;
		default:
			goto to_gw;
	}

	// Trace port Bouncing Table for error flow
	for ( nCount = 0 ; nCount < PORT_BOUNCING_ITEM ; nCount++ )
	{
		if (portBouncingTable[nCount].age == 0)
			continue;
		if (pkthdr->ph_dip != portBouncingTable[nCount].gatewayIpAddr)
			continue;
		if (errFlowType != portBouncingTable[nCount].flowType)
			continue;

		if (pkthdr->ph_sip == portBouncingTable[nCount].clientIpAddr)
		{	// Client to Server
			if (((eip->ip_p == IPPROTO_TCP) || (eip->ip_p == IPPROTO_UDP)) &&
				(esport == portBouncingTable[nCount].gatewayPort) &&
				(edport == portBouncingTable[nCount].clientPort))
			{	// Error flow found
				PKTSET_UINT32_UNALIGNED(portBouncingTable[nCount].serverIpAddr, &eip->ip_src);
				PKTSET_UINT32_UNALIGNED(portBouncingTable[nCount].gatewayIpAddr, &eip->ip_dst);
				_RTL8651_L3_ADJUST_CHECKSUM( portBouncingTable[nCount].serverIpAddr, portBouncingTable[nCount].clientIpAddr, eip->ip_sum);
				if (eip->ip_p == IPPROTO_TCP)
				{	// tcp
					etc->th_sport = htons(portBouncingTable[nCount].serverPort);
				}else
				{	// udp
					eud->uh_sport = htons(portBouncingTable[nCount].serverPort);
				}
			}else if (eid == portBouncingTable[nCount].id)
			{	// Error flow found
				PKTSET_UINT32_UNALIGNED(portBouncingTable[nCount].serverIpAddr, &eip->ip_src);
				PKTSET_UINT32_UNALIGNED(portBouncingTable[nCount].gatewayIpAddr, &eip->ip_dst);
				_RTL8651_L3_ADJUST_CHECKSUM( portBouncingTable[nCount].serverIpAddr, portBouncingTable[nCount].clientIpAddr, eip->ip_sum);
			}else
			{
				goto checkS2C;
			}
			pkthdr->ph_sip = pkthdr->ph_dip;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
			pkthdr->ph_dip = portBouncingTable[nCount].serverIpAddr;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
			goto out;
		}
checkS2C:
		if (pkthdr->ph_sip == portBouncingTable[nCount].serverIpAddr)
		{	// Server to Client
			if (((eip->ip_p == IPPROTO_TCP) || (eip->ip_p == IPPROTO_UDP)) &&
				(esport == portBouncingTable[nCount].clientPort) &&
				(edport == portBouncingTable[nCount].serverPort))
			{	// Error flow found
				PKTSET_UINT32_UNALIGNED(portBouncingTable[nCount].clientIpAddr, &eip->ip_src);
				PKTSET_UINT32_UNALIGNED(portBouncingTable[nCount].gatewayIpAddr, &eip->ip_dst);
				_RTL8651_L3_ADJUST_CHECKSUM(portBouncingTable[nCount].clientIpAddr, portBouncingTable[nCount].serverIpAddr, eip->ip_sum);
				if (eip->ip_p == IPPROTO_TCP)
				{	// tcp
					etc->th_dport = htons(portBouncingTable[nCount].gatewayPort);
				}else
				{	// udp
					eud->uh_dport = htons(portBouncingTable[nCount].gatewayPort);
				}
			}else if (eid == portBouncingTable[nCount].id)
			{	// Error flow found
				PKTSET_UINT32_UNALIGNED(portBouncingTable[nCount].clientIpAddr, &eip->ip_src);
				PKTSET_UINT32_UNALIGNED(portBouncingTable[nCount].gatewayIpAddr, &eip->ip_dst);
				_RTL8651_L3_ADJUST_CHECKSUM(portBouncingTable[nCount].clientIpAddr, portBouncingTable[nCount].serverIpAddr, eip->ip_sum);
			
			}else
			{
				continue;
			}
			pkthdr->ph_sip = pkthdr->ph_dip;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
			pkthdr->ph_dip = portBouncingTable[nCount].clientIpAddr;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
			goto out;
		}
	}
//failed:
	return FAILED;	// drop
to_gw:
	return FWDENG_TO_GW;	// trap to protocol stack
out:
	return SUCCESS;	// need Port bouncing
}

/*
	NAPT Port Bouncing data-path
		* exchange IP/Port
*/
static inline int32 _rtl8651_l4NaptPortBouncing(struct rtl_pktHdr *pkthdr, struct ip *pip)
{
	struct _RTL8651_PORT_BOUNCING_ENTRY *pb_entry = NULL;
	rtl8651_tblDrv_vlanTable_t * ingress_vlan;
	uint32 netIdx;
	struct tcphdr *tc=NULL;
	struct udphdr *ud=NULL;
	struct icmp *ic=NULL;
	ipaddr_t sip = pkthdr->ph_sip;
	ipaddr_t dip = pkthdr->ph_dip;
	uint16 sport=0;
	uint16 dport=0;
	uint16 id=0;
	int16 idx = -1;
	int16 nCount;

	// get gatway ip's network interface
	if (_rtl8651_extIpAddr(dip, &netIdx) != TRUE)
	{
		goto to_gw;
	}

	ingress_vlan = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	if (!ingress_vlan->internal)
	{
		goto to_gw;
	}

#ifdef _RTL_NAPT_FRAG
	if (ntohs(pip->ip_off) & IP_OFFMASK)
	{
		ipaddr_t mod_ip;

		if (_rtl8651_getFragCache(BOUNCING_FRAG, sip, dip, ntohs(pip->ip_id), &mod_ip, NULL, NULL, &(pkthdr->ph_dsid)) != SUCCESS)
		{
			goto to_gw;
		}
		pkthdr->ph_sip = pkthdr->ph_dip;
		pkthdr->ph_dip = mod_ip;
		PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
		PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
		goto out;
	}
#else
	if (ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK))
	{
		goto to_gw;
	}
#endif /* _RTL_NAPT_FRAG */


	switch (pkthdr->ph_proto)
	{
		case PKTHDR_TCP:
			tc = (struct tcphdr*)L4HDR_ADDR(pip);
			sport = ntohs(tc->th_sport);
			dport = ntohs(tc->th_dport);
			break;
		case PKTHDR_UDP:
			ud = (struct udphdr*)L4HDR_ADDR(pip);
			sport = ntohs(ud->uh_sport);
			dport = ntohs(ud->uh_dport);
			break;
		case PKTHDR_ICMP:
			ic = (struct icmp*)L4HDR_ADDR(pip);
			id = ntohs(ic->icmp_id);
			break;
		default:
			goto to_gw;
	}

	// Check port bouncing table for existing flow
	for ( nCount = 0 ; nCount < PORT_BOUNCING_ITEM ; nCount++ )
	{
		if (portBouncingTable[nCount].age == 0)
			continue;

		if (pkthdr->ph_proto != portBouncingTable[nCount].flowType)
			continue;

		if (dip != portBouncingTable[nCount].gatewayIpAddr)
			continue;

		if (sip == portBouncingTable[nCount].clientIpAddr)
		{	// Client to Server
			if( (((pkthdr->ph_proto == PKTHDR_TCP) ||(pkthdr->ph_proto == PKTHDR_UDP))) &&
				(sport == portBouncingTable[nCount].clientPort) &&
				(dport == portBouncingTable[nCount].gatewayPort))
			{	// Match TCP/UDP entry
				if(pkthdr->ph_proto == PKTHDR_TCP)
				{	// tcp
					tc->th_dport = htons(portBouncingTable[nCount].serverPort);
					portBouncingTable[nCount].age =
						_rtl8651_portBouncingTcpMonitor(tc, &(portBouncingTable[nCount].tcpFlag), TRUE);
				}else
				{	// udp
					int j;
					for(j=0;j < PORT_BOUNCING_ITEM ;j++ )
					{
						if (portBouncingTable[j].age == 0)
							continue;

						if (pkthdr->ph_proto != portBouncingTable[j].flowType)
							continue;

						if (sip != portBouncingTable[j].serverIpAddr)
							continue;					

						if (ud->uh_sport != portBouncingTable[j].serverPort)
							continue;					
						
						ud->uh_sport= portBouncingTable[j].gatewayPort;
						break;				
					}
					ud->uh_dport = htons(portBouncingTable[nCount].serverPort);
					portBouncingTable[nCount].age = rtl8651GlobalControl.udpTimeout;
				}
			}else if ((pkthdr->ph_proto == PKTHDR_ICMP) && (id == portBouncingTable[nCount].id))	// icmp
			{	// Match icmp entry
				switch (ic->icmp_type)
				{
					#if 0	// gateway send to DMZ Host will use internal interface instead of external interface
					case ICMP_ECHOREPLY:
					case ICMP_TSTAMPREPLY:
 						goto to_gw;	// corresponding ICMP_ECHO/TSTAMP is sent by gateway!
					#endif
					case ICMP_ECHO:
					case ICMP_TSTAMP:
						if (!rtl8651_DMZHostIcmpPassthroughEnable)
							goto to_gw;	// we don't flush this portBouncingEntry here, let it timeout (for un-replied ECHOREPLY/TSTAMPREPLY)
						break;
					default:
						break;
				}
				/* don't modify icmp header here */
				portBouncingTable[nCount].age = rtl8651GlobalControl.icmpTimeout;

			}else
			{
				goto checkS2C;
			}
			// Modify ip header
			pkthdr->ph_sip = pkthdr->ph_dip;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
			pkthdr->ph_dip = portBouncingTable[nCount].serverIpAddr;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
			idx = nCount;
			pb_entry = &(portBouncingTable[nCount]);
			goto out;	// success, out
		}
checkS2C:
		if (sip == portBouncingTable[nCount].serverIpAddr)
		{	//Server to Client
			if (((pkthdr->ph_proto == PKTHDR_TCP) || (pkthdr->ph_proto == PKTHDR_UDP)) &&
				(sport == portBouncingTable[nCount].serverPort) &&
				(dport == portBouncingTable[nCount].clientPort))
			{	// Match TCP/UDP entry
				if (pkthdr->ph_proto == PKTHDR_TCP)
				{	// tcp
					tc->th_sport = htons(portBouncingTable[nCount].gatewayPort);
					portBouncingTable[nCount].age =
						_rtl8651_portBouncingTcpMonitor(tc, &(portBouncingTable[nCount].tcpFlag), FALSE);
				}else
				{	// udp
					int j;
					for(j=0;j < PORT_BOUNCING_ITEM ;j++ )
					{
						if (portBouncingTable[j].age == 0)
							continue;

						if (pkthdr->ph_proto != portBouncingTable[j].flowType)
							continue;

						if (sip != portBouncingTable[j].clientIpAddr)
							continue;					

						if (ud->uh_sport != portBouncingTable[j].clientPort)
							continue;					
						
						ud->uh_dport= portBouncingTable[j].serverPort;
						break;				
					}				
					ud->uh_sport = htons(portBouncingTable[nCount].gatewayPort);
					portBouncingTable[nCount].age = rtl8651GlobalControl.udpTimeout;
				}
			}else if ((pkthdr->ph_proto == PKTHDR_ICMP) && (id == portBouncingTable[nCount].id))	// icmp
			{	// Match icmp entry
				/* don't modify icmp header here */
				portBouncingTable[nCount].age = rtl8651GlobalControl.icmpTimeout;
			}else
			{
				continue;
			}
			// Modify ip header
			pkthdr->ph_sip = pkthdr->ph_dip;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
			pkthdr->ph_dip = portBouncingTable[nCount].clientIpAddr;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
			idx = nCount;
			pb_entry = &(portBouncingTable[nCount]);
			goto out;	// success, out
		}
	}

	{
		/* 
			Now, we are going to establish a new flow, 
			but we have to check the ingress ACL of "L4NEWFLOW" type first.
		*/
		int32 local_val;
		rtl8651_tblDrv_vlanTable_t *local_vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx];
		local_val = _rtl8651_ingressAcl(pkthdr->ph_dsid, 0, local_vlanp, local_vlanp->vid, pkthdr->ph_mbuf->m_data, (uint8 *) pip, RTL8651_ACL_L4NEWFLOW, _RTL8651_OPERATION_OTHER);
		if (((local_val > 0) && (local_val & FWDENG_DROP)) ||(local_val == FWDENG_ACL_IN_USERTRAP))
			return local_val;
	}

// we can't find existing port bouncing flow entry for this packet, we try to find/create a new one
//FindNewPBFlow:
		if(idx == -1)
		{
			rtl8651_tblDrv_naptServerPortEntry_t * sp = NULL;
			if(pkthdr->ph_proto == PKTHDR_TCP)
			{
				sp = _rtl8651_getOutsideNaptServerPortMapping(TRUE, dip, dport);
				if (sp)
				{
					if ((pb_entry = _rtl8651_addPortBouncingEntry(PKTHDR_TCP,
						sip, sport,
						dip, dport,
						sp->localIp, (dport + sp->localPort  - sp->globalPort),
						0,
						TCP_OUTBOUND_SYN_RCVD_FLOW,
						rtl8651GlobalControl.tcpFastTimeout)) == NULL)
					{
						goto drop;
					}
					tc->th_dport = htons((dport + sp->localPort - sp->globalPort));
				}
			}else if(pkthdr->ph_proto == PKTHDR_UDP)
			{
				sp = _rtl8651_getOutsideNaptServerPortMapping(FALSE, dip, dport);
				if (sp)
				{
					if ((pb_entry = _rtl8651_addPortBouncingEntry(PKTHDR_UDP,
						sip, sport,
						dip, dport,
						sp->localIp, (dport + sp->localPort - sp->globalPort),
						0,
						0,
						rtl8651GlobalControl.udpTimeout)) == NULL)
					{
						goto drop;
					}
					ud->uh_dport = htons((dport + sp->localPort - sp->globalPort));
				}
			}
			if (sp)
			{	// set ip header
				pkthdr->ph_sip = pkthdr->ph_dip;
				PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
				pkthdr->ph_dip = sp->localIp;
				PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
				goto out;
			}
			assert(!sp);
//FinduPnP:
			if (rtl8651_drvProcessUpnpPortBouncing == TRUE)	/* we only check uPnP Mapping when rtl8651_drvProcessUpnpPortBouncing is set */
			{
				if ((pkthdr->ph_proto == PKTHDR_TCP) || (pkthdr->ph_proto == PKTHDR_UDP))
				{
					ipaddr_t localIp = 0;
					uint16 localPort = 0;

					/* check uPNP */
					_rtl8651_findInboundUpnpMapping((pkthdr->ph_proto == PKTHDR_TCP)?TRUE:FALSE, sip, sport, dip, dport, &localIp, &localPort, NULL, NULL);
					if (localIp)
					{	/* uPNP entry found */
						if ((pb_entry = _rtl8651_addPortBouncingEntry(pkthdr->ph_proto,
							sip, sport,
							dip, dport,
							localIp, localPort,
							0,
							(pkthdr->ph_proto == PKTHDR_TCP)?TCP_OUTBOUND_SYN_RCVD_FLOW:0,
							(pkthdr->ph_proto == PKTHDR_TCP)?rtl8651GlobalControl.tcpFastTimeout:rtl8651GlobalControl.udpTimeout)) == NULL)
						{
							goto drop;
						}
						if (pkthdr->ph_proto == PKTHDR_TCP)
							tc->th_dport = htons(localPort);
						else
							ud->uh_dport = htons(localPort);

						/* set packet hdr */
						pkthdr->ph_sip = pkthdr->ph_dip;
						PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
						pkthdr->ph_dip = localIp;
						PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
						goto out;
					}
				}
			}

/* 
	The reason to check PS action here is that the priority is ...
	(1) Server port
	(2) UPnP
	(3) PS action
	(4) DMZ host
	
*/
#ifdef _RTL_PS_ACTION
	if (rtl8651_drvEnPortBouncingPSaction) {
		/* Check if there's any rule for NAPT port bouncing pkt to protocol stack. */
		int32 val;

		val = _rtl8651_doProtoStackAction(pip);
		if (val == PS_ACTION_TO_PS)
			return FWDENG_TO_GW;
		else if (val == PS_ACTION_DROP)
			return FWDENG_USER_DROP;
	}
#endif	/* _RTL_PS_ACTION */



//FindDMZ:
			{
				ipaddr_t dmzhost;
				if (_rtl8651_getDmzHost(pkthdr->ph_dsid, dip, &dmzhost) == SUCCESS)
				{	// don't change port/id, chang ip only
					switch (pkthdr->ph_proto)
					{
						case PKTHDR_TCP:
							if ((pb_entry = _rtl8651_addPortBouncingEntry(PKTHDR_TCP,
								sip, sport,
								dip, dport,
								dmzhost, dport,
								0,
								TCP_OUTBOUND_SYN_RCVD_FLOW,
								rtl8651GlobalControl.tcpFastTimeout)) == NULL)
							{
								goto drop;
							}
							break;
						case PKTHDR_UDP:
							if ((pb_entry = _rtl8651_addPortBouncingEntry(PKTHDR_UDP,
								sip, sport,
								dip, dport,
								dmzhost, dport,
								0,
								0,
								rtl8651GlobalControl.udpTimeout)) == NULL)
							{
								goto drop;
							}
							break;
						case PKTHDR_ICMP:
							switch (ic->icmp_type)
							{
								#if 0	// gateway send to DMZ Host will use internal interface instead of external interface
								case ICMP_ECHOREPLY:
								case ICMP_TSTAMPREPLY:
									goto to_gw;	// corresponding ICMP_ECHO/TSTAMP is sent by gateway!
								#endif
								case ICMP_ECHO:
								case ICMP_TSTAMP:
									if (!rtl8651_DMZHostIcmpPassthroughEnable)
										goto to_gw;
									break;
								case ICMP_TIMXCEED:
								case ICMP_UNREACH:
								case ICMP_SOURCEQUENCH:
								case ICMP_PARAMPROB:
									goto icmp_errMsg;
								default:
									break;
							}
							if ((pb_entry = _rtl8651_addPortBouncingEntry(PKTHDR_ICMP,
								sip, 0,
								dip, 0,
								dmzhost, 0,
								id,
								0,
								rtl8651GlobalControl.icmpTimeout)) == NULL)
							{
								goto drop;
							}
							break;
					}
					pkthdr->ph_sip = pkthdr->ph_dip;
					PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
					pkthdr->ph_dip = dmzhost;
					PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
					goto out;
				}
			}
		}
		goto icmp_errMsg;	// it has no exist entry and it can not create new netry, trap to protocol stack
out:
	{
		// modify and flush pkthdr information
		rtl8651_tblDrv_networkIntfTable_t * netIface = &DrvTbl.netIface[netIdx];

		pkthdr->ph_vlanIdx = netIface->vid;
		pkthdr->ph_routeIdx = 0;
		pkthdr->ph_dsid = 0;
		pkthdr->ph_portlist = 0;
	}
#ifdef _RTL_NAPT_FRAG
	if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK)) == IP_MF))
	{
		if ((pip->ip_vhl & 0x0f) != 5)
		{
			_RTL8651_L3_ADJUST_CHECKSUM(pkthdr->ph_sip, sip, pip->ip_sum);
			_RTL8651_L3_ADJUST_CHECKSUM(pkthdr->ph_dip, dip, pip->ip_sum);
		}
		/* modify L4 Checksum */
		if ((ntohs(pip->ip_off) & IP_OFFMASK) == 0)
		{
			if (pkthdr->ph_proto == PKTHDR_TCP)
			{
				_RTL8651_L4_ADJUST_CHECKSUM(pkthdr->ph_sip, sip, tc->th_sport, htons(sport), tc->th_sum);
				_RTL8651_L4_ADJUST_CHECKSUM(pkthdr->ph_dip, dip, tc->th_dport, htons(dport), tc->th_sum);
			}else if ((pkthdr->ph_proto == PKTHDR_UDP) && (ud->uh_sum))
			{
				_RTL8651_L4_ADJUST_CHECKSUM(pkthdr->ph_sip, sip, ud->uh_sport, htons(sport), ud->uh_sum);
				_RTL8651_L4_ADJUST_CHECKSUM(pkthdr->ph_dip, dip, ud->uh_dport, htons(dport), ud->uh_sum);
			}
			/* for ICMP packet, we don't modify icmp id, and icmp checksum doesn't care ip modification, so we don't modify it */

			if (IS_FRAGMENT(pip->ip_off))
			{
				_rtl8651_updateFragCache(BOUNCING_FRAG, sip, dip, ntohs(pip->ip_id), pkthdr->ph_dip, 0, NULL, pkthdr->ph_dsid);
			}
		}
	}
#endif /* _RTL_NAPT_FRAG */
	if (_rtl8651_l3ActRouting(pkthdr, pkthdr->ph_dip, pip) != SUCCESS)
	{
		goto drop;
	}
	return SUCCESS;

icmp_errMsg:
	if (pkthdr->ph_proto == PKTHDR_ICMP)
	{
		int32 retval;
		retval = _rtl8651_l34PortBouncingIcmpErrMsg(pkthdr, pip, ic);
		if (retval == SUCCESS)
		{
			goto out;
		}else if (retval == FWDENG_TO_GW)
		{
			goto to_gw;
		}
		// failed: drop
	}else
	{
		goto to_gw;
	}
drop:
	if (pb_entry)
	{
		pb_entry->age = 0;
	}
	return FWDENG_DROP;
to_gw:
	// non-port bouncing: return failed and trapped to protocol stack
	return FAILED;
}

static inline int32 _rtl8651_l4NatPortBouncing(struct rtl_pktHdr *pkthdr, struct ip *pip)
{
	struct _RTL8651_PORT_BOUNCING_ENTRY *pb_entry = NULL;
	rtl8651_tblDrv_vlanTable_t * ingress_vlan;
	rtl8651_tblDrv_natEntry_t * nat;
	struct tcphdr *tc=0;
	struct udphdr *ud=0;
	ipaddr_t sip = pkthdr->ph_sip;
	ipaddr_t dip = pkthdr->ph_dip;
	uint16 sport=0, dport=0;
	struct icmp *ic=0;
	uint16 id=0;
	int16 idx = -1;
	int16 nCount;

	// get gatway ip's network interface
	if ((nat = _rtl8651_getNatExternalMappingEntry(dip, pkthdr)) == NULL)
	{
		goto to_gw;
	}

	ingress_vlan = &DrvTbl.vlan[pkthdr->ph_vlanIdx & (RTL8651_VLAN_NUMBER - 1)];
	if (!ingress_vlan->internal)
	{
		goto to_gw;
	}

	// for ip-unnumber
	if (nat->globalIp == nat->localIp)
	{
		goto out;
	}

#ifdef _RTL_NAPT_FRAG
	if (ntohs(pip->ip_off) & IP_OFFMASK)
	{
		ipaddr_t mod_ip;

		if (_rtl8651_getFragCache(BOUNCING_FRAG, sip, dip, ntohs(pip->ip_id), &mod_ip, NULL, NULL, &(pkthdr->ph_dsid)) != SUCCESS)
		{
			goto to_gw;
		}

		pkthdr->ph_sip = pkthdr->ph_dip;
		pkthdr->ph_dip = mod_ip;
		PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
		PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
		goto out;
	}
#else
	if (ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK))
	{
		goto to_gw;
	}
#endif /* _RTL_NAPT_FRAG */

	switch (pkthdr->ph_proto)
	{
		case PKTHDR_TCP:
			tc = (struct tcphdr*)L4HDR_ADDR(pip);
			sport = ntohs(tc->th_sport);
			dport = ntohs(tc->th_dport);
			break;
		case PKTHDR_UDP:
			ud = (struct udphdr*)L4HDR_ADDR(pip);
			sport = ntohs(ud->uh_sport);
			dport = ntohs(ud->uh_dport);
			break;
		case PKTHDR_ICMP:
			ic = (struct icmp*)L4HDR_ADDR(pip);
			id = ntohs(ic->icmp_id);
			break;
		default:
			goto to_gw;
	}

	// Check port bouncing table for existing flow
	for ( nCount = 0 ; nCount < PORT_BOUNCING_ITEM ; nCount++ )
	{
		if (portBouncingTable[nCount].age == 0)
			continue;

		if (pkthdr->ph_proto != portBouncingTable[nCount].flowType)
			continue;

		if (pkthdr->ph_dip != portBouncingTable[nCount].gatewayIpAddr)
			continue;

		if (pkthdr->ph_sip == portBouncingTable[nCount].clientIpAddr)
		{	// Client to Server
			if( (((pkthdr->ph_proto == PKTHDR_TCP) ||(pkthdr->ph_proto == PKTHDR_UDP))) &&
				(sport == portBouncingTable[nCount].clientPort) &&
				(dport == portBouncingTable[nCount].gatewayPort))
			{	// Match TCP/UDP entry
				if(pkthdr->ph_proto == PKTHDR_TCP)
				{	// tcp
					tc->th_dport = htons(portBouncingTable[nCount].serverPort);
					portBouncingTable[nCount].age =
						_rtl8651_portBouncingTcpMonitor(tc, &(portBouncingTable[nCount].tcpFlag), TRUE);
				}else
				{	// udp
					ud->uh_dport = htons(portBouncingTable[nCount].serverPort);
					portBouncingTable[nCount].age = rtl8651GlobalControl.udpTimeout;
				}
			}else if ((pkthdr->ph_proto == PKTHDR_ICMP) && (id == portBouncingTable[nCount].id))	// icmp
			{	// Match icmp entry
				portBouncingTable[nCount].age = rtl8651GlobalControl.icmpTimeout;

			}else
			{
				goto checkS2C;
			}
			// Modify ip header
			pkthdr->ph_sip = pkthdr->ph_dip;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
			pkthdr->ph_dip = portBouncingTable[nCount].serverIpAddr;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
			idx = nCount;
			pb_entry = &(portBouncingTable[nCount]);
			goto out;	// success, out
		}
checkS2C:
		if (pkthdr->ph_sip == portBouncingTable[nCount].serverIpAddr)
		{	//Server to Client
			if (((pkthdr->ph_proto == PKTHDR_TCP) || (pkthdr->ph_proto == PKTHDR_UDP)) &&
				(sport == portBouncingTable[nCount].serverPort) &&
				(dport == portBouncingTable[nCount].clientPort))
			{	// Match TCP/UDP entry
				if (pkthdr->ph_proto == PKTHDR_TCP)
				{	// tcp
					tc->th_sport = htons(portBouncingTable[nCount].gatewayPort);
					portBouncingTable[nCount].age =
						_rtl8651_portBouncingTcpMonitor(tc, &(portBouncingTable[nCount].tcpFlag), FALSE);
				}else
				{	// udp
					ud->uh_sport = htons(portBouncingTable[nCount].gatewayPort);
					portBouncingTable[nCount].age = rtl8651GlobalControl.udpTimeout;
				}
			}else if ((pkthdr->ph_proto == PKTHDR_ICMP) && (id == portBouncingTable[nCount].id))	// icmp
			{	// Match icmp entry
				portBouncingTable[nCount].age = rtl8651GlobalControl.icmpTimeout;
			}else
			{
				continue;
			}
			// Modify ip header
			pkthdr->ph_sip = pkthdr->ph_dip;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
			pkthdr->ph_dip = portBouncingTable[nCount].clientIpAddr;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
			idx = nCount;
			pb_entry = &(portBouncingTable[nCount]);
			goto out;	// success, out
		}
	}

	{
		/* 
			Now, we are going to establish a new flow, 
			but we have to check the ingress ACL of "L4NEWFLOW" type first.
		*/
		int32 local_val;
		rtl8651_tblDrv_vlanTable_t *local_vlanp = &DrvTbl.vlan[pkthdr->ph_vlanIdx];
		local_val = _rtl8651_ingressAcl(pkthdr->ph_dsid, 0, local_vlanp, local_vlanp->vid, pkthdr->ph_mbuf->m_data, (uint8 *) pip, RTL8651_ACL_L4NEWFLOW, _RTL8651_OPERATION_OTHER);
		if (((local_val > 0) && (local_val & FWDENG_DROP)) ||(local_val == FWDENG_ACL_IN_USERTRAP))
			return local_val;
	}
	
//FindNewPBFlow:

	if (idx == -1)
	{
		if (sip != nat->localIp)	// it will cause self-transmit
		{
			switch (pkthdr->ph_proto)
			{
				case PKTHDR_TCP:
					if ((pb_entry = _rtl8651_addPortBouncingEntry(PKTHDR_TCP,
						sip, sport,
						dip, dport,
						nat->localIp, dport,
						0,
						TCP_OUTBOUND_SYN_RCVD_FLOW,
						rtl8651GlobalControl.tcpFastTimeout)) == NULL)
					{
						goto drop;
					}
					break;
				case PKTHDR_UDP:
					if ((pb_entry = _rtl8651_addPortBouncingEntry(PKTHDR_UDP,
						sip, sport,
						dip, dport,
						nat->localIp, dport,
						0,
						0,
						rtl8651GlobalControl.udpTimeout)) == NULL)
					{
						goto drop;
					}
					break;
				case PKTHDR_ICMP:
					switch(ic->icmp_type)
					{
						case ICMP_TIMXCEED:
						case ICMP_UNREACH:
						case ICMP_SOURCEQUENCH:
						case ICMP_PARAMPROB:
							goto icmp_errMsg;
					}
					if ((pb_entry = _rtl8651_addPortBouncingEntry(PKTHDR_ICMP,
						sip, 0,
						dip, 0,
						nat->localIp, 0,
						id,
						0,
						rtl8651GlobalControl.icmpTimeout)) == NULL)
					{
						goto drop;
					}
					break;
			}
			pkthdr->ph_sip = pkthdr->ph_dip;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_sip, &pip->ip_src);
			pkthdr->ph_dip = nat->localIp;
			PKTSET_UINT32_UNALIGNED(pkthdr->ph_dip, &pip->ip_dst);
			goto out;
		}
	}

#ifdef _RTL_PS_ACTION
	if (rtl8651_drvEnPortBouncingPSaction) {
		/* Check if there's any rule for NAT port bouncing pkt to protocol stack. */
		int32 val;

		val = _rtl8651_doProtoStackAction(pip);
		if (val == PS_ACTION_TO_PS)
			return FWDENG_TO_GW;
		else if (val == PS_ACTION_DROP)
			return FWDENG_USER_DROP;
	}
#endif	/* _RTL_PS_ACTION */

	goto icmp_errMsg;	// no entry: test if it's icmp_error msg, or trap to protocol stack
out:
	{
		// modify and flush pkthdr information
		uint32 netIdx;
		int8 ret;
		rtl8651_tblDrv_networkIntfTable_t *netIface;
		// get network interface
		ret = _rtl8651_extNetIpAddr(nat->globalIp, &netIdx);
		assert(ret == TRUE);
		netIface = &DrvTbl.netIface[netIdx];
		pkthdr->ph_vlanIdx = netIface->vid;
		pkthdr->ph_routeIdx = 0;
		pkthdr->ph_dsid = 0;
		pkthdr->ph_portlist = 0;
	}
#ifdef _RTL_NAPT_FRAG
	if (((pip->ip_vhl & 0x0f) != 5) || ((ntohs(pip->ip_off) & (IP_MF|IP_OFFMASK)) == IP_MF))
	{
		if ((pip->ip_vhl & 0x0f) != 5)
		{
			_RTL8651_L3_ADJUST_CHECKSUM(pkthdr->ph_sip, sip, pip->ip_sum);
			_RTL8651_L3_ADJUST_CHECKSUM(pkthdr->ph_dip, dip, pip->ip_sum);
		}
		/* modify L4 Checksum */
		if ((ntohs(pip->ip_off) & IP_OFFMASK) == 0)
		{
			if (pkthdr->ph_proto == PKTHDR_TCP)
			{
				_RTL8651_L4_ADJUST_CHECKSUM(pkthdr->ph_sip, sip, tc->th_sport, htons(sport), tc->th_sum);
				_RTL8651_L4_ADJUST_CHECKSUM(pkthdr->ph_dip, dip, tc->th_dport, htons(dport), tc->th_sum);
			}else if ((pkthdr->ph_proto == PKTHDR_UDP) && (ud->uh_sum))
			{
				_RTL8651_L4_ADJUST_CHECKSUM(pkthdr->ph_sip, sip, ud->uh_sport, htons(sport), ud->uh_sum);
				_RTL8651_L4_ADJUST_CHECKSUM(pkthdr->ph_dip, dip, ud->uh_dport, htons(dport), ud->uh_sum);
			}
			/* for ICMP packet, we don't modify icmp id, and icmp checksum doesn't care ip modification, so we don't modify it */

			if (IS_FRAGMENT(pip->ip_off))
			{
				_rtl8651_updateFragCache(BOUNCING_FRAG, sip, dip, ntohs(pip->ip_id), pkthdr->ph_dip, 0, NULL, pkthdr->ph_dsid);
			}
		}
	}
#endif /* _RTL_NAPT_FRAG */
	if (_rtl8651_l3ActRouting(pkthdr, pkthdr->ph_dip, pip) != SUCCESS)
	{
		goto drop;
	}
	return SUCCESS;
icmp_errMsg:
	if (pkthdr->ph_proto == PKTHDR_ICMP)
	{
		int32 retval;
		retval = _rtl8651_l34PortBouncingIcmpErrMsg(pkthdr, pip, ic);
		if (retval == SUCCESS)
		{
			goto out;
		}else if (retval == FWDENG_TO_GW)
		{
			goto to_gw;
		}
		// failed: drop
	}else
	{
		goto to_gw;
	}
drop:
	if (pb_entry)
	{
		pb_entry->age = 0;
	}
	return FWDENG_DROP;
to_gw:

	return FAILED;
}

static inline uint32 _rtl8651_portBouncingTcpMonitor(struct tcphdr *tc, uint32 *currentState, int8 IsClientToServer)
{
	uint32 pktFlags;
	uint32 retval = 0;
	int8	*StateTable = NULL;
	
	/* software auto delete is alwyas on */

	pktFlags = tc->th_flags &  (TH_RST|TH_SYN|TH_FIN);
	assert((*currentState) <= (TCP_CLOSED_FLOW-1));

	if (IsClientToServer == TRUE)
	{
		StateTable = (int8*)_outNextState;
	}
	else
	{
		StateTable = (int8*)_inNextState;
	}

	if(pktFlags & TH_RST){
		*currentState = (TCP_CLOSED_FLOW - 1);
		retval = (rtl8651GlobalControl.tcpFastTimeout);
	}else if(pktFlags == 0){ 
resetage:
		//No special flags in packet, just update age timer
		retval = (*_ageTimer[(*currentState)+1]);
	}else if(pktFlags == TH_SYN){
		//only SYN set
		int8 state = StateTable[(8*0) + (*currentState)];
		assert(state >=0);
		if(state > 0){
			*currentState = (state - 1);
			retval = *(_ageTimer[(int)state]);
		}else if(state == 0) //No action, just reset age timer
			goto resetage;
	}else{
		//At least FIN set
		int8 state = StateTable[(8*1) + (*currentState)];
		assert(state >=0);
		if(state > 0){
			*currentState = (state - 1);
			retval = *(_ageTimer[(int)state]);
		}else if(state == 0) //No action, just reset age timer
			goto resetage;
	}
	return retval;
}
#endif  // _RTL_PORT_BOUNCING

#ifdef _RTL_TRIGGER_PORT
#define RTL8651_TRIGGER_PORT_FUNCTIONS
/* ==============================================================================
  *				- trigger port
  *
  *		Note: Some differences are shown in trigger port's hash algorithm:
  *			  1. Outbound flow only 'hash' using the target Network Interface ID,
  *			      (Actually we use the directly mapping to do this.)
  *			      because when system add the trigger port entry, Network Interface ID is the
  *			      only parameter to exactly identify one hash table entry. (other port range based
  *			      parameters is useless when packet coming.)
  *
  *			  2. For normal Inbound flows, hash algorithm using the external ip
  *			      because this information will be known when create cache flow.
  *
  *			  3. For static Inbound flow caches(Local IP is assigned, but don't care External IP),
  *			      all inbound traffic which don't match other trigger port cache will try to match them.
  *			      Ths data structure of them is just an simple linked list.
  *
  * ============================================================================== */

/* <- For normal Entry and Cache Link/Unlink -> */
#define _TRIGGERPORT_ENTRY_CACHE_LINK(HdrEntry,LinkCache) \
		do { \
			(LinkCache)->entryHdr=(HdrEntry); \
			(LinkCache)->entryList_next=(HdrEntry)->normalCache; \
			(LinkCache)->entryList_prev=NULL; \
			if ((HdrEntry)->normalCache){(HdrEntry)->normalCache->entryList_prev=(LinkCache);}\
			(HdrEntry)->normalCache=(LinkCache);\
		} while (0);

#define _TRIGGERPORT_ENTRY_CACHE_UNLINK(UnlinkCache) \
		do { \
			if ((UnlinkCache)->entryList_prev){(UnlinkCache)->entryList_prev->entryList_next=(UnlinkCache)->entryList_next;} \
			if ((UnlinkCache)->entryList_next){(UnlinkCache)->entryList_next->entryList_prev=(UnlinkCache)->entryList_prev;} \
			if ((UnlinkCache)->entryHdr->normalCache==(UnlinkCache)){(UnlinkCache)->entryHdr->normalCache=(UnlinkCache)->entryList_next;} \
		} while (0);

/*
  *
  *	Trigger Port system, data structure controlling function: free one cache.
  *
  */
static inline void _rtl8651_TriggerPort_FreeCache(uint8 isStatic, _rtl8651_triggerPort_cache_t *cache)
{
	/* check normal cache hashed linked list */
	_NORMAL_UNLINK(cache);
	HTLIST_REMOVE(cache, lru_next, lru_prev, triggerPort_cache_lruHdr, triggerPort_cache_lruTail);

	/* check normal-(Entry<->Cache)-link */
	if (isStatic == FALSE)
	{
		_TRIGGERPORT_ENTRY_CACHE_UNLINK(cache);
	}

	/* clear cache */
	memset(cache, 0, sizeof(_rtl8651_triggerPort_cache_t));

	/* free this cache */
	_PUT_TO_FREELIST(cache, triggerPort_cache_freeList);
}

/*
  *
  *	Trigger Port system, data structure controlling function: free one cache.
  *
  */
static inline _rtl8651_triggerPort_cache_t * _rtl8651_TriggerPort_getCache(void)
{
	_rtl8651_triggerPort_cache_t *cache = NULL;

	if (	(triggerPort_cache_freeList == NULL) &&
		(triggerPort_cache_lruHdr) )
	{
		_rtl8651_TriggerPort_FreeCache((cache->entryHdr)?FALSE:TRUE, cache);
		assert(triggerPort_cache_freeList);
	}

	/* get new static cache from free list */
	_GET_FROM_FREELIST(cache, triggerPort_cache_freeList);

	return cache;
}

/*
  *
  *	Trigger Port system, data structure controlling function: free one trigger port entry.
  *
  */
static inline void _rtl8651_TriggerPort_FreeEntry(_rtl8651_triggerPort_entry_t *entry)
{
	/* check normal entry hashed linked list */
	_NORMAL_UNLINK(entry);

	/* check normal-(Entry<->Cache)-link */
	while (entry->normalCache)
	{
		_rtl8651_TriggerPort_FreeCache(FALSE, entry->normalCache);
	}

	/* free this entry */
	_PUT_TO_FREELIST(entry, triggerPort_entry_freeList);
}
/*
  *
  *	Initialize the Trigger Port system
  *
  *	Allocate memory and initialize all data structures.
  */
static int32 _rtl8651_TriggerPortInit(rtl8651_fwdEngineInitPara_t *para)
{
	#ifdef TRIGGER_PORT_DEBUG
	printk("Init trigger port system.\n");
	#endif

	/* clear ALL data structures */
	triggerPort_entry_tbl = NULL;
	triggerPort_entry_freeList = NULL;
	BZERO_ARRAY(triggerPort_entry_Hdr);

	triggerPort_cache_tbl = NULL;
	triggerPort_cache_freeList = NULL;
	triggerPort_cache_staticEntry = NULL;
	triggerPort_cache_lruHdr = NULL;
	triggerPort_cache_lruTail = NULL;
	BZERO_ARRAY(triggerPort_cache_hashTbl);

	triggerPort_entry_cnt = para->triggerPortEntryCnt;
	triggerPort_cache_cnt = para->triggerPortCacheSize;

	if ( (triggerPort_entry_cnt == 0) || (triggerPort_cache_cnt == 0) ) /* table disabled */
	{
		#ifdef TRIGGER_PORT_DEBUG
		printk("Init trigger port system DONE ... Module disabled.\n");
		#endif

		return SUCCESS;
	}

	/* allocate trigger port entry */

	MEM_ALLOCATE_PTR(triggerPort_entry_tbl, triggerPort_entry_cnt * sizeof(_rtl8651_triggerPort_entry_t));

	if ( triggerPort_entry_tbl == NULL ) /* memory allocation fail */
	{

		#ifdef TRIGGER_PORT_DEBUG
		printk("Init trigger port system FAILED.\n");
		#endif

		return FAILED;
	}

	/* allocate trigger port cache */

	MEM_ALLOCATE_PTR(triggerPort_cache_tbl, triggerPort_cache_cnt * sizeof(_rtl8651_triggerPort_cache_t));

	if ( triggerPort_cache_tbl == NULL ) /* memory allocation fail */
	{

		#ifdef TRIGGER_PORT_DEBUG
		printk("Init trigger port system FAILED.\n");
		#endif

		return FAILED;
	}

	_rtl8651_TriggerPortReinit(); /* init data structures */

	#ifdef TRIGGER_PORT_DEBUG
	printk("Init trigger port system OK.\n");
	#endif

	return SUCCESS;
}

/*
  *
  *	Re-initialize the Trigger Port system
  *
  *	re-initialize all data structures.
  */
static void _rtl8651_TriggerPortReinit(void)
{
	_rtl8651_triggerPort_entry_t *entry;
	_rtl8651_triggerPort_cache_t *cache;
	uint32 idx;

	#ifdef TRIGGER_PORT_DEBUG
	printk("Re-init trigger port system.\n");
	#endif

	if ( (triggerPort_entry_tbl == NULL) || (triggerPort_cache_tbl == NULL) ) /* table disabled */
	{
		#ifdef TRIGGER_PORT_DEBUG
		printk("Re-init trigger port system DONE.\n");
		#endif

		return;
	}

	BZERO_ARRAY(triggerPort_entry_Hdr);			/* trigger port table */
	BZERO_ARRAY(triggerPort_cache_hashTbl);		/* dynamic cache */
	triggerPort_cache_staticEntry = NULL;				/* static cache */
	triggerPort_cache_lruHdr = NULL;					/* lru header */
	triggerPort_cache_lruTail = NULL;					/* lru tail */
	BZERO_PTR(triggerPort_entry_tbl, triggerPort_entry_cnt * sizeof(_rtl8651_triggerPort_entry_t));
	BZERO_PTR(triggerPort_cache_tbl, triggerPort_cache_cnt * sizeof(_rtl8651_triggerPort_cache_t));

	/* <-------------------------- Entry --------------------------> */
	entry = triggerPort_entry_freeList = triggerPort_entry_tbl;
	for ( idx=0 ; idx<(triggerPort_entry_cnt-1) ; idx++,entry++ )
	{
		entry->next = entry + 1;
	}

	/* <-------------------------- Cache --------------------------> */
	cache = triggerPort_cache_freeList = triggerPort_cache_tbl;
	for ( idx=0 ; idx<(triggerPort_cache_cnt-1) ; idx++, cache++ )
	{
		cache->next = cache + 1;
	}

	#ifdef TRIGGER_PORT_DEBUG
	printk("Re-init trigger port system DONE.\n");
	#endif

}

/*
@func void		| rtl8651_flushTriggerPortRules	| Flush all the records from Trigger Port Rule Table.
@comm 
This function will flush all the records in trigger port table.
@xref  <p rtl8651_addTriggerPortRule>
*/

void rtl8651_flushTriggerPortRules(void)
{
	rtlglue_drvMutexLock();
	_rtl8651_flushNetTriggerPortRules(0);
	rtlglue_drvMutexUnlock();
}

/*
@func void		| rtl8651a_flushNetTriggerPortRules	| Flush all the records for specific network interface from Trigger Port Rule Table.
@parm uint32		| netIntfId						| Network Interface ID, ROMEDRV treat it same as DSID now.
@comm 
This function will flush all the records related to the network interface ID in trigger port table.
@xref  <p rtl8651_addTriggerPortRule>
*/
void rtl8651a_flushNetTriggerPortRules(uint32 netIntfId)
{
	rtlglue_drvMutexLock();
	_rtl8651_flushNetTriggerPortRules(netIntfId);
	rtlglue_drvMutexUnlock();
}

static void _rtl8651_flushNetTriggerPortRules(uint32 netIntfId)
{
	uint32 index;
	_rtl8651_triggerPort_cache_t *cache, *next;

	#ifdef TRIGGER_PORT_DEBUG
	printk("Flush Trigger port rule for Network Interface (%d).\n", netIntfId);
	#endif

	/* Free trigger port rules and normal cache according to them */
	for ( index=0 ; index<RTL8651_MAX_DIALSESSION ; index++ )
	{
		_rtl8651_triggerPort_entry_t *entry;
		_rtl8651_triggerPort_entry_t *next;

		entry = triggerPort_entry_Hdr[index];

		while ( entry )
		{
			next = entry->next;

			if ( entry->netIntfId == netIntfId )
			{
				_rtl8651_TriggerPort_FreeEntry(entry);
			}
			entry = next;
		}
	}

	/* Free static trigger port cache */
	cache = triggerPort_cache_staticEntry;
	while ( cache )
	{
		next = cache->next;

		if ( cache->netIntfId == netIntfId )
		{
			#ifdef TRIGGER_PORT_DEBUG
			printk("Remove Trigger Port flow STATIC cache:\n\tNetwork Interface %d\n\tType: %s\n\tPort: %d ~ %d\n\tLocal IP Addr: %u.%u.%u.%u",
						cache->netIntfId,
						(cache->inType==0)?"TCP":(cache->inType==1)?"UDP":(cache->inType==2)?"BOTH":"ERROR !",
						cache->inPortStart,
						cache->inPortEnd,
						NIPQUAD(cache->localIpAddr));
			#endif

			_rtl8651_TriggerPort_FreeCache(TRUE, cache);
		}
		cache = next;
	}

}

/*
@func int32						|	rtl8651_addTriggerPortMappingRule	| General API to add one record into Tigger Port Table.
@parm uint32						|	netIntfId			| the network interface ID, ROMEDRV treat it same as DSID now.
@parm uint8						|	inType			| the protocol type of the incoming packets, you can assign 0:TCP 1:UDP 2:BOTH.
@parm uint16						|	inPortStart		| the port range of the incoming packets.
@parm uint16						|	inPortEnd			| the port range of the incoming packets.
@parm _rtl8651_triggerPort_param_t*	|	mappingInfo		| Further setting for STATIC and DYNAMIC trigger port rules / entries.
@rvalue SUCCESS					| 	Add one record into the table successfully.
@rvalue FAILED					| 	Add failed, maybe the table is full.
@comm 
The format of "mappingInfo" is as below:

	typedef struct _rtl8651_triggerPort_param_s
	{
		uint8 isStaticMapping;
		uint32 age;

		union {
			struct {
				uint8		__outType;
				ipaddr_t		__triggeringInsideLocalIpAddrStart;
				ipaddr_t		__triggeringInsideLocalIpAddrEnd;
				uint16		__outPortStart;
				uint16		__outPortEnd;
			} DYNAMIC_MAPPING;
			struct {
				ipaddr_t		__mappedInsideLocalIpAddr;
				uint8		__isStatic;
			} STATIC_MAPPING;
		} UN_MAPPING;

		#define	tp_outType							UN_MAPPING.DYNAMIC_MAPPING.__outType
		#define	tp_triggeringInsideLocalIpAddrStart		UN_MAPPING.DYNAMIC_MAPPING.__triggeringInsideLocalIpAddrStart
		#define	tp_triggeringInsideLocalIpAddrEnd		UN_MAPPING.DYNAMIC_MAPPING.__triggeringInsideLocalIpAddrEnd
		#define	tp_outPortStart						UN_MAPPING.DYNAMIC_MAPPING.__outPortStart
		#define	tp_outPortEnd							UN_MAPPING.DYNAMIC_MAPPING.__outPortEnd

		#define	tp_mappedInsideLocalIpAddr				UN_MAPPING.STATIC_MAPPING.__mappedInsideLocalIpAddr
		#define	tp_isStatic							UN_MAPPING.STATIC_MAPPING.__isStatic

	}_rtl8651_triggerPort_param_t;


	And 2 kinds of rule can be added into trigger port table:

	1. STATIC entry	(<p isStaticMapping> == TRUE)

		This is similar as what user call <p rtl8651_addTriggerPortRule> or <p rtl8651a_addTriggerPortRule> WITH 'localIpAddr' set.
		For this rule, the mapped Inside Local IP Address is given.
		All new inbound NAPT flows with {netIntfId, inType, inPortStart, inPortEnd} would match this rule and redirected to the mapped
		inside local address.

		The extra parameter for the static entry is <p isStatic>
		If <p isStatic> == TRUE, this STATIC entry would NEVER been replaced by LRU replacement mechanism.
		(BUT it still can be replaced by other STATIC entry addition function CALL).

	2. DYNAMIC entry	(<p isStaticMapping> == FALSE)

		This is similar as what user call <p rtl8651_addTriggerPortRule> or <p rtl8651a_addTriggerPortRule> WITHOUT 'localIpAddr' set.
		And the inbound port mapping would be triggered only when outbound traffic triggering it.
		In DYNAMIC entry, we can limit the triggering flow with outType, ourtPortStart-outPortEnd, sourceIPStart-sourceIPEnd.
		(If both <p tp_triggeringInsideLocalIpAddrStart> and <p tp_triggeringInsideLocalIpAddrEnd> are '0', trigger port module would ignore source check)
		NOTE that due to these kinds of port mapping is triggered dynamically, user CAN NOT keep the mapping entry from LRU replacement
		mechanism. Therefore, <p isStatic> setting in STATIC entry is not given.

	===========================================================================================

	AGE timer:

		ROMEDRV provide age timer for both STATIC and DYNAMIC entries and the unit of age is SECOND.
		If user set the <p age> of "_rtl8651_triggerPort_param_t" to '0', it means this entry will never be timed out and only
		can be removed by:

			1. Other triggering flow (for DYNAMIC entry only)
			2. Trigger Port Table flush

NOTE:
		ROMEDRV treat the 'netIntfId' as the dsid yet.
		In non-multi-session case, it is '0'. And user can add trigger port rule with calling <p rtl8651_addTriggerPortRule> .

@xref  <p rtl8651_flushTriggerPortRules>
@xref  <p rtl8651a_flushNetTriggerPortRules>
@xrec <p rtl8651a_addTriggerPortRule>
@xrec <p rtl8651_addTriggerPortRule>
*/
int32 rtl8651_addTriggerPortMappingRule (
	uint32 netIntfId,
	uint8 inType,
	uint16 inPortStart,
	uint16 inPortEnd,
	_rtl8651_triggerPort_param_t *mappingInfo)
{
	int32 retval;

	if (	(netIntfId >= RTL8651_MAX_DIALSESSION) ||
		(inPortStart > inPortEnd) ||
		(mappingInfo == NULL))
	{
		return FAILED;
	}
	rtlglue_drvMutexLock();
	retval = _rtl8651_addTriggerPortRule(	netIntfId,
										inType,
										inPortStart,
										inPortEnd,
										mappingInfo);
	rtlglue_drvMutexUnlock();
	return retval;
}

/*
@func int32		|	rtl8651_addTriggerPortRule	| Add one record into Tigger Port Rule Table.
@parm uint8		|	inType | the protocol type of the incoming packets, you can assign 0:TCP 1:UDP 2:BOTH.
@parm uint16		|	inPortStart |  the port range of the incoming packets.
@parm uint16		|	inPortEnd |  the port range of the incoming packets.
@parm uint8		|	outType | the protocol type of the outgoing(trigger) packets, you can assign 0:TCP 1:UDP 2:BOTH.
@parm uint16		|	outPortStart | the port range of the outgoing(trigger) packets.
@parm uint16		|	outPortEnd | the port range of the outgoing(trigger) packets.
@parm ipaddr_t	|	localIpAddr | the IP address of port forwarding, this is a LAN IP address, the packets which the protocol are conform to inType and port are between inPortStart and inPortFinish will forward to this IP address.
@rvalue SUCCESS	| 	Add one record into the table successfully.
@rvalue FAILED	| 	Add failed, maybe the table is full.
@comm 
If you want to add a trigger port rule, set the localIpAddr to zero.
If you want to add a port forwarding rule, set the outType, outPortStart, outPortFinish to zero.
"localIpAddr" is used for port forwarding, It is still be timed out, too. However, if this rule is set, the port forwarding function
can be used for all WAN IP Addresses of the specific network Interface "0".

NOTE:
		ROMEDRV treat the 'netIntfId' as the dsid yet.
		In non-multi-session case, it is '0'. And user can add trigger port rule with calling <p rtl8651_addTriggerPortRule> .

Example:
If an application's trigger ports are 30000-30020(TCP), and it's incoming ports are 3000-4000(TCP),5000(UDP),6000-7000(TCP and UDP), you can call functions as following.
rtl8651_flushTriggerPortRules();
rtl8651_addTriggerPortRule(0,3000,4000,0,30000,30020,0);
rtl8651_addTriggerPortRule(1,5000,5000,0,30000,30020,0);
rtl8651_addTriggerPortRule(2,6000,7000,0,30000,30020,0);
@xref  <p rtl8651_flushTriggerPortRules>
@xref  <p rtl8651a_flushNetTriggerPortRules>
@xrec <p rtl8651a_addTriggerPortRule>
*/
int32 rtl8651_addTriggerPortRule(
	uint8 inType, uint16 inPortStart, uint16 inPortEnd,
	uint8 outType,uint16 outPortStart,uint16 outPortEnd,
	ipaddr_t localIpAddr)
{
	return rtl8651a_addTriggerPortRule(	0,
										inType,
										inPortStart,
										inPortEnd,
										outType,
										outPortStart,
										outPortEnd,
										localIpAddr);
}
/*
@func int32		| rtl8651a_addTriggerPortRule	| Add one record into Tigger Port Rule Table for a specific Network Interface.
@parm uint32		|	netIntfId | the network interface ID, ROMEDRV treat it same as DSID now.
@parm uint8		| inType | the protocol type of the incoming packets, you can assign 0:TCP 1:UDP 2:BOTH.
@parm uint16		| inPortStart |  the port range of the incoming packets.
@parm uint16		| inPortEnd |  the port range of the incoming packets.
@parm uint8		| outType | the protocol type of the outgoing(trigger) packets, you can assign 0:TCP 1:UDP 2:BOTH.
@parm uint16		| outPortStart | the port range of the outgoing(trigger) packets.
@parm uint16		| outPortEnd | the port range of the outgoing(trigger) packets.
@parm ipaddr_t	| localIpAddr | the IP address of port forwarding, this is a LAN IP address, the packets which the protocol are conform to inType and port are between inPortStart and inPortFinish will forward to this IP address.
@rvalue SUCCESS	| 	Add one record into the table successfully.
@rvalue FAILED	| 	Add failed, maybe the table is full.
@comm 
If you want to add a trigger port rule, set the localIpAddr to zero.
If you want to add a port forwarding rule, set the outType, outPortStart, outPortFinish to zero.
"localIpAddr" is used for port forwarding, It is still be timed out, too. However, if this rule is set, the port forwarding function
can be use for all WAN IP Addresses of the specific network Interface defined by "netIntfId".

NOTE:
		ROMEDRV treat the 'netIntfId' as the dsid yet.
		In non-multi-session case, it is '0'. And user can add trigger port rule with calling <p rtl8651_addTriggerPortRule>.
		In multi-session case, user can add trigger port rule to specific interface using <p rtl8651a_addTriggerPortRule>.

Example:
For network interface '0', an application's trigger ports are 30000-30020(TCP), and it's incoming ports are 3000-4000(TCP),5000(UDP),6000-7000(TCP and UDP), you can call functions as following.
rtl8651_flushTriggerPortRules();
rtl8651a_addTriggerPortRule(0,0,3000,4000,0,30000,30020,0);
rtl8651a_addTriggerPortRule(0,1,5000,5000,0,30000,30020,0);
rtl8651a_addTriggerPortRule(0,2,6000,7000,0,30000,30020,0);
@xref  <p rtl8651_flushTriggerPortRules>
@xref  <p rtl8651a_flushNetTriggerPortRules>
@xrec <p rtl8651_addTriggerPortRule>
*/
int32 rtl8651a_addTriggerPortRule(
	uint32 netIntfId,
	uint8 inType, uint16 inPortStart, uint16 inPortEnd,
	uint8 outType,uint16 outPortStart,uint16 outPortEnd,
	ipaddr_t localIpAddr)
{
	int32 retval;
	_rtl8651_triggerPort_param_t mapInfo;


	if (localIpAddr)
	{	/* static */
		mapInfo.isStaticMapping = TRUE;
		mapInfo.age = 0;					/* We don't age out added entry. */

		mapInfo.tp_isStatic = TRUE;
		mapInfo.tp_mappedInsideLocalIpAddr = localIpAddr;
	} else
	{	/* dynamic */
		mapInfo.isStaticMapping = FALSE;
		mapInfo.age = RTL8651_TRIGGER_PORT_DEFAULT_TIMEOUT;	/* Default value of dynamic entry */

		mapInfo.tp_outType = outType;
		mapInfo.tp_outPortStart = outPortStart;
		mapInfo.tp_outPortEnd = outPortEnd;
		mapInfo.tp_triggeringInsideLocalIpAddrStart = 0;
		mapInfo.tp_triggeringInsideLocalIpAddrEnd = 0;
	}

	rtlglue_drvMutexLock();
	retval = _rtl8651_addTriggerPortRule(	netIntfId,
										inType,
										inPortStart,
										inPortEnd,
										&mapInfo);
	rtlglue_drvMutexUnlock();
	return retval;
}

/* ============================================================================
	Internal function to add trigger port rule.

		Including STATIC rules and DYNAMIC rules.
     ============================================================================ */
static int32 _rtl8651_addTriggerPortRule(
	uint32 netIntfId,
	uint8 inType, uint16 inPortStart, uint16 inPortEnd,
	_rtl8651_triggerPort_param_t *mappingInfo)

{
	assert(mappingInfo);
	assert(inPortStart <= inPortEnd);

	#ifdef TRIGGER_PORT_DEBUG
	if (mappingInfo->isStaticMapping == TRUE)
	{
		printk("Add STATIC Trigger port entry:\n\tNetIntfId[%d]\n\tinType[%d], inPort[%d - %d]\n\tAge: %d\n\tNO-LRU: %s\n",
			netIntfId,
			inType, inPortStart, inPortEnd,
			mappingInfo->age,
			mappingInfo->tp_isStatic==TRUE)?"YES":"NO";
	} else
	{
		printk("Add DYNAMIC Trigger port entry:\n\tNetIntfId[%d]\n\tinType[%d], inPort[%d - %d]\n\toutType[%d], outPort[%d - %d]\n\tTriggering IP[%u.%u.%u.%u - %u.%u.%u.%u]\n\tAge: %d\n",
			netIntfId,
			inType, inPortStart, inPortEnd,
			mappingInfo->tp_outType, mappingInfo->tp_outPortStart, mappingInfo->tp_outPortEnd,
			NIPQUAD(mappingInfo->tp_triggeringInsideLocalIpAddrStart), NIPQUAD(mappingInfo->tp_triggeringInsideLocalIpAddrEnd),
			mappingInfo->age);
	}
	#endif

	if ( mappingInfo->isStaticMapping == TRUE )
	{	/* static cache entry add */
		_rtl8651_triggerPort_cache_t *cache;

		/* find exist static cache */
		cache = triggerPort_cache_staticEntry;

		while ( cache )
		{
			if ( (cache->netIntfId==netIntfId) &&
			      (cache->inPortStart==inPortStart) &&
			      (cache->inPortEnd==inPortEnd))
			{
				if ( cache->inType==inType )
				{ /* update localIpAddr */
					goto updateCache;
				} else if ((cache->inType==2)||(inType==2))
				{ /* one of they is set to BOTH:conflict */
					goto errout;
				}
			}

			if ( (!(cache->inPortStart>inPortEnd||cache->inPortEnd<inPortStart)) &&
			      (cache->inType==inType||cache->inType==2||inType==2))
			{ /* port conflict */
				goto errout;
			}
			
			cache = cache->next;
		}

		/* add new static entry */
		if ( (cache = _rtl8651_TriggerPort_getCache()) == NULL )
		{
			goto errout;
		}

		_NORMAL_LINK(triggerPort_cache_staticEntry, cache);

		cache->ExternalIpAddr	= 0;
		cache->netIntfId		= netIntfId;
		cache->inPortStart		= inPortStart;
		cache->inPortEnd		= inPortEnd;
		cache->inType		= inType;
updateCache:
		cache->isStatic		= mappingInfo->tp_isStatic;
		cache->localIpAddr	= mappingInfo->tp_mappedInsideLocalIpAddr;
		cache->age			= mappingInfo->age;
		cache->maxAge		= mappingInfo->age;

		/* we always remove cache from LRU list, and add back to tail if it is not the static entry */
		HTLIST_REMOVE(cache, lru_next, lru_prev, triggerPort_cache_lruHdr, triggerPort_cache_lruTail);
		if (mappingInfo->tp_isStatic == FALSE)
		{
			HTLIST_ADD2TAIL(cache, lru_next, lru_prev, triggerPort_cache_lruHdr, triggerPort_cache_lruTail);
		}
	} else
	{	/* trigger port rule add */
		_rtl8651_triggerPort_entry_t *entry;

		if (	(mappingInfo->tp_outPortStart > mappingInfo->tp_outPortEnd) ||
			(mappingInfo->tp_triggeringInsideLocalIpAddrStart > mappingInfo->tp_triggeringInsideLocalIpAddrEnd))
		{
			goto errout;
		}

		/* find exist entry */
		entry = triggerPort_entry_Hdr[netIntfId];

		while ( entry )
		{
			if (	(entry->netIntfId==netIntfId) &&
				(entry->outPortStart==mappingInfo->tp_outPortStart) &&
				(entry->outPortEnd==mappingInfo->tp_outPortEnd) &&
				(entry->outType==mappingInfo->tp_outType) &&
				(entry->triggeringInsideLocalIpAddrStart==mappingInfo->tp_triggeringInsideLocalIpAddrStart) &&
				(entry->triggeringInsideLocalIpAddrEnd==mappingInfo->tp_triggeringInsideLocalIpAddrEnd) &&
				(entry->inPortStart==inPortStart) &&
				(entry->inPortEnd==inPortEnd))
			{ /* exist entry found! */
				if (entry->inType != inType)
				{ /* Of course, new type is BOTH */
					entry->inType = 2;
				}
				goto out;
			}
			entry = entry->next;
		}

		/* allocate new entry */
		_GET_FROM_FREELIST(entry, triggerPort_entry_freeList);
		if ( entry==NULL )
		{
			goto errout;
		}

		entry->netIntfId						= netIntfId;
		entry->outType						= mappingInfo->tp_outType;
		entry->outPortStart					= mappingInfo->tp_outPortStart;
		entry->outPortEnd						= mappingInfo->tp_outPortEnd;
		entry->triggeringInsideLocalIpAddrStart	= mappingInfo->tp_triggeringInsideLocalIpAddrStart;
		entry->triggeringInsideLocalIpAddrEnd	= mappingInfo->tp_triggeringInsideLocalIpAddrEnd;
		entry->inType						= inType;
		entry->inPortStart						= inPortStart;
		entry->inPortEnd						= inPortEnd;
		entry->cacheAge						= mappingInfo->age;
		entry->normalCache					= NULL;

		_NORMAL_LINK(triggerPort_entry_Hdr[netIntfId], entry);
	}
out:
	#ifdef TRIGGER_PORT_DEBUG
	printk("\t===> OK\n");
	#endif

	return SUCCESS;
errout:

	#ifdef TRIGGER_PORT_DEBUG
	printk("\t===> FAILED\n");
	#endif

	return FAILED;
}

static inline int32 _rtl8651_checkUsableTriggeredPortEntry(uint32 isTcp, uint32 netIntfId, ipaddr_t intIp, uint32 extPort)
{
	_rtl8651_triggerPort_cache_t *cache, *candidate = NULL;
	uint32 hashTblIdx;

	#ifdef TRIGGER_PORT_DEBUG
	printk("Find Entry:\n\tisTcp: %s\n\tNetwork Interface: %d\n\tinternal IP: %u.%u.%u.%u\n\textPort: %d\n",
				isTcp?"TRUE":"FALSE",
				netIntfId,
				NIPQUAD(intIp),
				extPort);
	#endif

	for ( hashTblIdx = 0 ; hashTblIdx < RTL8651_TRIGGER_PORT_HASH_CACHE_TBLSIZE ; hashTblIdx ++ )
	{
		cache = triggerPort_cache_hashTbl[hashTblIdx];

		while ( cache )
		{
			if (	(cache->netIntfId == netIntfId) &&
				(cache->inPortStart<=extPort) &&
				(cache->inPortEnd>=extPort) &&
				(cache->inType==2 || (isTcp&&cache->inType==0) || ((!isTcp)&&cache->inType==1)))
			{
				if (cache->localIpAddr==intIp)
				{
					#ifdef TRIGGER_PORT_DEBUG
					printk("\t==> FOUND!\n");
					#endif
					return TRUE;
				}
				candidate = cache;
			}

			cache = cache->next;
		}
	}

	if (candidate)
	{
		#ifdef TRIGGER_PORT_DEBUG
		printk("\t==> Triggered by other host: %u.%u.%u.%u !\n", NIPQUAD(candidate->localIpAddr));
		#endif
		return FALSE;
	}

	#ifdef TRIGGER_PORT_DEBUG
	printk("\t==> NO TRIGGER PORT ENTRY ABOUT THIS EXT PORT !\n");
	#endif

	return TRUE;

}

/*
  *
  *	Get exists trigger port cache for the specific {isTcp, network interface ID, external IP, external Port}.
  *   Note that the static cache is checked, too.
  *
  */
static inline _rtl8651_triggerPort_cache_t *_rtl8651_getTriggerPortCache(uint8 refreshAge, uint8 isTcp, uint32 netIntfId, ipaddr_t extIp, uint16 extPort)
{
	_rtl8651_triggerPort_cache_t *cache;
	uint32 hash = TWO_ITEM_EASYHASH(netIntfId, extIp, RTL8651_TRIGGER_PORT_HASH_CACHE_TBLSIZE);

	#ifdef TRIGGER_PORT_DEBUG
	printk("Get Cache Entry:\n\tisTcp: %s\n\tNetwork Interface: %d\n\tExt IP: %u.%u.%u.%u\n\tExt Port: %d\n",
				isTcp?"TRUE":"FALSE",
				netIntfId,
				NIPQUAD(extIp),
				extPort);
	#endif

	/* find normal cache ==> assumption: NORMAL cache has higher priority than STATIC cache */
	cache = triggerPort_cache_hashTbl[hash];
	while ( cache )
	{
		if ( (cache->netIntfId==netIntfId) &&
		      (cache->ExternalIpAddr==extIp) &&
		      (cache->inPortStart<=extPort) &&
		      (cache->inPortEnd>=extPort) &&
		      (cache->inType==2||(isTcp&&cache->inType==0)||(!isTcp&&cache->inType==1)))
		{
			goto found;
		}
		cache = cache->next;
	}
	
	/* find static cache */
	cache = triggerPort_cache_staticEntry;
	while ( cache )
	{
		if ( (cache->netIntfId==netIntfId) &&
		      (cache->inPortStart<=extPort) &&
		      (cache->inPortEnd>=extPort) &&
		      (cache->inType==2||(isTcp&&cache->inType==0)||(!isTcp&&cache->inType==1)))
		{
			goto found;
		}
		cache = cache->next;
	}

	#ifdef TRIGGER_PORT_DEBUG
	printk("\t==> NO CACHE FOUND !\n");
	#endif

	return NULL;

found:
	#ifdef TRIGGER_PORT_DEBUG
	printk("\t==> WE GOT IT !\n");
	#endif

	if (refreshAge == TRUE)
	{
		/* Update LRU */
		if (cache->isStatic == FALSE)
		{
			HTLIST_MOVE2TAIL(cache, lru_next, lru_prev, triggerPort_cache_lruHdr, triggerPort_cache_lruTail);
		}

		cache->age = cache->maxAge;
	}

	return cache;

}

/*
  *
  *	Add the trigger port entry for the corresponding
  *
  *
  *	NOTE: this function is called to add the DYNAMIC CACHE only.
  *
  */
static inline void _rtl8651_addTriggerPortCache(_rtl8651_triggerPort_entry_t *entry, ipaddr_t extIp, ipaddr_t localIpAddr)
{
	_rtl8651_triggerPort_cache_t *cache;
	uint32 hash = TWO_ITEM_EASYHASH((entry->netIntfId), extIp, RTL8651_TRIGGER_PORT_HASH_CACHE_TBLSIZE);

	#ifdef TRIGGER_PORT_DEBUG
	printk("Add Cache Entry:\n\tType: %s\n\tNetwork Interface: %d\n\tExt IP: %u.%u.%u.%u\n\tLocal IP Addr: %u.%u.%u.%u\n",
				(entry->inType==0)?"TCP":(entry->inType==1)?"UDP":(entry->inType==2)?"BOTH":"ERROR !",
				entry->netIntfId,
				NIPQUAD(extIp),
				NIPQUAD(localIpAddr));
	#endif

	/* find exist normal cache */
	cache = entry->normalCache;
	while ( cache )
	{
		if ( (cache->netIntfId==entry->netIntfId) &&
		      (cache->ExternalIpAddr==extIp) &&
		      (cache->inPortStart==entry->inPortStart) &&
		      (cache->inPortEnd>=entry->inPortEnd) &&
		      (cache->inType==entry->inType||cache->inType==2||entry->inType==2))
		{	/* update Exist entry */

			#ifdef TRIGGER_PORT_DEBUG
			printk("\t==> UPDATE EXIST\n");
			#endif

			goto updateCache;
		}
		cache = cache->entryList_next;
	}

	/* allocate new cache */
	if ( (cache = _rtl8651_TriggerPort_getCache()) == NULL )
	{	/* new cache allocation error */

		#ifdef TRIGGER_PORT_DEBUG
		printk("\t==> NO ENTRY\n");
		#endif

		return;
	}

	#ifdef TRIGGER_PORT_DEBUG
	printk("\t==> ADD NEW ENTRY\n");
	#endif

	_NORMAL_LINK(triggerPort_cache_hashTbl[hash], cache);
	_TRIGGERPORT_ENTRY_CACHE_LINK(entry, cache);

	cache->ExternalIpAddr = extIp;
	cache->netIntfId = entry->netIntfId;
	cache->inPortStart = entry->inPortStart;
	cache->inPortEnd = entry->inPortEnd;
	cache->inType = entry->inType;
	cache->isStatic = FALSE;					/* For dynamically added caches, it MUST can be removed by LRU mechanism */
updateCache:
	cache->localIpAddr = localIpAddr;
	cache->maxAge = cache->age = entry->cacheAge;

	/* this MACRO would remove from LRU list (if this cache is newly added, it would do nothing) and add to the tail of LRU list */
	HTLIST_MOVE2TAIL(cache, lru_next, lru_prev, triggerPort_cache_lruHdr, triggerPort_cache_lruTail);

	#ifdef TRIGGER_PORT_DEBUG
	printk("\t==> Local IP: %u.%u.%u.%u AGE: %d\n", NIPQUAD(localIpAddr), cache->age);
	#endif

}

/*
  *
  *	For outbound traffic, check the trigger port entries and try to trigger these trigger ports.
  *	If a trigger port entry is triggered, it will create an new trigger port cache for this activated entry.
  *
  *	This function would return the added count of new cache.
  *
  *	Note: that one outbound traffic might can trigger several trigger port entries.
  *
  */
static inline int32 _rtl8651_tryActiveTriggerPort(uint8 isTcp, uint32 netIntfId, uint16 outPort, ipaddr_t extIp, ipaddr_t localIpAddr)
{
	_rtl8651_triggerPort_entry_t *entry;

	#ifdef TRIGGER_PORT_DEBUG
	printk("Try to trigger:\n\tisTcp: %s\n\tNetwork Interface: %d\n\tOutPort: %d\n\tExternal IP: %u.%u.%u.%u\n\tLocal IP Addr: %u.%u.%u.%u\n",
			(isTcp)?"TRUE":"FALSE",
			netIntfId,
			outPort,
			NIPQUAD(extIp),
			NIPQUAD(localIpAddr));
	#endif

	entry = triggerPort_entry_Hdr[netIntfId];

	while ( entry )
	{
		#ifdef TRIGGER_PORT_DEBUG
		printk("\tEntry:\n");
		printk("\t\tNetwork Interface ID [%d] <--> %d\n", entry->netIntfId, netIntfId);
		printk("\t\tOutPort [%d - %d] <--> %d\n", entry->outPortStart, entry->outPortEnd, outPort);
		printk("\t\tSIP [%u.%u.%u.%u - %u.%u.%u.%u] <--> %u.%u.%u.%u\n", NIPQUAD(entry->triggeringInsideLocalIpAddrStart), NIPQUAD(entry->triggeringInsideLocalIpAddrEnd), NIPQUAD(localIpAddr));
		printk("\t\tType [%s] <--> %s\n",	(entry->outType==0)?"TCP":(entry->outType==1)?"UDP":(entry->outType==2)?"BOTH":"ERROR !",
										(isTcp==TRUE)?"TCP":"UDP");
		
		#endif


		if (	(entry->outPortStart<=outPort) &&
			(entry->outPortEnd>=outPort) &&
			(	(entry->triggeringInsideLocalIpAddrStart==0 && entry->triggeringInsideLocalIpAddrEnd==0)||
				(entry->triggeringInsideLocalIpAddrStart<=localIpAddr&&entry->triggeringInsideLocalIpAddrEnd>=localIpAddr)) &&
			((isTcp&&entry->outType==0)||((!isTcp)&&entry->outType==1)||(entry->outType==2)) &&
			(entry->netIntfId==netIntfId) )
		{

			#ifdef TRIGGER_PORT_DEBUG
			printk("\t==> Trigger Port Entry FOUND ! Trigger it!\n");
			#endif


			_rtl8651_addTriggerPortCache(entry, extIp, localIpAddr);
			return 1;
		}
		entry = entry->next;		
	}

	#ifdef TRIGGER_PORT_DEBUG
	printk("\t==> NO Trigger Port Entry for it!\n");
	#endif

	return 0;
}

/*
  *
  *	For inbound traffic, check the trigger port caches and get the insideLocal IP Address.
  *	
  */
static inline int32 _rtl8651_getTriggeredIpAddr(uint8 isTcp, uint32 netIntfId, ipaddr_t extIp, uint16 inPort, ipaddr_t *localIpAddrp)
{
	_rtl8651_triggerPort_cache_t *cache;

	#ifdef TRIGGER_PORT_DEBUG
	printk("Get to trigger cache:\n\tisTcp: %s\n\tNetwork Interface: %d\n\tExternal IP: %u.%u.%u.%u\n\tIncoming Port %d\n",
			isTcp?"TRUE":"FALSE",
			netIntfId,
			NIPQUAD(extIp),
			inPort);
	#endif

	cache = _rtl8651_getTriggerPortCache(TRUE, isTcp, netIntfId, extIp, inPort);
	if ( cache )
	{

		#ifdef TRIGGER_PORT_DEBUG
		printk("\t==> GET!! Local IP Addr: %u.%u.%u.%u\n", NIPQUAD(cache->localIpAddr));
		#endif

		*localIpAddrp = cache->localIpAddr;
		return SUCCESS;
	}
	return FAILED;
}

/*
  *
  *	Time update function of trigger port system, it time-out the trigger port caches which have phased out.
  *	
  */
static void _rtl8651_TriggerPortTimeUpdate(uint32 secPassed)
{
	uint32 idx;
	_rtl8651_triggerPort_cache_t *cache;
	_rtl8651_triggerPort_cache_t *next;

	/* update normal cache */
	for ( idx=0;idx<RTL8651_TRIGGER_PORT_HASH_CACHE_TBLSIZE;idx++ )
	{
		cache = triggerPort_cache_hashTbl[idx];
		while( cache )
		{
			next = cache->next;

			if (cache->age == 0)
			{
				/* DO NOTHING: never timed-out */

			} else if (cache->age>secPassed )
			{
				cache->age -= secPassed;
			}else
			{
				#ifdef TRIGGER_PORT_DEBUG
				printk("Trigger Port flow DYNAMIC cache timeout:\n\tNetwork Interface %d\n\tType: %s\n\tPort: %d ~ %d\n\tLocal IP Addr: %u.%u.%u.%u\n",
							cache->netIntfId,
							(cache->inType==0)?"TCP":(cache->inType==1)?"UDP":(cache->inType==2)?"BOTH":"ERROR TYPE!",
							cache->inPortStart,
							cache->inPortEnd,
							NIPQUAD(cache->localIpAddr));
				#endif

				_rtl8651_TriggerPort_FreeCache(FALSE, cache);
			}
			cache = next;
		}
	}
	/* update static cache */
	cache = triggerPort_cache_staticEntry;
	while ( cache )
	{
		next = cache->next;
		if (cache->age == 0)
		{
				/* DO NOTHING: never timed-out */

		} else if ( cache->age>secPassed )
		{
			cache->age -= secPassed;
		}else
		{
			#ifdef TRIGGER_PORT_DEBUG
			printk("Trigger Port flow STATIC cache timeout:\n\tNetwork Interface %d\n\tType: %s\n\tPort: %d ~ %d\n\tLocal IP Addr: %u.%u.%u.%u\n",
						cache->netIntfId,
						(cache->inType==0)?"TCP":(cache->inType==1)?"UDP":(cache->inType==2)?"BOTH":"ERROR !",
						cache->inPortStart,
						cache->inPortEnd,
						NIPQUAD(cache->localIpAddr));
			#endif

			_rtl8651_TriggerPort_FreeCache(TRUE, cache);
		}
		cache = next;
	}
}
#endif /* _RTL_TRIGGER_PORT */

#ifdef _RTL_QUEUE_ARP_PACKET
#define RTL8651_ARP_QUEUE_FUNCTIONS

static int32 _rtl8651_queueInit(void){
	queue_packet_tb = (struct _RTL8651_QUEUE_PACKET_ENTRY *)rtlglue_malloc(QUEUE_PACKET_ITEM * sizeof (struct _RTL8651_QUEUE_PACKET_ENTRY));
	memset(queue_packet_tb,0,QUEUE_PACKET_ITEM * sizeof (struct _RTL8651_QUEUE_PACKET_ENTRY));	
	return SUCCESS;
}

static int32 _rtl8651_queueReinit(void){
	memset(queue_packet_tb,0,QUEUE_PACKET_ITEM * sizeof (struct _RTL8651_QUEUE_PACKET_ENTRY));	
	return SUCCESS;
}



static int32 _rtl8651_enQueuePacket(struct rtl_pktHdr * pkthdrPtr, struct ip* ip, ipaddr_t waitIp, uint8 type){
	int32 i;
	struct _RTL8651_QUEUE_PACKET_ENTRY *tb;	
	tb = queue_packet_tb;

#if 0
	/* avoid dupilcate add */
	for(i = 0 ; i < QUEUE_PACKET_ITEM ; i++){
		if(tb->q_time &&   
			tb->q_ip == ip ){
			return SUCCESS;
		}
		tb++;
	}
#endif
	/* real add */
	tb = queue_packet_tb;	
	for(i = 0 ; i < QUEUE_PACKET_ITEM ; i++){
		if(tb->q_time== 0 && 
			tb->q_pkthdrPtr == NULL){
			   struct rtl_pktHdr *newPh=NULL;
			   struct rtl_mBuf  *newmb=NULL, *oldMbuf;
			   uint8 *newCluster, *newClusterData;
			   uint32 newClusterId;
			   int16 rxdesc;

			if (pkthdrPtr->ph_rxdesc<0)	// this packet is not from RX Ring, we don't need to allocate a new one and exchang it
			{
				newPh = pkthdrPtr;
			}else
			{
				oldMbuf=pkthdrPtr->ph_mbuf;

				//get a new pkt hdr and mbuf.
				if((newmb = mBuf_get(MBUF_DONTWAIT,MBUFTYPE_DATA, 1)) == NULL){
//					mBuf_driverFreeMbufChain(oldMbuf); //free this pkt since we can't queue it.
					return FAILED;
				}
				if((mBuf_getPkthdr(newmb, MBUF_DONTWAIT) == (struct rtl_mBuf *) NULL)||
					(mBuf_reserve(newmb,rtl8651_fwdEnginePara.mbufHeadroom))){
//					mBuf_driverFreeMbufChain(oldMbuf); //free this pkt since we can't queue it.				
					mBuf_freeMbufChain(newmb);//free new mbuf we just get.
					return FAILED;
				}
				assert(newmb);
				newPh=newmb->m_pkthdr;
				newCluster=newmb->m_extbuf;
				newClusterData=newmb->m_data;
				newClusterId=newmb->m_extClusterId;
				assert(newPh);

				*newmb=*oldMbuf;//copy mbuf content
				newmb->m_next=NULL;
				newmb->m_pkthdr=newPh;

				rxdesc = newPh->ph_rxdesc;
				*newPh=*pkthdrPtr;//copy pkthdr
				newPh->ph_mbuf=newmb;
				newPh->ph_rxdesc=rxdesc;

				oldMbuf->m_extClusterId=newClusterId;
				oldMbuf->m_data=newClusterData;
				oldMbuf->m_extbuf=newCluster;
				oldMbuf->m_len=0;
				mBuf_freeMbufChain(oldMbuf);
			}

			tb->q_time=Q_TIME;
			tb->q_waitIp = waitIp;
			tb->q_type = type;
			tb->q_id = ntohs(ip->ip_id);
			tb->q_ip = ip;
			tb->q_pkthdrPtr = newPh;

			return SUCCESS;
		}
		tb++;
	}
	return FAILED;
}

static struct rtl_pktHdr * _rtl8651_queryQuenePacket(struct ip* ip, uint8 type, 
					uint32 *dst_ip,  struct ip **ipp){
	int32 i;
	struct rtl_pktHdr * pkt;
	struct _RTL8651_QUEUE_PACKET_ENTRY *tb;
	
	tb = queue_packet_tb;
	for(i = 0 ; i < QUEUE_PACKET_ITEM ; i++){
		if(tb->q_time && tb->q_type == type && tb->q_pkthdrPtr){					
			switch(type){
			case Q_ARP:
				if(memcmp(&tb->q_waitIp, &ip->ip_dst, 4) == 0 ){
					pkt = tb->q_pkthdrPtr;
					if(dst_ip)
					{
						*dst_ip = PKTGET_UINT32_UNALIGNED(&tb->q_ip->ip_dst);
					}
					if(ipp)
					{
						*ipp = tb->q_ip;
					}
					memset(tb,0,	sizeof(struct _RTL8651_QUEUE_PACKET_ENTRY));
					return pkt;
				}
				break;
			}			
		}
		tb++;
	}
	return NULL;	
}


static int32 _rtl8651_queuePacketSecondTimer(void){
	int32 i;
	struct rtl_pktHdr * pkt;
	struct _RTL8651_QUEUE_PACKET_ENTRY *tb;	
	tb = queue_packet_tb;	
	
	for(i = 0 ; i < QUEUE_PACKET_ITEM ; i++){
		if(tb->q_time)
			tb->q_time--;
		
		if(tb->q_time == 0 && tb->q_pkthdrPtr){		
			pkt = tb->q_pkthdrPtr;
			tb->q_pkthdrPtr = NULL;
			tb->q_type = 0;
			tb->q_id = 0;
			tb->q_ip = NULL;
			mBuf_freeMbufChain(pkt->ph_mbuf);	
		}		
		tb++;
	}
	
	return SUCCESS;	
}

#endif /* _RTL_QUEUE_ARP_PACKET */

#define PPPOE_NOT_EXISTED_SESSION_TERMINATION

void _rtl8651_freeAllPppoeActiveSession(void) {
	_rtl8651_pppoeActiveSession_entry_t * tempPtr;
	while(pppoeActiveSessionHead) {
		tempPtr = pppoeActiveSessionHead;
		pppoeActiveSessionHead = pppoeActiveSessionHead->next;
		tempPtr->next = pppoeFreeActiveSession;
		pppoeFreeActiveSession = tempPtr;
	}
}

void _rtl8651_updatePppoeActiveSession(uint16 pppoeSessionId) {
	_rtl8651_pppoeActiveSession_entry_t * tempActivePppoeSessionPtr, * trackingActivePppoeSessionPtr, * removedActivePppoeSessionPtr=NULL;
	uint16 smallestKeepAlive;
	
	//Decade current active session counter. If entry becomes invalid, just move it from active list to free list
	smallestKeepAlive = 0xFFFF;
	tempActivePppoeSessionPtr = pppoeActiveSessionHead;
	while (tempActivePppoeSessionPtr) {
		if(tempActivePppoeSessionPtr->keepAlive == 0) {//Moving this session to free list
			if(tempActivePppoeSessionPtr == pppoeActiveSessionHead) {
				pppoeActiveSessionHead = pppoeActiveSessionHead->next;
				removedActivePppoeSessionPtr = tempActivePppoeSessionPtr;
				tempActivePppoeSessionPtr = tempActivePppoeSessionPtr->next;//tempActivePppoeSessionPtr finally point to the following entry to check
			} else {
				trackingActivePppoeSessionPtr = pppoeActiveSessionHead;
				while(trackingActivePppoeSessionPtr->next) {
					if(trackingActivePppoeSessionPtr->next == tempActivePppoeSessionPtr) {//Find the entry to delete
						trackingActivePppoeSessionPtr->next = trackingActivePppoeSessionPtr->next->next;
						removedActivePppoeSessionPtr = tempActivePppoeSessionPtr;
						tempActivePppoeSessionPtr = trackingActivePppoeSessionPtr->next;//tempActivePppoeSessionPtr finally point to the following entry to check
						break;
					}
					trackingActivePppoeSessionPtr = trackingActivePppoeSessionPtr->next;
				}
				assert(removedActivePppoeSessionPtr);
			}						
			removedActivePppoeSessionPtr->next = pppoeFreeActiveSession;
			pppoeFreeActiveSession = removedActivePppoeSessionPtr;
		}
		else {
			tempActivePppoeSessionPtr->keepAlive--;
			if(smallestKeepAlive > tempActivePppoeSessionPtr->keepAlive)
				smallestKeepAlive = tempActivePppoeSessionPtr->keepAlive;
			tempActivePppoeSessionPtr = tempActivePppoeSessionPtr->next;//tempActivePppoeSessionPtr finally point to the following entry to check
		}
	}
	//Find the existed active session entry
	tempActivePppoeSessionPtr = pppoeActiveSessionHead;
	while(tempActivePppoeSessionPtr) {
		if(tempActivePppoeSessionPtr->sessionId == pppoeSessionId)
			break;
		tempActivePppoeSessionPtr = tempActivePppoeSessionPtr->next;
	}
	if(tempActivePppoeSessionPtr == NULL) {
		if(pppoeFreeActiveSession) {//Entry found, allocate free list
			tempActivePppoeSessionPtr = pppoeFreeActiveSession;
			pppoeFreeActiveSession = pppoeFreeActiveSession->next;
			tempActivePppoeSessionPtr->next = pppoeActiveSessionHead;
			pppoeActiveSessionHead = tempActivePppoeSessionPtr;
		}
		else {//Entry not found, find the one to scrify
			trackingActivePppoeSessionPtr = pppoeActiveSessionHead;
			while(trackingActivePppoeSessionPtr) {
				if(trackingActivePppoeSessionPtr->keepAlive == smallestKeepAlive) {
					tempActivePppoeSessionPtr = trackingActivePppoeSessionPtr;//Find the entry with minimum keep alive value. This entry is the one to be replaced.
					break;
				}
				trackingActivePppoeSessionPtr = trackingActivePppoeSessionPtr->next;
			}
		}
		tempActivePppoeSessionPtr->sessionId = pppoeSessionId;
	}
	tempActivePppoeSessionPtr->keepAlive = 0xFFFF;

}

void _rtl8651_terminatePppoeActiveSession(uint16 pppoeSessionId) {
	_rtl8651_pppoeActiveSession_entry_t * tempActivePppoeSessionPtr, * trackingActivePppoeSessionPtr;
	
	//Find the existed active session entry
	tempActivePppoeSessionPtr = pppoeActiveSessionHead;
	while(tempActivePppoeSessionPtr) {
		if(tempActivePppoeSessionPtr->sessionId == pppoeSessionId)
			break;
		tempActivePppoeSessionPtr = tempActivePppoeSessionPtr->next;
	}
	if(tempActivePppoeSessionPtr) {
		if(tempActivePppoeSessionPtr == pppoeActiveSessionHead) {
			pppoeActiveSessionHead = pppoeActiveSessionHead->next;
		} else {
			trackingActivePppoeSessionPtr = pppoeActiveSessionHead;
			while(trackingActivePppoeSessionPtr->next) {
				if(trackingActivePppoeSessionPtr->next == tempActivePppoeSessionPtr) {//Find the entry to delete
					trackingActivePppoeSessionPtr->next = trackingActivePppoeSessionPtr->next->next;
					break;
				}
				trackingActivePppoeSessionPtr = trackingActivePppoeSessionPtr->next;
			}
		}						
		tempActivePppoeSessionPtr->next = pppoeFreeActiveSession;
		pppoeFreeActiveSession = tempActivePppoeSessionPtr;
	}
}

/*
@func int32 | _rtl8651_pppoeSessionActive	| Check whether PPPoE session ID active
@parm uint16 | pppoeSessionId | The PPPoE session ID to lookup whether active or not
@rvalue TRUE	| 	Entry active
@rvalue FALSE | 	Entry not active
@comm 
	This function is used to check the specified PPPoE session ID in active state
*/
inline int32 _rtl8651_pppoeSessionActive(uint16 pppoeSessionId) {
	_rtl8651_pppoeActiveSession_entry_t * tempActivePppoeSessionPtr;
	
	tempActivePppoeSessionPtr = pppoeActiveSessionHead;
	while(tempActivePppoeSessionPtr) {
		if(tempActivePppoeSessionPtr->sessionId == pppoeSessionId)
			break;
		tempActivePppoeSessionPtr = tempActivePppoeSessionPtr->next;
	}
	if(tempActivePppoeSessionPtr)
		return TRUE;
	return FALSE;
}

/*
@func uint16 *		| _rtl8651_getNextPppoeActiveSession	| Get next active session ID
@parm uint16 * | curPppoeSessionId | The pointer point to the variable store the current active session ID
@parm uint16 * | nextPppoeSessionId | The pointer point to the variable used to store the returned next active session ID
@comm 
	This function is used to go through all active session ID
*/
uint16 * _rtl8651_getNextPppoeActiveSession(uint16 *curPppoeSessionId, uint16 * nextPppoeSessionId) {
	_rtl8651_pppoeActiveSession_entry_t * tempActivePppoeSessionPtr;
	
	if(nextPppoeSessionId == NULL)
		return NULL;//No storage to store the session ID
	if(curPppoeSessionId == NULL) {
		if(pppoeActiveSessionHead) {
			*nextPppoeSessionId = pppoeActiveSessionHead->sessionId;
			return nextPppoeSessionId;
		}
		return NULL;//First session ID not exist
	}

	tempActivePppoeSessionPtr = pppoeActiveSessionHead;
	while(tempActivePppoeSessionPtr) {
		if(tempActivePppoeSessionPtr->sessionId == *curPppoeSessionId) {
			if(tempActivePppoeSessionPtr->next) {
				*nextPppoeSessionId = tempActivePppoeSessionPtr->next->sessionId;
				return nextPppoeSessionId;
			}
			return NULL;//Next not exist
		}
		tempActivePppoeSessionPtr = tempActivePppoeSessionPtr->next;
	}
	return NULL; //Specified session ID not exist
}

#ifdef RTL8651_FRAGMENT_SHORTESTPATH_FWD
#define RTL8651_FRAGMENT_SHORTESTPATH_FUNCTIONS
/*
	set sip/dip/id/mode/aliasIp,
	use pkthdr field "ph_fragOffset" to save caches index
*/
static inline void _rtl8651_fragmentShortestPathCacheSet_firstHalf(struct rtl_pktHdr *pkthdr, uint8*l4hdr, struct ip *pip, ipaddr_t sip, ipaddr_t dip, uint32 mode, ipaddr_t aliasIp)
{
	uint16 idx = ((sip+dip+htons(pip->ip_id))&(RTL8651_FRAGMENTSHORTESTPATH_TBLSIZE-1));

	/* set index first */
	pkthdr->ph_fragOffset = idx;

	/*
		If DOS is turned-ON and its check-item includes Fragment-reassemble related (tear-drop, UDP bomb),
		IP-level (ex. IP spoofing) or Packet-count-level (ex. flooding), forwarding engine can NOT apply
		fragment shortest path forwarding to ALL fragment packets.
	*/
	/* IP level and Reassemble related list */
	#define DOS_L3_SHORSTESTPATHFORBID_LIST	RTL8651_DOS_IPSPOOF|RTL8651_DOS_TEARDROP
	/* TCP realted, we don't include LAND Attack because system enable it and one packet is matched, it will be dropped and will not arrive this function */
	#define DOS_TCP_SHORTESTPATHFORBID_SYN	RTL8651_DOS_SYNFLOOD|RTL8651_DOS_SRC_SYNFLOOD
	#define DOS_TCP_SHORTESTPATHFORBID_FIN	RTL8651_DOS_STEALTHFIN|RTL8651_DOS_SRC_STEALTHFIN
	/* UDP realted, we don't include LAND Attack because system enable it and one packet is matched, it will be dropped and will not arrive this function */
	#define DOS_UDP_SHORTESTPATHFORBID_LIST	RTL8651_DOS_UDPBOMB|RTL8651_DOS_UDPFLOOD|RTL8651_DOS_SRC_UDPFLOOD

	/* check ignore cases: if this entry want to modify SIP: LAN-to-WAN, or it will be WAN-to-LAN */
	if ((_rtl8651_dos_enable == TRUE) && (_rtl8651_dosIgnoreTypeCheck(((mode==RTL8651_FRAGMENT_SHORTESTPATH_MODE_MODIFYSIP)?0:1)) == FALSE))
	{
		if (pip->ip_p == IPPROTO_UDP)
		{	/* UDP */
			if (_rtl8651_dos_sw[pkthdr->ph_dsid] & (DOS_L3_SHORSTESTPATHFORBID_LIST|DOS_UDP_SHORTESTPATHFORBID_LIST))
				return;
		}else
		{	/* TCP */
			struct tcphdr *tc;

			assert(pip->ip_p == IPPROTO_TCP);

			tc = (struct tcphdr *)l4hdr;
			if ((tc->th_flags & TH_SYN) && (_rtl8651_dos_sw[pkthdr->ph_dsid] & (DOS_L3_SHORSTESTPATHFORBID_LIST|DOS_TCP_SHORTESTPATHFORBID_SYN)))
			{
				return;
			}else if ((tc->th_flags & TH_FIN) && (_rtl8651_dos_sw[pkthdr->ph_dsid] & (DOS_L3_SHORSTESTPATHFORBID_LIST|DOS_TCP_SHORTESTPATHFORBID_FIN)))
			{
				return;
			}			
		}
	}
	/*
		We must check UDP-Blocking function of their DSID is turned-on
		or not. If system turns-ON the destination-DSID's UDP-Blocking Checking, forwarding engine can NOT
		apply fragment shortest path forwarding to these packets.
	*/
	if ((pip->ip_p == IPPROTO_UDP) && (_rtl8651_udpSizeThreshValue[pkthdr->ph_dsid] != 0xffff))
		return;

	/* when system call firsHalf set, it MUST be the fragment IP flow, so we only need to check ip header length */
	if (IPHDRLEN_IS_TWENTY(pip))
	{
		_rtl8651_FragmentShortestPathFwdCache_t *cache;
		
		/* get the cache entry */
		cache = &(fragmentShortestPathCache[idx]);

		cache->sip = sip;
		cache->dip = dip;
		cache->id = htons(pip->ip_id);
		cache->aliasIp = aliasIp;
		cache->mode = mode;
		cache->status = RTL8651_FRAGMENT_SHORTESTPATH_STATUS_HALFOK;
	}
}
/*
	set smac/dmac/vlanIdx/portlist/pppoeIdx/(ToPppoe?)
*/
inline void _rtl8651_fragmentShortestPathCacheSet_secondHalf(struct rtl_pktHdr *pkthdr, uint8 *m_data, struct ip*pip)
{
	if (IS_FRAGMENT(pip->ip_off) && (IPHDRLEN_IS_TWENTY(pip)))
	{
		_rtl8651_FragmentShortestPathFwdCache_t *cache;

		/* get the cache entry */
		cache = &(fragmentShortestPathCache[(pkthdr->ph_fragOffset&(RTL8651_FRAGMENTSHORTESTPATH_TBLSIZE-1))]);

		if (cache->status == RTL8651_FRAGMENT_SHORTESTPATH_STATUS_HALFOK)
		{
			rtl8651_memcpy(&cache->dmac, (void*)m_data, 6);
			rtl8651_memcpy(&cache->smac, (void*)(m_data + 6), 6);
		
			cache->vlanIdx = pkthdr->ph_vlanIdx;
			cache->portList = pkthdr->ph_portlist;
			cache->pppoeIdx = pkthdr->ph_pppoeIdx;
			cache->extPortList = pkthdr->ph_extPortList;
			cache->srcExtPortNum = pkthdr->ph_srcExtPortNum;
			cache->flags = (pkthdr->ph_flags & 0x000f);
			cache->status = RTL8651_FRAGMENT_SHORTESTPATH_STATUS_SETOK;
			cache->age = RTL8651_FRAGMENT_SHORTESTPATH_TIMEOUT;		/* set timer */
		}
	}
}
/*
	disable fragment cache if destination link-layer type is not VALN/PPPOE or mtu is changed
*/
inline void _rtl8651_fragmentShortestPathCache_clear(struct rtl_pktHdr *pkthdr, struct ip*pip)
{
	if (IS_FRAGMENT(pip->ip_off) && (IPHDRLEN_IS_TWENTY(pip)))
	{
		_rtl8651_FragmentShortestPathFwdCache_t *cache;

		/* get the cache entry */
		cache = &(fragmentShortestPathCache[(pkthdr->ph_fragOffset&(RTL8651_FRAGMENTSHORTESTPATH_TBLSIZE-1))]);

		cache->status = 0;
	}
}

static void _rtl8651_fragmentShortestPathCache_timeUpdate(uint32 passed)
{
	_rtl8651_FragmentShortestPathFwdCache_t *ptr;
	uint32 idx;

	ptr = fragmentShortestPathCache;

	for (idx = 0 ; idx < RTL8651_FRAGMENTSHORTESTPATH_TBLSIZE; idx ++, ptr ++)
	{
		if (ptr->age == 0)
			continue;

		if (ptr->age > passed)
			ptr->age = (ptr->age - passed);
		else
		{
			ptr->age = 0;
			ptr->status = 0;
		}
	}
}

static inline int32 _rtl8651_fragmentShortestPathFwd(struct rtl_pktHdr*pkthdr, struct ip *pip)
{
	if (IS_FRAGMENT_REMAIN(pip->ip_off) && IPHDRLEN_IS_TWENTY(pip))
	{
		_rtl8651_FragmentShortestPathFwdCache_t *cache;
		ipaddr_t sip;
		ipaddr_t dip;
		uint16 id;

		sip = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
		dip = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
		id = ntohs(pip->ip_id);

		/* get the cache entry */
		cache = &(fragmentShortestPathCache[(sip+dip+id)&(RTL8651_FRAGMENTSHORTESTPATH_TBLSIZE-1)]);

		if ((cache->status != RTL8651_FRAGMENT_SHORTESTPATH_STATUS_SETOK) ||
			(sip != cache->sip) || (dip != cache->dip) || (id != cache->id))
			return FAILED;

		/***** Flow matched, forward it! *****/
		if (pkthdr->ph_pppeTagged)
		{
			mBuf_adjHead(pkthdr->ph_mbuf, 8);
			*((uint16 *)((uint32)pip - (uint32)2)) = htons(0x0800);
			pkthdr->ph_pppeTagged = 0;
		}
		/* set ip */
		if (cache->mode == RTL8651_FRAGMENT_SHORTESTPATH_MODE_MODIFYSIP)
		{
			PKTSET_UINT32_UNALIGNED(cache->aliasIp, &pip->ip_src);
		}else
		{
			PKTSET_UINT32_UNALIGNED(cache->aliasIp, &pip->ip_dst);
		}
		/* ttl */
		pip->ip_ttl --;
		/* set mac */
		rtl8651_memcpy((void *)((uint32)pip - (uint32)14), &cache->dmac, 6);
		rtl8651_memcpy((void *)((uint32)pip - (uint32)8), &cache->smac, 6);
		/* set pkt hdr */
		pkthdr->ph_proto = PKTHDR_IP;
		pkthdr->ph_portlist = cache->portList;
		pkthdr->ph_extPortList = cache->extPortList;
		pkthdr->ph_srcExtPortNum = cache->srcExtPortNum;
		pkthdr->ph_pppoeIdx = cache->pppoeIdx;
		pkthdr->ph_vlanIdx = cache->vlanIdx;
		pkthdr->ph_flags = ((pkthdr->ph_flags & 0xfff0) | cache->flags);

		rtlglue_drvSend((void*)pkthdr);
		return SUCCESS;
	}
	return FAILED;
}

#endif /* RTL8651_FRAGMENT_SHORTESTPATH_FWD */


/* CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU */
/*
	If the gateway is a PPPoE client, it will receive PPPoE packets, too
	(the destination MAC is gateway MAC).
	However, the socket of upper application binds on external interface, not PPPoE passthru VLAN.
	Therefore, we must refine the VLAN index of this packet.

	Louis notes: cfliu says this function should be implemented in re_core.c.
*/
#if 0
void _modifyPPPoEPassthruVid( struct rtl_pktHdr * pkthdrPtr,
                              uint16 ethType, rtl8651_tblDrv_vlanTable_t *local_vlanp )
{
	if ( rtl8651_drvPppoePassthruEnable && 
	     ( ethType == 0x8863 || ethType == 0x8864 ) &&
	     local_vlanp->vid == rtl8651_gPPPoEPassthruVlanID )
	{
		ether_addr_t vlanGMAC;

		/* get VLAN Gateway MAC address */
		rtl8651_getVlanMacAddress( local_vlanp->vid, &vlanGMAC, NULL );

		/* the dst MAC address is GMAC address ==> it is NOT a PPPoE passthru packet. */
		if ( memcmp( pkthdrPtr->ph_mbuf->m_data, &vlanGMAC, sizeof(ether_addr_t) ) == 0 )
		{
			/* This packet is trapped due to DMAC is GatewayMAC. */
			/* Change the source VLAN index to external interface. */
			pkthdrPtr->ph_vlanIdx = rtl8651_gPPPoEExternalVlanIndex;
		}
	}
}
#endif /*0*/

/****************************************************************************
	Function:
			RTL napt redirect to protocol stack

			Some special flows need handled by protocol stack and we trap them up,
			these flows are co-operated by forwarding engine and protocol stack.

			These APIs include Register/Unregister/internal-process.
	
****************************************************************************/
#ifdef _RTL_NAPT_REDIRECT
#define RTL8651_NAPTREDIRECT_FUNCTIONS

#undef _RTL_NAPT_REDIRECT_DEBUG

/******************************************************************
	Initial	:
				Allocate memory and initiate all data structures

		redirectTblSize 			:	redirect table size
		flowCacheSize				:	redirect flow cache size
******************************************************************/
static int32 _rtl8651_naptRedirectInit(rtl8651_fwdEngineInitPara_t *para)
{
	uint32 redirectTblSize;
	uint32 flowCacheSize;

	redirectTblSize = para->redirectTblSize;
	flowCacheSize = para->redirectFlowCacheSize;
	/****************************************
		Init redirect table
	****************************************/
	/* initial data structure */
	_rtl8651_naptRedirectEntry_inuse = NULL;
	_rtl8651_naptRedirectEntry_freelist = NULL;

	/* allocate memory */
	if (redirectTblSize)
	{
		uint32 idx;
		_rtl8651_naptRedirectEntry_t *ptr;
		
		_rtl8651_naptRedirectEntry_freelist = (_rtl8651_naptRedirectEntry_t *)rtlglue_malloc(redirectTblSize * sizeof(_rtl8651_naptRedirectEntry_t));

		if (!_rtl8651_naptRedirectEntry_freelist)
			return FAILED;

		/* init entries */
		memset(_rtl8651_naptRedirectEntry_freelist, 0, (redirectTblSize*sizeof(_rtl8651_naptRedirectEntry_t)));

		for (idx = 0, ptr = _rtl8651_naptRedirectEntry_freelist ; idx < redirectTblSize ; idx ++, ptr ++)
		{
			if (idx < (redirectTblSize - 1))
				ptr->next = ptr + 1;
		}
	}

	/****************************************
		Init redirect flow cache
	****************************************/
	/* initial data structure */
	_rtl8651_naptRedirectFlowCache_lru_head = NULL;
	_rtl8651_naptRedirectFlowCache_lru_tail = NULL;
	_rtl8651_naptRedirectFlowCache_freelist = NULL;

	/* allocate memory */
	if (flowCacheSize)
	{
		uint32 idx;
		_rtl8651_naptRedirectFlow_t *ptr;

		_rtl8651_naptRedirectFlowCache_freelist = (_rtl8651_naptRedirectFlow_t *)rtlglue_malloc(flowCacheSize * sizeof(_rtl8651_naptRedirectFlow_t));

		if (!_rtl8651_naptRedirectFlowCache_freelist)
			return FAILED;

		/* init entries */
		memset(_rtl8651_naptRedirectFlowCache_freelist, 0, (flowCacheSize*sizeof(_rtl8651_naptRedirectFlow_t)));

		for (idx = 0 , ptr = _rtl8651_naptRedirectFlowCache_freelist ; idx < flowCacheSize ; idx ++, ptr ++)
		{
			if (idx < (flowCacheSize - 1))
				ptr->next = ptr + 1;
		}
	}

	return SUCCESS;
}
/******************************************************************
	Reinit	:
				- Clear all napt-redirect entries
				- Clear all flow caches
******************************************************************/
static void _rtl8651_naptRedirectReinit(void)
{
	/* free all entries */
	{
		_rtl8651_naptRedirectEntry_t *ptr;

		while ((ptr = _rtl8651_naptRedirectEntry_inuse) != NULL)
		{
			/* remove from inuse list */
			_rtl8651_naptRedirectEntry_inuse = _rtl8651_naptRedirectEntry_inuse->next;
			/* add to free list */
			ptr->next = _rtl8651_naptRedirectEntry_freelist;
			_rtl8651_naptRedirectEntry_freelist = ptr;
		}
	}

	/* free all flow caches */
	{
		/********************************************
			move all inused flow chain to free list:

			Before:
					lru_head									lru_tail
					     \										   /
					(cache 1) -- (cache 2) -- (cache 3) -- ..... -- (cache N)

					freelist
					     \
					(cache free1) -- (cache free2) -- ..... -- (cache freeN)

			After:
					lru_head	lru_tail
						\        /
						  (null)

					freelist
					     \
					     (cache 1) -- (cache 2) -- (cache 3) -- ..... -- (cache N) -- (cache free1) -- (cache free2) -- ..... -- (cache freeN)

		********************************************/
		if (_rtl8651_naptRedirectFlowCache_lru_tail)
		{
			_rtl8651_naptRedirectFlowCache_lru_tail->next = _rtl8651_naptRedirectFlowCache_freelist;
		}

		if (_rtl8651_naptRedirectFlowCache_lru_head)
		{
			_rtl8651_naptRedirectFlowCache_freelist = _rtl8651_naptRedirectFlowCache_lru_head;
		}

		_rtl8651_naptRedirectFlowCache_lru_head = NULL;
		_rtl8651_naptRedirectFlowCache_lru_tail = NULL;
	}
}

/******************************************************************
	Utilities
******************************************************************/
/*
	find exist entry in napt redirect table for a specific sip/sport/original dip/original dport
	if collision occurred, system return the first entry found.
*/
static _rtl8651_naptRedirectEntry_t *_rtl8651_findExistNaptRedirectEntry(
	uint8	isExceptionList,
	uint8	isTcp,
	ipaddr_t	sip,
	uint16	sport,
	ipaddr_t	originalDip,
	uint16	originalDport)
{
	_rtl8651_naptRedirectEntry_t *ptr;

#ifdef _RTL_NAPT_REDIRECT_DEBUG
	rtlglue_printf("\n\tFind %s entry for %u.%u.%u.%u(%d) -> %u.%u.%u.%u(%d)\n", isExceptionList?"Exception":"Redirect", NIPQUAD(sip), sport, NIPQUAD(originalDip), originalDport);
#endif /* _RTL_NAPT_REDIRECT_DEBUG */


	ptr = _rtl8651_naptRedirectEntry_inuse;

	while (ptr)
	{
#ifdef _RTL_NAPT_REDIRECT_DEBUG
		rtlglue_printf("\n\tTry match %s rule:\t(%d) %u.%u.%u.%u:%d -> %u.%u.%u.%u:%d ** %u.%u.%u.%u:%d\n",
				(ptr->isExceptionList==TRUE)?"Exception":"Redirect",
				ptr->isTcp, NIPQUAD(ptr->sip), ptr->sport, NIPQUAD(ptr->dip), ptr->dport, NIPQUAD(ptr->newDip), ptr->newDport);
#endif /* _RTL_NAPT_REDIRECT_DEBUG */

		if (ptr->isExceptionList != isExceptionList)
			goto next;
		if (ptr->isTcp != isTcp)
			goto next;
		if (ptr->sip && sip && (ptr->sip != sip))
			goto next;
		if (ptr->sport && sport && (ptr->sport != sport))
			goto next;
		if (ptr->dip && originalDip && (ptr->dip != originalDip))
			goto next;
		if (ptr->dport && originalDport && (ptr->dport != originalDport))
			goto next;
		/* entry found !*/
		return ptr;
	next:
		ptr = ptr->next;
	}

	return NULL;
}
/******************************************************************
	Used by forwarding engine:
		find the Napt Redirect Entry for new flow. If it's found,
		user can create cache for this flow if createCache==TRUE.
******************************************************************/
static inline int32 _rtl8651_findNaptRedirectEntry(
	uint8	isTcp,
	ipaddr_t	sip,
	uint16	sport,
	ipaddr_t	dip,
	uint16	dport,
	ipaddr_t	*pNewDip,
	uint16	*pNewDport,
	uint8	createCache,		/* create cache if entry is found */
	uint16	vlanIdx			/* incoming vlan index */
)
{
	_rtl8651_naptRedirectEntry_t *ptr;

	assert(pNewDip);
	assert(pNewDport);

#ifdef _RTL_NAPT_REDIRECT_DEBUG
	rtlglue_printf("[%s]\tFlow -- (%d) %u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
			__FUNCTION__, isTcp, NIPQUAD(sip), sport, NIPQUAD(dip), dport);
#endif /* _RTL_NAPT_REDIRECT_DEBUG */

	if ((ptr = _rtl8651_findExistNaptRedirectEntry(TRUE, isTcp, sip, sport, dip, dport)) != NULL)
	{	/* flow in exception list, we skip following NAPT redirect entry check */
		return FAILED;
	} else if ((ptr = _rtl8651_findExistNaptRedirectEntry(FALSE, isTcp, sip, sport, dip, dport)) != NULL)
	{	/* entry found */
		ipaddr_t	newDip = 0;
		uint16 newDport = 0;

		if (ptr->newDport >= 0)
		{
			/* set dport */
			newDport = ptr->newDport;

			/* set dip */
			if (ptr->newDip == 0)
			{	/* get find the dip (gateway's interface ip) for this entry */
				rtl8651_tblDrv_vlanTable_t *inVlan;
				rtl8651_tblDrv_networkIntfTable_t *netif;

				inVlan = &DrvTbl.vlan[vlanIdx];
				if (!inVlan->valid)
					return FAILED;

				netif = _rtl8651_getNetworkIntfEntryByVid(inVlan->vid);

				switch (netif->linkLayerType)
				{
					case RTL8651_LL_VLAN:
					{
						rtl8651_tblDrv_ipIntfEntry_t * ipIntfPtr;
						rtl8651_tblDrv_ipIntfIpAddrEntry_t * ipAddrPtr;

						ipIntfPtr = SLIST_FIRST(&netif->IpHead);
						if (!ipIntfPtr)
							return FAILED;
						ipAddrPtr = SLIST_FIRST(&ipIntfPtr->ipAddr);
						if (!ipAddrPtr)
							return FAILED;
						newDip = ipAddrPtr->ipAddr;
					}
						break;
					case RTL8651_LL_PPPOE:
#if defined(CONFIG_RTL865X_PPTPL2TP)||defined(CONFIG_RTL865XB_PPTPL2TP)
					case RTL8651_LL_PPTP:
					case RTL8651_LL_L2TP:
#endif /* CONFIG_RTL865X_PPTPL2TP , CONFIG_RTL865XB_PPTPL2TP */
					{
						rtl8651_tblDrvSession_t *sessionPtr;

						sessionPtr = CTAILQ_FIRST(&netif->sessionHead);
						if ((!sessionPtr) || (!sessionPtr->ip_t))
							return FAILED;
							newDip = sessionPtr->ip_t->ipAddr;
					}
						break;
					default:
						#ifdef RTL865X_DEBUG
						rtlglue_printf("%s[%d]ERROR: unknown l2 type: %d\n", __FUNCTION__, __LINE__, netif->linkLayerType);
						#endif
						break;
				}
			}else
			{	/* new dip is assigned by protocol stack */
				newDip = ptr->newDip;
			}
		}else
		{
			/* set dport */
			newDport = dport;
			/* set dip */
			newDip = dip;
		}
		/*********************
			add cache
		*********************/
		if (createCache == TRUE)
		{
			if (_rtl8651_addNaptRedirectFlow(isTcp, sip, sport, dip, dport, newDip, newDport) != SUCCESS)
			{
				return FAILED;
			}
		}

		/*********************
			set return value
		*********************/
		*pNewDip = newDip;
		*pNewDport = newDport;

		return SUCCESS;
	}

	/* don't match any entry */
	return FAILED;
}

/******************************************************************
		find the Napt Redirect flow cache.
		Return the pointer of entry found if it exists.

		* 2 indexing approaches:

			1. (isTcp, sip, sport, originalDip, originalDport, newDip, newDport)
				- if originalDip != 0

			2. (isTcp, sip, sport, newDip, newDport)
				- if originalDip == 0

		* (isExternalQuery == TRUE) means this query is called by external functions and we must
			increase the queryCount of the entry we found.

		* (refresh == TRUE) means that, for the entry we found, its query count must be reset to zero,
			its age must be re-fill to RTL8651_NAPTREDIRECT_CACHETIMEOUT and we must move it
			to the tail of lru list.
		
******************************************************************/
static inline _rtl8651_naptRedirectFlow_t *_rtl8651_findNaptRedirectFlow(
	uint8	isTcp,
	ipaddr_t	sip,
	uint16	sport,
	ipaddr_t	originalDip,
	uint16	originalDport,
	ipaddr_t	newDip,
	uint16	newDport,
	uint8	isExternalQuery,
	uint8	refresh
)
{
	_rtl8651_naptRedirectFlow_t *prev;
	_rtl8651_naptRedirectFlow_t *ptr;

	assert(sip);
	assert(newDip);

	prev = NULL;
	ptr = _rtl8651_naptRedirectFlowCache_lru_head;

	while (ptr)
	{
		if (	(isTcp == ptr->isTcp) &&
			(sip == ptr->sip) &&
			(sport == ptr->sport) &&
			(newDip == ptr->newDip) &&
			(newDport == ptr->newDport) &&
			(!originalDip || (originalDip == ptr->dip)) &&
			(!originalDip /* we don't use Dport because Dport can be zero ! */ || (originalDport == ptr->dport)))
		{
			goto found;
		}
		prev = ptr;
		ptr = ptr->next;
	}
	/* entry not be found, return NULL */
	return NULL;
found:
	assert(ptr); /* ptr is the entry found */

	if (isExternalQuery)
	{	/* add query count */
		ptr->queryCnt ++;
	}

	if (refresh)
	{	/* refresh entry's all control items */
		ptr->queryCnt = 0;
		ptr->age = RTL8651_NAPTREDIRECT_CACHETIMEOUT;
		/* move to the tail of lru list */
		if (ptr != _rtl8651_naptRedirectFlowCache_lru_tail)
		{
			/* remove from original position in linked list */
			if (prev)
			{	/* ptr is NOT the first entry of lru list */
				prev->next = ptr->next;
			}else
			{	/* ptr is the first entry of lru list */
				assert(ptr == _rtl8651_naptRedirectFlowCache_lru_head);
				_rtl8651_naptRedirectFlowCache_lru_head = _rtl8651_naptRedirectFlowCache_lru_head->next;
			}
			/* add to the tail of linked list */
			assert(_rtl8651_naptRedirectFlowCache_lru_tail);
			_rtl8651_naptRedirectFlowCache_lru_tail->next = ptr;
			ptr->next = NULL;
			_rtl8651_naptRedirectFlowCache_lru_tail = ptr;
		}
	}

	return ptr;
}

/******************************************************************
			Add the new flow cache entry
******************************************************************/
static inline int32 _rtl8651_addNaptRedirectFlow(
	uint8	isTcp,
	ipaddr_t	sip,
	uint16	sport,
	ipaddr_t	originalDip,
	uint16	originalDport,
	ipaddr_t	newDip,
	uint16	newDport
)
{
	_rtl8651_naptRedirectFlow_t *prev;
	_rtl8651_naptRedirectFlow_t *ptr;

	/*
		if this entry is found in cache, we refresh it and return SUCCESS immediately.
	*/
	ptr = _rtl8651_findNaptRedirectFlow(isTcp, sip, sport, originalDip, originalDport, newDip, newDport, FALSE, TRUE);
	if (ptr)
	{
		return SUCCESS;
	}
	
	/*
		allocate cache entry
	*/
		/* from freelist */
	if (_rtl8651_naptRedirectFlowCache_freelist)
	{
		ptr = _rtl8651_naptRedirectFlowCache_freelist;
		_rtl8651_naptRedirectFlowCache_freelist = _rtl8651_naptRedirectFlowCache_freelist->next;
	}else
	{	/* from lru list, entry which has been queried by protocol stack (queryCnt > 0) has higher priority to be allocated */

		/* check if there is cache entry which has been queried or not */
		prev = NULL;
		ptr = _rtl8651_naptRedirectFlowCache_lru_head;
		while (ptr)
		{
			if (ptr->queryCnt > 0)
			{	/* use it */
				if (prev)
				{
					prev->next = ptr->next;
				}else
				{
					_rtl8651_naptRedirectFlowCache_lru_head = ptr->next;
				}

				if (ptr == _rtl8651_naptRedirectFlowCache_lru_tail)
				{
					_rtl8651_naptRedirectFlowCache_lru_tail = prev;
				}
				goto setNewCache;
			}
			prev = ptr;
			ptr = ptr->next;
		}

		/* no any entry is queried, use the first one */
		if (!_rtl8651_naptRedirectFlowCache_lru_head)
			return FAILED;

		ptr = _rtl8651_naptRedirectFlowCache_lru_head;

		if (_rtl8651_naptRedirectFlowCache_lru_head == _rtl8651_naptRedirectFlowCache_lru_tail)
		{	/* only one entry in lru list */
			_rtl8651_naptRedirectFlowCache_lru_head = NULL;
			_rtl8651_naptRedirectFlowCache_lru_tail = NULL;
		}else
		{
			_rtl8651_naptRedirectFlowCache_lru_head = _rtl8651_naptRedirectFlowCache_lru_head->next;
		}
	}

/* start to set this new cache entry */
setNewCache:

	assert(ptr);

	ptr->isTcp		= isTcp;
	ptr->sip			= sip;
	ptr->sport		= sport;
	ptr->dip			= originalDip;
	ptr->dport		= originalDport;
	ptr->newDip		= newDip;
	ptr->newDport	= newDport;
	ptr->age			= RTL8651_NAPTREDIRECT_CACHETIMEOUT;
	ptr->queryCnt	= 0;
	ptr->next		= NULL;

	/*
		put it to tail of lru list
	*/
	if (_rtl8651_naptRedirectFlowCache_lru_tail)
		_rtl8651_naptRedirectFlowCache_lru_tail->next = ptr;
	_rtl8651_naptRedirectFlowCache_lru_tail = ptr;
	if (_rtl8651_naptRedirectFlowCache_lru_head == NULL)
		_rtl8651_naptRedirectFlowCache_lru_head = ptr;

	return SUCCESS;
}
/******************************************************************
			Process age of napt redirect flow cache
******************************************************************/
static void _rtl8651_naptRedirectFlowTimeUpdate(uint32 secPassed)
{
	_rtl8651_naptRedirectFlow_t *ptr;

	ptr = _rtl8651_naptRedirectFlowCache_lru_head;

	while (ptr)
	{
		if (ptr->age > secPassed)
		{
			ptr->age -= secPassed;
			ptr = ptr->next;
		}else
		{
			/* because it's lru list, so this entry must be the first entry in lru list */
			assert(_rtl8651_naptRedirectFlowCache_lru_head == ptr);

			/* remove this entry (it's the first one) from lru list */
			_rtl8651_naptRedirectFlowCache_lru_head = _rtl8651_naptRedirectFlowCache_lru_head->next;
			if (_rtl8651_naptRedirectFlowCache_lru_tail == ptr)
				_rtl8651_naptRedirectFlowCache_lru_tail = NULL;
			/* move to freelist */
			ptr->next = _rtl8651_naptRedirectFlowCache_freelist;
			_rtl8651_naptRedirectFlowCache_freelist = ptr;

			/* set new ptr to the first entry of lru list */
			ptr = _rtl8651_naptRedirectFlowCache_lru_head;
		}
	}
}
/******************************************************************************************
	External functions
******************************************************************************************/
/******************************************************************
	Register flows which need to be redirected to protocol stack
******************************************************************/
int32 rtl8651_registerRedirectOutboundNaptFlow(
		uint8	isTcp,			/* TRUE: tcp, FALSE: udp */
		ipaddr_t	sip,				/* val 0 for wildcard */
		uint16	sport,			/* val 0 for wildcard */
		ipaddr_t	originalDip,		/* val 0 for wildcard */
		uint16	originalDport,	/* val 0 for wildcard */
		ipaddr_t	newDip,			/* val 0 for don't care, automatically assigned by gateway's LAN IP */
		int32	newDport		/* must assign a new Dport to protocol stack, if < 0, both 'newDip' and 'newDport' will not be set(original dip/dport are kept) */
)
{
	int32 retval = FAILED;

	rtlglue_drvMutexLock();

	/*
		Will this entry cause collision in Redirect table?
	*/
	if (_rtl8651_findExistNaptRedirectEntry(FALSE, isTcp, sip, sport, originalDip, originalDport))
		goto out;
	/*
		check the correctness of input parameters
	*/
	if (newDport == 0)
		goto out;
	/*
		add this new entry
	*/
	{
		_rtl8651_naptRedirectEntry_t *ptr;

		if (_rtl8651_naptRedirectEntry_freelist == NULL)
			goto out;

		ptr = _rtl8651_naptRedirectEntry_freelist;
		_rtl8651_naptRedirectEntry_freelist = _rtl8651_naptRedirectEntry_freelist->next;

		ptr->isExceptionList	= FALSE;
		ptr->isTcp			= isTcp;
		ptr->sip				= sip;
		ptr->sport			= sport;
		ptr->dip				= originalDip;
		ptr->dport			= originalDport;
		ptr->newDip			= newDip;
		ptr->newDport		= newDport;

		ptr->next = _rtl8651_naptRedirectEntry_inuse;
		_rtl8651_naptRedirectEntry_inuse = ptr;
	}
	retval = SUCCESS;

out:
	rtlglue_drvMutexUnlock();
	return retval;
}
/******************************************************************
	Un-register flows which are trapped to protocol stack now
******************************************************************/
int32 rtl8651_unregisterRedirectOutboundNaptFlow(
		uint8	isTcp,			/* TRUE: tcp, FALSE: udp */
		ipaddr_t	sip,				/* val 0 for wildcard */
		uint16	sport,			/* val 0 for wildcard */
		ipaddr_t	originalDip,		/* val 0 for wildcard */
		uint16	originalDport,	/* val 0 for wildcard */
		ipaddr_t	newDip,			/* val 0 for wildcard */
		int32	newDport		/* val 0 for wildcard */
)
{
	/****************************************
		Exactly match and remove entry
	****************************************/
	_rtl8651_naptRedirectEntry_t *prev;
	_rtl8651_naptRedirectEntry_t *ptr;
	uint32 removeCnt;

	rtlglue_drvMutexLock();

	prev = NULL;
	ptr = _rtl8651_naptRedirectEntry_inuse;
	removeCnt = 0;

	while (ptr)
	{
		if (	(ptr->isExceptionList == FALSE) &&
			(isTcp == ptr->isTcp) &&
			(!sip || (sip == ptr->sip)) &&
			(!sport || (sport == ptr->sport)) &&
			(!originalDip || (originalDip == ptr->dip)) &&
			(!originalDport || (originalDport == ptr->dport)) &&
			(!newDip || (newDip == ptr->newDip)) &&
			(!newDport || (newDport == ptr->newDport)))
		{
			/* entry is found, remove it. */
			if (prev)
			{	/* entry is NOT the first one in redirect table */
				prev->next = ptr->next;
			}else
			{	/* entry is the first one in redirect table */
				_rtl8651_naptRedirectEntry_inuse = ptr->next;
			}
			ptr->next = _rtl8651_naptRedirectEntry_freelist;
			_rtl8651_naptRedirectEntry_freelist = ptr;

			removeCnt ++;	/* add the remove Count */
			/*
				set new prev/ptr
			*/
			if (prev)
			{	/* don't change prev */
				ptr = prev->next;
			}else
			{
				ptr = _rtl8651_naptRedirectEntry_inuse;
			}
		}else
		{
			prev = ptr;
			ptr = ptr->next;
		}
	}

	/* return the total remove count */
	rtlglue_drvMutexUnlock();
	return removeCnt;
}
/******************************************************************
	Query the original DIP/DPORT of a specific trapped flow
	allocated from flow cache
******************************************************************/
int32 rtl8651_queryRedirectOutboundNaptFlow(
		uint8	isTcp,
		ipaddr_t	sip,
		uint16	sport,
		ipaddr_t	*pOriginalDip,
		uint16	*pOriginalDport,
		ipaddr_t	newDip,
		uint16	newDport
)
{
	int32 retval = FAILED;
	rtlglue_drvMutexLock();
	retval = _rtl8651_queryRedirectOutboundNaptFlow(	isTcp,
													sip, sport,
													pOriginalDip, pOriginalDport,
													newDip, newDport);
	rtlglue_drvMutexUnlock();
	return retval;
}

static inline int32 _rtl8651_queryRedirectOutboundNaptFlow(
		uint8	isTcp,
		ipaddr_t	sip,
		uint16	sport,
		ipaddr_t	*pOriginalDip,
		uint16	*pOriginalDport,
		ipaddr_t	newDip,
		uint16	newDport
)
{
	_rtl8651_naptRedirectFlow_t *ptr;

	/* caller MUST provide these two references */
	assert(pOriginalDip);
	assert(pOriginalDport);

	ptr = _rtl8651_findNaptRedirectFlow(isTcp, sip, sport, 0, 0, newDip, newDport, TRUE, FALSE);

	if (ptr)
	{
		*pOriginalDip = ptr->dip;
		*pOriginalDport = ptr->dport;

		return SUCCESS;
	}
	return FAILED;
}

/******************************************************************
	resolve the modified IP/PORT back to the original IP/PORT
******************************************************************/
int32 _rtl8651_resolveProxyInboundFlow(rtl8651_tblDrv_vlanTable_t *dvlan, struct ip *pip)
{
	struct tcphdr *tc = NULL;
	struct udphdr *ud = NULL;
	uint8 isTcp = FALSE;
	ipaddr_t sip = 0;
	ipaddr_t dip = 0;
	uint16 sport = 0;
	uint16 dport = 0;
	/* ip/port to modify */
	ipaddr_t newSip = 0;
	uint16 newSport = 0;

	/* we only process packet from Gateway to Internal VLAN */
	if (!(dvlan->internal))
		goto out;

	/* fragment packet use fragment cache to get the original sip */
	if (IS_FRAGMENT_REMAIN(pip->ip_off))
	{
		if ((pip->ip_p == IPPROTO_TCP) || (pip->ip_p == IPPROTO_UDP))
		{
			sip = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
			dip = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

			if (_rtl8651_getFragCache(NAPT_REDIRECT_FRAG_INBOUND, sip, dip, ntohs(pip->ip_id), &newSip, NULL, NULL, NULL) == SUCCESS)
			{	/* cache found ! */
				/* modify source ip */
				PKTSET_UINT32_UNALIGNED(newSip, &pip->ip_src);
				/* adjust checksum */
				_RTL8651_L3_ADJUST_CHECKSUM(newSip, sip, pip->ip_sum);
			}
		}
		goto out;
	}

	/* we only process TCP and UDP */
	switch (pip->ip_p)
	{
		case IPPROTO_TCP:
			isTcp = TRUE;
			tc = (struct tcphdr *) L4HDR_ADDR(pip);
			sport = ntohs(tc->th_sport);
			dport = ntohs(tc->th_dport);
			break;
		case IPPROTO_UDP:
			isTcp = FALSE;
			ud = (struct udphdr *) L4HDR_ADDR(pip);
			sport = ntohs(ud->uh_sport);
			dport= ntohs(ud->uh_dport);
			break;
		default:
			goto out;
	}

	sip = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	dip = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);

	/* query redirect table to get original RP's ip/port information */
	if (_rtl8651_queryRedirectOutboundNaptFlow(isTcp, dip, dport, &newSip, &newSport, sip, sport) != SUCCESS)
		goto out;

	/* modify source ip/port, update cache, and adjust checksum */
	{
		uint16 *l4cksum;
	
		assert(newSip);

		/* modify packet info */
		PKTSET_UINT32_UNALIGNED(newSip, &pip->ip_src);
		if (isTcp == TRUE)
		{
			l4cksum = &(tc->th_sum);
			tc->th_sport = htons(newSport);
		}else
		{
			l4cksum = &(ud->uh_sum);
			ud->uh_sport = htons(newSport);
		}

		/* adjust checksum */
		if ((isTcp != FALSE) || (*l4cksum))
			_RTL8651_L4_ADJUST_CHECKSUM(newSip, sip, newSport, sport, *l4cksum);
		_RTL8651_L3_ADJUST_CHECKSUM(newSip, sip, pip->ip_sum);

		/* update cache */
		_rtl8651_updateFragCache(NAPT_REDIRECT_FRAG_INBOUND, sip, dip, ntohs(pip->ip_id), newSip, 0, NULL, 0);
	}
out:	
	return SUCCESS;
}

/******************************************************************
	Register flows which need to be redirected to protocol stack
******************************************************************/
int32 rtl8651_registerRedirectOutboundNaptExceptionList(
		uint8	isTcp,			/* TRUE: tcp, FALSE: udp */
		ipaddr_t	sip,				/* val 0 for wildcard */
		uint16	sport,			/* val 0 for wildcard */
		ipaddr_t	dip,				/* val 0 for wildcard */
		uint16	dport			/* val 0 for wildcard */
)
{
	int32 retval = FAILED;

	rtlglue_drvMutexLock();
	/*
		Will this entry cause collision in Redirect table?
	*/
	if (_rtl8651_findExistNaptRedirectEntry(TRUE, isTcp, sip, sport, dip, dport))
		goto out;
	/*
		add this new entry
	*/
	{
		_rtl8651_naptRedirectEntry_t *ptr;

		if (_rtl8651_naptRedirectEntry_freelist == NULL)
			goto out;

		ptr = _rtl8651_naptRedirectEntry_freelist;
		_rtl8651_naptRedirectEntry_freelist = _rtl8651_naptRedirectEntry_freelist->next;

		ptr->isExceptionList	= TRUE;
		ptr->isTcp			= isTcp;
		ptr->sip				= sip;
		ptr->sport			= sport;
		ptr->dip				= dip;
		ptr->dport			= dport;
		ptr->newDip			= 0;
		ptr->newDport		= 0;

		ptr->next = _rtl8651_naptRedirectEntry_inuse;
		_rtl8651_naptRedirectEntry_inuse = ptr;
	}
	retval = SUCCESS;

out:
	rtlglue_drvMutexUnlock();
	return retval;
}
/******************************************************************
	Un-register flows which are trapped to protocol stack now
******************************************************************/
int32 rtl8651_unregisterRedirectOutboundNaptExceptionList(
		uint8	isTcp,			/* TRUE: tcp, FALSE: udp */
		ipaddr_t	sip,				/* val 0 for wildcard */
		uint16	sport,			/* val 0 for wildcard */
		ipaddr_t	dip,				/* val 0 for wildcard */
		uint16	dport			/* val 0 for wildcard */
)
{
	/****************************************
		Exactly match and remove entry
	****************************************/
	_rtl8651_naptRedirectEntry_t *prev;
	_rtl8651_naptRedirectEntry_t *ptr;
	uint32 removeCnt;

	rtlglue_drvMutexLock();

	prev = NULL;
	ptr = _rtl8651_naptRedirectEntry_inuse;
	removeCnt = 0;

	while (ptr)
	{
		if (	(ptr->isExceptionList == TRUE) &&
			(isTcp == ptr->isTcp) &&
			(!sip || (sip == ptr->sip)) &&
			(!sport || (sport == ptr->sport)) &&
			(!dip || (dip == ptr->dip)) &&
			(!dport || (dport == ptr->dport)))
		{
			/* entry is found, remove it. */
			if (prev)
			{	/* entry is NOT the first one in redirect table */
				prev->next = ptr->next;
			}else
			{	/* entry is the first one in redirect table */
				_rtl8651_naptRedirectEntry_inuse = ptr->next;
			}
			ptr->next = _rtl8651_naptRedirectEntry_freelist;
			_rtl8651_naptRedirectEntry_freelist = ptr;

			removeCnt ++;	/* add the remove Count */
			/*
				set new prev/ptr
			*/
			if (prev)
			{	/* don't change prev */
				ptr = prev->next;
			}else
			{
				ptr = _rtl8651_naptRedirectEntry_inuse;
			}
		}else
		{
			prev = ptr;
			ptr = ptr->next;
		}
	}

	/* return the total remove count */

	rtlglue_drvMutexUnlock();

	return removeCnt;
}

#endif /* _RTL_NAPT_REDIRECT */

#ifdef _RTL_PPTPL2TP_HDR_REFILL
#define RTL8651_PPTPL2TP_HEADER_REFILL_FUNCTIONS
static int32 _rtl8651_pptpHdrCacheInit(rtl8651_fwdEngineInitPara_t *para)
{
	uint32 cnt;

	cnt = para->pptpHdrCacheSize;

	/* init all variables */
	memset(_rtl8651_pptpHdrCache_hashTbl, 0, sizeof(_rtl8651_pptpHdrCache_hashTbl));
	_rtl8651_pptpHdrCache_lru_head = NULL;
	_rtl8651_pptpHdrCache_lru_tail = NULL;
	_rtl8651_pptpHdrCache_unused = NULL;

	/* allocate all header cache */
	if (cnt)
	{
		uint32 idx;
		_rtl8651_pptpHdrCacheEntry_t *ptr;

		_rtl8651_pptpHdrCache_unused = (_rtl8651_pptpHdrCacheEntry_t*)rtlglue_malloc(cnt * sizeof(_rtl8651_pptpHdrCacheEntry_t));

		if (_rtl8651_pptpHdrCache_unused == NULL)
			return FAILED;

		/* init cache */
		memset(_rtl8651_pptpHdrCache_unused, 0, (cnt * sizeof(_rtl8651_pptpHdrCacheEntry_t)));

		ptr = _rtl8651_pptpHdrCache_unused;
		for (idx = 0 ; idx < (cnt-1) ; idx ++, ptr ++)
		{
			ptr->next = ptr + 1;
		}
	}

	return SUCCESS;
}

static void _rtl8651_pptpHdrCacheReinit(void)
{
	/* flush all caches */
	while (_rtl8651_pptpHdrCache_lru_head)
	{	/* cache exist */
		_rtl8651_pptpHdrCache_Del(_rtl8651_pptpHdrCache_lru_head);
	}
}

/* save the sequence number and ack number of each pptp header */
void _rtl8651_pptpHdrCache_Add(uint16 ip_id, uint16 ip_off, uint32 seq, uint32 ack, uint8 type)
{
	uint32 hashIdx = RTL8651_IP_ID_EASYHASH(0, (uint32)ip_off, ip_id, RTL8651_PPTPHDR_CACHE_HASH_TBLSIZE);
	_rtl8651_pptpHdrCacheEntry_t **pHashTbl;
	_rtl8651_pptpHdrCacheEntry_t *ptr;

	/* allocate cache entry */
	if (_rtl8651_pptpHdrCache_unused)
	{	/* from unused list */
fromUnused:
		ptr = _rtl8651_pptpHdrCache_unused;
		_rtl8651_pptpHdrCache_unused = _rtl8651_pptpHdrCache_unused->next;
	}else if (_rtl8651_pptpHdrCache_lru_head)	/* from lru list */
	{
		_rtl8651_pptpHdrCache_Del(_rtl8651_pptpHdrCache_lru_head);
		assert(_rtl8651_pptpHdrCache_unused);
		goto fromUnused;
	}else
		goto out;	/* can't allocate entry, system stop cache and return */
	
	ptr->hdr = NULL;
	ptr->next = NULL;
	ptr->prev = NULL;
	ptr->lru_next = NULL;
	ptr->lru_prev = NULL;


	/* set variables to new entry */
	ptr->id		= ip_id;
	ptr->offset	= ip_off;
	ptr->seq		= seq;
	ptr->ack		= ack;
	ptr->type	= type;
	/* put this new entry into hash table */
	pHashTbl		= &(_rtl8651_pptpHdrCache_hashTbl[hashIdx]);
	ptr->hdr		= pHashTbl;
	ptr->prev	= NULL;
	ptr->next	= *pHashTbl;
	if (*pHashTbl)
		(*pHashTbl)->prev = ptr;
	*pHashTbl = ptr;
	/* put this new entry into the tail of lru list */
	if (_rtl8651_pptpHdrCache_lru_tail)
	{	/* there are last 1 entry in lru list */
		ptr->lru_next = NULL;
		ptr->lru_prev = _rtl8651_pptpHdrCache_lru_tail;
		_rtl8651_pptpHdrCache_lru_tail->lru_next = ptr;
		_rtl8651_pptpHdrCache_lru_tail = ptr;
	}else
	{	/* lru list is empty */
		ptr->lru_next = NULL;
		ptr->lru_prev = NULL;
		_rtl8651_pptpHdrCache_lru_head = ptr;
		_rtl8651_pptpHdrCache_lru_tail = ptr;
	}
out:
	return;
}

/* remove cache entry */
static void _rtl8651_pptpHdrCache_Del(_rtl8651_pptpHdrCacheEntry_t *ptr)
{
	_rtl8651_pptpHdrCacheEntry_t **hashTbl;
	/* remove it from cached entry */
	assert(ptr->hdr);
	hashTbl = ptr->hdr;
	if (*hashTbl == ptr)
	{	/* ptr is the first entry of hash table */
		assert(ptr->prev == NULL);
		*hashTbl = ptr->next;
		if (ptr->next)
			ptr->next->prev = NULL;
	}else
	{	/* ptr is NOT the first entry of hash table */
		assert(ptr->prev);
		ptr->prev->next = ptr->next;
		if (ptr->next)
			ptr->next->prev = ptr->prev;
	}
	ptr->hdr = NULL;
	ptr->prev = NULL;
	ptr->next = NULL;

	/* remove it from lru list */
	if (ptr->lru_prev)
	{	/* ptr is not the first one */
		ptr->lru_prev->lru_next = ptr->lru_next;
	}else
	{	/* ptr is the first one */
		_rtl8651_pptpHdrCache_lru_head = ptr->lru_next;
	}
	if (ptr->lru_next)
	{	/* ptr is not the last one */
		ptr->lru_next->lru_prev = ptr->lru_prev;
	}else
	{	/* ptr is the last one */
		_rtl8651_pptpHdrCache_lru_tail = ptr->lru_prev;
	}
	ptr->lru_prev = NULL;
	ptr->lru_next = NULL;
	/* add it to unused list */
	ptr->next = _rtl8651_pptpHdrCache_unused;
	_rtl8651_pptpHdrCache_unused = ptr;
}

/* find entry for the specific ip header id/offset */
static inline int32 _rtl8651_pptpHdrCache_Find(uint16 ip_id, uint16 ip_off, uint32 *pSeq, uint32 *pAck, uint8 *pType)
{
	uint32 hashIdx = RTL8651_IP_ID_EASYHASH(0, (uint32)ip_off, ip_id, RTL8651_PPTPHDR_CACHE_HASH_TBLSIZE);
	_rtl8651_pptpHdrCacheEntry_t *ptr;

	assert(pSeq && pAck && pType);

	ptr = _rtl8651_pptpHdrCache_hashTbl[hashIdx];
	while (ptr)
	{
		if ((ptr->id == ip_id) && (ptr->offset == ip_off))
		{	/* entry found !*/
			*pSeq = ptr->seq;
			*pAck = ptr->ack;
			*pType = ptr->type;
			_rtl8651_pptpHdrCache_Del(ptr);
			return SUCCESS;
		}
		ptr = ptr->next;
	}
	return FAILED;
}

/*
	refill pptp/l2tp header before they are trapped to protocol stack

	FIXME: if the pptp/l2tp packet is fragmented, we CAN NOT process it!
*/
static inline int32 _rtl8651_refillHeader(int32 result, struct rtl_pktHdr *pkt, struct ip*pktip, rtl8651_tblDrv_vlanTable_t *svlan, uint16 srcPortList)
{
	uint8 wanType;
	rtl8651_tblDrv_routeTable_t *rt_t;
	rtl8651_tblDrv_pppTable_t *pppEntry_t;
	struct rtl_mBuf *mbuf;
	uint8 *data;
	struct rtl_pktHdr * pkthdr;
	struct ip*pip;

	pkthdr = pkt;
	pip = pktip;

	/* only all header for packet from loopback vlan and port */
	if (tunnel.valid != TRUE)
		goto out;

	if ((svlan->vid != tunnel.loopBackVid) || (srcPortList != tunnel.loopbackPort))
		goto out;

	/* get the default routing entry and check if its PPTP/L2TP */
	wanType = 0;
	rt_t = DrvTbl.route + (rtl8651_tblDrvPara.routingTableSize - 1);
	if ((rt_t->valid == FALSE) || (rt_t->ifrt_flag != FALSE))
		goto out;

	pppEntry_t = rt_t->rt_nhSession_t;
	//pppEntry_t = (rtl8651_tblDrv_pppTable_t *)rt_t->rt_nhstart_t->nexthop_t;
	if ((!pppEntry_t) || (!pppEntry_t->initiated))
		goto out;
	if (rt_t->netif_t->linkLayerType == RTL8651_LL_PPTP)
	{
		wanType = RTL8651_LL_PPTP;
	} else if (rt_t->netif_t->linkLayerType == RTL8651_LL_L2TP)
	{
		wanType = RTL8651_LL_L2TP;
	}else
	{	/* it's not PPTP or L2TP: we don't add header for it */
		goto out;
	}
	/* check source mac : it must from pptp server's mac */
	while (pkthdr)
	{
		mbuf = pkthdr->ph_mbuf;
		data = mbuf->m_data;

		assert(pppEntry_t->macInfo);

		if (memcmp(data+6, &(pppEntry_t->macInfo->macAddr.octet), 6))
			goto out;

		/* we modify portlist and vlanIdx here */
		pkthdr->ph_vlanIdx = rtl8651_vlanTableIndex(tunnel.wanVid);
		pkthdr->ph_portlist = tunnel.wanPort;


		if (wanType == RTL8651_LL_PPTP)
		{	/* refill PPTP header */
			if ((pppEntry_t->greBuf2_ == NULL) || (pppEntry_t->greOffset2_ == 0))
				goto out;
			if (((uint32)data - (uint32)(mbuf->m_extbuf)) >= pppEntry_t->greOffset2_)
			{
				uint16 pipLen;
				struct ip *ipBuf;
#ifndef RTL8651_VPNCOMPRESS_UNALIGNHANDLE
				struct ip *ipHdr;
#endif /* RTL8651_VPNCOMPRESS_UNALIGNHANDLE */
				struct wan_gre_header *greBuf;
				struct wan_gre_header *greHdr;
				uint16 pipId;
				uint16 pipOff;
				uint32 gre_seq;
				uint32 gre_ack;
				uint8 gre_type;
				int16 offset;
				uint8 *remain;
				int16 remainLen;
				uint8 *hdrRemain;

				/* get the sequence and type of gre header */
				pipId = ntohs(pip->ip_id);
				pipOff = ntohs(pip->ip_off);
				if (_rtl8651_pptpHdrCache_Find(pipId, pipOff, &gre_seq, &gre_ack, &gre_type) != SUCCESS)
				{	/* we assume there are no sequence/ack */
					assert(0);
					gre_seq = 0;
					gre_ack = 0;
					gre_type = 0;
				}
				/* 1. calculate original packet's checksum */
				/* ---> we DONT re-calculate checksum for original packet, because if packet sip/dip sport/dport
					is modified in forwarding engine and need to be trapped to procotol stack without modify
					checksum by forwarding engine, we can claim this packet must be useless and need to drop
					in protocol stack */

				assert(((uint32)pip - (uint32)data) == 14);

				/* 2. start encapsulation procedure */
				pipLen = ntohs(pip->ip_len);
				ipBuf = (struct ip *)(pppEntry_t->greBuf2_);
				greBuf = (struct wan_gre_header *)L4HDR_ADDR(ipBuf);
				remain = (uint8 *)((uint32)greBuf + sizeof(struct wan_gre_header));

				/* modify ip id/offset */
				ipBuf->ip_id = htons(ntohs(ipBuf->ip_id) + 1);
				ipBuf->ip_off = htons(ntohs(ipBuf->ip_off) & ~(IP_OFFMASK));

				/* get the remaining data length and pointer */
				if (!(greBuf->flags & 0x10))	/* S bit */
				{
					remain -= 4;
				}
				if (!(greBuf->ver& 0x80))	/* A bit */
				{
					remain -= 4;
				}

				remainLen = pppEntry_t->greOffset2_ - ((uint32)remain - (uint32)ipBuf);

				/* calculate the new header length must add to packet */
				offset = IPHDR_LEN(ipBuf) + sizeof(struct wan_gre_header) + remainLen;
				if (!(gre_type & RTL8651_PPTPHDR_SEQ_SET))
				{
					offset -= 4;
				}
				if (!(gre_type & RTL8651_PPTPHDR_ACK_SET))
				{
					offset -= 4;
				}

				/* length */
				ipBuf->ip_len = htons(offset + pipLen);
				greBuf->payload_len = htons(remainLen + pipLen);

				/* modify packet header */
				pkthdr->ph_len += offset;
				mbuf->m_len += offset;
				mbuf->m_data -= offset;

				/* add PPTP header */
				data = mbuf->m_data;	/* new packet head */

#ifdef RTL8651_VPNCOMPRESS_UNALIGNHANDLE
				/* calculate ip checksum */
				ipBuf->ip_sum = 0;
				ipBuf->ip_sum = htons(_rtl8651_ipChecksum(ipBuf));

				/* copy PPTP header */
				/* ip + {gre without seq/ack} */
				rtl8651_memcpy(data + 14, pppEntry_t->greBuf2_, (IPHDR_LEN(ipBuf) + (sizeof(struct wan_gre_header) - 8)));

				greHdr = (struct wan_gre_header *)(data+14+IPHDR_LEN(ipBuf));	/* fix unalignment */
				hdrRemain = (uint8*)greHdr + (sizeof(struct wan_gre_header) - 8);

				if (gre_type & RTL8651_PPTPHDR_SEQ_SET)
				{
					greHdr->flags |= 0x10;
					PKTSET_UINT32_UNALIGNED(gre_seq, &greHdr->seq);	/* fix unalignment */
					hdrRemain += 4;
				}else
				{
					greHdr->flags &= ~(0x10);
				}
				if (gre_type & RTL8651_PPTPHDR_ACK_SET)
				{
					greHdr->ver |= 0x80;
					if (gre_type & RTL8651_PPTPHDR_SEQ_SET)
					{	/* seq is set, too */
						PKTSET_UINT32_UNALIGNED(htonl(gre_ack - pppEntry_t->delta_), &greHdr->ack);	/* fix unalignment */
					}else
					{	/* seq is not set, so ack is set in sequence file */
						PKTSET_UINT32_UNALIGNED(htonl(gre_ack - pppEntry_t->delta_), &greHdr->seq);	/* fix unalignment */
					}
					hdrRemain += 4;
				}else
				{
					greHdr->ver &= ~(0x80);
				}
				/* copy remaining data */
				rtl8651_memcpy(hdrRemain, remain, remainLen);

#else	/* RTL8651_VPNCOMPRESS_UNALIGNHANDLE */

				/* copy PPTP header */
				/* ip + {gre without seq/ack} */
				rtl8651_memcpy(data + 14, pppEntry_t->greBuf2_, (IPHDR_LEN(ipBuf) + (sizeof(struct wan_gre_header) - 8)));

				ipHdr = (struct ip*)((uint32)data + 14);
				greHdr = (struct wan_gre_header *)L4HDR_ADDR(ipHdr);
				hdrRemain = (uint8*)greHdr + (sizeof(struct wan_gre_header) - 8);
				if (gre_type & RTL8651_PPTPHDR_SEQ_SET)
				{
					greHdr->flags |= 0x10;
					PKTSET_UINT32_UNALIGNED(gre_seq, &(greHdr->seq));
					hdrRemain += 4;
				}else
				{
					greHdr->flags &= ~(0x10);
				}
				if (gre_type & RTL8651_PPTPHDR_ACK_SET)
				{
					greHdr->ver |= 0x80;
					if (gre_type & RTL8651_PPTPHDR_SEQ_SET)
					{	/* seq is set, too */
						PKTSET_UINT32_UNALIGNED(htonl(gre_ack - pppEntry_t->delta_), &(greHdr->ack));
					}else
					{	/* seq is not set, so ack is set in sequence file */
						PKTSET_UINT32_UNALIGNED(htonl(gre_ack - pppEntry_t->delta_), &(greHdr->seq));
					}
					hdrRemain += 4;
				}else
				{
					greHdr->ver &= ~(0x80);
				}
				/* copy remaining data */
				rtl8651_memcpy(hdrRemain, remain, remainLen);

#endif	/* RTL8651_VPNCOMPRESS_UNALIGNHANDLE */


				/* generate mac */
				rtl8651_memcpy(data, &(svlan->macAddr), 6);	/* dmac == gateway mac */
				rtl8651_memcpy(data + 6, &(pppEntry_t->macInfo->macAddr), 6);	/* smac == l2tp server mac */
				*(uint16*)((uint32)data + 12) = htons(0x0800);
#ifndef RTL8651_VPNCOMPRESS_UNALIGNHANDLE
				/* calculate ip checksum */
				ipHdr->ip_sum = 0;
				ipHdr->ip_sum = htons(_rtl8651_ipChecksum(ipHdr));
#endif	/* RTL8651_VPNCOMPRESS_UNALIGNHANDLE */
			}
		}else
		{	/* refill L2TP header */
			if ((pppEntry_t->l2tpBuf2_ == NULL) || (pppEntry_t->l2tpOffset2_ == 0))
				goto out;
			if (((uint32)data - (uint32)(mbuf->m_extbuf)) >= pppEntry_t->l2tpOffset2_)
			{
				uint16 pipLen;
				struct ip *ipBuf;
				struct udphdr *udpBuf;

				/* 1. calculate original packet's checksum */
				/* ---> we DONT re-calculate checksum for original packet, because if packet sip/dip sport/dport
					is modified in forwarding engine and need to be trapped to procotol stack without modify
					checksum by forwarding engine, we can claim this packet must be useless and need to drop
					in protocol stack */

				/* 2. start encapsulation procedure */
				pipLen = ntohs(pip->ip_len);
				ipBuf = (struct ip *)(pppEntry_t->l2tpBuf2_);
				udpBuf = (struct udphdr*)L4HDR_ADDR(ipBuf);

				ipBuf->ip_id = htons(ntohs(ipBuf->ip_id) + 1);
				ipBuf->ip_off = htons(ntohs(ipBuf->ip_off) & ~(IP_OFFMASK));

				ipBuf->ip_len = htons(pipLen + pppEntry_t->l2tpOffset2_);
				udpBuf->uh_ulen = htons(pipLen + pppEntry_t->l2tpOffset2_ - IPHDR_LEN(ipBuf));

				/* check if Length bit is set or not */
				if (*(uint8 *)((uint32)udpBuf + sizeof(struct udphdr)) & 0x40)	/* 'L' bit is set */
				{
					uint16 *pLen = (uint16 *)((uint32)udpBuf + sizeof(struct udphdr) + 2);
					if ((pppEntry_t->l2tpBuf2_[pppEntry_t->l2tpOffset2_ - 2] == 0) &&
						(pppEntry_t->l2tpBuf2_[pppEntry_t->l2tpOffset2_ - 3] == 3))
					{	/* ppp header is not be compressed, it's 4 bytes */
						*pLen = htons(pipLen + 12);	/* l2tp(8) + ppp(4) */
					}else
					{	/* ppp header is compressed, it's 1 bytes */
						*pLen = htons(pipLen + 9);		/* l2tp(8) + ppp(1) */
					}
				}

				assert(((uint32)pip - (uint32)data) == 14);

				/* modify packet header */
				pkthdr->ph_len += pppEntry_t->l2tpOffset2_;
				mbuf->m_len += pppEntry_t->l2tpOffset2_;
				mbuf->m_data -= pppEntry_t->l2tpOffset2_;

				/* add L2TP header */
				data = mbuf->m_data;	/* new packet head */
				/* copy L2TP header */
				rtl8651_memcpy(data + 14, pppEntry_t->l2tpBuf2_, pppEntry_t->l2tpOffset2_);
				/* generate mac */
				rtl8651_memcpy(data, &(svlan->macAddr), 6);	/* dmac == gateway mac */
				rtl8651_memcpy(data + 6, &(pppEntry_t->macInfo->macAddr), 6);	/* smac == l2tp server mac */
				*(uint16*)((uint32)data + 12) = htons(0x0800);
				/* calculate ip/udp checksum */
				ipBuf = (struct ip *)(data + 14);
				udpBuf = (struct udphdr *)L4HDR_ADDR(ipBuf);
				ipBuf->ip_sum = 0;
				udpBuf->uh_sum = 0;
				udpBuf->uh_sum = htons(_rtl8651_tcpChecksum(ipBuf));
				ipBuf->ip_sum = htons(_rtl8651_ipChecksum(ipBuf));
			}
		}

		/* encapsulation done, check next header */
		if (result != FWDENG_QUEUED_TO_GW)
			goto out;
		/* next entry */
		pkthdr = pkthdr->ph_nextHdr;
		if (pkthdr)
		{
			assert(pkthdr->ph_mbuf && pkthdr->ph_mbuf->m_data && pkthdr->ph_iphdrOffset);
			pip = (struct ip *)((uint32)(pkthdr->ph_mbuf->m_data) + pkthdr->ph_iphdrOffset);
		}
	}
out:
	return SUCCESS;
}


/*
 *  <ghhuang>:
 *  This callback function is added in order to support Diffserv packet marking.
 *  It is registered currently in rtl8651_tblDrv.c
 */
int _rtl8651_markDiffserv( int32 ruleNo, struct rtl_pktHdr *pkthdr, struct ip *pip, void *userDefined )
{
//	printk("Rule No %d; Old %d; New %d\n", ruleNo, pip->ip_tos, *(uint8 *)userDefined);


	if (IPHDRLEN_IS_TWENTY(pip))
	{
		pip->ip_tos = *(uint8 *)userDefined;
	} else
	{
		uint8 oldTos, newTos;

		oldTos = pip->ip_tos;
		newTos = (*(uint8 *)userDefined);

		pip->ip_tos = newTos;

		_RTL8651_L3_ADJUST_CHECKSUM(newTos, oldTos, pip->ip_sum);
	}

	return FAILED;
}

#endif /* _RTL_PPTPL2TP_HDR_REFILL */


#define RTL8651_FWDENG_RETURN_STATISTIC_FUNCTION
#ifdef _RTL8651_FWDENG_RETURN_STATISTIC

void _rtl8651_fwdEngineRetValueStatisticInit(void)
{
	memset(rtl8651_returnValue, 0, TOTAL_RETURNVALUE);
}
void _rtl8651_fwdEngineRetValueStatisticReset(void)
{
	_rtl8651_fwdEngineRetValueStatisticInit();
}
void _rtl8651_fwdEngineRetValueStatisticUpdate(int32 retval)
{
	if (retval == FAILED)
	{
		rtl8651_returnValue[FWDENG_TOTALERRS + 1] ++;	/* failed */
		return;
	}
	if ((retval > 0) &&(retval & FWDENG_DROP))
	{
		rtl8651_returnValue[FWDENG_TOTALERRS] ++;	/* drop */
		retval &= ~FWDENG_DROP;
	}
	if (retval > 0 && retval < FWDENG_TOTALERRS)
	{
		rtl8651_returnValue[retval] ++;
		return;
	}
	rtl8651_returnValue[FWDENG_TOTALERRS + 2] ++;	/* other */
	return;
}
int32 _rtl8651_fwdEngineRetValueStaticDump(void)
{
	int i;
	printk("fwdEngine non-success return value Statistic:\n");
	for ( i = 0 ; i < FWDENG_TOTALERRS ; i ++)
	{
		if (rtl8651_returnValue[i])
		{
			printk("Return [%d] -- count: %d\n", i, rtl8651_returnValue[i]);
		}
	}
	printk("---other ---\n");
	/* failed */
	printk("FAILED: %d\n", rtl8651_returnValue[FWDENG_TOTALERRS + 1]);
	/* fwdeng_drop */
	printk("FWD DROP: %d\n", rtl8651_returnValue[FWDENG_TOTALERRS]);
	/* other */
	if (rtl8651_returnValue[FWDENG_TOTALERRS + 2])
	{
		printk("other return values: %d\n", rtl8651_returnValue[FWDENG_TOTALERRS + 2]);
	}
	return SUCCESS;
}
#endif /* _RTL8651_FWDENG_RETURN_STATISTIC */


#ifdef _RTL_TRAPPING_DISPATCH
#define RTL8651_TRAPPING_DISPATCH_FUNCTIONS

/*
	Init function of trapping packet dispatch system
*/
static int32 _rtl8651_trappingDispatchInit(rtl8651_fwdEngineInitPara_t *para)
{
	uint32 trappingDispatchTblSize, i;
	_rtl8651_trapping_dispatch_entry_t *ptr;

	/* init data structure */
	rtl8651_trappingDispatchSystem_inUsed = NULL;
	rtl8651_trappingDispatchSystem_freeHdr = NULL;
	bzero(&rtl8651_trappingDispatchSystem_cache, sizeof(rtl8651_trappingDispatchSystem_cache));

	trappingDispatchTblSize = para->trappingPktDispatchTblSize;

	/* disable trapping dispatch system */
	if (trappingDispatchTblSize == 0)
	{
		return SUCCESS;
	}

	rtl8651_trappingDispatchSystem_freeHdr =
		(_rtl8651_trapping_dispatch_entry_t *)rtlglue_malloc(trappingDispatchTblSize * sizeof(_rtl8651_trapping_dispatch_entry_t));

	/* memory allocation fail */
	if (rtl8651_trappingDispatchSystem_freeHdr == NULL)
	{
		return FAILED;
	}

	/* init all entries and data structure */	
	bzero(rtl8651_trappingDispatchSystem_freeHdr, (trappingDispatchTblSize * sizeof(_rtl8651_trapping_dispatch_entry_t)));

	for (i = 0, ptr = rtl8651_trappingDispatchSystem_freeHdr;
		i < (trappingDispatchTblSize - 1) ;
		i ++, ptr ++)
	{
		ptr->next = ptr + 1;
	}

	return SUCCESS;
}

/*
	Re-Init function of trapping packet dispatch system
*/
static void _rtl8651_trappingDispatchReinit(void)
{
	while (rtl8651_trappingDispatchSystem_inUsed)
	{
		#ifdef RTL8651_DEBUG
		uint32 delCnt;
		delCnt =
		#endif
			_rtl8651_removeTrappingDispatchEntry(rtl8651_trappingDispatchSystem_inUsed);

		#ifdef RTL8651_DEBUG
		assert(delCnt == 1);
		#endif
	}
}


/*	find the category of the specific packet in trapping dispatch
	accoding to the {l4Proto, srcIp, dstIp, srcPort, dstPort}.
*/
static _rtl8651_trapping_dispatch_entry_t *_rtl8651_findTrappingDispatchEntry(
	uint8 l4Proto,
	ipaddr_t srcIp,
	ipaddr_t dstIp,
	int32 srcPort,
	int32 dstPort,
	uint8 isExactlyMatch
)
{
	_rtl8651_trapping_dispatch_entry_t *ptr;

	ptr = rtl8651_trappingDispatchSystem_inUsed;

	if (isExactlyMatch == TRUE)
	{	/* only allow exactly match */
		while (ptr)
		{
			if ((l4Proto == ptr->l4Proto) &&			/* [both of them are "don't care"] is [both of them are 0], so the semantics is same as [they are the same] */
			     (srcIp == ptr->srcIp) &&			/* [both of them are "don't care"] is [both of them are 0], so the semantics is same as [they are the same] */
			     (dstIp == ptr->dstIp) &&			/* [both of them are "don't care"] is [both of them are 0], so the semantics is same as [they are the same] */
			     ((srcPort < 0 && ptr->srcPort < 0) || (srcPort == ptr->srcPort)) &&	/* both [both of them are "don't care"] or [they are the same] are OK */
			     ((dstPort < 0 && ptr->dstPort < 0) || (dstPort == ptr->dstPort)))		/* both [both of them are "don't care"] or [they are the same] are OK */
			{	/* entry found, we return it */
				return ptr;
			}
			ptr = ptr->next;
		}
	}else
	{	/* allow wide-card match */

		/* try to get from cache */
		if (	(l4Proto == rtl8651_trappingDispatchSystem_cache.l4Proto) &&
			(srcIp == rtl8651_trappingDispatchSystem_cache.srcIp) &&
			(dstIp == rtl8651_trappingDispatchSystem_cache.dstIp) &&
			(srcPort == rtl8651_trappingDispatchSystem_cache.srcPort) &&
			(dstPort == rtl8651_trappingDispatchSystem_cache.dstPort) &&
			rtl8651_trappingDispatchSystem_cache.ptr)
		{
			return rtl8651_trappingDispatchSystem_cache.ptr;
		}

		/* check data base */
		while (ptr)
		{
			if ( (ptr->l4Proto == 0 || l4Proto == ptr->l4Proto) &&
			     (ptr->srcIp == 0 || srcIp == ptr->srcIp) &&
			     (ptr->dstIp == 0 || dstIp == ptr->dstIp) &&
			     (ptr->srcPort < 0 || srcPort == ptr->srcPort) &&
			     (ptr->dstPort < 0 || dstPort == ptr->dstPort))
			{	/* entry found, we return it */

				/* update cache */
				rtl8651_trappingDispatchSystem_cache.l4Proto = l4Proto;
				rtl8651_trappingDispatchSystem_cache.srcIp = srcIp;
				rtl8651_trappingDispatchSystem_cache.dstIp = dstIp;
				rtl8651_trappingDispatchSystem_cache.srcPort = srcPort;
				rtl8651_trappingDispatchSystem_cache.dstPort = dstPort;
				rtl8651_trappingDispatchSystem_cache.ptr = ptr;

				return ptr;
			}
			ptr = ptr->next;
		}
	}
	/* not found : return NULL */
	return NULL;
}

/*
@func int32		| rtl8651_registerTrappingDispatchEntry	| Register/Update the category of Trapping Dispatch Entry
@parm uint8		| l4Proto		| The Layer 4 protocol of this new trapping dispatch entry
@parm ipaddr_t	| srcIp		| The source IP address that user want to dispatch ( 0 for widecard ).
@parm ipaddr_t	| dstIp		| The destination IP address that user want to dispatch ( 0 for widecard ).
@parm int32		| srcPort		| The source Port that user want to dispatch ( <0 for widecard ).
@parm int32		| dstPort		| The destination Port that user want to dispatch ( <0 for widecard ).
@parm uint16		| category	| The category of this trapping dispatch entry (RTL8651_DEFAULT_TRAPPING_DISPATCH_CATEGORY is default category).
@rvalue SUCCESS	| Register/Category update successfully.
@rvalue FAILED	| Failure occurs (might system has no any new entry).
@comm 
ROMEDRV allow user to dispatch packets trapped to upper-protocol-stack into several categories.
User can decide how to dispatch packet according to its l4Proto/srcIp/dstIp/srcPort(for TCP/UDP)/dstPort(for TCP/UDP),
give each with one specific category-ID, and register to Trapping Dispatch System in ROMEDRV.
Then ROMEDRV will dispatch packets (Only for IP packet with DMAC==gateway MAC yet) as user registered.
<p rtl8651_registerTrappingDispatchEntry> is the registration API for users to register the trapping
dispatch entry.

NOTE: If l4Proto is not IPPROTO_TCP(6) or IPPROTO_UDP(17), both srcPort/dstPort will be ignored.
 */
int32 rtl8651_registerTrappingDispatchEntry(
	uint8 l4Proto,
	ipaddr_t srcIp,
	ipaddr_t dstIp,
	int32 srcPort,
	int32 dstPort,
	uint16 category
)
{
	int32 retval;

	#ifdef RTL8651_DEBUG
	if (category == RTL8651_DEFAULT_TRAPPING_DISPATCH_CATEGORY)
	{	/* add this entry and the specific category same as default one */
		rtlglue_printf("[WARNING] %s(%d) The Category or this registration is [%d], same as the default one.",
			__FUNCTION__, __LINE__, category);
	}
	#endif

	if ( (l4Proto != IPPROTO_TCP) &&
	      (l4Proto != IPPROTO_UDP))
	{	/* disable Layer 4 Port check */
		srcPort = -1;
		dstPort = -1;
	}

	rtlglue_drvMutexLock(); /* Lock resource */

	retval = _rtl8651_registerTrappingDispatchEntry(l4Proto,
											 srcIp,
											 dstIp,
											 srcPort,
											 dstPort,
											 category);

	rtlglue_drvMutexUnlock(); /* Release resource */

	return retval;
}

/*
	register the trapping dispatch entry

	* SUCCESS: the entry exists or has been added successfully.
	* FAILED: add fail.
*/
static int32 _rtl8651_registerTrappingDispatchEntry(
	uint8 l4Proto,
	ipaddr_t srcIp,
	ipaddr_t dstIp,
	int32 srcPort,
	int32 dstPort,
	uint16 category
)
{
	_rtl8651_trapping_dispatch_entry_t *newEntry;
	int32 retval = FAILED;

	/* check if this entry is added into UDP trapping system */
	if ((newEntry = _rtl8651_findTrappingDispatchEntry(
				l4Proto,
				srcIp,
				dstIp,
				srcPort,
				dstPort,
				TRUE /* exactly match*/)) != NULL)
	{ /* entry exist, we update it's category */
		newEntry->category = category;
		retval = SUCCESS;
		goto out;
	}

	/* this entry is the new entry and system need to get one empty entry to store it */
	if (rtl8651_trappingDispatchSystem_freeHdr == NULL)
	{
		retval = FAILED;
		goto out;
	}

	/* add new entry */

	/* invalidate cache */
	bzero(&rtl8651_trappingDispatchSystem_cache, sizeof(rtl8651_trappingDispatchSystem_cache));

	/* get entry from freeList */
	newEntry = rtl8651_trappingDispatchSystem_freeHdr;
	rtl8651_trappingDispatchSystem_freeHdr = rtl8651_trappingDispatchSystem_freeHdr->next;

	/* add this new entry into inUsed list */
	newEntry->next = rtl8651_trappingDispatchSystem_inUsed;
	newEntry->prev = NULL;

	if (rtl8651_trappingDispatchSystem_inUsed)
	{
		rtl8651_trappingDispatchSystem_inUsed->prev = newEntry;
	}
	rtl8651_trappingDispatchSystem_inUsed = newEntry;

	/* set information of this newEntry */
	newEntry->l4Proto = l4Proto;
	newEntry->srcIp = srcIp;
	newEntry->dstIp = dstIp;
	newEntry->srcPort = srcPort;
	newEntry->dstPort = dstPort;
	newEntry->category = category;

	retval = SUCCESS;

out:
	return retval;
}

/*
	Remove the trapping dispatch entry,
	update pointers and put the entry back to freelist.
*/
static void _rtl8651_removeTrappingDispatchEntry(_rtl8651_trapping_dispatch_entry_t *delEntry)
{
	/* check cache for invalidation */
	if (rtl8651_trappingDispatchSystem_cache.ptr == delEntry)
	{
		bzero(&rtl8651_trappingDispatchSystem_cache, sizeof(rtl8651_trappingDispatchSystem_cache));
	}

	/* remove from database */
	if (delEntry->prev)
	{
		delEntry->prev->next = delEntry->next;
	}
	if (delEntry->next)
	{
		delEntry->next->prev = delEntry->prev;
	}
	if (rtl8651_trappingDispatchSystem_inUsed == delEntry)
	{
		rtl8651_trappingDispatchSystem_inUsed = delEntry->next;
	}

	/* put if back to freelist */
	delEntry->next = rtl8651_trappingDispatchSystem_freeHdr;
	rtl8651_trappingDispatchSystem_freeHdr = delEntry;
}

/*
@func int32		| rtl8651_unregisterTrappingDispatchEntry	| Remove the category of Trapping Dispatch Entry
@parm uint8		| l4Proto		| The Layer 4 protocol of the trapping dispatch entry which user want to remove.
@parm ipaddr_t	| srcIp		| The source IP address of the trapping dispatch entry which user want to remove.
@parm ipaddr_t	| dstIp		| The destination IP address of the trapping dispatch entry which user want to remove.
@parm int32		| srcPort		| The source Port of the trapping dispatch entry which user want to remove.
@parm int32		| dstPort		| The destination Port of the trapping dispatch entry which user want to remove.
@rvalue count		| Count of entry removed from trapping dispatch system.
@comm 
<p rtl8651_unregisterTrappingDispatchEntry> allow user to remove the specific entry from
trapping dispatch system. Note the all of the information given when calling this API must same as
the entry that user want to remove. Or different entry might be removed.

NOTE: If l4Proto is not IPPROTO_TCP(6) or IPPROTO_UDP(17), both srcPort/dstPort will be ignored.
 */
int32 rtl8651_unregisterTrappingDispatchEntry(
	uint8 l4Proto,
	ipaddr_t srcIp,
	ipaddr_t dstIp,
	int32 srcPort,
	int32 dstPort
)
{
	int32 retval;

	if ( (l4Proto != IPPROTO_TCP) &&
	      (l4Proto != IPPROTO_UDP))
	{	/* disable Layer 4 Port check */
		srcPort = -1;
		dstPort = -1;
	}

	rtlglue_drvMutexLock(); /* Lock resource */
	retval = _rtl8651_unregisterTrappingDispatchEntry(
				l4Proto,
				srcIp,
				dstIp,
				srcPort,
				dstPort);
	rtlglue_drvMutexUnlock(); /* Release resource */
	return retval;
}

/*
	remove the trapping dispatch entry.

	NOTE: return the deleted count of trapping dispatch entry.
*/
static int32 _rtl8651_unregisterTrappingDispatchEntry(
	uint8 l4Proto,
	ipaddr_t srcIp,
	ipaddr_t dstIp,
	int32 srcPort,
	int32 dstPort
)
{
	_rtl8651_trapping_dispatch_entry_t *delEntry;

	delEntry = _rtl8651_findTrappingDispatchEntry(
				l4Proto,
				srcIp,
				dstIp,
				srcPort,
				dstPort,
				TRUE);

	if (delEntry)
	{
		_rtl8651_removeTrappingDispatchEntry(delEntry);

		/* delete 1 entry */
		return 1;
	}

	/* don't delete any entry */
	return 0;

}

/*
	For the fragment-remaining packet, the port information is N/A, only the 'srcIp/dstIp/ipId' are used to
	determine the trapping category of this packet.
*/
static uint16 _rtl8651_getTrappingDispatchCategory(struct ip *pip)
{
	ipaddr_t srcIp, dstIp;
	int32 srcPort, dstPort;
	uint8 l4Proto;
	uint16 offset;
	uint16 category = RTL8651_DEFAULT_TRAPPING_DISPATCH_CATEGORY;
	_rtl8651_trapping_dispatch_entry_t *ptr;

	/* get Layer 3 information */
	srcIp = PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	dstIp = PKTGET_UINT32_UNALIGNED(&pip->ip_dst);
	offset = pip->ip_off;	/* DONT do Endian modification here, it will be done in IS_FRAGMENT_XXXX */
	l4Proto = pip->ip_p;

	if (IS_FRAGMENT_REMAIN(offset))
	{
		/* check fragment remaining flow */
		uint16 ipId;

		ipId = ntohs(pip->ip_id);

		if (_rtl8651_getFragCache(TRAP_DISPATCH_FRAG,
				srcIp,
				dstIp,
				ipId,
				(ipaddr_t *)(&category), NULL, NULL, NULL) == SUCCESS)
		{
			goto out;
		}

		/* framgnet packet without cache store : try to match L3 only information: invalidate port match */
		srcPort = -1;
		dstPort = -1;
		goto find;
	}

	/* get Layer 4 information */
	switch (l4Proto)
	{
		case IPPROTO_TCP:
		{
			struct tcphdr *tcpHdr;

			tcpHdr = (struct tcphdr *)L4HDR_ADDR(pip);
			srcPort = ntohs(tcpHdr->th_sport);
			dstPort = ntohs(tcpHdr->th_dport);
			break;
		}

		case IPPROTO_UDP:
		{
			struct udphdr *udpHdr;

			udpHdr = (struct udphdr *)L4HDR_ADDR(pip);
			srcPort = ntohs(udpHdr->uh_sport);
			dstPort = ntohs(udpHdr->uh_dport);
			break;
		}

		default:
			/* don't match port information */
			srcPort = -1;
			dstPort = -1;
	}

find:
	/* L4 information is available */
	if ((ptr = _rtl8651_findTrappingDispatchEntry(
				l4Proto,
				srcIp,
				dstIp,
				srcPort,
				dstPort,
				FALSE)) != NULL)
	{
		/* get the category of the entry */
		category = ptr->category;
	
		if (IS_FRAGMENT_FIRST(offset))
		{
			/* add to fragment cache */
			uint16 ipId;

			ipId = ntohs(pip->ip_id);

			_rtl8651_updateFragCache(TRAP_DISPATCH_FRAG,
				srcIp,
				dstIp,
				ipId,
				(ipaddr_t)category,
				0, NULL, 0);
		}
		goto out; /* flow found */
	}

	category = RTL8651_DEFAULT_TRAPPING_DISPATCH_CATEGORY;

out:
	return category;
}
#endif /* _RTL_TRAPPING_DISPATCH */

