/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : CLE command rtl8651_tblDrv_cmd.cf
* Abstract : 
* Author : Edward Jin-Ru Chen (jzchen@realtek.com.tw)               
* $Id: rtl8651_tblDrv_cmd.c,v 1.1.1.1 2006/07/26 05:38:24 blue Exp $
*/

#include "rtl_types.h"
#include <linux/string.h>
#ifdef RTL865X_DEBUG
#include <linux/proc_fs.h>
#endif
#include <linux/autoconf.h>
#include "types.h"
#include "assert.h"
#include "rtl_cle.h" 
#include "cle_utility.h" 
#include "rtl8651_tblDrv.h"
#include "rtl8651_tblDrvFwd.h"
#include "rtl8651_tblAsicDrv.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvStatistic.h"
#include "rtl8651_tblDrvPatch.h"
#include "rtl8651_alg_pptp.h"
#include "rtl8651_tblDrvStatistic.h"
#include "asicRegs.h"
#include "../swNic2.h"
#ifdef CONFIG_RTL865X_MULTICAST
#include "rtl8651_multicast.h"
#endif /* CONFIG_RTL865X_MULTICAST */
#ifdef CONFIG_RTL865X_IPSEC
#include "rtl8651_ipsec_local.h"
#endif /* CONFIG_RTL865X_IPSEC */
#ifdef CONFIG_RTL865X_ROMEPERF
#include "../romeperf.h"
#endif /* CONFIG_RTL865X_ROMEPERF */
#include "rtl8651_dos.h"
#ifdef RTL865X_DEBUG
#include <linux/proc_fs.h>
#endif
#include "rtl_glue.h"
#ifdef CONFIG_RTL865X_VOIP
#include "voip_support.h"
#endif

#include "rtl8651_dns_domainBlock.h"
extern rtl8651_PS_Action_Entry_t *rtl8651_ProtoStackActions;

#define STP_STATES 	5
static int8 _stpStates[STP_STATES][12]={
	"disabled", "blocking", "listening", "learning", "forwarding"
};

#define ACL_ACTIONS 3
static int8 _aclActions[ACL_ACTIONS][12]={
	"permit", "drop", "trap", 
};



static int32 _rtl8651_autoMdiMidxCmd(uint32 userId, int32 argc, int8 **saved)
{
	int32 size;
	uint32 port, isMdi;
	int8 *nextToken;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "enable") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "all") == 0)
			return rtl8651_autoCrossOverDetect(TRUE);
		port = U32_value(nextToken);
		return rtl8651_portAutoCrossOverDetect(port, TRUE);
	}
	else if (strcmp(nextToken, "disable") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "all") == 0)
			return rtl8651_autoCrossOverDetect(FALSE);
		port = U32_value(nextToken);
		return rtl8651_portAutoCrossOverDetect(port, FALSE);
	}
	else if (strcmp(nextToken, "select") == 0) 	{
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "mdi") == 0)
			isMdi = TRUE;
		else if (strcmp(nextToken, "mdix") == 0)
			isMdi = FALSE;
		else return FAILED;
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "all") == 0)
			return rtl8651_setMdiMdix(isMdi);
		port = U32_value(nextToken);
		return rtl8651_setPortMdiMdix(port, isMdi);
	}
	return FAILED;
}


static int32 _rtl8651_sysMtuCmd(uint32 userId, int32 argc, int8 **saved)
{
	int32 size;
	uint32 mtu;
	int8 *nextToken;


	cle_getNextCmdToken(&nextToken,&size,saved);
	mtu = U32_value(nextToken);
	rtl8651_setMtu(mtu);
	return SUCCESS;
}


static int32 _rtl8651_pbNatCmd(uint32 userId,  int32 argc,int8 **saved)
{
	int32 size;
	int8 *nextToken;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "enable") == 0) 
	{
		rtl8651_EnableProtocolBasedNAT(TRUE);
		return SUCCESS;
	}
	else if (strcmp(nextToken, "disable") == 0) 
	{
		rtl8651_EnableProtocolBasedNAT(FALSE);
		return SUCCESS;
	}
	else if (strcmp(nextToken, "add") == 0) 
	{
		ipaddr_t ExtIp, IntIp;
		uint32 Int32;
		uint8 protocol;
		
		cle_getNextCmdToken(&nextToken,&size,saved);
		Int32 = *(uint32 *)nextToken;
		protocol = Int32; // cast
		cle_getNextCmdToken(&nextToken,&size,saved);
		ExtIp = *((ipaddr_t *)nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved);
		IntIp = *((ipaddr_t *)nextToken);

		return rtl8651_addProtocolBasedNAT( protocol, ExtIp, IntIp );
	}
	else if (strcmp(nextToken, "del") == 0) 
	{
		ipaddr_t ExtIp, IntIp;
		uint32 Int32;
		uint8 protocol;
		
		cle_getNextCmdToken(&nextToken,&size,saved);
		Int32 = *(uint32 *)nextToken;
		protocol = Int32; // cast
		cle_getNextCmdToken(&nextToken,&size,saved);
		ExtIp = *((ipaddr_t *)nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved);
		IntIp = *((ipaddr_t *)nextToken);

		return rtl8651_delProtocolBasedNAT( protocol, ExtIp, IntIp );
	}
	else return FAILED;
}


static int32 _rtl8651_pppoePassthruCmd(uint32 userId,  int32 argc,int8 **saved)
{
	int32 size;
	int8 *nextToken;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "enable") == 0) 
	{
		rtl8651_EnablePppoePassthru(TRUE);
		return SUCCESS;
	}
	else if (strcmp(nextToken, "disable") == 0) 
	{
		rtl8651_EnablePppoePassthru(FALSE);
		return SUCCESS;
	}
	else if (strcmp(nextToken, "dropunknown") == 0) 
	{
		cle_getNextCmdToken(&nextToken,&size,saved);
		
		if (strcmp(nextToken, "enable") == 0) 
		{
			rtl8651_enableDropUnknownPppoePADT(TRUE);
			return SUCCESS;
		}
		else if (strcmp(nextToken, "disable") == 0) 
		{
			rtl8651_enableDropUnknownPppoePADT(FALSE);
			return SUCCESS;
		}
		else return FAILED;
	}
	else return FAILED;
}


static int32 _rtl8651_ipv6PassthruCmd(uint32 userId,  int32 argc,int8 **saved)
{
	int32 size;
	int8 *nextToken;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "enable") == 0) 
	{
		rtl8651_EnableIpv6Passthru(TRUE);
		return SUCCESS;
	}
	else if (strcmp(nextToken, "disable") == 0) 
	{
		rtl8651_EnableIpv6Passthru(FALSE);
		return SUCCESS;
	}
	else return FAILED;
}

static int32 _rtl8651_pptpPassthruCmd(uint32 userId,  int32 argc,int8 **saved) {
	int32 size;
	int8 *nextToken;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "dump") == 0) 
	{
		rtl8651_l4AliasPptpDump();
		return SUCCESS;
	}
	else return FAILED;
}

static int32 _rtl8651_romerealCmd(uint32 userId,  int32 argc,int8 **saved) 
{
#ifdef CONFIG_RTL865X_ROMEREAL
	int32 size;
	int8 *nextToken;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "dump") == 0) 
	{
		rtl8651_romerealDump();
		return SUCCESS;
	}
	else if (strcmp(nextToken, "start") == 0) 
	{
		rtl8651_romerealEnable( TRUE );
		return SUCCESS;
	}
	else if (strcmp(nextToken, "stop") == 0) 
	{
		rtl8651_romerealEnable( FALSE );
		return SUCCESS;
	}
	else if (strcmp(nextToken, "init") == 0) 
	{
		void *pBuf;
		uint32 bufSize;
		uint32 mode;

		cle_getNextCmdToken(&nextToken,&size,saved);
		pBuf = (void*) U32_value( nextToken );

		cle_getNextCmdToken(&nextToken,&size,saved);
		bufSize = U32_value( nextToken );

		cle_getNextCmdToken(&nextToken,&size,saved);
		mode = U32_value( nextToken );

		rtl8651_romerealInit( pBuf, bufSize, mode );
		return SUCCESS;
	}
	else if (strcmp(nextToken, "filter") == 0) 
	{
		uint32 filter;

		cle_getNextCmdToken(&nextToken,&size,saved);
		filter = U32_value( nextToken );

		rtl8651_romerealFilter( filter );
		return SUCCESS;
	}
	else if (strcmp(nextToken, "save") == 0) 
	{
		cle_getNextCmdToken(&nextToken,&size,saved);
		strtok( nextToken, "\r\n" );
		rtl8651_romerealSaveAsFile( nextToken );
		return SUCCESS;
	}
	else return FAILED;

#else/*CONFIG_RTL865X_ROMEREAL*/

	rtlglue_printf("\nROMEreal is not included in this image, please 'make menuconfig'\n\n");
	return FAILED;

#endif/*CONFIG_RTL865X_ROMEREAL*/
}


#ifdef CONFIG_RTL865X_ROMEPERF
#define _TEST_IRAM_SIZE_ 8192 /* IRAM 8K */
#define _TEST_DRAM_SIZE_ 4096 /* IRAM 4K */
void lx5280_setIram( void* IW_base, uint32 size );
void lx5280_getIram( void** pIW_base, void** pIW_top );
void lx5280_getCCTL( uint32* pCCTL );
void lx5280_setDram( void* DW_base, uint32 size );
void lx5280_getDram( void** pDW_base, void** pDW_top );
extern uint32 __iram;
extern uint32 __dram;

/* I-ram and D-ram test mechanism:
    If the I-ram (D-ram) is erroneously configured, reading at an illegal address would 
    NOT match the prevously written value. Note that this mechanism may cause system 
    crash while testing I-ram.*/
static int32 cmdTestIram(void)
{
    uint32 (*app_start)(void);
    uint8* allocedIram = rtlglue_malloc( _TEST_IRAM_SIZE_ * 2 ); /* to ensure alignment */
    uint8* alignedIram = (void*)((((uint32)allocedIram) + _TEST_IRAM_SIZE_ - 1 ) & ~(_TEST_IRAM_SIZE_-1) );
    uint8 *cachedIram = (uint8 *)(((uint32)alignedIram)&~0x20000000);
    uint8 *uncachedIram = (uint8 *)(((uint32)alignedIram)|0x20000000);
    void *IWBase, *IWTop;
    uint32 size = _TEST_IRAM_SIZE_;
    uint32 R2, expected;
    uint32 ret;
    uint32 CCTL;
    
    rtlglue_printf("\nStart Iram test. If there is any exception or lexra bus timeout, this configuration is illegal ...\n");

    lx5280_getIram(&IWBase,&IWTop);
	rtlglue_printf("Before test: iwbase=%p  iwtop=%p\n", IWBase, IWTop );

	lx5280_getCCTL( &CCTL );
    rtlglue_printf("CCTL=0x%08x\n",CCTL);

    rtlglue_printf("STATUS=0x%08x\n", read_c0_status() );

	rtlglue_printf("ld.script: __iram=%p\n",&__iram);

    app_start = (uint32 (*)(void)) (cachedIram);
    rtlglue_printf("app_start=%p\n",app_start);

    cli();
	/*******************************************************************************************
	 *  Fill test code
	 */
    bzero(uncachedIram, size);
    {
    	uint32 i;
    	for( i = 0; i < size; i+=4 )
    		* (uint32 *)( uncachedIram+i ) = 0x20420001;  /* opcode 'addi r2,r2,1' */
    }
    * (uint32 *) (uncachedIram           ) = 0x24020000; /* opcode 'addiu r2,zero,0x0000' */
    * (uint32 *) (uncachedIram + size - 8) = 0x03e00008; /* opcode 'jr $ra' */
    * (uint32 *) (uncachedIram + size - 4) = 0x20420001; /* opcode 'addi r2,r2,1' */ 
    lx5280_setIram(alignedIram, size);

    lx5280_getIram(&IWBase,&IWTop);
	rtlglue_printf("\nNEW: iwbase=%p  iwtop=%p\n\n", IWBase, IWTop );

    ret = SUCCESS;

	/*******************************************************************************************
	 *  Cached Address and meet IRAM
	 */
	{
		register uint32 i;
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_IRAM_1);
#endif
		for( i = 0 ; i < 10000; i++ )
		{
	    	app_start();
	    }
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfExitPoint(ROMEPERF_INDEX_IRAM_1);
#endif
	}
	R2 = app_start();
    expected = ( size / 4 ) - 2; /* exclude 'addiu r2,zero,0' and 'jr ra'. */
    if ( R2 != expected )  
    {
    	rtlglue_printf( "[FAILED]  expected: %d  returned: %d\n", expected, R2 );
    	ret = FAILED;
    }
	else
	{
		rtlglue_printf( "[PASSED]  Setting IRAM size to %d bytes seems OK.  Continue next test....\n", size );
	}

	//
	//  The above code only can test the IRAM size, but cannot test if IRAM can improve performance.
	//  The newly added code is to test if the IRAM really works!
	//  We will modify the code in SDRAM.
	//  Once the IRAM had beed fetched, the running result should be the same as above.
	//
	// Louis' method
    {
    	uint32 i;
    	for( i = 0; i < size; i+=4 )
    		* (uint32 *)( uncachedIram+i ) = 0x20420002;  /* opcode 'addi r2,r2,2' */
    }
    * (uint32 *) (uncachedIram           ) = 0x24020000; /* opcode 'addiu r2,zero,0x0000' */
    * (uint32 *) (uncachedIram + size - 8) = 0x03e00008; /* opcode 'jr $ra' */
    * (uint32 *) (uncachedIram + size - 4) = 0x20420002; /* opcode 'addi r2,r2,2' */ 

	/*******************************************************************************************
	 *  Cached Address and meet IRAM ( test if IRAM exists? )
	 */
    R2 = app_start();
    expected = ( size / 4 ) - 2; /* exclude 'addiu r2,zero,0' and 'jr ra'. */
    if ( R2 != expected )  
    {
    	rtlglue_printf( "[FAILED]  expected: %d  returned: %d\n", expected, R2 );
    	ret = FAILED;
    }
    else
    {
		rtlglue_printf( "[PASSED]  IRAM really works !!!\n" );
	}

	/*****************************************************************************************
	 *  The following code is used to test the performance executing directly from SDRAM.
	 *	Uncached Address and meet IRAM
	 */
    app_start = (uint32 (*)(void)) (uncachedIram);
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_IRAM_2);
#endif
    R2 = app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_IRAM_2);
#endif
    expected = ( ( size / 4 ) - 2 ) * 2; /* exclude 'addiu r2,zero,0' and 'jr ra', and each one increase two. */
    if ( R2 != expected )  
    {
    	rtlglue_printf( "[FAILED]  Performance test from SDRAM. expected: %d  returned: %d\n", expected, R2 );
    	ret = FAILED;
    }
    else
    {
	   	rtlglue_printf( "[PASSED]  Performance test from SDRAM is done.\n" );
	}

	// restore IRAM
    lx5280_setIram(&__iram, size);
    lx5280_getIram(&IWBase,&IWTop);
	rtlglue_printf("Restore: iwbase=%p  iwtop=%p\n", IWBase, IWTop );



	/*****************************************************************************************
	 *	test ICACHE
	 */
	size = 4096; /* ICACHE 4K */
    * (uint32 *) (uncachedIram + size - 8) = 0x03e00008; /* opcode 'jr $ra' */
    * (uint32 *) (uncachedIram + size - 4) = 0x20420002; /* opcode 'addi r2,r2,2' */ 
    
	/*****************************************************************************************
	 *	Cached Address and out of IRAM
	 */
    app_start = (uint32 (*)(void)) (cachedIram);
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_IRAM_3);
#endif
	/* 100* app_start() */
    R2 = app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
    app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
    app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
    app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
    app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
    app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
    app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
    app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
    app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
    app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_IRAM_3);
#endif
    expected = ( ( size / 4 ) - 2 ) * 2; /* exclude 'addiu r2,zero,0' and 'jr ra', and each one increase two. */
    if ( R2 != expected )  
    {
    	rtlglue_printf( "[FAILED]  Performance test (cached address and out of IRAM) expected: %d  returned: %d\n", expected, R2 );
    	ret = FAILED;
    }
    else
    {
	   	rtlglue_printf( "[PASSED]  Performance test (cached address and out of IRAM) is done.\n" );
	}

	/*****************************************************************************************
	 *	Uncached Address and out of IRAM
	 */
    app_start = (uint32 (*)(void)) (uncachedIram);
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_IRAM_4);
#endif
    R2 = app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();app_start();
#ifdef CONFIG_RTL865X_ROMEPERF
	rtl8651_romeperfExitPoint(ROMEPERF_INDEX_IRAM_4);
#endif
    expected = ( ( size / 4 ) - 2 ) * 2; /* exclude 'addiu r2,zero,0' and 'jr ra', and each one increase two. */
    if ( R2 != expected )  
    {
    	rtlglue_printf( "[FAILED]  Performance test (uncached address and out of IRAM) expected: %d  returned: %d\n", expected, R2 );
    	ret = FAILED;
    }
    else
    {
	   	rtlglue_printf( "[PASSED]  Performance test (uncached address and out of IRAM) is done.\n" );
	}
	sti();

	/*******************************************************************************************
	 *  Dump Results
	 */
	rtlglue_printf("\n");
	rtl8651_romeperfDump( ROMEPERF_INDEX_IRAM_1, ROMEPERF_INDEX_IRAM_4 );

	rtlglue_free( allocedIram );
    return ret;
}


static int32 cmdTestDram(void)
{
    uint32* allocedDram = rtlglue_malloc( _TEST_DRAM_SIZE_ * 2 ); /* to ensure alignment */
    uint32* alignedDram = (void*)((((uint32)allocedDram) + _TEST_DRAM_SIZE_ - 1 ) & ~(_TEST_DRAM_SIZE_-1) );
    uint32 *cachedDram = (uint32 *)(((uint32)alignedDram)&~0x20000000);
    uint32 *uncachedDram = (uint32 *)(((uint32)alignedDram)|0x20000000);
    void *DWBase, *DWTop;
    uint32 size = _TEST_DRAM_SIZE_;
    uint32 expected, sum;
    uint32 ret;
    uint32 CCTL;
    uint32 i;
    uint32* backupDram = rtlglue_malloc( _TEST_DRAM_SIZE_ ); /* backup original DRAM content */
    
    rtlglue_printf("\nStart Dram test.\n");

    lx5280_getDram(&DWBase,&DWTop);
	rtlglue_printf("Before test: dwbase=%p  dwtop=%p\n", DWBase, DWTop );

	lx5280_getCCTL( &CCTL );
    rtlglue_printf("CCTL=0x%08x\n",CCTL);

    rtlglue_printf("STATUS=0x%08x\n", read_c0_status() );

	rtlglue_printf("ld.script: __dram=%p\n",&__dram);

	cli();
	/*******************************************************************************************
	 *  backup DRAM
	 */
	for( i = 0; i < _TEST_DRAM_SIZE_/4; i++ )
	{
		backupDram[i] = ((uint32*)((uint32)DWBase+0x80000000))[i];
	}
	
	/*******************************************************************************************
	 *  Configure DRAM
	 */
    lx5280_setDram(alignedDram, size);
    lx5280_getDram(&DWBase,&DWTop);
    rtlglue_printf("alignedDram:%p\n", alignedDram );
	rtlglue_printf("NEW: dwbase=%p  dwtop=%p\n\n", DWBase, DWTop );
	
	/*******************************************************************************************
	 *  Fill test data
	 */
    bzero(uncachedDram, size);
    {
    	uint32 i;
    	for( i = 0; i < size/4; i++ )
    		uncachedDram[i] = 0x00000001;
    }

    ret = SUCCESS;

	/*******************************************************************************************
	 *  Cached Address and meet DRAM
	 */
	{
		register uint32 i, j, k;
		register uint32* addr = cachedDram;
		
		k = 0;
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_DRAM_1);
#endif
		for( i = 0; i < 10; i++ )
		{
			for( j = 0; j < _TEST_DRAM_SIZE_/4; j++ )
			{
				k += addr[j];
			}
		}
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfExitPoint(ROMEPERF_INDEX_DRAM_1);
#endif
		sum = k;
	}
    expected = ( size / 4 ) * 10;
    if ( sum != expected )  
    {
    	rtlglue_printf( "[FAILED]  expected: %d  returned: %d\n", expected, sum );
    	ret = FAILED;
    }
	else
	{
		rtlglue_printf( "[PASSED]  test DRAM performance OK.  Continue next test....\n" );
	}


	/*******************************************************************************************
	 *  Uncached Address and meet DRAM
	 */
	{
		register uint32 i, j, k;
		register uint32* addr = uncachedDram;
		
		k = 0;
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_DRAM_2);
#endif
		for( i = 0; i < 10; i++ )
		{
			for( j = 0; j < _TEST_DRAM_SIZE_/4; j++ )
			{
				k += addr[j];
			}
		}
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfExitPoint(ROMEPERF_INDEX_DRAM_2);
#endif
		sum = k;
	}
    expected = ( size / 4 ) * 10;
    if ( sum != expected )  
    {
    	rtlglue_printf( "[FAILED]  expected: %d  returned: %d\n", expected, sum );
    	ret = FAILED;
    }
	else
	{
		rtlglue_printf( "[PASSED]  test uncached DRAM performance OK.  Continue next test....\n" );
	}


	/*******************************************************************************************
	 *  restore DRAM
	 */
    lx5280_setDram(&__dram, size);
    lx5280_getDram(&DWBase,&DWTop);
	rtlglue_printf("Restore: dwbase=%p  dwtop=%p\n", DWBase, DWTop );
	for( i = 0; i < _TEST_DRAM_SIZE_/4; i++ )
	{
		((uint32*)((uint32)DWBase+0x80000000))[i] = backupDram[i];
	}
	
	/*******************************************************************************************
	 *  Refill test data
	 */
    bzero(uncachedDram, size);
    {
    	uint32 i;
    	for( i = 0; i < size/4; i++ )
    		uncachedDram[i] = 0x00000001;
    }

	/*******************************************************************************************
	 *  Cached Address and out of DRAM
	 */
	{
		register uint32 i, j, k;
		register uint32* addr = cachedDram;
		
		k = 0;
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_DRAM_3);
#endif
		for( i = 0; i < 100; i++ )
		{
			for( j = 0; j < _TEST_DRAM_SIZE_/4; j++ )
			{
				k += addr[j];
			}
		}
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfExitPoint(ROMEPERF_INDEX_DRAM_3);
#endif
		sum = k;
	}
    expected = ( size / 4 ) * 100;
    if ( sum != expected )  
    {
    	rtlglue_printf( "[FAILED]  expected: %d  returned: %d\n", expected, sum );
    	ret = FAILED;
    }
	else
	{
		rtlglue_printf( "[PASSED]  test DCACHE performance OK.  Continue next test....\n" );
	}


	/*******************************************************************************************
	 *  Uncached Address and out of DRAM
	 */
	{
		register uint32 i, j, k;
		register uint32* addr = uncachedDram;
		
		k = 0;
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_DRAM_4);
#endif
		for( i = 0; i < 10; i++ )
		{
			for( j = 0; j < _TEST_DRAM_SIZE_/4; j++ )
			{
				k += addr[j];
			}
		}
#ifdef CONFIG_RTL865X_ROMEPERF
		rtl8651_romeperfExitPoint(ROMEPERF_INDEX_DRAM_4);
#endif
		sum = k;
	}
    expected = ( size / 4 ) * 10;
    if ( sum != expected )  
    {
    	rtlglue_printf( "[FAILED]  expected: %d  returned: %d\n", expected, sum );
    	ret = FAILED;
    }
	else
	{
		rtlglue_printf( "[PASSED]  test uncached access performance OK.  Continue next test....\n" );
	}
	sti();	


	/*******************************************************************************************
	 *  Dump Results
	 */
	rtlglue_printf("\n");
	rtl8651_romeperfDump( ROMEPERF_INDEX_DRAM_1, ROMEPERF_INDEX_DRAM_4 );

	rtlglue_free( allocedDram );
	return ret;
}
#endif

static int32 _rtl8651_romeperfCmd(uint32 userId,  int32 argc,int8 **saved) 
{
#ifdef CONFIG_RTL865X_ROMEPERF
	int32 size;
	int8 *nextToken;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "dump") == 0) 
	{
		int32 from, to;
		
		if ( cle_getNextCmdToken(&nextToken,&size,saved)!=FAILED )
			from = U32_value(nextToken);
		else
			from = ROMEPERF_INDEX_MIN;
		if ( cle_getNextCmdToken(&nextToken,&size,saved)!=FAILED )
			to = U32_value(nextToken);
		else
			to = ROMEPERF_INDEX_MAX-1;

		rtl8651_romeperfDump( from, to );
		return SUCCESS;
	}
	else if (strcmp(nextToken, "reset") == 0) 
	{
		rtl8651_romeperfReset();
		return SUCCESS;
	}
	else if (strcmp(nextToken, "init") == 0) 
	{
		rtl8651_romeperfInit();
		return SUCCESS;
	}
	else if (strcmp(nextToken, "iram") == 0) 
	{
		return cmdTestIram();
	}
	else if (strcmp(nextToken, "dram") == 0) 
	{
		return cmdTestDram();
	}
	else if (strcmp(nextToken, "bmp") == 0) 
	{
		struct rtl_mBuf *mbufPtr;
		static uint16 jumpTbl[256];
		char *pattern = "free-quit-kit.how-quit-smoking.com/picture-of-someone-smoking-a-cigarette.how-quit-smoking.com";
		int i, patLen, loop = 1000;
		uint32 seed = 0x5566;
		char randChar[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz:/?&";

		/* generate URL Jump Table */
		mBuf_getBMjmpTable( pattern, jumpTbl, strlen( pattern ), TRUE );
		
		rtlglue_printf( "Performance test fot mBuf BMP algorithm.\n" );
		rtlglue_printf( "  Test Loop=%d\n", loop );
		rtlglue_printf( "  Random Seed=%d\n", seed );
		rtlglue_printf( "  pattern='%s'\n", pattern );

		/* allocate a mbuf */
		if((mbufPtr = mBuf_get(MBUF_DONTWAIT, MBUFTYPE_DATA, 1)) == NULL)
		{
			rtlglue_printf( "%s():%d mBuf_get() returns NULL.\n", __FUNCTION__, __LINE__ );
			return FAILED;
		}
		if(mBuf_getPkthdr(mbufPtr, MBUF_DONTWAIT) == (struct rtl_mBuf *) NULL)
		{
			mBuf_freeMbufChain(mbufPtr);
			rtlglue_printf( "%s():%d mBuf_getPkthdr() returns NULL.\n", __FUNCTION__, __LINE__ );
			return FAILED;
		}
		mbufPtr->m_len = 1500;

		for( patLen = 1; patLen < strlen(pattern); patLen+=8 )
		{
			rtl8651_romeperfReset();
			for( loop = 1000; loop > 0; loop-- )
			{
				/*rtlglue_printf( "  packet content='" );*/
				for( i = mbufPtr->m_len-1; i >= 0; i-- )
				{
					mbufPtr->m_data[i] = randChar[seed%(sizeof(randChar)-1)];
					/*rtlglue_printf( "%c", mbufPtr->m_data[i] );*/
					seed = seed * 0x13579bdf + 0xfdb97531;
				}
				/*rtlglue_printf( "'\n" );*/

				/* first, build Jump Table */
				mBuf_getBMjmpTable( pattern, jumpTbl, patLen , TRUE);

				rtl8651_romeperfEnterPoint(ROMEPERF_INDEX_BMP);
				/* Serach it ... */
				mBuf_BMpatternMatch(mbufPtr, 1500, pattern, patLen, jumpTbl, TRUE);
				rtl8651_romeperfExitPoint(ROMEPERF_INDEX_BMP);
			}
			
			/*******************************************************************************************
			 *  Dump Results
			 */
			rtlglue_printf( "patLen=%3d:\n", patLen );
			rtl8651_romeperfDump( ROMEPERF_INDEX_BMP, ROMEPERF_INDEX_BMP );
		}
		
		/* Free packet */
		mBuf_freeMbufChain( mbufPtr );
		
		return SUCCESS;
	}
	else if (strcmp(nextToken, "morethan4096") == 0) 
	{
		uint32 IntIp = ntohl( 0xc0a80101/*192.168.1.1*/ );
		uint16 IntPort = 1024;
		uint32 RemIp = ntohl( 0x8c710101/*140.113.1.1*/ );
		uint16 RemPort = 80;
		ipaddr_t ExtIp; // out
		uint16 ExtPort;     // out
		int32 ret, i;

		rtlglue_printf( "-------------------------------------------------------------------------\n" );
		rtlglue_printf( "We first flush all NAPT connection and then add %d new connections.\n", rtl8651_tblDrvPara.naptTcpUdpFlowSize );
		rtl8651_romeperfReset();
		rtl8651_flushAllNaptConnections();
		/* We first ensure NAPT table is full */
		for( i = 0; i < rtl8651_tblDrvPara.naptTcpUdpFlowSize; i++, IntPort++ )
		{
			ret = rtl8651_addNaptConnection( FALSE, TCP_ESTABLISHED_FLOW,
			                                 IntIp, IntPort, 
			                                 &ExtIp, &ExtPort,
			                                 RemIp, RemPort );
			if ( ret != SUCCESS )
				return FAILED;
		}
		rtl8651_romeperfDump( ROMEPERF_INDEX_NAPT_ADD, ROMEPERF_INDEX_NAPT_ADD );

		/* Then, add one more connection to test performance. */
		rtlglue_printf( "\n-------------------------------------------------------------------------\n" );
		rtlglue_printf( "Then, add another %d connections.\n", rtl8651_tblDrvPara.naptTcpUdpFlowSize );
		rtl8651_romeperfReset();
		for( i = 0; i < rtl8651_tblDrvPara.naptTcpUdpFlowSize; i++, IntPort++ )
		{
			ret = rtl8651_addNaptConnection( FALSE, TCP_ESTABLISHED_FLOW,
			                                 IntIp, IntPort, 
			                                 &ExtIp, &ExtPort,
			                                 RemIp, RemPort );
			if ( ret != SUCCESS )
				return FAILED;
		}
		rtl8651_romeperfDump( ROMEPERF_INDEX_NAPT_ADD, ROMEPERF_INDEX_NAPT_ADD );
		return SUCCESS;
	}
	else if (strcmp(nextToken, "pause") == 0) 
	{
		return rtl8651_romeperfPause();
	}
	else if (strcmp(nextToken, "resume") == 0) 
	{
		return rtl8651_romeperfResume();
	}
	else return FAILED;

#else/*CONFIG_RTL865X_ROMEPERF*/

	rtlglue_printf("\nROMEperf is not included in this image, please 'make menuconfig'\n\n");
	return FAILED;

#endif/*CONFIG_RTL865X_ROMEPERF*/
}


int32 readKeySeq( uint8 *input, uint32 *output, int32 maxOutputLen )
{
	int in, out, nibble;

	if ( !strncmp( input, "0x", 2 ) )
	{
		input += 2;
	}
	
	for( in = 0, out = 0, nibble = 0, output[0] = 0;
	     out < maxOutputLen; )
	{
		switch( input[in] )
		{
			case '_':
				in++; /* skip it. */
				break;
			case '0':case '1':case '2':case '3':case '4':case '5':case '6':case '7':case '8':case '9':
				output[out] = (output[out]<<4) | (input[in]-'0');
				nibble++;
				in++;
				break;
			case 'a':case 'b':case 'c':case 'd':case 'e':case 'f':
				output[out] = (output[out]<<4) | (input[in]-'a'+10);
				nibble++;
				in++;
				break;
			case 'A':case 'B':case 'C':case 'D':case 'E':case 'F':
				output[out] = (output[out]<<4) | (input[in]-'A'+10);
				nibble++;
				in++;
				break;
			case '\0':
				return SUCCESS;
			default:
				return FAILED;
		}
		if ( nibble == 8 )
		{
			nibble = 0;
			out++;
			if ( out < maxOutputLen )
				output[out] = '\0';
		}
	}
	
	return FAILED;
}

static int32 _rtl8651_ipsecCmd(uint32 userId,  int32 argc,int8 **saved) 
{
#ifdef CONFIG_RTL865X_IPSEC
	int32 size;
	int8 *nextToken;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "dump") == 0) 
	{
		_rtl8651_dumpIpsec();
		return SUCCESS;
	}
	else if (strcmp(nextToken, "init") == 0) 
	{
		_rtl8651_initIpsec();
		return SUCCESS;
	}
	else if (strcmp(nextToken, "flush") == 0) 
	{
		_rtl8651_flushIpsec();
		return SUCCESS;
	}
	else if (strcmp(nextToken, "spi") == 0) 
	{
		ipaddr_t edstIp;
		uint32 spi;
		uint32 proto;
		uint32 cryptoType;
		uint8 cryptoKey[RTL8651_MAX_IPSEC_KEY_LENGTH];
		uint32 authType;
		uint8 authKey[RTL8651_MAX_IPSEC_KEY_LENGTH];
		ipaddr_t srcIp, dstIp;

		/* spi'Security Parameter Index' af inet edst %ip'Effetive Destination IP address' spi %d'SPI' proto " */
		cle_getNextCmdToken(&nextToken,&size,saved); /* af */
		cle_getNextCmdToken(&nextToken,&size,saved); /* inet */
		cle_getNextCmdToken(&nextToken,&size,saved); /* edst */
		
		cle_getNextCmdToken(&nextToken,&size,saved); /* %ip */
		edstIp = *((ipaddr_t *)nextToken);

		cle_getNextCmdToken(&nextToken,&size,saved); /* spi */
		cle_getNextCmdToken(&nextToken,&size,saved); /* %d */
		spi = U32_value(nextToken);
		
		cle_getNextCmdToken(&nextToken,&size,saved); /* proto */
		cle_getNextCmdToken(&nextToken,&size,saved); /* esp|tun */
		if ( !strcmp( nextToken, "esp" ) )
		{
			/* { esp src %ip'Source IP address' esp {3des-md5-96|3des-sha1-96} enckey %s'Encryption Key' authkey %s'Authentication Key' | " */
			proto = IPSEC_ESP;
			
			cle_getNextCmdToken(&nextToken,&size,saved); /* src */
			cle_getNextCmdToken(&nextToken,&size,saved); /* %ip */
			srcIp = *((ipaddr_t *)nextToken);
			
			cle_getNextCmdToken(&nextToken,&size,saved); /* esp */
			cle_getNextCmdToken(&nextToken,&size,saved); /* 3des-md5-96|3des-sha1-96| des-md5-96|des-sha1-96| aes-md5-96|aes-sha1-96 */
			if ( strstr( nextToken, "3des" ) )
				cryptoType = IPSEC_CBC_3DES;
			else if ( strstr( nextToken, "des" ) )
				cryptoType = IPSEC_CBC_DES;
			else if ( strstr( nextToken, "aes" ) )
				cryptoType = IPSEC_CBC_AES256;
			else if ( strstr( nextToken, "aes128" ) )
				cryptoType = IPSEC_CBC_AES128;
			else if ( strstr( nextToken, "aes192" ) )
				cryptoType = IPSEC_CBC_AES192;
			else if ( strstr( nextToken, "aes256" ) )
				cryptoType = IPSEC_CBC_AES256;
			else
				return FAILED;
			if ( strstr( nextToken, "md5" ) )
				authType = IPSEC_HMAC_MD5;
			else if ( strstr( nextToken, "sha1" ) )
				authType = IPSEC_HMAC_SHA1;
			else
				return FAILED;
			
			cle_getNextCmdToken(&nextToken,&size,saved); /* enckey */
			cle_getNextCmdToken(&nextToken,&size,saved); /* %s */
			if ( readKeySeq( nextToken, (uint32*)cryptoKey, RTL8651_MAX_IPSEC_KEY_LENGTH ) != SUCCESS )
			{
				rtlglue_printf("parse cryptoKey error: %s\n", nextToken );
				return FAILED;
			}

			cle_getNextCmdToken(&nextToken,&size,saved); /* authkey */
			cle_getNextCmdToken(&nextToken,&size,saved); /* %s */
			if ( readKeySeq( nextToken, (uint32*)authKey, RTL8651_MAX_IPSEC_KEY_LENGTH ) != SUCCESS )
			{
				rtlglue_printf("parse authKey error: %s\n", nextToken );
				return FAILED;
			}

			dstIp = 0;
			return _rtl8651_addIpsecSpi( edstIp, spi, proto, srcIp, cryptoType, cryptoKey, authType, authKey, dstIp, -1, 64, IPSEC_AUTH_DATA_PRESENT, NULL );
		}
		else
		{
			/*   tun src %ip'Source IP address' dst %ip'Destination IP address' } |" */
			proto = IPSEC_IPIP;

			cle_getNextCmdToken(&nextToken,&size,saved); /* src */
			cle_getNextCmdToken(&nextToken,&size,saved); /* %ip */
			srcIp = *((ipaddr_t *)nextToken);
			
			cle_getNextCmdToken(&nextToken,&size,saved); /* dst */
			cle_getNextCmdToken(&nextToken,&size,saved); /* %ip */
			dstIp = *((ipaddr_t *)nextToken);

			cryptoType = authType = 0;
			
			return _rtl8651_addIpsecSpi( edstIp, spi, proto, srcIp, cryptoType, cryptoKey, authType, authKey, dstIp, -1, 64, IPSEC_AUTH_DATA_PRESENT, NULL );
		}
		
		return SUCCESS;
	}
	else if (strcmp(nextToken, "spigrp") == 0) 
	{
		/* spigrp'SPI Group' inet %ip'IP address 1' %d'SPI1' %s'Proto1' [inet %ip'IP address 2' %d'SPI2' %s'Proto2' "
		                    [inet %ip'IP address 3' %d'SPI3' %s'Proto3' [inet %ip'IP address 4' %d'SPI4' %s'Proto4']]] |" */
		ipaddr_t edstIp[4];
		uint32 spi[4];
		uint32 proto[4];
		int i;

		memset( edstIp, 0, sizeof(edstIp) );
		memset( spi, 0, sizeof(spi) );
		memset( proto, 0, sizeof(proto) );
		for( i = 0; i < 4; i++ )
		{
			if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) break; /* inet */

			if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) break; /* %ip */
			edstIp[i] = *((ipaddr_t *)nextToken);

			if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) break; /* spi */
			spi[i] = U32_value( nextToken );
			
			if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) break; /* proto */
			if ( !strcmp( nextToken, "esp" ) )
				proto[i] = IPSEC_ESP;
			else
				proto[i] = IPSEC_IPIP;
		}

		return _rtl8651_addIpsecSpiGrp( edstIp[0], spi[0], proto[0],
		                                edstIp[1], spi[1], proto[1],
		                                edstIp[2], spi[2], proto[2],
		                                edstIp[3], spi[3], proto[3],
		                                0, NULL );
	}
	else if (strcmp(nextToken, "eroute") == 0) 
	{
		/* eroute'Enhanced Route' add eraf inet src %net'Source Net' dst %net'Destination Net' said %s'said(tun0x1001@192.168.154.240)' } " */
		ipaddr_t srcNet, dstNet;
		int8 srcCIDR, dstCIDR;
		ipaddr_t edstIp;
		uint32 spi;
		char *first, *second, *ip0, *ip1, *ip2, *ip3;
		
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* add */
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* eraf */
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* inet */
		
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* src */
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* %ip */
		srcNet = *((ipaddr_t *)nextToken);
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* %d */
		srcCIDR = U32_value( nextToken );
		
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* dst */
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* %ip */
		dstNet = *((ipaddr_t *)nextToken);
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* %d */
		dstCIDR = U32_value( nextToken );
		
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* said */
		if ( cle_getNextCmdToken(&nextToken,&size,saved)==FAILED ) return FAILED; /* %s */
		nextToken+=3; /* skip 'tun' */
		if ( NULL == ( first = strtok( nextToken, "@" ) ) ) return FAILED;
		if ( NULL == ( second = strtok( NULL, "@" ) ) ) return FAILED;
		spi = simple_strtol( first, NULL, 0 );
		if ( NULL == ( ip0 = strtok( second, "." ) ) ) return FAILED;
		if ( NULL == ( ip1 = strtok( NULL, "." ) ) ) return FAILED;
		if ( NULL == ( ip2 = strtok( NULL, "." ) ) ) return FAILED;
		if ( NULL == ( ip3 = strtok( NULL, "." ) ) ) return FAILED;
		edstIp = (simple_strtol(ip0, NULL, 10 )<<24) |
		         (simple_strtol(ip1, NULL, 10 )<<16) |
		         (simple_strtol(ip2, NULL, 10 )<<8) |
		         (simple_strtol(ip3, NULL, 10 )) ;
		
		return _rtl8651_addIpsecERoute( srcNet, ~((((uint32)1)<<(32-srcCIDR))-1), 
		                                dstNet, ~((((uint32)1)<<(32-dstCIDR))-1),
		                                edstIp, spi, IPSEC_IPIP, 0, NULL );
	}
	else return FAILED;

#else/*CONFIG_RTL865X_IPSEC*/

	rtlglue_printf("\nIPSec is not included in this image, please 'make menuconfig'\n\n");
	return FAILED;

#endif/*CONFIG_RTL865X_IPSEC*/
}

static int _rtl8651_dnsCmd(uint32 userId,  int32 argc,int8 **saved)
{
	int32 size;
	int8 *nextToken;

	if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;	/* set */
	if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;	/* ttl-suppression */

	if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;

	if ( strcmp(nextToken, "enable") == 0 )
	{
		uint32 ttl = 0;
		if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;	/* TTL */

		if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;
		ttl = U32_value( nextToken );

		return (rtl8651_dns_setDnsTtlSuppression(TRUE, ttl) == TRUE)?SUCCESS:FAILED;

	}
	else if ( strcmp(nextToken, "disable") == 0 )
	{
		return (rtl8651_dns_setDnsTtlSuppression(FALSE, 0) == FALSE)?SUCCESS:FAILED;
	}
	else return FAILED;

}

static int32 _rtl8651_domainBlockCmd(uint32 userId,  int32 argc,int8 **saved)
{
	int32 size;
	int8 *nextToken;
	int8 add;
	ipaddr_t sip, smask;
	char *domain;
	rtl8651_domainBlockingInfo_t info;

	if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;

	if (strcmp(nextToken, "add") == 0)
	{
		add = TRUE;
	}
	else if (strcmp(nextToken, "del") == 0)
	{
		add = FALSE;
	}
	else return FAILED;

	if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;	/* src */
	if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;
	sip = *((ipaddr_t*)nextToken);
	if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;	/* mask */
	if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;
	smask = *((ipaddr_t*)nextToken);
	if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;	/* domain */
	if ( cle_getNextCmdToken(&nextToken, &size, saved) == FAILED ) return FAILED;
	domain = nextToken;
	info.sip = sip;
	info.sipMask = smask;

	if (add == TRUE)
	{
		return rtl8651_domainBlocking_addDomainBlockingEntry(domain, &info);
	} else
	{
		return rtl8651_domainBlocking_delDomainBlockingEntry(domain, &info);
	}
	

	return SUCCESS;
}

static int32 _rtl8651_macFilterCmd(uint32 userId,  int32 argc,int8 **saved)
{
	int32 size;
	int8 *nextToken;
	ether_addr_t macAddr, macMask;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "enable") == 0) 
		return rtl8651_enableMacAccessControl(TRUE);
	else if (strcmp(nextToken, "disable") == 0) 
		return rtl8651_enableMacAccessControl(FALSE);
	else if (strcmp(nextToken, "default") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "permit") == 0)
			return rtl8651_setMacAccessControlDefaultPermit(TRUE);
		else return rtl8651_setMacAccessControlDefaultPermit(FALSE);
	}
	else if (strcmp(nextToken, "add") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		macAddr = *((ether_addr_t *)nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved);
		macMask = *((ether_addr_t *)nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "permit") == 0)
			return rtl8651_addMacAccessControl(&macAddr, &macMask, TRUE);
		else return rtl8651_addMacAccessControl(&macAddr, &macMask, FALSE);
	}
	else if (strcmp(nextToken, "remove") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		macAddr = *((ether_addr_t *)nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved);
		macMask = *((ether_addr_t *)nextToken);
		return rtl8651_delMacAccessControl(&macAddr, &macMask);
	} 
	else if (strcmp(nextToken, "flush") == 0) 
		return rtl8651_flushMacAccessControl();
	else return FAILED;
}


static int32 _rtl8651_qosCmd(uint32 userId,  int32 argc,int8 **saved)
{
	int32 size;
	uint32 port, dscp;
	int8 *nextToken;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "rr-ratio") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "always-high") == 0)
			return rtl8651_setQueueScheduling(RTL8651_QOS_RR, RTL8651_RR_H_FIRST);
		else if (strcmp(nextToken, "h16_l1") == 0)
			return rtl8651_setQueueScheduling(RTL8651_QOS_RR, RTL8651_RR_H16_L1);
		else if (strcmp(nextToken, "h8_l1") == 0)
			return rtl8651_setQueueScheduling(RTL8651_QOS_RR, RTL8651_RR_H8_L1);
		else if (strcmp(nextToken, "h4_l1") == 0)
			return rtl8651_setQueueScheduling(RTL8651_QOS_RR, RTL8651_RR_H4_L1);
	} else if (strcmp(nextToken, "port") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		port = U32_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "high") == 0)
			return rtl8651_setPortPriorityQueue(port, TRUE);
		else if (strcmp(nextToken, "low") == 0)
			return rtl8651_setPortPriorityQueue(port, FALSE);
	} else if (strcmp(nextToken, "diffserv") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		dscp = U32_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "high") == 0)
			return rtl8651_setDiffServDSCP(dscp, TRUE);
		else if (strcmp(nextToken, "low") == 0)
			return rtl8651_setDiffServDSCP(dscp, FALSE);
	}
	return FAILED;
}


static int32 _rtl8651_flowControlCmd(uint32 userId,  int32 argc,int8 **saved)
{
	int32 size, retval=FAILED;
	uint32 port;
	int8 *nextToken;

	cle_getNextCmdToken(&nextToken,&size,saved);
	if (strcmp(nextToken, "threshold") == 0) {
		uint32 threshold;
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "inq-high") == 0) {
			cle_getNextCmdToken(&nextToken,&size,saved);
			threshold = U32_value(nextToken);
			return rtl8651_flowContrlThreshold(3, threshold);
		}
		else if (strcmp(nextToken, "outq-high") == 0) {
			cle_getNextCmdToken(&nextToken,&size,saved);
			threshold = U32_value(nextToken);
			return rtl8651_flowContrlThreshold(1, threshold);
		}
		else if (strcmp(nextToken, "inq-low") == 0) {
			cle_getNextCmdToken(&nextToken,&size,saved);
			threshold = U32_value(nextToken);
			return rtl8651_flowContrlThreshold(2, threshold);
		}
		else if (strcmp(nextToken, "outq-low") == 0) {
			cle_getNextCmdToken(&nextToken,&size,saved);
			threshold = U32_value(nextToken);
			return rtl8651_flowContrlThreshold(0, threshold);
		}
		else if (strcmp(nextToken, "inopt") == 0) {
			cle_getNextCmdToken(&nextToken,&size,saved);
			threshold = U32_value(nextToken);
			return rtl8651_flowContrlPrimeThreshold(1, threshold);
		}
		else if (strcmp(nextToken, "outopt") == 0) {
			cle_getNextCmdToken(&nextToken,&size,saved);
			threshold = U32_value(nextToken);
			return rtl8651_flowContrlPrimeThreshold(0, threshold);
		}
		return FAILED;
	}
	else {
		port = U32_value(nextToken);
		
		cle_getNextCmdToken(&nextToken,&size,saved);
		if (strcmp(nextToken, "enable") == 0) 
			retval = rtl8651_setFlowControl(port, TRUE);
		else if (strcmp(nextToken, "disable") == 0)
			retval = rtl8651_setFlowControl(port, FALSE);
		else if (strcmp(nextToken, "auto") == 0)
			retval = rtl8651_setFlowControl(port, 2);
		return retval;
	}
}


static void _rtl8651_phyRegConfig(uint32 address, uint32 data) {
 while ( (READ_MEM32(SWTACR) & ACTION_MASK) != ACTION_DONE );//Wait for command done
 WRITE_MEM32(TCR0, data);
 WRITE_MEM32(SWTAA, address);//Fill address
  WRITE_MEM32(SWTACR, ACTION_START | CMD_FORCE);//Activate add command
 while ( (READ_MEM32(SWTACR) & ACTION_MASK) != ACTION_DONE );//Wait for command done
}
 
static int32	_rtl8651_phyCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size, retval;
	int8 *nextToken;
	uint32 port;
	uint32 type, speed;
	
	cle_getNextCmdToken(&nextToken,&size,saved);
	port=U32_value(nextToken);

	cle_getNextCmdToken(&nextToken,&size,saved);
	if(!strcmp(nextToken, "loopback")){
		int32 flag;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(!strcmp(nextToken, "on"))
			flag=TRUE;
		else	flag=FALSE;
		if(rtl8651_setAsicEthernetPHYLoopback(port, flag)==FAILED)
			return FAILED;
		return SUCCESS;
	}
	else if (!strcmp(nextToken, "auto-negotiated")) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(!strcmp(nextToken, "100Full"))
			type = RTL8651_ETHER_AUTO_100FULL;
		else if(!strcmp(nextToken, "100Half"))
			type = RTL8651_ETHER_AUTO_100HALF;
		else if(!strcmp(nextToken, "10Full"))
			type = RTL8651_ETHER_AUTO_10FULL;
		else 	type = RTL8651_ETHER_AUTO_10HALF;
		if (rtl8651_setEthernetPortAutoNegotiation( port, TRUE, type) != SUCCESS)
			return FAILED;
		return SUCCESS;
	} 
	else if (!strcmp(nextToken, "force")) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(!strcmp(nextToken, "100Full")) {
			speed = 100;
			type = TRUE;
		}
		else if (!strcmp(nextToken, "100Half")) {
			speed = 100;
			type = FALSE;
		} 
		else if (!strcmp(nextToken, "10Full")) {
			speed = 10;
			type = TRUE;
		}
		else {
			speed = 10;
			type = FALSE;
		}
		if (rtl8651_setEthernetPortAutoNegotiation( port, FALSE, RTL8651_ETHER_AUTO_100FULL) != SUCCESS)
			return FAILED;
		if (rtl8651_setEthernetPortDuplexMode(port, type) != SUCCESS)
			return FAILED;
		if (rtl8651_setEthernetPortSpeed(port, speed) != SUCCESS)
			return FAILED;
		return SUCCESS;					
	}else if (!strcmp(nextToken, "cable-meter")) {
		int32 rx,tx;
		if((retval=rtl8651_asicEthernetCableMeter( port, &rx,&tx))!=SUCCESS)
			return retval;
		printk("Rx ");
		if(rx&0x400){
			rx&=0x1ff;
			printk("short at %d.%d meters away, ", rx>>1, rx&1?5:0);
		}else if(rx&0x200){
			rx&=0x1ff;
			printk("open at %d.%d meters away, ", rx>>1, rx&1?5:0);
		}else{
			printk("connected, ");
		}
		printk("Tx ");
		if(tx&0x400){
			tx&=0x1ff;
			printk("short at %d.%d meters away, ", tx>>1, tx&1?5:0);
		}else if(tx&0x200){
			tx&=0x1ff;
			printk("open at %d.%d meters away, ", tx>>1, tx&1?5:0);
		}else{
			printk("connected, ");
		}			
		return SUCCESS;
	}
	else if (!strcmp(nextToken, "power"))
	{
		cle_getNextCmdToken(&nextToken,&size,saved);
		
		if (!strcmp(nextToken, "up"))
		{
			_rtl8651_phyRegConfig( PORT0_PHY_CONTROL+port*0x20, READ_MEM32( PORT0_PHY_CONTROL ) & ~( 1 << 11 ) );
		}
		else if (!strcmp(nextToken, "down"))
		{
			_rtl8651_phyRegConfig( PORT0_PHY_CONTROL+port*0x20, READ_MEM32( PORT0_PHY_CONTROL ) | ( 1 << 11 ) );
		}
		else return FAILED;
	
		return SUCCESS;
	}
	return FAILED;
}



static int32	_rtl8651_stpCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size;
	int8 *nextToken;
	uint16 sid, port;
	uint8 portstate;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	sid=U16_value(nextToken);
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "create")){
		return rtl8651_addSpanningTreeInstance( sid);
	}else if(!strcmp(nextToken, "delete")){
		return rtl8651_delSpanningTreeInstance(sid);
	}else if(!strcmp(nextToken, "port")){
		cle_getNextCmdToken(&nextToken,&size,saved); 
		port=U16_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		portstate=U8_value(nextToken);
		return rtl8651_setSpanningTreeInstancePortState(sid, port, portstate);
	}else if(!strcmp(nextToken, "protocol")) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		if(!strcmp(nextToken, "enable"))
			return rtl8651_setSpanningTreeInstanceProtocolWorking(sid, TRUE);
		else //"disable"
			return rtl8651_setSpanningTreeInstanceProtocolWorking(sid, FALSE);
	}
	return FAILED;
}

static int32	_rtl8651_filterDBCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size;
	int8 *nextToken;
	uint16 fid, sid;
	uint32 portmask;
	ether_addr_t mac;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	fid=U16_value(nextToken);
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "create")){
		return rtl8651_addFilterDatabase( fid);
	}else if(!strcmp(nextToken, "delete")){
		return rtl8651_delFilterDatabase(fid);
	}else if(!strcmp(nextToken, "add")){
		uint32 type=0;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		mac=*((ether_addr_t *)nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(!strcmp(nextToken, "forward")){
			type =RTL8651_FORWARD_MAC ;
		}else if(!strcmp(nextToken, "dst-block")){
			type =RTL8651_DSTBLOCK_MAC ;
		}else if(!strcmp(nextToken, "src-block")){
			type =RTL8651_SRCBLOCK_MAC ;
		}else if(!strcmp(nextToken, "trap")){
			type =RTL8651_TOCPU_MAC ;
		}
		cle_getNextCmdToken(&nextToken,&size,saved); 
		portmask=U32_value(nextToken);
		return rtl8651_addFilterDatabaseEntry(fid, &mac,type,portmask);
		//return SUCCESS;
	}else if(!strcmp(nextToken, "remove")){
		cle_getNextCmdToken(&nextToken,&size,saved); 
		mac=*((ether_addr_t *)nextToken);
		return rtl8651_delFilterDatabaseEntry(fid, &mac);
	}else if(!strcmp(nextToken, "sid")){
		cle_getNextCmdToken(&nextToken,&size,saved); 
		sid=U16_value(nextToken);
		return rtl8651_specifyFilterDatabaseSpanningTreeInstance(fid, sid);
	}
	return FAILED;
}


static int32	_rtl8651_portCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size;
	int8 *nextToken;
	uint16 aggregator, pvid;
	uint32 port;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	aggregator=U32_value(nextToken);
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "pvid")){
		cle_getNextCmdToken(&nextToken,&size,saved); 
		pvid=U16_value(nextToken);
		return rtl8651_setPvid( aggregator, pvid);
	}else if(!strcmp(nextToken, "individual")){
		return rtl8651_setAggregatorIndividual(aggregator, TRUE);
	}else if(!strcmp(nextToken, "member-port")){
		cle_getNextCmdToken(&nextToken,&size,saved); 
		port=U32_value(nextToken);
		return rtl8651_setPortAggregator(port, aggregator);
	}
	return FAILED;
}

static int32	_rtl8651_vlanCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size, retval;
	int8 *nextToken, *value;
	uint16 vid, macAddrNumber, fid;
	ether_addr_t macAddr;
	uint32 port;

	cle_getNextCmdToken(&nextToken,&size,saved); 
	vid=U16_value(nextToken);
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "create")){
		return rtl8651_addVlan( vid);
	}else if(!strcmp(nextToken, "delete")){
		return rtl8651_delVlan(vid);
	}else{
		do {
			if(!strcmp(nextToken, "create")){
				return rtl8651_addVlan( vid);
			}else if(!strcmp(nextToken, "delete")){
				return rtl8651_delVlan(vid);
			}else if(!strcmp(nextToken, "base-mac")){
				cle_getNextCmdToken(&nextToken,&size,saved); 
				macAddr=*((ether_addr_t *)nextToken);
				cle_getNextCmdToken(&nextToken,&size,saved); 
				macAddrNumber=U16_value(nextToken);
				if((retval=rtl8651_assignVlanMacAddress( vid, &macAddr,  macAddrNumber))!=SUCCESS)
					return retval;
			}else if(!strcmp(nextToken, "max-macs")){
#if 0
				cle_getNextCmdToken(&nextToken,&size,saved); 
				macAddrNumber=U16_value(nextToken);
				if((retval=rtl8651_allocateVlanMacAddress(vid, macAddrNumber))!=SUCCESS)
					return retval;
#endif
			}else if(!strcmp(nextToken, "filter-db")){
				cle_getNextCmdToken(&nextToken,&size,saved); 
				fid=U16_value(nextToken);
				if((retval=rtl8651_specifyVlanFilterDatabase(vid, fid))!=SUCCESS)
					return retval;
			}else if(!strcmp(nextToken, "member-port")){
				cle_getNextCmdToken(&nextToken,&size,saved); 
				cle_getNextCmdToken(&value,&size,saved); 
				port=U16_value(value);
				if(!strcmp(nextToken, "add")){
					if((retval=rtl8651_addVlanPortMember( vid, port))!=SUCCESS)
						return retval;
				}else if(!strcmp(nextToken, "remove")){
					if((retval=rtl8651_delVlanPortMember(vid, port))!=SUCCESS)
						return retval;
				}else if(!strcmp(nextToken, "tagged")){
					if((retval=rtl8651_setVlanPortUntag(vid,  port, FALSE))!=SUCCESS)
						return retval;
				}
			} else if(!strcmp(nextToken, "fwdTxMirror")) {
				cle_getNextCmdToken(&nextToken,&size,saved); 
				if(!strcmp(nextToken, "enable")){
					if((retval=rtl8651_setVlanFwdTxMirror(vid, TRUE)) != SUCCESS)
						return retval;
				}
				else if(!strcmp(nextToken, "disable")) {
					if((retval=rtl8651_setVlanFwdTxMirror(vid, FALSE)) != SUCCESS)
						return retval;
				}
			}
		}while(cle_getNextCmdToken(&nextToken,&size,saved) !=FAILED);
		return SUCCESS;
	}}

static int32	_rtl8651_pppoeCmd(uint32 userId,  int32 argc,int8 **saved){
	printk("_rtl8651_pppoeCmd()..........\n");
	return SUCCESS;
}

static int32	_rtl8651_ipUnnumCmd(uint32 userId,  int32 argc,int8 **saved){
	printk("_rtl8651_ipUnnumCmd()...........\n");
	return SUCCESS;
}

static int32	_rtl8651_ifaceCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size;
	int8 *nextToken;
	int8 ifName[16];
	uint16 mtu;
	uint32 llType, llIndex;
	ipaddr_t ipAddr, ipMask;

	cle_getNextCmdToken(&nextToken,&size,saved); 
	strncpy(ifName,nextToken, sizeof(ifName));

	cle_getNextCmdToken(&nextToken,&size,saved); 
	if (!strcmp(nextToken, "mtu")) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		mtu = U16_value(nextToken);	
		return rtl8651_setNetMtu(ifName, mtu);
	} 
	else 	if(!strcmp(nextToken, "create")){
		return rtl8651_addNetworkIntf( ifName );
	}else if(!strcmp(nextToken, "delete")){
		return rtl8651_delNetworkIntf(ifName);
	} else if (!strcmp(nextToken, "acl")) {
		cle_getNextCmdToken(&nextToken,&size,saved); 

		if (!strcmp(nextToken, "def-acl")) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(!strcmp(nextToken, "permit"))
				return rtl8651_setDefaultAcl(ifName, RTL8651_ACL_PERMIT);
		else if(!strcmp(nextToken, "drop"))
				return rtl8651_setDefaultAcl(ifName, RTL8651_ACL_DROP);
			else if(!strcmp(nextToken, "cpu"))
				return rtl8651_setDefaultAcl(ifName, RTL8651_ACL_CPU);
		} else if (!strcmp(nextToken, "flush")) {
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if (!strcmp(nextToken, "Ingress")) {
				return rtl8651_flushAclRule(ifName, TRUE);
			} else if (!strcmp(nextToken, "Egress")) {
				return rtl8651_flushAclRule(ifName, FALSE);
			} else if (!strcmp(nextToken, "all")) {
				int32 retval;

				if ((retval=rtl8651_flushAclRule(ifName, TRUE)) != SUCCESS)
					return retval;
				if ((retval=rtl8651_flushAclRule(ifName, FALSE)) != SUCCESS)
					return retval;
				return SUCCESS;
			}
		}
	}else if(!strcmp(nextToken, "type")){
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(!strcmp(nextToken, "ether"))
			llType=RTL8651_LL_VLAN;
		else if(!strcmp(nextToken, "pppoe"))
			llType=RTL8651_LL_PPPOE;
		else  if(!strcmp(nextToken, "unnumbered"))
			llType=RTL8651_LL_WIPUNNUMBERED;
		else
			return FAILED;
		cle_getNextCmdToken(&nextToken,&size,saved); //ignore 'vid'
		cle_getNextCmdToken(&nextToken,&size,saved); 
		llIndex=U32_value(nextToken);	
		return rtl8651_specifyNetworkIntfLinkLayerType(ifName, llType,  llIndex);
	}else if(!strcmp(nextToken, "detach")){
		uint32 ifid, type;
		if(_rtl8651_getNetworkIntfLinkLayerType(ifName, &type, &ifid)==SUCCESS&&
			rtl8651_removeNetworkIntfLinkLayerType(ifName)==SUCCESS){
			//if(type == RTL8651_LL_PPPOE)	
			//	rtl_ppp_free(ifid);
			return SUCCESS;
		}
		printk("detach failed!!!\n");
	}else if(!strcmp(nextToken, "ip")){
		int32 add=FALSE;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(!strcmp(nextToken, "add")){
			add=TRUE;
		}else if(!strcmp(nextToken, "remove")){
			add=FALSE;
		}
		cle_getNextCmdToken(&nextToken,&size,saved); 
		ipAddr=*((ipaddr_t *)nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		ipMask=*((ipaddr_t *)nextToken);	
		if(add)
			return rtl8651_addIpIntf(ifName, ipAddr, ipMask);
		else
			return rtl8651_delIpIntf(ifName, ipAddr, ipMask);
	}else if(!strcmp(nextToken, "WAN")){
		return rtl8651_addExtNetworkInterface(ifName);
	}else if(!strcmp(nextToken, "LAN")){
		return rtl8651_delExtNetworkInterface(ifName);
#if 0
	}else if(!strcmp(nextToken, "dial")){
		uint32 type, vid;
		if(_rtl8651_getNetworkIntfLinkLayerType(ifName, &type, &vid)==SUCCESS &&	type >= RTL8651_LL_PPPOE){
			while(cle_getNextCmdToken(&nextToken,&size,saved) !=FAILED){
				uint32 pppoeid =U32_value(nextToken);
				//vid is not checked against pppoeid
				if(rtl_ppp_dial( pppoeid)){
					printk("Failed to dail pppoe %d on %s\n", pppoeid, ifName);
					return FAILED;
				}
			};
		}else
			return FAILED;
	}else if(!strcmp(nextToken, "hangup")){
		uint32 vid, type;
		if(_rtl8651_getNetworkIntfLinkLayerType(ifName, &type, &vid)==SUCCESS &&	type >= RTL8651_LL_PPPOE){
			while(cle_getNextCmdToken(&nextToken,&size,saved) !=FAILED){
				uint32 pppoeid =U32_value(nextToken);
				//vid is not checked against pppoeid
				if(rtl_ppp_hangup( pppoeid)){
					printk("Failed to hangup pppoe %d on %s\n", pppoeid, ifName);
					return FAILED;
				}else
					return SUCCESS;
			};
			return SUCCESS;
		}else
			return FAILED;
#endif		
	}
	return FAILED;
}

static int32	_rtl8651_arpCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size;
	int8 *nextToken, *ifName;
	uint32 port;
	ipaddr_t ip;
 	ether_addr_t macAddr;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "add")){
		cle_getNextCmdToken(&nextToken,&size,saved); 
		ip =*((ipaddr_t *)nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		macAddr=*((ether_addr_t *)nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		ifName=nextToken;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		port=U32_value(nextToken);	
		return rtl8651_addArp(ip, &macAddr, ifName, port);
	}else if(!strcmp(nextToken, "remove")){
		cle_getNextCmdToken(&nextToken,&size,saved); 
		ip =*((ipaddr_t *)nextToken);	
		return rtl8651_delArp(ip);
	}
	return FAILED;

}

static int32	_rtl8651_natCmd(uint32 userId,  int32 argc,int8 **saved){
	int32  size,add;
	int8 *nextToken;
	ipaddr_t extip, intip;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "add"))
		add=TRUE;
	else if(!strcmp(nextToken, "remove"))
		add=FALSE;
	else return FAILED;
		
	cle_getNextCmdToken(&nextToken,&size,saved); 
	extip =*((ipaddr_t *)nextToken);	
	cle_getNextCmdToken(&nextToken,&size,saved); 
	intip=*((ipaddr_t *)nextToken);	

	if(add)
		return rtl8651_addNatMapping(extip, intip);
	else
		return rtl8651_delNatMapping(extip, intip);
}

static int32	_rtl8651_serverPortCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size;
	int8 *nextToken, isTcp;
	ipaddr_t globalIp, localIp;
	uint16 globalPort, localPort;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "add")) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		if(!strcmp(nextToken, "tcp")) 
			isTcp = TRUE;
		else	if(!strcmp(nextToken, "udp")) 
			isTcp = FALSE;
		else 
			return FAILED;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		globalIp =*((ipaddr_t *)nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		globalPort =U16_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		localIp =*((ipaddr_t *)nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		localPort =U16_value(nextToken);
		return rtl8651_addNaptServerPortMapping(isTcp, globalIp, globalPort, localIp, localPort);
	}
	else if(!strcmp(nextToken, "remove")) {
		cle_getNextCmdToken(&nextToken,&size,saved);
		if(!strcmp(nextToken, "tcp")) 
			isTcp = TRUE;
		else	if(!strcmp(nextToken, "udp")) 
			isTcp = FALSE;
		else
			return FAILED;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		globalIp =*((ipaddr_t *)nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		globalPort =U16_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		localIp =*((ipaddr_t *)nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		localPort =U16_value(nextToken);
		return rtl8651_delNaptServerPortMapping(isTcp, globalIp, globalPort, localIp, localPort);
	}

	return FAILED;
}


static int32	_rtl8651_localserverCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size,add;
	int8 *nextToken;
	ipaddr_t serverip;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "add"))
		add=TRUE;
	else if(!strcmp(nextToken, "remove"))
		add=FALSE;
	else return FAILED;
		
	cle_getNextCmdToken(&nextToken,&size,saved); 
	serverip =*((ipaddr_t *)nextToken);	

	if(add)
		return rtl8651_addLocalServer(serverip);
	else
		return rtl8651_delLocalServer(serverip);
}



static int32	_rtl8651_routeCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size;
	int8 *nextToken;
	ipaddr_t dest, mask, gw;
 	int8 ifName[16];
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "add")){
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(!size){
			mask=dest=0; //add a default route
			cle_getNextCmdToken(&nextToken,&size,saved); //skip 'use' keyword
		}else{
			dest =*((ipaddr_t *)nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); //may be 'use' or network mask
			if(size){ //it's a network mask, a network route
				mask=*((ipaddr_t *)nextToken);	
				cle_getNextCmdToken(&nextToken,&size,saved); //skip 'use' keyword
			}else //it's a host route
				mask=0xffffffff;
		}
		cle_getNextCmdToken(&nextToken,&size,saved); //interface name
		strncpy(ifName,nextToken, sizeof(ifName));	
		cle_getNextCmdToken(&nextToken,&size,saved); //skip 'via' keyword
		cle_getNextCmdToken(&nextToken,&size,saved); 
		gw=*((ipaddr_t *)nextToken);	
		return rtl8651_addRoute(dest, mask, ifName, gw);
	}else if(!strcmp(nextToken, "remove")){
		cle_getNextCmdToken(&nextToken,&size,saved); 

		if(!size){
			mask=dest=0; //remove default route
		}else{
			dest =*((ipaddr_t *)nextToken);
			if(cle_getNextCmdToken(&nextToken,&size,saved)==FAILED) //delete a host route
				mask= 0xffffffff;
			else //delete a network route
				mask =*((ipaddr_t *)nextToken);
		}
		return rtl8651_delRoute(dest, mask);
	}
	return FAILED;
}

static int32	_rtl8651_naptCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size, retval;
	int8 *nextToken;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "timeout")){
		int8 *value;
		uint32 timeout=0;

		while(cle_getNextCmdToken(&nextToken,&size,saved) !=FAILED){
			cle_getNextCmdToken(&value,&size,saved); 
			if(size)
				timeout =U32_value(value);				
			if(!strcmp(nextToken, "adaptive")){
				if(size)
					_rtl8651_naptTimerAdaptiveAdjust(1, timeout, 0); //timeout is a  threshold for adaptive timeout adjust.
				else
					_rtl8651_naptTimerAdaptiveAdjust(0, 0, 0);
			}else if(!strcmp(nextToken, "smart-swap")){
				if(size)
					_rtl8651_naptSmartFlowSwapByPPS(1, timeout); //timeout is a  threshold for smart swapping
				else
					_rtl8651_naptSmartFlowSwapByPPS(0, 0);
			}else if(!strcmp(nextToken, "smart-swap-tcp")){
				_rtl8651_naptSmartFlowSwapByPPSTcp(timeout);
			}else if(!strcmp(nextToken, "smart-swap-udp")){
				_rtl8651_naptSmartFlowSwapByPPSUdp(timeout);
			}else if(!strcmp(nextToken, "smart-swap-cdr")){
				_rtl8651_naptSmartFlowSwapByPPSCdrouter(timeout);
			}else if(!strcmp(nextToken, "icmp")&&rtl8651_setNaptIcmpTimeout(timeout))
				return FAILED;
			else if(!strcmp(nextToken, "icmp-fast")&&rtl8651_setNaptIcmpFastTimeout(timeout))
				return FAILED;
			else if(!strcmp(nextToken, "udp")&&rtl8651_setNaptUdpTimeout(timeout))
				return FAILED;
			else if(!strcmp(nextToken, "tcp-long")&& rtl8651_setNaptTcpLongTimeout(timeout))
				return FAILED;
			else if(!strcmp(nextToken, "tcp-medium")&&rtl8651_setNaptTcpMediumTimeout(timeout))
				return FAILED;
			else if(!strcmp(nextToken, "tcp-short")&&rtl8651_setNaptTcpFastTimeout(timeout))
				return FAILED;
			else if(!strcmp(nextToken, "tcp-fin")&&rtl8651_setNaptTcpFinTimeout(timeout))
				return FAILED;
		}
		return SUCCESS;
	}else if(!strcmp(nextToken, "portrange")){		
		uint16 start, end;
		cle_getNextCmdToken(&nextToken,&size,saved); //first port
		start =U32_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); //skip 'to'
		cle_getNextCmdToken(&nextToken,&size,saved); // last port
		end =U32_value(nextToken);
		return rtl8651_setNaptTcpUdpPortRange(start, end); 
	}else if(!strcmp(nextToken, "flush")){
		retval = rtl8651_flushAllNaptConnections();
		//_rtl8651_tblDrvStatisticReset();
		return retval;
	}else if(!strcmp(nextToken, "autolearn")){
		cle_getNextCmdToken(&nextToken,&size,saved); //first port
		if(!strcmp(nextToken, "enable")){
			rtl8651_enableNaptAutoAdd(TRUE);
		}else{
			rtl8651_enableNaptAutoAdd(FALSE);
		}
		return SUCCESS;
	}else if(!strcmp(nextToken, "add")){
		#define NAPT_ICMP 	1
		#define NAPT_TCP	2
		#define NAPT_UDP	3
		int32 protocol=NAPT_TCP;
		ipaddr_t insideLocal, insideGlobal, dstIpAddr;
		uint16 insideLocalPort, insideGlobalPort, dstPort, insideLocalID, insideGlobalID;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(size){
			ipaddr_t newNaptIp, dmzhostIp;
			newNaptIp =*((ipaddr_t *)nextToken);
			if(cle_getNextCmdToken(&nextToken,&size,saved)==FAILED)
				return rtl8651_addNaptMapping(newNaptIp);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			dmzhostIp =*((ipaddr_t *)nextToken);
			return rtl8651_addDmzHost( newNaptIp, dmzhostIp);
		}else if(!strcmp(nextToken, "icmp"))
			protocol= NAPT_ICMP;
		else if(!strcmp(nextToken, "udp"))
			protocol= NAPT_UDP;
		else if(!strcmp(nextToken, "tcp"))
			protocol= NAPT_TCP;			


		if(protocol!=NAPT_ICMP){
			int32 retval;
			cle_getNextCmdToken(&nextToken,&size,saved); 
			insideLocal =*((ipaddr_t *)nextToken);//source IP
			cle_getNextCmdToken(&nextToken,&size,saved); 
			insideLocalPort =U16_value(nextToken);//source port

			insideGlobal =dstIpAddr=0;
			insideGlobalPort=dstPort=0;
			while(cle_getNextCmdToken(&nextToken,&size,saved) !=FAILED){
				if(strcmp(nextToken,"use")==0){
					//user specified which IP, port to use..
					cle_getNextCmdToken(&nextToken,&size,saved); 
					insideGlobal =*((ipaddr_t *)nextToken);
					cle_getNextCmdToken(&nextToken,&size,saved); 
					insideGlobalPort =U16_value(nextToken);
				}else if(strcmp(nextToken,"dst")==0){
					cle_getNextCmdToken(&nextToken,&size,saved); 
					dstIpAddr =*((ipaddr_t *)nextToken);
					cle_getNextCmdToken(&nextToken,&size,saved); 
					dstPort =U16_value(nextToken);
				}
			};

			retval= rtl8651_addNaptConnection(insideGlobal?TRUE:FALSE, protocol== NAPT_TCP? TCP_ESTABLISHED_FLOW:UDP_FLOW, 
				insideLocal,insideLocalPort,&insideGlobal,&insideGlobalPort,dstIpAddr,dstPort);
			if(insideGlobal&&retval==SUCCESS){
				int8 ipStr[sizeof "255.255.255.255"];
				struct in_addr ip;
				ip.s_addr= htonl(insideGlobal);
				inet_ntoa_r(*((ipaddr_t *)&ip), ipStr);
				printk("Entry added. translated IP is %s, port is %d\n",ipStr, insideGlobalPort);
			}
			return retval;
		}else{
			cle_getNextCmdToken(&nextToken,&size,saved); 
			insideLocal =*((ipaddr_t *)nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			insideLocalID =U16_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); //skip 'dst'
			cle_getNextCmdToken(&nextToken,&size,saved); 
			dstIpAddr =*((ipaddr_t *)nextToken);
			if(cle_getNextCmdToken(&nextToken,&size,saved)==SUCCESS){ //skip 'use' if present
				//user specified which IP to use..
				cle_getNextCmdToken(&nextToken,&size,saved); 
				insideGlobal =*((ipaddr_t *)nextToken);
				cle_getNextCmdToken(&nextToken,&size,saved); 
				insideGlobalID =U16_value(nextToken);
				return rtl8651_addNaptIcmpFlow(TRUE,insideLocal,insideLocalID,&insideGlobal,&insideGlobalID,dstIpAddr);
			}else{
				int32 retval= rtl8651_addNaptIcmpFlow(FALSE, insideLocal,insideLocalID,&insideGlobal,&insideGlobalID,dstIpAddr);
				if(retval==SUCCESS){
					int8 ipStr[sizeof"255.255.255.255"];
					struct in_addr ip;
					ip.s_addr=htonl(insideGlobal);
					inet_ntoa_r(*((ipaddr_t *)&ip), ipStr);
					printk("Entry added. translated IP is %s, ICMP ID is %d\n",ipStr, insideGlobalID);
				}
				return retval;
			}
		}
	}else if(!strcmp(nextToken, "remove")){
		int32 protocol=NAPT_TCP;
		ipaddr_t insideLocal,  dstIpAddr;
		uint16 insideLocalPort, dstPort, insideLocalID;

		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(size){
			ipaddr_t killNaptIp;
			cle_getNextCmdToken(&nextToken,&size,saved); 
			killNaptIp =*((ipaddr_t *)nextToken);

			if(cle_getNextCmdToken(&nextToken,&size,saved)==FAILED)
				return rtl8651_delNaptMapping(killNaptIp);
			return rtl8651_delDmzHost( killNaptIp);
		}else if(!strcmp(nextToken, "icmp"))
			protocol= NAPT_ICMP;
		else if(!strcmp(nextToken, "udp"))
			protocol= NAPT_UDP;
		else if(!strcmp(nextToken, "tcp"))
			protocol= NAPT_TCP;			

		if(protocol!=NAPT_ICMP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			insideLocal =*((ipaddr_t *)nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			insideLocalPort =U16_value(nextToken);
			dstIpAddr = dstPort=0;
			while(cle_getNextCmdToken(&nextToken,&size,saved) !=FAILED){
				cle_getNextCmdToken(&nextToken,&size,saved); 
				dstIpAddr =*((ipaddr_t *)nextToken);
				cle_getNextCmdToken(&nextToken,&size,saved); 
				dstPort =U16_value(nextToken);
			};

			return rtl8651_delNaptConnection( protocol== NAPT_TCP? TRUE:FALSE, insideLocal, insideLocalPort, dstIpAddr, dstPort);
		}else{
			cle_getNextCmdToken(&nextToken,&size,saved); 
			insideLocal =*((ipaddr_t *)nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			insideLocalID =U16_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); //skip 'dst'
			cle_getNextCmdToken(&nextToken,&size,saved); 
			dstIpAddr =*((ipaddr_t *)nextToken);
			return rtl8651_delNaptIcmpFlow(insideLocal,insideLocalID,dstIpAddr);
		}
	}
	return FAILED;
}

static int32	_rtl8651_miiCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size, phyAddr;
	int8 *nextToken;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(size){//config MII Phy address
		int32 retval;
		phyAddr = U32_value(nextToken);
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(!strcmp(nextToken, "phy"))
			retval=rtl8651_setAsicEthernetMII(phyAddr,P5_LINK_MII_PHY, TRUE);
		else
			retval=rtl8651_setAsicEthernetMII(phyAddr,P5_LINK_MII_MAC, TRUE);
		if(retval==0)
			printk("MII port enabled using phy address %d\n", phyAddr);
		return retval;
	}else{
		int32 phyAddr;
		rtl8651_getAsicEthernetMII(&phyAddr);
		if(phyAddr>=0){
			int8 linkUp;
			 int8  autoNegotiation;
			 uint32 advCapability,speed;
			 int8 fullDuplex;
			 int32 loopback;
			rtl8651_getAsicEthernetLinkStatus(5, &linkUp);
			rtl8651_getAsicEthernetPHY(5, &autoNegotiation, &advCapability, &speed, &fullDuplex);
			rtl8651_getAsicEthernetPHYLoopback(5, &loopback);
			printk("MII port enabled using phy address %d, link is %s\n", phyAddr, linkUp?"Up": "Down");
			printk("AutoNeg: %s, " ,autoNegotiation? "Yes":"No");
			printk("Speed: %s, " ,speed? "100M":"10M");
			printk("%s dulplex " ,fullDuplex? "Full":"Half");
			if(loopback)
				printk(" loopbacked\n");
			else
				printk("\n");
		}else{
			printk("MII port is not enabled\n");
		}
	}
	return SUCCESS;
}
static int32	_rtl8651_macACLCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size;
	int8 *nextToken;
	int8 ifName[16];
	int8 action;
	uint16 ingress=FALSE, add=FALSE;
	uint8 mask[]={ 0xff,0xff,0xff,0xff,0xff,0xff};
	rtl8651_tblDrvAclRule_t acl;
	
	memset(&acl, 0, sizeof(rtl8651_tblDrvAclRule_t));
	cle_getNextCmdToken(&nextToken,&size,saved); 
	strncpy(ifName,nextToken, sizeof(ifName));
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "add"))
		add=TRUE;
	else if(!strcmp(nextToken, "remove"))
		add=FALSE;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if (!strcmp(nextToken, "permit"))
		action = RTL8651_ACL_PERMIT;
	else if (!strcmp(nextToken, "drop"))
		action = RTL8651_ACL_DROP;
	else action = RTL8651_ACL_CPU;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "ingress"))
		ingress=TRUE;
	else if(!strcmp(nextToken, "egress"))
		ingress=FALSE;

	while(cle_getNextCmdToken(&nextToken,&size,saved) !=FAILED){
next:
		if(!strcmp(nextToken, "smac")){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.srcMac_ = *((ether_addr_t *)nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.srcMacMask_ =*((ether_addr_t *)nextToken);
			else {
				acl.srcMacMask_ =*((ether_addr_t *)&mask[0]);
				goto next;
			}
		}else if(!strcmp(nextToken, "dmac")){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.dstMac_ = *((ether_addr_t *)nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.dstMacMask_ =*((ether_addr_t *)nextToken);
			else {
				acl.dstMacMask_ =*((ether_addr_t *)&mask[0]);
				goto next;
			}

		}else if(!strcmp(nextToken, "ethtype")){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.typeLen_ = U16_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size) 
				acl.typeLenMask_ = U16_value(nextToken);
			else {
				acl.typeLenMask_ =0xffff;
				goto next;
			}
		}
	};
	acl.ruleType_ = RTL8651_ACL_MAC;
	acl.actionType_ = action;
	if(add)
		return rtl8651_addAclRule(ifName, ingress, &acl);
	else
		return rtl8651_delAclRule(ifName, ingress, &acl);}

static int32	_rtl8651_ipACLCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size;
	int8 *nextToken;
	int8 ifName[16], isRange=FALSE;
	uint16 ingress=FALSE, add=FALSE, ruleType=0;
	rtl8651_tblDrvAclRule_t acl;

	memset(&acl, 0, sizeof(rtl8651_tblDrvAclRule_t));
	cle_getNextCmdToken(&nextToken,&size,saved); 
		strncpy(ifName, nextToken, sizeof(ifName));
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "add"))
		add=TRUE;
	else if(!strcmp(nextToken, "remove"))
		add=FALSE;
	cle_getNextCmdToken(&nextToken,&size,saved); //operation type, used later
	if (!strcmp(nextToken, "permit"))
		acl.actionType_= RTL8651_ACL_PERMIT;
	else if (!strcmp(nextToken, "drop"))
		acl.actionType_= RTL8651_ACL_DROP;
	else if (!strcmp(nextToken, "drop-notify"))
		acl.actionType_ = RTL8651_ACL_DROP_NOTIFY;
	else acl.actionType_= RTL8651_ACL_CPU;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if(!strcmp(nextToken, "ingress"))
			ingress=TRUE;
		else if(!strcmp(nextToken, "egress"))
			ingress=FALSE;

	ruleType=RTL8651_ACL_IP;	
	while(cle_getNextCmdToken(&nextToken,&size,saved) !=FAILED){
		
next:
		if (strcmp(nextToken, "range") ==0 ) {
			isRange = TRUE;
		}
		else if(!strcmp(nextToken, "sip")){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.srcIpAddr_ = *((ipaddr_t *)nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.srcIpAddrMask_ = *((ipaddr_t *)nextToken);
			else {
				acl.srcIpAddrMask_ = 0x00;//0xffffffff;
				goto next;
			}
		}else if(!strcmp(nextToken, "dip")){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.dstIpAddr_ = *((ipaddr_t *)nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.dstIpAddrMask_ = *((ipaddr_t *)nextToken);
			else {
				acl.dstIpAddrMask_ = 0x00;//0xffffffff;
				goto next;
			}
		}else if(!strcmp(nextToken, "tos")){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.tos_ = U8_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.tosMask_ = U8_value(nextToken);
			else{
				acl.tosMask_ = 0x00;//0xff;
				goto next;
			}
		}else if(!strcmp(nextToken, "ip")){
			ruleType=RTL8651_ACL_IP;
		}else if(!strcmp(nextToken, "icmp")){
			ruleType=RTL8651_ACL_ICMP;
		}else if(!strcmp(nextToken, "igmp")){
			ruleType=RTL8651_ACL_IGMP;
		}else if(!strcmp(nextToken, "tcp")){
			ruleType=RTL8651_ACL_TCP;
			acl.tcpSrcPortLB_ = 0;
			acl.tcpSrcPortUB_ = 65535;
			acl.tcpDstPortLB_ = 0;
			acl.tcpDstPortUB_ = 65535;
		}else if(!strcmp(nextToken, "udp")){
			ruleType=RTL8651_ACL_UDP;
			acl.udpSrcPortLB_ = 0;
			acl.udpSrcPortUB_ = 65535;
			acl.udpDstPortLB_ = 0;
			acl.udpDstPortUB_ = 65535;
		}else if(!strcmp(nextToken, "ipproto")&&ruleType==RTL8651_ACL_IP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.ipProto_= U8_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.ipProtoMask_ = U8_value(nextToken);
			else{
				acl.ipProtoMask_ = 0x00;//0xff;
				goto next;
			}
		}else if(!strcmp(nextToken, "ipflag")&&ruleType==RTL8651_ACL_IP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.ipFlag_ = U8_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.ipFlagMask_ = U8_value(nextToken);
			else{ 
				acl.ipFlagMask_ = 0x00;//0xff;
				goto next;
			}
		}else if(!strcmp(nextToken, "icmptype")&&ruleType==RTL8651_ACL_ICMP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.icmpType_ = U8_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.icmpTypeMask_ = U8_value(nextToken);
			else{
				acl.icmpTypeMask_ = 0x00;//0xff;
				goto next;
			}
		}else if(!strcmp(nextToken, "code")&&ruleType==RTL8651_ACL_ICMP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.icmpCode_ = U8_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.icmpCodeMask_ = U8_value(nextToken);
			else{
				acl.icmpCodeMask_ = 0x00;//0xff;
				goto next;
			}
		}else if(!strcmp(nextToken, "igmptype")&&ruleType==RTL8651_ACL_IGMP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.igmpType_ = U8_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.igmpTypeMask_ = U8_value(nextToken);
			else {
				acl.igmpTypeMask_ = 0x00;//0xff;
				goto next;
			}
		}else if(!strcmp(nextToken, "tcpflag")&&ruleType==RTL8651_ACL_TCP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.tcpFlag_= U8_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.tcpFlagMask_=U8_value(nextToken);
			else {
				acl.tcpFlagMask_=0x00;//0xff;
				goto next;
			}
		}else if(!strcmp(nextToken, "tcpflag")&&ruleType==RTL8651_ACL_TCP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.tcpFlag_ = U8_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.tcpFlagMask_ = U8_value(nextToken);
			else {
				acl.tcpFlagMask_ = 0x00;//0xff;
				goto next;
			}

		}else if(!strcmp(nextToken, "sport")&&ruleType==RTL8651_ACL_TCP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.tcpSrcPortLB_= U16_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.tcpSrcPortUB_=U16_value(nextToken);
			else {
				acl.tcpSrcPortUB_=acl.tcpSrcPortLB_;
				goto next;
			}

		}else if(!strcmp(nextToken, "dport")&&ruleType==RTL8651_ACL_TCP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.tcpDstPortLB_= U16_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.tcpDstPortUB_=U16_value(nextToken);
			else {
				acl.tcpDstPortUB_=acl.tcpSrcPortLB_;
				goto next;
			}

		}else if(!strcmp(nextToken, "sport")&&ruleType==RTL8651_ACL_UDP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.udpSrcPortLB_= U16_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.udpSrcPortUB_=U16_value(nextToken);
			else {
				acl.udpSrcPortUB_=acl.udpSrcPortLB_;
				goto next;
			}

		}else if(!strcmp(nextToken, "dport")&&ruleType==RTL8651_ACL_UDP){
			cle_getNextCmdToken(&nextToken,&size,saved); 
			acl.udpDstPortLB_= U16_value(nextToken);
			cle_getNextCmdToken(&nextToken,&size,saved); 
			if(size)
				acl.udpDstPortUB_=U16_value(nextToken);
			else {
				acl.udpDstPortUB_=acl.udpSrcPortLB_;
				goto next;
			}
		}
	};

	acl.ruleType_ = ruleType;
	if(add){
		if (isRange == FALSE)
			return rtl8651_addAclRule(ifName, ingress, &acl);
		else return rtl8651_addAclRuleExt(ifName, ingress, &acl);
	}else{
		if (isRange == FALSE)
			return rtl8651_delAclRule(ifName, ingress, &acl);
		else rtl8651_delAclRuleExt(ifName, ingress, &acl);
	}
	return SUCCESS;
}


static int32	_rtl8651_algTrapCmd(uint32 userId,  int32 argc,int8 **saved){
	int32 size;
	int8 *nextToken;
	int32  add=FALSE, isTcp=FALSE, isClient=0, isServer=0;
	uint16 low, high;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "add"))
		add=TRUE;
	else if(!strcmp(nextToken, "remove"))
		add=FALSE;
	else return FAILED;
		
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(!strcmp(nextToken, "tcp")){
		isTcp=TRUE;
		cle_getNextCmdToken(&nextToken,&size,saved); 
	}
	if(!strcmp(nextToken, "client"))
		isClient=TRUE;
	else if(!strcmp(nextToken, "server"))
		isServer=TRUE;
	else if(!strcmp(nextToken, "both")){
		isClient = isServer = TRUE;
	}
	cle_getNextCmdToken(&nextToken,&size,saved); //skip 'port' token
	
	cle_getNextCmdToken(&nextToken,&size,saved); 
	low =U16_value(nextToken);	
	if(cle_getNextCmdToken(&nextToken,&size,saved)==SUCCESS){//skip 'to' if present
		cle_getNextCmdToken(&nextToken,&size,saved); 
		high=U16_value(nextToken);	
	}else
		high = low;

	if(low>high){
		low = high;
		high = low;
	}

	if(add)
		return rtl8651_addALGRule(low, high, isTcp, isServer, isClient);
	else
		return rtl8651_delALGRule(low, high, isTcp, isServer, isClient);}

static int32 _rtl8651_protoTrapCmd(uint32 userId, int32 argc, int8 **saved) {
	int8 *nextToken, add=0;
	uint16 trapContent;
	uint8 trapType;
	int32 size;

	cle_getNextCmdToken(&nextToken,&size,saved); 
	if (strcmp(nextToken, "add") == 0)
		add = TRUE;
	else if (strcmp(nextToken, "remove") == 0)
		add = FALSE;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if (strcmp(nextToken, "ether") == 0) 
		trapType = RTL8651_PROTOTRAP_ETHER;
	else if (strcmp(nextToken, "ip") == 0)
		trapType = RTL8651_PROTOTRAP_IP;
	else if (strcmp(nextToken, "tcp-port") == 0)
		trapType = RTL8651_PROTOTRAP_TCPPORT;
	else if (strcmp(nextToken, "udp-port") == 0)
		trapType = RTL8651_PROTOTRAP_UDPPORT;
	else return FAILED;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	trapContent = U32_value(nextToken);
	if (add == TRUE)
		return rtl8651_addProtoTrap(trapType, trapContent);
	else return rtl8651_delProtoTrap(trapType, trapContent);
}

static int32	_rtl8651_dumpCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken;
	int32 size;

	cle_getNextCmdToken(&nextToken,&size,saved); 
	if ( 0 )
	{
		// never been here ....
		// just for 'else if' statements ....
	}
	else if (strcmp(nextToken, "acldb") == 0) 
	{
		int i;
		int start, end;

		if ( cle_getNextCmdToken((int8 **)&nextToken,&size,saved) == SUCCESS )
			start = *((uint32 *)nextToken);			
		else
			start = 0;

		if ( cle_getNextCmdToken((int8 **)&nextToken,&size,saved) == SUCCESS )
			end = *((uint32 *)nextToken);			
		else
			end = RTL8651_ACLTBL_SIZE-1;

		for( i = start; i <= end; i++ )
		{
			rtlglue_printf("[%3d] func=%p userDefined=%p count=%d\n", i,
				acldb[i].func, acldb[i].userDefined, acldb[i].count );
		}
	}
	else if (strcmp(nextToken, "ipsec") == 0) 
	{
#ifdef CONFIG_RTL865X_IPSEC
		_rtl8651_dumpIpsec();
#else
		rtlglue_printf("\nNot supported, you should 'make menuconfig' to enable 'CONFIG_RTL865X_IPSEC'.\nThen, re-compile the kernel.\n\n" );
#endif
	}
	else if (strcmp(nextToken, "voip") == 0) 
	{
#ifdef CONFIG_RTL865X_VOIP
		voip_dump_callback();
		voip_dump_RTPport();
#else
		rtlglue_printf("\nNot supported, you should 'make menuconfig' to enable 'CONFIG_RTL865X_VOIP'.\nThen, re-compile the kernel.\n\n" );
#endif
	}
	else if (strcmp(nextToken, "policy-route") == 0) {
		extern _rtl8651_tblDrvPolicyRoute_t rtStaticPolicy[];
		//extern _rtl8651_tblDrvPolicyRoute_t rtDynamicPolicy[];
		_rtl8651_tblDrvPolicyRoute_t *policy_t;
		int32 entry;

		printk(">>Static Policy route:\n");
		for(policy_t=rtStaticPolicy, entry=0; entry<32; entry++, policy_t++) {
			if (policy_t->type == 0)
				continue;
			printk("  [*]  ");
			switch (policy_t->type)
			{
			case TRIGGER_DSTIP:
				 printk("dst ip: (%d.%d.%d.%d - %d.%d.%d.%d) ", policy_t->ip_start>>24, (policy_t->ip_start&0x00ff0000)>>16,
				 	(policy_t->ip_start&0x0000ff00)>>8, (policy_t->ip_start&0xff), policy_t->ip_end>>24, 
				 	(policy_t->ip_end&0x00ff0000)>>16, (policy_t->ip_start&0x0000ff00)>>8, (policy_t->ip_end&0xff));
				 break;
				 
			case TRIGGER_SRCIP:
				 printk("src ip: (%d.%d.%d.%d - %d.%d.%d.%d) ", policy_t->ip_start>>24, (policy_t->ip_start&0x00ff0000)>>16,
				 	(policy_t->ip_start&0x0000ff00)>>8, (policy_t->ip_start&0xff), policy_t->ip_end>>24, 
				 	(policy_t->ip_end&0x00ff0000)>>16, (policy_t->ip_start&0x0000ff00)>>8, (policy_t->ip_end&0xff));
				 break;

			case TRIGGER_SRCGLOBALIP:
				 printk("src global ip: (%d.%d.%d.%d - %d.%d.%d.%d) ", policy_t->ip_start>>24, (policy_t->ip_start&0x00ff0000)>>16,
				 	(policy_t->ip_start&0x0000ff00)>>8, (policy_t->ip_start&0xff), policy_t->ip_end>>24, 
				 	(policy_t->ip_end&0x00ff0000)>>16, (policy_t->ip_end&0x0000ff00)>>8, (policy_t->ip_end&0xff));
				 break;

			case TRIGGER_DSTTCPPORT:
				 printk("dst port: (%d - %d) ", policy_t->trigger.dport.start, policy_t->trigger.dport.end);
				 break;

			case TRIGGER_DSTUDPPORT:
				 printk("dst port: (%d - %d) ", policy_t->trigger.dport.start, policy_t->trigger.dport.end);
				 break;

			default: printk("Unknown policy type......\n");

			}
			printk("\n\tnapt ip: %d.%d.%d.%d ", policy_t->ip_alias>>24, (policy_t->ip_alias&0x00ff0000)>>16, 
				(policy_t->ip_alias&0x0000ff00)>>8, policy_t->ip_alias&0xff);
			if (policy_t->rt) {
				rtl8651_tblDrv_routeTable_t *rt = policy_t->rt;
				printk("rtIndex: %d\n", rt-DrvTbl.route);
			} else printk("\n");
		}


	}
	else if (strcmp(nextToken, "mdimdix") == 0) {
		uint32 port;
		uint32 isEnable, isMdi;
		printk("\n");
		for(port=0; port<RTL8651_PORT_NUMBER-1; port++) {
			if (rtl8651_getAutoMdiMdix(port, &isEnable) != SUCCESS)
				return FAILED;
			printk("\tport[%d]: %s", port, ((isEnable==TRUE)? "enable": "disable"));
			if (isEnable == FALSE) {
				if (rtl8651_getSelectMdiMdix(port, &isMdi) != SUCCESS)
					return FAILED;
				printk(" , %s\n", ((isMdi==TRUE)? "mdi": "mdix"));
			} else printk("\n");
		}
		return SUCCESS;
	}
#ifdef RTL865X_DEBUG
	else if (strcmp(nextToken, "timer") == 0) {
		printk("Timer0: _sysUpSeconds=%d, jiffies=%lu\n", _sysUpSeconds, jiffies );
		printk("Timer1: timer1Sec=%d, accJiffies=%d, tuRemind=%d (TC0DATA=0x%08x)\n", timer1Sec, accJiffies, tuRemind, READ_MEM32(TC0DATA) );
	
		printk("\n");
	}
#endif
	else if (strcmp(nextToken, "rate-limit") == 0) {
		rtl865x_tblAsicDrv_rateLimitParam_t asic_rl;
		uint32 entry;

		for(entry=0; entry<RTL8651_RATELIMITTBL_SIZE; entry++) {
			if (rtl8651_getAsicRateLimitTable(entry, &asic_rl) == SUCCESS) {
				printk(" [%d]  Token(%u)  MaxToken(%u)  remainTime Unit(%u)  \n\trefillTimeUnit(%u)  refillToken(%u)\n",
					entry, asic_rl.token, asic_rl.maxToken, asic_rl.t_remainUnit, asic_rl.t_intervalUnit, asic_rl.refill_number);
			}
			else printk(" [%d]  Invalid entry\n", entry);
		}
		printk("\n");
	}
	else if (strcmp(nextToken, "phy") == 0) {
		int8 linkUp, fullduplex, autoNeg;
		uint16 speed;
		uint32 port;
		printk("\n");
		for(port=0; port<6; port++) {
			rtl8651_getEthernetPortLinkStatus(port, &linkUp, &speed, &fullduplex, &autoNeg);
			printk("\tport %d: %s, %d<%s> %s\n", port, (linkUp==TRUE)? "up": "down",
				speed, (fullduplex==TRUE)? "full": "half", (autoNeg==TRUE)? "auto": "non-auto");
		}

	}
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
	else if (strcmp(nextToken, "drv_pppoepassthru") == 0)
	{
		printk("\n<<PPPoE Passthru table>>\n\n");

		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			printk("*** PPPoE Passthru Enabled ***\n\n");
		
			if ( rtl8651_drvDropUnknownPppoePADT )
			{
				printk("    Drop Unknown PPPoE PADT Enabled.\n");
				printk("    Learned inside MAC=%02x:%02x:%02x:%02x:%02x:%02x\n",
				       rtl8651_knownPppoePassthruEthHdr.ether_shost[0],
				       rtl8651_knownPppoePassthruEthHdr.ether_shost[1],
				       rtl8651_knownPppoePassthruEthHdr.ether_shost[2],
				       rtl8651_knownPppoePassthruEthHdr.ether_shost[3],
				       rtl8651_knownPppoePassthruEthHdr.ether_shost[4],
				       rtl8651_knownPppoePassthruEthHdr.ether_shost[5] );
				printk("\n");
			}
			else
			{
				printk("    Drop Unknown PPPoE PADT Disabled.\n");
			}
			
			if ( rtl8651_gPPPoEExternalVlanIndex == RTL8651_VLAN_NUMBER )
				printk("External VLAN Index:  (INVALID)\n" );
			else
				printk("External VLAN Index:  %d\n", rtl8651_gPPPoEExternalVlanIndex );
				
			printk("        Gateway MAC:  %02x:%02x:%02x:%02x:%02x:%02x\n", 
			       rtl8651_gPPPoEExternalGatewayMAC.octet[0], rtl8651_gPPPoEExternalGatewayMAC.octet[1], 
			       rtl8651_gPPPoEExternalGatewayMAC.octet[2], rtl8651_gPPPoEExternalGatewayMAC.octet[3], 
			       rtl8651_gPPPoEExternalGatewayMAC.octet[4], rtl8651_gPPPoEExternalGatewayMAC.octet[5] );

			if ( rtl8651_gPPPoEPassthruVlanID == RTL8651_PPPOEPASSTHRU_VLANID_IS_NOT_SET )
				printk("PPPoE Passthru VLAN Index: (INVALID)\n" );
			else
				printk("PPPoE Passthru VLAN Index: %d\n", rtl8651_gPPPoEPassthruVlanID );

			printk("\n");
		}
		else
		{
			printk("*** PPPoE Passthru Disabled ***\n\n");
		}

	}
#else//CONFIG_RTL865XB
	else if (strcmp(nextToken, "drv_pppoepassthru") == 0)
	{
		printk("\n<<PPPoE Passthru table>>\n\n");

		if ( rtl8651_drvPppoePassthruEnable )
		{
			printk("*** PPPoE Passthru Enabled ***\n\n");
		}
		else
		{
			printk("*** PPPoE Passthru Disabled ***\n\n");
		}
		
		return SUCCESS;
	}
#endif//CONFIG_RTL865XB
#if 1//CONFIG_RTL865XB_EXP_IPV6_PASSTHRU
	else if (strcmp(nextToken, "drv_ipv6passthru") == 0)
	{
		printk("\n<<IPv6 Passthru can be Hardware Accelerated>>\n\n");

		if ( rtl8651_l2protoPassThrough & PASSTHRU_IPV6 )
		{
			printk("*** Current configuration: IPv6 Passthru enabled ***\n\n");
		
			if ( rtl8651_gIPv6ExternalVlanIndex == RTL8651_VLAN_NUMBER )
				printk("External VLAN Index:  (INVALID)\n" );
			else
				printk("External VLAN Index:  %d\n", rtl8651_gIPv6ExternalVlanIndex );
				
			printk("        Gateway MAC:  %02x:%02x:%02x:%02x:%02x:%02x\n", 
			       rtl8651_gIPv6ExternalGatewayMAC.octet[0], rtl8651_gIPv6ExternalGatewayMAC.octet[1], 
			       rtl8651_gIPv6ExternalGatewayMAC.octet[2], rtl8651_gIPv6ExternalGatewayMAC.octet[3], 
			       rtl8651_gIPv6ExternalGatewayMAC.octet[4], rtl8651_gIPv6ExternalGatewayMAC.octet[5] );

			if ( rtl8651_gIPv6PassthruVlanID == RTL8651_IPV6PASSTHRU_VLANID_IS_NOT_SET )
				printk("IPv6 Passthru VLAN Index: (INVALID)\n" );
			else
				printk("IPv6 Passthru VLAN Index: %d\n", rtl8651_gIPv6PassthruVlanID );

			if ( rtl8651_pbv_ipv6passthru == RTL8651_IPV6PASSTHRU_PBVID_IS_NOT_SET )
				printk("Protocol-Based VLAN Control Register for IPv6 Passthru: (INVALID)\n" );
			else
				printk("Protocol-Based VLAN Control Register for IPv6 Passthru: %d\n", rtl8651_pbv_ipv6passthru );

			printk("\n");
		}
		else
		{
			printk("*** Current configuration: IPv6 Passthru disabled ***\n\n");
		}

	}
#else//CONFIG_RTL865XB
	else if (strcmp(nextToken, "drv_ipv6passthru") == 0)
	{
		printk("\n<<IPv6 Passthru is handled by software>>\n\n");

		if ( rtl8651_drvIpv6PassthruEnable )
		{
			printk("*** Current configuration: IPv6 Passthru enabled ***\n\n");
		}
		else
		{
			printk("*** Current configuration: IPv6 Passthru disabled ***\n\n");
		}
		
		return SUCCESS;
	}
#endif//CONFIG_RTL865XB


#ifdef CONFIG_RTL865XB
	else if (strcmp(nextToken, "pbvlan") == 0)
	{
		int32 ruleNo, port;

		printk("\n<<Protocol-Based VLAN Table>>\n\n");

		for( ruleNo = RTL8651_PBV_RULE_USR1; ruleNo <= RTL8651_PBV_RULE_USR2; ruleNo++ )
		{
			uint8 ProtocolType;
			uint16 ProtocolValue;
			
			printk("Rule %d : ", ruleNo);
			rtl8651_queryProtocolBasedVLAN( ruleNo, &ProtocolType, &ProtocolValue );
			
			if ( ProtocolType == 0x0 && ProtocolValue == 0x0000 )
			{
				printk("(Not Used)\n");
			}
			else
			{
				switch( ProtocolType )
				{
					case 0: printk( "EthernetII, Type:" ); break;
					case 1: printk( "RFC-1042  , Type:" ); break;
					case 2: printk( "LLC-Other , {DSAP,SSAP}:" ); break;
					case 3: printk( "Reserved" ); break;
					default:break;
				}
				printk( "0x%04x\n", ProtocolValue );
			}
		}
		
		printk("\nProtocol      P0 P1 P2 P3 P4 P5 E0 E1 E2\n");
		
		for( ruleNo = RTL8651_PBV_RULE_IPX; ruleNo <= RTL8651_PBV_RULE_USR2; ruleNo++  )
		{
			switch( ruleNo )
			{
				case 1: printk("%-12s  ", "IPX"); break;
				case 2: printk("%-12s  ", "NetBIOS"); break;
				case 3: printk("%-12s  ", "PPPoEControl"); break;
				case 4: printk("%-12s  ", "PPPoESession"); break;
				case 5: printk("%-12s  ", "Rule 5"); break;
				case 6: printk("%-12s  ", "Rule 6"); break;
			}

			for( port = 0; port < RTL8651_PORT_NUMBER + 3; port++ )
			{
				uint8 valid, vlanIdx;
				rtl8651_getProtocolBasedVLAN( ruleNo, port, &valid, &vlanIdx );

				if ( valid )
					printk( " %1d ", vlanIdx );
				else
					printk( " - " );
			}
			printk("\n");
		}
		printk("\n");
		return SUCCESS;
	}
#endif//CONFIG_RTL865XB

	else if (strcmp(nextToken, "idle_timeout") == 0)
	{
		int i;
		char *NA="N/A ";
		char *EXCEED="<-->";

		printk("\n<<Idle Timeout>>\n");

		for ( i = 0 ; i < RTL8651_MAX_DIALSESSION ; i ++ )
		{
			printk("[%d]\t", i);
			if (	(rtl8651_sessionHangUp[i].idletime_limit == 0) ||
				(rtl8651_sessionHangUp[i].p_callBack == NULL))
			{
				printk("%s\n", NA);
			} else
			{
				printk("Idle Timeout: %-4d ", rtl8651_sessionHangUp[i].idletime_limit);

				printk("Recv timer: ");
				if (rtl8651_sessionHangUp[i].last_received > 0)
				{
					if ((_sysUpSeconds - rtl8651_sessionHangUp[i].last_received) >= rtl8651_sessionHangUp[i].idletime_limit)
						printk("%s", EXCEED);
					else
						printk("%-4d", _sysUpSeconds-rtl8651_sessionHangUp[i].last_received);
				} else
				{
					printk("%s", NA);
				}
				printk(" ");

				printk("Xmit timer: ");
				if (rtl8651_sessionHangUp[i].last_transmit > 0)
				{
					if ((_sysUpSeconds - rtl8651_sessionHangUp[i].last_transmit) >= rtl8651_sessionHangUp[i].idletime_limit)
						printk("%s", EXCEED);
					else
						printk("%-4d", _sysUpSeconds-rtl8651_sessionHangUp[i].last_transmit);
				} else
				{
					printk("%s", NA);
				}
				printk(" ");

				printk("Session: %-2d", rtl8651_sessionHangUp[i].sessionID);

				printk("\n");
			}
		}
	}

	else if (strcmp(nextToken, "pbnat") == 0) 
	{
		rtl8651_tblDrv_protocolBasedNat_t *pbNat;

		printk("\n<<Protocol-Based NAT>>\n\n");

		if ( rtl8651_drvProtocolBasedNatEnable )
		{
			printk("*** Protocol-Based NAT Enabled ***\n\n");
		
			printk("Protocol    ExtIp            IntIp\n");
			
			for( pbNat = SLIST_FIRST(&DrvTbl.inuseList.protocolBasedNat);
			     pbNat != NULL;
			     pbNat = SLIST_NEXT(pbNat,next) )
			{
				printk( "%3d         %3d.%3d.%3d.%3d  %3d.%3d.%3d.%3d\n",
				        pbNat->protocol,
				        pbNat->ExtIp>>24, (pbNat->ExtIp>>16)&0xff, (pbNat->ExtIp>>8)&0xff, pbNat->ExtIp&0xff,
				        pbNat->IntIp>>24, (pbNat->IntIp>>16)&0xff, (pbNat->IntIp>>8)&0xff, pbNat->IntIp&0xff );
			}
		}
		else
		{
			printk("*** Protocol-Based NAT Disabled ***\n\n");
		}
	}
	else if (strcmp(nextToken, "url-filter") == 0) {
		extern urlFiltering_t	_urlFilter;
		uint32 entry;
		for(entry=0; entry<rtl8651_tblDrvPara.urlfilterDBSize*rtl8651_tblDrvPara.dialSessionNumber; entry++) {
			if (_urlFilter.urlFilterStrLen[entry] != 0) {
				urlPktFiltering_t *url_pf_t;
				printk("[%d] -%s- pattern: %s\n", entry, _urlFilter.urlExactlyMatch[entry] == 1?"EXACTLY":"KEYWORD", _urlFilter.urlFilterDB[entry]);
				if (_urlFilter.urlExactlyMatch[entry]==0)
					printk("\n");
				SLIST_FOREACH(url_pf_t, &(_urlFilter.pktFilter[entry]), nextUse)
				{
					switch (url_pf_t->rule_type)
					{
						case RTL8651_URLPKTFILTER_TYPE_SIP:
							printk("\t\t=> When SIP in\t(%u.%u.%u.%u) ~ (%u.%u.%u.%u)\n", NIPQUAD(url_pf_t->sip_start), NIPQUAD(url_pf_t->sip_end));
							break;
						case RTL8651_URLPKTFILTER_TYPE_SMAC:
							printk("\t\t=> When SMAC is\t%02x:%02x:%02x:%02x:%02x:%02x\n", url_pf_t->smac.octet[0]
																						, url_pf_t->smac.octet[1]
																						, url_pf_t->smac.octet[2]
																						, url_pf_t->smac.octet[3]
																						, url_pf_t->smac.octet[4]
																						, url_pf_t->smac.octet[5]);
							break;
					}
				}
			}
		}
	}
	else if (strcmp(nextToken, "mac-filter") == 0) {
		rtl8651_tblDrv_macACEntry_t *rule;
		uint32 entry=0;
		CTAILQ_FOREACH(rule, &DrvTbl.macACInfo->macAccessControl, nextMAC) {
			printk(" [%d] src: %x:%x:%x:%x:%x:%x  Mask: %x:%x:%x:%x:%x:%x  %s\n", entry,
				rule->mac.octet[0], rule->mac.octet[1], rule->mac.octet[2], 
				rule->mac.octet[3], rule->mac.octet[4], rule->mac.octet[5], 
				rule->macMask.octet[0], rule->macMask.octet[1], rule->macMask.octet[2], 
				rule->macMask.octet[3], rule->macMask.octet[4], rule->macMask.octet[5], 
				((rule->actionType==RTL8651_ACL_PERMIT)? "permit": "drop")
			);
			entry++;
		}
		printk("\n =>default: %s, %s\n\n", ((DrvTbl.macACInfo->def_permit==TRUE)? "permit": "drop"), 
			((DrvTbl.macACInfo->enable==TRUE)? "on": "off")
		);
		return SUCCESS;
	}
	else if (strcmp(nextToken, "qos") == 0) {
		uint32 qoscr, dscr0, dscr1, bit;

		if (rtl8651_getAsicQoSControlRegister(&qoscr) == FAILED)
			return FAILED;
		if (rtl8651_getAsicDiffServReg(&dscr0, &dscr1) == FAILED)
			return FAILED;
		printk(">> QoS:\n\n");
		printk("\tQoS : %s\n", ((qoscr & (1<<20))? "enable": "disable"));
		switch((qoscr&0xc0000000)>>30)
		{
		case RTL8651_RR_H_FIRST:	printk("\tRR Ratio: Always high priority queue first\n"); break;
		case RTL8651_RR_H16_L1:	printk("\tRR Ratio: H16:L1\n"); break; 
		case RTL8651_RR_H8_L1:	printk("\tRR Ratio: H8:L1\n"); break;
		case RTL8651_RR_H4_L1:	printk("\tRR Ratio: H4:L1\n"); break;
		default: assert(0);
		}
		printk("\t802.1p Vlan Tag Priority: %s\n", ((qoscr&(1<<27))? "disable": "enable")); /* 0: enable, 1: disable */
		printk("\tPort 5 Priority: %s\n", ((qoscr&(1<<26))? "high": "low"));
		printk("\tPort 4 Priority: %s\n", ((qoscr&(1<<25))? "high": "low"));
		printk("\tPort 3 Priority: %s\n", ((qoscr&(1<<24))? "high": "low"));
		printk("\tPort 2 Priority: %s\n", ((qoscr&(1<<23))? "high": "low"));
		printk("\tPort 1 Priority: %s\n", ((qoscr&(1<<22))? "high": "low"));
		printk("\tPort 0 Priority: %s\n", ((qoscr&(1<<21))? "high": "low"));
		
		for(bit=0; bit<32; bit++) {
			if (bit % 4 == 0)
				printk("\n\t");
			printk("%2x : <%s>  ", bit, ((dscr0 & (1<<bit))? "high": "low"));
		}
		for(bit=0; bit<32; bit++) {
			if (bit % 4 == 0)
				printk("\n\t");
			printk("%2x : <%s>  ", bit+32, ((dscr1 & (1<<bit))? "high": "low"));
		}
		printk("\n");	
		return SUCCESS;
	} else if (strcmp(nextToken, "flow-control") == 0) {
		uint32 fcren;

		if (rtl8651_getAsicFlowControlRegister(&fcren) == FAILED)
			return FAILED;
		printk(">> Flow Control Register:\n\n");
		printk("\tFlow Control for CPU:   input: %s,  output: %s\n",   ((fcren&0x80000000)? "<enable>": "<disable>"), ((fcren&0x01000000)? "<enable>": "<disable>"));
		printk("\tFlow Control for port0: input: %s,  output: %s\n", ((fcren&0x02000000)? "<enable>": "<disable>"), ((fcren&0x00040000)? "<enable>": "<disable>"));
		printk("\tFlow Control for port1: input: %s,  output: %s\n", ((fcren&0x04000000)? "<enable>": "<disable>"), ((fcren&0x00080000)? "<enable>": "<disable>"));
		printk("\tFlow Control for port2: input: %s,  output: %s\n", ((fcren&0x08000000)? "<enable>": "<disable>"), ((fcren&0x00100000)? "<enable>": "<disable>"));
		printk("\tFlow Control for port3: input: %s,  output: %s\n", ((fcren&0x10000000)? "<enable>": "<disable>"), ((fcren&0x00200000)? "<enable>": "<disable>"));
		printk("\tFlow Control for port4: input: %s,  output: %s\n", ((fcren&0x20000000)? "<enable>": "<disable>"), ((fcren&0x00400000)? "<enable>": "<disable>"));
		printk("\tFlow Control for port5: input: %s,  output: %s\n", ((fcren&0x40000000)? "<enable>": "<disable>"), ((fcren&0x00800000)? "<enable>": "<disable>"));
		printk("\n");
		return SUCCESS;
	} else if(strcmp(nextToken, "queueInfo") == 0) {
		int32 i;
		uint32 inCount, oHCount, oLCount, gdcsrValue;
		gdcsrValue = REG32(0xbc803050);
		for(i=0; i<6; i++) {
			oLCount=REG32(0xbc803034 + (i<<2));
			oHCount = ((oLCount >>10)&0x3FF);
			inCount = ((oLCount >>20)&0x3FF);
			oLCount = oLCount & 0x3FF;
			printk("Port %d\tInQueue %u (%u/%u), Out High Queue %u, Out Low Queue %u (%u/%u)\n", 
				i, inCount, (REG32(0xBC805018)>>24)&0xFF, (REG32(0xBC805018)>>16)&0xFF, 
				oHCount, oLCount, (REG32(0xBC805018)>>8)&0xFF, REG32(0xBC805018)&0xFF);
			printk("\tIn FC %s ", (REG32(0xbc805014) & (0x1<<(25+i)))?"ENABLED":"DISABLED");
			printk("Out FC %s High queue %s Low queue %s\n", 
				(REG32(0xbc805014) & (0x1<<(18+i)))?"ENABLED":"DISABLED", 
				gdcsrValue&(1<<(i+7))?"Congested":"OK", gdcsrValue&(1<<i)?"Congested":"OK");
		}
		oLCount=REG32(0xbc80304c);
		oHCount = ((oLCount >>10)&0x3FF);
		inCount = ((oLCount >>20)&0x3FF);
		oLCount = oLCount & 0x3FF;
		printk("CPU\n\tInput Queue %u (%u/%u), Output High Queue %u, Output Low Queue %u (%u/%u)\n", 
			inCount,  (REG32(0xBC805018)>>24)&0xFF, (REG32(0xBC805018)>>16)&0xFF, 
			oHCount, oLCount, (REG32(0xBC805018)>>8)&0xFF, REG32(0xBC805018)&0xFF);
		printk("\tIn FC %s ", (REG32(0xbc805014) & (0x1<<31))?"ENABLED":"DISABLED");
		printk("Out FC %s \n", (REG32(0xbc805014) & (0x1<<24))?"ENABLED":"DISABLED");
		printk("Global status\n");
		printk("\t Descriptor %s Used descriptor %u In queue FC %s Out queue FC %s\n", 
			gdcsrValue&(1<<26)?"Run-out":"Enough", (gdcsrValue>>16)&0x3FF,
			gdcsrValue&(1<<15)?"ENABLED":"DISABLE", gdcsrValue&(1<<14)?"ENABLED":"DISABLE");
		printk("\t In queueTH %u Out queueTH %u\n", (REG32(0xBC805028)>>16)&0xFFFF, REG32(0xBC805028)&0xFFFF);
		printk("\t Descriptor threshold %u\n", REG32(0xBC805014)&0x3FF);
		
	}else if(strcmp(nextToken, "hstamp") == 0){
#if defined(RTL865X_TEST) || defined(RTL865X_MODEL_USER) || defined(RTL865X_MODEL_KERNEL)
		rtlglue_printf("NOT supported when any of RTL865X_TEST/RTL865X_MODEL_USER/RTL865X_MODEL_KERNEL is defined.\n" );
#else
		hs_dumpHsb(0);
		hs_dumpHsa(0);
#endif
	}else if(strcmp(nextToken, "l2") == 0) {
		rtl865x_tblAsicDrv_l2Param_t asic_l2;
 		uint32 row, col, port, m=0;
  			
		printk(">>ASIC L2 Table:\n");
		for(row=0; row<RTL8651_L2TBL_ROW; row++)
			for(col=0; col<RTL8651_L2TBL_COLUMN; col++) {
				if (rtl8651_getAsicL2Table(row, col, &asic_l2) == FAILED)
					continue;
					
//#ifdef CONFIG_RTL865X_WLAN_ACCELERATION					
				if (asic_l2.isStatic && asic_l2.ageSec==0 && asic_l2.cpu && asic_l2.memberPortMask == 0)
					continue;
//#endif
				printk("%4d.[%3d,%d] %2x:%2x:%2x:%2x:%2x:%2x mbr(",m, row, col, 
						asic_l2.macAddr.octet[0], asic_l2.macAddr.octet[1], asic_l2.macAddr.octet[2], 
						asic_l2.macAddr.octet[3], asic_l2.macAddr.octet[4], asic_l2.macAddr.octet[5]
				);
				m++;
				for(port=0; port<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; port++){
					if(asic_l2.memberPortMask & (1<<port))
						printk("%d ", port);
					else 
						printk("  ");
				}
				printk(")");
				printk("%s %s %s %s age:%d ",asic_l2.cpu?"CPU":"FWD", asic_l2.isStatic?"STA":"DYN",  asic_l2.srcBlk?"BLK":"   ", asic_l2.nhFlag?"NH":"  ", asic_l2.ageSec);
//#ifdef CONFIG_RTL865X_WLAN_ACCELERATION					
				if(0==(asic_l2.macAddr.octet[0]&1)&&asic_l2.memberPortMask&peripheralExtPortMask){
					 rtl8651_tblDrv_filterDbTableEntry_t * entry;
					 uint8 queryMac[6];
					 memcpy(queryMac,&asic_l2.macAddr,6);
					 queryMac[5]=row^asic_l2.macAddr.octet[0]^asic_l2.macAddr.octet[1]^asic_l2.macAddr.octet[2]^asic_l2.macAddr.octet[3]^asic_l2.macAddr.octet[4];
					entry=_rtl8651_getVlanFilterDatabaseEntry(0, (ether_addr_t *)queryMac) ;
					if (entry)
						printk("W%d",entry->linkId );
				}
//#endif
				printk("\n");
			}
	}

	else if(strcmp(nextToken, "l3") == 0) {

		rtl865x_tblAsicDrv_routingParam_t asic_l3;
		int8 *str[] = { "PPPoE", "L2", "ARP", " ", "CPU", "NxtHop", "DROP", " " };
		uint32 idx, mask;

		printk(">>L3 Routing Table:\n");
		for(idx=0; idx<RTL8651_ROUTINGTBL_SIZE; idx++) {
			if (rtl8651_getAsicRouting(idx, &asic_l3) == FAILED) {
				printk("\t[%d]  (Invalid)\n", idx);
				continue;
			}
			if (idx == RTL8651_ROUTINGTBL_SIZE-1)
				mask = 0;
			else for(mask=32; !(asic_l3.ipMask&0x01); asic_l3.ipMask=asic_l3.ipMask>>1)
					mask--;
			printk("\t[%d]  %d.%d.%d.%d/%d process(%s)\n", idx, (asic_l3.ipAddr>>24),
				((asic_l3.ipAddr&0x00ff0000)>>16), ((asic_l3.ipAddr&0x0000ff00)>>8), (asic_l3.ipAddr&0xff), 
				mask, str[asic_l3.process]
			);
			switch(asic_l3.process) 
			{
			case 0x00:	/* PPPoE */
				printk("\t           dvidx(%d)  pppidx(%d) nxthop(%d)\n", asic_l3.vidx, asic_l3.pppoeIdx, (asic_l3.nextHopRow<<2)+asic_l3.nextHopColumn);
				break;
				
			case 0x01:	/* L2 */
				printk("              dvidx(%d) nexthop(%d)\n", asic_l3.vidx, (asic_l3.nextHopRow<<2)+asic_l3.nextHopColumn);
				break;

			case 0x02:	/* ARP */
				printk("             dvidx(%d) ARPSTA(%d) ARPEND(%d) IPIDX(%d)\n", asic_l3.vidx, asic_l3.arpStart, asic_l3.arpEnd, asic_l3.arpIpIdx);
				break;

			case 0x03:	/* Reserved */
				assert(0);

			case 0x04:	/* CPU */
				printk("             dvidx(%d)\n", asic_l3.vidx);
				break;

			case 0x05:	/* NAPT Next Hop */
				printk("              NHSTA(%d) NHNUM(%d) NHNXT(%d) NHALGO(%d) IPDOMAIN(%d)\n", asic_l3.nhStart,
					asic_l3.nhNum, asic_l3.nhNxt, asic_l3.nhAlgo, asic_l3.ipDomain);
				break;

			case 0x06:	/* DROP */
				printk("             dvidx(%d)\n", asic_l3.vidx);
				break;

			case 0x07:	/* Reserved */
				/* pass through */
			default: assert(0);
			}
		}
	}
	else if (strcmp(nextToken, "soft-l3") ==0) {
		rtl8651_tblDrv_filterDbTableEntry_t *macInfo;
		rtl8651_tblDrv_routeTable_t *rt_t;
		uint32 entry, i, mask;
		uint8 tempStr[20];
	
		printk(">>Software Routing Table:\n\n");
		printk("        Destination/Mask    Gateway          Iface   DVID  Process\n");
		printk("        ==================  ===============  ======  ====  ================\n");
		rt_t = DrvTbl.route;
		for(entry=0; entry<rtl8651_tblDrvPara.routingTableSize; entry++, rt_t++) {
			if (rt_t->valid == 0) {
				printk("  [%2d]  * *\n", entry);
				continue;
			}
			for(i=mask=0; i<32; i++)
				if (rt_t->ipMask&(1<<i))
					mask++;
			sprintf(tempStr, "%d.%d.%d.%d/%d", rt_t->ipAddr>>24, (rt_t->ipAddr&0x00ff0000)>>16, 
					(rt_t->ipAddr&0x0000ff00)>>8, rt_t->ipAddr&0xff, mask
			);
			printk("  [%2d]  %-18s  ", entry, tempStr);			

			if (rt_t->ifrt_flag == 1) {
				printk("%-15s  %6s  %4d  arp[%d:%d]\n", "*", rt_t->netif_t->ifName, rt_t->dvid, rt_t->rt_nharpsta, rt_t->rt_nharpend);
				continue;
			}
			sprintf(tempStr, "%d.%d.%d.%d", rt_t->nextHop>>24, (rt_t->nextHop&0x00ff0000)>>16, (rt_t->nextHop&0x0000ff00)>>8, rt_t->nextHop&0xff);
			printk("%-15s  %6s  %4d  ", tempStr, rt_t->netif_t->ifName, rt_t->dvid);
			if (rt_t->process ==RT_NEXTHOP)
				printk("nhIdx[%d:%d]  ", rt_t->rt_nhstart_t->entryIndex, rt_t->rt_nhend_t->entryIndex);
			if (rt_t->nextHopType==RTL8651_LL_VLAN)
				macInfo = (rt_t->rt_nharp_t)? rt_t->rt_nharp_t->macInfo: NULL;
			else macInfo = (rt_t->rt_nhSession_t)? rt_t->rt_nhSession_t->macInfo: NULL;
			if (macInfo == NULL)
				printk("L2[no]\n");
			else printk("L2[%d:%d]\n\n", rtl8651_filterDbIndex(&macInfo->macAddr), macInfo->asicPos);
				


		}
	}
	else if(strcmp(nextToken, "arp") == 0) {
		rtl865x_tblAsicDrv_routingParam_t asic_l3;
		rtl865x_tblAsicDrv_arpParam_t asic_arp;
		rtl865x_tblAsicDrv_l2Param_t asic_l2;
	
		uint32	i, j, port;
		ipaddr_t ipAddr;
		int8 ipBuf[sizeof"255.255.255.255"];

		
		printk(">>Arp Table:\n");
		for(i=0; i<RTL8651_ARPTBL_SIZE; i++) {
			if (rtl8651_getAsicArp(i,  &asic_arp) == FAILED)
				continue;
			for(j=0; j<RTL8651_ROUTINGTBL_SIZE; j++) {
				if (rtl8651_getAsicRouting(j, &asic_l3) == FAILED || asic_l3.process!= 2)
					continue;
				if(asic_l3.arpStart <= (i>>3) &&  (i>>3) <= asic_l3.arpEnd) {
					ipAddr = (asic_l3.ipAddr & asic_l3.ipMask) + (i - (asic_l3.arpStart<<3));
					if(rtl8651_getAsicL2Table_Patch(asic_arp.nextHopRow, asic_arp.nextHopColumn, &asic_l2) == FAILED){
						inet_ntoa_r(ipAddr, ipBuf);
						printk("%-16s [%3d,%d] ", ipBuf, asic_arp.nextHopRow, asic_arp.nextHopColumn);
					}else {
						inet_ntoa_r(ipAddr, ipBuf);
						printk("%-16s %02x-%02x-%02x-%02x-%02x-%02x (", ipBuf, asic_l2.macAddr.octet[0], asic_l2.macAddr.octet[1], asic_l2.macAddr.octet[2], asic_l2.macAddr.octet[3], asic_l2.macAddr.octet[4], asic_l2.macAddr.octet[5]);
						for(port=0; port< RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; port++){
							if(asic_l2.memberPortMask& (1<<port))
								printk("%d ", port);
							else
								printk("  ");
						}							
						printk(") %us", asic_l2.ageSec);
					}
					continue;
				}
			}
			printk(" ARP:%3d  L2:%3d,%d\n", i, asic_arp.nextHopRow, asic_arp.nextHopColumn);

		}


		
	}
	else if (strcmp(nextToken, "drv-arp") == 0) {
		rtl8651_dumpDriverArpInfo();
	}
	else if(strcmp(nextToken, "pppoe") == 0) {
		rtl865x_tblAsicDrv_pppoeParam_t asic_pppoe;
		int32	i;

		printk(">>PPPOE Table:\n");
		for(i=0; i<RTL8651_PPPOETBL_SIZE; i++) {
			if (rtl8651_getAsicPppoe(i,  &asic_pppoe) == FAILED)
				continue;
			printk("\t[%d]  sessionID(%d)  ageSec(%d)\n", i, asic_pppoe.sessionId, asic_pppoe.age);
		}
	}
	else if(strcmp(nextToken, "napt") == 0) {
		uint32	i;
		uint32	total = 0,total2=0,lines=0, mode;
		int32 ret;
		uint32 start, end, extport=0;

		mode=3;
		start=0;
		end=1023;
		cle_getNextCmdToken(&nextToken,&size,saved); 

		if(!strcmp("drv", (int8 *)nextToken)){
			mode=2;
			cle_getNextCmdToken(&nextToken,&size,saved); 
		}else	if(!strcmp("asic", (int8 *)nextToken)){
			mode=1;
			cle_getNextCmdToken(&nextToken,&size,saved); 
		}
		
		//Get start, end position		
		if(size>0){
			start=end=*((uint32 *)nextToken);
		}else if(!strcmp("extport", (int8*)nextToken)){
			cle_getNextCmdToken((int8 **)&nextToken,&size,saved); //extport number
			extport = *((uint32 *)nextToken);			
			start= 0;
			end=1023;
		}else if(!strcmp("from", (int8 *)nextToken)){
			cle_getNextCmdToken((int8 **)&nextToken,&size,saved); //starting index
			start = *((uint32 *)nextToken);
			cle_getNextCmdToken((int8 **)&nextToken,&size,saved);
			if(!strcmp("all", (int8 *)nextToken))
				end = 1023;
			else if(!strcmp("until", (int8 *)nextToken)){
				cle_getNextCmdToken((int8 **)&nextToken,&size,saved);
				end = *((uint32 *)nextToken);
				if(end>1023)
					end=1023;
			}
		}else if(!strcmp("until", (int8 *)nextToken)){
				cle_getNextCmdToken((int8 **)&nextToken,&size,saved);
				end = *((uint32 *)nextToken);
				if(end>1023)
					end=1023;
				start = 0;
		}else if(!strcmp("all", (int8 *)nextToken)){
				end = 1023;
				start = 0;
		}

		printk(">>TCP/UDP Table:\n");
		printk("\tA)utolearnChk.\n\t"
				"a)sic flow. ASIC forwards packets on this direction\n\t"
				"B)idirectional.outgoing hash index=inbound index\n\t"
				"C)ollided entry. More connections in driver\n\t"
				"c)andidate flow. May write back to ASIC\n\t"
				"D)edicated entry. 8651B specific entry\n\t"
				"f)orwarding engine created.\n\t"
				"I)nbound flow.\n\t"
				"L)iberal hashing flow. 8651b specific.\n\t"
				"O)utbound flow\n\t"
				"Q)uiet flow. 8651B specific\n\t"
				"S)tatic entry. Never replaced by ASIC.\n\t"
				"s)oftware flow. Pure software flow. Never write to ASIC\n\t"
				"T)CP\n\t"
				"U)DP\n\t"
				"V)alid entry\n\t");
		printk("\n");
		
		for(i=start; i<=end; i++) {
			if(mode&1){
				ret=_rtl8651_dumpAsicNaptTableEntry(i, extport);
				if(ret==FAILED)
					continue;
				if(ret==1)
					total+=1;
			}
			if(mode&2){
				int32 count=0;
				//if (mode==2)
				//	printk("  [%4d]", i);
				count=_rtl8651_dumpNaptTableEntry(i, extport);
				if(count)
					lines++;
				total2+=count;
			}
		}
		printk("%d ASIC entries listed.\n", total);
		printk("%d flows listed, on %d lines.\n", total2, lines);
	}
	else if(strcmp(nextToken, "icmp") == 0) {
		uint32 i;
		rtl865x_tblAsicDrv_naptIcmpParam_t naptIcmp;
		uint32 total = 0;
		printk(">>ICMP Table:\n");
		for(i=0; i<32; i++) {
			if (rtl8651_getAsicNaptIcmpTable(i, &naptIcmp) == FAILED)
				continue;

			if (naptIcmp.isValid ||naptIcmp.isStatic) {
				printk("GId(%x-%x) %s Local %d.%d.%d.%d-%x %s %s %s %s Age:%d Cnt:%u \n",
						i, naptIcmp.offset,naptIcmp.isPptp?"PPTP": "ICMP",
						(naptIcmp.insideLocalIpAddr>>24)&0xff, (naptIcmp.insideLocalIpAddr>>16)&0xff, (naptIcmp.insideLocalIpAddr>>8)&0xff, naptIcmp.insideLocalIpAddr&0xff,
						naptIcmp.insideLocalId, 
						naptIcmp.isCollision? "Col":"",
						naptIcmp.isStatic? "STA": "", 
						naptIcmp.isSpi? "SPI": "", 
						naptIcmp.isValid? "VAL":"",
						naptIcmp.ageSec,
						naptIcmp.count);
				total++;
			}
		}
		printk("Total entry: %u\n", total);
	}
	else if (strcmp(nextToken, "drv_icmp") == 0) {
		_rtl8651_dumpNaptIcmpTable();
	}
	else if(strcmp(nextToken, "ip") == 0) {
		rtl865x_tblAsicDrv_extIntIpParam_t asic_ip;
		int32	i;
		int8 intIpBuf[sizeof"255.255.255.255"];
		int8 extIpBuf[sizeof"255.255.255.255"];

		printk(">>IP Table:\n");
		for(i=0; i<RTL8651_IPTABLE_SIZE; i++) {
			if (rtl8651_getAsicExtIntIpTable(i,  &asic_ip) == FAILED) {
				printk("  [%d] (Invalid)\n", i);
				continue;
			}
			inet_ntoa_r(asic_ip.intIpAddr, intIpBuf);
			inet_ntoa_r(asic_ip.extIpAddr,extIpBuf);			
			printk("  [%d] intip(%-14s) extip(%-14s) type(%s) nhIdx(%d) rfcnt(%d)\n",
					i, intIpBuf,extIpBuf,
					(asic_ip.localPublic==TRUE? "LP" : (asic_ip.nat==TRUE ? "NAT" : "NAPT")), asic_ip.nhIndex, DrvTbl.iptbl[i].ref_cnt);
		}
	}
	else if(strcmp(nextToken, "serverPort") == 0) {
		rtl865x_tblAsicDrv_serverPortParam_t asic_sp;
		int32 i, entry;
		int8 intIpBuf[sizeof"255.255.255.255"];
		int8 extIpBuf[sizeof"255.255.255.255"];

		printk(">>ServerPort Table:\n");
		for(entry=0, i=0; i<RTL8651_SERVERPORTTBL_SIZE; i++) {
			if (rtl8651_getAsicServerPortTable(i,  &asic_sp) == FAILED) {
				printk("  [%d] (Invalid)\n", i);
				continue;
			}	
			inet_ntoa_r(asic_sp.intIpAddr,intIpBuf);
			inet_ntoa_r(asic_sp.extIpAddr,extIpBuf);			
			printk("  [%d] extIP(%-16s) extPort(%d)\n", i, extIpBuf, asic_sp.extPort );
			printk("      intIP(%-16s) intPort(%d) nextHop(%d) isRange(%d)\n", intIpBuf, asic_sp.intPort, asic_sp.nhIndex, asic_sp.portRange);
			entry++;
		}
		printk("total entry: %d\n", entry);
	}
	else if(strcmp(nextToken, "alg") == 0) {
		rtl865x_tblAsicDrv_algParam_t asic_alg;
		int32	i, entry;


		printk(">>ALG Table:\n");
		for(entry=0, i=0; i<RTL8651_ALGTBL_SIZE; i++) {
			if (rtl8651_getAsicAlg(i, &asic_alg) == SUCCESS) {
				printk("\t[%d] port(%d)\n", i, asic_alg.port);
				entry++;
			}
		}
		printk("total entry: %d\n", entry);
	}
	else if(strcmp(nextToken, "multicast") == 0) {

	}
	else if(strcmp(nextToken, "vlan") == 0) {
		uint8	*pst[] = { "DIS", "BLK", "LIS", "LRN", "FWD" };
		uint8	*mac;
		int32	i, j;


		printk(">>ASIC Vlan Table:\n\n");
		for(i=1; i<4096; i++) {
			rtl865x_tblAsicDrv_vlanParam_t vlan;

			if (rtl8651_getAsicVlan( i, &vlan ) == FAILED)
				continue;

			if(vlan.promiscuous)
				printk("  VID[%d] Promiscuous Mode\n", i);
			else {
				mac = (uint8 *)&vlan.macAddr.octet[0];
				printk("  VID[%d] %x:%x:%x:%x:%x:%x", 
					i, mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
				printk("  %s, L3/4 HW acc %s %s\n",
					vlan.internal==TRUE? "Internal": "External", vlan.enableRoute==TRUE? "enabled": "disabled",
					vlan.broadcastToCpu==TRUE?",L2 bcast to SW": "");
				printk("      ingress ");
				if(RTL8651_ACLTBL_DROP_ALL<=vlan.inAclStart){
					if(vlan.inAclStart==RTL8651_ACLTBL_PERMIT_ALL)
						printk("permit all,");
					if(vlan.inAclStart==RTL8651_ACLTBL_ALL_TO_CPU)
						printk("all to cpu,");
					if(vlan.inAclStart==RTL8651_ACLTBL_DROP_ALL)
						printk("drop all,");
				}else
					printk("ACL %d-%d, ",vlan.inAclStart, vlan.inAclEnd);
				printk("  egress ");
				if(RTL8651_ACLTBL_DROP_ALL<=vlan.outAclStart){
					if(vlan.outAclStart==RTL8651_ACLTBL_PERMIT_ALL)
						printk("permit all,");
					if(vlan.outAclStart==RTL8651_ACLTBL_ALL_TO_CPU)
						printk("all to cpu,");
					if(vlan.outAclStart==RTL8651_ACLTBL_DROP_ALL)
						printk("drop all,");
				}else
					printk("ACL %d-%d, ",vlan.outAclStart, vlan.outAclEnd);
				printk("\n      Untag member ports:");
				for(j=0; j<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; j++)
					if(vlan.untagPortMask & (1<<j))
						printk("%d ", j);
				printk("\n      Active member ports:");
				for(j=0; j<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; j++)
					if(vlan.memberPortMask & (1<<j))
						printk("%d ", j);
				
				printk("\n      %d MAC Addresses, MTU %d Bytes\n", vlan.macAddrNumber, vlan.mtu);
				printk("      Port state(");
				for(j=0; j<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; j++)
				{
					if((vlan.memberPortMask & (1<<j))==0)
						continue;
					if ( vlan.portState[j] > 4 )
						printk("--- ");
					else
						printk("%d:%s ", j, pst[(int32)vlan.portState[j]]);
				}
				printk(")\n\n");
			}
		}
	}
	else if(strcmp(nextToken, "prototrap") == 0) {
		uint16 trapContent;
		uint8 trapType;
		int8 isEnable;
		uint32 i;

		for(i=0; i<RTL8651_PROTOCOLTRAP_SIZE; i++) {
			printk("\t[%d] ", i);
			if (convert_getAsicProtoTrap(i, &trapType, &trapContent) == FAILED)
				printk("Invalid\n");
			else printk("Type(%x) Content(%x)\n", trapType, trapContent);
		}
		printk("\n");
		rtl8651_getHardwiredProtoTrap(0x00, 0x0806, &isEnable); 
		printk("\t[0] ARP %s\n", isEnable==TRUE? "on": "off");
		rtl8651_getHardwiredProtoTrap(0x00, 0x8035, &isEnable); 
		printk("\t[1] RARP %s\n", isEnable==TRUE? "on": "off");
		rtl8651_getHardwiredProtoTrap(0x00, 0x8863, &isEnable); 
		printk("\t[2] PPPoE %s\n", isEnable==TRUE? "on": "off");
		rtl8651_getHardwiredProtoTrap(0x01, 0x0002, &isEnable); 
		printk("\t[3] IGMP, DVMRP and PIMv1 %s\n", isEnable==TRUE? "on": "off");
		rtl8651_getHardwiredProtoTrap(0x01, 0x0043, &isEnable); 
		printk("\t[4] BOOTPs/DHCP %s\n", isEnable==TRUE? "on": "off");
		rtl8651_getHardwiredProtoTrap(0x01, 0x0044, &isEnable); 
		printk("\t[5] BOOTPc/DHCP %s\n", isEnable==TRUE? "on": "off");
		rtl8651_getHardwiredProtoTrap(0x01, 0x0059, &isEnable); 
		printk("\t[6] OSPF %s\n", isEnable==TRUE? "on": "off");
		rtl8651_getHardwiredProtoTrap(0x03, 0x0208, &isEnable); 
		printk("\t[7] RIP %s\n", isEnable==TRUE? "on": "off");
	}
	else if (strcmp(nextToken, "mirror-port") == 0) {
		uint32 mirrorRx, mirrorTx, mirrorPort;
		
		rtl8651_getAsicPortMirror(&mirrorRx, &mirrorTx, &mirrorPort);
		printk(">>Mirror Control Register:\n\n");
		printk("  Mirror Rx: 0x%x\n", mirrorRx);
		printk("  Mirror Tx: 0x%x\n", mirrorTx);
		printk("  Mirror Port: 0x%x\n", mirrorPort);
	}
	else if(strcmp(nextToken, "acl") == 0) {

		//int8 *actionT[] = { "", "permit", "drop", "cpu", "drop and log", "redirect to ethernet", "redirect to pppoe" };
		int8 *actionT[] = { "", "permit", "drop", "cpu", "drop log", "drop notify", "redirect to ethernet", "redirect to pppoe", "mirror", "mirro keep match", "drop rate exceed pps", "log rate exceed pps", "drop rate exceed bps", "log rate exceed bps" };
		_rtl8651_tblDrvAclRule_t asic_acl;
		rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
		uint32 start, end;

		uint16 vid;
		int8 outRule;


				
		printk(">>ASIC ACL Table:\n\n");
		for(vid=0; vid<4095; vid++ ) {
			/* Read VLAN Table */
			if (rtl8651_getAsicVlan(vid, &asic_vlan) == FAILED)
				continue;

			outRule = FALSE;
			start = asic_vlan.inAclStart; end = asic_vlan.inAclEnd;
	again:
			if (outRule == FALSE)
				printk("\n<<Ingress Rule for Vlan %d:>>\n", vid);
			else printk("\n<<Egress Rule for Vlan %d>>:\n", vid);
			for( ; start<=end; start++) {
				if (rtl8651_getAsicAclRule(start, &asic_acl) == FAILED)
					assert(0);
				switch(asic_acl.ruleType_)
				{
				case RTL8651_ACL_MAC:
					printk(" [%d] rule type: %s   rule action: %s\n", start, "Ethernet", actionT[asic_acl.actionType_]);
					printk("\tether type: %x   ether type mask: %x\n", asic_acl.typeLen_, asic_acl.typeLenMask_);
					printk("\tDMAC: %x:%x:%x:%x:%x:%x  DMACM: %x:%x:%x:%x:%x:%x\n",
						asic_acl.dstMac_.octet[0], asic_acl.dstMac_.octet[1], asic_acl.dstMac_.octet[2],
						asic_acl.dstMac_.octet[3], asic_acl.dstMac_.octet[4], asic_acl.dstMac_.octet[5],
						asic_acl.dstMacMask_.octet[0], asic_acl.dstMacMask_.octet[1], asic_acl.dstMacMask_.octet[2],
						asic_acl.dstMacMask_.octet[3], asic_acl.dstMacMask_.octet[4], asic_acl.dstMacMask_.octet[5]
					);
					printk("\tSMAC: %x:%x:%x:%x:%x:%x  SMACM: %x:%x:%x:%x:%x:%x\n",
						asic_acl.srcMac_.octet[0], asic_acl.srcMac_.octet[1], asic_acl.srcMac_.octet[2],
						asic_acl.srcMac_.octet[3], asic_acl.srcMac_.octet[4], asic_acl.srcMac_.octet[5],
						asic_acl.srcMacMask_.octet[0], asic_acl.srcMacMask_.octet[1], asic_acl.srcMacMask_.octet[2],
						asic_acl.srcMacMask_.octet[3], asic_acl.srcMacMask_.octet[4], asic_acl.srcMacMask_.octet[5]
					);
					break;

				case RTL8651_ACL_IP:
					printk(" [%d] rule type: %s   rule action: %s\n", start, "IP", actionT[asic_acl.actionType_]);
					printk("\tdip: %d.%d.%d.%d dipM: %d.%d.%d.%d\n", (asic_acl.dstIpAddr_>>24),
						((asic_acl.dstIpAddr_&0x00ff0000)>>16), ((asic_acl.dstIpAddr_&0x0000ff00)>>8),
						(asic_acl.dstIpAddr_&0xff), (asic_acl.dstIpAddrMask_>>24), ((asic_acl.dstIpAddrMask_&0x00ff0000)>>16),
						((asic_acl.dstIpAddrMask_&0x0000ff00)>>8), (asic_acl.dstIpAddrMask_&0xff)
					);
					printk("\tsip: %d.%d.%d.%d sipM: %d.%d.%d.%d\n", (asic_acl.srcIpAddr_>>24),
						((asic_acl.srcIpAddr_&0x00ff0000)>>16), ((asic_acl.srcIpAddr_&0x0000ff00)>>8),
						(asic_acl.srcIpAddr_&0xff), (asic_acl.srcIpAddrMask_>>24), ((asic_acl.srcIpAddrMask_&0x00ff0000)>>16),
						((asic_acl.srcIpAddrMask_&0x0000ff00)>>8), (asic_acl.srcIpAddrMask_&0xff)
					);
					printk("\tTos: %x   TosM: %x   ipProto: %x   ipProtoM: %x   ipFlag: %x   ipFlagM: %x\n",
						asic_acl.tos_, asic_acl.tosMask_, asic_acl.ipProto_, asic_acl.ipProtoMask_, asic_acl.ipFlag_, asic_acl.ipFlagMask_
					);
					printk("\t<FOP:%x> <FOM:%x> <http:%x> <httpM:%x> <IdentSdip:%x> <IdentSdipM:%x> \n",
						asic_acl.ipFOP_, asic_acl.ipFOM_, asic_acl.ipHttpFilter_, asic_acl.ipHttpFilterM_, asic_acl.ipIdentSrcDstIp_,
						asic_acl.ipIdentSrcDstIpM_
					);
					printk("\t<DF:%x> <MF:%x>\n", asic_acl.ipDF_, asic_acl.ipMF_); 
					break;			

				case RTL8651_ACL_ICMP:
					printk(" [%d] rule type: %s   rule action: %s\n", start, "ICMP", actionT[asic_acl.actionType_]);
					printk("\tdip: %d.%d.%d.%d dipM: %d.%d.%d.%d\n", (asic_acl.dstIpAddr_>>24),
						((asic_acl.dstIpAddr_&0x00ff0000)>>16), ((asic_acl.dstIpAddr_&0x0000ff00)>>8),
						(asic_acl.dstIpAddr_&0xff), (asic_acl.dstIpAddrMask_>>24), ((asic_acl.dstIpAddrMask_&0x00ff0000)>>16),
						((asic_acl.dstIpAddrMask_&0x0000ff00)>>8), (asic_acl.dstIpAddrMask_&0xff)
					);
					printk("\tsip: %d.%d.%d.%d sipM: %d.%d.%d.%d\n", (asic_acl.srcIpAddr_>>24),
						((asic_acl.srcIpAddr_&0x00ff0000)>>16), ((asic_acl.srcIpAddr_&0x0000ff00)>>8),
						(asic_acl.srcIpAddr_&0xff), (asic_acl.srcIpAddrMask_>>24), ((asic_acl.srcIpAddrMask_&0x00ff0000)>>16),
						((asic_acl.srcIpAddrMask_&0x0000ff00)>>8), (asic_acl.srcIpAddrMask_&0xff)
					);
					printk("\tTos: %x   TosM: %x   type: %x   typeM: %x   code: %x   codeM: %x\n",
						asic_acl.tos_, asic_acl.tosMask_, asic_acl.icmpType_, asic_acl.icmpTypeMask_, 
						asic_acl.icmpCode_, asic_acl.icmpCodeMask_);
					break;

				case RTL8651_ACL_IGMP:
					printk(" [%d] rule type: %s   rule action: %s\n", start, "IGMP", actionT[asic_acl.actionType_]);
					printk("\tdip: %d.%d.%d.%d dipM: %d.%d.%d.%d\n", (asic_acl.dstIpAddr_>>24),
						((asic_acl.dstIpAddr_&0x00ff0000)>>16), ((asic_acl.dstIpAddr_&0x0000ff00)>>8),
						(asic_acl.dstIpAddr_&0xff), (asic_acl.dstIpAddrMask_>>24), ((asic_acl.dstIpAddrMask_&0x00ff0000)>>16),
						((asic_acl.dstIpAddrMask_&0x0000ff00)>>8), (asic_acl.dstIpAddrMask_&0xff)
					);
					printk("\tsip: %d.%d.%d.%d sipM: %d.%d.%d.%d\n", (asic_acl.srcIpAddr_>>24),
						((asic_acl.srcIpAddr_&0x00ff0000)>>16), ((asic_acl.srcIpAddr_&0x0000ff00)>>8),
						(asic_acl.srcIpAddr_&0xff), (asic_acl.srcIpAddrMask_>>24), ((asic_acl.srcIpAddrMask_&0x00ff0000)>>16),
						((asic_acl.srcIpAddrMask_&0x0000ff00)>>8), (asic_acl.srcIpAddrMask_&0xff)
					);
					printk("\tTos: %x   TosM: %x   type: %x   typeM: %x\n", asic_acl.tos_, asic_acl.tosMask_,
						asic_acl.igmpType_, asic_acl.igmpTypeMask_
					);
					break;

				case RTL8651_ACL_TCP:
					printk(" [%d] rule type: %s   rule action: %s\n", start, "TCP", actionT[asic_acl.actionType_]);
					printk("\tdip: %d.%d.%d.%d dipM: %d.%d.%d.%d\n", (asic_acl.dstIpAddr_>>24),
						((asic_acl.dstIpAddr_&0x00ff0000)>>16), ((asic_acl.dstIpAddr_&0x0000ff00)>>8),
						(asic_acl.dstIpAddr_&0xff), (asic_acl.dstIpAddrMask_>>24), ((asic_acl.dstIpAddrMask_&0x00ff0000)>>16),
						((asic_acl.dstIpAddrMask_&0x0000ff00)>>8), (asic_acl.dstIpAddrMask_&0xff)
					);
					printk("\tsip: %d.%d.%d.%d sipM: %d.%d.%d.%d\n", (asic_acl.srcIpAddr_>>24),
						((asic_acl.srcIpAddr_&0x00ff0000)>>16), ((asic_acl.srcIpAddr_&0x0000ff00)>>8),
						(asic_acl.srcIpAddr_&0xff), (asic_acl.srcIpAddrMask_>>24), ((asic_acl.srcIpAddrMask_&0x00ff0000)>>16),
						((asic_acl.srcIpAddrMask_&0x0000ff00)>>8), (asic_acl.srcIpAddrMask_&0xff)
					);
					printk("\tTos:%x  TosM:%x  sportL:%d  sportU:%d  dportL:%d  dportU:%d\n",
						asic_acl.tos_, asic_acl.tosMask_, asic_acl.tcpSrcPortLB_, asic_acl.tcpSrcPortUB_,
						asic_acl.tcpDstPortLB_, asic_acl.tcpDstPortUB_
					);
					printk("\tflag: %x  flagM: %x  <URG:%x> <ACK:%x> <PSH:%x> <RST:%x> <SYN:%x> <FIN:%x>\n",
						asic_acl.tcpFlag_, asic_acl.tcpFlagMask_, asic_acl.tcpURG_, asic_acl.tcpACK_,
						asic_acl.tcpPSH_, asic_acl.tcpRST_, asic_acl.tcpSYN_, asic_acl.tcpFIN_
					);
					break;

				case RTL8651_ACL_UDP:
					printk(" [%d] rule type: %s   rule action: %s\n", start, "UDP", actionT[asic_acl.actionType_]);
					printk("\tdip: %d.%d.%d.%d dipM: %d.%d.%d.%d\n", (asic_acl.dstIpAddr_>>24),
						((asic_acl.dstIpAddr_&0x00ff0000)>>16), ((asic_acl.dstIpAddr_&0x0000ff00)>>8),
						(asic_acl.dstIpAddr_&0xff), (asic_acl.dstIpAddrMask_>>24), ((asic_acl.dstIpAddrMask_&0x00ff0000)>>16),
						((asic_acl.dstIpAddrMask_&0x0000ff00)>>8), (asic_acl.dstIpAddrMask_&0xff)
					);
					printk("\tsip: %d.%d.%d.%d sipM: %d.%d.%d.%d\n", (asic_acl.srcIpAddr_>>24),
						((asic_acl.srcIpAddr_&0x00ff0000)>>16), ((asic_acl.srcIpAddr_&0x0000ff00)>>8),
						(asic_acl.srcIpAddr_&0xff), (asic_acl.srcIpAddrMask_>>24), ((asic_acl.srcIpAddrMask_&0x00ff0000)>>16),
						((asic_acl.srcIpAddrMask_&0x0000ff00)>>8), (asic_acl.srcIpAddrMask_&0xff)
					);
					printk("\tTos:%x  TosM:%x  sportL:%d  sportU:%d  dportL:%d  dportU:%d\n",
						asic_acl.tos_, asic_acl.tosMask_, asic_acl.udpSrcPortLB_, asic_acl.udpSrcPortUB_,
						asic_acl.udpDstPortLB_, asic_acl.udpDstPortUB_
					);
					break;				

				case RTL8651_ACL_IFSEL:
					printk(" [%d] rule type: %s   rule action: %s\n", start, "UDP", actionT[asic_acl.actionType_]);
					printk("\tgidxSel: %x\n", asic_acl.gidxSel_);
					break;

				case RTL8651_ACL_SRCFILTER:
					printk(" [%d] rule type: %s   rule action: %s\n", start, "Source Filter", actionT[asic_acl.actionType_]);
					printk("\tSMAC: %x:%x:%x:%x:%x:%x  SMACM: %x:%x:%x:%x:%x:%x\n", 
						asic_acl.srcFilterMac_.octet[0], asic_acl.srcFilterMac_.octet[1], asic_acl.srcFilterMac_.octet[2], 
						asic_acl.srcFilterMac_.octet[3], asic_acl.srcFilterMac_.octet[4], asic_acl.srcFilterMac_.octet[5],
						asic_acl.srcFilterMacMask_.octet[0], asic_acl.srcFilterMacMask_.octet[1], asic_acl.srcFilterMacMask_.octet[2],
						asic_acl.srcFilterMacMask_.octet[3], asic_acl.srcFilterMacMask_.octet[4], asic_acl.srcFilterMacMask_.octet[5]
					);
					printk("\tsvidx: %d   svidxM: %x   sport: %d   sportM: %x   ProtoType: %x\n",
						asic_acl.srcFilterVlanIdx_, asic_acl.srcFilterVlanIdxMask_, asic_acl.srcFilterPort_, asic_acl.srcFilterPortMask_,
						(asic_acl.srcFilterIgnoreL3L4_==TRUE? 2: (asic_acl.srcFilterIgnoreL4_ == 1? 1: 0))
					);
					printk("\tsip: %d.%d.%d.%d   sipM: %d.%d.%d.%d\n", (asic_acl.srcFilterIpAddr_>>24),
						((asic_acl.srcFilterIpAddr_&0x00ff0000)>>16), ((asic_acl.srcFilterIpAddr_&0x0000ff00)>>8),
						(asic_acl.srcFilterIpAddr_&0xff), (asic_acl.srcFilterIpAddrMask_>>24),
						((asic_acl.srcFilterIpAddrMask_&0x00ff0000)>>16), ((asic_acl.srcFilterIpAddrMask_&0x0000ff00)>>8),
						(asic_acl.srcFilterIpAddrMask_&0xff)
					);
					printk("\tsportL: %d   sportU: %d\n", asic_acl.srcFilterPortLowerBound_, asic_acl.srcFilterPortUpperBound_);
					break;

				case RTL8651_ACL_DSTFILTER:
					printk(" [%d] rule type: %s   rule action: %s\n", start, "Deatination Filter", actionT[asic_acl.actionType_]);
					printk("\tDMAC: %x:%x:%x:%x:%x:%x  DMACM: %x:%x:%x:%x:%x:%x\n", 
						asic_acl.dstFilterMac_.octet[0], asic_acl.dstFilterMac_.octet[1], asic_acl.dstFilterMac_.octet[2], 
						asic_acl.dstFilterMac_.octet[3], asic_acl.dstFilterMac_.octet[4], asic_acl.dstFilterMac_.octet[5],
						asic_acl.dstFilterMacMask_.octet[0], asic_acl.dstFilterMacMask_.octet[1], asic_acl.dstFilterMacMask_.octet[2],
						asic_acl.dstFilterMacMask_.octet[3], asic_acl.dstFilterMacMask_.octet[4], asic_acl.dstFilterMacMask_.octet[5]
					);
					printk("\tdvidx: %d   dvidxM: %x  ProtoType: %x   dportL: %d   dportU: %d\n",
						asic_acl.dstFilterVlanIdx_, asic_acl.dstFilterVlanIdxMask_, 
						(asic_acl.dstFilterIgnoreL3L4_==TRUE? 2: (asic_acl.dstFilterIgnoreL4_ == 1? 1: 0)), 
						asic_acl.dstFilterPortLowerBound_, asic_acl.dstFilterPortUpperBound_
					);
					printk("\tdip: %d.%d.%d.%d   dipM: %d.%d.%d.%d\n", (asic_acl.dstFilterIpAddr_>>24),
						((asic_acl.dstFilterIpAddr_&0x00ff0000)>>16), ((asic_acl.dstFilterIpAddr_&0x0000ff00)>>8),
						(asic_acl.dstFilterIpAddr_&0xff), (asic_acl.dstFilterIpAddrMask_>>24),
						((asic_acl.dstFilterIpAddrMask_&0x00ff0000)>>16), ((asic_acl.dstFilterIpAddrMask_&0x0000ff00)>>8),
						(asic_acl.dstFilterIpAddrMask_&0xff)
					);
					break;

				default: assert(0);

				}


				/* Action type */
			switch (asic_acl.actionType_) {

			case RTL8651_ACL_PERMIT: /* 0x00 */
			case RTL8651_ACL_REDIRECT: /* 0x01 */
			case RTL8651_ACL_CPU: /* 0x03 */
			case RTL8651_ACL_DROP: /* 0x02, 0x04 */
			case RTL8651_ACL_DROP_LOG: /* 0x05 */
			case RTL8651_ACL_MIRROR: /* 0x06 */
			case RTL8651_ACL_REDIRECT_PPPOE: /* 0x07 */
			case RTL8651_ACL_MIRROR_KEEP_MATCH: /* 0x09 */
				printk("\tdvidx: %d   hp: %d   pppoeIdx: %d   nxtHop:%d  ", asic_acl.dvid_, asic_acl.priority_,
						asic_acl.pppoeIdx_, asic_acl.nextHop_);
				break;

			case RTL8651_ACL_POLICY: /* 0x08 */
				printk("\thp: %d   nxtHopIdx: %d  ", asic_acl.priority_, asic_acl.nhIndex);
				break;

			case RTL8651_ACL_DROP_RATE_EXCEED_PPS: /* 0x0a */
			case RTL8651_ACL_LOG_RATE_EXCEED_PPS: /* 0x0b */
			case RTL8651_ACL_DROP_RATE_EXCEED_BPS: /* 0x0c */
			case RTL8651_ACL_LOG_RATE_EXCEED_BPS: /* 0x0d */
				printk("\trlIdx: %d  ", asic_acl.rlIndex);
				break;
			default: assert(0);
			
			}
			printk("pktOpApp: %d\n", asic_acl.pktOpApp);
			
#if 0
				printk("\tpkgopt: %x\n", asic_acl.pktOpApp);
				if (asic_acl.actionType_ < 8 || asic_acl.actionType_ == 9) {
					printk("\tdvidx: %d   hp: %d   pppoeIdx: %d   nxtHop:%d\n", asic_acl.dvid_, asic_acl.priority_,
						asic_acl.pppoeIdx_, asic_acl.nextHop_
					);
				} else if (asic_acl.actionType_ == 8) {
					printk("\thp: %d   nxtHopIdx: %d\n", asic_acl.priority_, asic_acl.nhIndex);
				} else {
					printk("\trlIdx: %d\n", asic_acl.rlIndex);
				}
#endif
			}

			if (outRule == FALSE) {
				start = asic_vlan.outAclStart; end = asic_vlan.outAclEnd;
				outRule = TRUE;
				goto again;
			}
		}
	}
	else if(strcmp(nextToken, "pvid") == 0) {
		uint32 vidp[9];
		int32  i;

		for(i=0; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++)
			assert(rtl8651_getAsicPvid(i, &vidp[i]) == SUCCESS);
		printk(">> PVID Reg:\n");
		for(i=0; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++)
			printk("p%d: %d,", i, vidp[i]);
		printk("\n");
	}
	else if(strcmp(nextToken, "trunk") == 0) {
		int32	i;
		uint32	portMask, mapping;
		int32 retval;

		retval = rtl8651_getAsicLinkAggregator(&portMask, &mapping);
		assert( retval == SUCCESS );
		printk(">> Trunk Reg:\n");
		printk("   ");
		for(i=0; i<8; i++)
			printk("LMPR%-1d:%d  ", i, ((0x07<<(i*3))& mapping)>>(i*3));
		printk("\n   Trunk1PortMask: %x\n", portMask);
	}
	else if(strcmp(nextToken, "gidx") == 0) {
		uint32 i, gidxReg;
		int32 retval;

		retval = rtl8651_getAsicGidxRegister(&gidxReg);
		assert( retval == SUCCESS );
		printk(">> GIDXReg:\n");
		for(i=0; i<8; i++)
			printk("   gidx%d: %d", i, (gidxReg>>i)&0x7);
		printk("\n");
	}
	else if(strcmp(nextToken, "global") == 0) {
		int8 bool1, bool2;
		uint32 value, i;
		
		printk("NAPT control parameters:\n");
		rtl8651_getAsicAgingFunction(&bool1, &bool2);
		printk("\tAging %s\n", bool2==TRUE? "Enabled": "Disabled");
		rtl8651_getAsicNaptAutoAddDelete(&bool1, &bool2);
		printk("\tAuto Add %s Auto Delete %s\n", bool1==TRUE? "Enabled": "Disabled", bool2==TRUE? "Enabled": "Disabled");
		rtl8651_getAsicNaptIcmpTimeout(&value);
		printk("\tICMP timeout %u seconds\n", value);
		rtl8651_getAsicNaptUdpTimeout(&value);
		printk("\tUDP timeout %u seconds\n", value);
		rtl8651_getAsicNaptTcpLongTimeout(&value);
		printk("\tTCP long timeout %u seconds\n", value);
		rtl8651_getAsicNaptTcpMediumTimeout(&value);
		printk("\tTCP medium timeout %u seconds\n", value);
		rtl8651_getAsicNaptTcpFastTimeout(&value);
		printk("\tTCP fast timeout %u seconds\n", value);
		printk("\tForwarding engine Tx Mirror port ");
		for(i=0;i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum; i++)
			if(rtl8651_tblDrvCpuMirrorPortMask & (1<<i))
				printk("%d ", i);
		printk("\n");	
	}
	else if(strcmp(nextToken, "version") == 0) {
		printk("Table ASIC driver version %s\n", rtl8651_tblAsicDrv_Id);
		printk("Table driver version %s\n", rtl8651_tblDrv_Id);
		printk("Driver forwarding engine version %s\n", rtl8651_tblDrvFwd_Id);
#if 0 //#ifndef CYGWIN
		printk("Switch NIC version %s\n", swNic_Id);
		printk("Net job version %s\n", netJob_Id);
#endif
	}
	else if (strcmp(nextToken, "bandwidth") == 0) {
		int8 *bwd[] = { "full", "128Kbps", "256Kbps", "512Kbps", "1Mbps", "2Mbps", "4Mbps", "8Mbps", "16Mbps", "32Mbps" };
		uint32 rate, port;
		int8 x4;
		rtl8651_getAsicEthernetBandwidthControlX4(&x4);
		for(port=0; port<RTL8651_PORT_NUMBER; port++) {
			if (rtl8651_getAsicEthernetBandwidthControl(port, TRUE, &rate) != SUCCESS)
				return FAILED;
			printk("port %d => in(%s), ", port, rate==0?bwd[rate]:x4?bwd[rate+2]:bwd[rate]);
			if (rtl8651_getAsicEthernetBandwidthControl(port, FALSE, &rate) != SUCCESS)
				return FAILED;
			printk("out(%s)\n", rate==0?bwd[rate]:x4?bwd[rate+2]:bwd[rate]);
		}
		printk("Bandwidth ctrl not available for extension port.\n");
	}
	else if (strcmp(nextToken, "igmp") == 0) {
		MCast_DumpMCast();
	}
	else if (strcmp(nextToken, "mcast") == 0) {
		rtl865x_tblAsicDrv_multiCastParam_t asic;
		uint32 entry;
		
		printk(">>ASIC Multicast Table:\n");
		for(entry=0; entry<RTL8651_MULTICASTTBL_SIZE; entry++) {
			if (rtl8651_getAsicIpMulticastTable(entry, &asic) != SUCCESS) {
				printk("\t[%d]  (Invalid Entry)\n", entry);
				continue;
			}
			printk("\t[%d]  Mip(%d.%d.%d.%d) sip(%d.%d.%d.%d) MBR(%x)\n", entry,
				asic.dip>>24, (asic.dip&0x00ff0000)>>16, (asic.dip&0x0000ff00)>>8, (asic.dip&0xff), 
				asic.sip>>24, (asic.sip&0x00ff0000)>>16, (asic.sip&0x0000ff00)>>8, (asic.sip&0xff),
				asic.mbr);
			printk("\t       svid:%d, spa:%d, EXtIP:%d, age:%d, cpu:%d\n", asic.svid, asic.port, asic.extIdx,
				asic.age, asic.cpu);
		}
	}else if (strcmp(nextToken, "sip-block") == 0) {
		rtl8651_dosProc_blockSip_dumpJail();
	}
	else if (strcmp(nextToken, "conn-ctrl") == 0) {
		rtl8651_dumpConnCtrl();
	}
#ifdef _RTL8651_FWDENG_RETURN_STATISTIC
	else if (strcmp(nextToken, "retval-dump") == 0) {
		_rtl8651_fwdEngineRetValueStaticDump();
	}
#endif
	else if (strcmp(nextToken, "mnqueue") == 0) {
		rtl8651_dumpMNQueueSystemInfo();
	}
	else if (strcmp(nextToken, "nexthop") == 0) 
	{
		rtl865x_tblAsicDrv_nextHopParam_t asic_nxthop;
		rtl8651_tblDrv_nextHopEntry_t *nexthop_t;
		uint32 idx, refcnt, rt_flag;

		printk(">>ASIC Next Hop Table:\n");
		for(idx=0; idx<RTL8651_NEXTHOPTBL_SIZE; idx++) {
			refcnt = rt_flag = 0;
			if (rtl8651_getAsicNextHopTable(idx, &asic_nxthop) == FAILED)
				continue;
			SLIST_FOREACH(nexthop_t , &DrvTbl.inuseList.nexthop, next) {
				if (nexthop_t->entryIndex == idx) {
					refcnt = nexthop_t->refcount;
					rt_flag = nexthop_t->rt_flag;
					break;
				}
			}
			printk("  [%d]  type(%s) IPIdx(%d) dstVid(%d) pppoeIdx(%d) nextHop(%d) rf(%d) rt(%d)\n", idx,
				(asic_nxthop.isPppoe==TRUE? "pppoe": "ethernet"), asic_nxthop.extIntIpIdx, 
				asic_nxthop.dvid, asic_nxthop.pppoeIdx, (asic_nxthop.nextHopRow<<2)+asic_nxthop.nextHopColumn, refcnt, rt_flag);
		}
	}
	else if (strcmp(nextToken, "dns") == 0)
	{
		rtl8651_dns_dataBase_dump();
		rtl8651_dns_dnsProc_dump();
	}
	else if (strcmp(nextToken,"domain-block") == 0)
	{
		rtl8651_domainBlocking_dump();
	}
	else if (strcmp(nextToken, "psa") == 0)
	{
		int i;		
		rtl8651_PS_Action_Entry_t *psa = rtl8651_ProtoStackActions;

		printk(">>Protocol Stack Action Table:\n");
		
		/* lookup the PS action table */
		for (i = 0; i < rtl8651_fwdEnginePara.protoStackMaxActions; i++, psa++)
		{
			printk("       [%d] ", i);
			if (psa->action == PS_ACTION_NONE) {		/* invalid entry */
				printk("invalid\n");
				continue;
			}
			else {
		      		printk("LB(%u.%u.%u.%u) UB(%u.%u.%u.%u)\n",
				  	NIPQUAD(psa->lowerIp), NIPQUAD(psa->upperIp));

				printk("           dstIp(%u.%u.%u.%u) LB(%u) UB(%u)\n", 
					NIPQUAD(psa->dstIp), psa->orgPort, psa->orgPortEnd);

				/* protocol */
				printk("           fwdPort(%u) protocol", psa->fwdPort);
				if (psa->protocol == IPPROTO_TCP)
					printk("(TCP)");
				else if (psa->protocol == IPPROTO_UDP)
					printk("(UDP)");
				else if (psa->protocol == IPPROTO_ICMP)
					printk("(ICMP)");
				else
					printk("(%d)", psa->protocol);

				/* action */
				printk(" action");
				if (psa->action == PS_ACTION_NONE)
					printk("(None)");
				else if (psa->action == PS_ACTION_DROP)
					printk("(Drop)");
				else if (psa->action == PS_ACTION_TO_PS)
					printk("(PS)");
				else if (psa->action == PS_ACTION_DONT_TIMEOUT)
					printk("(NoTmout)");
				else
					printk("(Err)");

				/* role */
				printk(" PS_role");
				if (psa->protocol == IPPROTO_ICMP)
					printk("(none)");
				else {
					if (psa->role == PS_ACTION_CLIENT)
						printk("(client)");
					else if (psa->role == PS_ACTION_SERVER)
						printk("(server)");
					else
						printk("(err)");
				}

				printk("\n");
			}

		}
				

	}
	return SUCCESS;
}




static int32 _rtl8651_asicL34SwitchCmd(uint32 userId, int32 argc, int8 **saved) {
	int8 *nextToken;
	int32 size;

	cle_getNextCmdToken(&nextToken, &size, saved);
	if (strcmp(nextToken, "disable") == 0) //turn off asic layer 3/4 switching. All layer3/4 traffic handled by software
		return rtl8651_pureSoftwareFwd(TRUE);
	else if (strcmp(nextToken, "enable") == 0)//turn on asic layer 3/4 switching.
		return rtl8651_pureSoftwareFwd(FALSE);
	return FAILED;
}


static int32 _rtl8651_counterCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken;
	int32 size;
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if(strcmp(nextToken, "asic") == 0) {
		if(cle_getNextCmdToken(&nextToken,&size,saved)!=FAILED){
			if(strcmp(nextToken, "clear") == 0) {

#ifndef CYGWIN
				rtl8651_clearAsicCounter();
#endif
			} else if(strcmp(nextToken, "complex") == 0) {
#ifndef CYGWIN
				uint32 i;

				printk("Complex counter for port: ");
				for(i=0; i<RTL8651_PORT_NUMBER;i++)
					if(READ_MEM32(MIB_CONTROL) & (1<<(i+24)))
						printk("%d ", i);
				for(i=0; i<rtl8651_totalExtPortNum;i++)
					if(READ_MEM32(MIB_CONTROL) & (1<<(i+21)))
						printk("%d ", i+RTL8651_PORT_NUMBER);
				printk("\n");
				printk("SwitchCore Input counters\n");
				printk("\tRcv %u bytes, Drop %u pkts, CRCAlignErr %u, FragErr %u, JabberErr %u\n", rtl8651_returnAsicCounter(0), rtl8651_returnAsicCounter(0x4), rtl8651_returnAsicCounter(0x8), rtl8651_returnAsicCounter(0xc), rtl8651_returnAsicCounter(0x10));
				printk("\tUnicast %u pkts, Multicast %u pkts, Broadcast %u pkts\n", rtl8651_returnAsicCounter(0x14), rtl8651_returnAsicCounter(0x18), rtl8651_returnAsicCounter(0x1c));
				printk("\t<64 %u, 64 %u, -127 %u, -255 %u, - 511 %u, - 1023 %u, - 1518 %u, oversize %u\n", rtl8651_returnAsicCounter(0x20), rtl8651_returnAsicCounter(0x24), rtl8651_returnAsicCounter(0x28), rtl8651_returnAsicCounter(0x2c), rtl8651_returnAsicCounter(0x30), rtl8651_returnAsicCounter(0x34), rtl8651_returnAsicCounter(0x38), rtl8651_returnAsicCounter(0x3c));
				printk("\tControl unknown %u pkts, Pause %u pkts\n", rtl8651_returnAsicCounter(0x40), rtl8651_returnAsicCounter(0x44));
				printk("SwitchCore Output counters\n");
				printk("\tSnd %u bytes, Unicast %u pkts, Multicast %u pkts, Broadcast %u pkts\n", rtl8651_returnAsicCounter(0x48), rtl8651_returnAsicCounter(0x4c), rtl8651_returnAsicCounter(0x50), rtl8651_returnAsicCounter(0x54));
				printk("\tLate collision %u, Deferred transmission %u \n", rtl8651_returnAsicCounter(0x58), rtl8651_returnAsicCounter(0x5c));
				printk("\tCollisions %u Single collision %u Multiple collision %u pause %u\n", rtl8651_returnAsicCounter(0x60), rtl8651_returnAsicCounter(0x64), rtl8651_returnAsicCounter(0x68), rtl8651_returnAsicCounter(0x6c));

				printk("Simple Port: ");
				for(i=0; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum;i++)
					if(READ_MEM32(MIB_CONTROL_1) & (1<<i))
						printk("%d ",i);
				printk("\n");
				printk("Rx %u Bytes, Rx %u Pkts, Drop %u pkts, ToCPU %u pkts\n", rtl8651_returnAsicCounter(0x80),rtl8651_returnAsicCounter(0x84),rtl8651_returnAsicCounter(0x88),rtl8651_returnAsicCounter(0x8c));
				printk("Tx %u Bytes, Tx %u Pkts\n", rtl8651_returnAsicCounter(0xa0),rtl8651_returnAsicCounter(0xa4));

				printk("Simple Port: ");
				for(i=0; i<RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum;i++)
					if(READ_MEM32(MIB_CONTROL_2) & (1<<i))
						printk("%d ",i);
				printk("\n");
				printk("Rx %u Bytes, Rx %u Pkts, Drop %u pkts, ToCPU %u pkts\n", rtl8651_returnAsicCounter(0x90),rtl8651_returnAsicCounter(0x94),rtl8651_returnAsicCounter(0x98),rtl8651_returnAsicCounter(0x9c));
				printk("Tx %u Bytes, Tx %u Pkts\n", rtl8651_returnAsicCounter(0xa8),rtl8651_returnAsicCounter(0xac));

#endif
			}

		}else{

#ifndef CYGWIN
			uint32 i;

			printk("Port: ");
				for(i=0; i<RTL8651_PORT_NUMBER;i++)
					if(READ_MEM32(MIB_CONTROL) & (1<<(i+24)))
						printk("%d ", i);
				for(i=0; i<rtl8651_totalExtPortNum;i++)
					if(READ_MEM32(MIB_CONTROL) & (1<<(i+21)))
						printk("%d ", i+RTL8651_PORT_NUMBER);
			printk("\n");
			printk("Rx %u Bytes, Rx %u Pkts, Drop %u pkts, ToCPU %u pkts, Rx Pause %u pkts\n", rtl8651_returnAsicCounter(0),
				rtl8651_returnAsicCounter(0x20) + rtl8651_returnAsicCounter(0x24) + rtl8651_returnAsicCounter(0x28) + rtl8651_returnAsicCounter(0x2c) + rtl8651_returnAsicCounter(0x30) + rtl8651_returnAsicCounter(0x34) + rtl8651_returnAsicCounter(0x38) + rtl8651_returnAsicCounter(0x3c),
				rtl8651_returnAsicCounter(0x8), rtl8651_returnAsicCounter(0x74), rtl8651_returnAsicCounter(0x44));
			printk("Tx %u Bytes, Tx %u Pkts Tx Pause %u pkts\n", rtl8651_returnAsicCounter(0x48),
				rtl8651_returnAsicCounter(0x4c) + rtl8651_returnAsicCounter(0x50) + rtl8651_returnAsicCounter(0x54),
				rtl8651_returnAsicCounter(0x6c));

			printk("Port: ");
				for(i=0; i<RTL8651_PORT_NUMBER;i++)
					if(READ_MEM32(MIB_CONTROL_1) & (1<<(i+24)))
						printk("%d ", i);
				for(i=0; i<rtl8651_totalExtPortNum;i++)
					if(READ_MEM32(MIB_CONTROL_1) & (1<<(i+21)))
						printk("%d ", i+RTL8651_PORT_NUMBER);
			printk("\n");
			printk("Rx %u Bytes, Rx %u Pkts, Drop %u pkts, ToCPU %u pkts\n", rtl8651_returnAsicCounter(0x80),rtl8651_returnAsicCounter(0x84),rtl8651_returnAsicCounter(0x88),rtl8651_returnAsicCounter(0x8c));
			printk("Tx %u Bytes, Tx %u Pkts\n", rtl8651_returnAsicCounter(0xa0),rtl8651_returnAsicCounter(0xa4));

			printk("Port: ");
				for(i=0; i<RTL8651_PORT_NUMBER;i++)
					if(READ_MEM32(MIB_CONTROL_2) & (1<<(i+24)))
						printk("%d ", i);
				for(i=0; i<rtl8651_totalExtPortNum;i++)
					if(READ_MEM32(MIB_CONTROL_2) & (1<<(i+21)))
						printk("%d ", i+RTL8651_PORT_NUMBER);
			printk("\n");
			printk("Rx %u Bytes, Rx %u Pkts, Drop %u pkts, ToCPU %u pkts\n", rtl8651_returnAsicCounter(0x90),rtl8651_returnAsicCounter(0x94),rtl8651_returnAsicCounter(0x98),rtl8651_returnAsicCounter(0x9c));
			printk("Tx %u Bytes, Tx %u Pkts\n", rtl8651_returnAsicCounter(0xa8),rtl8651_returnAsicCounter(0xac));

#endif

		}
	}	
	else if(strcmp(nextToken, "pppoe") == 0) {
		uint32 curSessionId, nextSessionId, * sessionIdPtr;
		rtl865x_tblDrv_basicCounterParam_t pppoeCntr;
		sessionIdPtr = rtl8651_getNextPppoeSession(NULL, &nextSessionId);

		if (sessionIdPtr)
		{
			do {
				if(rtl8651_getPppoeCounter(*sessionIdPtr, &pppoeCntr)==SUCCESS) {
					printk("PPPoE session ID %u\n", *sessionIdPtr);
					printk("\tReceive %u packets %u Bytes %u Errors\n", pppoeCntr.rxPackets, pppoeCntr.rxBytes, pppoeCntr.rxErrors);
					printk("\tTransmit %u Packets %u Bytes Drop %u packets\n", pppoeCntr.txPackets, pppoeCntr.txBytes, pppoeCntr.drops);
				}
				curSessionId = *sessionIdPtr;
			} while((sessionIdPtr = rtl8651_getNextPppoeSession(&curSessionId, &nextSessionId))!=NULL);
		}else
		{
			printk("No session present.\n");
		}
	}
	else if(strcmp(nextToken, "fwd-engine") == 0) {

		struct rtl_mBufStatus  mbs;		
		rtl8651_tblDrvFwdEngineCounts_t counter;
		uint32 packetCount=0, ignoredPacketCount=0, linkChangeCount=0;

		if(cle_getNextCmdToken(&nextToken,&size,saved)!=FAILED){//any more tokens means user want a reset
			extern int32 rtl8651_resetFwdEngineCounter(void);
			
			//rtl8651_resetPacketDriverCounters();
			rtl8651_resetFwdEngineCounter();
			return SUCCESS;	
		}
		
		if(mBuf_getBufStat(&mbs)==SUCCESS){
			printk( "\tSizeof\tTotal\tFree\n");
			printk( "Cluster\t%d\t%d\t%d\n",mbs.m_mclbytes, mbs.m_totalclusters, mbs.m_freeclusters);
			printk( "Mbuf\t%d\t%d\t%d\n",mbs.m_msize, mbs.m_totalmbufs, mbs.m_freembufs);
			printk( "Pkthdr\t%d\t%d\t%d\n",mbs.m_pkthdrsize, mbs.m_totalpkthdrs, mbs.m_freepkthdrs);
		}

		//rtl8651_getPacketDriverCounters(&packetCount, &ignoredPacketCount, &linkChangeCount);
		rtl8651_getFwdEngineCounter(&counter);

		printk("RTL8651 table driver forwarding engine:\n");
		printk("\t%u packets received, %u packets to protocol stack\n\t(L34action:%u,  L34failed:%u)\n", packetCount, ignoredPacketCount, counter.l34actionLookupFailed, counter.l34actionFailed);
		printk("\tlink change events %d\n",linkChangeCount);
		
		printk("ARP\t%u arp recv, %u arp replied, %u arp generated\n", 
			counter.arpReceive,  counter.arpReply, counter.arpGenerate);
		printk("\t%u arp entry add success, %u arp entry add failed, %u arp entry deleted\n",
			counter.arpAddSuccess, counter.arpAddFail, counter.arpDelete);
		printk("IP\t%u recv %u, recv with err\n",
			counter.ipReceive, counter.ipErrorReceive);
		printk("ICMP\t%u recv, %u recv with err, %u replied, %u reply failed, %u drop, %u fwd\n",
			counter.icmpReceive, counter.icmpErrorReceive, counter.icmpReply, counter.icmpReplyFailed,
			counter.icmpDrop, counter.icmpSend);
		printk("IGMP\t%u recv, %u recv with err\n",
			counter.igmpReceive, counter.igmpErrorReceive);
		printk("TCP\t%u recv, %u recv with err, %u fwd\n",
			counter.tcpReceive, counter.tcpErrorReceive, counter.tcpSend);
		printk("UDP\t%u recv, %u recieved with err, %d fwd\n",
			counter.udpReceive, counter.udpErrorReceive,counter.udpSend);
		printk("ACL(in)\t%u permit, %u drop, %d cpu\n",
			counter.inAclPermit, counter.inAclDrop,counter.inAclCpu);		
		printk("(out)\t%u permit, %u drop ,%d cpu\n",
			counter.outAclPermit, counter.outAclDrop,counter.outAclCpu);
		printk("Action lookup failed: %d\n", counter.l34actionLookupFailed);
	}

	else if(strcmp(nextToken, "cpuload") == 0) {
#if 0//def RTL865X_DEBUG
		uint32 i;
		printk("Recent CPU loading percentage:\n");
		for(i=0; i<CPU_UTILIZATION_HISTORY; i++)
			printk("[%d]  %8d  / %8d  \n", i, (maxLoopCount-loopCount[i])>>15,maxLoopCount>>15);
#endif

	} else if (strcmp(nextToken, "napt") == 0) {
			uint32	i;
			uint16 insideLocalPort, insideGlobalPort;
			ipaddr_t insideLocalIpAddr;
			int8 isStatic, isTcp, isCollision, isValid, tcpFlag;
			uint32 collision = 0, valid = 0, ageSec, staticCount=0, purecollision=0;
			int8 isCollision2, isDedicated, selExtIpIdx;
			uint16 selEIdx;
			int32 TopN = 5;
			int32 listN = 3;

			if(cle_getNextCmdToken(&nextToken,&size,saved)!=FAILED)
			{
				if(strcmp(nextToken, "clear") == 0)
				{
					_rtl8651_tblDrvStatisticReset();
					return SUCCESS;	
				}
				else
				{
					TopN = I32_value(nextToken);

					if(cle_getNextCmdToken(&nextToken,&size,saved)!=FAILED)
					{
						listN = I32_value(nextToken);
					}
				}
			}
			
			for(i=0; i<RTL8651_TCPUDPTBL_SIZE; i++) {
				if (convert_getAsicNaptTcpUdpTable(NULL, i, &insideLocalIpAddr, &insideLocalPort,
					&insideGlobalPort, &ageSec, &isStatic, &isTcp, &tcpFlag, &isCollision, &isValid,
					&isCollision2, &isDedicated, &selEIdx, &selExtIpIdx) == FAILED)
					continue;
				if (isValid == TRUE)
					valid ++;
				if (isCollision == TRUE){
					collision++;
					if(isValid==FALSE)
						purecollision++;
				}
				if(isStatic==TRUE)
					staticCount++;
			}
			printk("\nASIC statistics\n\n");
			printk("Inuse: %d (%d of them are static)   Free:%d (%d of them are pure collision)\n", valid, staticCount, 1024-valid, purecollision);
			printk("Collision entries: %d", collision);
			rtl8651_getTblDrvNaptInfo( TopN, listN );
#if 0
	} else if (strcmp(nextToken, "icmp") == 0) {
			rtl8651_tblDrvIcmpCounter_t counter;
			if (rtl8651_getTblDrvNaptIcmpCounter(&counter) == FAILED)
				return FAILED;
			printk("\n");
			printk("Active:%d  Inactive:%d  Drvier table:%d  ASIC table:%d  from Asic:%d\n", counter.curActiveCount, counter.curInactiveCount, counter.curInDriverCount, counter.curInAsicCount, counter.curFromAsicCount);
			printk("Diff Hash Algo:%d  MThread wrote back:%d  Sync Fail:%d  Add Fail:%d\n", counter.curDiffHashCount, counter.threadWriteBackCount, counter.syncFailure, counter.addFailure);
			printk("Total Protocol add:%d  Total Protocol remove:%d\n", counter.protoAddCount, counter.protoDelCount);
			printk("Total Driver add:%d  Total Driver remove:%d  Total diff Hash Algo:%d\n", counter.drvAddCount, counter.drvDelCount, counter.diffHashCount);
#endif
	} 
	else if (strcmp(nextToken, "ipsec") == 0) 
	{
#ifdef CONFIG_RTL865X_IPSEC
		rtlglue_printf( "Counter for IPSec\n" );
		rtlglue_printf( "=============================================\n" );
		rtlglue_printf( "SPI count:%d  max=%d\n", CTAILQ_TOTAL( &IpsecEntry.inuseList.spi ), rtl865x_ipsecMaxConcurrentSpi );
		rtlglue_printf( "SPIGrp count:%d  max=%d\n", CTAILQ_TOTAL( &IpsecEntry.inuseList.spigrp ), rtl865x_ipsecMaxConcurrentSpiGrp );
		rtlglue_printf( "ERoute count:%d  max=%d\n", CTAILQ_TOTAL( &IpsecEntry.inuseList.eroute ), rtl865x_ipsecMaxConcurrentERoute );
#else
		rtlglue_printf( "Please enable IPSEC, and re-compile.\n" );
#endif
	}


	return SUCCESS;
}


static int32 _rtl8651_mirrorCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken, isAdd=0, isRx=0;
	int32 size, retval;
	uint16 rxtxport, port;
	
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if (strcmp(nextToken, "reset") == 0) {
		rtl8651_resetRxTxMirror();
		return SUCCESS;
	}
	else if (strcmp(nextToken, "add") == 0) {
		isAdd = TRUE;
	}
	else if (strcmp(nextToken, "remove") == 0) {
		isAdd = FALSE;
	}
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if (strcmp(nextToken, "rx") == 0) {
		isRx = TRUE;
	}
	else if (strcmp(nextToken, "tx") == 0) {
		isRx = FALSE;
	}
	cle_getNextCmdToken(&nextToken,&size,saved); 
	rxtxport = U16_value(nextToken);	
	if ((retval=rtl8651_addRxTxMirror(isRx, rxtxport)) != SUCCESS)
		return retval;
	if (cle_getNextCmdToken(&nextToken,&size,saved) == FAILED)
		return SUCCESS;
	if (strcmp(nextToken, "mirror-port") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		port = U16_value(nextToken);
		if (isAdd == TRUE)
			return rtl8651_addMirrorPort(port);
		else return rtl8651_delMirrorPort(port);
	}
	return FAILED;
}


static int32 _rtl8651_urlFilterCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken;
	int32 size, len;
	
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if (strcmp(nextToken, "add") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		len=strlen(nextToken);
		return rtl8651_addURLfilterString(nextToken, len);
	}else{
		cle_getNextCmdToken(&nextToken,&size,saved); 
		len=strlen(nextToken);
		return rtl8651_delURLfilterString(nextToken, len);
	}
	return FAILED;
}

static int32 _rtl8651_miscCmd(uint32 userId,  int32 argc,int8 **saved){
	int8 *nextToken;
	int32 size;
	
	cle_getNextCmdToken(&nextToken,&size,saved); 
	if (strcmp(nextToken, "mtu") == 0) {
		cle_getNextCmdToken(&nextToken,&size,saved); 
		if (strcmp(nextToken, "set") == 0) {
			cle_getNextCmdToken(&nextToken,&size,saved); 
			rtl8651_setMtu(U16_value(nextToken));
		} else if (strcmp(nextToken, "show") == 0) {
			uint16 mtu;
			rtl8651_getMtu(&mtu);
			printk("Current mtu limit is %u\n", mtu);
		}
		return SUCCESS;
	} else if (strcmp(nextToken, "psnet") == 0) {
		if (strcmp(nextToken, "enable") == 0) {
			rtl8651_fwdEnginePppoeNotExistedSessionTermination(TRUE);
		} else if (strcmp(nextToken, "disable") == 0) {
			rtl8651_fwdEnginePppoeNotExistedSessionTermination(FALSE);
		}
	
	} else if (strcmp(nextToken, "reset") == 0) {
		printk("Start switch core reset\n");
		rtl8651_resetSwitchCoreStore();
		swNic_descRingCleanup();

		rtl8651_resetSwitchCoreActionAndConfigure();
#ifdef CONFIG_RTL865XC
		swNic_setup(totalRxPkthdr0,totalRxPkthdr1,totalRxPkthdr2,totalRxPkthdr3,totalRxPkthdr4,totalRxPkthdr5,
					totalRxMbuf0,
					totalTxPkthdr0,totalTxPkthdr1);
#else
		swNic_setup(totalRxPkthdr, totalRxMbuf, totalTxPkthdr);
#endif
		printk("Finish switch core reset\n");
	}
	return FAILED;
}

static  int32 _checkSTPstate(void **token, int8 *input, int32 size, cle_exec_t **nextCmd){
	int32 i,numsize =cle_checkDynCmdParam_String(token, input, size, nextCmd);
	if(numsize)
		return FAILED;
	for(i=0;i < STP_STATES;i++)
		if(strcmp(input, _stpStates[i])==0)
			break;
	if(i==STP_STATES)
		return FAILED;
	 *((int32 *) *((memaddr *)token))=i;
	return 4;

}


static int32 _enumSTPstate(int32 entries, struct enumList *list){
	int32 i;

	if (list && entries < STP_STATES)
		return FAILED; //No enough buffer space
	else if(list){
		for(i=0; i < STP_STATES; i++){
			list[i].tokenName =(int8 *)_stpStates[i];
			list[i].helpString= "";
		}
	}
	return STP_STATES;
}


static  int32 _checkACLaction(void **token, int8 *input, int32 size, cle_exec_t **nextCmd){
	int8 *dummy;
	int32 i,numsize =cle_checkDynCmdParam_String((void **)&dummy, input, size, nextCmd);
	if(numsize)
		return FAILED;
	for(i=0;i < ACL_ACTIONS;i++)
		if(strcmp(input, _aclActions[i])==0)
			break;
	if(i==ACL_ACTIONS)
		return FAILED;
	 *((int32 *) *((memaddr *)token))=i+1; //avoid returns 0
	return 4;
}


static int32 _enumACLaction(int32 entries, struct enumList *list){
	int32 i;

	if (list && entries < ACL_ACTIONS)
		return FAILED; //No enough buffer space
	else if(list){
		for(i=0; i < ACL_ACTIONS; i++){
			list[i].tokenName =(int8 *)_aclActions[i];
			list[i].helpString= "";
		}
	}
	return ACL_ACTIONS;
}







static cle_paramType_t _checkFunc[] ={
	{"%stpState", 4, _checkSTPstate, 	_enumSTPstate,	"Spanning tree port state"},
	{"%aclAction", 4, _checkACLaction, 	_enumACLaction,	"ACL action type"},
};

cle_exec_t rtl8651_tblDrv_cmds[] = {
	{	"alg-trap",
		"Configure ALG trapping ports",
		"{ add'Add ALG trapping port(s)' | remove'Remove ALG trapping port(s)' } [ tcp'trap if protocol is TCP' ] { client'trap if internal host acts as client' | server'trap if internal host acts as server' | both'trap whichever sport or dport matches' } port'specify port(s) to trap' %d'port number to trap(lowest)'  [ to'specify a port range for trapping' %d'highest port number to trap' ] ",
		_rtl8651_algTrapCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"arp",
		"Configure the ARP table",
		"{ add'Add a new ARP entry' %ip'IP address' %mac'MAC address' %s'Network interface name' %d'Port number(0xffffffff for all ports)' | remove'Remove an existing ARP entry' %ip'IP to remove' } ",
		_rtl8651_arpCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"asic-L34switching", 
		"Turn on/off ASIC Layer3 and Layer4 features.",
		"{ enable'(default)Enable ASIC Layer 3/4 switching engine.' | disable'All Layer 3/4 traffic handled by software.' }",
		_rtl8651_asicL34SwitchCmd,
		CLE_USECISCOCMDPARSER,
		0,
		NULL
	},	
	{	"counter",
		"Display statistic information",
		"{ { asic'Show ASIC counters' [ clear'Reset ASIC counter values'  | complex'Get full set counter' ] } | "
		"  { pppoe'Show statistics of PPPoE sessions' } | "
		"  { napt'Show statistics of NAPT from ASIC and table driver' [ clear'Clear all NAPT counters' | %d'TopN' [ %d'listN' ] ] } | "
		"  { fwd-engine'Show statistics collected at forwarding engine' [ clear'Reset fwd engine counters' ] } | "
		"  { cpuload'Show recent CPU load average' } | "
		"  { ipsec'Show IPSec counter' } | "
		"  { icmp } }",
		_rtl8651_counterCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},	
	{	"dump",
		"Dump all ASIC table",
		"{ hstamp | queueInfo | qos'QoS' | mac-filter'Mac Filter' | url-filter'URL Filter' | "
		"  voip'VoIp Supporting' | ipsec'Dump IPSec' | "
		"l2'Dump Layer2 table' | bandwidth'Dump port bandwidth' | flow-control'Flow Control for each port' | "
		"l3'Dump Layer3 table' | soft-l3'Dump soft l3 table' | arp'Dump ARP table' | drv-arp'Dump Driver ARP table' | pppoe'Dump PPPoE session ID table' | "
		"napt'Dump NAPT table' [ drv'List NAPT entries in driver' | asic'List NAPT entries in ASIC' ] "
			"{ extport'search external port in table' %d'external port number' | "
			"  %d'entry index'  | "
			"  { all'List all'  | until'Specify end index'  %d'End index' | "
			"    from'Specify starting index'  %d'Starting  index' { all'List all'  | until'Specify end index' %d'End index' } "
			"  } "
			"} | "
		"icmp'Dump ICMP table' | "
		"drv_icmp'Dump ICMP entries in driver' | "
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		"drv_pppoepassthru'Dump PPPoE Passthru configuration in driver' | "
#else//CONFIG_RTL865XB
		"drv_pppoepassthru'Dump PPPoE Passthru entries in driver' | "
#endif//CONFIG_RTL865XB
#if 1//CONFIG_RTL865XB_EXP_IPV6_PASSTHRU
		"drv_ipv6passthru'Dump IPv6 Passthru configuration in driver' | "
#else//CONFIG_RTL865XB
		"drv_ipv6passthru'Dump IPv6 Passthru entries in driver' | "
#endif//CONFIG_RTL865XB
#ifdef CONFIG_RTL865XB
		"pbvlan'Dump Protocol-based VLAN' | "
#endif//CONFIG_RTL865XB
		"idle_timeout'Dump Idle Timeout counter in driver' | "
		"pbnat'Dump Protocol=based NAT in driver' | "
		"mirror-port'Dump Mirror Port' | "
		"igmp'Dump IGMP table and setting in driver' | "
		"sip-block'Dump blocked source ip address in driver' | "
		"conn-ctrl'Dump NAPT connection counting ctrl system in driver' | "
#ifdef _RTL8651_FWDENG_RETURN_STATISTIC
		"retval-dump'Dump Forwarding engine return value count' | "
#endif
		"mcast'Dump ASIC multicast table' | phy'PHY' | rate-limit | "
		"ip'Dump IP table' | serverPort'Dump server port table' | alg'Dump ALG table' | vlan'Dump VLAN table' | prototrap'Dump protocol trap table' | "
		"acl'Dump ACL table'  | pvid'Dump PVID of each port' | trunk'Dump trunking ports' | gidx'Dump IP group index register' | global'Show global control parameters' | version'Show version id of each software module' | "
		"nexthop'Dump ASIC nexthop table' | mdimdix'Dump auto MDI/MDIX status' | timer'Dump Timer Information' | policy-route'Dump policy route' | "
		"mnqueue'Dump MN queue information' | "
		"dns'Dump DNS module information' | "
		"psa'Dump PS Action' | "
		"domain-block'Dump domain-blocking database information' | "
		"acldb'Dump ACLDB table' [ %d'start rule number' %d'end rule number' ] } ",
		_rtl8651_dumpCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"filter-database",
		"Configure the filter database",
		"%d'Filter database id' { create'Create this filter DB' | delete'delete this filter DB' | { add'Add a new MAC entry' %mac'MAC to add'  { forward'Normal forwarding' | dst-block'Block dest MAC' | src-block'Block src MAC' | trap'Trap pkt to software'  }  %d'Portmask (in hexadecimal format)' } "
							"| remove'Remove an existing entry' %mac'MAC to remove' | sid'Specify spanning tree id' %d'Spanning tree id' } ",
		_rtl8651_filterDBCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},		
	{	"interface",
		"Configure network interfaces",
		" { %str'Network interface name' { create'Create this network interface' | delete'Delete this network interface' | "
		" acl'Set ACL attribute' { { def-acl'Default ACL action' { permit | drop | cpu } | flush'Flush ACL rules' { Ingress | Egress | all } } }  | "
		" type'specify link layer type' { { ether'Link layer is ethernet bcast interface' | pppoe'Link layer is PPP over Ethernet' | unnumbered'Link layer is PPPoE/Ethernet' } vid %d'specify belonging vlan id' } | detach'detach from link layer'  | "
		"{ ip'Configure IP subnets/interfaces' { add'Add a new IP interface' | remove'Remove an existing IP interface' } %ip'IP address' %ip'Network mask'  { LAN'Configure interface as LAN' | WAN'Configure interface as WAN' } }  | "
		" mtu %d'MTU' |  "
//		"{ dial'Dial PPPoE session(s)' (1~8)%d'PPPoE ID' } | { hangup'Hangup PPPoE session(s)' (1~8)%d'PPPoE ID' }  "
		" } } ",
		_rtl8651_ifaceCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},	
	{	"ip-acl",
		"Configure TCP/IP/ICMP/IGMP/UDP access control list",
		" { %str'network interface name' } { { add | remove } %str'Specify ACL action' { ingress'Ingress filter rule' | egress'Egress filter rule' } "
		"{ [ range'IP Range' ]  [ sip'filter source ip' %ip'src IP' [  %ip'sip mask' ] ] [ dip'filter dest IP' %ip'dest IP' [ %ip'dip mask' ] ] [ tos'filter IP TOS' %d'TOS' [ %d'TOS mask' ] ] "
		" [ { ip'filter more IP specific fields...' { [ ipproto'filter IP protocol type' %d'Protocol' [ %d'Protocol mask' ] ] [ ipflag'filter IP flags' %d'IP flags' [ %d'IP flag mask' ] ] } | "
		"{ icmp'filter more ICMP specific fields...' { [ icmptype'filter ICMP type' %d'ICMP type' [ %d'ICMP type mask' ] ] [ code'filter ICMP code' %d'ICMP code' [ %d'ICMP code mask' ] ] } | "
		"{ igmp'filter more IGMP specific fields...' { [ igmptype'filter IGMP type' %d'ICMP type' [ %d'ICMP type mask' ] ] } } | "
		"{ tcp'filter more TCP specific fields...' { [ tcpflag'filter TCP flags' %d'TCP flags' [ %d'TCP flag mask' ] ] [ sport'filter TCP src port' %d'src port(lowest)'  [ %d'last(highest) src port' ] ] [ dport'specify TCP dst port' %d'dest port'  [ %d'last dest port' ] ] } } | "
		"{ udp'filter more UDP specific fields...' { [ sport'filter UDP src port' %d'src port'  [ %d'last src port' ] ] [ dport'filter UDP dst port' %d'dest port(lowest)'  [ %d'last(highest) dest port' ] ] } } "
		"} ] } } }",
		_rtl8651_ipACLCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},	
	{	"ip-unnumbered",
		"Specify IP unnumbered interfaces",
		" { %s'WAN side network interface name'  [ %s'LAN side network interface name' ] %ip'unnumbered IP subnet mask' } ",
		_rtl8651_ipUnnumCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"local-server",
		"Configure the local server IP table",
		"{ add'Add a new local server IP' | remove'Remove a local server IP' } %ip'Local server IP address' ",
		_rtl8651_localserverCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},	
	{	"mac-acl",
		"Configure MAC access control lists",
		" %str'network interface name' { { add'Add a MAC ACL rule' | remove'Remove a MAC ACL rule' } %aclAction'Specify ACL action' { ingress'Ingress filter rule' | egress'Egress filter rule' } { [ smac'filter source MAC' %mac'src MAC' [ %mac'specify src MAC mask' ] ] [ dmac'filter dest MAC' %mac'dst MAC' [ %mac'48bit dst MAC mask' ] ] [ ethtype'Filter ether type' %d'EtherType' [ %d'EtherType mask' ] ] } } ",
		_rtl8651_macACLCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"mii",
		"Configure MII port",
		"{ { %d'set MII Phy address' { mac'mac mode' | phy'phy mode' } } | get'Get MII status' }",
		_rtl8651_miiCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"mirror",
		"Configure mirror port",
		"{ reset'Reset mirror port control register' | "
		"{ add'Add' | remove'Remove' } "
		"{ rx'set Mirror Rx' %d'Specify port number' | "
		"tx'set Mirror Tx' %d'Specify port number' } "
		"[ mirror-port'Set Mirror port' %d'Specify port number' ] } ",
		_rtl8651_mirrorCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"napt",
		"Configure the NAPT table",
		"{ timeout'specify table entry timeout values' { "
		"  [ adaptive'enable/disable adaptive adjustment of NAPT expiration time' { %d'threshold' | off } ] "
		"  [ smart-swap'enable/disable smart swapping of NAPT entries by load' { %d'pkts/sec threshold' | off } ] "
		"  [ smart-swap-tcp'smart swapping threshold of TCP entries' { %d'pkts/sec threshold' } ] "
		"  [ smart-swap-udp'smart swapping threshold of UDP entries' { %d'pkts/sec threshold' } ] "
		"  [ smart-swap-cdr'smart swapping threshold for CDRouter' { %d'pkts/sec threshold' } ] "
		"  [ icmp'specify ICMP entry age out time' %d'timeout value' ] "
		"  [ icmp-fast'specify ICMP ACKed entry age out time' %d'timeout value' ] "
		"  [ udp'specify UDP entry age out time' %d'timeout value' ] "
		"  [ tcp-long'specify established TCP age out time' %d'timeout value' ] "
		"  [ tcp-medium'specify closing TCP entry age out time'  %d'timeout value' ] "
		"  [ tcp-short'specify establishing TCP entry age out time' %d'timeout value' ] } | "
		" autolearn'Config ASIC NAPT auto learn feature' { enable'Let ASIC create NAPT entries automatically' | disable'Let software creates all NAPT connections' } | "
		" portrange'specify port range used for TCP/UDP NAPT' %d'first port number' to'specify last port number' %d'last port number' | "
		" flush'flush all NAPT entries' | "
		" { add'Add a new NAPT entry' "
			"{ { { tcp'add a TCP entry' | udp'add a UDP entry' } %ip'inside local IP' %d'inside local port' { [ dst'specify destination (outside local) IP and port' %ip'destination (outside local) IP'  %d'destination(outside local) port' ] "
			" [ use'specify inside global IP and port dst use in NAPT' %ip'inside global IP'  %d'inside global port' ] } } | "
			" { icmp'add an ICMP entry' %ip'inside local IP' %d'inside local icmp ID' %ip'destination IP'  [ use'specify an inside global IP' %ip'inside global IP for NAPT' %d'inside global icmp ID' ] } | "
			" %ip'Add an IP into NAPT pool' [ dmzhost'Add DMZ host for this NAPT IP' %ip'Intenal host IP' ] } } | "
		"{ remove'remove an NAPT entry' "
			"{ { { tcp'remove a TCP entry' | udp'remove a UDP entry' } %ip'inside local IP' %d'inside local port' [ dst'specify destination (outside local) IP and port' %ip'destination (outside local) IP'  %d'destination(outside local) port' ] } | "
			" icmp'remove an ICMP entry' %ip'inside local IP' %d'inside local icmp ID' dst'specify destination IP' %ip'destination IP'  use'specify an inside global IP' %ip'inside global IP for NAPT' %d'inside global icmp ID' | "
			" %ip'Remove an IP from NAPT pool' [ dmzhost'Remove DMZ host of this NAPT IP' ] } } }",
		_rtl8651_naptCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"nat",
		"Configure the NAT address mapping table",
		"{ add'Add a new NAT address mapping' | remove'Remove NAT mapping' } %ip'inside global IP address' %ip'inside local IP address' ",
		_rtl8651_natCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"phy-port",
		"Configure physical ports",
		"%d'Port number(s)' { "
		"  cable-meter'Cable diagnositic ' | "
		"  { loopback { on | off } } | "
		"  auto-negotiated { [ 100Full | 100Half | 10Full | 10Half ] } | "
		"  force { [ 10Half | 10Full | 100Half | 100Full ] } | "
		"  { bandwidth { in [ full | 128K | 256K | 512K | 1M | 2M | 4M | 8M ] | "
		"                out [ full | 128K | 256K | 512K | 1M | 2M | 4M | 8M ] } } | "
		"  power { up | down } "
		"}",
		_rtl8651_phyCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},	
	{	"port",
		"Configure (logical) ports",
		"%d'Port number' { pvid'set PVID' %d'PVID' | { individual'Do not aggregate with any other ports' | member-port'Specify member ports' %d'new member port number' } } ",
		_rtl8651_portCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},	
	{	"pppoe",
		"Configure pppoe session table",
		"%d'PPPoE id' { create-on-vlan'Create and attach this PPPoE session on VLAN' %d'VLAN id' | "
			"delete'Delete this PPPoE session' | reset'Reset PPPoE properties'  | bind'bind pppoe object to PPP network interface' %s'PPP interface name' | get'Get PPPoE info'  { src-mac'Get source MAC' } | "
			" set'Set PPPoE properties' { [ user'set username' %s'Username' ] [ passwd'set password' %s'Password' ] [ session'set session ID, peer MAC and port.' %d'PPPoE session Id'  %mac'peer MAC' %d'port number' ] [ { nailup-interval'specify delay before redial if link hanged-up by server' %d'nailup interval (seconds)' | no-nailup'do not redial after server forced hangup link' } ] "
			"   [ { inactivity-timeout'specify max inactivity timeout before hangup' %d'inactivity time(seconds)' | always-online'always online. do not active hangup connection' } ] [ retry'specify #retries' %d'max # of retries' ] [ host-uniq'Host unique tag' %s'Host unique tag' ] }  } ",
		_rtl8651_pppoeCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},	
	{
		"proto-trap",
		"Configure Protocol Trap",
		"{ add'Add Protogol Trapping' | remove'Remove Protocol Trapping' } { ether'trap ether type' %d | ip'trap ip protocol' %d | tcp-port'trap TCP port number' %d | udp-port'trap UDP port number' %d }",
		_rtl8651_protoTrapCmd,
		CLE_USECISCOCMDPARSER,
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"route",
		"Configure the routing table",
		"{ add'Add a new routing entry' { default'Add default route' | { %ip'Destination host or network IP' [ %ip'Network mask(for network route)' ] } } use'sepcify outgoing interface name' %str'outgoing interface name' via'specify next hop' %ip'specify next hop IP'  | remove'Remove a routing entry' { default'Remove default route' | { %ip'destination host or network IP to delete' [ %ip'netmask for network route' ] } } } ",
		_rtl8651_routeCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"server-port",
		"Configure the server port mapping table",
		"{ add'Add a new internal server IP and port pair' | remove'Remove an internal server IP and port mapping' } { tcp | udp } %ip'Internal server global IP address' %d'Internal server global port' %ip'Internal server local IP address' %d'Internal server local port'",
		_rtl8651_serverPortCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{	"spanning-tree",
		"Configure spanning tree instances",
		"%d'Spanning tree id'  { create'Create this spanning tree instance' | delete'delete this spanning tree instance' | port'Set port status' %d'port number' %stpState | protocol'Spanning tree protocol running status' { enable'Spanning tree protocol is running' | disable'Spanning tree protocol is not running' } }",
		_rtl8651_stpCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},	
	{	"url-filter",
		"Configure URL filter",
		" { add'Add a URL filter string' | remove'Remove a URL filter string' } %s'Filter string' ",
		_rtl8651_urlFilterCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},	
	{	"vlan",
		"Configure VLANs ",
		"%d'VLAN id' { create'Create this VLAN' | delete'Delete this VLAN' | "
			"{ [ base-mac'Specify base MAC' %mac'base MAC' %d'Specify # of continuous MACs to reserve after base MAC' ] [ max-macs'Specify max MACs' %d'max MAC number' ] "
			"[ filter-db'specify which filter DB to use' %d'filter DB id' ]  "
			"[ member-port'Configure member ports' { add | remove | tagged } %d'Member port number' ] } "
			"[ fwdTxMirror'Configure VLAN to mirror forwarding engine packet transmission' { enable | disable } ] }",
		_rtl8651_vlanCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"flow-control",
		"Configure Flow Control",
		"{ %d'Port number' { enable'Enable the specified port Flow Control' | disable'Disable the specified port Flow Control' | auto'Auto' } | threshold { [ inq-high | inq-low | outq-high | outq-low | inopt | outopt ]  } %d'threshold value' } ",
		_rtl8651_flowControlCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"qos",
		"Configure QoS",
		" { rr-ratio'Configure Round-Robin ratio' [ { always-high'Always high priority queue first' | h16_l1'16:1' | h8_l1'8:1' | h4_l1'4:1' } ]  |"
		" port'Configure Port High Low Priority' [ %d'Port number' { high | low } ] | "
		" diffserv'Configure DiffServer' [ %d'dscp value' { high | low } ] }  ",
		_rtl8651_qosCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"mac-filter",
		"Configure MAC Filter.",
		" { enable'Enable MAC Filter' | disable'Disable MAC Filter' | default'Set default action' [ { permit | drop } ]  | "
		" add'Add MAC Filter rule' [ %mac'MAC address' %mac'MAC Mask' [ { permit | drop } ] ] | "
		" remove'Remove MAC Filter rule' [ %mac'MAC address' %mac'MAC Mask' ]  | flush'Flush all rule' } ",
		_rtl8651_macFilterCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"pppoepassthru",
		"Enable/Disable PPPoE Passthru.",
		" { enable'Enable PPPoE Passthru' | disable'Disable PPPoE Passthru' |"
		"   dropunknown'Drop Unknown PPPoE PADT' { enable'Enable' | disable'Disable' } } ",
		_rtl8651_pppoePassthruCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"pbnat",
		"Enable/Disable Protocol-based NAT.",
		" { enable'Enable Protocol-based NAT' | disable'Disable Protocol-based NAT' | "
		" add'Add an NAT entry' %d'protocol' %ip'External IP' %ip'Internal IP' | "
		" del'Delete an NAT entry' %d'protocol' %ip'External IP' %ip'Internal IP' } ",
		_rtl8651_pbNatCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"misc",
		"Other simple command",
		" { mtu { set %d'Configured MSS limitation' | show'Display current MSS Limitation' } | psnet'PPPoE Session Not Existed Termination' { enable | disable } | reset'Reset switch core' }",
		_rtl8651_miscCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"ipv6passthru",
		"Enable/Disable IPv6 Passthru.",
		" { enable'Enable PPPoE Passthru' | disable'Disable PPPoE Passthru' } ",
		_rtl8651_ipv6PassthruCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"pptppassthru",
		"PPTP Passthru command",
		" { dump'Show PPTP pass-through session information' } ",
		_rtl8651_pptpPassthruCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"sys-mtu",
		"Configure whole system's MTU.",
		" { %d'MTU value' } ",
		_rtl8651_sysMtuCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"romereal",
		"ROME Driver sniffer",
		" { init'Initialize buffer' %d'buffer address' %d'buffer size' %d'mode'| "
		"   start'start capture packet' | "
		"   stop'stop capture packet' | "
		"   dump'dump captured packets' | "
		"   filter'set capture filter' %d'filter'| "
		"   save'save captured packets into libpcap file format' %s'filename' } ",
		_rtl8651_romerealCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"romeperf",
		"ROME Performance Profile",
		" { reset'reset statistic data' | "
		"   dump'dump statistic results' { [ %d'from index' %d'to index' ] } | "
		"   iram'test IRAM performance' | "
		"   dram'test DRAM performance' | "
		"   bmp'test Boyer-Moore performance' | "
		"   morethan4096'test performance when napt has more than 4096 connections' | "
		"   pause'pause profilling' | "
		"   resume'resume profilling' } ",
		_rtl8651_romeperfCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"mdimdix",
		"MDI/MDIX Configuration.",
		" { { [ select'Select MDI/MDIX' { mdi | mdix } ] | "
		" enable'Enable Auto CrossOver Detection' | "
		" disable'Disable Auto CrossOver detection' } [ %d'port0-port4' | all'apply to all' ] } ",
		_rtl8651_autoMdiMidxCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"ipsec",
		"IPSEC Configuration.",
		" { init'Initialization' | dump'dump SPI/SPIGRP/EROUTE' | flush'flush all SPI/SPIGRP/EROUTE' |"
		"   spi'Security Parameter Index' af inet edst %ip'Effetive Destination IP address' spi %d'SPI' proto "
		"       { esp src %ip'Source IP address' esp { 3des-md5-96 | 3des-sha1-96 | des-md5-96 | des-sha1-96 | aes-md5-96 | aes-sha1-96 } enckey %s'Encryption Key' authkey %s'Authentication Key' | "
		"         tun src %ip'Source IP address' dst %ip'Destination IP address' ip4 } |"
		"   spigrp'SPI Group' inet %ip'IP address 1' %d'SPI1' %s'Proto1' [ inet %ip'IP address 2' %d'SPI2' %s'Proto2' "
		"                   [ inet %ip'IP address 3' %d'SPI3' %s'Proto3' [ inet %ip'IP address 4' %d'SPI4' %s'Proto4' ] ] ] |"
		"   eroute'Enhanced Route' add eraf inet src %ip'Source Net' %d'Source CIDR' dst %ip'Destination Net' %d'Destination CIDR' said %s'said(tun0x1001@192.168.154.240)' } ",
		_rtl8651_ipsecCmd,
		CLE_USECISCOCMDPARSER,	
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"dns",
		"DNS module Configuration.",
		" set ttl-suppression'suppress A/PTR/CNAME RRs TTL field' { enable'enable suppression' ttl %d'TTL upperbound' | disable'disable suppression' } ",
		_rtl8651_dnsCmd,
		CLE_USECISCOCMDPARSER,
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
	{
		"domain-block",
		"DOMAIN BLOCKING Configuration.",
		" { add'add domain-blocking entry' | del'del domain-blocking entry' } src %ip'Source IP address' mask %ip'Source IP mask' domain %s'blocked domain' ",
		_rtl8651_domainBlockCmd,
		CLE_USECISCOCMDPARSER,
		sizeof(_checkFunc)/sizeof(cle_paramType_t),
		(cle_paramType_t *)&_checkFunc
	},
};


