
/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : Switch core table driver testing code
* Abstract : 
* Author : Chih-Hua Huang (chhuang@realtek.com.tw)               
* $Id: tblTest.c,v 1.1.1.1 2006/07/26 05:38:29 blue Exp $
*/
/* Implementation note
1. All information are implemented by host order
2. Expect tblAsicDrv to process the order issue
3. 
*/


#include "drvTest.h"
#include "mbufGen.h"
#include "rtl8651_layer2.h"
#include "rtl8651_layer2local.h"
#include "rtl8651_tblDrv.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblAsicDrv.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrv.h"
#include "asicRegs.h"



#define RTL8651_TESTING_DEFAULT_CONFIGURATION

static int32 default_configuration1(uint32 type1, uint32 type2)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask;

	/* VLAN 1: subnet 1 */	
	ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("vlan1") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan1", type1, 1) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.1.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("vlan1", ipaddr, mask) == SUCCESS);

	/* VLAN 2: subnet 2 */
	mac.octet[5] = 0x02;
	ASSERT(rtl8651_addVlan(2) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("vlan2") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan2", type2, 2) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.2.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("vlan2", ipaddr, mask) == SUCCESS);
	return SUCCESS;
}


static int32 default_configuration2(uint32 type1, uint32 type2)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask;

	/* WAN: subnet 1: */	
	ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("wan") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("wan", type1, 1) == SUCCESS);
	ipaddr = ntohl(inet_addr("140.113.1.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("wan", ipaddr, mask) == SUCCESS);
	ASSERT(rtl8651_addExtNetworkInterface("wan") == SUCCESS);

	/* VLAN: subnet 2 */
	mac.octet[5] = 0x02;
	ASSERT(rtl8651_addVlan(2) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("vlan") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan", type2, 2) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.1.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("vlan", ipaddr, mask) == SUCCESS);
	return SUCCESS;
}


static int32 default_configuration3(void)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask;

	/* VLAN 1: subnet 1 */	
	ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("pppoe") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("pppoe", RTL8651_LL_PPPOE, 1) == SUCCESS);
	ASSERT(rtl8651_addPppoeSession(1000, 1) == SUCCESS);

	/* VLAN 2: subnet 2 */
	mac.octet[5] = 0x02;
	ASSERT(rtl8651_addVlan(2) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("lan") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("lan", RTL8651_LL_VLAN, 2) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.2.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("lan", ipaddr, mask) == SUCCESS);
	return SUCCESS;

}


static int32 default_configuration3_1(void)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask;

	/* VLAN 1: subnet 1 */	
	ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("pppoe") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("pppoe", RTL8651_LL_PPPOE, 1) == SUCCESS);
	ASSERT(rtl8651_addPppoeSession(1000, 1) == SUCCESS);

	/* VLAN 2: subnet 2 */
	mac.octet[5] = 0x02;
	ASSERT(rtl8651_addVlan(2) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("lan0") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("lan0", RTL8651_LL_VLAN, 2) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.1.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("lan0", ipaddr, mask) == SUCCESS);

	/* VLAN 3: subnet 3 */
	mac.octet[5] = 0x03;
	ASSERT(rtl8651_addVlan(3) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(3, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("lan1") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("lan1", RTL8651_LL_VLAN, 3) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.2.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("lan1", ipaddr, mask) == SUCCESS);	
	return SUCCESS;
}




static int32 default_configuration4(void)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask;

	/* IP Unnumber: wan0 */
	ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("wan0") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("wan0", RTL8651_LL_PPPOE, 1) == SUCCESS);
	ASSERT(rtl8651_addPppoeSession(1000, 1) == SUCCESS);
	ASSERT(rtl8651_addExtNetworkInterface("wan0") == SUCCESS);

	/* IP Unnumber: wan1 */
	mac.octet[5] = 0x02;
	ASSERT(rtl8651_addVlan(2) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("wan1") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("wan1", RTL8651_LL_VLAN, 2) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.1.1"));
	mask = ntohl(inet_addr("255.255.255.240"));
	ASSERT(rtl8651_addIpIntf("wan1", ipaddr, mask) == SUCCESS);

	/* Lan: */
	mac.octet[5] = 0x03;
	ASSERT(rtl8651_addVlan(3) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(3, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("lan") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("lan", RTL8651_LL_VLAN, 3) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.2.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("lan", ipaddr, mask) == SUCCESS);
}

static int32 default_configuration5(uint32 type1, uint32 type2)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask;

	/* WAN: subnet 1: port 0, 6 */	
	ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("wan") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("wan", type1, 1) == SUCCESS);
	ipaddr = ntohl(inet_addr("140.113.1.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("wan", ipaddr, mask) == SUCCESS);
	ASSERT(rtl8651_addExtNetworkInterface("wan") == SUCCESS);

	/* VLAN: subnet 2: port 1, 2, 3, 4, 5 */
	mac.octet[5] = 0x02;
	ASSERT(rtl8651_addVlan(2) == SUCCESS);
	ASSERT(rtl8651_delVlanPortMember(1, 1) == SUCCESS);
	ASSERT(rtl8651_addVlanPortMember(2, 1) == SUCCESS);
	ASSERT(rtl8651_delVlanPortMember(1, 2) == SUCCESS);
	ASSERT(rtl8651_addVlanPortMember(2, 2) == SUCCESS);
	ASSERT(rtl8651_delVlanPortMember(1, 3) == SUCCESS);
	ASSERT(rtl8651_addVlanPortMember(2, 3) == SUCCESS);
	ASSERT(rtl8651_delVlanPortMember(1, 4) == SUCCESS);
	ASSERT(rtl8651_addVlanPortMember(2, 4) == SUCCESS);
	ASSERT(rtl8651_delVlanPortMember(1, 5) == SUCCESS);
	ASSERT(rtl8651_addVlanPortMember(2, 5) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("vlan") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan", type2, 2) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.1.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("vlan", ipaddr, mask) == SUCCESS);
	return SUCCESS;
}


struct rtl_mBuf *fakeTcpPkt(int8* data, uint16 len, int32 isPPPoE)
{
	rtl8651_PktConf_t	conf;

	memset(&conf, 0, sizeof(rtl8651_PktConf_t));

	conf.pktType						= _PKT_TYPE_TCP;					/* tcp packet */
	conf.conf_Flags					|= (CSUM_IP_OK|CSUM_L4_OK);

	if (isPPPoE == TRUE)
		conf.l2Flag |= L2_PPPoE;

//	CP_MAC (mac_npi, conf.conf_smac);
//	memset(conf.);
//	CP_MAC (mac_gw_int, conf.conf_dmac);
	conf.conf_sip						= ntohl(inet_addr("192.168.2.2"));;
	conf.conf_dip						= ntohl(inet_addr("149.113.1.1"));;
	conf.ip.id							= 0xa;
	conf.ip.mf						= 0;
	conf.ip.offset						= 0;
	conf.ip.ttl							= 255;
//	conf.conf_sport					= port_npi;
//	conf.conf_dport					= port_rp;
	conf.conf_tcp_seq					= 0xb;
	conf.conf_tcp_ack					= 0xc;
//	conf.conf_tcp_flag					= TH_SYN;
	conf.conf_tcp_win					= 0xfffe;

	conf.payload.content				= data;
	conf.payload.length					= len;

	return genMbufPkt(&conf);
}




/*===================================================================================
  * Table Driver testing codes
  *===================================================================================*/


#define RTL865X_URL_FILTER_TESTING_CODE



#define SAMPLE1 \
"\
GET /search?q=linux&ie=UTF-8&oe=UTF-8&hl=zh-TW&btnG=Google+%E6%90%9C%E5%B0%8B&lr= HTTP/1.1\r\n\
Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/vnd.ms-excel, application/vnd.ms-powerpoint, application/msword, application/x-shockwave-flash, */*\r\n\
Referer: http://www.google.com.tw/\r\n\
Accept-Language: zh-tw\r\n\
Accept-Encoding: gzip, deflate\r\n\
User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)\r\n\
Host: www.google.com.tw\r\n\
Connection: Keep-Alive\r\n\
Cookie: PREF=ID=12086cec7438ba07:LD=zh-TW:TM=1061463358:LM=1061463358:S=C8pv71LUKYQIY3_E\r\n\r\n\
"

#define SAMPLE2 \
"\
GET /imgvirus.jpg HTTP/1\r\n\
Accept: */*\r\n\
Referer: http:/www.nctu.edu.tw/\r\n\
Accept-Languge: zh-tw\r\n\
Accet-Encoding: gzi, deflate\r\n\
If-Mified-Since: Fri, 19 Sep 2003 09:50:08 GMT\r\n\
IfNone-Match: \"16e5008a-163c-a85af400\"\r\n\
User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT5.1)\r\n\
Host: www.nctu.edu.tw\r\n\
Connection: Keep-Alive\r\n\r\n\
"

#define SAMPLE3 \
"\
GET /img/txt2-1.jpg HTTP/1.1\r\n\
Accept: */*\r\n\
Referer: http://www.nctu.edu.tw/\r\n\
Accept-Language: zh-tw\r\n\
Accept-Encoding: gzip, deflate\r\n\
If-Modified-Since: Fri, 13 Apr 2001 14:21:41 GMT\r\n\
If-None-Match: \"16e50068-1dfc-d2920740\"\r\n\
User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)\r\n\
Host: www.nctu.edu.tw\r\n\
Connection: Keep-Alive\r\n\r\n\
"

#define SAMPLE4 	"GET http://www.w3.org/pub/WWW/TheProject.html HTTP/1.1\r\n\r\n"


#define SAMPLE_WWW_GOOGLE_COM \
"\
GET /search?q=linux&ie=UTF-8&oe=UTF-&btnG=Google+%E6%90%9C%E5%B0%8B&lr= HTTP/1.1\r\n\
Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/vnd.ms-excel, application/vnd.ms-powerpoint, application/msword, application/x-shockwave-flash, */*\r\n\
Referer: http://www.google.com/\r\n\
Accept-Language: zh-tw\r\n\
Accept-Encoding: gzip, deflate\r\n\
User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)\r\n\
Host: www.google.com\r\n\
Connection: Keep-Alive\r\n\
Cookie: PREF=ID=12086cec7438ba07:LD=zh-TW:TM=1061463358:LM=1061463358:S=C8pv71LUKYQIY3_E\r\n\r\n\
"


#define SAMPLE_WWW_GOOGLE_COM_TW \
"\
GET /search?q=linux&ie=UTF-8&oe=UTF-8&hl=zh-TW&btnG=Google+%E6%90%9C%E5%B0%8B&lr= HTTP/1.1\r\n\
Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/vnd.ms-excel, application/vnd.ms-powerpoint, application/msword, application/x-shockwave-flash, */*\r\n\
Referer: http://www.google.com.tw/\r\n\
Accept-Language: zh-tw\r\n\
Accept-Encoding: gzip, deflate\r\n\
User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)\r\n\
Host: www.google.com.tw\r\n\
Connection: Keep-Alive\r\n\
Cookie: PREF=ID=12086cec7438ba07:LD=zh-TW:TM=1061463358:LM=1061463358:S=C8pv71LUKYQIY3_E\r\n\r\n\
"

#define SAMPLE_WWW_GOOGLE_COM_DOT \
"\
GET /search?q=linux&ie=UTF-8&oe=UTF-&btnG=Google+%E6%90%9C%E5%B0%8B&lr= HTTP/1.1\r\n\
Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/vnd.ms-excel, application/vnd.ms-powerpoint, application/msword, application/x-shockwave-flash, */*\r\n\
Referer: http://www.google.com./\r\n\
Accept-Language: zh-tw\r\n\
Accept-Encoding: gzip, deflate\r\n\
User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)\r\n\
Host: www.google.com.\r\n\
Connection: Keep-Alive\r\n\
Cookie: PREF=ID=12086cec7438ba07:LD=zh-TW:TM=1061463358:LM=1061463358:S=C8pv71LUKYQIY3_E\r\n\r\n\
"

#define SAMPLE_ABC_WWW_GOOGLE_COM \
"\
GET /search?q=linux&ie=UTF-8&oe=UTF-&btnG=Google+%E6%90%9C%E5%B0%8B&lr= HTTP/1.1\r\n\
Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/vnd.ms-excel, application/vnd.ms-powerpoint, application/msword, application/x-shockwave-flash, */*\r\n\
Referer: http://abc.www.google.com/\r\n\
Accept-Language: zh-tw\r\n\
Accept-Encoding: gzip, deflate\r\n\
User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)\r\n\
Host: abc.www.google.com\r\n\
Connection: Keep-Alive\r\n\
Cookie: PREF=ID=12086cec7438ba07:LD=zh-TW:TM=1061463358:LM=1061463358:S=C8pv71LUKYQIY3_E\r\n\r\n\
"

#define SAMPLE_ABC_GOOGLE_COM \
"\
GET /search?q=linux&ie=UTF-8&oe=UTF-&btnG=Google+%E6%90%9C%E5%B0%8B&lr= HTTP/1.1\r\n\
Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/vnd.ms-excel, application/vnd.ms-powerpoint, application/msword, application/x-shockwave-flash, */*\r\n\
Referer: http://abc.google.com/\r\n\
Accept-Language: zh-tw\r\n\
Accept-Encoding: gzip, deflate\r\n\
User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)\r\n\
Host: abc.google.com\r\n\
Connection: Keep-Alive\r\n\
Cookie: PREF=ID=12086cec7438ba07:LD=zh-TW:TM=1061463358:LM=1061463358:S=C8pv71LUKYQIY3_E\r\n\r\n\
"



typedef struct urlFilteringPattern_s {
	uint32 sessionId;
	int8 	pattern[64];
	int8  sample[1024];
	int8	isMatch;
	int8	*dn;
} urlFilteringPattern_t;

#ifdef _RTL_NEW_LOGGING_MODEL
static uint32 _logNo_ = RTL8651_LOG_URL_MatchUrlFilter;
static int8 _msgMatch_;
static int8 _msgString_[128];
int32 __LOG_FUNC__
(
	unsigned long	moduleId,
	unsigned long	logNo,
	rtl8651_logInfo_t	*info
)
{
	if (_msgMatch_ == TRUE)
	{
		ASSERT(_logNo_ == logNo);
		ASSERT(strcmp(info->url_string, _msgString_) == 0);
	}
	return SUCCESS;
}
#else
static int8 _msgString_[128];
static int8 _msgMatch_;

int32 __LOG_FUNC__(uint32 sessionId, uint32 moduleId, uint8 proto, int8 direction, 
	ipaddr_t sip, ipaddr_t dip, 
	uint16 sport, uint16 dport,
	uint8 type, uint8 action, 
	int8 *msg)
{
	if (_msgMatch_ == TRUE) {
		ASSERT(strcmp(msg, _msgString_) == 0);
	}
	return SUCCESS;
}
#endif /* _RTL_NEW_LOGGING_MODEL */


/*
@func int32 |url_Filter_test1 | URL filter testing. 
@comm
Given a pre-defined http pattern, check if the URL filtering functionality is correct or not in a single PPPoE session environment.
Beside also check if logging mechanism is workable.
*/
int32 url_Filter_test1(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	struct rtl_mBuf *pkt_t;
	struct rtl_pktHdr *pktHdr_t;
	struct ip   *iphdr;
	ipaddr_t ipaddr, mask;
	int8 string[65];
	int32 len, entryIndex, isFound;
	urlFilteringPattern_t __urlFiltering[] = 
	{
		{ 1000, 	"google", 		SAMPLE1, 	TRUE,	"www.google.com.tw"	},
		{ 1000, 	"gle.com",	SAMPLE1, 	TRUE,	"www.google.com.tw" 	},
		{ 1000,	".",			SAMPLE1,		TRUE,	"www.google.com.tw"	},
		{ 1000, 	"t",			SAMPLE1, 	TRUE,	"www.google.com.tw"	},
		{ 1000, 	"w.g",		SAMPLE1,		TRUE,	"www.google.com.tw"	},
		{ 1000, 	"abcdef",		SAMPLE1, 	FALSE,	"www.google.com.tw"	},
		{ 1000,	"Accept",		SAMPLE1,		FALSE,	"www.google.com.tw"	},
		{ 1000, 	".com",		SAMPLE2,		FALSE,	"www.nctu.edu.tw"		},
		{ 1000, 	"w.w3",		SAMPLE4,		TRUE,	"www.w3.org"			},
		{ 0, "", "" },
	};
	

	START_TEST_CASE();
	{
		/*=====================================================
		  * Test:
		  *	1. Single Session environment
		  *	2. Install Logging function & turn on logging mechanism
		  *	3. Patterns are encapsulated as mbufs and feed to the url Filtering APIs
		  *=====================================================*/
		default_configuration3();

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

		ASSERT(rtl8651_installLoggingFunction(__LOG_FUNC__) == SUCCESS);
		ASSERT(rtl8651a_enableLogging(1000, 99999, TRUE) == SUCCESS);

		for(entryIndex=0; __urlFiltering[entryIndex].sessionId != 0; entryIndex++) {
			strcpy(_msgString_, __urlFiltering[entryIndex].dn);
			_msgMatch_ = __urlFiltering[entryIndex].isMatch;
			/* Add Url filter string */
			ASSERT(rtl8651a_addURLfilterString(1000, __urlFiltering[entryIndex].pattern , strlen(__urlFiltering[entryIndex].pattern)) == SUCCESS);
			/* Get Mbuf */
			pkt_t = fakeTcpPkt(__urlFiltering[entryIndex].sample, 1024, TRUE);
			pktHdr_t = pkt_t->m_pkthdr;
			iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
			len = 64;
			/* Pattern Matching...... */
			isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
			if (isFound != FAILED) {
				ASSERT(__urlFiltering[entryIndex].isMatch == TRUE);
				ASSERT(strncmp(string, __urlFiltering[entryIndex].pattern, len) == 0);
			} else {
				ASSERT(__urlFiltering[entryIndex].isMatch == FALSE);
			}
			/* Remove Url filter string */
			ASSERT(rtl8651a_delURLfilterString(1000, __urlFiltering[entryIndex].pattern , strlen(__urlFiltering[entryIndex].pattern)) == SUCCESS);
			freeMbuf(pkt_t);
		}
	}
	return SUCCESS;
}


/*
@func int32 |url_Filter_test2 | URL filter testing. 
@comm
Given a pre-defined http pattern, check if the URL filtering functionality is correct or not.
*/
int32 url_Filter_test2(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	struct rtl_mBuf *pkt_t;
	struct rtl_pktHdr *pktHdr_t;
	struct ip   *iphdr;
	ipaddr_t ipaddr, mask;
	int8 string[65];
	int32 len, isFound, round;
	urlFilteringPattern_t __urlFiltering[] = 
	{
		{ 1000,	".com",		SAMPLE3,		FALSE,	"www.nctu.edu.tw" },
		{ 1000, 	".com",		SAMPLE2,		FALSE,	"www.nctu.edu.tw" },
		{ 0, "", "" },
	};
	

	START_TEST_CASE();
	{
		/*=====================================================
		  * Test:
		  *	1. Single Session environment
		  *	2. Install Logging function & turn on logging mechanism
		  *	3. Patterns are encapsulated as mbufs and feed to the url Filtering APIs
		  *=====================================================*/
		default_configuration3();

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

		ASSERT(rtl8651_installLoggingFunction(__LOG_FUNC__) == SUCCESS);
		ASSERT(rtl8651a_enableLogging(1000, 99999, TRUE) == SUCCESS);

		for(round=0; __urlFiltering[round].sessionId ; round++) {
			strcpy(_msgString_, __urlFiltering[round].dn);
			_msgMatch_ = __urlFiltering[round].isMatch;
			/* Add Url filter string */
			ASSERT(rtl8651_addURLfilterString(__urlFiltering[round].pattern , strlen(__urlFiltering[round].pattern)) == SUCCESS);
			/* Get Mbuf */
			pkt_t = fakeTcpPkt(__urlFiltering[round].sample, 1024, TRUE);
			pktHdr_t = pkt_t->m_pkthdr;
			iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
			len = 64;
			/* Pattern Matching...... */
			isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
			if (isFound != FAILED) {
				ASSERT(__urlFiltering[round].isMatch == TRUE);
			}
			else {
				ASSERT(__urlFiltering[round].isMatch == FALSE);
			}
			freeMbuf(pkt_t);

			/* Remove Url filter string */
			ASSERT(rtl8651_delURLfilterString(__urlFiltering[round].pattern , strlen(__urlFiltering[round].pattern)) == SUCCESS);

		}
	}
	
	return SUCCESS;
}


/*
@func int32 |url_Filter_test3 | URL filter testing. 
@comm
Test url-filtering extension for source IP blocking.
*/

int32 url_Filter_test3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	struct rtl_mBuf *pkt_t;
	struct rtl_pktHdr *pktHdr_t;
	struct ip   *iphdr;
	ipaddr_t ipaddr, mask;
	int8 string[65];
	int32 len, isFound, round;
	int8 *sample = SAMPLE1;
	

	START_TEST_CASE();

	{
		/*=====================================================
		  * Test: src ip block for url-filtering
		  *=====================================================*/
		default_configuration3();

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

		ASSERT(rtl8651_installLoggingFunction(__LOG_FUNC__) == SUCCESS);
		ASSERT(rtl8651a_enableLogging(1000, 99999, TRUE) == SUCCESS);

		/* (1) wild-card source IP */
		ASSERT(rtl8651_addURLFilterRule(1000, "com",  3, 0, 0) == SUCCESS);
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		ASSERT(rtl8651_delURLFilterRule(1000, "com",  3, 0, 0) == SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* (2) non wild-card source IP, source IP is not blocking IP, not IP range */
		ipaddr = ntohl(inet_addr("192.168.2.10"));
		ASSERT(rtl8651_addURLFilterRule(1000, "com",  3, ipaddr, ipaddr) == SUCCESS);
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		ASSERT(rtl8651_delURLFilterRule(1000, "com",  3, ipaddr, ipaddr) == SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);


		/* (3) non wild-card source IP, source IP is not blocking IP, IP range */
		ipaddr = ntohl(inet_addr("192.168.2.10"));
		ASSERT(rtl8651_addURLFilterRule(1000, "com",  3, ipaddr, ipaddr+10) == SUCCESS);
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		ASSERT(rtl8651_delURLFilterRule(1000, "com",  3, ipaddr, ipaddr+10) == SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);
		

		/* (4) non wild-card source IP, source IP is blocking IP, not IP range */
		ipaddr = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addURLFilterRule(1000, "com",  3, ipaddr, ipaddr) == SUCCESS);
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		ASSERT(rtl8651_delURLFilterRule(1000, "com",  3, ipaddr, ipaddr) == SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);


		/* (5) non wild-card source IP, source IP is blocking IP, IP range */
		ipaddr = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addURLFilterRule(1000, "com",  3, ipaddr, ipaddr+10) == SUCCESS);
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		ASSERT(rtl8651_delURLFilterRule(1000, "com",  3, ipaddr, ipaddr+10) == SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);

	}
	return SUCCESS;
}

/*
@func int32 |url_Filter_test4 | URL filter testing. 
@comm
Given a pre-defined http pattern, check if the URL filtering functionality is correct or not in a single PPPoE session environment.
Beside also check if logging mechanism is workable.
This test case is for exactly match.
*/
int32 url_Filter_test4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	struct rtl_mBuf *pkt_t;
	struct rtl_pktHdr *pktHdr_t;
	struct ip   *iphdr;
	ipaddr_t ipaddr, mask;
	int8 string[65];
	int32 len, entryIndex, isFound;
	urlFilteringPattern_t __urlFiltering[] = 
	{
		{ 1000, 	"www.google.com", 			SAMPLE_WWW_GOOGLE_COM, 		TRUE,	"www.google.com"	},
		{ 1000, 	"www.google.com", 			SAMPLE_ABC_WWW_GOOGLE_COM, 	FALSE,	"abc.www.google.com"	},
		{ 1000, 	"www.google.com", 			SAMPLE_WWW_GOOGLE_COM, 		TRUE,	"www.google.com"	},
		{ 1000, 	"www.google.com",				SAMPLE_WWW_GOOGLE_COM_TW, 	FALSE,	"www.google.com.tw" 	},
		{ 1000,	"www.google.com",				SAMPLE_ABC_WWW_GOOGLE_COM,	FALSE,	"abc.www.google.com"	},
		{ 1000, 	"www.google.com",				SAMPLE_WWW_GOOGLE_COM_DOT, 	FALSE,	"www.google.com."	},
		{ 1000, 	"www.google.com.",			SAMPLE_WWW_GOOGLE_COM_DOT,	TRUE,	"www.google.com."	},
		{ 1000, 	"www.google.com.",			SAMPLE_WWW_GOOGLE_COM_TW, 	FALSE,	"www.google.com.tw" 	},
		{ 1000,	"www.google.com.",			SAMPLE_ABC_WWW_GOOGLE_COM,	FALSE,	"abc.www.google.com"	},
		{ 1000, 	"www.google.com.",			SAMPLE_WWW_GOOGLE_COM,		FALSE,	"www.google.com"	},
		{ 0, "", "" },
	};
	

	START_TEST_CASE();
	{
		/*=====================================================
		  * Test:
		  *	1. Single Session environment
		  *	2. Install Logging function & turn on logging mechanism
		  *	3. Patterns are encapsulated as mbufs and feed to the url Filtering APIs
		  *=====================================================*/
		default_configuration3();

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

		ASSERT(rtl8651_installLoggingFunction(__LOG_FUNC__) == SUCCESS);
		ASSERT(rtl8651a_enableLogging(1000, 99999, TRUE) == SUCCESS);

		for(entryIndex=0; __urlFiltering[entryIndex].sessionId != 0; entryIndex++) {
			strcpy(_msgString_, __urlFiltering[entryIndex].dn);
			_msgMatch_ = __urlFiltering[entryIndex].isMatch;
			/* Add Url filter string */
			ASSERT(rtl8651a_addURLExactlyfilterString(1000, __urlFiltering[entryIndex].pattern , strlen(__urlFiltering[entryIndex].pattern)) == SUCCESS);
			/* Get Mbuf */
			pkt_t = fakeTcpPkt(__urlFiltering[entryIndex].sample, 1024, TRUE);
			pktHdr_t = pkt_t->m_pkthdr;
			iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
			len = 64;
			/* Pattern Matching...... */
			isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
			if (isFound != FAILED) {
				ASSERT(__urlFiltering[entryIndex].isMatch == TRUE);
				ASSERT(strncmp(string, __urlFiltering[entryIndex].pattern, len) == 0);
			} else {
				ASSERT(__urlFiltering[entryIndex].isMatch == FALSE);
			}
			/* Remove Url filter string */
			ASSERT(rtl8651a_delURLExactlyfilterString(1000, __urlFiltering[entryIndex].pattern , strlen(__urlFiltering[entryIndex].pattern)) == SUCCESS);
			freeMbuf(pkt_t);
		}
	}
	return SUCCESS;
}

/*
@func int32 |url_Filter_test5 | URL filter testing. 
@comm
Given mulitple pre-defined http pattern, check if the URL filtering functionality is correct or not.
*/
/*
@func int32 |url_Filter_test3 | URL filter testing. 
@comm
Test url-filtering extension for source IP blocking.
*/

int32 url_Filter_test5(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	struct rtl_mBuf *pkt_t;
	struct rtl_pktHdr *pktHdr_t;
	struct ip   *iphdr;
	ipaddr_t ipaddr, ipaddr2, mask;
	int8 string[65];
	int32 len, isFound, round;
	int8 *sample_wwwgooglecom = SAMPLE_WWW_GOOGLE_COM;
	int8 *sample_abcgooglecom = SAMPLE_ABC_GOOGLE_COM;
	int8 *sample_abcwwwgooglecom = SAMPLE_ABC_WWW_GOOGLE_COM;
	int8 *sample_wwwgooglecomtw = SAMPLE_WWW_GOOGLE_COM_TW;
	int8 *sample_wwwgooglecomdot = SAMPLE_WWW_GOOGLE_COM_DOT;
	

	START_TEST_CASE();

	{
		/*=====================================================
		  * Test: src ip block for url-filtering
		  *=====================================================*/
		default_configuration3();

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

		ASSERT(rtl8651_installLoggingFunction(__LOG_FUNC__) == SUCCESS);
		ASSERT(rtl8651a_enableLogging(1000, 99999, TRUE) == SUCCESS);

		/* (1) wild-card source IP, 2 rules */
		ASSERT(rtl8651_addURLExactlyFilterRule(1000, "www.google.com", strlen("www.google.com"), 0, 0) == SUCCESS);
		ASSERT(rtl8651_addURLFilterRule(1000, "www.google.com",  strlen("www.google.com"), 0, 0) == SUCCESS);

		/* match rule 1 */
		pkt_t = fakeTcpPkt(sample_wwwgooglecom, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* match rule 2 */
		pkt_t = fakeTcpPkt(sample_abcwwwgooglecom, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* match non of 2 rules */
		pkt_t = fakeTcpPkt(sample_abcgooglecom, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* remove 2 rules */
		ASSERT(rtl8651_delURLExactlyFilterRule(1000, "www.google.com",  strlen("www.google.com"), 0, 0) == SUCCESS);
		ASSERT(rtl8651_delURLFilterRule(1000, "www.google.com",  strlen( "www.google.com"), 0, 0) == SUCCESS);

		/* (2) wild-card source IP, 2 rules */
		ASSERT(rtl8651_addURLExactlyFilterRule(1000, "www.google.com",  strlen("www.google.com"), 0, 0) == SUCCESS);
		ASSERT(rtl8651_addURLFilterRule(1000, "google.com",  strlen("google.com"), 0, 0) == SUCCESS);

		/* match both of rules */
		pkt_t = fakeTcpPkt(sample_wwwgooglecom, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* remove one of 2 rules */
		ASSERT(rtl8651_delURLExactlyFilterRule(1000, "www.google.com",  strlen("www.google.com"), 0, 0) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample_wwwgooglecom, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* remove another rule */
		ASSERT(rtl8651_delURLFilterRule(1000, "google.com",  strlen("google.com"), 0, 0) == SUCCESS);
		ASSERT(rtl8651_addURLExactlyFilterRule(1000, "www.google.com",  strlen("www.google.com"), 0, 0) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample_wwwgooglecom, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* remove both of them */
		ASSERT(rtl8651_delURLExactlyFilterRule(1000, "www.google.com",  strlen("www.google.com"), 0, 0) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample_wwwgooglecom, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* (3) non wild-card source IP, not IP range, 2 rules and one is keyword match and one is exactly match */
		ipaddr = ntohl(inet_addr("192.168.2.10"));
		ipaddr2 = ntohl(inet_addr("192.168.2.11"));
		ASSERT(rtl8651_addURLExactlyFilterRule(1000, "www.google.com.",  strlen("www.google.com."), ipaddr, ipaddr) == SUCCESS);
		ASSERT(rtl8651_addURLFilterRule(1000, "www.google",  strlen("www.google"), ipaddr2, ipaddr2) == SUCCESS);

		/* sip is not in the IP range 1/2 */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* sip is in the IP range of rule 1 : match */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* sip is in the IP range of rule 2 : match */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr2;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* remove rule 1 */
		ASSERT(rtl8651_delURLExactlyFilterRule(1000, "www.google.com.",  strlen("www.google.com."), ipaddr, ipaddr) == SUCCESS);

		/* sip is in the IP range of rule 1 : pass */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* sip is in the IP range of rule 2 : match */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr2;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* remove rule 2 */
		ASSERT(rtl8651_delURLFilterRule(1000,"www.google", strlen("www.google"), ipaddr2, ipaddr2) == SUCCESS);

		/* sip is in the IP range of rule 1 : pass */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* sip is in the IP range of rule 2 : pass */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr2;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* (4) non wild-card source IP, IP range, 2 rules and one is keyword match and one is exactly match */
		ipaddr = ntohl(inet_addr("192.168.2.10"));
		ipaddr2 = ntohl(inet_addr("192.168.2.20"));
		ASSERT(rtl8651_addURLExactlyFilterRule(1000, "www.google.com.",  strlen("www.google.com."), ipaddr, ipaddr+9) == SUCCESS);
		ASSERT(rtl8651_addURLFilterRule(1000, "www.google",  strlen("www.google"), ipaddr2, ipaddr2+9) == SUCCESS);

		/* sip is not in the IP range 1/2 */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.30"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* sip is in the IP range of rule 1 : match */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr+5;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* sip is in the IP range of rule 2 : match */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr2+5;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* remove rule 1 */
		ASSERT(rtl8651_delURLExactlyFilterRule(1000, "www.google.com.",  strlen("www.google.com."), ipaddr, ipaddr+9) == SUCCESS);

		/* sip is in the IP range of rule 1 : pass */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr+5;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* sip is in the IP range of rule 2 : match */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr2+5;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* remove rule 2 */
		ASSERT(rtl8651_delURLFilterRule(1000,"www.google", strlen("www.google"), ipaddr2, ipaddr2+9) == SUCCESS);

		/* sip is in the IP range of rule 1 : pass */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr+5;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* sip is in the IP range of rule 2 : pass */
		pkt_t = fakeTcpPkt(sample_wwwgooglecomdot, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ipaddr2+5;
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		freeMbuf(pktHdr_t->ph_mbuf);
		
	}
	return SUCCESS;
}

/*
@func int32 |url_Filter_test6 | URL filter testing. 
@comm
Test url-filtering extension for source MAC blocking.
*/

int32 url_Filter_test6(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl8651_tblDrv_urlPktFilter_t pktFilter;
	rtl8651_tblDrv_urlFilter_t urlFilter;
	struct rtl_mBuf *pkt_t;
	struct rtl_pktHdr *pktHdr_t;
	struct ip   *iphdr;
	ipaddr_t ipaddr, mask;
	int8 string[65];
	int32 len, isFound, round;
	int8 *sample = SAMPLE1;
	

	START_TEST_CASE();

	{
		/*=====================================================
		  * Test: src ip block for url-filtering
		  *=====================================================*/
		default_configuration3();

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

		ASSERT(rtl8651_installLoggingFunction(__LOG_FUNC__) == SUCCESS);
		ASSERT(rtl8651a_enableLogging(1000, 99999, TRUE) == SUCCESS);

		/* (1) wild-card source MAC */
		pktFilter.rule_type = RTL8651_URLPKTFILTER_TYPE_SMAC;
		memcpy(&pktFilter.smac, &mac, 6);

		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.urlFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.urlFilterString, "com", 3);
		urlFilter.urlFilterStrLen = 3;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		memcpy(&(pktHdr_t->ph_mbuf->m_data[6]), &mac, 6);
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)!= SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* (2) non wild-card source MAC, source MAC is not blocking MAC */
		pktFilter.rule_type = RTL8651_URLPKTFILTER_TYPE_SMAC;
		memcpy(&pktFilter.smac, &mac, 6);

		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.urlFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.urlFilterString, "com", 3);
		urlFilter.urlFilterStrLen = 3;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, &pktFilter) == SUCCESS);
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		memcpy(&(pktHdr_t->ph_mbuf->m_data[6]), &mac, 6);
		pktHdr_t->ph_mbuf->m_data[6] ++;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, &pktFilter)== SUCCESS);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, &pktFilter)!= SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);


		/* (3) non wild-card source MAC, source MAC is blocking MAC */
		pktFilter.rule_type = RTL8651_URLPKTFILTER_TYPE_SMAC;
		memcpy(&pktFilter.smac, &mac, 6);

		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.urlFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.urlFilterString, "com", 3);
		urlFilter.urlFilterStrLen = 3;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, &pktFilter) == SUCCESS);
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		memcpy(&(pktHdr_t->ph_mbuf->m_data[6]), &mac, 6);
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, &pktFilter)== SUCCESS);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, &pktFilter)!= SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);

	}
	return SUCCESS;
}

/*
@func int32 |url_Filter_test7 | URL filter testing. 
@comm
Test url-filtering extension for source MAC blocking.
For Exactly match.
*/

int32 url_Filter_test7(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl8651_tblDrv_urlPktFilter_t pktFilter;
	rtl8651_tblDrv_urlFilter_t urlFilter;
	struct rtl_mBuf *pkt_t;
	struct rtl_pktHdr *pktHdr_t;
	struct ip   *iphdr;
	ipaddr_t ipaddr, mask;
	int8 string[65];
	int32 len, isFound, round;
	int8 *sample = SAMPLE1;
	

	START_TEST_CASE();

	{
		/*=====================================================
		  * Test: src ip block for url-filtering
		  *=====================================================*/
		default_configuration3();

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

		ASSERT(rtl8651_installLoggingFunction(__LOG_FUNC__) == SUCCESS);
		ASSERT(rtl8651a_enableLogging(1000, 99999, TRUE) == SUCCESS);

		/* (1) wild-card source MAC */
		pktFilter.rule_type = RTL8651_URLPKTFILTER_TYPE_SMAC;
		memcpy(&pktFilter.smac, &mac, 6);

		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.urlFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.urlFilterString, "www.google.com.tw", 17);
		urlFilter.urlFilterStrLen = 17;
		urlFilter.urlExactlyMatch = TRUE;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);
		
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		memcpy(&(pktHdr_t->ph_mbuf->m_data[6]), &mac, 6);
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)!= SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);

		/* (2) non wild-card source MAC, source MAC is not blocking MAC */
		pktFilter.rule_type = RTL8651_URLPKTFILTER_TYPE_SMAC;
		memcpy(&pktFilter.smac, &mac, 6);

		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.urlFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.urlFilterString, "www.google.com.tw", 17);
		urlFilter.urlFilterStrLen = 17;
		urlFilter.urlExactlyMatch = TRUE;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, &pktFilter) == SUCCESS);
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		memcpy(&(pktHdr_t->ph_mbuf->m_data[6]), &mac, 6);
		pktHdr_t->ph_mbuf->m_data[6] ++;
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound == FAILED);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, &pktFilter)== SUCCESS);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, &pktFilter)!= SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);


		/* (3) non wild-card source MAC, source MAC is blocking MAC */
		pktFilter.rule_type = RTL8651_URLPKTFILTER_TYPE_SMAC;
		memcpy(&pktFilter.smac, &mac, 6);

		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.urlFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.urlFilterString, "www.google.com.tw", 17);
		urlFilter.urlFilterStrLen = 17;
		urlFilter.urlExactlyMatch = TRUE;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, &pktFilter) == SUCCESS);
		/* Get Mbuf */
		pkt_t = fakeTcpPkt(sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		memcpy(&(pktHdr_t->ph_mbuf->m_data[6]), &mac, 6);
		iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		pktHdr_t->ph_sip = ntohl(inet_addr("192.168.2.2"));
		/* Pattern Matching...... */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
		ASSERT(isFound != FAILED);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, &pktFilter)== SUCCESS);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, &pktFilter)!= SUCCESS);
		freeMbuf(pktHdr_t->ph_mbuf);

	}
	return SUCCESS;
}

/*
@func int32 |url_Filter_test8 | URL filter testing. 
@comm
Test url-filtering extension for URI-path blocking.
*/
int32 url_Filter_test8(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	struct rtl_mBuf *pkt_t;
	struct rtl_pktHdr *pktHdr_t;
	struct ip   *iphdr;
	ipaddr_t ipaddr, mask;
	int8 string[65];
	int32 len, entryIndex, isFound;
	rtl8651_tblDrv_urlPktFilter_t pktFilter;
	rtl8651_tblDrv_urlFilter_t urlFilter;
	urlFilteringPattern_t __urlFiltering[] = 
	{
		{ 1000, 	"google", 		SAMPLE1, 	TRUE,	"www.google.com.tw"	},
		{ 1000, 	"gle.com",	SAMPLE1, 	TRUE,	"www.google.com.tw" 	},
		{ 1000,	".",			SAMPLE1,		TRUE,	"www.google.com.tw"	},
		{ 1000, 	"t",			SAMPLE1, 	TRUE,	"www.google.com.tw"	},
		{ 1000, 	"w.g",		SAMPLE1,		TRUE,	"www.google.com.tw"	},
		{ 1000, 	"abcdef",		SAMPLE1, 	FALSE,	"www.google.com.tw"	},
		{ 1000,	"Accept",		SAMPLE1,		FALSE,	"www.google.com.tw"	},
		{ 1000, 	".com",		SAMPLE2,		FALSE,	"www.nctu.edu.tw"		},
		{ 1000, 	"w.w3",		SAMPLE4,		TRUE,	"www.w3.org"			},
		{ 0, "", "" },
	};
	

	START_TEST_CASE();
	{
		/*=====================================================
		  * Test:
		  *	1. Single Session environment
		  *	2. Install Logging function & turn on logging mechanism
		  *	3. Patterns are encapsulated as mbufs and feed to the url Filtering APIs
		  *=====================================================*/
		default_configuration3();

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

		ASSERT(rtl8651_installLoggingFunction(__LOG_FUNC__) == SUCCESS);
		ASSERT(rtl8651a_enableLogging(1000, 99999, TRUE) == SUCCESS);

		/* (1) */
		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.pathFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.pathFilterString, "imgvirus", 8);
		urlFilter.pathFilterStrLen = 8;	
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(__urlFiltering[7].sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *) &pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		/* Pattern Matching */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, RTL8651_WAN2LAN, string, &len);
		ASSERT(isFound != FAILED);
		ASSERT(strncmp(string, "imgvirus", len) == 0);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		freeMbuf(pkt_t);

		/* (2) */
		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.pathFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.pathFilterString, "imgvirus", 8);
		urlFilter.pathFilterStrLen = 8;	
		urlFilter.pathExactlyMatch = TRUE;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(__urlFiltering[7].sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *) &pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		/* Pattern Matching */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, RTL8651_WAN2LAN, string, &len);
		ASSERT(isFound == FAILED);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		freeMbuf(pkt_t);

		/* (3) */
		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.pathFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.pathFilterString, "imgvirus.jpg", 12);
		urlFilter.pathFilterStrLen = 12;	
		urlFilter.pathExactlyMatch = TRUE;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(__urlFiltering[7].sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *) &pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		/* Pattern Matching */
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, RTL8651_WAN2LAN, string, &len);
		ASSERT(isFound != FAILED);
		ASSERT(strncmp(string, "imgvirus.jpg", len) == 0);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		freeMbuf(pkt_t);

#if 0
		for(entryIndex=0; __urlFiltering[entryIndex].sessionId != 0; entryIndex++) {
			strcpy(_msgString_, __urlFiltering[entryIndex].dn);
			_msgMatch_ = __urlFiltering[entryIndex].isMatch;
			/* Add Url filter string */
			ASSERT(rtl8651a_addURLfilterString(1000, __urlFiltering[entryIndex].pattern , strlen(__urlFiltering[entryIndex].pattern)) == SUCCESS);
			/* Get Mbuf */
			pkt_t = fakeTcpPkt(__urlFiltering[entryIndex].sample, 1024, TRUE);
			pktHdr_t = pkt_t->m_pkthdr;
			iphdr = (struct ip *)&pktHdr_t->ph_mbuf->m_data[22];
			len = 64;
			/* Pattern Matching...... */
			isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, 1, string, &len);
			if (isFound != FAILED) {
				ASSERT(__urlFiltering[entryIndex].isMatch == TRUE);
				ASSERT(strncmp(string, __urlFiltering[entryIndex].pattern, len) == 0);
			} else {
				ASSERT(__urlFiltering[entryIndex].isMatch == FALSE);
			}
			/* Remove Url filter string */
			ASSERT(rtl8651a_delURLfilterString(1000, __urlFiltering[entryIndex].pattern , strlen(__urlFiltering[entryIndex].pattern)) == SUCCESS);
			freeMbuf(pkt_t);
		}
#endif
	}
	return SUCCESS;
}

/* =============================================================== */
static uint8 _urlFilterCallBackFlag = FALSE;
static void urlFilterCallBack(uint32 sessionId, struct rtl_pktHdr *pkt, struct ip *iphdr, const char *urlFiltered, const char *pathFiltered)
{
	_urlFilterCallBackFlag = TRUE;
}

/*
@func int32 |url_Filter_test9 | URL filter testing. 
@comm
Test url-filtering extension for CallBack function usage.
*/
int32 url_Filter_test9(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	struct rtl_mBuf *pkt_t;
	struct rtl_pktHdr *pktHdr_t;
	struct ip   *iphdr;
	ipaddr_t ipaddr, mask;
	int8 string[65];
	int32 len, entryIndex, isFound;
	rtl8651_tblDrv_urlPktFilter_t pktFilter;
	rtl8651_tblDrv_urlFilter_t urlFilter;
	urlFilteringPattern_t __urlFiltering[] = 
	{
		{ 1000, 	"google", 		SAMPLE1, 	TRUE,	"www.google.com.tw"	},
		{ 1000, 	"gle.com",	SAMPLE1, 	TRUE,	"www.google.com.tw" 	},
		{ 1000,	".",			SAMPLE1,		TRUE,	"www.google.com.tw"	},
		{ 1000, 	"t",			SAMPLE1, 	TRUE,	"www.google.com.tw"	},
		{ 1000, 	"w.g",		SAMPLE1,		TRUE,	"www.google.com.tw"	},
		{ 1000, 	"abcdef",		SAMPLE1, 	FALSE,	"www.google.com.tw"	},
		{ 1000,	"Accept",		SAMPLE1,		FALSE,	"www.google.com.tw"	},
		{ 1000, 	".com",		SAMPLE2,		FALSE,	"www.nctu.edu.tw"		},
		{ 1000, 	"w.w3",		SAMPLE4,		TRUE,	"www.w3.org"			},
		{ 0, "", "" },
	};
	

	START_TEST_CASE();
	{
		default_configuration3();

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

		ASSERT(rtl8651_installLoggingFunction(__LOG_FUNC__) == SUCCESS);
		ASSERT(rtl8651a_enableLogging(1000, 99999, TRUE) == SUCCESS);

		/* (1.1 with callBack function) */
		ASSERT(rtl8651_registerURLFilterCallBackFunction(urlFilterCallBack) == SUCCESS);
		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.pathFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.pathFilterString, "imgvirus", 8);
		urlFilter.pathFilterStrLen = 8;	
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(__urlFiltering[7].sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *) &pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		/* Pattern Matching */
		_urlFilterCallBackFlag = FALSE;
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, RTL8651_WAN2LAN, string, &len);
		ASSERT(_urlFilterCallBackFlag == TRUE);
		ASSERT(isFound != FAILED);
		ASSERT(strncmp(string, "imgvirus", len) == 0);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		freeMbuf(pkt_t);
		_urlFilterCallBackFlag = FALSE;
		ASSERT(rtl8651_registerURLFilterCallBackFunction(NULL) == SUCCESS);

		/* (1.2 without callBack function) */
		ASSERT(rtl8651_registerURLFilterCallBackFunction(NULL) == SUCCESS);
		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.pathFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.pathFilterString, "imgvirus", 8);
		urlFilter.pathFilterStrLen = 8;	
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(__urlFiltering[7].sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *) &pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		/* Pattern Matching */
		_urlFilterCallBackFlag = FALSE;
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, RTL8651_WAN2LAN, string, &len);
		ASSERT(_urlFilterCallBackFlag == FALSE);
		ASSERT(isFound != FAILED);
		ASSERT(strncmp(string, "imgvirus", len) == 0);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		freeMbuf(pkt_t);
		_urlFilterCallBackFlag = FALSE;
		ASSERT(rtl8651_registerURLFilterCallBackFunction(NULL) == SUCCESS);

		/* (2.1 with callBack function) */
		ASSERT(rtl8651_registerURLFilterCallBackFunction(urlFilterCallBack) == SUCCESS);
		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.pathFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.pathFilterString, "imgvirus", 8);
		urlFilter.pathFilterStrLen = 8;	
		urlFilter.pathExactlyMatch = TRUE;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(__urlFiltering[7].sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *) &pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		/* Pattern Matching */
		_urlFilterCallBackFlag = FALSE;
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, RTL8651_WAN2LAN, string, &len);
		ASSERT(_urlFilterCallBackFlag == FALSE);
		ASSERT(isFound == FAILED);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		freeMbuf(pkt_t);
		_urlFilterCallBackFlag = FALSE;
		ASSERT(rtl8651_registerURLFilterCallBackFunction(NULL) == SUCCESS);

		/* (2.2 without callBack function) */
		ASSERT(rtl8651_registerURLFilterCallBackFunction(NULL) == SUCCESS);
		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.pathFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.pathFilterString, "imgvirus", 8);
		urlFilter.pathFilterStrLen = 8;	
		urlFilter.pathExactlyMatch = TRUE;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(__urlFiltering[7].sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *) &pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		/* Pattern Matching */
		_urlFilterCallBackFlag = FALSE;
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, RTL8651_WAN2LAN, string, &len);
		ASSERT(_urlFilterCallBackFlag == FALSE);
		ASSERT(isFound == FAILED);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		freeMbuf(pkt_t);
		_urlFilterCallBackFlag = FALSE;
		ASSERT(rtl8651_registerURLFilterCallBackFunction(NULL) == SUCCESS);


		/* (3.1 with callBack function) */
		ASSERT(rtl8651_registerURLFilterCallBackFunction(urlFilterCallBack) == SUCCESS);
		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.pathFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.pathFilterString, "imgvirus.jpg", 12);
		urlFilter.pathFilterStrLen = 12;	
		urlFilter.pathExactlyMatch = TRUE;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(__urlFiltering[7].sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *) &pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		/* Pattern Matching */
		_urlFilterCallBackFlag = FALSE;
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, RTL8651_WAN2LAN, string, &len);
		ASSERT(_urlFilterCallBackFlag == TRUE);
		ASSERT(isFound != FAILED);
		ASSERT(strncmp(string, "imgvirus.jpg", len) == 0);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		freeMbuf(pkt_t);
		_urlFilterCallBackFlag = FALSE;
		ASSERT(rtl8651_registerURLFilterCallBackFunction(NULL) == SUCCESS);

		/* (3.2 without callBack function) */
		ASSERT(rtl8651_registerURLFilterCallBackFunction(NULL) == SUCCESS);
		bzero(&urlFilter, sizeof(rtl8651_tblDrv_urlFilter_t));
		urlFilter.sessionId = 1000;
		urlFilter.pathFilterString = (int8 *)malloc(32);
		_strncpy(urlFilter.pathFilterString, "imgvirus.jpg", 12);
		urlFilter.pathFilterStrLen = 12;	
		urlFilter.pathExactlyMatch = TRUE;
		ASSERT(rtl8651_addGenericURLFilterRule(&urlFilter, NULL) == SUCCESS);

		/* Get Mbuf */
		pkt_t = fakeTcpPkt(__urlFiltering[7].sample, 1024, TRUE);
		pktHdr_t = pkt_t->m_pkthdr;
		iphdr = (struct ip *) &pktHdr_t->ph_mbuf->m_data[22];
		len = 64;
		/* Pattern Matching */
		_urlFilterCallBackFlag = FALSE;
		isFound = _rtl8651_urlFilter(_rtl8651_SessionIDToDSID(1000), pktHdr_t, iphdr, RTL8651_WAN2LAN, string, &len);
		ASSERT(_urlFilterCallBackFlag == FALSE);
		ASSERT(isFound != FAILED);
		ASSERT(strncmp(string, "imgvirus.jpg", len) == 0);
		ASSERT(rtl8651_delGenericURLFilterRule(&urlFilter, NULL)== SUCCESS);
		freeMbuf(pkt_t);
		_urlFilterCallBackFlag = FALSE;
		ASSERT(rtl8651_registerURLFilterCallBackFunction(NULL) == SUCCESS);

	}
	return SUCCESS;
}


#if 0 /* assumption change */
int32 url_Filter_test3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	ipaddr_t ipaddr, mask;
	uint32 port, pattern, patternM;
	int32 operation;
	int8 *pchar;
	

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: If there exists patterns, set "TTP/" to ASIC port pattern
		  *	      register; otherwise, remove it from ASIC port pattern reg.
		  *==============================================*/
		default_configuration3();

		/* (1) dial up session 1 */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

		/* (2) dial up session 2 */
		ASSERT(rtl8651_addPppoeSession(1001, 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1001) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1001, 22222, &mac, 0x01) == SUCCESS);

		/* Verify Port Pattern match: */
		for(port=0; port<RTL8651_PORT_NUMBER; port++) {
			ASSERT(rtl8651_getAsicPortPatternMatch(port, &pattern, &patternM, &operation) != SUCCESS);
		}

		/* (3) Add Pattern match */
		ASSERT(rtl8651a_addURLfilterString(1000 ,"abcdef", 6) == SUCCESS);
		
		/* Verify Port Pattern Match */
		for(port=0; port<RTL8651_PORT_NUMBER; port++) {
			ASSERT(rtl8651_getAsicPortPatternMatch(port, &pattern, &patternM, &operation) == SUCCESS);
			ASSERT(patternM == 0xffffffff);
			ASSERT(operation == 0x02);
			pchar = (uint8 *)&pattern;
			ASSERT(pchar[0] == 'T');
			ASSERT(pchar[1] == 'T');
			ASSERT(pchar[2] == 'P');
			ASSERT(pchar[3] == '/');
		}

		/* (4) remove the Pattern match */
		ASSERT(rtl8651a_delURLfilterString(1000 ,"abcdef", 6) == SUCCESS);
		
		/* Verify Port Pattern match: */
		for(port=0; port<RTL8651_PORT_NUMBER; port++) {
			ASSERT(rtl8651_getAsicPortPatternMatch(port, &pattern, &patternM, &operation)  == FAILED);
		}
	}
	return SUCCESS;
}

#endif







#define RTL865X_ROUTING_TESTING_CODE


#if 0
/* OLD MODEL */



/*
@func int32 | l3_test2 | Routing Table testing.
@comm
Test host route addition and deletion. If the route lacks arp entry, the route should become to toCPU to trap all packets match
this route to CPU to handle arp issue. Latter, if arp entry is found, the route should chenge to a normal host route.
*/
int32 l3_test4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, mask, nexthop;
	uint32 entry, repeat = 0;
	

	START_TEST_CASE();

	{
		/*=======================================
		  * Test: add/delete host route (VLAN type )
		  *	      If no arp is present, the host route is traped to
		  *	      CPU. Once the arp is learnt, the process is set
		  *	      to L2.
		  *	      repeat N times
		  *=======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

again:
		/* (1) add a host route */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) == SUCCESS);

		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.255"));
		asic_rt_1.process 	= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		
		/* (2) Add nexthop's arp entry */
		ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= ntohl(inet_addr("140.113.1.1"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.255"));
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= 51;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		
		/* (3) remove the nexthop's arp entry */
		ASSERT(rtl8651_delArp(nexthop) == SUCCESS);
		
		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.255"));
		asic_rt_1.process 	= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (4) add nexthop's arp entry again */
		ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= ntohl(inet_addr("140.113.1.1"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.255"));
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= 51;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (5) Remove the host Route */
		ASSERT(rtl8651_delRoute(ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

		/* Verify L3 Table entry */
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) != SUCCESS);
		
		if (++repeat < 10)
			goto again;
	}
	return SUCCESS;
}






#endif 




/*
@fun int32 | l3_test1 | Routing table Testing.
@comm
Test if an interface route is automatically added while a subnet is created. Contrarily, check if the interface is automatically
removed while a subnet is removed.
*/
int32 l3_test1(uint32 caseNo)
{
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, mask;
	uint32 entry;
	

	START_TEST_CASE();

	{
		/*========================================
		  * Test: Interface route add/delete
		  *	      If a ethernet type subnet is added, a interface route
		  *	      should be added automatically.
		  *========================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		/* Verify L3 Table for VLAN 1 */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("192.168.1.1")) & ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process 	= RT_ARP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.arpStart	= 0;
		asic_rt_1.arpEnd	= 31;
		ASSERT(rtl8651_getAsicRouting(0,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L3 Table for VLAN 2 */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("192.168.2.1")) & ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process 	= RT_ARP;
		asic_rt_1.vidx		= 2;
		asic_rt_1.arpStart	= 32;
		asic_rt_1.arpEnd	= 63;
		ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		for(entry=2; entry<RTL8651_ROUTINGTBL_SIZE-1; entry++) {
			/* Other L3 entries should be invalid */
			ASSERT(rtl8651_getAsicRouting(entry,  &asic_rt_2) != SUCCESS);
		}

		

		/* Remove Subnet */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_delIpIntf("vlan1", ipaddr, mask) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.2.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_delIpIntf("vlan2", ipaddr, mask) == SUCCESS);

		/* Verify L3 Table */
		for(entry=0; entry<RTL8651_ROUTINGTBL_SIZE-1; entry++) {
			ASSERT(rtl8651_getAsicRouting(entry,  &asic_rt_2) != SUCCESS);
		}
	}
	return SUCCESS;
}


/*
@func int32 | l3_test2 | Routing Table testing.
@comm
Test ASIC default route. If no default is present and PPPoE is not allocated, the default route entry in the ASIC should
be DROP (2005.7) => DROP has been modified to CPU.
*/

int32 l3_test2(uint32 caseNo)
{
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	

	START_TEST_CASE();

	{
		/*===================================
		  * Test: If no default route, the last one entry should
		  *	      be DROP
		  *===================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* Verify the Last one L3 entry */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;

		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process 	= RT_DROP;
		asic_rt_1.process 	= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
	}
	return SUCCESS;
}


/*
@func int32 | l3_test3 | Routing Table testing.
@comm
Test if a subnet created is due to PPPoE dial up, it is no needed to add an interface route automatically.
*/
int32 l3_test3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, mask;
	uint32 entry;
	

	START_TEST_CASE();

	{
		/*==========================================
		  * Test: If PPPoE session, it is no needed to add an interface
		  *	      route for it.
		  *==========================================*/
		default_configuration3();

		/* (1) dial up the PPPoE session */
		ipaddr = ntohl(inet_addr("140.131.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		
		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("192.168.2.1")) & ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process 	= RT_ARP;
		asic_rt_1.vidx		= 2;
		asic_rt_1.arpStart	= 0;
		asic_rt_1.arpEnd	= 31;
		ASSERT(rtl8651_getAsicRouting(0,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		for(entry=1; entry<RTL8651_ROUTINGTBL_SIZE-1; entry++) {
			/* Others should be invalid */
			ASSERT(rtl8651_getAsicRouting(entry,  &asic_rt_2) != SUCCESS);
		}

		/* (2) Reset the PPPoE session */
		ipaddr = ntohl(inet_addr("140.131.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		ASSERT(rtl8651_delIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("192.168.2.1")) & ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process 	= RT_ARP;
		asic_rt_1.vidx		= 2;
		asic_rt_1.arpStart	= 0;
		asic_rt_1.arpEnd	= 31;
		ASSERT(rtl8651_getAsicRouting(0,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		for(entry=1; entry<RTL8651_ROUTINGTBL_SIZE-1; entry++) {
			/* Others should be invalid */
			ASSERT(rtl8651_getAsicRouting(entry,  &asic_rt_2) != SUCCESS);
		}
	}
	return SUCCESS;
}


/*
@func int32 | l3_test2 | Routing Table testing.
@comm
Test host route addition and deletion. If the route lacks arp entry, the route should become to toCPU to trap all packets match
this route to CPU to handle arp issue. Latter, if arp entry is found, the route should chenge to a normal host route.
*/
int32 l3_test4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	ipaddr_t ipaddr, mask, nexthop;
	uint32 entry, repeat = 0;
	

	START_TEST_CASE();

	{
		/*=======================================
		  * Test: add/delete host route (VLAN type )
		  *	      If no arp is present, the host route is traped to
		  *	      CPU. Once the arp is learnt, the process is set
		  *	      to L2.
		  *	      repeat N times
		  *=======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

again:
		/* (1) add a host route */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) == SUCCESS);

		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.255"));
#if 0 //Old model
		asic_rt_1.process 	= RT_CPU;
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		
		/* (2) Add nexthop's arp entry */
		ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= ntohl(inet_addr("140.113.1.1"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.255"));
#if 0 //Old model
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= 51;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= rtl8651_filterDbIndex(&mac); 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif


		/* (3) remove the nexthop's arp entry */
		ASSERT(rtl8651_delArp(nexthop) == SUCCESS);
		
		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.255"));
#if 0 //Old model
		asic_rt_1.process 	= RT_CPU;
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);


		/* (4) add nexthop's arp entry again */
		ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= ntohl(inet_addr("140.113.1.1"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.255"));
#if 0 //Old model
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= 51;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= rtl8651_filterDbIndex(&mac); 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif

		/* (5) Remove the host Route */
		ASSERT(rtl8651_delRoute(ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

		/* Verify L3 Table entry */
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) != SUCCESS);
		
		if (++repeat < 10)
			goto again;
	}
	return SUCCESS;
}



/*
@func int32 | l3_test5 | Routing Table testing.
@comm
Test network route addition and deletion. If lack the network route's nexthop ARP, the entry should be set toCPU to trap
all packets which match this route.
*/
int32 l3_test5(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	ipaddr_t ipaddr, mask, nexthop;
	uint32  repeat = 0;
	
	
	START_TEST_CASE();

	{
		/*=====================================
		  * Test: Network route add/delete (VLAN type )
		  *	      repeat N times
		  *=====================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

again:
		/* (1) add a host route */
		ipaddr = ntohl(inet_addr("140.113.1.0"));
		mask = ntohl(inet_addr("255.255.255.0"));
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) == SUCCESS);

		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.1.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
#if 0 //Old model
		asic_rt_1.process 	= RT_CPU;
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (2) Add nexthop's arp entry */
		ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= ntohl(inet_addr("140.113.1.0"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.0"));
#if 0 //Old model
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= 51;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else

		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= rtl8651_filterDbIndex(&mac); 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif


		/* (3) remove the nexthop's arp entry */
		ASSERT(rtl8651_delArp(nexthop) == SUCCESS);
		
		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.1.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
#if 0 //Old model
		asic_rt_1.process 	= RT_CPU;
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (4) add nexthop's arp entry again */
		ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

		/* Verify L3 Table for the host route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= ntohl(inet_addr("140.113.1.0"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.0"));
#if 0 //Old model
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= 51;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= rtl8651_filterDbIndex(&mac); 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif
		/* (5) Remove the host Route */
		ASSERT(rtl8651_delRoute(ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

		/* Verify L3 Table entry */
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) != SUCCESS);
		
		if (++repeat < 10)
			goto again;

	}
	return SUCCESS;
}



/*
@func int32 | l3_test6 | Routing Table testing.
@comm
Test default route addition and deletion and the interaction with nexthop's arp.
*/
int32 l3_test6(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	ipaddr_t nexthop;
	uint32 repeat ;
	

	START_TEST_CASE();

	{
		/*============================================
		  * Test: add/delete default route (VLAN type)
		  *	      Repeat N times
		  *============================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		for(repeat=0; repeat<100; repeat ++){
			/* (1) Add default route */
			nexthop = ntohl(inet_addr("192.168.1.254"));
			ASSERT(rtl8651_addRoute(0, 0, "vlan1", nexthop) == SUCCESS);

			/* Verify Default Route */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
#if 0 //Old model
			asic_rt_1.process 	= RT_CPU;
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
			
#endif
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);


			/* (2) add nexthop's arp entry */
			ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

			/* Verify Default route entry */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
#if 0 //Old model
			asic_rt_1.process 			= RT_L2;
			asic_rt_1.vidx				= 1;
			asic_rt_1.nextHopRow		= 51;
			asic_rt_1.nextHopColumn	= 0;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1 ,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 51; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

			/* (3) remove nexthop's arp entry */
			ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

			/* Verify Default Route */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
#if 0 //Old model
			asic_rt_1.process 	= RT_CPU;
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* (4) add nexthop's arp entry again! */
			ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

#if 0 //Old model
			/* Verify Default route entry */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
			asic_rt_1.process 			= RT_L2;
			asic_rt_1.vidx				= 1;
			asic_rt_1.nextHopRow		= 51;
			asic_rt_1.nextHopColumn	= 0;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1 ,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 51; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif
			/* (5) remove default route */
			ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);
			ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

			/* Verify Default Route entry */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
			/* 2005.7 => DROP has been modified to CPU. */
			//asic_rt_1.process 			= RT_DROP;
			asic_rt_1.process 			= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		}
	}
	return SUCCESS;
}



/*
@func int32 | l3_test7 | Routing Table testing.
@comm
Test routing entry addition and deletion in PPPoE environment. The route should be set to toCPU if PPPoE doesn't dial up,
otherwise the entry should be set to PPPoE session.
*/
int32 l3_test7(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	ipaddr_t ipaddr, mask;
	uint32 repeat ;
	

	START_TEST_CASE();

	{
		/*============================================
		  * Test: add/delete route to PPPoE session
		  *	      Repeat N times
		  *============================================*/
		/* for testing */
		ASSERT(rtl8651_addVlan(5) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(5, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("testing") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("testing", RTL8651_LL_PPPOE, 5) == SUCCESS);
		ASSERT(rtl8651_addPppoeSession(1001, 5) == SUCCESS);
		
		default_configuration3();

		for(repeat=0; repeat<100; repeat++) {
			/* (1) dial up one PPPoE session */
			ipaddr = ntohl(inet_addr("140.113.1.1"));
			mask = ntohl(inet_addr("255.255.255.255"));
			ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
			ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
			ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

			/* (2) add default route to the PPPoE session */
			ASSERT(rtl8651_addRoute(0, 0, "pppoe", 0) == SUCCESS);

			/* Verify L3 Table for default route */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
#if 0 //Old model
			asic_rt_1.process 			= RT_PPPOE;
			asic_rt_1.vidx				= 1;
			asic_rt_1.pppoeIdx			= 1;
			asic_rt_1.nextHopRow		= 51;
			asic_rt_1.nextHopColumn	= 0;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 51; 
			asic_nh2.pppoeIdx			= 1;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif


			/* (2) reset the PPPoE session */
			ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);
			ASSERT(rtl8651_delIpIntf("pppoe", ipaddr, mask) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
#if 0 //Old model
			asic_rt_1.process 			= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 1;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

			/* (3) remove the default route */
			ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
			/* 2005.7 => DROP has been modified to CPU. */
			//asic_rt_1.process 			= RT_DROP;
			asic_rt_1.process 			= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
			

			/* (4) add the default route again! */
			ASSERT(rtl8651_addRoute(0, 0, "pppoe", 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
#if 0 //Old model
			asic_rt_1.process 			= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 1;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

			/* (7) dial up the PPPoE session */
			ipaddr = ntohl(inet_addr("140.113.1.1"));
			mask = ntohl(inet_addr("255.255.255.255"));
			ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
			ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
			ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

			/* Verify L3 Table for default route */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
#if 0 //Old model
			asic_rt_1.process 			= RT_PPPOE;
			asic_rt_1.vidx				= 1;
			asic_rt_1.pppoeIdx			= 1;
			asic_rt_1.nextHopRow		= 51;
			asic_rt_1.nextHopColumn	= 0;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 51; 
			asic_nh2.pppoeIdx			= 1;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

			/* (8) remove the default route */
			ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
			/* 2005.7 => DROP has been modified to CPU. */
			//asic_rt_1.process 			= RT_DROP;
			asic_rt_1.process			= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* (9) reset the PPPoE Session */
			ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);
			ASSERT(rtl8651_delIpIntf("pppoe", ipaddr, mask) == SUCCESS);
		}
	}
	return SUCCESS;
}


/*
@func int32 | l3_test8 | Routing Table testing.
@comm
Test host route addition and deletion in PPPoE environment.
*/
int32 l3_test8(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	ipaddr_t ipaddr, mask;
	uint32 repeat ;
	

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: add/delete host route (Single PPPoE)
		  *	      Repeat N times
		  *==============================================*/
		/* for testing */
		ASSERT(rtl8651_addVlan(5) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(5, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("testing") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("testing", RTL8651_LL_PPPOE, 5) == SUCCESS);
		ASSERT(rtl8651_addPppoeSession(1001, 5) == SUCCESS);
		
		default_configuration3();

		for(repeat=0; repeat<100; repeat++) {
			/* (1) dial up session 1 */
			ipaddr = ntohl(inet_addr("140.113.1.1"));
			mask = ntohl(inet_addr("255.255.255.255"));
			ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
			ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
			ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

			/* (2) add host route */
			ipaddr = ntohl(inet_addr("140.112.1.1"));
			ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "pppoe", 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= ipaddr;
			asic_rt_1.ipMask			= 0xffffffff;
#if 0 //Old model
			asic_rt_1.process 			= RT_PPPOE;
			asic_rt_1.vidx				= 1;
			asic_rt_1.pppoeIdx			= 1;
			asic_rt_1.nextHopRow		= 51;
			asic_rt_1.nextHopColumn	= 0;
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 51; 
			asic_nh2.pppoeIdx			= 1;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
			/* (3) reset the PPPoE session */
			ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);
			ASSERT(rtl8651_delIpIntf("pppoe", ntohl(inet_addr("140.113.1.1")), 0xffffffff) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= ipaddr;
			asic_rt_1.ipMask			= 0xffffffff;
#if 0 //Old model
			asic_rt_1.process 			= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 1;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

			/* (4) remove the host route */
			ipaddr = ntohl(inet_addr("140.112.1.1"));
			ASSERT(rtl8651_delRoute(ipaddr, 0xffffffff) == SUCCESS);

			/* Verify L3 Table */
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) != SUCCESS);

			/* (5) dial up the PPPoE Session */
			ipaddr = ntohl(inet_addr("140.113.1.1"));
			mask = ntohl(inet_addr("255.255.255.255"));
			ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
			ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
			ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

			/* (6) add Host route */
			ipaddr = ntohl(inet_addr("140.112.1.1"));
			ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "pppoe", 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= ipaddr;
			asic_rt_1.ipMask			= 0xffffffff;
#if 0 //Old model
			asic_rt_1.process 			= RT_PPPOE;
			asic_rt_1.vidx				= 1;
			asic_rt_1.pppoeIdx			= 1;
			asic_rt_1.nextHopRow		= 51;
			asic_rt_1.nextHopColumn	= 0;
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 51; 
			asic_nh2.pppoeIdx			= 1;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif

			/* (7) remove Host Route */
			ASSERT(rtl8651_delRoute(ipaddr, 0xffffffff) == SUCCESS);
			
			/* Verify L3 Table */
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) != SUCCESS);

			/* (8) reset PPPoE session */
			ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);
			ASSERT(rtl8651_delIpIntf("pppoe", ntohl(inet_addr("140.113.1.1")), 0xffffffff) == SUCCESS);
		}
	}
	return SUCCESS;
}


/*
@func int32 | l3_test9 | Routing Table testing.
@comm
Test netwrok route addition and deletion in PPPoe environment.
*/
int32 l3_test9(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	ipaddr_t ipaddr, mask;
	uint32 repeat ;
	

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: add/delete host route (Single PPPoE)
		  *	      Repeat N times
		  *==============================================*/
		/* for testing */
		ASSERT(rtl8651_addVlan(5) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(5, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("testing") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("testing", RTL8651_LL_PPPOE, 5) == SUCCESS);
		ASSERT(rtl8651_addPppoeSession(1001, 5) == SUCCESS);
		
		default_configuration3();

		for(repeat=0; repeat<100; repeat++) {
			/* (1) dial up session 1 */
			ipaddr = ntohl(inet_addr("140.113.1.1"));
			mask = ntohl(inet_addr("255.255.255.255"));
			ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
			ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
			ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

			/* (2) add network route */
			ipaddr = ntohl(inet_addr("140.112.1.0"));
			mask = ntohl(inet_addr("255.255.255.0"));
			ASSERT(rtl8651_addRoute(ipaddr, mask, "pppoe", 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= ipaddr;
			asic_rt_1.ipMask			= 0xffffff00;
#if 0 //Old model
			asic_rt_1.process 			= RT_PPPOE;
			asic_rt_1.vidx				= 1;
			asic_rt_1.pppoeIdx			= 1;
			asic_rt_1.nextHopRow		= 51;
			asic_rt_1.nextHopColumn	= 0;
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 51; 
			asic_nh2.pppoeIdx			= 1;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif

			/* (3) reset the PPPoE session */
			ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);
			ASSERT(rtl8651_delIpIntf("pppoe", ntohl(inet_addr("140.113.1.1")), 0xffffffff) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= ipaddr;
			asic_rt_1.ipMask			= 0xffffff00;
#if 0 //Old model
			asic_rt_1.process 			= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 1;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif

			/* (4) remove the network route */
			ipaddr = ntohl(inet_addr("140.112.1.0"));
			mask = ntohl(inet_addr("255.255.255.0"));
			ASSERT(rtl8651_delRoute(ipaddr, mask) == SUCCESS);

			/* Verify L3 Table */
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) != SUCCESS);

			/* (5) dial up the PPPoE Session */
			ipaddr = ntohl(inet_addr("140.113.1.1"));
			mask = ntohl(inet_addr("255.255.255.255"));
			ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
			ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
			ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);

			/* (6) add Host route */
			ipaddr = ntohl(inet_addr("140.112.1.0"));
			mask = ntohl(inet_addr("255.255.255.0"));
			ASSERT(rtl8651_addRoute(ipaddr, mask, "pppoe", 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= ipaddr;
			asic_rt_1.ipMask			= 0xffffff00;
#if 0 //Old model
			asic_rt_1.process 			= RT_PPPOE;
			asic_rt_1.vidx				= 1;
			asic_rt_1.pppoeIdx			= 1;
			asic_rt_1.nextHopRow		= 51;
			asic_rt_1.nextHopColumn	= 0;
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 51; 
			asic_nh2.pppoeIdx			= 1;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

			/* (7) remove Host Route */
			ASSERT(rtl8651_delRoute(ipaddr, mask) == SUCCESS);
			
			/* Verify L3 Table */
			ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) != SUCCESS);

			/* (8) reset PPPoE session */
			ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);
			ASSERT(rtl8651_delIpIntf("pppoe", ntohl(inet_addr("140.113.1.1")), 0xffffffff) == SUCCESS);
		}
	}
	return SUCCESS;

}


/*
@func int32 | l3_test10 | Routing Table testing.
@comm
Test if only 7 interface routes are allowed. 
*/

int32 l3_test10(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, mask;
	uint32 repeat ;
	

	START_TEST_CASE();

	{
		/*=========================================
		  * Test: Interface route can not exceed 7 entries
		  *=========================================*/
		/* VLAN 1: subnet 1 */	
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan1") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan1", RTL8651_LL_VLAN, 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.252"));
		ASSERT(rtl8651_addIpIntf("vlan1", ipaddr, mask) == SUCCESS);

		/* VLAN 2: subnet 2 */
		mac.octet[5] = 0x02;
		ASSERT(rtl8651_addVlan(2) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan2") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan2", RTL8651_LL_VLAN, 2) == SUCCESS);

		/* Create IP Interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.252"));
		for(repeat=1; repeat<RTL8651_ROUTINGTBL_SIZE; repeat++) {
			/* (1) add Ip Interface */
			ipaddr += (repeat<<8);
			if (repeat == RTL8651_ROUTINGTBL_SIZE-1) {
				ASSERT(rtl8651_addIpIntf("vlan2", ipaddr, mask) != SUCCESS);
				break;
			} else {
				ASSERT(rtl8651_addIpIntf("vlan2", ipaddr, mask) == SUCCESS);
			}
			
			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= ipaddr&mask;
			asic_rt_1.ipMask			= mask;
			asic_rt_1.process 			= RT_ARP;
			asic_rt_1.vidx				= 2;
			asic_rt_1.arpStart			= repeat;
			asic_rt_1.arpEnd			= repeat;
			ASSERT(rtl8651_getAsicRouting(repeat,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		}


		/* Remove Ip interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.252"));
		ASSERT(rtl8651_delIpIntf("vlan1", ipaddr, mask) == SUCCESS);
		/* Verify L3 Table */
		ASSERT(rtl8651_getAsicRouting(0,  &asic_rt_2) != SUCCESS);

		for(repeat=1; repeat<RTL8651_ROUTINGTBL_SIZE-1; repeat++) {
			ipaddr += (repeat<<8);
			ASSERT(rtl8651_delIpIntf("vlan2", ipaddr, mask) == SUCCESS);

			/* Verify L3 Table */
			ASSERT(rtl8651_getAsicRouting(repeat,  &asic_rt_2) != SUCCESS);
		}
	}
	return SUCCESS;
}

/*
@func int32 | l3_test11 | Routing Table testing.
@comm
Overflow ASIC routing table and remove it.
*/

int32 l3_test11(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	ipaddr_t ipaddr, mask, nexthop;
	uint32 repeat, rtSize ;
	

	START_TEST_CASE();
	rtSize = rtl8651_tblDrvPara.routingTableSize;
	rtl8651_tblDrvPara.routingTableSize = 8;
	{
		/*==========================================
		  * Test: Routes not include default route & interface route 
		  *	      can not exceed 7 entries
		  *==========================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add nexthop's arp entry */
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);
		
		/* add Route */
		ipaddr = ntohl(inet_addr("140.112.1.0"));
		mask = ntohl(inet_addr("255.255.255.0"));
		for(repeat=2; repeat<RTL8651_ROUTINGTBL_SIZE; repeat++) {
			ipaddr += (repeat << 8);
			if (repeat == RTL8651_ROUTINGTBL_SIZE-1) {
				ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) != SUCCESS);
				break;
			} else {
				ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) == SUCCESS);
			}
			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= ipaddr;
			asic_rt_1.ipMask			= mask;
#if 0 //Old model
			asic_rt_1.process 			= RT_L2;
			asic_rt_1.vidx				= 1;
			asic_rt_1.nextHopRow		= 51;
			asic_rt_1.nextHopColumn	= 0;
			ASSERT(rtl8651_getAsicRouting(repeat,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2*(repeat-1);
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2*(repeat-1);
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(repeat,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 51; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
		}

		/* Remove Route */
		ipaddr = ntohl(inet_addr("140.112.1.0"));
		mask = ntohl(inet_addr("255.255.255.0"));
		for(repeat=2; repeat<RTL8651_ROUTINGTBL_SIZE-1; repeat++) {
			/* Remove route */
			ipaddr += (repeat << 8);
			ASSERT(rtl8651_delRoute(ipaddr, mask) == SUCCESS);
			ASSERT(rtl8651_getAsicRouting(repeat,  &asic_rt_2) != SUCCESS);
			/* Verify L3 Table */
		}		
	}
	rtl8651_tblDrvPara.routingTableSize = rtSize;
	return SUCCESS;
}



/*
@func int32 | l3_test12 | Routing Table testing.
@comm
Test route addition and deletion in Multiple-Session environment. If no PPPoE session is specified, the rome driver will
select the first added session as nexthop, otherwise use user specified session as nexthop.
*/

int32 l3_test12(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	ipaddr_t ipaddr, mask;
	uint32 repeat, cn ;
	

	START_TEST_CASE();

	{
		/*===============================================
		  * Test: add default route to PPPoE session 
		  *	      (Multiple-Session PPPoE environment)
		  *	      (Old PPPoE model)
		  *
		  *	      Creat 4 PPPoE session and attach them to "pppoe" network
		  *	      interface. Add route if no specify which session to use, the
		  *          table Driver will always select the first session as the route's
		  *	      nexthop; otherwise, use user specifies session as route's
		  *	      nexthop.
		  *===============================================*/
		default_configuration3();

		ASSERT(rtl8651_delPppoeSession(1000) == SUCCESS);
		ASSERT(rtl8651_addVlan(5) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(5, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("test") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("test", RTL8651_LL_PPPOE, 5) == SUCCESS);
		ASSERT(rtl8651_addPppoeSession(1000, 5) == SUCCESS);

		/* (1) Create 4 PPPoE sessions */
		ASSERT(rtl8651_addPppoeSession(1001, 1) == SUCCESS);
		ASSERT(rtl8651_addPppoeSession(1002, 1) == SUCCESS);
		ASSERT(rtl8651_addPppoeSession(1003, 1) == SUCCESS);
		ASSERT(rtl8651_addPppoeSession(1004, 1) == SUCCESS);

		/* (2) dial up all PPPoE session */
		for(repeat=0; repeat<4; repeat++) {
			mac.octet[5] ++;
			ipaddr = ntohl(inet_addr("140.113.1.1")) + (repeat<<8);
			mask = ntohl(inet_addr("255.255.255.255"));
			ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, mask) == SUCCESS);
			ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000+(repeat+1)) == SUCCESS);
			ASSERT(rtl8651_setPppoeSessionProperty(1000+(repeat+1), 11111*(repeat+1), &mac, 0x01) == SUCCESS);
		}

		/* (3) add default route: no session is specified */
		ASSERT(rtl8651_addRoute(0, 0, "pppoe", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
#if 0 //Old model
		asic_rt_1.process 			= RT_PPPOE;
		asic_rt_1.vidx				= 1;
		asic_rt_1.pppoeIdx			= 1; /* use the firstadding one */
		asic_rt_1.nextHopRow		= 60;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= TRUE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 60; 
		asic_nh2.pppoeIdx			= 1;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

		/* (4) remove the default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process 			= RT_DROP;
		asic_rt_1.process			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		for(cn=0; cn<4; cn++) {
			/* (5) add default route: specify which session to use */
			ASSERT(rtl8651_addRoute(0, 0, "pppoe", 1001+cn) == SUCCESS);
			
			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
#if 0 //Old model
			asic_rt_1.process 			= RT_PPPOE;
			asic_rt_1.vidx				= 1;
			asic_rt_1.pppoeIdx			= 1+cn; /* use the last adding one */
			asic_rt_1.nextHopRow		= 60+cn;
			asic_rt_1.nextHopColumn	= 0;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= TRUE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 60+cn; 
			asic_nh2.pppoeIdx			= 1+cn;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif

			/* Remove default route */
			ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr			= 0;
			asic_rt_1.ipMask			= 0;
			/* 2005.7 => DROP has been modified to CPU. */
			//asic_rt_1.process 			= RT_DROP;
			asic_rt_1.process			= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		}
	}
	return SUCCESS;
}



/*
@func int32 | l3_test13 |Routing Table testing.
@comm
Test add a default route to a class B network and add/delete the route's nexthop arp.
*/

int32 l3_test13(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x77 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, mask, nexthop;
	

	START_TEST_CASE();

	{
		/*==================================================
		  * Test: In Class B network, add a default route and check to see:
		  *	      (1) lack nexthop's arp entry, the default route should be to CPU
		  *	      (2) add nexthop's arp entry, the default route should be L2
		  *	      (3) remove nexthop's arp entry, the default route should be to CPU
		  *==================================================*/
		/* VLAN 1: subnet 1 */	
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan1") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan1", RTL8651_LL_VLAN, 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.154.1"));
		mask = ntohl(inet_addr("255.255.0.0"));
		ASSERT(rtl8651_addIpIntf("vlan1", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);

		/* VLAN 2: subnet 2 */
		mac.octet[5] = 0x02;
		ASSERT(rtl8651_addVlan(2) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan2") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan2", RTL8651_LL_VLAN, 2) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.169.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("vlan2", ipaddr, mask) == SUCCESS);
		
		/* (1) add default route */
		nexthop =  ntohl(inet_addr("192.168.154.254"));
		ASSERT(rtl8651_addRoute(0, 0, "vlan1", nexthop) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= ntohl(inet_addr("192.168.0.0"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.0.0"));
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(0,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		asic_rt_1.ipAddr			= ntohl(inet_addr("192.169.1.0"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process 			= RT_ARP;
		asic_rt_1.vidx				= 2;
		asic_rt_1.arpStart			= 0;
		asic_rt_1.arpEnd			= 31;
		ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
#if 0 //Old model
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

		/* (2) add ARP for nexthop */
		mac.octet[5] = 0xf0;
		netif_t = GET_NETWORK_INTERFACE("vlan1", NULL);
		ASSERT(((uint32)netif_t)) ;
		ASSERT(_rtl8651_addLocalArp(TRUE, nexthop, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
#if 0 //Old model
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= 180;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 180; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

		/* (3) remove nextho's arp entry */
		ASSERT(_rtl8651_delLocalArp(TRUE, nexthop, netif_t) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
#if 0 //Old model
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
	}
	return SUCCESS;
}


/*
@func int32 | l3_test14 |Routing Table testing.
@comm
Overflow routing table (include ASIC and Software entries). Check to see no more than 7 entries will be added to ASIC
routing table.
*/

int32 l3_test14(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, mask, nexthop;
	uint32 round;
	

	START_TEST_CASE();

	{
		/*===============================================
		  * Test: add routes (exceed ASIC capability)
		  *	      use host route
		  *===============================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		for(round=0; round<rtl8651_tblDrvPara.routingTableSize-3 ; round++) {
			ipaddr = ntohl(inet_addr("140.113.1.1")) + (round << 8);
			nexthop = ntohl(inet_addr("192.168.1.2")) + round;
			ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "vlan1", nexthop) == SUCCESS);
		}
		ipaddr += (1<<8); nexthop ++;
		ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "vlan1", nexthop) != SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ntohl(inet_addr("192.168.1.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process	= RT_ARP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.arpStart	= 0;
		asic_rt_1.arpEnd	= 31;
		ASSERT(rtl8651_getAsicRouting(0, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		asic_rt_1.ipAddr	= ntohl(inet_addr("192.168.2.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process	= RT_ARP;
		asic_rt_1.vidx		= 2;
		asic_rt_1.arpStart	= 32;
		asic_rt_1.arpEnd	= 63;
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);		
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.255"));
#if 0 //Old model
		asic_rt_1.process	= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.2.1"));
#if 0 //Old model
		ASSERT(rtl8651_getAsicRouting(3, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(3,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.3.1"));
#if 0 //Old model
		ASSERT(rtl8651_getAsicRouting(4, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(4,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.4.1"));
#if 0 //Old model
		ASSERT(rtl8651_getAsicRouting(5, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(5,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.5.1"));
#if 0
		ASSERT(rtl8651_getAsicRouting(6, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(6,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
		asic_rt_1.ipAddr	= 0x00;
		asic_rt_1.ipMask	= 0x00;
#if 0 //Old model
		ASSERT(rtl8651_getAsicRouting(7, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
	}
	return SUCCESS;
}


/*
@func int32 | l3_test15 | Routing Table testing.
@comm
If routing table is full, remove the second entry. Check to see if refill mechanism is workable. This testing case in under
no default route environment. Hence if total entries > ASIC capability (7), the default route should be automatically added
a toCPU entry to trap mismatch packets to CPU to continue routing lookup. Otherwise, the default route should be set to
DROP to drop mismatch packets.
*/
int32 l3_test15(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, mask, nexthop;
	uint32 round;
	

	START_TEST_CASE();

	{
		/*===============================================
		  * Test: del routes (exceed ASIC capability)
		  *	      use host route
		  *	      no default route
		  *===============================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		for(round=0; round<rtl8651_tblDrvPara.routingTableSize-3 ; round++) {
			ipaddr = ntohl(inet_addr("140.113.1.1")) + (round << 8);
			nexthop = ntohl(inet_addr("192.168.1.2")) + round;
			ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "vlan1", nexthop) == SUCCESS);
		}

		for(round=0; round < rtl8651_tblDrvPara.routingTableSize-1-7-1; round++) {
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_1) == SUCCESS);
			ASSERT(rtl8651_delRoute(asic_rt_1.ipAddr, 0xffffffff) == SUCCESS);

			/* Verify L3 entry */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.6.1")) + (round << 8);
			asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.255"));
#if 0 //Old model
			asic_rt_1.process	= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process			= RT_NEXTHOP;
			asic_rt_1.nhStart			= 2;
			asic_rt_1.nhNum				= 2;
			asic_rt_1.nhNxt				= 2;
			asic_rt_1.nhAlgo			= 2;
			asic_rt_1.ipDomain			= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

			/* Verify default route */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.process	= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(7, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		}

		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_1) == SUCCESS);
		ASSERT(rtl8651_delRoute(asic_rt_1.ipAddr, 0xffffffff) == SUCCESS);
		/* Verify default route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process 		=RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_1) == SUCCESS);
		ASSERT(rtl8651_delRoute(asic_rt_1.ipAddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_1) != SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | l3_test16 | Routing Table testing.
@comm
If routing table is full, remove the second entry. Check to see if refill mechanism is workable. This testing case in under
haveing default route environment. Hence if total entries > ASIC capability (7), the default route should be automatically added
a toCPU entry to trap mismatch packets to CPU to continue routing lookup. Otherwise, the default route should be set to
default route which we added.
*/
int32 l3_test16(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, mask, nexthop;
	uint32 round;
	

	START_TEST_CASE();

	{
		/*===============================================
		  * Test: del routes (exceed ASIC capability)
		  *	      use host route
		  *	      with default route
		  *===============================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		/* add default route */
		nexthop = ntohl(inet_addr("192.168.1.254")) ;
		ASSERT(rtl8651_addRoute(0, 0, "vlan1", nexthop) == SUCCESS);
		ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

		for(round=0; round<rtl8651_tblDrvPara.routingTableSize-3 ; round++) {
			ipaddr = ntohl(inet_addr("140.113.1.1")) + (round << 8);
			nexthop = ntohl(inet_addr("192.168.1.2")) + round;
			ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "vlan1", nexthop) == SUCCESS);
		}

		for(round=0; round < rtl8651_tblDrvPara.routingTableSize-1-7-1; round++) {
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_1) == SUCCESS);
			ASSERT(rtl8651_delRoute(asic_rt_1.ipAddr, 0xffffffff) == SUCCESS);

			/* Verify L3 entry */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.6.1")) + (round << 8);
			asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.255"));
#if 0 //Old model
			asic_rt_1.process	= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process			= RT_NEXTHOP;
			asic_rt_1.nhStart			= 2;
			asic_rt_1.nhNum				= 2;
			asic_rt_1.nhNxt				= 2;
			asic_rt_1.nhAlgo			= 2;
			asic_rt_1.ipDomain			= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

			/* Verify default route */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
#if 0 //Old model
			asic_rt_1.process	= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(7, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else 
			asic_rt_1.process			= RT_NEXTHOP;
			asic_rt_1.nhStart			= 2;
			asic_rt_1.nhNum				= 2;
			asic_rt_1.nhNxt				= 2;
			asic_rt_1.nhAlgo			= 2;
			asic_rt_1.ipDomain			= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
		}

		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_1) == SUCCESS);
		ASSERT(rtl8651_delRoute(asic_rt_1.ipAddr, 0xffffffff) == SUCCESS);
		/* Verify default route */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0x00;
		asic_rt_1.ipMask			= 0x00;
#if 0 //Old model
		asic_rt_1.process			= RT_L2;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= 85;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(7, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 85; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_1) == SUCCESS);
		ASSERT(rtl8651_delRoute(asic_rt_1.ipAddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_1) != SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | l3_test17 | Routing Table testing.
@comm
Add routing to overflow routing table and remove it one by one in reverse order.
*/
int32 l3_test17(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, mask, nexthop;
	int32 round;
	

	START_TEST_CASE();

	{
		/*===============================================
		  * Test: remove route in reverse oder
		  *	      use host route
		  *===============================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		for(round=0; round<rtl8651_tblDrvPara.routingTableSize-3 ; round++) {
			ipaddr = ntohl(inet_addr("140.113.1.1")) + (round << 8);
			nexthop = ntohl(inet_addr("192.168.1.2")) + round;
			ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "vlan1", nexthop) == SUCCESS);
		}

		for(round=rtl8651_tblDrvPara.routingTableSize-4; round>=0; round--) {
			ipaddr = ntohl(inet_addr("140.113.1.1")) + (round << 8);
			ASSERT(rtl8651_delRoute(ipaddr, 0xffffffff) == SUCCESS);
			if (round < 6) {
				if (round <5) {		
					ASSERT(rtl8651_getAsicRouting(round+2, &asic_rt_2) != SUCCESS);
				}
				continue;
			}
				
			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= ntohl(inet_addr("192.168.1.0"));
			asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
			asic_rt_1.process	= RT_ARP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.arpStart	= 0;
			asic_rt_1.arpEnd	= 31;
			ASSERT(rtl8651_getAsicRouting(0, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
			asic_rt_1.ipAddr	= ntohl(inet_addr("192.168.2.0"));
			asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
			asic_rt_1.process	= RT_ARP;
			asic_rt_1.vidx		= 2;
			asic_rt_1.arpStart	= 32;
			asic_rt_1.arpEnd	= 63;
			ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);		
			asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.1.1"));
			asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.255"));
#if 0
			asic_rt_1.process	= RT_CPU;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process			= RT_NEXTHOP;
			asic_rt_1.nhStart			= 2;
			asic_rt_1.nhNum				= 2;
			asic_rt_1.nhNxt				= 2;
			asic_rt_1.nhAlgo			= 2;
			asic_rt_1.ipDomain			= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif
			asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.2.1"));
#if 0 //Old model
			ASSERT(rtl8651_getAsicRouting(3, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process			= RT_NEXTHOP;
			asic_rt_1.nhStart			= 2;
			asic_rt_1.nhNum				= 2;
			asic_rt_1.nhNxt				= 2;
			asic_rt_1.nhAlgo			= 2;
			asic_rt_1.ipDomain			= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(3,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
			asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.3.1"));
#if 0 //Old model
			ASSERT(rtl8651_getAsicRouting(4, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process			= RT_NEXTHOP;
			asic_rt_1.nhStart			= 2;
			asic_rt_1.nhNum				= 2;
			asic_rt_1.nhNxt				= 2;
			asic_rt_1.nhAlgo			= 2;
			asic_rt_1.ipDomain			= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(4,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
			asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.4.1"));
#if 0 //Old model
			ASSERT(rtl8651_getAsicRouting(5, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process			= RT_NEXTHOP;
			asic_rt_1.nhStart			= 2;
			asic_rt_1.nhNum				= 2;
			asic_rt_1.nhNxt				= 2;
			asic_rt_1.nhAlgo			= 2;
			asic_rt_1.ipDomain			= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(5,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

			asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.5.1"));
#if 0 //Old model
			ASSERT(rtl8651_getAsicRouting(6, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process			= RT_NEXTHOP;
			asic_rt_1.nhStart			= 2;
			asic_rt_1.nhNum				= 2;
			asic_rt_1.nhNxt				= 2;
			asic_rt_1.nhAlgo			= 2;
			asic_rt_1.ipDomain			= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(6,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif
			asic_rt_1.ipAddr	= 0x00;
			asic_rt_1.ipMask	= 0x00;
#if 0 //Old model
			ASSERT(rtl8651_getAsicRouting(7, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
			asic_rt_1.process			= RT_NEXTHOP;
			asic_rt_1.nhStart			= 2;
			asic_rt_1.nhNum				= 2;
			asic_rt_1.nhNxt				= 2;
			asic_rt_1.nhAlgo			= 2;
			asic_rt_1.ipDomain			= 6;

			/* Verify nexthop Table */
			ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
			asic_nh2.dvid				= 1;
			asic_nh2.extIntIpIdx		= 0;
			asic_nh2.isPppoe			= FALSE;
			asic_nh2.nextHopColumn		= 0;
			asic_nh2.nextHopRow			= 5; 
			asic_nh2.pppoeIdx			= 0;
			NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

		}

	}
	return SUCCESS;
}




int32 l3_test20(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t nexthop;
	int32 round;
	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add defaut route, VLAN type, nexthop Table, L3 only
		 *===========================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add default route */
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addRoute(0, 0, "vlan1", nexthop) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {

			/* (2) add arp for nexthop */
			ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= 0;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 85;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

			/* (3) remove arp */
			ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= 0;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}

		/* (4) remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process 		=RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}



int32 l3_test21(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ipaddr_t nexthop, naptip;
	int32 round;
	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add defaut route, VLAN type, nexthop Table, L4 
		 *===========================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (0) add NAPT IP */
		ASSERT(rtl8651_addLocalServer(ntohl(inet_addr("140.113.1.2"))) == SUCCESS);
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* Check GIDX. It is needed is due to ASIC bug */
		{
			uint32 GIDX;
			rtl8651_getAsicGidxRegister(&GIDX);
			//ASSERT(GIDX == 0x249249);
			ASSERT(GIDX == 0x00000000);
		}

		/* (1) add default route */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addRoute(0, 0, "wan", nexthop) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;//1;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {

			/* (2) add arp for nexthop */
			ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= 0;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 85;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

			/* (3) remove arp */
			ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;//1 //because we make sure the napt ip is ranged from 0 to 1;
			asic_nexthop.isPppoe		= 0;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}

		/* (4) remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 0;
		//ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
	}
	return SUCCESS;
}



int32 l3_test22(uint32 caseNo)
{
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x54 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add defaut route, PPPoE type, nexthop Table, L3 Only
		 *===========================================================*/
		default_configuration3();
		
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (1) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "pppoe", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 16;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= TRUE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPPoE Property */
			ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= TRUE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 16;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}
		
		/* (4) Remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


int32 l3_test23(uint32 caseNo)
{
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x54 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add defaut route, PPPoE type, nexthop Table, L4
		 *===========================================================*/
		default_configuration3();
		ASSERT(rtl8651_addExtNetworkInterface("pppoe") == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.2"));
		ASSERT((netif_t = GET_NETWORK_INTERFACE("lan", NULL)) != NULL);
		ASSERT(_rtl8651_addIpEntry(ipaddr, ipaddr, IP_ATTR_LS, netif_t) != NULL);
		

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_addNaptMapping(ipaddr) == SUCCESS);

		/* (1) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "pppoe", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;//1;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 16;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= TRUE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPPoE Property */
			ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= TRUE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 16;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}
		
		/* (4) Remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process 		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		//ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
	}
	return SUCCESS;
}


int32 l3_test24(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, mask, pptp_serv;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add defaut route, PPTP type, nexthop Table, L3 Only
		 *===========================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;
		
		ASSERT(rtl8651_delExtNetworkInterface("pptp") == SUCCESS);
		pptp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", pptp_serv, mask) == SUCCESS);
		pptp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(pptp_serv, &mac, "pptp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);		
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 12345, pptp_serv, 0) == SUCCESS);
		ipaddr = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		

		/* (1) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPPoE Property */
			ASSERT(rtl8651_setPptpProperty(1000, 11111, 1234, pptp_serv, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 1;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}
		
		/* (4) Remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}



int32 l3_test25(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, mask, pptp_serv;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add defaut route, PPTP type, nexthop Table, L4
		 *===========================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;
		
		ipaddr = ntohl(inet_addr("192.168.1.2"));
		ASSERT((netif_t = GET_NETWORK_INTERFACE("vlan", NULL)) != NULL);
		ASSERT(_rtl8651_addIpEntry(ipaddr, ipaddr, IP_ATTR_LS, netif_t) != NULL);
	
		pptp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", pptp_serv, mask) == SUCCESS);
		pptp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(pptp_serv, &mac, "pptp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);		
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 12345, pptp_serv, 0) == SUCCESS);
		ipaddr = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_addNaptMapping(ipaddr) == SUCCESS);
		

		/* (1) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;//1;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPTP Property */
			ASSERT(rtl8651_setPptpProperty(1000, 11111, 1234, pptp_serv, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 1;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}

		/* (4) Remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		//ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
	}
	return SUCCESS;
}



int32 l3_test26(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, mask, l2tp_serv;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add defaut route, L2TP type, nexthop Table, L3 Only
		 *===========================================================*/
		defaultConfig_l2tp1();
		DrvTbl.vlan[0].valid = 1;
		
		ASSERT(rtl8651_delExtNetworkInterface("l2tp") == SUCCESS);
		l2tp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", l2tp_serv, mask) == SUCCESS);
		l2tp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(l2tp_serv, &mac, "l2tp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);		
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 12345, l2tp_serv, 0) == SUCCESS);
		ipaddr = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		

		/* (1) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "l2tp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetL2tpProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPPoE Property */
			ASSERT(rtl8651_setL2tpProperty(1000, 11111, 1234, l2tp_serv, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 1;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}
		
		/* (4) Remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


int32 l3_test27(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, mask, l2tp_serv;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add defaut route, L2TP type, nexthop Table, L4
		 *===========================================================*/
		defaultConfig_l2tp1();
		DrvTbl.vlan[0].valid = 1;
		
		ipaddr = ntohl(inet_addr("192.168.1.2"));
		ASSERT((netif_t = GET_NETWORK_INTERFACE("vlan", NULL)) != NULL);
		ASSERT(_rtl8651_addIpEntry(ipaddr, ipaddr, IP_ATTR_LS, netif_t) != NULL);
	
		l2tp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", l2tp_serv, mask) == SUCCESS);
		l2tp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(l2tp_serv, &mac, "l2tp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);		
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 12345, l2tp_serv, 0) == SUCCESS);
		ipaddr = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_addNaptMapping(ipaddr) == SUCCESS);
		

		/* (1) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "l2tp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;//1;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetL2tpProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPTP Property */
			ASSERT(rtl8651_setL2tpProperty(1000, 11111, 1234, l2tp_serv, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= 0;
			asic_rt_1.ipMask	= 0;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 1;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}

		/* (4) Remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0;
		asic_rt_1.ipMask	= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		//ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
	}
	return SUCCESS;
}


int32 l3_test28(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t nexthop, ipaddr;
	int32 round;
	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add host route, VLAN type, nexthop Table, L3 only
		 *===========================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add default route */
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ipaddr = ntohl(inet_addr("10.10.10.10"));
		ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "vlan1", nexthop) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ipaddr;
		asic_rt_1.ipMask	= 0xffffffff;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {

			/* (2) add arp for nexthop */
			ASSERT(rtl8651_addArp(nexthop, &mac, "vlan1", 0x01) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= ipaddr;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= 0;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 85;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

			/* (3) remove arp */
			ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= ipaddr;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= 0;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}

		/* (4) remove default route */
		ASSERT(rtl8651_delRoute(ipaddr, 0xffffffff) == SUCCESS);

		/* Verify L3 Table */
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) != SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}

int32 l3_test29(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ipaddr_t nexthop, naptip, ipaddr;
	int32 round;
	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add host route, VLAN type, nexthop Table, L4 
		 *===========================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (0) add NAPT IP */
		ASSERT(rtl8651_addLocalServer(ntohl(inet_addr("140.113.1.2"))) == SUCCESS);
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* (1) add default route */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ipaddr = ntohl(inet_addr("10.10.10.10"));
		ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "wan", nexthop) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= ipaddr;
		asic_rt_1.ipMask	= 0xffffffff;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;//1;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {

			/* (2) add arp for nexthop */
			ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= ipaddr;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= 0;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 85;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

			/* (3) remove arp */
			ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= ipaddr;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= 0;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}

		/* (4) remove default route */
		ASSERT(rtl8651_delRoute(ipaddr, 0xffffffff) == SUCCESS);

		/* Verify L3 Table */
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) != SUCCESS);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 0;
		//ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
	}
	return SUCCESS;
}


int32 l3_test30(uint32 caseNo)
{
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x54 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, host;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add host route, PPPoE type, nexthop Table, L3 Only
		 *===========================================================*/
		default_configuration3();
		
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (1) add default route */
		host = ntohl(inet_addr("10.10.10.10"));
		ASSERT(rtl8651_addRoute(host, 0xffffffff, "pppoe", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= host;
		asic_rt_1.ipMask	= 0xffffffff;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 16;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= TRUE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPPoE Property */
			ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= TRUE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 16;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}
		
		/* (4) Remove default route */
		ASSERT(rtl8651_delRoute(host, 0xffffffff) == SUCCESS);

		/* Verify L3 Table */
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) != SUCCESS);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


int32 l3_test31(uint32 caseNo)
{
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x54 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, host;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add host route, PPPoE type, nexthop Table, L4
		 *===========================================================*/
		default_configuration3();
		ASSERT(rtl8651_addExtNetworkInterface("pppoe") == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.2"));
		ASSERT((netif_t = GET_NETWORK_INTERFACE("lan", NULL)) != NULL);
		ASSERT(_rtl8651_addIpEntry(ipaddr, ipaddr, IP_ATTR_LS, netif_t) != NULL);
		

		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_addNaptMapping(ipaddr) == SUCCESS);

		/* (1) add default route */
		host = ntohl(inet_addr("10.10.10.10"));
		ASSERT(rtl8651_addRoute(host, 0xffffffff, "pppoe", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= host;
		asic_rt_1.ipMask	= 0xffffffff;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;//1;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 16;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= TRUE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPPoE Property */
			ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= TRUE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 16;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}
		
		/* (4) Remove default route */
		ASSERT(rtl8651_delRoute(host, 0xffffffff) == SUCCESS);

		/* Verify L3 Table */
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) != SUCCESS);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
//		ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
	}
	return SUCCESS;
}



int32 l3_test32(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, mask, pptp_serv, host;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add host route, PPTP type, nexthop Table, L3 Only
		 *===========================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;
		
		ASSERT(rtl8651_delExtNetworkInterface("pptp") == SUCCESS);
		pptp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", pptp_serv, mask) == SUCCESS);
		pptp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(pptp_serv, &mac, "pptp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);		
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 12345, pptp_serv, 0) == SUCCESS);
		ipaddr = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		
		/* (1) add host route */
		host = ntohl(inet_addr("10.10.10.10"));
		ASSERT(rtl8651_addRoute(host, 0xffffffff, "pptp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= host;
		asic_rt_1.ipMask	= 0xffffffff;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		/* 
			Because the interface route of the outer subnet is removed, 
			the host route of "10.10.10.10" will be added in the entry 1 of L3 table.
		*/
		//ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);


		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPPoE Property */
			ASSERT(rtl8651_setPptpProperty(1000, 11111, 1234, pptp_serv, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			/* 
				Although the interface route of the outer subnet is removed, 
				the host route of "10.10.10.10" will not be removed to the entry 1 of L3 table.
				Therefore, we still verify entry 2 of L3 table.
			*/
			ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) != SUCCESS);
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 1;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}
		
		/* (4) Remove default route */
		ASSERT(rtl8651_delRoute(host, 0xffffffff) == SUCCESS);

		/* Verify L3 Table */

		/* 
			Because the interface route of outer subnet is removed, 
			we should verify entry 1 of L3 table.
		*/
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) != SUCCESS);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


int32 l3_test33(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, mask, pptp_serv, host;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add host route, PPTP type, nexthop Table, L4
		 *===========================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;
		
		ipaddr = ntohl(inet_addr("192.168.1.2"));
		ASSERT((netif_t = GET_NETWORK_INTERFACE("vlan", NULL)) != NULL);
		ASSERT(_rtl8651_addIpEntry(ipaddr, ipaddr, IP_ATTR_LS, netif_t) != NULL);
	
		pptp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", pptp_serv, mask) == SUCCESS);
		pptp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(pptp_serv, &mac, "pptp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);		
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 12345, pptp_serv, 0) == SUCCESS);
		ipaddr = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_addNaptMapping(ipaddr) == SUCCESS);
		

		/* (1) add host route */
		host = ntohl(inet_addr("10.10.10.10"));
		ASSERT(rtl8651_addRoute(host, 0xffffffff, "pptp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= host;
		asic_rt_1.ipMask	= 0xffffffff;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		/* 
			Because the interface route of the outer subnet is removed, 
			the host route of "10.10.10.10" will be added in the entry 1 of L3 table.
		*/
		//ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;//1;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPTP Property */
			ASSERT(rtl8651_setPptpProperty(1000, 11111, 1234, pptp_serv, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			/* 
				Although the interface route of the outer subnet is removed, 
				the host route of "10.10.10.10" will not be removed to the entry 1 of L3 table.
				Therefore, we still verify entry 2 of L3 table.
			*/
			ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) != SUCCESS);
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 1;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}

		/* (4) Remove host route */
		ASSERT(rtl8651_delRoute(host, 0xffffffff) == SUCCESS);

		/* Verify L3 Table */
		/* 
			Because the interface route of outer subnet is removed, 
			we should verify entry 1 of L3 table.
		*/
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) != SUCCESS);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
//		ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
	}
	return SUCCESS;
}


int32 l3_test34(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, mask, l2tp_serv, host;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add host route, L2TP type, nexthop Table, L3 Only
		 *===========================================================*/
		defaultConfig_l2tp1();
		DrvTbl.vlan[0].valid = 1;
		
		ASSERT(rtl8651_delExtNetworkInterface("l2tp") == SUCCESS);
		l2tp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", l2tp_serv, mask) == SUCCESS);
		l2tp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(l2tp_serv, &mac, "l2tp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);		
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 12345, l2tp_serv, 0) == SUCCESS);
		ipaddr = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		

		/* (1) add host route */
		host = ntohl(inet_addr("10.10.10.10"));
		ASSERT(rtl8651_addRoute(host, 0xffffffff, "l2tp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= host;
		asic_rt_1.ipMask	= 0xffffffff;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		/* 
			Because the interface route of outer subnet is removed, 
			we should verify entry 1 of L3 table.
		*/
		//ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetL2tpProperty(1000) == SUCCESS);
			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);


			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPPoE Property */
			ASSERT(rtl8651_setL2tpProperty(1000, 11111, 1234, l2tp_serv, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			/* 
				Although the interface route of the outer subnet is removed, 
				the host route of "10.10.10.10" will not be removed to the entry 1 of L3 table.
				Therefore, we still verify entry 2 of L3 table.
			*/
			ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) != SUCCESS);
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 1;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}
		
		/* (4) Remove host route */
		ASSERT(rtl8651_delRoute(host, 0xffffffff) == SUCCESS);

		/* Verify L3 Table */
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) != SUCCESS);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


int32 l3_test35(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, mask, l2tp_serv, host;
	int32 round;

	
	START_TEST_CASE();
	{
		/*===========================================================
		 * Test: add host route, L2TP type, nexthop Table, L4
		 *===========================================================*/
		defaultConfig_l2tp1();
		DrvTbl.vlan[0].valid = 1;
		
		ipaddr = ntohl(inet_addr("192.168.1.2"));
		ASSERT((netif_t = GET_NETWORK_INTERFACE("vlan", NULL)) != NULL);
		ASSERT(_rtl8651_addIpEntry(ipaddr, ipaddr, IP_ATTR_LS, netif_t) != NULL);
	
		l2tp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", l2tp_serv, mask) == SUCCESS);
		l2tp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(l2tp_serv, &mac, "l2tp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);		
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 12345, l2tp_serv, 0) == SUCCESS);
		ipaddr = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ASSERT(rtl8651_addNaptMapping(ipaddr) == SUCCESS);
		

		/* (1) add host route */
		host = ntohl(inet_addr("10.10.10.10"));
		ASSERT(rtl8651_addRoute(host, 0xffffffff, "l2tp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= host;
		asic_rt_1.ipMask	= 0xffffffff;
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		/* 
			Because the interface route of outer subnet is removed, 
			we should verify entry 1 of L3 table.
		*/
		//ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) != SUCCESS);		
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;//1;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		for(round=0; round<1000; round++) {
			/* (2) Rest PPPoE Property */
			ASSERT(rtl8651_resetL2tpProperty(1000) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
			/* (3) set PPTP Property */
			ASSERT(rtl8651_setL2tpProperty(1000, 11111, 1234, l2tp_serv, 0) == SUCCESS);

			/* Verify L3 Table */
			bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
			asic_rt_1.ipAddr	= host;
			asic_rt_1.ipMask	= 0xffffffff;
			asic_rt_1.process	= RT_NEXTHOP;
			asic_rt_1.vidx		= 1;
			asic_rt_1.nhStart	= 2;
			asic_rt_1.nhNum		= 2;
			asic_rt_1.nhNxt		= 2;
			asic_rt_1.nhAlgo	= 2;
			asic_rt_1.ipDomain	= 6;
			/* 
				Although the interface route of the outer subnet is removed, 
				the host route of "10.10.10.10" will not be removed to the entry 1 of L3 table.
				Therefore, we still verify entry 2 of L3 table.
			*/
			ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) != SUCCESS);
			ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) == SUCCESS);
			L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			/* Verify Nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 3;
			asic_nexthop.extIntIpIdx	= 0;//1;
			asic_nexthop.isPppoe		= FALSE;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow		= 1;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		}

		/* (4) Remove host route */
		ASSERT(rtl8651_delRoute(host, 0xffffffff) == SUCCESS);

		/* Verify L3 Table */
		ASSERT(rtl8651_getAsicRouting(1, &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(2, &asic_rt_2) != SUCCESS);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
//		ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
	}
	return SUCCESS;
}

int32 l3_test36(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask, nexthop;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;

	START_TEST_CASE();
	{
		//default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* VLAN 1: subnet 1 */	
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan1") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan1", RTL8651_LL_VLAN, 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("vlan1", ipaddr, mask) == SUCCESS);

		/* Add route: 140.113.100.0/24 */
		ipaddr = ntohl(inet_addr("140.113.100.0"));
		mask = ntohl(inet_addr("255.255.255.0"));
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) == SUCCESS);

		/* VLAN 2: subnet 2 */
		mac.octet[5] = 0x02;
		ASSERT(rtl8651_addVlan(2) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan2") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan2", RTL8651_LL_VLAN, 2) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.2.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("vlan2", ipaddr, mask) == SUCCESS);

		/* Add route: 140.113.1.0/24 */
		ipaddr = ntohl(inet_addr("140.113.1.0"));
		mask = ntohl(inet_addr("255.255.255.0"));
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) == SUCCESS);
		/* Add route: 140.113.2.0/24 */
		ipaddr = ntohl(inet_addr("140.113.2.0"));
		mask = ntohl(inet_addr("255.255.255.0"));
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) == SUCCESS);
		/* Add route: 140.114.0.0/16 */
		ipaddr = ntohl(inet_addr("140.114.0.0"));
		mask = ntohl(inet_addr("255.255.0.0"));
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) == SUCCESS);
		/* Add route: 140.114.1.0/24 */
		ipaddr = ntohl(inet_addr("140.114.1.0"));
		mask = ntohl(inet_addr("255.255.255.0"));
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) == SUCCESS);
		/* Add route: 140.114.1.1/32 */
		ipaddr = ntohl(inet_addr("140.114.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));
		nexthop = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addRoute(ipaddr, mask, "vlan1", nexthop) == SUCCESS);

		/* Check the order of the routes in ASIC L3 Table. */
		/* 192.168.1.0/24 */
		ASSERT(rtl8651_getAsicRouting(0,  &asic_rt_2) == SUCCESS);
		asic_rt_1.ipAddr	= ntohl(inet_addr("192.168.1.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process	= RT_ARP;
		asic_rt_1.vidx		= 1;
		asic_rt_1.arpStart	= 0;
		asic_rt_1.arpEnd	= 31;
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* 192.168.2.0/24 */
		ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
		asic_rt_1.ipAddr	= ntohl(inet_addr("192.168.2.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process	= RT_ARP;
		asic_rt_1.vidx		= 2;
		asic_rt_1.arpStart	= 32;
		asic_rt_1.arpEnd	= 63;
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* 140.114.1.1/32 */
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) == SUCCESS);
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.114.1.1"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.255"));
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 12;
		asic_rt_1.nhNum	= 2;
		asic_rt_1.nhNxt	= 12;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* 140.113.100.0/24 */
		ASSERT(rtl8651_getAsicRouting(3,  &asic_rt_2) == SUCCESS);
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.100.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum	= 2;
		asic_rt_1.nhNxt	= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* 140.113.1.0/24 */
		ASSERT(rtl8651_getAsicRouting(4,  &asic_rt_2) == SUCCESS);
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.1.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 4;
		asic_rt_1.nhNum	= 2;
		asic_rt_1.nhNxt	= 4;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* 140.113.2.0/24 */
		ASSERT(rtl8651_getAsicRouting(5,  &asic_rt_2) == SUCCESS);
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.113.2.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 6;
		asic_rt_1.nhNum	= 2;
		asic_rt_1.nhNxt	= 6;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* 140.114.1.0/24 */
		ASSERT(rtl8651_getAsicRouting(6,  &asic_rt_2) == SUCCESS);
		asic_rt_1.ipAddr	= ntohl(inet_addr("140.114.1.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 10;
		asic_rt_1.nhNum	= 2;
		asic_rt_1.nhNxt	= 10;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* default route */
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		asic_rt_1.ipAddr	= ntohl(inet_addr("0.0.0.0"));
		asic_rt_1.ipMask	= ntohl(inet_addr("0.0.0.0"));
		asic_rt_1.process	= RT_CPU;
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

	}
	return SUCCESS;
}











#define RTL865X_ARP_TESTING_CODE


/*
@func int32 | arp_test1 | ARP entry testing.
@comm
Check to see if a subnet is created, a broadcast arp should be added automatically.
*/

int32 arp_test1(uint32 caseNo)
{
	ether_addr_t mac = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_arpParam_t asic_arp_1, asic_arp_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;


	START_TEST_CASE();

	{
		/*===============================================
		  * Test: If a Subnet is created, a broadcast arp should be added
		  *	      automatically.
		  *===============================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* Verify ARP Table */
		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= 0;
		asic_arp_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(255,  &asic_arp_2) == SUCCESS);
		ARP_ENTRY_CMP(&asic_arp_1, &asic_arp_2);
		ASSERT(rtl8651_getAsicArp(511,  &asic_arp_2) == SUCCESS);
		ARP_ENTRY_CMP(&asic_arp_1, &asic_arp_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		//asic_l2_1.	memberPortMask	= 63;  //for RTL8651
		asic_l2_1.memberPortMask	= 511; // for RTL8650B
		asic_l2_1.	ageSec			= 100;
		asic_l2_1.	cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.nhFlag			= 0;		
		ASSERT(rtl8651_getAsicL2Table(0, 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;
}


/*
@func int32 | arp_test2 | ARP entry testing.
@comm
Add and delete a ARP entry.
*/

int32 arp_test2(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_arpParam_t asic_arp_1, asic_arp_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	ipaddr_t ipaddr;


	START_TEST_CASE();

	{
		/*==================================================
		  * Test: add/delete a ARP entry
		  *==================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add ARP entry */
		ipaddr = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addArp(ipaddr, &mac, "vlan1", 0x01) == SUCCESS);

		/* Verify ARP Table */
		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= rtl8651_filterDbIndex(&mac);
		asic_arp_2.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(254, &asic_arp_2) == SUCCESS);
		ARP_ENTRY_CMP(&asic_arp_1, &asic_arp_2);
	
		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		asic_l2_1.	memberPortMask	= 0x02;
		asic_l2_1.	ageSec			= 100;
		asic_l2_1.	cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;  /* XXXXXX: Currently, we always write it as dynamic entry and set */
		asic_l2_1.nhFlag			= 1;	/* nhFlag to avoid round up when writing it as static entry */	
		ASSERT(rtl8651_getAsicL2Table(rtl8651_filterDbIndex(&mac), 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (2) Remove the arp entry */
		ASSERT(rtl8651_delArp(ipaddr) == SUCCESS);

		/* Verify ARP Table */
		ASSERT(rtl8651_getAsicArp(254, &asic_arp_2) != SUCCESS);

		/* Verify L2 Table */
		ASSERT(rtl8651_getAsicL2Table(rtl8651_filterDbIndex(&mac), 0,  &asic_l2_2) != SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | arp_test3 | ARP entry testing.
@comm
ARP doesn't belong to gateway's any network interface will be forbidden.
*/

int32 arp_test3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	ipaddr_t ipaddr;


	START_TEST_CASE();

	{
		/*==================================================
		  * Test: add a unknown arp
		  *==================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		ipaddr = ntohl(inet_addr("192.168.9.2"));
		ASSERT(rtl8651_addArp(ipaddr, &mac, "vlan1", 0x01) != SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | arp_test4 | ARP entry testing.
@comm
Add a ARP entry and check to see if its corresponding L2 entry's member port is correct.
*/
int32 arp_test4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_arpParam_t asic_arp_1, asic_arp_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	ipaddr_t ipaddr;
	int32 port;
	

	START_TEST_CASE();

	{
		/*==================================================
		  * Test: (ARP) L2 entry's  member port
		  *==================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		ipaddr = ntohl(inet_addr("192.168.1.254"));
		for(port=0; port<RTL8651_PORT_NUMBER; port++) {
			ASSERT(rtl8651_addArp(ipaddr, &mac, "vlan1", port) == SUCCESS);

			/* Verify L2 Table */
			bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
			bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
			memcpy(&asic_l2_1.macAddr, &mac, 6);
			asic_l2_1.	memberPortMask	= (1<<port);
			asic_l2_1.	ageSec			= 100;
			asic_l2_1.	cpu				= 0;
			asic_l2_1.srcBlk			= 0;
			asic_l2_1.isStatic			= 0;  /* XXXXXX: Currently, we always write it as dynamic entry and set */
			asic_l2_1.nhFlag			= 1;	/* nhFlag to avoid round up when writing it as static entry */	
			ASSERT(rtl8651_getAsicL2Table(rtl8651_filterDbIndex(&mac), 0,  &asic_l2_2) == SUCCESS);
			L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
			
			ASSERT(rtl8651_delArp(ipaddr) == SUCCESS);	
		}

		/* add Arp entry: port mask = 0xffffffff */
		ASSERT(rtl8651_addArp(ipaddr, &mac, "vlan1", 0xffffffff) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		//asic_l2_1.	memberPortMask	= 63; //for RTL8651
		asic_l2_1.	memberPortMask	= 511; //for RTL8650B
		asic_l2_1.	ageSec			= 100;
		asic_l2_1.	cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;  /* XXXXXX: Currently, we always write it as dynamic entry and set */
		asic_l2_1.nhFlag			= 1;	/* nhFlag to avoid round up when writing it as static entry */	
		ASSERT(rtl8651_getAsicL2Table(rtl8651_filterDbIndex(&mac), 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
		
	}
	return SUCCESS;
}


/*
@func int32 | arp_test5 | ARP entry testing.
@comm
Make sure the duplicate arp adding (same IP but different MAC) will result in the previous arp entry be updated.
*/
int32 arp_test5(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_arpParam_t asic_arp_1, asic_arp_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr;
	uint32 temp;
	

	START_TEST_CASE();

	{
		/*==================================================
		  * Test: Internal API: the duplicate adding will modify the previous
		  *	      arp info.
		  *==================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		netif_t = GET_NETWORK_INTERFACE("vlan1", NULL) ;
		ASSERT(netif_t != NULL);

		/* (1) Add Arp entry */
		ipaddr = ntohl(inet_addr("192.168.1.254"));
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify ARP Table */
		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= rtl8651_filterDbIndex(&mac);
		asic_arp_2.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(254, &asic_arp_2) == SUCCESS);
		ARP_ENTRY_CMP(&asic_arp_1, &asic_arp_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		asic_l2_1.	memberPortMask	= 0x02;
		asic_l2_1.	ageSec			= 100;
		asic_l2_1.	cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;  /* XXXXXX: Currently, we always write it as dynamic entry and set */
		asic_l2_1.nhFlag			= 1;	/* nhFlag to avoid round up when writing it as static entry */	
		ASSERT(rtl8651_getAsicL2Table(rtl8651_filterDbIndex(&mac), 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (2) adding the same ARP (same Ip address) */
		temp = rtl8651_filterDbIndex(&mac);
		mac.octet[5] ++;
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify ARP table */
		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= rtl8651_filterDbIndex(&mac);
		asic_arp_2.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(254, &asic_arp_2) == SUCCESS);
		ARP_ENTRY_CMP(&asic_arp_1, &asic_arp_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		asic_l2_1.	memberPortMask	= 0x02;
		asic_l2_1.	ageSec			= 100;
		asic_l2_1.	cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;  /* XXXXXX: Currently, we always write it as dynamic entry and set */
		asic_l2_1.nhFlag			= 1;	/* nhFlag to avoid round up when writing it as static entry */	
		ASSERT(rtl8651_getAsicL2Table(rtl8651_filterDbIndex(&mac), 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);		

		ASSERT(rtl8651_getAsicL2Table(temp, 0,  &asic_l2_2) != SUCCESS);

		/* (3) remove the arp entry */
		ASSERT(_rtl8651_delLocalArp(TRUE, ipaddr, netif_t) == SUCCESS);

		/* Verify ARP Table */
		ASSERT(rtl8651_getAsicArp(254, &asic_arp_2) != SUCCESS);

		/* Verify L2 Table */
		ASSERT(rtl8651_getAsicL2Table(rtl8651_filterDbIndex(&mac), 0,  &asic_l2_2) != SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | arp_test6 | ARP entry testing.
@comm
Make sure the duplicate arp adding (same IP but different MAC) will result in the previous arp entry be updated. In this case
if a route references the arp entry, the route should be updated.
*/

int32 arp_test6(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_arpParam_t asic_arp_1, asic_arp_2;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr;
	uint32 temp;
	

	START_TEST_CASE();

	{
		/*==================================================
		  * Test: Internal API: the duplicate adding will modify the previous
		  *	      arp info. In this case, if a route references the entry, the
		  *	      route should update is arp info.
		  *==================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		netif_t = GET_NETWORK_INTERFACE("vlan1", NULL) ;
		ASSERT(netif_t != NULL);

		/* (1) Add Arp entry */
		ipaddr = ntohl(inet_addr("192.168.1.254"));
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify ARP Table */
		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= rtl8651_filterDbIndex(&mac);
		asic_arp_2.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(254, &asic_arp_2) == SUCCESS);
		ARP_ENTRY_CMP(&asic_arp_1, &asic_arp_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		asic_l2_1.	memberPortMask	= 0x02;
		asic_l2_1.	ageSec			= 100;
		asic_l2_1.	cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;  /* XXXXXX: Currently, we always write it as dynamic entry and set */
		asic_l2_1.nhFlag			= 1;	/* nhFlag to avoid round up when writing it as static entry */	
		ASSERT(rtl8651_getAsicL2Table(rtl8651_filterDbIndex(&mac), 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (2) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "vlan1", ipaddr) == SUCCESS);

#if 0 //Old model
		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= rtl8651_filterDbIndex(&mac);
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh2, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= rtl8651_filterDbIndex(&mac); 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif

		/* (3) add arp again (different MAC addr) */
		mac.octet[5] ++;
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x02) == SUCCESS);

#if 0 //Old model
		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= rtl8651_filterDbIndex(&mac);
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh2, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= FALSE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= rtl8651_filterDbIndex(&mac); 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif
	}
	return SUCCESS;
}


/*
@func int32 | arp_test7 | ARP entry testing.
@comm
In public API, the duplicate arp adding is forbidden.
*/
int32 arp_test7(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	ipaddr_t ipaddr;


	START_TEST_CASE();

	{
		/*==================================================
		  * Test: Released API: duplicate adding is not allowed.
		  *==================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		ipaddr = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addArp(ipaddr, &mac, "vlan1", 0x01) == SUCCESS);
		mac.octet[5]++;
		ASSERT(rtl8651_addArp(ipaddr, &mac, "vlan1", 0x01) != SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | arp_test8 | ARP entry testing.
@comm
Test internal API, rtl8651_addArp() is correct.
*/
int32 arp_test8(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl8651_tblDrv_arpEntry_t *arp_t;
	ipaddr_t ipaddr;


	START_TEST_CASE();

	{
		/*========================================
		  * Test: Internal API: _rtl8651_getArpEntry()
		  *========================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add ARP entry */
		ipaddr = ntohl(inet_addr("192.168.1.250"));
		ASSERT(rtl8651_addArp(ipaddr, &mac, "vlan1", 0x01) == SUCCESS);
		mac.octet[5]++;
		ipaddr = ntohl(inet_addr("192.168.1.254"));
		ASSERT(rtl8651_addArp(ipaddr, &mac, "vlan1", 0x01) == SUCCESS);

		/* (2) find the specified ARP */
		ipaddr = ntohl(inet_addr("192.168.1.254"));
		arp_t = _rtl8651_getArpEntry(NULL, ipaddr, FALSE);
		ASSERT( arp_t != NULL);
		ASSERT(arp_t->ipAddr == ipaddr);

		/* (3) find the specified ARP */
		ipaddr = ntohl(inet_addr("192.168.1.250"));
		netif_t = GET_NETWORK_INTERFACE("vlan1", NULL) ;
		ASSERT(netif_t != NULL);
		arp_t = _rtl8651_getArpEntry(netif_t, ipaddr, FALSE) ;
		ASSERT(arp_t != NULL);
		ASSERT(arp_t->ipAddr == ipaddr);
	}
	return SUCCESS;
}





#define RTL865X_L2_TABLE_TESTING_CODE

int32 l2_test1(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x08, 0x02, 0x97, 0x00, 0x73 } };
	rtl865x_tblAsicDrv_arpParam_t asic_arp_1, asic_arp_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr;
	
	{
		/*====================================
		  * Test: Trust new l2 info.
		  *====================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add arp in wan */
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_addArp(ipaddr, &mac, "wan", 0) == SUCCESS);

		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= 238;
		asic_arp_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(2,  &asic_arp_2) == SUCCESS);

		/* Verify l2 */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		asic_l2_1.memberPortMask	= 1; 
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.nhFlag			= 1;		
		ASSERT(rtl8651_getAsicL2Table(238, 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (2) move to lan */
		ipaddr = ntohl(inet_addr("192.168.1.2"));
		ASSERT(rtl8651_addArp(ipaddr, &mac, "vlan", 3) == SUCCESS);

		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= 238;
		asic_arp_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(258,  &asic_arp_2) == SUCCESS);
		ASSERT(rtl8651_getAsicArp(2,  &asic_arp_2) != SUCCESS);

		/* Verify l2 */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		asic_l2_1.memberPortMask	= 8; 
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.nhFlag			= 1;		
		ASSERT(rtl8651_getAsicL2Table(238, 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (3) move to wan */
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_addArp(ipaddr, &mac, "wan", 0) == SUCCESS);

		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= 238;
		asic_arp_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(2,  &asic_arp_2) == SUCCESS);

		/* Verify l2 */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		asic_l2_1.memberPortMask	= 1; 
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.nhFlag			= 1;		
		ASSERT(rtl8651_getAsicL2Table(238, 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;
}

int32 l2_test2(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x08, 0x02, 0x97, 0x00, 0x73 } };
	rtl865x_tblAsicDrv_arpParam_t asic_arp_1, asic_arp_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr;
	
	{
		/*====================================
		  * Test: Trust new l2 info.
		  *====================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add arp in wan */
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("wan", NULL)) != NULL);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0) == SUCCESS);

		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= 238;
		asic_arp_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(2,  &asic_arp_2) == SUCCESS);

		/* Verify l2 */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		asic_l2_1.memberPortMask	= 1; 
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.nhFlag			= 1;		
		ASSERT(rtl8651_getAsicL2Table(238, 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (2) move to lan */
		ipaddr = ntohl(inet_addr("192.168.1.2"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("vlan", NULL)) != NULL);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 3) == SUCCESS);

		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= 238;
		asic_arp_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(258,  &asic_arp_2) == SUCCESS);
		ASSERT(rtl8651_getAsicArp(2,  &asic_arp_2) != SUCCESS);

		/* Verify l2 */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		asic_l2_1.memberPortMask	= 8; 
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.nhFlag			= 1;		
		ASSERT(rtl8651_getAsicL2Table(238, 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (3) move to wan */
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("wan", NULL)) != NULL);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0) == SUCCESS);

		bzero(&asic_arp_1, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		bzero(&asic_arp_2, sizeof(rtl865x_tblAsicDrv_arpParam_t));
		asic_arp_1.nextHopRow		= 238;
		asic_arp_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicArp(2,  &asic_arp_2) == SUCCESS);

		/* Verify l2 */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		memcpy(&asic_l2_1.macAddr, &mac, 6);
		asic_l2_1.memberPortMask	= 1; 
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.srcBlk			= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.nhFlag			= 1;		
		ASSERT(rtl8651_getAsicL2Table(238, 0,  &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;
}





#define RTL865X_IP_INTERFACE_TESTING_CODE


/* 
@func int32 | ip_interface_test_1 | IP Interface (subnet) testing.
@comm
Add and delete a IP interface (subnet).
*/
int32 ip_interface_test_1(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	ipaddr_t ipaddr, mask;
	uint32 round;

	START_TEST_CASE();

	{
		/*====================================
		  * Test: IP interface add/delete
		  *====================================*/
		/* (1) add Ip interface */
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan1") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan1", RTL8651_LL_VLAN, 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));

		for(round=0; round<1000; round++)  {
			ASSERT(rtl8651_addIpIntf("vlan1", ipaddr, mask) == SUCCESS);

			/* (2) remove Ip interface */
			ASSERT(rtl8651_delIpIntf("vlan1", ipaddr,  mask) == SUCCESS);
		}
	}
	return SUCCESS;
}


/* 
@func int32 | ip_interface_test_2 | IP Interface (subnet) testing.
@comm
When add or delete a IP interface (subnet), the specified network interface name should exist.
*/

int32 ip_interface_test_2(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	ipaddr_t ipaddr, mask;
	

	START_TEST_CASE();

	{
		/*====================================
		  * Test: add/delete to a non-exist network interface
		  *====================================*/
		/* (1) add Ip interface */
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan1") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan1", RTL8651_LL_VLAN, 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("vlan2", ipaddr, mask) != SUCCESS);

		/* (2) remove Ip interface */
		ASSERT(rtl8651_delIpIntf("vlan2", ipaddr,  mask) != SUCCESS);
	}
	return SUCCESS;
}



/* 
@func int32 | ip_interface_test_3 | IP Interface (subnet) testing.
@comm
Add and Delete a Class A/B subnet.
*/

int32 ip_interface_test_3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, mask;
	uint32 count=0;
	

	START_TEST_CASE();

	{
		/*====================================
		  * Test: 
		  *====================================*/
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan1") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan1", RTL8651_LL_VLAN, 1) == SUCCESS);
again:
		
		ipaddr = ntohl(inet_addr("110.1.1.1"));
		mask = ntohl(inet_addr("255.0.0.0"));
		ASSERT(rtl8651_addIpIntf("vlan1", ipaddr, mask) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= ipaddr&mask;
		asic_rt_1.ipMask			= mask;
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(0,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		ASSERT(rtl8651_delIpIntf("vlan1", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_getAsicRouting(0,  &asic_rt_2) != SUCCESS);

		if (count < 10000) {
			count ++;
			goto again;
		}
	}
	return SUCCESS;
}


int32 ip_interface_test_4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	ipaddr_t ipaddr, mask;
	uint32 round;

	START_TEST_CASE();

	{
		/*====================================
		  * Test: IP interface add/delete
		  *====================================*/
		/* (1) add Ip interface */
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan1") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan1", RTL8651_LL_PPTP, 1) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		rtl8651_setLoopbackPortPHY(RTL8651_MII_PORT, TRUE);
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "vlan1", 1) == SUCCESS);
		
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));

		for(round=0; round<1000; round++)  {
			ASSERT(rtl8651_addIpIntf("vlan1", ipaddr, mask) == SUCCESS);
			ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

			/* (2) remove Ip interface */
			ASSERT(rtl8651_delIpIntf("vlan1", ipaddr,  mask) == SUCCESS);
		}
	}
	return SUCCESS;
}


int32 ip_interface_test_5(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	ipaddr_t ipaddr, mask;
	uint32 round;

	START_TEST_CASE();

	{
		/*====================================
		  * Test: IP interface add/delete with mask=0xffffffff
		  *====================================*/
		/* (1) add Ip interface */
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan1") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan1", RTL8651_LL_VLAN, 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.255"));

		for(round=0; round<1000; round++)  {
			ASSERT(rtl8651_addIpIntf("vlan1", ipaddr, mask) != SUCCESS);

			/* (2) remove Ip interface */
			ASSERT(rtl8651_delIpIntf("vlan1", ipaddr,  mask) != SUCCESS);
		}
	}
	return SUCCESS;
}








#define RTL865X_NETWORK_INTERFACE_TESTING_CODE





#define RTL865X_L2_TABLE_TESTING_CODE



#define RTL865X_NAT_MAPPING_TESTING_CODE






#define RTL865X_NAPT_MAPPING_TESTING_CODE











#define RTL865X_LOCAL_PUBLIC_TESTING_CODE



#define RTL865X_PPPoE_TESTING_CODE


/*
@func int32 | pppoe_test1 | PPPoE testing.
@comm
Add and delete PPPoE session. Overflow PPPoE table and then remove it one by one.
*/
int32 pppoe_test1(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	uint32 session;


	START_TEST_CASE();

	{
		/*======================================= 
		  * Test: add/delete PPPoE session (old model)
		  *	      MAX session: 8
		  *=======================================*/
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("pppoe") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("pppoe", RTL8651_LL_PPPOE, 1) == SUCCESS);

		/* (1) add session: up to 8 sessions */
		for(session=0; session<RTL8651_PPPOETBL_SIZE; session++) {
			ASSERT(rtl8651_addPppoeSession(1000+session, 1) == SUCCESS);
		}
		ASSERT(rtl8651_addPppoeSession(2000, 1) != SUCCESS);

		/* (2) remove session */
		for(session=0; session<RTL8651_PPPOETBL_SIZE; session++) {		
			ASSERT(rtl8651_delPppoeSession(1000+session) == SUCCESS);
		}
		ASSERT(rtl8651_delPppoeSession(2000) != SUCCESS);
	}
	return SUCCESS;
}



/*
@func int32 | pppoe_test2 | PPPoE testing.
@comm
Dial up PPPoE session and then hang up it.
*/

int32 pppoe_test2(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	rtl865x_tblAsicDrv_pppoeParam_t asic_pppoe_1, asic_pppoe_2;
	uint32 session;
	

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: set/reset PPPoE property
		  *==============================================*/
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("pppoe") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("pppoe", RTL8651_LL_PPPOE, 1) == SUCCESS);
		mac.octet[5] = 0x10;

		/* (1) Add PPPoE session */
		for(session=0; session<RTL8651_PPPOETBL_SIZE; session++) {
			/* add PPPoE session and set its property */
			ASSERT(rtl8651_addPppoeSession(1000+session, 1) == SUCCESS);
			ASSERT(rtl8651_setPppoeSessionProperty(1000+session, 11111*(session+1), &mac, 0x01) == SUCCESS);

			/* Verify PPPoe Table */
			bzero(&asic_pppoe_1, sizeof(rtl865x_tblAsicDrv_pppoeParam_t));
			bzero(&asic_pppoe_1, sizeof(rtl865x_tblAsicDrv_pppoeParam_t));
			asic_pppoe_1.sessionId	= 11111*(session+1);
			ASSERT(rtl8651_getAsicPppoe(session, &asic_pppoe_2) == SUCCESS);
			PPPOE_ENTRY_CMP(&asic_pppoe_1, &asic_pppoe_2);
			mac.octet[5] ++;
		}

		/* (2) Del PPPoE Session */
		for(session=0, mac.octet[5]=0x10; session<RTL8651_PPPOETBL_SIZE; session++, mac.octet[5]++) {
			/* Reset PPPoE session */
			ASSERT(rtl8651_resetPppoeSessionProperty(session+1000) == SUCCESS);
		}
	}
	return SUCCESS;
}


int32 pppoe_test3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	ipaddr_t ipaddr;
	uint32 session;
	

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: 
		  *==============================================*/
		default_configuration3();

		/* (1) add default route to PPPoE */
		ASSERT(rtl8651_addRoute(0, 0, "pppoe", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0x00;
		asic_rt_1.ipMask	= 0x00;
#if 0 //Old model
		asic_rt_1.process	= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= TRUE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

		/* (2) dial up PPPoE session */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0x00;
		asic_rt_1.ipMask	= 0x00;
#if 0 //Old model
		asic_rt_1.process	= RT_PPPOE;
		asic_rt_1.vidx		= 1;
		asic_rt_1.pppoeIdx	= 0;
		asic_rt_1.nextHopRow		= 69;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= TRUE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 69; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif
	}
	return SUCCESS;
}



static int32 pppoe_default_configuration(void)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask;

	/* VLAN 1: subnet 1 */	
	ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("pppoe") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("pppoe", RTL8651_LL_PPPOE, 1) == SUCCESS);
	ASSERT(rtl8651_addExtNetworkInterface("pppoe") == SUCCESS);

	/* VLAN 2: subnet 2 */
	mac.octet[5] = 0x02;
	ASSERT(rtl8651_addVlan(2) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("vlan") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan", RTL8651_LL_VLAN, 2) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.2.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("vlan", ipaddr, mask) == SUCCESS);		
	return SUCCESS;
}



/*
@func int32 | pppoe_test10 | PPPoE testing.
@comm
Test add and delete PPPoE session (use new Model).
*/
int32 pppoe_test10(uint32 caseNo)
{
	
	START_TEST_CASE();

	{
		/*==============================================
		  * Test: add/delete PPPoE session by using new API
		  *==============================================*/
		pppoe_default_configuration();

		/* original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 1000, "pppoe", 1) == SUCCESS);
		ASSERT(_rtl8651_getPPPInfo(1000, 0) != NULL);
		ASSERT(rtl8651_delSession(1000) == SUCCESS);
		ASSERT(_rtl8651_getPPPInfo(1000, 0) == NULL);

		/* new network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 1000, "pppoe1", 1) == SUCCESS);
		ASSERT(_rtl8651_getPPPInfo(1000, 0) != NULL);
		ASSERT(rtl8651_delSession(1000) == SUCCESS);
		ASSERT(_rtl8651_getPPPInfo(1000, 0) == NULL);
	}
	return SUCCESS;
}


/*
@func int32 | pppoe_test11 | PPPoE testing.
@comm
Dial up PPPoE session and then hang up it. (use new Model)
*/

int32 pppoe_test11(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_pppoeParam_t asic_pppoe1, asic_pppoe2;
	ipaddr_t ipaddr;

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: set/resest pppoe property
		  *==============================================*/
		pppoe_default_configuration();
		/* original network interface */
		/* (1) set pppoe session property */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 1000, "pppoe", 1) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* Verify PPPoE TAble */
		bzero(&asic_pppoe1, sizeof(rtl865x_tblAsicDrv_pppoeParam_t));
		bzero(&asic_pppoe2, sizeof(rtl865x_tblAsicDrv_pppoeParam_t));
		asic_pppoe1.age		= 0 ;
		asic_pppoe1.sessionId	= 11111;
		ASSERT(rtl8651_getAsicPppoe(0,  &asic_pppoe2) == SUCCESS);
		PPPOE_ENTRY_CMP(&asic_pppoe1, &asic_pppoe2);

		/* (2) reset pppoe session */
		ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);
		ASSERT(rtl8651_delIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);

		/* new network interface */
		/* (3) set pppoe1 session property */
		mac.octet[5] ++;
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 1001, "pppoe1", 1) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(1001, 22222, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe1", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1001) == SUCCESS);

		/* Verify PPPoE TAble */
		bzero(&asic_pppoe1, sizeof(rtl865x_tblAsicDrv_pppoeParam_t));
		bzero(&asic_pppoe2, sizeof(rtl865x_tblAsicDrv_pppoeParam_t));
		asic_pppoe1.age		= 0 ;
		asic_pppoe1.sessionId	= 22222;
		ASSERT(rtl8651_getAsicPppoe(1,  &asic_pppoe2) == SUCCESS);
		PPPOE_ENTRY_CMP(&asic_pppoe1, &asic_pppoe2);
		
		/* (2) reset pppoe1 session */
		ASSERT(rtl8651_resetPppoeSessionProperty(1001) == SUCCESS);
		ASSERT(rtl8651_delIpIntf("pppoe1", ipaddr, 0xffffffff) == SUCCESS);
	}
	return SUCCESS;
}



/*
@func int32 | pppoe_test12 | PPPoE testing.
@comm
Add route to PPPoE session. Check the route if correct in dial up state and hang up state separately. The route is added before
the session dial up.
*/

int32 pppoe_test12(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	int8 *ifname[] = { "pppoe", "pppoe1" };
	ipaddr_t ipaddr;
	uint32 round=0;
	

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: add route to pppoe
		  *	      add route before dialing up the session
		  *==============================================*/
		pppoe_default_configuration();

again:
		/* (1) add pppoe session */
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 1000, ifname[round], 1) == SUCCESS);

		/* (2) add default route to the pppoe session */
		ASSERT(rtl8651_addRoute(0, 0, ifname[round], 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0x00;
		asic_rt_1.ipMask	= 0x00;
#if 0 //Old model
		asic_rt_1.process	= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= TRUE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

		/* (3) dial up the pppoe session */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf(ifname[round], ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0x00;
		asic_rt_1.ipMask			= 0x00;
#if 0 //Old model
		asic_rt_1.process			= RT_PPPOE;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= 85;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= TRUE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 85; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

		/* (4) reset pppoe session */
		ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);
		ASSERT(rtl8651_delIpIntf(ifname[round], ipaddr, 0xffffffff) == SUCCESS);
		
		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0x00;
		asic_rt_1.ipMask	= 0x00;
#if 0 //Old model
		asic_rt_1.process	= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= TRUE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 5; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);
#endif

		/* (5) remove route */
		ASSERT(rtl8651_delSession(1000) != SUCCESS);
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);
		
		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0x00;
		asic_rt_1.ipMask	= 0x00;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (6) remove session */
		ASSERT(rtl8651_delSession(1000) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0x00;
		asic_rt_1.ipMask	= 0x00;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		if (round < 1) {
			round ++;
			goto again;
		}
	}
	return SUCCESS;
}



/*
@func int32 | pppoe_test13 | PPPoE testing.
@comm
Add route to PPPoE session. Check the route if correct in dial up state and hang up state separately. The route is added after
the session dial up.
*/

int32 pppoe_test13(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x11 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_nextHopParam_t asic_nh1, asic_nh2;
	int8 *ifname[] = { "pppoe", "pppoe1" };
	ipaddr_t ipaddr;
	uint32 round=0;
	

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: add route to pppoe 
		  *	      add route after dialing up the session
		  *==============================================*/
		pppoe_default_configuration();

again:
		/* (1) add pppoe session */
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 1000, ifname[round], 1) == SUCCESS);

		/* (2) dial up the pppoe session */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf(ifname[round], ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (3) add default route to the pppoe session */
		ASSERT(rtl8651_addRoute(0, 0, ifname[round], 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0x00;
		asic_rt_1.ipMask			= 0x00;
#if 0 //Old model
		asic_rt_1.process			= RT_PPPOE;
		asic_rt_1.vidx				= 1;
		asic_rt_1.nextHopRow		= 85;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
#else
		asic_rt_1.process	= RT_NEXTHOP;
		asic_rt_1.nhStart	= 2;
		asic_rt_1.nhNum		= 2;
		asic_rt_1.nhNxt		= 2;
		asic_rt_1.nhAlgo	= 2;
		asic_rt_1.ipDomain	= 6;

		/* Verify nexthop Table */
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nh1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(asic_rt_2.nhStart, &asic_nh1) == SUCCESS);
		asic_nh2.dvid				= 1;
		asic_nh2.extIntIpIdx		= 0;
		asic_nh2.isPppoe			= TRUE;
		asic_nh2.nextHopColumn		= 0;
		asic_nh2.nextHopRow			= 85; 
		asic_nh2.pppoeIdx			= 0;
		NXTHOP_ENTRY_CMP(&asic_nh1, &asic_nh2);

#endif

		/* (4) remove route */
		ASSERT(rtl8651_delSession(1000) != SUCCESS);
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);
		
		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0x00;
		asic_rt_1.ipMask	= 0x00;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process 		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (5) reset pppoe session */
		ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);
		ASSERT(rtl8651_delIpIntf(ifname[round], ipaddr, 0xffffffff) == SUCCESS);
		
		/* (6) remove session */
		ASSERT(rtl8651_delSession(1000) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr	= 0x00;
		asic_rt_1.ipMask	= 0x00;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process	= RT_DROP;
		asic_rt_1.process		= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);


		if (round < 1) {
			round ++;
			goto again;
		}

	}
	return SUCCESS;
}
























#define RTL865X_PPTP_TESTING_CODE

static int32 defaultConfig_pptp1(void)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask;
	extern int32 rtl8651_loopbackPortPHY; 


	/* VLAN 1: subnet 1 */	
	ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("pptp") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("pptp", RTL8651_LL_PPTP, 1) == SUCCESS);
	ASSERT(rtl8651_addExtNetworkInterface("pptp") == SUCCESS);

	/* VLAN 2: subnet 2 */
	mac.octet[5] = 0x02;
	ASSERT(rtl8651_addVlan(2) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("vlan") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan", RTL8651_LL_VLAN, 2) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.2.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("vlan", ipaddr, mask) == SUCCESS);

	//rtl8651_setLoopbackPortPHY(RTL8651_MII_PORT, TRUE);
	rtl8651_setLoopbackPortPHY(0x07, TRUE);
	return SUCCESS;
}	


#if 0

/* OLD MODEL: no nexthop table */



/*
@func int32 | pptp_test4 | Testing PPTP.
@comm
After dial up a PPTP session, add the default route to the PPTP session. Make sure the route is correct.
Then hang up the PPTP session, the route information should change to incomplete state.
*/
int32 pptp_test4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp after dial pptp session
		  *	      delete default route
		  *=================================================*/
		defaultConfig_pptp1();

		/* (0) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);
		
		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);

		/* (2) add IP Interface */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (3) add arp & set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* (4) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 0;
		asic_rt_1.nextHopRow		= 68;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80; //0x20;   1000 0000
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_rt_1.nextHopRow, asic_rt_1.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
		
		/* (5) remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_DROP;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;
}




/*
@func int32 | pptp_test5 | Testing PPTP.
@comm
Before dial up a PPTP session, add the default route to the PPTP session. Make sure the route is correct.
Then dial up the PPTP session, the route should update its information.
*/
int32 pptp_test5(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp before dial pptp session
		  *	      delete default route
		  *=================================================*/
		defaultConfig_pptp1();

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);


		/* (3) add default route  to pptp */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Entry */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (4) add arp & set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) dial up pptp session */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 0;
		asic_rt_1.nextHopRow		= 68;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);


		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_rt_1.nextHopRow, asic_rt_1.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (6) remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_DROP;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;
}



/*
@func int32 | pptp_test6 | Testing PPTP.
@comm
Add default route to a PPTP session. Test if it is correct when set/reset (dial/hang-up) the PPTP session.
*/
int32 pptp_test6(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp before dial pptp session
		  *	      set/reset pptp session 
		  *=================================================*/
		defaultConfig_pptp1();

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);


		/* (3) add default route  to pptp */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Entry */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (4) add arp & set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) dial up pptp session */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 0;
		asic_rt_1.nextHopRow		= 68;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);


		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_rt_1.nextHopRow, asic_rt_1.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (6) rest pptp session */
		ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
 
	}
	return SUCCESS;
}


#endif


/*
@func int32 | pptp_test1 | Test PPTP.
@comm
Make sure adding and deleting PPTP session is workable.
*/
int32 pptp_test1(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan_1, asic_vlan_2;
	int8 portState[6] = { RTL8651_PORTSTA_FORWARDING, RTL8651_PORTSTA_FORWARDING, RTL8651_PORTSTA_FORWARDING, 
			  	          RTL8651_PORTSTA_FORWARDING, RTL8651_PORTSTA_FORWARDING, RTL8651_PORTSTA_FORWARDING };
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	extern rtl8651_tblDrv_miiTunneling_t tunnel;
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	int32 curr;
	uint16 vidx;
	
	
	START_TEST_CASE();

	{
		/*===================================================
  		  * Test: add/delete a session for PPTP
  		  *===================================================*/
  		defaultConfig_pptp1();

		curr = 0;
  		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);

		/* Verify Vlan Table for Virtual VLAN */
		bzero(&asic_vlan_1, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		bzero(&asic_vlan_2, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		asic_vlan_1.broadcastToCpu		= 1;
		asic_vlan_1.enableRoute			= 1;
		asic_vlan_1.inAclEnd			= 1;
		asic_vlan_1.inAclStart			= 0;
		asic_vlan_1.internal			= 0;			
		asic_vlan_1.macAddrNumber		= 1;
		asic_vlan_1.memberPortMask		= 0x80;
		asic_vlan_1.mtu					= 1500;
		asic_vlan_1.outAclEnd			= RTL8651_ACLTBL_PERMIT_ALL;
		asic_vlan_1.outAclStart			= RTL8651_ACLTBL_PERMIT_ALL;		
		asic_vlan_1.promiscuous			= 0;
		asic_vlan_1.untagPortMask		= 0x80;
		asic_vlan_1.valid				= 1;
		asic_vlan_1.vid					= 2048;
		rtl8651_memcpy(&asic_vlan_1.macAddr, &mac, sizeof(ether_addr_t));
		rtl8651_memcpy(asic_vlan_1.portState, portState, 6);
		ASSERT(rtl8651_getAsicVlan(2048,  &asic_vlan_2) == SUCCESS);
		VLAN_ENTRY_CMP(&asic_vlan_1, &asic_vlan_2);

	 	curr++; // Trap packets for PPTP/L2TP

		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_	= RTL8651_ACL_CPU;
		rule1.srcMac_		= mac;
		memset(&rule1.srcMacMask_, 0xff, 6);
		ASSERT(rtl8651_getAsicAclRule(asic_vlan_1.inAclStart, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		ACL_DEFAULT_PERMIT(asic_vlan_1.inAclEnd);
		ASSERT((asic_vlan_1.inAclStart+curr) == asic_vlan_1.inAclEnd);

		curr = 0;
	 	//curr++; // Trap packets for PPTP/L2TP
		/* Verify ACL: vlan 1 */
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE ) curr++;
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
#ifdef TRAP_TCP_SYN_FIN_RST
		curr += 3; /* tcp syn, fin, rst */
#else
		curr += 1; /* tcp syn */
#endif
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_	= RTL8651_ACL_MAC;
		rule1.actionType_	= RTL8651_ACL_CPU;
		rule1.dstMac_		= mac;
		memset(&rule1.dstMacMask_, 0xff, 6);
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		ACL_DEFAULT_PERMIT(asic_vlan.inAclEnd);
		ASSERT((asic_vlan.inAclStart+curr) == asic_vlan.inAclEnd);

		curr = 0;
	 	//curr++; // Trap packets for PPTP/L2TP
	 	curr += 4; /* for tcp syn, fin ,rst */
		/* Verify ACL: vlan 2 */
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE ) curr++;
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
//		ACL_DEFAULT_PERMIT(asic_vlan.inAclEnd);
//		ASSERT(asic_vlan.inAclStart == asic_vlan.inAclEnd);

		ACL_DEFAULT_PERMIT(asic_vlan.inAclStart+curr);
		ASSERT(asic_vlan.outAclStart == asic_vlan.outAclEnd);
		ASSERT(tunnel.valid == 1);
		
	
		ASSERT(rtl8651_delSession(1000) == SUCCESS);
		ASSERT(tunnel.valid == 0);

		/* Verify */
		ASSERT(rtl8651_getAsicVlan(2048,  &asic_vlan_2) != SUCCESS);

		/* (2) create another new network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp1", 1) == SUCCESS);
		ASSERT(rtl8651_delSession(1000) == SUCCESS);
		ASSERT(GET_NETWORK_INTERFACE("pptp1", NULL) == SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | pptp_test2 | Test PPTP.
@comm
Make sure IP interface (subnet) can be added to and removed from a PPTP session.
*/
int32 pptp_test2(uint32 caseNo)
{
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl8651_tblDrvSession_t *session_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add/delete pptp session with IP address
		  *=================================================*/
		defaultConfig_pptp1();

		/* (1) add to original network interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);
		ASSERT(rtl8651_delSession(1000) == SUCCESS);
		
		/* (1.1) add broadcast netowrk to PPTP */
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);
		/* Verify L3 */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= ntohl(inet_addr("192.168.2.0"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process 			= RT_ARP;
		asic_rt_1.vidx				= 2;
		asic_rt_1.arpStart			= 0;
		asic_rt_1.arpEnd			= 31;
		ASSERT(rtl8651_getAsicRouting(0,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		asic_rt_1.ipAddr			= ntohl(inet_addr("192.168.1.0"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.0"));

 		/*
		  *   When the linkLayerType is PPTP,
		  *   the PROCESS of the interface routing entry is always to CPU.
		  *   This is for "pptp dial demand (when pinging  pptp server IP)" spec.
		  */

		//asic_rt_1.process 			= RT_ARP;
		asic_rt_1.process 			= RT_CPU;
#if 0
		asic_rt_1.vidx				= 1;
		asic_rt_1.arpStart			= 32;
		asic_rt_1.arpEnd			= 63;
#endif
		ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process 			= RT_DROP;
		asic_rt_1.process				= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(3,  &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(4,  &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(5,  &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(6,  &asic_rt_2) != SUCCESS);

		/* (1.2) add tunneling network  */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		/* Verify */
		ASSERT((netif_t = GET_NETWORK_INTERFACE("pptp", NULL)) != NULL);
		ASSERT((session_t=CTAILQ_FIRST(&netif_t->sessionHead)) != SUCCESS);
		ASSERT(session_t->ip_t);
		ASSERT(session_t->ip_t->ipAddr == ipaddr);
		ASSERT(session_t->ip_t->nextIpAddr.sle_next == NULL);
	

		/* (2) add to new network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1001, "pptp1", 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp1", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1001) == SUCCESS);
		/* Verify */
		ASSERT((netif_t = GET_NETWORK_INTERFACE("pptp1", NULL)) != NULL);
		ASSERT((session_t=CTAILQ_FIRST(&netif_t->sessionHead)) != SUCCESS);
		ASSERT(session_t->ip_t);
		ASSERT(session_t->ip_t->ipAddr == ipaddr);
		ASSERT(session_t->ip_t->nextIpAddr.sle_next == NULL);
	
		/* (3) Remove PPTP  */
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_delIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_delIpIntf("pptp1", ipaddr, 0xffffffff) == SUCCESS);

		/* Verify */
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp1", NULL)) != NULL);
		ASSERT((session_t=CTAILQ_FIRST(&netif_t->sessionHead)) != SUCCESS);
		ASSERT(session_t->ip_t == NULL);
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != NULL);
		ASSERT((session_t=CTAILQ_FIRST(&netif_t->sessionHead)) != SUCCESS);
		ASSERT(session_t->ip_t == NULL);

		/* (4) remove Session */
		ASSERT(rtl8651_delSession(1000) == SUCCESS);
		ASSERT(rtl8651_delSession(1001) == SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | pptp_test3 | Test PPTP.
@comm
Test if rtl8651_setPptpProperty() and rtl8651_resetPptpProperty() are workable.
*/
int32 pptp_test3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl865x_tblAsicDrv_routingParam_t asic_rt;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: set/reset pptp property
		  *=================================================*/
		defaultConfig_pptp1();

		/* (0) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);

		/* Verify dvidx field when process of the routing entry is CPU. */
		ASSERT(rtl8651_getAsicRouting(1,  &asic_rt) == SUCCESS);
		ASSERT(asic_rt.vidx == rtl8651_vlanTableIndex(1));

		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);

		/* (2) add IP Interface */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (3) set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) != SUCCESS);

		/* (4) add arp & set pptp property */
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* (5) reset pptp property */
		ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | pptp_test4 | Testing PPTP.
@comm
After dial up a PPTP session, add the default route to the PPTP session. Make sure the route is correct.
Then hang up the PPTP session, the route information should change to incomplete state.
*/
int32 pptp_test4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp after dial pptp session
		  *	      delete default route
		  *=================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;

		/* (0) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);
		
		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);

		/* (2) add IP Interface */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (3) add arp & set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* (4) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 68;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80; //0x20;   1000 0000
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_nexthop.nextHopRow, asic_nexthop.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process 			= RT_DROP;
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		
		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;
}


/*
@func int32 | pptp_test5 | Testing PPTP.
@comm
Before dial up a PPTP session, add the default route to the PPTP session. Make sure the route is correct.
Then dial up the PPTP session, the route should update its information.
*/
int32 pptp_test5(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp before dial pptp session
		  *	      delete default route
		  *=================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);

		/* (3) add default route  to pptp */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);


		/* (4) add arp & set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) dial up pptp session */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 68;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		
		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_nexthop.nextHopRow, asic_nexthop.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (6) remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process 			= RT_DROP;
		asic_rt_1.process				= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


/*
@func int32 | pptp_test6 | Testing PPTP.
@comm
Add default route to a PPTP session. Test if it is correct when set/reset (dial/hang-up) the PPTP session.
*/
int32 pptp_test6(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp before dial pptp session
		  *	      set/reset pptp session 
		  *=================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;


		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);


		/* (3) add default route  to pptp */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Entry */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		

		/* (4) add arp & set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		
		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) dial up pptp session */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 68;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);


		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_nexthop.nextHopRow, asic_nexthop.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (6) rest pptp session */
		ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

	}
	return SUCCESS;
}


/*
@Func int32 | pptp_test7 | Testing PPTP.
@comm
When a PPTP session is added, a VLAN should be created stealthily and the MII port should be the VLAN's 
member port.
*/
int32 pptp_test7(uint32 caseNo)
{
	ipaddr_t ipaddr, ipMask;
	uint32 pvid;

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: When a route is set to PPTP, the route's dvid should be 
		  *	      modified to MII VID
		  *=================================================*/
		defaultConfig_pptp1();

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);


		/* (3) add default route  to pptp */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);
		ASSERT(DrvTbl.route[rtl8651_tblDrvPara.routingTableSize-1].dvid == 2048);

		/* Verify PVID of MII port */
		//ASSERT(rtl8651_getAsicPvid(RTL8651_MII_PORT, &pvid) == SUCCESS);
		ASSERT(rtl8651_getAsicPvid(0x07, &pvid) == SUCCESS); /* pptp/l2tp extension port is 7 */
		ASSERT(pvid == 0);  // 2048 % 8 = 0
	}
	return SUCCESS;
}


/*
@func int32 | pptp_test9 | Testing PPTP.
@comm
Test when fwd learns a ARP entry which is the PPTP server's ARP, the L2 entry's member port still be the MII port instead of
the physical port which the PPTP server belongs to.
*/
int32 pptp_test9(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	uint32 pvid;

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: Fwd learns and adds an ARP entry
		  *=================================================*/
		defaultConfig_pptp1();

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);

		/* (3) add IP Interface */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (4) add arp & set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80; /* change port to extension Port */
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (6) assume fwd learns the same arp entry */
		//Becaue the l2 informatin is reference by both arp and pptp property, if try to
		//modify the arp information, it will faill. ===> modify to success
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L2 Table: should be static and MBR=0x20 not 0x01 */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80; /* change port to MII Port */
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;

}


int32 pptp_test8(uint32 caseNo)
{
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	
	START_TEST_CASE();

	{
		/*=================================================
		  * Test: default route when pptp session is added
		  *=================================================*/
		defaultConfig_pptp1();

		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process 			= RT_DROP;
		asic_rt_1.process				= RT_CPU;
		asic_rt_1.vidx				= 1;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (2) remove PPTP Session */
		ASSERT(rtl8651_delSession(1000) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process 			= RT_DROP;
		asic_rt_1.process 			= RT_CPU;
		asic_rt_1.vidx				= 1;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
	}
	return SUCCESS;
}


int32 pptp_test14(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	
	START_TEST_CASE();

	{
		/*=================================================
		  * Test: default route when pptp session is added
		  *=================================================*/
		defaultConfig_pptp1();

		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);

		/* (1) add to original network interface */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != NULL);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);
		
		/* (2) remove arp */
		ASSERT(_rtl8651_delLocalArp(TRUE, ipaddr, netif_t) == SUCCESS);

		/* Verify that l2 entry should exists, because pptp property is set */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (3) reset pptp property */
		ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) != SUCCESS);
	}
	return SUCCESS;
}


int32 pptp_test15(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl8651_tblDrv_pppTable_t *pppInfo_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: set/reset pptp property
		  *=================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;


		/* (0) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);

		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);

		/* (2) add IP Interface */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (3) set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) != SUCCESS);

		/* (4) add arp & set pptp property */
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* (5) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 68;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_nexthop.nextHopRow, asic_nexthop.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
		
		/* (6) reset pptp property */
		ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);


//		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);


		/*===============================================================
		 * Change PPTP Server to 192.168.154.220 
		 *===============================================================*/
		/* (1) set pptp property */
		mac.octet[5] ++;
		ipaddr = ntohl(inet_addr("192.168.1.220"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 69;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_nexthop.nextHopRow, asic_nexthop.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
		
	}
	return SUCCESS;
}




int32 pptp_test16(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl8651_tblDrv_pppTable_t *pppInfo_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: set/reset pptp property
		  *=================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;


		/* (0) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);

		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);

		/* (2) add IP Interface */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (3) set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) != SUCCESS);

		/* (4) add arp & set pptp property */
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* (5) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 68;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_nexthop.nextHopRow, asic_nexthop.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
		
		/* (6) reset pptp property */
		ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);


		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/*===============================================================
		 * Change PPTP Server to 192.168.154.220 
		 *===============================================================*/
		/* (1) set pptp property */
		mac.octet[5] ++;
		ipaddr = ntohl(inet_addr("192.168.1.220"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		ASSERT(rtl8651_addRoute(0, 0, "pptp", 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 69;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_nexthop.nextHopRow, asic_nexthop.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
		
	}
	return SUCCESS;
}



int32 pptp_test17(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	int32 entry;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp before dial pptp session
		  *	      delete default route
		  *=================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);

		/* (3) add route */
		for(entry=0; entry<10; entry++) {
			ipaddr = ntohl(inet_addr("192.168.2.10")) + entry;
			ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "vlan", ipaddr) == SUCCESS);
		}
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addRoute(ipaddr, 0xffffffff, "pptp", 0) == SUCCESS);

		/* (4) add arp & set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pptp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) dial up pptp session */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("pptp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
		
		/* (6) reset pptp */
		ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;
}










#define RTL865X_L2TP_TESTING_CODE

static int32 defaultConfig_l2tp1(void)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask;
	extern int32 rtl8651_loopbackPortPHY; 

	/* VLAN 1: subnet 1 */	
	ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("l2tp") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("l2tp", RTL8651_LL_L2TP, 1) == SUCCESS);
	ASSERT(rtl8651_addExtNetworkInterface("l2tp") == SUCCESS);

	/* VLAN 2: subnet 2 */
	mac.octet[5] = 0x02;
	ASSERT(rtl8651_addVlan(2) == SUCCESS);
	ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
	ASSERT(rtl8651_addNetworkIntf("vlan") == SUCCESS);
	ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan", RTL8651_LL_VLAN, 2) == SUCCESS);
	ipaddr = ntohl(inet_addr("192.168.2.1"));
	mask = ntohl(inet_addr("255.255.255.0"));
	ASSERT(rtl8651_addIpIntf("vlan", ipaddr, mask) == SUCCESS);

	rtl8651_setLoopbackPortPHY(RTL8651_MII_PORT, TRUE);
	return SUCCESS;
}	


#if 0
/* OLD MODEL: without nexthop */



/*
@func int32 | l2tp_test4 | Testing L2TP.
@comm
After dial up a L2TP session, add the default route to the L2TP session. Make sure the route is correct.
Then hang up the L2TP session, the route information should change to incomplete state.
*/

int32 l2tp_test4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp after dial pptp session
		  *	      delete default route
		  *=================================================*/
		defaultConfig_l2tp1();

		/* (0) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);
		
		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);

		/* (2) add IP Interface */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (3) add arp & set l2tp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* (4) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "l2tp", 0) == SUCCESS);

 		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 0;
		asic_rt_1.nextHopRow		= 68;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_rt_1.nextHopRow, asic_rt_1.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
		
		/* (5) remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_DROP;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;
}



/*
@func int32 | l2tp_test5 | Testing L2TP.
@comm
Before dial up a L2TP session, add the default route to the L2TP session. Make sure the route is correct.
Then dial up the L2TP session, the route should update its information.
*/
int32 l2tp_test5(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp before dial pptp session
		  *	      delete default route
		  *=================================================*/
		defaultConfig_l2tp1();

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);


		/* (3) add default route  to l2tp */
		ASSERT(rtl8651_addRoute(0, 0, "l2tp", 0) == SUCCESS);

		/* Verify L3 Entry */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (4) add arp & set l2tp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) dial up l2tp session */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 0;
		asic_rt_1.nextHopRow		= 68;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);


		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_rt_1.nextHopRow, asic_rt_1.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (6) remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_DROP;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;
}





/*
@func int32 | l2tp_test6 | Testing L2TP.
@comm
Add default route to a L2TP session. Test if it is correct when set/reset (dial/hang-up) the L2TP session.
*/

int32 l2tp_test6(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp before dial pptp session
		  *	      set/reset pptp session 
		  *=================================================*/
		defaultConfig_l2tp1();

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);


		/* (3) add default route  to l2tp */
		ASSERT(rtl8651_addRoute(0, 0, "l2tp", 0) == SUCCESS);

		/* Verify L3 Entry */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* (4) add arp & set l2tp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) dial up l2tp session */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_L2;
		asic_rt_1.vidx				= 0;
		asic_rt_1.nextHopRow		= 68;
		asic_rt_1.nextHopColumn	= 0;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);


		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_rt_1.nextHopRow, asic_rt_1.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (6) rest l2tp session */
		ASSERT(rtl8651_resetL2tpProperty(1000) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
 
	}
	return SUCCESS;
}



#endif



/*
@func int32 | l2tp_test1 | Test L2TP.
@comm
Make sure adding and deleting L2TP session is workable.
*/

int32 l2tp_test1(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan_1, asic_vlan_2;
	int8 portState[6] = { RTL8651_PORTSTA_FORWARDING, RTL8651_PORTSTA_FORWARDING, RTL8651_PORTSTA_FORWARDING, 
			  	          RTL8651_PORTSTA_FORWARDING, RTL8651_PORTSTA_FORWARDING, RTL8651_PORTSTA_FORWARDING };
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	extern rtl8651_tblDrv_miiTunneling_t tunnel;
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	int32 curr;
	
	
	START_TEST_CASE();

	{
		/*===================================================
  		  * Test: add/delete a session for L2TP
  		  *===================================================*/
  		defaultConfig_l2tp1();

  		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);

		curr = 0;
		/* Verify Vlan Table for Virtual VLAN */

		// Louis notes: I have no idea about what is it.
		curr++;

		bzero(&asic_vlan_1, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		bzero(&asic_vlan_1, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		asic_vlan_1.broadcastToCpu		= 1;
		asic_vlan_1.enableRoute			= 1;
		asic_vlan_1.inAclEnd			= 1;
		asic_vlan_1.inAclStart			= 0;
		asic_vlan_1.internal			= 0;			
		asic_vlan_1.macAddrNumber		= 1;
		asic_vlan_1.memberPortMask		= 0x80;
		asic_vlan_1.mtu					= 1500;
		asic_vlan_1.outAclEnd			= RTL8651_ACLTBL_PERMIT_ALL;
		asic_vlan_1.outAclStart			= RTL8651_ACLTBL_PERMIT_ALL;		
		asic_vlan_1.promiscuous			= 0;
		asic_vlan_1.untagPortMask		= 0x80;
		asic_vlan_1.valid				= 1;
		asic_vlan_1.vid					= 2048;
		rtl8651_memcpy(&asic_vlan_1.macAddr, &mac, sizeof(ether_addr_t));
		rtl8651_memcpy(asic_vlan_1.portState, portState, 6);
		ASSERT(rtl8651_getAsicVlan(2048,  &asic_vlan_2) == SUCCESS);
		VLAN_ENTRY_CMP(&asic_vlan_1, &asic_vlan_2);

		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_	= RTL8651_ACL_MAC;
		rule1.actionType_	= RTL8651_ACL_CPU;
		rule1.srcMac_		= mac;
		memset(&rule1.srcMacMask_, 0xff, 6);
		ASSERT(rtl8651_getAsicAclRule(asic_vlan_1.inAclStart, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		ACL_DEFAULT_PERMIT(asic_vlan_1.inAclEnd);
		ASSERT((asic_vlan_1.inAclStart+curr) == asic_vlan_1.inAclEnd);


		curr = 0;
		/* Verify ACL: vlan 1 */
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE ) curr++;

		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_	= RTL8651_ACL_CPU;
		rule1.dstMac_		= mac;
		memset(&rule1.dstMacMask_, 0xff, 6);
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		ACL_DEFAULT_PERMIT(asic_vlan.inAclEnd);
		
		curr = 0;
	 	//curr++; // Trap packets for PPTP/L2TP
	 	curr += 3; /* for tcp syn, fin, rst */
		/* Verify ACL: vlan 2 */
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE ) curr++;
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
//		ACL_DEFAULT_PERMIT(asic_vlan.inAclEnd);
//		ASSERT(asic_vlan.inAclStart == asic_vlan.inAclEnd);
		
		ACL_DEFAULT_PERMIT(asic_vlan.inAclStart+curr);
		ASSERT(asic_vlan.outAclStart == asic_vlan.outAclEnd);
		ASSERT(tunnel.valid == 1);
		
	
		ASSERT(rtl8651_delSession(1000) == SUCCESS);
		ASSERT(tunnel.valid == 0);

		/* Verify */
		ASSERT(rtl8651_getAsicVlan(2048,  &asic_vlan_2) != SUCCESS);

		/* (2) create another new network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp1", 1) == SUCCESS);
		ASSERT(rtl8651_delSession(1000) == SUCCESS);
		ASSERT(GET_NETWORK_INTERFACE("l2tp1", NULL) == SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | l2tp_test2 | Test L2TP.
@comm
Make sure IP interface (subnet) can be added to and removed from a L2TP session.
*/
int32 l2tp_test2(uint32 caseNo)
{
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl8651_tblDrvSession_t *session_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add/delete l2tp session with IP address
		  *=================================================*/
		defaultConfig_l2tp1();

		/* (1) add to original network interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);
		ASSERT(rtl8651_delSession(1000) == SUCCESS);
		
		/* (1.1) add broadcast netowrk to L2TP */
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);

		/* Verify L3 */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= ntohl(inet_addr("192.168.2.0"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.0"));
		asic_rt_1.process 			= RT_ARP;
		asic_rt_1.vidx				= 2;
		asic_rt_1.arpStart			= 0;
		asic_rt_1.arpEnd			= 31;
		ASSERT(rtl8651_getAsicRouting(0,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		asic_rt_1.ipAddr			= ntohl(inet_addr("192.168.1.0"));
		asic_rt_1.ipMask			= ntohl(inet_addr("255.255.255.0"));
 		/*
		  *   When the linkLayerType is L2TP,
		  *   the PROCESS of the interface routing entry is always to CPU.
		  *   This is for "l2tp dial demand (when pinging  l2tp server IP)" spec.
		  */

		//asic_rt_1.process 			= RT_ARP;
		asic_rt_1.process 			= RT_CPU;
#if 0
		asic_rt_1.vidx				= 1;
		asic_rt_1.arpStart			= 32;
		asic_rt_1.arpEnd			= 63;
#endif
		ASSERT(rtl8651_getAsicRouting(1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process 			= RT_DROP;
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		ASSERT(rtl8651_getAsicRouting(2,  &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(3,  &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(4,  &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(5,  &asic_rt_2) != SUCCESS);
		ASSERT(rtl8651_getAsicRouting(6,  &asic_rt_2) != SUCCESS);

		/* (1.2) add tunneling network  */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		/* Verify */
		ASSERT((netif_t = GET_NETWORK_INTERFACE("l2tp", NULL)) != NULL);
		ASSERT((session_t=CTAILQ_FIRST(&netif_t->sessionHead)) != SUCCESS);
		ASSERT(session_t->ip_t);
		ASSERT(session_t->ip_t->ipAddr == ipaddr);
		ASSERT(session_t->ip_t->nextIpAddr.sle_next == NULL);
	

		/* (2) add to new network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1001, "l2tp1", 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("l2tp1", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1001) == SUCCESS);
		/* Verify */
		ASSERT((netif_t = GET_NETWORK_INTERFACE("l2tp1", NULL)) != NULL);
		ASSERT((session_t=CTAILQ_FIRST(&netif_t->sessionHead)) != SUCCESS);
		ASSERT(session_t->ip_t);
		ASSERT(session_t->ip_t->ipAddr == ipaddr);
		ASSERT(session_t->ip_t->nextIpAddr.sle_next == NULL);
	
		/* (3) Remove PPTP  */
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_delIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_delIpIntf("l2tp1", ipaddr, 0xffffffff) == SUCCESS);

		/* Verify */
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp1", NULL)) != NULL);
		ASSERT((session_t=CTAILQ_FIRST(&netif_t->sessionHead)) != SUCCESS);
		ASSERT(session_t->ip_t == NULL);
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp", NULL)) != NULL);
		ASSERT((session_t=CTAILQ_FIRST(&netif_t->sessionHead)) != SUCCESS);
		ASSERT(session_t->ip_t == NULL);

		/* (4) remove Session */
		ASSERT(rtl8651_delSession(1000) == SUCCESS);
		ASSERT(rtl8651_delSession(1001) == SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | l2tp_test3 | Test L2TP.
@comm
Test if rtl8651_setPptpProperty() and rtl8651_resetL2tpProperty() are workable.
*/
int32 l2tp_test3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: set/reset l2tp property
		  *=================================================*/
		defaultConfig_l2tp1();

		/* (0) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);

		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);

		/* (2) add IP Interface */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (3) set l2tp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 22222, ipaddr, 0) != SUCCESS);

		/* (4) add arp & set l2tp property */
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* (5) reset l2tp property */
		ASSERT(rtl8651_resetL2tpProperty(1000) == SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | l2tp_test4 | Testing L2TP.
@comm
After dial up a L2TP session, add the default route to the L2TP session. Make sure the route is correct.
Then hang up the L2TP session, the route information should change to incomplete state.
*/

int32 l2tp_test4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp after dial pptp session
		  *	      delete default route
		  *=================================================*/
		defaultConfig_l2tp1();
		DrvTbl.vlan[0].valid = 1;


		/* (0) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);
		
		/* (1) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);

		/* (2) add IP Interface */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (3) add arp & set l2tp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* (4) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "l2tp", 0) == SUCCESS);

 		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 68;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);


		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_nexthop.nextHopRow, asic_nexthop.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process 			= RT_DROP;
		asic_rt_1.process				= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

	}
	return SUCCESS;
}


/*
@func int32 | l2tp_test5 | Testing L2TP.
@comm
Before dial up a L2TP session, add the default route to the L2TP session. Make sure the route is correct.
Then dial up the L2TP session, the route should update its information.
*/
int32 l2tp_test5(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp before dial pptp session
		  *	      delete default route
		  *=================================================*/
		defaultConfig_l2tp1();
		DrvTbl.vlan[0].valid = 1;


		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);


		/* (3) add default route  to l2tp */
		ASSERT(rtl8651_addRoute(0, 0, "l2tp", 0) == SUCCESS);

		/* Verify L3 Entry */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (4) add arp & set l2tp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);


		/* (5) dial up l2tp session */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 68;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);


		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_nexthop.nextHopRow, asic_nexthop.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (6) remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		/* 2005.7 => DROP has been modified to CPU. */
		//asic_rt_1.process 			= RT_DROP;
		asic_rt_1.process 			= RT_CPU;
		ASSERT(rtl8651_getAsicRouting(7,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}



/*
@func int32 | l2tp_test6 | Testing L2TP.
@comm
Add default route to a L2TP session. Test if it is correct when set/reset (dial/hang-up) the L2TP session.
*/

int32 l2tp_test6(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: add default route to pptp before dial pptp session
		  *	      set/reset pptp session 
		  *=================================================*/
		defaultConfig_l2tp1();
		DrvTbl.vlan[0].valid = 1;

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);


		/* (3) add default route  to l2tp */
		ASSERT(rtl8651_addRoute(0, 0, "l2tp", 0) == SUCCESS);

		/* Verify L3 Entry */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		
		/* (4) add arp & set l2tp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) dial up l2tp session */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 68;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(asic_nexthop.nextHopRow, asic_nexthop.nextHopColumn, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (6) rest l2tp session */
		ASSERT(rtl8651_resetL2tpProperty(1000) == SUCCESS);

		/* Verify L3 Table */
		bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
		asic_rt_1.ipAddr			= 0;
		asic_rt_1.ipMask			= 0;
		asic_rt_1.process 			= RT_NEXTHOP;
		asic_rt_1.nhStart			= 2;
		asic_rt_1.nhNum				= 2;
		asic_rt_1.nhNxt				= 2;
		asic_rt_1.nhAlgo			= 2;
		asic_rt_1.ipDomain			= 6;
		ASSERT(rtl8651_getAsicRouting(RTL8651_ROUTINGTBL_SIZE-1,  &asic_rt_2) == SUCCESS);
		L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
		
		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* Verify Nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

 
	}
	return SUCCESS;
}


/*
@Func int32 | l2tp_test7 | Testing L2TP.
@comm
When a L2TP session is added, a VLAN should be created stealthily and the MII port should be the VLAN's 
member port.
*/

int32 l2tp_test7(uint32 caseNo)
{
	ipaddr_t ipaddr, ipMask;
	uint32 pvid;

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: When a route is set to L2TP, the route's dvid should be 
		  *	      modified to MII VID
		  *=================================================*/
		defaultConfig_l2tp1();

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);


		/* (3) add default route  to pptp */
		ASSERT(rtl8651_addRoute(0, 0, "l2tp", 0) == SUCCESS);
		ASSERT(DrvTbl.route[rtl8651_tblDrvPara.routingTableSize-1].dvid == 2048);

		/* Verify PVID of MII port */
		ASSERT(rtl8651_getAsicPvid(0x07, &pvid) == SUCCESS); /* l2tp/pptp extension port is 7 */
		ASSERT(pvid == 0);  // 2048 % 8 = 0
	}
	return SUCCESS;
}


/*
@func int32 | l2tp_test9 | Testing L2TP.
@comm
Test when fwd learns a ARP entry which is the L2TP server's ARP, the L2 entry's member port still be the MII port instead of
the physical port which the L2TP server belongs to.
*/

int32 l2tp_test9(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	uint32 pvid;

	START_TEST_CASE();

	{
		/*=================================================
		  * Test: Fwd learns and adds an ARP entry
		  *=================================================*/
		defaultConfig_l2tp1();

		/* (1) add IP interface */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);
		
		/* (2) add to original network interface */
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);

		/* (3) add IP Interface */
		ipaddr = ntohl(inet_addr("140.112.1.2"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(ipaddr, 1000) == SUCCESS);

		/* (4) add arp & set pptp property */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp", NULL)) != SUCCESS);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (5) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "l2tp", 0) == SUCCESS);

		/* Verify L2 Table */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80; /* change port to MII Port */
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (6) assume fwd learns the same arp entry */
		//Becaue the l2 informatin is reference by both arp and l2tp property, if try to
		//modify the arp information, it will faill. ==> change to SUCCESS
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);

		/* Verify L2 Table: should be static and MBR=0x20 not 0x01 */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 1;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x80; /* change port to MII Port */
		asic_l2_1.nhFlag			= 0;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);
	}
	return SUCCESS;

}



int32 l2tp_test14(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0x11, 0x22, 0x33, 0x44 } };
	rtl865x_tblAsicDrv_l2Param_t asic_l2_1, asic_l2_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	ipaddr_t ipaddr, ipMask;
	
	START_TEST_CASE();

	{
		/*=================================================
		  * Test: default route when l2tp session is added
		  *=================================================*/
		defaultConfig_l2tp1();

		ipaddr = ntohl(inet_addr("192.168.1.1"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", ipaddr, ipMask) == SUCCESS);

		/* (1) add to original network interface */
		ipaddr = ntohl(inet_addr("192.168.1.253"));
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);
		ASSERT((netif_t=GET_NETWORK_INTERFACE("l2tp", NULL)) != NULL);
		ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 22222, ipaddr, 0) == SUCCESS);
		
		/* (2) remove arp */
		ASSERT(_rtl8651_delLocalArp(TRUE, ipaddr, netif_t) == SUCCESS);

		/* Verify that l2 entry should exists, because pptp property is set */
		bzero(&asic_l2_1, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		bzero(&asic_l2_2, sizeof(rtl865x_tblAsicDrv_l2Param_t));
		asic_l2_1.ageSec			= 100;
		asic_l2_1.cpu				= 0;
		asic_l2_1.isStatic			= 0;
		asic_l2_1.macAddr			= mac;
		asic_l2_1.memberPortMask	= 0x02;
		asic_l2_1.nhFlag			= 1;
		asic_l2_1.srcBlk			= 0;
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) == SUCCESS);
		L2_ENTRY_CMP(&asic_l2_1, &asic_l2_2);

		/* (3) reset l2tp property */
		ASSERT(rtl8651_resetL2tpProperty(1000) == SUCCESS);
		ASSERT(rtl8651_getAsicL2Table(68, 0, &asic_l2_2) != SUCCESS);
	}
	return SUCCESS;
}


























#define RTL865X_IP_UNNUMBERED_TESTING_CODE



#define RTL865X_ACL_TESTING_CODE


#define RTL865X_VLAN_TESTING_CODE


#define RTL865XB_DCUT_SWVLAN_TESTING_CODE
#ifdef RTL865XB_DCUT_SWVLAN
int32 swVlan_test1(uint32 caseNo)
{
	uint32 vid;

	/* Testing rtl8651_addSwVlan() / rtl8651_delSwVlan() API. */
	START_TEST_CASE();
	{
		/* Add an illegal vid to software VLAN table. */
		ASSERT(rtl8651_addSwVlan(0) == TBLDRV_EINVALIDVLANID);
		ASSERT(rtl8651_addSwVlan(4095) == TBLDRV_EINVALIDVLANID);

		/* Add an default vid = 1 (in ASIC) to software VLAN table. */
		ASSERT(rtl8651_addSwVlan(1) == TBLDRV_EVLANALREADYEXISTS);

		/* Add an already-existing vid (in ASIC) to software VLAN table. */
		ASSERT(rtl8651_addVlan(2) == SUCCESS);
		ASSERT(rtl8651_addSwVlan(2) == TBLDRV_EVLANALREADYEXISTS);
		ASSERT(rtl8651_delVlan(2) == SUCCESS);

		/* Add vid = 2 ~ 4094 to software VLAN table. */
		for (vid = 2; vid < 4095; vid++)
			ASSERT(rtl8651_addSwVlan(vid) == SUCCESS);

		/* Add an already-existing vid (in software VLAN) to ASIC VLAN table. */
		for (vid = 2; vid < 4095; vid++)
			ASSERT(rtl8651_addVlan(vid) == TBLDRV_EVLANALREADYEXISTS);

		/* Delete an illegal vid from software VLAN table. */
		ASSERT(rtl8651_delSwVlan(0) == TBLDRV_EINVALIDVLANID);
		ASSERT(rtl8651_delSwVlan(4095) == TBLDRV_EINVALIDVLANID);

		/* Delete an default vid = 1 (in ASIC) from software VLAN table. */
		ASSERT(rtl8651_delSwVlan(1) == TBLDRV_EINVALIDVLANID);
		
		/* Delete vid = 2 ~ 4094 from software VLAN table. */
		for (vid = 2; vid < 4095; vid++)
			ASSERT(rtl8651_delSwVlan(vid) == SUCCESS);

		return SUCCESS;
	}
}

int32 swVlan_test2(uint32 caseNo)
{
	uint32 vid, port;
	
	START_TEST_CASE();
	{
		printf("rtl8651_totalExtPortNum: %d\n", rtl8651_totalExtPortNum);

		for (vid = 2; vid < 4095; vid++) {
			ASSERT(rtl8651_addSwVlan(vid) == SUCCESS);
			/* Add each port to VLAN */
			for (port = 0; port < RTL8651_PORT_NUMBER + rtl8651_totalExtPortNum; port++)
				ASSERT(rtl8651_addVlanPortMember(vid, port) == SUCCESS);
		}			

		/* Link up */
		for (port = 0; port < RTL8651_PORT_NUMBER + rtl8651_totalExtPortNum; port++)
			rtl8651_setEthernetPortLinkStatus(port, TRUE);

		for (vid = 2; vid < 4095; vid++) {
			/* Delete each port to VLAN */
			for (port = 0; port < RTL8651_PORT_NUMBER + rtl8651_totalExtPortNum; port++)
				ASSERT(rtl8651_delVlanPortMember(vid, port) == SUCCESS);
			ASSERT(rtl8651_delSwVlan(vid) == SUCCESS);
		}
		
		return SUCCESS;
	}
}
#endif
#define RTL865X_ETHERNET_PORT_TETING_CODE


#define RTL865X_TRUNKING_TESTING_CODE


#define RTL865X_ETHERNET_PORT_ATTRIBUTE_TESTING_CODE


#define RTL865X_MIRROR_PORT_TESTING_CODE



#define RTL865X_SPANNING_TREE_TESTING_CODE



#define RTL865X_LARGE_IP_CLASS

struct array_s {
	ipaddr_t dip;
	uint32 load;
	rtl8651_tblDrv_arpEntry_t *arp_t;
} ;

int32 large_ipClass_getMax(struct array_s *array_t, uint32 entry)
{
	uint32 index, max=0;

	for(index=0; index<entry; index++) {
		if (array_t[max].load < array_t[index].load && array_t[index].load>IP_CLASS_THRESHOLD)
			max = index;
	}
	if (array_t[max].load > IP_CLASS_THRESHOLD) {
		array_t[max].load = 0;
		return max;
	}
	else return -1;
}


/*
@func int32 | large_ipClass_1 | Test Class A/B IP subnet.
@comm
Test add a connection to Class A/B monitoring.
*/

int32 large_ipClass_1(uint32 caseNo)
{
	ipaddr_t ipaddr;
	uint32 round, load, i;

	START_TEST_CASE();

	{
		/*===================================================
		  * Test: _rtl8651_ipClassMonitoring()
		  *===================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		ipaddr = ntohl(inet_addr("192.168.1.2"));
		for(round=0; round<20; round++) {
			load = 200 + (uint32)(2000.0 * rand() / (RAND_MAX + 1.0));
			if (round < IP_CLASS_MAX_ENTRY) {
				ASSERT(_rtl8651_ipClassMonitoring(ipaddr, 0x123456) == SUCCESS);
			}
			else {
				ASSERT(_rtl8651_ipClassMonitoring(ipaddr, 0x123456) != SUCCESS);
			}
			ipaddr ++;
		}
	} 
	return SUCCESS;
}



/*
@func int32 | large_ipClass_2 | Test Class A/B IP subnet.
@comm
If the packet count > threshold, the connection should be written into ASIC L3 Table as a host route.
*/

int32 large_ipClass_2(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x00 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl8651_tblDrv_arpEntry_t *arp_t;
	ipaddr_t ipaddr, array[8] = { 0 };
	uint32 index, entry, i, size;

	
	START_TEST_CASE();

	{
		/*===================================================
		  * Test: One connection is under monitoring per time
		  *===================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		array[0] = array[1] = array[7] = 0xffffffff;
		netif_t = GET_NETWORK_INTERFACE("vlan1", NULL) ;
		ASSERT(netif_t);

		ipaddr = ntohl(inet_addr("192.168.1.2"));
		for(index=0; index<250; index++) {
			ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
			mac.octet[5] ++;
			ipaddr ++;
		}

		ipaddr = ntohl(inet_addr("192.168.1.2"));
		for(index=0, entry=6; index<250; index++) {
			arp_t = _rtl8651_getArpEntry(netif_t, ipaddr, FALSE);
			for(i=0; i<1000; i++)
				ASSERT(_rtl8651_ipClassMonitoring(ipaddr, arp_t) == SUCCESS);
			_rtl8651_ipClassTimeUpdate();

			size = 0;
			while (array[entry] == 0xffffffff && size <7) {
				if (--entry == 0)
					entry = 6;
				size ++;
			}
			array[entry] = ipaddr;
			if (--entry == 0)
				entry = 6;

			/* Verify L3 Table */
			for(i=0; i<RTL8651_ROUTINGTBL_SIZE; i++) {
				if (array[i] == 0xffffffff || array[i] == 0)
					continue;
				arp_t = _rtl8651_getArpEntry(netif_t, array[i], FALSE);
				ASSERT(arp_t);
				bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
				bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
				asic_rt_1.ipAddr			= array[i];
				asic_rt_1.ipMask			= 0xffffffff;
				asic_rt_1.process			= RT_L2;
				asic_rt_1.vidx				= 1;
				asic_rt_1.nextHopRow		= rtl8651_filterDbIndex(&arp_t->macInfo->macAddr);
				asic_rt_1.nextHopColumn	= arp_t->macInfo->asicPos;
				ASSERT(rtl8651_getAsicRouting(i, &asic_rt_2) == SUCCESS);
				L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
			}
			ipaddr ++;		
		}
	}
	return SUCCESS;
}




/*
@func int32 | large_ipClass_3 | Test Class A/B IP subnet.
@comm
Test more than one connection are under Class A/B monitoring.  And  select some connection too write to ASC L3 Table,
per second.
*/

int32 large_ipClass_3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x00 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	struct asicTable_s {
		ipaddr_t ipaddr;
		rtl8651_tblDrv_arpEntry_t *arp_t;
	} asicTable[8];
	struct array_s array[IP_CLASS_MAX_ENTRY] ;
	ipaddr_t ipaddr;
	int32 round, maxConn, i, conn, curIndex=0, maxIdx, kk;

	
	START_TEST_CASE();

	{
		memset(asicTable, 0, sizeof(struct asicTable_s) * 8);
		memset(array, 0, sizeof(struct array_s) * IP_CLASS_MAX_ENTRY);
		/*===================================================
		  * Test: more than one connection are under monitoring per time
		  *===================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		bzero(asicTable, sizeof(struct asicTable_s)*8);

		/* add arp */
		ipaddr = ntohl(inet_addr("192.168.1.2"));
		netif_t = GET_NETWORK_INTERFACE("vlan1", NULL);
		ASSERT(netif_t);
		for(round=0; round<IP_CLASS_MAX_ENTRY; round++) {
			mac.octet[5] ++;
			ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
			array[round].dip = ipaddr;
			array[round].arp_t = _rtl8651_getArpEntry(netif_t, array[round].dip, FALSE);
			ASSERT(array[round].arp_t);
			ipaddr ++;

			#if 0
			printf("array[%d].dip: %u.%u.%u.%u, array[%d].load: %d, array[%d].arp: %x.%x.%x.%x.%x.%x\n",
				round, NIPQUAD(array[round].dip), round, array[round].load, round, 
				(array[round].arp_t)->macInfo->macAddr.octet[0], (array[round].arp_t)->macInfo->macAddr.octet[1],
				(array[round].arp_t)->macInfo->macAddr.octet[2], (array[round].arp_t)->macInfo->macAddr.octet[3],
				(array[round].arp_t)->macInfo->macAddr.octet[4], (array[round].arp_t)->macInfo->macAddr.octet[5]);
			#endif
		}

		for(round=0; round<10000; round++) {
			/* select max session */
			maxConn = 1 + (uint32)(2000.0 * rand() / (RAND_MAX + 1.0));
			maxConn = (maxConn % IP_CLASS_MAX_ENTRY)+1;
			for(kk=0; kk<10000; kk++) {
				conn = (uint32)(maxConn * rand() / (RAND_MAX + 1.0));
				array[conn].load ++;
				ASSERT(_rtl8651_ipClassMonitoring(array[conn].dip, array[conn].arp_t) == SUCCESS);
				if (array[conn].load == (IP_CLASS_THRESHOLD+1)) {
					for(i=0; i<5&&(maxIdx=large_ipClass_getMax(&array, maxConn))!=-1; i++) {
						asicTable[6-curIndex].ipaddr = array[maxIdx].dip;
						asicTable[6-curIndex].arp_t = array[maxIdx].arp_t;
						curIndex = (curIndex+1)%5;
					}

				}
				
			}
	
			/* Verify L3 Table */
			for(i=6; i>1; i--) {
				if (asicTable[i].arp_t == NULL)
					continue;
				bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
				bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
				asic_rt_1.ipAddr			= asicTable[i].ipaddr;
				asic_rt_1.ipMask			= 0xffffffff;
				asic_rt_1.process			= RT_L2;
				asic_rt_1.vidx				= 1;
				asic_rt_1.nextHopRow		= rtl8651_filterDbIndex(&(asicTable[i].arp_t->macInfo->macAddr));
				asic_rt_1.nextHopColumn	= asicTable[i].arp_t->macInfo->asicPos;
				ASSERT(rtl8651_getAsicRouting(i, &asic_rt_2) == SUCCESS);
				L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);
			}
		}

	}
	return SUCCESS;
}


#if 0
int32 large_ipClass_3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x00 } };
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	struct asicTable_s {
		ipaddr_t ipaddr;
		rtl8651_tblDrv_arpEntry_t *arp_t;
	} asicTable[8];
	struct array_s array[IP_CLASS_MAX_ENTRY] ;
	ipaddr_t ipaddr;
	int32 round, maxConn, i, conn, curIndex=0, maxIdx;

	
	START_TEST_CASE();

	{
		/*===================================================
		  * Test: more than one connection are under monitoring per time
		  *===================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		bzero(asicTable, sizeof(struct asicTable_s)*8);

		/* add arp */
		ipaddr = ntohl(inet_addr("192.168.1.2"));
		netif_t = GET_NETWORK_INTERFACE("vlan1", NULL);
		ASSERT(netif_t);
		for(round=0; round<IP_CLASS_MAX_ENTRY; round++) {
			mac.octet[5] ++;
			ASSERT(_rtl8651_addLocalArp(TRUE, ipaddr, &mac, netif_t, 0x01) == SUCCESS);
			array[round].dip = ipaddr;
			array[round].arp_t = _rtl8651_getArpEntry(netif_t, array[round].dip, FALSE);
			ASSERT(array[round].arp_t);
			ipaddr ++;
		}

		for(round=0; round<50000; round++) {
			/* select max session */
			maxConn = 1 + (uint32)(2000.0 * rand() / (RAND_MAX + 1.0));
			maxConn = (maxConn % IP_CLASS_MAX_ENTRY)+1;
			for(conn=0; conn<maxConn; conn++) {
				array[conn].load = 200 + (uint32)(2000.0 * rand() / (RAND_MAX + 1.0));
				for(i=0; i<array[conn].load; i++) {
					ASSERT(_rtl8651_ipClassMonitoring(array[conn].dip, array[conn].arp_t) == SUCCESS);
				}
			}
			_rtl8651_ipClassTimeUpdate();

			for(i=0; i<5&&(maxIdx=large_ipClass_getMax(&array, maxConn))!=-1; i++) {
				asicTable[6-curIndex].ipaddr = array[maxIdx].dip;
				asicTable[6-curIndex].arp_t = array[maxIdx].arp_t;
				curIndex = (curIndex+1)%5;
			}

			/* Verify L3 Table */
			for(i=6; i>1; i--) {
				if (asicTable[i].arp_t == NULL)
					continue;
				bzero(&asic_rt_1, sizeof(rtl865x_tblAsicDrv_routingParam_t));
				bzero(&asic_rt_2, sizeof(rtl865x_tblAsicDrv_routingParam_t));
				asic_rt_1.ipAddr			= asicTable[i].ipaddr;
				asic_rt_1.ipMask			= 0xffffffff;
				asic_rt_1.process			= RT_L2;
				asic_rt_1.vidx				= 1;
				asic_rt_1.nextHopRow		= rtl8651_filterDbIndex(&(asicTable[i].arp_t->macInfo->macAddr));
				asic_rt_1.nextHopColumn	= asicTable[i].arp_t->macInfo->asicPos;
				ASSERT(rtl8651_getAsicRouting(i, &asic_rt_2) == SUCCESS);
				L3_ENTRY_CMP(&asic_rt_1, &asic_rt_2);

			}
		}

	}
	return SUCCESS;
}
#endif




#define RTL865X_QOS_TESTING_CODE


/*
@func int32 | qos_test1 | QoS Testing.
@comm
Test flow control configuration.
*/
int32 qos_test1(uint32 caseNo)
{
	uint32 ports[ ] = { RTL8651_PORT_0, RTL8651_PORT_1, RTL8651_PORT_2, RTL8651_PORT_3,
				      RTL8651_PORT_4, RTL8651_MII_PORT, RTL8651_CPU_PORT };
	int8 result[8] = { TRUE, TRUE, TRUE, TRUE, TRUE, TRUE };
	uint32 ran_value, round, port, reg, i;	
	int8 enable;
	

	START_TEST_CASE();
	{
		/*================================================
		  * Test: test API: rtl8651_setFlowControl()
		  *================================================*/
		for(i=0; i<5; i++) {
			(*((uint32 *)(PHY_BASE + (i<<5) + 0x10))) |= CAPABLE_PAUSE|CAPABLE_100BASE_TX_FD; /* Reg4 */
			(*((uint32 *)(PHY_BASE + (i<<5) + 0x14))) |= CAPABLE_PAUSE|CAPABLE_100BASE_TX_FD; /* Reg5 */
			(*((uint32 *)(PHY_BASE + (i<<5) + 0x00))) |= (1<<12); /* Reg0 n.12 */
			{
				extern uint32 linkChangePendingCount;
				linkChangePendingCount = 1;
				READ_MEM32(PHY_BASE+(i<<5) + 0x4) |= (1<<2);
				rtl8651_updateLinkStatus();
			}
		}

		for(round=0; round<500000; round++) {
			port = 1 + (uint32)(2000.0 * rand() / (RAND_MAX + 1.0));
			port = port % 5; 
			enable = (port % 2)? TRUE: FALSE;
			ASSERT(rtl8651_setFlowControl(ports[port], enable) == SUCCESS);
			result[ports[port]] = enable;
			/* Verify */
			ASSERT(rtl8651_getAsicFlowControlRegister(&reg) == SUCCESS);
			{
				extern uint32 linkChangePendingCount;
				linkChangePendingCount = 1;
				rtl8651_updateLinkStatus();
			}
	
			for(i=0; i<5; i++) {
				if (result[i] == TRUE) {
					ASSERT((reg & (1<<(18+i))) != 0);
					ASSERT((reg & (1<<(25+i))) != 0);
				}
				else {
					ASSERT((reg & (1<<(18+i))) == 0);
					ASSERT((reg & (1<<(25+i))) == 0);
				}
			}
		}
	}
	return SUCCESS;
}


/*
@func in32 | qos_test2 | Test QoS.
@comm
Test Queue Scheduling API.
*/
int32 qos_test2(uint32 caseNo)
{
	uint32 ratio[4] = { RTL8651_RR_H4_L1, RTL8651_RR_H8_L1, RTL8651_RR_H16_L1, RTL8651_RR_H_FIRST };
	uint32 round, sel, reg;
		

	START_TEST_CASE();
	{
		/*================================================
		  * Test: test API: rtl8651_setQueueScheduling()
		  *================================================*/
		ASSERT(rtl8651_setQueueScheduling(0xff, 0) != SUCCESS);
		for(round=0; round<500000; round++) {
			sel = 1 + (uint32)(2000.0 * rand() / (RAND_MAX + 1.0));
			sel = sel %  4;
			ASSERT(rtl8651_setQueueScheduling(RTL8651_QOS_RR, sel) == SUCCESS);

			/* Verify */
			ASSERT(rtl8651_getAsicQoSControlRegister(&reg) == SUCCESS);
			reg = reg >> 30;
			ASSERT(reg == ratio[sel]);
		}
	}
	return SUCCESS;
}


/*
@func int32 | qos_test3 | Test QoS.
@comm
Test Port based QoS configuration.
*/

int32 qos_test3(uint32 caseNo)
{
	uint32 ports[6] = { FALSE, FALSE, FALSE, FALSE, FALSE, FALSE };
	uint32 port, ishigh, round, i, reg;


	START_TEST_CASE();	
	{
		/*================================================
		  * Test: test API: rtl8651_setPortPriorityQueue()
		  *================================================*/
		for(round=0; round>500000; round++)
		{
			port = 1 + (uint32)(2000.0 * rand() / (RAND_MAX + 1.0));
			port = port % 6;
			ishigh = ((1 + (uint32)(2000.0 * rand() / (RAND_MAX + 1.0)))%2)? TRUE: FALSE;
			ASSERT(rtl8651_setPortPriorityQueue(port, ishigh) == SUCCESS);
			ports[port] = ishigh;

			/* Verify */
			ASSERT(rtl8651_getAsicQoSControlRegister(&reg) == SUCCESS);
			for(i=0; i<6; i++) {
				if (ports[i] == TRUE) {
					ASSERT((reg & (1<< (i+21))) != 0);
				}
				else {
					ASSERT((reg & (1 << (i+21))) == 0);
				}
			}
		}
	}
	return SUCCESS;
}


/*
@func int32 | qos_test4 | Test QoS.
@comm
Test Diff-Serv QoS configuration.
*/

int32 qos_test4(uint32 caseNo)
{
	int8 array[64] = { 0 }, ishigh;
	uint32 round, dscp, reg0, reg1, i;

	START_TEST_CASE();	
	{
		/*================================================
		  * Test: test API: rtl8651_setDiffServDSCP()
		  *================================================*/
		for(round=0; round<500000; round++) {
			dscp = 1 + (uint32)(2000.0 * rand() / (RAND_MAX + 1.0));
			ishigh = (dscp % 2)? TRUE: FALSE;
			dscp = dscp % 64;
			ASSERT(rtl8651_setDiffServDSCP(dscp, ishigh) == SUCCESS);
			array[dscp] = ishigh;

			/* Verify */
			ASSERT(rtl8651_getAsicDiffServReg(&reg0, &reg1) == SUCCESS);
			for(i=0; i<32; i++) {
				if (array[i] == TRUE) {
					ASSERT((reg0 & (1 << i)) != 0);
				}
				else {
					ASSERT((reg0 & (1 << i)) == 0);
				}
			}
			for(i=0; i<32; i++) {
				if (array[i+32] == TRUE) {
					ASSERT((reg1 & (1 << i)) != 0);
				}
				else {
					ASSERT((reg1 & (1 << i)) == 0);
				}
			}
			
		}
	}
	return SUCCESS;
}


/* 
@func int32 | qos_test5 | Test QoS.
@comm
Test policy-based QoS configuration.
*/

int32 qos_test5(uint32 caseNo)
{
	rtl8651_tblDrv_vlanTable_t vlan_t;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl8651_tblDrvPolicy_t policy;
	int32 curr;

	START_TEST_CASE();	
	{
		/*================================================
		  * Test: add policy-based Qos
		  *================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		
		bzero(&policy, sizeof(rtl8651_tblDrvPolicy_t));
		policy.proto		= RTL8651_POLICY_TCP;
		policy.dip     		= ntohl(inet_addr("140.112.1.0")) ;
		policy.dipMask		= ntohl(inet_addr("255.255.255.0"));
		policy.sip			= ntohl(inet_addr("192.168.1.0")) ;
		policy.sipMask		= ntohl(inet_addr("255.255.255.0"));
		policy.sport_start	= 1000;
		policy.sport_end	= 10000;
		policy.dport_start	= 2000;
		policy.dport_end	= 20000;
		ASSERT(rtl8651_addPolicyBasedQoS("vlan2", &policy, TRUE) == SUCCESS);
		ASSERT(rtl8651_addPolicyBasedQoS("vlan1", &policy, TRUE) == SUCCESS);

		curr = 1;
		/* Verify ACL Rule: for VLAN 2 */
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE ) curr++;
#ifdef TRAP_TCP_SYN_FIN_RST
		curr += 3; /* tcp syn, fin, rst */
#else
		//curr += 1; /* tcp syn */
#endif
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.srcIpAddr_		= policy.sip;
		rule1.srcIpAddrMask_	= policy.sipMask;
		rule1.dstIpAddr_		= policy.dip ;
		rule1.dstIpAddrMask_	= policy.dipMask;
		rule1.tcpSrcPortUB_		= policy.sport_end;
		rule1.tcpSrcPortLB_		= policy.sport_start;
		rule1.tcpDstPortUB_		= policy.dport_end;
		rule1.tcpDstPortLB_		= policy.dport_start;
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		rule1.priority_			= 1;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclEnd, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.outAclEnd, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		ASSERT(asic_vlan.outAclStart == asic_vlan.outAclEnd);

		curr = 1;
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE ) curr++;
		curr += 3; /* for tcp syn, fin, rst */
		/* Verify ACL Rule: for VLAN 1 */
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.srcIpAddr_		= policy.sip;
		rule1.srcIpAddrMask_	= policy.sipMask;
		rule1.dstIpAddr_		= policy.dip ;
		rule1.dstIpAddrMask_	= policy.dipMask;
		rule1.tcpSrcPortUB_		= policy.sport_end;
		rule1.tcpSrcPortLB_		= policy.sport_start;
		rule1.tcpDstPortUB_		= policy.dport_end;
		rule1.tcpDstPortLB_		= policy.dport_start;
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		rule1.priority_			= 1;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.outAclEnd, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		ASSERT(asic_vlan.outAclStart == asic_vlan.outAclEnd);
	}

	{
		/*================================================
		  * Test: del policy-based Qos
		  *================================================*/
		ASSERT(rtl8651_delPolicyBasedQoS("vlan2", &policy) == SUCCESS);
		/* Verify ACL Rule: for VLAN 2 */
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclEnd, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.outAclEnd, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		ASSERT(asic_vlan.outAclStart == asic_vlan.outAclEnd);
	}

	{
		
		/*================================================
		  * Test: flush policy-based Qos
		  *================================================*/
		ASSERT(rtl8651_flushPolicyBasedQoS() == SUCCESS);
		/* Verify ACL Rule: for VLAN 1 */
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclEnd, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.outAclEnd, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		ASSERT(asic_vlan.outAclStart == asic_vlan.outAclEnd);

	}
	return SUCCESS;
}






#define RTL865X_POLICY_ROUTE_TESTING_CODE

/*
@func int32 | policyRoute_1 | Test Policy Route 
@comm
Test add and delete demand route.
*/

int32 policyRoute_1(uint32 caseNo)
{
	//ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	rtl8651_tblDrvPolicy_t policy;
	_rtl8651_tblDrvPolicy_t *policy_t;
	ipaddr_t ipaddr;
	uint32 round, sessionId;
	

	START_TEST_CASE();	
	{
		/*===========================================
		  * Test: add/delete a Demand route
		  *===========================================*/
		default_configuration3();

		/* add the second PPPoE session */
		ASSERT(rtl8651_addPppoeSession(1001, 1) == SUCCESS);

		for(round=0; round<10000; round++) {
			/* (2) add a demand route */
			bzero(&policy, sizeof(rtl8651_tblDrvPolicy_t));
			policy.proto		= RTL8651_POLICY_TCP;
			policy.dip     		= ntohl(inet_addr("140.112.1.0")) + (( (uint32)(256.0 * rand() / (RAND_MAX + 1.0))<<8));
			policy.dipMask		= ntohl(inet_addr("255.255.255.0"));
			policy.sip			= ntohl(inet_addr("192.168.1.0")) + (( (uint32)(256.0 * rand() / (RAND_MAX + 1.0))<<8));
			policy.sipMask		= ntohl(inet_addr("255.255.255.0"));
			policy.sport_start	= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy.sport_end	= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy.dport_start	= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy.dport_end	= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			sessionId			 = (policy.dport_end%2)? 1000: 1001;
			/* add a non-exist session */
			ASSERT(rtl8651_addDemandRoute_1(&policy, 9000, (void *)123456) != SUCCESS);
			ASSERT(rtl8651_addDemandRoute_1(&policy, sessionId, (void *)0x123456) == SUCCESS);

			
			/* Verify Policy List */
			policy_t = TAILQ_FIRST(&DrvTbl.inuseList.policyRoute);
			ASSERT( policy_t && TAILQ_NEXT(policy_t, nextPolicy_t)==NULL);
			ASSERT(memcmp(policy_t, &policy, sizeof(rtl8651_tblDrvPolicy_t)) == 0);

			/* (3) remove the demand route */
			ASSERT(rtl8651_delDemandRoute(&policy) == SUCCESS);

			/* Verify */
			ASSERT(TAILQ_FIRST(&DrvTbl.inuseList.policyRoute) == NULL);
		}
	}
	return SUCCESS;
}


/*
@func int32 | policyRoute_2 | Test Policy Route 
@comm
Random test for demand route. The rtl8651_flushDemandRoute() also be tested.
*/

int32 policyRoute_2(uint32 caseNo)
{
	rtl8651_tblDrvPolicy_t policy[8];
	_rtl8651_tblDrvPolicy_t *policy_t;
	ipaddr_t ipaddr;
	uint32 round, sessionId[8], map[8] = { 0 }, sel;
	
	
	START_TEST_CASE();	
	{
		/*===========================================
		  * Test: add/delete randomly Demand route
		  *	      testing rtl8651_flushDemandRoute()
		  *===========================================*/
		default_configuration3();

		/* add the second PPPoE session */
		ASSERT(rtl8651_addPppoeSession(1001, 1) == SUCCESS);

		for(round=0; round<8; round++) {
			bzero(&policy[round], sizeof(rtl8651_tblDrvPolicy_t));
			policy[round].proto			= RTL8651_POLICY_TCP;
			policy[round].dip     			= ntohl(inet_addr("140.112.1.0")) + (round<<8);
			policy[round].dipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sip			= ntohl(inet_addr("192.168.1.0")) + (round<<8);
			policy[round].sipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].sport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			sessionId[round]			= (policy[round].dport_end%2)? 1000: 1001;
			map[round] = 0;
		}

		 for(round=0; round<100000; round++) {
			sel =  (uint32)(9.0 * rand() / (RAND_MAX + 1.0));
			if (sel == 8) {
				ASSERT(rtl8651_flushDemandRoute_1(1000) == SUCCESS);
				ASSERT(rtl8651_flushDemandRoute_1(1001) == SUCCESS);
				bzero(map, sizeof(uint32)*8);
			}
			else {
				if (map[sel] == 0) {
					ASSERT(rtl8651_addDemandRoute_1(&policy[sel], sessionId[sel], (void *)0x123456) == SUCCESS);
					map[sel] = 1;
				}
				else {
					ASSERT(rtl8651_delDemandRoute(&policy[sel]) == SUCCESS);
					map[sel] = 0;
				}
			}
			/* Verify */
			for(sel=0; sel<8; sel++) {
				TAILQ_FOREACH(policy_t, &DrvTbl.inuseList.policyRoute, nextPolicy_t) {
					if (!memcmp(policy_t, &policy[sel], sizeof(rtl8651_tblDrvPolicy_t))) {
						ASSERT(map[sel] == 1);
						goto keep;
					}
				}
				ASSERT(map[sel] == 0);
keep:
			}
		 }
	}
	return SUCCESS;
}


/*
@func int32 | policyRoute_3 | Test Policy Route.
@comm
Test add and delete static policy route.
*/

int32 policyRoute_3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	rtl8651_tblDrvPolicy_t policy;
	_rtl8651_tblDrvPolicy_t *policy_t;
	ipaddr_t ipaddr;
	uint32 round;

	
	START_TEST_CASE();	
	{
		/*===========================================
		  * Test: add/delete a static policy route
		  *===========================================*/
		default_configuration3();

		/* add the second PPPoE session */
		ASSERT(rtl8651_addPppoeSession(1001, 1) == SUCCESS);

		/* dial up PPPoE session */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_setPppoeSessionProperty(1001, 22222, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1001) == SUCCESS);
		

		for(round=0; round<10000; round++) {
			/* (2) add a demand route */
			bzero(&policy, sizeof(rtl8651_tblDrvPolicy_t));
			policy.proto		= RTL8651_POLICY_TCP;
			policy.dip     		= ntohl(inet_addr("140.112.1.0")) + (( (uint32)(256.0 * rand() / (RAND_MAX + 1.0))<<8));
			policy.dipMask		= ntohl(inet_addr("255.255.255.0"));
			policy.sip			= ntohl(inet_addr("192.168.1.0")) + (( (uint32)(256.0 * rand() / (RAND_MAX + 1.0))<<8));
			policy.sipMask		= ntohl(inet_addr("255.255.255.0"));
			policy.sport_start	= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy.sport_end	= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy.dport_start	= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy.dport_end	= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy.naptip		= 0xf0f0f0;
			/* add a non-exist naptip */
			ASSERT(rtl8651_addPolicyRoute_1(STATIC_POLICY_ROUTE, &policy) != SUCCESS);
			policy.naptip		= (policy.dport_end % 2)?  ntohl(inet_addr("140.113.1.1")): ntohl(inet_addr("140.113.1.2"));
			ASSERT(rtl8651_addPolicyRoute_1(STATIC_POLICY_ROUTE, &policy) == SUCCESS);

			
			/* Verify Policy List */
			policy_t = TAILQ_FIRST(&DrvTbl.inuseList.policyRoute);
			ASSERT( policy_t && TAILQ_NEXT(policy_t, nextPolicy_t)==NULL);
			ASSERT(memcmp(policy_t, &policy, sizeof(rtl8651_tblDrvPolicy_t)) == 0);

			/* (3) remove the demand route */
			ASSERT(rtl8651_delPolicyRoute(STATIC_POLICY_ROUTE, &policy) == SUCCESS);

			/* Verify */
			ASSERT(TAILQ_FIRST(&DrvTbl.inuseList.policyRoute) == NULL);
		}		  
	}
	return SUCCESS;
}



/*
@func int32 | policyRoute_4 | Test Policy Route.
@comm
Random test for static policy route. The rtl8651_flushPolicyRoute() also be tested.
*/

int32 policyRoute_4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	rtl8651_tblDrvPolicy_t policy[8];
	_rtl8651_tblDrvPolicy_t *policy_t;
	ipaddr_t ipaddr, naptip[8];
	uint32 round,  map[8] = { 0 }, sel;

	
	START_TEST_CASE();	
	{
		/*===========================================
		  * Test: add/delete randomly static policy route
		  *	      test rtl8651_flushPolicyRoute()
		  *===========================================*/
		default_configuration3();

		/* add the second PPPoE session */
		ASSERT(rtl8651_addPppoeSession(1001, 1) == SUCCESS);

		/* dial up PPPoE session */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_setPppoeSessionProperty(1001, 22222, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1001) == SUCCESS);
		
		for(round=0; round<8; round++) {
			bzero(&policy[round], sizeof(rtl8651_tblDrvPolicy_t));
			policy[round].proto			= RTL8651_POLICY_TCP;
			policy[round].dip     			= ntohl(inet_addr("140.112.1.0")) + (round<<8);
			policy[round].dipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sip			= ntohl(inet_addr("192.168.1.0")) + (round<<8);
			policy[round].sipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].sport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].naptip			= (policy[round].dport_end%2)? ntohl(inet_addr("140.113.1.1")): ntohl(inet_addr("140.113.1.2"));
			map[round] = 0;
		}

		 for(round=0; round<100000; round++) {
			sel =  (uint32)(9.0 * rand() / (RAND_MAX + 1.0));
			if (sel == 8) {
				ASSERT(rtl8651_flushPolicyRoute_1(ntohl(inet_addr("140.113.1.1"))) == SUCCESS);
				ASSERT(rtl8651_flushPolicyRoute_1(ntohl(inet_addr("140.113.1.2"))) == SUCCESS);
				bzero(map, sizeof(uint32)*8);
			}
			else {
				if (map[sel] == 0) {
					ASSERT(rtl8651_addPolicyRoute_1(STATIC_POLICY_ROUTE, &policy[sel]) == SUCCESS);
					map[sel] = 1;
				}
				else {
					ASSERT(rtl8651_delPolicyRoute(STATIC_POLICY_ROUTE, &policy[sel]) == SUCCESS);
					map[sel] = 0;
				}
			}
			/* Verify */
			for(sel=0; sel<8; sel++) {
				TAILQ_FOREACH(policy_t, &DrvTbl.inuseList.policyRoute, nextPolicy_t) {
					if (!memcmp(policy_t, &policy[sel], sizeof(rtl8651_tblDrvPolicy_t))) {
						ASSERT(map[sel] == 1);
						goto keep;
					}
				}
				ASSERT(map[sel] == 0);
keep:
			}
		 }
	}
	return SUCCESS;
}



/*
@func int32 | policyRoute_5 | Test Policy Route.
@comm
Test dynamic policy route's addition, deletion and flush.
*/
int32 policyRoute_5(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	rtl8651_tblDrvPolicy_t policy[10];
	_rtl8651_tblDrvPolicy_t *policy_t;
	ipaddr_t ipaddr;
	uint32 round, entry;

	
	START_TEST_CASE();	
	{
		/*===========================================
		  * Test: add/flush dynamic route
		  *===========================================*/
		default_configuration3();

		/* add the second PPPoE session */
		ASSERT(rtl8651_addPppoeSession(1001, 1) == SUCCESS);

		/* dial up PPPoE session */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_setPppoeSessionProperty(1001, 22222, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1001) == SUCCESS);

		/* add policy route */
		for(round=0; round<10; round++) {
			bzero(&policy[round], sizeof(rtl8651_tblDrvPolicy_t));
			policy[round].proto			= RTL8651_POLICY_TCP;
			policy[round].dip     			= ntohl(inet_addr("140.112.1.0")) + (round<<8);
			policy[round].dipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sip			= ntohl(inet_addr("192.168.1.0")) + (round<<8);
			policy[round].sipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].sport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].naptip			= (policy[round].dport_end%2)? ntohl(inet_addr("140.113.1.1")): ntohl(inet_addr("140.113.1.2"));
			ASSERT(rtl8651_addPolicyRoute_1(DYNAMIC_POLICY_ROUTE, &policy[round]) == SUCCESS);
		}

		/* Verify */
		entry = 0;
		CIRCLEQ_FOREACH(policy_t, &DrvTbl.dynamicPolicy.dyHead, nextTime_t) 
			if (policy_t->pppoe_t)
				entry ++;
		ASSERT(entry == 10);
		for(round=0; round<10; round++) {
			CIRCLEQ_FOREACH(policy_t, &DrvTbl.dynamicPolicy.dyHead, nextTime_t) {
				if (policy_t->pppoe_t == NULL)
					continue;
				if(!memcmp(policy_t, &policy[round], sizeof(rtl8651_tblDrvPolicy_t)))
					break;
			}
			ASSERT(policy_t != NULL);
		}

		/* flush policy route */
		ASSERT(rtl8651_flushPolicyRoute_1(ntohl(inet_addr("140.113.1.1"))) == SUCCESS);
		ASSERT(rtl8651_flushPolicyRoute_1(ntohl(inet_addr("140.113.1.2"))) == SUCCESS);
		
		/* Verfiy */
		CIRCLEQ_FOREACH(policy_t, &DrvTbl.dynamicPolicy.dyHead, nextTime_t)
			ASSERT(policy_t->pppoe_t == NULL);
	}
	return SUCCESS;
}


int32 policyRoute_6(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	rtl8651_tblDrvPolicy_t policy[64];
	_rtl8651_tblDrvPolicy_t *policy_t;
	ipaddr_t ipaddr;
	uint32 round, entry, head, map[32] = { 0 };


	START_TEST_CASE();	
	{
		/*===========================================
		  * Test: The dynamic policy route should have the following 
		  *	      behavior (N entry):
		  *	      1 2 3 4 5 .........N
		  *	      If add N+1, then the 1 should be replaced by N+1
		  *	      Note: assume N=32. make sure the 
		  *	     configPara.dynamicPolicy = 32
		  *===========================================*/
		default_configuration3();

		/* add the second PPPoE session */
		ASSERT(rtl8651_addPppoeSession(1001, 1) == SUCCESS);

		/* dial up PPPoE session */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_setPppoeSessionProperty(1001, 22222, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1001) == SUCCESS);

		/* Generate 64 policy route */
		for(round=0; round<64; round++) {
			bzero(&policy[round], sizeof(rtl8651_tblDrvPolicy_t));
			policy[round].proto			= RTL8651_POLICY_TCP;
			policy[round].dip     			= ntohl(inet_addr("140.112.1.0")) + (round<<8);
			policy[round].dipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sip			= ntohl(inet_addr("192.168.1.0")) + (round<<8);
			policy[round].sipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].sport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].naptip			= (policy[round].dport_end%2)? ntohl(inet_addr("140.113.1.1")): ntohl(inet_addr("140.113.1.2"));
		}

		/* Add dynamic Policy route */
		for(round=0, head=0; round<64; round++) {
			ASSERT(rtl8651_addPolicyRoute_1(DYNAMIC_POLICY_ROUTE, &policy[round]) == SUCCESS);
			map[head] = round;
			head = (head+1) % 32;
			/* Verify */
			entry = head;
			CIRCLEQ_FOREACH(policy_t, &DrvTbl.dynamicPolicy.dyHead, nextTime_t) {
				if (map[entry%32] != 0) {
					if (memcmp(policy_t, &policy[map[entry%32]], sizeof(rtl8651_tblDrvPolicy_t)))
						ASSERT(0);
				}
				entry ++;
			}
		}
	}
	return SUCCESS;
}


int32 policyRoute_7(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	rtl8651_tblDrvPolicy_t policy[10];
	_rtl8651_tblDrvPolicy_t *policy_t;
	ipaddr_t ipaddr;
	uint32 round, entry, map[10] = { 0 };


	START_TEST_CASE();	
	{
		/*===========================================
		  * Test: Remove dynamic policy
		  *	      assume N=32, make sure configPara.dynamicPolicy = 32
		  *===========================================*/
		default_configuration3();

		/* add the second PPPoE session */
		ASSERT(rtl8651_addPppoeSession(1001, 1) == SUCCESS);

		/* dial up PPPoE session */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_setPppoeSessionProperty(1001, 22222, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1001) == SUCCESS);

		/* Generate 10 policy route and add it */
		for(round=0; round<10; round++) {
			bzero(&policy[round], sizeof(rtl8651_tblDrvPolicy_t));
			policy[round].proto			= RTL8651_POLICY_TCP;
			policy[round].dip     			= ntohl(inet_addr("140.112.1.0")) + (round<<8);
			policy[round].dipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sip			= ntohl(inet_addr("192.168.1.0")) + (round<<8);
			policy[round].sipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].sport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].naptip			= (policy[round].dport_end%2)? ntohl(inet_addr("140.113.1.1")): ntohl(inet_addr("140.113.1.2"));
			ASSERT(rtl8651_addPolicyRoute_1(DYNAMIC_POLICY_ROUTE, &policy[round]) == SUCCESS);
		}

		/* Verify */
		entry = 0;
		CIRCLEQ_FOREACH(policy_t, &DrvTbl.dynamicPolicy.dyHead, nextTime_t) {
			if (entry < 22) {
				entry ++;
				continue;
			}
			if (memcmp(policy_t, &policy[entry-22], sizeof(rtl8651_tblDrvPolicy_t)))
				ASSERT(0);
			entry ++;
		}

		/* Remove dynamic policy */
		for(round=0; round<10; round++) {
			ASSERT(rtl8651_delPolicyRoute(DYNAMIC_POLICY_ROUTE, &policy[round]) == SUCCESS);
			/* Verify */
			entry = 0;
			CIRCLEQ_FOREACH(policy_t, &DrvTbl.dynamicPolicy.dyHead, nextTime_t) {
				if (entry < (22+round+1)) {
					entry ++;
					continue;
				}
				if (memcmp(policy_t, &policy[entry-22], sizeof(rtl8651_tblDrvPolicy_t)))
					ASSERT(0);
				entry ++;
			}
		}
	}
	return SUCCESS;
}



int32 policyRoute_8(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	rtl8651_tblDrvPolicy_t policy[10];
	_rtl8651_tblDrvPolicy_t *policy_t;
	ipaddr_t ipaddr;
	uint32 round, entry, i, map[10] = { 0 };
	int8 proto;


	START_TEST_CASE();	
	{
		/*===========================================
		  * Test: dynamic policy route lookup (if dynamic policy look up,
		  *	      imply static and demand policy look up ok)
		  *
		  *===========================================*/
		default_configuration3();

		/* add the second PPPoE session */
		ASSERT(rtl8651_addPppoeSession(1001, 1) == SUCCESS);

		/* dial up PPPoE session */
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1000) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_setPppoeSessionProperty(1001, 22222, &mac, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", ipaddr, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindPppoeSession(ipaddr, 1001) == SUCCESS);

		/* Generate 10 policy route and add it */
		for(round=0; round<10; round++) {
			bzero(&policy[round], sizeof(rtl8651_tblDrvPolicy_t));
			policy[round].proto			= RTL8651_POLICY_TCP;
			policy[round].dip     			= ntohl(inet_addr("140.112.1.0")) + (round<<8);
			policy[round].dipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sip			= ntohl(inet_addr("192.168.1.0")) + (round<<8);
			policy[round].sipMask		= ntohl(inet_addr("255.255.255.0"));
			policy[round].sport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].sport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_start		= 1 + (uint32)(10000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].dport_end		= 10000 + 1 + (uint32)(50000.0 * rand() / (RAND_MAX + 1.0));
			policy[round].naptip			= (policy[round].dport_end%2)? ntohl(inet_addr("140.113.1.1")): ntohl(inet_addr("140.113.1.2"));
			ASSERT(rtl8651_addPolicyRoute_1(DYNAMIC_POLICY_ROUTE, &policy[round]) == SUCCESS);
			map[round] = round;
		}

		/* Policy route look up */
		for(round=0; round<10; round++) {
			if (policy[round].proto == RTL8651_POLICY_TCP)
				proto = IPPROTO_TCP;
			else if (policy[round].proto == RTL8651_POLICY_UDP)
				proto = IPPROTO_UDP;
			else proto = IPPROTO_IP;
			policy_t = _rtl8651_getPolicyRoute(proto, policy[round].sip+1, policy[round].sport_start+1, policy[round].dip+1, policy[round].dport_start+1);
			ASSERT(policy_t != NULL);
			
			/* Verify */
			for(i=0, entry=round+1; i<10; i++) {
				map[i] = entry%10;
				entry = (entry+1)%10;
			}
			entry = 0;
			CIRCLEQ_FOREACH(policy_t, &DrvTbl.dynamicPolicy.dyHead, nextTime_t) {
				if (entry < 22) {
					entry ++;
					continue;
				}
				if (memcmp(policy_t, &policy[map[entry-22]], sizeof(rtl8651_tblDrvPolicy_t)))
					ASSERT(0);
				entry ++;
			}
		}
	}
	return SUCCESS;
}




int32 policyRoute_10(uint32 caseNo)
{
	ether_addr_t mac1 = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	ether_addr_t mac2 = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x02 } };
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl8651_tblDrvPolicy_t policyRoute;
	ipaddr_t naptip1, naptip2;
	int32 curr = 1;
	
	
	START_TEST_CASE();	
	{
		/*=============================================================
		 * Test: two PPPoE sessions
		 *=============================================================*/
		default_configuration3();
		ASSERT(rtl8651_addExtNetworkInterface("pppoe") == SUCCESS);

		/* (1) configure session 1: default napt */
		naptip1 = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac1, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", naptip1, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip1, 1000) == SUCCESS);
		ASSERT(rtl8651_addNaptMapping(naptip1) == SUCCESS);

		/* (2) configure session 2 */
		naptip2 = ntohl(inet_addr("10.1.1.2"));
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 2000, "pppoe1", 1) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(2000, 22222, &mac2, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe1", naptip2, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip2, 2000) == SUCCESS);
		ASSERT(rtl8651_addDriverNaptMapping(naptip2) == SUCCESS);

		/* (3) add Policy to session 2 */
		bzero(&policyRoute, sizeof(rtl8651_tblDrvPolicy_t));
		policyRoute.proto		= RTL8651_POLICY_TCP;
		policyRoute.dip			= ntohl(inet_addr("140.113.1.0"));
		policyRoute.dipMask		= ntohl(inet_addr("255.255.255.0"));
		policyRoute.dport_start	= 1000;
		policyRoute.dport_end	= 10000;
		policyRoute.sport_start	= 0;
		policyRoute.sport_end	= 65535;
		policyRoute.naptip	= naptip2;
		ASSERT(rtl8651_addPolicyRoute_1(STATIC_POLICY_ROUTE, &policyRoute) == SUCCESS);

		/* Verify ACL for VLAN Ingress */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_POLICY;
		rule1.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		rule1.tcpSrcPortUB_		= 65535;
		rule1.tcpSrcPortLB_		= 0;
		rule1.tcpDstPortUB_		= 10000;
		rule1.tcpDstPortLB_		= 1000;
		rule1.nhIndex			= 1;
		
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);

		/* (4) Reset Session 2: used for policy route */
		ASSERT(rtl8651_resetPppoeSessionProperty(2000) == SUCCESS);

		/* Verify ACL for VLAN Ingress */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_POLICY;
		rule1.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		rule1.tcpSrcPortUB_		= 65535;
		rule1.tcpSrcPortLB_		= 0;
		rule1.tcpDstPortUB_		= 10000;
		rule1.tcpDstPortLB_		= 1000;
		rule1.nhIndex			= 1;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
	}
	return SUCCESS;
}


int32 policyRoute_11(uint32 caseNo)
{
	ether_addr_t mac1 = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	ether_addr_t mac2 = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x02 } };
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl8651_tblDrvPolicy_t policyRoute;
	ipaddr_t naptip1, naptip2;
	int32 curr = 1;
	
	
	START_TEST_CASE();	
	{
		/*=============================================================
		 * Test: two PPPoE sessions
		 *=============================================================*/
		default_configuration3();
		ASSERT(rtl8651_addExtNetworkInterface("pppoe") == SUCCESS);

		/* (1) configure session 1: default napt */
		naptip1 = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac1, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", naptip1, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip1, 1000) == SUCCESS);
		
		/* (2) configure session 2 */
		naptip2 = ntohl(inet_addr("10.1.1.2"));
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 2000, "pppoe1", 1) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(2000, 22222, &mac2, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe1", naptip2, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip2, 2000) == SUCCESS);
		ASSERT(rtl8651_addDriverNaptMapping(naptip2) == SUCCESS);

		/* (3) add Policy to session 2 */
		bzero(&policyRoute, sizeof(rtl8651_tblDrvPolicy_t));
		policyRoute.proto		= RTL8651_POLICY_TCP;
		policyRoute.dip			= ntohl(inet_addr("140.113.1.0"));
		policyRoute.dipMask		= ntohl(inet_addr("255.255.255.0"));
		policyRoute.dport_start	= 1000;
		policyRoute.dport_end	= 10000;
		policyRoute.sport_start	= 0;
		policyRoute.sport_end	= 65535;
		policyRoute.naptip	= naptip2;
		ASSERT(rtl8651_addPolicyRoute_1(STATIC_POLICY_ROUTE, &policyRoute) == SUCCESS);

		/* Verify ACL for VLAN Ingress */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_POLICY;
		rule1.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		rule1.tcpSrcPortUB_		= 65535;
		rule1.tcpSrcPortLB_		= 0;
		rule1.tcpDstPortUB_		= 10000;
		rule1.tcpDstPortLB_		= 1000;
		rule1.nhIndex			= 1;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);

		/* (4) add default NAPT IP: session 1 */
		ASSERT(rtl8651_addNaptMapping(naptip1) == SUCCESS);

		/* Verify ACL for VLAN Ingress */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_POLICY;
		rule1.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		rule1.tcpSrcPortUB_		= 65535;
		rule1.tcpSrcPortLB_		= 0;
		rule1.tcpDstPortUB_		= 10000;
		rule1.tcpDstPortLB_		= 1000;
		rule1.nhIndex			= 2;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);	

		/* (5) remove default NAPT */
		ASSERT(rtl8651_delNaptMapping(naptip1) == SUCCESS);

		/* Verify ACL for VLAN Ingress */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_POLICY;
		rule1.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		rule1.tcpSrcPortUB_		= 65535;
		rule1.tcpSrcPortLB_		= 0;
		rule1.tcpDstPortUB_		= 10000;
		rule1.tcpDstPortLB_		= 1000;
		rule1.nhIndex			= 1;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);	
	}
	return SUCCESS;
}











#define RTL865X_ACL_CODE

/*
@func int32 | acl_test1 | ACL Testing.
@comm
Test add and delete MAC type ACL rule.
*/

int32 acl_test1(uint32 caseNo)
{
	uint8 ruleType[] = { RTL8651_ACL_TCP, RTL8651_ACL_UDP, RTL8651_ACL_IP, RTL8651_ACL_MAC, RTL8651_ACL_ICMP };
	uint8 ruleAction[]  = { RTL8651_ACL_DROP, RTL8651_ACL_PERMIT, RTL8651_ACL_CPU };
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	ether_addr_t mMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	rtl8651_tblDrvAclRule_t rule;
	uint32 ruleCount, actionCount;
	int32 curr;
	
	
	START_TEST_CASE();	
	{
		/*======================================
		  * Test: add/delete ACL rule
		  *======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		for(ruleCount=0; ruleCount<5; ruleCount++) {
			for(actionCount=0; actionCount<3; actionCount++) {
				bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
				switch(ruleType[ruleCount]) {
				case RTL8651_ACL_MAC:
					rule.ruleType_ 			= RTL8651_ACL_MAC;
					memcpy(&rule.dstMac_, &mac, 6);
					memcpy(&rule.dstMacMask_, &mMask, 6); mac.octet[5] ++;
					memcpy(&rule.srcMac_, &mac, 6); 
					memcpy(&rule.srcMacMask_, &mMask, 6);
					rule.typeLen_			= 0x08;
					rule.typeLenMask_		= 0xff;
					break;
					
				case RTL8651_ACL_TCP:
					rule.ruleType_ 			= RTL8651_ACL_TCP;
					rule.tcpSrcPortUB_		= 6000;
					rule.tcpSrcPortLB_		= 7000;
					rule.tcpDstPortUB_		= 10000;
					rule.tcpDstPortLB_		= 20000;
					rule.tcpFlagMask_		= 0xff;
					rule.tcpFlag_			= 0x04;
					goto commL3;

				case RTL8651_ACL_UDP:
					rule.ruleType_ 			= RTL8651_ACL_UDP;
					rule.udpSrcPortUB_		= 1000;
					rule.udpSrcPortLB_		= 20000;
					rule.udpDstPortUB_		= 3000;
					rule.udpDstPortLB_		= 40000;
					goto commL3;
					
				case RTL8651_ACL_ICMP:
					rule.ruleType_ 			= RTL8651_ACL_ICMP;
					rule.icmpType_			= 0x02;
					rule.icmpTypeMask_		= 0xff;
					rule.icmpCode_		= 0x04;
					rule.icmpCodeMask_	= 0xff;
					goto commL3;
				
				case RTL8651_ACL_IP:
					rule.ruleType_ 			= RTL8651_ACL_IP;
					rule.tos_				= 0x01;
					rule.tosMask_			= 0xff;
commL3:				rule.srcIpAddr_		= 0x12345678;
					rule.srcIpAddrMask_	= 0xffffff00;
					rule.dstIpAddr_		= 0x87654321;
					rule.dstIpAddrMask_	= 0xffff0000;
					break;

				default: ASSERT(0);

				}
				rule.actionType_ = ruleAction[actionCount];
				ASSERT(rtl8651_addAclRule("vlan2", TRUE, &rule) == SUCCESS);
				
				curr = 0;
				/* Verify */
				ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
				// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
				if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE ) curr++;
#ifdef RTL865XB_URL_FILTER
				if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
				bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
				bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
				memcpy(&rule1, &rule, sizeof(rtl8651_tblDrvAclRule_t));
				ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);

				ACL_ENTRY_CMP(&rule1, &rule2);
				curr++;

				/* Remove ACL */
				ASSERT(rtl8651_delAclRule("vlan2", TRUE, &rule) == SUCCESS);

			}

		}

	}
	return SUCCESS;
}






int32 acl_test11(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl8651_tblDrvAclRule_t rule;
	rtl8651_EnablePppoePassthru( TRUE );
	int curr; // current offset of testing ACL rule
	ether_addr_t mac_vlan1 = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ether_addr_t mac_vlan2 = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x02 } };
	

	START_TEST_CASE();	
	{
		/*======================================
		  * Test: Test IP Range ACL rule for Ingress rule
		  *======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add TCP rule: trap port */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.tcpSrcPortUB_		= 6000;
		rule.tcpSrcPortLB_		= 7000;
		rule.tcpDstPortUB_		= 10000;
		rule.tcpDstPortLB_		= 20000;
		ASSERT(rtl8651_addAclRule("vlan1", TRUE, &rule) == SUCCESS);

		/* (2) add TCP rule: trap destination IP */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addAclRule("vlan1", TRUE, &rule) == SUCCESS);

		/* (3) add TCP rule: trap destination IP RANGE!! */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddrStart_	= ntohl(inet_addr("140.112.1.10"));
		rule.dstIpAddrEnd_		= ntohl(inet_addr("140.112.1.200"));
		rule.tcpSrcPortUB_		= 0xffff;
		rule.tcpSrcPortLB_		= 0x0000;
		rule.tcpDstPortUB_		= 0xffff;
		rule.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_addAclRuleExt("vlan1", TRUE, &rule) == SUCCESS);

		/* (4) add UDP rule: trap port */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_UDP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.udpSrcPortUB_		= 10000;
		rule.udpSrcPortLB_		= 1000;
		rule.udpDstPortUB_		= 20000;
		rule.udpDstPortLB_		= 2000;		
		ASSERT(rtl8651_addAclRule("vlan1", TRUE, &rule) == SUCCESS);

		curr = 0;
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
		/* Verify ACL Rule: VLAN 1 */
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
		//0x8863
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.typeLen_			= 0x8863;
		rule1.typeLenMask_		= 0xffff;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//0x8864
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.typeLen_			= 0x8864;
		rule1.typeLenMask_		= 0xffff;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#endif// CONFIG_RTL865XB
		//first rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.tcpSrcPortUB_		= 6000;
		rule1.tcpSrcPortLB_		= 7000;
		rule1.tcpDstPortUB_		= 10000;
		rule1.tcpDstPortLB_		= 20000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//second rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#if 0	// chenyl (2005/03/02): Edward: IP Range ACL is only for L34 process, so we don't set DEFAULT-TO-CPU if ip range rule exist
		//This last rule for this vlan is toCPU due to IP Range acl rule exists
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#else
		#if 0	// shliu: we change the implementation of ip range acl rule.
		// third rule we added: Trap all IP packet with L3/L4 action need
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.typeLen_		= 0x0800;
		rule1.typeLenMask_	= 0xffff;
		memcpy(&rule1.dstMac_, &mac_vlan1, 6);
		memset(&rule1.dstMacMask_, 0xff, 6);
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		#endif

		// 192.168.1.10/31
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.10"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.254"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.12/30
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.12"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.252"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		
		// 192.168.1.16/28
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.16"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.240"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.32/27
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.32"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.224"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.64/26
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.64"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.192"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.200/32
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.200"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.255"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.192/29
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.192"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.248"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.128/26
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.128"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.192"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		
		// fourth rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_UDP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.udpSrcPortUB_		= 10000;
		rule1.udpSrcPortLB_		= 1000;
		rule1.udpDstPortUB_		= 20000;
		rule1.udpDstPortLB_		= 2000;		
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#endif
		//Egress
		ASSERT(asic_vlan.outAclStart == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.outAclEnd == RTL8651_ACLTBL_PERMIT_ALL);

		curr = 0;
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
		//curr += 3;
		/* Verify ACL Rule: VLAN 2 */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
		//0x8863
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.typeLen_			= 0x8863;
		rule1.typeLenMask_		= 0xffff;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//0x8864
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.typeLen_			= 0x8864;
		rule1.typeLenMask_		= 0xffff;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#endif// CONFIG_RTL865XB
		//permit rule
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//Egress
		ASSERT(asic_vlan.outAclStart == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.outAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
	}

	{
		rtl8651_tblDrv_networkIntfTable_t *netif_t;
		struct rtl_mBuf* mbuf_in;
		rtl8651_PktConf_t	conf;
		_rtl8651_tblDrvAclRule_t *aclRule;
		ether_addr_t smac = { { 0x00, 0x00, 0x01, 0x02, 0x03, 0x04 } };
		ether_addr_t dmac = { { 0x00, 0x00, 0x02, 0x02, 0x03, 0x04 } };
		ipaddr_t sip = ntohl(inet_addr("192.168.1.1"));
		ipaddr_t dip[] = { ntohl(inet_addr("140.112.1.50")), ntohl(inet_addr("140.112.1.200")), ntohl(inet_addr("140.112.1.10")),  
					     ntohl(inet_addr("140.112.1.5")), ntohl(inet_addr("140.112.1.250")) };
		uint16 sport = 12345;
		uint16 dport = 23456;
		int8 *ethdr, *iphdr;
		int32 round;
		
		/* acl lookup for ip range acl rule */
		for (round=0; round<5; round++ ) {
			memset(&conf, 0, sizeof(rtl8651_PktConf_t));
			conf.pktType						= _PKT_TYPE_TCP;					/* tcp packet */
			memcpy (&smac, conf.conf_smac, 6);
			memcpy (&dmac, conf.conf_dmac, 6);
			conf.conf_sip						= sip;
			conf.conf_dip						= dip[round];
			conf.conf_sport					= sport;
			conf.conf_dport					= dport;
			mbuf_in = genMbufPkt(&conf);

			netif_t = GET_NETWORK_INTERFACE("vlan1", NULL);
			ASSERT(netif_t);
			ethdr = (int8 *)mbuf_in->m_data;
			iphdr = (int8 *)(ethdr + 14);
			aclRule = _rtl8651_aclLookUp(0, netif_t, TRUE, 0, ethdr, iphdr, NULL, RTL8651_ACL_GENERIC, _RTL8651_OPERATION_OTHER);
			if (round <= 2) {
				ASSERT(aclRule != NULL);
				ASSERT(aclRule->isIPRange_ == 1);
			}
			else {
				ASSERT(aclRule == NULL);
			}
			freeMbuf(mbuf_in);
		
		}
	}

	{
		/* remove IP Range acl rule */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddrStart_	= ntohl(inet_addr("140.112.1.10"));
		rule.dstIpAddrEnd_		= ntohl(inet_addr("140.112.1.200"));
		rule.tcpSrcPortUB_		= 0xffff;
		rule.tcpSrcPortLB_		= 0x0000;
		rule.tcpDstPortUB_		= 0xffff;
		rule.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_delAclRuleExt("vlan1", TRUE, &rule) == SUCCESS);

		curr = 0;
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
		/* Verify ACL Rule for VLAN 1 */
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		if ( rtl8651_drvDropUnknownPppoePADT )
		{
			//0x8863
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_MAC;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.typeLen_			= 0x8863;
			rule1.typeLenMask_		= 0xffff;
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}
#if 0/* 5XB no longer use ACL */
		//0x8864
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.typeLen_			= 0x8864;
		rule1.typeLenMask_		= 0xffff;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#endif/*0*/
		//first rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.tcpSrcPortUB_		= 6000;
		rule1.tcpSrcPortLB_		= 7000;
		rule1.tcpDstPortUB_		= 10000;
		rule1.tcpDstPortLB_		= 20000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//second rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//udp rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_UDP;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.udpSrcPortUB_	= 10000;
		rule1.udpSrcPortLB_	= 1000;
		rule1.udpDstPortUB_	= 20000;
		rule1.udpDstPortLB_	= 2000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//curr += 3; /* for TCP syn, rst, fin */
		//permit rule: default rule
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//Egress rule for VLAN 1
		ASSERT(asic_vlan.outAclStart == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.outAclEnd == RTL8651_ACLTBL_PERMIT_ALL);

		curr = 0;
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
		/* Verify ACL rule for VLAN 2 */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		if ( rtl8651_drvDropUnknownPppoePADT )
		{
			//0x8863
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_MAC;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.typeLen_			= 0x8863;
			rule1.typeLenMask_		= 0xffff;
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}
#if 0/* 5XB no longer use ACL */
		//0x8864
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.typeLen_			= 0x8864;
		rule1.typeLenMask_		= 0xffff;
		ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#endif/*0*/
		//curr += 3; /* for tcp syn, fin, rst */
		//permit rule
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//Egress
		ASSERT(asic_vlan.outAclStart == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.outAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
		
	}
	return SUCCESS;
}



int32 acl_test12(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl8651_tblDrvAclRule_t rule;
	int32 curr;
	ether_addr_t mac_vlan1 = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ether_addr_t mac_vlan2 = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x02 } };

	

	START_TEST_CASE();	
	{
		/*======================================
		  * Test: Test IP Range ACL rule for Egress rule
		  *======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add TCP rule: trap port */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.tcpSrcPortUB_		= 6000;
		rule.tcpSrcPortLB_		= 7000;
		rule.tcpDstPortUB_		= 10000;
		rule.tcpDstPortLB_		= 20000;
		ASSERT(rtl8651_addAclRule("vlan1", FALSE, &rule) == SUCCESS);

		/* (2) add TCP rule: trap destination IP */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addAclRule("vlan1", FALSE, &rule) == SUCCESS);

		/* (3) add TCP rule: trap destination IP RANGE!! */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddrStart_	= ntohl(inet_addr("140.112.1.10"));
		rule.dstIpAddrEnd_		= ntohl(inet_addr("140.112.1.200"));
		rule.tcpSrcPortUB_		= 0xffff;
		rule.tcpSrcPortLB_		= 0x0000;
		rule.tcpDstPortUB_		= 0xffff;
		rule.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_addAclRuleExt("vlan1", FALSE, &rule) == SUCCESS);

		/* (4) add UDP rule: trap port */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_UDP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.udpSrcPortUB_		= 10000;
		rule.udpSrcPortLB_		= 1000;
		rule.udpDstPortUB_		= 20000;
		rule.udpDstPortLB_		= 2000;		
		ASSERT(rtl8651_addAclRule("vlan1", FALSE, &rule) == SUCCESS);

		curr = 0;
		/* Verify ACL Rule: VLAN 1 */
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		//first rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.tcpSrcPortUB_		= 6000;
		rule1.tcpSrcPortLB_		= 7000;
		rule1.tcpDstPortUB_		= 10000;
		rule1.tcpDstPortLB_		= 20000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//second rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#if 0	// chenyl (2005/03/02): Edward: IP Range ACL is only for L34 process, so we don't set DEFAULT-TO-CPU if ip range rule exist
		//This last rule for this vlan is toCPU due to IP Range acl rule exists
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#else
		#if 0
		// third rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.typeLen_		= 0x0800;
		rule1.typeLenMask_	= 0xffff;
		memcpy(&rule1.dstMac_, &mac_vlan1, 6);
		memset(&rule1.dstMacMask_, 0xff, 6);
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		#endif

		// 192.168.1.10/31
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.10"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.254"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.12/30
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.12"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.252"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		
		// 192.168.1.16/28
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.16"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.240"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.32/27
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.32"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.224"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.64/26
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.64"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.192"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.200/32
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.200"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.255"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.192/29
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.192"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.248"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.128/26
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.128"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.192"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// fourth rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_UDP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.udpSrcPortUB_		= 10000;
		rule1.udpSrcPortLB_		= 1000;
		rule1.udpDstPortUB_		= 20000;
		rule1.udpDstPortLB_		= 2000;		
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#endif
		//Ingress
		curr = 0;
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		//0x8863
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_MAC;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.typeLen_			= 0x8863;
			rule1.typeLenMask_		= 0xffff;
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
#ifdef TRAP_TCP_SYN_FIN_RST
		//curr += 3; /* tcp syn, fin, rst */
#else
		//curr += 1; /* tcp syn */
#endif
		//default rule
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);


		curr = 0;
		/* Verify ACL Rule: VLAN 2 */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		//0x8863
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_MAC;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.typeLen_			= 0x8863;
			rule1.typeLenMask_		= 0xffff;
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
		//curr += 3; /* tcp syn, fin, rst */
		//permit rule
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//Egress
		ASSERT(asic_vlan.outAclStart == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.outAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
	}


	{
		rtl8651_tblDrv_networkIntfTable_t *netif_t;
		struct rtl_mBuf* mbuf_in;
		rtl8651_PktConf_t	conf;
		_rtl8651_tblDrvAclRule_t *aclRule;
		ether_addr_t smac = { { 0x00, 0x00, 0x01, 0x02, 0x03, 0x04 } };
		ether_addr_t dmac = { { 0x00, 0x00, 0x02, 0x02, 0x03, 0x04 } };
		ipaddr_t sip = ntohl(inet_addr("192.168.1.1"));
		ipaddr_t dip[] = { ntohl(inet_addr("140.112.1.50")), ntohl(inet_addr("140.112.1.200")), ntohl(inet_addr("140.112.1.10")),  
					     ntohl(inet_addr("140.112.1.5")), ntohl(inet_addr("140.112.1.250")) };
		uint16 sport = 12345;
		uint16 dport = 23456;
		int8 *ethdr, *iphdr;
		int32 round;
		
		/* acl lookup for ip range acl rule */
		for (round=0; round<5; round++ ) {
			memset(&conf, 0, sizeof(rtl8651_PktConf_t));
			conf.pktType						= _PKT_TYPE_TCP;					/* tcp packet */
			memcpy (&smac, conf.conf_smac, 6);
			memcpy (&dmac, conf.conf_dmac, 6);
			conf.conf_sip						= sip;
			conf.conf_dip						= dip[round];
			conf.conf_sport					= sport;
			conf.conf_dport					= dport;
			mbuf_in = genMbufPkt(&conf);

			netif_t = GET_NETWORK_INTERFACE("vlan1", NULL);
			ASSERT(netif_t);
			ethdr = (int8 *)mbuf_in->m_data;
			iphdr = (int8 *)(ethdr + 14);
			aclRule = _rtl8651_aclLookUp(0, netif_t, FALSE, 0, ethdr, iphdr, NULL, RTL8651_ACL_GENERIC, _RTL8651_OPERATION_OTHER);
			if (round <= 2) {
				ASSERT(aclRule != NULL);
				ASSERT(aclRule->isIPRange_ == 1);
			}
			else {
				ASSERT(aclRule == NULL);
			}
			freeMbuf(mbuf_in);
		
		}
	}


	{
		/* remove IP Range acl rule */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddrStart_	= ntohl(inet_addr("140.112.1.10"));
		rule.dstIpAddrEnd_		= ntohl(inet_addr("140.112.1.200"));
		rule.tcpSrcPortUB_		= 0xffff;
		rule.tcpSrcPortLB_		= 0x0000;
		rule.tcpDstPortUB_		= 0xffff;
		rule.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_delAclRuleExt("vlan1", FALSE, &rule) == SUCCESS);

		curr = 0;
		/* Verify ACL Rule for VLAN 1 */
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		//0x8863
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_MAC;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.typeLen_			= 0x8863;
			rule1.typeLenMask_		= 0xffff;
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
#ifdef TRAP_TCP_SYN_FIN_RST
		//curr += 3; /* tcp syn, fin, rst */
#else
		//curr += 1; /* tcp syn */
#endif
		//permit rule: default rule
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		curr = 0;
		//Egress rule
		//first rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.tcpSrcPortUB_		= 6000;
		rule1.tcpSrcPortLB_		= 7000;
		rule1.tcpDstPortUB_		= 10000;
		rule1.tcpDstPortLB_		= 20000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//second rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_TCP;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.dstIpAddr_		= ntohl(inet_addr("140.113.1.0"));
		rule1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//udp rule we added
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_UDP;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.udpSrcPortUB_	= 10000;
		rule1.udpSrcPortLB_	= 1000;
		rule1.udpDstPortUB_	= 20000;
		rule1.udpDstPortLB_	= 2000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;


		curr = 0;
		/* Verify ACL rule for VLAN 2 */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		//0x8863
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_MAC;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.typeLen_			= 0x8863;
			rule1.typeLenMask_		= 0xffff;
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
		//curr += 3; /* for tcp syn, fin, rst */
		//permit rule
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		//Egress
		ASSERT(asic_vlan.outAclStart == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.outAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
		
	}
	return SUCCESS;
}

int32 acl_test13(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl8651_tblDrvAclRule_t rule;
	int32 round;
	int32 NumToBeAdded;
	int32 curr;
	ether_addr_t mac_vlan1 = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
	NumToBeAdded = 122;
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB

	if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		NumToBeAdded --;

#ifdef RTL865XB_URL_FILTER
	if(_urlFilter.urlFilterCount >= 0)
		NumToBeAdded --;
#endif/*RTL865XB_URL_FILTER*/
	
	START_TEST_CASE();	
	{
		/*======================================
		  * Test: Test IP Range ACL in out of asic space condition
		  *======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add 121 UDP rule not include ip range acl */
		for(round=0; round<NumToBeAdded; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_UDP;
			rule.actionType_		= RTL8651_ACL_CPU;
			rule.udpSrcPortUB_		= 6000 + round;
			rule.udpSrcPortLB_		= 7000 + round;
			rule.udpDstPortUB_		= 10000 + round;
			rule.udpDstPortLB_		= 20000 + round;
			ASSERT(rtl8651_addAclRule("vlan1", TRUE, &rule) == SUCCESS);
		}

		curr = 0;
		/* Verify ACL rule */
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		//0x8863
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_MAC;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.typeLen_			= 0x8863;
			rule1.typeLenMask_		= 0xffff;
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
		for(round=(asic_vlan.inAclStart+curr); round<NumToBeAdded; round++) {
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_UDP;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.udpSrcPortUB_		= 6000 + round-(asic_vlan.inAclStart+curr);
			rule1.udpSrcPortLB_		= 7000 + round-(asic_vlan.inAclStart+curr);
			rule1.udpDstPortUB_		= 10000 + round-(asic_vlan.inAclStart+curr);
			rule1.udpDstPortLB_		= 20000 + round-(asic_vlan.inAclStart+curr);
			ASSERT(rtl8651_getAsicAclRule(round, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);

		}
		curr+=NumToBeAdded;
		//default rule: permit
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule(curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);

		/* add ip Range ACL: this ip range ACL will replace the original default rule. */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddrStart_	= ntohl(inet_addr("140.112.1.10"));
		rule.dstIpAddrEnd_		= ntohl(inet_addr("140.112.1.200"));
		rule.tcpSrcPortUB_		= 0xffff;
		rule.tcpSrcPortLB_		= 0x0000;
		rule.tcpDstPortUB_		= 0xffff;
		rule.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_addAclRuleExt("vlan1", TRUE, &rule) == SUCCESS);

		#if 0
		/* Verify ip Range ACL rule. */
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.typeLen_		= 0x0800;
		rule1.typeLenMask_	= 0xffff;
		memcpy(&rule1.dstMac_, &mac_vlan1, 6);
		memset(&rule1.dstMacMask_, 0xff, 6);
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		#endif

		// 192.168.1.0/24
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_IP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.0"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);

		//last two rules are changed to "L2 permit" & "CPU" because rule count exceed */
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		rule1.pktOpApp		= 1;
		ASSERT(rtl8651_getAsicAclRule(123, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		ASSERT(rtl8651_getAsicAclRule(124, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);

		//Egress
		ASSERT(asic_vlan.outAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.outAclStart == RTL8651_ACLTBL_PERMIT_ALL);

		/* Verify VLAN 2 */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

		//Ingress:
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) {
			// Because ASIC ACL table is full, 
			// last two rules are changed to "L2 permit" & "CPU" because rule count exceed
			ASSERT(asic_vlan.inAclStart == 123);
			ASSERT(asic_vlan.inAclEnd == 124);
		}
#else	/* RTL865XB_URL_FILTER */

#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// PPPoE uses Protocol-Based VLAN, not ACL
		ASSERT(asic_vlan.inAclStart == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.inAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			ASSERT(asic_vlan.inAclStart == 123);
			ASSERT(asic_vlan.inAclEnd == 124);
		}
		else
		{
			ASSERT(asic_vlan.inAclStart == RTL8651_ACLTBL_PERMIT_ALL);
			ASSERT(asic_vlan.inAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
		}
#endif	/* RTL865XB_URL_FILTER */


		//Egress:
		ASSERT(asic_vlan.outAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.outAclStart == RTL8651_ACLTBL_PERMIT_ALL);

		/* remove the ip range acl rule */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddrStart_	= ntohl(inet_addr("140.112.1.10"));
		rule.dstIpAddrEnd_		= ntohl(inet_addr("140.112.1.200"));
		rule.tcpSrcPortUB_		= 0xffff;
		rule.tcpSrcPortLB_		= 0x0000;
		rule.tcpDstPortUB_		= 0xffff;
		rule.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_delAclRuleExt("vlan1", TRUE, &rule) == SUCCESS);

		//last one rule is changed to PERMT */
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		ASSERT(rtl8651_getAsicAclRule(curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		
	}
	return SUCCESS;
}



int32 acl_test14(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl8651_tblDrvAclRule_t rule;
	int32 round;
	int32 curr;
	int32 NumToBeAdded;
	ether_addr_t mac_vlan1 = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ether_addr_t mac_vlan2 = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x02 } };
	
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
	NumToBeAdded = 122;
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB

	if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		NumToBeAdded --;

#ifdef RTL865XB_URL_FILTER
	if(_urlFilter.urlFilterCount >= 0)
		NumToBeAdded --;
#endif/*RTL865XB_URL_FILTER*/

	START_TEST_CASE();	
	{
		/*======================================
		  * Test: Test IP Range ACL in out of asic space condition
		  *======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* add ip Range ACL */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddrStart_	= ntohl(inet_addr("140.112.1.10"));
		rule.dstIpAddrEnd_		= ntohl(inet_addr("140.112.1.200"));
		rule.tcpSrcPortUB_		= 0xffff;
		rule.tcpSrcPortLB_		= 0x0000;
		rule.tcpDstPortUB_		= 0xffff;
		rule.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_addAclRuleExt("vlan1", TRUE, &rule) == SUCCESS);

		for(round=0; round<NumToBeAdded; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_UDP;
			rule.actionType_		= RTL8651_ACL_CPU;
			rule.udpSrcPortUB_		= 6000 + round;
			rule.udpSrcPortLB_		= 7000 + round;
			rule.udpDstPortUB_		= 10000 + round;
			rule.udpDstPortLB_		= 20000 + round;
			ASSERT(rtl8651_addAclRule("vlan1", TRUE, &rule) == SUCCESS);
		}

		curr = 0;
		/* Verify ACL Rule */
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// PPPoE uses Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		//0x8863
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_MAC;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.typeLen_			= 0x8863;
			rule1.typeLenMask_		= 0xffff;
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
#if 0	// chenyl (2005/03/02): Edward: IP Range ACL is only for L34 process, so we don't set DEFAULT-TO-CPU if ip range rule exist
		//toCPU rule
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
#else
		#if 0
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		rule1.typeLen_		= 0x0800;
		rule1.typeLenMask_	= 0xffff;
		memcpy(&rule1.dstMac_, &mac_vlan1, 6);
		memset(&rule1.dstMacMask_, 0xff, 6);
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		#endif

		// 192.168.1.10/31
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.10"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.254"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.12/30
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.12"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.252"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		
		// 192.168.1.16/28
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.16"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.240"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.32/27
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.32"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.224"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.64/26
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.64"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.192"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.200/32
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.200"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.255"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.192/29
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.192"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.248"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.128/26
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.128"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.192"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

#endif

		/* remove ip range acl rule */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddrStart_	= ntohl(inet_addr("140.112.1.10"));
		rule.dstIpAddrEnd_		= ntohl(inet_addr("140.112.1.200"));
		rule.tcpSrcPortUB_		= 0xffff;
		rule.tcpSrcPortLB_		= 0x0000;
		rule.tcpDstPortUB_		= 0xffff;
		rule.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_delAclRuleExt("vlan1", TRUE, &rule) == SUCCESS);

		curr = 0;
		/* Verify ACL rule: VLAN 1 */
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// PPPoE uses Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		//0x8863
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_MAC;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.typeLen_			= 0x8863;
			rule1.typeLenMask_		= 0xffff;
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
		for(round=0; round<NumToBeAdded; round++) {
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_UDP;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.udpSrcPortUB_		= 6000 + round;
			rule1.udpSrcPortLB_		= 7000 + round;
			rule1.udpDstPortUB_		= 10000 + round;
			rule1.udpDstPortLB_		= 20000 + round;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+round+curr), &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
		}

		ASSERT(asic_vlan.inAclEnd == asic_vlan.inAclStart + curr + NumToBeAdded);
	}
	return SUCCESS;
}

/* Test for "srcMac & dstPort" ACL rule when we have the arp entry of this srcMac. */
int32 acl_test15(uint32 caseNo)
{
	uint8 ruleType[] = { RTL8651_ACL_TCP, RTL8651_ACL_UDP};
	uint8 ruleAction[]  = { RTL8651_ACL_DROP, RTL8651_ACL_PERMIT, RTL8651_ACL_CPU };
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	ether_addr_t mac1 = { { 0x00, 0x00, 0xab, 0x11, 0x55, 0x00 } };
	ether_addr_t mMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	rtl8651_tblDrvAclRule_t rule;
	rtl8651_tblDrv_arpEntry_t *arp_t;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	uint32 ruleCount, actionCount;
	int32 curr;
	ipaddr_t ipAddr = ntohl(inet_addr("192.168.2.200"));
		
	START_TEST_CASE();	
	{
		/*======================================
		  * Test: add/delete srcMac & dstPort ACL rule
		  *======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* add arp */
		ASSERT(rtl8651_addArp(ipAddr, &mac1, "vlan2", 0x02) == SUCCESS);

		for(ruleCount=0; ruleCount<2; ruleCount++) {
			for(actionCount=0; actionCount<3; actionCount++) {
				bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
				switch(ruleType[ruleCount]) {					
				case RTL8651_ACL_TCP:
					rule.ruleType_ 		= RTL8651_ACL_TCP;
					rule.tcpl2srcMac_		= mac1;
					rule.tcpl2srcMacMask_	= mMask;
					rule.tcpSrcPortUB_		= 6000;
					rule.tcpSrcPortLB_		= 7000;
					rule.tcpDstPortUB_		= 10000;
					rule.tcpDstPortLB_		= 20000;
					break;

				case RTL8651_ACL_UDP:
					rule.ruleType_ 		= RTL8651_ACL_UDP;
					rule.udpl2srcMac_		= mac1;
					rule.udpl2srcMacMask_	= mMask;
					rule.udpSrcPortUB_	= 1000;
					rule.udpSrcPortLB_		= 20000;
					rule.udpDstPortUB_	= 3000;
					rule.udpDstPortLB_		= 40000;
					break;

				default: ASSERT(0);

				}
				rule.actionType_ = ruleAction[actionCount];
				
				ASSERT(rtl8651_addAclRule("vlan2", TRUE, &rule) == SUCCESS);

				/* Because the arp of this srcMac exists, we find out the corresponding IP address of this srcMac. */
				netif_t = GET_NETWORK_INTERFACE("vlan2", NULL);
				ASSERT(netif_t != NULL);
				arp_t = _rtl8651_getArpEntryByMac(netif_t, &mac1, FALSE);
				ASSERT(arp_t != NULL);								
				rule.srcIpAddr_ 		= arp_t->ipAddr;
				rule.srcIpAddrMask_ 	= 0xffffffff;

				curr = 0;
				/* Verify */
				ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

				if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE ) curr++;
#ifdef RTL865XB_URL_FILTER
				if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/

				bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
				bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
				memcpy(&rule1, &rule, sizeof(rtl8651_tblDrvAclRule_t));
				ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);

				ACL_ENTRY_CMP(&rule1, &rule2);
				curr++;

				/* Remove ACL */
				rule.srcIpAddr_ 		= 0;
				rule.srcIpAddrMask_ 	= 0;

				ASSERT(rtl8651_delAclRule("vlan2", TRUE, &rule) == SUCCESS);

			}

		}
	}
	return SUCCESS;

}

/* Test for "srcMac & dstPort" ACL rule when we don't have the arp entry of this srcMac. */
int32 acl_test16(uint32 caseNo)
{
	uint8 ruleType[] = { RTL8651_ACL_TCP, RTL8651_ACL_UDP};
	uint8 ruleAction[]  = { RTL8651_ACL_DROP, RTL8651_ACL_PERMIT, RTL8651_ACL_CPU };
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	ether_addr_t mac1 = { { 0x00, 0x00, 0xab, 0x11, 0x55, 0x00 } };
	ether_addr_t mMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	rtl8651_tblDrvAclRule_t rule, tmpRule;
	rtl8651_tblDrv_arpEntry_t *arp_t;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	uint32 ruleCount, actionCount;
	int32 curr;
		
	START_TEST_CASE();	
	{
		/*======================================
		  * Test: add/delete srcMac & dstPort ACL rule
		  *======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		for(ruleCount=0; ruleCount<2; ruleCount++) {
			for(actionCount=0; actionCount<3; actionCount++) {
				bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
				switch(ruleType[ruleCount]) {					
				case RTL8651_ACL_TCP:
					rule.ruleType_ 		= RTL8651_ACL_TCP;
					rule.tcpl2srcMac_		= mac1;
					rule.tcpl2srcMacMask_	= mMask;
					rule.tcpSrcPortUB_		= 6000;
					rule.tcpSrcPortLB_		= 7000;
					rule.tcpDstPortUB_		= 10000;
					rule.tcpDstPortLB_		= 20000;
					break;

				case RTL8651_ACL_UDP:
					rule.ruleType_ 		= RTL8651_ACL_UDP;
					rule.udpl2srcMac_		= mac1;
					rule.udpl2srcMacMask_	= mMask;
					rule.udpSrcPortUB_	= 1000;
					rule.udpSrcPortLB_		= 20000;
					rule.udpDstPortUB_	= 3000;
					rule.udpDstPortLB_		= 40000;
					break;

				default: ASSERT(0);

				}
				rule.actionType_ = ruleAction[actionCount];
				
				ASSERT(rtl8651_addAclRule("vlan2", TRUE, &rule) == SUCCESS);

				//===============================================================
				curr = 0;
				
				if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE ) curr++;
#ifdef RTL865XB_URL_FILTER
				if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/

				//===============================================================
				bzero(&tmpRule, sizeof(rtl8651_tblDrvAclRule_t));
				tmpRule.ruleType_ 	= RTL8651_ACL_MAC;
				tmpRule.actionType_	= RTL8651_ACL_CPU;
				memcpy(&tmpRule.srcMac_, &mac1, 6); 
				memcpy(&tmpRule.srcMacMask_, &mMask, 6);				

				/* Verify */
				ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

				bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
				bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
				memcpy(&rule1, &tmpRule, sizeof(rtl8651_tblDrvAclRule_t));
				ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);

				ACL_ENTRY_CMP(&rule1, &rule2);

				/* Remove ACL */
				ASSERT(rtl8651_delAclRule("vlan2", TRUE, &rule) == SUCCESS);

			}

		}
	}
	return SUCCESS;

}

/* Test for "srcMac & dstPort" ACL rule. */
int32 acl_test17(uint32 caseNo)
{
	uint8 ruleType[] = { RTL8651_ACL_TCP, RTL8651_ACL_UDP};
	uint8 ruleAction[]  = { RTL8651_ACL_DROP, RTL8651_ACL_PERMIT, RTL8651_ACL_CPU };
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x01 } };
	ether_addr_t mac1 = { { 0x00, 0x00, 0xab, 0x11, 0x55, 0x00 } };
	ether_addr_t mMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	rtl8651_tblDrvAclRule_t rule, tmpRule;
	rtl8651_tblDrv_arpEntry_t *arp_t;
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	uint32 ruleCount, actionCount;
	int32 curr;
	ipaddr_t ipAddr = ntohl(inet_addr("192.168.2.200"));
		
	START_TEST_CASE();	
	{
		/*======================================
		  * Test: add/delete srcMac & dstPort ACL rule
		  *======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		for(ruleCount=0; ruleCount<2; ruleCount++) {
			for(actionCount=0; actionCount<3; actionCount++) {
				bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
				switch(ruleType[ruleCount]) {					
				case RTL8651_ACL_TCP:
					rule.ruleType_ 		= RTL8651_ACL_TCP;
					rule.tcpl2srcMac_		= mac1;
					rule.tcpl2srcMacMask_	= mMask;
					rule.tcpSrcPortUB_		= 6000;
					rule.tcpSrcPortLB_		= 7000;
					rule.tcpDstPortUB_		= 10000;
					rule.tcpDstPortLB_		= 20000;
					break;

				case RTL8651_ACL_UDP:
					rule.ruleType_ 		= RTL8651_ACL_UDP;
					rule.udpl2srcMac_		= mac1;
					rule.udpl2srcMacMask_	= mMask;
					rule.udpSrcPortUB_	= 1000;
					rule.udpSrcPortLB_		= 20000;
					rule.udpDstPortUB_	= 3000;
					rule.udpDstPortLB_		= 40000;
					break;

				default: ASSERT(0);

				}
				rule.actionType_ = ruleAction[actionCount];
				
				ASSERT(rtl8651_addAclRule("vlan2", TRUE, &rule) == SUCCESS);

				//===============================================================
				curr = 0;
				
				if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE ) curr++;
#ifdef RTL865XB_URL_FILTER
				if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/

				//===============================================================
				bzero(&tmpRule, sizeof(rtl8651_tblDrvAclRule_t));
				tmpRule.ruleType_ 	= RTL8651_ACL_MAC;
				tmpRule.actionType_	= RTL8651_ACL_CPU;
				memcpy(&tmpRule.srcMac_, &mac1, 6); 
				memcpy(&tmpRule.srcMacMask_, &mMask, 6);				

				/* Verify */
				ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

				bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
				bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
				memcpy(&rule1, &tmpRule, sizeof(rtl8651_tblDrvAclRule_t));
				ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);

				ACL_ENTRY_CMP(&rule1, &rule2);

				//================================================================

				/* Add arp */
				ASSERT(rtl8651_addArp(ipAddr, &mac1, "vlan2", 0x02) == SUCCESS);				

				/* Because the arp of this srcMac exists, we find out the corresponding IP address of this srcMac. */
				netif_t = GET_NETWORK_INTERFACE("vlan2", NULL);
				ASSERT(netif_t != NULL);
				arp_t = _rtl8651_getArpEntryByMac(netif_t, &mac1, FALSE);
				ASSERT(arp_t != NULL);
				
				rule.srcIpAddr_ 		= arp_t->ipAddr;
				rule.srcIpAddrMask_ 	= 0xffffffff;

				/* Verify ASIC ACL table */
				ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

				bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
				bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
				memcpy(&rule1, &rule, sizeof(rtl8651_tblDrvAclRule_t));
				ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);

				ACL_ENTRY_CMP(&rule1, &rule2);

				//================================================================
				
				/* Delete arp */
				ASSERT(rtl8651_delArp(ipAddr) == SUCCESS);
				bzero(&tmpRule, sizeof(rtl8651_tblDrvAclRule_t));
				tmpRule.ruleType_ 	= RTL8651_ACL_MAC;
				tmpRule.actionType_	= RTL8651_ACL_CPU;
				memcpy(&tmpRule.srcMac_, &mac1, 6); 
				memcpy(&tmpRule.srcMacMask_, &mMask, 6);				

				/* Verify */
				ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

				bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
				bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
				memcpy(&rule1, &tmpRule, sizeof(rtl8651_tblDrvAclRule_t));
				ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);

				ACL_ENTRY_CMP(&rule1, &rule2);

				//================================================================
				/* Remove ACL */
				rule.srcIpAddr_ 		= 0;
				rule.srcIpAddrMask_ 	= 0;

				ASSERT(rtl8651_delAclRule("vlan2", TRUE, &rule) == SUCCESS);

			}

		}
	}
	return SUCCESS;

}

int32 acl_test18(uint32 caseNo)
{
	uint8 ruleType[] = { RTL8651_ACL_TCP, RTL8651_ACL_UDP, RTL8651_ACL_IP, RTL8651_ACL_MAC, RTL8651_ACL_ICMP };
	uint8 ruleAction[]  = { RTL8651_ACL_DROP, RTL8651_ACL_PERMIT, RTL8651_ACL_CPU };
	ether_addr_t mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x00 } };
	ether_addr_t bak_mac = { { 0x00, 0x33, 0x44, 0x55, 0x66, 0x00 } };
	ether_addr_t mMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	rtl8651_tblDrvAclRule_t rule;
	uint32 ruleCount, actionCount, i;
	int32 curr;
	
	
	START_TEST_CASE();	
	{
		/*======================================
		  * Test: add 200 ACL rules
		  *======================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
#if 0
		ether_addr_t Gw_mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x02 } };
		ipaddr_t ipaddr, mask;

		/* VLAN 2: subnet 2 */
		ASSERT(rtl8651_addVlan(2) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(2, &Gw_mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan2") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan2", RTL8651_LL_VLAN, 2) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.2.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("vlan2", ipaddr, mask) == SUCCESS);

		/* VLAN 1: subnet 1 */	
		mac.octet[5] = 0x01;
		ASSERT(rtl8651_assignVlanMacAddress(1, &Gw_mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan1") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan1", RTL8651_LL_VLAN, 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("vlan1", ipaddr, mask) == SUCCESS);
#endif

		/* Add 200 'ethernet' type egress rules */
		for (i = 0; i < 200; i++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_ 	= RTL8651_ACL_MAC;
			rule.actionType_ 	= RTL8651_ACL_DROP;
			memcpy(&rule.dstMac_, &mac, 6);
			memcpy(&rule.dstMacMask_, &mMask, 6); mac.octet[5]++;
			memcpy(&rule.srcMac_, &mac, 6); 
			memcpy(&rule.srcMacMask_, &mMask, 6);
			rule.typeLen_			= 0x08;
			rule.typeLenMask_		= 0xff;
			ASSERT(rtl8651_addAclRule("vlan1", TRUE, &rule) == SUCCESS);
		}

		for (i = 0; i < 100; i++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_ 	= RTL8651_ACL_MAC;
			rule.actionType_ 	= RTL8651_ACL_DROP;
			memcpy(&rule.dstMac_, &bak_mac, 6);
			memcpy(&rule.dstMacMask_, &mMask, 6); bak_mac.octet[5]++;
			memcpy(&rule.srcMac_, &bak_mac, 6); 
			memcpy(&rule.srcMacMask_, &mMask, 6);
			rule.typeLen_			= 0x08;
			rule.typeLenMask_		= 0xff;
			ASSERT(rtl8651_delAclRule("vlan1", TRUE, &rule) == SUCCESS);
		}

		/* Verify */
		curr = 0;

#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/

		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);

		for (i = 0; i < 100; i++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_ 	= RTL8651_ACL_MAC;
			rule.actionType_ 	= RTL8651_ACL_DROP;
			memcpy(&rule.dstMac_, &bak_mac, 6);
			memcpy(&rule.dstMacMask_, &mMask, 6); bak_mac.octet[5]++;
			memcpy(&rule.srcMac_, &bak_mac, 6); 
			memcpy(&rule.srcMacMask_, &mMask, 6);
			rule.typeLen_			= 0x08;
			rule.typeLenMask_		= 0xff;

			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			memcpy(&rule1, &rule, sizeof(rtl8651_tblDrvAclRule_t));

			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart + curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}

	}

	return SUCCESS;
}

int32 acl_test19(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t rule1, rule2;
	rtl8651_tblDrvAclRule_t rule;
	int32 round;
	int32 NumToBeAdded;
	int32 curr;
	ether_addr_t mac_vlan1 = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
	NumToBeAdded = 118;
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB

	if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		NumToBeAdded --;

#ifdef RTL865XB_URL_FILTER
	if(_urlFilter.urlFilterCount >= 0)
		NumToBeAdded --;
#endif/*RTL865XB_URL_FILTER*/
	
	START_TEST_CASE();	
	{
		/*================================================
		  * Test: Test IP Range ACL in out of asic space condition when default drop
		  *================================================ */
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add 118 UDP rule not include ip range acl */
		for(round=0; round<NumToBeAdded; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_UDP;
			rule.actionType_		= RTL8651_ACL_CPU;
			rule.udpSrcPortUB_	= 6000 + round;
			rule.udpSrcPortLB_		= 7000 + round;
			rule.udpDstPortUB_	= 10000 + round;
			rule.udpDstPortLB_		= 20000 + round;
			ASSERT(rtl8651_addAclRule("vlan1", TRUE, &rule) == SUCCESS);
		}
		ASSERT(rtl8651_setDefaultAcl("vlan1", RTL8651_ACL_DROP) == SUCCESS);

		curr = 0;
		/* Verify ACL rule */
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// use Protocol-Based VLAN, not ACL
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		//0x8863
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_MAC;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.typeLen_			= 0x8863;
			rule1.typeLenMask_		= 0xffff;
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclStart+curr, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);
			curr++;
		}
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) curr++;
#endif/*RTL865XB_URL_FILTER*/
		for(round=(asic_vlan.inAclStart+curr); round<NumToBeAdded; round++) {
			bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
			rule1.ruleType_		= RTL8651_ACL_UDP;
			rule1.actionType_		= RTL8651_ACL_CPU;
			rule1.udpSrcPortUB_		= 6000 + round-(asic_vlan.inAclStart+curr);
			rule1.udpSrcPortLB_		= 7000 + round-(asic_vlan.inAclStart+curr);
			rule1.udpDstPortUB_		= 10000 + round-(asic_vlan.inAclStart+curr);
			rule1.udpDstPortLB_		= 20000 + round-(asic_vlan.inAclStart+curr);
			ASSERT(rtl8651_getAsicAclRule(round, &rule2) == SUCCESS);
			ACL_ENTRY_CMP(&rule1, &rule2);

		}
		curr+=NumToBeAdded;
		//default rule: drop -- no need to compare
		#if 0
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_DROP;
		ASSERT(rtl8651_getAsicAclRule(curr, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		#endif

		/* add ip Range ACL: this ip range ACL will replace the original default rule. */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddrStart_	= ntohl(inet_addr("140.112.1.10"));
		rule.dstIpAddrEnd_		= ntohl(inet_addr("140.112.1.200"));
		rule.tcpSrcPortUB_		= 0xffff;
		rule.tcpSrcPortLB_		= 0x0000;
		rule.tcpDstPortUB_		= 0xffff;
		rule.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_addAclRuleExt("vlan1", TRUE, &rule) == SUCCESS);

		// Verify ip range ACL
		// 192.168.1.10/31
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.10"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.254"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.12/30
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.12"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.252"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;
		
		// 192.168.1.16/28
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.16"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.240"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.32/27
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.32"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.224"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		// 192.168.1.64/26
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_			= RTL8651_ACL_TCP;
		rule1.actionType_			= RTL8651_ACL_CPU;
		rule1.dstIpAddr_			= ntohl(inet_addr("140.112.1.64"));
		rule1.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.192"));
		rule1.tcpSrcPortUB_		= 0xffff;
		rule1.tcpSrcPortLB_		= 0x0000;
		rule1.tcpDstPortUB_		= 0xffff;
		rule1.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+curr), &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		curr++;

		//last two rules are changed to "L2 permit" & "CPU" because rule count exceed */
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_PERMIT;
		rule1.pktOpApp		= 1;
		ASSERT(rtl8651_getAsicAclRule(123, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);
		
		bzero(&rule1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&rule2, sizeof(_rtl8651_tblDrvAclRule_t));
		rule1.ruleType_		= RTL8651_ACL_MAC;
		rule1.actionType_		= RTL8651_ACL_CPU;
		ASSERT(rtl8651_getAsicAclRule(124, &rule2) == SUCCESS);
		ACL_ENTRY_CMP(&rule1, &rule2);

		//Egress
		ASSERT(asic_vlan.outAclEnd == RTL8651_ACLTBL_DROP_ALL);
		ASSERT(asic_vlan.outAclStart == RTL8651_ACLTBL_DROP_ALL);

		/* Verify VLAN 2 */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

		//Ingress:
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) {
			// Because ASIC ACL table is full, 
			// last two rules are changed to "L2 permit" & "CPU" because rule count exceed
			ASSERT(asic_vlan.inAclStart == 123);
			ASSERT(asic_vlan.inAclEnd == 124);
		}
#else	/* RTL865XB_URL_FILTER */

#if 1//CONFIG_RTL865XB_EXP_PPPOE_PASSTHRU
		// PPPoE uses Protocol-Based VLAN, not ACL
		ASSERT(asic_vlan.inAclStart == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.inAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
#else// CONFIG_RTL865XB
#endif// CONFIG_RTL865XB
		if ( rtl8651_l2protoPassThrough & PASSTHRU_PPPOE )
		{
			ASSERT(asic_vlan.inAclStart == 123);
			ASSERT(asic_vlan.inAclEnd == 124);
		}
		else
		{
			ASSERT(asic_vlan.inAclStart == RTL8651_ACLTBL_PERMIT_ALL);
			ASSERT(asic_vlan.inAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
		}
#endif	/* RTL865XB_URL_FILTER */


		//Egress:
		ASSERT(asic_vlan.outAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
		ASSERT(asic_vlan.outAclStart == RTL8651_ACLTBL_PERMIT_ALL);

		/* remove the ip range acl rule */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_TCP;
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.dstIpAddrStart_	= ntohl(inet_addr("140.112.1.10"));
		rule.dstIpAddrEnd_		= ntohl(inet_addr("140.112.1.200"));
		rule.tcpSrcPortUB_		= 0xffff;
		rule.tcpSrcPortLB_		= 0x0000;
		rule.tcpDstPortUB_		= 0xffff;
		rule.tcpDstPortLB_		= 0x0000;
		ASSERT(rtl8651_delAclRuleExt("vlan1", TRUE, &rule) == SUCCESS);		
	}
	return SUCCESS;
}


#define RTL865X_MULTICAST_TESTING_CODE


/*
@func int32 | muticast_test1 | Multicast testing.
@comm
Test multicast addition and deletion.
*/

int32 muticast_test1(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip;
	uint32 entry;

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: add Multicast
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add IP Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0c, 0, FALSE, 0) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 1;
		asic_mcast1.port	= 0;
		asic_mcast1.mbr	= 0x0c;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
		ASSERT(CTAILQ_NEXT(mCast_t, nextMCast) == NULL);

		/* (2) add Multicast member port */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0e, 0, FALSE, 0) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 1;
		asic_mcast1.port	= 0;
		asic_mcast1.mbr	= 0x0e;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
		ASSERT(CTAILQ_NEXT(mCast_t, nextMCast) == NULL);
	}

	{
		/*==============================================
		  * Test: del Multicast
		  *==============================================*/
		 /* (1) remove member port 0x02 */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_delMulticast(mAddr, 0x04, 0x02) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 1;
		asic_mcast1.port	= 0;
		asic_mcast1.mbr	= 0x0c;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
		ASSERT(CTAILQ_NEXT(mCast_t, nextMCast) == NULL);
	
		/* (2) remove member port 0x0c */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_delMulticast(mAddr, 0x04, 0x0c) == SUCCESS);
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT(CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry]) == NULL);
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) != SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | muticast_test2 | Multicast testing.
@comm
Test if pending and refill mechanism is workable.
*/

int32 muticast_test2(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip;
	uint32 entry, index=0;
	uint8 array[4];

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: pending and refill
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add IP Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.1"));
		array[index++] = sip&0xff; 

		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, (1<<(index+1)), 0, FALSE, 0) == SUCCESS);
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		
		while(((++sip)&0xff) < 255) {	
			if (rtl8651_ipMulticastTableIndex(sip, mAddr) != entry) 
				continue;	
			array[index++] = sip&0xff;
			ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, (1<<(index+1)), 0, FALSE, 0) == SUCCESS);
		}

		/* Verify */
		index = 0;
		sip = ntohl(inet_addr("192.168.1.0"));
		sip |= array[index];
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 1;
		asic_mcast1.port	= 0;
		asic_mcast1.mbr	= (1 << (index+2));
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);

		/* (2) remove IP Multicast */
		ASSERT(_rtl8651_delMulticast(mAddr, 0x04, (1<<2)) == SUCCESS);
		index = 1;
		while (index < 4) {
			sip = ntohl(inet_addr("192.168.1.0"));
			sip |= array[index];
			
			/* Verify */
			bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
			asic_mcast1.sip	= sip;
			asic_mcast1.dip	= mAddr;
			asic_mcast1.svid	= 1;
			asic_mcast1.port	= 0;
			asic_mcast1.mbr	= (1<<(index+2));
			asic_mcast1.extIdx	= 0;
			asic_mcast1.age	= 10;
			asic_mcast1.cpu	= 0;
			ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
			MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);

			ASSERT(_rtl8651_delMulticast(mAddr, 0x04, (1<<(index+2))) == SUCCESS);
			index ++;			
				
		}
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) != SUCCESS);
		ASSERT(CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry]) == NULL);
	}
	return SUCCESS;
}


/*
@func int32 | muticast_test3 | Multicast testing.
@comm
Test removing multicast entry by the specified multicast address.
*/

int32 muticast_test3(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip;
	uint32 entry, index=0;
	uint8 array[4];

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: remove by IP Multicast address
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add IP Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.1"));
		array[index++] = sip&0xff; 
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, (1<<(index+1)), 0, FALSE, 0) == SUCCESS);
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		
		while(((++sip)&0xff) < 255) {	
			if (rtl8651_ipMulticastTableIndex(sip, mAddr) != entry) 
				continue;	
			array[index++] = sip&0xff;
			ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, (1<<(index+1)), 0, FALSE, 0) == SUCCESS);
		}

		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.2"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0c, 0, FALSE, 0) == SUCCESS);
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.3"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0c, 0, FALSE, 0) == SUCCESS);

		/* (2) remove by Multicast address */
		ASSERT(_rtl8651_delMulticast(mAddr, 0, 0)== SUCCESS);

		/* Verify */
		for(index=0; index<64; index++) {
			ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) != SUCCESS);
			ASSERT(CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry]) == NULL);
		}
	}

	{
		/* (1) add IP Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.1"));
		array[index++] = sip&0xff;
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0e, 0, FALSE, 0) == SUCCESS);
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		
		while(((++sip)&0xff) < 255) {	
			if (rtl8651_ipMulticastTableIndex(sip, mAddr) != entry) 
				continue;	
			array[index++] = sip&0xff;
			ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0e, 0, FALSE, 0) == SUCCESS);
		}

		/* (2) remove by Multicast address */
		ASSERT(_rtl8651_delMulticast(mAddr, 0x04, 0x0e)== SUCCESS);

		/* Verify */
		for(index=0; index<64; index++) {
			ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) != SUCCESS);
			ASSERT(CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry]) == NULL);
		}
	}
	return SUCCESS;
}



/*
@func int32 | muticast_test4 | Multicast testing.
@comm
Test Multicast Internal API: _rtl8651_delMulticastUpStream()
*/

int32 muticast_test4(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip;
	uint32 entry;

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: _rtl8651_delMulticastUpStream()
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add IP Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0e, 0, FALSE, 0) == SUCCESS);
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);

		/* (2) remove */
		ASSERT(_rtl8651_delMulticastUpStream(mAddr, sip, 1, 0) == SUCCESS);
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) != SUCCESS);
		ASSERT(CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry]) == NULL);
	}
	return SUCCESS;
}



/*
@func int32 | muticast_test5 | adding a entry with toCPU bit set.
@comm
Adding a multicast entry with toCPU bit set. The entry should not be written into ASIC.
*/

int32 muticast_test5(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip;
	uint32 entry;

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: adding an entry with toCPU bit set
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add IP Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0e, 0, TRUE, 0) == SUCCESS);
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);

		/* Verify: should be empty entry */
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast1) != SUCCESS);

		/* (2) add an entry with no toCPU set */
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0e, 0, FALSE, 0) == SUCCESS);
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 1;
		asic_mcast1.port	= 0;
		asic_mcast1.mbr	= 0x0e;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);

		/* (3) remove the entry with no toCPU set */
		ASSERT(_rtl8651_delMulticastUpStream(mAddr, sip, 1, 0) == SUCCESS);
		/* Verify: should be empty entry */
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast1) != SUCCESS);
	}
	return SUCCESS;
}


/*
@func int32 | muticast_test6 | Swapping mechanism.
@comm
Heavy load entry will be swapped to ASIC.
*/

int32 muticast_test6(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip;
	uint32 entry, count;

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: swapping mechanism
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add IP Multicast : this entry should be written into ASIC */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0e, 0, FALSE, 0) == SUCCESS);
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);

		/* (2) this entry will not be written into ASIC */
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 6, 0x04, 0x0c, 0, FALSE, 0) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 1;
		asic_mcast1.port	= 0;
		asic_mcast1.mbr	= 0x0e;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);

		for(count=0; count<2000; count++) {
			ASSERT(_rtl8651_findMCast(mAddr, sip, 1, 6) != NULL);
		}

		/* (3) Now the second entry is in the ASIC */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 1;
		asic_mcast1.port	= 6;
		asic_mcast1.mbr	= 0x0c;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
	}
	return SUCCESS;
}


/*
@func int32 | muticast_test7 | Test _rtl8651_setMGroupAttribute() API.
@comm
Test _rtl8651_setMGroupAttribute() API.
*/

int32 muticast_test7(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip;
	uint32 entry, count;

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: _rtl8651_setMGroupAttribute()
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add IP Multicast : this entry should be written into ASIC */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 0, 0x04, 0x0e, 0, FALSE, 0) == SUCCESS);
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);

		/* (2) this entry will not be written into ASIC */
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 1, 6, 0x04, 0x0c, 0, FALSE, 0) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 1;
		asic_mcast1.port	= 0;
		asic_mcast1.mbr	= 0x0e;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);

		/* (3) set specified group to toCPU */
		ASSERT(_rtl8651_setMGroupAttribute(mAddr, TRUE) == SUCCESS);
		/* Verify: should be empty entry */
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast1) != SUCCESS);

		/* (4) not set to CPU */
		ASSERT(_rtl8651_setMGroupAttribute(mAddr, FALSE) == SUCCESS);
		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 1;
		asic_mcast1.port	= 0;
		asic_mcast1.mbr	= 0x0e;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
	}
	return SUCCESS;
}



/*
@func int32 | muticast_test8 | Test aging and swapping mechanism
@comm
Test aging and swapping mechanism.
*/

int32 muticast_test8(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip;
	uint32 entry, count, round, sel_port, times;
	int32 array[5] = { 0 };

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: aging test
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		srandom( caseNo );

		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);

		/* (1) add 5 entry */
		for(count=0; count<5; count++) {
			ASSERT(_rtl8651_addMulticast(mAddr, sip, 2, (count+1), 0x02, 0x21, 0, FALSE, 0) == SUCCESS);
		}
		/* Verify */
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= 1;
		asic_mcast1.mbr	= 0x21;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
		times = 0;
		
again:
		for(round=0; round<10000; round++) {
			sel_port = ((200 + (uint32)(2000.0 * rand() / (RAND_MAX + 1.0))) % 5) + 1;
			ASSERT(_rtl8651_findMCast(mAddr, sip, 2, sel_port) != NULL);
			array[sel_port - 1] ++;
		}
		for(count=round=0; round<5; round++) {	
			if (array[round] > count) {
				count = array[round];
				sel_port = round + 1;
			}
		}
		/* Verify */
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= sel_port;
		asic_mcast1.mbr	= 0x21;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);

		_rtl8651_timeUpdateMulticast(1);
		for(count=0; count<5; count++)
			array[count] = 0;
		if (times < 100) { 
			times++;
			goto again;
		}
		_rtl8651_timeUpdateMulticast(35);
		/* Verify */
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= sel_port;
		asic_mcast1.mbr	= 0x21;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
	}
	return SUCCESS;
}


/*
@func int32 | muticast_test9 | Test external IP setting
@comm
Test external IP setting.
*/

int32 muticast_test9(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip, extip;
	uint32 entry;

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: add Multicast
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add IP Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		extip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 2, 1, 0x02, 0x01, extip, FALSE, 0) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= 1;
		asic_mcast1.mbr	= 0x01;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
	}

	{
		/*==============================================
		  * Test: del Multicast
		  *==============================================*/
		 /* (1) remove member port 0x01 */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		extip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(_rtl8651_delMulticast(mAddr, 0x02, 0x01) == SUCCESS);

		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT(CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry]) == NULL);
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) != SUCCESS);
	}
	return SUCCESS;
}

/*
@func int32 | muticast_test10 | Test upload flag
@comm
Test upload flag in addMulticast().
*/

int32 muticast_test10(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip, extip;
	uint32 entry;

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: add Multicast
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add IP Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		extip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 2, 1, 0x02, 0x01, extip, FALSE, RTL865X_MULTICAST_UPLOADONLY) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= 1;
		asic_mcast1.mbr	= 0x01;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
	}

	{
		/*==============================================
		  * Test: add Multicast member
		  *==============================================*/
		/* (1) add IP Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		extip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(_rtl8651_addMulticastMember(mAddr, 0x04, 0x0c) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= 1;
		asic_mcast1.mbr	= 0x0c;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
	}

	{
		/*==============================================
		  * Test: del Multicast
		  *==============================================*/
		 /* (1) remove member port 0x01 */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		extip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(_rtl8651_delMulticast(mAddr, 0x04, 0x0c) == SUCCESS);

		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT(CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry]) == NULL);
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) != SUCCESS);
	}
	return SUCCESS;
}

/*
@func int32 | muticast_test11 | Multicast member addition testing.
@comm
Test multicast member addition.
*/

int32 muticast_test11(uint32 caseNo)
{
	rtl8651_tblDrv_mCast_t *mCast_t;
	rtl865x_tblAsicDrv_multiCastParam_t asic_mcast1, asic_mcast2;
	ipaddr_t mAddr, sip;
	uint32 entry;

	START_TEST_CASE();

	{
		/*==============================================
		  * Test: add Multicast
		  *==============================================*/
		default_configuration5(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add IP Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 2, 1, 0x04, 0x10, 0, FALSE, 0) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= 1;
		asic_mcast1.mbr	= 0x10;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);

		/* (2) add Multicast */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		/* this entry don't set into ASIC */
		ASSERT(_rtl8651_addMulticast(mAddr, sip, 2, 2, 0x04, 0x08, 0, FALSE, 0) == SUCCESS);

		/* Verify: only first entry in ASIC */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= 1;
		asic_mcast1.mbr	= 0x10;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
		ASSERT(CTAILQ_NEXT(mCast_t, nextMCast) != NULL);
	}

	{
		/*==============================================
		  * Test: add Multicast member
		  *==============================================*/

		/* add IP Multicast member */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_addMulticastMember(mAddr, 0x04, 0x10) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= 1;
		asic_mcast1.mbr	= 0x10;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);

	}

	{
		/*==============================================
		  * Test: del Multicast
		  *==============================================*/
		 /* (1) remove member port 0x10 */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_delMulticast(mAddr, 0x04, 0x10) == SUCCESS);

		/* Verify */
		/* first entry is removed, second entry exists */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= 2;
		asic_mcast1.mbr	= 0x08;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
		ASSERT(CTAILQ_NEXT(mCast_t, nextMCast) == NULL);
	}

	{
		/*==============================================
		  * Test: add Multicast member
		  *==============================================*/

		/* add IP Multicast member */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_addMulticastMember(mAddr, 0x04, 0x10) == SUCCESS);

		/* Verify */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= 2;
		asic_mcast1.mbr	= 0x18;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);

	}

	{
		/*==============================================
		  * Test: del Multicast
		  *==============================================*/
		 /* (1) remove member port 0x08 */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_delMulticast(mAddr, 0x04, 0x08) == SUCCESS);

		/* Verify */
		/* first entry is removed, second entry exists */
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT((mCast_t = CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry])));
		bzero(&asic_mcast1, sizeof(rtl865x_tblAsicDrv_multiCastParam_t));
		asic_mcast1.sip	= sip;
		asic_mcast1.dip	= mAddr;
		asic_mcast1.svid	= 2;
		asic_mcast1.port	= 2;
		asic_mcast1.mbr	= 0x10;
		asic_mcast1.extIdx	= 0;
		asic_mcast1.age	= 10;
		asic_mcast1.cpu	= 0;
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) == SUCCESS);
		MCAST_ENTRY_CMP(&asic_mcast1, &asic_mcast2);
		ASSERT(CTAILQ_NEXT(mCast_t, nextMCast) == NULL);

		/* (2) remove member port 0x10 */
		mAddr = ntohl(inet_addr("224.0.0.3"));
		sip = ntohl(inet_addr("192.168.1.100"));
		ASSERT(_rtl8651_delMulticast(mAddr, 0x04, 0x10) == SUCCESS);
		entry = rtl8651_ipMulticastTableIndex(sip, mAddr);
		ASSERT(CTAILQ_FIRST(&DrvTbl.inuseList.mCastTbl[entry]) == NULL);
		ASSERT(rtl8651_getAsicIpMulticastTable(entry, &asic_mcast2) != SUCCESS);
	}
	return SUCCESS;
}



#define RTL865X_MISC_TESTING_CODE

/*
@func int32 | misc_test1 | Test Network interface API.
@comm
Test rtl8651_specifyNetworkIntfLinkLayerType() and rtl8651_removeNetworkIntfLinkLayerType() APIs.
*/

int32 misc_test1(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };


	START_TEST_CASE();
	{
		/*=================================================
		  * Test: API rtl8651_specifyNetworkIntfLinkLayerType()
		  *	     rtl8651_removeNetworkIntfLinkLayerType()
		  *=================================================*/
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("net1") == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("net2") == SUCCESS);

		/* Note: Disallow more than one Session-Based to attach to single VLAN */
		
		/* (1) VLAN type */
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net1", RTL8651_LL_VLAN, 1) == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_VLAN, 1) != SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net1") == SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net2") != SUCCESS);

		/* (1.1) add PPPoE PPTP L2TP */
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net1", RTL8651_LL_VLAN, 1) == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_PPPOE, 1) != SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_PPTP, 1) != SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_L2TP, 1) != SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net1") == SUCCESS);


		
	
		/* Note: allow more than one  Session-Based (should all the same) to attach to single VLAN */
		
		/* (2) PPPoE type */
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net1", RTL8651_LL_PPPOE, 1) == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_PPPOE, 1) == SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net1") == SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net2") == SUCCESS);

		/* (2.1) add VLAN PPTP L2TP */
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net1", RTL8651_LL_PPPOE, 1) == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_VLAN, 1) != SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_PPTP, 1) != SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_L2TP, 1) != SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net1") == SUCCESS);
		

		/* (3) PPTP type */
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net1", RTL8651_LL_PPTP, 1) == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_PPTP, 1) == SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net1") == SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net2") == SUCCESS);

		/* (3.1) add VLAN PPPOE L2TP */
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net1", RTL8651_LL_PPTP, 1) == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_VLAN, 1) != SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_PPPOE, 1) != SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_L2TP, 1) != SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net1") == SUCCESS);
		

		/* (4) L2TP type */
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net1", RTL8651_LL_L2TP, 1) == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_L2TP, 1) == SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net1") == SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net2") == SUCCESS);

		/* (4.1) add VLAN PPPOE PPTP */
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net1", RTL8651_LL_L2TP, 1) == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_VLAN, 1) != SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_PPPOE, 1) != SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("net2", RTL8651_LL_PPTP, 1) != SUCCESS);
		ASSERT(rtl8651_removeNetworkIntfLinkLayerType("net1") == SUCCESS);
		
	}
	return SUCCESS;
}


/*
@func int32 | misc_test2 | Switch mode
@comm
Test switch mode auto configuration. (L2/L3/L4)
*/

int32 misc_test2(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	ipaddr_t ipaddr, mask;
	extern int32 rtl8651_operationlayer;

	
	START_TEST_CASE();

	{
		/*===============================================
		  * Test: configure switch layer 
		  *===============================================*/
		  
		/* (0) Initially, switch is configured to layer-2 switch */
		ASSERT(rtl8651_operationlayer == 2);

		/* (1) After adding network interface, the switch is still layer-2 switch */
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan", RTL8651_LL_VLAN, 1) == SUCCESS);
		ASSERT(rtl8651_operationlayer == 2);

		/* (3) After adding IP interface, the switch is changed to layer-3 mode */
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("vlan", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_operationlayer == 3);

		
	}
	return SUCCESS;
}



/*
@func int32 | misc_test3 | VLAN MAC address assignment test.
@comm
Test VLAN MAC Address assignment.
*/
int32 misc_test3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x01 } };
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan1, asic_vlan2;
	uint32 round;

	
	START_TEST_CASE();

	{
		/*===============================================
		  * Test: assign VLAN MAC address
		  *===============================================*/
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan", RTL8651_LL_VLAN, 1) == SUCCESS);
		for(round=0; round<50; round++) {
			ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
			bzero(&asic_vlan1, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
			bzero(&asic_vlan2, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
			asic_vlan1.broadcastToCpu	= 1; //always sent pkt to CPU
			asic_vlan1.DMZFlag			= 0;
			asic_vlan1.enableRoute		= 0;
			asic_vlan1.inAclEnd			= RTL8651_ACLTBL_PERMIT_ALL;
			asic_vlan1.inAclStart		= RTL8651_ACLTBL_PERMIT_ALL;
			asic_vlan1.internal			= 1;
			asic_vlan1.macAddr			= mac;
			asic_vlan1.macAddrNumber	= 1;
			asic_vlan1.memberPortMask	= 0x1ff; //0x3f; /* 8650B: extension port */
			asic_vlan1.mtu				= 1500;
			asic_vlan1.outAclEnd		= RTL8651_ACLTBL_PERMIT_ALL;
			asic_vlan1.outAclStart		= RTL8651_ACLTBL_PERMIT_ALL;
			asic_vlan1.promiscuous		= 0;
			asic_vlan1.untagPortMask	= 0x1ff; //0x3f;
			asic_vlan1.valid			= 1; 
			asic_vlan1.vid				= 1;
			memset(&asic_vlan1.portState, RTL8651_PORTSTA_FORWARDING, sizeof(int8)*(RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum));
			ASSERT(rtl8651_getAsicVlan(1,  &asic_vlan2) == SUCCESS);
			VLAN_ENTRY_CMP(&asic_vlan1, &asic_vlan2);
			mac.octet[5] ++;
		}

		mac.octet[5] = 0x01;
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 2) != SUCCESS);
		mac.octet[5] = 0x02;
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 2) == SUCCESS);
		asic_vlan1.macAddr = mac;
		asic_vlan1.macAddrNumber = 2;
		ASSERT(rtl8651_getAsicVlan(1,  &asic_vlan2) == SUCCESS);
		VLAN_ENTRY_CMP(&asic_vlan1, &asic_vlan2);
	}
	return SUCCESS;
}

int32 misc_test4(uint32 caseNo)
{
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t ipaddr, mask;
	uint32 entry;
	uint16 vidx;
	

	START_TEST_CASE();

	{
		/*========================================
		  * Test: _rtl8651_getOneVidx() API.
		  *========================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		
		vidx = _rtl8651_getOneVidx(TRUE);
		ASSERT(vidx == rtl8651_vlanTableIndex(1));
		vidx = _rtl8651_getOneVidx(FALSE);
		ASSERT(vidx == rtl8651_vlanTableIndex(2));
	}
	return SUCCESS;
}






#define RTL865X_MTU_TESTING_CODE


int32 mtu_test1(uint32 caseNo)
{
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;

	
	START_TEST_CASE();

	{
		/*===============================================
		  * Test: default MTU testing
		  *===============================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* Default MTU is 1500 */
		ASSERT((netif_t=GET_NETWORK_INTERFACE("vlan1", NULL)) != NULL);
		ASSERT(netif_t->mtu == 1500);
		ASSERT((netif_t=GET_NETWORK_INTERFACE("vlan1", NULL)) != NULL);
		ASSERT(netif_t->mtu == 1500);
		bzero(&asic_vlan,  sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		ASSERT(asic_vlan.mtu == 1500);
		bzero(&asic_vlan,  sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		ASSERT(asic_vlan.mtu == 1500);
	}
	return SUCCESS;
}


int32 mtu_test2(uint32 caseNo)
{
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;

	
	START_TEST_CASE();

	{
		/*===============================================
		  * Test: set MTU testing
		  *===============================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) set vlan1: mtu */
		ASSERT(rtl8651_setNetMtu("vlan1", 1200) == SUCCESS);
		ASSERT((netif_t=GET_NETWORK_INTERFACE("vlan1", NULL)) != NULL);
		ASSERT(netif_t->mtu == 1200);
		bzero(&asic_vlan,  sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		ASSERT(asic_vlan.mtu == 1200);
		
		/* (2) set vlan2: mtu */
		ASSERT(rtl8651_setNetMtu("vlan2", 1000) == SUCCESS);
		ASSERT((netif_t=GET_NETWORK_INTERFACE("vlan2", NULL)) != NULL);
		ASSERT(netif_t->mtu == 1000);
		bzero(&asic_vlan,  sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		ASSERT(asic_vlan.mtu == 1000);

	}
	return SUCCESS;
}



int32 mtu_test3(uint32 caseNo)
{
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	uint32 dsid;
	uint16 mtu;

	
	START_TEST_CASE();

	{
		/*===============================================
		  * Test: set MTU testing, more than one netwrok interface in a single
		  *          VLAN
		  *===============================================*/
		default_configuration3();
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 1001, "pppoe1", 1) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 1002, "pppoe2", 1) == SUCCESS);

		/* (1) set pppoe mtu */
		ASSERT(rtl8651_setNetMtu("pppoe", 1000) == SUCCESS);
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pppoe", NULL)) != NULL);
		ASSERT(netif_t->mtu == 1000);
		bzero(&asic_vlan,  sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		ASSERT(asic_vlan.mtu == 1000 + 8); /* we need to add PPPoE header size (8), because this is the ASIC behavior */

		/* (2) set pppoe1 mtu */
		ASSERT(rtl8651_setNetMtu("pppoe1", 500) == SUCCESS);
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pppoe1", NULL)) != NULL);
		ASSERT(netif_t->mtu == 500);
		bzero(&asic_vlan,  sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		ASSERT(asic_vlan.mtu == 500+8); /* we need to add PPPoE header size (8), because this is the ASIC behavior */

		/* (1) set pppoe2 mtu */
		ASSERT(rtl8651_setNetMtu("pppoe2", 1200) == SUCCESS);
		ASSERT((netif_t=GET_NETWORK_INTERFACE("pppoe2", NULL)) != NULL);
		ASSERT(netif_t->mtu == 1200);
		bzero(&asic_vlan,  sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		ASSERT(asic_vlan.mtu == 500+8); /* we need to add PPPoE header size (8), because this is the ASIC behavior */

		/* Check pppoe_mss[] value */
		dsid = _rtl8651_SessionIDToDSID(1000);
		ASSERT(pppoe_mss[dsid] == (1000-40));
		dsid = _rtl8651_SessionIDToDSID(1001);
		ASSERT(pppoe_mss[dsid] == (500-40));
		dsid = _rtl8651_SessionIDToDSID(1002);
		ASSERT(pppoe_mss[dsid] == (1200-40));

		/* Test getmtu API */
		ASSERT(rtl8651_getNetMtu("pppoe", &mtu) == SUCCESS);
		ASSERT(mtu == 1000);
		ASSERT(rtl8651_getNetMtu("pppoe1", &mtu) == SUCCESS);
		ASSERT(mtu == 500);
		ASSERT(rtl8651_getNetMtu("pppoe2", &mtu) == SUCCESS);
		ASSERT(mtu == 1200);
	}
	return SUCCESS;
}



int32 mtu_test4(uint32 caseNo)
{
	rtl8651_tblDrv_networkIntfTable_t *netif_t;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	rtl8651_tblDrv_vlanTable_t *vlan_t;
	uint32 entry;

	
	START_TEST_CASE();
	{
		/*===============================================
		  * Test: rtl8651_setMtu()
		  *===============================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		rtl8651_setMtu(1200);
		netif_t = DrvTbl.netIface;
		for(entry=0; entry<rtl8651_tblDrvPara.networkIntfTableSize; entry++, netif_t++) {
			if (netif_t->valid == 0)
				continue;
			vlan_t = &DrvTbl.vlan[rtl8651_vlanTableIndex(netif_t->vid)];
			ASSERT(netif_t->mtu == 1200);
			ASSERT(vlan_t->mtu == 1200);
			bzero(&asic_vlan,  sizeof(rtl865x_tblAsicDrv_vlanParam_t));
			ASSERT(rtl8651_getAsicVlan(netif_t->vid, &asic_vlan) == SUCCESS);
			ASSERT(asic_vlan.mtu == 1200);
		}
	}
	return SUCCESS;
}







#define RTL865X_MAC_ACCESS_CONTROL_TESTING_CODE

/*
@func int32 | macAccessControl_test1 | MAC Access Control testing.
@comm
Test addition and deletion of MAC Access Control entry. Besides that the _rtl8651_isMacAccessControlPermit() internal API
is also tested. Note, this testing code always adds an entry with MAC Mask = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }.
*/

int32 macAccessControl_test1(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x33 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl8651_tblDrv_macACEntry_t *aclRule;


	START_TEST_CASE();
	
	{
		/*=====================================================
		  * Test: add/delete , mask = 0xffffffffffff
		  *=====================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		
		/* (1) add an entry */
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);
		ASSERT(_rtl8651_isMacAccessControlPermit(&mac) == FALSE);
		ASSERT(CTAILQ_TOTAL(&DrvTbl.macACInfo->macAccessControl) == 1);

		/* (2) remove the entry */
		ASSERT(rtl8651_delMacAccessControl(&mac, &macMask) == SUCCESS);
		ASSERT(_rtl8651_isMacAccessControlPermit(&mac) == TRUE);
		ASSERT(CTAILQ_TOTAL(&DrvTbl.macACInfo->macAccessControl) == 0);
	}
	return SUCCESS;
}


/*
@func int32 | macAccessControl_test2 | MAC Access Control testing.
@comm
Test addition and deletion of MAC Access Control entry. Besides that the _rtl8651_isMacAccessControlPermit() internal API
is also tested. Note, this testing code always adds an entry with MAC Mask = { 0xff, 0xff, 0xff, 0xff, 0x00, 0x00 }.
*/

int32 macAccessControl_test2(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x33 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0x00, 0x00 } };
	rtl8651_tblDrv_macACEntry_t *aclRule;


	START_TEST_CASE();
	
	{
		/*=====================================================
		  * Test: add/delete , mask = 0xffffffff0000
		  *=====================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		
		/* (1) add an entry */
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);
		mac.octet[5] = 0xff;
		ASSERT(_rtl8651_isMacAccessControlPermit(&mac) == FALSE);
		ASSERT(CTAILQ_TOTAL(&DrvTbl.macACInfo->macAccessControl) == 1);

		/* (2) remove the entry */
		ASSERT(rtl8651_delMacAccessControl(&mac, &macMask) == SUCCESS);
		mac.octet[5] = 0xff;
		ASSERT(_rtl8651_isMacAccessControlPermit(&mac) == TRUE);
		ASSERT(CTAILQ_TOTAL(&DrvTbl.macACInfo->macAccessControl) == 0);
	}
	return SUCCESS;
}


/*
@func int32 | macAccessControl_test3 | MAC Access Control testing.
@comm
Test Overflow the MAC Access Control Table and then remove it one by one.
*/

int32 macAccessControl_test3(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl8651_tblDrv_macACEntry_t *aclRule;
	int32 entry;


	START_TEST_CASE();
	
	{
		/*=====================================================
		  * Test: add/delete , (overflow table)
		  *=====================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		
		/* (1) overflow table */
		for(entry=0; entry<rtl8651_tblDrvPara.macAccessControllSize; entry++) {
			ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);
			ASSERT(_rtl8651_isMacAccessControlPermit(&mac) == FALSE);
			ASSERT(CTAILQ_TOTAL(&DrvTbl.macACInfo->macAccessControl) == (entry+1));
			mac.octet[5] ++;
		}

		/* (2) remove all entry */
		mac.octet[5] = 0x01;
		for(entry=0; entry<rtl8651_tblDrvPara.macAccessControllSize; entry++) {
			ASSERT(rtl8651_delMacAccessControl(&mac, &macMask) == SUCCESS);
			ASSERT(_rtl8651_isMacAccessControlPermit(&mac) == TRUE);
			ASSERT(CTAILQ_TOTAL(&DrvTbl.macACInfo->macAccessControl) == (rtl8651_tblDrvPara.macAccessControllSize-entry-1));
			mac.octet[5] ++;
		}
	}
	return SUCCESS;
}


/*
@func int32 | macAccessControl_test4 | MAC Access Control testing.
@comm
Test Overflow the MAC Access Control Table and then remove it by using rtl8651_flushMacAccessControl() API.
*/

int32 macAccessControl_test4(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl8651_tblDrv_macACEntry_t *aclRule;
	int32 entry;


	START_TEST_CASE();
	
	{
		/*=====================================================
		  * Test: add / flush
		  *=====================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		
		/* (1) overflow table */
		for(entry=0; entry<rtl8651_tblDrvPara.macAccessControllSize; entry++) {
			ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);
			ASSERT(_rtl8651_isMacAccessControlPermit(&mac) == FALSE);
			ASSERT(CTAILQ_TOTAL(&DrvTbl.macACInfo->macAccessControl) == (entry+1));
			mac.octet[5] ++;
		}
		ASSERT(rtl8651_flushMacAccessControl() == SUCCESS);
		ASSERT(CTAILQ_TOTAL(&DrvTbl.macACInfo->macAccessControl) == 0);
	}
	return SUCCESS;
}



/*
@func int32 | macAccessControl_test5 | MAC Access Control testing.
@comm
Add NAT entry but lack internal host's ARP. This kind of NAT entry never matches Mac Access Control rule, hence it should
be added to ASIC IP Table.
*/

int32 macAccessControl_test5(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip_1, asic_ip_2;
	ipaddr_t extip, intip;
	
	
	START_TEST_CASE();
	
	{
		/*=================================================================
		  * Test: add NAT entry but lack internal host's arp hence it will not match Mac Access Control.
		  *=================================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);

		/* (1)  add NAT Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNatMapping(extip, intip) == SUCCESS);

		/* Verify NAT Table */
		bzero(&asic_ip_1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip_2, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip_1.extIpAddr		= extip;
		asic_ip_1.intIpAddr		= intip;
		asic_ip_1.localPublic		= FALSE;
		asic_ip_1.nat			= TRUE;
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip_1, &asic_ip_2);
	}
	return SUCCESS;
}


/*
@func int32 | macAccessControl_test6 | MAC Access Control testing.
@comm
Add NAT entry with internal host's ARP present, but the ARP doesn't match MAC Access Control rule. This will result in the NAT
entry be added to ASIC IP Table.
*/

int32 macAccessControl_test6(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip_1, asic_ip_2;
	ipaddr_t extip, intip;
	
	
	START_TEST_CASE();
	
	{
		/*====================================================================
		  * Test: add NAT entry with internal host's arp present, but the MAC not match MAC Access
		  *	      Control.
		  *====================================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);

		/* (1) add ARP entry */
		mac.octet[5] = 0xfe;
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);

		/* (2)  add NAT Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNatMapping(extip, intip) == SUCCESS);

		/* Verify NAT Table */
		bzero(&asic_ip_1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip_2, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip_1.extIpAddr		= extip;
		asic_ip_1.intIpAddr		= intip;
		asic_ip_1.localPublic		= FALSE;
		asic_ip_1.nat			= TRUE;
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip_1, &asic_ip_2);
		
	}
	 return SUCCESS;
}	 


/*
@func int32 | macAccessControl_test7 | MAC Access Control testing.
@comm
Add NAT entry with internal host's ARP present, but the ARP  matches MAC Access Control rule. This will result in the NAT
entry be added to softwareIP Table only.
*/

int32 macAccessControl_test7(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip_1, asic_ip_2;
	ipaddr_t extip, intip;
	
	
	START_TEST_CASE();
	
	{
		/*====================================================================
		  * Test: add NAT entry with internal host arp present, the MAC matches MAC Access
		  *	      Control.
		  *====================================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);

		/* (1) add ARP entry */
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);

		/* (2)  add NAT Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNatMapping(extip, intip) == SUCCESS);

		/* Verify NAT Table */
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) != SUCCESS);		
	}
	return SUCCESS;
}


/*
@func int32 | macAccessControl_test8 | MAC Access Control testing.
@comm
Add NAT entry with internal host's ARP present. Then add a  MAC Access Control rule to match the internal host's arp.
This will result in the NAT entry be removed from ASIC IP Table. Then remove the MAC Access Control rule, the NAT
entry should be written back to ASIC.
*/

int32 macAccessControl_test8(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip_1, asic_ip_2;
	ipaddr_t extip, intip;
	
	
	START_TEST_CASE();
	
	{
		/*====================================================================
		  * Test: add NAT entry with internal host arp present, then add a MAC Access Control entry to
		  *	      match the ARP.
		  *====================================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);

		/* (1) add ARP entry */
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);

		/* (2)  add NAT Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNatMapping(extip, intip) == SUCCESS);

		/* Verify NAT Table */
		bzero(&asic_ip_1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip_2, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip_1.extIpAddr		= extip;
		asic_ip_1.intIpAddr		= intip;
		asic_ip_1.localPublic		= FALSE;
		asic_ip_1.nat			= TRUE;
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip_1, &asic_ip_2);

		/* (3) add MAC Access Control rule to match the NAT entry's internal host */
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) != SUCCESS);	

		/* (4) remove the MAC Access Control rule */
		ASSERT(rtl8651_delMacAccessControl(&mac,&macMask) == SUCCESS);

		/* Verify NAT Table */
		bzero(&asic_ip_1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip_2, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip_1.extIpAddr		= extip;
		asic_ip_1.intIpAddr		= intip;
		asic_ip_1.localPublic		= FALSE;
		asic_ip_1.nat			= TRUE;
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip_1, &asic_ip_2);
	}
	return SUCCESS;
}


/*
@func int32 | macAccessControl_test9 | MAC Access Control testing.
@comm
Add NAT entry with no internal host's ARP present. After the arp is learnt and the arp matches the MAC Access Control rule,
the NAT entry should be removed from ASIC IP Table. Then remove the arp, the NAT entry should be written back to ASIC.
*/

int32 macAccessControl_test9(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip_1, asic_ip_2;
	ipaddr_t extip, intip;
	
	
	START_TEST_CASE();
	
	{
		/*====================================================================
		  * Test: add NAT entry but lack its internal host's arp. When its arp is learnt and its MAC mathes
		  *	      MAC Access Control rule, then the NAT entry should be removed from ASIC IP Table.
		  *====================================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);

		/* (1)  add NAT Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNatMapping(extip, intip) == SUCCESS);

		/* Verify NAT Table */
		bzero(&asic_ip_1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip_2, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip_1.extIpAddr		= extip;
		asic_ip_1.intIpAddr		= intip;
		asic_ip_1.localPublic		= FALSE;
		asic_ip_1.nat			= TRUE;
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip_1, &asic_ip_2);

		/* (2) add Arp entry */
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) != SUCCESS);

		/* (3) remove Arp entry */
		ASSERT(rtl8651_delArp(intip) == SUCCESS);

		/* Verify NAT Table */
		bzero(&asic_ip_1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip_2, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip_1.extIpAddr		= extip;
		asic_ip_1.intIpAddr		= intip;
		asic_ip_1.localPublic		= FALSE;
		asic_ip_1.nat			= TRUE;
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip_1, &asic_ip_2);
	}
	return SUCCESS;
}



/*
@func int32 | macAccessControl_test10 | MAC Access Control testing.
@comm
Add Server Port entry but lack internal host's ARP. The internal host never matches Mac Access Control rule, hence it should
be added to ASIC Server Port Table.
*/

int32 macAccessControl_test10(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_serverPortParam_t asic_sp_1, asic_sp_2;
	ipaddr_t extip, intip;
	
	
	START_TEST_CASE();
	
	{
		/*=================================================================
		  * Test: add SP entry but lack internal host's arp hence it will not match Mac Access Control.
		  *=================================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);

		/* (1)  add NAT Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNaptMapping(extip) == SUCCESS);
		ASSERT(rtl8651_addNaptServerPortMapping(TRUE, extip, 1000, intip, 2000) == SUCCESS);

#ifdef SERVERPORT_PARK_TCPUDP
		/* In this mode, Server Port will be handled by software. */
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
#else/*SERVERPORT_PARK_TCPUDP*/
		/* Verify Server Port Table */
		bzero(&asic_sp_1, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		bzero(&asic_sp_2, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		asic_sp_1.extIpAddr	= extip;
		asic_sp_1.extPort		= 1000;
		asic_sp_1.intIpAddr		= intip;
		asic_sp_1.intPort		= 2000;
		asic_sp_1.nhIndex		= 0;
		asic_sp_1.portRange	= 0;
		asic_sp_1.valid		= 1;
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) == SUCCESS);
		SP_ENTRY_CMP(&asic_sp_1, &asic_sp_2);
#endif/*SERVERPORT_PARK_TCPUDP*/
	}
	return SUCCESS;
}



/*
@func int32 | macAccessControl_test11 | MAC Access Control testing.
@comm
Add Server Port entry with internal host's ARP present, but the ARP doesn't match MAC Access Control rule. This will result 
in Server Port entry be added to ASIC Server Port Table.
*/

int32 macAccessControl_test11(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_serverPortParam_t asic_sp_1, asic_sp_2;
	ipaddr_t extip, intip;
	
	
	START_TEST_CASE();
	
	{
		/*====================================================================
		  * Test: add SP entry with internal host's arp present, but the MAC not match MAC Access
		  *	      Control.
		  *====================================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);

		/* (1) add ARP entry */
		mac.octet[5] = 0xfe;
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);

		/* (2)  add NAT Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNaptMapping(extip) == SUCCESS);
		ASSERT(rtl8651_addNaptServerPortMapping(TRUE, extip, 1000, intip, 2000) == SUCCESS);

#ifdef SERVERPORT_PARK_TCPUDP
		/* In this mode, Server Port will be handled by software. */
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
#else/*SERVERPORT_PARK_TCPUDP*/
		/* Verify Server Port Table */
		bzero(&asic_sp_1, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		bzero(&asic_sp_2, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		asic_sp_1.extIpAddr	= extip;
		asic_sp_1.extPort		= 1000;
		asic_sp_1.intIpAddr		= intip;
		asic_sp_1.intPort		= 2000;
		asic_sp_1.nhIndex		= 0;
		asic_sp_1.portRange	= 0;
		asic_sp_1.valid		= 1;
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) == SUCCESS);
		SP_ENTRY_CMP(&asic_sp_1, &asic_sp_2);		
#endif/*SERVERPORT_PARK_TCPUDP*/
	}
	 return SUCCESS;
}	 



/*
@func int32 | macAccessControl_test12 | MAC Access Control testing.
@comm
Add Server Port entry with internal host's ARP present, but the ARP  matches MAC Access Control rule. This will result in the 
Server Port entry be added to softwareIP Table only.
*/

int32 macAccessControl_test12(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_serverPortParam_t asic_sp_1, asic_sp_2;
	ipaddr_t extip, intip;
	
	
	START_TEST_CASE();
	
	{
		/*====================================================================
		  * Test: add SP entry with internal host arp present, the MAC matches MAC Access
		  *	      Control.
		  *====================================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);

		/* (1) add ARP entry */
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);

		/* (2)  add SP Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNaptMapping(extip) == SUCCESS);
		ASSERT(rtl8651_addNaptServerPortMapping(TRUE, extip, 1000,  intip, 2000) == SUCCESS);

		/* Verify SP Table */
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
	}
	return SUCCESS;
}



/*
@func int32 | macAccessControl_test13 | MAC Access Control testing.
@comm
Add Server Port entry with internal host's ARP present. Then add a  MAC Access Control rule to match the internal host's arp.
This will result in the Server Port entry be removed from ASIC Server Port Table. Then remove the MAC Access Control rule, 
the Server Port entry should be written back to ASIC.
*/

int32 macAccessControl_test13(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_serverPortParam_t asic_sp_1, asic_sp_2;
	ipaddr_t extip, intip;
	
	
	START_TEST_CASE();
	
	{
		/*====================================================================
		  * Test: add SP entry with internal host arp present, then add a MAC Access Control entry to
		  *	      match the ARP.
		  *====================================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);

		/* (1) add ARP entry */
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);

		/* (2)  add SP Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNaptMapping(extip) == SUCCESS);
		ASSERT(rtl8651_addNaptServerPortMapping(TRUE, extip, 1000,  intip, 2000) == SUCCESS);

#ifdef SERVERPORT_PARK_TCPUDP
		/* In this mode, Server Port will be handled by software. */
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
#else/*SERVERPORT_PARK_TCPUDP*/
		/* Verify Server Port Table */
		bzero(&asic_sp_1, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		bzero(&asic_sp_2, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		asic_sp_1.extIpAddr	= extip;
		asic_sp_1.extPort		= 1000;
		asic_sp_1.intIpAddr		= intip;
		asic_sp_1.intPort		= 2000;
		asic_sp_1.nhIndex		= 0;
		asic_sp_1.portRange	= 0;
		asic_sp_1.valid		= 1;
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) == SUCCESS);
		SP_ENTRY_CMP(&asic_sp_1, &asic_sp_2);	
#endif/*SERVERPORT_PARK_TCPUDP*/

		/* (3) add MAC Access Control rule to match the SP entry's internal host */
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);	

		/* (4) remove the MAC Access Control rule */
		ASSERT(rtl8651_delMacAccessControl(&mac,&macMask) == SUCCESS);

#ifdef SERVERPORT_PARK_TCPUDP
		/* In this mode, Server Port will be handled by software. */
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
#else/*SERVERPORT_PARK_TCPUDP*/
		/* Verify Server Port Table */
		bzero(&asic_sp_1, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		bzero(&asic_sp_2, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		asic_sp_1.extIpAddr	= extip;
		asic_sp_1.extPort		= 1000;
		asic_sp_1.intIpAddr		= intip;
		asic_sp_1.intPort		= 2000;
		asic_sp_1.nhIndex		= 0;
		asic_sp_1.portRange	= 0;
		asic_sp_1.valid		= 1;
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) == SUCCESS);
		SP_ENTRY_CMP(&asic_sp_1, &asic_sp_2);
#endif/*SERVERPORT_PARK_TCPUDP*/
	}
	return SUCCESS;
}



/*
@func int32 | macAccessControl_test14 | MAC Access Control testing.
@comm
Add Server Port entry with no internal host's ARP present. After the arp is learnt and the arp matches the MAC Access Control rule,
the Server Port entry should be removed from ASIC Server Port Table. Then remove the arp, the Server Port entry should be 
written back to ASIC.
*/

int32 macAccessControl_test14(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_serverPortParam_t asic_sp_1, asic_sp_2;
	ipaddr_t extip, intip;
	
	
	START_TEST_CASE();
	
	{
		/*====================================================================
		  * Test: add SP entry but lack its internal host's arp. When its arp is learnt and its MAC mathes
		  *	      MAC Access Control rule, then the SP entry should be removed from ASIC SP Table.
		  *====================================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);

		/* (1)  add NAT Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNaptMapping(extip) == SUCCESS);
		ASSERT(rtl8651_addNaptServerPortMapping(TRUE, extip, 1000,  intip, 2000) == SUCCESS);

#ifdef SERVERPORT_PARK_TCPUDP
		/* In this mode, Server Port will be handled by software. */
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
#else/*SERVERPORT_PARK_TCPUDP*/
		/* Verify Server Port Table */
		bzero(&asic_sp_1, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		bzero(&asic_sp_2, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		asic_sp_1.extIpAddr	= extip;
		asic_sp_1.extPort		= 1000;
		asic_sp_1.intIpAddr		= intip;
		asic_sp_1.intPort		= 2000;
		asic_sp_1.nhIndex		= 0;
		asic_sp_1.portRange	= 0;
		asic_sp_1.valid		= 1;
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) == SUCCESS);
		SP_ENTRY_CMP(&asic_sp_1, &asic_sp_2);
#endif/*SERVERPORT_PARK_TCPUDP*/
		
		/* (2) add Arp entry */
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);

		/* (3) remove Arp entry */
		ASSERT(rtl8651_delArp(intip) == SUCCESS);

#ifdef SERVERPORT_PARK_TCPUDP
		/* In this mode, Server Port will be handled by software. */
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
#else/*SERVERPORT_PARK_TCPUDP*/
		/* Verify Server Port Table */
		bzero(&asic_sp_1, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		bzero(&asic_sp_2, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
		asic_sp_1.extIpAddr	= extip;
		asic_sp_1.extPort		= 1000;
		asic_sp_1.intIpAddr		= intip;
		asic_sp_1.intPort		= 2000;
		asic_sp_1.nhIndex		= 0;
		asic_sp_1.portRange	= 0;
		asic_sp_1.valid		= 1;
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) == SUCCESS);
		SP_ENTRY_CMP(&asic_sp_1, &asic_sp_2);
#endif/*SERVERPORT_PARK_TCPUDP*/
	}
	return SUCCESS;
}



/*
@func int32 | macAccessControl_test15 | MAC Access Control testing.
@comm
When enable or disable MAC filtering function, the pending Server Port entry and NAT entry should be written back to ASIC.
*/

int32 macAccessControl_test15(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_serverPortParam_t asic_sp_1, asic_sp_2;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip_1, asic_ip_2;
	ipaddr_t extip, intip;
	int32 round;
	
	
	START_TEST_CASE();
	
	{
		/*=========================================================
		  * Test for NAT
		  *=========================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);

		/* (1) add ARP entry */
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);

		/* (2)  add NAT Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNatMapping(extip, intip) == SUCCESS);

		/* Verify NAT Table */
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) != SUCCESS);

		for(round=0; round<1000; round++) {
			/* (3) turn off MAC filter */
			ASSERT(rtl8651_enableMacAccessControl(FALSE) == SUCCESS);
			
			/* Verify NAT Table */
			bzero(&asic_ip_1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			bzero(&asic_ip_2, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			asic_ip_1.extIpAddr		= extip;
			asic_ip_1.intIpAddr		= intip;
			asic_ip_1.localPublic		= FALSE;
			asic_ip_1.nat			= TRUE;
			ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) == SUCCESS);
			IP_ENTRY_CMP(&asic_ip_1, &asic_ip_2);

			/* (4) turn on MAC filter */
			ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);

			/* Verify NAT Table */
			ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) != SUCCESS);
		}
		
	}

	ASSERT(rtl8651_tblDrvReinit() == SUCCESS);
	
	{
		/*=========================================================
		  * Test for Server Port
		  *=========================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);

		/* (1)  add Arp */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);	

		/* (2) add Server Port Mapping */
		ASSERT(rtl8651_addNaptMapping(extip) == SUCCESS);
		ASSERT(rtl8651_addNaptServerPortMapping(TRUE, extip, 1000, intip, 2000) == SUCCESS);
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);

		for(round=0; round<1000; round++) {
			/* (3) turn off MAC filter */
			ASSERT(rtl8651_enableMacAccessControl(FALSE) == SUCCESS);
			
#ifdef SERVERPORT_PARK_TCPUDP
		/* In this mode, Server Port will be handled by software. */
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
#else/*SERVERPORT_PARK_TCPUDP*/
			/* Verify Server Port Table */
			bzero(&asic_sp_1, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
			bzero(&asic_sp_2, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
			asic_sp_1.extIpAddr	= extip;
			asic_sp_1.extPort		= 1000;
			asic_sp_1.intIpAddr		= intip;
			asic_sp_1.intPort		= 2000;
			asic_sp_1.nhIndex		= 0;
			asic_sp_1.portRange	= 0;
			asic_sp_1.valid		= 1;
			ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) == SUCCESS);
			SP_ENTRY_CMP(&asic_sp_1, &asic_sp_2);
#endif/*SERVERPORT_PARK_TCPUDP*/

			/* (4) turn on MAC Filter */
			ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
			ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
		}
	}
	return SUCCESS;
}




/*
@func int32 | macAccessControl_test16 | MAC Access Control testing.
@comm
Pendding or not pendding Server Port and NAT entry should be according to match MAC access control rule or not.
*/

int32 macAccessControl_test16(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x00, 0xcc, 0x11, 0x22, 0x01 } };
	ether_addr_t macMask = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
	rtl865x_tblAsicDrv_serverPortParam_t asic_sp_1, asic_sp_2;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip_1, asic_ip_2;
	ipaddr_t extip, intip;
	int32 round;
	
	
	START_TEST_CASE();
	
	{
		/*=========================================================
		  * Test for NAT
		  *=========================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);
		ASSERT(rtl8651_setMacAccessControlDefaultPermit(FALSE) == SUCCESS);

		/* (1) add ARP entry */
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);

		/* (2)  add NAT Mapping */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addNatMapping(extip, intip) == SUCCESS);

		/* Verify NAT Table */
		ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) != SUCCESS);

again:
		for(round=0; round<1000; round++) {
			/* (3) set default permit */
			ASSERT(rtl8651_setMacAccessControlDefaultPermit(TRUE) == SUCCESS);
			
			/* Verify NAT Table */
			bzero(&asic_ip_1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			bzero(&asic_ip_2, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			asic_ip_1.extIpAddr		= extip;
			asic_ip_1.intIpAddr		= intip;
			asic_ip_1.localPublic		= FALSE;
			asic_ip_1.nat			= TRUE;
			ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) == SUCCESS);
			IP_ENTRY_CMP(&asic_ip_1, &asic_ip_2);

			/* (4) set default drop */
			ASSERT(rtl8651_setMacAccessControlDefaultPermit(FALSE) == SUCCESS);

			/* Verify NAT Table */
			ASSERT(rtl8651_getAsicExtIntIpTable(0,  &asic_ip_2) != SUCCESS);
		}

		if (mac.octet[5] != 0xfd) {
			mac.octet[5] = 0xfd;
			ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);
			goto again;
		}
	}

	ASSERT(rtl8651_tblDrvReinit() == SUCCESS);
	
	{
		/*=========================================================
		  * Test for Server Port
		  *=========================================================*/
		default_configuration1(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ASSERT(rtl8651_enableMacAccessControl(TRUE) == SUCCESS);
		ASSERT(rtl8651_setMacAccessControlDefaultPermit(FALSE) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("vlan1") == SUCCESS);

		/* (1)  add Arp */
		extip = ntohl(inet_addr("192.168.1.1"));
		intip	 = ntohl(inet_addr("192.168.2.2"));
		ASSERT(rtl8651_addArp(intip, &mac, "vlan2", 0x04) == SUCCESS);	

		/* (2) add Server Port Mapping */
		ASSERT(rtl8651_addNaptMapping(extip) == SUCCESS);
		ASSERT(rtl8651_addNaptServerPortMapping(TRUE, extip, 1000, intip, 2000) == SUCCESS);
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);

again1:
		for(round=0; round<1000; round++) {
			/* (3) set default permit */
			ASSERT(rtl8651_setMacAccessControlDefaultPermit(TRUE) == SUCCESS);
			
#ifdef SERVERPORT_PARK_TCPUDP
		/* In this mode, Server Port will be handled by software. */
		ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
#else/*SERVERPORT_PARK_TCPUDP*/
			/* Verify Server Port Table */
			bzero(&asic_sp_1, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
			bzero(&asic_sp_2, sizeof(rtl865x_tblAsicDrv_serverPortParam_t));
			asic_sp_1.extIpAddr	= extip;
			asic_sp_1.extPort		= 1000;
			asic_sp_1.intIpAddr		= intip;
			asic_sp_1.intPort		= 2000;
			asic_sp_1.nhIndex		= 0;
			asic_sp_1.portRange	= 0;
			asic_sp_1.valid		= 1;
			ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) == SUCCESS);
			SP_ENTRY_CMP(&asic_sp_1, &asic_sp_2);
#endif/*SERVERPORT_PARK_TCPUDP*/

			/* (4) set default drop */
			ASSERT(rtl8651_setMacAccessControlDefaultPermit(FALSE) == SUCCESS);
			ASSERT(rtl8651_getAsicServerPortTable(0,  &asic_sp_2) != SUCCESS);
		}

		if (mac.octet[5] != 0xfd) {
			mac.octet[5] = 0xfd;
			ASSERT(rtl8651_addMacAccessControl(&mac, &macMask, FALSE) == SUCCESS);
			goto again1;
		}
	}
	return SUCCESS;
}








#define RTL865X_RATE_LIMIT_TESTING_CODE

int32 rateLimit_test1(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;

	START_TEST_CASE();

	{
		/*========================================================
		  * Test: add/remove ratelimt group
		  *========================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		
		/* (1) add an entry to ratelimit table: 1.5KMbps, max 1.6Mbps */
		ASSERT(rtl8651_addRateLimitGroup(1000, "vlan", 1500, 1600, TRUE) == SUCCESS);

		/* Verify */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 2666;
		asic_rl1.refill_number	= 2500;
		asic_rl1.t_intervalUnit		= 1;
		asic_rl1.t_remainUnit		= 1;
		asic_rl1.token			= 2500;
		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		
		/* (2) remove the entry added before */
		ASSERT(rtl8651_delRateLimitGroup(1001) != SUCCESS);
		ASSERT(rtl8651_delRateLimitGroup(1000) == SUCCESS);
	}
	return SUCCESS;
}


int32 rateLimit_test2(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	int32 round, sel, count;

	START_TEST_CASE();

	{
		/*========================================================
		  * Test: add/remove ratelimt group: overflow table
		  *========================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) overflow table */
		for(round=0; round<RTL8651_RATELIMITTBL_SIZE; round++) {
			ASSERT(rtl8651_addRateLimitGroup(1000+round, "vlan", 1500+round*3, 1800+round*3, TRUE) == SUCCESS);
			/* Verify */
			for(count=0; count<=round; count++) {
				bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
				bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
				asic_rl1.maxToken		= 3000 + count *5 ;
				asic_rl1.refill_number	= 2500 + count *5;
				asic_rl1.t_intervalUnit		= 1;
				asic_rl1.t_remainUnit		= 1;
				asic_rl1.token			= 2500 + count *5;
				ASSERT(rtl8651_getAsicRateLimitTable(count, &asic_rl2) == SUCCESS);
				RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);
			}
		}
		ASSERT(rtl8651_addRateLimitGroup(1000+round, "vlan", 1500+round*3, 1600+round*3, TRUE) != SUCCESS);

		/* (2) remove entry one by one to emtpry the table */
		for(round=0; round<RTL8651_RATELIMITTBL_SIZE; round++) {
			ASSERT(rtl8651_delRateLimitGroup(1000+round) == SUCCESS);
			for(count=0; count<RTL8651_RATELIMITTBL_SIZE; count++) {
				if (count <= round) {
					ASSERT(rtl8651_getAsicRateLimitTable(count, &asic_rl2) != SUCCESS);
				}
				else {
					bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
					bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
					asic_rl1.maxToken		= 3000 + count*5;
					asic_rl1.refill_number	= 2500 + count*5;
					asic_rl1.t_intervalUnit		= 1;
					asic_rl1.t_remainUnit		= 1;
					asic_rl1.token			= 2500 + count*5;
					ASSERT(rtl8651_getAsicRateLimitTable(count, &asic_rl2) == SUCCESS);
					RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);
				}
			}
		}
	}
	return SUCCESS;
}



int32 rateLimit_test3(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0;
	

	START_TEST_CASE();

	{
		/*===================================================
		  * Test: add/remove rate limit rule
		  *===================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add 2 rate limit groups */
		ASSERT(rtl8651_addRateLimitGroup(2000, "wan", 1500, 1600, TRUE) == SUCCESS);
		ASSERT(rtl8651_addRateLimitGroup(1000, "vlan", 1500, 1600, TRUE) == SUCCESS);
		
		/* (2) add a rate limit rule: IP rule */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= ntohl(inet_addr("192.168.1.0"));
		rule.srcIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addRateLimitRule(&rule, 1001) != SUCCESS);
		ASSERT(rtl8651_addRateLimitRule(&rule, 1000) == SUCCESS);

		/* Verify ACL */
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		memcpy(&asic_acl1, &rule, sizeof(rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ = RTL8651_ACL_DROP_RATE_EXCEED_BPS;
		asic_acl1.pktOpApp = 0x06;
		asic_acl1.rlIndex	= 1;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) cur++;
#endif/*RTL865XB_URL_FILTER*/
		cur +=4; /* for SYN+FIN+RST */
				
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ 	= RTL8651_ACL_PERMIT;
		asic_acl1.ruleType_	   	= RTL8651_ACL_MAC;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);

		/* (3) remove the rate limit rule */
		ASSERT(rtl8651_delRateLimitRule(&rule, 1000) == SUCCESS);

		cur = 0;
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) cur++;
#endif/*RTL865XB_URL_FILTER*/
		cur += 4;

		/* Verify ACL */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ 	= RTL8651_ACL_PERMIT;
		asic_acl1.ruleType_	   	= RTL8651_ACL_MAC;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+1), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);

	}
	return SUCCESS;
}


int32 rateLimit_test4(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0, round;
	

	START_TEST_CASE();

	{
		/*===================================================
		  * Test: rate limit reference count
		  *===================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add 2 rate limit groups */
		ASSERT(rtl8651_addRateLimitGroup(2000, "wan", 1500, 1600, TRUE) == SUCCESS);
		ASSERT(rtl8651_addRateLimitGroup(1000, "vlan", 1500, 1600, FALSE) == SUCCESS);

		/* (2) add 10 rate limit rules(use goup 1000): TCP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_TCP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(rtl8651_addRateLimitRule(&rule, 1000) == SUCCESS);
		}
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

		/* Verify ACL */
		for(round=0; round<10; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_TCP;
			asic_acl1.actionType_		= RTL8651_ACL_LOG_RATE_EXCEED_PPS;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.tcpSrcPortUB_		= 60000 + round;
			asic_acl1.tcpSrcPortLB_		= 0 + round;
			asic_acl1.tcpDstPortUB_	 	= 20000 + round;
			asic_acl1.tcpDstPortLB_		= 1000 + round;
			asic_acl1.rlIndex			= 1;
			asic_acl1.pktOpApp			= 0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}


		/* (3) add 10 rate limit rules(use goup 2000): UDP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_UDP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("141.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(rtl8651_addRateLimitRule(&rule, 2000) == SUCCESS);
		}
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);

		/* Verify ACL */
		for(round=0; round<10; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_UDP;
			asic_acl1.actionType_		= RTL8651_ACL_LOG_RATE_EXCEED_BPS;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("141.168.1.0")) + (round << 8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.tcpSrcPortUB_		= 60000 + round;
			asic_acl1.tcpSrcPortLB_		= 0 + round;
			asic_acl1.tcpDstPortUB_	 	= 20000 + round;
			asic_acl1.tcpDstPortLB_		= 1000 + round;
			asic_acl1.rlIndex			= 0;
			asic_acl1.pktOpApp			= 0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}


		/* (4) remove10 rate limit rules(use goup 1000): TCP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_TCP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(rtl8651_delRateLimitGroup(1000) != SUCCESS);
			ASSERT(rtl8651_delRateLimitRule(&rule, 1000) == SUCCESS);
		}
		
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

		cur = 0;
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) cur++;
#endif/*RTL865XB_URL_FILTER*/
		cur += 3; /* for tcp syn, fin ,rst */
		/* Verify ACL */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ 	= RTL8651_ACL_PERMIT;
		asic_acl1.ruleType_	   	= RTL8651_ACL_MAC;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+1), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);

		ASSERT(rtl8651_delRateLimitGroup(1000) == SUCCESS);

		/* (5) remove 10 rate limit rules(use goup 2000): UDP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_UDP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("141.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(rtl8651_delRateLimitGroup(2000) != SUCCESS);
			ASSERT(rtl8651_delRateLimitRule(&rule, 2000) == SUCCESS);
		}
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		ASSERT(rtl8651_delRateLimitGroup(2000) == SUCCESS);
	}
	return SUCCESS;
}


int32 rateLimit_test5(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0, round;
	

	START_TEST_CASE();

	{
		/*===================================================
		  * Test: flush Rate Limit
		  *===================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add 2 rate limit entries */
		ASSERT(rtl8651_addRateLimitGroup(2000, "wan", 1600, 1700, TRUE) == SUCCESS);
		ASSERT(rtl8651_addRateLimitGroup(1000, "vlan", 1500, 1800, FALSE) == SUCCESS);

		/* Verify Rate Limit Table */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 2833;
		asic_rl1.refill_number	= 2666;
		asic_rl1.t_intervalUnit		= 1;
		asic_rl1.t_remainUnit		= 1;
		asic_rl1.token			= 2666;
		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 3000;
		asic_rl1.refill_number	= 2500;
		asic_rl1.t_intervalUnit		= 1;
		asic_rl1.t_remainUnit		= 1;
		asic_rl1.token			= 2500;
		ASSERT(rtl8651_getAsicRateLimitTable(1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		/* (2) add 10 rate limit rules(use goup 1000): TCP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_TCP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(rtl8651_addRateLimitRule(&rule, 1000) == SUCCESS);
		}
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

		/* Verify ACL */
		for(round=0; round<10; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_TCP;
			asic_acl1.actionType_		= RTL8651_ACL_LOG_RATE_EXCEED_PPS;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.tcpSrcPortUB_		= 60000 + round;
			asic_acl1.tcpSrcPortLB_		= 0 + round;
			asic_acl1.tcpDstPortUB_	 	= 20000 + round;
			asic_acl1.tcpDstPortLB_		= 1000 + round;
			asic_acl1.rlIndex			= 1;
			asic_acl1.pktOpApp			= 0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}

		/* (3) flush Rate Limit */
		ASSERT(rtl8651_flushRateLimitGroup("vlan") == SUCCESS);

		/* Verify Rate Limit Table */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 2833;
		asic_rl1.refill_number	= 2666;
		asic_rl1.t_intervalUnit		= 1;
		asic_rl1.t_remainUnit		= 1;
		asic_rl1.token			= 2666;
		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);
		ASSERT(rtl8651_getAsicRateLimitTable(1, &asic_rl2) != SUCCESS);

		
	}
	return SUCCESS;
}





#define RTL865X_RATE_LIMIT_COUNTER_TESTING_CODE

int32 rateLimitCounter_test1(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	int32 idx;

	START_TEST_CASE();

	{
		/*========================================================
		  * Test: add/remove ratelimt counter
		  *========================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		for (idx=0;idx<RATELIMIT_COUNTER_RESV_SIZE;idx++)
		{
			/* (1) remove ratelimit counter table */
			ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("vlan", NULL), idx) == TBLDRV_EENTRYNOTFOUND);

			/* (2) add an entry to ratelimit counter table */
			ASSERT(_rtl8651_addRateLimitCounter(GET_NETWORK_INTERFACE("vlan", NULL), idx) == SUCCESS);

			/* Verify */
			bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
			bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
			asic_rl1.maxToken		= 0xffffff;
			asic_rl1.refill_number	= 0;
			asic_rl1.t_intervalUnit	= 1;
			asic_rl1.t_remainUnit	= 1;
			asic_rl1.token			= 0xffffff;
			ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) == SUCCESS);	// get idx[0] due to each counter would reuse it
			RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

			/* (3) remove the entry added before */
			ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), idx) == SUCCESS);
			ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), idx) == TBLDRV_EENTRYNOTFOUND);
		}
	}
	return SUCCESS;

}

int32 rateLimitCounter_test2(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	int32 idx;

	START_TEST_CASE();

	{
		/*========================================================
		  * Test: add/remove ratelimt counter
		  *========================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		for (idx=0;idx<RATELIMIT_COUNTER_RESV_SIZE;idx++)
		{
			/* (1) remove ratelimit counter table */
			ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("vlan", NULL), idx) == TBLDRV_EENTRYNOTFOUND);

			/* (2) add an entry to ratelimit counter table */
			ASSERT(_rtl8651_addRateLimitCounter(GET_NETWORK_INTERFACE("vlan", NULL), idx) == SUCCESS);

			/* Verify */
			bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
			bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
			asic_rl1.maxToken		= 0xffffff;
			asic_rl1.refill_number	= 0;
			asic_rl1.t_intervalUnit	= 1;
			asic_rl1.t_remainUnit	= 1;
			asic_rl1.token			= 0xffffff;
			ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) == SUCCESS);	// get idx[0] due to each counter would reuse it
			RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		}

		for (idx=RATELIMIT_COUNTER_RESV_SIZE;idx<5+RATELIMIT_COUNTER_RESV_SIZE;idx++)
		{
			/* (3) remove non-exist ratelimit counter table */
			ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("vlan", NULL), idx) == TBLDRV_EENTRYNOTFOUND);
			/* (4) add an entry to ratelimit counter table, table full */
			ASSERT(_rtl8651_addRateLimitCounter(GET_NETWORK_INTERFACE("vlan", NULL), idx) == TBLDRV_ENOFREEBUFFER);
		}

		for (idx=0;idx<RATELIMIT_COUNTER_RESV_SIZE;idx++)
		{
			/* (5) remove the entry added before */
			ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), idx) == SUCCESS);
			ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), idx) == TBLDRV_EENTRYNOTFOUND);
		}
	}
	return SUCCESS;

}

int32 rateLimitCounter_test3(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0;
	

	START_TEST_CASE();

	{
		/*===================================================
		  * Test: add/remove rate limit counter rule
		  *===================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add rate limit counter */
		ASSERT(_rtl8651_addRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), 0) == SUCCESS);
		
		/* (2) add a rate limit counter rule: IP rule */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= ntohl(inet_addr("192.168.1.0"));
		rule.srcIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(_rtl8651_addRateLimitCounterRule(&rule, GET_NETWORK_INTERFACE("vlan", 0), 0) == SUCCESS);

		/* Verify ACL */
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		memcpy(&asic_acl1, &rule, sizeof(rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ = RTL8651_ACL_DROP_RATE_EXCEED_PPS;
		asic_acl1.pktOpApp = 0x00; //0x06;
		asic_acl1.rlIndex	= 0;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);

		/* (3) remove the rate limit counter rule */
		ASSERT(_rtl8651_delRateLimitCounterRule(&rule, GET_NETWORK_INTERFACE("vlan", 0), 0) == SUCCESS);

		cur = 0;
#ifdef RTL865XB_URL_FILTER
		if(_urlFilter.urlFilterCount >= 0) cur++;
#endif/*RTL865XB_URL_FILTER*/
		cur += 4;

		/* Verify ACL */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ 	= RTL8651_ACL_PERMIT;
		asic_acl1.ruleType_	   	= RTL8651_ACL_MAC;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+1), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);

	}
	return SUCCESS;
}


int32 rateLimitCounter_test4(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0, round;
	

	START_TEST_CASE();

	{
		/*===================================================
		  * Test: rate limit counter reference count
		  *===================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add 2 rate limit counters */
		ASSERT(_rtl8651_addRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), 0) == SUCCESS);
		ASSERT(_rtl8651_addRateLimitCounter(GET_NETWORK_INTERFACE("wan", 0), 0) == SUCCESS);

		/* (2) add 10 rate limit rules(to vlan): TCP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_TCP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(_rtl8651_addRateLimitCounterRule(&rule, GET_NETWORK_INTERFACE("vlan", 0), 0) == SUCCESS);
		}
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

		/* Verify ACL */
		for(round=0; round<10; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_TCP;
			asic_acl1.actionType_		= RTL8651_ACL_LOG_RATE_EXCEED_PPS;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.tcpSrcPortUB_		= 60000 + round;
			asic_acl1.tcpSrcPortLB_		= 0 + round;
			asic_acl1.tcpDstPortUB_	 	= 20000 + round;
			asic_acl1.tcpDstPortLB_		= 1000 + round;
			asic_acl1.rlIndex			= 0;
			asic_acl1.pktOpApp			= 0x00; //0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}


		/* (3) add 10 rate limit counter rules(to wan): UDP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_UDP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("141.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(_rtl8651_addRateLimitCounterRule(&rule, GET_NETWORK_INTERFACE("wan", 0), 0) == SUCCESS);
		}
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);

		/* Verify ACL */
		for(round=0; round<10; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_UDP;
			asic_acl1.actionType_		= RTL8651_ACL_LOG_RATE_EXCEED_PPS;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("141.168.1.0")) + (round << 8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.tcpSrcPortUB_		= 60000 + round;
			asic_acl1.tcpSrcPortLB_		= 0 + round;
			asic_acl1.tcpDstPortUB_	 	= 20000 + round;
			asic_acl1.tcpDstPortLB_		= 1000 + round;
			asic_acl1.rlIndex			= 1;
			asic_acl1.pktOpApp			= 0x00; //0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}


		/* (4) remove10 rate limit counter rules(from vlan): TCP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_TCP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(_rtl8651_delRateLimitCounterRule(&rule, GET_NETWORK_INTERFACE("vlan", 0), 0) == SUCCESS);
		}
		
		cur = 0;
		/* Verify ACL */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ 	= RTL8651_ACL_PERMIT;
		asic_acl1.ruleType_	   	= RTL8651_ACL_MAC;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);

		/* Verify Rate Limit Counter : This counter has been deleted when all rules are removed */
		ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), 0) == TBLDRV_EENTRYNOTFOUND);

		/* (5) remove 10 rate limit counter rules(from wan): UDP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_UDP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("141.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(_rtl8651_delRateLimitCounterRule(&rule, GET_NETWORK_INTERFACE("wan", 0), 0) == SUCCESS);
		}

		/* Verify Rate Limit Counter : This counter has been deleted when all rules are removed */
		ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("wan", 0), 0) == TBLDRV_EENTRYNOTFOUND);

	}
	return SUCCESS;
}

int32 rateLimitCounter_test5(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0, round;
	

	START_TEST_CASE();

	{
		/*===================================================
		  * Test: flush Rate Limit Counter
		  *===================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add 2 rate limit counters */
		ASSERT(_rtl8651_addRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), 0) == SUCCESS);
		ASSERT(_rtl8651_addRateLimitCounter(GET_NETWORK_INTERFACE("wan", 0), 0) == SUCCESS);

		/* (2) add 10 rate limit rules(to vlan): TCP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_TCP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(_rtl8651_addRateLimitCounterRule(&rule, GET_NETWORK_INTERFACE("vlan", 0), 0) == SUCCESS);
		}
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

		/* Verify ACL */
		for(round=0; round<10; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_TCP;
			asic_acl1.actionType_		= RTL8651_ACL_LOG_RATE_EXCEED_PPS;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.tcpSrcPortUB_		= 60000 + round;
			asic_acl1.tcpSrcPortLB_		= 0 + round;
			asic_acl1.tcpDstPortUB_	 	= 20000 + round;
			asic_acl1.tcpDstPortLB_		= 1000 + round;
			asic_acl1.rlIndex			= 0;
			asic_acl1.pktOpApp			= 0x00; //0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}


		/* (3) add 10 rate limit counter rules(to wan): UDP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_UDP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("141.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(_rtl8651_addRateLimitCounterRule(&rule, GET_NETWORK_INTERFACE("wan", 0), 0) == SUCCESS);
		}
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);

		/* Verify ACL */
		for(round=0; round<10; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_UDP;
			asic_acl1.actionType_		= RTL8651_ACL_LOG_RATE_EXCEED_PPS;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("141.168.1.0")) + (round << 8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.tcpSrcPortUB_		= 60000 + round;
			asic_acl1.tcpSrcPortLB_		= 0 + round;
			asic_acl1.tcpDstPortUB_	 	= 20000 + round;
			asic_acl1.tcpDstPortLB_		= 1000 + round;
			asic_acl1.rlIndex			= 1;
			asic_acl1.pktOpApp			= 0x00; //0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}


		/* (4) Flush Rate Limit Counter */
		_rtl8651_flushRateLimitCounter();
				
		cur = 0;
		/* Verify ACL */
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ 	= RTL8651_ACL_PERMIT;
		asic_acl1.ruleType_	   	= RTL8651_ACL_MAC;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);


		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ 	= RTL8651_ACL_PERMIT;
		asic_acl1.ruleType_	   	= RTL8651_ACL_MAC;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);

		/* Verify Rate Limit Counter : This counter has been deleted when all rules are removed */
		ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), 0) == TBLDRV_EENTRYNOTFOUND);
		ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("wan", 0), 0) == TBLDRV_EENTRYNOTFOUND);

		
	}
	return SUCCESS;
}

int32 rateLimitCounter_test6(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0, round;
	

	START_TEST_CASE();

	{
		/*===================================================
		  * Test: Rate Limit Counter Auto-removing
		  *===================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add 2 rate limit counters */
		ASSERT(_rtl8651_addRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), 0) == SUCCESS);
		ASSERT(_rtl8651_addRateLimitCounter(GET_NETWORK_INTERFACE("wan", 0), 0) == SUCCESS);

		/* (2) add 10 rate limit rules(to vlan): TCP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_TCP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(_rtl8651_addRateLimitCounterRule(&rule, GET_NETWORK_INTERFACE("vlan", 0), 0) == SUCCESS);
		}
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);

		/* Verify ACL */
		for(round=0; round<10; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_TCP;
			asic_acl1.actionType_		= RTL8651_ACL_LOG_RATE_EXCEED_PPS;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("140.168.1.0")) + (round << 8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.tcpSrcPortUB_		= 60000 + round;
			asic_acl1.tcpSrcPortLB_		= 0 + round;
			asic_acl1.tcpDstPortUB_	 	= 20000 + round;
			asic_acl1.tcpDstPortLB_		= 1000 + round;
			asic_acl1.rlIndex			= 0;
			asic_acl1.pktOpApp			= 0x00; //0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}


		/* (3) add 10 rate limit counter rules(to wan): UDP rule */
		for(round=0; round<10; round++) {
			bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
			rule.ruleType_			= RTL8651_ACL_UDP;
			rule.actionType_		= RTL8651_ACL_DROP_LOG;
			rule.dstIpAddr_		= ntohl(inet_addr("141.168.1.0")) + (round << 8);
			rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			rule.tcpSrcPortUB_		= 60000 + round;
			rule.tcpSrcPortLB_		= 0 + round;
			rule.tcpDstPortUB_	 	= 20000 + round;
			rule.tcpDstPortLB_		= 1000 + round;
			ASSERT(_rtl8651_addRateLimitCounterRule(&rule, GET_NETWORK_INTERFACE("wan", 0), 0) == SUCCESS);
		}
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);

		/* Verify ACL */
		for(round=0; round<10; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_UDP;
			asic_acl1.actionType_		= RTL8651_ACL_LOG_RATE_EXCEED_PPS;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("141.168.1.0")) + (round << 8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.tcpSrcPortUB_		= 60000 + round;
			asic_acl1.tcpSrcPortLB_		= 0 + round;
			asic_acl1.tcpDstPortUB_	 	= 20000 + round;
			asic_acl1.tcpDstPortLB_		= 1000 + round;
			asic_acl1.rlIndex			= 1;
			asic_acl1.pktOpApp			= 0x00; //0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}

		/* (4) Modify Rate Limit Counter (for vlan) */
		asic_rl1.maxToken		= 0xffffff;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit	= 1;
		asic_rl1.t_remainUnit	= 1;
		asic_rl1.token			= 0xfffff;	/* remaining token become 0xfffff */
		ASSERT(rtl8651_setAsicRateLimitTable(0, &asic_rl1) == SUCCESS);

		/* Time update */
		_rtl8651_rateLimitCounterTimeUpdate(1);

		/* (5) Check if Rate Limit Counter for Vlan is removed */				
		cur = 0;
		/* Verify ACL */
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ 	= RTL8651_ACL_PERMIT;
		asic_acl1.ruleType_	   	= RTL8651_ACL_MAC;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);

		/* Verify Rate Limit Counter table */
		ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("vlan", 0), 0) == TBLDRV_EENTRYNOTFOUND);

		/* (6) Modify Rate Limit Counter (for wan) */
		asic_rl1.maxToken		= 0xffffff;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit	= 1;
		asic_rl1.t_remainUnit	= 1;
		asic_rl1.token			= 0xffffff - 100;
		ASSERT(rtl8651_setAsicRateLimitTable(1, &asic_rl1) == SUCCESS);

		/* Check if Rate Limit Counter for wan exists */
		ASSERT(_rtl8651_getRateLimitCounter(GET_NETWORK_INTERFACE("wan", 0), 0, TRUE) == 100);

		/* Verify refill function */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 0xffffff;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit	= 1;
		asic_rl1.t_remainUnit	= 1;
		asic_rl1.token			= 0xffffff;
		ASSERT(rtl8651_getAsicRateLimitTable(1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		/* (4) Modify Rate Limit Counter (for wan) */
		asic_rl1.maxToken		= 0xffffff;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit	= 1;
		asic_rl1.t_remainUnit	= 1;
		asic_rl1.token			= 0xfffff;	/* remaining token become 0xfffff */
		ASSERT(rtl8651_setAsicRateLimitTable(1, &asic_rl1) == SUCCESS);

		/* Time update */
		_rtl8651_rateLimitCounterTimeUpdate(1);

		/* (5) Check if Rate Limit Counter for Vlan is removed */				
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.actionType_ 	= RTL8651_ACL_PERMIT;
		asic_acl1.ruleType_	   	= RTL8651_ACL_MAC;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.outAclStart+cur), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);

		/* Verify Rate Limit Counter : This counter has been deleted when all rules are removed */
		ASSERT(_rtl8651_delRateLimitCounter(GET_NETWORK_INTERFACE("wan", 0), 0) == TBLDRV_EENTRYNOTFOUND);

		
	}
	return SUCCESS;
}






#define RTL865X_M_N_QUEUE_TESTING_CODE


int32 qos_MNQueue_test1(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0, round;

	START_TEST_CASE();
	{
		/*==============================================================
		  * Test: M-N Queue add/delete
		  *==============================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		
		/* (1) set total bandwidth to 100Mbps */
		ASSERT(rtl8651_setUpstreamBandwidth("wan", 100*1000*1000) == SUCCESS); 

		/* (2) add a flow to M queue */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.1.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToUnlimitedQueue("vlan", &rule, 20, 100) == SUCCESS);

		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.2.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToUnlimitedQueue("vlan", &rule, 70, 200) == SUCCESS);

		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.3.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToUnlimitedQueue("vlan", &rule, 20, 300) != SUCCESS);

		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.3.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToUnlimitedQueue("vlan", &rule, 10, 400) == SUCCESS);

		/* Verify: Rate Limit Table: Entry 1 */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 0; //((20*1000*1000) >> 3) * 5 / 3;
		asic_rl1.refill_number	= 0; //((20*1000*1000) >> 3) * 5 / 3;
		asic_rl1.t_intervalUnit		= 63;
		asic_rl1.t_remainUnit		= 63;
		asic_rl1.token			= 0; //((20*1000*1000) >> 3) * 5 / 3;
		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		/* Verify: Rate Limit Table: Entry 2 */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 0; //((70*1000*1000) >> 3) * 5 / 3;
		asic_rl1.refill_number	= 0; //((70*1000*1000) >> 3) * 5 / 3;
		asic_rl1.t_intervalUnit		= 63;
		asic_rl1.t_remainUnit		= 63;
		asic_rl1.token			= 0; //((70*1000*1000) >> 3) * 5 / 3;
		ASSERT(rtl8651_getAsicRateLimitTable(1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		/* Verify: Rate Limit Table: Entry 4 */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 0; //((10*1000*1000) >> 3) * 5 / 3;
		asic_rl1.refill_number	= 0; //((10*1000*1000) >> 3) * 5 / 3;
		asic_rl1.t_intervalUnit		= 63;
		asic_rl1.t_remainUnit		= 63;
		asic_rl1.token			= 0; //((10*1000*1000) >> 3) * 5 / 3;
		ASSERT(rtl8651_getAsicRateLimitTable(2, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);


		/* Verify ACL rule */
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		for(round=0; round<3; round++) {		
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_IP;
			asic_acl1.actionType_		= RTL8651_ACL_DROP_RATE_EXCEED_BPS;
			asic_acl1.srcIpAddr_		= 0;
			asic_acl1.srcIpAddrMask_	= 0;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("140.113.1.1")) + (round<<8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.priority_			= 1;
			asic_acl1.rlIndex			= round;
			asic_acl1.pktOpApp		= 0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}
		cur = 1;
#ifdef TRAP_TCP_SYN_FIN_RST
		cur += 3; /* tcp syn, fin, rst */
#else
		//cur += 1; /* tcp syn */
#endif
		for(round=0; round<3; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_IP;
			asic_acl1.actionType_		= RTL8651_ACL_PERMIT;
			asic_acl1.srcIpAddr_		= 0;
			asic_acl1.srcIpAddrMask_	= 0;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("140.113.1.1")) + (round<<8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.priority_			= 1;
			asic_acl1.pktOpApp		= 0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+round+3), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}

		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.ruleType_			= RTL8651_ACL_IP;
		asic_acl1.actionType_		= RTL8651_ACL_DROP_RATE_EXCEED_BPS;
		asic_acl1.srcIpAddr_		= 0;
		asic_acl1.srcIpAddrMask_	= 0;
		asic_acl1.dstIpAddr_		= 0;
		asic_acl1.dstIpAddrMask_	= 0;
		asic_acl1.priority_			= 0;
		asic_acl1.rlIndex			= 31;
		asic_acl1.pktOpApp		= 0x06;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+round+3), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		
		/* Flush */
		ASSERT(rtl8651_setUpstreamBandwidth("wan", 0) == SUCCESS);

		/* Verify */
		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) != SUCCESS);
		ASSERT(rtl8651_getAsicRateLimitTable(1, &asic_rl2) != SUCCESS);
		ASSERT(rtl8651_getAsicRateLimitTable(2, &asic_rl2) != SUCCESS);

		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		if (cur < 1) {
			ASSERT(asic_vlan.inAclStart+cur == RTL8651_ACLTBL_PERMIT_ALL);
			ASSERT(asic_vlan.inAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
		}
		else {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclEnd, &asic_acl2) == SUCCESS);
			asic_acl1.ruleType_ 	= RTL8651_ACL_MAC;
			asic_acl1.actionType_	= RTL8651_ACL_PERMIT;
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}
		
	}
	return SUCCESS;
}


int32 qos_MNQueue_test2(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0, round;

	START_TEST_CASE();
	{
		/*==============================================================
		  * Test: M-N Queue add/delete
		  *==============================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		
		/* (1) set total bandwidth to 100Mbps */
		ASSERT(rtl8651_setUpstreamBandwidth("wan", 100*1000*1000) == SUCCESS); 

		/* (2) add flows to M queue */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.1.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToUnlimitedQueue("vlan", &rule, 40, 100) == SUCCESS);

		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.2.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToUnlimitedQueue("vlan", &rule, 60, 200) == SUCCESS);

		/* (3) add flows to N queue */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.112.1.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToLimitedQueue("vlan", &rule, 20, 300) == SUCCESS);

		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.112.2.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToLimitedQueue("vlan", &rule, 75, 500) == SUCCESS);

		/* Verify: Rate Limit Table: Entry 1 */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 0; //((40*1000*1000) >> 3) * 5 / 3;
		asic_rl1.refill_number	= 0; //((40*1000*1000) >> 3) * 5 / 3;
		asic_rl1.t_intervalUnit		= 63;
		asic_rl1.t_remainUnit		= 63;
		asic_rl1.token			= 0; //((40*1000*1000) >> 3) * 5 / 3;
		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		/* Verify: Rate Limit Table: Entry 2 */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 0; //((60*1000*1000) >> 3) * 5 / 3;
		asic_rl1.refill_number	= 0; //((60*1000*1000) >> 3) * 5 / 3;
		asic_rl1.t_intervalUnit		= 63;
		asic_rl1.t_remainUnit		= 63;
		asic_rl1.token			= 0; //((60*1000*1000) >> 3) * 5 / 3;
		ASSERT(rtl8651_getAsicRateLimitTable(1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		/* Verify: Rate Limit Table: Entry 3 */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 0;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 63;
		asic_rl1.t_remainUnit		= 63;
		asic_rl1.token			= 0;
		ASSERT(rtl8651_getAsicRateLimitTable(2, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		/* Verify: Rate Limit Table: Entry 4 */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 0;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 63;
		asic_rl1.t_remainUnit		= 63;
		asic_rl1.token			= 0;
		ASSERT(rtl8651_getAsicRateLimitTable(3, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);


		/* Verify ACL rule */
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		for(round=0; round<2; round++) {		
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_IP;
			asic_acl1.actionType_		= RTL8651_ACL_DROP_RATE_EXCEED_BPS;
			asic_acl1.srcIpAddr_		= 0;
			asic_acl1.srcIpAddrMask_	= 0;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("140.113.1.1")) + (round<<8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.priority_			= 1;
			asic_acl1.rlIndex			= round;
			asic_acl1.pktOpApp		= 0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+round), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}

		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		for(round=0; round<2; round++) {		
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_IP;
			asic_acl1.actionType_		= RTL8651_ACL_DROP_RATE_EXCEED_BPS;
			asic_acl1.srcIpAddr_		= 0;
			asic_acl1.srcIpAddrMask_	= 0;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("140.112.1.1")) + (round<<8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.priority_			= 0;
			asic_acl1.rlIndex			= round+2;
			asic_acl1.pktOpApp		= 0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+round+2), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}

		cur = 1;
#ifdef TRAP_TCP_SYN_FIN_RST
		cur += 3; /* tcp syn, fin, rst */
#else
		//cur += 1; /* tcp syn */
#endif
		for(round=0; round<2; round++) {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
			asic_acl1.ruleType_			= RTL8651_ACL_IP;
			asic_acl1.actionType_		= RTL8651_ACL_PERMIT;
			asic_acl1.srcIpAddr_		= 0;
			asic_acl1.srcIpAddrMask_	= 0;
			asic_acl1.dstIpAddr_		= ntohl(inet_addr("140.113.1.1")) + (round<<8);
			asic_acl1.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
			asic_acl1.priority_			= 1;
			asic_acl1.pktOpApp		= 0x06;
			ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+round+4), &asic_acl2) == SUCCESS);
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}


		/* Flush M Queue */
		ASSERT(rtl8651_setUpstreamBandwidth("wan", 0) == SUCCESS);
		/* Verify */
		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) != SUCCESS);
		ASSERT(rtl8651_getAsicRateLimitTable(1, &asic_rl2) != SUCCESS);
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		if (cur == 0) {
			ASSERT(asic_vlan.inAclStart == RTL8651_ACLTBL_PERMIT_ALL);
			ASSERT(asic_vlan.inAclEnd == RTL8651_ACLTBL_PERMIT_ALL);
		}
		else {
			bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
			ASSERT(rtl8651_getAsicAclRule(asic_vlan.inAclEnd, &asic_acl2) == SUCCESS);
			asic_acl1.ruleType_ 	= RTL8651_ACL_MAC;
			asic_acl1.actionType_	= RTL8651_ACL_PERMIT;
			ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
		}

		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) != SUCCESS);
		ASSERT(rtl8651_getAsicRateLimitTable(1, &asic_rl2) != SUCCESS);
		ASSERT(rtl8651_getAsicRateLimitTable(2, &asic_rl2) != SUCCESS);
		ASSERT(rtl8651_getAsicRateLimitTable(3, &asic_rl2) != SUCCESS);
		
	}
	return SUCCESS;
}


int32 qos_MNQueue_test3(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl8651_tblDrvAclRule_t rule;
	uint32 ratio[] = { 40, 60, 0, 0, 0 };
	int32  round;

	START_TEST_CASE();
	{
		/*==============================================================
		  * Test: M-N Queue: timer
		  *==============================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		
		/* (1) set total bandwidth to 100Mbps */
		ASSERT(rtl8651_setUpstreamBandwidth("wan", 100*1000*1000) == SUCCESS); 

		/* (2) add flows to M queue */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.1.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToUnlimitedQueue("vlan", &rule, 40, 100) == SUCCESS);

		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.2.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToUnlimitedQueue("vlan", &rule, 60, 200) == SUCCESS);

		/* (3) add flows to N queue */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.112.1.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToLimitedQueue("vlan", &rule, 20, 300) == SUCCESS);

		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.112.2.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToLimitedQueue("vlan", &rule, 50, 400) == SUCCESS);

		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.112.2.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToLimitedQueue("vlan", &rule, 20, 500) == SUCCESS);

printf("sec 1:\n");
		/* (4) trigger timer: 1 second */
		_rtl8651_mnQueueTimer();

		/* Verify for (4) */
		for(round=0; round<5; round++) {
			bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
			bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
			asic_rl1.maxToken		= ((((100*1000*1000)>>3) * ratio[round]) / 100);
			asic_rl1.refill_number	= 0;
			asic_rl1.t_intervalUnit		= 61;
			asic_rl1.t_remainUnit		= 61;
			asic_rl1.token			= ((((100*1000*1000)>>3) * ratio[round]) / 100);
			ASSERT(rtl8651_getAsicRateLimitTable(round, &asic_rl2) == SUCCESS);
			RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);
		}

printf("sec: 2\n");
		/* (5) trigger timer: 2 second */
		_rtl8651_mnQueueTimer();

printf("sec: 3\n");
		/* (6) trigger timer: 3 second */
		_rtl8651_mnQueueTimer();

printf("sec: 4\n");
		/* (7) trigger timer: 4 second */
		_rtl8651_mnQueueTimer();
		
	}
	return SUCCESS;
}


int32 qos_MNQueue_test4(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0, round;

	START_TEST_CASE();
	{
		/*==============================================================
		  * Test: M-N Queue add/delete
		  *==============================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) set total bandwidth to 100Mbps */
		ASSERT(rtl8651_setUpstreamBandwidth("wan", 100*1000*1000) == SUCCESS); 

		/* (2) add flows to M queue */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.113.1.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToUnlimitedQueue("vlan", &rule, 40, 100) == SUCCESS);

		/* Verify if default rate limit entry is crated */
		/* Verify: Rate Limit Table: Entry 3 */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= 0;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 61;
		asic_rl1.t_remainUnit		= 61;
		asic_rl1.token			= 0;
		ASSERT(rtl8651_getAsicRateLimitTable(RTL8651_RATELIMITTBL_SIZE-1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);


		/* Verify ACL */
		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan) == SUCCESS);
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.ruleType_			= RTL8651_ACL_MAC;
		asic_acl1.actionType_		= RTL8651_ACL_DROP_RATE_EXCEED_BPS;
		asic_acl1.rlIndex			= RTL8651_RATELIMITTBL_SIZE-1;
		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+2), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);

		bzero(&asic_vlan, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		ASSERT(rtl8651_getAsicVlan(1, &asic_vlan) == SUCCESS);
		bzero(&asic_acl1, sizeof(_rtl8651_tblDrvAclRule_t));
		bzero(&asic_acl2, sizeof(_rtl8651_tblDrvAclRule_t));
		asic_acl1.ruleType_			= RTL8651_ACL_MAC;
		asic_acl1.actionType_		= RTL8651_ACL_PERMIT;
 		ASSERT(rtl8651_getAsicAclRule((asic_vlan.inAclStart+cur+1), &asic_acl2) == SUCCESS);
		ACL_ENTRY_CMP(&asic_acl1, &asic_acl2);
	}
	return SUCCESS;
}


int32 qos_MNQueue_test5(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0, round;

	START_TEST_CASE();
	{
		/*==============================================================
		  * Test: M-N Queue add/delete
		  *==============================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) set total bandwidth to 100Mbps */
		ASSERT(rtl8651_setUpstreamBandwidth("wan", 100*1000*1000) == SUCCESS); 

		/* (3) add flows to N queue */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.112.1.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToLimitedQueue("vlan", &rule, 20, 100) == SUCCESS);

		_rtl8651_mnQueueTimer();
		/* Verify: Rate Limit Table: Entry 3 */
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= (100*1000*1000 * 20 / 100) >> 3;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 61;
		asic_rl1.t_remainUnit		= 61;
		asic_rl1.token			= (100*1000*1000 * 20 / 100) >> 3;
		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= (100*1000*1000 /100 *80) >> 3;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 61;
		asic_rl1.t_remainUnit		= 61;
		asic_rl1.token			= (100*1000*1000  / 100 * 80) >> 3;
		ASSERT(rtl8651_getAsicRateLimitTable(RTL8651_RATELIMITTBL_SIZE-1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		_rtl8651_mnQueueTimer();
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= (100*1000*1000 * 20 / 100) >> 3;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 61;
		asic_rl1.t_remainUnit		= 61;
		asic_rl1.token			= (100*1000*1000 * 20 / 100) >> 3;
		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= (100*1000*1000 /100 *80) >> 3;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 61;
		asic_rl1.t_remainUnit		= 61;
		asic_rl1.token			= (100*1000*1000  / 100 * 80) >> 3;
		ASSERT(rtl8651_getAsicRateLimitTable(RTL8651_RATELIMITTBL_SIZE-1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);
		


		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.112.2.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToLimitedQueue("vlan", &rule, 70, 200) == SUCCESS);
		
		 _rtl8651_mnQueueTimer();
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= (100*1000*1000 * 20 / 100) >> 3;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 61;
		asic_rl1.t_remainUnit		= 61;
		asic_rl1.token			= (100*1000*1000 * 20 / 100) >> 3;
		ASSERT(rtl8651_getAsicRateLimitTable(0, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= (100*1000*1000  / 100 * 70) >> 3;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 61;
		asic_rl1.t_remainUnit		= 61;
		asic_rl1.token			= (100*1000*1000 / 100 * 70) >> 3;
		ASSERT(rtl8651_getAsicRateLimitTable(1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= (100*1000*1000 /100 *10) >> 3;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 61;
		asic_rl1.t_remainUnit		= 61;
		asic_rl1.token			= (100*1000*1000  / 100 * 10) >> 3;
		ASSERT(rtl8651_getAsicRateLimitTable(RTL8651_RATELIMITTBL_SIZE-1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);



		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.actionType_		= RTL8651_ACL_DROP;
		rule.srcIpAddr_		= 0;
		rule.srcIpAddrMask_	= 0;
		rule.dstIpAddr_		= ntohl(inet_addr("140.112.3.1"));
		rule.dstIpAddrMask_	= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addFlowToLimitedQueue("vlan", &rule, 20) != SUCCESS);
	}
	return SUCCESS;
}


int32 qos_MNQueue_test6(uint32 caseNo)
{
	rtl865x_tblAsicDrv_rateLimitParam_t asic_rl1, asic_rl2;
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan;
	_rtl8651_tblDrvAclRule_t asic_acl1, asic_acl2;
	rtl8651_tblDrvAclRule_t rule;
	int32 cur = 0, round;

	START_TEST_CASE();
	{
		/*==============================================================
		  * Test: M-N Queue add/delete
		  *==============================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) set total bandwidth to 100Mbps */
		ASSERT(rtl8651_setUpstreamBandwidth("wan", 100*1000*1000) == SUCCESS); 

		_rtl8651_mnQueueTimer();
		
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= (100*1000*1000 /100 *100) >> 3;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 61;
		asic_rl1.t_remainUnit		= 61;
		asic_rl1.token			= (100*1000*1000  / 100 * 100) >> 3;
		ASSERT(rtl8651_getAsicRateLimitTable(RTL8651_RATELIMITTBL_SIZE-1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);

		ASSERT(rtl8651_setUpstreamBandwidth("wan", 0) == SUCCESS);

		ASSERT(rtl8651_setUpstreamBandwidth("wan", 100*1000*1000) == SUCCESS); 

		_rtl8651_mnQueueTimer();
		
		bzero(&asic_rl1, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		bzero(&asic_rl2, sizeof(rtl865x_tblAsicDrv_rateLimitParam_t));
		asic_rl1.maxToken		= (100*1000*1000 /100 *100) >> 3;
		asic_rl1.refill_number	= 0;
		asic_rl1.t_intervalUnit		= 61;
		asic_rl1.t_remainUnit		= 61;
		asic_rl1.token			= (100*1000*1000  / 100 * 100) >> 3;
		ASSERT(rtl8651_getAsicRateLimitTable(RTL8651_RATELIMITTBL_SIZE-1, &asic_rl2) == SUCCESS);
		RATELIMIT_ENTRY_CMP(&asic_rl1, &asic_rl2);
	}
	return SUCCESS;
}




















/*
@func int32		| ProtocolBasedNATError	| test if the protocol based NAT related function can detect error.
@rvalue SUCCESS	| test OK
@comm 
This function can test if the protocol based NAT related function can detect error.
*/
int32 ProtocolBasedNATError( uint32 caseNo )
{
	int32 ret, i;
	
	START_TEST_CASE();

	ConfigNaptTestEnvironment();

	/* un-support protocol */
	ret = rtl8651_addProtocolBasedNAT( IPPROTO_ICMP, ntohl(inet_addr("140.113.214.102")), ntohl(inet_addr("192.168.1.4")) );
	ASSERT( ret == TBLDRV_EPBNAT_PROTOCOL_NOT_SUPPORTED );
	ret = rtl8651_addProtocolBasedNAT( IPPROTO_IGMP, ntohl(inet_addr("140.113.214.102")), ntohl(inet_addr("192.168.1.4")) );
	ASSERT( ret == TBLDRV_EPBNAT_PROTOCOL_NOT_SUPPORTED );
	ret = rtl8651_addProtocolBasedNAT( IPPROTO_TCP, ntohl(inet_addr("140.113.214.102")), ntohl(inet_addr("192.168.1.4")) );
	ASSERT( ret == TBLDRV_EPBNAT_PROTOCOL_NOT_SUPPORTED );
	ret = rtl8651_addProtocolBasedNAT( IPPROTO_UDP, ntohl(inet_addr("140.113.214.102")), ntohl(inet_addr("192.168.1.4")) );
	ASSERT( ret == TBLDRV_EPBNAT_PROTOCOL_NOT_SUPPORTED );
	ret = rtl8651_addProtocolBasedNAT( IPPROTO_GRE, ntohl(inet_addr("140.113.214.102")), ntohl(inet_addr("192.168.1.4")) );
	ASSERT( ret == TBLDRV_EPBNAT_PROTOCOL_NOT_SUPPORTED );

	/* test no free entry */
	for( i = 0; i < rtl8651_tblDrvPara.protocolBasedNatSize; i++ )
	{
		ret = rtl8651_addProtocolBasedNAT( IPPROTO_GRE+1+i, ntohl(inet_addr("140.113.214.102")), ntohl(inet_addr("192.168.1.4")) + i );
		ASSERT( ret == SUCCESS );
	}
	ret = rtl8651_addProtocolBasedNAT( IPPROTO_GRE+1+i, ntohl(inet_addr("140.113.214.102")), ntohl(inet_addr("192.168.1.4")) + i );
	ASSERT( ret == TBLDRV_EPBNAT_NO_FREE_ENTRY );

	/* test duplicate entry */
	ret = rtl8651_addProtocolBasedNAT( IPPROTO_GRE+1, ntohl(inet_addr("140.113.214.102")), ntohl(inet_addr("192.168.1.10")) );
	ASSERT( ret == TBLDRV_EPBNAT_ENTRY_EXIST );
	
	/* test delete entry */
	for( i = 0; i < rtl8651_tblDrvPara.protocolBasedNatSize; i++ )
	{
		ret = rtl8651_delProtocolBasedNAT( IPPROTO_GRE+1+i, ntohl(inet_addr("140.113.214.102")), ntohl(inet_addr("192.168.1.4")) + i );
		ASSERT( ret == SUCCESS );
	}
	ret = rtl8651_delProtocolBasedNAT( IPPROTO_GRE+1+i, ntohl(inet_addr("140.113.214.102")), ntohl(inet_addr("192.168.1.4")) + i );
	ASSERT( ret == TBLDRV_EPBNAT_ENTRY_NOT_FOUND );

	return SUCCESS;
}


/*
@func int32		| ProtocolBasedNATRandomTest	| random test for add/del protocol based NAT
@rvalue SUCCESS	| test OK
@comm 
This function can random test for add/del protocol based NAT.
*/
int32 ProtocolBasedNATRandomTest( uint32 caseNo )
{
	int32 ret, i, j;
	int32 loop;
	int32 addedNum, totalNum;
	
	START_TEST_CASE();

	ConfigNaptTestEnvironment();

	srandom( caseNo );

	totalNum = rtl8651_tblDrvPara.protocolBasedNatSize;
	addedNum = 0;

	uint8 protocol[totalNum];
	ipaddr_t ExtIp[totalNum];
	ipaddr_t IntIp[totalNum];

	for( loop = 0; loop < 10000; loop++ )
	{
		//printf( "[loop:%d]\n", loop );
		for( i = 0; i < totalNum; i++ )
		{
regen_protocol:
			protocol[i] = random() % 255;
			switch( protocol[i] )
			{
				case IPPROTO_ICMP:
				case IPPROTO_IGMP:
				case IPPROTO_TCP:
				case IPPROTO_UDP:
				case IPPROTO_GRE:
					goto regen_protocol;
			}
			for( j = 0; j < i; j++ )
			{
				if ( protocol[j] == protocol[i] ) goto regen_protocol;
			}
			
			ExtIp[i] = ntohl( inet_addr("140.113.214.102") );

regen_intip:			
			IntIp[i] = ntohl( inet_addr("192.168.1.1") ) + ( random() % 253 );
			for( j = 0; j < i; j++ )
			{
				if ( IntIp[j] == IntIp[i] ) goto regen_intip;
			}
			
			//printf( "Protocol:%d  ExtIp:0x%08x  IntIp:0x%08x\n", protocol[i], ExtIp[i], IntIp[i] );
		}

		// add
		for( i = 0; i < totalNum; i++ )
		{
			ret = rtl8651_addProtocolBasedNAT( protocol[i], ExtIp[i], IntIp[i] );
			ASSERT( ret == SUCCESS );
		}

		// get inbound and outbound
		for( i = 0; i < totalNum; i++ )
		{
			ipaddr_t ExpectedIp;

			// inbound
			ret = _rtl8651_getInboundProtocolBasedNAT( protocol[i], ExtIp[i], &ExpectedIp );
			ASSERT( ret == SUCCESS );
			//printf( "Expected:0x%08x  Int:0x%08x\n", ExpectedIp, IntIp[i] );
			ASSERT( ExpectedIp == IntIp[i] );

			// outbound
			ret = _rtl8651_getOutboundProtocolBasedNAT( protocol[i], &ExpectedIp, IntIp[i] );
			ASSERT( ret == SUCCESS );
			ASSERT( ExpectedIp == ExtIp[i] );
		}

		// del
		for( i = 0; i < totalNum; i++ )
		{
			ret = rtl8651_delProtocolBasedNAT( protocol[i], ExtIp[i], IntIp[i] );
			ASSERT( ret == SUCCESS );
		}
	}

	return SUCCESS;
}

/*
@func int32 | counter_test1 | Confirm the counter clear funciton and member port control function work properly.
@comm
Configure Rome driver as normal gateway and try whether the clear function and member port control function work well.
1. Force the globle variable to be 865xB mode for extension port available.
2. Configure the member port with some member ports
3. Clear the counter and evaluate the member port setting is the same as pre-configured one
4. Clear each individual counter and evaluate the member port setting is the same as pre-configured one
5. Using Add/Delete function and check the member port is configured correctly
*/

int32 counter_test1(uint32 caseNo)
{
	uint32 cntrIdx, port, i;
	
	
	START_TEST_CASE();
	
	{
		rtl8651_totalExtPortNum=3; //this replaces all RTL8651_EXTPORT_NUMBER defines
		rtl8651_allExtPortMask = 0x7<<RTL8651_MAC_NUMBER; //this replaces all RTL8651_EXTPORTMASK defines
		//Make sure the member port will not changed during CLEAR operation
		WRITE_MEM32(MIB_CONTROL, (1<<24)|(1<<21));
		WRITE_MEM32(MIB_CONTROL_1, (1<<25)|(1<<22));
		WRITE_MEM32(MIB_CONTROL_2, (1<<26)|(1<<23));
		rtl8651_clearAsicCounter();
		ASSERT(READ_MEM32(MIB_CONTROL)==(1<<24)|(1<<21));
		ASSERT(READ_MEM32(MIB_CONTROL_1)==(1<<25)|(1<<22));
		ASSERT(READ_MEM32(MIB_CONTROL_2)==(1<<26)|(1<<23));
		//Make sure the member port will not changed during CLEAR operation
		WRITE_MEM32(MIB_CONTROL, (1<<25)|(1<<22));
		WRITE_MEM32(MIB_CONTROL_1, (1<<26)|(1<<23));
		WRITE_MEM32(MIB_CONTROL_2, (1<<27)|(1<<21));
		rtl8651_clearAsicSpecifiedCounter(0);
		rtl8651_clearAsicSpecifiedCounter(1);
		rtl8651_clearAsicSpecifiedCounter(2);
		ASSERT(READ_MEM32(MIB_CONTROL)==(1<<24)|(1<<21));
		ASSERT(READ_MEM32(MIB_CONTROL_1)==(1<<25)|(1<<22));
		ASSERT(READ_MEM32(MIB_CONTROL_2)==(1<<26)|(1<<23));
		rtl8651_resetAsicCounterMemberPort(0);
		ASSERT(READ_MEM32(MIB_CONTROL)==0);
		ASSERT(READ_MEM32(MIB_CONTROL_1)==(1<<25)|(1<<22));
		ASSERT(READ_MEM32(MIB_CONTROL_2)==(1<<26)|(1<<23));
		WRITE_MEM32(MIB_CONTROL, (1<<25)|(1<<22));
		rtl8651_resetAsicCounterMemberPort(1);
		ASSERT(READ_MEM32(MIB_CONTROL)==(1<<24)|(1<<21));
		ASSERT(READ_MEM32(MIB_CONTROL_1)==0);
		ASSERT(READ_MEM32(MIB_CONTROL_2)==(1<<26)|(1<<23));
		WRITE_MEM32(MIB_CONTROL_1, (1<<26)|(1<<23));
		rtl8651_resetAsicCounterMemberPort(2);
		ASSERT(READ_MEM32(MIB_CONTROL)==(1<<24)|(1<<21));
		ASSERT(READ_MEM32(MIB_CONTROL_1)==(1<<25)|(1<<22));
		ASSERT(READ_MEM32(MIB_CONTROL_2)==0);
		for(cntrIdx=0; cntrIdx<3; cntrIdx++)
			rtl8651_resetAsicCounterMemberPort(cntrIdx);
		for(cntrIdx=0; cntrIdx<3; cntrIdx++)
			for(port=0; port<9; port++) {
				rtl8651_addAsicCounterMemberPort(cntrIdx, port);
				if(port<6) {
					switch(cntrIdx) {
						case 0:
						ASSERT(READ_MEM32(MIB_CONTROL)==(1<<port)<<24);
						ASSERT(READ_MEM32(MIB_CONTROL_1)==0);
						ASSERT(READ_MEM32(MIB_CONTROL_2)==0);
						break;
						case 1:
						ASSERT(READ_MEM32(MIB_CONTROL)==0);
						ASSERT(READ_MEM32(MIB_CONTROL_1)==(1<<port)<<24);
						ASSERT(READ_MEM32(MIB_CONTROL_2)==0);
						break;
						case 2:
						ASSERT(READ_MEM32(MIB_CONTROL)==0);
						ASSERT(READ_MEM32(MIB_CONTROL_1)==0);
						ASSERT(READ_MEM32(MIB_CONTROL_2)==(1<<port)<<24);
						break;
					}
				}
				else {
					switch(cntrIdx) {
						case 0:
						ASSERT(READ_MEM32(MIB_CONTROL)==(1<<(port-6))<<21);
						ASSERT(READ_MEM32(MIB_CONTROL_1)==0);
						ASSERT(READ_MEM32(MIB_CONTROL_2)==0);
						break;
						case 1:
						ASSERT(READ_MEM32(MIB_CONTROL)==0);
						ASSERT(READ_MEM32(MIB_CONTROL_1)==(1<<(port-6))<<21);
						ASSERT(READ_MEM32(MIB_CONTROL_2)==0);
						break;
						case 2:
						ASSERT(READ_MEM32(MIB_CONTROL)==0);
						ASSERT(READ_MEM32(MIB_CONTROL_1)==0);
						ASSERT(READ_MEM32(MIB_CONTROL_2)==(1<<(port-6))<<21);
						break;
					}
				}
				rtl8651_delAsicCounterMemberPort(cntrIdx, port);
				ASSERT(READ_MEM32(MIB_CONTROL)==0);
				ASSERT(READ_MEM32(MIB_CONTROL_1)==0);
				ASSERT(READ_MEM32(MIB_CONTROL_2)==0);
			}
			rtl8651_totalExtPortNum = 0;
	}
	return SUCCESS;
}

/*
@func int32 | pppoeSessionNotExistedTermination_test1 | Testing the active PPPoE tracking mechanism working correctly
@comm
Assumption 1. the entry number is 4.
Configure Rome driver as normal gateway and try whether the clear function and member port control function work well.
1. Add session ID separately and check whether entry is added successfully. First 4 session ID, 0x31, 0x33, 0x45 and 0x19 are stored in new allocated entry
2. Add other session ID 0x58 which make tracking system overflow. Force session 0x33 pushed out.
3. Update session ID 0x58 and make sure all entries keep valid during this update period.
4. Continue to update session ID 0x58 and this time force some entry aged out. They are either removed from head and middle.
5. Using free all function to release all entry and make sure entries are no longer exist.
6. Add session IDs once again and make sure they are active.
7. Using termination function to remove session ID. The removed position are either in the from of list or in the middle of the list.
8. Free the list once again and make sure not entries exist.
*/
int32 pppoeSessionNotExistedTermination_test1(uint32 caseNo) {
	uint16 *retSessionIdPtr, sessionId, idx;
	
	START_TEST_CASE();

	_rtl8651_updatePppoeActiveSession(0x31);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	_rtl8651_updatePppoeActiveSession(0x33);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	_rtl8651_updatePppoeActiveSession(0x45);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	_rtl8651_updatePppoeActiveSession(0x19);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	_rtl8651_updatePppoeActiveSession(0x45);
	_rtl8651_updatePppoeActiveSession(0x31);
	for(idx=0; idx<0xFFFD; idx++) {
	_rtl8651_updatePppoeActiveSession(0x58);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == FALSE);//Should be push out due to longest period not touched
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == TRUE);
	}
	retSessionIdPtr = _rtl8651_getNextPppoeActiveSession(NULL, &sessionId);
	ASSERT(retSessionIdPtr && *retSessionIdPtr == 0x19);
	retSessionIdPtr = _rtl8651_getNextPppoeActiveSession(retSessionIdPtr, &sessionId);
	ASSERT(retSessionIdPtr && *retSessionIdPtr == 0x45);
	retSessionIdPtr = _rtl8651_getNextPppoeActiveSession(retSessionIdPtr, &sessionId);
	ASSERT(retSessionIdPtr && *retSessionIdPtr == 0x58);
	retSessionIdPtr = _rtl8651_getNextPppoeActiveSession(retSessionIdPtr, &sessionId);
	ASSERT(retSessionIdPtr && *retSessionIdPtr == 0x31);
	retSessionIdPtr = _rtl8651_getNextPppoeActiveSession(retSessionIdPtr, &sessionId);
	ASSERT(retSessionIdPtr == NULL);
	_rtl8651_updatePppoeActiveSession(0x58);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == FALSE);//Should be push out due to longest period not touched
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == TRUE);
	_rtl8651_updatePppoeActiveSession(0x58);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == FALSE);//Should be push out due to longest period not touched
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == TRUE);
	_rtl8651_updatePppoeActiveSession(0x58);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == FALSE);//Should be push out due to longest period not touched
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == TRUE);
	_rtl8651_freeAllPppoeActiveSession();
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	_rtl8651_updatePppoeActiveSession(0x31);
	_rtl8651_updatePppoeActiveSession(0x33);
	_rtl8651_updatePppoeActiveSession(0x45);
	_rtl8651_updatePppoeActiveSession(0x19);
	_rtl8651_terminatePppoeActiveSession(0x58);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	_rtl8651_terminatePppoeActiveSession(0x19);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	_rtl8651_terminatePppoeActiveSession(0x33);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	_rtl8651_terminatePppoeActiveSession(0x31);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == TRUE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	_rtl8651_terminatePppoeActiveSession(0x45);
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	_rtl8651_freeAllPppoeActiveSession();
	ASSERT(_rtl8651_pppoeSessionActive(0x31) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x33) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x45) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x19) == FALSE);
	ASSERT(_rtl8651_pppoeSessionActive(0x58) == FALSE);
	retSessionIdPtr = _rtl8651_getNextPppoeActiveSession(NULL, &sessionId);
	ASSERT(retSessionIdPtr == NULL);
	
	return SUCCESS;

//void _rtl8651_terminatePppoeActiveSession(uint16 pppoeSessionId);
//int32 _rtl8651_pppoeSessionActive(uint16 pppoeSessionId);
//uint16 * _rtl8651_getNextPppoeActiveSession(uint16 *curPppoeSessionId, uint16 * nextPppoeSessionId);
}

/*
@func int32 | portBandwidthControl_test1 | Confirm the port bandwidth control function work properly.
@comm
Configure Rome driver as normal gateway and try whether the clear function and member port control function work well.
1. Force the globle variable to be 865xB mode for extension port available.
2. Configure the member port with some member ports
3. Clear the counter and evaluate the member port setting is the same as pre-configured one
4. Clear each individual counter and evaluate the member port setting is the same as pre-configured one
5. Using Add/Delete function and check the member port is configured correctly
*/

int32 portBandwidthControl_test1(uint32 caseNo)
{
	uint32 port, rate, asicRate;
	int8 input, asicX4Enable;
	
	START_TEST_CASE();

	for(port=0; port<RTL8651_PORT_NUMBER; port++) 
		for(input=0; input<2; input++)
			rtl8651_setEthernetPortBandwidthControl(port, input, RTL8651_BC_FULL);

	for(port=0; port<RTL8651_PORT_NUMBER; port++) 
		for(input=0; input<2; input++)
			for(rate=0; rate<RTL8651_BC_32M; rate++) {
				ASSERT(rtl8651_setEthernetPortBandwidthControl(port, input, rate)==SUCCESS);
				rtl8651_getAsicEthernetBandwidthControl(port, input, &asicRate);
				rtl8651_getAsicEthernetBandwidthControlX4(&asicX4Enable);
				if(rate < RTL8651_BC_16M) {
					ASSERT(asicRate==rate);
					ASSERT(asicX4Enable==FALSE);
				}
				else {
					ASSERT(asicRate==(rate-2));
					ASSERT(asicX4Enable==TRUE);
				}
				rtl8651_setEthernetPortBandwidthControl(port, input, RTL8651_BC_FULL);
			}

	ASSERT(rtl8651_setEthernetPortBandwidthControl(0, TRUE, RTL8651_BC_32M)==SUCCESS);
	for(port=0; port<RTL8651_PORT_NUMBER; port++) 
		for(input=0; input<2; input++)
			if(port != 0 && input != TRUE) {
				ASSERT(rtl8651_setEthernetPortBandwidthControl(port, input, RTL8651_BC_128K)!=SUCCESS);
				ASSERT(rtl8651_setEthernetPortBandwidthControl(port, input, RTL8651_BC_256K)!=SUCCESS);
			}
	ASSERT(rtl8651_setEthernetPortBandwidthControl(0, TRUE, RTL8651_BC_16M)==SUCCESS);
	for(port=0; port<RTL8651_PORT_NUMBER; port++) 
		for(input=0; input<2; input++)
			if(port != 0 && input != TRUE) {
				ASSERT(rtl8651_setEthernetPortBandwidthControl(port, input, RTL8651_BC_128K)!=SUCCESS);
				ASSERT(rtl8651_setEthernetPortBandwidthControl(port, input, RTL8651_BC_256K)!=SUCCESS);
			}
	ASSERT(rtl8651_setEthernetPortBandwidthControl(0, TRUE, RTL8651_BC_128K)==SUCCESS);
	for(port=0; port<RTL8651_PORT_NUMBER; port++) 
		for(input=0; input<2; input++)
			if(port != 0 && input != TRUE) {
				ASSERT(rtl8651_setEthernetPortBandwidthControl(port, input, RTL8651_BC_16M)!=SUCCESS);
				ASSERT(rtl8651_setEthernetPortBandwidthControl(port, input, RTL8651_BC_32M)!=SUCCESS);
			}
	ASSERT(rtl8651_setEthernetPortBandwidthControl(0, TRUE, RTL8651_BC_256K)==SUCCESS);
	for(port=0; port<RTL8651_PORT_NUMBER; port++) 
		for(input=0; input<2; input++)
			if(port != 0 && input != TRUE) {
				ASSERT(rtl8651_setEthernetPortBandwidthControl(port, input, RTL8651_BC_16M)!=SUCCESS);
				ASSERT(rtl8651_setEthernetPortBandwidthControl(port, input, RTL8651_BC_32M)!=SUCCESS);
			}

	return SUCCESS;
}






#define RTL865X_IP_TABLE

int32 iptbl_test1(uint32 caseNo)
{
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ipaddr_t naptip;
	int32 round;

	
	START_TEST_CASE();
	{
		/*======================================================
		 * Test: add NAPT IP for VLAN type
		 *======================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		naptip = ntohl(inet_addr("140.113.1.1"));
		for(round=0; round<2000; round++) {
			/* (1) add NAPT IP */
			ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

			/* Verify IP Table */
			bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
			asic_ip.intIpAddr	= 0;
			asic_ip.localPublic	= FALSE;
			asic_ip.nat			= FALSE;
			asic_ip.nhIndex		= 0;
			ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
			IP_ENTRY_CMP(&asic_ip, &asic_ip1);

			/* (2) remove NAPT IP */
			ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);

			/* Verify IP Table */
			ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) != SUCCESS);
		}
	}
	return SUCCESS;
}


int32 iptbl_test2(uint32 caseNo)
{
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x54 } };
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ipaddr_t naptip;
	int32 round;

	
	START_TEST_CASE();
	{
		/*======================================================
		 * Test: add NAPT IP for PPPoE type
		 *======================================================*/
		default_configuration3();

		mac_pppoe.octet[5] ++;
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", naptip, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip, 1000) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("pppoe") == SUCCESS);


		for(round=0; round<2000; round++) {
			/* (1) add NAPT IP */
			ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

			/* Verify IP Table */
			bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
			asic_ip.intIpAddr	= 0;
			asic_ip.localPublic	= FALSE;
			asic_ip.nat			= FALSE;
			asic_ip.nhIndex		= 1;
			ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
			IP_ENTRY_CMP(&asic_ip, &asic_ip1);

			/* (2) remove NAPT IP */
			ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);

			/* Verify IP Table */
			ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) != SUCCESS);
		}
	}
	return SUCCESS;
}


int32 iptbl_test3(uint32 caseNo)
{
	rtl865x_tblAsicDrv_routingParam_t asic_rt_1, asic_rt_2;
	ipaddr_t naptip, ipAddr, ipMask, nexthop;
	int32 round;

	
	START_TEST_CASE();
	{
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		ipAddr = ntohl(inet_addr("140.113.1.2"));
		ipMask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("wan", ipAddr, ipMask) == SUCCESS);
		
		for(round = 0; round < 2000; round++) {		
			/* (1) add NAPT IP */
			naptip = ntohl(inet_addr("140.113.1.1"));		// default NAPT IP
			ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

			naptip = ntohl(inet_addr("140.113.1.2"));		// non-default NAPT IP
			ASSERT(rtl8651_addDriverNaptMapping(naptip) == SUCCESS);

			/* (3) remove NAPT IP */
			naptip = ntohl(inet_addr("140.113.1.1"));		// remove default NAPT IP
			ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);

			naptip = ntohl(inet_addr("140.113.1.2"));		// remove non-default NAPT IP
			ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);	// remove this NAPT IP in software
			//ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);	// remove NAPT in hardware
		}
	}
	return SUCCESS;

}








#define RTL865X_NEXTHOP_TABLE

int32 nexthop_test1(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ipaddr_t naptip, nexthop;
	

	START_TEST_CASE();
	{
		/*======================================================
		  * Test: for NAPT adding, VLAN type
	  	  * 1. lack both default route and arp
	  	  * 2. present default route but no arp
	  	  * 3. present both default route and arp
	  	  *======================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add NAPT mapping */
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 0;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		
		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 0;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(0, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (2) add default route */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addRoute(0, 0, "wan", nexthop) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		
		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);


		/* (3) add arp for the nexthop */
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		
		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

	}
	return SUCCESS;
}


int32 nexthop_test2(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ipaddr_t naptip, nexthop;
	

	START_TEST_CASE();
	{
		/*======================================================
		  * Test: for NAPT adding, VLAN type
	  	  * 1. lack both default route and arp
	  	  * 2. present arp but not default route
	  	  * 3. present both default route and arp
	  	  *======================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add NAPT mapping */
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 0;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		
		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 0;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(0, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (2) add arp for the nexthop */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 0;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		
		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 0;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(0, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);


		/* (3) add default route */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addRoute(0, 0, "wan", nexthop) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		
		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


int32 nexthop_test3(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ipaddr_t naptip, nexthop;
	

	START_TEST_CASE();

	{
		/*=============================================================
		  * Test: for NAPT adding, VLAN type
		  * Remove napt ip
		  *=============================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add NAPT mapping */
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* (2) add default route */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addRoute(0, 0, "wan", nexthop) == SUCCESS);

		/* (3) add arp for the nexthop */
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);

		/* (4) remove NAPT IP address */
		ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);
 
		/* Verify IP Table */
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) != SUCCESS);
		
		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 0;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= 0;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 0;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


int32 nexthop_test4(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ipaddr_t naptip, nexthop;
	int32 round;
	

	START_TEST_CASE();
	{
		/*==================================================
		  * Test: arp remove/add for napt , for NAPT adding, VLAN type
		  *==================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add NAPT mapping */
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* (2) add default route */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addRoute(0, 0, "wan", nexthop) == SUCCESS);

		/* (3) add arp for the nexthop */
		for(round=0; round<5000; round++) {
			ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);
			/* Verify IP Table */
			bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
			asic_ip.intIpAddr	= 0;
			asic_ip.localPublic	= FALSE;
			asic_ip.nat		= FALSE;
			asic_ip.nhIndex	= 1;
			ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
			IP_ENTRY_CMP(&asic_ip, &asic_ip1);
			
			/* Verify nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx		= 0;
			asic_nexthop.isPppoe		= 0;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow	= 1;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

			/* remove arp */
			ASSERT(rtl8651_delArp(nexthop) == SUCCESS);
			/* Verify IP Table */
			bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
			asic_ip.intIpAddr	= 0;
			asic_ip.localPublic	= FALSE;
			asic_ip.nat		= FALSE;
			asic_ip.nhIndex	= 1;
			ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
			IP_ENTRY_CMP(&asic_ip, &asic_ip1);
			
			/* Verify nexthop Table */
			bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
			asic_nexthop.dvid			= 1;
			asic_nexthop.extIntIpIdx		= 0;
			asic_nexthop.isPppoe		= 0;
			asic_nexthop.nextHopColumn	= 0;
			asic_nexthop.nextHopRow	= 5;
			asic_nexthop.pppoeIdx		= 0;
			ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
			NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
			
		}
	}
	return SUCCESS;
}


int32 nexthop_test5(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x54 } };
	ipaddr_t naptip, nexthop;
	

	START_TEST_CASE();
	{
		/*======================================================
		  * Test: for NAPT adding, PPPoE type, old model
	  	  * 1. lack both default route and arp
	  	  * 2. present default route but no arp
	  	  * 3. present both default route and arp
	  	  *======================================================*/
		default_configuration3();
		ASSERT(rtl8651_addExtNetworkInterface("pppoe") == SUCCESS);

		/* (1) set PPPoE Property */
		mac_pppoe.octet[5] ++;
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", naptip, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip, 1000) == SUCCESS);
		
		/* (2) add NAPT mapping */
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 17;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		
		/* (3) reset pppoe property */
		ASSERT(rtl8651_resetPppoeSessionProperty(1000) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (4) set PPPoE property */
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 17;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (5)Remove NAPT IP */
		ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) != SUCCESS);
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		
	}
	return SUCCESS;
}


int32 nexthop_test6(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x54 } };
	ipaddr_t naptip, nexthop;
	

	START_TEST_CASE();
	{
		/*======================================================
		  * Test: for NAPT adding, PPPoE type, new model
	  	  * 1. lack both default route and arp
	  	  * 2. present default route but no arp
	  	  * 3. present both default route and arp
	  	  *======================================================*/
		default_configuration3();
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 1111, &mac_pppoe, 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 2000,  "pppoe1", 1) == SUCCESS);

		ASSERT(rtl8651_addExtNetworkInterface("pppoe1") == SUCCESS);
		mac_pppoe.octet[5] ++;
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(2000, 22222, &mac_pppoe, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe1", naptip, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip, 2000) == SUCCESS);

		/* (1) add NAPT IP address */
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);
		
		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 17;
		asic_nexthop.pppoeIdx		= 1;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


int32 nexthop_test7(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x55 } };
	ipaddr_t naptip, pptp_serv, mask;
	

	START_TEST_CASE();
	{
		/*======================================================
		  * Test: for NAPT adding, PPTP type, new model
	  	  *======================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;
		
		pptp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", pptp_serv, mask) == SUCCESS);
		pptp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(pptp_serv, &mac, "pptp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);		
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 12345, pptp_serv, 0) == SUCCESS);
		naptip = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("pptp", naptip, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip, 1000) == SUCCESS);


		/* (1) add NAPT IP address */
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);
		
		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (2) reset pptp property */
		ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (3) set property */
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 12345, pptp_serv, 0) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


int32 nexthop_test8(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x55 } };
	ipaddr_t naptip, pptp_serv, mask;
	

	START_TEST_CASE();
	{
		/*======================================================
		  * Test: for NAPT adding, L2TP type, new model
	  	  *======================================================*/
		defaultConfig_l2tp1();
		DrvTbl.vlan[0].valid = 1;
		
		pptp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", pptp_serv, mask) == SUCCESS);
		pptp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(pptp_serv, &mac, "l2tp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);		
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 12345, pptp_serv, 0) == SUCCESS);
		naptip = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("l2tp", naptip, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip, 1000) == SUCCESS);


		/* (1) add NAPT IP address */
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);
		
		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (2) reset l2tp property */
		ASSERT(rtl8651_resetL2tpProperty(1000) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (3) set property */
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 12345, pptp_serv, 0) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		
	}
	return SUCCESS;
}








int32 nexthop_test10(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ipaddr_t naptip, nexthop;
	int32 round;
	

	START_TEST_CASE();
	{
		/*=====================================================
		  * Test: add hash 2 entry, VLAN type, use default route's nexthop 
		  *=====================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add default route */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addRoute(0, 0, "wan", nexthop) == SUCCESS);

		/* (2) add arp for the default nexthop */
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);
		
		/* (3) add NAPT IP */
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);
		  
		/* Verify IP Table & Nexthop Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (4) add hash entry */
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_addHash2ToNexthop(0, naptip) == 0);
		}
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_addHash2ToNexthop(nexthop, naptip) == 0);
		}

		/* Verify IP Table & Nexthop Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* The refernece count of the ip entry should be 2001 */
		ASSERT(DrvTbl.iptbl[0].ref_cnt == 2001);
		ASSERT(DrvTbl.iptbl[0].def_flag == 1);

		/* A. remove default route */
		ASSERT(rtl8651_delRoute(0, 0) == SUCCESS);

		/* Verify IP Table & Nexthop Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 0;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* B. add default route */
		ASSERT(rtl8651_addRoute(0, 0 , "wan", nexthop) == SUCCESS);

		/* Verify IP Table & Nexthop Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);


		/* (5) remove it one by one */
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_delHash2FromNexthop(nexthop, naptip) == SUCCESS);
			ASSERT(DrvTbl.iptbl[0].ref_cnt == 2001 - (round+1));
		}
		ASSERT(DrvTbl.iptbl[0].ref_cnt == 1001);
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_delHash2FromNexthop(0, naptip) == SUCCESS);
			ASSERT(DrvTbl.iptbl[0].ref_cnt == 1001 - (round+1));
		}
		ASSERT(DrvTbl.iptbl[0].ref_cnt == 1);

		/* Verify IP Table & Nexthop Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (6) remove the NAPT IP */
		ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);
		
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) != SUCCESS);
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);	
	}
	return SUCCESS;
}


int32 nexthop_test11(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ipaddr_t naptip, nexthop, mask;
	int32 round;
	

	START_TEST_CASE();
	{
		/*=====================================================
		  * Test: add hash 2 entry, VLAN type, use specified nexthop 
		  *=====================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add default route */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addRoute(0, 0, "wan", nexthop) == SUCCESS);

		/* (2) add arp for the default nexthop */
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);
		
		/* (3) add NAPT IP */
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* (4) add the second NAPT IP */
		naptip = ntohl(inet_addr("140.113.1.2"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("wan", naptip, mask) == SUCCESS);
		ASSERT(rtl8651_addNaptMappingExt(naptip) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);  
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.2"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);  

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (5) add Hash 2 entry */
		mac.octet[5] ++;
		nexthop = ntohl(inet_addr("140.113.1.200"));
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_addHash2ToNexthop(nexthop, naptip) == 2);
			ASSERT(DrvTbl.iptbl[2].ref_cnt == (round+1));
			ASSERT(DrvTbl.iptbl[2].def_flag == 0);
		}
		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.2"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(2, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 2;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (6) remove hash2 entry */
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_delHash2FromNexthop(nexthop, naptip) == SUCCESS);
			ASSERT(DrvTbl.iptbl[2].ref_cnt == (1000 - (round+1)));
			ASSERT(DrvTbl.iptbl[2].def_flag == 0);
		}

		/* Verify IP Table */
		ASSERT(rtl8651_getAsicExtIntIpTable(2, &asic_ip1) != SUCCESS);
		
		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(DrvTbl.route[rtl8651_tblDrvPara.routingTableSize -1].rt_nhstart_t->entryIndex == 2);
		ASSERT(DrvTbl.route[rtl8651_tblDrvPara.routingTableSize -1].rt_nhend_t->entryIndex == 3);
	}
	return SUCCESS;
}



int32 nexthop_test12(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x54 } };
	ipaddr_t naptip, nexthop;
	int32 round;
	

	START_TEST_CASE();
	{
		/*=====================================================
		 * Test: add hash 2 entry, PPPoE type 
		 *=====================================================*/
		default_configuration3();
		ASSERT(rtl8651_addExtNetworkInterface("pppoe") == SUCCESS);

		/* (1) set PPPoE Property 1 */
		mac_pppoe.octet[5] ++;
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_setPppoeSessionProperty(1000, 11111, &mac_pppoe, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe", naptip, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip, 1000) == SUCCESS);
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* (2) set PPPoE Property 2 */
		mac_pppoe.octet[5] ++;
		naptip = ntohl(inet_addr("140.113.1.2"));
		ASSERT(rtl8651_addSession(RTL8651_PPPOE, 2000, "pppoe1", 1) == SUCCESS);
		ASSERT(rtl8651_setPppoeSessionProperty(2000, 22222, &mac_pppoe, 0x01) == SUCCESS);
		ASSERT(rtl8651_addIpIntf("pppoe1", naptip, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip, 2000) == SUCCESS);
		ASSERT(rtl8651_addNaptMappingExt(naptip) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 2;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.2"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 18;
		asic_nexthop.pppoeIdx		= 1;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 17;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (3) add Hash 2 entry */
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_addHash2ToNexthop(0, naptip) == 1);
			ASSERT(DrvTbl.iptbl[1].def_flag == 0);
			ASSERT(DrvTbl.iptbl[1].ref_cnt == (round+2));
		}

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 2;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.2"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 18;
		asic_nexthop.pppoeIdx		= 1;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 17;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (3) reset property */
		ASSERT(rtl8651_resetPppoeSessionProperty(2000) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 2;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.2"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 1;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 17;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (4) set property */
		ASSERT(rtl8651_setPppoeSessionProperty(2000, 22222, &mac_pppoe, 0x01) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 2;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.2"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 18;
		asic_nexthop.pppoeIdx		= 1;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= TRUE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 17;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

	}
	return SUCCESS;
}



int32 nexthop_test13(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x54 } };
	ipaddr_t naptip, pptp_serv, mask;
	int32 round;
	

	START_TEST_CASE();
	{
		/*=====================================================
		  * Test: add hash 2 entry, PPTP type
		  *=====================================================*/
		defaultConfig_pptp1();
		DrvTbl.vlan[0].valid = 1;
		
		pptp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("pptp", pptp_serv, mask) == SUCCESS);
		pptp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(pptp_serv, &mac, "pptp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_PPTP, 1000, "pptp", 1) == SUCCESS);		
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 12345, pptp_serv, 0) == SUCCESS);
		naptip = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("pptp", naptip, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip, 1000) == SUCCESS);

		/* (1) add NAPT IP */
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (2) add Hash 2 entry */
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_addHash2ToNexthop(0, naptip) == SUCCESS);
			ASSERT(DrvTbl.iptbl[0].ref_cnt == (round+2));
			ASSERT(DrvTbl.iptbl[0].def_flag == 0);
		}

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (3) reset PPTP property */
		ASSERT(rtl8651_resetPptpProperty(1000) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (4) set PPTP Property */
		ASSERT(rtl8651_setPptpProperty(1000, 11111, 12345, pptp_serv, 0) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		
		/* (5) remove hash 2 */
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_delHash2FromNexthop(0, naptip) == SUCCESS);
			ASSERT(DrvTbl.iptbl[0].ref_cnt == (1001-(round+1)) );
			ASSERT(DrvTbl.iptbl[0].def_flag == 0);
		}

		/* (6) remove NAPT IP */
		ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);

		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) != SUCCESS);
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}


int32 nexthop_test14(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ether_addr_t mac_pppoe = { { 0x00, 0x11, 0x22, 0x33, 0x44, 0x54 } };
	ipaddr_t naptip, l2tp_serv, mask;
	int32 round;
	

	START_TEST_CASE();
	{
		/*=====================================================
		  * Test: add hash 2 entry, L2TP type
		  *=====================================================*/
		defaultConfig_l2tp1();
		DrvTbl.vlan[0].valid = 1;
		
		l2tp_serv = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("l2tp", l2tp_serv, mask) == SUCCESS);
		l2tp_serv = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(l2tp_serv, &mac, "l2tp", 0x01) == SUCCESS);
		ASSERT(rtl8651_addSession(RTL8651_L2TP, 1000, "l2tp", 1) == SUCCESS);		
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 12345, l2tp_serv, 0) == SUCCESS);
		naptip = ntohl(inet_addr("10.1.1.1"));
		ASSERT(rtl8651_addIpIntf("l2tp", naptip, 0xffffffff) == SUCCESS);
		ASSERT(rtl8651_bindSession(naptip, 1000) == SUCCESS);

		/* (1) add NAPT IP */
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (2) add Hash 2 entry */
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_addHash2ToNexthop(0, naptip) == SUCCESS);
			ASSERT(DrvTbl.iptbl[0].ref_cnt == (round+2));
			ASSERT(DrvTbl.iptbl[0].def_flag == 0);
		}

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (3) reset L2TP property */
		ASSERT(rtl8651_resetL2tpProperty(1000) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (4) set L2TP Property */
		ASSERT(rtl8651_setL2tpProperty(1000, 11111, 12345, l2tp_serv, 0) == SUCCESS);

		/* Verify IP Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("10.1.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 3;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		
		/* (5) remove hash 2 */
		for(round=0; round<1000; round++) {
			ASSERT(_rtl8651_delHash2FromNexthop(0, naptip) == SUCCESS);
			ASSERT(DrvTbl.iptbl[0].ref_cnt == (1001-(round+1)) );
			ASSERT(DrvTbl.iptbl[0].def_flag == 0);
		}

		/* (6) remove NAPT IP */
		ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);

		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) != SUCCESS);
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
	}
	return SUCCESS;
}





int32 nexthop_test15(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ipaddr_t naptip, nexthop;
	int32 round;
	

	START_TEST_CASE();
	{
		/*=====================================================
		 * Test: add hash 2 entry, VLAN type
		 *		 when different nexthop, should add a NAPT IP
		 *=====================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add default route */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addRoute(0, 0, "wan", nexthop) == SUCCESS);

		/* (2) add arp for the default nexthop */
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);
		
		/* (3) add NAPT IP */
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);
		  
		/* Verify IP Table & Nexthop Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat		= FALSE;
		asic_ip.nhIndex	= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx		= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow	= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (4) add arp for nexthop */
		mac.octet[5] += 2;
		nexthop = ntohl(inet_addr("140.113.1.200"));
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);
		
		/* (5) use hash2 to add same NAPT IP, but different nexthop */
		ASSERT(_rtl8651_addHash2ToNexthop(nexthop, naptip) != FAILED);

		/* Verify IP Table & Nexthop Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 2;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(1, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);

		/* Verify NextHop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 3;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(1, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (6) overflow table */
		for(round=0; round<14; round++) {
			nexthop = ntohl(inet_addr("140.113.1.200")) + round + 1;
			ASSERT(_rtl8651_addHash2ToNexthop(nexthop, naptip) != FAILED);
		}
		nexthop = ntohl(inet_addr("140.113.1.100")) ;
		ASSERT(_rtl8651_addHash2ToNexthop(nexthop, naptip) == FAILED);

		/* Verify IP Table & Nexthop Table */
		for(round=0; round<16; round++) {
			bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
			asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
			asic_ip.intIpAddr	= 0;
			asic_ip.localPublic	= FALSE;
			asic_ip.nat			= FALSE;
			asic_ip.nhIndex		= 16-round;
			ASSERT(rtl8651_getAsicExtIntIpTable(round, &asic_ip1) == SUCCESS);
			IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		}
	}
	return SUCCESS;
}



int32 nexthop_test16(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ipaddr_t naptip, nexthop, localIp;
	int32 round;
	uint16 extPort;
	

	START_TEST_CASE();
	{
		/*=====================================================
		 * Test: add NAPT IP and hash2 to use the NAPT IP, then
		 *		 remove the NAPT IP, the hash 2 should be
		 *		 removed.
		 *=====================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		
		/* (1) add default route */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addRoute(0, 0, "wan", nexthop) == SUCCESS);

		/* (2) add arp for the default nexthop */
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x01) == SUCCESS);
		
		/* (3) add NAPT IP */
		naptip = ntohl(inet_addr("140.113.1.1"));
		ASSERT(rtl8651_addNaptMapping(naptip) == SUCCESS);

		/* (4) add NAPT IP address use hash 2 */
		localIp = ntohl(inet_addr("192.168.1.2"));
		extPort = 100;
		ASSERT(rtl8651_addNaptConnection(TRUE, UDP_FLOW, localIp, 1000, &naptip, &extPort, nexthop, 10000) == SUCCESS);

		/* Verify IP Table & Nexthop Table */
		bzero(&asic_ip, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		bzero(&asic_ip1, sizeof(rtl865x_tblAsicDrv_extIntIpParam_t));
		asic_ip.extIpAddr	= ntohl(inet_addr("140.113.1.1"));
		asic_ip.intIpAddr	= 0;
		asic_ip.localPublic	= FALSE;
		asic_ip.nat			= FALSE;
		asic_ip.nhIndex		= 1;
		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) == SUCCESS);
		IP_ENTRY_CMP(&asic_ip, &asic_ip1);
		
		/* (5) remove NAPT IP */
		ASSERT(rtl8651_delNaptMapping(naptip) == SUCCESS);

		ASSERT(rtl8651_getAsicExtIntIpTable(0, &asic_ip1) != SUCCESS);
		
	}
	return SUCCESS;
}


int32 nexthop_test17(uint32 caseNo)
{
	rtl865x_tblAsicDrv_nextHopParam_t asic_nexthop, asic_nexthop1;
	rtl865x_tblAsicDrv_extIntIpParam_t asic_ip, asic_ip1;
	ether_addr_t mac = { { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05 } };
	ipaddr_t naptip, nexthop, localIp;
	int32 round;
	uint16 extPort;
	

	START_TEST_CASE();
	{
		/*====================================================================
		 * Test: Arp add and delete
		 *====================================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* (1) add arp */
		nexthop = ntohl(inet_addr("140.113.1.254"));
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x1) == SUCCESS);
		
		/* (2) add default route */
		ASSERT(rtl8651_addRoute(0, 0, "wan", nexthop) == SUCCESS);

		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (3) remove arp */
		ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (4) add Arp again */
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x1) == SUCCESS);

		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (5) add L2 entry */
		ASSERT(rtl8651_addFilterDatabaseEntry(0, &mac, RTL8651_FORWARD_MAC, 0x02) == SUCCESS);

		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (6) remove arp */
		ASSERT(rtl8651_delArp(nexthop) == SUCCESS);

		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 5;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (7) add Arp again */
		ASSERT(rtl8651_addArp(nexthop, &mac, "wan", 0x1) == SUCCESS);

		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);

		/* (2) remove L2 entry */
		ASSERT(rtl8651_delFilterDatabaseEntry(0, &mac) == SUCCESS);

		/* Verify nexthop Table */
		bzero(&asic_nexthop, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		bzero(&asic_nexthop1, sizeof(rtl865x_tblAsicDrv_nextHopParam_t));
		asic_nexthop.dvid			= 1;
		asic_nexthop.extIntIpIdx	= 0;
		asic_nexthop.isPppoe		= FALSE;
		asic_nexthop.nextHopColumn	= 0;
		asic_nexthop.nextHopRow		= 1;
		asic_nexthop.pppoeIdx		= 0;
		ASSERT(rtl8651_getAsicNextHopTable(2, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		ASSERT(rtl8651_getAsicNextHopTable(3, &asic_nexthop1) == SUCCESS);
		NXTHOP_ENTRY_CMP(&asic_nexthop, &asic_nexthop1);
		
	}
	return SUCCESS;
}




#define RTL865X_PSEUDO_VLAN


int32 pseudo_vlan_test1(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x02 } };
	ipaddr_t ipaddr, mask;

	
	START_TEST_CASE();
	{
		/*==================================================
		 * Test: 
		 *==================================================*/
		ASSERT(rtl8651_delVlanPortMember(1, 0) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 1) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 2) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 3) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 4) == SUCCESS);
		 
		/* WAN: subnet 1: */	
		ASSERT(rtl8651_addVlanPortMember(1, 0) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("wan") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("wan", RTL8651_LL_VLAN, 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("wan", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("wan") == SUCCESS);

		/* VLAN: subnet 2 */
		ASSERT(rtl8651_addSubVlan(5, 3, 0x03) != SUCCESS);
		mac.octet[5] = 0x02;
		ASSERT(rtl8651_addVlan(2) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 1) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 2) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 3) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 4) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addSubVlan(2, 3, 0x03) != SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan", RTL8651_LL_VLAN, 2) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("vlan", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_addSubVlan(2, 3, 0x06) == SUCCESS);

	}
	return SUCCESS;
}


int32 pseudo_vlan_test2(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan1, asic_vlan2;
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x02 } };
	int32 round;
	
	
	START_TEST_CASE();
	{
		/*========================================================================
		 * Test: The pseudo vlan's parent vlan should be successfully configured,
		 *		 then we can create a pseudo vlan.
		 *========================================================================*/
		ASSERT(rtl8651_addSubVlan(2, 3, 0x02) != SUCCESS);
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* handle member port */
		ASSERT(rtl8651_addVlanPortMember(2, 1) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 2) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 3) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 4) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 5) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 1) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 2) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 3) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 4) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 5) == SUCCESS);

		/* set PVID */
		ASSERT(rtl8651_setPvid(1, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(2, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(3, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(4, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(5, 2) == SUCCESS);
		
		
		/* (1) add sub vlan 3 */
		ASSERT(rtl8651_addSubVlan(2, 3, 0x02) == SUCCESS);
		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 3);
		ASSERT(rtl8651_getPvid(2) == 2);
		ASSERT(rtl8651_getPvid(3) == 2);
		ASSERT(rtl8651_getPvid(4) == 2);
		ASSERT(rtl8651_getPvid(5) == 2);

		/* (2) add sub vlan 4 */
		ASSERT(rtl8651_addSubVlan(2, 4, 0x04) == SUCCESS);
		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 3);
		ASSERT(rtl8651_getPvid(2) == 4);
		ASSERT(rtl8651_getPvid(3) == 2);
		ASSERT(rtl8651_getPvid(4) == 2);
		ASSERT(rtl8651_getPvid(5) == 2);

		/* (3) add sub vlan 5 */
		ASSERT(rtl8651_addSubVlan(2, 5, 0x08) == SUCCESS);
		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 3);
		ASSERT(rtl8651_getPvid(2) == 4);
		ASSERT(rtl8651_getPvid(3) == 5);
		ASSERT(rtl8651_getPvid(4) == 2);
		ASSERT(rtl8651_getPvid(5) == 2);

		/* (4) add sub vlan 6 */
		ASSERT(rtl8651_addSubVlan(2, 6, 0x10) == SUCCESS);
		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 3);
		ASSERT(rtl8651_getPvid(2) == 4);
		ASSERT(rtl8651_getPvid(3) == 5);
		ASSERT(rtl8651_getPvid(4) == 6);
		ASSERT(rtl8651_getPvid(5) == 2);

		/* (5) add sub vlan 7 */
		ASSERT(rtl8651_addSubVlan(2, 7, 0x20) == SUCCESS);
		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 3);
		ASSERT(rtl8651_getPvid(2) == 4);
		ASSERT(rtl8651_getPvid(3) == 5);
		ASSERT(rtl8651_getPvid(4) == 6);
		ASSERT(rtl8651_getPvid(5) == 7);
		
		/* Verify VLAN */
		for(round=2; round<8; round++) {
			bzero(&asic_vlan1, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
			bzero(&asic_vlan2, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
			asic_vlan1.broadcastToCpu		= TRUE;
			asic_vlan1.DMZFlag				= FALSE;
			asic_vlan1.enableRoute			= TRUE;
			//asic_vlan1.inAclEnd				= RTL8651_ACLTBL_PERMIT_ALL;
			//asic_vlan1.inAclStart			= RTL8651_ACLTBL_PERMIT_ALL;

			/* Index 0 & 1 of ACL table is used by VLAN 1 in current _rtl8651_arrangeAllAcl(). */
			asic_vlan1.inAclEnd			= 3;
			asic_vlan1.inAclStart			= 2;

			asic_vlan1.internal				= 1;
			asic_vlan1.macAddr				= mac;
			asic_vlan1.macAddrNumber		= 1;
			asic_vlan1.memberPortMask		= (round==2)? 0x3e: (1<<(round-2));
			asic_vlan1.mtu					= 1500;
			asic_vlan1.outAclEnd			= RTL8651_ACLTBL_PERMIT_ALL;
			asic_vlan1.outAclStart			= RTL8651_ACLTBL_PERMIT_ALL;
			asic_vlan1.portState[0]			= RTL8651_PORTSTA_FORWARDING;
			asic_vlan1.portState[1]			= RTL8651_PORTSTA_FORWARDING; 
			asic_vlan1.portState[2]			= RTL8651_PORTSTA_FORWARDING;
			asic_vlan1.portState[3]			= RTL8651_PORTSTA_FORWARDING;
			asic_vlan1.portState[4]			= RTL8651_PORTSTA_FORWARDING;
			asic_vlan1.portState[5]			= RTL8651_PORTSTA_FORWARDING;
			asic_vlan1.portState[6]			= RTL8651_PORTSTA_FORWARDING;
			asic_vlan1.portState[7]			= RTL8651_PORTSTA_FORWARDING;
			asic_vlan1.portState[8]			= RTL8651_PORTSTA_FORWARDING;
			asic_vlan1.promiscuous			= FALSE;
			asic_vlan1.untagPortMask		= (round==2)? 0x3e: (1<<(round-2));
			asic_vlan1.valid				= 1;
			asic_vlan1.vid					= round;
			ASSERT(rtl8651_getAsicVlan(round, &asic_vlan2) == SUCCESS);

			VLAN_ENTRY_CMP(&asic_vlan1, &asic_vlan2);
		}

		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);
		for(round=1; round<RTL8651_PORT_NUMBER; round++) {		
			ASSERT(rtl8651_getPvid(round) == (round+2));
		}
	}
	return SUCCESS;
}



int32 pseudo_vlan_test3(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan1;
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x02 } };
	int32 round;
	
	
	START_TEST_CASE();
	{
		/*========================================================================
		 * Test: Rmove Sub VLAN
		 *========================================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);

		/* handle member port */
		ASSERT(rtl8651_addVlanPortMember(2, 1) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 2) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 3) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 4) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 5) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 1) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 2) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 3) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 4) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 5) == SUCCESS);

		/* set PVID */
		ASSERT(rtl8651_setPvid(1, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(2, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(3, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(4, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(5, 2) == SUCCESS);
		
		
		/* (1) add sub vlan */
		ASSERT(rtl8651_addSubVlan(2, 3, 0x02) == SUCCESS);
		ASSERT(rtl8651_addSubVlan(2, 4, 0x04) == SUCCESS);
		ASSERT(rtl8651_addSubVlan(2, 5, 0x08) == SUCCESS);
		ASSERT(rtl8651_addSubVlan(2, 6, 0x10) == SUCCESS);
		ASSERT(rtl8651_addSubVlan(2, 7, 0x20) == SUCCESS);
		
		/* (2) remove sub VLAN: 3 */
		ASSERT(rtl8651_delSubVlan(3) == SUCCESS);
		ASSERT(rtl8651_getAsicVlan(3, &asic_vlan1) != SUCCESS);

		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 2);
		ASSERT(rtl8651_getPvid(2) == 4);
		ASSERT(rtl8651_getPvid(3) == 5);
		ASSERT(rtl8651_getPvid(4) == 6);
		ASSERT(rtl8651_getPvid(5) == 7);

		/* (3) remove sub VLAN: 4 */
		ASSERT(rtl8651_delSubVlan(4) == SUCCESS);
		ASSERT(rtl8651_getAsicVlan(4, &asic_vlan1) != SUCCESS);

		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 2);
		ASSERT(rtl8651_getPvid(2) == 2);
		ASSERT(rtl8651_getPvid(3) == 5);
		ASSERT(rtl8651_getPvid(4) == 6);
		ASSERT(rtl8651_getPvid(5) == 7);

		/* (4) remove sub VLAN: 5 */
		ASSERT(rtl8651_delSubVlan(5) == SUCCESS);
		ASSERT(rtl8651_getAsicVlan(5, &asic_vlan1) != SUCCESS);

		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 2);
		ASSERT(rtl8651_getPvid(2) == 2);
		ASSERT(rtl8651_getPvid(3) == 2);
		ASSERT(rtl8651_getPvid(4) == 6);
		ASSERT(rtl8651_getPvid(5) == 7);

		/* (5) remove sub VLAN: 6 */
		ASSERT(rtl8651_delSubVlan(6) == SUCCESS);
		ASSERT(rtl8651_getAsicVlan(6, &asic_vlan1) != SUCCESS);

		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 2);
		ASSERT(rtl8651_getPvid(2) == 2);
		ASSERT(rtl8651_getPvid(3) == 2);
		ASSERT(rtl8651_getPvid(4) == 2);
		ASSERT(rtl8651_getPvid(5) == 7);

		/* (2) remove sub VLAN: 6 */
		ASSERT(rtl8651_delSubVlan(7) == SUCCESS);
		ASSERT(rtl8651_getAsicVlan(7, &asic_vlan1) != SUCCESS);

		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 2);
		ASSERT(rtl8651_getPvid(2) == 2);
		ASSERT(rtl8651_getPvid(3) == 2);
		ASSERT(rtl8651_getPvid(4) == 2);
		ASSERT(rtl8651_getPvid(5) == 2);
	}
	return SUCCESS;
}


int32 pseudo_vlan_test4(uint32 caseNo)
{

	START_TEST_CASE();
	{
		/*=============================================================
		 * Test: Can not remove sub vlan by rtl8651_delVlan() API.
		 *=============================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		
		/* handle member port */
		ASSERT(rtl8651_addVlanPortMember(2, 1) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 2) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 3) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 4) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 5) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 1) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 2) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 3) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 4) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 5) == SUCCESS);

		/* set PVID */
		ASSERT(rtl8651_setPvid(1, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(2, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(3, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(4, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(5, 2) == SUCCESS);

		/* (1) add sub vlan */
		ASSERT(rtl8651_addSubVlan(2, 3, 0x02) == SUCCESS);

		/* (2) try to remove the sub vlan by rtl8651_delVlan() API */
		ASSERT(rtl8651_delVlan(3) != SUCCESS);
	}
	return SUCCESS;
}


int32 pseudo_vlan_test5(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan1, asic_vlan2;
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x02 } };
	
	START_TEST_CASE();
	{
		/*=====================================================================
		 * Test: more than one ports are added to the same sub vlan
		 *=====================================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		
		/* handle member port */
		ASSERT(rtl8651_addVlanPortMember(2, 1) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 2) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 3) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 4) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 5) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 1) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 2) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 3) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 4) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 5) == SUCCESS);

		/* set PVID */
		ASSERT(rtl8651_setPvid(1, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(2, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(3, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(4, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(5, 2) == SUCCESS);

		/* (1) add port 1, 2, 3 to sub vlan 3 */
		ASSERT(rtl8651_addSubVlan(2, 3, 0x0e) == SUCCESS);

		/* Verify VLAN */
		bzero(&asic_vlan1, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		bzero(&asic_vlan2, sizeof(rtl865x_tblAsicDrv_vlanParam_t));
		asic_vlan1.broadcastToCpu		= TRUE;
		asic_vlan1.DMZFlag				= FALSE;
		asic_vlan1.enableRoute			= TRUE;
//		asic_vlan1.inAclEnd				= RTL8651_ACLTBL_PERMIT_ALL;
//		asic_vlan1.inAclStart			= RTL8651_ACLTBL_PERMIT_ALL;

		/* Index 0 & 1 of ACL table is used by VLAN 1 in current _rtl8651_arrangeAllAcl(). */
		asic_vlan1.inAclEnd			= 3;
		asic_vlan1.inAclStart			= 2;

		asic_vlan1.internal				= 1;
		asic_vlan1.macAddr				= mac;
		asic_vlan1.macAddrNumber		= 1;
		asic_vlan1.memberPortMask		= 0x0e;
		asic_vlan1.mtu					= 1500;
		asic_vlan1.outAclEnd			= RTL8651_ACLTBL_PERMIT_ALL;
		asic_vlan1.outAclStart			= RTL8651_ACLTBL_PERMIT_ALL;
		asic_vlan1.portState[0]			= RTL8651_PORTSTA_FORWARDING;
		asic_vlan1.portState[1]			= RTL8651_PORTSTA_FORWARDING; 
		asic_vlan1.portState[2]			= RTL8651_PORTSTA_FORWARDING;
		asic_vlan1.portState[3]			= RTL8651_PORTSTA_FORWARDING;
		asic_vlan1.portState[4]			= RTL8651_PORTSTA_FORWARDING;
		asic_vlan1.portState[5]			= RTL8651_PORTSTA_FORWARDING;
		asic_vlan1.portState[6]			= RTL8651_PORTSTA_FORWARDING;
		asic_vlan1.portState[7]			= RTL8651_PORTSTA_FORWARDING;
		asic_vlan1.portState[8]			= RTL8651_PORTSTA_FORWARDING;
		asic_vlan1.promiscuous			= FALSE;
		asic_vlan1.untagPortMask		= 0x0e;
		asic_vlan1.valid				= 1;
		asic_vlan1.vid					= 3;
		ASSERT(rtl8651_getAsicVlan(3, &asic_vlan2) == SUCCESS);
		VLAN_ENTRY_CMP(&asic_vlan1, &asic_vlan2);

		/* Verify PVID */
		ASSERT(rtl8651_getPvid(0) == 1);		
		ASSERT(rtl8651_getPvid(1) == 3);
		ASSERT(rtl8651_getPvid(2) == 3);
		ASSERT(rtl8651_getPvid(3) == 3);
		ASSERT(rtl8651_getPvid(4) == 2);
		ASSERT(rtl8651_getPvid(5) == 2);
	}
	return SUCCESS;
}


int32 pseudo_vlan_test6(uint32 caseNo)
{
	rtl865x_tblAsicDrv_vlanParam_t asic_vlan1, asic_vlan2;
	ether_addr_t mac1 = { { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } };
	rtl8651_tblDrvAclRule_t rule;
	
	START_TEST_CASE();
	{
		/*=====================================================================
		 * Test: more than one ports are added to the same sub vlan
		 *=====================================================================*/
		default_configuration2(RTL8651_LL_VLAN, RTL8651_LL_VLAN);
		
		/* handle member port */
		ASSERT(rtl8651_addVlanPortMember(2, 1) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 2) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 3) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 4) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 5) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 1) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 2) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 3) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 4) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 5) == SUCCESS);

		/* set PVID */
		ASSERT(rtl8651_setPvid(1, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(2, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(3, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(4, 2) == SUCCESS);
		ASSERT(rtl8651_setPvid(5, 2) == SUCCESS);

		/* (1) add Sub Vlan */
		ASSERT(rtl8651_addSubVlan(2, 3, 0x02) == SUCCESS);

		/* Compare ACL index bewteen Sub VLAN and VLAN 2 */
		ASSERT(rtl8651_getAsicVlan(3, &asic_vlan1) == SUCCESS);
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan2) == SUCCESS);
		ASSERT(asic_vlan1.inAclStart == asic_vlan2.inAclStart);
		ASSERT(asic_vlan1.inAclEnd == asic_vlan2.inAclEnd);
		ASSERT(asic_vlan1.outAclStart == asic_vlan2.outAclStart);
		ASSERT(asic_vlan1.outAclEnd == asic_vlan2.outAclEnd);

		/* (2) add ACL rule VLAN 2 */
		bzero(&rule, sizeof(rtl8651_tblDrvAclRule_t));
		rule.actionType_		= RTL8651_ACL_CPU;
		rule.ruleType_			= RTL8651_ACL_IP;
		rule.dstIpAddr_			= ntohl(inet_addr("120.113.1.0"));
		rule.dstIpAddrMask_		= ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addAclRule("vlan", TRUE, &rule) == SUCCESS);

		/* Compare ACL index bewteen Sub VLAN and VLAN 2 */
		ASSERT(rtl8651_getAsicVlan(3, &asic_vlan1) == SUCCESS);
		ASSERT(rtl8651_getAsicVlan(2, &asic_vlan2) == SUCCESS);
		ASSERT(asic_vlan1.inAclStart == asic_vlan2.inAclStart);
		ASSERT(asic_vlan1.inAclEnd == asic_vlan2.inAclEnd);
		ASSERT(asic_vlan1.outAclStart == asic_vlan2.outAclStart);
		ASSERT(asic_vlan1.outAclEnd == asic_vlan2.outAclEnd);


	}
	return SUCCESS;
}


int32 pseudo_vlan_test7(uint32 caseNo)
{
	ether_addr_t mac = { { 0x00, 0x03, 0x04, 0x05, 0x06, 0x02 } };
	ipaddr_t ipaddr, mask;

	
	START_TEST_CASE();
	{
		/*==================================================
		 * Test: 
		 *==================================================*/
		ASSERT(rtl8651_delVlanPortMember(1, 0) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 1) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 2) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 3) == SUCCESS);
		ASSERT(rtl8651_delVlanPortMember(1, 4) == SUCCESS);
		 
		/* WAN: subnet 1: */	
		ASSERT(rtl8651_addVlanPortMember(1, 0) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(1, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("wan") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("wan", RTL8651_LL_VLAN, 1) == SUCCESS);
		ipaddr = ntohl(inet_addr("140.113.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("wan", ipaddr, mask) == SUCCESS);
		ASSERT(rtl8651_addExtNetworkInterface("wan") == SUCCESS);

		/* VLAN: subnet 2 */
		ASSERT(rtl8651_addSubVlan(5, 3, 0x03) != SUCCESS);
		mac.octet[5] = 0x02;
		ASSERT(rtl8651_addVlan(2) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 1) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 2) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 3) == SUCCESS);
		ASSERT(rtl8651_addVlanPortMember(2, 4) == SUCCESS);
		ASSERT(rtl8651_assignVlanMacAddress(2, &mac, 1) == SUCCESS);
		ASSERT(rtl8651_addSubVlan(2, 3, 0x03) != SUCCESS);
		ASSERT(rtl8651_addNetworkIntf("vlan") == SUCCESS);
		ASSERT(rtl8651_specifyNetworkIntfLinkLayerType("vlan", RTL8651_LL_VLAN, 2) == SUCCESS);
		ipaddr = ntohl(inet_addr("192.168.1.1"));
		mask = ntohl(inet_addr("255.255.255.0"));
		ASSERT(rtl8651_addIpIntf("vlan", ipaddr, mask) == SUCCESS);
		
		ASSERT(rtl8651_addSubVlan(2, 3, 0x03) != SUCCESS);
		ASSERT(rtl8651_addSubVlan(2, 3, 0x05) != SUCCESS);
		ASSERT(rtl8651_addSubVlan(2, 3, 0x07) != SUCCESS);
		ASSERT(rtl8651_addSubVlan(2, 3, 0x06) == SUCCESS);
	}
	return SUCCESS;
}


